/**
 * 3DView
 *
 * @author KADA
 * @since 9.2.2015
 */
/**
 * 3DView
 *
 * @author KADA
 * @since 9.2.2015
 */
import _ from "underscore";

import $ from "jquery";
import WebViewer3DAPIEXT from "@esko/3d-web-viewer/WebViewer3DAPIEXT";
import View3DView from "./3DViewerLauncher-view";
/**
 * @param {JQueryElement} element - element where the viewer should be shown
 * @param {String} [docVersionId] - docVersion ID for which the 3D view should be shown. if not defined, it is taken from data-docVersionId
 * @param {Function} [callback] - called when 3D has loaded. First parameter indicates succes. Second the error if there is any error.
 * @param {Object} [position] - the initial camera position. See setCameraPosition.
 */
var Viewer3DLauncher = function(element, docVersionId, callback, position) {
	// private components
	this._View3DView = new View3DView(element);

	// Initialize the viewer.
	try {
		this._viewer = new WebViewer3DAPIEXT(this._View3DView.getViewerJQElement().get(0), this.processLoadingFeedback.bind(this));

		// Set Transparency Level (between 1 and 10, recommended default = 7)
		// and Redraw flag (recommended default = true)
		this._viewer.SetQualityLevel(7, true);

		// Set Interaction Optimization Flag (recommended default = true)
		this._viewer.SetInteractionOptimization(true);

		if (!docVersionId) {
			this._docVersionId = this._View3DView.getJQElement().attr("data-docVersionId");
		} else {
			this._docVersionId = docVersionId;
		}

		//turn off reflections as it is known to cause performance issues on FF
		this._viewer.SetReflections("OFF");

		// A lot of display issues are solve by already resizing here.
		this._viewer.Resize();

		//furthermore need to register a 'window resize' callback for resizing the viewport
		//NOTE: _.throttle is used to prevent calling viewer.Resize() on every single window resize event
		var self = this;
		$(window).on(
			"resize",
			_.throttle(function() {
				self._viewer.Resize();
			}, 40)
		);

		this.load3DForDocVersion(this._docVersionId, callback, position);
	} catch (error) {
		this._View3DView.displayErrorMessageLangTag("COULD_NOT_LOAD_VIEWER_WEBGL", function(e) {
			window.open("https://wiki.esko.com/pages/viewpage.action?pageId=170058865");
		});
		if (window.console && window.console.error) {
			window.console.error(error);
		}
	}
};

/**
 * Load 3D for given document version id
 * @param {String} docVersionId - docVersion ID for which the 3D view should be shown.
 * @param {Function} [callback] - called when 3D has loaded. First parameter indicates succes. Second the error if there is any error.
 * @param {Object} [position] - the initial camera position. See setCameraPosition.
 */
Viewer3DLauncher.prototype.load3DForDocVersion = function(docVersionId, callback, position) {
	throw "Not Implemented: Viewer3DLauncher.prototype.load3DForDocVersion";
};

/**
 * Get the current camera position if the 3D viewer is loaded or null otherwise.
 * @returns {Object} @example
 *                   {scale: 1.5,
 * 					 distance: 200,
 *                   azimuth: 46.6,
 *                   elevation: -52,
 *                   panx: 24.0,
 *                   pany: -12.8,
 *                   environment: {_x: .268, _y: -.2157, _z: .684, _w: .5498}}
 */
Viewer3DLauncher.prototype.getCameraPosition = function() {
	if (this._viewer) {
		var angles = this._viewer.GetCameraAngles();
		var scale = this._viewer.GetCameraScale();
		var offSet = this._viewer.GetCameraPanOffset();
		var envRot = this._viewer.GetEnvironmentRotation();
		var dist = this._viewer.GetCameraDistance();
		if (angles !== undefined && scale !== undefined && angles.azimuth !== undefined && angles.elevation !== undefined) {
			var position = {
				scale: scale,
				azimuth: angles.azimuth,
				elevation: angles.elevation,
				panx: offSet.x,
				pany: offSet.y,
				environment: envRot,
				distance: dist
			};
			return position;
		}
	}
	return null;
};

/**
 * Set the current camera position of the 3D widget if the 3D viewer is loaded
 * @param {Object} position -
 *                 @example
 *                    {scale: 1.5,
 * 					  distance: 200,
 *                    azimuth: 46.6,
 *                    elevation: -52,
 *                    panx: 24.0,
 *                    pany: -12.8,
 *                    environment: {_x: .268, _y: -.2157, _z: .684, _w: .5498}}
 */
Viewer3DLauncher.prototype.setCameraPosition = function(position) {
	if (this._viewer && position) {
		const actions = [];
		if (position.scale != null) {
			actions.push(redraw => {
				this._viewer.SetCameraScale(position.scale, redraw);
			});
		}
		if (position.azimuth != null && position.elevation != null) {
			actions.push(redraw => {
				this._viewer.SetCameraAngles(position.azimuth, position.elevation, redraw);
			});
		}
		if (position.panx != null && position.pany != null) {
			actions.push(redraw => {
				this._viewer.SetCameraPanOffset(position.panx, position.pany, redraw);
			});
		}
		if (position.environment != null) {
			actions.push(redraw => {
				this._viewer.SetEnvironmentRotation(position.environment, redraw);
			});
		}
		if (position.distance != null) {
			actions.push(redraw => {
				this._viewer.SetCameraDistance(position.distance, redraw);
			});
		}

		actions.forEach((action, i) => {
			action(i === actions.length - 1);
		});
	}
};

/**
 * Resize the viewer.
 */
Viewer3DLauncher.prototype.resize = function() {
	this._viewer.Resize();
};

/**
 *
 * @param arg0
 * @param arg1
 * @param arg2
 * @param arg3
 * @param arg4
 */
Viewer3DLauncher.prototype.processLoadingFeedback = function(arg0, arg1, arg2, arg3, arg4) {
	// When we get a progress update, postpone the automatically hiding of the progress bar.
	if (this._autoHideProgressAction) {
		clearTimeout(this._autoHideProgressAction);
	}

	// Check for error ...
	if (arg0 === 0) {
		// No error, findout progress ...
		var geometryLoadedP = arg1;
		var textureLoadedP = arg2;
		var numFilesToQuery = arg3;
		var numQueriedFiles = arg4;

		// Check if loading is done ...
		if (geometryLoadedP >= 100 && textureLoadedP >= 100 && numQueriedFiles >= numFilesToQuery) {
			// hide loading message.
			this._View3DView.hideAnyMessage();
		} else {
			// Determine percentage of files loaded.
			var fileLoadingP = (numQueriedFiles / numFilesToQuery) * 100;

			// Create global percentage number for these three criteria.
			var totalP = (geometryLoadedP + textureLoadedP + fileLoadingP) / 3;

			// Round this percentage to ...
			var rounded = parseFloat(Math.round(totalP * 10) / 10).toFixed(1);

			// Display progress...
			this._View3DView.displayLoadingMessage("(" + rounded + "%)");

			// Set that the progress update message should be hidden in three seconds.
			// This will get cancelled when this function, processLoadingFeedback, is called again within this method.
			// This way we avoid the scenario where the progress updater will keep to be shown when no new call this
			// method will be made anymore from the 3d viewer (which appears to happen on FF).
			this._autoHideProgressAction = setTimeout(
				function() {
					this._View3DView.hideAnyMessage();
				}.bind(this),
				5000
			);
		}
	} else {
		// Display progress...
		this._View3DView.displayErrorMessage("Error occured while loading 3D model. (code:" + arg0 + ")");
	}
};

/**
 * Sets the a callback function that will be fired whenever the user stops interacting with the model
 *
 * @param {Function} callback
 */
Viewer3DLauncher.prototype.setEndCameraInteractionCallback = function(callback) {
	this._viewer.SetEndCameraInteractionCallback(callback);
};

/**
 * Get a snapshot of the current viewer state as specified in format
 *
 * @function GetSnapshot
 * @param {string} [format = "png"] - image format (accepted: "png, jpeg")
 * @param {number} [width = context canvas width or 1] - image width in pixels (accepted: [1:2048] higher depend on graphics card and browser)
 * @param {number} [height = context canvas height or 1] - image height in pixels (accepted: [1:2048] higher depend on graphics card and browser)
 * @returns {Object} image data
 */
Viewer3DLauncher.prototype.getSnapShot = function(format, width, height) {
	return this._viewer.GetSnapshot(format, width, height);
};

/**
 * Get a jQuery wrapped reference to the encapsulating div of this component.
 *
 * @returns {jQuery} The jQuery wrapped reference to the encapsulating div of this component.
 */
Viewer3DLauncher.prototype.getJQElement = function() {
	return this._View3DView.getJQElement();
};

/**
 * Method to automatically apply this module to an input or select element ...
 */
Viewer3DLauncher.wcrinit = function(element) {
	new Viewer3DLauncher(element);
};

/**
 * Method to automatically destroy this component once initialized for a div...
 *
 * @param div {DomElement} the DomElement to destroy again
 */
Viewer3DLauncher.wcrdestroy = function(element) {};

export default Viewer3DLauncher;
