/**
 * WebViewer3D provides an HTML5-based 3D viewing component
 *
 * @module WebViewer3D
 * @author DOGO, PHST
 * @contributor ADCL
 * @copyright 2014 Esko Stonecube Ltd. All rights reserved.
 * @version 1.0.x
 */

define(["@esko/3d-web-viewer/modules/ViewerController", "@esko/3d-web-viewer/modules/Constants", "@esko/3d-web-viewer/modules/LogMessage", "@esko/3d-web-viewer/BuildVersion"], function(
	WV3DViewerController,
	WV3DConstants,
	LogMessage,
	WV3DBuildVersion
) {
	"use strict";

	/**
	 * The WebViewer3DAPI object is the entry point for WebViewer3D. To use,
	 * create a new instance and provide the DOM element in which the viewer
	 * should be inserted. Dispose() function has to be called before this object
	 * goes out of scope.
	 * Warning: most of the API functions should be called asynchronuously, when the
	 * scene has been loaded succesfully, therefore, most of them should be called in
	 * __defaultClientLoadedSceneCallback (this constructor) OR
	 * __clientLoadedSceneCallback (loadScene function)
	 *
	 * @alias module:WebViewer3D
	 * @class WebViewer3DAPI
	 * @param {Object} __context [undefined] - a reference to the DOM element in which the viewer
	 *            should be inserted - can be null if no need for 3D canvas (e.g. Snapshots only)
	 * @param {function} __clientLoadedSceneCallback [undefined] - default callback function pointer to pass on
	 *            various error states back to client. eg., when loading of
	 *            scene is finished.
	 *            The client callback gives:
	 *            arg[0] => error status
	 *             == 0 indicates success with no errors.
	 *             == 1 indicates an error during http request (initial or reference to external file)
	 *             == 2 indicates an error during texture request
	 *             == 3 indicates an error during other resource request
	 *            arg[1] => the current file geometry percentage
	 *            arg[2] => the current file texture percentage
	 *            arg[3] => the current number of queried files
	 *            arg[4] => the current number of successfully retrieved files
	 *            e.g. the client just needs to test if (arg[1]=100 & arg[2]=100 & arg[3] = arg[4]) to
	 *            know if "everything" (the entire scene) has been loaded (and
	 *            displayed at least once).
	 * @param {function} __clientUIAddElementCallback [undefined] - default callback function
	 *            giving the current node path (arg[0]) and a boolean indicating
	 *            if it is a leaf (arg[1])
	 * @param {function} __clientLogMessageCallback [undefined] - default callback function
	 *            for output message from a string (arg[0]) and from a callback that can retrieve the current call stack (arg[1])
	 * @param {function} __urlPreprocessCallback [undefined] - default callback function for intercepting all http requests
	 *                      arg[0] => the current url that the webviewer would use when this callback is undefined
	 *                      arg[1] => the query callback that must be called to query the file using the replacement URL as a parameter
	 *                  e.g., here is a potential definition of such a url preprocess callback
	 *    <pre>
	 *    var urlPreProcCallback = function(__currentURL, __queryCallback)
	 *    {
	 *        var deferredCallback = function()
	 *        {
	 *            //Here we decorate our url using a time stamp, but we could
	 *            //litteraly swap it with a redirected url, upon user's choices
	 *            //if, for example, file are located on a different server than
	 *            //the initial URL the webviewer is trying to deduce
	 *            var newURL = __currentURL+"?t="+((new Date()).getTime());
	 *            __queryCallback(newURL);
	 *        }
	 *        setTimeout(deferredCallback, 3000);
	 *    }
	 *    </pre>
	 * The following is the default behavior of the webviewer, when __urlPreprocessCallback
	 * is omitted in the constructor and SetURLPreprocessCallback has not been used
	 *    <pre>
	 *    var urlPreProcCallbackSimple = function(__currentURL, __queryCallback)
	 *    {
	 *        __queryCallback(__currentURL);
	 *    }
	 *    </pre>
	 */
	const WebViewer3DAPI = function(
		__context,
		__clientLoadedSceneCallback,
		__clientUIAddElementCallback,
		__clientLogMessageCallback,
		__urlPreprocessCallback
	) {
		console.log("ESKO HTML Viewer - Client V" + WV3DBuildVersion.SuiteNumber + "/" + WV3DBuildVersion.P4VChangelistNumber);
		LogMessage.SetCallbackFunction(__clientLogMessageCallback);
		LogMessage.Log("Creating WebViewer3DAPI Object");

		this._controller = new WV3DViewerController(
			__context,
			__clientLoadedSceneCallback,
			__clientUIAddElementCallback,
			__urlPreprocessCallback
		);
	};

	/**
	 * Clear the scene and release all GPU resources. Has to be called before API object
	 * goes out of scope to avoid memory leaks.
	 *
	 * @function Dispose
	 */
	WebViewer3DAPI.prototype.Dispose = function() {
		LogMessage.Log("API Dispose");
		this._controller.Dispose();
	};

	/**
	 * Set Log message callback
	 *
	 * @function SetLogMessageCallback
	 * @param {function} __clientLogMessageCallback - default callback funtion
	 *            for output message from a string (arg[1])
	 */
	WebViewer3DAPI.prototype.SetLogMessageCallback = function(__clientLogMessageCallback) {
		LogMessage.SetCallbackFunction(__clientLogMessageCallback);
	};

	/**
	 * Set URL preprocess callback
	 *
	 * @function SetURLPreprocessCallback
	 * @param {function} __urlPreprocessCallback - default callback function for intercepting all http requests
	 *                      arg[1] => the current url that the webviewer would use when this callback is undefined
	 *                      arg[2] => the query callback that must be called to query the file using the replacement URL as a parameter
	 *                      (check constructor of WebViewer3DAPI for usage example)
	 */
	WebViewer3DAPI.prototype.SetURLPreprocessCallback = function(__urlPreprocessCallback) {
		this._controller.SetURLPreprocessCallback(__urlPreprocessCallback);
	};

	/**
	 * Set end camera interaction callback
	 *
	 * @function SetEndCameraInteractionCallback
	 * @param {function} __clientEndCameraInteractionCallback - callback function at the end of camera interaction
	 */
	WebViewer3DAPI.prototype.SetEndCameraInteractionCallback = function(__clientEndCameraInteractionCallback) {
		this._controller.SetEndCameraInteractionCallback(__clientEndCameraInteractionCallback);
	};

	/**
	 * Trigger a resize event that allows the WebGLViewer to adapt to the current context size
	 * This function is useful only when you do not want to use window resize event trigger
	 * User should use window.trigger('resize’) otherwise.
	 * This Resize function is useful especially if the webgl canvas is embedded in an independant
	 * and resizable dom element and avoid unnecessary refresh for the whole window. (i.e. a dockable
	 * widget)
	 * Warning: this function should not be called when Headless API is being used, since it might cause
	 * scene not to be rendered.
	 *
	 * @function Resize
	 */
	WebViewer3DAPI.prototype.Resize = function() {
		this._controller.get("_view").Resize();
	};

	/**
	 * Function deprecated
	 *
	 * @function Initialize: <i> deprecated </i>
	 */
	WebViewer3DAPI.prototype.Initialize = function() {
		console.error(
			"WebViewer3DAPI.Initialize: Deprecated API function (No need to call that function anymore: this function will be removed from next builds)"
		);
	};

	/**
	 * Load a scene from a URL that returns a JSON file containing the scene description
	 *
	 * @function LoadScene
	 * @param {String} __sceneURL - URL to the scene data
	 * @param {Number} __coplanarityEpsilon [0.02] - if > 0 used to tune tolerance for
	 * smooth surfaces so that these are considered as one single highlighted surface, i.e., only
	 * their silhouette will appear.
	 * @param {function} __clientLoadedSceneCallback [undefined] - default callback function pointer to pass on
	 *            various error states back to client. eg., when loading of
	 *            scene is finished.
	 *            The client callback gives:
	 *            arg[0] => error status
	 *             == 0 indicates success with no errors.
	 *             == 1 indicates an error during http request (initial or reference to external file)
	 *             == 2 indicates an error during texture request
	 *             == 3 indicates an error during other resource request
	 *            arg[1] => the current file geometry percentage
	 *            arg[2] => the current file texture percentage
	 *            arg[3] => the current number of queried files
	 *            arg[4] => the current number of successfully retrieved files
	 *            e.g. the client just needs to test if (arg[1]=100 & arg[2]=100 & arg[3] = arg[4]) to
	 *            know if "everything" (the entire scene) has been loaded (and
	 *            displayed at least once).
	 * @param {function} __clientUIAddElementCallback [undefined] - default callback function
	 *            giving the current node path (arg[0]) and a boolean indicating
	 *            if it is a leaf (arg[1])
	 * @param {Number} __boundaryhardedges [true] - if true, builds the hard edges highlights at surface boundaries
	 * @param {Object} __materialsAndTextures - JSON object containing materials and corresponding texture assets.
	 *                                           These will be used in the scene and will replace the material of the corresponding printable part.
	 */
	WebViewer3DAPI.prototype.LoadScene = function(
		__sceneURL,
		__coplanarityEpsilon,
		__clientLoadedSceneCallback,
		__clientUIAddElementCallback,
		__boundaryhardedges,
		__materialsAndTextures
	) {
		LogMessage.Log("API Load Scene from URL");

		this._controller.LoadSceneFromURL(
			__sceneURL,
			__coplanarityEpsilon,
			__clientLoadedSceneCallback,
			__clientUIAddElementCallback,
			__boundaryhardedges,
			__materialsAndTextures
		);
	};

	/**
	 * Load a scene from a local JSON file containing the scene description
	 *
	 * @function LoadSceneFromLocalFile
	 * @param {File} __file - A File API file object
	 * @param {Number} __coplanarityEpsilon [0.02] - if > 0 used to tune tolerance for
	 * smooth surfaces so that these are considered as one single highlighted surface, i.e., only
	 * their silhouette will appear.
	 * @param {function} __clientLoadedSceneCallback [undefined] - default callback function pointer to pass on
	 *            various error states back to client. eg., when loading of
	 *            scene is finished.
	 *            The client callback gives:
	 *            arg[0] => error status
	 *             == 0 indicates success with no errors.
	 *             == 1 indicates an error during http request (initial or reference to external file)
	 *             == 2 indicates an error during texture request
	 *             == 3 indicates an error during other resource request
	 *            arg[1] => the current file geometry percentage
	 *            arg[2] => the current file texture percentage
	 *            arg[3] => the current number of queried files
	 *            arg[4] => the current number of successfully retrieved files
	 *            e.g. the client just needs to test if (arg[1]=100 & arg[2]=100 & arg[3] = arg[4]) to
	 *            know if "everything" (the entire scene) has been loaded (and
	 *            displayed at least once).
	 * @param {function} __clientUIAddElementCallback [undefined] - default callback function
	 *            giving the current node path (arg[1]) and a boolean indicating
	 *            if it is a leaf (arg[2])
	 * @param {Number} __boundaryhardedges [true] - if true, builds the hard edges highlights at surface boundaries
	 * @param {String} _defaultURL [null] - this allows to set the default URL path used for external file referenced by the local file
	 */
	WebViewer3DAPI.prototype.LoadSceneFromLocalFile = function(
		__file,
		__coplanarityEpsilon,
		__clientLoadedSceneCallback,
		__clientUIAddElementCallback,
		__boundaryhardedges,
		_defaultURL
	) {
		LogMessage.Log("API Load Scene from local file");

		const reader = new FileReader();

		reader.onload = function(e) {
			this.LoadSceneFromObject(
				JSON.parse(reader.result),
				__coplanarityEpsilon,
				__clientLoadedSceneCallback,
				__clientUIAddElementCallback,
				__boundaryhardedges,
				_defaultURL
			);
		}.bind(this);

		reader.readAsText(__file);
	};

	/**
	 * Load a scene from an object
	 *
	 * @function LoadSceneFromObject
	 * @param {Object} __object - a JSON scene object (not a string!)
	 * @param {Number} __coplanarityEpsilon [0.02] - if > 0 used to tune tolerance for
	 * smooth surfaces so that these are considered as one single highlighted surface, i.e., only
	 * their silhouette will appear.
	 * @param {function} __clientLoadedSceneCallback [undefined] - default callback function pointer to pass on
	 *            various error states back to client. eg., when loading of
	 *            scene is finished.
	 *            The client callback gives:
	 *            arg[0] => error status
	 *             == 0 indicates success with no errors.
	 *             == 1 indicates an error during http request (initial or reference to external file)
	 *             == 2 indicates an error during texture request
	 *             == 3 indicates an error during other resource request
	 *            arg[1] => the current file geometry percentage
	 *            arg[2] => the current file texture percentage
	 *            arg[3] => the current number of queried files
	 *            arg[4] => the current number of successfully retrieved files
	 *            e.g. the client just needs to test if (arg[1]=100 & arg[2]=100 & arg[3] = arg[4]) to
	 *            know if "everything" (the entire scene) has been loaded (and
	 *            displayed at least once).
	 * @param {function} __clientUIAddElementCallback [undefined] - default callback function
	 *            giving the current node path (arg[1]) and a boolean indicating
	 *            if it is a leaf (arg[2])
	 * @param {Number} __boundaryhardedges [true] - if true, builds the hard edges highlights at surface boundaries
	 * @param {String} _defaultURL [null] - this allows to set the default URL path used for external file referenced by the local file
	 */
	WebViewer3DAPI.prototype.LoadSceneFromObject = function(
		__object,
		__coplanarityEpsilon,
		__clientLoadedSceneCallback,
		__clientUIAddElementCallback,
		__boundaryhardedges,
		_defaultURL,
		__materialsAndTextures
	) {
		LogMessage.Log("API Load Scene from object");

		this._controller.LoadSceneFromJSONObject(
			__object,
			__coplanarityEpsilon,
			__clientLoadedSceneCallback,
			__clientUIAddElementCallback,
			__boundaryhardedges,
			_defaultURL,
			__materialsAndTextures
		);
	};

	/**
	 * Reset visiblity of the main scene to its default
	 * default visiblity of the main scene is true at startup
	 *
	 * @function ResetMainSceneVisibility
	 */
	WebViewer3DAPI.prototype.ResetMainSceneVisibility = function() {
		LogMessage.Log("API Reset visiblity of main scene");

		this._controller.ResetMainSceneVisibility();
	};

	/**
	 * Toggle visiblity of a node using its URL in the scene
	 * default visiblity of a node is true at startup
	 *
	 * @function ToggleNodeVisibility
	 * @param {String} __nodeURL - URL identifying a scene node
	 * @param {Boolean} __recurseChildren - If true, traverse all children to toggle their visibility state
	 * @param {Boolean} __recurseParents - If true, traverse all parents to toggle their visibility state
	 */
	WebViewer3DAPI.prototype.ToggleNodeVisibility = function(__nodeURL, __recurseChildren, __recurseParents) {
		this._controller.ToggleNodeVisibility(__nodeURL, __recurseChildren, __recurseParents);
	};

	/**
	 * Set visiblity state of a node using its URL in the scene
	 * default visiblity of a node is true at startup
	 *
	 * @function SetNodeVisibility
	 * @param {String} __nodeURL - URL identifying a scene node
	 * @param {Boolean} __visible - Set the object visibility to __visible
	 * @param {Boolean} __recurseChildren - If true, traverse all children to set their visibility to __visible
	 * @param {Boolean} __recurseParents - If true, traverse all parents to set their visibility to __visible
	 */
	WebViewer3DAPI.prototype.SetNodeVisibility = function(__nodeURL, __visible, __recurseChildren, __recurseParents) {
		this._controller.SetNodeVisibility(__nodeURL, __visible, __recurseChildren, __recurseParents);
	};

	/**
	 * Set the angular position of the camera relative to the default front position of the scene
	 *
	 * @function SetCameraAngles
	 * @param {Number} __azimuthAngle - the walk-around angle measured counterclockwise from the default front view
	 * @param {Number} __elevationAngle - the elevantion angle - positive values lift the camera above the scene centre
	 * @param {Boolean} [__redraw = true] If true, redraw the scene after this call
	 */
	WebViewer3DAPI.prototype.SetCameraAngles = function(__azimuthAngle, __elevationAngle, __redraw) {
		this._controller.SetCameraAngles(__azimuthAngle, __elevationAngle, __redraw);
	};

	/**
	 * Get an object containing the angular position of the camera relative to the default front position of the scene
	 *
	 * @function GetCameraAngles
	 * @returns {{azimuth: Number, elevation:Number} azimuth - the walk-around angle measured counterclockwise from the default
	 * front view. elevation - the elevantion angle - positive values lift the camera above the scene centre
	 *
	 */
	WebViewer3DAPI.prototype.GetCameraAngles = function() {
		return this._controller.GetCameraAngles();
	};

	/**
	 * Convenience function to reset the default camera to its default positions. This function call is same as calling
	 * SetCameraAngles with 0.0 for both angles and SetCameraScale with 1.0.
	 *
	 * @function ResetCamera
	 */
	WebViewer3DAPI.prototype.ResetCamera = function() {
		this._controller.ResetCamera();
	};

	/**
	 * Control interactive functionality. Pass 'false' to not have any interactive control on the 3D view.
	 * pass 'true' again to turn on interactivity.
	 *
	 * @function ToggleInteractiveCameraControl
	 * @param __onOrOff
	 */
	WebViewer3DAPI.prototype.ToggleInteractiveCameraControl = function(__onOrOff) {
		this._controller.SetInteractiveCameraControl(__onOrOff);
	};

	/**
	 * Control interactive functionality. "WV3DConstants.interaction.NONE", "WV3DConstants.interaction.PAN", "WV3DConstants.interaction.SCALE",
	 * "WV3DConstants.interaction.ROTATE", "WV3DConstants.interaction.ENVROT", "WV3DConstants.interaction.ALL" or an OR combination of these
	 *
	 * @function SetInteractiveCameraControlMask
	 * @param {Number} __mask - an or combination of the following flags "WV3DConstants.interaction.NONE", "WV3DConstants.interaction.PAN",
	 * "WV3DConstants.interaction.SCALE", "WV3DConstants.interaction.ROTATE", "WV3DConstants.interaction.ENVROT", "WV3DConstants.interaction.ALL".
	 */
	WebViewer3DAPI.prototype.SetInteractiveCameraControlMask = function(__mask) {
		this._controller.SetInteractiveCameraControlMask(__mask);
	};

	/**
	 * Set the size of the scene within the camera view. This is achieved by automatically
	 * adjusting the camera properties (e.g. changing the camera distance for a perspective camera)
	 *
	 * @function SetCameraScale
	 * @param {Number} __scale - the size of the scene bounding box in the view. A value of 1 will result in the scene filling the view in at least one dimension
	 * @param {Boolean} [__redraw = true] If true, redraw the scene after this call
	 */
	WebViewer3DAPI.prototype.SetCameraScale = function(__scale, __redraw) {
		this._controller.SetCameraScale(__scale, __redraw);
	};

	/**
	 * Get the size of the scene within the camera view.
	 *
	 * @function GetCameraScale
	 * @returns {Number} the size of the scene bounding box in the view. A value of 1 will result in the scene filling the view in at least one dimension
	 */
	WebViewer3DAPI.prototype.GetCameraScale = function() {
		return this._controller.GetCameraScale();
	};

	/**
	 * Set the absolute distance of the camera with respect to the scene center. This will update the current scale factor
	 * to match the desired distance but does not guarantee that the scene boundaries are enclosed within the frustum.
	 * Clamping will still occur to prevent the camera from entering the scene or zooming out too much.
	 *
	 * @function SetCameraDistance
	 * @param {Number} __distance - the absolute distance of the camera with respect to the scene center
	 * @param {Boolean} [__redraw = true] If true, redraw the scene after this call
	 */
	WebViewer3DAPI.prototype.SetCameraDistance = function(__distance, __redraw) {
		this._controller.SetCameraDistance(__distance, __redraw);
	};

	/**
	 * Get the absolute distance of the camera with respect to the scene center
	 *
	 * @function GetCameraDistance
	 * @returns {Number} the absolute distance of the camera with respect to the scene center
	 */
	WebViewer3DAPI.prototype.GetCameraDistance = function() {
		return this._controller.GetCameraDistance();
	};

	/**
	 * Set the panning offset of the camera.
	 *
	 * @function SetCameraPanOffset
	 * @param {Number} __x_offset - horizontal offset
	 * @param {Number} __y_offset - vertical offset
	 * @param {boolean} [__redraw = true] - indicates if the scene has to be redrawn after this call. Default is true.
	 */
	WebViewer3DAPI.prototype.SetCameraPanOffset = function(__x_offset, __y_offset, __redraw) {
		this._controller.SetCameraPanOffset(__x_offset, __y_offset, __redraw);
	};

	/**
	 * Get the current panning offset of the current camera. Returns an object with properties "x" and "y" indicating the
	 * offset in each of these dimensions.
	 *
	 * @function GetCameraPanOffset
	 * @returns {Object} containing pan properties 'x' and 'y'
	 */
	WebViewer3DAPI.prototype.GetCameraPanOffset = function() {
		return this._controller.GetCameraPanOffset();
	};

	/**
	 * Animate the current camera to given parameters. If the camera is currently animating the function returns false.
	 *
	 * @function AnimateToView
	 * @param {Number} __scale - scaling factor
	 * @param {Number} __azimuth - azimuth angle
	 * @param {Number} __elevation - elevation angle
	 * @param {Number} __x_offset - pan horizontal offset
	 * @param {Number} __y_offset - pan vertical offset
	 * @param {Number} [__duration = 1] - animation duration in seconds
	 * @returns {Boolean} false if an animation is already running else returns true
	 */
	WebViewer3DAPI.prototype.AnimateToView = function(__scale, __azimuth, __elevation, __x_offset, __y_offset, __duration) {
		return this._controller.AnimateToView(__scale, __azimuth, __elevation, __x_offset, __y_offset, __duration);
	};

	/**
	 * SetAnimations enabled or disabled
	 *
	 * @function SetAnimations deprecated
	 */
	WebViewer3DAPI.prototype.SetAnimations = function() {
		LogMessage.Log("SetAnimations API deprecated !!! Use SetCameraSpinAnimation instead.");
	};

	/**
	 * SetSpinAnimation enabled or disabled
	 *
	 * @function SetSpinAnimation
	 * @param {boolean} [__enableAnimations=true] - enables the camera animation if true, disable it if false
	 */
	WebViewer3DAPI.prototype.SetCameraSpinAnimation = function(__enableAnimations) {
		LogMessage.Log("API enable/disable animations: " + __enableAnimations);

		this._controller.SetCameraSpinAnimation(__enableAnimations);
	};

	/**
	 * Set the environment rotation.
	 *
	 * @function SetEnvironmentRotation
	 * @param {Number} __quaternion - rotation as a quaternion
	 * @param {boolean} [__redraw = true] - indicates if the scene has to be redrawn after this call. Default is true.
	 */
	WebViewer3DAPI.prototype.SetEnvironmentRotation = function(__quaternion, __redraw) {
		this._controller.SetEnvironmentRotation(__quaternion, __redraw);
	};

	/**
	 * Get the current environment rotation. Returns the rotation as a quaternion.
	 *
	 * @function GetEnvironmentRotation
	 */
	WebViewer3DAPI.prototype.GetEnvironmentRotation = function() {
		return this._controller.GetEnvironmentRotation();
	};

	/**
	 * Enables a specific rendering mode (for interaction) in which the best suitable scale of the scene
	 * is applied to the current view. This neglects previously set scale factor.
	 *
	 * @function SetOptimisedViewRendering
	 * @param {{x:Number,y:Number}} __border set the minimum gap between the projected bbox of the scene (labels included) and the borders of the image plane
	 * @param {Boolean} __enableOptimisation indicate the width of the viewport. default is the current canvas width.
	 * @param {Boolean} [__redraw = true] If true, redraw the scene after this call
	 */
	WebViewer3DAPI.prototype.SetOptimisedViewRendering = function(__border, __enableOptimisation, __redraw) {
		this._controller.SetOptimisedViewRendering(__border, __enableOptimisation, __redraw);
	};

	/**
	 * Set up a perspective camera. The camera distance from the scene centre will be set automatically based on the current scale setting.
	 *
	 * @function SetPerspectiveCamera
	 * @param {Number} [__fieldOfView=45] - the field of view in degrees
	 * @param {Boolean} [__redraw = true] If true, redraw the scene after this call
	 */
	WebViewer3DAPI.prototype.SetPerspectiveCamera = function(__fieldOfView, __redraw) {
		this._controller.SetPerspectiveCamera(__fieldOfView, __redraw);
	};

	/**
	 * Set up an orthographic camera. The viewport size will be set automaticall based on the current scale setting.
	 *
	 * @function SetOrthographicCamera
	 * @param {Boolean} [__redraw = true] If true, redraw the scene after this call
	 */
	WebViewer3DAPI.prototype.SetOrthographicCamera = function(__redraw) {
		this._controller.SetOrthographicCamera(__redraw);
	};

	/**
	 * Focus camera from a node's bbox using its URL in the scene
	 * default focus of camera is on the root node of the scene
	 *
	 * @function SetCameraFocus
	 * @param {String} __nodeURL - URL identifying the scene node
	 * @param {Boolean} [__redraw = true] If true, redraw the scene after this call
	 */
	WebViewer3DAPI.prototype.SetCameraFocus = function(__nodeURL, __redraw) {
		this._controller.SetCameraFocus(__nodeURL, __redraw);
	};

	/**
	 * Set specific callback for when a pointer event is detected
	 *
	 * @function SetPointerCallback
	 * @param {function} __pointerCallback - callback function
	 *            giving the x and y coordinates of picked point (arg[1], arg[2])
	 *            and which should return false to disable the default
	 *            mouse handling of the 3D viewer, true otherwise
	 */
	WebViewer3DAPI.prototype.SetPointerCallback = function(__pointerCallback) {
		this._controller.SetPointerCallback(__pointerCallback);
	};

	/**
	 * Returns a list of intersections containing information on intersected
	 * objects. Each of those is a javascript object containing the following fields:
	 * - "objectId" of type "string" containing the object uuid
	 * - "objectURL" of type "string" containing the object URL (can be used with SetNodeVisibility or SetLabelsVisibility)
	 * - "materialId" of type "string" containing the initial material assigned to the object
	 * - "instanceId" of type "number" (integer) refering to the object instance intersected
	 *      (useful if more than one instance of the object in the scene)
	 * - "faceId" of type "number" (integer) refering to the object intersected face
	 * - "point" of type THREE.Vector3 refering to the 3d intersection point
	 * - "uv" of type THREE.Vector2 refering to the uv coordinate corresponding
	 *      to the intersection point (interpolated from the face vertices uv coordinates,
	 *      see https://computergraphics.stackexchange.com/questions/1866/how-to-map-square-texture-to-triangle
	 *      for more details)
	 *
	 * @function Pick
	 * @param {Number} __x - pixel X axis coordinate of picked point (between 0 and context width)
	 * @param {Number} __y - pixel Y axis coordinate of picked point (between 0 and context height)
	 */
	WebViewer3DAPI.prototype.Pick = function(__x, __y) {
		return this._controller.Pick(__x, __y);
	};

	/**
	 * SetHighlight enabled or disabled
	 *
	 * @function SetHighlights
	 * @param {Boolean} [__enableHighlights=true] - enables the objects highlighting if true, disable it if false
	 * @param {Boolean} [__redraw = true] If true, redraw the scene after this call
	 */
	WebViewer3DAPI.prototype.SetHighlights = function(__enableHighlights, __redraw) {
		this._controller.SetHighlights(__enableHighlights, __redraw);
	};

	/**
	 * SetHighlightsConfiguration enables and disables each highlight component
	 * Call function SetHighlights first to enable highlights
	 * By default, __enableHorizonHighlights and __enableHardEdgeHighlights are set to true
	 * @function
	 * @param {boolean} __enableHorizonHighlights - enables the horizon highlighting if true and highlighting is enabled too, disable it otherwise
	 * @param {boolean} __enableHardEdgeHighlights - enables the hard edge highlighting if true and highlighting is enabled too, disable it otherwise
	 * @param {Boolean} [__redraw = true] If true, redraw the scene after this call
	 */
	WebViewer3DAPI.prototype.SetHighlightsConfiguration = function(__enableHorizonHighlights, __enableHardEdgeHighlights, __redraw) {
		this._controller.SetHighlightsConfiguration(__enableHorizonHighlights, __enableHardEdgeHighlights, __redraw);
	};

	/**
	 * SetHighlightColor
	 *
	 * @function SetHighlightsColor
	 * @param {Number} __red - red component (0 - 255)
	 * @param {Number} __green - green component (0 - 255)
	 * @param {Number} __blue - blue component (0 - 255)
	 * @param {Boolean} [__redraw = true] If true, redraw the scene after this call
	 */
	WebViewer3DAPI.prototype.SetHighlightsColor = function(__red, __green, __blue, __redraw) {
		this._controller.SetHighlightsColor(parseFloat(__red), parseFloat(__green), parseFloat(__blue), __redraw);
	};

	/**
	 * Set the background color for the current scene
	 * Setting an alpha lower than 255 might result in some visual artifacts in reflections since transparent objects on top of transparent background
	 * is still experimental.
	 * This function must be called when __clientCallback of API returns a fully loaded scene.
	 *
	 * @function SetBackgroundColor
	 * @param {Number} __red - red component (0 - 255)
	 * @param {Number} __green - green component (0 - 255)
	 * @param {Number} __blue - blue component (0 - 255)
	 * @param {Number} __alpha - __alpha component (0 - 255) (represent the opacity of the background, useful for snapshot)
	 * @param {Boolean} [__redraw = true] If true, redraw the scene after this call
	 */
	WebViewer3DAPI.prototype.SetBackgroundColor = function(__red, __green, __blue, __alpha, __redraw) {
		this._controller.SetBackgroundColor(parseFloat(__red), parseFloat(__green), parseFloat(__blue), parseFloat(__alpha), __redraw);
	};

	/**
	 * Set the backdrop texture
	 *
	 * @function SetBackdropTexture
	 * @param {String} __textureURL - URL identifying the texture URL to be used
	 * @param {function} __successCallback - callback funtion used when rendering has been updated
	 * @param {Boolean} [__redraw = true] If true, redraw the scene after this call
	 */
	WebViewer3DAPI.prototype.SetBackdropTexture = function(__textureURL, __successCallback, __redraw) {
		LogMessage.Log("API SetBackdrop texture");

		this._controller.SetBackdropTexture(__textureURL, __successCallback, __redraw);
	};

	/**
	 * Show or Hide all labels in the scene. Per label visibility settings are not altered with this function.
	 *
	 * @function SetLabelsVisibility
	 * @param __show {Boolean} [__show=true] - true to show all labels in the scene and false otherwise
	 * @param {Boolean} [__redraw = true] If true, redraw the scene after this call
	 */
	WebViewer3DAPI.prototype.SetLabelsVisibility = function(__show, __redraw) {
		this._controller.SetLabelsVisibility(__show, __redraw);
	};

	/**
	 * Updates properties of Label assets for a particular node that are specified in JSON files.
	 * The URL identifies the label asset for a particular node in the scene.
	 *
	 * @function UpdateLabelProperties
	 * @param __nodeURL {String} - URL identifying the label in the scene.
	 *                          eg., http://<yourhost>/path/to/jsonfile.json/InstanceGroup/MyLabel1
	 * @param __properties {Object} - Label properties as JSON object. Should include one or more of
	 *                          'text', 'display_style', 'orientation', 'preferred_loc_v', 'preferred_loc_h', 'pixel_offset' and 'visible'
	 *                          'display_style' should specify full styling and not a reference to scene's label display styles.
	 *                          Valid values for 'visible' are true and false. A particular label's visibility can be controlled with this property.
	 *                          A label that is already turned off particular node using "label_name":null mechanism in labelsettingsmap cannot be updated using this API
	 *                          eg., JSON.parse({ "text": "new label text", "display_style": "{ "fontsize": 36, "fontface": "Arial", "borderColor": {"r":0, "g":0, "b":0, "a":1.0}, "backgroundColor": {"r":255, "g":255, "b":255, "a":1.0} },
	 *                          "orientation": "horizontal", "preferred_loc_h": "right", "preferred_loc_v": "top" }")
	 * @return false - If the __nodeURL is invalid OR if the label with __nodeURL is not found in the scene.
	 *         true - In case of successful execution. (no other error case is hit)
	 * @note Throws exceptions in the case of an undefined __nodeURL or __properties or if the label with __nodeURL is not found in the scenegraph
	 */
	WebViewer3DAPI.prototype.UpdateLabelProperties = function(__nodeURL, __properties) {
		return this._controller.UpdateLabelProperties(__nodeURL, __properties);
	};

	/**
	 * Updates properties of all label assets in the scene. A Convenience function that iterates
	 * over all labels in the scene updating their display properties. Effectively same as calling the
	 * UpdateLabelProperties() function for each label node in the scene. Typically used to update display_style
	 * settings of labels in one go (eg., update fontsize to increase text size in all labels!)
	 *
	 * @function UpdateAllLabelsProperties
	 * @param __properties {Object} - Label properties as JSON object. Should include one or more of
	 *                          'text', 'display_style', 'orientation', 'preferred_loc_v', 'preferred_loc_h', 'pixel_offset' and 'visible'
	 *                          'display_style' should specify full styling and not a reference to scene's label display styles.
	 *                          Valid values for 'visible' are true and false. A particular label's visibility can be controlled with this property.
	 *                          A label that is already turned off particular node using "label_name":null mechanism in labelsettingsmap cannot be updated using this API
	 *                          eg., JSON.parse({ "display_style": "{ "fontsize": 36 }" })
	 * @return false if the __properties is not valid JSON (atleast one allowed parameters) or if there are no labels present in the scene. Otherwise returns true
	 * @note Throws exception in the case of an undefined __properties object
	 */
	WebViewer3DAPI.prototype.UpdateAllLabelsProperties = function(__properties) {
		return this._controller.UpdateAllLabelsProperties(__properties);
	};

	/**
	 * Retrieve the list of the parts that can support annotations
	 *
	 * @function GetAnnotablePartsList
	 * @return array of strings
	 */
	WebViewer3DAPI.prototype.GetAnnotablePartsList = function() {
		return this._controller.GetAnnotablePartsList();
	};

	/**
	 * Set the segment list of annotations
	 *
	 * @function SetPartAnnotationsSegments
	 * @param {string} __jsonMaterialName - a string matching the material name (json file)
	 * @param {array} __segmentList - an array containing each segment with the following properties: "startU","startV","endU","endV"
	 * @param {boolean} [__redraw = true] - indicates if the scene has to be redrawn after this call. Default is true.
	 * @return true if the json material name was found and was assigned correctly the annotation segments
	 */
	WebViewer3DAPI.prototype.SetAnnotations = function(__jsonMaterialName, __segmentList, __redraw) {
		return this._controller.SetAnnotations(__jsonMaterialName, __segmentList, __redraw);
	};

	/**
	 * Set thickness of annotations
	 *
	 * @function SetAnnotationsThickness
	 * @param {number} __thickness - [1 - 50] a number indicating the annotation segments thickness
	 * @param {boolean} [__redraw = true] - indicates if the scene has to be redrawn after this call. Default is true.
	 */
	WebViewer3DAPI.prototype.SetAnnotationsThickness = function(__thickness, __redraw) {
		this._controller.SetAnnotationsThickness(__thickness, __redraw);
	};

	/**
	 * Set color of annotations
	 *
	 * @function SetAnnotationsColor
	 * @param {Number} __red - red component (0 - 255)
	 * @param {Number} __green - green component (0 - 255)
	 * @param {Number} __blue - blue component (0 - 255)
	 * @param {Boolean} [__redraw = true] If true, redraw the scene after this call
	 */
	WebViewer3DAPI.prototype.SetAnnotationsColor = function(__red, __green, __blue, __redraw) {
		this._controller.SetAnnotationsColor(parseFloat(__red), parseFloat(__green), parseFloat(__blue), __redraw);
	};

	/**
	 * SetReflections sets the reflection on, off, or relative to what is provided in the Json file.
	 *
	 * @function SetReflections
	 * @param {string} __config - a string, "ON", "OFF", or "JSON" ("JSON" is already set by default)
	 * @param {number} [__opacity = JSON-Value ] - [0 1] opacity of reflections (only if "ON" mode chosen)
	 * @param {boolean} [__fade = JSON-Value ] - turns ON or OFF fading of reflections (only if "ON" mode chosen)
	 * @param {number} [__position = JSON-Value ] - [0 1] position of reflections (only if "ON" mode chosen)
	 * @param {number} [__fadingSize = JSON-Value ] - [1 2] fading size of reflections (only if "ON" mode chosen)
	 * @param {boolean} [__redraw = true] - indicates if the scene has to be redrawn after this call. Default is true.
	 */
	WebViewer3DAPI.prototype.SetReflections = function(__config, __opacity, __fade, __position, __fadingSize, __redraw) {
		LogMessage.Log("API SetReflections: " + __config);
		this._controller.SetReflections(__config, __opacity, __fade, __position, __fadingSize, __redraw);
	};

	/**
	 * SetInteractionOptimization enabled or disabled
	 *
	 * @function SetInteractionOptimization
	 * @param {boolean} [__enableInteractionOptimization=true] - enables the scene optimization while interacting with the model if true, disable it if false
	 */
	WebViewer3DAPI.prototype.SetInteractionOptimization = function(__enableInteractionOptimization) {
		LogMessage.Log("API enable/disable interaction optimization: " + __enableInteractionOptimization);

		this._controller.SetInteractionOptimization(__enableInteractionOptimization);
	};

	/**
	 * SetQualityLevel
	 *
	 * @function SetQualityLevel sets maximum quality when no interaction with the model or when interaction optimization is disabled
	 * by setting the number of transparency peels for transparent models. This quality level does not affect quality of opaque models.
	 * @param {number} __qualityLevel - an integer between 1 and 10 (included)
	 * @param {boolean} [__redraw = true] - indicates if the scene has to be redrawn after this call. Default is true.
	 */
	WebViewer3DAPI.prototype.SetQualityLevel = function(__qualityLevel, __redraw) {
		LogMessage.Log("API SetQualityLevel: " + __qualityLevel);

		this._controller.SetQualityLevel(__qualityLevel, __redraw);
	};

	/**
	 * SetLightingContrast sets the lighting contrast (in designer quality only)
	 * by reducing the directional light and increasing the ambient light components.
	 *
	 * @function SetLightingContrast
	 * @param {number} __lightingContrast - a floating number between 0 and 1 (included). default value is 0.5.
	 * @param {boolean} [__redraw = true] - indicates if the scene has to be redrawn after this call. Default is true.
	 */
	WebViewer3DAPI.prototype.SetLightingContrast = function(__lightingContrast, __redraw) {
		LogMessage.Log("API SetLightingContrast: " + __lightingContrast);

		this._controller.SetLightingContrast(__lightingContrast, __redraw);
	};

	/**
	 * SetDefaultResourceRelativePath (optional function when loading scene from samples, useful to specify Resource folder location
	 * not set to './Resources')
	 *
	 * @function SetDefaultResourceRelativePath
	 * @param {String} [__path = './Resources'] - overwrite the default resource relative path
	 */
	WebViewer3DAPI.prototype.SetDefaultResourceRelativePath = function(__path) {
		this._controller.SetDefaultResourceRelativePath(__path);
	};

	/**
	 * Get an indication of the degree to which the current system is supported for 3D viewing
	 *
	 * @function GetSystemSupportLevel
	 * @returns {{ "issues": Number, "support_level": supportLevel.FULL | supportLevel.PARTIAL | supportLevel.NONE}} An object containing "issues"
	 * (Array of issues) and a "support_level" (supportLevel.FULL | supportLevel.PARTIAL | supportLevel.NONE) property
	 */
	WebViewer3DAPI.prototype.GetSystemSupportLevel = function() {
		return this._controller.GetSystemSupportLevel();
	};

	/**
	 * Get a technical report containing system info
	 *
	 * @function GetSystemReport
	 * @returns {String} a report containing system info
	 */
	WebViewer3DAPI.prototype.GetSystemReport = function() {
		return this._controller.GetSystemReport();
	};

	/**
	 * Get a snapshot of the current viewer state as specified in format
	 *
	 * @function GetSnapshot
	 * @param {string} [__format = "png"] - image format (accepted: "png, jpeg")
	 * @param {number} [__width = context canvas width or 1] - image width in pixels (accepted: [1:2048] higher depend on graphics card and browser)
	 * @param {number} [__height = context canvas height or 1] - image height in pixels (accepted: [1:2048] higher depend on graphics card and browser)
	 * @returns {Object} image data
	 */
	WebViewer3DAPI.prototype.GetSnapshot = function(__format, __width, __height) {
		return this._controller.GetSnapshot(__format, __width, __height);
	};

	return WebViewer3DAPI;
});
