/**
 * @author DOGO
 * @module WebViewer3D
 * @copyright   2014 Esko Stonecube Ltd. All rights reserved.
 */

define([
    "underscore",
    "backbone",
    "@esko/3d-web-viewer/modules/HttpRequestManager",
    "@esko/3d-web-viewer/modules/LogMessage"
    ], function (
        _,
        bb,
        HttpRequestManager,
        LogMessage
        ) {

    'use strict';

    /**
     * @alias module:WebViewer3D
     * @class WV3DAssetLoader
     */
    return bb.Model.extend({

    initialize: function(__assetLoadedCallback, __assetLoadErrorCallback, __urlPreprocessCallback)
    {
        LogMessage.Log("Creating WV3DAssetLoader Object");
        this.set('_httprequestmanager', new HttpRequestManager(__assetLoadedCallback, __assetLoadErrorCallback, __urlPreprocessCallback));
    },

    /**
     * Set the current resource to read data from
     */
    SetResource:function (resource)
    {
        this.set('_jsondata', resource);
    },

    /**
     * Check a string and return true if it is a URL
     */
    IsURL:function (__sceneURL)
    {
        return this.get('_httprequestmanager').IsURL(__sceneURL);
    },

    /**
     * Return the file name + extension from the URL
     *
     * @param {String} __string URL to the file
     */
    GetFileName: function(__string)
    {
        return this.get('_httprequestmanager').GetFileName(__string);
    },

    /**
     * Return the asset name + extension from the URL
     *
     * @param {String} __string URL to the asset
     */
    GetAssetName: function(__string)
    {
        return this.get('_httprequestmanager').GetAssetName(__string);
    },

    /**
     * Load a url and return to the caller when data is ready
     */
    LoadURL:function (__sceneURL, __parentSceneURL)
    {
        this.get('_httprequestmanager').LoadURL(__sceneURL, __parentSceneURL);
    },

    /**
     * Apply somefunction to the object
     */
    Printresourceinfo: function()
    {

        LogMessage.Log("AssetLoader Printresourceinfo function");

        let str = "", data = this.get('_jsondata');

        for(let i = 0, keys = Object.keys(data), l = keys.length; i < l; ++i)
        {
            const attribname = keys[i];
            str = str + attribname + ": " + data[attribname] + " / ";
        }

        LogMessage.Log(str);

    },


    /**
     * Returns the set in which the resource shall be pushed (or not if already cached)
     */
    Getresourcelocation: function(resource)
    {
        if(resource.hasOwnProperty('AssetType'))
        {
            switch(resource.AssetType)
            {
                case "Scene": return "Reference";
                default: return "NoLocation";
            }
        }else if(resource.hasOwnProperty('AssetRef'))
        {
            return "NoLocation";
        }
    },


    /**
     * Apply somefunction to the object
     */
    Getrootnode: function()
    {

        LogMessage.Log("AssetLoader GetRootnode function");
        return this.get('_jsondata');

    },

    /**
     * Apply somefunction to the object
     */
    Getrootnoderefname: function()
    {

        LogMessage.Log("AssetLoader GetRootnode function");
        return this.get('_jsondata').TopLevelAsset.AssetRef;

    },

    /**
     * Return the list of matching elements in the queried object hierarchy
     */
    GetAssetFromName: function(obj,key,val)
    {

        var objects = [];
        for (var j = 0, keys = Object.keys(obj), l = keys.length; j < l; ++j)
        {
            var i = keys[j];
            if (typeof obj[i] === 'object') {
                objects = objects.concat(this.GetAssetFromName(obj[i], key, val));
            } else if (i === key && obj[key] === val) {
                objects.push(obj);
            }
        }
        return objects;
    },

    /**
     * Return the list containing the children or the assetref of the queried element
     */
    GetRefChildrenFromName: function(resource, __sceneURL)
    {
        // LogMessage.Log("Getting references");
        let list = [];

        if(resource.hasOwnProperty('Children') && resource.Children !== null && resource.Children !== "")
        {
            if(resource.Children === null) return list;
            if(resource.Children.length === 0)
            {
                LogMessage.Log("No Children found for this node");
            }
            list = list.concat(resource.Children);

        }else if(resource.hasOwnProperty('AssetRef') && resource.AssetRef !== null)
        {
            let ref_list = [];
            if(this.IsURL(resource.AssetRef))
            {
                // URL reference
                // Here the referenced scene or specific resource is translated from a relative
                // to an absolute path and keeps the ref to the specific asset (#.../.../)
                // if and only if the path has not already been changed to an absolute URL
                if(this.get('_httprequestmanager').IsRelativeURL(resource.AssetRef))
                {
                    resource.AssetRef = this.get('_httprequestmanager').GetRealURLWithAppendedRef(resource.AssetRef, __sceneURL);
                }
            }else
            {
                // Standard reference
                // Here is a standard fetch of the items contained in the current json structure
                ref_list = ref_list.concat(this.get('_jsondata')['Assets'][resource.AssetRef]);
            }

            if(ref_list.length === 0)
            {
                LogMessage.Log("No Reference found with AssetId: "+resource.AssetRef);
            }else
            {
                if(ref_list.length > 1)
                {
                    LogMessage.Log("Warning: AssetId in json structure is supposed to be unique: only one item returned for AssetId: "+resource.AssetRef);
                }
                list.push(ref_list[0]);
            }

        }else
        {
            for (let j = 0, keys = Object.keys(resource), l = keys.length; j < l; ++j)
            {
                const elt = keys[j];
                if( resource[elt] &&
                    ( resource[elt].hasOwnProperty('Children')
                    || resource[elt].hasOwnProperty('AssetRef')
                    || resource[elt].hasOwnProperty('Assets')))
                {
                    list = list.concat(resource[elt]);
                }
            }
        }

        return list;
    }

    });

});
