/**
 * @author DOGO
 * @module WebViewer3D
 * @copyright   2014 Esko Stonecube Ltd. All rights reserved.
 */

define([
    "underscore",
    "backbone",
    "@esko/3d-web-viewer/modules/AssetLoader",
    "@esko/3d-web-viewer/modules/LogMessage"
], function (
    _,
    bb,
    AssetLoader,
    LogMessage
)
{

    "use strict";

    /**
     * @alias module:WebViewer3D
     * @class WV3DAssetManager
     */
    return bb.Model.extend({


        initialize: function (__assetLoadedCallback, __assetLoadErrorCallback, __urlPreprocessCallback)
        {
            this.set("_assetloader", new AssetLoader(__assetLoadedCallback, __assetLoadErrorCallback, __urlPreprocessCallback));
            this.set("_replacement_materials_and_textures", {});// Materials and the corresponding textures to be injected into the scene.

            /** Default resource path */
            this.set("_resourcePath", "../Resources");

            this.Clear();
        },

        /**
         * Clear the cache and reinit
         */
        Clear: function (resource)
        {
            this.set("_cache", {
                "rootnodes": {}//in use: contains the loaded scenes using their absolute URL as a key
            });
            this.set("_urlCache", {});
            this.set("_pendingqueriedassets", []);

            this.set("_queriedassetnumber", 0);
            this.set("_resolvedassetnumber", 0);
            this.set("_textureCounter", 0);
            this.set("_textureLoadedCounter", 0);
        },

        SetDefaultResourceRelativePath: function (__path)
        {
            this.set("_resourcePath", __path);
        },

        /**
         * @param {Object} __materialsAndTextures - JSON object containing materials and corresponding texture assets.
         *                                            These will be used in the scene and will replace the material of the corresponding printable part.
         */
        SetReplacementMaterialsAndTextures: function (__materialsAndTextures)
        {
            this.set("_replacement_materials_and_textures", __materialsAndTextures);
        },

        /**
         * Replace the materials (and the corresponding textures) in the asset pool belonging to the scene with id \param __sceneId.
         * When a certain asset is already present in the original asset pool, it will be overriden.
         *
         * @param {Number} __sceneId - ID of the scene which determines the asset pool to use.
         */
        InsertMaterialsAndTexturesInAssetPool: function (__sceneId)
        {
            const current_asset_pool = this.get("_cache")["rootnodes"][__sceneId]["Assets"];
            const replacement_assets = this.get("_replacement_materials_and_textures");
            for (let i = 0, keys = Object.keys(replacement_assets), length = keys.length; i < length; ++i)
            {
                const element = keys[i];

                // A material element should be already present in the original asset pool to be able to replace it.
                if (!current_asset_pool.hasOwnProperty(element) && replacement_assets[element].AssetType === "Material")
                {
                    LogMessage.Log("ERROR: The material to be replaced is not present in the scene.");
                }
                else
                {
                    current_asset_pool[element] = replacement_assets[element];
                }
            }
        },


        /**
         * Return URL preprocesscallback
         * IMPROVEME: move texture loaders and material loaders of viewer view in subclasses (MaterialManager/TextureManager)
         */
        GetURLPreprocessCallback: function ()
        {
            return this.get("_assetloader").get("_httprequestmanager").get("_urlPreprocessCallback");
        },

        /**
         * Sets URL preprocesscallback
         * IMPROVEME: move texture loaders and material loaders of viewer view in subclasses (MaterialManager/TextureManager)
         */
        SetURLPreprocessCallback: function (__urlPreprocessCallback)
        {
            this.get("_assetloader").get("_httprequestmanager").set("_urlPreprocessCallback", __urlPreprocessCallback);
        },

        /**
         * Returns the set in which the resource shall be pushed (or not if already cached)
         */
        Getresourcelocation: function (resource)
        {

            // LogMessage.Log("AssetManager Getresourcelocation function");
            const cache = this.get("_cache");
            const resourcetypename = this.get("_assetloader").Getresourcelocation(resource);
            switch (resourcetypename)
            {
                //This is the location were we put our fetched scenes (one for each file actually)
                case "Reference":
                    return cache.rootnodes;
                //This location is mainly for debug purposes
                case "NoLocation":
                default:
                    if (this.get("_cache")[resourcetypename] === undefined)
                    {
                        this.get("_cache")[resourcetypename] = [];
                    }
                    return this.get("_cache")[resourcetypename];
            }
        },

        /**
         * Request the key type of unique id we should be looking for when trying to cache a resource
         */
        RequestType: function (resource)
        {
            if (resource.hasOwnProperty("AssetType") && resource["AssetType"] === "Scene")
            {
                return "AssetId";
            }
            else
            {
                // LogMessage.Log("No unique name to reference this resource. Cannot say if in pending queue list.");
                return null;
            }
        },

        /**
         * Return the file name + extension from the URL
         *
         * @param {String} __string URL to the file
         */
        GetFileName: function (__string)
        {
            return this.get("_assetloader").GetFileName(__string);
        },

        /**
         * Return the asset name + extension from the URL
         *
         * @param {String} __string URL to the asset
         */
        GetAssetName: function (__string)
        {
            return this.get("_assetloader").GetAssetName(__string);
        },

        /**
         * Updates the query counter
         */
        IncrementQueryCounter: function ()
        {
            this.set("_queriedassetnumber", this.get("_queriedassetnumber") + 1);
        },

        /**
         * Update a result query counter
         */
        IncrementResolvedQueryCounter: function ()
        {
            this.set("_resolvedassetnumber", this.get("_resolvedassetnumber") + 1);
        },

        /**
         * Update query counter for textures
         */
        IncrementTextureQueryCounter: function (__nb_queried_textures)
        {
            this.set("_textureCounter", this.get("_textureCounter") + __nb_queried_textures);
        },

        /**
         * Update query counter for textures
         */
        IncrementTextureLoadCounter: function ()
        {
            this.set("_textureLoadedCounter", this.get("_textureLoadedCounter") + 1);
        },

        /**
         * Load a url and return to the caller when data is ready
         */
        LoadURL: function (__sceneURL, __parentSceneURL)
        {
            const real_URL = this.get("_assetloader").get("_httprequestmanager").GetRealURL(__sceneURL, __parentSceneURL);
            if (!this.get("_urlCache").hasOwnProperty(real_URL))
            {
                this.IncrementQueryCounter();

                //Keeps track of previously cached URLs during that session so that there
                //is no duplicate and overheads in AssetManager cache construction
                this.get("_urlCache")[real_URL] = __parentSceneURL;
                this.get("_assetloader").LoadURL(real_URL, __parentSceneURL);
            }
        },

        /**
         * Load a resource file in JSON format and put its elements in the cache
         */
        LoadAsset: function (__resource, __sceneURL)
        {

            LogMessage.Log("Asset Manager LoadAsset function");

            this.get("_assetloader").SetResource(__resource);

            this.Printresourceinfo();
            const scene_name = this.Loadroot(__sceneURL);
            this.Loadpendingqueriedassets(__sceneURL);

            this.IncrementResolvedQueryCounter();
            return scene_name;
        },

        /**
         * Load a resource file in JSON format and put its elements in the cache
         */
        Loadroot: function (__sceneURL)
        {
            LogMessage.Log("Asset Manager LoadAsset function");

            const root = this.get("_assetloader").Getrootnode();
            //Replace root name if it is an url
            if (__sceneURL) root.AssetId = __sceneURL;
            if (!root.AssetId) root.AssetId = "Default_scene_name";
            if (!root.ServerBuildNumber) LogMessage.Log("Asset generated with ESKO HTML Viewer - Server Component version anterior to 691651");
            else LogMessage.Log("Asset generated with ESKO HTML Viewer - Server Component V" + root.ServerBuildNumber);

            const location = this.Getresourcelocation(root);
            if (!this.Iscached(root, location))
            {
                this.get("_cache").rootnodes[root.AssetId] = root;
                this.Pushinpendinglist(root, location);
            }
            else
            {
                LogMessage.Log("Root already chached");
            }

            return root.AssetId;
        },

        /**
         * Load queried assets
         */
        Loadpendingqueriedassets: function (__sceneURL)
        {
            LogMessage.Log("Asset Manager Recurse function");

            while (this.get("_pendingqueriedassets").length > 0)
            {
                //Get first element
                const elttofetch = this.get("_pendingqueriedassets")[0];
                //Remove it from pending list
                this.get("_pendingqueriedassets").splice(0, 1);
                //Add it to the cache
                this.Pushincache(elttofetch.res, elttofetch.loc);
                //Add its children to the pending list
                const list = this.get("_assetloader").GetRefChildrenFromName(elttofetch.res, __sceneURL);
                this.Pushlistinpendingqueue(list, __sceneURL);
            }
            LogMessage.Log("Size of pending list is: " + this.get("_pendingqueriedassets").length);
        },

        /**
         * Put list resources elements in the cache if not already in it
         */
        Pushlistinpendingqueue: function (list, __sceneURL)
        {
            for (let i = list.length - 1; i >= 0; i--)
            {
                const loc = this.Getresourcelocation(list[i]);
                if (!this.Isinpendinglist(list[i]) && !this.Iscached(list[i], loc))
                {
                    this.Pushinpendinglist(list[i], loc);
                }
                //Query referenced urls and push them as nodes
                if (this.get("_assetloader").IsURL(list[i].AssetRef))
                {
                    //Check if the current AssetRef is a URL and add it to the cache if needed
                    LogMessage.Log("Loading external asset from URL");
                    this.LoadURL(list[i].AssetRef, __sceneURL);
                }
            }
        },


        /**
         * Put a resource in the cache
         */
        Pushincache: function (resource, location)
        {
            const request_type = this.RequestType(resource);
            if (request_type === null)
            {
                // LogMessage.Log("No request type: object not cached");
                return false;
            }
            // Simple children are not added to the cache
            // But still need to be processed via the pending
            // query queue

            LogMessage.Log("AssetManager Pushincache function : Object \"" + resource[request_type] + "\"");
            location[resource[request_type]] = resource;
        },

        /**
         * Returns true if an asset is already loaded in the cache plausible location
         */
        Iscached: function (resource, plausiblelocation)
        {

            // LogMessage.Log("AssetManager Iscached function");
            const request_type = this.RequestType(resource);
            if (request_type === null) return false;

            for (const i in plausiblelocation)
            {
                if (plausiblelocation.hasOwnProperty(i))
                {
                    if (plausiblelocation[i][request_type] === resource[request_type])
                    {
                        // LogMessage.Log("Object \""+i+"\" already cached !");
                        return true;
                    }
                }
            }
        },

        /**
         * Put a resource in the cache
         */
        Pushinpendinglist: function (resource, location)
        {
            // LogMessage.Log("AssetManager PushinpendingList");
            this.get("_pendingqueriedassets").push({"res": resource, "loc": location});
        },

        /**
         * Put a resource in the cache
         */
        Isinpendinglist: function (resource)
        {
            // LogMessage.Log("AssetManager Isinpendinglist function");
            const request_type = this.RequestType(resource);
            if (request_type === null) return false;

            for (const i in this.get("_pendingqueriedassets"))
            {
                if (this.get("_pendingqueriedassets").hasOwnProperty(i))
                {
                    if (this.get("_pendingqueriedassets")[i].res[request_type] === resource[request_type])
                    {
                        // LogMessage.Log("Object \""+i+"\" already in pending list !");
                        return true;
                    }
                }
            }
            return false;
        },

        /**
         * Load a resource file in JSON format and put its elements in the cache
         */
        Printresourceinfo: function ()
        {
            this.get("_assetloader").Printresourceinfo();
        },
    });
});
