/**
 * @author DOGO
 * @module WebViewer3D
 * @copyright   2014 Esko Stonecube Ltd. All rights reserved.
 */

define([
	"underscore",
	"backbone",
	"@esko/3d-web-viewer/libs/cbor",
	"@esko/3d-web-viewer/modules/LogMessage",
	"@esko/3d-web-viewer/modules/CallbackWrapper",
	"@esko/3d-web-viewer/modules/Timers"
], function(_, bb, cbor, LogMessage, CallbackWrapper, WV3DTimers) {
	"use strict";

	/**
	 * Retrieve the file extension from a given __url
	 * Warning: if the url contains only http://www.esko.com#test=var
	 * This function will return "com". Therefore the caller must be sure that the url
	 * contains an extension or is aware of the risk
	 * @param {function} __url - the url that may contain an extension
	 */
	const GetURLExtension = function(__url) {
		return __url
			.split(/\#|\?/)[0]
			.split(".")
			.pop()
			.trim();
	};

	/**
	 * Call error callback
	 *
	 * @param {function} __callback - function to be called (with no arguments).
	 * @param __resourceURL
	 */
	const ErrorCallBack = function(__callback, __resourceURL) {
		try {
			console.warn("ERROR: The HTTP request failed for " + __resourceURL);
			CallbackWrapper(__callback)(__resourceURL);
		} catch (e) {
			LogMessage.Log(e);
		}
	};

	/**
	 * @alias module:WebViewer3D
	 * @class WV3DHttpRequestHandler
	 */
	return bb.Model.extend({
		/**
		 * Load a scene from a URL resource
		 *
		 * @param {String} __resourceURL - URL to the scene data. Internal url we use to reference our scene elements
		 * @param {function} __callback - Function to be called when the HTTP request is executed successfully.
		 *                                It accepts following parameters:
		 *                                  - A JSON object representing the resource returned by the HTTP call.
		 *                                  - The resource URL
		 * @param {function} __errorCallback - Function to be called when the HTTP request has failed (e.g. because the __resourceURL was invalid).
		 *                                     It accepts no parameters.
		 * @param {function} __urlPreprocessCallback - default callback function for intercepting all http requests
		 *                      arg[1] => the current url that the webviewer would use when this callback is undefined
		 *                      arg[2] => the query callback that must be called to query the file using the replacement URL as a parameter
		 *
		 */
		initialize: function(__resourceURL, __callback, __errorCallback, __urlPreprocessCallback) {
			LogMessage.Log("Creating WV3DHttpRequestHandler Object");

			const queryCallback = function(__url) {
				// Do the HTTP request
				// We use __url (External url provided by __urlPreprocessCallback) for http
				// requests, but we keep our internal url to store using __resourceURL

				if (GetURLExtension(__url) === "cbor") {
					//Resource target is CBOR encoded, therefore, we use XMLHttpRequest to query binary data
					const oReq = new XMLHttpRequest();
					oReq.open("GET", __url, true);
					oReq.responseType = "arraybuffer";

					oReq.onload = function(oEvent) {
						const arrayBuffer = oReq.response;
						try {
							WV3DTimers.reset("CBOR Decoding");
							const resource = cbor.decode(arrayBuffer);
							WV3DTimers.stop("CBOR Decoding");
							// When the returned resource is a string or encoded, it should first be converted to a JSON object.
							CallbackWrapper(__callback)(resource, __resourceURL);
						} catch (e) {
							LogMessage.Log(e);
							ErrorCallBack(__errorCallback, __resourceURL);
						}
					};

					oReq.onerror = function(oEvent) {
						ErrorCallBack(__errorCallback, __resourceURL);
					};

					oReq.send();
				} else {
					//Standard ajax call
					$.ajax({
						url: __url
					})
						.done(function(__resource) {
							try {
								// When the returned resource is a string or encoded, it should first be converted to a JSON object.
								const resource_object = typeof __resource === "string" ? JSON.parse(__resource) : __resource;
								CallbackWrapper(__callback)(resource_object, __resourceURL);
							} catch (e) {
								LogMessage.Log(e);
								ErrorCallBack(__errorCallback, __resourceURL);
							}
						})
						.fail(function() {
							ErrorCallBack(__errorCallback, __resourceURL);
						});
				}
			};

			if (__urlPreprocessCallback) {
				__urlPreprocessCallback(__resourceURL, queryCallback);
			} else {
				queryCallback(__resourceURL);
			}
		}
	});
});
