/**
 * @author DOGO
 * @module WebViewer3D
 * @copyright 2017 Esko Software BVBA. All rights reserved.
 */

define([
    "underscore",
    "backbone",
    "three",
    "@esko/3d-web-viewer/modules/LogMessage",
    "@esko/3d-web-viewer/modules/SystemChecker",
    "@esko/3d-web-viewer/modules/Math",
    "@esko/3d-web-viewer/modules/HttpRequestHandler",
    "@esko/3d-web-viewer/modules/Utils",
    "@esko/3d-web-viewer/modules/shaders/ShaderUtils",
    "@esko/3d-web-viewer/modules/shaders/AlphaFixShader",
    "@esko/3d-web-viewer/modules/shaders/DesignerShader",
    "@esko/3d-web-viewer/modules/shaders/D0Shader",
    "@esko/3d-web-viewer/modules/shaders/D1Shader",
    "@esko/3d-web-viewer/modules/shaders/D2Shader",
    "@esko/3d-web-viewer/modules/shaders/DepthOnlyShader",
    "@esko/3d-web-viewer/modules/shaders/BackdropShader",
    "@esko/3d-web-viewer/modules/shaders/ColorBlendShader",
    "@esko/3d-web-viewer/modules/shaders/CopyShader",
    "@esko/3d-web-viewer/modules/shaders/LineBasicShader",
    "@esko/3d-web-viewer/modules/shaders/UniformColorShader",
    "@esko/3d-web-viewer/modules/shaders/EdgeShader2",//Choose between EdgeShader, EdgeShader2, EdgeShader3 (EdgeShader3 seems to be the less expensive with better results)
    "@esko/3d-web-viewer/modules/shaders/TransparencyBlendShader",
    "@esko/3d-web-viewer/modules/shaders/FilterShaderPM",
    "@esko/3d-web-viewer/modules/shaders/FilterShaderNPM",
    "@esko/3d-web-viewer/modules/shaders/ReflectanceShaderPM",
    "@esko/3d-web-viewer/modules/shaders/ReflectanceShaderNPM",
    "@esko/3d-web-viewer/modules/shaders/IrradianceShaderPM",
    "@esko/3d-web-viewer/modules/shaders/IrradianceShaderNPM",
    "@esko/3d-web-viewer/modules/shaders/DiffractionShaderPM"
], function (
    _,
    bb,
    THREE,
    LogMessage,
    WV3DSystemChecker,
    WV3DMath,
    WV3DHttpRequestHandler,
    WV3DUtils,
    ShaderUtils,
    AlphaFixShader,
    DesignerShader,
    D0Shader,
    D1Shader,
    D2Shader,
    DepthOnlyShader,
    BackdropShader,
    ColorBlendShader,
    CopyShader,
    LineBasicShader,
    UniformColorShader,
    EdgeShader,
    TransBlendShader,
    FilterShaderPM,
    FilterShaderNPM,
    ReflectanceShaderPM,
    ReflectanceShaderNPM,
    IrradianceShaderPM,
    IrradianceShaderNPM,
    DiffractionShaderPM
)
{

    "use strict";

    /* PRIVATE */
    const GetTextureScaling = function (__mat_struct)
    {
        return new THREE.Vector2(__mat_struct["dUdP"] || 1, __mat_struct["dVdP"] || 1);
    };

    const ReleaseMaterial = function (__material, __destroy_textures)
    {
        //Dispose related material and all related uniforms
        if (__material)
        {
            __material.dispose();
            if (__material.uniforms)
            {
                for (let mat_uni_idx = 0, keys = Object.keys(__material.uniforms), l = keys.length;
                     mat_uni_idx < l; ++mat_uni_idx)
                {
                    const uniform = __material.uniforms[keys[mat_uni_idx]];
                    if (uniform["value"])
                    {
                        //We want to dispose of any uniform that is not a texture
                        //and has a dispose function, in any case
                        //In addition, we want to release textures uniforms if
                        //__destroy_textures is true (or default)
                        if ((typeof(uniform["value"]["dispose"]) === "function") &&
                            ((__destroy_textures === undefined) || __destroy_textures || !(uniform["value"] instanceof THREE.Texture)))
                        {
                            uniform["value"].dispose();
                        }
                    }
                }
            }
        }
    };

    const ReleaseMaterialsAndTextures = function (__mat_struct)
    {
        if (__mat_struct)
        {
            const meshestoupdate = __mat_struct["boundmeshlist"];
            //Update material reference in meshes that are already bound to this material
            if (meshestoupdate)
            {
                if (Object.keys(meshestoupdate).length === 0) LogMessage.Log("ViewerView Warning: No meshes to update");
                for (let j = 0, keys1 = Object.keys(meshestoupdate), L = keys1.length; j < L; ++j)
                {
                    const renderpass = keys1[j];
                    if (Object.keys(meshestoupdate[renderpass]).length === 0) LogMessage.Log("ViewerView Warning: No render passes to update");
                    for (let i = 0, keys2 = Object.keys(meshestoupdate[renderpass]), l = keys2.length; i < l; ++i)
                    {
                        const side_mat = keys2[i];
                        for (let spass_cpt = 0, spass_keys = Object.keys(meshestoupdate[renderpass][side_mat]), l_spass = spass_keys.length;
                             spass_cpt < l_spass; spass_cpt++)
                        {
                            const spass_id = spass_keys[spass_cpt];
                            ReleaseMaterial(__mat_struct["mat"]["untextured_mat"][renderpass][spass_id]);
                            if (Object.keys(meshestoupdate[renderpass][side_mat][spass_id]).length === 0) LogMessage.Log("ViewerView Warning: No sided materials to update");
                            for (let k = 0, keys3 = Object.keys(meshestoupdate[renderpass][side_mat][spass_id]), m = keys3.length; k < m; ++k)
                            {
                                const surfID = keys3[k];
                                //Release material and textures
                                ReleaseMaterial(__mat_struct["mat"]["textured_mat"][renderpass][spass_id][surfID]);
                                delete __mat_struct["mat"]["textured_mat"][renderpass][spass_id][surfID];
                                delete meshestoupdate[renderpass][side_mat][spass_id][surfID];
                            }
                            delete meshestoupdate[renderpass][side_mat][spass_id];
                        }
                        delete meshestoupdate[renderpass][side_mat];
                        delete meshestoupdate[renderpass];
                    }
                }
                delete __mat_struct["boundmeshlist"];
            }
            else
            {
                LogMessage.Log("ViewerView Warning: No meshes to update");
            }

            if (__mat_struct["diff_texture"]) __mat_struct["diff_texture"].dispose();
            if (__mat_struct["spec_texture"]) __mat_struct["spec_texture"].dispose();
            if (__mat_struct["int_spec_texture"]) __mat_struct["int_spec_texture"].dispose();
            if (__mat_struct["normal_map_texture"]) __mat_struct["normal_map_texture"].dispose();
            if (__mat_struct["diff_grating_texture"]) __mat_struct["diff_grating_texture"].dispose();
            if (__mat_struct["diff_grating_filter_texture"]) __mat_struct["diff_grating_filter_texture"].dispose();
            if (__mat_struct["annotationsTexture"]) __mat_struct["annotationsTexture"].dispose();

        }
    };

    const ReleaseScenesMaterialsAndTextures = function (__rendererCache)
    {
        for (let n = 0, keys = Object.keys(__rendererCache.get("materials")["materiallist"]), o = keys.length; n < o; ++n)
        {
            const mat_structs = __rendererCache.get("materials")["materiallist"][keys[n]];
            ReleaseMaterialsAndTextures(mat_structs.front);
            ReleaseMaterialsAndTextures(mat_structs.back);
        }

        if (__rendererCache.get("textures")["Designer"]["irradiancemap"]["data"])
            __rendererCache.get("textures")["Designer"]["irradiancemap"]["data"].dispose();
        if (__rendererCache.get("textures")["Designer"]["envmap"]["glossy"]["data"])
            __rendererCache.get("textures")["Designer"]["envmap"]["glossy"]["data"].dispose();
        if (__rendererCache.get("textures")["Designer"]["envmap"]["matte"]["data"])
            __rendererCache.get("textures")["Designer"]["envmap"]["matte"]["data"].dispose();

        if (__rendererCache.get("textures")["Visualizer"]["irradiancemap"]["data"])
            __rendererCache.get("textures")["Visualizer"]["irradiancemap"]["data"].dispose();
        if (__rendererCache.get("textures")["Visualizer"]["envmap"]["data"])
            for (let envmap_id = 0; envmap_id < __rendererCache.get('textures')["Visualizer"]["envmap"]["data"].length; envmap_id++)
                if (__rendererCache.get("textures")["Visualizer"]["envmap"]["data"][envmap_id])
                    __rendererCache.get("textures")["Visualizer"]["envmap"]["data"][envmap_id].dispose();
        if (__rendererCache.get("textures")["Visualizer"]["spectrum"]["data"])
            __rendererCache.get("textures")["Visualizer"]["spectrum"]["data"].dispose();

        if (__rendererCache.get("textures")["backdrop"]["data"])
            __rendererCache.get("textures")["backdrop"]["data"].dispose();
    };

    const ApplyFctToMaterialStruct = function (__mat_struct, __callback)
    {
        for (let n = 0, keys0 = Object.keys(__mat_struct["mat"]["untextured_mat"]), o = keys0.length; n < o; ++n)
        {
            const pass = __mat_struct["mat"]["untextured_mat"][keys0[n]];
            for (let m = 0, keys1 = Object.keys(pass), p = keys1.length; m < p; ++m)
            {
                const material = pass[keys1[m]];
                __callback(material);
                material.needsUpdate = true;
            }
        }
        for (let n = 0, keys0 = Object.keys(__mat_struct["mat"]["textured_mat"]), o = keys0.length; n < o; ++n)
        {
            const pass = __mat_struct["mat"]["textured_mat"][keys0[n]];
            for (let m = 0, keys1 = Object.keys(pass), p = keys1.length; m < p; ++m)
            {
                const meshID = pass[keys1[m]];
                for (let l = 0, keys2 = Object.keys(meshID), q = keys2.length; l < q; ++l)
                {
                    const material = meshID[keys2[l]];
                    __callback(material);
                    material.needsUpdate = true;
                }
            }
        }
    };

    const InitializeShaderUniforms = function(__mat_struct, __material_options, __shadername, __uniforms)
    {
        //TEXTURE OFFSET AND SCALE
        const uvScaleMap = __mat_struct["diff_texture"] || __mat_struct["spec_texture"];
        if (uvScaleMap)
        {
            const offset = uvScaleMap.offset;
            const repeat = uvScaleMap.repeat;
            __uniforms["offsetRepeat"]["value"] = new THREE.Vector4(offset.x, offset.y, repeat.x, repeat.y);
        }
        else
        {
            __uniforms["offsetRepeat"]["value"] = new THREE.Vector4(0, 0, 1, 1);
        }

        switch (__shadername)
        {
            case "designer":
                __uniforms["diffuse"]["value"] = __mat_struct["diff_col"];
                __uniforms["opacity"]["value"] = __material_options.opaque ? 1 : __mat_struct["opacity"];
                __uniforms["specular"]["value"] = __mat_struct["spec_col"];
                __uniforms["shininess"]["value"] = Math.max(0.001, __mat_struct["shininess"]);
                __uniforms["map"]["value"] = __mat_struct["diff_texture"];
                __uniforms["specularMap"]["value"] = __mat_struct["spec_texture"];
                __uniforms["reflectivity"]["value"] = 0.5;
                __uniforms["irr_tex"]["value"] = __material_options.irrmap;
                __uniforms["envMap"]["value"] = __material_options.envmap;
                break;

            case "irradiancepm":
                __uniforms["diffuseColor"]["value"] = __mat_struct["diff_col"];
                __uniforms["opacity"]["value"] = __material_options.opaque ? 1 : __mat_struct["opacity"];
                __uniforms["diffuseTexture"]["value"] = __mat_struct["diff_texture"] ? __mat_struct["diff_texture"] : null;
                __uniforms["normalMap"]["value"] = __mat_struct["normal_map_texture"] ? __mat_struct["normal_map_texture"] : null;
                __uniforms["DX1RT"]["value"] = null;
                __uniforms["DX2RT"]["value"] = null;
                __uniforms["irradianceCubeMap"]["value"] = __material_options.irrmap;
                __uniforms["backTexture"]["value"] = __mat_struct["backTexture"] ? 1 : 0;
                break;

            case "irradiancenpm":
                __uniforms["diffuseColor"]["value"] = __mat_struct["diff_col"];
                __uniforms["opacity"]["value"] = __material_options.opaque ? 1 : __mat_struct["opacity"];
                __uniforms["diffuseTexture"]["value"] = __mat_struct["diff_texture"] ? __mat_struct["diff_texture"] : null;
                __uniforms["DX1RT"]["value"] = null;
                __uniforms["DX2RT"]["value"] = null;
                __uniforms["irradianceCubeMap"]["value"] = __material_options.irrmap;
                __uniforms["backTexture"]["value"] = __mat_struct["backTexture"] ? 1 : 0;
                break;

            case "filterpm":
                __uniforms["diffuseColor"]["value"] = __mat_struct["diff_col"];
                __uniforms["opacity"]["value"] = __material_options.opaque ? 1 : __mat_struct["opacity"];
                __uniforms["diffuseTexture"]["value"] = __mat_struct["diff_texture"] ? __mat_struct["diff_texture"] : null;
                __uniforms["DX1RT"]["value"] = null;
                __uniforms["DX2RT"]["value"] = null;
                __uniforms["backTexture"]["value"] = __mat_struct["backTexture"] ? 1 : 0;
                break;

            case "filternpm":
                __uniforms["diffuseColor"]["value"] = __mat_struct["diff_col"];
                __uniforms["opacity"]["value"] = __material_options.opaque ? 1 : __mat_struct["opacity"];
                __uniforms["diffuseTexture"]["value"] = __mat_struct["diff_texture"] ? __mat_struct["diff_texture"] : null;
                __uniforms["DX1RT"]["value"] = null;
                __uniforms["DX2RT"]["value"] = null;
                __uniforms["backTexture"]["value"] = __mat_struct["backTexture"] ? 1 : 0;
                break;

            case "reflectancepm":
                if (WV3DSystemChecker.get("_glmaxtextureimageunits") >= 11)
                {
                    __uniforms["haze1CubeMap"]["value"] = __material_options.envmap ? __material_options.envmap[0] : null;
                    __uniforms["haze2CubeMap"]["value"] = __material_options.envmap ? __material_options.envmap[1] : null;
                    __uniforms["haze3CubeMap"]["value"] = __material_options.envmap ? __material_options.envmap[2] : null;
                    __uniforms["haze4CubeMap"]["value"] = __material_options.envmap ? __material_options.envmap[3] : null;
                    __uniforms["haze5CubeMap"]["value"] = __material_options.envmap ? __material_options.envmap[4] : null;
                }
                else
                {
                    __uniforms["haze1CubeMap"]["value"] = __material_options.envmap ? __material_options.envmap[0] : null;
                    __uniforms["haze2CubeMap"]["value"] = __material_options.envmap ? __material_options.envmap[4] : null;
                }
                __uniforms["specularColor"]["value"] = __mat_struct["spec_col"];
                __uniforms["shininess"]["value"] = __mat_struct["shininess"];
                __uniforms["specularTexture"]["value"] = __mat_struct["spec_texture"] ? __mat_struct["spec_texture"] : null;
                __uniforms["internalSpecTex"]["value"] = __mat_struct["int_spec_texture"] ? __mat_struct["int_spec_texture"] : null;
                __uniforms["normalMap"]["value"] = __mat_struct["normal_map_texture"] ? __mat_struct["normal_map_texture"] : null;
                __uniforms["DX1RT"]["value"] = null;
                __uniforms["DX2RT"]["value"] = null;
                __uniforms["backTexture"]["value"] = __mat_struct["backTexture"] ? 1 : 0;
                break;

            case "reflectancenpm":
                if (WV3DSystemChecker.get("_glmaxtextureimageunits") >= 11)
                {
                    __uniforms["haze1CubeMap"]["value"] = __material_options.envmap ? __material_options.envmap[0] : null;
                    __uniforms["haze2CubeMap"]["value"] = __material_options.envmap ? __material_options.envmap[1] : null;
                    __uniforms["haze3CubeMap"]["value"] = __material_options.envmap ? __material_options.envmap[2] : null;
                    __uniforms["haze4CubeMap"]["value"] = __material_options.envmap ? __material_options.envmap[3] : null;
                    __uniforms["haze5CubeMap"]["value"] = __material_options.envmap ? __material_options.envmap[4] : null;
                }
                else
                {
                    __uniforms["haze1CubeMap"]["value"] = __material_options.envmap ? __material_options.envmap[0] : null;
                    __uniforms["haze2CubeMap"]["value"] = __material_options.envmap ? __material_options.envmap[4] : null;
                }
                __uniforms["DX1RT"]["value"] = null;
                __uniforms["DX2RT"]["value"] = null;
                __uniforms["specularCol"]["value"] = __mat_struct["spec_col"];
                __uniforms["shininess"]["value"] = __mat_struct["shininess"];
                __uniforms["backTexture"]["value"] = __mat_struct["backTexture"] ? 1 : 0;
                break;

            case "holographicspm":
                __uniforms["uvdhTexture"]["value"] = __mat_struct["diff_grating_texture"] ? __mat_struct["diff_grating_texture"] : null;
                __uniforms["diffFilterTexture"]["value"] = __mat_struct["diff_grating_filter_texture"] ? __mat_struct["diff_grating_filter_texture"] : null;
                __uniforms["normalMap"]["value"] = __mat_struct["normal_map_texture"] ? __mat_struct["normal_map_texture"] : null;
                __uniforms["DX1RT"]["value"] = null;
                __uniforms["DX2RT"]["value"] = null;
                __uniforms["grating_spacing_scale"]["value"] = __mat_struct["grating_spacing_scale"];

                __uniforms["spectrumTexture"]["value"] = __material_options.textures["Visualizer"]["spectrum"]["data"];
                __uniforms["defaultDiffSpacing"]["value"] = __mat_struct["default_diff_spacing"];
                __uniforms["defaultDiffHaze"]["value"] = __mat_struct["default_diff_haze"];
                __uniforms["backTexture"]["value"] = __mat_struct["backTexture"] ? 1 : 0;

                if (__material_options.textures["Visualizer"]["spectrum"]["loaded"] && __material_options.lights.hasOwnProperty("holo_light"))
                {
                    __uniforms["lightPos"]["value"] = [];
                    __uniforms["lightIntensity"]["value"] = [];
                    __uniforms["lightRadius"]["value"] = [];
                    __uniforms["lightHardness"]["value"] = [];
                    for (let i = 0; i < __material_options.lights.holo_light.Lights.length && (i < 4); ++i)
                    {
                        // holographics_light = [ x, y, z, intensity, radius, hardness, spectrumIndex ]
                        const holographics_light = __material_options.lights.holo_light.Lights[i];
                        __uniforms["lightPos"]["value"].push(new THREE.Vector3(holographics_light[0], holographics_light[1], holographics_light[2]));
                        __uniforms["lightIntensity"]["value"].push(holographics_light[3]);
                        __uniforms["lightRadius"]["value"].push(holographics_light[4]);
                        __uniforms["lightHardness"]["value"].push(holographics_light[5]);
                    }
                }
                break;

            case "transparentblend":
                __uniforms["diffuseColor"]["value"] = __mat_struct["diff_col"];
                __uniforms["opacity"]["value"] = __mat_struct["opacity"];
                __uniforms["diffuseTexture"]["value"] = __mat_struct["diff_texture"] ? __mat_struct["diff_texture"] : null;
                __uniforms["DX1RT"]["value"] = null;
                __uniforms["DX2RT"]["value"] = null;
                __uniforms["backTexture"]["value"] = __mat_struct["backTexture"] ? 1 : 0;
                break;

            default:
                LogMessage.Error("Unknown shader type");

        }
    };

    const InitializeCustomShaderUniforms = function (__mat_struct, __uniforms)
    {
        if (__mat_struct["shader"])
        {
            const shader = __mat_struct["shader"];
            switch (shader["type"])
            {
                case "Wrap":
                    if (shader["parameters"].hasOwnProperty("strip_size"))
                        __uniforms["stripSize"]["value"] = shader["parameters"]["strip_size"];
                    if (shader["parameters"].hasOwnProperty("gap_size"))
                        __uniforms["gapSize"]["value"] = shader["parameters"]["gap_size"];
                    break;
                default:
                    LogMessage.Error("Unknown custom shader type");

            }
        }
    };

    const InitializeShaderDefines = function(__mat_struct, __material_options, __shadername, __defines)
    {
        switch (__shadername)
        {
            case "designer":
                __defines["ENVMAP_TYPE_CUBE"] = "";
                __defines["ENVMAP_MODE_REFLECTION"] = "";
                __defines["ENVMAP_BLENDING_ADD"] = "";

                __defines["ANNOTATION_NUMBER"] = "0";
                __defines["ANNOTATION_TEXTURE_SIZE"] = "0";
                __defines["MAX_LINES_PER_CELL"] = "1";
                __defines["ACCEL_GRID_SIZE"] = "0";
                __defines["LINES_START_OFFSET"] = "0";
                break;

            case "irradiancepm":
                if (__mat_struct["normal_map_texture"])
                    __defines["USE_NORMALMAP_TEXTURE"] = "";
                break;

            case "irradiancenpm"://No defines
                break;

            case "filterpm"://No defines
                break;

            case "filternpm"://No defines
                break;

            case "reflectancepm":
                if (__mat_struct["int_spec_texture"])
                    __defines["USE_INT_SPEC_TEXTURE"] = "";
                if (__mat_struct["normal_map_texture"])
                    __defines["USE_NORMALMAP_TEXTURE"] = "";

                __defines["ANNOTATION_NUMBER"] = "0";
                __defines["ANNOTATION_TEXTURE_SIZE"] = "0";
                __defines["MAX_LINES_PER_CELL"] = "1";
                __defines["ACCEL_GRID_SIZE"] = "0";
                __defines["LINES_START_OFFSET"] = "0";
                break;

            case "reflectancenpm":
                break;

            case "holographicspm":
                if (__mat_struct["double_grating"])
                    __defines["DOUBLE_GRATING"] = "";
                if (__mat_struct["normal_map_texture"])
                    __defines["USE_NORMALMAP_TEXTURE"] = "";

                if (__material_options.textures["Visualizer"]["spectrum"]["loaded"] && __material_options.lights.hasOwnProperty("holo_light"))
                {
                    __defines["USE_HOLOGRAPHICS"] = "";
                }
                break;

            case "common":

                //COMMON DEFINES for DESIGNER, VISUALISERPM and VISUALISERNPM
                //The first 5 defines names are the same that are used internally by three.js
                //so we keep them as they are for comparison/understanding purposes.
                //However, we could change them as long as we change the corresponding
                //defines names in DesignerShader.js, FilterShaderPM/NPM.js, IrradianceShaderPM/NPM.js,
                // ReflectanceShaderPM/NPM.js, DiffractionShaderPM.js, etc.

                if (__mat_struct["double_sided"])
                    __defines["DOUBLE_SIDED"] = "";

                if (__mat_struct["diff_texture"])
                    __defines["USE_MAP"] = "";

                if (__mat_struct["spec_texture"])
                    __defines["USE_SPECULARMAP"] = "";

                if (__material_options.irrmap)
                    __defines["USE_IRRMAP"] = "";

                if (__material_options.envmap)
                    __defines["USE_ENVMAP"] = "";

                //////////////////

                if (__mat_struct["diff_texture"] && __mat_struct["diff_texture"]["remappedAlphaChannel"])
                    __defines["REMAPPED_DIFFUSE_ALPHA_CHANNEL"] = "";

                if (__mat_struct["spec_texture"] && __mat_struct["spec_texture"]["remappedAlphaChannel"])
                    __defines["REMAPPED_SPECULAR_ALPHA_CHANNEL"] = "";

                if (__mat_struct["int_spec_texture"] && __mat_struct["int_spec_texture"]["remappedAlphaChannel"])
                    __defines["REMAPPED_INT_SPECULAR_ALPHA_CHANNEL"] = "";

                if (__mat_struct["diff_grating_texture"] && __mat_struct["diff_grating_texture"]["remappedAlphaChannel"])
                    __defines["REMAPPED_DIFF_GRATING_ALPHA_CHANNEL"] = "";

                if (__mat_struct["shader"] && __mat_struct["shader"]["type"] === "Wrap")
                    __defines["WRAP_MATERIAL"] = "1";

                break;

            case "transparentblend":

                __defines["ANNOTATION_NUMBER"] = "0";
                __defines["ANNOTATION_TEXTURE_SIZE"] = "0";
                __defines["MAX_LINES_PER_CELL"] = "1";
                __defines["ACCEL_GRID_SIZE"] = "0";
                __defines["LINES_START_OFFSET"] = "0";

                if (__mat_struct["diff_texture"])
                    __defines["USE_MAP"] = "";

                if (!__mat_struct["printable"])
                    __defines["IS_NPM"] = "";

                if (__mat_struct["diff_texture"] && __mat_struct["diff_texture"]["remappedAlphaChannel"])
                    __defines["REMAPPED_DIFFUSE_ALPHA_CHANNEL"] = "";

                if (__mat_struct["spec_texture"] && __mat_struct["spec_texture"]["remappedAlphaChannel"])
                    __defines["REMAPPED_SPECULAR_ALPHA_CHANNEL"] = "";

                if (__mat_struct["int_spec_texture"] && __mat_struct["int_spec_texture"]["remappedAlphaChannel"])
                    __defines["REMAPPED_INT_SPECULAR_ALPHA_CHANNEL"] = "";

                if (__mat_struct["diff_grating_texture"] && __mat_struct["diff_grating_texture"]["remappedAlphaChannel"])
                    __defines["REMAPPED_DIFF_GRATING_ALPHA_CHANNEL"] = "";

                break;
        }
    };

    /**
     * @alias module:WebViewer3D
     * @class WV3DMaterialManager
     * This class is responsible for creating the 3JS materials for each render pass (generic shaders created in InitializeShaders)
     * and for each colored render pass, which is per-object specific (customized shaders created in CreateMaterial)
     * It also uses _texturemanager to query the textures needed for the matching materials.
     */
    return bb.Model.extend({

        /* PUBLIC */

        initialize: function (__rendererCache, __assetManager, __textureManager, __executionQueue)
        {
            LogMessage.Log("Creating WV3DMaterialManager Object");
            //IMPROVEME: We could consider using singletons for some of these attributes
            this.set("_renderercache", __rendererCache);
            this.set("_assetmanager", __assetManager);
            this.set("_texturemanager", __textureManager);
            this.set("_exec_queue", __executionQueue);

            this.listenTo(__textureManager, "updatematerialsgloballighting", this.UpdateMaterialsGlobalLighting);
            this.listenTo(__textureManager, "updatetexturedmaterial", this.UpdateTexturedMaterial);
            this.listenTo(__textureManager, "updatematerialtexture", this.UpdateMaterialTexture);
        },

        Reset: function ()
        {
            this.get("_texturemanager").Reset();
        },

        InitializeShaders: function ()
        {
            //Set up default material shader (fallback material)
            if (!this.get("_renderercache").get("materials")["default_material"])
            {
                const designer_uniforms = THREE.UniformsUtils.clone(DesignerShader.uniforms);
                designer_uniforms["diffuse"]["value"] = new THREE.Vector3(1, 1, 1);
                designer_uniforms["specular"]["value"] = new THREE.Vector3(0, 0, 0);

                const enablePolygonOffset = this.get("_renderercache").get("materials")["enablehighlights"] && this.get("_renderercache").get("materials")["enablehardedgehighlights"];
                this.get("_renderercache").get("materials")["default_material"] = new THREE.ShaderMaterial({
                    "name": "designer",
                    "uniforms": designer_uniforms,
                    "vertexShader": DesignerShader.vertexShader,
                    "fragmentShader": DesignerShader.fragmentShader,
                    "depthTest": true,
                    "vertexColors": THREE.FaceColors,
                    "side": THREE.FrontSide,
                    "transparent": false,
                    "defines": {
                        "ENVMAP_TYPE_CUBE": "",
                        "ENVMAP_MODE_REFLECTION": "",
                        "ENVMAP_BLENDING_ADD": "",
                        "ANNOTATION_NUMBER": "0",
                        "ANNOTATION_TEXTURE_SIZE": "0",
                        "LINES_START_OFFSET": "0"
                    },
                    "lights": true,
                    "polygonOffset": true,
                    "polygonOffsetFactor": enablePolygonOffset ? this.get("_renderercache").get("_polygon_offset_factor") : 0,
                    "polygonOffsetUnits": enablePolygonOffset ? this.get("_renderercache").get("_polygon_offset_units") : 0
                });
            }

            //Set up depth peeling shaders
            if (!this.get("_renderercache").get("materials")["D0ShaderMaterial"])
            {
                const D0_uniforms = THREE.UniformsUtils.clone(D0Shader.uniforms);
                this.get("_renderercache").get("materials")["D0ShaderMaterial"] = new THREE.ShaderMaterial({
                    "name": "D0ShaderMaterial",
                    "uniforms": D0_uniforms,
                    "vertexShader": D0Shader.vertexShader,
                    "fragmentShader": D0Shader.fragmentShader,
                    "side": THREE.FrontSide,
                    "depthWrite": true,
                    "transparent": false
                });
            }

            if (!this.get("_renderercache").get("materials")["D1ShaderMaterial"])
            {
                const D1_uniforms = THREE.UniformsUtils.clone(D1Shader.uniforms);
                D1_uniforms["D0RT"]["value"] = null;
                this.get("_renderercache").get("materials")["D1ShaderMaterial"] = new THREE.ShaderMaterial({
                    "name": "D1ShaderMaterial",
                    "uniforms": D1_uniforms,
                    "vertexShader": D1Shader.vertexShader,
                    "fragmentShader": D1Shader.fragmentShader,
                    "side": THREE.FrontSide,
                    "depthWrite": true,
                    "transparent": false
                });
            }

            if (!this.get("_renderercache").get("materials")["D2ShaderMaterial"])
            {
                const D2_uniforms = THREE.UniformsUtils.clone(D2Shader.uniforms);
                D2_uniforms["DX1RT"]["value"] = null;
                D2_uniforms["DX2RT"]["value"] = null;
                this.get("_renderercache").get("materials")["D2ShaderMaterial"] = new THREE.ShaderMaterial({
                    "name": "D2ShaderMaterial",
                    "uniforms": D2_uniforms,
                    "vertexShader": D2Shader.vertexShader,
                    "fragmentShader": D2Shader.fragmentShader,
                    "side": THREE.FrontSide,
                    "depthWrite": true,
                    "transparent": false
                });
            }

            if (!this.get("_renderercache").get("materials")["DepthOnlyShaderMaterial"])
            {
                const depth_only_uniforms = THREE.UniformsUtils.clone(DepthOnlyShader.uniforms);
                depth_only_uniforms["DX1RT"]["value"] = null;
                depth_only_uniforms["DX2RT"]["value"] = null;
                this.get("_renderercache").get("materials")["DepthOnlyShaderMaterial"] = new THREE.ShaderMaterial({
                    "name": "DepthOnlyShaderMaterial",
                    "uniforms": depth_only_uniforms,
                    "vertexShader": DepthOnlyShader.vertexShader,
                    "fragmentShader": DepthOnlyShader.fragmentShader,
                    "side": THREE.FrontSide,
                    "depthWrite": true,
                    "colorWrite": false,
                    "transparent": false
                });
            }

            if (!this.get("_renderercache").get("materials")["BackdropShaderMaterial"])
            {
                const Backdrop_uniforms = THREE.UniformsUtils.clone(BackdropShader.uniforms);
                Backdrop_uniforms["backdrop"]["value"] = null;
                Backdrop_uniforms["reflection"]["value"] = null;
                this.get("_renderercache").get("materials")["BackdropShaderMaterial"] = new THREE.ShaderMaterial({
                    "name": "BackdropShaderMaterial",
                    "uniforms": Backdrop_uniforms,
                    "vertexShader": BackdropShader.vertexShader,
                    "fragmentShader": BackdropShader.fragmentShader,
                    "side": THREE.FrontSide,
                    "depthWrite": false,
                    "transparent": false
                });
            }

            if (!this.get("_renderercache").get("materials")["ColorBlendShaderMaterial"])
            {
                const ColorBlend_uniforms = THREE.UniformsUtils.clone(ColorBlendShader.uniforms);
                ColorBlend_uniforms["CX1RT"]["value"] = null;
                ColorBlend_uniforms["TransparencyStack"]["value"] = null;
                ColorBlend_uniforms["OpaqueRT"]["value"] = null;
                ColorBlend_uniforms["ColorStackRT"]["value"] = null;
                this.get("_renderercache").get("materials")["ColorBlendShaderMaterial"] = new THREE.ShaderMaterial({
                    "name": "ColorBlendShaderMaterial",
                    "uniforms": ColorBlend_uniforms,
                    "vertexShader": ColorBlendShader.vertexShader,
                    "fragmentShader": ColorBlendShader.fragmentShader,
                    "side": THREE.FrontSide,
                    "depthWrite": true,
                    "transparent": true,
                    "blending": THREE.AdditiveBlending
                });
            }

            if (!this.get("_renderercache").get("materials")["CopyShaderMaterial"])
            {
                const Copy_uniforms = THREE.UniformsUtils.clone(CopyShader.uniforms);
                Copy_uniforms["CXRT"]["value"] = null;
                Copy_uniforms["OpaqueRT"]["value"] = null;
                this.get("_renderercache").get("materials")["CopyShaderMaterial"] = new THREE.ShaderMaterial({
                    "name": "CopyShaderMaterial",
                    "uniforms": Copy_uniforms,
                    "vertexShader": CopyShader.vertexShader,
                    "fragmentShader": CopyShader.fragmentShader,
                    "side": THREE.FrontSide,
                    "depthWrite": true,
                    "transparent": false,
                    "blending": THREE.AdditiveBlending
                });
            }

            //Set up other shaders...
            this.ResetOutlineShaders();
        },

        ResetOutlineShaders: function ()
        {
            //Set up hard edges highlights materials
            const materials = this.get("_renderercache").get("materials");
            if (!materials["highlight"])
            {
                const basicLineUniforms = THREE.UniformsUtils.clone(LineBasicShader.uniforms);
                const color = this.get("_renderercache").get("materials")["defaultcolor"];
                basicLineUniforms["diffuse"]["value"] = new THREE.Vector3(color[0] / 255, color[1] / 255, color[2] / 255);
                materials["highlight"] = new THREE.ShaderMaterial({
                    "name": "ColoredLineBasicMaterialInstanced",
                    "uniforms": basicLineUniforms,
                    "vertexShader": LineBasicShader.vertexShader,
                    "fragmentShader": LineBasicShader.fragmentShader,
                    "side": THREE.DoubleSide,
                    "linewidth": 1,
                    "depthTest": true
                });
            }

            if (!materials["blackhighlight"])
            {
                const basicLineUniforms = THREE.UniformsUtils.clone(LineBasicShader.uniforms);
                basicLineUniforms["diffuse"]["value"] = new THREE.Vector3(0, 0, 0);
                materials["blackhighlight"] = new THREE.ShaderMaterial({
                    "name": "BlackLineBasicMaterialInstanced",
                    "uniforms": basicLineUniforms,
                    "vertexShader": LineBasicShader.vertexShader,
                    "fragmentShader": LineBasicShader.fragmentShader,
                    "side": THREE.DoubleSide,
                    "linewidth": 1,
                    "depthTest": true
                });
            }

            //Set up silhouette/horizon highlights materials

            if (!materials["colordepth"])
            {
                const colordepth_uniforms = THREE.UniformsUtils.clone(UniformColorShader.uniforms);
                materials["colordepth"] = new THREE.ShaderMaterial({
                    "name": "ColorDepth",
                    "uniforms": colordepth_uniforms,
                    "vertexShader": UniformColorShader.vertexShader,
                    "fragmentShader": UniformColorShader.fragmentShader,
                    "side": THREE.DoubleSide
                });
            }


            if (!materials["alphafix"])
            {
                materials["alphafix"] = new THREE.ShaderMaterial({
                    "name": "alphafix",
                    "uniforms": THREE.UniformsUtils.clone(AlphaFixShader.uniforms),
                    "vertexShader": AlphaFixShader.vertexShader,
                    "fragmentShader": AlphaFixShader.fragmentShader,
                    "depthFunc": THREE.GreaterDepth
                });
            }

            if (!materials["edge"])
            {
                const color = this.get("_renderercache").get("materials")["defaultcolor"];
                const edge_uniforms = THREE.UniformsUtils.clone(EdgeShader.uniforms);
                edge_uniforms["color"]["value"] = new THREE.Vector3(color[0] / 255, color[1] / 255, color[2] / 255);
                edge_uniforms["depth_texture"]["value"] = null;
                materials["edge"] = new THREE.ShaderMaterial({
                    "name": "Edge",
                    "uniforms": edge_uniforms,
                    "vertexShader": EdgeShader.vertexShader,
                    "fragmentShader": EdgeShader.fragmentShader,
                    "wireframe": true
                });
            }
        },

        UpdateReflectionUniforms: function (__opacity, __fade, __position, __fadingSize)
        {
            const rendererCache = this.get("_renderercache");
            const update_reflection_uniforms = function (__shaderMaterialName)
            {
                rendererCache.get("materials")[__shaderMaterialName]["uniforms"]["reflection_opacity"]["value"] = (__opacity !== undefined) ? __opacity : rendererCache.get("reflections")["opacity"];
                rendererCache.get("materials")[__shaderMaterialName]["uniforms"]["reflection_fade"]["value"] = (__fade !== undefined) ? __fade : rendererCache.get("reflections")["fade"];
                rendererCache.get("materials")[__shaderMaterialName]["uniforms"]["reflection_fade_position"]["value"] = (__position !== undefined) ? __position : rendererCache.get("reflections")["fade_position"];
                rendererCache.get("materials")[__shaderMaterialName]["uniforms"]["reflection_fade_size"]["value"] = (__fadingSize !== undefined) ? __fadingSize : rendererCache.get("reflections")["fade_size"];
            };
            update_reflection_uniforms("BackdropShaderMaterial");
            update_reflection_uniforms("ColorBlendShaderMaterial");
            update_reflection_uniforms("CopyShaderMaterial");
        },

        LoadMaterial: function (__jsonMaterialName, __matnode, __assets, __current_scene_id, __backTexture)
        {
            if (__matnode["AssetType"] === "Material")
            {
                if (__matnode.hasOwnProperty("Properties"))
                {
                    const matprops = __matnode["Properties"];
                    let rd = 1, gd = 1, bd = 1, rs = 1, gs = 1, bs = 1, ra = 1, ga = 1, ba = 1;

                    if (matprops["front"].hasOwnProperty("diffuse"))
                    {
                        rd = matprops["front"]["diffuse"][0];
                        gd = matprops["front"]["diffuse"][1];
                        bd = matprops["front"]["diffuse"][2];
                    }
                    if (matprops["front"].hasOwnProperty("specular"))
                    {
                        rs = matprops["front"]["specular"][0];
                        gs = matprops["front"]["specular"][1];
                        bs = matprops["front"]["specular"][2];
                    }
                    if (matprops["front"].hasOwnProperty("ambient"))
                    {
                        ra = matprops["front"]["ambient"][0];
                        ga = matprops["front"]["ambient"][1];
                        ba = matprops["front"]["ambient"][2];
                    }

                    const dif_col = new THREE.Color(rd, gd, bd);
                    const spec_col = new THREE.Color(rs, gs, bs);
                    const amb_col = new THREE.Color(ra, ga, ba);

                    const shininess = matprops["front"].hasOwnProperty("shininess") ? WV3DMath.clamp(parseFloat(matprops["front"]["shininess"]), 0, 1) : 0.5;
                    const double_sided = matprops.hasOwnProperty("double_sided") ? (WV3DUtils.ToBool(matprops["double_sided"])) : false;
                    let opacity = matprops.hasOwnProperty("opacity") ? WV3DMath.clamp(parseFloat(matprops["opacity"]), 0, 1) : 1;
                    const double_grating = matprops["front"].hasOwnProperty("double_grating") ? (WV3DUtils.ToBool(matprops["front"]["double_grating"])) : false;
                    const grating_spacing_scale = matprops["front"].hasOwnProperty("grating_spacing_scale") ? parseFloat(matprops["front"]["grating_spacing_scale"]) : null;
                    const default_diff_haze = matprops["front"].hasOwnProperty("default_diff_haze") ? parseFloat(matprops["front"]["default_diff_haze"]) : null;
                    const default_diff_spacing = matprops["front"].hasOwnProperty("default_diff_spacing") ? parseFloat(matprops["front"]["default_diff_spacing"]) : null;

                    const holographic = matprops["front"].hasOwnProperty("diff_grating_tex") || matprops["front"].hasOwnProperty("diff_grating_filter_tex");
                    const printable = (this.get("_renderercache").get("_RenderingQuality") === "Designer") || holographic || matprops["front"].hasOwnProperty("int_specular_tex") || matprops["front"].hasOwnProperty("normal_map_tex");

                    //Material opacity can be overriden if a transparent diffuse texture is attached,
                    //a default opacity parameter (see below in material_struct initialization)
                    //takes over the default one to ensure that material will be seen transparent
                    //as long as we are waiting for the attached texture (which has the right opacity
                    //value for each texel)
                    let transparent_diffuse_texture = false;
                    if (matprops["front"].hasOwnProperty("diffuse_tex")
                        && __assets[matprops["front"]["diffuse_tex"]]["Properties"].hasOwnProperty("transparent"))
                    {
                        //This only happens for printable parts (which have a diffuse alpha channel)
                        //Check if the alpha channel of the diffuse texture has any transparency
                        transparent_diffuse_texture = WV3DUtils.ToBool(__assets[matprops["front"]["diffuse_tex"]]["Properties"]["transparent"]);
                        //Conditionally replace the opacity of the untextured material
                        opacity = transparent_diffuse_texture ? 0.5 : opacity;
                    }

                    //This specific structure gathers the material data, depending on
                    //which textures are being loaded and allowed to be displayed (front is the list of
                    //current materials bound to the meshes in boundmeshlist, the one to be displayed)
                    const material_struct =
                        {
                            "JSONMaterialName": __jsonMaterialName,
                            //Here live the threejs materials,
                            "mat": {"untextured_mat": null, "textured_mat": {}},
                            //Here are the meshes that use these materials
                            "boundmeshlist": null,
                            "reflectance_model": matprops["reflectance_model"],
                            "diff_texture": null,
                            "spec_texture": null,
                            "int_spec_texture": null,
                            "normal_map_texture": null,
                            "diff_grating_texture": null,
                            "diff_grating_filter_texture": null,
                            "diff_col": dif_col,
                            "spec_col": spec_col,
                            "amb_col": amb_col,
                            "shininess": shininess,
                            "double_sided": double_sided,
                            "opacity": opacity,
                            "transp_texture": transparent_diffuse_texture,
                            "double_grating": double_grating,
                            "grating_spacing_scale": grating_spacing_scale,
                            "default_diff_haze": default_diff_haze,
                            "default_diff_spacing": default_diff_spacing,
                            "holographic": holographic,
                            "printable": printable,
                            "backTexture": __backTexture,
                            "shader": matprops.hasOwnProperty("shader") ? matprops.shader : null
                        };

                    //By default, we display the untextured material, waiting for the textures to be loaded asynchronuously
                    material_struct["mat"]["untextured_mat"] = this.CreateMaterial(material_struct);

                    let path;
                    if (this.get("_assetmanager").get("_assetloader").get("_httprequestmanager").IsAbsoluteURL(__current_scene_id))
                    {
                        path = __current_scene_id.substring(0, __current_scene_id.lastIndexOf("/") + 1);
                    }
                    else
                    {
                        path = this.get("_assetmanager").get("_resourcePath") + "/";//default resource location for samplescenedata
                    }

                    this.get("_texturemanager").StartAsyncTextureLoad(__assets, matprops, "diffuse_tex", "diff_texture", material_struct, path, printable ? THREE.RGBAFormat : THREE.RGBFormat, 4);
                    this.get("_texturemanager").StartAsyncTextureLoad(__assets, matprops, "specular_tex", "spec_texture", material_struct, path, THREE.RGBAFormat, 4);
                    if (this.get("_renderercache").get("_RenderingQuality") === "Visualizer")
                    {
                        this.get("_texturemanager").StartAsyncTextureLoad(__assets, matprops, "int_specular_tex", "int_spec_texture", material_struct, path, THREE.RGBAFormat, 4);
                        this.get("_texturemanager").StartAsyncTextureLoad(__assets, matprops, "normal_map_tex", "normal_map_texture", material_struct, path, THREE.RGBFormat, 4);
                        const dif_grat = this.get("_texturemanager").StartAsyncTextureLoad(__assets, matprops, "diff_grating_tex", "diff_grating_texture", material_struct, path, THREE.RGBAFormat, 4);
                        if(dif_grat) this.get("_texturemanager").StartAsyncTextureLoad(__assets, matprops, "diff_grating_filter_tex", "diff_grating_filter_texture", material_struct, path, THREE.RGBAFormat, 4);
                    }

                    if (this.get("_renderercache").get("_RenderingQuality") === "Visualizer" && holographic &&
                        !this.get("_renderercache").get("textures")[this.get("_renderercache").get("_RenderingQuality")]["spectrum"]["queried"])
                    {
                        //Load needed resources for holographics
                        const path = this.get("_assetmanager").get("_resourcePath") + "/" + this.get("_renderercache").get("textures")[this.get("_renderercache").get("_RenderingQuality")]["path"];

                        //The two next asynchronuous calls are treated separately, however when and only when
                        //both are completed holographic can be enabled.
                        this.get("_renderercache").get("textures")[this.get("_renderercache").get("_RenderingQuality")]["spectrum"]["queried"] = true;

                        //Lights positions and other parameters
                        new WV3DHttpRequestHandler(path + "PointLightSources.json", this.AsynchronousLightLoad.bind(this), this.AsynchronousLightLoadFail.bind(this), this.get("_assetmanager").GetURLPreprocessCallback());
                        //Here we use the texture counter (used for resources)
                        this.get("_assetmanager").IncrementTextureQueryCounter(1);

                        //Spectrum texture
                        this.get("_texturemanager").StartAsyncGlobalTextureLoad("spectrum", path);
                    }

                    return material_struct;
                }
                else
                {
                    LogMessage.Log("ViewerView Warning: Material properties not defined");
                }
            }
            else
            {
                LogMessage.Log("ViewerView Warning: Queried asset is not a material");
            }
        },

        /**
         * CreateMaterial
         * __renderpass is optional: if set, only create the required renderpass
         */
        CreateMaterial: function (__mat_struct, __renderpass)
        {
            //Select custom material to be assigned
            const shader_passes = {};
            const support_transparency = WV3DSystemChecker.get("_fullSupportFloatingPointTextures") || WV3DSystemChecker.get("_fullSupportHFloatingPointTextures");
            const material_options =
                {
                    // for designer quality, envmap is a texture. For visualizer quality, it is an array of textures
                    envmap: null,
                    irrmap: null,
                    lights: null,
                    textures: null,//shared textures (for holographics)
                    opaque: __mat_struct["opacity"] >= 1 && !__mat_struct["transp_texture"] || !support_transparency
                };

            //Build usual shading passes (irradiance, diffuse, holographics, reflectance)
            //And keep them all in a set called "opaque pass" or "transparent pass"
            if (!__renderpass
                || (__renderpass === "opaque_pass")
                || (__renderpass === "transparent_pass"))
            {
                const shader_properties = {};
                switch (this.get("_renderercache").get("_RenderingQuality"))
                {
                    case "Designer":

                        //Designer Quality
                        //Attaching irradiance and environment maps if loaded
                        if (__mat_struct["shininess"] >= 0.5)
                        {
                            if (this.get("_renderercache").get("textures")["Designer"]["envmap"]["glossy"]["loaded"])
                            {
                                material_options.envmap = this.get("_renderercache").get("textures")["Designer"]["envmap"]["glossy"]["data"];
                                //Enable irradiancemap only when envmap is enabled
                                if (this.get("_renderercache").get("textures")["Designer"]["irradiancemap"]["loaded"])
                                {
                                    material_options.irrmap = this.get("_renderercache").get("textures")["Designer"]["irradiancemap"]["data"];
                                }
                            }
                        }
                        else
                        {
                            if (this.get("_renderercache").get("textures")["Designer"]["envmap"]["matte"]["loaded"])
                            {
                                material_options.envmap = this.get("_renderercache").get("textures")["Designer"]["envmap"]["matte"]["data"];
                                //Enable irradiancemap only when envmap is enabled
                                if (this.get("_renderercache").get("textures")["Designer"]["irradiancemap"]["loaded"])
                                {
                                    material_options.irrmap = this.get("_renderercache").get("textures")["Designer"]["irradiancemap"]["data"];
                                }
                            }
                        }

                        const shaderName = "designer";

                        const designer_uniforms = THREE.UniformsUtils.clone(DesignerShader.uniforms);
                        InitializeShaderUniforms(__mat_struct, material_options, shaderName, designer_uniforms);
                        InitializeCustomShaderUniforms(__mat_struct, designer_uniforms);

                        const designer_defines = {};
                        InitializeShaderDefines(__mat_struct, material_options, shaderName, designer_defines);

                        //This is the single and only pass in designer quality. i.e. one scene construction
                        shader_properties[1] = {
                            "name": shaderName,
                            "uniforms": designer_uniforms,
                            "vertexShader": DesignerShader.vertexShader,
                            "fragmentShader": DesignerShader.fragmentShader,
                            "depthTest": true,
                            "vertexColors": THREE.FaceColors,
                            "side": THREE.FrontSide,
                            "transparent": !material_options.opaque,
                            "blending": THREE.NormalBlending,
                            "defines": designer_defines,
                            "lights": true
                        };

                        break;

                    case "Visualizer":

                        //Visualizer Quality
                        //Attaching irradiance and environment maps if loaded
                        const local_envmaps = {};
                        let all_loaded = true;
                        for (let i = 0; i < 5; ++i)
                        {
                            if (this.get("_renderercache").get("textures")["Visualizer"]["envmap"]["loaded"][i])
                            {
                                local_envmaps[i] = this.get("_renderercache").get("textures")["Visualizer"]["envmap"]["data"][i];
                            }
                            else
                            {
                                all_loaded = false;
                                break;
                            }
                        }

                        if (all_loaded && this.get("_renderercache").get("textures")["Visualizer"]["irradiancemap"]["loaded"])
                        {
                            material_options.irrmap = this.get("_renderercache").get("textures")["Visualizer"]["irradiancemap"]["data"];
                            material_options.envmap = local_envmaps;
                        }

                        //HERE, ACCORDING TO THE MATERIAL WE WANT TO CREATE, WE ORDER OUR SHADERS PASSES
                        if (__mat_struct["printable"])
                        {
                            //// IRRADIANCE (+BUMP) PASS
                            let shaderName = "irradiancepm";

                            const irradiance_shader_pm_uniforms_pass_0 = THREE.UniformsUtils.clone(IrradianceShaderPM.uniforms);
                            InitializeShaderUniforms(__mat_struct, material_options, shaderName, irradiance_shader_pm_uniforms_pass_0);
                            InitializeCustomShaderUniforms(__mat_struct, irradiance_shader_pm_uniforms_pass_0);

                            const irradiance_shader_pm_defines_pass_0 = {};
                            InitializeShaderDefines(__mat_struct, material_options, shaderName, irradiance_shader_pm_defines_pass_0);

                            shader_properties [1] = {
                                "name": shaderName,
                                "uniforms": irradiance_shader_pm_uniforms_pass_0,
                                "vertexShader": IrradianceShaderPM.vertexShader,
                                "fragmentShader": IrradianceShaderPM.fragmentShader,
                                "depthTest": true,
                                "vertexColors": THREE.FaceColors,
                                "side": THREE.FrontSide,
                                "defines": irradiance_shader_pm_defines_pass_0,
                                "blending": material_options.opaque ? THREE.NormalBlending : THREE.CustomBlending, // Coincident transparent surfaces need special blending
                                "blendSrc": THREE.OneFactor,//For CustomBlending only
                                "blendDst": THREE.OneMinusSrcAlphaFactor,//For CustomBlending only
                                "transparent": true,//<-- transparency is necessary to enable blending
                                "extensions": {
                                    derivatives: true, // set to use derivatives
                                    fragDepth: false, // set to use fragment depth values
                                    drawBuffers: false, // set to use draw buffers
                                    shaderTextureLOD: false // set to use shader texture LOD
                                }
                            };

                            //// FILTER (DIFFUSE) PASS
                            shaderName = "filterpm";

                            const filter_shader_pm_uniforms_pass_1 = THREE.UniformsUtils.clone(FilterShaderPM.uniforms);
                            InitializeShaderUniforms(__mat_struct, material_options, shaderName, filter_shader_pm_uniforms_pass_1);
                            InitializeCustomShaderUniforms(__mat_struct, filter_shader_pm_uniforms_pass_1);

                            const filter_shader_pm_defines_pass_1 = {};
                            InitializeShaderDefines(__mat_struct, material_options, shaderName, filter_shader_pm_defines_pass_1);

                            shader_properties [3] = {
                                "name": shaderName,
                                "uniforms": filter_shader_pm_uniforms_pass_1,
                                "vertexShader": FilterShaderPM.vertexShader,
                                "fragmentShader": FilterShaderPM.fragmentShader,
                                "depthTest": true,
                                "vertexColors": THREE.FaceColors,
                                "side": THREE.FrontSide,
                                "defines": filter_shader_pm_defines_pass_1,
                                "blending": THREE.CustomBlending,
                                "blendSrc": THREE.DstColorFactor,
                                "blendDst": THREE.ZeroFactor,
                                "transparent": true //<--this is necessary to enable blending
                            };


                            //// REFLECTANCE (ENVIRONMENT + BUMP) PASS
                            shaderName = "reflectancepm";
                            let reflectance_shader_pm_uniforms_pass_2;
                            if (WV3DSystemChecker.get("_glmaxtextureimageunits") >= 11)
                            {
                                reflectance_shader_pm_uniforms_pass_2 = THREE.UniformsUtils.clone(ReflectanceShaderPM.EXTuniforms);
                            }
                            else
                            {
                                reflectance_shader_pm_uniforms_pass_2 = THREE.UniformsUtils.clone(ReflectanceShaderPM.uniforms);
                            }
                            InitializeShaderUniforms(__mat_struct, material_options, shaderName, reflectance_shader_pm_uniforms_pass_2);
                            InitializeCustomShaderUniforms(__mat_struct, reflectance_shader_pm_uniforms_pass_2);

                            const reflectance_shader_pm_defines_pass_2 = {};
                            InitializeShaderDefines(__mat_struct, material_options, shaderName, reflectance_shader_pm_defines_pass_2);

                            shader_properties [6] = {
                                "name": shaderName,
                                "uniforms": reflectance_shader_pm_uniforms_pass_2,
                                "vertexShader": ReflectanceShaderPM.vertexShader,
                                "fragmentShader": (WV3DSystemChecker.get("_glmaxtextureimageunits") >= 11) ? ReflectanceShaderPM.EXTfragmentShader : ReflectanceShaderPM.fragmentShader,
                                "depthTest": true,
                                "vertexColors": THREE.FaceColors,
                                "side": THREE.FrontSide,
                                "defines": reflectance_shader_pm_defines_pass_2,
                                "blending": THREE.CustomBlending,
                                "blendSrc": THREE.OneFactor,
                                "blendDst": THREE.SrcAlphaFactor, //To be able to draw non-additive annotations in the reflectance pass
                                "transparent": true, //<--this is necessary to enable blending
                                "extensions": {
                                    derivatives: true, // set to use derivatives
                                    fragDepth: false, // set to use fragment depth values
                                    drawBuffers: false, // set to use draw buffers
                                    shaderTextureLOD: false // set to use shader texture LOD
                                }
                            };

                            if (__mat_struct["holographic"])
                            {

                                //// DIFFRACTION PASS
                                shaderName = "holographicspm";
                                material_options.lights = this.get("_renderercache").get("lights");
                                material_options.textures = this.get("_renderercache").get("textures");

                                const diffraction_shader_pm_uniforms_pass_3 = THREE.UniformsUtils.clone(DiffractionShaderPM.uniforms);
                                InitializeShaderUniforms(__mat_struct, material_options, shaderName, diffraction_shader_pm_uniforms_pass_3);

                                const diffraction_shader_pm_defines_pass_3 = {};
                                InitializeShaderDefines(__mat_struct, material_options, shaderName, diffraction_shader_pm_defines_pass_3);

                                shader_properties [5] = {
                                    "name": shaderName,
                                    "uniforms": diffraction_shader_pm_uniforms_pass_3,
                                    "vertexShader": DiffractionShaderPM.vertexShader,
                                    "fragmentShader": DiffractionShaderPM.fragmentShader,
                                    "depthTest": true,
                                    "vertexColors": THREE.FaceColors,
                                    "side": THREE.FrontSide,
                                    "defines": diffraction_shader_pm_defines_pass_3,
                                    "blending": THREE.CustomBlending,
                                    "blendSrc": THREE.OneMinusDstColorFactor,
                                    "blendDst": THREE.OneFactor,
                                    "transparent": true, //<--this is necessary to enable blending
                                    "extensions": {
                                        derivatives: true, // set to use derivatives
                                        fragDepth: false, // set to use fragment depth values
                                        drawBuffers: false, // set to use draw buffers
                                        shaderTextureLOD: false // set to use shader texture LOD
                                    }
                                };
                            }
                        }
                        else
                        {
                            //// IRRADIANCE (NO BUMP) PASS
                            let shaderName = "irradiancenpm";

                            const irradiance_shader_npm_uniforms_pass_4 = THREE.UniformsUtils.clone(IrradianceShaderNPM.uniforms);
                            InitializeShaderUniforms(__mat_struct, material_options, shaderName, irradiance_shader_npm_uniforms_pass_4);
                            InitializeCustomShaderUniforms(__mat_struct, irradiance_shader_npm_uniforms_pass_4);

                            const irradiance_shader_npm_defines_pass_4 = {};
                            InitializeShaderDefines(__mat_struct, material_options, shaderName, irradiance_shader_npm_defines_pass_4);

                            shader_properties [2] = {
                                "name": shaderName,
                                "uniforms": irradiance_shader_npm_uniforms_pass_4,
                                "vertexShader": IrradianceShaderNPM.vertexShader,
                                "fragmentShader": IrradianceShaderNPM.fragmentShader,
                                "depthTest": true,
                                "vertexColors": THREE.FaceColors,
                                "side": THREE.FrontSide,
                                "defines": irradiance_shader_npm_defines_pass_4,
                                "blending": material_options.opaque ? THREE.NormalBlending : THREE.CustomBlending, // Coincident transparent surfaces need special blending
                                "blendSrc": THREE.OneFactor,//For CustomBlending only
                                "blendDst": THREE.OneMinusSrcAlphaFactor,//For CustomBlending only
                                "transparent": true//<-- transparency is necessary to enable blending
                            };

                            //// FILTER (DIFFUSE) PASS
                            shaderName = "filternpm";

                            const filter_shader_npm_uniforms_pass_5 = THREE.UniformsUtils.clone(FilterShaderNPM.uniforms);
                            InitializeShaderUniforms(__mat_struct, material_options, shaderName, filter_shader_npm_uniforms_pass_5);
                            InitializeCustomShaderUniforms(__mat_struct, filter_shader_npm_uniforms_pass_5);

                            const filter_shader_npm_defines_pass_5 = {};
                            InitializeShaderDefines(__mat_struct, material_options, shaderName, filter_shader_npm_defines_pass_5);

                            shader_properties [4] = {
                                "name": shaderName,
                                "uniforms": filter_shader_npm_uniforms_pass_5,
                                "vertexShader": FilterShaderNPM.vertexShader,
                                "fragmentShader": FilterShaderNPM.fragmentShader,
                                "depthTest": true,
                                "vertexColors": THREE.FaceColors,
                                "side": THREE.FrontSide,
                                "defines": filter_shader_npm_defines_pass_5,
                                "blending": THREE.CustomBlending,
                                "blendSrc": THREE.DstColorFactor,
                                "blendDst": THREE.ZeroFactor,
                                "transparent": true //<--this is necessary to enable blending
                            };

                            //// REFLECTANCE (ENVIRONMENT + NO BUMP) PASS
                            shaderName = "reflectancenpm";
                            let reflectance_shader_npm_uniforms_pass_6;
                            if (WV3DSystemChecker.get("_glmaxtextureimageunits") >= 11)
                            {
                                reflectance_shader_npm_uniforms_pass_6 = THREE.UniformsUtils.clone(ReflectanceShaderNPM.EXTuniforms);
                            }
                            else
                            {
                                reflectance_shader_npm_uniforms_pass_6 = THREE.UniformsUtils.clone(ReflectanceShaderNPM.uniforms);
                            }

                            InitializeShaderUniforms(__mat_struct, material_options, shaderName, reflectance_shader_npm_uniforms_pass_6);
                            InitializeCustomShaderUniforms(__mat_struct, reflectance_shader_npm_uniforms_pass_6);

                            const reflectance_shader_npm_defines_pass_6 = {};
                            InitializeShaderDefines(__mat_struct, material_options, shaderName, reflectance_shader_npm_defines_pass_6);

                            shader_properties [7] = {
                                "name": shaderName,
                                "uniforms": reflectance_shader_npm_uniforms_pass_6,
                                "vertexShader": ReflectanceShaderNPM.vertexShader,
                                "fragmentShader": (WV3DSystemChecker.get("_glmaxtextureimageunits") >= 11) ? ReflectanceShaderNPM.EXTfragmentShader : ReflectanceShaderNPM.fragmentShader,
                                "depthTest": true,
                                "vertexColors": THREE.FaceColors,
                                "side": THREE.FrontSide,
                                "defines": reflectance_shader_npm_defines_pass_6,
                                "blending": THREE.CustomBlending,
                                "blendSrc": THREE.OneFactor,
                                "blendDst": THREE.OneFactor,
                                "transparent": true //<--this is necessary to enable blending
                            };
                        }
                        break;
                }

                //For each set of shader properties, we build the corresponding shader
                //Then we return the set of shaders that the Material requires
                for (let i = 0, keys = Object.keys(shader_properties), l = keys.length; i < l; ++i)
                {
                    //On-the-fly computation of the maximum needed number of passes
                    const id = keys[i];
                    if (shader_properties[id])
                    {
                        InitializeShaderDefines(__mat_struct, material_options, "common", shader_properties[id]["defines"]);

                        if (material_options.opaque && (!__renderpass || __renderpass === "opaque_pass"))
                        {
                            const newShader = new THREE.ShaderMaterial(shader_properties[id]);
                            newShader["polygonOffset"] = true;
                            const enableHardEdgeHighlights = this.get("_renderercache").get("materials")["enablehardedgehighlights"];
                            const enablePolygonOffset = this.get("_renderercache").get("materials")["enablehighlights"] && enableHardEdgeHighlights;
                            newShader["polygonOffsetFactor"] = enablePolygonOffset ? this.get("_renderercache").get("_polygon_offset_factor") : 0;
                            newShader["polygonOffsetUnits"] = enablePolygonOffset ? this.get("_renderercache").get("_polygon_offset_units") : 0;
                            newShader["JSONMaterialName"] = __mat_struct["JSONMaterialName"];

                            if (!shader_passes.hasOwnProperty("opaque_pass")) shader_passes["opaque_pass"] = [];
                            shader_passes["opaque_pass"].push(newShader);//Multiple shading passes material
                        }
                        else if (!material_options.opaque && (!__renderpass || __renderpass === "transparent_pass"))
                        {
                            const newShader = new THREE.ShaderMaterial(shader_properties[id]);
                            newShader["polygonOffset"] = false;
                            newShader["JSONMaterialName"] = __mat_struct["JSONMaterialName"];

                            if (!shader_passes.hasOwnProperty("transparent_pass")) shader_passes["transparent_pass"] = [];
                            shader_passes["transparent_pass"].push(newShader);//Multiple shading passes material
                        }
                        else
                        {
                            //Material not matching current pass begin built
                        }
                    }
                }
            }

            //For some rendering passes, we need a single pass shading version
            //of the scene. Here we push a simple diffuse shading pass into a
            //rendering pass name "opaque diffuse pass" or "transparent diffuse pass"
            if (!__renderpass || (__renderpass === "opaque_diffuse_pass") || (__renderpass === "transp_diffuse_pass"))
            {
                // This shader is mat_struct-dependent, so it cannot be put in the global material rendercache,
                // but should NOT be added to the usual renderpasses.
                const shaderName = "transparentblend";

                const trans_blend_uniforms = THREE.UniformsUtils.clone(TransBlendShader.uniforms);
                InitializeShaderUniforms(__mat_struct, material_options, shaderName, trans_blend_uniforms);

                const trans_blend_defines = {};
                InitializeShaderDefines(__mat_struct, material_options, shaderName, trans_blend_defines);

                const TransBlendShaderMaterial = new THREE.ShaderMaterial({
                    "name": shaderName,
                    "uniforms": trans_blend_uniforms,
                    "vertexShader": TransBlendShader.vertexShader,
                    "fragmentShader": TransBlendShader.fragmentShader,
                    "side": THREE.FrontSide,
                    "depthWrite": true,
                    "transparent": true,

                    // We use multiplicative blending to ensure that if several concident surfaces are drawn on the same peel,
                    // the result is the same as if they were each drawn on their own peel
                    "blending": THREE.CustomBlending,
                    "blendSrc": THREE.DstColorFactor,
                    "blendDst": THREE.ZeroFactor,
                    "lights": false,
                    "defines": trans_blend_defines
                });

                if (material_options.opaque && (!__renderpass || __renderpass === "opaque_diffuse_pass"))
                {
                    shader_passes["opaque_diffuse_pass"] = [TransBlendShaderMaterial];//single shading pass material
                }
                else if (!material_options.opaque && (!__renderpass || __renderpass === "transp_diffuse_pass"))
                {
                    shader_passes["transp_diffuse_pass"] = [TransBlendShaderMaterial];//single shading pass material
                }
                else
                {
                    //Material not matching current pass begin built
                }
            }
            return shader_passes;
        },

        AsynchronousLightLoad: function (Lights)
        {
            this.get("_exec_queue").whenFinished(function ()
            {
                this.get("_renderercache").get("lights")["holo_light"] = Lights;
                this.UpdateMaterialsGlobalLighting();
                this.get("_assetmanager").IncrementTextureLoadCounter();

                this.trigger("render");
                this.trigger("check_post_scene_load_actions");
            }, this);
        },

        AsynchronousLightLoadFail: function ()
        {
            this.get("_exec_queue").whenFinished(function ()
            {
                this.get("_assetmanager").IncrementTextureLoadCounter();
                const errorParams =
                    {
                        "errorCode": 3,
                        "message": "Point Light Sources"
                    };
                this.trigger("resource_query_error", errorParams);
            }, this);
        },

        /**
         * Update link from all meshes to their material
         */
        UpdateMaterialsGlobalLighting: function ()
        {
            const materials = this.get("_renderercache").get("materials");
            for (let k = 0, _mats = Object.keys(materials["materiallist"]), l2 = _mats.length; k < l2; ++k)
            {
                const mat = _mats[k];
                const mat_struct_front = materials["materiallist"][mat]["front"];
                const mat_struct_back = materials["materiallist"][mat]["back"];
                this.UpdateTexturedMaterial({"mat_struct": mat_struct_front});
                this.UpdateTexturedMaterial({"mat_struct": mat_struct_back});
            }
        },

        /**
         * Update link from mesh to material
         */
        UpdateTexturedMaterial: function (__params)
        {
            const __mat_struct = __params["mat_struct"];
            if (__mat_struct)
            {
                const meshestoupdate = __mat_struct["boundmeshlist"];
                //Update material reference in meshes that are already bound to this material
                if (meshestoupdate)
                {
                    if (Object.keys(meshestoupdate).length === 0) LogMessage.Log("ViewerView Warning: No meshes to update");
                    for (let j = 0, keys1 = Object.keys(meshestoupdate), L = keys1.length; j < L; ++j)
                    {
                        const renderpass = keys1[j];
                        //Create replacement material
                        const material = this.CreateMaterial(__mat_struct, renderpass);
                        if (Object.keys(meshestoupdate[renderpass]).length === 0) LogMessage.Log("ViewerView Warning: No render passes to update");
                        for (let i = 0, keys2 = Object.keys(meshestoupdate[renderpass]), l = keys2.length; i < l; ++i)
                        {
                            const side_mat = keys2[i];
                            if (Object.keys(meshestoupdate[renderpass][side_mat]).length === 0) LogMessage.Log("ViewerView Warning: No sided materials to update");
                            for (let k = 0, keys3 = Object.keys(meshestoupdate[renderpass][side_mat]), m = keys3.length; k < m; ++k)
                            {
                                const shadingpassID = keys3[k];
                                if (Object.keys(meshestoupdate[renderpass][side_mat][shadingpassID]).length === 0) LogMessage.Log("ViewerView Warning: No shading pass to update");
                                for (let q = 0, keys4 = Object.keys(meshestoupdate[renderpass][side_mat][shadingpassID]), p = keys4.length; q < p; ++q)
                                {
                                    const surfID = keys4[q];
                                    //Release before overwriting placeholder material
                                    ReleaseMaterial(__mat_struct["mat"]["textured_mat"][renderpass][shadingpassID][surfID], false);
                                    //Replace placeholder material and update bound meshes
                                    __mat_struct["mat"]["textured_mat"][renderpass][shadingpassID][surfID] = material[renderpass][shadingpassID];

                                    for (let n = 0, keys5 = Object.keys(meshestoupdate[renderpass][side_mat][shadingpassID][surfID]), o = keys5.length; n < o; ++n)
                                    {
                                        const meshtoupdate = meshestoupdate[renderpass][side_mat][shadingpassID][surfID][keys5[n]];
                                        meshtoupdate.material = material[renderpass][shadingpassID];
                                        meshtoupdate.geometry.buffersNeedUpdate = true;
                                        meshtoupdate.geometry.uvsNeedUpdate = true;
                                        if (meshtoupdate.geometry.hasOwnProperty("dUdP") && meshtoupdate.geometry.hasOwnProperty("dVdP"))
                                        {
                                            // Keep this lines for debugging: these are useful to know if different surfaces have different
                                            // dUdP and dVdP scaling factors.
                                            // if(__mat_struct.hasOwnProperty("dUdP") && __mat_struct.hasOwnProperty("dVdP") &&
                                            //     (__mat_struct["dUdP"] != meshtoupdate.geometry["dUdP"] || __mat_struct["dVdP"] != meshtoupdate.geometry["dVdP"]))
                                            // {
                                            //     console.warn("__mat_struct assigned to two different meshes, each having a different UV scaling factor. Developer error ?");
                                            // }
                                            __mat_struct["dUdP"] = meshtoupdate.geometry["dUdP"];
                                            __mat_struct["dVdP"] = meshtoupdate.geometry["dVdP"];
                                        }
                                    }
                                }
                            }
                        }
                    }
                }
                else
                {
                    LogMessage.Log("ViewerView Warning: No meshes to update");
                }
            }
        },

        /**
         * Use a new texture for the current material structure
         */
        UpdateMaterialTexture: function (__params)
        {
            const mat_struct = __params["mat_struct"];
            const textureURL = __params["texture_URL"];
            const newTexture = __params["new_texture"];
            if (mat_struct)
            {
                const passestoupdate = mat_struct["mat"]["textured_mat"];
                //Update material struct texture reference
                for (let idx = 0, keys = Object.keys(mat_struct), l = keys.length; idx < l; ++idx)
                {
                    const prop = keys[idx];
                    if (mat_struct[prop] && mat_struct[prop] instanceof THREE.Texture &&
                        mat_struct[prop].name !== "" && mat_struct[prop].name === textureURL)
                    {
                        mat_struct[prop].dispose();
                        mat_struct[prop] = newTexture;
                    }
                }

                //Update each material texture reference
                if (passestoupdate)
                {
                    if (Object.keys(passestoupdate).length === 0) LogMessage.Log("ViewerView Warning: No render passes to update");
                    for (let j = 0, keys1 = Object.keys(passestoupdate), L = keys1.length; j < L; ++j)
                    {
                        const renderpass = keys1[j];
                        if (Object.keys(passestoupdate[renderpass]).length === 0) LogMessage.Log("ViewerView Warning: No shading passes to update");
                        for (let i = 0, keys2 = Object.keys(passestoupdate[renderpass]), l = keys2.length; i < l; ++i)
                        {
                            const shadingpassID = keys2[i];
                            if (Object.keys(passestoupdate[renderpass][shadingpassID]).length === 0) LogMessage.Log("ViewerView Warning: No surface materials to update");
                            for (let k = 0, keys3 = Object.keys(passestoupdate[renderpass][shadingpassID]), m = keys3.length; k < m; ++k)
                            {
                                const surfID = keys3[k];
                                const material = passestoupdate[renderpass][shadingpassID][surfID];

                                if (material && material.uniforms)
                                {
                                    for (let mat_uni_idx = 0, keys = Object.keys(material.uniforms), lk = keys.length; mat_uni_idx < lk; ++mat_uni_idx)
                                    {
                                        const uniform = material.uniforms[keys[mat_uni_idx]];
                                        if (uniform["value"])
                                        {
                                            if (typeof(uniform["value"]["dispose"]) === "function" && uniform["value"] instanceof THREE.Texture)
                                            {
                                                const previous_texture = uniform["value"];
                                                if (previous_texture.name !== "" && previous_texture.name === textureURL)
                                                {
                                                    newTexture.format = previous_texture.format;
                                                    newTexture.unpackAlignment = previous_texture.unpackAlignment;
                                                    newTexture.anisotropy = previous_texture.anisotropy;
                                                    newTexture.repeat = previous_texture.repeat;
                                                    newTexture.remappedAlphaChannel = previous_texture.remappedAlphaChannel;
                                                    uniform["value"] = newTexture;
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                        }
                    }
                }
                else
                {
                    LogMessage.Log("ViewerView Warning: No textured material to update");
                }
            }
        },

        ReleaseGeomMatTex: function (__scenegraph)
        {
            if (__scenegraph)
            {
                const releaseGeometryAndMaterialCallback = function (__obj)
                {
                    if (__obj instanceof THREE.Mesh && __obj.geometry)
                    {
                        __obj.geometry.dispose();
                        ReleaseMaterial(__obj.material);
                    }
                };
                __scenegraph.traverse(releaseGeometryAndMaterialCallback);
            }
        },

        ReleaseAllMaterialsAndTextures: function ()
        {
            const rendererCache = this.get("_renderercache");

            ReleaseScenesMaterialsAndTextures(rendererCache);

            ReleaseMaterial(rendererCache.get("materials")["default_material"]);
            ReleaseMaterial(rendererCache.get("materials")["highlight"]);
            ReleaseMaterial(rendererCache.get("materials")["edge"]);
            ReleaseMaterial(rendererCache.get("materials")["colordepth"]);
            ReleaseMaterial(rendererCache.get("materials")["D0ShaderMaterial"]);
            ReleaseMaterial(rendererCache.get("materials")["D1ShaderMaterial"]);
            ReleaseMaterial(rendererCache.get("materials")["D2ShaderMaterial"]);
            ReleaseMaterial(rendererCache.get("materials")["BackdropShaderMaterial"]);
            ReleaseMaterial(rendererCache.get("materials")["ColorBlendShaderMaterial"]);
            ReleaseMaterial(rendererCache.get("materials")["CopyShaderMaterial"]);
            ReleaseMaterial(rendererCache.get("materials")["alphafix"]);
        },

        ApplyFctToAllMaterials: function (__callback)
        {
            //Update and trigger rebuild of default material shader
            __callback(this.get("_renderercache").get("materials")["default_material"]);
            this.get("_renderercache").get("materials")["default_material"].needsUpdate = true;

            // For the moment, we do not need to call any callback function on these materials
            // __callback(this.get('_renderercache').get('materials')["highlight"]);
            // __callback(this.get('_renderercache').get('materials')["D0ShaderMaterial"]);
            // __callback(this.get('_renderercache').get('materials')["D1ShaderMaterial"]);
            // __callback(this.get('_renderercache').get('materials')["D2ShaderMaterial"]);
            // __callback(this.get('_renderercache').get('materials')["BackdropShaderMaterial"]);
            // __callback(this.get('_renderercache').get('materials')["ColorBlendShaderMaterial"]);
            // __callback(this.get('_renderercache').get('materials')["CopyShaderMaterial"]);

            for (let n = 0, keys = Object.keys(this.get("_renderercache").get("materials")["materiallist"]), o = keys.length; n < o; ++n)
            {
                const mat_structs = this.get("_renderercache").get("materials")["materiallist"][keys[n]];
                ApplyFctToMaterialStruct(mat_structs.front, __callback);
                ApplyFctToMaterialStruct(mat_structs.back, __callback);
            }
        },

        UpdateMatStructAnnotationLines: function (__mat_struct, __segmentList)
        {
            // 64 is a magic number. This is the resolution that gives the best performance while rendering annotations
            const accelGridSize = 64;

            const texSizeAndMaxLinesPerCell = this.get("_texturemanager").UpdateAnnotationsTexture(__mat_struct, __segmentList, accelGridSize);
            const textureSize = texSizeAndMaxLinesPerCell.textureSize;
            const maxLinesPerCell = texSizeAndMaxLinesPerCell.maxLinesPerCell;
            const linesStartOffset = texSizeAndMaxLinesPerCell.linesStartOffset;

            ApplyFctToMaterialStruct(__mat_struct, function (__material)
            {
                //We only enable annotations in the Visualizer quality reflectance pass
                //or for designer quality pass
                if (__material.hasOwnProperty("uniforms") &&
                    __material.hasOwnProperty("name") &&
                    (__material.name === "reflectancepm" || __material.name === "designer" || __material.name === "transparentblend") &&
                    __material["defines"].hasOwnProperty("ANNOTATION_NUMBER"))
                {
                    __material["defines"]["ANNOTATION_NUMBER"] = "" + __segmentList.length;
                    __material["defines"]["ANNOTATION_TEXTURE_SIZE"] = "" + textureSize;
                    __material["defines"]["MAX_LINES_PER_CELL"] = "" + maxLinesPerCell;
                    __material["defines"]["ACCEL_GRID_SIZE"] = accelGridSize;
                    __material["defines"]["LINES_START_OFFSET"] = linesStartOffset;
                    __material["uniforms"]["annotationsTexture"]["value"] = __segmentList.length ? __mat_struct["annotationsTexture"] : null;
                    __material["uniforms"]["annotationsAspectRatio"]["value"] = GetTextureScaling(__mat_struct);
                }
            });
        },

        GetJSONMaterialStruct: function (__jsonMaterialName)
        {
            const materiallist = this.get("_renderercache").get("materials")["materiallist"];
            for (let i = 0, mat_keys = Object.keys(materiallist), l = mat_keys.length; i < l; ++i)
            {
                const mat_structs = materiallist[mat_keys[i]];
                //ANSWERME: should we add mat_structs["material_name"] and not just front or back mat_struct
                if (mat_structs["front"]["JSONMaterialName"] === __jsonMaterialName)
                {
                    return mat_structs["front"];
                }
                if (!mat_structs["sharedmaterial"] && mat_structs["back"]["JSONMaterialName"] === __jsonMaterialName)
                {
                    return mat_structs["back"];
                }
            }
            return null;
        },

        /**
         * Retrieve the list of the parts that can support annotations
         */
        GetAnnotablePartsList: function ()
        {
            const annotablePartsList = [];

            const materiallist = this.get("_renderercache").get("materials")["materiallist"];
            for (let i = 0, mat_keys = Object.keys(materiallist), l = mat_keys.length; i < l; ++i)
            {
                const mat_structs = materiallist[mat_keys[i]];
                if (mat_structs["front"]["printable"])
                {
                    //ANSWERME: should we add mat_structs["material_name"] and not just JSONMaterialName
                    annotablePartsList.push(mat_structs["front"]["JSONMaterialName"]);
                    if (!mat_structs["sharedmaterial"])
                    {
                        //Back is distinct from front, therefore, we also add it to the list
                        annotablePartsList.push(mat_structs["back"]["JSONMaterialName"]);
                    }
                }
            }
            return annotablePartsList;
        },

        /**
         * Set the list of annotation segments for the given part
         */
        SetAnnotations: function (__jsonMaterialName, __segmentList)
        {
            const mat_struct = this.GetJSONMaterialStruct(__jsonMaterialName);
            if (mat_struct)
            {
                this.UpdateMatStructAnnotationLines(mat_struct, __segmentList);
                return true;
            }
            else
                return false;
        },

        /**
         * Set the color of annotation segments for the given part
         */
        SetAnnotationsThickness: function (__thickness)
        {
            this.get("_renderercache").set("_annotationsThickness", __thickness);
            this.ApplyFctToAllMaterials(function (__material)
            {
                //We only enable annotations in the Visualizer quality reflectance pass
                //or for designer quality pass
                if (__material.hasOwnProperty("uniforms") &&
                    __material.hasOwnProperty("name") &&
                    (__material.name === "reflectancepm" || __material.name === "designer" || __material.name === "transparentblend") &&
                    __material["defines"].hasOwnProperty("ANNOTATION_NUMBER"))
                {
                    __material["uniforms"]["annotationsThickness"]["value"] = __thickness;
                }
            });
        },

        /**
         * Set the color of annotation segments for the given part
         */
        SetAnnotationsColor: function (__red, __green, __blue)
        {
            this.ApplyFctToAllMaterials(function (__material)
            {
                //We only enable annotations in the Visualizer quality reflectance pass
                //or for designer quality pass
                if (__material.hasOwnProperty("uniforms") &&
                    __material.hasOwnProperty("name") &&
                    (__material.name === "reflectancepm" || __material.name === "designer" || __material.name === "transparentblend") &&
                    __material["defines"].hasOwnProperty("ANNOTATION_NUMBER"))
                {
                    __material["uniforms"]["annotationsColor"]["value"] = new THREE.Vector3(__red, __green, __blue);
                }
            });
        }

    });
});
