/**
 * @author DOGO
 * @module WebViewer3D
 * @copyright   2014 Esko Stonecube Ltd. All rights reserved.
 */

define([
    "underscore",
    "backbone",
    "three"
], function (
    _,
    bb,
    THREE
)
{

    "use strict";

    /**
     * @alias module:WebViewer3D
     * @class WV3DMath
     */
    const WV3DMath = bb.Model.extend({

        // Converts from degrees to radians.
        radians: function (__degrees)
        {
            return __degrees * Math.PI / 180;
        },

        // Converts from radians to degrees.
        degrees: function (__radians)
        {
            return __radians * 180 / Math.PI;
        },

        // clamps a value between min and max.
        clamp: function (__number, __min, __max)
        {
            return Math.min(Math.max(__number, __min), __max);
        },

        isPowerOfTwo: function (x)
        {
            return (x & (x - 1)) == 0;
        },

        nextHighestPowerOfTwo: function (x)
        {
            --x;
            for (var i = 1; i < 32; i <<= 1)
            {
                x = x | x >> i;
            }
            return x + 1;
        },

        // Package-wrapping convex hull algo, see http://www.cs.princeton.edu/courses/archive/spr03/cs226/lectures/geo.4up.pdf
        CreateConvexHullFromPointlist: function (__points) // First M entries in __points is ordered convex hull, M is returned.
        {
            var N = __points.length;
            // console.log("input:");
            // for( var q = 0; q < N; q++)
            //     console.log(__points[q].x, ",", __points[q].y );

            // find point with lowest y-coordinate (rightmost for tie-breaking)
            var i, min, M, th, v, pt;
            for (min = 0, i = 1; i < N; i++)
                if (__points[i].y < __points[min].y ||
                    (__points[i].y == __points[min].y && __points[i].x > __points[min].x))
                    min = i;

            // store that at the end of the array
            __points[N] = __points[min];
            th = .0;
            for (M = 0; M < N; M++)
            {
                //console.log("Hullpoint ", M," = " ,__points[min].x, ",", __points[min].y);

                // swap most recently found convex hull point from end of entire array to end of hull array
                pt = __points[M];
                __points[M] = __points[min];
                __points[min] = pt;

                min = N; // sentinel value
                v = th;
                th = 2 * Math.PI;
                for (i = M + 1; i <= N; i++)
                {
                    var theta = Math.atan2(__points[i].y - __points[M].y, __points[i].x - __points[M].x);
                    if (theta < 0)
                        theta += 2 * Math.PI;

                    //console.log("considering pt",i," = " ,__points[i].x, ",", __points[i].y, " with theta ", theta);
                    if (theta > v)
                        if (theta < th)
                        {
                            //console.log("point taken");
                            min = i;
                            th = theta;
                        }
                }
                if (min == N) // if we have reached the sentinel, we're done.
                {
                    return M + 1; // original code had off-by-one error!
                }
            }
        },

        ConvexHullCenter: function (__points, N)
        {
            //Previous approach is still bugged when convex hull points are alined (returns center (0,0)),
            //therefore it is currently disabled and use a simple average center instead.
            var center = new THREE.Vector2(0, 0);
            for (var i = 0; i < N; i++)
            {
                center.add(__points[i]);
            }
            center.divideScalar(N);
            return center;

            // if (N == 1)
            //     return __points[0];
            // if (N == 2)
            //     return new THREE.Vector2( (__points[0].x + __points[1].x)*0.5,
            //                               (__points[0].y + __points[1].y)*0.5 );

            // // since a convex hull might be very unbalanced (f.e: 100 points in half a circle on the left side,
            // // and a single-point triangle on the right), we can't simply take the average of all hull vertices.
            // // To calculate the center-of-mass of a convex hull, split it in triangles, calculate the center of
            // // each triangle, then take the average of the centers, weighted by the area of each triangle.

            // var avg_center = new THREE.Vector2(0,0),
            //     total_area = 0;
            // // we split the convex polygon in N-2 triangles which share vertex 0
            // for (var i = 1; i < N-1; i++)
            // {
            //     var tri_center = new THREE.Vector2( __points[0].x, __points[0].y);
            //     tri_center.add( __points[i] );
            //     tri_center.add( __points[i+1] );
            //     tri_center.divideScalar( 3.);

            //     var tri_area = 0.5 * (-__points[i].x*__points[0].y+__points[i+1].x*__points[0].y
            //                           +__points[0].x*__points[i].y-__points[i+1].x*__points[i].y
            //                           -__points[0].x*__points[i+1].y+__points[i].x*__points[i+1].y);
            //     avg_center.addScaledVector( tri_center, tri_area);
            //     total_area += tri_area;
            // }
            // avg_center.divideScalar(total_area );
            // return avg_center;
        },

        ConvexHullContainsPoint: function (__points, N, __pt)
        {
            // Solution taken from https://www.ecse.rpi.edu/Homepages/wrf/Research/Short_Notes/pnpoly.html
            var result = false;
            for (var i = 0, j = N - 1; i < N; j = i++)
            {
                if ((__points[i].y > __pt.y) != (__points[j].y > __pt.y) && (__pt.x < (__points[j].x - __points[i].x) * (__pt.y - __points[i].y) / (__points[j].y - __points[i].y) + __points[i].x))
                {
                    result = !result;
                }
            }
            return result;
        },

        //Found in fiddle http://jsfiddle.net/justin_c_rounds/Gd2S2/
        CheckLineIntersection: function (line1StartX, line1StartY, line1EndX, line1EndY, line2StartX, line2StartY, line2EndX, line2EndY)
        {
            // if the lines intersect, the result contains the x and y of the intersection (treating the lines as infinite) and booleans for whether line segment 1 or line segment 2 contain the point
            var denominator, a, b, numerator1, numerator2, result = {
                x: null,
                y: null,
                onLine1: false,
                onLine2: false
            };
            denominator = ((line2EndY - line2StartY) * (line1EndX - line1StartX)) - ((line2EndX - line2StartX) * (line1EndY - line1StartY));
            if (denominator == 0)
            {
                return result;
            }
            a = line1StartY - line2StartY;
            b = line1StartX - line2StartX;
            numerator1 = ((line2EndX - line2StartX) * a) - ((line2EndY - line2StartY) * b);
            numerator2 = ((line1EndX - line1StartX) * a) - ((line1EndY - line1StartY) * b);
            a = numerator1 / denominator;
            b = numerator2 / denominator;

            // if we cast these lines infinitely in both directions, they intersect here:
            result.x = line1StartX + (a * (line1EndX - line1StartX));
            result.y = line1StartY + (a * (line1EndY - line1StartY));
            /*
            // it is worth noting that this should be the same as:
            x = line2StartX + (b * (line2EndX - line2StartX));
            y = line2StartX + (b * (line2EndY - line2StartY));
            */
            // if line1 is a segment and line2 is infinite, they intersect if:
            if (a > 0 && a < 1)
            {
                result.onLine1 = true;
            }
            // if line2 is a segment and line1 is infinite, they intersect if:
            if (b > 0 && b < 1)
            {
                result.onLine2 = true;
            }
            // if line1 and line2 are segments, they intersect if both of the above are true
            return result;
        },

        SemiStraightLineConvexHullIntersection: function (__points, N, __line_start, __line_direction)
        {
            var line_intersection, res;
            for (var i = 0; i < N; i++)
            {
                var dir_line_end = new THREE.Vector2(__line_start.x + __line_direction.x, __line_start.y + __line_direction.y);
                line_intersection = this.CheckLineIntersection(__points[i].x, __points[i].y, __points[(i + 1) % N].x, __points[(i + 1) % N].y,
                    __line_start.x, __line_start.y, dir_line_end.x, dir_line_end.y);
                if (line_intersection.onLine1)//allow line2 to be a straight line
                {
                    //There will probably be two intersections with the convex hull with that straight line
                    if (res === undefined)
                    {
                        //save the first result
                        res = new THREE.Vector2(line_intersection.x, line_intersection.y);
                    }
                    else
                    {
                        //override the first result with the second if the new result is closer to the end point
                        var second_res = new THREE.Vector2(line_intersection.x, line_intersection.y);
                        if (dir_line_end.distanceToSquared(res) > dir_line_end.distanceToSquared(second_res))
                        {
                            res = second_res;
                        }
                    }
                }
            }
            return res;
        },

        //Used for animation
        SmartEaseInEaseOutInterpolation: function (__startVal, __endVal, __currentTime, __totalTime)
        {
            var deltaVal = __endVal - __startVal;//signed distance
            var easeInTime, easeOutTime;

            //Here, we hardcode the time(duration) for ease-in and ease-out, for example
            //"1" second for ease-in if __totalTime > 10 seconds
            //otherwise 10% of __totalTime
            //"1" second for ease-in if __totalTime > 1.1111 seconds
            //otherwise 90% of __totalTime
            //In the future, these settings could also appear in the UI to control ease-in and ease-out

            easeInTime = Math.min(1000, 0.1 * __totalTime);
            easeOutTime = Math.min(1000, 0.9 * __totalTime);
            var middleTime = __totalTime - easeInTime - easeOutTime;
            var maxSpeed = deltaVal / (0.5 * (easeInTime + easeOutTime) + middleTime);//maximal speed to reach between ease-in and ease-out
            var easeInDelta = easeInTime * maxSpeed * 0.5;//distance for ease-in
            var easeOutDelta = easeOutTime * maxSpeed * 0.5;//distance for ease-out

            var currentDistance = 0.;
            if (__currentTime < easeInTime || __currentTime > easeInTime + middleTime)
            {
                if (__currentTime < easeInTime)
                {
                    //Ease-in using smoothstep function
                    var t = __currentTime / easeInTime;
                    var t3 = t * t * t;
                    var t4 = t3 * t;
                    var alphaPos = 2 * (t3 - t4 * 0.5);
                    currentDistance = (1 - alphaPos) * __startVal + alphaPos * (__startVal + easeInDelta);
                }
                else
                {
                    //Ease-out using smoothstep function
                    var deltatime = __currentTime - (easeInTime + middleTime);
                    var t = 1.0 - deltatime / easeOutTime;
                    var t3 = t * t * t;
                    var t4 = t3 * t;
                    var alphaPos = 2 * (t3 - t4 * 0.5);
                    currentDistance = alphaPos * (__endVal - easeOutDelta) + (1. - alphaPos) * __endVal;
                }
            }
            else
            {
                //Move at max speed
                var alphaPos = (__currentTime - easeInTime) / middleTime;
                currentDistance = (1 - alphaPos) * (__startVal + easeInDelta) + alphaPos * (__endVal - easeOutDelta);
            }
            return currentDistance;
        },

        IntersectCellSegment: function (__cell, __segment)
        {
            var minX = Math.min(__segment.startU, __segment.endU);
            var maxX = Math.max(__segment.startU, __segment.endU);

            if (maxX > __cell.right)
                maxX = __cell.right;

            if (minX < __cell.left)
                minX = __cell.left;

            if (minX > maxX)
                return false;

            var minY = __segment.startV;
            var maxY = __segment.endV;

            var dx = __segment.endU - __segment.startU;

            if (Math.abs(dx) > 0.0000001)
            {
                var a = (__segment.endV - __segment.startV) / dx;
                var b = __segment.startV - a * __segment.startU;
                minY = a * minX + b;
                maxY = a * maxX + b;
            }

            if (minY > maxY)
            {
                minY = [maxY, maxY = minY][0];//swap values
            }

            if (maxY > __cell.top)
                maxY = __cell.top;

            if (minY < __cell.bottom)
                minY = __cell.bottom;

            if (minY > maxY)
                return false;

            return true;
        }

    });


    return new WV3DMath;

});
