/**
 * @author DOGO
 * @module WebViewer3D
 * @copyright   2014 Esko Stonecube Ltd. All rights reserved.
 */

define([
    "underscore",
    "backbone",
    "three",
    "@esko/3d-web-viewer/modules/Utils",
    "@esko/3d-web-viewer/modules/Timers",
    "@esko/3d-web-viewer/modules/LogMessage"
], function (
    _,
    bb,
    THREE,
    WV3DUtils,
    WV3DTimers,
    LogMessage
)
{

    "use strict";

    //IMPROVEME: use const and let instead of var

    /* PRIVATE */
    const GetSignedArea = function (__uv1, __uv2, __uv3)
    {
        //see https://math.stackexchange.com/questions/10519/how-to-calculate-area-of-triangle-having-its-points-2d-coordinates
        //that explains how to get the signed area of a triangle using its coordinates
        var matTexCoord = new THREE.Matrix3();
        matTexCoord.set(__uv1.x, __uv1.y, 1,
            __uv2.x, __uv2.y, 1,
            __uv3.x, __uv3.y, 1);
        //if positive, points are ordered anticlockwise
        return matTexCoord.determinant();
    };

    const GetAveragedPdUdPdV = function (__trianglesCache)
    {
        if (__trianglesCache.length == 0)
            return new THREE.Vector2(1, 1);
        var dPdU = 0;
        var dPdV = 0;
        var areaSum = 0;
        //We also weigh each triangle using their own area
        for (var i = 0; i < __trianglesCache.length; ++i)
        {
            dPdU += __trianglesCache[i]["value"].x * __trianglesCache[i]["TriangleArea"];
            dPdV += __trianglesCache[i]["value"].y * __trianglesCache[i]["TriangleArea"];
            areaSum += __trianglesCache[i]["TriangleArea"];
        }
        dPdU /= areaSum;
        dPdV /= areaSum;
        return new THREE.Vector2(dPdU, dPdV);
    };

    /**
     * Creates the cache list of the X largest triangles in the surface
     * along with their area and the dPdUdPdV value
     * The dPdUdPdV value is the ratio of the distance (P) by the texture coordinate differential (U or V)
     * along the triangle.
     * __largestTrianglesCache is the output of this function.
     */
    const CalculatedPdUdPdVForLargestTriangles = function (__p1, __p2, __p3, __v1u, __v1v, __v2u, __v2v, __v3u, __v3v, __largestTrianglesCache)
    {
        var epsilon = 0.0000001;
        var maxCacheLength = 20;//We calculate the ratio among the X largest triangles
        var tc1 = new THREE.Vector2(__v1u, __v1v);
        var tc2 = new THREE.Vector2(__v2u, __v2v);
        var tc3 = new THREE.Vector2(__v3u, __v3v);

        var tcunitU = new THREE.Vector2(__v1u + 1, __v1v);
        var tcunitV = new THREE.Vector2(__v1u, __v1v + 1);

        var TriangleArea = GetSignedArea(tc1, tc3, tc2);
        if (Math.abs(TriangleArea) < epsilon)
        {
            return;//early rejection for degenerate triangle
        }
        var dPdUdPdV = {"TriangleArea": Math.abs(TriangleArea)};
        if (__largestTrianglesCache.length > maxCacheLength && dPdUdPdV["TriangleArea"] < __largestTrianglesCache[0])
        {
            return;//early rejection for small triangles
        }

        //We use the signed areas to determine the barycentric coordinates
        var UWeight3 = GetSignedArea(tc1, tcunitU, tc2) / TriangleArea;
        var UWeight1 = GetSignedArea(tc2, tcunitU, tc3) / TriangleArea;
        var UWeight2 = GetSignedArea(tc3, tcunitU, tc1) / TriangleArea;
        var VWeight3 = GetSignedArea(tc1, tcunitV, tc2) / TriangleArea;
        var VWeight1 = GetSignedArea(tc2, tcunitV, tc3) / TriangleArea;
        var VWeight2 = GetSignedArea(tc3, tcunitV, tc1) / TriangleArea;

        //Now we use the signed weight to determine the actual positions
        //of the unitU and unitV
        var punitU = __p1.clone().multiplyScalar(UWeight1).add(__p2.clone().multiplyScalar(UWeight2).add(__p3.clone().multiplyScalar(UWeight3)));
        var punitV = __p1.clone().multiplyScalar(VWeight1).add(__p2.clone().multiplyScalar(VWeight2).add(__p3.clone().multiplyScalar(VWeight3)));

        //Now that we have the actual positions of the unit U and V vectors
        //we subtract p1 to each of these and get the length to obtain the scaling
        //ratios in U and V directions
        dPdUdPdV["value"] = new THREE.Vector2(punitU.sub(__p1).length(), punitV.sub(__p1).length());
        var valid = (dPdUdPdV.value.x !== Infinity) && (dPdUdPdV.value.x > epsilon) && (dPdUdPdV.value.y !== Infinity) && (dPdUdPdV.value.y > epsilon);

        if (valid)
        {
            if (__largestTrianglesCache.length >= maxCacheLength)
            {
                var insertIndex = 0;
                for (var i = 1; i < __largestTrianglesCache.length; ++i)
                {
                    if (dPdUdPdV["TriangleArea"] > __largestTrianglesCache[i]["TriangleArea"])
                    {
                        insertIndex = i - 1;
                    }
                    else
                    {
                        break;
                    }
                }
                //Shift all items to the left, starting from insertIndex
                //This way, we avoid slicing or sorting the array,
                //which can be expensive if we do it several times.
                for (var i = 0; i < insertIndex; ++i)
                {
                    __largestTrianglesCache[i] = __largestTrianglesCache[i + 1];
                }
                //Insert the new triangle where it should be
                __largestTrianglesCache[insertIndex] = dPdUdPdV;
            }
            else
            {
                __largestTrianglesCache.push(dPdUdPdV);
                if (__largestTrianglesCache.length == maxCacheLength)
                {
                    //Now that we have filled the X largest triangle cache,
                    //we sort the array once so that we can proceed with smart
                    //insertions (see array values shifting above)
                    __largestTrianglesCache.sort(function (__a, __b)
                    {
                        var sign = __a.TriangleArea - __b.TriangleArea;
                        if (sign < 0)
                            return -1;
                        else if (sign > 0)
                            return 1;
                        else return 0;
                    });
                }
            }
        }
    };

    const DotNormals = function (__idxnormals, __idxvertices, __normals, __vertices)
    {
        var face_norm_idx_1 = __idxnormals[0] * 3;
        var face_norm_idx_2 = __idxnormals[1] * 3;
        var face_norm_idx_3 = __idxnormals[2] * 3;
        var normal_1 = new THREE.Vector3(__normals[face_norm_idx_1], __normals[face_norm_idx_1 + 1], __normals[face_norm_idx_1 + 2]);
        var normal_2 = new THREE.Vector3(__normals[face_norm_idx_2], __normals[face_norm_idx_2 + 1], __normals[face_norm_idx_2 + 2]);
        var normal_3 = new THREE.Vector3(__normals[face_norm_idx_3], __normals[face_norm_idx_3 + 1], __normals[face_norm_idx_3 + 2]);
        var average_normal = normal_1.add(normal_2).add(normal_3).normalize();

        var face_idx_1 = __idxvertices[0] * 3;
        var face_idx_2 = __idxvertices[1] * 3;
        var face_idx_3 = __idxvertices[2] * 3;
        var v_1 = new THREE.Vector3(__vertices [face_idx_1], __vertices [face_idx_1 + 1], __vertices [face_idx_1 + 2]);
        var v_2 = new THREE.Vector3(__vertices [face_idx_2], __vertices [face_idx_2 + 1], __vertices [face_idx_2 + 2]);
        var v_3 = new THREE.Vector3(__vertices [face_idx_3], __vertices [face_idx_3 + 1], __vertices [face_idx_3 + 2]);
        var vecv1v2 = v_2.sub(v_1).normalize();
        var vecv1v3 = v_3.sub(v_1).normalize();
        return vecv1v2.cross(vecv1v3).dot(average_normal);
    };

    const ColorFromHexa = function (__hexastring)
    {
        var color = parseInt(__hexastring, 16);//From hexadecimal
        var red = (color & 0xFF0000) >> 16;
        var green = (color & 0xFF00) >> 8;
        var blue = (color & 0xFF);
        return {r: red, g: green, b: blue};
    };

    const GetColor = function (__face_idx, __sources, __color_input)
    {
        var colors = __sources[__color_input["source"]];
        var idx_col = __color_input["indices"];
        var ci = __face_idx % idx_col.length;
        return ColorFromHexa(colors[idx_col[ci]]);
    };

    const DotNormalsInterleaved = function (__idx, __normals, __vertices, __chanoffset)
    {
        var face_norm_idx_1 = __idx[1] * 3;
        var face_norm_idx_2 = __idx[1 + __chanoffset] * 3;
        var face_norm_idx_3 = __idx[1 + 2 * __chanoffset] * 3;
        var normal_1 = new THREE.Vector3(__normals[face_norm_idx_1], __normals[face_norm_idx_1 + 1], __normals[face_norm_idx_1 + 2]);
        var normal_2 = new THREE.Vector3(__normals[face_norm_idx_2], __normals[face_norm_idx_2 + 1], __normals[face_norm_idx_2 + 2]);
        var normal_3 = new THREE.Vector3(__normals[face_norm_idx_3], __normals[face_norm_idx_3 + 1], __normals[face_norm_idx_3 + 2]);
        var average_normal = normal_1.add(normal_2).add(normal_3).normalize();

        var face_idx_1 = __idx[0] * 3;
        var face_idx_2 = __idx[__chanoffset] * 3;
        var face_idx_3 = __idx[2 * __chanoffset] * 3;
        var v_1 = new THREE.Vector3(__vertices [face_idx_1], __vertices [face_idx_1 + 1], __vertices [face_idx_1 + 2]);
        var v_2 = new THREE.Vector3(__vertices [face_idx_2], __vertices [face_idx_2 + 1], __vertices [face_idx_2 + 2]);
        var v_3 = new THREE.Vector3(__vertices [face_idx_3], __vertices [face_idx_3 + 1], __vertices [face_idx_3 + 2]);
        var vecv1v2 = v_2.sub(v_1).normalize();
        var vecv1v3 = v_3.sub(v_1).normalize();
        return vecv1v2.cross(vecv1v3).dot(average_normal);
    };

    const CreateCuboidFaces = function (__geometry, __properties)
    {
        return function ()
        {
            //POSITIONS
            var lengthHalf = __properties.length * 0.5;
            var widthHalf = __properties.width * 0.5;
            var heightHalf = __properties.height * 0.5;
            var boxPositions = [
                // Front +Z
                -1, 1, 1,
                1, 1, 1,
                -1, -1, 1,
                1, -1, 1,
                // Back -Z
                1, 1, -1,
                -1, 1, -1,
                1, -1, -1,
                -1, -1, -1,
                // Left -X
                -1, 1, -1,
                -1, 1, 1,
                -1, -1, -1,
                -1, -1, 1,
                // Right +X
                1, 1, 1,
                1, 1, -1,
                1, -1, 1,
                1, -1, -1,
                // Top +Y
                -1, 1, 1,
                1, 1, 1,
                -1, 1, -1,
                1, 1, -1,
                // Bottom -Y
                1, -1, 1,
                -1, -1, 1,
                1, -1, -1,
                -1, -1, -1
            ];
            var positions = boxPositions.map(function (currentValue, index)
            {
                switch (index % 3)
                {
                    case 0:
                        return currentValue * lengthHalf;
                    case 1:
                        return currentValue * widthHalf;
                    case 2:
                        return currentValue * heightHalf;
                }
            });
            var vertices = new THREE.BufferAttribute(new Float32Array(positions), 3);
            __geometry.addAttribute("position", vertices);

            //TEXTURE COORDINATES
            var athird = 1 / 3;
            var twothirds = athird * 2;
            var ahalf = 0.5;
            var uvs = new THREE.BufferAttribute(new Float32Array([
                // Front
                0, athird,
                ahalf, athird,
                0, 0,
                ahalf, 0,
                // Back
                ahalf, athird,
                1, athird,
                ahalf, 0,
                1, 0,
                // Left
                ahalf, 1,
                1, 1,
                ahalf, twothirds,
                1, twothirds,
                // Right
                0, 1,
                ahalf, 1,
                0, twothirds,
                ahalf, twothirds,
                // Top
                0, athird,
                ahalf, athird,
                0, twothirds,
                ahalf, twothirds,
                // Bottom
                ahalf, athird,
                1, athird,
                ahalf, twothirds,
                1, twothirds
            ]), 2);
            __geometry.addAttribute("uv", uvs);

            if (__properties.hasOwnProperty("colors") && (__properties["colors"] !== null))
            {
                //The cube has 24 vertices, so we need to couple these
                //to get the real cube face index and map that number
                //to the corresponding index in __properties.color list
                var colorarray = [];
                for (var i = 0; i < 24; i++)
                {
                    var face_index = Math.floor(i / 4);
                    var color_index;
                    switch (face_index)
                    {
                        case 0:
                            color_index = 2;
                            break;//+Z face
                        case 1:
                            color_index = 5;
                            break;//-Z face
                        case 2:
                            color_index = 3;
                            break;//-X face
                        case 3:
                            color_index = 0;
                            break;//+X face
                        case 4:
                            color_index = 1;
                            break;//+Y face
                        case 5:
                            color_index = 4;
                            break;//-Y face
                    }
                    var face_color = __properties.colors[color_index % __properties.colors.length];
                    var rgb = ColorFromHexa(face_color);
                    colorarray.push(rgb.r / 255);
                    colorarray.push(rgb.g / 255);
                    colorarray.push(rgb.b / 255);
                }
                var colors = new THREE.BufferAttribute(new Float32Array(colorarray), 3);
                __geometry.addAttribute("color", colors);
            }

            //FACES INDEXES
            var indices = new Uint16Array([
                0, 2, 1,
                2, 3, 1,
                4, 6, 5,
                6, 7, 5,
                8, 10, 9,
                10, 11, 9,
                12, 14, 13,
                14, 15, 13,
                16, 17, 18,
                18, 17, 19,
                20, 21, 22,
                22, 21, 23
            ]);
            __geometry.setIndex(new THREE.BufferAttribute(indices, 1));

            //NORMALS
            var normals = new THREE.BufferAttribute(new Float32Array([
                // Front
                0, 0, 1,
                0, 0, 1,
                0, 0, 1,
                0, 0, 1,
                // Back
                0, 0, -1,
                0, 0, -1,
                0, 0, -1,
                0, 0, -1,
                // Left
                -1, 0, 0,
                -1, 0, 0,
                -1, 0, 0,
                -1, 0, 0,
                // Right
                1, 0, 0,
                1, 0, 0,
                1, 0, 0,
                1, 0, 0,
                // Top
                0, 1, 0,
                0, 1, 0,
                0, 1, 0,
                0, 1, 0,
                // Bottom
                0, -1, 0,
                0, -1, 0,
                0, -1, 0,
                0, -1, 0
            ]), 3);

            __geometry.addAttribute("normal", normals);

        };
    };

    const CreateCuboidHardEdges = function (__geometry, __properties)
    {
        return function ()
        {
            //LINES FOR EDGE HIGHLIGHT
            var sx = -__properties.length / 2, sy = -__properties.width / 2, sz = -__properties.height / 2;
            var ex = __properties.length / 2, ey = __properties.width / 2, ez = __properties.height / 2;
            var fblc = [sx, sy, sz];
            var fbrc = [ex, sy, sz];
            var fulc = [sx, sy, ez];
            var furc = [ex, sy, ez];
            var oblc = [sx, ey, sz];
            var obrc = [ex, ey, sz];
            var oulc = [sx, ey, ez];
            var ourc = [ex, ey, ez];

            // LINE PIECES VERSION
            var positions = [].concat(fblc, fulc, fulc, furc, furc, fbrc, fulc, oulc,
                oulc, ourc, ourc, furc, oulc, oblc, oblc, obrc,
                obrc, ourc, oblc, fblc, fblc, fbrc, fbrc, obrc);

            // LINE STRIP VERSION (if needed, just change THREE.LinePieces to THREE.LineStrip at line 438)
            // var positions = [].concat(fblc, fbrc, fulc, furc, oblc, obrc, oulc, ourc);
            // var indices_array = [];
            // indices_array.push(0, 2, 3, 1, 0, 4, 5, 1, 3, 7, 6, 4, 5, 7, 6, 2);
            // var indexes = new THREE.BufferAttribute( new Uint16Array( indices_array ), 1 );
            // geometry.setIndex( indexes );

            var vertices = new THREE.BufferAttribute(new Float32Array(positions), 3);
            __geometry.addAttribute("position", vertices);
        };
    };

    const CreateMeshFaces = function (__geometry, __sources, __surface)
    {
        return function ()
        {
            var __position_input = __surface["input"]["position"];
            var __interleaved_input = __surface["input"]["interleaved"];
            var __color_input = __surface["input"]["color"];
            var __normal_input = __surface["input"]["normal"];
            var __texcoord_input = __surface["input"]["texcoord"];
            var __smoothe_normals = (__normal_input && __normal_input.hasOwnProperty("smooth") && WV3DUtils.ToBool(__normal_input["smooth"]));
            if (__interleaved_input)
            {
                ///INTERLEAVED DATA
                var vertices = __sources[__interleaved_input["sources"][0]];
                var idx = __interleaved_input["indices"];
                var normals = __sources[__interleaved_input["sources"][1]];
                var texcoord = null;
                var largestTrianglesCache = [];

                var vertex_buffer = [];
                var normal_buffer = [];
                var uvs = [];
                var uvs2 = [];
                var color_buffer = [];
                var index_buffer = [];

                var NBCHAN, CHANOFFSET;

                switch (__interleaved_input["format"])
                {
                    default:
                        LogMessage.Log("ThreeDObjectHandler Warning: Unrecognized interleaved format, switching to default PNT");
                    case "PNT":
                        texcoord = __sources[__interleaved_input["sources"][2]];
                        NBCHAN = 9;//3 indexes for vertices index + 3 for normal index + 3 for texture coordinates indexes
                        CHANOFFSET = 3;//3 indexes per type
                        break;

                    case "PN" :
                        NBCHAN = 6;//3 indexes for vertices index + 3 for normal index + 0 for texture coordinates indexes
                        CHANOFFSET = 2;//2 indexes per type
                        break;
                }
                switch (__surface["topology"])
                {
                    default:
                        LogMessage.Log("ThreeDObjectHandler Warning: Unrecognized topology type. Switching to default triangle soup");
                    case "TRIANGLE_SOUP":
                        //Check that winding is correct (CCW) using the first normal of the surface
                        var dot_normals = DotNormalsInterleaved(idx, normals, vertices, CHANOFFSET);
                        if (dot_normals == 0)
                        {
                            LogMessage.Log("ThreeDObjectHandler Warning: Problem in normal computation for winding");
                        }
                        if (dot_normals < 0 && __surface["winding"] == "CCW" ||
                            dot_normals > 0 && __surface["winding"] == "CW")
                        {
                            LogMessage.Log("ThreeDObjectHandler Warning: Surface winding is not coherent with input normals, please check input file.");
                        }

                        switch (__surface["winding"])
                        {
                            default:
                            case "CCW":
                                dot_normals = 1;
                                break;
                            case "CW":
                                dot_normals = -1;
                                break;
                        }

                        var vertexNormals = {};
                        for (var f = 0; f < __surface.count; ++f)
                        {
                            if (dot_normals < 0)
                            {
                                //Swap vertices index
                                idx[f * NBCHAN + CHANOFFSET] = [idx[f * NBCHAN + CHANOFFSET * 2], idx[f * NBCHAN + CHANOFFSET * 2] = idx[f * NBCHAN + CHANOFFSET]][0];
                                //Swap normal index
                                idx[f * NBCHAN + CHANOFFSET + 1] = [idx[f * NBCHAN + CHANOFFSET * 2 + 1], idx[f * NBCHAN + CHANOFFSET * 2 + 1] = idx[f * NBCHAN + CHANOFFSET + 1]][0];
                                //Swap texture index
                                if (CHANOFFSET >= 3) idx[f * NBCHAN + CHANOFFSET + 2] = [idx[f * NBCHAN + CHANOFFSET * 2 + 2], idx[f * NBCHAN + CHANOFFSET * 2 + 2] = idx[f * NBCHAN + CHANOFFSET + 2]][0];
                            }

                            var v0x, v0y, v0z, v1x, v1y, v1z, v2x, v2y, v2z;
                            v0x = vertices[idx[NBCHAN * f] * 3];
                            v0y = vertices[idx[NBCHAN * f] * 3 + 1];
                            v0z = vertices[idx[NBCHAN * f] * 3 + 2];
                            v1x = vertices[idx[NBCHAN * f + CHANOFFSET] * 3];
                            v1y = vertices[idx[NBCHAN * f + CHANOFFSET] * 3 + 1];
                            v1z = vertices[idx[NBCHAN * f + CHANOFFSET] * 3 + 2];
                            v2x = vertices[idx[NBCHAN * f + CHANOFFSET * 2] * 3];
                            v2y = vertices[idx[NBCHAN * f + CHANOFFSET * 2] * 3 + 1];
                            v2z = vertices[idx[NBCHAN * f + CHANOFFSET * 2] * 3 + 2];

                            var vertex_array = [v0x, v0y, v0z, v1x, v1y, v1z, v2x, v2y, v2z];
                            vertex_buffer.push.apply(vertex_buffer, vertex_array);

                            if (__smoothe_normals)
                            {
                                var accNormal = function (__idx, __vertexNormals, __normal)
                                {
                                    if (!vertexNormals.hasOwnProperty(__idx))
                                    {
                                        vertexNormals[__idx] = __normal;
                                    }
                                    else
                                    {
                                        vertexNormals[__idx].add(__normal);
                                    }
                                };
                                var v0 = new THREE.Vector3(v0x, v0y, v0z);
                                var v1 = new THREE.Vector3(v1x, v1y, v1z);
                                var v2 = new THREE.Vector3(v2x, v2y, v2z);
                                var v0v1 = v1.clone().sub(v0);
                                var v0v2 = v2.clone().sub(v0);
                                var normal = v0v1.cross(v0v2);//If no weighted average is required, normalize this cross product instead
                                accNormal(idx[NBCHAN * f], vertexNormals, normal);
                                accNormal(idx[NBCHAN * f + CHANOFFSET], vertexNormals, normal);
                                accNormal(idx[NBCHAN * f + CHANOFFSET * 2], vertexNormals, normal);
                            }
                            else
                            {
                                var normal_array = [normals[idx[NBCHAN * f + 1] * 3],
                                    normals[idx[NBCHAN * f + 1] * 3 + 1],
                                    normals[idx[NBCHAN * f + 1] * 3 + 2],
                                    normals[idx[NBCHAN * f + CHANOFFSET + 1] * 3],
                                    normals[idx[NBCHAN * f + CHANOFFSET + 1] * 3 + 1],
                                    normals[idx[NBCHAN * f + CHANOFFSET + 1] * 3 + 2],
                                    normals[idx[NBCHAN * f + CHANOFFSET * 2 + 1] * 3],
                                    normals[idx[NBCHAN * f + CHANOFFSET * 2 + 1] * 3 + 1],
                                    normals[idx[NBCHAN * f + CHANOFFSET * 2 + 1] * 3 + 2]];
                                normal_buffer.push.apply(normal_buffer, normal_array);
                            }

                            var texc_array;
                            if (texcoord)
                            {

                                var v1u = texcoord[idx[NBCHAN * f + 2] * 2];
                                var v1v = texcoord[idx[NBCHAN * f + 2] * 2 + 1];
                                var v2u = texcoord[idx[NBCHAN * f + CHANOFFSET + 2] * 2];
                                var v2v = texcoord[idx[NBCHAN * f + CHANOFFSET + 2] * 2 + 1];
                                var v3u = texcoord[idx[NBCHAN * f + CHANOFFSET * 2 + 2] * 2];
                                var v3v = texcoord[idx[NBCHAN * f + CHANOFFSET * 2 + 2] * 2 + 1];

                                texc_array = [v1u, v1v, v2u, v2v, v3u, v3v];

                                var p1 = new THREE.Vector3(v0x, v0y, v0z);
                                var p2 = new THREE.Vector3(v1x, v1y, v1z);
                                var p3 = new THREE.Vector3(v2x, v2y, v2z);

                                CalculatedPdUdPdVForLargestTriangles(p1, p2, p3, v1u, v1v, v2u, v2v, v3u, v3v, largestTrianglesCache);
                            }
                            else
                            {
                                texc_array = [0, 0, 0, 0, 0, 0];
                            }
                            uvs.push.apply(uvs, texc_array);
                            uvs2.push.apply(uvs2, texc_array);


                            if (__color_input)
                            {
                                var rgb = GetColor(f, __sources, __color_input);
                                color_buffer.push.apply(color_buffer, [rgb.r / 255, rgb.g / 255, rgb.b / 255]);
                                color_buffer.push.apply(color_buffer, [rgb.r / 255, rgb.g / 255, rgb.b / 255]);
                                color_buffer.push.apply(color_buffer, [rgb.r / 255, rgb.g / 255, rgb.b / 255]);
                            }
                        }
                        //We indeed calculate the average dUdP and dVdP to ensure stability since
                        //some textures are not mapped linearly (shrink sleeves for example).
                        var dPdUdPdV = GetAveragedPdUdPdV(largestTrianglesCache);
                        __geometry.dUdP = dPdUdPdV.x;
                        __geometry.dVdP = dPdUdPdV.y;

                        if (__smoothe_normals)
                        {
                            for (var j = 0, keys = Object.keys(vertexNormals), l = keys.length; j < l; ++j)
                            {
                                var elt = keys[j];
                                vertexNormals[elt].normalize();
                            }
                            for (var f = 0; f < __surface.count; ++f)
                            {

                                var normal_array = [vertexNormals[idx[NBCHAN * f]].x,
                                    vertexNormals[idx[NBCHAN * f]].y,
                                    vertexNormals[idx[NBCHAN * f]].z,
                                    vertexNormals[idx[NBCHAN * f + CHANOFFSET]].x,
                                    vertexNormals[idx[NBCHAN * f + CHANOFFSET]].y,
                                    vertexNormals[idx[NBCHAN * f + CHANOFFSET]].z,
                                    vertexNormals[idx[NBCHAN * f + CHANOFFSET * 2]].x,
                                    vertexNormals[idx[NBCHAN * f + CHANOFFSET * 2]].y,
                                    vertexNormals[idx[NBCHAN * f + CHANOFFSET * 2]].z];
                                normal_buffer.push.apply(normal_buffer, normal_array);
                            }
                        }
                        break;
                }

                //We do not use index arrays since vertices, normals, and tex coordinates can have different indexes
                //Therefore we build full arrays for each of these attributes (a bit of a shame but WebGL wouldn't support it otherwise)
                __geometry.addAttribute("position", new THREE.BufferAttribute(new Float32Array(vertex_buffer), 3));
                __geometry.addAttribute("normal", new THREE.BufferAttribute(new Float32Array(normal_buffer), 3));
                if (__color_input) __geometry.addAttribute("color", new THREE.BufferAttribute(new Float32Array(color_buffer), 3));
                __geometry.addAttribute("uv", new THREE.BufferAttribute(new Float32Array(uvs), 2));
                __geometry.addAttribute("uv2", new THREE.BufferAttribute(new Float32Array(uvs2), 2));

                __geometry.uvsNeedUpdate = true;

            }
            else
            {
                ///SURFACE WINDING
                //Check winding (Only when data is not interleaved, otherwise see previous block)
                var dot_normals = 1;
                if (__normal_input && __position_input &&
                    __normal_input.hasOwnProperty("indices") && __normal_input["indices"] && __normal_input["indices"].length >= 3 && __sources[__normal_input["source"]] &&
                    __position_input.hasOwnProperty("indices") && __position_input["indices"] && __position_input["indices"].length >= 3 && __sources[__position_input["source"]])
                {
                    var positions = __sources[__position_input["source"]];
                    var normals = __sources[__normal_input["source"]];
                    var idx_vert = __position_input["indices"];
                    var idx_norm = __normal_input["indices"];

                    //Check that winding is correct (CCW) using the first normal of the surface
                    dot_normals = DotNormals(idx_norm, idx_vert, normals, positions);

                    if (dot_normals == 0)
                    {
                        LogMessage.Log("ThreeDObjectHandler Warning: Problem in normal computation for winding");
                    }
                    if (dot_normals < 0 && __surface["winding"] == "CCW" || dot_normals > 0 && __surface["winding"] == "CW")
                    {
                        LogMessage.Log("ThreeDObjectHandler Warning: Surface winding is not coherent with input normals, please check input file.");
                    }
                }

                switch (__surface["winding"])
                {
                    default:
                    case "CCW":
                        dot_normals = 1;
                        break;
                    case "CW":
                        dot_normals = -1;
                        break;
                }

                ///POSITIONS
                var hasTextureCoordinates = __texcoord_input && __texcoord_input["indices"] && __texcoord_input["indices"].length && __sources[__texcoord_input["source"]];
                var positionVec3s = [];
                if (__position_input)
                {
                    var vertices = __sources[__position_input["source"]];
                    var idx = __position_input["indices"];

                    var vertex_buffer = [];
                    var color_buffer = [];

                    switch (__surface["topology"])
                    {
                        default:
                            LogMessage.Log("ThreeDObjectHandler Warning: Unrecognized topology type. Switching to default triangle soup");
                        case "TRIANGLE_SOUP":
                            for (var f = 0; f < __surface.count; ++f)
                            {
                                if (dot_normals < 0)
                                {
                                    //Swap second and third indexes (see http://stackoverflow.com/questions/16201656/how-to-swap-two-variables-in-javascript)
                                    idx[f * 3 + 2] = [idx[f * 3 + 1], idx[f * 3 + 1] = idx[f * 3 + 2]][0];
                                }

                                var v1x = vertices[parseInt(idx[f * 3]) * 3];
                                var v1y = vertices[parseInt(idx[f * 3]) * 3 + 1];
                                var v1z = vertices[parseInt(idx[f * 3]) * 3 + 2];
                                var v2x = vertices[parseInt(idx[f * 3 + 1]) * 3];
                                var v2y = vertices[parseInt(idx[f * 3 + 1]) * 3 + 1];
                                var v2z = vertices[parseInt(idx[f * 3 + 1]) * 3 + 2];
                                var v3x = vertices[parseInt(idx[f * 3 + 2]) * 3];
                                var v3y = vertices[parseInt(idx[f * 3 + 2]) * 3 + 1];
                                var v3z = vertices[parseInt(idx[f * 3 + 2]) * 3 + 2];

                                var vertex_array = [v1x, v1y, v1z, v2x, v2y, v2z, v3x, v3y, v3z];

                                if (vertex_array[0] === undefined || vertex_array[1] === undefined ||
                                    vertex_array[2] === undefined || vertex_array[3] === undefined ||
                                    vertex_array[4] === undefined || vertex_array[5] === undefined ||
                                    vertex_array[6] === undefined || vertex_array[7] === undefined ||
                                    vertex_array[8] === undefined)
                                {
                                    LogMessage.Log("ThreeDObjectHandler Warning: index out of vertices array bounds");
                                }
                                else
                                {
                                    vertex_buffer.push.apply(vertex_buffer, vertex_array);
                                }

                                if (hasTextureCoordinates)
                                {
                                    var v1 = new THREE.Vector3(v1x, v1y, v1z);
                                    var v2 = new THREE.Vector3(v2x, v2y, v2z);
                                    var v3 = new THREE.Vector3(v3x, v3y, v3z);
                                    positionVec3s.push(v1, v2, v3);
                                }

                                if (__color_input)
                                {
                                    var rgb = GetColor(f, __sources, __color_input);
                                    color_buffer.push.apply(color_buffer, [rgb.r / 255, rgb.g / 255, rgb.b / 255]);
                                    color_buffer.push.apply(color_buffer, [rgb.r / 255, rgb.g / 255, rgb.b / 255]);
                                    color_buffer.push.apply(color_buffer, [rgb.r / 255, rgb.g / 255, rgb.b / 255]);
                                }
                            }
                            break;
                    }

                    __geometry.addAttribute("position", new THREE.BufferAttribute(new Float32Array(vertex_buffer), 3));
                    if (__color_input) __geometry.addAttribute("color", new THREE.BufferAttribute(new Float32Array(color_buffer), 3));
                }
                ///NORMALS (FROM FILE OR CALCULATED ON THE FLY)
                var __normal_input = __surface["input"]["normal"];
                if (__normal_input)
                {
                    if (__normal_input.hasOwnProperty("smooth") && WV3DUtils.ToBool(__normal_input["smooth"]))
                    {
                        var vertices = __sources[__position_input["source"]];
                        var idx = __position_input["indices"];//we use the vertices index array, not the normals index array
                        var normals_buffer = [];
                        var vertexNormals = {};

                        var vA, vB, vC,

                            pA = new THREE.Vector3(),
                            pB = new THREE.Vector3(),
                            pC = new THREE.Vector3(),

                            cb = new THREE.Vector3(),
                            ab = new THREE.Vector3();
                        for (var f = 0; f < __surface.count; ++f)
                        {
                            //For each vertex of the face, we add the face normal to the vertex normal in the global vertex normal list
                            //We also add 1 to the number of faces that are linked to this vertex.

                            vA = idx[f * 3] * 3;
                            vB = idx[f * 3 + 1] * 3;
                            vC = idx[f * 3 + 2] * 3;

                            pA.fromArray(vertices, vA);
                            pB.fromArray(vertices, vB);
                            pC.fromArray(vertices, vC);

                            cb.subVectors(pC, pB);
                            ab.subVectors(pA, pB);
                            cb.cross(ab);//If no weighted average is required, normalize this cross product instead

                            var accVertexNormal = function (vNidx, vNacc)
                            {
                                if (!vertexNormals.hasOwnProperty(vNidx))
                                {
                                    vertexNormals[vNidx] = vNacc.clone();
                                }
                                else
                                {
                                    vertexNormals[vNidx].add(vNacc);
                                }
                            };
                            accVertexNormal(vA, cb);
                            accVertexNormal(vB, cb);
                            accVertexNormal(vC, cb);
                        }

                        for (var i = 0, keys = Object.keys(vertexNormals), l = keys.length; i < l; ++i)
                        {
                            vertexNormals[keys[i]].normalize();
                        }

                        for (var f = 0; f < __surface.count; ++f)
                        {
                            //For each vertex of the face, we retrieved the modulated normal of the vertex
                            //which has was calculated using neighbor faces normals

                            vA = idx[f * 3] * 3;
                            vB = idx[f * 3 + 1] * 3;
                            vC = idx[f * 3 + 2] * 3;

                            var normals_array = [vertexNormals[vA].x,
                                vertexNormals[vA].y,
                                vertexNormals[vA].z,
                                vertexNormals[vB].x,
                                vertexNormals[vB].y,
                                vertexNormals[vB].z,
                                vertexNormals[vC].x,
                                vertexNormals[vC].y,
                                vertexNormals[vC].z];

                            normals_buffer.push.apply(normals_buffer, normals_array);
                        }

                        __geometry.addAttribute("normal", new THREE.BufferAttribute(new Float32Array(normals_buffer), 3));
                    }
                    else if (__normal_input.hasOwnProperty("indices") && __normal_input["indices"] && __normal_input["indices"].length && __sources[__normal_input["source"]])
                    {
                        var normals = __sources[__normal_input["source"]];
                        var normals_index = __normal_input["indices"];
                        var normals_buffer = [];
                        for (var f = 0; f < __surface.count; ++f)
                        {
                            if (dot_normals < 0)
                            {
                                //Swap second and thrid indexes (see http://stackoverflow.com/questions/16201656/how-to-swap-two-variables-in-javascript)
                                normals_index[f * 3 + 2] = [normals_index[f * 3 + 1], normals_index[f * 3 + 1] = normals_index[f * 3 + 2]][0];
                            }

                            var normals_array = [normals[normals_index[f * 3] * 3],
                                normals[normals_index[f * 3] * 3 + 1],
                                normals[normals_index[f * 3] * 3 + 2],
                                normals[normals_index[f * 3 + 1] * 3],
                                normals[normals_index[f * 3 + 1] * 3 + 1],
                                normals[normals_index[f * 3 + 1] * 3 + 2],
                                normals[normals_index[f * 3 + 2] * 3],
                                normals[normals_index[f * 3 + 2] * 3 + 1],
                                normals[normals_index[f * 3 + 2] * 3 + 2]];
                            normals_buffer.push.apply(normals_buffer, normals_array);
                        }

                        __geometry.addAttribute("normal", new THREE.BufferAttribute(new Float32Array(normals_buffer), 3));
                    }
                    else if (__normal_input.hasOwnProperty("smooth") && !WV3DUtils.ToBool(__normal_input["smooth"]))
                    {
                        //Compute unsmooth normal buffer

                        var vertices = __sources[__position_input["source"]];
                        var idx = __position_input["indices"];//we use the vertices index array, not the normals index array
                        var normals_buffer = [];

                        var vA, vB, vC,

                            pA = new THREE.Vector3(),
                            pB = new THREE.Vector3(),
                            pC = new THREE.Vector3(),

                            cb = new THREE.Vector3(),
                            ab = new THREE.Vector3();
                        for (var f = 0; f < __surface.count; ++f)
                        {
                            vA = idx[f * 3] * 3;
                            vB = idx[f * 3 + 1] * 3;
                            vC = idx[f * 3 + 2] * 3;

                            pA.fromArray(vertices, vA);
                            pB.fromArray(vertices, vB);
                            pC.fromArray(vertices, vC);

                            cb.subVectors(pC, pB);
                            ab.subVectors(pA, pB);
                            cb.cross(ab);
                            cb.normalize();

                            var normals_array = [cb.x, cb.y, cb.z, cb.x, cb.y, cb.z, cb.x, cb.y, cb.z];

                            normals_buffer.push.apply(normals_buffer, normals_array);
                        }

                        __geometry.addAttribute("normal", new THREE.BufferAttribute(new Float32Array(normals_buffer), 3));
                    }
                    else
                    {
                        LogMessage.Log("Problem parsing json: no normal computation can be infered");
                    }
                }
                else
                {
                    __geometry.computeVertexNormals();
                }

                ///TEXTURE COORDINATES
                var uvs = [];
                var uvs2 = [];

                if (hasTextureCoordinates)
                {
                    var texcoord = __sources[__texcoord_input["source"]];
                    var texcoord_index = __texcoord_input["indices"];

                    var texcoord_buffer = [];
                    var largestTrianglesCache = [];
                    for (f = 0; f < __surface.count; f++)
                    {
                        if (dot_normals < 0)
                        {
                            //Swap second and thrid indexes (see http://stackoverflow.com/questions/16201656/how-to-swap-two-variables-in-javascript)
                            texcoord_index[f * 3 + 2] = [texcoord_index[f * 3 + 1], texcoord_index[f * 3 + 1] = texcoord_index[f * 3 + 2]][0];
                        }

                        var v1u = texcoord[texcoord_index[f * 3] * 2];
                        var v1v = texcoord[texcoord_index[f * 3] * 2 + 1];
                        var v2u = texcoord[texcoord_index[f * 3 + 1] * 2];
                        var v2v = texcoord[texcoord_index[f * 3 + 1] * 2 + 1];
                        var v3u = texcoord[texcoord_index[f * 3 + 2] * 2];
                        var v3v = texcoord[texcoord_index[f * 3 + 2] * 2 + 1];

                        var texc_array = [v1u, v1v, v2u, v2v, v3u, v3v];
                        uvs.push.apply(uvs, texc_array);
                        uvs2.push.apply(uvs2, texc_array);

                        var p1 = positionVec3s[f * 3];
                        var p2 = positionVec3s[f * 3 + 1];
                        var p3 = positionVec3s[f * 3 + 2];

                        CalculatedPdUdPdVForLargestTriangles(p1, p2, p3, v1u, v1v, v2u, v2v, v3u, v3v, largestTrianglesCache);
                    }
                    //We indeed calculate the average dUdP and dVdP to ensure stability since
                    //some textures are not mapped linearly (shrink sleeves for example).
                    var dPdUdPdV = GetAveragedPdUdPdV(largestTrianglesCache);
                    __geometry.dUdP = dPdUdPdV.x;
                    __geometry.dVdP = dPdUdPdV.y;
                    __geometry.uvsNeedUpdate = true;

                }
                else
                {
                    for (f = 0; f < __surface.count * 3; f++)
                    {
                        uvs.push.apply(uvs, [0, 0]);
                        uvs2.push.apply(uvs2, [0, 0]);
                    }
                }

                __geometry.addAttribute("uv", new THREE.BufferAttribute(new Float32Array(uvs), 2));
                __geometry.addAttribute("uv2", new THREE.BufferAttribute(new Float32Array(uvs2), 2));

                //We do not use index arrays since vertices, normals, and tex coordinates can have different indexes
                //Therefore we build full arrays for each of these attributes (a bit of a shame but WebGL wouldn't support it otherwise)
                //Because of that limitation, we prefer to build neighbor info or any other mesh navigational data


            }
        };
    };

    const AttributesInit = function (__geometry, __proxy_geometry)
    {
        return function ()
        {
            //Copy references to mesh buffers
            if (__proxy_geometry.attributes.position) __geometry.attributes.position = __proxy_geometry.attributes.position;
            if (__proxy_geometry.attributes.normal) __geometry.attributes.normal = __proxy_geometry.attributes.normal;
            if (__proxy_geometry.attributes.color) __geometry.attributes.color = __proxy_geometry.attributes.color;
            if (__proxy_geometry.attributes.uv) __geometry.attributes.uv = __proxy_geometry.attributes.uv;
            if (__proxy_geometry.attributes.uv2) __geometry.attributes.uv2 = __proxy_geometry.attributes.uv2;

            //Copy texture scaling factor
            if (__proxy_geometry.dUdP) __geometry.dUdP = __proxy_geometry.dUdP;
            if (__proxy_geometry.dVdP) __geometry.dVdP = __proxy_geometry.dVdP;

            //Copy reference to hard edge buffer
            if (__proxy_geometry.adjacentfacesedgeslines.attributes.position) __geometry.adjacentfacesedgeslines.attributes.position = __proxy_geometry.adjacentfacesedgeslines.attributes.position;
        };
    };

    const CreateMeshHardEdges = function (__geometry, __sources, __surface, __coplanarity_epsilon, __enable_boundary_hard_edges)
    {
        return function ()
        {
            var __position_input = __surface["input"]["position"];
            var __interleaved_input = __surface["input"]["interleaved"];
            var __color_input = __surface["input"]["color"];
            var __normal_input = __surface["input"]["normal"];
            var __texcoord_input = __surface["input"]["texcoord"];
            //No hard edges for smooth surfaces since normals are equal at the edges
            var enabled_hard_edges_build = !(__normal_input && __normal_input.hasOwnProperty("smooth") && WV3DUtils.ToBool(__normal_input["smooth"]));
            var uniqueedges = {};
            if (enabled_hard_edges_build || __enable_boundary_hard_edges)
            {
                var edges = {};
                for (var i = 0; i < __surface.count; ++i)
                {
                    var face = i;

                    var idxvert1, idxvert2, idxvert3;
                    if (__position_input)
                    {
                        var idx = __position_input["indices"];
                        idxvert1 = parseInt(idx[face * 3]);
                        idxvert2 = parseInt(idx[face * 3 + 1]);
                        idxvert3 = parseInt(idx[face * 3 + 2]);
                    }
                    else if (__interleaved_input)
                    {
                        var NBCHAN, CHANOFFSET;
                        switch (__interleaved_input["format"])
                        {
                            default:
                                LogMessage.Log("ThreeDObjectHandler Warning: Unrecognized interleaved format, switching to default PNT");
                            case "PNT":
                                var texcoord = __sources[__interleaved_input["sources"][2]];
                                var NBCHAN = 9;//3 indexes for vertices index + 3 for normal index + 3 for texture coordinates indexes
                                var CHANOFFSET = 3;//3 indexes per type
                                break;

                            case "PN" :
                                var NBCHAN = 6;//3 indexes for vertices index + 3 for normal index + 0 for texture coordinates indexes
                                var CHANOFFSET = 2;//2 indexes per type
                                break;
                        }
                        var idx = __interleaved_input["indices"];
                        idxvert1 = parseInt(idx[face * NBCHAN]);
                        idxvert2 = parseInt(idx[face * NBCHAN + CHANOFFSET]);
                        idxvert3 = parseInt(idx[face * NBCHAN + CHANOFFSET * 2]);
                    }

                    var edge1 = {"0": idxvert1, "1": idxvert2, "face1": face, "face2": null, "opedge": null, "edgeIDF1": 0, "edgeIDF2": null};
                    var edge1id = edge1["0"] + "/" + edge1["1"];
                    var edge1idop = edge1["1"] + "/" + edge1["0"];
                    if (edges.hasOwnProperty(edge1idop))//Edge already created: fetch the existing one
                    {
                        edges[edge1id] = edge1;
                        edges[edge1id]["face2"] = edges[edge1idop]["face1"];
                        edges[edge1id]["opedge"] = edges[edge1idop];
                        if (edges[edge1idop]["face2"]) LogMessage.Log("ThreeDObjectHandler Warning: Edge shared with more than two faces");
                        edges[edge1idop]["face2"] = face;
                        edges[edge1idop]["opedge"] = edge1;
                        edges[edge1idop]["edgeIDF2"] = 0;
                    }
                    else if (!edges.hasOwnProperty(edge1id))//Edge not existing: add it to the map
                    {
                        edges[edge1id] = edge1;
                        uniqueedges[edge1id] = edge1;
                    }
                    else
                    {
                        LogMessage.Log("ThreeDObjectHandler Warning: Edge already existing : not treated");
                    }

                    var edge2 = {"0": idxvert2, "1": idxvert3, "face1": face, "face2": null, "opedge": null, "edgeIDF1": 1, "edgeIDF2": null};
                    var edge2id = edge2["0"] + "/" + edge2["1"];
                    var edge2idop = edge2["1"] + "/" + edge2["0"];
                    if (edges.hasOwnProperty(edge2idop))//Edge already created: fetch the existing one
                    {
                        edges[edge2id] = edge2;
                        edges[edge2id]["face2"] = edges[edge2idop]["face1"];
                        edges[edge2id]["opedge"] = edges[edge2idop];
                        if (edges[edge2idop]["face2"]) LogMessage.Log("ThreeDObjectHandler Warning: Edge shared with more than two faces");
                        edges[edge2idop]["face2"] = face;
                        edges[edge2idop]["opedge"] = edge2;
                        edges[edge2idop]["edgeIDF2"] = 1;
                    }
                    else if (!edges.hasOwnProperty(edge2id))//Edge not existing: add it to the map
                    {
                        edges[edge2id] = edge2;
                        uniqueedges[edge2id] = edge2;
                    }
                    else
                    {
                        LogMessage.Log("Warning: Edge already existing : not treated");
                    }

                    var edge3 = {"0": idxvert3, "1": idxvert1, "face1": face, "face2": null, "opedge": null, "edgeIDF1": 2, "edgeIDF2": null};
                    var edge3id = edge3["0"] + "/" + edge3["1"];
                    var edge3idop = edge3["1"] + "/" + edge3["0"];
                    if (edges.hasOwnProperty(edge3idop))//Edge already created: fetch the existing one
                    {
                        edges[edge3id] = edge3;
                        edges[edge3id]["face2"] = edges[edge3idop]["face1"];
                        edges[edge3id]["opedge"] = edges[edge3idop];
                        if (edges[edge3idop]["face2"]) LogMessage.Log("ThreeDObjectHandler Warning: Edge shared with more than two faces");
                        edges[edge3idop]["face2"] = face;
                        edges[edge3idop]["opedge"] = edge3;
                        edges[edge3idop]["edgeIDF2"] = 2;
                    }
                    else if (!edges.hasOwnProperty(edge3id))//Edge not existing: add it to the map
                    {
                        edges[edge3id] = edge3;
                        uniqueedges[edge3id] = edge3;
                    }
                    else
                    {
                        LogMessage.Log("ThreeDObjectHandler Warning: Edge already existing : not treated");
                    }
                }
            }

            var positions = [];
            for (var i = 0, keys = Object.keys(uniqueedges), l = keys.length; i < l; ++i)
            {
                var e = keys[i];
                var edge = uniqueedges[e];
                if (!edge["face2"])//Boundary edges (not shared with any other face)
                {
                    if (__enable_boundary_hard_edges)
                    {
                        var start_point, end_point;
                        var idxvertexS = edge["0"];
                        var idxvertexE = edge["1"];
                        var vertices;
                        if (__position_input)
                        {
                            vertices = __sources[__position_input["source"]];
                        }
                        else if (__interleaved_input)
                        {
                            vertices = __sources[__interleaved_input["sources"][0]];
                        }

                        var offsetS = idxvertexS * 3;
                        var offsetE = idxvertexE * 3;
                        var start_point_x = vertices[offsetS];
                        var start_point_y = vertices[offsetS + 1];
                        var start_point_z = vertices[offsetS + 2];
                        var end_point_x = vertices[offsetE];
                        var end_point_y = vertices[offsetE + 1];
                        var end_point_z = vertices[offsetE + 2];

                        positions.push(start_point_x, start_point_y, start_point_z, end_point_x, end_point_y, end_point_z);
                    }
                }
                else if (enabled_hard_edges_build)
                {
                    //This was the old way of computing smooth edge (considering geometric construction of normal and not input)
                    var coplanar;
                    var normal_average = new THREE.Vector3();
                    if (__normal_input || __interleaved_input)
                    {
                        var normalf1e0, normalf1e1, normalf2e0, normalf2e1;
                        var normals, idx, idxface1norm1, idxface1norm2, idxface2norm1, idxface2norm2;
                        if (__position_input && __normal_input)
                        {
                            normals = __sources[__normal_input["source"]];
                            idx = __normal_input["indices"];
                            idxface1norm1 = edge["face1"] * 3 + edge["edgeIDF1"];
                            idxface1norm2 = edge["face1"] * 3 + ((edge["edgeIDF1"] + 1) % 3);
                            idxface2norm1 = edge["face2"] * 3 + edge["edgeIDF2"];
                            idxface2norm2 = edge["face2"] * 3 + ((edge["edgeIDF2"] + 1) % 3);
                        }
                        else //if(__interleaved_input)
                        {
                            var NBCHAN, CHANOFFSET;
                            switch (__interleaved_input["format"])
                            {
                                default:
                                    LogMessage.Log("ThreeDObjectHandler Warning: Unrecognized interleaved format, switching to default PNT");
                                case "PNT":
                                    var texcoord = __sources[__interleaved_input["sources"][2]];
                                    var NBCHAN = 9;//3 indexes for vertices index + 3 for normal index + 3 for texture coordinates indexes
                                    var CHANOFFSET = 3;//3 indexes per type
                                    break;

                                case "PN" :
                                    var NBCHAN = 6;//3 indexes for vertices index + 3 for normal index + 0 for texture coordinates indexes
                                    var CHANOFFSET = 2;//2 indexes per type
                                    break;
                            }
                            normals = __sources[__interleaved_input["sources"][1]];
                            idx = __interleaved_input["indices"];
                            idxface1norm1 = edge["face1"] * NBCHAN + 1 + CHANOFFSET * edge["edgeIDF1"];
                            idxface1norm2 = edge["face1"] * NBCHAN + 1 + CHANOFFSET * ((edge["edgeIDF1"] + 1) % 3);
                            idxface2norm1 = edge["face2"] * NBCHAN + 1 + CHANOFFSET * edge["edgeIDF2"];
                            idxface2norm2 = edge["face2"] * NBCHAN + 1 + CHANOFFSET * ((edge["edgeIDF2"] + 1) % 3);

                        }

                        var normalf1e0 = new THREE.Vector3();
                        var normalf1e1 = new THREE.Vector3();
                        var normalf2e0 = new THREE.Vector3();
                        var normalf2e1 = new THREE.Vector3();

                        normalf1e0.fromArray(normals, idx[idxface1norm1] * 3);
                        normalf1e1.fromArray(normals, idx[idxface1norm2] * 3);
                        normalf2e0.fromArray(normals, idx[idxface2norm1] * 3);
                        normalf2e1.fromArray(normals, idx[idxface2norm2] * 3);
                        normalf1e0.normalize();
                        normalf1e1.normalize();
                        normalf2e0.normalize();
                        normalf2e1.normalize();

                        coplanar = (Math.abs(Math.min(normalf1e0.dot(normalf2e1), normalf1e1.dot(normalf2e0))) > 1 - __coplanarity_epsilon);
                        normal_average.addVectors(normalf1e0, normalf1e1);
                        normal_average.add(normalf2e0).add(normalf2e1).divideScalar(4);

                    }
                    else
                    {
                        var normal1, normal2;

                        if (__position_input)
                        {
                            var vertices = __sources[__position_input["source"]];
                            var idx = __position_input["indices"];
                            var idxface1v1 = edge["face1"] * 3 + edge["edgeIDF1"];
                            var idxface1v2 = edge["face1"] * 3 + ((edge["edgeIDF1"] + 1) % 3);
                            var idxface1v3 = edge["face1"] * 3 + ((edge["edgeIDF1"] + 2) % 3);
                            var idxface2v1 = edge["face2"] * 3 + edge["edgeIDF2"];
                            var idxface2v2 = edge["face2"] * 3 + ((edge["edgeIDF2"] + 1) % 3);
                            var idxface2v3 = edge["face2"] * 3 + ((edge["edgeIDF2"] + 2) % 3);

                            //normals not smoothed in this case since no normal input data

                            var pA = new THREE.Vector3();
                            var pB = new THREE.Vector3();
                            var pC = new THREE.Vector3();
                            var cb = new THREE.Vector3();
                            var ab = new THREE.Vector3();

                            pA.fromArray(vertices, idx[idxface1v1] * 3);
                            pB.fromArray(vertices, idx[idxface1v2] * 3);
                            pC.fromArray(vertices, idx[idxface1v3] * 3);

                            cb.subVectors(pC, pB);
                            ab.subVectors(pA, pB);
                            cb.cross(ab);
                            cb.normalize();

                            normal1 = cb.clone();

                            pA.fromArray(vertices, idx[idxface2v1] * 3);
                            pB.fromArray(vertices, idx[idxface2v2] * 3);
                            pC.fromArray(vertices, idx[idxface2v3] * 3);

                            cb.subVectors(pC, pB);
                            ab.subVectors(pA, pB);
                            cb.cross(ab);
                            cb.normalize();

                            normal2 = cb.clone();
                        }
                        else
                        {
                            LogMessage.Log("Missing vertex data");
                            break;
                        }

                        coplanar = (Math.abs(normal1.dot(normal2)) > 1 - __coplanarity_epsilon);
                        normal_average.addVectors(normal1, normal2);
                        normal_average.divideScalar(4);
                    }
                    //this condition could be less restrictive (e.g. > 1 - treshold)
                    //for smoothe surfaces
                    // coplanar = false;//debug
                    if (!coplanar)//Hard edge
                    {
                        var start_point, end_point;
                        var idxvertexS = edge["0"];
                        var idxvertexE = edge["1"];
                        var idx, vertices;
                        if (__position_input)
                        {
                            vertices = __sources[__position_input["source"]];
                        }
                        else if (__interleaved_input)
                        {
                            vertices = __sources[__interleaved_input["sources"][0]];
                        }

                        var offsetS = idxvertexS * 3;
                        var offsetE = idxvertexE * 3;
                        var start_point_x = vertices[offsetS];
                        var start_point_y = vertices[offsetS + 1];
                        var start_point_z = vertices[offsetS + 2];
                        var end_point_x = vertices[offsetE];
                        var end_point_y = vertices[offsetE + 1];
                        var end_point_z = vertices[offsetE + 2];

                        positions.push(start_point_x, start_point_y, start_point_z, end_point_x, end_point_y, end_point_z);
                    }
                }
            }

            var vertices = new THREE.BufferAttribute(new Float32Array(positions), 3);
            __geometry.addAttribute("position", vertices);
        };
    };

    //Ordering comparator
    const CompareSurfacesInGlueGroup = function (__orderForFrontView)
    {
        return function (__surfA, __surfB)
        {
            if (parseFloat(__surfA.geometry["coincidentgroup"]) != parseFloat(__surfB.geometry["coincidentgroup"]))
            {
                if (parseFloat(__surfA.geometry["coincidentgroup"]) < 0 && parseFloat(__surfB.geometry["coincidentgroup"]) < 0)
                    return false;
                return (parseFloat(__surfA.geometry["coincidentgroup"]) - parseFloat(__surfB.geometry["coincidentgroup"]));
            }
            else
            {
                if (__orderForFrontView)
                    return (Math.abs(__surfA.geometry["coincidentpriority"]) - Math.abs(__surfB.geometry["coincidentpriority"]));
                else
                    return (Math.abs(__surfB.geometry["coincidentpriority"]) - Math.abs(__surfA.geometry["coincidentpriority"]));
            }
        };
    };

    //Add directly surfaces that are alone in a glue group to render list
    const splitSingleSurfaces = function (__groupedSurfacesSorted)
    {
        var singles = [];
        var sortedNonSingles = [];
        var keys_surfaces = Object.keys(__groupedSurfacesSorted);
        if (keys_surfaces.length != 0)
        {
            if (keys_surfaces.length == 1)
            {
                singles = __groupedSurfacesSorted;
            }
            else
            {
                for (var j = 0; j !== keys_surfaces.length; j++)
                {
                    var isSingle = !__groupedSurfacesSorted[keys_surfaces[j]].geometry.hasOwnProperty("coincidentgroup");

                    if (!isSingle)
                    {
                        var current_glue_group = __groupedSurfacesSorted[keys_surfaces[j]].geometry["coincidentgroup"];
                        if (j == 0)
                        {
                            var next_glue_group = __groupedSurfacesSorted[keys_surfaces[j + 1]].geometry["coincidentgroup"];
                            if (current_glue_group != next_glue_group)
                            {
                                isSingle = true;
                            }
                        }
                        else if (j == keys_surfaces.length - 1)
                        {
                            var prev_glue_group = __groupedSurfacesSorted[keys_surfaces[j - 1]].geometry["coincidentgroup"];
                            if (current_glue_group != prev_glue_group)
                            {
                                isSingle = true;
                            }
                        }
                        else
                        {
                            var next_glue_group = __groupedSurfacesSorted[keys_surfaces[j + 1]].geometry["coincidentgroup"];
                            var prev_glue_group = __groupedSurfacesSorted[keys_surfaces[j - 1]].geometry["coincidentgroup"];
                            if ((current_glue_group != prev_glue_group) && (current_glue_group != next_glue_group))
                            {
                                isSingle = true;
                            }
                        }
                    }

                    if (isSingle) singles.push(__groupedSurfacesSorted[keys_surfaces[j]]);
                    else sortedNonSingles.push(__groupedSurfacesSorted[keys_surfaces[j]]);
                }
            }

        }
        return {"singles": singles, "sorted": sortedNonSingles};
        // return {"singles":[], "sorted":sortedNonSingles};
        // return {"singles":singles, "sorted":[]};
    };

    //Flag inner/outer surfaces of sorted front and back lists
    const tagInnerSurfaces = function (__groupedSurfacesSorted, __backside)
    {
        var previous_pos_group, previous_neg_group, previous_max_pos_priority, previous_max_neg_priority;
        var current_index;
        var current_surface, current_group, current_priority;
        previous_pos_group = previous_neg_group = "-1";//set it to an unused name

        if (!__backside)
        {
            current_index = __groupedSurfacesSorted.length - 1;
            while (current_index >= 0)
            {
                current_surface = __groupedSurfacesSorted[current_index];
                current_group = current_surface.geometry["coincidentgroup"];
                current_priority = current_surface.geometry["coincidentpriority"];
                var fromARD;
                switch (current_surface.geometry["coincidentgrouptag"])
                {
                    case "planar" :
                        fromARD = true;
                        break;
                    default:
                    case "tris-same":
                    case "shape-same":
                        fromARD = false;
                        break;
                }
                if (current_priority > 0)
                {
                    if (current_group != previous_pos_group)
                    {
                        previous_max_pos_priority = current_priority;
                    }
                    else
                    {
                        if (current_priority < previous_max_pos_priority)
                            current_surface.material.transparencylayer = true && !fromARD;
                    }
                    previous_pos_group = current_group;
                }
                else
                {
                    current_surface.material.transparencylayer = true && !fromARD;
                }
                current_surface.material.fromARD = fromARD;
                current_index--;
            }
        }
        else
        {
            current_index = 0;
            while (current_index < __groupedSurfacesSorted.length)
            {
                current_surface = __groupedSurfacesSorted[current_index];
                current_group = current_surface.geometry["coincidentgroup"];
                current_priority = current_surface.geometry["coincidentpriority"];
                var fromARD;
                switch (current_surface.geometry["coincidentgrouptag"])
                {
                    case "planar" :
                        fromARD = true;
                        break;
                    default:
                    case "tris-same":
                    case "shape-same":
                        fromARD = false;
                        break;
                }
                if (current_priority < 0)
                {
                    if (current_group != previous_neg_group)
                    {
                        previous_max_neg_priority = current_priority;
                    }
                    else
                    {
                        if (current_priority < previous_max_neg_priority)
                            current_surface.material.transparencylayer = true && !fromARD;
                    }
                    previous_neg_group = current_group;
                }
                else
                {
                    current_surface.material.transparencylayer = true && !fromARD;
                }
                current_surface.material.fromARD = fromARD;
                current_index++;
            }
        }
    };

    /**
     * @alias module:WebViewer3D
     * @class ThreeDObjectHandler
     * This class is responsible for geometry parsing and instantiation
     * It creates 3JS cuboids or Meshes depending on what info is in the json.
     * It uses the execution queue to split geometry surface into smaller tasks.
     * It also caches geometry to avoid duplicates in geometry creation (and memory),
     * so that different 3JS scene leaf nodes can safely hold a reference to it.
     * It also handles hardware instancing by bringing the root to leaf transform
     * matrix and use it to feed the instantiation attibutes inst_translation, inst_rotation,
     * and inst_scale used in the instanced versions of the vertex shaders.
     */
    return bb.Model.extend({

        /* PUBLIC */

        initialize: function (__execution_queue)
        {
            LogMessage.Log("Creating ThreeDObjectHandler Object");
            //IMPROVEME: We could consider using singletons for this attribute
            this.set("_exec_queue", __execution_queue);
            this.set("_geometrycounter", 0);
            this.set("_geomsources", {});
            this.set("_proxy_geometries", {});
            this.set("_geom_register_map", {});
        },

        ClearBuildMaps: function ()
        {
            this.set("_geomsources", {});
            this.set("_proxy_geometries", {});
            this.set("_geom_register_map", {});
        },

        /* Add geometry asset to node */
        /* return true if the current asset should be added to the upper scene graph node*/
        InstantiateGeometry: function (__geometry_type, __unique_geometry_id, __mode, __properties, __geometries, __materials, __local_materials_refs, __coplanarity_epsilon, __enable_boundary_hard_edges, __coincident_group_id, __transmat, __forceSWInstPath)
        {
            //GEOMETRY CACHE FETCHING
            var geometry;//Fetch geometry (for Cuboid)
            var material;//Fetch material (for Cuboid)
            var geometry_list;//or Fetch geometry list (for Mesh)
            var material_list;//and Fetch material list (for Mesh)
            var mat_name;
            if (__forceSWInstPath === undefined) __forceSWInstPath = "";

            // WV3DTimers.start("3DObjectHandler Geometry Processing");
            var register_instance = false;
            switch (__geometry_type)
            {
                case "Cuboid":
                    if (__properties.hasOwnProperty("surfaceID") &&
                        __local_materials_refs.hasOwnProperty(__properties["surfaceID"]))
                    {
                        mat_name = __local_materials_refs[__properties["surfaceID"]]["material_name"];
                    }
                    else
                    {
                        mat_name = "default_material";
                    }

                    if (!__geometries.hasOwnProperty(__forceSWInstPath + __unique_geometry_id))
                    {
                        __geometries[__forceSWInstPath + __unique_geometry_id] = {};
                    }

                    if (!this.get("_geom_register_map").hasOwnProperty(__forceSWInstPath + __unique_geometry_id))
                    {
                        this.get("_geom_register_map")[__forceSWInstPath + __unique_geometry_id] = {};
                    }
                    if (!this.get("_geom_register_map")[__forceSWInstPath + __unique_geometry_id].hasOwnProperty(mat_name + "#" + __mode))
                    {
                        this.get("_geom_register_map")[__forceSWInstPath + __unique_geometry_id][mat_name + "#" + __mode] = {};
                        register_instance = true;
                    }

                    //Put data into buffers only for first pass (which is in our case "opaque_pass")
                    if (!__geometries[__forceSWInstPath + __unique_geometry_id].hasOwnProperty(mat_name) && (__mode == "opaque_pass"))
                    {
                        this.set("_geometrycounter", this.get("_geometrycounter") + 1);
                        //Create the first instance
                        geometry = new THREE.InstancedBufferGeometry();
                        geometry.maxInstancedCount = 1;//Should stay "1"

                        geometry.adjacentfacesedgeslines = new THREE.InstancedBufferGeometry();
                        geometry.adjacentfacesedgeslines.maxInstancedCount = 1;

                        var instance_position = new THREE.Vector3();
                        var instance_rotation = new THREE.Quaternion();
                        var instance_scale = new THREE.Vector3(1.0, 1.0, 1.0);
                        __transmat.decompose(instance_position, instance_rotation, instance_scale);

                        var offsets = new THREE.InstancedBufferAttribute(new Float32Array(geometry.maxInstancedCount * 3), 3, 1).setDynamic(true);
                        offsets.setXYZ(0, instance_position.x, instance_position.y, instance_position.z);
                        var orientations = new THREE.InstancedBufferAttribute(new Float32Array(geometry.maxInstancedCount * 4), 4, 1).setDynamic(true);
                        orientations.setXYZW(0, instance_rotation.x, instance_rotation.y, instance_rotation.z, instance_rotation.w);
                        var scales = new THREE.InstancedBufferAttribute(new Float32Array(geometry.maxInstancedCount * 3), 3, 1).setDynamic(true);
                        scales.setXYZ(0, instance_scale.x, instance_scale.y, instance_scale.z);

                        //Mesh buffer initialization and Hard edges buffer initialization
                        if (!this.get("_proxy_geometries").hasOwnProperty(__unique_geometry_id))
                        {
                            this.set("_geometrycounter", this.get("_geometrycounter") + 1);
                            this.get("_proxy_geometries")[__unique_geometry_id] = geometry;
                            CreateCuboidFaces(geometry, __properties)();
                            CreateCuboidHardEdges(geometry.adjacentfacesedgeslines, __properties)();
                        }
                        else
                        {
                            geometry.copy(this.get("_proxy_geometries")[__unique_geometry_id]);
                            geometry.adjacentfacesedgeslines.copy(this.get("_proxy_geometries")[__unique_geometry_id].adjacentfacesedgeslines);
                        }

                        //Instantiation buffers are shared between mesh and highlights
                        geometry.addAttribute("inst_translation", offsets); // per mesh translation
                        geometry.addAttribute("inst_rotation", orientations); // per mesh orientation
                        geometry.addAttribute("inst_scale", scales); // per mesh scale

                        geometry.adjacentfacesedgeslines.addAttribute("inst_translation", offsets); // per mesh translation
                        geometry.adjacentfacesedgeslines.addAttribute("inst_rotation", orientations); // per mesh orientation
                        geometry.adjacentfacesedgeslines.addAttribute("inst_scale", scales); // per mesh scale

                        geometry.attributes["inst_translation"].jarray = [instance_position.x, instance_position.y, instance_position.z];
                        geometry.attributes["inst_rotation"].jarray = [instance_rotation.x, instance_rotation.y, instance_rotation.z, instance_rotation.w];
                        geometry.attributes["inst_scale"].jarray = [instance_scale.x, instance_scale.y, instance_scale.z];

                        //ADD TO MAP
                        __geometries[__forceSWInstPath + __unique_geometry_id][mat_name] = geometry;
                    }
                    else
                    {
                        //Add data into buffers only for first pass (which is in our case "opaque_pass")
                        //otherwise, simply reuse the existing buffers
                        geometry = __geometries[__forceSWInstPath + __unique_geometry_id][mat_name];
                        if (__mode == "opaque_pass")
                        {
                            var instance_position = new THREE.Vector3();
                            var instance_rotation = new THREE.Quaternion();
                            var instance_scale = new THREE.Vector3(1.0, 1.0, 1.0);
                            __transmat.decompose(instance_position, instance_rotation, instance_scale);

                            geometry.attributes["inst_translation"].jarray.push(instance_position.x, instance_position.y, instance_position.z);
                            geometry.attributes["inst_rotation"].jarray.push(instance_rotation.x, instance_rotation.y, instance_rotation.z, instance_rotation.w);
                            geometry.attributes["inst_scale"].jarray.push(instance_scale.x, instance_scale.y, instance_scale.z);

                        }
                    }
                    break;

                case "Mesh":

                    var sources = null;
                    // var total_geom_size = 0;
                    for (var idx_surface = 0, surface_keys = Object.keys(__properties["triangles"]), nb_surfaces = surface_keys.length;
                         idx_surface < nb_surfaces;
                         ++idx_surface)
                    {
                        //Determine the current material name
                        var s = surface_keys[idx_surface];
                        var surface = __properties["triangles"][s];
                        if (surface.count > 0)
                        {
                            if (__local_materials_refs.hasOwnProperty(s))
                            {
                                mat_name = __local_materials_refs[s]["material_name"];
                            }
                            else
                            {
                                mat_name = "default_material";
                            }

                            //If the mesh does not exists yet, initialize arrays
                            if (!__geometries.hasOwnProperty(__forceSWInstPath + __unique_geometry_id))
                            {
                                __geometries[__forceSWInstPath + __unique_geometry_id] = geometry_list = [];
                            }
                            else
                            {
                                geometry_list = __geometries[__forceSWInstPath + __unique_geometry_id];
                            }
                            //If the source buffers haven't been initialised
                            if (!this.get("_geomsources").hasOwnProperty(__unique_geometry_id))
                            {
                                sources = {};

                                if (__properties["sources"])
                                {
                                    for (var i = 0, keys = Object.keys(__properties["sources"]), l = keys.length; i < l; ++i)
                                    {
                                        var source_buffer = keys[i];
                                        if (typeof source_buffer == "string")
                                        {
                                            var buffer_values = [];
                                            var stride = __properties["sources"][source_buffer]["stride"];
                                            var format = __properties["sources"][source_buffer]["format"];
                                            var count = __properties["sources"][source_buffer]["count"];
                                            var values = __properties["sources"][source_buffer]["values"];

                                            switch (format)
                                            {
                                                case "ST":
                                                case "UV":
                                                    for (var v = 0; v < count; ++v)
                                                    {
                                                        buffer_values.push.apply(buffer_values, [parseFloat(values[v * stride]), parseFloat(values[v * stride + 1])]);
                                                    }
                                                    break;
                                                case "XYZ":
                                                case "RGB":
                                                    for (var v = 0; v < count; ++v)
                                                    {
                                                        buffer_values.push.apply(buffer_values, [parseFloat(values[v * stride]), parseFloat(values[v * stride + 1]), parseFloat(values[v * stride + 2])]);
                                                    }
                                                    break;
                                                case "XYZW":
                                                    for (var v = 0; v < count; ++v)
                                                    {
                                                        buffer_values.push.apply(buffer_values, [parseFloat(values[v * stride]), parseFloat(values[v * stride + 1]), parseFloat(values[v * stride + 2]), parseFloat(values[v * stride + 3])]);
                                                    }
                                                    break;
                                                default:
                                                    LogMessage.Log("ThreeDObjectHandler Warning: Unhandled format string: default single buffer selected");
                                                case "C":
                                                    buffer_values = values;
                                                    break;
                                            }
                                            sources[source_buffer] = buffer_values;
                                        }
                                    }
                                }
                                this.get("_geomsources")[__unique_geometry_id] = sources;
                            }
                            else
                            {
                                sources = this.get("_geomsources")[__unique_geometry_id];
                            }

                            //Create new buffered geometry if the current geometryID/surfaceID/materialID
                            //combination does not exist ("if" branch), otherwise we just increase the
                            //number of instances ("else" branch).
                            if (!geometry_list.hasOwnProperty(s))
                            {
                                geometry_list[s] = {};
                            }

                            if (!this.get("_geom_register_map").hasOwnProperty(__forceSWInstPath + __unique_geometry_id))
                            {
                                this.get("_geom_register_map")[__forceSWInstPath + __unique_geometry_id] = {};
                            }
                            if (!this.get("_geom_register_map")[__forceSWInstPath + __unique_geometry_id].hasOwnProperty(mat_name + "#" + __mode) || surface.hasOwnProperty("coincident-group"))
                            {
                                this.get("_geom_register_map")[__forceSWInstPath + __unique_geometry_id][mat_name + "#" + __mode] = {};
                                register_instance = true;
                            }

                            if (!geometry_list[s].hasOwnProperty(mat_name) || surface.hasOwnProperty("coincident-group"))
                            {
                                if (__mode == "opaque_pass")//i.e. first pass
                                {
                                    var use_software_instancing = false;
                                    if (surface.hasOwnProperty("coincident-group"))
                                    {
                                        //Each coincident surface has to be treated as a separate surface (non hardware instantiated)
                                        //in order to allow coincident surfaces to be drawn in a specific order.
                                        //If we find the same surface in the geometry list which has only one instance
                                        //then, we can use it as a software instance. If we do so, the matrix given to
                                        //the shader should be identity and we have to return a flag indicating that
                                        //we want to treat the mesh within a regular scene graph structure with
                                        //transformations applied to each of these nodes.
                                        if (geometry_list[s])
                                        {
                                            for (var j = 0, keys = Object.keys(geometry_list[s]), l = keys.length; j < l; ++j)
                                            {
                                                var geom = keys[j];
                                                if (geom != "default_cg" && geometry_list[s][geom].maxInstancedCount == 1)
                                                {
                                                    use_software_instancing = true;
                                                    geometry = geometry_list[s][geom];
                                                    break;
                                                }
                                            }
                                        }
                                    }
                                    if (!use_software_instancing)
                                    {
                                        //Create the first instance
                                        geometry = new THREE.InstancedBufferGeometry();
                                        geometry.maxInstancedCount = 1;//Should stay "1"
                                        geometry.adjacentfacesedgeslines = new THREE.InstancedBufferGeometry();
                                        geometry.adjacentfacesedgeslines.maxInstancedCount = 1;

                                        //SURFACE ORDERING INFORMATION
                                        var geom_coincidentmeshID = __unique_geometry_id;
                                        var geom_coincidentgroup = surface.hasOwnProperty("coincident-group") ? parseFloat(surface["coincident-group"]) : 0;
                                        var geom_coincidentpriority = surface.hasOwnProperty("coincident-priority") ? parseFloat(surface["coincident-priority"]) : 0;
                                        var geom_coincidentgrouptag = surface.hasOwnProperty("coincident-group-tag") ? surface["coincident-group-tag"] : "tris-same";
                                        geometry["useSoftwareInstancing"] = surface.hasOwnProperty("coincident-group") ? true : use_software_instancing;
                                        geometry.adjacentfacesedgeslines["useSoftwareInstancing"] = surface.hasOwnProperty("coincident-group") ? true : use_software_instancing;
                                        geometry["coincidentmeshID"] = geom_coincidentmeshID;
                                        geometry["coincidentgroup"] = geom_coincidentgroup;
                                        geometry["coincidentpriority"] = geom_coincidentpriority;
                                        geometry["coincidentgrouptag"] = geom_coincidentgrouptag;

                                        //Instance transformation
                                        var offsets = new THREE.InstancedBufferAttribute(new Float32Array(geometry.maxInstancedCount * 3), 3, 1).setDynamic(true);
                                        var orientations = new THREE.InstancedBufferAttribute(new Float32Array(geometry.maxInstancedCount * 4), 4, 1).setDynamic(true);
                                        var scales = new THREE.InstancedBufferAttribute(new Float32Array(geometry.maxInstancedCount * 3), 3, 1).setDynamic(true);
                                        var instance_position = new THREE.Vector3();
                                        var instance_rotation = new THREE.Quaternion();
                                        var instance_scale = new THREE.Vector3(1.0, 1.0, 1.0);
                                        if (!(surface.hasOwnProperty("coincident-group")))
                                        {
                                            __transmat.decompose(instance_position, instance_rotation, instance_scale);
                                        }
                                        else
                                        {
                                            //Initialize with identity, transform will be used one level up
                                            use_software_instancing = true;
                                        }
                                        offsets.setXYZ(0, instance_position.x, instance_position.y, instance_position.z);
                                        orientations.setXYZW(0, instance_rotation.x, instance_rotation.y, instance_rotation.z, instance_rotation.w);
                                        scales.setXYZ(0, instance_scale.x, instance_scale.y, instance_scale.z);

                                        //Initialize instance attributes (shared between mesh and hard edges)
                                        geometry.addAttribute("inst_translation", offsets); // per mesh translation
                                        geometry.addAttribute("inst_rotation", orientations); // per mesh orientation
                                        geometry.addAttribute("inst_scale", scales); // per mesh scale

                                        geometry.adjacentfacesedgeslines.addAttribute("inst_translation", offsets); // per mesh translation
                                        geometry.adjacentfacesedgeslines.addAttribute("inst_rotation", orientations); // per mesh orientation
                                        geometry.adjacentfacesedgeslines.addAttribute("inst_scale", scales); // per mesh scale

                                        geometry.attributes["inst_translation"].jarray = [instance_position.x, instance_position.y, instance_position.z];
                                        geometry.attributes["inst_rotation"].jarray = [instance_rotation.x, instance_rotation.y, instance_rotation.z, instance_rotation.w];
                                        geometry.attributes["inst_scale"].jarray = [instance_scale.x, instance_scale.y, instance_scale.z];

                                        //Mesh buffer initialization and Hard edges buffer initialization
                                        var proxy_geometry;
                                        if (!this.get("_proxy_geometries").hasOwnProperty(__unique_geometry_id))
                                        {
                                            this.get("_proxy_geometries")[__unique_geometry_id] = {};
                                        }
                                        if (!this.get("_proxy_geometries")[__unique_geometry_id].hasOwnProperty(s))
                                        {
                                            this.set("_geometrycounter", this.get("_geometrycounter") + 1);
                                            //Create proxy "non instanced" object that will contain buffers (these will be referenced by instanced objects)
                                            this.get("_proxy_geometries")[__unique_geometry_id][s] = proxy_geometry = new THREE.BufferGeometry();
                                            proxy_geometry.adjacentfacesedgeslines = new THREE.BufferGeometry();
                                            //Defer costly operation (faces construction)
                                            this.get("_exec_queue").run(CreateMeshFaces(proxy_geometry, sources, surface), this, false, null, null, true);
                                            //Defer costly operation (hard edges construction)
                                            this.get("_exec_queue").run(CreateMeshHardEdges(proxy_geometry.adjacentfacesedgeslines, sources, surface, __coplanarity_epsilon, __enable_boundary_hard_edges), this, false, null, null, true);
                                        }
                                        else
                                        {
                                            proxy_geometry = this.get("_proxy_geometries")[__unique_geometry_id][s];
                                        }

                                        //Defer attributes initialization (since buffers initialization defered too, see previous lines)
                                        this.get("_exec_queue").run(AttributesInit(geometry, proxy_geometry), this, false, null, null, true);

                                        if (!geometry_list[s].hasOwnProperty(mat_name))
                                        {
                                            geometry_list[s][mat_name] = {};
                                        }
                                        if (!(surface.hasOwnProperty("coincident-group")))
                                        {
                                            geometry_list[s][mat_name]["default_cg"] = geometry;
                                        }
                                        else
                                        {
                                            geometry_list[s][mat_name][geom_coincidentgroup] = geometry;
                                        }
                                    }
                                }
                            }
                            else
                            {
                                geometry = geometry_list[s][mat_name]["default_cg"];
                                if (geometry instanceof THREE.InstancedBufferGeometry && (__mode == "opaque_pass") && (__coincident_group_id === undefined))
                                {
                                    //In the case we have a new instance of a non coincident surface using
                                    //the same material as a previous instance we use hardware instancing

                                    var instance_position = new THREE.Vector3();
                                    var instance_rotation = new THREE.Quaternion();
                                    var instance_scale = new THREE.Vector3(1.0, 1.0, 1.0);
                                    __transmat.decompose(instance_position, instance_rotation, instance_scale);

                                    geometry.attributes["inst_translation"].jarray.push(instance_position.x, instance_position.y, instance_position.z);
                                    geometry.attributes["inst_rotation"].jarray.push(instance_rotation.x, instance_rotation.y, instance_rotation.z, instance_rotation.w);
                                    geometry.attributes["inst_scale"].jarray.push(instance_scale.x, instance_scale.y, instance_scale.z);

                                }
                            }
                        }
                    }
                    // var total_geom_size = total_geom_size + WV3DUtils.SizeOf(geometry)/1000000;
                    // LogMessage.Log("geometry_list size: "+ total_geom_size + "MB");
                    // Next line is for Chrome only, with option --enable-precise-memory-info
                    // LogMessage.Log("Memory usage: "+performance.memory.usedJSHeapSize+"/"+performance.memory.totalJSHeapSize+"/"+performance.memory.jsHeapSizeLimit);

                    break;

                default:
                    LogMessage.Log("ThreeDObjectHandler Warning: Unrecognized type of geometry - Accepted: Cuboid, Mesh");
                    break;
            }
            // WV3DTimers.stop("3DObjectHandler Geometry Processing");


            // WV3DTimers.start("THREEJS Geometry Creation");

            var objects = [];
            if (register_instance)
            {
                switch (__mode)
                {
                    case "highlightnode":
                        switch (__geometry_type)
                        {
                            case "Cuboid":
                                //WE CREATE HIGHLIGHTS FROM THE BOX EDGES (see above adjacentfacesedgeslines)
                                objects.push(new THREE.LineSegments(geometry.adjacentfacesedgeslines, __materials["highlight"]));
                                break;
                            case "Mesh":
                                //WE CREATE HIGHLIGHTS FROM THE EDGES BETWEEN TWO NON COPLANAR FACES
                                var mat_name;
                                for (var i = 0, keys = Object.keys(geometry_list), l = keys.length; i < l; ++i)
                                {
                                    var s = keys[i];
                                    var geometry = geometry_list[s];
                                    if (__local_materials_refs.hasOwnProperty(s))
                                    {
                                        mat_name = __local_materials_refs[s]["material_name"];
                                    }
                                    else
                                    {
                                        mat_name = "default_material";
                                    }
                                    if (geometry[mat_name])
                                    {
                                        for (var j = 0, keys2 = Object.keys(geometry[mat_name]), l2 = keys2.length; j < l2; ++j)
                                        {
                                            var cgobjects = keys2[j];
                                            objects.push(new THREE.LineSegments(geometry[mat_name][cgobjects].adjacentfacesedgeslines, __materials["highlight"]));
                                        }
                                    }
                                }
                                break;
                            default:
                                LogMessage.Log("ThreeDObjectHandler Warning: Unrecognized type of geometry - Accepted: Cuboid, Mesh");
                                break;
                        }
                        break;

                    default :
                        switch (__geometry_type)
                        {
                            case "Cuboid":
                                //CREATE A MESH OBJECT
                                if (__properties.hasOwnProperty("surfaceID") && __local_materials_refs.hasOwnProperty(__properties["surfaceID"]))
                                {
                                    var s = __properties["surfaceID"];
                                    var mat_struct = __local_materials_refs[s]["front"];//always only one "front" element for cuboids material
                                    var transparent_material = mat_struct["opacity"] < 1 || mat_struct["transp_texture"];
                                    this.trigger("non_empty_scene", {"transparent": transparent_material});
                                    //If a shadermaterial exists for this pass
                                    if (mat_struct["mat"]["untextured_mat"][__mode])
                                    {
                                        if (!mat_struct["boundmeshlist"])
                                        {
                                            mat_struct["boundmeshlist"] = {};
                                        }
                                        if (!mat_struct["boundmeshlist"].hasOwnProperty(__mode))
                                        {
                                            mat_struct["boundmeshlist"][__mode] = {};
                                        }
                                        if (!mat_struct["boundmeshlist"][__mode].hasOwnProperty("front_mat"))
                                        {
                                            mat_struct["boundmeshlist"][__mode]["front_mat"] = {};
                                        }
                                        var placeholder_material = mat_struct["mat"]["untextured_mat"][__mode];

                                        for (var spass_cpt = 0, spass_keys = Object.keys(placeholder_material), l_spass = spass_keys.length;
                                             spass_cpt < l_spass; spass_cpt++)
                                        {
                                            var spass_id = spass_keys[spass_cpt];
                                            var spass = placeholder_material[spass_keys[spass_id]];

                                            if (!mat_struct["mat"]["textured_mat"].hasOwnProperty(__mode))
                                            {
                                                mat_struct["mat"]["textured_mat"][__mode] = {};
                                            }
                                            if (!mat_struct["mat"]["textured_mat"][__mode].hasOwnProperty(spass_id))
                                            {
                                                mat_struct["mat"]["textured_mat"][__mode][spass_id] = {};
                                            }
                                            if (!mat_struct["mat"]["textured_mat"][__mode][spass_id].hasOwnProperty(s))
                                            {
                                                mat_struct["mat"]["textured_mat"][__mode][spass_id][s] = spass;
                                            }

                                            if (!mat_struct["boundmeshlist"][__mode]["front_mat"].hasOwnProperty(spass_id))
                                            {
                                                mat_struct["boundmeshlist"][__mode]["front_mat"][spass_id] = {};
                                            }
                                            if (!mat_struct["boundmeshlist"][__mode]["front_mat"][spass_id].hasOwnProperty(s))
                                            {
                                                mat_struct["boundmeshlist"][__mode]["front_mat"][spass_id][s] = [];
                                            }
                                            object = new THREE.Mesh(geometry, spass);
                                            objects.push(object);
                                            mat_struct["boundmeshlist"][__mode]["front_mat"][spass_id][s].push(object);
                                            // {
                                            // console.log("surface: "+s+" / mode: "+__mode+" / side"+side_mat+" / spass_id: "+spass_id+" / ordered_render_list size: "+Object.keys(ordered_render_list).length);
                                            // console.log("coinc-group: "+boundgeom.coincidentmeshID+"/"+boundgeom.coincidentgroup+"/"+boundgeom.coincidentpriority+"/"+boundgeom.coincidentgrouptag);
                                            // }
                                        }
                                    }
                                }
                                else
                                {
                                    object = new THREE.Mesh(geometry, __materials["default_material"]);
                                    objects.push(object);
                                    this.trigger("non_empty_scene", {"transparent": false});
                                }
                                break;

                            case "Mesh":
                                //CREATE MESH OBJECTS
                                //Here, all coincident surfaces of this mesh are present in geometry_list
                                var ordered_render_list = this.OrderRenderList(__mode, geometry_list, __coincident_group_id, __local_materials_refs);

                                for (var i = 0, keys = Object.keys(ordered_render_list), l = keys.length; i < l; ++i)
                                {
                                    var object;
                                    var s = ordered_render_list[i].surfaceID;
                                    var boundgeom = ordered_render_list[i].geometry;
                                    var material_param = ordered_render_list[i].material;
                                    if (__local_materials_refs.hasOwnProperty(s))
                                    {
                                        //get the untextured material and set it for the current mesh
                                        var mat_struct, material;
                                        var side_mat = (material_param["side"] == THREE.BackSide) ? "back_mat" : "front_mat";//allows side_mat to be THREE.DoubleSide

                                        if (material_param["side"] == THREE.BackSide)
                                        {
                                            mat_struct = __local_materials_refs[s]["back"];
                                        }
                                        else
                                        {
                                            mat_struct = __local_materials_refs[s]["front"];
                                        }

                                        if (mat_struct)
                                        {
                                            //front and back should be in sync for tranparency/opacity
                                            var transparent_material = mat_struct["opacity"] < 1 || mat_struct["transp_texture"];
                                            if (!material_param.transparencylayer || transparent_material)
                                            {
                                                //If a shadermaterial exists for this pass
                                                if (mat_struct["mat"]["untextured_mat"][__mode])
                                                {

                                                    var placeholder_material = mat_struct["mat"]["untextured_mat"][__mode];
                                                    for (var spass_cpt = 0, spass_keys = Object.keys(placeholder_material), l_spass = spass_keys.length;
                                                         spass_cpt < l_spass; spass_cpt++)
                                                    {
                                                        var spass_id = spass_keys[spass_cpt];
                                                        var spass = placeholder_material[spass_id];

                                                        //Create a placeholder copy that has its own side parameters
                                                        //this copy is waiting for asynchronuous updates (updatetexturedmaterial in materialmanager.js)
                                                        var mconstructor = spass.constructor;
                                                        material = new mconstructor({"name": "placeholder_material"});
                                                        material.copy(spass);
                                                        //The following code is made for a limitation of AMD Radeon Graphics cards that
                                                        //do not retessellate/rasterize triangles touching the screen border the same way if being
                                                        //rendered using Front VS Back face culling, causing z-fighting when a bag seal
                                                        //gets close to the openGL canvas border. See STU-5906 for more details.
                                                        //To fix it, we change the winding order (this first part is made in
                                                        //ObjectBuilder::InvertGeometryWinding ) and forces the material to cull
                                                        //back faces instead of front faces (see below)
                                                        //Now that the geometry knows its material (see above "front_mat" or "back_mat") using material_param["side"]
                                                        //we set the material to FrontSide (enabling back-face culling)
                                                        if (material_param["side"] === THREE.BackSide && !boundgeom["HasInvertedWinding"])
                                                            console.error("Back geometry not inverted !!!");
                                                        material["side"] = THREE.FrontSide;

                                                        if (!mat_struct["mat"]["textured_mat"].hasOwnProperty(__mode))
                                                        {
                                                            mat_struct["mat"]["textured_mat"][__mode] = {};
                                                        }
                                                        if (!mat_struct["mat"]["textured_mat"][__mode].hasOwnProperty(spass_id))
                                                        {
                                                            mat_struct["mat"]["textured_mat"][__mode][spass_id] = {};
                                                        }
                                                        if (!mat_struct["mat"]["textured_mat"][__mode][spass_id].hasOwnProperty(s))
                                                        {
                                                            mat_struct["mat"]["textured_mat"][__mode][spass_id][s] = material;
                                                        }

                                                        if (!mat_struct["boundmeshlist"])
                                                        {
                                                            mat_struct["boundmeshlist"] = {};
                                                        }
                                                        if (!mat_struct["boundmeshlist"].hasOwnProperty(__mode))
                                                        {
                                                            mat_struct["boundmeshlist"][__mode] = {};
                                                        }
                                                        if (!mat_struct["boundmeshlist"][__mode].hasOwnProperty(side_mat))
                                                        {
                                                            mat_struct["boundmeshlist"][__mode][side_mat] = {};
                                                        }
                                                        if (!mat_struct["boundmeshlist"][__mode][side_mat].hasOwnProperty(spass_id))
                                                        {
                                                            mat_struct["boundmeshlist"][__mode][side_mat][spass_id] = {};
                                                        }
                                                        if (!mat_struct["boundmeshlist"][__mode][side_mat][spass_id].hasOwnProperty(s))
                                                        {
                                                            mat_struct["boundmeshlist"][__mode][side_mat][spass_id][s] = [];
                                                        }

                                                        object = new THREE.Mesh(boundgeom, material);
                                                        objects.push(object);
                                                        mat_struct["boundmeshlist"][__mode][side_mat][spass_id][s].push(object);
                                                        // {
                                                        // console.log("surface: "+s+" / mode: "+__mode+" / side"+side_mat+" / spass_id: "+spass_id+" / ordered_render_list size: "+Object.keys(ordered_render_list).length);
                                                        // console.log("coinc-group: "+boundgeom.coincidentmeshID+"/"+boundgeom.coincidentgroup+"/"+boundgeom.coincidentpriority+"/"+boundgeom.coincidentgrouptag);
                                                        // }
                                                    }
                                                }
                                            }
                                        }
                                        else
                                        {
                                            LogMessage.Log("ThreeDObjectHandler Warning: No front material found for this surface");
                                        }
                                    }
                                    else
                                    {
                                        object = new THREE.Mesh(boundgeom, __materials["default_material"]);
                                        objects.push(object);
                                        this.trigger("non_empty_scene", {"transparent": false});
                                    }
                                }
                                break;

                            default:
                                LogMessage.Log("ThreeDObjectHandler Warning: Unrecognized type of geometry - Only accepted: Cuboid, Mesh");
                                break;
                        }
                        break;
                }
            }
            // WV3DTimers.stop("THREEJS Geometry Creation");
            return {"register": register_instance, "objects": objects};
        },

        // OrderRenderList() is a very special purpose routine
        // currently used by Studio (Illustrator plugin) and by Studio Toolkit for Flexibles when rendering bag and Collada files,
        // so it is not relevant for folding carton (.ard).
        //
        // NOTE: this method has been adapted in 3D4.0 to be able to handle the glue groups from Collada, where double sided surfaces are possible.
        //
        // The main problem it deals with is avoiding Z-fighting between intentionally glued surfaces in .bag / Collada files.
        //
        // SetSurfaceVisibilityBasedOnGlueingAndPriorities() consults the Solid surfaces
        // and sets their front and back visibility flags based on their GlueGroup and Priority values,
        // such that the Solid will be rendered correctly by the current Solid::DrawWithOpenGL() implementation.
        //
        // Currently the following combinations of surface glueGroupID and priority are known to occur in .bag files:
        // - GlueGroupID==-1 and Priority==0:
        //   For pre-release Studio Toolkit for Flexibles bag files that did not yet have glueGroup Info.
        //   These bags also only had max 2 surfaces glued together (always back-to-back)
        //
        // - GlueGroupID>=0 and Priority==0:
        //   For pre-release Studio Toolkit for Flexibles bag files that grouped intentionally glued surfaces
        //   into glueGroups to assist Visualizer with dealing with the resulting Z-fighting problem.
        //   These bags still only had max 2 surfaces glued together (always back-to-back)
        //
        // - GlueGroupID>=0 and Priority!=0:
        //   For Release versions of Studio Toolkit for Flexibles bag files.
        //   These bag files can have more than 2 surfaces glued together. The priority value determines the order in which
        //   the surfaces are glued together, and also whether the glueing is back-to-back or back-to-front.
        //   The priority assignment as done in Studio Toolkit for Flexibles is currently as follows:
        //   For a stack of surfaces that are glued together, choose one of the two possible sides to look at the stack.
        //   Then, while looking at the stack in that way, start with a positive (>0) priority value for the backmost surface.
        //   Working from back to front, assign each surface a higher priority value than the one before.
        //   When all surfaces in this glueGroup are assigned a priority this way, go through the stack again in the same order,
        //   and change the sign of the priority value of the surface if it faces backwards (still under that same choice
        //   of looking at the surface stack), and leave the sign unmodified otherwise.
        //   Priority values assigned this way describe the way the surfaces in a glueGroup are connected to one another,
        //   and can be used to render them without Z-fighting.
        //
        //   SetSurfaceVisibilityBasedOnGlueingAndPriorities() will basically hide all surfaces (i.e. set their front & back
        //   visibility flags to "invisible") except for the "outer" surfaces of a stack of glued surfaces. This works
        //   because we're only interested in correctly rendering *opaque* surfaces. If we want to render transparent
        //   surfaces the trick with SetSurfaceVisibilityBasedOnGlueingAndPriorities() does not suffice and we have
        //   to use a more powerful rendering algorithm in Solid::DrawWithOpenGL() instead.

        OrderRenderList: function (__mode, __renderList, __coincident_group_id, __local_materials_refs)
        {
            //sort the surfaces before rendering if possible
            //the renderlist will be iterated from its end towards its beginning
            var orderedRenderList = [];
            var quicksortRenderList = [];
            var mat_name;

            for (var keys = Object.keys(__renderList), i = keys.length - 1; i !== -1; i--)
            {
                var geometry = __renderList[keys[i]];

                //Determine the current material name
                var s = keys[i];
                if (__local_materials_refs.hasOwnProperty(s))
                {
                    mat_name = __local_materials_refs[s]["material_name"];
                }
                else
                {
                    mat_name = "default_material";
                }

                if (geometry[mat_name])
                {
                    for (var j = 0, keys2 = Object.keys(geometry[mat_name]), l = keys2.length; j < l; ++j)
                    {
                        var cgobjects = keys2[j];
                        var object =
                            {
                                "surfaceID": keys[i],
                                "geometry": geometry[mat_name][cgobjects]
                            };
                        quicksortRenderList.push(object);
                    }
                }

                var transparent = false;
                if (__local_materials_refs.hasOwnProperty(keys[i]))
                {
                    var mat_struct_front = __local_materials_refs[keys[i]]["front"];
                    var mat_struct_back = __local_materials_refs[keys[i]]["back"];
                    var transparent_front_mat = mat_struct_front["opacity"] < 1 || mat_struct_front["transp_texture"];
                    var transparent_back_mat = mat_struct_back["opacity"] < 1 || mat_struct_back["transp_texture"];
                    transparent = transparent_front_mat || transparent_back_mat;
                }
                this.trigger("non_empty_scene", {"transparent": transparent});
            }

            //Sort surfaces to determine singles and non singles
            quicksortRenderList.sort(CompareSurfacesInGlueGroup(true));
            var split_render_list = splitSingleSurfaces(quicksortRenderList);
            //Add singles to render list if not coincident group pass
            if (__coincident_group_id === undefined)
            {
                var singlesFrontAndBack = this.DuplicateSurfaceForFrontAndBack(split_render_list.singles, __local_materials_refs);
                Array.prototype.push.apply(orderedRenderList, singlesFrontAndBack.ungrouped);
                Array.prototype.push.apply(orderedRenderList, singlesFrontAndBack.front);
                Array.prototype.push.apply(orderedRenderList, singlesFrontAndBack.back);
            }

            //Filter
            var filtered_render_list = this.FilterRenderList(split_render_list.sorted, __coincident_group_id, __local_materials_refs);
            var frontandback = this.DuplicateSurfaceForFrontAndBack(filtered_render_list, __local_materials_refs);
            frontandback.front.sort(CompareSurfacesInGlueGroup(true));
            frontandback.back.sort(CompareSurfacesInGlueGroup(false));
            if (__coincident_group_id === undefined)
                Array.prototype.push.apply(orderedRenderList, frontandback.ungrouped);

            //Tag Surfaces for transparency flag and ard flag
            tagInnerSurfaces(frontandback.front, false);

            //uncomment this line if priorities have to be inverted for ard coincident groups
            // if(__coincident_group_id == undefined) frontandback.front.reverse();

            //Reverse the array for processing
            frontandback.back.reverse();
            tagInnerSurfaces(frontandback.back, true);
            frontandback.back.reverse();

            //uncomment this line if priorities have to be inverted for ard coincident groups
            // if(__coincident_group_id == undefined) frontandback.back.reverse();

            // if(__coincident_group_id === undefined)
            Array.prototype.push.apply(orderedRenderList, frontandback.back);
            // if(__coincident_group_id === undefined)
            Array.prototype.push.apply(orderedRenderList, frontandback.front);

            //uncomment this line if priorities have to be inverted for ard coincident groups
            // if(__coincident_group_id !== undefined) orderedRenderList.reverse();

            return orderedRenderList;
        },

        InvertGeometryWinding: function (__invertedGeometry, __geometry)
        {
            //The following code helps to circumvent a limitation of AMD Radeon Graphics cards that
            //do not retessellate triangles touching the screen border the same way if being
            //rendered using Front VS Back face culling, causing z-fighting when a bag seal
            //gets close to the openGL canvas border. See STU-5906 for more details.
            //To fix it, we change the winding order (below) and force the material to cull back faces
            //instead of front faces (this second part is made in ObjectBuilder::InstantiateGeometry )
            if (!__invertedGeometry.hasOwnProperty("HasInvertedWinding"))
            {
                __invertedGeometry["HasInvertedWinding"] = true;
                //Inversion is deferred until buffers are filled in.
                this.get("_exec_queue").run(function ()
                {

                    if (__invertedGeometry.attributes.hasOwnProperty("index"))
                    {
                        //Make a clone before changing winding order
                        var indexes = __invertedGeometry.attributes["index"].clone();
                        for (var idx = 0; idx < indexes.count; idx += 3)
                        {
                            //swap two last indexes of every triangle
                            var _i1 = indexes[idx + 1];
                            var _i2 = indexes[idx + 2];

                            indexes[idx + 1] = _i2;
                            indexes[idx + 2] = _i1;
                        }
                        __invertedGeometry.attributes["index"] = indexes;
                    }
                    else
                    {
                        //Make a clone before changing winding order
                        var positions = __invertedGeometry.attributes["position"].clone();
                        var normals = __invertedGeometry.attributes["normal"].clone();
                        var uvs = __invertedGeometry.attributes["uv"].clone();
                        var uvs2 = __invertedGeometry.attributes["uv2"].clone();
                        var colors = __invertedGeometry.attributes.hasOwnProperty("color") ? __invertedGeometry.attributes["color"].clone() : null;
                        for (var idx = 0; idx < positions.count; idx += 3)
                        {
                            //swap XYZ of two last vertices of every triangle
                            var _X1 = positions.getX(idx + 1);
                            var _Y1 = positions.getY(idx + 1);
                            var _Z1 = positions.getZ(idx + 1);
                            var _X2 = positions.getX(idx + 2);
                            var _Y2 = positions.getY(idx + 2);
                            var _Z2 = positions.getZ(idx + 2);

                            positions.setXYZ(idx + 1, _X2, _Y2, _Z2);
                            positions.setXYZ(idx + 2, _X1, _Y1, _Z1);

                            var _X0 = normals.getX(idx);
                            var _Y0 = normals.getY(idx);
                            var _Z0 = normals.getZ(idx);
                            _X1 = normals.getX(idx + 1);
                            _Y1 = normals.getY(idx + 1);
                            _Z1 = normals.getZ(idx + 1);
                            _X2 = normals.getX(idx + 2);
                            _Y2 = normals.getY(idx + 2);
                            _Z2 = normals.getZ(idx + 2);

                            normals.setXYZ(idx + 0, -_X0, -_Y0, -_Z0);
                            normals.setXYZ(idx + 1, -_X2, -_Y2, -_Z2);
                            normals.setXYZ(idx + 2, -_X1, -_Y1, -_Z1);

                            _X1 = uvs.getX(idx + 1);
                            _Y1 = uvs.getY(idx + 1);
                            _X2 = uvs.getX(idx + 2);
                            _Y2 = uvs.getY(idx + 2);

                            uvs.setXY(idx + 1, _X2, _Y2);
                            uvs.setXY(idx + 2, _X1, _Y1);

                            _X1 = uvs2.getX(idx + 1);
                            _Y1 = uvs2.getY(idx + 1);
                            _X2 = uvs2.getX(idx + 2);
                            _Y2 = uvs2.getY(idx + 2);

                            uvs2.setXY(idx + 1, _X2, _Y2);
                            uvs2.setXY(idx + 2, _X1, _Y1);

                            if (colors)
                            {
                                var _R1 = colors.getX(idx + 1);
                                var _G1 = colors.getY(idx + 1);
                                var _B1 = colors.getZ(idx + 1);
                                var _R2 = colors.getX(idx + 2);
                                var _G2 = colors.getY(idx + 2);
                                var _B2 = colors.getZ(idx + 2);

                                colors.setXYZ(idx + 1, _R2, _G2, _B2);
                                colors.setXYZ(idx + 2, _R1, _G1, _B1);
                            }
                        }
                        positions.needsUpdate = true;
                        normals.needsUpdate = true;
                        uvs.needsUpdate = true;
                        uvs2.needsUpdate = true;
                        //Overwrite original object attributes since the previous ones
                        //are still shared with the non inverted winding geometry.
                        //Instantiation buffers stay the same and can be directly
                        //referenced though.
                        __invertedGeometry.attributes = {
                            "position": positions,
                            "normal": normals,
                            "uv": uvs,
                            "uv2": uvs2,
                            "inst_translation": __invertedGeometry.attributes.inst_translation,
                            "inst_rotation": __invertedGeometry.attributes.inst_rotation,
                            "inst_scale": __invertedGeometry.attributes.inst_scale
                        };
                        if (colors)
                        {
                            __invertedGeometry.attributes["color"] = colors;
                        }

                        __invertedGeometry.verticesNeedUpdate = true;
                        __invertedGeometry.normalsNeedUpdate = true;
                        __invertedGeometry.uvsNeedUpdate = true;
                    }

                    //Same dUdP and dVdP are applied to the inverted winding geometry
                    if (__geometry.hasOwnProperty("dUdP") && __geometry.hasOwnProperty("dVdP"))
                    {
                        __invertedGeometry.dUdP = __geometry.dUdP;
                        __invertedGeometry.dVdP = __geometry.dVdP;
                    }
                }, this, false, null, null, true);
            }
        },

        GetInvertedGeometry: function (__geometry)
        {
            var reversewindinggeometry;
            if (!__geometry.hasOwnProperty("InvertedWindingGeometry"))
            {
                var geomconstructor = __geometry.constructor;
                reversewindinggeometry = new geomconstructor();
                //Copy geometry properties into reversewindinggeometry
                for (var geom_keys = Object.keys(__geometry), geom_key_idx = geom_keys.length - 1; geom_key_idx >= 0; geom_key_idx--)
                {
                    reversewindinggeometry[geom_keys[geom_key_idx]] = __geometry[geom_keys[geom_key_idx]];
                }
                this.InvertGeometryWinding(reversewindinggeometry, __geometry);

                __geometry["InvertedWindingGeometry"] = reversewindinggeometry;
            }
            else
            {
                reversewindinggeometry = __geometry["InvertedWindingGeometry"];
            }
            return reversewindinggeometry;
        },

        DuplicateSurfaceForFrontAndBack: function (__renderList, __local_materials_refs)
        {
            //Arrange into separate coincident groups the surfaces
            var groupedSurfacesForFront = [];
            var groupedSurfacesForBack = [];
            var ungroupedSurfaces = [];
            // For all surfaces in a valid glue group we want to separate them into a set that contains all triangles facing
            // forward according to the group's reference direction, and a set facing backwards
            for (var keys = Object.keys(__renderList), i = keys.length - 1; i !== -1; i--)
            {
                var geometry = __renderList[keys[i]].geometry;
                if (geometry.hasOwnProperty(["coincidentgroup"]) &&
                    geometry.hasOwnProperty(["coincidentpriority"]) &&
                    geometry.hasOwnProperty(["coincidentgrouptag"]) &&
                    geometry["coincidentgroup"] >= 0 &&
                    geometry["coincidentpriority"] !== 0)
                {
                    var coincidentmeshID = geometry["coincidentmeshID"];
                    var group = geometry["coincidentgroup"];
                    var priority = geometry["coincidentpriority"];
                    var group_tag = geometry["coincidentgrouptag"];

                    var object_front =
                        {
                            "surfaceID": __renderList[keys[i]].surfaceID,
                            "material": {
                                "transparencylayer": false, //this flag indicates if this layer has to be used only when transparency is displayed
                                "coincidentsurface": true,
                                "fromARD": false,
                                "side": (priority > 0) ? THREE.FrontSide : THREE.BackSide
                            }
                        };
                    var object_back =
                        {
                            "surfaceID": __renderList[keys[i]].surfaceID,
                            "material": {
                                "transparencylayer": false, //this flag indicates if this layer has to be used only when transparency is displayed
                                "coincidentsurface": true,
                                "fromARD": false,
                                "side": (priority > 0) ? THREE.BackSide : THREE.FrontSide
                            }
                        };

                    switch (geometry["coincidentgrouptag"])
                    {
                        case "planar" :
                            //discard back appearance for ards
                            if (priority > 0)
                            {
                                object_front["geometry"] = geometry;
                                groupedSurfacesForFront.unshift(object_front);
                            }
                            else
                            {
                                object_back["geometry"] = geometry;
                                groupedSurfacesForBack.unshift(object_back);
                            }
                            break;
                        default:
                        case "tris-same":
                        case "shape-same":

                            var reversewindinggeometry = this.GetInvertedGeometry(geometry);

                            if (priority > 0)
                            {
                                object_front["geometry"] = geometry;
                                object_back["geometry"] = reversewindinggeometry;
                            }
                            else
                            {
                                object_front["geometry"] = reversewindinggeometry;
                                object_back["geometry"] = geometry;
                            }
                            groupedSurfacesForFront.unshift(object_front);
                            groupedSurfacesForBack.unshift(object_back);

                            break;
                    }
                }
                else
                {
                    //Objects that do not need to be ordered are pushed first into the list
                    //and we keep their visibility flag intact
                    var double_sided = true;
                    if (__local_materials_refs.hasOwnProperty(__renderList[keys[i]].surfaceID))
                    {
                        var mat_struct_front = __local_materials_refs[__renderList[keys[i]].surfaceID]["front"];
                        var mat_struct_back = __local_materials_refs[__renderList[keys[i]].surfaceID]["back"];
                        if (mat_struct_front["double_sided"] !== mat_struct_back["double_sided"])
                            console.warn("Front and back material not coherent on double_sided flag");
                        double_sided = mat_struct_front["double_sided"] || mat_struct_back["double_sided"];
                    }

                    var object_front =
                        {
                            "surfaceID": __renderList[keys[i]].surfaceID,
                            "geometry": geometry,
                            "material": {
                                "transparencylayer": false, //this flag indicates if this layer has to be used only when transparency is displayed
                                "coincidentsurface": false,
                                "fromARD": false,
                                "side": THREE.FrontSide
                            }
                        };
                    ungroupedSurfaces.push(object_front);

                    if (double_sided)
                    {
                        var reversewindinggeometry = this.GetInvertedGeometry(geometry);
                        var object_back_inv =
                            {
                                "surfaceID": __renderList[keys[i]].surfaceID,
                                "geometry": reversewindinggeometry,
                                "material": {
                                    "transparencylayer": false, //this flag indicates if this layer has to be used only when transparency is displayed
                                    "coincidentsurface": false,
                                    "fromARD": false,
                                    "side": THREE.BackSide
                                }
                            };
                        ungroupedSurfaces.unshift(object_back_inv);
                    }
                }
            }
            return {"front": groupedSurfacesForFront, "back": groupedSurfacesForBack, "ungrouped": ungroupedSurfaces};
        },

        FilterRenderList: function (__geometry_list, __coincident_group_id, __local_materials_refs)
        {
            var filteredRenderList = [];
            for (var keys = Object.keys(__geometry_list), i = keys.length - 1; i !== -1; i--)
            {
                var s = __geometry_list[keys[i]].surfaceID;
                if (__local_materials_refs.hasOwnProperty(s))
                {
                    var mat_struct_front = __local_materials_refs[s]["front"];
                    var mat_struct_back = __local_materials_refs[s]["back"];
                    var transparent_front_mat = mat_struct_front["opacity"] < 1 || mat_struct_front["transp_texture"];
                    var transparent_back_mat = mat_struct_back["opacity"] < 1 || mat_struct_back["transp_texture"];
                    var transparent = transparent_front_mat || transparent_back_mat;

                    var fromARD = false;
                    var geometry = __geometry_list[keys[i]].geometry;
                    if (geometry.hasOwnProperty(["coincidentgroup"]) &&
                        geometry.hasOwnProperty(["coincidentpriority"]) &&
                        geometry.hasOwnProperty(["coincidentgrouptag"]) &&
                        geometry["coincidentgroup"] >= 0 &&
                        geometry["coincidentpriority"] !== 0)
                    {
                        switch (geometry["coincidentgrouptag"])
                        {
                            case "planar" :
                                fromARD = true;
                                break;
                            default:
                            case "tris-same":
                            case "shape-same":
                                fromARD = false;
                                break;
                        }
                    }

                    if (__coincident_group_id === undefined)
                    {
                        if ((fromARD || transparent)
                            && geometry.hasOwnProperty("coincidentgroup")
                            && geometry["coincidentgroup"] >= 0
                            && geometry["coincidentpriority"] !== 0)
                        {
                            //Create a new coincident group pass (defer coincident processing),
                            //dedicated to ard SURFACES and transparent coincident groups
                            this.trigger("new_coincident_group_pass", {"coincidentgroup": geometry["coincidentgroup"], "transparent": transparent});
                        }
                        else
                        {
                            //Here we keep NON ard SURFACES and NON transparent coincident groups
                            filteredRenderList.push(__geometry_list[i]);
                        }
                    }
                    else
                    {
                        if ((fromARD || transparent)
                            && geometry.hasOwnProperty("coincidentgroup")
                            && geometry["coincidentgroup"] >= 0
                            && geometry["coincidentpriority"] !== 0
                            && (__coincident_group_id == geometry["coincidentgroup"]))
                        {
                            //Here we keep ard SURFACES and transparent coincident groups
                            filteredRenderList.push(__geometry_list[i]);
                        }
                        else
                        {
                            //This case is already handled in first build pass
                        }
                    }
                }
                else
                {
                    LogMessage.Log("ThreeDObjectHandler Warning: Surface '" + s + "' not found in material map :(");
                    filteredRenderList.push(__geometry_list[i]);
                }
            }
            return filteredRenderList;
        }

    });

});
