/**
 * @author DOGO
 * @module WebViewer3D
 * @copyright 2017 Esko Software BVBA. All rights reserved.
 */

define([
    "underscore",
    "backbone",
    "three",
    "@esko/3d-web-viewer/modules/LogMessage",
    "@esko/3d-web-viewer/modules/SceneBuilder",
    "@esko/3d-web-viewer/modules/Utils"
], function (
    _,
    bb,
    THREE,
    LogMessage,
    WV3DSceneBuilder,
    WV3DUtils
)
{

    "use strict";

    //IMPROVEME: use const and let instead of var

    /* PRIVATE */

    const ReleaseNodeViewProperties = function (__nodeviewprops)
    {
        for (var node_url_id = 0, keys = Object.keys(__nodeviewprops), l = keys.length; node_url_id < l; node_url_id++)
        {
            var threejsnodes = __nodeviewprops[keys[node_url_id]].threejsnodes;
            for (var node_id = 0, keys2 = Object.keys(threejsnodes), l2 = keys2.length; node_id < l2; node_id++)
            {
                //Delete nodes in all coincident groups
                for (var node_cg = 0, keys3 = threejsnodes[keys2[node_id]], l3 = keys3.length; node_cg < l3; node_cg++)
                {
                    delete threejsnodes[keys2[node_id]][keys3[node_cg]];
                }
                delete threejsnodes[keys2[node_id]];
            }
            delete __nodeviewprops[keys[node_url_id]].threejsnodes;
            delete __nodeviewprops[keys[node_url_id]];
        }
    };

    /**
     * @alias module:WebViewer3D
     * @class WV3DSceneManager
     * This class takes care of creating/deleting/holding a reference to each 3JS scene.
     * It also triggers scene building via the _scenebuilder
     */
    return bb.Model.extend({

        /* PUBLIC */

        initialize: function (__rendererCache, __model, __materialManager, __objectBuilder, __notifier, __executionQueue)
        {
            LogMessage.Log("Creating WV3DSceneManager Object");
            //IMPROVEME: We could consider using singletons for some of these attributes
            this.set("_renderercache", __rendererCache);
            this.set("_materialmanager", __materialManager);
            this.set("_notifier", __notifier);
            this.set("_exec_queue", __executionQueue);
            this.set("_scenebuilder", new WV3DSceneBuilder(__rendererCache, __model, this, __materialManager, __objectBuilder, __executionQueue));

            this.listenTo(__objectBuilder, "non_empty_scene", this.SetNonEmptyScene);
        },

        NotifyUpdateLightingContrast: function ()
        {
            this.trigger("update_lighting_contrast");
        },

        NotifyAddUIElement: function (__path, __isleaf)
        {
            this.get("_notifier").AddUIElement(__path, __isleaf);
        },

        /**
         * Reset the scene and creates a new one
         */
        ResetScene: function (__coplanarityEpsilon, __boundaryhardedges)
        {
            this.ReleaseScene();

            //Reset the execution queue (shared between notifier, scene builder and object builder)
            this.get("_materialmanager").Reset();
            this.get("_scenebuilder").Reset();

            //Create the scene objects using the asset manager
            this.set("_3JSsceneOp", new THREE.Scene());
            this.set("_3JSsceneOpCoincidentGroups", {});
            this.set("_3JSsceneOpDepth", new THREE.Scene());
            this.set("_3JSsceneOpCoincidentGroupsDepth", []);
            this.set("_3JSsceneOpaqueDiffuse", new THREE.Scene());
            this.set("_3JSsceneOpCoincidentGroupsDiffuse", []);
            this.set("_3JSsceneTransp", new THREE.Scene());
            this.set("_3JSsceneTranspCoincidentGroups", {});
            this.set("_3JSsceneTranspDepth", new THREE.Scene());
            this.set("_3JSsceneTranspCoincidentGroupsDepth", []);
            this.set("_3JSsceneTranspDiffuse", new THREE.Scene());
            this.set("_3JSsceneTranspCoincidentGroupsDiffuse", []);
            this.set("_3JSsceneHightlights", new THREE.Scene());

            this.set("_3JSsceneLabels", new THREE.Scene());
            this.set("_labelCreationCallbackList", []);

            this.set("_3JSsceneBBox", new THREE.Box3());
            this.set("_3JSsceneBBoxFocus", new THREE.Box3());

            this.get("_renderercache").Reset(__coplanarityEpsilon, __boundaryhardedges);
            this.get("_materialmanager").InitializeShaders();

            this.AddDirectionalAndAmbientLightsToScene(this.get("_3JSsceneOp"), "cg_0", false);
            this.AddDirectionalAndAmbientLightsToScene(this.get("_3JSsceneTransp"), "cg_0", true);
        },

        AddDirectionalAndAmbientLightsToScene: function (__scene, __coincident_group_id, __transparent)
        {
            // add a directional light
            if (!this.get("_renderercache").get("lights")["dir_light"]) this.get("_renderercache").get("lights")["dir_light"] = {};
            var directionalLight = new THREE.DirectionalLight(0xaaaaaa, 0.5);//this light will be updated afterwards when updating lighting contrast in UpdateLightingContrast
            this.get("_renderercache").get("lights")["dir_light"][(__transparent ? "tr_" : "") + __coincident_group_id] = directionalLight;
            __scene.add(directionalLight);

            // add subtle ambient lighting
            if (!this.get("_renderercache").get("lights")["amb_light"]) this.get("_renderercache").get("lights")["amb_light"] = {};
            var ambientLight = new THREE.AmbientLight(0x888888);//this light will be updated afterwards when updating lighting contrast in UpdateLightingContrast
            this.get("_renderercache").get("lights")["amb_light"][(__transparent ? "tr_" : "") + __coincident_group_id] = ambientLight;
            __scene.add(ambientLight);
        },

        InitOpaqueCoincidentGroupScene: function (__coincident_group_id)
        {
            if (!this.get("_3JSsceneOpCoincidentGroups").hasOwnProperty(__coincident_group_id))
            {
                this.get("_3JSsceneOpCoincidentGroups")[__coincident_group_id] = new THREE.Scene();
                this.AddDirectionalAndAmbientLightsToScene(this.get("_3JSsceneOpCoincidentGroups")[__coincident_group_id], __coincident_group_id, false);
                this.get("_3JSsceneOpCoincidentGroupsDepth")[__coincident_group_id] = new THREE.Scene();
                this.get("_3JSsceneOpCoincidentGroupsDiffuse")[__coincident_group_id] = new THREE.Scene();
            }
        },

        InitTransparentCoincidentGroupScene: function (__coincident_group_id)
        {
            if (!this.get("_3JSsceneTranspCoincidentGroups").hasOwnProperty(__coincident_group_id))
            {
                this.get("_3JSsceneTranspCoincidentGroups")[__coincident_group_id] = new THREE.Scene();
                this.AddDirectionalAndAmbientLightsToScene(this.get("_3JSsceneTranspCoincidentGroups")[__coincident_group_id], __coincident_group_id, true);
                this.get("_3JSsceneTranspCoincidentGroupsDepth")[__coincident_group_id] = new THREE.Scene();
                this.get("_3JSsceneTranspCoincidentGroupsDiffuse")[__coincident_group_id] = new THREE.Scene();
            }
        },

        /**
         * Build the scene using data from the asset manager
         */
        SetUpScene: function ()
        {
            this.get("_scenebuilder").SetUpScene();

            this.get("_exec_queue").whenFinished(function ()
            {
                var bboxop = new THREE.Box3();
                bboxop.setFromObject(this.get("_3JSsceneOp"));
                for (var cg = 0; cg < this.get('_3JSsceneOpCoincidentGroupsPasses').length; ++cg)
                {
                    var bboxopcg = new THREE.Box3();
                    var cg_id = this.get("_3JSsceneOpCoincidentGroupsPasses")[cg];
                    bboxopcg.setFromObject(this.get("_3JSsceneOpCoincidentGroups")[cg_id]);
                    bboxop.union(bboxopcg);
                }
                var bboxtr = new THREE.Box3();
                bboxtr.setFromObject(this.get("_3JSsceneTransp"));
                bboxop.union(bboxtr);
                for (var cg = 0; cg < this.get('_3JSsceneTranspCoincidentGroupsPasses').length; ++cg)
                {
                    var bboxopcg = new THREE.Box3();
                    var cg_id = this.get("_3JSsceneTranspCoincidentGroupsPasses")[cg];
                    bboxopcg.setFromObject(this.get("_3JSsceneTranspCoincidentGroups")[cg_id]);
                    bboxop.union(bboxopcg);
                }

                this.set("_3JSsceneBBox", bboxop);
                this.set("_3JSsceneBBoxFocus", bboxop);

            }, this);
        },

        /**
         * TRIGGERED EVENT (object_builder::'set_non_empty'): Make sure to trigger proper rendering capabilities
         */
        SetNonEmptyScene: function (__params)
        {
            var __transparent = __params["transparent"];
            if (__transparent)
            {
                this.set("_emptyTransparentScene", false);
            }
            else
            {
                this.set("_emptyOpaqueScene", false);
            }
        },

        ReleaseScene: function ()
        {
            // Texture cache cleared temporarily here since not allowing
            // texture to be properly linked to the material (see use of loadTexture
            // and linkedmaterial in ViewerView.js). Callback is being called before
            // texture can be linked to material when texture already present in cache
            // However, the browser is already caching images, then no overhead can be detected.
            this.get("_materialmanager").ReleaseAllMaterialsAndTextures();
            this.ReleaseSGGeometriesAndMaterials();
            this.ReleaseThreejsnodesReferences();
            THREE.Cache.clear();

            //Reset all attributes
            this.set("_3JSsceneOp", null);
            this.set("_3JSsceneOpDepth", null);
            this.set("_3JSsceneOpaqueDiffuse", null);
            this.set("_3JSsceneTransp", null);
            this.set("_3JSsceneTranspDepth", null);
            this.set("_3JSsceneTranspDiffuse", null);
            this.set("_3JSsceneHightlights", null);
            this.set("_3JSsceneLabels", null);

            this.set("_3JSsceneBBox", null);
            this.set("_3JSsceneBBoxFocus", null);

            this.set("_3JSsceneOpCoincidentGroupsIds", {});
            this.set("_3JSsceneOpCoincidentGroupsPasses", []);
            this.set("_3JSsceneNbOpCoincidentGroups", 0);

            this.set("_3JSsceneTranspCoincidentGroupsIds", {});
            this.set("_3JSsceneTranspCoincidentGroupsPasses", []);
            this.set("_3JSsceneNbTranspCoincidentGroups", 0);

            this.set("_emptyTransparentScene", true);
            this.set("_emptyOpaqueScene", true);

        },

        ReleaseSGGeometriesAndMaterials: function ()
        {
            this.ApplyFctToAllScenes(this.get("_materialmanager").ReleaseGeomMatTex.bind(this.get("_materialmanager")));
        },

        ReleaseThreejsnodesReferences: function ()
        {
            var nodeviewprops = this.get("_renderercache").get("nodeviewproperties");
            if (nodeviewprops)
            {
                ReleaseNodeViewProperties(nodeviewprops);
            }
        },

        ClearInstanceArrays: function ()
        {
            var clear_temp_instanciation_array = function (obj)
            {
                if (obj instanceof THREE.Mesh)
                {
                    if (obj.geometry instanceof THREE.InstancedBufferGeometry)
                    {
                        if (obj.geometry.attributes["inst_translation"].jarray)
                            obj.geometry.attributes["inst_translation"].jarray = null;
                        if (obj.geometry.attributes["inst_rotation"].jarray)
                            obj.geometry.attributes["inst_rotation"].jarray = null;
                        if (obj.geometry.attributes["inst_scale"].jarray)
                            obj.geometry.attributes["inst_scale"].jarray = null;
                    }
                }
            };

            this.TraverseAllScenes(clear_temp_instanciation_array);
        },

        //Scene traversal functions

        ApplyFctToAllColoredScenes: function (__callback)
        {
            //These are the scenes that are contributing to the color of the render
            //each geometry here should be referenced once and only once
            __callback(this.get("_3JSsceneOp"));
            if (this.get("_3JSsceneOpCoincidentGroupsPasses"))
            {
                for (var cg = 0; cg < this.get('_3JSsceneOpCoincidentGroupsPasses').length; ++cg)
                {
                    var cg_id = this.get("_3JSsceneOpCoincidentGroupsPasses")[cg];
                    __callback(this.get("_3JSsceneOpCoincidentGroups")[cg_id]);
                }
            }
            __callback(this.get("_3JSsceneTransp"));
            if (this.get("_3JSsceneTranspCoincidentGroupsPasses"))
            {
                for (var cg = 0; cg < this.get('_3JSsceneTranspCoincidentGroupsPasses').length; ++cg)
                {
                    var cg_id = this.get("_3JSsceneTranspCoincidentGroupsPasses")[cg];
                    __callback(this.get("_3JSsceneTranspCoincidentGroups")[cg_id]);
                }
            }
        },

        ApplyFctToAllNonColoredScenes: function (__callback)
        {
            //These are the scenes that are used to support render passes
            //each geometry is referenced twice, once for depth, once for
            //diffuse.
            __callback(this.get("_3JSsceneOpDepth"));
            __callback(this.get("_3JSsceneOpaqueDiffuse"));
            if (this.get("_3JSsceneOpCoincidentGroupsPasses"))
            {
                for (var cg = 0; cg < this.get('_3JSsceneOpCoincidentGroupsPasses').length; ++cg)
                {
                    var cg_id = this.get("_3JSsceneOpCoincidentGroupsPasses")[cg];
                    __callback(this.get("_3JSsceneOpCoincidentGroupsDepth")[cg_id]);
                    __callback(this.get("_3JSsceneOpCoincidentGroupsDiffuse")[cg_id]);
                }
            }
            __callback(this.get("_3JSsceneTranspDepth"));
            __callback(this.get("_3JSsceneTranspDiffuse"));
            if (this.get("_3JSsceneTranspCoincidentGroupsPasses"))
            {
                for (var cg = 0; cg < this.get('_3JSsceneTranspCoincidentGroupsPasses').length; ++cg)
                {
                    var cg_id = this.get("_3JSsceneTranspCoincidentGroupsPasses")[cg];
                    __callback(this.get("_3JSsceneTranspCoincidentGroupsDepth")[cg_id]);
                    __callback(this.get("_3JSsceneTranspCoincidentGroupsDiffuse")[cg_id]);
                }
            }
        },

        ApplyFctToAllAuxiliaryScenes: function (__callback)
        {
            //These are scene that are used for other rendering elements
            __callback(this.get("_3JSsceneHightlights"));
            __callback(this.get("_3JSsceneLabels"));
        },

        ApplyFctToAllScenes: function (__callback)
        {
            this.ApplyFctToAllColoredScenes(__callback);
            this.ApplyFctToAllNonColoredScenes(__callback);
            this.ApplyFctToAllAuxiliaryScenes(__callback);
        },

        TraverseAllColoredScenes: function (__callback)
        {
            this.ApplyFctToAllColoredScenes(function (__scene)
            {
                __scene.traverse(__callback);
            });
        },

        TraverseAllNonColoredScenes: function (__callback)
        {
            this.ApplyFctToAllNonColoredScenes(function (__scene)
            {
                __scene.traverse(__callback);
            });
        },

        TraverseAllAuxiliaryScenes: function (__callback)
        {
            this.ApplyFctToAllAuxiliaryScenes(function (__scene)
            {
                __scene.traverse(__callback);
            });
        },

        TraverseAllScenes: function (__callback)
        {
            this.ApplyFctToAllScenes(function (__scene)
            {
                __scene.traverse(__callback);
            });
        },


        /* Label creation and update related functions */
        CreateLabels: function ()
        {
            if (this.get("_labelCreationCallbackList").length > 0)
            {
                //Call back the creation of labels and set the list to empty to avoid duplication
                for (var i = 0; i < this.get("_labelCreationCallbackList").length; ++i)
                {
                    this.get("_labelCreationCallbackList")[i]();
                }
                this.set("_labelCreationCallbackList", []);

                //Create ThreeJS sprites for Label assets from JSON
                if (this.get("_3JSsceneLabels").children > 0)
                {
                    LogMessage.Log("ViewerView Warning: Label scene seems not to be empty. Please make sure to call this.get('_3JSsceneLabels') = new THREE.Scene();");
                }
                this.get("_scenebuilder").CreateSprites();
            }
        },

        UpdateLabelProperties: function (__nodeURL, __properties)
        {
            var shortNodeURL = WV3DUtils.URLToKeyString(__nodeURL);

            if (!(shortNodeURL in this.get("_renderercache").get("labels")["list"]))
                return false; //Invalid URL

            var label_asset = this.get("_renderercache").get("labels")["list"][shortNodeURL];
            if (label_asset === undefined)
                throw "UpdateLabelProperties: Cannot find requested label";

            this.get("_scenebuilder").SetLabelProperties(label_asset, __properties);

            return true;
        },

        UpdateAllLabelsProperties: function (__properties)
        {
            var lbl = 0;
            for (var lbl_nodes = Object.keys(this.get("_renderercache").get("labels")["list"]), n_lbls = lbl_nodes.length; lbl < n_lbls; ++lbl)
            {
                var label_asset = this.get("_renderercache").get("labels")["list"][lbl_nodes[lbl]];
                if (label_asset === undefined)
                    throw "UpdateAllLabelsProperties: Cannot find intended label!";

                this.get("_scenebuilder").SetLabelProperties(label_asset, __properties);
            }
            return lbl > 0;
        },

        /*Visibility related functions

        /**
        * Reset visiblity of the main scene to its default (in json structure??? to define)
        * default visiblity of the main scene is true at startup
        *
        * @function
        * @param {Boolean} __redraw - If true, redraw the scene after visibility change - default is true
        */
        ResetMainSceneVisibility: function (__redraw)
        {
            var resetvisibilitycallback = function (obj)
            {
                if (!(obj instanceof THREE.Scene || obj instanceof THREE.Mesh))
                {
                    obj.visible = true;
                }
            };

            this.TraverseAllScenes(resetvisibilitycallback);
            if (__redraw == undefined || __redraw) this.trigger("render");
        },

        /**
         * Picking
         * __x and __y are the pixel coordinates (range [0,screensize in pixels] )
         * returns information on selected object
         */
        Pick: function (__raycaster)
        {
            // LogMessage.Log("Object picked at ("+__x+","+__y+")");
            var objects = [];
            var intersectionCallback = function (__scene)
            {
                //Here we push all objects into the same array (needs to be sorted afterwards)
                var filterShadingPasses = function (__object)
                {
                    if (__object instanceof THREE.Mesh)
                    {
                        //Filter unnecessary shading passes (API user does not care about these)
                        //which would cause an unnecessary overhead (duplicates of same intersections)
                        if (__object.material.name == "filterpm" || __object.material.name == "designer")
                        {
                            Array.prototype.push.apply(objects, __raycaster.intersectObject(__object, true));
                        }
                    }
                };
                __scene.traverse(filterShadingPasses);
            };
            this.ApplyFctToAllColoredScenes(intersectionCallback);

            //Need to sort all objects by depths and coincident surface priorities
            objects.sort(function (a, b)
            {
                //Additionally to checking if surfaces belong to the same object instances
                //and coincident groups, we check the distance since we cannot be sure that
                // a different object, also having coincident groups does not have similar
                //coincident group names. We cannot check for strict equality (even if geometry
                //strictly identical) since due to winding inversion for front VS back, we have
                //a numerical precision issue kicking in.
                //IMPROVEME: Make sure that coincident group geometry share a unique identifier
                //instead of using this epsilon, ObjectBuilder. (would that be overkill?)
                var epsilon = 0.0000001;

                if (a.instanceIndex == b.instanceIndex &&
                    a.object.geometry.hasOwnProperty("coincidentgroup") &&
                    b.object.geometry.hasOwnProperty("coincidentgroup") &&
                    a.object.geometry.coincidentgroup == b.object.geometry.coincidentgroup &&
                    Math.abs(a.distance - b.distance) < epsilon)
                {
                    var coincidentGroupA = parseFloat(a.object.geometry.coincidentgroup);
                    var coincidentPriorityA = parseFloat(a.object.geometry.coincidentpriority);
                    var coincidentGroupB = parseFloat(b.object.geometry.coincidentgroup);
                    var coincidentPriorityB = parseFloat(b.object.geometry.coincidentpriority);

                    var normalA = a.face.normal.clone();
                    var normalB = b.face.normal.clone();
                    //Safety checks
                    if (!normalA.equals(normalB) ||
                        normalA.dot(__raycaster.ray.direction) > 0 ||
                        normalB.dot(__raycaster.ray.direction) > 0)
                    {
                        console.warn("Inconsistent normal orientation: Dev Error ?");
                    }

                    var orderForFrontViewA = (coincidentPriorityA > 0 && !a.object.geometry.hasOwnProperty("HasInvertedWinding")) ||
                        (coincidentPriorityA < 0 && a.object.geometry.hasOwnProperty("HasInvertedWinding"));
                    var orderForFrontViewB = (coincidentPriorityB > 0 && !b.object.geometry.hasOwnProperty("HasInvertedWinding")) ||
                        (coincidentPriorityB < 0 && b.object.geometry.hasOwnProperty("HasInvertedWinding"));

                    if (orderForFrontViewA != orderForFrontViewB)
                    {
                        console.warn("Compared surfaces have non consistent view order: Dev Error ?");
                    }

                    //In this case we want to order surfaces from the near-most to the further-most position
                    //so that ray cast intersections can be ordered correctly for the user.
                    //This order is different from how coincident surfaces are ordered for rendering
                    //(further-most to near-most)
                    if (orderForFrontViewA)
                        return Math.abs(coincidentPriorityB) - Math.abs(coincidentPriorityA);//a closer/lower than b if a priority higher
                    else
                        return Math.abs(coincidentPriorityA) - Math.abs(coincidentPriorityB);//b closer/lower than a if a priority higher
                }
                else
                {
                    return a.distance - b.distance;
                }
            });

            var intersectionsInfo = [];
            for (var k = 0, _objectsKeys = Object.keys(objects), l = objects.length; k < l; ++k)
            {
                var intersection = objects[_objectsKeys[k]];
                var intersectionInfo =
                    {
                        "point": intersection.point,
                        "uv": intersection.uv,
                        "faceId": intersection.faceIndex,
                        "instanceId": intersection.instanceIndex,
                        "objectId": intersection.object.uuid,
                        "geometryId": intersection.object.geometry.uuid,
                        "materialId": intersection.object.material.JSONMaterialName,
                        "objectURL": intersection.object.parent.view_prop_nodes ? intersection.object.parent.view_prop_nodes.nodeURL : null
                    };
                intersectionsInfo.push(intersectionInfo);
            }

            return intersectionsInfo;
        },


        /* Stats and debug */

        /**
         * Prints a memory report
         * @function LogMemoryReport
         */
        LogMemoryReport: function ()
        {
            LogMessage.Log("Node View Properties Size: " + WV3DUtils.SizeOf(this.get("_renderercache").get("nodeviewproperties")) / 1000000 + "MB");
            LogMessage.Log("_3JSsceneOp Size: " + WV3DUtils.SizeOf(this.get("_3JSsceneOp")) / 1000000 + "MB");
            LogMessage.Log("_3JSsceneOpDepth Size: " + WV3DUtils.SizeOf(this.get("_3JSsceneOpDepth")) / 1000000 + "MB");
            LogMessage.Log("_3JSsceneOpaqueDiffuse Size: " + WV3DUtils.SizeOf(this.get("_3JSsceneOpaqueDiffuse")) / 1000000 + "MB");
            LogMessage.Log("_3JSsceneTransp Size: " + WV3DUtils.SizeOf(this.get("_3JSsceneTransp")) / 1000000 + "MB");
            LogMessage.Log("_3JSsceneTranspDepth Size: " + WV3DUtils.SizeOf(this.get("_3JSsceneTranspDepth")) / 1000000 + "MB");
            LogMessage.Log("_3JSsceneTranspDiffuse Size: " + WV3DUtils.SizeOf(this.get("_3JSsceneTranspDiffuse")) / 1000000 + "MB");

            //Get size of opaque scene for each coincident group, with its own shading passes
            for (var cg = 0; cg < this.get('_3JSsceneOpCoincidentGroupsPasses').length; ++cg)
            {
                var cg_id = this.get("_3JSsceneOpCoincidentGroupsPasses")[cg];
                LogMessage.Log("_3JSsceneOpCoincidentGroups[" + cg_id + "] Size: " + WV3DUtils.SizeOf(this.get("_3JSsceneOpCoincidentGroups")[cg_id]) / 1000000 + "MB");
                LogMessage.Log("_3JSsceneOpCoincidentGroupsDepth[" + cg_id + "] Size: " + WV3DUtils.SizeOf(this.get("_3JSsceneOpCoincidentGroupsDepth")[cg_id]) / 1000000 + "MB");
                LogMessage.Log("_3JSsceneOpCoincidentGroupsDiffuse[" + cg_id + "] Size: " + WV3DUtils.SizeOf(this.get("_3JSsceneOpCoincidentGroupsDiffuse")[cg_id]) / 1000000 + "MB");
            }

            //Get size of opaque scene for each transparent coincident group, with its own shading passes
            for (var cg = 0; cg < this.get('_3JSsceneTranspCoincidentGroupsPasses').length; ++cg)
            {
                var cg_id = this.get("_3JSsceneTranspCoincidentGroupsPasses")[cg];
                LogMessage.Log("_3JSsceneTranspCoincidentGroups[" + cg_id + "] Size: " + WV3DUtils.SizeOf(this.get("_3JSsceneTranspCoincidentGroups")[cg_id]) / 1000000 + "MB");
                LogMessage.Log("_3JSsceneTranspCoincidentGroupsDepth[" + cg_id + "] Size: " + WV3DUtils.SizeOf(this.get("_3JSsceneTranspCoincidentGroupsDepth")[cg_id]) / 1000000 + "MB");
                LogMessage.Log("_3JSsceneTranspCoincidentGroupsDiffuse[" + cg_id + "] Size: " + WV3DUtils.SizeOf(this.get("_3JSsceneTranspCoincidentGroupsDiffuse")[cg_id]) / 1000000 + "MB");
            }

            LogMessage.Log("_3JSsceneHightlights Size: " + WV3DUtils.SizeOf(this.get("_3JSsceneHightlights")) / 1000000 + "MB");
            LogMessage.Log("_3JSsceneLabels Size: " + WV3DUtils.SizeOf(this.get("_3JSsceneLabels")) / 1000000 + "MB");
        },

        /**
         * Print list of opaque scene vertices in world coordinates
         */
        PrintOpaqueSceneVertices: function ()
        {
            LogMessage.Log("List of vertices (opaques scene)");

            var v1 = new THREE.Vector3();

            var OuputVertex = function (v)
            {
                if (v instanceof THREE.Vector3 && !isNaN(v.x) && !isNaN(v.y) && !isNaN(v.z))
                {
                    LogMessage.Log("(" + v.x + "," + v.y + "," + v.z + ")");
                }
            };

            var PrintNodeVertices = function (node)
            {
                var geometry = node.geometry;
                if (geometry !== undefined)
                {
                    //We have to distinguish between standard threejs geometry type
                    //and threejs buffered geometry since the access method is different
                    //for the two types. (vector type copy VS index array access)
                    //In addition to that we must transformation again instanced buffer
                    //geometry (inherits buffered geometry) that contains additional transformation
                    // in its 'inst_translation', 'inst_rotation', and 'inst_scale' (instantiation) attributes
                    if (geometry instanceof THREE.Geometry)
                    {
                        var vertices = geometry.vertices;
                        for (var i = 0, il = vertices.length; i < il; ++i)
                        {
                            v1.copy(vertices[i]);
                            v1.applyMatrix4(node.matrixWorld);
                            OuputVertex(v1);
                        }
                    }
                    else if (geometry instanceof THREE.BufferGeometry && geometry.attributes["position"] !== undefined)
                    {
                        var positions = geometry.attributes["position"].array;
                        var localTempBox;
                        if (geometry instanceof THREE.InstancedBufferGeometry)
                        {
                            localTempBox = new THREE.Box3();
                        }
                        for (var i = 0, il = positions.length; i < il; i += 3)
                        {

                            v1.fromArray(positions, i);
                            v1.applyMatrix4(node.matrixWorld);

                            if (geometry instanceof THREE.InstancedBufferGeometry
                                && geometry.attributes["inst_translation"] !== undefined
                                && geometry.attributes["inst_rotation"] !== undefined
                                && geometry.attributes["inst_scale"] !== undefined)
                            {
                                for (var ins = 0, nbins = geometry.maxInstancedCount; ins < nbins; ++ins)
                                {
                                    var transform = THREE.GetInstanceTransform(geometry.attributes["inst_translation"],
                                        geometry.attributes["inst_rotation"],
                                        geometry.attributes["inst_scale"], ins);

                                    OuputVertex(THREE.Vector3.transformVertex(v1, transform));
                                }
                            }
                            else
                            {
                                OuputVertex(v1);
                            }
                        }
                    }
                }
            };

            this.get("_3JSsceneOp").updateMatrixWorld(true);
            this.get("_3JSsceneOp").traverse(PrintNodeVertices);
            for (var cg = 0; cg < this.get('_3JSsceneOpCoincidentGroupsPasses').length; ++cg)
            {
                var cg_id = this.get("_3JSsceneOpCoincidentGroupsPasses")[cg];
                this.get("_3JSsceneOpCoincidentGroups")[cg_id].updateMatrixWorld(true);
                this.get("_3JSsceneOpCoincidentGroups")[cg_id].traverse(PrintNodeVertices);
            }
        }


    });

});
