/**
 * @author DOGO
 * @module WebViewer3D
 * @copyright   2014 Esko Stonecube Ltd. All rights reserved.
 */

define([
	"underscore",
	"backbone",
	"@esko/3d-web-viewer/libs/detect", //IMPROVEME: use another library since this one is not supported by NPM and has then to be copied manually
	"@esko/3d-web-viewer/modules/Constants",
	"@esko/3d-web-viewer/modules/LogMessage"
], function(_, bb, detect, WV3DConstants, LogMessage) {
	"use strict";

	//IMPROVEME: use const and let instead of var

	/**
	 * @alias module:WebViewer3D
	 * @class WV3DSystemChecker
	 * Warning : this class is a singleton, any modification to one of its component will change for each reference to it.
	 */
	const WV3DSystemChecker = bb.Model.extend({
		initialize: function() {
			this.set("_ua", detect.parse(navigator.userAgent));

			this.UpdateSystemSupportLevel();
			this.UpdateSystemReport();
		},

		/**
		 * Implementation details to get an indication of the degree to which the current system is supported for 3D viewing
		 *
		 * @function
		 */
		UpdateSystemSupportLevel: function() {
			LogMessage.Log("SystemChecker: Update system support level");

			this.set("_issues", {});
			var wglSupport = WV3DConstants.wglSsupportLevel.NONE;
			if (!!window.WebGLRenderingContext) {
				var canvas = document.createElement("canvas"),
					names = ["webgl", "experimental-webgl", "moz-webgl", "webkit-3d"],
					context = false;

				for (var i = 0; i < 4; i++) {
					try {
						context = canvas.getContext(names[i]);
						if (context && typeof context.getParameter == "function") wglSupport = WV3DConstants.wglSsupportLevel.FULL;
					} catch (e) {}
				}

				if (wglSupport != WV3DConstants.wglSsupportLevel.FULL) {
					this.get("_issues")["webgl"] = "Rendering context supported, but WebGL probably user disabled!";
					wglSupport = WV3DConstants.wglSsupportLevel.PARTIAL; //WebGL disabled by user!?
				}
			} else {
				this.get("_issues")["webgl"] = "WebGL not supported!";
				wglSupport = WV3DConstants.wglSsupportLevel.NONE;
			}

			//do other support checks (but what? GPU/CPU info looks like are not possible!)
			var OSName = "Unknown OS";
			if (navigator.appVersion.indexOf("Win") != -1) this.set("_os_name", "Windows");
			if (navigator.appVersion.indexOf("Mac") != -1) this.set("_os_name", "MacOS");
			if (navigator.appVersion.indexOf("X11") != -1) this.set("_os_name", "UNIX");
			if (navigator.appVersion.indexOf("Linux") != -1) this.set("_os_name", "Linux");

			var overallSupportLevel = WV3DConstants.supportLevel.FULL;
			if (wglSupport == WV3DConstants.wglSsupportLevel.PARTIAL /*but if other checks also turned out as partial*/)
				overallSupportLevel = WV3DConstants.supportLevel.PARTIAL;
			else if (wglSupport == WV3DConstants.wglSsupportLevel.NONE)
				//No need to bother other checks, no webgl available!
				overallSupportLevel = WV3DConstants.supportLevel.NONE;

			this.set("_overallSupportLevel", overallSupportLevel);
		},

		/**
		 * Get the degree to which the current system is supported for 3D viewing
		 *
		 * @function
		 * @param {Array} issues - an array of issues causing the limited support level
		 * @returns supportLevel.FULL | supportLevel.PARTIAL | supportLevel.NONE
		 */
		GetSystemSupportLevel: function() {
			LogMessage.Log("SystemChecker: Get system support level");
			return { issues: this.get("_issues"), support_level: this.get("_overallSupportLevel") };
		},

		/**
		 * Implementation to update the technical report containing system info
		 * sets local attributes for further query - see example of antialising ability
		 * query in ViewerView.js for polygonoffset
		 *
		 * @function
		 */
		UpdateSystemReport: function() {
			LogMessage.Log("SystemChecker: Update readable system report");

			var canvaselem = document.createElement("canvas");
			var glNames = ["webgl", "experimental-webgl", "moz-webgl", "webkit-3d"];
			var gl, glMPC;
			var s = "",
				JSE;

			for (var i = 0; i < 4; i++) {
				gl = canvaselem.getContext(glNames[i]);
				glMPC = canvaselem.getContext(glNames[i], { failIfMajorPerformanceCaveat: true });
				if (gl && typeof gl.getParameter == "function") break;
			}

			if (gl === null) {
				if (
					confirm(
						"This viewer could not be started since your browser does not support WebGL. Do you want to enable WebGL in your browser?"
					)
				) {
					window.open("https://wiki.esko.com/pages/viewpage.action?pageId=170058865");
				}
				console.error("No WebGL support detected");
				return;
			}

			this.RunRenderTargetTests(gl);

			//The following iteration is not unsafe javascript we actually rely on JS engine
			// difference of behavior to tell us which version it is.
			for (var x in { 3: 3, 1: 1 }) {
				s += x;
			}
			if (s == "31") JSE = "JSC";
			else JSE = "V8";

			var numplugins = navigator.plugins.length;
			var installedplugs = "";
			for (var i = 0; i < numplugins; i++) installedplugs += navigator.plugins[i].name + ", ";

			this.set("_glversion", gl.getParameter(gl.VERSION));
			this.set("_glshadinglanguage", gl.getParameter(gl.SHADING_LANGUAGE_VERSION));
			this.set("_glvendor", gl.getParameter(gl.VENDOR));
			this.set("_glrenderer", gl.getParameter(gl.RENDERER));
			this.set("_glantialias", gl.getContextAttributes().antialias);
			this.set("_glstencilbits", gl.getParameter(gl.STENCIL_BITS));
			this.set("_glmaxrenderbuffersize", gl.getParameter(gl.MAX_RENDERBUFFER_SIZE));
			this.set("_glmaxcombinedtextureimageunits", gl.getParameter(gl.MAX_COMBINED_TEXTURE_IMAGE_UNITS));
			this.set("_glmaxcubemaptexturesize", gl.getParameter(gl.MAX_CUBE_MAP_TEXTURE_SIZE));
			this.set("_glmaxfragmentuniformvectors", gl.getParameter(gl.MAX_FRAGMENT_UNIFORM_VECTORS));
			this.set("_glmaxtextureimageunits", gl.getParameter(gl.MAX_TEXTURE_IMAGE_UNITS));
			this.set("_glmaxtexturesize", gl.getParameter(gl.MAX_TEXTURE_SIZE));
			this.set("_glmaxvaryingvectors", gl.getParameter(gl.MAX_VARYING_VECTORS));
			this.set("_dlmaxvertexattribs", gl.getParameter(gl.MAX_VERTEX_ATTRIBS));
			this.set("_glmaxvertextextureimageunits", gl.getParameter(gl.MAX_VERTEX_TEXTURE_IMAGE_UNITS));
			this.set("_glmaxvertexuniformvectors", gl.getParameter(gl.MAX_VERTEX_UNIFORM_VECTORS));
			this.set("_glmaxviewportdims", gl.MAX_VIEWPORT_DIMS);
			this.set("_glaliasedlinewidthrange0", gl.getParameter(gl.ALIASED_LINE_WIDTH_RANGE)[0]);
			this.set("_glaliasedlinewidthrange1", gl.getParameter(gl.ALIASED_LINE_WIDTH_RANGE)[1]);
			this.set("_glaliasedpointsizerange0", gl.getParameter(gl.ALIASED_POINT_SIZE_RANGE)[0]);
			this.set("_glaliasedpointsizerange1", gl.getParameter(gl.ALIASED_POINT_SIZE_RANGE)[1]);
			this.set(
				"_gltexanistpyavail",
				gl.getExtension("EXT_texture_filter_anisotropic") ||
					gl.getExtension("WEBKIT_EXT_texture_filter_anisotropic") ||
					gl.getExtension("MOZ_EXT_texture_filter_anisotropic")
			);
			if (this.get("_gltexanistpyavail")) {
				this.set("_glmaxtexanistpy", gl.getParameter(this.get("_gltexanistpyavail").MAX_TEXTURE_MAX_ANISOTROPY_EXT));
			} else {
				this.set("_glmaxtexanistpy", "NA");
			}
			this.set("_vshighfloatprecision", gl.getShaderPrecisionFormat(gl.VERTEX_SHADER, gl.HIGH_FLOAT).precision);
			this.set("_vsmediumfloatprecision", gl.getShaderPrecisionFormat(gl.VERTEX_SHADER, gl.MEDIUM_FLOAT).precision);
			this.set("_vslowfloatprecision", gl.getShaderPrecisionFormat(gl.VERTEX_SHADER, gl.LOW_FLOAT).precision);
			this.set("_fshighfloatprecision", gl.getShaderPrecisionFormat(gl.FRAGMENT_SHADER, gl.HIGH_FLOAT).precision);
			this.set("_fsmediumfloatprecision", gl.getShaderPrecisionFormat(gl.FRAGMENT_SHADER, gl.MEDIUM_FLOAT).precision);
			this.set("_fslowfloatprecision", gl.getShaderPrecisionFormat(gl.FRAGMENT_SHADER, gl.LOW_FLOAT).precision);
			this.set("_fshighintprecision", gl.getShaderPrecisionFormat(gl.FRAGMENT_SHADER, gl.HIGH_INT).rangeMax);

			//simulate mobile architecture (debug only)
			// this.set('_glmaxtextureimageunits',8);

			var systemReport =
				this.get("_ua").os.family +
				"<BR/>" +
				this.get("_ua").browser.family +
				"<BR/>" +
				navigator.userAgent +
				"<BR/>" +
				"CPU Cores# " +
				navigator.hardwareConcurrency +
				"<BR/>" + //Available on Chrome and Opera, soon on safari
				"JavaScript Engine: " +
				JSE +
				"<BR/>" +
				"GL_VERSION: " +
				this.get("_glversion") +
				"<BR/>" +
				"SHADING_LANGUAGE_VERSION: " +
				this.get("_glshadinglanguage") +
				"<BR/>" +
				"VENDOR: " +
				this.get("_glvendor") +
				"<BR/>" +
				"RENDERER: " +
				this.get("_glrenderer") +
				"<BR/>" +
				"Anti-Aliasing: " +
				this.get("_glantialias") +
				"<BR/>" +
				"Stencil Bits: " +
				this.get("_glstencilbits") +
				"<BR/>" +
				"MajorPerformance Caveat: " +
				!glMPC +
				"<BR/>" +
				"MAX_RENDERBUFFER_SIZE: " +
				this.get("_glmaxrenderbuffersize") +
				"<BR/>" +
				"MAX_COMBINED_TEXTURE_IMAGE_UNITS: " +
				this.get("_glmaxcombinedtextureimageunits") +
				"<BR/>" +
				"MAX_CUBE_MAP_TEXTURE_SIZE: " +
				this.get("_glmaxcubemaptexturesize") +
				"<BR/>" +
				"MAX_FRAGMENT_UNIFORM_VECTORS: " +
				this.get("_glmaxfragmentuniformvectors") +
				"<BR/>" +
				"MAX_TEXTURE_IMAGE_UNITS: " +
				this.get("_glmaxtextureimageunits") +
				"<BR/>" +
				"MAX_TEXTURE_SIZE: " +
				this.get("_glmaxtexturesize") +
				"<BR/>" +
				"MAX_VARYING_VECTORS: " +
				this.get("_glmaxvaryingvectors") +
				"<BR/>" +
				"MAX_VERTEX_ATTRIBS: " +
				this.get("_dlmaxvertexattribs") +
				"<BR/>" +
				"MAX_VERTEX_TEXTURE_IMAGE_UNITS: " +
				this.get("_glmaxvertextextureimageunits") +
				"<BR/>" +
				"MAX_VERTEX_UNIFORM_VECTORS: " +
				this.get("_glmaxvertexuniformvectors") +
				"<BR/>" +
				"MAX_VIEWPORT_DIMS: " +
				this.get("_glmaxviewportdims") +
				"<BR/>" +
				"ALIASED_LINE_WIDTH_RANGE: " +
				this.get("_glaliasedlinewidthrange0") +
				", " +
				this.get("_glaliasedlinewidthrange1") +
				"<BR/>" +
				"ALIASED_POINT_SIZE_RANGE: " +
				this.get("_glaliasedpointsizerange0") +
				", " +
				this.get("_glaliasedpointsizerange1") +
				"<BR/>" +
				"MAX_TEXTURE_ANISITROPY: " +
				this.get("_glmaxtexanistpy") +
				"<BR/>" +
				"VERTEX_SHADER HIGH_FLOAT Precision: " +
				this.get("_vshighfloatprecision") +
				"<BR/>" +
				"VERTEX_SHADER MEDIUM_FLOAT Precision: " +
				this.get("_vsmediumfloatprecision") +
				"<BR/>" +
				"VERTEX_SHADER LOW_FLOAT Precision: " +
				this.get("_vslowfloatprecision") +
				"<BR/>" +
				"FRAG_SHADER HIGH_FLOAT Precision: " +
				this.get("_fshighfloatprecision") +
				"<BR/>" +
				"FRAG_SHADER MEDIUM_FLOAT Precision: " +
				this.get("_fsmediumfloatprecision") +
				"<BR/>" +
				"FRAG_SHADER LOW_FLOAT Precision: " +
				this.get("_fslowfloatprecision") +
				"<BR/>" +
				"FRAG_SHADER HIGH_INT Precision: " +
				this.get("_fshighintprecision") +
				"<BR/>" +
				"Supported EXTS: " +
				gl.getSupportedExtensions() +
				"<BR/>" +
				"Installed Plugins: " +
				installedplugs +
				"<BR/>" +
				"Floating point texture support: " +
				this.get("_fullSupportFloatingPointTextures") +
				"<BR/>" +
				"Half Floating point texture support: " +
				this.get("_fullSupportHFloatingPointTextures");

			this.set("_systemReport", systemReport);
		},

		RunRenderTargetTests: function(__gl) {
			function glEnum(gl, v) {
				//The following iteration is not unsafe javascript we actually rely on prototype
				// propertie give us the matching human readable glEnum name.
				for (var key in gl) {
					if (gl[key] === v) {
						return key;
					}
				}
				return "0x" + v.toString(16);
			}

			function getExt(gl, name, msg) {
				var ext = gl.getExtension(name);
				LogMessage.Log("SystemChecker " + (ext ? "can " : "can **NOT** ") + msg);
				return ext;
			}

			var testFloat = getExt(__gl, "OES_texture_float", "make floating point textures");
			getExt(__gl, "OES_texture_float_linear", "linear filter floating point textures");
			var testHalfFloat = getExt(__gl, "OES_texture_half_float", "make half floating point textures");
			getExt(__gl, "OES_texture_half_float_linear", "linear filter half floating point textures");

			__gl.HALF_FLOAT_OES = 0x8d61;

			// setup GLSL program
			var vshader_source =
				" attribute vec4 a_position;\
                            void main()\
                            {\
                              gl_Position = a_position;\
                            }";
			var fshader_source =
				"  precision mediump float;\
                              uniform vec4 u_color;\
                              uniform sampler2D u_texture;\
                              void main()\
                              {\
                                gl_FragColor = texture2D(u_texture, vec2(0.5, 0.5)) * u_color;\
                              }";
			var program = __gl.createProgram();

			var fragmentShader = __gl.createShader(__gl.FRAGMENT_SHADER);
			var vertexShader = __gl.createShader(__gl.VERTEX_SHADER);

			__gl.shaderSource(vertexShader, vshader_source);
			__gl.shaderSource(fragmentShader, fshader_source);

			__gl.compileShader(vertexShader);
			// LogMessage.Log("Vertex shader info: "+__gl.getShaderInfoLog( vertexShader ));
			__gl.compileShader(fragmentShader);
			// LogMessage.Log("Vertex shader info: "+__gl.getShaderInfoLog( fragmentShader ));

			__gl.attachShader(program, vertexShader);
			__gl.attachShader(program, fragmentShader);

			__gl.linkProgram(program);

			__gl.useProgram(program);

			// look up where the vertex data needs to go.
			var positionLocation = __gl.getAttribLocation(program, "a_position");
			var colorLoc = __gl.getUniformLocation(program, "u_color");

			// provide texture coordinates for the rectangle.
			var positionBuffer = __gl.createBuffer();
			__gl.bindBuffer(__gl.ARRAY_BUFFER, positionBuffer);
			__gl.bufferData(
				__gl.ARRAY_BUFFER,
				new Float32Array([-1.0, -1.0, 1.0, -1.0, -1.0, 1.0, -1.0, 1.0, 1.0, -1.0, 1.0, 1.0]),
				__gl.STATIC_DRAW
			);
			__gl.enableVertexAttribArray(positionLocation);
			__gl.vertexAttribPointer(positionLocation, 2, __gl.FLOAT, false, 0, 0);

			var whiteTex = __gl.createTexture();
			__gl.bindTexture(__gl.TEXTURE_2D, whiteTex);
			__gl.texImage2D(__gl.TEXTURE_2D, 0, __gl.RGBA, 1, 1, 0, __gl.RGBA, __gl.UNSIGNED_BYTE, new Uint8Array([255, 255, 255, 255]));

			function test(format) {
				var tex = __gl.createTexture();
				__gl.bindTexture(__gl.TEXTURE_2D, tex);
				__gl.texImage2D(__gl.TEXTURE_2D, 0, __gl.RGBA, 1, 1, 0, __gl.RGBA, format, null);
				__gl.texParameteri(__gl.TEXTURE_2D, __gl.TEXTURE_MIN_FILTER, __gl.NEAREST);
				__gl.texParameteri(__gl.TEXTURE_2D, __gl.TEXTURE_MAG_FILTER, __gl.NEAREST);

				var fb = __gl.createFramebuffer();
				__gl.bindFramebuffer(__gl.FRAMEBUFFER, fb);
				__gl.framebufferTexture2D(__gl.FRAMEBUFFER, __gl.COLOR_ATTACHMENT0, __gl.TEXTURE_2D, tex, 0);
				var status = __gl.checkFramebufferStatus(__gl.FRAMEBUFFER);
				if (status !== __gl.FRAMEBUFFER_COMPLETE) {
					LogMessage.Log("SystemChecker can **NOT** render to " + glEnum(__gl, format) + " texture");
					return;
				}

				// Draw the rectangle.
				__gl.bindTexture(__gl.TEXTURE_2D, whiteTex);
				__gl.uniform4fv(colorLoc, [0, 10, 20, 1]);
				__gl.drawArrays(__gl.TRIANGLES, 0, 6);

				__gl.bindTexture(__gl.TEXTURE_2D, tex);
				__gl.bindFramebuffer(__gl.FRAMEBUFFER, null);

				__gl.clearColor(1, 0, 0, 1);
				__gl.clear(__gl.COLOR_BUFFER_BIT);

				__gl.uniform4fv(colorLoc, [0, 1 / 10, 1 / 20, 1]);
				__gl.drawArrays(__gl.TRIANGLES, 0, 6);

				var pixel = new Uint8Array(4);
				__gl.readPixels(0, 0, 1, 1, __gl.RGBA, __gl.UNSIGNED_BYTE, pixel);

				if (pixel[0] !== 0 || pixel[1] < 248 || pixel[2] < 248 || pixel[3] < 254) {
					LogMessage.Log("SystemChecker was not able to actually render to " + glEnum(__gl, format) + " texture");
					return false;
				} else {
					LogMessage.Log("SystemChecker succesfully rendered to " + glEnum(__gl, format) + " texture");
					return true;
				}
			}

			this.set("_fullSupportFloatingPointTextures", testFloat && test(__gl.FLOAT));
			this.set("_fullSupportHFloatingPointTextures", testHalfFloat && test(__gl.HALF_FLOAT_OES));

			// this.set('_fullSupportFloatingPointTextures',false);//debug - remove before submit
		},

		/**
		 * Implementation to get the technical report containing system info
		 *
		 * @function
		 * @returns {String} a report containing system info
		 */
		GetSystemReport: function() {
			LogMessage.Log("SystemChecker: Get readable system report");
			return this.get("_systemReport");
		},

		IsWindows: function() {
			return this.get("_ua").os.family.search("Windows") != -1;
		},

		IsMac: function() {
			return this.get("_ua").os.family.search("Mac") != -1;
		},

		IsiOSFirstGen: function() {
			return this.get("_ua").os.family.search("iOS") != -1 && this.get("_glversion").search("SGX") != -1;
		},

		IsiOSSecondGen: function() {
			return this.get("_ua").os.family.search("iOS") != -1 && this.get("_glversion").search("SGX") == -1;
		},

		IsiPad: function() {
			return this.get("_ua").os.family.search("iPad") != -1 || this.get("_ua").source.search("iPad") != -1;
		},

		IsiPod: function() {
			return this.get("_ua").os.family.search("iPod") != -1 || this.get("_ua").source.search("iPod") != -1;
		},

		IsiPhone: function() {
			return (
				this.get("_ua").os.family.search("iPhone") != -1 ||
				(this.get("_ua").source.search("iPhone") != -1 && this.get("_ua").source.search("iPod") == -1)
			);
		},

		IsAndroid: function() {
			return this.get("_ua").os.family.search("Android") != -1;
		},

		IsIE: function() {
			return this.get("_ua").browser.family.search("IE") != -1;
		},

		IsFirefox: function() {
			return this.get("_ua").browser.family.search("Firefox") != -1;
		},

		IsChrome: function() {
			return this.get("_ua").browser.family.search("Chrome") != -1;
		},

		IsSafari: function() {
			return this.get("_ua").browser.family.search("Safari") != -1;
		},

		/** Test function */
		ControllerTestFunction: function() {
			LogMessage.Log("Controller Test function");

			this.get("_model").ModelTestFunction();
		},

		/** Test function 2 */
		ControllerTestFunction2: function() {
			LogMessage.Log("Controller Test function 2 ");
		}
	});

	return new WV3DSystemChecker(); //New is what makes it a singleton, see usage in ViewerController.js and shaders
});
