/**
 * @author DOGO
 * @module WebViewer3D
 * @copyright 2017 Esko Software BVBA. All rights reserved.
 */

define([
    "underscore",
    "backbone",
    "three",
    "@esko/3d-web-viewer/modules/LogMessage",
    "@esko/3d-web-viewer/modules/Constants",
    "@esko/3d-web-viewer/modules/Math",
    "@esko/3d-web-viewer/modules/Utils"
], function (
    _,
    bb,
    THREE,
    LogMessage,
    WV3DConstants,
    WV3DMath,
    WV3DUtils
)
{

    "use strict";

    //IMPROVEME: use const and let instead of var

    /* PRIVATE */

    /* Select texture wrapping mode*/
    const SelectTextureWrappingMode = function (__texture, __properties, __mat_struct)
    {
        if (__properties.hasOwnProperty("original_width"))
        {
            __mat_struct["originalTexWidth"] = parseFloat(__properties["original_width"]);
        }
        if (__properties.hasOwnProperty("original_height"))
        {
            __mat_struct["originalTexHeight"] = parseFloat(__properties["original_height"]);
        }
        switch (__properties["swrap"])
        {
            default:
            case "GL_REPEAT":
                __texture.wrapS = THREE.RepeatWrapping;
                // THREE.MirroredRepeatWrapping;
                break;
            case "GL_CLAMP":
            case "GL_CLAMP_TO_BORDER":
            case "GL_CLAMP_TO_EDGE":
                __texture.wrapS = THREE.ClampToEdgeWrapping;
                break;
        }
        switch (__properties["twrap"])
        {
            default:
            case "GL_REPEAT":
                __texture.wrapT = THREE.RepeatWrapping;
                // THREE.MirroredRepeatWrapping;
                break;
            case "GL_CLAMP":
            case "GL_CLAMP_TO_BORDER":
            case "GL_CLAMP_TO_EDGE":
                __texture.wrapT = THREE.ClampToEdgeWrapping;
                break;
        }
    };

    const GetFileList = function (__path, __files)
    {
        var files = [];
        for (var i = 0, keys = Object.keys(__files), l = keys.length; i < l; ++i)
        {
            var file_idx = keys[i];
            var filename = __path + __files[file_idx];
            files.push(filename);
        }
        return files;
    };

    /**
     * @alias module:WebViewer3D
     * @class WV3DSceneBuilder
     * This class is responsible for querying the png/jpeg images
     * for irradiance, environments, diffuse, specular, and holographics 3JS
     * textures from the json assets, and updating the given material structure
     * it is assigned to.
     */
    return bb.Model.extend({

        /* PUBLIC */

        initialize: function (__rendererCache, __assetManager, __executionQueue)
        {
            LogMessage.Log("Creating WV3DTextureManager Object");
            //IMPROVEME: We could consider using singletons for some of these attributes
            this.set("_renderercache", __rendererCache);
            this.set("_assetmanager", __assetManager);
            this.set("_exec_queue", __executionQueue);
        },

        Reset: function ()
        {
            this.set("_QueriedGlobalTextures", false);
            this.set("_TextureSizeWarningMessageDisplayed", false);
            //IMPROVEME: this flag is not set to false and will dictate
            //all scene notifications error return in CheckPostSceneLoadActions,
            //which is a bit weird
            this.set("_TextureSizeByPass", true);
        },

        CheckTextureSizeByPass: function (__texture)
        {
            var confirmed = true;
            if (__texture["image"]["naturalHeight"] > WV3DConstants.textureSupport.MAX_NATURAL_HEIGHT
                || __texture["image"]["naturalWidth"] > WV3DConstants.textureSupport.MAX_NATURAL_WIDTH)
            {
                if (!this.get("_TextureSizeWarningMessageDisplayed"))
                {
                    confirmed = window.confirm("WARNING ! Some textures have a size greater than " + WV3DConstants.textureSupport.MAX_NATURAL_WIDTH + "x" + WV3DConstants.textureSupport.MAX_NATURAL_HEIGHT + ". This might not be supported by your system and you might experience browser or graphics card crashes.");
                    this.set("_TextureSizeWarningMessageDisplayed", true);
                    this.set("_TextureSizeByPass", confirmed);
                }
                else
                {
                    confirmed = this.get("_TextureSizeByPass");//Keep user choice for large textures
                }
            }
            return confirmed;
        },

        /**
         * Start the asynchronous loading of a texture, if required
         */
        StartAsyncTextureLoad: function (__assets, __matprops, __prop_tex_name, __mat_tex_name, __material_struct, __path, __format, __unpack_alignment)
        {
            if (__matprops["front"].hasOwnProperty(__prop_tex_name) &&
                __matprops["front"][__prop_tex_name] &&
                __assets.hasOwnProperty(__matprops["front"][__prop_tex_name]))
            {
                var texture_node = __assets[__matprops["front"][__prop_tex_name]];
                var remappedAlphaChannel = texture_node["Properties"].hasOwnProperty("remapped_alpha_channel") ? WV3DUtils.ToBool(texture_node["Properties"]["remapped_alpha_channel"]) : false;
                //We try to load the textures, if there is a ref to it
                var onload_tex = function (__texture)
                {
                    this.get("_exec_queue").whenFinished(function ()
                    {
                        this.get("_assetmanager").IncrementTextureLoadCounter();
                        var confirmed = this.CheckTextureSizeByPass(__texture);
                        //If the texture has been successfully loaded we attach the geometries to the textured material
                        if (confirmed && __material_struct)
                        {
                            var scale_x, scale_y;
                            scale_x = scale_y = 1.;
                            if (__material_struct.hasOwnProperty("originalTexWidth"))
                            {
                                scale_x = (__material_struct["originalTexWidth"] - 1) / __texture["image"]["naturalWidth"];
                            }
                            if (__material_struct.hasOwnProperty("originalTexHeight"))
                            {
                                scale_y = (__material_struct["originalTexHeight"] - 1) / __texture["image"]["naturalHeight"];
                            }
                            __texture.repeat.set(scale_x, scale_y);
                            __texture["remappedAlphaChannel"] = remappedAlphaChannel;
                            __material_struct[__mat_tex_name] = __texture;
                            this.trigger("updatetexturedmaterial", {"mat_struct": __material_struct});
                            this.trigger("render");
                        }
                        this.trigger("check_post_scene_load_actions");
                    }, this);
                }.bind(this);
                var onerror = function ()
                {
                    this.get("_exec_queue").whenFinished(function ()
                    {
                        this.get("_assetmanager").IncrementTextureLoadCounter();
                        var errorParams =
                            {
                                "errorCode": 2,
                                "message": "Texture: " + __prop_tex_name
                            };
                        this.trigger("resource_query_error", errorParams);
                    }, this);
                }.bind(this);

                var texpath;
                if (this.get("_assetmanager").get("_assetloader").get("_httprequestmanager").IsAbsoluteURL(texture_node["Properties"]["filepath"]))
                {
                    texpath = texture_node["Properties"]["filepath"];
                }
                else
                {
                    texpath = __path + texture_node["Properties"]["filepath"];
                }

                var queryCallback = function (__url)
                {
                    var loader = new THREE.TextureLoader();
                    //allow cross origin loading
                    loader.crossOrigin = "";
                    var texture = loader.load(__url, function (__texture)
                    {
                        onload_tex(__texture);
                    }, undefined, onerror);
                    texture.name = texpath;
                    texture.format = __format;
                    texture.unpackAlignment = __unpack_alignment;
                    texture.anisotropy = 16;
                    // texture.minFilter = THREE.NearestFilter;
                    // texture.generateMipmaps = true;
                    SelectTextureWrappingMode(texture, texture_node["Properties"], __material_struct);
                }.bind(this);

                this.get("_assetmanager").IncrementTextureQueryCounter(1);
                this.get("_assetmanager").GetURLPreprocessCallback()(texpath, queryCallback);

                //Add reference to this texture url in the cache
                if (!this.get("_renderercache").get("textures")["FromFile"].hasOwnProperty(texpath))
                {
                    this.get("_renderercache").get("textures")["FromFile"][texpath] = null;
                }

                return true;
            }
            else
            {
                return false;
            }
        },

        LoadGlobalLightingTextures: function ()
        {
            if (!this.get("_QueriedGlobalTextures"))
            {
                var path = this.get("_assetmanager").get("_resourcePath") + "/" + this.get("_renderercache").get("textures")[this.get("_renderercache").get("_RenderingQuality")]["path"];

                if (this.get("_renderercache").get("materials")["enableirradiancemap"])
                    this.StartAsyncGlobalTextureLoad("irradiancemap", path);

                if (this.get("_renderercache").get("materials")["enableenvmap"])
                    this.StartAsyncGlobalTextureLoad("envmap", path);

                this.set("_QueriedGlobalTextures", true);
            }
        },

        StartAsyncGlobalTextureLoad: function (__type, __path)
        {
            var onload_global_tex = function (__texture)
            {
                this.get("_assetmanager").IncrementTextureLoadCounter();
                var confirmed = this.CheckTextureSizeByPass(__texture);
                if (confirmed) this.trigger("render");
                this.trigger("check_post_scene_load_actions");
            }.bind(this);
            var onload_standard_tex = function (__texture)
            {
                this.get("_exec_queue").whenFinished(function ()
                {
                    this.get("_renderercache").get("textures")[this.get("_renderercache").get("_RenderingQuality")][__type]["loaded"] = true;
                    this.trigger("updatematerialsgloballighting");
                    onload_global_tex(__texture);
                }, this);
            }.bind(this);
            var onload_env_map_glos_tex = function (__texture)
            {
                this.get("_exec_queue").whenFinished(function ()
                {
                    this.get("_renderercache").get("textures")[this.get("_renderercache").get("_RenderingQuality")]["envmap"]["glossy"]["loaded"] = true;
                    this.trigger("updatematerialsgloballighting");
                    onload_global_tex(__texture);
                }, this);
            }.bind(this);
            var onload_env_map_matte_tex = function (__texture)
            {
                this.get("_exec_queue").whenFinished(function ()
                {
                    this.get("_renderercache").get("textures")[this.get("_renderercache").get("_RenderingQuality")]["envmap"]["matte"]["loaded"] = true;
                    this.trigger("updatematerialsgloballighting");
                    onload_global_tex(__texture);
                }, this);
            }.bind(this);
            var onerror = function ()
            {
                this.get("_exec_queue").whenFinished(function ()
                {
                    this.get("_assetmanager").IncrementTextureLoadCounter();
                    var errorParams =
                        {
                            "errorCode": 2,
                            "message": "Texture: " + __type
                        };
                    this.trigger("resource_query_error", errorParams);
                }, this);
            }.bind(this);

            var preprocess_urls_and_query_cube_texture = function (__files, __type, __subtype, __onLoadCallback)
            {
                var preProcfiles = [];
                var queryCallback = function (__url)
                {
                    preProcfiles.push(__url);
                    if (preProcfiles.length == __files.length)
                    {
                        var loader = new THREE.CubeTextureLoader();
                        var texture = loader.load(preProcfiles, __onLoadCallback, undefined, onerror);
                        texture.format = THREE.RGBFormat;
                        switch (__type)
                        {
                            case "irradiancemap":
                                this.get("_renderercache").get("textures")[this.get("_renderercache").get("_RenderingQuality")][__type]["data"] = texture;
                                break;
                            case "envmap":
                                switch (this.get("_renderercache").get("_RenderingQuality"))
                                {
                                    case "Designer":
                                        this.get("_renderercache").get("textures")[this.get("_renderercache").get("_RenderingQuality")][__type][__subtype]["data"] = texture;
                                        break;
                                    case "Visualizer":
                                        this.get("_renderercache").get("textures")[this.get("_renderercache").get("_RenderingQuality")][__type]["data"][__subtype] = texture;
                                        break;
                                }
                                break;
                        }
                    }
                }.bind(this);

                for (var k = 0, _filesKeys = Object.keys(__files), l = __files.length; k < l; ++k)
                {
                    this.get("_assetmanager").GetURLPreprocessCallback()(__files[_filesKeys[k]], queryCallback);
                }
            }.bind(this);

            switch (__type)
            {
                case "irradiancemap":
                    var cacheTextures = this.get("_renderercache").get("textures")[this.get("_renderercache").get("_RenderingQuality")];
                    var files = GetFileList(__path, cacheTextures[__type]["files"]);
                    preprocess_urls_and_query_cube_texture(files, __type, null, onload_standard_tex);
                    this.get("_assetmanager").IncrementTextureQueryCounter(1);
                    break;

                case "envmap":
                    switch (this.get("_renderercache").get("_RenderingQuality"))
                    {
                        case "Designer":
                            var cacheTextures = this.get("_renderercache").get("textures")[this.get("_renderercache").get("_RenderingQuality")];
                            //Load glossy files and create cube texture
                            var files_glossy = GetFileList(__path, cacheTextures[__type]["glossy"]["files"]);
                            preprocess_urls_and_query_cube_texture(files_glossy, __type, "glossy", onload_env_map_glos_tex);

                            //Load matte files and create cube texture
                            var files_matte = GetFileList(__path, cacheTextures[__type]["matte"]["files"]);
                            preprocess_urls_and_query_cube_texture(files_matte, __type, "matte", onload_env_map_matte_tex);

                            this.get("_assetmanager").IncrementTextureQueryCounter(2);
                            break;

                        case "Visualizer":
                            //Here five cube textures haze levels are to be loaded
                            var cacheTextures = this.get("_renderercache").get("textures")[this.get("_renderercache").get("_RenderingQuality")];
                            cacheTextures[__type]["data"] = {};
                            cacheTextures[__type]["loaded"] = {};

                            var loadIFile = function (__i)
                            {
                                var onload_env_map_haze_i_tex = function (__texture)
                                {
                                    this.get("_exec_queue").whenFinished(function ()
                                    {
                                        cacheTextures[__type]["loaded"][__i] = true;
                                        this.trigger("updatematerialsgloballighting");
                                        onload_global_tex(__texture);
                                    }, this);
                                }.bind(this);
                                var files = GetFileList(__path, cacheTextures[__type]["files"].slice(6 * __i, 6 * (__i + 1)));
                                preprocess_urls_and_query_cube_texture(files, __type, __i, onload_env_map_haze_i_tex);
                            }.bind(this);

                            for (var i = 0; i < 5; ++i)
                            {
                                loadIFile(i);
                            }

                            this.get("_assetmanager").IncrementTextureQueryCounter(5);
                            break;
                    }
                    break;

                case "spectrum":
                    var cacheTextures = this.get("_renderercache").get("textures")[this.get("_renderercache").get("_RenderingQuality")];
                    var texpath = __path + cacheTextures[__type]["file"];
                    var spectrumQueryCallback = function (__url)
                    {
                        var loader = new THREE.TextureLoader();
                        var texture = loader.load(__url, onload_standard_tex, undefined, onerror);
                        texture.format = THREE.RGBFormat;
                        texture.name = texpath;//Should we assign the queried URL instead or keep the internal url ?
                        cacheTextures[__type]["data"] = texture;
                    }.bind(this);

                    this.get("_assetmanager").IncrementTextureQueryCounter(1);
                    this.get("_assetmanager").GetURLPreprocessCallback()(texpath, spectrumQueryCallback);
                    break;

                case "backdrop":

                    this.get("_assetmanager").IncrementTextureQueryCounter(1);
                    this.get("_assetmanager").GetURLPreprocessCallback()(__path, this.BackdropQueryCallback(__path, onload_global_tex));
                    break;

                default:
                    LogMessage.Log("ViewerView Warning: Unhandled global texture type !");

            }
        },

        BackdropQueryCallback: function (__textureURL, __callback)
        {
            var backdropTexCache = this.get("_renderercache").get("textures")["backdrop"];
            return function (__url)
            {
                var onload_backdrop_tex = function (__texture)
                {
                    this.get("_exec_queue").whenFinished(function ()
                    {
                        backdropTexCache["loaded"] = true;
                        if (!backdropTexCache["original_width"]) backdropTexCache["original_width"] = __texture["width"];
                        if (!backdropTexCache["original_height"]) backdropTexCache["original_height"] = __texture["height"];
                        __callback(__texture);
                    }, this);
                }.bind(this);

                var onerror = function (__event)
                {
                    this.get("_exec_queue").whenFinished(function ()
                    {
                        this.get("_assetmanager").IncrementTextureLoadCounter();
                        var errorParams =
                            {
                                "errorCode": 2,
                                "message": "Texture: " + __url
                            };
                        this.trigger("resource_query_error", errorParams);
                    }, this);
                }.bind(this);

                var loader = new THREE.TextureLoader();
                //allow cross origin loading
                loader.crossOrigin = "";
                var texture = loader.load(__textureURL, onload_backdrop_tex, undefined, onerror);
                texture.format = THREE.RGBFormat;
                texture.name = __textureURL;//Should we assign the queried URL instead or keep the internal url ?
                backdropTexCache["enabled"] = true;
                backdropTexCache["data"] = texture;
            }.bind(this);
        },

        LoadBackdrop: function (__root_node, __scene_id)
        {
            if (__root_node["Properties"].hasOwnProperty("backdrop"))
            {
                var texIndex = __root_node["Properties"]["backdrop"]["image_tex"];
                var texProps = __root_node["Assets"][texIndex]["Properties"];
                var path;
                if (this.get("_assetmanager").get("_assetloader").get("_httprequestmanager").IsAbsoluteURL(texProps["filepath"]))
                {
                    path = texProps["filepath"];
                }
                else
                {
                    path = __scene_id.substring(0, __scene_id.lastIndexOf("/") + 1) + texProps["filepath"];
                }
                this.StartAsyncGlobalTextureLoad("backdrop", path);

                this.get("_renderercache").get("textures")["backdrop"]["original_height"] = texProps["original_height"];
                this.get("_renderercache").get("textures")["backdrop"]["original_width"] = texProps["original_width"];
            }
        },

        SetBackdropTexture: function (__textureURL, __callback, __redraw)
        {
            var redrawCallback = function (args)
            {
                if (__redraw === undefined || __redraw) this.trigger("render");
                __callback(arguments);
            }.bind(this);

            var backdropTexCache = this.get("_renderercache").get("textures")["backdrop"];
            backdropTexCache["original_width"] = null;
            backdropTexCache["original_height"] = null;
            this.get("_assetmanager").GetURLPreprocessCallback()(__textureURL, this.BackdropQueryCallback(__textureURL, redrawCallback));
        },

        /**
         * Reload all textures
         */
        ReloadAllTextures: function (__callback, __redraw)
        {
            //Reload each texture
            var loadcounter = 0;
            for (var k = 0, _texts = Object.keys(this.get("_renderercache").get("textures")["FromFile"]), l = _texts.length; k < l; ++k)
            {
                var text_url = _texts[k];
                var cb = function ()
                {
                    ++loadcounter;
                    if (loadcounter == l)
                    {
                        if (__redraw === undefined || __redraw) this.trigger("render");
                        if (typeof(__callback) == "function")
                        {
                            __callback();
                        }
                    }
                }.bind(this);
                this.ReloadTexture(text_url, text_url, cb, true, false);
            }

        },

        /**
         * Reload a texture
         */
        ReloadTexture: function (__textureURL, __newURL, __callback, __keep_old_path, __redraw)
        {
            var onload_tex = function (__textureURL, __newTexture)
            {
                var materials = this.get("_renderercache").get("materials");
                for (var k = 0, _mats = Object.keys(materials["materiallist"]), l2 = _mats.length; k < l2; ++k)
                {
                    var mat = _mats[k];
                    var mat_struct_front = materials["materiallist"][mat]["front"];
                    var mat_struct_back = materials["materiallist"][mat]["back"];

                    var params_front =
                        {
                            "mat_struct": mat_struct_front,
                            "texture_URL": __textureURL,
                            "new_texture": __newTexture
                        };
                    var params_back =
                        {
                            "mat_struct": mat_struct_front,
                            "texture_URL": __textureURL,
                            "new_texture": __newTexture
                        };

                    this.trigger("updatematerialtexture", params_front);
                    this.trigger("updatematerialtexture", params_back);
                }

                if (__redraw === undefined || __redraw) this.trigger("render");

                if (typeof(__callback) == "function")
                {
                    __callback();
                }
            }.bind(this);

            var onerror = function (__event)
            {
                this.get("_exec_queue").whenFinished(function ()
                {
                    this.get("_assetmanager").IncrementTextureLoadCounter();
                    var errorParams =
                        {
                            "errorCode": 2,
                            "message": "Texture: " + __newURL
                        };
                    this.trigger("resource_query_error", errorParams);
                }, this);
            }.bind(this);

            var queryCallback = function (__url)
            {
                var loader = new THREE.TextureLoader();
                //allow cross origin texture loading
                loader.crossOrigin = "";
                var texture = loader.load(__url, function (__texture)
                {
                    onload_tex(__textureURL, __texture);
                }, undefined, onerror);
                if (__keep_old_path || __keep_old_path == undefined)
                {
                    //Should we keep the old texture path as reference ?
                    texture.name = __textureURL;
                }
                else
                {
                    //Should the new name (path) replace the old one ?
                    texture.name = __newURL;
                }
            }.bind(this);

            //Adding "?t="+((new Date()).getTime()) allows to bypass the browser
            //cache since it is very likely that the file has been updated on the disk
            this.get("_assetmanager").GetURLPreprocessCallback()(__newURL + "?t=" + ((new Date()).getTime()), queryCallback);
        },

        /**
         * Creates or update the annotation texture containing the line segment
         * return the size, in pixels, of the texture side and the maximum number
         * of lines per cell as an object
         * {"textureSize":squareSideLength, "maxLinesPerCell":maxLinesPerCell};
         */
        UpdateAnnotationsTexture: function (__mat_struct, __segmentList, __accelGridSize)
        {
            var accelGrid = [];
            var lineReferences = 0;
            var maxLinesPerCell = 0;
            for (var i = 0; i < __accelGridSize; ++i)
            {
                var left = i / __accelGridSize;
                var right = (i + 1) / __accelGridSize;
                for (var j = 0; j < __accelGridSize; ++j)
                {
                    var bottom = j / __accelGridSize;
                    var top = (j + 1) / __accelGridSize;
                    var cellLines = [];
                    for (var s = 0; s < __segmentList.length; s++)
                    {
                        var cell = {"left": left, "right": right, "bottom": bottom, "top": top};
                        if (WV3DMath.IntersectCellSegment(cell, __segmentList[s]))
                        {
                            cellLines.push({"lineIdx": s, "segment": __segmentList[s]});
                        }
                    }
                    accelGrid.push(cellLines);
                    //lineReferences counter stays unchanged if we only have one single
                    //referenced line, to avoid fetching twice in shader
                    if (cellLines.length > 1)
                    {
                        lineReferences += cellLines.length;
                        if (cellLines.length > maxLinesPerCell)
                        {
                            //Update maximum line number per cell. This avoids
                            //looping over many unnecessary elements in the shader
                            maxLinesPerCell = cellLines.length;
                        }
                    }
                }
            }

            /* First part of the texture is for the acceleration grid structure
             __accelGridSize * __accelGridSize, each cell contains 2 floats:
             - a number of lines
             - the starting index of its first line index */
            /* Second part contains the list of indexes of the cell lines:
             - 1 float per index
            /* Third part of the texture contains the lines themselves:
             - 4 floats per line */

            var squareSideLength = WV3DMath.nextHighestPowerOfTwo(Math.ceil(Math.sqrt(__accelGridSize * __accelGridSize + lineReferences + __segmentList.length)));
            var newSize = squareSideLength * squareSideLength;
            var data;
            if (__mat_struct["annotationTextureSize"] !== newSize)
            {
                data = new Float32Array(4 * newSize);
            }
            else
            {
                // Optimization recycles texture if its size did not change
                data = __mat_struct["annotationsTexture"].image.data;
            }

            //Fill in the data
            var currentCellRefOffset = 0;
            for (var i = 0; i < accelGrid.length; i++)
            {
                //Fill regular grid
                var stride = i * 4;
                var cellLines = accelGrid[i];
                //Number of lines in this cell
                data[stride] = cellLines.length;
                if (cellLines.length > 1)
                {
                    //Address of line index map for the current cell
                    //as currentCellRefOffset can be quite a large number, some mobile devices
                    //do not have enough precision to represent it (half float textures)
                    //to fix it, we encode that number into three different channels
                    //e.g., an analogous case would be with powers of 10 (instead of 256):
                    //if currentCellRefOffset == "123456", we would end up with
                    //channel1 == "12", channel2 == "34", channel3 == "56"
                    var channel1 = Math.floor(currentCellRefOffset / 65535);
                    var channel2 = Math.floor((currentCellRefOffset - channel1 * 65535) / 255);
                    var channel3 = Math.floor((currentCellRefOffset - channel1 * 65535 - channel2 * 255));

                    data[stride + 1] = channel1;
                    data[stride + 2] = channel2;
                    data[stride + 3] = channel3;

                    //Fill indexes list
                    for (var j = 0; j < cellLines.length; j++)
                    {
                        var indexStride = (accelGrid.length + currentCellRefOffset + j) * 4;
                        data[indexStride] = cellLines[j].lineIdx;
                        data[indexStride + 1] = -1;//unused
                        data[indexStride + 2] = -1;//unused
                        data[indexStride + 3] = -1;//unused
                    }
                    currentCellRefOffset += cellLines.length;
                }
                else if (cellLines.length == 1)
                {
                    //IMPROVEME: this branch do not use the index decomposition used for the other branch of this if/else block
                    //If some artifacts start to show on mobile devices when the number of annotations becomes high, it may be due
                    //to the same limitation and we may have to also encode/decode it in the same manner (see
                    //ShaderUtils:AnnotationDrawing chunck)

                    //Direct address of the only line for the current cell
                    //No need to create any index, saving an extra fetch
                    data[stride + 1] = cellLines[0].lineIdx;
                    data[stride + 2] = -1;//unused
                    data[stride + 3] = -1;//unused
                    // console.warn("Direct reference index: "+data[ stride + 1 ]);
                }
                else
                {
                    //Nothing, at all...
                    data[stride + 1] = -1;//unused
                    data[stride + 2] = -1;//unused
                    data[stride + 3] = -1;//unused
                }
            }
            //Fill line segments coordinates
            for (var i = 0; i < __segmentList.length; i++)
            {

                var stride = (accelGrid.length + lineReferences + i) * 4;
                data[stride] = __segmentList[i]["startU"];
                data[stride + 1] = __segmentList[i]["startV"];
                data[stride + 2] = __segmentList[i]["endU"];
                data[stride + 3] = __segmentList[i]["endV"];
            }

            if (__mat_struct["annotationTextureSize"] !== newSize)
            {
                //release existing texture if any, then use the new one
                if (__mat_struct["annotationsTexture"]) __mat_struct["annotationsTexture"].dispose();

                __mat_struct["annotationsTexture"] = new THREE.DataTexture(data, squareSideLength, squareSideLength, THREE.RGBAFormat,
                    THREE.FloatType, THREE.UVMapping, THREE.RepeatWrapping, THREE.RepeatWrapping,
                    THREE.NearestFilter, THREE.NearestFilter);
                __mat_struct["annotationTextureSize"] = newSize;
            }

            __mat_struct["annotationsTexture"].needsUpdate = true;

            return {"textureSize": squareSideLength, "maxLinesPerCell": maxLinesPerCell, "linesStartOffset": accelGrid.length + lineReferences};
        }


    });

});
