/**
 * @authors DOGO, PHST, RAVO, SVN, ADCL
 * @module WebViewer3D
 * @copyright   2014 Esko Stonecube Ltd. All rights reserved.
 */

define([
	"backbone",
	"underscore",
	"three",
	"@esko/3d-web-viewer/modules/threeEXT",
	"@esko/3d-web-viewer/modules/LogMessage",
	"@esko/3d-web-viewer/modules/Notifier",
	"@esko/3d-web-viewer/modules/Math",
	"@esko/3d-web-viewer/modules/SystemChecker",
	"@esko/3d-web-viewer/modules/ObjectBuilder",
	"@esko/3d-web-viewer/modules/TextureManager",
	"@esko/3d-web-viewer/modules/MaterialManager",
	"@esko/3d-web-viewer/modules/SceneManager",
	"@esko/3d-web-viewer/modules/RendererCache",
	"@esko/3d-web-viewer/modules/Utils",
	"@esko/3d-web-viewer/modules/FPSCounter",
	"@esko/3d-web-viewer/modules/Timers",
	"@esko/3d-web-viewer/modules/ExecutionQueue"
], function(
	bb,
	_,
	THREE,
	WV3DThreeEXT,
	LogMessage,
	WV3DNotifier,
	WV3DMath,
	WV3DSystemChecker,
	WV3DObjectBuilder,
	WV3DTextureManager,
	WV3DMaterialManager,
	WV3DSceneManager,
	WV3DRendererCache,
	WV3DUtils,
	WV3DFPSCounter,
	WV3DTimers,
	ExecutionQueue
) {
	"use strict";

	/**
	 * @alias module:WebViewer3D
	 * @class WV3DViewerView
	 * This class is responsible for detecting model changes (new json resource being added) via AddedResource
	 * and triggering scene building via scenemanager. This class is also responsible for update scene setting
	 * such as camera position, highlights, background and rendering the scene (IMPROVEME: move the rendering part
	 * into a separate class - see below)
	 */

	/**
	 * This method calculates the smallest distance between a camera and a given (bounding) box along that camera's view direction.
	 */
	const GetClosestCameraDistanceToBox = function(__box, __camera_azimuth, __camera_elevation, __epsilon) {
		const view_direction = new THREE.Vector3(0, 0, 0);

		view_direction.x = Math.cos(WV3DMath.radians(__camera_elevation)) * Math.cos(WV3DMath.radians(__camera_azimuth));
		view_direction.z = Math.cos(WV3DMath.radians(__camera_elevation)) * Math.sin(WV3DMath.radians(__camera_azimuth));
		view_direction.y = Math.sin(WV3DMath.radians(__camera_elevation));

		const viewing_ray = new THREE.Ray(__box.getCenter(), view_direction);
		const box_intersection = viewing_ray.intersectBox(__box);

		// Also add a (tiny) buffer region between the minimal camera position and the scene bounding box.
		// We use 2 * epsilon to avoid clipping since the near plane is already a distance of at least one epsilon in front of the camera
		return (
			box_intersection
				.clone()
				.sub(__box.getCenter())
				.length() +
			2 * __epsilon
		);
	};

	/**
	 * This method calculates the new camera radius given the current scale, preferred camera distance and several constraints that prevent the camera
	 * from ever entering the scene. This radius is determined by a unit distance and a scaling factor (radius = unit distance for scale = 1).
	 * When zooming in/out, the scale factor is increased/decreased respectively and from this the corresponding camera radius is derived.
	 */
	const ComputeNewCameraRadiusAndScale = function(
		__asset_camera,
		__min_radius,
		__sphere_radius,
		__unit_distance,
		__override_camera_distance,
		__scale_min,
		__scale_max
	) {
		let radius;

		if (__override_camera_distance) {
			// Clamp the distance so we won't obtain a negative divisor when calculating the scale below
			const min_distance = __asset_camera.cam_type == "PERSPECTIVE" ? __min_radius : __sphere_radius;
			const distance = Math.max(__override_camera_distance, min_distance);
			// Calculate, clamp and set the scale given the desired camera distance
			__asset_camera.scale = Math.min(Math.max(__unit_distance / (distance - __sphere_radius), __scale_min), __scale_max);
			// Now calculate the radius given the (possibly clamped) scale factor
			radius = __sphere_radius + __unit_distance / __asset_camera.scale;
		} else {
			radius = __sphere_radius + __unit_distance / __asset_camera.scale;
			// Push the camera back a bit if we are getting too close to the scene bounding box
			if (radius < __min_radius) {
				radius = __min_radius;
				// Keep the scale factor in sync with the new radius
				__asset_camera.scale = Math.min(Math.max(__unit_distance / (radius - __sphere_radius), __scale_min), __scale_max);
			}
		}
		return radius;
	};

	/**
	 * This method calculates the desired near plane given a camera position, its viewing direction and the scene bounding box.
	 * For this we first determine the point on the bounding box that is closest to the projection plane and then define the near plane accordingly.
	 */
	const GetNearPlaneDistance = function(__camera_position, __view_direction, __bbox, __epsilon) {
		// Note: this function will be used for both the scene and the scene focus bounding box respectively

		// First find the bbox point (vertice) closest to the camera projection plane
		const near_plane = new THREE.Plane();
		near_plane.setFromNormalAndCoplanarPoint(__view_direction.normalize(), __camera_position);

		let closest_point = new THREE.Vector3(__bbox.min.x, __bbox.min.y, __bbox.min.z);
		const bbox_vertices = [
			new THREE.Vector3(__bbox.max.x, __bbox.min.y, __bbox.min.z),
			new THREE.Vector3(__bbox.min.x, __bbox.max.y, __bbox.min.z),
			new THREE.Vector3(__bbox.min.x, __bbox.min.y, __bbox.max.z),
			new THREE.Vector3(__bbox.max.x, __bbox.max.y, __bbox.min.z),
			new THREE.Vector3(__bbox.min.x, __bbox.max.y, __bbox.max.z),
			new THREE.Vector3(__bbox.max.x, __bbox.min.y, __bbox.max.z),
			new THREE.Vector3(__bbox.max.x, __bbox.max.y, __bbox.max.z)
		];

		for (let i = 0; i < bbox_vertices.length; i++) {
			const vertex = bbox_vertices[i];
			// Beware: the distance between the plane and point is signed, so make sure the view direction points towards the scene center
			if (near_plane.distanceToPoint(vertex) < near_plane.distanceToPoint(closest_point)) closest_point = vertex;
		}

		return near_plane.distanceToPoint(closest_point) - __epsilon;
	};

	/**
	 * Simple helper method to derive the cartesian coordinates of the current camera position given the following properties
	 * @param __scene_center: the cartesian coordinates of the center of the current scene
	 * @param __radius: the distance between the scene center and the camera position
	 * @param __azimuth: the current azimuth angle of the camera
	 * @param __elevation: the current elevation angle of the camera
	 */
	const GetCameraCoordinates = function(__scene_center, __radius, __azimuth, __elevation) {
		const coordinates = new THREE.Vector3(__scene_center.x, __scene_center.y, __scene_center.z);

		coordinates.x += __radius * Math.cos(WV3DMath.radians(__elevation)) * Math.cos(WV3DMath.radians(__azimuth));
		coordinates.z -= __radius * Math.cos(WV3DMath.radians(__elevation)) * Math.sin(WV3DMath.radians(__azimuth));
		coordinates.y += __radius * Math.sin(WV3DMath.radians(__elevation));

		return coordinates;
	};

	/**
	 * This method calculates the direction of the camera left
	 */
	const CalculateLeftCamVecDir = function(__radius, __azimuth, __elevation) {
		const leftcamvecdir = new THREE.Vector3(0, 0, 0);

		leftcamvecdir.x += __radius * Math.cos(WV3DMath.radians(__elevation)) * Math.cos(WV3DMath.radians(__azimuth) - Math.PI / 2);
		leftcamvecdir.z -= __radius * Math.cos(WV3DMath.radians(__elevation)) * Math.sin(WV3DMath.radians(__azimuth) - Math.PI / 2);
		leftcamvecdir.normalize();

		return leftcamvecdir;
	};

	return bb.Model.extend({
		convertScaleToDistance: function(scale) {
			const box_size = this.get("_scenemanager")
				.get("_3JSsceneBBoxFocus")
				.getSize();
			const max_box_length = Math.sqrt(box_size.x * box_size.x + box_size.y * box_size.y + box_size.z * box_size.z);
			// The 'old' scale values were calculated using the bounding sphere instead of the inner sphere
			const bounding_sphere_radius = max_box_length * 0.5;

			// The unit distance equals the 'minimum camera distance' according to the Field Of View
			const fov = this.get('_renderercache').get('cameras')['defaultcamera'].fieldOfView;
        	const aspectratio = Math.max(this.get('_context').clientWidth / this.get('_context').clientHeight, 1);
			const min_fov = Math.min(fov,aspectratio*fov);
			const unit_distance = Math.abs(max_box_length/(2*Math.tan(WV3DMath.radians(min_fov)/2)));

			return unit_distance / scale + bounding_sphere_radius;
		},

		initialize: function(__context, __WebViewerModel, __clientUIAddElementCallback) {
			LogMessage.Log("Creating WV3DViewerView Object");

			this.set("_context", __context);
			this.set("_model", __WebViewerModel);

			const threejsproperties = {
				alpha: true,
				antialias: WV3DSystemChecker.get("_glantialias"),
				preserveDrawingBuffer: true //this allows to take screenshots using toDataURL("image/jpeg")
				// logarithmicDepthBuffer: true//Warning: this only works with a perspective camera, otherwise z-test may fail
			};
			/** Extend threejs before using it*/
			new WV3DThreeEXT();
			this.set("_3JSrenderer", new THREE.WebGLRenderer(threejsproperties));
			this.get("_3JSrenderer").sortObjects = false; //We are responsible for sorting surfaces (especially for coincident surfaces)

			//IMPROVEME: We could consider using singletons for some of these attributes
			this.set("_renderercache", new WV3DRendererCache());

			//We create the execution queue only once and share it among all components
			//FIXME: see other 'fixme' below. If we keep current replacement, we should then remove the exec_queue from the following lines
			this.set("_exec_queue", new ExecutionQueue(/*set to true for debug only*/ false, false, this.NotifySceneBuildProgress.bind(this)));

			this.set(
				"_texturemanager",
				new WV3DTextureManager(this.get("_renderercache"), this.get("_model").get("_assetmanager"), this.get("_exec_queue"))
			);
			this.set(
				"_materialmanager",
				new WV3DMaterialManager(
					this.get("_renderercache"),
					this.get("_model").get("_assetmanager"),
					this.get("_texturemanager"),
					this.get("_exec_queue")
				)
			);
			this.set(
				"_notifier",
				new WV3DNotifier(
					this.get("_model"),
					this.get("_materialmanager"),
					this.get("_texturemanager"),
					this.get("_exec_queue"),
					__clientUIAddElementCallback
				)
			);
			this.set("_objectbuilder", new WV3DObjectBuilder(this.get("_exec_queue")));
			this.set(
				"_scenemanager",
				new WV3DSceneManager(
					this.get("_renderercache"),
					this.get("_model"),
					this.get("_materialmanager"),
					this.get("_objectbuilder"),
					this.get("_notifier"),
					this.get("_exec_queue")
				)
			);

			this.set("_current_fps", 0);
			//We want at least an interactive framerate for scenes with transparency
			this.set("_target_fps_min", 25);
			this.set("_target_fps_max", 60);
			//We want at least transparency to be "fine" during interaction
			this.set("_nb_max_peels", 7);
			this.set("_nb_min_peels", 2);

			this.set("_reflection_config", "JSON");
			this.set("_enableInteractionOptimization", false);

			this.set("_lighting_contrast", 0.5);

			this.set("_optimised_view", false);
			this.set("_optimised_view_border", 0);

			this.set("_scale_min", 0.1);
			this.set("_scale_max", 20);

			// This epsilon value determines the minimal buffer distance between the camera and the scene bounding box
			this.set("_epsilon_camera", 0.1);

			this.set("_obj_in_frustum_prev", true);
			this.set("_obj_in_frustum", true);

			/** subscribe to any change in the model */
			this.listenTo(this.get("_model"), "added_scene_resource", this.AddedResource);
			this.listenTo(this.get("_scenemanager"), "update_lighting_contrast", this.UpdateLightingContrast);
			this.listenTo(this.get("_scenemanager"), "render", this.Render);
			this.listenTo(this.get("_materialmanager"), "render", this.Render);
			this.listenTo(this.get("_texturemanager"), "render", this.Render);
			this.listenTo(this.get("_notifier"), "render", this.Render);
			this.listenTo(this.get("_notifier"), "post_scene_load_actions", this.PostSceneLoadActions);

			this.set("_resizecallback", this.Resize.bind(this));
			window.addEventListener("resize", this.get("_resizecallback"), false);
			this.Resize();

			this.set("_backgroundColor", new THREE.Color(0.37, 0.37, 0.37));
			this.set("_backgroundAlpha", 1);

			this.AppendRendererToDocument();
		},

		NotifySceneBuildProgress: function(__geometry_percent, __texture__percent, __error__code) {
			this.get("_notifier").NotifySceneBuildProgress(__geometry_percent, __texture__percent, __error__code);
		},

		/** Append the Renderer to the target DOM element */
		AppendRendererToDocument: function() {
			if (typeof this.get("_context").appendChild === "function") {
				this.get("_context").appendChild(this.get("_3JSrenderer").domElement);
			}
		},

		DisposeThreeJSContext: function() {
			this.get("_3JSrenderer").dispose();
		},

		/**
		 * Release the scene
		 */
		ReleaseScene: function() {
			this.get("_scenemanager").ReleaseScene();
		},

		/**
		 * Reset the scene and creates a new one
		 */
		ResetScene: function(__coplanarityEpsilon, __boundaryhardedges) {
			//Sometimes, some events stayed stuck in the execution queue
			//especially if we load another scene without waiting for the
			//current one to be loaded. Therefore, we simply, clear it.
			// this.get('_exec_queue').Reset(false,false,this.NotifySceneBuildProgress.bind(this));
			// this.get('_exec_queue').clear();

			//FIXME: resetting the execution queue does not work when loading consecutively two scenes,
			//therefore, as a workaround, we create and replace the existing execution queue
			//We create the execution queue only once and share it among all components
			this.set("_exec_queue", new ExecutionQueue(/*set to true for debug only*/ false, false, this.NotifySceneBuildProgress.bind(this)));
			this.get("_notifier").set("_exec_queue", this.get("_exec_queue"));
			this.get("_scenemanager").set("_exec_queue", this.get("_exec_queue"));
			this.get("_scenemanager")
				.get("_scenebuilder")
				.set("_exec_queue", this.get("_exec_queue"));
			this.get("_texturemanager").set("_exec_queue", this.get("_exec_queue"));
			this.get("_materialmanager").set("_exec_queue", this.get("_exec_queue"));
			this.get("_objectbuilder").set("_exec_queue", this.get("_exec_queue"));

			this.get("_scenemanager").ResetScene(__coplanarityEpsilon, __boundaryhardedges);
			this.ResetScreenQuads();

			this.set("_InitialisedSceneSettings", false);
			this.set("_SceneLoaded", false);

			this.set("_annotationsEnabled", false);

			this.set("_current_camera_distance", 0);
		},

		ResetScreenQuads: function() {
			this.set("_screenQuadSceneCopyShader", null);
			this.set("_screenQuadSceneColorBlend", null);
			this.set("_screenQuadSceneBackdropShader", null);
			this.set("_screenQuadSceneReflectionShader", null);
		},

		/**
		 * This function is triggered after all scene elements have been loaded
		 */
		PostSceneLoadActions: function() {
			this.set("_SceneLoaded", true);
			this.get("_objectbuilder").ClearBuildMaps();
			this.get("_scenemanager").ClearInstanceArrays();
			this.get("_scenemanager").CreateLabels();
			this.get("_materialmanager").UpdateReflectionUniforms();

			this.CameraUpdate(true);
		},

		/** TRIGGERED EVENT (model::'added_scene_resource'): Handle changes to the model */
		AddedResource: function() {
			this.SetUpScene();

			this.get("_notifier").NotifyAddedResource();
		},

		/** Build the scene using data from the asset manager */
		SetUpScene: function() {
			//We add a default empty callback to 3DobjectHandlerExecutionQueue to avoid the whenFinished callbacks
			//to be triggered as soon as callback is added to the execution queue (LoadGlobalLightingTextures,
			//AsyncTextureLoad, SetUpScene, etc.)
			this.get("_exec_queue").run(function() {}, this);

			//Initialize scene quality, backdrop and reflection properties
			this.get("_exec_queue").run(function() {
				const cache = this.get("_model")
					.get("_assetmanager")
					.get("_cache");
				const root_node_unique_id = this.get("_model").get("SceneId");
				const root_node = cache["rootnodes"][root_node_unique_id];
				this.InitializeSceneSettings(root_node);
			}, this);

			//Load environment and irradiance maps
			this.get("_exec_queue").run(function() {
				this.LoadGlobalLightingTextures();
			}, this.get("_texturemanager"));

			this.get("_scenemanager").SetUpScene();

			this.get("_exec_queue").whenFinished(function() {
				//Prepare reflections (rely on bounding box)
				this.SetupReflections();

				//Ajust camera to scene (rely on bounding box and reflections)
				this.CameraUpdate(false);
				//The camera position needs to be set before computing the distance to scene center
				this.set(
					"_InitialDistanceToSceneCenter",
					this.get("_scenemanager")
						.get("_3JSsceneBBox")
						.getCenter()
						.clone()
						.sub(this.get("_3JScamera").position)
						.length()
				);
			}, this);
		},

		ReleaseRenderTarget: function(__renderTarget) {
			if (__renderTarget) __renderTarget.dispose();
		},

		RemoveListeners: function() {
			window.removeEventListener("resize", this.get("_resizecallback"), false);
		},

		ReleaseAllRenderTargets: function() {
			this.ReleaseRenderTarget(this.get("_OpaqueDepthRT"));
			this.ReleaseRenderTarget(this.get("_D1RT"));
			this.ReleaseRenderTarget(this.get("_D2RT"));
			this.ReleaseRenderTarget(this.get("_ColorStackRT"));
			this.ReleaseRenderTarget(this.get("_PeelColorRT"));
			this.ReleaseRenderTarget(this.get("_TransparencyStackRT"));
			this.ReleaseRenderTarget(this.get("_OpaqueBackgroundMaskRT"));
			this.ReleaseRenderTarget(this.get("_MirrorOpaqueDepthRT"));
			this.ReleaseRenderTarget(this.get("_MirrorColorStackRT"));
			this.ReleaseRenderTarget(this.get("_MirrorColorStackRT2"));
			this.ReleaseRenderTarget(this.get("_MirrorPeelColorRT"));
			this.ReleaseRenderTarget(this.get("_MirrorTransparencyStackRT"));
			this.ReleaseRenderTarget(this.get("_MirrorOpaqueBackgroundMaskRT"));
		},

		InitializeRenderTargets: function(__overrideWidth, __overrideHeight) {
			//Initialize render targets for depth peeling
			const width = __overrideWidth ? __overrideWidth : this.get("_context").clientWidth;
			const height = __overrideHeight ? __overrideHeight : this.get("_context").clientHeight;
			// width = WV3DMath.isPowerOfTwo(width)?width:WV3DMath.nextHighestPowerOfTwo(width);
			// height = WV3DMath.isPowerOfTwo(height)?height:WV3DMath.nextHighestPowerOfTwo(height);
			let depthTargetsFormat, colorTargetsFormat;

			this.ReleaseAllRenderTargets();

			//Depth encoding in the depth shaders allows use of RGBA textures instead
			//of floating point textures to support many mobile devices (except Ipads
			//which are apparently not very good at anything... :/)
			if (WV3DSystemChecker.get("_fullSupportFloatingPointTextures")) {
				depthTargetsFormat = {
					minFilter: THREE.LinearFilter,
					format: THREE.RGBAFormat,
					type: THREE.FloatType
				};
				colorTargetsFormat = {
					minFilter: THREE.LinearFilter,
					format: THREE.RGBAFormat
				};

				this.set("_OpaqueDepthRT", new THREE.WebGLRenderTarget(width, height, depthTargetsFormat));
				this.set("_D1RT", new THREE.WebGLRenderTarget(width, height, depthTargetsFormat));
				this.set("_D2RT", new THREE.WebGLRenderTarget(width, height, depthTargetsFormat));
				this.set("_ColorStackRT", new THREE.WebGLRenderTarget(width, height, colorTargetsFormat));
				this.set("_PeelColorRT", new THREE.WebGLRenderTarget(width, height, colorTargetsFormat));
				this.set("_TransparencyStackRT", new THREE.WebGLRenderTarget(width, height, colorTargetsFormat));
				this.set("_OpaqueBackgroundMaskRT", new THREE.WebGLRenderTarget(width, height, colorTargetsFormat));
				this.set("_MirrorOpaqueDepthRT", new THREE.WebGLRenderTarget(width, height, depthTargetsFormat));
				this.set("_MirrorColorStackRT", new THREE.WebGLRenderTarget(width, height, colorTargetsFormat));
				this.set("_MirrorColorStackRT2", new THREE.WebGLRenderTarget(width, height, colorTargetsFormat));
				this.set("_MirrorPeelColorRT", new THREE.WebGLRenderTarget(width, height, colorTargetsFormat));
				this.set("_MirrorTransparencyStackRT", new THREE.WebGLRenderTarget(width, height, colorTargetsFormat));
				this.set("_MirrorOpaqueBackgroundMaskRT", new THREE.WebGLRenderTarget(width, height, colorTargetsFormat));
			} else if (WV3DSystemChecker.get("_fullSupportHFloatingPointTextures")) {
				depthTargetsFormat = {
					minFilter: THREE.LinearFilter,
					magFilter: THREE.NearestFilter,
					format: THREE.RGBAFormat,
					type: THREE.HalfFloatType
				};
				colorTargetsFormat = {
					minFilter: THREE.LinearFilter,
					magFilter: THREE.NearestFilter,
					format: THREE.RGBAFormat
				};

				this.set("_OpaqueDepthRT", new THREE.WebGLRenderTarget(width, height, depthTargetsFormat));
				this.set("_D1RT", new THREE.WebGLRenderTarget(width, height, depthTargetsFormat));
				this.set("_D2RT", new THREE.WebGLRenderTarget(width, height, depthTargetsFormat));
				this.set("_ColorStackRT", new THREE.WebGLRenderTarget(width, height, colorTargetsFormat));
				this.set("_PeelColorRT", new THREE.WebGLRenderTarget(width, height, colorTargetsFormat));
				this.set("_TransparencyStackRT", new THREE.WebGLRenderTarget(width, height, colorTargetsFormat));
				this.set("_OpaqueBackgroundMaskRT", new THREE.WebGLRenderTarget(width, height, colorTargetsFormat));
				this.set("_MirrorOpaqueDepthRT", new THREE.WebGLRenderTarget(width, height, depthTargetsFormat));
				this.set("_MirrorColorStackRT", new THREE.WebGLRenderTarget(width, height, colorTargetsFormat));
				this.set("_MirrorColorStackRT2", new THREE.WebGLRenderTarget(width, height, colorTargetsFormat));
				this.set("_MirrorPeelColorRT", new THREE.WebGLRenderTarget(width, height, colorTargetsFormat));
				this.set("_MirrorTransparencyStackRT", new THREE.WebGLRenderTarget(width, height, colorTargetsFormat));
				this.set("_MirrorOpaqueBackgroundMaskRT", new THREE.WebGLRenderTarget(width, height, colorTargetsFormat));
			} else {
				colorTargetsFormat = {
					minFilter: THREE.LinearFilter,
					magFilter: THREE.NearestFilter,
					format: THREE.RGBAFormat
				};
				this.set("_ColorStackRT", new THREE.WebGLRenderTarget(width, height, colorTargetsFormat));
				this.set("_PeelColorRT", new THREE.WebGLRenderTarget(width, height, colorTargetsFormat));
			}
		},

		InitializeEnvironmentSettings: function(__root_node) {
			if (
				__root_node["Properties"].hasOwnProperty("rendering_quality") &&
				(WV3DSystemChecker.get("_fullSupportFloatingPointTextures") || WV3DSystemChecker.get("_fullSupportHFloatingPointTextures"))
			) {
				this.get("_renderercache").set("_RenderingQuality", __root_node["Properties"]["rendering_quality"]);
				if (__root_node["Properties"].hasOwnProperty("environment_name")) {
					this.get("_renderercache").set("_EnvironmentName", __root_node["Properties"]["environment_name"]);
					this.get("_renderercache").get("textures")["Visualizer"]["path"] =
						this.get("_renderercache").get("textures")["Visualizer"]["path"] + __root_node["Properties"]["environment_name"] + "/";
				}
			}
		},

		InitializeCameraSettings: function(__root_node) {
			if (__root_node["Properties"].hasOwnProperty("view_settings")) {
				const view_settings = __root_node["Properties"]["view_settings"];
				this.get("_renderercache").get("view_settings")["mode"] = parseInt(view_settings["view_mode"]);
				switch (this.get("_renderercache").get("view_settings")["mode"]) {
					default:
						LogMessage.Log("View mode not implemented yet. Camera view settings will not be loaded.");
						break;

					// case 0:
					// ...
					// break;

					case 1:
						const camera_preset = __root_node["Assets"][view_settings["camera_asset_ref"]];
						if (camera_preset["AssetType"] !== "Camera") {
							LogMessage.Log("Referenced camera asset not having proper type. Are you sure this camera asset id is uniquely named.");
						} else {
							//ROTATION
							//Rotation extraction inspired from http://www.euclideanspace.com/maths/geometry/rotations/conversions/matrixToEuler/index.htm
							/*  [m00 m01 m02]
                            [m10 m11 m12]
                            [m20 m21 m22]   */
							const camera_props = camera_preset["Properties"];
							const orientation_matrix = camera_props["azim_elev_cam_xform"]["orientation_matrix"];
							const m = {
								m00: parseFloat(orientation_matrix[0]),
								m01: parseFloat(orientation_matrix[1]),
								m02: parseFloat(orientation_matrix[2]),
								m10: parseFloat(orientation_matrix[3]),
								m11: parseFloat(orientation_matrix[4]),
								m12: parseFloat(orientation_matrix[5]),
								m20: parseFloat(orientation_matrix[6]),
								m21: parseFloat(orientation_matrix[7]),
								m22: parseFloat(orientation_matrix[8])
							};

							const camera_transform = new THREE.Matrix4();
							camera_transform.set(
								parseFloat(orientation_matrix[0]),
								parseFloat(orientation_matrix[1]),
								parseFloat(orientation_matrix[2]),
								parseFloat(model_settings["x"]),
								parseFloat(orientation_matrix[3]),
								parseFloat(orientation_matrix[4]),
								parseFloat(orientation_matrix[5]),
								parseFloat(model_settings["y"]),
								parseFloat(orientation_matrix[6]),
								parseFloat(orientation_matrix[7]),
								parseFloat(orientation_matrix[8]),
								parseFloat(model_settings["z"]),
								0,
								0,
								0,
								1
							);

							let heading, attitude;
							if (m.m10 > 0.998) {
								// singularity at north pole
								heading = Math.atan2(m.m02, m.m22);
								attitude = Math.PI / 2;
								// bank = 0;//unused
								return; //Why do we return here ?
							}
							if (m.m10 < -0.998) {
								// singularity at south pole
								heading = Math.atan2(m.m02, m.m22);
								attitude = -Math.PI / 2;
								// bank = 0;//unused
								return; //Why do we return here ?
							}
							heading = Math.atan2(-m.m20, m.m00);
							//bank = Math.atan2(-m.m12,m.m11);//unused
							attitude = Math.asin(m.m10);
							this.SetCameraAngles(WV3DMath.degrees(-heading - Math.PI / 2), WV3DMath.degrees(-attitude), false);

							//PANNING
							//Store the eye position to determine panning and scale when scene has finished loading
							const eye_pos = new THREE.Vector3(
								parseFloat(camera_props["eye_pos"]["x"]),
								parseFloat(camera_props["eye_pos"]["y"]),
								parseFloat(camera_props["eye_pos"]["z"])
							);
							//Use this one added to eye position ?
							const azim_elev_pos = new THREE.Vector3(
								parseFloat(camera_props["azim_elev_cam_xform"]["x"]),
								parseFloat(camera_props["azim_elev_cam_xform"]["y"]),
								parseFloat(camera_props["azim_elev_cam_xform"]["z"])
							);

							this.get("_renderercache").get("camera_preset")["eye_position"] = eye_pos; //.add(azim_elev_pos);

							//FIELD OF VIEW
							const vfov = parseFloat(camera_props["vfov"]);
							this.SetPerspectiveCamera(vfov, false);

							// const dir_of_sight = camera_props["dir_of_sight"];not needed ? if so, can be exported from collada
						}
						break;
				}
			}
		},

		InitializeModelSettings: function(__root_node) {
			if (__root_node["Properties"].hasOwnProperty("model_settings")) {
				switch (this.get("_renderercache").get("view_settings")["mode"]) {
					default:
						LogMessage.Log("View mode not implemented yet. Model view settings will not be loaded.");
						break;
					// case 0:
					// const model_settings = __root_node['Properties']["model_settings"]["floating_model_xform"];
					// ...
					// break;
					case 1:
						const model_settings = __root_node["Properties"]["model_settings"]["azim_elev_model_xform"];
						const orientation_matrix = model_settings["orientation_matrix"];
						const model_transform = new THREE.Matrix4();
						model_transform.set(
							parseFloat(orientation_matrix[0]),
							parseFloat(orientation_matrix[1]),
							parseFloat(orientation_matrix[2]),
							parseFloat(model_settings["x"]),
							parseFloat(orientation_matrix[3]),
							parseFloat(orientation_matrix[4]),
							parseFloat(orientation_matrix[5]),
							parseFloat(model_settings["y"]),
							parseFloat(orientation_matrix[6]),
							parseFloat(orientation_matrix[7]),
							parseFloat(orientation_matrix[8]),
							parseFloat(model_settings["z"]),
							0,
							0,
							0,
							1
						);
						this.get("_renderercache").get("view_settings")["model_xform"] = model_transform;
						break;
				}
			}
		},

		InitializeReflectionSettings: function(__root_node) {
			if (__root_node["Properties"].hasOwnProperty("reflection_settings")) {
				this.get("_renderercache").get("reflections")["enabled"] = __root_node["Properties"]["reflection_settings"].hasOwnProperty(
					"reflective_table_enabled"
				)
					? WV3DUtils.ToBool(__root_node["Properties"]["reflection_settings"]["reflective_table_enabled"])
					: false;
				if (this.get("_renderercache").get("reflections")["enabled"]) {
					this.get("_renderercache").get("reflections")["opacity"] = __root_node["Properties"]["reflection_settings"].hasOwnProperty(
						"reflection_opacity"
					)
						? parseFloat(__root_node["Properties"]["reflection_settings"]["reflection_opacity"])
						: 0.25;

					if (
						__root_node["Properties"]["reflection_settings"].hasOwnProperty("reflection_fade_size") &&
						__root_node["Properties"]["reflection_settings"].hasOwnProperty("reflection_fade_position")
					) {
						this.get("_renderercache").get("reflections")["fade"] = true;
						this.get("_renderercache").get("reflections")["fade_size"] = parseFloat(
							__root_node["Properties"]["reflection_settings"]["reflection_fade_size"]
						); // Optional.
						this.get("_renderercache").get("reflections")["fade_position"] = parseFloat(
							__root_node["Properties"]["reflection_settings"]["reflection_fade_position"]
						); //Optional.
					}
				}
			}
		},

		InitializeSceneSettings: function(__root_node) {
			if (!this.get("_InitialisedSceneSettings")) {
				//Default values for rendering quality and environment
				if (__root_node["Properties"]) {
					//Set environment settings
					this.InitializeEnvironmentSettings(__root_node);
					//The two following line have been disabled due to product owner's
					//choice: therefore, beware if uncommenting
					//this.InitializeCameraSettings(__root_node);
					//this.InitializeModelSettings(__root_node);
					this.InitializeReflectionSettings(__root_node);
					// Check backdrop, if any, otherwise sets default background color
				}
				this.LoadBackdrop(__root_node);
				this.set("_InitialisedSceneSettings", true);
			}
		},

		/*This function allows to update scene settings after scene has been loaded*/
		UpdateSceneSettings: function() {
			if (this.get("_renderercache").get("camera_preset")["eye_position"]) {
				//Make sure that camera has already been set up according to the scene bounding box
				this.CameraUpdate(false);

				//Use camera preset position to calculate proper panning and scaling offset
				//this approach is inspired from http://stackoverflow.com/questions/5227373/minimal-perpendicular-vector-between-a-point-and-a-line

				//Get the bounding box center
				const scenecenter = this.get("_scenemanager")
					.get("_3JSsceneBBox")
					.getCenter();
				//Get the current view vector
				const current_campos = this.get("_3JScamera").position;
				const pres_campos = this.get("_renderercache")
					.get("camera_preset")
					["eye_position"].clone();

				const viewvec = scenecenter.clone().sub(current_campos);
				const view_dir_offset = viewvec.clone().multiplyScalar(
					pres_campos
						.clone()
						.sub(current_campos)
						.dot(viewvec)
				);

				const upvec = this.get("_3JScamera").up;
				const rightvec = viewvec
					.clone()
					.normalize()
					.cross(upvec);
				const right_dir_offset = rightvec.clone().multiplyScalar(
					pres_campos
						.clone()
						.sub(current_campos)
						.dot(rightvec)
				);
				const up_dir_offset = upvec.clone().multiplyScalar(
					pres_campos
						.clone()
						.sub(current_campos)
						.dot(upvec)
				);

				this.SetCameraPanOffset({ x: right_dir_offset.length(), y: up_dir_offset.length() }, false);

				const prev_length = viewvec.length();
				const new_length = viewvec
					.clone()
					.sub(view_dir_offset)
					.length();
				this.SetCameraScale(new_length / prev_length);
			}
		},

		/*= REFLECTIONS =*/

		/**
		 * SetReflections
		 *
		 * @function SetReflections sets the reflection on, off, or relative to what is set in the Json file.
		 * @param __config "ON", "JSON", "OFF".
		 * @param __opacity only taken into account if __config is "ON"
		 * @param __fade only taken into account if __config is "ON"
		 * @param __position only taken into account if __config is "ON"
		 * @param __fadingSize only taken into account if __config is "ON"
		 * @param __redraw redraw if true, otherwise does not
		 */
		SetReflections: function(__config, __opacity, __fade, __position, __fadingSize, __redraw) {
			this.set("_reflection_config", __config);
			if (__config === "ON") {
				//Use parameters from API
				this.get("_materialmanager").UpdateReflectionUniforms(__opacity, __fade, __position, __fadingSize);
			} else if (__config === "JSON") {
				//Use default parameters set from JSON asset
				this.get("_materialmanager").UpdateReflectionUniforms();
			}
			if (__redraw === undefined || __redraw) this.Render();
		},

		SetupReflections: function() {
			//Epsilon is needed to offset the reflection, otherwise it can be coincident with the object's bottom
			const Y_axis_symmetry_transform_matrix = new THREE.Matrix4();
			Y_axis_symmetry_transform_matrix.set(1, 0, 0, 0, 0, -1, 0, 0, 0, 0, 1, 0, 0, 0, 0, 1);
			this.set("_reflection_Y_scale", Y_axis_symmetry_transform_matrix);
			this.set("_mirror_offset", 0.01);
			const reflection_Y_trans = this.get("_scenemanager").get("_3JSsceneBBox").min.y * 2 - this.get("_mirror_offset");
			const Y_axis_translation_transform_matrix = new THREE.Matrix4();
			Y_axis_translation_transform_matrix.makeTranslation(0, reflection_Y_trans, 0);
			this.set("_reflection_Y_translation", Y_axis_translation_transform_matrix);
			const Y_axis_opp_translation_transform_matrix = new THREE.Matrix4();
			Y_axis_opp_translation_transform_matrix.makeTranslation(0, -reflection_Y_trans, 0);
			this.set("_reflection_Y_opp_translation", Y_axis_opp_translation_transform_matrix);
		},

		/*= BACKGROUND =*/

		/**
		 * Check backdrop, if any, otherwise sets default background color
		 */
		LoadBackdrop: function(__root_node) {
			if (__root_node["Properties"] && __root_node["Properties"].hasOwnProperty("backdrop")) {
				if (__root_node["Properties"]["backdrop"]["type"] === "COLOR") {
					const bdColor = __root_node["Properties"]["backdrop"]["color"];
					this.SetBackgroundColor(bdColor[0], bdColor[1], bdColor[2], 1, false);
				} else if (__root_node["Properties"]["backdrop"]["type"] === "IMAGE") {
					this.get("_texturemanager").LoadBackdrop(__root_node, this.get("_model").get("SceneId"));
				} else {
					LogMessage.Log("ViewerView Warning: Backdrop type not supported (default to grey).");
					this.SetBackgroundColor(0.37, 0.37, 0.37, 1, false); // default background of plugin
				}
			} else {
				this.SetBackgroundColor(0.37, 0.37, 0.37, 1, false); // default background of plugin
			}
		},

		/**
		 * SetBackdropTexture
		 */
		SetBackdropTexture: function(__textureURL, __callback, __redraw) {
			this.get("_texturemanager").SetBackdropTexture(__textureURL, __callback, __redraw);
		},

		/** SetBackgroundColor */
		SetBackgroundColor: function(__red, __green, __blue, __alpha, __redraw) {
			this.set("_backgroundColor", new THREE.Color(__red, __green, __blue));
			this.set("_backgroundAlpha", __alpha);
			this.get("_renderercache").get("textures")["backdrop"]["enabled"] = false;

			if (__redraw === undefined || __redraw) this.Render();
		},

		/**
		 * Reload all textures
		 *
		 * @function ReLoad all textures (only material textures) that have been loaded by the json file.
		 * @param {function} __callback - callback funtion used when rendering has been updated
		 * @param {Boolean} [__redraw = true] If true, redraw the scene after this call
		 */
		ReloadAllTextures: function(__callback, __redraw) {
			LogMessage.Log("Controller ReLoadAllTextures function");
			this.get("_texturemanager").ReloadAllTextures(__callback, __redraw);
		},

		/**
		 * Reload a texture
		 *
		 * @function ReloadTexture
		 * @param {String} __textureURL - URL identifying the initial texture relative path from json to texture
		 * @param {String} __newURL - URL identifying the new texture relative path from json to texture
		 * @param {function} __callback - callback funtion used when rendering has been updated
		 * @param {Boolean} [__keep_old_path = true] - indicates if the old VS new path has to be used now on to reference the texture
		 * @param {Boolean} [__redraw = true] If true, redraw the scene after this call
		 */
		ReloadTexture: function(__textureURL, __newURL, __callback, __keep_old_path, __redraw) {
			LogMessage.Log("Controller ReLoadTexture function");
			this.get("_texturemanager").ReloadTexture(__textureURL, __newURL, __callback, __keep_old_path, __redraw);
		},

		/*= LABELS =*/

		UpdateLabelProperties: function(__nodeURL, __properties) {
			const retVal = this.get("_scenemanager").UpdateLabelProperties(__nodeURL, __properties);
			if (retVal) {
				this.CameraUpdate(true);
			}
			return retVal;
		},

		UpdateAllLabelsProperties: function(__properties) {
			const retVal = this.get("_scenemanager").UpdateAllLabelsProperties(__properties);
			if (retVal) {
				this.CameraUpdate(true);
			}
			return retVal;
		},

		RotateOutOfBoundsVisibleLabels: function(
			__camera,
			__globalBBox,
			__screenwidth,
			__screenheight,
			__marginscreenwidth,
			__marginscreenheight
		) {
			let success = true;
			for (let idx_key = 0, keys = Object.keys(this.get("_renderercache").get("labels")["list"]), l = keys.length; idx_key < l; ++idx_key) {
				const id = keys[idx_key];
				const label_asset = this.get("_renderercache").get("labels")["list"][id];
				//Get the labels bounding box
				let label_bbox = this.GetLabelBoundingBox(label_asset, label_asset["orientation"]);
				let modified_layout = undefined;
				//Rotate label if out of bounds in one direction
				if (label_asset["orientation"] !== "vertical") {
					if (label_bbox.max.x > __marginscreenwidth || label_bbox.min.x < -__marginscreenwidth) {
						label_asset["threejs_sprite"].material.rotation = 1.5708;
						modified_layout = "vertical";
					} else {
						if (label_bbox.max.y > __marginscreenheight || label_bbox.min.y < -__marginscreenheight) {
							success = false;
						}
						label_asset["threejs_sprite"].material.rotation = 0;
					}
				} else {
					if (label_bbox.max.y > __marginscreenheight || label_bbox.min.y < -__marginscreenheight) {
						label_asset["threejs_sprite"].material.rotation = 0;
						modified_layout = "horizontal";
					} else {
						if (label_bbox.max.x > __marginscreenwidth || label_bbox.min.x < -__marginscreenwidth) {
							success = false;
						}
						label_asset["threejs_sprite"].material.rotation = 1.5708;
					}
				}
				if (modified_layout) {
					//Update the position (label might end up in a different place)
					this.UpdateLabelPosition(label_asset, __camera, __screenwidth / 2, __screenheight / 2, modified_layout);
					//Get its new bbox
					label_bbox = this.GetLabelBoundingBox(label_asset, modified_layout);
					//Test against it new bounding box
					if (
						label_bbox.max.x > __marginscreenwidth ||
						label_bbox.min.x < -__marginscreenwidth ||
						label_bbox.max.y > __marginscreenheight ||
						label_bbox.min.y < -__marginscreenheight
					) {
						success = false;
					}
				}
				//Now update the global bounding box
				__globalBBox.union(label_bbox);
			}
			return success;
		},

		GetLabelBoundingBox: function(__label_asset, __overrideLayout) {
			const sprite_object = __label_asset["threejs_sprite"];
			const label_bbox = new THREE.Box2();
			const spriteDimensions = { spriteWidth: sprite_object.scale.x, spriteHeight: sprite_object.scale.y };
			const spriteHWidth =
				0.5 *
				((__overrideLayout ? __overrideLayout : __label_asset["orientation"]) == "vertical"
					? spriteDimensions.spriteHeight
					: spriteDimensions.spriteWidth);
			const spriteHHeight =
				0.5 *
				((__overrideLayout ? __overrideLayout : __label_asset["orientation"]) == "vertical"
					? spriteDimensions.spriteWidth
					: spriteDimensions.spriteHeight);

			label_bbox.expandByPoint(new THREE.Vector2(sprite_object.position.x - spriteHWidth, sprite_object.position.y - spriteHHeight));
			label_bbox.expandByPoint(new THREE.Vector2(sprite_object.position.x - spriteHWidth, sprite_object.position.y + spriteHHeight));
			label_bbox.expandByPoint(new THREE.Vector2(sprite_object.position.x + spriteHWidth, sprite_object.position.y - spriteHHeight));
			label_bbox.expandByPoint(new THREE.Vector2(sprite_object.position.x + spriteHWidth, sprite_object.position.y + spriteHHeight));
			return label_bbox;
		},

		UpdateLabelsPositions: function(__camera, __overrideWidth, __overrideHeight) {
			__camera.updateMatrixWorld();
			const screenwidth = __overrideWidth ? __overrideWidth : this.get("_context").clientWidth;
			const screenheight = __overrideHeight ? __overrideHeight : this.get("_context").clientHeight;

			for (let idx_key = 0, keys = Object.keys(this.get("_renderercache").get("labels")["list"]), l = keys.length; idx_key < l; ++idx_key) {
				const id = keys[idx_key];
				const label_asset = this.get("_renderercache").get("labels")["list"][id];
				this.UpdateLabelPosition(label_asset, __camera, screenwidth / 2, screenheight / 2);
			}
		},

		UpdateLabelPosition: function(__label_asset, __camera, __screenwidth, __screenheight, __overrideLayout) {
			//Position each label contained in rendercache list
			const sprite_object = __label_asset["threejs_sprite"];

			//Get the list of attachment points
			const attachment_points = [];
			const attConvexHull = []; // convex hull of pointlist or edges endpoints
			let vcenter; // projected center, in case of ellipse label
			switch (__label_asset["bounds_type"]) {
				case "edge_list":
					for (let ed = 0, ed_keys = Object.keys(__label_asset.edge_list), n_eds = ed_keys.length; ed < n_eds; ++ed) {
						const v1 = __label_asset.edge_list[ed_keys[ed]]["v1"].clone();
						v1.project(__camera);
						attConvexHull.push(new THREE.Vector2(v1.x * __screenwidth, v1.y * __screenheight));
						const v2 = __label_asset.edge_list[ed_keys[ed]]["v2"].clone();
						v2.project(__camera);
						attConvexHull.push(new THREE.Vector2(v2.x * __screenwidth, v2.y * __screenheight));
						const att_pt = v1.add(v2).multiplyScalar(0.5);

						//Offset mid point of the edge in the direction of user preference
						attachment_points.push(new THREE.Vector2(att_pt.x * __screenwidth, att_pt.y * __screenheight));
					}
					break;
				case "point_list":
					for (let ed = 0, ed_keys = Object.keys(__label_asset.point_list), n_eds = ed_keys.length; ed < n_eds; ++ed) {
						const att_pt = __label_asset.point_list[ed_keys[ed]].clone();
						att_pt.project(__camera);

						attConvexHull.push(new THREE.Vector2(att_pt.x * __screenwidth, att_pt.y * __screenheight));

						attachment_points.push(new THREE.Vector2(att_pt.x * __screenwidth, att_pt.y * __screenheight));
					}
					break;
				case "ellipse":
					//We use the equations mentioned at http://mathforum.org/library/drmath/view/63373.html
					vcenter = __label_asset["ellipse"]["center"].clone();
					let vradius1 = __label_asset["ellipse"]["radius1_vec"].clone();
					let vradius2 = __label_asset["ellipse"]["radius2_vec"].clone();

					//Find the outer most points based on ellipse plane's direction w.r.t. camera
					//Project ellipse vertices into view space

					const ellipse_plane_normal = vradius1
						.clone()
						.cross(vradius2)
						.normalize();
					ellipse_plane_normal.x = Math.abs(ellipse_plane_normal.x);
					ellipse_plane_normal.y = Math.abs(ellipse_plane_normal.y);
					ellipse_plane_normal.z = Math.abs(ellipse_plane_normal.z);
					const horizontal_layout = ellipse_plane_normal.angleTo(new THREE.Vector3(0, 1, 0)) < Math.PI * 0.25;

					//Radii are converted into points before projection
					vradius1.add(vcenter);
					vradius2.add(vcenter);

					vcenter.project(__camera);
					vradius1.project(__camera);
					vradius2.project(__camera);

					//Radii are converted back to vectors after projection
					vradius1.sub(vcenter);
					vradius2.sub(vcenter);

					//Ensure radius 1 is the longest of the two radii (swap them if necessary)
					if (vradius1.length() < vradius2.length()) {
						vradius1 = [vradius2, (vradius2 = vradius1)][0];
					}

					//At this stage vcenter, vradius1, and vradius2 contain the projected ellipse in view space
					const a = vradius1.length(),
						b = vradius2.length();
					const unit_vradius1 = vradius1.clone().normalize();
					const unit_vradius2 = vradius2.clone().normalize();

					const xc = vcenter.x;
					const yc = vcenter.y;
					const zc = vcenter.z;

					//intersections with Cx axis
					const t1 = Math.atan2(-a * unit_vradius1.y, b * unit_vradius2.y);
					//min_x
					let x1 = a * Math.cos(t1) * unit_vradius1.x + b * Math.sin(t1) * unit_vradius2.x;
					let y1 = a * Math.cos(t1) * unit_vradius1.y + b * Math.sin(t1) * unit_vradius2.y;

					//max_x
					const t2 = t1 + Math.PI;
					let x2 = a * Math.cos(t2) * unit_vradius1.x + b * Math.sin(t2) * unit_vradius2.x;
					let y2 = a * Math.cos(t2) * unit_vradius1.y + b * Math.sin(t2) * unit_vradius2.y;

					//intersections with Cy axis
					// const t3 = Math.atan2(-a * unit_vradius1.x, b * unit_vradius2.x);//initial intention
					const t3 = t2 + Math.PI * 0.5;
					//min_y
					let x3 = a * Math.cos(t3) * unit_vradius1.x + b * Math.sin(t3) * unit_vradius2.x;
					let y3 = a * Math.cos(t3) * unit_vradius1.y + b * Math.sin(t3) * unit_vradius2.y;

					//max_y
					const t4 = t3 + Math.PI;
					let x4 = a * Math.cos(t4) * unit_vradius1.x + b * Math.sin(t4) * unit_vradius2.x;
					let y4 = a * Math.cos(t4) * unit_vradius1.y + b * Math.sin(t4) * unit_vradius2.y;

					x1 = x1 + xc;
					y1 = y1 + yc;
					x2 = x2 + xc;
					y2 = y2 + yc;
					x3 = x3 + xc;
					y3 = y3 + yc;
					x4 = x4 + xc;
					y4 = y4 + yc;
					const att_pt1 = new THREE.Vector2(x1 * __screenwidth, y1 * __screenheight);
					const att_pt2 = new THREE.Vector2(x2 * __screenwidth, y2 * __screenheight);
					const att_pt3 = new THREE.Vector2(x3 * __screenwidth, y3 * __screenheight);
					const att_pt4 = new THREE.Vector2(x4 * __screenwidth, y4 * __screenheight);

					//From these attachement points, we want to select the ones that are
					//located far from each other.

					const lengtp1p2 = att_pt1.distanceToSquared(att_pt2);
					const lengtp1p3 = att_pt1.distanceToSquared(att_pt3);
					const lengtp1p4 = att_pt1.distanceToSquared(att_pt4);
					const lengtp2p3 = att_pt2.distanceToSquared(att_pt3);
					const lengtp2p4 = att_pt2.distanceToSquared(att_pt4);
					const lengtp3p4 = att_pt3.distanceToSquared(att_pt4);

					let sel_att_pt1,
						sel_att_pt2,
						max_length = 0;
					if (lengtp1p2 >= max_length) {
						max_length = lengtp1p2;
						sel_att_pt1 = att_pt1;
						sel_att_pt2 = att_pt2;
					}
					if (lengtp1p3 >= max_length) {
						max_length = lengtp1p3;
						sel_att_pt1 = att_pt1;
						sel_att_pt2 = att_pt3;
					}
					if (lengtp1p4 >= max_length) {
						max_length = lengtp1p4;
						sel_att_pt1 = att_pt1;
						sel_att_pt2 = att_pt4;
					}
					if (lengtp2p3 >= max_length) {
						max_length = lengtp2p3;
						sel_att_pt1 = att_pt2;
						sel_att_pt2 = att_pt3;
					}
					if (lengtp2p4 >= max_length) {
						max_length = lengtp2p4;
						sel_att_pt1 = att_pt2;
						sel_att_pt2 = att_pt4;
					}
					if (lengtp3p4 >= max_length) {
						max_length = lengtp3p4;
						sel_att_pt1 = att_pt3;
						sel_att_pt2 = att_pt4;
					}

					attachment_points.push(sel_att_pt1);
					attachment_points.push(sel_att_pt2);
					attConvexHull.push(new THREE.Vector2(vcenter.x, vcenter.y));
					break;
				default:
					LogMessage.Log("Warning! Unsupported label_asset type.");
					return;
			}

			let convexHullCtr,
				nbHullPts,
				x = 0,
				y = 0;
			const n_attps = attachment_points.length;
			switch (n_attps) {
				case 0:
					LogMessage.Log("Warning: no attachment point found.");
					break;
				case 1:
					x = attachment_points[0].x;
					y = attachment_points[0].y;
					break;
				default:
					//Get 2D convex hull of attachment points
					const attPtsProjBBox = new THREE.Box2();
					for (let attPtIdx = 0, n_attpt = attachment_points.length; attPtIdx < n_attpt; ++attPtIdx) {
						const attPt = attachment_points[attPtIdx];
						attPtsProjBBox.expandByPoint(new THREE.Vector2(attPt.x, attPt.y));
					}

					//Check which atachment points are valid relatively to itself
					const valid_att_points = [];
					for (let attPtIdx = 0, n_attpt = attachment_points.length; attPtIdx < n_attpt; ++attPtIdx) {
						const attPt = attachment_points[attPtIdx];

						if (
							attPt.x === attPtsProjBBox.min.x ||
							attPt.x === attPtsProjBBox.max.x ||
							attPt.y === attPtsProjBBox.min.y ||
							attPt.y === attPtsProjBBox.max.y
						) {
							valid_att_points.push(attPt);
						}
					}

					//if no valid point (meaning there is probably an issue)
					switch (valid_att_points.length) {
						case 0:
							//calculate main direction for the entire point cloud and choose one anyway according to preferences
							console.log("There is probably an issue here. at least for vertices and edges");
							break;
						case 1:
							//if only one is valid choose it
							(x = valid_att_points[0].x), (y = valid_att_points[0].y);
							break;
						default:
							//using user preferences we have to select the best attachment point
							if (__label_asset.preferred_loc_h === "center" && __label_asset.preferred_loc_v === "center") {
								nbHullPts = WV3DMath.CreateConvexHullFromPointlist(attConvexHull);
								// convexHullCtr = WV3DMath.ConvexHullCenter( attConvexHull, nbHullPts );
								// We use all points contained in convex hull list otherwise the center is wrongly calculated
								// when convex hull vertices are alined (easy to spot in orthographics view)
								convexHullCtr = WV3DMath.ConvexHullCenter(attConvexHull, attConvexHull.length - 1);
								x = convexHullCtr.x;
								y = convexHullCtr.y;
							} else if (__label_asset.preferred_loc_h === "center" || __label_asset.preferred_loc_v === "center") {
								x = valid_att_points[0].x;
								y = valid_att_points[0].y;
								for (let attPtIdx = 1, n_attpt = valid_att_points.length; attPtIdx < n_attpt; ++attPtIdx) {
									const attPt = valid_att_points[attPtIdx];
									if (__label_asset.preferred_loc_h === "left") {
										if (attPt.x < x) {
											x = attPt.x;
											y = attPt.y;
										}
									} else if (__label_asset.preferred_loc_h === "right") {
										if (attPt.x > x) {
											x = attPt.x;
											y = attPt.y;
										}
									} else if (__label_asset.preferred_loc_v === "bottom") {
										if (attPt.y < y) {
											x = attPt.x;
											y = attPt.y;
										}
									} else if (__label_asset.preferred_loc_v === "top") {
										if (attPt.y > y) {
											x = attPt.x;
											y = attPt.y;
										}
									} else {
										LogMessage.Log("Unrecognised Label user preference !");
										x = 0;
										y = 0;
									}
								}
							} else {
								//calculate main direction for the point cloud and choose one according to preferences
								const cumul_diff = new THREE.Vector2();
								for (let attPtIdx = 0, n_attpt = valid_att_points.length; attPtIdx < n_attpt; ++attPtIdx) {
									const attPt = valid_att_points[attPtIdx];
									for (let attPtIdx2 = attPtIdx + 1; attPtIdx2 < n_attpt; ++attPtIdx2) {
										const attPt2 = valid_att_points[attPtIdx2];
										const delta = attPt.clone().sub(attPt2);
										cumul_diff.add(new THREE.Vector2(Math.abs(delta.x), Math.abs(delta.y)));
									}
								}
								x = valid_att_points[0].x;
								y = valid_att_points[0].y;
								if (cumul_diff.x > cumul_diff.y) {
									// console.log("horizontal layout ! dx/dy: "+cumul_diff.x+"/"+cumul_diff.y);
									//select right or left most vertex from valid_att_points according to user preferences
									for (let attPtIdx = 1, n_attpt = valid_att_points.length; attPtIdx < n_attpt; ++attPtIdx) {
										const attPt = valid_att_points[attPtIdx];
										if (__label_asset.preferred_loc_h === "left") {
											if (attPt.x < x) {
												(x = attPt.x), (y = attPt.y);
											}
										} else if (__label_asset.preferred_loc_h === "right") {
											if (attPt.x > x) {
												(x = attPt.x), (y = attPt.y);
											}
										} else {
											LogMessage.Log("Unrecognised horizontal layout: " + __label_asset.preferred_loc_h);
										}
									}
								} else {
									// console.log("vertical layout ! dx/dy: "+cumul_diff.x+"/"+cumul_diff.y);
									//select top or bottom most vertex from valid_att_points according to user preferences
									for (let attPtIdx = 1, n_attpt = valid_att_points.length; attPtIdx < n_attpt; ++attPtIdx) {
										const attPt = valid_att_points[attPtIdx];
										if (__label_asset.preferred_loc_v === "bottom") {
											if (attPt.y < y) {
												x = attPt.x;
												y = attPt.y;
											}
										} else if (__label_asset.preferred_loc_v === "top") {
											if (attPt.y > y) {
												x = attPt.x;
												y = attPt.y;
											}
										} else {
											LogMessage.Log("Unrecognised vertical layout: " + __label_asset.preferred_loc_v);
										}
									}
								}
							}
					}
					break;
			}

			//Offset labels
			let px_offset_x = 0,
				px_offset_y = 0;

			// calculate length of string + scaling
			const spriteDimensions = {
				spriteWidth: __label_asset["threejs_sprite"].scale.x,
				spriteHeight: __label_asset["threejs_sprite"].scale.y
			};
			const spriteHWidth =
				0.5 *
				((__overrideLayout ? __overrideLayout : __label_asset["orientation"]) === "vertical"
					? spriteDimensions.spriteHeight
					: spriteDimensions.spriteWidth);
			const spriteHHeight =
				0.5 *
				((__overrideLayout ? __overrideLayout : __label_asset["orientation"]) === "vertical"
					? spriteDimensions.spriteWidth
					: spriteDimensions.spriteHeight);
			const user_pixel_offset = new THREE.Vector2(__label_asset.pixel_offset[0], __label_asset.pixel_offset[1]);

			if (
				(n_attps === 1 && __label_asset["bounds_type"] === "point_list") ||
				(__label_asset.preferred_loc_h === "center" && __label_asset.preferred_loc_v === "center")
			) {
				// don't do anything here as we can't derive a horizontal/vertical layout from a single point.
				// only add user offset if any
				let x_dir = 0,
					y_dir = 0;
				if (__label_asset.preferred_loc_h === "right") {
					x_dir = 1;
				} else if (__label_asset.preferred_loc_h === "left") {
					x_dir = -1;
				}
				if (__label_asset.preferred_loc_v === "top") {
					y_dir = 1;
				} else if (__label_asset.preferred_loc_v === "bottom") {
					y_dir = -1;
				}
				px_offset_x += user_pixel_offset.x * Math.sign(x_dir);
				px_offset_y += user_pixel_offset.y * Math.sign(y_dir);
			} else if (__label_asset["bounds_type"] === "ellipse") {
				// for an ellipse, push away from the center of the ellipse.
				// vCenter in camera space, but not in pixel coordinates
				const center = new THREE.Vector2(vcenter.x * __screenwidth, vcenter.y * __screenheight);
				px_offset_x = x - center.x + user_pixel_offset.x;
				px_offset_y = y - center.y + user_pixel_offset.y;
			} // edge list or multiple points
			else {
				// calculate convex hull and calculate direction in which to push the attachment point
				const n_hullPts = nbHullPts ? nbHullPts : WV3DMath.CreateConvexHullFromPointlist(attConvexHull);
				// const center = convexHullCtr? convexHullCtr : WV3DMath.ConvexHullCenter( attConvexHull, n_hullPts );
				const center = convexHullCtr ? convexHullCtr : WV3DMath.ConvexHullCenter(attConvexHull, attConvexHull.length - 1);

				const dir = new THREE.Vector2(x - center.x, y - center.y).normalize();

				// push label by moving according to layout and user preference
				const cum_diff_hull = new THREE.Vector2();
				for (let attPtIdx = 0, n_attpt = attConvexHull.length - 1; attPtIdx < n_attpt; ++attPtIdx) {
					const attPt = attConvexHull[attPtIdx];
					for (let attPtIdx2 = attPtIdx + 1; attPtIdx2 < n_attpt; ++attPtIdx2) {
						const attPt2 = attConvexHull[attPtIdx2];
						const delta = attPt.clone().sub(attPt2);
						cum_diff_hull.add(new THREE.Vector2(Math.abs(delta.x), Math.abs(delta.y)));
					}
				}
				let x_dir, y_dir;
				if (cum_diff_hull.x > cum_diff_hull.y) {
					if (__label_asset.preferred_loc_v === "top") {
						x_dir = 0;
						y_dir = 1;
					} else if (__label_asset.preferred_loc_v === "bottom") {
						x_dir = 0;
						y_dir = -1;
					} else if (__label_asset.preferred_loc_h === "right") {
						x_dir = 1;
						y_dir = 0;
					} else if (__label_asset.preferred_loc_h === "left") {
						x_dir = -1;
						y_dir = 0;
					}
				} else {
					if (__label_asset.preferred_loc_h === "right") {
						x_dir = 1;
						y_dir = 0;
					} else if (__label_asset.preferred_loc_h === "left") {
						x_dir = -1;
						y_dir = 0;
					} else if (__label_asset.preferred_loc_v === "top") {
						x_dir = 0;
						y_dir = 1;
					} else if (__label_asset.preferred_loc_v === "bottom") {
						x_dir = 0;
						y_dir = -1;
					}
				}
				const epsilon = 0.001;
				px_offset_x = (spriteHWidth + epsilon) * x_dir;
				px_offset_y = (spriteHHeight + epsilon) * y_dir;

				//Move iteratively the label to ensure that it is entirely (corners) outside of the convex hull
				let nb_max_iterations = 10000; //Safety for following loop
				let contains = true;
				while (nb_max_iterations && contains) {
					nb_max_iterations--;
					const label_corners = [
						new THREE.Vector2(x + spriteHWidth + px_offset_x, y + spriteHHeight + px_offset_y),
						new THREE.Vector2(x - spriteHWidth + px_offset_x, y + spriteHHeight + px_offset_y),
						new THREE.Vector2(x + spriteHWidth + px_offset_x, y - spriteHHeight + px_offset_y),
						new THREE.Vector2(x - spriteHWidth + px_offset_x, y - spriteHHeight + px_offset_y)
					];

					//Test box corners against convex hull
					if (
						WV3DMath.ConvexHullContainsPoint(attConvexHull, n_hullPts, label_corners[0]) ||
						WV3DMath.ConvexHullContainsPoint(attConvexHull, n_hullPts, label_corners[1]) ||
						WV3DMath.ConvexHullContainsPoint(attConvexHull, n_hullPts, label_corners[2]) ||
						WV3DMath.ConvexHullContainsPoint(attConvexHull, n_hullPts, label_corners[3])
					) {
						px_offset_x += dir.x;
						px_offset_y += dir.y;
						continue;
					}

					//Test each convex hull vertex against label bbox
					const label_box = new THREE.Box2();
					label_box.setFromPoints(label_corners);
					for (let idx_cvx_hull = 0; idx_cvx_hull < attConvexHull; ++idx_cvx_hull) {
						if (label_box.containsPoint(attConvexHull[idx_cvx_hull])) {
							px_offset_x += dir.x;
							px_offset_y += dir.y;
						}
					}

					contains = false;
				}
				if (nb_max_iterations === 0) {
					LogMessage.Log("ViewerView Warning: too many iterations when pushing label outside of its convex hull.");
				}
				px_offset_x += user_pixel_offset.x * Math.sign(x_dir);
				px_offset_y += user_pixel_offset.y * Math.sign(y_dir);
			}

			//Update labels positions
			sprite_object.position.set(x + px_offset_x, y + px_offset_y, 0);
		},

		/*= ANNOTATIONS *=/

    /**
    * Retrieve the list of the parts that can support annotations
    */
		GetAnnotablePartsList: function() {
			return this.get("_materialmanager").GetAnnotablePartsList();
		},

		/**
		 * Set the list of annotation segments for the given part
		 */
		SetAnnotations: function(__jsonMaterialName, __segmentList, __redraw) {
			const success = this.get("_materialmanager").SetAnnotations(__jsonMaterialName, __segmentList, __redraw);
			if (success && (__redraw === undefined || __redraw)) {
				let anyMaterialHasAnnotations = false;

				this.get("_materialmanager").ApplyFctToAllMaterials(function(__material) {
					if (
						__material.hasOwnProperty("defines") &&
						__material["defines"].hasOwnProperty("ANNOTATION_NUMBER") &&
						__material["defines"]["ANNOTATION_NUMBER"] > 0
					) {
						anyMaterialHasAnnotations = true;
					}
				});

				this.set("_annotationsEnabled", anyMaterialHasAnnotations);

				this.Render();
			}
			return success;
		},

		/**
		 * Set the thickness of annotation segments
		 */
		SetAnnotationsThickness: function(__thickness, __redraw) {
			this.get("_materialmanager").SetAnnotationsThickness(__thickness, __redraw);
			if (__redraw === undefined || __redraw) this.Render();
		},

		/**
		 * Set the color of annotation segments
		 */
		SetAnnotationsColor: function(__red, __green, __blue, __redraw) {
			this.get("_materialmanager").SetAnnotationsColor(__red / 255, __green / 255, __blue / 255, __redraw);
			if (__redraw === undefined || __redraw) this.Render();
		},

		/*= CAMERA =*/

		/**
		 * This function returns the smallest distance between the center of the scene bounding box and any of its faces,
		 * which can be visualized as the radius of the largest possible sphere still entirely contained inside this box.
		 * It is used as a camera angle independent (hence the sphere) reference point with which we can later define the
		 * appropriate unit distance and scale values. We use this distance because, regardless of the angle, the camera
		 * should NEVER get any closer to the scene center than this value.
		 */
		GetSceneInnerSphereRadius: function() {
			const bbox = this.get("_scenemanager").get("_3JSsceneBBoxFocus");
			const bbox_max_relative = bbox.max.clone().sub(bbox.getCenter());
			return Math.min(Math.min(bbox_max_relative.x, bbox_max_relative.y), bbox_max_relative.z);
		},

		/**
		 * Set camera settings according to presets
		 */
		CameraUpdate: function(__redraw, __overrideWidth, __overrideHeight, __overrideCameraDistance) {
			const sceneBBoxRef = this.get("_scenemanager").get("_3JSsceneBBox");
			if (sceneBBoxRef && !sceneBBoxRef.isEmpty()) {
				const sceneBBox = sceneBBoxRef.clone();

				// Adjust camera to scene (using focus bounding box)
				this.SetCamera(this.get("_scenemanager").get("_3JSsceneBBoxFocus"), __overrideWidth, __overrideHeight, __overrideCameraDistance);

				// Setup the direction light to provide light from above right of camera
				const camPos = this.get("_3JScamera").position.clone();
				const camUp = this.get("_3JScamera")
					.up.clone()
					.normalize();
				const camPoint = sceneBBox
					.getCenter()
					.clone()
					.sub(camPos);

				camPoint.normalize();
				const camRight = camPoint.clone().cross(camUp);

				const lightDir = camPoint
					.clone()
					.sub(camUp.clone().multiplyScalar(0.5))
					.sub(camRight.clone().multiplyScalar(0.7)); //camPoint - 0.3 * camUp - 0.3 * camRight

				for (let i = 0, cg_keys = Object.keys(this.get("_renderercache").get("lights")["dir_light"]), l = cg_keys.length; i < l; ++i) {
					const dir_light = this.get("_renderercache").get("lights")["dir_light"][cg_keys[i]];
					if (dir_light) {
						//for default and designer quality materials
						dir_light.position.set(-lightDir.x, -lightDir.y, -lightDir.z);
						dir_light.needsUpdate = true;
					}
				}

				//Also need to update near and far clipping planes of camera relatively to the reflected scene
				const reflectedpoint = this.get("_scenemanager")
					.get("_3JSsceneBBox")
					.min.clone();
				const trans_reflection = new THREE.Vector3(
					0,
					this.get("_scenemanager")
						.get("_3JSsceneBBox")
						.getSize().y + this.get("_mirror_offset"),
					0
				);
				reflectedpoint.sub(trans_reflection);
				sceneBBox.expandByPoint(reflectedpoint);

				// Adjust camera to scene
				this.UpdateCamNearAndFar(sceneBBox, __overrideWidth, __overrideHeight);

				// update camera matrices and frustum
				if (this.get("_3JScamera").parent === null) this.get("_3JScamera").updateMatrixWorld();

				const _projScreenMatrix = new THREE.Matrix4();
				const _frustum = new THREE.Frustum();
				this.get("_3JScamera").matrixWorldInverse.getInverse(this.get("_3JScamera").matrixWorld);
				_projScreenMatrix.multiplyMatrices(this.get("_3JScamera").projectionMatrix, this.get("_3JScamera").matrixWorldInverse);
				_frustum.setFromMatrix(_projScreenMatrix);

				const bboxgeom = new THREE.Geometry();
				bboxgeom.vertices.push(this.get("_scenemanager").get("_3JSsceneBBox").min, this.get("_scenemanager").get("_3JSsceneBBox").max);
				const bboxmesh = new THREE.Mesh(bboxgeom, null);

				//Check if scene is outside of frustum
				this.set("_obj_in_frustum_prev", this.get("_obj_in_frustum"));
				const intersect_frustum = _frustum.intersectsObject(bboxmesh);
				this.set("_obj_in_frustum", intersect_frustum);

				if (__redraw === undefined || __redraw) this.Render();

				return true;
			} else return false;
		},

		/**
		 * Set the angular position of the camera relative to the default front position of the scene
		 *
		 * @function
		 * @param {Number} __azimuthAngle - the walk-around angle measured counterclockwise from the default front view
		 * @param {Number} __elevationAngle - the elevantion angle - positive values lift the camera above the scene centre
		 * @param __redraw
		 */
		SetCameraAngles: function(__azimuthAngle, __elevationAngle, __redraw) {
			const prev_az = this.get("_renderercache").get("cameras")["defaultcamera"].azimuthAngle;
			const prev_el = this.get("_renderercache").get("cameras")["defaultcamera"].elevationAngle;
			this.get("_renderercache").get("cameras")["defaultcamera"].azimuthAngle = __azimuthAngle;
			this.get("_renderercache").get("cameras")["defaultcamera"].elevationAngle = __elevationAngle;
			this.CameraUpdate(__redraw);

			if (!this.get("_obj_in_frustum") && this.get("_obj_in_frustum_prev")) {
				//Object was previously in bounds but went out of frustum because
				//of these new parameters: reverting to previous state
				this.get("_renderercache").get("cameras")["defaultcamera"].azimuthAngle = prev_az;
				this.get("_renderercache").get("cameras")["defaultcamera"].elevationAngle = prev_el;
				this.CameraUpdate(__redraw);
			}
		},

		GetCameraAngles: function() {
			return {
				azimuth: this.get("_renderercache").get("cameras")["defaultcamera"].azimuthAngle,
				elevation: this.get("_renderercache").get("cameras")["defaultcamera"].elevationAngle
			};
		},

		/**
		 * Set the size of the scene within the camera view. This is achieved by automatically
		 * adjusting the camera properties (e.g. changing the camera distance for a perspective camera)
		 *
		 * @function
		 * @param {Number} __scale - the size of the scene bounding box in the view. A value of 1 will result in the scene filling the view in at least one dimension
		 * @param __redraw
		 */
		SetCameraScale: function(__scale, __redraw) {
			const prev_scale = this.get("_renderercache").get("cameras")["defaultcamera"].scale;
			// Clamp the scale factor
			const scale = Math.min(Math.max(__scale, this.get("_scale_min")), this.get("_scale_max"));
			this.get("_renderercache").get("cameras")["defaultcamera"].scale = scale;
			this.CameraUpdate(__redraw);

			if (!this.get("_obj_in_frustum") && this.get("_obj_in_frustum_prev")) {
				//Object was previously in bounds but went out of frustum because
				//of these new parameters: reverting to previous state
				this.get("_renderercache").get("cameras")["defaultcamera"].scale = prev_scale;
				this.CameraUpdate(__redraw);
			}
		},

		GetCameraScale: function() {
			return this.get("_renderercache").get("cameras")["defaultcamera"].scale;
		},

		/**
		 * Set the absolute distance of the camera with respect to the scene center. This will keep the current
		 * scale factor but does not guarantee that the scene boundaries are enclosed within the frustum.
		 *
		 * @function
		 * @param {Number} __distance - the absolute distance of the camera with respect to the scene center
		 * @param {Boolean} [__redraw = true] If true, redraw the scene after this call
		 */
		SetCameraDistance: function(__distance, __redraw) {
			const prev_scale = this.get("_renderercache").get("cameras")["defaultcamera"].scale;
			this.CameraUpdate(__redraw, undefined, undefined, __distance);

			if (!this.get("_obj_in_frustum") && this.get("_obj_in_frustum_prev")) {
				// Object was previously in bounds but went out of frustum because
				// of these new parameters: reverting to previous state
				this.get("_renderercache").get("cameras")["defaultcamera"].scale = prev_scale;
				this.CameraUpdate(__redraw);
			}
		},

		/**
		 * Get the absolute distance of the camera with respect to the scene center.
		 *
		 * @function
		 */
		GetCameraDistance: function() {
			return this.get("_current_camera_distance");
		},

		/**
		 * Set/Get the panning offset of the camera.
		 *
		 * @function
		 * @param {Object} __panOffset {x:Number,y:Number} - An object containing two properties X, horizontal offset, and Y vertical offset
		 * @param __redraw
		 */
		SetCameraPanOffset: function(__panOffset, __redraw) {
			const prev_pan = this.get("_renderercache").get("cameras")["defaultcamera"].panOffset;
			this.get("_renderercache").get("cameras")["defaultcamera"].panOffset = __panOffset;
			this.CameraUpdate(__redraw);

			if (!this.get("_obj_in_frustum") && this.get("_obj_in_frustum_prev")) {
				//Object was previously in bounds but went out of frustum because
				//of these new parameters: reverting to previous state
				this.get("_renderercache").get("cameras")["defaultcamera"].panOffset = prev_pan;
				this.CameraUpdate(__redraw);
			}
		},

		GetCameraPanOffset: function() {
			const pan_offset = this.get("_renderercache").get("cameras")["defaultcamera"].panOffset;
			return { x: pan_offset.x, y: pan_offset.y };
		},

		/**
		 * Set up a perspective camera. The camera distance from the scene centre will be set automaticall based on the current scale setting.
		 *
		 * @function
		 * @param {Number} [__fieldOfView=45] - the field of view in degrees
		 * @param __redraw
		 */
		SetPerspectiveCamera: function(__fieldOfView, __redraw) {
			this.get("_renderercache").get("cameras")["defaultcamera"].cam_type = "PERSPECTIVE";
			this.get("_renderercache").get("cameras")["defaultcamera"].fieldOfView = __fieldOfView;
			this.CameraUpdate(__redraw);
		},

		/**
		 * Set up an orthographic camera. The viewport size will be set automaticall based on the current scale setting.
		 *
		 * @function
		 */
		SetOrthographicCamera: function(__redraw) {
			this.get("_renderercache").get("cameras")["defaultcamera"].cam_type = "ORTHOGRAPHIC";
			this.CameraUpdate(__redraw);
		},

		/**
		 * Here we recompute the near and far plane according to the bounding box of the entire scene, as opposed to just the scene focus.
		 * This needs to happen AFTER extending the scene bounding box with reflections (see CameraUpdate()).
		 *
		 */
		UpdateCamNearAndFar: function(__scene_bbox, __overrideWidth, __overrideHeight) {
			const camera = this.get("_3JScamera");
			const bbox_size = __scene_bbox.getSize();
			const bbox_center = __scene_bbox.getCenter();

			// The maximal scene radius is the diagonal length of the scene bbox
			const max_scene_radius = Math.sqrt(bbox_size.x * bbox_size.x + bbox_size.y * bbox_size.y + bbox_size.z * bbox_size.z) / 2;

			//Calculate camera position according to the view angles and its distance to the scene center
			const cam_proj_position = this.get("_3JScamera").position.clone();
			const __asset_camera = this.get("_renderercache").get("cameras")["defaultcamera"];
			cam_proj_position.addScaledVector(this.get("_cameraleftdirvec"), __asset_camera.panOffset.x);
			cam_proj_position.addScaledVector(this.get("_3JScamera").up, -__asset_camera.panOffset.y);
			const camera_distance = bbox_center
				.clone()
				.sub(cam_proj_position)
				.length();

			// NEAR PLANE: This value should always be positive, otherwise the near plane would lie behind the camera.
			// We also want to provide a minimum value in order to avoid z-fighting with very close surfaces such as highlights (see OutlineShader.js).
			// Note: For the distances we account for the scene bounding box, but for the view direction we need to use the scene focus bounding box instead!
			const view_direction = this.get("_scenemanager")
				.get("_3JSsceneBBoxFocus")
				.getCenter()
				.clone()
				.sub(cam_proj_position);
			const near = Math.min(
				camera.near,
				GetNearPlaneDistance(cam_proj_position, view_direction, __scene_bbox, this.get("_epsilon_camera"))
			);
			// When inside the scene bbox, near will be negative so we clamp it again so it stays in front of the camera
			camera.near = Math.max(near, this.get("_epsilon_camera"));
			// FAR PLANE: This plane should always be behind the entire scene, else clipping will occur for certain angles if the scene focus
			// does not contain the whole scene. To be entirely sure no clipping would ever occur we add the maximal scene radius.
			camera.far = Math.max(camera_distance + max_scene_radius, 2 * this.get("_epsilon_camera"));

			camera.updateProjectionMatrix();
		},

		GetOptimisedScaleForCurrentView: function(__overrideWidth, __overrideHeight) {
			const __asset_camera = this.get("_renderercache").get("cameras")["defaultcamera"];

			const screenwidth = __overrideWidth ? __overrideWidth : this.get("_context").clientWidth;
			const screenheight = __overrideHeight ? __overrideHeight : this.get("_context").clientHeight;
			const bbox = this.get("_scenemanager").get("_3JSsceneBBoxFocus");
			const max = bbox.max;
			const min = bbox.min;

			const fov = __asset_camera.fieldOfView;
			const aspectratio = screenwidth / screenheight;
			let box_size = bbox.getSize();
			const max_box_length = Math.sqrt(box_size.x * box_size.x + box_size.y * box_size.y + box_size.z * box_size.z);

			//Optimize scale factor to make the bbox fit the screen
			//exactly depending on the current view angles

			//Center bbox coordinates relatively to its centroid
			let min_x = min.x - (min.x + max.x) / 2;
			let min_y = min.y - (min.y + max.y) / 2;
			let min_z = min.z - (min.z + max.z) / 2;
			let max_x = max.x - (min.x + max.x) / 2;
			let max_y = max.y - (min.y + max.y) / 2;
			let max_z = max.z - (min.z + max.z) / 2;

			//Transform the 8 vertices of the bbox using opposed camera angles
			const az = WV3DMath.radians(__asset_camera.azimuthAngle);
			const el = WV3DMath.radians(-__asset_camera.elevationAngle);

			const rot_mat = new THREE.Matrix3();
			rot_mat.set(
				Math.cos(el) * Math.cos(az),
				-Math.sin(el),
				Math.cos(el) * -Math.sin(az),
				Math.sin(el) * Math.cos(az),
				Math.cos(el),
				Math.sin(el) * -Math.sin(az),
				Math.sin(az),
				0,
				Math.cos(az)
			);

			const p1 = new THREE.Vector3(min_x, min_y, min_z);
			const p2 = new THREE.Vector3(min_x, min_y, max_z);
			const p3 = new THREE.Vector3(min_x, max_y, min_z);
			const p4 = new THREE.Vector3(min_x, max_y, max_z);
			const p5 = new THREE.Vector3(max_x, min_y, min_z);
			const p6 = new THREE.Vector3(max_x, min_y, max_z);
			const p7 = new THREE.Vector3(max_x, max_y, min_z);
			const p8 = new THREE.Vector3(max_x, max_y, max_z);

			p1.applyMatrix3(rot_mat);
			p2.applyMatrix3(rot_mat);
			p3.applyMatrix3(rot_mat);
			p4.applyMatrix3(rot_mat);
			p5.applyMatrix3(rot_mat);
			p6.applyMatrix3(rot_mat);
			p7.applyMatrix3(rot_mat);
			p8.applyMatrix3(rot_mat);

			min_x = Math.min(p1.x, Math.min(p2.x, Math.min(p3.x, Math.min(p4.x, Math.min(p5.x, Math.min(p6.x, Math.min(p7.x, p8.x)))))));
			min_y = Math.min(p1.y, Math.min(p2.y, Math.min(p3.y, Math.min(p4.y, Math.min(p5.y, Math.min(p6.y, Math.min(p7.y, p8.y)))))));
			min_z = Math.min(p1.z, Math.min(p2.z, Math.min(p3.z, Math.min(p4.z, Math.min(p5.z, Math.min(p6.z, Math.min(p7.z, p8.z)))))));
			max_x = Math.max(p1.x, Math.max(p2.x, Math.max(p3.x, Math.max(p4.x, Math.max(p5.x, Math.max(p6.x, Math.max(p7.x, p8.x)))))));
			max_y = Math.max(p1.y, Math.max(p2.y, Math.max(p3.y, Math.max(p4.y, Math.max(p5.y, Math.max(p6.y, Math.max(p7.y, p8.y)))))));
			max_z = Math.max(p1.z, Math.max(p2.z, Math.max(p3.z, Math.max(p4.z, Math.max(p5.z, Math.max(p6.z, Math.max(p7.z, p8.z)))))));

			if (__asset_camera.cam_type === "PERSPECTIVE") {
				//Fit the frustum to englobe the scene bbox for perpective view
				//by computing the minimal distance to the scene according to the fov
				const fov_y = fov;
				const fov_z = aspectratio * fov;
				box_size = new THREE.Vector3(Math.abs(max_x - min_x), Math.abs(max_y - min_y), Math.abs(max_z - min_z));
				const min_camera_distance_y = Math.abs(box_size.y / (2 * Math.tan(WV3DMath.radians(fov_y) / 2)));
				const min_camera_distance_z = Math.abs(box_size.z / (2 * Math.tan(WV3DMath.radians(fov_z) / 2)));
				const min_camera_distance = Math.max(min_camera_distance_z, min_camera_distance_y);
				//This is the radius we want to use
				const radius = Math.abs(max_x - min_x) * 0.5 + min_camera_distance;

				//For this we have to calculate the matching scale in the current view configuration (perspective)
				const min_fov = Math.min(fov, aspectratio * fov);
				const min_camera_distance_normal_view = Math.abs(max_box_length / (2 * Math.tan(WV3DMath.radians(min_fov) / 2)));
				const sphere_radius = this.GetSceneInnerSphereRadius();

				// The scale factor is defined as the ratio between the (extended) min_distance and the distance to the scene bounding sphere.
				// As the distances are currently calculated relative to the inner sphere instead of the bounding sphere, we need to adjust for
				// this by adding the bounding sphere radius (max_box_length * 0.5) and subtracting the inner sphere radius.
				return (min_camera_distance_normal_view + max_box_length * 0.5 - sphere_radius) / (radius - sphere_radius);
			} else {
				//Fit the frustum to englobe the scene bbox for perpective view
				//by computing the appropriate scaling factor
				const ratiowidth = Math.abs(max_z - min_z) / screenwidth;
				const ratioheight = Math.abs(max_y - min_y) / screenheight;
				const ratio = Math.max(ratiowidth, ratioheight);
				const width_scale = ratio * screenwidth;
				const height_scale = ratio * screenheight;

				const ratiowidth_normal_view = max_box_length / screenwidth;
				const ratioheight_normal_view = max_box_length / screenheight;
				const ratio_normal_view = Math.max(ratiowidth_normal_view, ratioheight_normal_view);
				const width_scale_normal_view = ratio_normal_view * screenwidth;
				const height_scale_normal_view = ratio_normal_view * screenheight;

				const scale_x = width_scale_normal_view / width_scale;
				const scale_y = height_scale_normal_view / height_scale;

				return Math.max(scale_x, scale_y);
			}
		},

		SetOptimisedViewRendering: function(__border, __enableOptimisation, __redraw) {
			this.set("_optimised_view", __enableOptimisation);
			this.set("_optimised_view_border", __border);
			this.CameraUpdate(__redraw);
		},

		/**
		 * Set camera position according to the view angles or override with a fixed distance if provided.
		 */
		ComputeNewCamera: function(
			__asset_camera,
			__point,
			__panX,
			__panY,
			__max_box_length,
			__unit_distance,
			__override_camera_distance,
			__screenwidth,
			__screenheight
		) {
			// The following adjustments allow closer close-ups (for the perspective mode, distances should remain untouched for orthographic mode)
			// The radius is the distance between the scene focus center and the camera, while the sphere_radius is the distance between the scene focus center
			// and the furthest point the camera is ever allowed to go. In other words, the camera will never be able to enter this sphere centered around the scene focus box center.
			const bbox = this.get("_scenemanager").get("_3JSsceneBBoxFocus");
			let sphere_radius;
			if (__asset_camera.cam_type === "PERSPECTIVE") {
				// For consistent camera scale behavior, we will base our camera distances on the inner radius of the bounding box (equal for all angles, see GetSceneInnerSphereRadius())
				sphere_radius = this.GetSceneInnerSphereRadius();
				// Because we used the bounding sphere radius before and wish to maintain the exact same behavior for scale = 1 as before, we need to adjust the unit distance as following:
				__unit_distance += __max_box_length / 2 - sphere_radius;
			} else if (__asset_camera.cam_type === "ORTHOGRAPHIC") {
				// For the orthographic view there is no need to ever go inside the bounding sphere of the scene (distance does not matter)
				// Thus for safety and simplicity we use the bounding sphere radius here.
				sphere_radius = 0.5 * __max_box_length;
			}

			// Determine the minimum distance between the camera and the scene focus bounding box
			const min_radius = GetClosestCameraDistanceToBox(
				bbox,
				__asset_camera.azimuthAngle,
				__asset_camera.elevationAngle,
				this.get("_epsilon_camera")
			);

			//Compute the desired camera radius and adjust the scale accordingly (should be in sync with the radius)
			const radius = ComputeNewCameraRadiusAndScale(
				__asset_camera,
				min_radius,
				sphere_radius,
				__unit_distance,
				__override_camera_distance,
				this.get("_scale_min"),
				this.get("_scale_max")
			);
			this.set("_current_camera_distance", radius);

			// Determine the new camera position using the calculated radius
			const position = GetCameraCoordinates(__point, radius, __asset_camera.azimuthAngle, __asset_camera.elevationAngle);

			// Clamp the elevation angle
			let el_angle;
			if (__asset_camera.elevationAngle >= 90) {
				el_angle = __asset_camera.elevationAngle = 89.99;
			} else {
				el_angle = __asset_camera.elevationAngle;
			}
			// Now determine the view direction and up vector
			const leftcamvecdir = CalculateLeftCamVecDir(radius, __asset_camera.azimuthAngle, el_angle);
			const viewdir = new THREE.Vector3(__point.x, __point.y, __point.z);
			viewdir.sub(position);
			const upvector = viewdir.cross(leftcamvecdir).normalize();
			this.set("_cameraleftdirvec", leftcamvecdir);

			// To determine the near and far planes, we first determine which point on the scene focus bounding box lies closest to the camera.
			// Note: the near and far plane value will still be adjusted to fit the whole scene bbox afterwards (see ComputeCamNearAndFar)
			const closest_point = new THREE.Vector3();
			bbox.clampPoint(position, closest_point);
			const cam_to_bbox_segment = closest_point.clone().sub(position);
			// Now we adjust for the angle the distance to the closest point makes with the viewing direction, so we ensure the edges of the view would not be clipped
			// We also clamp this angle to be at most FOV/2, in those cases the closest point is no longer within the camera view anyway
			const angle_diff = Math.min(cam_to_bbox_segment.angleTo(viewdir), (__asset_camera.fieldOfView / 2) * (Math.PI / 180));
			// IMPROVEME Try to reuse GetNearPlaneDistance here
			const near = cam_to_bbox_segment.length() * Math.abs(Math.cos(angle_diff));

			// For the far plane we make use of the bounding box symmetry and just mirror the near plane
			const far = 2 * radius - near;

			let camera;
			if (__asset_camera.cam_type === "PERSPECTIVE") {
				const aspectratio = __screenwidth / __screenheight;

				camera = new THREE.PerspectiveCamera(__asset_camera.fieldOfView, aspectratio, near, far);
			} else if (__asset_camera.cam_type === "ORTHOGRAPHIC") {
				// Fit the frustum to englobe the scene bbox for orthographic view
				const ratiowidth = __max_box_length / __screenwidth;
				const ratioheight = __max_box_length / __screenheight;
				const ratio = Math.max(ratiowidth, ratioheight);
				const width_scale = (ratio * __screenwidth) / __asset_camera.scale;
				const height_scale = (ratio * __screenheight) / __asset_camera.scale;

				camera = new THREE.OrthographicCamera(width_scale / -2, width_scale / 2, height_scale / 2, height_scale / -2, near, far);
			}

			camera.up = upvector;
			camera.position.set(position.x, position.y, position.z);
			camera.lookAt(__point);

			// Set the proper panning offset
			__asset_camera.panOffset.x = __panX;
			__asset_camera.panOffset.y = __panY;

			camera.position.addScaledVector(leftcamvecdir, -__panX);
			camera.position.addScaledVector(upvector, __panY);

			return camera;
		},

		/** Set up the camera according to the current camera settings in the asset manager renderercache*/
		SetCamera: function(__scene_focus_bbox, __overrideWidth, __overrideHeight, __overrideCameraDistance) {
			const camera_asset = this.get("_renderercache").get("cameras")["defaultcamera"];

			const screenwidth = __overrideWidth ? __overrideWidth : this.get("_context").clientWidth;
			const screenheight = __overrideHeight ? __overrideHeight : this.get("_context").clientHeight;
			const point = __scene_focus_bbox.getCenter();

			const fov = camera_asset.fieldOfView;
			const aspectratio = screenwidth / screenheight;

			const box_size = __scene_focus_bbox.getSize();
			const max_box_length = Math.sqrt(box_size.x * box_size.x + box_size.y * box_size.y + box_size.z * box_size.z);

			//Fit the frustum to englobe the scene bbox for perpective view
			//by computing the minimal distance to the scene according to the fov
			const min_fov = Math.min(fov, aspectratio * fov);
			const min_camera_distance = Math.abs(max_box_length / (2 * Math.tan(WV3DMath.radians(min_fov) / 2)));

			let camera;
			if (this.get("_optimised_view")) {
				//This optimisation takes into account the projection of the bounding box onto the image plane
				//to make it fit perfectly inside it. This override the scale factor.
				camera_asset.scale = this.GetOptimisedScaleForCurrentView(__overrideWidth, __overrideHeight);
				camera = this.ComputeNewCamera(
					camera_asset,
					point,
					0,
					0,
					max_box_length,
					min_camera_distance,
					__overrideCameraDistance,
					screenwidth,
					screenheight
				);

				//In optimised view, the object will always fit into screen. Optimized labels' position will then be according to a default 0,0 panning.
				//Update labels positions at least once before we start to optimize them
				this.UpdateLabelsPositions(camera, screenwidth, screenheight);
				//Update labels position and retrieve their projected bbox.
				const pt1 = __scene_focus_bbox.min.clone();
				const pt2 = new THREE.Vector3(__scene_focus_bbox.min.x, __scene_focus_bbox.min.y, __scene_focus_bbox.max.z);
				const pt3 = new THREE.Vector3(__scene_focus_bbox.min.x, __scene_focus_bbox.max.y, __scene_focus_bbox.min.z);
				const pt4 = new THREE.Vector3(__scene_focus_bbox.min.x, __scene_focus_bbox.max.y, __scene_focus_bbox.max.z);
				const pt5 = new THREE.Vector3(__scene_focus_bbox.max.x, __scene_focus_bbox.min.y, __scene_focus_bbox.min.z);
				const pt6 = new THREE.Vector3(__scene_focus_bbox.max.x, __scene_focus_bbox.min.y, __scene_focus_bbox.max.z);
				const pt7 = new THREE.Vector3(__scene_focus_bbox.max.x, __scene_focus_bbox.max.y, __scene_focus_bbox.min.z);
				const pt8 = __scene_focus_bbox.max.clone();

				const marginscreenwidth = screenwidth / 2 - this.get("_optimised_view_border").x;
				const marginscreenheight = screenheight / 2 - this.get("_optimised_view_border").y;

				const globalBBox = new THREE.Box2();
				globalBBox.expandByPoint(pt1.clone().project(camera));
				globalBBox.expandByPoint(pt2.clone().project(camera));
				globalBBox.expandByPoint(pt3.clone().project(camera));
				globalBBox.expandByPoint(pt4.clone().project(camera));
				globalBBox.expandByPoint(pt5.clone().project(camera));
				globalBBox.expandByPoint(pt6.clone().project(camera));
				globalBBox.expandByPoint(pt7.clone().project(camera));
				globalBBox.expandByPoint(pt8.clone().project(camera));
				globalBBox.min.x *= screenwidth / 2;
				globalBBox.min.y *= screenheight / 2;
				globalBBox.max.x *= screenwidth / 2;
				globalBBox.max.y *= screenheight / 2;

				//Add user defined border
				let sceneBBsuccess =
					globalBBox.min.x >= -marginscreenwidth &&
					globalBBox.max.x <= marginscreenwidth &&
					globalBBox.min.y >= -marginscreenheight &&
					globalBBox.max.y <= marginscreenheight;
				let success =
					sceneBBsuccess &&
					this.RotateOutOfBoundsVisibleLabels(camera, globalBBox, screenwidth, screenheight, marginscreenwidth, marginscreenheight);
				const currentPanOffset = new THREE.Vector2(0, 0);
				let nb_max_iterations_opt = 100;
				while (!success && camera_asset.scale > this.get("_scale_min")) {
					const labelsize = new THREE.Vector2();
					globalBBox.getSize(labelsize);

					if (labelsize.x > marginscreenwidth || labelsize.y > marginscreenheight || !nb_max_iterations_opt) {
						//panning would not even be enough: need rescale
						camera_asset.scale -= 0.005;
					} else {
						//try to pan towards the center of the projection
						const boxcenter = globalBBox.getCenter();
						camera_asset.panOffset.x = currentPanOffset.x += Math.sign(boxcenter.x) * 1.5;
						camera_asset.panOffset.y = currentPanOffset.y += Math.sign(boxcenter.y) * 1.5;
						nb_max_iterations_opt--;
					}
					camera = this.ComputeNewCamera(
						camera_asset,
						point,
						currentPanOffset.x,
						currentPanOffset.y,
						max_box_length,
						min_camera_distance,
						undefined,
						screenwidth,
						screenheight
					);
					//Try with the current camera
					this.UpdateLabelsPositions(camera, screenwidth, screenheight);
					//Update labels position and retrieve their projected bbox.
					globalBBox.makeEmpty();
					globalBBox.expandByPoint(pt1.clone().project(camera));
					globalBBox.expandByPoint(pt2.clone().project(camera));
					globalBBox.expandByPoint(pt3.clone().project(camera));
					globalBBox.expandByPoint(pt4.clone().project(camera));
					globalBBox.expandByPoint(pt5.clone().project(camera));
					globalBBox.expandByPoint(pt6.clone().project(camera));
					globalBBox.expandByPoint(pt7.clone().project(camera));
					globalBBox.expandByPoint(pt8.clone().project(camera));
					globalBBox.min.x *= screenwidth / 2;
					globalBBox.min.y *= screenheight / 2;
					globalBBox.max.x *= screenwidth / 2;
					globalBBox.max.y *= screenheight / 2;
					sceneBBsuccess =
						globalBBox.min.x >= -marginscreenwidth &&
						globalBBox.max.x <= marginscreenwidth &&
						globalBBox.min.y >= -marginscreenheight &&
						globalBBox.max.y <= marginscreenheight;
					success =
						sceneBBsuccess &&
						this.RotateOutOfBoundsVisibleLabels(camera, globalBBox, screenwidth, screenheight, marginscreenwidth, marginscreenheight);
				}
			} else {
				camera = this.ComputeNewCamera(
					camera_asset,
					point,
					camera_asset.panOffset.x,
					camera_asset.panOffset.y,
					max_box_length,
					min_camera_distance,
					__overrideCameraDistance,
					screenwidth,
					screenheight
				);

				this.UpdateLabelsPositions(camera, screenwidth, screenheight);
			}
			this.set("_3JScamera", camera);
		},

		/**
		 * Focus camera from a node's bbox using its URL in the scene
		 * default focus of camera is on the root node of the scene
		 *
		 * @function SetCameraFocus
		 * @param {String} __nodeURL - URL identifying the scene node
		 * @param {Boolean} __redraw - If true, redraw the scene after box update - default is true
		 */
		SetCameraFocus: function(__nodeURL, __redraw) {
			const shortNodeURL = WV3DUtils.URLToKeyString(__nodeURL);
			for (
				let i = 0, keys = Object.keys(this.get("_renderercache").get("nodeviewproperties")[shortNodeURL]["threejsnodes"]), l = keys.length;
				i < l;
				++i
			) {
				//Set the focus based on the first non empty object bbox found in the first pass available (then redraw and exit)
				const key = keys[i];
				const objects = this.get("_renderercache").get("nodeviewproperties")[shortNodeURL]["threejsnodes"][key];
				const box = new THREE.Box3();
				//Set camera focus based on the bbox of all objects in all coincident groups
				for (let node_cg = 0, keys2 = Object.keys(objects), l2 = keys2.length; node_cg < l2; node_cg++) {
					const boxob = new THREE.Box3();
					boxob.setFromObject(objects[keys2[node_cg]]);
					if (!boxob.isEmpty()) box.union(boxob);
				}
				if (!box.isEmpty()) {
					this.get("_scenemanager").set("_3JSsceneBBoxFocus", box);
					this.CameraUpdate(__redraw);
					break;
				}
			}
		},

		/*= ENVIRNOMENT ROTATION =*/

		/**
		 * Set environment rotation from quaternion
		 *
		 * @function SetEnvironmentRotation
		 *
		 */
		SetEnvironmentRotation: function(__quaternion, __redraw) {
			this.get("_renderercache").get("environment_settings")["rotation"] = __quaternion.clone();

			//Iterate through all materials to update environment rotation quaternion uniform
			const applyRotation = this.UpdateMaterialEnvironmentRotationFunction(
				this.get("_renderercache").get("environment_settings")["rotation"]
			);
			this.get("_materialmanager").ApplyFctToAllMaterials(applyRotation);

			if (__redraw === undefined || __redraw) this.Render();
		},

		/**
		 * Get environment rotation quaternion
		 *
		 * @function GetEnvironmentRotation
		 *
		 */
		GetEnvironmentRotation: function() {
			return this.get("_renderercache")
				.get("environment_settings")
				["rotation"].clone();
		},

		UpdateMaterialEnvironmentRotationFunction: function(__quaternion) {
			const rotation = new THREE.Vector4(__quaternion.x, __quaternion.y, __quaternion.z, __quaternion.w);
			return function(__material) {
				if (__material) {
					if (__material.hasOwnProperty("uniforms")) {
						if (__material["uniforms"].hasOwnProperty("envMapRotQuat")) {
							__material["uniforms"]["envMapRotQuat"]["value"] = rotation;
						}
						if (__material["uniforms"].hasOwnProperty("lightPos")) {
							for (let i = 0; i < this.get("_renderercache").get("lights")["holo_light"]["Lights"].length && i < 4; ++i) {
								const light = this.get("_renderercache").get("lights")["holo_light"]["Lights"][i];
								const lightPos = new THREE.Vector3(light[0], light[1], light[2]);
								lightPos.applyQuaternion(__quaternion);
								__material["uniforms"]["lightPos"]["value"][i] = lightPos;
							}
						}
					}
				}
			}.bind(this);
		},

		/*= RENDERING FUNCTIONS =*/
		//Flip and render functions
		//The flip functions are used to render the reflected scene
		//IMPROVEME: move flip and render functions into a separate file.

		FlipMaterialSide: function(__flipped) {
			return function(__material) {
				//No need to flip if the material is double sided
				if (__material) {
					//This mecanism ensures that a material cannot be flipped more than once if it is shared between several surfaces
					if (!__material.hasOwnProperty("flipped") || __material["flipped"] !== __flipped) {
						if (__material["side"] !== THREE.DoubleSide) {
							__material["side"] = __material["side"] === THREE.FrontSide ? THREE.BackSide : THREE.FrontSide;
						}
						if (__material.hasOwnProperty("uniforms") && __material["uniforms"].hasOwnProperty("isFloorReflection")) {
							__material["uniforms"]["isFloorReflection"]["value"] = THREE.FrontSide === __flipped ? 0 : 1;
						}
					}
					__material["flipped"] = __flipped;
				}
			};
		},

		IsReflectionsRenderingEnabled: function() {
			//enable if scene loaded entirely
			//and
			//camera above ground
			//and
			//((config JSON and json refl enabled) or (config ON))
			return (
				this.get("_SceneLoaded") &&
				this.get("_scenemanager").get("_3JSsceneBBox").min.y < this.get("_3JScamera").position.y &&
				((this.get("_reflection_config") === "JSON" && this.get("_renderercache").get("reflections")["enabled"]) ||
					this.get("_reflection_config") === "ON")
			);
		},

		FlipScene: function(__scene) {
			//Apply scaling matrix * -1 on Y axis
			__scene.applyMatrix(this.get("_reflection_Y_scale"));
			//Translate the object so that the bottom of the object nearly (offset) touches the top of the reflection
			__scene.applyMatrix(this.get("_reflection_Y_translation"));
		},

		UnflipScene: function(__scene) {
			__scene.applyMatrix(this.get("_reflection_Y_opp_translation"));
			__scene.applyMatrix(this.get("_reflection_Y_scale"));
		},

		FlipAllScenes: function() {
			this.FlipScene(this.get("_scenemanager").get("_3JSsceneOp"));
			this.FlipScene(this.get("_scenemanager").get("_3JSsceneOpaqueDiffuse"));
			for (let cg = 0; cg < this.get("_scenemanager").get("_3JSsceneOpCoincidentGroupsPasses").length; ++cg) {
				const cg_id = this.get("_scenemanager").get("_3JSsceneOpCoincidentGroupsPasses")[cg];
				this.FlipScene(this.get("_scenemanager").get("_3JSsceneOpCoincidentGroups")[cg_id]);
				this.FlipScene(this.get("_scenemanager").get("_3JSsceneOpCoincidentGroupsDiffuse")[cg_id]);
			}
			this.FlipScene(this.get("_scenemanager").get("_3JSsceneTransp"));
			this.FlipScene(this.get("_scenemanager").get("_3JSsceneTranspDiffuse"));
			for (let cg = 0; cg < this.get("_scenemanager").get("_3JSsceneTranspCoincidentGroupsPasses").length; ++cg) {
				const cg_id = this.get("_scenemanager").get("_3JSsceneTranspCoincidentGroupsPasses")[cg];
				this.FlipScene(this.get("_scenemanager").get("_3JSsceneTranspCoincidentGroups")[cg_id]);
				this.FlipScene(this.get("_scenemanager").get("_3JSsceneTranspCoincidentGroupsDiffuse")[cg_id]);
			}
		},

		UnFlipAllScenes: function() {
			this.UnflipScene(this.get("_scenemanager").get("_3JSsceneOp"));
			this.UnflipScene(this.get("_scenemanager").get("_3JSsceneOpaqueDiffuse"));
			for (let cg = 0; cg < this.get("_scenemanager").get("_3JSsceneOpCoincidentGroupsPasses").length; ++cg) {
				const cg_id = this.get("_scenemanager").get("_3JSsceneOpCoincidentGroupsPasses")[cg];
				this.UnflipScene(this.get("_scenemanager").get("_3JSsceneOpCoincidentGroups")[cg_id]);
				this.UnflipScene(this.get("_scenemanager").get("_3JSsceneOpCoincidentGroupsDiffuse")[cg_id]);
			}
			this.UnflipScene(this.get("_scenemanager").get("_3JSsceneTransp"));
			this.UnflipScene(this.get("_scenemanager").get("_3JSsceneTranspDiffuse"));
			for (let cg = 0; cg < this.get("_scenemanager").get("_3JSsceneTranspCoincidentGroupsPasses").length; ++cg) {
				const cg_id = this.get("_scenemanager").get("_3JSsceneTranspCoincidentGroupsPasses")[cg];
				this.UnflipScene(this.get("_scenemanager").get("_3JSsceneTranspCoincidentGroups")[cg_id]);
				this.UnflipScene(this.get("_scenemanager").get("_3JSsceneTranspCoincidentGroupsDiffuse")[cg_id]);
			}
		},

		FlipOpaqueScene: function() {
			this.FlipScene(this.get("_scenemanager").get("_3JSsceneOp"));
			for (let cg = 0; cg < this.get("_scenemanager").get("_3JSsceneOpCoincidentGroupsPasses").length; ++cg) {
				const cg_id = this.get("_scenemanager").get("_3JSsceneOpCoincidentGroupsPasses")[cg];
				this.FlipScene(this.get("_scenemanager").get("_3JSsceneOpCoincidentGroups")[cg_id]);
			}
		},

		UnFlipOpaqueScene: function() {
			this.UnflipScene(this.get("_scenemanager").get("_3JSsceneOp"));
			for (let cg = 0; cg < this.get("_scenemanager").get("_3JSsceneOpCoincidentGroupsPasses").length; ++cg) {
				const cg_id = this.get("_scenemanager").get("_3JSsceneOpCoincidentGroupsPasses")[cg];
				this.UnflipScene(this.get("_scenemanager").get("_3JSsceneOpCoincidentGroups")[cg_id]);
			}
		},

		FlipTransparentScene: function() {
			this.FlipScene(this.get("_scenemanager").get("_3JSsceneTransp"));
			for (let cg = 0; cg < this.get("_scenemanager").get("_3JSsceneTranspCoincidentGroupsPasses").length; ++cg) {
				const cg_id = this.get("_scenemanager").get("_3JSsceneTranspCoincidentGroupsPasses")[cg];
				this.FlipScene(this.get("_scenemanager").get("_3JSsceneTranspCoincidentGroups")[cg_id]);
			}
		},

		UnFlipTransparentScene: function() {
			this.UnflipScene(this.get("_scenemanager").get("_3JSsceneTransp"));
			for (let cg = 0; cg < this.get("_scenemanager").get("_3JSsceneTranspCoincidentGroupsPasses").length; ++cg) {
				const cg_id = this.get("_scenemanager").get("_3JSsceneTranspCoincidentGroupsPasses")[cg];
				this.UnflipScene(this.get("_scenemanager").get("_3JSsceneTranspCoincidentGroups")[cg_id]);
			}
		},

		FlipLights: function() {
			for (let i = 0, cg_keys = Object.keys(this.get("_renderercache").get("lights")["dir_light"]), l = cg_keys.length; i < l; ++i) {
				const dir_light = this.get("_renderercache").get("lights")["dir_light"][cg_keys[i]];
				if (dir_light) {
					dir_light.position.set(-dir_light.position.x, dir_light.position.y, -dir_light.position.z);
					dir_light.needsUpdate = true;
				}
			}
		},

		UnFlipLights: function() {
			this.FlipLights();
		},

		RenderScene: function(__scene, __renderTarget) {
			this.get("_3JSrenderer").render(__scene, this.get("_3JScamera"), __renderTarget);
		},

		RenderOpaqueScene: function(__renderTarget) {
			//  Render opaque non coincident groups (standard) + non ard coincident groups (bags)
			this.RenderScene(this.get("_scenemanager").get("_3JSsceneOp"), __renderTarget);

			//  Render opaque ard coincident groups (ard)
			const setWriteDepth = function(__depthwrite) {
				return function(obj) {
					if (obj instanceof THREE.Mesh) {
						if (obj["material"]) {
							obj["material"]["depthWrite"] = __depthwrite;
						}
					}
				};
			};

			this.get("_3JSrenderer").autoClear = false;
			const sceneOpCoincidentGroupsPasses = this.get("_scenemanager").get("_3JSsceneOpCoincidentGroupsPasses");
			for (let cg = 0; cg < sceneOpCoincidentGroupsPasses.length; ++cg) {
				//      -Render once with zwrite disabled (update color only)
				const cg_id = sceneOpCoincidentGroupsPasses[cg];
				const sceneOpCoincidentGroups = this.get("_scenemanager").get("_3JSsceneOpCoincidentGroups");
				sceneOpCoincidentGroups[cg_id].traverse(setWriteDepth(false));
				this.RenderScene(sceneOpCoincidentGroups[cg_id], __renderTarget);

				//      -Render once with colorwrite disabled (update depth only)
				this.get("_3JSrenderer").context.colorMask(false, false, false, false);
				sceneOpCoincidentGroups[cg_id].traverse(setWriteDepth(true));
				this.RenderScene(sceneOpCoincidentGroups[cg_id], __renderTarget);
				this.get("_3JSrenderer").context.colorMask(true, true, true, true);
			}
			this.get("_3JSrenderer").autoClear = true;
		},

		RenderOpaqueSceneDepth: function(__renderTarget, __overrideMaterial) {
			//Instead of having to build a scene only for rendering it with our depth material
			//we borrow the single shading pass opaque diffuse scene and override the material

			const sceneOpaqueDiffuse = this.get("_scenemanager").get("_3JSsceneOpaqueDiffuse");
			sceneOpaqueDiffuse.overrideMaterial = __overrideMaterial
				? __overrideMaterial
				: this.get("_renderercache").get("materials")["D0ShaderMaterial"];
			this.RenderScene(sceneOpaqueDiffuse, __renderTarget);
			sceneOpaqueDiffuse.overrideMaterial = null;

			this.get("_3JSrenderer").autoClear = false;
			const sceneOpCoincidentGroupsPasses = this.get("_scenemanager").get("_3JSsceneOpCoincidentGroupsPasses");
			for (let cg = 0; cg < sceneOpCoincidentGroupsPasses.length; ++cg) {
				const cg_id = sceneOpCoincidentGroupsPasses[cg];
				const sceneOpCoincidentGroupsDiffuse = this.get("_scenemanager").get("_3JSsceneOpCoincidentGroupsDiffuse");
				sceneOpCoincidentGroupsDiffuse[cg_id].overrideMaterial = __overrideMaterial
					? __overrideMaterial
					: this.get("_renderercache").get("materials")["D0ShaderMaterial"];
				this.RenderScene(sceneOpCoincidentGroupsDiffuse[cg_id], __renderTarget);
				sceneOpCoincidentGroupsDiffuse[cg_id].overrideMaterial = null;
			}
			this.get("_3JSrenderer").autoClear = true;
		},

		RenderOpaqueSceneDiffuse: function(__renderTarget) {
			//  Render opaque non coincident groups (standard) + non ard coincident groups (bags)
			this.RenderScene(this.get("_scenemanager").get("_3JSsceneOpaqueDiffuse"), __renderTarget);

			//  Render opaque ard coincident groups (ard)
			const setWriteDepth = function(__depthwrite) {
				return function(obj) {
					if (obj instanceof THREE.Mesh) {
						if (obj["material"]) {
							obj["material"]["depthWrite"] = __depthwrite;
						}
					}
				};
			};

			this.get("_3JSrenderer").autoClear = false;
			const sceneOpCoincidentGroupsPasses = this.get("_scenemanager").get("_3JSsceneOpCoincidentGroupsPasses");
			for (let cg = 0; cg < sceneOpCoincidentGroupsPasses.length; ++cg) {
				//      -Render once with zwrite disabled (update color only)
				const cg_id = sceneOpCoincidentGroupsPasses[cg];
				const sceneOpCoincidentGroupsDiffuse = this.get("_scenemanager").get("_3JSsceneOpCoincidentGroupsDiffuse");
				sceneOpCoincidentGroupsDiffuse[cg_id].traverse(setWriteDepth(false));
				this.RenderScene(sceneOpCoincidentGroupsDiffuse[cg_id], __renderTarget);

				//      -Render once with colorwrite disabled (update depth only)
				this.get("_3JSrenderer").context.colorMask(false, false, false, false);
				sceneOpCoincidentGroupsDiffuse[cg_id].traverse(setWriteDepth(true));
				this.RenderScene(sceneOpCoincidentGroupsDiffuse[cg_id], __renderTarget);
				this.get("_3JSrenderer").context.colorMask(true, true, true, true);
			}
			this.get("_3JSrenderer").autoClear = true;
		},

		RenderTransparentScene: function(__renderTarget, __sceneGraphCallbackFct, __writedepth) {
			//  Render opaque ard coincident groups (ard)
			const setWriteDepth = function(__depthwrite) {
				return function(obj) {
					if (obj instanceof THREE.Mesh) {
						if (obj["material"]) {
							obj["material"]["depthWrite"] = __depthwrite && __writedepth;
						}
					}
				};
			};

			//  Render transparent non coincident groups (standard) + non ard coincident groups (bags)
			const sceneTransp = this.get("_scenemanager").get("_3JSsceneTransp");
			if (typeof __sceneGraphCallbackFct === "function") {
				sceneTransp.traverse(__sceneGraphCallbackFct);
			}
			sceneTransp.traverse(setWriteDepth(false));
			this.RenderScene(sceneTransp, __renderTarget);

			this.get("_3JSrenderer").autoClear = false;
			const sceneTranspCoincidentGroupsPasses = this.get("_scenemanager").get("_3JSsceneTranspCoincidentGroupsPasses");
			for (let cg = 0; cg < sceneTranspCoincidentGroupsPasses.length; ++cg) {
				const cg_id = sceneTranspCoincidentGroupsPasses[cg];
				const sceneTranspCoincidentGroups = this.get("_scenemanager").get("_3JSsceneTranspCoincidentGroups");
				if (typeof __sceneGraphCallbackFct === "function") {
					sceneTranspCoincidentGroups[cg_id].traverse(__sceneGraphCallbackFct);
				}

				//      -Render once with zwrite disabled (update color only)
				sceneTranspCoincidentGroups[cg_id].traverse(setWriteDepth(false));
				this.RenderScene(sceneTranspCoincidentGroups[cg_id], __renderTarget);

				//      -Render once with colorwrite disabled (update depth only)
				this.get("_3JSrenderer").context.colorMask(false, false, false, false);
				sceneTranspCoincidentGroups[cg_id].traverse(setWriteDepth(true));
				this.RenderScene(sceneTranspCoincidentGroups[cg_id], __renderTarget);
				this.get("_3JSrenderer").context.colorMask(true, true, true, true);
			}
			this.get("_3JSrenderer").autoClear = true;
		},

		RenderTransparentSceneDepth: function(__renderTarget, __overrideMaterial) {
			//Instead of having to build a scene only for rendering it with our depth material
			//we borrow the single shading pass transparent diffuse pass and override the material

			const sceneTranspDiffuse = this.get("_scenemanager").get("_3JSsceneTranspDiffuse");
			sceneTranspDiffuse.overrideMaterial = __overrideMaterial;
			this.RenderScene(sceneTranspDiffuse, __renderTarget);
			sceneTranspDiffuse.overrideMaterial = null;

			this.get("_3JSrenderer").autoClear = false;
			const sceneTranspCoincidentGroupsPasses = this.get("_scenemanager").get("_3JSsceneTranspCoincidentGroupsPasses");
			for (let cg = 0; cg < sceneTranspCoincidentGroupsPasses.length; ++cg) {
				const cg_id = sceneTranspCoincidentGroupsPasses[cg];
				const sceneTranspCoincidentGroupsDiffuse = this.get("_scenemanager").get("_3JSsceneTranspCoincidentGroupsDiffuse");
				sceneTranspCoincidentGroupsDiffuse[cg_id].overrideMaterial = __overrideMaterial;
				this.RenderScene(sceneTranspCoincidentGroupsDiffuse[cg_id], __renderTarget);
				sceneTranspCoincidentGroupsDiffuse[cg_id].overrideMaterial = null;
			}
			this.get("_3JSrenderer").autoClear = true;
		},

		RenderTransparentSceneDiffuse: function(__renderTarget, __sceneGraphCallbackFct) {
			//  Render opaque non coincident groups (standard) + non ard coincident groups (bags)
			const sceneTranspDiffuse = this.get("_scenemanager").get("_3JSsceneTranspDiffuse");
			sceneTranspDiffuse.traverse(__sceneGraphCallbackFct);
			this.RenderScene(sceneTranspDiffuse, __renderTarget);

			//  Render opaque ard coincident groups (ard)
			const setWriteDepth = function(__depthwrite) {
				return function(obj) {
					if (obj instanceof THREE.Mesh) {
						if (obj["material"]) {
							obj["material"]["depthWrite"] = __depthwrite;
						}
					}
				};
			};

			this.get("_3JSrenderer").autoClear = false;
			const sceneTranspCoincidentGroupsPasses = this.get("_scenemanager").get("_3JSsceneTranspCoincidentGroupsPasses");
			for (let cg = 0; cg < sceneTranspCoincidentGroupsPasses.length; ++cg) {
				const cg_id = sceneTranspCoincidentGroupsPasses[cg];
				const sceneTranspCoincidentGroupsDiffuse = this.get("_scenemanager").get("_3JSsceneTranspCoincidentGroupsDiffuse");
				sceneTranspCoincidentGroupsDiffuse[cg_id].traverse(__sceneGraphCallbackFct);

				//      -Render once with zwrite disabled (update color only)
				sceneTranspCoincidentGroupsDiffuse[cg_id].traverse(setWriteDepth(false));
				this.RenderScene(sceneTranspCoincidentGroupsDiffuse[cg_id], __renderTarget);

				//      -Render once with colorwrite disabled (update depth only)
				this.get("_3JSrenderer").context.colorMask(false, false, false, false);
				sceneTranspCoincidentGroupsDiffuse[cg_id].traverse(setWriteDepth(true));
				this.RenderScene(sceneTranspCoincidentGroupsDiffuse[cg_id], __renderTarget);
				this.get("_3JSrenderer").context.colorMask(true, true, true, true);
			}
			this.get("_3JSrenderer").autoClear = true;
		},

		SetSizeOfRTSrc: function(__width, __height) {
			return function(obj) {
				if (obj instanceof THREE.Mesh) {
					if (obj["material"].hasOwnProperty("uniforms")) {
						if (obj["material"]["uniforms"].hasOwnProperty("canvas_size")) {
							obj["material"]["uniforms"]["canvas_size"]["value"] = new THREE.Vector2(__width, __height);
						}
					}
				}
			};
		},

		SetAsRTSrc: function(__width, __height, __rTSrc1, __rTSrc2) {
			const nb_arg = Math.max(arguments.length - 2, 0); //Remove width and height from the count
			return function(obj) {
				if (obj instanceof THREE.Mesh) {
					if (obj["material"].hasOwnProperty("uniforms")) {
						if (
							obj["material"]["uniforms"].hasOwnProperty("DX1RT") &&
							obj["material"]["uniforms"].hasOwnProperty("DX2RT") &&
							obj["material"]["uniforms"].hasOwnProperty("NBRTSRC") &&
							obj["material"]["uniforms"].hasOwnProperty("canvas_size")
						) {
							obj["material"]["uniforms"]["canvas_size"]["value"] = new THREE.Vector2(__width, __height);
							obj["material"]["uniforms"]["DX1RT"]["value"] = __rTSrc1 ? __rTSrc1.texture : null;
							obj["material"]["uniforms"]["DX2RT"]["value"] = __rTSrc2 ? __rTSrc2.texture : null;
							obj["material"]["uniforms"]["NBRTSRC"]["value"] = nb_arg;
						}
						if (obj["material"]["uniforms"].hasOwnProperty("transparent")) {
							obj["material"]["uniforms"]["transparent"]["value"] = __rTSrc1 ? 1 : 0;
						}
					}
				}
			};
		},

		RenderDepthPeel: function(
			__drawHardEdgeHighlights,
			__cameraRTT,
			__screen_plane,
			__width,
			__height,
			__colorStackRT,
			__transparencyStackRT,
			__opaqueDepthRT,
			__d1RT,
			__d2RT,
			__peelColorRT,
			__opaqueBackgroundMaskRT,
			__blackClearDepthColor,
			__whiteClearDepthColor
		) {
			// 1) Render opaque objects with an override material that writes depth as grey/color (D0)
			this.get("_3JSrenderer").setClearColor(__blackClearDepthColor, 1);
			this.RenderOpaqueSceneDepth(__opaqueDepthRT);

			// 2A) Render 1st peel depth in __colorStackRT, using OpaqueDepthRT. This enables a later pass to use blending, needed for coincident surfaces.
			this.get("_3JSrenderer").autoClearDepth = true;
			this.get("_renderercache").get("materials")["DepthOnlyShaderMaterial"]["uniforms"]["DX1RT"]["value"] = __opaqueDepthRT.texture;
			this.get("_renderercache").get("materials")["DepthOnlyShaderMaterial"]["uniforms"]["NBRTSRC"]["value"] = 1;

			this.RenderTransparentSceneDepth(__colorStackRT, this.get("_renderercache").get("materials")["DepthOnlyShaderMaterial"]);
			this.get("_3JSrenderer").autoClearDepth = false;

			// 2B: Render 1st peel color in __colorStackRT, using __opaqueDepthRT.
			this.get("_3JSrenderer").setClearColor(__blackClearDepthColor, 1); // color peel must have black BG and Alpha of background
			this.RenderTransparentScene(__colorStackRT, this.SetAsRTSrc(__width, __height, __opaqueDepthRT), true);

			// 2C) Render 1st peel depth in __transparencyStackRT, using __opaqueDepthRT.
			this.get("_3JSrenderer").autoClearDepth = true;
			this.RenderTransparentSceneDepth(__transparencyStackRT, this.get("_renderercache").get("materials")["DepthOnlyShaderMaterial"]);

			// 2D) Render 1st peel transparency in __transparencyStackRT, using __opaqueDepthRT.
			this.get("_3JSrenderer").autoClearDepth = false;
			this.get("_3JSrenderer").setClearColor(__whiteClearDepthColor, 1); // start 100% transparent, so background shows through unused pixels

			this.RenderTransparentSceneDiffuse(__transparencyStackRT, this.SetAsRTSrc(__width, __height, __opaqueDepthRT));

			this.get("_3JSrenderer").autoClearDepth = true;

			//Force number of peels for debug
			// this.set("_current_nb_peels", 1 );//remove after debug
			if (this.get("_current_nb_peels") > 1 && this.get("_SceneLoaded")) {
				// 3) RENDER NEXT PEELS. For this we need the depth D1 of the first (previous) peel.
				// Render transp objects with an override material using GL_LEQUAL and D0 to render depth into D1
				// here we draw only the fragments that are in front of opaque surfaces
				this.get("_3JSrenderer").setClearColor(__blackClearDepthColor, 1);
				this.get("_renderercache").get("materials")["D1ShaderMaterial"]["uniforms"]["D0RT"]["value"] = __opaqueDepthRT.texture;

				this.RenderTransparentSceneDepth(__d1RT, this.get("_renderercache").get("materials")["D1ShaderMaterial"]);

				this.get("_renderercache").get("materials")["D2ShaderMaterial"]["uniforms"]["DX1RT"]["value"] = __opaqueDepthRT.texture;

				this.get("_3JSrenderer").setClearColor(__blackClearDepthColor, 1); // color peel must have black BG
				for (
					let i = 0;
					i < this.get("_current_nb_peels") - 1;
					++i //This should be removed when occlusion query is implemented (with a max_nb_peel for safety)
				) {
					// 4A) Render the next peels depth in __peelColorRT. This enables a later pass to use blending, needed for coincident surfaces
					this.get("_3JSrenderer").autoClearDepth = true;
					this.get("_3JSrenderer").autoClearColor = true;
					this.get("_renderercache").get("materials")["DepthOnlyShaderMaterial"]["uniforms"]["DX2RT"]["value"] =
						i % 2 ? __d2RT.texture : __d1RT.texture;
					this.get("_renderercache").get("materials")["DepthOnlyShaderMaterial"]["uniforms"]["NBRTSRC"]["value"] = 2;

					this.RenderTransparentSceneDepth(__peelColorRT, this.get("_renderercache").get("materials")["DepthOnlyShaderMaterial"]);
					this.get("_3JSrenderer").autoClearDepth = false;

					// 4B Render the next peels color in __peelColorRT, using __opaqueDepthRT.
					this.RenderTransparentScene(__peelColorRT, this.SetAsRTSrc(__width, __height, __opaqueDepthRT, i % 2 ? __d2RT : __d1RT), true);
					this.get("_3JSrenderer").autoClearDepth = true;

					// 5) Blend the current peel color with the color stack, using the transparency stack
					this.get("_3JSrenderer").autoClearColor = false;
					this.get("_renderercache").get("materials")["ColorBlendShaderMaterial"]["uniforms"]["CX1RT"]["value"] = __peelColorRT.texture; //current peel
					this.get("_renderercache").get("materials")["ColorBlendShaderMaterial"]["uniforms"]["TransparencyStack"]["value"] =
						__transparencyStackRT.texture;
					this.get("_renderercache").get("materials")["ColorBlendShaderMaterial"]["uniforms"]["EnableBGBlend"]["value"] = 0;

					this.get("_3JSrenderer").render(__screen_plane, __cameraRTT, __colorStackRT);
					this.get("_3JSrenderer").autoClearColor = true;

					// 6A) Render the next peel depth in the transparency stack, without overwriting the color. Needed for coincident surfaces.
					this.get("_3JSrenderer").autoClearDepth = true;
					this.get("_3JSrenderer").autoClearColor = false;

					this.RenderTransparentSceneDepth(__transparencyStackRT, this.get("_renderercache").get("materials")["DepthOnlyShaderMaterial"]);
					this.get("_3JSrenderer").autoClearDepth = false;

					// 6B) Blend the next peel's transparency with the transparency stack. Blending is done via the material blending.
					this.get("_3JSrenderer").autoClearColor = false;

					this.RenderTransparentSceneDiffuse(
						__transparencyStackRT,
						this.SetAsRTSrc(__width, __height, __opaqueDepthRT, i % 2 ? __d2RT : __d1RT)
					);
					this.get("_3JSrenderer").autoClearColor = true;
					this.get("_3JSrenderer").autoClearDepth = true;

					// 7) Render transp objects with an override material using GL_LEQUAL and D0 to render depth into D1
					//here we draw only the fragments that have depth strictly between D0 and D1- with LEQUAL depth function
					this.get("_renderercache").get("materials")["D2ShaderMaterial"]["uniforms"]["DX2RT"]["value"] =
						i % 2 ? __d2RT.texture : __d1RT.texture;

					this.RenderTransparentSceneDepth(i % 2 ? __d1RT : __d2RT, this.get("_renderercache").get("materials")["D2ShaderMaterial"]);
				}
			}

			////Render opaque objects + hard edges highlights into PeelColor
			this.get("_3JSrenderer").setClearColor(__blackClearDepthColor, 1); //Restore clear color
			this.RenderOpaqueScene(__peelColorRT);
			if (__drawHardEdgeHighlights) {
				this.get("_3JSrenderer").autoClear = false;
				this.get("_3JSrenderer").render(this.get("_scenemanager").get("_3JSsceneHightlights"), this.get("_3JScamera"), __peelColorRT);
				this.get("_3JSrenderer").autoClear = true;
			}

			// 9) RENDER ALPHA OF OPAQUE PEEL.
			this.get("_3JSrenderer").setClearColor(__whiteClearDepthColor, 1); // start 100% opaque, so background shows through unused pixels
			this.RenderOpaqueSceneDiffuse(__opaqueBackgroundMaskRT);
			if (__drawHardEdgeHighlights) {
				//Also draw highlight edges to update opaque mask
				this.get("_3JSrenderer").autoClear = false;
				this.get("_scenemanager").get("_3JSsceneHightlights").overrideMaterial = this.get("_renderercache").get("materials")[
					"blackhighlight"
				];
				this.get("_3JSrenderer").render(
					this.get("_scenemanager").get("_3JSsceneHightlights"),
					this.get("_3JScamera"),
					__opaqueBackgroundMaskRT
				);
				this.get("_scenemanager").get("_3JSsceneHightlights").overrideMaterial = null;
				this.get("_3JSrenderer").autoClear = true;
			}
		},

		/** Render the scene */
		//IMPROVEME: Split this function into sub-functions
		//IMPROVEME: Move some initializations, such as cameraRTT, and screen plane geometry
		Render: function(__offscreen, __overrideWidth, __overrideHeight) {
			//If the scenes do not contain any object, rendering them is useless (return)
			//for debug only, or if we only want to render the scene only when finished loading (reduces loading time)
			//if(!this.get('_SceneLoaded')) return;
			if (this.get("_scenemanager").get("_emptyOpaqueScene") && this.get("_scenemanager").get("_emptyTransparentScene")) return;

			const width = __overrideWidth ? __overrideWidth : this.get("_context").clientWidth;
			const height = __overrideHeight ? __overrideHeight : this.get("_context").clientHeight;

			if (__offscreen || this.get("_resetRenderTargets")) {
				//If the camera was not initialized (no bounding box, i.e. no geometry)
				if (!this.CameraUpdate(false, width, height)) return;
				this.InitializeRenderTargets(width, height);
				this.ResetScreenQuads();
				this.set("_3JSHUDcamera", new THREE.OrthographicCamera(-width / 2, width / 2, height / 2, -height / 2, 0, 10));
				this.get("_3JSrenderer").setSize(width, height);
				this.set("_resetRenderTargets", false);
			}

			const enableHighlights = this.get("_renderercache").get("materials")["enablehighlights"];
			const enableHardEdgeHighlights = this.get("_renderercache").get("materials")["enablehardedgehighlights"];
			const enableHorizonHighlights = this.get("_renderercache").get("materials")["enablehorizonhighlights"];
			const enableLabels = this.get("_renderercache").get("labels")["visible"];

			//Prepare quad geometry to render it with any render target attached as texture
			const ColorStackRT = this.get("_ColorStackRT");
			const clipPlaneDistance = 100000;
			const cameraRTT = new THREE.OrthographicCamera(-width / 2, width / 2, height / 2, -height / 2, -clipPlaneDistance, clipPlaneDistance);
			cameraRTT.position.z = 10000;
			const screen_plane = new THREE.PlaneGeometry(width, height);
			// Because we're using an orthographic camera, it doesn't matter what the Z-distance is for blending, copying etc.
			//  We put it near the far clipplane for the annotations alpha fix.
			for (let i = 0; i < 4; i++) screen_plane.vertices[i].z = -clipPlaneDistance + cameraRTT.position.z;

			//Compute backdrop (background image) parameters if needed
			let backdropTex, bd_scaleX, bd_scaleY, bd_offsetX, bd_offsetY;
			if (
				this.get("_renderercache").get("textures")["backdrop"]["loaded"] &&
				this.get("_renderercache").get("textures")["backdrop"]["enabled"]
			) {
				backdropTex = this.get("_renderercache").get("textures")["backdrop"]["data"];

				let bd_orig_width = this.get("_renderercache").get("textures")["backdrop"]["original_width"];
				let bd_orig_height = this.get("_renderercache").get("textures")["backdrop"]["original_height"];
				const bd_nat_width = backdropTex["image"]["naturalWidth"];
				const bd_nat_height = backdropTex["image"]["naturalHeight"];
				if (!bd_orig_width) bd_orig_width = bd_nat_width;
				if (!bd_orig_height) bd_orig_height = bd_nat_height;

				const canvasRatio = width / height;
				const backdropRatio = bd_nat_width / bd_orig_height;

				bd_scaleX = 1;
				bd_scaleY = 1;
				if (canvasRatio < backdropRatio) bd_scaleX = canvasRatio / backdropRatio;
				else bd_scaleY = backdropRatio / canvasRatio;
				bd_scaleX *= bd_orig_width / bd_nat_width;
				bd_scaleY *= bd_orig_height / bd_nat_height;

				bd_offsetX = -0.5 * bd_scaleX + (0.5 * bd_orig_width) / bd_nat_width;
				bd_offsetY = -0.5 * bd_scaleY + (0.5 * bd_orig_height) / bd_nat_height;
			}

			WV3DFPSCounter.Update();
			const current_fps = WV3DFPSCounter.GetCountPerMinute();
			// Uncomment these three lines to check performances - comment at the end of the story
			if (current_fps !== this.get("_current_fps")) {
				LogMessage.Log("Average framerate: " + current_fps);
				this.set("_current_fps", current_fps);
			}

			//Render scene, with depth peeling algorithm if containing some transparent elements and using Visualizer quality
			const blackColor = new THREE.Color(0, 0, 0);
			if (
				!this.get("_scenemanager").get("_emptyTransparentScene") &&
				this.get("_renderercache").get("_RenderingQuality") === "Visualizer"
			) {
				if (this.get("_interactiveFrameRate") && this.get("_enableInteractionOptimization")) {
					//Calculate interactively the number of peels according to the performances
					//Current average window is set to 2 seconds in WV3DFPSCounter

					if (current_fps > this.get("_target_fps_max")) {
						//We are loose on performance, let's increase the number of peels
						this.set("_current_nb_peels", this.get("_current_nb_peels") + 1);
					} else if (current_fps < this.get("_target_fps_min")) {
						//We are losing performance, let's decrease the number of peels
						this.set("_current_nb_peels", this.get("_current_nb_peels") - 1);
					}
					//Clamp the number of peels
					this.set(
						"_current_nb_peels",
						Math.max(Math.min(this.get("_current_nb_peels"), this.get("_nb_max_peels")), this.get("_nb_min_peels"))
					);
				} else {
					this.set("_current_nb_peels", this.get("_nb_max_peels"));
				}

				//0) Initialize render targets
				const OpaqueDepthRT = this.get("_OpaqueDepthRT");
				const D1RT = this.get("_D1RT");
				const D2RT = this.get("_D2RT");
				const PeelColorRT = this.get("_PeelColorRT");
				const TransparencyStackRT = this.get("_TransparencyStackRT");
				const OpaqueBackgroundMaskRT = this.get("_OpaqueBackgroundMaskRT");

				const MirrorColorStackRT = this.get("_MirrorColorStackRT");
				const MirrorColorStackRT2 = this.get("_MirrorColorStackRT2");
				const MirrorOpaqueDepthRT = this.get("_MirrorOpaqueDepthRT");
				const MirrorPeelColorRT = this.get("_MirrorPeelColorRT");
				const MirrorTransparencyStackRT = this.get("_MirrorTransparencyStackRT");
				const MirrorOpaqueBackgroundMaskRT = this.get("_MirrorOpaqueBackgroundMaskRT");

				this.get("_scenemanager")
					.get("_3JSsceneOp")
					.traverse(this.SetSizeOfRTSrc(width, height));
				this.get("_scenemanager")
					.get("_3JSsceneTransp")
					.traverse(this.SetSizeOfRTSrc(width, height));
				this.get("_scenemanager")
					.get("_3JSsceneTranspDiffuse")
					.traverse(this.SetSizeOfRTSrc(width, height));
				this.get("_scenemanager")
					.get("_3JSsceneOpaqueDiffuse")
					.traverse(this.SetSizeOfRTSrc(width, height));
				this.get("_3JSrenderer").autoClear = true;

				const whiteClearDepthColor = new THREE.Color(1, 1, 1);
				const blackClearDepthColor = new THREE.Color(0, 0, 0);
				const backgroundColor = this.get("_backgroundColor");
				const backgroundAlpha = this.get("_backgroundAlpha");

				this.get("_renderercache").get("materials")["CopyShaderMaterial"]["uniforms"]["canvas_size"]["value"] = new THREE.Vector2(
					width,
					height
				);
				const colorBlendShaderMaterialUniforms = this.get("_renderercache").get("materials")["ColorBlendShaderMaterial"]["uniforms"];
				colorBlendShaderMaterialUniforms["canvas_size"]["value"] = new THREE.Vector2(width, height);
				colorBlendShaderMaterialUniforms["EnableBGBlend"]["value"] = 0; //disable background blending for next uses
				colorBlendShaderMaterialUniforms["EnableBackdrop"]["value"] = 0;
				colorBlendShaderMaterialUniforms["Backdrop"]["value"] = null;
				colorBlendShaderMaterialUniforms["CX1RT"]["value"] = null;
				colorBlendShaderMaterialUniforms["TransparencyStack"]["value"] = null;
				colorBlendShaderMaterialUniforms["OpaqueRT"]["value"] = null;
				colorBlendShaderMaterialUniforms["ColorStackRT"]["value"] = null;
				this.get("_renderercache").get("materials")["DepthOnlyShaderMaterial"]["uniforms"]["canvas_size"]["value"] = new THREE.Vector2(
					width,
					height
				);
				this.get("_renderercache").get("materials")["D1ShaderMaterial"]["uniforms"]["canvas_size"]["value"] = new THREE.Vector2(
					width,
					height
				);
				this.get("_renderercache").get("materials")["D2ShaderMaterial"]["uniforms"]["canvas_size"]["value"] = new THREE.Vector2(
					width,
					height
				);

				if (!this.get("_screenQuadSceneColorBlend")) {
					const screen_quad_color_blend = new THREE.Mesh(
						screen_plane,
						this.get("_renderercache").get("materials")["ColorBlendShaderMaterial"]
					);
					const screen_quad_scene_color_blend = new THREE.Scene();
					screen_quad_scene_color_blend.add(screen_quad_color_blend);
					this.set("_screenQuadSceneColorBlend", screen_quad_scene_color_blend);
				}

				if (this.IsReflectionsRenderingEnabled()) {
					//Flip all scenes
					this.FlipAllScenes();

					//Flip GL_FRONT to GL_BACK and vice versa to compensate scaling *-1 on Y axis
					this.get("_materialmanager").ApplyFctToAllMaterials(this.FlipMaterialSide(THREE.BackSide));

					//Flip lights positions (for designer quality), otherwise lighting is wrong in reflection (since normal hasn't been flipped)
					this.FlipLights();

					//Depth peel for reflected scenes
					this.RenderDepthPeel(
						false,
						cameraRTT,
						this.get("_screenQuadSceneColorBlend"),
						width,
						height,
						MirrorColorStackRT,
						MirrorTransparencyStackRT,
						MirrorOpaqueDepthRT,
						D1RT,
						D2RT,
						MirrorPeelColorRT,
						MirrorOpaqueBackgroundMaskRT,
						blackClearDepthColor,
						whiteClearDepthColor
					);

					//Unflip all scenes
					this.UnFlipAllScenes();

					this.get("_materialmanager").ApplyFctToAllMaterials(this.FlipMaterialSide(THREE.FrontSide));

					this.UnFlipLights();
				}
				//Depth peel for normal scene
				this.RenderDepthPeel(
					enableHighlights && enableHardEdgeHighlights,
					cameraRTT,
					this.get("_screenQuadSceneColorBlend"),
					width,
					height,
					ColorStackRT,
					TransparencyStackRT,
					OpaqueDepthRT,
					D1RT,
					D2RT,
					PeelColorRT,
					OpaqueBackgroundMaskRT,
					blackClearDepthColor,
					whiteClearDepthColor
				);

				//Blend with opaque color !
				//Clear alpha channel
				this.get("_3JSrenderer")
					.getContext()
					.colorMask(false, false, false, true);
				this.get("_3JSrenderer").setClearColor(blackClearDepthColor, 0);
				this.get("_3JSrenderer").clearTarget(ColorStackRT, true, false, false);
				this.get("_3JSrenderer")
					.getContext()
					.colorMask(true, true, true, true);
				this.get("_3JSrenderer").setClearColor(blackClearDepthColor, 1);

				this.get("_3JSrenderer").autoClearColor = false;

				this.get("_renderercache").get("materials")["ColorBlendShaderMaterial"]["uniforms"]["EnableBGBlend"]["value"] = 1;
				this.get("_renderercache").get("materials")["ColorBlendShaderMaterial"]["uniforms"]["BGColor"]["value"] = backgroundColor;
				this.get("_renderercache").get("materials")["ColorBlendShaderMaterial"]["uniforms"]["BGAlpha"]["value"] = backgroundAlpha;
				this.get("_renderercache").get("materials")["ColorBlendShaderMaterial"]["uniforms"]["Backdrop"]["value"] = backdropTex;
				this.get("_renderercache").get("materials")["ColorBlendShaderMaterial"]["uniforms"]["EnableBackdrop"]["value"] = backdropTex
					? 1
					: 0;
				this.get("_renderercache").get("materials")["ColorBlendShaderMaterial"]["uniforms"]["EnableBackdropAlpha"]["value"] = backdropTex
					? 0
					: 1;
				this.get("_renderercache").get("materials")["ColorBlendShaderMaterial"]["uniforms"]["Backdrop_scale"]["value"] = new THREE.Vector2(
					bd_scaleX,
					bd_scaleY
				);
				this.get("_renderercache").get("materials")["ColorBlendShaderMaterial"]["uniforms"]["Backdrop_offset"]["value"] = new THREE.Vector2(
					bd_offsetX,
					bd_offsetY
				);

				if (this.IsReflectionsRenderingEnabled()) {
					this.get("_3JSrenderer").autoClearColor = true;
					this.get("_3JSrenderer").setClearColor(blackClearDepthColor, 0);

					//Render the reflection first onto standard background
					this.get("_renderercache").get("materials")["ColorBlendShaderMaterial"]["uniforms"]["EnableMirror"]["value"] = 1;
					this.get("_renderercache").get("materials")["ColorBlendShaderMaterial"]["uniforms"]["CX1RT"]["value"] = MirrorPeelColorRT.texture;
					this.get("_renderercache").get("materials")["ColorBlendShaderMaterial"]["uniforms"]["TransparencyStack"]["value"] =
						MirrorTransparencyStackRT.texture;
					this.get("_renderercache").get("materials")["ColorBlendShaderMaterial"]["uniforms"]["OpaqueRT"]["value"] =
						MirrorOpaqueBackgroundMaskRT.texture;
					this.get("_renderercache").get("materials")["ColorBlendShaderMaterial"]["uniforms"]["ColorStackRT"]["value"] =
						MirrorColorStackRT.texture;
					this.get("_3JSrenderer").render(this.get("_screenQuadSceneColorBlend"), cameraRTT, MirrorColorStackRT2);

					//Render normal scene using previous render target as background
					this.get("_renderercache").get("materials")["ColorBlendShaderMaterial"]["uniforms"]["EnableMirror"]["value"] = 0;
					this.get("_renderercache").get("materials")["ColorBlendShaderMaterial"]["uniforms"]["EnableBackdrop"]["value"] = 1;
					this.get("_renderercache").get("materials")["ColorBlendShaderMaterial"]["uniforms"]["EnableBackdropAlpha"]["value"] = 1;
					this.get("_renderercache").get("materials")["ColorBlendShaderMaterial"]["uniforms"]["Backdrop_scale"][
						"value"
					] = new THREE.Vector2(1, 1);
					this.get("_renderercache").get("materials")["ColorBlendShaderMaterial"]["uniforms"]["Backdrop_offset"][
						"value"
					] = new THREE.Vector2(0, 0);
					this.get("_renderercache").get("materials")["ColorBlendShaderMaterial"]["uniforms"]["Backdrop"]["value"] =
						MirrorColorStackRT2.texture;
					this.get("_renderercache").get("materials")["ColorBlendShaderMaterial"]["uniforms"]["CX1RT"]["value"] = PeelColorRT.texture;
					this.get("_renderercache").get("materials")["ColorBlendShaderMaterial"]["uniforms"]["TransparencyStack"]["value"] =
						TransparencyStackRT.texture;
					this.get("_renderercache").get("materials")["ColorBlendShaderMaterial"]["uniforms"]["OpaqueRT"]["value"] =
						OpaqueBackgroundMaskRT.texture;
					this.get("_renderercache").get("materials")["ColorBlendShaderMaterial"]["uniforms"]["ColorStackRT"]["value"] =
						ColorStackRT.texture;
					this.get("_3JSrenderer").render(this.get("_screenQuadSceneColorBlend"), cameraRTT, MirrorColorStackRT);

					this.get("_renderercache").get("materials")["ColorBlendShaderMaterial"]["uniforms"]["EnableMirror"]["value"] = 0;
				} else {
					this.get("_3JSrenderer").autoClearColor = true;
					this.get("_3JSrenderer").setClearColor(blackClearDepthColor, 0);
					this.get("_renderercache").get("materials")["ColorBlendShaderMaterial"]["uniforms"]["EnableMirror"]["value"] = 0;
					this.get("_renderercache").get("materials")["ColorBlendShaderMaterial"]["uniforms"]["CX1RT"]["value"] = PeelColorRT.texture;
					this.get("_renderercache").get("materials")["ColorBlendShaderMaterial"]["uniforms"]["TransparencyStack"]["value"] =
						TransparencyStackRT.texture;
					this.get("_renderercache").get("materials")["ColorBlendShaderMaterial"]["uniforms"]["OpaqueRT"]["value"] =
						OpaqueBackgroundMaskRT.texture;
					this.get("_renderercache").get("materials")["ColorBlendShaderMaterial"]["uniforms"]["ColorStackRT"]["value"] =
						ColorStackRT.texture;
					this.get("_3JSrenderer").render(this.get("_screenQuadSceneColorBlend"), cameraRTT, MirrorColorStackRT);
				}

				this.get("_3JSrenderer").autoClearColor = true;

				if ((enableHighlights && enableHorizonHighlights) || enableLabels) {
					this.get("_3JSrenderer").clear(false, true, false); // clear the hardware Z-buffer in last active render target

					if (enableHighlights && enableHorizonHighlights) {
						// THIRD PASS: Render outline of objects (silhouette of smooth surfaces, opaque + transparent)
						this.get("_3JSrenderer").autoClear = true;
						this.get("_3JSrenderer").setClearColor(blackColor, 0);
						this.RenderOpaqueSceneDepth(OpaqueDepthRT, this.get("_renderercache").get("materials")["colordepth"]);
						this.get("_3JSrenderer").autoClear = false;
						this.RenderTransparentSceneDepth(OpaqueDepthRT, this.get("_renderercache").get("materials")["colordepth"]);
						this.get("_3JSrenderer").autoClear = false; //previous transp scene draw call is setting autoclear back to true

						this.get("_renderercache").get("materials")["edge"]["uniforms"]["canvas_size"]["value"] = new THREE.Vector2(width, height);
						this.get("_renderercache").get("materials")["edge"]["uniforms"]["depth_texture"]["value"] = OpaqueDepthRT.texture;
						this.RenderOpaqueSceneDepth(MirrorColorStackRT, this.get("_renderercache").get("materials")["edge"]);
						this.get("_3JSrenderer").autoClear = false;
						this.RenderTransparentSceneDepth(MirrorColorStackRT, this.get("_renderercache").get("materials")["edge"]);
						this.get("_3JSrenderer").autoClear = false;
					}

					// Render labels
					if (enableLabels) {
						this.get("_3JSrenderer").autoClear = false;
						this.get("_3JSrenderer").render(
							this.get("_scenemanager").get("_3JSsceneLabels"),
							this.get("_3JSHUDcamera"),
							MirrorColorStackRT
						);
					}

					this.get("_3JSrenderer").autoClear = true;
				}

				if (!__offscreen) {
					// Copy colorstack to screen if needed.
					this.get("_3JSrenderer").setClearColor(blackClearDepthColor, 0); //black ?
					this.get("_renderercache").get("materials")["CopyShaderMaterial"]["transparent"] = true;
					this.get("_renderercache").get("materials")["CopyShaderMaterial"]["blending"] = THREE.AdditiveBlending;
					this.get("_renderercache").get("materials")["CopyShaderMaterial"]["uniforms"]["canvas_size"]["value"] = new THREE.Vector2(
						width,
						height
					);
					this.get("_renderercache").get("materials")["CopyShaderMaterial"]["uniforms"]["CXRT"]["value"] = MirrorColorStackRT.texture;
					this.get("_renderercache").get("materials")["CopyShaderMaterial"]["uniforms"]["OpaqueRT"]["value"] = null;
					this.get("_renderercache").get("materials")["CopyShaderMaterial"]["uniforms"]["EnableMask"]["value"] = 0;
					this.get("_renderercache").get("materials")["CopyShaderMaterial"]["uniforms"]["EnableRefOpacity"]["value"] = 0;
					this.get("_renderercache").get("materials")["CopyShaderMaterial"]["uniforms"]["EnableInvRefOpacity"]["value"] = 0;

					if (!this.get("_screenQuadSceneCopyShader")) {
						const screen_quad_copy_shader = new THREE.Mesh(screen_plane, this.get("_renderercache").get("materials")["CopyShaderMaterial"]);
						const screen_quad_scene_copy_shader = new THREE.Scene();
						screen_quad_scene_copy_shader.add(screen_quad_copy_shader);
						this.set("_screenQuadSceneCopyShader", screen_quad_scene_copy_shader);
					}

					this.get("_3JSrenderer").render(this.get("_screenQuadSceneCopyShader"), cameraRTT, null);
				}

				// this.get('_3JSrenderer').setClearColor( backgroundColor, 1 );//Restore clear color
				this.get("_3JSrenderer").autoClear = false;
			} else {
				//No transparency or basic designer quality transparency, render it using standard shading algorithm,
				//this version has no aliasing problems since we use only the screen as render target.

				//borrow render target to render reflections
				let ReflectionRT = null;
				if (this.IsReflectionsRenderingEnabled()) {
					ReflectionRT = this.get("_PeelColorRT");
					this.get("_3JSrenderer").setClearColor(blackColor, 1);

					this.FlipOpaqueScene();
					this.FlipTransparentScene();

					//Flip GL_FRONT to GL_BACK and vice versa to compensate scaling *-1 on Y axis
					this.get("_materialmanager").ApplyFctToAllMaterials(this.FlipMaterialSide(THREE.BackSide));

					//Flip lights positions (for designer quality), otherwise lighting is wrong in reflection (since normal hasn't been flipped)
					this.FlipLights();

					this.RenderOpaqueScene(ReflectionRT, false);
					this.get("_3JSrenderer").autoClear = false;
					this.RenderTransparentScene(ReflectionRT, undefined, false);

					this.UnFlipOpaqueScene();
					this.UnFlipTransparentScene();

					this.get("_materialmanager").ApplyFctToAllMaterials(this.FlipMaterialSide(THREE.FrontSide));
					this.UnFlipLights();
				}

				if (backdropTex) {
					this.get("_renderercache").get("materials")["BackdropShaderMaterial"]["uniforms"]["canvas_size"]["value"] = new THREE.Vector2(
						width,
						height
					);
					this.get("_renderercache").get("materials")["BackdropShaderMaterial"]["uniforms"]["backdrop"]["value"] = backdropTex;
					this.get("_renderercache").get("materials")["BackdropShaderMaterial"]["uniforms"]["backdrop_enabled"]["value"] = 1;
					this.get("_renderercache").get("materials")["BackdropShaderMaterial"]["uniforms"]["backdrop_scale"]["value"] = new THREE.Vector2(
						bd_scaleX,
						bd_scaleY
					);
					this.get("_renderercache").get("materials")["BackdropShaderMaterial"]["uniforms"]["backdrop_offset"]["value"] = new THREE.Vector2(
						bd_offsetX,
						bd_offsetY
					);

					this.get("_renderercache").get("materials")["BackdropShaderMaterial"]["uniforms"]["reflection"]["value"] = ReflectionRT;
					this.get("_renderercache").get("materials")["BackdropShaderMaterial"]["uniforms"]["reflection_enabled"][
						"value"
					] = this.IsReflectionsRenderingEnabled() ? 1 : 0;

					if (!this.get("_screenQuadSceneBackdropShader")) {
						const backdrop_plane = new THREE.Mesh(screen_plane, this.get("_renderercache").get("materials")["BackdropShaderMaterial"]);
						const backdrop_shader_scene = new THREE.Scene();
						backdrop_shader_scene.add(backdrop_plane);
						this.set("_screenQuadSceneBackdropShader", backdrop_shader_scene);
					}
					this.get("_3JSrenderer").render(this.get("_screenQuadSceneBackdropShader"), cameraRTT, __offscreen ? ColorStackRT : null);
					this.get("_3JSrenderer").autoClear = false;
				} else {
					if (this.IsReflectionsRenderingEnabled()) {
						this.get("_3JSrenderer").setClearColor(this.get("_backgroundColor"), this.get("_backgroundAlpha"));

						//Here, we render the reflections using the Backdrop shader material (we will then draw the scene on top of it)
						this.get("_renderercache").get("materials")["BackdropShaderMaterial"]["uniforms"]["canvas_size"]["value"] = new THREE.Vector2(
							width,
							height
						);
						this.get("_renderercache").get("materials")["BackdropShaderMaterial"]["uniforms"]["backdrop"]["value"] = null;
						this.get("_renderercache").get("materials")["BackdropShaderMaterial"]["uniforms"]["backdrop_enabled"]["value"] = 0;
						this.get("_renderercache").get("materials")["BackdropShaderMaterial"]["uniforms"]["reflection"]["value"] = ReflectionRT.texture;
						this.get("_renderercache").get("materials")["BackdropShaderMaterial"]["uniforms"]["reflection_enabled"]["value"] = 1;
						this.get("_renderercache").get("materials")["BackdropShaderMaterial"]["uniforms"]["default_bg_color"][
							"value"
						] = new THREE.Vector4(
							this.get("_backgroundColor").r,
							this.get("_backgroundColor").g,
							this.get("_backgroundColor").b,
							this.get("_backgroundAlpha")
						);

						if (!this.get("_screenQuadSceneReflectionShader")) {
							const reflection_plane = new THREE.Mesh(screen_plane, this.get("_renderercache").get("materials")["BackdropShaderMaterial"]);
							const reflection_shader_scene = new THREE.Scene();
							reflection_shader_scene.add(reflection_plane);
							this.set("_screenQuadSceneReflectionShader", reflection_shader_scene);
						}
						this.get("_3JSrenderer").render(this.get("_screenQuadSceneReflectionShader"), cameraRTT, __offscreen ? ColorStackRT : null);
						this.get("_3JSrenderer").autoClear = false;
					} else {
						this.get("_3JSrenderer").setClearColor(this.get("_backgroundColor"), this.get("_backgroundAlpha"));
						this.get("_3JSrenderer").autoClear = true;
					}
				}

				// Render opaque objects
				this.RenderOpaqueScene(__offscreen ? ColorStackRT : null);
				this.get("_3JSrenderer").autoClear = false;
				// Render transparent objects
				this.RenderTransparentScene(__offscreen ? ColorStackRT : null, undefined, false);
				this.get("_3JSrenderer").autoClear = false;

				if (this.get("_annotationsEnabled") && this.get("_renderercache").get("_RenderingQuality") === "Visualizer") {
					// fix the alpha of the annotations to prevent them from blending with the HTML background
					const screen_quad_fix_alpha = new THREE.Mesh(screen_plane, this.get("_renderercache").get("materials")["alphafix"]);
					const screen_quad_scene_fix_alpha = new THREE.Scene();
					screen_quad_scene_fix_alpha.add(screen_quad_fix_alpha);

					this.get("_3JSrenderer").context.colorMask(false, false, false, true);
					this.get("_3JSrenderer").render(screen_quad_scene_fix_alpha, cameraRTT, __offscreen ? ColorStackRT : null);
					this.get("_3JSrenderer").context.colorMask(true, true, true, true);
				}

				if (enableHighlights) {
					if (enableHorizonHighlights) {
						// THIRD PASS: Render outline of objects (silhouette of smooth surfaces, opaque + transparent)
						this.get("_3JSrenderer").autoClear = true;
						const OpaqueDepthRT = this.get("_OpaqueDepthRT");
						this.get("_3JSrenderer").setClearColor(blackColor, 0);
						this.RenderOpaqueSceneDepth(OpaqueDepthRT, this.get("_renderercache").get("materials")["colordepth"]);
						this.get("_3JSrenderer").autoClear = false;

						this.get("_renderercache").get("materials")["edge"]["uniforms"]["canvas_size"]["value"] = new THREE.Vector2(width, height);
						this.get("_renderercache").get("materials")["edge"]["uniforms"]["depth_texture"]["value"] = OpaqueDepthRT.texture;
						this.RenderOpaqueSceneDepth(__offscreen ? ColorStackRT : null, this.get("_renderercache").get("materials")["edge"]);
						this.get("_3JSrenderer").autoClear = false;
					}

					if (enableHardEdgeHighlights) {
						// FOURTH PASS: Render highlights (hard edges)
						this.get("_3JSrenderer").render(
							this.get("_scenemanager").get("_3JSsceneHightlights"),
							this.get("_3JScamera"),
							__offscreen ? ColorStackRT : null
						);
					}
				}

				// Render labels
				if (enableLabels)
					this.get("_3JSrenderer").render(
						this.get("_scenemanager").get("_3JSsceneLabels"),
						this.get("_3JSHUDcamera"),
						__offscreen ? ColorStackRT : null
					);
			}

			this.get("_3JSrenderer").autoClear = true;

			if (__offscreen) {
				//Reset render targets sizes after offscreen render
				this.set("_resetRenderTargets", true);
			}
		},

		/*= INTERACTIVE FRAMERATE & OPTIMIZATION =*/

		StartInteractiveFramerate: function() {
			this.set("_interactiveFrameRate", true);
			this.set("_current_nb_peels", this.get("_nb_min_peels"));
		},

		EndInteractiveFramerate: function(__redraw) {
			this.set("_interactiveFrameRate", false);
			if (__redraw === undefined || __redraw) this.Render();
		},

		/** SetInteractionOptimization */
		SetInteractionOptimization: function(__enableInteractionOptimization) {
			this.set("_enableInteractionOptimization", __enableInteractionOptimization);
		},

		/**
		 * SetQualityLevel
		 *
		 * @function SetQualityLevel sets maximum quality when no interaction with the model or when interaction optimization is disabled
		 * by setting the number of transparency peels for transparent models. This quality level does not affect quality of opaque models.
		 * @param {number} __qualityLevel - an integer between 1 and 10 (included)
		 * @param __redraw
		 */
		SetQualityLevel: function(__qualityLevel, __redraw) {
			this.set("_nb_max_peels", __qualityLevel);
			if (__redraw === undefined || __redraw) this.Render();
		},

		/*= RESIZE =*/

		Resize: function(__event, __overrideWidth, __overrideHeight) {
			this.set("_resetRenderTargets", true);
			this.get("_exec_queue").whenFinished(function() {
				this.Render(false, __overrideWidth, __overrideHeight);
			}, this);
		},

		/*= HIGHLIGHTS =*/

		UpdatePolygonOffset: function(__enablePolygonOffset) {
			return function(__material) {
				if (__material && __material["polygonOffset"]) {
					__material["polygonOffsetFactor"] = __enablePolygonOffset ? this.get("_renderercache").get("_polygon_offset_factor") : 0;
					__material["polygonOffsetUnits"] = __enablePolygonOffset ? this.get("_renderercache").get("_polygon_offset_units") : 0;
				}
			}.bind(this);
		},

		SetHighlights: function(__enableHighlights, __redraw) {
			const materials = this.get("_renderercache").get("materials");
			materials["enablehighlights"] = __enableHighlights;

			const enablePolygonOffset = __enableHighlights && materials["enablehardedgehighlights"];
			this.get("_materialmanager").ApplyFctToAllMaterials(this.UpdatePolygonOffset(enablePolygonOffset));

			if (__redraw === undefined || __redraw) this.Render();
		},

		/** SetHighlightsConfiguration */
		SetHighlightsConfiguration: function(__enableHorizonHighlights, __enableHardEdgeHighlights, __redraw) {
			const materials = this.get("_renderercache").get("materials");
			materials["enablehardedgehighlights"] = __enableHardEdgeHighlights;
			materials["enablehorizonhighlights"] = __enableHorizonHighlights;

			const enablePolygonOffset = materials["enablehighlights"] && __enableHardEdgeHighlights;
			this.get("_materialmanager").ApplyFctToAllMaterials(this.UpdatePolygonOffset(enablePolygonOffset));

			if (__redraw === undefined || __redraw) this.Render();
		},

		/** SetHighlightColor */
		SetHighlightsColor: function(__red, __green, __blue, __redraw) {
			//Change highlight color for hard edges
			this.get("_renderercache").get("materials")["highlight"]["uniforms"]["diffuse"]["value"] = new THREE.Vector3(
				__red / 255,
				__green / 255,
				__blue / 255
			);
			//Change highlight color for smoothe edges
			this.get("_renderercache").get("materials")["edge"]["uniforms"]["color"]["value"] = new THREE.Vector3(
				__red / 255,
				__green / 255,
				__blue / 255
			);

			if (this.get("_renderercache").get("materials")["enablehighlights"] && (__redraw === undefined || __redraw)) this.Render();
		},

		/*= LIGHTING CONTRAST (DESIGNER QUALITY RENDERING ONLY) =*/

		/**
		 * SetLightingContrast
		 *
		 * @function SetLightingContrast sets the lighting contrast (in designer quality only)
		 * by reducing the directional light and increasing the ambient light components.
		 * @param {number} __lightingContrast - a floating number between 0 and 1 (included)
		 * @param __redraw
		 */
		SetLightingContrast: function(__lightingContrast, __redraw) {
			this.set("_lighting_contrast", __lightingContrast);
			this.UpdateLightingContrast();
			if (__redraw === undefined || __redraw) this.Render();
		},

		UpdateLightingContrast: function() {
			// set directional light intensity
			// default is THREE.DirectionalLight( 0xffffff, 0.5 ) for __lightingContrast == 0.5;
			for (
				let cg_idx = 0, cg_keys = Object.keys(this.get("_renderercache").get("lights")["dir_light"]), l = cg_keys.length;
				cg_idx < l;
				cg_idx++
			) {
				const __coincident_group_id = cg_keys[cg_idx];
				const directionalLight = this.get("_renderercache").get("lights")["dir_light"][__coincident_group_id];
				directionalLight.intensity = this.get("_lighting_contrast");
			}

			// set ambient lighting intensity
			// default is THREE.AmbientLight(0x777777) for __lightingContrast == 0.5;
			for (
				let cg_idx = 0, cg_keys = Object.keys(this.get("_renderercache").get("lights")["amb_light"]), l = cg_keys.length;
				cg_idx < l;
				cg_idx++
			) {
				const __coincident_group_id = cg_keys[cg_idx];
				const ambientLight = this.get("_renderercache").get("lights")["amb_light"][__coincident_group_id];
				ambientLight.intensity = 1 + (0.5 - this.get("_lighting_contrast"));
			}
		},

		/*= VISIBILITY CONTROL =*/

		/**
		 * SetLabels visibility
		 *
		 * @param __show true to show all labels and false to hide them all
		 * @param __redraw
		 */
		SetLabelsVisible: function(__show, __redraw) {
			this.get("_renderercache").get("labels")["visible"] = __show;

			if (__redraw === undefined || __redraw) this.CameraUpdate(true);
		},

		/**
		 * Reset visiblity of the main scene to its default (in json structure??? to define)
		 * default visiblity of the main scene is true at startup
		 *
		 * @function
		 * @param {Boolean} __redraw - If true, redraw the scene after visibility change - default is true
		 */
		ResetMainSceneVisibility: function(__redraw) {
			this.get("_scenemanager").ResetMainSceneVisibility(__redraw);
		},

		/**
		 * Toggle visiblity of a node using its URL in the scene
		 * default visiblity of a node is true at startup
		 *
		 * @function
		 * @param {String} __nodeURL - URL identifying a scene node
		 * @param {Boolean} __recurseChildren - If true, traverse all children to toggle their visibility state
		 * @param {Boolean} __recurseParents - If true, traverse all parents to toggle their visibility state
		 * @param {Boolean} __redraw - If true, redraw the scene after visibility change - default is true
		 */
		ToggleNodeVisibility: function(__nodeURL, __recurseChildren, __recurseParents, __redraw) {
			this.SetNodeVisibility(__nodeURL, null, __recurseChildren, __recurseParents, __redraw);
		},

		/**
		 * Set visiblity state of a node using its URL in the scene
		 * default visiblity of a node is true at startup
		 *
		 * @function
		 * @param {String} __nodeURL - URL identifying a scene node
		 * @param {Boolean} __visible - Set the object visibility to __visible
		 * @param {Boolean} __recurseChildren - If true, traverse all children to set their visibility to __visible
		 * @param {Boolean} __recurseParents - If true, traverse all parents to set their visibility to __visible
		 * @param {Boolean} __redraw - If true, redraw the scene after visibility change - default is true
		 */
		SetNodeVisibility: function(__nodeURL, __visible, __recurseChildren, __recurseParents, __redraw) {
			const shortNodeURL = WV3DUtils.URLToKeyString(__nodeURL);

			if (this.get("_renderercache").get("nodeviewproperties")[shortNodeURL] === undefined) return;

			for (
				let i = 0, keys = Object.keys(this.get("_renderercache").get("nodeviewproperties")[shortNodeURL]["threejsnodes"]), l = keys.length;
				i < l;
				++i
			) {
				const setvisibilitycallback = function(obj) {
					if (!(obj instanceof THREE.Scene || obj instanceof THREE.Mesh)) {
						if (__visible === null) {
							obj.visible = !obj.visible; //Simply toggle
						} else {
							obj.visible = __visible; //Force visibility state
						}
					}
				};

				//Get all objects found in the available passes/modes
				const objects = this.get("_renderercache").get("nodeviewproperties")[shortNodeURL]["threejsnodes"][keys[i]];

				for (let node_cg = 0, keys2 = Object.keys(objects), l2 = keys2.length; node_cg < l2; node_cg++) {
					const object = objects[keys2[node_cg]];

					setvisibilitycallback(object);
					if (__recurseChildren !== undefined && __recurseChildren) {
						for (let j = 0, keys3 = Object.keys(object.children), l3 = keys3.length; j < l3; ++j) {
							const id = keys3[j];
							object.children[id].traverse(setvisibilitycallback);
						}
					}
					if (__recurseParents !== undefined && __recurseParents) {
						if (object.parent) {
							object.parent.traverseParents(setvisibilitycallback);
						}
					}
				}
			}
			if (__redraw === undefined || __redraw) this.Render();
		},

		/**
		 * Get the queried data info of a node
		 * @function
		 * @param {string} __object_node - the type of information we want to retrieve from node
		 * @param {string} __datatype -  the type of information we want to retrieve from node
		 * to choose among the following list:
		 * - NODEOBJECTVIS returns (Boolean)
		 * @returns object, boolean, string, depending on the type of info
		 */
		GetNodeInfo: function(__object_node, __datatype) {
			switch (__datatype) {
				case "NODEOBJECTVIS":
					//Get the first object found in the first pass available
					const objects = __object_node["threejsnodes"][Object.keys(__object_node["threejsnodes"])[0]];
					const object = objects[Object.keys(objects)[0]];
					return object.visible;
				default:
					LogMessage.Log("Unknown queried data type");
			}
		},

		/**
		 * Get the queried data info of a node using its URL
		 *
		 * @function
		 * @param {string} __nodeURL -  the type of information we want to retrieve from node
		 * @param {string} __datatype -  the type of information we want to retrieve from node
		 * to choose among the following list:
		 * - NODEOBJECTVIS returns (Boolean)
		 * @returns object, boolean, string, depending on the type of info
		 */
		Get: function(__nodeURL, __datatype) {
			const shortNodeURL = WV3DUtils.URLToKeyString(__nodeURL);
			const object_ref = this.get("_renderercache").get("nodeviewproperties")[shortNodeURL];
			if (object_ref) {
				return this.GetNodeInfo(object_ref, __datatype);
			} else {
				LogMessage.Log("ViewerView Warning: No existing node for " + shortNodeURL);
			}
		},

		/**
		 * Picking
		 */
		Pick: function(__x, __y) {
			const raycaster = new THREE.Raycaster();
			const coordinates = new THREE.Vector2(
				(__x / this.get("_context").clientWidth) * 2.0 - 1.0,
				-((__y / this.get("_context").clientHeight) * 2.0 - 1.0)
			);
			raycaster.setFromCamera(coordinates, this.get("_3JScamera"));

			return this.get("_scenemanager").Pick(raycaster);
		},

		/**
		 * Get a snapshot of the current viewer state as specified in format using offscreen rendering
		 *
		 * @function
		 * @param {string} [__format="png"] - image format
		 * @param __width
		 * @param __height
		 * @returns {Object} image data
		 */
		GetSnapshot: function(__format, __width, __height) {
			const flipV = function(__imageData) {
				// loop through over half of row of pixels
				for (let row = 0; row < __imageData.height / 2; row++) {
					// loop over every column
					for (let col = 0; col < __imageData.width; col++) {
						// source pixel
						const sp = __imageData.width * row + col;

						// destination pixel
						const dp = __imageData.width * (__imageData.height - 1 - row) + col;

						// source and destination indexes, will always reference the red pixel
						const si = sp * 4;
						const di = dp * 4;

						//swap red channel
						let temp_swap = __imageData.data[di];
						__imageData.data[di] = __imageData.data[si];
						__imageData.data[si] = temp_swap;
						//swap green channel
						temp_swap = __imageData.data[di + 1];
						__imageData.data[di + 1] = __imageData.data[si + 1];
						__imageData.data[si + 1] = temp_swap;
						//swap blue channel
						temp_swap = __imageData.data[di + 2];
						__imageData.data[di + 2] = __imageData.data[si + 2];
						__imageData.data[si + 2] = temp_swap;
						//swap alpha channel
						temp_swap = __imageData.data[di + 3];
						__imageData.data[di + 3] = __imageData.data[si + 3];
						__imageData.data[si + 3] = temp_swap;
					}
				}
			};

			this.Render(true, __width, __height);

			const gl = this.get("_3JSrenderer").getContext();
			const screenshotdata = new Uint8Array(__width * __height * 4);
			//If bug on safari check screenshotdata['data'] !!!
			gl.readPixels(0, 0, __width, __height, gl.RGBA, gl.UNSIGNED_BYTE, screenshotdata);

			const canvas = document.createElement("canvas");
			canvas.width = __width;
			canvas.height = __height;
			const context = canvas.getContext("2d");

			// Copy the pixels to a 2D canvas
			const imageData = context.createImageData(__width, __height);
			imageData.data.set(screenshotdata);

			// Flip the image data since opengl buffer is mirrored compared to png
			flipV(imageData);
			context.putImageData(imageData, 0, 0);

			if (typeof this.get("_context").appendChild === "function") {
				this.CameraUpdate(true, this.get("_context").clientWidth, this.get("_context").clientHeight);
			}

			return canvas.toDataURL("image/" + __format);
		}
	});
});
