/**
 * @author SVN & DOGO (Original by RAVO)
 * @module WebViewer3D
 * @copyright   2015 Esko Stonecube Ltd. All rights reserved.
 */

define(["three",
        "@esko/3d-web-viewer/modules/shaders/VertexShaderNPM",
        "@esko/3d-web-viewer/modules/shaders/ShaderUtils"
    ],
    function (THREE,
              VertexShaderNPM,
              ShaderUtils)
    {
        /**
         * @alias module:WebViewer3D
         * @class Reflectance Shader for Non Printable Materials
         */

        const _uniforms =
            {
                "specularCol": {type: "c", value: new THREE.Color(0xffffff)},
                "shininess": {type: "f", value: 0.5},

                "isFloorReflection": {type: "i", value: 0},
                "haze1CubeMap": {type: "t", value: null},
                "haze2CubeMap": {type: "t", value: null},
                "envMapRotQuat": {type: "v4", value: new THREE.Vector4(0, 0, 0, 1)},

                "transparent": {type: "i", value: 0},

                "canvas_size": {type: "v2", value: new THREE.Vector2(1, 1)},
                "DX1RT": {type: "t", value: null},
                "DX2RT": {type: "t", value: null},
                "NBRTSRC": {type: "i", value: 0},

                "stripSize": {type: "f", value: 250},
                "gapSize": {type: "f", value: 35}
            };

        const EXT_uniforms =
            {
                "haze3CubeMap": {type: "t", value: null},
                "haze4CubeMap": {type: "t", value: null},
                "haze5CubeMap": {type: "t", value: null}
            };

        const fragmentShader_part1 =
            [
                "varying vec4 vPosition;",
                "varying vec4 vWorldPosition;",
                "varying vec3 vNormal;",

                "uniform vec3 specularCol;",
                "uniform float shininess;",

                "uniform vec4 envMapRotQuat;",

                "#ifdef USE_ENVMAP",
                    "uniform bool isFloorReflection;",
                    "uniform samplerCube haze1CubeMap;",
                    "uniform samplerCube haze2CubeMap;"
            ].join("\n");

        const fragmentShader_ext1 =
            [
                    "uniform samplerCube haze3CubeMap;",
                    "uniform samplerCube haze4CubeMap;",
                    "uniform samplerCube haze5CubeMap;"
            ].join("\n");

        const fragmentShader_part2 =
            [
                "#endif",

                "vec3 resolveHaze(float haze, vec3 reflSamples[5])",
                "{",
                    // work out the contribution from each reflection map based on the haze value
                    "float haze4 = 4.0 * haze;",
                    "vec3 result = reflSamples[0] * (1.0 - clamp(abs(0.0 - haze4), 0.0, 1.0));",
                    "result +=     reflSamples[1] * (1.0 - clamp(abs(1.0 - haze4), 0.0, 1.0));",
                    "result +=     reflSamples[2] * (1.0 - clamp(abs(2.0 - haze4), 0.0, 1.0));",
                    "result +=     reflSamples[3] * (1.0 - clamp(abs(3.0 - haze4), 0.0, 1.0));",
                    "result +=     reflSamples[4] * (1.0 - clamp(abs(4.0 - haze4), 0.0, 1.0));"
            ].join("\n");

        const fragmentShader_part2Alt =
            [
                "#endif",

                // work out the contribution from each reflection/environment map based on the haze value
                "vec3 resolveHaze(float haze, vec3 reflSamples[5])",
                "{",
                    "vec3 result = reflSamples[0] * (1.0 - clamp(abs(0.0 - haze), 0.0, 1.0));",
                    "result +=     reflSamples[1] * (1.0 - clamp(abs(1.0 - haze), 0.0, 1.0));"
            ].join("\n");

        const fragmentShader_part3 =
            [
                    "return result;",
                "}",

                ShaderUtils.get("DeclareDepthSamplersChunck"),
                ShaderUtils.get("EncodeDepthChunck"),
                ShaderUtils.get("DecodeDepthChunck"),
                ShaderUtils.get("EnvironmentRotationChunck"),
                ShaderUtils.get("Noise"),
                ShaderUtils.get("Distort"),

                "void main()",
                "{",
                    ShaderUtils.get("SelectPeelChunck"),

                    "#ifdef USE_ENVMAP",
                        "vec3 pN = normalize( vNormal);",
                        ShaderUtils.get("WrapDistortNormal"),

                        "vec3 eyeVector = normalize(vec3(vWorldPosition) - cameraPosition.xyz);",
                        "vec3 reflection = reflect(eyeVector, pN);",
                        "if(isFloorReflection)",
                        "{",
                            "reflection.y = -reflection.y;",
                        "}",

                        "reflection = rotate_vec_from_quat(envMapRotQuat, reflection);",

                        // Get reflection map samples
                        "vec3 reflSamples[5], reflResult;",
                        "reflSamples[0] = textureCube(haze1CubeMap, reflection).rgb;",
                        "reflSamples[1] = textureCube(haze2CubeMap, reflection).rgb;"
            ].join("\n");

        const fragmentShader_ext3 =
            [
                        "reflSamples[2] = textureCube(haze3CubeMap, reflection).rgb;",
                        "reflSamples[3] = textureCube(haze4CubeMap, reflection).rgb;",
                        "reflSamples[4] = textureCube(haze5CubeMap, reflection).rgb;"
            ].join("\n");

        const fragmentShader_part4 =
            [
                        // calculate reflections using haze and reflection coefficients
                        "reflResult = specularCol.rgb * resolveHaze(1.0 - shininess, reflSamples);",
                    "#else",
                        "vec3 reflResult = specularCol.rgb;",
                    "#endif",

                    // calculate reflections using haze and reflection coefficients
                    "gl_FragColor.rgb = reflResult;",
                    "gl_FragColor.a = 0.0;",
                "}"
            ].join("\n");

        return {

            vertexShader: VertexShaderNPM.vertexShader,

            //On some mobile devices, textures units are limited
            //In this case, we downgrade quality of environment

            //Mobile/Compatibility version
            uniforms: THREE.UniformsUtils.merge([
                VertexShaderNPM.uniforms,
                _uniforms
            ]),

            fragmentShader: [
                fragmentShader_part1,
                fragmentShader_part2Alt,
                fragmentShader_part3,
                fragmentShader_part4
            ].join("\n"),

            //Desktop/Full version
            EXTuniforms: THREE.UniformsUtils.merge([
                VertexShaderNPM.uniforms,
                _uniforms,
                EXT_uniforms
            ]),

            EXTfragmentShader: [
                fragmentShader_part1,
                fragmentShader_ext1,
                fragmentShader_part2,
                fragmentShader_part3,
                fragmentShader_ext3,
                fragmentShader_part4
            ].join("\n")
        };
    });
