/**
 * @author DOGO
 * @contributor SVN
 * @module WebViewer3D
 * @copyright   2015 Esko Stonecube Ltd. All rights reserved.
 */

define(["three",
        "backbone",
        "@esko/3d-web-viewer/modules/SystemChecker",
        "@esko/3d-web-viewer/modules/LogMessage"
    ],
    function (THREE,
              bb,
              WV3DSystemChecker,
              LogMessage)
    {
        /**
         * @alias module:WebViewer3D
         * @class ShaderUtils gather all shader functions that can be reused in other shaders
         * Warning : this class is a singleton, any modification to one of its component will change for each reference to it.
         */


        const Noise =
            [
                "#if defined(WRAP_MATERIAL)",
				
					// Returns a random value between 0 and 1, using a 3d point as a seed. No continuity.
                    "float hash( vec3 p )", 
                    "{",
                        "float h = dot( p, vec3( 127.1, 311.7, 233.2 ) );",
                        "return fract( sin( h ) * 43758.5453123 );",
                    "}",

					// Returns a random value between -1 and 1, using the integer values of a 3d point as a seed.
					// The fractional part is used to interpolate smoothly between the grid points. C1 continuous.
                    "float Noise( vec3 p )", 
                    "{",
                        "vec3 i = floor( p );", // clamp the position to a grid point nearby
                        "vec3 f = fract( p );",  // store the fractional part for interpolation						
                        "vec3 u = f * f * ( 3.0 - 2.0 * f );", // convert linear to C1 continous interpolation
						// Get 8 random values in 8 corners of a grid cell and smoothly interpolate between them.
                        "return -1.0 + 2.0 * mix(    mix(   mix(   hash( i + vec3(0.0,0.0,0.0) ),",
                                                                    "hash( i + vec3(1.0,0.0,0.0) ),",
                                                                    "u.x),",
                                                            "mix(   hash( i + vec3(0.0,1.0,0.0) ),",
                                                                    "hash( i + vec3(1.0,1.0,0.0) ),",
                                                                    "u.x),",
                                                            "u.y),",
                                                    "mix(   mix(    hash( i + vec3(0.0,0.0,1.0) ),",
                                                                    "hash( i + vec3(1.0,0.0,1.0) ), ",
                                                                    "u.x),",
                                                            "mix(   hash( i + vec3(0.0,1.0,1.0) ),",
                                                                    "hash( i + vec3(1.0,1.0,1.0) ),",
                                                                    "u.x), ",
                                                            "u.y),",
                                                    "u.z);",
                    "}",
                "#endif"

            ].join("\n");

        const Distort =
            [
                "#if defined(WRAP_MATERIAL)",
                // camera rotation : pitch, yaw
                "mat3 rotationXY( vec2 angle )",
                "{",
                    "vec2 c = cos( angle );",
                    "vec2 s = sin( angle );",

                    "return mat3(",
                        "c.y      ,  0.0, s.y,",
                        "-s.y * s.x,  c.x,  -c.y * s.x,",
                        "-s.y * c.x, s.x,  c.y * c.x",
                    ");",
                "}",

                "vec3 Distort(vec3 position, vec3 scale)",
                "{",
                    "vec3 scaled_position = position * scale * 0.01;", // Match the scale of the wrinkles to the scale of the scene,
                    "vec3 distort = 0.001 * vec3( Noise(scaled_position), Noise(scaled_position + 100.0 ), Noise(scaled_position + 237. ));", // make 3 "independant" noise lookups by adding some random delta.
                    
                    // For added realism, we want many small wrinkles and fewer big ones, so we apply a power curve.
					// Since pow() doesn't work on negative numers, we have to save and restore the sign of the distortion.
					"vec3 sign1 = sign(distort);",
                    "distort = abs(distort);",
                    "distort = pow(distort, vec3(.6));",
                    "distort *= sign1;",
                        "return distort;",
                "}",

				// The base layer does not use banding, it applies horizontal wrinkles everywhere\
				// so even the overlapping gaps of other layers never create empty space.
                "vec3 DistortBaseLayer(vec3 position)",
                "{",
				    // Scale the Y dimension by 30 to compress wrinkles in height (= stretch them in the X and Z dimensions)
                    "vec3 scale = vec3( 1.0, 30.0 + 1.0 * Noise( position.yyy * 0.1 ), 1.0);", //Noise( position.yyy * 0.1 ) brings a subtle high frequency noise
                    "return Distort(position, scale);",
                "}",

                "uniform float stripSize;", // width of a band
                "uniform float gapSize;", // width of the gap between bands
                "const float extraLayerContrast = 0.35;", // parameter agreed with PO. not exposed in JSON

				// extra layers don't add wrinkles everywhere, only in parallel bands separated by gaps.
                "vec4 DistortExtraLayer(vec3 position)",
                "{",
                    "float band_position = mod(position.y,stripSize+gapSize);", // find the position inside the one band (+gap) we are in
                    "if ( band_position > gapSize)",
                    "{",
                        // In the reference picture, the areas closer to the edge of a band had more wrinkles and looked less transparent. Try to emulate that."
                        "band_position = (band_position - gapSize);",						
						// get the distance to the center of the band. Value between 0 (center) and 0.5 * stripSize (edge)
						"float center_distance = abs( band_position - 0.5 * stripSize);",
						// edge_factor is 0.2 in the middle 40% of the band, and smoothly rises to .4 at the outer 10% of both sides of the band.
						"float edge_factor = 0.2 + 0.2 * smoothstep( stripSize * 0.2, stripSize * 0.4, center_distance);", 
						
						// Scale the Y dimension by [20 to 30] to compress wrinkles in height (= stretch them in the X and Z dimensions)
                        "vec3 scale = vec3( 1.0, 10.0 * edge_factor + 20.0, 1.0 );",

                        "return vec4(Distort(position, scale), edge_factor * extraLayerContrast);",
                    "}",
                    "else",
                    "{",
                        "return vec4(0.,0.,0.,0.);",
                    "}",
                "}",
                "#endif"

            ].join("\n");

        const WrapDistort =
            [
                //Base layer
                "vec4 distort = vec4(DistortBaseLayer(vPosition.xyz), 0.);",

                //Additional layers. Tilt them so that the combination looks like a continuous band
				// wrapping around the object several times. 

                // tilted XY bands
                "vec3 pos = rotationXY( vec2( 0.2, 0.2 ) ) * vPosition.xyz;",
                "distort += DistortExtraLayer(pos);",

                // tilted YZ bands
                "pos.zyx = rotationXY( vec2( -0.2, -0.2 ) ) * vPosition.zyx;",
                "distort += DistortExtraLayer(pos);",

                // tilted XZ bands
                "pos.xzy = rotationXY( vec2( 0.3, 0.3 ) ) * vPosition.xzy;",
                "distort += DistortExtraLayer(pos);"

            ].join("\n");

        const WrapDistortDesigner =
            [
                "#if defined(WRAP_MATERIAL)",

                WrapDistort,

                //Disturb normal
                "normal += distort.xyz;",

                //Disturb diffuse color,
                "material.diffuseColor *= (0.5 + distort.a * 0.5);",

                "#endif"

            ].join("\n");

        const WrapDistortFilterColor =
            [
                "#if defined(WRAP_MATERIAL)",

                WrapDistort,

                //Disturb diffuse color,
                "diffColour *= (0.5 + distort.a * 0.5);",
                "diffColour.a += distort.a;",

                "#endif"

            ].join("\n");


        const WrapDistortNormal =
            [
                "#if defined(WRAP_MATERIAL)",

                WrapDistort,

                //Disturb normal
                "pN += distort.xyz;",

                "#endif"

            ].join("\n");

        const SegmentDistance =
            [
                "#if ANNOTATION_NUMBER",
                "float distToSegment(vec2 p, vec2 v, vec2 w)",
                "{",
                    "p=(p-v)*annotationsAspectRatio;",
                    "if (v == w)",
                    "return length(p);",

                    "w=(w-v)*annotationsAspectRatio;",
                    "float t = dot(p, w) / dot(w,w);",
                    "t = clamp(t, 0.0, 1.0);",
                    "return length(p - t * w);",
                "}",

                "vec4 texelRefFetch(int idx)",
                "{",
                    //Get index in the texture referential, i.e. x and y integer coordinates
                    "int reindexedX = idx - ANNOTATION_TEXTURE_SIZE * (idx / ANNOTATION_TEXTURE_SIZE);",//i.e. idx % ANNOTATION_TEXTURE_SIZE
                    "int reindexedY = idx / ANNOTATION_TEXTURE_SIZE;",
                    //we now need to convert these between 0 and 1
                    "vec2 loc = vec2((float(reindexedX))/float(ANNOTATION_TEXTURE_SIZE), (float(reindexedY))/float(ANNOTATION_TEXTURE_SIZE));",
                    "return texture2D(annotationsTexture, loc);",
                "}",

                "vec4 accelGridFetch(int idX, int idY)",
                "{",
                    //Get index in the texture referential and fetch the texel
                    "int indexCell = idX * ACCEL_GRID_SIZE + idY;",
                    "return texelRefFetch(indexCell);",
                "}",
                "#endif"
            ].join("\n");

        const AnnotationDrawing =
            [
                "#if ANNOTATION_NUMBER",
                    "bool annotated = false;",
                    "if(!isFloorReflection){",
                        "float linearCamZ = length(vWorldPosition.xyz - cameraPosition);",
                        "float thick = linearCamZ * annotationsThickness * 0.001;",// 1 millimeter

                        "int minAccelIdX = int(floor(clamp(texcoordVarying.x - thick / annotationsAspectRatio.x, 0., 1.) * float(ACCEL_GRID_SIZE)));",
                        "int maxAccelIdX = int(ceil(clamp(texcoordVarying.x + thick / annotationsAspectRatio.x, 0., 1.) * float(ACCEL_GRID_SIZE)));",
                        "int minAccelIdY = int(floor(clamp(texcoordVarying.y - thick / annotationsAspectRatio.y, 0., 1.) * float(ACCEL_GRID_SIZE)));",
                        "int maxAccelIdY = int(ceil(clamp(texcoordVarying.y + thick / annotationsAspectRatio.y, 0., 1.) * float(ACCEL_GRID_SIZE)));",

                        "for(int x = 0; x < ACCEL_GRID_SIZE; x++)",
                        "{",
                            //The two next lines are an ugly hack since for loops
                            //cannot accept non constant expressions as initializer
                            "if(x < minAccelIdX) continue;",
                            "if(annotated || x >= maxAccelIdX) break;",
                            "for(int y = 0; y < ACCEL_GRID_SIZE; y++)",
                            "{",
                                "if(y < minAccelIdY) continue;",
                                "if(y >= maxAccelIdY) break;",
                                //vec2 accelGridinfo:
                                // - r channel contains the number of line refs
                                // - g contains the index of the first line ref
                                "vec4 accelGridInfo = accelGridFetch(x,y);",
                                "if(int(accelGridInfo.r) == 0) continue;",

                                "if(int(accelGridInfo.r) == 1)",
                                "{",
                                    //IMPROVEME: this branch do not use the index recomposition used for the other branch of this if/else block
                                    //If some artifacts start to show on mobile devices when the number of annotations becomes high, it may be due
                                    //to the same limitation and we may have to also encode/decode it in the same manner (see
                                    //TextureManager:UpdateAnnotationsTexture function)
                                    "int lineStartIdx = int(accelGridInfo.g);",
                                    "vec4 segment = texelRefFetch(lineStartIdx + LINES_START_OFFSET ).rgba;",//we now have access to the segment we want to test distance against
                                    "float dist = distToSegment(texcoordVarying, segment.rg, segment.ba );",
                                    "if(dist < thick )",
                                    "{",
                                        "annotated = true;",
                                        "break;",
                                    "}",
                                "}",
                                "else",
                                "{",
                                    //Decode index (this allows to encode large numbers in half float textures for some mobile devices)
                                    "int lineStartIdxRecomposed = int(dot( accelGridInfo.gba, vec3(65535., 255., 1.)));",
                                    "for(int lineIdx = 0; lineIdx < MAX_LINES_PER_CELL ; lineIdx++)",
                                    "{",
                                        "if(annotated || lineIdx >= int(accelGridInfo.r)) break;",//is int() cast expensive ?
                                        "int lineRef = int(texelRefFetch(ACCEL_GRID_SIZE * ACCEL_GRID_SIZE + lineStartIdxRecomposed + lineIdx).r);",//r channel contains an integer reference to the line
                                        "vec4 segment = texelRefFetch(lineRef + LINES_START_OFFSET).rgba;",//we know have access to the segment we want to test distance against
                                        "float dist = distToSegment(texcoordVarying, segment.rg, segment.ba );",
                                        "if(dist < thick )",
                                        "{",
                                            "annotated = true;",
                                            "break;",
                                        "}",
                                    "}",
                                "}",
                            "}",
                        "}",
                        "if(annotated) reflResult = vec4(annotationsColor.r,annotationsColor.g,annotationsColor.b,0.);", // alpha is 0 to disable additive blending of annotations
                    "}",
                "#endif"
            ].join("\n");

        const FullEncodeDepthChunck = [
            "vec4 EncodeFloatRGBA( float v ) {",
                "float remainder = v;",
                "vec4 res, r;",

                "res[0] = floor(remainder * 255.0) / 255.0;",
                "remainder -= res[0];",
                "res[1] = floor(remainder * 65535.0) / 65535.0;",
                "remainder -= res[1];",
                "res[2] = floor(remainder * 16777215.0) / 16777215.0;",

                "r[0] = clamp(floor((v) * 255.0),0.,255.);",
                "r[1] = clamp(floor((v - res[0]) * 65535.0),0.,255.);",
                "r[2] = clamp(floor((v - res[0] - res[1]) * 16777215.0),0.,255.);",
                "r[3] = clamp(floor((v - res[0] - res[1] - res[2]) * 4294967295.0),0.,255.);",

                "r[0] = r[0] / 255.0;",
                "r[1] = r[1] / 255.0;",
                "r[2] = r[2] / 255.0;",
                "r[3] = r[3] / 255.0;",
                "return r;",
            "}"
        ].join("\n");

        const FullDecodeDepthChunck = [
            "float DecodeFloatRGBA( vec4 rgba ) {",
                "return dot( rgba, vec4(1.0, 1./255., 1./65535., 1./16777215.) );",
            "}"
        ].join("\n");

        const NoEncodeDepthChunck = [
            "vec4 EncodeFloatRGBA( float v ) {",
                "return vec4(v, 0., 0., 0.);",
            "}"
        ].join("\n");

        const NoDecodeDepthChunck = [
            "float DecodeFloatRGBA( vec4 rgba ) {",
                "return rgba.r;",
            "}"
        ].join("\n");

        const Ipad2EncodeDepthChunck = [
            "vec4 EncodeFloatRGBA( float v ) {",
            "float remainder = v;",
            "vec4 res, r;",

            "res[0] = floor(remainder * 127.0) / 127.0;",
            "remainder -= res[0];",
            "res[1] = floor(remainder * 16383.0) / 16383.0;",
            "remainder -= res[1];",
            "res[2] = floor(remainder * 2097151.0) / 2097151.0;",

            "r[0] = clamp(floor((v) * 127.0),0.,127.);",
            "r[1] = clamp(floor((v - res[0]) * 16383.0),0.,127.);",
            "r[2] = clamp(floor((v - res[0] - res[1]) * 2097151.0),0.,127.);",

            "r.r = r[0] / 127.0;",
            "r.g = r[1] / 127.0;",
            "r.b = r[2] / 127.0;",
            "r.a = 1.;",
            "return r;",
            "}"
        ].join("\n");

        const Ipad2DecodeDepthChunck = [
            "float DecodeFloatRGBA( vec4 rgba ) {",
                "return rgba.r + rgba.g/127. + rgba.b/16383.;",
            "}"
        ].join("\n");

        const Simulate32BitEncodingChunck = [
            "depth = floor(depth * 255.) * 0.003921568627450980392156862745098;"
        ];

        let EncodeDepthChunck;
        let DecodeDepthChunck;
        let SimulateEncodingChunckD0, SimulateEncodingChunckD1, SimulateEncodingChunckD2;

        LogMessage.Log("Configuration set up for:");
        if (WV3DSystemChecker.get("_fullSupportFloatingPointTextures"))//Desktop platforms
        {
            if (WV3DSystemChecker.IsWindows())//Windows
            {
                LogMessage.Log("Windows");
                if (WV3DSystemChecker.IsIE())//tested ok
                {
                    LogMessage.Log("IE");
                    SimulateEncodingChunckD0 = SimulateEncodingChunckD1 = SimulateEncodingChunckD2 = "";
                    EncodeDepthChunck = NoEncodeDepthChunck;
                    DecodeDepthChunck = NoDecodeDepthChunck;
                }
                else //tested ok -- back of some printable parts (clearspeed) cannot be seen however
                {
                    LogMessage.Log("Not IE");
                    SimulateEncodingChunckD0 = SimulateEncodingChunckD1 = SimulateEncodingChunckD2 = Simulate32BitEncodingChunck;
                    EncodeDepthChunck = FullEncodeDepthChunck;
                    DecodeDepthChunck = FullDecodeDepthChunck;
                }
            }
            else if (WV3DSystemChecker.IsMac())//Mac -- tested on 10.10.4 -- back of some printable parts (clearspeed) cannot be seen however
            {
                LogMessage.Log("Mac");
                SimulateEncodingChunckD0 = SimulateEncodingChunckD1 = SimulateEncodingChunckD2 = "";
                EncodeDepthChunck = NoEncodeDepthChunck;
                DecodeDepthChunck = NoDecodeDepthChunck;
            }
            else
            {
                LogMessage.Log("Unsupported OS (for now): " + WV3DSystemChecker.get("_ua").os.family + " / " + WV3DSystemChecker.get("_ua").browser.family);
                SimulateEncodingChunckD0 = SimulateEncodingChunckD1 = SimulateEncodingChunckD2 = Simulate32BitEncodingChunck;
                EncodeDepthChunck = FullEncodeDepthChunck;
                DecodeDepthChunck = FullDecodeDepthChunck;
            }
        }
        else//All Mobile device ?
        {
            if (WV3DSystemChecker.IsAndroid())//Android -- to test and ajust relatively to the GPU (needs to be tested on some android devices)
            {
                LogMessage.Log("Android");
                SimulateEncodingChunckD0 = SimulateEncodingChunckD1 = SimulateEncodingChunckD2 = "";
                EncodeDepthChunck = Ipad2EncodeDepthChunck;
                DecodeDepthChunck = Ipad2DecodeDepthChunck;
            }
            else if (WV3DSystemChecker.IsiOSFirstGen())//Ipad 2 -- tested (still a few z-fighting issues on back of some printable parts, e.g. speed clear)
            {
                LogMessage.Log("iOS First Generation");
                SimulateEncodingChunckD0 = SimulateEncodingChunckD1 = SimulateEncodingChunckD2 = Simulate32BitEncodingChunck;
                EncodeDepthChunck = Ipad2EncodeDepthChunck;
                DecodeDepthChunck = Ipad2DecodeDepthChunck;
            }
            else //iPhone -- to test -- iPod -- tested ok (minor z-fighting issue on back of labels)
            {
                LogMessage.Log("Default Mobile Device");
                SimulateEncodingChunckD0 = SimulateEncodingChunckD1 = SimulateEncodingChunckD2 = "";
                EncodeDepthChunck = Ipad2EncodeDepthChunck;
                DecodeDepthChunck = Ipad2DecodeDepthChunck;
                //todo, try FullencodeB here and for IE and mac ?
            }
        }

        const DeclareDepthSamplerChunck = [
            "#ifdef GL_FRAGMENT_PRECISION_HIGH",
                "uniform highp sampler2D D0RT;",
            "#else",
                "uniform mediump sampler2D D0RT;",
            "#endif",
            "uniform vec2 canvas_size;"
        ].join("\n");

        const DeclareDepthSamplersChunck = [
            "uniform vec2 canvas_size;",
            "uniform int NBRTSRC;",
            "#ifdef GL_FRAGMENT_PRECISION_HIGH",
                "uniform highp sampler2D DX1RT;",
                "uniform highp sampler2D DX2RT;",
            "#else",
                "uniform mediump sampler2D DX1RT;",
                "uniform mediump sampler2D DX2RT;",
            "#endif"
        ].join("\n");

        const SelectPeelChunck = [
                "vec2 loc = vec2(gl_FragCoord.x/canvas_size.x, gl_FragCoord.y/canvas_size.y);",
                "vec4 dx1_rgba = texture2D(DX1RT, loc);",
                "vec4 dx2_rgba = texture2D(DX2RT, loc);",//Even if unused, needs to stay out of control flow
                "if(canvas_size.x > 0. && NBRTSRC > 0)",
                "{",
                    "float z = DecodeFloatRGBA(EncodeFloatRGBA(1.0 - gl_FragCoord.z));",

                    "if (NBRTSRC == 2)",
                    "{",
                        "float zmax = DecodeFloatRGBA(dx2_rgba);",

                        "if (z >= zmax)",
                        "{",
                            "discard;",
                        "} ",
                        "float zmin = DecodeFloatRGBA(dx1_rgba);",

                        "if (z <= zmin)",
                        "{",
                            "discard;",
                        "} ",
                    "}else if (NBRTSRC == 1){",
                        "float zmin = DecodeFloatRGBA(dx1_rgba);",

                        "if (z <= zmin)",
                        "{",
                            "discard;",
                        "} ",
                    "}",
                "}"
            ].join("\n");

        const ReflectionFadeChunck = [
            "if(reflection_fade)",
            "{",
                //Calculate start and end of fading according to reflection_fade_size
                //reflection_fade_size 1. --> sharp line
                //reflection_fade_size 2. --> fade takes half the height of the screen
                //get the distance of loc.y to the start of the fade (loc.y - startfade)
                //interpolate linearly

                "float inv_position = 1. - reflection_fade_position;",
                "float offset = (reflection_fade_size - 1.) * 0.1;",
                "float fade_start = inv_position + offset;",
                "float fade_end = inv_position - offset;",
                "float factor = (fade_start - loc.y) / (fade_end - fade_start) + 1.;",

                "if (loc.y > fade_start) refopacity *= 1.;",//keep it as it is
                "else if (loc.y < fade_end) refopacity *= 0.;",//Beyond expected fade limit
                "else refopacity *= factor ;",//Faded using a linear function
            "}"
        ].join("\n");

        const ReflectionCosineFadeChunck = [
            "if(reflection_fade)",
            "{",
                //Calculate start and end of fading according to reflection_fade_size
                //reflection_fade_size 1. --> sharp line
                //reflection_fade_size 2. --> fade takes half the height of the screen
                //get the distance of loc.y to the start of the fade (loc.y - startfade)
                //divide by the length (startfade-endfade) to get a value between 0 and PI
                //Then calculate the opacity according to a cosine function

                "float inv_position = 1. - reflection_fade_position;",
                "float offset = (reflection_fade_size - 1.) * 0.1;",
                "float fade_start = inv_position + offset;",
                "float fade_end = inv_position - offset;",
                "float factor = (fade_start - loc.y) / (fade_start - fade_end) * 3.1415926535;",

                "if (loc.y > fade_start) refopacity *= 1.;",//keep it as it is
                "else if (loc.y < fade_end) refopacity *= 0.;",//Beyond expected fade limit
                "else refopacity *= cos(factor)*0.5 + 0.5 ;",//Faded using a cosine function
            "}"
        ].join("\n");

        const EnvironmentRotationChunck = [
            "vec3 rotate_vec_from_quat(vec4 envMapRotQuat, vec3 vector)",
            "{",
                //fast quaternion multiplication. source: https://molecularmusings.wordpress.com/2013/05/24/a-faster-quaternion-vector-multiplication/
                "vec3 tempVec = 2.0 * cross(envMapRotQuat.xyz, vector);",
                "vector = vector + envMapRotQuat.w * tempVec + cross(envMapRotQuat.xyz, tempVec);",
                "return vector;",
            "}"
        ].join("\n");


        const ShaderUtils = bb.Model.extend({
            defaults:
                {
                    Noise: Noise,
                    Distort: Distort,
                    WrapDistortDesigner: WrapDistortDesigner,
                    WrapDistortFilterColor: WrapDistortFilterColor,
                    WrapDistortNormal: WrapDistortNormal,

                    SegmentDistance: SegmentDistance,
                    AnnotationDrawing: AnnotationDrawing,

                    EncodeDepthChunck: EncodeDepthChunck,
                    DecodeDepthChunck: DecodeDepthChunck,
                    SimulateEncodingChunckD2: SimulateEncodingChunckD2,
                    SimulateEncodingChunckD1: SimulateEncodingChunckD1,
                    SimulateEncodingChunckD0: SimulateEncodingChunckD0,

                    DeclareDepthSamplerChunck: DeclareDepthSamplerChunck,
                    DeclareDepthSamplersChunck: DeclareDepthSamplersChunck,

                    SelectPeelChunck: SelectPeelChunck,
                    ReflectionFadeChunck: ReflectionFadeChunck,
                    ReflectionCosineFadeChunck: ReflectionCosineFadeChunck,
                    EnvironmentRotationChunck: EnvironmentRotationChunck
                }
        });

        return new ShaderUtils;//New is what makes it a singleton, see usage in shaders

    });


