import Tooltip from "@esko/tooltip/Tooltip";
import $ from "jquery";
import _ from "underscore";
/**
 * A Button is an interface element which provides an action and can have a text, content (can be an element or a string).
 * Specific Buttons extend this class for specific buttons.
 *
 * @exports Button
 * @constructor
 * @param {I18n} i18n - The language string provider
 * @param {JQueryObj} tooltipContainer - The container and restrict element of the tooltip for the button
 * @param {Boolean} isMobile - True if we are on mobile, false otherwise
 * @param {String} Name - The name of the button
 * @param {String | JQueryObj } content - The text to be shown as button, or your own jQuery element.
 * @param {Boolean} [forMobile] - If true the button should react on touchend/touchstart (otherwise the default "click" is used)
 * @param {Function} [action] - The function which is called when the button is pressed (also see listenToTool). Context (this) is the button.
 */
var Button = function(i18n, tooltipContainer, isMobile, name, content, forMobile, action) {
	this._i18n = i18n;
	this._tooltipContainer = tooltipContainer;
	this._isMobile = isMobile;

	this.name = name;
	this.content = content;
	this.forMobile = forMobile === undefined ? true : forMobile;
	this.action = action;
	this.element = $("<a/>").attr("name", name);
	this.enabled = true;

	this.tooltip = undefined;
	this._enabledTooltipContent = [];
	this._disabledTooltipContent = [];

	if (this.forMobile && this._isMobile) {
		var eventType = "touchend";
		if (window.PointerEvent) {
			eventType = "pointerup";
		}
		this.element.on(eventType, _.bind(this.onClick, this));
	} else {
		this.element.on("click", _.bind(this.onClick, this));
	}
	this.element.on("contextmenu", this.onContextMenu.bind(this));

	this.element.append(this.content);
};

Button.prototype.addClass = function(cssClass) {
	this.element.addClass(cssClass);
};

/**
 * Change button content
 * @param content
 */
Button.prototype.setContent = function(content) {
	this.content = content;
	this.element.empty();
	this.element.append(this.content);
};

Button.prototype.onContextMenu = function(e) {
	if (this.tooltip) {
		this.tooltip.hide();
	}
};

/**
 * Handler for click event, meaning executing the action
 * @param {DOMEvent} e - Event
 */
Button.prototype.onClick = function(e) {
	if (!this.isEnabled()) {
		return;
	}
	if (this.tooltip) {
		this.tooltip.hide();
	}
	if (this.action) {
		this.action(e);
	}

	// for mobile we need the events to bubble up, for desktop we don't want this bubbling
	if (!this._isMobile) {
		e.stopPropagation();
	}
};

/**
 * Disable/enable the button
 * @param {Boolean} enabled
 */
Button.prototype.setEnabled = function(enabled) {
	this.element.toggleClass("disabled", !enabled);

	if (this.tooltip !== undefined) {
		if (this._disabledTooltipContent.length > 0) {
			this.tooltip.enableTooltip();
			// we have a changing tooltip when the button is disabled
			if (enabled) {
				this.tooltip.setTooltipContent(this._enabledTooltipContent[0], this._enabledTooltipContent[1], this._enabledTooltipContent[2]);
			} else {
				this.tooltip.setTooltipContent(this._disabledTooltipContent[0], this._disabledTooltipContent[1], this._disabledTooltipContent[2]);
			}
		} else {
			// the tooltip should be disabled when the button is disabled
			if (!enabled) {
				this.tooltip.disableTooltip();
			} else {
				this.tooltip.enableTooltip();
			}
		}
	}

	this.enabled = enabled;
	this._updateIcon();
};

/**
 * Is the button enabled?
 * @returns {Boolean}
 */
Button.prototype.isEnabled = function() {
	return this.enabled;
};

/**
 * TODO: Content can be changed here when button gets disabled, at the moment only used in MiniButton
 */
Button.prototype._updateIcon = function() {};

/**
 * Set the action of the button
 * @param {Function} newAction
 */
Button.prototype.setAction = function(newAction) {
	this.action = newAction;
};

/**
 * Set Tooltip for button.
 * @param {String} tooltipOrBundle - tooltip to show if langKey is undefined or language bundle
 * @param {String} [langKey] - not required
 * @param {Array.<String>} [stringArgs] - not required, parameters that need to be filled in the language string
 * @param {Object} [popoverOptions] Additional options to pass to the tooltip
 */
Button.prototype.setTooltip = function(tooltipOrBundle, langKey, stringArgs, popoverOptions) {
	popoverOptions = popoverOptions || {};
	this._enabledTooltipContent = [tooltipOrBundle, langKey, stringArgs];
	if (this.tooltip === undefined) {
		this.tooltip = new Tooltip(
			this._i18n,
			this.element,
			popoverOptions.container || this._tooltipContainer,
			this._tooltipContainer,
			this._isMobile,
			tooltipOrBundle,
			langKey,
			stringArgs,
			popoverOptions
		);
	} else {
		if (this.enabled) {
			this.tooltip.setTooltipContent(tooltipOrBundle, langKey, stringArgs);
		}
	}
};

/**
 * Set the disabled Tooltip for button (only works when there is also an enabled tooltip!) this tooltip is shown when the button is disabled
 * @param {String} tooltipOrBundle - tooltip to show if langKey is undefined or language bundle
 * @param {String} [langKey] - not required
 * @param {Array.<String>} [stringArgs] - not required, parameters that need to be filled in the language string
 */
Button.prototype.setDisabledTooltip = function(tooltipOrBundle, langKey, stringArgs) {
	this._disabledTooltipContent = [tooltipOrBundle, langKey, stringArgs];
	if (!this.enabled && this.tooltip) {
		this.tooltip.enableTooltip();
		this.tooltip.setTooltipContent(tooltipOrBundle, langKey, stringArgs);
	}
};

/**
 * Hide the tooltip of the button if any defined
 */
Button.prototype.hideTooltip = function() {
	if (this.tooltip) {
		this.tooltip.hide();
	}
};

/**
 * Is button active
 * @returns {Boolean} True if button is active, false otherwise
 */
Button.prototype.isActive = function() {
	return this.element.hasClass("active");
};

/**
 * Set the active status of the button
 * @param {Boolean} active - True to set active, false to deactivate
 */
Button.prototype.setActive = function(active) {
	if (!this.isEnabled()) {
		return;
	}
	if (active) {
		// If already active, we should not re-activate it to avoid flickering
		if (!this.isActive()) {
			this.element.addClass("active");
		}
	} else {
		if (this.isActive()) {
			this.element.removeClass("active");
		}
	}
};

/**
 * Called when the button is moved
 */
Button.prototype.moved = function() {
	// do nothing
};

/**
 * Show the button
 */
Button.prototype.show = function() {
	this.element.show();
};

/**
 * Hide the button
 */
Button.prototype.hide = function() {
	this.element.hide();
};

export default Button;
