import Observable from "@esko/observable";
import _ from "underscore";
import $ from "jquery";
import "spectrum-colorpicker/spectrum";
import i18n from "webi18n-loader?bundle=colorpickerinput/locales!";
import "spectrum-colorpicker/spectrum.css";

/**
 * Create an input group with a good colorpicker with swatches
 * You can listen for changes on the "changed" event
 * @mixes{Observable}
 * @param {Object} options see spectrum.js
 */
var ColorPickerInput = function(options) {
	Observable.mixin(this, "ColorPicker");

	// default options
	this._options = _.extend(
		{
			showInput: true,
			showInitial: true,
			allowEmpty: true,
			showPalette: true,
			preferredFormat: "hex",
			localStorageKey: "wcr.colorpicker",
			palette: [
				["#5cb85c", "#5bc0de", "#f0ad4e", "#d9534f"],
				["#334444", "#aabb11", "#7F7F7F", "#17BECF"],
				["#1F77B4", "#FF7F0E", "#2CA02C", "#D62728"],
				["#9467BD", "#8C564B", "#E377C2", "#BCBD22"]
			],
			hideAfterPaletteSelect: true,
			maxSelectionSize: 4,
			replacerClassName: "input-group-addon",
			cancelText: i18n.t("CANCEL"),
			chooseText: i18n.t("CHOOSE"),
			togglePaletteMoreText: i18n.t("PALETTE_MORE"),
			togglePaletteLessText: i18n.t("PALETTE_LESS"),
			clearText: i18n.t("CLEAR"),
			noColorSelectedText: i18n.t("NO_COLOR_SELECTED")
		},
		options
	);

	this._allowChangeEvent = true;

	var changeListener = this._options.change;
	this._options.change = function() {
		var color = this._colorPicker.spectrum("get");
		if (color === null) {
			color = "";
		} else {
			color = color.toHexString();
		}
		this._input.val(color);
		if (this._allowChangeEvent) {
			this._emit("changed", color);
		}
		if (changeListener) {
			changeListener(color);
		}
	}.bind(this);

	this._element = $("<div>", {
		class: "input-group colorpickerinput"
	});
	this._input = $("<input>", {
		class: "form-control",
		type: "text",
		title: i18n.t("TOOLTIP")
	}).on(
		"change",
		function() {
			var value = this._input.val();

			if (value === "" || this._validateHex(value)) {
				this.setValue(value);
				this._emit("changed", value);
			}
		}.bind(this)
	);

	this._colorPicker = $("<input>", {
		class: "input-group-addon"
	});
	this._element.append(this._input).append(this._colorPicker);
	this._colorPicker.spectrum(this._options);
};

/**
 * Get the element
 * @returns {JQueryObject}
 */
ColorPickerInput.prototype.getElement = function() {
	return this._element;
};

/**
 * Get the Value
 * @returns {String} hex color string
 */
ColorPickerInput.prototype.getValue = function() {
	return this._input.val();
};

/**
 * Set the value
 * @param {String} value - hex color string
 */
ColorPickerInput.prototype.setValue = function(value) {
	this._input.val(value);
	this._allowChangeEvent = false;
	this._colorPicker.spectrum("set", value);
	this._allowChangeEvent = true;
};

/**
 * Enable/disable the colorpicker
 * @param {Boolean} disabled
 */
ColorPickerInput.prototype.setDisabled = function(disabled) {
	this._input.prop("disabled", disabled);
	this._disabled = disabled;
	if (disabled) {
		this._colorPicker.spectrum("disable");
	} else {
		this._colorPicker.spectrum("enable");
	}
};

/**
 * Checks if the given hex string is valid (eg. #fff or #ffffff)
 * @private
 * @param   {string} hex
 * @returns {boolean} True if valid, otherwise false
 */
ColorPickerInput.prototype._validateHex = function(hex) {
	return /^#([0-9a-f]{3}){1,2}$/i.test(hex);
};

/**
 * Is the color picker input disabled?
 * @returns {Boolean}
 */
ColorPickerInput.prototype.isDisabled = function() {
	return this._disabled;
};

/**
 * Add  input-group addon
 * @param {DOMElement} addon
 */
ColorPickerInput.prototype.addAddon = function(addon) {
	this._element.append(addon);
};

/**
 * WCR Init method
 * @param {object} element [[Description]]
 */
ColorPickerInput.wcrinit = function(element) {
	var originalElement = $(element);
	var picker = new ColorPickerInput();
	picker.setValue(originalElement.val());
	originalElement.data("colorpickerinput", picker);

	var handleChanged = true;
	picker.on("changed", function() {
		handleChanged = false;
		originalElement.val(picker.getValue());
		originalElement.trigger("change");
		handleChanged = true;
	});

	originalElement.on("changed", function() {
		if (handleChanged) {
			picker.setValue(originalElement.val());
		}
	});

	originalElement.attr("type", "hidden");
	originalElement.after(picker.getElement());
};

export default ColorPickerInput;
