import Block from "./core/Block";
import Layout from "./core/Layout";
import _ from "underscore";
import "./Dashboard.css";
/**
 * Dashboard has a (interchangeable) layout, which displays other blocks
 * @constructor
 * @class
 * @augments Block
 * @param {String} id
 * @param {Context} [context]
 */
var Dashboard = function(id, context) {
	Block.call(this, id, context);

	this._layoutID = null;
	this._layoutPromise = null;
	this._layout = null;
	this._layoutError = null;

	this._builderPromise = null;
	this._builder = null;
	this._builderError = null;

	this._observe(this._context, "editModeActive", this._render);
};

Dashboard.prototype = Object.create(Block.prototype);
Dashboard.prototype.constructor = Dashboard;

Dashboard.prototype.getType = function() {
	return "dashboard/Dashboard";
};

/**
 * @Override
 */
Dashboard.prototype._doConfigure = function(config, idMap) {
	var self = this;

	return this._context.getLanguageBundle("Dashboard").then(function(bundle) {
		self._i18n = bundle;
		// initialize the layout
		return self._setLayout(config.layout, idMap);
	});
};

/**
 * Make sure to start loading the builder when intialized in edit mode
 */
Dashboard.prototype._doInit = function() {
	if (this._context.isEditModeActive()) {
		this._loadBuilder();
	}
};

/**
 * @Override
 */
Dashboard.prototype._getBlockConfiguration = function(willPersist) {
	return this._layoutPromise
		.then(function(layout) {
			return layout.getConfiguration(willPersist);
		})
		.then(function(layoutConfig) {
			return {
				layout: layoutConfig
			};
		});
};

/**
 * All blocks which are not layoutblocks and persistent should be included
 * Layout blocks should be stored separately in the dashboard config
 * @private
 * @returns {Promise.<Array.<Block>>}
 */
Dashboard.prototype._findChildrenForPersistence = function() {
	return this._layoutPromise.then(function(layout) {
		// find all persistent children, which are not Layout blocks
		return layout.findChildren(
			function(block) {
				return block.isPersistent() && !(block instanceof Layout);
			},
			true,
			true
		);
	});
};

/**
 * Initialize the layout of the Dashboard
 * @private
 * @param   {Object}  JSONLayout - Object definition of the layout
 * @returns {Promise} resolved when layout is initialized
 */
Dashboard.prototype._setLayout = function(JSONLayout, IDMap) {
	if (this._layout) {
		this.detachChild(this._layout);
		this._layout = null;
	}

	var newLayoutID = (this._layoutID = this._context.registerBlock(JSONLayout, IDMap));
	this._layoutPromise = this._context.loadBlock(this._layoutID);
	this._layoutError = null;

	// register that we are loading the layout
	var whenFutureLayoutAvailable = this._registerFutureChild(newLayoutID);

	this._render();

	var self = this;
	return this._layoutPromise
		.then(function(layout) {
			if (self._layoutID === newLayoutID) {
				self._layout = layout;
				self._render();
			}
		})
		.catch(function(error) {
			if (self._layoutID === newLayoutID) {
				if (console) {
					console.log("Error initializing layout for Dashboard", arguments);
				}

				// Something went wrong, display user friendly error
				self._layoutError = self._generateUserfriendlyException(error, JSONLayout);
				self._render();
			}
		})
		.then(function() {
			self.once("render", whenFutureLayoutAvailable);
		});
};

/**
 * Start loading the Dashboard builder
 * @private
 */
Dashboard.prototype._loadBuilder = function() {
	var self = this;

	var builderConfig = {
		type: "dashboard/builder/DashboardBuilder",
		config: {
			dashboard: self,
			layout: null
		}
	};
	var whenFutureBuilderAvailable = this._registerFutureChild("BUILDER");

	this._layoutPromise.then(function() {
		// only now the layout is available
		builderConfig.config.layout = self._layout;
		self._builderPromise = self._context
			.registerAndLoadBlock(builderConfig)
			.then(function(builder) {
				self.once("render", whenFutureBuilderAvailable);
				return builder;
			})
			.then(function(builder) {
				self._builder = builder;
				self._render();
			})
			.catch(function(error) {
				// Something went wrong, display user friendly error
				self._builderError = self._generateUserfriendlyException(error, builderConfig);
				self._render();
			});
	});
};

/**
 * Get a reference to the layout of the dashboard
 * @returns {Promise.<Layout>}
 */
Dashboard.prototype.getLayout = function() {
	return this._layoutPromise;
};

/**
 * Prepend a block to the dashboard
 * @param   {String}  blockID
 * @returns {Promise} resolved once the block is prepended
 */
Dashboard.prototype.prependBlock = function(blockID) {
	return this._layoutPromise.then(function(layout) {
		return layout.prependBlock(blockID);
	});
};

/**
 * Prepend a list of blocks to the dashboard
 * @param   {Array.<String>}  blockID
 * @returns {Promise} resolved once the block is prepended
 */
Dashboard.prototype.prependBlocks = function(blocks) {
	return this._layoutPromise.then(function(layout) {
		return layout.prependBlocks(blocks);
	});
};

/**
 * Update the rendering of the Dashboard
 * This is called whenever the layout or builder is loaded (or the state changes from loading to loaded/error)
 * @private
 */
Dashboard.prototype._doRender = function() {
	// empty element by detaching its children
	this._element.contents().detach();

	if (this._context.isEditModeActive()) {
		if (this._layout) {
			this._layout.setInDOM(false);
		}
		if (this._builder) {
			this._registerChild(this._builder);
			this._element.append(this._builder.getElement());
			this._builder.activate(true);
			this._builder.setInDOM(this.isInDOM());
		} else if (this._builderError) {
			this._element.append(this._builderError);
		} else {
			if (!this._builderPromise) {
				this._loadBuilder();
			}
			// loading...
			this._element.text(this._i18n.t("LOADING_BLOCK"));
		}
	} else {
		if (this._builder) {
			this._builder.activate(false);
			this._builder.setInDOM(false);
		}

		if (this._layout) {
			this._registerChild(this._layout);
			this._element.append(this._layout.getElement());
			this._layout.setInDOM(this.isInDOM());
		} else if (this._layoutError) {
			this._element.append(this._layoutError);
		} else {
			// loading...
			this._element.text(this._i18n.t("LOADING_BLOCK"));
		}

		var calculateDimensions = this._context.getOption("calculateDimensions");
		if (calculateDimensions) {
			var height = calculateDimensions(this._element);
			this._element.css("height", Math.ceil(height) + "px");
			this._layout.getElement().css("height", "100%");
		}
	}
};

Dashboard.prototype._generateUserfriendlyException = function(error, blockDefinition, extraInformation) {
	var div = document.createElement("div");
	div.className = "alert alert-danger";

	this._context.getLanguageBundle("Dashboard").then(function(i18n) {
		var errorMessage = document.createElement("p");
		errorMessage.textContent = i18n.t("ERROR_UNEXPECTED_BLOCK");
		div.appendChild(errorMessage);

		if ((error && (error.message || error.stack)) || blockDefinition || extraInformation) {
			var detailsLink = document.createElement("a");
			detailsLink.textContent = i18n.t("ERROR_UNEXPECTED_DETAILS");

			var isDetailsOpen = false;

			var errorDetails = null;

			detailsLink.addEventListener("click", function() {
				if (isDetailsOpen) {
					div.removeChild(errorDetails);
					detailsLink.textContent = i18n.t("ERROR_UNEXPECTED_DETAILS");
					isDetailsOpen = false;
				} else {
					errorDetails = document.createElement("pre");

					var errorDetailsSections = [];

					if (error.message) {
						errorDetailsSections.push(error.message);
					}

					if (error.stack) {
						errorDetailsSections.push(error.stack);
					}

					if (blockDefinition) {
						try {
							errorDetailsSections.push("Block Definition: \r\n" + JSON.stringify(blockDefinition, null, "  ")); // no translation needed, this is information only needed for support/investigating the error
						} catch (e) {
							errorDetailsSections.push("Could not generate block definition \r\n" + e);
						}
					}
					if (extraInformation) {
						errorDetailsSections.push(extraInformation);
					}

					errorDetails.textContent = errorDetailsSections.join("\r\n\r\n---------------------------------------------\r\n\r\n");

					detailsLink.textContent = i18n.t("ERROR_UNEXPECTED_HIDE_DETAILS");
					div.appendChild(errorDetails);
					isDetailsOpen = true;
				}
			});

			div.appendChild(detailsLink);
		}
	});

	return div;
};

/**
 * @Override
 */
Dashboard.prototype._childInDOM = function(child) {
	return child == (this._context.isEditModeActive() ? this._builder : this._layout);
};

export default Dashboard;
