import Block from "../core/Block";
import _ from "underscore";
/**
 * Basic implementation of an Error Block
 * This block can display an error
 *
 * @class
 * @constructor
 * @augments Block
 * @param {String} id - ID
 * @param {Context} [context]
 * @param {String} message - the error message
 * @param {BlockDefinition} blockDefinition
 * @param {String} errorDetails - extra information that should be shown in the error details
 */
var ErrorBlock = function(id, context, message, blockDefinition, errorDetails) {
	Block.call(this, id, context);

	this._message = message;

	this._blockDefinition = blockDefinition;

	this._errorDetails = errorDetails || null;
};

ErrorBlock.prototype = Object.create(Block.prototype);
ErrorBlock.prototype.constructor = ErrorBlock;

/**
 * @Override
 */
ErrorBlock.prototype.getType = function() {
	return "dashboard/blocks/ErrorBlock";
};

/**
 * @Override
 */
ErrorBlock.prototype.getConfiguration = function() {
	return this._blockDefinition.config;
};

/**
 * Configure the default parameters of the block
 * @override
 */
ErrorBlock.prototype._initializeDefaultParameters = function() {};

/**
 * @Override
 */
ErrorBlock.prototype._doInit = function(config) {
	this._error = this._generateUserfriendlyException(this._message, this._blockDefinition, this._errorDetails);
	this._render();
};

/**
 * @Override
 */
ErrorBlock.prototype._doRender = function() {
	this._element.html(this._error);
};

/**
 * @Override
 */
ErrorBlock.prototype.isPersistent = function() {
	return !_.isEmpty(this._blockDefinition.config);
};

/**
 * @Override
 */
ErrorBlock.prototype.canConfigure = function() {
	return false;
};

/**
 * Generate a HTML element for displaying an unexpected exception
 * The error message shows a user friendly message with a button to display more details of the error.
 * @param {Error} [error] - the error that was thrown
 * @param {Object} [blockDefinition] - Object that describes the block for which an error occurred
 * @param {String} [extraInformation] - extra information that should be shown in the error details
 * @returns {HTMLObject} element to display the error
 */
ErrorBlock.prototype._generateUserfriendlyException = function(error, blockDefinition, extraInformation) {
	var div = document.createElement("div");
	div.className = "alert alert-danger";

	this._context.getLanguageBundle("Dashboard").then(function(i18n) {
		var errorMessage = document.createElement("p");
		errorMessage.textContent = i18n.t("ERROR_UNEXPECTED_BLOCK");
		div.appendChild(errorMessage);

		if ((error && (error.message || error.stack)) || blockDefinition || extraInformation) {
			var detailsLink = document.createElement("a");
			detailsLink.textContent = i18n.t("ERROR_UNEXPECTED_DETAILS");

			var isDetailsOpen = false;

			var errorDetails = null;

			detailsLink.addEventListener("click", function() {
				if (isDetailsOpen) {
					div.removeChild(errorDetails);
					detailsLink.textContent = i18n.t("ERROR_UNEXPECTED_DETAILS");
					isDetailsOpen = false;
				} else {
					errorDetails = document.createElement("pre");

					var errorDetailsSections = [];

					if (error.message) {
						errorDetailsSections.push(error.message);
					}

					if (error.stack) {
						errorDetailsSections.push(error.stack);
					}

					if (blockDefinition) {
						// no translation needed, this is information only needed for support/investigating the error
						try {
							errorDetailsSections.push("Block Definition: \r\n" + JSON.stringify(blockDefinition, null, "  "));
						} catch (e) {
							errorDetailsSections.push("Could not generate block definition \r\n" + e);
						}
					}
					if (extraInformation) {
						errorDetailsSections.push(extraInformation);
					}

					errorDetails.textContent = errorDetailsSections.join("\r\n\r\n---------------------------------------------\r\n\r\n");

					detailsLink.textContent = i18n.t("ERROR_UNEXPECTED_HIDE_DETAILS");
					div.appendChild(errorDetails);
					isDetailsOpen = true;
				}
			});

			div.appendChild(detailsLink);
		}
	});

	return div;
};

export default ErrorBlock;
