import ParameterConfigurationPanel from "../ParameterConfigurationPanel";
import SDrop from "component/sdrop/SDrop";
import StaticSDropDataProvider from "component/sdrop/dataproviders/StaticSDropDataProvider";
import Observable from "@esko/observable";
import $ from "jquery";
import _ from "underscore";
/**
 *
 * @param {Parameter} parameter
 * @param {Array.<Object>} data      This object defines a list of options in dropdowns,
 *                                   each of these options can have sub-dropdowns of their own.
 *                                   Example structure:
 *   [
 *       {
 *          "value":"ROLES",
 *          "text":"Roles",
 *          "children":[] //no sub-dropdowns
 *       },
 *       {
 *          "value":"PROJINFO",
 *          "text":"Project Info",
 *          "children":[ // when PROJINFO is selected then the next sub-dropdown will have these options
 *             {
 *                "value":"db_salesPerson",
 *                "text":"Salesperson"
 *             },
 *             {
 *                "value":"db_status",
 *                "text":"Status"
 *             }
 *          ]
 *       }
 *   ]
 *
 */
var CascadingDropdownConfigurationPanel = function(parameter, data) {
	Observable.mixin(this, "CascadingDropdownConfigurationPanel");
	this._optionsData = data;
	this._selectedValues = [];
	this._firstSelect = null;

	ParameterConfigurationPanel.call(this, parameter);
};

CascadingDropdownConfigurationPanel.prototype = Object.create(ParameterConfigurationPanel.prototype);
CascadingDropdownConfigurationPanel.prototype.constructor = CascadingDropdownConfigurationPanel;

/**
 * Initializes all the dropdowns and appends them to element
 * @private
 */
CascadingDropdownConfigurationPanel.prototype._init = function() {
	if (!_.isEmpty(this._optionsData)) {
		this._element.addClass("form-group");

		this._element.append(
			$("<label>", {
				text: this._parameter.getUserFriendlyName()
			})
		);

		this._inputGroup = $("<div>", {
			class: "input-group"
		});

		this._element.append(
			$("<div>", {
				class: "form-entry"
			}).append(this._inputGroup)
		);

		this._optionsData = this._sort(this._optionsData);

		// Start adding all dropdowns and sub-dropdowns recursively
		this._firstSelect = this._buildDropdowns(this._optionsData, null, 0);
		this._onDropdownChange(this._firstSelect, false);
	}
};

CascadingDropdownConfigurationPanel.prototype._getCustomConfiguration = function() {
	return {
		value: this._selectedValues[this._selectedValues.length - 1]
	};
};

/**
 * Builds a dropdown based on the given data object. This will recursively call itself to add sub-dropdowns.
 *
 * @private
 * @param   {Array.<Object>} data The data that is used to build the dropdowns. See constructor for example data structure.
 * @param {String} previousDropdownValue The value of the option in the previous cascading dropdown.
 * @param {Integer} level This variable is used recursively to indicate the current level.
 * @returns {jQuery} The very first select element in list of cascading dropdowns
 *
 */
CascadingDropdownConfigurationPanel.prototype._buildDropdowns = function(data, previousDropdownValue, level) {
	// Wrap select object in div with class "dropdown-wrapper", this is necessary to let sdrop function properly.
	// A dropdownWrapper contains a single select, this select also has a reference to his wrapper.
	var dropdownWrapper = $("<div>", {
		class: "dropdown-wrapper"
	});

	this._inputGroup.append(dropdownWrapper);
	if (previousDropdownValue) {
		dropdownWrapper.addClass("hidden");
	}

	var select = $("<select>");
	dropdownWrapper.append(select);

	// Add data attributes to select, this is used during the change event to find the next cascading dropdowns.
	if (previousDropdownValue) {
		select.attr("data-linked-dropdown", previousDropdownValue);
	}
	select.attr("data-dropdown-level", level);
	select.data("dropdown-wrapper", dropdownWrapper);

	// Try and restore the original value of the dropdown
	var selectedValue = null;
	if (this._requestedSelectedValues) {
		if (level === 0 || (level > 0 && this._requestedSelectedValues[level - 1] === previousDropdownValue)) {
			selectedValue = this._requestedSelectedValues[level];
		}
	}

	// Build options + sub-dropdowns recursively
	_.each(
		data,
		function(el) {
			if (el.children) {
				this._buildDropdowns(el.children, el.value, level + 1);
			}
		}.bind(this)
	);

	// Store sdrop data, will be used later when the sdrop is injected
	select.data("sdrop-data", {
		initialValue: selectedValue,
		data: data,
		previousDropdownValue: previousDropdownValue,
		level: level
	});

	// Add change event listener
	select.on(
		"change",
		_.partial(function(self) {
			self._onDropdownChange($(this), true);
		}, this)
	);

	return select;
};

/**
 * This listener is fired when a dropdown changes selection
 * @private
 * @param {jQuery} select The select that changed value
 */
CascadingDropdownConfigurationPanel.prototype._onDropdownChange = function(select, doEmitEvent) {
	// Build the sdrop only when requested
	if (!select.data("sdrop")) {
		var sdropData = select.data("sdrop-data");
		var dataStore = this._getDataStore(sdropData.data, sdropData.previousDropdownValue, sdropData.level);
		select.data(
			"sdrop",
			new SDrop(select, null, dataStore, {
				initialValue: sdropData.initialValue ? sdropData.initialValue : dataStore.getData("*")[0].id,
				initializeDataProvider: false,
				manageStyleAndWidth: false,
				zIndex: 9999
			})
		);
	}
	select.data("dropdown-wrapper").removeClass("hidden");
	var level = parseInt(select.attr("data-dropdown-level"), 10);
	var dropdownWrapper = select.data("dropdown-wrapper");
	var nextDropdowns = dropdownWrapper.nextAll(".dropdown-wrapper");

	//Hide all next dropdowns
	nextDropdowns.addClass("hidden");

	// Remove all next selected values
	this._selectedValues.splice(level);

	//Store the selected value object in array on index level
	this._selectedValues[level] = {
		value: select.val(),
		label: select.find("option:selected").text()
	};

	//Find the next dropdown that corresponds to current selected value
	var nextSelect = nextDropdowns.find("select").filter(function() {
		return (
			$(this).attr("data-linked-dropdown") === select.val() &&
			parseInt($(this).attr("data-dropdown-level")) === parseInt(select.attr("data-dropdown-level"), 10) + 1
		);
	});

	//Show next dropdown (if it exists)
	if (nextSelect.length > 0) {
		dropdownWrapper.addClass("input-group-addon-sdrop");
		this._onDropdownChange(nextSelect, doEmitEvent);
	} else {
		dropdownWrapper.removeClass("input-group-addon-sdrop");
		if (doEmitEvent) {
			this._emit("changed");
		}
	}
};

/**
 * Creates datastores
 * @param {object}  data                  optionsData
 * @param {string}  previousDropdownValue id of previous dropdown value
 * @param {integer} level
 */
CascadingDropdownConfigurationPanel.prototype._getDataStore = _.memoize(
	function(data, previousDropdownValue, level) {
		var dataStore = new StaticSDropDataProvider();
		dataStore.setData(
			_.map(data, function(option) {
				return {
					id: option.value,
					name: option.text
				};
			}),
			true
		);
		return dataStore;
	},
	function(data, previousDropdownValue, level) {
		return level + "-" + previousDropdownValue;
	}
);

/**
 * Re-init the cascading dropdown with new options
 * @param {object}   data                    optionsData
 */
CascadingDropdownConfigurationPanel.prototype.setOptions = function(data, requestedSelectedValues) {
	this._optionsData = data;
	this._requestedSelectedValues = requestedSelectedValues;
	this._element.empty();
	this._init();
};

/**
 * Get the current selected values of the cascading dropdowns.
 * @returns {Array.<Object>} the selected values. Each objects in the array represents a level. An object contains the properties: label, value
 */
CascadingDropdownConfigurationPanel.prototype.getSelectedValues = function() {
	return this._selectedValues;
};

/**
 * Changes the cascading dropdowns to reflect the values of the given array
 * @returns {Array.<String>} array of the selected values
 */
CascadingDropdownConfigurationPanel.prototype.setSelectedValues = function(newSelectedValues) {
	var currentSelect = this._firstSelect;
	_.each(
		newSelectedValues,
		function(newValue) {
			if (currentSelect.length > 0 && currentSelect.val() !== newValue) {
				// Change sdrop current value to new value
				if (currentSelect.data("sdrop")) {
					currentSelect.data("sdrop").populate(newValue);
					this._onDropdownChange(currentSelect, false);
				}
			}

			// Look for new select
			var dropdownWrapper = currentSelect.data("dropdown-wrapper");
			var nextDropdowns = dropdownWrapper.nextAll(".dropdown-wrapper");
			currentSelect = nextDropdowns.find("select").filter(function() {
				return (
					$(this).attr("data-linked-dropdown") === currentSelect.val() &&
					parseInt($(this).attr("data-dropdown-level")) === parseInt(currentSelect.attr("data-dropdown-level"), 10) + 1
				);
			});
		}.bind(this)
	);
};

/**
 * Will recursively sort the data object and all children
 * @private
 * @param   {Array.<object>}    data See constructor for example data object
 * @returns {Array.<object>} sorted data object
 */
CascadingDropdownConfigurationPanel.prototype._sort = function(data) {
	var sortedDatas = data.sort(function(a, b) {
		return a.text.localeCompare(b.text);
	});
	_.each(
		sortedDatas,
		function(sortedData) {
			if (!_.isEmpty(sortedData.children)) {
				sortedData.children = this._sort(sortedData.children);
			}
		}.bind(this)
	);
	return sortedDatas;
};

export default CascadingDropdownConfigurationPanel;
