import ParameterConfigurationPanel from "../ParameterConfigurationPanel";
import UnitInputField from "../../../layout/flex/UnitInputField";
import $ from "jquery";
import _ from "underscore";
/**
 * Configuration panel for a number parameter
 * @augments ParameterConfigurationPanel
 * @param {Parameter} parameter
 */
var NumberParameterConfigurationPanel = function(parameter) {
	this._validNumber = parameter.getDefaultNumber();
	this._units = parameter.getUnits();

	ParameterConfigurationPanel.call(this, parameter);
};

NumberParameterConfigurationPanel.prototype = Object.create(ParameterConfigurationPanel.prototype);
NumberParameterConfigurationPanel.prototype.constructor = NumberParameterConfigurationPanel;

/**
 * By default a parameter is shown as a form-group with a label and an input field
 * @private
 */
NumberParameterConfigurationPanel.prototype._init = function() {
	this._$label = $("<label>", {
		text: this._parameter.getUserFriendlyName()
	});

	var constraints = this._parameter.getConstraints();
	if (this._units.length <= 1) {
		//Needed to provide compatibility with old implementation of cssSizeParameter
		var value = this._parameter.getValue();
		if (constraints.type === "int") {
			value = parseInt(this._parameter.getValue(), 10);
		} else if (constraints.type === "float") {
			value = parseFloat(this._parameter.getValue, 10);
		}

		this._element.addClass("form-group");

		this._$input = $("<input>", {
			type: "number",
			min: constraints.min,
			max: constraints.max,
			class: "form-control",
			value: value
		});
		var elementToBeAdded = this._$input;

		if (this._units.length === 1) {
			var _$inputgroup = $("<div>", {
				class: "input-group"
			});
			var _unitLabel = $("<span>", {
				class: "input-group-addon"
			}).append(this._units[0]);

			elementToBeAdded = _$inputgroup.append(this._$input).append(_unitLabel);
		}

		this._$input.attr("data-param-name", this._parameter.getName());

		this._$input.on(
			"blur",
			function(event) {
				if (this._$input.val() === "" || this._$input.val() === "-") {
					this._$input.val(this._parameter.getDefaultNumber());
				}
				if (this._parameter.getValue() !== this._validNumber.toString(10)) {
					this._parameter.undoableSetValue(this._validNumber.toString(10));
				}
			}.bind(this)
		);

		// Every change from the user is instantly checked and corrected if needed.
		this._$input.on(
			"input",
			function(event) {
				var defaultNumber = this._parameter.getDefaultNumber() ? parseInt(this._parameter.getDefaultNumber(), 10) : "";
				var value = this._$input.val();

				var number = value;
				var valid = true;
				if (value && !isNaN(Number(number))) {
					if (constraints) {
						if (constraints.type === "int") {
							number = parseInt(value, 10);
						} else if (constraints.type === "float") {
							number = parseFloat(value, 10);
						} else {
							//Wrong constraint type
							//TODO
						}
						if ("min" in constraints) {
							if (number < constraints.min) {
								//invalid => reset to min
								number = constraints.min;
							}
						}
						if ("max" in constraints) {
							if (number > constraints.max) {
								//invalid => rest to max
								number = constraints.max;
							}
						}
					}
					this._validNumber = number;
				} else {
					this._validNumber = defaultNumber;
				}

				if (value !== "" && value !== "-") {
					//allow input of the "-" for negative numbers
					this._$input.val(this._validNumber);
				}
			}.bind(this)
		);

		this._observeParam(
			function() {
				this._$input.val(this._parameter.getValue());
			}.bind(this)
		);

		this._element.append(this._$label).append(
			$("<div>", {
				class: "form-entry"
			})
				.append(elementToBeAdded)
				.append(this.getHelpElement())
		);
	} else {
		this._$input = new UnitInputField(
			this._parameter.getUserFriendlyName(),
			this._parameter.getValue(),
			this._units,
			true,
			{
				max: constraints.max,
				min: constraints.min
			},
			this.getHelpElement()
		);

		this._$input.on(
			"changed",
			_.bind(function() {
				this._parameter.undoableSetValue(this._$input.getValue());
			}, this)
		);
		this._$input.setValue(this._parameter.getValue());

		this._observeParam(
			function() {
				this._$input.setValue(this._parameter.getValue());
			}.bind(this)
		);

		this._element.append(this._$input.element);
	}
};

/**
 * Creates JQuery element to represent the help text of the parameter
 * @returns {JQueryObject}
 */
NumberParameterConfigurationPanel.prototype.getHelpElement = function() {
	var _$help = $();
	if (this._parameter.getHelpText()) {
		_$help = $("<div>", {
			class: "help-block"
		});
		_$help.append(
			$("<div>", {
				text: this._parameter.getHelpText(),
				class: "pull-left"
			})
		);
		if (this._parameter.getToolTipText()) {
			_$help.append(
				$("<div>", {
					class: "icon icon-WCR_Info-round12 icon-grey pull-left",
					title: this._parameter.getToolTipText()
				})
			);
		}
	}
	return _$help;
};

export default NumberParameterConfigurationPanel;
