import XDropDataProvider from "@esko/xdrop/dataproviders/XDropDataProvider";
import ResultEntry from "@esko/xdrop/dataproviders/ResultEntry";
import Observer from "@esko/observer";
import _ from "underscore";
/**
 * A Simple dataprovider which is a composite and applies a filter to the results of the composite providers
 *
 * @class
 * @augments CompositeDataProvider
 * @mixes Observer
 * @param {XDropDataProvider...,Array.<XDropDataProvider>} dataproviders - list of data provider
 */
var GalleryDataProvider = function(dashboardContext, options) {
	XDropDataProvider.apply(this, options);
	Observer.mixin(this);

	this._options = _.extend(
		{
			filter: function(datasource) {
				return true;
			}
		},
		options
	);

	this._context = dashboardContext;
	this._query = "";

	var self = this;

	self._throttledUpdated = _.throttle(this.updateResultSet, 250);
	self._startLoading("data");
	self._context.getBlockDefinitionDataSource().then(function(datasource) {
		self._observe(datasource, "changed.blockDefinitions", function() {
			self._throttledUpdated();
		});
		self._throttledUpdated();
		self._stopLoading("data");
	});
};

GalleryDataProvider.prototype = Object.create(XDropDataProvider.prototype);
GalleryDataProvider.prototype.constructor = GalleryDataProvider;

/**
 * Update the result lists based on the current filter
 */
GalleryDataProvider.prototype.updateResultSet = function() {
	if (this._cancelPendingUpdate) {
		this._cancelPendingUpdate();
	}
	var isCanceled = false;
	var cancelError = new Error("CANCEL ERROR");
	this._cancelPendingUpdate = function() {
		isCanceled = true;
	};

	this._startLoading("updateResultSet");
	this._context
		.getBlockDefinitionDataSource()
		.then(
			function(datasource) {
				return this._options.filter(datasource, this._query);
			}.bind(this)
		)
		.then(function(definitions) {
			if (isCanceled) {
				throw cancelError;
			}

			return _.sortBy(definitions, function(definition) {
				return definition.getDisplayName();
			});
		})
		.then(
			function(items) {
				if (isCanceled) {
					throw cancelError;
				}

				var existingEntries = this._resultSet.getEntriesImmutable();
				var resultEntries = _.map(items, function(blockDef) {
					var existing = _.find(existingEntries, function(entry) {
						return entry.getData() === entry || _.isEqual(entry.getData(), blockDef);
					});

					if (existing) {
						return existing;
					} else {
						return new ResultEntry("galeryItem", blockDef);
					}
				});
				// update the resultset
				this._resultSet.setEntries(resultEntries);
				this._resultSet.commitChanges();
			}.bind(this)
		)
		.catch(function(error) {
			if (error !== cancelError) {
				throw error;
			}
		})
		.then(
			function() {
				this._stopLoading("updateResultSet");
			}.bind(this)
		);
};

GalleryDataProvider.prototype.setQuery = function(queryString) {
	this._query = queryString;
	this._throttledUpdated();
};

export default GalleryDataProvider;
