import Context from "./Context";
import ParameterSet from "./parameters/ParameterSet";
import StringParameter from "./parameters/StringParameter";
import BooleanParameter from "./parameters/BooleanParameter";
import Observable from "@esko/observable";
import Observer from "@esko/observer";
import Promisify from "@esko/promisify";
import _ from "underscore";
import $ from "jquery";
/**
 * a Block is the base component of a Dashboard. Blocks can be nested in other blocks (using the composite pattern)
 *
 * @constructor
 * @class
 * @mixes {Observable}
 * @mixes {Observer}
 * @param {String} id - unique ID for the block
 * @param {Context} [context] - the context in which the block is initialized
 */
var Block = function(id, context) {
	if (!context) {
		context = new Context();
	}

	// Make a Block Observable and an Observer
	Observable.mixin(this, "Block_" + id);
	Observer.mixin(this);

	this._id = id;

	this._inDOM = false;

	this._isCollapsible = false;
	this._isCollapsed = false;

	/**
	 * @member {Context} context
	 * @protected
	 */
	this._context = context;

	this._parameters = new ParameterSet();
	var self = this;
	this._observeUnbound(this._parameters, "changed.*", function(oldValue, newValue) {
		var parameter = _.last(this.event.split("."));
		self._emit(["parameter.changed", parameter].join("."), oldValue, newValue);
	});

	/**
	 * @member {HTMLObject} DOM element
	 * !! WARNING !! NEVER REPLACE THIS PROPERTY, doing so while result in unstable behavior
	 * @protected
	 */
	this._element = $("<div>", {
		"data-block-id": id
	});
	this._element[0].__BLOCK = this;

	this._isInitialized = false;
	this._isModified = false;
	this._isInEditMode = false;
	this._parent = null;

	this._futureChildren = {};

	/**
	 * @member {Array.<Block>} children
	 * @protected
	 */
	this._children = [];

	this._didRequestAnimationFrame = null;
	this._renderPromise = Promisify.once(this, "render");

	this._isAttached = false;
	this._isRoot = false;

	/**
	 * @member {boolean} - indicates if a transparent frame should always be shown
	 * @protected
	 */
	this._showTransparentFrame = false;
};

/**
 * Init the parameters that define the blocktitle, framecolor and showFrame
 * @private
 */
Block.prototype._initTitleAndFrameParameters = function(displayName) {
	return this._context.getLanguageBundle("Block").then(
		function(i18n) {
			var showFrameParameter = new BooleanParameter("blockShowFrame", this._context, i18n.t("SHOW_FRAME"));
			showFrameParameter.setValue(false);
			showFrameParameter.dontPersistIfDefaultValue(false);
			this._addParameter(showFrameParameter);

			// Add the title to the parameters
			var blockTitleParameter = new StringParameter("blockTitle", this._context, i18n.t("BLOCK_TITLE"));
			blockTitleParameter.setValue(displayName || this.getID());
			blockTitleParameter.setVisible(false);
			this._addParameter(blockTitleParameter);

			// color	// color
			var colorParameter = new StringParameter("blockFrameColor", this._context, i18n.t("FRAME_COLOR"));
			colorParameter.setConfigurationGenerator(() =>
				import("../builder/configuration/parameters/ColorPickerConfigurationPanel").then(mod => mod.default)
			);
			colorParameter.setValue("#dedede"); // default color
			colorParameter.dontPersistIfDefaultValue("#dedede");
			colorParameter.setVisible(false);
			this._addParameter(colorParameter);

			// open the block in edit mode
			var showInEditModeParameter = new BooleanParameter("blockShowInEditMode", this._context, "");
			showInEditModeParameter.setValue(false);
			showInEditModeParameter.setVisible(false);
			showInEditModeParameter.setPersistent(false);
			this._addParameter(showInEditModeParameter);

			// observe the block show frame parameter
			this._observeParam("blockShowFrame", this._blockShowFrameChanged);
		}.bind(this)
	);
};

/**
 * Get the type of the block
 * @returns {String} type
 */
Block.prototype.getType = function() {
	throw new Error("This method should be overriden!");
};

/**
 * Get the Unique ID of the block
 * @returns {String}
 */
Block.prototype.getID = function() {
	return this._id;
};

/**
 * Get the title of the block
 * @returns {String}
 */
Block.prototype.getTitle = function() {
	var title = this._parameters.getParameter("blockTitle");
	return title != null ? title.getValue() : this.getID();
};

/**
 * Get the element with the title
 * @returns {HTMLObject}
 */
Block.prototype.getTitleElement = function() {
	var title = this.getTitle();
	return $("<h4>", {
		title: title,
		text: title
	});
};

/**
 * Get buttons to be added to the header of the BlockWrapper
 * @return Array.<JQueryObj>
 */
Block.prototype.getHeaderButtons = function() {
	//IMPLEMENT
};

/**
 * Should the frame be shown for this block
 * @returns {Boolean}
 */
Block.prototype.shouldShowFrame = function() {
	return this._parameters.getParameter("blockShowFrame").getValue();
};

/**
 * Get the desired frame color
 * @returns {String}
 */
Block.prototype.getFrameColor = function() {
	return this._parameters.getParameter("blockFrameColor").getValue();
};

/**
 * Return if a transparant frame should always be shown
 * @returns {boolean}
 */
Block.prototype.showAlwaysTransparentFrame = function() {
	return this._showTransparentFrame;
};

/**
 * Set the title of the block
 * @param {String} title
 */
Block.prototype.setTitle = function(title) {
	var param = this._parameters.getParameter("blockTitle");
	if (param != null) {
		param.setValue(title);
	}
};

/**
 * Set the config of the Block
 * Do NOT override this method, override _doConfigure()
 * @param {Object}          config - JSON configuration
 * @param {IDMap}           idMap  - blocks by ID, always use the ID map
 * @param {String} [displayName] - displayname
 * @return {Promise.<Block>} resolved with reference to itself once init is done
 */
Block.prototype.configure = function(config, idMap, displayName) {
	var self = this;

	return Promise.resolve()
		.then(function() {
			self._context.registerAndMapBlocks(config.blocks, idMap);
		})
		.then(function() {
			if (!self._isInitialized) {
				return self
					._initTitleAndFrameParameters(displayName)
					.then(function() {
						return self._initializeDefaultParameters();
					})
					.then(function() {
						return self._updateParameterOrder();
					});
			}
		})
		.then(function() {
			return self._configureParameters(config, idMap);
		})
		.then(function() {
			return self._doConfigure(config, idMap);
		})
		.then(function() {
			return self._loadPreferences();
		})
		.then(function() {
			// mark as initialized
			if (!self._isInitialized) {
				return Promise.resolve(self._doInit()).then(function() {
					self._isInitialized = true;
					self._emit("initialized");
				});
			}
		})
		.then(function() {
			self._emit("configured");
			return self;
		});
};

/**
 * Configure the parameters of the block
 * @private
 * @param   {object}   config [[Description]]
 * @returns {Promise} resolved when all parameters configured
 */
Block.prototype._configureParameters = function(config, idMap) {
	return Promise.all(
		_.map(
			config.parameters,
			function(parameterConfig) {
				if (this._parameters.getParameter(parameterConfig.name)) {
					return this._parameters.getParameter(parameterConfig.name).configure(parameterConfig.config, idMap);
				} else {
					return this._addParameterWithConfig(parameterConfig, idMap);
				}
			},
			this
		)
	);
};

/**
 * Add a parameter to the block
 * @private
 * @param {Parameter} parameter
 */
Block.prototype._addParameter = function(parameter) {
	this._parameters.setParameter(parameter.getName(), parameter);
};

/**
 * Find the children of a block
 * This method can search recursively if needed
 * @param   {Function}      predicate           - predicate function (return true if the child should be included)
 * @param   {Boolean}       recursive           - true if recursive search is needed
 * @param   {Boolean}       recursiveIfRejected - if true, the children of rejected blocks will also be search, recursiveIfRejected can also be a predicate function (with the rejected block as argument)
 * @returns {Array.<Block>} search result
 */
Block.prototype.findChildren = function(predicate, recursive, recursiveIfRejected) {
	var results = [];

	var visitBlock = function(block) {
		if (predicate(block)) {
			results.push(block);

			if (recursive) {
				_.each(block.getChildren(), visitBlock);
			}
		} else if (recursiveIfRejected && (!_.isFunction(recursiveIfRejected) || recursiveIfRejected(block))) {
			_.each(block.getChildren(), visitBlock);
		}
	};

	_.each(this.getChildren(), visitBlock);

	return results;
};

/**
 * Add a parameter from a config object
 * @private
 * @param {Object}  parameterConfig
 * @returns {Promise} resolved once parameter is added
 */
Block.prototype._addParameterWithConfig = function(parameterConfig) {
	// by default the framework ignores parameters which are not yet defined in the default parameters
	// by implementing this method you could dynamically add parameters
	console.warn("_addParameterWithConfig not implemented", parameterConfig);
};

/**
 * Initialize the default parameters
 * @private
 * @param {Object} config - Block configuration object
 */
Block.prototype._initializeDefaultParameters = function(config) {
	// to be implemented by specific implementation
};

/**
 * Set the order of parameters
 * @private
 */
Block.prototype._updateParameterOrder = function() {
	this._parameters.moveParameterToEndOfList("blockShowFrame");
	this._parameters.moveParameterToEndOfList("blockTitle");
	this._parameters.moveParameterToEndOfList("blockFrameColor");
};

/**
 * The ShowFrame parameter has changed
 * @private
 * @param {Boolean} oldValue
 * @param {Boolean} newValue
 */
Block.prototype._blockShowFrameChanged = function(oldValue, newValue) {
	this._parameters.getParameter("blockFrameColor").setVisible(newValue);
	this._parameters.getParameter("blockTitle").setVisible(newValue);
};

/**
 * Get an ordered list of all parameters
 * @returns {Array.<Parameter>}
 */
Block.prototype.getParameterList = function() {
	return this._parameters.getParameterList();
};

/**
 * Get an ordered list of all parameters of a certain type
 * @param {object} type
 * @returns {Array.<Parameter>}
 */
Block.prototype.getParameterListByType = function(type) {
	return this._parameters.getParameterListByType(type);
};

/**
 * Get a parameter by its name
 * @returns
 */
Block.prototype.getParameterByName = function(name) {
	return this._parameters.getParameter(name);
};

/**
 * Observe changes to a parameter
 * @private
 * @param {String} name    - name of the parameter
 * @param {Function} handler - function to handle the changed parameter update (automatically bound to the block)
 */
Block.prototype._observeParam = function(name, handler) {
	this._observe(this._parameters, ["changed", name].join("."), handler);
};

/**
 * Observe multiple parameters
 * @private
 * @param {Array.<String>} names
 *                                 @param {Function}       handler
 */
Block.prototype._observeParams = function(names, handler) {
	_.each(
		names,
		function(name) {
			this._observeParam(name, handler);
		},
		this
	);
};

/**
 * Configure the block
 * If the configuration is asynchronous, return a promise which resolves once the configuration is done
 * This method might be called multiple times, whenever configuration is changed by the user
 *
 * @param {Object}            config - JSON configuration
 * @param {IDMap}             idMap  - blocks by ID, always use the ID map
 * @returns {Promise,undefined}
 */
Block.prototype._doConfigure = function(config, idMap) {
	// implement this method to configure your block
	// if asynchronous, return a promise
};

/**
 * Called after first configuration
 */
Block.prototype._doInit = function() {
	// implement this method if you want to execute code after the first (initializing) configure() call
	// this could be used to start listening to changes of parameters and do a first render of the widget
};

/**
 * Get the configuration of a Block
 *
 * Do NOT override this method, override _getBlockConfiguration()
 *	@param {boolean} [willPersist] - The configuration will be stored and the new dashboard can be initialized, information should be serializable. (any children returned by getChildrenForPersistence() will be included)
 * @returns {Promise.<Object>} promise with configuration
 */
Block.prototype.getConfiguration = function(willPersist) {
	var self = this;

	return Promise.resolve()
		.then(function() {
			return Promise.all([
				self._getBlockConfiguration(willPersist),
				self._getParameterConfiguration(willPersist),
				self._getChildrenConfiguration(willPersist)
			]);
		})
		.then(function(configs) {
			configs.unshift({});
			if (configs[1].parameters || configs[1].blocks) {
				if (console && console.warn) {
					console.warn("'parameters' and 'blocks' are reserved keywords for the config object and will be ovewritten");
				}
			}

			return {
				type: self.getType(),
				id: self.getID(),
				config: _.extend.apply(_, configs)
			};
		});
};

/**
 * Utility method to generate a configuration objects for a set of blocks
 * @private
 * @param {Array.<Block>} blocks - list of blocks for which configuration should be generated
 * @returns {Promise.<Array.<Object>>} array with all configurations
 */
Block.prototype._getConfigurationForBlocks = function(blocks, willPersist) {
	return Promise.all(
		_.map(blocks, function(block) {
			return block.getConfiguration(willPersist);
		})
	);
};

/**
 * Get the configuration object for all the parameters of the Block
 * @private
 * @param {boolean} [willPersist] - The configuration will be stored and the new dashboard can be initialized, information should be serializable.
 * @returns {Promise.<Object>}
 */
Block.prototype._getParameterConfiguration = function(willPersist) {
	var parameters = this.getParameterList();
	if (parameters.length > 0) {
		return Promise.all(
			_.map(parameters, function(parameter) {
				if (parameter.isPersistent() || !willPersist) {
					return parameter.getConfiguration(willPersist);
				} else {
					return null;
				}
			})
		).then(function(values) {
			return {
				parameters: _.compact(values)
			};
		});
	} else {
		return Promise.resolve({});
	}
};

/**
 * Get the configuration for the children blocks that need to be persisted with this block
 * @private
 * @param   {Boolean} willPersist
 * @returns {Promise.<Object>}
 */
Block.prototype._getChildrenConfiguration = function(willPersist) {
	if (willPersist) {
		var self = this;
		return Promise.resolve(self._findChildrenForPersistence())
			.then(function(childrenForPersistence) {
				return self._getConfigurationForBlocks(childrenForPersistence, willPersist);
			})
			.then(function(configurations) {
				return {
					blocks: configurations
				};
			});
	} else {
		return Promise.resolve({});
	}
};

/**
 * Get all the children that need to be stored in the configuration of this block
 * when persisting
 * @private
 * @returns {Promise.<Array<Block>>}
 */
Block.prototype._findChildrenForPersistence = function() {
	return Promise.resolve([]);
};

/**
 * Get the block specific configuration
 * Override this method to include information that is not part of the default configuration object
 * @param {boolean} [willPersist] - The configuration will be stored and the new dashboard can be initialized, information should be serializable.
 * @returns {Promise.<Object>, Object} a promise with the configuration of the configuration itself
 */
Block.prototype._getBlockConfiguration = function(willPersist) {
	return {};
};

/**
 * The block is being destroyed, it should stop all its functionality
 * Destroy its children and detach itself from its parent
 */
Block.prototype.destroy = function() {
	_.each(this.getChildren(), function(child) {
		child.destroy();
	});

	if (this.getParent()) {
		this.getParent().detachChild(this);
	}

	// just for completeness... we are getting destroyed... we could also just ignore this one
	this.setInDOM(false);

	this._emit("destroyed");

	// remove all previously registered listeners
	this._stopObserveAll();

	// detach ourselves from the DOM
	this._element.detach();

	// override to implement destroy behavior, but make sure this original method is always called
};

/**
 * By default a block is persistent: its configuration/definition is included in the
 * JSON configuration for the context
 * A block can be non-persistent: it is a block termporarily created for displaying reasons,
 * it should not be stored in the configuration, but its children should
 */
Block.prototype.isPersistent = function() {
	return true;
};

/**
 * Is this block configurable?
 * @returns {boolean} true if configurable
 */
Block.prototype.canConfigure = function() {
	return true;
};

/**
 * Get the context of a certain block
 * @returns {Context}
 */
Block.prototype.getContext = function() {
	return this._context;
};

/**
 * Get the DOM Element of the block
 * @returns {JQueryObj}
 */
Block.prototype.getElement = function() {
	return this._element;
};

/**
 * Add a child to the Block
 * @param {Block} block - child block
 */
Block.prototype.addChild = function(block) {
	// OVERRIDE TO REALLY ADD THE CHILD TO THE BLOCK (we now just dump it in the DOM of the current block...)
	// make sure to register the child! and mark it inDOM

	this._registerChild(block);
	this._element.append(block.getElement());
	block.setInDOM(this.isInDOM());
};

/**
 * Use this to register that you are waiting for a child block te be loaded
 * This is used to know when a dashboard is fully loaded
 * @private
 * @param {String}  blockid - id of the block being loaded
 * @returns Function - needs to be executed once the block is no longer a future child, or markFutureChildResolved should be called with the blockid
 */
Block.prototype._registerFutureChild = function(blockid) {
	var promise;

	// check if we are trying to add a parent of the current block to this block...
	var current = this;
	while (current != null) {
		if (current.getID() === blockid) {
			throw new Error("Cannot register a child which is a parent of the current block");
		}
		current = current.getParent();
	}

	if (this._futureChildren[blockid]) {
		promise = this._futureChildren[blockid];
	} else {
		var resolver = function() {};
		promise = new Promise(function(resolve, reject) {
			resolver = resolve;
		});
		this._futureChildren[blockid] = promise;

		var self = this;
		promise.markAsResolved = function() {
			delete self._futureChildren[blockid];
			resolver();
		};
	}

	return function() {
		promise.markAsResolved();
	};
};

/**
 * To be called once a future child block was processed
 * @private
 * @param {String} blockid
 */
Block.prototype._markFutureChildResolved = function(blockid) {
	if (this._futureChildren[blockid]) {
		this._futureChildren[blockid].markAsResolved();
	}
};
/**
 * Register a child block
 * @param {Block} block
 */
Block.prototype._registerChild = function(block) {
	if (!_.contains(this._children, block)) {
		delete this._futureChildren[block.getID()];

		block.setParent(this);
		this._children.push(block);
		this._emit("child.attach", block);
	}
};

/**
 * Notify that the child is detached from the block
 * This will be called whenever a child block is moving to another parent block
 *
 * Note: this should not trigger a message "setParent()" to the block which was detached
 *
 * @author PJPR
 * @param {Block} block - child that is being detached from this block
 */
Block.prototype.detachChild = function(block) {
	if (_.contains(this._children, block)) {
		this._markFutureChildResolved(block.getID());
		this._children = _.without(this._children, block);
		this._emit("child.detach", block);
	}
};

/**
 * Set the parent of a child
 * If this block is already child of another parent, it is automatically detached from the previous parent
 * @param {[[Type]]} block [[Description]]
 */
Block.prototype.setParent = function(block) {
	if (this._parent != block) {
		if (block == null) {
			if (this._isAttached && !this._isRoot) {
				this.onDetached();
			}
		} else {
			if (!this._isAttached) {
				this.onAttached();
			}
		}
		if (this._parent) {
			// we should detach ourselves from our current parent
			this._parent.detachChild(this);
		}
		this._parent = block;
		this.setInDOM(false);
		this._emit("parent", block);
	}
};

/**
 * Only to be used by the context
 */
Block.prototype.__makeRootBlock = function() {
	this._isRoot = true;
	if (!this._isAttached) {
		this.onAttached();
	}
};

/**
 * Get the parent of this block
 * @returns {Block} parent
 */
Block.prototype.getParent = function() {
	return this._parent;
};

/**
 * Get the root node of this block
 * @returns {Block}
 */
Block.prototype.getRoot = function() {
	var root = this;

	while (root.getParent()) {
		root = root.getParent();
	}

	return root;
};

/**
 * Is the block attached (to the context)?
 * @returns {Boolean}
 */
Block.prototype.isAttached = function() {
	return this._isAttached || this._isRoot;
};

/**
 * Is the block in the DOM?
 * @returns {Boolean}
 */
Block.prototype.isInDOM = function() {
	return this._inDOM;
};

/**
 * Get the list of children
 * @returns Array.<Block>
 */
Block.prototype.getChildren = function() {
	return this._children;
};

/**
 * Request the dashboard to display the user configuration panel
 */
Block.prototype._requestUserEdit = function() {
	this._emit("request.useredit");
};

/**
 * Get the configuration panel of the block. This method could be called multiple times, it is possibel to have multiple configuration panels for 1 block
 * @returns {Promise.<BlockConfigurationPanel>}
 */
Block.prototype.getConfigurationPanel = function() {
	return import("../builder/configuration/BlockConfigurationPanel").then(
		({ default: BlockConfigurationPanel }) => new BlockConfigurationPanel(this)
	);
};

/**
 * If you only want to render in animation frames, call this method whenever you want to update the DOM
 * Implement .doRender() to really adapt the DOM
 *
 * @private
 */

Block.prototype._render = function() {
	if (!this._didRequestAnimationFrame) {
		if (!this._isInitialized) {
			this.once("initialized", this._render.bind(this));
			return this._renderPromise;
		}

		var self = this;
		this._renderPromise = Promisify.once(this, "render");
		this._didRequestAnimationFrame = window.requestAnimationFrame(function() {
			self._doRender();
			self._didRequestAnimationFrame = null;
			self._emit("render");
		});
	}

	return this._renderPromise;
};

/**
 * Implement this method to manipulate the DOM in animation frames.
 * Iin this method just adapt the DOM, trigger a render action by calling .doRender()
 * @private
 */
Block.prototype._doRender = function() {
	// implement
};

/**
 * Get a promise that resolves once the state is pushed to the DOM (waits for the render to be executed)
 */
Block.prototype.whenRendered = function() {
	var self = this;
	return this._renderPromise.then(function() {
		return self;
	});
};

/**
 * Wait for a block to be fully loaded
 * @param   {Boolean} [ignoreChildren] - if true, we only wait till all futureChildren are resolved
 * @returns {Promise.<Block>} Promise that resolves in to the block once fully loaded
 */
Block.prototype.whenFullyLoaded = function(ignoreChildren) {
	var waitFor = Promise.all(_.values(this._futureChildren));
	var self = this;

	if (!ignoreChildren) {
		// first wait for future children
		waitFor = waitFor.then(function() {
			return Promise.all(
				_.map(self.getChildren(), function(block) {
					return block.whenFullyLoaded();
				})
			);
		});
	}

	return waitFor
		.then(function() {
			if (!self._isInitialized) {
				return Promisify.once(self, "initialized");
			}
		})
		.then(function() {
			return self;
		});
};

/**
 * Called when the block is attached in the context.
 * Make sure to call this when overriding this method.
 */
Block.prototype.onAttached = function() {
	this._isAttached = true;
	_.each(this._children, function(child) {
		child.onAttached();
	});
};

/**
 * Called when the block is detached from the context.
 * Make sure to call this when overriding this method.
 */
Block.prototype.onDetached = function() {
	this._isAttached = false;
	_.each(this._children, function(child) {
		child.onDetached();
	});
};

/**
 * Notify the block that is added/removed to/from the DOM
 * This is propagated to the children
 *
 * You can either override this method (but be careful: children should also be notified), or listen to the "inDOM" event
 * Best is to avoid relying on the the fact whether a block is in DOM or not (but for legacy reasons it might be needed)
 * @param {Boolean} inDOM
 */
Block.prototype.setInDOM = function(inDOM) {
	if (this._inDOM != inDOM) {
		this._inDOM = inDOM;
		_.each(
			this._children,
			function(child) {
				child.setInDOM(inDOM && this._childInDOM(child));
			},
			this
		);

		this._emit("inDOM", inDOM);
	}
};

/**
 * Is a given child in the DOM? Override to change behavior
 * @param {Block} child
 * @returns {Boolean}
 */
Block.prototype._childInDOM = function(child) {
	return true;
};

/**
 * this element keeper div is shared with all blocks on a page
 */
var elementKeeper = null;

/**
 * Utility to keep the element of a block always in the DOM
 * @param {Boolean} keepInDOM
 * @returns {Object} call .stop() to stop the behavior
 */
Block.prototype._keepElementsInDOM = function(elements) {
	var self = this;
	var savedElements = [];
	var guardedElements = elements;

	var keepInDOM = function() {
		if (self._inDOM) {
			_.each(savedElements, function(pair) {
				pair.el.insertBefore(pair.guard);
				pair.guard.detach();
			});
			savedElements = [];
		} else {
			if (!elementKeeper) {
				elementKeeper = $("<div>", {
					class: "hidden dshElementInDOMKeeper"
				}).appendTo("body");
			}
			_.each(guardedElements, function(el) {
				el = $(el);
				var guard = $("<div>", {
					class: "dshDOMKeeperGuard"
				});
				guard.insertBefore(el);
				elementKeeper.append(el);
				savedElements.push({
					el: el,
					guard: guard
				});
			});
		}
	};

	keepInDOM();
	var subscription = this._observe(this, "inDOM", keepInDOM);
	return {
		stop: function() {
			if (elementKeeper) {
				elementKeeper.find(elements).detach();
				self._stopObserveSubscription(subscription);
			}
		}
	};
};

/**
 * Executed when configuration panel closes
 */
Block.prototype.onConfigurationPanelClosed = function() {
	//IMPLEMENT
};

/**
 * Is the block collapsible
 * @returns {Boolean}
 */
Block.prototype.isCollapsible = function() {
	return this._isCollapsible;
};

/**
 * Is the block collapsed
 * @returns {Boolean}
 */
Block.prototype.isCollapsed = function() {
	return this._isCollapsed;
};

/**
 * Collapse this block
 * @param {Boolean} collapse
 */
Block.prototype.collapse = function(collapse, userTriggered) {
	if (this.getParameterByName("blockShowFrame").getValue()) {
		this._isCollapsed = collapse;
		this._element.toggleClass("hidden", collapse);
		if (userTriggered) {
			this._savePreferences();
		}
	}
};

/**
 * Request a load of the preferences
 * @private
 */
Block.prototype._loadPreferences = function() {
	this._context.loadPreferences(this._id).then(
		function(preferences) {
			this._onLoadPreferences(preferences);
		}.bind(this)
	);
};

/**
 * Handle the loaded preferences
 * @private
 * @param {Object} preferences
 */
Block.prototype._onLoadPreferences = function(preferences) {
	//override
};

/**
 * Save the preferences of the block
 * @private
 */
Block.prototype._savePreferences = function() {
	if (this._isInitialized) {
		this._onSavePreferences().then(
			function(preferences) {
				this._context.savePreferences(this._id, preferences);
			}.bind(this)
		);
	}
};

/**
 * Prepare preferences to be saved.
 * OVERRIDE this method to save the preferences. Call _savePreferences to actually save them.
 * @private
 * @returns {Promise.<Object>} Promise that returns the object that have to be saved
 */
Block.prototype._onSavePreferences = function() {
	return Promise.resolve();
};

export default Block;
