import Observable from "@esko/observable";
import Observer from "@esko/observer";
import FilterUtils from "@esko/filterutils";
import _ from "underscore";
/**
 * Data Source for Block Definitions
 * @constructor
 * @class
 * @param {Array.<BlockDefinition>} [blockDefinitions]
 */
var BlockDefinitionDataSource = function(context, blockDefinitions) {
	Observable.mixin(this, "BlockDefinitionDataSource");
	Observer.mixin(this);
	if (!blockDefinitions) {
		blockDefinitions = [];
	}

	this._observingContext = false;
	this._context = context;
	this._blockDefinitions = blockDefinitions;

	this._observeContextIfNeeded();
};

/**
 * Get all the defined categories
 * @returns {Promise.<Array.<String>>}
 */
BlockDefinitionDataSource.prototype.getCategories = function() {
	var blockDefinitions = this._blockDefinitions;
	return Promise.resolve().then(function() {
		return _.unique(
			_.sortBy(
				_.map(blockDefinitions, function(blockDefinition) {
					return blockDefinition.getCategory();
				}),
				true
			)
		);
	});
};

/**
 * Get all the block definitions
 * @returns {Promise.<Array.<BlockDefinition>>}
 */
BlockDefinitionDataSource.prototype.getBlocks = function() {
	return Promise.resolve(this._blockDefinitions.slice(0));
};

/**
 * Get all blocks matching a certain category
 * @param {String} category
 * @returns {Promise.<Array.<BlockDefinition>>}
 */
BlockDefinitionDataSource.prototype.getBlocksForCategory = function(category) {
	var blockDefinitions = this._blockDefinitions;
	return Promise.resolve().then(
		function() {
			if (category) {
				return _.filter(
					blockDefinitions,
					function(blockDefinition) {
						return blockDefinition.getCategory() === category;
					},
					this
				);
			} else {
				return blockDefinitions;
			}
		}.bind(this)
	);
};

/**
 * Get all blocks matching a certain category and filter
 * @param {String} category
 * @param {String} filter
 * @returns {Promise.<Array.<BlockDefinition>>}
 */
BlockDefinitionDataSource.prototype.filterBlocks = function(category, filter) {
	var self = this;

	return this.getBlocksForCategory(category)
		.then(function(blocks) {
			var regex = FilterUtils.generateRegex(filter, true);

			return _.filter(blocks, function(blockDefinition) {
				return regex.test(blockDefinition.getDisplayName()) || regex.test(blockDefinition.getDescription());
			});
		})
		.then(function(blocks) {
			if (
				_.some(blocks, function(blockDefinition) {
					return !blockDefinition.allowMultipleInContext();
				})
			) {
				return self._context.whenFullyLoaded().then(function() {
					return Promise.all(
						_.map(blocks, function(blockDefinition) {
							if (blockDefinition.allowMultipleInContext()) {
								return blockDefinition;
							} else {
								return self._context.getAttachedBlocksByType(blockDefinition.getType()).then(function(blocksOfType) {
									if (blocksOfType.length === 0) {
										// there are no other blocks of this type in the context
										return blockDefinition;
									} else {
										return Promise.resolve()
											.then(function() {
												if (blockDefinition.config.uniqueParameter) {
													var parameter = _.find(blockDefinition.config.config.parameters, function(parameter) {
														return parameter.name === blockDefinition.config.uniqueParameter;
													});
													// Check if block with value for uniqueParameter is already used
													var isBlockUsed = _.find(blocksOfType, function(block) {
														return block.getParameterByName(blockDefinition.config.uniqueParameter).getValue() === parameter.config.value;
													});
													// return block definition if block is not used yet
													return isBlockUsed ? null : blockDefinition;
												} else {
													//only one per type is allowed
													return null;
												}
											})
											.catch(function(err) {
												console.error(err);
												return null;
											});
									}
								});
							}
						})
					);
				});
			} else {
				return blocks;
			}
		})
		.then(function(blocks) {
			return _.compact(blocks);
		});
};

/**
 * Set the block definitions
 * @param {Array.<BlockDefinition>}
 */
BlockDefinitionDataSource.prototype.setBlockDefinitions = function(blockDefinitions) {
	this._blockDefinitions = blockDefinitions;
	this._observeContextIfNeeded();
	this._emit("changed.blockDefinitions");
};

/**
 * Add a block definition
 * @param {BlockDefinition} blockDefinition
 */
BlockDefinitionDataSource.prototype.addBlockDefinition = function(blockDefinition) {
	this._blockDefinitions.push(blockDefinition);
	this._observeContextIfNeeded();
	this._emit("changed.blockDefinitions");
};

/**
 * Remove a block definition
 * @param {BlockDefinition} blockdefinition to remove
 */
BlockDefinitionDataSource.prototype.removeBlockDefinition = function(blockDefinition) {
	this._blockDefinitions = _.without(this._blockDefinitions, blockDefinition);
	this._emit("changed.blockDefinitions");
};

/**
 * Start listening to context changes if required (because some definitions only allow one instance per context)
 *
 * @private
 */
BlockDefinitionDataSource.prototype._observeContextIfNeeded = function() {
	if (
		!this._observingContext &&
		_.find(this._blockDefinitions, function(blockDef) {
			return !blockDef.allowMultipleInContext();
		})
	) {
		this._observe(this._context, "changed", function() {
			this._emit("changed.blockDefinitions");
		});
		this._observingContext = true;
	}
};
export default BlockDefinitionDataSource;
