import BlockDefinition from "./BlockDefinition";
import BlockDefinitionDataSource from "./BlockDefinitionDataSource";
import Factory from "./Factory";
import ParameterSet from "./parameters/ParameterSet";
import LinkedParameter from "./parameters/LinkedParameter";
import IDMap from "./IDMap";
import SharedState from "./SharedState";
import AsyncUndoRedoController from "@esko/undoredo/AsyncUndoRedoController";
import Command from "@esko/command";
import Observable from "@esko/observable";
import _ from "underscore";
import ErrorBlock from "../blocks/ErrorBlock";
/**
 * Base implementation of Dashboard Context.
 * A context contains mechanisms to load/initialize blocks and share information between the blocks
 * Typically a specialized context is defined for each functionality.
 * @class
 * @constructor
 */

var Context = function(ID, options) {
	if (!ID) {
		ID = "";
	}
	this._id = ID;

	this._options = _.extend(
		{
			calculateDimensions: false
		},
		options
	);

	// make the context observable
	Observable.mixin(this, "Context_" + ID);

	this._blocks = {};
	this._idGenerator = {};
	this._defaultIDMap = new IDMap();

	this._factory = new Factory();

	/**
	 * @member {ParameterSet} parameter functionality for the context
	 */
	this._parameters = new ParameterSet();

	/**
	 * @member {UndoRedoController} undo redo funcitonality
	 */
	this._undoRedoController = new AsyncUndoRedoController(500);

	/**
	 * @member {Promise.<DragDropHandler>}
	 */
	this._dragDropHandler = null;

	this._isEditModeActive = false;

	this._dashboardRootBlock = null;

	var self = this;
	/**
	 * Preference data provider
	 */
	this._preferenceProvider = {
		load: function(blockid) {
			//override
			return Promise.resolve().then(function() {
				if (process.env.NODE_ENV === "development") {
					console.log("No Preference Provider specified");
				}
			});
		},
		save: function(blockid, preferences) {
			//override
			return Promise.resolve().then(function() {
				if (process.env.NODE_ENV === "development") {
					console.log("No Preference Provider specified");
				}
			});
		}
	};

	/**
	 * @member {SharedState} Set of parameters which are publicly available for other blocks
	 */
	this._sharedState = new SharedState();
};

/**
 * Get all the blocks in de context
 * DO NOT MODIFY the result
 * @returns {Object.<id, BlockDefinition>}
 */
Context.prototype.getBlocks = function() {
	return this._blocks;
};

/**
 * Get the factory to construct a block of a certain type
 * @param {String} [blockType] - type of the block
 * @returns Promise.<Factory>
 */
Context.prototype.getFactory = function() {
	return this._factory;
};

Context.prototype.import = function(type) {
	return this._factory.import(type);
};

Context.prototype.addMapping = function(mapping) {
	this._factory.remap(mapping);
};

/**
 * Switch between configuration mode an normal viewing mode
 * @param {Boolean} isEditModeActive
 */
Context.prototype.toggleEditMode = function(isEditModeActive) {
	// FUTURE check permission
	if (this._isEditModeActive != isEditModeActive) {
		this._isEditModeActive = isEditModeActive;
		this._emit("editModeActive", isEditModeActive);
	}
};

/**
 * Is configuration mode active?
 * @returns {Boolean}
 */
Context.prototype.isEditModeActive = function() {
	return this._isEditModeActive;
};

/**
 * Set the factory of the context
 * SHOULD NOT BE CALLED BY BLOCKS!
 * @param {Factory|Promise.<Factory>} [factory] - Factory used for Block creation
 */
Context.prototype.setFactory = function(factory) {
	this._factory = Promise.resolve(factory);
};

/**
 * Get the Language Provider
 * @returns {Promise.<LanguageProvider>}
 */
Context.prototype.getLanguageProvider = function() {
	if (!this._languageProvider) {
		// we only load the LanguageProvider when really needed
		this._languageProvider = import("../i18n/LanguageProvider").then(({ default: LanguageProvider }) => new LanguageProvider());
	}

	return this._languageProvider;
};

/**
 * Get the DragDropHandler
 * @returns {Promise.<DragDropHandler>}
 */
Context.prototype.getDragDropHandler = function() {
	if (!this._dragDropHandler) {
		this._dragDropHandler = import("../builder/dragdrop/DragDropHandler").then(({ default: DragDropHandler }) => new DragDropHandler());
	}
	return this._dragDropHandler;
};

/**
 * Get the Undo Redo Controller for this context
 * @returns {UndoRedoController} the Undo Redo Controller
 */
Context.prototype.getUndoRedoController = function() {
	return this._undoRedoController;
};

/**
 * Execute an undoable action
 * This will first store the configurations of affected blocks and then execute the action
 * @param   {Array.<Block/Parameter>} affectedItems - blocks or parameters that might be affected by the action
 * @param   {Function}      action         - action to execute (this can optionally return a promise to indicate it is finished)
 * @param   {String}        message        - message to be shown in the undo redo controller
 * @returns {Promise}       resolved once the action is executed
 */
Context.prototype.undoableAction = function(affectedItems, action, message) {
	var self = this;
	var oldState = null;

	var command = new Command(
		function() {
			return Promise.resolve()
				.then(function() {
					if (oldState === null) {
						return Promise.all(
							_.map(affectedItems, function(item) {
								return Promise.resolve(item.getConfiguration()).then(function(config) {
									return {
										item: item,
										config: JSON.stringify(config)
									};
								});
							})
						).then(function(oldConfiguration) {
							oldState = oldConfiguration;
						});
					}
				})
				.then(function() {
					return action();
				})
				.then(function() {
					self._emit("changed");
				});
		},
		function() {
			return Promise.all(
				_.map(oldState, function(itemState) {
					var configuration = JSON.parse(itemState.config);
					return itemState.item.configure(configuration.config, self._defaultIDMap.createInstance());
				})
			).then(function() {
				self._emit("changed");
			});
		},
		message
	);

	return self.getUndoRedoController().execute(command);
};

/**
 * Was the dashboard context modified by the user?
 * (were there any undoable actions executed)
 * @returns {Boolean}
 */
Context.prototype.isModified = function() {
	return this.getUndoRedoController().undoCount() > 0;
};
/**
 * Change the configuration of a block (this action is undoable)
 * @param   {Block}   block         - block which should be configured
 * @param   {Object}  configuration - configuration object
 * @param   {String}  [message]     - Message to be shown in the undo redo controller\
 * @returns {Promise} resolved once the action is executed
 */
Context.prototype.configureBlock = function(block, configuration, message) {
	var self = this;
	return this.getLanguageBundle("Dashboard").then(
		function(i18n) {
			return this.undoableAction(
				[block],
				function() {
					return block.configure(configuration, self._defaultIDMap.createInstance());
				},
				message || i18n.t("CONFIGURE_BLOCK_UNDOREDOMESSAGE", block.getID())
			);
		}.bind(this)
	);
};

/**
 * Get the context ParameterSet to easily register observers for parameter changes
 * @returns {ParameterSet}
 */
Context.prototype.getParameterSet = function() {
	return this._parameters;
};

/**
 * Get all the parameters of a certain type
 * @param   {object}            type
 * @returns {Array.<Parameter>}
 */
Context.prototype.getParametersByType = function(type) {
	return this._parameters.getParameterListByType(type);
};

/**
 * Set a parameter
 * @param {String}    name      name of the parameter
 * @param {Parameter} parameter
 */
Context.prototype.setParameter = function(name, parameter) {
	this._parameters.setParameter(name, parameter);
};

/**
 * Get a parameter
 * @param   {String}    name
 * @returns {Parameter}
 */
Context.prototype.getParameter = function(name) {
	return this._parameters.getParameter(name);
};

/**
 * Set the LanguageProvider of the context
 * SHOULD NOT BE CALLED BY BLOCKS
 * @param {LanguageProvider|Promise.<LanguageProvider>} [languageProvider] - LanguageProvider which can be used by the blocks
 */
Context.prototype.setLanguageProvider = function(languageProvider) {
	this._languageProvider = Promise.resolve(languageProvider);
};
/**
 * @param {String} bundleName - name of the bundle which should be loaded
 * @returns {Promise.<LanguageBundle>
 */
Context.prototype.getLanguageBundle = function(bundleName) {
	return this.getLanguageProvider().then(function(provider) {
		return provider.getLanguageBundle(bundleName);
	});
};

/**
 * Load a block (and initialize if needed)
 * @returns Promise.<Block>
 */
Context.prototype.loadBlock = function(id) {
	var context = this;
	return Promise.resolve()
		.then(
			function() {
				if (!this._blocks[id]) {
					throw new Error(new Error("No definition for Block with id " + id));
				}

				if (!this._blocks[id].promise) {
					this._blocks[id].promise = this._initBlock(id);
				}

				return this._blocks[id].promise;
			}.bind(this)
		)
		.catch(function(error) {
			//make an error block

			if (!context._blocks[id]) {
				context._blocks[id] = new BlockDefinition(id, "dashboard/blocks/ErrorBlock", {});
			}

			var blockDefinition = context._blocks[id];

			var errorBlock = Promise.resolve(new ErrorBlock(id, context, error, blockDefinition.clone(id))).then(function(block) {
				return block.configure({}, context._defaultIDMap.createInstance());
			});
			blockDefinition.promise = errorBlock;

			return errorBlock;
		});
};

/**
 * Initialize a block
 * @private use loadBlock or addBlock
 * @param {String} id
 * @returns {Promise.<Block>}
 */
Context.prototype._initBlock = function(id) {
	var context = this;

	var blockDefinition = this._blocks[id];

	return this._factory
		.initBlock(id, context, blockDefinition.type, blockDefinition.config)
		.then(function(block) {
			var IDMap = context._defaultIDMap;

			if (blockDefinition.IDMap) {
				// we should use the ID map only once
				IDMap = blockDefinition.IDMap;
				blockDefinition.IDMap = null;
			}

			return block.configure(blockDefinition.config.config, IDMap.createInstance(), blockDefinition.getDisplayName());
		})
		.then(function(block) {
			blockDefinition.block = block;
			context._emit("initBlock", block);

			return block;
		});
};

/**
 * Get the block definitions of all blocks of a certain type
 * @param {String} type
 * @returns {Array.<BlockDefinition>}
 */
Context.prototype.getBlocksByType = function(type) {
	return _.where(this._blocks, {
		type: type
	});
};

/**
 * Returns all blocks of a certain type that are attached to the board
 * @param   {object}          type
 * @returns {Promise.<Block>}
 */
Context.prototype.getAttachedBlocksByType = function(type) {
	return this._getAttachedBlocks(this.getBlocksByType(type));
};

/**
 * Returns all blocks that are attached to the board
 * @param   {object}          type
 * @returns {Promise.<Block>}
 */
Context.prototype.getAttachedBlocks = function() {
	return this._getAttachedBlocks(this.getBlocks());
};

/**
 * Get the blocks that are attached from a list of blockdefinitions
 * @param   {Array.<BlockDefinition>} blockdefs
 * @returns {Promise.<Block>}
 */
Context.prototype._getAttachedBlocks = function(blockdefs) {
	var dashboardRoot = this._dashboardRootBlock;
	return Promise.all(
		_.map(blockdefs, function(blockDefinition) {
			if (blockDefinition.promise) {
				return blockDefinition.promise;
			} else {
				return null;
			}
		})
	).then(function(blocks) {
		return _.filter(blocks, function(block) {
			return block !== null && block.isAttached() && block.getRoot() === dashboardRoot;
		});
	});
};

/**
 * Small utility method to transform an type string to a valid ID string
 * @private
 * @param   {String} type - type string
 * @returns {String} ID string
 */
Context.prototype._makeTypeIDFriendly = function(type) {
	return type.split("/").join("_");
};

/**
 * Generate a unique ID for a certain type
 * @param {String} type
 * @returns {String} unique id for type
 */
Context.prototype._generateUniqueID = function(type, attempt) {
	var IDBase = this._makeTypeIDFriendly(type);

	if (attempt > 10000) {
		throw new Error("Could not generate unique ID after 10000 attempts");
	}

	if (!this._idGenerator[IDBase]) {
		var nextID = 0;
		_.each(this.getBlocksByType(type), function(blockDef) {
			var id = blockDef.getID();
			var blockDefBase = id.substr(0, id.lastIndexOf("_"));
			if (blockDefBase === IDBase) {
				var index = parseInt(id.substr(id.lastIndexOf("_") + 1), 10);
				if (!isNaN(index)) {
					nextID = Math.max(index + 1, nextID);
				}
			}
		});

		this._idGenerator[IDBase] = nextID;
	}
	var newID = IDBase + "_" + this._idGenerator[IDBase]++;

	if (!this._blocks[newID]) {
		return newID;
	} else {
		if (!attempt) {
			attempt = 0;
		}
		// ID already exists
		return this._generateUniqueID(type, attempt++);
	}
};

/**
 * Update the cache of unique IDs when adding a new block
 * @private
 * @param {string} id
 */
Context.prototype._updateUniqueIDs = function(id) {
	var IDBase = id.substr(0, id.lastIndexOf("_"));
	var index = parseInt(id.substr(id.lastIndexOf("_") + 1), 10);

	if (IDBase.length > 0 && !isNaN(index)) {
		if (!this._idGenerator[IDBase]) {
			this._idGenerator[IDBase] = 0;
		}
		this._idGenerator[IDBase] = Math.max(index + 1, this._idGenerator[IDBase]);
	}
};

Context.prototype.registerAndMapBlocks = function(blocks, idMap) {
	if (!idMap) {
		idMap = this._defaultIDMap;
	}

	// make sure we dont generate any collisions while generating new IDs for this set of blocks
	_.each(_.pluck(blocks, "id"), this._updateUniqueIDs, this);

	var mappedBlocks = _.map(
		blocks,
		function(blockConfig) {
			var id = blockConfig.id;
			var newId = id;

			if (this._blocks[id]) {
				// there is a collision, generate new id
				newId = this._generateUniqueID(blockConfig.type);
				idMap.addMapping(id, newId);
			}

			// we want a copy of the data
			var mappedConfig = JSON.parse(JSON.stringify(blockConfig));
			mappedConfig.id = newId;
			return mappedConfig;
		},
		this
	);

	_.each(
		mappedBlocks,
		function(blockConfig) {
			this._registerBlockWithID(blockConfig, idMap, blockConfig.id);
		},
		this
	);
};

/**
 * Register a Block definition without initializing.
 *
 * To add a block, first register it with this method and then load it
 *
 * @param {Object} config - configuration for the block
 * @param {IDMap] [IDMap]
 * @returns {String} id
 */
Context.prototype.registerBlock = function(config, idMap) {
	var id = this._generateUniqueID(config.type);
	this._registerBlockWithID(config, idMap, id);

	return id;
};

/**
 * Internal method for registering a block with a certain id
 * @private
 * @param {String} id
 * @param {object} config - config object
 * @param {IDMap}  idMap  - idmapping
 */
Context.prototype._registerBlockWithID = function(config, idMap, id) {
	if (!config.type) {
		throw new Error("Cannot register block without a type!", id, config);
	}

	if (this._blocks[id]) {
		throw new Error("Trying to add BlockDefinition with already existing ID");
	}

	if (!idMap) {
		idMap = this._defaultIDMap;
	}

	var definition = new BlockDefinition(id, config.type, config);
	definition.idMap = idMap;

	// FUTURE start loading the block immediatily if it requires it
	this._blocks[id] = definition;
	this._emit("changed");
};

/**
 * This will clone the block definition and register it in the context
 * This internally uses the BlockDefinition.clone method, so it only works for "plain configuration"
 * @param {BlockDefinition} blockDefinition
 * @returns {String} id
 */
Context.prototype.registerBlockFromDefinition = function(blockDefinition) {
	var newID = this._generateUniqueID(blockDefinition.type);
	var newDefinition = blockDefinition.clone(newID);

	this._blocks[newID] = newDefinition;
	this._emit("changed");
	return newID;
};

/**
 * Shorthand method to quickly start loading a block
 * @param {Object}          config - config object
 * @returns {Promise.<Block>} resolved once block is initialized
 */
Context.prototype.registerAndLoadBlock = function(config, IDMap) {
	var self = this;
	var setRoot = false;
	if (!self._dashboardRootBlock && !self._rootBlockIsLoading) {
		setRoot = true;
		this._rootBlockIsLoading = true;
	}

	return Promise.resolve()
		.then(function() {
			return self.loadBlock(self.registerBlock(config, IDMap));
		})
		.then(function(block) {
			if (setRoot) {
				self._dashboardRootBlock = block;
				block.__makeRootBlock();

				delete self._rootBlockIsLoading;
			}

			return block;
		});
};

/**
 * Destroy a block
 * @param {Block} block - Block that hsould be destroyed
 */
Context.prototype.destroyBlock = function(block) {
	// FUTURE if we listen to block events, make sure we no longer listen to the block

	// the block will remove itself from the DOM and its parents
	block.destroy();

	delete this._blocks[block.getID()];
	this._emit("changed");
};

/**
 * Undoable removal of a block
 * @param {Array.<Block>} affectedBlocks
 * @param {Block} block to detach
 */
Context.prototype.detachBlock = function(affectedBlocks, block) {
	return this.getLanguageBundle("Dashboard").then(
		function(i18n) {
			return this.undoableAction(
				affectedBlocks,
				function() {
					block.setParent(null);
				},
				i18n.t("DELETE_BLOCK_UNDOREDOMESSAGE", block.getID())
			);
		}.bind(this)
	);
};
/**
 * Remove a block definition from the context
 * If a Block is (being) initialized, the block will be destroyed
 * @param {BlockDefinition} blockDefinition
 */
Context.prototype.removeBlockDefinition = function(blockDefinition) {
	var definition = this._blocks[blockDefinition.getID()];
	if (definition) {
		if (definition.promise) {
			definition.promise.then(
				function(block) {
					this.destroyBlock(block);
				}.bind(this)
			);
		} else {
			// not being loaded yet
			delete this._blocks[definition.getID()];
		}
	}
	this._emit("changed");
};

/**
 * @returns {Promise.<BlockDefinitionDataSource}
 */
Context.prototype.getBlockDefinitionDataSource = function() {
	if (!this._blockDefinitionDataSource) {
		this._blockDefinitionDataSource = Promise.resolve(new BlockDefinitionDataSource(this));
	}
	return this._blockDefinitionDataSource;
};

/**
 * Set the BlockDefinitionDataSource of the context
 * SHOULD NOT BE CALLED BY BLOCKS
 * @param {BlockDefinitionDataSource|Promise.<BlockDefinitionDataSource>} [blockDefinitionDataSource] - BlockDefinitionDataSource
 */
Context.prototype.setBlockDefinitionDataSource = function(blockDefinitionDataSource) {
	this._blockDefinitionDataSource = Promise.resolve(blockDefinitionDataSource);
};

/**
 * Get the BlockDefinition for a certain ID
 * @param   {String}          id - ID of the BlockDefinition
 * @returns {BlockDefinition} definition object for the requested id
 */
Context.prototype.getBlockDefinition = function(id) {
	return this._blocks[id];
};

/**
 * Wait until all currently loading blocks have been loaded completely
 * @returns {Promise.<Context>}
 */
Context.prototype.whenFullyLoaded = function() {
	var context = this;
	return Promise.all(
		_.map(this._blocks, function(blockdef) {
			if (!blockdef.promise) {
				return true;
			} else {
				return blockdef.promise
					.then(function(block) {
						return block.whenFullyLoaded(true);
					})
					.catch(function() {
						// if a block fails to load, the ongoing loadf for that block is finished!
						return true;
					});
			}
		})
	).then(function(block) {
		return context;
	});
};

/**
 * Store custom Legacy My Work block User display preferences
 * @param {Array.<Object>} legacyPrefsList - array of JSON-based representation of HomePageUserPrefs.java objects
 */
Context.prototype.setLegacyMyWorkUserPreferences = function(legacyPrefsList) {
	// TODO REMOVE FROM DASHBOARD FRAMEWORK
	this._legacyMyWorkUserPrefs = legacyPrefsList;
};

/**
 * Retrieve Legacy My Work block User display preferences for a specific My Work Section (by its ID)
 * @returns {Object} - null if not found
 */
Context.prototype.getLegacyMyWorkSectionUserPreferences = function(legacyMyWorkSectionID) {
	// TODO REMOVE FROM DASHBOARD FRAMEWORK
	if (this._legacyMyWorkUserPrefs) {
		var targetMyWorkSectionPrefs = _.find(this._legacyMyWorkUserPrefs, function(sectionPrefs) {
			return sectionPrefs._HomePageID === legacyMyWorkSectionID;
		});

		if (targetMyWorkSectionPrefs) {
			return targetMyWorkSectionPrefs;
		}
	}

	return null;
};

/**
 * Get the id
 * @returns {String}
 */
Context.prototype.getID = function() {
	return this._id;
};

/**
 * Set a preference provider
 * @param {Object} preferenceProvider - with save and load method
 */
Context.prototype.setPreferenceProvider = function(preferenceProvider) {
	this._preferenceProvider = preferenceProvider;
};

/**
 * Save the preferences
 * @param {String} blockid
 * @param {Object} preferences
 */
Context.prototype.savePreferences = function(blockid, preferences) {
	this._preferenceProvider.save(blockid, preferences);
};

/**
 * Load the preferences
 * @param   {String} blockid
 * @returns {Object}
 */
Context.prototype.loadPreferences = function(blockid) {
	return this._preferenceProvider.load(blockid);
};

/**
 * Get an option from the context
 */
Context.prototype.getOption = function(key) {
	return this._options[key];
};

/**
 * Get the Shared State of the context
 * @param {String} name - name of the public parameter
 * @param {Function} onChange - called when parameter has changed
 * @param {Boolean} allowPushValue - set to TRUE if the block can make changes to the public parameter
 * @returns {LinkedParameter}
 */
Context.prototype.subscribeToPublicParameter = function(name, onChange, allowPushValue) {
	var linkedParameter = new LinkedParameter(name, this, null, this._sharedState, null, name, allowPushValue);
	linkedParameter.on("changed", function() {
		onChange(linkedParameter.getValue());
	});

	return linkedParameter;
};

/**
 * Add a public parameter to the Shared State of the Context
 * @param {String} name
 * @param {Parameter} parameter
 */
Context.prototype.setPublicParameter = function(name, parameter) {
	this._sharedState.setParameter(name, parameter);
};

/**
 * Get a public parameter from the Shared State of the Context
 * @param {String} name
 * @returns {Parameter}
 */
Context.prototype.getPublicParameter = function(name) {
	return this._sharedState.getParameter(name);
};

export default Context;
