import Block from "./Block";
import BlockWrapper from "../layout/BlockWrapper";
import _ from "underscore";
/**
 * Base implementation for a Layout
 * A layout is used to display other blocks in a predefined way
 * @constructor
 * @class
 * @augments Block
 * @param {String} id
 * @param {Context} context
 */
var Layout = function(id, context) {
	Block.call(this, id, context);

	this._element.addClass("dshLayout");
};

Layout.prototype = Object.create(Block.prototype);
Layout.prototype.constructor = Layout;

Layout.prototype.getType = function() {
	return "dashboard/core/Layout";
};

/**
 * @override
 * layouts have a different default frame color
 * @private
 */
Layout.prototype._initTitleAndFrameParameters = function() {
	return Block.prototype._initTitleAndFrameParameters.apply(this, arguments).then(
		function() {
			var frameColor = this._parameters.getParameter("blockFrameColor");
			frameColor.setValue("#ececec");
			frameColor.dontPersistIfDefaultValue("#ececec");
		}.bind(this)
	);
};

/**
 * Do not override unless you know what you are doing
 * Implement _getLayoutSpecificConfiguration to return the configuration of your layout
 *
 * @Override a Layout also includes all the configuration of first level child layout blocks
 * @private
 * @returns {object} [[Description]]
 */
Layout.prototype._getBlockConfiguration = function() {
	return this._getLayoutSpecificConfiguration();
};
Layout.prototype._findChildrenForPersistence = function() {
	return this.findChildren(
		function(child) {
			return child instanceof Layout;
		},
		false,
		function(block) {
			return block instanceof BlockWrapper;
		}
	);
};

/**
 * In builder mode layout always want to be framed
 * @override
 * @returns {Boolean} [[Description]]
 */
Layout.prototype.shouldShowFrame = function() {
	return Block.prototype.shouldShowFrame.call(this) || this._context.isEditModeActive();
};

/**
 * Implement this method to return the configuration for the Layout
 * @private
 * @returns {object} configuration of the Layout
 */
Layout.prototype._getLayoutSpecificConfiguration = function() {
	return {};
};

/**
 * Initialize a BlockWrapper for an item
 * this will cause the block to be loaded from the context
 * @protected
 * @param {String} itemID
 * @returns {Promise.<BlockWrapper>}
 */
Layout.prototype._loadBlockWrapper = function(itemID) {
	var promise = this._context.registerAndLoadBlock({
		type: "dashboard/layout/BlockWrapper",
		config: {
			blockID: itemID
		}
	});
	return promise;
};

/**
 * Prepend a block to the layout
 * @param   {String}  blockid
 * @returns {Promise} Resolved once the block is prepended
 */
Layout.prototype.prependBlock = function(blockid) {
	return this.prependBlocks([blockid]);
};

/**
 * Prepend a list of block to the layout
 * @param   {Array.<String>}  blockid
 * @returns {Promise} Resolved once the block is prepended
 */
Layout.prototype.prependBlocks = function(blocks) {
	var newList = blocks.concat(this.getItemList());

	return this.setItemList(newList);
};

/**
 * Append a block to the layout
 * @param   {String}  blockid
 * @returns {Promise} Resolved once the block is appended
 */
Layout.prototype.appendBlock = function(blockid) {
	return this.appendBlocks([blockid]);
};

/**
 * Append a list of block to the layout
 * @param   {Array.<String>}  blockid
 * @returns {Promise} Resolved once the block is prepended
 */
Layout.prototype.appendBlocks = function(blocks) {
	var newList = this.getItemList().concat(blocks);

	return this.setItemList(newList);
};

/**
 * Get a list of items in the layout
 * @returns {Array.<String>}  (ordered) list of IDs of items in layout
 */
Layout.prototype.getItemList = function() {
	// needs to be implemented by the layout
	return Promise.reject("Not implemented");
};

/**
 * Set the list of items in the layout
 * @param {Array.<String>} list - (ordered) list of IDs in layout
 */
Layout.prototype.setItemList = function(list) {
	// needs to be implemented by the layout
	return Promise.reject("Not implemented");
};

export default Layout;
