import Parameter from "./Parameter";
import ParameterSet from "./ParameterSet";
import _ from "underscore";
/**
 * A parameter that consists out of other parameters
 * @param {String}  name             - name of the parameter
 *                                   @param {Context} context
 * @param {String}  [type]           - type of the parameter
 * @param {String}  userFriendlyName - user friendly name of the parameter
 */
var CompositeParameter = function(name, context, type, userFriendlyName) {
	// fall back to composite parameter type
	if (!type) {
		type = "dashboard/core/parameters/CompositeParameter";
	}
	Parameter.call(this, name, context, type, userFriendlyName);

	this._parameters = new ParameterSet();
	this._parameters.on("changed.*", this._emitChanged.bind(this));
};

CompositeParameter.prototype = Object.create(Parameter.prototype);
CompositeParameter.prototype.constructor = CompositeParameter;

/**
 * Get the ordered list of parameters
 * @returns {Array.<Parameter>}
 */
CompositeParameter.prototype.getParameterList = function() {
	return this._parameters.getParameterList();
};

/**
 * Add a sub parameter
 * @param {Parameter} parameter
 */
CompositeParameter.prototype.addParameter = function(parameter) {
	this._parameters.setParameter(parameter.getName(), parameter);
};
/**
 * Get a sub parameter
 * @param   {String}    name - name of the sub parameter
 * @returns {Parameter}
 */
CompositeParameter.prototype.getParameter = function(name) {
	return this._parameters.getParameter(name);
};

/**
 * Get the value of sub parameter
 * @param   {String} name - name of the sub paramter
 * @returns {Any}
 */
CompositeParameter.prototype.getParameterValue = function(name) {
	return this._parameters.getValue(name);
};

/**
 * @Override
 */
CompositeParameter.prototype.configure = function(config) {
	_.each(
		config.parameters,
		function(parameterConfig) {
			var parameter = this._parameters[parameterConfig.name];
			if (parameter) {
				parameter.configure(parameterConfig.config);
			} else {
				throw new Error("Trying to configure non-existing subparameter");
			}
		},
		this
	);
};

/**
 * @Override
 */
CompositeParameter.prototype._getParameterSpecificConfig = function() {
	return Promise.all(
		_.map(this.getParameterList(), function(parameter) {
			return parameter.getConfiguration();
		})
	).then(function(parameterConfig) {
		return {
			parameters: parameterConfig
		};
	});
};

/**
 * @override
 */
CompositeParameter.prototype._getConfigurationPanelComponent = function() {
	return import("../../builder/configuration/parameters/CompositeParameterConfigurationPanel").then(mod => mod.default);
};

export default CompositeParameter;
