import Parameter from "./Parameter";
import Observer from "@esko/observer";
import Promisify from "@esko/promisify";
/**
 * Parameter which is a 1:1 link to a parameter	from (another) parameterSet
 * @augments Parameter
 * @mixes {Observer}
 * @param {String}       name             - name of the parameter
 * @param {String}       userFriendlyName - user friendly representation of the parameter
 * @param {ParameterSet} parameterSet     - the parameterSet in which the linked parameter resides
 * @param {String} linkedType - the type of the linked parameter
 * @param {String} linkedName - the name of the linked parameter in the linked parameterset
 * @param {Boolean} allowPushValue - if true, this parameter can change the value of the linked parameter
 */
var LinkedParameter = function(name, context, userFriendlyName, parameterSet, linkedType, linkedName, allowPushValue) {
	Parameter.call(this, name, context, "dashboard/core/parameters/LinkedParameter", userFriendlyName);

	Observer.mixin(this);

	this._linkedSet = parameterSet;
	this._allowPushValue = allowPushValue;
	// FUTURE use this linkedType to check if we are making valid connections or not
	this._linkedType = linkedType;
	this._linkedName = null;

	this.setLinkedName(linkedName);
};

LinkedParameter.prototype = Object.create(Parameter.prototype);
LinkedParameter.prototype.constructor = LinkedParameter;

/**
 * Change the link
 * @param {String} name - name of the linked parameter
 */
LinkedParameter.prototype.setLinkedName = function(name) {
	if (this._linkedName !== name) {
		var oldValue = this.getValue();

		// remove old listeners
		this._stopObserveAll();
		this._linkedName = name;

		this._observe(this._linkedSet, ["changed", name].join("."), this._emitChanged);
		this._emitChanged(oldValue, this.getValue());
	}
};

/**
 * Get the name of the linked parameter
 * @returns {String}
 */
LinkedParameter.prototype.getLinkedName = function() {
	return this._linkedName;
};

/**
 * Get the type of the linked parameter
 * @returns {String}
 */
LinkedParameter.prototype.getLinkedType = function() {
	return this._linkedType;
};

/**
 * Set the linked Type
 * @param {String} type
 */
LinkedParameter.prototype.setLinkedType = function(type) {
	this._linkedType = type;
};

/**
 * (re)configure a parameter
 * @param {object} config - configuration object of the parameter
 */
LinkedParameter.prototype.configure = function(config) {
	this.setLinkedName(config.linkedName);
	this.setLinkedType(config.linkedType);
};

/**
 * Internal method for getting to the linked parameter
 * @private
 * @returns {Parameter} might return undefined!
 */
LinkedParameter.prototype._getLinkedParameter = function() {
	return this._linkedSet.getParameter(this._linkedName);
};

/**
 * Get the value of the linked Parameter
 * @returns {Any}
 */
LinkedParameter.prototype.getValue = function() {
	var parameter = this._getLinkedParameter();
	if (parameter) {
		// FUTURE extra check to check if the parameter is of the correct type
		return parameter.getValue();
	} else {
		return undefined;
	}
};

/**
 * Change the value of the linked parameter
 * @param {Any} newValue
 */
LinkedParameter.prototype.setValue = function(newValue) {
	if (this._allowPushValue) {
		if (this._getLinkedParameter()) {
			this._getLinkedParameter().setValue(newValue);
		}
	}
};

/**
 * @Override
 */
LinkedParameter.prototype.isConfigurableByUser = function() {
	return this._configurableByUser && this._allowPushValue;
};

/**
 * Get the configuration object for the linkedparameter
 * @private
 * @returns {object}
 */
LinkedParameter.prototype._getParameterSpecificConfig = function() {
	return {
		linkedName: this.getLinkedName(),
		linkedType: this.getLinkedType()
	};
};

/**
 * Get the configuration element for the parameter
 * @returns {Promise.<ConfigurationPanel>}
 */
LinkedParameter.prototype.getConfigurationPanel = function() {
	return import("../../builder/configuration/parameters/LinkedParameterConfigurationPanel").then(
		({ default: ParameterConfigurationPanel }) => new ParameterConfigurationPanel(this)
	);
};

export default LinkedParameter;
