import Observable from "@esko/observable";
import Promisify from "@esko/promisify";
/**
 * a Parameter is some configurable/changing information of a Block
 * A parameter could be linked to the context or have a static value, or use a service to update information
 * @param {String} name             - name of the parameter
 * @param {String} type             - type of the parameter
 * @param {String} userFriendlyName - user friendly name of the parameter
 * @param {String} [helpText]       - help text to be displayed right under the input field
 * @param {String} [toolTipText]	- extra info in tooltip
 */
var Parameter = function(name, context, type, userFriendlyName, helpText, toolTipText) {
	Observable.mixin(this, "Parameter_" + name);

	this._name = name;
	this._context = context;

	// we want to fall back to the default parameter if no type is set
	if (!type) {
		type = "dashboard/core/parameters/Parameter";
	}

	this._configurationGenerator = () => import("../../builder/configuration/ParameterConfigurationPanel").then(mod => mod.default);

	this._type = type;
	this._userFriendlyName = userFriendlyName;
	this._helpText = helpText;
	this._toolTipText = toolTipText;
	this._restrictedSet = null;
	this._configurableByUser = true;
	this._isVisible = true;
	this._isPeristent = true;

	this._isPersistentWhenDefaultValue = true;
	this._defaultValue = null;
};

/**
 * Get the name of the parameter
 * @returns {String}
 */
Parameter.prototype.getName = function() {
	return this._name;
};

/**
 * Get the Type of the parameter
 * @returns {String}
 */
Parameter.prototype.getType = function() {
	return this._type;
};

/**
 * Get the user friendly name of the parameter
 * @returns {String}
 */
Parameter.prototype.getUserFriendlyName = function() {
	return this._userFriendlyName;
};

/**
 * Get the help text for the parameter
 * @returns {String}
 */
Parameter.prototype.getHelpText = function() {
	return this._helpText;
};

/**
 * Set the help text for the parameter
 * @returns {String}
 */
Parameter.prototype.setHelpText = function(helpText) {
	this._helpText = helpText;
};

/**
 * Get the tooltip text for the parameter
 * @returns {String}
 */
Parameter.prototype.getToolTipText = function() {
	return this._toolTipText;
};

/**
 * Set the tooltip text for the parameter
 * @returns {String}
 */
Parameter.prototype.setTooltipText = function(toolTipText) {
	this._toolTipText = toolTipText;
};

/**
 * Set the value of the parameter
 * @param {Any} value - new value for the parameter
 */
Parameter.prototype.setValue = function(value) {
	var oldValue = this._value;
	this._value = value;
	if (!this._valueEquals(oldValue, value)) {
		this._emitChanged(oldValue, value);
	}
};

/**
 * Check if two values are equal
 * @private
 * @param   {Any}     oldValue
 * @param   {Any}     newValue
 * @returns {Boolean}
 */
Parameter.prototype._valueEquals = function(oldValue, newValue) {
	return oldValue === newValue;
};

/**
 * Sets the value of the parameter in a undoable way.
 * @returns {Promise} resolved once the value was changed
 */
Parameter.prototype.undoableSetValue = function(value) {
	// FUTURE also incorporate parameters that are directly linked to this parameter...

	return this._context.getLanguageBundle("Dashboard").then(
		function(i18n) {
			return this._context.undoableAction(
				[this],
				function() {
					this.setValue(value);
				}.bind(this),
				i18n.t("CONFIGURE_PARAMETER_UNDOREDOMESSAGE", this.getUserFriendlyName())
			);
		}.bind(this)
	);
};

/**
 * A restricted set restricts the possible values of the parameter (the configuration panel will show a dropdown)
 *
 * @param {Object} restrictedSet - {value: "label", value2: "label 2"}
 */
Parameter.prototype.setRestrictedSet = function(restrictedSet) {
	this._restrictedSet = restrictedSet;
};

/**
 * Get the restricted set of the parameter
 * @returns {Object} restrictedSet - {value: "label", value2: "label 2"}
 */
Parameter.prototype.getRestrictedSet = function() {
	return this._restrictedSet;
};

/**
 * Emit a change event for the parameters value
 * @private
 * @param {Any} oldValue - previous value of the parameter
 * @param {Any} newValue - new value of the parameter
 */
Parameter.prototype._emitChanged = function(oldValue, newValue) {
	this._emit("changed", oldValue, newValue);
};
/**
 * Get the value of the parameter
 * @returns {Any} value of the parameter
 */
Parameter.prototype.getValue = function() {
	return this._value;
};

/**
 * Configure the parameter from a configuration object
 * @param {object} config
 */
Parameter.prototype.configure = function(config, idMap) {
	this.setValue(config.value);
};

/**
 * Get the configuration of the parameter
 * @param {boolean} [willPersist] - The configuration will be stored and the new dashboard can be initialized, information should be serializable.
 * @returns {Promise.<Object>}
 */
Parameter.prototype.getConfiguration = function(willPersist) {
	return Promise.resolve(this._getParameterSpecificConfig(willPersist)).then(
		function(config) {
			return {
				name: this.getName(),
				type: this.getType(),
				config: config
			};
		}.bind(this)
	);
};

/**
 * Get the parameter specific configuration
 * @private
 * @param {boolean} [willPersist] - The configuration will be stored and the new dashboard can be initialized, information should be serializable.
 * @returns {Promise.<Object>,Object} parameter specific configuration
 */
Parameter.prototype._getParameterSpecificConfig = function(willPersist) {
	return Promise.resolve({
		value: this.getValue()
	});
};

/**
 * Get the (requirejs) location of the configuration for this panel
 * this is used by the default getConfigurationPanel implementation
 * @private
 * @returns {string}
 */
Parameter.prototype._getConfigurationPanelComponent = function() {
	return this._configurationGenerator();
};

/**
 * Change which component should be used for configuring this parameter
 * @param {String} componentLocation
 */
Parameter.prototype.setConfigurationGenerator = function(componentGenerator) {
	this._configurationGenerator = componentGenerator;
};

/**
 * Should the parameter be configurable by the user
 * @param {Boolean} configurable
 */
Parameter.prototype.setConfigurableByUser = function(configurable) {
	if (this._configurableByUser !== configurable) {
		this._configurableByUser = configurable;
		this._emitChanged(this.getValue(), this.getValue());
	}
};

/**
 * Should the parameter be visbible (in configuration panel)
 * @returns {Boolean}
 */
Parameter.prototype.isVisible = function() {
	return this.isConfigurableByUser() && this._isVisible;
};

/**
 * Show/hide the parameter in the configuration panel
 * @param {Boolean} visible
 */
Parameter.prototype.setVisible = function(visible) {
	if (this._isVisible !== visible) {
		this._isVisible = visible;
		this._emitChanged(this.getValue(), this.getValue());
	}
};

/**
 * Is the parameter configurable by the user
 * @returns {Boolean}
 */
Parameter.prototype.isConfigurableByUser = function() {
	return this._configurableByUser;
};

/**
 * Get the configuration panel for the parameter.
 * This panel is used to configure this specific parameter
 * @returns {Promise.<ConfigurationPanel>}
 */
Parameter.prototype.getConfigurationPanel = function() {
	return this._getConfigurationPanelComponent().then(
		function(ParameterConfigurationPanel) {
			return new ParameterConfigurationPanel(this, this._context);
		}.bind(this)
	);
};

/**
 * By default a parameter is persistent: its configuration/definition is included in the
 * JSON configuration for the context
 * A parameter can be non-persistent: it is used for internal functionality while the dashboard is running
 * however its value should not be stored in a dashboard configuration
 *
 * @returns {Boolean}
 */
Parameter.prototype.isPersistent = function() {
	return this._isPeristent && (this._isPersistentWhenDefaultValue || this._defaultValue !== this.getValue());
};

/**
 * Dont persist the parameter if the value is equal to defaultValue
 * @param {Any} defaultValue
 */
Parameter.prototype.dontPersistIfDefaultValue = function(defaultValue) {
	this._defaultValue = defaultValue;
	this._isPersistentWhenDefaultValue = false;
};

/**
 * Change the persistency of the parameter
 * @param {Boolean} isPersistent
 */
Parameter.prototype.setPersistent = function(isPersistent) {
	this._isPeristent = isPersistent;
};

export default Parameter;
