import Observable from "@esko/observable";
import Observer from "@esko/observer";
import OrderedObject from "@esko/orderedobject";
import _ from "underscore";
/**
 * Utility class to manage a set of parameters
 * @constructor
 * @class
 */
var ParameterSet = function ParameterSet() {
	Observable.mixin(this, "ParameterSet");
	Observer.mixin(this);

	/**
	 * @member {OrderedObject.<String, Parameter>}
	 * @protected
	 */
	this._parameters = new OrderedObject();
};

/**
 * Get an ordered list of all parameters
 * @returns {Array.<Parameter>}
 */
ParameterSet.prototype.getParameterList = function() {
	return this._parameters.getOrderedObjectList();
};

/**
 * Get an ordered list of all parameters of a certain type
 * @param   {String}   type - type
 * @returns {Array.<Parameter>} ordered list
 */
ParameterSet.prototype.getParameterListByType = function(type) {
	return _.filter(this.getParameterList(), function(parameter) {
		return parameter.getType() === type;
	});
};

/**
 * Get a parameter by name
 * @param   {String}    name
 * @returns {Parameter}
 */
ParameterSet.prototype.getParameter = function(name) {
	return this._parameters.getItemByKey(name);
};

/**
 * Set/Replace a parameter
 * @param {String}    name
 * @param {Parameter} parameter
 */
ParameterSet.prototype.setParameter = function(name, parameter) {
	if (this._parameters.getItemByKey(name) !== parameter) {
		if (this._parameters.getItemByKey(name)) {
			this._stopObserveObject(this._parameters.getItemByKey(name));
		}

		var oldValue = this.getValue(name);

		this._parameters.addOrReplaceItem(name, parameter);
		this._emitParameterChanged(name, oldValue, this.getValue(name));

		if (parameter) {
			this._observe(parameter, "changed", _.partial(this._emitParameterChanged, name));
		}
	}
};
/**
 * Internal method to emit parameter change
 * @private
 * @param {String} name     - name of the parameter
 * @param {Any}    oldValue
 * @param {Any}    newValue
 */
ParameterSet.prototype._emitParameterChanged = function(name, oldValue, newValue) {
	this._emit(["changed", name].join("."), oldValue, newValue);
};

/**
 * Get the value of a parameter
 * @param   {String} name - name of the parameter
 * @returns {Any}    value for the parameter
 */
ParameterSet.prototype.getValue = function(name) {
	var param = this._parameters.getItemByKey(name);

	if (param) {
		return param.getValue();
	}

	// we will return undefined if param does not exist
	return;
};

/**
 * Set the value of a parameter
 * Note if the parameter "name" does not exist, the code will fail!
 *
 * @param {String} name  - name of the parameter
 * @param {Any}    value - new value for the parameter
 */
ParameterSet.prototype.setValue = function(name, value) {
	this._parameters.getItemByKey(name).setValue(value);
};

/**
 * Move a parameter to the end of the list
 * @param {String} name name of the parameter
 */
ParameterSet.prototype.moveParameterToEndOfList = function(name) {
	this._parameters.moveItemByKeyToEnd(name);
};

/**
 * Move a parameter to the front of the list
 * @param {String} name
 */
ParameterSet.prototype.moveParametertoFrontOfList = function(name) {
	this._parameters.moveItemByKeyToIndex(name, 0);
};

export default ParameterSet;
