import Block from "../core/Block";
import ColorUtils from "@esko/colorutils";
import Modal from "@esko/modal/Modal";
import $ from "jquery";
import _ from "underscore";
import webcss from "./BlockWrapper.css";
//timeout before loading indicator is shown
var LOADING_INDICATOR_TIMEOUT = 250;

var MINIMUM_CONFIG_WIDTH = 250;

/**
 * Proxy block used for layouting and organizing dashboards
 * @augments Block
 * @constructor
 * @class
 * @param {String} id
 * @param {Context} context
 */
var BlockWrapper = function(id, context) {
	Block.call(this, id, context);

	this._block = null;
	this._isConfiguring = false;
	this._dragDropHandler = null;
	this._element.addClass("dshBlockWrapper");
	this._hasCustomHeaderButtons = false;

	// prevent doing unneeded DOM operations
	this._previousState = {
		divDisplayed: null,
		showFrame: false,
		showHeaderButtons: false,
		showLoadingIndicator: false,
		initial: true
	};
};

BlockWrapper.prototype = Object.create(Block.prototype);
BlockWrapper.prototype.constructor = BlockWrapper;

/**
 * @Override
 */
BlockWrapper.prototype.getType = function() {
	return "dashboard/layout/BlockWrapper";
};

/**
 * @override
 */
BlockWrapper.prototype.isPersistent = function() {
	return false;
};

/**
 * Get the block for which the wrapper is made
 * @returns {Block}
 */
BlockWrapper.prototype.getBlock = function() {
	return this._block;
};

/**
 * Get the Block
 * @returns {Promise.<Block>}
 */
BlockWrapper.prototype.getBlockPromise = function() {
	return this._blockPromise;
};

/**
 * When the child block is detached, it no longer makes sense to exists... request the context to destroy the wrapper
 * @override
 */
BlockWrapper.prototype.detachChild = function(block) {
	Block.prototype.detachChild.call(this, block);

	// a block wrapper without a child is stupid... detach the wrapper from the parent
	// and destroy it
	if (this._block === block) {
		this._block = null;
		this._stopObserveObject(block);
		this._context.destroyBlock(this);
	}
};

/**
 * Initialize the header for the wrapper
 * @private
 */
BlockWrapper.prototype._initHeader = function() {
	if (!this._header) {
		this._header = $("<div>", {
			class: "header"
		});
		this._headerHandle = $("<div>", {
			class: "headerHandle"
		});

		if (this._block) {
			this._headerTitle = this._headerHandle.append(this._block.getTitleElement());

			if (this._block.isCollapsible()) {
				var self = this;
				var isCollapsed = this._block.isCollapsed();
				var icon = $("<i>", {
					class: isCollapsed ? "icon icon-WCR_Right" : "icon icon-WCR_Down"
				});
				this._collapseBtn = $("<a>", {
					class: "btn btn-no-border dshToggleBtn",
					title: isCollapsed ? this._i18n.t("EXPAND") : this._i18n.t("COLLAPSE")
				})
					.append(icon)
					.on("click", function(e) {
						e.preventDefault();
						$(this).toggleClass("collapsed");
						var collapsed = $(this).hasClass("collapsed");

						$(this).attr("title", collapsed ? self._i18n.t("EXPAND") : self._i18n.t("COLLAPSE"));
						icon.toggleClass("icon-WCR_Right", collapsed);
						icon.toggleClass("icon-WCR_Down", !collapsed);

						self._block.collapse(collapsed, true);
					});
				this._collapseBtn.toggleClass("collapsed", isCollapsed);
				this._headerHandle.prepend(this._collapseBtn);
			}
		} else {
			this._headerTitle = this._headerHandle.append($("<h4>"));
		}

		this._headerButtons = $("<div>", {
			class: "buttons"
		});

		if (this._block) {
			var customHeaderButtons = this._block.getHeaderButtons();
			if (customHeaderButtons && customHeaderButtons.length > 0) {
				this._hasCustomHeaderButtons = true;
				this._headerButtons.append(customHeaderButtons);
			}
		}

		this._deleteBtn = $("<a>", {
			class: "btn btn-no-border",
			title: this._i18n.t("DELETE_BLOCK")
		})
			.append(
				$("<i>", {
					class: "icon icon-WCR_Actions-Delete12"
				})
			)
			.on(
				"click",
				function(e) {
					e.preventDefault();
					this.deleteBlock();
				}.bind(this)
			);

		this._configureBtn = $("<a>", {
			class: "btn btn-no-border",
			title: this._i18n.t("EDIT_BLOCK_CONFIGURATION")
		})
			.append(
				$("<i>", {
					class: "icon icon icon-WCR_Actions-Edit12"
				})
			)
			.on(
				"click",
				function(e) {
					// FUTURE prevent losing unsaved operations

					this._enableBlockConfiguration();
				}.bind(this)
			);

		this._saveButton = $("<a>", {
			class: "btn btn-no-border save hidden",
			title: this._i18n.t("PREVIEW_BLOCK_CONFIGURATION")
		}).append(
			$("<i>", {
				class: "icon icon-WCR_Actions-Eye12"
			})
		);

		this._cancelButton = $("<a>", {
			class: "btn btn-no-border hidden",
			title: this._i18n.t("CANCEL_BLOCK_CONFIGURATION")
		}).append(
			$("<i>", {
				class: "icon icon-WCR_Revert12"
			})
		);

		this._saveButton.on(
			"click",
			function() {
				this.setConfiguring(false);
				this._block.onConfigurationPanelClosed();
			}.bind(this)
		);

		this._headerButtons
			.append(this._deleteBtn)
			.append(this._configureBtn)
			.append(this._saveButton);
		this._header.append(this._headerHandle);
	}
};

/**
 * Display the block in edit configuration mode
 */
BlockWrapper.prototype._enableBlockConfiguration = function() {
	this.setConfiguring(true);

	if (this.isConfiguring() && !this._configurationPanel) {
		this._blockPromise
			.then(function(block) {
				return block.getConfigurationPanel();
			})
			.then(
				function(configPanel) {
					this._configurationPanel = configPanel;
					this._render();
				}.bind(this)
			);
	} else if (this.isConfiguring() && this._configurationPanel) {
		this._render();
	}
};

/**
 * Delete the block (--> detach it, we don't want to jinx everything)
 * @returns {Promise} resolved once the block is deleted
 */
BlockWrapper.prototype.deleteBlock = function() {
	return this._blockPromise.then(
		function(block) {
			this._context.detachBlock([this.getParent()], block);
		}.bind(this)
	);
};

/**
 * Get the ID of the wrapped block
 * @returns {String} id of the wrapped block
 */
BlockWrapper.prototype.getWrappedBlockID = function() {
	return this._blockID;
};

/**
 * @override
 */
BlockWrapper.prototype._doConfigure = function(config) {
	var self = this;

	self._blockID = config.blockID;

	return self._context.getLanguageBundle("Dashboard").then(function(bundle) {
		self._i18n = bundle;

		// start loading block, we do not want the block initialization till block is loaded
		var whenFutureBlockAvailable = self._registerFutureChild(self._blockID);
		self._blockPromise = self._context
			.loadBlock(self._blockID)
			.then(function(block) {
				self._block = block;

				self._observe(block, "parameter.changed.blockTitle", self._render);
				self._observe(block, "parameter.changed.blockShowFrame", self._render);
				self._observe(block, "parameter.changed.blockFrameColor", self._render);
				self._observe(block, "parameter.changed.blockShowInEditMode", self._render);
				self._observe(block, "configured", self._blockConfigured);
				self._registerChild(block);
			})
			.then(function() {
				self.once("render", whenFutureBlockAvailable);
				self._render();
				return self._block;
			});

		self._registerLoadingStart();
		self._render();
	});
};

/**
 * This is used to only show the loading indicator after a given amount of time
 * @private
 */
BlockWrapper.prototype._registerLoadingStart = function() {
	this._loadingStart = Date.now();
};

/**
 * @override
 */
BlockWrapper.prototype._doInit = function() {
	this._observe(this._context, "editModeActive", this._render);
};

/**
 * Can the block be configured?
 * @returns {Boolean}
 */
BlockWrapper.prototype.canConfigure = function() {
	return this.allowConfigure() && this._block && this._block.canConfigure();
};

/**
 * Is the dashboard in configuration mode? Should we show buttons to configure
 * @returns {Boolean}
 */
Block.prototype.allowConfigure = function() {
	return this._context.isEditModeActive();
};

/**
 * Is the BlockWrapper in configure mode
 * @returns {Boolean}
 */
BlockWrapper.prototype.isConfiguring = function() {
	return this._isConfiguring;
};

/**
 * toggle configuration mode
 * @param {Boolean} isConfiguring
 */
BlockWrapper.prototype.setConfiguring = function(isConfiguring) {
	if (this._isConfiguring != isConfiguring) {
		this._isConfiguring = isConfiguring;
		this._registerLoadingStart();
		this._render();
	}
	if (!isConfiguring) {
		if (this._configurationPanel) {
			this._configurationPanel.destroy();
			this._configurationPanel = null;
		}
	}
};

/**
 * Should the frame be shown?
 * --> is the block configured to always show frame or is the dashboard in configuration mode
 * @private
 * @returns {Boolean}
 */
BlockWrapper.prototype._shouldShowFrame = function() {
	return this.allowConfigure() || (this._block && (this._block.shouldShowFrame() || this._block.showAlwaysTransparentFrame()));
};

/**
 * Update the rendering of the BlockWrapper
 * @private
 */
BlockWrapper.prototype._doRender = function() {
	// Display edit mode if block can be configured and if widget size is not small
	var editModeParam = this._block ? this._block.getParameterByName("blockShowInEditMode") : undefined;
	if (editModeParam && this.canConfigure() && editModeParam.getValue() && this._element[0].offsetWidth >= MINIMUM_CONFIG_WIDTH) {
		editModeParam.setValue(false);
		this._enableBlockConfiguration();
	} else {
		var showFrame = this._shouldShowFrame();

		var newState = {
			divDisplayed: null,
			showFrame: showFrame || false,
			showHeader: showFrame || (this._block && this._block.showAlwaysTransparentFrame()),
			showHeaderButtons: false,
			showLoadingIndicator: false,
			headerElement: this._headerTitle,
			blockInDOM: false
		};

		if (showFrame) {
			this._initHeader();

			if (this._block) {
				newState.headerElement = this._block.getTitleElement();

				// update frame color
				var frameColor = "";
				var contrastColor = "";
				if (this._block.shouldShowFrame()) {
					frameColor = this._block.getFrameColor();
					contrastColor = ColorUtils.calculateContrastColor(frameColor);
				} else if (!this._context.isEditModeActive() && this._block.showAlwaysTransparentFrame()) {
					frameColor = "transparent";
				}

				this._header.css({
					"background-color": frameColor,
					color: contrastColor
				});

				var borderColor = frameColor;

				if (this.allowConfigure() && this._block && !this._block.shouldShowFrame()) {
					borderColor = "transparent";
				}

				this._element.css("border-color", borderColor);

				if (this._headerButtons) {
					// we want the default icon color if contrast color is 0000000
					if (contrastColor == "#000000") {
						contrastColor = "";
					}
					this._headerButtons.find("i").css("color", contrastColor);
				}
			}

			// update header buttons
			newState.showHeaderButtons = this.allowConfigure() || this._hasCustomHeaderButtons;

			this._saveButton.toggleClass("hidden", !this.isConfiguring());
			this._configureBtn.toggleClass("hidden", !this.canConfigure() || this.isConfiguring());
			this._deleteBtn.toggleClass("hidden", !this.allowConfigure());
			if (this._collapseBtn) {
				this._collapseBtn.toggleClass("hidden", this.allowConfigure());
			}
		}

		if ((this._block && !this.isConfiguring()) || this._configurationPanel) {
			if (this.isConfiguring()) {
				if (this._configurationModal || this._element[0].offsetWidth < MINIMUM_CONFIG_WIDTH) {
					if (this._block) {
						newState.divDisplayed = this._block.getElement();
						newState.blockInDOM = true;
					}
					this._openConfigurationPanelInModal();
				} else {
					newState.divDisplayed = this._configurationPanel.getElement();
				}
			} else {
				newState.divDisplayed = this._block.getElement();
				newState.blockInDOM = true;
			}
		} else {
			if (Date.now() - this._loadingStart > LOADING_INDICATOR_TIMEOUT) {
				// show loading indicator
				newState.showLoadingIndicator = true;
			} else {
				// don't show loading indicator yet, but request rerender after timeout has passed
				setTimeout(this._render.bind(this), LOADING_INDICATOR_TIMEOUT);
			}
		}

		// update rendering if needed
		if (!_.isEqual(this._previousState, newState)) {
			if (this._headerTitle) {
				this._headerTitle.children().detach();
			}
			if (newState.headerElement) {
				this._headerTitle = this._headerHandle.append(newState.headerElement);
			}
			if (this._collapseBtn) {
				this._headerHandle.prepend(this._collapseBtn);
			}

			if (
				this._previousState.divDisplayed != newState.divDisplayed ||
				newState.showLoadingIndicator ||
				this._previousState.showHeader != newState.showHeader ||
				this._previousState.initial
			) {
				this._element.children().detach();
				this._element.empty();
				this._element.append(this._header);
				if (newState.showLoadingIndicator) {
					this._element.append(
						$("<span>", {
							text: this._i18n.t("LOADING_BLOCK")
						})
					);
				} else {
					this._element.append(newState.divDisplayed);
				}
			}

			this._element.toggleClass("showFrame", newState.showFrame);
			this._element.toggleClass("allowDrag", this.allowConfigure());

			if (this._header) {
				this._header.toggleClass("hidden", !newState.showHeader);
			}
			if (newState.showHeaderButtons) {
				this._header.append(this._headerButtons);
			} else if (this._headerButtons) {
				this._headerButtons.detach();
			}
			this._element.toggleClass("blockloading", newState.showLoadingIndicator || newState.divDisplayed === null);

			if (this._block) {
				this._block.setInDOM(newState.blockInDOM && this.isInDOM());
			}

			this._previousState = newState;
		}

		// update draggable
		if (this._dragDropHandler) {
			this.toggleDragging();
		} else if (this.allowConfigure()) {
			// start loading the dragdrop handler and enable dragging
			this._context.getDragDropHandler().then(
				function(dragDropHandler) {
					this._dragDropHandler = dragDropHandler;
					this.toggleDragging();
				}.bind(this)
			);
		}

		if (!this.isConfiguring() && this._configurationPanel) {
			this._configurationPanel.destroy();
			this._configurationPanel = null;
		}
	}
};

/**
 * Opens a modal showing the configuration panel
 * @private
 */
BlockWrapper.prototype._openConfigurationPanelInModal = function() {
	if (!this._configurationModal) {
		this._configurationModal = new Modal(this._configurationPanel.getElement(), {
			widthClass: "modal-lg",
			onClose: function() {
				this.setConfiguring(false);
				var modal = this._configurationModal;
				modal.destroy();
				this._configurationModal = null;
			}.bind(this),
			closeOnBackdropClick: false
		});
		this._configurationModal.setTitle(this._block.getTitle());
		this._configurationModal.open();

		this._configurationModal.setFooter(
			$("<div>").append(
				$("<a>", {
					class: "btn btn-primary",
					text: this._i18n.t("SAVE_BLOCK_CONFIGURATION")
				}).on(
					"click",
					function() {
						this._configurationModal.close();
					}.bind(this)
				)
			)
		);
	}
};

/**
 * Enable/disable Dragging
 */
BlockWrapper.prototype.toggleDragging = function() {
	this._dragDropHandler.toggleDragging(
		this.allowConfigure(),
		this._context,
		this._context.getBlockDefinition(this._blockID),
		this.getParent(),
		this._element,
		this._headerHandle
	);
};

/**
 * The wrapped block was configured, if in configuration mode replace the panel
 * @private
 */
BlockWrapper.prototype._blockConfigured = function() {
	if (this.isConfiguring()) {
		this._block.getConfigurationPanel().then(
			function(configPanel) {
				this._configurationPanel = configPanel;
				this._render();
			}.bind(this)
		);
	}
};

/**
 * If we no longer have a parent, remove our child
 * @override
 */
BlockWrapper.prototype.setParent = function(parent) {
	Block.prototype.setParent.call(this, parent);

	if (parent === null && this._block) {
		this._block.setParent(null);
	}
};

/**
 * @override
 */
BlockWrapper.prototype._childInDOM = function(child) {
	return child == this._block && this._previousState.blockInDOM;
};

export default BlockWrapper;
