import Layout from "../core/Layout";
import CssSizeParameter from "../core/parameters/CssSizeParameter";
import _ from "underscore";
import "./ListLayout.css";
/**
 * A list layout shows blocks in a list! (crazy huh?)
 *
 * @constructor
 * @class
 * @augments Layout
 * @param {String} id
 * @param {Context} context
 */
var ListLayout = function(id, context) {
	Layout.call(this, id, context);

	this._element.addClass("ListLayout");
	this._list = [];
};

ListLayout.prototype = Object.create(Layout.prototype);
ListLayout.prototype.constructor = ListLayout;

/**
 * @Override
 */
ListLayout.prototype.getType = function() {
	return "dashboard/layout/ListLayout";
};

/**
 * @Override
 */
ListLayout.prototype.canConfigure = function() {
	return true;
};

/**
 * @override
 */
ListLayout.prototype._doConfigure = function(config, IDMap) {
	if (config.items) {
		var newList = _.map(
			config.items,
			function(item) {
				var id = IDMap.mapID(item.id);

				var listItem = _.findWhere(this._list, {
					id: id
				});
				if (listItem) {
					return listItem;
				} else {
					return {
						id: id
					};
				}
			},
			this
		);

		return this._updateList(newList);
	} else {
		this._render();
	}
};

/**
 * @override
 */
ListLayout.prototype._doInit = function() {
	this._observe(this._context, "editModeActive", this._initializeDropzone);
	this._initializeDropzone();
	this._observe(this._context, "CssSizeParameter", this._render);
};

/**
 * @override
 */
ListLayout.prototype._doRender = function() {
	//add margin to the blocks in the list layout
	var margin = this._parameters.getParameter("CssSizeParameter").getValue();
	for (var i = 0; i < this._list.length - 1; i++) {
		var item = this._list[i];
		item.blockWrapper._element.css("margin-bottom", margin);
	}
};

/**
 * Wait until all the blockwrappers of a given list are loaded
 * Once the list is processed, always call _markListProcessed
 * @private
 * @param   {Array.<ListLayoutItem>}           list - the list for which all blockwrappers should be loaded
 * @returns {Promise.<Array.<ListLayoutItem>>} resolved once all blockwrappers are available
 */
ListLayout.prototype._waitForList = function(list) {
	return Promise.all(
		_.map(
			list,
			function(item) {
				if (!item.blockWrapperPromise) {
					item = {
						id: item.id,
						blockWrapperPromise: this._loadBlockWrapper(item.id)
					};

					this._registerFutureChild(item.id);
				}

				return item.blockWrapperPromise.then(function(blockWrapper) {
					item.blockWrapper = blockWrapper;
					return item;
				});
			}.bind(this)
		)
	);
};

/**
 * Whenever you wait for a list, you should also call this method
 * to make sure the future children is set correctly
 * @private
 * @param {Array.<ListLayoutItem>} list
 */
ListLayout.prototype._markListProcessed = function(list) {
	_.each(
		list,
		function(item) {
			this._markFutureChildResolved(item.id);
		},
		this
	);
};

/**
 * Update the rendering of the widget, by updating the list
 * @private
 * @param   {Array.<ListLayoutItem>} newList - new configuration of the layout
 * @returns {Promise} resolved once the rendering is updated
 */
ListLayout.prototype._updateList = function(newList) {
	var self = this;

	newList = _.map(
		newList,
		function(item) {
			var listItem = _.findWhere(this._list, {
				id: item.id
			});
			if (listItem) {
				return listItem;
			} else {
				return {
					id: item.id
				};
			}
		},
		this
	);

	var oldList = self._list.slice(0);

	// wait for the lists to be complete, then update the rendering
	return Promise.all([this._waitForList(oldList), this._waitForList(newList)]).then(function(lists) {
		var item;
		var oldList = lists[0];
		var newList = lists[1];

		var oldListPos = 0;

		// run over the new list
		for (var newListPos = 0; newListPos < newList.length; newListPos++) {
			if (oldList[oldListPos] === newList[newListPos]) {
				// item is at the same location as in the old list
				oldListPos++;
			} else {
				// another item is in the new list
				item = newList[newListPos];
				var positionInOldList = _.indexOf(oldList, item);
				if (positionInOldList >= 0) {
					// item already exists in the old list, remove it from there
					oldList.splice(positionInOldList, 1);
				}

				// make sure to register the item as a child
				self._registerChild(item.blockWrapper);

				// now insert the "new" item at the correct location
				if (newList[newListPos - 1]) {
					newList[newListPos - 1].blockWrapper.getElement().after(item.blockWrapper.getElement());
				} else {
					self._element.prepend(item.blockWrapper.getElement());
				}
				item.blockWrapper.setInDOM(self.isInDOM());
			}
		}

		// detach the elements which we don't want anymore
		while (oldListPos < oldList.length) {
			item = oldList[oldListPos++];

			item.blockWrapper.setInDOM(false);
			item.blockWrapper.getElement().detach();
			item.blockWrapper.setParent(null);
		}

		self._list = newList;
		self._render();

		self._markListProcessed(oldList);
		self._markListProcessed(newList);
	});
};

/**
 * Get the configuration for the List
 * @Override
 * @returns {object}
 */
ListLayout.prototype._getLayoutSpecificConfiguration = function() {
	return {
		items: _.map(this._list, function(item) {
			return {
				id: item.id
			};
		})
	};
};

/**
 * Make sure to remove the child from our internal list once detacted
 * @override
 * @param   {object}   block - Block that is being removed
 */
ListLayout.prototype.detachChild = function(block) {
	Layout.prototype.detachChild.call(this, block);

	if (block.getWrappedBlockID) {
		this._list = _.filter(this._list, function(item) {
			return item.id !== block.getWrappedBlockID();
		});
	}
};

/**
 * Look up an item for a certain id
 * @private
 * @param   {String} id - id of the item
 * @returns {Object} list item
 */
ListLayout.prototype._getItemForID = function(id) {
	return _.findWhere(this._list, {
		id: id
	});
};

/**
 * Configure the default parameters of the block
 * @override
 */
ListLayout.prototype._initializeDefaultParameters = function() {
	return this._context.getLanguageBundle("ListLayout").then(
		function(i18n) {
			var parameter = new CssSizeParameter("CssSizeParameter", this._context, i18n.t("SPACING"));
			parameter.setValue("10px");
			this._addParameter(parameter);
			this._observeParam(
				"CssSizeParameter",
				function() {
					this._render();
				}.bind(this)
			);
		}.bind(this)
	);
};

/**
 * Initialize the dropzone
 * @private
 */
ListLayout.prototype._initializeDropzone = function() {
	if (this._context.isEditModeActive()) {
		import("./ListLayoutDragDrop").then(
			function() {
				this._ddInitializeDropzone();
			}.bind(this)
		);
	}

	this._element.toggleClass("isDroppable", this._context.isEditModeActive());
};

/**
 * Get a list of items in the layout
 * @override
 * @returns {Array.<String>}
 */
ListLayout.prototype.getItemList = function() {
	return _.pluck(this._list, "id");
};

/**
 * Set the list of items in the layout
 * @override
 * @param {Array.<String>} list - ordered list of ids
 */
ListLayout.prototype.setItemList = function(list) {
	var mappedList = _.map(list, function(value) {
		return { id: value };
	});

	return this._updateList(mappedList);
};

export default ListLayout;
