import ListLayout from "./ListLayout";
import $ from "jquery";
import _ from "underscore";
/**
 * This file extends the ListLayout prototype, to include the Drag And Drop functionality
 */

/**
 * Initialize the Dropzone
 * @private
 */
ListLayout.prototype._ddInitializeDropzone = function() {
	var self = this;
	if (!this._ddInitialized) {
		this._ddInitialized = true;

		this._context.getDragDropHandler().then(function(dragDropHandler) {
			dragDropHandler.registerDropzone(
				self,
				self.getElement(),
				{
					accept: self._ddAccept,
					over: self._ddOver,
					out: self._ddOut,
					drag: self._ddDrag,
					drop: self._ddDrop
				},
				self
			);
		});
	}
};

/**
 * Handle over event of the droppable
 * @private
 * @param {Event}  event - event
 * @param {Object} ui    - JQueryUI extra info
 */
ListLayout.prototype._ddOver = function(event, ui) {
	if (this._ddGhost) {
		this._ddGhost.remove();
	}

	this._ddInsertAt = 0;

	this._ddGhost = $("<div>", {
		style: "border: 1px dotted #ccc; padding: 10px; margin: 5px;"
	});

	this._ddCalculateInsertPosition(event, ui);
	this._ddUpdateGhost();
};

/**
 * Recalculate the position for the block/ghost, based on the UI helper
 * @private
 * @param {Event}  event - event
 * @param {Object} ui    - JQueryUI extra info
 */
ListLayout.prototype._ddCalculateInsertPosition = function(event, ui) {
	var helperDomElement = ui.helper.get(0);
	var helperBoundingBox = helperDomElement.getBoundingClientRect();
	var blockDefinition = ui.helper.data("BlockDefinition");
	var helperTop = helperBoundingBox.top;

	var i = 0;

	while (
		i < this._list.length &&
		(!this._list[i].blockWrapper ||
			this._list[i].id === blockDefinition.getID() ||
			this._list[i].blockWrapper
				.getElement()
				.get(0)
				.getBoundingClientRect().top < helperTop)
	) {
		i++;
	}

	this._ddInsertAt = i;
};

/**
 * Update the location of the ghost in the DOM
 * @private
 */
ListLayout.prototype._ddUpdateGhost = function() {
	this._ddInsertElementAt(this._ddInsertAt, this._ddGhost);
};

/**
 * Handle the droppable out event
 * @private
 */
ListLayout.prototype._ddOut = function() {
	this._ddGhost.remove();
};

/**
 * Handle the draggable drag event
 * @private
 * @param {Event}  event - event
 * @param {Object} ui    - JQueryUI extra info
 */
ListLayout.prototype._ddDrag = function(event, ui) {
	this._ddCalculateInsertPosition(event, ui);
	this._ddUpdateGhost();
};

/**
 * Handle the droppable drop event
 * @private
 * @param {Event}  event - event
 * @param {Object} ui    - JQueryUI extra info
 */
ListLayout.prototype._ddDrop = function(event, ui) {
	this._ddGhost.remove();

	var blockDefinition = ui.helper.data("BlockDefinition");
	var insertAt = this._ddInsertAt;
	var currentParent = ui.helper.data("CurrentParent");
	var newListInformation = this._ddGenerateNewList(blockDefinition, insertAt, currentParent);

	var newList = newListInformation.list;

	// compare lists
	if (
		this._list.length === newList.length &&
		_.every(this._list, function(item, index) {
			return newList[index] && item.id === newList[index].id;
		})
	) {
		// the order has stayed the same

		return true;
	} else {
		// return a promise, so the draggable is only shown once it is moved
		return this._context.getLanguageBundle("Dashboard").then(
			function(i18n) {
				// we need to recalculate the list... there still might have been an user action ongoing
				var finalList = null;
				return this._context.undoableAction(
					newListInformation.affectedBlocks,
					function() {
						if (finalList === null) {
							finalList = this._ddGenerateNewList(blockDefinition, insertAt, currentParent).list;
						}

						return this._updateList(finalList);
					}.bind(this),
					i18n.t("MOVE_BLOCK_UNDOREDOMESSAGE", blockDefinition.getDisplayName())
				);
			}.bind(this)
		);
	}
};

/**
 * Generate the new list and affectedblocks based on the insert postion
 * @private
 * @param   {BlockDefinition} blockDefinition - definition of the block that is being dropped
 * @param   {Number}          insertPosition  - insert position
 * @param   {Block}           currentParent   - the current parent of the block that is being dropped
 * @returns {object}          {list: newList, affectedBlock: [blocks, that, are, affected]}
 */
ListLayout.prototype._ddGenerateNewList = function(blockDefinition, insertPosition, currentParent) {
	// block is already in the layout
	var itemToInsertBefore = this._list[insertPosition];

	var item = this._getItemForID(blockDefinition.getID());

	// first remove the item from the list(this also works if the item is not yet in the list)
	var newList = _.without(this._list, item);
	var newlyCreatedItem = null;
	var insertAt = _.indexOf(newList, itemToInsertBefore);
	if (insertAt < 0) {
		insertAt = newList.length;
	}

	var affectedBlocks = [this];

	if (!item) {
		// item comes from another location
		newlyCreatedItem = {
			id: blockDefinition.getID()
		};

		item = newlyCreatedItem;

		if (currentParent) {
			affectedBlocks.push(currentParent);
		}
	}

	newList.splice(insertAt, 0, item);

	return { list: newList, affectedBlocks: affectedBlocks };
};

/**
 * Insert an element in the DOM at a certain position
 * @private
 * @param {Number}    insertAt - position before which we want to insert the element
 * @param {JQueryObj} element  - element to be inserted
 */
ListLayout.prototype._ddInsertElementAt = function(insertAt, element) {
	while (insertAt < this._list.length && !this._list[insertAt].blockWrapper) {
		insertAt++;
	}

	if (this._list[insertAt] && this._list[insertAt].blockWrapper) {
		element.insertBefore(this._list[insertAt].blockWrapper.getElement());
	} else {
		//append element in the back
		this._element.append(element);
	}
};
