import Parameter from "../../core/parameters/Parameter";
import FlexItem from "./FlexItem";
import _ from "underscore";
/**
 * Flex Column Parameter
 * @augments Parameter
 * @param {String} name             - name of the parameter
 * @param {String} userFriendlyName - user friendly representation of the parameter
 */
var FlexColumnParameter = function(name, context, layout, userFriendlyName) {
	Parameter.call(this, name, context, "dashboard/layout/flex/ColumnParameter", userFriendlyName);

	/**
	 * @member {Array.<FlexItem>} array containing the flexItems of the coloumns
	 */
	this._columns = [];

	this._layout = layout;

	this._toBeDeleted = [];
};

FlexColumnParameter.prototype = Object.create(Parameter.prototype);
FlexColumnParameter.prototype.constructor = FlexColumnParameter;

/**
 * Get the configuration element for the parameter
 * @returns {Promise.<ConfigurationPanel>}
 */
FlexColumnParameter.prototype.getConfigurationPanel = function() {
	return Promise.all([import("./FlexLayoutConfigurationPanel"), this._context.getLanguageBundle("DashboardFlexLayout")]).then(
		function(values) {
			var FlexLayoutConfigurationPanel = values[0].default;
			var i18n = values[1];

			return new FlexLayoutConfigurationPanel(this, i18n);
		}.bind(this)
	);
};

/**
 * @override
 */
FlexColumnParameter.prototype.configure = function(config, idMap) {
	_.each(config.value.columns, function(columnDefinition) {
		columnDefinition.blockid = idMap.mapID(columnDefinition.blockid);
	});

	this.setValue(config.value);
};

/**
 * Returns an array of flexItems (copy)
 * @returns {Array.<FlexItem>}
 */
FlexColumnParameter.prototype.getItemList = function() {
	return this._columns.slice(0);
};

/**
 * Sets a new item
 */
FlexColumnParameter.prototype.setItem = function(columnNR, blockID) {
	this.getItemList()[columnNR].setItem(blockID);
	this.setValue(this._getSpecification());
};

/**
 * converts the columns into the value for this parameter
 */
FlexColumnParameter.prototype._getSpecification = function() {
	return {
		columns: _.map(this.getItemList(), function(column) {
			return column.getConfiguration();
		})
	};
};

/**
 * Used by config item to show block title
 */
FlexColumnParameter.prototype.getBlockTitle = function(blockId) {
	return this._context.loadBlock(blockId).then(function(block) {
		return block.getTitle();
	});
};

/**
 * Updates the parameter with the current configuration in the itemlist.
 * (To be called when the itemlist is adapted outside this parameter)
 */
FlexColumnParameter.prototype.update = function() {
	this.setValue(this._getSpecification());
};

/**
 * Return the items that should be deleted in the DOM and flushes the list
 * @returns {Array.<FlexItem>}
 */
FlexColumnParameter.prototype.getToBeDeletedItems = function() {
	return this._toBeDeleted.splice(0, this._toBeDeleted.length);
};

/**
 * Should the columns wrap?
 * @returns {Boolean}
 */
FlexColumnParameter.prototype.shouldWrap = function() {
	return _.some(this._columns, function(flexItem) {
		return flexItem.hasWidthDefined();
	});
};

/**
 * Sets the value of this parameter
 * @param {object} value
 */
FlexColumnParameter.prototype.setValue = function(value) {
	this._layout._element.css("overflow-x", "auto");

	var previousSetup = this._columns.slice(0); // copy
	var previousSetupPos = 0;
	this._columns = [];

	var totalCols = value.columns.length;
	for (var newSetupPos = 0; newSetupPos < value.columns.length; newSetupPos++) {
		var flexItem = previousSetup[previousSetupPos];
		var columnDefinition = value.columns[newSetupPos];

		if (flexItem && flexItem.getBlockID() === columnDefinition.blockid) {
			// It is the same flexItem, so we can copy the flexItem from the previous setup
			this._columns[newSetupPos] = flexItem;
			this._columns[newSetupPos].configure(columnDefinition, totalCols);
			// we don't need to do anything
			previousSetupPos++;
		} else {
			var i = previousSetupPos;
			flexItem = null;
			// Search for the corresponding flexItem in the previous configuration
			while (i < previousSetup.length && !flexItem) {
				if (previousSetup[i].getBlockID() === columnDefinition.blockid) {
					flexItem = previousSetup[i];
					previousSetup.splice(i, 1);
				} else {
					i++;
				}
			}

			if (!flexItem) {
				// A new column is added => create new FlexItem
				flexItem = new FlexItem(this._layout, newSetupPos, columnDefinition, totalCols);
			} else {
				flexItem.configure(columnDefinition, totalCols);
			}

			// make sure the ID is always set correctly
			this._columns[newSetupPos] = flexItem;
		}

		this._columns[newSetupPos].setID(newSetupPos);
	}

	while (previousSetupPos < previousSetup.length) {
		var item = previousSetup[previousSetupPos++];
		this._toBeDeleted.push(item);
	}

	Parameter.prototype.setValue.apply(this, arguments);
};

export default FlexColumnParameter;
