import UnitInputField from "./UnitInputField";
import $ from "jquery";
import jqueryUI from "jqueryUI";
import Observable from "@esko/observable";
import _ from "underscore";
/**
 * A simple UI element to display configuration for a FlexColumn
 * @param {object}   columnConfig - column config object
 * @param {LanguageBundle} i18n         - language bundle for FlexLayout
 */
var FlexConfigItem = function(columnConfig, i18n, parameter) {
	Observable.mixin(this, "FlexConfigItem");
	this._parameter = parameter;
	this._columnConfig = _.clone(columnConfig);

	this._i18n = i18n;

	this._element = $("<div>").addClass("flexConfigItem");
	this._element.data("flexConfigItem", this);

	this._deleteBtn = $("<a>", {
		class: "btn btn-no-border pull-right",
		title: i18n.t("REMOVE_COLUMN")
	}).append(
		$("<i>", {
			class: "icon icon-WCR_Actions-Delete12"
		})
	);

	var title = $("<div>", {
		class: "flexConfig-title"
	});
	title.append(this._deleteBtn);
	this._blockTitle = $("<h5>");

	title.append(this._blockTitle);
	this._element.append(title);
	this._updateBlockTitle();

	this._minWidth = new UnitInputField(i18n.t("COLUMN_MIN_WIDTH"), columnConfig.minWidth, ["px", "%"]);
	this._element.append(this._minWidth.element);
	this._minWidth.element.addClass("minWidth");

	this._minWidthIsContentWidth = $("<label>", {
		class: "checkbox-label"
	}).append(
		$("<span>", {
			text: i18n.t("COLUMN_FIT_CONTENT")
		})
	);
	this._minWidthIsContentWidthCheckbox = $("<input>", {
		type: "checkbox"
	});
	this._minWidthIsContentWidthCheckbox.prop("checked", columnConfig.minWidthIsContent);

	this._minWidthIsContentWidth.prepend(this._minWidthIsContentWidthCheckbox);
	this._minWidth.element.find("label").after(this._minWidthIsContentWidth);

	this._width = new UnitInputField(i18n.t("COLUMN_WIDTH"), columnConfig.width, ["px", "%"]);
	this._width.element.addClass("width");
	this._element.append(this._width.element);

	this._widthIsContentWidth = $("<label>", {
		class: "checkbox-label"
	}).append(
		$("<span>", {
			text: i18n.t("COLUMN_FIT_CONTENT")
		})
	);
	this._widthIsContentWidthCheckbox = $("<input>", {
		type: "checkbox"
	});
	this._widthIsContentWidthCheckbox.prop("checked", columnConfig.fitContent);

	this._widthIsContentWidth.prepend(this._widthIsContentWidthCheckbox);
	this._width.element.find("label").after(this._widthIsContentWidth);

	this._maxWidth = new UnitInputField(i18n.t("COLUMN_MAX_WIDTH"), columnConfig.maxWidth, ["px", "%"]);
	this._element.append(this._maxWidth.element);

	this._width.input.on("keyup", _.debounce(this._updateFieldVisibility.bind(this), 40));
	this._width.unitSelector.on("change", this._updateFieldVisibility.bind(this));

	this._updateFieldVisibility();

	if (columnConfig.blockid) {
		$("i", this._deleteBtn).addClass("icon-warning");
		this._deleteBtn.attr("title", i18n.t("WARNING_COLUMN_CONTAINS_CONTENT"));
	}

	this._deleteBtn.on(
		"click",
		function() {
			this._element.remove();
			this._emit("deleted");
		}.bind(this)
	);
};

/**
 * Update the visibility of the input fields
 * @private
 */
FlexConfigItem.prototype._updateFieldVisibility = function() {
	var enableMinAndMaxWidth = this._isMinMaxEnabled();

	this._width.element.find(".input-group").toggleClass("disabled", this._isFitContent());
	this._minWidth.element.toggleClass("disabled", !enableMinAndMaxWidth);
	this._minWidth.element.find(".input-group").toggleClass("disabled", this._isMinWidthContent());
	this._maxWidth.element.toggleClass("disabled", !enableMinAndMaxWidth);
};

/**
 * Is fit content enabled?
 * @private
 * @returns {Boolean}
 */
FlexConfigItem.prototype._isFitContent = function() {
	return this._widthIsContentWidthCheckbox.prop("checked");
};

FlexConfigItem.prototype._isMinWidthContent = function() {
	return this._minWidthIsContentWidthCheckbox.prop("checked");
};

/**
 * Is defining a mininum and max width possible
 * @private
 * @returns {Boolean}
 */
FlexConfigItem.prototype._isMinMaxEnabled = function() {
	return !this._isFitContent() && (this._width.getValue() === "" || this._width.getUnit() === "%");
};

/**
 * Get the DOM element of the FlexConfigItem
 * @returns {JQueryObj}
 */
FlexConfigItem.prototype.getElement = function() {
	return this._element;
};

/**
 * Get the Configuration of the FLexColumn
 * @returns {Object}
 */
FlexConfigItem.prototype.getConfiguration = function() {
	var config = _.clone(this._columnConfig);

	var growshrink = this._width.getValue() === "" ? 1 : 0;

	config.minWidthIsContent = this._isMinWidthContent();
	config.fitContent = this._isFitContent();
	config.grow = growshrink;
	config.shrink = growshrink;

	config.width = this._width.getValue();

	config.minWidth = this._minWidth.getValue();
	config.maxWidth = this._maxWidth.getValue();

	if (!this._isMinMaxEnabled()) {
		delete config.minWidth;
		delete config.maxWidth;
	}

	if (this._isFitContent()) {
		config.grow = 0;
		config.shrink = 0;
		config.width = "";
	}

	if (this._isMinWidthContent()) {
		config.shrink = 0;
		config.minWidth = "";
	}

	return config;
};

/**
 * Observe all input fields of a FlexConfigItem
 * @param {function} handler - function that have to be executed when something is changed
 */
FlexConfigItem.prototype.observe = function(handler) {
	this._widthIsContentWidthCheckbox.on("change", handler);
	this._minWidthIsContentWidthCheckbox.on("change", handler);
	this._width.on("changed", handler);
	this._minWidth.on("changed", handler);
	this._maxWidth.on("changed", handler);
	this.on("deleted", handler);
};

/**
 * Update DOM elements with the given values
 * @param {object} columnConfig
 */
FlexConfigItem.prototype.update = function(columnConfig) {
	this._columnConfig = _.clone(columnConfig);
	// Update values
	this._width.setValue(columnConfig.width);
	this._minWidth.setValue(columnConfig.minWidth);
	this._maxWidth.setValue(columnConfig.maxWidth);
	this._minWidthIsContentWidthCheckbox.prop("checked", columnConfig.minWidthIsContent);
	this._widthIsContentWidthCheckbox.prop("checked", columnConfig.fitContent);

	this._updateBlockTitle();

	// Update rendering
	this._updateFieldVisibility();
	if (columnConfig.blockid) {
		$("i", this._deleteBtn).addClass("icon-warning");
		this._deleteBtn.attr("title", this._i18n.t("WARNING_COLUMN_CONTAINS_CONTENT"));
	} else {
		$("i", this._deleteBtn).removeClass("icon-warning");
		this._deleteBtn.removeAttr("title");
	}
};

/**
 * Render the blocktitle correctly
 */
FlexConfigItem.prototype._updateBlockTitle = function() {
	if (!this._columnConfig.blockid) {
		this._blockTitle.text("");
	} else {
		var blockid = this._columnConfig.blockid;
		this._parameter.getBlockTitle(this._columnConfig.blockid).then(
			function(title) {
				if (this._columnConfig.blockid === blockid) {
					this._blockTitle.text(title).attr("title", title);
				}
			}.bind(this)
		);
	}
};
/**
 * Enable/disable the columnremoval button
 * @param {Boolean} allowRemoval
 */
FlexConfigItem.prototype.allowColumnRemoval = function(allowRemoval) {
	this._deleteBtn.toggleClass("disabled", !allowRemoval);
};

export default FlexConfigItem;
