import $ from "jquery";
import _ from "underscore";
/**
 * Flex Item, data object to represent a Column in a Flex Layout
 * @class
 *
 * @param {FlexLayout} layout
 * @param {Number} ID - number of the column
 * @param {Object} columnConfig
 */
var FlexItem = function(layout, ID, columnConfig, totalColumns) {
	this._layout = layout;
	this._columnID = ID;

	this._blockID = null;
	this._blockPromise = null;
	this._block = null;
	this._element = $("<div>", {
		class: "dshFlexItem"
	});
	this._wrapper = $("<div>");
	this._element.append(this._wrapper);

	this._contentElement = null;
	this._isPlaceholder = null;

	this.configure(columnConfig, totalColumns);
	this._layout._initializePlaceholder(this);
};

/**
 * Set the ID of the FlexItem (sequence number)
 * @param {Number} ID
 */
FlexItem.prototype.setID = function(ID) {
	if (this._columnID != ID) {
		this._columnID = ID;
		this.applyColumnStyles();
	}
};

/**
 * Get the ID of the FlexItem (sequence number)
 * @returns {Number} ID
 */
FlexItem.prototype.getID = function() {
	return this._columnID;
};

/**
 * Update the configuration of the FlexItem
 * @param {object} columnConfig
 */
FlexItem.prototype.configure = function(columnConfig, totalColumns) {
	// we are only interested in a subset of configuration keys
	this._config = _.pick(
		columnConfig,
		"fitContent",
		"grow",
		"shrink",
		"width",
		"height",
		"maxWidth",
		"minWidth",
		"lockedColumn",
		"minWidthIsContent",
		"className",
		"showVerticalScrollbar"
	);
	this._totalColumns = totalColumns;
	if (columnConfig.blockid) {
		this.setItem(columnConfig.blockid);
	} else {
		this.setItem(null);
	}
	this._element.removeClass();
	this._element.addClass("dshFlexItem");
	if (this._config.className) {
		this._element.addClass(this._config.className);
	}
	this.applyColumnStyles();
};

/**
 * Set the item for this column
 * @param {String}  id           - ID of the item
 * @param {Promise.<Block>} blockPromise - promise for the item
 */
FlexItem.prototype.setItem = function(id) {
	if (this._blockID !== id) {
		// detach current item
		if (this._block) {
			var oldBlock = this._block;
			this._block = null;
			oldBlock.setInDOM(false);
			this._layout.detachChild(oldBlock);
		}

		if (!id) {
			this._blockID = null;
			this._blockPromise = null;
			this._block = null;
			// this is an empty column, fill with placeholder
			this._layout._initializePlaceholder(this);
		} else {
			this._layout._context.getDragDropHandler().then(
				function(dragDropHandler) {
					//remove the droppable functionality when a placeholder is replaced
					if (this._isPlaceholder) {
						dragDropHandler.removeDropzone(this.getElement());
						this._isPlaceholder = null;
					}
				}.bind(this)
			);

			this._blockID = id;

			var whenFutureBlockAvailable = this._layout._registerFutureChild(this._blockID);

			if (this._config.lockedColumn) {
				this._blockPromise = this._layout.getContext().loadBlock(this._blockID);
			} else {
				this._blockPromise = this._layout.loadBlockWrapper(this._blockID);
			}

			this._blockPromise.then(
				function(blockWrapper) {
					if (this._blockID === id) {
						// we still want this block to be in the column, do the replace
						this._replaceElement(blockWrapper.getElement());

						this._layout._registerChild(blockWrapper);
						blockWrapper.setInDOM(this._layout.isInDOM());
						this._block = blockWrapper;
					}

					whenFutureBlockAvailable();

					if (this._config.showVerticalScrollbar) {
						blockWrapper.getElement().css("height", "inherit");
						blockWrapper.getBlockPromise().then(function(block) {
							block.getElement().css("overflow-y", "auto");
						});
					}
				}.bind(this)
			);
		}
	}
};

/**
 * Get the promise of the block that is being displayed in the column
 * @returns {Promise.<Block>}
 */
FlexItem.prototype.getBlockPromise = function() {
	return this._blockPromise;
};

/**
 * Replaces the current element with a paceholder element.
 * @param {JQueryObject} ghost - the placeholder element
 */
FlexItem.prototype.makePlaceholder = function(ghost) {
	this._replaceElement(ghost);
	this._isPlaceholder = true;
};

/**
 * Replace the current element with another element
 * @private
 * @param {JQueryObject} newElement - element
 */
FlexItem.prototype._replaceElement = function(newElement) {
	this._wrapper.children().detach();
	this._wrapper.append(newElement);

	this._contentElement = newElement;
	this.applyColumnStyles();
};

/**
 * Get the configuration of the Column
 * @returns {Object} configuration
 */
FlexItem.prototype.getConfiguration = function() {
	var config = _.pick(
		this._config,
		"fitContent",
		"grow",
		"shrink",
		"width",
		"height",
		"maxWidth",
		"minWidth",
		"lockedColumn",
		"minWidthIsContent",
		"className"
	);
	if (this._blockID) {
		config.blockid = this._blockID;
	}
	return config;
};

/**
 * Get the element of the Column
 *
 * @returns {JQueryObj}
 */
FlexItem.prototype.getElement = function() {
	return this._element;
};

/**
 * Get the ID of the block in the column
 * @returns {String}
 */
FlexItem.prototype.getBlockID = function() {
	return this._blockID;
};

/**
 * Is there any with defined?
 * @returns {Boolean}
 */
FlexItem.prototype.hasWidthDefined = function() {
	return (
		(this._config.width || this._config.minWidth || this._config.maxWidth || this._config.fitContent || this._config.minWidthIsContent) &&
		true
	);
};

/**
 * Apply column styles
 */
FlexItem.prototype.applyColumnStyles = function() {
	this._element.attr("data-column", this._columnID);

	var elemclasswidthdecided = true;
	var elemclassfittocontent = false;

	var flexWidth = this._config.width || this._config.minWidth || this._config.maxWidth;
	if (this._config.fitContent) {
		flexWidth = "auto";
	}
	if (!flexWidth) {
		flexWidth = Math.floor(100 / this._totalColumns) + "%";
	}

	var cssStyles = {
		"-webkit-flex": this._config.grow + " " + this._config.shrink + " " + flexWidth,
		"-ms-flex": this._config.grow + " " + this._config.shrink + " " + flexWidth,
		flex: this._config.grow + " " + this._config.shrink + " " + flexWidth
	};

	var isEmpty = !this._blockID && this._layout._context.isEditModeActive();

	//indicate placeholder if the column is empty
	if (isEmpty) {
		if (!this._contentElement) {
			this._contentElement = $("<div>", {
				css: {
					width: "100px"
				}
			});
			this._wrapper.append(this._contentElement);
		}
		this._contentElement.css({
			border: "1px dotted #ccc",
			padding: "10px",
			margin: "5px"
		});
	}

	if (this._config.width) {
		cssStyles.width = this._config.width;
	} else if (this._config.grow > 0 && this._config.shrink > 0 && !this._config.fitContent) {
		// IE fix ...
		cssStyles.width = "100%";
		elemclasswidthdecided = false;
	} else if (this._config.fitContent) {
		cssStyles.width = "auto";
		elemclassfittocontent = true;
		elemclasswidthdecided = false;
	}

	if (this._config.height) {
		cssStyles.height = this._config.height;
	}

	if (this._config.maxWidth) {
		cssStyles["max-width"] = this._config.maxWidth;
	}

	if (this._config.minWidth) {
		cssStyles["min-width"] = this._config.minWidth;
	} else if (!this._config.minWidthIsContent) {
		cssStyles["min-width"] = 0;
	} else {
		cssStyles["min-width"] = "";
	}

	if (this._config.showVerticalScrollbar) {
		cssStyles["height"] = "inherit";
	}

	this._element.css(cssStyles);

	var margin = this._layout._spacingParameter.getValue();
	this._wrapper.css({
		"padding-left": margin,
		"padding-bottom": margin,
		height: "inherit"
	});

	// Tell the items in this column that the column width has been decided
	// They need to adapt to fit
	// Mainly used for the design panel & the results table
	this._element.toggleClass("columnwidthdecided", elemclasswidthdecided);
	this._element.toggleClass("columnwidthfittocontent", elemclassfittocontent);
};

/**
 * Get the block which is assigned to the column
 * @returns {Block}
 */
FlexItem.prototype.getBlock = function() {
	return this._block;
};

/**
 * get the layout to which this flex items belongs
 * @return {FlexLayout}
 */
FlexItem.prototype.getLayout = function() {
	return this._layout;
};

export default FlexItem;
