import ParameterConfigurationPanel from "../../builder/configuration/ParameterConfigurationPanel";
import FlexConfigItem from "./FlexConfigItem";
import $ from "jquery";
import _ from "underscore";
import jqueryUI from "jqueryUI";
import webcss from "./FlexLayoutConfigurationPanel.css";
/**
 * FlexLayoutConfigurationPanel
 * @augments ParameterConfigurationPanel
 * @param {FlexLayout}     flexLayout - flex layout
 * @param {LanguageBundle} i18n       - language bundle
 */
var FlexLayoutConfigurationPanel = function(parameter, i18n) {
	this._i18n = i18n;

	/**
	 * @member {Array.<FlexConfigItem>} list of flexConfigItems
	 * @protected
	 */
	this._flexConfigItems = [];

	ParameterConfigurationPanel.call(this, parameter);
};

FlexLayoutConfigurationPanel.prototype = Object.create(ParameterConfigurationPanel.prototype);
FlexLayoutConfigurationPanel.prototype.constructor = FlexLayoutConfigurationPanel;

/**
 * Initialize the DOM and interactionof the panel
 * @private
 */
FlexLayoutConfigurationPanel.prototype._init = function() {
	this._flexConfigItems = [];

	this._stageWrapper = $("<div>").addClass("flexStageWrapper");
	this._stage = $("<div>").addClass("flexConfigurationStage");

	_.each(
		this._parameter.getItemList(),
		function(flexItem) {
			this._addNewConfigItem(flexItem.getConfiguration());
		},
		this
	);

	this._addItemButton = $("<a>", {
		class: "btn btn-no-border"
	})
		.append(
			$("<i>", {
				class: "icon-large icon icon-WCR_Add12"
			})
		)
		.on(
			"click",
			function() {
				var flexConfigItem = this._addNewConfigItem();

				this._stage.scrollLeft(this._stage.scrollLeft() + flexConfigItem.getElement().position().left - this._stage.position().left);

				this._updateParameter();
			}.bind(this)
		);

	this._stage.sortable({
		axis: "x",
		items: ".flexConfigItem",
		handle: "h5"
	});

	this._stage.on(
		"sortstart",
		function(event, ui) {
			this._sortStartPos = ui.item.index();
		}.bind(this)
	);
	this._stage.on(
		"sortstop",
		function(event, ui) {
			var _sortStopPos = ui.item.index();
			if (_sortStopPos === this._sortStartPos) {
				return;
			}
			var draggedItem = this._flexConfigItems[this._sortStartPos];

			var head;
			var tail;
			var movingColumns;
			if (this._sortStartPos < _sortStopPos) {
				//dragged to right
				head = this._flexConfigItems.slice(0, this._sortStartPos);
				tail = this._flexConfigItems.slice(_sortStopPos + 1);
				movingColumns = this._flexConfigItems.slice(this._sortStartPos + 1, _sortStopPos + 1);

				this._flexConfigItems = head
					.concat(movingColumns)
					.concat([draggedItem])
					.concat(tail);
			} else {
				// dragged to left
				head = this._flexConfigItems.slice(0, _sortStopPos);
				tail = this._flexConfigItems.slice(this._sortStartPos + 1);
				movingColumns = this._flexConfigItems.slice(_sortStopPos, this._sortStartPos);
				this._flexConfigItems = head
					.concat([draggedItem])
					.concat(movingColumns)
					.concat(tail);
			}

			this._updateParameter();
		}.bind(this)
	);

	this._stageWrapper.append(this._stage).append(this._addItemButton);
	this._element.append(this._stageWrapper);

	this._observeParam(
		function() {
			this._update();
		}.bind(this)
	);
};

/**
 * Update the parameter with the current configuration of the panel
 * @private
 */
FlexLayoutConfigurationPanel.prototype._updateParameter = function() {
	this._parameter.undoableSetValue(this._getCustomConfiguration());
};

/**
 * Update the UI
 * @private
 */
FlexLayoutConfigurationPanel.prototype._update = function() {
	var flexItems = this._parameter.getItemList();
	// if there are columns removed, remove them from the DOM
	if (this._flexConfigItems.length > flexItems.length) {
		for (var i = this._flexConfigItems.length - 1; i > flexItems.length - 1; i--) {
			this._flexConfigItems[i].getElement().remove();
			this._flexConfigItems.splice(i, 1);
		}
	}
	// if there are columns added, add them to the DOM
	if (this._flexConfigItems.length < flexItems.length) {
		for (var i = flexItems.length - 1; i > this._flexConfigItems.length - 1; i--) {
			this._flexConfigItems.push(null);
		}
	}

	// update DOM elements with the new configuration
	for (var i = 0; i < flexItems.length; i++) {
		var flexItem = flexItems[i];
		var flexConfigItem = this._flexConfigItems[i];
		if (!flexConfigItem) {
			this._addNewConfigItem(flexItem.getConfiguration(), i);
		} else {
			flexConfigItem.update(flexItem.getConfiguration());
		}
	}
};

/**
 * Get the new configuration for the FlexLayout
 * @private
 * @returns {object}
 */
FlexLayoutConfigurationPanel.prototype._getCustomConfiguration = function() {
	return {
		columns: _.map($(".flexConfigItem", this._stage), function(item) {
			return $(item)
				.data("flexConfigItem")
				.getConfiguration();
		}),
		flexDirection: this._parameter.getValue().flexDirection
	};
};

/**
 * Creates a new config item, adds it to the dom and adds it to _flexConfigItems
 * @private
 * @param {Object} configuration - initial configuration
 * @param {Integer} index - the index in _flexConfigItems where the new configItem have to be added (optional).
 *                        if not defined the new item is added at the end
 * @returns {FlexConfigItem} the new item
 */
FlexLayoutConfigurationPanel.prototype._addNewConfigItem = function(configuration, index) {
	var flexConfigItem = new FlexConfigItem(configuration || {}, this._i18n, this._parameter);

	//add observers
	flexConfigItem.observe(this._updateParameter.bind(this));
	flexConfigItem.on(
		"deleted",
		function() {
			var index = this._flexConfigItems.indexOf(flexConfigItem);
			this._flexConfigItems.splice(index, 1);
			this.updateColumnRemoval();
		}.bind(this)
	);
	//add to DOM
	this._stage.append(flexConfigItem.getElement());

	if (index) {
		this._flexConfigItems[index] = flexConfigItem;
	} else {
		this._flexConfigItems.push(flexConfigItem);
	}

	this.updateColumnRemoval();
	return flexConfigItem;
};

/**
 * Update the column removal property of all flex config items
 */
FlexLayoutConfigurationPanel.prototype.updateColumnRemoval = function() {
	var allowColumnRemoval = this._flexConfigItems.length > 1;
	_.each(this._flexConfigItems, function(configItem) {
		configItem.allowColumnRemoval(allowColumnRemoval);
	});
};
export default FlexLayoutConfigurationPanel;
