import $ from "jquery";
import Observable from "@esko/observable";
import _ from "underscore";
/**
 * A simple UI component to show an inputfield with units
 * @param {String}         label           - label for the input field
 * @param {String}         value           - value of the input field
 * @param {Array.<String>} units           - available units
 * @param {boolean}        usedAsParameter - defines if the input group is used as form element or not (different kind of rendering)
 * @param {Object} 		   options         - attributes added to the input field
 * @param {JQueryObject}   helptextElement - JQueryObject that contains a helptext
 */
var UnitInputField = function(label, value, units, usedAsParameter, options, helptextElement) {
	Observable.mixin(this, "UnitInputField");

	this._units = units;

	this._previousValue = value;

	this._options = options || {};

	if (usedAsParameter) {
		this.element = $("<div>", {
			class: "form-group"
		});
	} else {
		this.element = $("<div>");
	}

	this.element.append(
		$("<label>", {
			text: label
		})
	);
	var inputgroup = $("<div>", {
		class: "input-group"
	});

	if (usedAsParameter) {
		this.element.append(
			$("<div>", {
				class: "form-entry"
			})
				.append(inputgroup)
				.append(helptextElement)
		);
	} else {
		this.element.append(inputgroup);
	}

	this.input = $("<input>", {
		class: "form-control",
		attr: this._options
	});
	this.input.on(
		"input",
		function() {
			var intValue = parseInt(this.input.val(), 10);
			if (this.input.val() && isNaN(intValue)) {
				//invalid input (e.g. letters)
				this.input.val(0);
			} else if (options && options.min != null && intValue < options.min) {
				this.input.val(options.min);
			} else if (options && options.max != null && intValue > options.max) {
				this.input.val(options.max);
			} else {
				if (!isNaN(intValue)) {
					this.input.val(intValue);
				} //else input is empty and can remain like this (waiting for user input)
			}
		}.bind(this)
	);

	this.input.on(
		"change blur",
		function() {
			// Used both change &  blur , because change event is not triggered when value is set by JQuery
			var currentValue = this.getValue();
			if (currentValue !== this._previousValue) {
				this._previousValue = currentValue;
				this._emit("changed");
			}
		}.bind(this)
	);

	inputgroup.append(this.input);

	this.unitSelector = $("<select>");
	_.each(
		units,
		function(unit) {
			this.unitSelector.append(
				$("<option>", {
					value: unit,
					text: unit
				})
			);
		},
		this
	);
	this.unitSelector.on(
		"change",
		function() {
			this._emit("changed");
		}.bind(this)
	);

	inputgroup.append(
		$("<div>", {
			class: "input-group-addon-select"
		}).append(this.unitSelector)
	);

	this.setValue(value);
};

/**
 * Set the value of the input field
 * @param   {String}   value - including the unit!
 */
UnitInputField.prototype.setValue = function(value) {
	this._previousValue = value;

	var number = this.parseValue(value) || "";

	var unit = _.find(this._units, function(unit) {
		return (value + "").indexOf(unit) >= 0;
	});
	if (!unit) {
		unit = _.first(this._units);
	}

	this.input.val(number);
	this.unitSelector.val(unit);
};

/**
 * UnitInput field value contains "px" ,"%" , so parse it and get the number only
 *
 * @param value
 * @returns {String}
 */
UnitInputField.prototype.parseValue = function(value) {
	if (!value || (value + "").trim() === "") {
		return "";
	} else {
		var number = parseFloat(value);
		if (isNaN(number)) {
			number = "";
		}
		return number + "";
	}
};

/**
 * Get the currently selected unit
 * @returns {String}
 */
UnitInputField.prototype.getUnit = function() {
	return this.unitSelector.val();
};

/**
 * Get the full value, this includes the unit.
 * An empty string is returned if nothing if filled in
 * @returns {string}
 */
UnitInputField.prototype.getValue = function() {
	if (!this.input.val().trim()) {
		return "";
	} else {
		return this.input.val().trim() + this.getUnit();
	}
};

export default UnitInputField;
