import css from "./dropdownmenu-style.css";
import Popover from "@esko/popover/Popover";
import Button from "@esko/button/Button";
import Observable from "@esko/observable";
import $ from "jquery";
import _ from "underscore";
/**
 * @class Dropdown menu with either actions or options.
 *
 * @constructor
 * @exports DropdownMenu
 * @mixes Observable
 *
 * @param {JQueryObj} container - the container where the dropdown menu is added to (in DOM). It will also be used as restricting element.
 * @param {Boolean} isMobile - True if it is for mobile, false otherwise
 * @param {String} popoverPosition - place of the dropdown menu, One of "top" | "right" | "bottom" | "left" | "innerLeft".
 * @param {String} popoverClass - name of the class of the popover
 * @param {Number} [minHeight] - the minimum height of the dropdown menu
 * @param {Array.<String>} [restrictedPositions] - array with all the positions that are allowed
 * @param {JQueryObj} [restrictedContainer] - Restricts popover from expanding outside this container
 */
var DropdownMenu = function(i18n, container, isMobile, popoverPosition, popoverClass, minHeight, restrictedPositions, restrictedContainer) {
	Observable.mixin(this, "DropdownMenu");

	/**
	 * the container where the dropdown menu is added to (in DOM)
	 */
	this._container = container;

	/**
	 * Restricts popover from expanding outside this container
	 */
	this._restrictedContainer = restrictedContainer;
	/**
	 * Is this for a mobile device or not
	 */
	this._isMobile = isMobile;

	/**
	 * Is the menu opened or not?
	 */
	this.active = false;
	/**
	 * Is the menu attached to an element?
	 */
	this._attachedTo = undefined;

	/**
	 * array witht the positions to restrict the dropdown menu to
	 */
	this._restrictedPositions = restrictedPositions;

	/**
	 * items shown in the menu
	 */
	this.items = {};
	/**
	 * Order of the items
	 */
	this.itemStack = [];
	/**
	 * counter for giving the dividers an unique name
	 */
	this.dividerNum = 0;

	/**
	 * counter for giving the headers an unique name
	 */
	this.headerNum = 0;

	/**
	 * counter for giving the content an unique name
	 */
	this.contentNum = 0;

	/**
	 * <li> elements of the list, to make it possible to hide/show items in the list
	 */
	this._listElements = {};

	/**
	 * popover to show the dropdown menu
	 */
	this.menu = undefined;

	this.autoHideTimeout = 750;

	/**
	 * @member {Boolean} if true will no auto hide the popup on mouse leave
	 * @private
	 */
	this._disableAutoHide = false;

	this._minHeight = minHeight;
	this._popoverPosition = popoverPosition || "bottom";
	this._popoverClass = popoverClass;
};

/**
 * Initialize the dropdown menu (create needed dom elements)
 *
 * @private
 */
DropdownMenu.prototype._initialize = function() {
	this.menu = $("<ul/>").addClass("wcr-dropdown-menu");
	this.popover = new Popover(
		this._container,
		this._container,
		this._restrictedContainer || this._container,
		this._popoverPosition,
		"#ffffff",
		this._popoverClass,
		0,
		this._minHeight,
		true
	);
	if (this._restrictedPositions) {
		this.popover.setPlacementOptions(this._restrictedPositions);
	}

	this.popover.setContent(this.menu);

	_.each(
		this.itemStack,
		function(item) {
			this._addItemToDOM(item);
			if (!item.isVisible) {
				this.hideElement(item.name);
			}
		},
		this
	);

	this._emit("initialized");

	this._isInitialized = true;
};

/**
 * Attach the dropdown menu to a certain element,
 * if the user clicks on the element, the dropdown is shown relative to that element
 * @param {JQueryObj} element - the element to which the dropdown is attached
 */
DropdownMenu.prototype.attachTo = function(element) {
	var eventType = "click";
	if (this._isMobile && window.PointerEvent) {
		eventType = "pointerup";
	} else if (this._isMobile) {
		eventType = "touchend";
	}
	if (this.attachedEvent) {
		this._attachedTo.off(eventType, this.attachedEvent);
	}

	this._attachedTo = element;
	this.attachedEvent = _.bind(function() {
		if (this.isActive()) {
			this.hide();
		} else {
			this.show();
		}
	}, this);

	this._attachedTo.on(eventType, this.attachedEvent);
};

/**
 * Open the dropdown menu, if the menu is not attached to an element, an element can be passed
 * @param element
 */
DropdownMenu.prototype.show = function(element) {
	if (element === undefined) {
		element = this._attachedTo;
	}

	if (!this._isInitialized) {
		this._initialize();
	}

	this.popover.setContext(element);
	this.popover.show();
	if (!this._disableAutoHide) {
		this.startAutoHide(element);
	}
	this.active = true;
};

DropdownMenu.prototype.hide = function() {
	if (this.popover) {
		this.popover.hide();
	}
	this.active = false;
};

DropdownMenu.prototype.toggle = function(toggle) {
	if (toggle === undefined) {
		toggle = !this.active;
	}

	if (toggle) {
		this.show();
	} else {
		this.hide();
	}
};

/**
 * Is the dropdown menu openend?
 * @returns {Boolean}
 */
DropdownMenu.prototype.isActive = function() {
	return this.active;
};

/**
 * Add an action, this is a list item with a text/icon and an action bound to it
 *
 * @param {Object} action - description of the action
 * @example
 * {
 * 	name: 'nameOfButton',
 * 	content: Text|JQueryObj,
 * 	icon: undefined|''|'nameOfIcon'|JQueryObj,
 * 	action: function{},
 *  enabled: Boolean,
 *  [cssClass: String]
 * 	[optional: doNotHideOnClick]
 * }
 */
DropdownMenu.prototype.addAction = function(action) {
	action.type = "action";
	if (action.enabled === undefined) {
		action.enabled = true;
	}
	this._addItem(action, action.after);
};

/**
 * Remove an action from the dropdown menu.
 * @param {String} name - name (id) of the item to be removed
 */
DropdownMenu.prototype.removeAction = function(name) {
	if (this._listElements[name]) {
		//are the items allready added to the
		this._listElements[name].remove();
		delete this._listElements[name];
	} else {
		var index = _.indexOf(this.itemStack, this.items[name]);
		if (index >= 0) {
			this.itemStack.splice(index, 1);
		}
	}
};

/**
 * Add a group to the dropdown, has the same data as addAction, but adds a div element to add action elements to the group
 * @param {Object} group
 */
DropdownMenu.prototype.addGroup = function(group) {
	group.type = "group";
	if (group.enabled === undefined) {
		group.enabled = true;
	}
	this._addItem(group);
};

/**
 * Enable/disable an action
 *
 * if an action is not defined, nothing happens
 * @param {String}  name    - name of the action to enable
 * @param {Boolean} enabled - enable/disable
 */
DropdownMenu.prototype.setActionEnabled = function(name, enabled) {
	if (this.items[name]) {
		this.items[name].enabled = enabled;
		if (this._isInitialized) {
			this.items[name].btn.setEnabled(enabled);
		}
	}
};

DropdownMenu.prototype.isActionEnabled = function(name) {
	return this.items[name] && this.items[name].enabled;
};

/**
 * Add a divider in the menu
 */
DropdownMenu.prototype.addDivider = function() {
	var divider = {
		name: "divider" + this.dividerNum++,
		type: "divider"
	};
	this._addItem(divider);
	return divider.name;
};

/**
 * Add a header in the menu
 */
DropdownMenu.prototype.addHeader = function(headerText) {
	var header = {
		name: "header" + this.headerNum++,
		type: "header",
		text: headerText
	};
	this._addItem(header);
	return header.name;
};

/**
 * Add some generic content in the menu
 *
 * @param element - jQuery element.
 */
DropdownMenu.prototype.addContent = function(element) {
	var content = {
		name: "content" + this.contentNum++,
		type: "content",
		content: element
	};
	this._addItem(content);
	return content.name;
};

/**
 * Add a checkable option
 * @param {Object} option
 * @example
 * {
 * 	name,
 * 	content,
 *  enabled,
 * 	selected,
 * 	onchange,
 * 	[doNotHideOnClick]
 *}
 *
 */
//
DropdownMenu.prototype.addOption = function(option) {
	option.type = "option";
	if (option.selected === undefined) {
		option.selected = false;
	}
	if (option.enabled === undefined) {
		option.enabled = true;
	}
	option.toggle = function(toggle) {
		if (toggle === undefined) {
			option.selected = !option.selected;
		} else {
			option.selected = toggle > 1 || toggle;
		}

		// update icon if needed
		if (option.iconElement) {
			option.iconElement.toggleClass("icon-WCR_Checked", option.selected).toggleClass("unchecked", !option.selected);
		}
	};

	// when the option is clicked, change the selected state and toggle the icon
	option.action = function() {
		if (!(option.userCannotUncheck && option.selected)) {
			option.toggle();
			option.onchange(option.selected);
		}
	};
	this._addItem(option);
};

DropdownMenu.prototype.toggleOption = function(name, toggle) {
	if (this.items[name]) {
		this.items[name].toggle(toggle);
	}
};

DropdownMenu.prototype.getOptionValue = function(name) {
	if (this.items[name]) {
		return this.items[name].selected;
	}
};

/**
 * Add an item to the itemstack
 * @private
 * @param {Object} item
 * @param {String} after - name of the item after which it needs to be added
 */
DropdownMenu.prototype._addItem = function(item, after) {
	this.items[item.name] = item;
	if (after) {
		var index = _.indexOf(this.itemStack, this.items[after]);
		if (index >= 0) {
			this.itemStack.splice(index + 1, 0, item);
		} else {
			this.itemStack.unshift(item);
		}
	} else {
		this.itemStack.push(item);
	}

	if (item.isVisible === undefined) {
		item.isVisible = true;
	}

	if (this._isInitialized && item.isVisible) {
		this._addItemToDOM(item, after);
	}
};

/**
 * Add an item to the DOM
 * @param {Object} item
 * @param {Boolean} prepend
 */
DropdownMenu.prototype._addItemToDOM = function(item, after) {
	if (item.type == "action" || item.type == "option" || item.type == "group") {
		if (item.type == "option") {
			// if we have an option action, use the checkbox icon
			item.icon = item.selected ? "icon-WCR_Checked" : "unchecked";
		}

		// create the button
		item.btn = new Button(
			this._i18n,
			this._container,
			this._isMobile,
			"btn_" + item.name,
			item.content,
			this._isMobile,
			_.partial(function(that, e) {
				item.action.call(this, e);

				if (!item.doNotHideOnClick) {
					that.hide();
				}
			}, this)
		);

		if (item.cssClass) {
			item.btn.element.addClass(item.cssClass);
		}

		item.btn.setEnabled(item.enabled);
		var listElement = $("<li/>")
			.addClass(item.name)
			.addClass("action")
			.append(item.btn.element)
			.attr("data-name", item.name);
		if (item.type == "group") {
			var groupElementWrapper = $("<li/>")
				.addClass("group")
				.addClass(item.name);
			var groupElement = $("<ul/>");
			listElement = groupElementWrapper.append(groupElement.append(listElement));

			this._listElements[item.name] = groupElement;
		} else {
			this._listElements[item.name] = listElement;
		}
		if (item.group !== undefined) {
			this._listElements[item.group].append(listElement);
		} else if (after) {
			var appendAfter = this._listElements[after];
			if (appendAfter) {
				appendAfter.after(listElement);
			} else {
				this.menu.prepend(listElement);
			}
		} else {
			this.menu.append(listElement);
		}

		this._prependIcon(item);

		if (!item.isVisible) {
			//make sure to hide the item if it shouldn't be visible
			this._listElements[item.name].hide();
		}
	} else if (item.type == "divider") {
		var listElement = $("<li/>").addClass("divider");
		this._listElements[item.name] = listElement;
		this.menu.append(listElement);
		if (!item.isVisible) {
			listElement.hide();
		}
	} else if (item.type == "header") {
		var listElement = $("<h3/>")
			.addClass("dropdownmenu-header")
			.text(item.text);
		this._listElements[item.name] = listElement;
		this.menu.append(listElement);
		if (!item.isVisible) {
			listElement.hide();
		}
	} else if (item.type == "content") {
		this._listElements[item.name] = item.content;
		this.menu.append(item.content);
		if (!item.isVisible) {
			listElement.hide();
		}
	} else {
		this.f.emit("error.SEVERE", "DropdownMenu shouldn't come into this state");
	}
};

/**
 * Clear the contents of the dropdown
 */
DropdownMenu.prototype.clear = function() {
	if (this.menu) {
		this.menu.empty();
	}
	this.itemStack = [];
	this.items = {};
	this._listElements = {};
};

/**
 * leaivng the dropdown will no longer hide the dropdownmenu
 */
DropdownMenu.prototype.disableAutoHide = function() {
	this._disableAutoHide = true;
};

DropdownMenu.prototype.startAutoHide = function(element) {
	if (!this._isMobile) {
		var timer;
		var onLeave;

		var onEnter = function() {
			clearTimeout(timer);
		};

		onLeave = _.bind(function() {
			if (timer) {
				clearTimeout(timer);
			}
			timer = setTimeout(
				_.bind(function() {
					this.hide();
					// unregister the events on both the element to which the menu is attached to and the popover
					element.add(this.popover.element).off({
						mouseenter: onEnter,
						mouseleave: onLeave
					});
				}, this),
				this.autoHideTimeout
			);
		}, this);

		// register the events on both the element to which the menu is attached to and the popover
		element.add(this.popover.element).on({
			mouseenter: onEnter,
			mouseleave: onLeave
		});
	} else {
		if (!this._oneClickAutoHide) {
			this._oneClickAutoHide = _.bind(function(e) {
				if (
					this.popover.element.has(e.target).length === 0 &&
					e.target != this.popover.element[0] &&
					(!element || (element.has(e.target).length === 0 && element[0] !== e.target))
				) {
					this.hide();

					var eventType = "click touchend";
					if (window.PointerEvent) {
						eventType = "pointerup";
					}
					$(document.body).off(eventType, this._oneClickAutoHide);
				}
			}, this);
		}

		var eventType = "click touchend";
		if (window.PointerEvent) {
			eventType = "pointerup";
		}
		$(document.body).on(eventType, this._oneClickAutoHide);
	}
};

/**
 * Sets the action text
 * @param {String} name
 * @param {String} text
 */
DropdownMenu.prototype.setActionText = function(name, text) {
	var item = this.items[name];
	item.content = text;
	if (this._isInitialized) {
		item.btn.setContent(text);
		this._prependIcon(item);
	}
};

/**
 * Hides the element in the dropdown list
 * @param {String} name - identifier of the element
 */
DropdownMenu.prototype.hideElement = function(name) {
	if (this.items[name] !== undefined) {
		if (this._listElements[name] !== undefined) {
			this._listElements[name].isVisible = false;
			this._listElements[name].hide();
		}
		this.items[name].isVisible = false;
	}
};

/**
 * show the element in the dropdown list
 * @param {String} name - identifier of the element
 */
DropdownMenu.prototype.showElement = function(name) {
	if (this.items[name] !== undefined) {
		this.items[name].isVisible = true;
		if (this._listElements[name] !== undefined) {
			this._listElements[name].show();
			this._listElements[name].isVisible = true;
		} else if (this._isInitialized) {
			this._addItemToDOM(this.items[name]);
		}
	}
};

/**
 * checks if the element is visible in the list
 * @param {String} name - identifier of the element
 * @returns {Boolean}
 */
DropdownMenu.prototype.isElementVisible = function(name) {
	if (this.items[name] !== undefined) {
		return this.items[name].isVisible;
	} else {
		return undefined;
	}
};

/**
 * checks if icon should be prepended
 * @param {Object} item - item to which icon should be prepended
 */
DropdownMenu.prototype._prependIcon = function(item) {
	if (item.icon !== undefined && item.icon !== "") {
		if (_.isString(item.icon)) {
			// a css class
			item.iconElement = $("<i/>")
				.addClass("icon")
				.addClass(item.icon);
			item.btn.element.prepend(item.iconElement);
		} else {
			// a jquery elem
			item.btn.element.prepend(item.icon);
		}
	}
};

export default DropdownMenu;
