import _ from "underscore";
/**
 * This module provides internationalization. Can be subclassed in order to add extra custom bundles or custom logic.
 *
 * @param {String} locale requested language
 * @param {String} localStorageKey Key that is used to cache the language bundles in HTML5 local storage
 * @param {Number} timestamp [optional] this is used a cache busting parameter for requesting language strings
 */
var I18n = function(locale, localStorageKey, timestamp) {
	this._locale = locale ? locale : "en";
	this._timestamp = timestamp ? timestamp : 0;
	this._localStorageKey = localStorageKey + "-" + locale;

	this._isLocalStorageLoaded = false;

	/**
	 * @member {Object} all loaded bundles
	 */
	this.bundles = {};
	/**
	 * @member {Array.<String>} list of bundles currently being fetched
	 */
	this._bundlesBeingFetched = [];
	/**
	 * @member {Array.<String>} list of bundles that should be fetched in the next fetching batch
	 */
	this._fetchQueue = [];

	/**
	 * @member {Object} Maps bundle names to BundleResolver in order to resolve the path of the bundle.
	 */
	this._bundleToResolver = {};

	/**
	 * @member {BundleResolver} This BundleResolver is used when no BundleResolver is given for a requested bundle.
	 */
	this._fallbackBundleResolver = null;

	this.bundleOrder = [];
	this.bundleLoadedListeners = {};

	this._throttledFetch = _.throttle(_.bind(this._fetchBundles, this), 50, {
		leading: false
	});
};

/**
 * Get the locale of the module this object is used for
 * @returns {String}
 */
I18n.prototype.getLocale = function() {
	return this._locale;
};

/**
 * Load one or more bundles
 *
 * If fallback is enabled, the fallback will be loaded automatically!
 *
 * @param {String} bundleName - the name of the bundle
 * @param {Function} [callback] - is called when the bundle is loaded
 * @param {Object} [binding] - bind the callback to "binding"
 */
I18n.prototype.loadBundle = function(bundleName, callback, binding) {
	if (callback && binding) {
		callback = _.bind(callback, binding);
	}

	if (_.isArray(bundleName)) {
		// multiple bundles
		// make new callback, which only is executed after bundName.length-1 times
		var multiCallback = _.after(bundleName.length, _.once(callback));
		_.each(
			bundleName,
			_.bind(function(bundle) {
				this.loadSingleBundle(bundle, multiCallback);
			}, this)
		);
	} else {
		this.loadSingleBundle(bundleName, callback);
	}
};

/**
 * Load a bundle
 *
 * WARNING about newlines: "\n" in the translation file is interpreted as a JavaScript newline character!! (see I18n:loadSingleBundle)
 *
 * If fallback is enabled, the fallback will be loaded automatically!
 *
 * @param {String} bundleName - the name of the bundle
 * @param {Function} [callback] - is called when the bundle is loaded
 * @param {Function} [callback] - is called when the bundle failed to load
 */
I18n.prototype.loadSingleBundle = function(bundleName, callback, errorcallback) {
	this._readLocalStorageIfNeeded();

	if (this.bundles[bundleName]) {
		// make sure to update the bundle order, so that the findByKey works correctly
		if (!_.contains(this.bundleOrder, bundleName)) {
			this.bundleOrder.push(bundleName);
		}
		if (callback) {
			callback(this.bundles[bundleName]);
		}
	} else {
		// bundle info not yet available, register callback
		if (callback) {
			this.whenBundleLoaded(bundleName, callback);
		}

		if (!_.contains(this._bundlesBeingFetched, bundleName)) {
			this._fetchQueue.push({
				bundle: bundleName,
				errorcallback: errorcallback
			});
			this._bundlesBeingFetched.push(bundleName);
			this.bundleOrder.push(bundleName);

			this._throttledFetch();
		}
	}
};

/**
 * Execute a batched fetch operation
 * @private
 */
I18n.prototype._fetchBundles = function() {
	if (this._fetchQueue.length > 0) {
		// only fetch if there is need to fetch
		var requestedBundles = _.pluck(this._fetchQueue, "bundle");
		this._fetchQueue = [];

		_.each(
			requestedBundles,
			function(bundleName) {
				var bundleResolver = this._bundleToResolver[bundleName] || this._fallbackBundleResolver;
				if (bundleResolver) {
					bundleResolver.fetch(
						bundleName,
						this._locale,
						this._timestamp,
						function(bundle, err) {
							if (err) {
								window.console.error(err);
								this.fireBundleLoaded(bundleName, {});
							} else if (bundle) {
								this.bundles[bundleName] = bundle;
								this.fireBundleLoaded(bundleName, this.bundles[bundleName]);
							}
							this._bundlesBeingFetched = _.without(this._bundlesBeingFetched, bundleName);
							this._saveToLocalStorage();
						}.bind(this)
					);
				} else {
					throw "No BundleResolver found for bundle: " + bundleName;
				}
			}.bind(this)
		);
	}
};

/**
 * Read the data from the local storage
 * (will only be executed once)
 * @private
 */
I18n.prototype._readLocalStorageIfNeeded = function() {
	if (!this._isLocalStorageLoaded) {
		try {
			var localStorageData = JSON.parse(window.localStorage.getItem(this._localStorageKey));

			if (localStorageData && localStorageData.timestamp == this._timestamp) {
				this.bundles = localStorageData.bundles;
			}
		} catch (e) {
			if (window.console) {
				window.console.log("There was a problem while reading from Local Storage", e);
			}
		}

		this._isLocalStorageLoaded = true;
	}
};

/**
 * Extend the local storage information with all the bundles currently loaded
 * @private
 */
I18n.prototype._saveToLocalStorage = function() {
	var localData;
	try {
		localData = JSON.parse(window.localStorage.getItem(this._localStorageKey));
	} catch (e) {
		// safe to ignore, localStorageData might not havebeen set yet
	}

	if (!localData || localData.timestamp !== this._timestamp || !localData.bundles) {
		// no local data is set, or the local data is corrupt or outdated
		localData = {
			timestamp: this._timestamp,
			bundles: {}
		};
	}

	//  copy over the bundles
	_.each(
		this.bundles,
		function(bundle, bundleName) {
			localData.bundles[bundleName] = bundle;
		},
		this
	);

	try {
		// store in local storage
		window.localStorage.setItem(this._localStorageKey, JSON.stringify(localData));
	} catch (e) {
		if (window.console) {
			window.console.log("There was a problem while writing to Local Storage", e);
		}
	}
};

/**
 * Register a callback to be fired when bundle with bundleName is loaded
 * @param {String} bundleName
 * @param {Function} callback
 */
I18n.prototype.whenBundleLoaded = function(bundleName, callback) {
	if (!this.bundleLoadedListeners[bundleName]) {
		this.bundleLoadedListeners[bundleName] = [];
	}
	this.bundleLoadedListeners[bundleName].push(callback);
};

/**
 * Fire all callbacks that were registered
 * @param {String} bundleName
 * @param {Object} bundle
 */
I18n.prototype.fireBundleLoaded = function(bundleName, bundle) {
	if (this.bundleLoadedListeners[bundleName]) {
		_.each(this.bundleLoadedListeners[bundleName], function(callback) {
			callback(bundle);
		});
		delete this.bundleLoadedListeners[bundleName];
	}
};

/**
 * Try to find a value for key, respecting the order in which bundles are loaded
 * @param {String} key
 * @returns {String|undefined}
 */
I18n.prototype.findByKey = function(key) {
	for (var i = this.bundleOrder.length - 1; i >= 0; i--) {
		if (this.bundles[this.bundleOrder[i]] && this.bundles[this.bundleOrder[i]][key]) {
			return this.bundles[this.bundleOrder[i]][key];
		}
	}

	return undefined;
};

/**
 * Try to find a value for a key in a certain bundle.
 */
I18n.prototype.findByKeyInBundle = function(key, bundleName) {
	if (this.bundles[bundleName] && this.bundles[bundleName][key]) {
		return this.bundles[bundleName][key];
	} else {
		return undefined;
	}
};

/**
 * Get the translation of a certain key
 *
 * @param {String} key - key from which you want the translation
 * @param {...String} [replacements] - if extra parameters are given, they replace {0}, {1}, ... in the translated string
 * @returns translation - if no translation is found, the key is returned
 */
I18n.prototype.t = function(key) {
	var value = this.findByKey(key);

	if (value !== undefined) {
		// replace all given options
		for (var i = 1; i < arguments.length; i++) {
			value = value.replace("{" + (i - 1) + "}", arguments[i]);
		}

		return value;
	}

	// we didn't find any value for key, return key
	return key;
};

/**
 * Get the translation of a date based on a key
 */
I18n.prototype.d = function(key, value) {
	var dateFormat = this.findByKey(key);
	if (dateFormat !== undefined) {
		return value.toString(dateFormat);
	} else {
		return undefined;
	}
};

/**
 * Adds a bundle. This bundle will be fetched through the given bundle resolver
 *
 * @param {String} bundleName Name of bundle
 * @param {BundleResolver} bundleResolver Responsible for "fetching" the actual bundle
 */
I18n.prototype.addBundle = function(bundleName, bundleResolver) {
	this._bundleToResolver[bundleName] = bundleResolver;
};

/**
 * This BundleResolver is used when no BundleResolver is given for a requested bundle.
 *
 * @param {BundleResolver} bundleResolver
 */
I18n.prototype.setFallbackBundleResolver = function(bundleResolver) {
	this._fallbackBundleResolver = bundleResolver;
};

export default I18n;
