import $ from "jquery";
import _ from "underscore";
import Map from "@esko/map";
import Observable from "@esko/observable";
import css from "./loadingindicator-style.css";
import loaderIcon from "./loaderTransparent.gif";
/**
 * A loading indicator singleton.
 * This shows a (unique) loading indicator with optionally a description of what is loading (default on the left).
 * The wrapping element can be accessed by .element.
 * The loading indicator is also observable
 * @exports LoadingIndicator
 * @constructor
 * @param {Boolean} [asSpan = false] - If true, the element is constructed as a span instead of a div
 */
var LoadingIndicator = function(asSpan) {
	Observable.mixin(this, "LoadingIndicator");

	/**
	 * @member {Map<String, String>} A map of ongoing loads (key -> description)
	 */
	this._ongoingLoads = new Map();

	/**
	 * @member {JQueryObj} Wrapping element
	 */
	if (asSpan) {
		this.element = $("<span />");
	} else {
		this.element = $("<div />");
	}
	this.element.addClass("wcr-loadingIndicator");

	/**
	 * @member {JQueryObj} Text element
	 */
	this._text = $("<span />").addClass("text");

	/**
	 * @member {String} used for only removing the text when the loading action stopped
	 */
	this._showingTextFor = null;

	/**
	 * @member {Boolean} true when the text label is being shown
	 * @private
	 */
	this._isTextVisible = false;

	this.element.append(this._text);

	/**
	 * @member {JQueryObj} spinner element
	 */
	this._spin = $("<span />").addClass("spin");
	this.element.append(this._spin);

	this.element.hide();

	/**
	 * @member {Boolean} Text on the left of the spinner or not
	 */
	this._textOnTheLeft = null;

	this.setTextPosition(true);

	/**
	 * Debounced update of the loading indicator visibility
	 */
	this.updateLoadingIndicatorDebounced = _.debounce(
		_.bind(function() {
			this._updateLoadingIndicator();
		}, this),
		40
	);
};

/**
 * Set the text on the left or on the right of the spinner.
 * Default is on the left.
 * @param {Boolean} onTheLeft - True if text should be on the left, false to set it on the right.
 */
LoadingIndicator.prototype.setTextPosition = function(onTheLeft) {
	if (this._textOnTheLeft !== onTheLeft) {
		this._textOnTheLeft = onTheLeft;
		if (this._textOnTheLeft) {
			this._text.after(this._spin);
			this._text.removeClass("right").addClass("left");
		} else {
			this._text.before(this._spin);
			this._text.removeClass("left").addClass("right");
		}
	}
};

/**
 * Update the icon (show /hide) + text
 * @param {String} [key] - Optional text to update
 * @param {String} [text] - Optional text to update
 */
LoadingIndicator.prototype._updateLoadingIndicator = function(key, text) {
	if (this._ongoingLoads.size() > 0) {
		if (!this._spinner) {
			this._spinner = $("<img>", {
				src: loaderIcon
			});
			this._spin.append(this._spinner);
		}
		if (text) {
			this._text.text(text);
			if (!this._isTextVisible && !this._showingTextFor) {
				this._text.show();
				this._isTextVisible = true;
			}
			this._showingTextFor = key;
		}
		this.element.show();
	} else {
		this.element.fadeOut();
	}

	if (this._isTextVisible && !this._showingTextFor) {
		this._isTextVisible = false;
		this._text.fadeOut();
	}
};

/**
 * Start a loading indication. If it was already loading, it will update the text with the new description
 * @param {String]} key - What is loading key
 * @param {String} [description] - Description of what is loading
 *
 */
LoadingIndicator.prototype.start = function(key, description) {
	if (description === undefined) {
		description = "";
	}
	this._ongoingLoads.add(key, description);
	this._updateLoadingIndicator(key, description);
	this._emit("changed.ongoingloads");
	return this.element;
};

/**
 * Stop a loading indication
 * @param {String} key - What is stopped loading key
 */
LoadingIndicator.prototype.stop = function(key) {
	this._ongoingLoads.remove(key);
	if (this._showingTextFor === key) {
		this._showingTextFor = null;
	}
	this.updateLoadingIndicatorDebounced();
	this._emit("changed.ongoingloads");
};

/**
 * Has the loading indicator ongoing loads?
 * @returns {Boolean} True if loads are ongoing, false otherwise
 */
LoadingIndicator.prototype.hasOngoingLoads = function() {
	return !this._ongoingLoads.isEmpty();
};

export default LoadingIndicator;
