import _ from "underscore";
import MessageType from "./MessageType";
/**
 * NEVER include this file directly, but use dependency component/message/MessageComponent instead
 *
 * This is the model of a message used by the message model and widget.
 * @exports Message
 * @constructor
 * @param {MessageType | String} type - Type of the message. This can be the message type itself (e.g. MessageType.WARN) or based on the MessageType name property, but case insensitive  (e.g. "warn").
 * @param {String} message - Message to show.
 * @param {Function | String} action - Function executed when clicking on the message or if string URL accessed when clicking on the message.
 * @param {Boolean} autoHide - True if the message should be auto hidden, false otherwise
 * @param {String} [tag] - A label for the message (used for marking messages obsolete)
 * @param {Boolean} [allowClose] - Indicates if the user can remove the message or not (default true)
 */
var Message = function(type, message, action, autoHide, tag, allowClose) {
	if (_.isString(type)) {
		this._type = MessageType.getTypeFromString(type);
	} else {
		this._type = type;
	}
	this._message = message;
	this._action = action;
	this._autoHide = autoHide;
	this._tag = tag;
	this._allowClose = allowClose === undefined ? true : allowClose;
};

/**
 * Get the message type
 * @returns {MessageType}
 */
Message.prototype.getType = function() {
	return this._type;
};

/**
 * Get the message to show
 * @returns {String}
 */
Message.prototype.getMessage = function() {
	return this._message;
};

/**
 * Get the action function executed when clicking on the message or if string URL accessed when clicking on the message
 * @returns {Function | String}
 */
Message.prototype.getAction = function() {
	return this._action;
};

/**
 * Get the autoHide boolean. True if the message should be auto hidden, false otherwise.
 * @returns {Boolean}
 */
Message.prototype.getAutoHide = function() {
	return this._autoHide;
};

/**
 * Get the message tag label for the message (used for marking messages obsolete)
 * @returns {String}
 */
Message.prototype.getTag = function() {
	return this._tag;
};

/**
 * Get if the user is allowed to close the message
 * @returns {boolean}
 */
Message.prototype.allowClose = function() {
	return this._allowClose;
};

export default Message;
