import css from "./messagewidget-style.css";
import MessageType from "./MessageType";
import Message from "./Message";
import $ from "jquery";
import _ from "underscore";
/**
 * NEVER include this file directly, but use dependency component/message/MessageComponent instead
 * MessageWidget is responsible for showing messages to the user.
 * @exports MessageWidget
 * @constructor
 * @param {MessageModel} messageModel - The message model providing the messages
 * @param {JQueryObj} context - The context where this message widget should be added to
 */
var MessageWidget = function(messageModel, context, i18n) {
	/**
	 * @member {Number} Auto Hide delay
	 */
	this._AUTOHIDE_DELAY = 3000; // 3s

	/**
	 * @member {Boolean} Showing message or not
	 */
	this._showing = false;

	/**
	 * @member {JQueryObj} Current div object that is shown
	 */
	this._currentMessage = undefined;

	/**
	 * @member {Message} Current message that is shown
	 */
	this._currentMessageObject = null;

	/**
	 * @member {MessageModel} Model serving the messages
	 */
	this._messageModel = messageModel;

	/**
	 * @member {JQueryObj} Context where this message widget should be attached to
	 */
	this._context = context;
	this._context.addClass("messageWidget");

	this.i18n = i18n;

	// Start listening
	this._messageModel.on("added", _.bind(this._showNextMessage, this));
	this._messageModel.on("obsolete", _.bind(this._markObsolete, this));

	// Start it there are already
	this._showNextMessage();
};

/**
 * Get the css class based on the message type
 * @param {MessageType} messageType - The message type
 */
MessageWidget.prototype._getCSSClass = function(messageType) {
	switch (messageType) {
		case MessageType.ERROR:
			return "alert-danger";
		case MessageType.WARN:
			return "";
		case MessageType.INFO:
			return "alert-info";
		case MessageType.SUCCESS:
			return "alert-success";
	}
};

/**
 * Show the given message
 * @param {Message} message The message to show
 */
MessageWidget.prototype._showMessage = function(message) {
	this._showing = true;
	this._currentMessageObject = message;
	// Create new message
	var button = $("<button />")
		.attr("type", "button")
		.addClass("close")
		.append(
			$("<i>", {
				class: "icon icon-small icon-WCR_Close12"
			})
		);
	var title = $("<strong />").append(this.i18n.t(message.getType().name));
	var messageSpan = $("<span />", {
		class: "message-text"
	})
		.append(title)
		.append(" ");
	var action = message.getAction();
	if (action) {
		var actionLink = $("<a/>")
			.addClass("alert-link")
			.append(message.getMessage());
		if (typeof action === "string") {
			actionLink.attr("href", action);
		} else {
			actionLink.attr("href", "#").on(
				"click",
				_.bind(function(e) {
					// On action executed, destroy and show next
					e.stopPropagation();
					this._destroyCurrentMessage(
						_.bind(function() {
							message.getAction()();
							this._showNextMessage();
						}, this)
					);
				}, this)
			);
		}
		messageSpan.append(actionLink);
	} else {
		messageSpan.append(message.getMessage());
	}
	this._currentMessage = $("<div />")
		.addClass("alert")
		.addClass(this._getCSSClass(message.getType()));
	if (message.allowClose()) {
		this._currentMessage.append(button);
	}
	this._currentMessage.append(messageSpan).hide();
	this._currentMessage.on("contextmenu mouseenter mouseleave mousedown click mousemove mouseup dblclick", function(e) {
		e.stopPropagation();
	});

	// select the message when double clicking on the span
	messageSpan.on("dblclick", function(e) {
		if (document.selection) {
			var range = document.body.createTextRange();
			range.moveToElementText(this);
			range.select();
		} else if (window.getSelection) {
			var range = document.createRange();
			range.selectNodeContents(this);
			window.getSelection().removeAllRanges();
			window.getSelection().addRange(range);
		}
		e.stopPropagation();
		e.preventDefault();
	});

	// Auto hide if needed
	var timer;
	if (message.getAutoHide()) {
		timer = window.setTimeout(
			_.bind(function() {
				this._destroyCurrentMessage(
					_.bind(function() {
						this._showNextMessage();
					}, this)
				);
			}, this),
			this._AUTOHIDE_DELAY
		);
	}

	// On click destroy and show next
	button.on(
		"click",
		_.bind(function(e) {
			e.stopPropagation();
			if (timer) {
				window.clearTimeout(timer);
			}
			this._destroyCurrentMessage(
				_.bind(function() {
					this._showNextMessage();
				}, this)
			);
		}, this)
	);

	// Show it
	this._context.empty().append(this._currentMessage);
	this._currentMessage.fadeIn("fast");
};

/**
 * Destroy the current message
 * @param {Function} onDestroyedCallback - Callback executed when the message is destroyed
 */
MessageWidget.prototype._destroyCurrentMessage = function(onDestroyedCallback) {
	this._currentMessage.fadeOut(
		"fast",
		_.bind(function() {
			this._currentMessage.remove();
			this._showing = false;
			onDestroyedCallback();
		}, this)
	);
};

/**
 * Show the next message if there is any
 */
MessageWidget.prototype._showNextMessage = function() {
	if (!this._showing) {
		var m = this._messageModel.getNextMessage();
		if (m) {
			this._showMessage(m);
		}
	}
};

/**
 * Check if the widget is currently showing a message with a certain tag
 * @param {String} tag - The tag to check
 */
MessageWidget.prototype._isShowingMessageWithTag = function(tag) {
	return this._showing && this._currentMessage && this._currentMessageObject && this._currentMessageObject.getTag() === tag;
};

/**
 * Mark the current message as obsolete if it has the tag provided
 * @param {String} tag - the tag of the messages to mark obsolete
 */
MessageWidget.prototype._markObsolete = function(tag) {
	if (this._isShowingMessageWithTag(tag)) {
		this._destroyCurrentMessage(
			_.bind(function() {
				this._showNextMessage();
			}, this)
		);
	}
};

/**
 * Get the context where this message widget is added to
 * @returns {JQueryObj}
 */
MessageWidget.prototype.getContext = function() {
	return this._context;
};

/**
 * Get the number of messages (including the currently displayed message)
 * @returns {Number} [[Description]]
 */
MessageWidget.prototype.getMessageCount = function() {
	var count = this._messageModel.getMessageCount();
	if (this._showing) {
		count++;
	}
	return count;
};

export default MessageWidget;
