import $ from "jquery";
import jqueryUI from "jqueryUI";
import _ from "underscore";
import "bootstrap";
/**
 * Dynamically create a Bootstrap Modal
 * @param {JQueryObj|HTMLObj} content
 * @param {object}            options
 */
var Modal = function(content, options) {
	// available options with their defaults
	this._options = _.extend(
		{
			allowClose: true, // allow the modal be closed by a close icon, hitting escape or clicking the backdrop
			closeOnBackdropClick: true, // if false, the modal won't close when clicking on the backdrop
			widthClass: "", // add a custom class for defining the width of the modal
			customClass: "", // add a custom class for the full modal
			animate: true, // animate the modal
			customCloseHandler: null, // you could disable allow close, but still provide a custom close handler function, in that case a close icon is shown, on click this function is called
			onClose: null, // register on close callback
			destroyOnClose: false, // if true, modal destroys itself on close
			onOpen: null, // register on open callback
			onShown: null, // register on shown callback
			draggable: false, // allow dragging the modal around on the page from the header
			autoFocus: true, // automatically focus the first form control if no focusel is set
			beforeClose: function() {
				return true;
			}
		},
		options
	);

	// Content should always be wrapped by jQuery, do this if not ...
	if (!(content instanceof $)) {
		content = $(content);
	}

	// The array that will hold all the functions that need to be called on close.
	this._onCloseFunctions = [];
	this._onOpenFunctions = [];
	this._onShownFunctions = [];

	// Parse the callbacks for the modal.
	if (this._options.onClose) {
		this._onCloseFunctions.push(this._options.onClose);
	}
	if (this._options.onOpen) {
		this._onOpenFunctions.push(this._options.onOpen);
	}
	if (this._options.onShown) {
		this._onShownFunctions.push(this._options.onShown);
	}

	// Create the content
	this._content = content;

	var animate = " fade ";
	if (!this._options.animate) {
		animate = " ";
	}

	// Create the other elements of the modal
	this._wrapper = $("<div>", {
		class: "modal fl-scrolls-viewport" + animate + this._options.customClass
	});
	this._modalDialog = $("<div>", {
		class: "modal-dialog " + this._options.widthClass
	});
	this._modalContent = $("<div>", {
		class: "modal-content"
	});
	this._modalHeader = $("<div>", {
		class: "modal-header"
	});
	this._modalCloseButton = $("<button>", {
		class: "close",
		type: "button",
		"data-dismiss": "modal"
	});
	this._modalHeaderTitle = $("<h4>", {
		class: "modal-title"
	});
	this._modalBody = $("<div>", {
		class: "modal-body"
	});
	this._modalFooter = $("<div>", {
		class: "modal-footer"
	});

	this._helpIcon = null;

	this._focusEl = null;

	// Only add the close button when we allow the user to dismiss this modal.
	if (this._options.allowClose) {
		this._modalCloseButton.append(
			$("<i>", {
				class: "icon icon-WCR_Reject"
			})
		);

		if (this._options.customCloseHandler) {
			this._modalCloseButton.on("click", this._options.customCloseHandler);
		}

		this._modalHeader.append(this._modalCloseButton);
	}

	this._modalHeader.append(this._modalHeaderTitle);
	this._modalBody.append(this._content);

	this._modalContent.append(this._modalBody);
	this._modalDialog.append(this._modalContent);
	this._wrapper.append(this._modalDialog);

	var modalOptions = {
		show: false
	};

	if (!this._options.closeOnBackdropClick) {
		modalOptions.backdrop = "static";
	}
	// Apply some settings to the modal for when we do not allow it to be closed ...
	// or when we have custom close handler
	if (!this._options.allowClose || this._options.customCloseHandler) {
		modalOptions.backdrop = "static";
		modalOptions.keyboard = false;
		this._wrapper.modal(modalOptions);
	} else {
		this._wrapper.modal(modalOptions);
	}

	// Make the modal draggable if needed
	if (this._options.draggable) {
		this._modalDialog.draggable({
			handle: this._modalHeader
		});
		this._modalHeader.addClass("modal-header-draggable");
	}

	// When a model hidden event is fired do ...
	this._wrapper.on(
		"hide.bs.modal",
		function(e) {
			if (this._options.beforeClose() === false) {
				return false;
			}
			// execute all callbacks ...
			$.each(
				this._onCloseFunctions,
				function(index, callback) {
					callback();
				}.bind(this)
			);
		}.bind(this)
	);

	// When a model show event is fired do ...
	this._wrapper.on(
		"show.bs.modal",
		function(e) {
			// execute all callbacks ...
			fixZIndex.call(this._wrapper);
			$.each(
				this._onOpenFunctions,
				function(index, callback) {
					callback();
				}.bind(this)
			);
		}.bind(this)
	);

	// When a model shown event is fired (after animations) do ...
	this._wrapper.on(
		"shown.bs.modal",
		function(e) {
			if (!this._focusEl && this._options.autoFocus) {
				this._focusEl = this._wrapper.find(".form-control").first();
			}
			if (this._focusEl) {
				this._focusEl.trigger("focus");
			}

			// execute all callbacks ...
			$.each(
				this._onShownFunctions,
				function(index, callback) {
					callback();
				}.bind(this)
			);
		}.bind(this)
	);

	if (this._options.destroyOnClose) {
		this._wrapper.on(
			"hidden.bs.modal",
			function() {
				this.destroy();
			}.bind(this)
		);
	}
};

// Set the content.
Modal.prototype.setContent = function(content) {
	this._content = content;
	this._modalBody.empty().append(this._content);
};

// The title is optional, when not set, it will not be shown.
Modal.prototype.setTitle = function(newTitle) {
	this._modalHeaderTitle.text(newTitle);
	if (this._helpIcon) {
		this._modalHeaderTitle.append(this._helpIcon);
	}
	// If the header is not yet part of the modal, add it.
	if (this._modalHeader.parent()[0] !== this._modalContent[0]) {
		this._modalContent.prepend(this._modalHeader);
	} else if (newTitle === undefined || newTitle === "") {
		// If the passed element is undefined, remove the header again from the modal.
		this._modalContent.remove(this._modalHeader);
	}
};

// The footer is optional, when not set, it will not be shown.
Modal.prototype.setFooter = function(jQElement) {
	this._modalFooter.empty().append(jQElement);
	// If the footer is not yet part of the modal, add it.
	if (this._modalFooter.parent()[0] !== this._modalContent[0]) {
		this._modalContent.append(this._modalFooter);
	} else if (jQElement === undefined) {
		// If the passed element is undefined, remove the footer again from the modal.
		this._modalContent.remove(this._modalFooter);
	}
};

/**
 * Get a reference to the modal dialog element
 * @returns {JQueryObj}
 */
Modal.prototype.getModalDialog = function() {
	return this._modalDialog;
};

Modal.prototype.close = function() {
	this._wrapper.modal("hide");
};

/**
 * Open the popup
 * @param {boolean} fullscreen - make the popup fill most of the screen height, don't shrink to content height
 */
Modal.prototype.open = function(fullscreen) {
	this._wrapper.modal("show");

	if (fullscreen) {
		this._modalBody.css("min-height", "80vh"); //Temporarily give an aproximate min-height to prevent big jumps. Will become final in onShown.
		this.onShown(
			function() {
				//When the animation is finished and all elements are added to the DOM
				var windowHeight = $(window).height();
				var marginTop = this._modalDialog.css("marginTop").replace("px", "");
				var headerHeight = this._modalHeader.outerHeight(true) || null; //include margin height
				var footerHeight = this._modalFooter.outerHeight(true) || null;
				var marginBottom = this._modalDialog.css("marginBottom").replace("px", "");
				var bodyHeight = Math.floor(windowHeight - marginTop - headerHeight - footerHeight - marginBottom - 2); //-2 to counteract rounding errors and to make sure the scrollbars are not shown unnecessarily
				this._modalBody.css("min-height", bodyHeight + "px"); //Set the final min-height
			}.bind(this)
		);
	}
};

Modal.prototype.update = function() {
	this._wrapper.modal("handleUpdate");
};

Modal.prototype.onClose = function(onCloseFunction) {
	this._onCloseFunctions.push(onCloseFunction);
};

Modal.prototype.onOpen = function(onOpenFunction) {
	this._onOpenFunctions.push(onOpenFunction);
};

Modal.prototype.onShown = function(onShownFunction) {
	this._onShownFunctions.push(onShownFunction);
};

Modal.prototype.getJQElement = function() {
	return this._wrapper;
};

Modal.prototype.destroy = function() {
	this._wrapper.remove();
	fixModalOverlay();
};

Modal.prototype.getModalFooter = function() {
	return this._modalFooter;
};

Modal.prototype.getModalContent = function() {
	return this._modalContent;
};
/**
 * Add a Help Icon to the title which links to a certain topic
 * @param {String} topic
 */
Modal.prototype.setHelpTopic = function(topic) {
	import("component/helpIcon/HelpIcon").then(({ default: HelpIcon }) => {
		this._helpIcon = HelpIcon.getInlineHelpTopicJQuery(topic);
		// by triggering a setTitle, the help Icon will be inserted correctly
		this.setTitle(this._modalHeaderTitle.text());
	});
};

/**
 * Set which element should be focus when the modal is shown
 * @param {JQueryObj} focusEl - or null
 */
Modal.prototype.setFocusEl = function(focusEl) {
	this._focusEl = focusEl;
};

/***
 * The bootstraphack/code below fixes issues that arise when making multiple bootstrap modals at the same moment.
 * It fixes z-index issues and makes sure modal-open class is not removed when other modals are still available
 */

function fixZIndex() {
	var zIndex = 1050;
	if (!$(this).is(".modal-stack")) {
		// first time the modal is shown
		$(".modal:visible").each(function(i, item) {
			zIndex = Math.max(zIndex, parseInt($(item).css("z-index"), 10));
		});
		zIndex += 2;
	} else {
		// modal already has a z-index
		zIndex = parseInt($(this).css("z-index"), 10);
	}

	$(this).modal("setStackZIndex", zIndex);
}

/**
 * Are there any visible modals on the page?
 * @returns {Boolean} true if there are modals
 */
function hasModals() {
	var modals = $(".modal").filter(function() {
		var modal = $(this);
		return modal.is(":visible") || modal.is(".in") || modal.css("display") === "block";
	});

	return modals.length > 0;
}
/**
 * Fix the modal-open class on the body
 */
function fixModalOverlay() {
	// restore the modal-open class to the body element, so that scrolling works
	// properly after de-stacking a modal.

	$(document.body).toggleClass("modal-open", hasModals());
}

// when a modal is opened fix its zindex
$(document.body).on("show.bs.modal shown.bs.modal", ".modal", fixZIndex);

/**
 * Function to set the StackZIndex of a modal
 * @param {Number} zIndex
 */
$.fn.modal.Constructor.prototype.setStackZIndex = function(zIndex) {
	this.$element.css("z-index", zIndex).addClass("modal-stack");
	var that = this;
	setTimeout(function() {
		if (that.$backdrop) {
			var focused = that.$element.find(":focus");
			that.$element.insertAfter(that.$backdrop).css("z-index", zIndex);
			that.$backdrop.css("z-index", zIndex - 1).addClass("modal-stack");
			focused.trigger("focus");
		}
	}, 5);
};

// override the hideModal function of Bootstrap modal
// so it doesn't remove the modal-open if unneeded
// bootstrap always removed the modal-open class.. but we only want to do this if there's no other modals visible
// normally we would listen to hideden.bs.modal but this event doesn't bubble if the the element is removed from the DOM
$.fn.modal.Constructor.prototype.hideModal = function() {
	var that = this;
	this.$element.hide();
	this.backdrop(function() {
		that.resetAdjustments();
		that.resetScrollbar();
		that.$element.trigger("hidden.bs.modal");

		fixModalOverlay();
	});
};

export default Modal;
