import ShapeUtils from "@esko/shapeutils";
import css from "./popover-style.css";
import $ from "jquery";
import _ from "underscore";
/**
 * @class A popover is an element (some sort of dialog) that has an arrow and some content.
 * It can be attached to an element in a certain direction.
 * When the element cannot be placed into the chosen direction, the popover tries to figure out where it can be put.
 * If no place can be found, the popover is hidden.
 *
 * Use the constructor to initialize the popover once. The popover is not added to container (and thus hidden) until show is called once.
 * Use the public functions for adapting content and for showing and hiding the popover.
 * If you set content after you are showing the popover, you should call show again to recalculate the position.
 *
 * @exports Popover
 * @constructor
 * @param {JQueryObj} context - Element where the popover is attached to.
 * @param {JQueryObj} container - Container (in the DOM) where the popover should be added to.
 * @param {JQueryObj} restrict - Element to restrict the popover position to.
 * @param {String} [placement = "bottom"] - One of "top" | "right" | "bottom" | "left" | "innerLeft".
 * @param {String} [color] - Background-color of the popover and arrow. E.g. "#ffffff" for white popover with white arrow.
 * @param {String} [cssClass] - Class added to the popover element
 * @param {Number} [restrictBorder = 10] - Border that is maintained between the popover and the restrict element.
 * @param {Number} [minHeight] - the popover can have to much items to be shown, hides the arrow and shows scrollbars if needed,
 * 										keeps the popover on the placement and doesn't try to change placement, unless smaller than minHeight
 * @param {Boolean} [maxSizeOverwritesCss] - does the popover decide the max-width and max-height when a css class was provided
 * @param {String} [arrowColor = color] - Background-color of the arrow. This defaults to the general color of the popover
 * @param {boolean} forceOverflow - Force the tooltip to have overflow in case the tooltip breaks out of restrict element
 * @param {Object} {
 * 						inheritSizeFromContext //makes sure that the popup has the same width or height as the context (currently on supported for placements 'top' and 'bottom')
 * 					}
 */
var Popover = function(
	context,
	container,
	restrict,
	placement,
	color,
	cssClass,
	restrictBorder,
	minHeight,
	maxSizeOverwritesCss,
	arrowColor,
	forceOverflow,
	options
) {
	this.context = context;
	this.container = container;
	this.restrict = restrict;
	this.minHeight = minHeight;

	this._placementOptions = ["top", "right", "bottom", "left"];
	this._hasCustomPlacementOptions = false;
	this.placement = placement ? placement : "bottom";

	// Popup wrapper element
	this.element = $("<div />").addClass("wcr-popover");

	this.color = color;
	this.arrowColor = color;
	if (arrowColor) {
		this.arrowColor = arrowColor;
	}
	if (this.color) {
		this.element.css("background-color", color);
	}

	this.cssClass = cssClass;
	if (cssClass) {
		this.element.addClass(cssClass);
	}

	if (forceOverflow) {
		this.element.addClass("wcr-popover-force-overflow");
	}

	this.maxSizeOverwritesCss = maxSizeOverwritesCss;
	this.options = options || {};

	/**
	 * Recalculate the position of the popover when a contentResized event occurs
	 */
	this.element.on(
		"contentResized",
		_.bind(function(e) {
			this._recalculatePositions();
			e.stopPropagation();
		}, this)
	);

	// Border to restricting element
	this.restrictBorder = restrictBorder === undefined ? 10 : restrictBorder;

	this._arrow = $("<div />");
	this._content = $("<div />").addClass("content");
	if (this.minHeight === undefined) {
		this.element.append(this._arrow);
	}
	this.element.append(this._content);

	if (cssClass === undefined || maxSizeOverwritesCss) {
		// Defaults in pixels
		var maxWidth = 800;
		var maxHeight = this.restrict.innerHeight() * this.restrictBorder;
		// Min of css defined max-width and calculated max width
		maxWidth = Math.min(maxWidth, this.restrict.innerWidth() - 2 * this.restrictBorder);
		this._content.css({
			"max-height": maxHeight,
			"max-width": maxWidth
		});
	}
	// All possible placements
	this._placements = undefined;
	this._createPlacements();

	// Max recursion depth to avoid deadlock
	this.MAX_DEPTH = this._placements.length;

	// Indicate whether the popover is already part of the container
	this._addedToContainer = false;

	// Callbacks when added to container
	this._onAddedToContainerCallbacks = [];
	// Callbacks when destroyed
	this._onDestroyCallbacks = [];

	// Is the popover shown or not
	this._isActive = false;
};

/**
 * Create an array with all the possible placements.
 * The order is clockwise, so "top", "right", etc.
 * The "innerLeft" placement is added just before the current placement to have it as last in cyclic order beginning from the current placement (@see _getNextPlacement).
 * @returns {Array.<String>} Array of possible placements.
 */
Popover.prototype._createPlacements = function() {
	// All possible placements
	this._placements = this._placementOptions.slice(0); // take copy

	if (!this._hasCustomPlacementOptions) {
		// Add innerLeft as last possible placement (so just before the requested placement)
		var i = _.indexOf(this._placements, this.placement);
		if (i < 0) {
			// if not found (placement is innerLeft), just add it at the beginning of the list.
			i = 0;
		}
		this._placements.splice(i, 0, "innerLeft");
	}
};

/**
 * Will restrict popover to the given placements.
 * Possible options are top, right, bottom, left, innerLeft.
 * @param {Array} placementOptions - Array containing the allowed placement options.
 */
Popover.prototype.setPlacementOptions = function(placementOptions) {
	this._placementOptions = placementOptions.splice(0);
	this.placement = this._placementOptions[0] ? this._placementOptions[0] : "bottom";
	this._hasCustomPlacementOptions = true;
	this._createPlacements();
};

/**
 * Get the shape of the context element
 * @returns {ShapeUtils.Rect} The positions left (x), top (y) and the dimensions width and height of the context element.
 */
Popover.prototype._getContextRect = function() {
	return ShapeUtils.rectMake(
		this.context.offset().left,
		this.context.offset().top,
		this.context.outerWidth(true),
		this.context.outerHeight(true)
	);
};

/**
 * Get the shape of the restrict elemnet
 * @returns {ShapeUtils.Rect} The positions left (x), top (y) and the dimensions width and height of the restrict elemnet.
 */
Popover.prototype._getRestrictRect = function() {
	return ShapeUtils.rectMake(
		this.restrict.offset().left,
		this.restrict.offset().top,
		this.restrict.innerWidth(),
		this.restrict.innerHeight()
	);
};

/**
 * Get the element size
 * @returns {ShapeUtils.Size} The dimensions width and height of the popover.
 */
Popover.prototype._getElementSize = function() {
	return new ShapeUtils.Size(this.element.outerWidth(true), this.element.outerHeight(true));
};

/**
 * Get the arrow size
 * @returns {ShapeUtils.Size} The dimensions width and height of the arrow.
 */
Popover.prototype._getArrowSize = function() {
	return new ShapeUtils.Size(this._arrow.outerWidth(true), this._arrow.outerHeight(true));
};

/**
 * Restrict the top and left of an element to the borders of a restricting element
 * @param {Object} elementTopLeft - The positions left , top (y) of the element to restrict.
 * @param {ShapeUtils.Size} elementSize - The width and height of the element to restrict.
 * @param {ShapeUtils.Rect} restrict - The positions left (x), top (y) and the dimensions width and height from the restricting element.
 */
Popover.prototype._restrictToBorders = function(elementTopLeft, elementSize, restrict) {
	if (elementTopLeft.left < restrict.left() + this.restrictBorder) {
		elementTopLeft.left = restrict.left() + this.restrictBorder;
	}
	if (elementTopLeft.top < restrict.top() + this.restrictBorder) {
		elementTopLeft.top = restrict.top() + this.restrictBorder;
	}
	if (elementTopLeft.left + elementSize.width > restrict.right() - this.restrictBorder) {
		elementTopLeft.left = Math.max(0, restrict.right() - elementSize.width - this.restrictBorder); // keep within browser window
	}
	if (
		(this.minHeight === undefined || restrict.bottom() - this.restrictBorder - elementTopLeft.top < this.minHeight) &&
		elementTopLeft.top + elementSize.height > restrict.bottom() - this.restrictBorder
	) {
		elementTopLeft.top = Math.max(0, restrict.bottom() - elementSize.height - this.restrictBorder); // keep within browser window
	}
};

/**
 * Get the next placement given a placement.
 * @param {String} placement - One of "top" | "right" | "bottom" | "left" | "innerLeft".
 * @returns The next placement as specified by the ordered array _placements. E.g "top" will return "right", "right" will return "bottom", etc. and "innerLeft" will come last.
 */
Popover.prototype._getNextPlacement = function(placement) {
	return this._placements[(_.indexOf(this._placements, placement) + 1) % this._placements.length];
};

/**
 * Reset the arrow color to default (= from css).
 */
Popover.prototype._resetArrowColor = function() {
	this._arrow.css("border-top-color", "");
	this._arrow.css("border-bottom-color", "");
	this._arrow.css("border-left-color", "");
	this._arrow.css("border-right-color", "");
};

/**
 * Set the color and arrow direction for the given placement
 * @param {String} placement One of "top" | "right" | "bottom" | "left" | "innerLeft"
 */
Popover.prototype._setArrow = function(placement) {
	this._resetArrowColor();
	this._arrow.removeClass();
	switch (placement) {
		case "top":
			this._arrow.addClass("arrow-down");
			if (this.arrowColor) {
				this._arrow.css("border-top-color", this.arrowColor);
			}
			break;
		case "bottom":
			this._arrow.addClass("arrow-up");
			if (this.arrowColor) {
				this._arrow.css("border-bottom-color", this.arrowColor);
			}
			break;
		case "left":
			this._arrow.addClass("arrow-right");
			if (this.arrowColor) {
				this._arrow.css("border-left-color", this.arrowColor);
			}
			break;
		case "right":
			this._arrow.addClass("arrow-left");
			if (this.arrowColor) {
				this._arrow.css("border-right-color", this.arrowColor);
			}
			break;
		case "innerLeft":
			this._arrow.addClass("arrow-left");
			if (this.arrowColor) {
				this._arrow.css("border-right-color", this.arrowColor);
			}
			break;
	}
};

/**
 * Calculate the position of the arrow and the content, based on the placement.
 * The depth is a parameter that avoids a deadlock in recursion, so to be sure this function always returns something.
 * @param {Number} depth - Recursion depth
 * @param {String} placement - One of "top" | "right" | "bottom" | "left" | "innerLeft"
 * @returns {Object} The top and left position of arrow and content, and the contentMaxHeight if needed or undefined if popover can't be put right
 */
Popover.prototype._calculatePositions = function(depth, placement) {
	/**
	 * Avoid deadlock
	 */
	// We can't place the popover correct
	if (depth >= this.MAX_DEPTH) {
		// Reset it to the one we had
		this._setArrow(this.placement);
		return undefined;
	}

	/**
	 * Set the arrow direction for the given placement
	 */
	this._setArrow(placement);

	/**
	 * Calculate positions
	 */
	var cx = this._getContextRect(); // context
	var a = this._getArrowSize(); // arrow
	var cn = this._getElementSize(); // content
	var rs = this._getRestrictRect(); // restrict

	var arrowPos;
	var contentPos;

	switch (placement) {
		case "top":
			arrowPos = {
				top: cx.top() - a.height,
				left: cx.left() + cx.size.width / 2 - a.width / 2
			};
			contentPos = {
				top: cx.top() - a.height - cn.height,
				left: cx.left() + (this.options.inheritSizeFromContext ? 0 : cx.size.width / 2 - cn.width / 2)
			};
			break;
		case "bottom":
			arrowPos = {
				top: cx.top() + cx.size.height,
				left: cx.left() + cx.size.width / 2 - a.width / 2
			};
			contentPos = {
				top: cx.top() + a.height + cx.size.height,
				left: cx.left() + (this.options.inheritSizeFromContext ? 0 : cx.size.width / 2 - cn.width / 2)
			};
			break;
		case "left":
			arrowPos = {
				top: cx.top() + cx.size.height / 2 - a.height / 2,
				left: cx.left() - a.width
			};
			contentPos = {
				top: cx.top() + cx.size.height / 2 - cn.height / 2,
				left: cx.left() - a.width - cn.width
			};
			break;
		case "right":
			arrowPos = {
				top: cx.top() + cx.size.height / 2 - a.height / 2,
				left: cx.left() + cx.size.width
			};
			contentPos = {
				top: cx.top() + cx.size.height / 2 - cn.height / 2,
				left: cx.left() + cx.size.width + a.width
			};
			break;
		case "innerLeft":
			arrowPos = {
				top: cx.top() + cx.size.height / 2 - a.height / 2,
				left: cx.left()
			};
			contentPos = {
				top: cx.top() + cx.size.height / 2 - cn.height / 2,
				left: cx.left() + a.width
			};
			break;
		case "bottomLeft":
			arrowPos = {
				top: cx.top() + cx.size.height / 2 - a.height / 2,
				left: cx.left() + a.width / 2
			};
			contentPos = {
				top: cx.top() + a.height + cx.size.height,
				left: cx.left()
			};
	}

	/**
	 * SMART Popover
	 */
	// Calculate best arrow position with respect to restricting element
	// check if there is enough space at the top
	if (arrowPos.top <= rs.top() + this.restrictBorder) {
		return this._calculatePositions(depth + 1, this._getNextPlacement(placement));
	}
	// check if there is enough space at the bottom
	if (arrowPos.top + a.height >= rs.bottom() - this.restrictBorder) {
		return this._calculatePositions(depth + 1, this._getNextPlacement(placement));
	}
	// check if there is enough space at the left side
	if (arrowPos.left <= rs.left() + this.restrictBorder) {
		return this._calculatePositions(depth + 1, this._getNextPlacement(placement));
	}
	// check if there is enough space at the right side
	if (arrowPos.left + a.width >= rs.right() - this.restrictBorder) {
		return this._calculatePositions(depth + 1, this._getNextPlacement(placement));
	}

	// Restrict content to borders
	this._restrictToBorders(contentPos, cn, rs, this.minHeight);

	// If content moved over arrow by restricting, search another place
	switch (placement) {
		case "top":
			if (contentPos.top + (this.minHeight === undefined ? cn.height : Math.min(cn.height, this.minHeight)) > arrowPos.top) {
				return this._calculatePositions(depth + 1, this._getNextPlacement(placement));
			}
			break;
		case "bottom":
			if (contentPos.top < arrowPos.top + a.height) {
				return this._calculatePositions(depth + 1, this._getNextPlacement(placement));
			}
			break;
		case "left":
			if (contentPos.left + cn.width > arrowPos.left) {
				return this._calculatePositions(depth + 1, this._getNextPlacement(placement));
			}
			break;
		case "right":
			if (contentPos.left < arrowPos.left + a.width) {
				return this._calculatePositions(depth + 1, this._getNextPlacement(placement));
			}
			break;
		case "innerLeft":
			if (contentPos.left < arrowPos.left + a.width) {
				return this._calculatePositions(depth + 1, this._getNextPlacement(placement));
			}
			break;
	}

	arrowPos.top = Math.floor(arrowPos.top);
	arrowPos.left = Math.floor(arrowPos.left);
	contentPos.top = Math.floor(contentPos.top);
	contentPos.left = Math.floor(contentPos.left);
	var result = {
		arrow: arrowPos,
		content: contentPos,
		placement: placement
	};

	if (this.cssClass !== undefined && this.maxSizeOverwritesCss) {
		switch (placement) {
			case "bottom":
			case "bottomLeft":
				result.contentMaxHeight = this.restrict.offset().top + this.restrict.innerHeight() - this.restrictBorder - contentPos.top;
				break;
			case "top":
				result.contentMaxHeight = arrowPos.top - this.restrict.offset().top - this.restrictBorder;
				break;
		}
	}
	return result;
};

/**
 * Recalculate position of the popover.
 */
Popover.prototype._recalculatePositions = function() {
	if (this._addedToContainer) {
		var scroll = this._content.scrollTop();

		// reset max height, as this can affect the position where it will be displayed
		// (we need to take the full size into account during calculations)
		this.element.css({
			"max-height": "",
			"max-width": ""
		});
		this._content.css({
			"max-height": "",
			"max-width": ""
		});
		var positions = this._calculatePositions(0, this.placement);
		if (positions) {
			_.each(
				this._placementOptions,
				function(option) {
					this.element.removeClass("opened-" + option);
				},
				this
			);
			this.element.addClass("opened-" + positions.placement);
			this.element.offset(positions.content); // order matters!
			if (this.options.inheritSizeFromContext) {
				this.element.css("width", this.context.outerWidth(true));
			}
			this._arrow.offset(positions.arrow);

			if (positions.contentMaxHeight !== undefined) {
				this._content.css("max-height", positions.contentMaxHeight);
			}

			// this._content might break outside of the restrict element
			// recalculate max-height/max-width of content to trigger overflows
			if (this.restrict.length > 0) {
				var contentClientRect = this._content[0].getBoundingClientRect();
				var restrictClientRect = this.restrict[0].getBoundingClientRect();
				var maxHeight = restrictClientRect.height - contentClientRect.top + restrictClientRect.top;
				var maxWidth = restrictClientRect.width - contentClientRect.left + restrictClientRect.left;
				this._content.css({
					"max-height": maxHeight,
					"max-width": maxWidth
				});
			}
		} else {
			// if we can't find a good position, hide it
			this.hide();
		}

		this._content.scrollTop(scroll);
	}
};

/**
 * Set the context element where the popover is attached to
 */
Popover.prototype.setContext = function(context) {
	this.context = context;
};

/**
 * Set the preferred placement
 * @param {String} placement - One of "top" | "right" | "bottom" | "left" | "innerLeft".
 */
Popover.prototype.setPreferredPlacement = function(placement) {
	this.placement = placement;
	// recreate the list of possible placements so innerLeft comes last
	this._createPlacements();
};

/**
 * Show the popover
 */
Popover.prototype.show = function() {
	if (!this._addedToContainer) {
		this.container.append(this.element);
		this._addedToContainer = true;
		_.each(this._onAddedToContainerCallbacks, function(onAddedToContainerCallback) {
			onAddedToContainerCallback();
		});
	}
	this.element.show();
	this._recalculatePositions();
	this._isActive = true;
};

/**
 * Hide the popover
 */
Popover.prototype.hide = function() {
	this.element.hide();
	this._isActive = false;
};

/**
 * Is the popover active (shown) or not
 * @returns {Boolean} True if shown, false otherwise
 */
Popover.prototype.isActive = function() {
	return this._isActive;
};

/**
 * Destroy the popover
 */
Popover.prototype.destroy = function() {
	_.each(this._onDestroyCallbacks, function(onDestroyCallback) {
		onDestroyCallback();
	});
	this._content.detach();
	this.element.remove();
};

/**
 * Set the content of the popover.
 * Call show to recalculate the position.
 * @param {JQueryObj} newContent - New content to replace the old content
 */
Popover.prototype.setContent = function(newContent) {
	this._content.empty().append(newContent);
};

/**
 * Set the container of the popover
 * @param container
 */
Popover.prototype.setContainer = function(container) {
	this._container = container;
};

/**
 * Add content to the end of the popover
 * Call show to recalculate the position.
 * @param {JQueryObj} newContent - New content to add after the old content
 */
Popover.prototype.appendContent = function(newContent) {
	this._content.append(newContent);
};

/**
 * Add a callback for when the element is added to the container
 * @param {Function} onAddedToContainerCallback - Callback to be executed when popover is added to container
 */
Popover.prototype.onAddedToContainer = function(onAddedToContainerCallback) {
	this._onAddedToContainerCallbacks.push(onAddedToContainerCallback);
};

/**
 * Add a callback just before the element is destroyed
 * @param {Function} onDestroyCallback  - Callback to be executed just before the popover is destroyed
 */
Popover.prototype.onDestroy = function(onDestroyCallback) {
	this._onDestroyCallbacks.push(onDestroyCallback);
};

/**
 * Get the JQueryObj of the content div
 * @returns {JQueryObj}
 */
Popover.prototype.getContentDiv = function() {
	return this._content;
};

/**
 * Get the JQueryObject of the popover
 * @returns {JQueryObj}
 */
Popover.prototype.getElement = function() {
	return this.element;
};

export default Popover;
