import $ from "jquery";
import SliderMulti from "./SliderMulti";
import SliderSingle from "./SliderSingle";
import SliderConstants from "./SliderConstants";
import "./slider-style.css";
/**
 * Generic HTML5/JS slider component
 *
 * @param {Object} options Mandatory object that can contains following options
 * {
 * 		max: {Number} represents maximum value of the slider,
 * 		[min]: {Number} minimum value this slider can represent (default: 0),
 * 		[class]: {String} classname that is added to the element,
 * 		[on]: {Object} contains a mapping of eventnames and callbacks
 * 			{
 * 				"drag.start": function.., // is called when the user starts to move the slider
 * 				"drag.end": function.., // is called when the user lets go of the slider
 * 				"value.change": function..., // is called whenever the user changes the slider manually in SliderConstants.Mode.SINGLE
 * 				"values.change": function... // is called whenever the user changes the slider manually in SliderConstants.Mode.MULTI
 * 			}
 * 		[tooltip]: {Function} This function returns a string that can be used to generate tooltips over the indicators
 * 		[mode]: {SliderConstants.Mode} indicates the initial way of displaying the slider,
 * 		[single]: {Object} Contains options specific for the mode SliderConstants.Mode.SINGLE,
 * 					can also be passed when calling slider.setMode(SliderConstants.Mode.SINGLE, <option object>)
 * 			{
 * 				[value]: {Number} the initial value of the slider (default: 0),
 * 			}
 *		[multi]: {Object} Contains options specific for the mode SliderConstants.Mode.MULTI,
 *					can also be passed when calling slider.setMode(SliderConstants.Mode.MULTI, <option object>)
 * 			{
 * 				values: {Array<Number>} the initial values of the sliders
 * 				[fixedValues]: {Array<Number>} list of indices of indicators that cannot by sliding.
 * 				[keepOrder = false]: {Boolean} 	if true, the index of indicator the user is dragging will not change.
 * 										If the user tries to drag the indicator beyond another indicator,
 * 										the value of the latter will change to the value of the first.
 * 										When false, the user will start dragging the latter indicator rather than the first one.
 *
 *
 * 			}
 * }
 */
var Slider = function(options) {
	this._events = {};
	this._parseOptions(options);

	this.element = $("<div>", {
		class: (options.class || "") + " esko-slider"
	});

	this._layerWrapper = $("<div>", {
		class: "esko-slider-layers"
	}).appendTo(this.element);

	this._layers = {};
	[SliderConstants.Layer.TRACK, SliderConstants.Layer.TRACK_OVERLAY, SliderConstants.Layer.INDICATOR].forEach(
		function(layerName) {
			this._layers[layerName] = $("<div>", {
				class: "esko-slider-layer",
				"data-layer": layerName
			}).appendTo(this._layerWrapper);
		}.bind(this)
	);

	this._modes = {};
	this._modes[SliderConstants.Mode.SINGLE] = new SliderSingle(this);
	this._modes[SliderConstants.Mode.MULTI] = new SliderMulti(this);

	this.setMode(options.mode, options[options.mode]);

	if (window.PointerEvent) {
		// Pointer events are supported.
		this._registerPointerListeners();
	} else {
		// ADD ALL MOUSE LISTENERS TO SUPPORT DESKTOP
		this._registerMouseListeners();

		// ADD ALL TOUCH LISTENERS TO SUPPORT MOBILE
		this._registerTouchListeners();
	}
};

/**
 * Parses the options objects and checks for missing or invalid properties
 *
 * @param {Object} options
 */
Slider.prototype._parseOptions = function(options) {
	if (!options) {
		throw new Error("An options object must be provided!");
	} else {
		this._options = options;
	}

	if (!options.max) {
		throw new Error("Mandatory property 'max' is not specified");
	}
	options.min = options.min || 0;

	if (options.min >= options.max) {
		throw new Error("Property 'min' is larger than 'max'");
	}

	if (!options.single) {
		options.single = {};
	}

	options.single.value = options.single.value || options.min;

	if (options.single.value > options.max || options.single.value < options.min) {
		throw new Error("Initial value must be between min and max (" + options.min + "," + options.max + ")");
	}

	options.mode = options.mode || SliderConstants.Mode.SINGLE;

	options.tooltip =
		options.tooltip ||
		function() {
			return null;
		};

	// Register callbacks
	if (options.on) {
		for (var eventName in options.on) {
			if (options.on.hasOwnProperty(eventName) && options.on[eventName]) {
				this.on(eventName, options.on[eventName]);
			}
		}
	}
};

/**
 * Registers all mouse listeners
 *
 */
Slider.prototype._registerMouseListeners = function() {
	var dragging = false;
	this.element.on(
		"mousedown",
		function(e) {
			dragging = true;
			this.element.addClass("esko-slider-active");
			this.emit("drag.start");
			this._onDragStart(e.clientX);
		}.bind(this)
	);

	$(window.document).on(
		"mousemove",
		function(e) {
			if (dragging) {
				this._onDrag(e.clientX);
			}
		}.bind(this)
	);

	$(window.document).on(
		"mouseup",
		function(e) {
			if (dragging) {
				this.emit("drag.end");
				this._onDragEnd(e.clientX);
			}
			dragging = false;
			this.element.removeClass("esko-slider-active");
		}.bind(this)
	);
};

/**
 * Registers all touch listeners
 *
 */
Slider.prototype._registerTouchListeners = function() {
	var ongoingTouches = [];
	this.element[0].addEventListener(
		"touchstart",
		function(e) {
			if (e.cancelable) {
				e.preventDefault();
			}

			this.element.addClass("esko-slider-active");
			this.emit("drag.start");

			var touches = e.changedTouches;
			for (var i = 0; i < touches.length; i++) {
				ongoingTouches.push(touches[i].identifier);
				this._onDragStart(touches[i].clientX);
			}
		}.bind(this)
	);

	window.document.body.addEventListener(
		"touchmove",
		function(e) {
			var touches = e.changedTouches;
			for (var i = 0; i < touches.length; i++) {
				if (ongoingTouches.indexOf(touches[i].identifier) !== -1) {
					e.preventDefault();
					this._onDrag(touches[i].clientX);
				}
			}
		}.bind(this)
	);

	window.document.body.addEventListener(
		"touchend",
		function(e) {
			var touches = e.changedTouches;
			for (var i = 0; i < touches.length; i++) {
				var idx = ongoingTouches.indexOf(touches[i].identifier);
				if (idx >= 0) {
					e.preventDefault();
					this.emit("drag.end");
					this._onDragEnd(touches[i].clientX);
					this.element.removeClass("esko-slider-active");
					ongoingTouches.splice(idx, 1);
				}
			}
		}.bind(this)
	);

	window.document.body.addEventListener("touchcancel", function(e) {
		var touches = e.changedTouches;
		for (var i = 0; i < touches.length; i++) {
			var idx = ongoingTouches.indexOf(touches[i].identifier);
			if (idx >= 0) {
				e.preventDefault();
				ongoingTouches.splice(idx, 1);
			}
		}
	});
};

Slider.prototype._registerPointerListeners = function() {
	var dragging = false;
	this.element.on(
		"pointerdown",
		function(e) {
			e.preventDefault();
			e.stopPropagation();
			e.target.setPointerCapture(e.originalEvent.pointerId);
			dragging = true;
			this.element.addClass("esko-slider-active");
			this.emit("drag.start");
			this._onDragStart(e.clientX);
		}.bind(this)
	);

	this.element.on(
		"pointermove",
		function(e) {
			e.preventDefault();
			e.stopPropagation();
			if (dragging) {
				this._onDrag(e.clientX);
			}
		}.bind(this)
	);

	this.element.on(
		"pointerup pointercancel",
		function(e) {
			e.preventDefault();
			e.stopPropagation();
			if (dragging) {
				this.emit("drag.end");
				this._onDragEnd(e.clientX);
			}
			dragging = false;
			this.element.removeClass("esko-slider-active");
			this.element[0].releasePointerCapture(e.originalEvent.pointerId);
		}.bind(this)
	);
};

/**
 * Forwards drag event to current mode
 *
 * @param {Number} x clientX of input event
 */
Slider.prototype._onDrag = function(x) {
	this.mode.onDrag(this._parseEventToValue(x));
};

/**
 * Forwards drag start event to current mode
 *
 * @param {Number} x clientX of input event
 */
Slider.prototype._onDragStart = function(x) {
	this.mode.onDragStart(this._parseEventToValue(x));
};

/**
 * Forwards drag end event to current mode
 *
 * @param {Number} x clientX of input event
 */
Slider.prototype._onDragEnd = function(x) {
	this.mode.onDragEnd(this._parseEventToValue(x));
};

/**
 * Calculates the value on a slider for a given ClientX coordinate of an event
 *
 * @param {Number} x clientX of event
 * @returns value between specified min and max
 */
Slider.prototype._parseEventToValue = function(x) {
	var clientRect = this.element[0].getBoundingClientRect();
	var percentage = (x - clientRect.left) / clientRect.width;
	if (percentage < 0) {
		percentage = 0;
	} else if (percentage > 1) {
		percentage = 1;
	}

	var range = this._options.max - this._options.min;
	var value = percentage * range + this._options.min;
	return value;
};

/**
 * Utility function used to convert a value to a percentage of this slider
 *
 * @param {Number} value between min and max
 * @returns percentage (0-100)
 */
Slider.prototype._valueToPercentage = function(value) {
	var range = this._options.max - this._options.min;
	return ((value - this._options.min) / range) * 100;
};

/**
 * Renders extra marks on the track for a list of given mark objects.
 * Removes the previous marks every time that this function is called.
 *
 * Example:
 * 		slider.renderMarks([{
 * 		 	color: "red",
 * 			start: 5,
 * 			end: 20
 * 		}]);
 *
 * @param {Array.Object} marks
 */
Slider.prototype.renderMarks = function(marks) {
	window.requestAnimationFrame(
		function() {
			// Remove old marks
			this._layers[SliderConstants.Layer.TRACK_OVERLAY].empty();

			// Insert new marks
			if (marks) {
				marks.forEach(
					function(mark) {
						this._layers[SliderConstants.Layer.TRACK_OVERLAY].append(
							$("<div>", {
								class: "esko-slider-mark",
								css: {
									"background-color": mark.color,
									left: this._valueToPercentage(mark.start) + "%",
									right: 100 - this._valueToPercentage(mark.end) + "%"
								}
							})
						);
					}.bind(this)
				);
			}
		}.bind(this)
	);
};

/**
 * Changes the current display mode of the slider.
 *
 * Example:
 * 		slider.setMode(SliderConstants.Mode.MULTI, {
 *			values: [325, 575, ...]
 *		});
 *
 * @param {SliderConstants.Mode} newMode Possible values are SliderConstants.Mode.SINGLE and SliderConstants.Mode.MULTI
 * @param {Object} options options that are passed to the mode. See constructor for details.
 */
Slider.prototype.setMode = function(newMode, options) {
	// Stop the current mode
	if (this.mode) {
		this.mode.onStop();
	}
	// Fetch new mode and initialize it
	if (this._modes.hasOwnProperty(newMode)) {
		this.currentMode = newMode;
		this.mode = this._modes[newMode];
		this.mode.onStart(options);
	} else {
		throw new Error("Mode specified does not exist: " + newMode);
	}
};

/**
 * Registers a callback that is called for a given event. See constructor for a list of possible event names
 *
 * @param {String} eventName
 * @param {Function} callback
 */
Slider.prototype.on = function(eventName, callback) {
	if (!this._events.hasOwnProperty(eventName)) {
		this._events[eventName] = [];
	}
	this._events[eventName].push(callback);
};

/**
 * Removes a callback that is called for a given event
 *
 * @param {String} eventName
 * @param {Function} callback
 */
Slider.prototype.off = function(eventName, callback) {
	if (this._events.hasOwnProperty(eventName)) {
		this._events[eventName].splice(callback, 1);
	}
};

/**
 * Emits an event with the given name and triggers all registered callbacks
 *
 * @param {String} eventName
 * @param {Any} options passed to the callback
 */
Slider.prototype.emit = function(eventName) {
	var args = Array.prototype.slice.call(arguments);
	args.shift();
	if (this._events.hasOwnProperty(eventName) && this._events[eventName]) {
		this._events[eventName].forEach(function(callback) {
			callback.apply(undefined, args);
		});
	}
};

/**
 * Get the layer element of a given layer name
 *
 * @param {SliderConstants.Layer} layerName
 * @returns jquery object
 */
Slider.prototype._getLayer = function(layerName) {
	return this._layers[layerName];
};

export default Slider;
