import $ from "jquery";
import SliderConstants from "./SliderConstants";

/**
 * Responsible for displaying a range of values on the slider
 *
 * @param {Slider} slider
 */
var SliderMulti = function(slider) {
	this._slider = slider;

	this._indicators = [];
	this._tracks = [];
	this._currentActiveIndicator = 0;

	this._fixedValues = [];

	// Inject getter/setter to slider object
	slider.getValues = function() {
		return this._values;
	}.bind(this);

	slider.setValues = function(values) {
		if (values.length !== this._values.length) {
			throw new Error("Different amount of values given when slider was initialized");
		}
		this._values = values;
		this._adjustPositions();
	}.bind(this);
};

/**
 * Called when this mode is activated
 *
 * @param {Object} options
 */
SliderMulti.prototype.onStart = function(options) {
	this._slider.element.addClass("esko-slider-multi");

	this._values = options.values;
	if (this._values && this._values.length > 1) {
		for (var i = 0; i < this._values.length - 1; i++) {
			if (this._values[i] > this._values[i + 1]) {
				throw new Error("Value at position " + i + " is larger than the next value");
			}
		}
	} else {
		throw new Error("No values specified in options");
	}
	this._fixedValues = options.fixedValues || [];
	this._keepOrder = options.keepOrder;
	for (var i = 0; i < this._values.length - 1; i++) {
		var track = $("<div/>", {
			class: "esko-slider-track esko-slider-track-" + i
		});
		this._slider._getLayer(SliderConstants.Layer.TRACK).append(track);
		this._tracks.push(track);
	}

	for (var i = 0; i < this._values.length; i++) {
		var indicator = $("<div/>", {
			class: "esko-slider-indicator esko-slider-indicator-" + i
		});
		this._slider._getLayer(SliderConstants.Layer.INDICATOR).append(indicator);
		this._indicators.push(indicator);
	}

	this._adjustPositions();
};

/**
 * Called when this mode is de-activated
 *
 */
SliderMulti.prototype.onStop = function() {
	this._slider.element.removeClass("esko-slider-multi");
	this._values = [];
	this._indicators.forEach(function(indicator) {
		indicator.remove();
	});
	this._indicators = [];
	this._tracks.forEach(function(track) {
		track.remove();
	});
	this._tracks = [];
};

SliderMulti.prototype._calculateActiveIndicator = function(value) {
	var maxDist = -1;
	var indicator = 0;
	var indicatorValue = 0;
	for (var i = 0; i < this._values.length; i++) {
		if (this._fixedValues.indexOf(i) === -1) {
			var dist = Math.abs(this._values[i] - value);
			if (maxDist < 0 || dist < maxDist) {
				maxDist = dist;
				indicator = i;
				indicatorValue = this._values[i];
			} else if (dist === maxDist && value > indicatorValue) {
				indicator = i;
			}
		}
	}
	return indicator;
};

/**
 * Called when the user starts dragging somewhere on the slider
 *
 * @param {Number} value
 */
SliderMulti.prototype.onDragStart = function(value) {
	this._currentActiveIndicator = this._calculateActiveIndicator(value);
	this.onDrag(value);
};

/**
 * Called when the user is dragging somewhere on the slider
 *
 * @param {Number} value The value on the slider where the mouse/pointer currently sits
 */
SliderMulti.prototype.onDrag = function(value) {
	if (this._keepOrder) {
		for (var i = 0; i < this._values.length; i++) {
			if (i > this._currentActiveIndicator && this._values[i] < value && this._fixedValues.indexOf(i) !== -1) {
				value = this._values[i];
			} else if (i < this._currentActiveIndicator && this._values[i] > value && this._fixedValues.indexOf(i) !== -1) {
				value = this._values[i];
			}
		}

		this._values[this._currentActiveIndicator] = value;
		this._values = this._values.map(
			function(curValue, i) {
				if (i > this._currentActiveIndicator && this._values[i] < value) {
					return value;
				} else if (i < this._currentActiveIndicator && this._values[i] > value) {
					return value;
				}
				return curValue;
			}.bind(this)
		);
	} else {
		var doReOrder = true;
		while (doReOrder) {
			if (this._currentActiveIndicator > 0 && this._values[this._currentActiveIndicator - 1] > value) {
				this._values[this._currentActiveIndicator] = this._values[this._currentActiveIndicator - 1];
				this._values[this._currentActiveIndicator - 1] = value;
				this._currentActiveIndicator--;
				doReOrder = true;
			} else if (this._currentActiveIndicator < this._values.length - 1 && this._values[this._currentActiveIndicator + 1] < value) {
				this._values[this._currentActiveIndicator] = this._values[this._currentActiveIndicator + 1];
				this._values[this._currentActiveIndicator + 1] = value;
				this._currentActiveIndicator++;
				doReOrder = true;
			} else {
				this._values[this._currentActiveIndicator] = value;
				doReOrder = false;
			}
		}
	}
	this._adjustPositions();
	this._slider.emit("values.change", this._values, this._currentActiveIndicator);
};

/**
 * Called when the user stops dragging
 *
 */
SliderMulti.prototype.onDragEnd = function() {
	// Do nothing...
};

/**
 * Re-calculates all the css offsets based on the current list of values
 *
 */
SliderMulti.prototype._adjustPositions = function() {
	window.requestAnimationFrame(
		function() {
			for (var i = 0; i < this._values.length; i++) {
				var percentageStart = this._slider._valueToPercentage(this._values[i]);
				this._indicators[i].css("left", percentageStart + "%");
				this._indicators[i].attr("data-slider-tooltip", this._slider._options.tooltip(SliderConstants.Mode.MULTI, this._values[i]));

				if (i < this._values.length - 1) {
					var percentageEnd = this._slider._valueToPercentage(this._values[i + 1]);
					this._tracks[i].css("left", percentageStart + "%");
					this._tracks[i].css("right", 100 - percentageEnd + "%");
				}
			}
		}.bind(this)
	);
};

export default SliderMulti;
