import $ from "jquery";
import _ from "underscore";
/**
 * Create a new Template instance
 *
 * @exports Template
 * @memberof Layout
 *
 * @class Template A templating mechanism providing i18n and variables
 * @example
 * <div>
 * 	<h1>This is an example template for <%=d.someVariablename%></h1>
 * 	<p>
 * 		template syntax: <%=d.variable%> <%=d.i18n('Language_key')%>
 *  </p>
 * </div>
 *
 * Template usage:
 *
 * var t = new Template(i18n, "some tempalte code");
 * t.d.someVariable = 'a value';
 * t.d.someOtherVariable = 'another value';
 *
 * t.setContext($("<div/>"));
 * t.render();
 *
 * @param {Object} i18n - This translator object should have at least a method 't' to translate keys
 * @param {String} template - The template to set
 */
var Template = function(i18n, template) {
	this.bindings = {};
	this._context = undefined;

	this._subTemplates = {};

	this.d = {
		i18n: function() {
			return i18n.t.apply(i18n, arguments);
		},
		/**
		 * Call another sub-template.
		 * The template file should already be loaded and registered before calling this function!!
		 *
		 * call registerTemplate to register a sub-template
		 *
		 * @param {String} templateName - Name of the sub template e.g. "pagelist/PagelistBlock"
		 * @param {Object} [data] - Optional extra data for the sub template
		 * @param {String} [dataName] - Optional name of where the extra data can be accessed (via d.dataName). If undefined "d" (so d.d) contains the data (if any)
		 */
		callTemplate: _.bind(function(templateName, data, dataName) {
			if (dataName === undefined) {
				dataName = "d";
			}
			if (data !== undefined) {
				this.d[dataName] = data;
			}

			return this._subTemplates[templateName](this.d);
		}, this)
	};

	this.setTemplate(template);
};

/**
 * Set the template
 * @param {String} template
 */
Template.prototype.setTemplate = function(template) {
	this._parsedTemplate = template;
};

/**
 * Register subtemplate
 */
Template.prototype.registerTemplate = function(name, template) {
	this._subTemplates[name] = template;
};

/**
 * Register a subtemplates to be used with d.callTemplate
 */
Template.prototype.registerTemplates = function(templates) {
	this._subTemplates = { ...templates };
};

/**
 * Set the HTMLObject in which the template should be rendered
 * @param {JQueryObj} context
 */
Template.prototype.setContext = function(context) {
	if (this._context) {
		this.unbind();
	}
	this._context = context;
	if (context) {
		this.bind();
	}
};

/**
 * Return the context of the template.
 *
 * @returns {JQueryObj} The context of the template.
 */
Template.prototype.getContext = function() {
	return this._context;
};

/**
 * Render the template
 */
Template.prototype.render = function() {
	if (this._context) {
		this._context.html(this._parsedTemplate(this.d));
	}
};

/**
 * Return the rendered HTML
 */
Template.prototype.getRenderedHtml = function() {
	return this._parsedTemplate(this.d);
};

/**
 * Bind an event to the template contents
 * @param {String} type - The type of the event i.e. mouseover, click, ...
 * @param {String|JQueryIdentifier} identifier - Filter to identify the elements too which the event should be bound
 * @param {Function} callback - callback which should be callend when the event occurs (arguments: source, event)
 * @param {Object} [binding] - bind the callback to "binding"
 */
Template.prototype.bind = function(type, identifier, callback, binding) {
	if (type === undefined) {
		// rebind all events
		_.each(
			this.bindings,
			function(bindingsForType, type) {
				this._context.on(type, bindingsForType.handler);
			},
			this
		);
	} else {
		if (!this.bindings[type]) {
			// no bindings yet for this type, add binding handler
			this.bindings[type] = {};
			this.bindings[type].handler = _.bind(this._handleEvent, this, type);
			this.bindings[type].identifiers = {};
			if (this._context) {
				this._context.on(type, this.bindings[type].handler);
			}
		}

		var handlers = this.bindings[type];
		if (!handlers.identifiers[identifier]) {
			// no callbacks with the same identifier defined, add it
			handlers.identifiers[identifier] = [];
		}

		// push the callback
		if (binding) {
			callback = _.bind(callback, binding);
		}
		handlers.identifiers[identifier].push(callback);
	}
};

/**
 * Unbind an event
 *
 * CAUTION: unbind only works for events bound without the "binding" parameter
 * @param {String} type - The type of the event i.e. mouseover, click, ...
 * @param {String|JQueryIdentifier} identifier - Filter to identify the elements too which the event should be bound (note same identifier should be specified as in in the binding step)
 * @param {Function} callback - callback which should be callend when the event occurs (arguments: source, event) (note same callback should be specified as in the binding step
 */
Template.prototype.unbind = function(type, identifier, callback) {
	if (type === undefined) {
		// unbind all handlers
		_.each(
			this.bindings,
			function(bindingsForType, type) {
				this._context.off(type, bindingsForType.handler);
			},
			this
		);
	} else {
		if (this.bindings[type]) {
			if (this.bindings[type][identifier]) {
				this.bindings[type][identifier] = _.without(this.bindings[type][identifier], callback);
			}

			if (identifier === undefined && callback === undefined) {
				if (this._context) {
					this._context.off(type, this.bindings[type].handler);
				}

				delete this.bindings[type];
			} else if (callback === undefined) {
				delete this.bindings[type][identifier];
			}
		}
	}
};

/**
 * Handle an event, execute the correct callbacks
 * @param type
 * @param event
 */
Template.prototype._handleEvent = function(type, event) {
	// find all possible identifiers for this event type
	_.each(
		this.bindings[type].identifiers,
		function(callbacks, identifier) {
			// find elements mathcing the identifier in the event bubble chain
			var targets = $(event.target).parentsUntil(this._context, identifier);
			// the target itself can also be a possible target!
			if ($(event.target).is(identifier)) {
				targets = $(event.target, targets);
			}

			// find all elements matching the identifier on the event bubble path
			_.each(
				targets,
				function(target) {
					_.each(callbacks, function(callback) {
						callback(target, event);
					});
				},
				this
			);
		},
		this
	);
};

export default Template;
