import $ from "jquery";
import _ from "underscore";
/**
 * A ButtonGroup is a group of buttons
 *
 * @exports {ButtonGroup}
 * @class
 *
 * @param name {string} the name of the buttongroup
 * @param [buttons] {Array<Button>} the list of buttons which are in the buttongroup
 */
var ButtonGroup = function(name, buttons) {
	this.name = name;
	this.element = $("<div/>")
		.attr("name", name)
		.addClass("btn-group");
	this.btngroup = this.element;
	this.titleElement = undefined;
	this.buttons = [];

	/**
	 * add button at the beginning of the group
	 * @param {AbstractButton} button - the button to prepend
	 */
	this.prependButton = function(button) {
		this.btngroup.prepend(button.element);
		this.buttons.unshift(button);
	};

	/**
	 * Add a button to the group (if already present, the button is moved to the end)
	 * @param {Button} button - button to add
	 */
	this.addButton = function(button) {
		this.detachButton(button);

		this.btngroup.append(button.element);
		this.buttons.push(button);
	};

	/**
	 * Remove a button from the group
	 * @param {Button} button - button to remove
	 */
	this.detachButton = function(button) {
		var buttonCount = this.buttons.length;
		this.buttons = _.without(this.buttons, button);

		if (buttonCount != this.buttons.length) {
			button.element.detach();
		}
	};

	/**
	 * Remove all buttons from the buttongroup
	 */
	this.clearButtons = function() {
		_.each(this.buttons, function(button) {
			button.element.detach();
		});
		this.buttons = [];
	};

	/**
	 * Add a separator in the buttongroup
	 */
	this.addSeparator = function() {
		this.btngroup.append($("<span/>").addClass("separator"));
	};

	// add all buttons
	if (buttons) {
		for (var key in buttons) {
			if (buttons.hasOwnProperty(key)) {
				this.addButton(buttons[key]);
			}
		}
	}

	/**
	 * Hide the buttongroup
	 */
	this.hide = function() {
		this.element.addClass("displayNone");
	};

	/**
	 * Show the buttongroup
	 */
	this.show = function() {
		this.element.removeClass("displayNone");
	};
};
/**
 * Set title for buttongroup, the title is added under the buttongroup
 * @param i18n {I18n}
 * @param [langBundle] {String}
 * @param langKey {String}
 */
ButtonGroup.prototype.setTitle = function(i18n, langBundle, langKey) {
	this.element = $("<div/>")
		.addClass("btnGroupWithTitle")
		.append(this.element);
	if (this.titleElement === undefined) {
		this.titleElement = $("<div/>").addClass("btnGroupTitle");
		this.element.append(this.titleElement);
	}
	if (langBundle) {
		i18n.loadBundle(
			langBundle,
			_.bind(function() {
				this.titleElement.text(i18n.t(langKey));
			}, this)
		);
	} else {
		this.titleElement.text(i18n.t(langKey));
	}
};

/**
 * Called when the button group is moved
 */
ButtonGroup.prototype.moved = function() {
	_.each(this.buttons, function(button) {
		button.moved();
	});
};

export default ButtonGroup;
