import Button from "@esko/button/Button";
import Popover from "@esko/popover/Popover";
import $ from "jquery";
import _ from "underscore";
/**
 * A ToolbarButton is an interface element which provides an action and can have a text and icon description
 *
 * @exports ToolbarButton
 * @constructor
 * @param {String} Name - The name of the Toolbarbutton
 * @param {String | JQueryObj } icon - The class which is used for displaying the icon, or your own jQuery element
 * @param {Boolean} [forMobile] - If true the Toolbarbutton is useful on a mobile device
 * @param {Function} [action] - The function which is called when the Toolbarbutton is pressed (also see listenToTool)
 * @param {String} [tooltipText] the text for the button tooltip
 */

var ToolbarButton = function(i18n, context, isMobile, name, icon, forMobile, action, tooltipText) {
	if (_.isString(icon)) {
		this.iconClass = icon;
		icon = $("<i/>")
			.addClass("icon")
			.addClass(icon);
		if (isMobile) {
			icon.addClass("icon-mobile");
		} else {
			icon.addClass("icon-normal");
		}
	}
	Button.call(this, i18n, context, isMobile, name, icon, forMobile, action);

	if (tooltipText) {
		this.setTooltip(tooltipText);
	}
	this.context = context;
	this.icon = icon;
	this.element.addClass("btn").addClass("btn-default");
	this.subtoolbarComponents = [];
	this.subtoolbar = undefined;
	this.badgeText = "";

	this._listeningToToolOptionValue = undefined; // value of the tool option to listen to
};

ToolbarButton.prototype = Object.create(Button.prototype);
ToolbarButton.prototype.constructor = ToolbarButton;

/**
 * Add a subtoolbar component
 * @param {ToolbarComponent} toolbarComponent - ToolbarComponent to add to subtoolbar
 */
ToolbarButton.prototype.addSubtoolbarComponent = function(toolbarComponent, position) {
	this.subtoolbarComponents.push(toolbarComponent);
	if (!this.subtoolbar) {
		this.subtoolbar = new Popover(
			this.element,
			this.context,
			this.context,
			position || "bottom",
			"#bbbbbb",
			"subtoolbar",
			null,
			null,
			null,
			null,
			true
		);
	}
	this.subtoolbar.appendContent(toolbarComponent.element);
};

/**
 * Detach subtoolbar component
 * @param {ToolbarComponent} toolbarComponent - ToolbarComponent to add to subtoolbar
 */
ToolbarButton.prototype.detachSubtoolbarComponent = function(toolbarComponent) {
	this.subtoolbarComponents = _.without(this.subtoolbarComponents, toolbarComponent);
	toolbarComponent.element.detach();
};

/**
 * Set icon of Toolbarbutton
 * @param {String} icon - CSS class that contains the icon
 * @param {Array.<String>} [subicons] - list of sub icons
 */
ToolbarButton.prototype.setIcon = function(icon, subicons) {
	this.icon.empty();
	this.icon.removeClass();
	this.iconClass = icon;
	this.icon.addClass("icon").addClass(icon);
	if (this.forMobile) {
		this.icon.addClass("icon-mobile");
	}

	if (subicons) {
		this.icon.empty().addClass("containerIcon");
		_.each(
			subicons,
			function(subIcon) {
				this.icon.append($("<i>", { class: "icon subIcon " + subIcon }));
			},
			this
		);
	}
};

/**
 * Set the active status of the Toolbarbutton
 * @param {Boolean} active - True to set active, false to deactivate
 */
ToolbarButton.prototype.setActive = function(active) {
	if (!this.isEnabled()) {
		return;
	}

	if (active) {
		// If already active, we should not re-activate it to avoid flickering
		if (!this.isActive()) {
			this.element.addClass("active");
			if (this.subtoolbar) {
				this.subtoolbar.show();
			}
		}
	} else {
		if (this.isActive()) {
			this.element.removeClass("active");
			if (this.subtoolbar) {
				this.subtoolbar.hide();
			}
		}
	}
};

/**
 * Called when the button is moved
 * @override
 */
ToolbarButton.prototype.moved = function() {
	if (this.subtoolbar !== undefined && this.isActive()) {
		// recalculate positions
		this.subtoolbar.show();
	}
};

/**
 * Show the button
 * @Override
 */
ToolbarButton.prototype.show = function() {
	this.element.css("display", "inline-block");
};

/**
 * Set the badge text of the button
 *
 * @param {String} text - if undefined no badge will be shown
 * @param {String} [cssClass] - css class(es) to apply to the badge
 */
ToolbarButton.prototype.setBadge = function(text, cssClass) {
	this.badgeText = text;

	if (this.badgeText) {
		if (!this.badge) {
			this.badge = $("<span />");
			this.icon.append(this.badge);
		}
		this.badge.removeClass().addClass("badge");
		if (cssClass) {
			this.badge.addClass(cssClass);
		}

		this.badge.text(this.badgeText);
		this.badge.show();
	} else if (this.badge) {
		this.badge.hide();
	}
};
export default ToolbarButton;
