import Popover from "@esko/popover/Popover";
import css from "./tooltip-style.css";
import $ from "jquery";
import _ from "underscore";
/**
 * A Tooltip is a Popover with only text content. It can be added to other elements (e.g. Button) and shows
 * additional information about the element when the user hovers (mouseover) over the element.
 *
 * @exports Tooltip
 * @constructor
 * @augments Layout.Popover
 * @param {I18n} [i18n] - The language string provider optional if langKey is not defined
 * @param {JQueryObj} context - The element to which the tooltip should be binded
 * @param {JQueryObj} container - Container (in the DOM) where the tooltip should be added to.
 * @param {JQueryObj} restrict - Element to restrict the tooltip position to.
 * @param {Boolean} isMobile - True if it is for mobile, false otherwise
 * @param {String} tooltipTextorBundle - if langKey is undefined this is considered to be tooltipText
 * @param {String} [langKey] - language key of the text for the tooltip
 * @param {String} [stringArgs] - string arguments for the text
 * @param {Object} [popoverOptions] - An object containing the style arguments of Popover
 */
var Tooltip = function(i18n, context, container, restrict, isMobile, tooltipTextOrBundle, langKey, stringArgs, popoverOptions) {
	var color = "#333";
	if (popoverOptions && popoverOptions.color) {
		color = popoverOptions.color;
	}
	var cssClass = "tooltip";
	if (popoverOptions && popoverOptions.cssClass) {
		cssClass = popoverOptions.cssClass;
	}
	var restrictBorder = 3;
	if (popoverOptions && popoverOptions.restrictBorder) {
		restrictBorder = popoverOptions.restrictBorder;
	}
	var prefPosition = isMobile ? "top" : "bottom";
	if (popoverOptions && popoverOptions.prefPosition) {
		prefPosition = popoverOptions.prefPosition;
	}

	Popover.call(this, context, container, restrict, prefPosition, color, cssClass, restrictBorder);
	// maximum number of chars on one line
	this.MAX_CHARS = 65;

	this._i18n = i18n;
	this._isMobile = isMobile;

	this.tooltip = $("<div>");

	this._touchAndHoldShow = _.bind(this._delayedShow, this);
	this._show = _.bind(this.show, this);
	this._hide = _.bind(this.hide, this);

	this.setTooltipContent(tooltipTextOrBundle, langKey, stringArgs);

	this._timer = null;
	this._enabled = false;
	this.enableTooltip();
};

Tooltip.prototype = Object.create(Popover.prototype);
Tooltip.prototype.constructor = Tooltip;

/**
 * Set tooltip content. The langKey parameter can be omitted.
 * @param {String} tooltipTextOrBundle - tooltip text if langKey is undefined or language bundle name
 * @param {String} [langKey] - not required
 * @param {Array.<String>} [stringArgs] - not required, parameters that need to be filled in the language string
 */
Tooltip.prototype.setTooltipContent = function(tooltipTextOrBundle, langKey, stringArgs) {
	if (langKey) {
		this._i18n.loadBundle(
			tooltipTextOrBundle,
			function() {
				if (stringArgs) {
					stringArgs.unshift(langKey);
				} else {
					stringArgs = [langKey];
				}
				var text = this._i18n.t.apply(this._i18n, stringArgs);
				this.tooltip.empty().append(this._fitLine(text));
				this.setContent(this.tooltip);
			},
			this
		);
	} else {
		this.tooltip.empty().append(this._fitLine(tooltipTextOrBundle));
		this.setContent(this.tooltip);
	}

	// Listeners should be refreshed after updating text for tool tip.
	this.refereshListeners();
};

/**
 * Split the given text in multiple lines if it's too long (MAX_CHARS) to fit on one line. The split point is a space.
 * @param {String} text - The text to split.
 * @returns {String} The split text (with <br /> tags) or original text if no possible split is found (no spaces).
 */
Tooltip.prototype._fitLine = function(text) {
	if (text.length < this.MAX_CHARS || text.lastIndexOf("<br/>") !== -1) {
		return text;
	} else {
		// Avoided tail recursion
		var multiLineText = "";
		do {
			// split on spaces
			var p = text.lastIndexOf(" ", this.MAX_CHARS);
			if (p != -1) {
				// split found, append substr and search further for possibly other splits
				multiLineText += text.substr(0, p) + "<br />";
				text = text.substr(p);
			} else {
				// split can't be found, append everything
				multiLineText += text;
				text = "";
			}
		} while (text.length >= this.MAX_CHARS);

		// If there was still text left, append it
		if (text.length > 0) {
			multiLineText += text;
		}
		return multiLineText;
	}
};

/**
 * Disable the tooltip
 */
Tooltip.prototype.disableTooltip = function() {
	this.context.removeClass("hasTooltip");
	if (this._isMobile) {
		this._disableTooltipTouch();
	} else {
		this._disableTooltipMouse();
	}
	this._enabled = false;
};

/**
 * Enable the tooltip
 */
Tooltip.prototype.enableTooltip = function() {
	if (!this._enabled) {
		this.context.addClass("hasTooltip");
		if (this._isMobile) {
			this._enableTooltipTouch();
		} else {
			this._enableTooltipMouse();
		}
		this._enabled = true;
	}
};

/**
 * show the popup after a delay
 */
Tooltip.prototype._delayedShow = function() {
	if (!this._timer) {
		this._timer = setTimeout(
			_.bind(function() {
				this._show();
				setTimeout(this._hide, 2500);
			}, this),
			600
		);
	}
};

/**
 * Hide the Tooltip
 * @override
 */
Tooltip.prototype.hide = function() {
	if (this._timer) {
		clearTimeout(this._timer);
		this._timer = null;
	}

	Popover.prototype.hide.call(this);
};

/**
 * Refreshes the mouse listeners on the context-element
 */
Tooltip.prototype.refereshListeners = function() {
	this.disableTooltip();
	this.enableTooltip();
};

/**
 * Disable the tooltip for mouse interaction
 */
Tooltip.prototype._disableTooltipMouse = function() {
	this.context.off("mouseover", this._show);
	this.context.off("mouseout", this._hide);
};

/**
 * Enable tooltip for mouse interaction
 */
Tooltip.prototype._enableTooltipMouse = function() {
	this.context.on("mouseover", this._show);
	this.context.on("mouseout", this._hide);
};

/**
 * Disable the tooltip for mouse interaction
 */
Tooltip.prototype._disableTooltipTouch = function() {
	this.context.off("touchstart pointerdown", this._touchAndHoldShow);
	this.context.off("touchend touchcancel", this._hide);
	this.context.off("contextmenu", this.contextMenuListener);
};

/**
 * Enable tooltip for mouse interaction
 */
Tooltip.prototype._enableTooltipTouch = function() {
	this.context.on("touchstart pointerdown", this._touchAndHoldShow);
	this.context.on("touchend touchcancel", this._hide);
	this.context.on("contextmenu", this.contextMenuListener);
};

/**
 * Listener to prevent the context menu
 * @param {Event} e
 */
Tooltip.prototype.contextMenuListener = function(e) {
	e.stopPropagation();
	e.preventDefault();
	return false;
};

Tooltip.wcrinit = function(context) {
	if (!(context instanceof $)) {
		context = $(context);
	}
	var body = $("body");
	var text = context.attr("data-tooltip");
	var isMobile = context.attr("data-formobile");
	if (text) {
		new Tooltip(undefined, context, body, body, isMobile, text);
	}
};

export default Tooltip;
