/**
 * @namespace Utils
 */
/**
 * @namespace Utils
 */
import OrderedObject from "@esko/orderedobject";

import _ from "underscore";
/**
 * @memberof Utils
 * @exports Cache
 * @class
 * A simple LRU cache implementation (With loading mechanism)
 * @constructor
 * @param {Number} [cacheSize] - size of the cache (default: 30)
 * @param {Boolean} [executeCallbacksIfInCache] - if true, callbacks will also be executed when the item is available in the cache (default: true)
 */
var Cache = function(cacheSize, executeCallbacksIfInCache) {
	this.store = new OrderedObject();
	this.deletedItems = {};

	this.executeCallbackIfInCache = executeCallbacksIfInCache !== undefined ? executeCallbacksIfInCache : true;

	cacheSize = cacheSize || 30;

	this.getCacheSize = function() {
		return cacheSize;
	};
};

/**
 * Get a property from the cache, or load it if needed
 * @param {String} name - name of the item
 * @param {Function} callback - callback to be called when item is available
 * @returns {Any}
 */
Cache.prototype.get = function(name, callback) {
	var storeItem = this.store.getItemByKey(name);

	// if no cache hit is found, try to see if there is an alternative in the cache
	if (!storeItem) {
		var alternativeName = this.getAlternativeName(name);
		if (alternativeName !== null) {
			// an alternative might have been found
			storeItem = this.store.getItemByKey(alternativeName);
		}
	}

	if (!storeItem) {
		// item was never loaded
		storeItem = {
			name: name,
			item: undefined,
			callbacks: [],
			loaded: false,
			error: false
		};

		// insert in store
		this.store.addItem(name, storeItem);

		// push the callback
		if (callback) {
			storeItem.callbacks.push(callback);
		}

		// start loading
		this._loadItem(name);
	} else {
		// the item is already in the cache in some way
		if (storeItem.loaded) {
			// the item is in the cache and is loaded, maybe we need to fire the callback
			if (callback && this.executeCallbackIfInCache) {
				// we need to execute the callback
				callback(storeItem.item);
			}
		} else {
			// the item is still being loaded, add callback if needed
			if (callback) {
				storeItem.callbacks.push(callback);
			}
		}

		// move the item to the end of the store (so the cache is LRU)
		this.store.moveItemByKeyToEnd(name);
	}

	return storeItem.item;
};

/**
 * This method should be overridden so it loads the requested item
 * When the item is loaded, _registerItem should be called
 * @param {String} name - name of the item
 */
Cache.prototype._loadItem = function(name) {
	console.error("This method should be overriden!");
};

/**
 * Register an item in the cache
 * @param {String} name - name of the item
 * @param {Any|undefined} item - the item, if undefined, the item couldn't be loaded (callbacks will be called with "undefined" as parameter, and item is removed from cache)
 */
Cache.prototype._registerItem = function(name, item) {
	var inCache = true;
	var storeItem = this.store.getItemByKey(name);
	if (storeItem === undefined) {
		inCache = false;
		// item was already removed from cache, check if it is still in deleted
		storeItem = this.deletedItems[name];

		if (storeItem === undefined) {
			throw new Error("Trying to register item that was never requested");
		}

		// remove the item from the deleted items list
		delete this.deletedItems[name];
	}

	// item is still in cache
	storeItem.loaded = true;
	storeItem.item = item;
	storeItem.error = item === undefined;

	if (inCache) {
		// the item is in the cache
		if (storeItem.error) {
			// remove it
			this.store.removeByKey(name);
		} else {
			// make space for the new item that was loaded
			while (this.store.getLength() > this.getCacheSize()) {
				var deleted = this.store.removeByIndex(0);
				// if an item is still loading and has callbacks, mark it as deleted
				if (!deleted.loaded && deleted.callbacks.length > 0) {
					this.deletedItems[deleted.name] = deleted;
				}
			}
		}
	}

	_.each(storeItem.callbacks, function(callback) {
		callback(item);
	});

	storeItem.callbacks = [];
};

/**
 * Remove item from cache
 * @param name
 */
Cache.prototype.remove = function(name) {
	var storeItem = this.store.removeByKey(name);
	if (!storeItem.loaded && storeItem.callbacks.length > 0) {
		this.deletedItems[storeItem.name] = storeItem;
	}
};

/**
 * In this method an alternative name could be generated for an item
 * @param name
 * @returns {String}
 */
Cache.prototype.getAlternativeName = function(name) {
	return null;
};

export default Cache;
