import $ from "jquery";
import _ from "underscore";
var ConvertUtils = {
	/**
	 * Make a capitalized version of the first letter of the given string
	 * @param {String} string - String of which you want to capitalize the first letter
	 * @returns {String} String with a capitalized first letter
	 */
	capitaliseFirstLetter: function(string) {
		return string.charAt(0).toUpperCase() + string.slice(1);
	},

	/**
	 * Count the number of lines and characters in a raw HTML string
	 * @param {String} rawHtmlString - The raw HTML string
	 * @returns {Object} o.lineCount contains the line count, o.charCount the character count
	 */
	getCharacterLineCount: function(rawHtmlString) {
		var lineCount = 1;
		var charCount = 0;

		if (rawHtmlString) {
			var text = this.convertHTMLStringToRaw(rawHtmlString);
			text = text.replace(/\r\n/g, "\n");

			// count lines
			var newLines = 0;
			var n = text.indexOf("\n", 0);
			while (n != -1) {
				lineCount++;
				newLines++;
				n = text.indexOf("\n", n + 1);
			}
			// count characters
			charCount = text.length - newLines;
		}

		return {
			lineCount: lineCount,
			charCount: charCount
		};
	},
	/**
	 * Convert an html string to RAW string (remove all html tags and convert html entitites
	 * @param {String} htmlString
	 * @param {String} [newLineCharacter=\n]  - the character used for separating/indicating newlines in the output
	 * @returns {String} the raw text string
	 */
	convertHTMLStringToRaw: function(htmlString, newLineCharacter) {
		if (newLineCharacter === undefined) {
			newLineCharacter = "\n";
		}
		// remove all white space between tags, except real spaces
		var rawString = htmlString.replace(
			/>[\f\n\r\t\v\u00a0\u1680\u180e\u2000\u2001\u2002\u2003\u2004\u2005\u2006\u2007\u2008\u2009\u200a\u2028\u2029\u202f\u205f\u3000]*([ ]*)[\f\n\r\t\v\u00a0\u1680\u180e\u2000\u2001\u2002\u2003\u2004\u2005\u2006\u2007\u2008\u2009\u200a\u2028\u2029\u202f\u205f\u3000]*</g,
			">$1<"
		);

		// get body only
		rawString = rawString.replace(/.*<body>((\r\n|\n|.)*)<\/body>.*/i, "$1");

		// remove the first blocklevel element
		rawString = rawString.replace(
			/^<(blockquote|body|center|dd|dir|div|dl|dt|form|h1|h2|h3|h4|h5|h6|head|hr|html|isindex|li|menu|noframes|ol|p|pre|td|th|title|ul)[^>]*>/,
			""
		);
		// all elements that cause a line break are replaced by a \n
		rawString = rawString.replace(
			/<(blockquote|body|br|center|dd|dir|div|dl|dt|form|h1|h2|h3|h4|h5|h6|head|hr|html|isindex|li|menu|noframes|ol|p|pre|td|th|title|ul)[^>]*>/g,
			newLineCharacter
		);
		// strip all html elements
		rawString = rawString.replace(/<[^>]*>/g, "");
		// convert spaces
		rawString = rawString.replace(/&nbsp;|&#160;/gi, " ");
		// decode entitites
		rawString = this.decodeHtmlEntities(rawString);

		return rawString;
	},

	/**
	 * Convert a RAW string to xHTML (change \r\n or \n to <br/>)
	 * @param {String} rawString
	 * @returns {String} the xHTML string
	 */
	convertRawStringToHTML: function(rawString) {
		var htmlString = this.encodeHtmlEntities(rawString);

		// replace newlines with breaks
		htmlString = htmlString.replace(/(\r)?\n/g, "<br />");

		return htmlString;
	},
	/**
	 * Decode HTML entities in the HTML string
	 * @param html - HTML string to decode
	 * @returns {String} html parameter with decoded entities
	 */
	decodeHtmlEntities: function(html) {
		return $("<textarea>")
			.html(html)
			.text();
	},

	/**
	 * Encode HTML entities in the HTML string
	 * @param text - text string to encode
	 * @returns {String} text parameter with encoded entities
	 */
	encodeHtmlEntities: function(text) {
		return $("<div>")
			.text(text)
			.html();
	},
	/**
	 * Format a number.
	 * source: http://stackoverflow.com/questions/149055/how-can-i-format-numbers-as-money-in-javascript/
	 * @example format(12345678.9, 2, 3, '.', ',');  // "12.345.678,90"
	 * 			format(123456.789, 4, 4, ' ', ':');  // "12 3456:7890"
	 *			format(12345678.9, 0, 3, '-');       // "12-345-679"
	 * @param   {Number} number - the number to format
	 * @param   {Number} n - the length of decimal
	 * @param   {Number} x - the length of the whole part
	 * @param   {Char} s - the sections delimiter
	 * @param   {Char} c - the decimal delimiter
	 * @returns {String} The formatted number
	 */
	format: function(number, n, x, s, c) {
		var re = "\\d(?=(\\d{" + (x || 3) + "})+" + (n > 0 ? "\\D" : "$") + ")",
			num = number.toFixed(Math.max(0, ~~n));

		return (c ? num.replace(".", c) : num).replace(new RegExp(re, "g"), "$&" + (s || ","));
	}
};

export default ConvertUtils;
