import _ from "underscore";
import ShapeUtils from "@esko/shapeutils";
function drawDashedLine(ctx, x, y, x2, y2, dashArray, zoom) {
	if (!dashArray) {
		dashArray = [10, 5];
	}

	if (zoom !== undefined && zoom !== null) {
		dashArray = _.map(dashArray, function(value) {
			return Math.abs(value / zoom);
		});
	}
	var supportsLinesDash = true;
	if (ctx.setLineDash !== undefined) {
		//check if setLineDash exists and dashed lines are supported
		ctx.setLineDash(dashArray);
	} else if (ctx.mozDash !== undefined) {
		//for older versions of firefox
		ctx.mozDash = dashArray;
	} else {
		supportsLinesDash = false; //dashed lines are not supported
	}

	ctx.save();
	ctx.beginPath();
	var dx = x2 - x,
		dy = y2 - y;
	var len = Math.sqrt(dx * dx + dy * dy);
	var rot = Math.atan2(dy, dx);

	ctx.translate(x, y);
	ctx.moveTo(0, 0);
	ctx.rotate(rot);
	if (supportsLinesDash) {
		ctx.lineTo(len, 0);
	} else {
		var dc = dashArray.length;
		var di = 0,
			draw = true;
		x = 0;
		while (len > x) {
			x += dashArray[di++ % dc];
			if (x > len) {
				x = len;
			}
			if (draw) {
				ctx.lineTo(x, 0);
			} else {
				ctx.moveTo(x, 0);
			}
			draw = !draw;
		}
	}
	ctx.stroke();
	ctx.restore();
}

/**
 * @memberof Utils
 * @exports DashUtils
 */
var DashUtils = {
	/**
	 * Draw a dashed line between two points
	 * @param {CanvasRenderingContext2D} ctx - Canvas context
	 * @param {Number} x - x coordinate of the first point
	 * @param {Number} y - y coordinate of the first point
	 * @param {Number} x2 - x coorindate of the second point
	 * @param {Number} y2 - x coordinate of the second point
	 * @param {Array<Number>} dashArray - dash lenghts
	 */
	drawDash: function(context, x, y, x2, y2, dashArray) {
		var dashCount = dashArray.length;

		var dx = x2 - x,
			dy = y2 - y;
		var xSlope = dx > dy;
		var slope = xSlope ? dy / dx : dx / dy;

		if (slope > 9999) {
			slope = 9999;
		} else if (slope < -9999) {
			slope = -9999;
		}

		var distRemaining = Math.sqrt(dx * dx + dy * dy);
		var dashIndex = 0,
			draw = true;
		while (distRemaining >= 0.1) {
			var dashLength = dashArray[dashIndex++ % dashCount];
			if (dashLength === 0) {
				dashLength = 0.001;
			}
			if (dashLength > distRemaining) {
				dashLength = distRemaining;
			}
			var step = Math.sqrt((dashLength * dashLength) / (1 + slope * slope));
			if (xSlope) {
				x += dx < 0 && dy < 0 ? step * -1 : step;
				y += dx < 0 && dy < 0 ? slope * step * -1 : slope * step;
			} else {
				x += dx < 0 && dy < 0 ? slope * step * -1 : slope * step;
				y += dx < 0 && dy < 0 ? step * -1 : step;
			}
			context[draw ? "lineTo" : "moveTo"](x, y);
			distRemaining -= dashLength;
			draw = !draw;
		}

		context.moveTo(x2, y2);
	},
	/**
	 * Draw a dashed line
	 * @param {Context2D} ctx - canvas context
	 * @param {Number} x - start x
	 *                   @param {Number} y - start y
	 * @param {Number} x2 - end x
	 * @param {Number} y2 - end y
	 * @param {Array.<Number>} dashArray - dash pattern [px draw, px empty, px draw, px empty, ...]
	 */
	drawDashedLine: drawDashedLine,
	/**
	 * Draw a dashed rectangle
	 * @param {Context2D} ctx - canvas context
	 * @param {ShapeUtils.Rect} rect - rectangle to draw
	 * @param {Array.<Number>} dashArray - dash pattern
	 * @param {Number} [zoom] - which zoomlevel is used (use this if you want to define the dasharray in pixels and not world coordinates)
	 */
	drawDashedRect: function(ctx, rect, dashArray, zoom) {
		// we have to subtract/add half the lineWidth to be able to draw sharp lines
		var left = rect.left() + 0.5 * ctx.lineWidth;
		var right = rect.right() - 0.5 * ctx.lineWidth;
		var top = rect.top() + 0.5 * ctx.lineWidth;
		var bottom = rect.bottom() - 0.5 * ctx.lineWidth;

		// top
		drawDashedLine(ctx, left, top, right, top, dashArray, zoom);
		// bottom
		drawDashedLine(ctx, left, bottom, right, bottom, dashArray, zoom);
		// left
		drawDashedLine(ctx, left, top + 0.5 * ctx.lineWidth, left, bottom - 0.5 * ctx.lineWidth, dashArray, zoom);
		// right
		drawDashedLine(ctx, right, top + 0.5 * ctx.lineWidth, right, bottom - 0.5 * ctx.lineWidth, dashArray, zoom);
	}
};
export default DashUtils;
