import _ from "underscore";
/**
 * @class Map a Map is an implementation of the Map datastructure. It can contain unique keys (with a value assigned to it)
 * @constructor
 */
var Map = function() {
	this._map = {};
};

/**
 * Add an entry to the set
 * @param {String} key - the entry name
 * @param {Any} value - the value for the entry (should not be undefined!)
 */
Map.prototype.add = function(key, value) {
	if (value === undefined) {
		throw new Error("value parameter should not be undefined!");
	}

	this._map[key] = value;
};

/**
 * Check if an entry exists in the set
 * @param {String} key - the entry name
 * @returns {Boolean}
 */
Map.prototype.has = function(key) {
	return this._map.hasOwnProperty(key);
};

/**
 * Get the value for an entry
 * @param {String} key - the entry name
 * @returns {Any} the value of the entry, or undefined if entry does not exist
 */
Map.prototype.get = function(key) {
	if (this._map.hasOwnProperty(key)) {
		return this._map[key];
	} else {
		return undefined;
	}
};

/**
 * Remove an entry
 * @param {String} key - the entry name to remove
 * @returns {Object} the value stored for the key
 */
Map.prototype.remove = function(key) {
	var oldValue = this.get(key);
	delete this._map[key];
	return oldValue;
};

/**
 * Clear the set
 */
Map.prototype.clear = function() {
	this._map = {};
};

/**
 * Get a list of entry names
 * @returns {Array.<String>}
 */
Map.prototype.entries = function() {
	return _.keys(this._map);
};

/**
 * Get a list of values
 * @returns {Array.<Any>}
 * @returns
 */
Map.prototype.values = function() {
	return _.values(this._map);
};

/**
 * Get the size of the map
 * @returns {Number}
 */
Map.prototype.size = function() {
	return _.size(this._map);
};

/**
 * Is the map empty
 * @returns {Boolean} True if there are no elements in the map (size = 0), false otherwise
 */
Map.prototype.isEmpty = function() {
	return this.size() === 0;
};

/**
 * Get access to the map object with key/value pairs
 * IMMUTABLE (only use for iterating)
 * @returns {Object}
 */
Map.prototype.asObject = function() {
	return this._map;
};

/**
 * Loop through all key value pairs.
 * @param {Function} callback - a function with two parameters: "key" & "value"
 */
Map.prototype.each = function(callback) {
	_.each(
		_.keys(this._map),
		_.bind(function(key, i) {
			var value = this.get(key);
			callback(key, value);
		}, this)
	);
};

/**
 * Change each key by the transform function and return as new Map
 * @param {Function} transformFunction - Function(key) to transform the key
 * @returns {Map} This map as new Map with all keys transformed by the given function
 */
Map.prototype.transformKeys = function(transformFunction) {
	var newMap = new Map();
	this.each(function(key, value) {
		newMap.add(transformFunction(key), value);
	});
	return newMap;
};

export default Map;
