import EventEmitter2 from "eventemitter2";
import _ from "underscore";
/**
 * Collection of instances for the observables installed on objects on the current page.
 * Components can shooce to use the default event bus or specific private event bus.
 * Collection is here for the singleton pattern.
 */
var observableInstances = {};

/**
 * Observable is a singleton defintion to install an event bus and make the object observable.
 * The constructor of the class in not publically available; only possible to use mixin-method.
 */
var Observable = function(eventBusID) {
	this._eventBusID = eventBusID;

	this.ebus = new EventEmitter2({
		wildcard: true, // should the event emitter use wildcards.
		delimiter: ".", // the delimiter used to segment namespaces,
		newListener: true // if you want to emit the newListener event
	});

	var maxListeners = 0;

	if (process.env.NODE_ENV === "development") {
		// during development we want maximum 20 listeners per channel, in production mode this can be much more
		maxListeners = 20;
	}

	// IE9 fix
	if (!window.console) {
		window.console = {};
	}
	if (!window.console.trace) {
		window.console.trace = function() {};
	}
	//>>excludeEnd("PExclude");

	this.ebus.setMaxListeners(maxListeners);
};

export default {
	/**
	 * Mixin Observable functionality to an object
	 * @param {Object} object - the object in which the Observable functionality should be mixed in
	 * @param {String|Array} ID - the unique identifier of the observable object
	 * @param {String} [eventBusID] - If not provided a new EventBus is created; which instance of event bus should be used for the observable events?
	 */
	mixin: function(object, ID, eventBusID) {
		if (!ID) {
			throw new Error("ID should be defined");
		}

		if (_.isArray(ID)) {
			ID = ID.join(".");
		}

		var eventBus;

		if (eventBusID) {
			if (!observableInstances[eventBusID]) {
				observableInstances[eventBusID] = new Observable(eventBusID);
			}
			eventBus = observableInstances[eventBusID].ebus;
		} else {
			eventBus = new Observable(ID).ebus;
		}

		/**
		 * Subscribe to a channel on the Observable
		 * @param {String} channel - the channel to which to subscribe
		 * @param {Function} callback - the callback called when the channel event occurs
		 */
		object.on = function(channel, callback) {
			arguments[0] = [ID, channel].join(".");
			eventBus.on.apply(eventBus, arguments);
		};

		/**
		 * Subscribe to multiple channels on the Observable
		 * @param {Array<String>} channels - the channels to which to subscribe
		 * @param {Function} callback - the callback called when the channel event occurs
		 */
		object.onAll = function(channels, callback) {
			_.each(channels, function(channel) {
				object.on(channel, callback);
			});
		};

		/**
		 * Subscribe to a channel on the Observable
		 * @param {String} channel - the channel to which to subscribe
		 * @param {Function} callback - the callback called when the channel event occurs
		 */
		object.onAny = function(callback) {
			eventBus.onAny(function() {
				if (arguments[0].indexOf(ID) === 0) {
					arguments[0] = arguments[0].substr(ID.length + 1);
					callback.apply(eventBus, arguments);
				}
			});
		};

		/**
		 * Emit an event on a certain sub channel
		 */
		object._emit = function(channel) {
			arguments[0] = [ID, channel].join(".");
			eventBus.emit.apply(eventBus, arguments);
		};

		/**
		 * Subscribe once to a channel on the Observable (after first event, the subscription is removed)
		 * @param {String} channel - the channel to which to subscribe
		 * @param {Function} callback - the callback called when the channel event occurs
		 */
		object.once = function(channel, callback) {
			arguments[0] = [ID, channel].join(".");
			eventBus.once.apply(eventBus, arguments);
		};

		/**
		 * Subscribe once to multiple channels on the Observable (after first event, the subscription is removed)
		 * @param {Array<String>} channels - the channels to which to subscribe
		 * @param {Function} callback - the callback called when the channel event occurs
		 */
		object.onceAll = function(channels, callback) {
			_.each(channels, function(channel) {
				object.once(channel, callback);
			});
		};

		/**
		 * Remove subscription to a specific channel
		 * callback should be identical as when registered with on or once
		 * @param {String} channel - the channel to which to unsubscribe
		 * @param {Function} callback - the callback called when the channel event occurs
		 */
		object.off = function(channel, callback) {
			arguments[0] = [ID, channel].join(".");
			eventBus.off.apply(eventBus, arguments);
		};

		/**
		 * Remove subscription to a specific list of channels
		 * @param {Array<String>} channels - the channels to which to unsubscribe
		 * @param {Function} callback - the callback called when the channel event occurs
		 */
		object.offAll = function(channels, callback) {
			_.each(channels, function(channel) {
				object.off(channel, callback);
			});
		};

		/**
		 * Remove all observers that where registered on this observable.
		 * NOTE! if the same ID is used amongst other observables (which is a bad idea), these listeners are also removed...)
		 */
		object._destroyObservable = function() {
			eventBus.removeAllListeners([ID, "**"].join("."));
		};

		/**
		 * Get the ID of the observable
		 * @returns {String}
		 */
		object.getObservableID = function() {
			return ID;
		};
	}
};
