import _ from "underscore";
function removeSubscriptions(observer, predicate) {
	// filter out all subscriptions to observable
	var tmp = _.partition(observer._observer_subscriptions, predicate);

	// remove subscriptions
	_.each(tmp[0], observer._stopObserveSubscription);

	// object observer subscription list
	observer._subscriptions = tmp[1];
}

export default {
	/**
	 * Mixin Observer functionality to an object
	 * @param {Object} object - the object in which the Obser functionality should be installed
	 */
	mixin: function(object) {
		object._observer_subscriptions = [];

		/**
		 * Subscribe to a certain channel of an observable
		 * !Note! this automatically binds the callback to the observer (if not needed use _observeUnbound)
		 * @param {Observable} observable - the object to subscribe to
		 * @param {String} channel - the channel to subscribe to
		 * @param {Function} callback - the callback executed when message arrives in the channel. This callback is automatically bound to the observer!
		 * @returns {SubscriptionObject} this object can be used as an argument for _stopObserSubscription
		 */
		object._observe = function(observable, channel, callback) {
			return this._observeUnbound(observable, channel, _.bind(callback, object));
		};

		/**
		 * Subscribe to a list of channels of an observable
		 * !Note! this automatically binds the callback to the observer (if not needed use _observeUnbound)
		 * @param {Observable} observable - the object to subscribe to
		 * @param {Array<String>} channel - the channels to subscribe to
		 * @param {Function} callback - the callback executed when message arrives in the channel. This callback is automatically bound to the observer!
		 */
		object._observeAll = function(observable, channels, callback) {
			_.each(
				channels,
				function(channel) {
					this._observeUnbound(observable, channel, _.bind(callback, object));
				}.bind(this)
			);
		};

		/**
		 * Subscribe to a certain channel of a observable
		 * @param {Observable} observable - the object to subscribe to
		 * @param {String|Array<String>} channel - the channel(s) to subscribe to
		 * @param {Function} callback - the callback executed when message arrives in the channel.
		 * @returns {SubscriptionObject} this object can be used as an argument for _stopObserSubscription
		 */
		object._observeUnbound = function(observable, channel, callback) {
			var subscription = {
				observable: observable,
				channel: channel,
				callback: callback
			};

			this._observer_subscriptions.push(subscription);

			observable.on(channel, callback);

			return subscription;
		};

		/**
		 * Stop a certain specific subscription
		 * @param {SubscriptionObject} subscription - a subscription object as returned by _observe and _observeUnbound
		 */
		object._stopObserveSubscription = function(subscription) {
			subscription.observable.off(subscription.channel, subscription.callback);
		};

		/**
		 * Stop all subscriptions registered by the observer
		 */
		object._stopObserveAll = function() {
			_.each(this._observer_subscriptions, this._stopObserveSubscription);
		};

		/**
		 * Stop observing a specific object
		 * This will remove all subscriptions registered by the observer on the observable object
		 * @param {Observable} observable - to object for which all subscription should be removed
		 */
		object._stopObserveObject = function(observable) {
			removeSubscriptions(this, function(subscription) {
				return subscription.observable === observable;
			});
		};

		/**
		 * Stop observing a specific channel on a specific observable
		 * @param {Observable} observable - the object for which all subscription on the channel should be removed
		 * @param {String} channel - the channel for which subscriptions should be removed
		 */
		object._stopObserveObjectChannel = function(observable, channel) {
			removeSubscriptions(this, function(subscription) {
				return subscription.observable === observable && subscription.channel == channel;
			});
		};

		/**
		 * Remove all subscriptions registerd by the observer on a certain channel
		 * It doesn't matter on which observable the subscription was made
		 * @param {String} channel - the channel for which all subscriptions should be removed
		 */
		object._stopObserveChannel = function(channel) {
			removeSubscriptions(this, function(subscription) {
				return subscription.channel == channel;
			});
		};
	}
};
