import _ from "underscore";
/**
 * Create a new ordered object.
 * @exports OrderedObject
 * @memberof Utils
 * @class
 * An ordered object is an object that supports the advantages of an object (key lookup),
 * while having the items ordered as well (index lookup).
 * @constructor
 */
var OrderedObject = function() {
	// Index -> Key
	this.array = [];
	// Key -> Item
	this.object = {};
};

/**
 * Add the given item for the given key at the given index.
 * @param {String} key - Key of the item
 * @param {Object} item - Any item
 * @param {Number} index - The item is added at the specified index
 */
OrderedObject.prototype.addItemAtIndex = function(key, item, index) {
	if (this.getIndexByKey(key) >= 0) {
		throw new Error("Duplicate key exception");
	}
	this.array.splice(index, 0, key);
	this.object[key] = item;
};

/**
 * Add the given item for the given key. The item is added at the end.
 * @param {String} key - Key of the item
 * @param {Object} item - Any item
 */
OrderedObject.prototype.addItem = function(key, item) {
	if (this.getIndexByKey(key) >= 0) {
		throw new Error("Duplicate key exception");
	}
	this.array.push(key);
	this.object[key] = item;
};

/**
 * If a certain key is already existent, update the item for that key, else add the item in the end of the list
 * @param {String} key - Key of the item
 * @param {Object} item - Any item
 */
OrderedObject.prototype.addOrReplaceItem = function(key, item) {
	if (this.getIndexByKey(key) >= 0) {
		// replace item
		this.object[key] = item;
	} else {
		this.addItem(key, item);
	}
};

/**
 * Sets the order array for the orderedObject, array has as values the keys of the orderedobject
 * @param {Array} newOrder
 * @throws {Error} when newOrder doesn't contain the same elements as the current ordered object keys
 */
OrderedObject.prototype.setOrder = function(newOrder) {
	if (_.difference(this.array, newOrder).length === 0) {
		this.array = newOrder.slice(0);
	} else {
		throw new Error("Invalid array given for order. Should contain the same elements as in the ordered object.");
	}
};

/**
 * Remove the item with the given key
 * @param {String} key - Key of the item
 * @returns {Any} the deleted object, undefined if nothing was removed
 */
OrderedObject.prototype.removeByKey = function(key) {
	var deleted;
	var i = this.getIndexByKey(key);
	if (this.object[key] !== undefined && i >= 0) {
		this.array.splice(i, 1);

		//this.array = _.without(this.array, key);
		deleted = this.object[key];
		delete this.object[key];
	}

	return deleted;
};

/**
 * Remove the item at the given index
 * @param {Number} index - Index of the item
 * @returns {Any} the deleted object, undefined if nothing was removed
 */
OrderedObject.prototype.removeByIndex = function(index) {
	var deleted;

	if (this.array[index]) {
		var k = this.array[index];
		deleted = this.object[k];
		delete this.object[k];
		this.array.splice(index, 1);
	}
	return deleted;
};
/**
 * Omit all items which are not mentioned in keys
 * @param {Array.<String>} keys - only these items can stay
 */
OrderedObject.prototype.omitOthers = function(keys) {
	var toBeOmmitted = _.difference(this.array, keys);

	this.array = _.difference(this.array, toBeOmmitted);
	this.object = _.omit(this.object, toBeOmmitted);
};

/**
 * Clear all the elements in the Ordered Object
 */
OrderedObject.prototype.clear = function() {
	this.array = [];
	this.object = {};
};

/**
 * Get the index of the item with the given key
 * @param key - Key of the item
 * @returns {Number} Index of the item with the given key, -1 if not found
 */
OrderedObject.prototype.getIndexByKey = function(key) {
	return this.array.indexOf(key);
};

/**
 * Get the key of the item by index
 * @param {Number} index - Index of the item
 * @returns {String} Key of the item at the given index, undefined if not defined
 */
OrderedObject.prototype.getKeyByIndex = function(index) {
	return this.array[index];
};

/**
 * Get the key of the item with the given value
 * @param {Object} value - value of the item
 * @returns {String} Key of the item at the given index, undefined if not defined
 */
OrderedObject.prototype.getKeyByValue = function(value) {
	var key;

	_.find(
		this.object,
		function(v, k) {
			if (value === v) {
				key = k;
				return true;
			} else {
				return false;
			}
		},
		this
	);

	return key;
};

/**
 * Get the item at the specified index
 * @param {Number} index - Index of the item
 * @returns {Object} The item at the specified index
 */
OrderedObject.prototype.getItemByIndex = function(index) {
	return this.object[this.array[index]];
};

/**
 * Get the item by key
 * @param {String} key - Key of the item
 * @returns {Object} The item with the specified key
 */
OrderedObject.prototype.getItemByKey = function(key) {
	return this.object[key];
};

/**
 * Check if  a certain key is present in the ordered object
 * @param {String} key - key of the item
 * @returns {Boolean}
 */
OrderedObject.prototype.containsKey = function(key) {
	return this.object[key] !== undefined;
};

/**
 * Get the keys, this list is ordered correctly
 * @returns {Array.<String>} The keys as array ordered.
 */
OrderedObject.prototype.getKeys = function() {
	return this.array.slice(0);
};

/**
 * Get the items as object
 * @returns {Object} The items as object
 */
OrderedObject.prototype.getItemsAsObject = function() {
	return _.clone(this.object);
};

/**
 * Get the number of items
 * @returns {Number} Number of items
 */
OrderedObject.prototype.getLength = function() {
	return this.array.length;
};

/**
 * Move an item from a certain position to another position
 * @param {Number} fromIndex - original index of the item to move
 * @param {Number} toIndex - new index for element
 */
OrderedObject.prototype.moveItemByIndex = function(fromIndex, toIndex) {
	// first splice out the element at from index, and then inject it in the correct location
	var elementToMove = this.array.splice(fromIndex, 1)[0];
	this.array.splice(toIndex, 0, elementToMove);
};

/**
 * Move an item from a certain position to another position
 * @param {String} key - Key of the item
 * @param {Number} toIndex - new index for element
 */
OrderedObject.prototype.moveItemByKeyToIndex = function(key, toIndex) {
	this.moveItemByIndex(this.getIndexByKey(key), toIndex);
};

/**
 * Move an item to the end of the list
 * @param {String} key - Key of the item
 */
OrderedObject.prototype.moveItemByKeyToEnd = function(key) {
	this.moveItemByKeyToIndex(key, this.array.length - 1);
};

/**
 * Gets the next Object, if there isn't a next Object it returns undefined
 * @param {String} key - Key of the item
 * @returns {Object|undefined}
 */
OrderedObject.prototype.getNext = function(key) {
	if (this.hasNext(key)) {
		var index = this.getIndexByKey(key);
		var nextKey = this.array[++index];
		return this.object[nextKey];
	} else {
		return undefined;
	}
};

/**
 * Gets the previous object, if there isn't a previous object it returns undefined
 * @param {String} key - Key of the item
 * @returns {Object|undefined}
 */
OrderedObject.prototype.getPrevious = function(key) {
	if (this.hasPrevious(key)) {
		var index = this.getIndexByKey(key);
		var previousKey = this.array[--index];
		return this.object[previousKey];
	} else {
		return undefined;
	}
};

/**
 * returns true if there is a next Object
 * @param {String} key
 * @returns {Boolean}
 */
OrderedObject.prototype.hasNext = function(key) {
	var index = this.getIndexByKey(key);
	return index !== -1 && ++index < this.array.length;
};

/**
 * returns true if there is a previous object
 * @param {String} key
 * @returns {Boolean}
 */
OrderedObject.prototype.hasPrevious = function(key) {
	var index = this.getIndexByKey(key);
	return index !== -1 && --index >= 0;
};

/**
 * Return an array of the objects as they are ordered
 * @returns {Array}
 */
OrderedObject.prototype.getOrderedObjectList = function() {
	var orderedObjectList = [];
	_.each(
		this.array,
		_.bind(function(key) {
			orderedObjectList.push(this.object[key]);
		}, this)
	);
	return orderedObjectList;
};

/**
 * Get the first item of the ordered object
 * @returns {Any}
 */
OrderedObject.prototype.getFirst = function() {
	return this.getItemByIndex(0);
};

/**
 * Get the last item of the ordered object
 * @returns {Any}
 */
OrderedObject.prototype.getLast = function() {
	return this.getItemByIndex(this.getLength() - 1);
};

export default OrderedObject;
