import $ from "jquery";
import _ from "underscore";

export default {
	/**
	 * Turn a JQuery $.Deferred in a Promises/A compliant promise
	 * @param {$.Deferred} deferred
	 * @return Promise
	 */
	deferred: function(deferred) {
		return new Promise(function(resolve, reject) {
			deferred
				.done(function(data) {
					resolve(data);
				})
				.fail(function(jqXHR, errorStatus, errorString) {
					// check if a response error was shown
					try {
						errorString =
							$(jqXHR.responseText)
								.find("msg")
								.text() || errorString;
					} catch (e) {
						// Use standard error string
					}
					reject(errorString);
				});
		});
	},

	/**
	 * Make the promise "Querable", e.g. you can now ask your promise whether it resolved, failed or was fulfilled.
	 *
	 * @param promise
	 * 			An ordinary promise.
	 * @returns A Queryable promise with the "isResolved", "isRejected" and "isFulfilled" methods.
	 */
	makeQueryable: function(promise) {
		// Inspired from: http://stackoverflow.com/questions/21485545/is-there-a-way-to-tell-if-an-es6-promise-is-fulfilled-rejected-resolved

		// Don't create a wrapper for promises that can already be queried.
		if (promise.isResolved) {
			return promise;
		}

		var isResolved = false;
		var isRejected = false;

		// Observe the promise, saving the fulfillment in a closure scope.
		var result = promise.then(
			function(v) {
				isResolved = true;
				return v;
			},
			function(e) {
				isRejected = true;
				throw e;
			}
		);
		result.isFulfilled = function() {
			return isResolved || isRejected;
		};
		result.isResolved = function() {
			return isResolved;
		};
		result.isRejected = function() {
			return isRejected;
		};
		return result;
	},

	/**
	 * Creates a promise that will resolve when the observable emits the provided event.
	 */
	once: function(observable, event) {
		return new Promise(function(resolve) {
			observable.once(event, resolve);
		});
	},
	formSubmit: function(form, ajaxOptions) {
		var options = _.extend(
			{
				type: "POST",
				url: form.attr("action"),
				data: form.serialize() // serializes the form's elements.
			},
			ajaxOptions
		);

		return this.deferred($.ajax(options));
	}
};
