import _ from "underscore";
/**
 * Create a new point
 * @class Point
 * @namespace ShapeUtils
 * @param {Number} [x = 0] x coordinate of the point
 * @param {Number} [y = 0] y coordinate of the point
 */
var Point = function(x, y) {
	this.x = x ? x : 0;
	this.y = y ? y : 0;
};

/**
 * String representation of the point.
 * @returns {String} "xxy" where the second x means \times
 */
Point.prototype.toString = function() {
	return this.x + "x" + this.y;
};

/**
 * Make a deep copy of this point
 * @returns {ShapeUtils.Point} point with the same x and y
 */
Point.prototype.copy = function() {
	return new Point(this.x, this.y);
};

/**
 * Check if the point equals otherpoint
 * @param {ShapeUtils.Point} otherpoint - point to check equality of this point with
 * @returns {Boolean} True if x and y are the same, false otherwise
 */
Point.prototype.equals = function(otherpoint) {
	return this.x == otherpoint.x && this.y == otherpoint.y;
};

/**
 * Create a size object from this point to the given point to
 * @param {ShapeUtils.Point} to - To point
 * @returns {ShapeUtils.Size} Size object from this point to the given to point calculated as to.x - this.x, to.y-this.y.
 */
Point.prototype.sizeTo = function(to) {
	return new ShapeUtils.Size(to.x - this.x, to.y - this.y);
};

/**
 * Add an offset to this point
 * @param {ShapeUtils.Size} offsetSize - width and height to add as offset
 */
Point.prototype.offset = function(offsetSize) {
	this.x += offsetSize.width;
	this.y += offsetSize.height;
};

/**
 * Add an offset to this point
 * @param {Number} x - Number to add to x field of point
 * @param {Number} y - Number to add to y field of point
 */
Point.prototype.offsetXY = function(x, y) {
	this.x += x;
	this.y += y;
};

/**
 * Check if this point is near another point within a certain margin
 * @param {ShapeUtils.Point} otherpoint - other point to check closeness with
 * @param {Number} margin - max distance (included) between otherpoint and this point
 * @returns {Boolean} True if the other point is within margin distance in absolute value of this point for x AND y, false otherwise.
 */
Point.prototype.near = function(otherpoint, margin) {
	return Math.abs(this.x - otherpoint.x) <= margin && Math.abs(this.y - otherpoint.y) <= margin;
};

/**
 * Get the euclidian distance between to points
 * @param {ShapeUtils.Point} otherpoint
 */
Point.prototype.getDistance = function(otherpoint) {
	return Math.sqrt(Math.pow(this.x - otherpoint.x, 2) + Math.pow(this.y - otherpoint.y, 2));
};

/**
 * Round the x and y values to the nearest integer
 * @returns {ShapeUtils.Point} this point so you can use chaining
 */
Point.prototype.round = function() {
	this.x = Math.round(this.x);
	this.y = Math.round(this.y);
	return this;
};

/**
 * If this Point isn't in/on Rect already the Point will be altered to the closest point on the border of rect. otherwise nothing will happen
 * @param {ShapeUtils.Rect} rect
 */
Point.prototype.containPointInRect = function(rect) {
	if (this.x > rect.normRight()) {
		this.x = rect.normRight();
	} else if (this.x < rect.normLeft()) {
		this.x = rect.normLeft();
	}
	if (this.y > rect.normBottom()) {
		this.y = rect.normBottom();
	} else if (this.y < rect.normTop()) {
		this.y = rect.normTop();
	}
};

/**
 * Create new Size object
 *
 * @class Size
 * @namespace ShapeUtils
 * @param {Number} width - Width of the size object
 * @param {Number} height - Height of the size object
 */
var Size = function(width, height) {
	this.width = width ? width : 0;
	this.height = height ? height : 0;
};

/**
 * Create a string representation of the size object
 * @returns {String} "widthxheight" where the x means \times
 */
Size.prototype.toString = function() {
	return this.width + " x " + this.height;
};

/**
 * Deep copy of this size object
 * @returns {ShapeUtils.Size} size object with the same width and height
 */
Size.prototype.copy = function() {
	return new Size(this.width, this.height);
};

/**
 * Check if this size object equals another one
 * @param {ShapeUtils.Size} other - size object to check equality with
 * @returns {Boolean} True if width and height are equal, false otherwise
 */
Size.prototype.equals = function(other) {
	return this.width == other.width && this.height == other.height;
};

/**
 * Add a width/height to this object's width/height
 * @param {ShapeUtils.Size} size - width and height to add to this object's width and height
 */
Size.prototype.add = function(size) {
	this.width += size.width;
	this.height += size.height;
};

/**
 * Create new Rect(angle) object
 *
 * @class Rect
 * @namespace ShapeUtils
 * @param {ShapeUtils.Point} origin - point object containing x and y
 * @param {ShapeUtils.Size} size - size object containing width and height
 */
var Rect = function(origin, size) {
	this.origin = origin ? origin : new Point();
	this.size = size ? size : new Size();
};

/**
 * Normalizes rects with negative width or height to positive width or height (changes origin and inverts signs of width/height)
 */
Rect.prototype.normalize = function() {
	if (this.size.width < 0) {
		this.origin.x += this.size.width;
		this.size.width = Math.abs(this.size.width);
	}
	if (this.size.height < 0) {
		this.origin.y += this.size.height;
		this.size.height = Math.abs(this.size.height);
	}
};

/**
 * Check if this rectangle contains the given point
 * NOTE PJPR: this probably will go wrong on non normalized rectangles
 * @param {ShapeUtils.Point} pointToCheck - point to check
 * @returns {Boolean} True if the pointToCheck lies in this rectangle, false otherwise
 */
Rect.prototype.containsPoint = function(pointToCheck) {
	var contains = false;
	if (this.origin.x <= pointToCheck.x && pointToCheck.x <= Number(this.origin.x) + Number(this.size.width)) {
		if (this.origin.y <= pointToCheck.y && pointToCheck.y <= Number(this.origin.y) + Number(this.size.height)) {
			contains = true;
		}
	}
	return contains;
};

/**
 * String representation of this rectangle
 * @returns {String} string representation of this rectangle of the format "Origin: xxy size: widthxheight" with x meaning \times
 */
Rect.prototype.toString = function() {
	return "Origin: " + this.origin.toString() + " size: " + this.size.toString();
};

/**
 * Check if this rectangle equals another rectangle within a certain precision
 * @param {ShapeUtils.Rect} r1 - Rectangle to check equality of this rectangle with
 * @param {Number} [precision = 0.000001] - Precision of the equality check
 * @returns {Boolean} True if the difference of x, y, width and height of this rectangle with the respective values of r1 are smaller than precision, false otherwise
 */
Rect.prototype.rectEquals = function(r1, precision) {
	if (r1 === undefined || r1 === null) {
		return false;
	}
	if (precision === undefined) {
		precision = 0.000001;
	}

	return (
		Math.abs(r1.origin.x - this.origin.x) <= precision &&
		Math.abs(r1.origin.y - this.origin.y) <= precision &&
		Math.abs(r1.size.width - this.size.width) <= precision &&
		Math.abs(r1.size.height - this.size.height) <= precision
	);
};

/**
 * Check if two rectangle overlap or intersect
 *
 * @param {ShapeUtils.Rect} r - the rectangle to check intersection with
 * @param {Float} [p] percentage that this rect overlaps r
 * @returns {Boolean} true if there is an overlap or some points of the rectangles intersect
 */
Rect.prototype.rectIntersect = function(r, p) {
	var r1 = r;
	var r2 = this;
	var intersect = !(
		r1.normLeft() > r2.normRight() ||
		r1.normRight() < r2.normLeft() ||
		r1.normTop() > r2.normBottom() ||
		r1.normBottom() < r2.normTop()
	);
	if (p && intersect) {
		var overlap = this.calculateOverlap(r);
		overlap.normalize();
		var areaOverlap = overlap.size.width * overlap.size.height;
		var areaP = r.size.width * r.size.height;
		return areaOverlap / areaP > p;
	} else {
		return intersect;
	}
};

/**
 * Get the right coordinate
 * DANGER this doesn't work correctly for non-normalized rectangles!
 * @returns {Number} The right coordinate calculated as x + width
 */
Rect.prototype.right = function() {
	return this.origin.x + this.size.width;
};

/**
 * Get the left coordinate
 * DANGER this doesn't work correctly for non-normalized rectangles!
 * @returns {Number} The left coordinate calculated as x
 */
Rect.prototype.left = function() {
	return this.origin.x;
};

/**
 * Get the top coordinate
 * DANGER this doesn't work correctly for non-normalized rectangles!
 * @returns {Number} The top coordinate calculated as y
 */
Rect.prototype.top = function() {
	return this.origin.y;
};

/**
 * Get the bottom coordinate
 * DANGER this doesn't work correctly for non-normalized rectangles!
 * @returns {Number} The bottom coordinate calculated as y + height
 */
Rect.prototype.bottom = function() {
	return this.origin.y + this.size.height;
};

/**
 * Get the width of the rectangle
 * DANGER this doesn't work correctly for non-normalized rectangles!
 * @returns {Number} The width of the rectangle
 */
Rect.prototype.width = function() {
	return this.size.width;
};

/**
 * Get the height of the rectangle
 * DANGER this doesn't work correctly for non-normalized rectangles!
 * @returns {Number} The height of the rectangle
 */
Rect.prototype.height = function() {
	return this.size.height;
};

/**
 * Get the top-left coordinate
 * DANGER this doesn't work correctly for non-normalized rectangles!
 * @returns {ShapeUtils.Point} The top-left coordinate calculated as left(), top()
 */
Rect.prototype.topleft = function() {
	return new Point(this.left(), this.top());
};

/**
 * Get the top-right coordinate
 * DANGER this doesn't work correctly for non-normalized rectangles!
 * @returns {ShapeUtils.Point} The top-right coordinate calculated as right(), top()
 */
Rect.prototype.topright = function() {
	return new Point(this.right(), this.top());
};

/**
 * Get the bottom-left coordinate
 * DANGER this doesn't work correctly for non-normalized rectangles!
 * @returns {ShapeUtils.Point} The bottom-left coordinate calculated as left(), bottom()
 */
Rect.prototype.bottomleft = function() {
	return new Point(this.left(), this.bottom());
};

/**
 * Get the bottom-right coordinate
 * DANGER this doesn't work correctly for non-normalized rectangles!
 * @returns {ShapeUtils.Point} The bottom-right coordinate calculated as right(), bottom()
 */
Rect.prototype.bottomright = function() {
	return new Point(this.right(), this.bottom());
};

/**
 * Get the top-center coordinate
 * DANGER this doesn't work correctly for non-normalized rectangles!
 * @returns {ShapeUtils.Point} The top-center coordinate calculated as center().x, top()
 */
Rect.prototype.topcenter = function() {
	return new Point(this.center().x, this.top());
};

/**
 * Get the bottom-center coordinate
 * DANGER this doesn't work correctly for non-normalized rectangles!
 * @returns {ShapeUtils.Point} The bottom-center coordinate calculated as center().x, bottom()
 */
Rect.prototype.bottomcenter = function() {
	return new Point(this.center().x, this.bottom());
};

/**
 * Get the center-left coordinate
 * DANGER this doesn't work correctly for non-normalized rectangles!
 * @returns {ShapeUtils.Point} The center-left coordinate calculated as left(), center().y
 */
Rect.prototype.centerleft = function() {
	return new Point(this.left(), this.center().y);
};

/**
 * Get the center-right coordinate
 * DANGER this doesn't work correctly for non-normalized rectangles!
 * @returns {ShapeUtils.Point} The center-right coordinate calculated as right(), center().y
 */
Rect.prototype.centerright = function() {
	return new Point(this.right(), this.center().y);
};

// ---- following functions return correct top/right/bottom/left values for non-normalized rectangles

/**
 * Right side of the rectangle (x coordinate)
 * works correctly for non-normalized rectangles
 * @returns {Number}
 */
Rect.prototype.normRight = function() {
	if (this.size.width >= 0) {
		return this.origin.x + this.size.width;
	} else {
		return this.origin.x;
	}
};

/**
 * Left side of the rectangle (x coordinate)
 * works correctly for non-normalized rectangles
 * @returns {Number}
 */
Rect.prototype.normLeft = function() {
	if (this.size.width >= 0) {
		return this.origin.x;
	} else {
		return this.origin.x + this.size.width;
	}
};

/**
 * Top side of the rectangle (y coordinate)
 * works correctly for non-normalized rectangles
 * @returns {Number}
 */
Rect.prototype.normTop = function() {
	if (this.size.height >= 0) {
		return this.origin.y;
	} else {
		return this.origin.y + this.size.height;
	}
};

/**
 * Bottom side of the rectangle (y coordinate)
 * works correctly for non-normalized rectangles
 * @returns {Number}
 */
Rect.prototype.normBottom = function() {
	if (this.size.height >= 0) {
		return this.origin.y + this.size.height;
	} else {
		return this.origin.y;
	}
};

/**
 * Get the normalized (absolute) height of the rectangle
 * @returns {Number}
 */
Rect.prototype.normHeight = function() {
	return Math.abs(this.size.height);
};

/**
 * Get the normalized (absolute) width of the rectangle
 * @returns {Number}
 */
Rect.prototype.normWidth = function() {
	return Math.abs(this.size.height);
};

/**
 * Get the center of the rectangle
 * @returns {ShapeUtils.Point} - point containing the center of the rectangle calculated as x + width/2 and y + height/2
 */
Rect.prototype.center = function() {
	return new Point(this.origin.x + this.size.width / 2, this.origin.y + this.size.height / 2);
};

/**
 * Set the center of the rectangle. The x and y coordinates are adapted.
 * @param {ShapeUtils.Point} center - New center of the rectangle
 */
Rect.prototype.setCenter = function(center) {
	this.origin.x = center.x - this.size.width / 2;
	this.origin.y = center.y - this.size.height / 2;
};

/**
 * Check if this rect contains otherRect within a certain margin
 * @param {ShapeUtils.Rect} otherRect - other rectangle that is contained by this rectangle
 * @param {Number} margin - extra margin that is required for the otherRect within this rectangle
 * @returns {Boolean} True if this rectangle contains otherRect, false otherwise.
 */
Rect.prototype.containsRect = function(otherRect, margin) {
	if (!margin) {
		margin = 0;
	}
	var x1 = this.origin.x - margin;
	var y1 = this.origin.y - margin;
	var w1 = this.size.width + margin;
	var h1 = this.size.height + margin;
	var x2 = otherRect.origin.x;
	var y2 = otherRect.origin.y;
	var w2 = otherRect.size.width;
	var h2 = otherRect.size.height;

	if ((w1 | h1 | w2 | h2) < 0) {
		// At least one of the dimensions is negative...
		return false;
	}
	if (x2 < x1 || y2 < y1) {
		return false;
	}
	w1 += x1;
	w2 += x2;
	if (w2 <= x2) {
		// X+W overflowed or W was zero, return false if...
		// either original w or W was zero or
		// x+w did not overflow or
		// the overflowed x+w is smaller than the overflowed X+W
		if (w1 >= x1 || w2 > w1) {
			return false;
		}
	} else {
		// X+W did not overflow and W was not zero, return false if...
		// original w was zero or
		// x+w did not overflow and x+w is smaller than X+W
		if (w1 >= x1 && w2 > w1) {
			return false;
		}
	}
	h1 += y1;
	h2 += y2;
	if (h2 <= y2) {
		if (h1 >= y1 || h2 > h1) {
			return false;
		}
	} else {
		if (h1 >= y1 && h2 > h1) {
			return false;
		}
	}
	return true;
};

/**
 * Make a deep copy of this rectangle
 * @returns {ShapeUtils.Rect} rectangle with the same origin and size
 */
Rect.prototype.copy = function() {
	return new Rect(this.origin.copy(), this.size.copy());
};

/**
 * Update rectangle
 * @param {ShapeUtils.Point} p1 - origin (or new endpoint if only 1 argument)
 * @param {ShapeUtils.Point} p2 - endpoint
 */
Rect.prototype.update = function(p1, p2) {
	if (arguments.length == 1) {
		this.size.width = p1.x - this.origin.x;
		this.size.height = p1.y - this.origin.y;
	} else {
		this.origin.x = p1.x;
		this.origin.y = p1.y;

		this.size.width = p2.x - p1.x;
		this.size.height = p2.y - p2.y;
	}
};

/**
 * Calculate the rectangle that covers the overlap of this rectangle and rectangle b
 * @param {ShapeUtils.Rect} b - rectangle to calculate overlap with
 * @returns {ShapeUtils.Rect} the overlap rectangle
 */
Rect.prototype.calculateOverlap = function(b) {
	var origin = new Point(Math.max(this.normLeft(), b.normLeft()), Math.max(this.normTop(), b.normTop()));
	var endpoint = new Point(Math.min(this.normRight(), b.normRight()), Math.min(this.normBottom(), b.normBottom()));

	var result = new Rect(origin);
	result.update(endpoint);
	return result;
};

/**
 * Calculates the combination of two rectangles
 * @param {ShapeUtils.Rect} b - rectangle to combine this rectangle with
 * @returns {ShapeUtils.Rect} An area that covers this rectangle and rectangle b
 */
Rect.prototype.calculateArea = function(b) {
	var origin = new Point(Math.min(this.normLeft(), b.normLeft()), Math.min(this.normTop(), b.normTop()));
	var endpoint = new Point(Math.max(this.normRight(), b.normRight()), Math.max(this.normBottom(), b.normBottom()));

	var result = new Rect(origin);
	result.update(endpoint);
	return result;
};

/**
 * Returns a normalized rectangle definition
 * @returns {NormalizedRectArray} [x1, y1, x2, y2]
 */
Rect.prototype.getNormalizedRectArray = function() {
	return [this.normLeft(), this.normTop(), this.normRight(), this.normBottom()];
};

/**
 * Create a new Vector. This represents a vector in polar coordinates.
 * Note: A vector 0,0 has length 0 and by default angle 0, if angleWhenZero is set this is the angle for the 0,0 vector
 * @class Vector
 * @namespace ShapeUtils
 * @param {Number} [x = 0] - x coordinate of the Vector
 * @param {Number} [y = 0] - y coordinate of the Vector
 * @param {Number} [angle = 0] - the angle (in radians) when the length is 0
 */
var Vector = function(x, y, angle) {
	this.x = x ? x : 0;
	this.y = y ? y : 0;

	if (this.x == 0 && this.y == 0) {
		this.angleWhenZero = angle ? angle : 0;
	} else {
		this.angleWhenZero = this.getAngleInRadians();
	}
};

/**
 * String representation of the Vector.
 * @returns {String} "x,y"
 */
Vector.prototype.toString = function() {
	return this.x + "," + this.y;
};

/**
 * Make a deep copy of this Vector
 * @returns {Vector} Vector with the same x and y
 */
Vector.prototype.copy = function() {
	return new Vector(this.x, this.y, this.angleWhenZero);
};

/**
 * Check if the Vector equals otherVector
 * @param {ShapeUtils.Vector} otherVector - Vector to check equality of this Vector with
 * @returns {Boolean} True if x and y are the same, false otherwise
 */
Vector.prototype.equals = function(otherVector) {
	return this.x == otherVector.x && this.y == otherVector.y && this.getAngleInRadians() == otherVector.getAngleInRadians();
};

/**
 * Return the length (norm) of the vector
 * @returns {Number} Length of the vector
 */
Vector.prototype.getLength = function() {
	return Math.sqrt(Math.pow(this.x, 2) + Math.pow(this.y, 2));
};

/**
 * Set the length (norm) of the vector
 * It changes the x and y accordingly.
 */
Vector.prototype.setLength = function(newLength) {
	var length = this.getLength();
	var angle = this.getAngleInRadians();
	if (length === 0) {
		this.x = Math.cos(angle) * newLength;
		this.y = Math.sin(angle) * newLength;
	} else {
		var factor = newLength / length;
		this.x = this.x * factor;
		this.y = this.y * factor;
	}

	this.angleWhenZero = angle;
};

/**
 * Get the angle of the vector in degrees in [-180, 180]
 * @returns {Number} Angle of the vector in degrees in [-180, 180]
 */
Vector.prototype.getAngleInDegrees = function() {
	return (this.getAngleInRadians() * 180) / Math.PI;
};

/**
 * Get the angle of the vector in radians
 * @returns {Number} Angle of the vector in radians in [-PI, PI]
 */
Vector.prototype.getAngleInRadians = function() {
	if (this.x === 0 && this.y === 0) {
		return this.angleWhenZero;
	}
	return Math.atan2(this.y, this.x);
};

/**
 * Set the angle of the vector in degrees.
 * Keep the length the same.
 * @param angleInDegrees [-180, 180]
 */
Vector.prototype.setAngleInDegrees = function(angleInDegrees) {
	this.setAngleInRadians((angleInDegrees * Math.PI) / 180);
};

/**
 * Set the angle of the vector in radians
 * Keep the length the same.
 * @param angleInRadians [-PI, PI]
 */
Vector.prototype.setAngleInRadians = function(angleInRadians) {
	var length = this.getLength();

	this.angleWhenZero = angleInRadians;
	this.x = length * Math.cos(angleInRadians);
	this.y = length * Math.sin(angleInRadians);
};

/**
 * ShapeUtils contains utils for geometrical shapes (Point, Size, Rect)
 *
 * @exports ShapeUtils
 * @borrows Point as Point
 * @borrows Size as Size
 * @borrows Rect as Rect
 */
var ShapeUtils = {
	Point: Point,
	Size: Size,
	Rect: Rect,
	Vector: Vector,

	/**
	 * Create a new Rect from x,y, width and height
	 * @param {Number} x
	 * @param {Number} y
	 * @param {Number} width
	 * @param {Number} height
	 * @returns {ShapeUtils.Rect}
	 */
	rectMake: function(x, y, width, height) {
		return new Rect(new Point(x, y), new Size(width, height));
	},
	/**
	 * Create a ShapeUtils.Rect from a normalized rect array ([x1, y1, x2,y2])
	 * @param {NormalizedRectArray} normalizedRect
	 * @returns {ShapeUtils.Rect}
	 */
	rectFromNormalizedRectArray: function(normalizedRect) {
		return new Rect(
			new Point(normalizedRect[0], normalizedRect[1]),
			new Size(normalizedRect[2] - normalizedRect[0], normalizedRect[3] - normalizedRect[1])
		);
	},
	/**
	 * Calculate the bounding box of two normalized rects
	 * @param {NormalizedRectArray} rectA
	 * @param {NormalizedRectArray} rectB
	 * @returns {NormalizedRectArray} the bounding box for the two rects
	 */
	calculateBoundingBoxRectArrays: function(rectA, rectB) {
		return [Math.min(rectA[0], rectB[0]), Math.min(rectA[1], rectB[1]), Math.max(rectA[2], rectB[2]), Math.max(rectA[3], rectB[3])];
	},
	/**
	 * Check if two normalized rectangles are equal
	 * @param {NormalizedRectArray} rectA
	 * @param {NormalizedRectArray} rectB
	 * @returns {Boolean} true if rectangles equal
	 */
	normalizedRectEquals: function(rectA, rectB) {
		return rectA[0] === rectB[0] && rectA[1] === rectB[1] && rectA[2] === rectB[2] && rectA[3] === rectB[3];
	},
	/**
	 * Create a new rectangle from the overlap of two rectangles
	 * @param {ShapeUtils.Rect} r1
	 * @param {ShapeUtils.Rect} r2
	 * @returns {ShapeUtils.Rect}
	 */
	rectIntersection: function(r1, r2) {
		var x1, x2, y1, y2;

		x1 = Math.max(r1.left(), r2.left());
		y1 = Math.max(r1.top(), r2.top());
		x2 = Math.min(r1.right(), r2.right());
		y2 = Math.min(r1.bottom(), r2.bottom());

		return this.rectMake(x1, y1, x2 - x1, y2 - y1);
	},

	/**
	 * Check if a rectangle is empty/null
	 * @param {ShapeUtils.Rect} r
	 * @returns {Boolean} true if the rectangle is positioned at (0,0) and has a size of (0,0)
	 */
	rectIsNull: function(r) {
		return r.origin.x == 0 && r.origin.y == 0 && r.size.width == 0 && r.size.height == 0;
	}
};

export default ShapeUtils;
