import _ from "underscore";
import ShapeUtils from "@esko/shapeutils";
/**
 * Create a new tranformation object
 * See http://www.willamette.edu/~gorr/classes/GeneralGraphics/Transforms/transforms2d.htm#Translations
 * and http://docs.geotools.org/latest/userguide/tutorial/affinetransform.html
 *
 * [a c e]
 * [b d f]
 * [0 0 1]
 *
 * @param {Number} a
 * @param {Number} b
 * @param {Number} c
 * @param {Number} d
 * @param {Number} e
 * @param {Number} f
 */
var Transformation = function(a, b, c, d, e, f) {
	this.a = a !== undefined ? a : 1;
	this.b = b !== undefined ? b : 0;
	this.c = c !== undefined ? c : 0;
	this.d = d !== undefined ? d : 1;
	this.e = e !== undefined ? e : 0;
	this.f = f !== undefined ? f : 0;
};

/**
 * Sets this transform to the matrix specified by the 6 numbers.
 *
 * [a c e]
 * [b d f]
 * [0 0 1]
 *
 * @param {Number} a
 * @param {Number} b
 * @param {Number} c
 * @param {Number} d
 * @param {Number} e
 * @param {Number} f
 */
Transformation.prototype.setTransform = function(a, b, c, d, e, f) {
	this.a = a ? a : 1;
	this.b = b ? b : 0;
	this.c = c ? c : 0;
	this.d = d ? d : 1;
	this.e = e ? e : 0;
	this.f = f ? f : 0;
};

/**
 * Create a copy of this transformation
 * @returns {Transformation} Copy of this transformation with same a,b,c,d,e,f
 */
Transformation.prototype.copy = function() {
	return new Transformation(this.a, this.b, this.c, this.d, this.e, this.f);
};

/**
 * Check if this transformation equals another transformation
 * @param {Transformation} otherTransformation - Other transformation
 * @returns {Boolean} True if a,b,c,d,e,f are equal, false otherwise
 */
Transformation.prototype.equals = function(otherTransformation) {
	return (
		this.a == otherTransformation.a &&
		this.b == otherTransformation.b &&
		this.c == otherTransformation.c &&
		this.d == otherTransformation.d &&
		this.e == otherTransformation.e &&
		this.f == otherTransformation.f
	);
};

/**
 * Create the inverse transformation of this transformation.
 * @returns {Transformation} The inverse transformation of this transformation.
 */
Transformation.prototype.createInverse = function() {
	var inverse = this.copy();

	var d = 1 / (this.a * this.d - this.b * this.c);
	var m0 = this.d * d;
	var m1 = -this.b * d;
	var m2 = -this.c * d;
	var m3 = this.a * d;
	var m4 = d * (this.c * this.f - this.d * this.e);
	var m5 = d * (this.b * this.e - this.a * this.f);
	inverse.a = m0;
	inverse.b = m1;
	inverse.c = m2;
	inverse.d = m3;
	inverse.e = m4;
	inverse.f = m5;

	return inverse;
};

/**
 * Sets this transform's scaling part. The matrix representing this transform becomes:
 *          [   sx   c    e   ]
 *          [   b    sy   f   ]
 *          [   0    0    1   ]
 * @param {Number} sx
 * @param {Number} sy
 */
Transformation.prototype.setScale = function(sx, sy) {
	this.a = sx;
	this.d = sy;
};

/**
 * Scale this transformation.
 * The matrix representing this transform becomes:
 *          [   a*sx   c*sy   e   ]
 *          [   b*sx   d*sy   f   ]
 *          [   0      0      1   ]
 * @param {Number} sx
 * @param {Number} sy
 */
Transformation.prototype.scale = function(sx, sy) {
	this.a *= sx;
	this.b *= sx;
	this.c *= sy;
	this.d *= sy;
};

/**
 * Prescale the transformation
 * The matrix representing this transform becomes:
 *          [   a*sx   c*sy   e*sx   ]
 *          [   b*sx   d*sy   f*sy   ]
 *          [   0      0      1      ]
 * @param {Number} sx
 * @param {Number} sy
 */
Transformation.prototype.preScale = function(sx, sy) {
	this.a *= sx;
	this.b *= sx;
	this.e *= sx;
	this.c *= sy;
	this.d *= sy;
	this.f *= sy;
};

/**
 * Sets this transform's translation part.
 * The matrix representing this transform becomes:
 *           [   a    c    dx  ]
 *           [   b    d    dy  ]
 *           [   0    0    1   ]
 * @param {Number} dx
 * @param {Number} dy
 */
Transformation.prototype.setTranslate = function(dx, dy) {
	this.e = dx;
	this.f = dy;
};

/**
 * Translate this transformation.
 * The matrix representing this transform becomes:
 *           [   a    c    e += dx*a+dy*c  ]
 *           [   b    d    f += dx*b+dy*d  ]
 *           [   0    0    1   ]
 * @param {Number} dx
 * @param {Number} dy
 */
Transformation.prototype.translate = function(dx, dy) {
	this.e += dx * this.a + dy * this.c;
	this.f += dx * this.b + dy * this.d;
};

/**
 * Pretranslate the transformation
 * The matrix representing this transform becomes:
 *          [   a   c    e+=dx   ]
 *          [   b   d    d+=dy   ]
 *          [   0   0    1       ]
 * @param {Number} dx
 * @param {Number} dy
 */
Transformation.prototype.preTranslate = function(dx, dy) {
	this.e += dx;
	this.f += dy;
};

/**
 * Get a rotation transformation instance:
 * [ cos(theta) -sin(theta) x-x*cos+y*sin ]
 * [ sin(theta)  cos(theta) y-x*sin-y*cos ]
 * [     0           0            1       ]
 * @param {Number} theta - The rotation angle (positive in radians)
 * @param {Number} [x = 0] - The x coordinate of the pivot point
 * @param {Number} [y = 0] - The y coordinate of the pivot point
 * @returns {Transformation}
 */
Transformation.prototype.rotationTransformation = function(theta, x, y) {
	x = x === undefined ? 0 : x;
	y = y === undefined ? 0 : y;
	var cos = Math.cos(theta);
	var sin = Math.sin(theta);
	return new Transformation(cos, sin, -sin, cos, x - x * cos + y * sin, y - x * sin - y * cos);
};

/**
 * Get a translation transformation instance:
 * [ 1 0 dx ]
 * [ 0 1 dy ]
 * [ 0 0 1  ]
 * @param {Number} dx - the x translation distance
 * @param {Number} dy - the y translation distance
 * @returns {Transformation}
 */
Transformation.prototype.translateTransformation = function(dx, dy) {
	return new Transformation(1, 0, 0, 1, dx, dy);
};

/**
 * Rotate this transformation
 * @param {Number} theta - the rotation angle (in radians)
 * @param {Number} [x = 0] - The x coordinate of the pivot point
 * @param {Number} [y = 0] - The y coordinate of the pivot point
 */
Transformation.prototype.rotate = function(theta, x, y) {
	var rotationMatrix = this.rotationTransformation(theta, x, y);
	this.concatenate(rotationMatrix);
};

/**
 * Pre-rotate this transformation
 * @param {Number} theta - the rotation angle (in radians)
 * @param {Number} [x = 0] - The x coordinate of the pivot point
 * @param {Number} [y = 0] - The y coordinate of the pivot point
 * @returns
 */
Transformation.prototype.preRotate = function(theta, x, y) {
	return this.preConcatenate(this.rotationTransformation(theta, x, y));
};

/**
 * Get a flip transformation instance:
 * [-1  0  x ]
 * [ 0  1  y ]
 * [ 0  0  1 ]
 * @param {Number} x
 * @param {Number} y
 * @returns {Transformation}
 */
Transformation.prototype.flipTransformation = function(x, y) {
	return new Transformation(-1, 0, 0, 1, x, y);
};

/**
 * Flip this transformation
 * @param {Number} x
 * @param {Number} y
 */
Transformation.prototype.flip = function(x, y) {
	this.concatenate(this.flipTransformation(x, y));
};

/**
 * preFlip this transformation
 * @param {Number} x
 * @param {Number} y
 */
Transformation.prototype.preFlip = function(x, y) {
	this.preConcatenate(this.flipTransformation(x, y));
};

/**
 * Create a string representation of this matrix
 * @returns {String}
 */
Transformation.prototype.toString = function() {
	return "Transformation[[" + this.a + ", " + this.c + ", " + this.e + "], [" + this.b + ", " + this.d + ", " + this.f + "]]";
};

/**
 * Concatenate the transformation tx to this transformation:
 * In matrix form:
 * [ this ] = [ this ] x [ tx ]
 * @param {Transformation} tx - transformation to preconcatenate
 */
Transformation.prototype.concatenate = function(tx) {
	var m11 = this.a * tx.a + this.c * tx.b;
	var m12 = this.b * tx.a + this.d * tx.b;

	var m21 = this.a * tx.c + this.c * tx.d;
	var m22 = this.b * tx.c + this.d * tx.d;

	var dx = this.a * tx.e + this.c * tx.f + this.e;
	var dy = this.b * tx.e + this.d * tx.f + this.f;

	this.a = m11;
	this.b = m12;
	this.c = m21;
	this.d = m22;
	this.e = dx;
	this.f = dy;
};

/**
 * Preconcatenate tx to this transformation:
 * In matrix form:
 * [ this ] = [ tx ] x [ this ]
 * @param {Transformation} tx - transformation to preconcatenate
 * @returns {Transformation} this
 */
Transformation.prototype.preConcatenate = function(tx) {
	var n00 = tx.a * this.a + tx.c * this.b;
	var n01 = tx.a * this.c + tx.c * this.d;
	var n02 = tx.a * this.e + tx.c * this.f + tx.e;
	var n10 = tx.b * this.a + tx.d * this.b;
	var n11 = tx.b * this.c + tx.d * this.d;
	var n12 = tx.b * this.e + tx.d * this.f + tx.f;

	this.a = n00;
	this.c = n01;
	this.e = n02;
	this.b = n10;
	this.d = n11;
	this.f = n12;

	return this;
};

/**
 * Transform a point according to this transformation
 * @param {ShapeUtils.Point} point - point to transform
 * @returns {ShapeUtils.Point} transformed point
 */
Transformation.prototype.transformedPoint = function(point) {
	var x = point.x;
	var y = point.y;
	var px = x * this.a + y * this.c + this.e;
	var py = x * this.b + y * this.d + this.f;
	return new ShapeUtils.Point(px, py);
};

/**
 * Pretransform a point according to this transformation
 * @param {ShapeUtils.Point} point
 * @returns {ShapeUtils.Point}
 */
Transformation.prototype.preTransformedPoint = function(point) {
	return new ShapeUtils.Point(this.a * point.x + this.b * point.y + this.e, this.c * point.x + this.d * point.y + this.f);
};

/**
 * Transform a size according to this transformation
 * @param {ShapeUtils.Size} size - size to transform
 * @returns {ShapeUtils.Size} transformed size
 */
Transformation.prototype.transformedSize = function(size) {
	var p1 = this.transformedPoint(new ShapeUtils.Point());
	var p2 = this.transformedPoint(new ShapeUtils.Point(size.width, size.height));
	var s = new ShapeUtils.Size(p2.x - p1.x, p2.y - p1.y);

	return s;
};

/**
 * Transform a rectangle according to thsi transformation
 * @param {ShapeUtils.Rect} rect - rect to transform
 * @returns {ShapeUtils.Rect} transformed rect
 */
Transformation.prototype.transformedRect = function(rect) {
	var worldRect = rect.copy();

	var topleft = this.transformedPoint(worldRect.origin);
	var p1x = topleft.x;
	var p1y = topleft.y;
	var p2x = topleft.x;
	var p2y = topleft.y;

	var topright = this.transformedPoint(new ShapeUtils.Point(worldRect.right(), worldRect.top()));
	p1x = Math.min(p1x, topright.x);
	p1y = Math.min(p1y, topright.y);
	p2x = Math.max(p2x, topright.x);
	p2y = Math.max(p2y, topright.y);

	var bottomleft = this.transformedPoint(new ShapeUtils.Point(worldRect.left(), worldRect.bottom()));
	p1x = Math.min(p1x, bottomleft.x);
	p1y = Math.min(p1y, bottomleft.y);
	p2x = Math.max(p2x, bottomleft.x);
	p2y = Math.max(p2y, bottomleft.y);

	var bottomright = this.transformedPoint(new ShapeUtils.Point(worldRect.right(), worldRect.bottom()));
	p1x = Math.min(p1x, bottomright.x);
	p1y = Math.min(p1y, bottomright.y);
	p2x = Math.max(p2x, bottomright.x);
	p2y = Math.max(p2y, bottomright.y);

	var p1 = new ShapeUtils.Point(p1x, p1y);
	var p2 = new ShapeUtils.Point(p2x, p2y);

	worldRect.origin = p1;
	worldRect.size = new ShapeUtils.Size(p2.x - p1.x, p2.y - p1.y);

	return worldRect;
};

/**
 * Get the scale x
 * @returns {Number} scale x calculated as Math.sqrt(a^2 + b^2)
 */
Transformation.prototype.getScaleX = function() {
	return Math.sqrt(this.a * this.a + this.b * this.b);
};

/**
 * Get the scale y
 * @returns {Number} scale y calculated as Math.sqrt(c^2 + d^2)
 */
Transformation.prototype.getScaleY = function() {
	return Math.sqrt(this.c * this.c + this.d * this.d);
};

/**
 * Get the rotation that is represented by this transformation
 * @returns {Number} the rotation in degrees
 */
Transformation.prototype.getRotation = function() {
	return Math.atan2(this.c, this.d);
};

export default Transformation;
