import UndoRedoController from "./UndoRedoController";
/**
 * Implementation of a simple Undo Redo controller
 * Using the Command pattern it is possible to execute commands that can be undone or redone
 *
 * If Commands are async and return a promise, the next command is only executed once the promise is resolved
 *
 * @exports AsyncUndoRedoController
 * @class
 *
 * @mixes Observable
 * @param {Number} [stackSize=1000] stackSize - maximum size of the undo stack. If more actions are added, the oldes undo actions are removed from the stack
 */
var AsyncUndoRedoController = function(stackSize) {
	UndoRedoController.call(this, stackSize);

	this._currentOperation = Promise.resolve();
};

AsyncUndoRedoController.prototype = Object.create(UndoRedoController.prototype);
AsyncUndoRedoController.prototype.constructor = AsyncUndoRedoController;

AsyncUndoRedoController.prototype._synchronizedOperation = function(operation) {
	this._currentOperation = this._currentOperation.then(operation.bind(this));
	return this._currentOperation;
};

/**
 * Execute a Command, this clears the redo stack
 * @param {Command} command - the "do()" method of the command is executed, and the command is added to the undo stack
 * @param {Boolean} [skipDo=false] - if true, the command.executeDo() is not called automatically
 * @param {Boolean} [doNotClearRedoStack = false] - if true, the redoStack will not be emptied !! USE WITH CAUTION !!
 * @returns {Any} the result of the command execution
 */
AsyncUndoRedoController.prototype.execute = function(command, skipDo, doNotClearRedoStack) {
	return this._synchronizedOperation(function() {
		return UndoRedoController.prototype.execute.call(this, command, skipDo, doNotClearRedoStack);
	});
};

/**
 * Undo the last executed command, the command is added to the redo stack
 * @returns {Any} the result of the undo command execution
 */
AsyncUndoRedoController.prototype.undo = function() {
	return this._synchronizedOperation(function() {
		return UndoRedoController.prototype.undo.call(this);
	});
};

/**
 * Redo the command that was last undone. The command is added to the undo stack
 * @returns {Any} the result of the redo command execution
 */
AsyncUndoRedoController.prototype.redo = function() {
	return this._synchronizedOperation(function() {
		return UndoRedoController.prototype.redo.call(this);
	});
};

/**
 * Clear the Undo Redo stack
 */
AsyncUndoRedoController.prototype.clear = function() {
	return this._synchronizedOperation(function() {
		return UndoRedoController.prototype.clear.call(this);
	});
};

export default AsyncUndoRedoController;
