import Observable from "@esko/observable";
import _ from "underscore";
/**
 * Implementation of a simple Undo Redo controller
 * Using the Command pattern it is possible to execute commands that can be undone or redone
 *
 * @exports UndoRedoController
 * @class
 *
 * @mixes Observable
 * @param {Number} [stackSize=1000] stackSize - maximum size of the undo stack. If more actions are added, the oldes undo actions are removed from the stack
 */
var UndoRedoController = function(stackSize) {
	Observable.mixin(this, "undoRedoManager");

	if (!stackSize) {
		stackSize = 1000;
	}

	this._stackSize = stackSize;
	this._stackSizeLimitReached = false;
	this._undoStack = [];
	this._redoStack = [];
};

/**
 * Execute a Command, this clears the redo stack
 * @param {Command} command - the "do()" method of the command is executed, and the command is added to the undo stack
 * @param {Boolean} [skipDo=false] - if true, the command.executeDo() is not called automatically
 * @param {Boolean} [doNotClearRedoStack = false] - if true, the redoStack will not be emptied !! USE WITH CAUTION !!
 * @returns {Any} the result of the command execution
 */
UndoRedoController.prototype.execute = function(command, skipDo, doNotClearRedoStack) {
	if (!doNotClearRedoStack) {
		this._redoStack = [];
	}
	this._addToUndoStack(command);

	var result;
	if (!skipDo) {
		result = command.executeDo();
	}

	this._emit("execute", command);
	this._emit("changed");

	return result;
};

/**
 * Undo the last executed command, the command is added to the redo stack
 * @returns {Any} the result of the undo command execution
 */
UndoRedoController.prototype.undo = function() {
	var command = this._undoStack.pop();
	var result;
	if (command) {
		this._redoStack.push(command);
		result = command.executeUndo();
		this._emit("undo", command);
		this._emit("changed");
	}

	return result;
};

/**
 * Redo the command that was last undone. The command is added to the undo stack
 * @returns {Any} the result of the redo command execution
 */
UndoRedoController.prototype.redo = function() {
	var command = this._redoStack.pop();
	var result;
	if (command) {
		this._addToUndoStack(command);
		result = command.executeDo();
		this._emit("redo", command);
		this._emit("changed");
	}

	return result;
};

/**
 * Add a command to undo stack
 * @param {Command} command to add to the undo stack
 */
UndoRedoController.prototype._addToUndoStack = function(command) {
	if (this._undoStack.length >= this._stackSize) {
		this._undoStack = _.last(this._undoStack, this._stackSize - 1);
		this._stackSizeLimitReached = true;
	}

	this._undoStack.push(command);
};

/**
 * Look at the next undoable command
 * @param {Number} [peekLevel=0] - peek level (to look deeper down the stack)
 * @returns {Command}
 */
UndoRedoController.prototype.peekUndo = function(peekLevel) {
	if (typeof peekLevel === "undefined") {
		peekLevel = 0;
	}
	return this._undoStack[this._undoStack.length - 1 - peekLevel];
};

/**
 * Look at the next redoable command
 * @param {Number} [peekLevel=0] - peek level (to look deeper down the stack)
 * @returns {Command}
 */
UndoRedoController.prototype.peekRedo = function(peekLevel) {
	if (typeof peekLevel === "undefined") {
		peekLevel = 0;
	}
	return this._redoStack[this._redoStack.length - 1 - peekLevel];
};

/**
 * Get the number of undo actions on the stack
 * @returns {Number}
 */
UndoRedoController.prototype.undoCount = function() {
	return this._undoStack.length;
};

/**
 * Get the number of redo actions on the stack
 * @returns {Number}
 */
UndoRedoController.prototype.redoCount = function() {
	return this._redoStack.length;
};

/**
 * Get the Undo Stack
 * !!IMMUTABLE!!! don't change the return Array
 * @returns {Array.<Command>}
 */
UndoRedoController.prototype.getUndoStack = function() {
	return this._undoStack;
};
/**
 * Get the Redo Stack
 * !!IMMUTABLE!!! don't change the return Array
 * @returns {Array.<Command>}
 */
UndoRedoController.prototype.getRedoStack = function() {
	return this._redoStack;
};

/**
 * Clear the Undo Redo stack
 */
UndoRedoController.prototype.clear = function() {
	this._undoStack = [];
	this._redoStack = [];

	this._emit("clear");
	this._emit("changed");
};

/**
 * If the maximum stack size was reached (and undo commands where thrown away), this returns true
 * @returns {Boolean}
 */
UndoRedoController.prototype.didReachStackSizeLimit = function() {
	return this._stackSizeLimitReached;
};
export default UndoRedoController;
