import LineworkWrapper from "@esko/web-viewer/utils/LineworkWrapper";
import LineworkPoint from "@esko/web-viewer/utils/LineworkPoint";
import ViewerUtils from "@esko/web-viewer/utils/ViewerUtils";
import GEOUtils from "@esko/web-viewer/utils/GEOUtils";
import DrawUtils from "@esko/web-viewer/utils/DrawUtils";
import AnnotationConstants from "@esko/web-viewer/annotation/AnnotationConstants";
import PrivateChannel from "@esko/web-viewer/core/PrivateChannel";

import ShapeUtils from "@esko/shapeutils";
import XMLUtils from "@esko/xmlutils";
import StringSet from "@esko/stringset";
import _ from "underscore";
/**
 * @class Represents an annotation. It includes the linework and the data of the annotation.
 * @exports Annotation
 * @memberof AnnotationModule
 * @augments Objects.LineworkWrapper
 * @constructor
 * @param {Core.ViewerFacade} f
 * @param {String} originType - The origin of the annotation, defaults to "user", which indicates a user generated annotation
 */
var Annotation = function(f, originType) {
	LineworkWrapper.call(this);
	PrivateChannel.installTo(this, f, "Annotation");
	this.f = f;

	/** creation date */
	this.creationdate = new Date();
	/** modification date */
	this.date = new Date();
	/** name object of the annotation (will be used to construct username/firstname/lastname */
	this.name = null;
	/** page number */
	this.page = 0;
	/** subject of the annotation */
	this.subject = "";
	/** title */
	this.title = "";
	/** content as a string (without BODY tag) */
	this.content = "";

	/** origin of the annotation, this is derived from the xfdf file the annotations originates from
	 *  AnnotationConstants.AnnotationOriginTypes has the set of the supported types
	 */
	this.originType = originType || AnnotationConstants.AnnotationOriginTypes.user.id;

	/**
	 * The 3D view camera position
	 * @example
	 * {distance: 200.5,
	 * azimuth: 46.6,
	 * elevation: -52}
	 */
	this._view3DPosition = null;

	/** user name */
	this.username = "";
	/** first name */
	this.firstname = "";
	/** last name */
	this.lastname = "";
	/** message number. This is generated by the server. If it's a non-saved annotation, it will be temporarily assigned to the highest number */
	this.messagenumber = -1;
	/** Type of annotation {AnnotationConstants.AnnotationType} */
	this.annottype = null;

	/** parent {Annotation} object. In case of inreplyto or state or link */
	this.parent = null;

	/** Array of replies to this annotation */
	this.replies = [];
	/** {Annotation} the annotation that this annotation is in reply to. If null, there is no reply. The actual annotation object should be in the "parent" object */
	this.inreplyto = null;

	/** Array of links to this annotation */
	this.links = [];
	/** {Annotation} Indication if the annotation is linked to another annotation (will be in the parent object) */
	this.linkedto = null;

	/** Is this annotation saved? */
	this.saved = false;
	/** Is this annotation new (never saved on the server)? */
	this.isNew = true;

	/** Array of states for this annotation -> {Annotation} objects */
	this.states = [];
	/** State name if the annotation is a state-object */
	this.state = null;
	/** {String } State model, attribute in the annotation xml, if the annotation is a state-object */
	this.statemodel = null;

	/** {String} - attribute flags in the xml (by default just 'print') */
	this.flags = "print";

	/** Rectangle of the popup - calculated from bottom left corner*/
	this.popuprect = new ShapeUtils.Rect();

	/** Page of the popup */
	this.popuppage = 0;
	/** Indication if the popup is/was opened when saved */
	this.popupIsOpen = false;
	/** this is the content of the reply while it is not yet saved */
	this.unsavedReplyText = null;
	/** Modified content of the annotation **/
	this.modifiedContent = null;
	/** Original annotation color, modified sits in this.annotationcolor **/
	this.originalAnnotationColor = null;

	/** keep track of what modifications have been done*/
	this._modifications = new StringSet();

	/** {ShapeUtils.Point} the scaling of the annotation, null if there is no scaling */
	this._scaling = null;

	/**
	 * If the rectangle is changed, set the annotation on modified
	 */
	this._override("setRect", function(rect, flip) {
		this._super(rect, flip);
		this.setModified("LINEWORK");
	});

	/**
	 * If the annotation is moved, set the annotation on modified
	 */
	this._override("move", function(size) {
		this._super(size);
		if (size.width !== 0 || size.height !== 0) {
			this.setModified("LINEWORK");
		}
	});

	/**
	 * If annotation's start and end point changed, set the annotation on modified
	 */
	this._override("setStartEnd", function(start, end) {
		this._super(start, end);
		this.setModified("LINEWORK");
	});
};
// inherit
Annotation.prototype = Object.create(LineworkWrapper.prototype);
// correct the constructor pointer
Annotation.prototype.constructor = Annotation;

/**
 * @returns {AnnotationConstants.AccessRight} access right object of the annotation
 */
Annotation.prototype.getPermission = function() {
	if (this.f.getCurrentUser().getUsername() == this.username) {
		return AnnotationConstants.AccessRight.READ_WRITE;
	} else {
		return AnnotationConstants.AccessRight.READ_ONLY;
	}
};

/**
 * @returns {Boolean} permission for a certain action on the annotation
 */
Annotation.prototype.getPermissionFor = function(action) {
	return action !== AnnotationConstants.ActionPermissions.DUPLICATE_WITHOUT_TEXT;
};

/**
 * @returns {String} sort by attribute of the annotation
 */
Annotation.prototype.getSortByAttr = function() {
	return this.content;
};

/**
 * Get the outline of the rect of all annotations + its links
 * @returns {ShapeUtils.Rect} The outline of the rect of all annotations + its links
 */
Annotation.prototype.getOutliningRect = function() {
	var annotationGroup = _.union(this.links, [this]); //annotation with all its linked annotations

	var x0 = _.min(annotationGroup, function(annotation) {
		return annotation.rect.left();
	}).rect.left();
	var y0 = _.max(annotationGroup, function(annotation) {
		return annotation.rect.top();
	}).rect.top();
	var x1 = _.max(annotationGroup, function(annotation) {
		return annotation.rect.right();
	}).rect.right();
	var y1 = _.min(annotationGroup, function(annotation) {
		return annotation.rect.bottom();
	}).rect.bottom();

	return new ShapeUtils.Rect(new ShapeUtils.Point(x0, y0), new ShapeUtils.Size(x1 - x0, y1 - y0));
};

/**
 * @returns {String} class to represent the icon of the annotation (depends on its type)
 */
Annotation.prototype.getIconClass = function() {
	return AnnotationConstants.AnnotationIcons[this.annottype];
};

/**
 * @returns {Boolean} true if creation date is the same as the modification date
 */
Annotation.prototype.isAnnotationModified = function() {
	return this.date - this.creationdate === 0;
};

Annotation.prototype.setStartEndTime = function(start, end) {
	var startModified = false;
	var endModified = false;
	start = parseFloat(start, 10);
	end = parseFloat(end, 10);
	if (start !== undefined && this.startTime != start) {
		this.startTime = start;
		startModified = true;
	}
	if (end) {
		if (this.endTime != end) {
			this.endTime = end;
			endModified = true;
		}
	} else if (this.endTime != start) {
		this.endTime = start;
		endModified = true;
	}

	// Preserve correct order of start and end
	if (this.startTime > this.endTime) {
		var tmp = this.startTime;
		this.startTime = this.endTime;
		this.endTime = tmp;
		endModified = true;
		startModified = true;
	}

	this.setModified("TIME");
	this._emit("modified.startEndTime", this, startModified, endModified);
};

/**
 * Initialize an annotation based on a type, color and if it's selected
 * @param {AnnotationConstants.AnnotationType} type
 * @param {String} color - annotation color
 * @param {Boolean} selected - if the annotation is initialized as being selected
 * @param {AnnotationModel} [annotationModel] - needed when initializing annotation that has no name yet
 * @param {ShapeUtils.Rect} [rect] - rectangle to initialize the annotation with, if not specified 0,0 rect will be used
 */
Annotation.prototype.initAnnotation = function(type, color, selected, annotationModel, rect) {
	if (!this.name || this.name.length === 0) {
		if (annotationModel === undefined) {
			this.f.emit("error.SEVERE", "Problem initializing annotation", "No model specified for generating unique name!");
		} else {
			this.name = annotationModel.generateUniqueName();
		}
	}

	if (!rect) {
		rect = new ShapeUtils.Rect();
	}
	this.rect = rect;
	if (annotationModel) {
		this._scaling = annotationModel.getCurrentScaling();
	}

	if (!color) {
		color = this.annotationcolor;
	}
	this.originalAnnotationColor = color;
	this.annotationcolor = color;
	if (this.annotationcolor === null) {
		this.annotationColor = this._getDefaultAnnotationColor(type);
	}

	this.color = color;

	this.annottype = type;
	this.setSelected(selected);
	if (!this.subject || this.subject.length === 0) {
		this.subject = (this.inreplyto != null ? "Re" + ": " : "") + AnnotationConstants.AnnotationReplyName[this.annottype];
	}

	if (annotationModel && annotationModel.getTimeStamp() != null) {
		this.startTime = parseFloat(annotationModel.getTimeStamp(), 10);
		this.endTime = parseFloat(annotationModel.getTimeStamp(), 10);
	}

	switch (type) {
		case AnnotationConstants.AnnotationType.ANNOT_RECT:
			this.color = color;
			this.type = DrawUtils.LWTYPE_RECT;
			break;

		case AnnotationConstants.AnnotationType.ANNOT_ELLIPSE:
			this.color = color;
			this.type = DrawUtils.LWTYPE_ELLIPSE;
			break;

		case AnnotationConstants.AnnotationType.ANNOT_HIGHLIGHT:
			this.type = DrawUtils.LWTYPE_RECT;
			this.treatAsOutline = true;
			this.color = null;
			break;

		case AnnotationConstants.AnnotationType.ANNOT_LINE:
			this.type = DrawUtils.LWTYPE_LINE;
			this.color = color;

			this.pointArray[0] = new LineworkPoint();
			this.pointArray[0].makeMoveTo(this.rect.origin.x, this.rect.origin.y);
			this.pointArray[1] = new LineworkPoint();
			this.pointArray[1].makeLineTo(this.rect.origin.x + this.rect.size.width, this.rect.origin.y + this.rect.size.height);

			break;

		case AnnotationConstants.AnnotationType.ANNOT_ARROW:
			this.type = DrawUtils.LWTYPE_ARROW;
			this.color = color;

			this.pointArray[0] = new LineworkPoint();
			this.pointArray[0].makeMoveTo(this.rect.origin.x, this.rect.origin.y);
			this.pointArray[1] = new LineworkPoint();
			this.pointArray[1].makeLineTo(this.rect.origin.x + this.rect.size.width, this.rect.origin.y + this.rect.size.height);

			break;

		case AnnotationConstants.AnnotationType.ANNOT_PIN:
			this.type = DrawUtils.LWTYPE_IMAGE;
			this.imagehotspot = new ShapeUtils.Point(
				AnnotationConstants.AnnotationPin.imagehotspotx,
				AnnotationConstants.AnnotationPin.imagehotspoty
			);
			this.imagetouchspot = new ShapeUtils.Rect(
				new ShapeUtils.Point(AnnotationConstants.AnnotationPin.imagetouchspotx, AnnotationConstants.AnnotationPin.imagetouchspoty),
				new ShapeUtils.Size(AnnotationConstants.AnnotationPin.imagetouchspotw, AnnotationConstants.AnnotationPin.imagetouchspoth)
			);
			break;

		case AnnotationConstants.AnnotationType.ANNOT_FREEFORM:
			this.type = DrawUtils.LWTYPE_LINEARRAY;
			this.color = color;
			break;
	}
};

/**
 * Init an annotation from XML
 * @param {Element} xml - The XML element that is used to initialize the annotation
 * @param {Object} boxOffsets - The offsets of the media and trim boxes origins
 */
Annotation.prototype.initWithXML = function(xml, boxOffsets) {
	this.saved = true; //set saved to true
	this.isNew = false; //set new to false
	this.pointArray = []; //reset the point array to avoid duplication

	this.name = XMLUtils.attributeValue(xml, "name");
	this.subject = XMLUtils.attributeValue(xml, "subject");
	this.title = XMLUtils.attributeValue(xml, "title");
	this.flags = XMLUtils.attributeValue(xml, "flags");

	// For backwards compatibility reasons, check for attributes with uppercase characters too (<= WCR 16.1.1)
	this._view3DPosition = {
		distance: ViewerUtils.parseFloat(XMLUtils.attributeValue(xml, "view3ddistance")),
		scale: ViewerUtils.parseFloat(XMLUtils.attributeValue(xml, "view3dscale") || XMLUtils.attributeValue(xml, "view3Dscale")),
		azimuth: ViewerUtils.parseFloat(XMLUtils.attributeValue(xml, "view3dazimuth") || XMLUtils.attributeValue(xml, "view3Dazimuth")),
		elevation: ViewerUtils.parseFloat(XMLUtils.attributeValue(xml, "view3delevation") || XMLUtils.attributeValue(xml, "view3Delevation")),
		panx: ViewerUtils.parseFloat(XMLUtils.attributeValue(xml, "view3dpanx")) || 0,
		pany: ViewerUtils.parseFloat(XMLUtils.attributeValue(xml, "view3dpany")) || 0
	};

	this._view3DVersion = parseInt(XMLUtils.attributeValue(xml, "view3dversion"), 10);

	if (isNaN(this._view3DPosition.azimuth) || isNaN(this._view3DPosition.elevation)) {
		this._view3DPosition = null;
	}

	/**
	 * Parse the title and split it into "messagenumber", last name, first name and username
	 * This can be different depending on the annotation server this was created in
	 *
	 * Order of trying:
	 * 	- 11 Cagnie, Ruben (RUCA)
	 *  - 11 ADMIN (ADMIN)
	 *  - 11 USERNAME
	 *  - Cagnie, Ruben (RUCA)
	 */
	var numberExtraction = /((\d+)\s)?(.+)/.exec(this.title);
	if (numberExtraction) {
		this.messagenumber = numberExtraction[2];
		if (this.messagenumber === undefined) {
			this.messagenumber = -1;
		}

		var fullNameString = numberExtraction[3];
		// Check if string format is "11 USERNAME" or "11 ADMIN (ADMIN)"
		if (fullNameString[fullNameString.length - 1] === ")") {
			// Find username
			var brCount = 0;
			var index = fullNameString.length - 1;
			do {
				if (fullNameString[index] === ")") {
					brCount++;
				}
				if (fullNameString[index] === "(") {
					brCount--;
				}
				index--;
			} while (brCount > 0 && index >= 0);
			this.username = fullNameString.substring(index + 2, fullNameString.length - 1);

			// Find first and last name (if present)
			var firstAndLastName = fullNameString.substring(0, index);
			var nameExtraction = /([^,]+)(,\s(.+))?\s*/.exec(firstAndLastName);
			if (nameExtraction) {
				this.lastname = nameExtraction[1];
				this.firstname = nameExtraction[3] || "";
			} else {
				this.lastname = "";
				this.firstname = "";
			}
		} else {
			this.username = fullNameString;
			this.lastname = "";
			this.firstname = "";
		}
	}

	this.messagenumber = parseInt(this.messagenumber, 10);

	this.page = XMLUtils.attributeValue(xml, "page");

	this.date = ViewerUtils.dateFromString(XMLUtils.attributeValue(xml, "date"));
	this.creationdate = ViewerUtils.dateFromString(XMLUtils.attributeValue(xml, "creationdate"));

	this.state = XMLUtils.attributeValue(xml, "state");
	this.statemodel = XMLUtils.attributeValue(xml, "statemodel");

	this._parseContentsFromXML(xml);

	var replyto = XMLUtils.attributeValue(xml, "inreplyto");
	if (replyto) {
		this.inreplyto = replyto;
	}

	var linkedto = XMLUtils.attributeValue(xml, "linkedto");
	if (linkedto) {
		this.linkedto = linkedto;
	}

	var startTime = XMLUtils.attributeValue(xml, "startTime");
	if (startTime) {
		this.startTime = parseFloat(startTime, 10);
	}

	var endTime = XMLUtils.attributeValue(xml, "endTime");
	if (endTime) {
		this.endTime = parseFloat(endTime, 10);
	}

	var rectexp = new RegExp(/(.+),(.+),(.+),(.+)/);
	var rectsplit = rectexp.exec(XMLUtils.attributeValue(xml, "rect"));

	var x1 = GEOUtils.convertPSPointsToMM(rectsplit[1]);
	var y2 = GEOUtils.convertPSPointsToMM(rectsplit[2]);
	var x2 = GEOUtils.convertPSPointsToMM(rectsplit[3]);
	var y1 = GEOUtils.convertPSPointsToMM(rectsplit[4]);
	var coords = this._modifyAnnotatationCoords(x1, y2, x2, y1, boxOffsets);
	x1 = coords[0];
	y2 = coords[1];
	x2 = coords[2];
	y1 = coords[3];

	var type = xml.nodeName;
	switch (type) {
		case "square":
			this.initAnnotation(AnnotationConstants.AnnotationType.ANNOT_RECT);
			break;
		case "circle":
			this.initAnnotation(AnnotationConstants.AnnotationType.ANNOT_ELLIPSE);
			break;
		case "highlight":
			this.initAnnotation(AnnotationConstants.AnnotationType.ANNOT_HIGHLIGHT);
			break;
		case "text":
			this.rect.size.width = 0;
			this.rect.size.height = 0;
			this.initAnnotation(AnnotationConstants.AnnotationType.ANNOT_PIN);
			break;
		case "ink":
			this.initAnnotation(AnnotationConstants.AnnotationType.ANNOT_FREEFORM);

			if (!this.inreplyto) {
				var inklist = xml.getElementsByTagName("inklist")[0];
				if (inklist) {
					var gesture = inklist.getElementsByTagName("gesture")[0];
					var node = gesture;
					var pointArray = "";
					// Firefox splits large text regions into multiple Text objects (4096 chars in
					// each). Glue it back together.
					for (var x = 0; x < node.childNodes.length; x++) {
						pointArray += node.childNodes[x].nodeValue;
					}
					if (pointArray.length > 0) {
						var chuncs = pointArray.split(";");

						var startpoint = new LineworkPoint();
						var chunkpoint = chuncs[0].split(",");
						if (chunkpoint[0] !== undefined && chunkpoint[1] !== undefined) {
							startpoint.makeMoveTo(GEOUtils.convertPSPointsToMM(chunkpoint[0]), GEOUtils.convertPSPointsToMM(chunkpoint[1]));
							this.pointArray.push(startpoint);
							for (var i = 1; i < chuncs.length; i++) {
								var movepoint = new LineworkPoint();
								chunkpoint = chuncs[i].split(",");
								if (chunkpoint[0] !== undefined && chunkpoint[1] !== undefined) {
									var px = GEOUtils.convertPSPointsToMM(chunkpoint[0]);
									var py = GEOUtils.convertPSPointsToMM(chunkpoint[1]);
									movepoint.makeLineTo(px, py);
									this.pointArray.push(movepoint);
								} else {
									this.f.emit("error.DEBUG", "chunkpoint[0] || chunkpoint[1] is undefined...");
								}
							}
						} else {
							this.f.emit("error.DEBUG", "chunkpoint[0] || chunkpoint[1] is undefined...");
						}
					}
					this.calculateRectFromPointArray();
				}
			}
			break;
		case "line":
			var subject = XMLUtils.attributeValue(xml, "subject");
			if (subject && subject == "Arrow") {
				this.initAnnotation(AnnotationConstants.AnnotationType.ANNOT_ARROW);
			} else {
				this.initAnnotation(AnnotationConstants.AnnotationType.ANNOT_LINE);
			}

			if (!this.inreplyto) {
				var pointexp = new RegExp(/(.+),(.+)/);

				var start = XMLUtils.attributeValue(xml, "start");
				var end = XMLUtils.attributeValue(xml, "end");

				var startsplit = pointexp.exec(start);
				var endsplit = pointexp.exec(end);

				var startx = GEOUtils.convertPSPointsToMM(startsplit[1]);
				var starty = GEOUtils.convertPSPointsToMM(startsplit[2]);
				var endx = GEOUtils.convertPSPointsToMM(endsplit[1]);
				var endy = GEOUtils.convertPSPointsToMM(endsplit[2]);

				var startpoint = new LineworkPoint();
				startpoint.makeMoveTo(startx, starty);
				var endpoint = new LineworkPoint();
				endpoint.makeLineTo(endx, endy);

				this.pointArray[0] = startpoint;
				this.pointArray[1] = endpoint;

				var sp = new ShapeUtils.Point(startpoint.x, startpoint.y);
				var ep = new ShapeUtils.Point(endpoint.x, endpoint.y);
				this.rect = new ShapeUtils.Rect(sp, sp.sizeTo(ep));
				this.rect.normalize();
			}
			break;
	}

	this.rect.origin.x = x2 < x1 ? x2 : x1;
	this.rect.origin.y = y2 < y1 ? y2 : y1;
	this.rect.size.width = Math.abs(x2 - x1);
	this.rect.size.height = Math.abs(y2 - y1);

	var color = XMLUtils.attributeValue(xml, "color");
	if (!/^#?[a-f\d]{6}$/i.test(color)) {
		color = this._getDefaultAnnotationColor(this.annottype);
	}
	this.color = color;
	this.originalAnnotationColor = color;
	this.annotationcolor = color;

	var popup = xml.getElementsByTagName("popup")[0];
	if (popup) {
		this.popuppage = XMLUtils.attributeValue(popup, "page");
		this.popupIsOpen = XMLUtils.attributeValue(popup, "open") == "yes";

		var popuprectsplit = rectexp.exec(XMLUtils.attributeValue(popup, "rect"));

		var popupx1 = GEOUtils.convertPSPointsToMM(popuprectsplit[1]);
		var popupy1 = GEOUtils.convertPSPointsToMM(popuprectsplit[2]);
		var popupx2 = GEOUtils.convertPSPointsToMM(popuprectsplit[3]);
		var popupy2 = GEOUtils.convertPSPointsToMM(popuprectsplit[4]);

		let coords = this._modifyAnnotationPopupCoords(popupx1, popupy1, popupx2, popupy2);

		popupx1 = coords[0];
		popupy1 = coords[1];
		popupx2 = coords[2];
		popupy2 = coords[3];

		this.popuprect.origin.x = popupx1;
		this.popuprect.origin.y = popupy1;
		this.popuprect.size.width = popupx2 - popupx1;
		this.popuprect.size.height = popupy2 - popupy1;
	}

	var thumbnail = XMLUtils.attributeValue(xml, "thumbnail");
	if (thumbnail) {
		this._thumbnail = thumbnail;
	}

	var detailedThumbnail = XMLUtils.attributeValue(xml, "detailed-thumbnail");
	if (detailedThumbnail) {
		this._detailedThumbnail = detailedThumbnail;
	}
};

/**
 * Allows the annotation to modify it's coordinates when creating the annotation from XML
 * This lets subclasses define other coordinates if for example the coordinates in the xml are relative to the media box instead of the trim box
 * Defaults to coordinates relative to trim box
 * @param {Number} x1 - bottom left x coordinate
 * @param {Number} y2 - bottom left y coordinate
 * @param {Number} x2 - top right x coordinate
 * @param {Number} y1 - top right y coordinate
 * @param {Object} boxOffests - object containing the mediaboxOffset and trimboxOffset
 */
Annotation.prototype._modifyAnnotatationCoords = function(x1, y2, x2, y1, boxOffsets) {
	return [x1, y2, x2, y1];
};

/**
 * Allows the annotation to modify it's popup coordinates when creating the annotation from XML
 * This lets subclasses define other coordinates
 * @param {Number} x1 - bottom left x coordinate
 * @param {Number} y1 - bottom left y coordinate
 * @param {Number} x2 - top right x coordinate
 * @param {Number} y2 - top right y coordinate
 * @param {Object} boxOffests - object containing the mediaboxOffset and trimboxOffset
 */
Annotation.prototype._modifyAnnotationPopupCoords = function(x1, y1, x2, y2) {
	return [x1, y1, x2, y2];
};

/**
 * Duplicate an annotation
 * @param {Boolean} keepReplies - True if you want to copy the replies to the duplicated annotation, false otherwise
 * @param {Boolean} keepStates - True if you want to copy the states to the duplicated annotation, false otherwise
 * @param {Boolean} keepInklist - True if you want to copy the pointArray to the duplicated annotation, false otherwise
 * @param {Boolean} keepLinkedAnnotations - True if you want to copy the linked annotations to the duplicated annotation, false otherwise
 * @param {AnnotationModel} annotationModel - The annotation model to ask a unique new name to
 * @returns {Annotation} the duplicated annotation
 */
Annotation.prototype.duplicate = function(keepReplies, keepStates, keepInklist, keepLinkedAnnotations, annotationModel) {
	var newAnnotation = new this.constructor(this.f);
	if (!keepInklist) {
		//copies properties of the annotation
		_.each(this, function(value, key) {
			newAnnotation[key] = value;
		});
		newAnnotation.modifiedContent = null;
		// make sure the rectangle is cloned, note copied
		if (newAnnotation.rect) {
			newAnnotation.rect = newAnnotation.rect.copy();
		}
		// same for the popuprect
		if (newAnnotation.popuprect) {
			newAnnotation.popuprect = newAnnotation.popuprect.copy();
		}
	} else {
		newAnnotation.initAnnotation(this.annottype, this.annotationcolor, true, annotationModel);
		newAnnotation.content = this.getTextContent();
	}

	newAnnotation.isNew = true;
	newAnnotation.name = annotationModel.generateUniqueName();

	if (!keepReplies) {
		newAnnotation.replies = [];
	}
	if (!keepStates) {
		newAnnotation.states = [];
	}
	if (!keepInklist) {
		newAnnotation.pointArray = [];
	} else {
		//check if the annotation is a reply, if this is the case use parent annotation's properties else is this the mainAnnotation
		var mainAnnotation = this.inreplyto ? this.parent : this;
		newAnnotation.pointArray = [];
		_.each(mainAnnotation.pointArray, function(point) {
			newAnnotation.pointArray.push(_.clone(point));
		});
		newAnnotation.color = _.clone(mainAnnotation.color);
		newAnnotation.rect = mainAnnotation.rect.copy();
		newAnnotation.popuprect = mainAnnotation.popuprect.copy();
	}
	if (keepLinkedAnnotations) {
		newAnnotation.links = [];
		_.each(this.links, function(link) {
			var newLink = link.duplicate(false, false, true, false, annotationModel);
			newLink.linkedto = newAnnotation.name;
			newLink.parent = newAnnotation;
			newAnnotation.links.push(newLink);
		});
	} else {
		newAnnotation.links = [];
	}

	newAnnotation.startTime = this.startTime;
	newAnnotation.endTime = this.endTime;

	return newAnnotation;
};

/**
 * @returns {Boolean} true if the annotation needs a selection box (anchor points) to be edited
 */
Annotation.prototype.needsSelectionBox = function() {
	return this.annottype != AnnotationConstants.AnnotationType.ANNOT_PIN;
};

/**
 * Set the annotation selected
 * @param {Boolean} selected - true if selected
 */
Annotation.prototype.setSelected = function(selected) {
	this.selected = selected;
	if (selected) {
		this.linewidth = 3;
	} else {
		this.linewidth = 2;
	}

	if (this.annottype === AnnotationConstants.AnnotationType.ANNOT_HIGHLIGHT) {
		this.fillcolor = "rgba(255, 255, 0, " + (selected ? ".8" : ".3") + ")";
	}
	if (this.annottype === AnnotationConstants.AnnotationType.ANNOT_PIN) {
		this.imageRef = selected ? AnnotationConstants.AnnotationPin.highlighted : AnnotationConstants.AnnotationPin.normal;
	}
};

/**
 * Update the indication number (depending on the links) and the indication color
 */
Annotation.prototype.updateIndicator = function() {
	this.indicatorText = "" + this.messagenumber;
	if (this.linkedto) {
		if (this.parent) {
			this.indicatorText += "-" + (_.indexOf(this.parent.links, this) + 2); //add 2, because 1 is the main annotation and an array starts from 0...
		}
	} else {
		var p = this.getGodfather();
		if (p.links.length > 0) {
			this.indicatorText += "-1";
		}
	}
	this.indicatorTextColor = this.getHeaderTextColor();
	this.indicatorColor = this.annotationcolor;
};

/**
 * Get the full username of the user
 * @returns {String} user name in following form: "Cagnie, Ruben (RUCA)" or "Admin (ADMIN")
 */
Annotation.prototype.getUserName = function() {
	if (this.firstname && this.lastname && this.username) {
		return this.lastname + ", " + this.firstname + " (" + this.username + ")";
	} else if (this.lastname && this.username) {
		return this.lastname + " (" + this.username + ")";
	} else if (this.username) {
		return this.username;
	} else {
		return this.lastname;
	}
};

/**
 * @returns {Object|null} the annotation color in RGB colors
 */
Annotation.prototype.rgbColor = function() {
	var result = /^#?([a-f\d]{2})([a-f\d]{2})([a-f\d]{2})$/i.exec(this.annotationcolor);
	return result
		? {
				r: parseInt(result[1], 16),
				g: parseInt(result[2], 16),
				b: parseInt(result[3], 16)
		  }
		: null;
};

/**
 * @returns {String} 'white' if the annotation color is a dark color, 'black' if the annotation color is a light color
 */
Annotation.prototype.getHeaderTextColor = function() {
	var bg = this.rgbColor();
	var y = 0.3 * (bg.r / 255.0) + (0.59 * bg.g) / 255.0 + (0.11 * bg.b) / 255.0;

	if (y > 0.5) {
		return "black";
	} else {
		return "white";
	}
};

/**
 * Indicate that the annotation has been modified. The modification date will be set to "now" as well
 * @param {String} what - what is modified
 */
Annotation.prototype.setModified = function(what) {
	this._modifications.add(what);
	this.saved = this._modifications.size() === 0;
	this.date = new Date();
	this.f.emit("annotation.modified", this, what);
};

/**
 * Set if the annotation is new or not, change
 * @param {Boolean} isNew
 */
Annotation.prototype.setIsNew = function(isNew) {
	this.isNew = isNew;
};

/**
 * Cancels the modified state.
 * @param {String} [what] - which modification is canceled? if not provided, all modifications are cancelled
 */
Annotation.prototype.cancelModified = function(what) {
	if (what === "CONTENT") {
		this.modifiedContent = null;
	}
	if (what !== undefined) {
		this._modifications.remove(what);
	} else {
		this._modifications.clear();
	}

	this.saved = this._modifications.size() === 0;
};

/**
 * The annotation is being modified, the changes are not yet saved
 * The modification date is not yet updated.
 */
Annotation.prototype.setBeingModified = function(modifiedContent) {
	if (modifiedContent === undefined) {
		modifiedContent = null;
	}

	if (this.modifiedContent !== modifiedContent) {
		this._modifications.add("CONTENT");
		this.modifiedContent = modifiedContent;
		this.saved = false;
	}
};

/**
 * Is the annotation being modified
 * @returns {Boolean}
 */
Annotation.prototype.isBeingModified = function() {
	return !this.saved;
};

/**
 * Run through an annotation to get all its children (replies and/or states)
 * @param {Boolean} self - indicates if the callback function should also be called on the annotation itself
 * @param {Boolean} replies - the callback will be called for every reply (and its replies)
 * @param {Boolean} states - the callback will be called for every state
 * @param {Function} callback - will be called (with the annotation) for every annotation that was found
 */
Annotation.prototype._runThroughAnnotation = function(annotation, replies, states, callback) {
	if (replies) {
		//making a shalow copy of the array makes it possible to remove it in the callback
		var annotationReplies = _.clone(annotation.replies);
		_.each(
			annotationReplies,
			_.bind(function(annot) {
				if (!callback(annot)) {
					this._runThroughAnnotation(annot, replies, states, callback);
				}
			}, this)
		);
	}
	if (states) {
		var annotationStates = _.clone(annotation.states);
		_.each(
			annotationStates,
			_.bind(function(annot) {
				if (!callback(annot)) {
					this._runThroughAnnotation(annot, replies, states, callback);
				}
			}, this)
		);
	}
};

/**
 * Run through an annotation to get all its children (replies and/or states)
 * @param {Boolean} self - indicates if the callback function should also be called on the annotation itself
 * @param {Boolean} replies - the callback will be called for every reply (and its replies)
 * @param {Boolean} states - the callback will be called for every state
 * @param {Function} callback - will be called (with the annotation) for every annotation that was found
 */
Annotation.prototype.runThroughAnnotation = function(self, replies, states, callback) {
	if (self) {
		if (callback(this)) {
			return;
		}
	}
	this._runThroughAnnotation(this, replies, states, callback);
};

/**
 * Makes an XML representation of the annotation
 * @param {Document} document - the owner document that is needed to create nodes
 * @returns {Element} element that represents the annotation (NOT appended to the document yet)
 */
Annotation.prototype.toXML = function(document) {
	/*
    <circle color="#FF0000" creationdate="D:20130608142640-04'00'" date="D:20130608223906-04'00'" flags="print" name="MfUwftyuPc4ZRLmCWavqpG" page="0" rect="116.65092240883808,193.72593822615414,247.69028039670718,291.9917703696743" subject="Oval" title="1 Cagnie, Ruben (RUCA)">
        <contents-richtext>
            <body>
                <p>Eggs</p>
            </body>
        </contents-richtext>
        <popup open="no" page="0" rect="255.2249117711603,158.17305818469373,405.22491177116024,270.67305818469373"/>
    </circle>
    */
	var annot = document.createElementNS(null, AnnotationConstants.AnnotationElementName[this.annottype]);

	if (this.isNew && !this.linkedto) {
		annot.setAttribute("title", this.getUserName());
	} else {
		annot.setAttribute("title", this.messagenumber + " " + this.getUserName());
	}

	if (!this.name || this.name.length === 0) {
		this.f.emit("error.SEVERE", "Trying to convert an annotation without a name to XML");
	}

	annot.setAttribute("name", this.name);

	annot.setAttribute("color", this.annotationcolor);
	annot.setAttribute("creationdate", ViewerUtils.stringFromDate(this.creationdate));
	annot.setAttribute("date", ViewerUtils.stringFromDate(this.date));

	annot.setAttribute("flags", this.flags);

	annot.setAttribute("page", this.page);

	//rect="116.65092240883808,193.72593822615414,247.69028039670718,291.9917703696743"
	var rect = this.rect;
	if (this._scaling) {
		rect = new ShapeUtils.Rect(
			new ShapeUtils.Point((this.rect.origin.x * 1) / this._scaling.x, (this.rect.origin.y * 1) / this._scaling.y),
			new ShapeUtils.Size((this.rect.size.width * 1) / this._scaling.x, (this.rect.size.height * 1) / this._scaling.y)
		);
	}
	annot.setAttribute("rect", ViewerUtils.rectToString(rect));

	annot.setAttribute("subject", this.subject);

	if (this.state) {
		annot.setAttribute("state", this.state);
	}
	if (this.statemodel) {
		annot.setAttribute("statemodel", this.statemodel);
	}
	if (this.inreplyto) {
		annot.setAttribute("inreplyto", this.inreplyto);
	}
	if (this.linkedto) {
		annot.setAttribute("linkedto", this.linkedto);
	}

	if (this.startTime != null || this.endTime != null) {
		// if for some reason, one of then doesn't exist, make sure to store a consistent state
		annot.setAttribute("startTime", this.startTime || 0);
		annot.setAttribute("endTime", this.endTime || this.startTime);
	}

	if (this._view3DPosition) {
		annot.setAttribute("view3dversion", "1");
		annot.setAttribute("view3ddistance", this._view3DPosition.distance);
		annot.setAttribute("view3dazimuth", this._view3DPosition.azimuth);
		annot.setAttribute("view3delevation", this._view3DPosition.elevation);
		annot.setAttribute("view3dpanx", this._view3DPosition.panx);
		annot.setAttribute("view3dpany", this._view3DPosition.pany);
	}

	if (!this.inreplyto) {
		if (AnnotationConstants.AnnotationType.ANNOT_FREEFORM == this.annottype) {
			/*
    <inklist>
        <gesture>-285.60323374911894,598.6597735920734;-285.60323374911894,598.3248898662756;-286.94276865231024,596.4830293743876;-296.6543967004471,584.2597733827671;-302.84974562770685,575.0504709233269;-303.18462935350465,572.7062848427423;-302.51486190190906,570.0272150363597;-300.3381176842231,565.8411684638869;-297.15672228914383,562.8272149317065;-294.3102106198623,561.8225637543129;-291.29625708768197,559.6458195366272;-290.62648963608626,557.8039590447391;-290.62648963608626,554.1202380609631;-291.63114081347976,552.7807031577718;-294.64509434566014,550.2690752142881;-296.82183856334603,549.9341914884903;-304.0218386679992,547.7574472708044;-304.0218386679992,545.9155867789165;-303.18462935350465,536.8737261823752;-301.84509445031335,534.6969819646893;-294.64509434566014,524.6504701907547;-292.13346640217645,523.1434934246645;-289.4543965957939,519.9620980295851;-281.58462903954506,514.1016328281232;-278.4032336444658,513.0969816507296;-263.50090784646284,510.0830281185492;-255.9660240160117,510.0830281185492;-211.2590466220026,510.92023743304384;-196.3567208239996,512.7620979249318;-195.352069646606,514.436516553921;-195.352069646606,516.278377045809;-197.0264882755952,524.6504701907547;-198.53346504168533,527.1620981342382;-216.95206996056564,558.8086102221325;-221.64044212173508,568.6876801331682;-226.3288142829044,575.5527965120236;-231.6869538956696,581.5807035763844;-236.04044233104128,584.9295408343627;-249.43579136295418,593.971401430904;-254.12416352412376,595.813261922792;-258.4776519594954,596.4830293743876;-275.72416383808314,597.4876805517811;-276.89625687837554,596.4830293743876;-278.068349918668,594.3062851567017;-278.4032336444658,593.4690758422072;-278.4032336444658,593.4690758422072</gesture>
    </inklist>
         */
			var inklist = document.createElementNS(null, "inklist");
			var gesture = document.createElementNS(null, "gesture");

			var pointarray = "";
			for (var i = 0; i < this.pointArray.length; i++) {
				var point = this.pointArray[i];
				if (i !== 0) {
					pointarray += ";";
				}
				if (this._scaling) {
					pointarray +=
						GEOUtils.convertMMToPSPoints((point.x * 1) / this._scaling.x) +
						"," +
						GEOUtils.convertMMToPSPoints((point.y * 1) / this._scaling.y);
				} else {
					pointarray += GEOUtils.convertMMToPSPoints(point.x) + "," + GEOUtils.convertMMToPSPoints(point.y);
				}
			}
			gesture.textContent = pointarray;

			inklist.appendChild(gesture);
			annot.appendChild(inklist);
		} else if (
			AnnotationConstants.AnnotationType.ANNOT_LINE == this.annottype ||
			AnnotationConstants.AnnotationType.ANNOT_ARROW == this.annottype
		) {
			//end="374.1086344293871,459.21259580886755"
			//start="284.02491218977303,541.7614342180307"
			var startpoint = this.pointArray[0];
			var endpoint = this.pointArray[1];

			var rect;
			if (this._scaling) {
				rect = new ShapeUtils.Rect(
					new ShapeUtils.Point((rect.origin.x * 1) / this._scaling.x, (rect.origin.y * 1) / this._scaling.y),
					new ShapeUtils.Size((rect.size.width * 1) / this._scaling.x, (rect.size.height * 1) / this._scaling.y)
				);
			} else {
				rect = new ShapeUtils.Rect(
					new ShapeUtils.Point(startpoint.x, endpoint.y),
					new ShapeUtils.Size(endpoint.x - startpoint.x, startpoint.y - endpoint.y)
				);
			}
			annot.setAttribute("rect", ViewerUtils.rectToString(rect));

			var start = GEOUtils.convertMMToPSPoints(startpoint.x) + "," + GEOUtils.convertMMToPSPoints(startpoint.y);
			var end = GEOUtils.convertMMToPSPoints(endpoint.x) + "," + GEOUtils.convertMMToPSPoints(endpoint.y);
			if (this._scaling) {
				start =
					GEOUtils.convertMMToPSPoints((startpoint.x * 1) / this._scaling.x) +
					"," +
					GEOUtils.convertMMToPSPoints((startpoint.y * 1) / this._scaling.y);
				end =
					GEOUtils.convertMMToPSPoints((endpoint.x * 1) / this._scaling.x) +
					"," +
					GEOUtils.convertMMToPSPoints((endpoint.y * 1) / this._scaling.y);
			}
			annot.setAttribute("start", start);
			annot.setAttribute("end", end);

			if (this.annottype == AnnotationConstants.AnnotationType.ANNOT_ARROW) {
				annot.setAttribute("head", "None");
				annot.setAttribute("tail", "OpenArrow");
			}
		}
	}
	if (!this.linkedto || this.statemodel) {
		var content = document.createElement("contents-richtext");

		if (this.modifiedContent !== null) {
			this.content = this.modifiedContent;
		}
		if (this.content) {
			var markupElement = this._parseMarkupToXMLDocument(this.content);
			var bodyTag = markupElement.getElementsByTagName("body")[0];

			if (bodyTag) {
				content.appendChild(copyNode(bodyTag, document));
			} else {
				var newBody = document.createElement("body");
				// copy children
				if (markupElement.hasChildNodes && markupElement.hasChildNodes()) {
					for (var i = 0; i < markupElement.childNodes.length; i++) {
						var copy = copyNode(markupElement.childNodes[i], document);
						if (copy != null) {
							newBody.appendChild(copy);
						}
					}
				}
				content.appendChild(newBody);
			}
		}

		annot.appendChild(content);
	}

	if (this.popuprect.origin.x === 0 && this.popuprect.origin.y === 0) {
		this.popuprect.origin.x = this.rect.right();
		this.popuprect.origin.y = this.rect.top();
	}

	var popup = document.createElementNS(null, "popup");
	popup.setAttribute("rect", ViewerUtils.rectToString(this.popuprect));
	popup.setAttribute("page", this.popuppage);
	popup.setAttribute("open", this.popupIsOpen ? "yes" : "no");
	annot.appendChild(popup);

	if (this._thumbnail) {
		annot.setAttribute("thumbnail", this._thumbnail);
	}

	if (this._detailedThumbnail) {
		annot.setAttribute("detailed-thumbnail", this._detailedThumbnail);
	}

	return annot;
};

/**
 * This helper function copies a node and al its attributes and children to a new node (that is created in the document given)
 *
 * In this way, copy without the conflicts of namespaces is possible
 * Only works for TextNode and Element (otherwise null is returned)
 * @param {Node} node - the node that needs to be copied
 * @param {Document} document - the document in which the copy should be created
 * @returns {Node} the copied node or null if it is not a text or element node
 */
function copyNode(node, document) {
	var newElement = null;

	if (node.nodeType === Node.TEXT_NODE) {
		newElement = document.createTextNode(node.textContent);
	} else if (node.nodeType === Node.ELEMENT_NODE) {
		newElement = document.createElement(node.nodeName.toLowerCase());

		// copy attributes if available
		if (node.hasAttributes && node.hasAttributes()) {
			for (var i = 0; i < node.attributes.length; i++) {
				newElement.setAttribute(node.attributes[i].name, node.attributes[i].value);
			}
		}

		// copy children
		if (node.hasChildNodes && node.hasChildNodes()) {
			for (var i = 0; i < node.childNodes.length; i++) {
				var childCopy = copyNode(node.childNodes[i], document);
				if (childCopy != null) {
					newElement.appendChild(childCopy);
				}
			}
		}
	}

	return newElement;
}

/**
 * This function can parse an xhtml string in to a Document
 *
 * If DOMParser(text/html) is available, this is used, otherwise a workarround  is used
 * (works in Chrome/Firefox/IE9 and up)
 *
 * @param {String} markup
 * @returns {DocumentElement} the documentelement of the parsed markup
 */
Annotation.prototype._parseMarkupToXMLDocument = function(markup) {
	// replace non-breaking spaces with normal spaces, otherwise they will get lost in the parsers...
	markup = markup.replace(/\&nbsp;/g, " ");

	// Firefox/Opera/IE throw errors on unsupported types
	try {
		// WebKit returns null on unsupported types
		if (new DOMParser().parseFromString("", "text/html")) {
			// text/html parsing is natively supported
			var parser = new DOMParser();
			var markupDoc = parser.parseFromString(markup, "text/html");

			// on iOS when running the viewer in a UIWebView, documentElement is sometimes null
			// this is a very obscure bug by Apple
			// if documentElement is not defined/null, we fall back to method using document.innerHTML (after the try catch)
			if (markupDoc && markupDoc.documentElement) {
				return markupDoc.documentElement;
			}
		}
	} catch (ex) {}

	// the native DOMParser is not available...
	// create HTMLDocument
	var doc = document.implementation.createHTMLDocument("");

	// find the contents of the <body> tag (or <html>-tag is no body tag is found)
	var matches = /<body>(.*)<\/body>/.exec(markup);
	if (matches && matches.length >= 1) {
		markup = matches[1];
	}

	matches = /<html>(.*)<\/html>/.exec(markup);
	if (matches && matches.length >= 1) {
		markup = matches[1];
	}

	doc.body.innerHTML = markup;

	return doc.documentElement;
};

/**
 * Returns the state-annotation object that is for a given username
 * @param {String} username
 * @returns {Annotation} state annotation object that is used for the given user
 */
Annotation.prototype.getUserStateAnnotation = function(username) {
	return _.find(this.states, function(state) {
		return state.username === username;
	});
};

/**
 * @param {Number} max - Maximum number of characters
 * @returns {String} the content as a string (without HTML tags)
 */
Annotation.prototype.getStringContent = function(max) {
	var c = this.content.replace(/<([^<]*)>/g, "").trim();
	if (max && c.length > max) {
		c = c.substring(0, max) + "...";
	}
	return c;
};

/**
 * Set the color of an annotation (and update all its links/replies/states)
 * @param {String} color - new color of the annotation
 */
Annotation.prototype.setColor = function(color) {
	if (color === null) {
		color = this._getDefaultAnnotationColor(this.annottype);
	}
	this.runThroughAnnotation(true, true, true, function(annot) {
		if (annot.annotationcolor != color) {
			annot.annotationcolor = color;
			annot.color = color;
			annot.setModified("COLOR");
		}
	});
	var p = this.getGodfather();
	_.each(p.links, function(annot) {
		if (annot.annotationcolor != color) {
			annot.annotationcolor = color;
			annot.color = color;
			annot.setModified("COLOR");
		}
	});
};

/**
 * Cancel the modified indication for color (also do this for all the links that are effected when doing color modification
 */
Annotation.prototype.cancelModifiedColor = function() {
	this.cancelModified("COLOR");
	this.runThroughAnnotation(true, true, true, function(annot) {
		annot.cancelModified("COLOR");
	});
	var p = this.getGodfather();
	_.each(p.links, function(annot) {
		annot.cancelModified("COLOR");
	});
};

/**
 * Reset the annotation color to the last saved one
 */
Annotation.prototype.resetAnnotationColor = function() {
	if (this.originalAnnotationColor) {
		this.setColor(this.originalAnnotationColor);
	}
};

/**
 * Get the godfather of this annotation (meaning the most main annotation)
 * @returns {Annotation} - Godfather of this annotation
 */
Annotation.prototype.getGodfather = function() {
	var mainAnnotation = this;
	while (mainAnnotation.parent) {
		mainAnnotation = mainAnnotation.parent;
	}
	return mainAnnotation;
};

/**
 * Get the type of the annotation
 * @returns {AnnotationModule.AnnotationConstants.AnnotationType} type of the annotation
 */
Annotation.prototype.getType = function() {
	return this.annottype;
};

/**
 * returns the text content of the annotation
 * @return {String}
 */
Annotation.prototype.getTextContent = function() {
	return _.clone(this.content);
};

/**
 * returns the message number of this annotation
 * @return {Number}
 */
Annotation.prototype.getMessageNumber = function() {
	return this.messagenumber;
};

/**
 * Set the content of the reply (but don't save it yet)
 * @param {String} text
 */
Annotation.prototype.setUnsavedReply = function(text) {
	this.setModified("REPLY");
	this.unsavedReplyText = text;
	if (text === "") {
		this.cancelModified("REPLY");
	}
};

/**
 * Reset the value of the unsaved reply text
 */
Annotation.prototype.resetUnsavedReply = function() {
	this.unsavedReplyText = null;
	this.cancelModified("REPLY");
};

/**
 * Get the unsaved reply text
 * @returns {String} the value of the unsaved reply
 */
Annotation.prototype.getUnsavedReply = function() {
	return this.unsavedReplyText;
};

/**
 * Check if the annotation has an unsaved reply
 * @returns {Boolean} true if there is an unsaved reply
 */
Annotation.prototype.hasUnsavedReply = function() {
	return this.unsavedReplyText !== null && this.unsavedReplyText !== "";
};

/**
 * Is this annotation used as a state object
 * @returns {Boolean}
 */
Annotation.prototype.isStateObject = function() {
	return this.state !== null;
};

/**
 * Set the scaling of the annotation, all points of the annotation (linework) are scaled witht his value
 * @param {ShapeUtils.Point|null} scaling
 */
Annotation.prototype.setScale = function(scaling) {
	this.scale(scaling);
	this._scaling = scaling;
};

/**
 * Is the annotation scaled
 * @returns {Boolean}
 */
Annotation.prototype.isScaled = function() {
	return this._scaling !== null && this._scaling !== undefined;
};

/**
 * Get the scaling for this annotation
 * @returns {ShapeUtils.Point|null}
 */
Annotation.prototype.getScale = function() {
	return this._scaling;
};

/**
 * Returns the default annotation color for the given annotation type
 * @param   {AnnotationConstants.AnnotationType} annotType
 * @returns {String}
 */
Annotation.prototype._getDefaultAnnotationColor = function(annotType) {
	if (annotType == AnnotationConstants.AnnotationType.ANNOT_PIN) {
		return this.f.getCurrentUser().getColor("PREFERENCES_ANNOTATION_NOTE_COLOR");
	} else {
		return this.f.getCurrentUser().getColor("PREFERENCES_ANNOTATION_RECTANGLE_COLOR");
	}
};

/**
 * parses xml and puts it in the contents
 * @param {NodeList} xml - the full annotation XML
 */
Annotation.prototype._parseContentsFromXML = function(xml) {
	var contentEls = xml.getElementsByTagName("contents-richtext");

	this.content = "<p></p>";
	if (contentEls.length > 0) {
		var el = contentEls[0].getElementsByTagName("body")[0];
		if (!el) {
			el = contentEls[0];
		}

		if (typeof window.XMLSerializer != "undefined") {
			this.content = new XMLSerializer().serializeToString(el);
		} else if (typeof el.xml != "undefined") {
			this.content = el.xml; // internet explorer doesn't have XMLSerializer
		}
	}
};

/**
 * Object with extra info to show before content or null
 * @example Object {
 *           element: myJQueryElement
 *           tooltip: a tooltip for the element
 * 			}
 * @returns {Object}
 */
Annotation.prototype.getExtraContent = function() {
	return null;
};

/**
 * Save the 3D View position in this annotation
 * @param {Object} position
 *                 @example
 *                   {scale: 1.5,
 *                   azimuth: 46.6,
 *                   elevation: -52,
 * 					 panx: -5,
 * 					 pany: 0
 * 					}
 */
Annotation.prototype.set3DViewPosition = function(position) {
	this._view3DPosition = position;
	this.setModified("3DVIEW");
};

/**
 * Get the 3D View camera position of this annotation or null if there is none
 * @returns {Object} position
 *                 @example
 *                   {scale: 1.5,
 *                   azimuth: 46.6,
 *                   elevation: -52,
 * 					 panx: -5,
 * 					 pany: 0
 * 					}
 */
Annotation.prototype.get3DViewPosition = function() {
	return this._view3DPosition;
};

/**
 * Change the base64 encoded thumbnail of this annotation
 *
 * @param {String} base64String
 */
Annotation.prototype.setThumbnail = function(base64String) {
	this._thumbnail = base64String;
	this.setModified("THUMBNAIL");
};

/**
 * Change the base64 encoded detail thumbnail of this annotation
 *
 * @param {String} base64String
 */
Annotation.prototype.setDetailedThumbnail = function(base64String) {
	this._detailedThumbnail = base64String;
	this.setModified("DETAILEDTHUMBNAIL");
};

export default Annotation;
