import PrivateChannel from "@esko/web-viewer/core/PrivateChannel";
import Popup from "@esko/web-viewer/layout/component/Popup";
import ViewerToolbarButton from "@esko/web-viewer/layout/component/ViewerToolbarButton";
import Template from "@esko/template";
import AnnotationConstants from "@esko/web-viewer/annotation/AnnotationConstants";
import mainTpl from "@esko/web-viewer/annotation/AnnotationCustomFilterPopup.tpl.html";
import subTpl from "@esko/web-viewer/annotation/AnnotationCustomFilterRow.tpl.html";
import $ from "jquery";
import _ from "underscore";
/**
 * Create a popup for creating a custom annotation filter
 * @exports AnnotationCustomFilterPopup
 * @memberof AnnotationModule
 * @constructor
 * @augments Layout.Popup
 *
 * @param {Core.ViewerFacade} f - Viewerfacade
 * @param {String} popupId - Id of the popup used for its private channel
 */
var AnnotationCustomFilterPopup = function(f, popupId, annotationModel) {
	Popup.call(this, f, popupId, f.getContext());
	PrivateChannel.installTo(this, f, ["annotationFilterPopup"]);

	this.element.addClass("annotation-custom-filter-popup");
	this.element.attr("data-sahi-id", "annotation-custom-filter-popup");

	/**
	 * @member {AnnotationModel}
	 * @private
	 */
	this._annotationModel = annotationModel;

	/**
	 * @member {AnnotationFilterModel}
	 * @private
	 */
	this._filterModel = this._annotationModel.getAnnotationFilterModel();

	/**
	 * @member {Array.<Object>} saves the state of the filter when the popup was opened to be able to restore it
	 * @private
	 */
	this._savedFilterState = null;

	/**
	 * @member {Boolean} remember if the custom filter was enabled when the popup was opened, to be able to restore the filter correctly
	 * @private
	 */
	this._wasCustomFilterActive = false;

	/**
	 * @member {JQueryObject} content of the entire filter in the popup (not the action buttons)
	 * @private
	 */
	this._filterContent = $("<div>").addClass("annotationFilter");

	/**
	 * @member {ShapeUtils.Point} location the popup should be rendered
	 * @private
	 */
	this._location = null;

	this.content.append(this._filterContent);

	/**
	 * @member {ViewerToolbarButton}
	 * @private
	 */
	this._okButton = new ViewerToolbarButton(
		this.f,
		"AnnotationCustomFilterOkButton",
		$("<span/>"),
		false,
		_.bind(function() {
			this._filterModel.setCustomFilterActive(true); //set the filter to be active
			//to make sure the state doesn't get persisted with every small change in the popup, only persist when the popup is closed
			this._filterModel.fireFilterStateChanged(); //filter is applied and the popup gets closed.
			this.hide();
		}, this)
	);
	this._okButton.addClass("btn-success");

	/**
	 * @member {ViewerToolbarButton}
	 * @private
	 */
	this._cancelButton = new ViewerToolbarButton(
		this.f,
		"AnnotationCustomFilterCancelButton",
		$("<span/>"),
		false,
		_.bind(function() {
			this._filterModel.setFilterState(this._savedFilterState);
			this._savedFilterState = null;
			if (!this._wasCustomFilterActive) {
				this._filterModel.setCustomFilterActive(false);
			}
			this.hide();
		}, this)
	);
	this._cancelButton.addClass("btn");

	//append the buttons to a general container present in the popup (bottom right corner, location to add buttons)
	this.addButton(this._cancelButton.element);
	this.addButton(this._okButton.element);

	//initialize the template to be used for this popup
	this.t = new Template(this.f.i18n, mainTpl);
	this.t.registerTemplate("web-viewer/annotation/AnnotationCustomFilterRow", subTpl);

	this.t.setContext(this._filterContent); //the content is the pre-defined div

	//register several functions on the template
	//returns all the filter rows, adds the default filter row if there are none
	this.t.d.getFilterRows = _.bind(function() {
		var annotationFilters = this._filterModel.getFilterRows();

		return annotationFilters;
	}, this);

	//get the filter types that are available for the annotation filter
	this.t.d.getFilterTypes = _.bind(function() {
		return _.values(AnnotationConstants.FilterTypeNameLocs);
	}, this);

	//get all the filter options for a specific filter type
	this.t.d.getOptions = _.bind(function(selectedFilter, selectedOption) {
		var users = this._annotationModel.getUsers(AnnotationConstants.AnnotationOriginTypes.user.id);
		if (selectedFilter == AnnotationConstants.FilterTypes.AUTHOR) {
			if (selectedOption && users[selectedOption] === undefined) {
				users["hidden"] = this.f.i18n.t("AnnotFilterUserNotAvailable");
			}
			return users;
		} else if (
			selectedFilter == AnnotationConstants.FilterTypes.REVIEWSTATE ||
			selectedFilter == AnnotationConstants.FilterTypes.REVIEWAUTHORANDSTATE
		) {
			//for both review state and review author and state, translate review state string are needed
			var reviewStates = {}; //construct an object with same keys but translated review states
			_.each(
				AnnotationConstants.AnnotationReviewStateLocs,
				function(value, key) {
					reviewStates[key] = this.f.i18n.t(value); //translation possible, template rendering is only done if the language bundle is loaded
				},
				this
			);

			if (selectedFilter == AnnotationConstants.FilterTypes.REVIEWSTATE) {
				return reviewStates;
			} else {
				//REVIEWAUTHORANDSTATE
				var reviewAuthorAndState = {}; //construct object with a combination of the user name and review state
				_.each(
					users,
					function(user, index) {
						_.each(reviewStates, function(reviewState, key) {
							reviewAuthorAndState[user + " " + key] = user + " " + reviewState;
						});

						if (index !== _.last(_.keys(users))) {
							//between all review states of an user add a delimiter
							reviewAuthorAndState[user + "Delimiter"] = "──────────";
						}
					},
					this
				);
				if (selectedOption && reviewAuthorAndState[selectedOption] === undefined) {
					reviewAuthorAndState["hidden"] = this.f.i18n.t("AnnotFilterUserNotAvailable");
				}
				return reviewAuthorAndState;
			}
		}
	}, this);

	//add the listener to the popup, to be able to use the close button on the top right
	this.on(
		"closed",
		_.bind(function() {
			this.hide();
		}, this)
	);

	//add listeneres to key events to selectors and buttons of the by template rendered popup
	this.t.bind(
		"click",
		".addFilterRow",
		function(source, event) {
			var filterId = this._filterModel.addEmptyRow();
			this._addFilterRow(filterId);
		},
		this
	);

	this.t.bind(
		"click",
		".removeFilterRow",
		function(source, event) {
			var filterId = $(source).attr("data-bind");
			this._removeFilterRow(filterId);
		},
		this
	);

	//value in first dropdown changes, the third dropdownlist needs to be updated
	this.t.bind(
		"change",
		".filterType",
		function(source, event) {
			var filterId = $(source)
				.parent()
				.attr("data-bind");
			var selectedValue = $(source).val(); //get the currently selected option of the dropdown
			var options = this.t.d.getOptions(selectedValue); //get the options for the selected filter type (to be able to set default option)
			this._filterModel.updateFilterType(filterId, selectedValue, _.first(_.keys(options))); //take the first options as the default selected option
			//rerender the changed filter tow to make sure the options dropdown is filled with correct values
			this._rerenderFilterRow(filterId);
		},
		this
	);

	this.t.bind(
		"change",
		".filterCondition",
		function(source, event) {
			var filterId = $(source)
				.parent()
				.attr("data-bind");
			var selectedValue = $(source).val();
			this._filterModel.updateFilterCondition(filterId, selectedValue);
		},
		this
	);

	this.t.bind(
		"change",
		".filterOption",
		function(source, event) {
			var filterId = $(source)
				.parent()
				.attr("data-bind");
			var selectedValue = $(source).val();
			this._filterModel.updateFilterOption(filterId, selectedValue);
		},
		this
	);

	this.t.bind(
		"change",
		".filterOperatorSelector",
		function(source, event) {
			var selectedValue = $(source).val(); //operator (ALL or ANY) changed
			this._filterModel.setFilterOperator(selectedValue);
		},
		this
	);

	//load the language bundle to set the correct popup title and contents of the ok and cancel buttons
	this.f.i18n.loadBundle(
		"viewer",
		function() {
			this.setHeader(
				$("<span/>")
					.addClass("popupTitle")
					.html(this.f.i18n.t("CustomAnnotationFilter"))
			);

			this._okButton.element.html(this.f.i18n.t("ApplyCustomFilter"));
			this._cancelButton.element.html(this.f.i18n.t("AppDialogCancel"));
		},
		this
	);
};

AnnotationCustomFilterPopup.prototype = Object.create(Popup.prototype);
AnnotationCustomFilterPopup.prototype.constructor = AnnotationCustomFilterPopup;

/**
 * Renders the popup template
 */
AnnotationCustomFilterPopup.prototype.render = function() {
	this.f.i18n.loadBundle(
		"viewer",
		function() {
			this.t.d.filterOperator = this._filterModel.getFilterOperator();
			this.t.render();
		},
		this
	);
};

/**
 * Rerender a specific filter row (uses the subTemplate)
 * @param {Number} filterId
 */
AnnotationCustomFilterPopup.prototype._rerenderFilterRow = function(filterId) {
	var filterRow = this._filterModel.getFilterRow(filterId);
	this.content
		.find(".filterRow[data-bind='" + filterId + "']")
		.html(this.t.d.callTemplate("web-viewer/annotation/AnnotationCustomFilterRow", filterRow, "filter"));
};

/**
 * Add a row the filter, adds a row in the popup
 * @param {Number} filterId
 */
AnnotationCustomFilterPopup.prototype._addFilterRow = function(filterId) {
	var filterRow = this._filterModel.getFilterRow(filterId);
	this.content.find(".filterRows").append(this.t.d.callTemplate("web-viewer/annotation/AnnotationCustomFilterRow", filterRow, "filter"));
	this.recalculateSize(this._location, 475);
};

/**
 * Remove a filter row from the popup
 * @param {Number} filterId
 */
AnnotationCustomFilterPopup.prototype._removeFilterRow = function(filterId) {
	this.content.find(".filterRow[data-bind='" + filterId + "']").remove();
	this._filterModel.removeFilterRow(filterId);
	this.recalculateSize(this._location, 475);
};

/**
 * Show the popup, opening the popup sets the costom filter active
 * @param {ShapeUtils.Point} location The location of the popup
 */
AnnotationCustomFilterPopup.prototype.show = function(location) {
	this._savedFilterState = this._filterModel.getFilterState(); //save the custom filter state to be able to restore it
	this._wasCustomFilterActive = this._filterModel.isCustomFilterActive(); //save the custom filter state
	this.render(); //render the template
	this._location = location;
	this._filterModel.setCustomFilterActive(true);
	Popup.prototype.show.call(this, location, 475);
};

export default AnnotationCustomFilterPopup;
