import _ from "underscore";
import AbstractTool from "@esko/web-viewer/viewport/tools/AbstractTool";
import ShapeUtils from "@esko/shapeutils";
import GEOUtils from "@esko/web-viewer/utils/GEOUtils";
import DrawUtils from "@esko/web-viewer/utils/DrawUtils";
import AnnotationFactory from "@esko/web-viewer/annotation/AnnotationFactory";
import LineworkPoint from "@esko/web-viewer/utils/LineworkPoint";
import ListenerUtil from "@esko/web-viewer/utils/ListenerUtil";
import AnnotationConstants from "@esko/web-viewer/annotation/AnnotationConstants";
import ViewerUtils from "@esko/web-viewer/utils/ViewerUtils";
import CoreConstants from "@esko/web-viewer/core/CoreConstants";
/**
 * This is the tool for drawing annotations.
 * Also see it's related tool AnnotationNoteDrawTool.
 * @exports AnnotationDrawTool
 * @memberof AnnotationModule
 * @constructor
 *
 * @param {Core.ViewerFacade} f
 * @param {AnnotationModel} annotationModel - the model used by the module for annotations
 * @param {AnnotationModule} annotationModule - the annotation module reference
 */
var AnnotationDrawTool = function(f, annotationModel, annotationModule) {
	AbstractTool.call(this, f, CoreConstants.TOOLS.ANNOTATION_DRAW);
	ListenerUtil.mixin(this, f);

	var DEFAULTSTATE = "default";
	var DRAWSTATE = "draw";
	var DRAWINGSTATE = "drawing";

	var threshold = 0; //use this threshold to debug the freeform tool. This allows to only draw a couple of segments. When set to 0, it's harmless

	var newAnnotation;

	this.drawLinkedAnnotation = false;

	this._annotationModule = annotationModule;

	this._annotationModel = annotationModel;

	/**
	 * Set the type of annotation to draw
	 */
	this.setAnnotationType = function(type) {
		f.setOption("PREFERENCES_ANNOTATION_TOOLMODE", type);
	};

	/**
	 * Get the type of annotation to draw
	 */
	this.getAnnotationType = function() {
		return f.getOption("PREFERENCES_ANNOTATION_TOOLMODE");
	};

	/**
	 * Set the annotation color to draw with
	 */
	this.setAnnotationColor = function(color) {
		f.getCurrentUser().setColor("PREFERENCES_ANNOTATION_RECTANGLE_COLOR", color);
	};

	/**
	 * Get the annotation color to draw with
	 */
	this.getAnnotationColor = function() {
		if (f.getProperty("annotationcolor")) {
			return f.getProperty("annotationcolor");
		} else {
			return f.getCurrentUser().getColor("PREFERENCES_ANNOTATION_RECTANGLE_COLOR");
		}
	};

	/**
	 * @override
	 */
	this.getDefaultCursor = function() {
		return "cursor-drawAnnotation";
	};

	/**
	 * Adds a global annotation to the opened document(s)
	 */
	this.addGlobalAnnotation = function() {
		if (!this.drawLinkedAnnotation) {
			annotationModel.unselectAllAnnotations(true, true);
		}
		var worldBounds = f.getModule("viewport").getWorldBounds(); //get the world bounds, this is the rect of the annotation
		var l = getAnnotationLayer("Main"); //get the annotation layer for the Main-View
		if (l) {
			newAnnotation = new AnnotationFactory(f).create();
			newAnnotation.initAnnotation(this.getAnnotationType(), this.getAnnotationColor(), false, annotationModel);
			//set the rect to the world rectangle converted to scene coordinates to make sure it get converted back to the correct annotation world coordinates
			newAnnotation.rect = f
				.getModule("viewport")
				.getWorld()
				.getSceneRectFromWorld(worldBounds);
			doneDrawing("Main"); //trigger a done drawing and use the annotation layer of the Main-view
		}
	};
	//register the listener to the model to create global annotations
	this._registerListener(
		"addGlobalAnnotation",
		function() {
			if (this.f.isParameterUpdateOngoing()) {
				var paramChangedListener = this._registerListener(
					"parameters.changed",
					function() {
						this._unregisterListener(paramChangedListener);
						this.addGlobalAnnotation();
					},
					this,
					this.f.getDocumentController()
				);
			} else {
				this.addGlobalAnnotation();
			}
		},
		this,
		this._annotationModel
	);

	/**
	 * Get the annotation layer
	 */
	var getAnnotationLayer = _.bind(function(e) {
		var v = this._widget.requestView(e);
		if (v) {
			var l = v.getLayer("Annotations");
			if (l) {
				return l;
			}
		}
		return null;
	}, this);

	/**
	 * This is the mode when the tool is active,
	 * it will zoom on click and go the couldStartDraggingMode if left mousedown is clicked
	 */
	this.registerState(DEFAULTSTATE, {
		mousedown: function(e) {},
		mouseup: function(e) {},
		mousemove: function(e) {},
		mouseout: function() {},
		contextmenu: function(e) {
			// prevent the context menu
			e.preventDefault();
			e.stopPropagation();
			return false;
		},
		mouseleave: function() {},
		click: function() {},
		canBeInterrupted: true
	});

	/**
	 * Draw state
	 */
	this.registerState(
		DRAWSTATE,
		_.defaults(
			{
				mousedown: _.bind(function(e) {
					if (e.which === 1) {
						// Annotation created by this tool are always user annotation, so set the filter to user annotations
						this._annotationModel.getAnnotationFilterModel().setOriginType(AnnotationConstants.AnnotationOriginTypes.user.id);
						if (!this.drawLinkedAnnotation) {
							annotationModel.unselectAllAnnotations(true, true);
						}

						this.requestCanvasLayer(e);

						var touchPoint = this.keepPointWithinWorldBounds(this.getWorldPoint(e), true);
						newAnnotation = new AnnotationFactory(f).create();

						newAnnotation.initAnnotation(this.getAnnotationType(), this.getAnnotationColor(), false, annotationModel);
						newAnnotation.rect = new ShapeUtils.Rect(touchPoint, new ShapeUtils.Size());

						this.changeState(DRAWINGSTATE);
					}
					e.originalEvent.preventDefault();
				}, this)
			},
			this.getState(DEFAULTSTATE)
		)
	);

	/**
	 * Function that should be called when the drawing is finished (does correct cleanups etc)
	 */
	var doneDrawing = _.bind(function(e) {
		cleanupCanvas(); //clear the canvas

		if (!e) {
			newAnnotation = null; //if there was no event passes, put the newAnnotation to null
		}

		//If the new annotation is a line or an arrow (pointArray==2), make sure that the points are not on the same location
		if (newAnnotation && newAnnotation.pointArray.length === 2) {
			if (
				newAnnotation.pointArray[0].x == newAnnotation.pointArray[1].x &&
				newAnnotation.pointArray[0].y == newAnnotation.pointArray[1].y
			) {
				newAnnotation = null;
			}
		}

		//Make sure the annotation has at least some width or some height
		if (newAnnotation !== null && (Math.abs(newAnnotation.rect.size.width) === 0 || Math.abs(newAnnotation.rect.size.height) <= 0)) {
			//If both the width and height are 0, just return (not really useful to have a 0-sized annotation)
			if (Math.abs(newAnnotation.rect.size.width) === 0 && Math.abs(newAnnotation.rect.size.height) <= 0) {
				newAnnotation = null;
			} else {
				//If the box has either a 0-width or 0-height AND it's not a line/arrow, do not save the annotation
				if (
					newAnnotation.annottype != AnnotationConstants.AnnotationType.ANNOT_ARROW &&
					newAnnotation.annottype != AnnotationConstants.AnnotationType.ANNOT_LINE
				) {
					newAnnotation = null;
				}
			}
		}

		//If we have an annotation, let's save it
		if (newAnnotation) {
			var l = getAnnotationLayer(e);
			if (l) {
				//Convert the drawed annotation from Scene (pixels) to World (mm)
				newAnnotation.rect = l.getWorldRectFromScene(newAnnotation.rect); //first the box
				if (this._annotationModel.getPage().isReaderSpread()) {
					this._annotationModule.promptIfAnnotationOutsidePage(
						newAnnotation.rect,
						_.bind(function(isCreatingLinkedAnnot) {
							var shiftPressed = this.drawLinkedAnnotation;
							if (isCreatingLinkedAnnot && !shiftPressed) {
								this.drawLinkedAnnotation = true;
							}
							this.createAnnotation(newAnnotation, l);
							if (isCreatingLinkedAnnot && !shiftPressed) {
								this.drawLinkedAnnotation = false;
							}
						}, this),
						function() {
							newAnnotation = null;
						},
						this.drawLinkedAnnotation
					);
				} else {
					this.createAnnotation(newAnnotation, l);
				}
			} else {
				newAnnotation = null;
			}
		}

		this.changeState(DRAWSTATE);
	}, this);

	/**
	 * creates the annotation, makes a new annotation object, this will not save the annotation to the back-end
	 */
	this.createAnnotation = function(newAnnotation, annotationLayer) {
		for (var i = 0; i < newAnnotation.pointArray.length; i++) {
			//then the points
			var point = newAnnotation.pointArray[i];
			var p = new ShapeUtils.Point(point.x, point.y);
			var newp = annotationLayer.getWorldPointFromScene(p);
			point.x = newp.x;
			point.y = newp.y;
		}

		//Check if it need to be a linked annotation or a single new annotation
		if (this.drawLinkedAnnotation) {
			annotationModel.addLinkedAnnotation(newAnnotation);
		} else {
			annotationModel.addAnnotation(newAnnotation);
			if (this.isActive()) {
				this.requestDeactivation();
			}
		}
		newAnnotation = null;
	};

	/**
	 * Render the annotation via the draw utils
	 */
	var renderAnnotation = _.throttle(
		_.bind(function() {
			var ctx = this.getCanvasCtx();
			if (ctx) {
				// could be that the layer is already removed (caused by throttle)
				var csize = this.getCanvasSize();
				ctx.clearRect(0, 0, csize.width, csize.height);

				if (newAnnotation) {
					DrawUtils.drawOverlay(this.f, ctx, [newAnnotation]);
				}
			}
		}, this),
		40
	);

	/**
	 * Drawing state
	 */
	this.registerState(
		DRAWINGSTATE,
		_.defaults(
			{
				enterState: function() {
					f.emit("viewport.startInteractive");
					f.emit("layout.disableTextSelection");
				},
				leaveState: function() {
					f.emit("viewport.stopInteractive");
					setTimeout(function() {
						f.emit("layout.enableTextSelection");
					}, 500);
				},
				mousemove: _.bind(function(e) {
					var newRect = newAnnotation.rect.copy();
					var touchPoint = this.keepPointWithinWorldBounds(this.getWorldPoint(e), true);

					if (this.getAnnotationType() == AnnotationConstants.AnnotationType.ANNOT_FREEFORM) {
						var point = new LineworkPoint();
						if (newAnnotation.pointArray.length === 0) {
							point.makeMoveTo(newAnnotation.rect.origin.x, newAnnotation.rect.origin.y);
						} else {
							var previousPoint = newAnnotation.pointArray[newAnnotation.pointArray.length - 1];
							if (Math.abs(previousPoint.x - touchPoint.x) < threshold || Math.abs(previousPoint.y - touchPoint.y) < threshold) {
								return;
							}
							point.makeLineTo(touchPoint.x, touchPoint.y);
						}
						newAnnotation.pointArray.push(point);

						//Check if the width needs to be updated
						var width = Math.abs(point.x - newRect.origin.x);
						if (point.x < newRect.origin.x) {
							var diff = Math.abs(newRect.origin.x - point.x);
							newRect.origin.x = point.x;
							newRect.size.width += diff;
						} else if (width > newRect.size.width) {
							newRect.size.width = width;
						}

						//Check if the height needs to be updated
						var height = Math.abs(point.y - newRect.origin.y);
						if (point.y < newRect.origin.y) {
							var diff = Math.abs(newRect.origin.y - point.y);
							newRect.origin.y = point.y;
							newRect.size.height += diff;
						} else if (height > newRect.size.height) {
							newRect.size.height = height;
						}
					} else {
						newRect.size = newAnnotation.rect.origin.sizeTo(touchPoint);
					}
					newAnnotation.setRect(newRect, true);

					renderAnnotation();
				}, this),
				mouseclick: function(e) {},
				mouseup: function(e) {
					doneDrawing(e);
				},
				mouseleave: function(e) {
					doneDrawing();
				},
				mousedown: function(e) {
					e.originalEvent.preventDefault();
				},
				canBeInterrupted: false
			},
			this.getState(DEFAULTSTATE)
		)
	);

	/**
	 * Can the tool be interrupted ?
	 */
	this.canBeInterrupted = function() {
		return false;
	};

	this.activeKeyEvents = [];
	this.persistentKeyEvents = [];
	/**
	 * @override
	 */
	this._override("initialize", function(widget, toolManager) {
		this._super(widget, toolManager);
		this.changeState(DEFAULTSTATE);
		this.registerActiveEvents(["mousedown", "mousemove", "mouseup", "mouseout", "mouseleave", "contextmenu"]);

		this.activeKeyEvents.push([
			"shift",
			_.bind(function(e) {
				this.drawLinkedAnnotation = true;
			}, this),
			"keydown"
		]);

		this.activeKeyEvents.push([
			"shift",
			_.bind(function(e) {
				this.drawLinkedAnnotation = false;
			}, this),
			"keyup"
		]);
	});

	/**
	 * Cleanup canvas
	 */
	var cleanupCanvas = _.bind(function() {
		this.releaseCanvasLayer();
	}, this);

	/**
	 * @override
	 */
	this.canBeActivated = function() {
		return annotationModel.canChangeAnnotations();
	};

	/**
	 * @override
	 */
	this._override("activate", function(interruptRequest) {
		this._super(interruptRequest);

		/*
         * Initialize defaults
         */
		this.getAnnotationType();
		this.getAnnotationColor();

		annotationModel.setLockSelection(true);

		/*
         * Reconsider the state of the tool
         */
		this.changeState(DRAWSTATE);
	});

	/**
	 * @override
	 */
	this._override("deactivate", function(interruptRequest) {
		this._super(interruptRequest);
		this.changeState(DEFAULTSTATE);
		cleanupCanvas();

		this.drawLinkedAnnotation = false;
		annotationModel.setLockSelection(false);
	});

	f.getCurrentUser().on(
		"options.changed.PREFERENCES_ANNOTATIONTOOLS_VISIBLE",
		_.bind(function(oldValue, newValue) {
			if (!newValue && this.isActive()) {
				this.requestDeactivation();
			}
		}, this)
	);
};

// inherit
AnnotationDrawTool.prototype = Object.create(AbstractTool.prototype);
// correct the constructor pointer
AnnotationDrawTool.prototype.constructor = AnnotationDrawTool;

export default AnnotationDrawTool;
