import _ from "underscore";
import PrivateChannel from "@esko/web-viewer/core/PrivateChannel";
import AnnotationConstants from "@esko/web-viewer/annotation/AnnotationConstants";
/**
 * @class
 * Controller for filtering annotation, saves the filtered state in the filter model
 * @exports AnnotationFilter
 * @memberof AnnotationModule
 * @constructor
 * @param {Core.ViewerFacade} f - Viewer facade
 * @param {AnnotationFilterModel} annotationFilterModel
 */
var AnnotationFilter = function(f, annotationFilterModel) {
	PrivateChannel.installTo(this, f, ["annotationFilter"]);

	this._annotationFilterModel = annotationFilterModel;
};

/**
 * filter the annotation
 * @param {Array.<Annotation>} annotations
 * @returns {Array.<Annotation>}
 */
AnnotationFilter.prototype.filterAnnotations = function(annotations) {
	var filteredAnnotations = [];
	if (this._annotationFilterModel.isCustomFilterActive() && annotations) {
		filteredAnnotations = _.filter(
			annotations,
			_.bind(function(annotation) {
				return this.compliesToCustomFilter(annotation);
			}, this)
		);
		this._annotationFilterModel.setNumberOfFilteredAnnotations(filteredAnnotations.length);
		return filteredAnnotations;
	} else if (annotations) {
		filteredAnnotations = _.filter(
			annotations,
			_.bind(function(annotation) {
				return this.compliesToFilter(annotation);
			}, this)
		);
		this._annotationFilterModel.setNumberOfFilteredAnnotations(filteredAnnotations.length);
		return filteredAnnotations;
	} else {
		this._annotationFilterModel.setNumberOfFilteredAnnotations(0);
		return [];
	}
};

/**
 * Filters on which annotations need te be shown (depending on the viewer wide filtering of annotation authors)
 * Deletes review states and replies from the annotationModel of the page to be sure they are hidden in the annotation list
 * @param   {Array.<Annotation>} annotations
 * @returns {Array.<Annotation>}  array of the filtered annotations
 */
AnnotationFilter.prototype.filterIncomingAnnotations = function(annotations) {
	var filters = this._annotationFilterModel.getViewerUserFilter();
	if (filters.length > 0) {
		var annotsToRemove = []; //save top-level annotations that can be removed
		//now run over all our filtered annotations, and run through their replies/states to check the authors of those
		_.each(
			annotations,
			function(annotation) {
				//run through the review states and replies of the main annotation, the main annotation doesn't need be checked
				annotation.runThroughAnnotation(
					true,
					true,
					true,
					_.bind(function(annot) {
						//check if the author of the annotaiton complies to the filter (any of the filter authors is correct)
						var authorComplies = _.some(
							filters,
							function(filter) {
								return this._compliesToAuthor(annot, filter);
							},
							this
						);
						if (!authorComplies) {
							//if it doesn't comply to the filter remove the review state/reply
							if (annot.inreplyto && !annot.state) {
								//it's a reply, remove the reply from the parent
								var index = _.indexOf(annot.parent.replies, annot);
								annot.parent.replies.splice(index, 1); //save to remove, runThroughAnnotation makes this possible
								return true; //no need to check the children of the reply (review states)
							} else if (annot.state && annot.parent) {
								//it's a review state
								var index = _.indexOf(annot.parent.states, annot);
								annot.parent.states.splice(index, 1);
								return true;
							} else if (!annot.parent || annot.linkedto) {
								//if top level annotation add it to array to remove after foreach
								annotsToRemove.push(annot);
								return true; //no need to run over the replies and states of the annotation
							}
						}
						return false;
					}, this)
				);
			},
			this
		);
		//remove all the annotation that do not comply from the annotations list
		_.each(annotsToRemove, function(removeAnnot) {
			var index = _.indexOf(annotations, removeAnnot);
			annotations.splice(index, 1);
		});

		return annotations; //return the collection of the filteredAnnotations
	} else {
		//no filter was defined.
		return annotations;
	}
};

/**
 * check if the annotation needs to be shown with the user filter and state filter
 * @param {Annotation} annotation
 * @returns {Boolean}
 */
AnnotationFilter.prototype.compliesToFilter = function(annotation) {
	// If it isn't a user annotation, the filters don't really make sense
	var originType = this._annotationFilterModel.getOriginType();
	if (originType !== AnnotationConstants.AnnotationOriginTypes.user.id && originType === annotation.originType) {
		return true;
	}
	return this._compliesToUser(annotation) && this._compliesToState(annotation) && this._compliesToType(annotation);
};

/**
 * Does the annotation comply to the custom filter that has been set up
 * @param   {Annotation} annotation
 * @returns {Boolean}
 */
AnnotationFilter.prototype.compliesToCustomFilter = function(annotation) {
	var filters = this._annotationFilterModel.getFilterRows();

	if (this._annotationFilterModel.getFilterOperator() == AnnotationConstants.FilterOperators.ALL) {
		return _.every(
			filters,
			function(filter) {
				return this._checkCustomFilterRule(annotation, filter);
			},
			this
		);
	} else if (this._annotationFilterModel.getFilterOperator() == AnnotationConstants.FilterOperators.ANY) {
		return _.some(
			filters,
			function(filter) {
				return this._checkCustomFilterRule(annotation, filter);
			},
			this
		);
	}
};

/**
 * Check if the annotation complies to the custom filter rule
 * @param   {Annotation} annotation
 * @param   {Object}   filter
 * @returns {Boolean}
 * @private
 */
AnnotationFilter.prototype._checkCustomFilterRule = function(annotation, filter) {
	if (filter.filterType == AnnotationConstants.FilterTypes.AUTHOR) {
		if (filter.filterCondition == AnnotationConstants.FilterCondition.IS) {
			return this._compliesToUser(annotation, filter.filterOption);
		} else {
			return !this._compliesToUser(annotation, filter.filterOption);
		}
	} else if (filter.filterType == AnnotationConstants.FilterTypes.REVIEWSTATE) {
		if (filter.filterCondition == AnnotationConstants.FilterCondition.IS) {
			return this._compliesToCustomFilterState(annotation, filter.filterOption);
		} else {
			return !this._compliesToCustomFilterState(annotation, filter.filterOption);
		}
	} else if (filter.filterType == AnnotationConstants.FilterTypes.REVIEWAUTHORANDSTATE) {
		if (filter.filterCondition == AnnotationConstants.FilterCondition.IS) {
			return this._compliesToReviewAuthorAndState(annotation, filter.filterOption);
		} else {
			return !this._compliesToReviewAuthorAndState(annotation, filter.filterOption);
		}
	}
};

/**
 * does the annotation comply to the user filter
 * @private
 * @param {Annotation} annotation
 * @returns {Boolean}
 */
AnnotationFilter.prototype._compliesToUser = function(annotation, user) {
	var complies = true;
	if (user === undefined) {
		user = this._annotationFilterModel.getSelectedUser();
	}
	if (user) {
		complies = false;
		annotation.runThroughAnnotation(
			true,
			true,
			true,
			_.bind(function(annot) {
				if (annot.username == user) {
					complies = true;
					return true;
				}
			}, this)
		);
	}
	return complies;
};

AnnotationFilter.prototype._compliesToType = function(annotation) {
	return annotation.originType === this._annotationFilterModel.getOriginType();
};

/**
 * Does the annotation comply to the author
 * @private
 * @param   {Annotation}   annotation
 * @param   {String} author
 * @returns {Boolean}
 */
AnnotationFilter.prototype._compliesToAuthor = function(annotation, author) {
	if (author) {
		if (annotation.username == author) {
			return true;
		}
	}
	return false;
};

/**
 * Does the annotation comply to the author and state of the review
 * @private
 * @param   {Annotation} annotation
 * @param   {String} reviewAuthorAndState
 * @returns {Boolean}
 */
AnnotationFilter.prototype._compliesToReviewAuthorAndState = function(annotation, reviewAuthorAndState) {
	var complies = false;
	if (reviewAuthorAndState) {
		var reviewStateKeys = _.invert(AnnotationConstants.AnnotationReviewState);

		annotation.runThroughAnnotation(
			true,
			true,
			false,
			_.bind(function(annot) {
				var states = annot.states;
				complies = _.some(
					states,
					_.bind(function(state) {
						var annotReviewAuthorState = state.getUserName() + " " + reviewStateKeys[state.state];
						return reviewAuthorAndState === annotReviewAuthorState;
					}, this)
				);
				if (complies) {
					return true;
				}
			}, this)
		);
	}
	return complies;
};

/**
 * does the annotation comply to the state filter
 * @private
 * @param {Annotation} annotation
 * @returns {Boolean}
 */
AnnotationFilter.prototype._compliesToState = function(annotation) {
	var complies = true;
	var reviewState = this._annotationFilterModel.getSelectedState();
	if (reviewState) {
		complies = false;

		if (reviewState == "Any") {
			annotation.runThroughAnnotation(
				true,
				!this._annotationFilterModel.isInverseState(),
				false,
				_.bind(function(annot) {
					var states = annot.states;
					complies = complies || states.length > 0;
				}, this)
			);
		} else {
			// only check replies if it is not an inverse State
			annotation.runThroughAnnotation(
				true,
				!this._annotationFilterModel.isInverseState(),
				false,
				_.bind(function(annot) {
					var states = annot.states;
					var annotComplies = _.some(
						states,
						_.bind(function(state) {
							return state.state == reviewState;
						}, this)
					);
					if (annotComplies) {
						complies = true;
					}
				}, this)
			);
		}

		if (this._annotationFilterModel.isInverseState()) {
			complies = !complies;
		}
	}
	return complies;
};

/**
 * Does the annotation (with all its replies) comply to the custom filter state
 * @private
 * @param   {Annotation} annotation
 * @param   {String} reviewState
 * @returns {Boolean}
 */
AnnotationFilter.prototype._compliesToCustomFilterState = function(annotation, reviewState) {
	var complies = false;
	annotation.runThroughAnnotation(
		true,
		true,
		false,
		_.bind(function(annot) {
			var states = annot.states;
			var annotComplies = _.some(
				states,
				_.bind(function(state) {
					return state.state == AnnotationConstants.AnnotationReviewState[reviewState];
				}, this)
			);
			if (annotComplies) {
				complies = true;
			}
		}, this)
	);
	return complies;
};

export default AnnotationFilter;
