import PrivateChannel from "@esko/web-viewer/core/PrivateChannel";
import AnnotationConstants from "@esko/web-viewer/annotation/AnnotationConstants";
import AbstractModel from "@esko/web-viewer/core/AbstractModel";
import _ from "underscore";
import $ from "jquery";
/**
 * Model specific for the annotations filter
 * @exports AnnotationFilterModel
 * @memberof AnnotationModule
 * @class
 * @constructor
 *
 * @param {Core.ViewerFacade} f - Viewer facade
 */
var AnnotationFilterModel = function(f) {
	AbstractModel.call(this, f, "annotationFilterModel");
	this.f = f;

	/**
	 * @private
	 * @member {Array.<Object>} collection of all the filter rows
	 */
	this._filterRows = [];

	/**
	 * @private
	 * @member {AnnotationConstants.FilterOperators} the operator the combine different filter rows
	 */
	this._filterOperator = AnnotationConstants.FilterOperators.ALL;

	/**
	 * @private
	 * @member {Boolean} is the custom filter enabled or not
	 */
	this._isCustomFilterEnabled = false;

	/**
	 * @private
	 * @member {String|null} the selected user in the default filter
	 */
	this._selectedUser = null;

	/**
	 * @private
	 * @member {String|null} the selected state in the default filter
	 */
	this._selectedState = null;

	/**
	 * @private
	 * @member {Boolean} is the review state filter in the reverse state
	 */
	this._inverseState = false;

	/**
	 * @private
	 * @member {Number} the number of annotations after applying the filter
	 */
	this._numberOfFilteredAnnotations = 0;

	/**
	 * @private
	 * @member {String} the origin of annotation to filter on
	 */
	this._originType = AnnotationConstants.AnnotationOriginTypes.user.id;

	var viewerUserFilter = this.f.getProperty("showuser");
	if (viewerUserFilter) {
		this._viewerUserFilter = viewerUserFilter.split("|");
	} else {
		this._viewerUserFilter = [];
	}

	/**
	 * @private
	 * @member {Boolean} is the pinned state of the annotation filter enabled or not
	 */
	this._annotationFilterPinnedState = this.f.getOption("ANNOTATION_FILTER") !== undefined && this.f.getOption("ANNOTATION_FILTER") !== "";

	//if the pinned state is enabled, the option is set load the filter from this option
	if (this._annotationFilterPinnedState) {
		this.setFilterState(this.f.getOption("ANNOTATION_FILTER"));
	}
};

/**
 * reset the annotation filter
 */
AnnotationFilterModel.prototype.resetFilter = function() {
	this._selectedUser = null;
	this._selectedState = null;
	this._inverseState = false;
	this._originType = AnnotationConstants.AnnotationOriginTypes.user.id;
	this._filterRows = [];
	this._filterRows.push(this.getDefaultFilterRow());
	this._filterOperator = AnnotationConstants.FilterOperators.ALL;
	this._isCustomFilterEnabled = false;
	this._numberOfFilteredAnnotations = 0;
	this._emit("filterReset", this);
	this._emit("filterChanged", this);
};

/**
 * set the user to filter on
 * @param {String|null} user
 */
AnnotationFilterModel.prototype.setSelectedUser = function(user) {
	this._selectedUser = user;
	this._isCustomFilterEnabled = false;
	this._emit("filterChanged", this);
	this.fireFilterStateChanged();
};

/**
 * set the annotation origin type to filter on
 * @param {String} type
 */
AnnotationFilterModel.prototype.setOriginType = function(type) {
	if (type !== this._originType) {
		this._originType = type;
		this._isCustomFilterEnabled = false;
		this._emit("filterChanged", this);
		this.fireFilterStateChanged();
	}
};

/**
 * get the currently selected user
 * @returns {String}
 */
AnnotationFilterModel.prototype.getSelectedUser = function() {
	return this._selectedUser;
};

/**
 * get the currently selected origin
 * @returns {String}
 */
AnnotationFilterModel.prototype.getOriginType = function() {
	return this._originType;
};

/**
 * set the annotation state to filter on
 * @param {String|null} state
 * @param {Boolean} inverse - is it an inverse state, every but the state
 */
AnnotationFilterModel.prototype.setSelectedState = function(state, inverse) {
	this._selectedState = state;
	this._inverseState = inverse;
	this._isCustomFilterEnabled = false;
	this._emit("filterChanged", this);
	this.fireFilterStateChanged();
};

/**
 * get the currectly selected state
 * @returns {String}
 */
AnnotationFilterModel.prototype.getSelectedState = function() {
	return this._selectedState;
};

/**
 * Is the state the inverse state
 * @returns {Boolean}
 */
AnnotationFilterModel.prototype.isInverseState = function() {
	return this._inverseState;
};

/**
 * Add a filter row to the annotation filters
 *
 * @param {Number} filterRowId
 * @param {AnnotationConstants.FilterTypes} filterType
 * @param {AnnotationConstants.FilterCondition} filterCondition
 * @param {String} filterOption
 */
AnnotationFilterModel.prototype.addFilterRow = function(filterRowId, filterType, filterCondition, filterOption) {
	var filterRow = {
		filterId: filterRowId,
		filterType: filterType,
		filterCondition: filterCondition,
		filterOption: filterOption
	};

	this._filterRows[filterRowId] = filterRow;
};

/**
 * Add the default row to the annotation filter
 * @returns {Number}
 */
AnnotationFilterModel.prototype.addEmptyRow = function() {
	this._filterRows.push(this.getDefaultFilterRow());
	this._emit("filterChanged", this);
	return this._filterRows.length - 1;
};

/**
 * Update the filter type of a filter row, need to provide the default selected option for selected type
 * @param {Number} filterId
 * @param {AnnotationConstants.FilterTypes} filterType
 * @param {String} defaultOption
 */
AnnotationFilterModel.prototype.updateFilterType = function(filterId, filterType, defaultOption) {
	if (this._filterRows[filterId]) {
		this._filterRows[filterId].filterType = filterType;
		this._filterRows[filterId].filterOption = defaultOption;
		this._emit("filterChanged", this);
	}
};

/**
 * Update the filter option of a filter row
 * @param {Number} filterId
 * @param {String} filterOption
 */
AnnotationFilterModel.prototype.updateFilterOption = function(filterId, filterOption) {
	if (this._filterRows[filterId]) {
		this._filterRows[filterId].filterOption = filterOption;
		this._emit("filterChanged", this);
	}
};

/**
 * Update the condition of a filter row
 * @param {Number} filterId
 * @param {AnnotationConstants.FilterCondition} filterCondition
 */
AnnotationFilterModel.prototype.updateFilterCondition = function(filterId, filterCondition) {
	if (this._filterRows[filterId]) {
		this._filterRows[filterId].filterCondition = filterCondition;
		this._emit("filterChanged", this);
	}
};

/**
 * Remove a filter row from the model
 * @param {Number} filterRowId
 */
AnnotationFilterModel.prototype.removeFilterRow = function(filterRowId) {
	delete this._filterRows[filterRowId];
	this._emit("filterChanged", this);
};

/**
 * Remove all the filter rows from the model (including the default row)
 */
AnnotationFilterModel.prototype.removeAllFilterRows = function() {
	this._filterRows = [];
	this._emit("filterChanged", this);
};

/**
 * Get the filter row information for the row with filterRowId
 * @param   {Number} filterRowId
 * @returns {Object} object with all the information of specific row
 */
AnnotationFilterModel.prototype.getFilterRow = function(filterRowId) {
	return this._filterRows[filterRowId];
};

/**
 * Returns all the filter rows.
 * @returns {Array.<Object>}
 */
AnnotationFilterModel.prototype.getFilterRows = function() {
	if (this._filterRows.length === 0) {
		this._filterRows.push(this.getDefaultFilterRow());
	}
	return _.without(this._filterRows, null, undefined);
};

/**
 * Set all the filter rows
 * @param {Array.<Object>} filterRows
 */
AnnotationFilterModel.prototype.setFilterRows = function(filterRows) {
	this._filterRows = filterRows;
	this._emit("filterChanged", this);
};

/**
 * Get the default filter row (selected values are by default review state with option accepted)
 * @returns {Object}
 */
AnnotationFilterModel.prototype.getDefaultFilterRow = function() {
	return {
		filterId: this._filterRows.length,
		filterType: AnnotationConstants.FilterTypes.REVIEWSTATE,
		filterCondition: AnnotationConstants.FilterCondition.IS,
		filterOption: "STATUS_ACCEPTED"
	};
};

/**
 * Set the active state of the custom filter
 * @param {Boolean} setActive
 */
AnnotationFilterModel.prototype.setCustomFilterActive = function(setActive) {
	this._isCustomFilterEnabled = setActive;
	if (setActive && this._filterRows.length === 0) {
		this._filterRows.push(this.getDefaultFilterRow());
	}
	this._emit("filterChanged", this);
};

/**
 * Is the custom filter active
 * @returns {Boolean}
 */
AnnotationFilterModel.prototype.isCustomFilterActive = function() {
	return this._isCustomFilterEnabled;
};

/**
 * Set the filter operator.
 * @param {AnnotationConstants.FilterOperators} filterOperator
 */
AnnotationFilterModel.prototype.setFilterOperator = function(filterOperator) {
	this._filterOperator = filterOperator;
	this._emit("filterChanged", this);
};

/**
 * Returns the currently active filter operator.
 * @returns {AnnotationConstants.FilterOperators}
 */
AnnotationFilterModel.prototype.getFilterOperator = function() {
	return this._filterOperator;
};

/**
 * Set the number of filtered annotations.
 * @param {Number} number
 */
AnnotationFilterModel.prototype.setNumberOfFilteredAnnotations = function(number) {
	this._numberOfFilteredAnnotations = number;
};

/**
 * Get the number of filtered annotations.
 * @returns {Number}
 */
AnnotationFilterModel.prototype.getNumberOfFilteredAnnotations = function() {
	return this._numberOfFilteredAnnotations;
};

/**
 * Is the annotation filter pinned at this moment
 * @returns {Boolean}
 */
AnnotationFilterModel.prototype.isAnnotationFilterPinned = function() {
	return this._annotationFilterPinnedState;
};

/**
 * Toggle the annotation pinned state
 * @param {Boolean} [isEnabled] set the pinned state to
 */
AnnotationFilterModel.prototype.toggleAnnotationPinnedState = function(isEnabled) {
	if (isEnabled !== undefined) {
		this._annotationFilterPinnedState = isEnabled;
	} else {
		this._annotationFilterPinnedState = !this._annotationFilterPinnedState;
	}

	if (this._annotationFilterPinnedState) {
		this.fireFilterStateChanged(); //fire a change in the filter state so it gets persisted
	} else {
		this._emit("filterStateChanged", false); //reset the persisted filter
	}
};

/**
 * Get the state of the annotation filter
 * @returns {String} JSON string to be saved as option
 */
AnnotationFilterModel.prototype.getFilterState = function() {
	var stateObj = {
		customFilterOperator: this._filterOperator,
		customFilterRows: _.compact(this._filterRows), //compact needed to remove falsy values (null) from the array
		isCustomFilterActive: this.isCustomFilterActive(),
		selectedUser: this._selectedUser,
		selectedState: this._selectedState,
		inverseState: this._inverseState
	};
	return JSON.stringify(stateObj);
};

/**
 * Set the state of the annotation filter
 * @param {String} filterState
 */
AnnotationFilterModel.prototype.setFilterState = function(filterState) {
	try {
		filterState = JSON.parse(filterState);
		if (filterState) {
			this._isCustomFilterEnabled = filterState.isCustomFilterActive;
			this._filterOperator = filterState.customFilterOperator;
			this._filterRows = filterState.customFilterRows;
			_.each(
				this._filterRows,
				function(row, index) {
					row.filterId = index; //compensate for the compact in the get filter state
				},
				this
			);
			this._selectedUser = filterState.selectedUser;
			this._selectedState = filterState.selectedState;
			this._inverseState = filterState.inverseState;
		}
	} catch (e) {
		// something wrong with the filter state
		// fall back to default filter
	}
	this._emit("filterChanged", this);
};

/**
 * Fire a change in the filter state, this causes the filter state to be persisted
 */
AnnotationFilterModel.prototype.fireFilterStateChanged = function() {
	this._emit("filterStateChanged", true);
};

/**
 * Get the viewer filter on users, only show annotations, replies and states of the users in this array
 * @returns {Array.<String>}
 */
AnnotationFilterModel.prototype.getViewerUserFilter = function() {
	return this._viewerUserFilter;
};

export default AnnotationFilterModel;
