import Annotation from "@esko/web-viewer/annotation/Annotation";
import AnnotationConstants from "@esko/web-viewer/annotation/AnnotationConstants";
import GVAnnotationConstants from "@esko/web-viewer/annotation/GVAnnotationConstants";
import XMLUtils from "@esko/xmlutils";
import AnnotationGVTemplate from "./AnnotationGV.tpl.html";
import $ from "jquery";
import _ from "underscore";
/**
 * @class
 * @augments Annotation
 * @exports AnnotationGV
 * An annotation created by AE's GV nodes, these are read_only
 * @param {Core.ViewerFacade}
 * @param {String} originType - origin of the annotation, in this case "braille", "barcode", "text", "spelling" and "artwork" are correct.
 */
var AnnotationGV = function(f, originType) {
	Annotation.call(this, f, originType);

	/**
	 * @member{String} a wrapper containing images if there are any
	 */
	this.imageWrapper = null;
	/**
	 * @member{String} the tooltip for the image wrapper if any
	 */
	this.imageTooltip = null;
};

AnnotationGV.prototype = Object.create(Annotation.prototype);
AnnotationGV.prototype.constructor = AnnotationGV;

/**
 * @override
 */
AnnotationGV.prototype._parseContentsFromXML = function(xml) {
	var contentEls = xml.getElementsByTagName("contents-richtext");

	if (contentEls.length === 0) {
		contentEls = xml.getElementsByTagName("contents");
	}

	// Content
	var dataToRender = [];
	if (contentEls.length > 0) {
		var el = contentEls[0].getElementsByTagName("body")[0];
		if (!el) {
			el = contentEls[0];
		}

		// The contents of the file is one big string.
		// To parse it we go looking for certain Keys and create a substring up until the next newLine or up until a certain stop character
		var text = el.textContent;
		if (GVAnnotationConstants.GVTypes[this.originType] && GVAnnotationConstants.GVTypes[this.originType]["contentProperties"]) {
			// GVAnnotationConstants has defined the keys to display

			_.each(
				GVAnnotationConstants.GVTypes[this.originType]["contentProperties"],
				function(obj, key) {
					var begin = text.indexOf(key);

					if (begin === -1) {
						return;
					}
					var substr = text.substr(begin + key.length);

					var substrLength;

					// Check how far the substring should go, if no 'stop' property was defined, go to the first newline character
					if (!obj.stop) {
						substrLength = substr.indexOf("\n");
						if (substrLength === -1) {
							substrLength = undefined;
						}
					}
					// Else just keep the rest of the string

					var value = substr.substr(0, substrLength).trim();

					// Check for I18n of the value
					if (
						GVAnnotationConstants.GVTypes[this.originType]["valueProperties"] &&
						GVAnnotationConstants.GVTypes[this.originType]["valueProperties"][value]
					) {
						value = this.f.i18n.t(GVAnnotationConstants.GVTypes[this.originType]["valueProperties"][value]);
					}

					dataToRender.push({
						key: this.f.i18n.t(obj.langKey),
						value: value
					});
				},
				this
			);
			this.content = AnnotationGVTemplate({
				lines: dataToRender
			});
		} else {
			this.content = "<p>" + el.textContent + "</p>";
		}
	}

	// Images
	var tMaster = XMLUtils.attributeValue(xml, "tMaster"); // returns null when not present
	var tSample = XMLUtils.attributeValue(xml, "tSample");
	if (tMaster !== null && tSample !== null) {
		var masterImage = "<img class='masterImage' src='data:image/png;base64," + tMaster + "' alt='Master'/>";
		var sampleImage = "<img class='sampleImage' src='data:image/png;base64," + tSample + "' alt='Sample'/>";

		this.imageWrapper = "<div class='imageWrapper'>" + masterImage + sampleImage + "</div>";

		// if both present, use image and add tooltip
		this.content = "";

		var tooltip = "";
		_.each(dataToRender, function(data, index) {
			if (index > 0) {
				tooltip += "<br />";
			}
			tooltip += data.key + " " + data.value;
		});
		this.imageTooltip = tooltip;
	} else {
		this.imageWrapper = null;
		this.imageTooltip = null;
	}

	// Set the title
	this.lastname = this.f.i18n.t(this.originType + "Title");
};

/**
 * @override
 */
AnnotationGV.prototype.initWithXML = function(xml, boxOffsets, msgnr) {
	Annotation.prototype.initWithXML.call(this, xml, boxOffsets, msgnr);
	this.messagenumber = msgnr;
};

/**
 * @override
 */
AnnotationGV.prototype._modifyAnnotatationCoords = function(x1, y2, x2, y1, boxOffsets) {
	// GV Annotations are relative to the mediabox
	// y offset doesn't get a minus because the world origin is top left, while the trim and mediaboxes origin is bottom left
	var xOffset = -(boxOffsets.trimboxOffset.width - boxOffsets.mediaboxOffest.width);
	var yOffset = boxOffsets.trimboxOffset.height - boxOffsets.mediaboxOffest.height;
	return [x1 + xOffset, y2 + yOffset, x2 + xOffset, y1 + yOffset];
};

/**
 * @override
 */
AnnotationGV.prototype.getPermission = function() {
	return AnnotationConstants.AccessRight.READ_ONLY;
};

/**
 * override
 */
AnnotationGV.prototype.getPermissionFor = function(action) {
	switch (action) {
		case AnnotationConstants.ActionPermissions.DUPLICATE:
			return this.originType !== AnnotationConstants.AnnotationOriginTypes.artwork.id;
		case AnnotationConstants.ActionPermissions.DUPLICATE_WITHOUT_TEXT:
			return true;
		case AnnotationConstants.ActionPermissions.CAN_BE_REPLIED_TO:
			return false;
		case AnnotationConstants.ActionPermissions.ZOOM_TO_ANNOTATION:
			return true;
	}
};

/**
 * GV annotations are special in the sense that duplicating them creates a non-GV user annotation.
 * Annotation.prototype.duplicate however uses this.constructor to create a duplicate. We don't want to create a new GV however, so we proxy it with the Annotation constructor
 */
AnnotationGV.prototype.duplicate = function() {
	var myProxy = Object.create(this);
	myProxy.constructor = Annotation;
	return Annotation.prototype.duplicate.apply(myProxy, arguments);
};

/**
 * @override
 */
AnnotationGV.prototype.getExtraContent = function() {
	if (this.imageWrapper) {
		return {
			element: $(this.imageWrapper),
			tooltip: this.imageTooltip
		};
	}
	return null;
};

/**
 * @override
 */
AnnotationGV.prototype.getTextContent = function() {
	if (this.imageTooltip) {
		return this.imageTooltip;
	} else {
		return Annotation.prototype.getTextContent.call(this);
	}
};

/**
 * @override
 */
AnnotationGV.prototype.getIconClass = function() {
	if (
		this.originType === AnnotationConstants.AnnotationOriginTypes.braille.id ||
		this.originType === AnnotationConstants.AnnotationOriginTypes.barcode.id ||
		this.originType === AnnotationConstants.AnnotationOriginTypes.text.id ||
		this.originType === AnnotationConstants.AnnotationOriginTypes.artwork.id ||
		this.originType === AnnotationConstants.AnnotationOriginTypes.spelling.id
	) {
		return AnnotationConstants.AnnotationOriginTypes[this.originType].subIcon;
	} else {
		return Annotation.prototype.getIconClass.call(this);
	}
};

export default AnnotationGV;
