import _ from "underscore";
import AbstractTool from "@esko/web-viewer/viewport/tools/AbstractTool";
import ViewportLayerModel from "@esko/web-viewer/viewport/model/ViewportLayerModel";
import ShapeUtils from "@esko/shapeutils";
import AnnotationConstants from "@esko/web-viewer/annotation/AnnotationConstants";
import CoreConstants from "@esko/web-viewer/core/CoreConstants";
/**
 * @class Tool to highlight annotations on the viewport
 * @exports AnnotationIndicatorTool
 * @memberof AnnotationModule
 * @constructor
 *
 * @param {Core.ViewerFacade} f
 * @param {AnnotationModel} annotationModel - the model used by the module for annotations
 */
var AnnotationIndicatorTool = function(f, annotationModel) {
	AbstractTool.call(this, f, CoreConstants.TOOLS.ANNOTATION_INDICATOR);
	this.f = f;

	this.layerModel = new ViewportLayerModel(f, "annotationIndication");
	this.layerModel.annotation = null;
	this.layerModel.tool = this;
	this.layerModel.annotationModel = annotationModel;
	this.layerModel.alpha = 0.7;

	/**
	 * Render the annotation indicator, highlights the annotation
	 * @param {Context2D} ctx
	 * @param {ShapeUtils.Rect} drawrect
	 */
	this.layerModel.render = function(ctx, drawrect) {
		ctx.fillStyle = "rgba(0, 0, 0, " + this.alpha + ")";

		ctx.clearRect(drawrect.origin.x, drawrect.origin.y, drawrect.size.width, drawrect.size.height);

		ctx.beginPath();
		ctx.rect(drawrect.origin.x, drawrect.origin.y, drawrect.size.width, drawrect.size.height);

		ctx.fill();

		ctx.save();
		if (this.annotation) {
			var inset = Math.abs(this.tool.sceneToWorldSize(new ShapeUtils.Size(10, 10)).width);
			var page = this.annotationModel.getAnnotationsPage();
			if (page !== null && page !== undefined) {
				this.resetTransformation();
				this.getTransform().scale(1, -1);

				var rotation = page.getRotation();

				var worldBounds = this.f.getModule("viewport").getWorldBounds();

				this.calculateTransformation(
					worldBounds,
					page.getPageSize(),
					this.annotationModel.getAnnotationsOffset(),
					rotation,
					page.isFlipped()
				);
			}

			var r = this.annotation.rect.copy();

			r.normalize();

			ctx.clearRect(r.origin.x - inset, r.origin.y + r.size.height + inset, r.size.width + 2 * inset, -r.size.height - 2 * inset);

			var links = this.annotationModel.getAnnotationLinks(this.annotation);
			for (var i = 0; i < links.length; i++) {
				r = links[i].rect.copy();

				r.normalize();

				ctx.clearRect(r.origin.x - inset, r.origin.y + r.size.height + inset, r.size.width + 2 * inset, -r.size.height - 2 * inset);
			}
		}
		ctx.restore();
	};

	this._override("activate", function() {
		this._super();

		this.f.getModule("viewport").addLayer("annotationIndication", this.layerModel);
	});

	this._override("deactivate", function(interruptRequest) {
		this._super(interruptRequest);

		this.f.getModule("viewport").removeLayer("annotationIndication");
	});
};

// inherit
AnnotationIndicatorTool.prototype = Object.create(AbstractTool.prototype);
// correct the constructor pointer
AnnotationIndicatorTool.prototype.constructor = AnnotationIndicatorTool;

/**
 * Can the tool be interrupted ?
 */
AnnotationIndicatorTool.prototype.canBeInterrupted = function() {
	return false;
};

/**
 * Set the currently active annotation
 * @param {Annotation} annotation
 */
AnnotationIndicatorTool.prototype.setAnnotation = function(annotation) {
	this.layerModel.annotation = annotation;
};

export default AnnotationIndicatorTool;
