import ViewportLayerModel from "@esko/web-viewer/viewport/model/ViewportLayerModel";
import ShapeUtils from "@esko/shapeutils";
import ListenerUtil from "@esko/web-viewer/utils/ListenerUtil";
import CoreConstants from "@esko/web-viewer/core/CoreConstants";
import _ from "underscore";

const ANNOTATION_TOOLS = [
	CoreConstants.TOOLS.ANNOTATION_TOUCH_DRAW,
	CoreConstants.TOOLS.ANNOTATION_TOUCH,
	CoreConstants.TOOLS.ANNOTATION_DRAW,
	CoreConstants.TOOLS.ANNOTATION_NOTE_DRAW,
	CoreConstants.TOOLS.ANNOTATION_INDICATOR
];

/**
 * Create a new annotation layer model
 * @exports AnnotationLayerModel
 * @memberof AnnotationModule
 *
 * @class This class is responsible for drawing annotations	 *
 * @constructor
 * @param {Core.ViewerFacade} f - Viewerfacade
 * @param {String} ID - Id of the layer
 * @param {AnnotationModel} annotationModel - Model of the annotations.
 */
var AnnotationLayerModel = function(f, ID, annotationModel) {
	ViewportLayerModel.call(this, f, ID);
	ListenerUtil.mixin(this, f);

	this.f = f;
	this.annotationModel = annotationModel;

	ANNOTATION_TOOLS.forEach(tool => {
		this._registerListener("tool." + tool + ".activated", this.modelChanged, this);
	});
	this._registerListener("modelChanged", this._calculateTransformation, this, this.annotationModel);
	this._registerListener("modelContentChanged", this.modelChanged, this, this.annotationModel);
	this._registerListener("parameters.changed", this._calculateTransformation, this, this.f.getDocumentController());
	this._registerListener("transform.viewportWidget.transformChanged", this._calculateTransformation, this);
};

AnnotationLayerModel.prototype = Object.create(ViewportLayerModel.prototype);
AnnotationLayerModel.prototype.constructor = AnnotationLayerModel;

/**
 * Calculates the transformation for the annotation layer, takes compensation for distortion and scaling into account
 */
AnnotationLayerModel.prototype._calculateTransformation = function() {
	var page = this.annotationModel.getAnnotationsPage();
	if (page !== null && page !== undefined) {
		this.resetTransformation();
		this.getTransform().scale(1, -1);

		var rotation = page.getRotation();

		var worldBounds = this.f.getModule("viewport").getWorldBounds();

		var offset = new ShapeUtils.Size(); //start constructing the offset needed for the origin of the annotations
		offset.add(this.annotationModel.getAnnotationsOffset()); //add the annotation offset (used in compare)

		//calculate the rest of the tranformation, for rotation, flipped state
		this.calculateTransformation(worldBounds, page.getPageSize(), offset, rotation, page.isFlipped());
	}
	this.modelChanged();
};

/**
 * @override
 */
AnnotationLayerModel.prototype.getLinework = function(animated, rotationanimation) {
	var visibleAnnotations = this.annotationModel.getVisibleAnnotations();
	var linkedAnnotations = [];
	_.each(visibleAnnotations, function(annotation) {
		_.each(annotation.links, function(linkedAnnot) {
			linkedAnnotations.push(linkedAnnot);
		});
	});

	visibleAnnotations = _.union(visibleAnnotations, linkedAnnotations);

	/**
	 * Box with the anchor points to allow resizing of the annotation
	 */
	var selectionBoxes = [];

	/**
	 * Show sequence number
	 */
	var showNumber = this.f.getOption("PREFERENCES_ANNOTATION_SEQLABEL_VISIBLE") && !rotationanimation;

	/**
	 * Iterate over all visible annotations and add the correct info to it
	 */
	for (var i = 0; i < visibleAnnotations.length; i++) {
		var annot = visibleAnnotations[i];

		if (this.annotationModel.isAnnotationOrReplySelected(annot)) {
			annot.setSelected(true);
			if (annot.needsSelectionBox() && this.annotationModel.canChangeAnnotation(annot)) {
				selectionBoxes[i] = annot.getSelectionBox();
			}
		} else {
			annot.setSelected(false);
		}

		if (showNumber) {
			annot.updateIndicator();
			annot.showIndicator = true;
		} else {
			annot.showIndicator = false;
		}
	}

	//Return the union of the annotations + their selection boxes (flatten to make it 1 dimensional array)
	return _.union(visibleAnnotations, linkedAnnotations, _.without(_.flatten(selectionBoxes), undefined, null)).map(
		function(annot) {
			if (this.f.getProperty("forceAnnotationColor")) {
				annot.annotationcolor = this.f.getProperty("forceAnnotationColor");
				annot.color = this.f.getProperty("forceAnnotationColor");
			}
			return annot;
		}.bind(this)
	);
};

export default AnnotationLayerModel;
