import PrivateChannel from "@esko/web-viewer/core/PrivateChannel";
import AnnotationFilter from "@esko/web-viewer/annotation/AnnotationFilter";
import AnnotationFactory from "@esko/web-viewer/annotation/AnnotationFactory";
import AnnotationConstants from "@esko/web-viewer/annotation/AnnotationConstants";
import AnnotationFilterModel from "@esko/web-viewer/annotation/AnnotationFilterModel2";
import Lock from "@esko/web-viewer/annotation/Lock";
import ShapeUtils from "@esko/shapeutils";
import ListenerUtil from "@esko/web-viewer/utils/ListenerUtil";
import ViewerUtils from "@esko/web-viewer/utils/ViewerUtils";
import ModalPopup from "@esko/web-viewer/layout/component/ModalPopup";
import ViewerToolbarButton from "@esko/web-viewer/layout/component/ViewerToolbarButton";
import LayoutConstants from "@esko/web-viewer/layout/LayoutConstants";
import OrderedObject from "@esko/orderedobject";
import _ from "underscore";
import $ from "jquery";
/**
 * Constructs a new annotation model
 * @exports AnnotationModel
 * @memberof AnnotationModule
 * @class
 * The annotation model is the main model within the viewer that keeps track of the current annotations and their selection. The model has its main function "setAnnotations". It keeps track of the selected annotations
 * @constructor
 * @param {Core.ViewerFacade} f - Viewer facade
 * @param {String} IID - Unique channel ID of the annotation model
 */
var AnnotationModel = function(f, IID) {
	PrivateChannel.installTo(this, f, ["annotationModel", IID]);
	ListenerUtil.mixin(this, f);

	/**
	 * Facade
	 */
	this.f = f;

	/**
	 * In here every reason to disable annotation editing is logged
	 * {
	 * 	reason: reason,
	 *	hideAnnotations: hideAnnotations,
	 *	dontRemove: dontRemove
	 *	}
	 */
	this._disabledReasons = new OrderedObject();

	/**
	 * The page too which the current annotations belong
	 * @private
	 * @type {Data.AbstractDocumentPage}
	 */
	this._page = null;

	/**
	 * Annotations of the main document
	 */
	this._annotations = [];
	/**
	 * Annotations that are removed by the user, but not yet saved
	 */
	this._removedannotations = [];

	/**
	 * Keeps the list of annotations that are selected
	 */
	this._selectedAnnotations = [];
	/**
	 * Keeps the list of the annotations that are filtered
	 */
	this._filteredAnnotations = [];

	/**
	 * for which layout type does the wiget need to be rendered
	 */
	this._layoutType = null;

	/**
	 * the model specific for the annotation filter
	 */
	this._annotationFilterModel = new AnnotationFilterModel(this.f);

	/**
	 * variable used for stepping through annotations
	 */
	this._orderedAnnotations = [];
	//index gets first increment before it is used, make sure the first one is index 0
	this._currentAnnotationIndex = -1;

	/**
	 *  The currently active scaling, all annotation are scaled with this scaling. This is used to initialize new annotations.
	 */
	this._currentScaling = null;

	/**
	 * Can be set to lock the selection of annotations
	 */
	this.lockSelection = false; // A tool/module can decide that the selection will be locked and take the ownership for deselecting annotations
	// This is needed to prevent 2 tools (e.g. AnnotationTool and AnnotationDrawTool to get in each other ways

	/**
	 * Listeners used in adding global annotation on load (from SDK) to make sure all of annotations are loaded.
	 */
	this._addAnnotationsEnabledListener = null;
	this._disabledReasonChangedListener = null;

	/**
	 * Indicates if the annotation table is visible
	 */
	this._annotationListVisible = false; // Is the list visible? Needed for visibility options of annotations & popups

	/**
	 * Indicates if the annotations are currently being saved.
	 */
	this._isSavingAnnotations = false;

	/**
	 * Indicates if all annotation popups should be visible, whether the popup is pinned or not.
	 */
	this._showAllAnnotationPopups = this.f.getOption("PREFERENCES_SHOW_ALL_ANNOTATION_POPUPS") > 0;

	/**
	 * Indicates if the viewer should auto fit view to the selected annotation;
	 */
	this._autoFitViewToSelectedAnnotation = this.f.getOption("PREFERENCES_ANNOTATION_FITVIEW") > 0;

	var currentUser = f.getCurrentUser();
	/**
	 * If the Sequence label visibility changes, redraw the annotations
	 */
	currentUser.on(
		"options.changed.PREFERENCES_ANNOTATION_SEQLABEL_VISIBLE",
		_.bind(function(oldValue, newValue) {
			this.modelChanged();
		}, this)
	);

	/**
	 * If the preferences changes, redraw the annotations
	 */
	currentUser.on(
		"options.changed.PREFERENCES_ANNOTATION_NEXT_MODE",
		_.bind(function(oldValue, newValue) {
			this._orderedAnnotations.length = 0;
			this.modelChanged();
		}, this)
	);

	// listen to the datesort mode
	currentUser.on(
		"options.changed.PREFERENCES_ANNOTATION_DATESORT_MODE",
		_.bind(function(oldValue, newValue) {
			this.setSortingMode(newValue);
		}, this)
	);

	currentUser.on(
		"options.changed.PREFERENCES_ANNOTATION_SHOWSELECTEDONLY",
		_.bind(function() {
			this.modelChanged();
		}, this)
	);

	this.setSortingMode(f.getOption("PREFERENCES_ANNOTATION_DATESORT_MODE"));

	this.setDisableReason("LOADING_MODULES", this.f.t("Loading"), true);

	f.whenAllModulesAreLoaded(
		function() {
			f.loadBundle(
				"viewer",
				function() {
					if (this._page) {
						this._emit("setdisable", this._page);
						this.removeDisableReason("LOADING_MODULES");
					}
				}.bind(this)
			);
		}.bind(this)
	);

	/**
	 * Variable that indicates if a warning should be shown if the annotations should changed
	 */
	this.warnWhenChangingAnnotations = new Lock();

	/**
	 * Variable that remembers the confirmation of the warning
	 */
	this._confirmedOnce = false;

	/**
	 * Filter object that keeps track of the current filter
	 */
	this._filter = new AnnotationFilter(f, this._annotationFilterModel);

	/**
	 * If the filter changes, refilter all the annotations
	 */
	this._annotationFilterModel.on(
		"filterChanged",
		_.bind(function(filter) {
			this._filterAllAnnotations();
			this._orderedAnnotations = [];
		}, this)
	);

	/**
	 * The annotations can be temporarily disabled with a certain reason
	 * the reason is shown in the annotation list
	 * if enableAnnotationToggling is true, buttons to show either current document or reference anntoations are available
	 * if showingDifferenceAnnotations is true then the annotations of the differnce document are shown, these cannot be changed
	 */
	this._temporarilyDisabled = {
		isTemporarilyDisabled: false,
		reason: "",
		enableAnnotationToggling: false,
		showingCurrentAnnotations: false,
		showingDifferenceAnnotations: false
	};

	this.f.whenModuleAvailable(
		"video",
		function(videoModule) {
			if (videoModule) {
				this._videoModule = videoModule;
				this._videoModule.whenVideoAvailable(
					function() {
						this._emit("modelChanged");
					}.bind(this)
				);
			}
		}.bind(this)
	);
};

/**
 * Sets the variable that indicates if the annotation table is visible
 * @param {Boolean} annotationListVisible
 */
AnnotationModel.prototype.setAnnotationListVisible = function(annotationListVisible) {
	if (this._annotationListVisible !== annotationListVisible) {
		this._annotationListVisible = annotationListVisible;
		this._emit("annotationListVisible", annotationListVisible);
	}
};

/**
 * Is the annotation list/table visible?
 * Used by e.g. AnnotationPopupsLayerModel to show the popups based on the visible state
 * @returns {Boolean} True if the annotation list/table is visible
 */
AnnotationModel.prototype.isAnnotationListVisible = function() {
	return this._annotationListVisible;
};

/**
 * Get the current ordering function
 * @returns {Function} function to order the annotations
 */
AnnotationModel.prototype.currentOrdening = function() {
	return this.orderByCreationDate;
};

/**
 * Toggle the annotation visibility
 * @param {Boolean} [toggle] - force visibility
 */
AnnotationModel.prototype.toggleVisibility = function(toggle) {
	if (toggle === undefined) {
		toggle = this.f.getOption("PREFERENCES_ANNOTATIONTOOLS_VISIBLE") == 1;
		toggle = !toggle;
	} else if (toggle && this.f.getOption("PREFERENCES_ANNOTATIONTOOLS_VISIBLE") == 1) {
		// yes! we compare with == and not with === (we are not sure what is in the preference (true, false)
		// nothing changed
		return;
	}

	if (!toggle) {
		this.f.setOption("PREFERENCES_ANNOTATIONTOOLS_VISIBLE", 0);
		this.unselectAllAnnotations();
	} else {
		this.f.setOption("PREFERENCES_ANNOTATIONTOOLS_VISIBLE", 1);
	}
	this.modelChanged();
};

/**
 * Temporarily disable the module
 * @param {Boolean} temporarilyDisable - if true, annotations are temporarily disabled (this setting is not saved)
 * @param {Boolean} nableTogglingAnnotations - if true the user will have the option to toggle annotation on and off
 * @param {String} reason - the reason why annotation is temporarily disabled, this is displayed to the user
 * @param {String} [whichToEnable] - [difference|current] if provided, either the difference or current annotations will be shown
 */
AnnotationModel.prototype.temporarilyDisable = function(temporarilyDisable, enableTogglingAnnotations, reason, whichToEnable) {
	this._temporarilyDisabled.isTemporarilyDisabled = temporarilyDisable;
	this._temporarilyDisabled.reason = reason;
	this._temporarilyDisabled.enableAnnotationToggling = enableTogglingAnnotations;

	this.unselectAllAnnotations();

	if (whichToEnable !== undefined) {
		// if which to enable is set, immediately show these temporary annotations
		if (whichToEnable == "difference") {
			this.showTemporarilyDifferencePageAnnotations();
		} else if (whichToEnable == "current") {
			this.showTemporarilyActivePageAnnotations();
		}
		this.f.setOption("PREFERENCES_ANNOTATIONTOOLS_VISIBLE", 1, true);
	} else {
		// just disable annotations, don't show anything special
		this._temporarilyDisabled.showingCurrentAnnotations = false;
		this._temporarilyDisabled.showingDifferenceAnnotations = false;
	}

	this.updateTemporarilyDisabledReason();

	this._emit("temporarilyDisableChanged");

	if (!temporarilyDisable) {
		this.setPage(this.f.getActivePage());
		this._temporarilyDisabled.showingCurrentAnnotations = true;
	}
	this.modelChanged();
};

/**
 * Register the disable reason
 */
AnnotationModel.prototype.updateTemporarilyDisabledReason = function() {
	this.f.loadBundle(
		"viewer",
		function() {
			if (this.isTemporarilyDisabled() && !this.isTemporarilyCurrentPageAnnotationsShown()) {
				this.setDisableReason("TEMPORARY_DISABLE", this._temporarilyDisabled.reason);

				if (this.isTemporarilyDifferencePageAnnotationsShown()) {
					this.setDisableReason("TEMPORARY_DISABLE", this.f.t("AnnDisabledDiffMode"));
				}
			} else {
				this.removeDisableReason("TEMPORARY_DISABLE");
			}
		},
		this
	);
};

/**
 * Check wether the annotation module is temporarily disabled and no annotations are visible
 * @returns {Boolean}
 */
AnnotationModel.prototype.isTemporarilyDisabledAndNoAnnotationsVisible = function() {
	return (
		(this._temporarilyDisabled.isTemporarilyDisabled &&
			!this._temporarilyDisabled.showingCurrentAnnotations &&
			!this._temporarilyDisabled.showingDifferenceAnnotations) ||
		!this.isAllowedToShowAnnotations()
	);
};

/**
 * Are annotations temporarily disabled, only checks if there are temporarily disabled reason, not other reasons!
 * @returns {Boolean}
 */
AnnotationModel.prototype.isTemporarilyDisabled = function() {
	return this._temporarilyDisabled.isTemporarilyDisabled;
};

/**
 * Are annotations temporarily disabled and the user cannot edit
 * @returns {Boolean}
 */
AnnotationModel.prototype.isTemporarilyDisabledAndCannotEdit = function() {
	return (
		(this._temporarilyDisabled.isTemporarilyDisabled && !this._temporarilyDisabled.showingCurrentAnnotations) ||
		!this.isAllowedToShowAnnotations()
	);
};
/**
 * Get the mssage to be shown when no annotations are visible because of temporarily disable
 * @returns {String}
 */
AnnotationModel.prototype.getTemporarilyDisabledMessage = function() {
	return this._temporarilyDisabled.reason;
};

/**
 * Is the annotation module temporarily disabled and the annotations for the current document are shown
 * @returns {Boolean}
 */
AnnotationModel.prototype.isTemporarilyCurrentPageAnnotationsShown = function() {
	return this._temporarilyDisabled.showingCurrentAnnotations && this.isAllowedToShowAnnotations();
};

/**
 * True if annotations for the difference document are shown
 * @returns {Boolean}
 */
AnnotationModel.prototype.isTemporarilyDifferencePageAnnotationsShown = function() {
	return this._temporarilyDisabled.showingDifferenceAnnotations && this.isAllowedToShowAnnotations();
};

/**
 * Can the user re-enable annotations while the annotationmodule is temporarily disabled
 * @returns {Boolean}
 */
AnnotationModel.prototype.isTemporarilyAnnotationDocumentTogglingEnabled = function() {
	return this._temporarilyDisabled.enableAnnotationToggling && this.isAllowedToShowAnnotations();
};

/**
 * Checks if there are disable reasons that don't allow to show annotations at all
 * @returns {Boolean}
 */
AnnotationModel.prototype.isAllowedToShowAnnotations = function() {
	return !_.some(this._disabledReasons.getItemsAsObject(), function(item) {
		return item.hideAnnotations;
	});
};

/**
 * Temporarily show the annotations for the difference document, only if there are no reasons that don't allow showing annotations
 */
AnnotationModel.prototype.showTemporarilyDifferencePageAnnotations = function() {
	if (this.isAllowedToShowAnnotations()) {
		var differencePage = this.f.getDocumentController().getDifferencePage();
		this.unselectAllAnnotations();

		this._temporarilyDisabled.showingDifferenceAnnotations = true;
		this._temporarilyDisabled.showingCurrentAnnotations = false;
		this.updateTemporarilyDisabledReason();

		this.setPage(
			differencePage,
			_.bind(function() {
				this.updateTemporarilyDisabledReason();
				this._emit("temporarilyDisableChanged");
				this.modelChanged();
			}, this)
		);
	}
};

/**
 * Temporarily show the annotations for the active document
 */
AnnotationModel.prototype.showTemporarilyActivePageAnnotations = function() {
	if (this.isAllowedToShowAnnotations()) {
		var activePage = this.f.getActivePage();
		this.unselectAllAnnotations();
		this.setPage(
			activePage,
			_.bind(function() {
				this._temporarilyDisabled.showingDifferenceAnnotations = false;
				this._temporarilyDisabled.showingCurrentAnnotations = true;

				this.updateTemporarilyDisabledReason();
				this._emit("temporarilyDisableChanged");
				this.modelChanged();
			}, this)
		);
	}
};

/**
 * Hide the temporarily shown annotations
 */
AnnotationModel.prototype.hideTemporarilyAnnotations = function() {
	this._temporarilyDisabled.showingDifferenceAnnotations = false;
	this._temporarilyDisabled.showingCurrentAnnotations = false;
	this.setPage(this.f.getDocumentController().getActivePage());

	this.updateTemporarilyDisabledReason();
	this._emit("temporarilyDisableChanged");
	this.modelChanged();
};

/**
 * Get the page for which the annotations are visible
 * @returns {Data.AbstractDocumentPage}
 */
AnnotationModel.prototype.getAnnotationsPage = function() {
	return this._page;
};

/**
 * Can the current user change the annotations?
 * @param {Boolean} [dontConfirm] - if set to false, a warning message could be displayed
 * @returns {Boolean} true if the annotations are editable
 */
AnnotationModel.prototype.canChangeAnnotations = function(dontConfirm) {
	if (
		this.isTemporarilyDisabledAndCannotEdit() ||
		this._cannotChangeAnnotationsOfCurrentPage() ||
		this._disabledReasons.getLength() !== 0
	) {
		return false;
	} else if (!this._confirmedOnce && this.warnWhenChangingAnnotations.locked && !dontConfirm) {
		var msg = this.warnWhenChangingAnnotations.reasons.join("; ");
		this._confirmedOnce = window.confirm(msg);
		return this._confirmedOnce;
	} else {
		return true;
	}
};

/**
 * Can the given annotation be changed
 * @param {Annotation} annotation - The annotation to be checked
 * @returns {Boolean} True if the current user can change the given annotation, false otherwise
 */
AnnotationModel.prototype.canChangeAnnotation = function(annotation) {
	if (this.isTemporarilyDisabledAndCannotEdit() || this._cannotChangeAnnotationsOfCurrentPage()) {
		return false;
	}

	var access = AnnotationConstants.AccessRight.READ_ONLY;
	if (this._disabledReasons.getLength() === 0) {
		access = annotation.getPermission();
	}
	return access == AnnotationConstants.AccessRight.READ_WRITE;
};

/**
 * Can the annotations NOT be changed of the current page.
 * This is false if the page is not part of the latest document version or because of e.g. pharma settings
 */
AnnotationModel.prototype._cannotChangeAnnotationsOfCurrentPage = function() {
	return this._page && (!this._page.getDocumentVersion().isLatestVersion() || !this._page.canChangeAnnotations());
};

/**
 * Check if an annotation can be deleted
 * @param {Annotation} annotation - the annotation to be deleted
 * @returns {Boolean}
 */
AnnotationModel.prototype.canDeleteAnnotation = function(annotation) {
	if (this.f.getProperty("annotationblockdeletesaved") && !annotation.isNew) {
		return false;
	} else {
		return this.canChangeAnnotation(annotation);
	}
};

/**
 * Set the annotations objects for this model
 * @param {Array<Annotation>} annots - List of annotations
 */
AnnotationModel.prototype.setAnnotations = function(annots) {
	// Only load the allowed annotations if the config key 'annotationModule.filter' has an array of annotation names.
	var annotationFilter = this.f.getProperty("AnnotationModel.annotationFilter");
	if (annotationFilter && annots.length > 0) {
		if (annotationFilter.allowedAnnotations.length > 0) {
			annots = annots.filter(function(obj) {
				var matchingFilter = _.find(annotationFilter.allowedAnnotations, function(annotation) {
					return annotation.name === obj.name;
				});
				if (matchingFilter) {
					if (!matchingFilter.allowLinkedAnnotations) {
						obj.links = [];
						obj.linkedto = null;
						obj.parent = null;
					}
					return true;
				}
				return false;
			});
		}

		if (annotationFilter.resetFilter) {
			this.getAnnotationFilterModel().resetFilter();
		}
	}
	var annotList = this._filter.filterIncomingAnnotations(annots);
	this.scaleAllAnnotations(null);
	this._annotations = annotList;
	//TODO allow the reference page the compensated, the Java Viewer does not allow this, as it returns wrong parameter values
	if (this._page && this._page !== this.f.getDocumentController().getDifferencePage()) {
		this.scaleAllAnnotations(this._page.getCurrentScaling());
	}
	this._confirmedOnce = false;
	this.setSortingMode(this.f.getOption("PREFERENCES_ANNOTATION_DATESORT_MODE"), true);
	this._filterAllAnnotations(true);

	this.modelContentChanged();
};

/**
 * @returns {Array.<Annotation>} annotations of the model (unfiltered)
 */
AnnotationModel.prototype.getAnnotations = function() {
	return this._annotations;
};

/**
 * @returns {Array.<String>} origin types of the annotations
 */
AnnotationModel.prototype.getAnnotationOrigins = function() {
	return _.uniq(
		_.map(this._annotations, function(annotation) {
			return annotation.originType;
		})
	);
};

/**
 * Get the number of annotations (that are not linked)
 * @returns {Number}
 */
AnnotationModel.prototype.getAnnotationCount = function() {
	return _.filter(
		this._annotations,
		function(annotation) {
			return annotation.getGodfather() === annotation;
		},
		this
	).length;
};

/**
 * Filter all annotations to the active filter
 * @private
 * @param {Boolean} [silent] - if true, no events are send out
 */
AnnotationModel.prototype._filterAllAnnotations = function(silent) {
	var annotations = _.filter(
		this._annotations,
		function(annotation) {
			return annotation.getGodfather() === annotation;
		},
		this
	);
	this._filteredAnnotations = this._filter.filterAnnotations(annotations);
	this.currentOrdening();

	if (!silent) {
		this.modelChanged();
		this.filterChanged();
	}
};

/**
 * Get the list of users that have made an annotation + the current user
 * @param {String} [originType] - type of origin to get the users of
 * @returns {Object} where key is the user name and the value is the full user name
 * {
 * 	"RUCA" : "Cagnie, Ruben (RUCA)"
 *  "BAMT" : "Matte, Bart (BAMT)"
 * }
 */
AnnotationModel.prototype.getUsers = function(originType) {
	var users = {};
	_.each(this._annotations, function userIterator(annotation) {
		if (!users[annotation.username] && (!originType || annotation.originType === originType)) {
			users[annotation.username] = annotation.getUserName();
		}

		if (annotation.replies.length > 0) {
			_.each(annotation.replies, userIterator);
		}
	});

	if (users[this.f.getCurrentUser().getUsername()] === undefined) {
		users[this.f.getCurrentUser().getUsername()] = this.f.getCurrentUser().getFullName();
	}

	return users;
};

/**
 * Delete an annotation OBJECT (can be annotation/reply/review state)
 * @private
 * @param {Annotation} annotation - Annotation to delete
 */
AnnotationModel.prototype._deleteOneAnnotation = function(annotation) {
	this._annotations.splice(_.indexOf(this._annotations, annotation), 1);
	if (_.indexOf(this._filteredAnnotations, annotation) >= 0) {
		this._filteredAnnotations.splice(_.indexOf(this._filteredAnnotations, annotation), 1);
	}

	if (!annotation.isNew) {
		this._removedannotations.push(annotation);
	}
};

/**
 * Renumber all the not saved annotations so the numbers are continuous. This is called when a not saved annotation is removed
 * @param deletedNumber - the sequencenumber of the removed annotation
 */
AnnotationModel.prototype._renumberNewAnnotations = function(deletedNumber) {
	var unsavedAnnotations = this.getUnsavedAnnotations();
	_.each(
		unsavedAnnotations,
		_.bind(function(ann) {
			if (ann.messagenumber > deletedNumber) {
				ann.messagenumber = ann.messagenumber - 1;
				ann.updateIndicator();
			}
		}, this)
	);
	this._emit("modelContentChanged");
};

/**
 * Delete an annotation/reply with all its replies, review states and links. If the annotation is the main annotation, a message will be shown ("Do you wish to delete this annotation thread?")
 * If the annotations can't be changed, this function will not be executed
 * @param {Annotation} annotation - The annotation to delete
 */
AnnotationModel.prototype.deleteAnnotation = function(annotation) {
	if (!this.canDeleteAnnotation(annotation)) {
		this.f.i18n.loadBundle(
			"viewer",
			function() {
				this.f.emit("error.WARNING", this.f.i18n.t("AnnReadOnlyError"));
			},
			this
		);
		return;
	}

	this.f.i18n.loadBundle(
		"viewer",
		_.bind(function() {
			if (!annotation.inreplyto) {
				if (this._deleteAnnotationPopup === undefined) {
					this._deleteAnnotationPopup = new ModalPopup(this.f, "deleteAnnotationPopup", this.f.getContext(), undefined, undefined, {
						cancelButtonSahiID: "cancelDeleteAnnotation",
						okButtonSahiID: "confirmDeleteAnnotation"
					});

					// Header + Content
					this._deleteAnnotationPopup.setHeader(
						$("<span/>")
							.addClass("popupTitle")
							.text(this.f.i18n.t("DeleteAnnotBtnTT"))
					);
					this._deleteAnnotationPopup.setContent(
						$("<div/>")
							.addClass("unsavedItemsContent")
							.text(this.f.i18n.t("AnnDelete"))
					);
					this._deleteAnnotationPopup.setOkBtnText(this.f.i18n.t("YesResponse"));
					this._deleteAnnotationPopup.setCancelBtnText(this.f.i18n.t("NoResponse"));
				}
				this._deleteAnnotationPopup.setOkAction(
					_.bind(function() {
						this._deleteAnnotationPopup.hide();
						this._deleteAnnotationWithoutConfirmation(annotation);
					}, this)
				);

				this._deleteAnnotationPopup.show();
				this._deleteAnnotationPopup.recalculateSize(400);
			} else {
				this._deleteAnnotationWithoutConfirmation(annotation);
			}
		}, this)
	);
};

/**
 * Delete an annotation/reply with all its replies, review states and links.
 * @param {Annotation} annotation - The annotation to delete
 */
AnnotationModel.prototype._deleteAnnotationWithoutConfirmation = function(annotation) {
	if (annotation.parent) {
		var replyIndex = _.indexOf(annotation.parent.replies, annotation);
		if (replyIndex > -1) {
			annotation.parent.replies.splice(replyIndex, 1);
		} else {
			var stateIndex = _.indexOf(annotation.parent.states, annotation);
			if (stateIndex > -1) {
				annotation.parent.states.splice(stateIndex, 1);
			} else {
				var linkIndex = _.indexOf(annotation.parent.links, annotation);
				if (linkIndex > -1) {
					this._deleteAnnotationWithoutConfirmation(annotation.parent);
				}
			}
		}

		if (!annotation.isNew) {
			this._removedannotations.push(annotation);
		}
	} else {
		this.unselectAllAnnotations(true);
		_.each(
			annotation.links,
			function(annot) {
				this._deleteOneAnnotation(annot);
			},
			this
		);
		this._deleteOneAnnotation(annotation);
		this._filterAllAnnotations(true);
	}

	this._emit("annotationDeleted", annotation);

	if (annotation.isNew) {
		this._renumberNewAnnotations(annotation.messagenumber);
	}
	this.modelChanged();
};

/**
 * Reload the list of annotations.
 * If the annotations have been changed since, a messages will be shown if the user wants to save the annotations first ("Annotations have been changed. Do you wish to save these changes?")
 * @param {Boolean} silent - if silent is true, the user won't be prompted to save unsaved annotations
 */
AnnotationModel.prototype.reload = function(silent) {
	var reloadFunction = _.bind(function() {
		this._emit("reload.started");
		this.unselectAllAnnotations();
		this.resetAllAnnotations();
		if (this._page) {
			this.modelChanged();
			this._page.reloadAnnotations(
				_.bind(function(success) {
					if (success) {
						this.setAnnotations(this._page.getAnnotations());
					}
					this.modelChanged();
					this._emit("reload.finished");
				}, this)
			);
		} else {
			this._emit("reload.finished");
		}
	}, this);

	if (silent) {
		// just throw away changes
		reloadFunction();
	} else {
		// prompt
		var hasUnsavedAnnotations = this.promptToSaveUnsavedAnnotations(reloadFunction, function() {}, reloadFunction);
		if (!hasUnsavedAnnotations) {
			//no prompt for unsaved annotations is shown
			reloadFunction(); //reload the annotations
		}
	}
};

/**
 * Save 1 or all annotations
 * @param {Annotation} [annotation] if specified, only this annotation will be saved. If not, all annotations will be saved
 * @param {Function} [callbackFunction] will be called when the annotations have been saved
 */
AnnotationModel.prototype.save = function(annotation, callbackFunction) {
	if (!this._isSavingAnnotations) {
		this._emit("save.started");
		this._isSavingAnnotations = true;
		if (this.canChangeAnnotations() && this._page) {
			this.unselectAllAnnotations();

			// check if there are unsaved replies, if so, add them in the model
			_.each(
				this._annotations,
				function(annotation) {
					if (annotation.hasUnsavedReply()) {
						var text = annotation.getUnsavedReply();
						annotation.resetUnsavedReply();
						this.replyToAnnotation(annotation, text, false);
					}
				},
				this
			);

			var annotationsToBeSaved;
			var annotationToBeRemoved = this._removedannotations.slice(0);
			this._removedannotations = [];

			if (!annotation) {
				annotationsToBeSaved = this.getUnsavedAnnotations();
			} else {
				annotationToBeRemoved = [];
				annotationsToBeSaved = [];

				annotationsToBeSaved.push(annotation);

				var p = annotation.getGodfather();
				_.each(p.links, function(annot) {
					if (!annot.saved && _.indexOf(annotationsToBeSaved, annot) == -1) {
						annotationsToBeSaved.push(annot);
					}
				});
			}

			this._page.saveAnnotations(
				annotationsToBeSaved,
				annotationToBeRemoved,
				_.bind(
					function(page, annots) {
						this._isSavingAnnotations = false;
						if (this._page == page) {
							// we only update the annotation, if the page that was saved is still active, otherwise don't do anything
							this.setAnnotations(annots);
						}

						if (callbackFunction) {
							callbackFunction();
						}
						this._emit("save.finished");

						this.modelChanged();
					},
					this,
					this._page
				)
			);
		} else {
			this._isSavingAnnotations = false;
			// we didn't need to save, notify that the safe action is finsihed
			this._emit("save.finished");
		}
	} else {
		this.f.emit("error.WARNING", this.f.i18n.t("AnnAreAlreadyBeingSaved"), true);
	}
};

/**
 *
 * @returns {Number} the count of all changed annotations (deleted + unsaved)
 */
AnnotationModel.prototype.getUnsavedAnnotationCount = function() {
	var unsavedCount = this.getUnsavedAnnotations().length;
	var deletedCount = this.getDeletedAnnotations().length;
	return unsavedCount + deletedCount;
};

/**
 * Get an array of unsaved annotations.
 * @returns {Array.<Annotation>} Array of unsaved annotations
 */
AnnotationModel.prototype.getUnsavedAnnotations = function() {
	var unsavedAnnotations = [];
	if (this.canChangeAnnotations(true)) {
		_.each(
			this._annotations,
			_.bind(function(annotation) {
				annotation.runThroughAnnotation(
					true,
					true,
					true,
					_.bind(function(annot) {
						if (
							(!annot.saved || annot.hasUnsavedReply()) &&
							_.indexOf(unsavedAnnotations, annot) == -1 &&
							this.canChangeAnnotation(annot)
						) {
							unsavedAnnotations.push(annot);
						}
					}, this)
				);

				var p = annotation.getGodfather();
				_.each(
					p.links,
					_.bind(function(annot) {
						if (
							(!annot.saved || annot.hasUnsavedReply()) &&
							_.indexOf(unsavedAnnotations, annot) == -1 &&
							this.canChangeAnnotation(annot) &&
							!annot.isStateObject()
						) {
							unsavedAnnotations.push(annot);
						}
					}, this)
				);
			}, this)
		);
	}

	return unsavedAnnotations;
};

/**
 * Get an array of saved annotations.
 * @returns {Array.<Annotation>} - Array of saved annotations
 */
AnnotationModel.prototype.getSavedAnnotations = function() {
	var savedAnnotations = [];
	_.each(this._annotations, function(annotation) {
		if (annotation.saved) {
			savedAnnotations.push(annotation);
		}
	});
	return savedAnnotations;
};

/**
 * Get an array of removed annotations
 * @returns {Array.<Annotation>} Array of deleted annotations
 */
AnnotationModel.prototype.getDeletedAnnotations = function() {
	return this._removedannotations;
};

/**
 * @param {Boolean} ignoreTimestamp - True to disable the timestamp visibility check
 * @returns {Array.<Annotation>} the visible annotations to be displayed on the viewport
 */
AnnotationModel.prototype.getVisibleAnnotations = function(ignoreTimestamp) {
	var visible;
	// if annotation tools are disabled, or the module is temporarily disabled, do not show any annotations
	if (this.f.getOption("PREFERENCES_ANNOTATIONTOOLS_VISIBLE") === 0 || this.isTemporarilyDisabledAndNoAnnotationsVisible()) {
		visible = [];
	} else {
		if (this.f.getOption("PREFERENCES_ANNOTATION_SHOWSELECTEDONLY") && this._selectedAnnotations.length > 0) {
			visible = this._selectedAnnotations;
		} else {
			visible = this._filteredAnnotations;
		}
	}

	// Filter out annotations outside of the current timestamp
	if (!ignoreTimestamp) {
		var timestamp = this.getTimeStamp();
		if (timestamp != null) {
			visible = _.filter(
				visible,
				_.bind(function(annot) {
					if (annot.startTime == null || annot.endTime == null) {
						return true;
					} else {
						return timestamp >= parseFloat(annot.startTime, 10) - 0.25 && timestamp <= parseFloat(annot.endTime, 10) + 0.25;
					}
				}, this)
			);
		}
	}

	// Filter annotations by sending an event, other places can listen to this event and prevent an annotation from being shown
	visible = _.filter(
		visible,
		function(annot) {
			var hidden = _.some(
				this._interceptAnnotationVisibilityCallbacks,
				function(cb) {
					return !cb(annot);
				}.bind(this)
			);
			return !hidden;
		}.bind(this)
	);

	return visible;
};

/**
 * Get the filtered annotations
 * @returns {Array.<Annotation>} Array of filtered annotations
 */
AnnotationModel.prototype.getFilteredAnnotations = function() {
	return this._filteredAnnotations;
};

/**
 * Select an annotation. All other selected annotations will be unselected first
 * @param {Annotation} annot - Annotation to be selected
 * @param {Boolean} force - makes the unselectAllAnnotations forced
 * @param {Boolean} isNewAnnotation - Indicates if the function is called because a new annotation has been added
 */
AnnotationModel.prototype.selectAnnotation = function(annot, force, isNewAnnotation) {
	var i;
	if (this._selectedAnnotations.length > 0) {
		var sa = this._selectedAnnotations[0];
		if (annot === sa) {
			return;
		}
	}

	this.unselectAllAnnotations(true, force);

	this._selectedAnnotations.push(annot);
	var mainAnnotation = annot.getGodfather();
	if (mainAnnotation.links) {
		for (i = 0; i < mainAnnotation.links.length; i++) {
			this._selectedAnnotations.push(mainAnnotation.links[i]);
		}
	}
	if (annot.linkedto) {
		annot = this.getAnnotation(annot.linkedto);
		if (annot) {
			for (i = 0; i < annot.links.length; i++) {
				if (annot.links[i] != annot) {
					this._selectedAnnotations.push(annot.links[i]);
				}
			}
			this._selectedAnnotations.push(annot);
		}
	}

	this.ensureVisibility();

	this._emit("selectedAnnotation", annot, isNewAnnotation);
	this.modelChanged();
};

/**
 * Make sure that the annotations are visible. If not, set the visible property back to 1
 */
AnnotationModel.prototype.ensureVisibility = function() {
	if (!this.f.getOption("PREFERENCES_ANNOTATIONTOOLS_VISIBLE")) {
		this.f.setOption("PREFERENCES_ANNOTATIONTOOLS_VISIBLE", 1);
	}
};

/**
 * Get an annotation based on the annotation name
 * @param {String} annotationName - internal name of the annotation e.g. name="kkIy7drJAKxuER0gspHfIS"
 * @returns {Annotation} the annotation for a given name, null if no annotation was found
 */
AnnotationModel.prototype.getAnnotation = function(annotationName) {
	for (var i = 0; i < this._annotations.length; i++) {
		if (this._annotations[i].name == annotationName) {
			return this._annotations[i];
		}
	}
	return null;
};

/**
 * Get the all the links of an annotation (also of the parents and the links of links)
 * @param {Annotation} annot - reference annotation
 * @returns {Array.<Annotation>}
 */
AnnotationModel.prototype.getAnnotationLinks = function(annot) {
	var i;
	var annots = [];
	if (annot.links) {
		for (i = 0; i < annot.links.length; i++) {
			annots.push(annot.links[i]);
		}
	}
	if (annot.linkedto) {
		for (i = 0; i < this._annotations.length; i++) {
			if (this._annotations[i].name == annot.linkedto) {
				annots.push(this._annotations[i]);
			}
		}
	}
	return annots;
};

/**
 * @param {Annotation} annotation
 * @returns {Array.<Annotation>} all the replies to one annotation
 */
AnnotationModel.prototype.getReplies = function(annotation) {
	var replies = [];
	annotation.runThroughAnnotation(false, true, false, function(annot) {
		replies.push(annot);
	});
	return replies;
};

/**
 * Unselect all the annotations
 * @param {Boolean} [donotbroadcast] - if set, the model will not broadcast that this happened (only used when the caller will call modelChanged() itself)
 * @param {Boolean} [force] - if set, it will ignore the lockSelection flag
 */
AnnotationModel.prototype.unselectAllAnnotations = function(donotbroadcast, force) {
	if (this._selectedAnnotations.length <= 0 || (this.lockSelection && !force)) {
		return;
	}

	var selectedAnnotations = this._selectedAnnotations;
	this._selectedAnnotations = [];

	_.each(
		selectedAnnotations,
		_.bind(function(annotation) {
			this._emit("unselectedAnnotation", annotation);
		}, this)
	);

	if (!donotbroadcast) {
		this.modelChanged();
	}
};

/**
 * Resets all the annotations in the model, all changes to annotations are lost.
 */
AnnotationModel.prototype.resetAllAnnotations = function() {
	this.scaleAllAnnotations(null);
	this.unselectAllAnnotations(); //first unselect all annotations
	this._annotations = []; //set annonations of to 0
	this._removedannotations = []; //set no removed annotation
	this._page.resetAnnotations();
};

/**
 * Check if an annotation or one of the replies of the annotations is selected.
 * @param {Annotation} annot - Annotation to check
 * @returns {Boolean} True if the annotation or one of its replies are selected, false otherwise
 */
AnnotationModel.prototype.isAnnotationOrReplySelected = function(annot) {
	if (annot !== null) {
		//first check if annotation is selected
		if (_.indexOf(this._selectedAnnotations, annot) > -1) {
			return true;
		}
		//if annotation is not selected first check if it has replies
		if (annot.replies.length === 0) {
			return false;
		} else {
			//returns true the first time a reply of annot is selected
			return _.some(
				annot.replies,
				_.bind(function(reply) {
					return _.indexOf(this._selectedAnnotations, reply) > -1;
				}, this)
			);
		}
	} else {
		return false;
	}
};

/**
 * Get the selected annotations
 * @returns {Array.<Annotation>} list of selected annotations
 */
AnnotationModel.prototype.getSelectedAnnotations = function() {
	return this._selectedAnnotations;
};

/**
 * Move an annotation
 * @param {Annotation} annot- Annotation to move
 * @param {ShapeUtils.Size} movesize - World size of how much the annotation should move
 * @returns {Boolean} True if the annotation was moved, false otherwise (e.g. when you cannot edit the annotation)
 */
AnnotationModel.prototype.moveAnnotation = function(annot, movesize) {
	if (!this.canChangeAnnotation(annot)) {
		return false;
	}

	annot.move(movesize);
	this.modelChanged();

	return true;
};

/**
 * Transform the frame of an annotation
 * @param {Annotation} annot - Annotation to transform
 * @param {ShapeUtils.Rect} rect - New world rect for the annotation
 * @returns {Boolean} True if the annotation was transformed, false otherwise (e.g. when you cannot edit the annotation)
 */
AnnotationModel.prototype.transformAnnotation = function(annot, rect) {
	if (!this.canChangeAnnotation(annot)) {
		return false;
	}

	annot.setRect(rect);
	this.modelChanged(annot);

	return true;
};

/**
 * Change the start and end point of an annotation
 * @param {Annotation} annot - annotation to change
 * @param {ShapeUtils.Point} start - start point
 * @param {ShapeUtils.Point} end - end point
 */
AnnotationModel.prototype.updateStartEndOfAnnotation = function(annot, start, end) {
	if (!this.canChangeAnnotation(annot)) {
		return false;
	}

	annot.setStartEnd(start, end);
	this.modelChanged();

	return true;
};

/**
 * Add an annotation with a given rect
 * e.g. used to create an annotation from the measure module or to create a global annotation
 * @param {ShapeUtils.Rect}  rect - the new rect
 * @param {String} [content] - Content for the annotation
 * @param {AnnotationConstants.AnnotationType} [type] - type of the annotation to be added
 */
AnnotationModel.prototype.addAnnotationWithRect = function(rect, content, type) {
	var annotationColor;
	if (this.f.getProperty("annotationcolor")) {
		annotationColor = this.f.getProperty("annotationcolor");
	} else {
		annotationColor = this.f.getCurrentUser().getColor("PREFERENCES_ANNOTATION_RECTANGLE_COLOR");
	}

	var annot = new AnnotationFactory(this.f).create();
	annot.initAnnotation(type, annotationColor, false, this, rect);
	if (content) {
		annot.content = content;
	}

	this.addAnnotation(annot);
};

/**
 * Get the next message number for a new annotation
 * @private
 * @returns {Number} <the highest messagenumber> + 1
 */
AnnotationModel.prototype._nextNumberForOriginType = function(originType) {
	var number = 0;
	for (var i = 0; i < this._annotations.length; i++) {
		if (this._annotations[i].messagenumber > number && this._annotations[i].originType === originType) {
			number = this._annotations[i].messagenumber;
		}
	}
	number++;
	return number;
};

/**
 * Add an annotation. If the annotation is not a main annotation (e.g. reply/review state) the main annotation should be set
 * @param {Annotation} annot - the annotation to add
 * @param {Annotation} [mainannotation] - the main annotation (the godfather annotation when adding a reply or review state)
 * @param {Boolean} [silent = false] - don't send events
 */
AnnotationModel.prototype.addAnnotation = function(annot, mainannotation, silent) {
	if (this.checkCanChangeAnnotationsAndDisplayError()) {
		if (!annot.linkedto) {
			annot.messagenumber = this._nextNumberForOriginType(annot.originType);
		}
		annot.username = this.f.getCurrentUser().getUsername();
		annot.firstname = this.f.getCurrentUser().getFirstName();
		annot.lastname = this.f.getCurrentUser().getLastName();
		annot.creationdate = new Date();
		annot.date = new Date();

		annot.setModified("NEW");

		this.ensureVisibility();

		if (!mainannotation) {
			this._annotations.push(annot);
			this._filterAllAnnotations(true);
			var compliesWithFilter = this._filter.compliesToFilter(annot);
			var compliesWithCustomFilter = true;
			if (this._annotationFilterModel.isCustomFilterActive()) {
				compliesWithCustomFilter = this._filter.compliesToCustomFilter(annot);
			}
			if (compliesWithFilter && compliesWithCustomFilter) {
				if (!annot.linkedto && !silent) {
					this._emit("annotationAdded", annot, false);
				}
				this._filteredAnnotations.push(annot);
				if (!silent) {
					this.selectAnnotation(annot, undefined, true);
				}
			} else {
				/**
				 * The current annotation does not match with the current filter... Inform the user, so they know...
				 */
				if (!silent) {
					this._emit("annotationAdded", annot, true);
				}
				this.f.i18n.loadBundle(
					"viewer",
					function() {
						this.f.emit(
							"userMessage",
							"warn",
							this.f.i18n.t("AnnNotViewable"),
							_.bind(function() {
								this._annotationFilterModel.resetFilter();
							}, this)
						);
					},
					this
				);
			}
		} else {
			if (!silent) {
				this.modelChanged(mainannotation, annot);
			}
		}
	}
};

/**
 * Add a linked annotation to the currently selected annotation
 * @param {Annotation} annotation - Annotation to be added
 */
AnnotationModel.prototype.addLinkedAnnotation = function(annotation, mainAnnotation) {
	var annot = mainAnnotation || this.getSelectedAnnotations()[0];

	if (annot) {
		if (!this.canChangeAnnotation(annot)) {
			// we cannot add a linked annotation, just add a new annotation
			this.f.i18n.loadBundle(
				"viewer",
				function() {
					this.f.emit("userMessage", "warn", this.f.i18n.t("AnnCannotAddLinkedToNonOwnedAnnotation"), true);
				},
				this
			);
		} else {
			if (this.checkCanChangeAnnotationsAndDisplayError()) {
				annot = annot.getGodfather();

				if (annot.linkedto) {
					annot = this.getAnnotation(annot.linkedto);
				}

				annotation.linkedto = annot.name;
				annotation.parent = annot;

				annot.links.push(annotation);

				annotation.messagenumber = annot.messagenumber;
				this.addAnnotation(annotation);

				annotation.annotationcolor = annot.annotationcolor;
				annotation.color = annot.color;

				this.selectAnnotation(annot, true);
			}
		}
	} else {
		this.f.i18n.loadBundle(
			"viewer",
			function() {
				this.f.emit("userMessage", "warn", this.f.i18n.t("NoAnnotationSelected"), true);
			},
			this
		);
	}
};

/**
 * Reply to an annotation
 * @param {Annotation} annotation - Annotation to reply to
 * @param {String} replyText - reply text
 * @param {Boolean} save - If true, save the annotation directly
 */
AnnotationModel.prototype.replyToAnnotation = function(annotation, replyText, save) {
	if (this.checkCanChangeAnnotationsAndDisplayError()) {
		annotation.resetUnsavedReply();
		var annot = annotation.duplicate(false, false, false, false, this);
		annot.inreplyto = annotation.name;
		annot.content = replyText;
		annot.parent = annotation;

		annotation.replies.push(annot);

		this.addAnnotation(annot, annotation);

		if (save) {
			this.save(
				null,
				_.bind(function() {
					this.selectAnnotation(annotation, true);
					annot.cancelModified();
					annot.setIsNew(false);
				}, this)
			);
		}
	} else {
		annotation.resetUnsavedReply();
		this.modelChanged();
	}
};

/**
 * Set the state of an annotation (for the current user)
 * If the state is already set, the state is changed
 * If the state has not been set before, make a new review state object
 *
 * @param {Annotation} annotation - Annotation for which the state needs to be set
 * @param {String} state - new state
 */
AnnotationModel.prototype.setState = function(annotation, state) {
	if (this.checkCanChangeAnnotationsAndDisplayError()) {
		this.f.i18n.loadBundle(
			"viewer",
			_.bind(function() {
				var text = this.f.i18n.t("AnnStatusSetByText", state, this.f.getCurrentUser().getFullName());
				var bodytext = "<p>" + text + "</p>";

				var stateAnnotation = annotation.getUserStateAnnotation(this.f.getCurrentUser().getUsername());
				if (stateAnnotation) {
					if (stateAnnotation.state != state) {
						stateAnnotation.state = state;
						stateAnnotation.content = bodytext;
						stateAnnotation.setModified("STATE");
						this.modelChanged(annotation);
					}
				} else {
					var annot = annotation.duplicate(false, false, false, false, this);

					annot.inreplyto = annotation.name;
					annot.content = bodytext;
					annot.parent = annotation;
					annot.state = state;
					annot.statemodel = "Review";
					annot.parent = annotation;
					annot.linkedto = annotation.name;
					annotation.states.push(annot);

					this.addAnnotation(annot, annotation);
				}
			}, this)
		);
	}
};

/**
 * Check if the user is allowed to change annotations, display an error if needed
 * @returns {Boolean} canChangeAnnotations
 */
AnnotationModel.prototype.checkCanChangeAnnotationsAndDisplayError = function() {
	if (!this.canChangeAnnotations()) {
		if (this.getDisabledReason().length > 0) {
			// only show a warning message, if we have something to say...
			this.f.emit("userMessage", "warn", this.getDisabledReason());
		}

		return false;
	} else {
		return true;
	}
};

/**
 * Inform the listeners that the model has changed.
 * If a mainannotation or annotation is provided, the function will also emit an "annotationUpdated" message
 * @param {Annotation} [mainannotation] - the "main" annotation that has been updated
 * @param {Annotation} [annotation] - the "reference" annotation (reply/review state/link) that has been updated
 */
AnnotationModel.prototype.modelChanged = function(mainannotation, annotation) {
	if (annotation || mainannotation) {
		this._emit("annotationUpdated", mainannotation, annotation);
	}
	this._emit("modelChanged");
};

/**
 * Inform the listeners that the complete content has changed for the model (i.o.w, the _annotations are changed)
 */
AnnotationModel.prototype.modelContentChanged = function() {
	this._orderedAnnotations.length = 0;
	this._emit("modelContentChanged");
};

/**
 * Inform that the filter has changed (so the GUI can be updated)
 */
AnnotationModel.prototype.filterChanged = function() {
	this._emit("filterChanged");
};

/**
 * Orders the annotations by modification date. Only main annotation get ordered but the last modified date of all the
 * replies and states are taken into account.
 * @private
 */
AnnotationModel.prototype.orderByModificationDate = function() {
	this._filteredAnnotations = _.sortBy(this._filteredAnnotations, function(ann) {
		var lastModifiedDate = new Date(0);
		ann.runThroughAnnotation(true, true, true, function(annotation) {
			if (annotation.date > lastModifiedDate) {
				lastModifiedDate = annotation.date;
			}
		});
		return lastModifiedDate;
	});
	this.currentOrdening = this.orderByModificationDate;
};

/**
 * Orders the annotations by creation date.
 * @private
 */
AnnotationModel.prototype.orderByCreationDate = function() {
	this._filteredAnnotations = _.sortBy(this._filteredAnnotations, function(ann) {
		return ann.creationdate;
	});
	this.currentOrdening = this.orderByCreationDate;
};

/**
 * Calculates the _orderedAnnotations array according to the step by option and sets the index to 0.
 */
AnnotationModel.prototype.orderAnnotationsForStep = function() {
	var stepByCreationDate =
		this.f.getOption("PREFERENCES_ANNOTATION_DATESORT_MODE") === AnnotationConstants.AnnotationSortType.SORT_BY_CREATION_DATE;

	var annotationList = [];

	//when step by creation date is selected (also replies)
	if (this.f.getOption("PREFERENCES_ANNOTATION_NEXT_MODE") === 0) {
		_.each(this._filteredAnnotations, function(ann) {
			//go through all replies of the annotation and add them to the temp collection
			ann.runThroughAnnotation(true, true, false, function(annotation) {
				annotationList.push(annotation);
			});
		});
	} else if (this.f.getOption("PREFERENCES_ANNOTATION_NEXT_MODE") === 3) {
		// step by area/position, each part of a linked annotation needs to be added
		_.each(
			this._filteredAnnotations,
			function(annotation) {
				annotationList.push(annotation);
				annotationList = annotationList.concat(this.getAnnotationLinks(annotation));
			},
			this
		);
	} else {
		annotationList = this._filteredAnnotations;
	}

	//Order the temp collection
	this._orderedAnnotations = _.sortBy(annotationList, function(ann) {
		if (stepByCreationDate) {
			return -ann.creationdate;
		} else {
			return -ann.date;
		}
	});

	this._currentAnnotationIndex = _.indexOf(this._orderedAnnotations, this._selectedAnnotations[0]);
};

/**
 *	Gets the next annotation according to Step by setting. When the orderedAnnotations collection is empty it
 *	calculates the new collection by calling orderAnnotationsForStep.
 *	@private
 *
 *	@returns {Annotation} the next annotation
 */
AnnotationModel.prototype.getNextAnnotation = function() {
	if (this._orderedAnnotations.length === 0) {
		this.orderAnnotationsForStep();
	}
	if (this._selectedAnnotations.length > 0) {
		this._currentAnnotationIndex = _.indexOf(this._orderedAnnotations, this._selectedAnnotations[0]);
	}
	return this._orderedAnnotations[++this._currentAnnotationIndex % this._orderedAnnotations.length];
};

/**
 *	Gets the previous annotation according to Step by setting. When the orderedAnnotations collection is empty it
 *	calculates the new collection by calling orderAnnotationsForStep.
 *	@private
 *
 *	@returns {Annotation} the previous annotation
 */
AnnotationModel.prototype.getPreviousAnnotation = function() {
	if (this._orderedAnnotations.length === 0) {
		this.orderAnnotationsForStep();
	}
	if (this._selectedAnnotations.length > 0) {
		this._currentAnnotationIndex = _.indexOf(this._orderedAnnotations, this._selectedAnnotations[0]);
	}

	if (--this._currentAnnotationIndex < 0) {
		this._currentAnnotationIndex = this._orderedAnnotations.length - 1;
	}

	return this._orderedAnnotations[this._currentAnnotationIndex % this._orderedAnnotations.length];
};

/**
 * Selects the next annotation.
 *
 * @param {Boolean} isNext - should it go to the next annotation, if false it goes to the previous annotation
 */
AnnotationModel.prototype.goToAnnotation = function(isNext) {
	var nextAnnotation = isNext ? this.getNextAnnotation() : this.getPreviousAnnotation();
	this.selectAnnotation(nextAnnotation, true);
	var annotationNextMode = this.f.getOption("PREFERENCES_ANNOTATION_NEXT_MODE");
	//when zoom to annotation is needed
	if (annotationNextMode === 2 || annotationNextMode === 3) {
		this.f.emit("module.annotation.zoomToAnnotation", nextAnnotation);
	}
};

/**
 * Update the annotation color of the current user to use the colors defined in his settings
 */
AnnotationModel.prototype.updateAllAnnotationColors = function() {
	var pinColor = this.f.getCurrentUser().getColor("PREFERENCES_ANNOTATION_NOTE_COLOR");
	var otherColor = this.f.getCurrentUser().getColor("PREFERENCES_ANNOTATION_RECTANGLE_COLOR");

	_.each(
		this._annotations,
		_.bind(function(annotation) {
			if (this.canChangeAnnotation(annotation)) {
				if (annotation.annottype == AnnotationConstants.AnnotationType.ANNOT_PIN) {
					annotation.setColor(pinColor);
				} else {
					annotation.setColor(otherColor);
				}
				this._emit("annotationUpdated", annotation);
			}
		}, this)
	);

	this.modelChanged();
};

/**
 * Set the sorting type of the annotations
 * @param {Number} mode
 * @param {Boolean} [silent] - if true, no events are sent out
 */
AnnotationModel.prototype.setSortingMode = function(mode, silent) {
	if (mode === AnnotationConstants.AnnotationSortType.SORT_BY_CREATION_DATE) {
		this.orderByCreationDate();
	} else if (mode === AnnotationConstants.AnnotationSortType.SORT_BY_MODIFICATION_DATE) {
		this.orderByModificationDate();
	}

	if (!silent) {
		this.filterChanged();
	}
};

/**
 * Generate a unique name (looking to all existing annotations)
 * This method will try up to 100 times to generate an unique name
 * @returns {String} Unique random name for an annotation containing 22 characters from [0-9a-zA-Z]
 */
AnnotationModel.prototype.generateUniqueName = function() {
	var trialsLeft = 100;
	var finalName;

	while (finalName === undefined && trialsLeft > 0) {
		var candidate = ViewerUtils.randomString(22);

		var match = _.find(this._annotations, function(annotation) {
			return annotation.name === candidate;
		});

		if (match === undefined) {
			finalName = candidate;
		}

		trialsLeft--;
	}

	if (finalName === undefined) {
		this.f.emit("error.SEVERE", "Error while generating ID for new annotation");
	}

	return finalName;
};

/**
 * Set the page for which annotations should be displayed (this method will fetch the annotation for the page if necessary)
 * @param {Data.AbstractDocumentPage} newPage - page for whichannotatiosn should be displayed
 * @param {Function} [callback] - called when the page has changed
 */
AnnotationModel.prototype.setPage = function(newPage, callback) {
	if (newPage != this._page) {
		// the page has changed, so we need to do something

		// reset annotations
		this.setAnnotations([]);
		this._emit("changed.page");
		this.unselectAllAnnotations(false, true);
		this._page = newPage;
		if (this._addAnnotationsEnabledListener) {
			this._unregisterListener(this._addAnnotationsEnabledListener);
		}
		if (this._disabledReasonChangedListener) {
			this._unregisterListener(this._disabledReasonChangedListener);
		}

		this._clearDisableReasons();

		if (newPage) {
			// load the annotations for the newPage
			newPage.whenParametersAvailable(function() {
				newPage.requestAnnotations(
					_.bind(function(success) {
						this._page = newPage;
						this.f.loadBundle(
							"viewer",
							function() {
								if (!newPage.canChangeAnnotations()) {
									this.setDisableReason("CANCHANGEANNOTATIONS", this.f.t("AnnReadOnlyCannotChangeAnnotations"));
								} else {
									this.removeDisableReason("CANCHANGEANNOTATIONS");
								}

								if (!newPage.getDocumentVersion().isLatestVersion()) {
									this.setDisableReason("NOT_LATEST", this.f.t("AnnReadOnlyNotLatestVersion"));
								} else {
									this.removeDisableReason("NOT_LATEST");
								}
							},
							this
						);

						if (success) {
							this.setAnnotations(newPage.getAnnotations());

							this._checkForZoomToAnnotation();

							if (this.getAnnotationCount() === 0 && this.f.getConfig().get("annotationaddglobal") === "1") {
								//if annotations are not completly done loading (either still disabled reason or user cannot edit yet)
								if (this.getDisabledReason().length > 0 || !this.canChangeAnnotations()) {
									//listen to both change is disabled reasons (either during loading or when opened in compare) and edit rights to annotations (load time)
									this._addAnnotationsEnabledListener = this._registerListener(
										"readWriteChanged",
										function() {
											if (this.getDisabledReason().length === 0 && this.canChangeAnnotations() && this.getAnnotationCount() === 0) {
												this._emit("addGlobalAnnotation");
												this._unregisterListener(this._addAnnotationsEnabledListener);
												this._unregisterListener(this._disabledReasonChangedListener);
											}
										},
										this,
										this
									);
									this._disabledReasonChangedListener = this._registerListener(
										"disabledReasonChanged",
										function() {
											if (this.getDisabledReason().length === 0 && this.canChangeAnnotations() && this.getAnnotationCount() === 0) {
												this._emit("addGlobalAnnotation");
												this._unregisterListener(this._addAnnotationsEnabledListener);
												this._unregisterListener(this._disabledReasonChangedListener);
											}
										},
										this,
										this
									);
								} else {
									this._emit("addGlobalAnnotation");
								}
							}
						} else {
							this.f.debug.error("Annotations were NOT fetched succesfully");
						}

						this.updateTemporarilyDisabledReason();

						if (callback) {
							callback();
						}
					}, this)
				);
			}, this);
			this.removeDisableReason("LOADING_MODULES");
		}

		this._emit("setdisable", newPage);
	} else {
		if (callback !== undefined) {
			callback();
		}
	}

	if (!this._annotationFilterModel.isAnnotationFilterPinned()) {
		this._annotationFilterModel.resetFilter();
	}
	this.modelChanged();
};

/**
 * Get the page for which annotations are shown currently
 * @returns {Data.AbstractDocumentPage}
 */
AnnotationModel.prototype.getPage = function() {
	return this._page;
};

/**
 * Returns the offset of the page
 * @returns {ShapeUtils.Size}
 */
AnnotationModel.prototype.getAnnotationsOffset = function() {
	var annotPageOffset = this._page.getAnnotationsOffset();
	return annotPageOffset;
};

/**
 * Get the value of the warning message
 * @returns {String}
 */
AnnotationModel.prototype.getDisabledReason = function() {
	if (this._disabledReasons.getLength() > 0) {
		var reason = this._disabledReasons.getLast().reason;
		if (reason === "") {
			_.some(this._disabledReasons.getItemsAsObject(), function(item) {
				reason = item.reason;
				return reason !== "";
			});
		}
		return reason;
	} else {
		return "";
	}
};

/**
 * Add a new reason why the annotationmodule is blocked
 * @param {String} id
 * @param {String} reason - message shown to the user
 * @param {Boolean} [hideAnnotations] - do the annotations need to be completly hidden?
 * @param {Boolean} [dontRemove] - don't remove the disabled reason from the collection when the page changes, the module who added the reason is responsible of removing it again
 */
AnnotationModel.prototype.setDisableReason = function(id, reason, hideAnnotations, dontRemove) {
	var disabledReason = {
		reason: reason,
		hideAnnotations: hideAnnotations,
		dontRemove: dontRemove
	};
	this._disabledReasons.addOrReplaceItem(id, disabledReason);
	this._disabledReasons.moveItemByKeyToEnd(id);
	this._emit("disabledReasonChanged");
};

/**
 * Remove a reason why the annotaitonmodule was blocked
 * @param id
 */
AnnotationModel.prototype.removeDisableReason = function(id) {
	this._disabledReasons.removeByKey(id);
	this._emit("disabledReasonChanged");
};

/**
 * Clear out all disabled reasons, with exeption of the reasons set by other modules, they are responsible to clear those reasons
 */
AnnotationModel.prototype._clearDisableReasons = function() {
	var keepReasons = []; //construct an array with all the keys of reason that don't need to be cleared out of the reasons object
	_.each(
		this._disabledReasons.getKeys(),
		function(key) {
			//run over all the reasons in the disabledReasons collection
			if (this._disabledReasons.getItemByKey(key).dontRemove) {
				//if the don't remove flag is set, push the key in the array
				keepReasons.push(key);
			}
		},
		this
	);
	this._disabledReasons.omitOthers(keepReasons); //omit all the items from the ordered object but the reasons that need to be kept
	this._emit("disabledReasonChanged");
};

/**
 * Prompt to save unsaved annotations if there are any
 * if there are unsaved anntations, save them
 *
 * NOTE: this is not a pure model function, but to limit regressions, we keep it this way...
 * @param {Function} callbackOnOK - function which is called when the popup is closed and annotations are saved
 * @param {Function} [callbackOnCancel] - if provided, this callback is called when the user presses cancel
 * @param {Function} [callbackOnDontSave] - adds the possibility to not save unsaved annotation but to continue with the action, on the contrary, cancel will cancel the action and don't save. Note that reloading the annotations should be done by the caller!
 * @returns {Boolean} true if there are unsaved items and the popup is shown, false if not
 */
AnnotationModel.prototype.promptToSaveUnsavedAnnotations = function(callbackOnOK, callbackOnCancel, callbackOnDontSave) {
	if (callbackOnCancel === undefined) {
		callbackOnCancel = callbackOnOK;
	}

	if (this._page && this.getUnsavedAnnotationCount() > 0 && this.canChangeAnnotations(true)) {
		this.f.loadBundle(
			"viewer",
			function() {
				// show popup if there are unsaved items
				var modalPopup = new ModalPopup(this.f, "unsavedAnnotationsPopup", this.f.getContext());
				modalPopup.setRemoveFromDomAfterUsage(true);
				modalPopup.setOkAction(
					_.bind(function() {
						modalPopup.enableOkBtn(false);

						// save annotations
						this.save(
							null,
							_.bind(function() {
								// saving succeeded, so hide
								modalPopup.hide();
								if (callbackOnOK) {
									callbackOnOK();
								}
								this.toggleVisibility(true);
							}, this)
						);
					}, this)
				);

				modalPopup.setCancelAction(
					_.bind(function() {
						if (callbackOnCancel) {
							callbackOnCancel();
						}
						this.toggleVisibility(true);
					}, this)
				);

				if (callbackOnDontSave) {
					var dontSaveButton = new ViewerToolbarButton(
						this.f,
						"dontSaveAnnotations",
						$("<span>").append(this.f.t("RevertAnnChanges")),
						false,
						_.bind(function() {
							modalPopup.hide();
							this.toggleVisibility(true);

							if (callbackOnDontSave) {
								callbackOnDontSave();
							}
						}, this)
					);
					modalPopup.addButton(dontSaveButton);
				}

				// Header + Content
				modalPopup.setHeader(
					$("<span/>")
						.addClass("popupTitle")
						.text(this.f.i18n.t("AnnTitle"))
				);
				modalPopup.setContent(
					$("<div/>")
						.addClass("unsavedItemsContent")
						.text(this.f.i18n.t("AnnSavePrompt"))
				);
				modalPopup.setOkBtnText(this.f.i18n.t("Save"));
				this.toggleVisibility(false);
				modalPopup.show();
				modalPopup.recalculateSize(600);
			},
			this
		);

		return true;
	} else {
		return false;
	}
};

/**
 * Set to lock the selection (true) of annotations or not (false)
 * @param {Boolean} lockSelection - True to lock the selection, false to not lock the selection
 */
AnnotationModel.prototype.setLockSelection = function(lockSelection) {
	this.lockSelection = lockSelection;
};

/**
 * Set the layout type, the widget needs to be rendered for
 * @param layoutType
 */
AnnotationModel.prototype.setLayoutType = function(layoutType) {
	this._layoutType = layoutType;
};

/**
 * Get the layout type the widget should be rendered for
 * @returns {LayoutConstants.LayoutType)
 */
AnnotationModel.prototype.getLayoutType = function() {
	return this._layoutType;
};

/**
 * Scale all the annotation of the annotation model with a given factor, both horizontal as vertical, changes all the points and rectangle of the annotation,
 * first undos the scaling and than applies the new scaling, if no scaleFactor is provided the current scaling is undone
 * @param {ShapeUtils.Point} scaleFactor
 */
AnnotationModel.prototype.scaleAllAnnotations = function(scaleFactor) {
	if (this._currentScaling) {
		_.each(
			this._annotations,
			function(annotation) {
				if (annotation.isScaled()) {
					annotation.scale(new ShapeUtils.Point(1 / annotation.getScale().x, 1 / annotation.getScale().y));
				}
			},
			this
		);
	}
	this._currentScaling = scaleFactor;
	_.each(this._annotations, function(annotation) {
		annotation.setScale(scaleFactor);
	});
};

/**
 * Get the currently active scaling
 * @returns {ShapeUtils.Point}
 */
AnnotationModel.prototype.getCurrentScaling = function() {
	return this._currentScaling;
};

/**
 * Get the model for the annotation filter
 * @returns {AnnotationFilterModel}
 */
AnnotationModel.prototype.getAnnotationFilterModel = function() {
	return this._annotationFilterModel;
};

/**
 * This method checks if the property "AnnotationModel.zoom" is available and zooms to the annotation if present.
 */
AnnotationModel.prototype._checkForZoomToAnnotation = function() {
	var zoomToAnnotation = this.f.getProperty("AnnotationModel.zoom");
	if (zoomToAnnotation) {
		var selectedAnnotation = this.getAnnotation(zoomToAnnotation.annotationName);
		if (selectedAnnotation) {
			if (this._zoomToAnnotationListener) {
				this._unregisterListener(this._zoomToAnnotationListener);
			}
			this.unselectAllAnnotations(true, true);
			this.f.emit("module.annotation.zoomToAnnotation", selectedAnnotation, !zoomToAnnotation.animation, !zoomToAnnotation.select);
			// We need to listen to the repainted event and zoom again to combat concurrency issues
			this._zoomToAnnotationListener = this._registerListener(
				"widget.ViewportWidget.repainted",
				_.bind(function() {
					this.unselectAllAnnotations(true, true);
					this.f.emit("module.annotation.zoomToAnnotation", selectedAnnotation, !zoomToAnnotation.animation, !zoomToAnnotation.select);
					if (this._zoomToAnnotationListener) {
						this._unregisterListener(this._zoomToAnnotationListener);
						this._zoomToAnnotationListener = null;
					}
				}, this),
				this
			);
		}
	}
};

/**
 * Checks if the annotation popups should always be shown.
 * @returns {boolean} True if all should be visible, false otherwise.
 */
AnnotationModel.prototype.areAllAnnotationPopupsShown = function() {
	return this._showAllAnnotationPopups && !this.f.getProperty("AnnotationModel.hideAllPopups");
};

/**
 * Change the visibility of forcing visibility on all annotation popups
 * @param {boolean} newValue True if popups should be visible, false otherwise.
 */
AnnotationModel.prototype.setShowAllAnnotationPopups = function(newValue) {
	this._showAllAnnotationPopups = newValue;
	this._emit("showAllAnnotationPopupsChanged");
};

/**
 * Change the default behaviour when selecting an annotation
 * @param {boolean} newValue True if the viewer should auto fit to annotation when selected, false otherwise.
 */
AnnotationModel.prototype.setAutoFitViewToSelectedAnnotation = function(newValue) {
	this._autoFitViewToSelectedAnnotation = newValue;
	this._emit("fitViewToSelectedAnnotationChanged");
};

/**
 * Check if the viewer should auto fit view to newly selected annotation
 * @returns {boolean} True if should auto fit, false otherwise.
 */
AnnotationModel.prototype.shouldAutoFitViewToSelectedAnnotation = function() {
	return this._autoFitViewToSelectedAnnotation;
};

/**
 * @returns the time perceived by the annotation model
 */
AnnotationModel.prototype.getTimeStamp = function() {
	if (this.f.getActivePage().hasVideo() && this._videoModule) {
		var video = this._videoModule.getVideo();
		if (video) {
			return video.currentTime;
		}
	}
};

/**
 * This callback can be used to intercept the visibility of annotations.
 * If the callback returns false then the annotation is hidden.
 *
 * @param {Function} callback return true to show annotation, false to hide the annotation
 */
AnnotationModel.prototype.interceptAnnotationVisibility = function(callback) {
	if (!this._interceptAnnotationVisibilityCallbacks) {
		this._interceptAnnotationVisibilityCallbacks = [];
	}
	this._interceptAnnotationVisibilityCallbacks.push(callback);
};

export default AnnotationModel;
