/**
 *  @namespace AnnotationModule
 */
/**
 *  @namespace AnnotationModule
 */
import _ from "underscore";

import $ from "jquery";
import AbstractModule from "@esko/web-viewer/core/AbstractModule";
import AnnotationModel from "@esko/web-viewer/annotation/AnnotationModel";
import AnnotationLayerModel from "@esko/web-viewer/annotation/AnnotationLayerModel";
import AnnotationPopupsLayerModel from "@esko/web-viewer/annotation/AnnotationPopupsLayerModel";
import AnnotationTool from "@esko/web-viewer/annotation/AnnotationTool";
import AnnotationDrawTool from "@esko/web-viewer/annotation/AnnotationDrawTool";
import AnnotationIndicatorTool from "@esko/web-viewer/annotation/AnnotationIndicatorTool";
import AnnotationWidget from "@esko/web-viewer/annotation/AnnotationWidget";
import AnnotationFactory from "@esko/web-viewer/annotation/AnnotationFactory";
import DefaultAnnotationDataExtensions from "@esko/web-viewer/annotation/DefaultAnnotationDataExtensions";
import ShapeUtils from "@esko/shapeutils";
import AnnotationNoteDrawTool from "@esko/web-viewer/annotation/AnnotationNoteDrawTool";
import AnnotationTouchDrawTool from "@esko/web-viewer/annotation/AnnotationTouchDrawTool";
import AnnotationTouchTool from "@esko/web-viewer/annotation/AnnotationTouchTool";
import AnnotationToolbarComponent from "@esko/web-viewer/annotation/AnnotationToolbarComponent";
import AnnotationTouchToolbarComponent from "@esko/web-viewer/annotation/AnnotationTouchToolbarComponent";
import AnnotationConstants from "@esko/web-viewer/annotation/AnnotationConstants";
import ModalPopup from "@esko/web-viewer/layout/component/ModalPopup";
import LayoutConstants from "@esko/web-viewer/layout/LayoutConstants";
import ListenerUtil from "@esko/web-viewer/utils/ListenerUtil";
/**
 * @exports AnnotationModule
 * @memberof AnnotationModule
 * @class The annotation module is the main module that controls the loading and displaying of the annotations
 * @augments Core.AbstractModule
 * @param {Core.ViewerFacade} f - Viewer facade
 * @param {Boolean} silenced - if true, the module shouldn't publish tools/widgets/toolbarcomponents
 * @constructor
 */
var AnnotationModule = function(f, silenced) {
	AbstractModule.call(this, f, silenced, "annotation");
	ListenerUtil.mixin(this, f);

	DefaultAnnotationDataExtensions(f);

	/**
	 * Current annotation model
	 */
	this.annotationModel = this._createModel();

	/**
	 * Layer Models
	 */
	this.layerModels = {};

	/**
	 * Tools
	 */
	// NOTE we should only create the needed tools... based on the device mode
	this._drawTool = new AnnotationDrawTool(f, this.annotationModel, this);
	this._noteDrawTool = new AnnotationNoteDrawTool(f, this.annotationModel, this);
	this._indicatorTool = new AnnotationIndicatorTool(f, this.annotationModel);

	this._publishTool(this._indicatorTool);

	if (this.f.isMobile()) {
		// only publish the annotation touch draw tool
		this._touchDrawTool = new AnnotationTouchDrawTool(f, this.annotationModel, this);
		this._publishTool(this._touchDrawTool);
		this._selectTool = new AnnotationTouchTool(f, this.annotationModel);
		this._publishTool(this._selectTool);
	} else {
		// desktop
		// publish both the draw tool and not draw tool
		this._publishTool(this._drawTool);
		this._publishTool(this._noteDrawTool);

		this._selectTool = new AnnotationTool(f, this.annotationModel);
		this._publishTool(this._selectTool);
	}

	/**
	 * Component: annotate
	 */
	this._annotateComponent = null; // set the annotation tool bar component as unavailable

	// publish annotate tool bar component
	this._publishToolbarComponent(
		AnnotationToolbarComponent.toolbarName,
		_.bind(function(layoutType) {
			if (layoutType === LayoutConstants.LayoutType.MOBILE) {
				this._annotateComponent = new AnnotationTouchToolbarComponent(f, this.annotationModel, this._touchDrawTool);
			} else {
				this._annotateComponent = new AnnotationToolbarComponent(f, this.annotationModel, this._drawTool, this._noteDrawTool);
			}
			return this._annotateComponent;
		}, this)
	);

	/**
	 * Widget(s)
	 */
	this._annotationWidget = null; // set the annotation widget as unavailable

	// publish annotation widget
	this._publishWidget(
		AnnotationWidget.widgetName,
		_.bind(function(layoutType) {
			this.annotationModel.setLayoutType(layoutType);
			this._annotationWidget = new AnnotationWidget(f, this, this.annotationModel, layoutType);
			return this._annotationWidget;
		}, this)
	);

	this.on(
		"highlightAnnotation",
		_.bind(function(annotation) {
			this.annotationModel.selectAnnotation(annotation, true);
			this._indicatorTool.setAnnotation(annotation);

			f.emit("toolManager.request.activation", "annotationindicatortool");
		}, this)
	);

	this.on(
		"duplicateAnnotation",
		_.bind(function(annotation) {
			if (annotation.getPermissionFor(AnnotationConstants.ActionPermissions.DUPLICATE)) {
				this.annotationModel.getAnnotationFilterModel().setOriginType(AnnotationConstants.AnnotationOriginTypes.user.id);
				this._duplicateAnnotation(annotation, false, false, true, true);
			}
		}, this)
	);

	this.on(
		"duplicateWithoutText",
		_.bind(function(annotation) {
			if (annotation.getPermissionFor(AnnotationConstants.ActionPermissions.DUPLICATE_WITHOUT_TEXT)) {
				this.annotationModel.getAnnotationFilterModel().setOriginType(AnnotationConstants.AnnotationOriginTypes.user.id);
				this._duplicateAnnotation(annotation, false, false, true, true, true);
			}
		}, this)
	);

	this.on(
		"duplicateAllFilteredAnnotations",
		_.bind(function() {
			var filteredAnnotations = _.filter(this.annotationModel.getFilteredAnnotations(), function(annotation) {
				return annotation.getPermissionFor(AnnotationConstants.ActionPermissions.DUPLICATE);
			});
			if (filteredAnnotations.length > 0) {
				this.annotationModel.getAnnotationFilterModel().setOriginType(AnnotationConstants.AnnotationOriginTypes.user.id);
				_.each(
					filteredAnnotations,
					function(annotation) {
						this._duplicateAnnotation(annotation, false, false, true, true, false, true);
					},
					this
				);
				this.annotationModel.modelChanged();
				this.annotationModel.modelContentChanged();
			}
		}, this)
	);

	this.on(
		"duplicateAllFilteredAnnotationWithoutText",
		_.bind(function() {
			var filteredAnnotations = _.filter(this.annotationModel.getFilteredAnnotations(), function(annotation) {
				return annotation.getPermissionFor(AnnotationConstants.ActionPermissions.DUPLICATE_WITHOUT_TEXT);
			});
			if (filteredAnnotations.length > 0) {
				this.annotationModel.getAnnotationFilterModel().setOriginType(AnnotationConstants.AnnotationOriginTypes.user.id);
				_.each(
					filteredAnnotations,
					function(annotation) {
						this._duplicateAnnotation(annotation, false, false, true, true, true, true);
					},
					this
				);
				this.annotationModel.modelChanged();
				this.annotationModel.modelContentChanged();
			}
		}, this)
	);

	this.on("unselectAnnotations", _.partial(this.annotationModel.unselectAllAnnotations).bind(this.annotationModel));

	this.on(
		"zoomToAnnotation",
		_.bind(function(annotation, disableAnimation, doNotSelectAnnotation) {
			if (!annotation.getPermissionFor(AnnotationConstants.ActionPermissions.ZOOM_TO_ANNOTATION)) {
				return;
			}
			var options = {};

			/**
			 * Constrain the zoom so it never goes lower than 2 cm
			 */
			var zoomRect = annotation.rect.copy();
			if (f.getOption("PREFERENCES_ANNOTATION_NEXT_MODE") !== 3) {
				zoomRect = annotation.getOutliningRect();
			}
			var minSize = 20;

			var diff = null;
			if (Math.abs(zoomRect.size.width) < minSize) {
				diff = (minSize - zoomRect.size.width) / 2;
				zoomRect.origin.x -= diff;
				zoomRect.size.width = minSize;
			}
			if (Math.abs(zoomRect.size.height) < minSize) {
				diff = (minSize - zoomRect.size.height) / 2;
				zoomRect.origin.y -= diff;
				zoomRect.size.height = minSize;
			}

			options.rect = convertAnnotationRectToWorld(zoomRect);
			options.constrain = true;
			options.animated = !disableAnimation;

			this.annotationModel.ensureVisibility();

			// Select annotation
			if (this.annotationModel.getLayoutType() !== LayoutConstants.LayoutType.MOBILE && !doNotSelectAnnotation) {
				this.annotationModel.selectAnnotation(annotation, true);
			}

			f.emit("viewport.fitInRect", options);
		}, this)
	);

	// with this event annotation module can be temporarily disabled
	this.on(
		"temporarilyDisable",
		_.bind(function(disable, enableTogglingAnnotations, reason, whichToEnable) {
			this.annotationModel.temporarilyDisable(disable, enableTogglingAnnotations, reason, whichToEnable);
		}, this)
	);

	var getAnnotationTransform = _.bind(function() {
		return this.layerModels["AnnotationLayer"].getTransform().copy();
	}, this);

	function convertAnnotationRectToWorld(annotationRect) {
		return getAnnotationTransform().transformedRect(annotationRect);
	}

	this.convertWorldRectToAnnotationLayer = function(worldRect) {
		var transform = getAnnotationTransform().createInverse();
		return new ShapeUtils.Rect(transform.transformedPoint(worldRect.origin), transform.transformedSize(worldRect.size));
	};

	this.on(
		"createAnnotation",
		_.bind(function(rect, content, type) {
			if (type === undefined) {
				type = AnnotationConstants.AnnotationType.ANNOT_RECT;
			}
			this.annotationModel.addAnnotationWithRect(this.convertWorldRectToAnnotationLayer(rect), content, type);
		}, this)
	);

	this._registerListener(
		"intercept.activePageChange",
		function(context) {
			var newActivePage = context.newActivePage;
			var oldActivePage = context.oldActivePage;
			if (newActivePage !== oldActivePage) {
				var hasUnsavedAnnotations = this.promptToSaveUnsavedAnnotations(
					_.bind(function() {
						this.f.getDocumentController().changeActivePage(newActivePage);
					}, this),
					_.bind(function() {
						//Don't change the active page, cancel doesn't change the page and doesn't save the annotations
					}, this),
					_.bind(function() {
						this.annotationModel.resetAllAnnotations(); //don't save annotations removes all the unsaved annotations
						this.f.getDocumentController().changeActivePage(newActivePage); //change the active page, the annotations of the previous page are not saved
					}, this)
				);
				if (hasUnsavedAnnotations) {
					context.newActivePage = oldActivePage;
				}
			}
		},
		this
	);

	this._registerListener(
		"viewControls.compensationChanged",
		function() {
			if (this.annotationModel) {
				var page = this.annotationModel.getAnnotationsPage();

				//TODO allow the reference page the compensated, the Java Viewer does not allow this, as it returns wrong parameter values
				if (page !== this.f.getDocumentController().getDifferencePage()) {
					var scaling = page.getCurrentScaling();

					this.annotationModel.scaleAllAnnotations(scaling);
				}
			}
		},
		this,
		this.f
	);

	this._registerListener(
		"filterStateChanged",
		function(doPersist) {
			if (doPersist) {
				this.f.setOption("ANNOTATION_FILTER", this.annotationModel.getAnnotationFilterModel().getFilterState());
			} else {
				this.f.resetOption("ANNOTATION_FILTER");
			}
		},
		this,
		this.annotationModel.getAnnotationFilterModel()
	);

	// React on annotation added to save 3D View
	this._registerListener(
		"annotationAdded",
		function(annotation) {
			if (!this.f.isMobile()) {
				this._save3DView(annotation, undefined, true);
			}
		},
		this,
		this.annotationModel
	);

	this.on(
		"update3DView",
		_.bind(function(annotation) {
			this._save3DView(annotation);
		}, this)
	);

	this.on(
		"delete3DView",
		_.bind(function(annotation) {
			this._save3DView(annotation, true);
		}, this)
	);

	this.on(
		"show3DView",
		_.bind(function(annotation) {
			this._show3DView(annotation);
		}, this)
	);

	this._registerListener(
		"selectedAnnotation",
		function(annotation) {
			if (this.annotationModel.getLayoutType() !== LayoutConstants.LayoutType.MOBILE) {
				this._show3DView(annotation.getGodfather(), false);
			}
		},
		this,
		this.annotationModel
	);

	// When the 3D widget is available repaint the rows to update action visiblity
	this.f.whenModuleAvailable(
		"view3D",
		_.bind(function(view3DModule) {
			view3DModule.on(
				"3DWidget.enabled",
				_.bind(function(enabled) {
					this.annotationModel.modelContentChanged();
				}, this)
			);
		}, this)
	);

	// React on active page changes
	this.f
		.getDocumentController()
		.whenActiveOrDifferencePageChangedAndParametersAvailable(this._whenActiveOrDifferencePageChangedAndParametersAvailable, this);
	this.f
		.getDocumentController()
		.whenActiveOrDifferencePageChangedAndParametersAvailableOrChange(
			this._whenActiveOrDifferencePageChangedAndParametersAvailableOrChange,
			this
		);
};

// inherit
AnnotationModule.prototype = Object.create(AbstractModule.prototype);
// correct the constructor pointer
AnnotationModule.prototype.constructor = AnnotationModule;

/**
 * Show the 3D view of the annotation if the 3D module is available
 * @param {Annotation} annotation - The annotation to show the 3D view of
 * @param {Boolean} [doSelect = true] - Select the annotation also
 */
AnnotationModule.prototype._show3DView = function(annotation, doSelect) {
	var view3DModule = this.f.getModule("view3D");
	var position = annotation.get3DViewPosition();
	if (view3DModule && position) {
		// Extract scale out of position
		const { scale, ...newPosition } = position;
		// If distance is undefined, it means we're dealing with an old annotation that has a scale value.
		if (newPosition.distance == null) {
			//TODO STU-6914, wait for proper API call
			newPosition.distance = view3DModule._3DWidget._3DLauncher._viewer._controller.get("_view").convertScaleToDistance(scale);
		}
		if (this.annotationModel.getLayoutType() === LayoutConstants.LayoutType.MOBILE) {
			view3DModule.show3D(newPosition);
		} else {
			view3DModule.setCameraPosition(newPosition);
		}
		if (doSelect === undefined || doSelect) {
			this.annotationModel.selectAnnotation(annotation, true); // also select
		}
	}
};

/**
 * Save the current 3D view in the annotation
 * @param {Annotation} annotation - The annotation to set the 3D view of
 * @param {Boolean} doDelete - Indicating to clear the current 3D view rather than setting a new one
 * @param {Boolean} isNewAnnotation - Indicates if the function is called because a new annotation has been added
 */
AnnotationModule.prototype._save3DView = function(annotation, doDelete, isNewAnnotation) {
	// only for main annotations
	if (annotation.getGodfather() === annotation) {
		if (doDelete) {
			annotation.set3DViewPosition(null);
			this.annotationModel.selectAnnotation(annotation, true, isNewAnnotation); // also select
			this.annotationModel.modelChanged(annotation);
		} else {
			var view3DModule = this.f.getModule("view3D");
			if (view3DModule) {
				var position = view3DModule.getCameraPosition();
				if (position) {
					annotation.set3DViewPosition(position);
					this.annotationModel.selectAnnotation(annotation, true, isNewAnnotation); // also select
					this.annotationModel.modelChanged(annotation);
				}
			}
		}
	}
};

/**
 * Are there still unsaved element?
 * @override
 * @returns {Boolean}
 */
AnnotationModule.prototype.hasUnsavedElements = function() {
	return this.annotationModel.getUnsavedAnnotationCount() > 0;
};

/**
 * Save all unsaved elements
 * @override
 * @param callbackFunction
 */
AnnotationModule.prototype.saveAllUnsavedElements = function(callbackFunction) {
	this.annotationModel.save(null, callbackFunction);
};

/**
 * When the active page has changed, and the parameters are availabled for this page, reinitialize annotations model
 * @param {data.AbstractDocumentPage} page
 */
AnnotationModule.prototype._whenActiveOrDifferencePageChangedAndParametersAvailableOrChange = function(page, difference) {
	//handle new page
	if (this.layerModels["AnnotationLayer"] === undefined) {
		// Create the annotations layer
		this.layerModels["AnnotationLayer"] = new AnnotationLayerModel(this.f, "Annotations", this.annotationModel);
		// Create the annotation popups layer
		this.layerModels["AnnotationPopupsLayer"] = new AnnotationPopupsLayerModel(this.f, "AnnotationPopups", this.annotationModel);

		var viewportModule = this.f.getModule("viewport");
		// Global layer
		viewportModule.addLayer("Annotations", this.layerModels["AnnotationLayer"]);
		// Main layer only
		viewportModule.addLayer("AnnotationPopups", this.layerModels["AnnotationPopupsLayer"], "Main");
	}

	if (this.annotationModel.isTemporarilyDisabled()) {
		if (this.annotationModel.isTemporarilyCurrentPageAnnotationsShown()) {
			this.annotationModel.setPage(page);
			this.annotationModel.modelChanged();
		} else if (this.annotationModel.isTemporarilyDifferencePageAnnotationsShown()) {
			this.annotationModel.setPage(difference);
			this.annotationModel.modelChanged();
		}
	} else {
		this.annotationModel.setPage(page);
	}
};

/**
 * When active or difference page changes, calculate the annotation offsets for both active and/or reference CAD page, only for when those pages are in compare
 * @param {AbstractDocumentPage} activePage
 * @param {AbstractDocumentPage} referencePage
 * @param {AbstractDocumentPage|null} oldPage
 * @param {Boolean} isCurrentChanged
 */
AnnotationModule.prototype._whenActiveOrDifferencePageChangedAndParametersAvailable = function(
	activePage,
	referencePage,
	oldPage,
	isCurrentChanged
) {
	if (this._annotateComponent) {
		if (activePage.isBlankPage()) {
			this._annotateComponent.hide();
		} else {
			this._annotateComponent.show();
		}
	}
};

/**
 * Toggle whether the difference page annotations should be shown temporarily
 * @param {Boolean} [toggle]
 */
AnnotationModule.prototype.toggleDifferencePageAnnotations = function(toggle) {
	if (toggle === undefined) {
		toggle = !this.annotationModel.isTemporarilyDifferencePageAnnotationsShown();
	}
	if (toggle) {
		this.f.setOption("PREFERENCES_ANNOTATIONTOOLS_VISIBLE", 1, true);
		this.annotationModel.showTemporarilyDifferencePageAnnotations();
	} else {
		// hide them again
		this.annotationModel.hideTemporarilyAnnotations();
	}
};

/**
 * Toggle whether the active/current page annotations should be shown temporarily
 * @param {Boolean} [toggle]
 */
AnnotationModule.prototype.toggleCurrentPageAnnotations = function(toggle) {
	if (toggle === undefined) {
		toggle = !this.annotationModel.isTemporarilyCurrentPageAnnotationsShown();
	}
	if (toggle) {
		this.f.setOption("PREFERENCES_ANNOTATIONTOOLS_VISIBLE", 1, true);
		this.annotationModel.showTemporarilyActivePageAnnotations();
	} else {
		this.annotationModel.hideTemporarilyAnnotations();
	}
};

/**
 * @override
 * @returns {String}
 */
AnnotationModule.prototype.getUnsavedElementsName = function() {
	// note we don't wait for the bundle to be loaded... but it's not even possible in this situation
	// this language var should normally be loaded already...
	return this.f.i18n.t("Annotations");
};

/**
 * Are annotations for the current document currently shown (annotations are temporarily disabled because of compare)
 * @borrows {AnnotationModule.AnnotationModel~isTemporarilyCurrentPageAnnotationsShown}
 * @returns {Boolean}
 */
AnnotationModule.prototype.isTemporarilyCurrentPageAnnotationsShown = function() {
	return this.annotationModel.isTemporarilyCurrentPageAnnotationsShown();
};

/**
 * Are annotations for the difference document currently shown (annotations are temporarily disabled because of compare)
 * @borrows {AnnotationModule.AnnotationModel~isTemporarilyDifferencePageAnnotationsShown}
 * @returns {Boolean}
 */
AnnotationModule.prototype.isTemporarilyDifferencePageAnnotationsShown = function() {
	return this.annotationModel.isTemporarilyDifferencePageAnnotationsShown();
};

/**
 * @override
 */
AnnotationModule.prototype.getState = function() {
	var state = {
		// Annotation model
		annotationModel: {
			temporarilyDisabled: {
				temporarilyDisabled: this.annotationModel.isTemporarilyDisabled(),
				temporarilyCurrentPageAnnotationsShown: this.annotationModel.isTemporarilyCurrentPageAnnotationsShown(),
				temporarilyDifferencePageAnnotationsShown: this.annotationModel.isTemporarilyDifferencePageAnnotationsShown(),
				temporarilyDisabledMessage: this.annotationModel.getTemporarilyDisabledMessage(),
				temporaryAnnotationDocumentTogglingEnabled: this.annotationModel.isTemporarilyAnnotationDocumentTogglingEnabled()
			},
			annotationVisibility: this.f.getOption("PREFERENCES_ANNOTATIONTOOLS_VISIBLE") == 1,
			filter: this.annotationModel.getAnnotationFilterModel().getFilterState()
		}
	};
	return state;
};

/**
 * @override
 * @param {Object} state
 */
AnnotationModule.prototype.setState = function(state) {
	this.f.whenModuleAvailable(
		"annotation",
		function() {
			// This module is ready
			if (state !== undefined) {
				var annotationModelState = state.annotationModel;
				if (state.disableSelectTool) {
					this._selectTool = null;
					this._unpublishTool("annotationtool");
					this._indicatorTool = null;
					this._unpublishTool("annotationindicatortool");
				}
				// Restore Annotation model
				if (annotationModelState !== undefined) {
					var temporarilyDisabledState = annotationModelState.temporarilyDisabled;
					if (temporarilyDisabledState !== undefined) {
						this.annotationModel.temporarilyDisable(
							temporarilyDisabledState.temporarilyDisabled,
							temporarilyDisabledState.temporaryAnnotationDocumentTogglingEnabled,
							temporarilyDisabledState.temporarilyDisabledMessage
						);
						if (temporarilyDisabledState.temporarilyCurrentPageAnnotationsShown) {
							this.annotationModel.showTemporarilyActivePageAnnotations();
						} else if (temporarilyDisabledState.temporarilyDifferencePageAnnotationsShown) {
							this.annotationModel.showTemporarilyDifferencePageAnnotations();
						}
					}
					this.annotationModel.toggleVisibility(annotationModelState.annotationVisibility);

					if (annotationModelState.filter) {
						this.annotationModel.getAnnotationFilterModel().setFilterState(annotationModelState.filter);
					}
				}
			}
		},
		this
	);
};

/**
 * Prompt to save unsaved annotations if there are any
 * if there are unsaved annotations, save them
 *
 * @param {Function} callbackOnOK - function which is called when the popup is closed and annotations are saved
 * @param {Function} [callbackOnCancel] - if provided, this callback is called when the user presses cancel (if not defined, callbackOnOK is executed)
 * @param {Function} [callbackOnDontSave] - adds the possibility to not save unsaved annotation but to continue with the action, instead cancal will cancel the action and dnon't save
 * @returns {Boolean} true if there are unsaved items and the popup is shown, false if not
 */
AnnotationModule.prototype.promptToSaveUnsavedAnnotations = function(callbackOnOK, callbackOnCancel, callbackOnDontSave) {
	return this.annotationModel.promptToSaveUnsavedAnnotations(callbackOnOK, callbackOnCancel, callbackOnDontSave);
};

/**
 * Check if the annotation is outside the currently active page, if this is the case prompt to check if the user still wants to add the annotation
 * @param {ShapeUtils.Rect} rect
 * @param {Function} callbackOnOk
 * @param {Function} callbackOnCancel
 * @param {Boolean} [isCreatingLinkedAnnot]
 * @param {Boolean} [isWorldRect]
 */
AnnotationModule.prototype.promptIfAnnotationOutsidePage = function(
	rect,
	callbackOnOk,
	callbackOnCancel,
	isCreatingLinkedAnnot,
	isWorldRect
) {
	if (isWorldRect === true) {
		rect = this.convertWorldRectToAnnotationLayer(rect);
	}
	var page = this.f.getDocumentController().getActivePage();
	var isRectOnPage = true;
	if (!page.hasCad()) {
		isRectOnPage = page.isRectOnPage(rect);
	}

	//show modal popup with the question if the user wants to add an annotation that's not on the currently active page (if needed)
	if (!isRectOnPage) {
		var modalPopup = new ModalPopup(this.f, "annotationOnOtherPage", this.f.getContext());
		modalPopup.setRemoveFromDomAfterUsage(true);
		modalPopup.setOkAction(
			_.bind(function() {
				callbackOnOk(isCreatingLinkedAnnot);
				modalPopup.hide();
			}, this)
		);

		modalPopup.setCancelAction(
			_.bind(function() {
				callbackOnCancel();
			}, this)
		);

		// Header + Content
		modalPopup.setHeader(
			$("<span/>")
				.addClass("popupTitle")
				.text(this.f.i18n.t("AnnSaveTitle"))
		);
		modalPopup.setContent(
			$("<div/>")
				.addClass("unsavedItemsContent")
				.text(this.f.i18n.t("AnnSaveOutsidePrompt"))
		);
		modalPopup.setOkBtnText(this.f.i18n.t("YesResponse"));
		modalPopup.setCancelBtnText(this.f.i18n.t("NoResponse"));

		modalPopup.show();
	} else {
		callbackOnOk();
	}
};

/**
 * Reload the annotations (throwing away unsaved actions!)
 */
AnnotationModule.prototype.reload = function() {
	this.annotationModel.reload(true);
};

/**
 * Disable annotations with a given reason
 * @param {Boolean} disable - disable or enable the annotations for given reason (key is used as id for the reason)
 * @param {String} key - the key used in the reasons
 * @param {String} disabledText - text used to display in the annotationWidget for which reason the annotations were disabled
 * @param {Boolean} [showAnnotations] - Set to true to keep annotations visible, by default they are hidden
 */
AnnotationModule.prototype.disableAnnotations = function(disable, key, disabledText, showAnnotations) {
	if (disable) {
		this.annotationModel.setDisableReason(key, disabledText, !showAnnotations, true);
	} else {
		this.annotationModel.removeDisableReason(key);
	}
};

AnnotationModule.prototype.getWarnWhenChangingAnnotationsLock = function() {
	return this.annotationModel.warnWhenChangingAnnotations.getLock();
};

/**
 * duplicates the annotation
 * @param {Annotation} annotation - the annotation to duplicate
 * @param {Boolean} keepReplies - True if you want to copy the replies to the duplicated annotation, false otherwise
 * @param {Boolean} keepStates - True if you want to copy the states to the duplicated annotation, false otherwise
 * @param {Boolean} keepInklist - True if you want to copy the pointArray to the duplicated annotation, false otherwise
 * @param {Boolean} keepLinkedAnnotations - True if you want to copy the linked annotations to the duplicated annotation, false otherwise
 * @param {Boolean} [duplicateWithoutText = false] - whether or not to remove the content of the annotation
 * @param {Boolean} [silent = false] - whether or not to keep silent about the changes

 */
AnnotationModule.prototype._duplicateAnnotation = function(
	annotation,
	keepReplies,
	keepStates,
	keepInklist,
	keepLinkedAnnotations,
	duplicateWithoutText,
	silent
) {
	var ann = annotation.getGodfather().duplicate(keepReplies, keepStates, keepInklist, keepLinkedAnnotations, this.annotationModel);
	if (duplicateWithoutText) {
		ann.content = "";
	}

	if (annotation.inreplyto) {
		ann.content = _.clone(annotation.content);
	}

	// fix the color of the duplicate (to get the current drawing color we refer to the draw tools)
	var color = null;
	if (ann.getType() == AnnotationConstants.AnnotationType.ANNOT_PIN) {
		color = this._noteDrawTool.getAnnotationColor();
	} else {
		color = this._drawTool.getAnnotationColor();
	}
	ann.setColor(color);

	this.annotationModel.unselectAllAnnotations(true, true);
	// add annotation so it gets a number
	this.annotationModel.addAnnotation(ann, false, silent);
	if (ann.links.length > 0) {
		// copy all the links
		_.each(
			ann.links,
			function(linkedAnnotation) {
				linkedAnnotation.messagenumber = ann.messagenumber;
				this.annotationModel.addAnnotation(linkedAnnotation);
			},
			this
		);
	}
	// make sure the linked annotation are also in the correct color
	ann.setColor(color);
	if (!silent) {
		this.annotationModel.modelChanged(ann);
	}
};

/**
 * Creates a series of linked annotations.
 * @param {Array.<ShapeUtils.Rect>} rects - The array of world rects for which annotations should be made. The first rect is considered the main annotaiton to which the others link
 * @param {String} content - the content of the first annotation
 * @param {String} type - the type of annotation to make
 */
AnnotationModule.prototype.addLinkedAnnotations = function(rects, content, type) {
	var newAnnotation = null;
	var firstAnnotation = null;
	for (var i = 0; i < rects.length; i++) {
		// Build the annotation (first annotation has selected = true)
		newAnnotation = new AnnotationFactory(this.f).create();
		newAnnotation.initAnnotation(
			type,
			this._getAnnotationColor(),
			i === 0,
			this.annotationModel,
			this.convertWorldRectToAnnotationLayer(rects[i])
		);

		if (i === 0) {
			// Create the first annotation with textContent
			newAnnotation.content = content;
			firstAnnotation = newAnnotation;
			this.annotationModel.unselectAllAnnotations(true, true);
			this.annotationModel.addAnnotation(newAnnotation);
		} else {
			// Add the other annotations linked to the first annotation
			this.annotationModel.addLinkedAnnotation(newAnnotation, firstAnnotation);
		}
	}
};

/**
 * Get the annotation color
 */
AnnotationModule.prototype._getAnnotationColor = function() {
	if (this.f.getProperty("annotationcolor")) {
		return this.f.getProperty("annotationcolor");
	} else {
		return this.f.getCurrentUser().getColor("PREFERENCES_ANNOTATION_RECTANGLE_COLOR");
	}
};

AnnotationModule.prototype._createModel = function() {
	return new AnnotationModel(this.f, "annotation");
};

AnnotationModule.prototype.whenAnnotationsAvailableOrChanged = function(callback) {
	var page = this.f.getActivePage();
	if (page) {
		page.requestAnnotations(function() {
			var annots = page.getAnnotations();
			callback(annots);
		});
	}

	this.annotationModel.on(
		"modelChanged",
		_.bind(function() {
			var aPage = this.f.getActivePage();
			var annots = aPage.getAnnotations();
			callback(annots);
		}, this)
	);
};

export default AnnotationModule;
