import _ from "underscore";
import AbstractTool from "@esko/web-viewer/viewport/tools/AbstractTool";
import ShapeUtils from "@esko/shapeutils";
import AnnotationFactory from "@esko/web-viewer/annotation/AnnotationFactory";
import AnnotationConstants from "@esko/web-viewer/annotation/AnnotationConstants";
import ViewerUtils from "@esko/web-viewer/utils/ViewerUtils";
import CoreConstants from "@esko/web-viewer/core/CoreConstants";

/**
 * This is the tool for drawing note annotations.
 * Also see it's related tool AnnotationDrawTool.
 * @exports AnnotationNoteDrawTool
 * @memberof AnnotationModule
 * @constructor
 *
 * @param {Core.ViewerFacade} f
 * @param {AnnotationModel} annotationModel - the model used by the module for annotations
 * @param {AnnotationModule} annotationModule - the annotation module reference
 */
var AnnotationNoteDrawTool = function(f, annotationModel, annotationModule) {
	AbstractTool.call(this, f, CoreConstants.TOOLS.ANNOTATION_NOTE_DRAW);

	var defaultState = "default";

	this.drawLinkedAnnotation = false;

	this._annotationModel = annotationModel;

	this._annotationModule = annotationModule;

	/**
	 * Set the annotation color of the note annotation to draw
	 */
	this.setAnnotationColor = function(color) {
		f.getCurrentUser().setColor("PREFERENCES_ANNOTATION_NOTE_COLOR", color);
	};

	/**
	 * Get the annotation color of the note annotation to draw
	 */
	this.getAnnotationColor = function() {
		if (f.getProperty("annotationcolor")) {
			return f.getProperty("annotationcolor");
		} else {
			return f.getCurrentUser().getColor("PREFERENCES_ANNOTATION_NOTE_COLOR");
		}
	};

	/**
	 * @override
	 */
	this.getDefaultCursor = function() {
		return "cursor-drawComment";
	};

	/**
	 * Get the annotation layer
	 */
	var getAnnotationLayer = _.bind(function(e) {
		var v = this._widget.requestView(e);
		if (v) {
			var l = v.getLayer("Annotations");
			if (l) {
				return l;
			}
		}
		return null;
	}, this);

	/**
	 * This is the mode when the tool is active,
	 * it will zoom on click and go the couldStartDraggingMode if left mousedown is clicked
	 */
	this.registerState(defaultState, {
		contextmenu: function(e) {
			// prevent the context menu
			e.preventDefault();
			e.stopPropagation();
			return false;
		},
		click: _.bind(function(e) {
			var l = getAnnotationLayer(e);
			if (l) {
				//first get the current touchpoint, the point where was clicked
				var touchPoint = this.keepPointWithinWorldBounds(this.getWorldPoint(e), true);
				var newAnnotRect = new ShapeUtils.Rect(touchPoint, new ShapeUtils.Size()); //make a rect out of it

				//create the new annotation
				var newAnnotation = new AnnotationFactory(f).create();
				newAnnotation.initAnnotation(AnnotationConstants.AnnotationType.ANNOT_PIN, this.getAnnotationColor(), false, annotationModel);
				//make sure the rectangle is in world coordinates and rounded so detection of inside active page is correct
				newAnnotRect = l.getWorldRectFromScene(newAnnotRect);
				newAnnotRect.origin.x = ViewerUtils.parseFloat(newAnnotRect.origin.x.toFixed(2));
				newAnnotRect.origin.y = ViewerUtils.parseFloat(newAnnotRect.origin.y.toFixed(2));
				newAnnotation.rect = newAnnotRect;

				if (this._annotationModel.getPage().isReaderSpread()) {
					this._annotationModule.promptIfAnnotationOutsidePage(
						newAnnotation.rect,
						_.bind(function() {
							this.addAnnotation(newAnnotation);
						}, this),
						function() {
							newAnnotation = null;
						}
					);
				} else {
					this.addAnnotation(newAnnotation);
				}
			}
		}, this)
	});

	/**
	 * @override
	 */
	this.canBeInterrupted = function() {
		return true;
	};

	this.activeKeyEvents = [];
	this.persistentKeyEvents = [];
	/**
	 * @override
	 */
	this._override("initialize", function(widget, toolManager) {
		this._super(widget, toolManager);
		this.changeState(defaultState);
		this.registerActiveEvents(["contextmenu", "click"]);
		this.activeKeyEvents.push([
			"shift",
			_.bind(function(e) {
				this.drawLinkedAnnotation = true;
			}, this),
			"keydown"
		]);

		this.activeKeyEvents.push([
			"shift",
			_.bind(function(e) {
				this.drawLinkedAnnotation = false;
			}, this),
			"keyup"
		]);
	});

	/**
	 * @override
	 */
	this._override("activate", function(interruptRequest) {
		this._super(interruptRequest);
		annotationModel.setLockSelection(true);
	});

	/**
	 * @override
	 */
	this.canBeActivated = function() {
		return annotationModel.canChangeAnnotations();
	};

	/**
	 * @override
	 */
	this._override("deactivate", function(interruptRequest) {
		this._super(interruptRequest);
		this.drawLinkedAnnotation = false;
		annotationModel.setLockSelection(false);
	});

	f.getCurrentUser().on(
		"options.changed.PREFERENCES_ANNOTATIONTOOLS_VISIBLE",
		_.bind(function(oldValue, newValue) {
			if (!newValue && this.isActive()) {
				this.requestDeactivation();
			}
		}, this)
	);
};

// inherit
AnnotationNoteDrawTool.prototype = new AbstractTool();
// correct the constructor pointer
AnnotationNoteDrawTool.prototype.constructor = AnnotationNoteDrawTool;

/**
 * Adds a created annotation to the annotationmodel
 * @param {Annotation} the new annotation to be added
 */
AnnotationNoteDrawTool.prototype.addAnnotation = function(newAnnotation) {
	if (this.drawLinkedAnnotation) {
		this._annotationModel.addLinkedAnnotation(newAnnotation);
	} else {
		this._annotationModel.addAnnotation(newAnnotation);
		this.requestDeactivation();
	}
};

export default AnnotationNoteDrawTool;
