import $ from "jquery";
import _ from "underscore";
import ButtonGroup from "@esko/toolbar/ButtonGroup";
import ToolbarButton from "@esko/toolbar/ToolbarButton";
import AnnotationConstants from "@esko/web-viewer/annotation/AnnotationConstants";

/**
 * @class
 * @exports AnnotationOriginSelector
 * @param {Core.ViewerFacade} f
 * @param {AnnotationFilterModel2} filterModel
 * @param {AnnotationModel} annotationModel
 */
var AnnotationOriginSelector = function(f, filterModel, annotationModel, context, isMobile) {
	/**
	 * @member {AnnotationFilterModel2}
	 * @private
	 */
	this._filterModel = filterModel;

	/**
	 * @member {AnnotationModel}
	 * @private
	 */
	this._annotationModel = annotationModel;

	this._annotationModel.on("modelChanged", _.bind(this._enableToolbarButtons, this));
	this._annotationModel.on("modelContentChanged", _.bind(this._enableToolbarButtons, this));
	this._annotationModel.on("modelContentChanged", _.bind(this._activateToolbarButtons, this));
	this._filterModel.on("filterChanged", _.bind(this._activateToolbarButtons, this));

	/**
	 * @member {Core.ViewerFacade}
	 * @private
	 */
	this.f = f;

	/**
	 * @member {ButtonGroup}
	 * @private
	 */
	this._buttonGroup = new ButtonGroup("annotationTypeButtonGroup");

	/**
	 * @member {JQueryElement}
	 * @private
	 */
	this.element = $("<div class='annotation-origin-selector'/>").append(this._buttonGroup.element);

	/**
	 * @member {JQueryElement}
	 * @private
	 */
	this._context = context;

	/**
	 * @member {Boolean}
	 * @private
	 */
	this._isMobile = isMobile;

	/**
	 * @member {Object} id -> ToolbarButton
	 * @private
	 */
	this._buttons = {};

	this._initTypeFilters();
};

/**
 * creates the origin buttons from the annotation types defined in AnnotationConstants.AnnoationTypes
 */
AnnotationOriginSelector.prototype._initTypeFilters = function() {
	var annotationTypes = AnnotationConstants.AnnotationOriginTypes;
	_.each(
		annotationTypes,
		function(annotationType) {
			var id = annotationType.id;

			var annotationIcon = annotationType.icon;
			if (annotationType.subIcon) {
				annotationIcon = $("<i class='icon containerIcon icon-normal'/>").addClass(annotationIcon);
				annotationIcon.append($("<i class='icon subIcon icon-normal'/>").addClass(annotationType.subIcon));
			}

			var button = new ToolbarButton(
				this.f.i18n,
				this._context,
				this._isMobile,
				id,
				annotationIcon,
				this._isMobile,
				_.bind(function() {
					this._filterModel.setOriginType(id);
					this._annotationModel.unselectAllAnnotations();
				}, this),
				this.f.i18n.t(annotationType.langKey)
			);
			this._buttonGroup.addButton(button);
			this._buttons[id] = button;
		},
		this
	);
	this._enableToolbarButtons();
	this._activateToolbarButtons();
};

/**
 * enables the buttons depending on the availability of the types in the annotation model
 */
AnnotationOriginSelector.prototype._enableToolbarButtons = function() {
	var annotationsTypes = _.uniq(
		_.map(this._annotationModel.getAnnotations(), function(annotation) {
			return annotation.originType;
		})
	);
	var annotationGroups = _.uniq(
		_.map(annotationsTypes, function(type) {
			return AnnotationConstants.AnnotationOriginTypes[type].group;
		})
	);
	_.each(this._buttons, function(button, id) {
		var annotationsOfIdExist = _.find(annotationsTypes, function(type) {
			return id === type;
		});
		var annotationsOfGroupExist = _.find(annotationGroups, function(group) {
			return AnnotationConstants.AnnotationOriginTypes[id].group === group;
		});
		button.setEnabled(annotationsOfIdExist || id === AnnotationConstants.AnnotationOriginTypes.user.id);
		var originType = AnnotationConstants.AnnotationOriginTypes[id];
		if (originType && originType.hideBtnOnEmpty) {
			if (annotationsOfIdExist) {
				button.show();
			} else {
				button.hide();
			}
		} else if (originType && originType.hideBtnOnGroupEmpty) {
			if (annotationsOfGroupExist) {
				button.show();
			} else {
				button.hide();
			}
		}
	});
};

/**
 * activates the buttons depending on the filter model
 */
AnnotationOriginSelector.prototype._activateToolbarButtons = function() {
	var id = this._filterModel.getOriginType();
	_.each(this._buttons, function(button, i) {
		button.setActive(button.name === id);
	});
};

export default AnnotationOriginSelector;
