import AnnotationConstants from "@esko/web-viewer/annotation/AnnotationConstants";
import Popup from "@esko/web-viewer/layout/component/Popup";
import MiniButton from "@esko/web-viewer/layout/component/MiniButton";
import $ from "jquery";
import _ from "underscore";
/**
 * Create an annotation popup.
 * Initial (re)paint should be called by caller
 *
 * @exports AnnotationPopup
 * @memberof AnnotationModule
 * @class
 * The annotation popup controls the popup for annotation details
 * @constructor
 * @augments Layout.Popup
 * @example
 * Emitted events:
 * 		"pinned", {Boolean} - Fired when the popup is pinned/unpinned. The argument indicates whether the new state is pinned (true) or unpinned (false).
 * @param {Core.ViewerFacade} f - Viewerfacade
 * @param {String} id - Id of the popup used for its private channel
 * @param {AnnotationModule.Annotation} annotation - Annotation this popup belongs to
 * @param {JQueryObj} containment - The containment where the popup should stay in
 * @param {AnnotationModel} annotationModel - The model that holds all annotations
 */
var AnnotationPopup = function(f, id, annotation, containment, annotationModel, container) {
	this.f = f;
	Popup.call(this, f, id, containment, { container: container });

	this.annotation = annotation;
	this.element.addClass("annotationPopup");

	// Annotation number
	this.annotationNumber = $("<span />").addClass("annotationNumber");
	this.annotationNumber.get(0).dataset.sahiId = "annotationNumber";

	// Annotation icon
	this.icon = $("<i />");

	// Title
	this.title = $("<span />").addClass("annotationTitle");
	this.title.get(0).dataset.sahiId = "annotationTitle";

	this.annotationModel = annotationModel;
};

AnnotationPopup.prototype = Object.create(Popup.prototype);
AnnotationPopup.prototype.constructor = AnnotationPopup;

/**
 * Initialize the popup, create needed DOM events, listeners
 *
 * @private
 * @override
 */
AnnotationPopup.prototype._initialize = function() {
	if (!this._isInitialized) {
		Popup.prototype._initialize.call(this);

		// Styling
		this.headerWrapper.addClass("annotationPopupHeader");

		this.header.append(this.annotationNumber);
		this.header.append(this.icon);
		this.header.append(this.title);

		// Pin Icon
		this.pinIcon = new MiniButton(
			this.f,
			"pin",
			"pinned icon-WCR_Pin12",
			true,
			_.bind(function(e) {
				// Pin if unpinned
				this._emit("pinned", !this.annotation.popupIsOpen);
				e.stopPropagation();
			}, this)
		);
		this.headerIcons.prepend(this.pinIcon.element);

		this.annotationModel.on("showAllAnnotationPopupsChanged", this._toggleCloseButton.bind(this));
		this._toggleCloseButton();

		// Initial paint
		AnnotationPopup.prototype.repaint.call(this);
	}
};

/**
 * Repaint the popup.
 * The pin icon is repainted.
 * The header is repainted (annotation number, title, annotation type icon).
 */
AnnotationPopup.prototype.repaint = function() {
	if (this._isInitialized) {
		// Pin icon
		if (this.annotation.popupIsOpen) {
			this.pinIcon.setIcon("pinned icon-WCR_Pin12");
			this.pinIcon.setTooltip("viewer", "UnpinPopupTT");
		} else {
			this.pinIcon.setIcon("unpinned icon-WCR_Pin12");
			this.pinIcon.setTooltip("viewer", "PinPopupTT");
		}

		// Set color
		this.setAnnotationNumber(this.annotation.messagenumber);
		this.annotationNumber.css({
			color: this.annotation.getHeaderTextColor(),
			"background-color": this.annotation.annotationcolor
		});

		// Title
		this.setTitle(this.annotation.getUserName());

		// Icon
		this.setIcon(AnnotationConstants.AnnotationIcons[this.annotation.annottype]);
	}
};

/**
 * Set the title in the header
 * @param {String} title - Title to set in the header of the popup
 */
AnnotationPopup.prototype.setTitle = function(title) {
	this.title.text(title);
};

/**
 * Remove this popup
 * @override
 */
AnnotationPopup.prototype.remove = function() {
	this.removeAllListeners("pinned");
	Popup.prototype.remove.call(this);
};

/**
 * Set the icon (annotation type) in the header
 * @param {String} icon - Icon (css class) to set in the header of the popup
 */
AnnotationPopup.prototype.setIcon = function(icon) {
	this.icon
		.removeClass()
		.addClass("icon")
		.addClass(icon);
};

/**
 * Set the annotation number in the header
 * @param {String} number - Annotation number to set in the header of the popup
 */
AnnotationPopup.prototype.setAnnotationNumber = function(number) {
	this.annotationNumber.text(number);
};

/**
 * Hide/Show the close button
 * @private
 */
AnnotationPopup.prototype._toggleCloseButton = function() {
	if (this.annotationModel.areAllAnnotationPopupsShown()) {
		this.close.element.hide();
	} else {
		this.close.element.show();
	}
};

export default AnnotationPopup;
