import ViewportPopupModel from "@esko/web-viewer/viewport/model/ViewportPopupModel";
import AnnotationPopupNew from "@esko/web-viewer/annotation/AnnotationPopupNew";
import AnnotationPopupExisting from "@esko/web-viewer/annotation/AnnotationPopupExisting";
import AnnotationConstants from "@esko/web-viewer/annotation/AnnotationConstants";
import ShapeUtils from "@esko/shapeutils";
import _ from "underscore";
/**
 * Create a new Annotation Popup Model.
 * @exports AnnotationPopupModel
 * @memberof AnnotationModule
 * @class
 * The annotation popup model is responsible for an annotation popup and its content.
 * It binds the events of the popup and reacts on them.
 * This popup model is the model for both a new annotation and an existing one.
 * It will create the appropriate popup: AnnotationPopupNew if it's a new or AnnotationPopupExisting if it's an existing annotation.
 * @augments Viewport.ViewportPopupModel
 * @constructor
 * @param {Core.ViewerFacade} f - Viewerfacade
 * @param {String} id - Id of the popup model.
 * @param {Annotation} annotation - Annotation this model belongs to.
 * @param {AnnotationModel} annotationModel - Model of the annotations.
 */
var AnnotationPopupModel = function(f, id, annotation, annotationModel) {
	ViewportPopupModel.call(this, f, id);

	this.annotation = annotation;
	this.annotationModel = annotationModel;

	this.popup = null;
};

AnnotationPopupModel.prototype = Object.create(ViewportPopupModel.prototype);
AnnotationPopupModel.prototype.constructor = AnnotationPopupModel;

/**
 * Repaint the popup
 */
AnnotationPopupModel.prototype.repaint = function(refannotation) {
	if (this.popup != null) {
		this.popup.repaint(refannotation);
	}
};

/**
 * Get the Annotation popup
 * @param {ViewportLayer} viewportLayer - The layer where the popups should be shown.
 * @returns {AnnotationPopupNew | AnnotationPopupExisting} The annotation popup of this model.
 */
AnnotationPopupModel.prototype.getPopup = function(viewportLayer) {
	var containment = viewportLayer.getLayerElement();
	var container = this.f
		.getLayoutManager()
		.getLayout()
		.getMainContentElement();
	if (this.popup === null) {
		this._createPopup(containment, container);
	} else {
		this.popup.setContainment(containment, container);
	}
	return this.popup;
};

/**
 * Create an annotation popup {AnnotationPopupNew | AnnotationPopupExisting} for the given layer
 * and bind all events.
 * @param {JQueryObj} container - Container for the popup
 */
AnnotationPopupModel.prototype._createPopup = function(contaiment, container) {
	if (this.annotation.isNew) {
		this.popup = new AnnotationPopupNew(this.f, this.id, this.annotation, this.annotationModel, contaiment, container);
	} else {
		this.popup = new AnnotationPopupExisting(this.f, this.id, this.annotation, this.annotationModel, contaiment, container);
	}
	// we want the resized in world rect instead of scene rect, so don't bind them here, but in the layer, and let the layer call the model functions
	this.popup.on("closed", _.bind(this.onClosed, this));
	this.popup.on("pinned", _.bind(this.onPinned, this));
	this.popup.on("clicked", _.bind(this.onClicked, this));
};

/**
 * Method called when the model is destroyed.
 * Useful for some cleanup in the popup/content
 */
AnnotationPopupModel.prototype.onDestroy = function() {
	if (this.popup != null) {
		this.popup.remove();
	}
};

/**
 * @override
 * This method saves the rectangle in the annotation.
 * @param {ShapeUtils.Rect} rect - Rectangle (origin/size) of the popup in world coordinates
 * @param {Boolean} autoAdapted - True if the event is fired because the content was auto adapted (e.g. from a makeMeVisible trigger from an element in the content), false otherwise
 */
AnnotationPopupModel.prototype.onResized = function(worldRect, autoAdapted) {
	var copy = worldRect.copy();
	// compensate for origin in left bottom corner
	copy.origin.y -= copy.size.height;

	if (!this.annotation.popuprect.rectEquals(copy)) {
		this.annotation.popuprect = copy;

		if (!autoAdapted) {
			this._fireAnnotationChanged("POPUP_SIZE");
		}
	}
};

/**
 * @override
 * This method saves the origin in the annotation.
 * @param {ShapeUtils.Point} point - Point (origin) of the popup in world coordinates
 */
AnnotationPopupModel.prototype.onDragged = function(worldPoint) {
	var point = worldPoint.copy();
	point.y -= this.annotation.popuprect.size.height;

	if (!this.annotation.popuprect.origin.equals(point)) {
		this.annotation.popuprect.origin = point;
		this._fireAnnotationChanged("POPUP_POSITION");
	}
};

/**
 * This function is bind to the "closed" popup event and called when the popup is closed.
 */
AnnotationPopupModel.prototype.onClosed = function() {
	this.annotationModel.unselectAllAnnotations(true, true);

	if (this.annotation.popupIsOpen) {
		this.annotation.popupIsOpen = false;
		this.annotation.setModified("POPUP_PINNED");
	}
	this.annotationModel.modelChanged(this.annotation);
};

/**
 * This function is bind to the "pinned" annotation popup event and called when the popup is pinned/unpinned.
 */
AnnotationPopupModel.prototype.onPinned = function(pinned) {
	if (this.annotation.popupIsOpen !== pinned) {
		this.annotation.popupIsOpen = pinned;
		if (!this._fireAnnotationChanged("POPUP_PINNED")) {
			// trigger repaint for pinned icon if we cannot change annotations
			this.popup.repaint();
		}
	}
};

/**
 * This function is bind to the "clicked" annotation popup event and called when the popup is clicked.
 */
AnnotationPopupModel.prototype.onClicked = function() {
	this.annotationModel.selectAnnotation(this.annotation);
};

/**
 * Fire that an annotation has changed
 * @param {String} what - indicate what was changed
 * @returns {Boolean} True when fired, false otherwise
 */
AnnotationPopupModel.prototype._fireAnnotationChanged = function(what) {
	if (this.annotationModel.canChangeAnnotation(this.annotation)) {
		this.annotation.setModified(what);
		this.annotationModel.modelChanged(this.annotation);
		return true;
	} else {
		return false;
	}
};

/**
 * Get the origin of the popup in world coordinates
 * @returns {ShapeUtils.Point} - The origin of the popup in world coordinates
 */
AnnotationPopupModel.prototype.getWorldOrigin = function() {
	// origin in left bottom corner
	return new ShapeUtils.Point(this.annotation.popuprect.left(), this.annotation.popuprect.bottom());
};

/**
 * Get the size of the popup in world coordinates
 * @returns {ShapeUtils.Size} - The size of the popup in world coordinates
 */
AnnotationPopupModel.prototype.getWorldSize = function() {
	return this.annotation.popuprect.size;
};

/**
 * When the popup is for a pinned annotation, add an offset so the popup doesn't show above the icon
 * @returns {ShapeUtils.Size}
 */
AnnotationPopupModel.prototype.getSceneOffset = function() {
	if (this.annotation.getType() == AnnotationConstants.AnnotationType.ANNOT_PIN) {
		return new ShapeUtils.Size(AnnotationConstants.AnnotationPin.imagewidth, 0);
	}
};

/**
 * Override ViewportPopupModel isShown.
 * Determines visibility of annotation popup based on the visibility settings in the annotation widget.
 * @returns {boolean} True if it is visible, false otherwise
 */
AnnotationPopupModel.prototype.isShown = function() {
	var visibleAnnotations = this.annotationModel.getVisibleAnnotations();
	if (visibleAnnotations.length > 0 && visibleAnnotations.indexOf(this.annotation) !== -1) {
		return this.annotationModel.areAllAnnotationPopupsShown() || this.shown;
	}
	return false;
};

export default AnnotationPopupModel;
