import AnnotationPopup from "@esko/web-viewer/annotation/AnnotationPopup";
import ColorPicker from "@esko/web-viewer/layout/component/ColorPicker";
import ViewerToolbarButton from "@esko/web-viewer/layout/component/ViewerToolbarButton";
import ButtonGroup from "@esko/toolbar/ButtonGroup";
import RichTextEditorFactory from "@esko/web-viewer/layout/component/RichTextEditorFactory";
import AnnotationCharLineCount from "@esko/web-viewer/annotation/AnnotationCharLineCount";
import AnnotationRow from "@esko/web-viewer/annotation/AnnotationRow";
import AnnotationTimeRow from "@esko/web-viewer/annotation/AnnotationTimeRow";
import _ from "underscore";
/**
 * Create a new New Annotation Popup.
 * @exports AnnotationPopupNew
 * @memberof AnnotationModule
 * @class
 * This annotation popup is responsible for a new annotation popup and its content (the rich text editor).
 * @augments AnnotationModule.AnnotationPopup
 * @constructor
 * @param {Core.ViewerFacade} f - Viewerfacade
 * @param {String} id - Id of the popup.
 * @param {Annotation} annotation - Annotation this popup belongs to.
 * @param {AnnotationModel} annotationModel - Model of the annotations.
 * @param {JQueryObj} contaiment - The contaiment where the popup should stay in
 */
var AnnotationPopupNew = function(f, id, annotation, annotationModel, contaiment, container) {
	AnnotationPopup.call(this, f, id, annotation, contaiment, annotationModel, container);
	this.element.removeClass("annotationPopup").addClass("annotationPopupNew");
	// Default Size (in world size)
	if (!this.annotation.popuprect.size.width) {
		this.annotation.popuprect.size.width = 89; // min size for tinymce to show toolbar and timestamps on one line
	}
	if (!this.annotation.popuprect.size.height) {
		this.annotation.popuprect.size.height = 45;
		if (this.annotation.startTime != null || this.annotation.endTime != null) {
			this.annotation.popuprect.size.height = 65;
		}
	}

	// Default Location (top right corner of annotation rect)
	if (!this.annotation.popuprect.origin.x) {
		this.annotation.popuprect.origin.x = this.annotation.rect.right();
	}
	if (!this.annotation.popuprect.origin.y) {
		this.annotation.popuprect.origin.y = this.annotation.rect.bottom() - this.annotation.popuprect.size.height;
	}
};

AnnotationPopupNew.prototype = Object.create(AnnotationPopup.prototype);
AnnotationPopupNew.prototype.constructor = AnnotationPopupNew;

/**
 * @override
 */
AnnotationPopupNew.prototype._initialize = function() {
	// Super call
	AnnotationPopup.prototype._initialize.call(this);

	// Rich text editor
	this.richTextEditor = RichTextEditorFactory.createEditor(this.f);
	this.richTextEditor.setContent(this.annotation.content);

	var container = this.f
		.getLayoutManager()
		.getLayout()
		.getMainContentElement();

	// Show a color picker if no annotation color is set
	var annotationColorPicker;
	if (!this.f.getProperty("annotationcolor")) {
		annotationColorPicker = new ColorPicker(
			this.f,
			_.bind(function(newColor) {
				this.annotation.setColor("#" + newColor);
				this.annotationModel.modelChanged(this.annotation);
			}, this),
			undefined,
			{ appendTo: container }
		);
		annotationColorPicker.setColor(this.annotation.annotationcolor);
		annotationColorPicker.element.addClass("annotationColorPicker");
		annotationColorPicker.setTooltip(this.f, "viewer", "ChangeColor");
	}

	// Save button
	var saveAnnot = new ViewerToolbarButton(
		this.f,
		"saveAnnot",
		"icon-WCR_Save12 marginRight5",
		false,
		_.bind(function(e) {
			this.annotation.content = this.richTextEditor.getContent();
			this.annotationModel.save();
			e.stopPropagation();
		}, this)
	);
	// Delete button
	var deleteAnnot = new ViewerToolbarButton(
		this.f,
		"deleteAnnot",
		"icon-WCR_Trash12 marginRight5",
		false,
		_.bind(function(e) {
			this.annotationModel.deleteAnnotation(this.annotation);
			e.stopPropagation();
		}, this)
	);
	// Save and delete are a buttongroup
	var annotationActionsGroup = new ButtonGroup("AnnotationActionsGroup", [deleteAnnot, saveAnnot]);
	annotationActionsGroup.element.addClass("pull-right");

	// Set editor/button texts
	this.f.i18n.loadBundle(
		"viewer",
		_.bind(function() {
			saveAnnot.element
				.addClass("btn-mini")
				.addClass("btn-success")
				.append(this.f.i18n.t("Save"));
			deleteAnnot.element.addClass("btn-mini").append(this.f.i18n.t("Delete"));
		}, this)
	);

	// Character/line count
	this.charLineCount = new AnnotationCharLineCount(this.f);

	var buttons = [annotationColorPicker, annotationActionsGroup, this.charLineCount];

	// Time stamps for video
	if (this.annotation.startTime != null || this.annotation.endTime != null) {
		var timestamps = new AnnotationTimeRow(this.f, this.annotation, this.annotationModel, true);
		buttons.unshift({ element: timestamps.getElement() });
	}

	// Set the buttons and char/line count in the editor
	this.richTextEditor.setButtons(buttons);

	// Update count on text change
	this.richTextEditor.onTextChanged(
		_.bind(function() {
			this.charLineCount.update(this.richTextEditor.getContent());
			this.annotation.setBeingModified(this.richTextEditor.getContent());
		}, this)
	);

	// Set the content to the rich text editor
	this.setContent(this.richTextEditor.element);

	// Init rich text
	this.richTextEditor.init(
		_.bind(function() {
			this._repaintCharLineCount();

			// Resize the editor along with the popup
			this.on(
				"contentResized",
				_.bind(function(size) {
					this.richTextEditor.setHeight(size.height);
				}, this)
			);
			this.richTextEditor.setHeight(this._outerContent.height() || null);
		}, this)
	);
};

/**
 * @override
 */
AnnotationPopupNew.prototype.repaint = function() {
	if (this._isInitialized) {
		// Super call
		AnnotationPopup.prototype.repaint.call(this);
		this._repaintCharLineCount();
	}
};

/**
 * Repaint the char/line count
 */
AnnotationPopupNew.prototype._repaintCharLineCount = function() {
	if (this.f.getOption("PREFERENCES_ANNOTATION_SHOWCOUNT")) {
		this.charLineCount.show();
		this.charLineCount.update(this.richTextEditor.getContent());
	} else {
		this.charLineCount.hide();
	}
};

/**
 * @override
 */
AnnotationPopupNew.prototype.remove = function() {
	if (this.annotationColorPicker) {
		this.annotationColorPicker.destroy();
	}
	if (this.richTextEditor) {
		this.richTextEditor.destroy();
	}
	// Super call
	AnnotationPopup.prototype.remove.call(this);
};

export default AnnotationPopupNew;
