import ViewportLayerModel from "@esko/web-viewer/viewport/model/ViewportLayerModel";
import AnnotationPopupModel from "@esko/web-viewer/annotation/AnnotationPopupModel";
import _ from "underscore";
/**
 * @exports AnnotationPopupsLayerModel
 * @memberof AnnotationModule
 * @class
 * Create a new annotation layer model for annotation popups
 * @constructor
 * @param {Core.ViewerFacade} f - Viewerfacade
 * @param {String} ID - Id of the layer
 * @param {AnnotationModel} annotationModel - Model of the annotations.
 */
var AnnotationPopupsLayerModel = function(f, ID, annotationModel) {
	ViewportLayerModel.call(this, f, ID);

	this.f = f;
	this.annotationModel = annotationModel;

	// New annotations received
	this._createModels();
	this.annotationModel.on(
		"modelContentChanged",
		_.bind(function() {
			this.removeAllPopupModels();
			this._createModels();
		}, this)
	);

	// Annotation model changed events
	this.annotationModel.on(
		"modelChanged",
		_.bind(function() {
			var page = this.annotationModel.getAnnotationsPage();
			if (page !== null && page !== undefined) {
				this.resetTransformation();
				this.getTransform().scale(1, -1);

				var rotation = page.getRotation();

				var worldBounds = this.f.getModule("viewport").getWorldBounds();

				this.calculateTransformation(
					worldBounds,
					page.getPageSize(),
					this.annotationModel.getAnnotationsOffset(),
					rotation,
					page.isFlipped()
				);
			}
			this.modelChanged();
		}, this)
	);

	// Annotation selected
	this.annotationModel.on(
		"selectedAnnotation",
		_.bind(function(annotation) {
			// if show count, repaint the annotation popup when selected to draw the char/line count
			if (this.f.getOption("PREFERENCES_ANNOTATION_SHOWCOUNT")) {
				var mainAnnotation = annotation.getGodfather();
				var popupModel = this.getPopupModel(mainAnnotation.name);
				if (popupModel !== undefined) {
					if (mainAnnotation == annotation) {
						// the main annotation is selected
						popupModel.repaint();
					} else {
						// a reply is selected
						popupModel.repaint(annotation);
					}
				}
			}
			this._updateModels();
			this.modelChanged();
		}, this)
	);

	// Annotation unselected
	this.annotationModel.on(
		"unselectedAnnotation",
		_.bind(function(annotation) {
			// if show count, repaint the annotation popup when selected to hide the char/line count
			if (this.f.getOption("PREFERENCES_ANNOTATION_SHOWCOUNT")) {
				var popupModel = this.getPopupModel(annotation.getGodfather().name);
				if (popupModel !== undefined) {
					popupModel.repaint();
				}
			}
			this._updateModels();
		}, this)
	);

	// Annotation deleted
	this.annotationModel.on(
		"annotationDeleted",
		_.bind(function(annotation) {
			var mainAnnotation = annotation.getGodfather();
			if (annotation == mainAnnotation) {
				// deleted main annotation
				this.removePopupModel(annotation.name);
			} else {
				// deleted a sub annotation, so just repaint
				var popupModel = this.getPopupModel(mainAnnotation.name);
				if (popupModel !== undefined) {
					popupModel.repaint();
				}
			}
		}, this)
	);

	// Annotation added
	this.annotationModel.on(
		"annotationAdded",
		_.bind(function(annotation) {
			this._createModel(annotation);
		}, this)
	);

	// Annotation updated
	this.annotationModel.on(
		"annotationUpdated",
		_.bind(function(annotation, refannotation) {
			var popupModel = this.getPopupModel(annotation.getGodfather().name);
			if (popupModel !== undefined) {
				popupModel.repaint(refannotation);
			}
			this._updateModels();
		}, this)
	);

	// Annotation list visible
	this.annotationModel.on(
		"annotationListVisible",
		_.bind(function(annotationListVisible) {
			this._updateModels();
			this.modelChanged();
		}, this)
	);

	// Read/write changed
	this.annotationModel.on(
		"readWriteChanged",
		_.bind(function() {
			_.each(
				this.getPopupModels(),
				function(popupModel) {
					popupModel.repaint();
				},
				this
			);
			this.modelChanged();
		}, this)
	);

	// Annotations disabled
	this.annotationModel.on(
		"temporarilyDisableChanged",
		_.bind(function() {
			this._updateModels();
			//this.modelChanged(); already fired by the annotation model (avoid flickering)
		}, this)
	);

	// Triggered when the setting "Show All Annotation Popups" has changed
	this.annotationModel.on(
		"showAllAnnotationPopupsChanged",
		_.bind(function() {
			this.modelChanged();
		}, this)
	);

	// Show selected annotation only
	this.f.getCurrentUser().on(
		"options.changed.PREFERENCES_ANNOTATION_SHOWSELECTEDONLY",
		_.bind(function(oldValue, newValue) {
			this._updateModels();
			this.modelChanged();
		}, this)
	);

	// Show selected annotation popup only
	this.f.getCurrentUser().on(
		"options.changed.PREFERENCES_ANNOTATION_SHOWSELECTEDPOPUPONLY",
		_.bind(function(oldValue, newValue) {
			this._updateModels();
			this.modelChanged();
		}, this)
	);

	// Show annotation char/line count
	this.f.getCurrentUser().on(
		"options.changed.PREFERENCES_ANNOTATION_SHOWCOUNT",
		_.bind(function(oldValue, newValue) {
			_.each(
				this.getPopupModels(),
				function(popupModel) {
					var annotation = this.annotationModel.getAnnotation(popupModel.getId());
					if (this.annotationModel.isAnnotationOrReplySelected(annotation)) {
						popupModel.repaint();
					}
				},
				this
			);
			this.modelChanged();
		}, this)
	);

	// Annotations visible
	this.f.getCurrentUser().on(
		"options.changed.PREFERENCES_ANNOTATIONTOOLS_VISIBLE",
		_.bind(function(oldValue, newValue) {
			this._updateModels();
			//this.modelChanged(); already fired by the annotation model (avoid flickering)
		}, this)
	);
};

AnnotationPopupsLayerModel.prototype = Object.create(ViewportLayerModel.prototype);
AnnotationPopupsLayerModel.prototype.constructor = AnnotationPopupsLayerModel;

/**
 * Update the state of the models based on the options.
 * Does the show/hide logic for annotation popups
 */
AnnotationPopupsLayerModel.prototype._updateModels = function() {
	if (this.f.getOption("PREFERENCES_ANNOTATIONTOOLS_VISIBLE")) {
		// Annotations visible
		if (this.annotationModel.isAnnotationListVisible()) {
			// options apply
			if (this.f.getOption("PREFERENCES_ANNOTATION_SHOWSELECTEDONLY")) {
				if (this.annotationModel.getSelectedAnnotations().length > 0) {
					// there are selected annotations, so show only the selected ones
					_.each(
						this.getPopupModels(),
						function(popupModel) {
							this._showSelected(popupModel);
						},
						this
					);
				} else {
					// there are no selected annotations, show the pinned ones
					_.each(
						this.getPopupModels(),
						function(popupModel) {
							this._showPinnedOrSelected(popupModel);
						},
						this
					);
				}
			} else {
				// else because it can only apply if PREFERENCES_ANNOTATION_SHOWSELECTEDONLY is turned off
				if (this.f.getOption("PREFERENCES_ANNOTATION_SHOWSELECTEDPOPUPONLY")) {
					_.each(
						this.getPopupModels(),
						function(popupModel) {
							this._showSelected(popupModel);
						},
						this
					);
				} else {
					// show all popups that are pinned/selected if no (preference) options selected
					_.each(
						this.getPopupModels(),
						function(popupModel) {
							this._showPinnedOrSelected(popupModel);
						},
						this
					);
				}
			}
		} else {
			// options not valid if list not visible, just check pinned state (or selected)
			_.each(
				this.getPopupModels(),
				function(popupModel) {
					this._showPinnedOrSelected(popupModel);
				},
				this
			);
		}
	} else {
		// Annotations not visible, hide all popups
		_.each(
			this.getPopupModels(),
			function(popupModel) {
				popupModel.hide();
			},
			this
		);
	}
};

/**
 * Show the popup if it is selected
 * This method relies on the fact that there are no annotations selected if the annotations are disabled
 * @private
 * @param {AnnotationPopupModel} popupModel - The model of the popup to show if pinned or selected
 */
AnnotationPopupsLayerModel.prototype._showSelected = function(popupModel) {
	var annotation = this.annotationModel.getAnnotation(popupModel.getId());
	if (this.annotationModel.isAnnotationOrReplySelected(annotation)) {
		popupModel.show();
	} else {
		popupModel.hide();
	}
};

/**
 * Show the popup if it is pinned/selected and if the annotations are not disabled
 * @private
 * @param {AnnotationPopupModel} popupModel - The model of the popup to show if pinned or selected
 */
AnnotationPopupsLayerModel.prototype._showPinnedOrSelected = function(popupModel) {
	var annotation = this.annotationModel.getAnnotation(popupModel.getId());
	// if the annotations are not disabled and the annotation popup is open or selected, show it, else hide it
	if (
		!this.annotationModel.isTemporarilyDisabledAndNoAnnotationsVisible() &&
		annotation &&
		(annotation.popupIsOpen || this.annotationModel.isAnnotationOrReplySelected(annotation))
	) {
		popupModel.show();
	} else {
		popupModel.hide();
	}
};

/**
 * Create the annotation popup models
 */
AnnotationPopupsLayerModel.prototype._createModels = function() {
	// Create all models
	_.each(
		this.annotationModel.getAnnotations(),
		function(annotation) {
			this._createModel(annotation);
		},
		this
	);
	this._updateModels();
	this.modelChanged();
};

/**
 * Create the annotation popup model for the given annotation
 * @param {Annotation} annotation - Annotation to create the model of
 */
AnnotationPopupsLayerModel.prototype._createModel = function(annotation) {
	// If annotation is linked, the other one already exists,
	// so don't create a new model
	if (annotation.linkedto) {
		return;
	} else {
		// This will create the appropriate popup (AnnotationPopupNew vs AnnotationPopupExisting)
		var popupModel = new AnnotationPopupModel(this.f, annotation.name, annotation, this.annotationModel);
		this.addPopupModel(popupModel);
	}
};

export default AnnotationPopupsLayerModel;
