import AnnotationSingleRow from "@esko/web-viewer/annotation/AnnotationSingleRow";
import ViewerToolbarButton from "@esko/web-viewer/layout/component/ViewerToolbarButton";
import ButtonGroup from "@esko/toolbar/ButtonGroup";
import RichTextEditorFactory from "@esko/web-viewer/layout/component/RichTextEditorFactory";
import AnnotationCharLineCount from "@esko/web-viewer/annotation/AnnotationCharLineCount";
import AnnotationConstants from "@esko/web-viewer/annotation/AnnotationConstants";
import MiniButton from "@esko/web-viewer/layout/component/MiniButton";
import $ from "jquery";
import _ from "underscore";
/**
 * @exports AnnotationRow
 * @memberof AnnotationModule
 * @class An annotation div is an interface element which provides one line for an annotation
 * @constructor
 * @param {Annotation} annotation
 * @param {AnnotationModel} annotationModel
 * @param {Boolean} popup
 * @param {Number} replyLimit
 * @param {Boolean} isForMobile
 */
var AnnotationRow = function(f, annotation, annotationModel, popup, replyLimit, isForMobile) {
	this.f = f;
	this._replyLimit = replyLimit;
	if (!this._replyLimit) {
		this._replyLimit = 2;
	}

	this._annotation = annotation;
	this._annotationModel = annotationModel;

	this._limitReplyCount = true;
	this._shownReplyCount = 0;
	this._totalReplyCount = 0;
	this.name = this._annotation.name;
	this.element = $("<div>")
		.addClass("annotation")
		.toggleClass("cannot-reply", !this._annotation.getPermissionFor(AnnotationConstants.ActionPermissions.CAN_BE_REPLIED_TO));

	this.annotationRows = {};

	this.selected = false;

	this._isForMobile = isForMobile;

	/**
	 * Replies
	 */
	this._repliesDiv = $("<div/>").addClass("replies");

	/**
	 * More Replies
	 */
	this._moreReplies = $("<p/>").addClass("MoreReplies clearfix");
	this._moreRepliesLabel = $("<a/>");

	this._moreRepliesLabel.on(
		"click",
		_.bind(function(e) {
			this.toggleReplyLimit();
			e.stopPropagation();
		}, this)
	);
	this._moreReplies.append(this._moreRepliesLabel);
	this._moreReplies.hide();

	if (!popup) {
		this.title = $("<span/>").addClass("annotationNumber");
		this.element.append(this.title);
	}

	/**
	 * Reply input
	 */
	if (isForMobile) {
		var replyButton = new MiniButton(
			this.f,
			"annotationAddReply",
			"icon-WCR_Comment",
			true,
			_.bind(function(e) {
				this._showRichTextEditor(e);
			}, this)
		);
		replyButton.addButtonClass("annotAction");
		this._replyText = replyButton.element;
	} else {
		this._replyText = $("<input/>");
		this.f.i18n.loadBundle(
			"viewer",
			_.bind(function() {
				this._replyText.attr("placeholder", this.f.i18n.t("AnnReplyTT"));
			}, this)
		);
		// block click bubbling
		this._replyText.on("click", function(e) {
			e.preventDefault();
			e.stopPropagation();
		});
		this._replyText.on(
			"focus",
			_.bind(function(e) {
				this._showRichTextEditor(e);
			}, this)
		);
	}
	this._reply = $("<div/>")
		.append(this._replyText)
		.addClass("addReply");

	// Rich text
	this._richTextEditor = RichTextEditorFactory.createEditor(this.f);

	// Buttons in the editor
	// Save button
	this._saveReply = new ViewerToolbarButton(
		this.f,
		"saveReply",
		"icon-WCR_Save12 marginRight5",
		true,
		_.bind(function(e) {
			this._annotationModel.replyToAnnotation(this._annotation, this._richTextEditor.getContent(), true);
			this._closeRichText();
			e.stopPropagation();
		}, this)
	);
	// Cancel button
	this._cancelReply = new ViewerToolbarButton(
		this.f,
		"cancelReply",
		"icon-WCR_Revert12 marginRight5",
		true,
		_.bind(function(e) {
			this._closeRichText();
			this._richTextEditor.setContent("");
			this._annotation.resetUnsavedReply();
			this._annotationModel.modelChanged();
			e.stopPropagation();
		}, this)
	);
	// Save and cancel are a buttongroup
	this._replyActionsGroup = new ButtonGroup("replyActionsGroup", [this._cancelReply, this._saveReply]);
	this._replyActionsGroup.element.addClass("pull-right");

	// Set editor/button texts
	f.i18n.loadBundle(
		"viewer",
		_.bind(function() {
			this._saveReply.element
				.addClass("btn-mini")
				.addClass("btn-success")
				.append(this.f.i18n.t("Save"));
			this._cancelReply.element.addClass("btn-mini").append(this.f.i18n.t("CancelButton"));
		}, this)
	);

	// Character/line count in reply
	this._charLineCount = new AnnotationCharLineCount(f);
	this.showHideCharLineCount(this.f.getOption("PREFERENCES_ANNOTATION_SHOWCOUNT"));
	// On text change update charline count
	this._richTextEditor.onTextChanged(
		_.bind(function() {
			this._charLineCount.update(this._richTextEditor.getContent());
			this._annotation.setUnsavedReply(this._richTextEditor.getContent());
			this._annotationModel.modelChanged();
		}, this)
	);

	// Set the buttons
	this._richTextEditor.setButtons([this._replyActionsGroup, this._charLineCount]);

	/**
	 * Order of gui elements
	 */
	this._threadDiv = $("<div>").addClass("viewer-annotation-popup-thread");
	this.element.append(this._threadDiv);
	// if popup, don't show header
	this._mainRow = new AnnotationSingleRow(this.f, this._annotation, true, this._annotationModel, popup, this._replyText, this._isForMobile);
	this._threadDiv.append(this._mainRow.element);
	this.annotationRows[this._annotation.name] = this._mainRow;

	this._threadDiv.append(this._repliesDiv);
	this._repliesDiv.prepend(this._moreReplies);

	if (!isForMobile) {
		this.element.append(this._reply);
	}

	this.annotationUpdated();
	this.updateReplyLimitLabel();
};

/**
 * Close rich text editor and replace with reply
 */
AnnotationRow.prototype._closeRichText = function() {
	// on cancel, show reply
	this._replyText.show();
	this._richTextEditor.setContent("");
	this._richTextEditor.element.hide();
};

/**
 * show the richt text editor
 * @param {Object} e - event object
 */
AnnotationRow.prototype._showRichTextEditor = function(e) {
	this._replyText.hide();
	if (!this._richTextEditor.isInitialized()) {
		// only add first time
		this._replyText.after(this._richTextEditor.element);
	}
	this._richTextEditor.element.show();
	this._annotation.setUnsavedReply("");
	this._richTextEditor.setContent("<html><body></body></html>");
	this._richTextEditor.init(
		_.bind(function() {
			this._charLineCount.update(this._richTextEditor.getContent());
			// Scroll rich text editor in view
			this._richTextEditor.element.trigger("makeMeVisible");
			this._richTextEditor.element.trigger("contentResized");
		}, this)
	);
	e.preventDefault();
	e.stopPropagation();
};

/**
 * show/hide the character line count
 * @param {Boolean} show - show or hide the character line count
 * @param {Boolean} updateSingleRows - do all single rows need to be updated
 */
AnnotationRow.prototype.showHideCharLineCount = function(show, updateSingleRows) {
	if (show && this._annotationModel.isAnnotationOrReplySelected(this._annotation)) {
		this._charLineCount.show();
		this._charLineCount.update(this._richTextEditor.getContent());
	} else {
		this._charLineCount.hide();
	}
	if (updateSingleRows) {
		_.each(this.annotationRows, function(row) {
			row.showHideCharLineCount(show);
		});
	}
};

/**
 * Redraw the annotation row after the annotation was updated
 * @param {Annotation} refannotation
 */
AnnotationRow.prototype.annotationUpdated = function(refannotation) {
	this.element.css({
		"border-color": this._annotation.annotationcolor
	});

	if (this.title) {
		this.title.text(this._annotation.getMessageNumber());
		this.title.css({
			color: this._annotation.getHeaderTextColor(),
			"background-color": this._annotation.annotationcolor
		});
	}

	this._mainRow.setAnnotation(this._annotation);

	var replyStack = this._annotationModel.getReplies(this._annotation);
	this._totalReplyCount = replyStack.length;

	//Remove first the replies that might have been deleted
	var currentReplies = [];
	_.each(this.annotationRows, function(row, annotName) {
		currentReplies.push(annotName);
	});
	this._annotation.runThroughAnnotation(true, true, false, function(annot) {
		var i = _.indexOf(currentReplies, annot.name);
		if (i > -1) {
			currentReplies.splice(i, 1);
		}
	});
	_.each(
		currentReplies,
		function(reply) {
			this.annotationRows[reply].element.remove();
			delete this.annotationRows[reply];
		},
		this
	);

	/**
	 * Now redraw the replies
	 */
	var currentLimit = this._totalReplyCount;
	if (this._limitReplyCount) {
		currentLimit = this._replyLimit;
	}

	if (refannotation) {
		var indexOfRefAnnotation = replyStack.length - _.indexOf(replyStack, refannotation.parent); //minimum go to the annotation that is ref'd
		if (indexOfRefAnnotation <= replyStack.length) {
			currentLimit = Math.max(currentLimit, indexOfRefAnnotation);
		}
	}

	this._shownReplyCount = Math.min(currentLimit, this._totalReplyCount);

	for (var i = 0; i < this._totalReplyCount; i++) {
		var annot = replyStack[i];
		var row = this.annotationRows[annot.name];
		// show
		if (!row) {
			row = new AnnotationSingleRow(this.f, annot, false, this._annotationModel, undefined, undefined, this._isForMobile);
			this._repliesDiv.append(row.element);
			this.annotationRows[annot.name] = row;
			row.setSelected(this.selected);
			row.setVisible(false);
		}

		if (this._totalReplyCount - 1 - i < currentLimit) {
			row.setAnnotation(annot);
			row.setVisible(true);
		} else {
			// hide
			if (row) {
				row.setVisible(false);
			}
		}
	}
	this._repliesDiv.prepend(this._moreReplies);
	if (this.shouldShowMoreRepliesLink()) {
		// show morereplies link
		this._moreReplies.show();
	} else {
		this._moreReplies.hide();
	}
	this.updateReplyLimitLabel();

	this.showHideCharLineCount(this.f.getOption("PREFERENCES_ANNOTATION_SHOWCOUNT"));

	if (this._annotationModel.canChangeAnnotations(true) && !this._annotation.isNew) {
		this._reply.show();

		if (this._annotation.getUnsavedReply() === null) {
			this._replyText.show();
			this._closeRichText();
		} else {
			this._richTextEditor.setContent(this._annotation.getUnsavedReply());
		}
	} else {
		this._reply.hide();
		this._replyText.hide();
	}
};

/**
 * Should the more replies link be shown?
 * @returns {Boolean}
 */
AnnotationRow.prototype.shouldShowMoreRepliesLink = function() {
	return this._totalReplyCount > this._replyLimit;
};

/**
 * Toggle the reply limit, sets the setting so all annotation replies are shown/hidden
 * @param {Boolean} toggle - possible to set if the replies should be shown or hidden
 */
AnnotationRow.prototype.toggleReplyLimit = function(toggle) {
	if (toggle === undefined) {
		this._limitReplyCount = !this._limitReplyCount;
	} else {
		this._limitReplyCount = toggle;
	}

	this.annotationUpdated();
	this.updateReplyLimitLabel();
	this._moreReplies.trigger("contentResized");
};

/**
 * Update the reply limit label, changes the link to show/hide annotation replies
 */
AnnotationRow.prototype.updateReplyLimitLabel = function() {
	if (!this._limitReplyCount) {
		// showing all replies, show less link
		this._moreRepliesLabel.text(this.f.i18n.t("AnnShowLessReplies"));
	} else {
		// show only limited replies, show more link
		var repliesHidden = this._totalReplyCount - this._shownReplyCount;
		if (repliesHidden <= 1) {
			this._moreRepliesLabel.text(this.f.i18n.t("AnnShowOneMoreReply"));
		} else {
			this._moreRepliesLabel.text(this.f.i18n.t("AnnShowMoreReplies", repliesHidden));
		}
	}
};

/**
 * updates the timers of the annotation row
 */
AnnotationRow.prototype.updateTimers = function() {
	_.each(this.annotationRows, function(row) {
		row.updateTimer();
	});
};

/**
 * Set this row to be the selected row, this highlights the row in the annotation widget
 * @param {Boolean} selected - does the row needs to be hightlighted or not
 */
AnnotationRow.prototype.setSelected = function(selected) {
	this.selected = selected;
	_.each(this.annotationRows, function(row) {
		row.setSelected(selected);
	});
	this.element.toggleClass("selected", selected);
	this.showHideCharLineCount(this.f.getOption("PREFERENCES_ANNOTATION_SHOWCOUNT"));
};

/**
 * Detach the annotation row from the DOM
 */
AnnotationRow.prototype.onDetachFromDOM = function() {
	if (!this.detachedFromDOM) {
		this.detachedFromDOM = true;
		this._richTextEditor.beforeDetach();
		_.each(this.annotationRows, function(row) {
			row.onDetachFromDOM();
		});
	}
};

/**
 * Reinitialize TinyMCE, needed after reattaching the editor to the dom
 */
AnnotationRow.prototype.onReattachToDOM = function() {
	if (this.detachedFromDOM) {
		this.detachedFromDOM = false;
		this._richTextEditor.repair();
		_.each(this.annotationRows, function(row) {
			row.onReattachToDOM();
		});
	}
};

/**
 * Set the available height for the row (e.g. while resizing the popup)
 * @param {Number} height - Available height.
 */
AnnotationRow.prototype.setAvailableHeight = function(height) {
	// Adapt the reply editor size
	if (this._annotation.getUnsavedReply() !== null) {
		var leftOverHeight = height - 10; // minus 1 for border on .richText div, minus 5 for margin and minus 2x2 for padding on .addReply div

		if (this.shouldShowMoreRepliesLink()) {
			// the more replies link is shown
			leftOverHeight -= this._moreReplies.outerHeight(true) || null;
		}

		_.each(this.annotationRows, function(row) {
			if (row.isVisible()) {
				// the row is visible
				leftOverHeight -= row.element.outerHeight(true) || null;
			}
		});
		this._richTextEditor.setHeight(leftOverHeight);
	}
};

/**
 * Set the annotation for the AnnotationRow. This will trigger a render of the HTML
 * @param {Annotation} annotation - the annotation]
 */
AnnotationRow.prototype.setAnnotation = function(annotation) {
	this._annotation = annotation;
	this.annotationUpdated();
};

/**
 * Returns the annotation object assigned with the annotation row
 * @returns {Annotation}
 */
AnnotationRow.prototype.getAnnotation = function() {
	return this._annotation;
};

export default AnnotationRow;
