import AutoTimeLabel from "@esko/web-viewer/layout/AutoTimeLabel";
import AnnotationConstants from "@esko/web-viewer/annotation/AnnotationConstants";
import AnnotationCharLineCount from "@esko/web-viewer/annotation/AnnotationCharLineCount";
import AnnotationTimeRow from "@esko/web-viewer/annotation/AnnotationTimeRow";
import MiniButton from "@esko/web-viewer/layout/component/MiniButton";
import DropdownMiniButton from "@esko/web-viewer/layout/component/DropdownMiniButton";
import ViewerToolbarButton from "@esko/web-viewer/layout/component/ViewerToolbarButton";
import ButtonGroup from "@esko/toolbar/ButtonGroup";
import Tooltip from "@esko/tooltip/Tooltip";
import RichTextEditorFactory from "@esko/web-viewer/layout/component/RichTextEditorFactory";
import ColorPicker from "@esko/web-viewer/layout/component/ColorPicker";
import $ from "jquery";
import _ from "underscore";
import "@esko/web-viewer/utils/HTMLTruncator";
/**
 * @exports AnnotationSingleRow
 * @memberof AnnotationModule
 * @class An annotation div is an interface element which provides one line for an annotation
 * @constructor
 * @param {Core.ViewerFacade} f - Viewerfacade
 * @param {Annotation} annotation - Annotation this single row belongs to
 * @param {Boolean} main - indicates if this is the main annotation
 * @param {AnnotationModel} annotationModel - Annotation Model
 * @param {Boolean} hideHeader - True if you want to hide the header, false otherwise
 * @param {JQueryObj} extraElement - Extra element to be added to the actions div
 * @param {Boolean} isForMobile
 */
var AnnotationSingleRow = function(f, annotation, main, annotationModel, hideHeader, extraElement, isForMobile) {
	this.element = $("<div/>").addClass(main ? "main" : "reply");

	this.selected = false;

	this.f = f;

	this._MAX_CHARS = 300;

	this._hideHeader = hideHeader;

	this._annotation = annotation;

	this._main = main;

	this._annotationModel = annotationModel;

	this._isForMobile = isForMobile;

	this._visible = true;

	/**
	 * Header
	 */
	if (!this._hideHeader) {
		var header = $("<div/>").addClass("annotHeader");
		this.element.append(header);
		if (this._main) {
			header.append(
				$("<i>")
					.addClass("icon")
					.addClass(this._annotation.getIconClass())
			);
		}
		var userName = $("<span />")
			.addClass("userName")
			.text(this._annotation.getUserName());
		header.append(userName);

		if (this._main && !this._isForMobile) {
			/**
			 * 3D View button for desktop
			 */
			this._view3D = new DropdownMiniButton(this.f, "3DView", "icon icon-WCR_Actions-3D12");
			this._view3D.setTooltip("viewer", "annot3DView");
			header.append(this._view3D.element.addClass("view3d"));
		}

		this._timelabel = new AutoTimeLabel(this.f, this._annotation.date);
		this._timelabel.element.addClass("date").addClass("pull-right");
		header.append(this._timelabel.element);

		this.updateTimer = _.bind(function() {
			this._timelabel.date = this._annotation.date;
			this._timelabel.updateText(this._annotation.isAnnotationModified());
		}, this);
	}

	/**
	 * Content
	 */
	this._content = $("<div/>").addClass("annotContent");
	this.element.append(this._content);

	// Time stamps for video
	if ((this._annotation.startTime != null || this._annotation.endTime != null) && this._annotation.parent == null) {
		this.timeStamps = new AnnotationTimeRow(this.f, this._annotation, this._annotationModel, false);
		this._content.after(this.timeStamps.getElement());
	}

	/**
	 * Actions
	 */
	this._actionsDiv = $("<div/>")
		.addClass("annotActions")
		.addClass("clearfix");

	// Edit + delete actions
	this._moreActions = new DropdownMiniButton(this.f, "moreActions", "icon-WCR_More-horizontal");
	this._moreActions.addButtonClass("annotAction");
	this._moreActions.setTooltip("viewer", "moreActions");
	// Rich text
	this._richTextEditor = undefined;
	this._annotationColorPicker = undefined;

	this._actionsDiv.append(this._moreActions.element);

	// Buttons in the editor
	// Save button
	var saveEdit = new ViewerToolbarButton(
		this.f,
		"saveEdit",
		"icon-WCR_Save12 marginRight5",
		true,
		_.bind(function(e) {
			this._annotation.content = this._richTextEditor.getContent();
			this._annotation.setModified("CONTENT");
			this._annotationModel.save();
			this.closeRichText();
			e.stopPropagation();
		}, this)
	);
	// Cancel button
	var cancelEdit = new ViewerToolbarButton(
		this.f,
		"cancelReply",
		"icon-WCR_Revert12 marginRight5",
		true,
		_.bind(function(e) {
			this._annotation.resetAnnotationColor();
			this._annotationModel.modelChanged(this._annotation);
			this._annotation.cancelModifiedColor();
			this._richTextEditor.setContent("");
			this._annotation.cancelModified("CONTENT");
			this._annotationModel.modelChanged();
			this.closeRichText();
			e.stopPropagation();
		}, this)
	);
	// Save and cancel are a buttongroup
	this._editActionsGroup = new ButtonGroup("editActionsGroup", [cancelEdit, saveEdit]);
	this._editActionsGroup.element.addClass("pull-right");

	if (annotation.getPermissionFor(AnnotationConstants.ActionPermissions.DUPLICATE_WITHOUT_TEXT)) {
		// duplicateWithoutText
		this._duplicateWithoutText = new MiniButton(
			this.f,
			"annotationCopyWithoutText",
			"icon-WCR_Annotation-Duplicate-NoText",
			true,
			_.bind(function(e) {
				this.f.emit("module.annotation.duplicateWithoutText", this._annotation);
				e.stopPropagation();
			}, this)
		);
		this._duplicateWithoutText.addButtonClass("annotAction");
		this._duplicateWithoutText.setTooltip("viewer", "DuplicateWithoutTextAnnotBtnTT");
		this._actionsDiv.append(this._duplicateWithoutText.element);
	}

	// duplicate
	if (annotation.getPermissionFor(AnnotationConstants.ActionPermissions.DUPLICATE)) {
		this._duplicateAction = new MiniButton(
			this.f,
			"annotationDuplicate",
			"icon-WCR_Annotation-Duplicate",
			true,
			_.bind(function(e) {
				this.f.emit("module.annotation.duplicateAnnotation", this._annotation);
				e.stopPropagation();
			}, this)
		);
		this._duplicateAction.addButtonClass("annotAction");
		if (this._annotation.originType === AnnotationConstants.AnnotationOriginTypes.user.id) {
			this._duplicateAction.setTooltip("viewer", "DuplicateAnnotBtnTT");
		} else {
			this._duplicateAction.setTooltip("viewer", "DuplicateToUserAnnotBtnTT");
		}
		this._actionsDiv.append(this._duplicateAction.element);
	}

	if (this._main) {
		// 3D View mobile
		if (this._isForMobile) {
			this._view3DMobile = new MiniButton(
				this.f,
				"3DView",
				"icon icon-WCR_Actions-3D12",
				true,
				_.bind(function(e) {
					this.f.emit("module.annotation.show3DView", this._annotation);
				}, this)
			);
			this._view3DMobile.addButtonClass("annotAction");
			this._view3DMobile.setTooltip("viewer", "annot3DView");
			this._actionsDiv.append(this._view3DMobile.element);
		}

		if (annotation.getPermissionFor(AnnotationConstants.ActionPermissions.ZOOM_TO_ANNOTATION)) {
			// Zoom
			this._zoomAction = new MiniButton(
				this.f,
				"annotationZoom",
				"icon-WCR_ZoomFit",
				true,
				_.bind(function(e) {
					this.f.emit("module.annotation.zoomToAnnotation", this._annotation);
					e.stopPropagation();
				}, this)
			);
			this._zoomAction.addButtonClass("annotAction");
			this._zoomAction.setTooltip("viewer", "ZoomAnnotBtnTT");
			this._actionsDiv.append(this._zoomAction.element);
		}
	}

	// Review status
	this._statusAction = new DropdownMiniButton(this.f, "setReviewStatus", "icon-WCR_Approve-submenu");
	this._statusAction.addButtonClass("annotAction");
	this._statusAction.setTooltip("viewer", "SetReviewStateBtnTT");

	this._actionsDiv.append(this._statusAction.element);

	if (this._isForMobile && extraElement !== undefined) {
		this._actionsDiv.append(extraElement);
	}

	// Character/Line count for annotation
	this._charLineCount = new AnnotationCharLineCount(this.f);
	this._actionsDiv.append(this._charLineCount.element);

	this.element.append(this._actionsDiv);

	//reviews
	this._reviews = $("<ul/>")
		.addClass("Review")
		.hide();
	this.element.append(this._reviews);

	this.element
		.on(
			"mouseenter",
			_.bind(function() {
				this.mouseEntered = true;
				this.showOrHideOptions();
			}, this)
		)
		.on(
			"mouseleave",
			_.bind(function() {
				this.mouseEntered = false;
				this.showOrHideOptions();
			}, this)
		);

	this.f.i18n.loadBundle(
		"viewer",
		_.bind(function() {
			this._moreActions.addAction({
				name: "annotationEdit",
				content: this.f.i18n.t("EditAnnotBtnTT"),
				icon: "icon-WCR_Annotate",
				action: _.bind(function() {
					if (!this._annotation.isNew) {
						this._initializeRichTextEditor();
						if (!this._richTextEditor.isInitialized()) {
							// only add first time
							this._content.after(this._richTextEditor.element);
							this._richTextEditor.element.trigger("makeMeVisible");
						}
						// On click, show rich text
						this._richTextEditor.element.show();
						this._content.hide();

						// Hide the actions div, move the char line counter in the editor
						this._actionsDiv.hide();
						this._charLineCount.element.insertAfter(this._editActionsGroup.element);

						// update
						this._richTextEditor.setContent("<html>" + this._annotation.content + "</html>");
						this.updateColorPicker();

						this._richTextEditor.init();
					} else {
						this._annotationModel.selectAnnotation(this._annotation, true);
					}
				}, this)
			});

			this._moreActions.addAction({
				name: "duplicateAnnotation",
				content: this.f.i18n.t("DuplicateAnnotBtnTT"),
				icon: "icon-WCR_Annotation-Duplicate",
				action: _.bind(function(e) {
					this.f.emit("module.annotation.duplicateAnnotation", this._annotation);
					e.stopPropagation();
				}, this)
			});

			if (this._annotation.getPermissionFor(AnnotationConstants.ActionPermissions.DUPLICATE_WITHOUT_TEXT)) {
				this._moreActions.addAction({
					name: "duplicateWithoutText",
					content: this.f.i18n.t("DuplicateWithoutTextAnnotBtnTT"),
					icon: "icon-WCR_Annotation-Duplicate-NoText",
					action: _.bind(function(e) {
						if (e.which) {
							this.f.emit("module.annotation.duplicateWithoutText", this._annotation);
							e.stopPropagation();
						}
					}, this)
				});
			}

			this._moreActions.addAction({
				name: "annotationSave",
				content: this.f.i18n.t("SaveAnnotBtnTT"),
				icon: "icon-WCR_Save",
				action: _.bind(function(e) {
					this._annotationModel.save();
				}, this)
			});

			this._moreActions.addAction({
				name: "annotationDelete",
				content: this.f.i18n.t("DeleteAnnotBtnTT"),
				icon: "icon-WCR_Trash",
				action: _.bind(function(e) {
					this._annotationModel.deleteAnnotation(this._annotation);
				}, this)
			});

			// 3D button actions in more actions
			if (this._main) {
				if (!this._isForMobile) {
					this._moreActions.addAction({
						name: "update3DView",
						content: this.f.i18n.t("annot3DViewUpdate"),
						icon: "icon icon-WCR_Annotation-3D-Update",
						action: _.bind(function(e) {
							this.f.emit("module.annotation.update3DView", this._annotation);
						}, this)
					});

					this._moreActions.addAction({
						name: "delete3DView",
						content: this.f.i18n.t("annot3DViewDelete"),
						icon: "icon icon-WCR_Annotation-3D-Delete",
						action: _.bind(function(e) {
							this.f.emit("module.annotation.delete3DView", this._annotation);
						}, this)
					});
				}
			}

			saveEdit.element
				.addClass("btn-mini")
				.addClass("btn-success")
				.append(this.f.i18n.t("Save"));
			cancelEdit.element.addClass("btn-mini").append(this.f.i18n.t("CancelButton"));

			// Create buttons
			_.each(
				AnnotationConstants.AnnotationReviewStateLocs,
				_.bind(function(loc, name, i) {
					if (name != "STATUS_ANY") {
						this._statusAction.addAction({
							name: name,
							content: this.f.i18n.t(loc),
							action: _.bind(function() {
								this._annotationModel.setState(this._annotation, AnnotationConstants.AnnotationReviewState[name]);
							}, this),
							icon: AnnotationConstants.AnnotationReviewStateIcons[name]
						});
					}
				}, this)
			);

			// 3D button actions
			if (this._view3D) {
				this._view3D.addAction({
					name: "update3DView",
					content: this.f.i18n.t("annot3DViewUpdate"),
					icon: "icon icon-WCR_Annotation-3D-Update",
					action: _.bind(function(e) {
						this.f.emit("module.annotation.update3DView", this._annotation);
					}, this)
				});

				this._view3D.addAction({
					name: "delete3DView",
					content: this.f.i18n.t("annot3DViewDelete"),
					icon: "icon icon-WCR_Annotation-3D-Delete",
					action: _.bind(function(e) {
						this.f.emit("module.annotation.delete3DView", this._annotation);
					}, this)
				});
			}

			this._updateView3DButtons();
		}, this)
	);
};

/**
 * Close rich text editor and replace with content
 */
AnnotationSingleRow.prototype.closeRichText = function() {
	// reset the buttons
	this._actionsDiv.show();
	this._actionsDiv.append(this._charLineCount.element);
	// on cancel, show input
	this._content.show();
	this._richTextEditor.element.hide();
	if (this.timeStamps) {
		this.timeStamps.show();
	}
};

/**
 * Show or hide the character line count in the annotation row
 * @param {Boolean} show - show the char count
 */
AnnotationSingleRow.prototype.showHideCharLineCount = function(show) {
	var mainAnnotation = this._annotation.getGodfather();
	if (show && this._annotationModel.isAnnotationOrReplySelected(mainAnnotation)) {
		this._charLineCount.show();
		this._charLineCount.update(this._annotation.content);
	} else {
		this._charLineCount.hide();
	}
};

/**
 * Repaint the entire row
 */
AnnotationSingleRow.prototype.repaintRow = function() {
	if (!this._hideHeader) {
		this.updateTimer();
	}
	if (!this._annotation.isBeingModified() && this._richTextEditor !== undefined) {
		this.closeRichText();
	} else {
		this.updateColorPicker();
	}

	this._updateView3DButtons();

	this._content.empty();
	this._content.removeClass("noContent");

	// Extra content
	var extraContent = this._annotation.getExtraContent();
	if (extraContent) {
		if (extraContent.element) {
			this._content.append(extraContent.element);

			if (extraContent.tooltip) {
				if (this.extraContentTooltip) {
					this.extraContentTooltip.destroy();
				}
				this.extraContentTooltip = new Tooltip(
					this.f.i18n,
					extraContent.element,
					this.f.getContext(),
					this.f.getContext(),
					this.f.isMobile(),
					extraContent.tooltip
				);
			}
		}
	}

	if (this._annotation.content && this._annotation.content.length > 0) {
		var wrapper = $("<div>").html(this._annotation.content);
		this._content.append(wrapper);
		this.f.i18n.loadBundle(
			"viewer",
			_.bind(function() {
				wrapper.truncate({
					max_length: this._MAX_CHARS,
					more: this.f.i18n.t("ReadMore"),
					less: this.f.i18n.t("ReadLess"),
					css_more_class: "truncator-link truncator-more",
					css_less_class: "truncator-link truncator-less",
					link_prefix: "",
					link_suffix: ""
				});
			}, this)
		);
	} else {
		if (!extraContent) {
			this.f.i18n.loadBundle(
				"viewer",
				function() {
					this._content.addClass("noContent");
					if (this._annotation.isNew) {
						this._content.text(this.f.t("AnnNewClickToEdit"));
					} else {
						this._content.text(this.f.t("AnnNoContent"));
					}
				},
				this
			);
		}
	}

	this.showOrHideOptions();

	this._updateReviews(this._annotation.states);

	this.showHideCharLineCount(this.f.getOption("PREFERENCES_ANNOTATION_SHOWCOUNT"));

	if (this._annotationModel.canChangeAnnotations(true)) {
		if (this._annotationModel.canChangeAnnotation(this._annotation)) {
			this._moreActions.show();
			if (this._duplicateAction) {
				this._duplicateAction.hide();
			}
			if (this._duplicateWithoutText) {
				this._duplicateWithoutText.hide();
			}
		} else {
			this._moreActions.hide();
			if (this._duplicateAction) {
				this._duplicateAction.show();
			}
			if (this._duplicateWithoutText) {
				this._duplicateWithoutText.show();
			}
		}
		this._statusAction.show();

		this._moreActions.setActionEnabled("annotationDelete", this._annotationModel.canDeleteAnnotation(this._annotation));
	} else {
		if (this._duplicateAction) {
			this._duplicateAction.hide();
		}
		if (this._duplicateWithoutText) {
			this._duplicateWithoutText.hide();
		}
		this._moreActions.hide();
		this._statusAction.hide();
	}
	try {
		this.element.attr("data-sahi-id", "viewer-annotation" + this._annotation.messagenumber);
		this._statusAction.element.attr("data-sahi-id", "viewer-annotation-setReviewStatus-" + this._annotation.messagenumber);
		this._moreActions.element.get(0).dataset.sahiId = "viewer-annotation-MoreActions-" + this._annotation.messagenumber;
	} catch (e) {}
};

/**
 * Checks if the row should show all the option-buttons
 * @returns {Boolean}
 */
AnnotationSingleRow.prototype.shouldShowOptions = function() {
	if (this._isForMobile) {
		return this.selected;
	} else {
		return this.selected || this.mouseEntered || this._moreActions.isOpen() || this._statusAction.isOpen();
	}
};

/**
 * Shows or hides the options, uses shouldShowOptions to check whether to show or hide
 */
AnnotationSingleRow.prototype.showOrHideOptions = function() {
	this.showOptions(this.shouldShowOptions());
};

/**
 * Shows the extra annotation action-buttons (options)
 * @param {Boolean} show
 */
AnnotationSingleRow.prototype.showOptions = function(show) {
	if (show) {
		if (this._isForMobile) {
			this._actionsDiv.show();
		} else {
			this._actionsDiv.removeClass("hideOptions");
		}
	} else {
		if (this._isForMobile) {
			this._actionsDiv.hide();
		} else {
			this._actionsDiv.addClass("hideOptions");
		}
	}
};

/**
 * Update all the annotation review states
 * @param {Array.<Annotation>} states
 */
AnnotationSingleRow.prototype._updateReviews = function(states) {
	this.f.i18n.loadBundle(
		"viewer",
		function() {
			this._reviews.empty();
			for (var i = 0; i < states.length; i++) {
				var review = $("<li/>");

				var container = $("<a/>");
				var image = $("<i/>")
					.addClass("icon")
					.addClass(AnnotationConstants.AnnotationReviewStateIcons[_.invert(AnnotationConstants.AnnotationReviewState)[states[i].state]]);
				var reviewStateKeys = _.invert(AnnotationConstants.AnnotationReviewState);
				var reviewState = reviewStateKeys[states[i].state];
				var user = $("<span/>").text(
					'"' + this.f.i18n.t(AnnotationConstants.AnnotationReviewStateLocs[reviewState]) + '": ' + states[i].getUserName()
				);

				container.append(image);
				container.append(user);
				review.append(container);

				this._reviews.append(review);
			}
		},
		this
	);
	this._reviews.toggle(states.length > 0);
};

/**
 * initialize the rich text editor of the annotation row
 */
AnnotationSingleRow.prototype._initializeRichTextEditor = function() {
	if (!this._richTextEditor) {
		this._richTextEditor = RichTextEditorFactory.createEditor(this.f);

		// Show a color picker if no annotation color is set
		//color picker only added for main annotation
		if (!this.f.getProperty("annotationcolor") && this._main) {
			this._annotationColorPicker = new ColorPicker(
				this.f,
				_.bind(function(newColor) {
					this._annotation.setColor("#" + newColor);
					this._annotationModel.modelChanged(this._annotation);
				}, this)
			);

			this._annotationColorPicker.setColor(this._annotation.annotationcolor);
			this._annotationColorPicker.element.addClass("annotationColorPicker");
			this._annotationColorPicker.setTooltip(this.f, "general", "ChangeColor");
		}

		var buttons = [this._editActionsGroup, this._annotationColorPicker];

		// Time stamps for video
		if ((this._annotation.startTime != null || this._annotation.endTime != null) && this._annotation.parent == null) {
			var timestamps = new AnnotationTimeRow(this.f, this._annotation, this._annotationModel, true);
			buttons.unshift({ element: timestamps.getElement() });
			if (this.timeStamps) {
				this.timeStamps.hide();
			}
		}

		// Add to rich text editor
		this._richTextEditor.setButtons(buttons);

		// Let the counter listen to the richtext editor
		this._richTextEditor.onTextChanged(
			_.bind(function() {
				this._charLineCount.update(this._richTextEditor.getContent());
				this._annotation.setBeingModified(this._richTextEditor.getContent());
				this._annotationModel.modelChanged(this._annotation);
			}, this)
		);
	}
};

/**
 * Update 3D view button
 */
AnnotationSingleRow.prototype._updateView3DButtons = function() {
	var has3Dview = this._annotation.get3DViewPosition() !== null;

	var view3DModule = this.f.getModule("view3D");
	var view3DWidgetVisible = view3DModule && view3DModule.is3DWidgetEnabled();

	var canChange = this._annotationModel.canChangeAnnotation(this._annotation);

	if (this._view3D) {
		this._view3D.toggleButton(has3Dview);
		this._view3D.toggleElementVisibility("update3DView", canChange && view3DWidgetVisible);
		this._view3D.toggleElementVisibility("delete3DView", canChange && has3Dview);

		if (canChange) {
			this._view3D.enableDropdown();
		} else {
			this._view3D.disableDropdown();
		}
	}

	if (this._view3DMobile) {
		if (has3Dview) {
			this._view3DMobile.show();
		} else {
			this._view3DMobile.hide();
		}
	}

	if (this._moreActions) {
		this._moreActions.toggleElementVisibility("delete3DView", canChange && has3Dview);
		this._moreActions.toggleElementVisibility("update3DView", canChange && view3DWidgetVisible);
	}
};

/**
 * set if the annotation is selected, the row should be highlighted
 * @param {Boolean} selected
 */
AnnotationSingleRow.prototype.setSelected = function(selected) {
	this.selected = selected;
	this.element.toggleClass("selected", this.selected);
	this.repaintRow();
};

/**
 * Should be called when the annotation row is detached from the DOM
 */
AnnotationSingleRow.prototype.onDetachFromDOM = function() {
	if (this._richTextEditor) {
		this._richTextEditor.beforeDetach();
	}
	this._hideAllTooltips();
};

/**
 * Hide all tooltips
 */
AnnotationSingleRow.prototype._hideAllTooltips = function() {
	if (this.extraContentTooltip) {
		this.extraContentTooltip.hide();
	}

	if (this._duplicateAction) {
		this._duplicateAction.hideTooltip();
	}

	if (this._duplicateWithoutText) {
		this._duplicateWithoutText.hideTooltip();
	}

	if (this._zoomAction) {
		this._zoomAction.hideTooltip();
	}

	this._timelabel.tooltip.hide();
	this._moreActions.hideTooltip();
};

/**
 * Should be called when the annotation row gets retached to the DOM
 */
AnnotationSingleRow.prototype.onReattachToDOM = function() {
	if (this._richTextEditor) {
		this._richTextEditor.repair();
	}
};

/**
 * Updates the color picker
 */
AnnotationSingleRow.prototype.updateColorPicker = function() {
	if (this._annotationColorPicker) {
		this._annotationColorPicker.setColor(this._annotation.annotationcolor);
	}
};

/**
 * Set the row visible or not
 * @param {Boolean} visible - True for show, false for hide
 */
AnnotationSingleRow.prototype.setVisible = function(visible) {
	this._visible = visible;
	this.element.toggle(visible);
};

/**
 * Is the row currently visible or not
 * @returns {Boolean} True if visible, false otherwise
 */
AnnotationSingleRow.prototype.isVisible = function() {
	return this._visible;
};

/**
 * Set the Annotation for this row and repaint
 * @param {Annotation} annotation - the new annotation object
 */
AnnotationSingleRow.prototype.setAnnotation = function(annotation) {
	this._annotation = annotation;
	this.repaintRow();
};

export default AnnotationSingleRow;
