import _ from "underscore";
import AbstractTool from "@esko/web-viewer/viewport/tools/AbstractTool";
import ShapeUtils from "@esko/shapeutils";
import DrawUtils from "@esko/web-viewer/utils/DrawUtils";
import ViewerUtils from "@esko/web-viewer/utils/ViewerUtils";
import AnnotationConstants from "@esko/web-viewer/annotation/AnnotationConstants";
import Annotation from "@esko/web-viewer/annotation/Annotation";
import CoreConstants from "@esko/web-viewer/core/CoreConstants";
/**
 * Construct a new Annotation Tool
 * @exports AnnotationTool
 * @memberof AnnotationModule
 *
 * @class Tool to select and edit annotations
 * @constructor
 * @param {Core.ViewerFacade} f - Viewer facade
 * @param {AnnotationModel} annotationModel - Annotation model
 */
var AnnotationTool = function(f, annotationModel) {
	AbstractTool.call(this, f, CoreConstants.TOOLS.ANNOTATION);

	this.f = f;

	this._annotationModel = annotationModel;

	/**
	 * Different states constants
	 */
	this._defaultState = "defaultState";
	this._hoverState = "hoverState";
	this._draggingState = "draggingState";
	this._anchorState = "anchorState";

	/**
	 * Variable that contain the current situation
	 */
	this._anchor = undefined;
	this._targetAnnotation = null;

	/**
	 * Variable to keep track of where interaction started
	 */
	this._touchDown = null;

	/**
	 * The default state
	 */
	this.registerState(
		this._defaultState,
		(function(that) {
			return {
				mousemove: function(e) {
					var l = that._getAnnotationLayer(e);
					if (!l || !l.isVisible()) {
						return;
					}

					that._verifyObjects(e, l);
					if (that._targetAnnotation || that._anchor) {
						that.requestActivation();
					}
				},
				mouseup: function(e) {},
				mousedown: function(e) {
					var l = that._getAnnotationLayer(e);
					if (!l) {
						return;
					}

					annotationModel.unselectAllAnnotations();
					e.originalEvent.preventDefault();
				},
				mouseout: function(e) {
					if (that.isActive()) {
						f.emit("toolManager.request.deactivation", that);
					}
				},
				mouseleave: function() {
					if (that.isActive()) {
						f.emit("toolManager.request.deactivation", that);
					}
				},
				contextmenu: function(e) {
					e.preventDefault();
					e.stopPropagation();
					return false;
				}
			};
		})(this)
	);

	this.registerState(
		this._hoverState,
		_.defaults(
			(function(that) {
				return {
					mousedown: function(e) {
						if (e.which === 1) {
							that._verifyObjects(e);
							that._touchDown = that._getAnnotationLayer(e).getWorldPointFromScene(that.getScenePoint(e));
							if (that._anchor && that._targetAnnotation.annottype != AnnotationConstants.AnnotationType.ANNOT_FREEFORM) {
								f.emit("viewport.startInteractive");
								that.changeState(that._anchorState);
							} else if (that._targetAnnotation) {
								f.emit("viewport.startInteractive");
								annotationModel.selectAnnotation(that._targetAnnotation);
								that.changeState(that._draggingState);
							}
						}
						e.originalEvent.preventDefault();
					},
					mousemove: function(e) {
						that._verifyObjects(e);
						that.cursorShouldChange();
					}
				};
			})(this),
			this.getState(this._defaultState)
		)
	);

	/**
	 * Dragging state is when an annotation is selected and every mousemove the annotation is moved
	 */
	this.registerState(
		this._draggingState,
		_.defaults(
			(function(that) {
				return {
					mousemove: function(e) {
						that.cursorShouldChange();
						var worldPoint = that._getAnnotationLayer(e).getWorldPointFromScene(that.getScenePoint(e));
						var worldBounds = that._widget.getWorld().getWorldBounds(); //get the world bounds
						//get the converted worldbounds (with origin (0,0)), the origin gets added later on to not intervere with transforming the worldbounds to annotation coordinate system
						var worldRect = that
							._getAnnotationLayer(e)
							.convertWorldRectToLayer(new ShapeUtils.Rect(new ShapeUtils.Point(), worldBounds.size));
						worldRect.origin.y = -(worldRect.size.height + worldRect.origin.y) - worldBounds.origin.y; //calculate the origin in annotation coordinate system
						worldRect.origin.x += worldBounds.origin.x; //add the origin to the worldRect (this is not converted as it stays a constant)
						var moveSize = that.keepMoveSizeWithinRect(worldRect, that._targetAnnotation.rect, that._touchDown.sizeTo(worldPoint)); //keep move size within world bounds
						annotationModel.moveAnnotation(that._targetAnnotation, moveSize);
						that._touchDown = worldPoint;
					},
					mouseup: function(e) {
						f.emit("viewport.stopInteractive");
						that.changeState(that._hoverState);
					},
					mouseout: function(e) {
						f.emit("viewport.stopInteractive");
						that.changeState(that._hoverState);
					},
					enterState: function(e) {
						f.emit("viewport.startInteractive");
					},
					leaveState: function(e) {
						f.emit("viewport.stopInteractive");
					}
				};
			})(this),
			this.getState(this._defaultState)
		)
	);

	/**
	 * Anchor state is when clicked and an anchor is selected
	 */
	this.registerState(
		this._anchorState,
		_.defaults(
			(function(that) {
				/**
				 * @returns a suffix for the anchor point, so the right action can be calculated
				 */
				function suffixForAnchor() {
					var suffix = null;
					switch (that._anchor.anchor) {
						case DrawUtils.LWANCHOR_TOPLEFT:
							suffix = "NW";
							break;
						case DrawUtils.LWANCHOR_TOPCENTER:
							suffix = "N";
							break;
						case DrawUtils.LWANCHOR_TOPRIGHT:
							suffix = "NE";
							break;
						case DrawUtils.LWANCHOR_BOTTOMLEFT:
							suffix = "SW";
							break;
						case DrawUtils.LWANCHOR_BOTTOMCENTER:
							suffix = "S";
							break;
						case DrawUtils.LWANCHOR_BOTTOMRIGHT:
							suffix = "SE";
							break;
						case DrawUtils.LWANCHOR_CENTERLEFT:
							suffix = "W";
							break;
						case DrawUtils.LWANCHOR_CENTERRIGHT:
							suffix = "E";
							break;
						default:
							break;
					}
					return suffix;
				}

				return {
					mousemove: function(e) {
						var worldPoint = that._getAnnotationLayer(e).getWorldPointFromScene(that.getScenePoint(e));
						var suffix = suffixForAnchor();
						if (suffix != null) {
							var newrect = that._anchor.linework.rect.copy();

							//Figure out x-axis
							if (suffix.indexOf("E") > -1 || suffix.indexOf("W") > -1) {
								if (suffix.indexOf("W") > -1) {
									var sizeToX = that._anchor.linework.rect.left() - worldPoint.x;
									newrect.origin.x -= sizeToX;
									newrect.size.width += sizeToX;
								} else {
									var sizeToX = that._anchor.linework.rect.right() - worldPoint.x;
									newrect.size.width -= sizeToX;
								}
							}

							//Figure out y-axis
							if (suffix.indexOf("N") > -1 || suffix.indexOf("S") > -1) {
								if (suffix.indexOf("N") > -1) {
									var sizeToY = that._anchor.linework.rect.top() - worldPoint.y;
									newrect.origin.y -= sizeToY;
									newrect.size.height += sizeToY;
								} else {
									var sizeToY = that._anchor.linework.rect.bottom() - worldPoint.y;
									newrect.size.height -= sizeToY;
								}
							}

							annotationModel.transformAnnotation(that._anchor.linework, newrect);
						} else {
							if (that._anchor.anchor == DrawUtils.LWANCHOR_START) {
								annotationModel.updateStartEndOfAnnotation(that._anchor.linework, worldPoint, null);
							} else if (that._anchor.anchor == DrawUtils.LWANCHOR_END) {
								annotationModel.updateStartEndOfAnnotation(that._anchor.linework, null, worldPoint);
							}
						}
						that._touchDown = worldPoint;
					},
					mouseup: function(e) {
						f.emit("viewport.stopInteractive");
						that.cursorShouldChange();
						that.changeState(that._hoverState);
					},
					mouseleave: function(e) {
						f.emit("viewport.stopInteractive");
						that.cursorShouldChange();
						that.changeState(that._hoverState);
					},
					mouseout: function(e) {}
				};
			})(this),
			this.getState(this._defaultState)
		)
	);

	this._override(
		"initialize",
		_.bind(function(widget, toolManager) {
			this._super(widget, toolManager);

			// on pressing delete key, request to delete the currently selected annotation
			this.persistentKeyEvents.push([
				"del",
				_.bind(function(e) {
					if (this._annotationModel.getSelectedAnnotations().length > 0) {
						_.each(
							this._annotationModel.getSelectedAnnotations(),
							_.bind(function(annot) {
								this._annotationModel.deleteAnnotation(annot);
							}, this)
						);
					}
				}, this),
				"keyup"
			]);
			this.changeState(this._defaultState);
			this.registerPassiveEvents(["mousemove", "mousedown"]);
			this.registerActiveEvents(["mousemove", "mousedown", "mouseup", "mouseout", "mouseleave", "contextmenu"]);
		}, this)
	);

	this._override(
		"activate",
		_.bind(function(interruptRequest) {
			this._super(interruptRequest);
			this.changeState(this._hoverState);
		}, this)
	);

	this._override(
		"deactivate",
		_.bind(function(interruptRequest) {
			this._super(interruptRequest);
			this.changeState(this._defaultState);
		}, this)
	);
};

// inherit
AnnotationTool.prototype = Object.create(AbstractTool.prototype);
// correct the constructor pointer
AnnotationTool.prototype.constructor = AnnotationTool;

/**
 * Calculate the default cursor
 * 	- when an anchor is selected:
 * 		- line or arrow annotation: move-cursor
 * 		- show correct resize-cursor
 *  - when an annotation is selected
 *  	- show move-cursor unless the annotation can't be moved -> then show a not-allowed cursor
 *  - when nothing is selected: show select arrow
 */
AnnotationTool.prototype.getDefaultCursor = function() {
	if (this._anchor) {
		return this._cursorForAnchor();
	} else if (this.getCurrentStateName() == this._draggingState) {
		if (this._annotationModel.canChangeAnnotation(this._targetAnnotation)) {
			return "cursor-moveObject";
		} else {
			return "cursor-notAllowed";
		}
	} else {
		return "cursor-select";
	}
};

/**
 * Deactivate the current Annotation tool
 * @param that
 * @param e
 */
AnnotationTool.prototype._deActivate = function(e) {
	if (this.isActive()) {
		this.changeState(this._defaultState);
		this.f.emit("toolManager.request.deactivation", this);
	}
};

/**
 * @returns {String} the cursor for the anchor state
 */
AnnotationTool.prototype._cursorForAnchor = function() {
	var suffix = "cursor-moveObject";

	//Show anchor cursor, unless it's an arrow or line, then just show the move cursor
	if (
		this._anchor.linework.annottype == AnnotationConstants.AnnotationType.ANNOT_ARROW ||
		this._anchor.linework.annottype == AnnotationConstants.AnnotationType.ANNOT_LINE
	) {
		return suffix;
	}

	var l = this._getAnnotationLayer("Main");
	if (!l) {
		return suffix;
	}

	//because it's a static size, use the actual size of the rectangle
	var center = l.getScenePointFromWorld(this._anchor.rect.center());
	var size = new ShapeUtils.Size(this._anchor.rect.size.width, this._anchor.rect.size.height);
	var anchorrect = new ShapeUtils.Rect(new ShapeUtils.Point(center.x - size.width / 2, center.y - size.height / 2), size);

	var lwrect = l.getSceneRectFromWorld(this._anchor.linework.rect);

	if (anchorrect.containsPoint(lwrect.topleft())) {
		suffix = "NW";
	} else if (anchorrect.containsPoint(lwrect.topcenter())) {
		suffix = "N";
	} else if (anchorrect.containsPoint(lwrect.topright())) {
		suffix = "NE";
	} else if (anchorrect.containsPoint(lwrect.centerleft())) {
		suffix = "W";
	} else if (anchorrect.containsPoint(lwrect.centerright())) {
		suffix = "E";
	} else if (anchorrect.containsPoint(lwrect.bottomleft())) {
		suffix = "SW";
	} else if (anchorrect.containsPoint(lwrect.bottomcenter())) {
		suffix = "S";
	} else if (anchorrect.containsPoint(lwrect.bottomright())) {
		suffix = "SE";
	}

	return "cursor-resize" + suffix;
};

/**
 * Get the annotation layer based on a point
 * @param {String|Event} either the name of the view or an event that will calculate the view of the annotation layer
 * @returns {ViewportLayer} Annotations layer of null if not found
 */
AnnotationTool.prototype._getAnnotationLayer = function(e) {
	var v = this._widget.requestView(e);
	if (v) {
		var l = v.getLayer("Annotations");
		if (l) {
			return l;
		}
	}
	return null;
};

/**
 * Get the closest linework from a given point
 * @param {Event} e - mouse event to calcuate the point to
 * @param {ViewportLayer} [layer] - if the layer was already fetched
 * @returns {LineworkWrapper} the closest lineworkwrapper to a certain point
 */
AnnotationTool.prototype._getClosestLineworkFromPoint = function(e, layer) {
	var l = layer;
	if (!l) {
		l = this._getAnnotationLayer(e);
	}

	if (l) {
		return l.getClosestLineworkFromPoint(this.getScenePoint(e));
	} else {
		return null;
	}
};

/**
 * Converts a point from an event to the world, makes sure the point is containt in the document page
 * @param {Event} e - mouse event
 * @returns {ShapeUtils.Point} point in world coordinates
 */
AnnotationTool.prototype._getWorldPoint = function(e) {
	var activePage = this._annotationModel.getAnnotationsPage();
	var worldPoint = this.keepPointWithinWorldBounds(this.getWorldPoint(e), false);

	//convert the coordinate system from world coordinates to annotation coordinates
	worldPoint.y = activePage.getPageSize().height - worldPoint.y;

	var origin = activePage.getDocumentPageOrigin();

	worldPoint.offset(new ShapeUtils.Size(-origin.x, origin.y));

	var trimboxSize = activePage.getTrimboxRect().size;
	var missingHeight = 0; //pages with wrong XMP-data (page size) can have a missing part in the height, annotationlocation needs to be compensated for this
	if (trimboxSize.width !== 0 && trimboxSize.height !== 0) {
		missingHeight = activePage.getPageSize().height - (trimboxSize.height + activePage.getMargins().top + activePage.getMargins().bottom);
	}

	var margin = activePage.getMargins();
	if (activePage.isReaderSpread() && activePage.get("isRightPage")) {
		//check if the active page is on the right side of reader spread
		margin.left = -margin.left;
	}
	worldPoint.offset(new ShapeUtils.Size(-margin.left, -margin.bottom - missingHeight));

	return worldPoint;
};

/**
 * Verify the current situation, so calcutate where the mouse is for the moment on the canvas
 * @param {Event} e - mouse event to calcuate the point to
 * @param {ViewportLayer} [layer] - if the layer was already fetched
 */
AnnotationTool.prototype._verifyObjects = function(e, layer) {
	if (!layer) {
		layer = this._getAnnotationLayer(e);
		if (!layer) {
			return;
		}
	}
	var object = this._getClosestLineworkFromPoint(e, layer);

	this._targetAnnotation = null;
	this._anchor = null;
	if (object) {
		if (object instanceof Annotation) {
			this._targetAnnotation = object;
		} else if (object.anchor) {
			this._anchor = object;
			this._targetAnnotation = object.linework;
		}
	}
	if (!object && !this._targetAnnotation) {
		this._deActivate(e);
	}
};

export default AnnotationTool;
