import _ from "underscore";
import ToolbarComponent from "@esko/toolbar/ToolbarComponent";
import ButtonGroup from "@esko/toolbar/ButtonGroup";
import ViewerToolbarButton from "@esko/web-viewer/layout/component/ViewerToolbarButton";
import SubtoolbarButton from "@esko/web-viewer/layout/component/SubtoolbarButton";
import ColorPicker from "@esko/web-viewer/layout/component/ColorPicker";
import LayoutConstants from "@esko/web-viewer/layout/LayoutConstants";
import AnnotationConstants from "@esko/web-viewer/annotation/AnnotationConstants";
/**
 * @exports AnnotationToolbarComponent
 * @memberof AnnotationModule
 * @constructor
 * @param {Core.ViewerFacade} f - Viewer facade
 * @param {AnnotationModel} annotationsModel - Annotation model
 * @param {AnnotationDrawTool} drawTool - The annotation draw tool
 * @param {AnnotationNoteDrawTool} noteDrawTool - The annotation note draw tool
 */
var AnnotationToolbarComponent = function(f, annotationsModel, drawTool, noteDrawTool) {
	ToolbarComponent.call(this, LayoutConstants.ToolbarComponent.ANNOTATION, f);
	this.f = f;
	this.annotationModel = annotationsModel;
	/**
	 * Group: annotate
	 */
	var annotateGroup = new ButtonGroup("AnnotateGroup");
	annotateGroup.setTitle(f.i18n, "viewer", "AnnBtnGroupTitle");
	this.addButtonGroup(annotateGroup);

	// Annotate ViewerToolbarButton
	var hideAnnotationsButton = new ViewerToolbarButton(
		f,
		"HideAnnotationsButton",
		f.getOption("PREFERENCES_ANNOTATIONTOOLS_VISIBLE") ? "icon-WCR_Annotation-Hide" : "icon-WCR_Annotation-Show",
		true,
		function() {
			annotationsModel.toggleVisibility();
		}
	);
	// set initial tooltip
	hideAnnotationsButton.setTooltip(
		"viewer",
		f.getOption("PREFERENCES_ANNOTATIONTOOLS_VISIBLE") ? "HideAnnotationsBtnTT" : "ShowAnnotationsBtnTT"
	);
	// react on preference changes
	f.getCurrentUser().on("options.changed.PREFERENCES_ANNOTATIONTOOLS_VISIBLE", function(oldValue, newValue) {
		if (newValue) {
			hideAnnotationsButton.setTooltip("viewer", "HideAnnotationsBtnTT");
			hideAnnotationsButton.setIcon("icon-WCR_Annotation-Hide");
		} else {
			hideAnnotationsButton.setTooltip("viewer", "ShowAnnotationsBtnTT");
			hideAnnotationsButton.setIcon("icon-WCR_Annotation-Show");
		}
		updateAnnotationButtonsUsability();
	});
	annotateGroup.addButton(hideAnnotationsButton);

	// Annotate ViewerToolbarButton
	var annotateButton = new ViewerToolbarButton(f, "AnnotateButton", "icon icon-WCR_Annotation-Global-cont icon-danger", true, function() {
		//make sure annotations are visible
		annotationsModel.toggleVisibility(true);
	});
	annotateButton.setTooltip("viewer", "AddAnnotationBtnTT");
	annotateButton.listenToTool(drawTool, true);
	annotateGroup.addButton(annotateButton);
	// update the initial icon
	if (f.getOption("PREFERENCES_ANNOTATION_TOOLMODE") != AnnotationConstants.ANNOT_PIN) {
		annotateButton.setIcon(
			AnnotationConstants.AnnotationToolIcons[f.getOption("PREFERENCES_ANNOTATION_TOOLMODE")].icon,
			AnnotationConstants.AnnotationToolIcons[f.getOption("PREFERENCES_ANNOTATION_TOOLMODE")].sub
		);
	}
	// update the icon on events
	f.getCurrentUser().on("options.changed.PREFERENCES_ANNOTATION_TOOLMODE", function(oldValue, newValue) {
		if (newValue != AnnotationConstants.ANNOT_PIN) {
			annotateButton.setIcon(AnnotationConstants.AnnotationToolIcons[newValue].icon, AnnotationConstants.AnnotationToolIcons[newValue].sub);
		}
	});

	/**
	 * Subtoolbar Component: Annotation Selector
	 */
	var annotationSelectorSubtoolbar = new ToolbarComponent("AnnotationSelectorSubtoolbar");
	annotateButton.addSubtoolbarComponent(annotationSelectorSubtoolbar);

	/**
	 * Group: rectangle + oval + line + arrow + freeform + highlight + comment
	 */
	var annotateSubGroup = new ButtonGroup("AnnotateSubGroup");
	annotationSelectorSubtoolbar.addButtonGroup(annotateSubGroup);
	var annotationColorPicker = undefined;
	var modeType = "PREFERENCES_ANNOTATION_TOOLMODE";
	// Rectangle
	var rectangleButton = new SubtoolbarButton(
		f,
		"RectangleButton",
		AnnotationConstants.AnnotationToolIcons[AnnotationConstants.AnnotationType.ANNOT_RECT].icon,
		true,
		function() {
			drawTool.setAnnotationType(AnnotationConstants.AnnotationType.ANNOT_RECT);
			if (annotationColorPicker) {
				annotationColorPicker.setColor(drawTool.getAnnotationColor());
			}
		}
	);
	rectangleButton.listenToToolOption("annotationdrawtool", modeType, AnnotationConstants.AnnotationType.ANNOT_RECT);
	rectangleButton.setTooltip("viewer", "AnnRectTT");
	rectangleButton.setIcon(
		AnnotationConstants.AnnotationToolIcons[AnnotationConstants.AnnotationType.ANNOT_RECT].icon,
		AnnotationConstants.AnnotationToolIcons[AnnotationConstants.AnnotationType.ANNOT_RECT].sub
	);
	annotateSubGroup.addButton(rectangleButton);
	// Oval
	var ovalButton = new SubtoolbarButton(f, "OvalButton", "addOval", true, function() {
		drawTool.setAnnotationType(AnnotationConstants.AnnotationType.ANNOT_ELLIPSE);
		if (annotationColorPicker) {
			annotationColorPicker.setColor(drawTool.getAnnotationColor());
		}
	});
	ovalButton.listenToToolOption("annotationdrawtool", modeType, AnnotationConstants.AnnotationType.ANNOT_ELLIPSE);
	ovalButton.setTooltip("viewer", "AnnOvalTT");
	ovalButton.setIcon(
		AnnotationConstants.AnnotationToolIcons[AnnotationConstants.AnnotationType.ANNOT_ELLIPSE].icon,
		AnnotationConstants.AnnotationToolIcons[AnnotationConstants.AnnotationType.ANNOT_ELLIPSE].sub
	);
	annotateSubGroup.addButton(ovalButton);
	// Line
	var lineButton = new SubtoolbarButton(f, "LineButton", "addLine", true, function() {
		drawTool.setAnnotationType(AnnotationConstants.AnnotationType.ANNOT_LINE);
		if (annotationColorPicker) {
			annotationColorPicker.setColor(drawTool.getAnnotationColor());
		}
	});
	lineButton.listenToToolOption("annotationdrawtool", modeType, AnnotationConstants.AnnotationType.ANNOT_LINE);
	lineButton.setTooltip("viewer", "AnnLineTT");
	lineButton.setIcon(
		AnnotationConstants.AnnotationToolIcons[AnnotationConstants.AnnotationType.ANNOT_LINE].icon,
		AnnotationConstants.AnnotationToolIcons[AnnotationConstants.AnnotationType.ANNOT_LINE].sub
	);
	annotateSubGroup.addButton(lineButton);
	// Arrow
	var arrowButton = new SubtoolbarButton(f, "ArrowButton", "addArrow", true, function() {
		drawTool.setAnnotationType(AnnotationConstants.AnnotationType.ANNOT_ARROW);
		if (annotationColorPicker) {
			annotationColorPicker.setColor(drawTool.getAnnotationColor());
		}
	});
	arrowButton.listenToToolOption("annotationdrawtool", modeType, AnnotationConstants.AnnotationType.ANNOT_ARROW);
	arrowButton.setTooltip("viewer", "AnnArrowTT");
	arrowButton.setIcon(
		AnnotationConstants.AnnotationToolIcons[AnnotationConstants.AnnotationType.ANNOT_ARROW].icon,
		AnnotationConstants.AnnotationToolIcons[AnnotationConstants.AnnotationType.ANNOT_ARROW].sub
	);

	annotateSubGroup.addButton(arrowButton);
	// Freeform
	var freeformButton = new SubtoolbarButton(f, "FreeformButton", "addFreehand", true, function() {
		drawTool.setAnnotationType(AnnotationConstants.AnnotationType.ANNOT_FREEFORM);
		if (annotationColorPicker) {
			annotationColorPicker.setColor(drawTool.getAnnotationColor());
		}
	});
	freeformButton.listenToToolOption("annotationdrawtool", modeType, AnnotationConstants.AnnotationType.ANNOT_FREEFORM);
	freeformButton.setTooltip("viewer", "AnnFreehandTT");
	freeformButton.setIcon(
		AnnotationConstants.AnnotationToolIcons[AnnotationConstants.AnnotationType.ANNOT_FREEFORM].icon,
		AnnotationConstants.AnnotationToolIcons[AnnotationConstants.AnnotationType.ANNOT_FREEFORM].sub
	);
	annotateSubGroup.addButton(freeformButton);
	// Highlight
	var highlightButton = new SubtoolbarButton(f, "HighlightButton", "addHighlight", true, function() {
		drawTool.setAnnotationType(AnnotationConstants.AnnotationType.ANNOT_HIGHLIGHT);
		if (annotationColorPicker) {
			annotationColorPicker.setColor(drawTool.getAnnotationColor());
		}
	});
	highlightButton.listenToToolOption("annotationdrawtool", modeType, AnnotationConstants.AnnotationType.ANNOT_HIGHLIGHT);
	highlightButton.setTooltip("viewer", "AnnHighlightTT");
	highlightButton.setIcon(
		AnnotationConstants.AnnotationToolIcons[AnnotationConstants.AnnotationType.ANNOT_HIGHLIGHT].icon,
		AnnotationConstants.AnnotationToolIcons[AnnotationConstants.AnnotationType.ANNOT_HIGHLIGHT].sub
	);
	annotateSubGroup.addButton(highlightButton);

	/**
	 * Other options
	 */

	var annotateSubGroupGlobAction = new ButtonGroup("AnnotateSubGroupGlobAction");
	annotationSelectorSubtoolbar.addButtonGroup(annotateSubGroupGlobAction);
	var addGlobalAnnotationButton = new SubtoolbarButton(f, "GlobalAnnotation", "addGlobalAnnotation", true, function(e) {
		drawTool.setAnnotationType(AnnotationConstants.AnnotationType.ANNOT_RECT);
		drawTool.addGlobalAnnotation(e);
	});
	addGlobalAnnotationButton.setTooltip("viewer", "AnnNewRectGlobal");
	addGlobalAnnotationButton.setIcon("icon-WCR_Annotation-Global-cont icon-danger", ["icon icon-WCR_Annotation-Global-sub icon-white"]);
	annotateSubGroupGlobAction.addButton(addGlobalAnnotationButton);

	/**
	 * Subtoolbar Component: Annotation Color selector
	 */
	if (!f.getProperty("annotationcolor")) {
		var annotateSubGroupColor = new ButtonGroup("AnnotateSubGroupColor");
		annotationSelectorSubtoolbar.addButtonGroup(annotateSubGroupColor);

		annotationColorPicker = new ColorPicker(
			f,
			function(newColor) {
				drawTool.setAnnotationColor("#" + newColor);
			},
			undefined,
			{ sahiId: "AnnotationColorPicker" }
		);

		f.getCurrentUser().on("options.changed.PREFERENCES_ANNOTATION_RECTANGLE_COLOR", function(oldValue, newValue) {
			annotationColorPicker.setColor(f.getCurrentUser().getColor("PREFERENCES_ANNOTATION_RECTANGLE_COLOR"));
		});

		annotationColorPicker.setColor(f.getCurrentUser().getColor("PREFERENCES_ANNOTATION_RECTANGLE_COLOR"));

		var colorSelectorButton = new SubtoolbarButton(f, "ColorSelector", annotationColorPicker.element, true);
		colorSelectorButton.setTooltip("viewer", "AnnColorDefaultRect");

		var colorResetButton = new SubtoolbarButton(
			f,
			"ColorReset",
			"icon icon-WCR_WF-Undo",
			true,
			_.bind(function() {
				f.resetOption("PREFERENCES_ANNOTATION_RECTANGLE_COLOR");

				this.askToChangeAnnotationColors();
			}, this)
		);
		colorResetButton.setIcon("icon icon-WCR_WF-Undo");
		colorResetButton.setTooltip("viewer", "AnnResetDefaultColor");

		annotateSubGroupColor.addButton(colorSelectorButton);
		annotateSubGroupColor.addButton(colorResetButton);
	}

	// Comment
	var commentButton = new ViewerToolbarButton(f, "CommentButton", "addComment", true, function() {
		// make sure annotations are visible
		annotationsModel.toggleVisibility(true);
	});
	commentButton.listenToTool(noteDrawTool, true);
	commentButton.setTooltip("viewer", "AnnCommNote");
	commentButton.setIcon("icon icon-WCR_Annotation-Note-sub");
	annotateGroup.addButton(commentButton);

	/**
	 * Subtoolbar Component: Annotation Color selector
	 */
	if (!f.getProperty("annotationcolor")) {
		/**
		 * Subtoolbar Component: Annotation Selector
		 */
		var annotationNoteSelectorSubtoolbar = new ToolbarComponent("AnnotationNoteSelectorSubtoolbar");
		commentButton.addSubtoolbarComponent(annotationNoteSelectorSubtoolbar);

		/**
		 * Group: rectangle + oval + line + arrow + freeform + highlight + comment
		 */
		var annotateNoteSubGroup = new ButtonGroup("AnnotateNoteSubGroup");
		annotationNoteSelectorSubtoolbar.addButtonGroup(annotateNoteSubGroup);

		var annotationNoteColorPicker = new ColorPicker(
			f,
			function(newColor) {
				noteDrawTool.setAnnotationColor("#" + newColor);
			},
			undefined,
			{ sahiId: "NoteAnnotationColorPicker" }
		);

		f.getCurrentUser().on("options.changed.PREFERENCES_ANNOTATION_NOTE_COLOR", function(oldValue, newValue) {
			annotationNoteColorPicker.setColor(noteDrawTool.getAnnotationColor());
		});
		annotationNoteColorPicker.setColor(noteDrawTool.getAnnotationColor());

		var noteColorSelectorButton = new SubtoolbarButton(f, "ColorSelector", annotationNoteColorPicker.element, true);
		noteColorSelectorButton.setTooltip("viewer", "AnnColorDefaultNote");

		var noteColorResetButton = new SubtoolbarButton(
			f,
			"ColorReset",
			"icon icon-WCR_WF-Undo",
			true,
			_.bind(function() {
				f.resetOption("PREFERENCES_ANNOTATION_NOTE_COLOR");

				this.askToChangeAnnotationColors();
			}, this)
		);
		noteColorResetButton.setTooltip("viewer", "AnnResetNoteColor");

		annotateNoteSubGroup.addButton(noteColorSelectorButton);
		annotateNoteSubGroup.addButton(noteColorResetButton);
	}

	/**
	 * Disable/Enable the buttons depending on:
	 * 	1. If the user can't change the annotations (e.g. document is approved and annotationmode is set to "readonly when approved"
	 *  2. If "hideAllAnnotations" is true
	 */
	var updateAnnotationButtonsUsability = function() {
		if (!annotationsModel.canChangeAnnotations(true)) {
			var reason = "";
			if (annotationsModel.getDisabledReason().length > 0) {
				reason = annotationsModel.getDisabledReason();
			}

			// set the tooltip for when the button is disabled
			annotateButton.setDisabledTooltip(reason);
			commentButton.setDisabledTooltip(reason);
			// disable buttons
			if (drawTool.isActive()) {
				drawTool.requestDeactivation();
			}
			if (noteDrawTool.isActive()) {
				noteDrawTool.requestDeactivation();
			}
			annotateButton.setEnabled(false);
			commentButton.setEnabled(false);
		} else {
			// the buttons should be enabled
			annotateButton.setEnabled(true);
			commentButton.setEnabled(true);
		}

		hideAnnotationsButton.setEnabled(!annotationsModel.isTemporarilyDisabledAndNoAnnotationsVisible());
	};
	updateAnnotationButtonsUsability();

	annotationsModel.on("readWriteChanged", updateAnnotationButtonsUsability);
	annotationsModel.on("temporarilyDisableChanged", updateAnnotationButtonsUsability);
	annotationsModel.on("disabledReasonChanged", updateAnnotationButtonsUsability);
};

AnnotationToolbarComponent.prototype = Object.create(ToolbarComponent.prototype);
AnnotationToolbarComponent.prototype.constructor = AnnotationToolbarComponent;
AnnotationToolbarComponent.toolbarName = LayoutConstants.ToolbarComponent.ANNOTATION;

/**
 * Ask the user if the current annotations should be updated with newly set defaults
 * @private
 */
AnnotationToolbarComponent.prototype.askToChangeAnnotationColors = function() {
	this.f.i18n.loadBundle(
		"viewer",
		_.bind(function() {
			if (confirm(this.f.i18n.t("AnnUpdateColors"))) {
				this.annotationModel.updateAllAnnotationColors();
			}
		}, this)
	);
};

export default AnnotationToolbarComponent;
