import _ from "underscore";
import AbstractTool from "@esko/web-viewer/viewport/tools/AbstractTool";
import ShapeUtils from "@esko/shapeutils";
import GEOUtils from "@esko/web-viewer/utils/GEOUtils";
import DrawUtils from "@esko/web-viewer/utils/DrawUtils";
import TouchUtil from "@esko/web-viewer/utils/TouchUtil";
import AnnotationFactory from "@esko/web-viewer/annotation/AnnotationFactory";
import LineworkPoint from "@esko/web-viewer/utils/LineworkPoint";
import AnnotationConstants from "@esko/web-viewer/annotation/AnnotationConstants";
import ViewerUtils from "@esko/web-viewer/utils/ViewerUtils";
import CoreConstants from "@esko/web-viewer/core/CoreConstants";
var DEFAULTSTATE = "default";
var DRAWSTATE = "draw";
var DRAWINGSTATE = "drawing";

/**
 * This is the tool for drawing annotations with touch interaction
 * @exports AnnotationTouchDrawTool
 * @memberof AnnotationModule
 * @constructor
 *
 * @param {Core.ViewerFacade} f
 * @param {AnnotationModel} annotationModel - the model used by the module for annotations
 * @param {AnnotationModule} annotationModule - the annotation module reference
 */
var AnnotationTouchDrawTool = function(f, annotationModel, annotationModule) {
	AbstractTool.call(this, f, CoreConstants.TOOLS.ANNOTATION_TOUCH_DRAW);

	var threshold = 0; //use this threshold to debug the freeform tool. This allows to only draw a couple of segments. When set to 0, it's harmless

	var newAnnotation = undefined;

	this.drawLinkedAnnotation = false;
	this._annotationModel = annotationModel;
	this._annotationModule = annotationModule;

	/**
	 * Set the type of annotation to draw
	 */
	this.setAnnotationType = function(type) {};

	/**
	 * Get the type of annotation to draw
	 */
	this.getAnnotationType = function() {
		return AnnotationConstants.AnnotationType.ANNOT_FREEFORM;
	};

	/**
	 * Set the annotation color to draw with
	 */
	this.setAnnotationColor = function(color) {
		f.getCurrentUser().setColor("PREFERENCES_ANNOTATION_RECTANGLE_COLOR", color);
	};

	/**
	 * Get the annotation color to draw with
	 */
	this.getAnnotationColor = function() {
		if (f.getProperty("annotationcolor")) {
			return f.getProperty("annotationcolor");
		} else {
			return f.getCurrentUser().getColor("PREFERENCES_ANNOTATION_RECTANGLE_COLOR");
		}
	};

	/**
	 * @override
	 */
	this.getDefaultCursor = function() {
		return "cursor-drawAnnotation";
	};

	/**
	 * Get the annotation layer
	 */
	var getAnnotationLayer = _.bind(function(e) {
		var v = this._widget.requestView(e);
		if (v) {
			var l = v.getLayer("Annotations");
			if (l) {
				return l;
			}
		}
		return null;
	}, this);

	/**
	 * This is the mode when the tool is active,
	 * it will zoom on click and go the couldStartDraggingMode if left mousedown is clicked
	 */
	this.registerState(DEFAULTSTATE, {
		panstart: function(e) {},
		panmove: function(e) {},
		panend: function(e) {},
		canBeInterrupted: true
	});

	/**
	 * Draw state
	 */
	this.registerState(
		DRAWSTATE,
		_.defaults(
			{
				panstart: _.bind(function(e) {
					this.requestCanvasLayer(e.pointers[0]);
					var touchPoint = this.keepPointWithinWorldBounds(this.getWorldPoint(e.pointers[0]), true);
					newAnnotation = new AnnotationFactory(f).create();
					newAnnotation.initAnnotation(
						AnnotationConstants.AnnotationType.ANNOT_FREEFORM,
						this.getAnnotationColor(),
						false,
						annotationModel
					);
					newAnnotation.rect = new ShapeUtils.Rect(touchPoint, new ShapeUtils.Size());

					this.changeState(DRAWINGSTATE);
					e.preventDefault();
				}, this)
			},
			this.getState(DEFAULTSTATE)
		)
	);

	/**
	 * Function that should be called when the drawing is finished (does correct cleanups etc)
	 */
	var doneDrawing = _.bind(function(e) {
		cleanupCanvas(); //clear the canvas

		if (!e) {
			newAnnotation = null; //if there was no event passes, put the newAnnotation to null
		}

		//If the new annotation is a line or an arrow (pointArray==2), make sure that the points are not on the same location
		if (newAnnotation && newAnnotation.pointArray.length === 2) {
			if (
				newAnnotation.pointArray[0].x == newAnnotation.pointArray[1].x &&
				newAnnotation.pointArray[0].y == newAnnotation.pointArray[1].y
			) {
				newAnnotation = null;
			}
		}

		//Make sure the annotation has at least some width or some height
		if (newAnnotation !== null && (Math.abs(newAnnotation.rect.size.width) === 0 || Math.abs(newAnnotation.rect.size.height) <= 0)) {
			//If both the width and height are 0, just return (not really useful to have a 0-sized annotation)
			newAnnotation = null;
		}

		//If we have an annotation, let's save it
		if (newAnnotation) {
			var l = getAnnotationLayer(e);
			if (l) {
				//Convert the drawed annotation from Scene (pixels) to World (mm)
				newAnnotation.rect = l.getWorldRectFromScene(newAnnotation.rect); //first the box

				if (this._annotationModel.getPage().isReaderSpread()) {
					this._annotationModule.promptIfAnnotationOutsidePage(
						newAnnotation.rect,
						_.bind(function() {
							this.createAnnotation(newAnnotation, l);
						}, this),
						function() {
							newAnnotation = null;
						}
					);
				} else {
					this.createAnnotation(newAnnotation, l);
				}
			} else {
				newAnnotation = null;
			}
		}

		this.changeState(DRAWSTATE);
	}, this);

	/**
	 * creates the annotation, makes a new annotation object, this will not save the annotation to the back-end
	 */
	this.createAnnotation = function(newAnnotation, annotationLayer) {
		for (var i = 0; i < newAnnotation.pointArray.length; i++) {
			//then the points
			var point = newAnnotation.pointArray[i];
			var p = new ShapeUtils.Point(point.x, point.y);
			var newp = annotationLayer.getWorldPointFromScene(p);
			point.x = newp.x;
			point.y = newp.y;
		}

		//Check if it need to be a linked annotation or a single new annotation
		annotationModel.addAnnotation(newAnnotation);
		this.requestDeactivation();

		newAnnotation = null;
	};

	/**
	 * Render the annotation via the draw utils
	 */
	var renderAnnotation = _.throttle(
		_.bind(function() {
			var ctx = this.getCanvasCtx();
			if (ctx) {
				// could be that the layer is already removed (caused by throttle)
				var csize = this.getCanvasSize();
				ctx.clearRect(0, 0, csize.width, csize.height);

				if (newAnnotation) {
					DrawUtils.drawOverlay(this.f, ctx, [newAnnotation]);
				}
			}
		}, this),
		40
	);

	/**
	 * Drawing state
	 */
	this.registerState(
		DRAWINGSTATE,
		_.defaults(
			{
				enterState: function() {
					f.emit("viewport.startInteractive");
					f.emit("layout.disableTextSelection");
				},
				leaveState: function() {
					f.emit("viewport.stopInteractive");
					setTimeout(function() {
						f.emit("layout.enableTextSelection");
					}, 500);
				},
				panmove: _.bind(function(e) {
					var newRect = newAnnotation.rect.copy();
					var touchPoint = this.keepPointWithinWorldBounds(this.getWorldPoint(e.pointers[0]), true);

					var point = new LineworkPoint();
					if (newAnnotation.pointArray.length === 0) {
						point.makeMoveTo(newAnnotation.rect.origin.x, newAnnotation.rect.origin.y);
					} else {
						var previousPoint = newAnnotation.pointArray[newAnnotation.pointArray.length - 1];
						if (Math.abs(previousPoint.x - touchPoint.x) < threshold || Math.abs(previousPoint.y - touchPoint.y) < threshold) {
							return;
						}
						point.makeLineTo(touchPoint.x, touchPoint.y);
					}
					newAnnotation.pointArray.push(point);

					//Check if the width needs to be updated
					var width = Math.abs(point.x - newRect.origin.x);
					if (point.x < newRect.origin.x) {
						var diff = Math.abs(newRect.origin.x - point.x);
						newRect.origin.x = point.x;
						newRect.size.width += diff;
					} else if (width > newRect.size.width) {
						newRect.size.width = width;
					}

					//Check if the height needs to be updated
					var height = Math.abs(point.y - newRect.origin.y);
					if (point.y < newRect.origin.y) {
						var diff = Math.abs(newRect.origin.y - point.y);
						newRect.origin.y = point.y;
						newRect.size.height += diff;
					} else if (height > newRect.size.height) {
						newRect.size.height = height;
					}

					newAnnotation.setRect(newRect, true);

					e.preventDefault();
					renderAnnotation();
				}, this),
				panend: function(e) {
					doneDrawing(e.pointers[0]);
				},
				canBeInterrupted: false
			},
			this.getState(DEFAULTSTATE)
		)
	);

	/**
	 * Can the tool be interrupted ?
	 */
	this.canBeInterrupted = function() {
		return false;
	};

	this.activeKeyEvents = [];
	this.persistentKeyEvents = [];

	/**
	 * Cleanup canvas
	 */
	var cleanupCanvas = _.bind(function() {
		this.releaseCanvasLayer();
	}, this);

	/**
	 * @override
	 */
	this.canBeActivated = function() {
		return annotationModel.canChangeAnnotations();
	};

	f.getCurrentUser().on(
		"options.changed.PREFERENCES_ANNOTATIONTOOLS_VISIBLE",
		_.bind(function(oldValue, newValue) {
			if (!newValue && this.isActive()) {
				this.requestDeactivation();
			}
		}, this)
	);
};

// inherit
AnnotationTouchDrawTool.prototype = Object.create(AbstractTool.prototype);
// correct the constructor pointer
AnnotationTouchDrawTool.prototype.constructor = AnnotationTouchDrawTool;

/**
 * Initialize the annotation touch draw tool
 * @override
 */
AnnotationTouchDrawTool.prototype.initialize = function(widget, toolManager) {
	AbstractTool.prototype.initialize.apply(this, arguments);
	// note this only works on the viewportwidget
	this._touchUtil = new TouchUtil(widget.getElement()[0]);
	this.changeState(DEFAULTSTATE);
};

/**
 * Activate
 * @override
 */
AnnotationTouchDrawTool.prototype.doActivation = function() {
	AbstractTool.prototype.doActivation.call(this);

	this._annotationModel.setLockSelection(true);

	this._touchUtil.addPanStartListener(
		_.bind(function(e) {
			this.getCurrentState().panstart(e);
		}, this)
	);

	this._touchUtil.addPanMoveListener(
		_.bind(function(e) {
			this.getCurrentState().panmove(e);
		}, this)
	);

	this._touchUtil.addPanEndListener(
		_.bind(function(e) {
			this.getCurrentState().panend(e);
		}, this)
	);

	/*
     * Reconsider the state of the tool
     */
	this.changeState(DRAWSTATE);
	this.drawLinkedAnnotation = false;
};

/**
 * deactivate
 * @override
 */
AnnotationTouchDrawTool.prototype.doDeactivation = function() {
	AbstractTool.prototype.doDeactivation.call(this);

	this.changeState(DEFAULTSTATE);
	this.releaseCanvasLayer();

	this._touchUtil.removeAllListeners();

	this._annotationModel.setLockSelection(false);
};

export default AnnotationTouchDrawTool;
