import _ from "underscore";
import AbstractTool from "@esko/web-viewer/viewport/tools/AbstractTool";
import ShapeUtils from "@esko/shapeutils";
import TouchUtil from "@esko/web-viewer/utils/TouchUtil";
import Annotation from "@esko/web-viewer/annotation/Annotation";
import CoreConstants from "@esko/web-viewer/core/CoreConstants";
/**
 * Construct a new Annotation touch Tool (a touch optimized version to select and move annotations)
 * @exports AnnotationTouchTool
 * @memberof AnnotationModule
 *
 * @class Tool to select and edit annotations
 * @constructor
 * @param {Core.ViewerFacade} f - Viewer facade
 * @param {AnnotationModel} annotationModel - Annotation model
 */
var AnnotationTouchTool = function(f, annotationModel) {
	AbstractTool.call(this, f, CoreConstants.TOOLS.ANNOTATION_TOUCH);

	this.f = f;

	this._annotationModel = annotationModel;

	/**
	 * Different states constants
	 */
	this._defaultState = "defaultState";
	this._annotationSelectedState = "annotationSelectedState";
	this._annotationSelectedPanningState = "annotationSelectedPanningState";

	// no annotation is selected
	this.registerState(
		this._defaultState,
		(function(that) {
			return {
				enterState: function() {
					that._annotationModel.unselectAllAnnotations();
					that._activeAnnotation = null;
				},
				panstart: function(e) {},
				panmove: function(e) {},
				panend: function(e) {},
				singletap: function(event) {
					var location = that.getScenePointFromTouch(event);
					if (location.view) {
						var annotation = that.getClosestAnnotation(location.view, location.point);

						if (annotation) {
							that._annotationModel.selectAnnotation(annotation);
						} else {
							that._annotationModel.unselectAllAnnotations();
						}
					}
				}
			};
		})(this)
	);

	// an annotation was selected
	this.registerState(
		this._annotationSelectedState,
		(function(that) {
			return _.defaults(
				{
					enterState: function() {},
					panstart: function(event) {
						var location = that.getScenePointFromTouch(event);
						if (location.view) {
							var annotation = that.getClosestAnnotation(location.view, location.point);
							if (that._activeAnnotation === annotation && that._annotationModel.canChangeAnnotation(annotation)) {
								that.changeState(that._annotationSelectedPanningState);

								that._previousPoint = that._getAnnotationLayer(location.view).getWorldPointFromScene(that.getScenePoint(event.pointers[0]));
								that.requestActivation();
							}
						}
					},
					panmove: function(e) {},
					panend: function(e) {}
				},
				that.getState(that._defaultState)
			);
		})(this)
	);

	// an annotation was selected, and a panning operation started in the selected annotation
	this.registerState(
		this._annotationSelectedPanningState,
		(function(that) {
			return {
				enterState: function() {
					that.f.emit("viewport.startInteractive");
					that.f.emit("layout.disableTextSelection");
				},
				leaveState: function() {
					if (that.isActive()) {
						that.requestDeactivation();
					}
					that.f.emit("viewport.stopInteractive");
					setTimeout(function() {
						that.f.emit("layout.enableTextSelection");
					}, 500);
				},
				panstart: function(e) {},
				panmove: function(event) {
					if (that.isActive()) {
						var view = that._widget.requestView(event.pointers[0]);
						if (view) {
							var newPoint = that._getAnnotationLayer(view).getWorldPointFromScene(that.getScenePoint(event.pointers[0]));
							var worldBounds = that._widget.getWorld().getWorldBounds(); //get the world bounds
							//get the converted worldbounds (with origin (0,0)), the origin gets added later on to not intervere with transforming the worldbounds to annotation coordinate system
							var worldRect = that
								._getAnnotationLayer(view)
								.convertWorldRectToLayer(new ShapeUtils.Rect(new ShapeUtils.Point(), worldBounds.size));
							worldRect.origin.y = -(worldRect.size.height + worldRect.origin.y) - worldBounds.origin.y; //calculate the origin in annotation coordinate system
							worldRect.origin.x += worldBounds.origin.x; //add the origin to the worldRect (this is not converted as it stays a constant)
							var moveSize = that.keepMoveSizeWithinRect(worldRect, that._activeAnnotation.rect, that._previousPoint.sizeTo(newPoint)); //keep move size within world bounds
							annotationModel.moveAnnotation(that._activeAnnotation, moveSize);
							that._previousPoint = newPoint;
						}
					} else {
						that.changeState(that._annotationSelectedState);
					}
				},
				panend: function(e) {
					that.changeState(that._annotationSelectedState);
				}
			};
		})(this)
	);

	this._annotationModel.on(
		"unselectedAnnotation",
		_.bind(function() {
			this.changeState(this._defaultState);
		}, this)
	);

	this._annotationModel.on(
		"selectedAnnotation",
		_.bind(function(annot) {
			if (annot != this._activeAnnotation) {
				this._activeAnnotation = annot;
				this.changeState(this._annotationSelectedState);
			}
		}, this)
	);
};

// inherit
AnnotationTouchTool.prototype = Object.create(AbstractTool.prototype);
// correct the constructor pointer
AnnotationTouchTool.prototype.constructor = AnnotationTouchTool;

/**
 * Get the center scene point of a TouchEvent
 * @param {TouchEvent} event - the touch event
 * @param {ViewportWidget} widget - the ViewportWidget
 * @returns {Object} {view: reference to the View, point: scenePoint}
 */
AnnotationTouchTool.prototype.getScenePointFromTouch = function(event) {
	return this._touchUtil.getScenePointFromTouch(event, this._widget);
};

/**
 * Get the annotation near the given scenepoint
 * @param {ViewportView} view - the view in which the annotation should be searched
 * @param {ShapeUtils.Point} scenePoint - the scene point where the annotation should be searched
 * @returns {Annotation} the annotation near the scenePoint or null
 */
AnnotationTouchTool.prototype.getClosestAnnotation = function(view, scenePoint) {
	var closestLinework = view.getClosestLineworkFromPoint(scenePoint, ["Annotations"], 30);
	if (closestLinework) {
		if (closestLinework instanceof Annotation) {
			return closestLinework;
		} else if (closestLinework.anchor && closestLinework.linework instanceof Annotation) {
			return closestLinework.linework;
		}
	}
	return null;
};

/**
 * @override
 */
AnnotationTouchTool.prototype.initialize = function(widget, toolManager) {
	AbstractTool.prototype.initialize.apply(this, arguments);

	this._touchUtil = new TouchUtil(widget.getElement()[0]);

	this.changeState(this._defaultState);

	this._touchUtil.addSingleTapListener(
		_.bind(function(event) {
			this.getCurrentState().singletap(event);
		}, this)
	);
	this._touchUtil.addPanStartListener(
		_.bind(function(event) {
			this.getCurrentState().panstart(event);
		}, this)
	);
	this._touchUtil.addPanMoveListener(
		_.bind(function(event) {
			this.getCurrentState().panmove(event);
		}, this)
	);
	this._touchUtil.addPanEndListener(
		_.bind(function(event) {
			this.getCurrentState().panend(event);
		}, this)
	);
};

/**
 * Get the annotation layer based on a point
 * @param {ViewportView} view
 * @returns {ViewportLayer} Annotations layer of null if not found
 */
AnnotationTouchTool.prototype._getAnnotationLayer = function(view) {
	if (view) {
		var l = view.getLayer("Annotations");
		if (l) {
			return l;
		}
	}
	return null;
};
export default AnnotationTouchTool;
