import AbstractWidget from "@esko/web-viewer/layout/AbstractWidget";
import AnnotationRow from "@esko/web-viewer/annotation/AnnotationRow";
import AnnotationOriginSelector from "@esko/web-viewer/annotation/AnnotationOriginSelector";
import AnnotationCustomFilterPopup from "@esko/web-viewer/annotation/AnnotationCustomFilterPopup";
import $ from "jquery";
import AnnotationConstants from "@esko/web-viewer/annotation/AnnotationConstants";
import ShapeUtils from "@esko/shapeutils";
import MiniButton from "@esko/web-viewer/layout/component/MiniButton";
import Button from "@esko/button/Button";
import ButtonGroup from "@esko/toolbar/ButtonGroup";
import DropdownMiniButton from "@esko/web-viewer/layout/component/DropdownMiniButton";
import LayoutConstants from "@esko/web-viewer/layout/LayoutConstants";
import ViewerToolbarButton from "@esko/web-viewer/layout/component/ViewerToolbarButton";
import _ from "underscore";
/**
 * Construct the AnnotationWidget
 * @exports AnnotationWidget
 * @memberof AnnotationModule
 * @class
 * The AnnotationWidget displays the list of annotations and gives the user options to change the way annotations are displayed. It is possible to filter, order and step through annotations. Annotation can be edited in the widget
 * @augments Layout.AbstractWidget
 * @constructor
 * @param {Core.ViewerFacade} f - Viewer facade
 * @param {AnnotationModule} annotationModule
 * @param {AnnotationModel} annotationModel - The annotation model
 * @param {String} layoutType - the layout type the widget needs to be rendered for
 */
var AnnotationWidget = function(f, annotationModule, annotationModel, layoutType) {
	AbstractWidget.call(this, f, LayoutConstants.Widget.ANNOTATION);

	this.f = f;
	this.annotationModel = annotationModel;
	this.annotationModule = annotationModule;
	/**
	 * In this hash set all annotationRow objects are stored
	 */
	this.annotationRows = {};

	this._isForMobile = layoutType === LayoutConstants.LayoutType.MOBILE;

	/**
	 * Details
	 */
	this.detailsDiv = $("<div>").addClass("annotationDetails enableTextSelection");

	this.nonScrollablePart = $("<div>").addClass("nonScrollablePart");

	this._annotationOriginSelector = new AnnotationOriginSelector(
		f,
		annotationModel.getAnnotationFilterModel(),
		annotationModel,
		this.detailsDiv,
		this._isForMobile
	);
	this.nonScrollablePart.append(this._annotationOriginSelector.element);
	this.detailsDiv.append(this.nonScrollablePart);

	this.warningMessage = $("<div/>")
		.addClass("alert")
		.text(this.annotationModel.getDisabledReason())
		.toggle(this.annotationModel.getDisabledReason().length > 0);
	this.nonScrollablePart.append(this.warningMessage);

	// this div is shown when annotations are temporarily disabled
	this.temporarilyDisabledDiv = $("<div/>")
		.addClass("annotationsTemporarilyDisabled")
		.hide();
	this.nonScrollablePart.append(this.temporarilyDisabledDiv);

	// note i18n is done a bit later, I wanted the references to be available already (even if te language is not loaded yet)
	this.showCurrentFileAnnotationsButton = new Button(
		this.f.i18n,
		this.f.getContext(),
		this.f.isMobile(),
		"currentDocument",
		$("<span/>")
			.text("current")
			.prepend($("<i>").addClass("icon icon-WCR_Compare-Current")),
		true,
		_.bind(function() {
			this.annotationModule.toggleCurrentPageAnnotations();
		}, this)
	);

	this.showCurrentFileAnnotationsButton.element.addClass("btn");
	// note i18n is done a bit later, I wanted the references to be available already (even if te language is not loaded yet)
	this.showDifferenceFileAnnotationsButton = new Button(
		this.f.i18n,
		this.f.getContext(),
		this.f.isMobile(),
		"differenceDocument",
		$("<span/>")
			.text("reference")
			.prepend($("<i>").addClass("icon icon-WCR_Compare-Reference")),
		true,
		_.bind(function() {
			this.annotationModule.toggleDifferencePageAnnotations();
		}, this)
	);
	this.showDifferenceFileAnnotationsButton.element.addClass("btn");

	this.annotationToggleGroup = new ButtonGroup("annotationToggleBetweenCurrentAndDifference", [
		this.showCurrentFileAnnotationsButton,
		this.showDifferenceFileAnnotationsButton
	]);
	this.temporaryAnnotationToggler = $("<div/>").addClass("temporaryAnnotationToggle");

	// i18n for temporarily disabling
	f.i18n.loadBundle(
		"viewer",
		function() {
			this.temporaryAnnotationToggler
				.append($("<span/>").text(this.f.t("AnnShowAnnotationsFor") + ": "))
				.append($("<br/>"))
				.append(this.annotationToggleGroup.element);

			this.showCurrentFileAnnotationsButton.setContent(
				$("<span/>")
					.text(this.f.t("CaptionCurrent"))
					.prepend($("<i>").addClass("icon icon-WCR_Compare-Current"))
			);

			this.showDifferenceFileAnnotationsButton.setContent(
				$("<span/>")
					.text(this.f.t("CaptionPrevious"))
					.prepend($("<i>").addClass("icon icon-WCR_Compare-Reference"))
			);
		},
		this
	);
	this.temporarilyDisabledDiv.append(this.temporaryAnnotationToggler);
	this.temporaryAnnotationToggler.hide();

	// filter
	this.header = $("<div>").addClass("annotationFilter");
	this._annotationFilters = $("<div>").addClass("annotationFilterSelectors");
	this._customFilters = $("<div>").addClass("customFilterControls");
	this.header.append(this._annotationFilters);
	this._numberOfSelectedThreads = $("<span/>")
		.addClass("threadsCount")
		.attr("data-sahi-id", "annotations-threadsCount");
	this.header.append(this._customFilters);
	this._customFilters.append(this._numberOfSelectedThreads);

	this.bindKeys([
		[
			["esc"],
			_.bind(function(e) {
				e.preventDefault();
				e.stopPropagation();
				if (this.annotationModel) {
					this.annotationModel.unselectAllAnnotations();
				}
			}, this),
			"keydown"
		]
	]);

	this.annotationModel.getAnnotationFilterModel().on(
		"filterStateChanged",
		_.bind(function() {
			if (this.annotationModel.getAnnotationFilterModel().getOriginType() !== AnnotationConstants.AnnotationOriginTypes.user.id) {
				if (this._annotationFilters) {
					this._annotationFilters.hide();
				}
				if (this._pinFilterButton) {
					this._pinFilterButton.element.hide();
				}
				if (this._openFilterPopupButton) {
					this._openFilterPopupButton.element.hide();
				}
				if (this._duplicateFilteredAnnotationsButton) {
					if (
						_.all(this.annotationModel.getFilteredAnnotations(), function(annotation) {
							return annotation.getPermissionFor(AnnotationConstants.ActionPermissions.DUPLICATE);
						})
					) {
						this._duplicateFilteredAnnotationsButton.element.show();
					} else {
						this._duplicateFilteredAnnotationsButton.element.hide();
					}
				}
				if (this._copyFilteredAnnotationsWithoutText) {
					this._copyFilteredAnnotationsWithoutText.element.show();
				}
			} else {
				if (this._annotationFilters) {
					this._annotationFilters.show();
				}
				if (this._pinFilterButton) {
					this._pinFilterButton.element.show();
				}
				if (this._openFilterPopupButton) {
					this._openFilterPopupButton.element.show();
				}
				if (this._duplicateFilteredAnnotationsButton) {
					this._duplicateFilteredAnnotationsButton.element.hide();
				}
				if (this._copyFilteredAnnotationsWithoutText) {
					this._copyFilteredAnnotationsWithoutText.element.hide();
				}
			}
		}, this)
	);

	if (!this._isForMobile) {
		this._customFilterPopup = new AnnotationCustomFilterPopup(this.f, "customFilter", this.annotationModel);

		this.annotationModel.getAnnotationFilterModel().on(
			"filterReset",
			_.bind(function() {
				if (this._customFilterPopup.isVisible()) {
					this._customFilterPopup.hide();
				}
			}, this)
		);

		//add a button to pin the annotation filter
		this._pinFilterButton = new MiniButton(
			this.f,
			"pinFilter",
			"icon-WCR_Pin12",
			true,
			_.bind(function() {
				this.annotationModel.getAnnotationFilterModel().toggleAnnotationPinnedState();
				this._updateFilterPinnedIcon();
			}, this)
		);
		this._updateFilterPinnedIcon();
		this._pinFilterButton.element.addClass("pinAnnotationFilter");

		//add a button to copy without text all filtered annotations
		this._copyFilteredAnnotationsWithoutText = new MiniButton(
			this.f,
			"duplicateAllFilteredAnnotationWithoutText",
			" icon icon-WCR_Annotation-Duplicate-NoText12",
			true,
			_.bind(function() {
				this.f.emit("module.annotation.duplicateAllFilteredAnnotationWithoutText");
			}, this)
		);

		this._copyFilteredAnnotationsWithoutText.setTooltip("viewer", "duplicateAllFilteredAnnotationWithoutText");
		this._copyFilteredAnnotationsWithoutText.element.addClass("pinAnnotationFilter");
		this._copyFilteredAnnotationsWithoutText.element.hide();

		//add a button to copy all filtered annotations
		this._duplicateFilteredAnnotationsButton = new MiniButton(
			this.f,
			"duplicateAllFiltered",
			"icon icon-WCR_Annotation-Duplicate12",
			true,
			_.bind(function() {
				this.f.emit("module.annotation.duplicateAllFilteredAnnotations");
			}, this)
		);
		this._duplicateFilteredAnnotationsButton.setTooltip("viewer", "duplicateAllFilteredAnnotations");
		this._duplicateFilteredAnnotationsButton.element.addClass("pinAnnotationFilter");
		this._duplicateFilteredAnnotationsButton.element.hide();

		this._openFilterPopupButton = new MiniButton(
			this.f,
			"customFilter",
			"icon-WCR_Filter12",
			true,
			_.bind(function() {
				var left = this.context.offset().left < 250 ? this.context.width() || null : this.context.offset().left - 250;
				var location = new ShapeUtils.Point(left, this.context.offset().top);
				this._customFilterPopup.show(location);
			}, this)
		);
		this._openFilterPopupButton.element.addClass("customFilterBtn");
		this._openFilterPopupButton.setTooltip("viewer", "OpenCustomFilterPopup");

		this._customFilters.append(this._copyFilteredAnnotationsWithoutText.element);
		this._customFilters.append(this._duplicateFilteredAnnotationsButton.element);
	}

	// the list wrapper is used to only have a scrollbar for the annotationlist and not for the filter
	this.annotationsListWrapper = $("<div>").addClass("annotationListWrapper");
	this.annotationsListWrapper.on(
		"makeMeVisible",
		_.partial(function(that, e) {
			// add support for makeMeVisible trigger on the scrollable pane

			// element
			var et = $(e.target).offset().top;
			var eh = $(e.target).outerHeight() || null;
			// content
			var ct = $(this).offset().top;
			var ch = $(this).outerHeight() || null;

			// check whether we need to scroll
			if (et - ct < 0 || et + eh > ct + ch) {
				var top = $(this).scrollTop() + et - ct;
				$(this).scrollTop(top);
			}

			that._visibleItem = $(e.target);

			e.preventDefault();
			e.stopPropagation();
		}, this)
	);

	if (!this._isForMobile) {
		this._customFilters.append(this._openFilterPopupButton.element);
		this._customFilters.append(this._pinFilterButton.element);
	}

	this.detailsDiv.append(this.annotationsListWrapper);

	this.annotationList = $("<div>").addClass("annotationList");

	if (this._isForMobile) {
		this.annotationsListWrapper.append(this.header);
	} else {
		this.nonScrollablePart.append(this.header);
	}
	this.annotationsListWrapper.append(this.annotationList);

	f.i18n.loadBundle(
		"viewer",
		function() {
			this.setTitle(this.f.i18n.t("AnnWidgetTitle"));
		},
		this
	);

	this.setIcon("icon icon-WCR_Viewer-Annotation-List");

	/**
	 * In here for every annotation row that is in the DOM is noted
	 * @private
	 */
	this.inDOM = {};
	/**
	 * In this object all elements that should be removed from the DOM are noted
	 * @private
	 */
	this.toBeRemovedFromDOM = {};

	/**
	 * The button in the header to save the annotation
	 * @private
	 */
	this.saveButton = null;
	/**
	 * Should the button be disabled?
	 * @private
	 */
	this.disableSaveButton = false;

	this.annotationModel.on(
		"disabledReasonChanged",
		_.bind(function() {
			this.warningMessage.text(this.annotationModel.getDisabledReason());
			this.warningMessage.toggle(this.annotationModel.getDisabledReason().length > 0);
			this.resized();
		}, this)
	);
	/**
	 * When annotations are saved, update the save button
	 */
	this.annotationModel.on(
		"save.finished",
		_.bind(function() {
			this.disableSaveButton = false;
			this.updateSaveButton();
			this.reloadButton.setEnabled(true);
		}, this)
	);

	/**
	 * When the annotation has changed, update all rows
	 */
	this.annotationModel.on(
		"readWriteChanged",
		_.bind(function() {
			_.each(this.annotationRows, function(row) {
				row.annotationUpdated();
			});
		}, this)
	);

	/**
	 * Trigger a zoom when an annotation is selected
	 */
	this.annotationModel.on("selectedAnnotation", _.bind(this._fitViewToAnnotation, this));

	/**
	 * Trigger a zoom when "Fit View to Selected Annotation" setting has changed
	 */
	this.annotationModel.on(
		"fitViewToSelectedAnnotationChanged",
		_.bind(function() {
			var selectedAnnotations = this.annotationModel.getSelectedAnnotations();
			if (selectedAnnotations && selectedAnnotations.length > 0) {
				this._fitViewToAnnotation(selectedAnnotations[0]);
			}
		}, this)
	);

	/**
	 * Save annotations button
	 */
	this.saveButton = new MiniButton(
		f,
		"saveAnnotations",
		"icon-WCR_Save",
		true,
		_.bind(function() {
			// start saving annotations
			this.annotationModel.save();
		}, this),
		"fontIconDisabled icon-WCR_Save"
	);

	this.updateSaveButton();

	/**
	 * Step through annotation sbutton
	 */
	this.stepThroughButton = new MiniButton(
		f,
		"stepThroughAnn",
		"icon-WCR_Step",
		true,
		_.bind(function(e) {
			if (e.shiftKey) {
				//check if the shift key is pressed
				this.annotationModel.goToAnnotation(false);
			} else {
				this.annotationModel.goToAnnotation(true);
			}
		}, this)
	);
	this.stepThroughButton.setDisabledIcon("icon-WCR_Step fontIconDisabled");
	// on right click, step through backwards
	this.stepThroughButton.element.on(
		"contextmenu",
		_.bind(function(e) {
			e.preventDefault();
			if (this.stepThroughButton.isEnabled()) {
				this.annotationModel.goToAnnotation(false);
			}
		}, this)
	);
	this.stepThroughButton.setTooltip("viewer", "StepThroughAnnTT");

	/**
	 * Reload annotations button
	 */
	this.reloadButton = new MiniButton(
		f,
		"reloadAnnotations",
		"icon-WCR_Refresh",
		true,
		_.bind(function() {
			this.annotationModel.reload();
		}, this),
		"icon-WCR_Refresh fontIconDisabled"
	);
	this.reloadButton.setTooltip("viewer", "ReloadMiniBtnTT");

	// while saving, disable reload button
	this.annotationModel.on(
		"save.started",
		_.bind(function() {
			this.reloadButton.setEnabled(false);
			this.disableSaveButton = true;
			this.updateSaveButton();
		}, this)
	);

	// while reloading, disable reload button
	this.annotationModel.on(
		"reload.started",
		_.bind(function() {
			this.reloadButton.setEnabled(false);
		}, this)
	);
	this.annotationModel.on(
		"reload.finished",
		_.bind(function() {
			this.reloadButton.setEnabled(true);
		}, this)
	);

	this.annotationModel.on("temporarilyDisableChanged", _.bind(this.updateListVisibility, this));
	this.annotationModel.on("disabledReasonChanged", _.bind(this.updateListVisibility, this));
	this.updateListVisibility();

	/**
	 * Listen to user settings
	 */

	// set sort by mode
	f.getCurrentUser().on(
		"options.changed.PREFERENCES_ANNOTATION_DATESORT_MODE",
		_.bind(function(oldValue, newValue) {
			this.setSortByMode(newValue);
		}, this)
	);

	// step through mode
	f.getCurrentUser().on(
		"options.changed.PREFERENCES_ANNOTATION_NEXT_MODE",
		_.bind(function(oldValue, newValue) {
			this.setStepThroughMode(newValue);
		}, this)
	);

	// show only selected annotation
	f.getCurrentUser().on(
		"options.changed.PREFERENCES_ANNOTATION_SHOWSELECTEDONLY",
		_.bind(function(oldValue, newValue) {
			this.setShowSelectedAnnotationOnly(newValue > 0);
		}, this)
	);

	// show only selected annotation popup
	f.getCurrentUser().on(
		"options.changed.PREFERENCES_ANNOTATION_SHOWSELECTEDPOPUPONLY",
		_.bind(function(oldValue, newValue) {
			this.setShowSelectedAnnotationPopupOnly(newValue > 0);
		}, this)
	);

	// show sequence label
	f.getCurrentUser().on(
		"options.changed.PREFERENCES_ANNOTATION_SEQLABEL_VISIBLE",
		_.bind(function(oldValue, newValue) {
			this.setShowSequenceLabel(newValue > 0);
		}, this)
	);

	// show char/line counter
	f.getCurrentUser().on(
		"options.changed.PREFERENCES_ANNOTATION_SHOWCOUNT",
		_.bind(function(oldValue, newValue) {
			this.setShowCharLineCounter(newValue > 0);
		}, this)
	);

	/**
	 * Make the options dropdown
	 */
	this.optionButton = new DropdownMiniButton(f, "annotationSettings", "icon-WCR_Settings-submenu");

	f.i18n.loadBundle(
		"viewer",
		function() {
			// step through mode
			function toggleStepThrougMode(mode, toggle) {
				if (toggle) {
					this.setStepThroughMode(mode, true);
				}
			}
			this.optionButton.addOption({
				name: "stepByThread",
				content: this.f.i18n.t("AnnSortThread"),
				onchange: _.bind(toggleStepThrougMode, this, AnnotationConstants.StepThroughMode.BY_THREAD),
				selected: this.f.getOption("PREFERENCES_ANNOTATION_NEXT_MODE") == AnnotationConstants.StepThroughMode.BY_THREAD,
				userCannotUncheck: true
			});

			this.optionButton.addOption({
				name: "stepWithReplies",
				content: this.f.i18n.t("AnnSortDate"),
				onchange: _.bind(toggleStepThrougMode, this, AnnotationConstants.StepThroughMode.WITH_REPLIES),
				selected: this.f.getOption("PREFERENCES_ANNOTATION_NEXT_MODE") == AnnotationConstants.StepThroughMode.WITH_REPLIES,
				userCannotUncheck: true
			});

			this.optionButton.addOption({
				name: "stepWithZoom",
				content: this.f.i18n.t("AnnSortThreadFit"),
				onchange: _.bind(toggleStepThrougMode, this, AnnotationConstants.StepThroughMode.WITH_ZOOM),
				selected: this.f.getOption("PREFERENCES_ANNOTATION_NEXT_MODE") == AnnotationConstants.StepThroughMode.WITH_ZOOM,
				userCannotUncheck: true
			});

			this.optionButton.addOption({
				name: "stepWithLinked",
				content: this.f.i18n.t("AnnSortAreaFit"),
				onchange: _.bind(toggleStepThrougMode, this, AnnotationConstants.StepThroughMode.WITH_LINKED),
				selected: this.f.getOption("PREFERENCES_ANNOTATION_NEXT_MODE") == AnnotationConstants.StepThroughMode.WITH_LINKED,
				userCannotUncheck: true
			});

			this.optionButton.addDivider();

			// show selected annotation only
			this.optionButton.addOption({
				name: "showselann",
				content: this.f.i18n.t("AnnListShowSel"),
				onchange: _.bind(function(toggle) {
					this.setShowSelectedAnnotationOnly(toggle, true);
				}, this),
				selected: this.f.getOption("PREFERENCES_ANNOTATION_SHOWSELECTEDONLY") > 0
			});

			// show selected annotation popup only
			this.optionButton.addOption({
				name: "showselannpopup",
				content: this.f.i18n.t("AnnListShowSelPopup"),
				onchange: _.bind(function(toggle) {
					this.setShowSelectedAnnotationPopupOnly(toggle, true);
				}, this),
				selected: this.f.getOption("PREFERENCES_ANNOTATION_SHOWSELECTEDPOPUPONLY") > 0
			});

			this.optionButton.addDivider();

			// show all annotation popups
			this.optionButton.addOption({
				name: "showAllAnnotationPopups",
				content: this.f.i18n.t("AnnShowAllAnnotationPopups"),
				onchange: _.bind(function(toggle) {
					this.annotationModel.setShowAllAnnotationPopups(toggle);
					this.f.setOption("PREFERENCES_SHOW_ALL_ANNOTATION_POPUPS", toggle ? 1 : 0);
				}, this),
				selected: this.f.getOption("PREFERENCES_SHOW_ALL_ANNOTATION_POPUPS") > 0
			});

			// show sequence number
			this.optionButton.addOption({
				name: "shownumber",
				content: this.f.i18n.t("AnnSeqLabelShow"),
				onchange: _.bind(function(toggle) {
					this.setShowSequenceLabel(toggle, true);
				}, this),
				selected: this.f.getOption("PREFERENCES_ANNOTATION_SEQLABEL_VISIBLE") > 0
			});

			// show char/line counter
			this.optionButton.addOption({
				name: "showCharLineCount",
				content: this.f.i18n.t("AnnCountShow"),
				onchange: _.bind(function(toggle) {
					this.setShowCharLineCounter(toggle, true);
				}, this),
				selected: this.f.getOption("PREFERENCES_ANNOTATION_SHOWCOUNT") > 0
			});

			// auto fit view to selected annotation
			this.optionButton.addOption({
				name: "autoFitViewToSelectedAnnot",
				content: this.f.i18n.t("AnnListFitView"),
				onchange: _.bind(function(toggle) {
					this.annotationModel.setAutoFitViewToSelectedAnnotation(toggle);
					this.f.setOption("PREFERENCES_ANNOTATION_FITVIEW", toggle ? 1 : 0);
				}, this),
				selected: this.f.getOption("PREFERENCES_ANNOTATION_FITVIEW") > 0
			});

			this.optionButton.addDivider();

			// sort by options
			function toggleSortByMode(mode, toggle) {
				if (toggle) {
					this.setSortByMode(mode, true);
				}
			}
			this.optionButton.addOption({
				name: "sortbycreationdate",
				content: this.f.i18n.t("DateSortCreation"),
				onchange: _.bind(toggleSortByMode, this, AnnotationConstants.AnnotationSortType.SORT_BY_CREATION_DATE),
				selected: this.f.getOption("PREFERENCES_ANNOTATION_DATESORT_MODE") === AnnotationConstants.AnnotationSortType.SORT_BY_CREATION_DATE,
				userCannotUncheck: true
			});

			this.optionButton.addOption({
				name: "sortbymodificationdate",
				content: this.f.i18n.t("DateSortModification"),
				onchange: _.bind(toggleSortByMode, this, AnnotationConstants.AnnotationSortType.SORT_BY_MODIFICATION_DATE),
				selected:
					this.f.getOption("PREFERENCES_ANNOTATION_DATESORT_MODE") === AnnotationConstants.AnnotationSortType.SORT_BY_MODIFICATION_DATE,
				userCannotUncheck: true
			});

			// make sure the show selected annotation popup only is disable when show selected annotation only is true
			this.setShowSelectedAnnotationOnly(this.f.getOption("PREFERENCES_ANNOTATION_SHOWSELECTEDONLY") > 0, false);
			this.updateWarningBadge();
		},
		this
	);

	// Add callbacks for moving: find all rich text editors and re-init them
	this.on(
		"movingStarted",
		_.bind(function() {
			if (this.context) {
				this.context.find("[id^=richText_]").each(function() {
					var richTextObj = $(this).data("richTextObj");
					// Only destroy the already initialized ones
					if (richTextObj && richTextObj.isInitialized()) {
						richTextObj.destroy();
						richTextObj.isDestroyed = true;
					}
				});
			}
		}, this)
	);

	this.on(
		"movingStopped",
		_.bind(function() {
			if (this.context) {
				this.context.find("[id^=richText_]").each(function() {
					var richTextObj = $(this).data("richTextObj");
					// Only re-init the ones we destroyed ourselves.
					if (richTextObj && richTextObj.isDestroyed) {
						_.defer(function() {
							richTextObj.init();
						});
					}
				});
			}
		}, this)
	);

	/**
	 * Widget visibility
	 */
	this.on("visible", function(isVisible) {
		annotationModel.setAnnotationListVisible(isVisible);
	});
	annotationModel.setAnnotationListVisible(this.isVisible());

	annotationModel.on("selectedAnnotation", _.bind(this.selectAnnotation, this));
	annotationModel.on("unselectedAnnotation", _.bind(this.deselectAnnotation, this));
	annotationModel.on(
		"annotationAdded",
		_.bind(function(annotation, hidden) {
			if (!hidden) {
				this.addAnnotation(annotation, true);
			}
		}, this)
	);
	annotationModel.on("annotationUpdated", _.bind(this.updateAnnotation, this));
	annotationModel.on("annotationDeleted", _.bind(this.deleteAnnotation, this));

	/**
	 * Render the complete widget
	 */
	var renderWidget = _.bind(function() {
		this._updateButtons();
		this.renderCompleteWidget();
	}, this);

	/**
	 * Update both save and step through buttons
	 */
	var updateButtons = _.bind(function() {
		this._updateButtons();
	}, this);

	/**
	 * The widget gets initialized with its context
	 */
	this._override("setContext", function(context, headerExtra) {
		// call the old implementation of activate() (the implementation of AbstractWidget.js)
		this._super(context, headerExtra);

		//render the widget
		this.renderCompleteWidget();
		this.context.append(this.detailsDiv);

		// add buttons to the widget header
		this.headerExtra
			.append(this.saveButton.element)
			.append(this.stepThroughButton.element)
			.append(this.reloadButton.element)
			.append(this.optionButton.element);
		this._updateButtons();

		// start listening to the model
		this.annotationModel.on("filterChanged", renderWidget);
		this.annotationModel.on("modelContentChanged", renderWidget);
		this.annotationModel.on("modelChanged", updateButtons);
		this.annotationModel.on(
			"changed.page",
			_.bind(function() {
				// when a page changes, make sure to remove all DOM elements + clear the cache of AnnotationRow objects
				var toBeRemovedFromDOM = _.keys(this.inDOM);
				_.each(
					toBeRemovedFromDOM,
					function(name) {
						this._removeFromDOM(name);
					},
					this
				);
				this.annotationRows = {};
			}, this)
		);
		// heart beat for the timers (note PJPR, would be nice if this isn't managed by the widget, but by the autotimelabels themselves)
		this.startHeartBeat();

		// adjust the layout of the widget
		this.resized();
	});

	//Dectivation of the widget
	this._override("destroy", function() {
		// call the old implementation of activate() (the implementation of AbstractWidget.js)
		this._super();

		annotationModel.off("filterChanged", renderWidget);
		annotationModel.off("modelContentChanged", renderWidget);
		annotationModel.off("modelChanged", updateButtons);
		this.stopHeartBeat();
	});
};

AnnotationWidget.prototype = Object.create(AbstractWidget.prototype);
AnnotationWidget.prototype.constructor = AnnotationWidget;
AnnotationWidget.widgetName = LayoutConstants.Widget.ANNOTATION;

/**
 * Get the minimum size of the annotation details widget
 * @override
 * @returns {ShapeUtils.Size}
 */
AnnotationWidget.prototype.getMinimumSize = function() {
	return new ShapeUtils.Size(250, 150);
};

/**
 * The annotation widget is greedy, and wants to have all available extra space
 * @override
 * @returns {Boolean} true
 */
AnnotationWidget.prototype.willFill = function() {
	return true;
};

/**
 * Updates the save and step through buttons
 */
AnnotationWidget.prototype._updateButtons = function() {
	this.updateSaveButton();
	this.updateStepThroughButton(true);
};

/**
 * Update the save button display (disable/enable the button, change badge, change tooltip
 * @private
 */
AnnotationWidget.prototype.updateSaveButton = function() {
	if (this.saveButton) {
		if (this.disableSaveButton) {
			// the button is disabled
			this.saveButton.setEnabled(false);
		} else {
			// get the amount of unsaved annotations
			var unsavedCount = this.annotationModel.getUnsavedAnnotationCount();
			if (unsavedCount > 0) {
				this.saveButton.setEnabled(true);
				this.saveButton.setBadge(unsavedCount);
				if (this.annotationModel.getUnsavedAnnotations().length > 0) {
					if (this.annotationModel.getDeletedAnnotations().length > 0) {
						this.saveButton.setTooltip("viewer", "SaveMiniBtnTT", [
							this.annotationModel.getUnsavedAnnotations().length,
							this.annotationModel.getDeletedAnnotations().length
						]);
					} else {
						this.saveButton.setTooltip("viewer", "SaveMiniBtnAnnTT", [this.annotationModel.getUnsavedAnnotations().length]);
					}
				} else {
					this.saveButton.setTooltip("viewer", "SaveMiniBtnDelTT", [this.annotationModel.getDeletedAnnotations().length]);
				}
			} else {
				this.saveButton.setEnabled(false);
			}
		}
	}
};

/**
 * Show or hide the warning badge, update the tool tip
 */
AnnotationWidget.prototype.updateWarningBadge = function() {
	var showSelAnn = this.optionButton.getOptionValue("showselann");
	var showSelAnnPopup = this.optionButton.getOptionValue("showselannpopup");
	if (showSelAnn || showSelAnnPopup) {
		// show the warning badge
		this.optionButton.setWarning();
		this.f.i18n.loadBundle(
			"viewer",
			function() {
				// build up the tooltip text
				var tooltip = this.f.i18n.t("AnnWidgetOptionsTT") + ".";
				tooltip += "<br />" + this.f.i18n.t("Warning");
				if (showSelAnn) {
					tooltip += "  " + this.f.i18n.t("OnlySelectedAnnotationWarningTT");
				}
				if (showSelAnnPopup && this.optionButton.isActionEnabled("showselannpopup")) {
					if (showSelAnn) {
						// should not happen
						tooltip += "<br />";
					}
					tooltip += "  " + this.f.i18n.t("OnlySelectedAnnotationPopupWarningTT");
				}
				this.optionButton.setTooltip(tooltip);
			},
			this
		);
	} else {
		// hide warning badge and restore tooltip
		this.optionButton.removeWarning();
		this.optionButton.setTooltip("viewer", "AnnWidgetOptionsTT");
	}
};

/**
 * update the step through button, makes sure the button in disabled when there are no annotations
 * @param {Boolean} enable - should the button be enabled when there are annotations
 */
AnnotationWidget.prototype.updateStepThroughButton = function(enable) {
	if (_.size(this.annotationModel.getFilteredAnnotations()) === 0 || !enable) {
		this.stepThroughButton.setEnabled(false);
	} else {
		this.stepThroughButton.setEnabled(true);
	}
};

/**
 * Set the sort mode
 * @param {Number} sorting mode - toggle the option
 * @param {Boolean} [persist] - true if option needs to be persisted
 */
AnnotationWidget.prototype.setSortByMode = function(sortMode, persist) {
	if (persist) {
		this.f.setOption("PREFERENCES_ANNOTATION_DATESORT_MODE", sortMode);
	}

	this.optionButton.toggleOption("sortbycreationdate", sortMode == AnnotationConstants.AnnotationSortType.SORT_BY_CREATION_DATE);
	this.optionButton.toggleOption("sortbymodificationdate", sortMode == AnnotationConstants.AnnotationSortType.SORT_BY_MODIFICATION_DATE);
};

/**
 * Enable the option to show only the selected annotation
 * @param {Boolean} toggle - toggle the option
 * @param {Boolean} [persist] - true if option needs to be persisted
 */
AnnotationWidget.prototype.setShowSelectedAnnotationOnly = function(toggle, persist) {
	if (persist) {
		this.f.setOption("PREFERENCES_ANNOTATION_SHOWSELECTEDONLY", toggle ? 1 : 0);
	}

	// if "show only selected annotation" is enabled, disable the "show only selected annotation popup" option
	if (toggle) {
		this.optionButton.toggleOption("showselannpopup", false);
		this.f.setOption("PREFERENCES_ANNOTATION_SHOWSELECTEDPOPUPONLY", 0);
	}

	// display warning if needed
	this.updateWarningBadge();
};

/**
 * Enable the option to show only the selected annotation popup
 * @param {Boolean} toggle - toggle the option
 * @param {Boolean} [persist] - true if option needs to be persisted
 */
AnnotationWidget.prototype.setShowSelectedAnnotationPopupOnly = function(toggle, persist) {
	if (persist) {
		this.f.setOption("PREFERENCES_ANNOTATION_SHOWSELECTEDPOPUPONLY", toggle ? 1 : 0);
	}

	// if "show selected annotation popup only" is enabled, disable "show only selected annotation"
	if (toggle) {
		this.optionButton.toggleOption("showselann", false);
		this.f.setOption("PREFERENCES_ANNOTATION_SHOWSELECTEDONLY", 0);
	}

	// display warning if needed
	this.updateWarningBadge();
};

/**
 * Set the step through mode
 * @param {Number} mode - set AnnotationConstants.StepTroughModemode
 * @param {Boolean} [persist] -  true if option needs to be persisted
 */
AnnotationWidget.prototype.setStepThroughMode = function(mode, persist) {
	if (persist) {
		this.f.setOption("PREFERENCES_ANNOTATION_NEXT_MODE", mode);
	}

	// update toggles
	this.optionButton.toggleOption("stepWithReplies", mode == AnnotationConstants.StepThroughMode.WITH_REPLIES);
	this.optionButton.toggleOption("stepByThread", mode == AnnotationConstants.StepThroughMode.BY_THREAD);
	this.optionButton.toggleOption("stepWithZoom", mode == AnnotationConstants.StepThroughMode.WITH_ZOOM);
	this.optionButton.toggleOption("stepWithLinked", mode == AnnotationConstants.StepThroughMode.WITH_LINKED);
};

/**
 * Enable the option to show the sequence label
 * @param {Boolean} toggle - toggle the option
 * @param {Boolean} [persist] - true if option needs to be persisted
 */
AnnotationWidget.prototype.setShowSequenceLabel = function(toggle, persist) {
	if (persist) {
		this.f.setOption("PREFERENCES_ANNOTATION_SEQLABEL_VISIBLE", toggle ? 1 : 0);
	}

	this.optionButton.toggleOption("shownumber", toggle);
};

/**
 * Enable the option to show the char/line counter
 * @param {Boolean} toggle - toggle the option
 * @param {Boolean} [persist] - true if option needs to be persisted
 */
AnnotationWidget.prototype.setShowCharLineCounter = function(toggle, persist) {
	if (persist) {
		this.f.setOption("PREFERENCES_ANNOTATION_SHOWCOUNT", toggle ? 1 : 0);
	}

	this.optionButton.toggleOption("showCharLineCount", toggle);

	_.each(this.annotationRows, function(annotationRow) {
		annotationRow.showHideCharLineCount(toggle, true);
	});
};

/**
 * Render all the interface element of the widget
 */
AnnotationWidget.prototype.renderCompleteWidget = function() {
	if (!this._isDirty) {
		this._isDirty = true;
		this.f.i18n.loadBundle(
			"viewer",
			function() {
				// defer makes the priority of the full redraw lower (all current events on the stack are first processed before redraw begins
				_.defer(
					_.bind(function() {
						this._isDirty = false;
						this.renderHeader();
						this.renderAnnotationList();
					}, this)
				);
			},
			this
		);
	}
};

/**
 * render the annotation details header
 */
AnnotationWidget.prototype.renderHeader = function() {
	this._annotationFilters.empty();
	var annotationsOrigins = this.annotationModel.getAnnotationOrigins();
	if (
		annotationsOrigins.length > 1 ||
		(annotationsOrigins.length === 1 && annotationsOrigins[0] !== AnnotationConstants.AnnotationOriginTypes.user.id)
	) {
		this._annotationOriginSelector.element.show();
	} else {
		this._annotationOriginSelector.element.hide();
	}

	/**
	 * User selector
	 */
	this._usercombo = $("<select>");
	var all = $("<option/>")
		.text(this.f.i18n.t("AnnAllUsers"))
		.val("All");
	var customEnabled = $("<option selected disabled hidden value=''/>").text(this.f.i18n.t("AnnotFilterCustomEnabled"));

	if (this.annotationModel.getAnnotationFilterModel().isCustomFilterActive()) {
		this._usercombo.append(customEnabled);
	}

	this._usercombo.append(all);

	var users = this.annotationModel.getUsers(AnnotationConstants.AnnotationOriginTypes.user.id);
	_.each(
		users,
		function(loc, name) {
			var option = $("<option/>")
				.text(loc)
				.val(name);
			if (selectedUser == name) {
			}
			this._usercombo.append(option);
		},
		this
	);

	var selectedUser = this.annotationModel.getAnnotationFilterModel().getSelectedUser();
	if (selectedUser && !this.annotationModel.getAnnotationFilterModel().isCustomFilterActive()) {
		var hasSelectedUser =
			this._usercombo.find("option").filter(function(i, elem) {
				return $(elem).prop("value") === selectedUser;
			}).length > 0;
		if (hasSelectedUser) {
			this._usercombo.val(selectedUser);
		} else {
			var newUserOption = $("<option hidden selected/>").text(this.f.i18n.t("AnnotFilterUserNotAvailable"));
			this._usercombo.append(newUserOption);
		}
	}

	this._annotationFilters.append(this._usercombo);

	this._usercombo.on(
		"change",
		_.bind(function(e) {
			if (e.target.value != "All") {
				this.annotationModel.getAnnotationFilterModel().setSelectedUser(e.target.value);
			} else {
				this.annotationModel.getAnnotationFilterModel().setSelectedUser(null);
			}
		}, this)
	);

	/**
	 * State selector
	 */
	this._statecombo = $("<select>");
	var stateall = $("<option/>")
		.text(this.f.i18n.t("AnnAllReviewState"))
		.val("All");
	var customEnabledState = $("<option selected disabled hidden value=''/>").text(this.f.i18n.t("AnnotFilterCustomEnabled"));

	if (this.annotationModel.getAnnotationFilterModel().isCustomFilterActive()) {
		this._statecombo.append(customEnabledState);
	}

	this._statecombo.append(stateall);

	var sep = "------------------------------";
	this._statecombo.append(
		$("<option>")
			.text(sep)
			.prop("disabled", true)
	);

	var notprefix = "NOT_";

	var option = $("<option/>")
		.text(this.f.i18n.t("AnnStatusAny"))
		.val("Any");
	this._statecombo.append(option);
	var notoption = $("<option/>")
		.text(this.f.i18n.t("AnnNoReviewState"))
		.val("NOT_Any");
	this._statecombo.append(notoption);

	this._statecombo.append(
		$("<option>")
			.text(sep)
			.prop("disabled", true)
	);
	_.each(
		AnnotationConstants.AnnotationReviewState,
		function(state, key) {
			var option = $("<option/>")
				.text(this.f.i18n.t(AnnotationConstants.AnnotationReviewStateLocs[key]))
				.val(AnnotationConstants.AnnotationReviewState[key]);
			this._statecombo.append(option);
		},
		this
	);

	this._statecombo.append(
		$("<option>")
			.text(sep)
			.prop("disabled", true)
	);
	_.each(
		AnnotationConstants.AnnotationReviewState,
		function(state, key) {
			var notoption = $("<option/>")
				.text(this.f.i18n.t("AllExcept", this.f.i18n.t(AnnotationConstants.AnnotationReviewStateLocs[key])))
				.val(notprefix + AnnotationConstants.AnnotationReviewState[key]);
			this._statecombo.append(notoption);
		},
		this
	);

	this._annotationFilters.append(this._statecombo);
	this._statecombo.on(
		"change",
		_.bind(function(e) {
			if (e.target.value != "All") {
				var inverse = e.target.value.indexOf(notprefix) > -1;
				var status = e.target.value;
				if (inverse) {
					status = status.substring(notprefix.length);
				}
				this.annotationModel.getAnnotationFilterModel().setSelectedState(status, inverse);
			} else {
				this.annotationModel.getAnnotationFilterModel().setSelectedState(null, false);
			}
			return true;
		}, this)
	);

	var selectedState = this.annotationModel.getAnnotationFilterModel().getSelectedState();
	if (selectedState && !this.annotationModel.getAnnotationFilterModel().isCustomFilterActive()) {
		this._statecombo.val((this.annotationModel.getAnnotationFilterModel().isInverseState() ? notprefix : "") + selectedState);
	}

	/**
	 * The custom filter popup
	 */
	if (this._customFilterPopup && this._customFilterPopup.isVisible()) {
		this._customFilterPopup.render();
	}

	this._updateNumberOfSelectedAnnotations();
};

/**
 * Update the count of currently visbile and total number of annotations.
 * @private
 */
AnnotationWidget.prototype._updateNumberOfSelectedAnnotations = function() {
	this.f.i18n.loadBundle(
		"viewer",
		function() {
			var selectedThreads = this.f.i18n.t(
				"SelectedThreadsCount",
				this.annotationModel.getAnnotationFilterModel().getNumberOfFilteredAnnotations(),
				this.annotationModel.getAnnotationCount()
			);
			this._numberOfSelectedThreads.text(selectedThreads);
		},
		this
	);
};

/**
 * Render/update the complete annotation list
 */
AnnotationWidget.prototype.renderAnnotationList = function() {
	var filteredAnnots = this.annotationModel.getFilteredAnnotations();

	// mark all elements as elements that probably should be removed
	this.toBeRemovedFromDOM = _.clone(this.inDOM);

	// readd every annotation row from the filtered annotation list
	// if a row gets added to the DOM it is removed from the "toBeRemovedFromDOM" list
	_.each(
		filteredAnnots,
		function(annotation) {
			if (annotation.linkedto && annotation.linkedto.length > 0) {
				//do not show linked annotations in the table
				return;
			} else if (this.annotationRows[annotation.name]) {
				// this annotation row already exists, readd to DOM
				var row = this.annotationRows[annotation.name];
				row.element.next("hr").remove();
				row.setAnnotation(annotation);
				this._addToDom(row);
				row.element.show();
			} else {
				// there is no annotationRow for this annotation, create a new one
				this.addAnnotation(annotation);
			}
		},
		this
	);

	// remove all unneeded rows
	_.each(
		this.toBeRemovedFromDOM,
		function(value, name) {
			this._removeFromDOM(name);
		},
		this
	);

	this.toBeRemovedFromDOM = {};
	this._fireResized();
};

/**
 * Add an annotation row to the list
 * @private
 * @param {AnnotationRow} annotationRow
 */
AnnotationWidget.prototype._addToDom = function(annotationRow) {
	this.inDOM[annotationRow.name] = 1;
	delete this.toBeRemovedFromDOM[annotationRow.name];

	annotationRow.onDetachFromDOM();

	this.annotationList.append(annotationRow.element).append($("<hr/>"));
	annotationRow.onReattachToDOM();
};

/**
 * Remove an annotation row from the DOM
 * @private
 * @param {String} name - the name of the annotation row
 */
AnnotationWidget.prototype._removeFromDOM = function(name) {
	if (this.inDOM[name]) {
		var row = this.annotationRows[name];
		if (row) {
			row.element.next("hr").remove();
			row.onDetachFromDOM();
			row.element.detach();
		}
		delete this.toBeRemovedFromDOM[name];
		delete this.inDOM[name];
	}
};

/**
 * Create an annotationRow for a given Annotation, the annotation row is added to the DOM
 * @param {Annotation} annotation - the annotation that should be added
 * @param {Boolean} [fireResized] - if true, a resized event will be fired from the widget
 */
AnnotationWidget.prototype.addAnnotation = function(annotation, fireResized) {
	if (annotation.linkedto && annotation.linkedto.length > 0) {
		return; //do not show linked annotations in the table
	}

	var row = new AnnotationRow(this.f, annotation, this.annotationModel, undefined, undefined, this._isForMobile);
	this.annotationRows[annotation.name] = row;
	var annotationEl = row.element;

	this._addToDom(row);

	annotationEl.on(
		"click",
		_.bind(function(e) {
			this.annotationModel.selectAnnotation(row.getAnnotation(), true);
			e.stopPropagation();
		}, this)
	);

	var viewpressed = false;
	var viewactive = false;
	annotationEl.on(
		"mousedown",
		_.bind(function() {
			viewpressed = true;
			setTimeout(
				_.bind(function() {
					if (viewpressed) {
						viewactive = true;
						this.f.emit("module.annotation.highlightAnnotation", row.getAnnotation());
					}
				}, this),
				500
			);
		}, this)
	);

	annotationEl.on(
		"mouseup",
		_.bind(function() {
			if (viewpressed && viewactive) {
				this.f.emit("toolManager.deactivateTool", "annotationindicatortool");
			}
			viewpressed = false;
			viewactive = false;
		}, this)
	);

	annotationEl.on(
		"mouseout",
		_.bind(function() {
			if (viewpressed && viewactive) {
				this.f.emit("toolManager.deactivateTool", "annotationindicatortool");
			}
			viewpressed = false;
			viewactive = false;
		}, this)
	);

	annotationEl.on(
		"mouseleave",
		_.bind(function() {
			if (viewpressed && viewactive) {
				this.f.emit("toolManager.deactivateTool", "annotationindicatortool");
			}
			viewpressed = false;
			viewactive = false;
		}, this)
	);

	if (this.annotationModel.isAnnotationOrReplySelected(annotation)) {
		this.selectAnnotation(annotation);
	}

	this._updateNumberOfSelectedAnnotations();
	if (fireResized) {
		this._fireResized();
	}
};

/**
 * Select an annotation
 * @param {Annotation} annotation - the annotation that is selected
 */
AnnotationWidget.prototype.selectAnnotation = function(annotation) {
	if (this.annotationRows[annotation.inreplyto]) {
		this.annotationRows[annotation.inreplyto].toggleReplyLimit(false);
	}

	if (this.annotationRows[annotation.name] || this.annotationRows[annotation.inreplyto]) {
		var annotationRow = this.annotationRows[annotation.name] || this.annotationRows[annotation.inreplyto].annotationRows[annotation.name];
		annotationRow.setSelected(true);

		var element = annotationRow.element;
		// assure it is visible in the scrollable pane
		element.trigger("makeMeVisible");
	}
};

/**
 * Deselect an annotation
 * @param {Annotation} annotation - the annotation that is deselected
 */
AnnotationWidget.prototype.deselectAnnotation = function(annotation) {
	var annotationRow;

	if (this.annotationRows[annotation.name]) {
		annotationRow = this.annotationRows[annotation.name];
	}

	if (this.annotationRows[annotation.inreplyto]) {
		this.annotationRows[annotation.inreplyto].toggleReplyLimit();
		if (annotationRow === undefined) {
			annotationRow = this.annotationRows[annotation.inreplyto].annotationRows[annotation.name];
		}
	}

	if (annotationRow) {
		annotationRow.setSelected(false);
	}
};

/**
 * Update an annotation
 * @param {Annotation} annotation
 * @param {Annotation} [referenceAnnotation]
 */
AnnotationWidget.prototype.updateAnnotation = function(annotation, referenceAnnotation) {
	if (this.annotationRows[annotation.name]) {
		this.annotationRows[annotation.name].annotationUpdated(referenceAnnotation);
	} else if (annotation.linkedto && !annotation.parent && this.annotationRows[annotation.linkedto]) {
		this.annotationRows[annotation.linkedto].annotationUpdated();
	} else {
		var annot = annotation.getGodfather();
		if (annot !== annotation) {
			this.updateAnnotation(annot, annotation);
		}
	}
};

/**
 * Delete an annotation
 * @param {Annotation} annotation - Annotation to delete
 */
AnnotationWidget.prototype.deleteAnnotation = function(annotation) {
	if (this.annotationRows[annotation.name]) {
		this._removeFromDOM(annotation.name);
		delete this.annotationRows[annotation.name];
	} else if (annotation.linkedto && !annotation.parent && this.annotationRows[annotation.linkedto]) {
		this._removeFromDOM(annotation.linkedto);
		delete this.annotationRows[annotation.linkedto];
	} else {
		annotation = annotation.getGodfather();
		this.updateAnnotation(annotation);
	}

	this._updateNumberOfSelectedAnnotations();
	this._fireResized();
};

/**
 * Gets called when the widget was resized,
 * Update the height of the annotationsListWrapper
 * @override
 * @private
 */
AnnotationWidget.prototype.resized = function() {
	this.annotationsListWrapper.css("height", "");
	if (this.isContextAvailable() && this.detailsDiv.css("display") !== "flex") {
		_.defer(
			_.bind(function() {
				this.context.css("overflow", "hidden");
				this.annotationsListWrapper.css("height", "auto");

				var height = this.context.outerHeight(true) || null;
				if (!this._isForMobile) {
					height -= this.header.outerHeight(true) || null;
				}
				if (this.warningMessage.is(":visible")) {
					height -= this.warningMessage.outerHeight(true) || null;
				}
				if (this.temporarilyDisabledDiv.is(":visible")) {
					height -= this.temporarilyDisabledDiv.outerHeight(true) || null;
				}
				if (this._annotationOriginSelector.element.is(":visible")) {
					height -= this._annotationOriginSelector.element.outerHeight(true) || null;
				}
				if (height > 0) {
					this.annotationsListWrapper.css("height", height);
				}

				if (this._visibleItem) {
					this._visibleItem.trigger("makeMeVisible");
				}
				this.context.css("overflow", "auto");
			}, this)
		);
	}
};

/**
 * Get the preferred size
 * @override
 */
AnnotationWidget.prototype.getPreferredSize = function() {
	return new ShapeUtils.Size(
		this.annotationsListWrapper.outerWidth(true) || null,
		(this.annotationsListWrapper.outerHeight(true) || null) + (this.nonScrollablePart.outerHeight(true) || null)
	);
};

/**
 * Start the heartbeat for the auto time labels
 */
AnnotationWidget.prototype.startHeartBeat = function() {
	this.canRunTimer = true;
	this.updateTimers();
};

/**
 * Stop the heartbeat for the auto time labels
 */
AnnotationWidget.prototype.stopHeartBeat = function() {
	this.canRunTimer = false;
};

/**
 * Update the filter pinned icon
 */
AnnotationWidget.prototype._updateFilterPinnedIcon = function() {
	if (this.annotationModel.getAnnotationFilterModel().isAnnotationFilterPinned()) {
		this._pinFilterButton.setIcon("pinned icon-WCR_Pin12");
		this._pinFilterButton.setTooltip("viewer", "AnnotFilterPinOn");
	} else {
		this._pinFilterButton.setIcon("unpinned icon-WCR_Pin12");
		this._pinFilterButton.setTooltip("viewer", "AnnotFilterPinOff");
	}
};

/**
 * Update all autotimelabels
 */
AnnotationWidget.prototype.updateTimers = function() {
	if (!this.canRunTimer) {
		return;
	}

	var timer = 30000;
	_.each(this.annotationRows, function(row) {
		row.updateTimers();
	});

	setTimeout(
		_.bind(function() {
			this.updateTimers();
		}, this),
		timer
	);
};

/**
 * Update the visible of the annotation list + header
 * and show messages or toggling buttons for when in compare mode
 */
AnnotationWidget.prototype.updateListVisibility = function() {
	if (this.annotationModel.isTemporarilyDisabled()) {
		// temporarily disabled, show toggle buttons if possible
		// hide list if no annotations are shown, and show a message
		this.temporarilyDisabledDiv.toggle(this.annotationModel.isTemporarilyAnnotationDocumentTogglingEnabled());

		this.temporaryAnnotationToggler.toggle(this.annotationModel.isTemporarilyAnnotationDocumentTogglingEnabled());
		this.showCurrentFileAnnotationsButton.setActive(this.annotationModel.isTemporarilyCurrentPageAnnotationsShown());
		this.showDifferenceFileAnnotationsButton.setActive(this.annotationModel.isTemporarilyDifferencePageAnnotationsShown());

		if (this.annotationModel.isTemporarilyDisabledAndNoAnnotationsVisible()) {
			//no annotations are visible, show a message is one is provided
			this.header.hide();
			this.annotationsListWrapper.hide();
			this.updateStepThroughButton(false);
			this.reloadButton.setEnabled(false);
		} else {
			//there are some annotations, we don't need to show a message
			this.header.show();
			this.annotationsListWrapper.show();
			this.updateStepThroughButton(true);
			this.reloadButton.setEnabled(true);
		}
	} else {
		// module is not disabled, so show all things and hide the temporarilydisabled div
		this.header.show();
		this.annotationsListWrapper.show();
		this.temporarilyDisabledDiv.hide();
		this.updateStepThroughButton(true);
		this.reloadButton.setEnabled(true);
	}

	this.resized();
};

/**
 * Zooms to a given annotation only if the "Fit View to Selected Annotation" setting is enabled.
 * @private
 * @param {Annotation} annotation
 * @param {Boolean} isNewAnnotation - Indicates if the function is called because a new annotation has been added
 */
AnnotationWidget.prototype._fitViewToAnnotation = function(annotation, isNewAnnotation) {
	if (!isNewAnnotation && this.annotationModel.shouldAutoFitViewToSelectedAnnotation()) {
		this.f.emit("module.annotation.zoomToAnnotation", annotation);
	}
};

export default AnnotationWidget;
