import AnnotationFactory from "@esko/web-viewer/annotation/AnnotationFactory";
import AnnotationGV from "@esko/web-viewer/annotation/AnnotationGV";
import AbstractDocumentPage from "@esko/web-viewer/data/AbstractDocumentPage";
import ShapeUtils from "@esko/shapeutils";
import XMLUtils from "@esko/xmlutils";
import ViewerUtils from "@esko/web-viewer/utils/ViewerUtils";
import AnnotationConstants from "@esko/web-viewer/annotation/AnnotationConstants";
import _ from "underscore";
import $ from "jquery";
/**
 * Inject common annotation functionality in the data model
 * The AbstractDocumentPage prototype gets extended with methods for retrieving and storing annotations
 *
 * @exports CommonAnnotationDataExtensions
 * @memberof AnnotationModule
 * @param {Core.ViewerFacade} f - Viewer facade
 */
var CommonAnnotationDataExtensions = _.once(function(f) {
	/**
	 * Initialize the page with needed data object, so the annotation module can work correctly
	 * @private
	 */
	AbstractDocumentPage.prototype._initAnnotationModule = function() {
		if (!this._annotData) {
			/**
			 * dataStatus: [notAvailable | available | loading]
			 * onAnnotationsAvailableCallbacks: array of callbacks for when the annotations are available
			 * annotations: {Array.<Annotation>}
			 */
			this._annotData = {
				dataStatus: "notAvailable",
				onAnnotationsAvailableCallbacks: [],
				annotations: []
			};
		}
	};

	/**
	 * Load the annotation information for this page
	 * If the information is already available, it won't be reloaded (use reloadAnnotations if this is needed)
	 *
	 * @param {Function} onAnnotationsAvailable - Callback. First argument {Boolean} indicates whether the annotations are loaded successfully or not
	 */
	AbstractDocumentPage.prototype.requestAnnotations = function(onAnnotationsAvailable) {
		this._initAnnotationModule();
		if (this._annotData.dataStatus === "available") {
			onAnnotationsAvailable(true);
		} else {
			if (this._annotData.dataStatus === "loading") {
				this._annotData.onAnnotationsAvailableCallbacks.push(onAnnotationsAvailable);
			} else {
				// start loading information if needed
				if (this._shouldFetchAnnotations(AnnotationConstants.AnnotationOriginTypes.user.id)) {
					this._annotData.onAnnotationsAvailableCallbacks.push(onAnnotationsAvailable);
					this._fetchAnnotationsFromServer();
				} else {
					onAnnotationsAvailable(true);
				}
			}
		}
	};

	/**
	 * (re)load the annotation information for this page.
	 *
	 * If the information is already available, the information will be reloaded
	 *
	 * @param {Function} onAnnotationsAvailable - Callback. First argument {Boolean} indicates whether the annotations are loaded successfully or not
	 */
	AbstractDocumentPage.prototype.reloadAnnotations = function(onAnnotationsAvailable) {
		this._initAnnotationModule();

		if (this._annotData.dataStatus === "loading") {
			// the system is still loading, register callback
			this._annotData.onAnnotationsAvailableCallbacks.push(onAnnotationsAvailable);
		} else {
			// normal reload
			// start loading information
			this._annotData.onAnnotationsAvailableCallbacks.push(onAnnotationsAvailable);
			this._fetchAnnotationsFromServer();
		}
	};

	/**
	 * Reset the annotations so they will need to be reloaded next time
	 */
	AbstractDocumentPage.prototype.resetAnnotations = function() {
		this._annotData.dataStatus = "notAvailable";
		this._annotData.annotations = [];
	};

	/**
	 * Fire all onAnnotationsAvailableCallbacks with "result" as parameter
	 * @private
	 * @param {Boolean} result- True when the annotations were successfully loaded, false otherwise
	 */
	AbstractDocumentPage.prototype._fireOnAnnotationsAvailableCallbacks = function(result) {
		_.each(this._annotData.onAnnotationsAvailableCallbacks, function(callback) {
			if (callback) {
				callback(result);
			}
		});

		this._annotData.onAnnotationsAvailableCallbacks = [];
		this._annotData.dataStatus = result ? "available" : "notAvailable";
	};

	/**
	 * Fetch annotation information from the server
	 * @private
	 */
	AbstractDocumentPage.prototype._fetchAnnotationsFromServer = function() {
		var params = this._getAnnotationParams();

		this._annotData.dataStatus = "loading";

		var promises = [];
		if (this._shouldFetchAnnotations(AnnotationConstants.AnnotationOriginTypes.user.id)) {
			promises.push(this._getUserAnnotations(params));
		}
		if (this._shouldGVFetchAnnotations()) {
			promises.push(this._getGVAnnotations(params));
		}
		var deferred = $.Deferred();
		var noErrors = true;
		$.when
			.apply($, promises)
			.done(
				_.bind(function() {
					var annotations = [];
					_.each(arguments, function(annots) {
						annotations = annotations.concat(annots);
					});
					this._annotData.annotations = annotations;
					deferred.resolve(annotations);
				}, this)
			)
			.fail(
				_.bind(function(data) {
					f.i18n.loadBundle("viewer", function() {
						f.emit("error.SEVERE", f.i18n.t("AnnOpenError", ""), data);
					});
					noErrors = false;
					deferred.reject();
				}, this)
			)
			.always(
				_.bind(function() {
					this._fireOnAnnotationsAvailableCallbacks(noErrors);
				}, this)
			);

		return deferred.promise();
	};

	AbstractDocumentPage.prototype._shouldFetchAnnotations = function(type) {
		return this.hasAnnotations(type);
	};

	AbstractDocumentPage.prototype._shouldGVFetchAnnotations = function(type) {
		return this._hasGVAnnotations();
	};

	/**
	 * Define parameters for fetching annotations
	 * @private
	 */
	AbstractDocumentPage.prototype._getAnnotationParams = function() {
		// Override me in implementation specific annotation data extensions
		var params = {};
		return params;
	};

	/**
	 * checks if there are GV annotations
	 */
	AbstractDocumentPage.prototype._hasGVAnnotations = function() {
		return (
			this.hasAnnotations(AnnotationConstants.AnnotationOriginTypes.artwork.id) ||
			this.hasAnnotations(AnnotationConstants.AnnotationOriginTypes.braille.id) ||
			this.hasAnnotations(AnnotationConstants.AnnotationOriginTypes.barcode.id) ||
			this.hasAnnotations(AnnotationConstants.AnnotationOriginTypes.text.id) ||
			this.hasAnnotations(AnnotationConstants.AnnotationOriginTypes.spelling.id) ||
			this.hasAnnotations(AnnotationConstants.AnnotationOriginTypes.ext.id) ||
			this.hasAnnotations(AnnotationConstants.AnnotationOriginTypes.editor.id) ||
			this.hasAnnotations(AnnotationConstants.AnnotationOriginTypes.acrobat.id)
		);
	};

	/**
	 * Parse the received annotation data (xml)
	 * @param {XMLObject} data - XML object holding the annotation data in the tag annots
	 * @param {Array.<Annotation>} currentAnnotations - The annotations to check against if annotations already exist
	 * @param {AnnotationConstants.AnnotationOriginTypes.x.id} [originType] - The origin type of the annotations if known, otherwise it will be derived
	 * @returns {Array.<Annotation>} annotations
	 */
	AbstractDocumentPage.prototype._parseAnnotationData = function(data, currentAnnotations, originType) {
		//Make sure the annotations are initialized (e.g. when there were no annotations before)
		this._initAnnotationModule();

		// keep track of the message number, mapped by source
		var messageNrs = {};
		if (!data) {
			return [];
		}

		var annotations = [];

		try {
			var elements = data.getElementsByTagName("annots");
			var annots;
			var child;
			if (elements && elements.length >= 1) {
				annots = data.getElementsByTagName("annots")[0];
			}

			if (annots) {
				child = annots.firstChild;
			}

			/**
			 * Make a temporary dictionary so see if the annotations are new or an update
			 *
			 * this will make a flat list of the annotation list, so not considering replies, states etc
			 */
			var tmpAnnotationDictionary = {};
			if (currentAnnotations != null) {
				for (var i = 0; i < currentAnnotations.length; i++) {
					currentAnnotations[i].runThroughAnnotation(true, true, true, function(annot) {
						tmpAnnotationDictionary[annot.name] = annot;
					});
				}
			}

			var rawAnnotationList = {};
			var source;
			while (child) {
				if (child.nodeType === 1) {
					var name = child.getAttribute("name");

					var annot = null;
					if (tmpAnnotationDictionary[name]) {
						annot = tmpAnnotationDictionary[name];

						/*
                         * Remove all the current states/replies/links
                         */
						annot.replies = [];
						annot.links = [];
						annot.states = [];
						annot.resetUnsavedReply();
						annot.cancelModified();
						annot._scaling = null;

						delete tmpAnnotationDictionary[name];
					} else {
						source = this._getAnnotationSourceFromXML(child, originType);
						if (source === AnnotationConstants.AnnotationOriginTypes.user.id) {
							annot = new AnnotationFactory(f).create();
						} else {
							// external annotations
							annot = this._createAnnotation(source);
						}
					}
					source = annot.originType;
					messageNrs[source] = messageNrs[source] ? ++messageNrs[source] : 1;

					annot.initWithXML(
						child,
						{
							trimboxOffset: this.getAnnotationsOffset(),
							mediaboxOffest: this.getAnnotationsOffset(true)
						},
						messageNrs[source]
					);
					rawAnnotationList[annot.name] = annot;
				}
				child = child.nextSibling;
			}

			tmpAnnotationDictionary = null;

			/**
			 * The structure of the xml is that all of the annotations AND their replies are seen as annotations
			 * Only, if an annotation is a reply, it has the "inreplyto" field and should NOT be drawn...
			 */
			for (var annotName in rawAnnotationList) {
				if (rawAnnotationList.hasOwnProperty(annotName)) {
					var rawAnnotation = rawAnnotationList[annotName];
					if (rawAnnotation.inreplyto) {
						// the annotation is a reply or review state, add it to its parent
						var inreplyAnnot = rawAnnotationList[rawAnnotation.inreplyto];
						if (inreplyAnnot) {
							rawAnnotation.parent = inreplyAnnot;

							if (rawAnnotation.state) {
								inreplyAnnot.states.push(rawAnnotation);
							} else {
								inreplyAnnot.replies.push(rawAnnotation);
							}
						}
					} else {
						// root level annoation, push it on the annoations list
						if (rawAnnotation.linkedto) {
							if (rawAnnotationList[rawAnnotation.linkedto]) {
								// the annotation is linked to another annotation, establish link
								rawAnnotationList[rawAnnotation.linkedto].links.push(rawAnnotation);
								annotations.push(rawAnnotationList[annotName]);
								rawAnnotation.parent = rawAnnotationList[rawAnnotation.linkedto];
								//set linked annotation sequence number correct
								rawAnnotation.messagenumber = rawAnnotation.parent.messagenumber;
								rawAnnotation.updateIndicator();
							}
						} else {
							annotations.push(rawAnnotationList[annotName]);
						}
					}
				}
			}
		} catch (err) {
			f.i18n.loadBundle("viewer", function() {
				f.emit("error.SEVERE", f.i18n.t("AnnDisplayError"), err);
			});
		}
		return annotations;
	};

	/**
	 * gets the source of the annotation
	 * @param {NodeList} xml
	 * @param {AnnotationConstants.AnnotationOriginTypes.x.id} originType - the origin of the annotations if already known, otherwise the source will be derived
	 * @return {AnnotationConstants.AnnotationOriginTypes.id}
	 */
	AbstractDocumentPage.prototype._getAnnotationSourceFromXML = function(xml, originType) {
		if (originType) {
			return originType;
		}

		var subject = XMLUtils.attributeValue(xml, subject);
		if (subject) {
			if (ViewerUtils.startsWith(subject, "GVBarcode") || ViewerUtils.startsWith(subject, "Barcode")) {
				return AnnotationConstants.AnnotationOriginTypes.barcode.id;
			}
			if (ViewerUtils.startsWith(subject, "GVBraille") || ViewerUtils.startsWith(subject, "BrailleProof")) {
				return AnnotationConstants.AnnotationOriginTypes.braille.id;
			}
			if (ViewerUtils.startsWith(subject, "GVSpell")) {
				return AnnotationConstants.AnnotationOriginTypes.spelling.id;
			}
			if (ViewerUtils.startsWith(subject, "GVArtwork")) {
				return AnnotationConstants.AnnotationOriginTypes.artwork.id;
			}
			if (ViewerUtils.startsWith(subject, "GVText")) {
				return AnnotationConstants.AnnotationOriginTypes.text.id;
			}
			if (ViewerUtils.startsWith(subject, "Global Vision")) {
				var title = XMLUtils.attributeValue(xml, title);
				if (title) {
					if (ViewerUtils.endsWith(title, "Spelling")) {
						return AnnotationConstants.AnnotationOriginTypes.spelling.id;
					}
					if (ViewerUtils.endsWith(title, "Change")) {
						return AnnotationConstants.AnnotationOriginTypes.text.id;
					}
					if (ViewerUtils.endsWith(title, "Difference")) {
						return AnnotationConstants.AnnotationOriginTypes.artwork.id;
					}
				}
			}
		}

		var source = XMLUtils.attributeValue(xml, "source");
		if (source) {
			var indexOfDash = source.lastIndexOf("-");
			if (indexOfDash > 0) {
				return source.substr(indexOfDash + 1, source.lastIndexOf(".") - indexOfDash - 1);
			}
		}

		// default - anything else
		return AnnotationConstants.AnnotationOriginTypes.ext.id;
	};

	/**
	 * creates an annotation from a particular source
	 * @param {String} source - source of the annotation
	 */
	AbstractDocumentPage.prototype._createAnnotation = function(source) {
		return new AnnotationGV(f, source);
	};

	/**
	 * Get annotations
	 * @returns {Array.<AnnotationModule.Annotation>} Raw annotations
	 */
	AbstractDocumentPage.prototype.getAnnotations = function() {
		this._initAnnotationModule();
		return this._annotData.annotations;
	};

	// ANNOTATION STORAGE
	/**
	 * Save annotations
	 * @param {Array.<AnnotationModule.Annotation>} annotations
	 * @param {Array.<AnnotationModule.Annotation>} removedannotations
	 * @param {Function} callback - is called after the annotations were successfully saved
	 */
	AbstractDocumentPage.prototype.saveAnnotations = function(annotations, removedannotations, callback) {
		//Create first the XML
		var xfdf = document.implementation.createDocument("http://ns.adobe.com/xfdf/", "xfdf", null);

		// set character set (this is needed for IE (9))
		try {
			xfdf.characterSet = "UTF-8";
		} catch (exception) {
			// FireFox and possible other browsers might block this assignment as it is readonly-only property
			// but they assign the character set from the original document (which is UTF-8)
			// we can thus ignore this exception
		}

		try {
			xfdf.charset = "UTF-8";
		} catch (exception) {
			// FireFox and possible other browsers might block this assignment as it is readonly-only property
			// but they assign the character set from the original document (which is UTF-8)
			// we can thus ignore this exception
		}

		var annots = document.createElementNS(null, "annots");
		xfdf.documentElement.appendChild(annots);

		var addedAnnotations = [];
		var modifiedAnnotations = [];

		for (var i = 0; i < annotations.length; i++) {
			if (annotations[i].isNew) {
				addedAnnotations.push(annotations[i]);
			} else {
				modifiedAnnotations.push(annotations[i]);
			}
			annots.appendChild(annotations[i].toXML(xfdf));
		}

		var remove = [];
		if (removedannotations && removedannotations.length > 0) {
			_.each(removedannotations, function(an) {
				an.runThroughAnnotation(true, true, true, function(annot) {
					remove.push(annot.name);
				});
			});
		}

		var params = this._getAnnotationParams();

		if (remove && remove.length > 0) {
			params.remove = remove;
		}

		this._doSaveAnnotations(xfdf, params, addedAnnotations, modifiedAnnotations, removedannotations, callback);
	};

	/**
	 *
	 * Things that need to be done after parsing successful response from change annotations (in saveAnnotations).
	 * @private
	 * @param {Array.<Annotation>} addedAnnotations
	 * @param {Array.<Annotation>} modifiedAnnotations
	 * @param {Array.<Annotation>} removedAnnotations
	 * @param {Function} callback
	 */
	AbstractDocumentPage.prototype._processAfterParsingSavedAnnotations = function(
		addedAnnotations,
		modifiedAnnotations,
		removedAnnotations,
		callback
	) {
		// update the page that it has annotations or not
		var rawAnnotations = this.getAnnotations();
		var typeOfOrigins = _.pluck(rawAnnotations, "originType");
		var hasAnnotsObj = {};
		_.each(typeOfOrigins, function(origin) {
			hasAnnotsObj[origin] = true;
		});
		this.setHasAnnotations(_.size(rawAnnotations) > 0, hasAnnotsObj);

		var newCallback = function(status) {
			if (callback) {
				callback(rawAnnotations);
			}

			// only if everything was successfull, show this message
			if (status) {
				// SDK support
				var annotationCallbackUrl = f.getProperty("annotationcallbackurl");
				if (annotationCallbackUrl) {
					f.getDataAdapter().fetch(annotationCallbackUrl, {
						dataType: undefined, // intelligent guess of what will come back
						noAjaxParam: true
					});
				}

				f.i18n.loadBundle("viewer", function() {
					f.emit("userMessage", "success", f.i18n.t("AnnotationsSaved"), true);
				});
			}
		};

		this.reportAnnotations(addedAnnotations, modifiedAnnotations, removedAnnotations, newCallback);
	};

	/**
	 * Report to Webserver that there was a modification on the annotations for this document page
	 *
	 * @param {Array.<Annotation>} newannotations - List of the newly added annotations
	 * @param {Array.<Annotation>} modifiedannotations - List of the modified annotations
	 * @param {Array.<Annotation>} removedannotations - List of the removed annotatinos
	 * @param {Function} [callback] - Callback with parameter boolean success (true) or error (false)
	 */
	AbstractDocumentPage.prototype.reportAnnotations = function(newannotations, modifiedannotations, removedannotations, callback) {
		// Aggregate the callback
		var expectedNumberOfCalls = 0;
		if (newannotations) {
			expectedNumberOfCalls += newannotations.length;
		}
		if (modifiedannotations) {
			expectedNumberOfCalls += modifiedannotations.length;
		}
		if (removedannotations) {
			expectedNumberOfCalls += removedannotations.length;
		}
		if (expectedNumberOfCalls === 0) {
			callback(true);
		}
		var numberOfCalls = 0;
		var status = true;
		var aggregatedCallback = function(success) {
			if (!success) {
				// one of the calls failed
				status = false;
			}

			// count the number of calls to this function
			numberOfCalls++;
			if (numberOfCalls === expectedNumberOfCalls) {
				if (callback) {
					callback(status);
				}
			}
		};

		f.i18n.loadBundle(
			"viewer",
			_.bind(function() {
				if (newannotations) {
					_.each(
						newannotations,
						function(annotation) {
							annotation.cancelModified();
							this.alertOnAnnotationAdded(annotation, aggregatedCallback);
						},
						this
					);
				}
				if (modifiedannotations) {
					_.each(
						modifiedannotations,
						function(annotation) {
							annotation.cancelModified();
							this.alertOnAnnotationModified(annotation, aggregatedCallback);
						},
						this
					);
				}
				if (removedannotations) {
					_.each(
						removedannotations,
						function(annotation) {
							this.alertOnAnnotationDelete(annotation, aggregatedCallback);
						},
						this
					);
				}
			}, this)
		);
	};

	/**
	 * This function will determine the string for in the document history report
	 * For a reply: Reply (sequenceNrGodFater)
	 * For an annotation: Type annotation (sequenceNrIfKnown)
	 *
	 * @param {Annotation} annotation - the annotation for which to generate type string
	 * @return {String}
	 */
	function _getAnnotationType(annotation) {
		var sequenceNumber = "";
		if (annotation.inreplyto) {
			// the annotation is a Reply

			// find the godfather to determine the sequence number
			var godfather = annotation.getGodfather();

			if (godfather && godfather.messagenumber >= 0) {
				sequenceNumber = " (" + godfather.messagenumber + ")";
			}
			return f.i18n.t("AnnCommReply") + sequenceNumber;
		} else {
			if (annotation.messagenumber >= 0 && !annotation.isNew) {
				sequenceNumber = " (" + annotation.messagenumber + ")";
			}
			return f.i18n.t(AnnotationConstants.AnnotationComments[annotation.annottype]) + sequenceNumber;
		}
	}

	/**
	 * Add that an annotation was added to the document history
	 * @param {Annotation} annotation - Annotation that is added
	 * @param {Function} [callback] - Callback with parameter boolean success (true) or error (false)
	 * @example
	 * Line annoataion "Annotation" added by Cagnie, Ruben (RUCA) on Aug 21, 2013 at 10:15 EDT
	 * Rectangle annotation "http://www.esko.com" added by Cagnie, Ruben (RUCA) on Aug 21, 2013 at 10:42 EDT
	 */
	AbstractDocumentPage.prototype.alertOnAnnotationAdded = function(annotation, callback) {
		var sameDates = annotation.date.getTime() == annotation.creationdate.getTime();
		var type = _getAnnotationType(annotation);
		var text = f.i18n.t(
			sameDates ? "AnnCommAdded" : "AnnCommAdded2",
			type,
			annotation.getStringContent(130),
			annotation.getUserName(),
			f.i18n.d("AnnCommDate", annotation.creationdate),
			f.i18n.d("AnnCommDate", annotation.date)
		);
		if (this.getDocumentVersion().isMultiPage()) {
			//has multiple pages
			text = f.i18n.t("PagesString", this.getPagenum(), this.getPagenum(), this.getDocumentVersion().getPageCount()) + text;
		}
		this.updateApprovalState(text, callback);
	};

	/**
	 * Add that an annotation was deleted to the document history
	 * @param {Annotation} annotation - Annotation that is deleted
	 * @param {Function} [callback] - Callback with parameter boolean success (true) or error (false)
	 * @param {Boolean} [doNotReportReplies - If true, the deletion of replies to the deleted annotation is not reported (used internally)
	 * @example
	 * Rectangle annotation (27) "swaha" deleted by Cagnie, Ruben (RUCA) - was created by 27 Cagnie, Ruben (RUCA) on Aug 21, 2013 at 10:07 EDT and last modified on Aug 21, 2013 at 10:08 EDT
	 */
	AbstractDocumentPage.prototype.alertOnAnnotationDelete = function(annotation, callback, doNotReportReplies) {
		if (!doNotReportReplies && annotation.replies && annotation.replies.length > 0) {
			// we should report replies, and the annotation has 1 or more replies

			// generate a new callback which is only executed after the callback is executed for every reply
			var newCallback = _.after(
				annotation.replies.length,
				_.bind(function() {
					// report the main annotation, but don't report its replies
					this.alertOnAnnotationDelete(annotation, callback, true);
				}, this)
			);

			// for every reply, make a report
			_.each(
				annotation.replies,
				function(reply) {
					this.alertOnAnnotationDelete(reply, newCallback);
				},
				this
			);
		} else {
			// report the deletion
			var sameDates = annotation.date.getTime() == annotation.creationdate.getTime();
			var type = _getAnnotationType(annotation);
			var text = f.i18n.t(
				sameDates ? "AnnCommDeleted" : "AnnCommDeleted2",
				type,
				annotation.getStringContent(130),
				f.getCurrentUser().getFullName(),
				annotation.getUserName(),
				f.i18n.d("AnnCommDate", new Date()),
				f.i18n.d("AnnCommDate", annotation.creationdate),
				f.i18n.d("AnnCommDate", annotation.date)
			);
			if (this.getDocumentVersion().isMultiPage()) {
				//has multiple pages
				text = f.i18n.t("PagesString", this.getPagenum(), this.getPagenum(), this.getDocumentVersion().getPageCount()) + text;
			}
			this.updateApprovalState(text, callback);
		}
	};

	/**
	 * Add that an annotation was deleted to the document history
	 * @param {Annotation} annotation - Annotation that is modified
	 * @param {Function} [callback] - Callback with parameter boolean success (true) or error (false)
	 * @example
	 * Oval annotation (1) "Eggs" modified by Cagnie, Ruben (RUCA) at Aug 21, 2013 at 10:30 EDT - was created by Cagnie, Ruben (RUCA) on Jun 08, 2013 at 14:26 EDT
	 */
	AbstractDocumentPage.prototype.alertOnAnnotationModified = function(annotation, callback) {
		var type = _getAnnotationType(annotation);
		var text = f.i18n.t(
			"AnnCommModded",
			type,
			annotation.getStringContent(130),
			f.getCurrentUser().getFullName(),
			annotation.getUserName(),
			f.i18n.d("AnnCommDate", annotation.creationdate),
			f.i18n.d("AnnCommDate", new Date())
		);
		if (this.getDocumentVersion().isMultiPage()) {
			//has multiple pages
			text = f.i18n.t("PagesString", this.getPagenum(), this.getPagenum(), this.getDocumentVersion().getPageCount()) + text;
		}
		this.updateApprovalState(text, callback);
	};
});

export default CommonAnnotationDataExtensions;
