import AbstractDocumentPage from "@esko/web-viewer/data/AbstractDocumentPage";
import ShapeUtils from "@esko/shapeutils";

export default function(f) {
	/**
	 * Get the offset of the annotations on the current page, this depends on the trimbox origin and the rotation/flipped state
	 * @param {Boolean} relativeToMediaBox - If true, the annotations are not relative to the trimbox, but relative to the mediabox
	 * @returns {ShapeUtils.Size}
	 */
	AbstractDocumentPage.prototype.getAnnotationsOffset = function(relativeToMediaBox) {
		var viewport = this.f.getModule("viewport");
		var rotation = this.getRotation() % 360;
		var isFlipped = this.isFlipped();
		var worldBounds = viewport.getWorldBounds();
		var pageSize = this.getPageSize();
		if (this.hasGraphics()) {
			//for document pages with graphics:

			var trimboxRect;
			if (relativeToMediaBox) {
				trimboxRect = this.getMediaBoxRect();
			} else {
				trimboxRect = this.getTrimboxRect();
			}

			var annotationOrigin;

			//get the correct origin for the annotations depending on the rotation and flipped state to get the correct trimbox corner
			if ((!isFlipped && rotation === 0) || (isFlipped && rotation === 90)) {
				annotationOrigin = trimboxRect.bottomleft();
			} else if ((!isFlipped && rotation === 90) || (isFlipped && rotation === 180)) {
				annotationOrigin = trimboxRect.topleft();
			} else if ((!isFlipped && rotation === 180) || (isFlipped && rotation === 270)) {
				annotationOrigin = trimboxRect.topright();
			} else {
				annotationOrigin = trimboxRect.bottomright();
			}
			return new ShapeUtils.Size(annotationOrigin.x, -(worldBounds.size.height - annotationOrigin.y)); //make sure to invert the y-coordinate (annotation-coordinates start from bottom, not top)
		} else if (this.hasCad()) {
			//for CAD files use the origin location in the current world
			var origin = new ShapeUtils.Point();

			if (this.getCad() && this.getCad().getLocation()) {
				//if the location is defined (happens by the layer model) use it
				origin = this.getCad()
					.getLocation()
					.origin.copy();
			}
			//if the CAD is rotated use following calculations as the orign will always give the top left corner
			if (!isFlipped && rotation === 0) {
				origin.y += pageSize.height;
			} else if (!isFlipped && rotation === 90) {
				//nothing needs to change here
			} else if (!isFlipped && rotation === 180) {
				origin.x += pageSize.width;
			} else if (!isFlipped) {
				origin.x += pageSize.height;
				origin.y += pageSize.width;
			}
			//if the CAD is flipped and rotated use following calculations
			if (isFlipped && rotation === 90) {
				origin.y += pageSize.width;
			} else if (isFlipped && rotation === 180) {
				//nothing needs to change
			} else if (isFlipped && rotation === 270) {
				origin.x += pageSize.height;
			} else if (isFlipped) {
				origin.x += pageSize.width;
				origin.y += pageSize.height;
			}

			return new ShapeUtils.Size(origin.x, -(worldBounds.size.height - origin.y));
		} else if (this.hasSvg()) {
			if (this.isReaderSpread() && this.get("isRightPage")) {
				var diffPage = this.f.getDocumentController().getDifferencePage();
				if (diffPage) {
					return new ShapeUtils.Size(diffPage.getSVGSize().width, 0);
				} else {
					return new ShapeUtils.Size(this.getSVGSize().width, 0);
				}
			}
		} else if (this.is3D()) {
			const worldBounds = this.f.getModule("viewport").getWorldBounds();
			return new ShapeUtils.Size(worldBounds.size.width / 2, -worldBounds.size.height / 2);
		}
		return new ShapeUtils.Size();
	};
}
