import _ from "underscore";
import ShapeUtils from "@esko/shapeutils";
import XMLUtils from "@esko/xmlutils";
import CadLine from "@esko/web-viewer/cad/CadLine";
import CadText from "@esko/web-viewer/cad/CadText";
import CadConstants from "@esko/web-viewer/cad/CadConstants";
import ViewerUtils from "@esko/web-viewer/utils/ViewerUtils";
/**
 * @exports CadLayer
 * @memberof CadModule
 * @class CadLayer is an object that contains the information of a layer + its lines and oneups
 * @constructor
 * @param {String} layerclass - class of the layer
 * @param {String} layertype - type of the layer
 * @param {String} layername - name of the layer
 * @param {Boolean} [oneup] - if true, the xml node describes a oneup CadFile within a layout CadFile
 */
var CadLayer = function(layerclass, layertype, layername, oneup) {
	/**
	 * Class of the layer
	 */
	this.layerclass = layerclass;

	/**
	 * Type of the layer
	 */
	this.layertype = layertype;

	/**
	 * Layer name
	 */
	this.layername = layername;

	/**
	 * Array of CadLine/CadText objects
	 */
	this.lines = [];

	/**
	 * Array of oneups (CadFile objects)
	 */
	this.oneUps = [];
};

/**
 * @exports CadFile
 * @memberof CadModule
 * @class CadFile is an object that contains all needed info for a cad file (layers, styles...)
 * @constructor
 * @param {CadStyle} style - style to use to construct the CadFile (colors/width/...)
 * @param {XMLElement} node - xml to construct the CadFile
 * @param {ShapeUtils.Rect} [rect] - Fixed rectangle (if not, its calculated from the xml node)
 * @param {Boolean} [oneup] - if true, the xml node describes a oneup CadFile within a layout CadFile
 */
var CadFile = function(style, node, rect, oneup) {
	/**
	 * Plotting styles
	 */
	this.cstyles = style;

	/**
	 * Name of the CadFile (only used for a oneup CadFile)
	 */
	this.name = null;

	/**
	 * Cad layers (Array.<CadLayer>)
	 */
	this.cadlayers = [];

	/**
	 * Oneup layers (undefined in the beginning, but getOneupClasses())
	 */
	this.oneupClasses = null;

	/**
	 * {Array.<CadLayer>} objects that are selected
	 * @private
	 */
	this._selectedLayers = null;

	/**
	 * {Array.<String>} class names that are selected
	 * @private
	 */
	this._selectedClasses = null;

	/**
	 * {Number} transparency of the CadFile
	 */
	this.alpha = 1;

	/**
	 * {XMLElement} xml used to construct the CAD file
	 */
	this._xmlData = node; //used to recontruct the cad-file when getting a deep copy, prevent call to the server

	/*
     * If the CadFile is a oneup, use the rect from its parent
     */
	if (oneup) {
		this.rect = rect;
	} else {
		//Calculate the rect based on the attributes
		this.rect = new ShapeUtils.Rect();

		var minx = ViewerUtils.parseFloat(XMLUtils.attributeValue(node, "minx"));
		var miny = ViewerUtils.parseFloat(XMLUtils.attributeValue(node, "miny"));
		var maxx = ViewerUtils.parseFloat(XMLUtils.attributeValue(node, "maxx"));
		var maxy = ViewerUtils.parseFloat(XMLUtils.attributeValue(node, "maxy"));

		this.rect.origin.x = minx;
		this.rect.origin.y = miny;
		this.rect.size.width = maxx - minx;
		this.rect.size.height = maxy - miny;
	}

	/**
	 * Original origin of the CAD
	 */
	this.originalOrigin = this.rect.origin.copy();

	/**
	 * Doc type (1 = oneup, 2 = layout)
	 */
	this.doctype = XMLUtils.attributeValue(node, "doctype");

	/*
     * When the CadFile is a oneup, take the "OneupLayer"-elements, if not, take the "Layer"-elements
     */
	var layers;
	if (oneup) {
		this.name = XMLUtils.attributeValue(node, "name");
		layers = node.getElementsByTagName("OneupLayer");
	} else {
		layers = node.getElementsByTagName("Layer");
	}

	/*
     * Variables used to calculate the content rect (rect that contains all the cad lines)
     */
	var x1 = null;
	var y1 = null;
	var x2 = null;
	var y2 = null;

	/*
     * Iterate over the layers and store them inside of the layer objects
     */
	for (var i = 0; i < layers.length; i++) {
		var clayer = layers[i];

		var lclass = XMLUtils.attributeValue(clayer, "class");
		var ltype = XMLUtils.attributeValue(clayer, "type");
		var lname = XMLUtils.attributeValue(clayer, "name");

		var l = new CadLayer(lclass, ltype, lname, oneup);

		var children = clayer.childNodes;
		for (var j = 0; j < children.length; j++) {
			var child = children[j];
			if (child.nodeType === 1) {
				var cline;
				if (child.nodeName == "Path" || child.nodeName == "Nicks") {
					cline = new CadLine(child, this.cstyles.plottingStyles, this.rect.size.height);

					if (child.nodeName == "Path" && l.layerclass != "3" && l.layerclass != "4") {
						//Snap if it's a line, but not on layers of class 3/4 (Dimensions)
						cline.snapable = true;
					}

					l.lines.push(cline);

					//Check if the x/y values are larger than the ones we already had
					if (x1) {
						x1 = Math.min(x1, cline.rect.left());
					} else {
						x1 = cline.rect.left();
					}
					if (y1) {
						y1 = Math.min(y1, cline.rect.top());
					} else {
						y1 = cline.rect.top();
					}
					if (x2) {
						x2 = Math.max(x2, cline.rect.right());
					} else {
						x2 = cline.rect.right();
					}
					if (y2) {
						y2 = Math.max(y2, cline.rect.bottom());
					} else {
						y2 = cline.rect.bottom();
					}
				} else if (child.nodeName == "TextGroup") {
					var texts = child.getElementsByTagName("Text");
					for (var t = 0; t < texts.length; t++) {
						var tspans = texts[t].getElementsByTagName("Tspan");
						for (var ts = 0; ts < tspans.length; ts++) {
							cline = new CadText(texts[t], this.cstyles.plottingStyles, ts, this.rect.size.height);
							l.lines.push(cline);
						}
					}
				} else if (child.nodeName == "Oneup") {
					//If the node is a OneUp, make a new CadFile, specifically for oneups
					var oneupFile = new CadFile(style, child, this.rect, true);
					l.oneUps.push(oneupFile);
				}
			}
		}

		this.cadlayers.push(l);

		this.contentRect = new ShapeUtils.Rect(new ShapeUtils.Point(x1, y1), new ShapeUtils.Size(x2 - x1, y2 - y1));
		this.contentRect.normalize();
	}

	// Initialize the selected layers (select all by default)
	this._selectedLayers = _.clone(this.cadlayers);

	// Initialize the selected classes (select all by default)
	this._selectedClasses = _.clone(this.getOneupClasses());
};

/**
 * Get the size of the CadFile
 * @returns {ShapeUtils.Size}
 */
CadFile.prototype.getSize = function() {
	return this.rect.size.copy();
};

/**
 * Get the origin of the CadFile
 * @returns {ShapeUtils.Point}
 */
CadFile.prototype.getOrigin = function() {
	return this.rect.origin.copy();
};

/**
 * Returns the original origin
 * @returns {ShapeUtils.Point}
 */
CadFile.prototype.getOriginalOrigin = function() {
	return this.originalOrigin;
};

/**
 * Resets the origin
 */
CadFile.prototype.resetOrigin = function() {
	this.rect.origin = this.originalOrigin.copy();
};

/**
 *
 * @param newX
 * @param newY
 */
CadFile.prototype.setOrigin = function(newX, newY) {
	this.rect.origin.x = newX;
	this.rect.origin.y = newY;
};

/**
 * get the selected layers
 * @returns {Array.<CadLayer>}
 */
CadFile.prototype.getSelectedLayers = function() {
	return this._selectedLayers;
};

/**
 * get the selected classes
 * @returns {Array.<String>}
 */
CadFile.prototype.getSelectedClasses = function() {
	return this._selectedClasses;
};

/**
 * Set the location of the CAD file in the current world
 * @param {ShapeUtils.Rect} location
 */
CadFile.prototype.setLocation = function(location) {
	this._location = location;
};

/**
 * Returns the location of the CAD file in the current world
 * @returns {ShapeUtils.Rect}
 */
CadFile.prototype.getLocation = function() {
	return this._location;
};

/**
 * Toggle the visibility of a layer
 * @param {Number} index
 * @param {Boolean} hideAllLayers
 */
CadFile.prototype.toggleLayer = function(index, hideAllLayers) {
	if (index < 0) {
		if (this._selectedLayers.length === this.cadlayers.length && hideAllLayers) {
			this._selectedLayers = [];
		} else {
			this._selectedLayers = _.clone(this.cadlayers);
		}
	} else {
		var layer = this.cadlayers[index];
		if (layer) {
			index = _.indexOf(this._selectedLayers, layer);
			if (index >= 0) {
				this._selectedLayers.splice(index, 1);
			} else {
				this._selectedLayers.push(layer);
			}
		}
	}
};

/**
 * Toggle the visibility of a class
 * @param {Number} index
 * @param {Boolean} hideAllClasses
 */
CadFile.prototype.toggleClass = function(index, hideAllClasses) {
	if (index < 0) {
		if (this._selectedClasses.length === this.getOneupClasses().length && hideAllClasses) {
			this._selectedClasses = [];
		} else {
			this._selectedClasses = _.clone(this.getOneupClasses());
		}
	} else {
		var classname = this.getOneupClasses()[index];
		if (classname) {
			var i = _.indexOf(this._selectedClasses, classname);
			if (i < 0) {
				this._selectedClasses.push(classname);
			} else {
				this._selectedClasses.splice(i, 1);
			}
		}
	}
};

/**
 * hide all classes
 */
CadFile.prototype.hideClasses = function() {
	this._selectedClasses = [];
};

/**
 * hide all layers
 */
CadFile.prototype.hideLayers = function() {
	this._selectedLayers = [];
};

/**
 * Get the union of all the layer classes of the oneups
 * @returns {Array.String}
 */
CadFile.prototype.getOneupClasses = function() {
	if (!this.oneupClasses) {
		//calculate the layer names when requested the first time
		//Fetch all the unique layer classes first
		var classes = [];
		_.each(this.cadlayers, function(layer) {
			_.each(layer.oneUps, function(oneup) {
				_.each(oneup.cadlayers, function(oneupLayer) {
					var lname = oneupLayer.layerclass;
					if (_.indexOf(classes, lname) < 0) {
						classes.push(lname);
					}
				});
			});
		});

		this.oneupClasses = classes;
	}

	return this.oneupClasses;
};

export default CadFile;
