import _ from "underscore";
import LayerModel from "@esko/web-viewer/viewport/model/ViewportLayerModel";
import LineworkWrapper from "@esko/web-viewer/utils/LineworkWrapper";
import LineworkPoint from "@esko/web-viewer/utils/LineworkPoint";
import ShapeUtils from "@esko/shapeutils";
import ViewerUtils from "@esko/web-viewer/utils/ViewerUtils";
/**
 * @exports ViewserverCADLayerModel
 * @class The cad layer model keeps all the info for a cad document on a layer
 * @augments Viewport.LayerModel
 * @constructor
 * @param {Core.ViewerFacade} f
 * @param {String} ID
 * @param {CadModel} cadModel
 * @param {Boolean} isForReferencePage
 */
var CadLayerModel = function(f, ID, cadModel, isForReferencePage) {
	LayerModel.call(this, f, ID);

	this.cadModel = cadModel;

	this._isForReferencePage = isForReferencePage;

	this._buffer = null;

	//Take the offset of the cad file itself first
	this._offsetX = -this.cadModel.getCad().getOrigin().x;
	this._offsetY = this.cadModel.getCad().getOrigin().y;

	this._calculateTransformation();

	//Register for model change events
	this.cadModel.on(
		"modelChanged",
		_.bind(function() {
			this._calculateTransformation();
		}, this)
	);
};

// inherit
CadLayerModel.prototype = Object.create(LayerModel.prototype);
// correct the constructor pointer
CadLayerModel.prototype.constructor = CadLayerModel;

/**
 * returns the transparency of the layer
 * @returns {Number} alpha
 */
CadLayerModel.prototype.getAlpha = function() {
	return this.cadModel.getAlpha();
};

/**
 * Check if the layer model has CAD
 * @returns {Boolean}
 */
CadLayerModel.prototype.hasLinework = function() {
	return true;
};

/**
 * Does the layermodel have a buffer
 * @returns {Boolean}
 */
CadLayerModel.prototype.hasBuffer = function() {
	return this._buffer !== null;
};

/**
 * Returns the layers's buffer
 * @returns {null|Fragment}
 */
CadLayerModel.prototype.getBuffer = function() {
	return this._buffer;
};

/**
 * Set the buffer for the layer
 * @param {null|Fragment} buffer
 */
CadLayerModel.prototype.setBuffer = function(buffer) {
	this._buffer = buffer;
};

/**
 * Get the color for the cad linework
 * @returns {String} the color "#RGB"
 */
CadLayerModel.prototype.getLineworkColor = function() {
	return this.cadModel.getCadColor();
};

/**
 * @override
 */
CadLayerModel.prototype.getLinework = function(isAnimating, isAnimatingRotation, drawrect) {
	if (drawrect !== undefined && this.cadModel.getCad() !== null) {
		var cadSize = this.cadModel.getCad().getSize();
		var originLineX = new LineworkWrapper();

		var startX = new LineworkPoint();
		startX.makeMoveTo(0, drawrect.top());
		originLineX.pointArray.push(startX);
		var endX = new LineworkPoint();
		endX.makeLineTo(0, drawrect.bottom());
		originLineX.pointArray.push(endX);

		var originLineY = new LineworkWrapper();

		var startY = new LineworkPoint();
		startY.makeMoveTo(drawrect.left(), cadSize.height);
		originLineY.pointArray.push(startY);
		var endY = new LineworkPoint();
		endY.makeLineTo(drawrect.right(), cadSize.height);
		originLineY.pointArray.push(endY);

		originLineX.color = "#" + ViewerUtils.rgbToHex(0, 0, 255);
		originLineY.color = "#" + ViewerUtils.rgbToHex(0, 0, 255);
		originLineX.dash = [5, 4, 2, 4];
		originLineY.dash = [5, 4, 2, 4];

		var cadLines = this.cadModel.getVisibleCadLines().slice(0);
		cadLines.push(originLineX);
		cadLines.push(originLineY);

		return cadLines;
	} else {
		return this.cadModel.getVisibleCadLines();
	}
};

/**
 * @override
 */
CadLayerModel.prototype.getBackgroundOptions = function() {
	var options = {};

	//Get the background of the cad layer. This is typically 0 (transparent) when showing cad over graphics. And 1 for only cad
	//This could also be another value, example is when the graphics need to be dimmed to highlight the CAD more
	options.fillStyle = this.cadModel.getBackgroundString();

	return options;
};

/**
 * Get the rotation of the page in the CadModel
 * @returns {Number}
 */
CadLayerModel.prototype.getRotation = function() {
	return this.cadModel.getPage().getRotation();
};

/**
 * Is the page in the CadModel flipped
 * @returns {Boolean}
 */
CadLayerModel.prototype.isFlipped = function() {
	return this.cadModel.getPage().isFlipped();
};

/**
 * Calculates the transformation for the CAD layers (current CAD and reference CAD)
 */
CadLayerModel.prototype._calculateTransformation = function() {
	this.resetTransformation(); //first reset the transformation and calculate the new one

	var page = this.cadModel.getPage(); //get CAD page for which this layer draws the linework
	page.getCad().setLocation(null); //set the location of the CAD to null

	this._offsetX = -this.cadModel.getCad().getOrigin().x;
	this._offsetY = this.cadModel.getCad().getOrigin().y;

	var offset = new ShapeUtils.Size(this._offsetX, this._offsetY);
	if (page.hasStateParam("compareOffset")) {
		//if there is a compare offset defined add it to the general offset
		var compareOffset = page.getStateParam("compareOffset");
		offset.width -= compareOffset.x;
		offset.height += compareOffset.y;
	}

	var rotation = this.cadModel.getPage().getRotation();

	var isFlipped = this.cadModel.getPage().isFlipped();
	var worldBounds = this.f.getModule("viewport").getWorldBounds();
	if (!this._isForReferencePage) {
		//if the layer is not displaying the reference document page CAD

		this.calculateTransformation(worldBounds, page.getPageSize(), offset, rotation, isFlipped, 0, false);
	} else {
		// the layer is displaying the CAD of the reference document page
		var activeOrigin = this.f.getActivePage().getOrigin(true);
		var referenceOrigin = page.getOrigin(true);

		var activePageSize = this.f.getActivePage().getPageSize();
		var referencePageSize = page.getPageSize();

		//calculate the point it needs to rotate around (the centre of the active docuemnt page)
		var rotatePoint = new ShapeUtils.Point(activePageSize.width / 2, activePageSize.height / 2);
		//calculate the point to flip the reference document page around (the centre of the reference document page)
		var flipPoint = new ShapeUtils.Point(-activeOrigin.x + referenceOrigin.x + referencePageSize.width / 2, 0);
		this.calculateTransformation(worldBounds, page.getPageSize(), offset, 0, false, rotation, isFlipped, rotatePoint, flipPoint);

		//now pass the rectangle of the CAD through the transformation to get the new CAD rectangle (and set the location in the CadFIle)
		var origin = page.getCad().getOrigin();
		//need to get the origin point of the CAD relative to the world (0,0) point
		var point = new ShapeUtils.Point(
			origin.x - activeOrigin.x + referenceOrigin.x,
			-origin.y + activePageSize.height + activeOrigin.y - referenceOrigin.y
		);
		referencePageSize.height = -referencePageSize.height;
		var rect = new ShapeUtils.Rect(point, referencePageSize);

		var newRect = this.getTransform().transformedRect(rect);
		//set the new rect as location in the CadFile
		page.getCad().setLocation(newRect);
	}

	this.modelChanged();
};

export default CadLayerModel;
