import LineworkWrapper from "@esko/web-viewer/utils/LineworkWrapper";
import LineworkPoint from "@esko/web-viewer/utils/LineworkPoint";
import ShapeUtils from "@esko/shapeutils";
import XMLUtils from "@esko/xmlutils";
import ViewerUtils from "@esko/web-viewer/utils/ViewerUtils";
import DrawUtils from "@esko/web-viewer/utils/DrawUtils";
import CadConstants from "@esko/web-viewer/cad/CadConstants";
/**
 * @exports CadLine
 * @memberof CadModule
 * @class CadLine is a LineworkWrapper object that is used to display a line within a CadFile
 * @augments Utils.LineworkWrapper
 * @constructor
 * @param {XMLElement} node - xml to construct the CadLine
 * @param {Object} styles - plotting styles to use to construct the CadFile (colors/width/...)
 * @param {Number} height - height of the cad file to calculate the point on normal axis
 */
var CadLine = function(node, styles, height) {
	LineworkWrapper.call(this);

	var linename = node.nodeName;
	if (linename == "Nicks") {
		this.type = CadConstants.CadTypes.Nick;
		this.linewidth = 1;
		this.color = "black";
	} else {
		this.type = XMLUtils.attributeValue(node, "linetype");
	}
	var bopt = XMLUtils.attributeValue(node, "bridgeopt");
	if (bopt) {
		this.bridgeopt = parseInt(bopt, 10);
	}

	if (this.type === null) {
		var strokeColor = XMLUtils.attributeValue(node, "stroke");
		if (strokeColor !== null) {
			this.color = strokeColor;
		}
	} else {
		var pstyle = styles[this.type];
		if (pstyle) {
			this.c = pstyle.color; //color
			this.color = "#" + ViewerUtils.rgbToHex(this.c.r, this.c.g, this.c.b);
			this.linewidth = pstyle.width / 4;
			if (pstyle.style !== "0") {
				this.dash = CadConstants.LineTypes[pstyle.style];

				this.lineDecorator = pstyle.style;
			}
		} else {
			// plotting style is not defined... do not draw
			this.NODRAW = true;
		}
	}

	//test if fill color is defined for the CADLine, specify a fill color for the lineworkWrapper
	this.fillcolor = XMLUtils.attributeValue(node, "fill");
	if (this.fillcolor === "none") {
		//if "none" is specified as fill color set fillcolor to null
		this.fillcolor = null;
	}
	//if there is a stroke defined as none set the linewidth to the minimum (linethickness 1)
	if (this.fillcolor && XMLUtils.attributeValue(node, "stroke") === "none") {
		this.linewidth = 1;
	}

	var path = XMLUtils.attributeValue(node, "d");
	this.pointArray = [];

	var rectx1 = null;
	var rectx2 = null;
	var recty1 = null;
	var recty2 = null;

	var cmdRegEx = /[a-z][^a-z]*/gi;
	var pathchuncs = path.match(cmdRegEx);
	for (var i = 0; i < pathchuncs.length; i++) {
		var p = new LineworkPoint();

		var chunc = ViewerUtils.trim(pathchuncs[i]);

		var minXPoint = null;
		var maxXPoint = null;
		var minYPoint = null;
		var maxYPoint = null;

		var command = chunc.charAt(0);
		if (command == "A") {
			//A 22.00 22.00 0 0 0 13.585 323.300
			var arcregex = new RegExp(/A\s(.+)\s(.+)\s(.+)\s(.+)\s(.+)\s(.+)\s(.+)/);
			var arccommand = arcregex.exec(chunc);

			var rx = ViewerUtils.parseFloat(arccommand[1]);
			var ry = ViewerUtils.parseFloat(arccommand[2]);
			var angle = ViewerUtils.parseFloat(arccommand[3]);
			var largeArcFlag = ViewerUtils.parseFloat(arccommand[4]);
			var sweepFlag = ViewerUtils.parseFloat(arccommand[5]);
			var x1 = ViewerUtils.parseFloat(arccommand[6]);
			var y1 = ViewerUtils.parseFloat(arccommand[7]);

			p.makeArc(x1, height - y1, rx, ry, angle, largeArcFlag, sweepFlag);

			minXPoint = p.x;
			maxXPoint = p.x;
			minYPoint = p.y;
			maxYPoint = p.y;
		} else if (command == "C") {
			var circleregex = new RegExp(/C\s(.+)\s(.+)\s(.+)\s(.+)\s(.+)\s(.+)/);
			var circlecommand = circleregex.exec(chunc);

			var x1 = ViewerUtils.parseFloat(circlecommand[1]);
			var y1 = ViewerUtils.parseFloat(circlecommand[2]);
			var x2 = ViewerUtils.parseFloat(circlecommand[3]);
			var y2 = ViewerUtils.parseFloat(circlecommand[4]);
			var x3 = ViewerUtils.parseFloat(circlecommand[5]);
			var y3 = ViewerUtils.parseFloat(circlecommand[6]);

			//calculate the min and max x/y points for the rectangle that contains the bezier-curve
			//A bezier curve is completely contained in the convex hull of its control points
			minXPoint = Math.min(x1, x2, x3);
			maxXPoint = Math.max(x1, x2, x3);
			minYPoint = Math.min(height - y1, height - y2, height - y3); //the origin of CAD has inverted Y-axis
			maxYPoint = Math.max(height - y1, height - y2, height - y3);

			p.makeBezier(x1, height - y1, x2, height - y2, x3, height - y3);
		} else if (command == "N") {
			var nickregex = new RegExp(/N(.+)\s(.+)\s(.+)\s(.+)\s(.+)/);
			var nickcommand = nickregex.exec(chunc);
			var v1 = ViewerUtils.parseFloat(nickcommand[1]);
			var v2 = ViewerUtils.parseFloat(nickcommand[2]);
			var v3 = ViewerUtils.parseFloat(nickcommand[3]);
			var v4 = ViewerUtils.parseFloat(nickcommand[4]);
			var v5 = ViewerUtils.parseFloat(nickcommand[5]);

			p.makeNick(v1, height - v2, v3, v4, v5);

			minXPoint = p.x;
			maxXPoint = p.x;
			minYPoint = p.y;
			maxYPoint = p.y;
		} else if (command == "L" || command == "M") {
			var lineregex = new RegExp(/.(.+)\s(.+)/);
			var linecommand = lineregex.exec(chunc);
			var x = ViewerUtils.parseFloat(linecommand[1]);
			var y = height - ViewerUtils.parseFloat(linecommand[2]);

			if (command == "L") {
				p.makeLineTo(x, y);
			} else {
				p.makeMoveTo(x, y);
			}

			minXPoint = p.x;
			maxXPoint = p.x;
			minYPoint = p.y;
			maxYPoint = p.y;
		}
		if (p.type != -1) {
			this.pointArray.push(p);

			if (rectx1 === null || minXPoint < rectx1) {
				rectx1 = minXPoint;
			}
			if (rectx2 === null || maxXPoint > rectx2) {
				rectx2 = maxXPoint;
			}
			if (recty1 === null || minYPoint < recty1) {
				recty1 = minYPoint;
			}
			if (recty2 === null || maxYPoint > recty2) {
				recty2 = maxYPoint;
			}
		}
	}
	var r = new ShapeUtils.Rect(new ShapeUtils.Point(rectx1, recty1), new ShapeUtils.Size(rectx2 - rectx1, recty2 - recty1));
	r.normalize();
	this.rect = r;
};
// inherit
CadLine.prototype = Object.create(LineworkWrapper.prototype);
// correct the constructor pointer
CadLine.prototype.constructor = CadLine;

export default CadLine;
