import PrivateChannel from "@esko/web-viewer/core/PrivateChannel";
import CadConstants from "@esko/web-viewer/cad/CadConstants";
import CadOneupName from "@esko/web-viewer/cad/CadOneupName";
import ViewerUtils from "@esko/web-viewer/utils/ViewerUtils";
import _ from "underscore";
/**
 * @exports CadModel
 * @memberof CadModule
 * @class The cad model is the main model within the viewer that keeps track of the current cad and their layer visibility.
 * @constructor
 * @param {Core.ViewerFacade} f
 * @param {String} ID - used for the private channel of the CadModel
 */
var CadModel = function(f, ID) {
	PrivateChannel.installTo(this, f, ["cadModel", ID]);

	/**
	 * Facade
	 */
	this.f = f;

	/**
	 * {CadFile} Cad object for the model
	 * @private
	 */
	this._cad = null;

	/**
	 * {AbstractDocumentPage} The page which is used for the CAD
	 * @private
	 */
	this._page = null;

	/**
	 * {Boolean} indicated if the cad is the only content of the file
	 * @private
	 */
	this._onlycad = false;

	/**
	 * {Number} Visibility of Bridges and Nicks. 1: none, 2: only bridges, 3: only nicks, 4: both
	 * @private
	 */
	this._bridgesAndNicks = CadConstants.ShowBridgesNicks.NONE;

	/**
	 * {Number} Visibility of Cad and Graphics. 1: both, 2: only graphics, 3: only cad
	 * @private
	 */
	this._cadGraphicsMode = 1;

	/**
	 * {Number} Visibility of graphics. 1:normal, 2: lighten, 3: lighten more
	 * @private
	 */
	this._graphicsMode = 1;

	/**
	 * {Boolean} true if the viewport should show oneupnames
	 * @private
	 */
	this._showOneupNames = false; //Show oneup names

	/**
	 * {Array.<CadLine/CadText>} List of currently showed cad lines
	 * @private
	 */
	this._currentCadLines = null;

	/**
	 * @member {String} color of the cad linework ('#' + RGB string)
	 * @private
	 */
	this._cadColor = null;

	/**
	 * @member {String} color of the cad background in hex format (without #)
	 * @private
	 */
	this._cadBackgroundColor = null;
};

/**
 * Get the cad file
 * @returns {CadFile} Cad File
 */
CadModel.prototype.getCad = function() {
	return this._cad;
};

/**
 * @returns {Boolean} true if the cad is the only content
 */
CadModel.prototype.isOnlyCad = function() {
	return this._onlycad;
};

/**
 * @returns {Array.CadLine|CadText} the visible Cad Lines
 */
CadModel.prototype.getVisibleCadLines = function() {
	return this._currentCadLines;
};

/**
 * Set the cad file that is represented in the model
 * @param {CadFile} cad - cad file
 * @para {Boolean} onlyCad - is there only CAD (white background) or are there graphics
 * @param {ShapeUtils.Size} offset - offset for the CAD (when there is graphics with CAD)
 * @param {AbstractDocumentPage} page - the page set in the CadModel
 */
CadModel.prototype.setCad = function(cad, onlyCad, offset, page) {
	this._cad = cad;

	this._onlycad = onlyCad;

	this.offset = offset;

	this._page = page;

	this.invalidateLayers();
};

/**
 * Get the document page where the CAD sits in
 * @returns {AbstractDocumentPage}
 */
CadModel.prototype.getPage = function() {
	return this._page;
};

/**
 * Invalidate the current layers and reload the new ones
 */
CadModel.prototype.invalidateLayers = function() {
	this._currentCadLines = this._getLinesForSelectedLayers();
	this._emit("modelChanged");
};

/**
 * fire model changed on the cad model
 */
CadModel.prototype.modelChanged = function() {
	this._emit("modelChanged");
};

/**
 * Returns the CAD offset, returns null if there is no offset
 * @returns {Size|null}
 */
CadModel.prototype.getCadOffset = function() {
	if (this.offset !== undefined) {
		return this.offset;
	} else {
		return null;
	}
};

/**
 * Get the cad lines for the selected layers
 * @private
 * @returns {Array.<CadLine|CadText>}
 */
CadModel.prototype._getLinesForSelectedLayers = function() {
	var lines = [];
	for (var i = 0; i < this._cad.cadlayers.length; i++) {
		var l = this._cad.cadlayers[i];
		if (_.indexOf(this._cad.getSelectedLayers(), l) > -1) {
			lines.push(this._getCadLinesForLayer(l));

			//Iterate over the layer oneups and check if they need to be visible
			_.each(
				l.oneUps,
				_.bind(function(oneup) {
					_.each(
						oneup.cadlayers,
						_.bind(function(layer) {
							if (_.indexOf(this._cad.getSelectedClasses(), layer.layerclass) >= 0) {
								lines.push(this._getCadLinesForLayer(layer));
							}
						}, this)
					);

					if (this._showOneupNames) {
						var oneupName = new CadOneupName(oneup.name, oneup.contentRect);
						lines.push(oneupName);
					}
				}, this)
			);
		}
	}
	lines = _.flatten(lines);
	lines = _.compact(lines);
	return lines;
};

/**
 * Get CadLine/CadText for one specific layer
 * @private
 * @param {CadLayer} layer
 * @returns {Array.<CadLine|CadText>}
 */
CadModel.prototype._getCadLinesForLayer = function(layer) {
	var array = [];
	for (var j = 0; j < layer.lines.length; j++) {
		var lwwrapper = layer.lines[j];
		var add = false;

		if (lwwrapper.type === CadConstants.CadTypes.Nick) {
			var showNicks =
				this._bridgesAndNicks === CadConstants.ShowBridgesNicks.NICKS || this._bridgesAndNicks === CadConstants.ShowBridgesNicks.BOTH;
			add = showNicks;
		} else if (lwwrapper.bridgeopt) {
			var showBridges =
				this._bridgesAndNicks === CadConstants.ShowBridgesNicks.BRIDGES || this._bridgesAndNicks === CadConstants.ShowBridgesNicks.BOTH;

			if (layer.lines[j].bridgeopt === 1) {
				if (showBridges) {
					add = true;
				}
			} else {
				if (!showBridges) {
					add = true;
				}
			}
		} else {
			add = true;
		}

		if (add) {
			array.push(lwwrapper);
		}
	}
	if (array.length === 0) {
		array = layer.lines;
	}
	return array;
};

/**
 * Get all the cad layers
 * @returns {Array.<CadFile.layer>}
 */
CadModel.prototype.getAllCadLayers = function() {
	return this._cad.cadlayers;
};

/**
 * Get all the cad classes
 * @returns {Array.<String>}
 */
CadModel.prototype.getAllClasses = function() {
	return this._cad.getOneupClasses();
};

/**
 * @returns {Array.<CadFile.layer>} layers that are currently selected
 */
CadModel.prototype.getSelectedLayers = function() {
	return this._cad.getSelectedLayers();
};

/**
 * Toggle a layers visibility.
 * @param {Number} index - index of the layer (-1 for all)
 * @param {Boolean} [hideAllLayers] - if index = -1 define if layers should be hidden or not
 */
CadModel.prototype.toggleLayer = function(index, hideAllLayers) {
	this._cad.toggleLayer(index, hideAllLayers);
	this.invalidateLayers();
};

/**
 * Hides all the layers
 */
CadModel.prototype.hideLayers = function() {
	this._cad.hideLayers();
	this.invalidateLayers();
};

/**
 * Set the visibility of bridges and nicks
 * @param {Number} 1: none, 2: only bridges, 3: only nicks, 4: both
 */
CadModel.prototype.setBridgesAndNicks = function(bridgesAndNicks) {
	var oldBridgesAndNicks = this._bridgesAndNicks;
	if (oldBridgesAndNicks !== bridgesAndNicks) {
		this._bridgesAndNicks = bridgesAndNicks;
		this.invalidateLayers();
	}
};

/**
 * Returns the visibility of the nicks/bridges
 * @returns {Number} 1: none, 2: only bridges, 3: only nicks, 4: both
 */
CadModel.prototype.getBridgesAndNicksVisibility = function() {
	return this._bridgesAndNicks;
};

/**
 * Show the layer selector
 * @returns {Boolean} true if the layer dialog should be visible
 */
CadModel.prototype.showLayers = function() {
	return true;
};

/**
 * Show the class selector
 * @returns {Boolean} true if the class dialog should be visible
 */
CadModel.prototype.showClasses = function() {
	return this._cad.doctype == CadConstants.CadDocType.Layout && this._cad.getOneupClasses().length > 0;
};

/**
 * Get the list of the currently selected classes
 * @returns {Array.<String>} selected classes
 */
CadModel.prototype.getSelectedClasses = function() {
	return this._cad.getSelectedClasses();
};

/**
 * Toggle a class visibility.
 * @param {Number} classindex - index of the class (-1 for all)
 * @param {Boolean} [hideAllClasses] - if index = -1 define if layers should be hidden or not
 */
CadModel.prototype.toggleClass = function(index, hideAllClasses) {
	this._cad.toggleClass(index, hideAllClasses);

	this.invalidateLayers();
};

/**
 * Hides all the classes
 */
CadModel.prototype.hideClasses = function() {
	this._cad.hideClasses();
	this.invalidateLayers();
};

/**
 * Toggle the visibility of the oneup names
 * @param {Boolean} [toggle] True to show it, false otherwise
 */
CadModel.prototype.toggleShowOneupNames = function(toggle) {
	if (toggle !== undefined) {
		if (this._showOneupNames !== toggle) {
			this._showOneupNames = toggle;
			this.invalidateLayers();
		}
	} else {
		this._showOneupNames = !this._showOneupNames;
		this.invalidateLayers();
	}
};

/**
 * @returns {Boolean} true if the UI should show the oneup names
 */
CadModel.prototype.isShowOneupNames = function() {
	return this.canShowOneupNames() && this._showOneupNames;
};

/**
 * @returns {Boolean} true if the UI can show oneup names (typically only layout files)
 */
CadModel.prototype.canShowOneupNames = function() {
	return this._cad && this._cad.doctype == CadConstants.CadDocType.Layout;
};

/**
 * Set the graphics mode
 * @param {Number} mode - 1:normal, 2: lighten, 3: lighten more
 */
CadModel.prototype.setGraphicsMode = function(mode) {
	this._graphicsMode = mode;
	this._emit("modelChanged");
};

/**
 * Get the graphics mode
 * @returns {Number} mode - 1:normal, 2: lighten, 3: lighten more
 */
CadModel.prototype.getGraphicsMode = function() {
	return this._graphicsMode;
};

/**
 * Set the color of the cad in the model
 * @param {String} cadColor - '#' + RGB string
 */
CadModel.prototype.setCadColor = function(_cadColor) {
	this._cadColor = _cadColor;
	this._emit("modelChanged");
};

/**
 * Returns the cad linework color
 * @returns {String}
 */
CadModel.prototype.getCadColor = function() {
	return this._cadColor;
};

/**
 * Resets the color for the cad linework
 */
CadModel.prototype.resetCadColor = function() {
	this._cadColor = null;
};

/**
 * Get the cad graphics mode
 * @returns {Number} 1: both, 2: only graphics, 3: only cad
 */
CadModel.prototype.getCadGraphicsMode = function() {
	return this._cadGraphicsMode;
};

/**
 * Set the cad graphics mode
 * @param mode
 */
CadModel.prototype.setCadGraphicsMode = function(mode) {
	if (mode !== this._cadGraphicsMode) {
		this._cadGraphicsMode = mode;

		this._emit("modelChanged");
	}
};

/**
 * Get the alpha value of the background of the cad line
 * @returns {Number} - alpha value that needs to be used to display the cad background
 * - When only cad is shown OR cad graphics mode is to show only CAD, return 1 (full white background)
 * - When the graphics mode is "lighten", return an alpha of .4
 * - When the graphics mode is "lighten more", return an alpha of .8
 */
CadModel.prototype.getBackgroundAlpha = function() {
	if (this._onlycad || this._cadGraphicsMode === 3) {
		return 1;
	} else if (this.getGraphicsMode() === 2) {
		return 0.4;
	} else if (this.getGraphicsMode() === 3) {
		return 0.8;
	} else {
		return 0;
	}
};

/**
 * Get the cad background color that needs to be displayed.
 * @returns {String} - Color in hex format (without #).
 */
CadModel.prototype.getBackgroundColor = function() {
	if (this._cadBackgroundColor === null) {
		return "FFFFFF"; // default white
	} else {
		return this._cadBackgroundColor;
	}
};

/**
 * Set the cad background color
 * @param {String} [hexColor] - Background color in hex format (without #). If not set, it will reset to the default (white)
 */
CadModel.prototype.setBackgroundColor = function(hexColor) {
	if (hexColor === undefined) {
		hexColor = "FFFFFF";
	}
	if (this._cadBackgroundColor !== hexColor) {
		this._cadBackgroundColor = hexColor;
		this._emit("modelChanged");
	}
};

/**
 * Get the background as string "rgba(r,g,b,a)".
 * @returns {String}
 */
CadModel.prototype.getBackgroundString = function() {
	var backgroundAlpha = this.getBackgroundAlpha();
	var backgroundColor = this.getBackgroundColor();
	var rgbBackgroundColor = ViewerUtils.rgbColor(backgroundColor);
	if (rgbBackgroundColor !== null) {
		return "rgba(" + rgbBackgroundColor.r + "," + rgbBackgroundColor.g + "," + rgbBackgroundColor.b + "," + backgroundAlpha + ")";
	} else {
		return "rgba(255,255,255," + backgroundAlpha + ")"; // white
	}
};

/**
 * Get the transparency of the CAD
 * @returns {Float}
 */
CadModel.prototype.getAlpha = function() {
	return this._cadAlpha;
};

/**
 * Set the transparence of the CAD
 * @param {Number} cadAlpha
 */
CadModel.prototype.setAlpha = function(cadAlpha) {
	this._cadAlpha = cadAlpha;
};

export default CadModel;
