/**
 *  @namespace CadModule
 */
/**
 *  @namespace CadModule
 */
import _ from "underscore";

import AbstractModule from "@esko/web-viewer/core/AbstractModule";
import CadLayerModel from "@esko/web-viewer/cad/CadLayerModel";
import CadOverlayLayerModel from "@esko/web-viewer/cad/CadOverlayLayerModel";
import CadModel from "@esko/web-viewer/cad/CadModel";
import CadWidget from "@esko/web-viewer/cad/CadWidget";
import CadConstants from "@esko/web-viewer/cad/CadConstants";
import CompareConstants from "@esko/web-viewer/compare/CompareConstants";
import ListenerUtil from "@esko/web-viewer/utils/ListenerUtil";
import CadDataExensions from "@esko/web-viewer/cad/AbstractCadDataExtensions";
/**
 * @exports CadModule
 * @memberof CadModule
 * @class The cad module is the main module that controls the loading of cad and displaying
 * @augments Core.AbstractModule
 * @param {Core.ViewerFacade} f
 * @param {Boolean} silenced
 * @constructor
 */
var CadModule = function(f, silenced) {
	AbstractModule.call(this, f, silenced, "cad");
	ListenerUtil.mixin(this, f);
	CadDataExensions(f);
	this.f = f;

	/**
	 * CadModel
	 * @private
	 */
	this._cadModel = new CadModel(this.f, "MainCad");

	/**
	 * CadModel for difference compare
	 * @private
	 */
	this._referenceDiffCadModel = new CadModel(this.f, "referenceDiffCad");

	/**
	 * @member {CadModel} CadModel for side by side, reference document page
	 * @private
	 */
	this._referenceCadModel = new CadModel(this.f, "referenceCad");

	/**
	 * @member {CadModel} CadModel for difference compare in side by side, current document page
	 * @private
	 */
	this._currentDiffCadModel = new CadModel(this.f, "currentDiffCad");

	/**
	 * @member {CadModel} CadModel for the cad overlay
	 * @private
	 */
	this._cadOverlayModel = new CadModel(this.f, "cadOverlay");

	/**
	 * CadWidget
	 * @private
	 */
	this._widget = null;

	/**
	 * This is the model currently assigned to the CAD Widget
	 * @private
	 **/
	this._cadModelForWidget = this._cadModel;

	/**
	 * This is the model currently assigned to the CAD Widget for the reference document page
	 * @private
	 **/
	this._referenceCadModelForWidget = null;

	/**
	 * Layer models per view
	 * @private
	 */
	this._layerModels = {};

	/**
	 * Fetched cadStyles, cache to prevent multiple calls for the same cadStyle
	 * @private
	 */
	this._cadStyles = {};

	this.compareDiffMode = CompareConstants.diffMode.UNDEFINED;
	this.f.whenModuleAvailable(
		"compare",
		function(compare) {
			this.compareDiffMode = compare.getDiffMode();
		}.bind(this)
	);

	this.f.on(
		"module.compare.changed.diffMode",
		_.bind(function(oldMode, newMode) {
			if (this.compareDiffMode !== newMode) {
				this._whenCompareModeChanged(this.compareDiffMode, newMode);
			}
		}, this)
	);

	this.f.getCurrentUser().on(
		"options.changed.PREFERENCES_HIGHLIGHTCOLORS_DIFF_ADD",
		_.bind(function() {
			if (
				(this.compareDiffMode === CompareConstants.diffMode.SHOWBOTHCHANGES ||
					this.compareDiffMode === CompareConstants.diffMode.SHOWBOTHSIDEBYSIDE) &&
				this._referenceDiffCadModel !== null
			) {
				var color = this.f.getCurrentUser().getColor("PREFERENCES_HIGHLIGHTCOLORS_DIFF_ADD");
				if (this.compareDiffMode === CompareConstants.diffMode.SHOWBOTHSIDEBYSIDE) {
					this._currentDiffCadModel.setCadColor(color);
				} else {
					this._cadModel.setCadColor(color);
				}
			}
		}, this)
	);

	this.f.getCurrentUser().on(
		"options.changed.PREFERENCES_HIGHLIGHTCOLORS_DIFF_SUB",
		_.bind(function() {
			if (
				(this.compareDiffMode === CompareConstants.diffMode.SHOWBOTHCHANGES ||
					this.compareDiffMode === CompareConstants.diffMode.SHOWBOTHSIDEBYSIDE) &&
				this._cadModel !== null
			) {
				var color = this.f.getCurrentUser().getColor("PREFERENCES_HIGHLIGHTCOLORS_DIFF_SUB");
				this._referenceDiffCadModel.setCadColor(color);
			}
		}, this)
	);

	var docController = this.f.getDocumentController();

	// initialize the CAD before parameters available (and endUpdate, so the viewport will use correct info for verifyWorld)
	docController.on(
		"parameters.beforeEndUpdate",
		_.bind(function(activeChanged, differenceChanged, cadOverlayChanged) {
			var docController = this.f.getDocumentController();
			_.each(
				[docController.getActivePage(), docController.getDifferencePage(), docController.getCadOverlay()],
				function(page) {
					if (page && page.hasCad()) {
						page.initializeCadFile();
					}
				},
				this
			);

			if (activeChanged || differenceChanged || cadOverlayChanged) {
				this._whenPagesChangedAndParametersAvailable();
			} else if (this._layerModels["CadOverlay"]) {
				this._cadOverlayModel.modelChanged();
			}
		}, this)
	);

	docController.whenReferencePageStateParametersChange(function(activePage, differencePage) {
		if (this._layerModels["referenceDiffCad"]) {
			this._referenceDiffCadModel.modelChanged();
		}
		if (this._layerModels["referenceCad"]) {
			this._referenceCadModel.modelChanged();
		}
	}, this);

	docController.whenCadOverlayPageStateParametersChange(function(activePage, differencePage, cadOverlay) {
		if (this._layerModels["CadOverlay"]) {
			this._cadOverlayModel.modelChanged();
		}
	}, this);

	if (!this.f.isParameterUpdateOngoing() && this.f.getActivePage()) {
		this._whenPagesChangedAndParametersAvailable();
	}
};

// inherit
CadModule.prototype = Object.create(AbstractModule.prototype);
// correct the constructor pointer
CadModule.prototype.constructor = CadModule;

/**
 * Get an array with all the pages (either active, difference or overlay) having CAD
 * @returns {Array.<AbstractDocumentPage>} list of pages having CAD
 */
CadModule.prototype.getAllActivePagesHavingCad = function() {
	var docController = this.f.getDocumentController();
	return _.filter([docController.getActivePage(), docController.getDifferencePage(), docController.getCadOverlay()], function(page) {
		return page && page.hasCad();
	});
};

/**
 * Execute a callback function when all CAD files are successfully initialized and available
 * @param   {Function} callback - the callback, called when all cad files are loaded
 * @param   {Object}   [binding]  - binding for the callback
 */
CadModule.prototype.whenAllCadFilesInitialized = function(callback, binding) {
	if (binding) {
		callback = _.bind(callback, binding);
	}

	var pagesHavingCad = this.getAllActivePagesHavingCad();

	if (pagesHavingCad.length > 0) {
		callback = _.after(pagesHavingCad.length, callback);
		_.each(pagesHavingCad, function(page) {
			page.requestCad(callback);
		});
	} else {
		// no CAD files are loaded, still execute the callback
		callback();
	}
};

/**
 * Initialize the cad layer and widget when a document with cad data is loaded
 * @param {AbstractDocumentPage} currentPage - New page that is loaded
 */
CadModule.prototype._whenPagesChangedAndParametersAvailable = function(activePage, differencePage, cadOverlay) {
	this.whenAllCadFilesInitialized(this._updateCadRendering, this);
};

/**
 * React on a change of compare mode, if the cadModule wasn't initialized don't execute it
 * @private
 * @param {CompareConstants.diffMode} oldMode
 * @param {CompareConstants.diffMode} newMode
 */
CadModule.prototype._whenCompareModeChanged = function(oldMode, newMode) {
	if (newMode === CompareConstants.diffMode.SHOWBOTHCHANGESHIGHLIGHT || newMode === CompareConstants.diffMode.SHOWBOTHCHANGESONLY) {
		newMode = CompareConstants.diffMode.SHOWBOTHCHANGES;
	} else if (
		newMode === CompareConstants.diffMode.SHOWBOTHSIDEBYSIDEHIGHLIGHT ||
		newMode === CompareConstants.diffMode.SHOWBOTHSIDEBYSIDECHANGESONLY
	) {
		newMode = CompareConstants.diffMode.SHOWBOTHSIDEBYSIDE;
	}
	this.compareDiffMode = newMode;

	this.whenAllCadFilesInitialized(this._updateCadRendering, this);
};

/**
 * Update the rendering of CAD (configure CadModels correctly, add/remove layers, update colors for CadModels)
 * !!NOTE this method should only be called when CAD is available for all loaded documetns (active, difference, overlay)
 */
CadModule.prototype._updateCadRendering = function() {
	var docController = this.f.getDocumentController();

	var activePage = docController.getActivePage();
	var differencePage = docController.getDifferencePage();
	var cadOverlay = docController.getCadOverlay();

	// we use the viewport module for adding and removing layers
	this.f.whenModuleAvailable(
		"viewport",
		_.bind(function(viewportModule) {
			viewportModule.whenWidgetModelAvailable(function() {
				var subtractedHighlightColor = this.f.getCurrentUser().getColor("PREFERENCES_HIGHLIGHTCOLORS_DIFF_ADD");
				var addedHighlightColor = this.f.getCurrentUser().getColor("PREFERENCES_HIGHLIGHTCOLORS_DIFF_SUB");

				// reset the origins of all cad files
				_.each([activePage, differencePage, cadOverlay], function(page) {
					if (page && page.hasCad()) {
						page.getCad() && page.getCad().resetOrigin();
					}
				});

				if ((activePage && activePage.hasCad()) || (cadOverlay && cadOverlay.hasCad())) {
					// disable animations
					this.f.emit("viewport.toggleAnimation", "cad", false);

					this._referenceCadModelForWidget = null;

					// set the correct cad data
					if (cadOverlay) {
						this._cadOverlayModel.setCad(cadOverlay.getCad(), false, cadOverlay.getCadOffset(), cadOverlay);
						if (!this._layerModels["CadOverlay"]) {
							this._layerModels["CadOverlay"] = new CadOverlayLayerModel(this.f, "CadOverlay", this._cadOverlayModel, false);
							viewportModule.addLayer("CadOverlay", this._layerModels["CadOverlay"]);
						}
						this._cadModelForWidget = this._cadOverlayModel;
						if (this._layerModels["Cad"]) {
							viewportModule.removeLayer("Cad");
							delete this._layerModels["Cad"];
						}
					} else if (this.compareDiffMode !== CompareConstants.diffMode.DIFFVERSION) {
						if (this._layerModels["CadOverlay"]) {
							viewportModule.removeLayer("CadOverlay");
							delete this._layerModels["CadOverlay"];
						}
						this._cadModel.setCad(activePage.getCad(), true, activePage.getCadOffset(), activePage);
						this._addCadLayer("Cad", this._cadModel, false);
						this._cadModelForWidget = this._cadModel;
					}

					//objects will be used if the active/reference page has CAD
					var activeCad;
					var diffCad;

					if (
						(this.compareDiffMode === CompareConstants.diffMode.DIFFVERSION ||
							this.compareDiffMode === CompareConstants.diffMode.SHOWBOTHSIDEBYSIDE) &&
						!cadOverlay &&
						differencePage &&
						differencePage.hasCad()
					) {
						// we are in CAD COMPARE, show only difference
						activeCad = activePage.getCad();
						diffCad = differencePage.getCad();

						var activePageOffset = activeCad.getOrigin();
						var activeSize = activeCad.getSize();
						var diffSize = diffCad.getSize();
						diffCad.setOrigin(activePageOffset.x, activePageOffset.y - (diffSize.height - activeSize.height));
					}

					// now let's make sure compare is rendered correctly
					if (this.compareDiffMode === CompareConstants.diffMode.DIFFVERSION && !cadOverlay && differencePage && differencePage.hasCad()) {
						if (this._layerModels["Cad"]) {
							viewportModule.removeLayer("Cad", "Main");
							delete this._layerModels["Cad"];
						}

						this._referenceCadModel.resetCadColor();

						this._cadModelForWidget = this._referenceCadModel;
						this._referenceCadModel.setCad(diffCad, true, differencePage.getCadOffset(), differencePage);
						this._addCadLayer("referenceCad", this._referenceCadModel, true);
					} else if (
						this.compareDiffMode === CompareConstants.diffMode.SHOWBOTHSIDEBYSIDE &&
						!cadOverlay &&
						differencePage &&
						differencePage.hasCad()
					) {
						// we are in CAD Compare, side by side mode (no cad overlay is defined)
						var cadOffset = activePage.getCadOffset();
						var referenceCadOffset = differencePage.getCadOffset();

						this._cadModel.resetCadColor();

						// add layer for the reference CAD linework
						this._referenceCadModel.setCad(diffCad, true, referenceCadOffset, differencePage);
						this._addCadLayer("referenceCad", this._referenceCadModel, true);
						this._referenceCadModelForWidget = this._referenceCadModel;

						// add layer for showing a colored version of the active page
						this._currentDiffCadModel.setCadColor(subtractedHighlightColor);
						this._currentDiffCadModel.setCad(activeCad, true, cadOffset, activePage);

						this._addCadLayer("currentDiffCad", this._currentDiffCadModel, false);

						// add a layer for showing a colored version of the difference page
						this._referenceDiffCadModel.setCadColor(addedHighlightColor);
						this._addLayerForReferenceDiffCad(activePage, differencePage);
					} else if (
						(this.compareDiffMode === CompareConstants.diffMode.SHOWBOTHCHANGES ||
							this.compareDiffMode === CompareConstants.diffMode.SHOWBOTHONTOP) &&
						!cadOverlay &&
						differencePage &&
						differencePage.hasCad()
					) {
						// we are in CAD compare mode, difference or show-both-on-top mode (no cad overlay is defined)

						// show the reference cad
						this._addLayerForReferenceDiffCad(activePage, differencePage);

						if (this.compareDiffMode === CompareConstants.diffMode.SHOWBOTHCHANGES) {
							// we need to color the CAD linework
							this._cadModel.setCadColor(subtractedHighlightColor);
							this._referenceDiffCadModel.setCadColor(addedHighlightColor);
						} else {
							this._cadModel.resetCadColor();
							this._referenceDiffCadModel.resetCadColor();
						}

						this._referenceCadModelForWidget = this._referenceDiffCadModel;

						if (this._layerModels["referenceCad"]) {
							viewportModule.removeLayer("referenceCad");
							delete this._layerModels["referenceCad"];
						}
					} else {
						// remove layers from side by side compare or show both on top if needed
						_.each(
							["referenceCad", "currentDiffCad", "referenceDiffCad"],
							function(layerName) {
								if (this._layerModels[layerName]) {
									viewportModule.removeLayer(layerName);
								}
								delete this._layerModels[layerName];
							},
							this
						);

						// reset the CAD linework colors
						this._cadModel.resetCadColor();
						this._referenceDiffCadModel.resetCadColor();
					}

					// show widget
					if (!this._widget) {
						// the widget does not exist yet, publish it
						this._publishWidget(
							CadWidget.widgetName,
							_.bind(function(layoutType) {
								this._widget = new CadWidget(this.f, this._cadModelForWidget, this);

								// check if the widget is still needed
								if (this.getAllActivePagesHavingCad().length > 0) {
									this._widget.setCadModel(this._cadModelForWidget, this._referenceCadModelForWidget);
								} else {
									this._widget.setEnabled(false);
								}

								return this._widget;
							}, this)
						);
					} else {
						// set the correct CAD model for the widget (this could be the difference cad is "show reference mode" is activated
						this._widget.setCadModel(this._cadModelForWidget, this._referenceCadModelForWidget);
						// enable the widget
						this._widget.setEnabled(true);
					}

					this._updateCadGraphicsMode();
				} else {
					// NO CAD is defined

					// enable animations
					this.f.emit("viewport.toggleAnimation", "cad", true);

					// remove all added CAD layers
					_.each(this._layerModels, function(layerModel, layerName) {
						viewportModule.removeLayer(layerName);
					});
					this._layerModels = {};

					// disable widget
					if (this._widget) {
						this._widget.setEnabled(false);
					}

					viewportModule.toggleLayer("Graphics", true);
				}
			}, this);
		}, this)
	);
};

/**
 * Add a layer to the viewport (if the layer is not yet existing)
 * @param {String}   layerName - name of the CAD layer
 * @param {CadModel} cadModel  - the CAD model to be used for the layer
 * @param {Boolean} isForReferencePage - is the layer model for the reference page
 */
CadModule.prototype._addCadLayer = function(layerName, cadModel, isForReferencePage) {
	if (!this._layerModels[layerName]) {
		this._layerModels[layerName] = new CadLayerModel(this.f, layerName, cadModel, isForReferencePage);
		this.f.getModule("viewport").addLayer(layerName, this._layerModels[layerName]);
	}
};

/**
 * Add a layer for the reference cad file to the difference panel
 * @param {AbstractDocumentPage} cadDiff
 * @param {CadFile} currentCad - the active document page's CAD file
 * @private
 */
CadModule.prototype._addLayerForReferenceDiffCad = function(activePage, differencePage) {
	var activeCad = activePage.getCad();
	var diffCad = differencePage.getCad();

	var diffCadOffset = differencePage.getCadOffset();

	var activeCadOffset = activeCad.getOrigin();
	var activeSize = activeCad.getSize();
	var diffSize = diffCad.getSize();
	diffCad.setOrigin(activeCadOffset.x, activeCadOffset.y - (diffSize.height - activeSize.height));

	this._referenceDiffCadModel.setAlpha(0.5);
	this._referenceDiffCadModel.setCad(diffCad, false, diffCadOffset, differencePage);

	this._addCadLayer("referenceDiffCad", this._referenceDiffCadModel, true);
};

/**
 * @override
 */
CadModule.prototype.getState = function() {
	if (this._cadModel !== null && this._cadModel.getPage()) {
		var state = {
			// Cad model
			cadModel: {
				bridgesAndNicks: this._cadModel.getBridgesAndNicksVisibility(),
				cadGraphicsMode: this.getCadGraphicsMode(),
				graphicsMode: this._cadModel.getGraphicsMode(),
				showOneupNames: this._cadModel.isShowOneupNames(),
				cadColor: this._cadModel.getCadColor(),
				//run over all the layers and set value to true if the layer is selected
				selectedLayers: _.map(
					this._cadModel.getAllCadLayers(),
					_.bind(function(layer) {
						//maps layerId on a boolean, is the layer selected or not
						return _.indexOf(this._cadModel.getSelectedLayers(), layer) >= 0;
					}, this)
				),
				selectedClasses: _.map(
					this._cadModel.getAllClasses(),
					_.bind(function(className) {
						//maps ClassId on a boolean, is the class selected or not
						return _.indexOf(this._cadModel.getSelectedClasses(), className) >= 0;
					}, this)
				)
			}
		};
		return state;
	} else {
		return {};
	}
};

/**
 * @override
 * @param {Object} state
 */
CadModule.prototype.setState = function(state) {
	// This module is ready
	if (state !== undefined) {
		// Cad model
		var cadModelState = state.cadModel;

		if (cadModelState !== undefined) {
			this._cadModel.setBridgesAndNicks(cadModelState.bridgesAndNicks);

			this._cadModel.setGraphicsMode(cadModelState.graphicsMode);
			this._cadModel.toggleShowOneupNames(cadModelState.showOneupNames);
			this._cadModel.setCadColor(cadModelState.cadColor);
			if (cadModelState.selectedLayers !== undefined) {
				_.each(
					cadModelState.selectedLayers,
					_.bind(function(isLayerSelected, index) {
						if (!isLayerSelected) {
							this._cadModel.toggleLayer(index);
						}
					}, this)
				);
			}
			if (cadModelState.selectedClasses !== undefined) {
				_.each(
					cadModelState.selectedClasses,
					_.bind(function(isClassSelected, index) {
						if (!isClassSelected) {
							this._cadModel.toggleClass(index);
						}
					}, this)
				);
			}
			this.setCadGraphicsMode(cadModelState.cadGraphicsMode);
		}
		this._whenCompareModeChanged(-1, this.compareDiffMode);
	}
};

/**
 * Get the CadModel of the currently active document page
 * @returns {CadModel}
 */
CadModule.prototype.getCurrentPageCadModel = function() {
	if (
		this._currentDiffCadModel.getVisibleCadLines() !== null &&
		this._currentDiffCadModel.getVisibleCadLines().length > 0 &&
		this.compareDiffMode === CompareConstants.diffMode.SHOWBOTHSIDEBYSIDE
	) {
		return this._currentDiffCadModel;
	} else {
		return this._cadModel;
	}
};

/**
 * Get the CadModel of the reference document page
 * @returns {CadModel}
 */
CadModule.prototype.getReferencePageCadModel = function() {
	if (
		this.compareDiffMode === CompareConstants.diffMode.SHOWBOTHCHANGES ||
		this.compareDiffMode === CompareConstants.diffMode.SHOWBOTHSIDEBYSIDE ||
		this.compareDiffMode === CompareConstants.diffMode.SHOWBOTHONTOP
	) {
		return this._referenceDiffCadModel;
	} else if (this.compareDiffMode === CompareConstants.diffMode.DIFFVERSION) {
		return this._referenceCadModel;
	} else {
		return null;
	}
};

/**
 * Get the cadModel for the CAD overlay
 * @returns {CadModel}
 */
CadModule.prototype.getCadOverlayModel = function() {
	if (this._cadOverlayModel.getVisibleCadLines() !== null && this._cadOverlayModel.getVisibleCadLines().length > 0) {
		return this._cadOverlayModel;
	} else {
		return null;
	}
};

/**
 * Set the CAD background color
 * @param {String} [hexColor] - New background color in hex format (without #). If not set, it will reset to the default.
 */
CadModule.prototype.setCadBackgroundColor = function(hexColor) {
	// set it in all models
	this._cadModel.setBackgroundColor(hexColor);
	this._referenceDiffCadModel.setBackgroundColor(hexColor);
	this._referenceCadModel.setBackgroundColor(hexColor);
	this._currentDiffCadModel.setBackgroundColor(hexColor);
};

/**
 * Get the transformation for the reference CAD file
 * @returns {Transformation}
 */
CadModule.prototype.getReferenceTransform = function() {
	var layerModel = this._layerModels["referenceCad"];
	if (!layerModel) {
		layerModel = this._layerModels["referenceDiffCad"];
	}
	if (layerModel) {
		return layerModel.getTransform().copy();
	}
};

/**
 * Get the transformation for the CAD overlay
 * @returns {Transformation}
 */
CadModule.prototype.getOverlayTransform = function() {
	var layerModel = this._layerModels["CadOverlay"];
	if (layerModel) {
		return layerModel.getTransform().copy();
	}
};

/**
 * Get the currently active CADGraphics mode
 * @returns {CadConstants.CadGraphicsMode}
 */
CadModule.prototype.getCadGraphicsMode = function() {
	return this._cadModel.getCadGraphicsMode();
};

/**
 * Set the cad graphics mode
 * @param mode
 */
CadModule.prototype.setCadGraphicsMode = function(mode) {
	_.each([this._cadModel, this._referenceDiffCadModel, this._referenceCadModel, this._currentDiffCadModel, this._cadOverlayModel], function(
		model
	) {
		model.setCadGraphicsMode(mode);
	});

	this._updateCadGraphicsMode();
};

/**
 * Toggle the correct layers on the viewport (for CADGraphics rendering)
 */
CadModule.prototype._updateCadGraphicsMode = function() {
	var cadGraphicsMode = this.getCadGraphicsMode();

	var showGraphics =
		cadGraphicsMode === CadConstants.CadGraphicsMode.BOTH || cadGraphicsMode === CadConstants.CadGraphicsMode.GRAPHICS_ONLY;
	var showCad = cadGraphicsMode === CadConstants.CadGraphicsMode.BOTH || cadGraphicsMode === CadConstants.CadGraphicsMode.CAD_ONLY;

	var viewportModule = this.f.getModule("viewport");

	viewportModule.toggleLayer("Graphics", showGraphics);
	_.each(this._layerModels, function(model, layerName) {
		viewportModule.toggleLayer(layerName, showCad);
	});
};

export default CadModule;
