import _ from "underscore";
import ShapeUtils from "@esko/shapeutils";
import XMLUtils from "@esko/xmlutils";

/**
 * @exports CadColor
 * @memberof CadModule
 * @class A cad color describes the color definition of a color with a certain number (used in the plotting styles)
 * @constructor
 * @param {String} r - red value
 * @param {String} g - green value
 * @param {String} b - blue value
 * @param {String} number - color number
 */
var CadColor = function(r, g, b, number) {
	this.r = r;
	this.g = g;
	this.b = b;
	this.number = number;
};

/**
 * @exports CadPlottingStyle
 * @memberof CadModule
 * @class A plotting style describes how a cad line should be drawn
 * @constructor
 * @param {String} type
 * @param {String} tool
 * @param {String} style
 * @param {String} width
 * @param {CadColor} color
 * @param {String} name
 */
var CadPlottingStyle = function(type, tool, style, width, color, name) {
	this.type = type;
	this.tool = tool;
	this.style = style;
	this.width = width;
	this.color = color;
	this.name = name;
};

/**
 * @exports CadStyle
 * @memberof CadModule
 * @class
 * A cad style indicates how a cad line should be drawn.
 * It contains the color definitions (per colors, the rgb value), the plotting styles (line width, colorref, style...) and the name of a line type.
 * This can be altered in ArtiosCad to have a different visualisation per line type
 * @constructor
 * @param {XMLObject} stylesxml
 */
var CadStyle = function(stylesxml) {
	this.colors = {};
	this.names = {};
	this.plottingStyles = {};

	/*
     * Colors
     */
	var colordefs = stylesxml.getElementsByTagName("ColorDef")[0].getElementsByTagName("Color");
	for (var i = 0; i < colordefs.length; i++) {
		var colordef = colordefs[i];
		var c = new CadColor(
			XMLUtils.attributeValue(colordef, "red"),
			XMLUtils.attributeValue(colordef, "green"),
			XMLUtils.attributeValue(colordef, "blue"),
			XMLUtils.attributeValue(colordef, "color")
		);
		this.colors[c.number] = c;
	}

	/*
     * Names
     */
	var namedefs = stylesxml.getElementsByTagName("NameDef")[0].getElementsByTagName("Name");
	for (var i = 0; i < namedefs.length; i++) {
		var namedef = namedefs[i];
		var name = XMLUtils.attributeValue(namedef, "name");
		var type = XMLUtils.attributeValue(namedef, "type");
		this.names[type] = name;
	}

	/*
     * Plotting Styles
     */
	var pstyles = stylesxml.getElementsByTagName("PlottingStyle")[0].getElementsByTagName("Style");
	for (var i = 0; i < pstyles.length; i++) {
		var pstyle = pstyles[i];
		var styletype = XMLUtils.attributeValue(pstyle, "type");
		if (styletype == "*") {
			continue;
		}
		var stylecolor = XMLUtils.attributeValue(pstyle, "color");
		var styletool = XMLUtils.attributeValue(pstyle, "tool");
		var stylestyle = XMLUtils.attributeValue(pstyle, "style");
		var stylewidth = XMLUtils.attributeValue(pstyle, "width");

		var stylecolorobject = this.colors[stylecolor];
		var stylename = this.names[styletype];
		var plotstyle = new CadPlottingStyle(styletype, styletool, stylestyle, stylewidth, stylecolorobject, stylename);

		this.plottingStyles[styletype] = plotstyle;
	}
};

export default CadStyle;
