import AbstractWidget from "@esko/web-viewer/layout/AbstractWidget";
import tpl from "@esko/web-viewer/cad/CadWidget.tpl.html";
import CadConstants from "@esko/web-viewer/cad/CadConstants";
import Template from "@esko/template";
import ShapeUtils from "@esko/shapeutils";
import MiniButton from "@esko/web-viewer/layout/component/MiniButton";
import DropdownMiniButton from "@esko/web-viewer/layout/component/DropdownMiniButton";
import ViewerToolbarButton from "@esko/web-viewer/layout/component/ViewerToolbarButton";
import ButtonGroup from "@esko/toolbar/ButtonGroup";
import LayoutConstants from "@esko/web-viewer/layout/LayoutConstants";
import $ from "jquery";
import _ from "underscore";
/**
 * @exports CadWidget
 * @memberof CadModule
 * @augments Layout.AbstractWidget
 * @constructor
 * @param {Core.ViewerFacade} f
 * @param {CadModel} cadModel
 * @param {CadModule} cadModule
 */
var CadWidget = function(f, cadModel, cadModule) {
	AbstractWidget.call(this, f, LayoutConstants.Widget.CAD); // with this call all the methods of AbstractWidget.js are installed to this object

	/**
	 * Facade
	 * @private
	 */
	this._f = f;

	/**
	 * {CadModule}
	 * @private
	 */
	this._cadModule = cadModule;
	/**
	 * {CadModel}
	 * @private
	 */
	this._cadModel = cadModel;

	/**
	 * Div that holds the content of the template
	 * @private
	 */
	this._contentDiv = $("<div/>");

	this.setIcon("icon icon-WCR_CAD-Both");

	// initialize the layer template
	this.layerTemplate = new Template(f.i18n, tpl);
	this.layerTemplate.setContext(this._contentDiv);

	//Layers
	this.layerTemplate.d.layers = this._cadModel.getAllCadLayers();
	this.layerTemplate.d.selectedLayers = this._cadModel.getSelectedLayers();

	//Bind the cad layer selection to this event
	this.layerTemplate.bind(
		"mouseup",
		".cadLayer",
		function(source, event) {
			var id = $(source).attr("data-bind");
			if (id == "all") {
				id = -1;
			}

			var activeCadLayerCount = this._cadModel.getAllCadLayers().length;
			if (id >= activeCadLayerCount && this._referenceCadModel) {
				id = id - activeCadLayerCount;
				if (event.which === 3) {
					this._referenceCadModel.hideLayers();
					this._referenceCadModel.toggleLayer(id);
				} else {
					this._referenceCadModel.toggleLayer(id);
				}
			} else if (id === -1) {
				var hideAllLayers = this._areAllLayersSelected();
				this._cadModel.toggleLayer(id, hideAllLayers);
				if (this._referenceCadModel) {
					this._referenceCadModel.toggleLayer(id, hideAllLayers);
				}
			} else {
				if (event.which === 3) {
					//right click
					this._cadModel.hideLayers();
					this._cadModel.toggleLayer(id);
				} else {
					this._cadModel.toggleLayer(id);
				}
			}
		},
		this
	);

	this.layerTemplate.bind(
		"contextmenu",
		".cadLayer,.cadClass",
		function(el, e) {
			e.preventDefault();
			return false;
		},
		this
	);

	//Function that indicates if all the cad layers are selected or not
	this.layerTemplate.d.areAllLayersSelected = _.bind(this._areAllLayersSelected, this);

	//Function that indicates if all the cad layers are selected or not
	this.layerTemplate.d.isLayerSelected = _.bind(function(layer) {
		return (
			_.indexOf(this._cadModel.getSelectedLayers(), layer) >= 0 ||
			(this._referenceCadModel && _.indexOf(this._referenceCadModel.getSelectedLayers(), layer) >= 0)
		);
	}, this);

	//Classes
	this.layerTemplate.d.classes = this._cadModel.getAllClasses();
	this.layerTemplate.d.selectedClasses = this._cadModel.getSelectedClasses();
	this.layerTemplate.d.showClasses = this._cadModel.showClasses();

	//Function that indicates if all the cad layers are selected or not
	this.layerTemplate.d.areAllClassesSelected = _.bind(this._areAllClassesSelected, this);

	//Function that indicates if all the cad layers are selected or not
	this.layerTemplate.d.isClassSelected = _.bind(function(className, isCurrentCAD) {
		if (isCurrentCAD) {
			return _.indexOf(this._cadModel.getSelectedClasses(), className) >= 0;
		} else {
			return this._referenceCadModel && _.indexOf(this._referenceCadModel.getSelectedClasses(), className) >= 0;
		}
	}, this);

	//Get the localised name of the layer
	this.layerTemplate.d.getLayerLocName = _.bind(function(layer) {
		var name;
		if (layer.layername) {
			name = layer.layername;
		} else {
			name = this._f.i18n.t(CadConstants.CadLayerLocName[layer.layertype]);
			if (name === undefined) {
				name = layer.layertype;
			}
		}
		if (layer.layerclass !== undefined && layer.layerclass != "0") {
			var className = layer.layerclass;
			if (CadConstants.CadClassName[layer.layerclass] !== undefined) {
				className = this._f.i18n.t(CadConstants.CadClassName[layer.layerclass]);
			}
			if (className.toUpperCase() != name.toUpperCase()) {
				if (!name || name.length === 0) {
					name = className;
				} else {
					name += " (" + className + ")";
				}
			}
		}
		return name;
	}, this);

	//Get the localised name of the class
	this.layerTemplate.d.getClassLocName = _.bind(function(classNumber) {
		return this._f.i18n.t(CadConstants.CadClassName[classNumber]);
	}, this);

	//Bind the cad class selection to this event
	this.layerTemplate.bind(
		"mouseup",
		".cadClass",
		function(source, event) {
			var id = $(source).attr("data-bind");
			if (id == "all") {
				id = -1;
			}

			var activeCadClassesCount = this._cadModel.getCad().getOneupClasses().length;
			if (id >= activeCadClassesCount && this._referenceCadModel) {
				id = id - activeCadClassesCount;
				if (event.which === 3) {
					//right click
					this._referenceCadModel.hideClasses();
					this._referenceCadModel.toggleClass(id);
				} else {
					this._referenceCadModel.toggleClass(id);
				}
			} else if (id === -1) {
				var hideAllClasses = this._areAllClassesSelected();
				if (this._referenceCadModel) {
					this._referenceCadModel.toggleClass(id, hideAllClasses);
				}
				this._cadModel.toggleClass(id, hideAllClasses);
			} else {
				if (event.which === 3) {
					//right click
					this._cadModel.hideClasses();
					this._cadModel.toggleClass(id);
				} else {
					this._cadModel.toggleClass(id);
				}
			}
		},
		this
	);

	this._boundUpdateUI = _.bind(function() {
		this._updateUI();
	}, this);
	/*
     * When the model changes, update the toggles and the selected layers
     */
	this._cadModel.on("modelChanged", this._boundUpdateUI);

	/**
	 * Show both button
	 * @private
	 */
	this._showBothButton = new ViewerToolbarButton(
		f,
		"showBoth",
		"icon-WCR_CAD-Both",
		true,
		_.bind(function() {
			this._cadModule.setCadGraphicsMode(CadConstants.CadGraphicsMode.BOTH);
		}, this)
	);
	this._showBothButton.setTooltip("viewer", "CadViewBothTT");

	/**
	 * Show only graphics
	 * @private
	 */
	this._showOnlyGraphicsButton = new ViewerToolbarButton(
		f,
		"showOnlyGraphics",
		"icon-WCR_CAD-OnlyGraphics",
		true,
		_.bind(function() {
			this._cadModule.setCadGraphicsMode(CadConstants.CadGraphicsMode.GRAPHICS_ONLY);
		}, this)
	);
	this._showOnlyGraphicsButton.setTooltip("viewer", "CadViewGraphicsTT");

	/**
	 * Show only cad
	 * @private
	 */
	this._showOnlyCadButton = new ViewerToolbarButton(
		f,
		"showOnlyCad",
		"icon-WCR_CAD-OnlyCAD",
		true,
		_.bind(function() {
			this._cadModule.setCadGraphicsMode(CadConstants.CadGraphicsMode.CAD_ONLY);
		}, this)
	);
	this._showOnlyCadButton.setTooltip("viewer", "CadViewCadTT");

	/**
	 * Show 3D widget with a generated 3D model of the active document page
	 * @private
	 */
	this._3DView = new MiniButton(
		this.f,
		"show3DView",
		"icon-WCR_3D-Create",
		true,
		_.bind(function() {
			//first check if the compareModule exists, if that's the case let it check if the aligment is correctly saved
			var compareModule = this.f.getModule("compare");
			if (compareModule) {
				//pass opening of 3D through the compare module
				compareModule.checkAligmentAndShow3D();
			} else {
				//no compare module (only CAD is open) load the 3D View
				this.f.requestModule(
					"view3D",
					_.bind(function(view3DModule) {
						view3DModule.show3D();
					}, this),
					false
				);
			}
		}, this),
		"fontIconDisabled icon-WCR_3D-Create"
	);

	/**
	 * Button group to switch cadGraphicsMode
	 * @private
	 */
	this._showButtonGroup = new ButtonGroup("buttonGroup", [this._showBothButton, this._showOnlyGraphicsButton, this._showOnlyCadButton]);
	this._showButtonGroup.element.addClass("cadShowOptions toggleBtnGroup");

	/**
	 * Make the options dropdown
	 * @private
	 */
	this._optionButton = new DropdownMiniButton(f, "cadOptions", "icon-WCR_Settings-submenu");

	f.i18n.loadBundle(
		"viewer",
		_.bind(function() {
			this._optionButton.addOption({
				name: "CadBridgesShowNone",
				content: f.i18n.t("CadBridgesShowNone"),
				onchange: _.bind(this._showNone, this),
				userCannotUncheck: true
			});

			this._optionButton.addOption({
				name: "CadBridgesShowNicks",
				content: f.i18n.t("CadBridgesShowNicks"),
				onchange: _.bind(this._showNicks, this)
			});

			this._optionButton.addOption({
				name: "CadBridgesShowBridges",
				content: f.i18n.t("CadBridgesShowBridges"),
				onchange: _.bind(this._showBridges, this)
			});

			this._optionButton.addOption({
				name: "CadBridgesShowBoth",
				content: f.i18n.t("CadBridgesShowBoth"),
				onchange: _.bind(this._showBoth, this)
			});

			this._optionButton.toggleOption("CadBridgesShowNone", true);

			this._optionButton.addDivider();

			this._optionButton.addOption({
				name: "CadOneupTT",
				content: f.i18n.t("CadOneupTT"),
				onchange: _.bind(this._toggleShowOneupNames, this)
			});

			this._optionButton.addDivider();

			this._optionButton.addOption({
				name: "CadNormal",
				content: f.i18n.t("CadNormal"),
				onchange: _.bind(this._setGraphicsNormal, this)
			});

			this._optionButton.addOption({
				name: "CadLighten",
				content: f.i18n.t("CadLighten"),
				onchange: _.bind(this._setGraphicsLighten, this)
			});
			this._optionButton.addOption({
				name: "CadLightenMore",
				content: f.i18n.t("CadLightenMore"),
				onchange: _.bind(this._setGraphicsLightenMore, this)
			});

			this.setTitle(this.f.i18n.t("CadTitle"));

			this._updateUI();
		}, this)
	);

	/*
     * Overwrite the setContext to attach the header and the template
     */
	this._override("setContext", function(context, headerExtra) {
		this._super(context);

		headerExtra.append(this._3DView.element);
		this._3DView.setTooltip("viewer", "Open3DViewTT");

		//Append header
		headerExtra.append(this._optionButton.element);
		headerExtra.addClass("flexEnd");

		//Append content (buttons + content)
		context.append(this._showButtonGroup.element);
		context.append(this._contentDiv);
		context.addClass("cadWidget");

		//Load translations if needed
		f.i18n.loadBundle(
			"viewer",
			_.bind(function() {
				this._updateUI();
			}, this)
		);
	});
};

CadWidget.prototype = Object.create(AbstractWidget.prototype);
CadWidget.prototype.constructor = CadWidget;
CadWidget.widgetName = LayoutConstants.Widget.CAD;

/**
 * @override
 */
CadWidget.prototype.getMinimumSize = function() {
	return new ShapeUtils.Size(250, 150);
};

/**
 * @override
 */
CadWidget.prototype.willFill = function() {
	return false;
};

/**
 * Are all the layers currently selected (visible)?
 * @returns {Boolean}
 * @private
 */
CadWidget.prototype._areAllLayersSelected = function() {
	return (
		this._cadModel.getAllCadLayers().length == this._cadModel.getSelectedLayers().length &&
		(!this._referenceCadModel || this._referenceCadModel.getAllCadLayers().length == this._referenceCadModel.getSelectedLayers().length)
	);
};

/**
 * Are all classes currently selected (visible)?
 * @returns {Boolean}
 */
CadWidget.prototype._areAllClassesSelected = function() {
	return (
		this._cadModel.getCad().getOneupClasses().length == this._cadModel.getSelectedClasses().length &&
		(!this._referenceCadModel ||
			this._referenceCadModel.getCad().getOneupClasses().length == this._referenceCadModel.getSelectedClasses().length)
	);
};

/**
 * Change the cadModel of the widget
 * @param {CadModel} cadModel
 * @param {CadModel} referenceCadModel
 */
CadWidget.prototype.setCadModel = function(cadModel, referenceCadModel) {
	if (this._cadModel !== cadModel) {
		this._cadModel.off("modelChanged", this._boundUpdateUI);
		this._cadModel = cadModel;
		this._cadModel.on("modelChanged", this._boundUpdateUI);
	}
	if (this._referenceCadModel !== referenceCadModel) {
		if (this._referenceCadModel) {
			this._referenceCadModel.off("modelChanged", this._boundUpdateUI);
		}
		if (referenceCadModel) {
			this._referenceCadModel = referenceCadModel;
			this._referenceCadModel.on("modelChanged", this._boundUpdateUI);
			this._referenceCadModel.setBridgesAndNicks(this._cadModel.getBridgesAndNicksVisibility());
		} else {
			this._referenceCadModel = null;
		}
	}
	this._updateUI();
};

/**
 * Update the UI components
 * @private
 */
CadWidget.prototype._updateUI = function() {
	var activePage = this.f.getActivePage();
	var overlay = this.f.getDocumentController().getCadOverlay();
	if (this._3DView && this._3DView.element) {
		//check if the active page either has CAD or there is CAD overlay (linked or docref)
		if (
			!this.f.isMobile() &&
			activePage.get("canGenerate3D") &&
			(activePage.hasCad() || (activePage.hasGraphics() && overlay && (activePage.hasLinkedPage() || activePage.canSaveCadOverlay()))) &&
			activePage.canEditAttributes()
		) {
			this._3DView.element.removeClass("displayNone");
		} else {
			this._3DView.element.addClass("displayNone");
		}
	}

	// Nicks/Bridges visibility
	this._optionButton.toggleOption(
		"CadBridgesShowNone",
		this._cadModel.getBridgesAndNicksVisibility() === CadConstants.ShowBridgesNicks.NONE
	);
	this._optionButton.toggleOption(
		"CadBridgesShowBridges",
		this._cadModel.getBridgesAndNicksVisibility() === CadConstants.ShowBridgesNicks.BRIDGES
	);
	this._optionButton.toggleOption(
		"CadBridgesShowNicks",
		this._cadModel.getBridgesAndNicksVisibility() === CadConstants.ShowBridgesNicks.NICKS
	);
	this._optionButton.toggleOption(
		"CadBridgesShowBoth",
		this._cadModel.getBridgesAndNicksVisibility() === CadConstants.ShowBridgesNicks.BOTH
	);

	// Show one up names
	this._optionButton.setActionEnabled(
		"CadOneupTT",
		this._cadModel.canShowOneupNames() || (this._referenceCadModel && this._referenceCadModel.canShowOneupNames())
	);
	this._optionButton.toggleOption(
		"CadOneupTT",
		this._cadModel.isShowOneupNames() ||
			(this._referenceCadModel !== null && this._referenceCadModel !== undefined && this._referenceCadModel.isShowOneupNames())
	);

	// Graphics mode
	this._optionButton.toggleOption("CadNormal", this._cadModel.getGraphicsMode() === 1);
	this._optionButton.toggleOption("CadLighten", this._cadModel.getGraphicsMode() === 2);
	this._optionButton.toggleOption("CadLightenMore", this._cadModel.getGraphicsMode() === 3);

	var showOptions = !this._cadModel.isOnlyCad();
	this._optionButton.setActionEnabled("CadNormal", showOptions);
	this._optionButton.setActionEnabled("CadLighten", showOptions);
	this._optionButton.setActionEnabled("CadLightenMore", showOptions);

	//Set the tooltip of the "both" button with the mode of the graphics
	var mode = "CadNormal";
	if (this._cadModel.getGraphicsMode() === 2) {
		mode = "CadLighten";
	} else if (this._cadModel.getGraphicsMode() === 3) {
		mode = "CadLightenMore";
	}
	this._showBothButton.setTooltip("viewer", "CadViewBothTT", [this._f.i18n.t(mode)]);

	// Update template
	this.layerTemplate.d.classes = this._cadModel.getCad().getOneupClasses();
	this.layerTemplate.d.showClasses = this._cadModel.showClasses();
	this.layerTemplate.d.layers = this._cadModel.getAllCadLayers();

	if (this._referenceCadModel) {
		this.layerTemplate.d.referenceClasses = this._referenceCadModel.getCad().getOneupClasses();
		this.layerTemplate.d.showClasses = this._cadModel.showClasses() || this._referenceCadModel.showClasses();
		this.layerTemplate.d.layers = this.layerTemplate.d.layers.concat(this._referenceCadModel.getAllCadLayers());
	}

	this.layerTemplate.render();

	// Show graphics
	this._showBothButton.setActive(this._cadModel.getCadGraphicsMode() === CadConstants.CadGraphicsMode.BOTH);
	this._showOnlyGraphicsButton.setActive(this._cadModel.getCadGraphicsMode() === CadConstants.CadGraphicsMode.GRAPHICS_ONLY);
	this._showOnlyCadButton.setActive(this._cadModel.getCadGraphicsMode() === CadConstants.CadGraphicsMode.CAD_ONLY);

	//Hide cad/graphics selection if there is only cad
	if (this._cadModel.isOnlyCad()) {
		this._showButtonGroup.hide();
	} else {
		this._showButtonGroup.show();
	}

	this._fireResized();
};

/**
 * Show bridges
 * @private
 */
CadWidget.prototype._showBridges = function(toggle) {
	if (toggle) {
		this._cadModel.setBridgesAndNicks(CadConstants.ShowBridgesNicks.BRIDGES);
		if (this._referenceCadModel) {
			this._referenceCadModel.setBridgesAndNicks(CadConstants.ShowBridgesNicks.BRIDGES);
		}
	} else {
		this._showNone(true);
	}
};

/**
 * Show nicks
 * @private
 */
CadWidget.prototype._showNicks = function(toggle) {
	if (toggle) {
		this._cadModel.setBridgesAndNicks(CadConstants.ShowBridgesNicks.NICKS);
		if (this._referenceCadModel) {
			this._referenceCadModel.setBridgesAndNicks(CadConstants.ShowBridgesNicks.NICKS);
		}
	} else {
		this._showNone(true);
	}
};

/**
 * Show both bridges and nicks
 * @private
 */
CadWidget.prototype._showBoth = function(toggle) {
	if (toggle) {
		this._cadModel.setBridgesAndNicks(CadConstants.ShowBridgesNicks.BOTH);
		if (this._referenceCadModel) {
			this._referenceCadModel.setBridgesAndNicks(CadConstants.ShowBridgesNicks.BOTH);
		}
	} else {
		this._showNone(true);
	}
};

/**
 * Show no bridges or nicks
 * @private
 */
CadWidget.prototype._showNone = function(toggle) {
	if (toggle) {
		this._cadModel.setBridgesAndNicks(CadConstants.ShowBridgesNicks.NONE);
		if (this._referenceCadModel) {
			this._referenceCadModel.setBridgesAndNicks(CadConstants.ShowBridgesNicks.NONE);
		}
	}
};

/**
 * Show/hide oneup names
 * @private
 */
CadWidget.prototype._toggleShowOneupNames = function() {
	if (this._cadModel.canShowOneupNames()) {
		this._cadModel.toggleShowOneupNames();
	}
	if (this._referenceCadModel && this._referenceCadModel.canShowOneupNames()) {
		this._referenceCadModel.toggleShowOneupNames();
	}
};

/**
 * Set grapics level to normal
 * @private
 */
CadWidget.prototype._setGraphicsNormal = function() {
	this._cadModel.setGraphicsMode(1);
};

/**
 * Set grapics level to lighten
 * @private
 */
CadWidget.prototype._setGraphicsLighten = function() {
	this._cadModel.setGraphicsMode(2);
};

/**
 * Set grapics level to lighten more
 * @private
 */
CadWidget.prototype._setGraphicsLightenMore = function() {
	this._cadModel.setGraphicsMode(3);
};

export default CadWidget;
