import _ from "underscore";
import CompareMode from "@esko/web-viewer/compare/CompareMode";
import CompareConst from "@esko/web-viewer/compare/CompareConstants";
import ViewerUtils from "@esko/web-viewer/utils/ViewerUtils";
/**
 * @memberof CompareModule
 * @exports CompareDifferenceMode
 * @class Compare two documents in difference compare (two colors for added and removed parts)
 * @constructor
 * @param {Core.ViewerFacade} f
 * @param {Compare.CompareModel} compareModel - model for compare
 */
var CompareDifferenceMode = function(f, compareModel) {
	CompareMode.call(this, f, CompareConst.mode.DIFFERENCE, compareModel, true, "DiffDifferenceAnnotationsHidden");

	this.threshold = 0;
	this.setThreshold(parseInt(this.f.getCurrentUser().getOption("PREFERENCES_DIFFMODE_THRESHOLD"), 10));

	this.f.getCurrentUser().on(
		"options.changed.PREFERENCES_DIFFMODE_THRESHOLD",
		_.bind(function(oldValue, newValue) {
			// not newValue for translations to hex
			this.setThreshold(newValue);
		}, this)
	);

	//the setting is the same as in the javaviewer:
	// 0=DEFAULT, 1=CHANNELS, 2=HSV, 3=AUTOMATIC
	this.colorSpace = this.f.getOption("CONFIG_DIFF_COMPAREMODE");

	this.differenceMode = compareModel.getDiffMode();

	if (this.differenceMode < CompareConst.diffMode.SHOWBOTHCHANGES || this.differenceMode > CompareConst.diffMode.SHOWBOTHCHANGESONLY) {
		var diffMode = CompareConst.sideBySideToDiffMode[this.differenceMode];
		if (diffMode === undefined) {
			this.differenceMode = CompareConst.diffMode.SHOWBOTHCHANGES;
		} else {
			this.differenceMode = diffMode;
		}
	}

	// hex value with #
	this.addedHighlightColor = this.f.getCurrentUser().getColor("PREFERENCES_HIGHLIGHTCOLORS_DIFF_ADD");
	this.f.getCurrentUser().on(
		"options.changed.PREFERENCES_HIGHLIGHTCOLORS_DIFF_ADD",
		_.bind(function(oldValue, newValue) {
			// not newValue for translations to hex
			this.setAddedHighlightColor(this.f.getCurrentUser().getColor("PREFERENCES_HIGHLIGHTCOLORS_DIFF_ADD"));
		}, this)
	);

	// hex value with #
	this.subtractedHighlightColor = this.f.getCurrentUser().getColor("PREFERENCES_HIGHLIGHTCOLORS_DIFF_SUB");
	this.f.getCurrentUser().on(
		"options.changed.PREFERENCES_HIGHLIGHTCOLORS_DIFF_SUB",
		_.bind(function(oldValue, newValue) {
			// not newValue for translations to hex
			this.setSubtractedHighlightColor(this.f.getCurrentUser().getColor("PREFERENCES_HIGHLIGHTCOLORS_DIFF_SUB"));
		}, this)
	);

	this.on("changed.**", function() {
		compareModel.parametersChanged();
	});
};

CompareDifferenceMode.prototype = Object.create(CompareMode.prototype);
CompareDifferenceMode.prototype.constructor = CompareDifferenceMode;

/**
 * Start/stop show compare mode
 */
CompareDifferenceMode.prototype.toggle = function(toggle, diffMode) {
	CompareMode.prototype.toggle.call(this, toggle);
	if (toggle) {
		if (
			diffMode != undefined &&
			(diffMode === CompareConst.diffMode.SHOWBOTHCHANGES ||
				diffMode === CompareConst.diffMode.SHOWBOTHCHANGESHIGHLIGHT ||
				diffMode === CompareConst.diffMode.SHOWBOTHCHANGESONLY ||
				diffMode === CompareConst.diffMode.SHOWBOTHONTOP)
		) {
			// we only allow diffModes of this mode
			this.setDifferenceMode(diffMode);
		} else {
			this._compareModel.setDiffMode(this.differenceMode);
		}
	} else {
		this._compareModel.setDiffMode(CompareConst.diffMode.UNDEFINED);
	}
};

/**
 * get the DType parameter for the viewserver
 */
CompareDifferenceMode.prototype._getDType = function() {
	// the DType parameter is formed with diffHsvMode, diffEnhanche and diffChannelsMode (but we use default value for those: 0)
	// original formula: diffHsvMode*256 + diffEnhance*64 + colorSpace*8 + diffChannelsMode
	// we only use colorspace setting

	return this.colorSpace * 8;
};

/**
 * Intercept fragment url generation
 * @param {Object} context
 */
CompareDifferenceMode.prototype.interceptFragmentURL = function(context) {
	throw "Not Implemented";
};

/**
 * Get the overwrite options for fragmentURL generation
 * @returns {Object} object with the parameters which should be set
 */
CompareDifferenceMode.prototype.getOverwriteOptions = function() {
	throw "Not Implemented";
};
/**
 * Set a new threshold for difference compare
 * @param {Number} newThreshold - number between 0 and 100
 */
CompareDifferenceMode.prototype.setThreshold = function(newThreshold) {
	var oldThreshold = this.threshold;

	this.threshold = newThreshold;
	if (isNaN(this.threshold)) {
		this.threshold = 0;
	}
	this.threshold = Math.min(Math.max(this.threshold, 0), 100);

	this.f.setOption("PREFERENCES_DIFFMODE_THRESHOLD", this.threshold);
	this._fireChange(false, ["threshold"], oldThreshold, newThreshold);
};

/**
 * Get the currently used threshold
 * @returns {Number} threshold
 */
CompareDifferenceMode.prototype.getThreshold = function() {
	return this.threshold;
};

/**
 * Set the colorspace
 * @param {CompareModule.CompareConstants.colorSpace} newColorspace
 */
CompareDifferenceMode.prototype.setColorSpace = function(newColorspace) {
	var oldColorspace = this.colorSpace;
	this.colorSpace = newColorspace;
	this._fireChange(false, ["colorspace"], oldColorspace, newColorspace);
	//the setting is set the same way as in the java viewer, so making sure the setting is the same
	// 0=default, 1=channels, 2=hsv, 3=automatic (using the 4th and 5th bit)
	this.f.setOption("CONFIG_DIFF_COMPAREMODE", newColorspace);
};

/**
 * Get the currently used colorspace
 * @param {CompareModule.CompareConstants.colorSpace} colorspace
 */
CompareDifferenceMode.prototype.getColorSpace = function() {
	return this.colorSpace;
};

/**
 * Set the difference mode
 * @param {CompareModule.CompareConstants.showDifferenceMode} newDifferenceMode
 */
CompareDifferenceMode.prototype.setDifferenceMode = function(newDifferenceMode) {
	var oldDifferenceMode = this.differenceMode;
	this.differenceMode = newDifferenceMode;

	if (this.isEnabled()) {
		// only update the compare mode if active...
		this._compareModel.setDiffMode(this.differenceMode);
	}
	this._fireChange(false, ["differenceMode"], oldDifferenceMode, newDifferenceMode);
};

/**
 * Get the difference mode
 * @returns {CompareModule.CompareConstants.showDifferenceMode}
 */
CompareDifferenceMode.prototype.getDifferenceMode = function() {
	return this.differenceMode;
};

/**
 * Set the added highlight color
 * @param {String} addedHighlightColor - hex value with #
 */
CompareDifferenceMode.prototype.setAddedHighlightColor = function(addedHighlightColor) {
	var oldAddedHighlightColor = this.addedHighlightColor;
	this.f.getCurrentUser().setColor("PREFERENCES_HIGHLIGHTCOLORS_DIFF_ADD", addedHighlightColor);
	this.addedHighlightColor = this.f.getCurrentUser().getColor("PREFERENCES_HIGHLIGHTCOLORS_DIFF_ADD");
	this._fireChange(false, ["addedHighlightColor"], oldAddedHighlightColor, this.addedHighlightColor);
};

/**
 * Get the added highlight color
 * @returns {String} hex value with #
 */
CompareDifferenceMode.prototype.getAddedHighlightColor = function() {
	return this.addedHighlightColor;
};

/**
 * Set the subtracted highlight color
 * @param {String} subtractedHighlightColor - hex value with #
 */
CompareDifferenceMode.prototype.setSubtractedHighlightColor = function(subtractedHighlightColor) {
	var oldSubtractedHighlightColor = this.subtractedHighlightColor;
	this.f.getCurrentUser().setColor("PREFERENCES_HIGHLIGHTCOLORS_DIFF_SUB", subtractedHighlightColor);
	this.subtractedHighlightColor = this.f.getCurrentUser().getColor("PREFERENCES_HIGHLIGHTCOLORS_DIFF_SUB");
	this._fireChange(false, ["subtractedHighlightColor"], oldSubtractedHighlightColor, this.subtractedHighlightColor);
};

/**
 * Get the subtracted highlight color
 * @returns {String} hex value with #
 */
CompareDifferenceMode.prototype.getSubtractedHighlightColor = function() {
	return this.subtractedHighlightColor;
};

/**
 * Reset highlight colors
 */
CompareDifferenceMode.prototype.resetHighlightColors = function() {
	this.f.getCurrentUser().resetOption("PREFERENCES_HIGHLIGHTCOLORS_DIFF_ADD");
	this.f.getCurrentUser().resetOption("PREFERENCES_HIGHLIGHTCOLORS_DIFF_SUB");
	// it is automatically fired back to setAddedHighlightColor and setSubtractedHighlightColor
};

CompareDifferenceMode.prototype.isAvailable = function() {
	var activePage = this.f.getDocumentController().getActivePage();
	var difPage = this.f.getDocumentController().getDifferencePage();
	return _.reduce(
		[activePage, difPage],
		function(memo, page) {
			return memo && (!page || page.getAvailableCompareModes().indexOf(CompareConst.mode.DIFFERENCE) !== -1);
		},
		true
	);
};

export default CompareDifferenceMode;
