import ToolbarComponent from "@esko/toolbar/ToolbarComponent";
import SubtoolbarButton from "@esko/web-viewer/layout/component/SubtoolbarButton";
import DropdownSubtoolbarButton from "@esko/web-viewer/layout/component/DropdownSubtoolbarButton";
import ButtonGroup from "@esko/toolbar/ButtonGroup";
import Slider from "@esko/web-viewer/layout/component/Slider";
import ColorPicker from "@esko/web-viewer/layout/component/ColorPicker";
import DoubleColorBlock from "@esko/web-viewer/layout/component/DoubleColorBlock";
import CompareConstants from "@esko/web-viewer/compare/CompareConstants";
import ListenerUtil from "@esko/web-viewer/utils/ListenerUtil";
import _ from "underscore";
/**
 * @memberof CompareModule
 * @exports CompareDifferenceSubToolbarComponent
 * @class subtoolbarcomponent used for interaction with the difference compare mode
 * @constructor
 * @param {Core.ViewerFacade} f
 * @param {CompareModule.CompareDifferenceMode} compareDiffMode - the mode that should be controlled
 * @param {Layout.ViewerToolbarButton} parent - model for compareparent button
 */
var CompareDifferenceSubToolbarComponent = function(f, compareDiffMode, parent) {
	ListenerUtil.mixin(this, f);
	ToolbarComponent.call(this, "DifferenceCompareOptions", f);
	this.f = f;
	this.compareDiffMode = compareDiffMode;

	this.parent = parent;

	this._initColorSpace();
	this._initThresholdSlider();
	this._initHighlightColor();
	this._initDiffmode();

	this.f.getDocumentController().whenActivePageChanged(
		_.bind(function(oldPage, newPage) {
			this._whenActivePageChanged(oldPage, newPage);
		}, this)
	);
};

CompareDifferenceSubToolbarComponent.prototype = Object.create(ToolbarComponent.prototype);
CompareDifferenceSubToolbarComponent.prototype.constructor = CompareDifferenceSubToolbarComponent;

/**
 * The active page has changed
 * @param {Data.AbstractDocumentPage} oldPage
 * @param {Data.AbstractDocumentPage} newPage
 */
CompareDifferenceSubToolbarComponent.prototype._whenActivePageChanged = function(oldPage, newPage) {
	if (newPage.hasCad()) {
		this.colorspaceGroup.hide();
		this.differenceShowModeGroup.hide();
		this.slider.element.addClass("displayNone");
	} else {
		this.colorspaceGroup.show();
		this.differenceShowModeGroup.show();
		this.slider.element.removeClass("displayNone");
	}
};

/**
 * Initialize the colorspace
 */
CompareDifferenceSubToolbarComponent.prototype._initColorSpace = function() {
	this.colorSpaceButtons = [];
	var isMobile = this.f.isMobile();
	// colorspace 1: channels
	this.colorSpaceButtons[0] = new SubtoolbarButton(
		this.f,
		"colorspaceChannels",
		"icon-WCR_Compare-Channels",
		true,
		_.bind(this.compareDiffMode.setColorSpace, this.compareDiffMode, CompareConstants.colorSpace.CHANNELS)
	);

	// colorspace 2: HSV
	this.colorSpaceButtons[1] = new SubtoolbarButton(
		this.f,
		"colorspaceHSV",
		"icon-WCR_Compare-HSV",
		true,
		_.bind(this.compareDiffMode.setColorSpace, this.compareDiffMode, CompareConstants.colorSpace.HSV)
	);

	// colorspace 3: automatic
	this.colorSpaceButtons[2] = new SubtoolbarButton(
		this.f,
		"colorspaceAutomatic",
		"icon-WCR_Compare-Automatic",
		true,
		_.bind(this.compareDiffMode.setColorSpace, this.compareDiffMode, CompareConstants.colorSpace.AUTOMATIC)
	);

	// tool tips
	this.f.i18n.loadBundle(
		"viewer",
		function() {
			var prefix = this.f.i18n.t("DiffColorSpace") + " ";
			this.colorSpaceButtons[0].setTooltip(prefix + this.f.i18n.t("DiffColorSpaceChannels"));
			this.colorSpaceButtons[1].setTooltip(prefix + this.f.i18n.t("DiffColorSpaceHsv"));
			this.colorSpaceButtons[2].setTooltip(prefix + this.f.i18n.t("DiffColorSpaceAutomatic"));
		},
		this
	);

	this.colorspaceGroup = new ButtonGroup("colorspace", this.colorSpaceButtons);
	this.colorspaceGroup.element.addClass("toggleBtnGroup");
	this.addButtonGroup(this.colorspaceGroup);

	this.compareDiffMode.on("changed.colorspace", _.bind(this.updateColorSpaceButtons, this));
	this.updateColorSpaceButtons(0, this.compareDiffMode.getColorSpace());
};

/**
 * Update the color space buttons
 * @param oldColorSpace
 * @param newColorSpace
 */
CompareDifferenceSubToolbarComponent.prototype.updateColorSpaceButtons = function(oldColorSpace, newColorSpace) {
	_.each(
		this.colorSpaceButtons,
		function(button, index) {
			// the colorspace options are 1,2,3 and array indexes are 0,1,2 thus +1
			if (newColorSpace === CompareConstants.colorSpace.DEFAULT) {
				// note we have the same icon for colorspace 0 (default) and 1 (channels)
				newColorSpace = CompareConstants.colorSpace.CHANNELS;
			}
			button.setActive(index + 1 === newColorSpace);
		},
		this
	);
};

/**
 * Initialize threshold slider
 */
CompareDifferenceSubToolbarComponent.prototype._initThresholdSlider = function() {
	this.slider = new Slider(this.f, {
		onSliderChanged: _.bind(
			_.debounce(function() {
				this.compareDiffMode.setThreshold(this.slider.getValue());
			}, 250),
			this
		)
	});

	this.slider.setValue(this.compareDiffMode.getThreshold()); // initial value

	// synchronize between model and slider
	this._registerListener(
		"changed.threshold",
		function(oldValue, newValue) {
			this.slider.setValue(newValue);
		},
		this,
		this.compareDiffMode
	);

	this.f.i18n.loadBundle(
		"viewer",
		function() {
			this.slider.setTooltip(this.f.i18n.t("ChangeDifferenceThresholdTT"));
		},
		this
	);
	this.element.append(this.slider.element);
};

/**
 * Initialize the addSubtractedColorConfiguration
 */
CompareDifferenceSubToolbarComponent.prototype._initHighlightColor = function() {
	if (!this.f.isMobile()) {
		/**
		 * Highlight Color changes
		 */
		this.diffColorChangeGroup = new ButtonGroup("diffColorChangeGroup");

		this.addButtonGroup(this.diffColorChangeGroup);

		// Main button to change high light colors
		var mainDiffColorChangeIcon = new DoubleColorBlock(
			this.compareDiffMode.getAddedHighlightColor(),
			this.compareDiffMode.getSubtractedHighlightColor()
		);
		var mainDiffColorChangeButton = new DropdownSubtoolbarButton(this.f, "ChangeDiffColor", mainDiffColorChangeIcon.element, true);
		mainDiffColorChangeButton.setTooltip("viewer", "ChangeDifferenceHighlightColorsTT");
		this.diffColorChangeGroup.addButton(mainDiffColorChangeButton);

		// Submenu options

		// icons
		// added highlight color
		var changeAddedHighlightColorIcon = new ColorPicker(this.f, _.bind(this.compareDiffMode.setAddedHighlightColor, this.compareDiffMode));
		changeAddedHighlightColorIcon.setColor(this.compareDiffMode.getAddedHighlightColor());
		// subtracted highlight color
		var changeSubtractedHighlightColorIcon = new ColorPicker(
			this.f,
			_.bind(this.compareDiffMode.setSubtractedHighlightColor, this.compareDiffMode)
		);

		changeSubtractedHighlightColorIcon.setColor(this.compareDiffMode.getSubtractedHighlightColor());
		// both highlight colors
		var changeBothHighlightColorBlock = new DoubleColorBlock(
			this.compareDiffMode.getAddedHighlightColor(),
			this.compareDiffMode.getSubtractedHighlightColor()
		);
		var changeBothHighlightColorIcon = new ColorPicker(
			this.f,
			_.bind(function(newColor) {
				this.compareDiffMode.setAddedHighlightColor(newColor);
				this.compareDiffMode.setSubtractedHighlightColor(newColor);
			}, this),
			changeBothHighlightColorBlock.element
		);
		changeBothHighlightColorIcon.setColor(this.compareDiffMode.getAddedHighlightColor());

		// react on changes -> update all icons
		this.compareDiffMode.on("changed.addedHighlightColor", function(oldValue, newValue) {
			changeAddedHighlightColorIcon.setColor(newValue);
			mainDiffColorChangeIcon.setLeftColor(newValue);
			changeBothHighlightColorBlock.setLeftColor(newValue);
		});
		this.compareDiffMode.on("changed.subtractedHighlightColor", function(oldValue, newValue) {
			changeSubtractedHighlightColorIcon.setColor(newValue);
			mainDiffColorChangeIcon.setRightColor(newValue);
			changeBothHighlightColorBlock.setRightColor(newValue);
		});

		// options
		this.f.i18n.loadBundle(
			"viewer",
			function() {
				mainDiffColorChangeButton.addAction({
					name: "changeAddedHighlightColor",
					icon: changeAddedHighlightColorIcon.element,
					content: this.f.i18n.t("DiffAddColor"),
					enabled: true,
					action: function() {
						changeAddedHighlightColorIcon.show();
					}
				});

				mainDiffColorChangeButton.addAction({
					name: "changeSubtractedHighlightColor",
					icon: changeSubtractedHighlightColorIcon.element,
					content: this.f.i18n.t("DiffSubColor"),
					enabled: true,
					action: function() {
						changeSubtractedHighlightColorIcon.show();
					}
				});

				mainDiffColorChangeButton.addAction({
					name: "changeBothHighlightColor",
					icon: changeBothHighlightColorIcon.element,
					content: this.f.i18n.t("DiffBothColors"),
					enabled: true,
					action: function() {
						changeBothHighlightColorIcon.show();
					}
				});
			},
			this
		);

		// Reset highlight color changes
		var resetDiffColorButton = new SubtoolbarButton(
			this.f,
			"ResetDiffColor",
			"icon-WCR_WF-Undo",
			true,
			_.bind(this.compareDiffMode.resetHighlightColors, this.compareDiffMode)
		);
		resetDiffColorButton.setTooltip("viewer", "DiffResetColor");
		this.diffColorChangeGroup.addButton(resetDiffColorButton);
	}
};

/**
 * Initialize the Difference mode selection (show all, highlight & only difference)
 */
CompareDifferenceSubToolbarComponent.prototype._initDiffmode = function() {
	/**
	 * Mode buttons
	 */
	this.showModeButtons = [];

	var isMobile = this.f.isMobile();

	// dmode 2: view both
	this.showModeButtons[0] = new SubtoolbarButton(
		this.f,
		"differenceShow",
		"icon-WCR_Compare-BothDiff",
		true,
		_.bind(this.setDifferenceMode, this, CompareConstants.showDifferenceMode.SHOW)
	);

	// dmode 3: view both but highlight difference
	this.showModeButtons[1] = new SubtoolbarButton(
		this.f,
		"differenceHighlight",
		"icon-WCR_Compare-HighlightDiff",
		true,
		_.bind(this.setDifferenceMode, this, CompareConstants.showDifferenceMode.HIGHLIGHT)
	);

	// dmode 4: view both but only difference
	this.showModeButtons[2] = new SubtoolbarButton(
		this.f,
		"differenceOnly",
		"icon-WCR_Compare-ShowDiff",
		true,
		_.bind(this.setDifferenceMode, this, CompareConstants.showDifferenceMode.ONLY)
	);

	// tool tips
	this.f.i18n.loadBundle(
		"viewer",
		function() {
			this.showModeButtons[0].setTooltip(this.f.i18n.t("DiffBothTT"));
			this.showModeButtons[1].setTooltip(this.f.i18n.t("DiffBothHighlightTT"));
			this.showModeButtons[2].setTooltip(this.f.i18n.t("DiffBothShowTT"));
		},
		this
	);

	this.differenceShowModeGroup = new ButtonGroup("differenceShowMode", this.showModeButtons);
	this.differenceShowModeGroup.element.addClass("toggleBtnGroup");
	this.addButtonGroup(this.differenceShowModeGroup);

	this.compareDiffMode.on("changed.differenceMode", _.bind(this.updateDifferenceModeButtons, this));
	this.updateDifferenceModeButtons(0, this.compareDiffMode.getDifferenceMode());
};

/**
 * Update the difference mode buttons (activate/deactivate them)
 * @param {CompareConstants.showDifferenceMode} oldMode
 * @param {CompareConstants.showDifferenceMode} newMode
 */
CompareDifferenceSubToolbarComponent.prototype.updateDifferenceModeButtons = function(oldMode, newMode) {
	var diffModeIcons = ["icon-WCR_Compare-BothDiff", "icon-WCR_Compare-HighlightDiff", "icon-WCR_Compare-ShowDiff"];

	// difference mode 0 --> 2, 1 --> 3, 2 --> 4 thus +2
	_.each(this.showModeButtons, function(button, index) {
		button.setActive(index + 2 === newMode);
		button.element.removeClass("displayNone");
	});

	this.updateParentIcon(diffModeIcons[newMode - 2]);
};

/**
 * Update the icon of the parent button
 * @param {String} icon
 */
CompareDifferenceSubToolbarComponent.prototype.updateParentIcon = function(icon) {
	if (this.parent) {
		this.parent.setIcon(icon);
	}
};

/**
 * Change the difference mode
 * @param mode
 */
CompareDifferenceSubToolbarComponent.prototype.setDifferenceMode = function(mode) {
	this.compareDiffMode.setDifferenceMode(mode);
};

export default CompareDifferenceSubToolbarComponent;
