import AbstractModel from "@esko/web-viewer/core/AbstractModel";
import LayoutConstants from "@esko/web-viewer/layout/LayoutConstants";
import _ from "underscore";
/**
 * @exports CompareMode
 * @memberof CompareModule
 *
 * @class CompareMode is a base class for compare modes
 * @constructor
 * @param {Core.ViewerFacade} f
 * @param {String} name - name of the compare mode
 * @param {CompareModule.CompareModel} compareModel - the compare model
 * @param {Boolean} enableTogglingAnnotations - Enable the toggling of annotation or not
 * @param {String} annotationTooltipKey - Key (in the general bundle) of the reason message that should be shown when annotations are disabled
 * @param {String} [whichToEnable] - Which of the show annotation buttons to enable?
 */
var CompareMode = function(f, name, compareModel, enableTogglingAnnotations, annotationTooltipKey, whichToEnable) {
	var channel = compareModel.getPrivateChannelNameAsArray();
	channel.push("mode");
	channel.push(name);
	AbstractModel.call(this, f, channel);
	this.f = f;
	this.name = name;
	this._enabled = false;
	this._compareModel = compareModel;

	this._enableTogglingAnnotations = enableTogglingAnnotations;
	this._annotationTooltipKey = annotationTooltipKey;
	this._whichToEnable = whichToEnable;

	this._annotationModuleAvailable = false;
	this.f.whenModuleAvailable(
		"annotation",
		function(annotationModule) {
			this._annotationModuleAvailable = true;
			// by default annotation are shown, so only fire if not enabled
			if (this._enabled) {
				this._fireTemporarilyDisableAnnotations();
			}
		},
		this
	);
};

CompareMode.prototype = Object.create(AbstractModel.prototype);
CompareMode.prototype.constructor = CompareMode;

/**
 * Get the name of the mode
 * @returns {String}
 */
CompareMode.prototype.getName = function() {
	return this.name;
};

/**
 * Enable or disable the compare mode
 * @param {Boolean} toggle
 * @param {CompareModule.CompareConstants.diffMode} [diffMode] - The diff mode in the mode that should be activated
 */
CompareMode.prototype.toggle = function(toggle, diffMode) {
	if (toggle !== this._enabled) {
		this._enabled = toggle;
		// enable/disable annotations
		if (this._annotationModuleAvailable) {
			this._fireTemporarilyDisableAnnotations();
		}
	}
};

/**
 * Is the mode active or not?
 * @returns {Boolean}
 */
CompareMode.prototype.isEnabled = function() {
	return this._enabled;
};

/**
 * Intercept fragmentURL generation. inject the correct parameters
 * @param context
 */
CompareMode.prototype.interceptFragmentURL = function(context) {
	// do nothing;
};

/**
 * Fire the temporarily disable annotations options
 */
CompareMode.prototype._fireTemporarilyDisableAnnotations = function() {
	this.f.loadBundle(
		"viewer",
		function() {
			this.f.emit(
				"module.annotation.temporarilyDisable",
				this._enabled,
				this._enableTogglingAnnotations,
				this.f.t(this._annotationTooltipKey),
				this._whichToEnable
			);
		},
		this
	);
};

CompareMode.prototype.getLayoutMode = function() {
	var page = this.f.getDocumentController().getActivePage();
	if (page && page.is3D()) {
		return LayoutConstants.LayoutMode.COMPARE_3D;
	} else if (this.f.isMobile()) {
		return LayoutConstants.LayoutMode.MOBILE_COMPARE;
	}

	return LayoutConstants.LayoutMode.VIEW2D;
};

export default CompareMode;
