import AbstractModel from "@esko/web-viewer/core/AbstractModel";
import CompareConstants from "@esko/web-viewer/compare/CompareConstants";
import ShapeUtils from "@esko/shapeutils";
import _ from "underscore";
/**
 * @memberof CompareModule
 * @exports CompareModel
 *
 * @class The CompareModel class keeps track of the current compare mode and the needed settings
 * @constructor
 * @param {Core.ViewerFacade} f
 * @param {Array.<String>} channel
 */
var CompareModel = function(f, channel, diffMode) {
	AbstractModel.call(this, f, channel);

	/**
	 * @member The currently activate compare mode
	 * @private
	 * @type CompareModule.CompareMode
	 */
	this._currentMode = null;

	/**
	 * @member The current DMode
	 * @private
	 * @type CompareModule.CompareConstants.diffMode
	 */
	this._diffMode = diffMode ? diffMode : CompareConstants.diffMode.UNDEFINED;

	/**
	 * @member The available compare modes
	 * @private
	 */
	this._availableModes = {};

	/**
	 * @member {ShapeUtils.Point} offset of the reference document page for compare
	 */
	this._offset = new ShapeUtils.Point(0, 0);

	/**
	 * @member {ShapeUtils.Point} offset of the cad overlay document page for compare
	 */
	this._cadOverlayOffset = new ShapeUtils.Point(0, 0);

	/**
	 * @member {CompareConstants.cornerAutoAlign}
	 */
	this._cornerAlign = CompareConstants.cornerAutoAlign.TOPLEFT;

	/**
	 * @member {Boolean} is the reference document page flipped
	 */
	this._referenceFlipped = false;

	/**
	 * @member {Boolean} is the cad overlay document page flipped
	 */
	this._cadOverlayFlipped = false;

	/**
	 * @member {Boolean} is the reference document page negated
	 */
	this._referenceNegated = false;

	/**
	 * @member {CompareConstants.referenceRotation}
	 */
	this._referenceRotation = CompareConstants.referenceRotation.NOTROTATED;

	/**
	 * @member {CompareConstants.referenceRotation}
	 */
	this._cadOverlayRotation = CompareConstants.referenceRotation.NOTROTATED;

	/**
	 * @member {Boolean} use trim box align
	 */
	this._trimBoxAlign = false;

	/**
	 * The name of the most recently activated (or activating) compare mode
	 */
	this._newModeName = null;

	/**
	 * @member {Boolean} used to delay the firing of view server parameters changed
	 */
	this._silenceParametersChanged = false;
	this._fireParametersChanged = false;
};

CompareModel.prototype = Object.create(AbstractModel.prototype);
CompareModel.prototype.constructor = CompareModel;

/**
 * Change the current mode by name.
 * SHOULD BE CALLED FROM THE MODULE ONLY
 * @param {String} name - the name of the new mode
 * @param {CompareModule.CompareConstants.diffMode} [diffMode] - The diff mode that the mode should activate
 */
CompareModel.prototype.activateModeByName = function(name, diffMode) {
	if (this._availableModes[name]) {
		var oldMode = this._currentMode;
		var newMode = this._availableModes[name];

		if (newMode !== oldMode) {
			this.toggleCurrentMode(false);

			this._newModeName = name;

			this._currentMode = newMode;

			//activate the new mode
			newMode.toggle(true, diffMode);

			this._fireChange(false, ["mode"], oldMode, newMode);
		}
	}
};

/**
 * De-activate the current active mode
 * SHOULD BE CALLED FROM THE MODULE ONLY
 */
CompareModel.prototype.deactivateCurrentMode = function() {
	this.toggleCurrentMode(false);
	if (this._currentMode) {
		var oldMode = this._currentMode;
		this._newModeName = "";
		this._currentMode = null;
		this._fireChange(false, ["mode"], oldMode, null);
	}
};

/**
 * Toggle the current active mode
 * SHOULD BE CALLED FROM THE MODULE ONLY
 */
CompareModel.prototype.toggleCurrentMode = function(toggle) {
	if (this._currentMode && this._currentMode.isEnabled() !== toggle) {
		this._currentMode.toggle(toggle);
	}
};

/**
 * Get the mode name based on the given diff mode
 * @param {CompareModule.CompareConstants.diffMode} diffMode - The diff mode of which you want to get the mode name
 * @returns {String} diffMode
 */
CompareModel.prototype.getModeNameByDiffMode = function(diffMode) {
	// Get to the mode name based on the diffMode: first get the name of the diff mode
	var diffModeName = _.invert(CompareConstants.diffMode)[diffMode];
	// then use it in the diffModeToMode mapping to get the mode constant name
	var mode = CompareConstants.diffModeToMode[diffModeName];
	// and then get the value of the mode constant, meaning the mode name.
	return CompareConstants.mode[mode];
};

/**
 * Get a Compare mode by name
 * @param {String|CompareModule.CompareConstants.mode} name
 * @returns {CompareModule.CompareMode}  the requested mode or undefined if the mode doesn't exist
 */
CompareModel.prototype.getModeByName = function(name) {
	return this._availableModes[name];
};

/**
 * sets the offset of the reference document page for compare, set by alignment of the document page
 * @param {ShapeUtils.Point} offset
 */
CompareModel.prototype.setOffset = function(offset) {
	var oldOffset = this._offset;

	this._offset = offset.copy();

	this._fireChange(false, ["offset"], oldOffset, offset);
};

/**
 * Get the offset of the reference page for compare
 * @returns {ShapeUtils.Point}
 */
CompareModel.prototype.getOffset = function() {
	return this._offset.copy();
};

/**
 * reset the offset of the reference document page
 */
CompareModel.prototype.resetOffset = function() {
	this.setOffset(new ShapeUtils.Point(0, 0));
};

/**
 * Set the offset of the cad overlay in the model
 * @param {ShapeUtils.Point} offset
 */
CompareModel.prototype.setCadOverlayOffset = function(offset) {
	var oldOffset = this._cadOverlayOffset;

	this._cadOverlayOffset = offset.copy();

	this._fireChange(false, ["cadOverlayOffset"], oldOffset, offset);
};

/**
 * Get the offset of the CAD overlay
 * @returns {ShapeUtils.Point}
 */
CompareModel.prototype.getCadOverlayOffset = function() {
	return this._cadOverlayOffset.copy();
};

/**
 * Reset the offset of the CAD overlay
 */
CompareModel.prototype.resetCadOverlayOffset = function() {
	this.setCadOverlayOffset(new ShapeUtils.Point(0, 0));
};

/**
 * Set the corner alignment
 * @param {CompareConstants.cornerAutoAlign} corner
 */
CompareModel.prototype.setCornerAlign = function(corner) {
	this._cornerAlign = corner;
};

/**
 * Get the corner alignment
 * @returns {CompareConstants.cornerAutoAlign}
 */
CompareModel.prototype.getCornerAlign = function() {
	return this._cornerAlign;
};

/**
 * reset the corner alignment
 */
CompareModel.prototype.resetCornerAlign = function() {
	this._cornerAlign = CompareConstants.cornerAutoAlign.TOPLEFT;
};

/**
 * Get the rotation of the reference document page
 * @returns {CompareConstants.referenceRotation}
 */
CompareModel.prototype.getReferenceRotation = function() {
	return this._referenceRotation;
};

/**
 * Set the rotation of the reference document page
 * @param {CompareConstants.referenceRotation}
 */
CompareModel.prototype.setReferenceRotation = function(rotation) {
	this._referenceRotation = rotation;
};

/**
 * Flip the reference document page
 */
CompareModel.prototype.flipReference = function() {
	this._referenceFlipped = !this._referenceFlipped;
};

/**
 * Flip the CAD overlay
 * @param {Boolean} [flip] - if undefined just flip from the current flip state, otherwhise if true : page is flipped, if false page is not flipped
 */
CompareModel.prototype.flipCadOverlay = function(flip) {
	if (flip === undefined) {
		flip = !this._cadOverlayFlipped;
	}
	this._cadOverlayFlipped = flip;
};

/**
 * Is the reference document page flipped
 * @returns {Boolean}
 */
CompareModel.prototype.isReferenceFlipped = function() {
	return this._referenceFlipped;
};

/**
 * Is the CAD overlay flipped
 * @returns {Boolean}
 */
CompareModel.prototype.isCadOverlayFlipped = function() {
	return this._cadOverlayFlipped;
};

/**
 * negate the reference document page
 */
CompareModel.prototype.negateReference = function() {
	this._referenceNegated = !this._referenceNegated;
	this.parametersChanged();
};

/**
 * Is the reference document page nagated
 * @returns {Boolean}
 */
CompareModel.prototype.isReferenceNegated = function() {
	return this._referenceNegated;
};

/**
 * Reset all the orientation options of the reference document page (flip, negation and rotation)
 */
CompareModel.prototype.resetReferenceOrientation = function() {
	this._referenceFlipped = false;
	this._referenceNegated = false;
	this._referenceRotation = CompareConstants.referenceRotation.NOTROTATED;
};

/**
 * Reset the orientation options of the CAD overlay (this is the rotation and the flipped state)
 */
CompareModel.prototype.resetCadOverlayOrientation = function() {
	this._cadOverlayRotation = CompareConstants.referenceRotation.NOTROTATED;
	this._cadOverlayFlipped = false;
};

/**
 * toggle the trim box align option
 */
CompareModel.prototype.toggleTrimBoxAlign = function() {
	this._trimBoxAlign = !this._trimBoxAlign;
};

/**
 * Is the trim box align option on
 * @returns {Boolean}
 */
CompareModel.prototype.isTrimBoxAlign = function() {
	return this._trimBoxAlign;
};

/**
 * Reset the trim box algin option (not active)
 */
CompareModel.prototype.resetTrimBoxAlign = function() {
	this._trimBoxAlign = false;
};

/**
 * Set the dmode
 * @param {CompareModule.CompareConstants.diffMode} dmode
 */
CompareModel.prototype.setDiffMode = function(dmode) {
	var oldMode = this._diffMode;
	this._diffMode = dmode;

	if (this._currentMode === null) {
		oldMode = CompareConstants.diffMode.UNDEFINED;
	}

	if (oldMode != dmode) {
		this.parametersChanged();
	}

	this._fireChange(false, ["diffMode"], oldMode, this._diffMode);
};

/**
 * Get the configured dmode
 * @returns {CompareModule.CompareConstants.diffMode}
 */
CompareModel.prototype.getDiffMode = function() {
	return this._diffMode;
};

/**
 * Get the current active compare mode mode
 * @returns {CompareModule.CompareMode} or undefined when no comparison is active
 */
CompareModel.prototype.getActiveMode = function() {
	return this._currentMode;
};

/**
 * Register a new compare mode
 * @param {CompareModule.CompareMode} mode
 */
CompareModel.prototype.registerMode = function(mode) {
	if (this._availableModes[mode.getName()]) {
		throw new Error("Trying to register mode which already exists!", mode.getName(), mode);
	}

	var oldList = _.keys(this._availableModes);
	this._availableModes[mode.getName()] = mode;

	this._fireChange(false, ["availableModes"], oldList, _.keys(this._availableModes));
};

/**
 * Notify the viewport that parameters have changed
 */
CompareModel.prototype.parametersChanged = function() {
	if (this._silenceParametersChanged) {
		// fire it later
		this._fireParametersChanged = true;
	} else {
		this.f.emit("viewport.viewserverParametersChanged");
	}
};

/**
 * Check if the given dmode is a "real" (not current or reference or readerspread diff mode) compare mode.
 * @param {CompareModule.CompareConstants.diffMode} diffMode - The diff mode to check
 * @returns {Boolean} True if the given dmode is a "real" compare mode, false otherwise.
 */
CompareModel.prototype.isRealCompareMode = function(diffMode) {
	return (
		diffMode >= CompareConstants.diffMode.SHOWBOTHCHANGES &&
		diffMode <= CompareConstants.diffMode.SHOWBOTHSIDEBYSIDECHANGESONLY &&
		diffMode !== CompareConstants.diffMode.READERSPREAD
	);
};

/**
 * is the currently active compare mode read spread
 * @returns {Boolean}
 */
CompareModel.prototype.isReaderSpreadActive = function() {
	return this._newModeName === CompareConstants.mode.READERSPREAD;
};

/**
 * Silence parameters changed or not
 * @param {Boolean} toggle - True to silence, false to unsilence
 */
CompareModel.prototype.toggleSilenceParametersChanged = function(toggle) {
	if (toggle) {
		this._silenceParametersChanged = true;
	} else {
		this._silenceParametersChanged = false;
		if (this._fireParametersChanged) {
			this.parametersChanged();
			this._fireParametersChanged = false;
		}
	}
};

/**
 * Set the rotation of the CAD overlay
 * @param {Number} rotation The rotation in degrees
 */
CompareModel.prototype.setCadOverlayRotation = function(rotation) {
	this._cadOverlayRotation = rotation;
};

/**
 * Get the rotation of the CAD overlay
 * @returns {Number} the rotation in degrees
 */
CompareModel.prototype.getCadOverlayRotation = function() {
	return this._cadOverlayRotation;
};

export default CompareModel;
