/**
 * @namespace CompareModule
 */
/**
 * @namespace CompareModule
 */
import AbstractModule from "@esko/web-viewer/core/AbstractModule";

import CompareModel from "@esko/web-viewer/compare/CompareModel";
import CompareShowCurrentMode from "@esko/web-viewer/compare/CompareShowCurrentMode";
import CompareShowDifferenceMode from "@esko/web-viewer/compare/CompareShowDifferenceMode";
import CompareBlinkingMode from "@esko/web-viewer/compare/CompareBlinkingMode";
import CompareShowBothOnTopMode from "@esko/web-viewer/compare/CompareShowBothOnTopMode";
import Compare3DSideBySideMode from "@esko/web-viewer/compare/Compare3DSideBySideMode";
import Compare3DBlinking from "@esko/web-viewer/compare/Compare3DBlinking";
import Compare3DCurrent from "@esko/web-viewer/compare/Compare3DCurrent";
import Compare3DReference from "@esko/web-viewer/compare/Compare3DReference";
import CompareAlignmentWizard from "@esko/web-viewer/compare/alignment/CompareAlignmentWizard";
import CompareToolbarComponent from "@esko/web-viewer/compare/CompareToolbarComponent";
import CompareTouchToolbarComponent from "@esko/web-viewer/compare/CompareTouchToolbarComponent";
import CompareWidget from "@esko/web-viewer/compare/CompareWidget";
import CompareConstants from "@esko/web-viewer/compare/CompareConstants";
import ComparePagelistPopup from "@esko/web-viewer/compare/ComparePagelistPopup";
import AbstractCompareDataAdapter from "@esko/web-viewer/compare/AbstractCompareDataAdapter";
import DataConstants from "@esko/web-viewer/data/DataConstants";
import DocumentSelection from "@esko/web-viewer/layout/component/DocumentSelection";
import ModalPopup from "@esko/web-viewer/layout/component/ModalPopup";
import LayoutConstants from "@esko/web-viewer/layout/LayoutConstants";
import ListenerUtil from "@esko/web-viewer/utils/ListenerUtil";
import XMLUtils from "@esko/xmlutils";
import ShapeUtils from "@esko/shapeutils";
import ViewerUtils from "@esko/web-viewer/utils/ViewerUtils";
import Mousetrap from "@esko/mousetrap";
import _ from "underscore";
import $ from "jquery";
/**
 * Initialize the Compare Module
 *
 * @exports CompareModule
 * @memberof CompareModule
 * @class
 * The Compare module enables comparison in the viewer
 * @constructor
 * @augments {Core.AbstractModule}
 * @param {Core.ViewerFacade} f
 * @param {Boolean} silenced - if true, the module shouldn't publish tools/widgets/toolbarcomponents
 * @param {String} name - unique name of the module
 */
var CompareModule = function(f, silenced, dataAdapter) {
	AbstractModule.call(this, f, silenced, "compare", dataAdapter || new AbstractCompareDataAdapter(f));
	ListenerUtil.mixin(this, f);

	var diffMode = this.f.getOption("PREFERENCES_DIFFMODE_COMPARE");
	this._compareModel = new CompareModel(this.f, this.getPrivateChannelNameAsArray(), diffMode);

	/**
	 * @member {DocumentSelection} Active and Difference document selectors
	 */
	this._documentSelection = new DocumentSelection(
		this.f,
		this,
		_.bind(function() {
			// update widget, if it's already loaded
			if (this._compareWidget) {
				this._compareWidget.updateCompareWidget();
			}
		}, this),
		this.f.isMobile()
	);

	// REGISTER ALL AVAILABLE MODES
	this._compareModel.registerMode(new CompareShowCurrentMode(this.f, this._compareModel));
	this._compareModel.registerMode(new CompareShowDifferenceMode(this.f, this._compareModel));
	this._compareModel.registerMode(new CompareBlinkingMode(this.f, this._compareModel));
	this._compareModel.registerMode(new Compare3DSideBySideMode(this.f, this._compareModel));
	this._compareModel.registerMode(new Compare3DBlinking(this.f, this._compareModel));
	this._compareModel.registerMode(new Compare3DCurrent(this.f, this._compareModel));
	this._compareModel.registerMode(new Compare3DReference(this.f, this._compareModel));
	this._compareModel.registerMode(this.createCompareShowReaderSpreadMode());

	this._differenceMode = this.createCompareDifferenceMode();
	this._compareModel.registerMode(this._differenceMode);

	this._sideBySideMode = this.createCompareSideBySideMode();
	this._compareModel.registerMode(this._sideBySideMode);

	this._compareModel.registerMode(new CompareShowBothOnTopMode(this.f, this._compareModel));

	this._toolbarComponent = null;
	this._compareWidget = null;

	this._comparePagelistPopup = null;

	this._compareAlignmentWizard = null;

	this._publishToolbarComponentAndWidgetIfNeeded();

	this._docListener = this._registerListener("doc.loaded", this._publishToolbarComponentAndWidgetIfNeeded, this);

	this._registerListener("intercept.fragmentURL", this._interceptFragmentURL, this);

	this._registerListener("intercept.parameters", this._interceptFragmentURL, this);

	this._settingDefault = false;

	// Listen to the compare model for saving the diffMode in the user settings

	// first time, check if the viewer was opened in compare
	if (this.f.getProperty("compare") === "1" && this._compareModel.isRealCompareMode(diffMode)) {
		var diffPage = this.getDefaultDiffPage(this.f.getActivePage());
		if (diffPage) {
			this.activateModeByName(this.f.getActivePage().getAvailableCompareModes(diffPage)[0], diffMode, diffPage);
		}
		this._settingDefault = true;
	} else {
		this._compareModel.setDiffMode(CompareConstants.diffMode.UNDEFINED);
	}
	// listen to main modes
	this._registerListener(
		"changed.mode",
		function(oldValue, newValue) {
			var currentDiffMode = this.getCurrentDiffMode();
			if (newValue === null && this.f.getDocumentController().getDifferencePage() !== undefined) {
				this.changeReferenceOffset(new ShapeUtils.Point(0, 0));
			}
			if (this._compareModel.isRealCompareMode(currentDiffMode)) {
				this.f.setOption("PREFERENCES_DIFFMODE_COMPARE", currentDiffMode);
			}
		},
		this,
		this._compareModel
	);
	// also listen to difference mode sub modes
	this._registerListener(
		"changed.differenceMode",
		function(oldValue, newValue) {
			var compareMode = newValue;
			//first check if side-by-side is active
			if (this._compareModel.getActiveMode().getName() === CompareConstants.mode.SIDEBYSIDE) {
				//map the difference mode to a side-by-side difference mode
				if (compareMode === CompareConstants.diffMode.SHOWBOTHCHANGESHIGHLIGHT) {
					compareMode = CompareConstants.diffMode.SHOWBOTHSIDEBYSIDEHIGHLIGHT;
				} else if (compareMode === CompareConstants.diffMode.SHOWBOTHCHANGESONLY) {
					compareMode = CompareConstants.diffMode.SHOWBOTHSIDEBYSIDECHANGESONLY;
				} else {
					compareMode = CompareConstants.diffMode.SHOWBOTHSIDEBYSIDE;
				}
			}
			this.f.setOption("PREFERENCES_DIFFMODE_COMPARE", compareMode);
		},
		this,
		this._differenceMode
	);
	// don't listen to the user options (assuming this is the only place where the option will be changed, so no need to update on such events because they come from this module)

	// When we were not setting a default reference page (meaning, the user selected the page himself) and when a second document is selected and there was no previous document selected, activate the last saved compare mode
	this.f.getDocumentController().whenDifferencePageChanged(function(oldPage, newPage) {
		if (!this._settingDefault && oldPage === undefined && newPage && !this._compareModel.isReaderSpreadActive()) {
			this.activateModeByDiffMode(this.f.getOption("PREFERENCES_DIFFMODE_COMPARE"));
		}
	}, this);
	this.f.getDocumentController().whenActiveOrDifferencePageChanged(this._whenActiveOrDifferencePageChanged, this);
	this.f
		.getDocumentController()
		.whenActiveOrDifferencePageChangedAndParametersAvailableOrChange(
			this._whenActiveOrDifferencePageChangedAndParametersAvailableOrChange,
			this
		);

	// On changes in the document selector
	this._registerListener(
		"changeReferencePage",
		_.bind(function(newDiffPage) {
			if (
				this.f
					.getActivePage()
					.getDocumentVersion()
					.isMultiPage()
			) {
				// only possible when difference selector is set, so comparePagelistPopup is available then
				this._comparePagelistPopup.setComparePage(newDiffPage);
				this._comparePagelistPopup.show();
			} else {
				this.activateModeByDiffMode(this.f.getOption("PREFERENCES_DIFFMODE_COMPARE"), newDiffPage);
			}
		}, this),
		this,
		this._documentSelection
	);
	this._registerListener(
		"changeActivePage",
		_.bind(function(newActivePage, callback) {
			//first check if the new active page is a multipage and the current active mode is reader spread
			if (this._compareModel.isReaderSpreadActive() && !newActivePage.getDocumentVersion().isMultiPage()) {
				this.activateModeByName();
			}
			var isActivePageChanged = this.f.getDocumentController().changeActivePage(newActivePage);
			if (callback) {
				callback(isActivePageChanged);
			}
		}, this),
		this,
		this._documentSelection
	);

	// set the CAD overlay when a new cad overlay was selected in the document selector
	this._registerListener(
		"changeCadOverlay",
		_.bind(function(newCadOverlay) {
			this.f.getDocumentController().changeCadOverlay(newCadOverlay);
		}, this),
		this,
		this._documentSelection
	);

	this._registerListener("intercept.activePageChange", this._interceptActivePageChange, this);

	this.f.getDocumentController().whenDifferencePageChanged(function(oldPage, newPage) {
		var navigationModule = this.f.getModule("navigation");
		if (navigationModule) {
			navigationModule.resetExtraRotation();
		}
		if (this._compareModel.isRealCompareMode(this._compareModel.getDiffMode())) {
			this._previousDifferencePage = newPage;
		}
	}, this);

	this.f.getDocumentController().whenCadOverlayChanged(function(oldPage, newPage) {
		if (newPage) {
			//make sure the same data of the compare model is set to the CAD overlay, we keep the same alignment as before
			this.changeCadOverlayOffset(this._compareModel.getCadOverlayOffset());
			this.flipCadOverlay(this._compareModel.isCadOverlayFlipped());
			this.setCadOverlayRotation(this._compareModel.getCadOverlayRotation());
		}
	}, this);

	this.f.getDocumentController().whenActivePageChanged(function(oldPage, newPage) {
		var navigationModule = this.f.getModule("navigation");
		if (navigationModule) {
			navigationModule.resetExtraRotation();
		}

		this._compareModel.resetCadOverlayOffset(); //first reset the alignment data
		this._compareModel.resetCadOverlayOrientation();
		if (newPage && newPage.hasLinkedPage()) {
			//if there is a linked page, go set the correct alignment data
			var cadOverlayPage = newPage.getLinkedPage();

			if (cadOverlayPage) {
				var overlayOffset = new ShapeUtils.Point();
				var overlayIsFlipped = false;
				var overlayRotation = 0;

				if (
					this.f
						.getDocumentController()
						.getActivePage()
						.hasLinkedPage() &&
					this.f
						.getDocumentController()
						.getActivePage()
						.getLinkedPage() === cadOverlayPage
				) {
					overlayOffset = new ShapeUtils.Point(
						ViewerUtils.parseFloat(cadOverlayPage.get("offsetx")),
						ViewerUtils.parseFloat(cadOverlayPage.get("offsety"))
					);
					overlayIsFlipped = cadOverlayPage.get("overlayIsFlipped");
					overlayRotation = cadOverlayPage.get("overlayRotation");
				}

				this.changeCadOverlayOffset(overlayOffset);
				this.flipCadOverlay(overlayIsFlipped);
				this.setCadOverlayRotation(overlayRotation);
			}
		}

		if (this._3DWidget) {
			this._3DWidget.setEnabled(false);
		}
	}, this);

	this._registerListener(
		"newLinkedCad",
		function(oldCadOverlay, newCadOverlay) {
			if (oldCadOverlay !== newCadOverlay) {
				var oldCadId = null;
				if (oldCadOverlay) {
					oldCadId = oldCadOverlay.getUniqueID();
				}
				this._documentSelection.setNewlyCreatedCadOverlay(this.f.getActivePage(), oldCadId);
			}
		},
		this,
		this.f.getDocumentController()
	);

	// Keyboard shortcuts
	Mousetrap.bind(
		"ctrl+l",
		_.bind(function(e) {
			var diffPage = this.f.getDocumentController().getDifferencePage();
			if (diffPage === undefined || this._compareModel.isReaderSpreadActive() || diffPage.hasCad()) {
				return;
			}

			this.autoAlign();

			e.preventDefault();
			e.stopPropagation();
		}, this),
		"keydown"
	);
	Mousetrap.bind(
		"home",
		_.bind(function(e) {
			var pageId = this._documentSelection.getFirstSelectablePageId();
			if (pageId) {
				var newActivePage = this.f.getDocumentController().getDocumentPage(pageId);
				if (newActivePage) {
					this.f.getDocumentController().changeActivePage(newActivePage);
				}
			}
			e.preventDefault();
			e.stopPropagation();
		}, this)
	);

	Mousetrap.bind(
		"end",
		_.bind(function(e) {
			var pageId = this._documentSelection.getLastSelectablePageId();
			if (pageId) {
				var newActivePage = this.f.getDocumentController().getDocumentPage(pageId);
				if (newActivePage) {
					this.f.getDocumentController().changeActivePage(newActivePage);
				}
			}
			e.preventDefault();
			e.stopPropagation();
		}, this)
	);

	Mousetrap.bind(
		"pagedown",
		_.bind(function(e) {
			var pageId = this._documentSelection.getNextSelectablePageId();
			if (pageId) {
				var newActivePage = this.f.getDocumentController().getDocumentPage(pageId);
				if (newActivePage) {
					this.f.getDocumentController().changeActivePage(newActivePage);
				}
			}
			e.preventDefault();
			e.stopPropagation();
		}, this)
	);

	Mousetrap.bind(
		"pageup",
		_.bind(function(e) {
			var pageId = this._documentSelection.getPreviousSelectablePageId();
			if (pageId) {
				var newActivePage = this.f.getDocumentController().getDocumentPage(pageId);
				if (newActivePage) {
					this.f.getDocumentController().changeActivePage(newActivePage);
				}
			}
			e.preventDefault();
			e.stopPropagation();
		}, this)
	);

	Mousetrap.bind(
		"ctrl+s",
		_.bind(function(e) {
			var currentDiffMode = this.getCurrentDiffMode();
			if (currentDiffMode === CompareConstants.diffMode.UNDEFINED || this.isReaderSpreadActive()) {
				return;
			}
			var mode = this.getActiveMode().getName();

			if (CompareConstants.modes2D[mode]) {
				if (mode === CompareConstants.mode.SHOW_CURRENT) {
					// go to diff version
					this.activateModeByName(CompareConstants.mode.SHOW_DIFFERENCE);
				} else {
					// go to current version
					this.activateModeByName(CompareConstants.mode.SHOW_CURRENT);
				}
			} else if (CompareConstants.modes3D[mode]) {
				if (mode === CompareConstants.mode.CURRENT3D) {
					// go to diff version
					this.activateModeByName(CompareConstants.mode.REFERENCE3D);
				} else {
					// go to current version
					this.activateModeByName(CompareConstants.mode.CURRENT3D);
				}
			}

			e.preventDefault();
			e.stopPropagation();
		}, this),
		"keydown"
	);
};

CompareModule.prototype = Object.create(AbstractModule.prototype);
CompareModule.prototype.constructor = CompareModule;

/**
 * Get the current active diff mode in saveable format (meaning, blinking returns BLINKING_CURRENT instead of CURRENT/REFERENCE)
 * @returns {CompareConstants.diffMode} The current active diff mode
 */
CompareModule.prototype.getCurrentDiffMode = function() {
	var compareMode = this._compareModel.getActiveMode();
	var diffMode = null;
	// blinking switches between current/reference but we should save BLINKING (otherwise we cannot restore blinking)
	if (compareMode !== null && compareMode.getName() === CompareConstants.mode.BLINKING) {
		diffMode = CompareConstants.diffMode.BLINKING_CURRENT;
	} else {
		diffMode = this._compareModel.getDiffMode();
	}
	return diffMode;
};

/**
 * This method will activate the last used compare mode with the last used difference file
 */
CompareModule.prototype.activateLastUsedCompareMode = function() {
	var lastUsedDiffMode = this.f.getOption("PREFERENCES_DIFFMODE_COMPARE");
	if (this._compareModel.isRealCompareMode(lastUsedDiffMode)) {
		if (this._previousDifferencePage === this.f.getActivePage()) {
			delete this._previousDifferencePage;
		}

		this.activateModeByDiffMode(this.f.getOption("PREFERENCES_DIFFMODE_COMPARE"), this._previousDifferencePage);
		if (this._previousState) {
			this.setState(this._previousState);
			delete this._previousState;
		}
	}
};

/**
 * Change the current mode by mode name
 * @param {String|CompareModule.CompareConstants.mode} [name] - the name of the new mode, if undefined/null/falsy the current mode is unloaded
 * @param {CompareModule.CompareConstants.diffMode} [diffMode] - Diff mode to set in the mode specified by name
 * @param {AbstractDocumentPage} [diffPage] - Diff page to set if none is active
 */
CompareModule.prototype.activateModeByName = function(name, diffMode, diffPage) {
	if (name === undefined) {
		if (this._compareModel.isRealCompareMode(this._compareModel.getDiffMode())) {
			this._previousState = this.getState();
		} else {
			delete this._previousState;
		}

		// we deactivate the current mode
		this._compareModel.deactivateCurrentMode();
		this.f.getDocumentController().resetDifferencePage();
	} else {
		var activeMode = this._compareModel.getActiveMode();
		if (!activeMode || activeMode.getName() !== name) {
			this._compareModel.toggleSilenceParametersChanged(true);
			this._compareModel.toggleCurrentMode(false);

			if (name === CompareConstants.mode.READERSPREAD || name === CompareConstants.mode.ALIGNMENT) {
				// we go to readerspread; don't set diff page, reader spread will do it first
				// or we go to alignment mode, this doesn't need difference page when in cad overlay
				this._compareModel.activateModeByName(name, diffMode);

				diffPage = this.f.getDocumentController().getDifferencePage();
				if (diffPage && !diffPage.isBlankPage()) {
					diffPage.whenParametersAvailable(
						_.bind(function() {
							this._compareModel.toggleSilenceParametersChanged(false);
							if (this._compareWidget) {
								this._compareWidget.updateCompareWidget();
							}
						}, this)
					);
				} else {
					this._compareModel.toggleSilenceParametersChanged(false);
				}
			} else {
				// to another mode
				// Make sure the diff page is filled in
				this.f.getDocumentController().whenDocumentsLoaded(function() {
					if (!diffPage) {
						diffPage = this.f.getDocumentController().getDifferencePage();
					}
					if (!diffPage) {
						// we don't know which one to set, so do default logic
						diffPage = this.getDefaultDiffPage(this.f.getActivePage(), name);
					}

					// disable restore of last saved mode, see module constructor
					this._settingDefault = true;
					this.f.getDocumentController().changeDifferencePage(diffPage);
					this._settingDefault = false;

					diffPage = this.f.getDocumentController().getDifferencePage();
					if (diffPage) {
						this._compareModel.activateModeByName(name, diffMode);
						diffPage.whenParametersAvailable(
							_.bind(function() {
								this._compareModel.toggleSilenceParametersChanged(false);
								if (this._compareWidget) {
									this._compareWidget.updateCompareWidget();
								}
							}, this)
						);
					} else {
						this._compareModel.toggleSilenceParametersChanged(false);
					}
				}, this);
			}
		} else {
			// the mode should not be change, but maybe we should update the diff page
			if (diffPage) {
				// disable restore of last saved mode, see module constructor
				this._settingDefault = true;
				this.f.getDocumentController().changeDifferencePage(diffPage);
				this._settingDefault = false;
			}
		}
	}
};

/**
 * Change the current mode by diffMode
 * @param {CompareModule.CompareConstants.diffMode} diffMode - The diff mode
 * @param {AbstractDocumentPage} [diffPage] - Diff page that should be taken if none is active
 */
CompareModule.prototype.activateModeByDiffMode = function(diffMode, diffPage) {
	if (diffMode === CompareConstants.diffMode.UNDEFINED) {
		this.activateModeByName();
	} else {
		var modeName = this._compareModel.getModeNameByDiffMode(diffMode);
		var activePage = this.f.getDocumentController().getActivePage();
		var availableCompareModes = activePage.getAvailableCompareModes(diffPage);
		if (diffPage) {
			var referenceCompareModes = diffPage.getAvailableCompareModes(activePage);
			availableCompareModes = _.intersection(availableCompareModes, referenceCompareModes);
		}
		if (availableCompareModes.indexOf(modeName) !== -1) {
			this.activateModeByName(modeName, diffMode, diffPage);
		} else {
			this.activateModeByName(availableCompareModes[0], diffMode, diffPage);
		}
	}
};

/**
 * Get the default diff doc page for the current/given active document page
 * Make sure all documents are loaded before calling this function!
 * @param {AbstractDocumentPage} [activePage] - If defined, use this page to calculate the diff page; otherwise use the current active page
 * @param {String} [mode] - the mode in which the page will be diffed.
 * @return {AbstractDocumentPage} The default diff doc page for the current/given active document page or undefined if nothing useful can be found
 */
CompareModule.prototype.getDefaultDiffPage = function(activePage, mode) {
	if (!activePage) {
		activePage = this.f.getActivePage();
	}

	// First see if there is a linked compare page
	var linkedComparePage = activePage.getLinkedComparePage();
	if (linkedComparePage) {
		return linkedComparePage;
	} else {
		var diffVersion = this._getNextDiffDocumentVersion(activePage, undefined, mode);

		if (diffVersion) {
			// for multi page only, get corresponding page
			if (
				(activePage.getDocumentVersion().isMultiPage() && diffVersion.isMultiPage()) ||
				(activePage.getDocumentVersion().isPageList() && diffVersion.isPageList())
			) {
				var correspondingDiffPage = diffVersion.getPage(activePage.getPagenum());
				if (correspondingDiffPage && !correspondingDiffPage.isBlankPage()) {
					if (correspondingDiffPage === activePage) {
						// same one as the active page, so search another one
						diffVersion = this._getNextDiffDocumentVersion(activePage, diffVersion, mode);

						if (diffVersion) {
							correspondingDiffPage = diffVersion.getPage(activePage.getPagenum());
							if (correspondingDiffPage) {
								return correspondingDiffPage;
							} else {
								// if the active document has more pages than the diff, show the last
								// other case is a multipage active document with a normal diff document -> last = first in that case.
								return diffVersion.getLastPage();
							}
						} else {
							//can happen when pagelist has only 1 version
							// no other document found -> should not happen
							return undefined;
						}
					} else {
						return correspondingDiffPage;
					}
				} else {
					if (diffVersion.isPageList() && (correspondingDiffPage === undefined || correspondingDiffPage.isBlankPage())) {
						//check if blank or empty page
						var pageNum = activePage.getPagenum();
						pageNum++;
						//as long the page is empty or blank try the next page
						while (pageNum < diffVersion.getPageCount() && (!diffVersion.hasPage(pageNum) || diffVersion.getPage(pageNum).isBlankPage())) {
							pageNum++;
						}
						if (pageNum < diffVersion.getPageCount() && diffVersion.getPage(pageNum) !== activePage) {
							//there is a next page that's not empty or blank
							return diffVersion.getPage(pageNum);
						} else {
							//no next page is found, select the first non blank page for this document version
							return diffVersion.getFirstNotBlankPage();
						}
					}
					// if the active document has more pages than the diff, show the last
					// other case is a multipage active document with a normal diff document -> last = first in that case.
					return diffVersion.getLastPage();
				}
			} else {
				return diffVersion.getFirstPage();
			}
		} else {
			// no other document found, check if there are other pages
			if (activePage.getDocumentVersion().isMultiPage()) {
				// return the first if it's different
				var firstPage = activePage.getDocumentVersion().getFirstPage();
				if (activePage === firstPage) {
					var nextDiffPage = firstPage.getNextPage();
					while (nextDiffPage.isBlankPage() && nextDiffPage.getPagenum() < nextDiffPage.getDocumentVersion().getPageCount()) {
						nextDiffPage = nextDiffPage.getNextPage();
					}
					return nextDiffPage;
				} else {
					return firstPage;
				}
			}
			return undefined;
		}
	}
};

/**
 * Get the next diff document version based on the current active document page.
 * First another document is tried (latest version), then previous version.
 * Make sure all documents are loaded before calling this function!
 * @param {AbstractDocumentPage} activePage - Use this page to calculate the diff page
 * @param {AbstractDocumentVersion} [excludeVersion] - Version to skip in the search of a new diff version (next to active document page)
 * @return {AbstractDocumentVersion} Next diff document version or undefined if none found.
 */
CompareModule.prototype._getNextDiffDocumentVersion = function(activePage, excludeVersion, mode) {
	var activeDoc = activePage.getDocument();

	var newDoc = _.find(this.f.getDocumentController().getAllDocuments(), function(document) {
		// if exclude version is defined, make sure we exclude it
		return (
			(!excludeVersion || (excludeVersion && document !== excludeVersion.getDocument())) &&
			document !== activeDoc &&
			_.all(document.getDocumentVersion().getAllPages(), function(page) {
				return page.canBeComparedTo(activePage, mode);
			})
		);
	});

	var diffVersion;

	if (newDoc !== undefined) {
		// Get the latest version of that document
		diffVersion = newDoc.getDocumentVersion();
	} else {
		// There is no other document, but maybe other versions
		var activeDocVersion = activePage.getDocumentVersion();
		// if multiple versions
		if (activeDoc.hasMultipleVersions() && activeDocVersion.getVersionId() > 1) {
			// previous version
			diffVersion = activeDoc.getDocumentVersion(activeDocVersion.getVersionId() - 1);
		}
	}

	var oldDiffVersion = diffVersion;
	//if it's a page list make sure only the same pages of previous version are used to compare to, since only these can be picked in dropdown menu
	if (activePage.getDocumentVersion().isPageList() && diffVersion !== undefined && !diffVersion.hasPage(activePage.getPagenum())) {
		while (diffVersion !== undefined && diffVersion.getVersionId() > 0 && !diffVersion.hasPage(activePage.getPagenum())) {
			diffVersion = activeDoc.getDocumentVersion(diffVersion.getVersionId() - 1);
		}
		//if the page has no previous versions, set the current version as diffVersion
		if (diffVersion !== undefined && !diffVersion.hasPage(activePage.getPagenum())) {
			diffVersion = activePage.getDocumentVersion();
		}
	}

	if (excludeVersion && diffVersion === excludeVersion) {
		// we exclude it if needed
		if (oldDiffVersion && oldDiffVersion === excludeVersion) {
			return undefined;
		} else {
			return oldDiffVersion;
		}
	} else {
		return diffVersion;
	}
};

/**
 * @override
 */
CompareModule.prototype._interceptFragmentURL = function(c) {
	if (this._compareModel.getActiveMode()) {
		this.f.getDataAdapterForModule("compare").convertCompareParameters(this._compareModel, c.params);

		// Do mode specific things
		this._compareModel.getActiveMode().interceptFragmentURL(c);
	}
};

/**
 * Get the comparemodel
 * @returns {CompareModule.CompareModel}
 */
CompareModule.prototype.getCompareModel = function() {
	return this._compareModel;
};

/**
 * Get the diffMode from the model
 * @returns {CompareConstants.diffMode}
 */
CompareModule.prototype.getDiffMode = function() {
	return this._compareModel.getDiffMode();
};

/**
 * Get the active Compare mode
 * @returns {CompareModule.CompareMode} the active mode (or null)
 */
CompareModule.prototype.getActiveMode = function() {
	return this._compareModel.getActiveMode();
};

/**
 * Get the offset of the reference page for compare
 * @returns {ShapeUtils.Point}
 */
CompareModule.prototype.getReferenceOffset = function() {
	return this._compareModel.getOffset();
};

/**
 * Is it possible to use the compare functionallity in this session (based on which documents are loaded)
 * @returns {Boolean}
 */
CompareModule.prototype.isCompareAvailable = function() {
	if (!this._compareWidget && !this._toolbarComponent) {
		var show = this._shouldShowWidgetOrTools();

		return show.widget || show.tools;
	} else {
		return true;
	}
};

/**
 * Check if CAD overlay is allowed (graphical document is activated, and there is at least 1 other CAD file loaded (or graphics document with linked page))
 * @returns {Boolean}
 */
CompareModule.prototype.isCadOverlayAllowed = function() {
	var activePage = this.f.getActivePage();

	if (
		!activePage.hasCad() &&
		!activePage.is3D() &&
		!activePage.hasSvg() &&
		!activePage.hasVideo() &&
		this.f
			.getLayoutMode()
			.getMainPanelContent()
			.indexOf(LayoutConstants.Widget.VIEWPORT) !== -1
	) {
		if (activePage.canSaveCadOverlay() || activePage.hasLinkedPage()) {
			return true;
		}
		var activeDocument = this.f.getDocumentController().getActiveDocument();
		return _.some(this.f.getDocumentController().getDocuments(), function(document) {
			if (document != activeDocument) {
				var version = document.getDocumentVersion();
				if (version) {
					var page = version.getPage();
					if (page) {
						return page.hasCad() || page.hasLinkedPage();
					}
				}
			}
			return false;
		});
	}

	return false;
};

/**
 * Check if the CAD overlay settings can be saved for the currently active page
 * @returns {Boolean}
 */
CompareModule.prototype.canSaveCadOverlay = function() {
	var activePage = this.f.getActivePage();
	return activePage.canSaveCadOverlay();
};

/**
 * Is the CAD overlay changed since the current page was loaded
 * @returns {Boolean}
 */
CompareModule.prototype.isCadOverlayModified = function() {
	var activePage = this.f.getActivePage();
	var linkedCad = activePage.getLinkedPage();
	var cadOverlay = this.f.getDocumentController().getCadOverlay();
	return (
		linkedCad != cadOverlay ||
		(linkedCad &&
			linkedCad.documentType == DataConstants.documentPageType.CAD &&
			((linkedCad.getStateParam("compareOffset") !== undefined &&
				ViewerUtils.parseFloat(linkedCad.get("offsetx")) !== linkedCad.getStateParam("compareOffset").x) ||
				(linkedCad.getStateParam("compareOffset") !== undefined &&
					ViewerUtils.parseFloat(linkedCad.get("offsety")) !== linkedCad.getStateParam("compareOffset").y) ||
				(linkedCad.getStateParam("rotation") !== undefined &&
					ViewerUtils.parseFloat(linkedCad.get("overlayRotation")) !== linkedCad.getStateParam("rotation")) ||
				(linkedCad.getStateParam("flipped") !== undefined && linkedCad.get("overlayIsFlipped") !== linkedCad.getStateParam("flipped"))))
	);
};

/**
 * Save the CAD overlay for the current active page
 * @param {function} [callback] - called when the save either fails or is successful. The argument is an boolean indicating success
 */
CompareModule.prototype.saveCadOverlay = function(callback) {
	var graphicsPage = this.f.getActivePage();
	var context = {
		saveCadOverlay: true,
		callback: _.partial(_.bind(this.saveCadOverlay, this), callback)
	};

	this.f.emit("intercept.savingCadOverlay", context);

	if (context.saveCadOverlay) {
		this.f.getDataAdapterForModule("compare").saveCadOverlay(
			graphicsPage,
			this.f.getDocumentController().getCadOverlay(),
			this._compareModel.getCadOverlayRotation(),
			this._compareModel.isCadOverlayFlipped(),
			this._compareModel.getCadOverlayOffset(),
			_.bind(function(success) {
				if (success) {
					graphicsPage.set("isCadAlignmentSaved", true);
				}

				// update the compare widget
				if (this._compareWidget) {
					this._compareWidget.updateCompareWidget();
				}

				// execute callback
				if (callback) {
					callback(success);
				}
			}, this)
		);
	}
};

/**
 * Should the widget or compare tools be shown?
 * @returns {Object} Object with a boolean for widget and tools (whether they should be shown)
 */
CompareModule.prototype._shouldShowWidgetOrTools = function() {
	// it is posisble that we only show the widget and not the compare tool
	var showWidget = false;
	var showCompareTools = false;
	if (this.f.getDocumentController().getDocumentCount() > 2) {
		// there's more than 2 documents, show everything!
		showWidget = true;
		showCompareTools = true;
	} else {
		// there are either 2 documents or 1 document
		var documents = this.f.getDocumentController().getAllDocuments();
		if (
			_.some(documents, function(doc) {
				return doc.hasMultipleVersions();
			})
		) {
			// at least 1 document has more than 1 version, so show all controls
			showWidget = true;
			showCompareTools = true;
		} else {
			if (documents.length == 2) {
				showWidget = true;

				if (
					_.chain(documents[0].getDocumentVersions()[0].getAllPages())
						.union(documents[1].getDocumentVersions()[0].getAllPages())
						.reduce(function(memo, page) {
							if (memo === undefined) {
								return page.getAvailableCompareModes();
							} else {
								return _.intersection(memo, page.getAvailableCompareModes());
							}
						}, undefined)
						.value().length > 0
				) {
					showCompareTools = true;
				}
			} else {
				// there is 1 document with 1 version
				if (documents[0].getDocumentVersion().getPageCount() > 1) {
					// there are multiple pages
					showWidget = true;
					showCompareTools = true;
				}
				if (this.f.getActivePage().hasLinkedPage()) {
					showWidget = true;
					showCompareTools = false;
				}
			}
		}

		if (!showWidget) {
			//maybe we still need to show the widget when the user can save CAD Overlay Alignment
			if (
				_.some(documents, function(document) {
					return _.some(document.getDocumentVersion().getAllPages(), function(page) {
						return page.canSaveCadOverlay();
					});
				})
			) {
				showWidget = true;
			}
		}
	}
	return {
		widget: showWidget,
		tools: showCompareTools
	};
};

/**
 * Publish the toolbarcomponent if available
 */
CompareModule.prototype._publishToolbarComponentAndWidgetIfNeeded = function() {
	if (this._toolbarComponent === null || this._compareWidget === null) {
		//should be both either null or initialized

		var show = this._shouldShowWidgetOrTools();

		if (show.tools) {
			// publish the compare buttons, make the buttons work!
			this._publishToolbarComponent(
				CompareToolbarComponent.toolbarName,
				_.bind(function(layoutType) {
					if (layoutType === LayoutConstants.LayoutType.MOBILE) {
						this._toolbarComponent = new CompareTouchToolbarComponent(this.f, this);
					} else {
						this._toolbarComponent = new CompareToolbarComponent(this.f, this);
					}
					return this._toolbarComponent;
				}, this)
			);
		}

		if (show.widget) {
			// publish the compare widget
			this._publishWidget(
				CompareWidget.widgetName,
				_.bind(function(layoutType) {
					this._compareWidget = new CompareWidget(this.f, this, this._compareModel, this._documentSelection);

					this._comparePagelistPopup = new ComparePagelistPopup(this.f, "comparePagelistPopup");
					this._comparePagelistPopup.setOkAction(
						_.partial(function(that) {
							// that = module, this = popup
							switch (this._currentPagelistCompareRange) {
								case CompareConstants.pagelistCompareRange.CURRENT:
									that._doCurrentCompareRange(this._pagelistComparePage);
									break;
								case CompareConstants.pagelistCompareRange.CURRENTANDFOLLOWING:
									that._doCurrentAndFollowingCompareRange(this._pagelistComparePage);
									break;
								case CompareConstants.pagelistCompareRange.ALL:
									that._doAllCompareRange(this._pagelistComparePage);
									break;
							}

							that.activateModeByDiffMode(that.f.getOption("PREFERENCES_DIFFMODE_COMPARE"), this._pagelistComparePage);
							this.hide();
						}, this)
					);

					this._comparePagelistPopup.setCancelAction(
						_.bind(function() {
							this._documentSelection.setDifferenceSelectorPage(this.f.getDocumentController().getDifferencePage());
						}, this)
					);

					//listen to the widget for auto align action
					this._registerListener("autoAlign", this.autoAlign, this, this._compareWidget);
					//listen to the widget for corner align
					this._registerListener("doCornerAlign", this._doCornerAlign, this, this._compareWidget);
					//listen to the widget for rotate, flip and negate reference document page
					this._registerListener("rotateReference", this.rotateReference, this, this._compareWidget);
					this._registerListener("flipReference", this.flipReference, this, this._compareWidget);
					this._registerListener(
						"negateReference",
						_.bind(function() {
							this._compareModel.negateReference();
						}, this),
						this,
						this._compareWidget
					);
					//listen to set the trim box align on or off
					this._registerListener("trimBoxAlign", this._doTrimBoxAlign, this, this._compareWidget);
					//listen to when the offset is changed manually
					this._registerListener(
						"offsetXChanged",
						_.bind(function(offsetX) {
							var offset = this._compareModel.getOffset();
							offset.x = offsetX;
							this.changeReferenceOffset(offset);
						}, this),
						this,
						this._compareWidget
					);
					this._registerListener(
						"offsetYChanged",
						_.bind(function(offsetY) {
							var offset = this._compareModel.getOffset();
							offset.y = offsetY;
							this.changeReferenceOffset(offset);
						}, this),
						this,
						this._compareWidget
					);
					//listen to when the cad overlay offset is changed manually
					this._registerListener(
						"cadOverlayOffsetXChanged",
						_.bind(function(offsetX) {
							var offset = this._compareModel.getCadOverlayOffset();
							offset.x = offsetX;
							this.changeCadOverlayOffset(offset);
						}, this),
						this,
						this._compareWidget
					);
					this._registerListener(
						"cadOverlayOffsetYChanged",
						_.bind(function(offsetY) {
							var offset = this._compareModel.getCadOverlayOffset();
							offset.y = offsetY;
							this.changeCadOverlayOffset(offset);
						}, this),
						this,
						this._compareWidget
					);
					this._registerListener("rotateCadOverlay", this.rotateCadOverlay, this, this._compareWidget);
					this._registerListener("flipCadOverlay", this.flipCadOverlay, this, this._compareWidget);
					this._registerListener("saveCadOverlay", this.saveCadOverlay, this, this._compareWidget);

					if (this._docListener) {
						this._unregisterListener(this._docListener);
						this._docListener = null;
					}

					this._compareWidget.updateCompareWidget();
					return this._compareWidget;
				}, this)
			);
		}

		if (show.widget || show.tools) {
			this._emit("compareAvailable");
		}
	}
};

/**
 * Executed when active or difference page has changed and before the parameters are fetched...
 */
CompareModule.prototype._whenActiveOrDifferencePageChanged = function() {
	this._compareModel.resetOffset();

	// reset corner align
	this._compareModel.resetCornerAlign();
	this._compareModel.resetReferenceOrientation();
	this._compareModel.resetTrimBoxAlign();

	if (!this.f.getActivePage().canBeComparedTo(this.f.getDocumentController().getDifferencePage()).length === 0) {
		this.activateModeByName(); // If this page cannot be compared to, set compare mode to no compare
	}

	this._showOrHideCompareToolbarComponent();

	if (this._compareWidget) {
		this._compareWidget.updateCompareWidget();
	}
};

/**
 * Shows or Hides the toolbar component
 *
 */
CompareModule.prototype._showOrHideCompareToolbarComponent = function() {
	if (this._toolbarComponent !== null) {
		if (this.f.getActivePage().isBlankPage() || !this.f.getActivePage().canBeComparedTo()) {
			this._toolbarComponent.hide();
			if (this._compareWidget) {
				this._compareWidget.toggleDifferenceDocPageSelector(true);
			}
		} else {
			this._toolbarComponent.show();
			if (this._compareWidget) {
				this._compareWidget.toggleDifferenceDocPageSelector(false);
			}
		}
	}
};

/**
 * Executed when the active or difference page changed and the parameters are available
 * @private
 */
CompareModule.prototype._whenActiveOrDifferencePageChangedAndParametersAvailableOrChange = function(
	page,
	referencePage,
	oldPage,
	isActivePageChanged
) {
	if (this._compareWidget) {
		this._compareWidget.updateCompareWidget();
	}

	this._showOrHideCompareToolbarComponent();
};

/**
 * align reference document page to a corner (or center) of the current document page
 * @private
 * @param {CompareConstants.cornerAutoAlign} corner - or null to reset
 */
CompareModule.prototype._doCornerAlign = function(corner) {
	var docController = this.f.getDocumentController();
	docController.willNeedParameterUpdate("CORNER_ALIGN");

	this._compareModel.resetOffset();
	if (this._compareWidget) {
		this._compareWidget.updateCompareWidget();
	}
	this._compareModel.setCornerAlign(corner);
	this.changeReferenceOffset(this._compareModel.getOffset());
	docController.readyForParameterUpdate("CORNER_ALIGN");
};

/**
 * align the current document page and the reference document page on their trimboxes
 */
CompareModule.prototype._doTrimBoxAlign = function() {
	var docController = this.f.getDocumentController();
	docController.willNeedParameterUpdate("TRIMBOX_ALIGN");

	this._compareModel.toggleTrimBoxAlign();

	docController.readyForParameterUpdate("TRIMBOX_ALIGN");
};

/**
 * auto align the two document page
 */
CompareModule.prototype.autoAlign = function() {
	var viewport = this.f.getModule("viewport");
	var visibleRect = viewport.getVisibleRect();
	var context = {
		rectangle: visibleRect,
		// the servlet that should be used to fetch the image information
		servlet: "",
		// the parameters used for requesting the image
		params: {
			x1: 0,
			x2: 0,
			y1: 0,
			y2: 0,
			isTrimBoxAlign: this._compareModel.isTrimBoxAlign()
		}
	};

	// ask all interceptors to update the context. All listeners have access to the context object and can adapt the information that is in the object
	// there is no predefined order in which these interceptors are called, their actions thus should not overlap!
	this.f.emit("intercept.fragmentURL", context);

	this.doAutoAlign(context);
};

/**
 *
 */
CompareModule.prototype._getCornerAlignOffset = function() {
	var corner = this._compareModel.getCornerAlign();
	var cornerOffset = new ShapeUtils.Size();

	var page = this.f.getActivePage();
	var reference = this.f.getDocumentController().getDifferencePage();

	//get the size of the reference document page
	var diffSize = reference.getPageSize();

	//get the size of the current document page
	var pageSize = page.getPageSize();

	if (
		corner === CompareConstants.cornerAutoAlign.TOPLEFT ||
		corner === CompareConstants.cornerAutoAlign.CENTERLEFT ||
		corner === CompareConstants.cornerAutoAlign.BOTTOMLEFT
	) {
		//No offset for x
	} else if (
		corner === CompareConstants.cornerAutoAlign.TOPCENTER ||
		corner === CompareConstants.cornerAutoAlign.CENTERED ||
		corner === CompareConstants.cornerAutoAlign.BOTTOMCENTER
	) {
		cornerOffset.width = pageSize.width / 2 - diffSize.width / 2;
	} else if (
		corner === CompareConstants.cornerAutoAlign.TOPRIGHT ||
		corner === CompareConstants.cornerAutoAlign.CENTERRIGHT ||
		corner === CompareConstants.cornerAutoAlign.BOTTOMRIGHT
	) {
		cornerOffset.width = pageSize.width - diffSize.width;
	}

	if (
		corner === CompareConstants.cornerAutoAlign.TOPLEFT ||
		corner === CompareConstants.cornerAutoAlign.TOPCENTER ||
		corner === CompareConstants.cornerAutoAlign.TOPRIGHT
	) {
		//No offset for y
	} else if (
		corner === CompareConstants.cornerAutoAlign.CENTERLEFT ||
		corner === CompareConstants.cornerAutoAlign.CENTERED ||
		corner === CompareConstants.cornerAutoAlign.CENTERRIGHT
	) {
		cornerOffset.height = pageSize.height / 2 - diffSize.height / 2;
	} else if (
		corner === CompareConstants.cornerAutoAlign.BOTTOMLEFT ||
		corner === CompareConstants.cornerAutoAlign.BOTTOMCENTER ||
		corner === CompareConstants.cornerAutoAlign.BOTTOMRIGHT
	) {
		cornerOffset.height = pageSize.height - diffSize.height;
	}

	return cornerOffset;
};

/**
 * Change the offset of the current and reference document page, so both document pages are visible
 * @param {ShapeUtils.Point} offset
 */
CompareModule.prototype.changeReferenceOffset = function(referenceOffset) {
	this.f.getDocumentController().willNeedParameterUpdate("COMPARE_OFFSET");
	this._emit("referenceOffset.beforeChange");

	this._compareModel.setOffset(referenceOffset);

	if (this._compareWidget !== null) {
		this._compareWidget.updateCompareWidget();
	}

	var diff = this.f.getDocumentController().getDifferencePage();
	diff.setStateParam("compareOffset", referenceOffset);

	this.f.getDocumentController().readyForParameterUpdate("COMPARE_OFFSET");
	this.f.getDocumentController().whenParametersAvailable(function() {
		this._emit("referenceOffset.changed");
	}, this);
};

/**
 * Change the offset between the CAD overlay and the active document page (Graphics), it's the offset between the CAD overlay and the Graphics top-left corner(if in CAD is upright positioned)
 * @param {ShapeUtils.Point} cadOverlayerOffset
 */
CompareModule.prototype.changeCadOverlayOffset = function(cadOverlayerOffset) {
	this._compareModel.setCadOverlayOffset(cadOverlayerOffset);

	if (this._compareWidget !== null) {
		this._compareWidget.updateCompareWidget();
	}

	var cadOverlay = this.f.getDocumentController().getCadOverlay();
	if (cadOverlay) {
		this.f.getDocumentController().willNeedParameterUpdate("CAD_OVERLAY_OFFSET");
		this._emit("cadOverlayOffset.beforeChange");

		cadOverlay.setStateParam("compareOffset", cadOverlayerOffset);

		this.f.getDocumentController().readyForParameterUpdate("CAD_OVERLAY_OFFSET");
		this.f.getDocumentController().whenParametersAvailable(function() {
			this._emit("cadOverlayOffset.changed");
		}, this);
	}
};

/**
 * @override
 */
CompareModule.prototype.getState = function() {
	var state = {
		// Compare model
		compareModel: {
			diffMode: this.getCurrentDiffMode(),
			cornerAlign: this._compareModel.getCornerAlign(),
			trimBoxAlign: this._compareModel.isTrimBoxAlign(),
			offset: this._compareModel.getOffset(),
			cadOverlayOffset: this._compareModel.getCadOverlayOffset(),
			cadOVerlayRotation: this._compareModel.getCadOverlayRotation(),
			referenceFlipped: this._compareModel.isReferenceFlipped(),
			referenceNegated: this._compareModel.isReferenceNegated(),
			referenceRotation: this._compareModel.getReferenceRotation()
		},
		// Difference mode
		differenceMode: {
			threshold: this._differenceMode.getThreshold(),
			colorSpace: this._differenceMode.getColorSpace(),
			addedColor: this._differenceMode.getAddedHighlightColor(),
			subtractedColor: this._differenceMode.getSubtractedHighlightColor()
		},
		// Side by side mode
		sideBySideMode: {
			panelsLayoutMode: this._sideBySideMode.getPanelsLayoutMode(),
			showingDifference: this._sideBySideMode.isShowingDifference()
		}
	};
	return state;
};

/**
 * @override
 * @param state
 */
CompareModule.prototype.setState = function(state) {
	this.f.whenModuleAvailable(
		"compare",
		function() {
			if (state !== undefined) {
				// Compare model
				var compareModelState = state.compareModel;
				if (compareModelState !== undefined) {
					this.activateModeByDiffMode(compareModelState.diffMode);

					var reference = this.f.getDocumentController().getDifferencePage();
					if (reference !== undefined) {
						this._compareModel.setCornerAlign(compareModelState.cornerAlign);
						if (compareModelState.trimBoxAlign) {
							this._doTrimBoxAlign();
						}
						if (compareModelState.referenceFlipped) {
							this.flipReference();
						}
						if (compareModelState.referenceNegated) {
							this._compareModel.negateReference();
						}
						if (compareModelState.referenceRotation) {
							this.setReferenceRotation(compareModelState.referenceRotation);
						}
						if (!this._compareModel.isReaderSpreadActive()) {
							this.changeReferenceOffset(compareModelState.offset);
						}
						if (compareModelState.cadOverlayOffset) {
							this._compareModel.setCadOverlayOffset(compareModelState.cadOverlayOffset);
						}
						if (compareModelState.cadOverlayRotation) {
							this._compareModel.setCadOverlayRotation(compareModelState.cadOverlayRotation);
						}
					}
				}
				// Difference mode
				var differenceModeState = state.differenceMode;
				if (differenceModeState !== undefined) {
					this._differenceMode.setThreshold(differenceModeState.threshold);
					this._differenceMode.setColorSpace(differenceModeState.colorSpace);
					this._differenceMode.setAddedHighlightColor(differenceModeState.addedColor);
					this._differenceMode.setSubtractedHighlightColor(differenceModeState.subtractedColor);
				}
				// Side by side mode
				var sideBySideModeState = state.sideBySideMode;
				if (sideBySideModeState !== undefined) {
					this._sideBySideMode.setPanelsLayoutMode(sideBySideModeState.panelsLayoutMode);
					this._sideBySideMode.setShowingDifference(sideBySideModeState.showingDifference);
				}
			}
			this.f.debug.log("STATE", "compare state is injected.");
		},
		this
	);
};

/**
 * Is Reader spread active?
 * @returns {Boolean} True if active, false otherwise.
 */
CompareModule.prototype.isReaderSpreadActive = function() {
	return this._compareModel.isReaderSpreadActive();
};

/**
 * Do current compare range
 * @param {AbstractDocumentPage} diffPage - The selected diff page to link to the active page
 */
CompareModule.prototype._doCurrentCompareRange = function(diffPage) {
	var activePage = this.f.getActivePage();

	activePage.setLinkedComparePage(diffPage);
};

/**
 * Do current and following compare range
 * @param {AbstractDocumentPage} diffPage - The selected diff page to link to the active page (and serves as base for following pages)
 */
CompareModule.prototype._doCurrentAndFollowingCompareRange = function(diffPage) {
	var activePage = this.f.getActivePage();
	var activeDocumentVersion = activePage.getDocumentVersion();
	var diffDocumentVersion = diffPage.getDocumentVersion();

	var activePageNumber = activePage.getPagenum();
	var diffPageNumber = diffPage.getPagenum();
	while (activePageNumber <= activeDocumentVersion.getPageCount()) {
		activePage = activeDocumentVersion.getPage(activePageNumber);

		// if there are no pages left, keep the last one
		if (diffPageNumber <= diffDocumentVersion.getPageCount()) {
			diffPage = diffDocumentVersion.getPage(diffPageNumber);
		}
		if (activePage !== undefined && !activePage.isBlankPage()) {
			activePage.setLinkedComparePage(diffPage);
		}

		activePageNumber++;
		diffPageNumber++;
	}
};

/**
 * Do all compare range
 * @param {AbstractDocumentPage} diffPage - The selected diff page to link to the active page (and serves as base for following/previous pages)
 */
CompareModule.prototype._doAllCompareRange = function(diffPage) {
	// do current and next
	this._doCurrentAndFollowingCompareRange(diffPage);

	// do previous
	var activePage = this.f.getActivePage();
	var activeDocumentVersion = activePage.getDocumentVersion();
	var diffDocumentVersion = diffPage.getDocumentVersion();

	var activePageNumber = activePage.getPagenum() - 1; // current already done
	var diffPageNumber = diffPage.getPagenum() - 1; // current already done
	while (activePageNumber >= 1) {
		activePage = activeDocumentVersion.getPage(activePageNumber);

		// if there are no pages left, keep the first one
		if (diffPageNumber >= 1) {
			diffPage = diffDocumentVersion.getPage(diffPageNumber);
		}

		if (activePage !== undefined && !activePage.isBlankPage()) {
			activePage.setLinkedComparePage(diffPage);
		}

		activePageNumber--;
		diffPageNumber--;
	}
};

/**
 * Start up the auto alignment wizard
 */
CompareModule.prototype.startAlignmentWizard = function(cadOverlay) {
	if (!this._compareAlignmentWizard) {
		this._compareAlignmentWizard = new CompareAlignmentWizard(this.f, this, this._compareModel, this.getCompareAlignmentModeConstructor());
	}
	this._compareAlignmentWizard.setForCadOverlay(cadOverlay);

	this.f.startMode(this._compareAlignmentWizard);
};

/**
 * Intercept active page change to make sure the diff page is set correctly.
 * @param {Object} context - Context of the interceptor
 */
CompareModule.prototype._interceptActivePageChange = function(context) {
	var navigationModule = this.f.getModule("navigation");
	if (navigationModule) {
		navigationModule.resetExtraRotation();
	}

	var currentDiffPage = this.f.getDocumentController().getDifferencePage();
	if (currentDiffPage && !this.isReaderSpreadActive() && context.newActivePage.getDocumentVersion().isMultiPage()) {
		// there was already a diff page, and we are not in reader spread and the new one will be a multi page document, so update it if needed
		var newDiffPage = this.getDefaultDiffPage(context.newActivePage);
		if (newDiffPage) {
			// if nothing useful found, leave it like it is
			context.diffPage = newDiffPage;
		}
	}

	if (context.newActivePage.getDocumentVersion().isPageList()) {
		this._documentSelection.updateDocumentSelectorsForPageList(context.oldActivePage, context.newActivePage);
		if (!this.isReaderSpreadActive() && context.newActivePage.isBlankPage()) {
			context.diffPage = undefined;
			this.activateModeByName();
		}
	}
};

/**
 * Flip the reference page
 */
CompareModule.prototype.flipReference = function() {
	var docController = this.f.getDocumentController();
	docController.willNeedParameterUpdate("FLIP");

	this._compareModel.flipReference();
	var diff = docController.getDifferencePage();
	diff.flipPage();

	docController.readyForParameterUpdate("FLIP");
};

/**
 * Flip the cad overlay
 * @param {Boolean} [flip] - if undefined just flip from the current flip state, otherwhise if true : page is flipped, if false page is not flipped
 */
CompareModule.prototype.flipCadOverlay = function(flip) {
	this._compareModel.flipCadOverlay(flip);

	var docController = this.f.getDocumentController();
	var cadOverlay = docController.getCadOverlay();
	if (cadOverlay) {
		docController.willNeedParameterUpdate("FLIP_OVERLAY");

		cadOverlay.flipPage(flip);

		docController.readyForParameterUpdate("FLIP_OVERLAY");
	}
};

/**
 * Rotate the reference page with 1 step of 90 degrees
 */
CompareModule.prototype.rotateReference = function() {
	this.setReferenceRotation((this._compareModel.getReferenceRotation() + 90) % 360);
};

/**
 * Set the rotation of the reference page
 * @param {Number} rotation - rotation in degreees
 */
CompareModule.prototype.setReferenceRotation = function(rotation) {
	var docController = this.f.getDocumentController();
	docController.willNeedParameterUpdate("ROTATION");

	rotation = rotation % 360;
	this._compareModel.setReferenceRotation(rotation);

	var diff = this.f.getDocumentController().getDifferencePage();
	diff.setRotation(rotation);
	var navigationModule = this.f.getModule("navigation");
	if (navigationModule && diff.hasCad()) {
		navigationModule.setExtraRotation(rotation);
	}

	docController.readyForParameterUpdate("ROTATION");
};

/**
 * Rotate the CAD overlay
 */
CompareModule.prototype.rotateCadOverlay = function() {
	this.setCadOverlayRotation((this._compareModel.getCadOverlayRotation() + 90) % 360);
};

/**
 * Set the CAD overlay to a given rotation
 * @param {Number} rotation - rotation in degrees
 */
CompareModule.prototype.setCadOverlayRotation = function(rotation) {
	rotation = rotation % 360;
	this._compareModel.setCadOverlayRotation(rotation);

	var docController = this.f.getDocumentController();
	var cadOverlay = docController.getCadOverlay();
	if (cadOverlay) {
		docController.willNeedParameterUpdate("ROTATION_OVERLAY");

		cadOverlay.setRotation(rotation);
		var navigationModule = this.f.getModule("navigation");
		if (navigationModule) {
			navigationModule.setExtraRotation(rotation);
		}
		docController.readyForParameterUpdate("ROTATION_OVERLAY");
	}
};

/**
 * Publishes and shows the 3D widget
 */
CompareModule.prototype.checkAligmentAndShow3D = function() {
	if (this.isCadOverlayModified()) {
		this.f.loadBundle(
			"viewer",
			function() {
				var promptSaveCADAlignPopup = new ModalPopup(this.f, "saveCADAlignment", this.f.getContext());
				promptSaveCADAlignPopup.setRemoveFromDomAfterUsage(true);

				// Header + Content
				if (this.canSaveCadOverlay()) {
					promptSaveCADAlignPopup.setOkAction(
						_.bind(function() {
							this.saveCadOverlay(_.bind(this._load3DView, this));
							promptSaveCADAlignPopup.hide();
						}, this)
					);
					promptSaveCADAlignPopup.setHeader(
						$("<span/>")
							.addClass("popupTitle")
							.text(this.f.i18n.t("CadOverlayPromptTitle"))
					);
					promptSaveCADAlignPopup.setContent(
						$("<div/>")
							.addClass("unsavedItemsContent")
							.text(this.f.i18n.t("CadOverlayPromptContent"))
					);
					promptSaveCADAlignPopup.setOkBtnText(this.f.i18n.t("Save"));
				} else {
					promptSaveCADAlignPopup.setOkAction(
						_.bind(function() {
							this._load3DView();
							promptSaveCADAlignPopup.hide();
						}, this)
					);
					promptSaveCADAlignPopup.setHeader(
						$("<span/>")
							.addClass("popupTitle")
							.text(this.f.i18n.t("CadOverlayPromptOpen3D"))
					);
					promptSaveCADAlignPopup.setContent(
						$("<div/>")
							.addClass("unsavedItemsContent")
							.text(this.f.i18n.t("CadOverlayPromptOpen3DContent"))
					);
					promptSaveCADAlignPopup.setOkBtnText(this.f.i18n.t("CadOverlayPromptOpen3D"));
				}

				promptSaveCADAlignPopup.setCancelBtnText(this.f.i18n.t("AppDialogCancel"));

				promptSaveCADAlignPopup.show();
			},
			this
		);
	} else {
		if (!this.f.getActivePage().get("isCadAlignmentSaved") && this.canSaveCadOverlay()) {
			this.saveCadOverlay(_.bind(this._load3DView, this));
		} else {
			this._load3DView();
		}
	}
};

/**
 * load the 3D view, loads the module which will publish the widget if needed
 */
CompareModule.prototype._load3DView = function() {
	this.f.requestModule(
		"view3D",
		_.bind(function(view3DModule) {
			view3DModule.show3D();
		}, this),
		false
	);
};

CompareModule.prototype.createCompareDifferenceMode = function() {
	throw "Not Implemented: createCompareShowDifferenceMode";
};

CompareModule.prototype.createCompareSideBySideMode = function() {
	throw "Not Implemented: createCompareSideBySideMode";
};

CompareModule.prototype.createCompareShowReaderSpreadMode = function() {
	throw "Not Implemented: CompareShowReaderSpreadMode";
};

CompareModule.prototype.getCompareAlignmentModeConstructor = function() {
	throw "Not Implemented: getCompareAlignmentModeConstructor";
};

CompareModule.prototype.doAutoAlign = function context() {
	throw "Not Implemented: CompareShowReaderSpreadMode";
};

export default CompareModule;
