import _ from "underscore";
import CompareMode from "@esko/web-viewer/compare/CompareMode";
import CompareConstants from "@esko/web-viewer/compare/CompareConstants";
import ViewportConstants from "@esko/web-viewer/viewport/ViewportConstants";
import PagelistConstants from "@esko/web-viewer/pagelist/PagelistConstants";
import ListenerUtil from "@esko/web-viewer/utils/ListenerUtil";
import LayoutConstants from "@esko/web-viewer/layout/LayoutConstants";
/**
 * @memberof CompareModule
 * @exports CompareShowReaderSpreadMode
 * @class This mode is only available for multipage or pagelist files and shows two pages next to each other.
 * @constructor
 * @param {Core.ViewerFacade} f
 * @param {CompareModule.CompareModel} compareModel - model for compare
 */
var CompareShowReaderSpreadMode = function(f, compareModel) {
	CompareMode.call(this, f, CompareConstants.mode.READERSPREAD, compareModel, false, "", "current");
	ListenerUtil.mixin(this, f);

	/**
	 * @member {AbstractDocumentPage} - current diffpage
	 */
	this.diffPage = null;

	/**
	 * @member {Number} - parameter for url, 1 when the active page is on the left of reader spread, 2 if the active page is on the right of reader spread
	 */
	this.rs = CompareConstants.pagelistSide.RIGHTPAGE;
};

CompareShowReaderSpreadMode.prototype = Object.create(CompareMode.prototype);
CompareShowReaderSpreadMode.prototype.constructor = CompareShowReaderSpreadMode;

/**
 * Start/stop show compare mode
 *
 * @param {Boolean} toggle - should the reader spread mode be toggled on or off
 */
CompareShowReaderSpreadMode.prototype.toggle = function(toggle) {
	CompareMode.prototype.toggle.call(this, toggle);
	this.f.getDocumentController().willNeedParameterUpdate("READER_SPREAD");
	if (toggle) {
		var isFetchingParameters = this.f.getDocumentController().getDifferencePage() !== undefined;
		this.f.getDocumentController().resetDifferencePage();
		this._compareModel.setDiffMode(CompareConstants.diffMode.READERSPREAD);
		var activePage = this.f.getDocumentController().getActivePage();
		activePage.setReaderSpread(true);
		var diffPageno = this._calculateReferencePageNumber(activePage);

		this.diffPage = activePage.getDocumentVersion().getPage(diffPageno);
		if (this.diffPage !== undefined) {
			//is undefined when the active page is the first page
			this.diffPage.resetParameters();
			activePage.resetParameters();
			this.f.getDocumentController().changeDifferencePage(this.diffPage);
			this.diffPage.setReaderSpread(true);
		}
		if (this.rs === CompareConstants.pagelistSide.RIGHTPAGE) {
			// active page is shown on the right
			activePage.set("isRightPage", true); //set a property of the page to indicate if the active page is right or left in reader spread
			if (this.diffPage === undefined) {
				//if there is no reference page, calculate the width of the active page and set as offset
				if (isFetchingParameters) {
					activePage.resetParameters();
				}
			} else {
				this.diffPage.set("isRightPage", false);
			}
		} else {
			//for left sided pages, no offset is needed
			activePage.set("isRightPage", false); //set is right page to false, page is left side of reader spread
			if (this.diffPage) {
				this.diffPage.set("isRightPage", true);
			}
		}
		//register listener to when the active page is being changed, a difference pages needs to be set
		this._registerListener("intercept.activePageChange", this._whenActivePageChanges, this, this.f);
	} else {
		//turn the compare mode off
		var activePage = this.f.getDocumentController().getActivePage();
		activePage.reset("isRightPage");
		if (this.diffPage) {
			this.diffPage.reset("isRightPage");
			this.diffPage.setReaderSpread(false);
		}
		activePage.setReaderSpread(false); //make sure the readerSpread attribute of the page is set to false
		//reset the difference page, this page is not the default reference page wanted for other compare modes as its the previous or next page of the same version
		this.diffPage = null;
		this.f.getDocumentController().resetDifferencePage();
		this._compareModel.setDiffMode(CompareConstants.diffMode.UNDEFINED);
		this._unregisterAllListeners(); //unregister the listeners
		// see compare module for resetting diff page
	}
	this.f.getDocumentController().readyForParameterUpdate("READER_SPREAD");
};

/**
 * Intercept the fragment URL for construction of the URL
 * @param {Object} c - object with all the parameters for the URL
 */
CompareShowReaderSpreadMode.prototype.interceptFragmentURL = function(c) {
	throw "Not Implemented";
};

/**
 * When the active page changes, calculate the difference page and set the correct offset on the active page
 *
 * @param {AbstractDocumentPage} oldActivePage
 * @param {AbstractDocumentPage} activePage
 */
CompareShowReaderSpreadMode.prototype._whenActivePageChanges = function(pages) {
	if (this.isEnabled()) {
		//first make sure the compare mode is activated
		if (pages.oldActivePage !== undefined) {
			//set the reader spread of the old active document page to false
			pages.oldActivePage.setReaderSpread(false);
			pages.oldActivePage.reset("isRightPage");
		}
		if (this.diffPage) {
			this.diffPage.reset("isRightPage");
			this.diffPage.setReaderSpread(false);
		}

		pages.newActivePage.setReaderSpread(true); //set the reader spread for the new active page
		var diffPageno = this._calculateReferencePageNumber(pages.newActivePage);
		this.diffPage = pages.newActivePage.getDocumentVersion().getPage(diffPageno);

		if (this.diffPage === undefined) {
			pages.diffPage = undefined;
		} else {
			this.diffPage.resetParameters();
			this.diffPage.setReaderSpread(true);
			pages.diffPage = this.diffPage;
		}

		//set the offset of the active document page
		if (this.rs === CompareConstants.pagelistSide.RIGHTPAGE) {
			if (this.diffPage !== undefined) {
				pages.newActivePage.resetParameters();
				this.diffPage.set("isRightPage", false);
			}
			pages.newActivePage.set("isRightPage", true);
		} else {
			if (this.diffPage) {
				this.diffPage.set("isRightPage", true);
			}
			pages.newActivePage.set("isRightPage", false);
		}
	}
};

/**
 * Calculates the the number of the reference page for reader spread mode
 * @param {AbstractDocumentPage} activePage - The current active document page
 * @returns {Number}
 */
CompareShowReaderSpreadMode.prototype._calculateReferencePageNumber = function(activePage) {
	var activePagenum = activePage.getPagenum();
	var activePageBinding = activePage.getDocument().get("binding");
	var diffPagenum;
	if (activePagenum % 2 === 0) {
		// even pages
		diffPagenum = activePagenum + 1;
		if (activePageBinding === undefined || activePageBinding === PagelistConstants.binding.LEFT) {
			// even pages with left (or undefined) binding come left
			this.rs = CompareConstants.pagelistSide.LEFTPAGE;
		} else if (activePageBinding === PagelistConstants.binding.RIGHT) {
			// even pages with right binding come right
			this.rs = CompareConstants.pagelistSide.RIGHTPAGE;
		}
	} else {
		// odd pages
		diffPagenum = activePagenum - 1;
		if (activePageBinding === undefined || activePageBinding === PagelistConstants.binding.LEFT) {
			// odd pages with left (or undefined) binding come right
			this.rs = CompareConstants.pagelistSide.RIGHTPAGE;
		} else if (activePageBinding === PagelistConstants.binding.RIGHT) {
			// odd pages with right binding come left
			this.rs = CompareConstants.pagelistSide.LEFTPAGE;
		}
	}

	return diffPagenum;
};

CompareShowReaderSpreadMode.prototype.getLayoutMode = function() {
	if (this.f.isMobile()) {
		return LayoutConstants.LayoutMode.MOBILE_READERSPREAD;
	} else {
		return CompareMode.prototype.getLayoutMode.call(this);
	}
};

export default CompareShowReaderSpreadMode;
