import _ from "underscore";
import CompareMode from "@esko/web-viewer/compare/CompareMode";
import CompareConst from "@esko/web-viewer/compare/CompareConstants";
import ViewportConstants from "@esko/web-viewer/viewport/ViewportConstants";
import ListenerUtil from "@esko/web-viewer/utils/ListenerUtil";
/**
 * @memberof CompareModule
 * @exports CompareSideBySideMode
 * @class This mode shows a side by side view of the current and reference document. It enables the user to show a difference panel too (therefor it relies on compareShowDifferenceMode)
 * @constructor
 * @param {Core.ViewerFacade} f
 * @param {CompareModule.CompareModel} compareModel - model for compare
 * @param {CompareModule.CompareDifferenceMode} differenceMode - the difference mode implementation
 */
var CompareSideBySideMode = function(f, compareModel, differenceMode) {
	CompareMode.call(this, f, CompareConst.mode.SIDEBYSIDE, compareModel, true, "DiffSideBySideAnnotationsHidden");
	ListenerUtil.mixin(this, f);

	this.differenceCompareMode = differenceMode;

	/**
	 * Is the difference compare panel shown?
	 * @private
	 */
	this._showDifference = true;

	// is the mode already listening to the graphics model
	this._isListeningToGraphics = false;

	/**
	 * How are the panels layouted
	 * @private
	 */
	this._panelsLayoutMode = ViewportConstants.PanelsLayoutMode.AUTOMATIC;

	// load the user preference
	this.updateSettingsFromUserPreference(this.f.getOption("PREFERENCES_DIFF_SPLITMODE"));

	this.currentFilePanel = null;
	this.differenceFilePanel = null;
	this.differenceModePanel = null;

	this.viewport = this.f.getModule("viewport");

	f.getCurrentUser().on(
		"options.changed.PREFERENCES_DIFF_SPLITMODE",
		_.bind(function(oldValue, newValue) {
			this.updateSettingsFromUserPreference(newValue);
		}, this)
	);
};

CompareSideBySideMode.prototype = Object.create(CompareMode.prototype);
CompareSideBySideMode.prototype.constructor = CompareSideBySideMode;

/**
 * Start/stop show compare mode
 * @param {Boolean} toggle
 * @override
 */
CompareSideBySideMode.prototype.toggle = function(toggle, diffMode) {
	if (!this.isEnabled() && toggle) {
		CompareMode.prototype.toggle.call(this, toggle);

		if (diffMode === undefined) {
			diffMode = this.differenceCompareMode.getDifferenceMode();
			diffMode = parseInt(_.invert(CompareConst.sideBySideToDiffMode)[diffMode]);
		}

		// when enabled, set the diff mode and add the needed panels
		this._compareModel.setDiffMode(diffMode);

		this.f.getDocumentController().whenParametersAvailable(function() {
			if (this.isEnabled()) {
				if (!this.graphicsModule) {
					this.graphicsModule = this.f.getModule("graphics");
				}

				// set panel layout mode
				this.viewport.setPanelsLayoutMode(this._panelsLayoutMode);

				// get a reference to the default panel
				this.currentFilePanel = this.viewport.getPanelModel("default");
				this.currentFilePanel.toggleLayer("referenceCad", false);
				this.currentFilePanel.toggleLayer("currentDiffCad", false);
				this.currentFilePanel.toggleLayer("referenceDiffCad", false);
				this.currentFilePanel.toggleLayer("SVGDifferenceLayer", false);
				this.currentFilePanel.toggleLayer("Graphics", this.f.getActivePage().hasGraphics());

				// add difference file panel
				var disabledLayers = ["Cad", "currentDiffCad", "referenceDiffCad", "SVGLayer", "Video"];
				if (
					!this.f
						.getDocumentController()
						.getDifferencePage()
						.hasGraphics()
				) {
					disabledLayers.push("Graphics");
				}
				this.differenceFilePanel = this.viewport.addPanel(
					"compare_difference",
					this._getAllDifferenceFileOverwriteOptions(),
					disabledLayers
				);
				// show difference compare mode panel if needed && available
				this.toggleDifferencePanel(this.isShowingDifference());

				// update the overwrite options for all panels
				this.updateOverwriteOptions();

				// make sure to react to channel changes to update the overwrite options
				this.f.whenModuleAvailable(
					"graphics",
					function(module) {
						this.graphicsModule = module;

						if (this.isEnabled() && !this._isListeningToGraphics) {
							this._registerListener(
								"model.changed.channels",
								_.debounce(_.bind(this.updateOverwriteOptions, this), 40),
								undefined,
								this.graphicsModule
							);
							this._registerListener(
								"model.changed.channel.**",
								_.debounce(_.bind(this.updateOverwriteOptions, this), 40),
								undefined,
								this.graphicsModule
							);
							this._isListeningToGraphics = true;
							this.updateOverwriteOptions();
						}
					},
					this
				);

				this._registerListener("changed.**", this.updateOverwriteOptions, this, this.differenceCompareMode);
				this._registerListener("annotationModel.annotation.temporarilyDisableChanged", this.updateOverwriteOptions, this);
				this._registerListener("annotationModel.annotation.disabledReasonChanged", this.updateOverwriteOptions, this);
				this._registerListener("doc.page.active", this.updatePanelHeaders, this);
				this._registerListener("doc.page.difference", this.updatePanelHeaders, this);
				this._registerListener("doc.page.active", this._updateLayers.bind(this, this.currentFilePanel), this);
				this._registerListener("doc.page.difference", this._updateLayers.bind(this, this.differenceFilePanel), this);
				this.updatePanelHeaders();
			}
		}, this);
	} else if (this.isEnabled() && !toggle) {
		CompareMode.prototype.toggle.call(this, toggle);

		// remove panels
		this.viewport.removePanel("compare_difference");
		this.viewport.removePanel("compare_differenceMode");

		// the side by side compare is disabled, restore everything and remove the previously created panels
		this._compareModel.setDiffMode(CompareConst.diffMode.UNDEFINED);

		if (this.currentFilePanel) {
			// reset overwrite options
			this.currentFilePanel.deleteOverwriteOption("dmode");
			this.currentFilePanel.deleteOverwriteOption("colorBITWISE");
			this.currentFilePanel.toggleLayer("Annotations", true);
			this.currentFilePanel.toggleLayer("AnnotationPopups", true);
			this.currentFilePanel.toggleLayer("referenceCad", true);
			this.currentFilePanel.toggleLayer("currentDiffCad", true);
			this.currentFilePanel.toggleLayer("referenceDiffCad", true);
			this.currentFilePanel.toggleLayer("SVGDifferenceLayer", true);
			this.currentFilePanel.toggleLayer("Graphics", true);

			this.currentFilePanel.setHeaderText(null);
		}

		// reset references
		this.currentFilePanel = null;
		this.differenceFilePanel = null;
		this.differenceModePanel = null;

		// set panel layout to automatic
		this.viewport.setPanelsLayoutMode(ViewportConstants.PanelsLayoutMode.AUTOMATIC);

		// unregister all listeners
		this._unregisterAllListeners();
		this._isListeningToGraphics = false;
	}
};

CompareSideBySideMode.prototype._updateLayers = function(panel, oldPage, newPage) {
	panel.toggleLayer("Graphics", newPage && newPage.hasGraphics());
};

/**
 * Is the side by side mode showing the difference compare mode panel
 * @returns {Boolean}
 */
CompareSideBySideMode.prototype.isShowingDifference = function() {
	return this._showDifference && this._canFilesShowDifference();
};

/**
 * Can the current and reference file show a difference?
 */
CompareSideBySideMode.prototype._canFilesShowDifference = function() {
	return this.differenceCompareMode.isAvailable();
};

/**
 * Show/hide the difference compare mode panel
 * @param {Boolean} toggle
 */
CompareSideBySideMode.prototype.setShowingDifference = function(toggle) {
	var oldValue = this._showDifference;
	this._showDifference = toggle && this._canFilesShowDifference();

	if (this.isEnabled() && oldValue !== toggle) {
		// if the mode is enabled, show/hide the panel if needed
		this.toggleDifferencePanel(toggle);
	}

	this.storeUserPreference();
	this._fireChange(false, ["showDifference"], oldValue, toggle);
};

/**
 * Add/remove the difference compare mode panel
 * @param {Boolean} toggle
 */
CompareSideBySideMode.prototype.toggleDifferencePanel = function(toggle) {
	if (toggle) {
		this.differenceModePanel = this.viewport.addPanel("compare_differenceMode", this._getAllDifferenceModeOverwriteOptions(), [
			"Annotations",
			"AnnotationPopups",
			"Cad",
			"referenceCad"
		]);
		this.f.loadBundle(
			"viewer",
			function() {
				this.differenceModePanel.setHeaderText(this.f.t("CaptionCompared"));
			},
			this
		);
	} else {
		this.viewport.removePanel("compare_differenceMode");
		this.differenceModePanel = null;
	}

	// update the overwrite options
	this.updateOverwriteOptions();
};

/**
 * Configure the overwrite options for all created panels
 */
CompareSideBySideMode.prototype.updateOverwriteOptions = function() {
	if (this.isEnabled()) {
		if (this.graphicsModule === undefined) {
			this.graphicsModule = this.f.getModule("graphics");
		}
		if (this.annotationModule === undefined) {
			this.annotationModule = this.f.getModule("annotation");
		}

		this.currentFilePanel.setOverwriteOptions(this._getCurrentFileOverwriteOptions());

		this.differenceFilePanel.setOverwriteOptions(this._getAllDifferenceFileOverwriteOptions());
		if (this.differenceModePanel !== null) {
			this.differenceModePanel.setOverwriteOptions(this._getAllDifferenceModeOverwriteOptions());
		}

		if (this.annotationModule !== undefined) {
			this.currentFilePanel.toggleLayer("Annotations", this.annotationModule.isTemporarilyCurrentPageAnnotationsShown());
			this.currentFilePanel.toggleLayer("AnnotationPopups", this.annotationModule.isTemporarilyCurrentPageAnnotationsShown());
			this.differenceFilePanel.toggleLayer("Annotations", this.annotationModule.isTemporarilyDifferencePageAnnotationsShown());
			this.differenceFilePanel.toggleLayer("AnnotationPopups", this.annotationModule.isTemporarilyDifferencePageAnnotationsShown());
		}
	}
};

/**
 * Generate the overwrite options for the current file panel
 * @returns {Object}
 */
CompareSideBySideMode.prototype._getCurrentFileOverwriteOptions = function() {
	throw "Not Implemented";
};

/**
 * Generate the overwrite options for the difference file panel
 * @returns {Object}
 */
CompareSideBySideMode.prototype._getAllDifferenceFileOverwriteOptions = function() {
	var options = this._getDifferenceFileOverwriteOptions();
	options.isDefaultView = false;
	return options;
};

/**
 * Generate the overwrite options for the difference file panel, specifically for the implementation of this abstract class
 * @returns {Object}
 */
CompareSideBySideMode.prototype._getDifferenceFileOverwriteOptions = function() {
	throw "Not Implemented";
};

/**
 * Generate the overwrite options for the difference file panel
 * @returns {Object}
 */
CompareSideBySideMode.prototype._getAllDifferenceModeOverwriteOptions = function() {
	var options = this._getDifferenceModeOverwriteOptions();
	options.isDefaultView = false;
	return options;
};

/**
 * Generate the overwrite options for the difference compare mode panel
 * @returns {Object}
 */
CompareSideBySideMode.prototype._getDifferenceModeOverwriteOptions = function() {
	throw "Not Implemented";
};

/**
 * Restore the settings of the compare mode from the user preferences
 * @param {Number} value - some magical bitmasked number :) (see below)
 */
CompareSideBySideMode.prototype.updateSettingsFromUserPreference = function(value) {
	// the option is build up as follows"
	// value & 3 --> gives the panel layout mode (0=auto, 1=vertical, 2= horizontal)
	// value & 4 --> determines whether the differencecompare is shown or not

	this.setPanelsLayoutMode(+value & 3); // the + makes it a number
	this.setShowingDifference((value & 4) === 4);
};

/**
 * Save the user preference
 */
CompareSideBySideMode.prototype.storeUserPreference = function() {
	var value = this.getPanelsLayoutMode();
	if (this.isShowingDifference()) {
		value += 4;
	}
	this.f.setOption("PREFERENCES_DIFF_SPLITMODE", value);
};

/**
 * Change the panels layoutmode
 * @param {ViewportConstants.PanelsLayoutMode} mode
 */
CompareSideBySideMode.prototype.setPanelsLayoutMode = function(mode) {
	var oldValue = this._panelsLayoutMode;
	this._panelsLayoutMode = mode;
	if (this.isEnabled()) {
		// if side by side is enabled enabled, also update the viewport
		this.viewport.setPanelsLayoutMode(mode);
	}
	this.storeUserPreference();
	this._fireChange(false, ["panelLayout"], oldValue, this._panelsLayoutMode);
};

/**
 * Get the currently configure panels layout mode for side by side compare
 * @returns {ViewportConstants.PanelsLayoutMode}
 */
CompareSideBySideMode.prototype.getPanelsLayoutMode = function() {
	return this._panelsLayoutMode;
};

/**
 * Update the header text for the panels
 */
CompareSideBySideMode.prototype.updatePanelHeaders = function() {
	this.f.loadBundle(
		"viewer",
		function() {
			if (this.isEnabled()) {
				var current = "";
				var difference = "";

				var docController = this.f.getDocumentController();
				var activePage = docController.getActivePage();
				var activeDocVersion = activePage.getDocumentVersion();
				var differencePage = docController.getDifferencePage();
				var differenceDocVersion = differencePage.getDocumentVersion();

				if (activeDocVersion.getDocument() !== differenceDocVersion.getDocument()) {
					// we need to add document info too
					current += activeDocVersion.getDocument().get("docname") + ", ";
					difference += differenceDocVersion.getDocument().get("docname") + ", ";
				}

				current += this.f.t("WCHistVersion") + " " + activeDocVersion.getVersionId();
				difference += this.f.t("WCHistVersion") + " " + differenceDocVersion.getVersionId();

				current += " - (" + this.f.t("CaptionCurrent") + ")";
				difference += " - (" + this.f.t("CaptionPrevious") + ")";

				this.currentFilePanel.setHeaderText(current);
				this.differenceFilePanel.setHeaderText(difference);
			}
		},
		this
	);
};

/**
 * Set the dmode parameter correctly
 * @param {Object} c intercept context
 */
CompareSideBySideMode.prototype.interceptFragmentURL = function(c) {
	throw "Not Implemented";
};

export default CompareSideBySideMode;
