import ToolbarComponent from "@esko/toolbar/ToolbarComponent";
import ButtonGroup from "@esko/toolbar/ButtonGroup";
import DropdownToolbarButton from "@esko/web-viewer/layout/component/DropdownToolbarButton";
import ViewerToolbarButton from "@esko/web-viewer/layout/component/ViewerToolbarButton";
import LayoutConstants from "@esko/web-viewer/layout/LayoutConstants";
import CompareConstants from "@esko/web-viewer/compare/CompareConstants";
import CompareDifferenceSubToolbarComponent from "@esko/web-viewer/compare/CompareDifferenceSubToolbarComponent";
import CompareSideBySideSubToolbarComponent from "@esko/web-viewer/compare/CompareSideBySideSubToolbarComponent";
import _ from "underscore";
/**
 * create a new compare toolbar component
 * @memberof CompareModule
 * @exports CompareToolbarComponent
 * @class The toolbar component for the compare module
 * @constructor
 * @param {Core.ViewerFacade} f
 * @param {CompareModule.CompareModel} compareModel
 */
var CompareToolbarComponent = function(f, compareModule) {
	ToolbarComponent.call(this, LayoutConstants.ToolbarComponent.COMPARE, f);
	this.f = f;
	this.compareModule = compareModule;
	this.compareModel = compareModule.getCompareModel();
	this.buttons = {};
	this.modesGroup = new ButtonGroup("Compare", []);
	this.modesGroup.setTitle(this.f.i18n, "viewer", "CompareToolsTitle");
	this.addButtonGroup(this.modesGroup);

	this.alignGroup = new ButtonGroup("Align", []);
	this.alignGroup.setTitle(this.f.i18n, "viewer", "CompareToolsTitle2");
	this.addButtonGroup(this.alignGroup);
	this.addLockButton();
	this.addAlignButton();

	var updateButtonsActive = _.bind(function(oldMode, newMode) {
		if (oldMode && this.buttons[oldMode.getName()] !== undefined) {
			this.buttons[oldMode.getName()].setActive(false);
		}

		if (newMode && this.buttons[newMode.getName()] !== undefined) {
			this.buttons[newMode.getName()].setActive(true);
		}
	}, this);
	this.compareModel.on("changed.mode", updateButtonsActive);

	this.addButtonForMode([CompareConstants.mode.SHOW_CURRENT, CompareConstants.mode.CURRENT3D], "icon-WCR_Compare-Current", "DiffCurrentTT");
	this.addButtonForMode(
		[CompareConstants.mode.SHOW_DIFFERENCE, CompareConstants.mode.REFERENCE3D],
		"icon-WCR_Compare-Reference",
		"DiffPrevTT"
	);
	this.addButtonForMode(
		[CompareConstants.mode.BLINKING, CompareConstants.mode.BLINKING3D],
		"icon-WCR_Compare-Blinking",
		"DiffBothBlinkingTT"
	);
	this.addButtonForMode(CompareConstants.mode.SHOWBOTHONTOP, "icon-WCR_Compare", "DiffShowBothOnTopTT");

	var differenceCompareMode = this.compareModel.getModeByName(CompareConstants.mode.DIFFERENCE);
	if (differenceCompareMode !== undefined) {
		this.addDifferenceModeButton(differenceCompareMode);

		var updateBadges = _.debounce(_.bind(this._updateBadges, this), 0);

		differenceCompareMode.on("changed.threshold", updateBadges);
		this.compareModel.on("changed.mode", updateBadges);
		this.f.getDocumentController().whenActivePageChanged(updateBadges);

		var sideBySideCompareMode = this.compareModel.getModeByName(CompareConstants.mode.SIDEBYSIDE);
		if (sideBySideCompareMode !== undefined) {
			this.addSideBySideModeButton(differenceCompareMode, sideBySideCompareMode);
			sideBySideCompareMode.on("changed.showDifference", updateBadges);
		}

		this._updateBadges();
	}

	this.f.getDocumentController().whenActiveOrDifferencePageChanged(this._updateButtonVisibility.bind(this));
	this.f.on("module.compare.changed.mode", this._updateButtonVisibility.bind(this));
	this.f.on("view3d.widget.loaded", this._updateButtonVisibility.bind(this));
	this.f.on(
		"changed.lock",
		_.bind(function(value) {
			var view3DModule = this.f.getModule("view3D");
			if (view3DModule) {
				this._compare3DLockBtn.setActive(!value);
			}
		}, this)
	);
	updateButtonsActive(undefined, this.compareModel.getActiveMode());
};

CompareToolbarComponent.prototype = Object.create(ToolbarComponent.prototype);
CompareToolbarComponent.prototype.constructor = CompareToolbarComponent;
CompareToolbarComponent.toolbarName = LayoutConstants.ToolbarComponent.COMPARE;

/**
 * Add a button to active/deactivate a mode
 * @param {String|CompareModule.CompareConstants.mode|Array.<String>} name - names of the compare mode
 * @param {String} icon - name of the icon
 * @param {String} tooltip - tooltip
 */
CompareToolbarComponent.prototype.addButtonForMode = function(names, icon, tooltip) {
	names = _.flatten([names]);
	var button = new ViewerToolbarButton(
		this.f,
		"compare_" + names[0],
		icon,
		true,
		_.partial(function(that) {
			if (that.compareModel.getActiveMode() && names.indexOf(that.compareModel.getActiveMode().getName()) !== -1) {
				// pressing the toolbar button when it's active, deactivates the current mode.
				that.compareModule.activateModeByName();
			} else {
				// if the button is pressed, check which compare modes are available for the active and difference pages.
				var page = this.f.getDocumentController().getActivePage();
				var diff = this.f.getDocumentController().getDifferencePage();
				var availableCompareModes;
				if (diff) {
					availableCompareModes = _.intersection(page.getAvailableCompareModes(diff), diff.getAvailableCompareModes(page));
				} else {
					availableCompareModes = that._getAvailableCompareModes(page);
				}

				// there are twoD and threeD diff mode, check if the current mode is a twoD mode or a twoD mode
				var compareMode = that.compareModel.getActiveMode();
				var isThreeDMode = compareMode && CompareConstants.modes3D[compareMode.getName()];

				// if there already was a 3D mode active, we would like to see if there is another 3D that can be activated by this button
				var name = _.find(names, function(name) {
					return (
						availableCompareModes.indexOf(name) !== -1 &&
						((isThreeDMode && CompareConstants.modes3D[name]) || (!isThreeDMode && CompareConstants.modes2D[name]))
					);
				});

				if (!name) {
					// if name wasn't set yet, just take the first available one
					name = _.find(names, function(name) {
						return availableCompareModes.indexOf(name) !== -1;
					});
				}
				that.compareModule.activateModeByName(name);
			}
		}, this)
	);

	button.setTooltip("viewer", tooltip);
	this.modesGroup.addButton(button);

	names.forEach(
		function(name) {
			this.buttons[name] = button;
		}.bind(this)
	);
};

/**
 * Add a button to enable/disable synchronization between current 3D document and difference 3D document
 */
CompareToolbarComponent.prototype.addLockButton = function() {
	var name = "compare3D_lock";
	this._compare3DLockBtn = new DropdownToolbarButton(
		this.f,
		name,
		"icon-WCR_Viewer-Perspectives-Linked",
		true,
		_.partial(function(that) {
			that.f.emit("changed.lock", this.isActive());
		}, this)
	);

	this._compare3DLockBtn.addOption({
		name: "compare3D_lock_low_performance",
		content: this.f.i18n.t("View3DLockCameraLive"),
		selected: this.f.getOption("PREFERENCES_VIEW3D_CAMERA_LOCK_LIVE") === 1,
		onchange: function() {
			var isCameraLockLowPerf = this.f.getOption("PREFERENCES_VIEW3D_CAMERA_LOCK_LIVE") === 1;
			this.f.setOption("PREFERENCES_VIEW3D_CAMERA_LOCK_LIVE", isCameraLockLowPerf ? 0 : 1);
		}.bind(this)
	});
	this._compare3DLockBtn.setTooltip("viewer", "LockMovementTT");
	this._compare3DLockBtn.setActive(true); // default is on
	this.alignGroup.addButton(this._compare3DLockBtn);
};

/**
 * Add a button to align current 3D document and difference 3D document
 */
CompareToolbarComponent.prototype.addAlignButton = function() {
	var name = "compare3D_sync";
	var button = new ViewerToolbarButton(
		this.f,
		name,
		"icon-WCR_Viewer-Perspectives-Reset",
		true,
		_.partial(function(that) {
			var view3DModule = that.f.getModule("view3D");
			if (view3DModule) {
				view3DModule.syncCurrentWithDifference();
			}
		}, this)
	);

	button.setTooltip("viewer", "AlignViewsTT");
	this.alignGroup.addButton(button);
};

/**
 * Updates the button visibility based on the available compare modes of the page
 */
CompareToolbarComponent.prototype._updateButtonVisibility = function() {
	var activePage = this.f.getDocumentController().getActivePage();

	var diff = this.f.getDocumentController().getDifferencePage();
	var availableCompareModes;
	if (diff) {
		availableCompareModes = _.intersection(activePage.getAvailableCompareModes(diff), diff.getAvailableCompareModes(activePage));
	} else {
		availableCompareModes = this._getAvailableCompareModes(activePage);
	}

	// Hide all buttons that are not available
	_.chain(this.buttons)
		.filter(function(button, key) {
			return availableCompareModes.indexOf(key) === -1;
		})
		.uniq()
		.invoke("hide");

	// show all the ones that are
	_.chain(this.buttons)
		.filter(function(button, key) {
			return availableCompareModes.indexOf(key) !== -1;
		})
		.uniq()
		.invoke("show");

	// show/hide Lock button
	var currentMode = this.compareModule.getActiveMode();
	if (currentMode && _.keys(CompareConstants.modes3D).indexOf(currentMode.getName()) != -1) {
		var view3DModule = this.f.getModule("view3D");
		if (view3DModule && view3DModule.canSyncWidgets()) {
			this.alignGroup.show();
		} else {
			this.alignGroup.hide();
		}
	} else {
		this.alignGroup.hide();
	}
};

CompareToolbarComponent.prototype._getAvailableCompareModes = function(activePage) {
	return _.chain(this.f.getDocumentController().getAllDocuments())
		.map(function(doc) {
			// get document versions
			return doc.getDocumentVersions();
		})
		.flatten()
		.map(function(version) {
			// map to pages
			return version.getAllPages();
		})
		.flatten()
		.without(activePage)
		.reduce(function(memo, page) {
			return _.union(memo, _.intersection(activePage.getAvailableCompareModes(page), page.getAvailableCompareModes(activePage)));
		}, [])
		.value();
};

/**
 * Difference compare mode button
 * @param differenceCompareMode
 */
CompareToolbarComponent.prototype.addDifferenceModeButton = function(differenceCompareMode) {
	this.addButtonForMode(CompareConstants.mode.DIFFERENCE, "icon-WCR_Compare-BothDiff", "compareDifferenceMode");

	var button = this.buttons[CompareConstants.mode.DIFFERENCE];
	button.addSubtoolbarComponent(new CompareDifferenceSubToolbarComponent(this.f, differenceCompareMode, button));

	this._differenceCompareMode = differenceCompareMode;
};

/**
 * Side by side compare mode button
 * @param {CompareModule.CompareDifferenceMode} differenceCompareMode
 * @param {CompareModule.CompareSideBySideMode} sideBySideCompareMode
 */
CompareToolbarComponent.prototype.addSideBySideModeButton = function(differenceCompareMode, sideBySideCompareMode) {
	this.addButtonForMode(
		[CompareConstants.mode.SIDEBYSIDE, CompareConstants.mode.SIDEBYSIDE3D],
		"icon-WCR_Compare2-SideBySide",
		"DiffBothSplitTT"
	);

	var button = this.buttons[CompareConstants.mode.SIDEBYSIDE];
	var subToolbar = new CompareSideBySideSubToolbarComponent(this.f, differenceCompareMode, sideBySideCompareMode, button);
	var onModeChanged = function() {
		var mode = this.f.getModule("compare") && this.f.getModule("compare").getActiveMode();
		if (mode && mode.getName() === CompareConstants.mode.SIDEBYSIDE) {
			button.addSubtoolbarComponent(subToolbar);
		} else {
			button.detachSubtoolbarComponent(subToolbar);
		}
	}.bind(this);
	this.f.on("module.compare.changed.mode", onModeChanged);
	onModeChanged();
	this._sideBySideCompareMode = sideBySideCompareMode;
};

/**
 * Update the Warning Badges for the Threshold for difference and side by side mode
 */

CompareToolbarComponent.prototype._updateBadges = function() {
	var button;

	if (this.buttons[CompareConstants.mode.DIFFERENCE] && this._differenceCompareMode) {
		// difference compare mode is available
		button = this.buttons[CompareConstants.mode.DIFFERENCE];
		if (this._differenceCompareMode.isEnabled() && this._differenceCompareMode.getThreshold() > 0 && !this.f.getActivePage().hasCad()) {
			// show warning
			button.setBadge("!", "warning");
			button.setTooltip("viewer", "compareDifferenceThresholdWarning");
		} else {
			button.setBadge();
			button.setTooltip("viewer", "compareDifferenceMode");
		}
	}

	if (this.buttons[CompareConstants.mode.SIDEBYSIDE]) {
		// side by side compare mode is available
		button = this.buttons[CompareConstants.mode.SIDEBYSIDE];

		if (
			this._sideBySideCompareMode.isEnabled() &&
			this._sideBySideCompareMode.isShowingDifference() &&
			this._differenceCompareMode.getThreshold() > 0 &&
			!this.f.getActivePage().hasCad()
		) {
			button.setBadge("!", "warning");
			button.setTooltip("viewer", "compareDifferenceThresholdWarning");
		} else {
			button.setBadge();
			button.setTooltip("viewer", "DiffBothSplitTT");
		}
	}
};

export default CompareToolbarComponent;
