import ToolbarComponent from "@esko/toolbar/ToolbarComponent";
import ButtonGroup from "@esko/toolbar/ButtonGroup";
import ViewerToolbarButton from "@esko/web-viewer/layout/component/ViewerToolbarButton";
import LayoutConstants from "@esko/web-viewer/layout/LayoutConstants";
import CompareConstants from "@esko/web-viewer/compare/CompareConstants";
import _ from "underscore";
/**
 * create a new compare toolbar component
 * @memberof CompareModule
 * @exports CompareTouchToolbarComponent
 * @class The toolbar component for the compare module
 * @constructor
 * @param {Core.ViewerFacade} f
 * @param {CompareModule.CompareModel} compareModel
 */
var CompareTouchToolbarComponent = function(f, compareModule) {
	ToolbarComponent.call(this, LayoutConstants.ToolbarComponent.COMPARE, f);
	this.f = f;
	this._compareModule = compareModule;
	this._compareModel = compareModule.getCompareModel();
	this._buttons = {};
	this._modesGroup = new ButtonGroup("Compare", []);
	this._modesGroup.element.removeClass("btn-group");
	this._modesGroup.element.addClass("compareModes");
	this._modesGroup.setTitle(this.f.i18n, "viewer", "CompareToolsTitle");
	this.addButtonGroup(this._modesGroup);

	var differenceCompareMode = this._compareModel.getModeByName(CompareConstants.mode.DIFFERENCE);
	this.addDifferenceModeButton(differenceCompareMode);

	this.addButtonForMode(
		[CompareConstants.mode.BLINKING, CompareConstants.mode.BLINKING3D],
		"icon-WCR_Compare-Blinking",
		"DiffBothBlinkingTT"
	);

	var sideBySideCompareMode = this._compareModel.getModeByName(CompareConstants.mode.SIDEBYSIDE);
	this.addSideBySideModeButton(differenceCompareMode, sideBySideCompareMode);

	this.addButtonForMode(CompareConstants.mode.SHOWBOTHONTOP, "icon-WCR_Compare-Both", "DiffShowBothOnTopTT");

	this.addButtonForMode([CompareConstants.mode.SHOW_CURRENT, CompareConstants.mode.CURRENT3D], "icon-WCR_Compare-Current", "DiffCurrentTT");
	this.addButtonForMode(
		[CompareConstants.mode.SHOW_DIFFERENCE, CompareConstants.mode.REFERENCE3D],
		"icon-WCR_Compare-Reference",
		"DiffPrevTT"
	);

	this._compareModel.on("changed.mode", _.bind(this._updateActiveStatus, this));
	this._updateActiveStatus(undefined, this._compareModule.getActiveMode());
	this.f.getDocumentController().whenActiveOrDifferencePageChanged(this._updateButtonVisibility.bind(this));
};

CompareTouchToolbarComponent.prototype = Object.create(ToolbarComponent.prototype);
CompareTouchToolbarComponent.prototype.constructor = CompareTouchToolbarComponent;
CompareTouchToolbarComponent.toolbarName = LayoutConstants.ToolbarComponent.COMPARE;

/**
 * Update the active flag for the compare mode buttons
 * @param {String} oldMode
 * @param {String} newMode
 */
CompareTouchToolbarComponent.prototype._updateActiveStatus = function(oldMode, newMode) {
	if (oldMode && this._buttons[oldMode.getName()] !== undefined) {
		this._buttons[oldMode.getName()].setActive(false);
	}

	if (newMode && this._buttons[newMode.getName()] !== undefined) {
		this._buttons[newMode.getName()].setActive(true);
	}
};

/**
 * Add a button to active/deactivate a mode
 * @param {String|CompareModule.CompareConstants.mode} name - name of the compare mode
 * @param {String} icon - name of the icon
 * @param {String} tooltip - tooltip
 */
CompareTouchToolbarComponent.prototype.addButtonForMode = function(names, icon, tooltip) {
	names = _.flatten([names]);
	var button = new ViewerToolbarButton(
		this.f,
		"compare_" + names[0],
		icon,
		true,
		_.partial(function(that) {
			if (this.isActive()) {
				// pressing the toolbar button when it's active, deactivates the current mode.
				that._compareModule.activateModeByName();
			} else {
				// if the button is pressed, check which compare modes are available for the active and difference pages.
				var page = this.f.getDocumentController().getActivePage();
				var diff = this.f.getDocumentController().getDifferencePage();
				var availableCompareModes;
				if (diff) {
					availableCompareModes = _.intersection(page.getAvailableCompareModes(diff), diff.getAvailableCompareModes(page));
				} else {
					availableCompareModes = that._getAvailableCompareModes(page);
				}

				// there are twoD and threeD diff mode, check if the current mode is a twoD mode or a twoD mode
				var compareMode = that._compareModel.getActiveMode();
				var isThreeDMode = compareMode && CompareConstants.modes3D[compareMode.getName()];

				// if there already was a 3D mode active, we would like to see if there is another 3D that can be activated by this button
				var name = _.find(names, function(name) {
					return (
						availableCompareModes.indexOf(name) !== -1 &&
						((isThreeDMode && CompareConstants.modes3D[name]) || (!isThreeDMode && CompareConstants.modes2D[name]))
					);
				});

				if (!name) {
					// if name wasn't set yet, just take the first available one
					name = _.find(names, function(name) {
						return availableCompareModes.indexOf(name) !== -1;
					});
				}
				that._compareModule.activateModeByName(name);
			}
		}, this)
	);

	button.setTooltip("viewer", tooltip);
	this._modesGroup.addButton(button);

	names.forEach(
		function(name) {
			this._buttons[name] = button;
		}.bind(this)
	);
};

/**
 * Updates the button visibility based on the available compare modes of the page
 */
CompareTouchToolbarComponent.prototype._updateButtonVisibility = function() {
	var activePage = this.f.getDocumentController().getActivePage();

	var differencePage = this.f.getDocumentController().getDifferencePage();
	var activePageCompareModes = activePage.getAvailableCompareModes();
	var availableCompareModes = [];
	if (differencePage) {
		// if there is a difference page, only show the compare modes available between the two
		availableCompareModes = _.intersection(activePageCompareModes, differencePage.getAvailableCompareModes());
	} else {
		// else we only show compare modes that make sense for the loaded documents
		availableCompareModes = this._getAvailableCompareModes(activePage);
	}

	// Hide all buttons that are not available
	_.chain(this._buttons)
		.filter(function(button, key) {
			return availableCompareModes.indexOf(key) === -1;
		})
		.uniq()
		.invoke("hide");

	// show all the ones that are
	_.chain(this._buttons)
		.filter(function(button, key) {
			return availableCompareModes.indexOf(key) !== -1;
		})
		.uniq()
		.invoke("show");
};

CompareTouchToolbarComponent.prototype._getAvailableCompareModes = function(page) {
	var activePageCompareModes = page.getAvailableCompareModes();
	return _.chain(this.f.getDocumentController().getAllDocuments())
		.map(function(doc) {
			// get document versions
			return doc.getDocumentVersions();
		})
		.flatten()
		.map(function(version) {
			// map to pages
			return version.getAllPages();
		})
		.flatten()
		.without(page)
		.reduce(function(memo, page) {
			return _.union(memo, _.intersection(activePageCompareModes, page.getAvailableCompareModes(page)));
		}, [])
		.value();
};

/**
 * Difference compare mode button
 * @param differenceCompareMode
 */
CompareTouchToolbarComponent.prototype.addDifferenceModeButton = function(differenceCompareMode) {
	this.addButtonForMode(CompareConstants.mode.DIFFERENCE, "icon-WCR_Compare-BothDiff", "compareDifferenceMode");
};

/**
 * Side by side compare mode button
 * @param {CompareModule.CompareDifferenceMode} differenceCompareMode
 * @param {CompareModule.CompareSideBySideMode} sideBySideCompareMode
 */
CompareTouchToolbarComponent.prototype.addSideBySideModeButton = function(differenceCompareMode, sideBySideCompareMode) {
	this.addButtonForMode(
		[CompareConstants.mode.SIDEBYSIDE, CompareConstants.mode.SIDEBYSIDE3D],
		"icon-WCR_Compare-SideBySide",
		"DiffBothSplitTT"
	);
};
export default CompareTouchToolbarComponent;
