import AbstractWidget from "@esko/web-viewer/layout/AbstractWidget";
import ViewerToolbarButton from "@esko/web-viewer/layout/component/ViewerToolbarButton";
import DropdownToolbarButton from "@esko/web-viewer/layout/component/DropdownToolbarButton";
import MiniButton from "@esko/web-viewer/layout/component/MiniButton";
import ModalPopup from "@esko/web-viewer/layout/component/ModalPopup";
import UserUnitInputField from "@esko/web-viewer/layout/component/UserUnitInputField";
import CompareDifferenceSubToolbarComponent from "@esko/web-viewer/compare/CompareDifferenceSubToolbarComponent";
import CompareSideBySideSubToolbarComponent from "@esko/web-viewer/compare/CompareSideBySideSubToolbarComponent";
import LayoutConstants from "@esko/web-viewer/layout/LayoutConstants";
import CompareConstants from "@esko/web-viewer/compare/CompareConstants";
import $ from "jquery";
import _ from "underscore";
/**
 * @exports CompareWidget
 * @memberof Compare
 * @augments Layout.AbstractWidget
 * @class The CompareWidget displays the currently active page and the makes it possible to select the active and difference page from a dropdown menu.
 * @constructor
 *
 * @param {Core.ViewerFacade} f
 * @param {CompareModule} compareModule
 * @param {CompareModel} compareModel
 * @param {Layout.Component.DocumentSelection} documentSelection - contains the document selectors
 */
var CompareWidget = function CompareWidget(f, compareModule, compareModel, documentSelection) {
	AbstractWidget.call(this, f, LayoutConstants.Widget.COMPARE);

	/**
	 * @member {CompareModel} compareModel
	 */
	this._compareModel = compareModel;

	/**
	 * @member {CompareModule} compareModule
	 */
	this._compareModule = compareModule;

	/**
	 * @member {Core.ViewerFacade} the viewer facade
	 */
	this.f = f;

	/**
	 * @member {Layout.Component.DocumentSelection}
	 */
	this._documentSelection = documentSelection;

	this._compareOptionsInitialized = false;
	this._cadOverlayOptionsInitialized = false;

	/**
	 * Containers for all sections
	 **/
	this._extraControls = $("<div/>").addClass("extraControls displayNone");
	this._documentSelectionDiv = $("<div/>").addClass("documentSelection");
	this._compareOptions = $("<div/>").addClass("compareOptions displayNone");
	this._cadOverlay = $("<div>").addClass("cadOverlay displayNone");

	this._docTraversal = this.createDocTraversalElem();

	/**
	 * Reset compare pages
	 */
	this._resetLinkedComparePagesButton = new MiniButton(
		this.f,
		"resetLinkedComparePages",
		"icon-WCR_Revert",
		true,
		_.bind(function() {
			this.f.i18n.loadBundle(
				"viewer",
				function() {
					var modalPopup = new ModalPopup(this.f, "resetLinkedComparePagesPopup", this.f.getContext());
					modalPopup.setRemoveFromDomAfterUsage(true);

					modalPopup.setOkAction(
						_.bind(function() {
							this.f.getActivePage().resetLinkedComparePage();
							this._updateResetLinkedComparePagesBtn();
							modalPopup.hide();
						}, this)
					);

					modalPopup.addButton(
						new ViewerToolbarButton(
							this.f,
							"allPages",
							$("<span/>").append(f.i18n.t("AllPages")),
							true,
							_.bind(function() {
								_.each(
									this.f
										.getActivePage()
										.getDocumentVersion()
										.getAllPages(),
									function(page) {
										page.resetLinkedComparePage();
									}
								);
								this._updateResetLinkedComparePagesBtn();
								modalPopup.hide();
							}, this)
						)
					);

					// Header + Content
					modalPopup.setHeader(
						$("<span/>")
							.addClass("popupTitle")
							.text(this.f.i18n.t("ResetLinkedComparePagesTT"))
					);
					modalPopup.setContent(
						$("<div/>")
							.addClass("unsavedItemsContent")
							.text(this.f.t("ResetLinkedComparePages"))
					);
					modalPopup.setOkBtnText(this.f.i18n.t("PageOnly"));

					modalPopup.show();
				},
				this
			);
		}, this),
		"icon-WCR_Revert fontIconDisabled"
	);

	this._resetLinkedComparePagesButton.setTooltip("viewer", "ResetLinkedComparePagesTT");

	this._diffDocPageElement = $("<div/>");

	this.f.i18n.loadBundle(
		"viewer",
		_.bind(function() {
			this._documentSelectionDiv.append(
				$("<div />")
					.addClass("currentTitleBox")
					.append(
						$("<h4/>")
							.addClass("currentTitle")
							.append(this.f.t("DocumentListCurrentLabel"))
					)
					.append(this._docTraversal)
			);
			this._documentSelectionDiv.append(this._documentSelection.getActiveDocumentSelector().element);
			this._diffDocPageElement
				.append($("<h4/>").append(this.f.t("DocumentListDiffLabel") + " <br/>"))
				.append(this._documentSelection.getDifferenceDocumentSelector().element);
			this._documentSelectionDiv.append(this._diffDocPageElement);

			if (this.f.getActivePage().canBeComparedTo()) {
				this._diffDocPageElement.addClass("displayNone");
			}

			// set the widget title
			this.setTitle(this.f.t("DocumentListTitle"));
		}, this)
	);

	if (this.f.isMobile()) {
		// on mobile we want to update the title whenever the compare mode changes
		this.f.loadBundle(
			"viewer",
			function() {
				this._compareModule.on("changed.mode", _.bind(this._compareModeChanged, this));
				this._compareModeChanged();
			},
			this
		);
	}

	this.f.on("layout.*.changed.layoutMode", this.updateCompareWidget.bind(this));

	this.setIcon("icon icon-WCR_Document");

	this.f.getDocumentController().whenActivePageChanged(this.updateDocTraversalElem.bind(this));
};

CompareWidget.prototype = Object.create(AbstractWidget.prototype);
CompareWidget.prototype.constructor = CompareWidget;
CompareWidget.widgetName = LayoutConstants.Widget.COMPARE;

/**
 * Update the title to the active compare mode
 * And show the controls for difference or side by side compare
 *
 * NOTE: this method is only used on mobile
 */
CompareWidget.prototype._compareModeChanged = function() {
	var activeMode = this._compareModule.getActiveMode();
	var title;
	if (activeMode) {
		var activeModeName = activeMode.getName();
		title = this.f.t(CompareConstants.modeNameToLanguageKey[activeModeName]);

		if (this.f.isMobile()) {
			if (activeModeName === CompareConstants.mode.DIFFERENCE || activeModeName === CompareConstants.mode.SIDEBYSIDE) {
				var differenceCompareMode = this._compareModel.getModeByName(CompareConstants.mode.DIFFERENCE);
				if (activeModeName === CompareConstants.mode.DIFFERENCE) {
					if (!this._differenceCompareControls) {
						this._differenceCompareControls = new CompareDifferenceSubToolbarComponent(this.f, differenceCompareMode);
						this._extraControls.prepend(this._differenceCompareControls.element);
					}

					this._differenceCompareControls.element.removeClass("displayNone");
					if (this._sideBySideCompareControls) {
						this._sideBySideCompareControls.element.addClass("displayNone");
					}
				}

				if (activeModeName === CompareConstants.mode.SIDEBYSIDE) {
					if (!this._sideBySideCompareControls) {
						var sideBySideCompareMode = this._compareModel.getModeByName(CompareConstants.mode.SIDEBYSIDE);
						this._sideBySideCompareControls = new CompareSideBySideSubToolbarComponent(
							this.f,
							differenceCompareMode,
							sideBySideCompareMode
						);
						this._extraControls.prepend(this._sideBySideCompareControls.element);
					}

					this._sideBySideCompareControls.element.removeClass("displayNone");
					if (this._differenceCompareControls) {
						this._differenceCompareControls.element.addClass("displayNone");
					}
				}

				// show the compare mode extra controls
				this._extraControls.removeClass("displayNone");
			} else {
				// hide the extra controls
				this._extraControls.addClass("displayNone");
			}
			this._fireResized();
		}
	} else {
		// hide the extra controls
		this._extraControls.addClass("displayNone");
	}

	if (this.f.isMobile()) {
		// update the title for mobile
		if (!title) {
			title = this.f.t("DocumentListTitle");
		}

		this.setTitle(title);
	}
};

/**
 * initializes all the options for when both a current and difference page are selected, this are options for compare
 */
CompareWidget.prototype._initCompareOptions = function() {
	if (!this._compareOptionsInitialized) {
		this._compareOptionsInitialized = true;

		if (!this.f.isMobile()) {
			this._channelMatchingBtn = new ViewerToolbarButton(
				this.f,
				"channelMatching",
				"icon-WCR_Channel-Matching",
				false,
				_.bind(function() {
					var graphicsModule = this.f.getModule("graphics");
					if (graphicsModule !== undefined) {
						graphicsModule.showChannelMatchingPopup();
					}
				}, this)
			);
		}

		// auto align button
		this._autoAlign = new ViewerToolbarButton(
			this.f,
			"autoAlign",
			"icon-WCR_Align-Auto",
			true,
			_.bind(function() {
				this._emit("autoAlign");
			}, this)
		);
		this._autoAlign.element.addClass("autoAlign");

		//anchor align button
		this._anchorAlign = new ViewerToolbarButton(
			this.f,
			"anchorAlign",
			"icon-WCR_Align-Point",
			true,
			_.bind(function() {
				this._compareModule.startAlignmentWizard();
			}, this)
		);
		this._anchorAlign.setTooltip("viewer", "AnchorAlignWizardTitle");

		// corner align button
		this._cornerAlign = new DropdownToolbarButton(this.f, "cornerAlign", "cornerAlign", true);

		// rotate reference button
		this._rotateReference = new ViewerToolbarButton(
			this.f,
			"rotateReference",
			"icon-WCR_Orientation-Rotate1",
			true,
			_.bind(function() {
				this._emit("rotateReference");
				this._updateRotateBtn();
			}, this)
		);

		// flip reference button
		this._flipReference = new ViewerToolbarButton(
			this.f,
			"flipReference",
			"icon-WCR_Orientation-Mirror",
			true,
			_.bind(function() {
				this._emit("flipReference");
				this._updateFlipBtn();
			}, this)
		);

		// negate reference button
		this._negateReference = new ViewerToolbarButton(
			this.f,
			"negateReference",
			"icon-WCR_Invert",
			true,
			_.bind(function() {
				this._emit("negateReference");
				this._updateNegateBtn();
			}, this)
		);

		// trimbox align button
		this._trimBoxAlign = new ViewerToolbarButton(
			this.f,
			"trimBoxAlign",
			"icon-WCR_Align-Trimbox",
			true,
			_.bind(function() {
				this._emit("trimBoxAlign");
				this._trimBoxAlign.setActive(this._compareModel.isTrimBoxAlign());
			}, this)
		);
		this._trimBoxAlign.element.addClass("toggleBtn");

		this._offsetX = new UserUnitInputField(
			this.f,
			0,
			"x",
			_.bind(function(oldValue, newValue) {
				this._emit("offsetXChanged", newValue);
			}, this)
		);

		this._offsetY = new UserUnitInputField(
			this.f,
			0,
			"y",
			_.bind(function(oldValue, newValue) {
				this._emit("offsetYChanged", newValue);
			}, this)
		);

		this.f.i18n.loadBundle(
			"viewer",
			_.bind(function() {
				// initialize corner align actions
				this._cornerAlign.addAction({
					name: "topLeftAlign",
					content: this.f.t("AlignTopLeft"),
					icon: "icon-WCR_Align-TopLeft",
					action: _.bind(function() {
						this._emit("doCornerAlign", CompareConstants.cornerAutoAlign.TOPLEFT);
						this._cornerAlign.setIcon("icon-WCR_Align-TopLeft");
					}, this),
					enabled: true
				});

				this._cornerAlign.addAction({
					name: "topCenterAlign",
					content: this.f.t("AlignTopCenter"),
					icon: "icon-WCR_Align-TopCenter",
					action: _.bind(function() {
						this._emit("doCornerAlign", CompareConstants.cornerAutoAlign.TOPCENTER);
						this._cornerAlign.setIcon("icon-WCR_Align-TopCenter");
					}, this),
					enabled: true
				});

				this._cornerAlign.addAction({
					name: "topRightAlign",
					content: this.f.t("AlignTopRight"),
					icon: "icon-WCR_Align-TopRight",
					action: _.bind(function() {
						this._emit("doCornerAlign", CompareConstants.cornerAutoAlign.TOPRIGHT);
						this._cornerAlign.setIcon("icon-WCR_Align-TopRight");
					}, this),
					enabled: true
				});

				this._cornerAlign.addAction({
					name: "centerLeftAlign",
					content: this.f.t("AlignCenterLeft"),
					icon: "icon-WCR_Align-MiddleLeft",
					action: _.bind(function() {
						this._emit("doCornerAlign", CompareConstants.cornerAutoAlign.CENTERLEFT);
						this._cornerAlign.setIcon("icon-WCR_Align-MiddleLeft");
					}, this),
					enabled: true
				});

				this._cornerAlign.addAction({
					name: "centeredAlign",
					content: this.f.t("AlignCentered"),
					icon: "icon-WCR_Align-MiddleCenter",
					action: _.bind(function() {
						this._emit("doCornerAlign", CompareConstants.cornerAutoAlign.CENTERED);
						this._cornerAlign.setIcon("icon-WCR_Align-MiddleCenter");
					}, this),
					enabled: true
				});

				this._cornerAlign.addAction({
					name: "centerRightAlign",
					content: this.f.t("AlignCenterRight"),
					icon: "icon-WCR_Align-MiddleRight",
					action: _.bind(function() {
						this._emit("doCornerAlign", CompareConstants.cornerAutoAlign.CENTERRIGHT);
						this._cornerAlign.setIcon("icon-WCR_Align-MiddleRight");
					}, this),
					enabled: true
				});

				this._cornerAlign.addAction({
					name: "bottomLeftAlign",
					content: this.f.t("AlignBottomLeft"),
					icon: "icon-WCR_Align-BottomLeft",
					action: _.bind(function() {
						this._emit("doCornerAlign", CompareConstants.cornerAutoAlign.BOTTOMLEFT);
						this._cornerAlign.setIcon("icon-WCR_Align-BottomLeft");
					}, this),
					enabled: true
				});

				this._cornerAlign.addAction({
					name: "bottomCenterAlign",
					content: this.f.t("AlignBottomCenter"),
					icon: "icon-WCR_Align-BottomCenter",
					action: _.bind(function() {
						this._emit("doCornerAlign", CompareConstants.cornerAutoAlign.BOTTOMCENTER);
						this._cornerAlign.setIcon("icon-WCR_Align-BottomCenter");
					}, this),
					enabled: true
				});

				this._cornerAlign.addAction({
					name: "bottomRightAlign",
					content: this.f.t("AlignBottomRight"),
					icon: "icon-WCR_Align-BottomRight",
					action: _.bind(function() {
						this._emit("doCornerAlign", CompareConstants.cornerAutoAlign.BOTTOMRIGHT);
						this._cornerAlign.setIcon("icon-WCR_Align-BottomRight");
					}, this),
					enabled: true
				});

				// set tooltips
				if (this._channelMatchingBtn) {
					this._channelMatchingBtn.setTooltip(this.f.i18n.t("MapChanTitle"));
				}
				this._autoAlign.setTooltip(this.f.t("AlignAuto"));
				this._rotateReference.setTooltip(this.f.t("RotateReferenceTT"));
				this._flipReference.setTooltip(this.f.t("MirrorReferenceTT"));
				this._negateReference.setTooltip(this.f.t("ToolInvertReferenceTT"));
				this._cornerAlign.setTooltip(this.f.t("AlignCornerReferenceTT"));
				this._trimBoxAlign.setTooltip(this.f.t("AlignFromTrimBoxTT"));
				this._addCompareOptionsToDOM();
			}, this)
		);
	}
};

/**
 * Initialize the controls for CAD overlay
 */
CompareWidget.prototype._initCadOverlayOptions = function() {
	if (!this._cadOverlayOptionsInitialized) {
		this._cadOverlayOptionsInitialized = true;
		// initialize elements
		this._cadOverlaySaveToEmpty = new MiniButton(
			this.f,
			"saveCadOverlayToEmpty",
			"icon-WCR_Save",
			true,
			_.bind(function() {
				this._emit("saveCadOverlay");
			}, this),
			"icon-WCR_Save fontIconDisabled"
		);
		this._cadOverlaySaveToEmpty.setTooltip("viewer", "saveCadRefTT");
		this._cadOverlaySaveToEmpty.hide();

		this._cadOverlaySave = new ViewerToolbarButton(
			this.f,
			"saveCadOVerlay",
			"icon-WCR_Save",
			true,
			_.bind(function() {
				this._emit("saveCadOverlay");
			}, this)
		);

		this._cadOverlaySave.setTooltip("viewer", "SaveCadOverlayTT");

		this._open3DView = new ViewerToolbarButton(
			this.f,
			"open3DView",
			"icon-WCR_3D-Create",
			false,
			_.bind(function() {
				this._compareModule.checkAligmentAndShow3D();
			}, this)
		);
		this._open3DView.setTooltip("viewer", "Open3DViewTT");

		// orientation and alignment
		this._cadOverlayOrientation = $("<div>").addClass("alignment");

		// rotate reference button
		this._cadOverlayRotate = new ViewerToolbarButton(
			this.f,
			"rotateOverlay",
			"icon-WCR_Orientation-Rotate1",
			true,
			_.bind(function() {
				this._emit("rotateCadOverlay");
			}, this)
		);
		this._cadOverlayRotate.setTooltip("viewer", "RotateOverlayTT");

		// flip reference button
		this._cadOverlayFlip = new ViewerToolbarButton(
			this.f,
			"flipOverlay",
			"icon-WCR_Orientation-Mirror",
			true,
			_.bind(function() {
				this._emit("flipCadOverlay");
			}, this)
		);
		this._cadOverlayFlip.setTooltip("viewer", "FlipOverlayTT");

		// manual offset
		this._cadOverlayOffsetX = new UserUnitInputField(
			this.f,
			0,
			"x",
			_.bind(function(oldValue, newValue) {
				this._emit("cadOverlayOffsetXChanged", newValue);
			}, this)
		);

		this._cadOverlayOffsetY = new UserUnitInputField(
			this.f,
			0,
			"y",
			_.bind(function(oldValue, newValue) {
				this._emit("cadOverlayOffsetYChanged", newValue);
			}, this)
		);

		this.f.loadBundle(
			"viewer",
			function() {
				// insert into dom and do i18n

				var documentSelection = $("<div>")
					.addClass("documentSelection cadSelection")
					.append(this._documentSelection.getCadOverlayDocumentSelector().element);

				// header
				if (this.f.isMobile()) {
					this._cadOverlay.append(
						$("<h3>")
							.text(this.f.t("DocumentListCadOverlayTitle"))
							.append(this._cadOverlaySaveToEmpty.element)
					);
				} else {
					documentSelection.append(this._cadOverlaySaveToEmpty.element);
					this._cadOverlay.append($("<h4>").text(this.f.t("DocumentListCadLabel")));
				}

				this._cadOverlay.append(documentSelection);

				this._cadOverlay.append(this._cadOverlayOrientation);

				if (!this.f.isMobile()) {
					this._cadOverlayOrientation.append(
						$("<span>")
							.addClass("alignLabel")
							.text(this.f.t("DiffOrientation"))
					);
				}
				this._cadOverlayOrientation.append(this._cadOverlayRotate.element).append(this._cadOverlayFlip.element);
				this._cadOverlayOrientation.append("<br />");

				if (!this.f.isMobile()) {
					this._cadOverlayOrientation.append(
						$("<span>")
							.addClass("alignLabel")
							.text(this.f.t("AlignLabel"))
					);

					//anchor align button
					this._cadOverlayAnchorAlign = new ViewerToolbarButton(
						this.f,
						"anchorAlign",
						"icon-WCR_Align-Point",
						true,
						_.bind(function() {
							this._compareModule.startAlignmentWizard(true);
						}, this)
					);
					this._cadOverlayAnchorAlign.setTooltip("viewer", "AnchorAlignWizardTitle");
					this._cadOverlayOrientation.append(this._cadOverlayAnchorAlign.element);
				}

				var offsetGroup;
				if (this.f.isMobile()) {
					// mobile optimized html for offset fields
					offsetGroup = $("<div>").addClass("offsetFields");
					offsetGroup.append(
						$("<p>")
							.addClass("alignLabel")
							.text(this.f.t("AlignOffset"))
					);
					offsetGroup.append(
						$("<p>")
							.text("X: ")
							.append(this._cadOverlayOffsetX.element)
					);
					offsetGroup.append(
						$("<p>")
							.text("Y: ")
							.append(this._cadOverlayOffsetY.element)
					);
				} else {
					this._cadOverlayOffsetX.toggleLabel(false);
					offsetGroup = $("<div>").append(
						$("<span>")
							.addClass("alignLabel")
							.text(this.f.t("AlignOffset"))
					);
					offsetGroup.append(
						$("<div>")
							.addClass("offsetFields")
							.append(this._cadOverlayOffsetX.element)
							.append(this._cadOverlayOffsetY.element)
					);
				}

				this._cadOverlayOrientation.append(this._cadOverlaySave.element);
				this._cadOverlayOrientation.append(this._open3DView.element);

				this._cadOverlayOrientation.append(offsetGroup);
			},
			this
		);
	}
};

/**
 * Add the compare buttons and input fields to the DOM
 */
CompareWidget.prototype._addCompareOptionsToDOM = function() {
	if (this._compareOptionsInitialized && this.context && !this._compareOptionsAddedToDOM) {
		this._compareOptionsAddedToDOM = true;
		this.f.loadBundle(
			"viewer",
			function() {
				if (this.f.isMobile()) {
					// MOBILE LAYOUT
					this._documentSelectionTitle = $("<h3/>").text(this.f.t("DocumentListTitle"));
					this._extraControls.append(this._documentSelectionTitle);
				} else {
					// channel matchin is only available on desktop
					this._channelMatching = $("<div/>").addClass("channelMatching");
					this._channelMatching.append($("<h4/>").append(this.f.t("DocumentListChannelMatch") + ": "));
					this._channelMatching.append(this._channelMatchingBtn.element);
					this._compareOptions.append(this._channelMatching);
				}

				// ALIGNMENT AND ORIENTATION
				this._alignBtns = $("<div/>").addClass("alignment");

				var heading = $(this.f.isMobile() ? "<h3/>" : "<h4/>");

				this._alignBtns.append(heading.text(this.f.t("DocumentListAlignment")));

				if (!this.f.isMobile()) {
					this._alignBtns.append(
						$("<span>")
							.addClass("alignLabel")
							.text(this.f.t("DiffOrientation"))
					);
				}
				this._alignBtns.append(this._rotateReference.element);
				this._alignBtns.append(this._flipReference.element);
				this._alignBtns.append(this._negateReference.element);
				this._alignBtns.append("<br>");

				if (!this.f.isMobile()) {
					this._alignBtns.append(
						$("<span>")
							.addClass("alignLabel")
							.text(this.f.t("AlignLabel"))
					);
				}
				this._alignBtns.append(this._autoAlign.element);
				if (!this.f.isMobile()) {
					this._alignBtns.append(this._anchorAlign.element);
				}
				this._alignBtns.append(this._cornerAlign.element);
				this._alignBtns.append(this._trimBoxAlign.element);
				this._alignBtns.append("<br>");

				if (this.f.isMobile()) {
					// mobile optimized html for offset fields
					this._offsetFields = $("<div>").addClass("offsetFields");
					this._offsetFields.append(
						$("<p>")
							.addClass("alignLabel")
							.text(this.f.t("AlignOffset"))
					);
					this._offsetFields.append(
						$("<p>")
							.text("X: ")
							.append(this._offsetX.element)
					);
					this._offsetFields.append(
						$("<p>")
							.text("Y: ")
							.append(this._offsetY.element)
					);
					this._alignBtns.append(this._offsetFields);
				} else {
					this._alignBtns.append(
						$("<span>")
							.addClass("alignLabel")
							.text(this.f.t("AlignOffset"))
					);
					this._offsetFields = $("<div>").addClass("offsetFields");
					this._offsetX.toggleLabel(false);
					this._offsetFields.append(this._offsetX.element);
					this._offsetFields.append(this._offsetY.element);
					this._alignBtns.append(this._offsetFields).append(this._unitLabel);
				}
				this._compareOptions.append(this._alignBtns);

				this.headerExtra.append(this._resetLinkedComparePagesButton.element);

				this._compareModeChanged();
			},
			this
		);
	}
};

/**
 * Set the context of the widget
 * @override
 * @param context
 * @param header
 */
CompareWidget.prototype.setContext = function(context, header) {
	AbstractWidget.prototype.setContext.call(this, context, header);
	this.context.addClass("compareWidget");
	this.context
		.append(this._extraControls)
		.append(this._documentSelectionDiv)
		.append(this._compareOptions)
		.append(this._cadOverlay);

	this._addCompareOptionsToDOM();
	this.headerExtra.addClass("flexEnd");
	this._updateResetLinkedComparePagesBtn();
};

/**
 * Update the compare widget components
 */
CompareWidget.prototype.updateCompareWidget = function() {
	// show/hide align buttons
	var refPage = this.f.getDocumentController().getDifferencePage();

	if (!this.f.getActivePage().canBeComparedTo()) {
		this._diffDocPageElement.addClass("displayNone");
	} else {
		this._diffDocPageElement.removeClass("displayNone");
	}

	if (
		refPage === undefined ||
		this._compareModel.isReaderSpreadActive() ||
		!this.f.getActivePage().canBeComparedTo() ||
		!this.f.getActivePage().areCompareOptionsAllowed() ||
		!refPage.areCompareOptionsAllowed() ||
		(this._compareModule.getActiveMode() && CompareConstants.modes3D[this._compareModule.getActiveMode().name])
	) {
		if (this._compareOptions !== null) {
			this._compareOptions.addClass("displayNone");
		}
	} else {
		this._initCompareOptions();
		this._compareOptions.removeClass("displayNone");

		// update state
		var offset = this._compareModel.getOffset();

		this._offsetX.setValue(offset.x);
		this._offsetY.setValue(offset.y);

		this._trimBoxAlign.setActive(this._compareModel.isTrimBoxAlign());

		var icon = CompareConstants.cornerAlignIcons[this._compareModel.getCornerAlign()];
		if (icon) {
			this._cornerAlign.setIcon(icon);
		} else {
			this._cornerAlign.setIcon("icon-WCR_Align-TopLeft");
		}

		this._updateRotateBtn();
		this._updateFlipBtn();
		this._updateNegateBtn();

		if (refPage.hasCad()) {
			this._negateReference.hide();
			this._trimBoxAlign.hide();
			this._cornerAlign.hide();
			this._autoAlign.hide();
			if (this._channelMatching) {
				this._channelMatching.hide();
			}
		} else {
			this._negateReference.show();
			this._trimBoxAlign.show();
			this._cornerAlign.show();
			this._autoAlign.show();
			if (this._channelMatching) {
				this._channelMatching.show();
			}
		}
	}

	// show/hide cad overlay
	if (this._cadOverlay) {
		this._cadOverlay.toggleClass("displayNone", !this._compareModule.isCadOverlayAllowed());
	}
	var cadOverlay = this.f.getDocumentController().getCadOverlay();
	if (cadOverlay && cadOverlay.hasCad()) {
		this._initCadOverlayOptions();
		this._cadOverlayOrientation.removeClass("displayNone");

		// update state
		var cadOverlayOffset = this._compareModel.getCadOverlayOffset();

		this._cadOverlayOffsetX.setValue(cadOverlayOffset.x);
		this._cadOverlayOffsetY.setValue(cadOverlayOffset.y);

		if (this._compareModule.canSaveCadOverlay()) {
			this._cadOverlaySave.show();
			if (
				!this.f
					.getActivePage()
					.getDocumentVersion()
					.isMultiPage() &&
				this.f.getActivePage().get("canGenerate3D") &&
				this.f.getActivePage().canEditAttributes()
			) {
				// only show "Open 3D button" when user has EditProperties permission and a 3D view can be generated
				this._open3DView.show();
			} else {
				this._open3DView.hide();
			}

			if (!this._compareModule.isCadOverlayModified()) {
				this._cadOverlaySave.setEnabled(false);
			} else {
				this._cadOverlaySave.setEnabled(true);
			}
		} else {
			this._cadOverlaySave.hide();
			this._open3DView.hide();
		}

		if (
			this.f.getActivePage().hasLinkedPage() &&
			this.f.getActivePage().get("canGenerate3D") &&
			this.f.getActivePage().canEditAttributes()
		) {
			// only show "Open 3D button" when user has EditProperties permission and a 3D view can be generated
			this._open3DView.show();
		}

		this._cadOverlaySaveToEmpty.hide();
		this._updateCadOverlayButtons();
	} else {
		if (this._cadOverlayOrientation) {
			this._cadOverlayOrientation.addClass("displayNone");
		}
		if (this._compareModule.isCadOverlayAllowed()) {
			this._initCadOverlayOptions();
		}
		if (this._compareModule.canSaveCadOverlay()) {
			this._initCadOverlayOptions();
			if (this._cadOverlaySaveToEmpty) {
				this._cadOverlaySaveToEmpty.show();
				if (!this._compareModule.isCadOverlayModified()) {
					this._cadOverlaySaveToEmpty.setEnabled(false);
				} else {
					this._cadOverlaySaveToEmpty.setEnabled(true);
				}
			}
		} else {
			if (this._cadOverlaySaveToEmpty) {
				this._cadOverlaySaveToEmpty.hide();
			}
		}
	}

	this._updateResetLinkedComparePagesBtn();

	this._fireResized();
};

/**
 * @private
 * Update the rotate btn
 */
CompareWidget.prototype._updateRotateBtn = function() {
	var rotation = this._compareModel.getReferenceRotation();
	if (rotation === CompareConstants.referenceRotation.NOTROTATED) {
		this._rotateReference.setIcon("icon-WCR_Orientation-Rotate1");
	} else if (rotation === CompareConstants.referenceRotation.ROTATED90) {
		this._rotateReference.setIcon("icon-WCR_Orientation-Rotate2");
	} else if (rotation === CompareConstants.referenceRotation.ROTATED180) {
		this._rotateReference.setIcon("icon-WCR_Orientation-Rotate3");
	} else if (rotation === CompareConstants.referenceRotation.ROTATED270) {
		this._rotateReference.setIcon("icon-WCR_Orientation-Rotate4");
	}
};

/**
 * @private
 * update the flip btn
 */
CompareWidget.prototype._updateFlipBtn = function() {
	if (this._compareModel.isReferenceFlipped()) {
		this._flipReference.setIcon("icon-WCR_Orientation-MirrorReverse");
		this._flipReference.setTooltip(this.f.t("MirrorReferenceTT2"));
	} else {
		this._flipReference.setIcon("icon-WCR_Orientation-Mirror");
		this._flipReference.setTooltip(this.f.t("MirrorReferenceTT"));
	}
};

/**
 * @private
 * update the negate btn
 */
CompareWidget.prototype._updateNegateBtn = function() {
	if (this._compareModel.isReferenceNegated()) {
		this._negateReference.setIcon("icon-WCR_InvertReverse");
		this._negateReference.setTooltip(this.f.t("ToolInvertReferenceTT2"));
	} else {
		this._negateReference.setIcon("icon-WCR_Invert");
		this._negateReference.setTooltip(this.f.t("ToolInvertReferenceTT"));
	}
};

/**
 * @private
 * Update the reset linked compare pages button
 */
CompareWidget.prototype._updateResetLinkedComparePagesBtn = function() {
	var isMultiPage = this.f
		.getActivePage()
		.getDocumentVersion()
		.isMultiPage();
	if (isMultiPage) {
		this._resetLinkedComparePagesButton.setEnabled(
			_.some(
				this.f
					.getActivePage()
					.getDocumentVersion()
					.getAllPages(),
				function(page) {
					return page.hasLinkedComparePage();
				}
			)
		);
	}
};

/**
 * Toggle the visibility of the difference document selector
 * @param {Boolean} hide - show or hide the document selector
 */
CompareWidget.prototype.toggleDifferenceDocPageSelector = function(hide) {
	if (hide) {
		this._diffDocPageElement.hide();
	} else {
		this._diffDocPageElement.show();
	}
};

/**
 * Update the orientation buttons for the cad overlay
 */
CompareWidget.prototype._updateCadOverlayButtons = function() {
	var rotation = this._compareModel.getCadOverlayRotation();
	if (rotation === CompareConstants.referenceRotation.NOTROTATED) {
		this._cadOverlayRotate.setIcon("icon-WCR_Orientation-Rotate1");
	} else if (rotation === CompareConstants.referenceRotation.ROTATED90) {
		this._cadOverlayRotate.setIcon("icon-WCR_Orientation-Rotate2");
	} else if (rotation === CompareConstants.referenceRotation.ROTATED180) {
		this._cadOverlayRotate.setIcon("icon-WCR_Orientation-Rotate3");
	} else if (rotation === CompareConstants.referenceRotation.ROTATED270) {
		this._cadOverlayRotate.setIcon("icon-WCR_Orientation-Rotate4");
	}

	if (this._compareModel.isCadOverlayFlipped()) {
		this._cadOverlayFlip.setIcon("icon-WCR_Orientation-MirrorReverse");
		this._cadOverlayFlip.setTooltip(this.f.t("MirrorCadOverlayTT2"));
	} else {
		this._cadOverlayFlip.setIcon("icon-WCR_Orientation-Mirror");
		this._cadOverlayFlip.setTooltip(this.f.t("MirrorCadOverlayTT"));
	}
};

/**
 * Create element + content for content traversal
 * @returns {JQueryObject}
 */
CompareWidget.prototype.createDocTraversalElem = function() {
	var container = $("<div>").addClass("doctraversal");
	var docs = this.f.getDocumentController().getAllDocuments();

	if (docs.length > 1) {
		var nextFn = function(back) {
			var target = 0;
			var i = this.getCurrentDocIndex();
			if (back) {
				target = (i - 1 + docs.length) % docs.length;
			} else {
				target = (i + 1) % docs.length;
			}
			this.f.emit("selection.requestdoc", docs[target]);
		};
		var endFn = function(back) {
			var target = back ? 0 : docs.length - 1;
			this.f.emit("selection.requestdoc", docs[target]);
		};

		this._firstBtn = new MiniButton(this.f, "SelectFirstDoc", "icon icon-WCR_Table-Page-First12", true, _.bind(endFn, this, true));
		this._firstBtn.setTooltip("viewer", "NavigatorFirstTT");
		container.append(this._firstBtn.element);

		this._prevtBtn = new MiniButton(this.f, "SelectPreviousDoc", "icon icon-WCR_Table-Page-Previous12", true, _.bind(nextFn, this, true));
		this._prevtBtn.setTooltip("viewer", "NavigatorPreviousTT");
		container.append(this._prevtBtn.element);

		this._docIndexSpan = $("<span />").addClass("documentOverview");
		container.append(this._docIndexSpan);

		this._nextBtn = new MiniButton(this.f, "SelectNextDoc", "icon icon-WCR_Table-Page-Next12", true, _.bind(nextFn, this, false));
		this._nextBtn.setTooltip("viewer", "NavigatorNextTT");
		container.append(this._nextBtn.element);

		this._lastBtn = new MiniButton(this.f, "SelectLastDoc", "icon icon-WCR_Table-Page-Last12", true, _.bind(endFn, this, false));
		this._lastBtn.setTooltip("viewer", "NavigatorLastTT");
		container.append(this._lastBtn.element);

		this.updateDocTraversalElem();
		this.docTraversalInitialized = true;
	}
	return container;
};

/**
 * Update document traversal tool
 */
CompareWidget.prototype.updateDocTraversalElem = function() {
	if (!this.docTraversalInitialized) {
		return;
	}

	var index = this.getCurrentDocIndex();
	var docs = this.f.getDocumentController().getAllDocuments();

	this._docIndexSpan.empty().append(this.f.t("PageNumberOfCount", this.getCurrentDocIndex() + 1, docs.length));

	this._firstBtn.setEnabled(index !== 0);
	this._prevtBtn.setEnabled(index !== 0);
	this._nextBtn.setEnabled(index !== docs.length - 1);
	this._lastBtn.setEnabled(index !== docs.length - 1);
};

/**
 * get index of currently selected document in selector
 * @returns {Number}
 */
CompareWidget.prototype.getCurrentDocIndex = function() {
	var docs = this.f.getDocumentController().getAllDocuments();
	var activeDoc = this.f.getActiveDocument();
	return docs.indexOf(activeDoc);
};

export default CompareWidget;
