import _ from "underscore";
import ViewportLayerModel from "@esko/web-viewer/viewport/model/ViewportLayerModel";
import CompareConstants from "@esko/web-viewer/compare/CompareConstants";
import LineworkWrapper from "@esko/web-viewer/utils/LineworkWrapper";
import DrawUtils from "@esko/web-viewer/utils/DrawUtils";
import ShapeUtils from "@esko/shapeutils";
import SnapIndicator from "@esko/web-viewer/utils/SnapIndicator";
/**
 * @exports CompareAlignmentAnchorLayerModel
 * @memberof Compare
 * @augments ViewportLayerModel
 * @class This model holds 1 anchor point to draw in a layer
 * @constructor
 *
 * @param {Core.ViewerFacade} f - Viewer facade
 * @param {String} ID - the ID of the layer model
 */

var CompareAlignmentAnchorLayerModel = function(f, ID) {
	ViewportLayerModel.call(this, f, ID);
	this._anchor = null;
	this._isActive = false;
	this._snapIndicator = null;
	this.f.getDocumentController().whenActivePageParametersChange(_.bind(this._updateTransformation, this));
};

// inherit
CompareAlignmentAnchorLayerModel.prototype = Object.create(ViewportLayerModel.prototype);
// correct the constructor pointer
CompareAlignmentAnchorLayerModel.prototype.constructor = CompareAlignmentAnchorLayerModel;

/**
 * Set the position of the anchor
 * @param {ShapeUtils.Point} anchor
 */
CompareAlignmentAnchorLayerModel.prototype.setAnchor = function(anchor) {
	this._anchor = anchor;
	this.modelChanged();
};

/**
 * Move the anchor
 * @param {Number} x
 * @param {Number} y
 */
CompareAlignmentAnchorLayerModel.prototype.moveAnchor = function(x, y) {
	this._anchor.x += x;
	this._anchor.y += y;
	this.modelChanged();
};

/**
 * Change how the anchor should be displayed (active or non-active)
 * @param {Boolean} isActive
 */
CompareAlignmentAnchorLayerModel.prototype.setActive = function(isActive) {
	this._isActive = isActive;
};

/**
 * Update the transformation of the layer so that (0,0) is (0,0) in the current document
 */
CompareAlignmentAnchorLayerModel.prototype._updateTransformation = function() {
	this.resetTransformation();
	var page = this.f.getActivePage();
	var origin = page.getDocumentPageOrigin();
	this.transform.translate(origin.x, origin.y);
};

/**
 * Indicate CAD snap point
 * @param {ShapeUtils.Point} [point] - snap point to indicate
 */
CompareAlignmentAnchorLayerModel.prototype.indicateSnapPoint = function(point) {
	if (this._snapPoint !== point) {
		this._snapPoint = point;
		this.modelChanged();
	}
};

/**
 * @override
 */
CompareAlignmentAnchorLayerModel.prototype.getLinework = function() {
	var linework = [];

	if (this._anchor) {
		var anchor = new LineworkWrapper();
		anchor.type = DrawUtils.LWTYPE_IMAGE;
		anchor.linewidth = 1;
		anchor.imageRef = CompareConstants.anchorAlignment[this._isActive ? "ACTIVE" : "DEFAULT"];
		anchor.staticsize = true;
		anchor.setRect(new ShapeUtils.Rect(new ShapeUtils.Point(this._anchor.x, this._anchor.y)));
		anchor.imageOffset = new ShapeUtils.Size(CompareConstants.anchorAlignment.OFFSET.width, CompareConstants.anchorAlignment.OFFSET.height);

		linework.push(anchor);
	}

	if (this._snapPoint) {
		if (!this._snapIndicator) {
			this._snapIndicator = new SnapIndicator(this._snapPoint);
		} else {
			this._snapIndicator.setCenter(this._snapPoint);
		}

		linework.push(this._snapIndicator);
	}

	return linework;
};

export default CompareAlignmentAnchorLayerModel;
