import AbstractModel from "@esko/web-viewer/core/AbstractModel";
import ShapeUtils from "@esko/shapeutils";
import _ from "underscore";
/**
 * @exports CompareAlignmentAnchorModel
 * @memberof CompareModule
 * @class
 * This model contains 2 anchors, used for anchor alignment (both anchors can be null)
 * These points are stored in document coordinates (activePage) relative to activePage.documentPageOrigin()
 * @constructor
 * @param {Core.ViewerFacade} f - Viewerfacade
 * @param {String,Array} channelName - the name of the private channel
 */
var CompareAlignmentAnchorModel = function(f, channelName) {
	AbstractModel.call(this, f, channelName);

	this._currentAnchor = null;
	this._referenceAnchor = null;

	// by default the reference anchor is active
	this._isCurrentActive = false;
};

CompareAlignmentAnchorModel.prototype = Object.create(AbstractModel.prototype);
CompareAlignmentAnchorModel.prototype.constructor = CompareAlignmentAnchorModel;

/**
 * Change the position of the current anchor
 * @param {ShapeUtils.Point} point
 */
CompareAlignmentAnchorModel.prototype.setCurrentAnchor = function(point) {
	var oldCurrent = this._currentAnchor;
	this._currentAnchor = point !== null ? point.copy() : null;
	this._fireChange(false, ["current"], oldCurrent, this._currentAnchor !== null ? this._currentAnchor.copy() : null);
};

/**
 * Move the current Anchor
 * @param {Number} x - x offset
 * @param {Number} y - y offset
 */
CompareAlignmentAnchorModel.prototype.moveCurrentAnchor = function(x, y) {
	var oldCurrent = this._currentAnchor;

	if (this._currentAnchor === null) {
		this._currentAnchor = new ShapeUtils.Point();
	}
	this._currentAnchor = new ShapeUtils.Point(this._currentAnchor.x + x, this._currentAnchor.y + y);
	this._fireChange(false, ["current"], oldCurrent, this._currentAnchor.copy());
};

/**
 * Change the position of the reference anchor
 * @param {ShapeUtils.Point} point
 */
CompareAlignmentAnchorModel.prototype.setReferenceAnchor = function(point) {
	var oldReference = this._referenceAnchor;
	this._referenceAnchor = point !== null ? point.copy() : null;
	this._fireChange(false, ["reference"], oldReference, this._referenceAnchor !== null ? this._referenceAnchor.copy() : null);
};

/**
 * Move the reference anchor
 * @param {Number} x - x offset
 * @param {Number} y - y offset
 */
CompareAlignmentAnchorModel.prototype.moveReferencAnchor = function(x, y) {
	var oldReference = this._referenceAnchor;

	if (this._referenceAnchor === null) {
		this._referenceAnchor = new ShapeUtils.Point();
	}

	this._referenceAnchor = new ShapeUtils.Point(this._referenceAnchor.x + x, this._referenceAnchor.y + y);
	this._fireChange(false, ["reference"], oldReference, this._referenceAnchor.copy());
};

/**
 * Calculate the vector from current to reference anchor
 * @returns {ShapeUtils.Size}
 */
CompareAlignmentAnchorModel.prototype.getOffset = function() {
	if (this._currentAnchor !== null && this._referenceAnchor !== null) {
		return new ShapeUtils.Size(this._currentAnchor.x - this._referenceAnchor.x, this._currentAnchor.y - this._referenceAnchor.y);
	} else {
		return undefined;
	}
};

/**
 * Get the position of the current anchor (anchor 1)
 * @returns {ShapeUtils.Point}
 */
CompareAlignmentAnchorModel.prototype.getCurrentAnchor = function() {
	return this._currentAnchor !== null ? this._currentAnchor.copy() : null;
};

/**
 * Get the position of the reference anchor (anchor 2)
 * @returns {ShapeUtils.Point}
 */
CompareAlignmentAnchorModel.prototype.getReferenceAnchor = function() {
	return this._referenceAnchor !== null ? this._referenceAnchor.copy() : null;
};

/**
 * Get the position of the currently active anchor (depending on isCurrentActive)
 * @returns {ShapeUtils.Point}
 */
CompareAlignmentAnchorModel.prototype.getActiveAnchor = function() {
	if (this._isCurrentActive) {
		return this.getCurrentAnchor();
	} else {
		return this.getReferenceAnchor();
	}
};

/**
 * Change which anchor is active
 * @param {Boolean} isCurrentActive - true if the current anchor is active, false if the reference anchor is active
 */
CompareAlignmentAnchorModel.prototype.setCurrentActive = function(isCurrentActive) {
	var before = this._isCurrentActive;
	this._isCurrentActive = isCurrentActive;
	this._fireChange(false, ["isCurrentActive"], before, isCurrentActive);
};

/**
 * Check which anchor is active
 * @returns {Boolean} true if the current anchor is active
 */
CompareAlignmentAnchorModel.prototype.isCurrentActive = function() {
	return this._isCurrentActive;
};

/**
 * Change the position of the currently activated anchor
 * @param {ShapeUtils.Point} point
 */
CompareAlignmentAnchorModel.prototype.setActiveAnchorPosition = function(point) {
	if (this.isCurrentActive()) {
		this.setCurrentAnchor(point);
	} else {
		this.setReferenceAnchor(point);
	}
};

/**
 * Move the currently active anchor
 * @param {Number} x - x offset
 * @param {Number} y - y offset
 */
CompareAlignmentAnchorModel.prototype.moveActiveAnchor = function(x, y) {
	if (this.isCurrentActive()) {
		this.moveCurrentAnchor(x, y);
	} else {
		this.moveReferencAnchor(x, y);
	}
};

/**
 * Reset the anchor model (this resets the position of both anchors)
 */
CompareAlignmentAnchorModel.prototype.reset = function() {
	this.setCurrentAnchor(null);
	this.setReferenceAnchor(null);
};

export default CompareAlignmentAnchorModel;
