import _ from "underscore";
import CompareMode from "@esko/web-viewer/compare/CompareMode";
import CompareConst from "@esko/web-viewer/compare/CompareConstants";
import ViewportConstants from "@esko/web-viewer/viewport/ViewportConstants";
import ListenerUtil from "@esko/web-viewer/utils/ListenerUtil";
import ShapeUtils from "@esko/shapeutils";
/**
 * @memberof CompareModule
 * @exports CompareAlignmentMode
 * @class This mode shows a side by side view of the current and reference document. It enables the user to quickly align two files. (the result of alignment is shown on the right)
 * @constructor
 * @param {Core.ViewerFacade} f
 * @param {CompareModule.CompareModel} compareModel - model for compare
 * @param {CompareModule.CompareAlignmentAnchorLayerModel} currentAnchorLayerModel - the layer model which displays the current anchor
 * @param {CompareModule.CompareAlignmentAnchorLayerModel} referenceAnchorLayerModel - the layer model which displays the reference anchor
 */
var CompareAlignmentMode = function(f, compareModel, currentAnchorLayerModel, referenceAnchorLayerModel) {
	CompareMode.call(this, f, CompareConst.mode.ALIGNMENT, compareModel, false, "DiffSideBySideAnnotationsHidden");
	ListenerUtil.mixin(this, f);

	// is the mode already listening to the graphics model
	this._isListeningToGraphics = false;

	this._currentFilePanel = null;
	this._differenceFilePanel = null;
	this._resultPanel = null;

	this._currentAnchorLayerModel = currentAnchorLayerModel;
	this._referenceAnchorLayerModel = referenceAnchorLayerModel;

	this._forCadOverlay = false;

	this.viewport = this.f.getModule("viewport");
};

CompareAlignmentMode.prototype = Object.create(CompareMode.prototype);
CompareAlignmentMode.prototype.constructor = CompareAlignmentMode;

/**
 * Start/stop show compare mode
 * @param {Boolean} toggle
 * @override
 */
CompareAlignmentMode.prototype.toggle = function(toggle) {
	if (!this.isEnabled() && toggle) {
		CompareMode.prototype.toggle.call(this, toggle);

		// add layers for drawing the anchors
		this.viewport.addLayer("compare_alignment_currentAnchor", this._currentAnchorLayerModel);
		this.viewport.addLayer("compare_alignment_referenceAnchor", this._referenceAnchorLayerModel);

		if (!this.graphicsModule) {
			this.graphicsModule = this.f.getModule("graphics");
		}

		// when enabled, set the diff mode and add the needed panels
		this._compareModel.setDiffMode(CompareConst.diffMode.SHOWBOTHSIDEBYSIDEHIGHLIGHT);

		// set panel layout mode
		this.viewport.setPanelsLayoutMode(ViewportConstants.PanelsLayoutMode.VERTICAL);

		// get a reference to the default panel
		this._currentFilePanel = this.viewport.getPanelModel("default");
		this._currentFilePanel.toggleLayer("referenceCad", false);
		this._currentFilePanel.toggleLayer("currentDiffCad", false);
		this._currentFilePanel.toggleLayer("referenceDiffCad", false);
		this._currentFilePanel.toggleLayer("Annotations", false);
		this._currentFilePanel.toggleLayer("AnnotationPopups", false);
		this._currentFilePanel.toggleLayer("compare_alignment_referenceAnchor", false);
		this._currentFilePanel.toggleLayer("CadOverlay", !this._forCadOverlay);

		// add difference panel
		var disallowedLayersForDifferencePanel = [
			"Cad",
			"currentDiffCad",
			"referenceDiffCad",
			"Annotations",
			"AnnotationPopups",
			"compare_alignment_currentAnchor"
		];

		if (!this._forCadOverlay) {
			disallowedLayersForDifferencePanel.push("CadOverlay");
		}

		this._differenceFilePanel = this.viewport.addPanel(
			"compare_difference",
			this._getAllDifferenceFileOverwriteOptions(),
			disallowedLayersForDifferencePanel
		);
		// show the result panel
		this._resultPanel = this.viewport.addPanel("compare_align_result", this._getResultOverwriteOptions(), [
			"Annotations",
			"AnnotationPopups",
			"Cad",
			"referenceCad",
			"compare_alignment_referenceAnchor"
		]);

		this.updatePanelHeaders();

		// update the overwrite options for all panels
		this.updateOverwriteOptions();

		// make sure to react to channel changes to update the overwrite options
		this.f.whenModuleAvailable(
			"graphics",
			function(module) {
				this.graphicsModule = module;

				if (this.isEnabled() && !this._isListeningToGraphics) {
					this._registerListener(
						"model.changed.channels",
						_.debounce(_.bind(this.updateOverwriteOptions, this), 40),
						undefined,
						this.graphicsModule
					);
					this._registerListener(
						"model.changed.channel.**",
						_.debounce(_.bind(this.updateOverwriteOptions, this), 40),
						undefined,
						this.graphicsModule
					);
					this._isListeningToGraphics = true;
					this.updateOverwriteOptions();
				}
			},
			this
		);

		this._registerListener("doc.page.active", this.updatePanelHeaders, this);
		this._registerListener("doc.page.difference", this.updatePanelHeaders, this);
		this.updatePanelHeaders();
	} else if (this.isEnabled() && !toggle) {
		CompareMode.prototype.toggle.call(this, toggle);

		// remove panels
		this.viewport.removePanel("compare_difference");
		this.viewport.removePanel("compare_align_result");

		// the side by side compare is disabled, restore everything and remove the previously created panels
		this._compareModel.setDiffMode(CompareConst.diffMode.UNDEFINED);

		// reset overwrite options
		this._currentFilePanel.deleteOverwriteOption("dmode");
		this._currentFilePanel.deleteOverwriteOption("colorBITWISE");
		this._currentFilePanel.toggleLayer("Annotations", true);
		this._currentFilePanel.toggleLayer("AnnotationPopups", true);
		this._currentFilePanel.toggleLayer("referenceCad", true);
		this._currentFilePanel.toggleLayer("currentDiffCad", true);
		this._currentFilePanel.toggleLayer("referenceDiffCad", true);
		this._currentFilePanel.toggleLayer("CadOverlay", true);
		this._currentFilePanel.setHeaderText(null);

		// reset references
		this._currentFilePanel = null;
		this._differenceFilePanel = null;
		this._resultPanel = null;

		// set panel layout to automatic
		this.viewport.setPanelsLayoutMode(ViewportConstants.PanelsLayoutMode.AUTOMATIC);

		// unregister all listeners
		this._unregisterAllListeners();
		this._isListeningToGraphics = false;

		this.viewport.removeLayer("compare_alignment_currentAnchor");
		this.viewport.removeLayer("compare_alignment_referenceAnchor");
	}
};

/**
 * Configure the overwrite options for all created panels
 */
CompareAlignmentMode.prototype.updateOverwriteOptions = function() {
	if (this.isEnabled()) {
		if (this.graphicsModule === undefined) {
			this.graphicsModule = this.f.getModule("graphics");
		}
		if (this.annotationModule === undefined) {
			this.annotationModule = this.f.getModule("annotation");
		}

		this._currentFilePanel.setOverwriteOptions(this._getCurrentFileOverwriteOptions());
		this._differenceFilePanel.setOverwriteOptions(this._getAllDifferenceFileOverwriteOptions());
		this._resultPanel.setOverwriteOptions(this._getResultOverwriteOptions());
	}
};

/**
 * Generate the overwrite options for the current file panel
 * @returns {Object}
 */
CompareAlignmentMode.prototype._getCurrentFileOverwriteOptions = function() {
	throw "Not Implemented: CompareAlignmentMode._getCurrentFileOverwriteOptions";
};

/**
 * Generate the overwrite options for the difference file panel
 * @returns {Object}
 */
CompareAlignmentMode.prototype._getAllDifferenceFileOverwriteOptions = function() {
	var options = this._getDifferenceFileOverwriteOptions();
	options.isDefaultView = false;
	return options;
};
/**
 * Generate the overwrite options for the difference file panel
 * @returns {Object}
 */
CompareAlignmentMode.prototype._getDifferenceFileOverwriteOptions = function() {
	throw "Not Implemented: CompareAlignmentMode._getDifferenceFileOverwriteOptions";
};

/**
 * Generate the overwrite options for the difference compare mode panel
 * @returns {Object}
 */
CompareAlignmentMode.prototype._getResultOverwriteOptions = function() {
	throw "Not Implemented: CompareAlignmentMode._getResultOverwriteOptions";
};

/**
 * Update the header text for the panels
 */
CompareAlignmentMode.prototype.updatePanelHeaders = function() {
	this.f.loadBundle(
		"viewer",
		function() {
			if (this.isEnabled()) {
				if (!this._forCadOverlay) {
					var current = "";
					var difference = "";

					var docController = this.f.getDocumentController();
					var activePage = docController.getActivePage();
					var activeDocVersion = activePage.getDocumentVersion();

					var differencePage = docController.getDifferencePage();
					var differenceDocVersion = differencePage.getDocumentVersion();

					if (activeDocVersion.getDocument() !== differenceDocVersion.getDocument()) {
						// we need to add document info too
						current += activeDocVersion.getDocument().get("docname") + ", ";
						difference += differenceDocVersion.getDocument().get("docname") + ", ";
					}

					current += this.f.t("WCHistVersion") + " " + activeDocVersion.getVersionId();
					current += " - (" + this.f.t("CaptionCurrent") + ")";

					difference += this.f.t("WCHistVersion") + " " + differenceDocVersion.getVersionId();
					difference += " - (" + this.f.t("CaptionPrevious") + ")";

					this._currentFilePanel.setHeaderText(current);
					this._differenceFilePanel.setHeaderText(difference);

					if (this._resultPanel) {
						this._resultPanel.setHeaderText(this.f.t("AnchorAlignWizardResultPanelHeader"));
					}
				} else {
					// set headers for cad overlay
				}
			}
		},
		this
	);
};

/**
 * Enable the "CAD OVERLAY" alignment mode (needs to be done before mode is started
 * @param {Boolean} forCadOverlay - if true, enable cad overlay alignment mode
 */
CompareAlignmentMode.prototype.setForCadOverlay = function(forCadOverlay) {
	this._forCadOverlay = forCadOverlay;
};

export default CompareAlignmentMode;
