import ViewerMode from "@esko/web-viewer/core/ViewerMode";
import CoreConstants from "@esko/web-viewer/core/CoreConstants";
import CompareAlignmentAnchorModel from "@esko/web-viewer/compare/alignment/CompareAlignmentAnchorModel";
import CompareAlignmentAnchorLayerModel from "@esko/web-viewer/compare/alignment/CompareAlignmentAnchorLayerModel";
import CompareAnchorTool from "@esko/web-viewer/compare/alignment/CompareAnchorTool";
import CompareAlignmentWizardWidget from "@esko/web-viewer/compare/alignment/CompareAlignmentWizardWidget";
import ListenerUtil from "@esko/web-viewer/utils/ListenerUtil";
import ShapeUtils from "@esko/shapeutils";
import LayoutConstants from "@esko/web-viewer/layout/LayoutConstants";
import GraphicsConstants from "@esko/web-viewer/graphics/GraphicsConstants";
import CadConstants from "@esko/web-viewer/cad/CadConstants";
import Mousetrap from "@esko/mousetrap";
import _ from "underscore";
/**
 * @class CompareAlignmentWizard the Compare Alignment Wizard is a ViewerMode which starts up an interactive anchor alignment for the compare module
 * @memberof CompareModule
 * @exports CompareAlignmentWizard
 *
 * @augments {Core.ViewerMode}
 * @param {Core.ViewerFacade} f
 * @param {CompareModule} compareModule - the compareModule
 * @param {CompareModel} compareModel - the compare model
 * @param {CompareAlignmentMode} CompareAlignmentMode - constructor to create the CompareAlignmentMode
 */
var CompareAlignmentWizard = function(f, compareModule, compareModel, CompareAlignmentMode) {
	ViewerMode.call(this, f, "compareAlignmentWizardMode");
	ListenerUtil.mixin(this, f);

	this._compareModule = compareModule;
	this._compareModel = compareModel;

	this._alignmentAnchorModel = new CompareAlignmentAnchorModel(this.f, ["module", "compare", "alignment", "anchor"]);
	this._currentAnchorLayerModel = new CompareAlignmentAnchorLayerModel(this.f, "compare_alignment_currentAnchor");
	this._referenceAnchorLayerModel = new CompareAlignmentAnchorLayerModel(this.f, "compare_alignment_referenceAnchor");

	this._compareAlignmentMode = new CompareAlignmentMode(
		this.f,
		this._compareModel,
		this._currentAnchorLayerModel,
		this._referenceAnchorLayerModel
	);
	this._compareModel.registerMode(this._compareAlignmentMode);
	this._compareAlignmentWizardWidget = null;
	this._middleToActiveAnchor = null;
	this._anchorTool = null;
	this._forCadOverlay = false;

	this._offsetBeforeStart = null;
	this._before = {};
	this._activeAnchorToWorldOffset = new ShapeUtils.Size();

	this._compareDiffMode = null;

	this._onPressA = _.bind(function() {
		this.f.emit("toolManager.activateTool", this._anchorTool);
	}, this);
};

CompareAlignmentWizard.prototype = Object.create(ViewerMode.prototype);
CompareAlignmentWizard.prototype.constructor = CompareAlignmentWizard;

/**
 * The Wizard mode is started
 */
CompareAlignmentWizard.prototype._doStart = function() {
	this._before = {};

	// store the current compare offset (so we can restore it on cancel)
	if (!this._forCadOverlay) {
		this._offsetBeforeStart = this._compareModel.getOffset();
	} else {
		this._offsetBeforeStart = this._compareModel.getCadOverlayOffset();
	}

	// Save the current Channel view mode and set it to composite view
	var graphicsModule = this.f.getModule("graphics");
	if (graphicsModule) {
		this._before.channelViewMode = graphicsModule.getChannelViewMode();
		graphicsModule.setChannelViewMode(GraphicsConstants.ChannelViewMode.COMPOSITE_VIEW);
	}

	// Save the current CAD+Graphics mode and change it to show both
	var cadModule = this.f.getModule("cad");
	if (cadModule) {
		this._before.cadGraphicsMode = cadModule.getCadGraphicsMode();
		cadModule.setCadGraphicsMode(CadConstants.CadGraphicsMode.BOTH);
	}

	// listen to anchor model changes
	this._registerListener("changed.*", this._anchorsUpdated, this, this._alignmentAnchorModel);

	// listen to reference offset changes
	this._registerListener("referenceOffset.beforeChange", this._storeMiddleToAnchorOffset, this, this._compareModule);
	this._registerListener("cadOverlayOffset.beforeChange", this._storeMiddleToAnchorOffset, this, this._compareModule);

	// listen to compare model offset changes
	this._registerListener("referenceOffset.changed", this._offsetChanged, this, this._compareModel);
	this._registerListener("cadOverlayOffset.changed", this._offsetChanged, this, this._compareModel);

	this._alignmentAnchorModel.reset();

	this._currentAnchorLayerModel._updateTransformation();
	this._referenceAnchorLayerModel._updateTransformation();

	this._disallowAllToolsExcept([CoreConstants.TOOLS.PAN, CoreConstants.TOOLS.ZOOM, CoreConstants.TOOLS.COMPARE_ANCHOR_ALIGNMENT]);

	// store the current compare mode so we can restore it later on
	this._compareDiffMode = this._compareModule.getCurrentDiffMode();

	this._compareAlignmentMode.setForCadOverlay(this._forCadOverlay);

	// enable the alignment compare mode
	this._compareModule.activateModeByName("alignment");

	// initialize the anchor tool if non-existent
	if (!this._anchorTool) {
		this._initializeAnchorTool();
	}

	this._anchorTool.setForCadOverlay(this._forCadOverlay);

	this.f.emit("toolManager.activateTool", this._anchorTool);

	if (!this._compareAlignmentWizardWidget) {
		// we directly create the widget, the wizard can't really work without it
		this._compareAlignmentWizardWidget = new CompareAlignmentWizardWidget(this.f, this, this._anchorTool);
		// _publishWidget
		this.f.emit(
			"layout.addWidget",
			CompareAlignmentWizardWidget.widgetName,
			_.bind(function(layoutType) {
				return this._compareAlignmentWizardWidget;
			}, this)
		);
	}

	this._compareAlignmentWizardWidget.setEnabled(true);

	Mousetrap.bind("a", this._onPressA, "keypress");

	this.f.emit("layout.switchLayoutMode", LayoutConstants.LayoutMode.WIZARD);
	if (this._compareAlignmentWizardWidget) {
		this._compareAlignmentWizardWidget.render();
	}
};

/**
 * The wizard mode is ended
 */
CompareAlignmentWizard.prototype._doStop = function() {
	this._unregisterAllListeners();

	// allow all tools again
	this._allowAllTools();

	// disable the anchor tool
	if (this._anchorTool.isActive()) {
		this._anchorTool.requestDeactivation();
	}

	// activate the previous diffmode
	this._compareModule.activateModeByDiffMode(this._compareDiffMode);
	this._compareDiffMode = null;
	Mousetrap.unbind("a", this._onPressA, "keypress");

	this._compareAlignmentWizardWidget.setEnabled(false);

	this.f.emit("layout.switchLayoutMode");

	// restore the channel view mode
	var graphicsModule = this.f.getModule("graphics");
	if (graphicsModule) {
		graphicsModule.setChannelViewMode(this._before.channelViewMode);
	}

	// restore the CAD + Graphics mode
	var cadModule = this.f.getModule("cad");
	if (cadModule && this._before.cadGraphicsMode !== undefined) {
		cadModule.setCadGraphicsMode(this._before.cadGraphicsMode);
	}
};

/**
 * Update the anchors
 */
CompareAlignmentWizard.prototype._offsetChanged = function(oldValue, newValue) {
	// hide snapping indicator when chaning the transformation
	this._currentAnchorLayerModel.indicateSnapPoint(null);
	this._referenceAnchorLayerModel.indicateSnapPoint(null);
	this._currentAnchorLayerModel._updateTransformation();
	this._referenceAnchorLayerModel._updateTransformation();
	this._alignmentAnchorModel.setReferenceAnchor(this._alignmentAnchorModel.getCurrentAnchor());

	this._restoreMiddleToAnchorOffset();
};

/**
 * Change the reference offset of the CompareModule
 */
CompareAlignmentWizard.prototype._setReferenceOffset = function() {
	var anchorOffset = this._alignmentAnchorModel.getOffset();
	var cadModule;
	if (anchorOffset && !(anchorOffset.width === 0 && anchorOffset.height === 0)) {
		if (!this._forCadOverlay) {
			var compareOffset = this._compareModel.getOffset();
			if (this.f.getActivePage().hasCad()) {
				// CAD compare alignment
				// we need to apply the inverse transform from the reference CAD to our offset difference

				cadModule = this.f.getModule("cad");
				anchorOffset = cadModule
					.getReferenceTransform()
					.createInverse()
					.transformedSize(anchorOffset);
				//we need to change the x-offset to the other way for CAD
				anchorOffset.width = -anchorOffset.width;
			}

			this._compareModule.changeReferenceOffset(
				new ShapeUtils.Point(compareOffset.x + anchorOffset.width, compareOffset.y + anchorOffset.height)
			);
		} else {
			// CAD overlay
			var cadOverlayOffset = this._compareModel.getCadOverlayOffset();

			cadModule = this.f.getModule("cad");
			// apply the inverse transform for the CAD overlay to the offset difference
			anchorOffset = cadModule
				.getOverlayTransform()
				.createInverse()
				.transformedSize(anchorOffset);

			this._compareModule.changeCadOverlayOffset(
				new ShapeUtils.Point(cadOverlayOffset.x - anchorOffset.width, cadOverlayOffset.y + anchorOffset.height)
			);
		}
	}
};

/**
 * The anchors have changed positions, update the layermodel, and set the offset on the compare module
 */
CompareAlignmentWizard.prototype._anchorsUpdated = function() {
	this._referenceAnchorLayerModel.setAnchor(this._alignmentAnchorModel.getReferenceAnchor());
	this._currentAnchorLayerModel.setAnchor(this._alignmentAnchorModel.getCurrentAnchor());
	this._currentAnchorLayerModel.setActive(this._alignmentAnchorModel.isCurrentActive());
	this._referenceAnchorLayerModel.setActive(!this._alignmentAnchorModel.isCurrentActive());

	this._setReferenceOffset();
};

/**
 * Initialize the anchor tool
 */
CompareAlignmentWizard.prototype._initializeAnchorTool = function() {
	this._anchorTool = new CompareAnchorTool(
		this.f,
		this._alignmentAnchorModel,
		this._currentAnchorLayerModel,
		this._referenceAnchorLayerModel
	);
	this.f.emit("toolManager.addTool", this._anchorTool);
};

/**
 * Cancel the wizard
 */
CompareAlignmentWizard.prototype.cancel = function() {
	if (!this._forCadOverlay) {
		this._compareModule.changeReferenceOffset(this._offsetBeforeStart);
	} else {
		this._compareModule.changeCadOverlayOffset(this._offsetBeforeStart);
	}
	this.requestStop();
};

/**
 * Save the alignment (actually, leave it like it is) and stop the viewermode
 */
CompareAlignmentWizard.prototype.alignAndExit = function() {
	this.requestStop();
};

/**
 * Execute an auto-align action
 */
CompareAlignmentWizard.prototype.autoAlign = function() {
	this._compareModule.autoAlign();
};

/**
 * Store the current vector between Middle and the active anchor (for later usage)
 */
CompareAlignmentWizard.prototype._storeMiddleToAnchorOffset = function() {
	var anchor = this._alignmentAnchorModel.getActiveAnchor();
	if (anchor && !this._middleToActiveAnchor) {
		// apply offset from active document to get position in the world
		var docOrigin = this.f.getActivePage().getDocumentPageOrigin();
		anchor.offset(new ShapeUtils.Size(docOrigin.x, docOrigin.y));

		// calculate middle to anchor
		var viewport = this.f.getModule("viewport");
		var middle = viewport.getMiddle();
		var zoom = viewport.getZoom();

		// we store the vector in pixel domain! the zoom level can change when the offsets are updated (and the document gets bigger/smaller)
		this._middleToActiveAnchor = new ShapeUtils.Size((anchor.x - middle.x) * zoom, (anchor.y - middle.y) * zoom);
		this.f.emit("module.graphics.model.needsFragmentCleanup");
	}
};

/**
 * Restore the offset between the active anchor and the middle of the viewport
 * (so that the anchor is displayed on the same location as saved by _storeMiddleToAnchorOffset)
 */
CompareAlignmentWizard.prototype._restoreMiddleToAnchorOffset = function() {
	if (this._middleToActiveAnchor) {
		var anchor = this._alignmentAnchorModel.getActiveAnchor();
		// apply offset from active document to get position in the world
		var docOrigin = this.f.getActivePage().getDocumentPageOrigin();
		anchor.offset(new ShapeUtils.Size(docOrigin.x, docOrigin.y));

		// set middle of the viewport
		var viewport = this.f.getModule("viewport");
		var zoom = viewport.getZoom();
		viewport.transform({
			middle: new ShapeUtils.Point(anchor.x - this._middleToActiveAnchor.width / zoom, anchor.y - this._middleToActiveAnchor.height / zoom),
			animated: false,
			tmpDisableConstraints: true
		});
		this._middleToActiveAnchor = null;
	}
};

/**
 * Is auto align possible?
 * @returns {Boolean}
 */
CompareAlignmentWizard.prototype.isAutoAlignEnabled = function() {
	return !this.f.getActivePage().hasCad() && !this._forCadOverlay;
};

/**
 * Check if the wizard is being used for Cad Overlay alignment
 * @returns {Boolean}
 */
CompareAlignmentWizard.prototype.isForCadOverlay = function() {
	return this._forCadOverlay;
};

/**
 * Enable the "CAD OVERLAY" alignment mode (needs to be done before mode is started
 * @param {Boolean} forCadOverlay - if true, enable cad overlay alignment mode
 */
CompareAlignmentWizard.prototype.setForCadOverlay = function(forCadOverlay) {
	this._forCadOverlay = forCadOverlay;
};

export default CompareAlignmentWizard;
