import AbstractWidget from "@esko/web-viewer/layout/AbstractWidget";
import Template from "@esko/template";
import tpl from "@esko/web-viewer/compare/alignment/CompareAlignmentWizardWidget.tpl.html";
import ViewerToolbarButton from "@esko/web-viewer/layout/component/ViewerToolbarButton";
import ButtonGroup from "@esko/toolbar/ButtonGroup";
import Popover from "@esko/popover/Popover";
import $ from "jquery";
import _ from "underscore";
import CompareConstants from "@esko/web-viewer/compare/CompareConstants";
import LayoutConstants from "@esko/web-viewer/layout/LayoutConstants";
import CoreConstants from "@esko/web-viewer/core/CoreConstants";
import ListenerUtil from "@esko/web-viewer/utils/ListenerUtil";
/**
 * @constant {Number} the speed to render the widget
 */
var RENDER_SPEED = 40;

/**
 * @exports CompareAlignmentWizardWidget
 * @augments Layout.AbstractWidget
 * @class The CompareAlignmentWizardWidget displays the currently loaded document and logged in user
 * @constructor
 * @mixes {Utils.ListenerUtilMixin}
 * @param {Core.ViewerFacade} f
 */
var CompareAlignmentWizardWidget = function CompareAlignmentWizardWidget(f, wizard, anchorTool) {
	AbstractWidget.call(this, f, LayoutConstants.Widget.ALIGNWIZARD);
	ListenerUtil.mixin(this, f);

	this._wizard = wizard;
	this._anchorTool = anchorTool;

	this.templateContext = $("<td/>");

	// initialize template
	this.t = new Template(f.i18n, tpl);
	this.t.d.getExplanationText = function() {
		if (!wizard.isForCadOverlay()) {
			return this.i18n("AnchorAlignWizardExplanation");
		} else {
			return this.i18n("AnchorAlignWizardExplanationCadGraphics");
		}
	};

	this.render = _.throttle(_.bind(this._render, this), RENDER_SPEED);

	// the button bar at the bottom of the screen
	this._buttonBar = $("<td/>").addClass("buttons");

	this._anchorToolButton = new ViewerToolbarButton(this.f, "anchorTool", "icon-WCR_Align-Point", true);
	// tooltip shown when anchor tool is not active
	var tooltip = new Popover(this._anchorToolButton.element, this.f.getContext(), this.f.getContext(), "bottom", "#333", "tooltip", 5);
	this._tooltip = tooltip;
	var tmpAnchorSetActive = this._anchorToolButton.setActive;
	// when the button is inactive (because the tool is inactive) show the popover
	this._anchorToolButton.setActive = _.bind(
		function(widget, isActive) {
			tmpAnchorSetActive.call(this, isActive);
			if (!isActive && widget.isVisible() && widget.isEnabled()) {
				tooltip.show();
			} else {
				tooltip.hide();
			}
		},
		this._anchorToolButton,
		this
	);

	this._anchorToolButton.listenToTool(this._anchorTool, true);
	this._anchorToolButton.setTooltip("viewer", "AnchorAlignWizardAnchorTool");

	this._autoAlignButton = new ViewerToolbarButton(
		this.f,
		"autoAlign",
		"icon-WCR_Align-Auto",
		true,
		_.bind(function() {
			this._wizard.autoAlign();
		}, this)
	);

	this._autoAlignButton.setTooltip("viewer", "AlignAuto");
	this._autoAlignButton.element.addClass("autoAlign");

	this._alignGroup = new ButtonGroup("alignButtons", [this._anchorToolButton, this._autoAlignButton]);
	this._buttonBar.append(this._alignGroup.element);
	// Buttons in the editor
	// Cancel button
	this._cancelButton = new ViewerToolbarButton(
		this.f,
		"cancelAlignment",
		"icon-WCR_Revert",
		true,
		_.bind(function(e) {
			this._tooltip.hide();
			this._wizard.cancel();
		}, this)
	);
	// Save button
	this._alignButton = new ViewerToolbarButton(
		this.f,
		"saveAlignment",
		"icon-WCR_Save",
		true,
		_.bind(function(e) {
			this._tooltip.hide();
			this._wizard.alignAndExit();
		}, this)
	);
	this._alignButton.element.addClass("btn-success");

	// Save and cancel are a buttongroup
	this._cancelSaveGroup = new ButtonGroup("editActionsGroup", [this._cancelButton, this._alignButton]);
	this._buttonBar.append(this._cancelSaveGroup.element);

	// load language if needed
	this.f.loadBundle(
		"viewer",
		function() {
			this._tooltip.setContent($("<div/>").text(this.f.t("AnchorAlignWizardReactivateTool")));
			this._alignButton.setContent($("<span/>").text(this.f.t("AnchorAlignWizardAlign")));
			this._cancelButton.setContent($("<span/>").text(this.f.t("CancelButton")));

			this.render();
		},
		this
	);
};

CompareAlignmentWizardWidget.prototype = Object.create(AbstractWidget.prototype);
CompareAlignmentWizardWidget.prototype.constructor = CompareAlignmentWizardWidget;
CompareAlignmentWizardWidget.widgetName = LayoutConstants.Widget.ALIGNWIZARD;

/**
 * @override
 */
CompareAlignmentWizardWidget.prototype.setContext = function(context) {
	AbstractWidget.prototype.setContext.call(this, context);

	this.context.addClass("enableTextSelection");
	this.context.append(
		$("<table>").append(
			$("<tr>")
				.append(this.templateContext)
				.append(this._buttonBar)
		)
	);

	this.t.setContext(this.templateContext);

	this.context.addClass("CompareAlignmentWizardWidget");
	this.render();
};

/**
 * Update data and render
 */
CompareAlignmentWizardWidget.prototype._render = function() {
	if (!this._wizard.isAutoAlignEnabled()) {
		this._alignGroup.detachButton(this._autoAlignButton);
	} else {
		this._alignGroup.addButton(this._autoAlignButton);
	}
	this.t.render();
};

/**
 * @override
 * @returns {String}
 */
CompareAlignmentWizardWidget.prototype.getWidgetType = function() {
	return "wizard";
};

/**
 * Hide the tooltip when the widget is disabled
 * @override
 */
CompareAlignmentWizardWidget.prototype.setEnabled = function(isEnabled) {
	AbstractWidget.prototype.setEnabled.apply(this, arguments);

	if (!isEnabled) {
		this._tooltip.hide();
	}
};

/**
 * hide the tooltip if the widget is made invisible
 * @override
 * @param isVisible
 */
CompareAlignmentWizardWidget.prototype.setVisible = function(isVisible) {
	AbstractWidget.prototype.setVisible.call(this, isVisible);
	if (!isVisible) {
		// hide the tooltip when the widget is hidden
		this._tooltip.hide();
	}

	this.render();
};

/**
 * hide the tooltip if the widget is collapsed
 * @override
 * @param isVisible
 */
CompareAlignmentWizardWidget.prototype.setCollapsed = function(collapsed) {
	AbstractWidget.prototype.setCollapsed.call(this, collapsed);
	if (collapsed) {
		// hide the tooltip when the widget is hidden
		this._tooltip.hide();
	}
};

export default CompareAlignmentWizardWidget;
