import _ from "underscore";
import AbstractTool from "@esko/web-viewer/viewport/tools/AbstractTool";
import ShapeUtils from "@esko/shapeutils";
import CoreConstants from "@esko/web-viewer/core/CoreConstants";
/**
 * @exports CompareAnchorTool
 * @memberof Compare
 * @class
 * This tool is used to align two documents
 * @constructor
 * @param {Core.ViewerFacade} f - Viewer facade
 * @param {CompareAlignmentAnchorModel} compareAlignmentAnchorModel - model used for storing alignment anchors
 * @param {CompareAlignmentAnchorLayerModel} currentLayerModel - layer model for the current anchor
 * @param {CompareAlignmentAnchorLayerModel} referenceLayerModel - layer model for the reference anchor
 */
var CompareAnchorTool = function(f, compareAlignmentAnchorModel, currentLayerModel, referenceLayerModel) {
	AbstractTool.call(this, f, CoreConstants.TOOLS.COMPARE_ANCHOR_ALIGNMENT);

	/**
	 * The model used for the alignment anchors
	 */
	this._alignmentAnchorModel = compareAlignmentAnchorModel;

	/**
	 * The layer model for the current anchor
	 */
	this._currentLayerModel = currentLayerModel;

	/**
	 * The layer model for the reference anchor
	 */
	this._referenceLayerModel = referenceLayerModel;

	/**
	 * Is the tool used for CAD overlay Alignment
	 **/
	this._forCadOverlay = false;

	/* on Firefox on Mac there is a problem with cursor display none and custom cursors... so we show a default crosshair cursor in this case (by adding MacFF to the ViewerApp)*/
	if (window.mozInnerScreenX !== null && navigator.platform.indexOf("Mac") >= 0) {
		this._macFF = true;
		this.f.getContext().addClass("MacFF");
	} else {
		this._macFF = false;
	}

	/**
	 * DEFAULT STATE
	 */
	this.registerState("default", {
		/*
         * on mouse down, check if the mouse is in either the current or reference panel (and set the correct anchor as active)
         */
		mousedown: _.bind(function(e) {
			var panel = this.getPanelAt(e);
			var panelName = panel.getModel().getName();
			if (panelName === "default") {
				this._alignmentAnchorModel.setCurrentActive(true);
				this.changeState("validPanel");
			} else if (panelName === "compare_difference") {
				this._alignmentAnchorModel.setCurrentActive(false);
				this.changeState("validPanel");
			}
			e.preventDefault();
		}, this),

		mouseup: function() {},

		mousemove: _.bind(function(e) {
			// we don't need to use the result, but the snapping should be indicated
			this._getAndIndicateSnapPoint(e);
		}, this),

		/**
		 * no context menu!
		 */
		contextmenu: _.bind(function(e) {
			e.preventDefault();
			e.stopPropagation();
		}, this)
	});

	/**
	 * validPanel - the interaction is happening in a valid panel, move the anchors
	 */
	this.registerState(
		"validPanel",
		_.defaults(
			{
				mouseup: _.bind(function(e) {
					var anchorPoint = this._getAndIndicateSnapPoint(e);
					if (!anchorPoint) {
						var scenePoint = this.getScenePoint(e);
						anchorPoint = this.convertScenePointToWorldWithCorrectTransform(scenePoint);
					}

					this._alignmentAnchorModel.setActiveAnchorPosition(anchorPoint);
					this.changeState("default");
				}, this),
				mousemove: _.bind(function(e) {
					var anchorPoint = this._getAndIndicateSnapPoint(e);
					if (!anchorPoint) {
						var scenePoint = this.getScenePoint(e);
						anchorPoint = this.convertScenePointToWorldWithCorrectTransform(scenePoint);
					}
					this._liveSetAnchor(anchorPoint);
				}, this)
			},
			this.getState("default")
		)
	);

	this.registerActiveEvents(["mousedown", "mousemove", "mouseup", "contextmenu"]);

	/**
	 * Also support key bindins
	 */
	_.each(
		["up", "down", "left", "right"],
		function(key) {
			this.activeKeyEvents.push([key, _.bind(this._handleArrowKey, this, key), "keyup"]);
		},
		this
	);

	/**
	 * @type {ShapeUtils.Size} the movement the anchor should make at the end of key input
	 */
	this._keyMove = null;

	/**
	 * When the keyinput stopped, move the active anchor
	 */
	this._handleEndOfKeyInput = _.debounce(
		_.bind(function() {
			if (this._keyMove) {
				this._alignmentAnchorModel.moveActiveAnchor(this._keyMove.width, this._keyMove.height);
				this._keyMove = null;
			}

			this.f.emit("viewport.stopInteractive");
		}, this),
		500
	);
};

CompareAnchorTool.prototype = Object.create(AbstractTool.prototype);
CompareAnchorTool.prototype.constructor = CompareAnchorTool;

/**
 * The tool is activated
 * @param interruptRequest
 */
CompareAnchorTool.prototype.doActivation = function(interruptRequest) {
	AbstractTool.prototype.doActivation.call(this, interruptRequest);
	this.changeState("default");
};

/**
 * Get the layer model for the active anchor
 * @returns {CompareAlignmentAnchorLayerModel}
 */
CompareAnchorTool.prototype._getActiveAnchorLayerModel = function() {
	return this._alignmentAnchorModel.isCurrentActive() ? this._currentLayerModel : this._referenceLayerModel;
};

/**
 * Live update the layer model, but not the anchor model
 * @param destination
 */
CompareAnchorTool.prototype._liveSetAnchor = function(destination) {
	var layerModel = this._getActiveAnchorLayerModel();
	layerModel.setAnchor(destination);
};

/**
 * Returns the default cursor for the tool
 * @returns {String}
 */
CompareAnchorTool.prototype.getDefaultCursor = function() {
	return "cursor-anchor";
};

/**
 * Convert a scene point to the correct location relative to (0,0) of the current document
 * @param {ShapeUtils.Point} point
 * @returns {ShapeUtils.Point}
 */
CompareAnchorTool.prototype.convertScenePointToWorldWithCorrectTransform = function(point) {
	var world = this._widget.getWorld();
	/* World transform */
	var t = world.getTransformation();

	var docOrigin = this.f.getActivePage().getDocumentPageOrigin();
	t.translate(docOrigin.x, docOrigin.y);

	return t.createInverse().transformedPoint(point);
};

/**
 * Get the Point to snap to (and indicate it)
 * @param   {Object}           e - mouse event
 * @returns {ShapeUtils.Point} point to select (could be snapped or the original scenepoint) in active page coordinates (relative to the activepage DocumentOrigin)
 */
CompareAnchorTool.prototype._getAndIndicateSnapPoint = function(e) {
	var panel = this.getPanelAt(e);
	if (panel) {
		var panelName = panel.getModel().getName();

		if (panelName === "default" || panelName === "compare_difference") {
			var view = this._widget.requestView(e);
			var refpoint = this.getScenePoint(e);
			if (view) {
				var searchFilter;
				if (panelName === "default") {
					searchFilter = ["Cad"];
					if (!this._forCadOverlay) {
						searchFilter.push("CadOverlay");
					}
				} else if (panelName === "compare_difference") {
					if (this._forCadOverlay) {
						searchFilter = ["CadOverlay"];
					} else {
						searchFilter = ["referenceCad"];
					}
				}
				var point = view.getClosestLineworkPointFromPoint(refpoint, 16, searchFilter);

				if (point) {
					point = this.convertScenePointToWorldWithCorrectTransform(point); // convert to world coordinates (relative to the activepage DocumentOrigin)
					this._currentLayerModel.indicateSnapPoint(panelName === "default" ? point : null);
					this._referenceLayerModel.indicateSnapPoint(panelName === "compare_difference" ? point : null);
					this.setCursor("cursor-none");
				} else {
					point = this.convertScenePointToWorldWithCorrectTransform(refpoint);
					this._currentLayerModel.indicateSnapPoint(null);
					this._referenceLayerModel.indicateSnapPoint(null);
					this.setCursor("cursor-anchor");
				}
				return point;
			}
		}
	}
	return null;
};

/**
 * An arrow key is pressed, update the keymove size, and launch the debounced handleendofkeyinput
 * @param key
 * @param e
 */
CompareAnchorTool.prototype._handleArrowKey = function(key, e) {
	if (!this._keyMove) {
		this.f.emit("viewport.startInteractive");
		this._keyMove = new ShapeUtils.Size(0, 0);
	}

	// we want to move the anchor with 1 pixel, so based on the zoomlevel we know how much mm in the document it is
	var distanceInMM = 1 / this._widget.getWorld().getZoom();

	var keyIndex = _.indexOf(["up", "right", "down", "left"], key);

	var rotation = this._widget.getWorld().getRotation() % 360;
	var flip = this._widget.getWorld().isFlip();
	var isViewFromBack = this._widget.getWorld().isFlipViewFromBack();
	if (isViewFromBack) {
		if (rotation === 90 || rotation === 270) {
			rotation = (rotation + 180) % 360;
		}
		flip = !flip;
	}

	// reverse rotation
	keyIndex = (keyIndex + 4 - ((rotation / 90) % 4)) % 4;

	// reverse flip
	if (flip) {
		if (keyIndex == 1) {
			keyIndex = 3;
		} else if (keyIndex == 3) {
			keyIndex = 1;
		}
	}

	var x = 0,
		y = 0;

	switch (keyIndex) {
		case 0:
			// UP
			y -= distanceInMM;
			break;
		case 1:
			// RIGHT
			x += distanceInMM;
			break;
		case 2:
			// DOWN
			y += distanceInMM;
			break;
		case 3:
			// LEFT
			x -= distanceInMM;
			break;
	}

	// update the layer model
	var layerModel = this._getActiveAnchorLayerModel();
	layerModel.moveAnchor(x, y);

	// update key move
	this._keyMove.width += x;
	this._keyMove.height += y;

	// launch debounced handle end of key input
	this._handleEndOfKeyInput();
};

/*
 * Is the tool being used for CAD overlay or not?
 * @param {Boolean} forCadOverlay
 */
CompareAnchorTool.prototype.setForCadOverlay = function(forCadOverlay) {
	this._forCadOverlay = forCadOverlay;
};

/**
 * @override
 */
CompareAnchorTool.prototype.getIconOffset = function() {
	if (!this._macFF) {
		return new ShapeUtils.Size(5, 5);
	} else {
		// default crosshair is shown, no icon offset
		return new ShapeUtils.Size(0, 0);
	}
};

export default CompareAnchorTool;
