import AbstractDataAdapter from "@esko/web-viewer/data/AbstractDataAdapter";
import Extendable from "@esko/web-viewer/core/Extendable";
import _ from "underscore";
/**
 * Create new AbstractConfig
 *
 * @exports AbstractConfig
 * @memberof Core
 * @class Configuration for Web Viewer. A configuration provides the correct DataAdapter (to interact with the server) and sets up the debugger
 * @param {User} user
 */
var AbstractConfig = function(user) {
	Extendable.installTo(this);

	this._properties = {};
	this._user = user;
	this._modules = {};
	this._layout = {};

	// add mandatory modules
	this.addModule("featureDetector", () => import("@esko/web-viewer/featureDetector/FeatureDetectorModule"), true);
	this.addModule("message", () => import("@esko/web-viewer/message/MessageModule"), true);
	this.addModule("svg", () => import("@esko/web-viewer/svg/SvgModule"), true);
};

/**
 * Set a property in the config
 *
 * @param {String} key - The name of the property
 * @param {Any} value - The value of the property
 */
AbstractConfig.prototype.set = function(key, value) {
	this._properties[key] = value;
};

/**
 * Get a property from the config
 *
 * @param {String} key - The name of the property for which we want the value
 * @returns {Any} The value of the property
 */
AbstractConfig.prototype.get = function(key) {
	return this._properties[key];
};

/**
 * Initialize the DataAdapter
 *
 * (This method should be overridden by the a specific DataAdapter)
 *
 * @param {Core.ViewerFacade} f
 */
AbstractConfig.prototype.initializeAdapter = function(f) {
	this.adapter = new AbstractDataAdapter(f);
	return this.adapter;
};

/**
 * Get the user who is logged in
 *
 * @returns {User} the currently logged in user
 */
AbstractConfig.prototype.getUser = function() {
	return this._user;
};

/**
 * Get the modules to load
 *
 * @return {Array.<Object>} array of modules to load
 */
AbstractConfig.prototype.getModules = function() {
	return _.map(
		_.keys(this._modules),
		function(key) {
			return this._modules[key];
		}.bind(this)
	);
};

/**
 * Get a specific modules from the config
 *
 * @return {AbstractModule} the module if it exists
 */
AbstractConfig.prototype.findModule = function(name) {
	return this._modules[name];
};

/**
 * Add a module to be loaded.
 * @param {String} name 				- name of the module
 * @param {String | Function} [path] 	- path to the module or a constructor function for the module. 
                                          if no path is specified the module will be loaded from a specific file structure
 * @param {Boolean} [lazy = false] 		- whether or not the module should be loaded on startup or only when another module needs it.
 */
AbstractConfig.prototype.addModule = function(name, path, lazy) {
	var moduleInfo = {
		name: name,
		path: path,
		lazy: lazy
	};
	this._modules[name] = moduleInfo;
	return {
		setLazy: function(lazy) {
			moduleInfo.lazy = lazy;
			return this;
		},
		setArguments: function() {
			moduleInfo.arguments = Array.prototype.slice.call(arguments);
			return this;
		}
	};
};

/**
 * Initialize this adapter
 * @param {ViewerCore} f
 * @returns {DataAdapter} Initialized adapter
 */
AbstractConfig.prototype.getCustomModule = function(name) {
	return null;
};

/**
 * Registers an implementation for a given layouttype. This allows to override functionality of the Layout.
 *
 * @param {LayoutConstants.LayoutType} layoutType
 * @param {String} implementationPath
 */
AbstractConfig.prototype.addLayout = function(layoutType, implementationPath) {
	var layout = {
		path: implementationPath
	};
	this._layout[layoutType] = layout;
	return {
		setArguments: function() {
			layout.arguments = Array.prototype.slice.call(arguments);
			return this;
		}
	};
};

/**
 * Fetches an implementation for the given layouttype
 *
 * @param {LayoutConstants.LayoutType} layoutType
 * @returns The path that points towards an implementation of the layout
 */
AbstractConfig.prototype.findLayout = function(layoutType) {
	return this._layout[layoutType];
};

export default AbstractConfig;
