import PrivateChannel from "@esko/web-viewer/core/PrivateChannel";
import _ from "underscore";
/**
 * @memberof Core
 * @class AbstractModel - abstract model, providing base model functionality
 * @constructor
 * @mixes Core.PrivateChannel
 * @param {Core.ViewerFacade} f
 * @param {String} channelName - the channel on which the model will send out events
 */
var AbstractModel = function(f, channelName) {
	this.f = f;
	this.channelName = channelName;

	PrivateChannel.installTo(this, this.f, this.channelName);
};

/**
 * Fire a model change, if not in silent mode and oldValue and newValue are not the same.
 * The even will be fired on channelName.changed.[what] if oldValue and newValue are not the same
 *
 * @param {Boolean} silent - if true no event will be fired
 * @param {Array.<String>} what - specific description of which property has changed, this is an array i.e. ["channel", channelID, "enabled"] the event will be fired in the channel privateChannel.changed.channel.channelID.enabled
 * @param {Any} oldValue - the old value before the change
 * @param {Any} newValue - the new value after the change
 * @param {Any} extraArguments - Events can be fired with extra arguments, just put them after newValue
 */
AbstractModel.prototype._fireChange = function(silent, what, oldValue, newValue) {
	if (!silent) {
		var didChange = false;

		if (oldValue === newValue) {
			didChange = false; // identical objects are always the same... (this is check is here for "extra" performance)
		} else if (
			newValue !== undefined &&
			oldValue !== undefined &&
			newValue !== null &&
			oldValue !== null &&
			_.isFunction(newValue.equals) &&
			newValue.constructor === oldValue.constructor
		) {
			// an equals function was defined, and both objects are of the same type
			didChange = !oldValue.equals(newValue);
		} else if (_.isArray(oldValue) && _.isArray(newValue)) {
			// array comparison
			if (oldValue.length === newValue.length) {
				for (var i = 0; i < oldValue.length && !didChange; i++) {
					if (oldValue[i] != newValue[i]) {
						didChange = true;
					}
				}
			} else {
				// non equal length arrays
				didChange = true;
			}
		} else {
			didChange = oldValue !== newValue;
		}

		// if there is an equals method, use this for comparison, otherwise compare the primitive types
		if (didChange) {
			arguments[1].unshift("changed");
			this._emit.apply(this, _.rest(arguments));
		}
	}
};

/**
 * Install the _fireChange method in to an object, it will use the privateChannel of object, append "channelname" and what to it
 * @param {Object} object - the object on which the _fireChange method should be installed
 * @param {Array.<String>} channelName - the name of the channel for events --> [privateChannel, channelName, what]
 */
AbstractModel.mixin = function(object) {
	object._fireChange = function(silent, what, oldValue, newValue) {
		AbstractModel.prototype._fireChange.apply(this, silent, what, oldValue, newValue);
	};
};

export default AbstractModel;
