import Extendable from "@esko/web-viewer/core/Extendable";
import PrivateChannel from "@esko/web-viewer/core/PrivateChannel";
import _ from "underscore";
/**
 * Initialize the module
 * @exports AbstractModule
 * @memberof Core
 * @class
 * AbstractModule is an abstract class for defining a module
 *
 * A module is a piece of functionality that is loaded once, only 1 instance of the module is available in the system
 * The AbstractModule creates a private channel ("module.[name]") on which messages can be put. And foresees some methods for basic interaction
 *
 * The name of a module is unique in the system, a module is a singleton (and can thus only be instantiated once!)
 * @constructor
 * @mixes Core.PrivateChannel
 * @mixes Core.Extendable
 * @param {Core.ViewerFacade} f
 * @param {Boolean} silenced - if true, the module shouldn't publish tools/widgets/toolbarcomponents
 * @param {String} name - unique name of the module
 * @param {Object} dataAdapter - specific data adapter for this module
 */
var AbstractModule = function(f, silenced, name, dataAdapter) {
	this.f = f;
	this.silenced = silenced || false;
	this.name = name;
	// in here all publish events reside until the module is unsilenced
	this._onUnsilence = [];

	Extendable.installTo(this);
	PrivateChannel.installTo(this, f, "module." + name);

	if (dataAdapter) {
		this.f.getDataAdapter().registerDataAdapterForModule(name, dataAdapter);
	}
};

/**
 * Publish an event as soon as the module is unsilenced
 *
 * @private
 * @param {String|Array} channel
 * @param {...Any} arguments
 */
AbstractModule.prototype._publish = function() {
	if (this.silenced) {
		this._onUnsilence.push(arguments);
	} else {
		this.f.emit.apply(this.f, arguments);
	}
};

/**
 * Unsilence the module
 *
 * All outstanding events are emitted
 *
 * @private
 */
AbstractModule.prototype._unSilence = function() {
	// Publish all waiting events
	_.each(
		this._onUnsilence,
		_.bind(function(args) {
			this.f.emit.apply(this.f, args);
		}, this)
	);
	// Clear the waiting events list
	this._onUnsilence = [];
	this.silenced = false;
};

/**
 * Set the silenced mode of the module
 * If a module is silenced, it won't add widgets/toolbarcomponents/tools
 * @param {Boolean} silenced
 */
AbstractModule.prototype.setSilenced = function(silenced) {
	if (silenced) {
		this.silenced = true;
	} else {
		this._unSilence();
	}
};

/**
 * Publish a widget as soon as the module is unsilenced
 * @protected
 * @param {LayoutConstants.Widget} name - the name of the widget
 * @param {Function} widgetFactory - factory function to instantiate the widget (with 1 parameter: LayoutConstants.LayoutType)
 */
AbstractModule.prototype._publishWidget = function(name, widgetFactory) {
	this._publish("layout.addWidget", name, widgetFactory);
};

/**
 * Publish a toolbarcomponent as soon as the module is unsilenced
 * @protected
 * @param {LayoutConstants.ToolbarComponent} name - the name the toolbarcomponent
 * @param {Function} toolbarComponentFactory - factory function to instantiate the toolbarcomponent (with 1 parameter: LayoutConstants.LayoutType)
 */
AbstractModule.prototype._publishToolbarComponent = function(name, toolbarComponentFactory) {
	this._publish("layout.addToolbarComponent", name, toolbarComponentFactory);
};

/**
 * Publish a tool as soon as the module is unsilenced
 * @protected
 * @param {AbstractTool} tool - The tool should be published
 */
AbstractModule.prototype._publishTool = function(tool) {
	this._publish("toolManager.addTool", tool);
};

/**
 * Unpublish a tool
 * @protected
 * @param {String} tool - name of the tool
 */
AbstractModule.prototype._unpublishTool = function(tool) {
	this._publish("toolManager.removeTool", tool);
};

/**
 * Check if the module has unsaved elements, return false by default
 * @returns {Boolean}
 */
AbstractModule.prototype.hasUnsavedElements = function() {
	return false;
};

/**
 * The (localized) string representing the unsaved elements
 * @returns {String}
 */
AbstractModule.prototype.getUnsavedElementsName = function() {
	return this.name;
};

/**
 * Saves all unsaved elements of a module
 * @param {Function} [callback] - called when the save action is complete
 */
AbstractModule.prototype.saveAllUnsavedElements = function(callback) {
	if (callback) {
		callback();
	}
};

/**
 * Get the state of a module so it can be restored afterwards via setState.
 * This are typically filters of the widget, options of the models, etc.
 * @returns {Object} object containing the state of the module
 */
AbstractModule.prototype.getState = function() {
	// to be implemented by the module
};

/**
 * Restore the state of a module as obtained by getState
 * @param {Object} state - The state to restore of the module
 */
AbstractModule.prototype.setState = function(state) {
	// to be implemented by the module
};

export default AbstractModule;
