
/**
 * @namespace Core
 */

import ViewerFacade from "@esko/web-viewer/core/ViewerFacade";
import DocumentController from "@esko/web-viewer/data/DocumentController";
import LayoutManager from "@esko/web-viewer/layout/LayoutManager";
import FeatureDetectorModule from "@esko/web-viewer/featureDetector/FeatureDetectorModule";
import MessageModule from "@esko/web-viewer/message/MessageModule";
import ViewportModule from "@esko/web-viewer/viewport/ViewportModule";
import ViewerToolbarButton from "@esko/web-viewer/layout/component/ViewerToolbarButton";
import ModalPopup from "@esko/web-viewer/layout/component/ModalPopup";
import LayoutConstants from "@esko/web-viewer/layout/LayoutConstants";
import loaderTransparentGif from "@esko/web-viewer/layout/img/loaderTransparent.gif";
import ViewerModeManager from "@esko/web-viewer/core/ViewerModeManager";
import WebviewerModule from "@esko/web-viewer/webviewer/WebviewerModule";
import Debug from "@esko/web-viewer/core/DevDebug";
import ImportLoadingStrategy from "@esko/web-viewer/core/ImportLoadingStrategy";
import ObjectLoadingStrategy from "@esko/web-viewer/core/ObjectLoadingStrategy";

import LoadingIndicator from "@esko/loadingindicator/LoadingIndicatorSingleton";
import EventEmitter2 from "eventemitter2";
import $ from "jquery";
import _ from "underscore";
/**
 * EVENTS
 *
 *
 * module.request
 *
 * 		Request the instance of a module, if the module isn't loaded, it will be loaded, the callback function is called with the module
 *
 *
 * 		@param {String} name - The name of the module
 *		@param {Function} callback - Function which is called when the module is available (or couldn't be loaded), The first argument of this callback is the instance of the module (or null)
 * 		@param {Boolean} [silenced] - If true and the module is not loaded yet, then the module won't emit widgets/tools/toolbarcomponents until it loaded with silenced false/undefined or is unsilenced
 */

/**
 * The ViewerCore is the core of the application.
 *
 * It provides an Event Bus for communication between modules, all interaction with the core happens through the Facade.
 * The ViewerCode can load widgets and modules
 * @exports ViewerCore
 * @memberof Core
 * @constructor
 * @param {JQueryObj} context - the HTML object in which the viewer application should reside
 * @param {Core.AbstractConfig} config - configuration class
 */
var ViewerCore = function(context, config) {
	context = $(context);

	var initialLoader = $("<div>")
		.css({
			background: "url('" + loaderTransparentGif + "') no-repeat 2px center",
			"padding-left": "20px",
			font: "normal 11px/20px Verdana,Arial,Helvetica,sans-serif",
			margin: "20px",
			height: "20px"
		})
		.text("Loading...");
	context.empty().append(initialLoader);

	var maxListeners = 0; // in production environment we don't want to generate any memory leak warnings on the event bus (0=unlimited)

	if (process.env.NODE_ENV === "development") {
		// during development we want maximum 100 listeners per channels, in production mode this can be much more
		maxListeners = 100;
	}

	// Create event bus
	var ebus = new EventEmitter2({
		wildcard: true, // should the event emitter use wildcards.
		delimiter: ".", // the delimiter used to segment namespaces,
		newListener: true // if you want to emit the newListener event
	});

	ebus.setMaxListeners(maxListeners); // the max number of listeners that can be assigned to an event

	var modulesToLoadCount = -1;

	/**
	 * The coreAPI is the API the facade can use to interact with the core
	 * but all other components can only communicate with the core through the
	 * facade
	 */
	var coreAPI = {
		getProperty: function(key) {
			return config.get(key);
		},
		getConfig: function() {
			return config;
		},
		getAppContext: function() {
			return context;
		}
	};

	/**
	 * The currently logged in user.
	 * This variable is initialized later during the initialization phase
	 */
	var user;
	/**
	 * The dataAdapter.
	 * This variable is initialized later during initialization
	 */
	var dataAdapter;
	/**
	 * The documentController
	 * This variable is initialized later during initalization
	 */
	var docController;

	// initialize debugger
	var debug = new Debug(ebus);
	debug.log("CORE", "Debugger activated", debug);

	// initialize facade
	debug.log("CORE", "Initializing ViewerFacade", debug);
	var f = new ViewerFacade(coreAPI, ebus, debug);

	// create the loading indicator (we do this here, because we want it to listen before all loads start)
	// listen to all global loading.start/stop events
	f.on("loading.start", function(what) {
		LoadingIndicator.start(what);
	});
	f.on("loading.stop", function(what) {
		LoadingIndicator.stop(what);
	});

	if (process.env.NODE_ENV === "development") {
		// only expose the ViewerFacade and loading indicator during development
		window.gF = f;
		debug.log("CORE", "The ViewerFacade is now exposed as window.gF");

		window.gLoadingIndicator = LoadingIndicator;
		debug.log("CORE", "The loading indicator is now exposed as window.gLoadingIndicator");
	}

	// initialize viewermode manager
	debug.log("CORE", "Initializing viewerModeManager");
	var viewerModeManager = new ViewerModeManager(f);
	var isWaitingForCompare = false;

	coreAPI.stopMode = function(viewerMode) {
		return viewerModeManager.stopMode(viewerMode);
	};

	coreAPI.startMode = function(viewerMode) {
		return viewerModeManager.startMode(viewerMode);
	};

	coreAPI.stopActiveMode = function() {
		return viewerModeManager.stopActiveMode();
	};

	coreAPI.getActiveMode = function() {
		return viewerModeManager.getActiveMode();
	};

	coreAPI.canFetchFragments = function() {
		return !docController.isParameterUpdateOngoing() && !isWaitingForCompare;
	};

	// initialize module management
	var loadedModules = {};
	var loadingModules = {};
	var loadingSilencedModules = {};

	/**
	 * Initialize a module
	 *
	 * @param {Core.AbstractModule} Module - module constructor
	 * @param {String} name - the name of the module
	 * @param {Boolean} silenced - initialize module in silenced mode
	 * @returns {Core.AbstractModule} instantiated module
	 */
	function initializeModule(Module, name, silenced) {
		ebus.emit(["module", name, "startInitialization"]);
		var moduleInstance;
		if (loadedModules.hasOwnProperty(name)) {
			moduleInstance = loadedModules[name];
		} else {
			moduleInstance = new Module(f, silenced);
			loadedModules[name] = moduleInstance;
		}
		// notify possible listeners
		ebus.emit(["module", "loaded", name], moduleInstance);

		return moduleInstance;
	}

	/**
	 * Request the instance of a module, if the module isn't loaded, it will be loaded, the callback function is called with the module
	 * (or null if the module couldn't be loaded)
	 *
	 * @param {String} name - The name of the module
	 * @param {Function} [callback] - Function which is called when the module is available (or couldn't be loaded), The first argument of this callback is the instance of the module (or null)
	 * @param {Boolean} [silenced] - If true and the module is not loaded yet, then the module won't emit widgets/tools/toolbarcomponents until it loaded with silenced false/undefined or is unsilenced
	 */
	function requestModule(name, callback, silenced) {
		silenced = silenced || false;
		if (loadedModules.hasOwnProperty(name)) {
			// the module is already loaded
			var module = loadedModules[name];

			// unsilence it if needed
			if (!silenced) {
				module.setSilenced(false);
			}

			// call the callback if available
			if (callback) {
				callback(module);
			}
		} else if (loadingModules.hasOwnProperty(name)) {
			// the module is already being loaded, add callback if available
			if (loadingSilencedModules.hasOwnProperty(name) && !silenced) {
				delete loadingSilencedModules[name];
			}

			// if callback available, push it on the list
			if (callback) {
				loadingModules[name].push(callback);
			}
		} else {
			// the module is not yet loaded, request it, and register callback if any
			loadingModules[name] = [];

			if (silenced) {
				// the module should be loaded silenced
				loadingSilencedModules[name] = 1;
			}

			if (callback) {
				loadingModules[name].push(callback);
			}

			ebus.emit(["module", "loading", name]);

			var loadInfo = config.findModule(name);

			var loadingStrategy;
			if (!loadInfo || !loadInfo.path) {
				if (process.env.NODE_ENV === "development") {
					console.error('Could not find module "' + name + '" in the config.');
				}
				return callback && callback(null);
			}
			if (typeof loadInfo.path === "string") {
				throw new Error("Trying to load module with path: " + loadInfo.path);
			} else if (typeof loadInfo.path === "function") {
				loadingStrategy = new ImportLoadingStrategy(f, ebus); // if a path was specified, load the module based on that path
			} else {
				loadingStrategy = new ObjectLoadingStrategy(f, ebus); // a constructor function is expected.
			}

			loadingStrategy.load(loadInfo, function(Mod, err) {
				docController.whenActivePageChanged(
					_.once(function() {
						//the module is probably loaded!
						var moduleInstance = null;
						if (Mod) {
							// should the module be silenced?
							var silenced = loadingSilencedModules.hasOwnProperty(name);

							// there is an object, instantiate it with the facade
							moduleInstance = initializeModule(Mod, name, silenced);
						} else if (err) {
							// something went wrong while loading the module (module probably doesn't exist)
							console.error(err);
						} else {
							// No error obj and module doesn't exist, something is terribly wrong.
							ebus.emit("error.SEVERE", "Failed to load module " + name);
						}

						// now notify all registered callbacks about the result
						var callbacks = loadingModules[name];
						for (var i = 0; i < callbacks.length; i++) {
							callbacks[i](moduleInstance);
						}

						// delete the registered callbacks, the module is loaded (or has failed loading)
						delete loadingModules[name];
						delete loadingSilencedModules[name];
					})
				);
			});
		}
	}
	// for now the requestModule shouldn't work yet
	coreAPI.requestModule = function() {};

	/**
	 * Do an Ajax call
	 * @param {String} url - The URL to call
	 * @param {Object} options - Object containing the options for the ajax request. See jQuery API for the meaning
	 * @example
	 * defaultOptions = {
	 *	dataType: "xml",
	 *		success: function(data, textStatus, jqXHR) {},
	 *		error: function(jqXHR, textStatus, errorThrown) {},
	 *		cache: false,
	 *		processData: true,
	 *		contentType: "application/x-www-form-urlencoded; charset=UTF-8",
	 *		type: "GET",
	 *		data: "",
	 *		noAjaxParam: false // it will automatically add ajax=1 to the param list if this is false, setting this to true will NOT add it
	 *	};
	 *
	 */
	coreAPI.ajax = function(url, options) {
		// TODO BDGE replace all ajax calls in modules with a call to "connection" module
		var deferred = $.Deferred();
		requestModule("connection", function(connectionModule) {
			deferred.resolve(connectionModule.ajax(url, options));
		});
		return deferred.promise();
	};

	/**
	 * Executes the callback when all modules are loaded
	 * @param {Function} callback
	 */
	coreAPI.whenAllModulesAreLoaded = function(callback) {
		if (modulesToLoadCount === 0) {
			callback();
		} else {
			f.once("allModulesAreLoaded", callback);
		}
	};

	/**
	 * Returns the module is it is already loaded
	 * @param {String} name - name of the module
	 * @returns {Core.AbstractModule}
	 */
	coreAPI.getModuleIfAvailable = function(name) {
		return loadedModules[name];
	};

	/**
	 * Executes the callback when the module with the given name is loaded.
	 * If it is already there, the callback is immediately executed, otherwise it waits until it is loaded.
	 * @param {String} name - Name of the module
	 * @param {Function} callback - Parameters: moduleInstance
	 * @param {Any} [binding] - Context binding of the callback
	 */
	coreAPI.whenModuleAvailable = function(name, callback, binding) {
		if (binding !== undefined) {
			callback = _.bind(callback, binding);
		}

		var module = loadedModules[name];
		if (module !== undefined) {
			// it is already available
			callback(module);
		} else {
			ebus.once(["module", "loaded", name], callback);
		}
	};

	/**
	 * Get the loaded modules
	 * @returns {Object}
	 */
	coreAPI.getLoadedModules = function() {
		return loadedModules;
	};

	/**
	 * Inject state to the viewer
	 * @param {Object} state
	 */
	coreAPI.injectState = function(state) {
		coreAPI.whenAllModulesAreLoaded(
			_.bind(function() {
				//wait for all modules to be loaded
				if (state.compare !== undefined) {
					// first inject state for the compare module if it is not undefined
					var compareModule = loadedModules["compare"];
					if (compareModule !== undefined) {
						compareModule.setState(state.compare);
					}
					delete state.compare; //set to undefined to make sure it doesn't get injected again
				}
				var measureState = state.measure;
				delete state.measure;
				_.each(
					state,
					_.bind(function(moduleState, moduleName) {
						//inject state for all the other modules
						if (moduleState !== undefined) {
							//first make sure there is a state to inject
							coreAPI.whenModuleAvailable(
								moduleName,
								function(module) {
									module.setState(moduleState);
								},
								this
							);
						}
					}, this)
				);
				if (measureState !== undefined) {
					var measureModule = loadedModules["measure"];
					if (measureModule !== undefined) {
						measureModule.setState(measureState);
					}
				}
			}, this)
		);
	};

	/**
	 * Get the state of the current viewer, will ask the state of all loaded modules
	 * @returns {Object}
	 */
	coreAPI.getState = function() {
		var modules = this.getLoadedModules();
		var viewerState = {};
		_.each(modules, function(module, moduleName) {
			viewerState[moduleName] = module.getState();
		});
		return viewerState;
	};

	debug.debug("CORE", "Initializing Message module");
	// we first start by loading the message module
	if (!MessageModule) {
		window.alert("Couldn't load Message module.");
	}

	// initialize message module instance
	var messageModule = initializeModule(config.getCustomModule("message") || MessageModule, "message", true);

	// the message module is loaded (or it failed, and the user was warned. now we start initializing other components

	// Initialize user
	debug.debug("CORE", "Initializing user object");
	user = config.getUser();
	user.setFacade(f);
	coreAPI.getCurrentUser = function() {
		return user;
	};

	// setup data adapter
	debug.debug("CORE", "Initializing DataAdapter");
	dataAdapter = config.initializeAdapter(f);

	coreAPI.getDataAdapter = function() {
		return dataAdapter;
	};

	// initialize i18n and start loading general bundle
	debug.debug("CORE", "Initializing I18N");
	f.i18n = dataAdapter.getI18nProvider();
	f.i18n.loadBundle(["viewer", "messagewidget"], function() {
		debug.debug("I18N", "General language bundle loaded");
		initialLoader.text(f.i18n.t("Loading"));
	});

	// setup document controller
	debug.debug("CORE", "Initializing DocumentController");
	docController = new DocumentController(f, dataAdapter);

	/**
	 * Get the DocumentController
	 *
	 * @return {DocumentController}
	 */
	coreAPI.getDocumentController = function() {
		return docController;
	};

	/**
	 * Open document(s)
	 * @param {any} documentInformation - String or JSON object containing information on which document to load
	 */
	coreAPI.openDocuments = function openDocuments(documentsIdentifier) {
		docController.openDocuments(documentsIdentifier);
	};

	/**
	 * Called before the viewer is unloaded
	 */
	function onBeforeUnload() {
		var unsavedItems = "";
		_.each(loadedModules, function(module) {
			if (module.hasUnsavedElements()) {
				unsavedItems += "\r\n" + module.getUnsavedElementsName();
			}
		});

		if (unsavedItems.length > 0) {
			return f.i18n.t("ExitViewer") + unsavedItems;
		}
	}

	// to be sure we have strings available
	f.i18n.loadBundle("viewer", function() {
		$(window).on("beforeunload", onBeforeUnload);
	});

	f.on("viewer.request.close", function(returnURL, saveBeforeExitKey, exitWithoutSaveKey) {
		var unsavedModules = {};
		_.each(loadedModules, function(module, name) {
			if (module.hasUnsavedElements()) {
				unsavedModules[name] = module;
			}
		});

		saveBeforeExitKey = saveBeforeExitKey ? saveBeforeExitKey : "SaveAndExit";
		exitWithoutSaveKey = exitWithoutSaveKey ? exitWithoutSaveKey : "ExitWithoutSave";

		// temporary function to go the the return url
		var goToReturnURL = function() {
			if (window.location === returnURL) {
				window.location.reload();
			} else {
				window.location = returnURL;
			}
		};

		if (!_.isEmpty(unsavedModules)) {
			f.i18n.loadBundle("viewer", function() {
				var modalPopup = new ModalPopup(f, "unsavedItemsPopup", f.getContext(), undefined, undefined, {
					okButtonSahiID: "ConfirmSaveUnsavedChanges",
					cancelButtonSahiID: "CancelSaveUnsavedChanges"
				});
				modalPopup.setRemoveFromDomAfterUsage(true);
				modalPopup.setOkBtnText(f.i18n.t(saveBeforeExitKey));

				// extra button
				var exitBtn = new ViewerToolbarButton(f, "exit", $("<span/>").text(f.i18n.t(exitWithoutSaveKey)), false, function() {
					$(window).off("beforeunload", onBeforeUnload);
					goToReturnURL();
				});
				exitBtn.element.get(0).dataset.sahiId = "DiscardUnsavedChanges";
				modalPopup.addButton(exitBtn);

				// cancel button
				// nothing extra needed

				// ok button
				modalPopup.setOkAction(function() {
					var callback = function(name) {
						delete unsavedModules[name];
						if (_.isEmpty(unsavedModules)) {
							$(window).off("beforeunload", onBeforeUnload);
							goToReturnURL();
						}
					};
					if (!_.isEmpty(unsavedModules)) {
						modalPopup.enableOkBtn(false);
						_.each(unsavedModules, function(module, name) {
							module.saveAllUnsavedElements(function() {
								return callback(name);
							});
						});
					}
				});

				// Header + Content
				modalPopup.setHeader(
					$("<span/>")
						.addClass("popupTitle")
						.html(f.i18n.t("UnsavedItems"))
				);
				modalPopup.setContent(
					$("<div/>")
						.addClass("unsavedItemsContent")
						.append(f.i18n.t("UnsavedItemsContent"))
				);

				modalPopup.show();
				modalPopup.recalculateSize(400);
			});
		} else {
			goToReturnURL();
		}
	});

	// now start listening for module requests
	debug.debug("CORE", "Core is now listening to module requests");
	// requesting modules should also be possible from the facade
	coreAPI.requestModule = requestModule;

	// Load user options
	user.loadOptions(function() {
		// initialize feature Detector
		debug.debug("CORE", "Initializing Feature detector module");
		// initialize feature detector module instance
		var featureDetector = initializeModule(config.getCustomModule("featureDetector") || FeatureDetectorModule, "featureDetector", true);
		coreAPI.getFeatureDetector = function() {
			return featureDetector;
		};

		// Check if layout type is requested through the config
		var requestedLayoutType = config.get("ViewerCore.layouttype");
		if (requestedLayoutType) {
			layoutType = requestedLayoutType;
		} else {
			debug.debug("CORE", "Detecting layout type");
			// detect layout type
			var layoutType = user.getOption("LAYOUT_PREFERENCE");
			if (
				(featureDetector.isMobile() && featureDetector.hasTouch()) ||
				(layoutType === LayoutConstants.LayoutType.MOBILE && featureDetector.hasTouch()) ||
				(layoutType === LayoutConstants.LayoutType.AUTOMATIC && featureDetector.isMobile())
			) {
				// if preference is "automatic": use mobile if device is mobile detected. If preference is "mobile": only show mobile layout if touch support is enabled
				layoutType = LayoutConstants.LayoutType.MOBILE;
			} else {
				// Desktop, check which style
				var configViewerStyle = coreAPI.getProperty("viewerStyle"); // launch config takes priority
				if (
					configViewerStyle == LayoutConstants.ViewerStyle.SIMPLIFIED ||
					(!configViewerStyle && layoutType == LayoutConstants.LayoutType.FLEX)
				) {
					layoutType = LayoutConstants.LayoutType.FLEX;
				} else {
					layoutType = LayoutConstants.LayoutType.DESKTOP;
				}
			}
		}

		/**
		 * Return the type of layout we are in
		 * @returns {LayoutConstants.LayoutType}
		 */
		coreAPI.getLayoutType = function() {
			return layoutType;
		};

		/**
		 * Returns the layoutManager
		 *
		 * @returns
		 */
		coreAPI.getLayoutManager = function() {
			return layoutManager;
		};

		/**
		 *  Returns the active layout mode
		 */
		coreAPI.getLayoutMode = function() {
			return layoutManager.getLayoutMode();
		};

		// if compare is activated, the viewport module needs a repaint after loading
		// after 2: we need to wait for both the layoutmanager to be loaded and the documents (if they need "compare" functionality)
		var repaintViewportModuleIfNeeded = _.after(2, function() {
			coreAPI.whenModuleAvailable("viewport", function(viewportModule) {
				viewportModule.whenWidgetModelAvailable(function() {
					requestModule(
						"compare",
						function() {
							isWaitingForCompare = false;
							viewportModule.widget.repaint();
						},
						false
					); //now make it unsilenced to trigger a publish of the widget
				});
			});
		});

		// load the viewport first, then load the other modules
		debug.debug("CORE", "Start loading viewport module");
		if (!ViewportModule) {
			ebus.emit("error.SEVERE", "Couldn't load viewport module!");
		}
		var viewportModule = initializeModule(config.getCustomModule("viewport") || ViewportModule, "viewport", true); // SILENCED (wait till LayoutManager is ready)

		// load the webviewer module, this sets the correct viewportwidget model
		debug.debug("CORE", "Start loading webviewer module, this intializes the viewport widget with the correct widgetmodel");
		var webviewerModule = initializeModule(config.getCustomModule("webviewer") || WebviewerModule, "webviewer", true); // SILENCED (wait till LayoutManager is ready)

		// start loading compare module as soon as possible
		docController.whenDocumentsLoaded(
			function() {
				if (
					docController.getDocumentCount() > 1 ||
					_.some(docController.getDocuments(), function(document) {
						return (
							document.hasLinkedDocument() ||
							document.hasMultipleVersions() ||
							document.getDocumentVersion().isMultiPage() ||
							document.getDocumentVersion().isPageList()
						);
					})
				) {
					isWaitingForCompare = true;
					requestModule(
						"compare",
						function() {
							isWaitingForCompare = false;
						},
						true
					); //true becaused silenced (so no widgets get publised, compare widget needs viewport-module)

					repaintViewportModuleIfNeeded();
				}
			},
			this,
			true
		);

		// Start up layoutmanager
		debug.debug("CORE", "Initializing LayoutManager");
		var layoutManager = new LayoutManager(
			f,
			context,
			LoadingIndicator,
			initialLoader,
			_.bind(function() {
				debug.debug("CORE", "Initialized LayoutManager", layoutManager);

				/**
				 * Switch to another layout
				 * @param {LayoutConstants.LayoutType} layoutType - type of the layout
				 * @returns {Number}
				 */
				coreAPI.switchLayout = function(layoutType) {
					layoutManager.switchLayout(layoutType);
				};

				// unsilence the message module
				if (messageModule) {
					messageModule.setSilenced(false);
				}

				// unsilence viewport and webviewer
				debug.debug("CORE", "Unsilence Viewport and WebViewer module");
				viewportModule.setSilenced(false);
				webviewerModule.setSilenced(false);

				// we are ready for a repaint
				repaintViewportModuleIfNeeded();

				// start loading other modules
				debug.debug("CORE", "Start loading other modules");

				var modulesToLoad = _.chain(config.getModules())
					.reject({
						lazy: true
					})
					.pluck("name")
					.value();
				var whenAllModulesLoaded = _.after(modulesToLoad.length, function() {
					f.emit("allModulesAreLoaded");
				});

				_.each(modulesToLoad, function(moduleName) {
					requestModule(moduleName, whenAllModulesLoaded);
				});
				f.emit("layout.initialized");
			}, this)
		);
	});

	// expose the ViewerFacade as the interface to the Core
	return f;
};

export default ViewerCore;
