import AbstractDebug from "@esko/web-viewer/core/AbstractDebug";
import _ from "underscore";

/**
 * Create new DevDebug instance.
 * @exports DevDebug the devDebug debugger enables sending debugmessages to the console
 * @constructor
 *
 * @param {EventEmitter2} eventBus - the event bus used in the App
 */

var DevDebug = function(eventBus) {
	AbstractDebug.call(this, eventBus);

	// paths the Debug will listen to
	var channels = ["*"];

	/**
	 * This function logs the event information
	 */
	var _logEventFunction = _.partial(function(that) {
		that.debug(new Date().getTime(), "E", this.event, _.tail(arguments, 1));
	}, this);

	// log errors
	eventBus.on(
		"error.*",
		_.bind(function(msg, details) {
			this.error(msg, details);
		}, this)
	);

	/**
	 * Stop listening to channels on the eventBus
	 *
	 * @param {String} [channel] - if undefined: remove all debug listeners for all channels, if string stop listening to the specific channel
	 */
	this.stopListeningTo = function stopListening(channel) {
		if (channel) {
			if (channel == "*") {
				eventBus.offAny(_logEventFunction);
			} else {
				eventBus.off(channel, _logEventFunction);
			}
			this.debug("DEBUG", "stop listening to", channel);
		} else {
			// remove all listeners
			for (var i = 0; i < channels.length; i++) {
				this.stopListeningTo(channels[i]);
			}
			channels = [];
		}
	};

	/**
	 * Start listening to a certain channel on the eventBus
	 *
	 * @param {String} channel
	 */
	this.startListeningTo = function(channel) {
		this.debug("DEBUG", "start listening to", channel);
		if (channel == "*") {
			eventBus.onAny(_logEventFunction);
		} else {
			eventBus.on(channel, _logEventFunction);
		}
	};

	if (process.env.NODE_ENV === "development") {
		// listen to the standard paths
		for (var i = 0; i < channels.length; i++) {
			this.startListeningTo(channels[i]);
		}
	}
};

DevDebug.prototype = Object.create(AbstractDebug.prototype);
DevDebug.prototype.constructor = DevDebug;

// if you want the file name + line numbers next to your log messages, set this to true
var addLineInfo = false;
if (navigator.appName.indexOf("Internet Explorer") != -1) {
	addLineInfo = false;
}

/**
 * Skip certain lines in stack trace containing one of those strings
 */
var filters = ["DevDebug", "underscore", "EventEmitter2", "ViewerFacade", "PrivateChannel"];
/**
 * Extract line number from stack
 * @param stack Stacktrace
 * @returns Stack line where the log function is called.
 */
var extractLineNumberFromStack = function(stack) {
	var lines = stack.split("\n");
	var line;
	for (var i = 1; i < lines.length && !line; i++) {
		var j = 0;
		while (j < filters.length && lines[i].indexOf(filters[j]) < 0) {
			j++;
		}
		// filter not found
		if (j >= filters.length) {
			line = lines[i];
		}
	}
	if (!line) {
		return "";
	} else {
		// fix for various display text
		return line.indexOf(" (") >= 0 ? line.split(" (")[1].substring(0, line.length - 1) : line.split("at ")[1];
	}
};

/**
 * Get the suffix for a line, meaning the file name and line number of the caller of log.
 * @param error Error containing the stacktrace.
 * @returns File name and line number of the caller of log
 */
var getLineSuffix = function(error) {
	return (
		"@" +
		(error.lineNumber
			? error.fileName + ":" + error.lineNumber + ":1" // add arbitrary column value for chrome linking
			: extractLineNumberFromStack(error.stack))
	);
};

/**
 * Same behavior as console.log
 */
DevDebug.prototype.log = function() {
	var args = arguments;
	if (addLineInfo) {
		args = Array.prototype.slice.call(args, 0);
		args = args.concat([getLineSuffix(new Error())]);
	}
	if (console && console.log) {
		console.log.apply(console, args);
	}
};

/**
 * Same behavior as console.warn
 */
DevDebug.prototype.warn = function() {
	var args = arguments;
	if (addLineInfo) {
		args = Array.prototype.slice.call(args, 0);
		args = args.concat([getLineSuffix(new Error())]);
	}
	try {
		console.warn.apply(console, args);
	} catch (e) {
		this.log.apply(this, args);
	}
};

/**
 * Same behavior as console.error
 */
DevDebug.prototype.error = function() {
	var args = arguments;
	if (addLineInfo) {
		args = Array.prototype.slice.call(args, 0);
		args = args.concat([getLineSuffix(new Error())]);
	}
	try {
		console.error.apply(console, args);
	} catch (e) {
		this.log.apply(this, args);
	}
};

/**
 * Same behavior as console.debug
 */
DevDebug.prototype.debug = function() {
	var args = arguments;
	if (addLineInfo) {
		args = Array.prototype.slice.call(args, 0);
		args = args.concat([getLineSuffix(new Error())]);
	}

	try {
		console.debug.apply(console, args);
	} catch (e) {
		this.log.apply(this, args);
	}
};

export default DevDebug;
