import _ from "underscore";

/**
 * Extend an object to have a private channel with some convenience methods to listen to events on that channel
 *
 * to install the channel to an object
 *
 * PrivateChannel.installTo(obj, f, privateChannel);
 *
 * _emit, on, off, once and many will become available
 * @memberof Core
 * @exports PrivateChannel
 * @mixin
 */
var PrivateChannel = {
	/**
	 * Add a private channel to an object
	 * @param {Any} obj - The object that should get extended with a private channel
	 * @param {Core.ViewerFacade} f - the ViewerFacade
	 * @param {(String|Array)} privateChannel - the channel which should be installed to the object
	 */
	installTo: function(obj, f, privateChannel) {
		if (f && _.isArray(privateChannel)) {
			privateChannel = privateChannel.join(f.eDelimiter);
		}
		/**
		 * Merge a given channel with the installed privateChannel
		 *
		 * @param {(String|Array)} channel - the channel to append to privateChannel
		 */
		function mergeChannel(channel) {
			if (channel && channel.length > 0) {
				if (_.isArray(channel)) {
					channel = channel.join(f.eDelimiter);
				}
				return [privateChannel, channel].join(f.eDelimiter);
			} else {
				return privateChannel;
			}
		}

		/**
		 * Execute each of the listeners that may be listening for the specified
		 * event name in order with the list of arguments.
		 *
		 * The emit is done in the channel: privateChannel.channel
		 * emit(event, [arg1], [arg2], [...])
		 *
		 * @alias PrivateChannel~_emit
		 * @param {(String|Array|undefined)} channel - if defined, the channel gets appended to the installed private channel
		 * @param {...Any} parameters
		 */
		obj._emit = function(channel) {
			arguments[0] = mergeChannel(channel);

			f.emit.apply(f, arguments);
		};

		/**
		 * Adds a listener to the end of the listeners array for the specified
		 * event.
		 *
		 * @alias PrivateChannel~on
		 * @param {(String|Array|undefined)} channel - if defined, the channel gets appended to the installed private channel
		 * @param {Function} listener
		 */
		obj.on = function(channel) {
			arguments[0] = mergeChannel(channel);
			f.on.apply(f, arguments);
		};

		/**
		 * Remove a listener from the listener array for the specified event.
		 * Caution: changes array indices in the listener array behind the
		 * listener.
		 *
		 * @alias PrivateChannel~off
		 * @param {(String|Array|undefined)} channel - if defined, the channel gets appended to the installed private channel
		 * @param {Function} listener
		 */
		obj.off = function(channel) {
			arguments[0] = mergeChannel(channel);
			f.off.apply(f, arguments);
		};

		/**
		 * Adds a one time listener for the event. The listener is invoked only the first time the event is fired, after which it is removed.
		 *
		 * @alias PrivateChannel~once
		 * @param {(String|Array|undefined)} channel - if defined, the channel gets appended to the installed private channel
		 * @param {...Any} parameters
		 */
		obj.once = function(channel) {
			arguments[0] = mergeChannel(channel);
			f.once.apply(f, arguments);
		};

		/**
		 * Adds a listener that will execute n times for the event before being removed. The listener is invoked only the first time the event is fired, after which it is removed.
		 *
		 * @alias PrivateChannel~many
		 * @param {(String|Array|undefined)} channel - if defined, the channel gets appended to the installed private channel
		 * @param {Number} timesToListen
		 * @param {...Any} parameters
		 */
		obj.many = function(channel) {
			arguments[0] = mergeChannel(channel);
			f.many.apply(f, arguments);
		};

		/**
		 * Returns an array of listeners for the specified event. This array can be manipulated, e.g. to remove listeners.
		 *
		 * @alias PrivateChannel~listeners
		 * @param {String|Array} channel - Name of the event
		 * @returns {Array.<Function>} List of functions that listen to the specified event
		 */
		obj.listeners = function(channel) {
			arguments[0] = mergeChannel(channel);
			return f.listeners.apply(f, arguments);
		};

		/**
		 * Removes all listeners, or those of the specified event.
		 * Should be used with caution!
		 * @alias PrivateChannel~removeAllListeners
		 * @param {String|Array} [channel] - Name of the event
		 */
		obj.removeAllListeners = function(channel) {
			arguments[0] = mergeChannel(channel);
			f.removeAllListeners.apply(f, arguments);
		};

		/**
		 * Get the name of the currently installed private channel
		 * @alias PrivateChannel~getPrivateChannelName
		 * @returns {String} the installed private channel
		 */
		obj.getPrivateChannelName = function() {
			return privateChannel;
		};

		/**
		 * Get the name of the currently installed private channel
		 * @alias PrivateChannel~getPrivateChannelNameAsArray
		 * @returns {Array.<String>} the installed private channel
		 */
		obj.getPrivateChannelNameAsArray = function() {
			return privateChannel.split(f.eDelimiter);
		};
	}
};

export default PrivateChannel;
