import CoreConstants from "@esko/web-viewer/core/CoreConstants";
import LayoutConstants from "@esko/web-viewer/layout/LayoutConstants";
/**
 * The ViewerFacade is the API that the core provides to all other application components.
 *
 * The facade provides methods for communication over the EventBus, a debugger, a reference to the DocumentController, etc.
 * @exports ViewerFacade
 * @memberof Core
 * @class ViewerFacade The facade provides methods for communication over the EventBus, a debugger, a reference to the DocumentController, etc.
 * @param {Core.ViewerCore} coreAPI - The Viewer Core
 * @param {EventEmitter2} ebus - The event bus
 * @param {Debug} debug - The debug module
 */
var ViewerFacade = function(coreAPI, ebus, debug) {
	this.i18n = null; // filled in by the core

	/**
	 * Execute each of the listeners that may be listening for the specified
	 * event name in order with the list of arguments.
	 *
	 * emit(event, [arg1], [arg2], [...])
	 *
	 * @param {(String|Array)} channel
	 * @param {...Any} parameters
	 */
	this.emit = function() {
		// extra security checks can be added here
		ebus.emit.apply(ebus, arguments);
	};

	this.eDelimiter = ebus.delimiter;

	/**
	 * Adds a listener to the end of the listeners array for the specified
	 * event.
	 *
	 * on(event, listener)
	 *
	 * @param {(String|Array)} channel
	 * @param {Function} listener
	 */
	this.on = function() {
		ebus.on.apply(ebus, arguments);
	};

	/**
	 * Remove a listener from the listener array for the specified event.
	 * Caution: changes array indices in the listener array behind the
	 * listener.
	 *
	 * @param {(String|Array)} channel
	 * @param {Function} listener
	 */
	this.off = function() {
		ebus.off.apply(ebus, arguments);
	};

	/**
	 * Adds a one time listener for the event. The listener is invoked only
	 * the first time the event is fired, after which it is removed.
	 *
	 * @param {(String|Array)} channel
	 * @param {...Any} parameters
	 */
	this.once = function() {
		ebus.once.apply(ebus, arguments);
	};

	/**
	 * Adds a listener that will execute n times for the event before being
	 * removed. The listener is invoked only the first time the event is
	 * fired, after which it is removed.
	 *
	 * @param {(String|Array)} channel
	 * @param {Number} timesToListen
	 * @param {...Any} parameters
	 */
	this.many = function() {
		ebus.many.apply(ebus, arguments);
	};

	/**
	 * Returns an array of listeners for the specified event. This array can be manipulated, e.g. to remove listeners.
	 * @param {(String|Array)} channel
	 * @returns {Array.<Function>} List of functions that listen to the specified event
	 */
	this.listeners = function() {
		return ebus.listeners.apply(ebus, arguments);
	};

	/**
	 * Removes all listeners, or those of the specified event.
	 * Should be used with caution!
	 * @param {String|Array} [channel]
	 */
	this.removeAllListeners = function() {
		ebus.removeAllListeners.apply(ebus, arguments);
	};

	/**
	 * Open document(s)
	 * @param documentInformation {any} String or JSON object containing information on which document to load
	 */
	this.openDocuments = function(fileIdentifier) {
		coreAPI.openDocuments(fileIdentifier);
	};

	/**
	 * Get the DocumentController
	 *
	 * @returns {Data.DocumentController}
	 */
	this.getDocumentController = function() {
		return coreAPI.getDocumentController();
	};

	/**
	 * Get the Config
	 *
	 * @returns {Core.AbstractConfig}
	 */
	this.getConfig = function() {
		return coreAPI.getConfig();
	};

	/**
	 * Get the user who is logged in
	 *
	 * @returns {User}
	 */
	this.getCurrentUser = function() {
		return coreAPI.getCurrentUser();
	};

	/**
	 * Debugger
	 */
	this.debug = debug;

	/**
	 * Get configuration property
	 *
	 * @param {String} key - the key of the property
	 * @returns {Any} the value of the property
	 */
	this.getProperty = function(key) {
		return coreAPI.getProperty(key);
	};

	/**
	 * This returns the DataAdapter
	 *
	 * Usage of this adapter should be limited, and only when really needed!
	 * @returns {AbstractDataAdapter} the data adapter
	 */
	this.getDataAdapter = function() {
		return coreAPI.getDataAdapter();
	};

	/**
	 * This returns the DataAdapter for a specific module
	 *
	 * @param {String} module
	 * @returns {AbstractDataAdapter} the data adapter
	 */
	this.getDataAdapterForModule = function(module) {
		return coreAPI.getDataAdapter().getDataAdapterForModule(module);
	};

	/**
	 * Use with care! Do not use this for layout components.
	 * @return {JQueryObj} context
	 */
	this.getContext = function() {
		return coreAPI.getAppContext();
	};

	/**
	 * Request a module, if the module is not available yet, it will be loaded in silenced mode
	 * @param {String} name - name of the module
	 * @param {Function} callback - callback which is called when the module is available, the first argument is a reference to the module
	 * @param {Boolean} [silenced] - if undefined, the module is loaded in silenced mode!
	 */
	this.requestModule = function(name, callback, silenced) {
		if (silenced === undefined) {
			silenced = true;
		}
		coreAPI.requestModule(name, callback, silenced);
	};

	/**
	 * If a module is already loaded, this method will return the module, if the module is not loaded it will return undefined
	 *
	 * !!Note!!  this method will never start to load a module
	 * @param {String} name - name of the module
	 * @returns {Core.AbstractModule} module
	 */
	this.getModule = function(name) {
		return coreAPI.getModuleIfAvailable(name);
	};

	/**
	 * Executes the callback when the module with the given name is loaded.
	 * If it is already there, the callback is immediately executed, otherwise it waits until it is ever loaded.
	 *
	 * !! Note !! this method will never start to load a module
	 * @param {String} name - Name of the module
	 * @param {Function} callback - Parameters: moduleInstance
	 * @param {Any} [binding] - Context binding of the callback
	 */
	this.whenModuleAvailable = function(name, callback, binding) {
		return coreAPI.whenModuleAvailable(name, callback, binding);
	};

	/**
	 * Executes the callback when all modules are loaded.
	 * If this is already the case, the callback is immediately executed, otherwise it waits until it is ever loaded.
	 *
	 * @param {Function} callback
	 */
	this.whenAllModulesAreLoaded = function(callback) {
		return coreAPI.whenAllModulesAreLoaded(callback);
	};

	/**
	 * Get the loaded modules
	 */
	this.getLoadedModules = function() {
		return coreAPI.getLoadedModules();
	};

	/**
	 * Inject state to the current viewer. The viewerState is defined as and object with a entry for each module. The value of this entry is specific for
	 * each module.
	 *
	 * @param {Object} viewerState
	 */
	this.injectState = function(state) {
		return coreAPI.injectState(state);
	};

	/**
	 * Get the state of the viewer
	 * @returns {Object}
	 */
	this.getViewerState = function() {
		return coreAPI.getState();
	};

	/**
	 * Do an Ajax call
	 * @param {String} url - The URL to call
	 * @param {Object} options - Object containing the options for the ajax request. See jQuery API for the meaning
	 * @example
	 * defaultOptions = {
	 *	dataType: "xml",
	 *		success: function(data, textStatus, jqXHR) {},
	 *		error: function(jqXHR, textStatus, errorThrown) {},
	 *		cache: false,
	 *		processData: true,
	 *		contentType: "application/x-www-form-urlencoded; charset=UTF-8",
	 *		type: "GET",
	 *		data: ""
	 *	};
	 *
	 */
	this.ajax = function(url, options) {
		return coreAPI.ajax(url, options);
	};

	/**
	 * Start a ViewerMode
	 * @param {Core.ViewerMode} viewerMode - the viewermode which should be started
	 */
	this.startMode = function(viewerMode) {
		return coreAPI.startMode(viewerMode);
	};

	/**
	 * Stop a ViewerMode
	 * @param {Core.ViewerMode} viewerMode - the viewermode which should be stopped
	 */
	this.stopMode = function(viewerMode) {
		return coreAPI.stopMode(viewerMode);
	};

	/**
	 * Stop the currently active viewermode
	 */
	this.stopActiveMode = function() {
		return coreAPI.stopActiveMode();
	};

	/**
	 * Get the currently active mode
	 * @returns {Core.ViewerMode}
	 */
	this.getActiveMode = function() {
		return coreAPI.getActiveMode();
	};

	/**
	 * Return the type of layout we are in
	 * @returns {LayoutConstants.LayoutType}
	 */
	this.getLayoutType = function() {
		return coreAPI.getLayoutType();
	};

	/**
	 * Returns the active layout mode
	 */
	this.getLayoutMode = function() {
		return coreAPI.getLayoutMode();
	};

	/**
	 * Is the viewer in mobile mode?
	 * @returns {Boolean}
	 */
	this.isMobile = function() {
		return coreAPI.getLayoutType() === LayoutConstants.LayoutType.MOBILE;
	};
	/**
	 * Swith to another layout
	 * @param {LayoutConstants.LayoutType} layoutType - name of the layout to switch to
	 */
	this.switchLayout = function(layoutType) {
		coreAPI.switchLayout(layoutType);
	};

	/**
	 * Get the FeatureDetectorModule
	 * @returns {FeatureDetectorModule}
	 */
	this.getFeatureDetector = function() {
		return coreAPI.getFeatureDetector();
	};

	/**
	 * Is the state of the viewer stable? can fragments be fetched?
	 * @returns {Boolean} true if OK
	 */
	this.canFetchFragments = function() {
		return coreAPI.canFetchFragments();
	};

	/**
	 *
	 * Returns the layout manager
	 */
	this.getLayoutManager = function() {
		return coreAPI.getLayoutManager();
	};
};

/**
 * Get an option
 *
 * @returns value of the option
 */
ViewerFacade.prototype.getOption = function(option) {
	return this.getCurrentUser().getOption(option);
};

/**
 * Set an option
 * @param {String} key - Key of the option
 * @param {Object} value - Value of the option
 * @param {Boolean} [doNotPersist] - True if the option should not be persisted
 */
ViewerFacade.prototype.setOption = function(key, newValue, doNotPersist) {
	return this.getCurrentUser().setOption(key, newValue, doNotPersist);
};

/**
 * Set an option with the option to wait for when it's saved.
 * @param {String} key - Key of the option
 * @param {Object} value - Value of the option
 * @param {Function} onPersistedCallback - @param {Function} [onPersistedCallback] - Callback executed when the option is persisted. Called with parameter boolean successful (true) saved or not (false).
 */
ViewerFacade.prototype.setOptionCallback = function(key, newValue, onPersistedCallback) {
	return this.getCurrentUser().setOption(key, newValue, false, onPersistedCallback);
};

/**
 * Reset an option to its default value
 * @param option - the option to reset
 */
ViewerFacade.prototype.resetOption = function(option) {
	return this.getCurrentUser().resetOption(option);
};

/**
 * Get the active document
 * @returns {Data.AbstractDocument}
 */
ViewerFacade.prototype.getActiveDocument = function() {
	return this.getDocumentController().getActiveDocument();
};

/**
 * Get the active document
 * @returns {Data.AbstractDocument}
 */
ViewerFacade.prototype.getActivePage = function() {
	return this.getDocumentController().getActivePage();
};

/**
 * Get the size units for the current viewer. Tries to read from user preferences, but has a falback to the config.
 * @returns {CoreConstants.SizeUnits}
 */
ViewerFacade.prototype.getUnits = function() {
	var units = this.getCurrentUser().getOption("PREFERENCES_UNIT_SIZE");
	if (isNaN(units)) {
		// fallback on config
		units = this.getProperty("units");
		if (!isNaN(units)) {
			units = parseInt(units, 10);
		} else {
			units = CoreConstants.SizeUnits.MM;
		}
	}
	return units;
};

/**
 * Get the size units localized name
 * @returns {String} localized name
 */
ViewerFacade.prototype.getUnitsLocName = function(index) {
	var units = index;
	if (units === undefined) {
		units = this.getUnits();
	}
	if (
		units === CoreConstants.SizeUnits.PIXELS &&
		this.getActivePage() &&
		this.getActivePage().hasCad() &&
		!this.getActivePage().hasGraphics()
	) {
		return CoreConstants.UnitsLocNames[CoreConstants.SizeUnits.POINTS];
	} else {
		return CoreConstants.UnitsLocNames[units];
	}
};

/**
 * Get the fix for the precision of the size unit
 * @param {CoreConstants.SizeUnits} index - get unit fix or certain type
 * @returns {Number} number of precision for the size unit
 */
ViewerFacade.prototype.getUnitsFix = function(index) {
	if (index !== undefined) {
		return CoreConstants.UnitsFixes[index];
	} else {
		return CoreConstants.UnitsFixes[this.getUnits()];
	}
};

/**
 * Get the resolution units for the current viewer.
 * @returns {CoreConstants.ResolutionUnits}
 */
ViewerFacade.prototype.getResolutionUnits = function() {
	var units = this.getCurrentUser().getOption("PREFERENCES_UNIT_RESOLUTION");
	return units;
};

/**
 * Get the resolution units localized name
 * @returns {String} localized name
 */
ViewerFacade.prototype.getResolutionUnitsLocName = function() {
	return CoreConstants.ResUnitsLocNames[this.getResolutionUnits()];
};

/**
 * Get the fix for the precision of the resolution unit
 * @param {CoreConstants.ResolutionUnits} index - get unit fix or certain type
 * @returns {Number} number of precision for the resolution unit
 */
ViewerFacade.prototype.getResolutionUnitsFix = function(index) {
	if (index !== undefined) {
		return CoreConstants.ResUnitsFixes[index];
	} else {
		return CoreConstants.ResUnitsFixes[this.getResolutionUnits()];
	}
};

/**
 * Get the screen ruling units for the current viewer.
 * @returns {CoreConstants.ResolutionUnits}
 */
ViewerFacade.prototype.getScreenUnits = function() {
	var units = this.getCurrentUser().getOption("PREFERENCES_UNIT_SCREENRULING");
	return units;
};

/**
 * Get the sceen ruling units localized name
 * @returns {String} localized name
 */
ViewerFacade.prototype.getScreenUnitsLocName = function() {
	return CoreConstants.ScreenUnitsLocNames[this.getScreenUnits()];
};

/**
 * Get the fix for the precision of the screen ruling unit
 * @param {CoreConstants.ResolutionUnits} [index] - get unit fix or certain type
 * @returns {Number} number of precision for the resolution unit
 */
ViewerFacade.prototype.getScreenUnitsFix = function(index) {
	if (index !== undefined) {
		return CoreConstants.ScreenUnitsFixes[index];
	} else {
		return CoreConstants.ScreenUnitsFixes[this.getScreenUnits()];
	}
};

/**
 * Shortcut for loading a i18n bundle
 * @param {String|Array.<String>} bundlename - bundlenmame(s)
 * @param {Function} callback - callback called when bundle available
 * @param {Any} [binding]
 */
ViewerFacade.prototype.loadBundle = function(bundleName, callback, binding) {
	if (callback) {
		this.i18n.loadBundle(bundleName, callback, binding);
	} else {
		return new Promise(
			function(resolve, reject) {
				this.i18n.loadBundle(bundleName, resolve);
			}.bind(this)
		);
	}
};

/**
 * Shortcut for getting the translation of a certain key
 *
 * @param {String} key - key from which you want the translation
 * @param {...String} [replacements] - if extra parameters are given, they replace {0}, {1}, ... in the translated string
 * @returns translation - if no translation is found, the key is returned
 */
ViewerFacade.prototype.t = function(bundlename, callback, binding) {
	return this.i18n.t.apply(this.i18n, arguments);
};

/**
 * Check if a parameter update is pending or ongoing
 * @returns {Boolean}
 */
ViewerFacade.prototype.isParameterUpdateOngoing = function() {
	return this.getDocumentController().isParameterUpdateOngoing();
};

export default ViewerFacade;
