import _ from "underscore";
/**
 * @class ViewerMode - A ViewerMode is a modal modus which gets control over the viewer. Only 1 viewermode can be activate a certain point in time
 * @export ViewerMode
 * @memberof Core
 *
 * @param {Core.ViewerFacade} f
 * @param {String} name - the name of the mode
 */
var ViewerMode = function(f, name) {
	this.f = f;
	this._name = name;

	this._active = false;
};

/**
 * Start or stop this mode
 *
 * NOTE!! this should only be called by the ViewerModeManager
 * @param toggle
 */
ViewerMode.prototype.toggle = function(toggle) {
	// the mode is being started
	if (!this._active && toggle) {
		this._active = true;

		this._doStart();
	} else if (this._active && !toggle) {
		this._active = false;

		this._doStop();
	}
};

/**
 * Start the mode
 */
ViewerMode.prototype._doStart = function() {
	// implemented by the specific mode: execute all actions on start
};

/**
 * Stop the mode
 */
ViewerMode.prototype._doStop = function() {
	// implemented by the specific mode: execute all actions on stop (cleanup!)
};

/**
 * Check if the mode is active
 * @returns {Boolean}
 */
ViewerMode.prototype.isActive = function() {
	return this._active;
};

/**
 * Request the modemanager to stop this mode
 */
ViewerMode.prototype.requestStop = function() {
	if (this._active) {
		this.f.stopMode(this);
	}
};

/**
 * Allow all tools in the system
 */
ViewerMode.prototype._allowAllTools = function() {
	this.f.emit("toolManager.request.allowAllTools");
};

/**
 * Allow all tools to be used, except for the tools listed in disallowedTools
 * @param {Array.<String>} [disallowedTools] - an array of tools that should be disallowed
 */

ViewerMode.prototype._allowAllToolsExcept = function(disallowedTools) {
	this.f.emit("toolManager.request.allowAllToolsExcept", disallowedTools);
};

/**
 * Disallow all tools, except for the tools lited in allowedTools
 * @param {Array.<String>} [allowedTools] - an array of tools that are still allowed
 */
ViewerMode.prototype._disallowAllToolsExcept = function(allowedTools) {
	this.f.emit("toolManager.request.disallowAllToolsExcept", allowedTools);
};

export default ViewerMode;
