/**
 * @namespace Data
 */
/**
 * @namespace Data
 */
import PrivateChannel from "@esko/web-viewer/core/PrivateChannel";

/**
 * Construct the data adapter
 *
 * @exports AbstractDataAdapter
 * @memberof Data
 * @class DataAdapter is responsible for loading new documents and (their) data.
 * @constructor
 * @param {Core.ViewerFacade} f
 */
var AbstractDataAdapter = function(f) {
	/**
	 * @protected
	 */
	this.f = f;

	PrivateChannel.installTo(this, f, ["doc"]);

	this._dataAdapterModules = {};
};

AbstractDataAdapter.prototype.registerDataAdapterForModule = function(module, dataAdapter) {
	this._dataAdapterModules[module] = dataAdapter;
};

AbstractDataAdapter.prototype.getDataAdapterForModule = function(module) {
	var adapter = this._dataAdapterModules[module];

	if (!adapter) {
		throw "No DataAdapter Registered for Module " + module;
	}

	return adapter;
};

/**
 * Returns the number of documents that are loaded with the documentsIdentifier
 * @param {String|Object} documentsIdentifier
 * @returns {Number}
 */
AbstractDataAdapter.prototype.getNumberOfDocuments = function(documentsIdentifier) {
	throw "Not implemented: getNumberOfDocuments";
};

/**
 * Open document(s)
 * @param {any} documentsIdentifier - String or JSON object containing information on which document to load
 * @param {Function} callback - Callback fired with as first argument the document (if loading succeeded, otherwise null) and as second argument whether the document should be made active. Can be fired multiple times when there are multiple documents to be loaded.
 */
AbstractDataAdapter.prototype.openDocuments = function(documentsIdentifier, callback) {
	// open documents based on documentsIdentifier
	throw "Not implemented: openDocuments";
};

/**
 * Returns an instance for the I18n provider
 * @returns {I18n}
 */
AbstractDataAdapter.prototype.getI18nProvider = function() {
	// to be implemetened by specific dataAdapter
	throw "Not implemented: getI18nProvider";
};

/**
 * Load previous versions of the document for a given document page. If the version does not have the page with given pagenumber the version
 * will not be loaded (only happens for pagelist documents).
 *
 * @param {Object} data
 * @param {AbstractDocumentPage} document
 * @param {Function} callback
 * @param {AbstractDocumentPage} [documentPage]
 */
AbstractDataAdapter.prototype.loadVersionsOfDocumentPage = function(data, document, callback, documentPage) {
	//to be implemented by a specific dataAdapter
	throw "Not implemented: loadVersionsOfDocumentPage";
};

/**
 * Update the given pages with correct data.
 *
 * !! Note !! if two consecutive calls are done, the first (ongoing) call should be cancelled, only the last call result can update the parameters
 * @param {AbstractDocumentPage} currentPage - the currenct active page
 * @param {AbstractDocumentPage} differencePage - the current d
 * @param {AbstractDocumentPage} cadOverlay - the current cad overlay
 * @param {function} callback - callback to be called when parameter call has finished
 */
AbstractDataAdapter.prototype.updateDocumentParameters = function(currentPage, differencePage, cadOverlay, callback) {
	//to be implemented by a specific dataAdapter
	throw "Not implemented: updateDocumentParameters";
};

/**
 * Update the data of the document, also updates the document reference info
 * @param {AbstractDocumentPage}   page
 */
AbstractDataAdapter.prototype.updateDocumentData = function(page) {
	//to be implemented by a specific dataAdapter
	throw "Not implemented: updateDocumentData";
};

/**
 * If a parameter update is ongoing, it should be cancelled, and it it shouldn't update any documents or trigger events
 */
AbstractDataAdapter.prototype.cancelDocumentParametersUpdate = function() {
	// to be implemented  by a specific dataAdapter
	throw "Not implemented: cancelDocumentParametersUpdate";
};

/**
 * Fetches all pages of a document version, callback is used after all pages are initailized and added to the document version.
 * @param {AbstractDocumentVersion} docVersionId
 * @param {Function} callback
 */
AbstractDataAdapter.prototype.getPagesOfDocumentVersion = function(docVersion, callback) {
	//to be implemented
	throw "Not implemented: getPagesOfDocumentVersion";
};

export default AbstractDataAdapter;
