import Extendable from "@esko/web-viewer/core/Extendable";
import PrivateChannel from "@esko/web-viewer/core/PrivateChannel";
import DataConstants from "@esko/web-viewer/data/DataConstants";
import CompareConstants from "@esko/web-viewer/compare/CompareConstants";
import LayoutConstants from "@esko/web-viewer/layout/LayoutConstants";
import _ from "underscore";
import ShapeUtils from "@esko/shapeutils";
/**
 * @class Abstract representation of a Document page. Extend this class for a specific implementation of the document page.
 * @exports AbstractDocumentPage
 * @memberof Data
 * @constructor
 * @param {Core.ViewerFacade} f
 * @param {Data.AbstractDocumentVersion} docVersion - the documentversion which this page belongs to
 * @param {Number} pagenum - the number of the page
 */
var AbstractDocumentPage = function(f, docVersion, pagenum) {
	this.f = f;
	var privateChannel = ["doc", docVersion.getDocument().getIID()];
	privateChannel.push("version", docVersion.getVersionId());
	privateChannel.push("page", pagenum);
	PrivateChannel.installTo(this, f, privateChannel);

	/**
	 * Object containing general data of the page
	 * This information lives during the complete viewer session
	 * @private
	 */
	this._data = {};

	/**
	 * viewer parameters
	 * This information is constantly updated, and is very volatile.
	 * Caution should be taken when setting values on _parameters
	 * @private
	 */
	this._parameters = null;

	/**
	 * State data, this information is stored as long as a page is active or difference page
	 * @private
	 */
	this._state = {};
	/**
	 * Are there annotations available?
	 * @private
	 */
	this._hasAnnotations = false;

	/**
	 * The documentversion to which this page belongs
	 * @private
	 */
	this._docVersion = docVersion;

	/**
	 * The sequence number of the page (counting from 1)
	 * @private
	 */
	this._pagenum = parseInt(pagenum, 10);

	/**
	 * Is the page displayed in Readerspread
	 * @private
	 */
	this._isReaderSpread = false;

	/**
	 * Linked compare page (multi-page documents)
	 */
	this._linkedComparePage = null;

	/**
	 * Is this page a blank page, has no Graphics or CAD data
	 */
	this._isBlankPage = false;

	/**
	 * Indicates whether parameters are available or not
	 */
	this._isParametersAvailable = false;

	if (this.is3D() || this.hasVideo()) {
		this._isParametersAvailable = true;
		this._parameters = {};
	}
};

/**
 * Get a property of the page
 *
 * @param {String} key -Key of the property
 * @returns {Any} Value of the property with the given key or undefined when not found
 */
AbstractDocumentPage.prototype.get = function(key) {
	if (this._data[key] !== undefined) {
		return this._data[key];
	} else {
		return undefined;
	}
};

/**
 * Set a property of the page
 *
 * @param {String} key - Key of the property
 * @param {Any} Value of the property with the given key
 */
AbstractDocumentPage.prototype.set = function(key, value) {
	var oldValue = this._data[key];

	this._data[key] = value;

	if (oldValue != value) {
		// != is intentional
		this._emit(["property", key, "changed"], oldValue, value, this);
	}
};

/**
 * reset a property of the page
 *
 * @param {String} key - Key of the property
 */
AbstractDocumentPage.prototype.reset = function(key) {
	delete this._data[key];
};

/**
 * Get a copy of the page data
 * @returns {Object} Copy of the page data
 */
AbstractDocumentPage.prototype.getData = function() {
	return _.clone(this._data);
};

/**
 * @returns {Data.AbstractDocumentVersion} the documentversion that this page belongs to
 */
AbstractDocumentPage.prototype.getDocumentVersion = function() {
	return this._docVersion;
};

/**
 * @returns the number of the page (counting from 1)
 */
AbstractDocumentPage.prototype.getPagenum = function() {
	return this._pagenum;
};

/**
 * @returns {Data.AbstractDocument} the document that this page belongs to
 */
AbstractDocumentPage.prototype.getDocument = function() {
	return this._docVersion.getDocument();
};

/**
 * Get a viewer parameter
 * @param {String} key - Key of the property
 * @returns {Any} Viewer parameter value
 */
AbstractDocumentPage.prototype.getParam = function(key) {
	if (this._parameters !== null && this._parameters[key] !== undefined) {
		return this._parameters[key];
	} else {
		return undefined;
	}
};

/**
 * Returns if the document page has the param
 * @returns {Boolean}
 */
AbstractDocumentPage.prototype.hasParam = function(key) {
	return this._parameters !== null && this._parameters[key] !== undefined;
};

/**
 * Get a copy of all viewer parameters
 * @returns {Object} Copy of all viewer parameters
 */
AbstractDocumentPage.prototype.getParams = function() {
	return _.clone(this._parameters);
};

/**
 * Sets a parameter of the document page.
 * @param {String} key
 * @param {Any} value
 */
AbstractDocumentPage.prototype.setParam = function(key, value) {
	this._parameters[key] = value;
	this._emit("parameters.changed", this);
};

/**
 * Reset a parameter of the document page
 * @param {String} key
 */
AbstractDocumentPage.prototype.resetParam = function(key) {
	delete this._parameters[key];
	this._emit("parameters.changed", this);
};

/**
 * Rotate the document page over rotation (this adds rotation to the current rotation)
 * @param {Number} rotation
 */
AbstractDocumentPage.prototype.rotatePage = function(rotation) {
	this.setRotation(this.getRotation() + rotation);
};

/**
 * Set the rotation of the page (used for difference pages)
 * @param {Number} rotation
 */
AbstractDocumentPage.prototype.setRotation = function(rotation) {
	this.setStateParam("rotation", rotation % 360);
};

/**
 * Flip the document page
 * @param {Boolean} [flip] - if undefined just flip from the current flip state, otherwhise if true : page is flipped, if false page is not flipped
 */
AbstractDocumentPage.prototype.flipPage = function(flip) {
	if (flip === undefined) {
		flip = !this.getStateParam("flipped");
	}

	this.setStateParam("flipped", flip);
};

/**
 * get the rotation of the document page
 * @returns {Number}
 */
AbstractDocumentPage.prototype.getRotation = function() {
	var rotation = this.getStateParam("rotation");
	if (rotation) {
		return rotation * 1;
	} else {
		return 0;
	}
};

/**
 * @returns {Boolean} - is the document page flipped
 */
AbstractDocumentPage.prototype.isFlipped = function() {
	return this.getStateParam("flipped") === undefined ? false : this.getStateParam("flipped") && true;
};

/**
 * Returns the origin of the CAD-file, only if the CAD-origin is above and/or left of the graphics' (0,0)-point
 * if this is not the case it returns (0,0)
 *
 * @returns {ShapeUtils.Point}
 */
AbstractDocumentPage.prototype.getOrigin = function() {
	return new ShapeUtils.Point();
};

/**
 * Get the size of the page (note in readerspread this is the double width of the page)
 * @returns {ShapeUtils.Size}
 */
AbstractDocumentPage.prototype.getPageSize = function() {
	return new ShapeUtils.Size();
};

/**
 * Get the original size of the page
 * @returns {ShapeUtils.Size}
 */
AbstractDocumentPage.prototype.getOriginalSize = function() {
	return new ShapeUtils.Size();
};

/**
 * Get Image type
 * @returns {Number} image type or -1 if parameters not yet loaded
 */
AbstractDocumentPage.prototype.getImageType = function() {
	return this._parameters !== null ? parseInt(this._parameters.imagetype, 10) : -1;
};

/**
 * Load the URL to fetch an image based on parameters
 * @abstract
 * @param {Object} params - Parameters for imagedata
 * @returns {String} URL to fetch an image based on parameters
 */
AbstractDocumentPage.prototype.getImageURL = function(params) {
	this.f.emit("error.SEVERE", "No getImageURL implementation is provided");
	return "UNDEFINED";
};

/**
 * Load the URL to fetch info based on parameters
 * @abstract
 * @param {Object} params - Parameters for info
 * @returns {String} URL to fetch info based on parameters
 */
AbstractDocumentPage.prototype.getInfoURL = function(params) {
	this.f.emit("error.SEVERE", "No getInfoURL implementation is provided");
	return "UNDEFINED";
};

/**
 * Load info needed by a certain tool
 * params.tool indicates for which tool it is needed.
 * @abstract
 * @param {Object} params - Parameters for info
 * @param {Function} success - function to call on success, first argument should be info object containing info for tool
 * @param {Function} error - function to call on fail
 */
AbstractDocumentPage.prototype.getInfo = function(params, success, error) {
	this.f.emit("error.SEVERE", "No getInfo implementation is provided");
	return;
};

/**
 * Get the margin of the document, must be implemented
 * @abstract
 * @returns {Object} margin{bottom, top, left, right}
 */
AbstractDocumentPage.prototype.getMargins = function() {
	return undefined;
};

/**
 * Does the document have graphics
 * @returns {Boolean} True if image, false otherwise
 */
AbstractDocumentPage.prototype.hasGraphics = function() {
	return this._data.file !== undefined && !this._data.file.match(/\.svg$/i);
};

/**
 * Has annotations
 * @param {String} [originType] - checks if annotations are available for this origin type, if omitted will check if there are annotations in general
 * @returns {Boolean} True if it has annotations, false otherwise
 */
AbstractDocumentPage.prototype.hasAnnotations = function(originType) {
	if (originType && this._hasAnnotationsByOrigin) {
		return this._hasAnnotationsByOrigin[originType];
	}
	return this._hasAnnotations;
};

/**
 * Has CAD
 * @returns {Boolean} True if has cad, false otherwise
 */
AbstractDocumentPage.prototype.hasCad = function() {
	return this._data.cad ? true : false;
};

/**
 * is a 3D file
 * @returns {Boolean} True if is a 3D document, false otherwise
 */
AbstractDocumentPage.prototype.is3D = function() {
	return false;
};

/**
 * Has 3D perspective
 * @returns {Boolean} True if has 3D perspective or is a 3D document, false otherwise
 */
AbstractDocumentPage.prototype.has3D = function() {
	return (
		this.get("has3DPerspective") ||
		this.is3D() ||
		this.getDocument().getLinkedDocumentPages(DataConstants.LINKEDDOCUMENTTYPES.THREEDVIEW).length > 0
	);
};

/**
 * Does the document have svg view files
 */
AbstractDocumentPage.prototype.hasSvg = function() {
	return false;
};

/**
 * Does the document have video view files
 */
AbstractDocumentPage.prototype.hasVideo = function() {
	return this.getDocument().getDocumentType() === DataConstants.documentPageType.VIDEO;
};

/**
 * Can this page be compared to refPage in this mode?
 * if no refPage or mode has been specified, this call should return whether this page can be compared in a mode or not.
 */
AbstractDocumentPage.prototype.canBeComparedTo = function(refPage, mode) {
	if (!refPage) {
		return this.getAvailableCompareModes().length > 0;
	} else if (mode) {
		return _.intersection(this.getAvailableCompareModes(refPage), refPage.getAvailableCompareModes(this)).indexOf(mode) !== -1;
	} else {
		return _.intersection(this.getAvailableCompareModes(refPage), refPage.getAvailableCompareModes(this)).length > 0;
	}
};

/**
 * Returns the names of compare modes that are available for this document.
 * @return {Array.<String>}
 */
AbstractDocumentPage.prototype.getAvailableCompareModes = function() {
	return [];
};

/**
 * Checks if the document of this page is linked with another document (Graphics that are linked with CAD)
 * @returns {Boolean}
 */
AbstractDocumentPage.prototype.hasLinkedPage = function() {
	return this.getDocument().hasLinkedDocument();
};

/**
 * returns the linked document page
 * @returns {AbstractDocumentPage}
 */
AbstractDocumentPage.prototype.getLinkedPage = function() {
	return this.getDocument().getLinkedDocumentPage();
};

AbstractDocumentPage.prototype.getLinkedPages = function(type) {
	return this.getDocument().getLinkedDocumentPages(type);
};

/**
 * Get the unique ID for this page, build as follows: "docIID_versionID_pagenum"
 * @returns {String}
 */
AbstractDocumentPage.prototype.getUniqueID = function() {
	return this._docVersion.getDocument().getIID() + "_" + this._docVersion.getVersionId() + "_" + this._pagenum;
};

/**
 * Are the parameters available
 * @returns {Boolean}
 */
AbstractDocumentPage.prototype.isParametersAvailable = function() {
	return this._parameters !== null && this._isParametersAvailable;
};

/**
 * Execute some code when the parameters for this page are available
 * @param {Function} callback - first parameter of the callback is the page itself
 * @param {Object} [binding] - binding for callback (optional)
 */
AbstractDocumentPage.prototype.whenParametersAvailable = function(callback, binding) {
	if (binding) {
		callback = _.bind(callback, binding);
	}

	if (this.isParametersAvailable()) {
		callback(this);
	} else {
		this.once(
			"parameters.available",
			_.bind(function() {
				callback(this);
			}, this)
		);
	}
};

/**
 * Set the parameters.
 * @param {Object} parameters
 */
AbstractDocumentPage.prototype.setParameters = function(parameters) {
	this._parameters = parameters;
	this._emit("parameters.changed", this);
};

/**
 * Resets all the parameters so they need be loaded again
 */
AbstractDocumentPage.prototype.resetParameters = function() {
	this._parameters = null;
};

/**
 * Inject the needed parameters for the page when a fragment url is requested
 * @params {Object} c - the context of intercept.fragmentURL
 */
AbstractDocumentPage.prototype.interceptFragmentURL = function() {
	// nothing is added by default
};

/**
 * Set if the page has annotations or not
 * @param {Boolean} hasAnnotations
 * @param {Object} [hasAnnotationsByOrigin] - used to specify which annotation is available based on the origin
 */
AbstractDocumentPage.prototype.setHasAnnotations = function(hasAnnotations, hasAnnotationsByOrigin) {
	if (hasAnnotations !== this._hasAnnotations) {
		this._hasAnnotations = hasAnnotations;
		this._emit("hasAnnotations.changed", this);
	}
	this._hasAnnotationsByOrigin = hasAnnotationsByOrigin;
};

/**
 * Does this page equals another page?
 * @param {Data.AbstractDocumentPage} other
 */
AbstractDocumentPage.prototype.equals = function(other) {
	if (other && this.getUniqueID() == other.getUniqueID()) {
		if (this._docVersion == other._docVersion) {
			if (this._hasAnnotations == other._hasAnnotations) {
				if (
					_.every(this._data, function(value, key) {
						return value == other._data[key];
					})
				) {
					return true;
				}
			}
		}
	}
	return false;
};

/**
 * Indicate that parameters are available for this file
 * @param {Boolean} available - true if parameters are set
 */
AbstractDocumentPage.prototype.setParametersAvailable = function(available) {
	this._isParametersAvailable = available;

	if (available) {
		this._emit("parameters.available");
	}
};

/**
 * Notify that (raw) CAD data is avaialble
 * NOTE!! Should only be used by datadapter/documentcontroler
 */
AbstractDocumentPage.prototype.setCadDataAvailable = function() {
	this._emit("cadData.available");
};

/**
 * Set if the page is in reader spread or not
 * @param {Boolean} isReaderSpread
 */
AbstractDocumentPage.prototype.setReaderSpread = function(isReaderSpread) {
	this._isReaderSpread = isReaderSpread;
};

/**
 * Is the current document page in reader spread
 * @returns {Boolean}
 */
AbstractDocumentPage.prototype.isReaderSpread = function() {
	return this._isReaderSpread;
};

/**
 * Get the next page in the document version with respect to this page
 * @param {Number} [times] How many times next?
 * @param {Boolean} [skipBlankPages] indicates if the next page can or cannot be an empty page
 * @returns {Data.AbstractDocumentPage} The next page in the document version with respect to this page. The next of the last page, stays the last page.
 */
AbstractDocumentPage.prototype.getNextPage = function(times, skipBlankPages) {
	var documentVersion = this.getDocumentVersion();
	if (times === undefined) {
		times = 1;
	}
	var nextPageNumber = Math.min(this.getPagenum() + times, documentVersion.getPageCount());
	var nextPage = documentVersion.getPage(nextPageNumber);
	while ((nextPage === undefined || (skipBlankPages && nextPage.isBlankPage())) && ++nextPageNumber <= documentVersion.getPageCount()) {
		nextPage = documentVersion.getPage(nextPageNumber);
	}
	return nextPage;
};

/**
 * Get the previous page with respect to the current active page
 * @param {Number} [times] How many times previous?
 * @param {Boolean} [skipBlankPages] indicates if the next page can or cannot be an empty page
 * @returns {Data.AbstractDocumentPage} The previous page with respect to the current active page. The previous of the first page, stays the first page.
 */
AbstractDocumentPage.prototype.getPreviousPage = function(times, skipBlankPages) {
	var documentVersion = this.getDocumentVersion();
	if (times === undefined) {
		times = 1;
	}
	var previousPageNumber = Math.max(this.getPagenum() - times, 1);
	var previousPage = documentVersion.getPage(previousPageNumber);
	while ((previousPage === undefined || (skipBlankPages && previousPage.isBlankPage())) && --previousPageNumber >= 1) {
		previousPage = documentVersion.getPage(previousPageNumber);
	}
	return previousPage;
};

/**
 * Get the linked compare page of this page
 * @returns {AbstractDocumentPage} Linked compare page or null if none set.
 */
AbstractDocumentPage.prototype.getLinkedComparePage = function() {
	return this._linkedComparePage;
};

/**
 * Set the linked compare page of this page
 * @param {AbstractDocumentPage} linkedComparePage New linked compare page of this page.
 */
AbstractDocumentPage.prototype.setLinkedComparePage = function(linkedComparePage) {
	this._linkedComparePage = linkedComparePage;
};

/**
 * Reset the linked compare page
 */
AbstractDocumentPage.prototype.resetLinkedComparePage = function() {
	this._linkedComparePage = null;
};

/**
 * Has this page a linked compare page?
 * @returns {Boolean} True if it has a linked compare page, false otherwise
 */
AbstractDocumentPage.prototype.hasLinkedComparePage = function() {
	return this._linkedComparePage !== null;
};

/**
 * Is the rectangle on the page, only for part on the page returns true
 * @param {ShapeUtils.Rect} rect
 * @returns {Boolean}
 */
AbstractDocumentPage.prototype.isRectOnPage = function(rect) {
	//to be implemented
};

/**
 * Return the version Id for this page (used for page list document)
 * @returns {Number} - version id for this page
 */
AbstractDocumentPage.prototype.getPageVersionId = function() {
	//to be implemented, only use case is page list documents
	return null;
};

/**
 * Is this page a blank page?, has no Graphics or CAD data
 * @returns {Boolean}
 */
AbstractDocumentPage.prototype.isBlankPage = function() {
	return this._isBlankPage;
};

/**
 * Check if the page is a digital film
 * @returns {Boolean}
 */
AbstractDocumentPage.prototype.isDigitalFilmPage = function() {
	return false;
};

/**
 * Set a state parameter. This is information lives as long as a page is active or set as difference page
 * @param {String} key - Key of the state parameter
 * @param {Object} value - Value for the state parameter
 */
AbstractDocumentPage.prototype.setStateParam = function(key, value) {
	this._state[key] = value;
	this._emit("stateParameter.changed", this, key);
};

/**
 * Checks if the state parameter with the key is set in this document page
 * @param   {String} key - Key of the state parameter
 * @returns {Boolean}
 */
AbstractDocumentPage.prototype.hasStateParam = function(key) {
	return this._state[key] !== undefined;
};

/**
 * Get a state parameter
 * @param {String} key - Key of the state parameter
 * @returns {Object} value for the given key or undefined if not found
 */
AbstractDocumentPage.prototype.getStateParam = function(key) {
	return this._state[key];
};

/**
 * Reset the value of a state param
 * @param key
 */
AbstractDocumentPage.prototype.resetStateParam = function(key) {
	delete this._state[key];
};

/**
 * Reset all data stored in the state
 */
AbstractDocumentPage.prototype.resetState = function() {
	this._state = {};
};

/**
 * Get the url for fetching tumbnails (50x50)
 * @returns {String} - url for fetching the thumbdata of the page
 */
AbstractDocumentPage.prototype.getThumbnailUrl = function() {
	//to be implemented in subclass
	return null;
};

/**
 * Returns the information needed for a DocumentDropdownmenu to add an document to the dropdown menu
 * @returns {Object}
 */
AbstractDocumentPage.prototype.getInformationForDocumentSelector = function() {
	return {
		has3D: this.has3D(),
		hasGraphics: this.hasGraphics(),
		hasCad: this.hasCad(),
		hasVideo: this.hasVideo()
	};
};

/**
 * Can the CAD overlay be saved for this file?
 * @returns {Boolean}
 */
AbstractDocumentPage.prototype.canSaveCadOverlay = function() {
	// to be implemented in subclass
	return false;
};

/**
 * Does the document page need an update of its data
 * @returns {Boolean}
 */
AbstractDocumentPage.prototype.needsDataUpdate = function() {
	// to be implemented in subclass
	return false;
};

/**
 * Get the origin of the document page, calculates this from the trimbox origin and the margins
 * @returns {ShapeUtils.Point}
 */
AbstractDocumentPage.prototype.getDocumentPageOrigin = function() {
	// to be implemented in subclass
	throw "Not implemented";
};

/**
 * get the trimbox rectangle for the page
 * @returns {ShapeUtils.Rect}
 */
AbstractDocumentPage.prototype.getTrimboxRect = function() {
	// to be implemented in subclass
	throw "Not implemented";
};

/**
 * Get the current scaling of the document page, takes the compensation state (dist or scale or both) into account, returns null if no compensation present
 * @returns {ShapeUtils.Point|null}
 */
AbstractDocumentPage.prototype.getCurrentScaling = function() {
	// to be implemented in subclass
	throw "Not implemented";
};

/**
 * get the size of the trimbox, return null if the trimbox size is not known by the page (need to use margins)
 * @return {ShapeUtils.Size}
 */
AbstractDocumentPage.prototype.getTrimboxSize = function() {
	// to be implemented in subclass
	throw "Not implemented";
};

/**
 * Is the user allowed the edit/save te attributes of this page?
 * @returns {Boolean}
 */
AbstractDocumentPage.prototype.canEditAttributes = function() {
	// to be implemented in subclass
	throw "Not implemented";
};

/**
 * Can the annotations be changed for this document page (not latest version, pharma settings for non-active projects, ...)
 * @returns {Boolean}
 */
AbstractDocumentPage.prototype.canChangeAnnotations = function() {
	// to be implemented in subclass
	throw "Not implemented";
};

/**
 * Is used to conditionally render the compare options at bottom in the document compare widget
 * @returns {Boolean}
 */
AbstractDocumentPage.prototype.areCompareOptionsAllowed = function() {
	return true;
};

AbstractDocumentPage.prototype.getLayoutMode = function() {
	return LayoutConstants.LayoutMode.VIEW2D;
};

export default AbstractDocumentPage;
