import Extendable from "@esko/web-viewer/core/Extendable";
import PrivateChannel from "@esko/web-viewer/core/PrivateChannel";
import _ from "underscore";
/**
 * @class Abstract representation of a DocumentVersion that can be loaded in the viewer. No
 * data is fetched in the AbstractClass. Extend this class to have a working
 * version fetching the correct information
 *
 * @exports AbstractDocumentVersion
 * @memberof Data
 * @constructor
 * @param {Core.ViewerFacade} f
 * @param {AbstractDocument} doc - reference to the document
 * @param {Number} versionId - id of the version of the document
 */
var AbstractDocumentVersion = function(f, doc, versionId) {
	PrivateChannel.installTo(this, f, ["doc", doc.getIID(), "version", versionId]);
	/**
	 * @private
	 */
	this._f = f;

	/**
	 * reference to the document
	 * @private
	 */
	this._doc = doc;

	/**
	 * Number of the version of the document
	 * @private
	 */
	this._versionId = versionId;

	/**
	 * Specific data of the document version
	 * @private
	 */
	this._versionData = {};

	/**
	 * list of all the pages in this documentversion
	 * @private
	 */
	this._pages = [];
};

/**
 * Get a property of the document
 * @param {String} key - name of the property
 * @returns {Any} Property value for the given key, or undefined if not found
 */
AbstractDocumentVersion.prototype.get = function(key) {
	if (this._versionData[key] !== undefined) {
		return this._versionData[key];
	} else {
		return undefined;
	}
};

/**
 * Change a property of the document
 *
 * @param {String} key - name of the property
 * @param {Any} newValue - new value
 * @returns {Boolean} True if success
 */
AbstractDocumentVersion.prototype.set = function(key, newValue) {
	var oldValue = this._versionData[key];
	if (oldValue === undefined || (oldValue !== undefined && oldValue != newValue)) {
		this._versionData[key] = newValue;
	}
	return true;
};

/**
 * returns the document of this page
 * @returns {AbstractDocument}
 */
AbstractDocumentVersion.prototype.getDocument = function() {
	return this._doc;
};

/**
 * Get the version number of the documentVersion
 * @returns {Number}
 */
AbstractDocumentVersion.prototype.getVersionId = function() {
	return this._versionId;
};

/**
 * Request a certain page
 *
 * If the page is not loaded yet, load it, when the page is available, the callback is called with the page as first argument
 *
 * @param {Number} pagenum - number of needed page
 * @param {Function} callback - callback-function which is called when page is available
 */
AbstractDocumentVersion.prototype.requestPage = function(pagenum, callback) {
	var page = this.getPage(pagenum);
	if (page === undefined) {
		this.once(
			["page", pagenum, "loaded"],
			_.bind(function() {
				callback(this.getPage(pagenum));
			}, this)
		);
		this.loadPage(pagenum);
	} else if (page === null) {
		this.once(
			["page", pagenum, "loaded"],
			_.bind(function() {
				callback(this.getPage(pagenum));
			}, this)
		);
	} else {
		callback(page);
	}
};

/**
 * Load a new page
 *
 * @param {Number} pagenum - number of needed page
 * @param {Function} callback - callback-function which is called when page is available
 */
AbstractDocumentVersion.prototype.loadPage = function(pagenum) {
	this._pages[pagenum] = null; //if it fails loading it should place this._pages[pagenum] back to undefined
	this._emit("page.loading", pagenum);
	this._pages[pagenum] = undefined; //= page couldn't be loaded
	// to be implemented by implementation of AbstractDocumentVersion
};

/**
 * Add a page to the document version
 * @param {Number} pagenum
 * @param {AbstractDocumentPage} page
 */
AbstractDocumentVersion.prototype.addPage = function(pagenum, page) {
	this._pages[pagenum] = page;
	this._emit("page." + pagenum + ".loaded", page);
};

/**
 * Get the page object of a certain page
 * @param {Number} [pagenum] - Number of the page (if undefined the first page will be returned)
 * @returns {Data.AbstractDocumentPage} Page at requested pagenum or undefined if the requested page does not exist
 */
AbstractDocumentVersion.prototype.getPage = function(pagenum) {
	if (pagenum === undefined) {
		return this.getFirstPage();
	}
	return this._pages[pagenum];
};

/**
 * Check if the document version has a page with pagenum
 * @param {Number} pagenum
 * @returns {Boolean}
 */
AbstractDocumentVersion.prototype.hasPage = function(pagenum) {
	return this._pages[pagenum] !== undefined;
};

/**
 * Page count of the document
 * @returns {Number} Page count of the document
 */
AbstractDocumentVersion.prototype.getPageCount = function() {
	//to be implemented by specific document version
};

/**
 * Checks if this document version is the latest version
 * @returns {Boolean}
 */
AbstractDocumentVersion.prototype.isLatestVersion = function() {
	return this._doc.getDocumentVersion().getVersionId() === this._versionId;
};

/**
 * Check if the document version has multiple pages
 * @returns {Boolean}
 */
AbstractDocumentVersion.prototype.isMultiPage = function() {
	return this.getPageCount() > 1;
};

/**
 * Check if the document version is a page list
 * @returns {Boolean}
 */
AbstractDocumentVersion.prototype.isPageList = function() {
	// pagelist when pageindex is defined in the pages
	return this.get("isPageList") !== undefined && this.get("isPageList");
};

/**
 * Get the first page
 * If the first page is an empty page, take the first real page.
 * @returns {Data.AbstractDocumentPage} The first page
 */
AbstractDocumentVersion.prototype.getFirstPage = function() {
	var i = 0;
	while (i < this._pages.length && !this._pages[i]) {
		i++;
	}
	return this._pages[i];
};

/**
 * Get the first not blank/empty page in the document version
 *
 * @returns {Data.AbstractDocumentPage} the first not blank/empty page, returns undefined if no page is found
 */
AbstractDocumentVersion.prototype.getFirstNotBlankPage = function() {
	var i = 0;
	while (i < this._pages.length && (!this._pages[i] || this._pages[i].isBlankPage())) {
		i++;
	}
	return this._pages[i];
};

/**
 * Get the last page
 * @returns {Data.AbstractDocumentPage} The last page
 */
AbstractDocumentVersion.prototype.getLastPage = function() {
	return _.last(this._pages);
};

/**
 * Get all pages of this version
 * @returns {Array.<AbstractDocumentPage>} All pages of the version.
 */
AbstractDocumentVersion.prototype.getAllPages = function() {
	return _.without(this._pages.slice(0), null, undefined);
};

export default AbstractDocumentVersion;
