import PrivateChannel from "@esko/web-viewer/core/PrivateChannel";
import Map from "@esko/map";
import StringSet from "@esko/stringset";
import _ from "underscore";
import $ from "jquery";
/**
 * @class The DocumentController is responsible for keeping track of all loaded
 * documents. It has methods to load new documents, and set the active
 * document
 *
 * The DocumentController uses the doc private channel
 *
 * @exports DocumentController
 * @memberof Data
 * @constructor
 * @param f {Core.ViewerFacade}
 * @param adapter {AbstractDataAdapter} the DataAdapter which should be used by the DocumentController
 */
var DocumentController = function(f, adapter) {
	this.f = f;
	PrivateChannel.installTo(this, f, "doc");

	this._nextIID = 0;
	// {Number} iid -> {AbstractDocument} document
	this._documents = new Map();
	/**
	 * Linked documents contains all documents that are linked from another document
	 * they are not available in this._documents
	 */
	this._linkedDocuments = new Map();

	this._adapter = adapter;

	/**
	 * The page that is currently shown in the viewer
	 * {AbstractDocumentPage}
	 */
	this._activePage;

	/**
	 * The reference page that is used by compare
	 * {AbstractDocumentPage}
	 */
	this._differencePage;

	/**
	 * CAD overlay if any
	 */
	this._cadOverlay;

	/**
	 * This keeps track of which parameter updates are pending
	 */
	this._pendingParameterUpdateOrigins = new StringSet();
	/**
	 * This keeps tracks wether an parameter update is ongoing or not...
	 */
	this._parameterUpdateOngoing = true;

	/**
	 * Number of documents that are still loading
	 */
	this._documentsLoading = 0;

	/**
	 * If the page changes, the previous page is stored here until the parameters update has finished
	 */
	/**
	 * {Boolean}
	 */
	this._activePageHasChanged;
	/**
	 * {AbstractDocumentPage}
	 */
	this._previousActivePage;
	/**
	 * {Boolean}
	 */
	this._differencePageHasChanged;
	/**
	 * {AbstractDocumentPage}
	 */
	this._previousDifferencePage;

	/**
	 * Trigger a debounced parameter update
	 */
	this._debouncedExecuteParameterUpdateIfAllowed = _.debounce(_.bind(this._executeParameterUpdateIfAllowed, this), 20);

	// also trigger parametersUpdated when the parameters are changed on the page (without a call to the ViewServer (this is used by CAD)
	var debouncedNotifyParametersUpdated = _.debounce(_.bind(this._notifyParametersUpdated, this), 20);
	this.on(
		"**.page.*.parameters.changed",
		_.bind(function() {
			if (!this._parameterUpdateOngoing) {
				debouncedNotifyParametersUpdated();
			}
		}, this)
	);

	var debouncedNotifyStateParametersUpdated = _.debounce(_.bind(this._notifyStateParametersUpdated, this), 20);
	this.on(
		"**.page.*.stateParameter.changed",
		_.bind(function(page) {
			debouncedNotifyStateParametersUpdated(page);
		}, this)
	);

	this.f.on("intercept.fragmentURL", _.bind(this.interceptFragmentURL, this));
};

// /////////////////////////////////////
// EXTERNAL API
// /////////////////////////////////////
/**
 * Open document(s)
 * @param {any} fileIdentifier - String or JSON object containing information on which document to load
 */
DocumentController.prototype.openDocuments = function(fileIdentifier) {
	var defer = $.Deferred();
	var nrOfDocuments = this._adapter.getNumberOfDocuments(fileIdentifier);
	if (nrOfDocuments.then && typeof nrOfDocuments.then === "function") {
		nrOfDocuments.then(defer.resolve);
	} else {
		defer.resolve(nrOfDocuments);
	}
	defer.done(
		function(nrOfDocuments) {
			this._documentsLoading = nrOfDocuments;
			this._adapter.openDocuments(
				fileIdentifier,
				_.bind(function(doc, makeActive) {
					this._documentsLoading--;
					// doc can be null (in case no viewfiles were available) but this should still trigger whenDocumentsLoaded
					if (doc) {
						this.addDocument(doc, makeActive);
						if (doc.hasLinkedDocument()) {
							var linkedDocument = doc.getLinkedDocument();
							this._linkedDocuments.add(linkedDocument.getIID(), linkedDocument);
						}
					}
					if (this._documentsLoading === 0) {
						this._emit("loaded", null);
					}
				}, this)
			);
		}.bind(this)
	);
};

/**
 * Creates a copy from the cad overlay and adds this new document page to the linked pages collection.
 * It sets a link between this new page and the active (Graphics) document page. After creation the new page is set as cadOverlay.
 * @param {Boolean} [doNotActivate] do not activate the newly created CAD page
 * @param {AbstractDocumentPage} [cadToLinkFrom] the CAD that will be used to create the link from
 */
DocumentController.prototype.createNewLinkedCadFromOverlay = function(doNotActivate, cadToLinkFrom) {
	if (this._cadOverlay || cadToLinkFrom) {
		if (!cadToLinkFrom) {
			cadToLinkFrom = this._cadOverlay;
		}
		let oldLinkedPage = this._activePage.getDocument().getLinkedDocumentPage();
		let newLinkedPage = cadToLinkFrom;
		this._activePage.getDocument().setLinkedDocumentPage(newLinkedPage);

		this._linkedDocuments.add(newLinkedPage.getDocument().getIID(), newLinkedPage.getDocument());
		if (!doNotActivate) {
			this.changeCadOverlay(newLinkedPage);
		}
		this._emit("newLinkedCad", oldLinkedPage, newLinkedPage);
	} else {
		let oldLinkedPage = this._activePage.getDocument().getLinkedDocumentPage();
		this._activePage.getDocument().setLinkedDocumentPage(null);
		this._cadOverlay = null;
		this._emit("newLinkedCad", oldLinkedPage, null);
	}
};

/**
 * Add a newly created document to the linked documents, changes the linked document of the active document page
 * @param {AbstractDocument} linkedDocument
 * @param {Boolean} [doNotActivate] do not activate the newly added linked document
 */
DocumentController.prototype.addNewLinkedDocument = function(linkedDocument, doNotActivate) {
	var oldLinkedPage = this._activePage.getDocument().getLinkedDocumentPage();
	var linkedPage = linkedDocument.getDocumentVersion().getPage();
	this._activePage.getDocument().setLinkedDocumentPage(linkedPage);
	this._linkedDocuments.add(linkedDocument.getIID(), linkedDocument);
	if (!doNotActivate) {
		this.changeCadOverlay(linkedPage);
	}
	this._emit("newLinkedCad", oldLinkedPage, linkedPage);
};

/**
 * Load all pages for the given docId and version number
 * @param {Number} docId
 * @param {Number} versionNumber
 */
DocumentController.prototype.loadDocumentPages = function(docId, versionNumber) {
	var document = this.getDocument(docId);
	if (document !== undefined) {
		var documentVersion = document.getDocumentVersion(versionNumber);
		if (documentVersion !== undefined) {
			this._adapter.getPagesOfDocumentVersion(
				documentVersion,
				_.bind(function() {
					this._emit("loaded"); //after all pages are loaded emit loaded, so listeners know changes are made to loaded documents
				}, this)
			);
		}
	}
};

/**
 * Returns all linked documents
 * @returns {Array.<AbstractDocument>}
 */
DocumentController.prototype.getLinkedDocuments = function() {
	return this._linkedDocuments.values();
};

/**
 * Returns a document with the docId
 * @param {Number} docId
 * @returns {AbstractDocument}
 */
DocumentController.prototype.getDocument = function(docId) {
	var document = this._documents.get(docId);
	if (document) {
		return document;
	} else {
		return this._linkedDocuments.get(docId);
	}
};

/**
 * Retuns a document page
 * @param {String} pageId
 * @returns {AbstractDocumentPage}
 */
DocumentController.prototype.getDocumentPage = function(pageId) {
	var idArray = pageId.split("_");
	var document = this.getDocument(idArray[0]);
	if (document) {
		var version = document.getDocumentVersion(idArray[1]);
		if (version) {
			return version.getPage(idArray[2]);
		}
	}
};

/**
 * Call this method when you will change things that will trigger a parameter update in the end
 * @param {String} origin - an identifier, you also need to supply in readyForParameterUpdate
 */
DocumentController.prototype.willNeedParameterUpdate = function(origin) {
	if (!this._pendingParameterUpdateOrigins.has(origin)) {
		this._pendingParameterUpdateOrigins.add(origin);
		this._parameterUpdateOngoing = true;
		_.each(
			[this._activePage, this._differencePage, this._cadOverlay],
			function(page) {
				if (page) {
					page.setParametersAvailable(false);
				}
			},
			this
		);
		this._emit("parameters.willChange", origin);
		this._adapter.cancelDocumentParametersUpdate();
	}
};

/**
 * Call this method when you have applied your changes, and are ready for a parameter update
 * @param {String} origin - an identifier, the same as provided in the willNeedParameterUpdate call
 * @param {Function} [callback] - callback called when parameters are updated
 * @param {Object} [binding] - binding for the callback
 */
DocumentController.prototype.readyForParameterUpdate = function(origin, callback, binding) {
	var removed = this._pendingParameterUpdateOrigins.remove(origin);
	// only trigger an update if this origin was already registered by the "willNeedParameterUpdate"
	if (removed !== undefined) {
		this._emit("parameters.readyForUpdate", origin);

		if (callback) {
			this.whenParametersAvailable(callback, binding);
		}

		if (this._pendingParameterUpdateOrigins.size() === 0) {
			// no more pending parameter updates...
			this._debouncedExecuteParameterUpdateIfAllowed();
		}
	}
};

/**
 * Execute the parameter update if there are no more parameter updates pending
 */
DocumentController.prototype._executeParameterUpdateIfAllowed = function() {
	if (this._pendingParameterUpdateOrigins.size() === 0) {
		this._emit("parameters.startUpdate");
		this._updateParameters(
			_.bind(function() {
				var oldPage;

				// mark parameters as available
				_.each(
					[this._activePage, this._differencePage, this._cadOverlay],
					function(page) {
						if (page) {
							page.setParametersAvailable(true);
						}
					},
					this
				);

				// the parameters have been updated, pages have been updated....
				// before the viewport world is verified
				this._emit("parameters.beforeEndUpdate", this._activePageHasChanged, this._differencePageHasChanged, this._cadOverlayHasChanged);
				// the viewport can react on this event, so it can set itself correctly before other changes occur
				this._emit("parameters.endUpdate", this._activePageHasChanged, this._differencePageHasChanged, this._cadOverlayHasChanged);
				// this event can be used for doing changes to a model that should be in place when modules get the "parametersAvailable" event
				this._emit("parameters.afterEndUpdate", this._activePageHasChanged, this._differencePageHasChanged, this._cadOverlayHasChanged);

				// send out all events, based on parameters...
				var activePage = this.getActivePage();
				var differencePage = this.getDifferencePage();

				if (activePage) {
					this._emit("activePageParametersChange", activePage, differencePage);
				}

				if (this._activePageHasChanged) {
					// the active page has changed
					this._emit("activePageChangedAndParametersAvailable", this._previousActivePage, activePage);
				}

				if (activePage || this._activePageHasChanged) {
					this._emit("activePageChangedAndParametersAvailableOrChange", this._previousActivePage, activePage);
				}

				if (differencePage) {
					this._emit("differencePageParametersChange", activePage, differencePage);
				}

				if (this._differencePageHasChanged) {
					// the difference page has changed
					this._emit("differencePageChangedAndParametersAvailable", this._previousDifferencePage, differencePage);
				}

				if (differencePage || this._differencePageHasChanged) {
					this._emit("differencePageChangedAndParametersAvailableOrChange", this._previousDifferencePage, differencePage);
				}

				if (this._activePageHasChanged || this._differencePageHasChanged) {
					// the active or difference page has changed
					oldPage = this._activePageHasChanged ? this._previousActivePage : this._previousDifferencePage;
					this._emit(
						"activeOrDifferencePageChangedAndParametersAvailable",
						activePage,
						differencePage,
						oldPage,
						this._activePageHasChanged
					);
				}

				if (activePage || differencePage || this._activePageHasChanged || this._differencePageHasChanged) {
					if (this._activePageHasChanged) {
						oldPage = this._previousActivePage;
					} else if (this._differencePageHasChanged) {
						oldPage = this._previousDifferencePage;
					} else {
						oldPage = undefined;
					}

					this._emit(
						"activeOrDifferencePageChangedAndParametersAvailableOrChange",
						activePage,
						differencePage,
						oldPage,
						this._activePageHasChanged
					);
				}

				if (this._cadOverlayHasChanged) {
					this._emit("cadOverlayChangedAndParametersAvailable", this._previousCadOverlay, this._cadOverlay);
				}

				if (this._activePageHasChanged || this._differencePageHasChanged || this._cadOverlayHasChanged) {
					this._emit("activeOrDifferenceOrCadOverlayChangedAndParametersAvailable", activePage, differencePage, this._cadOverlay);
				}

				this._parameterUpdateOngoing = false;

				// parameters are up to date
				this._emit("parameters.changed", activePage, differencePage);

				this._activePageHasChanged = false;
				this._differencePageHasChanged = false;
				this._cadOverlayHasChanged = false;
				this._previousActivePage = undefined;
				this._previousDifferencePage = undefined;
			}, this)
		);
	}
};

/**
 * Send out Parameters changed evens if no parameter update is ongoing
 */
DocumentController.prototype._notifyParametersUpdated = function() {
	if (!this._parameterUpdateOngoing) {
		// send out all events, based on parameters...
		var activePage = this.getActivePage();
		var differencePage = this.getDifferencePage();

		if (activePage) {
			this._emit("activePageParametersChange", activePage, differencePage);
			this._emit("activePageChangedAndParametersAvailableOrChange", activePage, activePage);
		}

		if (differencePage) {
			this._emit("differencePageParametersChange", activePage, differencePage);
			this._emit("differencePageChangedAndParametersAvailableOrChange", differencePage, differencePage);
		}

		if (activePage || differencePage) {
			this._emit("activeOrDifferencePageChangedAndParametersAvailableOrChange", activePage, differencePage, differencePage, false);
		}
	}
};

/**
 * Notify listeners that the state parameters have been updated
 * @param {AbstractDocumentPage} page
 */
DocumentController.prototype._notifyStateParametersUpdated = function(page) {
	var differencePage = this.getDifferencePage();
	var cadOverlay = this.getCadOverlay();

	if (page === differencePage) {
		this._emit("referencePageStateParametersChanged", this.getActivePage(), this.getDifferencePage());
	} else if (page === cadOverlay) {
		this._emit("cadOverlayPageStateParametersChanged", this.getActivePage(), this.getDifferencePage(), this.getCadOverlay());
	}
};

/**
 * Check if a parameter update is pending or ongoing
 * @returns {Boolean}
 */
DocumentController.prototype.isParameterUpdateOngoing = function() {
	return this._parameterUpdateOngoing;
};

/**
 * Execute a callback when the parameters are available
 * (if the parameters are already available and no update is pending, the callback is executed immediately)
 *
 * @param {Function} callback - function that is called once the parameters are available (with parameters activePage, differencePage)
 * @param {Any} [binding] - binding for the callback
 */
DocumentController.prototype.whenParametersAvailable = function(callback, binding) {
	if (binding) {
		callback = _.bind(callback, binding);
	}
	if (this.isParameterUpdateOngoing()) {
		this.once("parameters.changed", callback);
	} else {
		callback(this.getActivePage(), this.getDifferencePage());
	}
};

/**
 * Update the parameters of the active and difference pages
 * @param {Function} callback
 */
DocumentController.prototype._updateParameters = function(callback) {
	this._adapter.updateDocumentParameters(this._activePage, this._differencePage, this._cadOverlay, callback);
};

/**
 * Get the currently active page
 * @returns {Data.AbstractPage} page - undefined if no document active
 */
DocumentController.prototype.getActivePage = function() {
	return this._activePage;
};

/**
 * @returns {Boolean} True if there is an active document page, false otherwise.
 */
DocumentController.prototype.hasActiveDocumentPage = function() {
	return this._activePage !== undefined;
};

/**
 * @returns {Boolean} True if there is an active document page with the parameters available
 */
DocumentController.prototype.hasActiveDocumentPageWithParametersAvailable = function() {
	return this._activePage !== undefined && this._activePage.isParametersAvailable();
};

/**
 * @returns {Boolean} true when the cad overlay is set and its parameters are available
 */
DocumentController.prototype.hasCadOverlaywWithParametersAvailable = function() {
	return this._cadOverlay && this._cadOverlay.isParametersAvailable();
};

/**
 *
 * @returns {Boolean} True if there is a difference page set, false otherwise.
 */
DocumentController.prototype.hasDifferenceDocumentPage = function() {
	return this._differencePage !== undefined;
};

/**
 * @returns {Boolean} True if there is a difference document page with the parameters available
 */
DocumentController.prototype.hasDifferenceDocumentPageWithParametersAvailable = function() {
	return this._differencePage !== undefined && this._differencePage.isParametersAvailable();
};

/**
 * returns the diffence page
 * @returns {AbstractDocumentPage}
 */
DocumentController.prototype.getDifferencePage = function() {
	return this._differencePage;
};

/**
 * Get the next IID id
 * @returns {Number} Next ID
 */
DocumentController.prototype.getNextIID = function() {
	return this._nextIID++;
};

/**
 * Add a document
 * @param {Data.AbstractDocument} doc - Document to add
 * @param {Boolean} makeActive - Make the given document the active document or not
 */
DocumentController.prototype.addDocument = function(doc, makeActive) {
	this._documents.add(doc.getIID(), doc);

	this._emit("loaded", doc);

	if (makeActive) {
		if (doc.getDocumentVersion() !== undefined && doc.getDocumentVersion().getFirstPage() !== undefined) {
			this.changeActivePage(doc.getDocumentVersion().getFirstPage());
		}
	}
};

/**
 * Returns the current activated document
 *
 * @returns {Data.AbstractDocument} Get the active document or null if none is active
 */
DocumentController.prototype.getActiveDocument = function() {
	if (this.hasActiveDocumentPage()) {
		var doc = this._activePage.getDocument();
		if (!this.hasDocument(doc)) {
			this.f.emit("error.WARNING", "Trying to get an already removed document");
			return null;
		}
		return doc;
	} else {
		return null;
	}
};

/**
 * Returns the current difference document
 *
 * @returns {Data.AbstractDocument} Get the difference document or null if none is active
 */
DocumentController.prototype.getDifferenceDocument = function() {
	if (this.hasDifferenceDocumentPage()) {
		var doc = this._differencePage.getDocument();
		if (!this.hasDocument(doc)) {
			this.f.emit("error.WARNING", "Trying to get an already removed document");
			return null;
		}
		return doc;
	} else {
		return null;
	}
};

/**
 * Returns the current activated document version
 *
 * @returns {Data.AbstractDocumentVersion} Get the active document version or null if none is active
 */
DocumentController.prototype.getActiveDocumentVersion = function() {
	if (this.hasActiveDocumentPage()) {
		var docVersion = this._activePage.getDocumentVersion();
		if (!this.hasDocument(docVersion.getDocument())) {
			this.f.emit("error.WARNING", "Trying to get an already removed document version");
			return null;
		}
		return docVersion;
	} else {
		return null;
	}
};

/**
 * Returns the current difference document version
 *
 * @returns {Data.AbstractDocumentVersion} Get the difference document version or null if none is active
 */
DocumentController.prototype.getDifferenceDocumentVersion = function() {
	if (this.hasDifferenceDocumentPage()) {
		var docVersion = this._differencePage.getDocumentVersion();
		if (!this.hasDocument(docVersion.getDocument())) {
			this.f.emit("error.WARNING", "Trying to get an already removed document version");
			return null;
		}
		return docVersion;
	} else {
		return null;
	}
};

/**
 * check if document exists in the collection
 * @param document {AbstractDocument} - check if this document is part of the collection of documents
 * @returns {Boolean}
 */
DocumentController.prototype.hasDocument = function(document) {
	return this._documents.has(document.getIID());
};

/**
 * change the active page
 * @param {AbstractPage} page
 */
DocumentController.prototype.changeActivePage = function(page) {
	if (!page) {
		this.f.emit("error.WARNING", "Trying to change the active page to undefined");
	} else {
		var document = page.getDocument();
		var oldPage = this._activePage;

		var oldDocumentVersion = null;
		if (oldPage) {
			oldDocumentVersion = oldPage.getDocumentVersion();
		}
		if (this.hasDocument(document)) {
			var context = {
				oldActivePage: this._activePage,
				newActivePage: page,
				diffPage: this._differencePage
			};
			this.f.emit("intercept.activePageChange", context);

			if (context.diffPage !== this._differencePage) {
				this.changeDifferencePage(context.diffPage);
			}

			this._activePage = context.newActivePage;
			var newDocumentVersion = this._activePage.getDocumentVersion();

			if (oldPage != this._activePage) {
				this.willNeedParameterUpdate("ACTIVE_PAGE_CHANGE");
				this._activePageHasChanged = true;
				this._previousActivePage = oldPage;

				//reset the state
				if (oldPage) {
					oldPage.resetState();
				}

				var finalPartOfActivePageChange = _.bind(function() {
					// check whether we should set a CAD overlay or not
					if (this._activePage.hasLinkedPage()) {
						// the page has a linked CAD file, set it as cadOverlay
						var cadOverlay = this._activePage.getLinkedPage();
						this.changeCadOverlay(cadOverlay);
					} else {
						// reset cadOverlay
						this.changeCadOverlay();
					}

					this._emit("page.active", oldPage, this._activePage);

					if (oldDocumentVersion != newDocumentVersion) {
						this._emit("documentVersion.active", oldDocumentVersion, newDocumentVersion);
					}

					this.readyForParameterUpdate("ACTIVE_PAGE_CHANGE");
				}, this);

				if (this._activePage.needsDataUpdate()) {
					this._adapter.updateDocumentData(
						this._activePage,
						_.bind(function() {
							finalPartOfActivePageChange();
						}, this)
					);
				} else {
					finalPartOfActivePageChange();
				}
				return true;
			} else {
				return false;
			}
		}
	}
};

/**
 * Reset the difference page to undefined
 */
DocumentController.prototype.resetDifferencePage = function() {
	var oldPage = this._differencePage;
	if (oldPage !== undefined) {
		this.willNeedParameterUpdate("RESET_DIFFERENCE_PAGE");
		this._differencePageHasChanged = true;
		this._previousDifferencePage = oldPage;
		this._differencePage = undefined;

		oldPage.resetState();

		this._emit("page.difference", oldPage, this._differencePage);

		this.readyForParameterUpdate("RESET_DIFFERENCE_PAGE");
	}
};

/**
 * Change the reference page
 * @param {AbstractPage} page
 */
DocumentController.prototype.changeDifferencePage = function(page) {
	if (page === undefined) {
		this.resetDifferencePage();
	} else {
		var document = page.getDocument();
		var oldPage = this._differencePage;

		if (this.hasDocument(document)) {
			this._differencePage = page;

			if (oldPage != this._differencePage) {
				this.willNeedParameterUpdate("DIFFERENCE_PAGE_CHANGE");
				this._differencePageHasChanged = true;
				this._previousDifferencePage = oldPage;

				// reset the state
				if (oldPage) {
					oldPage.resetState();
				}

				this._emit("page.difference", oldPage, this._differencePage);

				this.readyForParameterUpdate("DIFFERENCE_PAGE_CHANGE");
			}
		}
	}
};

/**
 * Register a callback for when the active document version changes, if there is already a page active,  the callback is called  with the currently active documentversion
 * @param {Function} callback - Called when active document version changes (oldDoc, newDoc)
 * @param {Object} [binding] - bind the callback to "binding"
 */
DocumentController.prototype.whenActiveDocumentVersionChanged = function(callback, binding) {
	if (binding) {
		callback = _.bind(callback, binding);
	}

	var activeDocument = this.getActiveDocumentVersion();
	if (activeDocument) {
		callback(undefined, activeDocument);
	}

	this.on("documentVersion.active", callback);
};

/**
 * Register a callback for when new documents are loaded. If all documents are already loaded the callback is called immediately.
 * There is already an active page set when this callback is executed.
 * @param {Function} callback
 * @param {Object} binding
 * @param {Boolean} keepListening - indicate if the callback needs to be called for every "loaded"-event
 */
DocumentController.prototype.whenDocumentsLoaded = function(callback, binding, keepListening) {
	if (binding) {
		callback = _.bind(callback, binding);
	}
	if (this._documentsLoading === 0 && this._documents.size() > 0) {
		// documents are loaded
		callback(this.getAllDocuments());
	}

	if (keepListening || this._documentsLoading !== 0) {
		// document are not loaded yet
		this.on(
			"loaded",
			_.bind(function(doc) {
				if (this._documentsLoading === 0) {
					//call the callback with all loaded documents
					callback(this.getAllDocuments());
				}
			}, this)
		);
	}
};

/**
 * Register a callback for when the documents collection changes. First checks if the documents are loaded if so register listener to "loaded"-event.
 * @param {Function} callback
 * @param {Object} binding
 * @param {Boolean} keepListening - indicate if the callback needs to be called for every "loaded"-event
 */
DocumentController.prototype.whenDocumentsLoadedOrChange = function(callback, binding) {
	this.whenDocumentsLoaded(callback, binding, true);
};

/**
 * When changes to loaded documents are made, call this function to let listeners know (needed for synchronisation problems)
 */
DocumentController.prototype.loadedDocumentChanged = function() {
	this._emit("loaded");
};

/**
 * Change the CAD overlay
 * @param {AbstractPage} page
 */
DocumentController.prototype.changeCadOverlay = function(page) {
	var oldPage = this._cadOverlay;

	if (page && this._activePage.hasCad()) {
		this.f.debug.error("Setting a CAD overlay is not allowed when the active document is a CAD document");
		page = undefined;
	}

	if (oldPage !== page) {
		this._cadOverlay = page;
		this.willNeedParameterUpdate("CAD_OVERLAY_CHANGE");
		this._cadOverlayHasChanged = true;
		this._previousCadOverlay = oldPage;

		// reset the state
		if (oldPage) {
			oldPage.resetState();
		}
		this._emit("page.cadOverlay", oldPage, this._cadOverlay);

		this.readyForParameterUpdate("CAD_OVERLAY_CHANGE");
	}
};

/**
 * Get CAD Overlay document (if any)
 * @returns {AbstractDocumentPage} the CAD overlay page
 */
DocumentController.prototype.getCadOverlay = function() {
	return this._cadOverlay;
};
/**
 * Register a callback for when the active page changes, if there is already a page active, the callback is called with the currently active page
 * @param {Function} callback - parameters: (oldPage, newPage)
 * @param {Object} [binding] - bind the callback to "binding"
 */
DocumentController.prototype.whenActivePageChanged = function(callback, binding) {
	if (binding) {
		callback = _.bind(callback, binding);
	}

	var page = this.getActivePage();
	if (page) {
		callback(undefined, page);
	}

	this.on("page.active", callback);
};

/**
 * Register a callback for when the difference page changes, if there is already a page active, the callback is called with the currently active page
 * @param {Function} callback - parameters: (oldPage, newPage)
 * @param {Object} [binding] - bind the callback to "binding"
 */
DocumentController.prototype.whenDifferencePageChanged = function(callback, binding) {
	if (binding) {
		callback = _.bind(callback, binding);
	}

	var page = this.getDifferencePage();
	if (page) {
		callback(undefined, page);
	}

	this.on("page.difference", callback);
};

/**
 * Register a callback for when the active or difference page changes. if there is already an active or difference page, the callback is called immediatily.
 * This callback occurs before the parameters are available or are being fetched!
 * @param {Function} callback - the callback, this callback has no parameters
 * @param {Object} [binding] - binds the callback to binding
 */
DocumentController.prototype.whenActiveOrDifferencePageChanged = function(callback, binding) {
	if (binding) {
		callback = _.bind(callback, binding);
	}

	if (this.getActivePage() || this.getDifferencePage()) {
		callback();
	}

	this.on("page.difference", callback);
	this.on("page.active", callback);
};

/**
 * call the callback when the active pages' parameters change
 * @param {Function} callback - parameters: activePage, diffPage
 * @param {Object} [binding] - bind the callback to "binding"
 */
DocumentController.prototype.whenActivePageParametersChange = function(callback, binding) {
	if (binding) {
		callback = _.bind(callback, binding);
	}

	this.on("activePageParametersChange", callback);
};

/**
 * call the callback when the reference pages' state parameters change
 * @param {Function} callback - parameters: activePage, diffPage
 * @param {Object} [binding] - bind the callback to "binding"
 */
DocumentController.prototype.whenReferencePageStateParametersChange = function(callback, binding) {
	if (binding) {
		callback = _.bind(callback, binding);
	}

	this.on("referencePageStateParametersChanged", callback);
};

/**
 * call the callback when the cad overlay pages' state parameters change
 * @param {Function} callback - parameters: activePage, diffPage
 * @param {Object} [binding] - bind the callback to "binding"
 */
DocumentController.prototype.whenCadOverlayPageStateParametersChange = function(callback, binding) {
	if (binding) {
		callback = _.bind(callback, binding);
	}

	this.on("cadOverlayPageStateParametersChanged", callback);
};

/**
 * call the callback when the difference pages' parameters change
 * @param {Function} callback - parameters: acitvePage, diffPage
 * @param {Object} [binding] - bind the callback to "binding"
 */
DocumentController.prototype.whenDifferencePageParametersChange = function(callback, binding) {
	if (binding) {
		callback = _.bind(callback, binding);
	}

	this.on("differencePageParametersChange", callback);
};

/**
 * Register a callback which is called when the active page has changed and its parameters are available
 * @param {Function} callback - parameters: oldPage, page
 * @param {Object} [binding] - bind the callback to "binding"
 */
DocumentController.prototype.whenActivePageChangedAndParametersAvailable = function(callback, binding) {
	if (binding) {
		callback = _.bind(callback, binding);
	}

	this.on("activePageChangedAndParametersAvailable", callback);
	if (this.hasActiveDocumentPageWithParametersAvailable()) {
		callback(undefined, this.getActivePage());
	} else if (this.isParameterUpdateOngoing() && !this._activePageHasChanged) {
		this.whenParametersAvailable(function(activePage, differencePage) {
			callback(undefined, activePage);
		});
	}
};

/**
 * Register a callback which is called when the active page has changed and its parameters are available
 * or when the parameters of the active page are changed
 * @param {Function} callback - parameters: oldPage, activePage
 * @param {Object} [binding] - bind the callback to "binding"
 */
DocumentController.prototype.whenActivePageChangedAndParametersAvailableOrChange = function(callback, binding) {
	if (binding) {
		callback = _.bind(callback, binding);
	}

	this.on("activePageChangedAndParametersAvailableOrChange", callback);
	if (this.hasActiveDocumentPageWithParametersAvailable()) {
		callback(undefined, this.getActivePage());
	}
};

/**
 * Register a callback which is called when the difference page has changed and its parameters are available
 * @param {Function} callback - parameters: oldDiffPage, diffPage
 * @param {Object} [binding] - bind the callback to "binding"
 */
DocumentController.prototype.whenDifferencePageChangedAndParametersAvailable = function(callback, binding) {
	if (binding) {
		callback = _.bind(callback, binding);
	}

	this.on("differencePageChangedAndParametersAvailable", callback);
	if (this.hasDifferenceDocumentPageWithParametersAvailable()) {
		callback(undefined, this._differencePage);
	} else if (this.isParameterUpdateOngoing() && this._differencePage && !this._differencePageHasChanged) {
		this.whenParametersAvailable(function(activePage, differencePage) {
			callback(undefined, differencePage);
		});
	}
};

/**
 * Register a callback which is called when the difference page has changed and its parameters are available
 * or when the parameters of the difference page are changed
 * @param {Function} callback - parameters: oldDiffPage,diffPage
 * @param {Object} [binding] - bind the callback to "binding"
 */
DocumentController.prototype.whenDifferencePageChangedAndParametersAvailableOrChange = function(callback, binding) {
	if (binding) {
		callback = _.bind(callback, binding);
	}

	this.on("differencePageChangedAndParametersAvailableOrChange", callback);
	if (this.hasDifferenceDocumentPageWithParametersAvailable()) {
		callback(undefined, this.getDifferencePage());
	}
};

/**
 * Register a callback to be called when the cadOverlay is available and changes
 * @param {Function} callback - parameters : oldCadOverlay, newCadOverlay
 * @param {[[Type]]} binding  [[Description]]
 */
DocumentController.prototype.whenCadOverlayChanged = function(callback, binding) {
	if (binding) {
		callback = _.bind(callback, binding);
	}
	this.on("page.cadOverlay", callback);

	if (this._cadOverlay) {
		callback(undefined, this._cadOverlay);
	}
};

/**
 * Register a callback which is called when the CAD overlay page has changed and its parameters are available
 * @param {Function} callback  - parameters: oldCadOverlayPage, cadOverlayPage
 * @param {Object}   [binding] - bind the callback to "binding"
 */
DocumentController.prototype.whenCadOverlayChangedAndParametersAvailable = function(callback, binding) {
	if (binding) {
		callback = _.bind(callback, binding);
	}

	this.on("cadOverlayChangedAndParametersAvailable", callback);
	if (this.hasCadOverlaywWithParametersAvailable()) {
		callback(undefined, this.getCadOverlay());
	} else if (this.isParameterUpdateOngoing() && this._cadOverlay && !this._cadOverlayHasChanged) {
		this.whenParametersAvailable(function(activePage, differencePage) {
			callback(undefined, this._cadOverlay);
		}, this);
	}
};

/**
 * Register a callback which is called when the difference page or active page has changed and its parameters are available
 * @param {Function} callback - parameters: activePage, diffPage, oldPage (either active or diff, depending on last boolean), activePageHasChanged
 * @param {Object} [binding] - bind the callback to "binding"
 */
DocumentController.prototype.whenActiveOrDifferencePageChangedAndParametersAvailable = function(callback, binding) {
	if (binding) {
		callback = _.bind(callback, binding);
	}

	this.on("activeOrDifferencePageChangedAndParametersAvailable", callback);
	if (this.hasActiveDocumentPageWithParametersAvailable() || this.hasDifferenceDocumentPageWithParametersAvailable()) {
		callback(this.getActivePage(), this.getDifferencePage(), undefined, true);
	} else if (this.isParameterUpdateOngoing() && !this._activePageHasChanged && !this._differencePageHasChanged) {
		this.whenParametersAvailable(function(activePage, differencePage) {
			callback(activePage, differencePage, undefined, true);
		});
	}
};

/**
 * Callback is triggered when either active, difference or cad over has changed and the parameters are available
 * @param {Function} callback - parameters: activePage, differencePage, cadOverlay
 * @param {Object} [binding] - bind the callback to "binding"
 */
DocumentController.prototype.whenActiveOrDifferenceOrCadOverlayChangedAndParametersAvailable = function(callback, binding) {
	if (binding) {
		callback = _.bind(callback, binding);
	}

	this.on("activeOrDifferenceOrCadOverlayChangedAndParametersAvailable", callback);
	if (
		this.hasActiveDocumentPageWithParametersAvailable() ||
		this.hasDifferenceDocumentPageWithParametersAvailable() ||
		this.hasCadOverlaywWithParametersAvailable()
	) {
		callback(this.getActivePage(), this.getDifferencePage(), this.getCadOverlay());
	} else if (
		this.isParameterUpdateOngoing() &&
		!this._activePageHasChanged &&
		!this._differencePageHasChanged &&
		!this._cadOverlayHasChanged
	) {
		this.whenParametersAvailable(function(activePage, differencePage) {
			callback(this._activePage, this._differencePage, this._cadOverlay);
		}, this);
	}
};
/**
 * Register a callback which is called when the active or difference page has changed and its parameters are available or change
 * @param {Function} callback - parameters: activePage, diffPage, oldPage (either active or diff, depending on last boolean), activePageHasChanged
 * @param {Object} [binding] - bind the callback to "binding"
 */
DocumentController.prototype.whenActiveOrDifferencePageChangedAndParametersAvailableOrChange = function(callback, binding) {
	if (binding) {
		callback = _.bind(callback, binding);
	}

	this.on("activeOrDifferencePageChangedAndParametersAvailableOrChange", callback);
	if (this.hasActiveDocumentPageWithParametersAvailable() || this.hasDifferenceDocumentPageWithParametersAvailable()) {
		callback(this.getActivePage(), this.getDifferencePage(), undefined, true);
	}
};

/**
 * Let the active document fill out the needed parameters
 * @param context
 */
DocumentController.prototype.interceptFragmentURL = function(context) {
	if (this._activePage) {
		this._activePage.interceptFragmentURL(context);
	}
};

/**
 * Check is the document with documentId has multiple versions
 * @param {Number} document IID
 * @returns {Boolean}
 */
DocumentController.prototype.hasDocumentMultipleVersions = function(documentIID) {
	return this._documents.get(documentIID).hasMultipleVersions();
};

/**
 * Get the number of documents
 * @returns {Number}
 */
DocumentController.prototype.getDocumentCount = function() {
	return this._documents.size();
};

/**
 * Get all loaded documents as a non-sorted array
 * @returns {Array.<AbstractDocument>} list of all loaded documents
 */
DocumentController.prototype.getDocuments = function() {
	return this._documents.values();
};

/**
 * Get all documents
 * @returns {Array.<AbstractDocument>}
 */
DocumentController.prototype.getAllDocuments = function() {
	var documents = _.sortBy(this._documents.values(), function(document) {
		if (document.get("docname")) {
			return document.get("docname").toLowerCase();
		} else {
			return document.getIID();
		}
	});
	return documents;
};

export default DocumentController;
