import _ from "underscore";
import Extendable from "@esko/web-viewer/core/Extendable";
import PrivateChannel from "@esko/web-viewer/core/PrivateChannel";
import ViewerUtils from "@esko/web-viewer/utils/ViewerUtils";
import PreferencesDefaults from "@esko/web-viewer/data/PreferencesDefaults";
/**
 * Represents a user
 *
 * @exports User
 * @memberof Data
 * @constructor
 * @param {String} id - Id of the user
 * @param {String} username - User name of the user
 * @param {String} firstname - The first name of the user
 * @param {String} lastname - Last name of the user
 * @parma {String} [locale] - the locale of the user (defaults to "en")
 */
var User = function User(id, username, first, lastname, locale) {
	/**
	 * @private
	 */
	this.f = null;

	/**
	 * Debounced saving
	 */
	this.persistDebounced = _.debounce(_.bind(this._persistUserOptions, this), 1000);

	Extendable.installTo(this);

	/**
	 * In _options all user preferences are stored. This state could be
	 * persisted by a LayoutManager
	 *
	 * All changes to the _options should be done with User.setOption so the
	 * correct events are sent out
	 *
	 * @access private
	 */
	this._options = {};

	/**
	 * In here previous values are stored when an option is not protected
	 * @private
	 */
	this._originalNonPersistedValues = {};

	/**
	 * @returns {String} Id of the user
	 */
	this.getID = function() {
		return id;
	};

	this.username = username;

	this.lastname = lastname;

	this.first = first;

	if (locale === undefined) {
		locale = "en";
	}
	this.locale = locale;

	/**
	 * Set the facade of the user, can be executed only once!
	 */
	this.setFacade = _.once(function(facade) {
		this.f = facade;
		PrivateChannel.installTo(this, this.f, ["user", id]);
	});
};

/**
 * Get full name of the user
 * @returns {String} Full name of the user
 */
User.prototype.getFullName = function() {
	return this.getUsername();
};

/**
 * Get last name of the user
 * @returns {String} Last name of the user
 */
User.prototype.getLastName = function() {
	return this.lastname;
};

/**
 * Get the first name of the user
 * @returns {String} first name of the user
 */
User.prototype.getFirstName = function() {
	return this.first;
};

/**
 * @returns {String} Username of the user
 */
User.prototype.getUsername = function() {
	return this.username;
};

/**
 * Get locale of the user
 * @returns {String} Locale of the user
 */
User.prototype.getLocale = function() {
	return this.locale;
};

/**
 * Change an option of the user
 *
 * If an option is changed, an event is send on
 * <user.[ID].options.changed.[key]>
 *
 * @param {String} key
 * @param {Object} newValue
 * @param {Boolean} [doNotPersist] if true the new value is not persisted!
 * @param {Function} [onPersistedCallback] - Callback executed when the option is persisted. Called with parameter boolean successful (true) saved or not (false)
 */
User.prototype.setOption = function(key, newValue, doNotPersist, onPersistedCallback) {
	var oldValue = this._options[key];
	if (oldValue !== newValue) {
		this._options[key] = newValue;
		this._emit("options.changed." + key, oldValue, newValue);
		if (!doNotPersist) {
			delete this._originalNonPersistedValues[key];
			if (onPersistedCallback) {
				this._persistUserOptions(onPersistedCallback);
			} else {
				this.persistDebounced();
			}
		} else if (this._originalNonPersistedValues[key] === undefined) {
			this._originalNonPersistedValues[key] = oldValue;
		}
	} else if (onPersistedCallback) {
		onPersistedCallback(true);
	}
};

/**
 * Change multiple options of the user. This can be used to inject the
 * state of the user
 *
 * @param options
 *            {key: value, key2: value2}
 */
User.prototype.setOptions = function(options) {
	_.each(
		options,
		function(value, key) {
			this.setOption(key, value);
		},
		this
	);
};

/**
 * Initialize an option (this doesn't trigger  a persist operation)
 * @param {String} key
 * @param {Object} newValue
 */
User.prototype.initializeOption = function(key, value) {
	var oldValue = this._options[key];
	if (oldValue !== value) {
		this._options[key] = value;
		this._emit("options.changed." + key, oldValue, value);
	}
};

/**
 * intialize the options of the user. This can be used to inject the
 * state of the user
 *
 * @param options
 *            {key: value, key2: value2}
 */
User.prototype.initializeOptions = function(options) {
	_.each(
		options,
		function(value, key) {
			this.initializeOption(key, value);
		},
		this
	);
};

/**
 * Get the options of the user. This method can be used to save the
 * state of a user
 *
 * @returns object with all options of the user
 */
User.prototype.getOptions = function() {
	return this._options;
};

/**
 * Get an option of the user
 * @param {String} option - Name of the option
 * @returns value of the option
 */
User.prototype.getOption = function(option) {
	return this._options[option];
};

/**
 * Get a hex color from an option
 */
User.prototype.getColor = function(option) {
	return "#" + ViewerUtils.hexFromInt(this.getOption(option));
};

/**
 * Set a hex color
 */
User.prototype.setColor = function(option, hexcolor) {
	if (hexcolor.indexOf("#") > -1) {
		hexcolor = hexcolor.replace("#", "");
	}
	this.setOption(option, ViewerUtils.intFromHex(hexcolor));
};

/**
 * persist the options to the backend
 * @param {Object} options - {key: value, key2: value2}
 * @param {Function} [onPersistedCallback] - Callback executed when the options are persisted. Called with parameter boolean successful (true) saved or not (false).
 */
User.prototype.persistOptions = function(options, onPersistedCallback) {
	// override
};

/**
 * Load/initialize the user options from the backend
 * @param {Function} onLoadedCallback - Callback executed when the options are loaded
 */
User.prototype.loadOptions = function(onLoadedCallback) {
	// override
};

/**
 * Reset an option to its default value
 * @param {String} option - the option to reset
 */
User.prototype.resetOption = function(option) {
	this.setOption(option, PreferencesDefaults[option]);
};

/**
 * Get the given option default value
 * @param {String} option - Name of the option
 * @returns {String | Number} Value of the option
 */
User.prototype.getOptionDefault = function(option) {
	return PreferencesDefaults[option];
};

/**
 * Save the user preferences
 *
 * @param {function} [onPersistedCallback] - called when preferences are saved
 */
User.prototype._persistUserOptions = function(onPersistedCallback) {
	// clone the current options
	var options = _.clone(this._options);
	// possible some options were overridden but shouldn't be persisted, restore these
	_.each(this._originalNonPersistedValues, function(value, key) {
		options[key] = value;
	});

	this.persistOptions(options, onPersistedCallback);
};

export default User;
