import CadConstants from "@esko/web-viewer/cad/CadConstants";
import ListenerUtil from "@esko/web-viewer/utils/ListenerUtil";
import DynamIcon from "@esko/web-viewer/layout/component/DynamIcon";
import $ from "jquery";
import _ from "underscore";
/**
 * View that displays information icons of the supplied document page (CAD-file). The icons change whenever parameters
 * of the page change and if the viewport rotates or flips.
 *
 * @memberof DocumentInfoModule
 * @exports DocumentInfoIcons
 * @class
 * @constructor
 * @param {Core.ViewerFacade} f
 */
var DocumentInfoIcons = function(f) {
	ListenerUtil.mixin(this, f);
	this.f = f;

	/**
	 * @member {AbstractDocumentPage}
	 */
	this._documentPage = null;

	/**
	 * @member {JQueryObj} - wrapper DOM element for the icons
	 */
	this._cadInfoIcons = $("<span>").addClass("activeDocumentInfo");

	/**
	 * @member {ViewportModule} - viewport module to fetch rotation/flip information of the world
	 */
	this._viewport = null;

	/**
	 *  Information icons for the Document page:
	 */
	this._defaultSide = new DynamIcon(this.f, CadConstants.DefaultSide.OUTSIDE, "icon-WCR_CAD-Outside-cont", "icon-WCR_CAD-Outside-sub");
	this._defaultSide.addIcon(CadConstants.DefaultSide.INSIDE, "icon-WCR_CAD-Inside-cont", "icon-WCR_CAD-Inside-sub");

	this._grainFluteDirection = new DynamIcon(this.f, CadConstants.GrainFluteDirection.VERTICAL, "icon-WCR_CAD-Direction-V");
	this._grainFluteDirection.addIcon(CadConstants.GrainFluteDirection.HORIZONTAL, "icon-WCR_CAD-Direction-H");

	this._dieSide = new DynamIcon(this.f, CadConstants.DieSide.BACKSIDE, "icon-WCR_CAD-Die-cont", "icon-WCR_CAD-DieBack-sub");
	this._dieSide.addIcon(CadConstants.DieSide.KNIFESIDE, "icon-WCR_CAD-Die-cont", "icon-WCR_CAD-DieKnife-sub");

	this._defaultSide.addTooltip(CadConstants.DefaultSide.OUTSIDE, "viewer", "OutsideDefaultSideTT");
	this._defaultSide.addTooltip(CadConstants.DefaultSide.INSIDE, "viewer", "InsideDefaultSideTT");
	this._grainFluteDirection.addTooltip(CadConstants.GrainFluteDirection.VERTICAL, "viewer", "VerticalGrainFluteTT");
	this._grainFluteDirection.addTooltip(CadConstants.GrainFluteDirection.HORIZONTAL, "viewer", "HorizontalGrainFluteTT");
	this._dieSide.addTooltip(CadConstants.DieSide.BACKSIDE, "viewer", "BackSideTT");
	this._dieSide.addTooltip(CadConstants.DieSide.KNIFESIDE, "viewer", "KnifeSideTT");

	//save the viewport module which can be used to fetch rotation/flip information of the document page
	this.f.whenModuleAvailable(
		"viewport",
		function(viewportModule) {
			this._viewport = viewportModule;
			this.updateIcons();
		},
		this
	);

	//register several listeners to update the icons (rotation/flip)
	this._registerListener("doc.parameters.changed", this.updateIcons, this);
	this._registerListener("transform.viewportWidget.rotated", this.updateIcons, this);
	this._registerListener("transform.viewportWidget.flipped", this.updateIcons, this);
};

/**
 * Update the information icons.
 */
DocumentInfoIcons.prototype.updateIcons = function() {
	//first check if the rotation/flip state of the document and the world causes the icons to change
	var isSideSwitched = false;
	var isGrainFluteSwitched = false;
	var isDieSideSwitched = false;
	if (this._viewport && this._documentPage) {
		var pageRotation = (this._viewport.getWorld().getRotation() + this._documentPage.getRotation()) % 360;
		isGrainFluteSwitched = pageRotation === 90 || pageRotation === 270;
		isSideSwitched = this._viewport.getWorld().isFlip() !== this._documentPage.isFlipped(); //logical XOR operation
		isDieSideSwitched = this._viewport.getWorld().isFlip() !== this._documentPage.isFlipped();
	}

	//update the default side icon
	if (this._documentPage && this._documentPage.get("defaultSide")) {
		var defaultSide = this._documentPage.get("defaultSide");

		if (isSideSwitched) {
			defaultSide = defaultSide == CadConstants.DefaultSide.OUTSIDE ? CadConstants.DefaultSide.INSIDE : CadConstants.DefaultSide.OUTSIDE;
		}
		//show the correct icon
		this._defaultSide.show(defaultSide);
	} else {
		this._defaultSide.hide();
	}
	//update the flute/grain direction icon
	if (this._documentPage && this._documentPage.get("grainDirection")) {
		var grainFluteDirection = this._documentPage.get("grainDirection");

		if (isGrainFluteSwitched) {
			grainFluteDirection =
				grainFluteDirection == CadConstants.GrainFluteDirection.VERTICAL
					? CadConstants.GrainFluteDirection.HORIZONTAL
					: CadConstants.GrainFluteDirection.VERTICAL;
		}

		this._grainFluteDirection.show(grainFluteDirection);
	} else {
		this._grainFluteDirection.hide();
	}
	//update the die side icon (this is only for mfg-files)
	if (this._documentPage && this._documentPage.get("dieSide")) {
		var dieSide = this._documentPage.get("dieSide");

		if (isDieSideSwitched) {
			dieSide = dieSide == CadConstants.DieSide.KNIFESIDE ? CadConstants.DieSide.BACKSIDE : CadConstants.DieSide.KNIFESIDE;
		}

		this._dieSide.show(dieSide);
	} else {
		this._dieSide.hide();
	}
};

/**
 * Set a new document page.
 * @param {AbstractDocumentPage} documentPage
 */
DocumentInfoIcons.prototype.setDocumentPage = function(documentPage) {
	this._documentPage = documentPage;
	this.updateIcons();
};

/**
 * Get the page icons to be displayed.
 * @returns {JQueryObj}
 */
DocumentInfoIcons.prototype.getPageIcons = function() {
	var documentInfoIcons = $("<span>");
	documentInfoIcons
		.append(this._grainFluteDirection.getElement())
		.append(this._defaultSide.getElement())
		.append(this._dieSide.getElement());
	return documentInfoIcons;
};

export default DocumentInfoIcons;
