import AbstractModule from "@esko/web-viewer/core/AbstractModule";
import DocumentSelection from "@esko/web-viewer/layout/component/DocumentSelection";
import DocumentInfoWidget from "@esko/web-viewer/documentInfo/DocumentInfoWidget";
import InfoWidget from "@esko/web-viewer/documentInfo/InfoWidget";
import DebugWidget from "@esko/web-viewer/documentInfo/DebugWidget";
import DocumentInfoModel from "@esko/web-viewer/documentInfo/DocumentInfoModel";
import MouseLocatorTool from "@esko/web-viewer/documentInfo/MouseLocatorTool";
import Mousetrap from "@esko/mousetrap";
import LayoutConstants from "@esko/web-viewer/layout/LayoutConstants";
import _ from "underscore";
import ToolbarComponent from "@esko/toolbar/ToolbarComponent";
import CoreConstants from "@esko/web-viewer/core/CoreConstants";
/**
 * This module is responsible for adding the document info widget (topbar)
 *
 * @exports DocumentInfoModule
 * @constructor
 * @augments Core.AbstractModule
 * @param {Core.ViewerFacade} f
 * @param {Boolean} silenced
 */
var DocumentInfoModule = function(f, silenced, overrideClasses) {
	AbstractModule.call(this, f, silenced, "documentInfo");

	overrideClasses = overrideClasses || {};

	/**
	 * @member {Layout.Component.DocumentSelection} - mobile document selector, active document only
	 */
	this._documentSelection = null;

	// create the widget
	this._publishWidget(
		DocumentInfoWidget.widgetName,
		_.bind(function(layoutType) {
			this._documentInfoWidget = new (overrideClasses.DocumentInfoWidget || DocumentInfoWidget)(f);
			return this._documentInfoWidget;
		}, this)
	);

	// create model, create tool
	this.documentInfoModel = new DocumentInfoModel(this.f, "documentInfo");
	var mouseLocatorTool = new MouseLocatorTool(f, this.documentInfoModel);
	this._publishTool(mouseLocatorTool);

	// create the Info widget
	this._publishWidget(
		InfoWidget.widgetName,
		_.bind(function(layoutType) {
			this._infoWidget = new (overrideClasses.InfoWidget || InfoWidget)(f, this.documentInfoModel);
			return this._infoWidget;
		}, this)
	);

	// only publish the debugging-infowidget once!
	this._showDebugWidget = _.once(
		_.bind(function() {
			this._publishWidget(DebugWidget.widgetName, function(layoutType) {
				return new DebugWidget(f);
			});
		}, this)
	);

	/**
	 * Toolbar Component: Document Selector
	 */
	this.f.i18n.loadBundle(
		"viewer",
		function() {
			// the DocumentSelection only works correctly once the general language bundle is available
			this._publishToolbarComponent(
				LayoutConstants.ToolbarComponent.DOCUMENTSELECTOR,
				_.bind(function(layoutType) {
					this._documentSelection = new DocumentSelection(this.f, undefined, undefined, true, true);
					var documentSelectorComponent = new ToolbarComponent(LayoutConstants.ToolbarComponent.DOCUMENTSELECTOR, f);
					documentSelectorComponent.addObject(this._documentSelection.getActiveDocumentSelector());
					return documentSelectorComponent;
				}, this)
			);
		},
		this
	);

	if (process.env.NODE_ENV === "development") {
		// on developer systems we can just enter "d e b u g (enter)" to open the widget
		// note, this code is not included in the compiled version!
		Mousetrap.bind("d e b u g enter", this._showDebugWidget);
	}

	// if the debug parameter is provided, show the info widget
	if (this.f.getProperty("debug")) {
		this._showDebugWidget();
	}

	// Listen to changes in selection (used in e.g. measure tools) and update model accordingly
	this.f.on("selection.changed", _.bind(this.documentInfoModel.setSelection, this.documentInfoModel));

	// Keyboard shortcuts to toggle units
	Mousetrap.bind("ctrl+u", _.bind(this._setNextSizeUnit, this), "keydown");
	Mousetrap.bind("ctrl+shift+u", _.bind(this._setNextResolutionUnit, this), "keydown");
};

// inherit
DocumentInfoModule.prototype = Object.create(AbstractModule.prototype);
// correct the constructor pointer
DocumentInfoModule.prototype.constructor = DocumentInfoModule;

/**
 * Set the next size unit (circular)
 * @param {Event} e - Event object
 */
DocumentInfoModule.prototype._setNextSizeUnit = function(e) {
	var currentSizeUnit = this.f.getUnits();
	var nextSizeUnit = (currentSizeUnit + 1) % _.size(CoreConstants.SizeUnits);
	this.f.setOption("PREFERENCES_UNIT_SIZE", nextSizeUnit, true);

	e.preventDefault();
	e.stopPropagation();
};

/**
 * Set the next resolution unit (circular)
 * @param {Event} e - Event object
 */
DocumentInfoModule.prototype._setNextResolutionUnit = function(e) {
	var currentResUnit = this.f.getResolutionUnits();
	var nextResUnit = (currentResUnit + 1) % _.size(CoreConstants.ResolutionUnits);
	this.f.setOption("PREFERENCES_UNIT_RESOLUTION", nextResUnit, true);
	this.f.setOption("PREFERENCES_UNIT_SCREENRULING", nextResUnit, true);

	e.preventDefault();
	e.stopPropagation();
};

export default DocumentInfoModule;
