import AbstractWidget from "@esko/web-viewer/layout/AbstractWidget";
import Template from "@esko/template";
import tpl from "@esko/web-viewer/documentInfo/DocumentInfoWidget.tpl.html";
import DocumentInfoIcons from "@esko/web-viewer/documentInfo/DocumentInfoIcons";
import ToolbarComponent from "@esko/toolbar/ToolbarComponent";
import ButtonGroup from "@esko/toolbar/ButtonGroup";
import MiniButton from "@esko/web-viewer/layout/component/MiniButton";
import ModalPopup from "@esko/web-viewer/layout/component/ModalPopup";
import $ from "jquery";
import _ from "underscore";
import CompareConstants from "@esko/web-viewer/compare/CompareConstants";
import ListenerUtil from "@esko/web-viewer/utils/ListenerUtil";
import ViewerUtils from "@esko/web-viewer/utils/ViewerUtils";
import Tooltip from "@esko/tooltip/Tooltip";
import LayoutConstants from "@esko/web-viewer/layout/LayoutConstants";
/**
 * @constant {Number} the speed to render the widget
 */
var RENDER_SPEED = 40;

/**
 * @exports DocumentInfoWidget
 * @augments Layout.AbstractWidget
 * @class The DocumentInfoWidget displays the currently loaded document and logged in user
 * @constructor
 * @mixes {Utils.ListenerUtilMixin}
 * @param {Core.ViewerFacade} f
 */
var DocumentInfoWidget = function DocumentInfoWidget(f) {
	AbstractWidget.call(this, f, LayoutConstants.Widget.DOCUMENTINFO);

	ListenerUtil.mixin(this, f);

	this.templateContext = $("<div>", {
		class: "infoSection"
	});
	this.buttonsComponent = new ToolbarComponent("buttonsComponent", f);
	this.isListeningToCompare = false;
	this.diffMode = -1;
	this.compareMode = null;

	this._activePageInfoIcons = new DocumentInfoIcons(f);
	this._referencePageInfoIcons = new DocumentInfoIcons(f);

	var returnURL = f.getProperty("returnurl");

	this.buttonGroup = new ButtonGroup("buttonsGroup");
	this.buttonGroup.element.addClass("pull-right");
	this.buttonsComponent.addButtonGroup(this.buttonGroup);

	// If return URL set, add close viewer button
	if (returnURL) {
		var returnButton = new MiniButton(f, "exit", "icon-WCR_Close12", true, function() {
			f.emit("viewer.request.close", returnURL);
		});
		returnButton.element.addClass("pull-right");
		returnButton.setTooltip("viewer", "exitViewerBtnTT");
		this.buttonGroup.addButton(returnButton);
	}

	//show a help button
	var helpButton = new MiniButton(f, "help", "icon-WCR_Menu-Help", true, function() {
		var helpModule = this.f.getModule("help");
		if (helpModule) {
			helpModule.openHelpPages();
		}
	});
	helpButton.element.addClass("pull-right");
	helpButton.setTooltip("viewer", "HelpTT");
	helpButton.hide();

	//show the shortcuts help button
	var shortcutsButton = new MiniButton(f, "shortcuts", "icon-WCR_KeyboardShortcuts-keyboard", true, function() {
		var helpModule = this.f.getModule("help");
		if (helpModule) {
			helpModule.openShortcutsPopup();
		}
	});
	shortcutsButton.element.addClass("pull-right");
	shortcutsButton.setTooltip("viewer", "ShortcutsTT");
	shortcutsButton.hide();

	this.buttonGroup.addButton(helpButton);
	this.buttonGroup.addButton(shortcutsButton);
	this.f.whenModuleAvailable(
		"help",
		function() {
			helpButton.show(); // show if help module available
			shortcutsButton.show();
		},
		this
	);

	// Viewer options button
	var resetLayoutButton = new MiniButton(
		f,
		"resetLayout",
		"icon-WCR_ResetLayout",
		true,
		_.bind(function() {
			this._resetLayout();
		}, this)
	);
	resetLayoutButton.element.addClass("pull-right");
	resetLayoutButton.setTooltip("viewer", "ResetLayout");
	this.buttonGroup.addButton(resetLayoutButton);
	this._resetLayoutPopup = null;

	// Show button to switch to mobile layout (if the device has touch)
	this.f.requestModule(
		"featureDetector",
		_.bind(function(featureDetector) {
			if (featureDetector.isMobile() || featureDetector.hasTouch()) {
				var switchToMobileButton = new MiniButton(
					this.f,
					"switchToMobile",
					"icon-WCR_Mobile",
					true,
					_.bind(function() {
						this.f.switchLayout(LayoutConstants.LayoutType.MOBILE);
					}, this)
				);
				switchToMobileButton.setTooltip("viewer", "SwitchToMobile");
				this.buttonGroup.prependButton(switchToMobileButton);
			}
		}, this)
	);

	// initialize template
	this.t = new Template(f.i18n, tpl);

	this.render = _.throttle(_.bind(this._render, this), RENDER_SPEED);
	// load language if needed
	this.f.loadBundle("viewer", this.render, this);
	this.f.getDocumentController().whenActiveOrDifferencePageChangedAndParametersAvailableOrChange(this.render, this);

	this.f.whenModuleAvailable(
		"print",
		function(printModule) {
			var printButton = new MiniButton(
				f,
				"print",
				"icon-WCR_Print",
				true,
				_.bind(function() {
					if (this.compareMode === CompareConstants.mode.BLINKING) {
						this.f.emit("error.WARNING", "Cannot print when blinking compare mode is active.");
					} else {
						printModule.openPrintPopup();
					}
				}, this)
			);
			printButton.element.addClass("pull-right");
			printButton.setTooltip("viewer", "PrintTT");
			this.buttonGroup.addButton(printButton);
		},
		this
	);

	this.f.whenModuleAvailable(
		"compare",
		function(compareModule) {
			this._registerListener(
				"changed.diffMode",
				function(oldValue, newValue) {
					this.diffMode = newValue;
					this.render();
				},
				this,
				compareModule
			);
			this._registerListener(
				"changed.mode",
				function(oldMode, newMode) {
					if (newMode === null) {
						this.compareMode = null;
					} else {
						this.compareMode = newMode.getName();
					}
				},
				this,
				compareModule
			);
			this.diffMode = compareModule.getDiffMode();
		},
		this
	);

	// Page controls
	this.t.bind("click", "a[name='firstPage']", this._firstPage, this);
	this.t.bind("click", "a[name='previousPage']", this._previousPage, this);
	this.t.bind("click", "a[name='nextPage']", this._nextPage, this);
	this.t.bind("click", "a[name='lastPage']", this._lastPage, this);
	this.t.bind("mousedown", "input[name='currentPage']", this._currentPageClick, this);
	this.t.bind("keydown", "input[name='currentPage']", this._currentPageEnter, this);

	// add tooltips
	this.whenContextAvailable(function() {
		this.tooltip = new Tooltip(this.f.i18n, this.context, this.f.getContext(), this.f.getContext(), this.f.isMobile(), "");
		this.tooltip.disableTooltip();
		// filter on all elements with attribute tooltip that are not disabled
		this.t.bind(
			"mouseover",
			"[tooltip]:not('.disabled')",
			function(source) {
				this.tooltip.setTooltipContent($(source).attr("tooltip"));
				this.tooltip.context = $(source);
				this.tooltip.show();
			},
			this
		);
		// filter on all elements with attribute tooltip that are not disabled
		this.t.bind(
			"mouseout",
			"[tooltip]:not('.disabled')",
			function() {
				this.tooltip.hide();
			},
			this
		);
	}, this);
};

DocumentInfoWidget.prototype = Object.create(AbstractWidget.prototype);
DocumentInfoWidget.prototype.constructor = DocumentInfoWidget;
DocumentInfoWidget.widgetName = LayoutConstants.Widget.DOCUMENTINFO;

/**
 * @override
 */
DocumentInfoWidget.prototype.setContext = function(context) {
	AbstractWidget.prototype.setContext.call(this, context);
	this.context.append($("<div>").append(this.templateContext, this.buttonsComponent.element));
	this.context.addClass("enableTextSelection");
	this.t.setContext(this.templateContext);
	this.context.addClass("documentInfoWidget");
	this.render();
};

/**
 * Set the needed data for the widget
 */
DocumentInfoWidget.prototype._render = function() {
	var docController = this.f.getDocumentController();
	var activePage = docController.getActivePage();
	var diffPage = docController.getDifferencePage();
	var overlay = docController.getCadOverlay();

	// reset template vars
	this.t.d.diffDocument = "";
	this.t.d.diffVersion = "";
	this.t.d.diffClass = "";
	this.t.d.documentClass = "";

	if (activePage) {
		// fill in active page details
		var doc = activePage.getDocument();
		this.t.d.extraTitles = [];
		this.t.d.document = doc.get("docname") || "";
		this.t.d.version = activePage.getDocumentVersion().get("docversion") || "";
		if (activePage.getDocumentVersion().get("docRevision")) {
			this.t.d.version += "(" + activePage.getDocumentVersion().get("docRevision") + ")";
		}
		this.t.d.showPageControls = activePage.getDocumentVersion().isMultiPage();
		this.t.d.pageNumber = activePage.getPagenum();
		this.t.d.pageCount = activePage.getDocumentVersion().getPageCount();
		this.t.d.firstPageClass =
			"mini-btn" +
			(activePage.getPagenum() <=
			activePage
				.getDocumentVersion()
				.getFirstPage()
				.getPagenum()
				? " disabled"
				: "");
		this.t.d.previousPageClass =
			"mini-btn" +
			(activePage.getPagenum() <=
			activePage
				.getDocumentVersion()
				.getFirstPage()
				.getPagenum()
				? " disabled"
				: "");
		this.t.d.nextPageClass =
			"mini-btn" +
			(activePage.getPagenum() >=
			activePage
				.getDocumentVersion()
				.getLastPage()
				.getPagenum()
				? " disabled"
				: "");
		this.t.d.lastPageClass =
			"mini-btn" +
			(activePage.getPagenum() >=
			activePage
				.getDocumentVersion()
				.getLastPage()
				.getPagenum()
				? " disabled"
				: "");

		// if there is a diff page, show this information
		if (diffPage !== undefined && this.compareMode !== CompareConstants.mode.READERSPREAD) {
			doc = diffPage.getDocument();
			this.t.d.diffDocument = doc.get("docname") || "";
			this.t.d.diffVersion = diffPage.getDocumentVersion().get("docversion") || "";
			if (diffPage.getDocumentVersion().get("docRevision")) {
				this.t.d.diffVersion += "(" + diffPage.getDocumentVersion().get("docRevision") + ")";
			}
			if (this.diffMode === CompareConstants.diffMode.DIFFVERSION || this.diffMode === CompareConstants.diffMode.SHOWDIFFERENCE3D) {
				this.t.d.diffClass = "active";
			} else if (
				this.diffMode === CompareConstants.diffMode.CURRENTVERSION ||
				this.diffMode === CompareConstants.diffMode.SHOWCURRENT3D ||
				this.diffMode === CompareConstants.diffMode.UNDEFINED
			) {
				this.t.d.documentClass = "active";
			}
		}
	}
	// render user
	this.t.d.user = this.f.getCurrentUser().getFullName();

	this.t.d = this._extendTplData(this.t.d);

	// render the template
	this.t.render();

	if ((activePage && activePage.hasCad()) || overlay) {
		var activeCad = overlay ? overlay : activePage;
		var activePageInfo = this.templateContext.find(".activeDocumentInfo");
		if (activePageInfo.length > 0) {
			this._activePageInfoIcons.setDocumentPage(activeCad);
			activePageInfo.append(this._activePageInfoIcons.getPageIcons());
		}
	}
	if (diffPage && diffPage.hasCad()) {
		var referencePageInfo = this.templateContext.find(".referenceDocumentInfo");
		if (referencePageInfo.length > 0) {
			this._referencePageInfoIcons.setDocumentPage(diffPage);
			referencePageInfo.append(this._referencePageInfoIcons.getPageIcons());
		}
	}
};

DocumentInfoWidget.prototype._extendTplData = function(data) {
	return data;
};

/**
 * @override
 * @returns {String}
 */
DocumentInfoWidget.prototype.getWidgetType = function() {
	return "topbar";
};

/**
 * Change to first page
 */
DocumentInfoWidget.prototype._firstPage = function() {
	var pageListModule = this.f.getModule("pagelist");
	if (pageListModule) {
		pageListModule.firstPage();
	} else {
		this.f.emit("error.SEVERE", "Pagelist module not loaded.");
	}
};

/**
 * Change to previous page
 */
DocumentInfoWidget.prototype._previousPage = function() {
	var pageListModule = this.f.getModule("pagelist");
	if (pageListModule) {
		pageListModule.previousPage();
	} else {
		this.f.emit("error.SEVERE", "Pagelist module not loaded.");
	}
};

/**
 * Change to next page
 */
DocumentInfoWidget.prototype._nextPage = function() {
	var pageListModule = this.f.getModule("pagelist");
	if (pageListModule) {
		pageListModule.nextPage();
	} else {
		this.f.emit("error.SEVERE", "Pagelist module not loaded.");
	}
};

/**
 * Change to last page
 */
DocumentInfoWidget.prototype._lastPage = function() {
	var pageListModule = this.f.getModule("pagelist");
	if (pageListModule) {
		pageListModule.lastPage();
	} else {
		this.f.emit("error.SEVERE", "Pagelist module not loaded.");
	}
};

/**
 * Executed on click of the input page control
 * @param el
 * @param event
 */
DocumentInfoWidget.prototype._currentPageClick = function(el, event) {
	if (!$(el).is(":focus")) {
		$(el).trigger("select");
		event.preventDefault();
	}
};

/**
 * Executed on enter of the input page control
 * @param el
 * @param event
 */
DocumentInfoWidget.prototype._currentPageEnter = function(el, event) {
	if (event.which == 13 || event.keyCode == 13) {
		var value = $(el).val();
		var parsedValue = parseInt(value, 10);
		if (!isNaN(parsedValue)) {
			// valid number
			var activeDocumentVersion = this.f.getDocumentController().getActiveDocumentVersion();
			// check boundaries
			if (parsedValue < 1) {
				parsedValue = 1;
			} else if (parsedValue > activeDocumentVersion.getPageCount()) {
				parsedValue = activeDocumentVersion.getPageCount();
			}
			// change page
			var newPage = activeDocumentVersion.getPage(parsedValue);
			if (newPage && newPage !== this.f.getDocumentController().getActivePage()) {
				this.f.getDocumentController().changeActivePage(newPage);
			} else {
				if (!newPage) {
					this.f.emit("userMessage", "warn", this.f.i18n.t("WarningGetNonexistentPage"), true);
				}
				// re-render to reset
				this.render();
			}
		} else {
			// not valid number, re-render to reset
			this.render();
		}
		event.preventDefault();
	}
};

/**
 * This method performs the reset of the layout (sends event to layout manager)
 * It first shows a popup to confirm.
 */
DocumentInfoWidget.prototype._resetLayout = function() {
	if (!this._resetLayoutPopup) {
		this._resetLayoutPopup = new ModalPopup(this.f, "resetLayoutConfirm", this.f.getContext());
		this._resetLayoutPopup.setOkAction(
			_.bind(function() {
				this.f.emit("layout.resetToDefault");
				this._resetLayoutPopup.hide();
			}, this)
		);
		// Header + Content
		this._resetLayoutPopup.setHeader(
			$("<span/>")
				.addClass("popupTitle")
				.text(this.f.i18n.t("ResetLayout"))
		);
		this._resetLayoutPopup.setContent(
			$("<div/>")
				.addClass("unsavedItemsContent")
				.text(this.f.i18n.t("ResetLayoutConfirm"))
		);
		this._resetLayoutPopup.setOkBtnText(this.f.i18n.t("YesResponse"));
		this._resetLayoutPopup.setCancelBtnText(this.f.i18n.t("NoResponse"));
	}
	this._resetLayoutPopup.show();
	this._resetLayoutPopup.recalculateSize(450);
};

export default DocumentInfoWidget;
