import AbstractWidget from "@esko/web-viewer/layout/AbstractWidget";
import Template from "@esko/template";
import LayoutConstants from "@esko/web-viewer/layout/LayoutConstants";
import tpl from "@esko/web-viewer/documentInfo/InfoWidget.tpl.html";
import CoreConstants from "@esko/web-viewer/core/CoreConstants";
import CompareConstants from "@esko/web-viewer/compare/CompareConstants";
import ListenerUtil from "@esko/web-viewer/utils/ListenerUtil";
import ViewerUtils from "@esko/web-viewer/utils/ViewerUtils";
import ShapeUtils from "@esko/shapeutils";
import Coordinates from "@esko/web-viewer/utils/Coordinates";
import DropdownMiniButton from "@esko/web-viewer/layout/component/DropdownMiniButton";
import Tooltip from "@esko/tooltip/Tooltip";
import _ from "underscore";
/**
 * @exports InfoWidget
 * @augments Layout.AbstractWidget
 * @class The InfoWidget shows some information on the document and the viewport
 * @constructor
 * @mixes {Utils.ListenerUtilMixin}
 * @param {Core.ViewerFacade} f
 *  @param {DocumentInfo.DocumentInfoModel} documentInfoModel
 */
var InfoWidget = function InfoWidget(f, documentInfoModel) {
	AbstractWidget.call(this, f, LayoutConstants.Widget.INFO);
	ListenerUtil.mixin(this, f);

	this._model = documentInfoModel;

	// initialize template
	this.t = new Template(f.i18n, tpl);

	/**
	 * {Layout.Component.Tooltip} tooltip for resolution
	 */
	this.tooltip = null;

	// default settings
	/**
	 * {CoreConstants.SizeUnits}
	 */
	this._sizeUnit = this.f.getOption("PREFERENCES_UNIT_SIZE");

	/**
	 * {CoreConstants.ResolutionUnits}
	 */
	this._resUnit = this.f.getOption("PREFERENCES_UNIT_RESOLUTION");

	/**
	 * {boolean}
	 */
	this._showDiffPage = false;

	/**
	 * {MeasureConstants.MeasureDrawType}
	 */
	this.t.d.selectionType = null;

	/**
	 * {MeasureConstants.MeasureDrawType}
	 */
	this._previousSelectionType = null;

	/**
	 * {Coordinates} - util for coordinate conversion
	 */
	this._coordinates = new Coordinates(this.f);

	/**
	 * Make the options dropdown
	 */
	this._optionButton = new DropdownMiniButton(f, "infoUnitSettings", "icon-WCR_Settings-submenu");

	this.setIcon("icon icon-WCR_Info");

	f.i18n.loadBundle(
		"viewer",
		function() {
			this.setTitle(this.f.i18n.t("InfoWidgetTitle"));

			// Units for size
			function toggleSizeUnit(mode, toggle) {
				if (toggle) {
					this.setSizeUnit(mode);
				}
			}

			this._optionButton.addOption({
				name: "sizeUnitPixels",
				content: this.f.i18n.t(this.f.getUnitsLocName(CoreConstants.SizeUnits.PIXELS)),
				onchange: _.bind(toggleSizeUnit, this, CoreConstants.SizeUnits.PIXELS),
				selected: this._sizeUnit === CoreConstants.SizeUnits.PIXELS,
				userCannotUncheck: true
			});

			this._optionButton.addOption({
				name: "sizeUnitMm",
				content: this.f.i18n.t(this.f.getUnitsLocName(CoreConstants.SizeUnits.MM)),
				onchange: _.bind(toggleSizeUnit, this, CoreConstants.SizeUnits.MM),
				selected: this._sizeUnit === CoreConstants.SizeUnits.MM,
				userCannotUncheck: true
			});

			this._optionButton.addOption({
				name: "sizeUnitCm",
				content: this.f.i18n.t(this.f.getUnitsLocName(CoreConstants.SizeUnits.CM)),
				onchange: _.bind(toggleSizeUnit, this, CoreConstants.SizeUnits.CM),
				selected: this._sizeUnit === CoreConstants.SizeUnits.CM,
				userCannotUncheck: true
			});

			this._optionButton.addOption({
				name: "sizeUnitInch",
				content: this.f.i18n.t(this.f.getUnitsLocName(CoreConstants.SizeUnits.INCH)),
				onchange: _.bind(toggleSizeUnit, this, CoreConstants.SizeUnits.INCH),
				selected: this._sizeUnit === CoreConstants.SizeUnits.INCH,
				userCannotUncheck: true
			});

			this._optionButton.addDivider();

			// units for resolution
			function toggleResolutionUnit(mode, toggle) {
				if (toggle) {
					this.setResolutionUnit(mode);
				}
			}

			this._optionButton.addOption({
				name: "resUnitPpmm",
				content: this.f.i18n.t("UnitResmm"),
				onchange: _.bind(toggleResolutionUnit, this, CoreConstants.ResolutionUnits.PPMM),
				selected: this._resUnit === CoreConstants.ResolutionUnits.PPMM,
				userCannotUncheck: true
			});

			this._optionButton.addOption({
				name: "resUnitPpcm",
				content: this.f.i18n.t("UnitRescm"),
				onchange: _.bind(toggleResolutionUnit, this, CoreConstants.ResolutionUnits.PPCM),
				selected: this._resUnit === CoreConstants.ResolutionUnits.PPCM,
				userCannotUncheck: true
			});

			this._optionButton.addOption({
				name: "resUnitPpi",
				content: this.f.i18n.t("UnitResInch"),
				onchange: _.bind(toggleResolutionUnit, this, CoreConstants.ResolutionUnits.PPI),
				selected: this._resUnit === CoreConstants.ResolutionUnits.PPI,
				userCannotUncheck: true
			});

			this._optionButton.setTooltip("viewer", "UnitSettings");

			this.whenContextAvailable(
				_.bind(function() {
					this.f.getDocumentController().whenActiveOrDifferencePageChangedAndParametersAvailableOrChange(function() {
						// if we have a CAD file, we need to show points instead of pixels and we don't shown resolution.
						// Note that viewerfacade handles the units, we just need to refresh when this changed
						var hasCad = this._getInfoPage() && this._getInfoPage().hasCad();
						if (hasCad) {
							this.f.requestModule(
								"cad",
								_.bind(function() {
									this._getInfoPage().requestCad(_.bind(this.render, this));
								}, this),
								true
							);
						}

						var isCadOnly = hasCad && !this._getInfoPage().hasGraphics();
						this.t.d.isCadOnly = isCadOnly;
						this._optionButton.setActionText("sizeUnitPixels", this.f.i18n.t(this.f.getUnitsLocName(CoreConstants.SizeUnits.PIXELS)));
						this._optionButton.toggleOption("sizeUnitPixels", this._sizeUnit === CoreConstants.SizeUnits.PIXELS); // recover icon if selected

						this.render();
					}, this);

					// listen to diffMode, so we can show the info of the correct document (active or difference)
					this._registerListener(
						"module.compare.changed.diffMode",
						function(oldMode, newMode) {
							this._showDiffPage = newMode === CompareConstants.diffMode.DIFFVERSION;
							this.render();
						},
						this
					);

					// rerender when unit settings change
					this.f.getCurrentUser().on(
						"options.changed.PREFERENCES_UNIT_SIZE",
						_.bind(function(oldValue, newValue) {
							this.setSizeUnit(newValue);
						}, this)
					);
					this.f.getCurrentUser().on(
						"options.changed.PREFERENCES_UNIT_RESOLUTION",
						_.bind(function(oldValue, newValue) {
							this.setResolutionUnit(newValue);
						}, this)
					);

					// rerender when mouselocation changes
					this._model.on(
						"changed.mouselocation",
						_.bind(function() {
							this.updateMouseLocation();
						}, this)
					);

					// rerender when mouselocation changes
					this._model.on(
						"changed.selection",
						_.bind(function() {
							this.updateSelection();
						}, this)
					);
				}, this)
			);
		},
		this
	);
};

InfoWidget.prototype = Object.create(AbstractWidget.prototype);
InfoWidget.prototype.constructor = InfoWidget;
InfoWidget.widgetName = LayoutConstants.Widget.INFO;

/**
 * @override
 */
InfoWidget.prototype.setContext = function(context, headerExtra) {
	AbstractWidget.prototype.setContext.call(this, context, headerExtra);

	this.context.addClass("enableTextSelection infoWidget");
	this.t.setContext(this.context);

	// add buttons to the widget header
	this.headerExtra.append(this._optionButton.element);

	this.render();
};

/**
 * Render the widget.
 * When information cannot be shown, make sure the template does not show it by removing the data.
 */
InfoWidget.prototype.render = function() {
	if (this.isVisible()) {
		// get size and resolution data
		try {
			var sizeUnit = this.f.getUnits();
			var sizeUnitLoc = this.f.getUnitsLocName();
			var sizeUnitFix = this.f.getUnitsFix();

			var resUnit = this.f.getResolutionUnits();
			var resUnitLoc = this.f.getResolutionUnitsLocName();
			var resUnitFix = this.f.getResolutionUnitsFix();

			// for calculations, use resolution of current page (can be active or diff)
			var currentPage = this._getInfoPage();
			var resX = ViewerUtils.parseFloat(currentPage.getParam("resolutionx"));
			resX = resX > 0 ? resX : undefined;
			var resY = ViewerUtils.parseFloat(currentPage.getParam("resolutiony"));
			resY = resY > 0 ? resY : undefined;

			var documentSize = currentPage.getPageSize();
			var margin = currentPage.getMargins();

			if (documentSize) {
				// convert size and resolution to correct units and put in template data
				this.t.d.documentSize = new ShapeUtils.Size(
					ViewerUtils.convertFromMmValue(documentSize.width, sizeUnit, sizeUnitFix, resX),
					ViewerUtils.convertFromMmValue(documentSize.height, sizeUnit, sizeUnitFix, resY)
				);
				var trimbox = new ShapeUtils.Rect(
					new ShapeUtils.Point(margin.left, margin.top),
					new ShapeUtils.Size(documentSize.width - margin.left - margin.right, documentSize.height - margin.top - margin.bottom)
				);
				var trimboxSize = trimbox.size;
				this.t.d.trimboxSize = new ShapeUtils.Size(
					ViewerUtils.convertFromMmValue(trimboxSize.width, sizeUnit, sizeUnitFix, resX),
					ViewerUtils.convertFromMmValue(trimboxSize.height, sizeUnit, sizeUnitFix, resY)
				);

				// for display, show both active and difference page
				var activePage = this.f.getActivePage();
				var resXActive = ViewerUtils.parseFloat(activePage.getParam("resolutionx"));
				var resYActive = ViewerUtils.parseFloat(activePage.getParam("resolutiony"));
				this.t.d.resolutionX = ViewerUtils.convertResolutionFromPpmmValue(resXActive, resUnit, resUnitFix);
				this.t.d.resolutionY = ViewerUtils.convertResolutionFromPpmmValue(resYActive, resUnit, resUnitFix);
				this.t.d.resolutionUnit = resUnitLoc;

				var diffPage = this.f.getDocumentController().getDifferencePage();
				if (diffPage) {
					var resXDiff = ViewerUtils.parseFloat(diffPage.getParam("resolutionx"));
					var resYDiff = ViewerUtils.parseFloat(diffPage.getParam("resolutiony"));
					if (resXDiff && resYDiff) {
						this.t.d.resolutionXDiff = ViewerUtils.convertResolutionFromPpmmValue(resXDiff, resUnit, resUnitFix);
						this.t.d.resolutionYDiff = ViewerUtils.convertResolutionFromPpmmValue(resYDiff, resUnit, resUnitFix);
					}
				} else {
					//the world size (from parameters) needs to be the same as the document size (since there is no reference document page)
					var worldDocumentSize = new ShapeUtils.Size(
						ViewerUtils.parseFloat(currentPage.getParam("sizex")),
						ViewerUtils.parseFloat(currentPage.getParam("sizey"))
					);
					if (
						!currentPage.hasCad() &&
						(!ViewerUtils.isCloseTo(worldDocumentSize.width, documentSize.width, 1) ||
							!ViewerUtils.isCloseTo(worldDocumentSize.height, documentSize.height, 1))
					) {
						this.t.d.documentSize = new ShapeUtils.Size(
							ViewerUtils.convertFromMmValue(worldDocumentSize.width, sizeUnit, sizeUnitFix, resX),
							ViewerUtils.convertFromMmValue(worldDocumentSize.height, sizeUnit, sizeUnitFix, resY)
						);
					}
					this.t.d.resolutionXDiff = null;
					this.t.d.resolutionYDiff = null;
				}
			} else {
				// nothing can be shown
				this.t.d.documentSize = null;
				this.t.d.trimboxSize = null;
			}
			// set units localized
			this.t.d.sizeUnit = sizeUnitLoc;

			this.updateMouseLocation(true);
			this.updateSelection(true);

			this._renderWidget();
		} catch (e) {
			// TODO DEV DEBUG.
			console.warn("could not render info widget: " + e);
		}
	}
};

/**
 * Update the mouse location and rerender the template.
 * Mouse point is converted to selected units.
 * @param {boolean} skipRender - don't render (render is done outside the function)
 */
InfoWidget.prototype.updateMouseLocation = function(skipRender) {
	var mousePos = this._model.getMousePosition();
	if (mousePos) {
		var activePage = this._getInfoPage();
		var resX = ViewerUtils.parseFloat(activePage.getParam("resolutionx"));
		var resY = ViewerUtils.parseFloat(activePage.getParam("resolutiony"));

		//convert to correct coordinate system
		var mousePoint = this._coordinates.getTransformedPoint(new ShapeUtils.Point(mousePos.x, mousePos.y));

		// convert and pass infor to template
		this.t.d.position = {
			x: ViewerUtils.convertFromMmValue(mousePoint.x, this.f.getUnits(), this.f.getUnitsFix(), resX),
			y: ViewerUtils.convertFromMmValue(mousePoint.y, this.f.getUnits(), this.f.getUnitsFix(), resY)
		};
	} else {
		// mouse is not in viewport, don't show coordinates
		this.t.d.position = null;
	}

	if (!skipRender) {
		this._renderWidget();
	}
};

/**
 * Update the selection and rerender the template
 * Selection is converted to selected units
 * @param {boolean} skipRender - don't render (render is done outside the function)
 */
InfoWidget.prototype.updateSelection = function(skipRender) {
	var selection = this._model.getSelection();
	var fireResized = false;
	if (selection === null) {
		// nothing can be shown
		this.t.d.selectionType = null;
		fireResized = this._previousSelectionType !== null;
		this._previousSelectionType = null;
	} else {
		fireResized = this._previousSelectionType !== selection.mode;
		this._previousSelectionType = selection.mode;
		this.t.d.selectionType = selection.mode;
		this.t.d.selection = {};

		// handle offsets and convert values
		var activePage = this._getInfoPage();
		var resX = ViewerUtils.parseFloat(activePage.getParam("resolutionx"));
		var resY = ViewerUtils.parseFloat(activePage.getParam("resolutiony"));
		var offsetX = ViewerUtils.parseFloat(activePage.get("offsetx"));
		if (isNaN(offsetX)) {
			offsetX = 0;
		}
		var offsetY = ViewerUtils.parseFloat(activePage.get("offsety"));
		if (isNaN(offsetY)) {
			offsetY = 0;
		}

		//convert to correct coordinate system
		var selectionPoint = this._coordinates.getTransformedPoint(new ShapeUtils.Point(selection.x, selection.y));

		// convert to selected units
		this.t.d.selection.x =
			selectionPoint !== undefined ? ViewerUtils.convertFromMmValue(selectionPoint.x, this.f.getUnits(), this.f.getUnitsFix(), resX) : null;
		this.t.d.selection.y =
			selectionPoint !== undefined ? ViewerUtils.convertFromMmValue(selectionPoint.y, this.f.getUnits(), this.f.getUnitsFix(), resY) : null;
		this.t.d.selection.w =
			selection.w !== undefined ? ViewerUtils.convertFromMmValue(selection.w, this.f.getUnits(), this.f.getUnitsFix(), resX) : null;
		this.t.d.selection.h =
			selection.h !== undefined ? ViewerUtils.convertFromMmValue(selection.h, this.f.getUnits(), this.f.getUnitsFix(), resY) : null;
		if (selection.d !== undefined && selection.a !== undefined) {
			// use combined resolution for diagonal
			var angleRad = (selection.a * Math.PI) / 180;
			var combinedRes = Math.sqrt(
				Math.cos(angleRad) * Math.cos(angleRad) * resX * resX + Math.sin(angleRad) * Math.sin(angleRad) * resY * resY
			);
			this.t.d.selection.d =
				selection.d !== undefined
					? ViewerUtils.convertFromMmValue(selection.d, this.f.getUnits(), this.f.getUnitsFix(), combinedRes)
					: null;
			if (selection.a !== undefined) {
				this.t.d.selection.a = this.f.getOption("PREFERENCES_RULER_ANGLE_MODE") ? -selection.a.toFixed(2) : selection.a.toFixed(2);
			} else {
				this.t.d.selection.a = null;
			}
		}
	}

	if (!skipRender) {
		this._renderWidget();

		// resize widget if needed
		if (fireResized) {
			this._fireResized();
		}
	}
};

/**
 * Set the unit for size. This does not persist.
 * @param {CoreConstants.SizeUnits} mode
 */
InfoWidget.prototype.setSizeUnit = function(mode) {
	if (this._sizeUnit !== mode) {
		this._sizeUnit = mode;

		// update toggles
		this._optionButton.toggleOption("sizeUnitPixels", mode === CoreConstants.SizeUnits.PIXELS);
		this._optionButton.toggleOption("sizeUnitMm", mode === CoreConstants.SizeUnits.MM);
		this._optionButton.toggleOption("sizeUnitCm", mode === CoreConstants.SizeUnits.CM);
		this._optionButton.toggleOption("sizeUnitInch", mode === CoreConstants.SizeUnits.INCH);

		this.f.setOption("PREFERENCES_UNIT_SIZE", mode, true);

		this.render();
	}
};

/**
 * Set the unit for resolution. This does not persist.
 * @param {CoreConstants.ResolutionUnits} mode
 */
InfoWidget.prototype.setResolutionUnit = function(mode) {
	if (this._resUnit !== mode) {
		this._resUnit = mode;

		// update toggles
		this._optionButton.toggleOption("resUnitPpmm", mode === CoreConstants.ResolutionUnits.PPMM);
		this._optionButton.toggleOption("resUnitPpcm", mode === CoreConstants.ResolutionUnits.PPCM);
		this._optionButton.toggleOption("resUnitPpi", mode === CoreConstants.ResolutionUnits.PPI);

		this.f.setOption("PREFERENCES_UNIT_RESOLUTION", mode, true);

		this.render();
	}
};

/**
 * Get the page which we should shown information of (active or difference)
 * @param {boolean} getOther - get the page that is not shown instead of the one that's shown
 * @returns {AbstractDocumentPage} active or difference page
 */
InfoWidget.prototype._getInfoPage = function(getOther) {
	var getDiff = this._showDiffPage;
	if (getOther) {
		getDiff = !getDiff;
	}

	if (getDiff) {
		return this.f.getDocumentController().getDifferencePage();
	} else {
		return this.f.getActivePage();
	}
};

/**
 * Render Widget and create tooltip
 */
InfoWidget.prototype._renderWidget = function(getOther) {
	this.t.render();

	// enable tooltip
	if (
		(this.t.d.resolutionX && this.t.d.resolutionXDiff && this.t.d.resolutionX !== this.t.d.resolutionXDiff) ||
		(this.t.d.resolutionY && this.t.d.resolutionYDiff && this.t.d.resolutionY !== this.t.d.resolutionYDiff)
	) {
		if (this.tooltip) {
			this.tooltip.setContext(this.context.find(".infoWidgetResolution"));
			this.tooltip.enableTooltip();
		} else {
			this.tooltip = new Tooltip(
				this.f.i18n,
				this.context.find(".infoWidgetResolution"),
				this.f.getContext(),
				this.f.getContext(),
				this.f.isMobile(),
				this.f.i18n.t("InfoWidgetResCurrentRefTT")
			);
		}
	}
};

export default InfoWidget;
