import PrivateChannel from "@esko/web-viewer/core/PrivateChannel";
import AbstractModel from "@esko/web-viewer/core/AbstractModel";
import ShapeUtils from "@esko/shapeutils";
import ViewerUtils from "@esko/web-viewer/utils/ViewerUtils";
import GraphicsConstants from "@esko/web-viewer/graphics/GraphicsConstants";
import OrderedObject from "@esko/orderedobject";
import CompareConstants from "@esko/web-viewer/compare/CompareConstants";
import ViewControlsConstants from "@esko/web-viewer/viewControls/ViewControlsConstants";
import _ from "underscore";
/**
 * Simple datastructure for holding a page and a difference
 */
function PageGroup(page, difference) {
	this.page = page;
	this.difference = difference;
}

PageGroup.prototype.equals = function(ref) {
	if (this.page && this.difference) {
		return this.page.equals(ref.page) && this.difference.equals(ref.difference);
	}
	return this.page === ref.page && this.difference === ref.difference;
};

/**
 * @exports GraphicsModel
 * @memberof Graphics
 * @class The Graphics model is used for representing all graphical information of a page
 * @constructor
 * @augments {Core.AbstractModel}
 *
 * @param {Core.ViewerFacade} f - Viewer facade
 * @param {String} privateChannelName
 * @param {Data.AbstractDocumentPage} page - document page
 * @param {Data.AbstractDocumentPage} difference - the page that is set as a difference (for comparison)
 */
var GraphicsModel = function(f, privateChannelName, page, difference) {
	AbstractModel.call(this, f, privateChannelName);

	this.f = f;
	this.page = null;
	this.difference = null;

	/**
	 * compare mode
	 * @default 0 = no comparison
	 */
	this.dmode = 0;

	// compare module is loaded before graphics, it could be that the diffmode has changed already
	var compareModule = this.f.getModule("compare");
	if (compareModule) {
		this.dmode = compareModule.getDiffMode();
	}
	/**
	 * Keep track of the channels, the channels are copied from the document to enable restore of original values
	 * @private
	 */
	this._channels = new OrderedObject();

	/**
	 *  The default ordering of the channels
	 *  @private
	 */
	this._defaultOrder = null;

	/**
	 * @member {ViewControlsConstants.trimboxMode|null}
	 * @private
	 */
	this._trimboxMode = this.f.getOption("PREFERENCES_DISPLAY_MARGINS") === 1 ? this.f.getOption("PREFERENCES_MARGIN_DISPLAY_MODE") : null;

	/**
	 * True to reverse the channel order in the fragment url, false otherwise
	 * @member {Boolean}
	 */
	this._reverseChannelOrder = false;

	/**
	 * Also fetch image data from server when resolution is higher than pixel level
	 */
	// TODO: in the future when we have more drawing capabilities, we shouldn't fetch the high resolution images, but draw them ourselves
	this.allowServerDeepZoom = true;

	/**
	 * The current channel view mode
	 * @default Composite View
	 */
	this.channelViewMode = GraphicsConstants.ChannelViewMode.COMPOSITE_VIEW;

	/**
	 * The current single channel mode
	 */
	this.singleChannelMode = this.f.getOption("PREFERENCES_CHANNELS_SINGESEPBLACK");

	// set the page
	this.setPage(page, difference, true);
};

GraphicsModel.prototype = Object.create(AbstractModel.prototype);
GraphicsModel.prototype.constructor = GraphicsModel;

/**
 * Get the trimbox of the page
 * @returns {ShapeUtils.Rect} the page trimbox
 */
GraphicsModel.prototype.getTrimBox = function() {
	var trimboxRect = this.page.getTrimboxRect();
	if (this.page.isReaderSpread() && this.page.isBlankPage() && this.page.get("isRightPage") && this.difference) {
		var offset = this.difference.getTrimboxRect().size.width + this.difference.getMargins().left - this.page.getMargins().left;
		trimboxRect.origin.offset(new ShapeUtils.Size(offset, 0));
	} else if (this.page.isReaderSpread() && this.page.isBlankPage() && this.page.get("isRightPage")) {
		trimboxRect.origin.offset(new ShapeUtils.Size(this.page.getPageSize().width, 0));
	}
	return trimboxRect;
};

/**
 * Get the trimbox of the reference document page, returns null if there is no reference page
 * @returns {ShapeUtils.Rect|null} the reference page trimbox
 */
GraphicsModel.prototype.getReferenceTrimBox = function() {
	//only return a trimbox if there is a reference page, if in reader spread only return the trimbox if the correct mode is set
	if (this.difference && !(this.page.isReaderSpread() && this._trimboxMode === ViewControlsConstants.trimboxMode.ONCURRENTPAGE)) {
		var trimboxRect = this.difference.getTrimboxRect();
		if (this.difference.isBlankPage() && !this.page.get("isRightPage") && this.page) {
			var offset = this.page.getTrimboxRect().size.width + this.page.getMargins().left - this.difference.getMargins().left;
			trimboxRect.origin.offset(new ShapeUtils.Size(offset, 0));
		}
		return trimboxRect;
	} else {
		return null;
	}
};

/**
 * Set the page (and difference page) for the graphics
 * @param {Data.AbstractDocumentPage} page - the current active page
 * @param {Data.AbstractDocumentPage} difference - the currently configured difference page
 * @param {Boolean} [silenced] - if true no event will be send (for internal use only)
 */
GraphicsModel.prototype.setPage = function(page, difference, silenced) {
	var oldGroup = new PageGroup(this.page, this.difference);
	this.page = page;
	this.difference = difference;
	var newGroup = new PageGroup(this.page, this.difference);
	if (page !== undefined) {
		this.initChannels(oldGroup, newGroup, silenced);
	}

	this._fireChange(silenced, ["page"], oldGroup, newGroup);
};

/**
 * Get the active page that is set on the graphicsmodel
 * @returns {Data.AbstractDocumentPage}
 */
GraphicsModel.prototype.getPage = function() {
	return this.page;
};

/**
 * Get the reference page that is set on the graphicsmodel
 * @returns {Data.AbstractDocumentPage}
 */
GraphicsModel.prototype.getReferencePage = function() {
	return this.difference;
};

//------------ CHANNELS ------------------------------------------------------------------------------

/**
 * Initialize the channel data, set all channels as selected
 * @param {PageGroup} oldPageGroup - the page/difference group before the init
 * @param {PageGroup} newPageGroup - the current page/difference group
 * @param {Boolean} silenced - if true, do not send out events
 */
GraphicsModel.prototype.initChannels = function(oldPageGroup, newPageGroup, silenced) {
	// old Channels
	var oldChannels = this.getChannels();

	// here all the keys of the current existing channels are stored
	var existingChannels = [];

	// this is a help function to add the channels
	function addChannel(channel) {
		existingChannels.push(channel.getID());

		var newChannel = channel.copy();
		newChannel.toggle(true);

		var oldDefinition = this._channels.getItemByKey(channel.getID());
		if (oldDefinition) {
			// copy over previous settings
			newChannel.toggle(oldDefinition.isEnabled());
			newChannel.setOpaque(oldDefinition.isChannelOpaque());
			newChannel.setRgb(oldDefinition.getRGB());
			newChannel.setName(oldDefinition.getNameWithoutSuffix());
		}

		this._channels.addOrReplaceItem(newChannel.getID(), newChannel);
	}

	_.each(this.page.getChannels(true), addChannel, this);
	if (this.difference) {
		_.each(this.difference.getChannels(true), addChannel, this);
	}

	this._channels.omitOthers(existingChannels);

	if (this.difference) {
		// synchronize "enabled" flag for matched channels
		for (var i = 0; i < this._channels.getLength(); i++) {
			var channel = this._channels.getItemByIndex(i);
			if (channel.isInCurrentAndMatchedToPrevious()) {
				var matchedChannel = this.getMatchedChannel(channel);
				if (matchedChannel) {
					this._toggleChannel(matchedChannel, channel.isEnabled());
				}
			}
		}
	}

	this._defaultOrder = this._channels.getKeys().slice(0);
	this._fireChange(silenced, ["channels"], oldChannels, this.getChannels());
};

/**
 * Returns a channel
 * @param {Number} id
 * @returns {Channel} - Channel
 */
GraphicsModel.prototype.getChannel = function(id) {
	return this._channels.getItemByKey(id);
};

/**
 * Checks if there are changed channel definitions or the channel order has changed
 * @returns {Boolean} True if the channel definition is changed or the order is changed, false if there are no changes
 */
GraphicsModel.prototype.hasChanges = function() {
	//test if channels have changed color def
	var hasChangedChannels = _.some(this.getChannels(), function(channel) {
		return channel.isChanged();
	});

	if (hasChangedChannels || !this.hasDefaultOrder()) {
		return true;
	} else {
		return false;
	}
};

/**
 * Checks if the channels have the default order
 * @returns {Boolean} True if the channels have the default order, false if they have another order
 */
GraphicsModel.prototype.hasDefaultOrder = function() {
	var orderedKeys = this._channels.getKeys().slice(0);
	var i = (this._defaultOrder && this._defaultOrder.length) || 0;
	while (i--) {
		if (orderedKeys[i] !== this._defaultOrder[i]) {
			return false;
		}
	}
	return true;
};

/**
 * Toggle a channel (enable or disable it) (if another is channel is matched with this channel, the linked channel is also toggled
 * @param {Number} channelID - the unique identifier of the channel
 * @param {Boolean} toggle - enable/disable
 * @param {Boolean} [silent] - if true no event will be fired
 */
GraphicsModel.prototype.toggleChannel = function(channelID, toggle, silent) {
	var channel = this.getChannel(channelID);
	var matchedChannel = this.getMatchedChannel(channel);

	this._toggleChannel(channel, toggle, silent);
	if (matchedChannel !== null) {
		this._toggleChannel(matchedChannel, toggle, silent);
	}
};

/**
 * Toggle the channel without toggling any matched channels
 * @private
 * @param {Channel} channel
 * @param {Boolean} toggle - enable/disable
 * @param {Boolean} [silent] - if true no event will be fired
 */
GraphicsModel.prototype._toggleChannel = function(channel, toggle, silent) {
	var oldValue = channel.isEnabled();
	channel.toggle(toggle);

	// fire change if needed
	this._fireChange(silent, ["channel", channel.getID(), "enabled"], oldValue, toggle, channel.getID());
};

/**
 * Get the matched channel
 * @param {Graphics.Channel} channel - the channel for which you want the matched channel
 * @returns {Graphics.Channel} the matched channel (or null if none is defined)
 */
GraphicsModel.prototype.getMatchedChannel = function(channel) {
	if (channel.isMatched()) {
		for (var i = 0; i < this._channels.getLength(); i++) {
			var match = this._channels.getItemByIndex(i);
			if (match !== channel && match.getViewserverID() === channel.getMatchedChannelServerId()) {
				return match;
			}
		}
	}
	return null;
};

/**
 * Check if a certain channel is enabled or not
 * @param {Number} channelID - the unique identifier of the channel
 * @returns {Boolean}
 */
GraphicsModel.prototype.isChannelEnabled = function(channelID) {
	return this._channels.getItemByKey(channelID).isEnabled();
};

/**
 * Get the selection of the given channel only
 * @param {Channel} Channel where you want the selection of
 * @returns {Any} The selection of the given channel only
 */
GraphicsModel.prototype.getSpecificChannelSelection = function(channel) {
	return this.getChannelSelection([channel]);
};

/**
 * Get the selection of the given list of channels
 * @param {Channel} Channels where you want the selection of
 * @returns {Any} The selection of the given channel only
 */
GraphicsModel.prototype.getChannelSelection = function(channels) {
	throw "Not Implemented: GraphicsModel.getChannelSelection";
};

/**
 * Get the selection for a certain diffMode
 * @param diffMode
 * @returns {Number}
 */
GraphicsModel.prototype.getChannelSelectionForDiffMode = function(diffMode) {
	var channels = [];
	_.each(this.getChannels(false, diffMode), function(channel) {
		if (channel.isEnabled()) {
			channels.push(channel);
		}
	});
	return this.getChannelSelection(channels);
};

/**
 * Get the single channel mode
 * @returns {GraphicsConstants.SingleChannelMode}
 */
GraphicsModel.prototype.getSingleChannelMode = function() {
	return this.singleChannelMode;
};

/**
 * Set the single channel mode
 * @param {GraphicConstants.SingleChannelMode} mode
 */
GraphicsModel.prototype.setSingleChannelMode = function(mode) {
	var oldValue = this.getSingleChannelMode();

	if (oldValue != mode) {
		this.singleChannelMode = mode;
		this.f.getCurrentUser().setOption("PREFERENCES_CHANNELS_SINGESEPBLACK", mode);
		this._fireChange(false, ["singleChannelMode"], oldValue, mode);
	}
};

/**
 * Returns the channels that are not returned by getChannels(). (these are the channels of the difference document / active document, dependent on the diffMode)
 * @returns {Array.<GraphicsModule.Channel>}
 */
GraphicsModel.prototype.getDisabledChannels = function() {
	var channels = this._channels.getOrderedObjectList();
	return _.difference(channels, this.getChannels());
};

/**
 * Get all channels that are in the current viewserver setup (so all channels from both current and difference document)
 * @returns {Array.<GraphicsModule.Channel>} all channels of the loaded document (both active and difference file)
 */
GraphicsModel.prototype.getAllChannels = function() {
	return this._channels.getOrderedObjectList();
};

/**
 * Get all channels for the current visible document
 *
 * This depends on the diffMode:
 * 	0: only channels that are in the current document (and the channels that are in the current document and matched with the difference doc)
 *  1: only channels that are in the difference document (and the channels that are in the difference document and are matched with the current document)
 * @param {Boolean} filterMatchedChannels - should the matched channels be filter out
 * @param {CompareConstants.diffMode} [dmode] - this parameter is only for internal usage (it can force the model to give only channels for a certain dmode
 * @returns channel information
 */
GraphicsModel.prototype.getChannels = function(filterMatchedChannels, dmode) {
	var channels = this._channels.getOrderedObjectList();
	if (dmode === undefined) {
		dmode = this.dmode;
	}
	return _.filter(
		channels,
		function(channel) {
			if (
				dmode === CompareConstants.diffMode.SHOWBOTHCHANGES ||
				dmode === CompareConstants.diffMode.SHOWBOTHCHANGESHIGHLIGHT ||
				dmode === CompareConstants.diffMode.SHOWBOTHCHANGESONLY ||
				dmode === CompareConstants.diffMode.SHOWBOTHSIDEBYSIDE ||
				dmode === CompareConstants.diffMode.SHOWBOTHONTOP ||
				dmode === CompareConstants.diffMode.READERSPREAD ||
				dmode === CompareConstants.diffMode.SHOWBOTHSIDEBYSIDEHIGHLIGHT ||
				dmode === CompareConstants.diffMode.SHOWBOTHSIDEBYSIDECHANGESONLY
			) {
				// difference compare, show all channels
				if (filterMatchedChannels) {
					return !channel.isInPreviousAndMatchedWithCurrent();
				} else {
					return true;
				}
			} else if (dmode === CompareConstants.diffMode.DIFFVERSION) {
				return (
					channel.getDiffMode() === null ||
					channel.getDiffMode() === GraphicsConstants.ChannelDiffMode.PREVIOUS ||
					channel.getDiffMode() === GraphicsConstants.ChannelDiffMode.IN_PREVIOUS_AND_MATCHED_WITH_CURRENT
				);
			} else {
				// default mode (show current version, or unrecognised/unsupported mode)
				return (
					channel.getDiffMode() === null ||
					channel.getDiffMode() === GraphicsConstants.ChannelDiffMode.CURRENT ||
					channel.getDiffMode() === GraphicsConstants.ChannelDiffMode.IN_CURRENT_AND_MATCHED_WITH_PREVIOUS
				);
			}
		},
		this
	);
};

/**
 * Set the channel view mode
 * @param {GraphicsConstants.ChannelViewMode} channelViewMode - The new channel view mode
 */
GraphicsModel.prototype.setChannelViewMode = function(channelViewMode) {
	if (this.channelViewMode != channelViewMode) {
		var oldValue = this.channelViewMode;
		this.channelViewMode = channelViewMode;
		this._fireChange(false, ["channelViewMode"], oldValue, this.channelViewMode);
	}
};

/**
 * Get the channel view mode
 * @returns {GraphicsConstants.ChannelViewMode} The current channel view mode
 */
GraphicsModel.prototype.getChannelViewMode = function() {
	return this.channelViewMode;
};

/**
 * Returns an Object of all the selected channels, keys are the channel ID's and values are the selected channels
 * @returns {Object}
 */
GraphicsModel.prototype.getSelectedChannels = function() {
	return _.filter(this.getChannels(), function(channel) {
		return channel.isEnabled();
	});
};

/**
 * Returns the first channel that is enabled
 * @returns {Channel}
 */
GraphicsModel.prototype.getFirstSelectedChannel = function() {
	return _.find(this.getChannels(), function(channel) {
		return channel.isEnabled();
	});
};

/**
 * Checks if the list of selected channels has holes in it
 * @returns {Boolean}
 */
GraphicsModel.prototype.hasEnabledChannelGaps = function() {
	var gapStarted = false;
	var previousIsEnabled = false;
	var channels = this.getChannels();
	for (var i = 0; i < channels.length; i++) {
		var channel = channels[i];
		if (i > 0) {
			if (!gapStarted && previousIsEnabled && !channel.isEnabled()) {
				gapStarted = true;
			} else if (gapStarted && channel.isEnabled()) {
				// GAP DETECTED
				return true;
			}
		}
		previousIsEnabled = channel.isEnabled();
	}
	return false;
};

/**
 * How many channels are enabled?
 * @param {Boolean} filterMatchedChannels - should the matched channels be filtered out
 * @returns {Number}
 */
GraphicsModel.prototype.getEnabledChannelCount = function(filterMatchedChannels) {
	return _.reduce(
		this.getChannels(filterMatchedChannels),
		function(memo, channel) {
			if (channel.isEnabled()) {
				memo++;
			}
			return memo;
		},
		0
	);
};

/**
 * Get the amount of channels
 * @returns {Number}
 */
GraphicsModel.prototype.getChannelCount = function() {
	return this._channels.getLength();
};

/**
 * Move the channel with key in the list to a specified index
 * @param {Number} key - Id of the channel
 * @param {Number} index - place to move the channel to
 */
GraphicsModel.prototype.moveChannelTo = function(channelID, index) {
	// copy the old order
	var oldOrder = this._channels.getKeys().slice(0);
	this._channels.moveItemByKeyToIndex(channelID, index);
	// emit event if needed
	this._fireChange(false, ["channelOrder"], oldOrder, this._channels.getKeys());
};

/**
 * Reset the order of channels to the default order
 */
GraphicsModel.prototype.resetDefaultChannelOrder = function() {
	var oldOrder = this._channels.getKeys().slice(0);
	this._channels.setOrder(this._defaultOrder);
	this._fireChange(false, ["channelOrder"], oldOrder, this._channels.getKeys());
};

/**
 * Reset all channels to default color definition
 */
GraphicsModel.prototype.resetAllChannelsColorDef = function() {
	_.each(
		this.getChannels(),
		_.bind(function(channel) {
			this.resetChannel(channel.getID());
		}, this)
	);
};

/**
 * Checks if any of the visible channels is opaque
 * @returns {Boolean}
 */
GraphicsModel.prototype.hasVisibleOpaqueChannels = function() {
	return _.some(this.getChannels(), function(channel) {
		return channel.isEnabled() && channel.isChannelOpaque();
	});
};

/**
 * Change the channel order
 * @param {Array<Number>} orderArray - list of the ChannelId's in the correct order
 */
GraphicsModel.prototype.setChannelOrder = function(orderArray) {
	var oldOrder = this._channels.getKeys().slice(0);
	var newOrder = oldOrder.slice(0);

	// when the order is changed, this is done for a limited subset of channels, we need to fit this in the whole set of channels
	var currentChannels = this.getChannels();
	if (currentChannels.length === orderArray.length) {
		_.each(this.getChannels(), function(channel, index) {
			if (_.indexOf(orderArray, channel.getID() >= 0)) {
				// we first find the position of the channel at index [index] in the oldOrder, on this place we put the channel from orderArray[index]
				var newIndex = _.indexOf(oldOrder, channel.getID());
				newOrder[newIndex] = orderArray[index];
			} else {
				// a channel was provide which was not in the previous order...
				newOrder = null;
			}
		});
	}

	if (newOrder !== null) {
		// the new order is ok, change the order and fire event
		this._channels.setOrder(newOrder);
		this._fireChange(false, ["channelOrder"], oldOrder, this._channels.getKeys());
	}
};

/**
 * Reverse the channel order or not (only internally for url fragment, not in the view widget).
 * @param reverseChannelOrder - True to reverse the channel order, false otherwise.
 */
GraphicsModel.prototype.setReverseChannelOrder = function(reverseChannelOrder) {
	this._reverseChannelOrder = reverseChannelOrder;
};

/**
 * Get the order of the channels
 * @returns {Array<Number>} - Array of the channelIds in the correct order
 */
GraphicsModel.prototype.getChannelKeysOrder = function() {
	return this._channels.getKeys();
};

/**
 * Get a channel at a certain index
 * @param {Number} index
 * @returns {Channel}
 */
GraphicsModel.prototype.getChannelAtIndex = function(index) {
	return this._channels.getItemByIndex(index);
};

/**
 * resets the definition of a channel to the default
 * @param {String} channelId
 */
GraphicsModel.prototype.resetChannel = function(channelId) {
	var channel = this.getChannel(channelId);
	var oldChannel = channel.copy();
	channel.resetToDefault();
	this._fireChange(false, ["channel", channel.getID(), "channelRGB"], oldChannel, channel, channel.isEnabled(), channel.getID());
};

/**
 * Returns the channel after the channel with channelID
 * @param {String} channelID
 * @returns {Channel}
 */
GraphicsModel.prototype.getNextChannel = function(channelID) {
	return this._channels.getNext(channelID);
};

/**
 * Returns true if there is a next channel
 * @param {String} channelID
 * @returns {Boolean}
 */
GraphicsModel.prototype.hasNextChannel = function(channelID) {
	return this._channels.hasNext(channelID);
};

/**
 * Returns the channel before the channel with channelID
 * @param {String} channelID
 * @returns {Channel}
 */
GraphicsModel.prototype.getPreviousChannel = function(channelID) {
	return this._channels.getPrevious(channelID);
};

/**
 * Returns true if there is a previous channel
 * @param {String} channelID
 * @returns {Boolean}
 */
GraphicsModel.prototype.hasPreviousChannel = function(channelID) {
	return this._channels.hasPrevious(channelID);
};

/**
 * Changes the definition of channel (first argument) with the definition from the channel with channelId (second argument)
 * @param {String} channelIdToChange - the channel whose definition needs to be changed
 * @param {String} channelIdToCopyFrom - the channelId of the channel which definition is copied to channel
 * @param {Boolean} changeWithOriginalDef - use original definition or changed definition of the channel to copy
 */
GraphicsModel.prototype.changeChannelDef = function(channelIdToChange, channelIdToCopyFrom, changeWithOriginalDef) {
	var channel = this.getChannel(channelIdToChange);
	var channelToCopyFrom = this.getChannel(channelIdToCopyFrom);
	var oldChannel = channel.copy();
	if (changeWithOriginalDef) {
		channel.setName(channelToCopyFrom.name);
		channel.setRgb(channelToCopyFrom.rgb);
	} else {
		channel.setName(channelToCopyFrom.getNameWithoutSuffix());
		channel.setRgb(channelToCopyFrom.displayedRgb);
	}
	this._fireChange(false, ["channel", channelIdToChange, "channelRGB"], oldChannel, channel, channel.isEnabled(), channel.getID());
};

/**
 * Set the matching between two channels.
 * @param {Number} currentChannelId
 * @param {Number} referenceChannelId
 */
GraphicsModel.prototype.setChannelMatching = function(currentChannelId, referenceChannelId) {
	var currentChannel = this.getChannel(currentChannelId);
	var referenceChannel = this.getChannel(referenceChannelId);
	referenceChannel.setMatchedChannelServerId(currentChannel.getViewserverID());
	currentChannel.setMatchedChannelServerId(referenceChannel.getViewserverID());
};

/**
 * reset the matching between all channels.
 */
GraphicsModel.prototype.resetChannelMatching = function() {
	_.each(this.getChannels(), function(channel, index) {
		channel.setMatchedChannelServerId();
	});
};

/**
 * Changes the rgb color of a channel
 * @param {String} channelId - channel to change the rgb
 * @param {Number} rgbColor
 */
GraphicsModel.prototype.changeChannelRgb = function(channelId, rgbColor) {
	var channel = this.getChannel(channelId);
	var oldChannel = channel.copy();
	channel.setRgb(rgbColor);
	this._fireChange(false, ["channel", channelId, "channelRGB"], oldChannel, channel, channel.isEnabled, channel.getID());
};

/**
 * Changes the opaque of a channel
 * @param {String} channelId
 * @param {Boolean} isOpaqueChecked
 */
GraphicsModel.prototype.setChannelOpaque = function(channelId, isOpaqueChecked) {
	var channel = this.getChannel(channelId);
	var wasOpaqueChecked = channel.isChannelOpaque();
	channel.setOpaque(isOpaqueChecked);
	this._fireChange(false, ["channel", channelId, "channelRGB"], wasOpaqueChecked, isOpaqueChecked, channel.isEnabled(), channel.getID());
};

/**
 * Parsed and formats the given channels info. Used in ChannelsWidget and MeasurePopoverModel.
 * result.channels contains all the channels with as value the formatted info of the given object, result.total contains the total
 * @param {Object} channelsInfo - channels information coming from the measure tools.
 * 					- channelsinfo.makeTotalCount can be set and will result in making the total and setting it in the result, otherwise result.total is the empty string.
 * 					- channelsinfo.unit can optionally be set and will be added as unit (default '')
 * 					- channelsinfo.unitFix can optionally be set and will be be used for rounding the values (default 2).
 * @return {Object} result.channels contains all the channels in order of this model with result.channels[i].value the formatted value and result.total contains the total if needed (see above)
 */
GraphicsModel.prototype.getParsedChannelsInfo = function(channelsInfo) {
	throw "Not Implemented: GraphicsModel.getParsedChannelsInfo";
};

//------------ IMAGE FETCHING ------------------------------------------------------------------------------

/**
 * Add correct parameters for fragment URL
 * @param {Object} c - context for interceptor
 * @returns {Object}
 */
GraphicsModel.prototype.interceptFragmentURL = function(c) {
	throw "Not Implemented: GraphicsModel.interceptFragmentURL";
};

/*
 * TODO: Figure out which resolution to use.
 *
 * For now, the implementation will not go over the maximum resolution of the document (to avoid too many server calls)
 *
 * Today, the java viewer does some magic tricks to simulate a higher resolution. The HTML Viewer could also use this, but it's at a cost (extra server calls)
 * The HTML Viewer could offer an option to allow extra server calls, or should make its own implementation of "nearest neighbor" calculation
 */
/**
 * Returns the maximum resolution
 * @returns {Number}
 */
GraphicsModel.prototype.getMaximumResolution = function() {
	if (this.allowServerDeepZoom) {
		return -1;
	} else {
		return ViewerUtils.parseFloat(this.page.getParam("resolutionx"));
	}
};

//------------ OTHER ------------------------------------------------------------------------------

/**
 * Get the current compare mode
 * @returns {Number}
 */
GraphicsModel.prototype.getCompareMode = function() {
	return this.dmode;
};

/**
 * Is the graphicsmodule in difference compare mode?
 * @returns {Boolean}
 */
GraphicsModel.prototype.isInDifferenceCompareMode = function() {
	return (
		this.dmode === CompareConstants.diffMode.SHOWBOTHCHANGES ||
		this.dmode === CompareConstants.diffMode.SHOWBOTHCHANGESHIGHLIGHT ||
		this.dmode === CompareConstants.diffMode.SHOWBOTHCHANGESONLY ||
		this.dmode === CompareConstants.diffMode.SHOWBOTHSIDEBYSIDE ||
		this.dmode === CompareConstants.diffMode.SHOWBOTHONTOP ||
		this.dmode === CompareConstants.diffMode.READERSPREAD ||
		this.dmode === CompareConstants.diffMode.SHOWBOTHSIDEBYSIDEHIGHLIGHT ||
		this.dmode === CompareConstants.diffMode.SHOWBOTHSIDEBYSIDECHANGESONLY
	);
};
/**
 * Set the current compare mode
 * @param {CompareModule.CompareConstants.diffMode} newMode
 */
GraphicsModel.prototype.setCompareMode = function(newMode) {
	var oldMode = this.dmode;
	var oldChannels = this.getChannels();
	this.dmode = newMode;

	this._fireChange(false, ["compareMode"], oldMode, newMode);
	this._fireChange(false, ["channels"], oldChannels, this.getChannels());
};

/**
 * Set if the trimbox needs to be shown or not and which trimbox mode needs to be used
 * @param {Boolean} showTrimbox
 * @param {ViewControlsConstants.trimboxMode} trimboxMode
 */
GraphicsModel.prototype.setShowTrimbox = function(showTrimbox, trimboxMode) {
	var oldMode = this._trimboxMode; //oldmode
	this._trimboxMode = showTrimbox ? trimboxMode : null; //set the mode to null if no trimbox needs to be shown
	this._fireChange(false, ["trimbox"], oldMode, this._trimboxMode);
};

/**
 * Is the trimbox shown
 * @returns {Boolean}
 */
GraphicsModel.prototype.isShowTrimbox = function() {
	return this._trimboxMode !== null;
};

export default GraphicsModel;
