/**
 *  @namespace Graphics
 */
/**
 *  @namespace Graphics
 */
import AbstractModule from "@esko/web-viewer/core/AbstractModule";

import GraphicsDataExtensions from "@esko/web-viewer/graphics/GraphicsDataExtensions";
import ChannelsWidget from "@esko/web-viewer/graphics/ChannelsWidget";
import GraphicsLayerModel from "@esko/web-viewer/graphics/GraphicsLayerModel";
import GraphicsViewportWidgetController from "@esko/web-viewer/graphics/GraphicsViewportWidgetController";
import ChannelMatchingModel from "@esko/web-viewer/graphics/ChannelMatchingModel";
import ChannelMatchingPopup from "@esko/web-viewer/graphics/ChannelMatchingPopup";
import ChannelMatchingController from "@esko/web-viewer/graphics/ChannelMatchingController";
import ModalPopup from "@esko/web-viewer/layout/component/ModalPopup";
import GraphicsConstants from "@esko/web-viewer/graphics/GraphicsConstants";
import _ from "underscore";
import $ from "jquery";
/**
 * @exports AbstractGraphicsModule
 * @memberof AbstractGraphicsModule
 * @class the graphics module adds the graphics layer if applicable
 * @constructor
 * @augments Core.AbstractModule
 *
 * @param {Core.ViewerFacade} f - Viewer facade
 * @param {Boolean} silenced - if true, the module shouldn't publish tools/widgets/toolbarcomponents
 */
var AbstractGraphicsModule = function(f, silenced) {
	AbstractModule.call(this, f, silenced, "graphics");

	// Inject graphics functionality in data objects
	GraphicsDataExtensions(f);

	this.graphicsViewportWidgetController = null;
	this.graphicsModel = null;
	this.graphicsLayerModel = null;
	this.channelsWidget = null;

	//channelmatching
	this.channelMatchingPopup = null;
	this.channelMatchingModel = null;
	this.channelMatchingController = null;
	this.isChannelMatchingActive = false;
	this.nonMatchedChannelsPopup = null;

	this.isGraphicsActive = false;

	this._waitingForParameterUpdate = false;

	this.f.on("intercept.fragmentURL", _.bind(this._interceptFragmentURL, this));

	this.f.on(
		"module.compare.changed.diffMode",
		_.bind(function(oldMode, newMode) {
			if (this.graphicsModel) {
				this.graphicsModel.setCompareMode(newMode);
			}
		}, this)
	);

	_.defer(
		_.bind(function() {
			// listen to page changes or parameter changes
			// Note that we want to defer this function because the constructor of the implementing class
			// is still ongoing
			this.f.getDocumentController().whenParametersAvailable(function() {
				this._whenParametersAvailable();
				this.f.getDocumentController().whenActiveOrDifferencePageChanged(this._whenActiveOrDifferencePageChanged, this);
			}, this);
		}, this)
	);
};

// inherit
AbstractGraphicsModule.prototype = Object.create(AbstractModule.prototype);
AbstractGraphicsModule.prototype.constructor = AbstractGraphicsModule;

/**
 * When the active or difference page has changed, wait for the parameters to become available
 * @private
 */
AbstractGraphicsModule.prototype._whenActiveOrDifferencePageChanged = function() {
	var docController = this.f.getDocumentController();
	if (!this._waitingForParameterUpdate) {
		if (docController.isParameterUpdateOngoing()) {
			this.f.getDocumentController().once("parameters.afterEndUpdate", _.bind(this._whenParametersAvailable, this));
			this._waitingForParameterUpdate = true;
		} else {
			// parameters are available already
			this._whenParametersAvailable();
		}
	}
};

/**
 * Execute when active document page has changed or new document page is loaded. Initializes all components of the Graphics module
 * @private
 */
AbstractGraphicsModule.prototype._whenParametersAvailable = function() {
	this._waitingForParameterUpdate = false;
	var channels;
	var activePage = this.f.getActivePage();
	var differencePage = this.f.getDocumentController().getDifferencePage();

	if (!activePage) {
		return;
	}

	//the active document page has changed, create the graphics model and update the widget
	if (activePage.hasGraphics() || (differencePage && differencePage.hasGraphics()) || activePage.isBlankPage()) {
		//create models if not existing
		if (this.graphicsModel == null) {
			this.graphicsModel = this._createGraphicsModel(activePage, differencePage);
			this.graphicsLayerModel = new GraphicsLayerModel(this.f, "Graphics", this.graphicsModel);
		} else {
			// if the active page has changed, remove all views
			if ((this.graphicsModel.getPage() != activePage || !this.isGraphicsActive) && this.graphicsViewportWidgetController) {
				this.graphicsViewportWidgetController.destructViews();
			}
			// update the graphics model
			this.graphicsModel.setPage(activePage, differencePage);
			if (this.isChannelMatchingActive) {
				channels = this.graphicsModel.getAllChannels();
				this.channelMatchingController.setChannels(channels);
				if (activePage !== undefined && differencePage !== undefined && this.nonMatchedChannelsPopup === null) {
					//only try opening the popup when the popup is null or is was set to null
					this.showChannelMatchingPopupIfNeeded();
				}
			}
		}

		var compareModule = this.f.getModule("compare");
		if (compareModule) {
			this.graphicsModel.setCompareMode(compareModule.getDiffMode());
		}

		if (!this.isGraphicsActive) {
			//add the layer if there weren't graphics active before
			this.f.getModule("viewport").addLayer("Graphics", this.graphicsLayerModel, "Main");
			this.isGraphicsActive = true;
		}

		if (this.graphicsViewportWidgetController == null) {
			this.graphicsViewportWidgetController = new GraphicsViewportWidgetController(this.f, this.graphicsModel);
			this.graphicsViewportWidgetController.updateTiles();
		}

		// it is an image file, it makes sense to display the channels widget
		if (!this.channelsWidget) {
			// the channels widget was never intialized, create it
			this._publishWidget(
				ChannelsWidget.widgetName,
				_.bind(function(layoutType) {
					this.channelsWidget = new ChannelsWidget(this.f, this.graphicsModel);
					return this.channelsWidget;
				}, this)
			);
		} else {
			//widget already existed, make sure it is visible
			this.channelsWidget.setEnabled(true);
		}
	} else if (this.graphicsModel) {
		this.f.getModule("viewport").removeLayer("Graphics", "Main");
		//remove all the views when there is no graphics on the page
		if (this.graphicsModel.getPage() != activePage && this.graphicsViewportWidgetController) {
			this.graphicsViewportWidgetController.destructViews();
		}
		// update the graphics model, no graphics files are active so both pages are undefined
		this.graphicsModel.setPage(undefined, undefined);
		this.isGraphicsActive = false;
		// not an image file, hide the channels widget if already available
		if (this.channelsWidget) {
			this.channelsWidget.setEnabled(false);
		}
		this.graphicsModel.setChannelViewMode(GraphicsConstants.ChannelViewMode.COMPOSITE_VIEW);
	}

	// channel matching
	if (!activePage.isReaderSpread()) {
		if (activePage !== undefined && differencePage !== undefined && activePage.hasGraphics() && differencePage.hasGraphics()) {
			// we are comparing two files... initialize channel matching correctly
			this.isChannelMatchingActive = true;
			channels = this.graphicsModel.getAllChannels();

			if (this.channelMatchingModel === null) {
				this.channelMatchingModel = new ChannelMatchingModel(this.f, channels);
			}
			if (this.channelMatchingPopup === null) {
				this.channelMatchingPopup = new ChannelMatchingPopup(this.f, "channelMatchingPopup", this.channelMatchingModel);
			}
			if (this.channelMatchingController === null) {
				this.channelMatchingController = new ChannelMatchingController(
					this.f,
					this.channelMatchingPopup,
					this.channelMatchingModel,
					this.graphicsModel
				);
			}
			this.channelMatchingController.setChannels(channels);

			this.showChannelMatchingPopupIfNeeded();

			if (this.channelsWidget) {
				this.channelsWidget.showChannelMatchingBtn(_.bind(this.showChannelMatchingPopup, this));
			}
		} else if (this.isChannelMatchingActive) {
			// channelmatching was still active.. clean it up
			this.isChannelMatchingActive = false;
			if (this.nonMatchedChannelsPopup) {
				// if the nonMatchedChannelsPopup exist, we want to clean it up completely
				this.nonMatchedChannelsPopup.remove();
			}
			this.nonMatchedChannelsPopup = null;
			if (this.channelsWidget) {
				//hide the channel matching button when not both current and reference document pages are selected
				this.channelsWidget.hideChannelMatchingBtn();
			}
		}
	} else {
		if (this.channelsWidget) {
			// in reader spread, hide the channel matching btn
			this.channelsWidget.hideChannelMatchingBtn();
		}
	}
};

/**
 * Show the channel matching popup if it is needed (there are unmatched channels)
 */
AbstractGraphicsModule.prototype.showChannelMatchingPopupIfNeeded = function() {
	//when not all the channels are matched, show popup asking to do channel matching if not on mobile
	if (this.channelMatchingModel !== undefined && this.channelMatchingModel.hasUnmatchedChannels() && !this.f.isMobile()) {
		if (this.nonMatchedChannelsPopup === null) {
			this.nonMatchedChannelsPopup = new ModalPopup(this.f, "nonMatchingChannelsPopup", this.f.getContext());
		}
		this.f.i18n.loadBundle(
			"viewer",
			_.bind(function() {
				this.nonMatchedChannelsPopup.setContent(
					$("<div>")
						.addClass("nonMatchingChannelsDialog")
						.html(this.f.i18n.t("DocumentListWarning"))
				);
				this.nonMatchedChannelsPopup.setOkBtnText(this.f.i18n.t("startChannelMatching"));
				this.nonMatchedChannelsPopup.setHeader(
					$("<span/>")
						.addClass("popupTitle")
						.html(this.f.i18n.t("MapChanTitle"))
				);
			}, this)
		);

		this.nonMatchedChannelsPopup.setOkAction(
			_.bind(function() {
				this.channelMatchingPopup.show();
				this.nonMatchedChannelsPopup.hide();
			}, this)
		);

		this.nonMatchedChannelsPopup.show();
		this.nonMatchedChannelsPopup.recalculateSize(820);
	}
};

/**
 * Get the graphics module for the currently active page
 * @returns {Graphics.GraphicsModel}
 *
 */
AbstractGraphicsModule.prototype.getGraphicsModel = function() {
	return this.graphicsModel;
};

/**
 * Get the selection for a certain diffMode
 * @param {CompareConstants.difMode} diffMode - the mode for which the selection should be calculated
 * @returns {Number}
 */
AbstractGraphicsModule.prototype.getChannelSelectionForDiffMode = function(diffMode) {
	if (this.graphicsModel !== null) {
		return this.graphicsModel.getChannelSelectionForDiffMode(diffMode);
	}
};

/**
 * Shows the popup for channel matching
 */
AbstractGraphicsModule.prototype.showChannelMatchingPopup = function() {
	var channels = this.graphicsModel.getAllChannels();
	this.channelMatchingController.setChannels(channels);
	this.channelMatchingPopup.show();
};

/**
 * @override
 */
AbstractGraphicsModule.prototype.getState = function() {
	if (this.graphicsModel !== null) {
		var channelsState = {};
		_.each(
			this.graphicsModel.getAllChannels(),
			function(channel) {
				var channelState = {
					enabled: channel.isEnabled(),
					displayedRgb: channel.getRGB(),
					isOpaque: channel.isChannelOpaque()
				};
				channelsState[channel.getID()] = channelState;
			},
			this
		);

		var state = {
			// Graphics model
			graphicsModel: {
				channelViewMode: this.graphicsModel.getChannelViewMode(),
				channelKeysOrder: this.graphicsModel.getChannelKeysOrder(),
				singleChannelMode: this.graphicsModel.getSingleChannelMode(),
				channels: channelsState
			}
		};

		if (this.channelMatchingModel !== null) {
			// Channel Matching model
			state.channelMatchingModel = {
				channelMatching: this.channelMatchingModel.getChannelMatching()
			};
		}
		return state;
	}
};

/**
 * @override
 * @param state
 */
AbstractGraphicsModule.prototype.setState = function(state) {
	this.f.whenModuleAvailable(
		"graphics",
		function() {
			if (this.graphicsModel === null) {
				this.graphicsModel = this._createGraphicsModel(
					this.f.getDocumentController().getActivePage(),
					this.f.getDocumentController().getDifferencePage()
				);
				this.graphicsLayerModel = new GraphicsLayerModel(this.f, "Graphics", this.graphicsModel);
			}
			// This module is ready
			if (state !== undefined) {
				// Graphics model
				var graphicsModelState = state.graphicsModel;
				if (graphicsModelState !== undefined) {
					this.graphicsModel.setChannelViewMode(graphicsModelState.channelViewMode);
					this.graphicsModel.setChannelOrder(graphicsModelState.channelKeysOrder);
					this.graphicsModel.setSingleChannelMode(graphicsModelState.singleChannelMode);

					// Channels
					var channelsState = graphicsModelState.channels;
					_.each(
						this.graphicsModel.getAllChannels(),
						function(channel) {
							var channelID = channel.getID();
							var channelState = channelsState[channelID];
							if (channelState !== undefined) {
								this.graphicsModel.toggleChannel(channelID, channelState.enabled);
								this.graphicsModel.setChannelOpaque(channelID, channelState.isOpaque);
								this.graphicsModel.changeChannelRgb(channelID, channelState.displayedRgb);
							}
						},
						this
					);
				}
				// Channel matching model
				var channelMatchingModelState = state.channelMatchingModel;
				if (channelMatchingModelState !== null && channelMatchingModelState !== undefined) {
					if (this.channelMatchingModel === null) {
						this.channelMatchingModel = new ChannelMatchingModel(this.f, undefined);
					}
					this.channelMatchingModel.setChannelMatching(channelMatchingModelState.channelMatching);
				}
			}
		},
		this
	);
};

/**
 * Get the enabled status of all channels
 * @returns {Object} with key-value pairs (key=channelID, value=enabled)
 */
AbstractGraphicsModule.prototype.getChannelSelection = function() {
	var channelSelection = {};
	_.each(this.graphicsModel.getAllChannels(), function(channel) {
		channelSelection[channel.getID()] = channel.isEnabled();
	});

	return channelSelection;
};

/**
 * Set the channel selection (restore a state retrieved from getChannelSelection())
 * @param {Object} channelSelection -  with key-value pairs with the enabled state of the channels (key=channelID, value=enabled)
 */
AbstractGraphicsModule.prototype.setChannelSelection = function(channelSelection) {
	_.each(
		channelSelection,
		function(enabled, ID) {
			this.graphicsModel.toggleChannel(ID, enabled);
		},
		this
	);
};

/**
 * Set the channel view mode
 * @param {GraphicsConstants.ChannelViewMode} channelViewMode - The new channel view mode
 */
AbstractGraphicsModule.prototype.setChannelViewMode = function(mode) {
	this.graphicsModel.setChannelViewMode(mode);
};

/**
 * Get the channel view mode
 * @returns {GraphicsConstants.ChannelViewMode} The current channel view mode
 */
AbstractGraphicsModule.prototype.getChannelViewMode = function() {
	return this.graphicsModel.getChannelViewMode();
};

/**
 * Set if the trimbox needs to be shown and which mode is used
 * @param {Boolean} showTrimbox
 * @param {ViewControlsConstants.trimboxMode} trimboxMode
 */
AbstractGraphicsModule.prototype.setShowTrimbox = function(showTrimbox, trimboxMode) {
	if (this.graphicsModel) {
		this.graphicsModel.setShowTrimbox(showTrimbox, trimboxMode);
	}
};

AbstractGraphicsModule.prototype._interceptFragmentURL = function(c) {
	throw "Not Implemented";
};

/**
 * Set the correct trimbox information in the context
 * @param {Object} c
 */
AbstractGraphicsModule.prototype._setTrimboxInformation = function(c) {
	throw "Not Implemented";
};

/**
 * Create a Graphics model of the specific implementation
 * @returns {AbstractGraphicsModel}
 */
AbstractGraphicsModule.prototype._createGraphicsModel = function(activePage, differencePage) {
	throw "Not Implemented";
};

export default AbstractGraphicsModule;
