import ViewerUtils from "@esko/web-viewer/utils/ViewerUtils";
import GraphicsConstants from "@esko/web-viewer/graphics/GraphicsConstants";
import _ from "underscore";

/**
 * @exports Channel
 * @memberof Graphics
 * @class Representation of a channel
 * The channel is enabled by default.
 * @constructor
 *
 * @param {String} id - Id of the channel. This is the channel number.
 * @param {Number} viewserverID - the number on the viewserver
 * @param {String} name - Name of the channel
 * @param {Number} rgb - Integer representation of the rgb color
 * @param {Boolean} isOpaque - is the channel opaque
 * @param {String} file - File of the channel
 * @param {GraphicsConstants.ChannelDiffMode} dmode
 * @param {Data.AbstractDocumentPage} page - the page to which the channel belongs
 * @param {Number} [matchedTo] - the channel-file too which this channel is matched
 */
var Channel = function(id, viewserverID, name, rgb, isOpaque, file, dmode, page, matchedWith) {
	this.id = id;
	this.viewserverID = viewserverID;
	this.name = name;
	this.displayedName = name;
	this.rgb = rgb;
	this.displayedRgb = rgb;
	this.wasOpaque = isOpaque;
	this.isOpaque = isOpaque;
	this.file = file;
	// set dmode (if undefined, make it null)
	this.dmode = dmode === undefined ? null : dmode;
	this.page = page;
	// this is the file-identifier for the channel to which this channel is matched
	this.matchedWith = matchedWith;
	if (this.matchedWith === undefined) {
		this.matchedWith = null;
	}
	// There is also valid but it should always be 3: this means Valid (bit 0) and RGB (bit 1)
	// we can ignore it because the view server always assures that rgb contains rgb values
	this.enabled = true;
};

/**
 * Copy this channel
 * @returns {Channel} New object with the same id, name, rgb, file, etc.
 */
Channel.prototype.copy = function() {
	var copiedChannel = new Channel(
		this.id,
		this.viewserverID,
		this.name,
		this.rgb,
		this.wasOpaque,
		this.file,
		this.dmode,
		this.page,
		this.matchedWith
	);
	copiedChannel.displayedRgb = this.displayedRgb;
	copiedChannel.isOpaque = this.isOpaque;
	copiedChannel.displayedName = this.displayedName;
	return copiedChannel;
};

/**
 * Toggle whether this channel is enabled or not
 * @param {Boolean} [toggle] - If defined, the channel will be enabled/disabled according to this parameter,
 * otherwise the value of 'enabled' is switched
 */
Channel.prototype.toggle = function(toggle) {
	if (toggle === undefined) {
		this.enabled = !this.enabled;
	} else {
		this.enabled = toggle;
	}
};

/**
 * Is this channel enabled?
 * @returns {Boolean} True if the channel is enabled, false otherwise.
 */
Channel.prototype.isEnabled = function() {
	return this.enabled;
};

/**
 * Get the id of the channel as used by the viewserver
 * @returns {Number}
 */
Channel.prototype.getViewserverID = function() {
	return this.viewserverID;
};

/**
 * Get id of the channel
 * @returns {String} Id of the channel
 */
Channel.prototype.getID = function() {
	return this.id;
};

/**
 * Get the name of the channel, if the name is changed a "*" character will be added
 * @returns {String} Name of the channel
 */
Channel.prototype.getName = function() {
	if (this.isChanged()) {
		return this.displayedName + " *";
	} else {
		return this.displayedName;
	}
};

/**
 * Get the original channel name
 * @returns {String}
 */
Channel.prototype.getOriginalName = function() {
	return this.name;
};

/**
 * Returns the displayed name of the channel without the optional *-char
 * @returns {String}
 */
Channel.prototype.getNameWithoutSuffix = function() {
	return this.displayedName;
};

/**
 * Get the RGB color of the channel in int format
 * @returns {Number} RGB value of the channel in int format
 */
Channel.prototype.getRGB = function() {
	return this.displayedRgb;
};

/**
 * Get the RGB color of the channel in hex format
 * @returns {String} RGB value of the channel in hex format
 */
Channel.prototype.getHexRGB = function() {
	return ViewerUtils.hexFromInt(this.displayedRgb);
};

/**
 * Get the file of the channel
 * @returns {String} File of the channel
 */
Channel.prototype.getFile = function() {
	return this.file;
};

/**
 * Get the diff mode for the channel, if no dmode is defined, it returns undefined
 * @returns {GraphicsConstants.ChannelDiffMode} diffMode or undefined
 */
Channel.prototype.getDiffMode = function() {
	return this.dmode;
};

/**
 * Resets the displayed name and rgb to the previous values
 */
Channel.prototype.resetToDefault = function() {
	this.displayedName = this.name;
	this.displayedRgb = this.rgb;
	this.isOpaque = this.wasOpaque;
};

/**
 * is the channel opaque
 * @returns {Boolean}
 */
Channel.prototype.isChannelOpaque = function() {
	return this.isOpaque;
};

/**
 * was the channel opaque by default
 * @returns {Boolean}
 */
Channel.prototype.wasChannelOpaque = function() {
	return this.wasOpaque;
};

/**
 * Returns true is the channel is changed
 * @returns {Boolean}
 */
Channel.prototype.isChanged = function() {
	return this.rgb != this.displayedRgb || this.isOpaque != this.wasOpaque || this.displayedName != this.name;
};

/**
 * Change the Rgb of the channel
 * @param {Number} Rgb
 */
Channel.prototype.setRgb = function(Rgb) {
	if (Rgb !== this.displayedRgb) {
		this.displayedRgb = Rgb;
	}
};

/**
 * Change the opaque of the channel
 * @param {Boolean} opaque
 */
Channel.prototype.setOpaque = function(opaque) {
	if (opaque !== this.isOpaque) {
		this.isOpaque = opaque;
	}
};

/**
 * sets the displayed name of the channel
 * @param {String} name
 */
Channel.prototype.setName = function(name) {
	this.displayedName = name;
};

/**
 * Checks if both channels can be considered equal, does not take default values into account
 * @param {Channel} channel - Channel to compare with
 * @returns {Boolean}
 */
Channel.prototype.equals = function(channel) {
	if (
		channel.getRGB() === this.displayedRgb &&
		channel.isChannelOpaque() === this.isOpaque &&
		channel.displayedName === this.displayedName
	) {
		return true;
	} else {
		return false;
	}
};

/**
 * Checks if both channel can be considered equal, only takes the default values into account
 * @param {Channel} channel
 * @returns {Boolean}
 */
Channel.prototype.equalsOriginalDef = function(channel) {
	if (channel.rgb === this.rgb && channel.wasChannelOpaque() === this.isOpaque) {
		return true;
	} else {
		return false;
	}
};

/**
 * Is the channel matched with another channel of another file
 * @returns {Boolean}
 */
Channel.prototype.isMatched = function() {
	return this.matchedWith !== null;
};

/**
 * Get the channel-file of the channel which is matched with this channel
 * @returns {Number} - viewerserverId
 */
Channel.prototype.getMatchedChannelServerId = function() {
	return this.matchedWith;
};

/**
 * Set the channel-file of the channel which is matched with this channel
 * @param {Channel} matchedChannel - if undefined the channel will no longer be matched
 */
Channel.prototype.setMatchedChannelServerId = function(matchedChannelServerId) {
	if (matchedChannelServerId === undefined) {
		this.matchedWith = null;
	} else {
		this.matchedWith = matchedChannelServerId;
	}
};

/**
 * Get the page to which this page belongs
 * @returns {Data.AbstractDocumentPage}
 */
Channel.prototype.getPage = function() {
	return this.page;
};

/**
 * Is the channel in the current file and matched with the difference file
 * @returns {Boolean}
 */
Channel.prototype.isInCurrentAndMatchedToPrevious = function() {
	return (
		this.dmode === GraphicsConstants.ChannelDiffMode.IN_CURRENT_AND_MATCHED_WITH_PREVIOUS ||
		(this.isMatched() && this.dmode === GraphicsConstants.ChannelDiffMode.CURRENT)
	);
};

/**
 * Is the channel in the difference file and matched with the current?
 * @returns {Boolean}
 */
Channel.prototype.isInPreviousAndMatchedWithCurrent = function() {
	return (
		this.dmode === GraphicsConstants.ChannelDiffMode.IN_PREVIOUS_AND_MATCHED_WITH_CURRENT ||
		(this.isMatched() && this.dmode === GraphicsConstants.ChannelDiffMode.PREVIOUS)
	);
};

/**
 * Is this channel subtracted in comparison to the document too which the viewer is comparing?
 * @returns {Boolean}
 */
Channel.prototype.isSubtracted = function() {
	return this.dmode === GraphicsConstants.ChannelDiffMode.PREVIOUS && !this.isMatched();
};

/**
 * Is this channel added in comparison to the document too which the viewer is comparing?
 * @returns {Boolean}
 */
Channel.prototype.isAdded = function() {
	return this.dmode === GraphicsConstants.ChannelDiffMode.CURRENT && !this.isMatched();
};

export default Channel;
