import ListenerUtil from "@esko/web-viewer/utils/ListenerUtil";
import ChannelMatchingModel from "@esko/web-viewer/graphics/ChannelMatchingModel";
import ChannelMatchingPopup from "@esko/web-viewer/graphics/ChannelMatchingPopup";
import GraphicsConstants from "@esko/web-viewer/graphics/GraphicsConstants";
import $ from "jquery";
import _ from "underscore";
/**
 *
 * @exports ChannelMatchingController
 * @memberof GraphicsModule
 * @class Controller for the channel matching popup. Listens to events of the channelMatchingPopup and saves changed data to the channelMatchingModel.
 * Implements the matching algorithm, channels can be matched by name and RGB. There are three levels of matching: same name and RGB definition, same name or
 * smart name matching. When the matching is complete, the controller will persist the changes to the graphicsmodel.
 * @constructor
 *
 * @param {Core.ViewerFacade} f - Viewer facade
 * @param {ChannelMatchingPopup} channelMatchingPopup - the view for the channel matching
 * @param {ChannelMatchingModel} channelMatchingModel - the model for the channel matching
 * @param {GraphicsModel} graphicsModel - used for persisting changes made in the channel matching model
 */
var ChannelMatchingController = function(f, channelMatchingPopup, channelMatchingModel, graphicsModel) {
	ListenerUtil.mixin(this, f);

	this.f = f;

	/**
	 * @member {ChannelMatchingPopup}
	 */
	this.channelMatchingPopup = channelMatchingPopup;

	/**
	 * @member {ChannelMatchingModel}
	 */
	this.channelMatchingModel = channelMatchingModel;

	/**
	 * @member {GraphicsModel}
	 */
	this._graphicsModel = graphicsModel;

	//Do initial matching (default or in user preferences saved matching)
	this._matchChannels(this.channelMatchingModel.getChannelMatching());

	//Register all the listeners to the channelMatchingPopup
	this._registerListener("currentChannelMatch", this._currentChannelMatchingChanged, this, this.channelMatchingPopup);
	this._registerListener("referenceChannelMatch", this._referenceChannelMatchingChanged, this, this.channelMatchingPopup);
	this._registerListener("channelMatching", this._matchChannels, this, this.channelMatchingPopup);
	this._registerListener(
		"channelMatchingComplete",
		_.bind(function() {
			this._completeChannelMatching(true);
		}, this),
		this,
		this.channelMatchingPopup
	);
	this._registerListener("matchedChannelClicked", this._activateMatchedChannel, this, this.channelMatchingPopup);
	this._registerListener("persistDefaultChannelMatching", this._persistChannelMatching, this, this.channelMatchingPopup);
};

/**
 * @private
 * Called when the list of matched channels of the current document page is changed. Will reset the definition of the moved channel if it moves from
 * the matched channels list to the unmatched channels list.
 *
 * @param {String} movedChannelId - Id of the moved channel
 * @param {Array.<String>} currentMatchedChannels - id list of the matched channels, id are strings
 * @param {Boolean} toMatchedList - is the channel moved to the list of matched channels
 */
ChannelMatchingController.prototype._currentChannelMatchingChanged = function(movedChannelId, currentMatchedChannels, toMatchedList) {
	this.channelMatchingModel.setMatchedCurrentChannels(currentMatchedChannels);
	if (!toMatchedList) {
		this.channelMatchingModel.getCurrentChannel(parseInt(movedChannelId, 10)).resetToDefault();
	}
	this.channelMatchingPopup.render();
};

/**
 * @private
 * Called when the list of matched channels of the reference document page is changed. Will reset the definition of the moved channel if it moves from
 * the matched channels list to the unmatched channels list.
 *
 * @param {String} movedChannelId - Id of the moved channel
 * @param {Array<String>} referenceMatchedChannels - id list of the matched channels, id are strings
 * @param {Boolean} toMatchedList - is the channel moved to the list of matched channels
 */
ChannelMatchingController.prototype._referenceChannelMatchingChanged = function(movedChannelId, referenceMatchedChannels, toMatchedList) {
	this.channelMatchingModel.setMatchedReferenceChannels(referenceMatchedChannels);
	if (!toMatchedList) {
		this.channelMatchingModel.getReferenceChannel(parseInt(movedChannelId, 10)).resetToDefault();
	}
	this.channelMatchingPopup.render();
};

/**
 * @private
 * Saves the channel matching to the user preferences. Reset these preferences to the default if no parameter or parameters is false.
 *
 * @param {Boolean} persistChannelMatching - does the channel matching need to be persisted
 */
ChannelMatchingController.prototype._persistChannelMatching = function(persistChannelMatching) {
	if (persistChannelMatching) {
		this.f.getCurrentUser().setOption("PREFERENCES_COLORMATCHING", this.channelMatchingModel.getChannelMatching());
	} else {
		this.f.resetOption("PREFERENCES_COLORMATCHING");
	}
	this.channelMatchingModel.setIsChannelMatchingPersisted(persistChannelMatching);
};

/**
 * Set the channels for the channelModel, it is possible to do initial matching for the new channels for the channelMatchingModel
 * @param {Array.<Channel>} channels
 */
ChannelMatchingController.prototype.setChannels = function(channels) {
	this.channelMatchingModel.setChannels(channels);
	this._matchChannels(this.channelMatchingModel.getChannelMatching());
	this._completeChannelMatching(false);
};

/**
 * @private
 * Matches the channels, the parameter determines which matching is used.
 *
 * @param {String} channelMatchKey
 */
ChannelMatchingController.prototype._matchChannels = function(channelMatchKey) {
	this.channelMatchingModel.resetChannelMatching(); //reset all the channel matching before trying to match again
	channelMatchKey = parseInt(channelMatchKey, 10); //convert the string to GraphicsConstants.ChannelMatching
	this.channelMatchingModel.setChannelMatching(channelMatchKey);
	var currentChannels = this.channelMatchingModel.getCurrentChannels();
	var referenceChannels = this.channelMatchingModel.getReferenceChannels();

	_.each(
		currentChannels,
		function(currentChannel) {
			//all the current channels
			var match = undefined;
			if (channelMatchKey === GraphicsConstants.ChannelMatching.SMART_NAME_MATCHING) {
				// we are matching with smart names, first try without smart names (this is to prevent following situation)
				/**
				 * @example
				 *
				 * File 1:  C, M, Y, K, PROCESS BLACK
				 * File 2:  C, M, Y, PROCESS BLACK, K
				 *
				 * if we just search using smart name matching, process black will be mapped on K and the other way arround... but we want to map k on k and process black on process black
				 */
				match = _.find(
					referenceChannels,
					function(referenceChannel) {
						return this._canChannelsBeMatched(currentChannel, referenceChannel, GraphicsConstants.ChannelMatching.NAME_MATCHING);
					},
					this
				);
			}

			if (match === undefined) {
				// either we are not in SMART_NAME_MATCHING, or nothing was found using the NAME_MATCHING (in this case we try again with SMAR_NAME_MATCHING)
				match = _.find(
					referenceChannels,
					function(referenceChannel) {
						return this._canChannelsBeMatched(currentChannel, referenceChannel, channelMatchKey);
					},
					this
				);
			}

			if (match !== undefined) {
				// there was a match
				var isStayMatched = this._canChannelsBeMatched(currentChannel, match, channelMatchKey, true);
				this.channelMatchingModel.setChannelMatch(currentChannel.getViewserverID(), match.getViewserverID(), true, isStayMatched);
			}
		},
		this
	);

	this.channelMatchingModel.orderMatchedChannels(); //to make sure the stay matched channels are on top
	this.channelMatchingPopup.render();
	this._persistChannelMatching(this.channelMatchingModel.isChannelMatchingPersisted());
};

/**
 * @private
 * Checks if the two channels can be matched to each other depending on the channelMatchKey
 *
 * @param {Channel} currentChannel - channel of the current document page
 * @param {Channel} referenceChannel - channel of the reference document page
 * @param {GraphicsConstants.ChannelMatching} channelMatchKey - key for the channel matching
 * @param {Boolean} useOriginalDef - determines whether the original definition of the channels is used or the changed definition
 *
 * @returns {Boolean}
 */
ChannelMatchingController.prototype._canChannelsBeMatched = function(currentChannel, referenceChannel, channelMatchKey, useOriginalDef) {
	var currentChannelName = currentChannel.getNameWithoutSuffix();
	var referenceChannelName = referenceChannel.getNameWithoutSuffix();
	if (useOriginalDef) {
		currentChannelName = currentChannel.getOriginalName();
		referenceChannelName = referenceChannel.getOriginalName();
	}
	if (channelMatchKey === GraphicsConstants.ChannelMatching.SMART_NAME_MATCHING) {
		//smart name matching
		//first remove the process, pantone key words from the name
		currentChannelName = currentChannelName.replace(/process\s*/i, "").replace(/pantone\s*/i, "");
		referenceChannelName = referenceChannelName.replace(/process\s*/i, "").replace(/pantone\s*/i, "");
		//see if the strings match:
		if (currentChannelName.toLowerCase() === referenceChannelName.toLowerCase()) {
			return true;
		}
		//match if names are similar: cyan and C, magenta and M, yellow and Y, black and K
		currentChannelName = currentChannelName
			.replace(/cyan/i, "c")
			.replace(/magenta/i, "m")
			.replace(/yellow/i, "y")
			.replace(/black/i, "k");
		referenceChannelName = referenceChannelName
			.replace(/cyan/i, "c")
			.replace(/magenta/i, "m")
			.replace(/yellow/i, "y")
			.replace(/black/i, "k");
		if (currentChannelName.toLowerCase() === referenceChannelName.toLowerCase()) {
			return true;
		} else {
			return false;
		}
	} else if (channelMatchKey === GraphicsConstants.ChannelMatching.NAME_MATCHING) {
		//match channels (ignore RGB differences)
		if (currentChannelName.toLowerCase() === referenceChannelName.toLowerCase()) {
			return true;
		} else {
			return false;
		}
	} else if (channelMatchKey === GraphicsConstants.ChannelMatching.IDENTICAL_MATCHING) {
		//only match if identical (no RGB differences)
		if (
			currentChannelName.toLowerCase() === referenceChannelName.toLowerCase() &&
			useOriginalDef &&
			currentChannel.equalsOriginalDef(referenceChannel)
		) {
			return true;
		} else if (
			currentChannelName.toLowerCase() === referenceChannelName.toLowerCase() &&
			!useOriginalDef &&
			currentChannel.equals(referenceChannel)
		) {
			return true;
		} else {
			return false;
		}
	}
};

/**
 * @private
 * Persist the channel matching to the graphicsmodel. Changes the definition of matched channels.
 *
 * @param {Boolean} resetChannelDefinition
 *
 */
ChannelMatchingController.prototype._completeChannelMatching = function(resetChannelDefinition) {
	this._graphicsModel.resetChannelMatching();
	var currentMatchedChannels = this.channelMatchingModel.getMatchedCurrentChannels();
	var referenceMatchedChannels = this.channelMatchingModel.getMatchedReferenceChannels();
	var copyCurrentMatchedChannels = this.channelMatchingModel.getIsCurrentChannelActive();

	_.each(
		currentMatchedChannels,
		_.bind(function(currentMatchedChannel, index) {
			//run over all matched channels
			var referenceMatchedChannel = referenceMatchedChannels[index];
			var copyCurrentMatchedChannel = copyCurrentMatchedChannels[index];
			if (referenceMatchedChannel !== undefined) {
				//check if current channel is matched with a reference channel, if not the case dont match the channel
				this._graphicsModel.setChannelMatching(currentMatchedChannel.getID(), referenceMatchedChannel.getID());
				if (copyCurrentMatchedChannel) {
					//which of the two channels is active and copies its definition to the other channel
					if (!currentMatchedChannel.isChanged()) {
						//if the channel is not changed (definition was reset) make sure to use the original definition of the channel
						this._graphicsModel.getChannel(currentMatchedChannel.getID()).resetToDefault();
					}
					this._graphicsModel.changeChannelDef(referenceMatchedChannel.getID(), currentMatchedChannel.getID(), false);
				} else {
					if (!referenceMatchedChannel.isChanged()) {
						//if the channel is not changed (definition was reset) make sure to use the original definition of the channel
						this._graphicsModel.getChannel(referenceMatchedChannel.getID()).resetToDefault();
					}
					this._graphicsModel.changeChannelDef(currentMatchedChannel.getID(), referenceMatchedChannel.getID(), false);
				}
			} else {
				this._graphicsModel.changeChannelDef(currentMatchedChannel.getID(), currentMatchedChannel.getID(), true);
			}
		}, this)
	);

	if (resetChannelDefinition) {
		//make sure reseted channels are saved to the graphicsModel
		var currentChannels = this.channelMatchingModel.getCurrentChannels();
		var referenceChannels = this.channelMatchingModel.getReferenceChannels();
		_.each(
			currentChannels,
			_.bind(function(channel) {
				this._graphicsModel.resetChannel(channel.getID());
			}, this)
		);
		_.each(
			referenceChannels,
			_.bind(function(channel) {
				this._graphicsModel.resetChannel(channel.getID());
			}, this)
		);
	}
};

/**
 * @private
 * Changes the active channel, the active channel is the channel which defintion will be copied to the matched channels.
 * @param {Number} channelId - identifier for the channel that becomes active
 */
ChannelMatchingController.prototype._activateMatchedChannel = function(channelId) {
	this.channelMatchingModel.setChannelActive(channelId);
	this.channelMatchingPopup.render();
};

export default ChannelMatchingController;
