import PrivateChannel from "@esko/web-viewer/core/PrivateChannel";
import GraphicsConstants from "@esko/web-viewer/graphics/GraphicsConstants";
import $ from "jquery";
import _ from "underscore";
/**
 *
 * @exports ChannelMatchingPopup
 * @memberof GraphicsModule
 * @class ChannelMatchingModel holds all the information for the channel matching. It saves all current channels, reference channels and all the matched channels.
 * @constructor
 *
 * @param {Core.ViewerFacade} f - Viewer facade
 * @param {Array.<Channel>} channels - all the channels of the current and reference document page
 */
var ChannelMatchingModel = function(f, channels) {
	PrivateChannel.installTo(this, f, "channelMatchingModel");

	/**
	 * @private
	 * @member {Array.<Channel>} the channels of the active document page
	 */
	this._currentChannels = {};

	/**
	 * @private
	 * @member {Array.<Channel>} the channels of the reference document page
	 */
	this._referenceChannels = {};

	/**
	 * @private
	 * @member {GraphicsConstants.ChannelMatching} how are the channels matched, saved in user preferences
	 */
	this._currentChannelMatching = f.getCurrentUser().getOption("PREFERENCES_COLORMATCHING");

	/**
	 * @private
	 * @member {Array.<Number>} ViewServerId's of all the matched channels of the current document page, index determines to which channel of the reference document page it's matched
	 */
	this._currentMatchedChannels = [];

	/**
	 * @private
	 * @member {Array.<Number>} ViewServerId's of all the matched channel of the reference document page
	 */
	this._referenceMatchedChannels = [];

	/**
	 * @private
	 * @member {Array.<Boolean>} is the channel of the current document page active (copies its definition to the matched channel)
	 */
	this._isCurrentChannelActive = [];

	/**
	 * @private
	 * @member {Array.<Boolean>} collection of all the channel id's of the current document page that have to stay matched, cannot be unmatched by the user
	 */
	this._stayMatchedCurrentChannels = [];

	/**
	 * @private
	 * @member {Array.<Boolean>} collection of all the channel id's of the reference document page that have to stay matched, cannot be unmatched by the user
	 */
	this._stayMatchedReferenceChannels = [];

	/**
	 * @private
	 * @member {Boolean} is the setting of the channel matching persisted (saved in the user preferences) or uses it the default
	 */
	this._isChannelMatchingPersisted = this._currentChannelMatching !== GraphicsConstants.SMART_NAME_MATCHING;

	this.setChannels(channels); // set the channels
};

/**
 * Get a channel of the current document page
 * @param {Number} viewserverId - channel id
 * @returns {Channel}
 */
ChannelMatchingModel.prototype.getCurrentChannel = function(viewserverId) {
	return this._currentChannels[viewserverId];
};

/**
 * Get a channel of the reference document page
 * @param {Number} viewserverId
 * @returns {Channel}
 */
ChannelMatchingModel.prototype.getReferenceChannel = function(viewserverId) {
	return this._referenceChannels[viewserverId];
};

/**
 * get channels of the active document page that are not matched
 * @returns {Array.<Channel>}
 */
ChannelMatchingModel.prototype.getCurrentChannels = function() {
	return _.reject(
		this._currentChannels,
		function(channel, key) {
			return _.contains(this._currentMatchedChannels, parseInt(key, 10));
		},
		this
	);
};

/**
 * get channels of the reference document page that are not matched
 * @returns {Array.<Channel>}
 */
ChannelMatchingModel.prototype.getReferenceChannels = function() {
	return _.reject(
		this._referenceChannels,
		function(channel, key) {
			return _.contains(this._referenceMatchedChannels, parseInt(key, 10));
		},
		this
	);
};

/**
 * Has the current AND the reference document page unmatched channels
 * @returns {Boolean}
 */
ChannelMatchingModel.prototype.hasUnmatchedChannels = function() {
	return this.getCurrentChannels().length !== 0 && this.getReferenceChannels().length !== 0;
};

/**
 * get the matched channels of the active document page
 * @returns {Array.<Channel>}
 */
ChannelMatchingModel.prototype.getMatchedCurrentChannels = function() {
	return _.map(
		this._currentMatchedChannels,
		_.bind(function(channelId) {
			return this._currentChannels[channelId];
		}, this)
	);
};

/**
 * get the matched channels of the reference document page
 * @returns {Array.<Channel>}
 */
ChannelMatchingModel.prototype.getMatchedReferenceChannels = function() {
	return _.map(
		this._referenceMatchedChannels,
		_.bind(function(channelId) {
			return this._referenceChannels[channelId];
		}, this)
	);
};

/**
 * Returns the array of the active channels in the channel matching
 * @returns {Array.<Boolean>}
 */
ChannelMatchingModel.prototype.getIsCurrentChannelActive = function() {
	return this._isCurrentChannelActive;
};

/**
 * Set if the channel matching is persisted or not
 * @param {Boolean} isPersisted
 */
ChannelMatchingModel.prototype.setIsChannelMatchingPersisted = function(isPersisted) {
	this._isChannelMatchingPersisted = isPersisted;
};

/**
 * Returns true if the used channel matching is saved in the user preferences
 * @returns {Boolean}
 */
ChannelMatchingModel.prototype.isChannelMatchingPersisted = function() {
	return this._isChannelMatchingPersisted;
};

/**
 * Sets a channel active, it deactivates the channel it is matched to. Can set current/reference channels active.
 * @param {Number} channelId
 */
ChannelMatchingModel.prototype.setChannelActive = function(channelId) {
	if (_.indexOf(this._currentMatchedChannels, channelId) !== -1) {
		this._isCurrentChannelActive[_.indexOf(this._currentMatchedChannels, channelId)] = true;
	} else if (_.indexOf(this._referenceMatchedChannels, channelId) !== -1) {
		this._isCurrentChannelActive[_.indexOf(this._referenceMatchedChannels, channelId)] = false;
	}
};

/**
 * Returns the currently used channel matching.
 * @returns {GraphicsConstants.ChannelMatching}
 */
ChannelMatchingModel.prototype.getChannelMatching = function() {
	return this._currentChannelMatching;
};

/**
 * Sets the currently used channel matching.
 * @param {GraphicsConstants.ChannelMatching} channelMatching
 */
ChannelMatchingModel.prototype.setChannelMatching = function(channelMatching) {
	this._currentChannelMatching = channelMatching;
};

/**
 * Resets the channel matching.
 */
ChannelMatchingModel.prototype.resetChannelMatching = function() {
	this._currentMatchedChannels = [];
	this._referenceMatchedChannels = [];
	this._isCurrentChannelActive = [];
	this._stayMatchedCurrentChannels = [];
	this._stayMatchedReferenceChannels = [];
};

/**
 * Set all the matched channels of the current document page. It first copies the stay matched channels to the current matched channels
 * @param {Array.<String>} currentMatchedChannels
 */
ChannelMatchingModel.prototype.setMatchedCurrentChannels = function(currentMatchedChannels) {
	//make sure the channels that need to stay matched stay in the matched channels array, they are not in currentMatchedChannels due to JQuery UI
	this._currentMatchedChannels = this._stayMatchedCurrentChannels.slice(0);
	this._currentMatchedChannels = _.union(
		this._currentMatchedChannels,
		_.map(currentMatchedChannels, function(channel) {
			return parseInt(channel, 10);
		})
	);
};

/**
 * Set all the matched channels of the reference document page. It first copies the stay matched channels to the reference matched channels
 * @param {Array.<String>} referenceMatchedChannels
 */
ChannelMatchingModel.prototype.setMatchedReferenceChannels = function(referenceMatchedChannels) {
	this._referenceMatchedChannels = this._stayMatchedReferenceChannels.slice(0);
	this._referenceMatchedChannels = _.union(
		this._referenceMatchedChannels,
		_.map(referenceMatchedChannels, function(channel) {
			return parseInt(channel, 10);
		})
	);
};

/**
 * Match a current and a reference channel.
 * @param {Number} currentChannelId
 * @param {Number} referenceChannelId
 * @param {Boolean} isCurrentCopied - does the reference channel take the definition of the current channel
 * @param {Boolean} stayMatched - does the channel need to stay matched
 */
ChannelMatchingModel.prototype.setChannelMatch = function(currentChannelId, referenceChannelId, isCurrentCopied, stayMatched) {
	if (isCurrentCopied === undefined) {
		isCurrentCopied = false;
	}
	if (_.contains(this._currentMatchedChannels, currentChannelId)) {
		var channelIndex = _.indexOf(this._currentMatchedChannels, currentChannelId);
		this._referenceMatchedChannels[channelIndex] = referenceChannelId;
		this._isCurrentChannelActive[channelIndex] = isCurrentCopied;
		if (stayMatched && !_.contains(this._stayMatchedCurrentChannels, currentChannelId)) {
			this._stayMatchedCurrentChannels.push(currentChannelId);
		}
	} else if (_.contains(this._referenceMatchedChannels, referenceChannelId)) {
		var channelIndex = _.indexOf(this._referenceMatchedChannels, referenceChannelId);
		this._currentMatchedChannels[channelIndex] = currentChannelId;
		this._isCurrentChannelActive[channelIndex] = isCurrentCopied;
		if (stayMatched && !_.contains(this._stayMatchedReferenceChannels, referenceChannelId)) {
			this._stayMatchedReferenceChannels.push(referenceChannelId);
		}
	} else {
		this._currentMatchedChannels.push(currentChannelId);
		this._referenceMatchedChannels.push(referenceChannelId);
		this._isCurrentChannelActive.push(isCurrentCopied);
		if (stayMatched) {
			if (!_.contains(this._stayMatchedCurrentChannels, currentChannelId)) {
				this._stayMatchedCurrentChannels.push(currentChannelId);
			}
			if (!_.contains(this._stayMatchedReferenceChannels, referenceChannelId)) {
				this._stayMatchedReferenceChannels.push(referenceChannelId);
			}
		}
	}
};

/**
 * Make sure that channels that stay matched are on top of the list.
 */
ChannelMatchingModel.prototype.orderMatchedChannels = function() {
	var currentMatchedChannels = this._stayMatchedCurrentChannels.slice(0);
	var referenceMatchedChannels = this._stayMatchedReferenceChannels.slice(0);

	_.each(
		this._currentMatchedChannels,
		_.bind(function(currentChannel) {
			if (!_.contains(currentMatchedChannels, currentChannel)) {
				currentMatchedChannels.push(currentChannel);
			}
		}, this)
	);
	_.each(
		this._referenceMatchedChannels,
		_.bind(function(referenceChannel) {
			if (!_.contains(referenceMatchedChannels, referenceChannel)) {
				referenceMatchedChannels.push(referenceChannel);
			}
		}, this)
	);

	this._currentMatchedChannels = currentMatchedChannels;
	this._referenceMatchedChannels = referenceMatchedChannels;
};

/**
 * Returns the channels of the current document page that have to stay matched
 * @returns {Array.<Number>}
 */
ChannelMatchingModel.prototype.getStayMatchedCurrentChannels = function() {
	return this._stayMatchedCurrentChannels;
};

/**
 * Returns the channels of the reference document page that have to stay matched
 * @returns {Array.<Number>}
 */
ChannelMatchingModel.prototype.getStayMatchedReferenceChannels = function() {
	return this._stayMatchedReferenceChannels;
};

/**
 * Sets the channels of the model. Removes all the previous channels.
 * @param {Array.<Channel>} channels
 */
ChannelMatchingModel.prototype.setChannels = function(channels) {
	this._referenceChannels = {};
	this._currentChannels = {};
	this._currentMatchedChannels = [];
	this._referenceMatchedChannels = [];
	this._isCurrentChannelActive = [];
	this._stayMatchedCurrentChannels = [];
	this._stayMatchedReferenceChannels = [];

	if (channels === undefined) {
		channels = [];
	}
	_.each(
		channels,
		_.bind(function(channel) {
			channel = channel.copy(); //make sure copied channel is used
			if (
				channel.getDiffMode() === GraphicsConstants.ChannelDiffMode.CURRENT ||
				channel.getDiffMode() === GraphicsConstants.ChannelDiffMode.IN_CURRENT_AND_MATCHED_WITH_PREVIOUS
			) {
				this._currentChannels[channel.getViewserverID()] = channel;
			} else {
				this._referenceChannels[channel.getViewserverID()] = channel;
			}
		}, this)
	);
};

export default ChannelMatchingModel;
