import PrivateChannel from "@esko/web-viewer/core/PrivateChannel";
import ListenerUtil from "@esko/web-viewer/utils/ListenerUtil";
import ShapeUtils from "@esko/shapeutils";
import ModalPopup from "@esko/web-viewer/layout/component/ModalPopup";
import Template from "@esko/template";
import tpl from "@esko/web-viewer/graphics/ChannelMatchingPopup.tpl.html";
import $ from "jquery";
import _ from "underscore";
/**
 *
 * @exports ChannelMatchingPopup
 * @memberof GraphicsModule
 * @augments layout.component.ModelPopup
 * @class View for the channel matching. A modal popup that allows channel matching with drag-and-drop operations.
 * @constructor
 *
 * @param {Core.ViewerFacade} f - Viewer facade
 * @param {String} id - Id of the popup. It is used in its private channel.
 * @param {ChannelMatchingModel} channelMatchingModel
 */
var ChannelMatchingPopup = function(f, id, channelMatchingModel) {
	ModalPopup.call(this, f, id, f.getContext());
	PrivateChannel.installTo(this, f, ["channelMatchingPopup"]);
	ListenerUtil.mixin(this, f);

	f.i18n.loadBundle(
		"viewer",
		_.bind(function() {
			this.setHeader(
				$("<span/>")
					.addClass("popupTitle")
					.html(f.i18n.t("MapChanTitle"))
			);
			this.setOkBtnText(f.i18n.t("saveChannelMatching"));
		}, this)
	);

	this.element.addClass("channelsMatchingPopup");

	/**
	 * @private
	 * @member {ChannelMatchingModel} - model used to construct the popup
	 */
	this._channelMatchingModel = channelMatchingModel;

	this.setOkAction(
		_.bind(function() {
			//sets the action when the okAction button is pressed
			this._emit("channelMatchingComplete");
			this.hide();
		}, this)
	);

	/**
	 * @member {Template} template for the popup
	 */
	this.t = new Template(this.f.i18n, tpl);
	this.t.setContext(this._popupContent);

	//action for the radio buttons to change the channel matching, emits channelMatchingPopup.channelMatching
	this.t.bind(
		"change",
		"[name=channelMatching]",
		function(source, event) {
			this._emit("channelMatching", source.value);
		},
		this
	);

	//action for matched channels that are clickable, where the defintion of the reference channel doesn't match the defintion of the current channel
	this.t.bind(
		"click",
		".clickable",
		_.bind(function(source, event) {
			this._emit("matchedChannelClicked", parseInt(source.getAttribute("data-bind"), 10));
		}, this)
	);

	//action for checkbox to persist the selected channel matching option to the user preferences
	this.t.bind(
		"change",
		"[name=saveChannelMatchingAsDefault]",
		_.bind(function(source, event) {
			this._emit("persistDefaultChannelMatching", source.checked);
		}, this)
	);

	this._registerListener("modelChanged", this.render, this, this._channelMatchingModel);
};

ChannelMatchingPopup.prototype = Object.create(ModalPopup.prototype);
ChannelMatchingPopup.prototype.constructor = ChannelMatchingPopup;

/**
 * Renders the popup for channel matching, it's a throttled function
 */
ChannelMatchingPopup.prototype.render = _.throttle(function() {
	this.t.d.currentChannels = this._channelMatchingModel.getCurrentChannels();
	this.t.d.referenceChannels = this._channelMatchingModel.getReferenceChannels();
	this.t.d.matchedCurrentChannels = this._channelMatchingModel.getMatchedCurrentChannels();
	this.t.d.matchedReferenceChannels = this._channelMatchingModel.getMatchedReferenceChannels();
	this.t.d.channelMatching = this._channelMatchingModel.getChannelMatching();
	this.t.d.currentMatchedChannelActive = this._channelMatchingModel.getIsCurrentChannelActive();
	this.t.d.stayMatchedCurrentChannels = this._channelMatchingModel.getStayMatchedCurrentChannels();
	this.t.d.stayMatchedReferenceChannels = this._channelMatchingModel.getStayMatchedReferenceChannels();
	this.t.d.isChannelMatchingPersisted = this._channelMatchingModel.isChannelMatchingPersisted();

	this.t.d.isMatchedChannelClickable = _.map(
		this.t.d.matchedCurrentChannels,
		_.bind(function(channel, key) {
			return this.t.d.matchedReferenceChannels[key] !== undefined && !channel.equals(this.t.d.matchedReferenceChannels[key]);
		}, this)
	);

	this.t.render();

	//Make the channel lists sortable
	var current = this._popupContent.find(".sortableCurrentMatched").sortable({
		items: "li.dragAndDropChannel:not(.stayMatched)", //list-items with the class stayMatched are not possible to be drag-and-dropped
		connectWith: ".sortableCurrent", //connect the list with the current channels list, possible to drag-and-drop to each other
		receive: function(event, ui) {
			ui.item.addClass("clickable"); //make item clickable when entering the list of matched channels
		},
		remove: function(event, ui) {
			ui.item.removeClass("clickable"); //remove clickable when item leaves the list of matched channels
		},
		update: _.bind(function(event, ui) {
			//when drag-and-dropping ended
			var currentMatchedChannels = this._popupContent.find(".sortableCurrentMatched").sortable("toArray", {
				attribute: "data-bind"
			});
			var toMatchedList = false;
			if (ui.sender !== null) {
				// if sender is not null, the item was dropped in the matched list
				toMatchedList = true;
			}
			this.recalculateSize();
			this._emit("currentChannelMatch", ui.item.context.getAttribute("data-bind"), currentMatchedChannels, toMatchedList);
		}, this)
	});

	//same as the list for the matched current channels
	var reference = this._popupContent.find(".sortableReferenceMatched").sortable({
		items: "li.dragAndDropChannel:not(.stayMatched)",
		connectWith: ".sortableReference",
		receive: function(event, ui) {
			ui.item.addClass("clickable");
		},
		remove: function(event, ui) {
			ui.item.removeClass("clickable");
		},
		update: _.bind(function(event, ui) {
			var referenceMatchedChannels = this._popupContent.find(".sortableReferenceMatched").sortable("toArray", {
				attribute: "data-bind"
			});
			var toMatchedList = false;
			if (ui.sender !== null) {
				toMatchedList = true;
			}
			this.recalculateSize();
			this._emit("referenceChannelMatch", ui.item.context.getAttribute("data-bind"), referenceMatchedChannels, toMatchedList);
		}, this)
	});

	//list for the channels of the current document page
	this._popupContent.find(".sortableCurrent").sortable({
		items: "li.dragAndDropChannel",
		connectWith: current, //connect the list with the matched channels of the current document page
		stop: _.bind(this.render, this) //make sure the list itself is not sortable, only possible to move items to match list
	});

	//list for the channels of the reference document page
	this._popupContent.find(".sortableReference").sortable({
		items: "li.dragAndDropChannel",
		connectWith: reference,
		stop: _.bind(this.render, this)
	});
}, 40);

/**
 * Shows the popup.
 */
ChannelMatchingPopup.prototype.show = function() {
	this.render();

	ModalPopup.prototype.show.call(this);
};

export default ChannelMatchingPopup;
