import Popup from "@esko/web-viewer/layout/component/Popup";
import Template from "@esko/template";
import tpl from "@esko/web-viewer/graphics/ChannelInformationPopup.tpl.html";
import ShapeUtils from "@esko/shapeutils";
import ButtonGroup from "@esko/toolbar/ButtonGroup";
import ColorPicker from "@esko/web-viewer/layout/component/ColorPicker";
import ViewerToolbarButton from "@esko/web-viewer/layout/component/ViewerToolbarButton";
import ViewerUtils from "@esko/web-viewer/utils/ViewerUtils";
import $ from "jquery";
import _ from "underscore";
/**
 * Create a popup for the channel information.
 * @exports ChannelPopup
 * @memberof Graphics
 * @constructor
 * @augments Layout.Popup
 *
 * @param {Core.ViewerFacade} f - Viewerfacade
 * @param {String} popupId - Id of the popup used for its private channel
 * @param {GraphicsModel} graphicsModel - the model that represents holds the information on all the channels
 */
var ChannelPopup = function(f, popupId, graphicsModel) {
	Popup.call(this, f, popupId, f.getContext());

	this.f = f;

	this.element.addClass("channelsInfoPopup");

	/**
	 * @private
	 */
	this._graphicsModel = graphicsModel;

	/**
	 * This colorpicker is used to pick another channel RGB value
	 * @private
	 */
	this._colorPicker = null;

	/**
	 * Channel that is selected and for which the information is shown in the popup
	 */
	this.channel = null;

	/**
	 * @private
	 */
	this._nextButton = new ViewerToolbarButton(
		this.f,
		"nextChannel",
		"icon-WCR_Table-Page-Next12",
		false,
		_.bind(function() {
			var nextChannel = this._graphicsModel.getNextChannel(this.channel.getID());
			if (nextChannel) {
				this.setSelectedChannel(nextChannel);
			}
		}, this)
	);

	/**
	 * @private
	 */
	this._previousButton = new ViewerToolbarButton(
		this.f,
		"previousChannel",
		"icon-WCR_Table-Page-Previous12",
		false,
		_.bind(function() {
			var previousChannel = this._graphicsModel.getPreviousChannel(this.channel.getID());
			if (previousChannel) {
				this.setSelectedChannel(previousChannel);
			}
		}, this)
	);

	this._stepButtonGroup = new ButtonGroup("stepThroughChannels", [this._previousButton, this._nextButton]);

	this._tplContent = $("<div>");
	this._popupBtns = $("<div>").addClass("channelPopupBtns");

	f.i18n.loadBundle(
		"viewer",
		_.bind(function() {
			/**
			 * @private
			 */
			this._revertButton = new ViewerToolbarButton(
				this.f,
				"undoChannelRgb",
				"icon-WCR_Actions-Undo12",
				false,
				_.bind(function() {
					this._graphicsModel.resetChannel(this.channel.getID());
				}, this)
			);
			this._revertButton.element.append(f.i18n.t("AppDialogRevert"));
			this._revertButton.element.addClass("undoChannelRgbBtn");

			this.content.append(this._tplContent);
			this._popupBtns.append(this._revertButton.element);
			this._popupBtns.append(this._stepButtonGroup.element);
			this._outerContent.append(this._popupBtns);

			this.setHeader(
				$("<span />")
					.addClass("popupTitle")
					.html(f.i18n.t("ChanEditTitle"))
			);
		}, this)
	);

	/**
	 * template for the popup
	 */
	this.t = new Template(this.f.i18n, tpl);
	this.t.setContext(this._tplContent);

	this.t.bind("change", "[name=r],[name=g],[name=b]", this._rgbValueChanged, this);
	this.t.bind("click", "[name=opaque]", this._opaqueValueChanged, this);
	this.t.bind("change", ".channelSelectName", this._channelDefinitionChanged, this);
	this.t.bind(
		"click",
		".colorChangeBlock",
		function(div, e) {
			if (this._colorPicker) {
				this._colorPicker.destroy();
			}
			this._colorPicker = new ColorPicker(
				this.f,
				_.bind(function(newColor) {
					this._changeChannelColor(ViewerUtils.intFromHex(newColor));
				}, this),
				$(div)
			);
			this._colorPicker.setColor(this.channel.getHexRGB());
			this._colorPicker.show();
			e.stopPropagation();
		},
		this
	);

	this.on(
		"closed",
		_.bind(function() {
			this.hide();
		}, this)
	);

	this._debouncedRender = _.debounce(this.render, 50);

	//Listen to model for change in channel definition
	this._graphicsModel.on(
		"changed.channel.*.channelRGB",
		_.bind(function() {
			if (this.visible) {
				this._debouncedRender();
			}
		}, this)
	);
};

ChannelPopup.prototype = Object.create(Popup.prototype);
ChannelPopup.prototype.constructor = ChannelPopup;

/**
 * Called when the checkbox for the opaque value changes
 * @param {HTMLObject} el - the element that was clicked
 * @param {Event} e - the event object
 */
ChannelPopup.prototype._opaqueValueChanged = function(el, e) {
	this._graphicsModel.setChannelOpaque(this.channel.getID(), el.checked);
};

/**
 * Called when one of the values (R,G,B) is changed, changes the values for RGB, emits that this has changed and rerenders the popup
 * @param {HTMLObject} el - the element that was clicked
 * @param {Event} e - the event object
 */
ChannelPopup.prototype._rgbValueChanged = function(el, e) {
	var colorRGB = ViewerUtils.rgbColor(this.channel.getHexRGB());
	if (el.name === "r") {
		colorRGB.r = parseInt(el.value, 10);
	} else if (el.name === "g") {
		colorRGB.g = parseInt(el.value, 10);
	} else if (el.name === "b") {
		colorRGB.b = parseInt(el.value, 10);
	}
	this._changeChannelColor(ViewerUtils.intFromHex(ViewerUtils.rgbToHex(colorRGB.r, colorRGB.g, colorRGB.b)));
};

/**
 * Called when the color of the selected channel is changed
 * @param {Number} color - Integer representation of the rgb color
 */
ChannelPopup.prototype._changeChannelColor = function(color) {
	this._graphicsModel.changeChannelRgb(this.channel.getID(), color);
};

/**
 * Called when the RGB of another channel will be used for the current channel, also changes the displayname of the channel
 * @param {HTMLObject} el - the element that was clicked
 * @param {Event} e - the event object
 */
ChannelPopup.prototype._channelDefinitionChanged = function(el, e) {
	var channelId = el.value;
	var isOriginalChannelDef = $(el)
		.find(":selected")
		.hasClass("originalChannelDef");
	this._graphicsModel.changeChannelDef(this.channel.getID(), channelId, isOriginalChannelDef);
};

/**
 * Set the current selected channel, should be called before showing the popup
 * @param {graphics.Channel} channel
 */
ChannelPopup.prototype.setSelectedChannel = function(channel) {
	this.channel = channel;
	this.updateStepButtons();
	this.updateRevertButton();
	this.render();
};

/**
 * Show the popup, be sure the correct channel has been set
 * @param {ShapeUtils.Point} location - location where the popup should be drawn
 */
ChannelPopup.prototype.show = function(location) {
	Popup.prototype.show.call(this, location, 250);

	this.render();
};

/**
 * check if the previous and nextbutton need to be enabled or disabled
 */
ChannelPopup.prototype.updateStepButtons = function() {
	this._previousButton.setEnabled(this._graphicsModel.hasPreviousChannel(this.channel.getID()));
	this._nextButton.setEnabled(this._graphicsModel.hasNextChannel(this.channel.getID()));
};

/**
 * Updates the revert button when needed
 */
ChannelPopup.prototype.updateRevertButton = function() {
	if (this.channel.isChanged()) {
		this._revertButton.setEnabled(true);
	} else {
		this._revertButton.setEnabled(false);
	}
};

/**
 * Sets the correct values for the template and renders the template
 */
ChannelPopup.prototype.render = function() {
	this.t.d.channelName = this.channel.getName();
	this.t.d.originalName = this.channel.name;
	this.t.d.channelColor = this.channel.getHexRGB();
	this.t.d.originalChannelColor = ViewerUtils.hexFromInt(this.channel.rgb);
	this.t.d.channels = this._graphicsModel.getChannels();
	this.t.d.rgb = ViewerUtils.rgbColor(this.channel.getHexRGB());
	this.t.d.originalRgb = ViewerUtils.rgbColor(ViewerUtils.hexFromInt(this.channel.rgb));
	this.t.d.isOpaque = this.channel.isChannelOpaque();
	this.t.d.wasOpaque = this.channel.wasChannelOpaque();
	this.updateStepButtons();
	this.updateRevertButton();
	this.t.render();
};

export default ChannelPopup;
