import AbstractWidget from "@esko/web-viewer/layout/AbstractWidget";
import Template from "@esko/template";
import tpl from "@esko/web-viewer/graphics/ChannelsWidget.tpl.html";
import GraphicsConstants from "@esko/web-viewer/graphics/GraphicsConstants";
import ViewerToolbarButton from "@esko/web-viewer/layout/component/ViewerToolbarButton";
import ButtonGroup from "@esko/toolbar/ButtonGroup";
import DropdownToolbarButton from "@esko/web-viewer/layout/component/DropdownToolbarButton";
import ColorPicker from "@esko/web-viewer/layout/component/ColorPicker";
import LayoutConstants from "@esko/web-viewer/layout/LayoutConstants";
import ChannelPopup from "@esko/web-viewer/graphics/ChannelPopup";
import ShapeUtils from "@esko/shapeutils";
import ViewerUtils from "@esko/web-viewer/utils/ViewerUtils";
import CompareConstants from "@esko/web-viewer/compare/CompareConstants";
import EyeDropperUnitSelector from "@esko/web-viewer/measure/EyeDropperUnitSelector";
import ScreenRulingUnitSelector from "@esko/web-viewer/measure/ScreenRulingUnitSelector";
import ListenerUtil from "@esko/web-viewer/utils/ListenerUtil";
import CoreConstants from "@esko/web-viewer/core/CoreConstants";
import $ from "jquery";
import _ from "underscore";
import "jquery-ui/ui/widgets/sortable";
import "jquery-ui-touch-punch";
/**
 * @exports ChannelsWidget
 * @memberof Graphics
 * @augments Layout.AbstractWidget
 * @class The ChannelsWidget
 * @constructor
 *
 * @param {Core.ViewerFacade} f
 * @param {GraphicsModel} graphicsModel
 */
var ChannelsWidget = function ChannelsWidget(f, graphicsModel) {
	AbstractWidget.call(this, f, LayoutConstants.Widget.CHANNELS);
	this.f = f;
	ListenerUtil.mixin(this, f);
	// set the model
	this._graphicsModel = null;
	this.setGraphicsModel(graphicsModel);

	this.isOrderingEnabled = true;
	this._contentDiv = $("<div/>");

	this._colorPicker = null;

	// Settings
	this._eyeDropperUnitSelector = null;
	this._screenRulingUnitSelector = null;

	this.setIcon("icon icon-WCR_Channel-Color");

	this.t = new Template(this.f.i18n, tpl);
	this.t.setContext(this._contentDiv);

	//Get the list of channels from the model
	this.t.d.getChannels = _.bind(function() {
		if (this._graphicsModel) {
			var channels = this._graphicsModel.getChannels(true);

			return channels;
		} else {
			return [];
		}
	}, this);

	this.t.d.getChannelsInfo = function(channel) {
		if (this.channelsInfo) {
			return this.channelsInfo[channel.getViewserverID()] || "";
		} else {
			return "";
		}
	};

	//check if all channels are enabled
	this.t.d.isAllEnabled = _.bind(function() {
		return this.isAllChannelsEnabled();
	}, this);

	// get the suffix for a channel (if it should be shown)
	this.t.d.getChannelSuffix = _.bind(function(channel) {
		if (this._graphicsModel.isInDifferenceCompareMode()) {
			if (channel.isAdded()) {
				return "(+)";
			}
			if (channel.isSubtracted()) {
				return "(-)";
			}
		} else {
			return "";
		}
	}, this);

	//listen to measure tools if information is added
	this.f.on(
		"channels.informationAdded",
		_.bind(function(channelsInfo) {
			this.t.d.channelsInfo = [];

			var parsedInfo = this._graphicsModel.getParsedChannelsInfo(channelsInfo);

			_.each(
				parsedInfo.channels,
				function(channel) {
					if (channel) {
						this.t.d.channelsInfo[channel.getViewserverID()] = channel.value;
					}
				},
				this
			);

			this.t.d.totalCount = parsedInfo.total;

			this.render();
		}, this)
	);

	this.f.on(
		"channels.informationNotVisible",
		_.bind(function() {
			this.t.d.channelsInfo = null;
			this.t.d.totalCount = "";
			this.render();
		}, this)
	);

	/**
	 * The popup that shows the channel information if the color box of that channel is clicked
	 */
	this.channelPopup = new ChannelPopup(f, "channelInformation", graphicsModel);

	// Split channel button
	this._splitChannels = new DropdownToolbarButton(
		f,
		"splitChannels",
		" icon-WCR_Tiling-Auto",
		true,
		_.bind(function(activated) {
			if (activated) {
				this.setChannelViewMode(GraphicsConstants.ChannelViewMode.AUTOMATIC_TILING);
			} else {
				this.setChannelViewMode(GraphicsConstants.ChannelViewMode.COMPOSITE_VIEW);
			}
		}, this)
	);
	this._splitChannels.element.addClass("toggleBtn");

	// single channel button
	this._singleChannelMode = new DropdownToolbarButton(
		f,
		"singleChannelMode",
		"icon-WCR_Channel-Black",
		true,
		_.bind(function(activated) {
			if (activated) {
				this.setSingleChannelMode(GraphicsConstants.SingleChannelMode.IN_BLACK);
			} else {
				this.setSingleChannelMode(GraphicsConstants.SingleChannelMode.IN_CHANNEL_COLOR);
			}
		}, this)
	);
	this._singleChannelMode.element.addClass("toggleBtn");

	this._resetChannelsButton = new DropdownToolbarButton(
		f,
		"resetChannels",
		"icon-WCR_Revert",
		true,
		_.bind(function() {
			this._graphicsModel.resetDefaultChannelOrder();
			this._graphicsModel.resetAllChannelsColorDef();
		}, this)
	);

	// Move channel(s) up button
	this._moveChannelsUp = new ViewerToolbarButton(
		f,
		"moveChannelUp",
		"icon-WCR_Up",
		false,
		_.bind(function() {
			var i;
			var channel;
			if (this._graphicsModel.hasEnabledChannelGaps()) {
				// there are gaps in the list, move all elements together
				var moveTo = null;
				for (i = 0; i < this._graphicsModel.getChannelCount(); i++) {
					channel = this._graphicsModel.getChannelAtIndex(i);
					if (channel.isEnabled()) {
						if (moveTo === null || moveTo === i) {
							// there was no enabled channel found yet, from now on, move all elements just after this element
							moveTo = i + 1;
						} else {
							// there was an enabled channel found, move it to the enabled channel block
							this._graphicsModel.moveChannelTo(channel.getID(), moveTo);
							moveTo++;
						}
					}
				}
			} else {
				// we can only move up if the first channel is not enabled
				if (this._graphicsModel.getChannelCount() >= 2 && !this._graphicsModel.getChannelAtIndex(0).isEnabled()) {
					// move every enabled channel 1 up
					i = 1;
					while (i < this._graphicsModel.getChannelCount()) {
						channel = this._graphicsModel.getChannelAtIndex(i);
						if (channel.isEnabled()) {
							this._graphicsModel.moveChannelTo(channel.getID(), i - 1);
						} else {
							i++;
						}
					}
				}
			}
		}, this)
	);

	// Move channel(s) down button
	this._moveChannelsDown = new ViewerToolbarButton(
		f,
		"moveChannelDown",
		"icon-WCR_Down",
		false,
		_.bind(function() {
			var i;
			var channel;
			if (this._graphicsModel.hasEnabledChannelGaps()) {
				// there are gaps in the list, move all elements together
				var moveTo = null;
				for (i = this._graphicsModel.getChannelCount() - 1; i >= 0; i--) {
					channel = this._graphicsModel.getChannelAtIndex(i);
					if (channel.isEnabled()) {
						if (moveTo === null || moveTo === i) {
							// there was no enabled channel found yet, from now on, move all elements just before this element
							moveTo = i - 1;
						} else {
							// there was an enabled channel found, move it to the enabled channel block
							this._graphicsModel.moveChannelTo(channel.getID(), moveTo);
							moveTo--;
						}
					}
				}
			} else {
				// we can only move down if the last channel is not enabled
				if (
					this._graphicsModel.getChannelCount() > 1 &&
					!this._graphicsModel.getChannelAtIndex(this._graphicsModel.getChannelCount() - 1).isEnabled()
				) {
					// move every enabled channel 1 up
					i = graphicsModel.getChannelCount() - 2;
					while (i >= 0) {
						channel = this._graphicsModel.getChannelAtIndex(i);
						if (channel.isEnabled()) {
							this._graphicsModel.moveChannelTo(channel.getID(), i + 1);
						} else {
							i--;
						}
					}
				}
			}
		}, this)
	);

	this._channelMatching = new ViewerToolbarButton(f, "channelMatching", "icon-WCR_Channel-Matching", false, function() {});
	this._channelMatching.element.hide();

	this._moveChannelsButtonGroup = new ButtonGroup("moveChannels", [this._moveChannelsUp, this._moveChannelsDown]);

	f.i18n.loadBundle(
		"viewer",
		_.bind(function() {
			// set the widget title
			this.setTitle(this.f.t("ChannelsTitle"));

			//set tooltips for the buttons
			this._splitChannels.setTooltip(f.i18n.t("ChanSplitChannelsTT"));
			this._singleChannelMode.setTooltip(f.i18n.t("ChanSingleTT"));
			this._moveChannelsDown.setTooltip(f.i18n.t("ChanDownTT"));
			this._moveChannelsUp.setTooltip(f.i18n.t("ChanUpTT"));
			this._resetChannelsButton.setTooltip(f.i18n.t("ChanResetDefsTT"));
			this._channelMatching.setTooltip(f.i18n.t("MapChanTitle"));

			// set the options
			this._splitChannels.addOption({
				name: "composite",
				content: f.i18n.t("SplitChannelsOff"),
				onchange: _.bind(function() {
					this.setChannelViewMode(GraphicsConstants.ChannelViewMode.COMPOSITE_VIEW);
				}, this),
				userCannotUncheck: true
			});

			this._splitChannels.addOption({
				name: "automatic",
				content: f.i18n.t("SplitChannelsTileAuto"),
				onchange: _.bind(function() {
					this.setChannelViewMode(GraphicsConstants.ChannelViewMode.AUTOMATIC_TILING);
				}, this),
				userCannotUncheck: true
			});

			this._splitChannels.addOption({
				name: "horizontal",
				content: f.i18n.t("SplitChannelsTileHor"),
				onchange: _.bind(function() {
					this.setChannelViewMode(GraphicsConstants.ChannelViewMode.HORIZONTAL_TILING);
				}, this),
				userCannotUncheck: true
			});

			this._splitChannels.addOption({
				name: "vertical",
				content: f.i18n.t("SplitChannelsTileVer"),
				onchange: _.bind(function() {
					this.setChannelViewMode(GraphicsConstants.ChannelViewMode.VERTICAL_TILING);
				}, this),
				userCannotUncheck: true
			});

			// initial option
			this.setChannelViewMode(this._graphicsModel.getChannelViewMode());

			// single channel as black

			this._singleChannelMode.addOption({
				name: "inChannelColor",
				content: f.i18n.t("ChanSingleOff"),
				onchange: _.bind(function() {
					this.setSingleChannelMode(GraphicsConstants.SingleChannelMode.IN_CHANNEL_COLOR);
				}, this),
				userCannotUncheck: true
			});

			this._singleChannelMode.addOption({
				name: "inBlack",
				content: f.i18n.t("ChanSingleOn"),
				onchange: _.bind(function() {
					this.setSingleChannelMode(GraphicsConstants.SingleChannelMode.IN_BLACK);
				}, this),
				userCannotUncheck: true
			});

			this._singleChannelMode.addOption({
				name: "inBlackWithShowAll",
				content: f.i18n.t("ChanSingleShowAll"),
				onchange: _.bind(function() {
					this.setSingleChannelMode(GraphicsConstants.SingleChannelMode.IN_BLACK_WITH_SHOW_ALL_EFFECT);
				}, this),
				userCannotUncheck: true
			});

			// initial option
			this.setSingleChannelMode(this._graphicsModel.getSingleChannelMode());

			this._resetChannelsButton.addAction({
				name: "resetOnlyChannelOrder",
				content: f.i18n.t("ChanResetOrder"),
				action: _.bind(function() {
					this._graphicsModel.resetDefaultChannelOrder();
				}, this)
			});

			this._resetChannelsButton.addAction({
				name: "resetOnlyChannelsColorDef",
				content: f.i18n.t("ChanResetColor"),
				action: _.bind(function() {
					this._graphicsModel.resetAllChannelsColorDef();
				}, this)
			});

			this._resetChannelsButton.addAction({
				name: "resetBothChannelsColorDefAndOrder",
				content: f.i18n.t("ChanResetBoth"),
				action: _.bind(function() {
					this._graphicsModel.resetDefaultChannelOrder();
					this._graphicsModel.resetAllChannelsColorDef();
				}, this)
			});
		}, this)
	);

	//Initialize keyevents for the channels widget
	this.keyEvents = [];
	this.keyEvents.push([
		["alt+1", "alt+2", "alt+3", "alt+4", "alt+5", "alt+6", "alt+7", "alt+8", "alt+9"],
		_.bind(function(e) {
			e.preventDefault();
			//get the number of the key pressed, 96 and up are the keys on the numpad, 48 and up are the number keys
			var number = e.keyCode - 96 < 0 ? e.keyCode - 48 : e.keyCode - 96;
			var channels = this._graphicsModel.getChannels(true);

			if (number <= channels.length) {
				//only if the channel exists
				_.each(
					channels,
					_.bind(function(channel, index) {
						this._graphicsModel.toggleChannel(channel.getID(), index == number - 1);
					}, this)
				);
			}
		}, this)
	]);
	//select all channels
	this.keyEvents.push([
		["alt+a"],
		_.bind(function(e) {
			e.preventDefault();
			_.each(
				this._graphicsModel.getChannels(),
				_.bind(function(channel, index) {
					this._graphicsModel.toggleChannel(channel.getID(), true);
				}, this)
			);
		}, this)
	]);
	//enable first selected channel if more than 1 is enabled, if not select the next channel
	this.keyEvents.push([
		["alt+n"],
		_.bind(function(e) {
			e.preventDefault();
			var channels = this._graphicsModel.getChannels(true);

			var channelID = _.find(channels, function(channel) {
				return channel.isEnabled();
			}).getID();

			if (this._graphicsModel.getEnabledChannelCount(true) > 1) {
				_.each(
					channels,
					_.bind(function(channel, index) {
						this._graphicsModel.toggleChannel(channel.getID(), channel.getID() == channelID); //only enable first selected channel
					}, this)
				);
			} else {
				this._graphicsModel.toggleChannel(channelID, false);

				var i = 0;
				while (i < channels.length && channels[i].getID() !== channelID) {
					i++;
				}
				var nextChannel = channels[(i + 1) % channels.length];
				this._graphicsModel.toggleChannel(nextChannel.getID(), true);
			}
		}, this)
	]);
	this.keyEvents.push([
		[
			"ctrl+alt+0",
			"ctrl+alt+1",
			"ctrl+alt+2",
			"ctrl+alt+3",
			"ctrl+alt+4",
			"ctrl+alt+5",
			"ctrl+alt+6",
			"ctrl+alt+7",
			"ctrl+alt+8",
			"ctrl+alt+9"
		],
		_.bind(function(e) {
			e.preventDefault();
			e.stopPropagation();
			//get the number of the key pressed, 96 and up are the keys on the numpad, 48 and up are the number keys
			var number = e.keyCode - 96 < 0 ? e.keyCode - 48 : e.keyCode - 96;
			number = number + 10; //because these are channels 11 to 19
			var channels = this._graphicsModel.getChannels(true);

			if (number - 1 < channels.length) {
				//only if the channel exists
				_.each(
					channels,
					_.bind(function(channel, index) {
						this._graphicsModel.toggleChannel(channel.getID(), index == number - 1); //-1 because array starts at index 0
					}, this)
				);
			}
		}, this),
		"keydown"
	]);
	this.bindKeys(this.keyEvents);

	// react on eyebox clicks
	if (!this.f.isMobile()) {
		this.t.bind("mouseup", ".eyeBox", this.eyeBoxClicked, this);
	} else {
		// double tap detection
		this._clickTimer = null;
		this._lastClickedElement = null;
		this.t.bind(
			"touchstart pointerdown",
			".eyeBox",
			function(el, event) {
				if (this._clickTimer == null || this._lastClickedElement !== el) {
					this._lastClickedElement = el;
					this._clickTimer = setTimeout(
						_.bind(function() {
							this._clickTimer = null;
							// single tap
							this.eyeBoxClicked(el, event);
						}, this),
						350
					);
				} else {
					clearTimeout(this._clickTimer);
					this._clickTimer = null;
					// double tap
					event.which = 3;
					this.eyeBoxClicked(el, event);
				}
			},
			this
		);
	}

	// react on colorbox clicks
	this.t.bind("click", ".colorBox,.channelName", this.colorBoxClicked, this);
	this.t.bind("contextmenu", ".colorBox", this.colorBoxRightClicked, this);

	// do not show context menus
	this.t.bind(
		"contextmenu",
		".eyeBox",
		function(el, e) {
			e.preventDefault();
			return false;
		},
		this
	);

	this.f.on(
		"module.compare.changed.mode",
		_.bind(function(oldMode, newMode) {
			var enableOrdering = true;
			if (
				newMode &&
				(newMode.getName() === CompareConstants.mode.BLINKING ||
					newMode.getName() === CompareConstants.mode.DIFFERENCE ||
					newMode.getName() === CompareConstants.mode.SHOWBOTHONTOP ||
					newMode.getName() === CompareConstants.mode.SIDEBYSIDE)
			) {
				enableOrdering = false;
			}
			this.toggleOrdering(enableOrdering);
		}, this)
	);

	// React on active page changes
	this.f.getDocumentController().whenActivePageChangedAndParametersAvailable(this._updateEyeDropperUnitSelectorButton, this);
	// React on tool activate changes
	this.f.on("tool.measureEyeDropperTool.activated", _.bind(this._updateEyeDropperUnitSelectorButton, this));
	this.f.on("tool.measureEyeDropperTool.deactivated", _.bind(this._updateEyeDropperUnitSelectorButton, this));
	this.f.on("tool.measureScreenRulingTool.activated", _.bind(this._updateScreenRulingUnitSelectorButton, this));
	this.f.on("tool.measureScreenRulingTool.deactivated", _.bind(this._updateScreenRulingUnitSelectorButton, this));
};

ChannelsWidget.prototype = Object.create(AbstractWidget.prototype);
ChannelsWidget.prototype.constructor = ChannelsWidget;
ChannelsWidget.widgetName = LayoutConstants.Widget.CHANNELS;

/**
 * Set the context of the widget
 * @override
 * @param context
 * @param header
 */
ChannelsWidget.prototype.setContext = function(context, header) {
	AbstractWidget.prototype.setContext.call(this, context, header);
	this.context.addClass("channelsWidget");

	// eye dropper unit selector for digital film if needed
	this._updateEyeDropperUnitSelectorButton();

	var buttons = $("<div/>").addClass("buttons");
	buttons.append(this._splitChannels.element);
	buttons.append(this._singleChannelMode.element);

	// only add reset, move and channelmatching for non mobile devices
	if (!this.f.isMobile()) {
		buttons.append(this._resetChannelsButton.element);
		buttons.append(this._moveChannelsButtonGroup.element);
		buttons.append(this._channelMatching.element);
	} else {
		// mobile
		this.headerExtra.addClass("flexEnd").append(this._resetChannelsButton.element);
	}
	this.context.append(buttons);
	this.context.append(this._contentDiv);
	this.render();
};

/**
 * Shows the button for channel matching
 * @param {Function} action
 */
ChannelsWidget.prototype.showChannelMatchingBtn = function(action) {
	this._channelMatching.setAction(action);
	this._channelMatching.element.show();
};

/**
 * hides the button for channel matching
 */
ChannelsWidget.prototype.hideChannelMatchingBtn = function() {
	this._channelMatching.element.hide();
};

/**
 * Set the graphicsModel for the widget
 * @param {GraphicsModel} graphicsModel
 */
ChannelsWidget.prototype.setGraphicsModel = function(graphicsModel) {
	this._unregisterAllListeners();

	this._graphicsModel = graphicsModel;

	var throttledRender = _.throttle(
		_.bind(function() {
			this.render();
		}, this),
		40
	);

	//render the widget when the graphicsmodel changes
	throttledRender();

	this._registerListener("changed.channel.**", throttledRender, this, this._graphicsModel);
	this._registerListener("changed.channels", throttledRender, this, this._graphicsModel);
	this._registerListener(
		"changed.channelViewMode",
		_.bind(function(oldValue, newValue) {
			this._updateChannelViewModeBtn(newValue);
		}, this),
		this,
		this._graphicsModel
	);
	this._registerListener("changed.channelOrder", throttledRender, this, this._graphicsModel);
};

/**
 * A channel eyebox was clicked, execute needed actions
 * @param {HTMLObject} el - the element that was clicked
 * @param {Event} event - the event object
 */
ChannelsWidget.prototype.eyeBoxClicked = function(el, event) {
	var channels;
	//get the channel id
	var id = $(el).attr("data-bind");
	if (id == "all") {
		// the "ALL" box was clicked
		if (this.isAllChannelsEnabled()) {
			channels = this._graphicsModel.getChannels();
			// if all channels are enabled, disable all channels, except the first
			_.each(
				channels,
				function(channel, index) {
					if (index > 0) {
						this._graphicsModel.toggleChannel(channel.getID(), false);
					}
				},
				this
			);
			this._graphicsModel.toggleChannel(channels[0].getID(), true);
		} else {
			// some channels were deselected, enable all channels again
			_.each(
				this._graphicsModel.getChannels(),
				function(channel, index) {
					this._graphicsModel.toggleChannel(channel.getID(), true);
				},
				this
			);
		}
	} else {
		if (event.which === 3) {
			// right click, only select the clicked channel
			_.each(
				this._graphicsModel.getChannels(),
				function(channel, index) {
					this._graphicsModel.toggleChannel(channel.getID(), false);
				},
				this
			);
			this._graphicsModel.toggleChannel(id, true);
		} else {
			var newToggleValue = !$(el).hasClass("selected");
			this._graphicsModel.toggleChannel(id, newToggleValue);

			if (this.isNoChannelEnabled()) {
				// no channels are selected anymore, find the next channel and select it
				channels = this._graphicsModel.getChannels();
				var i = 0;
				while (i < channels.length && channels[i].getID() !== id) {
					i++;
				}
				var nextChannel = channels[(i + 1) % channels.length];
				this._graphicsModel.toggleChannel(nextChannel.getID(), true);
			}
		}
	}
};

/**
 * A channel colorblock was clicked, set correct data for the popup and show popup
 * @param {HTMLObject} el - the element that was clicked
 * @param {Event} event - the event object
 */
ChannelsWidget.prototype.colorBoxClicked = function(el, event) {
	if (!this.f.isMobile()) {
		if (event.which < 3) {
			//get the channel id
			var id = $(el).attr("data-bind");
			var channel = null;
			if (id === "all") {
				// if the colorbox of "all" is clicked, show the channelpopup for the first channel
				channel = _.first(this._graphicsModel.getChannels(true));
			} else {
				channel = this._graphicsModel.getChannel(id);
			}

			this.channelPopup.setSelectedChannel(channel);

			//checks if there is room to the left or the right of the widget
			var left = this.context.offset().left < 250 ? this.context.width() || null : this.context.offset().left - 250;
			var location = new ShapeUtils.Point(left, this.context.offset().top);
			this.channelPopup.show(location);
		} else if (event.which === 3) {
			// right click... show the colorpicker
			this.colorBoxRightClicked(el, event);
		}
	}
};

/**
 * A channel colorblock was right clicked, show a color picker for the channel RGB definition
 * @param {HTMLObject} el - the element that was clicked
 * @param {Event} event - the event object
 */
ChannelsWidget.prototype.colorBoxRightClicked = function(el, event) {
	event.preventDefault();

	var id = $(el).attr("data-bind");

	if (id !== "all") {
		var channel = this._graphicsModel.getChannel(id);

		if (this._colorPicker) {
			this._colorPicker.destroy();
			this._colorPicker = null;
		}
		this._colorPicker = new ColorPicker(
			this.f,
			_.bind(function(newColor) {
				this._graphicsModel.changeChannelRgb(id, ViewerUtils.intFromHex(newColor));
			}, this),
			$("<div/>").insertAfter($(el))
		);

		this._colorPicker.setColor(channel.getHexRGB());
		this._colorPicker.show();
	}
};

/**
 * Render the widget (note, this is a debounced function to avoid unnecessary redraws
 */
ChannelsWidget.prototype.render = _.debounce(function() {
	//make drag and drop for channel list
	if (this.isContextAvailable()) {
		this.t.render();
		var sortableHandle;
		if (this.f.isMobile()) {
			sortableHandle = ".colorBox";
		}
		this.context.find(".sortableChannelsList").sortable({
			items: "tr.dragAndDropChannel",
			update: _.bind(function() {
				//when the channel is dropped on new location
				this._graphicsModel.channelOrderChanged = true;
				var list = this.context.find(".sortableChannelsList").sortable("toArray", {
					attribute: "data-bind"
				});
				this._graphicsModel.setChannelOrder(list);
				if (this.channelPopup.visible) {
					//hide channelPopup if the order is changed
					this.channelPopup.hide();
				}
				this.updateRevertButton();
			}, this),
			handle: sortableHandle
		});
		this.context.find(".sortableChannelsList").disableSelection();
		this.updateMoveChannelsButtons();
		this.updateRevertButton();
		this._fireResized();
	}
}, 40);

/**
 * @returns {Boolean} true if all channels are enabled
 */
ChannelsWidget.prototype.isAllChannelsEnabled = function() {
	return _.every(this._graphicsModel.getChannels(), function(channel) {
		return channel.isEnabled();
	});
};

/**
 * @returns {Boolean} true if no channel is enabled
 */
ChannelsWidget.prototype.isNoChannelEnabled = function() {
	return _.every(this._graphicsModel.getChannels(), function(channel) {
		return !channel.isEnabled();
	});
};

/**
 * Set the channel view mode
 * @param {GraphicsConstants.ChannelViewMode} mode
 */
ChannelsWidget.prototype.setChannelViewMode = function(mode) {
	this._updateChannelViewModeBtn(mode);
	this._graphicsModel.setChannelViewMode(mode);
};

/**
 * Update the channel view mode button
 * @param {GraphicsConstants.ChannelViewMode} mode
 */
ChannelsWidget.prototype._updateChannelViewModeBtn = function(mode) {
	// update toggles
	this._splitChannels.toggleOption("composite", mode === GraphicsConstants.ChannelViewMode.COMPOSITE_VIEW);
	this._splitChannels.toggleOption("automatic", mode === GraphicsConstants.ChannelViewMode.AUTOMATIC_TILING);
	this._splitChannels.toggleOption("horizontal", mode === GraphicsConstants.ChannelViewMode.HORIZONTAL_TILING);
	this._splitChannels.toggleOption("vertical", mode === GraphicsConstants.ChannelViewMode.VERTICAL_TILING);

	this._splitChannels.setActive(mode !== GraphicsConstants.ChannelViewMode.COMPOSITE_VIEW);
};

/**
 * Set the single channel mode
 * @param {GraphicsConstants.SingleChannelMode} mode
 */
ChannelsWidget.prototype.setSingleChannelMode = function(mode) {
	this._singleChannelMode.setActive(mode !== GraphicsConstants.SingleChannelMode.IN_CHANNEL_COLOR);

	this._singleChannelMode.toggleOption("inChannelColor", mode === GraphicsConstants.SingleChannelMode.IN_CHANNEL_COLOR);
	this._singleChannelMode.toggleOption("inBlack", mode === GraphicsConstants.SingleChannelMode.IN_BLACK);
	this._singleChannelMode.toggleOption("inBlackWithShowAll", mode === GraphicsConstants.SingleChannelMode.IN_BLACK_WITH_SHOW_ALL_EFFECT);

	this._singleChannelMode.setTooltip(
		"viewer",
		mode !== GraphicsConstants.SingleChannelMode.IN_CHANNEL_COLOR ? "ChanSingleTT2" : "ChanSingleTT"
	);

	this._graphicsModel.setSingleChannelMode(mode);
};

/**
 * Checks if the buttons for moving channels need to be disabled or enabled.
 */
ChannelsWidget.prototype.updateMoveChannelsButtons = function() {
	if (this.isContextAvailable()) {
		if (!this.isOrderingEnabled) {
			this._moveChannelsUp.setEnabled(false);
			this._moveChannelsDown.setEnabled(false);
			this.context.find(".sortableChannelsList").sortable("disable");
		} else {
			if (this._graphicsModel.hasEnabledChannelGaps()) {
				this._moveChannelsUp.setEnabled(true);
				this._moveChannelsDown.setEnabled(true);
			} else {
				var channels = this._graphicsModel.getChannels();
				if (channels.length > 0) {
					this._moveChannelsDown.setEnabled(!_.last(channels).isEnabled());
					this._moveChannelsUp.setEnabled(!_.first(channels).isEnabled());
				} else {
					this._moveChannelsDown.setEnabled(false);
					this._moveChannelsUp.setEnabled(false);
				}
			}
			this.context.find(".sortableChannelsList").sortable("enable");
		}
	}
};

/**
 * Enable or disable ordering
 * @param {Boolean} toggle
 */
ChannelsWidget.prototype.toggleOrdering = function(toggle) {
	if (this.isOrderingEnabled != toggle) {
		this.isOrderingEnabled = toggle;
		this.updateMoveChannelsButtons();
	}
};

/**
 * Updates the revert button
 */
ChannelsWidget.prototype.updateRevertButton = function() {
	if (this._graphicsModel.hasChanges()) {
		this._resetChannelsButton.setEnabled(true);
	} else {
		this._resetChannelsButton.setEnabled(false);
	}
};

/**
 * Update the eye dropper unit selector button.
 * It will create it when needed.
 */
ChannelsWidget.prototype._updateEyeDropperUnitSelectorButton = function() {
	if (this.isContextAvailable()) {
		this.f.emit(
			"toolManager.activeTool",
			_.bind(function(activeTool) {
				if (
					activeTool !== undefined &&
					activeTool.getIID() === CoreConstants.TOOLS.MEASURE_EYEDROPPER &&
					activeTool.isActive() &&
					this.f.getActivePage().isDigitalFilmPage()
				) {
					if (this._eyeDropperUnitSelector === null) {
						this._eyeDropperUnitSelector = new EyeDropperUnitSelector(this.f);
						this.headerExtra.append(this._eyeDropperUnitSelector.getButtonElement());
					} else {
						this._eyeDropperUnitSelector.getButtonElement().show();
					}
				} else {
					if (this._eyeDropperUnitSelector !== null) {
						this._eyeDropperUnitSelector.getButtonElement().hide();
					}
				}
			}, this)
		);
	}
};

/**
 * Update the screen ruling unit selector button.
 * It will create it when needed.
 */
ChannelsWidget.prototype._updateScreenRulingUnitSelectorButton = function() {
	if (this.isContextAvailable()) {
		this.f.emit(
			"toolManager.activeTool",
			_.bind(function(activeTool) {
				if (
					activeTool !== undefined &&
					activeTool.getIID() === CoreConstants.TOOLS.MEASURE_SCREEN_RULING &&
					activeTool.isActive() &&
					this.f.getActivePage().isDigitalFilmPage()
				) {
					if (this._screenRulingUnitSelector === null) {
						this._screenRulingUnitSelector = new ScreenRulingUnitSelector(this.f);
						this.headerExtra.append(this._screenRulingUnitSelector.getButtonElement());
					} else {
						this._screenRulingUnitSelector.getButtonElement().show();
					}
				} else {
					if (this._screenRulingUnitSelector !== null) {
						this._screenRulingUnitSelector.getButtonElement().hide();
					}
				}
			}, this)
		);
	}
};

/**
 * Fill on mobile
 * @returns {Boolean}
 * @override
 */
ChannelsWidget.prototype.willFill = function() {
	return this.f.isMobile();
};

export default ChannelsWidget;
