import _ from "underscore";
import $ from "jquery";
import ViewportLayerModel from "@esko/web-viewer/viewport/model/ViewportLayerModel";
import LineworkWrapper from "@esko/web-viewer/utils/LineworkWrapper";
import DrawUtils from "@esko/web-viewer/utils/DrawUtils";
import ShapeUtils from "@esko/shapeutils";
import LineworkPoint from "@esko/web-viewer/utils/LineworkPoint";
import ListenerUtil from "@esko/web-viewer/utils/ListenerUtil";
import ViewerUtils from "@esko/web-viewer/utils/ViewerUtils";
import CompareConstants from "@esko/web-viewer/compare/CompareConstants";
/**
 * @exports GraphicsLayerModel
 * @memberof Graphics
 * @augments ViewportLayerModel
 * @class This layer model is responsible to draw the bitmaps on the canvas.
 * It also contains an option to display the trimbox or not (by default off).
 * The main functions is "setOverwriteOption" which allows to call the exact same viewer calls, but with one or more options overwritten, like the compare mode
 * @constructor
 * @mixes Utils.ListenerUtilMixin
 *
 * @param {Core.ViewerFacade} f - Viewer facade
 * @param {String} ID - the ID of the layer model
 * @param {GraphicsModel} graphicsModel
 */

var GraphicsLayerModel = function(f, ID, graphicsModel) {
	ViewportLayerModel.call(this, f, ID);
	ListenerUtil.mixin(this, f);

	this._graphicsModel = graphicsModel;

	this._showMargin = false; //shows the margin of the graphics or not

	this.f.whenModuleAvailable(
		"viewport",
		function(viewportModule) {
			this._viewport = viewportModule;
		},
		this
	);

	this._hasGrid = false;

	this._buffer = null;

	/**
	 * Mark all fragments as dirty (debounced version)
	 */
	this.markAllAsDirty = _.throttle(_.bind(this._markAllAsDirty, this), 20);
	this._registerListener("changed.**", this.modelChanged, this, this._graphicsModel);

	this._registerListener("changed.channel.**", this.markAllAsDirty, undefined, this._graphicsModel);
	this._registerListener("changed.channelOrder", this.markAllAsDirty, undefined, this._graphicsModel);
	this._registerListener("changed.singleChannelMode", this.markAllAsDirty, undefined, this._graphicsModel);
	this._registerListener("viewport.viewserverParametersChanged", this.markAllAsDirty);
	this._registerListener("doc.parameters.changed", this.markAllAsDirty);
	this._registerListener("changed.page", this._cleanupFragments, this, this._graphicsModel);
	this._registerListener("needsFragmentCleanup", this._cleanupFragments, this, this._graphicsModel);
};

// inherit
GraphicsLayerModel.prototype = Object.create(ViewportLayerModel.prototype);
// correct the constructor pointer
GraphicsLayerModel.prototype.constructor = GraphicsLayerModel;

/**
 * Remove all listeners
 */
GraphicsLayerModel.prototype.destroy = function() {
	ViewportLayerModel.prototype.destroy.call(this);

	this._unregisterAllListeners();
};

/**
 * This model has graphics!
 * @returns {Boolean} true
 */
GraphicsLayerModel.prototype.hasGraphics = function() {
	return (
		(this._graphicsModel.page && this._graphicsModel.page.hasGraphics()) ||
		(this._graphicsModel.difference && this._graphicsModel.difference.hasGraphics())
	);
};

/**
 * Does the layermodel have a buffer, depends on the panel
 * @param {String} panelName
 * @returns {Boolean}
 */
GraphicsLayerModel.prototype.hasBuffer = function(panelName) {
	if (panelName === "default") {
		return this._buffer !== null;
	} else {
		return false;
	}
};

/**
 * Returns the layers's buffer
 * @returns {null|Fragment}
 */
GraphicsLayerModel.prototype.getBuffer = function() {
	return this._buffer;
};

/**
 * Set the buffer for the layer
 * @param {null|Fragment} buffer
 */
GraphicsLayerModel.prototype.setBuffer = function(buffer) {
	this._buffer = buffer;
};

/**
 * Check if the layer model has CAD, returns true, do the same as for CAD layer if the graphics shows a grid (grid is a linework)
 * @returns {Boolean}
 */
GraphicsLayerModel.prototype.hasLinework = function(panelName) {
	return this._hasGrid && panelName === "default";
};

/**
 * For a graphics layer, the layer should autofetch fragments (this is not a task for the graphicsLayerModel)
 * @override
 * @returns {Boolean} true
 */
GraphicsLayerModel.prototype.shouldAutoFetchFragments = function() {
	return true;
};

/**
 * @override
 * @param {Array.<LineworkWrapper>} lineworks
 * @param {Boolean} isAnimating
 * @param {Boolean} isAnimatingRotation
 * @param {ShapeUtils.Rect} drawrect
 * @param {String} viewName - name of the view to get the grid from
 * @param {String} panelName - the name of the panel this layer is part of
 * @param {Boolean} isZooming - is there zooming active at this moment
 */
GraphicsLayerModel.prototype.getLinework = function(isAnimating, isAnimatingRotation, drawrect, viewName, panelName, isZooming) {
	var lineworks = [];

	if (this._graphicsModel.page !== undefined && !isZooming && drawrect !== undefined) {
		this._getTrimbox(lineworks, drawrect, panelName);

		this._getGridLinework(lineworks, isAnimating, isAnimatingRotation, drawrect, viewName, panelName);

		this._getReaderSpreadLines(lineworks);
	}

	return lineworks;
};

/**
 * Construct the linework to show a grid on the graphics document if zoomed in on the document
 * @param {Array.<LineworkWrapper>} lineworks
 * @param {Boolean} isAnimating
 * @param {Boolean} isAnimatingRotation
 * @param {ShapeUtils.Rect} drawrect
 * @param {String} viewName - name of the view to get the grid from
 */
GraphicsLayerModel.prototype._getGridLinework = function(lineworks, isAnimating, isAnimatingRotation, drawrect, viewName, panelName) {
	//only return linework for the grid if the zoom lever is bigger than 1000 and deep zoom is not supported
	var activePage = this.f.getActivePage(); //get the currently active page from the viewerfacade for the correct resolution of the active page
	if (this._viewport && this._viewport.getWorld().getZoomPercentage() >= 1000 && !activePage.hasDeepZoom()) {
		//make sure the viewport is available (should be as the call for linework comes from the viewport)
		this._hasGrid = true;

		var view = this._viewport.getFirstViewOccurrence(viewName); //get the view for which grid needs te be drawn
		if (panelName === "default" && view) {
			//first check if this view is available, only draw grid on default panel
			var visibleRect = this._viewport.getVisibleRect(); //get the currently visible rectangle (scene-rect)
			//only the grid for the currently active page is shown in the viewport
			var resX = ViewerUtils.parseFloat(activePage.getParam("resolutionx")); //get the resolution for the x and the y axis
			var resY = ViewerUtils.parseFloat(activePage.getParam("resolutiony")); //resolution can be different for the y-axis, interval between grid lines will differ if this is the case
			var zoom = this._viewport.getWorld().getZoom(); //ppmm, get the current zoom of the world, this are screen pixels per mm
			var gridOffsetHeight = 0; //calculate the extra offset to shift the grid as the offset causes the active page to shift part of a document pixel
			var gridOffsetWidth = 0;

			//Calculate all the x-coordinates to draw lines
			var xCoord = Math.ceil(visibleRect.origin.x * resX) / resX + gridOffsetWidth; //mm, first value for the x-coordinate, will be changed to the next coordinate
			var initXDistancePx = Math.round((xCoord - visibleRect.origin.x) * zoom); //px, distance between the visibleRect border and the first grid line in pixels
			var intervalX = 1 / resX; //mmpp, mm per docucument pixel, is the interval between in mm between the grid lines
			var intervalPx = intervalX * zoom; //ppp, screen pixels per document pixel, amount of screen pixels between grid lines
			var endXCoord = visibleRect.origin.x + visibleRect.size.width; //mm, end of the visible rectangle
			var xPoints = []; //make an array for all the points of the grid on the x-axis
			while (xCoord < endXCoord) {
				//run over the width of the visible rect and add a grid line for every interval still inside the visible rect
				xPoints.push(xCoord);
				xCoord += intervalX;
			}
			xPoints.push(visibleRect.origin.x + visibleRect.size.width);

			//Calculate all the y-coordinates to draw grid lines
			var yCoord = Math.ceil(visibleRect.origin.y * resY) / resY + gridOffsetHeight; //mm, first value for the y-coordinate
			var initYDistancePx = Math.round((yCoord - visibleRect.origin.y) * zoom); //px, distance between the visibleRect border and the first grid line in pixels
			var endYCoord = visibleRect.origin.y + visibleRect.size.height; //mm, end of the visible rectangle
			var intervalY = 1 / resY; //mmpp, mm per docucument pixel, is the interval in mm between the grid lines
			var intervalPy = intervalY * zoom; //ppp, screen pixels per document pixel, amount of screen pixels between grid lines
			var yPoints = []; //make an array for all the points of the grid on the y axis
			while (yCoord < endYCoord) {
				// run over all the gridlines that fit in the visible rectangle
				yPoints.push(yCoord);
				yCoord += intervalY;
			}
			yPoints.push(visibleRect.origin.y + visibleRect.size.height);

			//The Graphics canvas is needed to reques the pixel data to calculate the correct RGB values for the grid
			this._graphicsLayer = view.getLayer("Graphics").getLayerElement();

			var rotation = this._viewport.getWorld().getRotation() % 360; //get the current rotation
			var flip = this._viewport.getWorld().isFlip(); //get the flipped values of the world
			var isViewFromBack = this._viewport.getWorld().isFlipViewFromBack(); // is flipped view from back
			if (isViewFromBack) {
				if (rotation === 90 || rotation === 270) {
					rotation = (rotation + 180) % 360;
				}
				flip = !flip;
			}

			//to be sure always the correct value for the color is selected, move to half-pixels
			var xhalfPx = Math.ceil(intervalPx / 2);
			var yhalfPx = Math.ceil(intervalPy / 2);

			var width = $(this._graphicsLayer).innerWidth() || null; //we'll use the width and height of the canvas, gives the amount of screen pixels for the width
			var height = $(this._graphicsLayer).innerHeight() || null;

			var worldBorderWidth = 0;
			if (visibleRect.origin.x === 0) {
				worldBorderWidth = width - Math.ceil(visibleRect.size.width * zoom);
			}
			var worldBorderHeight = 0;
			if (visibleRect.origin.y === 0) {
				worldBorderHeight = height - Math.ceil(visibleRect.size.height * zoom);
			}

			//only if the canvas has a width and height go further
			if (width !== 0 && height !== 0) {
				//request the RGB values for all the visible screen pixels, for each pixel 4 values are returned (red, green, blue and alpha)
				var imageColor = $(this._graphicsLayer)[0]
					.getContext("2d")
					.getImageData(0, 0, width, height); //return a 1D array with all the values row by row
				var color = imageColor.data; //make a copy of the array of pixels
				_.each(
					xPoints,
					function(xPoint, indexX) {
						//run over all the points for the grid to construct the grid, every pass construct two lines to the side of the pixel
						_.each(
							yPoints,
							function(yPoint, indexY) {
								//depending on rotation/flip the pixel to use for lines at the border of the viewport differs (try to keep it in the viewport)
								//the normal offset, yhalfPx/xhalfPx (to get to the center of the document pixel) is reverted or added twice, depending on the sign
								var ySign = 1;
								var xSign = 1;

								var x1 = xPoint; //current x-point
								var x0 = 0;
								if (indexX == 0) {
									//if it's the first point for the x-coordinate, take the origin of the visible rect as second x-coordinate
									x0 = visibleRect.origin.x;
								} else {
									//if not, take the previous point as second x-coordinate
									x0 = xPoints[indexX - 1];
								}

								//same as for the x-coordinate, construct the needed y-coordinates
								var y1 = yPoint;
								var y0 = 0;
								if (indexY == 0) {
									y0 = visibleRect.origin.y;
								} else {
									y0 = yPoints[indexY - 1];
								}

								//First correct values need to be calculated due to rotation and flipped state of the viewport. Visible rect rotates, the origin
								//changes when the world is rotated. The width and height stay the same and the image pixel data stays with its origin at the left-top corner.
								var indexCoordX = indexX; //make sure to not change the index, this will be changed when the world is rotated/flipped
								var indexCoordY = indexY; //number of the row
								var xOffset = initXDistancePx; //need to add an offset to get the correct rgb value from the pixeldata-object, use the width betweeen the first grid line and the visible rectangle border. (changed when rotated/flipped)
								var yOffset = initYDistancePx;
								if (rotation === 90 || rotation === 180) {
									//if the rotation is 90 or 180 degrees, the origin of the visible rectangle will be on the right not on the left
									indexCoordY = yPoints.length - indexY - 1; //the index to calculate the inverse RGB-value needs to be moved to the other side of the viewport
									if (rotation === 90) {
										//calculate the correct offset, as with the index, we need the offset from the left side, not the right side
										yOffset = (width - initYDistancePx) % Math.round(intervalPy); //width is the yOffset as the coordinate system is rotated (the pixel data is not), %interval to get the pixel between the grid line and the visible rect border
									} else {
										//180 degrees
										yOffset = (height - initYDistancePx) % Math.round(intervalPy); //for 180 also the y-offset needs to be from the top and not the bottom of the viewport
									}
									if (indexY === 0) {
										//a line at the border of the viewport
										ySign = -1; //add yhalfPx twice to get to a pixel that's visible
									}
								} else if (_.last(yPoints) == yPoint) {
									//a line at the border of the viewport
									ySign = -1; //add yhalfPx twice to get to a pixel that's visible
								}

								//do the same for the x-coordinates, it's needed to take the flipped-state into account
								if (((rotation === 180 || rotation === 270) && !flip) || ((rotation === 90 || rotation === 0) && flip)) {
									indexCoordX = xPoints.length - indexX - 1; //x-coordinate lays on the wrong side of the visible rectangle, move it to the correct

									if (rotation === 90 || rotation === 270) {
										//offset needs to be calculated as 90+flipped or 270 are wrong (and since it's rotated height is used)
										xOffset = (height - initXDistancePx) % Math.round(intervalPx);
									} else {
										//0°+flipped or 180° not flipped
										xOffset = (width - initXDistancePx) % Math.round(intervalPx);
									}
									if (indexX === 0) {
										//a line at the border of the viewport
										xSign = -1;
									}
								} else if (_.last(xPoints) == xPoint) {
									//a line at the border of the viewport
									xSign = -1;
								}

								//calculate the index in the 1D array for the color of the grid lines, rows are placed behind eachother in the array
								//For all Y-coordinates multiply with the width of the viewport to get to the correct 'row'. Everything is multiplied by 4
								//because every pixel has 4 values (RGB+Alpha).
								var colorIndex = 0;
								if (rotation == 90 || rotation == 270) {
									//for 90° and 270°, all the x- and y-coordinates are switched
									colorIndex =
										width * Math.round(indexCoordX * intervalPx) +
										xOffset * width +
										Math.round(intervalPy * indexCoordY) +
										yOffset -
										xhalfPx * width -
										yhalfPx;
									//if the color is for a line at the border of the viewport, compensate colorIndex to get a value of the correct (visible) pixel
									if (indexX === 0 || indexY === 0 || _.last(xPoints) == xPoint || _.last(yPoints) == yPoint) {
										colorIndex += xSign * (xhalfPx - 2) * width + ySign * (yhalfPx - 2); //either add the xhalfPx/yhalfPx twice or subtract it with a difference of 2px to avoid borders of document pixels
									}
								} else {
									colorIndex =
										width * Math.round(indexCoordY * intervalPy) + //multiply the number of the row with the number of pixels per document pixel and the width => y-coordinate
										yOffset * width + //multiply the initial distance between the first grid-line and visible rect with the width of the viewport => y-coordinate
										worldBorderHeight * width +
										worldBorderWidth +
										Math.round(intervalPx * indexCoordX) + //multiply the number of document pixel width the number of pixels per document pixel
										xOffset - //add the distance between the most left grid-line and the left of the visible rectangel
										yhalfPx * width -
										xhalfPx;
									//if the color is for a line at the border of the viewport, compensate colorIndex to get a value of the correct (visible) pixel
									if (indexX === 0 || indexY === 0 || _.last(xPoints) == xPoint || _.last(yPoints) == yPoint) {
										colorIndex += ySign * (yhalfPx - 2) * width + xSign * (xhalfPx - 2);
									}
								}

								colorIndex *= 4; //multiply with 4 as every pixel has 4 values
								//get the RGB values for the pixel from the pixel data array and invert them to get the values for the grid-lines
								var r = (255 - color[colorIndex + 0]).toString(16);
								var g = (255 - color[colorIndex + 1]).toString(16);
								var b = (255 - color[colorIndex + 2]).toString(16);

								if (r.length === 1) {
									r = "0" + r;
								}
								if (g.length === 1) {
									g = "0" + g;
								}
								if (b.length === 1) {
									b = "0" + b;
								}

								//Construct the two lines to draw arround the pixel
								var linework = new LineworkWrapper(); //use a lineworkwrapper to push in the linework array
								linework.type = DrawUtils.LWTYPE_LINE;
								linework.linewidth = 1;
								linework.color = "#" + r + g + b; //use the iverted RGB-values
								linework.pointArray[0] = new LineworkPoint();
								linework.pointArray[0].makeMoveTo(x0, y1); //make an horizontal line
								linework.pointArray[1] = new LineworkPoint();
								linework.pointArray[1].makeLineTo(x1, y1);
								lineworks.push(linework); //push the linework in the linework array
								linework = new LineworkWrapper();
								linework.type = DrawUtils.LWTYPE_LINE;
								linework.linewidth = 1;
								linework.color = "#" + r + g + b;
								linework.pointArray[0] = new LineworkPoint();
								linework.pointArray[0].makeMoveTo(x1, y0); //make a vertical line
								linework.pointArray[1] = new LineworkPoint();
								linework.pointArray[1].makeLineTo(x1, y1);
								lineworks.push(linework);
							},
							this
						);
					},
					this
				);
			}
		}
	} else {
		this._hasGrid = false;
	}
};

/**
 * Get the linework for the small lines for multipage/page list documents when in reader spread mode
 * @param {Array.<LineworkWrapper>} lineworks
 */
GraphicsLayerModel.prototype._getReaderSpreadLines = function(lineworks) {
	//if the document is a multi page document, and reader spread is active show 2 small lines indicating border of the active document
	if (this._graphicsModel.page.isReaderSpread()) {
		//create two lineworkwrappers for the top and bottom
		var lineworkTop = new LineworkWrapper();
		lineworkTop.type = DrawUtils.LWTYPE_LINE;
		lineworkTop.linewidth = 1;
		lineworkTop.color = "#ababab";
		var lineworkBottom = new LineworkWrapper();
		lineworkBottom.type = DrawUtils.LWTYPE_LINE;
		lineworkBottom.linewidth = 1;
		lineworkBottom.color = "#ababab";

		var page = this._graphicsModel.getPage();
		var pageSize = page.getPageSize();
		var referencePage = this._graphicsModel.getReferencePage();
		var lineLocationX = 0;
		if (page && page.get("isRightPage")) {
			if (referencePage) {
				if (referencePage.isBlankPage()) {
					lineLocationX = referencePage.getTrimboxSize().width;
					if (page.isBlankPage()) {
						lineLocationX += referencePage.getMargins().left;
					}
				} else {
					lineLocationX = referencePage.getPageSize().width - referencePage.getMargins().right;
				}
			} else {
				if (page.isBlankPage()) {
					lineLocationX = pageSize.width + page.getMargins().left;
				} else {
					lineLocationX = page.getTrimboxRect().origin.x - page.getMargins().left;
				}
			}
		} else {
			if (page.isBlankPage() && referencePage) {
				//page is blank and reference is not empty
				lineLocationX = page.getTrimboxSize().width;
				if (referencePage.isBlankPage()) {
					lineLocationX += page.getMargins().left;
				}
			} else {
				lineLocationX = pageSize.width - page.getMargins().right;
			}
		}

		lineworkTop.pointArray[0] = new LineworkPoint();
		lineworkTop.pointArray[0].makeMoveTo(lineLocationX, -(pageSize.height * 0.1));
		lineworkTop.pointArray[1] = new LineworkPoint();
		//don't draw against the document, leave some space (use 0.01 of the document height)
		lineworkTop.pointArray[1].makeLineTo(lineLocationX, -(pageSize.height * 0.01));

		lineworkBottom.pointArray[0] = new LineworkPoint();
		lineworkBottom.pointArray[0].makeMoveTo(lineLocationX, pageSize.height + pageSize.height * 0.1);
		lineworkBottom.pointArray[1] = new LineworkPoint();
		lineworkBottom.pointArray[1].makeLineTo(lineLocationX, pageSize.height + pageSize.height * 0.01);

		lineworks.push(lineworkTop);
		lineworks.push(lineworkBottom);
	}
};

/**
 * Get the linework for the trimbox
 * @param {Array.<LineworkWrapper>} lineworks
 */
GraphicsLayerModel.prototype._getTrimbox = function(lineworks, drawRect, panelName) {
	if (this._graphicsModel.isShowTrimbox()) {
		var trimbox = new LineworkWrapper();
		var activePageTrimbox = this._graphicsModel.getTrimBox();

		var isInSideBySide =
			this._graphicsModel.dmode === CompareConstants.diffMode.SHOWBOTHSIDEBYSIDE ||
			this._graphicsModel.dmode === CompareConstants.diffMode.SHOWBOTHSIDEBYSIDEHIGHLIGHT ||
			this._graphicsModel.dmode === CompareConstants.diffMode.SHOWBOTHSIDEBYSIDECHANGESONLY;

		if (
			activePageTrimbox !== null &&
			!activePageTrimbox.containsRect(drawRect) &&
			this._graphicsModel.dmode !== CompareConstants.diffMode.DIFFVERSION &&
			((isInSideBySide && (panelName === "default" || panelName === "compare_align_result" || panelName === "compare_differenceMode")) ||
				!isInSideBySide)
		) {
			trimbox.type = DrawUtils.LWTYPE_DASHEDBOX;
			trimbox.dash = [10, 3, 2, 3, 2, 3];
			trimbox.linewidth = 1;
			trimbox.color = this.f.getCurrentUser().getColor("PREFERENCES_HIGHLIGHTCOLORS_MARGINS");
			trimbox.fillcolor = this.f.getCurrentUser().getColor("PREFERENCES_HIGHLIGHTCOLORS_MARGINS");
			trimbox.rect = activePageTrimbox;
			lineworks.push(trimbox);
		}

		var referenceTrimBox = this._graphicsModel.getReferenceTrimBox();
		if (
			referenceTrimBox !== null &&
			!referenceTrimBox.containsRect(drawRect) &&
			this._graphicsModel.dmode !== CompareConstants.diffMode.CURRENTVERSION &&
			((isInSideBySide &&
				(panelName === "compare_difference" || panelName === "compare_align_result" || panelName === "compare_differenceMode")) ||
				!isInSideBySide)
		) {
			trimbox = new LineworkWrapper();
			trimbox.type = DrawUtils.LWTYPE_DASHEDBOX;
			trimbox.dash = [10, 3, 2, 3, 2, 3];
			trimbox.linewidth = 1;
			trimbox.color = this.f.getCurrentUser().getColor("PREFERENCES_HIGHLIGHTCOLORS_MARGINS_DIFF");
			trimbox.fillcolor = this.f.getCurrentUser().getColor("PREFERENCES_HIGHLIGHTCOLORS_MARGINS_DIFF");
			trimbox.rect = referenceTrimBox;
			lineworks.push(trimbox);
		}
	}
};

/**
 * @override
 */
GraphicsLayerModel.prototype.getMaximumResolution = function() {
	return this._graphicsModel.getMaximumResolution();
};

/**
 * @override
 */
GraphicsLayerModel.prototype.getBackgroundOptions = function() {
	var options = {};

	if (this._graphicsModel.page && this._graphicsModel.page.isBlankPage()) {
		options.fillStyle = "rgba(255,255,255,1)";
	}

	return options;
};

export default GraphicsLayerModel;
