import GraphicsConstants from "@esko/web-viewer/graphics/GraphicsConstants";
import GraphicsLayerModel from "@esko/web-viewer/graphics/GraphicsLayerModel";
import ViewportConstants from "@esko/web-viewer/viewport/ViewportConstants";
import ListenerUtil from "@esko/web-viewer/utils/ListenerUtil";
import _ from "underscore";

/**
 * @exports GraphicsViewportWidgetController
 * @memberof Graphics
 * @class This class controls the viewport widget model for graphics.
 * It creates for example the views for channels, etc.
 * @constructor
 * @mixes Util.ListenerUtilMixin
 * @param {Core.ViewerFacade} f
 * @param {GraphicsModel} graphicsModel - Model for graphics
 */
var GraphicsViewportWidgetController = function(f, graphicsModel) {
	this.f = f;
	ListenerUtil.mixin(this, f);
	this._graphicsModel = null;
	this.viewport = this.f.getModule("viewport");
	this._createdViewLayerModels = {};
	this._shownViews = [];

	this.setGraphicsModel(graphicsModel);

	//register listeners for the change of the trimbox colors in the settings, needs to update the viewport (trigger a rerender)
	this._registerListener(
		"options.changed.PREFERENCES_HIGHLIGHTCOLORS_MARGINS_DIFF",
		function(oldValue, newValue) {
			if (oldValue !== newValue) {
				//only if the color changed repaint the viewport
				this._repaintViewport();
			}
		},
		this,
		this.f.getCurrentUser()
	);

	this._registerListener(
		"options.changed.PREFERENCES_HIGHLIGHTCOLORS_MARGINS",
		function(oldValue, newValue) {
			if (oldValue !== newValue) {
				this._repaintViewport();
			}
		},
		this,
		this.f.getCurrentUser()
	);
};

/**
 * Change the graphics model
 * @param {GraphicsModule.GraphicsModel} graphicsModel
 */
GraphicsViewportWidgetController.prototype.setGraphicsModel = function(graphicsModel) {
	this._unregisterAllListeners();

	this._graphicsModel = graphicsModel;

	this._registerListener("changed.channels", this.updateTiles, this, this._graphicsModel);
	this._registerListener("changed.channelViewMode", this.updateTiles, this, this._graphicsModel);
	this._registerListener("changed.channel.*.enabled", this.updateTiles, this, this._graphicsModel);
	this._registerListener("changed.channelOrder", this.updateTiles, this, this._graphicsModel);
	this._registerListener("changed.trimbox", this._repaintViewport, this, this._graphicsModel);
	this._registerListener(
		"changed.page",
		function() {
			this.destructViews();
			this.updateTiles();
		},
		this,
		this._graphicsModel
	);
};

/**
 * get/generate the layermodel for a certain channel
 * If the layer does not exist yet, create a new one
 * @param {Graphics.Channel} channel - the channel for which a layermodel/layer needs to be created
 * @returns {Graphics.GraphicsLayerModel}
 */
GraphicsViewportWidgetController.prototype.getLayerModelForChannel = function(channel) {
	var layerModel = this._createdViewLayerModels[channel.getID()];

	if (!layerModel) {
		// NOTE: in the future we might need to limit the amount of existing layermodels...
		// create the layermodel and view
		this.viewport.createView(channel.getID(), channel.getName(), false);
		layerModel = new GraphicsLayerModel(this.f, "Graphics", this._graphicsModel);
		var channelSelection = this._graphicsModel.getSpecificChannelSelection(channel);
		layerModel.setOverwriteOption("color", channelSelection);
		this.viewport.addLayer("Graphics", layerModel, channel.getID());
		this._createdViewLayerModels[channel.getID()] = layerModel;
	}

	return layerModel;
};

/**
 * Called when the channel view mode has changed
 */
GraphicsViewportWidgetController.prototype.updateTiles = function() {
	var isTiledViewEnabled = this.isTiledViewEnabled();

	// disable tiled mode is there is only one extra view available
	if (isTiledViewEnabled) {
		if (this._graphicsModel.getChannelViewMode() !== GraphicsConstants.ChannelViewMode.COMPOSITE_VIEW) {
			if (this._graphicsModel.getEnabledChannelCount() <= 1) {
				isTiledViewEnabled = false;
			}
		}
	}

	// in here all the views that will be hidden are stored
	// by default mark all shown views to be hidden
	var viewsToHide = this._shownViews;

	// in here all the channels that need to be shown are stored
	var channelsToShow = [];

	if (isTiledViewEnabled) {
		// only run over all visible channels if the tiled mode is enabled
		_.each(
			this._graphicsModel.getChannels(),
			function(channel) {
				if (channel.isEnabled()) {
					if (this._graphicsModel.isInDifferenceCompareMode()) {
						if (channel.isInPreviousAndMatchedWithCurrent()) {
							// if viewer is in difference compare mode and the channel is from the difference file and matched with a channel from the current file, just hide id
							return;
						}
					}

					// the channel is visible. and it should be shown, remove it from the viewsToHide array and push in on the "show" array
					viewsToHide = _.without(viewsToHide, channel.getID());
					channelsToShow.push(channel);
				}
			},
			this
		);
	}

	// hide all views that need to be hidden
	_.each(
		viewsToHide,
		function(view) {
			this.viewport.toggleView(view, false);
		},
		this
	);

	// rebuild the _shownViews variable
	this._shownViews = [];
	_.each(
		channelsToShow,
		function(channel) {
			// get/generate the layer
			var layerModel = this.getLayerModelForChannel(channel);

			// get the color selection mode
			var channels = [channel];

			if (this._graphicsModel.isInDifferenceCompareMode() && channel.isInCurrentAndMatchedToPrevious()) {
				// this channel was matched onto a channel of the difference file, add matched channel
				var matchedChannel = this._graphicsModel.getMatchedChannel(channel);
				if (matchedChannel) {
					channels.push(matchedChannel);
				}
			}

			// overwrite the channel selection for this view
			layerModel.setOverwriteOption("color", this._graphicsModel.getChannelSelection(channels));

			this._shownViews.push(channel.getID());
			this.viewport.toggleView(channel.getID(), true);
		},
		this
	);

	if (isTiledViewEnabled) {
		// generate the needed order and push it to the viewport widget
		var order = [];
		_.each(channelsToShow, function(channel) {
			order.push(channel.getID());
		});
		// set the order
		this.viewport.setViewsOrder(order);
	}

	// disable animation if in tiled mode
	this.f.emit("viewport.toggleAnimation", "tiledMode", !isTiledViewEnabled);
	// to make it more readable in the GraphicsConstants the Channel View Mode 'names' are mapped to the Layout Mode 'names' of ViewportConstants
	this.viewport.setViewsLayoutMode(
		ViewportConstants.ViewsLayoutMode[
			GraphicsConstants.ChannelViewModeLayoutMode[_.invert(GraphicsConstants.ChannelViewMode)[this._graphicsModel.getChannelViewMode()]]
		]
	);
};

/**
 * is the tiled mode active
 * returns {Boolean}
 */
GraphicsViewportWidgetController.prototype.isTiledViewEnabled = function() {
	return this._graphicsModel.getChannelViewMode() !== GraphicsConstants.ChannelViewMode.COMPOSITE_VIEW;
};

/**
 * Trigger a repaint of the viewport
 */
GraphicsViewportWidgetController.prototype._repaintViewport = function() {
	this.f.emit("viewport.repaint");
};

/**
 * Destruct all views that are created
 */
GraphicsViewportWidgetController.prototype.destructViews = function() {
	_.each(
		this._createdViewLayerModels,
		function(value, key) {
			this.viewport.removeView(key);
			value.destroy();
		},
		this
	);
	this._createdViewLayerModels = {};
};

export default GraphicsViewportWidgetController;
