import Extendable from "@esko/web-viewer/core/Extendable";
import PrivateChannel from "@esko/web-viewer/core/PrivateChannel";
import $ from "jquery";
import ShapeUtils from "@esko/shapeutils";
import Mousetrap from "@esko/mousetrap";
import _ from "underscore";
/**
 * A widget is an interface element. It get's injected with its context (a
 * div in which it should display itself)
 *
 * @exports AbstractWidget
 * @memberof Layout
 * @mixes Core.Extendable
 * @mixes Core.PrivateChannel
 * @constructor
 *
 * @param {Core.ViewerFacade} f
 * @param {string} ID - the unique name for the widget
 */
var AbstractWidget = function(f, ID) {
	Extendable.installTo(this);
	PrivateChannel.installTo(this, f, ["widget", ID]);

	this.collapsed = false;

	this.visible = true;

	this._isModal = false;

	this.f = f;

	this._enabled = true;

	/**
	 * The title of the widget
	 * @access private
	 */
	this._title = ID;

	/**
	 * The icon of the widget
	 * @access private
	 */
	this._icon = "icon icon-WCR_Cont-Circle";

	/**
	 * The cursor that should be displayed when hovering over the widget
	 * @access private
	 */
	this._currentCursor = "cursor-default";

	/**
	 * In _options all user preferences are stored. This state could be
	 * persisted by a LayoutManager
	 *
	 * All changes to the _options should be done with AbstractWidget.setOption so the
	 * correct events are sent out
	 *
	 * @access private
	 */
	this._options = {};

	/**
	 * Get the unique name of the widget.
	 *
	 * This is a unique identifier for the widget, the widget sends events
	 * on widget.[ID].* channels
	 */
	this.getID = function() {
		return ID;
	};

	/**
	 * Change an option of the widget
	 *
	 * If an option is changed, an event is send on
	 * <widget.[ID].options.changed>
	 *
	 * @param {String} key
	 * @param {Object} newValue
	 */
	this.setOption = function(key, newValue) {
		var oldValue = this._options[key];
		if (oldValue !== newValue) {
			this._options[key] = newValue;
			this._emit("options.changed." + key, oldValue, newValue);
		}
	};

	/**
	 * Get an option of the widget based on the given key.
	 *
	 * @param {String} key - Key of the option
	 * @returns {Object} Option of the widget for the given key or undefined if not found
	 */
	this.getOption = function(key) {
		return this._options[key];
	};

	/**
	 * Get the options of the widget. This method can be used to save the
	 * state of a widget
	 *
	 * @returns {Object} object with all options of the Widget
	 */
	this.getOptions = function() {
		return _.clone(this._options);
	};
};

/**
 * Change multiple options of the widget. This can be used to inject the
 * state of the widget
 *
 * @param {Object} options - object with all options to be set (i.e. {key1: value1, key2: value2, ...}
 */
AbstractWidget.prototype.setOptions = function(options) {
	_.each(
		options,
		function(value, key) {
			this.setOption(key, value);
		},
		this
	);
};

/**
 * Get the title of the widget
 * @returns {String} current title of the widget
 */
AbstractWidget.prototype.getTitle = function() {
	return this._title;
};

/**
 * Set the title of the widget. Listeners get notified automatically
 * @param {String} title - the new title for the widget
 */
AbstractWidget.prototype.setTitle = function(title) {
	if (this._title !== title) {
		this._title = title;
		this._emit("titlechanged", title);
	}
};

/**
 * Get the icon of the widget
 * @returns {String} icon string
 */
AbstractWidget.prototype.getIcon = function() {
	return this._icon;
};

/**
 * Change the icon of the widget
 * @param {String} icon
 */
AbstractWidget.prototype.setIcon = function(icon) {
	if (this._icon !== icon) {
		this._icon = icon;
		this._emit("iconchanged", icon);
	}
};

/**
 * Set te context of the widget. The widget gets injected with a DOM element in which it should draw itself
 *
 * @params {JQueryObj} context - the element in which the widget should draw itself
 * @params {JQueryObj} headerExtra - a div in which extra icons for the header of the widget can be but (preferably MiniButtons)
 */
AbstractWidget.prototype.setContext = function(context, headerExtra) {
	if (!context) {
		this.f.emit("error.WARNING", "Can't call setContext without context!");
		return;
	}

	// set the context of the widget
	this.context = context;
	this.headerExtra = headerExtra;

	this.setCursor(this._currentCursor);
	this._emit("contextAvailable");
};

/**
 * The widget is destroyed
 */
AbstractWidget.prototype.destroy = function() {
	this._emit("destroyed");
};

/**
 * Check if the context of the widget has been set
 * @returns {Boolean} true if the context is set
 */
AbstractWidget.prototype.isContextAvailable = function() {
	return this.context !== undefined;
};

/**
 * Execute some code as soon as the context is available. If the context is already available execute the code immediately
 * @param {Function} callback - callback function to execute when the context of the widget is available
 * @param {Object} [binding] - bind the callback to "binding"
 */
AbstractWidget.prototype.whenContextAvailable = function(callback, binding) {
	if (binding) {
		callback = _.bind(callback, binding);
	}

	if (this.isContextAvailable()) {
		callback();
	} else {
		this.once("contextAvailable", callback);
	}
};

/**
 * Make the widget aware that it is visible or not
 * !! NOTE !! should only be called by Layout, not by any other module/component
 * to make a widget visible use the setEnabled property
 * @param {Boolean} isVisible - True if visible, false otherwise
 */
AbstractWidget.prototype.setVisible = function(visible) {
	if (this.visible !== visible) {
		this.visible = visible;
		this._emit("visible", visible);
	}
};

/**
 * Check if the widget is visible or not
 * @returns {Boolean} True if the widget is visible, false otherwse
 */
AbstractWidget.prototype.isVisible = function() {
	return this.visible;
};

/**
 * Notify the widget that is collapsed
 * @param {Boolean} collapsed
 */
AbstractWidget.prototype.setCollapsed = function(collapsed) {
	if (this.collapsed !== collapsed) {
		this.collapsed = collapsed;
		this._emit("collapsed", collapsed);
	}
};

/**
 * is the widget collapsed
 * @returns {Boolean}
 */
AbstractWidget.prototype.isCollapsed = function() {
	return this.collapsed;
};

/**
 * Notify the widget that it has become modal/not-modal
 * @param {Boolean} isModal
 */
AbstractWidget.prototype.setModal = function(isModal) {
	this._isModal = isModal;
};

/**
 * This method notifies the widget that it has been resized
 */
AbstractWidget.prototype.resized = function() {};

/**
 * Call this when the contents of the widget have changed, and the layout manager should try to adapt the layout to the new size
 * @protected
 */
AbstractWidget.prototype._fireResized = function() {
	this._emit("resized");
};

/**
 * Fire an event when the modal widget wrapper has been resized, used to persist this in a user preference
 */
AbstractWidget.prototype.fireModalResize = function() {
	this._emit("modalResize");
};

/**
 * Call this when the widget needs to react to a resizing event (from widgetwrapper)
 */
AbstractWidget.prototype.resizing = function() {
	//implement in specific widget
};

/**
 * Change the cursor of the widget
 *
 * @param {String} cursor - the class for the cursor that should be displayed
 */
AbstractWidget.prototype.setCursor = function(cursor) {
	this.whenContextAvailable(
		_.bind(function() {
			if (cursor != this._currentCursor) {
				this.context.removeClass(this._currentCursor);
				this.context.addClass(cursor);
				this._currentCursor = cursor;
			}
		}, this)
	);
};

/**
 * request the type of the widget
 *
 * @returns {"dockable"|"topbar"|"maincontent" | "messagebar" } widgettype
 */
AbstractWidget.prototype.getWidgetType = function() {
	return "dockable";
};

/**
 * Get the minimum size of the widget content
 *
 * note it is possible that this size is not respected by the layoutmanager
 *
 * @returns {ShapeUtils.Size}
 */
AbstractWidget.prototype.getMinimumSize = function() {
	return new ShapeUtils.Size(100, 30);
};

/**
 * Get the preferred size of the widget content
 *
 * note it is possible that this size is not respected by the layoutmanager
 *
 * @returns {ShapeUtils.Size} if not overridden the widget will determine its size dynamically
 */
AbstractWidget.prototype.getPreferredSize = function() {
	var clone = this.context.clone(false);
	clone.css({
		position: "absolute",
		visibility: "hidden",
		overflow: "auto",
		height: "auto",
		width: (this.context.width() || null) - 2
	});
	this.f.getContext().append(clone);
	clone.css({
		overflow: "hidden"
	});
	clone.css({
		overflow: "auto"
	});
	var size = new ShapeUtils.Size(clone.outerWidth(true) || null, clone.outerHeight(true) || null);
	clone.remove();
	return size;
};

/**
 * This method returns true if the widget can handle extra space above the preferred size
 * Credit to BAMT for this great method name!
 * @returns {Boolean} true if the widget is expandable and will make use of the extra gained space
 */
AbstractWidget.prototype.willFill = function() {
	return false;
};

/**
 * Get the maximum dimension of the widget content
 *
 * @returns {ShapeUtils.Size} if undefined the widget can be made as big as the user wants
 */
AbstractWidget.prototype.getMaximumSize = function() {
	return undefined;
};

/**
 * Called when moving in the DOM of the widget started
 */
AbstractWidget.prototype.movingStarted = function() {
	this._emit("movingStarted");
};

/**
 * Called when moving in the DOM of the widget stopped
 */
AbstractWidget.prototype.movingStopped = function() {
	this._emit("movingStopped");
};

/**
 * Called when widget is dragged to the specified top and left
 */
AbstractWidget.prototype.dragged = function(top, left) {
	this._emit("dragged", top, left);
};

/**
 * Binds event of a key to a function (can be keypress, keyup or keydown)
 * @param {Array.<Array>} array - 2D array where each record on the first level has three records on the second: the first is an array of the keys
 * (or combination of keys) to listen to, the second is the function that is called when the event happens and the third is optional
 * and can be the the actual event to listen to (keyup, keydown or keypress) by default this will be automatically use the most correct one
 */
AbstractWidget.prototype.bindKeys = function(array) {
	for (var i = 0; i < array.length; i++) {
		Mousetrap.bind(array[i][0], array[i][1], array[i][2]);
	}
};

/**
 * unbinds the events of keys
 * @param {Array.<Array>} array
 */
AbstractWidget.prototype.unbindKeys = function(array) {
	for (var i = 0; i < array.length; i++) {
		Mousetrap.unbind(array[i][0], array[i][1], array[i][2]);
	}
};

/**
 * Enable or disable the widget. If a widget is disabled it won't be visible in the layout
 * @param {Boolean} enabled
 */
AbstractWidget.prototype.setEnabled = function(enabled) {
	if (this._enabled !== enabled) {
		this._enabled = enabled;
		this._emit("enabled", enabled);
	}
};

/**
 * Is the widget enabled?
 * @returns {Boolean}
 */
AbstractWidget.prototype.isEnabled = function() {
	return this._enabled;
};

/**
 * If the widget is collapsed or disabled, it will be uncollapsed and enabled.
 * the widget emits its intention to be visible, the layout manager should make sure the widget is visible
 * (even if the widget is in a collapsed side panel or such)
 */
AbstractWidget.prototype.makeVisible = function() {
	if (!this.isEnabled()) {
		this.setEnabled(true);
	}

	this._didRequestMakeVisible = true;

	this._emit("makeVisible");
};

/**
 * Did the widget made a request to become visible that is not yet executed/accepted?
 * @returns {Boolean} true if the widget requested to become visible, but the request was not accepted
 */
AbstractWidget.prototype.didRequestMakeVisible = function() {
	return this._didRequestMakeVisible;
};

/**
 * Set the "makeVisible" request as accepted
 */
AbstractWidget.prototype.setMakeVisibleRequestAccepted = function() {
	this._didRequestMakeVisible = false;
};

export default AbstractWidget;
