import $ from "jquery";
import _ from "underscore";
import "jquery-ui/ui/widgets/draggable";
import "jquery-ui/ui/widgets/droppable";
/**
 * HELPER interface for draggable widgetwrappers
 * Provides structured storage for extra information needed for dragging
 */
function DraggableWidgetWrapper(widgetWrapper) {
	this.wrapper = widgetWrapper;
	this.ghost = $("<div/>").addClass("ghostWrapper");

	this.sizeBeforeDrag = {
		width: undefined,
		height: undefined
	};
}

/**
 * This method updates the dragging ghost element dimensions.
 * The ghost element is transparent, but has the same size as the original element being dragged
 */
DraggableWidgetWrapper.prototype.updateGhost = function() {
	this.ghost.css({
		width: this.wrapper.element.outerWidth(true) || null,
		height: this.wrapper.element.outerHeight(true) || null
	});
};

/**
 * Enable draggable/droppable/sortable widgets on a container
 *
 * When initialized widgets should be added (dragging enabled by enableDragging(...)) and dropzones should be defined
 */
var DraggableWidgets = function(container) {
	this.container = container;
	this._acceptSelector = ".widgetWrapper";
	this._siblingSelector = ".widgetWrapper";
	this.hoveredSortableDropzone = undefined;
};

/**
 * Enable a WidgetWrapper to be dragged
 * @param widgetWrapper
 */
DraggableWidgets.prototype.enableDragging = function(widgetWrapper) {
	//make the widgetwrapper draggable
	widgetWrapper.drag = new DraggableWidgetWrapper(widgetWrapper);

	$(widgetWrapper.element).draggable({
		containment: this.container,
		// only when the item is dragged more than 20px, we start dragging
		distance: 20,
		// dragging widgets doesn't introdicue scrolling of the page
		scroll: false,
		// when a dropzone is invalid, the widget is reverted to its orignal lcation
		revert: "invalid",
		// when dragging starts, create a semi transparent clone of the widget
		helper: function() {
			var clone = widgetWrapper.element.clone();
			clone.css({
				width: widgetWrapper.element.innerWidth() || null,
				opacity: 0.9
			});
			return clone;
		},
		// elements should be added to the container
		appendTo: this.container,
		// start dragging
		start: _.bind(function() {
			//update the ghost size and store the size before dragging or later usage
			this.drag.updateGhost();
			this.drag.sizeBeforeDrag.width = this.element.innerWidth() || null;
			this.drag.sizeBeforeDrag.height = this.element.innerHeight() || null;

			// hide the original element
			this.element.hide();
		}, widgetWrapper),

		// the element is being dragged
		drag: _.bind(function(event, ui) {
			// if a sortable zone is being hovered, insert the ghost at the correct location (this make the dropzone sortable!)
			if (this.hoveredSortableDropzone) {
				this.insertAt(this.hoveredSortableDropzone, widgetWrapper.drag.ghost, ui.offset.top - this.hoveredSortableDropzone.offset().top);
			}
		}, this),
		// dragging has stopped, remove the helper element, and detac the ghost, display the original element again
		stop: _.bind(function(event, ui) {
			this.drag.ghost.detach();
			this.element.show();

			this.widget.dragged(ui.offset.top, ui.offset.left);

			ui.helper.detach();
		}, widgetWrapper),

		// dragging is being done using the header element as handle
		handle: ".header"
	});
};

/**]
 * Define a sortable widget dropzone
 * @param panel
 */
DraggableWidgets.prototype.addSortableWidgetDropzone = function(panel, section, workspace) {
	$(panel).droppable({
		// only accept widgetwrappers to be dropped
		accept: this._acceptSelector,
		drop: _.bind(function(e, ui) {
			// drop in sortable widget dropzone, find the correct sibbling
			var sibling = this.findSibling(panel, ui.draggable, ui.offset.top - panel.offset().top);
			var widgetWrapper = ui.draggable[0].wrapper;
			if (sibling === undefined) {
				// no sibling found, just append to the section
				workspace.moveItem(widgetWrapper.getID(), section, undefined, true);
			} else {
				if (sibling.wrapper) {
					sibling = sibling.wrapper;
					workspace.moveItem(widgetWrapper.getID(), section, sibling.getID(), true);
				}
			}
		}, this),

		// a widgetwrapper is being dragged over this section, register this section as the current dropzone
		over: _.bind(function(e, ui) {
			this.hoveredSortableDropzone = panel;
		}, this),

		// the widgetwrapper is no long being dragged over this section
		out: _.bind(function(e, ui) {
			this.hoveredSortableDropzone = undefined;
		}, this),
		tolerance: "touch"
	});
};

/**
 * Add a DropZone in which widgets can be dropped on top of all content (widget becomes modal!)
 * @param panel
 */
DraggableWidgets.prototype.addModalWidgetDropzone = function(panel, section, workspace) {
	$(panel).droppable({
		accept: this._acceptSelector,
		drop: _.bind(function(e, ui) {
			var widgetWrapper = ui.draggable[0].wrapper;
			if (widgetWrapper) {
				widgetWrapper.setModal(true, ui.offset.top, ui.offset.left, widgetWrapper.drag.sizeBeforeDrag.width);

				workspace.moveItem(widgetWrapper.getID(), section);

				widgetWrapper.widget.fireModalResize();
			}
		}, this)
	});
};

/**
 * Remove a panel as dropzone
 * @param panel
 */
DraggableWidgets.prototype.removeDropzone = function(panel) {
	$(panel).droppable("disable");
};

/**
 * Insert a element in a panel at a certain position (used for enabling sorting)
 * @param list
 * @param el
 * @param position
 */
DraggableWidgets.prototype.insertAt = function(list, el, position) {
	var insertBefore = this.findSibling(list, el, position);
	if (insertBefore === undefined) {
		list.append(el);
	} else {
		$(insertBefore).before(el);
	}
};

/**
 * Find the correct position to insert element
 * @param list
 * @param el
 * @param position
 * @returns element before which the element should be inserted
 */
DraggableWidgets.prototype.findSibling = function(list, el, position) {
	var currentHeight = 0;
	var insertBefore;
	var canContinue = true;

	list.children(this._siblingSelector).each(function(key, el) {
		var outerHeight = $(el).outerHeight(true) || null;
		if (canContinue && position <= currentHeight + outerHeight / 2) {
			insertBefore = el;
			canContinue = false;
		}
		currentHeight += outerHeight;
	});

	return insertBefore;
};

export default DraggableWidgets;
