import ListenerUtil from "@esko/web-viewer/utils/ListenerUtil";
import _ from "underscore";
/**
 * @exports Layout
 * @class
 * Defines a general implementation for different layouts. Every layout has different modes it can change to.
 * @constructor
 * @param {core.ViewerFacade} f
 * @param {LayoutModel} layoutModel
 * @param {JQueryObj} context
 * @param {Object} layoutState
 * @param {LoadingIndicator} loadingIndicator - the loading indicator, which was created before all requests started
 * @param {JQueryObj} initialLoader - the JQuery reference to the element indicating the initial loader (needed to safely remove it without emptying the context)
 */
var Layout = function(f, layoutModel, context, layoutState, loadingIndicator, initialLoader) {
	ListenerUtil.mixin(this, f);

	this.f = f;

	/**
	 * Context of this layout, this is the DOM-object the layout will use
	 * @member {JQueryObj}
	 */
	this._context = context;

	/**
	 * The layout model, holds the layoutItems
	 * @member {LayoutModel}
	 */
	this._layoutModel = layoutModel;

	/**
	 * The loaded state of the layout.
	 */
	this._layoutState = layoutState;

	/**
	 * The loadingIndicator
	 */
	this._loadingIndicator = loadingIndicator;

	/**
	 * The DOM element used to show initial loader in the context
	 */
	this._initialLoader = initialLoader;

	/**
	 * Currently active mode for this layout
	 * @member {LayoutMode}
	 */
	this._activeMode = null; //to be set during initialization of the specific layout

	//register listener to the layoutModel for when the layout model is updated (item is added)
	this._registerListener(
		"itemAdded",
		function(layoutItem) {
			this._addItem(layoutItem);
		},
		this,
		this._layoutModel
	);
	this._registerListener("layout.switchLayoutMode", _.bind(this._switchLayoutMode, this), this, this.f);

	// Wait untill the subclass has finished registering all available LayoutModes before continuing
	this.f.once(
		"layout.finishRegisteringModes",
		function() {
			var switchLayoutMode = function() {
				var page = this.f.getDocumentController().getActivePage();
				var compareModule = this.f.getModule("compare");
				var activeCompareMode;
				var layoutMode;
				if (compareModule) {
					activeCompareMode = compareModule.getActiveMode();
				}
				if (activeCompareMode) {
					layoutMode = activeCompareMode.getLayoutMode();
				} else if (page) {
					layoutMode = page.getLayoutMode();
				}
				this.f.emit("layout.switchLayoutMode", layoutMode);
			}.bind(this);

			this.f.getDocumentController().whenActivePageChanged(switchLayoutMode);

			this.f.on(
				"module.compare.changed.mode",
				function() {
					// make sure the compare module is available (might not have been returned yet)
					this.f.whenModuleAvailable("compare", switchLayoutMode);
				}.bind(this)
			);

			_.defer(
				function() {
					// tick to indicate activity
					var tick = _.throttle(
						function() {
							this._emit("tick");
						}.bind(this),
						10000
					);
					this.getMainContentElement().on("mousemove", tick);
					this.getMainContentElement().on("touchmove", tick);
					this.getMainContentElement().on("pointermove", tick);
				}.bind(this)
			);
		}.bind(this)
	);
};

/**
 * Adds a layoutItem to the layout
 * @param {Object} layoutItem
 */
Layout.prototype._addItem = function(layoutItem) {
	//to be implemented by specific layout
};

/**
 * Decide which of the added items should be shown and which ones should be hidden.
 */
Layout.prototype._updateItemVisibility = function() {
	//to be implemented by specific layout
};

/**
 * The layout has changed, emit this.
 */
Layout.prototype._emitLayoutChanged = function() {
	this._emit("layout.changed");
};

/**
 * Check whether the layout should be persisted or not, depends on the active mode.
 * @returns {Boolean}
 */
Layout.prototype.isLayoutStatePersisted = function() {
	return this._activeMode.isPersisted();
};

/**
 * Get the state of the layout.
 * @returns {Object}
 */
Layout.prototype.getLayoutState = function() {
	//to be implemented by specific layout
};

/**
 * Switch the active mode.
 * @param {LayoutConstants.LayoutMode} [mode] - if not provided it will use the default mode for the layout
 */
Layout.prototype._switchLayoutMode = function(mode) {
	if (this._activeMode === null || mode !== this._activeMode.getName()) {
		var newMode = null;
		if (mode != undefined) {
			newMode = this._layoutModel.getLayoutMode(mode);
		} else {
			newMode = this._getDefaultLayoutMode();
		}
		if (newMode != null) {
			if (this._activeMode != null) {
				this._activeMode.toggle(false);
			}
			this._activeMode = newMode;
			this._activeMode.toggle(true);
			this._debouncedLayoutUpdate(newMode);
		}
	}
};

/**
 * debounce the update as multiple switch layouts could have been triggered.
 */
Layout.prototype._debouncedLayoutUpdate = _.debounce(function(newMode) {
	this._updateItemVisibility();
	this._emit("changed.layoutMode", newMode.getName());
}, 0);

/**
 * Returns the default layout mode for the layout.
 * @return {LayoutMode}
 */
Layout.prototype._getDefaultLayoutMode = function() {
	//to be implemented by specific layout
};

/**
 * Get a reference to the active layout mode
 * @returns {LayoutMode}
 */
Layout.prototype.getActiveLayoutMode = function() {
	return this._activeMode;
};

/**
 * Returns the type of the layout.
 * @return {Number}
 */
Layout.prototype.getLayoutType = function() {
	//to be implemented by specific layout
	return -1;
};

/**
 * Makes the center content pane fullscreen
 *
 * @returns True if making center pane fullscreen was succesful
 */
Layout.prototype.fullscreenViewport = function() {
	//To be decided by specific layout if they want to implement this
	return false;
};

Layout.prototype.getMainContentElement = function() {
	return this.f.getContext();
};

/**
 * return default tool or null if no default tool.
 */
Layout.prototype.getDefaultTool = function() {
	return this._activeMode && this._activeMode.getDefaultTool();
};

Layout.prototype.getLayoutMode = function(name) {
	return this._layoutModel.getLayoutMode(name);
};

export default Layout;
