import LayoutConstants from "@esko/web-viewer/layout/LayoutConstants";
import PrivateChannel from "@esko/web-viewer/core/PrivateChannel";
/**
 * @exports LayoutMode
 * @memberof Layout
 * @class
 * Defines a general implementation of layout mode, the mode decides if items should be added to the layout
 * and if the state should be persisted.
 * @constructor
 * @param {Core.ViewerFacade} f
 * @param {Layout} layout
 * @param {LayoutConstants.LayoutMode} name
 */
var LayoutMode = function(f, layout, name) {
	PrivateChannel.installTo(this, f, "LayoutMode-" + name);
	this.f = f;

	/**
	 * Name of the layoutMode
	 */
	this._name = name;

	/**
	 * Is the layoutMode active
	 */
	this._isActive = false;

	/**
	 * The layout for which this mode is active
	 */
	this._layout = layout;

	/**
	 * is the modepanel expanded
	 */
	this._isModePanelExpanded = true;
};

/**
 * Get the name of the mode.
 * @returns {LayoutConstants.LayoutMode}
 */
LayoutMode.prototype.getName = function() {
	return this._name;
};

/**
 * Turns On/Off the layoutMode
 * @param {Boolean} toggle
 */
LayoutMode.prototype.toggle = function(toggle) {
	//to be implemented by specific layout mode
};

/**
 * Checks if the item is allowed for this layout mode. Prevents channels and CAD widget from loading.
 *
 * @param {String} itemname
 * @returns {Boolean}
 */
LayoutMode.prototype.isItemAllowed = function(itemname) {
	//to be implemented by specific layout mode
};

/**
 * Checks if the state of the layout should be persisted.
 * @return {Boolean}
 */
LayoutMode.prototype.isPersisted = function() {
	//to be implemented by specific layout mode
};

/**
 * Checks if the mode is active at the moment.
 */
LayoutMode.prototype.isActive = function() {
	return this._isActive;
};

/**
 * Decide whether or not to enable the mode panel
 */
LayoutMode.prototype.isModePanelEnabled = function() {
	// should be overwritten for mobile layouts in which the mode panel should be shown
	return false;
};

/**
 * If the mode panel is enabled, should it be expanded by default?
 * @returns {Boolean} is de mode panel expanded by default?
 */
LayoutMode.prototype.isModePanelExpanded = function() {
	return this._isModePanelExpanded && this.isModePanelEnabled();
};

/**
 * Should the widget titles be shown?
 * @returns {Boolean}
 */
LayoutMode.prototype.showModePanelWidgetTitles = function() {
	return true;
};

/**
 * Is the modepnale expanded or not?
 * This method is called by the layout if a modepanel was expanded or collapsed
 * @param {Boolean} expanded
 */
LayoutMode.prototype.setIsmodePanelExpanded = function(expanded) {
	this._isModePanelExpanded = expanded;
};

/**
 * This method should return a list of widgets that should appear in the modepanel (or an empty array if none should be shown)
 * @returns {Array.<String>} widget names of the widgets shown in the mode panel
 */
LayoutMode.prototype.getModePanelContent = function() {
	return [];
};

/**
 * Get the name of the widget which should be shown in the Main Panel
 * @returns {Array.<String>}  name of the widgets shown in the main panel
 */
LayoutMode.prototype.getMainPanelContent = function() {
	return this.f.getActivePage() && this.f.getActivePage().is3D() ? [LayoutConstants.Widget.VIEW3D] : [LayoutConstants.Widget.VIEWPORT];
};

/**
 * This method can be overridden if a particular mode needs a different default tool
 */
LayoutMode.prototype.getDefaultTool = function() {};

export default LayoutMode;
