import PrivateChannel from "@esko/web-viewer/core/PrivateChannel";
import LayoutConstants from "@esko/web-viewer/layout/LayoutConstants";
import LayoutItem from "@esko/web-viewer/layout/LayoutItem";
/**
 *
 * @exports LayoutModel
 * @memberof Layout
 */
var LayoutModel = function(f) {
	PrivateChannel.installTo(this, f, ["layoutModel"]);

	this.f = f;

	/**
	 * Maps the name of the layoutItem on a reference to the layoutItem (typically widget, toolbarcomponent)
	 */
	this._items = {};

	/**
	 * Holds all the layoutmodes that are available for the loaded layout.
	 */
	this._layoutModes = {};
};

/**
 * Checks if the layoutItem with itemName is part of the layoutModel
 * @param {String} itemName
 * @returns {Boolean}
 */
LayoutModel.prototype.hasItem = function(itemName) {
	return this._items[itemName] !== undefined;
};

/**
 * Add a widget to the layoutModel
 * @param {LayoutConstants.Widget} name - the name of the widget
 * @param {Function} widgetFactory - factory function to instantiate the widget (with 1 parameter: LayoutConstants.LayoutType)
 */
LayoutModel.prototype.addWidget = function(name, widgetFactory) {
	var newItem = new LayoutItem(name, LayoutConstants.LayoutItemType.WIDGET, widgetFactory);
	this._items[newItem.name] = newItem;

	this._emit("itemAdded", newItem);
};

/**
 * Add a toolbarComponent to the layoutModel
 * @param {LayoutConstants.ToolbarComponent} name - the name the toolbarcomponent
 * @param {Function} toolbarComponentFactory - factory function to instantiate the toolbarcomponent (with 1 parameter: LayoutConstants.LayoutType)
 */
LayoutModel.prototype.addToolbarComponent = function(name, toolbarComponentFactory) {
	var newItem = new LayoutItem(name, LayoutConstants.LayoutItemType.TOOLBARCOMPONENT, toolbarComponentFactory);
	this._items[newItem.name] = newItem;

	this._emit("itemAdded", newItem);
};

/**
 * Get the layoutType of the layoutItem
 * @param {String} itemName
 * @returns {LayoutConstants.LayoutItemType}
 */
LayoutModel.prototype.getLayoutItemType = function(itemName) {
	if (this.hasItem(itemName)) {
		return this._items[itemName].type;
	} else {
		return undefined;
	}
};

/**
 * Get the layout item.
 * @param {String} itemName
 * @returns {Object}
 */
LayoutModel.prototype.getLayoutItem = function(itemName) {
	if (this.hasItem(itemName)) {
		return this._items[itemName].getItem();
	} else {
		return undefined;
	}
};

/**
 * Register a layoutMode to the model. All available mode for the loaded layout are saved in the model.
 * @param {LayoutMode} mode
 */
LayoutModel.prototype.registerLayoutMode = function(mode) {
	this._layoutModes[mode.getName()] = mode;
};

/**
 * Get the LayoutMode.
 * @param {LayoutConstants.LayoutMode} modeName
 * @returns {LayoutMode}
 */
LayoutModel.prototype.getLayoutMode = function(modeName) {
	return this._layoutModes[modeName];
};

export default LayoutModel;
