import _ from "underscore";

export default {
	/**
	 * Layout al WidgetWrappers in a certain element.
	 * It respects the "willFill" and minSize, maxSize and preferredSize of the widgets
	 * When there is not enough space to fit all widgets, it adds a scrollbar to containerElement
	 * @param {JQueryObj} containerElement - the element in which the widgets reside. The element should have a height for the algorithm to work
	 * @param {Number} widgetBorder - this is some extra spacing that is not taken into account by outerheight
	 */
	layoutWidgets: function(containerElement, widgetBorders) {
		var widgets = containerElement.find(".widgetWrapper");
		var innerHeight = (containerElement.innerHeight() || null) - widgetBorders;
		var data = [];
		var w;
		var currentHeight = 0;
		var max;
		var canShrink = [];
		var canGrow = [];

		// first build up the needed data structure
		for (var i = 0; i < widgets.length; i++) {
			var widget = widgets[i].wrapper;
			if (widget.isVisible()) {
				max = undefined;
				if (widget.getMaximumSize()) {
					max = widget.getMaximumSize().height;
				}
				w = {
					widget: widget,
					min: widget.getMinimumSize().height,
					pref: widget.getPreferredSize().height,
					max: max,
					height: 0
				};
				data.push(w);

				//calculate preferred height (taking into account min and max)

				w.height = Math.max(w.min, w.pref);
				if (w.max) {
					w.height = Math.min(w.max, w.height);
				}

				w.height = Math.min(w.height, innerHeight);

				// first we only want to shrink willFill elements
				// keep track if the element can be shrinked
				if (w.widget.willFill() && w.height > w.min) {
					canShrink.push(w);
				}
				currentHeight += w.height + widgetBorders; // 4: 2px padding and 2px border
			}
		}
		var pixelsNeeded = 0;
		// this function is used to shrink all shrinkable elements
		// it is defined as a inner function because it is used two times
		var shrinkShrinkables = function() {
			pixelsNeeded = currentHeight - innerHeight;
			//as long as we need to shrink, and we have shrinkable elements, try to shrink
			while (pixelsNeeded > 0 && canShrink.length > 0) {
				//we want to shrink proportionally
				var w = _.first(canShrink);
				var shrinkablesPrefHeight = _.reduce(
					canShrink,
					function(memo, w) {
						return memo + w.pref;
					},
					0
				);
				var diff = Math.ceil((w.pref / shrinkablesPrefHeight) * pixelsNeeded);

				canShrink = _.rest(canShrink, 1);

				var prevHeight = w.height;
				w.height = Math.max(w.min, w.height - diff);

				var pixelsRemoved = prevHeight - w.height;
				pixelsNeeded -= pixelsRemoved;
				currentHeight -= pixelsRemoved;
				if (w.height > w.min) {
					canShrink.push(w);
				}
			}
		};

		// first try to shrink all "willFill" elements... these are usually very flexible elements, so they care less if they are shrunk
		shrinkShrinkables();

		if (pixelsNeeded > 0) {
			// we need another round of shrinking...
			// now we take every element that could be shrunk (so its current height is still higher than its minimum size
			canShrink = _.filter(data, function(w) {
				return w.height > w.min;
			});

			shrinkShrinkables();
		}

		pixelsNeeded = currentHeight - innerHeight;
		if (pixelsNeeded > 0) {
			// oh ooh... we can't get small enough... show a scrollbar
			containerElement.css("overflow-y", "auto");
		} else {
			containerElement.css("overflow-y", "hidden");
		}

		// at this point, we should no longer exceed the maximum size (we could still be too big if an element has a big minimum size

		// check if we can extend some elements
		var pixelsLeft = innerHeight - currentHeight;
		if (pixelsLeft > 0) {
			_.each(data, function(w) {
				if ((!w.max || w.height < w.max) && (w.height < w.pref || w.widget.willFill())) {
					canGrow.push(w);
				}
			});

			while (pixelsLeft > 0 && canGrow.length > 0) {
				var diff = Math.ceil(pixelsLeft / canGrow.length);
				w = _.first(canGrow);
				canGrow = _.rest(canGrow, 1);

				var prevHeight = w.height;
				if (w.max) {
					w.height = Math.min(w.height + diff, w.max);
				} else {
					w.height = w.height + diff;
				}
				var pixelsAdded = w.height - prevHeight;

				pixelsLeft -= pixelsAdded;
				if ((!w.max || w.height < w.max) && (w.height < w.pref || w.widget.willFill())) {
					canGrow.push(w);
				}
			}
		}

		// set all heights
		_.each(data, function(w) {
			w.widget.setHeight(w.height);
		});
	}
};
