import $ from "jquery";
import _ from "underscore";
import ShapeUtils from "@esko/shapeutils";
/**
 * A widgetwrapper is a wrapper for a widget, which makes enables the widget to be displayed with some extra controls
 * @param {AbstractWidget} widget          - the widget to wrap
 * @param {String}         widgetName      - the name of the widget
 * @param {Boolean} allowCollapse - is collapsing the widget allowed?
 */
var WidgetWrapper = function(widget, widgetName, allowCollapse) {
	this._headerEnabled = true;
	// main wrapper
	this.element = $("<div/>").addClass("widgetWrapper");
	this.element[0].wrapper = this;

	// webkit has a bug with iFrames, which appearantly they won't be solving soon
	// if an input in an iframe is focused, the page scrolls to the iframe
	// this can occur when TinyMCE is used... we can fix it by scrolling to top whenever the context is scrolled
	this.element.on("scroll", function(event) {
		$(this).scrollTop(0);
	});

	this.name = widgetName;

	// header + title
	this.header = $("<div/>")
		.addClass("header")
		.attr("data-sahi-id", "viewer-" + widgetName + "-widget-header");
	this.headerText = $("<h3>", {
		text: widget.getTitle(),
		title: widget.getTitle()
	}).prepend($("<i>", { class: "icon-normal icon-WCR_Down-S" }));
	this.headerHandle = $("<div/>")
		.addClass("headerHandle")
		.append(this.headerText);
	if (allowCollapse) {
		this.headerHandle.on(
			"click",
			_.bind(function() {
				this.toggle();
			}, this)
		);
	}
	this.header.append(this.headerHandle);

	this.headerExtra = $("<div/>").addClass("headerExtra");
	this.header.append(this.headerExtra);

	// listen to title changed of widget
	widget.on("titlechanged", _.bind(this._onTitleChanged, this));

	// content holder
	this.content = $("<div/>").addClass("content");

	// create dom structure
	this.element.append(this.header, this.content);

	// activate the widget
	widget.setContext(this.content, this.headerExtra);

	//is the widget resizable
	this.resizable = false;

	// the widget
	this.widget = widget;

	this.isModal = false;

	this.isCollapsed = false;
	this._isVisible = true;

	//location; only use if the widgetwrapper is hidden
	this._locationWhenHidden = new ShapeUtils.Point();
};

/**
 * Convert the widget in to a modal widget or normal widget
 * @param modal {boolean} should the widget be modal or not?
 * @param top {int} position top
 * @param left {int} position left
 * @param width {int} width of modal window
 * @param height {int} height of modal window
 */
WidgetWrapper.prototype.setModal = function(modal, top, left, width, height) {
	this.widget.setModal(modal);
	if (modal) {
		this.element.css({
			position: "absolute",
			top: top,
			left: left,
			width: width || this.element.width() || undefined,
			height: height || this.element.height() || undefined
		});
		var newHeight = height || this.element.height() || 0;
		this.setHeight(newHeight);

		this.setResizable(!this.isCollapsed);
		this.isModal = true;
	} else {
		this.isModal = false;
		// restore normal css
		this.element.css({
			position: "relative",
			top: "",
			left: "",
			width: "auto"
		});

		this.setResizable(false);
	}
};

/**
 * Function called when the title of the widget changed
 * @param {String} newTitle - The new title
 */
WidgetWrapper.prototype._onTitleChanged = function(newTitle) {
	this.headerText.text(newTitle).attr("title", newTitle);
};

/**
 * Toggle collapse state of the widget
 */
WidgetWrapper.prototype.toggle = function() {
	this.setCollapsed(!this.isCollapsed);
};

/**
 * Hides/shows a widget
 * @param {Boolean} isVisible
 */
WidgetWrapper.prototype.setVisible = function(isVisible) {
	this._isVisible = isVisible;
	if (isVisible) {
		this.element.removeClass("hiddenElement");
	} else {
		var offset = this.element.offset();
		//location is only used when the widgetwrapper is hidden, if visible the element is used
		this._locationWhenHidden = new ShapeUtils.Point(offset.left, offset.top);
		this.element.addClass("hiddenElement");
	}
};

/**
 * Is the widgetwrapper visible.
 * @returns {Boolean}
 */
WidgetWrapper.prototype.isVisible = function() {
	return this._isVisible;
};

/**
 * (un)Collapse the widget
 * @param {Boolean} collapsed
 */
WidgetWrapper.prototype.setCollapsed = function(collapse) {
	var wasCollapsed = this.isCollapsed;
	this.element.toggleClass("collapsed", collapse);
	this.isCollapsed = collapse;
	this.widget.setCollapsed(collapse);
	this.headerText.find(".icon-normal").toggleClass("icon-WCR_Down-S", !collapse);
	this.headerText.find(".icon-normal").toggleClass("icon-WCR_Right-S", collapse);

	if (this.isModal && wasCollapsed !== collapse) {
		var newHeight = Math.min($(window).height() - (this.element.offset().top - $(window).scrollTop()), this.getPreferredSize().height);
		this.setHeight(newHeight);
		if (!collapse) {
			this.setResizable(true);
		}
		this.setResizable(!this.isCollapsed);
		this.widget.fireModalResize();
	}
};

WidgetWrapper.prototype.setResizable = function(resizable) {
	if (this.resizable && !resizable) {
		this.element.resizable("destroy");
		this.resizable = false;
	} else if (!this.resizable && resizable) {
		//make the element resizable
		var minimumSize = this.getMinimumSize();
		this.element.resizable({
			minWidth: Math.max(100, minimumSize.width),
			minHeight: Math.max(40, minimumSize.height),
			stop: _.bind(function(event, ui) {
				this.setHeight(ui.size.height);
				this.widget.fireModalResize();
			}, this),
			resize: _.throttle(
				_.bind(function(event, ui) {
					this.setHeight(ui.size.height);
					this.widget.resizing();
				}, this),
				40
			)
		});
		this.resizable = true;
	}
};

/**
 * Get the ID of the enclosed widget
 * @returns
 */
WidgetWrapper.prototype.getID = function() {
	return this.widget.getID();
};

/**
 * Get the minimum size of the widget content
 *
 * note it is possible that this size is not respected by the layoutmanager
 *
 * @returns {ShapeUtils.Size}
 */
WidgetWrapper.prototype.getMinimumSize = function() {
	if (this.isCollapsed) {
		return new ShapeUtils.Size(0, this.getHeaderHeight());
	} else {
		var minSize = this.widget.getMinimumSize().copy();
		minSize.height += this.getHeaderHeight() + 3;

		return minSize;
	}
};

/**
 * Get the preferred size of the widget content
 *
 * note it is possible that this size is not respected by the layoutmanager
 *
 * @returns {ShapeUtils.Size} if undefined, no preferred size is taken in to account by layoutmanager, the layout managers tries to guess the needed size
 */
WidgetWrapper.prototype.getPreferredSize = function() {
	if (this.isCollapsed) {
		return new ShapeUtils.Size(0, this.getHeaderHeight());
	} else {
		var prefSize = this.widget.getPreferredSize();
		if (prefSize) {
			prefSize = prefSize.copy();
			prefSize.height += this.getHeaderHeight() + 3;
			return prefSize;
		} else {
			return undefined;
		}
	}
};

/**
 * Get the maximum dimension of the widget content
 *
 * @returns {ShapeUtils.Size} if undefined the widget can be made as big as the user wants
 */
WidgetWrapper.prototype.getMaximumSize = function() {
	if (this.isCollapsed) {
		return new ShapeUtils.Size(1000, this.getHeaderHeight());
	} else {
		var maxSize = this.widget.getMaximumSize();
		if (maxSize) {
			maxSize = maxSize.copy();
			maxSize.height += this.getHeaderHeight() + 3;
			return maxSize;
		} else {
			return undefined;
		}
	}
};

/**
 * Get the height of the header
 * @returns
 */
WidgetWrapper.prototype.getHeaderHeight = function() {
	if (this._headerEnabled) {
		return this.header.outerHeight(true);
	} else {
		return 0;
	}
};

/**
 * Get the current content height of the widget
 * @returns
 */
WidgetWrapper.prototype.getContentHeight = function() {
	return this.content.getOuterHeight(true);
};

/**
 * Will the widgetwrapper extend if needed?
 * @returns
 */
WidgetWrapper.prototype.willFill = function() {
	return this.widget.willFill();
};

/**
 * Set the height of the widgetwrapper
 * @param height
 */
WidgetWrapper.prototype.setHeight = function(height) {
	var prevContentHeight = this.content.height();
	var newContentHeight = height - this.getHeaderHeight();
	this.element.css("height", height);
	this.content.css("height", newContentHeight);

	if (newContentHeight != prevContentHeight) {
		this.widget.resized();

		this.content.css("overflow", "hidden");
		_.defer(function(el) {
			el.css("overflow", "auto");
		}, this.content);
	}
};

/**
 * Get the size of the WidgetWrapper
 *
 * @returns {ShapeUtils.Size}
 */
WidgetWrapper.prototype.getSize = function() {
	return new ShapeUtils.Size(this.element.width(), this.element.height());
};

/**
 * Get the outer size of the widget wrapper, including borders and padding
 *
 * @returns {ShapeUtils.Size}
 */
WidgetWrapper.prototype.getOuterSize = function() {
	return new ShapeUtils.Size(this.element.outerWidth(true), this.element.outerHeight(true));
};

/**
 * Get the location of the WidgetWrapper
 *
 * @returns {ShapeUtils.Point}
 */
WidgetWrapper.prototype.getLocation = function() {
	if (this._isVisible) {
		var offset = this.element.offset();
		return new ShapeUtils.Point(offset.left, offset.top);
	} else {
		return this._locationWhenHidden;
	}
};

/**
 * Enable/Disable the header display of the widget wrapper
 * @param {Boolean} enable
 */
WidgetWrapper.prototype.setHeaderEnabled = function(enable) {
	this._headerEnabled = enable;
	this.header.toggleClass("displayNone", !enable);
};

export default WidgetWrapper;
