import ViewerUtils from "@esko/web-viewer/utils/ViewerUtils";
import Tooltip from "@esko/tooltip/Tooltip";
import _ from "underscore";
import $ from "jquery";
import "spectrum-colorpicker/spectrum";
import "spectrum-colorpicker/spectrum.css";
/**
 * @exports ColorPicker
 * @constructor
 * @param {Function} onColorPicked - callback called when a color is picked, parameter: {String} (hex)color without #
 * @param {HTMLObject} [element] - the element where the colorpicker needs to be added to
 */
var ColorPicker = function(f, onColorPicked, element, options) {
	// Callback on color picked
	this._onColorPicked = onColorPicked;

	// Current color (without #)
	this._color = undefined;

	// The color block
	this.element = undefined;
	if (element === undefined) {
		this.element = $("<span>").addClass("colorblock");
	} else {
		this.element = element;
	}

	this._lazyOnChange = _.debounce(_.bind(this._onChange, this), 50);
	options = options || {};
	var appendTo = $("<div>").appendTo(options.appendTo || f.getContext());
	// default options

	var cancelClicked = false;
	this._options = _.extend(
		{
			showInput: true,
			showInitial: true,
			preferredFormat: "hex",
			localStorageKey: "wcr.colorpicker",
			hideAfterPaletteSelect: true,
			maxSelectionSize: 4,
			cancelText: f.i18n.t("AppDialogCancel"),
			chooseText: f.i18n.t("OkButton"),
			change: this._lazyOnChange,
			hide: this._onChange.bind(this),
			move: this._lazyOnChange
		},
		options
	);

	this._options.appendTo = appendTo;
	this.element.spectrum(this._options);
	appendTo.find(".sp-cancel").on(
		"click",
		_.debounce(function() {
			if (options.revert) {
				options.revert();
			}
		}, 0)
	);

	if (options.sahiId) {
		this.element.get(0).dataset.sahiId = options.sahiId;
		appendTo.find(".sp-input").get(0).dataset.sahiId = options.sahiId + "-input";
		appendTo.find(".sp-choose").get(0).dataset.sahiId = options.sahiId + "-confirm";
		appendTo.find(".sp-cancel").get(0).dataset.sahiId = options.sahiId + "-cancel";
		appendTo.find(".sp-val").get(0).dataset.sahiId = options.sahiId + "-saturation";
		appendTo.find(".sp-hue").get(0).dataset.sahiId = options.sahiId + "-hue";
	}
	// Init colorpicker
};

/**
 * On Change of the color picker
 * @param {Object} hsb - HSB values
 * @param {String}   hex - Hex value
 * @param {Object} rgb - RGB values
 */
ColorPicker.prototype._onChange = function(color) {
	this._color = color.toHexString().substring(1);
	this._updateColor();
	this._onColorPicked(this._color);
};

/**
 * Set the color of the colorpicker.
 * @param {String} hexColor New color in hex format (with or without #)
 */
ColorPicker.prototype.setColor = function(hexColor) {
	if (ViewerUtils.startsWith(hexColor, "#")) {
		hexColor = hexColor.substring(1);
	}
	if (this._color != hexColor) {
		this.element.spectrum("set", hexColor);
		this._color = hexColor;
		this._updateColor();
	}
};

/**
 * Update the color of the color block if it was added by this class
 */
ColorPicker.prototype._updateColor = function() {
	if (this.element.hasClass("colorblock")) {
		this.element.css("background-color", "#" + this._color);
	}
};
/**
 * Get the color of the colorpicker
 * @return {String} current color in hex format (without #)
 */
ColorPicker.prototype.getColor = function() {
	return this._color;
};

/**
 * Show the color picker
 */
ColorPicker.prototype.show = function() {
	this.element.spectrum("show");
};

/**
 * Hide the color picker
 */
ColorPicker.prototype.hide = function() {
	this.element.spectrum("hide");
};

/**
 * Destroy the color picker
 */
ColorPicker.prototype.destroy = function() {
	this.element.spectrum("hide");
	this.element.spectrum("destroy");
	this.element.remove();
};

/**
 * Set Tooltip for the color picker element.
 * @param {String}         tooltipOrBundle - tooltip to show if langKey is undefined or language bundle
 * @param {String}         [langKey]       - not required
 * @param {Array.<String>} [stringArgs]    - not required, parameters that need to be filled in the language string
 */
ColorPicker.prototype.setTooltip = function(f, tooltipOrBundle, langKey, stringArgs) {
	if (this.tooltip === undefined) {
		this.tooltip = new Tooltip(f.i18n, this.element, f.getContext(), f.getContext(), f.isMobile(), tooltipOrBundle, langKey, stringArgs);
	} else {
		this.tooltip.setTooltipContent(tooltipOrBundle, langKey, stringArgs);
	}
};

export default ColorPicker;
