import DropdownMenu from "@esko/dropdownmenu/DropdownMenu";
import Template from "@esko/template";
import tpl from "@esko/web-viewer/layout/component/DocumentSelector.tpl.html";
import $ from "jquery";
import _ from "underscore";
/**
 *
 * @class A DocumentDropdownMenu is a dropdownmenu that when clicked shows a dropdown with a list of document pages. New
 * documents can be added to the list and the documents can be enabled/disabled. Changing the selected pages makes the previous
 * selected page back available in the the dropdown. It is also possible to group document pages. Group elements will hide all its elements,
 * when the group element is clicked all the pages of this group will become visible. Only document pages of the same type should be added
 * the same group.
 *
 * @exports Layout.Component.DocumentDropdownMenu
 * @constructor
 * @augments Layout.Component.DropdownMenu
 * @param {Core.ViewerFacade} f
 * @param {Function}          onDocumentSelect      - function to execute when a document is selected
 * @param {String}            [emptyListText]       - text shown in the list when no documents can be selected
 * @param {String}            [nothingSelectedText] - Text shown in the dropdown when no document is selected
 * @param {String}            [deselectText]        - Text used for the action to deselect the currently selected document, when not passed it is not possible to deselect the currently selected document
 * @param {Function}          [deselectCallback]    - callback to execute when the document gets deselected, first parameter is the deselected pageId
 * @param {Boolean}           [isMobile]            - adapt layout for mobile
 * @param {Boolean}           [isCompact]           - if true version number is put behind the document (on the same line)
 */
var DocumentDropdownMenu = function(
	f,
	onDocumentSelect,
	emptyListText,
	nothingSelectedText,
	deselectText,
	deselectCallback,
	isMobile,
	isCompact,
	options
) {
	this.f = f;
	DropdownMenu.call(this, this.f.i18n, this.f.getContext(), this.f.isMobile(), "bottom", "documentSelector", 300);

	this._options = _.extend({}, options);

	/**
	 * @member {Function} function that's executed when a new document gets selected
	 * @private
	 */
	this._onDocChange = onDocumentSelect;

	/**
	 * @member {JQueryObj} visible part of the dropdownmenu with the currently selected document page
	 */
	this.element = $("<div/>").addClass("selectedDocument");
	this.attachTo(this.element);

	this._isMobile = isMobile;

	/**
	 * @member {JQueryObj} element for the currently selected element
	 */
	this._selectedDocumentElement = $("<div>");
	if (this._isMobile) {
		this.element.append('<i class="icon-WCR_Down"></i>');
	}
	this.element.append(this._selectedDocumentElement);
	if (!this._isMobile) {
		this.element.append(
			$("<div>")
				.addClass("docSelectorCaret")
				.append($("<span/>").addClass("caret"))
		);
	}

	/**
	 * @member {Object} the selected document
	 * @private
	 */
	this._selectedDocumentPage = null;

	/**
	 * @member {Number} the number of enabled documents
	 * @private
	 */
	this._enabledDocuments = 0;

	/**
	 * @member {Object} maps a group name to an array of element names
	 * @private
	 */
	this._elementGrouping = {};

	/**
	 * @member {String} the currently active group, group of which all elements are visible
	 * @private
	 */
	this._activeGroup = null;

	/**
	 * @member {String} id of the divider element in the dropdown
	 */
	this.dividerId = null;

	/**
	 * @member {Template} template used to generate each row of document page information
	 */
	this.t = new Template(this.f.i18n, tpl);

	this.unveil = _.debounce(_.partial(this._unveil, true), 100);
	this.t.d.compact = isCompact;
	this._deselectText = deselectText || "";
	this._nothingSelectedText = nothingSelectedText || "";
	this._deselectCallback = deselectCallback;

	if (this._deselectText !== "" && this._nothingSelectedText !== "") {
		var html = $("<div />")
			.addClass("deselectElement")
			.html(this._nothingSelectedText);
		this.deselectElement = {
			name: "deselectAction",
			content: html,
			icon: undefined,
			action: _.bind(function() {
				var selectedPageId = this.getSelectedDocumentPageId();
				this.deselectCurrentPage(true);
				this._deselectCallback(selectedPageId);
			}, this),
			enabled: true
		};
		this.addAction(this.deselectElement);
		this.disableDocumentPage(this.deselectElement.name);
		this._selectedDocumentPage = this.deselectElement;
		this._selectedDocumentElement.html(this._selectedDocumentPage.content.clone());
		this._enabledDocuments++;
	}
	if (emptyListText !== "") {
		var html = $("<div />")
			.addClass("deselectElement")
			.html(emptyListText);
		this.emptyListElement = {
			name: "emptyList",
			content: html,
			icon: undefined,
			action: function() {},
			enabled: true
		};
		this.addAction(this.emptyListElement);
	}
};

DocumentDropdownMenu.prototype = Object.create(DropdownMenu.prototype);
DocumentDropdownMenu.prototype.constructor = DocumentDropdownMenu;

/**
 * Initialize the document dropdown menu
 * @private
 * @override
 */
DocumentDropdownMenu.prototype._initialize = function() {
	DropdownMenu.prototype._initialize.call(this);
	if (this._isMobile) {
		this.popover.setPlacementOptions(["bottomLeft"]);
	} else {
		this.popover.setPlacementOptions(["bottom", "top"]);
	}
	this.popover.options.inheritSizeFromContext = true;
};

/**
 * Add a document to the dropdown menu
 * @param {Object} documentInfo : description of the document
 * @param {String} groupName - name of the group the document belongs to
 * @example
 * {
 * 	pageId: 'uniquePageId',
 * 	project: 'projectName',
 * 	folder: 'folderName'
 * 	documentName: 'nameOfTheDocument',
 * 	[versionNumber: number,]
 * 	[pageNumber: page,]
 * 	[pageCount: numberOfPages,]
 *  thumbnailUrl: String,
 *  docType: docTypeConstant,
 *  [isVisible: Boolean,]
 *  [group: String]
 * }
 */
DocumentDropdownMenu.prototype.addDocumentPage = function(docInfo, groupName) {
	if (groupName !== undefined && this._elementGrouping[groupName] !== undefined) {
		if (this._elementGrouping[groupName].length === 0) {
			//no previous elements where added to the group, use the first page's information for the group element
			this.t.d.docInfo = _.clone(docInfo); //make sure to make copy, will be changing this object for generating html
			this.t.d.docInfo.isGroup = true; //identify this element as group
			this.t.d.docInfo.isOpen = this.items[groupName].isOpen;
			this.t.d.thumbnailUrl = docInfo.thumbnailUrl;
			var html = this.t.getRenderedHtml();
			this.setActionText(groupName, $(html)); //copy generated html in group element
		}
		this._elementGrouping[groupName].push(docInfo.pageId);
	}
	if (this.items[docInfo.pageId] === undefined) {
		//to make sure a document page doesn't get added twice
		this.t.d.docInfo = docInfo;
		this.t.d.thumbnailUrl = "";
		var html = this.t.getRenderedHtml();
		var dropdownElement = {
			name: docInfo.pageId,
			content: html,
			icon: undefined,
			action: _.bind(
				function(pageId) {
					this._itemClicked(pageId);
				},
				this,
				docInfo.pageId
			),
			enabled: true,
			docType: docInfo.docType,
			isVisible: groupName === undefined || this._activeGroup === groupName, //if page doesn't belong to a group, make it visible //if the groupName is the same as the currently active group

			group: groupName, //name of the group this element belongs to
			thumb: docInfo.thumbnailUrl,
			isUnveiled: false,
			after: docInfo.prepend && this.deselectElement ? this.deselectElement.name : undefined
		};
		this.addAction(dropdownElement);
		if (dropdownElement.isVisible) {
			//only if the element is visible at creation of the element add to the counter of currently active pages
			this._enabledDocuments++;
		}
		if (this.emptyListElement !== undefined && this._enabledDocuments > 0) {
			this.hideElement(this.emptyListElement.name);
			this.showElement(this.dividerId);
		}
	}
};

/**
 * Remove a document page from the document dropdown menu.
 * @param {String} pageId - the unique ID for the document page
 */
DocumentDropdownMenu.prototype.removeDocumentPage = function(pageId) {
	this.removeAction(pageId);
	delete this.items[pageId];
};

/**
 * Add a divider to the dropdown menu, there can only be one divider added to the dropdown menu
 */
DocumentDropdownMenu.prototype.addDivider = function() {
	if (this.dividerId === null) {
		this.dividerId = DropdownMenu.prototype.addDivider.call(this); //same the id to be able to hide when the emptylist element is shown
		if (this.emptyListElement !== undefined && this.emptyListElement.isVisible) {
			//if the emptylist element is already visible, make sure to hide the divider
			this.hideElement(this.dividerId);
		}
	}
};

DocumentDropdownMenu.prototype._addItemToDOM = function(item) {
	DropdownMenu.prototype._addItemToDOM.call(this, item);
	if (this._options.sahiId) {
		this._listElements[item.name].attr("data-sahi-id", this._options.sahiId + "-" + item.name);
	}
};

/**
 * Add a document group to the dropdown menu. After adding the group document pages can be added to that group.
 * The group element will be shown in the dropdown when the enabled document type is the same as the elements in the group and
 * the elements of the group are not enabled in the dropdown. Only document pages of the same document type should be added to
 * a document group.
 * @param {String} groupName - name of the group, is the name for the element in the dropdown
 * @param {Function} [beforeClickCallback] - callback will be called before the group element in the dropdown is clicked.
 * @param {Function} [afterClickCallback] - callback will be called after the group element in the dropdown is clicked.
 */
DocumentDropdownMenu.prototype.addDocumentGroup = function(groupName, beforeClickCallback, afterClickCallback) {
	this._elementGrouping[groupName] = [];
	var html = $("<div />"); //first create an empty element for the group, when adding document pages to the group, the first added page will be used for the dropdown element
	this.groupElement = {
		name: groupName,
		content: html,
		icon: undefined,
		isOpen: false,
		action: _.bind(function() {
			if (beforeClickCallback !== undefined) {
				beforeClickCallback(); //call the callback at the beginning of the click
			}

			if (this.isGroupOpen(groupName)) {
				this.hideElementGroup(groupName);
			} else {
				this.showElementGroup(groupName); //show the clicked group
			}

			// dropdown size has changed, so trigger repositioning and unveiling
			this.show();

			if (afterClickCallback !== undefined) {
				afterClickCallback(); //call the callback at the end of the click
			}
		}, this),
		enabled: true,
		doNotHideOnClick: true, //to make sure the dropdown will not be hidden when the group element is clicked in the dropdown
		isUnveiled: true,
		isGroup: true
	};
	this._enabledDocuments++; //add as an enabled document, the element should be hidden later if that's needed
	this.addGroup(this.groupElement); //add the group element to the dropdown, prepend group elements to make sure they are above the divider
};

/**
 * Set the selected document page
 * @param {String} pageId
 */
DocumentDropdownMenu.prototype.setSelectedDocumentPage = function(pageId) {
	if (this._selectedDocumentPage !== this.deselectElement && this._selectedDocumentPage !== null) {
		this.enableDocumentPage(this._selectedDocumentPage.name);
	} else if (this._selectedDocumentPage !== null) {
		this.deselectElement.content.html(this._deselectText);
		this.enableDocumentPage(this.deselectElement.name);
	}
	this._selectedDocumentPage = this.items[pageId];
	if (this._selectedDocumentPage !== undefined) {
		this._selectedDocumentElement.html(this._selectedDocumentPage.content);
		this._selectedDocumentElement.find(".thumbnail").attr("src", this._selectedDocumentPage.thumb);
	} else {
		this._selectedDocumentPage = null;
	}

	this.disableDocumentPage(pageId);
};

/**
 * Set a temp content for the document selector, this will not be added to the dropdown list when getting deactivated.
 * It sets the currently selected document page to null.
 * @param {Object} docInfo
 */
DocumentDropdownMenu.prototype.setTempSelectedDocumentPage = function(docInfo) {
	this.t.d.docInfo = docInfo;
	this.t.d.thumbnailUrl = docInfo.thumbnailUrl;
	var html = this.t.getRenderedHtml();
	if (this._selectedDocumentPage !== this.deselectElement && this._selectedDocumentPage !== null) {
		this.enableDocumentPage(this._selectedDocumentPage.name);
	} else if (this._selectedDocumentPage !== null) {
		this.deselectElement.content.html(this._deselectText);
		this.enableDocumentPage(this.deselectElement.name);
	}
	this._selectedDocumentElement.html(html);
	this._selectedDocumentPage = null;
};

/**
 * Is called when an item in the dropdown is clicked
 * @private
 * @param {String} pageId - the id of the document page
 */
DocumentDropdownMenu.prototype._itemClicked = function(pageId) {
	this.hide();
	this.setSelectedDocumentPage(pageId);
	this._onDocChange(pageId);
};

/**
 * Disable a document in the dropdown list, the document will no longer be visible
 * @param {String} pageId
 */
DocumentDropdownMenu.prototype.disableDocumentPage = function(pageId, hideGroupIfAllPagesDisabled) {
	//needed to check if it wasn't visible before, make sure the enabledDocuments count stays correct
	if (this.isElementVisible(pageId)) {
		this.hideElement(pageId);
		this._enabledDocuments--;
		if (this.emptyListElement !== undefined && this._enabledDocuments <= 0) {
			this.showElement(this.emptyListElement.name);
			this.hideElement(this.dividerId);
		}
	}
	if (hideGroupIfAllPagesDisabled && this.items[pageId] && this.items[pageId].group) {
		_.chain(this._elementGrouping[this.items[pageId].group])
			.reduce(
				function(memo, pageId) {
					return (memo === null || memo) && !this.isElementVisible(pageId);
				}.bind(this),
				null
			)
			.tap(
				function(allPagesHidden) {
					if (allPagesHidden) {
						this.disableDocumentPage(this.items[pageId].group);
					}
				}.bind(this)
			);
	}
};

/**
 * Enable a document in the dropdown list
 * @param {String} pageId
 */
DocumentDropdownMenu.prototype.enableDocumentPage = function(pageId) {
	//needed to check if it wasn't visible before, make sure the enabledDocuments count stays correct
	if (this.isElementVisible(pageId) !== undefined && !this.isElementVisible(pageId)) {
		//test if either the element is in no group or the pages that needs to be enabled belongs to the active group
		if (this.items[pageId].group === undefined || this.isGroupOpen(this.items[pageId].group)) {
			this.showElement(pageId);
			this._enabledDocuments++;
			if (this.emptyListElement !== undefined && this._enabledDocuments > 0) {
				this.hideElement(this.emptyListElement.name);
				this.showElement(this.dividerId);
			}
		}

		if (this.items[pageId] && this.items[pageId].group) {
			this.enableDocumentPage(this.items[pageId].group);
		}
	}
};

/**
 * Returns the unique id of the currently selected document page
 * @returns {String|undefined} - unique id of the document page or undefined if there is no selected document page
 */
DocumentDropdownMenu.prototype.getSelectedDocumentPageId = function() {
	if (this._selectedDocumentPage !== null && this._selectedDocumentPage !== this.deselectElement) {
		return this._selectedDocumentPage.name;
	} else {
		return undefined;
	}
};

/**
 * deselect the currently selected document page
 * @param {Boolean} enableDocument - should the deselected document be added to the dropdownlist
 */
DocumentDropdownMenu.prototype.deselectCurrentPage = function(enableDocument) {
	if (enableDocument && this._selectedDocumentPage !== null) {
		this.enableDocumentPage(this._selectedDocumentPage.name);
	}
	if (this._deselectText !== "" && this._nothingSelectedText !== "") {
		this._selectedDocumentPage = this.deselectElement;
		this.disableDocumentPage(this.deselectElement.name);
		this.deselectElement.content.html(this._nothingSelectedText);
		this._selectedDocumentElement.html(this._selectedDocumentPage.content.clone());
	} else {
		this._selectedDocumentElement.html("");
		this._selectedDocumentPage = null;
	}
};

/**
 * Hide all the elements belonging to a group, this will active the group element if there is no enabled document type
 * @param {String} groupName
 */
DocumentDropdownMenu.prototype.hideElementGroup = function(groupName) {
	if (groupName !== undefined) {
		if (this._listElements[groupName] !== undefined) {
			this._listElements[groupName]
				.first()
				.find(".showAllPages")
				.show();
			this._listElements[groupName]
				.first()
				.find(".hideAllPages")
				.hide();
		} else {
			this.items[groupName].content.find(".showAllPages").show();
			this.items[groupName].content.find(".hideAllPages").hide();
		}
		this.items[groupName].isOpen = false;
		_.each(
			this._elementGrouping[groupName],
			_.bind(function(groupElementName) {
				this.disableDocumentPage(groupElementName); //first hide all the members of the group
			}, this)
		);
		this.enableDocumentPage(groupName);
	}
};

/**
 * Shows all the elements belonging to a group, this will hide the group element
 * @param {String} groupName
 */
DocumentDropdownMenu.prototype.showElementGroup = function(groupName) {
	if (groupName !== undefined) {
		if (this._listElements[groupName] !== undefined) {
			this._listElements[groupName]
				.first()
				.find(".showAllPages")
				.hide();
			this._listElements[groupName]
				.first()
				.find(".hideAllPages")
				.show();
		} else {
			this.items[groupName].content.find(".showAllPages").hide();
			this.items[groupName].content.find(".hideAllPages").show();
		}
		this.items[groupName].isOpen = true;
		this._activeGroup = groupName;
		_.each(
			this._elementGrouping[groupName],
			_.bind(function(groupElementName) {
				this.enableDocumentPage(groupElementName);
			}, this)
		);
	}
	this._unveil();
};

/**
 * Hides all elements belonging to a group, will show all the group elements
 */
DocumentDropdownMenu.prototype.hideAllGroups = function() {
	_.each(
		this._elementGrouping,
		_.bind(function(group, groupName) {
			this.hideElementGroup(groupName);
		}, this)
	);
};

/**
 * Has the dropdown menu the document page with docId
 * @param {String} docId
 * @returns {Boolean}
 */
DocumentDropdownMenu.prototype.hasDocumentPage = function(docId) {
	return this.items[docId] !== undefined;
};

/**
 * Has the dropdown menu the document group with groupId
 * @param {String} groupId
 * @returns {Boolean}
 */
DocumentDropdownMenu.prototype.hasDocumentGroup = function(groupId) {
	return this.items[groupId] !== undefined;
};

/**
 * Does the dropdown menu have any groups
 * @returns {Boolean}
 */
DocumentDropdownMenu.prototype.hasGroups = function() {
	return !_.isEmpty(this._elementGrouping);
};

/**
 * Is the group with group name open
 * @param {String} groupName
 * @returns {Boolean}
 */
DocumentDropdownMenu.prototype.isGroupOpen = function(groupName) {
	return this.items[groupName].isOpen;
};

/**
 * Show the document dropdown menu
 */
DocumentDropdownMenu.prototype.show = function() {
	DropdownMenu.prototype.show.call(this); //show the drop down menu
	this._unveil(); //unveil thumbnails if needed
	if (!this.scrollListenerRegistered) {
		this.popover.getContentDiv().on("scroll", _.bind(this.unveil, this)); //register scroll listener
		this.scrollListenerRegistered = true;
	}
};

/**
 * Checks if the dropdown menu is open and gets all the elements that are visible in the list. If there are visible
 * elements where the thumbnail has not yet been unveiled sets the src-element to the correct value so the document
 * selector requests the thumbnail.
 *
 * @param {Boolean} [keepList] - should the previously constructed list of visible elements of the document selector be used, this happens when
 * 								no elements become visible/hidden (when scrolling for example)
 */
DocumentDropdownMenu.prototype._unveil = function(keepList) {
	//first order all the elements of the lsit so all the elements are in order of appearance in the docSelector
	//if the list doesn't exist yet(is empty) or the existing list shouldn't be used the check to unveil thumbnails, construct the list
	if (this.indexList === undefined || this.indexList.length === 0 || !keepList) {
		var orderedListElements = [];

		//run over all the elements that have been added to the DOM
		_.each(
			this._listElements,
			_.bind(function(element, key) {
				var item = this.items[key]; //get the item that contains the data of this DOM-element
				if (item.isVisible) {
					//check if the element in the document controller is visible
					var groupIndex;
					if (item.type === "group") {
						//if the element is a group
						groupIndex = element.closest(".group").index(); // The closest .group element is the encompassing group element
						orderedListElements[groupIndex] = []; //add an array at the location to fill with all elements of this group
						orderedListElements[groupIndex][0] = item; //add the group element to the array as first element
					} else if (item.group !== undefined) {
						//the element is part of a group and index is relative to the div-element of the group
						groupIndex = element.closest(".group").index();
						orderedListElements[groupIndex][element.index()] = item; //add the element to the array of the group element
					} else {
						//the element is not a group nor belonging to a group
						orderedListElements[element.index()] = item;
					}
				}
			}, this)
		);
		//to get all the elements as a 1 level deep array, flatten the array so elements of groups are placed at the correct location
		this.indexList = _.without(_.flatten(orderedListElements), undefined, null); //array in the correct order of appearance in the document selector
	}
	//only if there are elements in the list, document selector is not empty continue
	if (this.indexList.length > 0) {
		var scrollValue = 0; //how many pixels are the top of the element (not visible due to scroll)
		if (this.popover !== undefined) {
			//check if the dropdown is open and the popover is visible
			scrollValue = this.popover.getContentDiv().scrollTop();
		}
		var elementHeight = Math.ceil((this.menu.outerHeight(true) || null) / this.indexList.length); // we can't rely on outerheight of a single element (it could be a group, and margins are collapsed on eachother), we do now how many elements we have and how much space they take in total :)
		var startIndex = Math.floor(scrollValue / elementHeight); //start index of the visible elements, how many elements are above and not visible

		var visibleHeight = this.popover.getContentDiv().innerHeight() || null; //calculate the height of the dropdown menu
		var endIndex = startIndex + Math.ceil(visibleHeight / elementHeight); //get the index of the last visible element
		if (endIndex > this.indexList.length - 1) {
			//make sure we do not move outside the indexlist array when going over visible elements
			endIndex = this.indexList.length - 1;
		}

		//run over all the visible elements
		for (var i = startIndex; i <= endIndex; i++) {
			var item = this.indexList[i];
			if (item) {
				var listElement = this._listElements[item.name];
				if (!item.isUnveiled && listElement !== undefined) {
					//check whether the thumbnail is already unveiled
					listElement.find(".thumbnail").attr("src", item.thumb); //unveil thumbnail
					item.isUnveiled = true;
				}
			}
		}
	}
};

/**
 * Unveil thumbnails if this is needed.
 */
DocumentDropdownMenu.prototype.unveilThumbnails = function() {
	this._unveil();
};

/**
 * Get an ordered list of the items as they appear in the dropdown menu
 * @returns {Array.<Object>}
 * @private
 */
DocumentDropdownMenu.prototype._getOrderedItemList = function() {
	if (this.menu) {
		//the menu list was openend, all items are in the DOM,
		//use this since opened groups add documents to the end fo the itemStack
		return _.map(
			this.menu.find("li[data-name]"),
			function(itemElement) {
				return this.items[$(itemElement).attr("data-name")];
			},
			this
		);
	} else {
		return this.itemStack;
	}
};

/**
 * Get the page id of the first selectable item in the dropdown list
 * @returns {String|null} Page Id, e.g. "0_1_2"
 */
DocumentDropdownMenu.prototype.getFirstSelectablePageId = function() {
	var listItems = this._getOrderedItemList();
	var index = 1;
	var item = listItems[index];
	while (index < listItems.length && item && item.isGroup) {
		item = listItems[++index];
	}
	if (index < listItems.length && item) {
		return item.name;
	} else {
		return null;
	}
};

/**
 * Get the page id of the last selectable item in the dropdown list
 * @returns {String|null} Page Id, e.g. "0_1_2"
 */
DocumentDropdownMenu.prototype.getLastSelectablePageId = function() {
	var listItems = this._getOrderedItemList();
	var item = _.last(listItems);
	var index = _.indexOf(listItems, item);
	while (index > 0 && item && item.isGroup) {
		item = listItems[--index];
	}
	if (item && index > 0) {
		return item.name;
	} else {
		return null;
	}
};

/**
 * Get the page id of the next selectable item, relative to the currently active document page, in the dropdown list
 * @returns {String|null} Page Id, e.g. "0_1_2"
 */
DocumentDropdownMenu.prototype.getNextItemId = function() {
	var listItems = this._getOrderedItemList();
	var selectedPageId = this.getSelectedDocumentPageId();
	var index = _.indexOf(listItems, this.items[selectedPageId]);
	var item = listItems[++index];
	while (index < listItems.length && item && item.isGroup) {
		item = listItems[++index];
	}
	if (item && index < listItems.length) {
		return item.name;
	} else {
		return null;
	}
};

/**
 * Get the page id of the previous selectable item, relative to the currently active document page, in the dropdown list
 * @returns {String|null} Page Id, e.g. "0_1_2"
 */
DocumentDropdownMenu.prototype.getPreviousItemId = function() {
	var listItems = this._getOrderedItemList();
	var selectedPageId = this.getSelectedDocumentPageId();
	var index = _.indexOf(listItems, this.items[selectedPageId]);
	var item = listItems[--index];
	while (index > 0 && item && item.isGroup) {
		item = listItems[--index];
	}
	if (item && index > 0) {
		return item.name;
	} else {
		return null;
	}
};

export default DocumentDropdownMenu;
