import PrivateChannel from "@esko/web-viewer/core/PrivateChannel";
import DocumentDropdownMenu from "@esko/web-viewer/layout/component/DocumentDropdownMenu";
import ListenerUtil from "@esko/web-viewer/utils/ListenerUtil";
import DataConstants from "@esko/web-viewer/data/DataConstants";
import _ from "underscore";
/**
 * This component will take care of document selection.
 * It contains a document selector for active documents and for difference document.
 *
 * @exports DocumentSelection
 * @memberof Layout.Component
 * @constructor
 * @param {Core.ViewerFacade} f - The viewer facade
 * @param {Compare.CompareModule} [compareModule] - compareModule, needed for difference document selector
 * @param {Function} [onDocChangeCallback] - Called when active or difference document is changed
 * @param {Boolean} [isMobile] - is mobile layout
 * @param {Boolean} [isCompact] - if true version number is put behind the document (on the same line)
 */
var DocumentSelection = function(f, compareModule, onDocChangeCallback, isMobile, isCompact) {
	PrivateChannel.installTo(this, f, "documentSelection");
	this.f = f;
	ListenerUtil.mixin(this, f);

	//makes sure callbacks are only registered once when new documents are added
	this._callBacksAreRegistered = false;

	/**
	 * @Member {Function}
	 */
	this._onDocChangeCallback = onDocChangeCallback;

	/**
	 * @Member {compare.CompareModule}
	 */
	this.compareModule = compareModule;

	/**
	 * @member {DocumentDropdownMenu} dropdown menu for selecting the active document page
	 */
	this.activeDocPageSelector = null;

	/**
	 * @member {DocumentDropdownMenu} dropdown menu for selecting the difference document page
	 */
	this.differenceDocPageSelector = null;

	/**
	 * @member {DocumentDropdownMenu} dropdown menu for selecting the CAD overlay
	 */
	this.cadOverLaySelector = null;

	this.f.i18n.loadBundle(
		"viewer",
		_.bind(function() {
			this.activeDocPageSelector = new DocumentDropdownMenu(
				f,
				_.bind(this.changeActivePage, this),
				this.f.t("EmptyDocumentList"),
				undefined,
				undefined,
				undefined,
				isMobile,
				isCompact,
				{ sahiId: "viewer-active-page-selector" }
			);
			this.activeDocPageSelector.element.attr("data-sahi-id", "viewer-active-page-selector");
			if (this.compareModule) {
				this.differenceDocPageSelector = new DocumentDropdownMenu(
					f,
					_.bind(this.changeDifferencePage, this),
					this.f.t("EmptyDocumentList"),
					this.f.t("EmptyDifferenceDocSelector"),
					this.f.t("DocumentListCompareTurnOff"),
					_.bind(function(deselectedPageId) {
						this.activeDocPageSelector.enableDocumentPage(deselectedPageId);

						// disable compare
						this.compareModule.activateModeByName();
					}, this),
					isMobile,
					isCompact,
					{ sahiId: "viewer-difference-page-selector" }
				);
				this.differenceDocPageSelector.element.attr("data-sahi-id", "viewer-difference-page-selector");

				this.cadOverLaySelector = new DocumentDropdownMenu(
					f,
					_.bind(this.changeCadOverlay, this),
					this.f.t("EmptyDocumentList"),
					this.f.t("EmptyCadOverlayDocSelector"),
					this.f.t("DocumentListCadOverlayTurnOff"),
					_.bind(function(deselectedPageId) {
						// turn of CAD overlay
						this.f.getDocumentController().changeCadOverlay();
					}, this),
					isMobile,
					isCompact
				);
			}

			this.f.getDocumentController().whenDocumentsLoadedOrChange(this._whenDocumentsLoaded, this);
		}, this)
	);

	this.f.on(
		"selection.requestdoc",
		_.bind(function(doc) {
			this.changeActivePageImpl(doc.getDocumentVersion().getFirstPage());
			if (this.differenceDocPageSelector) {
				this.setDifferenceSelectorPage();
				this.compareModule.activateModeByName();
			}
		}, this)
	);
};

/**
 * Get active document selector
 * @return {DocumentDropdownMenu}
 */
DocumentSelection.prototype.getActiveDocumentSelector = function() {
	return this.activeDocPageSelector;
};

/**
 * Get difference document selector
 * @return {DocumentDropdownMenu}
 */
DocumentSelection.prototype.getDifferenceDocumentSelector = function() {
	return this.differenceDocPageSelector;
};

/**
 * Get the CAD overlay document selector
 * @returns {DocumentDropdownMenu} /
 */
DocumentSelection.prototype.getCadOverlayDocumentSelector = function() {
	return this.cadOverLaySelector;
};

/**
 * Changes the active document page
 * @param {String} id - unique id of the document page
 */
DocumentSelection.prototype.changeActivePage = function(id) {
	var newActivePage = this.f.getDocumentController().getDocumentPage(id);
	this.changeActivePageImpl(newActivePage);
};

/**
 * Changes the active document page
 * @param {String} id - unique id of the document page
 */
DocumentSelection.prototype.changeActivePageImpl = function(newActivePage) {
	var activePage = this.f.getDocumentController().getActivePage();

	this._emit(
		"changeActivePage",
		newActivePage,
		_.bind(function(isChanged) {
			if (isChanged && this.differenceDocPageSelector) {
				if (activePage !== undefined) {
					var pageId = activePage.getUniqueID();
					//make sure that the previous active document page can be selected as difference page
					this.differenceDocPageSelector.enableDocumentPage(pageId);
				}
				// reset difference when there is another type of document selected
				if (activePage.getDocument().getDocumentType() !== newActivePage.getDocument().getDocumentType()) {
					this.activeDocPageSelector.enableDocumentPage(this.differenceDocPageSelector.getSelectedDocumentPageId());
					this.compareModule.activateModeByName();
				}

				this._setPagesOfDifferenceSelector();
			}

			if (!isChanged) {
				this.activeDocPageSelector.setSelectedDocumentPage(this.f.getActivePage().getUniqueID());
			}
		}, this)
	);
};

/**
 * shows/hides the correct pages and grouds for the document selector based on the current active page
 */
DocumentSelection.prototype._setPagesOfDifferenceSelector = function() {
	var newActivePage = this.f.getDocumentController().getActivePage();
	// all available pages
	var pages = _.chain(this.f.getDocumentController().getAllDocuments())
		.map(function(doc) {
			// get document versions
			return doc.getDocumentVersions();
		})
		.flatten()
		.map(function(version) {
			// map to pages
			return version.getAllPages();
		})
		.flatten()
		.without(newActivePage)
		.value();

	var comparablePages = _.filter(pages, function(page) {
		return _.intersection(newActivePage.getAvailableCompareModes(page), page.getAvailableCompareModes(newActivePage)).length > 0;
	});

	var nonComparablePages = _.difference(pages, comparablePages);

	_.each(
		nonComparablePages,
		function(page) {
			this.differenceDocPageSelector.disableDocumentPage(page.getUniqueID(), true);
		},
		this
	);

	_.chain(comparablePages)
		.map(function(page) {
			return page.getDocumentVersion();
		})
		.flatten()
		.uniq()
		.filter(function(docVersion) {
			return docVersion.getPageCount() > 1;
		})
		.each(
			function(docVersion) {
				var uniqueId = docVersion.getFirstPage().getUniqueID();
				var item = this.differenceDocPageSelector.items[uniqueId];
				if (item && item.group) {
					this.differenceDocPageSelector.hideElementGroup(item.group);
				}
			}.bind(this)
		);

	_.each(
		comparablePages,
		function(page) {
			this.differenceDocPageSelector.enableDocumentPage(page.getUniqueID());
		},
		this
	);

	//the new active page can't be selected as difference document page
	this.differenceDocPageSelector.disableDocumentPage(newActivePage.getUniqueID());
};

/**
 * changes the difference document page
 * @param {String} id - unique id of the document page
 */
DocumentSelection.prototype.changeDifferencePage = function(id) {
	var differencePage = this.f.getDocumentController().getDifferencePage();

	if (differencePage !== undefined) {
		var pageId = differencePage.getUniqueID();
		//make sure that the previous difference document page can be selected as active page
		this.activeDocPageSelector.enableDocumentPage(pageId);
	}
	var newDiffPage = undefined;
	if (id) {
		this.activeDocPageSelector.disableDocumentPage(id);
		newDiffPage = this.f.getDocumentController().getDocumentPage(id);
	}
	this._emit("changeReferencePage", newDiffPage);
};

/**
 * Changes the cad voerlay
 * @param {String} id - unique id of the document page
 */
DocumentSelection.prototype.changeCadOverlay = function(id) {
	var docController = this.f.getDocumentController();
	var cadOverlayPage = docController.getDocumentPage(id);
	this._emit("changeCadOverlay", cadOverlayPage);
};

/**
 * Called when documents are loaded or when new documents are added
 * @param {Array.<AbstractDocument>} documents
 */
DocumentSelection.prototype._whenDocumentsLoaded = function(documents) {
	//this function is called when later on new documents are added or documents change (new version, new pages)
	//so make sure it takes this into account (check if page or group was already in the dropdown)

	var thumbnailPromises = [];
	_.each(
		documents,
		_.bind(function(document) {
			var docVersions = _.chain(document.getAllDocumentVersions())
				.filter(function(docVersion) {
					return docVersion && docVersion.getDocumentVersionID();
				})
				.sortBy(function(docVersion) {
					return docVersion.getVersionId() * -1;
				})
				.value();
			_.each(
				docVersions,
				_.bind(function(docVersion) {
					var groupName;
					var isGroupNeeded = docVersion.isMultiPage() || docVersions.length > 1;
					if (isGroupNeeded) {
						if (docVersion.isMultiPage()) {
							groupName = docVersion.getDocument().getIID() + "_" + docVersion.getVersionId() + "_Group";
						} else {
							groupName = docVersion.getDocument().getIID() + "_Group";
						}
						if (!this.activeDocPageSelector.hasDocumentGroup(groupName)) {
							//add a document group if it wasn't in the dropdown before, call loadPagesIfNeeded as callback to check wether pages still need to be loaded
							this.activeDocPageSelector.addDocumentGroup(
								groupName,
								// before show/hide
								_.bind(this._loadPagesIfNeeded, this, docVersion),
								// after show/hide
								_.bind(function() {
									// always disable the current visible active/diff page again if the whole group was shown again
									this.activeDocPageSelector.disableDocumentPage(this.activeDocPageSelector.getSelectedDocumentPageId());
									if (this.differenceDocPageSelector) {
										this.activeDocPageSelector.disableDocumentPage(this.differenceDocPageSelector.getSelectedDocumentPageId());
									}
								}, this)
							);
						}
						if (this.differenceDocPageSelector && !this.differenceDocPageSelector.hasDocumentGroup(groupName)) {
							this.differenceDocPageSelector.addDocumentGroup(
								groupName,
								// before show/hide
								_.bind(this._loadPagesIfNeeded, this, docVersion),
								// after show/hide
								_.bind(function() {
									// always disable the current visible active/diff page again if the whole group was shown again
									this.differenceDocPageSelector.disableDocumentPage(this.differenceDocPageSelector.getSelectedDocumentPageId());
									this.differenceDocPageSelector.disableDocumentPage(this.activeDocPageSelector.getSelectedDocumentPageId());
								}, this)
							);
						}
					}

					var pageCount = docVersion.getPageCount();
					_.range(pageCount).forEach(
						function(i) {
							var docPage = docVersion.getPage(i + 1);
							if (docPage !== undefined) {
								// this support both promises and non-promises
								var pageInfoP = Promise.resolve(docPage.getInformationForDocumentSelector());
								thumbnailPromises.push(
									pageInfoP.then(function(pageInfo) {
										return { pageInfo: pageInfo, groupName: groupName, docPage: docPage };
									})
								);
							}
						}.bind(this)
					);
				}, this)
			);
		}, this)
	);
	//after elements have been added, unveil thumbnails if needed (doc selector might be open)
	Promise.all(thumbnailPromises)
		.then(
			function(pageInfos) {
				pageInfos.forEach(
					function(pageInfo) {
						var groupName = pageInfo.groupName;
						var docPage = pageInfo.docPage;
						pageInfo = pageInfo.pageInfo;
						if (!this.activeDocPageSelector.hasDocumentPage(pageInfo.pageId)) {
							this.activeDocPageSelector.addDocumentPage(pageInfo, groupName);
						}
						if (
							docPage.canBeComparedTo() &&
							this.differenceDocPageSelector &&
							!this.differenceDocPageSelector.hasDocumentPage(pageInfo.pageId) &&
							!docPage.isBlankPage()
						) {
							this.differenceDocPageSelector.addDocumentPage(pageInfo, groupName);
						}

						if (this.cadOverLaySelector) {
							if (docPage.hasCad() || docPage.is3D()) {
								this.cadOverLaySelector.addDocumentPage(pageInfo);
							} else if (docPage.hasLinkedPage()) {
								var linkedCadDocInfo = docPage.getLinkedPage().getInformationForDocumentSelector();
								linkedCadDocInfo.referencedFrom = docPage.getDocument().get("docname");
								this.cadOverLaySelector.addDocumentPage(linkedCadDocInfo);
							}

							_.each(
								docPage.getLinkedPages(DataConstants.LINKEDDOCUMENTTYPES.THREEDVIEW),
								function(linkedPage) {
									var linkedInfo = linkedPage.page.getInformationForDocumentSelector();
									linkedInfo.referencedFrom = docPage.getDocument().get("docname");
									this.cadOverLaySelector.addDocumentPage(linkedInfo);
								},
								this
							);
						}
					}.bind(this)
				);
			}.bind(this)
		)
		.then(
			function() {
				this.activeDocPageSelector.unveilThumbnails();
				if (this.differenceDocPageSelector) {
					this.differenceDocPageSelector.unveilThumbnails();
					this._setPagesOfDifferenceSelector();
				}
				this._registerCallbacksIfNeeded();
			}.bind(this)
		);
};

/**
 * Registers the callbacks when the document selectors are initialized, makes sure they are only registered once.
 * @private
 */
DocumentSelection.prototype._registerCallbacksIfNeeded = function() {
	if (!this._callBacksAreRegistered) {
		//To make sure first the document selectors are filled with data before the active page is initialized
		this.f.getDocumentController().whenActivePageChangedAndParametersAvailable(function(oldPage, page) {
			//check if the active document page was changed from somewhere else than the document selector
			if (
				this.activeDocPageSelector.getSelectedDocumentPageId() === undefined ||
				page.getUniqueID() !== this.activeDocPageSelector.getSelectedDocumentPageId()
			) {
				if (page.getDocumentVersion().isMultiPage()) {
					this.activeDocPageSelector.hideAllGroups();
					var groupName = page.getDocument().getIID() + "_" + page.getDocumentVersion().getVersionId() + "_Group";
					this.activeDocPageSelector.showElementGroup(groupName);
				}
				this.activeDocPageSelector.setSelectedDocumentPage(page.getUniqueID());

				if (this.differenceDocPageSelector) {
					this.differenceDocPageSelector.disableDocumentPage(page.getUniqueID());
					this.differenceDocPageSelector.hideAllGroups();
					this._setPagesOfDifferenceSelector();

					if (page.getDocumentVersion().isPageList()) {
						_.each(
							page.getDocument().getAllDocumentVersions(),
							function(docVersion) {
								if (docVersion) {
									var diffPage = docVersion.getPage(page.getPagenum());
									if (diffPage) {
										this.differenceDocPageSelector.enableDocumentPage(diffPage.getUniqueID());
									}
								}
							},
							this
						);
					}

					//for the first active page make sure to load pages of previous version for pagelist documents
					//only for this case is the oldPage === undefined
					if (oldPage === undefined && page.getDocumentVersion().isPageList()) {
						this.updateDocumentSelectorsForPageList(undefined, page);
					}
				}
			}

			if (this._onDocChangeCallback) {
				this._onDocChangeCallback();
			}
		}, this);
		if (this.differenceDocPageSelector) {
			this.f.getDocumentController().whenDifferencePageChangedAndParametersAvailable(function(oldPage, page) {
				//deselect the current page in the difference selector if reader spread is active (enabled) and a page is selected in the doc selector
				var isReaderSpread = this.compareModule.isReaderSpreadActive();
				var activeMode = this.compareModule.getActiveMode();
				if (isReaderSpread && this.differenceDocPageSelector.getSelectedDocumentPageId() !== undefined && activeMode.isEnabled()) {
					this.differenceDocPageSelector.deselectCurrentPage(true);
					//if there is no active mode or the active mode is not enabled or the active mode is not reader spread
				} else if (activeMode === null || !activeMode.isEnabled() || !isReaderSpread) {
					if (page === undefined) {
						// set selected difference page to not selected
						this.activeDocPageSelector.enableDocumentPage(this.differenceDocPageSelector.getSelectedDocumentPageId());
						this.differenceDocPageSelector.deselectCurrentPage(false);
						this._setPagesOfDifferenceSelector();
					} else if (page.isBlankPage()) {
						// set selected difference page to not selected
						//this.differenceDocPageSelector.deselectCurrentPage(true);
						if (oldPage !== undefined) {
							this.activeDocPageSelector.enableDocumentPage(oldPage.getUniqueID());
						}
						this.differenceDocPageSelector.setTempSelectedDocumentPage(page.getInformationForDocumentSelector());
						this.f.emit("error.WARNING", this.f.i18n.t("ComparingWithBlankPage"));
					} else {
						//check if the difference document page is changed from somewhere other than the document selector
						if (
							this.activeDocPageSelector.getSelectedDocumentPageId() === undefined ||
							page.getUniqueID() !== this.differenceDocPageSelector.getSelectedDocumentPageId()
						) {
							this.differenceDocPageSelector.setSelectedDocumentPage(page.getUniqueID());
							this.activeDocPageSelector.disableDocumentPage(page.getUniqueID());
							if (oldPage) {
								this.activeDocPageSelector.enableDocumentPage(oldPage.getUniqueID());
							}
						}
					}
				}
				if (this._onDocChangeCallback) {
					this._onDocChangeCallback();
				}
			}, this);
		}

		if (this.cadOverLaySelector) {
			this.f.getDocumentController().whenCadOverlayChangedAndParametersAvailable(function(oldPage, newPage) {
				this.cadOverLaySelector.deselectCurrentPage(true);
				if (newPage) {
					this.cadOverLaySelector.setSelectedDocumentPage(newPage.getUniqueID());
				}

				if (this._onDocChangeCallback) {
					this._onDocChangeCallback();
				}
			}, this);
		}

		this._callBacksAreRegistered = true;
	}
};

/**
 * Set difference selector page to the given diff page. If undefined, the current page is de-selected and again added to the list of possible difference pages.
 * If reader-spread is active, the page is deselected and added to the list also if diffpage is defined.
 * @param {AbstractDocumentPage} [diffPage] - Diff page to set in the difference selector.
 */
DocumentSelection.prototype.setDifferenceSelectorPage = function(diffPage) {
	if (this.differenceDocPageSelector) {
		if (diffPage === undefined) {
			this.differenceDocPageSelector.deselectCurrentPage(true);
		} else {
			if (this.compareModule.isReaderSpreadActive()) {
				this.differenceDocPageSelector.deselectCurrentPage(true);
				this.activeDocPageSelector.enableDocumentPage(diffPage.getUniqueID());
			} else {
				this.differenceDocPageSelector.setSelectedDocumentPage(diffPage.getUniqueID());
				this.activeDocPageSelector.disableDocumentPage(diffPage.getUniqueID());
			}
		}
	}
};

/**
 * Update the document selectors if the selected active document page is part of a page list document. Make sure to test if the document is
 * a pagelist before calling this method! Since part is asynchronous, call this method before disabling/changing compare mode as takes
 * it into account when changing to the correct difference document page. This only happens if compare was already active.
 * @param {AbstractDocumentPage|undefined} activePage - if not provided, no previous pages will be disabled in the difference document selector
 * @param {AbstractDocumentPage} newActivePage - the new active document page, used to enable the correct document pages in the difference document selector
 */
DocumentSelection.prototype.updateDocumentSelectorsForPageList = function(oldActivePage, newActivePage) {
	if (oldActivePage !== undefined) {
		var pageNumber = oldActivePage.getPagenum();
		var document = oldActivePage.getDocument();
		//run over all previous document versions of the document, to disable all pages that were enabled for the old active document page
		for (var i = 1; i < oldActivePage.getDocumentVersion().getVersionId(); i++) {
			var documentVersion = document.getDocumentVersion(i);
			if (documentVersion !== undefined && documentVersion.hasPage(pageNumber)) {
				//it is possible that the document version/page was not (yet) loaded
				var pageID = documentVersion.getPage(pageNumber).getUniqueID(); //unique id is used as key for the document selector
				this.differenceDocPageSelector.removeDocumentPage(pageID); //disable the document page
			}
		}
	}
	//to make sure that the correct document page is set as difference page, if before changing active document page, a compare mode was active
	//this is needed because of asynchronous call to load version of the new active document page
	var wasComparing = false;
	if (this.compareModule) {
		wasComparing =
			this.compareModule.getActiveMode() !== null && !this.compareModule.isReaderSpreadActive() && !newActivePage.getLinkedComparePage();
	}

	//call to the server to request previous version of the document that has pages with the same page number as the active document page
	newActivePage.getDocument().loadVersionsForDocumentPage(
		newActivePage,
		_.bind(function(pages) {
			//execute callback for when the pages have been added to the document controller
			_.each(
				pages,
				_.bind(function(page) {
					//run over each pages to add to the difference document selector
					if (page) {
						if (this.differenceDocPageSelector.hasDocumentPage(page.getUniqueID())) {
							//first check if the page wasn't already added to the document selector
							this.differenceDocPageSelector.enableDocumentPage(page.getUniqueID()); //enable if that is the case
						} else {
							var docInfo = page.getInformationForDocumentSelector();
							docInfo.prepend = true; //pages need to be prepended to make sure they are added to the beginning of the document selector
							docInfo.docType = DataConstants.documentPageType.PAGELIST; //change docType to PageList to make sure they are not hidden by the document selector
							docInfo.pageCount = null; //set pageCount to null, this is not correct for page list documents as they don't load all pages of a version at the same time
							docInfo.pageNumber = undefined;
							this.differenceDocPageSelector.addDocumentPage(docInfo);
						}
					}
				}, this)
			);
			//check if the comparing was active before, if so set the difference document page to the same page of the last version
			if (wasComparing) {
				this.differenceDocPageSelector.setSelectedDocumentPage(_.last(pages).getUniqueID()); //_.last gets the page of the last version
				this.changeDifferencePage(_.last(pages).getUniqueID());
			}
		}, this)
	);
};

/**
 * Load pages of the given document version if these pages are not yet loaded, this to prevent loading at initialization
 * @private
 * @param {AbstractDocumentVersion} docVersion - document version to load pages for
 */
DocumentSelection.prototype._loadPagesIfNeeded = function(docVersion) {
	if (docVersion.isMultiPage() && docVersion.getPageCount() <= 1) {
		this.f.getDocumentController().loadDocumentPages(docVersion.getDocument().getIID(), docVersion.getVersionId());
	}
};

/**
 * Add a newly created document page in the cad overlay selector and remove an old cad overlay (linked document pages).
 * It makes the newly added document page as the active (selected) document page.
 * @param {AbstractDocumentPage} graphicsPage
 * @param {String} oldCadOverlayId
 */
DocumentSelection.prototype.setNewlyCreatedCadOverlay = function(graphicsPage, oldCadOverlayId) {
	// Only remove the old cad overlay when it's not in the active document selector too
	if (oldCadOverlayId && !this.activeDocPageSelector.hasDocumentPage(oldCadOverlayId)) {
		this.cadOverLaySelector.removeDocumentPage(oldCadOverlayId);
	}

	var linkedCadPage = graphicsPage.getLinkedPage();
	if (linkedCadPage) {
		var linkedCadDocInfo = linkedCadPage.getInformationForDocumentSelector();
		linkedCadDocInfo.referencedFrom = graphicsPage.getDocument().get("docname"); //make sure the link is displayed in the document selector
		this.cadOverLaySelector.addDocumentPage(linkedCadDocInfo);
		if (this.f.getDocumentController().getCadOverlay()) {
			this.cadOverLaySelector.setSelectedDocumentPage(linkedCadDocInfo.pageId);
		}
	}
};

/**
 * Get the first selectable page id
 * @returns {String|null}
 */
DocumentSelection.prototype.getFirstSelectablePageId = function() {
	return this.activeDocPageSelector.getFirstSelectablePageId();
};

/**
 * Get the last selectable page id
 * @returns {String|null}
 */
DocumentSelection.prototype.getLastSelectablePageId = function() {
	return this.activeDocPageSelector.getLastSelectablePageId();
};

/**
 * Get the next selectable page id, relative to the currently selected page in the active page selector
 * @returns {String|null}
 */
DocumentSelection.prototype.getNextSelectablePageId = function() {
	return this.activeDocPageSelector.getNextItemId();
};

/**
 * Get the previous selectable page id, relative to the currently selected page in the active page selector
 * @returns {String|null}
 */
DocumentSelection.prototype.getPreviousSelectablePageId = function() {
	return this.activeDocPageSelector.getPreviousItemId();
};

export default DocumentSelection;
