import MiniButton from "@esko/web-viewer/layout/component/MiniButton";
import DropdownMenu from "@esko/dropdownmenu/DropdownMenu";
import $ from "jquery";
import _ from "underscore";
/**
 *
 * A DropdownMiniButton is a Minibutton that when clicked shows a dropdownmenu.
 *
 * @exports Layout.Component.DropdownMiniButton
 * @constructor
 * @param {Core.ViewerFacade} f
 * @param {String} name - the name of the button
 * @param {String} icon - the icon of the button (css class for 16px * 16px icon)
 * @param {Array.<String>} [restrictedPositions] - array with all the positions that are allowed
 * @param {Function} [leftClickAction] - Action executed when left clicking. The dropdown will be shown on right click. If undefined the dropdown will be shown when left clicking.
 * Parameters are a boolean indicating whether the button is activated (true) or deactivated (false). The button itself is NOT ACTIVATED!
 */
var DropdownMiniButton = function(f, name, icon, restrictedDropdownPositions, leftClickAction) {
	MiniButton.call(this, f, name, icon, false, this.toggle);
	this.element.addClass("dropdown-mini-button");

	this.dropdown = new DropdownMenu(
		this.f.i18n,
		this.f.getContext(),
		this.f.isMobile(),
		"bottom",
		"wcr-dropdown",
		undefined,
		restrictedDropdownPositions
	);
	this._dropdownEnabled = true;

	this.leftClickAction = leftClickAction;
	if (this.leftClickAction === undefined) {
		// only left
		this.leftClickAction = _.bind(function() {
			if (this.isEnabled() && this._dropdownEnabled) {
				if (!this.dropdown.isActive()) {
					this.showDropdown();
				} else {
					this.dropdown.hide();
				}
			}
		}, this);
	} else {
		// we have both right/left
		this.element.on(
			"contextmenu",
			_.bind(function(e) {
				e.preventDefault();
				e.stopPropagation();

				this.hideTooltip();

				if (this.isEnabled() && this._dropdownEnabled) {
					this.showDropdown();
				}
			}, this)
		);
	}
	this.setAction(this.leftClickAction);
};

DropdownMiniButton.prototype = Object.create(MiniButton.prototype);
DropdownMiniButton.prototype.constructor = DropdownMiniButton;

DropdownMiniButton.prototype.isOpen = function() {
	return this.dropdown.isActive();
};

/**
 * @override
 */
DropdownMiniButton.prototype.show = function() {
	this.element.css("display", "inline-block");
};

/**
 * Show or hide the button
 * @param {Boolean} show - True to show the button, false to hide it
 */
DropdownMiniButton.prototype.toggleButton = function(show) {
	if (show) {
		this.show();
	} else {
		this.dropdown.hide();
		this.hide();
	}
};

/**
 * Set the text of a specific button by name
 * @param {String} name - Name of the button
 * @param {String} text - Text of the button
 */
DropdownMiniButton.prototype.setActionText = function(name, text) {
	this.dropdown.setActionText(name, text);
};

DropdownMiniButton.prototype.toggleOption = function(name, toggle) {
	this.dropdown.toggleOption(name, toggle);
};

DropdownMiniButton.prototype.getOptionValue = function(name) {
	return this.dropdown.getOptionValue(name);
};

DropdownMiniButton.prototype.setActionEnabled = function(name, enabled) {
	this.dropdown.setActionEnabled(name, enabled);
};

DropdownMiniButton.prototype.isActionEnabled = function(name) {
	return this.dropdown.isActionEnabled(name);
};
/**
 *
 */
DropdownMiniButton.prototype.setWarning = function() {
	this.badge.addClass("warning");
	this.setBadge("!");
};

DropdownMiniButton.prototype.removeWarning = function() {
	this.badge.removeClass("warning");
	this.setBadge("");
};

DropdownMiniButton.prototype.addAction = function(action) {
	this.dropdown.addAction(action);
};

DropdownMiniButton.prototype.addOption = function(option) {
	this.dropdown.addOption(option);
};

DropdownMiniButton.prototype.addDivider = function() {
	this.dropdown.addDivider();
};

DropdownMiniButton.prototype.addHeader = function(headerText) {
	this.dropdown.addHeader(headerText);
};

DropdownMiniButton.prototype.addContent = function(element) {
	this.dropdown.addContent(element);
};

/**
 * Clear out the dropdown list
 */
DropdownMiniButton.prototype.clear = function() {
	this.dropdown.clear();
};

/**
 * Enable the dropdown
 */
DropdownMiniButton.prototype.enableDropdown = function() {
	this._dropdownEnabled = true;
};

/**
 * Disable the dropdown
 */
DropdownMiniButton.prototype.disableDropdown = function() {
	this.dropdown.hide();
	this._dropdownEnabled = false;
};

/**
 * Show the dropdownmenu
 */
DropdownMiniButton.prototype.showDropdown = function() {
	this.dropdown.show(this.element);
};

/**
 * Show/hide the element with the given name
 * @param {String}   elementName - Name of the element to identify it. If the element is not found, this action does nothing.
 * @param {Boolean} show - Show or hide the action
 */
DropdownMiniButton.prototype.toggleElementVisibility = function(elementName, show) {
	if (show) {
		this.dropdown.showElement(elementName);
	} else {
		this.dropdown.hideElement(elementName);
	}
};

export default DropdownMiniButton;
