import ToolbarButton from "@esko/web-viewer/layout/component/ViewerToolbarButton";
import DropdownMenu from "@esko/dropdownmenu/DropdownMenu";
import $ from "jquery";
import _ from "underscore";
/**
 *
 * @class A DropdownToolbarButton is a ToolbarButton that when clicked shows a dropdownmenu.
 *
 * @exports Layout.Component.DropdownToolbarButton
 * @constructor
 * @param {Core.ViewerFacade} f
 * @param {String} name - the name of the button
 * @param {String | JQueryObj } icon - The class which is used for displaying the icon (css class for 16px * 16px icon), or your own jQuery element
 * @param {Boolean} forMobile - is the button needed for mobile usage?
 * @param {Function} [leftClickAction] - Action executed when left clicking. If undefined the dropdown will be shown when left clicking.
 * Parameters are a boolean indicating whether the button is activated (true) or deactivated (false). The button itself is NOT ACTIVATED!
 */
var DropdownToolbarButton = function(f, name, icon, forMobile, leftClickAction) {
	ToolbarButton.call(this, f, name, icon, forMobile, this.toggle);

	this.leftClickAction = leftClickAction;

	if (this.leftClickAction === undefined) {
		this.setAction(
			_.bind(function() {
				if (this.isEnabled() && this._dropdownEnabled) {
					if (!this.dropdown.isActive()) {
						this.showDropdown();
					} else {
						this.dropdown.hide();
					}
				}
			}, this)
		);
	} else {
		this.element.on(
			"contextmenu",
			_.bind(function(e) {
				e.preventDefault();
				e.stopPropagation();
				if (this.isEnabled() && this._dropdownEnabled) {
					this.showDropdown();
				}
			}, this)
		);
	}

	this._dropdownEnabled = true;

	this.dropdown = new DropdownMenu(this.f.i18n, this.f.getContext(), this.f.isMobile(), "bottom", "wcr-dropdown");
};

DropdownToolbarButton.prototype = Object.create(ToolbarButton.prototype);
DropdownToolbarButton.prototype.constructor = DropdownToolbarButton;

/**
 * Toggle the button
 */
DropdownToolbarButton.prototype.toggle = function() {
	if (this.isActive()) {
		this.leftClickAction(false);
	} else {
		this.leftClickAction(true);
	}
};

/**
 * Check whether the dropdown is open
 * @returns {Boolean} True if it is open, false otherwise
 */
DropdownToolbarButton.prototype.isOpen = function() {
	return this.dropdown.isActive();
};

/**
 * Show the dropdownmenu
 */
DropdownToolbarButton.prototype.showDropdown = function() {
	this.dropdown.show(this.element);
};

/**
 * Set the text of a specific button by name
 * @param {String} name - Name of the button
 * @param {String} text - Text of the button
 */
DropdownToolbarButton.prototype.setActionText = function(name, text) {
	this.dropdown.setActionText(name, text);
};

/**
 * Toggle an option
 * @param {String} name - Name of the option
 * @param {Boolean} toggle - True if should be 'checked', false otherwise
 */
DropdownToolbarButton.prototype.toggleOption = function(name, toggle) {
	this.dropdown.toggleOption(name, toggle);
};

/**
 * Get the value of an option
 * @param {String} name - Name of the option
 * @returns {Boolean} Value of the option
 */
DropdownToolbarButton.prototype.getOptionValue = function(name) {
	return this.dropdown.getOptionValue(name);
};

/**
 * Set whether the action is enabled or not
 * @param {String} name - Name of hte action
 * @param {Boolean} enabled - True if it should be enabled, false otherwise
 */
DropdownToolbarButton.prototype.setActionEnabled = function(name, enabled) {
	this.dropdown.setActionEnabled(name, enabled);
};

/**
 * @override
 * When setting a tooltip for DropdownToolbarButtons, we also add a 'right click for more options' string
 */
DropdownToolbarButton.prototype.setTooltip = function(tooltipOrBundle, langKey, stringArgs, popoverOptions) {
	if (langKey) {
		this._i18n.loadBundle(
			tooltipOrBundle,
			function() {
				if (stringArgs) {
					stringArgs.unshift(langKey);
				} else {
					stringArgs = [langKey];
				}
				var text = this._i18n.t.apply(this._i18n, stringArgs) + " (" + this.f.i18n.t("RightClickOptionTT") + ")";
				ToolbarButton.prototype.setTooltip.call(this, text, undefined, undefined, popoverOptions);
			},
			this
		);
	} else {
		tooltipOrBundle += " (" + this.f.i18n.t("RightClickOptionTT") + ")";
		ToolbarButton.prototype.setTooltip.call(this, tooltipOrBundle, undefined, undefined, popoverOptions);
	}
};

/**
 * Check whether the action is enabled or not
 * @param {String} name - Name of hte action
 * @returns {Boolean} True if it is enabled, false otherwise
 */
DropdownToolbarButton.prototype.isActionEnabled = function(name) {
	return this.dropdown.isActionEnabled(name);
};

/**
 * Add an action
 * @param {Object} action - Description of the action
 */
DropdownToolbarButton.prototype.addAction = function(action) {
	this.dropdown.addAction(action);
};

/**
 * Add an option
 * @param {Object} option - Description of the option
 */
DropdownToolbarButton.prototype.addOption = function(option) {
	this.dropdown.addOption(option);
};

/**
 * Add a divider line
 */
DropdownToolbarButton.prototype.addDivider = function() {
	this.dropdown.addDivider();
};

/**
 * Enable the dropdown
 */
DropdownToolbarButton.prototype.enableDropdown = function() {
	this._dropdownEnabled = true;
};

/**
 * Disable the dropdown
 */
DropdownToolbarButton.prototype.disableDropdown = function() {
	this._dropdownEnabled = false;
};

/**
 * Clear the contents of the dropdown menu
 */
DropdownToolbarButton.prototype.clear = function() {
	this.dropdown.clear();
};

export default DropdownToolbarButton;
