import Tooltip from "@esko/tooltip/Tooltip";
import $ from "jquery";
import _ from "underscore";
/**
 * Dynamic Icon makes it possible to save all possible icons by key and change the icon whithout the need
 * to create a new DOM-element or rerender an entire template because the icon is part of it. Every added icon
 * can have a seperate tooltip. Icons can be composed of a container- and sub-icon which are rendered on top
 * of each other. Do not use .element! use the getElement-method to get the DOM-object of the icon.
 *
 * @exports
 * @constructor
 * @param {Core.ViewerFacade} f - The viewer facade
 * @param {String} keyOrName - The name of the DynamIcon or the key of the given icon
 * @param {String} iconClass - default icon that is used for the DynamIcon
 * @param {String} [subIconClass] - default subicon that is used for the DynamIcon
 */
var DynamIcon = function(f, keyOrName, iconClass, subIconClass) {
	this.f = f;

	this._element = $("<span>").addClass("DynamIcon");

	this._icon = $("<i/>").addClass("icon");

	//are the icons composed with container- and sub-icons
	this._subIconsUsed = subIconClass ? true : false;

	if (this._subIconsUsed) {
		this._icon.addClass("containerIcon");
		this._subIcon = $("<i/>").addClass("icon subIcon");
		this._icon.append(this._subIcon);
	}

	this._tooltip = null;

	this._possibleIcons = {};

	this._possibleSubIcons = {};

	this._activeIconKey = null;

	this._tooltips = {};

	if (subIconClass) {
		this._possibleSubIcons[keyOrName] = subIconClass;
	}

	if (iconClass) {
		this._possibleIcons[keyOrName] = iconClass;
		this.setIcon(keyOrName);
	}

	this._element.append(this._icon);
};

/**
 * Set which icon needs to be displayed
 * @param {String} key
 */
DynamIcon.prototype.setIcon = function(key) {
	if (this._activeIconKey) {
		//first remove the class of the currently active icon
		this._icon.removeClass(this._possibleIcons[this._activeIconKey]);
		if (this._subIconsUsed) {
			//if subicons are used also remove the class of the subicon DOM-element
			this._subIcon.removeClass(this._possibleSubIcons[this._activeIconKey]);
		}
	}
	this._icon.addClass(this._possibleIcons[key]);
	if (this._subIconsUsed) {
		this._subIcon.addClass(this._possibleSubIcons[key]);
	}
	this._activeIconKey = key;
	this._setTooltip(key);
};

/**
 * Add a new icon, [subIcon] and tooltip to the map
 * @param {String} key
 * @param {String} iconClass
 * @param {String} [subIconClass]
 */
DynamIcon.prototype.addIcon = function(key, iconClass, subIconClass) {
	this._possibleIcons[key] = iconClass;
	if (subIconClass) {
		this._possibleSubIcons[key] = subIconClass;
	}
};

/**
 * Set Tooltip for icon with given key
 * @param {Stirng}		key
 * @param {String}         tooltipOrBundle - tooltip to show if langKey is undefined or language bundle
 * @param {String}         [langKey]       - not required
 * @param {Array.<String>} [stringArgs]    - not required, parameters that need to be filled in the language string
 */
DynamIcon.prototype.addTooltip = function(key, tooltipOrBundle, langKey, stringArgs) {
	this._tooltips[key] = {
		tooltipOrBundle: tooltipOrBundle,
		langKey: langKey,
		stringArgs: stringArgs
	};
	if (key === this._activeIconKey) {
		this._setTooltip(key);
	}
};

/**
 * Set which tooltip needs to be displayed on the currently active icon
 * @param {String} key
 * @private
 */
DynamIcon.prototype._setTooltip = function(key) {
	var tooltip = this._tooltips[key];
	if (tooltip) {
		if (this._tooltip === null) {
			this._tooltip = new Tooltip(
				this.f.i18n,
				this._element,
				this.f.getContext(),
				this.f.getContext(),
				this.f.isMobile(),
				tooltip.tooltipOrBundle,
				tooltip.langKey,
				tooltip.stringArgs
			);
		} else {
			this._tooltip.setTooltipContent(tooltip.tooltipOrBundle, tooltip.langKey, tooltip.stringArgs);
		}
	}
};

/**
 * Show the icon
 * @param {String} key of the icon to be shown
 */
DynamIcon.prototype.show = function(key) {
	if (key) {
		this.setIcon(key);
	}
	this._element.show();
};

/**
 * Hide the button
 */
DynamIcon.prototype.hide = function() {
	this._element.hide();
};

/**
 * Get the element of the icon
 * @returns {JQueryObj}
 */
DynamIcon.prototype.getElement = function() {
	//the tooltip is not rendered correctly if the icon was part of a template (all listeners get removed on its rerender)
	if (this._tooltip) {
		//if the tooltip is present make sure to refresh the mouse-listeners
		this._tooltip.refereshListeners();
	}
	//IE doesn't like dynamicons :(, and decides to destroys branches of the DOM-tree on rerender of the template
	if (this._element[0].children.length === 0) {
		//re-construct the structure of DOM-elements
		this._icon.append(this._subIcon);
		this._element.append(this._icon);
	}
	return this._element;
};

export default DynamIcon;
