import Tooltip from "@esko/tooltip/Tooltip";
import ViewerUtils from "@esko/web-viewer/utils/ViewerUtils";
import $ from "jquery";
import _ from "underscore";
import "jquery-ui/ui/widgets/slider";
/**
 * @class A log slider is a component that has a slider bar that slides logarithmic/exponential.
 * The minp and maxp parameters define the min/max number of the step in the slider.
 * The minv and maxv parameters define the min/max values that can be returned in the onSliderChanged callback
 * or can be set via setValue for the slider.
 * The wrapping element to add to a component sits in Slider.element
 * @exports LogSlider
 * @constructor
 * @param {Core.ViewerFacade} f - Viewer facade
 * @param {Object} options -
 *    {Number} minp: the minimum slider position (Default: 0)
 *    {Number} maxp: the maximum slider position (Default: 100)
 *    {Number} minv: the result minimum value (Default: 1)
 *    {Number} maxv: the result maximum value (Default: 100000)
 *    {Function} onSliderChanged: called when the slider value is changed. Arguments: value (Required)
 *    {String} tooltip: Tooltip text that should be shown when you hover over the component. (Optional)
 */
var LogSlider = function(f, options) {
	this.f = f;

	// Set Options
	this._minp = options.minp === undefined ? 0 : options.minp;
	this._maxp = options.maxp === undefined ? 100 : options.maxp;
	this.setOptions(options);
	this._onSliderChanged = options.onSliderChanged;

	// Wrapper element
	this.element = $("<div />").addClass("sliderWrapper");

	// Add slider
	this._slider = $("<div />").addClass("slider");
	this.element.append(this._slider);
	this._slider.slider({
		slide: _.bind(function(event, ui) {
			this._fireSliderChanged(ui.value);
		}, this),
		min: this._minp,
		max: this._maxp
	});

	// Tooltip
	if (options.tooltip !== undefined) {
		this._tooltip = new Tooltip(this.f.i18n, this.element, this.f.getContext(), this.f.getContext(), this.f.isMobile(), options.tooltip);
	}
};

/**
 * Calculate the value from a slider position
 * @param position
 * @returns
 */
LogSlider.prototype._toValue = function(position) {
	return Math.exp((position - this._minp) * this._scale + this._minv);
};

/**
 * Calculate the slider position from the given value
 * @param {Number} value - Value to calculate the slider position from
 * @returns {Number} Position of the slider for the given value
 */
LogSlider.prototype._toPosition = function(value) {
	return this._minp + (Math.log(value) - this._minv) / this._scale;
};

/**
 * Fire that the slider position has changed
 * @param {Number} value - The slider position
 */
LogSlider.prototype._fireSliderChanged = function(position) {
	var v = this._toValue(position);
	this._onSliderChanged(v);
};

/**
 * Set the value of the slider
 * @param {Number} newValue - New value of the slider
 */
LogSlider.prototype.setValue = function(newValue) {
	var p = this._toPosition(newValue);
	this._slider.slider("value", p);
};

/**
 * Update the initial options
 * @param {Object} options -
 *    {Number} minv: the result minimum value (Default: 1)
 *    {Number} maxv: the result maximum value (Default: 100000)
 */
LogSlider.prototype.setOptions = function(options) {
	this._minv = Math.log(options.minv === undefined ? 1 : options.minv);
	this._maxv = Math.log(options.maxv === undefined ? 100000 : options.maxv);
	this._scale = (this._maxv - this._minv) / (this._maxp - this._minp);
};

export default LogSlider;
