import Popup from "@esko/web-viewer/layout/component/Popup";
import ViewerToolbarButton from "@esko/web-viewer/layout/component/ViewerToolbarButton";
import ShapeUtils from "@esko/shapeutils";
import ListenerUtil from "@esko/web-viewer/utils/ListenerUtil";
import FullscreenUtil from "@esko/web-viewer/utils/FullscreenUtil";
import Mousetrap from "@esko/mousetrap";
import $ from "jquery";
import _ from "underscore";
/**
 * Create a new modal popup.
 * @exports ModalPopup
 * @class A modal popup is a popup where it is not possible to interact with things outside the popup
 * @constructor
 *
 * @param {Core.ViewerFacade} f - Viewer facade
 * @param {String} id - Id of the popup. It is used in its private channel.
 * @param {String|JQueryObj|String} containment - JQuery selector. The popup will be contained to the bounding box of the first element found by the selector. If no element is found, no containment will be set.
 *		| The popup will be contained to the bounding box of this element. | Possible values: "parent" and "document".
 *  @param {Boolean} [noCancelButton] - do no show a cancel button
 */
var ModalPopup = function(f, id, containment, noCancelButton, maxHeight, options) {
	options = options || {};
	if (FullscreenUtil.fullscreenElement) {
		options.container = $(FullscreenUtil.fullscreenElement);
	}
	Popup.call(this, f, id, containment, options);
	ListenerUtil.mixin(this, f);

	/**
	 * @private
	 * @member {JQueryObj} makes the popup modal, not possible to interact with other components of the viewer.
	 */
	this._background = $("<div/>").addClass("modalBackground");

	/**
	 * @private
	 * @member {Boolean} remove element from DOM when popup is hidden
	 */
	this._removeFromDomAfterUsage = false;

	this.element.addClass("modalPopup");
	/**
	 * @private
	 * @member {JQueryObj} content of the modal popup
	 */
	this._popupContent = $("<div/>");

	/**
	 * @private
	 * @member {JQueryObj} container for the cancel- and ok-action buttons of the popup
	 */
	this._actionBtns = $("<div/>")
		.addClass("pull-right")
		.addClass("modalPopupBtns");

	/**
	 * @private
	 * @member {JQueryObj} Container for the content of the popup
	 */
	this._popupContentContainer = $("<div/>").addClass("popupContentContainer clearfix");

	/**
	 * @private
	 * @member {Function} ok-action
	 */
	this._okAction = null;

	/**
	 * @private
	 * @member {Function} cancel-action
	 */
	this._cancelAction = function() {};

	/**
	 * @privfate
	 * @member {Function} close-action, this defaults to the cancel action
	 */
	this._closeAction = function() {
		this._cancelAction();
	};

	/**
	 * @private
	 * @member {Number} the maximum height for the modal popup
	 */
	this._maxHeight = maxHeight;

	if (this._maxHeight) {
		this.element.css("max-height", this._maxHeight);
	}

	/**
	 * @private
	 * @member {JQueryObj} ok-action button
	 */
	this._okBtn = new ViewerToolbarButton(
		this.f,
		"ok",
		$("<span/>"),
		false,
		_.bind(function() {
			this._okAction();
		}, this)
	);
	this._okBtn.addClass("btn-success");
	if (options.okButtonSahiID) {
		this._okBtn.element.get(0).dataset.sahiId = options.okButtonSahiID;
	}

	//close the modal popup with ESC
	Mousetrap.bind(
		"esc",
		_.bind(function(e) {
			this.hide();

			e.preventDefault();
			e.stopPropagation();
		}, this),
		"keydown"
	);

	/**
	 * @private
	 * @member {JQueryObj} cancel-action button
	 */
	this._cancelBtn = null;

	if (!noCancelButton) {
		this._cancelBtn = new ViewerToolbarButton(this.f, "cancel", $("<span/>"), false, _.bind(this.cancelAction, this));

		f.i18n.loadBundle(
			"viewer",
			_.bind(function() {
				this._cancelBtn.element.html(f.i18n.t("AppDialogCancel"));
			}, this)
		);
		this._actionBtns.append(this._cancelBtn.element);
		if (options.cancelButtonSahiID) {
			this._cancelBtn.element.get(0).dataset.sahiId = options.cancelButtonSahiID;
		}
	}
	this._actionBtns.append(this._okBtn.element);

	this.element.addClass("modal");
	this.content.append(this._popupContentContainer.append(this._popupContent).append(this._actionBtns));

	FullscreenUtil.addEventListener(
		"fullscreenchange",
		function() {
			if (FullscreenUtil.fullscreenElement) {
				this.setContainer($(FullscreenUtil.fullscreenElement));
			} else {
				this.setContainer(containment);
			}
		}.bind(this)
	);
};

ModalPopup.prototype = Object.create(Popup.prototype);
ModalPopup.prototype.constructor = ModalPopup;

/**
 * Set the content of the OK button
 * @param {String} okText
 */
ModalPopup.prototype.setOkBtnText = function(okText) {
	this._okBtn.element.html(okText);
};

/**
 * Set the content of the Cancel button
 * @param {String} cancelText
 */
ModalPopup.prototype.setCancelBtnText = function(cancelText) {
	if (this._cancelBtn) {
		this._cancelBtn.element.html(cancelText);
	}
};

/**
 * Show the modal popup
 */
ModalPopup.prototype.show = function() {
	if (!this._isInitialized) {
		this.on("closed", _.bind(this.closeAction, this));
		this._initialize();
	}

	this.element.before(this._background);
	this._background.show();

	Popup.prototype.show.call(this);

	this.recalculateSize();
};

/**
 * Recalculate the size of the modal popup
 * @param {Number} [popupWidth] - give a fixed with for the popup
 */
ModalPopup.prototype.recalculateSize = function(popupWidth) {
	var width = 0;
	if (popupWidth === undefined) {
		width = Math.ceil((this.containment.innerWidth() || null) * 0.55);
	} else {
		width = popupWidth;
	}

	this._popupContentContainer.outerWidth(width);
	var height = (this._popupContentContainer.outerHeight(true) || null) + (this.header.outerHeight(true) || null);
	this._popupContentContainer.outerWidth("auto");

	if (height > this._maxHeight) {
		height = this._maxHeight;
	}

	var x = (this.containment.innerWidth() || null) / 2 - width / 2;
	var y = (this.containment.innerHeight() || null) / 2 - height / 2;

	this.content.css("overflow", "hidden"); //fix for Chrome: dont show unnecessary scrollbar
	this.setRect(new ShapeUtils.Rect(new ShapeUtils.Point(x, y), new ShapeUtils.Size(width, height)));
	this.content.css("overflow", "auto");
};

/**
 * hide the modal popup
 */
ModalPopup.prototype.hide = function() {
	this._background.hide();
	Popup.prototype.hide.call(this);

	if (this._removeFromDomAfterUsage) {
		this.element.remove();
		this._background.remove();
	}
};

/**
 * Sets the action for the ok-action button
 * @param {Function} okAction
 */
ModalPopup.prototype.setOkAction = function(okAction) {
	this._okAction = okAction;
};

/**
 * Sets the action for the cancel-action button
 * @param {Function} cancelAction
 */
ModalPopup.prototype.setCancelAction = function(cancelAction) {
	this._cancelAction = cancelAction;
};

/**
 * Sets the action for the close-action button (if not set the cancel action is takeb by default)
 * @param {Function} closeAction
 */
ModalPopup.prototype.setCloseAction = function(closeAction) {
	this._closeAction = closeAction;
};

/**
 * Action for the cancel button
 */
ModalPopup.prototype.cancelAction = function() {
	this._cancelAction();
	this.hide();
};

/**
 * Action for the close button
 */
ModalPopup.prototype.closeAction = function() {
	this._closeAction();
	this.hide();
};

/**
 * Set the content of the modalpopup
 * @param {String|JQueryObj} content
 */
ModalPopup.prototype.setContent = function(content) {
	this._popupContent.html(content);
};

/**
 * Add an extra button just before the ok button
 * @param {Button} btn - the button to add
 */
ModalPopup.prototype.addButton = function(btn) {
	this._okBtn.element.before(btn.element);
};

/**
 * Set the ok button enabled or not
 * @param {Boolean} enable - True if you want to enable the button, false if not
 */
ModalPopup.prototype.enableOkBtn = function(enable) {
	this._okBtn.setEnabled(enable);
};

/**
 * Set whether or not the popup should be removed from DOM when it's hidden
 * @param {Boolean} removeFromDomAfterUsage - True if you want to remove from DOM
 */
ModalPopup.prototype.setRemoveFromDomAfterUsage = function(removeFromDomAfterUsage) {
	this._removeFromDomAfterUsage = removeFromDomAfterUsage;
};

/**
 * Display or hide the action buttons of the modal popup
 * @param {Boolean} toggle
 */
ModalPopup.prototype.toggleButtons = function(toggle) {
	this._actionBtns.toggleClass("displayNone", !toggle);
};

export default ModalPopup;
