import PrivateChannel from "@esko/web-viewer/core/PrivateChannel";
import ShapeUtils from "@esko/shapeutils";
import MiniButton from "@esko/web-viewer/layout/component/MiniButton";
import _ from "underscore";
import $ from "jquery";
import "jquery-ui/ui/widgets/draggable";
import "jquery-ui/ui/widgets/resizable";
/**
 * Create a new popup.
 * @exports Popup
 * @class
 * A popup is a dialog that has a header with a close button and some content.
 * You can also drag it/resize it inside its containment element.
 * The popup is added to the context of the facade.
 * @example
 * Emitted events:
 * 		"clicked" - Fired when a user clicks somewhere in the popup (except the close icon).
 * 		"closed" - Fired when a user clicks the close button.
 * 		"resized", ShapeUtils.Rect - Fired when a user stops resizing the popup. The argument is the new position and size of the popup relative to its containment element (meaning positioning the popup in the top left corner of the containment element is 0,0).
 * 		"dragged", ShapeUtils.Point - Fired when a user stops dragging the popup. The argument is the new position of the popup relative to its containment element (meaning positioning the popup in the top left corner of the containment element is 0,0).
 * @param {Core.ViewerFacade} f - Viewer facade
 * @param {String} id - Id of the popup. It is used in its private channel.
 * @param {String|JQueryObj|String} containment - JQuery selector. The popup will be contained to the bounding box of the first element found by the selector. If no element is found, no containment will be set.
 *		| The popup will be contained to the bounding box of this element. | Possible values: "parent" and "document".
 */
var Popup = function(f, id, containment, options) {
	PrivateChannel.installTo(this, f, ["popup", id]);
	this.containment = containment;
	this.f = f;

	this._options = options || {};
	if (options && options.allowClose === false) {
		this._allowClose = false;
	} else {
		this._allowClose = true;
	}

	/**
	 * is the popup initialized?
	 * @private
	 */
	this._isInitialized = false;

	// Wrapper
	this.element = $("<div>", {
		class: "popup",
		css: {
			position: "absolute"
		}
	});
	this.element.hide();

	// Header
	this.header = $("<td />").addClass("popup-header");

	// Header wrapper
	this.headerWrapper = $("<tr/>");

	// Right icons wrapper
	this.headerIcons = $("<td />").addClass("popup-headerIcons");

	//outer content with a scrollbar
	this._outerContent = $("<div>").addClass("popup-content");

	// Content
	this.content = $("<div />").addClass("inner-popup-content");

	this._outerContent.append(this.content);

	this._popupBtns = $("<div>");

	this._outerContent.append(this._popupBtns);

	// Visibility
	this.visible = false;

	// Minimum Size
	this.minWidth = 100;
	this.minHeight = 30;

	// React on makeMeVisible event when an element should be made visible in the popup
	this._outerContent.on(
		"makeMeVisible",
		_.partial(function(that, e) {
			// element
			var et = $(e.target).position().top;
			var eh = $(e.target).outerHeight() || null;
			// content
			var ct = $(this).position().top;
			var ch = $(this).outerHeight() || null;

			// If the content is too small, make the popup bigger
			if (eh > ch) {
				var ui = {
					position: {
						left: that.element.position().left,
						top: that.element.position().top
					},
					size: {
						width: that.element.outerWidth() || null,
						// element height + popup header
						height: eh + (that.header.outerHeight() || null)
					}
				};
				// the method is relative to containment, so compensate
				that.setRect(
					new ShapeUtils.rectMake(
						ui.position.left - that.containment.offset().left,
						ui.position.top - that.containment.offset().top,
						ui.size.width,
						ui.size.height
					)
				);
				that._fireResized(ui, true);
			}

			// check whether we need to scroll
			if (et - ct < 0 || et + eh > ct + ch) {
				var top = $(this).scrollTop() + et - ct;
				$(this).scrollTop(top);
			}
			e.preventDefault();
			e.stopPropagation();
		}, this)
	);

	// If a sub element triggered a contentResized event re-fire it on the eventbus
	this.content.on(
		"contentResized",
		_.bind(function(e) {
			this._fireContentResized();
			e.preventDefault();
			e.stopPropagation();
		}, this)
	);

	this.element.on("mousedown", function(e) {
		e.stopPropagation();
	});
};

/**
 * Initialize the resizable and draggable behavior of the popup
 * @private
 */
Popup.prototype._initialize = function() {
	if (!this._isInitialized) {
		// onclick
		this.element.on(
			"click",
			_.bind(function() {
				this._emit("clicked");
			}, this)
		);

		this.element.append(
			$("<table />")
				.addClass("popup-headerWrapper")
				.append(this.headerWrapper)
		);
		this.headerWrapper.append(this.header);
		this.headerWrapper.append(this.headerIcons);

		// Only add the close button when we allow the user to dismiss this modal.
		if (this._allowClose) {
			this.close = new MiniButton(
				this.f,
				"close",
				"icon-WCR_Close12",
				true,
				_.bind(function(e) {
					this._emit("closed");
					e.stopPropagation();
				}, this)
			);
			this.close.setTooltip("viewer", "ClosePopupTT");
			this.headerIcons.append(this.close.element);
		}

		this.element.append(this._outerContent);
		if (this._options.draggable !== false) {
			// Resizable
			this.element.resizable({
				handles: "all",
				autoHide: true,
				minWidth: this.minWidth,
				minHeight: this.minHeight,
				maxWidth: this.containment.innerWidth() || null,
				maxHeight: this.containment.innerHeight() || null,
				start: _.bind(function(event, ui) {
					this.f.emit("toolManager.interruptTools", true);
					event.stopPropagation();
				}, this),
				resize: _.throttle(
					_.bind(function(event, ui) {
						this._onResized();
						event.stopPropagation();
					}, this),
					10
				),
				stop: _.bind(function(event, ui) {
					this.f.emit("toolManager.interruptTools", false);
					this._fireResized(ui, false);
					event.stopPropagation();
				}, this)
			});

			// Draggable
			this.element.draggable({
				containment: this.containment,
				handle: this.headerWrapper,
				start: _.bind(function(event, ui) {
					this.f.emit("toolManager.interruptTools", true);
					event.stopPropagation();
				}, this),
				stop: _.bind(function(event, ui) {
					this.f.emit("toolManager.interruptTools", false);
					this._fireDragged(ui);
					event.stopPropagation();
				}, this)
			});
		}

		this._isInitialized = true;

		this.setContainer(this._options.container || this.f.getContext());

		this._onResized();
	}
};

/**
 * @private
 * Fire when the popup is dragged
 * @param {Object} ui - UI object from the dragged event (position is used)
 */
Popup.prototype._fireDragged = function(ui) {
	this._emit(
		"dragged",
		new ShapeUtils.Point(ui.offset.left - this.containment.offset().left, ui.offset.top - this.containment.offset().top)
	);
};

/**
 * @private
 * Fire when the popup is resized
 * @param {Object} ui - UI object from the resized event (position and size is used)
 * @param {Boolean} autoAdapted - True if the event is fired because the content was auto adapted (e.g. from a makeMeVisible trigger from an element in the content), false otherwise
 */
Popup.prototype._fireResized = function(ui, autoAdapted) {
	this._emit(
		"resized",
		new ShapeUtils.Rect(
			new ShapeUtils.Point(
				this.element.offset().left - this.containment.offset().left,
				this.element.offset().top - this.containment.offset().top
			),
			new ShapeUtils.Size(ui.size.width, ui.size.height)
		),
		autoAdapted
	);
};

/**
 * Set the header of the popup (left of the close button).
 * The previous header is erased.
 * @param {JQueryObj} header - The new header
 */
Popup.prototype.setHeader = function(header) {
	this.header.empty().append(header);
};

/**
 * Set the content of the popup.
 * The previous content is erased.
 * @param {JQueryObj} content - The new content
 */
Popup.prototype.setContent = function(content) {
	this.content.empty().append(content);
};

/**
 * Hide the popup
 */
Popup.prototype.hide = function() {
	if (this._isInitialized && this.visible) {
		this.element.hide();
	}

	this.visible = false;
};

/**
 * Show the popup
 * @param {ShapeUtils.Point} [location]
 * @param {Number} [width] - width of the popup, used to calculate the height and adapt the popup to its contents
 */
Popup.prototype.show = function(location, width) {
	if (!this._isInitialized) {
		this._initialize();
	}

	if (!this.visible) {
		this.element.show();
	}

	this.visible = true;

	if (location && width) {
		this.recalculateSize(location, width);
	}
};

/**
 * Is the popup visible
 * @returns {Boolean}
 */
Popup.prototype.isVisible = function() {
	return this.visible;
};

/**
 * Remove the popup from the DOM
 * It will remove all event listeners
 */
Popup.prototype.remove = function() {
	this.removeAllListeners("clicked");
	this.removeAllListeners("dragged");
	this.removeAllListeners("resized");
	this.removeAllListeners("closed");
	this.element.remove();
};

/**
 * @private
 * Called when the popup is resized to let the content follow the size of the parent
 */
Popup.prototype._onResized = function() {
	// show
	if (!this.visible) {
		this.element.show();
	}
	// let the content follow the (parent) element
	this._outerContent.css({
		width: this.element.outerWidth(false) || null,
		height: (this.element.outerHeight(false) || null) - (this.header.outerHeight(true) || null)
	});

	this._fireContentResized();

	// restore
	if (!this.visible) {
		this.element.hide();
	}
};

/**
 * Fire that the content of the popup is resized
 */
Popup.prototype._fireContentResized = function() {
	this._emit("contentResized", new ShapeUtils.Size(this._outerContent.width() || null, this._outerContent.height() || null));
};

/**
 * Set the rectangle of the popup (position and size) with the position relative to the containment element (top left corner of containment element is 0,0).
 * The position/size will be limited to the containment element.
 * @param {ShapeUtils.Rect} rect - Rectangle of the popup
 */
Popup.prototype.setRect = function(rect) {
	if (!this._isInitialized) {
		this._initialize();
	}

	var restrictedRect = this._restrictRectToContainment(rect);
	this.element.css({
		width: restrictedRect.size.width,
		height: restrictedRect.size.height,
		left: restrictedRect.origin.x,
		top: restrictedRect.origin.y
	});
	this._onResized();
};

/**
 * Set the size of the popup
 * @param {ShapeUtils.Size} size
 */
Popup.prototype.setSize = function(size) {
	if (!this._isInitialized) {
		this._initialize();
	}

	/**
	 * Restrict size
	 */
	var restrictedSize = size.copy();
	this._restrictSizeToContainment(restrictedSize);

	this.element.css({
		width: restrictedSize.width,
		height: restrictedSize.height
	});
	this._onResized();
};

/**
 * Update the minimum size of the popup.
 *
 * @param size
 */
Popup.prototype.setMinSize = function(size) {
	// Update the internal variables.
	this.minWidth = size.width;
	this.minHeight = size.height;
	// Update it in the element
	this.element.resizable({
		minWidth: this.minWidth,
		minHeight: this.minHeight
	});
};

/**
 * @private
 * Restrict the size to the containment element
 * The size parameter is changed so it fits in the containment element
 * @param {ShapeUtils.Size} size - Size to restrict
 */
Popup.prototype._restrictSizeToContainment = function(size) {
	var maxWidth = this.containment.innerWidth() || null;
	var maxHeight = this.containment.innerHeight() || null;
	if (size.width > maxWidth) {
		size.width = maxWidth;
	}
	if (size.height > maxHeight) {
		size.height = maxHeight;
	}
};

/**
 * @private
 * Restrict the rect to the containment element
 * @param {ShapeUtils.Rect} rect - Rect to restrict
 */
Popup.prototype._restrictRectToContainment = function(rect) {
	var restrictedRect = rect.copy();

	/**
	 * Restrict size
	 */
	this._restrictSizeToContainment(restrictedRect.size);

	/**
	 * Restrict position
	 */
	// popups are added to the viewer context, so take offset into account of containment
	// restrict to containment
	if (restrictedRect.origin.x + restrictedRect.size.width > (this.containment.innerWidth() || null)) {
		restrictedRect.origin.x = (this.containment.innerWidth() || null) - restrictedRect.size.width;
	}
	if (restrictedRect.origin.y + restrictedRect.size.height > (this.containment.innerHeight() || null)) {
		restrictedRect.origin.y = (this.containment.innerHeight() || null) - restrictedRect.size.height;
	}

	// make sure we don't have any negative offsets
	restrictedRect.origin.x = Math.max(0, restrictedRect.origin.x);
	restrictedRect.origin.y = Math.max(0, restrictedRect.origin.y);

	// Calculate origin x/y relative to nearest positioned parent
	var containRect = this.containment[0].getBoundingClientRect();
	restrictedRect.origin.x += containRect.left;
	restrictedRect.origin.y += containRect.top;
	let offsetParent = this.element[0].offsetParent;
	if (offsetParent) {
		let parentRect = offsetParent.getBoundingClientRect();
		restrictedRect.origin.x -= parentRect.left;
		restrictedRect.origin.y -= parentRect.top;
	}

	return restrictedRect;
};

/**
 * Set a new containment for the popup.
 * @param {String|JQueryObj|String} containment - JQuery selector. The popup will be contained to the bounding box of the first element found by the selector. If no element is found, no containment will be set.
 *		| The popup will be contained to the bounding box of this element. | Possible values: "parent" and "document".
 */
Popup.prototype.setContainment = function(containment) {
	if (this.containment !== containment) {
		this.containment = containment;

		// update resizable options
		if (this._isInitialized) {
			this.element.resizable("option", "maxWidth", this.containment.innerWidth() || null);
			this.element.resizable("option", "maxHeight", this.containment.innerHeight() || null);
			// update draggable options
			this.element.draggable("option", "containment", this.containment);
		}
	}
};

Popup.prototype.setContainer = function(container) {
	this.element.appendTo(container);
};

/**
 * Recalculate the size of the popup, make sure the popup is resized accordingly to its contents
 * @param {ShapeUtils.Point} location
 * @param {Number} popupWidth
 */
Popup.prototype.recalculateSize = function(location, popupWidth) {
	this._outerContent.outerWidth(popupWidth);
	var height =
		(this.content.outerHeight(true) || null) + (this.header.outerHeight(true) || null) + (this._popupBtns.outerHeight(true) || null);

	this._outerContent.css("overflow", "hidden"); //fix for Chrome: dont show unnecessary scrollbar
	this.setRect(new ShapeUtils.Rect(location, new ShapeUtils.Size(popupWidth, height)));
	this._outerContent.css("overflow", "auto");
};

/**
 * Add a button element to the buttons group.
 * @param {JQueryObj} buttonElement
 */
Popup.prototype.addButton = function(buttonElement) {
	this._popupBtns.addClass("popupBtns");
	this._popupBtns.append(buttonElement);
};

/**
 * Show/hide the header icons
 * @param {Boolean} toggle [[Description]]
 */
Popup.prototype.toggleHeaderIcons = function(toggle) {
	this.headerIcons.toggleClass("displayNone", !toggle);
};

export default Popup;
