import ViewerUtils from "@esko/web-viewer/utils/ViewerUtils";
import $ from "jquery";
import _ from "underscore";
/**
 * Create a new abstract rich text editor. The richtext editor is not shown by default, it will be initialized by calling .init()
 * @class
 * A rich text editor has a title, a rich text area and an action bar with some buttons.
 * This class hangs on the DOM element in the property "richTextObj"
 * @exports RichTextEditor
 */
var RichTextEditor = function(f) {
	this.f = f;
	this.onTextChangeCallbacks = [];

	// Wrapper
	this.id = ViewerUtils.generateUniqueDOMId("richText_");

	this.element = $("<div />")
		.addClass("richText")
		.attr("id", this.id);
	this.element.data("richTextObj", this);

	// Title bar
	this.titleBar = $("<h3 />").addClass("title");
	this.titleBar.hide();

	// Action bar
	this.actionBar = $("<div />")
		.addClass("actionBar")
		.addClass("clearfix");

	// Init text area
	this.textAreaDiv = $("<div />");
	this.textArea = $("<textarea />").addClass("text");
	this.textAreaId = "textArea_" + this.id;
	this.textArea.attr("id", this.textAreaId);

	// Is the editor initialized
	this.initialized = false;

	// Add them to the wrapper
	this.element
		.append(this.titleBar)
		.append(this.textAreaDiv.append(this.textArea))
		.append(this.actionBar);

	// should fire text changes?
	this._fireTextChanges = true;
};

/**
 * Set content
 * @param {String} content - Content to set in the text area
 */
RichTextEditor.prototype.setContent = function(content) {
	// to be implemented
};

/**
 * Get content as HTML
 * @returns {String} Content as HTML if the editor was initialized, otherwise ""
 */
RichTextEditor.prototype.getContent = function() {
	// to be implemented
	return "";
};

/**
 * Set the buttons
 * The previous action bar is erased.
 * @param {Array<Obj.element>} buttons - Can be every JQueryObj with a .element. If it's undefined, it's not added.
 */
RichTextEditor.prototype.setButtons = function(buttons) {
	this.actionBar.empty();
	_.each(
		buttons,
		function(button) {
			if (button) {
				this.actionBar.append(button.element);
			}
		},
		this
	);
};

/**
 * Set the title
 * @param {String} newTitle - New title
 */
RichTextEditor.prototype.setTitle = function(newTitle) {
	this.titleBar.text(newTitle);
	this.titleBar.show();
};

/**
 * Init the rich text editor
 * Should be called after the element is added to the DOM
 * @param {Function} [callback] - Callback when the editor is initialized (also fired when it was already initialized). From then on, you can safely use methods like getContent().
 */
RichTextEditor.prototype.init = function(callback) {
	// initialize the editor
};

/**
 * Fire a change of the contents
 */
RichTextEditor.prototype._fireTextChanged = function() {
	if (this._shouldFireTextChanged()) {
		_.each(this.onTextChangeCallbacks, function(onTextChangeCallback) {
			onTextChangeCallback();
		});
	}
};

/**
 * should a text change be fired?
 * @returns {Boolean}
 */
RichTextEditor.prototype._shouldFireTextChanged = function() {
	// override
	return this._fireTextChanges;
};

/**
 * Check whether the rich text editor is already initialized
 * @returns {Boolean} True if the editor was initialized, false otherwise
 */
RichTextEditor.prototype.isInitialized = function() {
	return this.initialized;
};

/**
 * Destroy the rich text editor.
 */
RichTextEditor.prototype.destroy = function() {
	// to be implemented
};

/**
 * Add a callback to the editor that is fired when the text changes
 * It includes onKeyUp, onUndo, onRedo
 * @param {Function} onTextChangeCallback - executed on text change of the editor
 */
RichTextEditor.prototype.onTextChanged = function(onTextChangeCallback) {
	this.onTextChangeCallbacks.push(onTextChangeCallback);
};

/**
 * Should be called when an editor is detached from the DOM
 */
RichTextEditor.prototype.beforeDetach = function() {
	// to be implemented
};

/**
 * Repair the TinyMCE instance by destroying it and reinializing it
 */
RichTextEditor.prototype.repair = function() {
	// to be implemented
};

/**
 * Set the height of the editor
 * @param {Number} newHeight - New height of the editor
 */
RichTextEditor.prototype.setHeight = function(newHeight) {
	if (this.isInitialized()) {
		var height = newHeight - (this.actionBar.outerHeight(true) || null);
		height = Math.max(height, 80); // min 80 px
		this.textAreaDiv.css("height", height);
	}
};

export default RichTextEditor;
