import RichTextEditor from "@esko/web-viewer/layout/component/RichTextEditor";
import ViewerUtils from "@esko/web-viewer/utils/ViewerUtils";
import tinyMCE from "tinymce";
import attachRTLPlugin from "@esko/tinymce-esko-plugins/rtl";
import $ from "jquery";
import _ from "underscore";
import "tinymce/themes/modern/theme";
import "tinymce/plugins/fullpage";
import "tinymce/plugins/directionality";
import "tinymce/plugins/paste";
import TINYMCE_LANGUAGE_MAP from "@esko/tinymce-esko-plugins/tinymcelanguages";

var DEFAULT_LANGUAGE = "en";

/**
 * Create a new TinyMCE rich text editor. The richtext editor is not shown by default, it will be initialized by calling .init()
 * @class
 * A rich text editor has a title, a rich text area and an action bar with some buttons.
 * This class hangs on the DOM element in the property "richTextObj"
 * @exports RichTextEditorTinyMCE
 */
var RichTextEditorTinyMCE = function(f) {
	RichTextEditor.call(this, f);

	this.editor = undefined;
};

RichTextEditorTinyMCE.prototype = Object.create(RichTextEditor.prototype);
RichTextEditorTinyMCE.prototype.constructor = Object.create(RichTextEditorTinyMCE);
/**
 * Set content
 * @param {String} content - Content to set in the text area
 * @override
 */
RichTextEditorTinyMCE.prototype.setContent = function(content) {
	this.textArea.val(content);
	if (this.initialized) {
		this.editor.setContent(content);
	}
};

/**
 * Get content as HTML
 * @returns {String} Content as HTML if the editor was initialized, otherwise ""
 * @override
 */
RichTextEditorTinyMCE.prototype.getContent = function() {
	if (this.initialized) {
		return this.editor.getContent();
	} else {
		return "";
	}
};

/**
 * Init the rich text editor
 * Should be called after the element is added to the DOM
 * @param {Function} [callback] - Callback when the editor is initialized (also fired when it was already initialized). From then on, you can safely use methods like getContent().
 * @override
 */
RichTextEditorTinyMCE.prototype.init = function(callback) {
	if (!this.initialized) {
		var userLocale = this.f.getCurrentUser().getLocale();
		this._initMCE(callback, userLocale);
	} else {
		if (callback) {
			callback();
		}
		this.editor.focus();
	}
};

/**
 * Start up TinyMCE
 * @param {Function} callback - callback function to be called when TinyMCE is initialized
 * @param {String} locale - the locale in which TinyMCE should be started
 */
RichTextEditorTinyMCE.prototype._initMCE = function(callback, locale) {
	var languageUrl = null;

	if (TINYMCE_LANGUAGE_MAP[locale]) {
		languageUrl = TINYMCE_LANGUAGE_MAP[locale];
	} else {
		languageUrl = TINYMCE_LANGUAGE_MAP[DEFAULT_LANGUAGE];
	}

	tinyMCE.init(
		attachRTLPlugin({
			allowTextAlign: false
		})({
			skin_url: process.env.ASSET_PATH + "tinymce/skins/lightgray",
			selector: "#" + this.textAreaId,
			menubar: false,
			statusbar: false,
			branding: false,
			plugins: "paste fullpage",
			toolbar: "bold italic underline strikethrough | subscript superscript | undo redo | ltr rtl",
			toolbar_items_size: "small",
			element_format: "xhtml",
			doctype: "<!DOCTYPE html PUBLIC '-//W3C//DTD XHTML 1.0 Strict//EN' " + "'http://www.w3.org/TR/xhtml1/DTD/xhtml1-strict.dtd'>",
			fullpage_default_doctype:
				"<!DOCTYPE html PUBLIC '-//W3C//DTD XHTML 1.0 Strict//EN' " + "'http://www.w3.org/TR/xhtml1/DTD/xhtml1-strict.dtd'>",
			inline_styles: true,
			language_url: languageUrl,
			valid_elements:
				"@[dir|style],p,p/li,div,span,span,br,nobr,sub,sup,strike/s,strong/b,i,u,s,a,h?,big,small,font,tt,code,var,kbd,em,samp,center,a[href|title]",
			auto_focus: this.textAreaId,
			setup: _.bind(function(ed) {
				this.editor = ed;
				ed.on(
					"keyup",
					_.bind(function(e) {
						this._fireTextChanged();
					}, this)
				);
				ed.on(
					"redo",
					_.bind(function(e) {
						this._fireTextChanged();
					}, this)
				);
				ed.on(
					"undo",
					_.bind(function(e) {
						this._fireTextChanged();
					}, this)
				);
				ed.on(
					"change",
					_.bind(function(e) {
						this._fireTextChanged();
					}, this)
				);
				ed.on(
					"init",
					_.bind(function(e) {
						this.initialized = true;
						if (callback) {
							callback();
						}
					}, this)
				);
			}, this),
			// this method is called before text is pasted in the richtext editor
			paste_preprocess: function(pl, o) {
				// convert the tables to div/p/span elements, as the java viewer doesn's support tables
				o.content = o.content.replace(/<(\/)?table[^>]*>/gi, "<$1div>");
				o.content = o.content.replace(/<(\/)?tr[^>]*>/gi, "<$1p>");
				o.content = o.content.replace(/<th[^>]*>/gi, "<span><strong>");
				o.content = o.content.replace(/<\/th[^>]*>/gi, "</strong></span>&nbsp;");
				o.content = o.content.replace(/<td[^>]*>/gi, "<span>");
				o.content = o.content.replace(/<\/td[^>]*>/gi, "</span>&nbsp;");
			}
		})
	);
};

/**
 * should a text change be fired?
 * @returns {Boolean}
 * @override
 */
RichTextEditorTinyMCE.prototype._shouldFireTextChanged = function() {
	return this.editor && this.editor.isDirty() && this._fireTextChanges;
};

/**
 * Destroy the rich text editor.
 * @override
 */
RichTextEditorTinyMCE.prototype.destroy = function() {
	if (this.initialized) {
		this.editor.destroy();
	}
	this.initialized = false;
};

/**
 * Should be called when an editor is detached from the DOM
 * @override
 */
RichTextEditorTinyMCE.prototype.beforeDetach = function() {
	if (this.isInitialized()) {
		this._fireTextChanges = false;
		this.destroy();
		this._fireTextChanges = true;
	}
};

/**
 * Repair the TinyMCE instance by destroying it and reinializing it
 * @override
 */
RichTextEditorTinyMCE.prototype.repair = function() {
	this.init();
};

/**
 * @override
 */
RichTextEditorTinyMCE.prototype.setHeight = function(newHeight) {
	if (this.isInitialized()) {
		// just calculating it here doesn't work as tinymce loads its css asynchronously
		var toolbarHeight = this.textAreaDiv.find(".mce-toolbar-grp").outerHeight(true) || null;
		// THIS IS A VERY UGLY HACK but there is no usable alternative known at the moment!
		// one toolbar row is 26px, 2px padding at top and bottom
		if (toolbarHeight % 26 !== 4) {
			toolbarHeight = 30;
		}
		var height = newHeight - (this.actionBar.outerHeight(true) || null) - toolbarHeight;
		height = Math.max(height, 80); // min 80 px
		var iframe = "#" + this.textAreaId + "_ifr";
		$(iframe).css("height", height);
	}
};

export default RichTextEditorTinyMCE;
