import Tooltip from "@esko/tooltip/Tooltip";
import Spinner from "@esko/web-viewer/layout/component/Spinner";
import ViewerUtils from "@esko/web-viewer/utils/ViewerUtils";
import _ from "underscore";
import $ from "jquery";
import "jquery-ui/ui/widgets/slider";
import "jquery-ui-touch-punch";
/**
 * @class A slider is a component that has a slider bar and optionally a spinner next to it.
 * The wrapping element to add to a component sits in Slider.element
 * @exports Slider
 * @constructor
 * @param {Core.ViewerFacade} f - Viewer facade
 * @param {Object} options -
 *    {Number} min: minimum value of the slider (Default: 0)
 *    {Number} max: maximum value of the slider (Default: 100)
 *    {Number} step: step value of the slider (Default: 1)
 *    {Function} onSliderChanged: called when the slider value is changed. Arguments: value (Required)
 *    {Boolean} showInput: Show an input field next to the slider (Default: true)
 *    {String} inputUnit: Unit of the input field (put behind the input field) (Default: %)
 *    {String} tooltip: Tooltip text that should be shown when you hover over the component. (Optional)
 */
var Slider = function(f, options) {
	this.f = f;

	// Set Options
	this._min = options.min === undefined ? 0 : options.min;
	this._max = options.max === undefined ? 100 : options.max;
	this._step = options.step === undefined ? 1 : options.step;
	this._onSliderChanged = options.onSliderChanged;
	this._showInput = options.showInput === undefined ? true : options.showInput;
	this._inputUnit = options.inputUnit === undefined ? "%" : options.inputUnit;

	// Wrapper element
	this.element = $("<div />").addClass("sliderWrapper");

	// Add slider
	this._slider = $("<div />").addClass("slider");
	this.element.append(this._slider);
	this._slider.slider({
		slide: _.bind(function(event, ui) {
			this._setInputValue(ui.value);
			this._fireSliderChanged(ui.value);
		}, this),
		min: this._min,
		max: this._max,
		step: this._step
	});

	// Add the input field if needed
	if (this._showInput) {
		this._spinner = new Spinner(f, {
			min: this._min,
			max: this._max,
			step: this._step,
			page: this._step * 10,
			onSpinnerChanged: _.bind(function(value) {
				this._setSliderValue(value);
			}, this),
			inputUnit: this._inputUnit
		});
		this.element.append(this._spinner.element);
	}

	// Tooltip
	if (options.tooltip !== undefined) {
		this._tooltip = new Tooltip(this.f.i18n, this.element, this.f.getContext(), this.f.getContext(), this.f.isMobile(), options.tooltip);
	}
};

/**
 * @private
 * Fire the value to the listener
 * @param {Number} value - New slider value
 */
Slider.prototype._fireSliderChanged = function(value) {
	this._onSliderChanged(value);
};

/**
 * @private
 * Set the value of the slider.
 * @param {Number} value - New value of the slider
 */
Slider.prototype._setSliderValue = function(value) {
	if (this.getValue() !== value) {
		this._slider.slider("value", value);
		this._fireSliderChanged(value);
	}
};

/**
 * @private
 * Set the value of the spinner. It doesn't fire yet.
 * @param {Number} value - New value of the spinner
 */
Slider.prototype._setInputValue = function(value) {
	if (this._showInput) {
		this._spinner.setValue(value);
	}
};

/**
 * Set the value of the slider
 * @param {Number} value - New value of the slider
 */
Slider.prototype.setValue = function(value) {
	if (this.getValue() !== value) {
		this._setSliderValue(value);
		this._setInputValue(value);
	}
};

/**
 * Get the value of the slider
 * @returns {Number} The value of the slider
 */
Slider.prototype.getValue = function() {
	return this._slider.slider("value");
};

/**
 * Is the slider disabled?
 * @returns {Boolean} True if the slider is disabled, false otherwise
 */
Slider.prototype.isDisabled = function() {
	return this._slider.slider("option", "disabled");
};

/**
 * Set the slider disabled
 * @param {Boolean} disabled - True if the slider should be disabled, false if the slider should be enabled.
 */
Slider.prototype.setDisabled = function(disabled) {
	this._slider.slider("option", "disabled", disabled);
	if (this._showInput) {
		this._input.spinner("option", "disabled", disabled);
	}
};

/**
 * Set the tool tip of the slider
 * @param {String} tooltip - Tool tip text
 */
Slider.prototype.setTooltip = function(tooltip) {
	if (this._tooltip === undefined) {
		this._tooltip = new Tooltip(this.f.i18n, this.element, this.f.getContext(), this.f.getContext(), this.f.isMobile(), tooltip);
	} else {
		this._tooltip.setTooltipContent(tooltip);
	}
};

/**
 * Update one of the initial options.
 * Currently supported: "min" and "max".
 * @param {String} optionName - Name of the option, e.g. "max"
 * @param {Object} value - New value for the option
 */
Slider.prototype.setOption = function(optionName, newValue) {
	var oldValue = this["_" + optionName];
	if (oldValue !== undefined && oldValue !== newValue) {
		// update the current value
		this["_" + optionName] = newValue;
		// update the slider
		this._slider.slider("option", optionName, newValue);

		// update the spinner if needed
		if (this._showInput) {
			this._spinner.setOption(optionName, newValue);
		}
	}
};

/**
 * Get one of the options.
 * Currently supported: "min" and "max".
 * @param {String} optionName - Name of the option, e.g. "max"
 * @returns {Object} value - The value of the option or undefined if not found
 */
Slider.prototype.getOption = function(optionName) {
	return this["_" + optionName];
};

export default Slider;
