import Tooltip from "@esko/tooltip/Tooltip";
import ViewerUtils from "@esko/web-viewer/utils/ViewerUtils";
import _ from "underscore";
import $ from "jquery";
import "jquery-ui/ui/widgets/spinner";
/**
 * @class A spinner is a component that has an input field with arrows next to it to spin through the values.
 * You can also use page keys (page) or arrow keys (step).
 * The wrapping element to add to a component sits in Spinner.element
 * @exports Spinner
 * @constructor
 * @param {Core.ViewerFacade} f - Viewer facade
 * @param {Object} options - put undefined for default values
 *    {Number} min: minimum value of the spinner (Default: 0)
 *    {Number} max: maximum value of the spinner (Default: 100)
 *    {Number} step: step value of the spinner (Default: 1);
 *    {Number} page: page size of the spinner (Default: 10);
 *    {Function} onSpinnerChanged: called when the spinner value is changed. Arguments: value (Required)
 *    {String} inputUnit: Unit of the input field (put behind the input field) (Default: %)
 *    {String} tooltip: Tooltip text that should be shown when you hover over the component. (Optional)
 */
var Spinner = function(f, options) {
	this.f = f;

	// Set Options
	this._min = options.min === undefined ? 0 : options.min;
	this._max = options.max === undefined ? 100 : options.max;
	this._step = options.step === undefined ? 1 : options.step;
	this._page = options.page === undefined ? 10 : options.page;
	this._onSpinnerChanged = options.onSpinnerChanged;
	this._inputUnit = options.inputUnit === undefined ? "%" : options.inputUnit;

	// Wrapper element
	this.element = $("<div />").addClass("spinnerWrapper");

	// Add the input field if needed
	this._input = $("<input />")
		.attr("size", ("" + this._max).length)
		.addClass("input"); // the input field
	this._input.on("keydown", _.bind(this._valueChanged, this));
	this._input.on("blur", _.bind(this._valueChanged, this));
	this._labelUnit = $("<span>" + this._inputUnit + "</span>");
	this._label = $("<label />")
		.addClass("unit")
		.append(this._input)
		.append(this._labelUnit); // the unit
	this.element.append(this._label);

	// spinner behavior
	this._input.spinner({
		min: this._min,
		max: this._max,
		step: this._step,
		page: this._page,
		spin: _.bind(function(event, ui) {
			this._fireSpinnerChanged(ui.value);
		}, this)
	});
	this.setValue(this._min);

	// Tooltip
	if (options.tooltip !== undefined) {
		this._tooltip = new Tooltip(this.f.i18n, this.element, this.f.getContext(), this.f.getContext(), this.f.isMobile(), options.tooltip);
	}
};

/**
 * Change the value of the spinner (by pressing enter in the input or on blur)
 * @param {Object} event
 */
Spinner.prototype._valueChanged = function(event) {
	if ((event.which == 13 || event.keyCode == 13) && event.type === "keydown") {
		// on enter, lose focus, let the blur event handle it
		this._input.trigger("blur");
	} else if (event.type !== "keydown") {
		var value = ViewerUtils.parseFloat($(event.target).val());
		if (isNaN(value)) {
			// if nothing is filled in or only letter(s), behave like entered the minimum
			value = this._min;
		}
		// limit to boundaries
		value = Math.max(Math.min(value, this._max), this._min);
		$(event.target).val(value); // by doing this letters added to the end will be filtered out and bounderies are enforced
		this._fireSpinnerChanged(value);
	}
};

/**
 * @private
 * Fire the value to the listener
 * @param {Number} value - New Spinner value
 */
Spinner.prototype._fireSpinnerChanged = function(value) {
	this._onSpinnerChanged(value);
};

/**
 * Set the value of the Spinner
 * @param {Number} value - New value of the Spinner
 */
Spinner.prototype.setValue = function(value) {
	if (this.getValue() !== value) {
		this._input.spinner("value", value);
	}
};

/**
 * Get the value of the Spinner
 * @returns {Number} The value of the Spinner
 */
Spinner.prototype.getValue = function() {
	return this._input.spinner("value");
};

/**
 * Is the Spinner disabled?
 * @returns {Boolean} True if the Spinner is disabled, false otherwise
 */
Spinner.prototype.isDisabled = function() {
	return this._input.spinner("option", "disabled");
};

/**
 * Set the Spinner disabled
 * @param {Boolean} disabled - True if the Spinner should be disabled, false if the Spinner should be enabled.
 */
Spinner.prototype.setDisabled = function(disabled) {
	this._input.spinner("option", "disabled", disabled);
};

/**
 * Set the tool tip of the Spinner
 * @param {String} tooltip - Tool tip text
 */
Spinner.prototype.setTooltip = function(tooltip) {
	if (this._tooltip === undefined) {
		this._tooltip = new Tooltip(this.f.i18n, this.element, this.f.getContext(), this.f.getContext(), this.f.isMobile(), tooltip);
	} else {
		this._tooltip.setTooltipContent(tooltip);
	}
};

/**
 * Update one of the initial options.
 * Currently supported: "min","max","step","page".
 * @param {String} optionName - Name of the option, e.g. "max"
 * @param {Object} value - New value for the option
 */
Spinner.prototype.setOption = function(optionName, newValue) {
	var oldValue = this["_" + optionName];
	if (oldValue !== undefined && oldValue !== newValue) {
		// update the current value
		this["_" + optionName] = newValue;
		// update the spinner
		this._input.spinner("option", optionName, newValue);
	}
};

/**
 * Set the label of the spinner and update the DOM element
 * @param {String} newValuet - new value of the label
 */
Spinner.prototype.setInputUnit = function(newValuet) {
	this._inputUnit = newValuet;
	this._labelUnit.empty();
	this._labelUnit.append(this._inputUnit);
};

export default Spinner;
