import ViewerUtils from "@esko/web-viewer/utils/ViewerUtils";
import CoreConstants from "@esko/web-viewer/core/CoreConstants";
import _ from "underscore";
import $ from "jquery";
/**
 * An input field that adapts to the user preference UNIT SIZE
 * It allows the easily get input from the users (as a mm value)
 * @memberof LayoutModule
 * @param {ViewerFacade}         f         - the ViewerFacade
 * @param {Number|String}        value     - the initial mm value of the field
 * @param {String}               direction - x or y the direction of the value (needed because resolutionx and resolutiony can be different in 1 file)
 * @param {Function}             onChange  - callback called when the value has changed. Its parameters are oldValue, newValue (both in mm)
 */
var UserUnitInputField = function(f, value, direction, onChange) {
	this.f = f;
	this.element = $("<span>").addClass("userUnitInputField");
	this._input = $("<input>")
		.attr("type", "number")
		.attr("step", "any")
		.appendTo(this.element);
	this._unitLabel = $("<span>")
		.addClass("unitLabel")
		.appendTo(this.element);

	this._direction = direction;

	value = ViewerUtils.parseFloat(value);

	this._mmValue = value;

	this.f.getCurrentUser().on("options.changed.PREFERENCES_UNIT_SIZE", _.bind(this._renderField, this));
	this._input.on("blur keydown", _.bind(this._inputChanged, this));
	this._renderField();

	this._onChange = onChange;
};

/**
 * Set the value (in MM)
 * @param {Number} value - value in mm
 */
UserUnitInputField.prototype.setValue = function(value) {
	var oldValue = this._mmValue;

	this._mmValue = ViewerUtils.parseFloat(value);
	this._renderField();

	if (oldValue !== this._mmValue) {
		this._onChange(oldValue, this._mmValue);
	}
};

/**
 * Render the field
 */
UserUnitInputField.prototype._renderField = function() {
	var convertedValue = ViewerUtils.convertFromMmValue(this._mmValue, this.f.getUnits(), this.f.getUnitsFix(), this._getResolution());
	this._input.val(convertedValue);

	this.f.loadBundle(
		"viewer",
		function() {
			this._unitLabel.text(this.f.t(this.f.getUnitsLocName()));
		},
		this
	);
};

/**
 * Get the resolution for unit conversion
 * @returns {Number} resolution of the page for the needed conversion
 */
UserUnitInputField.prototype._getResolution = function() {
	var page = this.f.getActivePage();
	if (page) {
		if (this._direction === "y") {
			return ViewerUtils.parseFloat(page.getParam("resolutiony"));
		} else if (this._direction === "x") {
			return ViewerUtils.parseFloat(page.getParam("resolutionx"));
		}
	}

	// by default return 1
	return 1;
};

/**
 * The user changed the input field, update the mmValue and send out correct event
 * @param {Object} event - Event Object
 */
UserUnitInputField.prototype._inputChanged = function(event) {
	if (event.which == 13 || event.keyCode == 13 || event.type !== "keydown") {
		var value = ViewerUtils.parseFloat(this._input.val());
		if (!value) {
			value = 0;
		}

		this._input.val(value);
		var oldValue = this._mmValue;
		this._mmValue = ViewerUtils.parseFloat(
			ViewerUtils.convertToMmValue(value, this.f.getUnits(), this.f.getUnitsFix(CoreConstants.SizeUnits.MM), this._getResolution())
		);

		if (this._mmValue != oldValue) {
			this._onChange(oldValue, this._mmValue);
		}
		this._input.trigger("blur");
	}
};

/**
 * Show/hide the unit label
 * @param {Boolean} toggle
 */
UserUnitInputField.prototype.toggleLabel = function(toggle) {
	this._unitLabel.toggleClass("displayNone", !toggle);
};

/**
 * Get the value of the field (in mm)
 * @returns {Number} the value of the field in mm
 */
UserUnitInputField.prototype.getValue = function() {
	return this._mmValue;
};

export default UserUnitInputField;
