import Button from "@esko/button/Button";
import Popover from "@esko/popover/Popover";
import ToolbarButton from "@esko/toolbar/ToolbarButton";
import $ from "jquery";
import _ from "underscore";
/**
 * A ViewerToolbarButton is an interface element which provides an action and can have a text and icon description
 *
 * @exports ViewerToolbarButton
 * @constructor
 * @param {Core.ViewerFacade} f - The viewer facade
 * @param {String} Name - The name of the ViewerToolbarButton
 * @param {String | JQueryObj } icon - The class which is used for displaying the icon, or your own jQuery element
 * @param {Boolean} [forMobile] - If true the ViewerToolbarButton is useful on a mobile device
 * @param {Function} [action] - The function which is called when the ViewerToolbarButton is pressed (also see listenToTool)
 */

var ViewerToolbarButton = function(f, name, icon, forMobile, action) {
	this.f = f;
	ToolbarButton.call(this, this.f.i18n, this.f.getContext(), this.f.isMobile(), name, icon, forMobile, action);
};

ViewerToolbarButton.prototype = Object.create(ToolbarButton.prototype);
ViewerToolbarButton.prototype.constructor = ViewerToolbarButton;

/**
 * Start/stop listening to a tool.
 * It automatically sets "toggle" behavior to the ViewerToolbarButton (next to the maybe already specified action).
 * @param {AbstractTool} tool - The tool it should listen to. If not specified or false, it stops listening to the tool.
 * @param {Boolean} enableDeactivation - If true, the tool will be deactivated when clicking again on the button, false disables this.
 */
ViewerToolbarButton.prototype.listenToTool = function(tool, enableDeactivation) {
	var onToolActivated = _.bind(function(interruptRequest) {
		//Only activate if the request was an actual request (not an interrupt)
		if (!interruptRequest) {
			this.setActive(true);
		}
	}, this);
	var onToolDeactivated = _.bind(function(interruptRequest) {
		//Only deactivate if the request was an actual request (not an interrupt)
		if (!interruptRequest) {
			this.setActive(false);
		}
	}, this);
	if (tool) {
		var toggleTool = _.bind(function() {
			if (this.isActive()) {
				if (enableDeactivation) {
					this.f.emit("toolManager.deactivateTool", tool);
				}
			} else {
				this.f.emit("toolManager.activateTool", tool);
			}
		}, this);
		// Set the action if not specified yet or append the action
		if (!this.action) {
			this.action = toggleTool;
		} else {
			this.origaction = this.action;
			this.action = _.bind(function() {
				this.origaction();
				toggleTool();
			}, this);
		}

		// Register for future tool activations/deactivations
		tool.on("activated", onToolActivated);
		tool.on("deactivated", onToolDeactivated);

		// Make sure that if the tool is already active, we activate the ViewerToolbarButton
		this.f.emit("toolManager.activeTool", function(activeTool) {
			if (activeTool && activeTool === tool) {
				onToolActivated();
			} else {
				onToolDeactivated();
			}
		});
	} else {
		tool.off("activated", onToolActivated);
		tool.off("deactivated", onToolDeactivated);
	}
};

/**
 * Start/stop listening to a tool option.
 * If not all params specified, it stops listening
 * @param {String} toolName - Name of the tool
 * @param {String} optionName - Name of the option
 * @param {Object} value - Value this ViewerToolbarButton should react on
 */
ViewerToolbarButton.prototype.listenToToolOption = function(toolName, optionName, value) {
	this._listeningToToolOptionValue = value;

	var onToolOptionChanged = _.bind(function(oldValue, newValue) {
		this.setActive(newValue == this._listeningToToolOptionValue);
	}, this);
	if (toolName && optionName && typeof value !== undefined) {
		this.f.getCurrentUser().on(["options", "changed", optionName], onToolOptionChanged);
	} else {
		this.f.getCurrentUser().off(["options", "changed", optionName], onToolOptionChanged);
	}

	this.setActive(this.f.getOption(optionName) == this._listeningToToolOptionValue);
};

export default ViewerToolbarButton;
