import _ from "underscore";
import CoreConstants from "@esko/web-viewer/core/CoreConstants";
import LayoutConstants from "@esko/web-viewer/layout/LayoutConstants";
import LayoutMode from "@esko/web-viewer/layout/LayoutMode";
import View3DLayerModel from "@esko/web-viewer/view3D/View3DLayerModel";
/**
 * @exports Collada3DMode
 * @memberof Layout
 * @class
 * Mode for mobile layout, shows the annotation widget
 * @constructor
 * @augments LayoutMode
 * @param {Core.ViewerFacade} f
 * @param {Layout} layout
 */
var Collada3DMode = function(f, layout) {
	LayoutMode.call(this, f, layout, LayoutConstants.LayoutMode.THREEDONLY);

	this.f.whenModuleAvailable(
		"annotation",
		function(annotModule) {
			f.whenModuleAvailable(
				"view3D",
				function(view3DModule) {
					var model = annotModule.annotationModel;

					// Intercept visibility of annotations (only show if camera angles match or when selected)
					model.interceptAnnotationVisibility(
						function(annot) {
							if (this._isActive) {
								const cameraAngle = view3DModule.getCameraPosition();
								const annotAngle = annot.get3DViewPosition();
								const isMatch = view3DModule.isAngleEqual(cameraAngle, annotAngle);
								return view3DModule.isUsingAnnotations() && (model.isAnnotationOrReplySelected(annot) || isMatch);
							}
							return true;
						}.bind(this)
					);

					//Store selected annotation before saving
					model.on(
						"save.started",
						function() {
							this._previouslySelectedAnnotation = _.first(model.getSelectedAnnotations());
						}.bind(this)
					);

					//Restore selected annotation after saving
					model.on(
						"save.finished",
						function() {
							if (this._previouslySelectedAnnotation && this._isActive) {
								var annot = _.find(
									model.getAnnotations(),
									function(annot) {
										return annot.name === this._previouslySelectedAnnotation.name;
									}.bind(this)
								);
								if (annot) {
									model.selectAnnotation(annot);
								}
							}
						}.bind(this)
					);
				}.bind(this)
			);
		}.bind(this)
	);
};

Collada3DMode.prototype = new LayoutMode();
Collada3DMode.prototype.constructor = Collada3DMode;

/**
 * Turns On/Off the layoutMode
 * @param {Boolean} toggle
 */
Collada3DMode.prototype.toggle = function(toggle) {
	if (toggle === undefined) {
		toggle = !this._isActive;
	}
	this._isActive = toggle;

	var world = this.f.getModule("viewport").getWorld();
	world.allowRotation(!this._isActive);
	world.allowPan(!this._isActive);
	world.allowZoom(!this._isActive);
	world.allowFlip(!this._isActive);
	world.allowMiddle(!this._isActive);

	this.f.whenModuleAvailable(
		"viewport",
		function(viewportModule) {
			viewportModule.whenWidgetModelAvailable(
				function() {
					if (this._isActive) {
						viewportModule.addLayer("View3D", new View3DLayerModel(this.f, "View3D", this));
					} else {
						viewportModule.removeLayer("View3D");
					}
				}.bind(this)
			);
		}.bind(this)
	);

	if (this._isActive) {
		this.f.emit("toolManager.request.allowAllToolsExcept", [CoreConstants.TOOLS.PAN, CoreConstants.TOOLS.ZOOM]);
	} else {
		this.f.emit("toolManager.request.allowAllTools");
	}
};

/**
 * Checks if the item is allowed for this layout mode. Prevents channels and CAD widget from loading.
 *
 * @param {String} itemname
 * @returns {Boolean}
 */
Collada3DMode.prototype.isItemAllowed = function(itemname) {
	return (
		itemname === LayoutConstants.Widget.COMPARE ||
		itemname === LayoutConstants.ToolbarComponent.APPROVAL ||
		itemname === LayoutConstants.Widget.APPROVAL ||
		itemname === LayoutConstants.Widget.MESSAGE ||
		itemname === LayoutConstants.Widget.DISCUSSION ||
		itemname === LayoutConstants.ToolbarComponent.DOCUMENTSELECTOR ||
		itemname === LayoutConstants.ToolbarComponent.COMPARE ||
		itemname === LayoutConstants.Widget.ATTRIBUTE_CATEGORY ||
		itemname === LayoutConstants.Widget.VIEWPORT ||
		itemname === LayoutConstants.ToolbarComponent.ANNOTATION ||
		itemname === LayoutConstants.Widget.ANNOTATION
	);
};

/**
 * We want the 3D widget as Main Content
 * @override
 */
Collada3DMode.prototype.getMainPanelContent = function() {
	// Collada3DMode ensures that an instance of Viewport3DLayer is added to the Viewport widget (in order to support annotations)
	return [LayoutConstants.Widget.VIEWPORT];
};

export default Collada3DMode;
