import Layout from "@esko/web-viewer/layout/Layout";
import DraggableWidgets from "@esko/web-viewer/layout/DraggableWidgets";
import WidgetWrapper from "@esko/web-viewer/layout/WidgetWrapper";
import ToolbarComponentWrapper from "@esko/web-viewer/layout/ToolbarComponentWrapper";
import LayoutConstants from "@esko/web-viewer/layout/LayoutConstants";
import Panel from "@esko/web-viewer/layout/desktop/Panel";
import SidePanel from "@esko/web-viewer/layout/desktop/SidePanel";
import BasicMode from "@esko/web-viewer/layout/desktop/BasicMode";
import RestrictedMode from "@esko/web-viewer/layout/desktop/RestrictedMode";
import WizardMode from "@esko/web-viewer/layout/desktop/WizardMode";
import Desktop3DMode from "@esko/web-viewer/layout/desktop/Desktop3DMode";
import Desktop2D3DMode from "@esko/web-viewer/layout/desktop/Desktop2D3DMode";
import Collada3DMode from "@esko/web-viewer/layout/desktop/Collada3DMode";
import DesktopVideoMode from "@esko/web-viewer/layout/desktop/DesktopVideoMode";
import PrivateChannel from "@esko/web-viewer/core/PrivateChannel";
import ShapeUtils from "@esko/shapeutils";
import ListenerUtil from "@esko/web-viewer/utils/ListenerUtil";
import FullscreenUtil from "@esko/web-viewer/utils/FullscreenUtil";
import CoreConstants from "@esko/web-viewer/core/CoreConstants";
import Compare3DMode from "@esko/web-viewer/view3D/Compare3DMode";
import _ from "underscore";
import $ from "jquery";
/**
 * @const minimum width of viewport before autocollapse occurs
 */
var AUTOCOLLAPSE_THRESHOLD = 45;
/**
 * @const default width of panels
 */
var DEFAULT_PANELWIDTH = 300;

/**
 * @class
 * Layout for desktop screens. This is the view initialized by the LayoutManager that will add all the layoutcomponents
 * to the viewer. It can generate a String representing the current state of the viewer and can be initialized by
 * a state-String.
 *
 * The desktop layout defines 4 main content areas: nord, west, east and center. These hold different kinds of
 * components. The layout is not responsible for rendering in the viewport.
 *
 * @exports DesktopLayout
 * @constructor
 * @param {Core.ViewerFacade} f
 * @param {LayoutModel} layoutModel
 * @param {JQueryObj} context - context where the layout will add its components to
 *                            @param {Object} layoutState
 * @param {LoadingIndicator} loadingIndicator - the loading indicator, which was created before all requests started
 * @param {JQueryObj} initialLoader - the JQuery reference to the element indicating the initial loader (needed to safely remove it without emptying the context)
 */
var DesktopLayout = function(f, layoutModel, context, layoutState, loadingIndicator, initialLoader) {
	Layout.call(this, f, layoutModel, context, layoutState, loadingIndicator, initialLoader);
	ListenerUtil.mixin(this, f);
	PrivateChannel.installTo(this, f, "layout.desktop");

	/**
	 * Default state of the layout, used to reset the layout
	 */
	this._defaultLayout = {};

	/**
	 * Mapping for all the items, maps items to their section
	 */
	this._sectionMapping = {};

	/**
	 * All the layout items mapped as Name -> WidgetWrapper
	 */
	this._items = {};

	/**
	 * The component that is used to display the content in the viewport: graphics, CAD, annotations,...
	 */
	this._mainContent = null;

	this._mainContentWidgetWrappers = [];

	/**
	 * Inidicates whether the CSS is available or not
	 */
	this._cssLoaded = false;

	this._wrapper = $("<div/>").addClass("wrapper disableTextSelection");
	this._north = $("<div/>").addClass("northPane");
	this._mainWrapper = $("<div/>");
	this._topbar = $("<div/>").addClass("topbar");
	this._messagebar = $("<div />").addClass("messagebar");
	this._west = $("<div/>").addClass("westPane");
	this._innerWest = $("<div/>")
		.addClass("accordeon")
		.addClass("inner");
	this._center = $("<div/>").addClass("centerPane");
	this._mainContentPanel = $("<div/>").addClass("mainContentPanel");
	this._east = $("<div/>").addClass("eastPane");
	this._innerEast = $("<div/>")
		.addClass("accordeon")
		.addClass("inner");
	this._aboveCenter = $("<div/>").addClass("aboveCenter");
	this._belowCenter = $("<div/>").addClass("belowCenter northPane");

	var modes = this._getLayoutModes();

	modes.forEach(
		function(mode) {
			this._layoutModel.registerLayoutMode(mode);
			mode.on(
				"changed",
				function() {
					this._updateMainContent();
				}.bind(this)
			);
		}.bind(this)
	);

	this.f.emit("layout.finishRegisteringModes");

	//initialize the correct mode for the desktoplayout
	this._switchLayoutMode();

	//DOM-elements to build the layout

	//construct the layout with the previously defined DOM-elements
	this._west.append(this._innerWest);
	this._east.append(this._innerEast);
	this._mainWrapper
		.append(this._west)
		.append(this._aboveCenter)
		.append(this._center.append(this._mainContentPanel).append(this._belowCenter))
		.append(this._east);
	this._wrapper
		.append(this._topbar)
		.append(this._north)
		.append(this._messagebar)
		.append(this._mainWrapper)
		.append(this._loadingIndicator.element);

	/**
	 * Information on the current layout (state of the panels,...)
	 */
	this._layout = {
		width: 0,
		height: 0,
		top: new Panel(f, "topbar", this._topbar),
		north: new Panel(f, "north", this._north),
		east: new SidePanel(f, "east", this._east, this._innerEast),
		west: new SidePanel(f, "west", this._west, this._innerWest),
		center: new Panel(f, "center", this._center),
		below: new Panel(f, "below", this._belowCenter)
	};

	this._layout.east.setWidth(DEFAULT_PANELWIDTH);
	this._layout.west.setWidth(DEFAULT_PANELWIDTH);

	this._enforceRestrictedMode();

	/**
	 * Maps all sections to their JQuery-objects
	 */
	this._sections = {
		default: this._innerWest,
		west: this._innerWest,
		north: this._north,
		east: this._innerEast,
		center: this._mainWrapper,
		below: this._belowCenter
	};

	//a debounced invalidate, to make sure not to many invalidates are triggered
	this._debouncedInvalidate = _.debounce(_.bind(this._invalidateLayout, this), 200);
	this._deferredInvalidate = function() {
		_.defer(_.bind(this._invalidateLayout, this));
	};

	/**
	 * Invalidate of the panels in the layout is debounced (with 100ms)
	 */
	this._debouncedPanelInvalidate = _.debounce(
		_.bind(function() {
			this._layout.east.invalidate();
			this._layout.west.invalidate();
			this._emitLayoutChanged();
		}, this),
		100
	);

	this._loadCss();

	this._cssLoaded = true;
	this._initialLoader.remove();
	delete this._initialLoader;
	this._context.append(this._wrapper);
	this._deferredInvalidate();

	//listen to resize events of the window
	$(window).on(
		"resize",
		_.bind(function(e) {
			// http://bugs.jqueryui.com/ticket/7514: resize of popups bubbles to window
			if (e.target === window) {
				this._debouncedInvalidate();
			}
		}, this)
	);

	// constrain all widgets when window is resized
	$(window).on(
		"resize",
		_.debounce(
			_.bind(function(e) {
				// http://bugs.jqueryui.com/ticket/7514: resize of popups bubbles to window
				if (e.target === window) {
					this._constrainAllModalWidgets();
				}
			}, this),
			500
		)
	);

	//make sure the default layout is initialized
	this._initializeDefaultLayout();
	//initialize the section mapping of the layoutItems
	this._initializeSectionMapping(this._layoutState);
	//initialize dragging for widgets
	this._initializeDragging();
	//initialize panels, size and visibility
	this._initializePanels();
	//initialize all behaviour for resizing/hiding panels
	this.initializePanelInteraction();
	//listen to widget events, to act on resized/collapsed/...
	this._registerOnLayoutEvents();

	// webkit has a bug with iFrames, which appearantly they won't be solving soon
	// if an input in an iframe is focused, the page scrolls to the iframe
	// this can occur when TinyMCE is used... we can fix it by scrolling to top whenever the context is scrolled
	context.on("scroll", function(event) {
		$(this).scrollTop(0);
	});

	this.on(
		"changed.layoutMode",
		_.bind(function() {
			this._updateMainContent();
		}, this)
	);

	this._invalidateLayout(); //this one is needed to provide the correct world coordinates
};

DesktopLayout.prototype = Object.create(Layout.prototype);
DesktopLayout.prototype.constructor = DesktopLayout;

/**
 * Initializes the default layout. This is the layout that's used when the viewer is opened for the first time.
 * The layout can be reseted to this layout. It maps all the components to the correct section.
 */
DesktopLayout.prototype._initializeDefaultLayout = function() {
	this._defaultLayout = {
		north: [],
		east: [],
		center: [],
		west: []
	};
	this._defaultLayout.north = [
		LayoutConstants.ToolbarComponent.PAGELIST,
		LayoutConstants.ToolbarComponent.NAVIGATE,
		LayoutConstants.ToolbarComponent.MEASURE,
		LayoutConstants.ToolbarComponent.TEXTSELECTION,
		LayoutConstants.ToolbarComponent.COMPARE,
		LayoutConstants.ToolbarComponent.ANNOTATION,
		LayoutConstants.ToolbarComponent.DYNCONTENT,
		LayoutConstants.ToolbarComponent.VIEW3D,
		LayoutConstants.ToolbarComponent.APPROVAL
	];
	this._defaultLayout.east = [
		LayoutConstants.Widget.NAVIGATOR,
		LayoutConstants.Widget.VIEW3D,
		LayoutConstants.Widget.VIEWCONTROLS,
		LayoutConstants.Widget.CHANNELS,
		LayoutConstants.Widget.CAD,
		LayoutConstants.Widget.DEBUG,
		LayoutConstants.Widget.APPROVAL,
		LayoutConstants.Widget.INFO,
		LayoutConstants.Widget.HELP,
		LayoutConstants.Widget.VIDEO_KEYFRAME
	];
	this._defaultLayout.west = [
		LayoutConstants.Widget.COMPARE,
		LayoutConstants.Widget.ATTRIBUTE_CATEGORY,
		LayoutConstants.Widget.ANNOTATION,
		LayoutConstants.Widget.PAGELIST,
		LayoutConstants.Widget.TEXTCONTENT,
		LayoutConstants.Widget.DISCUSSION,
		LayoutConstants.Widget.VIEWPORT,
		LayoutConstants.Widget.VIEW3D_COMPARE
	];
	this._defaultLayout.below = [LayoutConstants.ToolbarComponent.VIDEO];
};

DesktopLayout.prototype._getLayoutModes = function() {
	return [
		new RestrictedMode(this.f, this),
		new BasicMode(this.f, this),
		new WizardMode(this.f, this),
		new Collada3DMode(this.f, this),
		new Desktop3DMode(this.f, this),
		new DesktopVideoMode(this.f, this),
		new Desktop2D3DMode(this.f, this),
		new Compare3DMode(this.f, this)
	];
};

/**
 * Initializes the section mapping. All items will be mapped to their section so it's easier to find the item's section.
 * Items that aren't in the layoutstate but are in the default state will also be added to the mapping.
 *
 * @param {Object} [layoutState] - if not defined only the default layout is used for the mapping
 */
DesktopLayout.prototype._initializeSectionMapping = function(layoutState) {
	//first run over all the sections and items in the layoutState
	if (layoutState !== undefined && layoutState !== "" && layoutState != {}) {
		_.each(
			layoutState,
			function(section, sectionName) {
				//runs over all sections
				_.each(
					section.items,
					function(item) {
						//runs over all items in those sections
						if (item) {
							this._sectionMapping[item.name] = sectionName; //map each itemName->sectionName
						}
					},
					this
				);
			},
			this
		);
	}
	//run over the defaultLayout, if items of the default layout aren't added yet, add them to the mapping
	_.each(
		this._defaultLayout,
		function(items, sectionName) {
			_.each(
				items,
				function(item) {
					if (this._sectionMapping[item] === undefined) {
						//check if item isn't added by layoutState yet
						this._sectionMapping[item] = sectionName;
					}
				},
				this
			);
		},
		this
	);
};

/**
 *  Add a class to the wrapper based on the layout-mode
 */
DesktopLayout.prototype._switchLayoutMode = function(mode) {
	if (this._activeMode) {
		this._wrapper.removeClass("layout-mode-" + this._activeMode.getName());
	}
	Layout.prototype._switchLayoutMode.call(this, mode);
	if (this._activeMode) {
		this._wrapper.addClass("layout-mode-" + this._activeMode.getName());
	}
};

/**
 * Register listeners for general layout event.
 */
DesktopLayout.prototype._registerOnLayoutEvents = function() {
	this.f.on(
		"layout.disableTextSelection",
		_.bind(function() {
			this._disableTextSelection();
		}, this)
	);
	this.f.on(
		"layout.enableTextSelection",
		_.bind(function() {
			this._enableTextSelection();
		}, this)
	);
	this.f.on("layout.resetToDefault", _.bind(this.resetToDefault, this));
};

/**
 * Initialize draggable widgets, makes it possible to move widgets.
 */
DesktopLayout.prototype._initializeDragging = function() {
	this._draggableWidgets = new DraggableWidgets(this._wrapper);
	//add dropzones for widgets, these are the parts widgets can be moved to
	this._draggableWidgets.addModalWidgetDropzone(this._center, "center", this);
	this._draggableWidgets.addSortableWidgetDropzone(this._innerWest, "west", this);
	this._draggableWidgets.addSortableWidgetDropzone(this._innerEast, "east", this);
};

/**
 * Initialize panel interaction. Adds listeners to the global wrapper div-element. It checks if
 * the users wants to resize/collapse the side panels.
 */
DesktopLayout.prototype.initializePanelInteraction = function() {
	var THRESHOLD = 7;
	var draggingHandle;
	var previousDragX;
	var maxWidth;
	var minWidth = 100;
	var hideTreshold = 30;
	var dragDirection = 1;

	var inWestHandle = _.bind(function(x, y) {
		return this._layout.west.isVisible() && Math.abs(x - this._layout.west.getWidth()) < THRESHOLD && y >= this._layout.west.position().top;
	}, this);

	var inEastHandle = _.bind(function(x, y) {
		return (
			this._layout.east.isVisible() &&
			Math.abs(x - this._layout.west.getWidth() - this._layout.center.getWidth()) < THRESHOLD &&
			y >= this._layout.east.position().top
		);
	}, this);

	this._registerListener(
		"mousedown",
		function(e) {
			if (inWestHandle(e.clientX, e.clientY)) {
				draggingHandle = this._layout.west;
				dragDirection = 1;
			} else if (inEastHandle(e.clientX, e.clientY)) {
				draggingHandle = this._layout.east;
				dragDirection = -1;
			}
			if (draggingHandle) {
				previousDragX = e.clientX;
				maxWidth = draggingHandle.getWidth() + this._layout.center.getWidth() - (AUTOCOLLAPSE_THRESHOLD + 5);
				this.f.emit("layout.disableTextSelection");
			}
		},
		this,
		this._wrapper
	);

	this._registerListener(
		"mousemove",
		function(e) {
			if (draggingHandle !== undefined || inWestHandle(e.clientX, e.clientY) || inEastHandle(e.clientX, e.clientY)) {
				this._wrapper.toggleClass("panelResize", true);
			} else {
				this._wrapper.toggleClass("panelResize", false);
			}
			if (draggingHandle !== undefined) {
				var diff = dragDirection * (e.clientX - previousDragX);
				var newWidth = draggingHandle.getWidth() + diff;

				if (newWidth > minWidth && newWidth < maxWidth) {
					previousDragX = e.clientX;
					draggingHandle.setWidth(newWidth);
					this._layout.center.setWidth(
						this._layout.width - (this._layout.east.outerWidth(true) || null) - (this._layout.west.outerWidth(true) || null)
					);
				}

				if (newWidth < hideTreshold) {
					draggingHandle.hide();
					draggingHandle.setWidth(DEFAULT_PANELWIDTH);
					draggingHandle = undefined;
					this.f.emit("layout.enableTextSelection");
					this._deferredInvalidate();
				}
			}
		},
		this,
		this._wrapper
	);

	this._registerListener(
		"mouseout",
		function(e) {
			this._wrapper.toggleClass("panelResize", false);
			if (draggingHandle && e.target == this._wrapper[0]) {
				draggingHandle = undefined;
				this.f.emit("layout.enableTextSelection");
				this._deferredInvalidate();
			}
		},
		this,
		this._wrapper
	);

	this._registerListener(
		"mouseup",
		function(e) {
			if (draggingHandle) {
				this.f.emit("layout.enableTextSelection");
				draggingHandle = undefined;
				this._deferredInvalidate();
			}
		},
		this,
		this._wrapper
	);
};

/**
 * Disable all possible interaction with the side panels.
 */
DesktopLayout.prototype.disablePanelInteraction = function() {
	this._unregisterAllListenersForObject(this._wrapper);
};

/**
 * Initializes the panels, gets the collapsed/width state from the layoutState.
 */
DesktopLayout.prototype._initializePanels = function() {
	if (this._layoutState) {
		_.each(
			this._layoutState,
			function(section, sectionName) {
				//run over all sections
				if (this._layout[sectionName] !== undefined) {
					this._layout[sectionName].restoreState(section);
				}
			},
			this
		);
	}

	// listen to panels to invalidate layout when toggled
	this._registerListener("changed.visibility", this._deferredInvalidate, this, this._layout.west);
	this._registerListener("changed.visibility", this._deferredInvalidate, this, this._layout.east);

	this._toggleButtonWest = this._layout.west.toggleButton.element;
	this._toggleButtonEast = this._layout.east.toggleButton.element;
	this._wrapper.append(this._toggleButtonWest);
	this._wrapper.append(this._toggleButtonEast);
};

/**
 * Returns the default layout mode for the layout.
 * @return {LayoutMode}
 */
Layout.prototype._getDefaultLayoutMode = function() {
	if (this.f.getProperty("viewerStyle") == LayoutConstants.ViewerStyle.RESTRICTED) {
		//is the viewer loaded in restricted style (hide channels and CAD widgets)
		return this._layoutModel.getLayoutMode(LayoutConstants.LayoutMode.RESTRICTED);
	} else {
		return this._layoutModel.getLayoutMode(LayoutConstants.LayoutMode.VIEW2D);
	}
};

/**
 * Add an item to the layout. This is called by Layout to add an item.
 * @param {Object} layoutItem
 * @example
 * 	{ 	type: {LayoutConstants.LayoutItemType},
 * 		name: 'itemName',
 * 		item: {AbstractWidget/ToolbarComponent}
 * 	}
 */
DesktopLayout.prototype._addItem = function(layoutItem) {
	if (layoutItem.type === LayoutConstants.LayoutItemType.WIDGET) {
		var widget = layoutItem.getItem(this.getLayoutType());
		var widgetType = widget.getWidgetType();
		if (this._activeMode.getMainPanelContent().indexOf(layoutItem.name) !== -1) {
			widgetType = "maincontent";
		}
		switch (
			widgetType //different layoutitems are of different widgetType
		) {
			case "topbar": //widget needs to be added to the topbar
				widget.setContext(this._topbar);
				break;
			case "messagebar": //message bar is visible in the centerpanel
				widget.setContext(this._messagebar);
				break;
			case "wizard":
				widget.setContext(this._aboveCenter);
				this._items[widget.getID()] = widget;
				break;
			default:
				//default widgets are part of either the sidepanels or the center panel
				if (!this._items[layoutItem.name]) {
					var widgetWrapper = new WidgetWrapper(widget, layoutItem.name, true);

					this._draggableWidgets.enableDragging(widgetWrapper);
					this._items[layoutItem.name] = widgetWrapper;
					if (widgetType === "maincontent") {
						this._mainContentPanel.prepend(widgetWrapper.element);
						this._mainContentWidgetWrappers.push(widgetWrapper);
						this._updateMainContent();
					} else {
						this._addLayoutItem(layoutItem.name); //make sure the item is added on the correct location in one of the panels
						if (!this._layout[this._sectionMapping[layoutItem.name]].isVisible()) {
							widgetWrapper.widget.setVisible(false);
						}
					}
					this._registerListener(
						"collapsed",
						function() {
							if (widgetWrapper.isModal) {
								//if model no invalidate is needed as the widget is not part of a side panel
								this._emitLayoutChanged();
							} else {
								this._debouncedPanelInvalidate();
							}
						},
						this,
						widget
					);
					this._registerListener(
						"modalResize",
						function() {
							if (widgetWrapper.isModal) {
								//make sure the state is saved when modal widget is moved
								this._emitLayoutChanged();
							}
						},
						this,
						widget
					);
					this._registerListener(
						"dragged",
						function() {
							//the item is dragged, change its location
							if (widgetWrapper.isModal) {
								this._emitLayoutChanged();
							}
						},
						this,
						widget
					);

					this._registerListener(
						"enabled",
						function(isEnabled) {
							this._toggleWidget(layoutItem.name, widgetWrapper, isEnabled);
							this._debouncedPanelInvalidate();
						},
						this,
						widget
					);

					this._registerListener("resized", this._debouncedPanelInvalidate, this, widget);
					this._registerListener("makeVisible", _.bind(this._makeWidgetVisible, this, layoutItem.name, widgetWrapper), undefined, widget);

					if (widget.didRequestMakeVisible()) {
						this._makeWidgetVisible(layoutItem.name, widgetWrapper);
					}
				}
				//register listeners to the widget to make sure the layout gets invalidated (if needed) or the state is saved
				this._debouncedInvalidate();
				this._updateItemVisibility();

				break;
		}
	} else if (layoutItem.type === LayoutConstants.LayoutItemType.TOOLBARCOMPONENT) {
		//add a toolbarcomponent
		var toolbarItem = layoutItem.getItem(this.getLayoutType());
		var toolbarItemWrapper = new ToolbarComponentWrapper(toolbarItem);
		toolbarItemWrapper.element[0].name = layoutItem.name;
		this._items[layoutItem.name] = toolbarItemWrapper;
		this._addLayoutItem(layoutItem.name);
		this._registerListener(
			"changed.visibility",
			function() {
				this._invalidateNorthPane();
				this._debouncedInvalidate();
			},
			this,
			toolbarItem
		);
		this._debouncedInvalidate();
		this._updateItemVisibility();
	} else {
		//not supported in this layout
	}
};

/**
 * Add an item to the layout, the items needs to already added to the items-collection.
 * It adds the item to the corect element on the correct location (depends on the previous/default state and already added items)
 * @param {String} itemName
 */
DesktopLayout.prototype._addLayoutItem = function(itemName) {
	var section = this._sectionMapping[itemName]; //get the section the items belongs to
	if (section !== undefined) {
		//make sure there is a section to add the item to
		var sectionItems = null;
		var itemState = null;
		if (this._layoutState && this._layoutState[section]) {
			sectionItems = _.map(this._layoutState[section].items, function(item) {
				return item.name;
			}); //just get all the names of the items in this section, not other information
			itemState = _.find(this._layoutState[section].items, function(item) {
				return item.name === itemName;
			}); //get the state of the to be added item
		} else {
			sectionItems = this._defaultLayout[section]; //default state only holds names of the items
		}
		if (sectionItems !== null && sectionItems !== undefined) {
			//the section should at least hold the item itself
			var index = _.indexOf(sectionItems, itemName);
			var nextItem = null;
			var itemsToLookAt = null;
			if (index === -1) {
				index = _.indexOf(this._defaultLayout[section], itemName);
				itemsToLookAt = this._defaultLayout[section];
			} else {
				itemsToLookAt = sectionItems;
			}

			nextItem = this._items[itemsToLookAt[++index]];
			if (_.contains(this._mainContentWidgetWrappers, nextItem)) {
				nextItem = undefined;
			}
			while (index < itemsToLookAt.length && nextItem === undefined) {
				//as long nextItem is not yet added to the layout, search for a next one
				nextItem = this._items[itemsToLookAt[++index]];
				if (_.contains(this._mainContentWidgetWrappers, nextItem)) {
					nextItem = undefined;
				}
			}

			if (nextItem === undefined) {
				//no nextItem is found, the item is (for now) the last item in the section
				this._appendItemToSection(itemName, section); //append the item to the section
			} else {
				//add the item in front of the found nextItem
				this._prependItemBeforeItem(nextItem.name, itemName, section);
			}
			//now be sure to set the correct state of the item (collapsed/size/location)
			if (itemState !== null && itemState !== undefined) {
				this._setCollapsed(itemName, itemState.isCollapsed);
				if (itemState.width !== undefined || itemState.height !== undefined) {
					this._setSize(itemName, new ShapeUtils.Size(itemState.width, itemState.height));
				}
				if (itemState.left !== undefined || itemState.top !== undefined) {
					this._setLocation(itemName, new ShapeUtils.Point(itemState.left, itemState.top));
				}
			}
		}
	}
};

/**
 * Appends an already added item (to the items collection) to a section.
 * @param {String} itemName
 * @param {String} section
 */
DesktopLayout.prototype._appendItemToSection = function(itemName, section) {
	var layoutItem = this._items[itemName];
	if (layoutItem.widget) {
		layoutItem.widget.movingStarted();
	}

	this._sections[section].append(layoutItem.element);

	if (layoutItem.widget) {
		layoutItem.widget.movingStopped();
	}
};

/**
 * Prepends an already added item (to the items collection) before the nextItem in a given section.
 * @param {String} nextItemName
 * @param {String} itemName
 * @param {String} section
 */
DesktopLayout.prototype._prependItemBeforeItem = function(nextItemName, itemName, section) {
	var layoutItem = this._items[itemName];
	if (layoutItem.widget) {
		layoutItem.widget.movingStarted();
	}
	//add item before other item in the layout
	this._items[nextItemName].element.before(layoutItem.element);

	if (layoutItem.widget) {
		layoutItem.widget.movingStopped();
	}
};

/**
 * Moves an item to another section. Is used by draggable widgets to let the layout know it is moved.
 * @param {String} itemName
 * @param {String} sectionName
 * @param {String} nextItemName - item will be moved before this item
 * @param {Boolean} resetSizeAndLocation
 * @param {Boolean} resetCollapsed
 */
DesktopLayout.prototype.moveItem = function(itemName, sectionName, nextItemName, resetSizeAndLocation, resetCollapsed) {
	if (itemName !== nextItemName) {
		if (resetSizeAndLocation) {
			//resets the size and location of the item (moves from center to a side-panel)
			this._setLocation(itemName, null);
			this._setSize(itemName, null);
		}
		if (resetCollapsed) {
			//resets the collapsed state of the item
			this._setCollapsed(itemName, false);
		}
		var currentSection = this._sectionMapping[itemName];
		this._detachItem(itemName, currentSection); //removes the item from the previous section
		this._sectionMapping[itemName] = sectionName;
		if (nextItemName === undefined || this._sectionMapping[nextItemName] !== sectionName) {
			this._appendItemToSection(itemName, sectionName);
		} else {
			//check if no problems with after parameter
			this._prependItemBeforeItem(nextItemName, itemName, sectionName);
		}
	}
};

/**
 * Set the location of an item. It will make sure the item stays within the window bounds.
 * If no location is provide the item will no longer be modal (is moved to one of the side-panels)
 * @param {String} itemName
 * @param {ShapeUtils.Point} location
 */
DesktopLayout.prototype._setLocation = function(itemName, location) {
	var widgetWrapper = this._items[itemName];

	if (widgetWrapper !== undefined && widgetWrapper.widget) {
		if (location === null || location === undefined) {
			widgetWrapper.setModal(false);
		} else {
			this._constrainModalWidgetWrapper(widgetWrapper, location);
		}
	}
};

/**
 * Constrain a widgetwrapper to the windowsize/viewerapp size
 * @param {WidgetWrapper} widgetWrapper - the widgetWrapper that should be constrained to the window
 * @param {ShapeUtils.Size}   location - the wanted location of the the widgetwrapper
 */
DesktopLayout.prototype._constrainModalWidgetWrapper = function(widgetWrapper, location) {
	var itemSize = widgetWrapper.getOuterSize();

	//get the location so the widget stays inside the window, will change the location if the widget lays outside the window
	if (location.x > (this._context.innerWidth() || null) - itemSize.width) {
		location.x = (this._context.innerWidth() || null) - itemSize.width;
	}
	if (location.y > (this._context.innerHeight() || null) - itemSize.height) {
		location.y = (this._context.innerHeight() || null) - itemSize.height;
	}

	widgetWrapper.setModal(true, location.y, location.x);
};

/**
 * Make sure all modal widgets are visible
 */
DesktopLayout.prototype._constrainAllModalWidgets = function() {
	_.each(
		this._items,
		function(item) {
			if (item.isModal) {
				var widgetWrapper = this._items[item.name];
				if (widgetWrapper !== undefined && widgetWrapper.widget) {
					this._constrainModalWidgetWrapper(widgetWrapper, item.getLocation());
				}
			}
		},
		this
	);
};

/**
 * Set a widget to collapsed (only possible if the layout item is a widget)
 * @param {String} itemName
 * @param {Boolean} collapsed
 */
DesktopLayout.prototype._setCollapsed = function(itemName, collapsed) {
	var widgetWrapper = this._items[itemName];
	if (widgetWrapper !== undefined && widgetWrapper.widget) {
		widgetWrapper.setCollapsed(collapsed);
	}
};

/**
 * Hide side panel.
 * @param {String} panelName
 */
DesktopLayout.prototype.disablePanel = function(panelName) {
	this._layout[panelName].setDisabled(true);
};

/**
 * Show side panel.
 * @param {String} panelName
 */
DesktopLayout.prototype.enablePanel = function(panelName) {
	this._layout[panelName].setDisabled(false);
};

/**
 * Is the panel visible.
 * @param {String} panelName
 * @returns {Boolean}
 */
DesktopLayout.prototype.isPanelVisible = function(panelName) {
	return this._layout[panelName].isVisible();
};

/**
 * Sets the size of the item, if no size is provided the item will no longer be modal (it is added to one of the side panels)
 * @param {String} itemName
 * @param {ShapeUtils.Size} size
 */
DesktopLayout.prototype._setSize = function(itemName, size) {
	var widgetWrapper = this._items[itemName];
	if (widgetWrapper !== undefined && widgetWrapper.widget) {
		if (size === null || size === undefined) {
			widgetWrapper.setModal(false);
		} else {
			var widgetWrapperLocation = widgetWrapper.getLocation();
			widgetWrapper.setModal(true, widgetWrapperLocation.y, widgetWrapperLocation.x, size.width, size.height);
		}
	}
};

/**
 * Remove an item from a section, it removes the item from the DOM, but doesn't delete the DOM-element.
 * @param {String} itemName
 * @param {String} sectionName
 */
DesktopLayout.prototype._detachItem = function(itemName, sectionName) {
	var widgetWrapper = this._items[itemName];
	if (widgetWrapper) {
		if (widgetWrapper.widget) {
			widgetWrapper.widget.movingStarted();
		}

		//remove element from the DOM (doesn't delete DOM-element)
		this._items[itemName].element.detach();

		if (widgetWrapper.widget) {
			widgetWrapper.widget.movingStopped();
		}
		//the invalidate is only done after all other changes (e.g. re-add DOM-element) are done
		_.defer(
			_.bind(function() {
				this._invalidateLayout();
			}, this)
		);
	}
};

/**
 * Hide a layout item.
 * @param {LayoutConstants.Widget|LayoutConstants.ToolbarComponent} itemName
 */
DesktopLayout.prototype.hideItem = function(itemName) {
	var layoutItem = this._items[itemName];
	if (layoutItem !== undefined && layoutItem.context !== undefined) {
		layoutItem.context.hide();
	} else if (layoutItem !== undefined && layoutItem.element !== undefined) {
		layoutItem.element.hide();
	}
};

/**
 * Makes a hidden layout item visible.
 * @param {LayoutConstants.Widget|LayoutConstants.ToolbarComponent} itemName
 */
DesktopLayout.prototype.showItem = function(itemName) {
	var layoutItem = this._items[itemName];
	if (layoutItem !== undefined && layoutItem.context !== undefined) {
		layoutItem.context.show();
	} else if (layoutItem !== undefined && layoutItem.element !== undefined) {
		layoutItem.element.show();
	}
};

/**
 * Is the layout item visible.
 * @param {String} itemName
 * @returns {Boolean}
 */
DesktopLayout.prototype.isItemVisible = function(itemName) {
	var layoutItem = this._items[itemName];
	if (layoutItem !== undefined) {
		return layoutItem.isVisible();
	}
};

/**
 * Hide all modal widgets.
 * @returns {Array.<String>} - returns all the names of the hidden widgets
 */
DesktopLayout.prototype.hideAllModalWidgets = function() {
	var modalWidgets = [];
	_.each(
		this._items,
		function(item) {
			if (item.isModal) {
				modalWidgets.push(item.name);
				this.hideItem(item.name);
			}
		},
		this
	);
	return modalWidgets;
};

/**
 * Get all the currently visible popups
 * @returns {Array.<JQueryObj>}
 */
DesktopLayout.prototype.getAllVisiblePopups = function() {
	return this._context.find(".popup:visible");
};

/**
 * Disable text selection in the entire viewer.
 */
DesktopLayout.prototype._disableTextSelection = function() {
	this._context
		.find(".enableTextSelection")
		.toggleClass("enableTextSelection", false)
		.toggleClass("enableTextSelectionTempDisabled", true);
};

/**
 * Enable text selection in the entire viewer.
 */
DesktopLayout.prototype._enableTextSelection = function() {
	this._context
		.find(".enableTextSelectionTempDisabled")
		.toggleClass("enableTextSelection", true)
		.toggleClass("enableTextSelectionTempDisabled", false);
};

/**
 * Restore a certain state of the layout to the state provided. (not used at this moment)
 * @param {Object} state
 */
DesktopLayout.prototype.restore = function(state) {
	_.each(
		state,
		_.bind(function(section, sectionName) {
			this._layout[sectionName].restoreState(section);

			_.each(
				section.items,
				_.bind(function(item) {
					if (this._items[item.name] !== undefined && this._activeMode.isItemAllowed(item.name)) {
						if (item.isCollapsed) {
							this._items[item.name].setCollapsed(item.isCollapsed);
						}
						if (item.width && item.height) {
							this._items[item.name].setSize(new ShapeUtils.Size(item.width, item.height));
						}
						if (item.left && item.top) {
							this._items[item.name].setLocation(new ShapeUtils.Point(item.left, item.top));
						}
						this.moveItem(item.name, sectionName);
					}
				}, this)
			);
		}, this)
	);
};

/**
 * Reset the sate of the layout to the default state.
 */
DesktopLayout.prototype.resetToDefault = function() {
	_.each(
		this._defaultLayout,
		function(section, sectionName) {
			this._layout[sectionName].setDisabled(false); //make sure all panels are enabled
			this._layout[sectionName].toggle(true); //make all the panels visible (this is the default at this moment)
			if (sectionName === "west" || sectionName === "east") {
				//reset the width of the side panels to the default
				this._layout[sectionName].setWidth(DEFAULT_PANELWIDTH);
				this._layout[sectionName].show();
			}
			_.each(
				section,
				function(itemName) {
					//run over each item in the section
					var item = this._items[itemName];
					if (item !== undefined && this._activeMode.isItemAllowed(itemName) && !_.contains(this._mainContentWidgetWrappers, item)) {
						//move the item to its default location.
						this.moveItem(itemName, sectionName, undefined, true, true);
					}
				},
				this
			);
		},
		this
	);
	this._layoutState = undefined; //reset the state of the layout, it now is the default\

	this._enforceRestrictedMode();

	this._invalidateLayout();
};

/**
 * Generates the state of the layout as a string. It takes the not-visible elements into account so state of those
 * elements is not lost.
 * @returns {Object}
 */
DesktopLayout.prototype.getLayoutState = function() {
	var stateObject = {};
	_.each(
		this._sections,
		_.bind(function(section, sectionName) {
			if (this._layout[sectionName] !== undefined) {
				stateObject[sectionName] = this._layout[sectionName].getState();
				stateObject[sectionName].items = [];

				_.each(
					this._sections[sectionName].children(),
					_.bind(function(item) {
						if (item !== null && item !== undefined) {
							var itemWrapper = item.wrapper;
							var itemObject = {};
							if (itemWrapper === undefined) {
								itemObject.name = item.name;
								itemObject.isCollapsed = !$(item).is(":visible");
							} else {
								itemObject.name = itemWrapper.name;
								itemObject.isCollapsed = itemWrapper.isCollapsed;

								var itemLocation = itemWrapper.getLocation();
								if (itemLocation !== null && itemWrapper.isModal) {
									itemObject.left = itemLocation.x;
									itemObject.top = itemLocation.y;
								}
								var itemSize = itemWrapper.getSize();
								if (itemSize !== null && itemWrapper.isModal) {
									itemObject.width = itemSize.width;
									itemObject.height = itemSize.height;
								}
							}

							if (itemObject.name !== undefined) {
								stateObject[sectionName].items.push(itemObject);
							}
						}
					}, this)
				);
			}
		}, this)
	);

	//construct a mapping of all the items per section
	var layoutItemsPerSection = {};
	_.each(this._layoutState, function(section, sectionName) {
		layoutItemsPerSection[sectionName] = [];
		_.each(section.items, function(item) {
			if (item) {
				layoutItemsPerSection[sectionName].push(item.name);
			}
		});
	});

	//if there is a layout state defined get the state of non-visible element (elements that have not yet been added) from here.
	if (this._layoutState) {
		//runs over all the elements of the state, if the item wasn't added to the state yet, add it
		_.each(
			this._layoutState,
			function(section, sectionName) {
				_.each(
					section.items,
					function(item) {
						if (item) {
							//if the item is not part of the items-collection it has not yet been added.
							if (this._items[item.name] === undefined && this._layout[sectionName] !== undefined && this._sectionMapping[item.name]) {
								var section = layoutItemsPerSection[this._sectionMapping[item.name]]; //get the section for the item
								var itemIndex = _.indexOf(section, item.name); //index of the item in the section
								itemIndex++; //start running over all the items after the current item and check if they are in the items collection (present in the stateObject)
								while (itemIndex < section.length && this._items[section[itemIndex]] === undefined) {
									itemIndex++; //run over the default section untill an item is found that is present in the stateObject
								}
								if (itemIndex < section.length) {
									//there is an item found in items (sits in stateObject)
									//get the index of this item in the stateObject, place the new item before nextItem
									var nextItem = _.findWhere(stateObject[sectionName].items, {
										name: section[itemIndex]
									});
									var index = _.indexOf(stateObject[sectionName].items, nextItem);
									stateObject[sectionName].items.splice(index, 0, item);
								} else {
									stateObject[sectionName].items.push(item); //insert at the end, it's the last item
								}
							}
						}
					},
					this
				);
			},
			this
		);

		// The state of the current maincontentpanel should also be stored for when the layout mode switches and it no longer is the main content.
		// We will insert its state into 'stateObject' in the correct position
		var mainPanelContents = this._activeMode.getMainPanelContent();

		_.each(
			mainPanelContents,
			function(mainPanelContent) {
				// All states of layout items in the same section as the main content panel
				var layoutStateItems =
					this._layoutState[this._sectionMapping[mainPanelContent]] && this._layoutState[this._sectionMapping[mainPanelContent]].items;

				if (layoutStateItems) {
					// Map to the name to make the searching of index easier.
					var layoutStateItemsNames = layoutStateItems.map(function(item) {
						return item.name;
					});

					// this is the state of the main content panel
					var mainContentPanelLayout = _.find(layoutStateItems, function(item) {
						return item && item.name === mainPanelContent;
					});

					if (mainContentPanelLayout) {
						// these items are all items that come after the main content item.
						var itemsAfterMainContentItem = _.rest(layoutStateItemsNames, layoutStateItemsNames.indexOf(mainPanelContent) + 1);

						// to check where to insert it, we'll loop over the items of stateObject and find the first item that is also in itemsAfterMainContentItem
						var index = _.findIndex(stateObject[this._sectionMapping[mainPanelContent]].items, function(item) {
							return _.contains(itemsAfterMainContentItem, item.name);
						});
						if (index > -1) {
							stateObject[this._sectionMapping[mainPanelContent]].items.splice(index, 0, mainContentPanelLayout);
						} else {
							stateObject[this._sectionMapping[mainPanelContent]].items.push(mainContentPanelLayout);
						}
					}
				}
			}.bind(this)
		);
	}

	this._layoutState = stateObject;
	return stateObject;
};

/**
 * Invalidates the layout. Makes sure all panels have the correct size, resize is triggered on the correct elements,...
 */
DesktopLayout.prototype._invalidateLayout = function() {
	this._layout.height = $(this._context).innerHeight() || null;
	this._layout.width = $(this._context).innerWidth() || null;

	if (this._layout.width === 0 || this._layout.height === 0 || !this._cssLoaded) {
		// it makes no sense to invalidate the layout if the viewer is size(0,0) or no css is loaded
		return;
	}

	this._wrapper.css({
		height: this._layout.height,
		width: this._layout.width
	});

	var contentHeight = this._layout.height;

	// when the browser is zoomed in or out, getting width of elemens becomes very unreliable
	// widths can become floats, we can't rely on default jquery methods to get the correct width
	// if we make an element 1px to wide, the east panel can disappear
	// we can rely on the widths returned by getBoundClientRect() and the border width of the getComputedStyle
	// they return the float "width" values of the elements
	// make sure we ceil and floor (width of side panels should be ceiled, total width floored)
	// worst case scenario: 3 pixels are not used in the horizontal axis... (but users should zoom/in out anyway)
	var realWestWidth = this._layout.west.el.get(0).getBoundingClientRect().width;
	realWestWidth += parseFloat(getComputedStyle(this._layout.west.el.get(0)).borderRightWidth);
	realWestWidth = Math.ceil(realWestWidth);

	var realEastWidth = this._layout.east.el.get(0).getBoundingClientRect().width;
	realEastWidth += parseFloat(getComputedStyle(this._layout.east.el.get(0)).borderLeftWidth);
	realEastWidth = Math.ceil(realEastWidth);

	var realWidth = Math.floor(
		$(this._context)
			.get(0)
			.getBoundingClientRect().width
	);

	var centerWidth = realWidth - realWestWidth - realEastWidth;

	// if there is not enough space, auto collapse the side panels
	if (centerWidth < AUTOCOLLAPSE_THRESHOLD) {
		if (!this._layout.west.isAutoCollapsed() || !this._layout.east.isAutoCollapsed()) {
			this._layout.west.setAutoCollapsed(true);
			this._layout.east.setAutoCollapsed(true);

			// we did an autocollapse, invalidate the layout, and notify the user
			this._invalidateLayout();

			this.f.loadBundle(
				"viewer",
				function() {
					this.f.emit("userMessage", "info", this.f.t("infoSidepanelsAutoCollapsed"), undefined, false, "AUTOCOLLAPSE", true);
				},
				this
			);
			return;
		}
	} else if (this._layout.west.isAutoCollapsed() || this._layout.east.isAutoCollapsed()) {
		// if can have enough space with both panels open, open them again
		var possibleCenterWidth =
			this._layout.width - this._layout.west.getWidthWhenNotAutoCollapsed() - this._layout.east.getWidthWhenNotAutoCollapsed();
		if (possibleCenterWidth > AUTOCOLLAPSE_THRESHOLD) {
			this._layout.west.setAutoCollapsed(false);
			this._layout.east.setAutoCollapsed(false);

			// no need to show the autocollapse messages anymore
			this.f.emit("userMessage.markObsolete", "AUTOCOLLAPSE");
			this._invalidateLayout();
			return;
		}
	}

	this._invalidateNorthPane();

	contentHeight -= this._layout.top.outerHeight() || null;
	contentHeight -= this._layout.north.outerHeight() || null;

	if (this._aboveCenter.is(":visible")) {
		contentHeight -= this._aboveCenter.outerHeight() || null;
	}

	this._layout.west.setHeight(contentHeight);
	this._layout.east.setHeight(contentHeight);

	if (FullscreenUtil.fullscreenElement === this._layout.center.el[0]) {
		this._layout.center.setDim("", ""); // remove height, width properties when fullscreen
	} else {
		this._layout.center.setDim(centerWidth, contentHeight - 1); // minus 1: take into account the top border of the element
	}

	var centerPosition = this._layout.center.position();

	// Center message bar inside center
	this._messagebar.css({
		width: (this._layout.center.outerWidth() || null) - 36,
		top: $(this._context).position().top + (this._layout.top.outerHeight() || null) + (this._layout.north.outerHeight() || null) + 18,
		left: centerPosition.left + 18
	});

	// Layout loading indicator

	this._loadingIndicator.element.css({
		top: centerPosition.top + 5,
		left: centerPosition.left + this._layout.center.width - 29
	});

	// position toggle buttons
	this._toggleButtonWest.css({
		left: 3,
		top: centerPosition.top - 16 // 16: height of the icon
	});
	this._toggleButtonEast.css({
		left: this._layout.width - 21, // 22 width of the icon + padding
		top: centerPosition.top - 16
	});

	this._layout.east.invalidate();
	this._layout.west.invalidate();

	if (this._mainContent !== null) {
		this._mainContent.resized();
	}
	var length = _.filter(this._activeMainPanelContent, function(content) {
		return content.isVisible();
	}).length;
	_.each(
		this._activeMainPanelContent,
		function(activeMainPanelContent) {
			var verticalPaddingBorderAndMargin =
				(activeMainPanelContent.element.outerHeight(true) || null) - (activeMainPanelContent.element.height() || null);
			activeMainPanelContent.content.css({
				height: (this._center.innerHeight() || null) - (this._layout.below.outerHeight() || null) - verticalPaddingBorderAndMargin,
				width: (this._mainContentPanel.innerWidth() || null) / length
			});

			if (activeMainPanelContent.widget !== null) {
				activeMainPanelContent.widget.resized();
			}
		},
		this
	);

	_.each(
		this._items,
		function(layoutItem, itemName) {
			if (this._layoutModel.getLayoutItemType(itemName) === LayoutConstants.LayoutItemType.WIDGET && layoutItem.widget !== undefined) {
				layoutItem.widget.resized();
			} else if (this._layoutModel.getLayoutItemType(itemName) === LayoutConstants.LayoutItemType.TOOLBARCOMPONENT) {
				layoutItem.moved();
			}
		},
		this
	);
	this._emitLayoutChanged();
	this._updateItemVisibility();
};

/**
 * Position the toolbarcomponents in the northpane
 */
DesktopLayout.prototype._invalidateNorthPane = function() {
	// align the toolbar items to the left of the viewport
	// if not enough space, move the more to the left
	var northContentWidth = 0; // calculate content width (sum of all children)
	this._layout.north.el.children().each(function() {
		if ($(this).is(":visible")) {
			northContentWidth += ($(this).outerWidth(true) || null) + 2; // add to for borders that are not taken into account in some way...
		}
	});
	if (northContentWidth === 0) {
		// no children yet, so take the whole width
		northContentWidth = this._layout.north.outerWidth() || null;
	}

	var paddingLeft = this._layout.west.outerWidth() || null;
	var availableWidth = this._layout.width - 2 * 30; //-2 * 25 : make space for the collapse icons
	if (paddingLeft + northContentWidth >= availableWidth) {
		paddingLeft -= paddingLeft + northContentWidth - availableWidth;
	}
	paddingLeft = Math.max(paddingLeft, 30);

	this._layout.north.el.css("padding-left", paddingLeft);
};

/**
 * Returns the type of the layout.
 * @override
 * @return {Number}
 */
DesktopLayout.prototype.getLayoutType = function() {
	return LayoutConstants.LayoutType.DESKTOP;
};

/**
 * Update visibility of the added items based on whether or not they are allowed in the current mode.
 * E.g. item was added in previous mode, but should not be shown in current mode
 */
DesktopLayout.prototype._updateItemVisibility = function() {
	_.each(
		this._items,
		_.bind(function(item) {
			var layoutItemType = this._layoutModel.getLayoutItemType(item.name);
			if (layoutItemType === LayoutConstants.LayoutItemType.TOOLBARCOMPONENT) {
				//var toolbar = layoutItem.getItem(this.getLayoutType());
				item.setVisible(this._activeMode.isItemAllowed(item.name));
			} else if (layoutItemType === LayoutConstants.LayoutItemType.WIDGET) {
				item.setVisible(this._activeMode.isItemAllowed(item.name) && item.widget.isEnabled());
			}
		}, this)
	);
};
/**
 * Update the contents of the main panel
 * A mode can decide to show something else than the viewport widget
 */
DesktopLayout.prototype._updateMainContent = function() {
	var mainPanelContents = this._activeMode.getMainPanelContent();
	this._activeMainPanelContent = [];

	var copy = _.clone(this._mainContentWidgetWrappers);
	// run through all possible main content widget wrappers
	_.each(
		copy,
		function(wrapper) {
			var index = mainPanelContents.indexOf(wrapper.widget.getID());
			if (index !== -1) {
				// this wrapper should be visible
				wrapper.setVisible(true);
				wrapper.setCollapsed(false);
				wrapper.setHeaderEnabled(false);
				wrapper.element.css({
					height: "",
					width: "",
					top: "",
					left: "",
					position: "relative",
					order: index
				});
				this._activeMainPanelContent.push(wrapper);
			} else {
				// hide, the wrapper should not be active
				wrapper.setHeaderEnabled(true);
				wrapper.setVisible(false);
				var section = this._sections[this._sectionMapping[wrapper.name]];
				if (section) {
					this._mainContentWidgetWrappers = _.without(this._mainContentWidgetWrappers, wrapper);
					wrapper.content.css({
						height: "",
						width: ""
					});
					this._addLayoutItem(wrapper.name);
					this._toggleWidget(wrapper.name, wrapper, wrapper.widget.isEnabled());
				}
			}
		},
		this
	);

	var activeMainPanelContentNames = this._activeMainPanelContent.map(function(wrapper) {
		return wrapper.widget.getID();
	});

	_.chain(mainPanelContents)
		.filter(
			function(mainPanelContent) {
				// The item was added but isn't a standard main content widget wrapper
				return this._items[mainPanelContent] && activeMainPanelContentNames.indexOf(mainPanelContent) === -1;
			}.bind(this)
		)
		.each(
			function(mainPanelContent) {
				// Add it to the main content widget wrapper
				this._mainContentWidgetWrappers.push(this._items[mainPanelContent]);
				this._mainContentPanel.prepend(this._items[mainPanelContent].element);
			}.bind(this)
		)
		.tap(
			function(a) {
				// if there were widgets that weren't added yet,
				if (a.length > 0) {
					this._updateMainContent();
				} else {
					this._invalidateLayout();
				}
			}.bind(this)
		);
};

/**
 * Update visibility of a Widget
 * @param {String}        widgetName    - name of the widget
 * @param {WidgetWrapper} widgetWrapper - widgetWrapper in which the widget is displayed
 * @param {Boolean}       visible       - should the widget be visible?
 */
DesktopLayout.prototype._toggleWidget = function(widgetName, widgetWrapper, isEnabled) {
	if (isEnabled) {
		if (this._activeMode.isItemAllowed(widgetName)) {
			widgetWrapper.setVisible(true);
		}
	} else {
		widgetWrapper.setVisible(false);
	}
};

/**
 * Enforce restricted mode (collapse side panels)
 */
DesktopLayout.prototype._enforceRestrictedMode = function() {
	if (this.f.getProperty("viewerStyle") == LayoutConstants.ViewerStyle.RESTRICTED) {
		this._layout.east.toggle(false);
		this._layout.west.toggle(false);
	}
};

/**
 * Make sure a widget is visible
 * Uncollapse it if it is collapsed
 *
 * If the widget is modal, show it
 * If the widget is not modal, make the widget modal if it is in a hidden/collapsed sidebar
 * @param {String} layoutItemName - the name of the Layout Item
 * @param {WidgetWrapper} widgetWrapper - the widgetWrapper for the widget
 */
DesktopLayout.prototype._makeWidgetVisible = function(layoutItemName, widgetWrapper) {
	if (!widgetWrapper.isModal) {
		// the widget is not modal
		if (!this._layout[this._sectionMapping[layoutItemName]].isVisible() && this._mainContentWidgetWrappers.indexOf(widgetWrapper) === -1) {
			// the widget is in a collapsed sidepanel, make it modal
			this.moveItem(widgetWrapper.getID(), "center");

			var width = 450;
			var height = 350;
			width = Math.min(this._layout.center.width - 20, width);
			height = Math.min(this._layout.center.height - 20, height);

			var position = this._layout.center.position();
			var top = position.top + 10;
			var left = position.left + this._layout.center.width - 10 - width;
			widgetWrapper.setCollapsed(false);
			widgetWrapper.setModal(true, top, left, width, height);
		}
	}
	widgetWrapper.widget.setMakeVisibleRequestAccepted();
	widgetWrapper.setCollapsed(false);
};

/**
 * Defines the array of css files that should be loaded
 *
 * @returns Array<String> list of css files
 */
DesktopLayout.prototype._loadCss = function() {
	require("./less/desktop.less");
};

/**
 * Makes the center content pane fullscreen
 *
 * @returns True if making center pane fullscreen was succesful
 */
DesktopLayout.prototype.fullscreenViewport = function() {
	if (FullscreenUtil.fullscreenEnabled) {
		FullscreenUtil.requestFullscreen(this._center[0]);
		return true;
	}
	return false;
};

DesktopLayout.prototype.getMainContentElement = function() {
	return this._center;
};

/**
 * @override
 */
DesktopLayout.prototype.getDefaultTool = function() {
	return (this._activeMode && this._activeMode.getDefaultTool()) || CoreConstants.TOOLS.ZOOM;
};

export default DesktopLayout;
