import PrivateChannel from "@esko/web-viewer/core/PrivateChannel";
/**
 * Panel interface element (for sidepanels/topbar/centerpanel)
 *
 * @constructor
 * @memberof Layout
 * @param {ViewerFacade} f
 * @param {String} name - name of the panel
 * @param {JQueryObj} el - main object in which the panel is drawn
 */
var Panel = function(f, name, el) {
	this.f = f;
	PrivateChannel.installTo(this, f, ["panel", name]);

	this.name = name;
	this.el = el;

	this.width = 0;
	this.height = 0;

	this._visible = true;
	this._disabled = false;
};

/**
 * Set visibility of panel to visible
 */
Panel.prototype.show = function() {
	var oldVisibility = this.isVisible();

	this._visible = true;
	this._updateVisibility();

	if (oldVisibility != this.isVisible()) {
		this._emit("changed.visibility", oldVisibility, this.isVisible());
	}
};

/**
 * Set visiblity of panel to hidden
 */
Panel.prototype.hide = function() {
	var oldVisibility = this.isVisible();
	this._visible = false;
	this._updateVisibility();

	if (oldVisibility != this.isVisible()) {
		this._emit("changed.visibility", oldVisibility, this.isVisible());
	}
};

/**
 * Toggle the panel
 * @param {Boolean} [toggle]
 */
Panel.prototype.toggle = function(toggle) {
	if (toggle === undefined) {
		toggle = !this.getVisibility();
	}

	if (toggle) {
		this.show();
	} else {
		this.hide();
	}
};

/**
 * Check if the panel is shown or not (note, this doesn't take into account autocollapse or disabled)
 * @returns {Boolean}
 */
Panel.prototype.getVisibility = function() {
	return this._visible;
};

/**
 * Enable/disable the panel
 * @param {Boolean} disabled
 */
Panel.prototype.setDisabled = function(disabled) {
	var oldVisibility = this.isVisible();

	this._disabled = disabled;
	this._updateVisibility();

	if (oldVisibility != this.isVisible()) {
		this._emit("changed.visibility", oldVisibility, this.isVisible());
	}
};

/**
 * Check if the panel is disabled
 * @returns {Boolean}
 */
Panel.prototype.isDisabled = function() {
	return this._disabled;
};

/**
 * Set the width of the panel
 * @param {Number} width - width of the panel
 */
Panel.prototype.setWidth = function(width) {
	if (this.width !== width) {
		this.width = width;
		this.el.css("width", width);
	}
};

/**
 * Set the height of the panel
 * @param {Number} height - height of the panel
 */
Panel.prototype.setHeight = function(height) {
	if (this.height !== height) {
		this.height = height;
		this.el.css("height", height);
	}
};

/**
 * Get the height of the panel
 * @returns {Number}
 */
Panel.prototype.getHeight = function() {
	return this.isVisible() ? this.height : 0;
};

/**
 * Get the width of the panel
 * @returns {Number}
 */
Panel.prototype.getWidth = function() {
	return this.isVisible() ? this.width : 0;
};

/**
 * et the width of the panel (don't take into account visibility)
 * @returns {Number}
 */
Panel.prototype.getWidthWhenVisible = function() {
	return this.width;
};

/**
 * Get the height of the panel (don't take into account visibility)
 * @returns {Number}
 */
Panel.prototype.getHeightWhenVisible = function() {
	return this.height;
};

/**
 * Get the outerwidth of the panel
 * @returns {Number}
 */
Panel.prototype.outerWidth = function() {
	if (this.isVisible()) {
		return this.el.outerWidth(true) || null;
	} else {
		return 0;
	}
};

/**
 * Get the innerWidth of the panel
 * @returns {Number}
 */
Panel.prototype.innerWidth = function() {
	if (this.isVisible()) {
		return this.el.innerWidth() || null;
	} else {
		return 0;
	}
};

/**
 * Get the outerHeight of the pane;
 * @returns {Number}
 */
Panel.prototype.outerHeight = function() {
	if (this.isVisible()) {
		return this.el.outerHeight(true) || null;
	} else {
		return 0;
	}
};

/**
 * Get the position of the panel
 * @returns {Object} - position
 */
Panel.prototype.position = function() {
	return this.el.position();
};

/**
 * Set both width and height of the panel
 * @private
 * @param {Number} [width]
 * @param {Number} [height]
 */
Panel.prototype.setDim = function(width, height) {
	if (width !== undefined) {
		this.setWidth(width);
	}

	if (height !== undefined) {
		this.setHeight(height);
	}
};

/**
 * Check if a panel is visible (takes in to account all parameters (visible, autocollapsed, disabled)
 * @returns {Boolean}
 */
Panel.prototype.isVisible = function() {
	return this._visible && !this._disabled;
};

/**
 * Update the visibility of the panel
 */
Panel.prototype._updateVisibility = function() {
	if (this.isVisible()) {
		this.el.removeClass("displayNone");
	} else {
		this.el.addClass("displayNone");
	}
};

/**
 * Get the state of the panel (note only the visible property is saved in the state)
 * @returns {Object}
 */
Panel.prototype.getState = function() {
	return {
		visible: this._visible
	};
};

/**
 * Restore the state of the panel
 * @param {Object} state - as returned by getState
 */
Panel.prototype.restoreState = function(state) {
	this.toggle(state.visible);
};

export default Panel;
