import Panel from "@esko/web-viewer/layout/desktop/Panel";
import MiniButton from "@esko/web-viewer/layout/component/MiniButton";
import LayoutUtils from "@esko/web-viewer/layout/LayoutUtils";
import _ from "underscore";
/**
 * Panel interface element (for sidepanels/topbar/centerpanel) used by DesktopLayout
 *
 * @class SidePanel anel interface element (for sidepanels/topbar/centerpanel) used by DesktopLayout
 * @constructor
 * @augments Panel
 * @memberof Layout
 * @param {ViewerFacade} f
 * @param {String} name - name of the panel
 * @param {JQueryObj} el - main object in which the panel is drawn
 * @param {JQueryObj} [inner] - inner element of the panel
 */
var SidePanel = function(f, name, el, inner) {
	Panel.call(this, f, name, el);

	this.toggleButton = new MiniButton(
		this.f,
		"toggleHide" + name,
		name == "west" ? "icon-WCR_Panel-Left" : "icon-WCR_Panel-Right",
		false,
		_.bind(function() {
			this.toggle();
		}, this)
	);

	this.toggleButton.setTooltip("viewer", name == "west" ? "LayoutToggleLeftSidebar" : "LayoutToggleRightSidebar");
	this.toggleButton.element.addClass("panelToggleButton");
	this.inner = inner;
	this._autoCollapsed = false;
};

SidePanel.prototype = Object.create(Panel.prototype);
SidePanel.prototype.constructor = SidePanel;

/**
 * Autocollapse the panel (or undo it)
 * @param {Boolean} autoCollapsed
 */
SidePanel.prototype.setAutoCollapsed = function(autoCollapsed) {
	this._autoCollapsed = autoCollapsed;
	this._updateVisibility();
};

/**
 * Check if the panel is auto collapsed
 * @returns {Boolean}
 */
SidePanel.prototype.isAutoCollapsed = function() {
	return this._autoCollapsed;
};

/**
 * @override
 */
SidePanel.prototype.isVisible = function() {
	return Panel.prototype.isVisible.call(this) && !this._autoCollapsed;
};

/**
 * Show/hide the panel
 * @override
 */
SidePanel.prototype._updateVisibility = function() {
	Panel.prototype._updateVisibility.call(this);

	// Set the widgets on state invisible
	var widgets = this.inner.find(".widgetWrapper");
	_.each(
		widgets,
		function(widgetWrapperElement) {
			var wrapper = widgetWrapperElement.wrapper;
			// NOTE: this doesn't check whether a widget is allowed or not in the layout
			// SidePanel is only used in the DesktopLayout, so this is not a problem (for now)
			if (this.isVisible()) {
				if (wrapper.widget.isEnabled()) {
					wrapper.widget.setVisible(true);
				}
			} else {
				wrapper.widget.setVisible(false);
			}
		},
		this
	);

	// if the panel is disabled don't show the toggle buttons
	if (!this.isDisabled()) {
		this.toggleButton.element.show();
	} else {
		this.toggleButton.element.hide();
	}
};

/**
 * recalculate the position of all widgetwrappers inside the panel
 */
SidePanel.prototype.invalidate = function() {
	if (this.isVisible()) {
		LayoutUtils.layoutWidgets(this.inner, 2);
	}
};

/**
 * Also store the width
 * @override
 */
SidePanel.prototype.getState = function() {
	var state = Panel.prototype.getState.call(this);
	state.width = this.width;
	return state;
};

/**
 * Also restore the width of the panel
 * @param {Object} state
 */
SidePanel.prototype.restoreState = function(state) {
	Panel.prototype.restoreState.call(this, state);

	this.setWidth(state.width);
};

/**
 * Get the width of the panel when it would not be autocolapsed
 * @returns {Number}
 */
SidePanel.prototype.getWidthWhenNotAutoCollapsed = function() {
	if (!this.isDisabled() && this.getVisibility()) {
		return this.getWidthWhenVisible();
	} else {
		return 0;
	}
};

/**
 * @override
 * @param {Number} height
 */
SidePanel.prototype.setHeight = function(height) {
	Panel.prototype.setHeight.call(this, height);

	this.inner.css("height", height);
};

export default SidePanel;
