import LayoutConstants from "@esko/web-viewer/layout/LayoutConstants";
import LayoutMode from "@esko/web-viewer/layout/LayoutMode";
import _ from "underscore";
/**
 * @exports WizardMode
 * @memberof Layout
 * @class
 * Wizard mode of the desktoplayout. Will allow all the layout items to be added and the state is persisted.
 * @constructor
 * @param {Core.ViewerFacade} f
 * @param {Layout} layout
 */
var WizardMode = function(f, layout) {
	LayoutMode.call(this, f, layout, LayoutConstants.LayoutMode.WIZARD);

	/**
	 * Saves the names of all modal widgets that are hidden
	 * @member {Array.<String>}
	 */
	this._hiddenModalWidgets = [];

	/**
	 * Saves the visible state of toolbarComponent, maps ItemName to Boolean. Initializes to all the toolbarComponents.
	 * @member {Object.<String, Boolean>}
	 */
	this._wasToolbarComponentVisible = _.invert(_.values(LayoutConstants.ToolbarComponent));

	/**
	 * Save all the open popup when the wizard is visible to be able to hide them
	 */
	this._openPopups = [];
};

WizardMode.prototype = Object.create(LayoutMode.prototype);
WizardMode.prototype.constructor = WizardMode;

/**
 * Turns On/Off the layoutMode
 * @param {Boolean} toggle
 */
WizardMode.prototype.toggle = function(toggle) {
	toggle = toggle === undefined ? !this._isActive : toggle;
	this._isActive = toggle;
	if (toggle) {
		this._hideSidePanels();

		this._layout.showItem(LayoutConstants.Widget.ALIGNWIZARD); //show alignment widget

		this._hideModalWidgets();

		this._hidePopups();
	} else {
		this._showSidePanels();

		this._layout.hideItem(LayoutConstants.Widget.ALIGNWIZARD); //hide alignment widget

		this._showModalWidgets();

		this._showPopups();
	}
};

/**
 * Hide the side panels
 */
WizardMode.prototype._hideSidePanels = function() {
	this._layout.disablePanelInteraction(); //disable panel interaction, not possible to open side panels
	this._layout.disablePanel("east");
	this._layout.disablePanel("west");
};

/**
 * Show the side panels
 */
WizardMode.prototype._showSidePanels = function() {
	this._layout.initializePanelInteraction(); //make sure to enable panel interaction again
	this._layout.enablePanel("east");
	this._layout.enablePanel("west");
};

/**
 * Hide modal widgets
 */
WizardMode.prototype._hideModalWidgets = function() {
	//hides all modal widgets and returns the names of the hidden elements
	this._hiddenModalWidgets = this._layout.hideAllModalWidgets();
};
/**
 * Show the modal widget  that were previously hidden
 */
WizardMode.prototype._showModalWidgets = function() {
	_.each(
		this._hiddenModalWidgets,
		function(widgetName) {
			//run over the hidden modal widgets
			this._layout.showItem(widgetName); //show them again
		},
		this
	);
};

/**
 * Hide all popups in the viewport
 */
WizardMode.prototype._hidePopups = function() {
	//gets all popups to hide and returns those popups to show them when wizardMode ends
	this._openPopups = this._layout.getAllVisiblePopups();
	this._openPopups.hide();
};

/**
 * Show all the popups that were previously hidden
 */
WizardMode.prototype._showPopups = function() {
	this._openPopups.show();
	this._openPopups = [];
};

/**
 * Hide all toolbar components except for navigation controls
 */
WizardMode.prototype._hideToolbarComponents = function() {
	//run over all the toolbarComponents and check if they are visible
	_.each(
		this._wasToolbarComponentVisible,
		function(wasVisible, itemName) {
			//save the visible state of each toolbarcomponent
			this._wasToolbarComponentVisible[itemName] = this._layout.isItemVisible(itemName);
			if (
				this._wasToolbarComponentVisible[itemName] && //only hide items that are visible
				itemName !== LayoutConstants.ToolbarComponent.NAVIGATE
			) {
				//and are not the navigate controls
				this._layout.hideItem(itemName);
			}
		},
		this
	);
};

/**
 * Show all toolbar ocmponetns that were previously hidden
 */
WizardMode.prototype._showToolbarComponents = function() {
	_.each(
		this._wasToolbarComponentVisible,
		function(wasVisible, itemName) {
			//run over all toolbarcomponents
			if (wasVisible) {
				//only if it was visible, show it again
				this._layout.showItem(itemName);
			}
		},
		this
	);
};

/**
 * Checks if the item is allowed for this layout mode.
 *
 * @param {String} itemname
 * @returns {Boolean}
 */
WizardMode.prototype.isItemAllowed = function(itemname) {
	return (
		itemname === LayoutConstants.Widget.ALIGNWIZARD ||
		itemname === LayoutConstants.Widget.VIEWPORT ||
		itemname === LayoutConstants.ToolbarComponent.NAVIGATE
	);
};

/**
 * Checks if the state of the layout should be persisted.
 * @return {Boolean}
 */
WizardMode.prototype.isPersisted = function() {
	return false;
};

export default WizardMode;
