import Layout from "@esko/web-viewer/layout/Layout";
import LayoutConstants from "@esko/web-viewer/layout/LayoutConstants";
import PrivateChannel from "@esko/web-viewer/core/PrivateChannel";
import ListenerUtil from "@esko/web-viewer/utils/ListenerUtil";
import FullscreenUtil from "@esko/web-viewer/utils/FullscreenUtil";
import FlexBasicMode from "@esko/web-viewer/layout/flex/modes/FlexBasicMode";
import FlexWizardMode from "@esko/web-viewer/layout/flex/modes/FlexWizardMode";
import Collada3DMode from "@esko/web-viewer/layout/desktop/Collada3DMode";
import Desktop3DMode from "@esko/web-viewer/layout/desktop/Desktop3DMode";
import Desktop2D3DMode from "@esko/web-viewer/layout/desktop/Desktop2D3DMode";
import DesktopVideoMode from "@esko/web-viewer/layout/desktop/DesktopVideoMode";
import FlexWidgetWrapper from "@esko/web-viewer/layout/flex/FlexWidgetWrapper";
import ToolbarComponentWrapper from "@esko/web-viewer/layout/ToolbarComponentWrapper";
import FlexSidebar from "@esko/web-viewer/layout/flex/FlexSidebar";
import FlexSection from "@esko/web-viewer/layout/flex/FlexSection";
import FlexDrag from "@esko/web-viewer/layout/flex/FlexDrag";
import CoreConstants from "@esko/web-viewer/core/CoreConstants";
import Compare3DMode from "@esko/web-viewer/view3D/Compare3DMode";
import ShapeUtils from "@esko/shapeutils";
import _ from "underscore";
import $ from "jquery";
/**
 * @class
 * Icon Side Panel layout This is the view initialized by the LayoutManager that will add all the layoutcomponents
 * to the viewer. It can generate a an object representing the current state of the viewer and can be initialized by
 * a state-String.
 *
 * NOTE widget that are both MainPanelContent and ModelPanelContent are not supported at the moment!
 *
 * @exports FlexLayout
 * @constructor
 * @param {Core.ViewerFacade} f
 * @param {LayoutModel} layoutModel
 * @param {JQueryObj} context - context where the layout will add its components to
 * @param {Object} layoutState
 * @param {LoadingIndicator} loadingIndicator - the loading indicator, which was created before all requests started
 * @param {JQueryObj} initialLoader - the JQuery reference to the element indicating the initial loader (needed to safely remove it without emptying the context)
 */
var FlexLayout = function(f, layoutModel, context, layoutState, loadingIndicator, initialLoader, layoutConfiguration) {
	Layout.call(this, f, layoutModel, context, layoutState, loadingIndicator, initialLoader);
	PrivateChannel.installTo(this, f, "layout.flex");
	ListenerUtil.mixin(this, f);
	/**
	 * Default state of the layout, used to reset the layout
	 */

	this._configuration = layoutConfiguration;

	this._defaultLayout = this._getDefaultLayout();

	/**
	 * Mapping for all the items, maps items to their section
	 */
	this._sectionMapping = {};

	/**
	 * All the layout items mapped as Name -> FlexWidgetWrapper
	 */
	this._items = {};

	//a debounced invalidate, to make sure not to many invalidates are triggered
	this._debouncedInvalidate = _.debounce(_.bind(this._invalidateLayout, this), 200);

	/**
	 * In this array all items that are for the main content are kept
	 */
	this._mainContentWidgetWrappers = [];

	this._topbar = $("<div>", {
		class: "topbar"
	});
	this._toolbar = $("<div>", {
		class: "toolbar"
	});
	this._mainContentWrapper = $("<div>", {
		class: "maincontentwrapper"
	});

	var isEastUndefined = typeof this._configuration.east === "undefined" || typeof this._configuration.east.iconified === "undefined";
	var isWestUndefined = typeof this._configuration.west === "undefined" || typeof this._configuration.west.iconified === "undefined";

	this._centerContent = $("<div>", {
		class: "centercontent"
	});

	this._messagebar = $("<div />").addClass("messagebar");
	this._aboveCenter = $("<div />").addClass("aboveCenter");
	this._mainContent = $("<div />").addClass("mainContent");
	this._belowCenter = $("<div />").addClass("belowCenter toolbar toolbar-bottom");

	this._sections = {
		north: new FlexSection(this._toolbar),
		east: new FlexSidebar(this.f, "right", "right", this._mainContentWrapper, isEastUndefined ? true : this._configuration.east.iconified),
		west: new FlexSidebar(this.f, "left", "left", this._mainContentWrapper, isWestUndefined ? true : this._configuration.west.iconified),
		below: new FlexSection(this._belowCenter)
	};

	this._centerContent.append(this._aboveCenter, this._loadingIndicator.element, this._messagebar, this._mainContent, this._belowCenter);

	this._mainContentWrapper.append(
		this._sections.west.getTilebarContext(),
		this._sections.west.getWidgetbarContext(),
		this._centerContent,
		this._rightWidgetbar,
		this._sections.east.getWidgetbarContext(),
		this._sections.east.getTilebarContext()
	);

	// initialize the section mapping of the layoutItems
	this._initializeSectionMapping(this._layoutState);
	// initialize the sections
	this._initializeSections(this._layoutState);

	// by default load the view mode

	var modes = this._getLayoutModes();
	this._activeMode = modes[0];
	this._activeMode.toggle(true);

	modes.forEach(
		function(mode) {
			this._layoutModel.registerLayoutMode(mode);
			mode.on(
				"changed",
				function() {
					this._updateMainContent();
				}.bind(this)
			);
		}.bind(this)
	);

	this.f.emit("layout.finishRegisteringModes");

	this._loadCss();
	this._cssLoaded = true;
	this._initializeDOM();

	//listen to resize events of the window
	$(window).on(
		"resize",
		_.bind(function(e) {
			// http://bugs.jqueryui.com/ticket/7514: resize of popups bubbles to window
			if (e.target === window) {
				this._debouncedInvalidate();
			}
		}, this)
	);

	//listen to widget events, to act on resized/collapsed/...
	this._registerOnLayoutEvents();

	this._invalidateLayout(); //this one is needed to provide the correct world coordinates

	this._flexDrag = new FlexDrag(this._mainContentWrapper);
	this._flexDrag.addFlexSidebarDropzone(this._sections.west, "west", this);
	this._flexDrag.addFlexSidebarDropzone(this._sections.east, "east", this);
};

FlexLayout.prototype = Object.create(Layout.prototype);
FlexLayout.prototype.constructor = FlexLayout;

FlexLayout.prototype._initializeDOM = function() {
	this._initialLoader.remove();
	delete this._initialLoader;
	this._context.append(this._topbar, this._toolbar, this._mainContentWrapper);
	this._invalidateLayout();
};

FlexLayout.prototype._getLayoutModes = function() {
	return [
		new FlexBasicMode(this.f, this),
		new FlexWizardMode(this.f, this),
		new Collada3DMode(this.f, this),
		new Desktop2D3DMode(this.f, this),
		new Desktop3DMode(this.f, this),
		new DesktopVideoMode(this.f, this),
		new Compare3DMode(this.f, this)
	];
};

/**
 * Add an item to the layout, the items needs to already added to the items-collection.
 * It adds the item to the corect element on the correct location (depends on the previous/default state and already added items)
 * @param {String} itemName
 */
FlexLayout.prototype._addLayoutItem = function(itemName) {
	var section = this._sectionMapping[itemName]; //get the section the items belongs to
	if (section !== undefined) {
		//make sure there is a section to add the item to
		var sectionItems = null;
		var itemState = null;
		if (this._layoutState && this._layoutState[section]) {
			sectionItems = _.map(this._layoutState[section].items, function(item) {
				return item.name;
			}); //just get all the names of the items in this section, not other information
			itemState = _.find(this._layoutState[section].items, function(item) {
				return item.name === itemName;
			}); //get the state of the to be added item
		} else {
			sectionItems = this._defaultLayout[section]; //default state only holds names of the items
		}
		if (sectionItems !== null && sectionItems !== undefined) {
			//the section should at least hold the item itself
			var index = _.indexOf(sectionItems, itemName);
			var nextItem = null;
			var itemsToLookAt = null;
			if (index === -1) {
				index = _.indexOf(this._defaultLayout[section], itemName);
				itemsToLookAt = this._defaultLayout[section];
			} else {
				itemsToLookAt = sectionItems;
			}

			nextItem = this._items[itemsToLookAt[++index]]; //get the next item, so we know infront of which item the item needs to be added to
			if (_.contains(this._mainContentWidgetWrappers, nextItem)) {
				nextItem = undefined;
			}
			while (index < itemsToLookAt.length && nextItem === undefined) {
				//as long nextItem is not yet added to the layout, search for a next one
				nextItem = this._items[itemsToLookAt[++index]];
				if (_.contains(this._mainContentWidgetWrappers, nextItem)) {
					nextItem = undefined;
				}
			}

			if (nextItem === undefined) {
				//no nextItem is found, the item is (for now) the last item in the section
				this._appendItemToSection(itemName, section); //append the item to the section
			} else {
				//add the item in front of the found nextItem
				this._prependItemBeforeItem(nextItem.name, itemName, section);
			}
			//now be sure to set the correct state of the item (active and collapsed or not)
			if (itemState !== null && itemState !== undefined) {
				this._setActive(itemName, itemState.isActive, itemState.isCollapsed);
				var widgetWrapper = this._items[itemName];
				if (widgetWrapper !== undefined && widgetWrapper.widget) {
					if (itemState.stickyDimensions) {
						widgetWrapper.setStickyDimensions(itemState.stickyDimensions);
					} else {
						widgetWrapper.setStickyDimensions(null);
					}
				}
			}
		}
	} else {
		this.f.debug.warn("FlexLayout :: No section found for item " + itemName);
	}
};

/**
 * Set a widget active
 * @param {String} itemName
 * @param {Boolean} collapsed
 */
FlexLayout.prototype._setActive = function(itemName, active, collapsed) {
	var widgetWrapper = this._items[itemName];
	if (widgetWrapper !== undefined && widgetWrapper.widget) {
		var sectionName = this._sectionMapping[itemName];
		if (sectionName) {
			var section = this._sections[sectionName];
			if (section) {
				section.toggleWidgetWrapperActive(widgetWrapper, active);
				widgetWrapper.setCollapsed(collapsed);
			}
		}
	}
};

/**
 * Appends an already added item (to the items collection) to a section.
 * @param {String} itemName
 * @param {String} section
 */
FlexLayout.prototype._appendItemToSection = function(itemName, section) {
	var layoutItem = this._items[itemName];
	if (layoutItem.widget) {
		layoutItem.widget.movingStarted();
	}

	this._sections[section].appendItem(layoutItem);

	if (layoutItem.widget) {
		layoutItem.widget.movingStopped();
	}
};

/**
 * Prepends an already added item (to the items collection) before the nextItem in a given section.
 * @param {String} nextItemName
 * @param {String} itemName
 * @param {String} section
 */
FlexLayout.prototype._prependItemBeforeItem = function(nextItemName, itemName, section) {
	var layoutItem = this._items[itemName];
	if (layoutItem.widget) {
		layoutItem.widget.movingStarted();
	}

	//add item before other item in the layout
	this._sections[section].prependItemBeforeItem(layoutItem, this._items[nextItemName]);

	if (layoutItem.widget) {
		layoutItem.widget.movingStopped();
	}
};

/**
 * Moves an item to another section. Is used by draggable widgets to let the layout now it is moved.
 * @param {String} itemName
 * @param {String} sectionName
 * @param {String} nextItemName - item will be moved before this item
 * @param {Boolean} resetSizeAndLocation
 * @param {Boolean} resetCollapsed
 */
FlexLayout.prototype.moveItem = function(itemName, sectionName, nextItemName, resetSizeAndLocation, resetCollapsed) {
	if (itemName !== nextItemName) {
		/*	if (resetSizeAndLocation) { //resets the size and location of the item (moves from center to a side-panel)
                this._setLocation(itemName, null);
                this._setSize(itemName, null);
            }
            if (resetCollapsed) { //resets the collapsed state of the item
                this._setCollapsed(itemName, false);
            }*/
		var currentSection = this._sectionMapping[itemName];
		this._detachItem(itemName, currentSection); //removes the item from the previous section
		this._sectionMapping[itemName] = sectionName;
		if (nextItemName === undefined || this._sectionMapping[nextItemName] !== sectionName) {
			this._appendItemToSection(itemName, sectionName);
		} else {
			//check if no problems with after parameter
			this._prependItemBeforeItem(nextItemName, itemName, sectionName);
		}
	}
};

/**
 * Remove an item from a section, it removes the item from the DOM, but doesn't delete the DOM-element.
 * @param {String} itemName
 * @param {String} sectionName
 */
FlexLayout.prototype._detachItem = function(itemName, sectionName) {
	var widgetWrapper = this._items[itemName];
	if (widgetWrapper) {
		if (widgetWrapper.widget) {
			widgetWrapper.widget.movingStarted();
		}

		//remove element from the DOM (doesn't delete DOM-element)
		this._items[itemName].element.detach();

		if (widgetWrapper.widget) {
			widgetWrapper.widget.movingStopped();
		}
		//the invalidate is only done after all other changes (e.g. re-add DOM-element) are done
		_.defer(
			_.bind(function() {
				this._invalidateLayout();
			}, this)
		);
	}
};

/**
 * Initializes the section mapping. All items will be mapped to their section so it's easier to find the item's section.
 * Items that aren't in the layoutstate but are in the default state will also be added to the mapping.
 *
 * @param {Object} [layoutState] - if not defined only the default layout is used for the mapping
 */
FlexLayout.prototype._initializeSectionMapping = function(layoutState) {
	//first run over all the sections and items in the layoutState
	if (layoutState !== undefined && layoutState !== "" && _.size(layoutState) !== 0) {
		_.each(
			layoutState,
			function(section, sectionName) {
				//runs over all sections
				_.each(
					section.items,
					function(item) {
						//runs over all items in those sections
						this._sectionMapping[item.name] = sectionName; //map each itemName->sectionName
					},
					this
				);
			},
			this
		);
	}
	//run over the defaultLayout, if items of the default layout aren't added yet, add them to the mapping
	_.each(
		this._getDefaultLayout(),
		function(items, sectionName) {
			_.each(
				items,
				function(item) {
					if (this._sectionMapping[item] === undefined) {
						//check if item isn't added by layoutState yet
						this._sectionMapping[item] = sectionName;
					}
				},
				this
			);
		},
		this
	);
};

/**
 * Initializes the sections, get state from the layoutState.
 */
FlexLayout.prototype._initializeSections = function(layoutState) {
	if (layoutState !== undefined && layoutState !== "" && _.size(layoutState) !== 0) {
		_.each(
			layoutState,
			function(section, sectionName) {
				//run over all sections
				if (this._sections[sectionName] !== undefined && this._sections[sectionName].restoreState) {
					this._sections[sectionName].restoreState(section);
				}
			},
			this
		);
	}

	var onSidebarChange = _.bind(function() {
		this._emitLayoutChanged();
		this._debouncedInvalidate();
	}, this);

	this._sections.east.on("changed.*", onSidebarChange);
	this._sections.west.on("changed.*", onSidebarChange);
};

/**
 * Register listeners for general layout event.
 */
FlexLayout.prototype._registerOnLayoutEvents = function() {
	this.f.on(
		"layout.disableTextSelection",
		_.bind(function() {
			this._disableTextSelection();
		}, this)
	);

	this.f.on(
		"layout.enableTextSelection",
		_.bind(function() {
			this._enableTextSelection();
		}, this)
	);

	this.f.on("layout.resetToDefault", _.bind(this.resetToDefault, this));

	this.on(
		"changed.layoutMode",
		_.bind(function() {
			this._updateMainContent();
		}, this)
	);
};

/**
 * Add an item to the layout. This is called by Layout to add an item.
 * @param {Object} layoutItem
 * @example
 * 	{ 	type: {LayoutConstants.LayoutType},
 * 		name: 'itemName',
 * 		item: {AbstractWidget/ToolbarComponent}
 * 	}
 */
FlexLayout.prototype._addItem = function(layoutItem) {
	if (layoutItem.type === LayoutConstants.LayoutItemType.WIDGET) {
		var widget = layoutItem.getItem(this.getLayoutType());
		var widgetType = widget.getWidgetType();

		if (this._activeMode.getMainPanelContent().indexOf(layoutItem.name) !== -1) {
			widgetType = "maincontent";
		}

		switch (widgetType) {
			case "messagebar":
				widget.setContext(this._messagebar);

				break;
			case "topbar": //widget needs to be added to the topbar
				widget.setContext(this._topbar);
				break;
			case "wizard":
				widget.setContext(this._aboveCenter);
				this._items[widget.getID()] = widget;
				break;

			case "maincontent":
			default:
				var widgetWrapper = new FlexWidgetWrapper(widget, layoutItem.name, true);
				this._items[layoutItem.name] = widgetWrapper;

				if (widgetType === "maincontent") {
					this._mainContent.append(widgetWrapper.element);
					this._mainContentWidgetWrappers.push(widgetWrapper);
					this._updateMainContent();
				} else {
					widgetWrapper.setActive(false);
					this._addLayoutItem(layoutItem.name); //make sure the item is added on the correct location in one of the panels
				}
				this._flexDrag.enableDragging(widgetWrapper);

				this._registerListener(
					"enabled",
					function(isEnabled) {
						this._toggleWidget(layoutItem.name, widgetWrapper, isEnabled);
						this._emitLayoutChanged();
					},
					this,
					widget
				);

				this._registerListener(
					"visible",
					function() {
						this._emitLayoutChanged();
					},
					this,
					widget
				);

				this._registerListener(
					"collapsed",
					function() {
						this._emitLayoutChanged();
					},
					this,
					widget
				);

				this._registerListener(
					"dragged",
					function() {
						this._emitLayoutChanged();
					},
					this,
					widget
				);

				this._registerListener(
					"modalResize",
					function() {
						this._emitLayoutChanged();
					},
					this,
					widget
				);

				this._registerListener("makeVisible", _.bind(this._makeWidgetVisible, this, layoutItem.name, widgetWrapper), undefined, widget);

				if (widget.didRequestMakeVisible()) {
					this._makeWidgetVisible(layoutItem.name, widgetWrapper);
				}

				this._updateItemVisibility();

				break;
		}
	} else if (layoutItem.type === LayoutConstants.LayoutItemType.TOOLBARCOMPONENT) {
		//add a toolbarcomponent
		var toolbarItem = layoutItem.getItem(this.getLayoutType());
		var toolbarItemWrapper = new ToolbarComponentWrapper(toolbarItem);
		toolbarItemWrapper.element[0].name = layoutItem.name;
		this._items[layoutItem.name] = toolbarItemWrapper;
		this._addLayoutItem(layoutItem.name);
		this._registerListener(
			"changed.visibility",
			function() {
				this._debouncedInvalidate();
			},
			this,
			toolbarItem
		);
		this._debouncedInvalidate();
		this._updateItemVisibility();
	} else {
		//not supported in this layout
	}
};

/**
 *  Add a class to the wrapper based on the layout-mode
 */
FlexLayout.prototype._switchLayoutMode = function(mode) {
	if (this._activeMode) {
		this._toolbar.removeClass("layout-mode-" + this._activeMode.getName());
	}
	Layout.prototype._switchLayoutMode.call(this, mode);
	if (this._activeMode) {
		this._toolbar.addClass("layout-mode-" + this._activeMode.getName());
	}
};

/**
 * Update visibility of the added items based on whether or not they are allowed in the current mode.
 * E.g. item was added in previous mode, but should not be shown in current mode
 */
FlexLayout.prototype._updateItemVisibility = function() {
	_.each(
		this._items,
		_.bind(function(item) {
			var layoutItemType = this._layoutModel.getLayoutItemType(item.name);
			if (layoutItemType === LayoutConstants.LayoutItemType.TOOLBARCOMPONENT) {
				//var toolbar = layoutItem.getItem(this.getLayoutType());
				item.setVisible(this._activeMode.isItemAllowed(item.name));
			} else if (layoutItemType === LayoutConstants.LayoutItemType.WIDGET) {
				item.setVisible(this._activeMode.isItemAllowed(item.name) && item.widget.isEnabled());
			}
		}, this)
	);
};

/**
 * Disable text selection in the entire viewer.
 */
FlexLayout.prototype._disableTextSelection = function() {
	this._context
		.find(".enableTextSelection")
		.toggleClass("enableTextSelection", false)
		.toggleClass("enableTextSelectionTempDisabled", true);
};

/**
 * Enable text selection in the entire viewer.
 */
FlexLayout.prototype._enableTextSelection = function() {
	this._context
		.find(".enableTextSelectionTempDisabled")
		.toggleClass("enableTextSelection", true)
		.toggleClass("enableTextSelectionTempDisabled", false);
};

/**
 * Reset the sate of the layout to the default state.
 */
FlexLayout.prototype.resetToDefault = function() {
	var iconifiedWest = this._configuration.west && this._configuration.west.iconified;
	var iconifiedEast = this._configuration.east && this._configuration.east.iconified;

	this._sections.east.setIconified(iconifiedEast);
	this._sections.east.setWidgetBarWidth(null);
	this._sections.west.setIconified(iconifiedWest);
	this._sections.west.setWidgetBarWidth(null);

	_.each(
		this._getDefaultLayout(),
		function(section, sectionName) {
			if (sectionName === "west" || sectionName === "east") {
				//reset the width of the side panels to the default

				_.each(
					section,
					function(itemName) {
						//run over each item in the section
						var item = this._items[itemName];
						if (item !== undefined && this._activeMode.isItemAllowed(itemName) && !_.contains(this._mainContentWidgetWrappers, item)) {
							//move the item to its default location.
							this.moveItem(itemName, sectionName, undefined, true, true);
						}

						if (item && item.wrapper) {
							item.wrapper.setStickyDimensions(null);
						}
					},
					this
				);
			}
		},
		this
	);

	this._layoutState = this._getDefaultLayout(); //reset the state of the layout, it now is the default\
	this._updateItemVisibility();
	this._invalidateLayout();
};

/**
 * Get default layout state
 * @returns {Object}
 */
FlexLayout.prototype._getDefaultLayout = function() {
	return this._configuration.layout;
};

/**
 * Generates the state of the layout as a string. It takes the not-visible elements into account so state of those
 * elements is not lost.
 * @returns {Object}
 */
FlexLayout.prototype.getLayoutState = function() {
	var stateObject = {};
	_.each(
		this._sections,
		_.bind(function(section, sectionName) {
			var context;
			if (section instanceof FlexSection) {
				context = section.getContext();
				stateObject[sectionName] = {};
			} else {
				context = this._sections[sectionName].getWidgetbarContainer();
				stateObject[sectionName] = this._sections[sectionName].getState();
			}

			stateObject[sectionName].items = [];

			_.each(
				context.children(),
				_.bind(function(item) {
					if (item !== null && item !== undefined) {
						var itemWrapper = item.wrapper;
						var itemObject = {};
						if (itemWrapper) {
							itemObject.name = itemWrapper.name;
							itemObject.isCollapsed = itemWrapper.isCollapsed;
							itemObject.isActive = itemWrapper.isActive();
							itemObject.stickyDimensions = itemWrapper.getStickyDimensions();
							stateObject[sectionName].items.push(itemObject);
						}
					}
				}, this)
			);
		}, this)
	);

	if (this._layoutState) {
		// The state of the current maincontentpanel should also be stored for when the layout mode switches and it no longer is the main content.
		// We will insert its state into 'stateObject' in the correct position
		var mainPanelContents = this._activeMode.getMainPanelContent();

		_.each(
			mainPanelContents,
			function(mainPanelContent) {
				// All states of layout items in the same section as the main content panel
				var layoutStateItems =
					this._layoutState[this._sectionMapping[mainPanelContent]] && this._layoutState[this._sectionMapping[mainPanelContent]].items;

				if (layoutStateItems) {
					// Map to the name to make the searching of index easier.
					var layoutStateItemsNames = layoutStateItems.map(function(item) {
						return item.name;
					});

					// this is the state of the main content panel
					var mainContentPanelLayout = _.find(layoutStateItems, function(item) {
						return item && item.name === mainPanelContent;
					});

					if (mainContentPanelLayout) {
						// these items are all items that come after the main content item.
						var itemsAfterMainContentItem = _.rest(layoutStateItemsNames, layoutStateItemsNames.indexOf(mainPanelContent) + 1);

						// to check where to insert it, we'll loop over the items of stateObject and find the first item that is also in itemsAfterMainContentItem
						var index = _.findIndex(stateObject[this._sectionMapping[mainPanelContent]].items, function(item) {
							return _.contains(itemsAfterMainContentItem, item.name);
						});
						if (index > -1) {
							stateObject[this._sectionMapping[mainPanelContent]].items.splice(index, 0, mainContentPanelLayout);
						} else {
							stateObject[this._sectionMapping[mainPanelContent]].items.push(mainContentPanelLayout);
						}
					}
				}
			}.bind(this)
		);
	}
	this._layoutState = stateObject;
	return stateObject;
};

/**
 * Invalidates the layout. Makes sure all panels have the correct size, resize is triggered on the correct elements,...
 */
FlexLayout.prototype._invalidateLayout = function() {
	if (!this._cssLoaded) {
		return;
	}

	// make sure subtoolbar are moved when new items are added
	_.each(
		this._items,
		function(layoutItem, itemName) {
			if (this._layoutModel.getLayoutItemType(itemName) === LayoutConstants.LayoutItemType.TOOLBARCOMPONENT) {
				layoutItem.moved();
			}
		},
		this
	);

	this._sections.east.invalidateLayout();
	this._sections.west.invalidateLayout();
	this._updateItemVisibility();
};

/**
 * Returns the type of the layout.
 * @override
 * @return {Number}
 */
FlexLayout.prototype.getLayoutType = function() {
	return LayoutConstants.LayoutType.FLEX;
};

/**
 * Get all the currently visible popups
 * @returns {Array.<JQueryObj>}
 */
FlexLayout.prototype.getAllVisiblePopups = function() {
	return this._context.find(".popup:visible");
};

/**
 * Enable/disable fullscreen
 * Note: this method should be called from a user generated event... (or it won't work...)
 * @param enabled
 */
FlexLayout.prototype.toggleFullscreen = function(enabled) {
	if (enabled) {
		this._requestFullscreen();
	} else {
		this._exitFullscreen();
	}
};

/**
 * Check if the app is fullscreen
 * @returns {Boolean}
 */
FlexLayout.prototype.isFullscreen = function() {
	return FullscreenUtil.fullscreenElement;
};

/**
 * Check if the browser supports the fullescren API
 * @returns {Boolean}
 */
FlexLayout.prototype.supportsFullscreen = function() {
	return FullscreenUtil.fullscreenEnabled;
};

/**
 * Request fullscreen for the whole document.body
 */
FlexLayout.prototype._requestFullscreen = function() {
	var viewerDiv = this.f.getContext().get(0);
	FullscreenUtil.requestFullscreen(viewerDiv);
};

/**
 * Exit fullscreen mode
 */
FlexLayout.prototype._exitFullscreen = function() {
	FullscreenUtil.exitFullscreen();
};

/**
 * Update the contents of the main panel
 * A mode can decide to show something else than the viewport widget
 */
FlexLayout.prototype._updateMainContent = function() {
	var mainPanelContents = this._activeMode.getMainPanelContent();
	this._activeMainPanelContent = [];

	var copy = _.clone(this._mainContentWidgetWrappers);

	// Partition wrapers, wrappers that should be hidden are hidden first, only then are new wrappers shown
	var partition = _.partition(copy, function(wrapper) {
		return mainPanelContents.indexOf(wrapper.widget.getID()) === -1;
	});

	_.each(
		partition[0],
		function(wrapper) {
			wrapper.setVisible(false);
			wrapper.setHeaderEnabled(true);

			var section = this._sections[this._sectionMapping[wrapper.name]];
			if (section) {
				// If the widget belongs to a section we should remove it from the main content and add it to that section
				this._mainContentWidgetWrappers = _.without(this._mainContentWidgetWrappers, wrapper);
				wrapper.content.css({
					height: "",
					width: "",
					"min-height": "",
					"min-width": ""
				});
				wrapper.element.css({
					height: "",
					width: ""
				});
				var previousState = _.find(this._layoutState[this._sectionMapping[wrapper.name]].items, function(item) {
					return item.name === wrapper.name;
				});
				if (previousState) {
					wrapper.setActive(previousState.isActive);
				}
				this._addLayoutItem(wrapper.name);
				if (wrapper.widget.getMinimumSize().height && wrapper.widget.getMinimumSize().height > 0) {
					wrapper.content.css({
						flex: "1 0 auto",
						height: wrapper.widget.getMinimumSize().height + "px"
					});
				}
			}
		}.bind(this)
	);

	_.each(
		partition[1],
		function(wrapper) {
			wrapper.unstick();
			wrapper.setSidebar(null);
			wrapper.setHeaderEnabled(false);
			wrapper.setActive(true, true);
			wrapper.setVisible(true);

			this._activeMainPanelContent.push(wrapper);
			wrapper.content.css({
				flex: "",
				height: "100%",
				width: "100%"
			});

			wrapper.element.css({
				order: mainPanelContents.indexOf(wrapper.widget.getID()),
				"min-height": "",
				"min-width": ""
			});
		}.bind(this)
	);

	var activeMainPanelContentNames = this._activeMainPanelContent.map(function(wrapper) {
		return wrapper.widget.getID();
	});

	_.chain(mainPanelContents)
		.filter(
			function(mainPanelContent) {
				// The item was added but isn't a standard main content widget wrapper
				return this._items[mainPanelContent] && activeMainPanelContentNames.indexOf(mainPanelContent) === -1;
			}.bind(this)
		)
		.each(
			function(mainPanelContent) {
				// Add it to the main content widget wrapper
				this._mainContentWidgetWrappers.push(this._items[mainPanelContent]);
				this._mainContent.append(this._items[mainPanelContent].element);
			}.bind(this)
		)
		.tap(
			function(a) {
				// if there were widgets that weren't added yet,
				if (a.length > 0) {
					this._updateMainContent();
				} else {
					this._debouncedInvalidate();
				}
			}.bind(this)
		);
};

/**
 * Enable/disable a widget
 * @param {String}        widgetName    - name of the widget
 * @param {WidgetWrapper} widgetWrapper - widgetWrapper in which the widget is displayed
 * @param {Boolean}       isEnabled     - should the widget be enabled
 */
FlexLayout.prototype._toggleWidget = function(widgetName, widgetWrapper, isEnabled) {
	if (isEnabled) {
		if (this._activeMode.isItemAllowed(widgetName)) {
			widgetWrapper.setVisible(true);
		}
	} else {
		widgetWrapper.setVisible(false);
	}
};

/**
 * Returns the default layout mode for the layout.
 * @return {LayoutMode}
 */
FlexLayout.prototype._getDefaultLayoutMode = function() {
	return this._layoutModel.getLayoutMode(LayoutConstants.LayoutMode.VIEW2D);
};

FlexLayout.prototype.isItemAllowed = function(itemname) {
	return !_.contains(this._configuration.disabledItems, itemname);
};

/**
 * Enable/disable the sidebars
 * @param {Boolean} enabled
 */
FlexLayout.prototype.setSidebarsEnabled = function(enabled) {
	this._sections.east.setEnabled(enabled);
	this._sections.west.setEnabled(enabled);
};

/**
 * Hide a layout item.
 * @param {LayoutConstants.Widget|LayoutConstants.ToolbarComponent} itemName
 */
FlexLayout.prototype.hideItem = function(itemName) {
	var layoutItem = this._items[itemName];
	if (layoutItem !== undefined && layoutItem.context !== undefined) {
		layoutItem.context.hide();
	} else if (layoutItem !== undefined && layoutItem.element !== undefined) {
		layoutItem.element.hide();
	}
};

/**
 * Makes a hidden layout item visible.
 * @param {LayoutConstants.Widget|LayoutConstants.ToolbarComponent} itemName
 */
FlexLayout.prototype.showItem = function(itemName) {
	var layoutItem = this._items[itemName];
	if (layoutItem !== undefined && layoutItem.context !== undefined) {
		layoutItem.context.show();
	} else if (layoutItem !== undefined && layoutItem.element !== undefined) {
		layoutItem.element.show();
	}
};

/**
 * Is the layout item visible.
 * @param {String} itemName
 * @returns {Boolean}
 */
FlexLayout.prototype.isItemVisible = function(itemName) {
	var layoutItem = this._items[itemName];
	if (layoutItem !== undefined) {
		return layoutItem.isVisible();
	}
};
/**
 * Make sure a widget is visible
 * Uncollapse it if it is collapsed
 *
 * If the widget is modal, show it
 * If the widget is not modal, make the widget modal if it is in a hidden/collapsed sidebar
 * @param {String} layoutItemName - the name of the Layout Item
 * @param {WidgetWrapper} widgetWrapper - the widgetWrapper for the widget
 */
FlexLayout.prototype._makeWidgetVisible = function(layoutItemName, widgetWrapper) {
	var section = this._sections[this._sectionMapping[layoutItemName]];
	if (section && !_.contains(this._activeMainPanelContent, widgetWrapper)) {
		section.toggleWidgetWrapperActive(widgetWrapper, true);
		widgetWrapper.setCollapsed(false);
	} else {
		widgetWrapper.setVisible(true);
	}
	widgetWrapper.widget.setMakeVisibleRequestAccepted();
};

/**
 * Defines the array of css files that should be loaded
 *
 * @returns Array<String> list of css files
 */
FlexLayout.prototype._loadCss = function() {
	require("./less/flex.less");
};

/**
 * Makes the center content pane fullscreen
 *
 * @returns True if making center pane fullscreen was succesful
 */
FlexLayout.prototype.fullscreenViewport = function() {
	if (FullscreenUtil.fullscreenEnabled) {
		FullscreenUtil.requestFullscreen(this._centerContent[0]);
		return true;
	}
	return false;
};

FlexLayout.prototype.getMainContentElement = function() {
	return this._centerContent;
};

/**
 * @override
 */
FlexLayout.prototype.getDefaultTool = function() {
	return (this._activeMode && this._activeMode.getDefaultTool()) || CoreConstants.TOOLS.ZOOM;
};

export default FlexLayout;
