import PrivateChannel from "@esko/web-viewer/core/PrivateChannel";
import _ from "underscore";
import $ from "jquery";
import "jquery-ui/ui/widgets/resizable";
/**
 * Sidebar that can be used in a flex layout
 * it has a Tile bar and a widgetbar and can toggle between the modes
 * @parma {ViewerFacade} f
 * @param {String} name - unique name for the sidebar
 * @param {String} positionClass - position of the sidebar (left|right)
 * @param {Boolean} isIconified - isIconified
 */
var FlexSidebar = function(f, name, positionClass, container, isIconified) {
	PrivateChannel.installTo(this, f, ["layout", "flex", "sidebar", name]);

	this._isResizable = false;
	this._isIconified = true;
	this._widgetBarWidth = null;
	this._container = container;

	this._tilebar = $("<div>", {
		class: "sidebar iconified " + positionClass
	});
	this._widgetbar = $("<div>", {
		class: "widgetbar iconified " + positionClass
	});

	// To enable resizable with a scrolling div we unfortunately need an extra container div for the widgets...
	this._widgetbarContainer = $("<div>", {
		class: "widgetbarcontainer"
	}).appendTo(this._widgetbar);

	this._toggleIconified = $("<a>", {
		title: f.t(this.isRight() ? "LayoutToggleRightSidebar" : "LayoutToggleLeftSidebar"),
		class: "iconifierToggle",

		on: {
			dblclick: _.bind(function() {
				this._widgetbar.css("width", "");
				this._widgetBarWidth = null;
			}, this),

			click: _.bind(function() {
				this.setIconified(!this._isIconified);
			}, this)
		}
	}).append(
		$("<i>", {
			class: "icon icon-WCR_flexsidebarToggleIcon " + (this.isRight() ? "icon-WCR_Viewer-Sidebar-Left8" : "icon-WCR_Viewer-Sidebar-Right8")
		})
	);

	this._tilebar.append(this._toggleIconified);
	this._widgetWrappers = [];
	this._activeWrapper = null;

	this.setIconified(isIconified);
};

/**
 * Get the context element for the tile bar
 * @returns {JQueryObj}
 */
FlexSidebar.prototype.getTilebarContext = function() {
	return this._tilebar;
};

/**
 * Get the context element for the widget bar
 * @returns {JQueryObj}
 */
FlexSidebar.prototype.getWidgetbarContext = function() {
	return this._widgetbar;
};

/**
 * Get the container in which widgets reside
 * @returns {JQueryObject}
 */
FlexSidebar.prototype.getWidgetbarContainer = function() {
	return this._widgetbarContainer;
};

/**
 * Append a WidgetWrapper to the sidebar
 * @param {FlexWidgetWrapper} item
 */
FlexSidebar.prototype.appendItem = function(item) {
	this._widgetbarContainer.append(item.element);
	this._tilebar.append(item.getTileElement());

	this._handleNewItem(item);
};

/**
 * Prepend a WidgetWrapper begore another existing WidgetWrapper
 * @param {FlexWidgetWrapper} itemToPrepend - the new widgetwrapper to add to the bar
 * @param {FlexWidgetWrapper} existingItem - the existing WidgetWrapper before which the new item should be prepended
 */
FlexSidebar.prototype.prependItemBeforeItem = function(itemToPrepend, existingItem) {
	existingItem.element.before(itemToPrepend.element);
	existingItem.getTileElement().before(itemToPrepend.getTileElement());

	this._handleNewItem(itemToPrepend);
};

/**
 * Internal method to handle a new item
 * @param {FlexWidgetWrapper} item
 */
FlexSidebar.prototype._handleNewItem = function(item) {
	item.setSidebar(this);
	if (this._isIconified) {
		if (item.isActive()) {
			item.unstick();
			item.setActive(false);
			_.defer(
				_.bind(function() {
					this.toggleWidgetWrapperActive(item, true);
				}, this)
			);
		} else {
			this._fixActiveWrapperPosition();
		}
	} else {
		item.setActive(true);
		item.unstick();
	}
};

/**
 * Unregister a widgetwrapper from the sidebar
 * @param {FlexWidgetWrapper} widgetWrapper
 */
FlexSidebar.prototype.unregisterWidgetWrapper = function(widgetWrapper) {
	this._widgetWrappers = _.without(this._widgetWrappers, widgetWrapper);
	if (this._activeWrapper === widgetWrapper) {
		this._activeWrapper = null;
	}

	this._fixActiveWrapperPosition();
};

/**
 * Register a widgetwrapper at the sidebar (used for active behavior in iconified mode)
 * @param {FlexWidgetWrapper} widgetWrapper
 */
FlexSidebar.prototype.registerWidgetWrapper = function(widgetWrapper) {
	this._widgetWrappers.push(widgetWrapper);
};

/**
 * Toggle a widget wrapper to be active
 * @param {FlexWidgetWrapper} the widgetwrapper to change
 * @param {Boolean} [active] - should the wrapper become active of not (if undefined, the active state is toggled)
 */
FlexSidebar.prototype.toggleWidgetWrapperActive = function(widgetWrapper, active) {
	if (active === undefined) {
		active = !widgetWrapper.isActive();
	}

	if (this._isIconified) {
		if (active) {
			_.each(this._widgetWrappers, function(wrapper) {
				wrapper.setActive(wrapper === widgetWrapper, true);
			});
			this._activeWrapper = widgetWrapper;
			this._activeWrapper.stickToTile();
		} else {
			widgetWrapper.setActive(false);
			if (this._activeWrapper === widgetWrapper) {
				this._activeWrapper = null;
			}
		}
	}
};

/**
 * Make the widgetbar resizable
 * @param {Boolean} resizable
 */
FlexSidebar.prototype.setResizable = function(resizable) {
	if (!resizable && this._isResizable) {
		this._widgetbar.resizable("destroy");
		this._isResizable = false;
	}

	if (resizable) {
		var sizeBeforeStart;
		this._widgetbar.resizable({
			handles: this.isRight() ? "w" : "e",
			start: function(event, ui) {
				sizeBeforeStart = ui.size.width;
				$(this).css("position", "");
			},
			resize: function(event, ui) {
				ui.position.left = "0";
				$(this).css("left", "");
			},
			stop: _.bind(function(event, ui) {
				this._widgetbar.css({
					position: "",
					left: "",
					top: ""
				});
				if (sizeBeforeStart != ui.size.width) {
					this._widgetBarWidth = ui.size.width;
					_.each(this._widgetWrappers, function(wrapper) {
						wrapper.widget.resizing();
					});
					this._emit("changed.resize");
				}
			}, this),
			maxWidth: "50%",
			minWidth: 150
		});
		this._isResizable = true;
	}
};

/**
 * Switch between iconified and expanded mode
 * @param {Boolean} iconify - true if sidebar should become iconified
 */
FlexSidebar.prototype.setIconified = function(iconify) {
	if (this._isIconified != iconify) {
		_.each(this._widgetWrappers, function(wrapper) {
			wrapper.setActive(!iconify);
			if (!iconify) {
				wrapper.unstick();
			}
		});

		this._isIconified = iconify;

		this._tilebar.toggleClass("hiddenElement", !iconify);

		this._emit("changed.iconified", !iconify, iconify);

		this._widgetbar.toggleClass("iconified", iconify);
		if (!iconify) {
			this._activeWrapper = null;
			this._widgetbarContainer.prepend(this._toggleIconified);

			if (this._widgetBarWidth !== null) {
				this._widgetbar.css("width", this._widgetBarWidth);
			}
		} else {
			this._tilebar.prepend(this._toggleIconified);
			this._widgetbar.css({
				width: "",
				height: ""
			});
		}

		this._toggleIconified
			.find(".icon-WCR_flexsidebarToggleIcon")
			.toggleClass("icon-WCR_Viewer-Sidebar-Right8", (!this.isRight() && iconify) || (this.isRight() && !iconify));
		this._toggleIconified
			.find(".icon-WCR_flexsidebarToggleIcon")
			.toggleClass("icon-WCR_Viewer-Sidebar-Left8", (!this.isRight() && !iconify) || (this.isRight() && iconify));

		this.setResizable(!iconify);
	}
};
/**
 * Check if the sidebar is iconified
 * @returns {Boolean}
 */
FlexSidebar.prototype.isIconified = function() {
	return this._isIconified;
};

/**
 * Fix the position of the active wrapper
 */
FlexSidebar.prototype._fixActiveWrapperPosition = function() {
	if (this._activeWrapper) {
		this._activeWrapper.stickToTile();
	}
};

/**
 * Get the state of the sidebar
 * @returns {Object} serializable state object
 */
FlexSidebar.prototype.getState = function() {
	return {
		iconified: this.isIconified(),
		widgetbarWidth: this._widgetBarWidth
	};
};

/**
 * Restore the state of the sidebar
 * @param {Object} state - serialized state as returned from getState()
 */
FlexSidebar.prototype.restoreState = function(state) {
	this.setIconified(state.iconified);
	this.setWidgetBarWidth(state.widgetbarWidth);
};

/**
 * Set the sidebar enabled
 * @param {Boolean} enabled - if false the sidebar is hidden
 */
FlexSidebar.prototype.setEnabled = function(enabled) {
	if (!enabled) {
		this._tilebar.toggleClass("hiddenElement", true);
		this._widgetbar.toggleClass("hiddenElement", true);
	} else {
		this._tilebar.toggleClass("hiddenElement", !this.isIconified());
		this._widgetbar.toggleClass("hiddenElement", false); // widgetbar should also be visible in iconified mode (the sticky widget are still in the widgetbar container when they are stuck to their tile)
		this._fixActiveWrapperPosition();
	}

	this._emit("changed.enabled", enabled);
};

/**
 * Set the width of the widgetbar
 * @param {Number} widgetbarWidth - set the width of the widgetbar, if null, the width is based on its contents
 */
FlexSidebar.prototype.setWidgetBarWidth = function(widgetbarWidth) {
	if (!this.isIconified()) {
		if (widgetbarWidth !== null) {
			this._widgetbar.css("width", widgetbarWidth);
		} else {
			this._widgetbar.css("width", "");
		}
	}

	this._widgetBarWidth = widgetbarWidth;
};

/**
 * Is this a right of left aligned sidebar?
 * @returns {Boolean} true if is right
 */
FlexSidebar.prototype.isRight = function() {
	return this._tilebar.is(".right");
};

/**
 * Get the container of the FlexSideBar
 * @returns {JQueryObj}
 */
FlexSidebar.prototype.getContainer = function() {
	return this._container;
};

/**
 * Make sure the layout looks OK when resized:
 * Fix the resizable properties when in iconified mode
 */
FlexSidebar.prototype.invalidateLayout = function() {
	this._fixActiveWrapperPosition();
};

export default FlexSidebar;
