import WidgetWrapper from "@esko/web-viewer/layout/WidgetWrapper";
import ShapeUtils from "@esko/shapeutils";
import _ from "underscore";
import $ from "jquery";
import "jquery-ui/ui/widgets/resizable";
/**
 * A widgetwrapper is a wrapper for a widget, which makes enables the widget to be displayed with some extra controls
 * @exports FlexWidgetWrapper
 * @augments WidgetWrapper
 * @param {AbstractWidget} widget          - the widget to wrap
 * @param {String}         widgetName      - the name of the widget
 * @param {Boolean} allowCollapse - is collapsing the widget allowed?
 */
var FlexWidgetWrapper = function(widget, widgetName, allowCollapse) {
	WidgetWrapper.call(this, widget, widgetName, allowCollapse);

	/**
	 * @member {Boolean} is the widgetwrapper currently resizable
	 */
	this._isResizable = false;
	/**
	 * @member {Object} contains the dimensions for the widget, if null, the widgetwrapper should try to fit around the containt of the widget
	 */
	this._stickyDimensions = null;

	/**
	 * @member {Boolean} tru if the widgtet is sticky
	 */
	this._isSticky = false;

	/**
	 * @member {JQueryObj} Tile to represent the widget
	 */
	this._tile = $("<a>", {
		class: "tile",
		title: widget.getTitle()
	});

	this._tile[0].wrapper = this;

	/**
	 * @member {JQueryObj} icon used for the tile
	 */
	this._tileIcon = $("<i>", {
		class: widget.getIcon()
	});
	/**
	 * @member {JQueryObj} Label for the tile
	 */
	this._tileSpan = $("<span>", {
		text: widget.getTitle()
	});

	this._tile.on(
		"click",
		_.bind(function() {
			if (this._sidebar) {
				this._sidebar.toggleWidgetWrapperActive(this);
			}
		}, this)
	);

	// reset dimensions on double click
	this._tile.on(
		"dblclick",
		_.bind(function() {
			this.setStickyDimensions(null);
		}, this)
	);

	if (widget.willFill()) {
		this.element.addClass("flex");
	}

	widget.on(
		"iconchanged",
		function(newIcon) {
			this._tileIcon.removeClass().addClass(newIcon);
		}.bind(this)
	);

	widget.on("resized", this._onWidgetResized.bind(this));

	this._tile.append(this._tileIcon, this._tileSpan);

	/**
	 * @member {FlexSidebar} sidebar in which the widget is placed
	 */
	this._sidebar = null;

	/**
	 * @member {Boolean}
	 */
	this._isActive = true;
};

FlexWidgetWrapper.prototype = Object.create(WidgetWrapper.prototype);
FlexWidgetWrapper.prototype.constructor = FlexWidgetWrapper;

/**
 * Get a reference to the tile element of the widget wrapper
 * @returns {JQueryObj}
 */
FlexWidgetWrapper.prototype.getTileElement = function() {
	return this._tile;
};

/**
 * @Override
 */
FlexWidgetWrapper.prototype._onTitleChanged = function(newTitle) {
	WidgetWrapper.prototype._onTitleChanged.call(this, newTitle);
	this._tile.attr("title", newTitle);
	this._tileSpan.text(newTitle);
};

/**
 * Stick the widget wrapper to its tile
 * This makes the widgetwrapper absolute positioned
 */
FlexWidgetWrapper.prototype.stickToTile = function() {
	var css = this._calculateStickyCSS();
	// for a flexbox bug in Safare we quickly set the widgetwrapper to display block, and then remove it later on...
	// if we don't do this hack, the sidepanel container is visible...
	this.element
		.addClass("absolute-position")
		.css("display", "block")
		.css(css);
	this._setHeightFixForIE();
	this._isSticky = true;
	this.setResizable(true, css);
	// remove the temporarily set display css property
	this.element.css("display", "");
	this._updateWidgetVisibility();
};

/**
 * Calculate CSS for a sticky widget
 * This calculates the correct min/max dimensions and position
 * @returns {Object} key-value pairs of CSS properties to be set
 */
FlexWidgetWrapper.prototype._calculateStickyCSS = function() {
	var offset = this._tile.position();
	var tileWidth = this._tile.outerWidth() || null;
	var isRight = this._tile.parents(".sidebar").is(".right");
	var topPos = offset.top;
	var leftPos = "";
	var rightPos = "";
	if (isRight) {
		rightPos = tileWidth;
	} else {
		leftPos = tileWidth;
	}
	var container = this._sidebar.getContainer();

	var maxHeight = (container.innerHeight() || null) - topPos;
	// only allow widgets to come to the middle of the container
	var maxWidth = Math.floor(0.495 * ((container.innerWidth() || null) - 2 * tileWidth));

	if (this.getMaximumSize()) {
		var maxSize = this.getMaximumSize();
		maxHeight = maxSize.height;
		maxWidth = maxSize.width;
	}

	var minSize = this.widget.getMinimumSize() || {};

	var css = {
		top: topPos,
		left: leftPos,
		right: rightPos,
		"max-height": Math.floor(maxHeight),
		"max-width": maxWidth,
		"min-width": minSize.width,
		"min-height": minSize.height
	};

	if (this._stickyDimensions) {
		css.width = this._stickyDimensions.width ? Math.min(this._stickyDimensions.width, maxWidth) : "";
		css.height = this._stickyDimensions.height ? Math.min(this._stickyDimensions.height, maxHeight) : "";
	} else {
		css.width = "";
		css.height = "";
	}

	return css;
};

/**
 * This sets height of the flex widget wrapper as IE's flexbox doesn't work well with max-heights.
 * The content div will overflow the FlexWidgetWrapper without a scrollbar. Setting the height of the
 * FlexWidgetWrapper makes sure that the content does overflow with a scrollbar.
 */
FlexWidgetWrapper.prototype._setHeightFixForIE = function() {
	if (this._sidebar && this._tile && this.content && this.header) {
		var container = this._sidebar.getContainer();
		var topPos = this._tile[0].offsetTop;

		var maxHeight = (container.innerHeight() || null) - topPos;
		var contentHeight = this.content.outerHeight() || null;
		var headerHeight = this.header.outerHeight() || null;

		// If the height of the content + header is bigger than the max size of the widgetwrapper we should
		// set the height of the widget wrapper.
		if (maxHeight < contentHeight + headerHeight) {
			this.element.height(maxHeight);
		}
	}
};

/**
 * Because IE can't handle flexbox nicely with max-height and overflow scroll, we need to set the height on widget resize
 * See also _setHeightFixForIE
 */
FlexWidgetWrapper.prototype._onWidgetResized = function() {
	var isVisible = this.isVisible() && this.isActive();
	if (isVisible) {
		this._setHeightFixForIE();
	}
};

/**
 * Reset the sticky dimensions of the widget wrapper
 */
FlexWidgetWrapper.prototype.resetStickyDimensions = function() {
	this._stickyDimensions = null;
	if (this._isSticky) {
		this.element.css({
			width: "",
			height: ""
		});
	}
};

/**
 * Unstick the widgetwrapper
 */
FlexWidgetWrapper.prototype.unstick = function() {
	this._isSticky = false;

	this.setResizable(false);

	this.element.removeClass("absolute-position").css({
		"max-height": "",
		top: "",
		left: "",
		right: "",
		width: "",
		height: "",
		"max-width": "",
		position: ""
	});
};

/**
 * Set the sidebar of the widgetwrapper
 * @param {FlexSidebar} sidebar
 */
FlexWidgetWrapper.prototype.setSidebar = function(sidebar) {
	if (this._sidebar && this._sidebar != sidebar) {
		this._sidebar.unregisterWidgetWrapper(this);
	}

	this._sidebar = sidebar;
	if (this._sidebar) {
		this._tile.removeClass("hiddenElement");
		this._sidebar.registerWidgetWrapper(this);
		this._updateWidgetVisibility();
	} else {
		this.element.addClass("hiddenElement");
		this._tile.addClass("hiddenElement");
	}
};

/**
 * Get the sidebar of the widgetwrapper
 * @return {FlexSidebar} sidebar
 */
FlexWidgetWrapper.prototype.getSidebar = function() {
	return this._sidebar;
};

/**
 * Set the widget as active
 * @param {Boolean} active
 * @param {Boolean} [uncollapse] - if active and uncollapse true, the widget is uncollapsed
 */
FlexWidgetWrapper.prototype.setActive = function(active, uncollapse) {
	if (active & uncollapse) {
		this.setCollapsed(false);
	}

	this._isActive = active;
	this._tile.toggleClass("active", active);
	this._updateWidgetVisibility();
};
/**
 * Is the widgetwrapper active?
 * @returns {Boolean}
 */
FlexWidgetWrapper.prototype.isActive = function() {
	return this._isActive;
};

/**
 * @override
 */
FlexWidgetWrapper.prototype.setVisible = function(isVisible) {
	this._isVisible = isVisible;
	if (!this._isVisible) {
		var offset = this.element.offset();
		//location is only used when the widgetwrapper is hidden, if visible the element is used
		this._locationWhenHidden = new ShapeUtils.Point(offset.left, offset.top);
		this._tile.addClass("hiddenElement");
	} else if (this._sidebar) {
		this._tile.removeClass("hiddenElement");
	}

	this._updateWidgetVisibility();
};

/**
 * Update the visibility of the widget (based on active and visible
 */
FlexWidgetWrapper.prototype._updateWidgetVisibility = function() {
	var isVisible = this.isVisible() && this.isActive();
	this.element.toggleClass("hiddenElement", !isVisible);

	this.widget.setVisible(isVisible);
	if (isVisible) {
		this._setHeightFixForIE();
		this.widget.resized();
	}
};

/**
 * Toggle resizable functionality of the widget wrapper
 * @param {Boolean} resizable
 * @param {Object} [stickyCSS] - if not provided, sticky CSS is calculated automatically
 */
FlexWidgetWrapper.prototype.setResizable = function(resizable, stickyCSS) {
	if (this.resizable && !resizable) {
		this.element.resizable("destroy");
		this.resizable = false;
	} else if (resizable) {
		if (!stickyCSS) {
			stickyCSS = this._calculateStickyCSS();
		}

		//make the element resizable
		var minimumSize = this.getMinimumSize();
		_.defer(
			_.bind(function() {
				if (this._isSticky) {
					var resizeHandler = _.throttle(
						_.bind(function(event, ui) {
							this.setHeight(ui.size.height);
							this.widget.resizing();
						}, this),
						40
					);

					var resizableOptions = {
						maxWidth: stickyCSS["max-width"],
						maxHeight: stickyCSS["max-height"],
						minWidth: Math.max(100, minimumSize.width),
						minHeight: Math.max(40, minimumSize.height),
						stop: _.bind(function(event, ui) {
							if (!this._stickyDimensions) {
								this._stickyDimensions = {};
							}

							this._stickyDimensions.width = ui.size.width;
							this._stickyDimensions.height = ui.size.height;

							this.setHeight(ui.size.height);
							this.widget.fireModalResize();
						}, this),
						resize: resizeHandler
					};

					if (this._sidebar.isRight()) {
						resizableOptions.handles = "w, sw, s";
						resizableOptions.resize = function(event, ui) {
							$(this).css("left", "");
							resizeHandler(event, ui);
						};
					} else {
						resizableOptions.handles = "e, se, s";
					}

					this.element.resizable(resizableOptions);
					this.resizable = true;
				}
			}, this)
		);
	}
};

/**
 * Only notify the widget that is was resized
 * @override
 */
FlexWidgetWrapper.prototype.setHeight = function() {
	this.widget.resized();
};

/**
 * Get the sticky dimensions
 * @returns {Object}
 */
FlexWidgetWrapper.prototype.getStickyDimensions = function() {
	return this._stickyDimensions;
};

/**
 * Restore the sticky dimensions
 */
FlexWidgetWrapper.prototype.setStickyDimensions = function(stickyDimensions) {
	this._stickyDimensions = stickyDimensions;
	if (this._isSticky) {
		this.stickToTile();
	}
};

/**
 * Set the widget wrapper collapsed. If collapsed is false, the widget becomes resizable, else it's sticked to tile if sticky.
 * @param {Boolean} collapsed - True to set collapsed, false otherwise
 */
FlexWidgetWrapper.prototype.setCollapsed = function(collapsed) {
	WidgetWrapper.prototype.setCollapsed.call(this, collapsed);

	if (this._sidebar) {
		this.setResizable(!collapsed);
		if (collapsed) {
			this.element.css("height", "");
		} else {
			if (this._isSticky) {
				this.stickToTile();
			}
		}
	}
};

export default FlexWidgetWrapper;
