import _ from "underscore";
import LayoutConstants from "@esko/web-viewer/layout/LayoutConstants";
import LayoutMode from "@esko/web-viewer/layout/LayoutMode";
/**
 * @exports Mobile3DViewMode
 * @memberof Layout
 * @class
 * Mode for mobile layout, shows the annotation widget
 * @constructor
 * @augments LayoutMode
 * @param {Core.ViewerFacade} f
 * @param {Layout} layout
 */
var Mobile3DViewMode = function(f, layout) {
	LayoutMode.call(this, f, layout, LayoutConstants.LayoutMode.THREEDONLY);

	// by default this mode is disabled
	this._emitModeAvailability(false);

	var coldBootCheck3DPerspective = _.bind(function() {
		this._emitModeAvailability(this.f.getActivePage().get("has3DPerspective"));
	}, this);

	this.f.on("doc.activePageParametersChange", coldBootCheck3DPerspective);

	// check if a 3D perspective is availble for the current page (this check is needed because the 3D module is not loaded by default)
	if (this.f.getActivePage() && this.f.getActivePage().isParametersAvailable()) {
		coldBootCheck3DPerspective();
	}

	this.f.whenModuleAvailable(
		"view3D",
		function(view3DModule) {
			// whenever 3D module becomes available, start listening to the module
			// stop listening to active page parameters changes
			this.f.off("doc.activePageParametersChange", coldBootCheck3DPerspective);

			// if 3D view is started (and this widget will be published), switch to 3D mode
			view3DModule.on(
				"start3D",
				_.bind(function() {
					this._emitModeAvailability(true);
				}, this)
			);

			// 3D view is stopped (this will be because of switching documents)
			view3DModule.on(
				"stop3D",
				_.bind(function() {
					// if the mobile 3D view mode is active, and there is a 3D perspective for the active file: make sure to iniate and show the 3D view of the current fikel
					if (this.isActive() && view3DModule.has3DPerspective()) {
						view3DModule.show3D();
					}
				}, this)
			);

			// whenever the 3D perspective availability changes, hide/show the 3D view mode
			view3DModule.on(
				"has3DPerspective",
				_.bind(function(is3DAvailable) {
					this._emitModeAvailability(is3DAvailable);
				}, this)
			);
			this._emitModeAvailability(view3DModule.has3DPerspective());
		},
		this
	);
};

Mobile3DViewMode.prototype = new LayoutMode();
Mobile3DViewMode.prototype.constructor = Mobile3DViewMode;

/**
 * Notify the change in availability of the mode
 * @param {Boolean} available [[Description]]
 */
Mobile3DViewMode.prototype._emitModeAvailability = function(available) {
	this.f.emit("layout.modeAvailabilityChanged", this, available, {
		icon: "icon-WCR_3D-Create",
		title: "View3DBtnGroupTitle"
	});

	if (this.isActive() && !available) {
		// if the 3D view mode is active, we switch to the default mode (since it doesn't make sense anymore to have it)
		this.f.emit("layout.switchLayoutMode");
	}
};

/**
 * Turns On/Off the layoutMode
 * @param {Boolean} toggle
 */
Mobile3DViewMode.prototype.toggle = function(toggle) {
	if (toggle === undefined) {
		toggle = !this._isActive;
	}
	this._isActive = toggle;

	if (toggle) {
		// request the view3D module to show the 3D widget
		this.f.requestModule(
			"view3D",
			_.bind(function(view3DModule) {
				view3DModule.show3D();
			}, this),
			false
		);
	}
};

/**
 * Checks if the item is allowed for this layout mode. Prevents channels and CAD widget from loading.
 *
 * @param {String} itemname
 * @returns {Boolean}
 */
Mobile3DViewMode.prototype.isItemAllowed = function(itemname) {
	// only allow viewport widget
	return (
		itemname === LayoutConstants.ToolbarComponent.DOCUMENTSELECTOR ||
		itemname === LayoutConstants.ToolbarComponent.APPROVAL ||
		itemname === LayoutConstants.Widget.MESSAGE ||
		itemname === LayoutConstants.Widget.VIEW3D
	);
};

/**
 * Checks if the state of the layout should be persisted.
 * @return {Boolean}
 */
Mobile3DViewMode.prototype.isPersisted = function() {
	return false;
};

/**
 * Decide whether or not to show the mode panel
 * @override
 */
Mobile3DViewMode.prototype.isModePanelEnabled = function() {
	return false;
};

/**
 * @override
 */
Mobile3DViewMode.prototype.showModePanelWidgetTitles = function() {
	return false;
};

/**
 * list of needed widgets in modepanel for annotation view mode
 * @returns {Array}
 * @override
 */
Mobile3DViewMode.prototype.getModePanelContent = function() {
	return [];
};

/**
 * We want the 3D widget as Main Content
 * @override
 */
Mobile3DViewMode.prototype.getMainPanelContent = function() {
	return [LayoutConstants.Widget.VIEW3D];
};

export default Mobile3DViewMode;
