import _ from "underscore";
import CompareConstants from "@esko/web-viewer/compare/CompareConstants";
import LayoutConstants from "@esko/web-viewer/layout/LayoutConstants";
import LayoutMode from "@esko/web-viewer/layout/LayoutMode";
/**
 * @exports MobileCompare3DMode
 * @memberof Layout
 * @class
 * Mode that is used to Compare two 3D widgets on mobile
 * @constructor
 * @augments LayoutMode
 * @param {Core.ViewerFacade} f
 * @param {Layout} layout
 */
var MobileCompare3DMode = function(f, layout) {
	LayoutMode.call(this, f, layout, LayoutConstants.LayoutMode.COMPARE_3D);

	this.setIsmodePanelExpanded(false);
	this._mainPanelContents = {};
	this._mainPanelContents[CompareConstants.diffMode.SHOW3DSIDEBYSIDE] = [
		LayoutConstants.Widget.VIEW3D,
		LayoutConstants.Widget.VIEW3D_COMPARE
	];
	this._mainPanelContents[CompareConstants.diffMode.SHOWCURRENT3D] = [LayoutConstants.Widget.VIEW3D];
	this._mainPanelContents[CompareConstants.diffMode.SHOWDIFFERENCE3D] = [LayoutConstants.Widget.VIEW3D_COMPARE];

	this.f.whenModuleAvailable(
		"compare",
		function(compareModule) {
			var isCompareAvailable = compareModule.isCompareAvailable();
			this._emitModeAvailability(isCompareAvailable && f.getActivePage().is3D());

			compareModule.on(
				"changed.mode",
				_.bind(function(oldMode, newMode) {
					this._emitModeAvailability(isCompareAvailable && newMode && newMode.getLayoutMode() === LayoutConstants.LayoutMode.COMPARE_3D);
				}, this)
			);

			this.f.getDocumentController().whenActivePageChanged(function() {
				var page = this.f.getDocumentController().getActivePage();
				this._emitModeAvailability(isCompareAvailable && page.canBeComparedTo() && page.is3D());
			}, this);

			this.f.on(
				"changed.layoutMode",
				function() {
					// turn off compare when leaving this mode
					if (
						!this.isActive() &&
						compareModule.getActiveMode() &&
						compareModule.getActiveMode().getLayoutMode() === LayoutConstants.LayoutMode.COMPARE_3D
					) {
						compareModule.activateModeByName();
					}
				}.bind(this)
			);
		}.bind(this)
	);

	// This is here for blinking compare. Blinking compare switches from current diffmode to reference diffmode.
	this.f.on(
		"module.compare.changed.diffMode",
		function(oldMode, newMode) {
			if (this.isActive() && newMode !== CompareConstants.diffMode.UNDEFINED) {
				this._emit("changed"); // emit that our contents have changed.
			}
		}.bind(this)
	);
};

MobileCompare3DMode.prototype = new LayoutMode();
MobileCompare3DMode.prototype.constructor = MobileCompare3DMode;

/**
 * Turns On/Off the layoutMode
 * @param {Boolean} toggle
 */
MobileCompare3DMode.prototype.toggle = function(toggle) {
	if (toggle === undefined) {
		toggle = !this._isActive;
	}
	this._isActive = toggle;
};

/**
 * Notify the change in availability of the mode
 * @param {Boolean} available [[Description]]
 */
MobileCompare3DMode.prototype._emitModeAvailability = function(available) {
	this.f.emit("layout.modeAvailabilityChanged", this, available, {
		icon: "icon-WCR_Compare",
		title: "CompareToolsTitle"
	});
};
/**
 * Checks if the item is allowed for this layout mode. Prevents channels and CAD widget from loading.
 *
 * @param {String} itemname
 * @returns {Boolean}
 */
MobileCompare3DMode.prototype.isItemAllowed = function(itemname) {
	if (itemname === LayoutConstants.Widget.VIEW3D_COMPARE || itemname === LayoutConstants.Widget.VIEW3D) {
		var compareModule = this.f.getModule("compare");
		if (compareModule) {
			var mainPanelContent = this._mainPanelContents[compareModule.getDiffMode()];
			if (mainPanelContent) {
				return mainPanelContent.indexOf(itemname) !== -1;
			}
		}
		return false;
	}

	return (
		itemname === LayoutConstants.Widget.COMPARE ||
		itemname === LayoutConstants.ToolbarComponent.COMPARE ||
		itemname === LayoutConstants.ToolbarComponent.APPROVAL ||
		itemname === LayoutConstants.Widget.MESSAGE ||
		itemname === LayoutConstants.ToolbarComponent.VIEW3D
	);
};

/**
 * We want the 3D widget as Main Content
 * @override
 */
MobileCompare3DMode.prototype.getMainPanelContent = function() {
	var compareModule = this.f.getModule("compare");
	if (compareModule) {
		return this._mainPanelContents[compareModule.getDiffMode()] || [];
	}
	return [];
};

/**
 * Checks if the state of the layout should be persisted.
 * @return {Boolean}
 */
MobileCompare3DMode.prototype.isPersisted = function() {
	return this._layout.getLayoutMode(LayoutConstants.LayoutMode.MOBILE_COMPARE).isPersisted();
};

/**
 * Decide whether or not to show the mode panel
 * @override
 */
MobileCompare3DMode.prototype.isModePanelEnabled = function() {
	return this._layout.getLayoutMode(LayoutConstants.LayoutMode.MOBILE_COMPARE).isModePanelEnabled();
};

/**
 * @override
 */
MobileCompare3DMode.prototype.showModePanelWidgetTitles = function() {
	return this._layout.getLayoutMode(LayoutConstants.LayoutMode.MOBILE_COMPARE).showModePanelWidgetTitles();
};

/**
 * list of needed widgets in modepanel for mobile compare mode
 * @returns {Array}
 * @override
 */
MobileCompare3DMode.prototype.getModePanelContent = function() {
	return this._layout.getLayoutMode(LayoutConstants.LayoutMode.MOBILE_COMPARE).getModePanelContent();
};

export default MobileCompare3DMode;
