import Layout from "@esko/web-viewer/layout/Layout";
import LayoutConstants from "@esko/web-viewer/layout/LayoutConstants";
import ModeBar from "@esko/web-viewer/layout/mobile/ModeBar";
import NavigationBar from "@esko/web-viewer/layout/mobile/NavigationBar";
import PrivateChannel from "@esko/web-viewer/core/PrivateChannel";
import ListenerUtil from "@esko/web-viewer/utils/ListenerUtil";
import MobileViewMode from "@esko/web-viewer/layout/mobile/MobileViewMode";
import MobileVideoMode from "@esko/web-viewer/layout/mobile/MobileVideoMode";
import AnnotationViewMode from "@esko/web-viewer/layout/mobile/AnnotationViewMode";
import MobileCompareMode from "@esko/web-viewer/layout/mobile/MobileCompareMode";
import Mobile3DViewMode from "@esko/web-viewer/layout/mobile/Mobile3DViewMode";
import MobileCompare3DMode from "@esko/web-viewer/layout/mobile/MobileCompare3DMode";
import WidgetWrapper from "@esko/web-viewer/layout/WidgetWrapper";
import ToolbarComponentWrapper from "@esko/web-viewer/layout/ToolbarComponentWrapper";
import ModePanel from "@esko/web-viewer/layout/mobile/ModePanel";
import ShapeUtils from "@esko/shapeutils";
import MobileReaderSpreadMode from "@esko/web-viewer/layout/mobile/MobileReaderSpreadMode";
import FullscreenUtil from "@esko/web-viewer/utils/FullscreenUtil";
import CoreConstants from "@esko/web-viewer/core/CoreConstants";
import _ from "underscore";
import $ from "jquery";
/**
 * @class
 * Layout for mobile screens. This is the view initialized by the LayoutManager that will add all the layoutcomponents
 * to the viewer. It can generate a an object representing the current state of the viewer and can be initialized by
 * a state-String.
 *
 * NOTE widget that are both MainPanelContent and ModelPanelContent are not supported at the moment!
 *
 * @exports MobileLayout
 * @constructor
 * @param {Core.ViewerFacade} f
 * @param {LayoutModel} layoutModel
 * @param {JQueryObj} context - context where the layout will add its components to
 * @param {Object} layoutState
 * @param {LoadingIndicator} loadingIndicator - the loading indicator, which was created before all requests started
 * @param {JQueryObj} initialLoader - the JQuery reference to the element indicating the initial loader (needed to safely remove it without emptying the context)
 */
var MobileLayout = function(f, layoutModel, context, layoutState, loadingIndicator, initialLoader) {
	Layout.call(this, f, layoutModel, context, layoutState, loadingIndicator, initialLoader);
	PrivateChannel.installTo(this, f, "layout.mobile");
	ListenerUtil.mixin(this, f);

	if (!this._layoutState) {
		this._layoutState = this._getDefaultLayout();
	}

	/**
	 * All the layout items mapped as Name -> WidgetWrapper
	 */
	this._items = {};

	/**
	 * The component that is used to display the content in the viewport: graphics, CAD, annotations,...
	 */
	this._mainContent = null;

	/**
	 * In this array all items that are for the main content are kept
	 */
	this._mainContentWidgetWrappers = [];

	this._centerContent = $("<div>").addClass("centerContent");

	this._modeBar = new ModeBar(this.f, $("<div>"), this);
	this._modeBar.setState(this._layoutState.modeBar);

	this._appContentWrapper = $("<div>").addClass("appContentWrapper");

	this._messagebar = $("<div />").addClass("messagebar");

	this._navigationBar = new NavigationBar(this.f, $("<div>"), this);

	this._modePanel = new ModePanel(f, "modePanel");

	// stop the bouncing effect in iOS safari (we can only stop propagation if we are not in an interactive element (or button or mode)
	$(context).on(
		"touchmove",
		_.bind(function(e) {
			if (
				e.target === this._modeBar.element[0] ||
				(this._modeBar.element.find(e.target).length > 0 &&
					$(e.target).parentsUntil(this._modeBar.element, ".btn,.interactive").length === 0) ||
				e.target === this._navigationBar.element[0] ||
				(this._navigationBar.element.find(e.target).length > 0 &&
					$(e.target).parentsUntil(this._navigationBar.element, ".btn,.interactive,.mode").length === 0)
			) {
				e.preventDefault();
				e.stopPropagation();
			}
		}, this)
	);

	this._appContentWrapper.append(this._navigationBar.element);
	this._centerContent.append(this._modePanel.element);
	this._centerContentWidgets = $("<div>", {
		class: "center-content-widgets-wrapper"
	}).appendTo(this._centerContent);

	this._aboveContent = $("<div>", {
		class: "above-content"
	});

	this._registerListener(
		"layout.toggleAboveContent",
		function(show) {
			if (show) {
				this._aboveContent.addClass("above-content-visible");
			} else {
				this._aboveContent.removeClass("above-content-visible");
			}
		}.bind(this)
	);

	this._appContentWrapper.append(this._centerContent);
	this._appContentWrapper.append(this._aboveContent);

	this._appContentWrapper.append(this._messagebar);

	// by default load the view mode
	this._activeMode = new MobileViewMode(this.f, this);
	this._activeMode.toggle(true);

	// register all modes
	var modes = [
		this._activeMode,
		new MobileVideoMode(this.f, this),
		new AnnotationViewMode(this.f, this),
		new MobileReaderSpreadMode(this.f, this),
		new MobileCompareMode(this.f, this),
		new Mobile3DViewMode(this.f, this),
		new MobileCompare3DMode(this.f, this)
	];

	modes.forEach(
		function(mode) {
			this._layoutModel.registerLayoutMode(mode);
			mode.on(
				"changed",
				function() {
					this._updateModePanel();
					this._updateMainContent();
				}.bind(this)
			);
		}.bind(this)
	);

	this.f.emit("layout.finishRegisteringModes");

	//a debounced invalidate, to make sure not to many invalidates are triggered
	this._debouncedInvalidate = _.debounce(_.bind(this._invalidateLayout, this), 200);

	this._loadCss();

	this._cssLoaded = true;
	this._initialLoader.remove();
	delete this._initialLoader;
	this._context
		.append(this._modeBar.element)
		.append(this._appContentWrapper)
		.addClass("disableTextSelection")
		.append(this._loadingIndicator.element);
	this._debouncedInvalidate();

	//listen to resize events of the window
	$(window).on(
		"resize",
		_.bind(function(e) {
			// http://bugs.jqueryui.com/ticket/7514: resize of popups bubbles to window
			if (e.target === window) {
				this._debouncedInvalidate();
			}
		}, this)
	);

	//listen to widget events, to act on resized/collapsed/...
	this._registerOnLayoutEvents();

	// listen to the modebar
	this._registerListener("visible", this._debouncedInvalidate, this, this._modeBar);

	//listen to changes in the visibility of the mode panel
	this._registerListener("changed.visibility", this._invalidateLayout, this, this._modePanel);

	this._invalidateLayout(); //this one is needed to provide the correct world coordinates

	this._updateModePanel();

	// invalidate layout when modebar or modepanel transition is finished
	this._modeBar.element.on("webkitAnimationEnd oanimationend msAnimationEnd animationend transitionend ", this._debouncedInvalidate);
	this._modePanel.element.on("webkitAnimationEnd oanimationend msAnimationEnd animationend transitionend ", this._debouncedInvalidate);

	this._registerListener("changed.expanded", this._modePanelChanged, this, this._modePanel);
};

MobileLayout.prototype = Object.create(Layout.prototype);
MobileLayout.prototype.constructor = MobileLayout;

/**
 * Register listeners for general layout event.
 */
MobileLayout.prototype._registerOnLayoutEvents = function() {
	this.f.on(
		"layout.disableTextSelection",
		_.bind(function() {
			this._disableTextSelection();
		}, this)
	);

	this.f.on(
		"layout.enableTextSelection",
		_.bind(function() {
			this._enableTextSelection();
		}, this)
	);

	this.f.on("layout.resetToDefault", _.bind(this.resetToDefault, this));

	this.on(
		"changed.layoutMode",
		_.bind(function() {
			this._updateModePanel();
			this._updateMainContent();
		}, this)
	);
};

/**
 * Add an item to the layout. This is called by Layout to add an item.
 * @param {Object} layoutItem
 * @example
 * 	{ 	type: {LayoutConstants.LayoutType},
 * 		name: 'itemName',
 * 		item: {AbstractWidget/ToolbarComponent}
 * 	}
 */
MobileLayout.prototype._addItem = function(layoutItem) {
	// the mobile layout only supports the ViewportWidget for now...
	if (layoutItem.type === LayoutConstants.LayoutItemType.WIDGET) {
		var widget = layoutItem.getItem(this.getLayoutType());
		var widgetType = widget.getWidgetType();

		if (widgetType == "messagebar") {
			widget.setContext(this._messagebar);
		} else {
			var widgetWrapper = new WidgetWrapper(widget, layoutItem.name, false);
			this._items[layoutItem.name] = widgetWrapper;

			this._modePanel.element
				.find(".widgetWrapper:visible")
				.removeClass("first")
				.first()
				.addClass("first");
			this._modePanel.invalidate();

			this._registerListener(
				"enabled",
				function(isEnabled) {
					this._toggleWidget(layoutItem.name, widgetWrapper, isEnabled);
				},
				this,
				widget
			);

			this._registerListener("resized", this._invalidateModePanel, this, widget);

			if (widgetType == "maincontent" || this._activeMode.getMainPanelContent().indexOf(layoutItem.name) !== -1) {
				// show as main content
				widgetWrapper.setHeaderEnabled(false);
				this._centerContentWidgets.append(widgetWrapper.element);
				this._mainContentWidgetWrappers.push(widgetWrapper);
				this._updateMainContent();
			} else {
				var modePanelContent = this._activeMode.getModePanelContent();
				var nextItem;
				var index = _.indexOf(modePanelContent, layoutItem.name);
				while (index > -1 && index < modePanelContent.length && !nextItem) {
					nextItem = this._items[modePanelContent[++index]];
				}

				if (nextItem) {
					nextItem.element.before(widgetWrapper.element);
				} else if (index > -1) {
					this._modePanel.append(widgetWrapper.element);
				}
			}
		}
	} else if (layoutItem.type === LayoutConstants.LayoutItemType.TOOLBARCOMPONENT) {
		var toolbar = layoutItem.getItem(this.getLayoutType());
		var toolbarComponent = new ToolbarComponentWrapper(toolbar);
		this._items[layoutItem.name] = toolbarComponent;
		switch (toolbar.name) {
			case LayoutConstants.ToolbarComponent.APPROVAL:
			case LayoutConstants.ToolbarComponent.ANNOTATION:
				this._navigationBar.prependToolbarRight(toolbarComponent);
				break;
			case LayoutConstants.ToolbarComponent.VIDEO:
				this._aboveContent.append(toolbar.element);
			default:
				this._navigationBar.appendToolbarLeft(toolbarComponent);
		}
		this._updateItemVisibility();
	}
};

/**
 * Invalidate the ModePanel (update heights of the widgets)
 */
MobileLayout.prototype._invalidateModePanel = function() {
	this._modePanel.invalidate();
};

/**
 * Update visibility of the added items based on whether or not they are allowed in the current mode.
 * E.g. item was added in previous mode, but should not be shown in current mode
 */
MobileLayout.prototype._updateItemVisibility = function() {
	_.each(
		this._items,
		_.bind(function(item) {
			var layoutItemType = this._layoutModel.getLayoutItemType(item.name);
			if (layoutItemType === LayoutConstants.LayoutItemType.TOOLBARCOMPONENT) {
				item.setVisible(this._activeMode.isItemAllowed(item.name));
			} else if (layoutItemType === LayoutConstants.LayoutItemType.WIDGET) {
				item.setVisible(this._activeMode.isItemAllowed(item.name) && item.widget.isEnabled());
			}
		}, this)
	);
};

/**
 * Disable text selection in the entire viewer.
 */
MobileLayout.prototype._disableTextSelection = function() {
	this._context
		.find(".enableTextSelection")
		.toggleClass("enableTextSelection", false)
		.toggleClass("enableTextSelectionTempDisabled", true);
};

/**
 * Enable text selection in the entire viewer.
 */
MobileLayout.prototype._enableTextSelection = function() {
	this._context
		.find(".enableTextSelectionTempDisabled")
		.toggleClass("enableTextSelection", true)
		.toggleClass("enableTextSelectionTempDisabled", false);
};

/**
 * Restore a certain state of the layout to the state provided. (not used at this moment)
 * @param {Object} state
 */
MobileLayout.prototype.restore = function(state) {
	this._modeBar.setState(state);
};

/**
 * Reset the sate of the layout to the default state.
 */
MobileLayout.prototype.resetToDefault = function() {
	this._layoutState = this._getDefaultLayout(); //reset the state of the layout, it now is the default
	this._modeBar.setState(this._layoutState.modeBar);
	this._invalidateLayout();
};

/**
 * Get default layout state
 * @returns {Object}
 */
MobileLayout.prototype._getDefaultLayout = function() {
	return {
		modeBar: {
			visible: true
		}
	};
};

/**
 * Generates the state of the layout as a string. It takes the not-visible elements into account so state of those
 * elements is not lost.
 * @returns {Object}
 */
MobileLayout.prototype.getLayoutState = function() {
	return {
		modeBar: this._modeBar.getState()
	};
};

/**
 * Toggle the mode bar
 */
MobileLayout.prototype.toggleModeBar = function(toggle) {
	this._modeBar.toggle();
};

/**
 * Invalidates the layout. Makes sure all panels have the correct size, resize is triggered on the correct elements,...
 */
MobileLayout.prototype._invalidateLayout = function() {
	if (!this._cssLoaded) {
		return;
	}

	this._context.toggleClass("hideModeBar", !this._modeBar.isVisible());

	var height = (this._context.innerHeight() || null) - (this._navigationBar.element.outerHeight(true) || null);
	this._modePanel.element.css("height", height);

	_.each(
		this._activeMainPanelContent,
		function(widgetWrapper) {
			widgetWrapper.content.css({
				height: (this._context.innerHeight() || null) - (this._navigationBar.element.outerHeight(true) || null)
			});

			if (widgetWrapper.widget !== null) {
				widgetWrapper.widget.resized();
			}
		},
		this
	);

	// Layout loading indicator
	var navigationBarHeight = this._navigationBar.element.outerHeight(true) || null;
	this._loadingIndicator.element.css({
		top: navigationBarHeight + 5,
		left: (this._context.innerWidth() || null) - 29
	});

	// Center message bar inside content wrapper
	var contentPosition = this._appContentWrapper.position();
	this._messagebar.css({
		width: (this._appContentWrapper.outerWidth() || null) - 20,
		top: contentPosition.top + navigationBarHeight + 10,
		left: 10
	});

	this._modePanel.invalidate();

	this._emitLayoutChanged();
};

/**
 * Returns the type of the layout.
 * @override
 * @return {Number}
 */
MobileLayout.prototype.getLayoutType = function() {
	return LayoutConstants.LayoutType.MOBILE;
};

/**
 * Enable/disable fullscreen
 * Note: this method should be called from a user generated event... (or it won't work...)
 * @param enabled
 */
MobileLayout.prototype.toggleFullscreen = function(enabled) {
	if (enabled) {
		this._requestFullscreen();
	} else {
		this._exitFullscreen();
	}
};

/**
 * Check if the app is fullscreen
 * @returns {Boolean}
 */
MobileLayout.prototype.isFullscreen = function() {
	return FullscreenUtil.fullscreenElement;
};

/**
 * Check if the browser supports the fullescren API
 * @returns {Boolean}
 */
MobileLayout.prototype.supportsFullscreen = function() {
	return FullscreenUtil.fullscreenEnabled;
};

/**
 * Request fullscreen for the whole document.body
 */
MobileLayout.prototype._requestFullscreen = function() {
	var viewerDiv = this.f.getContext().get(0);
	FullscreenUtil.requestFullscreen(viewerDiv);
};

/**
 * Exit fullscreen mode
 */
MobileLayout.prototype._exitFullscreen = function() {
	FullscreenUtil.exitFullscreen();
};

/**
 * Show or hide mode panel, depending on the active mode
 */
MobileLayout.prototype._updateModePanel = function() {
	this._modePanel.setEnabled(this._activeMode.isModePanelEnabled());

	// update the modepanel content
	// hide existing items
	this._modePanel.element.find(".widgetWrapper").each(function(index, element) {
		element.wrapper.setVisible(false);
	});

	var modePanelContent = this._activeMode.getModePanelContent();
	var first = true;
	_.each(
		modePanelContent,
		function(itemName) {
			if (this._items[itemName]) {
				var item = this._items[itemName];
				this._modePanel.append(item.element);
				var visible = item.widget.isVisible() && item.widget.isEnabled();
				item.setVisible(visible);
				if (visible) {
					item.element.toggleClass("first", first);
					first = false;
				}
			}
		},
		this
	);
	this._modePanel.element.toggleClass("hideWidgetTitles", !this._activeMode.showModePanelWidgetTitles());
	this._modePanel.invalidate();
	this._modePanel.setExpanded(this._activeMode.isModePanelExpanded());
};

/**
 * Update the contents of the main panel
 * A mode can decide to show something else than the viewport widget
 */
MobileLayout.prototype._updateMainContent = function() {
	var mainPanelContents = this._activeMode.getMainPanelContent();
	this._activeMainPanelContent = [];

	// run through all possible main content widget wrappers
	_.each(
		this._mainContentWidgetWrappers,
		function(wrapper) {
			var index = mainPanelContents.indexOf(wrapper.widget.getID());
			if (index > -1) {
				// this wrapper should be visible
				wrapper.setVisible(true);
				wrapper.setHeaderEnabled(false);
				this._activeMainPanelContent.push(wrapper);
				wrapper.element.append(this._aboveContent);
				wrapper.element.css({
					order: index
				});
			} else {
				// hide, the wrapper should not be active
				wrapper.setVisible(false);
				wrapper.setHeaderEnabled(true);
			}
		},
		this
	);

	var activeMainPanelContentNames = this._activeMainPanelContent.map(function(wrapper) {
		return wrapper.widget.getID();
	});

	_.chain(mainPanelContents)
		.filter(
			function(mainPanelContent) {
				// The item was added but isn't a standard main content widget wrapper
				return this._items[mainPanelContent] && activeMainPanelContentNames.indexOf(mainPanelContent) === -1;
			}.bind(this)
		)
		.each(
			function(mainPanelContent) {
				// Add it to the main content widget wrapper
				this._mainContentWidgetWrappers.push(this._items[mainPanelContent]);
				this._centerContentWidgets.append(this._items[mainPanelContent].element);
			}.bind(this)
		)
		.tap(
			function(a) {
				// if there were widgets that weren't added yet,
				if (a.length > 0) {
					this._updateMainContent();
				} else {
					this._debouncedInvalidate();
				}
			}.bind(this)
		);
};

/**
 * React on collapse/expand operation of the modepanel
 * @param {Boolean} oldValue
 * @param {Boolean} newValue
 */
MobileLayout.prototype._modePanelChanged = function(oldValue, newValue) {
	this._activeMode.setIsmodePanelExpanded(newValue);
	this._debouncedInvalidate();
};

/**
 * Enable/disable a widget
 * @param {String}        widgetName    - name of the widget
 * @param {WidgetWrapper} widgetWrapper - widgetWrapper in which the widget is displayed
 * @param {Boolean}       isEnabled     - should the widget be enabled
 */
MobileLayout.prototype._toggleWidget = function(widgetName, widgetWrapper, isEnabled) {
	if (isEnabled) {
		if (this._activeMode.isItemAllowed(widgetName)) {
			widgetWrapper.setVisible(true);
		}
	} else {
		widgetWrapper.setVisible(false);
	}
};

/**
 * Returns the default layout mode for the layout.
 * @return {LayoutMode}
 */
MobileLayout.prototype._getDefaultLayoutMode = function() {
	return this._layoutModel.getLayoutMode(LayoutConstants.LayoutMode.VIEW2D);
};

/**
 * Defines the array of css files that should be loaded
 *
 * @returns Array<String> list of css files
 */
MobileLayout.prototype._loadCss = function() {
	require("./less/mobile.less");
};

/**
 * @override
 */
MobileLayout.prototype.getDefaultTool = function() {
	return (this._activeMode && this._activeMode.getDefaultTool()) || CoreConstants.TOOLS.TOUCH;
};

export default MobileLayout;
