import LayoutConstants from "@esko/web-viewer/layout/LayoutConstants";
import PrivateChannel from "@esko/web-viewer/core/PrivateChannel";
import ViewerToolbarButton from "@esko/web-viewer/layout/component/ViewerToolbarButton";
import _ from "underscore";
import $ from "jquery";
/**
 * @class ModeBar The ModeBar is a mobile specific component. It displays the available modes on the left side of the application, and enables the user to switch between them
 * @exports ModeBar
 * @memberof Layout
 * @constructor
 *
 * @param {Core.ViewerFacade} f
 * @param {JQueryObj} element - the element in which the mode bar should be rendered
 * @param {Layout.MobileLayout} mobileLayout - a reference to the mobile layout
 */
var ModeBar = function(f, element, mobileLayout) {
	this.f = f;
	PrivateChannel.installTo(this, f, "layout.mobile.modebar");

	this.element = element;
	this.element.addClass("modeBar");
	this._mobileLayout = mobileLayout;
	this._visible = true;

	/**
	 * Modes registered for the mode bar
	 * {mode: {layout.LayoutMode},
	 *  icon: {String},
	 *  name: {String}}
	 */
	this._availableModes = {};

	this.switchMode(LayoutConstants.LayoutMode.VIEW2D);
	this.renderModeBar();

	this.f.on(
		"layout.modeAvailabilityChanged",
		_.bind(function(mode, available, options) {
			if (available) {
				this._registerMode(mode, options.icon, options.title);
			} else {
				this._unregisterMode(mode.getName());
			}
		}, this)
	);

	mobileLayout.on("changed.layoutMode", _.bind(this.renderModeBar, this));
	this.f.loadBundle("viewer", this.renderModeBar, this);
};

/**
 * Toggle the mode bars visibility
 */
ModeBar.prototype.toggle = function() {
	this.setVisible(!this.isVisible());
};

/**
 * Change the visibility of the mode bar
 * @param {Boolean} visible
 */
ModeBar.prototype.setVisible = function(visible) {
	var oldValue = this._visible;
	this._visible = visible;
	if (oldValue !== visible) {
		this._emit("visible", visible);
	}
};

/**
 * Check if the mode bar is visible
 * @returns {Boolean}
 */
ModeBar.prototype.isVisible = function() {
	return this._visible;
};

/**
 * Get the width of the mode bar
 * @returns {Number} the width of the bar
 */
ModeBar.prototype.getWidth = function() {
	return this.isVisible() ? this.element.outerWidth(true) || null : 0;
};

/**
 * Get the state of the mode bar, used for restoring the layout state
 * @returns {Object}
 */
ModeBar.prototype.getState = function() {
	return {
		visible: this.isVisible()
	};
};

/**
 * Restore the state of the mode bar
 * @param {Object} state - state as returned by getState()
 */
ModeBar.prototype.setState = function(state) {
	this.setVisible(state.visible);
};

/**
 * Register a mode in the Mode bar
 */
ModeBar.prototype._registerMode = function(mode, icon, title) {
	this._availableModes[mode.getName()] = {
		mode: mode,
		icon: icon,
		title: title
	};
	this.renderModeBar();
};

/**
 * Unregister a mode in the Mode bar
 */
ModeBar.prototype._unregisterMode = function(name) {
	if (this._availableModes[name]) {
		delete this._availableModes[name];
	}
	this.renderModeBar();
};

/**
 * Render mode bar
 */
ModeBar.prototype.renderModeBar = function() {
	this.element.html("");
	_.each(
		LayoutConstants.MobileLayoutModeBarOrder,
		_.bind(function(modeName) {
			var mobileMode = this._availableModes[modeName];
			if (mobileMode) {
				var button = $("<div>").addClass("mode");
				if (mobileMode.mode.isActive()) {
					button.addClass("active");
				}
				button.append(
					$("<i>")
						.addClass(mobileMode.icon)
						.addClass("icon-large")
				);
				button.append($("<span>").html(this.f.i18n.t(mobileMode.title)));
				button.on(
					"touchstart pointerdown",
					_.bind(
						function(name) {
							this.switchMode(name);
						},
						this,
						mobileMode.mode.getName()
					)
				);
				this.element.append(button);
			}
		}, this)
	);
};

/**
 * Activate the mode by name
 * @param {String} name
 */
ModeBar.prototype.switchMode = function(name) {
	this.f.emit("layout.switchLayoutMode", name);
	this.renderModeBar();
};

export default ModeBar;
