import PrivateChannel from "@esko/web-viewer/core/PrivateChannel";
import MiniButton from "@esko/web-viewer/layout/component/MiniButton";
import LayoutUtils from "@esko/web-viewer/layout/LayoutUtils";
import ShapeUtils from "@esko/shapeutils";
import $ from "jquery";
import _ from "underscore";
/**
 * ModePanel interface element (for mobile, information of the mode can be displayed here), it is possible to collapse the modal panel.
 *
 * @constructor
 * @memberof Layout/mobile
 * @param {ViewerFacade} f
 * @param {String} name - name of the panel
 */
var ModePanel = function(f, name) {
	PrivateChannel.installTo(this, f, ["modePanel", name]);

	this.f = f;

	/**
	 * @member {String} name of the mode panel
	 */
	this.name = name;

	/**
	 * @member {JQueryObj} Element the mode panel uses to add the collapse button to
	 */
	this.element = $("<div/>").addClass("modePanel");

	/**
	 * @member {JQueryObj} div that contains the content of the mode panel
	 */
	this._content = $("<div/>");

	/**
	 * @member {Map.<String,JQueryObj>} keeps all available content and maps the name of the content to the content
	 */
	this._contentMap = {};

	/**
	 * @member {JQueryObj} content that is currently active in the mode panel
	 */
	this._activeContent = null;

	//append the content to the mode panel element
	this.element.append(this._content);

	/**
	 * @member {MiniButton} button used to collapse the mode panel
	 */
	this._collapseButton = new MiniButton(
		f,
		"collapseButton",
		"icon-WCR_Widget-Collapse",
		true,
		_.bind(function() {
			this._togglePanel();
		}, this)
	);

	/**
	 * @member {JQueryObj} div that has both the minibutton and the containing div-element
	 */
	this._collapseButtonElement = $("<div>")
		.addClass("collapseButton")
		.append(this._collapseButton.element);

	//append the collapse button
	this.element.append(this._collapseButtonElement);

	/**
	 * @member {Boolean} is the panel expanded or collapsed
	 */
	this._expanded = true;

	/**
	 * @member {Boolean} is the panel disabled or not
	 */
	this._disabled = false;

	/**
	 * Invalidate the panel, layout all element in the panel
	 */
	this.invalidate = _.debounce(_.bind(this._invalidate, this), 50);
};

/**
 * Collapse or Expand the mode panel
 * @param {Boolean} expanded
 */
ModePanel.prototype.setExpanded = function(expanded) {
	if (this._expanded !== expanded) {
		this._togglePanel();
	}
};

/**
 * Collapse or uncollapse the panel
 */
ModePanel.prototype._togglePanel = function() {
	var wasVisible = this.isVisible();

	this.element.toggleClass("collapsed");
	this._expanded = !this._expanded;

	this._collapseButton.setIcon(this._expanded ? "icon-WCR_Widget-Collapse" : "icon-WCR_Widget-Expand");
	var isVisible = this.isVisible();
	if (wasVisible != isVisible) {
		this._emit("changed.expanded", wasVisible, isVisible);
	}
};

/**
 * Is the mode panel disabled or enabled?
 * @returns {Boolean}
 */
ModePanel.prototype.isEnabled = function() {
	return !this._disabled;
};

/**
 * Is the mode panel visible or not
 * @returns {Boolean}
 */
ModePanel.prototype.isVisible = function() {
	return this._expanded && !this._disabled;
};

/**
 * Get the width of the mode panel
 * @returns {Number}
 */
ModePanel.prototype.getWidth = function() {
	if (this.isVisible()) {
		return this.element.outerWidth(true) || null;
	} else {
		return 0;
	}
};

/**
 * Get the height of the mode panel
 * @returns {Number}
 */
ModePanel.prototype.getHeight = function() {
	return this.element.outerHeight(true) || null;
};

/**
 * Enable/disable the mode panel
 * @param {Boolean} enabled
 */
ModePanel.prototype.setEnabled = function(enabled) {
	if (enabled) {
		this.enable();
	} else {
		this.disable();
	}
};

/**
 * Enable the mode panel (shows the panel)
 */
ModePanel.prototype.enable = function() {
	this.element.show();
	this._disabled = false;
};

/**
 * Disable the mode panel, hides it completely (including the hide/show button)
 */
ModePanel.prototype.disable = function() {
	this.element.hide();
	this._disabled = true;
};

/**
 * Invalidate the layout of the modepanel, resize all widgets
 */
ModePanel.prototype._invalidate = function() {
	if (this.isEnabled()) {
		LayoutUtils.layoutWidgets(this._content, 0);
	}
};

/**
 * Append an element to the panel
 * @param {JQueryObj} element - the object to add to the panel
 */
ModePanel.prototype.append = function(element) {
	this._content.append(element);
};

export default ModePanel;
