import ViewerToolbarButton from "@esko/web-viewer/layout/component/ViewerToolbarButton";
import LayoutConstants from "@esko/web-viewer/layout/LayoutConstants";
import ViewerUtils from "@esko/web-viewer/utils/ViewerUtils";
import _ from "underscore";
import $ from "jquery";
/**
 * @class NavigationBar The NavigationBar is a mobile specific component. It displays the navigation bar with options to show widgets, toggle the mode bar and see the document title
 * @exports NavigationBar
 * @memberof Layout
 * @constructor
 *
 * @param {Core.ViewerFacade} f
 * @param {JQueryObj} element - the element in which the navigation bar should be rendered
 * @param {Layout.MobileLayout} mobileLayout - a reference to the mobile layout
 */
var NavigationBar = function(f, element, mobileLayout) {
	this.f = f;
	this._mobileLayout = mobileLayout;

	/**
	 * Maps names of components to the context in which they are shown
	 */
	this._components = {};

	this.element = element;
	this.element.addClass("navigationBar");
	this._mobileLayout = mobileLayout;
	this._left = $("<div>")
		.addClass("navigationLeft")
		.appendTo(this.element);
	this._right = $("<div>")
		.addClass("navigationRight")
		.appendTo(this.element);

	/**
	 * @member {ViewerToolbarButton}
	 */
	this._hamburger = new ViewerToolbarButton(
		this.f,
		"menu",
		"icon-WCR_Hamburger",
		true,
		_.bind(function() {
			this._mobileLayout.toggleModeBar();
		}, this)
	);

	this._hamburger.element.addClass("hamburgerButton");
	this._left.append(this._hamburger.element);

	// show the exit icon if needed
	if (f.getProperty("returnurl")) {
		this._returnButton = new ViewerToolbarButton(
			this.f,
			"exit",
			"icon-WCR_Exit",
			true,
			_.bind(function() {
				this.f.emit("viewer.request.close", f.getProperty("returnurl"));
			}, this)
		);
		this._right.append(this._returnButton.element);
	}

	if (mobileLayout.supportsFullscreen()) {
		this._fullscreenButton = new ViewerToolbarButton(
			this.f,
			"fullscreen",
			"icon-WCR_FullScreen",
			true,
			_.bind(function() {
				mobileLayout.toggleFullscreen(!mobileLayout.isFullscreen());
				this._updateFullscreenButton();
			}, this)
		);

		$(window).on("resize", _.debounce(_.bind(this._updateFullscreenButton, this), 100));
		this._updateFullscreenButton();

		this._right.prepend(this._fullscreenButton.element);
	}

	// show the button to switch to desktop layout if the device is not mobile
	this.f.requestModule(
		"featureDetector",
		_.bind(function(featureDetector) {
			if (!featureDetector.isMobile() || !featureDetector.hasTouch()) {
				// not a mobile device, but the mobile layout was loaded
				this._switchLayoutModeButton = new ViewerToolbarButton(
					this.f,
					"switchToDesktop",
					"icon-WCR_Desktop",
					false,
					_.bind(function() {
						this.f.switchLayout(LayoutConstants.LayoutType.DESKTOP);
					}, this)
				);

				this._right.prepend(this._switchLayoutModeButton.element);
			}
		}, this)
	);

	this._mobileLayout.on("changed.layoutMode", _.bind(this._updateButtonVisibility, this));
};

/**
 * Update full screen button
 */
NavigationBar.prototype._updateFullscreenButton = function() {
	this._fullscreenButton.setIcon(this._mobileLayout.isFullscreen() ? "icon-WCR_NormalScreen" : "icon-WCR_FullScreen");
};

/**
 * Update the visibility of the goto desktop and full screen buttons based on the active mode
 */
NavigationBar.prototype._updateButtonVisibility = function() {
	var activeLayoutMode = this._mobileLayout.getActiveLayoutMode();
	if (activeLayoutMode) {
		// full screen button is available in view mode only
		if (this._fullscreenButton) {
			this._fullscreenButton.show();
		}
	} else {
		// hide full screen button in other modes
		if (this._fullscreenButton) {
			this._fullscreenButton.hide();
		}
	}
};

/**
 * Add toolbar at left side. Add to extra div, so they can be shown and hidden ignoring the module behavior.
 * e.g. toolbar can be hidden by module, so we don't want to show it when it becomes available
 * @param {ToolbarComponent} toolbar
 */
NavigationBar.prototype.appendToolbarLeft = function(toolbar) {
	var context = $("<div />");
	this._left.append(toolbar.element);
	this._components[toolbar.name] = toolbar.element;
};

/**
 * Add toolbar at right side. Add to extra div, so they can be shown and hidden ignoring the module behavior.
 * e.g. toolbar can be hidden by module, so we don't want to show it when it becomes available
 * @param {ToolbarComponent} toolbar
 */
NavigationBar.prototype.prependToolbarRight = function(toolbar) {
	var context = $("<div />");
	this._right.prepend(toolbar.element);
	this._components[toolbar.name] = toolbar.element;
};

/**
 * Toggle visibility of a component on the navigation bar.
 * @param {String} name - name of component
 * @param {Boolean} show - will show if true and hide if false.
 */
NavigationBar.prototype.toggleComponent = function(name, show) {
	if (this._components[name]) {
		if (show) {
			this._components[name].show();
		} else {
			this._components[name].hide();
		}
	}
};

export default NavigationBar;
