import _ from "underscore";
import AbstractTool from "@esko/web-viewer/viewport/tools/AbstractTool";
import ShapeUtils from "@esko/shapeutils";
import GEOUtils from "@esko/web-viewer/utils/GEOUtils";
import DrawUtils from "@esko/web-viewer/utils/DrawUtils";
import ViewerUtils from "@esko/web-viewer/utils/ViewerUtils";
import MeasureLayerModel from "@esko/web-viewer/measure/MeasureLayerModel";
import MeasureConstants from "@esko/web-viewer/measure/MeasureConstants";
import MeasurePopoverModel from "@esko/web-viewer/measure/MeasurePopoverModel";
import XMLUtils from "@esko/xmlutils";
import ConversionConstants from "@esko/web-viewer/utils/ConversionConstants";
import Cache from "@esko/cache";
/**
 * AbstractMeasureTool, defines the drawing functionality for the measure tools and the functionality to get the information for specific
 * measure tools from the viewserver.
 *
 * @exports AbstractMeasureTool
 * @memberof MeasureModule
 * @class AbstractMeasureTool
 * @augments Viewport.Tools.AbstractTool
 *
 * @param {Core.ViewerFacade} f
 * @param {String} IID
 */
var AbstractMeasureTool = function(f, IID) {
	AbstractTool.call(this, f, IID);

	this.f = f;

	this.popover = null;

	/**
	 * States
	 */
	this.defaultState = "default";
	this.drawingState = "drawing";
	this.clickState = "click";

	this.model = new MeasureLayerModel(f, IID);

	this.touchPoint = null;
	this.minimum = 16; //minimum that a user has to drag to start dragging (in pixels)

	this.snapDistance = 16; //distance of the mouse to snap to a linework

	this.infoFetchedFor = null;
	this.isPopoverShown = false;

	// initialize a cache for getting info from the server (this is to prevent flickering in blinking compare mode)
	this._infoCache = new Cache(4, true);
	this._infoCache._loadItem = _.bind(function(paramsString) {
		var page = this.f.getActivePage();
		var parameters = JSON.parse(paramsString); // stringified JSON object is used for the Cache's hash, parse it again to get the actual info
		page.getInfo(
			parameters,
			_.bind(function(info) {
				// show error to user when loading info fails
				if (info && info.status && info.status.value == "-1") {
					this.f.emit("userMessage", "warn", info.status.message);
				}

				// register item in cache (so callbacks will be executed)
				this._infoCache._registerItem(paramsString, info);
			}, this),
			_.bind(function(jqXHR, textStatus, errorThrown) {
				// ajax request failed, remove from cache
				this._infoCache._registerItem(paramsString);
				this.f.i18n.loadBundle(
					"viewer",
					function() {
						// show error
						this.f.emit("error.SEVERE", this.f.i18n.t("MeasurementInfoError"), jqXHR, textStatus, errorThrown);
					},
					this
				);
			}, this)
		);
	}, this);

	/**
	 * This is the mode when the tool is active,
	 * it will zoom on click and go the couldStartDraggingMode if left mousedown is clicked
	 */
	this.registerState(
		this.defaultState,
		(function(that) {
			return {
				enterState: function() {
					// do nothing
				},
				mousedown: function(e) {
					that.removePopover();
					that.model.reset();
					that.didReset();

					that.touchPoint = that.keepPointWithinWorldBounds(that.getWorldPoint(e), true);

					var point = that.getPointToSnapTo(e, that.snapDistance);
					if (!point) {
						point = that.keepPointWithinWorldBounds(that.getWorldPoint(e), false);
					}
					that.model.setStartPoint(point);
					that.model.hidePopover();
					that.changeState(that.drawingState);
					e.originalEvent.preventDefault();
				},
				mouseup: function(e) {
					if (that.model.ruler.startPoint && !that.model.ruler.endPoint) {
						that.changeState(that.clickState);
					} else {
						that.changeState(that.defaultState);
						if (that.model.ruler.startPoint && that.model.ruler.endPoint) {
							that.calculate();
						}
					}
				},
				mousemove: function(e) {
					that.checkForSnapPoint(e);
				},
				mouseout: function() {},
				contextmenu: function(e) {
					// prevent the context menu
					e.preventDefault();
					e.stopPropagation();
					return false;
				},
				mouseleave: function() {}
			};
		})(this)
	);

	this.registerState(
		this.drawingState,
		_.defaults(
			(function(that) {
				return {
					enterState: function() {
						f.emit("layout.disableTextSelection");
						f.emit("viewport.startInteractive");
					},
					leaveState: function() {
						setTimeout(function() {
							f.emit("layout.enableTextSelection");
						}, 500);
						f.emit("viewport.stopInteractive");
					},
					mousemove: function(e) {
						var size = that.keepPointWithinWorldBounds(that.getWorldPoint(e), true).sizeTo(that.touchPoint);
						if (that.model.ruler.endPoint || Math.abs(size.width) >= that.minimum || Math.abs(size.height) > that.minimum) {
							that.registerEndPoint(e);
						}
						that.checkForSnapPoint(e);
					}
				};
			})(this),
			this.getState(this.defaultState)
		)
	); //copies the not overridden attributes from defaultstate to this state

	this.registerState(
		this.clickState,
		_.defaults(
			(function(that) {
				return {
					enterState: function() {
						that.changeState(that.defaultState);
					}
				};
			})(this),
			this.getState(this.drawingState)
		)
	);

	// always show the crosshair cursor when active
	this.getDefaultCursor = function() {
		return "rulerCursor";
	};

	this._override("initialize", function(widget, toolManager) {
		this._super(widget, toolManager);
		this.changeState(this.defaultState);
		this.registerActiveEvents(["mousedown", "mousemove", "mouseup", "mouseout", "mouseleave", "contextmenu"]);
	});

	this._override("activate", function(interruptRequest) {
		this._super(interruptRequest);
		if (!interruptRequest) {
			this.f.getModule("viewport").addLayer(this.getIID(), this.model);
			if (this.model.ruler.startPoint && this.model.ruler.endPoint) {
				if (this.popover) {
					this.model.addPopoverModel(this.popover);
					this.f.emit("channels.informationAdded", this.popover.getChannelsInformation());
				}
			}
		}
	});

	this._override("deactivate", function(interruptRequest) {
		this._super(interruptRequest);
		if (!interruptRequest) {
			this.model.modelChanged(true);
			this.removePopover();
			this.f.getModule("viewport").removeLayer(this.getIID());
			f.emit("channels.informationNotVisible");
		}
	});
};

// extend from the AbstractTool
AbstractMeasureTool.prototype = new AbstractTool();
AbstractMeasureTool.prototype.constructor = AbstractMeasureTool;

/**
 * Abstract method to set the specific options for the call to the viewserver, calls getInfo()
 */
AbstractMeasureTool.prototype.calculate = function() {};

AbstractMeasureTool.prototype.endPointSet = function() {};

AbstractMeasureTool.prototype.didReset = function() {};

/**
 * Register the end point based on a mouse event. This will check if the endpoint is not near a snap point etc
 * @param {MouseEvent} e - mouse event
 */
AbstractMeasureTool.prototype.registerEndPoint = function(e) {
	var point = this.getPointToSnapTo(e, this.snapDistance);
	if (!point) {
		point = this.keepPointWithinWorldBounds(this.getWorldPoint(e), false);
	}

	this.model.setEndPoint(point, e.shiftKey);
	this.endPointSet();
};

/**
 * Register the start point based on a mouse event. This will check if the startpoint is not near a snap point etc
 * @param {MouseEvent} e - mouse event
 */
AbstractMeasureTool.prototype.registerStartPoint = function(e) {
	var point = this.getPointToSnapTo(e, this.snapDistance);
	if (!point) {
		point = this.keepPointWithinWorldBounds(this.getWorldPoint(e), false);
	}

	this.model.setStartPoint(point, e.shiftKey);
	this.endPointSet();
};

/**
 * Get a point to snap to, if needed
 * @param {Event} event - reference event
 * @param {Number} distance - maximum distance to the mouse to snap
 * @returns {ShapeUtils.Point}
 */
AbstractMeasureTool.prototype.getPointToSnapTo = function(e, distance) {
	return null;
};

/**
 * Check if there is a point to snap to
 * @param {ShapeUtils.Point} refpoint - reference point
 */
AbstractMeasureTool.prototype.checkForSnapPoint = function(e) {
	var point = this.getPointToSnapTo(e, this.snapDistance);
	this.model.setSnapIndicatorPoint(point);
};

/**
 * Abstract method to get the info from the received object and show it in a popover
 */
AbstractMeasureTool.prototype.showInfo = function(object, requestParameters) {};

/**
 * Gets the required information from the viewserver and converts the xml to an object before passing it to showInfo
 *
 * @param {Object} options
 */
AbstractMeasureTool.prototype.getInfo = function(options) {
	var parameters = this._generateParameters();
	var generateParams = _.clone(parameters);
	if (parameters != null) {
		_.each(options, function(value, key) {
			parameters[key] = value;
		});

		var paramsString = JSON.stringify(parameters); // Stringify the parameters as hash for the Cache.
		this._infoCache.get(
			paramsString,
			_.bind(function(info) {
				if (info) {
					// the info is available
					this.infoFetchedFor = generateParams;
					this.showInfo(info, parameters);
				}
			}, this)
		);
	}
};

/**
 * Generate the needed parameters for a data request
 * @returns {Object} the needed parameters
 */
AbstractMeasureTool.prototype._generateParameters = function() {
	if (this.model.ruler.startPoint && this.model.ruler.endPoint) {
		var rect = new ShapeUtils.Rect(this.model.ruler.startPoint.copy(), this.model.ruler.startPoint.sizeTo(this.model.ruler.endPoint));
		rect.normalize();

		//lib on viewserver expects this constant value for calculating barcodes and brailles
		var RESOLUTION = ConversionConstants.FIXED_BARCODE_DPMM; //300 dpi converted to dpmm;

		var context = {
			rectangle: rect,
			params: {
				tool: this.getIID(),
				x1: 0,
				y1: 0,
				x2: 0,
				y2: 0
			}
		};

		// ask all interceptors to update the context. All listeners have access to the context object and can adapt the information that is in the object
		// there is no predefined order in which these interceptors are called, their actions thus should not overlap!
		this.f.emit("intercept.fragmentURL", context);

		var data = context.params;
		data.res = RESOLUTION;

		return data;
	} else {
		return null;
	}
};

/**
 * @override
 */
AbstractMeasureTool.prototype.canBeInterrupted = function() {
	return this.getCurrentStateName() != this.drawingState;
};

/**
 * Can the tool be used on the currently active page?
 *
 * @returns {Boolean}
 */
AbstractMeasureTool.prototype.canBeUsedOnActivePage = function() {
	return true;
};

/**
 * Update measure information when it is deemed to be outdated
 * @param {Boolean} forceUpdate - forces the measure tool to update
 */
AbstractMeasureTool.prototype.updateIfNeeded = function(forceUpdate) {
	// on force or when we have already data and we have to update it
	if (forceUpdate || (this.infoFetchedFor && this.infoFetchedFor != this._generateParameters())) {
		this.calculate();
	}

	if (this.popover && this.isPopoverShown) {
		this.popover.update();
		this.model.modelChanged();
	}
};

/**
 * Show/update information popover
 * @param {Object} channelInfo
 * @param {LineworkWrapper} (area)
 * @param {JQuery.Element} headerExtra - element that is shown in the header of popover
 */
AbstractMeasureTool.prototype.showPopover = function(channelInfo, area, title, headerExtra) {
	if (!area) {
		area = this.model.ruler.getSquareLinework()[0];
	}

	if (!this.popover) {
		this.popover = this.createPopoverModel(channelInfo, area, title, headerExtra);
	} else {
		this.popover.setArea(area); // order matters!
		this.popover.setChannelsInfo(channelInfo);
		this.popover.setHeader(title);
		this.popover.setHeaderExtra(headerExtra);
	}

	// note as this popover keeps his ID, it should be added only once
	this.model.addPopoverModel(this.popover);
	this.isPopoverShown = true;

	this.model.modelChanged();
};

/**
 * Create a popover model for the tool
 * @param {Object} channelInfo
 * @param {LineworkWrapper} area
 * @param {String OR JQuery.Element} An object that is shown in the title (defaults to "Channels" when null)
 * @param {JQuery.Element} headerExtra - element that is shown in the header of popover
 * @returns {MeasurePopoverModel}
 */
AbstractMeasureTool.prototype.createPopoverModel = function(channelInfo, area, title, headerExtra) {
	return new MeasurePopoverModel(
		this.f,
		area,
		"Channels",
		title ? title : this.f.i18n.t("ChannelsTitle"),
		channelInfo,
		this._needsTotalCount(),
		headerExtra
	);
};

/**
 * Remove the popover
 */
AbstractMeasureTool.prototype.removePopover = function() {
	if (this.popover) {
		this.model.removePopoverModel(this.popover.getID());
	}
	this.isPopoverShown = false;
};

/**
 * Does this tool need a total count
 * @returns {Boolean} False
 */
AbstractMeasureTool.prototype._needsTotalCount = function() {
	return false;
};

/**
 * Get the state of the measure tool
 * @returns {Object}
 */
AbstractMeasureTool.prototype.getToolState = function() {
	var state = {
		startPoint: this.model.getStartPoint(),
		endPoint: this.model.getEndPoint(),
		isActive: this.isActive()
	};
	return state;
};

/**
 * Set the state of the measure tool
 * @param {Object} state
 */
AbstractMeasureTool.prototype.setToolState = function(state) {
	if (state.startPoint !== null) {
		this.model.setStartPoint(state.startPoint);
	}
	if (state.endPoint !== null) {
		this.model.setEndPoint(state.endPoint);
		this.endPointSet();
	}
};

/**
 * @override
 */
AbstractMeasureTool.prototype.doActivation = function() {
	this.model.modelChanged();
	AbstractTool.prototype.doActivation.call(this);
};

export default AbstractMeasureTool;
