import _ from "underscore";
import $ from "jquery";
import AbstractMeasureTool from "@esko/web-viewer/measure/AbstractMeasureTool";
import MeasureConstants from "@esko/web-viewer/measure/MeasureConstants";
import ViewerUtils from "@esko/web-viewer/utils/ViewerUtils";
import CoreConstants from "@esko/web-viewer/core/CoreConstants";
import Tooltip from "@esko/tooltip/Tooltip";
/**
 * Braille measure tool,  You can see the braille readout of your selected area
 *
 * @exports BrailleMeasureTool
 * @memberof MeasureModule
 * @class BrailleMeasureTool
 * @augments Measure.AbstractMeasureTool
 *
 * @param {Core.ViewerFacade} f
 * @param {MeasureModule} measureModule
 */
var BrailleMeasureTool = function(f, measureModule) {
	AbstractMeasureTool.call(this, f, CoreConstants.TOOLS.MEASURE_BRAILLE);

	this.model.setRulerType(MeasureConstants.MeasureType.Braille);

	if (MeasureConstants.BrailleLanguage[f.getOption("PREFERENCES_BRAILLE_LANGUAGE")]) {
		this.language = MeasureConstants.BrailleLanguage[f.getOption("PREFERENCES_BRAILLE_LANGUAGE")].url;
	} else {
		this.language = MeasureConstants.BrailleLanguage[MeasureConstants.LanguageDefault].url;
	}
	this.f = f;
	this.measureModule = measureModule;

	//Change the braille language
	f.getCurrentUser().on(
		"options.changed.PREFERENCES_BRAILLE_LANGUAGE",
		_.bind(function(oldValue, newValue) {
			this.language = MeasureConstants.BrailleLanguage[newValue].url;
			this.removePopover();
			this.calculate();
		}, this)
	);
};

// extend from the AbstractMeasureTool
BrailleMeasureTool.prototype = Object.create(AbstractMeasureTool.prototype);
BrailleMeasureTool.prototype.constructor = BrailleMeasureTool;

/**
 * @override
 *
 *  Gets the information from the to object converted xml and creates a MeasurePopoverModel that shows the information and tells the AbstractModel to add the popover
 *
 *  @param {Object} object
 *  @param {Object} requestParameters - parameters used to gather information
 */
BrailleMeasureTool.prototype.showInfo = function(object, requestParameters) {
	var passBrailleCheck = true;
	var containsBraille = false;
	var brailleChannels = [];
	if (object) {
		if (object.status.info) {
			var brailleInfo = object.status.info.braille;
			//possible that braille consists of multiple colors or is in both documents of side-by-side compare
			if (!(brailleInfo instanceof Array)) {
				brailleInfo = [brailleInfo];
			}
			//run over all braille information of different colors
			_.each(
				brailleInfo,
				function(braille) {
					//it is also possible that there are multiple regions of braille
					var regions = braille.xml.Regions.Braille;
					if (!(regions instanceof Array)) {
						//make sure that regions is an array
						regions = [regions];
					}
					//run over all braille regions
					_.each(
						regions,
						function(region, index) {
							containsBraille = true;
							if (region.attributes.Status !== "Pass") {
								passBrailleCheck = false;
							}

							var regionIndex = "";
							//if there are multiple regions make sure the region index is places before each line number
							if (regions.length > 1) {
								regionIndex = index + 1 + ",";
							}
							if (brailleChannels[braille.color] === undefined) {
								brailleChannels[braille.color] = "";
							}
							//when more lines of braille are detected display the index of the line
							if (region.Line instanceof Array) {
								_.each(region.Line, function(brailleLine, lineIndex) {
									if (lineIndex >= 1) {
										brailleChannels[braille.color] += " ";
									}
									brailleChannels[braille.color] += "[" + regionIndex + (lineIndex + 1) + "] " + brailleLine;
								});
							} else if (regions.length > 1) {
								//if there are multiple region always add region number and line number (even if only one line)
								brailleChannels[braille.color] += "[" + regionIndex + "1" + "] " + region.Line + " ";
							} else {
								//only one region and one line
								brailleChannels[braille.color] += region.Line;
							}
						},
						this
					);
				},
				this
			);
		}
		this.f.emit("channels.informationAdded", brailleChannels);
	}

	var icon = null;
	if (containsBraille) {
		icon = passBrailleCheck ? this._createPassIcon() : this._createFailIcon();
	}

	var title = $("<span>")
		.append($("<i>").addClass("icon icon-normal icon-WCR_Measure-Braille"))
		.append(
			$("<span>")
				.addClass("titleText")
				.text(requestParameters.brlang)
		);
	this.showPopover(brailleChannels, null, title, icon);
};

/**
 * @override
 *
 * Sets the required options for the call to the viewserver
 */
BrailleMeasureTool.prototype.calculate = function() {
	var viewServerLang = this.language;
	if (viewServerLang === undefined || viewServerLang === MeasureConstants.BrailleLanguage[MeasureConstants.LanguageAutomatic].url) {
		// select the language automatically, default to English
		viewServerLang = MeasureConstants.BrailleLanguage[MeasureConstants.LanguageLocales[this.f.getCurrentUser().getLocale()]].url;
		if (viewServerLang === undefined) {
			viewServerLang = MeasureConstants.BrailleLanguage[MeasureConstants.LanguageDefault].url;
		}
	}
	var options = {
		brlang: viewServerLang
	};

	this.getInfo(options);
};

BrailleMeasureTool.prototype.canBeUsedOnActivePage = function() {
	var page = this.f.getActivePage();
	return page && page.hasGraphics();
};

/**
 * Creates a success icon with appropriate tooltip (used to display braille check results)
 * @private
 * @returns {jQuery.Element} the icon
 */
BrailleMeasureTool.prototype._createPassIcon = function() {
	var icon = $("<i>").addClass("icon icon-normal icon icon-WCR_Msg-Success-round");
	var wrapper = $("<div>").append(icon);
	new Tooltip(this.f.i18n, icon, wrapper, this.f.getContext(), false, "viewer", "BraillePass");
	return wrapper;
};

/**
 * Creates a fail icon with appropriate tooltip (used to display braille check results)
 * @private
 * @returns {jQuery.Element} the icon
 */
BrailleMeasureTool.prototype._createFailIcon = function() {
	var icon = $("<i>").addClass("icon icon-normal icon icon-WCR_Msg-Warning");
	var wrapper = $("<div>").append(icon);
	new Tooltip(this.f.i18n, icon, wrapper, this.f.getContext(), false, "viewer", "BrailleFail");
	return wrapper;
};

export default BrailleMeasureTool;
