import _ from "underscore";
import ViewportLayerModel from "@esko/web-viewer/viewport/model/ViewportLayerModel";
import CaliperPopoverModel from "@esko/web-viewer/measure/CaliperPopoverModel";
import DrawUtils from "@esko/web-viewer/utils/DrawUtils";
import ShapeUtils from "@esko/shapeutils";
import LineworkWrapper from "@esko/web-viewer/utils/LineworkWrapper";
import LineworkPoint from "@esko/web-viewer/utils/LineworkPoint";
import ListenerUtil from "@esko/web-viewer/utils/ListenerUtil";
import MeasureConstants from "@esko/web-viewer/measure/MeasureConstants";
/**
 * @exports CaliperLayerModel
 * @memberof MeasureModule
 * @augments ViewportLayerModel
 * @mixes {Utils.ListenerUtil}
 * @class This class gives the linework to render for the caliper tool.
 * @constructor
 *
 * @param {Core.ViewerFacade} f - Viewer facade
 * @param {String} ID - the ID of the layer model
 * @param {CaliperModel} caliperModel - Model of the caliper tool
 */
var CaliperLayerModel = function(f, ID, caliperModel) {
	ViewportLayerModel.call(this, f, ID);
	ListenerUtil.mixin(this, f);

	// Variables
	this._caliperModel = caliperModel;
	this._caliperPopover = null;
	this._popoverFrame = new ShapeUtils.Rect();
	/**
	 * {ShapeUtils.Point}
	 * The start point in world coordinates (so in mm).
	 * The end point/line is calculated as start point/line + gap of the current gap in the caliper model.
	 */
	this._startPoint = null;

	// Listen to the model
	this._registerListener("changed.currentGap", this.modelChanged, this, this._caliperModel);
	this._registerListener("changed.currentGap.vector", this.modelChanged, this, this._caliperModel);
};

// inherit
CaliperLayerModel.prototype = Object.create(ViewportLayerModel.prototype);
// correct the constructor pointer
CaliperLayerModel.prototype.constructor = CaliperLayerModel;

/**
 * Set a new start point
 * @param {ShapeUtils.Point} startPoint - The new start point
 */
CaliperLayerModel.prototype.setStartPoint = function(startPoint) {
	this._startPoint = startPoint;
	this.modelChanged();
};

/**
 * Get the current start point
 * @returns {ShapeUtils.Point}
 */
CaliperLayerModel.prototype.getStartPoint = function() {
	return this._startPoint;
};

/**
 * @override
 */
CaliperLayerModel.prototype.getLinework = function(animated, rotationanimation, drawrect) {
	if (this._startPoint) {
		var startLine = new LineworkWrapper(); // first line
		var endLine = new LineworkWrapper(); // second line
		var beginRect = new LineworkWrapper(); // rect before start line
		var endRect = new LineworkWrapper(); // rect after end line

		var vector = this._caliperModel.getCurrentGap().getVector();
		var gap = vector.getLength();
		var angle = vector.getAngleInDegrees();

		if (this._caliperModel.isMeasuringHeight(vector)) {
			// measuring heights

			if (angle === 0 || angle === 90) {
				// vector pointed to the top but we are drawing from top,left
				gap = -gap;
			}

			// upper rect
			var upperRect = new ShapeUtils.Rect(
				new ShapeUtils.Point(drawrect.left(), drawrect.top()),
				new ShapeUtils.Size(drawrect.right() - drawrect.left(), Math.min(this._startPoint.y, this._startPoint.y + gap) - drawrect.top())
			);
			upperRect.normalize();
			beginRect.rect = upperRect;

			// lower rect
			var max = Math.max(this._startPoint.y, this._startPoint.y + gap);
			var lowerRect = new ShapeUtils.Rect(
				new ShapeUtils.Point(drawrect.left(), max),
				new ShapeUtils.Size(drawrect.right() - drawrect.left(), drawrect.bottom() - max)
			);
			lowerRect.normalize();
			endRect.rect = lowerRect;

			// first line
			var startLineBegin = new LineworkPoint();
			startLineBegin.makeMoveTo(drawrect.left(), this._startPoint.y);
			startLine.pointArray.push(startLineBegin);

			var startLineEnd = new LineworkPoint();
			startLineEnd.makeLineTo(drawrect.right(), this._startPoint.y);
			startLine.pointArray.push(startLineEnd);

			// second line
			var endLineBegin = new LineworkPoint();
			endLineBegin.makeMoveTo(drawrect.left(), this._startPoint.y + gap);
			endLine.pointArray.push(endLineBegin);

			var endLineEnd = new LineworkPoint();
			endLineEnd.makeLineTo(drawrect.right(), this._startPoint.y + gap);
			endLine.pointArray.push(endLineEnd);

			this._popoverFrame.origin.x = drawrect.left();
			this._popoverFrame.origin.y = this._startPoint.y + gap;
			this._popoverFrame.size.width = drawrect.size.width;
			this._popoverFrame.size.height = 0.00000001;
		} else if (this._caliperModel.isMeasuringWidth(vector)) {
			// measuring widths

			if (angle === 180) {
				// pointed to the left
				gap = -gap;
			}

			// left rect
			var leftRect = new ShapeUtils.Rect(
				new ShapeUtils.Point(drawrect.left(), drawrect.top()),
				new ShapeUtils.Size(Math.min(this._startPoint.x, this._startPoint.x + gap) - drawrect.left(), drawrect.bottom() - drawrect.top())
			);
			leftRect.normalize();
			beginRect.rect = leftRect;

			// right rect
			var max = Math.max(this._startPoint.x, this._startPoint.x + gap);
			var rightRect = new ShapeUtils.Rect(
				new ShapeUtils.Point(max, drawrect.top()),
				new ShapeUtils.Size(drawrect.right() - max, drawrect.bottom() - drawrect.top())
			);
			rightRect.normalize();
			endRect.rect = rightRect;

			// first line
			var startLineBegin = new LineworkPoint();
			startLineBegin.makeMoveTo(this._startPoint.x, drawrect.top());
			startLine.pointArray.push(startLineBegin);

			var startLineEnd = new LineworkPoint();
			startLineEnd.makeLineTo(this._startPoint.x, drawrect.bottom());
			startLine.pointArray.push(startLineEnd);

			// second line
			var endLineBegin = new LineworkPoint();
			endLineBegin.makeMoveTo(this._startPoint.x + gap, drawrect.top());
			endLine.pointArray.push(endLineBegin);

			var endLineEnd = new LineworkPoint();
			endLineEnd.makeLineTo(this._startPoint.x + gap, drawrect.bottom());
			endLine.pointArray.push(endLineEnd);

			this._popoverFrame.origin.x = this._startPoint.x + gap;
			this._popoverFrame.origin.y = drawrect.top();
			this._popoverFrame.size.width = 0.00000001;
			this._popoverFrame.size.height = drawrect.size.height;
		}

		// Properties
		startLine.dash = [4, 3];
		startLine.color = "rgba(210,210,210,0.9)";
		startLine.type = DrawUtils.LWTYPE_LINE;

		endLine.dash = [4, 3];
		endLine.color = "rgba(210,210,210,0.9)";
		endLine.type = DrawUtils.LWTYPE_LINE;

		beginRect.type = DrawUtils.LWTYPE_RECT;
		beginRect.fillcolor = "rgba(0, 0, 0, 0.40)";
		beginRect.color = "rgba(0, 0, 0, 1.00)";
		beginRect.linewidth = 1;

		endRect.type = DrawUtils.LWTYPE_RECT;
		endRect.fillcolor = "rgba(0, 0, 0, 0.40)";
		endRect.color = "rgba(0, 0, 0, 1.00)";
		endRect.linewidth = 1;

		return [beginRect, endRect, startLine, endLine];
	}
};

/**
 * Get the end point.
 * @returns {ShapeUtils.Point} - The point on the end line as start point + gap (vector)
 */
CaliperLayerModel.prototype.getEndPoint = function() {
	if (this._startPoint) {
		var vector = this._caliperModel.getCurrentGap().getVector();
		var gap = vector.getLength();
		var angle = vector.getAngleInDegrees();

		// calculate end point
		if (this._caliperModel.isMeasuringHeight(vector)) {
			// measuring heights
			if (angle === 0 || angle === 90) {
				// pointed to the top but we are drawing from top,left
				gap = -gap;
			}
			return new ShapeUtils.Point(this._startPoint.x, this._startPoint.y + gap);
		} else if (this._caliperModel.isMeasuringWidth(vector)) {
			// measuring widths

			if (angle === 180) {
				// pointed to the left
				gap = -gap;
			}
			return new ShapeUtils.Point(this._startPoint.x + gap, this._startPoint.y);
		}
	} else {
		return null;
	}
};

/**
 * Set new start and end point.
 * Either start point can be changed or end point (other one given by the getEndPoint and getStartPoint methods), but not both
 * @param {ShapeUtils.Point} startPoint New start point in world coordinates
 * @param {ShapeUtils.Point} endPoint - New end point in world coordinates
 * @param {Boolean} keepVector - True to keep the current vector (e.g. for fixed gaps)
 * @param {Boolean} [keepAngle] - if true, the angle won;t be changed
 */
CaliperLayerModel.prototype.setPoints = function(startPoint, endPoint, keepVector, keepAngle) {
	var oldStartPoint = this.getStartPoint();
	var oldEndPoint = this.getEndPoint();
	var currentVector = this._caliperModel.getCurrentGap().getVector();

	if (!keepVector) {
		// Vector can be changed (e.g. for freeform, depending on the largest x/y distance)
		var angle = 0;
		var gap = 0;

		var distX = endPoint.x - startPoint.x;
		var distY = endPoint.y - startPoint.y;
		var distXAbs = Math.abs(distX);
		var distYAbs = Math.abs(distY);
		var shouldMeasureWidth = distXAbs > distYAbs;

		if (keepAngle) {
			shouldMeasureWidth = this._caliperModel.isMeasuringWidth(currentVector);
		}

		if (shouldMeasureWidth) {
			// we are measuring widths

			// change orientation if needed
			if (distX < 0) {
				// to the left of the begin point
				angle = 180;
			} else {
				// to the right of the begin point
				angle = 0;
			}
			gap = distXAbs;
		} else {
			// we are measuring heights

			// change orientation if needed
			if (distY < 0) {
				// more to the top than begin point
				angle = 90;
			} else {
				// more to the bottom than begin point
				angle = -90;
			}
			gap = distYAbs;
		}
		currentVector.setAngleInDegrees(angle);
		currentVector.setLength(gap);
	} else {
		// Vector is kept
		// Check if the end point changed -> change start point accordingly
		var distEndY = endPoint.y - oldEndPoint.y;
		var distEndX = endPoint.x - oldEndPoint.x;

		if (distEndY !== 0) {
			startPoint = new ShapeUtils.Point(oldStartPoint.x, oldStartPoint.y + distEndY);
		} else if (distEndX !== 0) {
			startPoint = new ShapeUtils.Point(oldStartPoint.x + distEndX, oldStartPoint.y);
		}
	}

	// start point changed or not?
	if (!startPoint.equals(oldStartPoint)) {
		this._startPoint = startPoint;
	}

	this._caliperModel.setCurrentGapVector(currentVector);
	this.modelChanged();
};

/**
 * Start showing the caliper popover
 */
CaliperLayerModel.prototype.showPopover = function() {
	if (!this._caliperPopover) {
		this._caliperPopover = new CaliperPopoverModel(this.f, null, "CaliperPopover", this);
		this.modelChanged();

		this.addPopoverModel(this._caliperPopover);
	}
};

/**
 * Hide the caliper popover
 */
CaliperLayerModel.prototype.hidePopover = function() {
	if (this._caliperPopover) {
		this.removePopoverModel(this._caliperPopover.getID());
		this.removePopover(this._caliperPopover.getID());
		this._caliperPopover = null;
	}
};

/**
 * Override the modelchanged, so the popovermodel is updated when needed. Also sends out current selection.
 * @param {Boolean} deactivate - True when tool using this model gets deactivated (see CaliperTool)
 * @override
 */
CaliperLayerModel.prototype.modelChanged = function(deactivate) {
	var selectionObject = null;

	// deactivate can also be other variables, so when it is other variable,or when it is a boolean and it is false (= activate) execute block
	if (!_.isBoolean(deactivate) || (_.isBoolean(deactivate) && !deactivate)) {
		// first update the popover
		if (this._caliperPopover) {
			this._caliperPopover.setGap(this._caliperModel.getCurrentGap());
			this._caliperPopover.setEndPoint(this.getEndPoint());
			this._caliperPopover.setMeasuringHeight(this._caliperModel.isMeasuringHeight(this._caliperModel.getCurrentGap().getVector()));
			this._caliperPopover.setFrame(this._popoverFrame);
		}

		// created updated updated selection object
		if (
			this._startPoint &&
			this._caliperModel.getCurrentGap() &&
			this._caliperModel
				.getCurrentGap()
				.getVector()
				.getLength() !== 0
		) {
			if (this._caliperModel.isMeasuringHeight(this._caliperModel.getCurrentGap().getVector())) {
				selectionObject = {
					mode: MeasureConstants.MeasureDrawType.HEIGHT,
					y: this._startPoint.y,
					h: this._caliperModel
						.getCurrentGap()
						.getVector()
						.getLength()
				};
			} else {
				selectionObject = {
					mode: MeasureConstants.MeasureDrawType.WIDTH,
					x: this._startPoint.x,
					w: this._caliperModel
						.getCurrentGap()
						.getVector()
						.getLength()
				};
			}
		}
	}

	// structure of selectionObject: see {DocumentInfoModel.setSelection}
	this.f.emit("selection.changed", selectionObject);

	// make sure the normal model changed behavior is still executed
	ViewportLayerModel.prototype.modelChanged.call(this);
};

export default CaliperLayerModel;
