import AbstractModel from "@esko/web-viewer/core/AbstractModel";
import CaliperGap from "@esko/web-viewer/measure/CaliperGap";
import MeasureConstants from "@esko/web-viewer/measure/MeasureConstants";
import ShapeUtils from "@esko/shapeutils";
import _ from "underscore";
/**
 * @exports CaliperModel
 * @memberof MeasureModule
 * @class The Caliper model is used for representing all caliper tool gap info
 * @constructor
 * @augments {Core.AbstractModel}
 *
 * @param {Core.ViewerFacade} f - Viewer facade
 * @param {String} privateChannelName - The privateChannelName of the model
 */
var CaliperModel = function(f, privateChannelName) {
	AbstractModel.call(this, f, privateChannelName);

	// ids of the different gaps (for this session only)
	// These ids are not saved to the view server. They are just used to identify the different gaps.
	this._nextId = 0;

	// default freeform vector
	this._freeformVector = new ShapeUtils.Vector(0, 0, -Math.PI / 2); // a zero vector with an angle of 90 degrees
	this._freeform = new CaliperGap(this._getNextId(), MeasureConstants.CaliperStyle.Freeform, this._freeformVector);

	// default fixed vector
	var fixedVector = new ShapeUtils.Vector(0, 0.8);
	this._fixed = new CaliperGap(this._getNextId(), MeasureConstants.CaliperStyle.Fixed, fixedVector);

	// array of CaliperGaps
	this._presetGaps = [];
	this._readPresets(); // read them from the view server

	// current active gap
	this._currentGap = null;
};

CaliperModel.prototype = Object.create(AbstractModel.prototype);
CaliperModel.prototype.constructor = CaliperModel;

/**
 * Reset the vector of the current caliper gap if it's freeform to its default value
 */
CaliperModel.prototype.resetFreeformVector = function() {
	var currentGap = this.getCurrentGap();
	if (currentGap && currentGap.getStyle() === MeasureConstants.CaliperStyle.Freeform) {
		// a freeform selection is active, reset the vector
		this.setCurrentGapVector(this._freeformVector);
	}
};

/**
 * Get the next id.
 * @returns {String} The next id
 */
CaliperModel.prototype._getNextId = function() {
	var id = this._nextId;
	this._nextId++;
	return "" + id;
};

/**
 * Get the freeform gap
 * @returns {CaliperGap} The freeform gap.
 */
CaliperModel.prototype.getFreeformGap = function() {
	return this._freeform;
};

/**
 * Get the fixed gap
 * @returns {CaliperGap}
 */
CaliperModel.prototype.getFixedGap = function() {
	return this._fixed;
};

/**
 * Get all gaps, including freeform, fixed and all presets.
 * @returns {Array<CaliperGap>} Array of all gaps.
 */
CaliperModel.prototype.getAllGaps = function() {
	var result = [this._freeform, this._fixed];
	var presets = _.sortBy(this._presetGaps, function(presetGap) {
		// first sort by angle, then by length
		return Math.abs(presetGap.getVector().getAngleInDegrees()) * 10000 + presetGap.getVector().getLength();
	});

	return _.union(result, presets);
};

/**
 * Get the current active gap.
 * @returns {CaliperGap} Current active gap
 */
CaliperModel.prototype.getCurrentGap = function() {
	return this._currentGap;
};

/**
 * Get a gap by id
 * @param {String} id - Id of the gap
 * @returns {CaliperGap} The caliper gap with the given id or undefined if not found.
 */
CaliperModel.prototype.getGapById = function(id) {
	return _.find(this.getAllGaps(), function(gap) {
		return gap.getId() === id;
	});
};

/**
 * Set the current gap by id
 * @param {String} id - Id of the gap to set
 */
CaliperModel.prototype.setCurrentGapById = function(id) {
	var gap = this.getGapById(id);
	if (gap) {
		var oldGap = this._currentGap;
		this._currentGap = gap;
		this._fireChange(false, ["currentGap"], oldGap, this._currentGap);
	} else {
		this.f.debug.error("Gap with id " + id + " not found");
	}
};

/**
 * Set the vector of the current active gap.
 * @param {ShapeUtils.Vector} newGapVector - The new gap vector
 */
CaliperModel.prototype.setCurrentGapVector = function(newGapVector) {
	var oldVector = this._currentGap.getVector();
	this._currentGap.setVector(newGapVector);
	this._fireChange(false, ["currentGap", "vector"], oldVector, newGapVector);
};

/**
 * Save the current active gap as preset to the view server.
 */
CaliperModel.prototype.saveCurrentGapAsPreset = function() {
	var newFixedGap = new CaliperGap(this._getNextId(), MeasureConstants.CaliperStyle.Preset, this._currentGap.getVector());
	this._presetGaps.push(newFixedGap);

	this._persistPresets();

	this._fireChange(false, ["presetGaps", "new"], null, newFixedGap);
};

/**
 * Read the presets from the view server and put them in presetGaps.
 */
CaliperModel.prototype._readPresets = function() {
	var vectorsString = this.f.getOption("CALIPER_PRESETS");
	if (vectorsString) {
		var vectors = JSON.parse(vectorsString);
		this._presetGaps = _.map(
			vectors,
			function(vector) {
				return new CaliperGap(this._getNextId(), MeasureConstants.CaliperStyle.Preset, new ShapeUtils.Vector(vector.x, vector.y));
			},
			this
		);
	}
};

/**
 * Persist the presets to the view server.
 */
CaliperModel.prototype._persistPresets = function() {
	// only vectors
	var vectors = _.map(this._presetGaps, function(gap) {
		return gap.getVector();
	});
	var jsonnedPresets = JSON.stringify(vectors);
	this.f.setOption("CALIPER_PRESETS", jsonnedPresets);
};

/**
 * Delete the current active gap from the presets.
 */
CaliperModel.prototype.deleteCurrentGapFromPresets = function() {
	if (this._currentGap.getStyle() !== MeasureConstants.CaliperStyle.Preset) {
		this.f.debug.error("Cannot delete a non-preset style");
	}
	this._presetGaps = _.without(this._presetGaps, this._currentGap);

	this._persistPresets();

	this._fireChange(false, ["presetGaps", "deleted"], this._currentGap, null);
};

/**
 * Check whether a preset with the same vector as similarPreset already exists.
 * @param {CaliperGap} similarPreset - The preset to check with the existing presets
 * @returns {CaliperGap} The preset that is similar to the given similarPreset or undefined if not found
 */
CaliperModel.prototype.getSimilarPreset = function(similarPreset) {
	return _.find(this._presetGaps, function(presetGap) {
		return presetGap.isSimilarTo(similarPreset);
	});
};

/*
 * UTILITY FUNCTIONS
 */

/**
 * Check if the given vector is representing a height measurement
 * This depends on the angle of the vector
 * @param {ShapeUtils.Vector} vector - The vector to check.
 * @returns {Boolean} True if measuring heights, false otherwise.
 */
CaliperModel.prototype.isMeasuringHeight = function(vector) {
	var angle = vector.getAngleInDegrees();

	return angle === 90 || angle === -90;
};

/**
 * Check if the given vector is representing a width measurement
 * This depends on the angle of the vector
 * @param {ShapeUtils.Vector} vector - The vector to check.
 * @returns {Boolean} True if measuring widths, false otherwise
 */
CaliperModel.prototype.isMeasuringWidth = function(vector) {
	var angle = vector.getAngleInDegrees();

	return angle === 0 || angle === 180;
};

export default CaliperModel;
