import ViewportPopoverModel from "@esko/web-viewer/viewport/model/ViewportPopoverModel";
import tpl from "@esko/web-viewer/measure/CaliperPopoverTemplate.tpl.html";
import ShapeUtils from "@esko/shapeutils";
import ViewerUtils from "@esko/web-viewer/utils/ViewerUtils";
import Template from "@esko/template";
import CoreConstants from "@esko/web-viewer/core/CoreConstants";
import _ from "underscore";
import $ from "jquery";

/**
 * PoveroverModel for the measure tools, defines the layout (uses the the MeasurePopoverTemplate) of the Popover for measure tools
 *
 * @exports CaliperPopoverModel
 * @memberof MeasureModule
 * @class CaliperPopoverModel
 * @augments Viewport.Model.ViewportPopoverModel
 * @constructor
 *
 * @param {Core.ViewerFacade} f
 * @param {LineworkWrapper} area
 * @param {String} name
 * @param {Array<Number> | Array<String>} channels
 * @param {Boolean} [makeTotalCount] Defines if a total of the different channel values needs to be calculated, only possible if the
 * 										values of the channels array are numbers
 * @param {String} unit The String to be displayed behind the the values of the different channels
 * @param {CaliperLayerModel} caliperLayerModel
 */
var CaliperPopoverModel = function(f, area, name) {
	ViewportPopoverModel.call(this, f, name);

	this.div = $("<div/>").addClass("measureInfo caliperPopover");

	// initialize template
	this.t = new Template(f.i18n, tpl);
	this.t.setContext(this.div);

	this.t.d.isMeasuringHeight = true;
	this._endPoint = new ShapeUtils.Point();
	this._endPointValueInMm = null;
	this._gapinMm = null;

	// load language if needed
	f.i18n.loadBundle(
		"viewer",
		_.bind(function() {
			this.t.render();
		}, this)
	);

	// listen to unit changes
	this.f.getCurrentUser().on("options.changed.PREFERENCES_UNIT_SIZE", _.bind(this.recalculateGap, this));

	this.frame = null;

	this.getFrame = function() {
		if (this.frame) {
			return this.frame;
		} else {
			return new ShapeUtils.Rect(new ShapeUtils.Point(0, 0), new ShapeUtils.Size(1, 1));
		}
	};

	this.getContent = function() {
		return this.div;
	};

	this.getPreferredPosition = function() {
		// override
		return this.t.d.isMeasuringHeight ? "bottom" : "right";
	};

	this.getRestrictBorder = function() {
		return 0;
	};

	/**
	 * Is this an interactive popover
	 * @returns {Boolean} True
	 */
	this.isInteractive = function() {
		return true;
	};

	this.getContext = function() {
		return this.f.getContext().find(".viewport");
	};

	this.render = _.throttle(
		_.bind(function() {
			this.t.render();
		}, this),
		100
	);

	/**
	 * Get the class for the popover
	 */
	this.getClass = function() {
		return "transparent";
	};
};

// inherit
CaliperPopoverModel.prototype = Object.create(ViewportPopoverModel.prototype);
// correct the constructor pointer
CaliperPopoverModel.prototype.constructor = CaliperPopoverModel;

CaliperPopoverModel.prototype.getChannelsInformation = function() {
	return this._channelsInfo;
};

/**
 * Update the popover
 */
CaliperPopoverModel.prototype.setGap = function(caliperGap) {
	this._gapinMm = caliperGap.getVector().getLength();
	this.t.d.gap = this._convertToCorrectUnits(this._gapinMm);
	this.render();
};

/**
 * Recalculate the content of the popup (correct units) and render
 */
CaliperPopoverModel.prototype.recalculateGap = function() {
	if (this._gapinMm !== null && this._endPointValueInMm !== null) {
		this.t.d.endPointValue = this._convertToCorrectUnits(this._endPointValueInMm);
		this.t.d.gap = this._convertToCorrectUnits(this._gapinMm);
		this.render();
	}
};

CaliperPopoverModel.prototype.setEndPoint = function(endPoint) {
	this._endPoint = endPoint;
	if (this._endPoint) {
		this.t.d.endPointValue = this._convertToCorrectUnits(this.t.d.isMeasuringHeight ? this._endPoint.y : this._endPoint.x);
	}
};

CaliperPopoverModel.prototype._convertToCorrectUnits = function(value) {
	var resolution = null;
	if (this.f.getUnits() === CoreConstants.SizeUnits.PIXELS) {
		var page = this.f.getActivePage();
		if (this.t.d.isMeasuringHeight) {
			resolution = page.getParam("resolutiony");
		} else {
			resolution = page.getParam("resolutionx");
		}
	}
	var converted = ViewerUtils.convertFromMmValue(value, this.f.getUnits(), this.f.getUnitsFix(), ViewerUtils.parseFloat(resolution));
	var units = this.f.i18n.t(this.f.getUnitsLocName());

	return converted + " " + units;
};

/**
 * Set the area of the popover
 * @param {ShapeUtils.Rect} area for which the popover should be shown
 */
CaliperPopoverModel.prototype.setFrame = function(frame) {
	this.frame = frame;
};

/**
 * Change the mode of the popover
 * @param isMeasuringHeight
 */
CaliperPopoverModel.prototype.setMeasuringHeight = function(isMeasuringHeight) {
	this.t.d.isMeasuringHeight = isMeasuringHeight;
	if (this._endPoint) {
		if (this.t.d.isMeasuringHeight) {
			this.t.d.endPointLabel = "Y";
			// convert to correct coordinate system
			var endPointY = this._endPoint.copy();
			var transformedEndPointY = this._coordinates.getTransformedPoint(endPointY);
			this.t.d.endPointValue = this._convertToCorrectUnits(transformedEndPointY.y);
			this._endPointValueInMm = transformedEndPointY.y;
		} else {
			this.t.d.endPointLabel = "X";
			// convert to correct coordinate system
			var endPointX = this._endPoint.copy();
			var transformedEndPointX = this._coordinates.getTransformedPoint(endPointX);
			this.t.d.endPointValue = this._convertToCorrectUnits(transformedEndPointX.x);
			this._endPointValueInMm = transformedEndPointX.x;
		}
	}
	this.render();
};
export default CaliperPopoverModel;
