import ToolbarComponent from "@esko/toolbar/ToolbarComponent";
import ButtonGroup from "@esko/toolbar/ButtonGroup";
import SubtoolbarButton from "@esko/web-viewer/layout/component/SubtoolbarButton";
import Spinner from "@esko/web-viewer/layout/component/Spinner";
import CoreConstants from "@esko/web-viewer/core/CoreConstants";
import MeasureConstants from "@esko/web-viewer/measure/MeasureConstants";
import ViewerUtils from "@esko/web-viewer/utils/ViewerUtils";
import ListenerUtil from "@esko/web-viewer/utils/ListenerUtil";
import $ from "jquery";
import _ from "underscore";

/**
 * @exports CaliperSubtoolbarComponent
 * @memberof MeasureModule
 * @class This is the subtoolbar of the caliper tool
 * @augments ToolbarComponent
 * @mixes {Utils.ListenerUtil}
 * @constructor
 * @param {Core.ViewerFacade} f - Viewer facade
 * @param {CaliperModel} caliperModel - The model of the caliper tool
 */
var CaliperSubtoolbarComponent = function(f, caliperModel) {
	ToolbarComponent.call(this, "CaliperSubtoolbar", f);
	ListenerUtil.mixin(this, f);

	this.f = f;
	this._caliperModel = caliperModel;

	this.f.i18n.loadBundle(
		"viewer",
		function() {
			// Style selector
			this._stylesLabel = $("<span />")
				.addClass("styleLabel")
				.text(this.f.i18n.t("GapStyle") + ":");
			this._styles = $("<select />").addClass("styleSelect mousetrap");
			this._refreshGaps();
			this._styles.on("change", _.bind(this._styleSelectorChanged, this));
			this.element.append(
				$("<div class='caliperStyle'/>")
					.append(this._stylesLabel)
					.append(this._styles)
			);

			// Gap
			this._gapLabel = $("<span />")
				.addClass("gapLabel")
				.text(this.f.i18n.t("GapSize") + ":");
			this._gapSpinner = new Spinner(this.f, {
				min: 0,
				max: 9999,
				step: 0.01,
				page: 10,
				onSpinnerChanged: _.bind(this._spinnerChanged, this),
				inputUnit: this.f.i18n.t(this.f.getUnitsLocName())
			});
			this.element.append(
				$("<div class='caliperGap'/>")
					.append(this._gapLabel)
					.append(this._gapSpinner.element.addClass("mousetrap"))
			);

			// Width/Height buttons
			this._heightWidthButtonGroup = new ButtonGroup("heightWidthButtonGroup");
			this._heightWidthButtonGroup.element.addClass("heightWidthButtonGroup");
			this._heightButton = new SubtoolbarButton(this.f, "checkHeight", "icon-WCR_Height", true, _.bind(this._heightButtonClicked, this));
			this._heightButton.element.append(this.f.i18n.t("GapHeight"));
			this._heightWidthButtonGroup.addButton(this._heightButton);
			this._widthButton = new SubtoolbarButton(this.f, "checkWidth", "icon-WCR_Width", true, _.bind(this._widthButtonClicked, this));
			this._widthButton.element.append(this.f.i18n.t("GapWidth"));
			this._heightWidthButtonGroup.addButton(this._widthButton);
			this.addButtonGroup(this._heightWidthButtonGroup);

			// Buttons Group
			this._saveButtonGroup = new ButtonGroup("saveButtonGroup");
			// Save button
			this._saveGapButton = new SubtoolbarButton(this.f, "SaveGap", "icon-WCR_Save", true, _.bind(this._saveClicked, this));
			this._saveGapButton.setTooltip("viewer", "SaveGapTT");
			this._saveButtonGroup.addButton(this._saveGapButton);

			// Delete button
			this._deleteGapButton = new SubtoolbarButton(
				this.f,
				"DeleteGap",
				"icon-WCR_Actions-Delete12",
				true,
				_.bind(this._deleteClicked, this)
			);
			this._deleteGapButton.setTooltip("viewer", "DeleteGapTT");
			this._saveButtonGroup.addButton(this._deleteGapButton);
			this.addButtonGroup(this._saveButtonGroup);

			// Listen to model
			this._registerListener("changed.currentGap", this._gapChanged, this, this._caliperModel);
			this._registerListener("changed.currentGap.vector", this._gapVectorChanged, this, this._caliperModel);
			this._registerListener("changed.presetGaps.new", this._refreshGaps, this, this._caliperModel);
			this._registerListener("changed.presetGaps.deleted", this._deletedPresetGap, this, this._caliperModel);

			// listen to unit change
			this.f.getCurrentUser().on("options.changed.PREFERENCES_UNIT_SIZE", _.bind(this._unitsUpdated, this));

			// Set initial gap to freeform
			this._caliperModel.setCurrentGapById(this._caliperModel.getFreeformGap().getId());
		},
		this
	);
};

CaliperSubtoolbarComponent.prototype = Object.create(ToolbarComponent.prototype);
CaliperSubtoolbarComponent.prototype.constructor = CaliperSubtoolbarComponent;

/**
 * On change of style selector.
 * @param {EventObj} e
 */
CaliperSubtoolbarComponent.prototype._styleSelectorChanged = function(e) {
	// if switching between freeform and fixed, copy the vector
	var currentCaliperGap = this._caliperModel.getCurrentGap();
	var newCaliperGap = this._caliperModel.getGapById(e.target.value);
	if (
		(newCaliperGap.getStyle() === MeasureConstants.CaliperStyle.Fixed &&
			currentCaliperGap.getStyle() === MeasureConstants.CaliperStyle.Freeform) ||
		(newCaliperGap.getStyle() === MeasureConstants.CaliperStyle.Freeform &&
			currentCaliperGap.getStyle() === MeasureConstants.CaliperStyle.Fixed)
	) {
		newCaliperGap.setVector(currentCaliperGap.getVector());
	}
	this._caliperModel.setCurrentGapById(e.target.value);
};

/**
 * On change of the gap spinner
 * @param {Number} value - New value
 */
CaliperSubtoolbarComponent.prototype._spinnerChanged = function(value) {
	var resolution = null;
	if (this.f.getUnits() === CoreConstants.SizeUnits.PIXELS) {
		resolution = this._getResolution(this._caliperModel.getCurrentGap().getVector());
	}
	value = ViewerUtils.convertToMmValue(
		value,
		this.f.getUnits(),
		this.f.getUnitsFix(CoreConstants.SizeUnits.MM),
		ViewerUtils.parseFloat(resolution)
	);

	if (this._caliperModel.getCurrentGap().getStyle() === MeasureConstants.CaliperStyle.Preset) {
		var oldGap = this._caliperModel.getCurrentGap();
		this._caliperModel.setCurrentGapById(this._caliperModel.getFixedGap().getId());
		this._caliperModel.setCurrentGapVector(oldGap.getVector()); // copy settings
	}
	var currentVector = this._caliperModel.getCurrentGap().getVector();
	currentVector.setLength(value);
	this._caliperModel.setCurrentGapVector(currentVector);
};

/**
 * On click of the height button
 * @param {EventObj} e
 */
CaliperSubtoolbarComponent.prototype._heightButtonClicked = function(e) {
	if (this._caliperModel.getCurrentGap().getStyle() === MeasureConstants.CaliperStyle.Preset) {
		var oldGap = this._caliperModel.getCurrentGap();
		this._caliperModel.setCurrentGapById(this._caliperModel.getFixedGap().getId());
		this._caliperModel.setCurrentGapVector(oldGap.getVector()); // copy settings
	}
	var currentVector = this._caliperModel.getCurrentGap().getVector();
	currentVector.setAngleInDegrees(-90);
	this._caliperModel.setCurrentGapVector(currentVector);
	e.stopPropagation();
};

/**
 * On click of the width button
 * @param {EventObj} e
 */
CaliperSubtoolbarComponent.prototype._widthButtonClicked = function(e) {
	if (this._caliperModel.getCurrentGap().getStyle() === MeasureConstants.CaliperStyle.Preset) {
		var oldGap = this._caliperModel.getCurrentGap();
		this._caliperModel.setCurrentGapById(this._caliperModel.getFixedGap().getId());
		this._caliperModel.setCurrentGapVector(oldGap.getVector()); // copy settings
	}
	var currentVector = this._caliperModel.getCurrentGap().getVector();
	currentVector.setAngleInDegrees(0);
	this._caliperModel.setCurrentGapVector(currentVector);
	e.stopPropagation();
};

/**
 * Called when the gap of the model changed.
 * @param {CaliperGap} oldValue - Old gap
 * @param {CaliperGap} newValue - New gap
 */
CaliperSubtoolbarComponent.prototype._gapChanged = function(oldValue, newValue) {
	this._styles.find('option[value="' + newValue.getId() + '"]').prop("selected", true);

	this._gapVectorChanged(null, newValue.getVector());
	this._updateButtons();
};

/**
 * Called when the gap vector of current caliper gap of the model changed
 * @param {ShapeUtils.Vector} oldValue - Old vector
 * @param {ShapeUtils.Vector} newValue - New vector
 */
CaliperSubtoolbarComponent.prototype._gapVectorChanged = function(oldValue, newValue) {
	// angle
	if (this._caliperModel.isMeasuringHeight(newValue)) {
		this._heightButton.setActive(true);
		this._widthButton.setActive(false);
	} else if (this._caliperModel.isMeasuringWidth(newValue)) {
		this._widthButton.setActive(true);
		this._heightButton.setActive(false);
	}

	// gap size
	var resolution = null;
	if (this.f.getUnits() === CoreConstants.SizeUnits.PIXELS) {
		resolution = this._getResolution(newValue);
	}
	var newGap = newValue.getLength();
	newGap = ViewerUtils.convertFromMmValue(newGap, this.f.getUnits(), this.f.getUnitsFix(), ViewerUtils.parseFloat(resolution));
	this._gapSpinner.setValue(newGap);
};

/**
 * Refresh the gaps.
 * The style selector will be replaced with all gaps from the caliper model.
 * If newValue is passed, it will become the selected gap.
 * @param {CaliperGap} [oldValue] - Old caliper gap
 * @param {CaliperGap} [newValue] - New caliper gap
 */
CaliperSubtoolbarComponent.prototype._refreshGaps = function(oldValue, newValue) {
	this._styles.empty();

	_.each(
		this._caliperModel.getAllGaps(),
		function(caliperGap) {
			this._newGap(caliperGap);
		},
		this
	);

	// a new gap -> select it as current gap
	if (newValue !== undefined) {
		this._caliperModel.setCurrentGapById(newValue.getId());
	}
};

/**
 * Handle a change of units: recalculate and redraw toolbar
 */
CaliperSubtoolbarComponent.prototype._unitsUpdated = function() {
	this._gapSpinner.setInputUnit(this.f.i18n.t(this.f.getUnitsLocName()));
	this._gapChanged(this._caliperModel.getCurrentGap(), this._caliperModel.getCurrentGap());
	this._refreshGaps();
};

/**
 * Called when a new gap is added to the presets of the model (or initially), see _refreshGaps.
 * @param {CaliperGap} newValue - New caliper gap
 */
CaliperSubtoolbarComponent.prototype._newGap = function(newValue) {
	// new value is the new gap
	var style = $("<option />")
		.html(this._toString(newValue))
		.val(newValue.getId());

	this._styles.append(style);
};

/**
 * Called when a preset gap is deleted from the model
 * @param {CaliperGap} oldValue - This is the removed preset
 * @param {CaliperGap} newValue - Always null
 */
CaliperSubtoolbarComponent.prototype._deletedPresetGap = function(oldValue, newValue) {
	// old value is the removed fixed gap
	this._styles.find('option[value="' + oldValue.getId() + '"]').remove();

	// switch back to fixed
	this._caliperModel.setCurrentGapById(this._caliperModel.getFixedGap().getId());
	this._caliperModel.setCurrentGapVector(oldValue.getVector()); // copy settings
};

/**
 * Called when the save button is clicked in the sub tool bar
 * Saves the current gap as preset.
 */
CaliperSubtoolbarComponent.prototype._saveClicked = function() {
	var similarPreset = this._caliperModel.getSimilarPreset(this._caliperModel.getCurrentGap());
	if (similarPreset) {
		// there already exists a similar preset, so just switch to that one instead of saving a new one
		this._caliperModel.setCurrentGapById(similarPreset.getId());
	} else {
		this._caliperModel.saveCurrentGapAsPreset();
	}
};

/**
 * Called when the delete button is clicked in the sub tool bar
 * Deletes the current gap from the presets.
 */
CaliperSubtoolbarComponent.prototype._deleteClicked = function() {
	this._caliperModel.deleteCurrentGapFromPresets();
};

/**
 * Update the buttons save/delete (decides which ones to show)
 */
CaliperSubtoolbarComponent.prototype._updateButtons = function() {
	var currentGap = this._caliperModel.getCurrentGap();
	switch (currentGap.getStyle()) {
		case MeasureConstants.CaliperStyle.Freeform:
			this._deleteGapButton.element.addClass("hiddenElement");
			this._saveGapButton.element.removeClass("hiddenElement");
			break;
		case MeasureConstants.CaliperStyle.Fixed:
			this._deleteGapButton.element.addClass("hiddenElement");
			this._saveGapButton.element.removeClass("hiddenElement");
			break;
		case MeasureConstants.CaliperStyle.Preset:
			this._saveGapButton.element.addClass("hiddenElement");
			this._deleteGapButton.element.removeClass("hiddenElement");
			break;
	}
};

/**
 * Convert a gap to a String representation for the drop down.
 * i18n general bundle should be available!
 * @param {CaliperGap} caliperGap - Gap to represent as string
 * @returns {String} String representation of the given caliper gap
 */
CaliperSubtoolbarComponent.prototype._toString = function(caliperGap) {
	switch (caliperGap.getStyle()) {
		case MeasureConstants.CaliperStyle.Freeform:
			return this.f.i18n.t("GapStyleFreeform");
			break;
		case MeasureConstants.CaliperStyle.Fixed:
			return this.f.i18n.t("GapStyleFixed");
			break;
		case MeasureConstants.CaliperStyle.Preset:
			var resolution = null;
			if (this.f.getUnits() === CoreConstants.SizeUnits.PIXELS) {
				resolution = this._getResolution(caliperGap.getVector());
			}
			var gap = ViewerUtils.convertFromMmValue(
				caliperGap.getVector().getLength(),
				this.f.getUnits(),
				this.f.getUnitsFix(),
				ViewerUtils.parseFloat(resolution)
			);
			var units = this.f.i18n.t(this.f.getUnitsLocName());
			var orientation = "";
			if (this._caliperModel.isMeasuringHeight(caliperGap.getVector())) {
				// heights
				orientation = this.f.i18n.t("GapHeight");
			} else if (this._caliperModel.isMeasuringWidth(caliperGap.getVector())) {
				// widths
				orientation = this.f.i18n.t("GapWidth");
			}
			return gap + " " + units + " (" + orientation + ")";
			break;
	}
};

/**
 * Get the resolution to use for unit conversion
 */
CaliperSubtoolbarComponent.prototype._getResolution = function(vector) {
	var page = this.f.getActivePage();
	var resolution = null;
	if (this._caliperModel.isMeasuringHeight(vector)) {
		resolution = page.getParam("resolutiony");
	} else if (this._caliperModel.isMeasuringWidth(vector)) {
		resolution = page.getParam("resolutionx");
	}
	return resolution;
};

export default CaliperSubtoolbarComponent;
