import _ from "underscore";
import AbstractTool from "@esko/web-viewer/viewport/tools/AbstractTool";
import ShapeUtils from "@esko/shapeutils";
import CaliperLayerModel from "@esko/web-viewer/measure/CaliperLayerModel";
import MeasureConstants from "@esko/web-viewer/measure/MeasureConstants";
import CoreConstants from "@esko/web-viewer/core/CoreConstants";
/**
 * @exports CaliperTool
 * @memberof MeasureModule
 * @class The Caliper tool allows to measure widths/heights
 * @constructor
 * @augments {Core.AbstractTool}
 *
 * @param {Core.ViewerFacade} f - Viewer facade
 * @param {CaliperModel} caliperModel - The model of the caliper tool
 */
var CaliperTool = function(f, caliperModel) {
	AbstractTool.call(this, f, CoreConstants.TOOLS.MEASURE_CALIPER);

	// the distance (in pixels) of a line for going to the hover state
	this.CLOSETODISTANCE = 16;

	this._caliperModel = caliperModel;
	this._caliperLayerModel = new CaliperLayerModel(f, this.getIID(), caliperModel);

	/**
	 * DEFAULT STATE -> nothing set
	 */
	this.registerState("default", {
		enterState: _.bind(function() {
			this._caliperLayerModel.hidePopover();
			this._caliperLayerModel.setStartPoint(null);
			this._caliperModel.resetFreeformVector();
		}, this),

		mousedown: _.bind(function(e) {
			var startPoint = this.keepPointWithinWorldBounds(this.getWorldPoint(e), false);
			this._caliperLayerModel.setStartPoint(startPoint);
			if (!this._caliperLayerModel.getEndPoint()) {
				// if the end point is not defined yet, move the endline to the start point (in this way we prevent some flickering in the UI)
				this._moveEndLine(e);
			}
			this.changeState("dragging");
		}, this),

		mouseup: function() {},

		mousemove: function() {},

		mouseout: function() {},

		mouseleave: function() {},

		/**
		 * no context menu!
		 */
		contextmenu: _.bind(function(e) {
			e.preventDefault();
			e.stopPropagation();
		}, this),

		canBeInterrupted: true
	});

	/**
	 * DRAGGING STATE -> after setting start point: end line can be set (freeform), or a new start point can be set (fixed)
	 */
	this.registerState(
		"dragging",
		_.defaults(
			{
				enterState: _.bind(function() {
					this._caliperLayerModel.showPopover();
					this.f.emit("viewport.startInteractive");
				}, this),
				leaveState: _.bind(function() {
					this.f.emit("viewport.stopInteractive");
				}, this),
				mousedown: function() {},
				mouseup: _.bind(function() {
					var currentGap = this._caliperModel.getCurrentGap();
					var currentStyle = currentGap.getStyle();
					if (
						currentStyle === MeasureConstants.CaliperStyle.Fixed ||
						currentStyle === MeasureConstants.CaliperStyle.Preset ||
						currentGap.getVector().getLength() > 0
					) {
						this.changeState("drawn");
					}
				}, this),

				mousemove: _.bind(function(e) {
					var currentGap = this._caliperModel.getCurrentGap();
					var currentStyle = currentGap.getStyle();
					if (currentStyle === MeasureConstants.CaliperStyle.Fixed || currentStyle === MeasureConstants.CaliperStyle.Preset) {
						this._moveStartLine(e);
					} else {
						this._moveEndLine(e);
					}
				}, this),

				canBeInterrupted: false
			},
			this.getState("default")
		)
	);

	/**
	 * DRAWN STATE -> end point also set
	 */
	this.registerState(
		"drawn",
		_.defaults(
			{
				enterState: function() {},
				mousemove: _.bind(function(e) {
					if (this._isHovering(e)) {
						this.changeState("hover");
					}
				}, this),

				canBeInterrupted: true
			},
			this.getState("default")
		)
	);

	/**
	 * HOVER STATE -> hover over a line
	 */
	this.registerState(
		"hover",
		_.defaults(
			{
				enterState: function() {},
				mousedown: _.bind(function(e) {
					var worldPoint = this.getWorldPoint(e);
					var startPoint = this._caliperLayerModel.getStartPoint();
					var endPoint = this._caliperLayerModel.getEndPoint();

					if (this._isCloseToLine(startPoint, worldPoint)) {
						// close to start line -> move start line
						this.changeState("movingStartLine");
					} else if (this._isCloseToLine(endPoint, worldPoint)) {
						// close to end line -> move end line
						this.changeState("movingEndLine");
					}
				}, this),

				mousemove: _.bind(function(e) {
					if (!this._isHovering(e)) {
						this.changeState("drawn");
					}
				}, this),

				mouseleave: _.bind(function(e) {
					this.changeState("drawn");
				}, this),

				canBeInterrupted: true
			},
			this.getState("default")
		)
	);

	/**
	 * MOVE END LINE STATE -> end line is being moved
	 */
	this.registerState(
		"movingEndLine",
		_.defaults(
			{
				enterState: _.bind(function() {
					this.f.emit("viewport.startInteractive");
				}, this),
				leaveState: _.bind(function() {
					this.f.emit("viewport.stopInteractive");
				}, this),
				mouseleave: _.bind(function(e) {
					this.changeState("drawn");
				}, this),

				mouseup: _.bind(function(e) {
					this.changeState("drawn");
				}, this),

				mousemove: _.bind(function(e) {
					this._moveEndLine(e, true);
				}, this),

				canBeInterrupted: false
			},
			this.getState("default")
		)
	);

	/**
	 * MOVE START LINE STATE -> start line is being moved
	 */
	this.registerState(
		"movingStartLine",
		_.defaults(
			{
				enterState: _.bind(function() {
					this.f.emit("viewport.startInteractive");
				}, this),
				leaveState: _.bind(function() {
					this.f.emit("viewport.stopInteractive");
				}, this),
				mouseleave: _.bind(function(e) {
					this.changeState("drawn");
				}, this),

				mouseup: _.bind(function(e) {
					this.changeState("drawn");
				}, this),

				mousemove: _.bind(function(e) {
					this._moveStartLine(e);
				}, this),

				canBeInterrupted: false
			},
			this.getState("default")
		)
	);

	this.registerActiveEvents(["mousedown", "mousemove", "mouseup", "mouseleave", "contextmenu"]);

	// Esc allows to reset selection
	this.activeKeyEvents.push([
		"esc",
		_.bind(function(e) {
			this.changeState("default");
			e.preventDefault();
			e.stopPropagation();
		}, this),
		"keyup"
	]);

	/**
	 * @override
	 */
	this._override("initialize", function(widget, toolManager) {
		this._super(widget, toolManager);
		this.changeState("default");
	});

	/**
	 * @override
	 */
	this._override("activate", function(interruptRequest) {
		this._super(interruptRequest);
		if (!interruptRequest) {
			this.f.getModule("viewport").addLayer(this.getIID(), this._caliperLayerModel);
		}
	});

	/**
	 * @override
	 */
	this._override("deactivate", function(interruptRequest) {
		this._super(interruptRequest);
		if (!interruptRequest) {
			this._caliperLayerModel.modelChanged(true);
			this.f.getModule("viewport").removeLayer(this.getIID());
		}

		if (!interruptRequest || _.contains(["movingStartLine", "movingEndLine", "dragging"], this.getCurrentStateName())) {
			this.changeState("default");
		}
	});
};

// extend from the AbstractTool
// inherit
CaliperTool.prototype = Object.create(AbstractTool.prototype);
CaliperTool.prototype.constructor = CaliperTool;

/**
 * Move the end line to the new point given by e
 * @param {EventObj} e
 * @param {Boolean} [keepAngle] - if the true, the angle of te vector is not changed/altered by moving the endline
 */
CaliperTool.prototype._moveEndLine = function(e, keepAngle) {
	var currentStyle = this._caliperModel.getCurrentGap().getStyle();
	var newPoint = this.keepPointWithinWorldBounds(this.getWorldPoint(e), false);
	var startPoint = this._caliperLayerModel.getStartPoint();
	if (currentStyle === MeasureConstants.CaliperStyle.Freeform) {
		this._caliperLayerModel.setPoints(startPoint, newPoint, false, keepAngle);
	} else if (currentStyle === MeasureConstants.CaliperStyle.Fixed || currentStyle === MeasureConstants.CaliperStyle.Preset) {
		this._caliperLayerModel.setPoints(startPoint, newPoint, true, keepAngle);
	}
};

/**
 * Move the start line to the new point given by e
 * @param {Event} e - mouse event
 */
CaliperTool.prototype._moveStartLine = function(e) {
	var currentStyle = this._caliperModel.getCurrentGap().getStyle();
	var newPoint = this.keepPointWithinWorldBounds(this.getWorldPoint(e), false);
	if (currentStyle === MeasureConstants.CaliperStyle.Freeform) {
		var endPoint = this._caliperLayerModel.getEndPoint();
		this._caliperLayerModel.setPoints(newPoint, endPoint, false, true);
	} else if (currentStyle === MeasureConstants.CaliperStyle.Fixed || currentStyle === MeasureConstants.CaliperStyle.Preset) {
		this._caliperLayerModel.setStartPoint(newPoint);
	}
};

/**
 * Check if hovered over begin or end line by point given by e
 * Also sets the cursor correctly.
 * @param {EventObj} e
 * @returns {Boolean} True if hovering over begin or end line, false otherwise
 */
CaliperTool.prototype._isHovering = function(e) {
	var worldPoint = this.getWorldPoint(e);
	var startPoint = this._caliperLayerModel.getStartPoint();
	var endPoint = this._caliperLayerModel.getEndPoint();

	if (this._isCloseToLine(startPoint, worldPoint) || this._isCloseToLine(endPoint, worldPoint)) {
		var rotation = this._widget.getWorld().getRotation();
		if (this._caliperModel.isMeasuringHeight(this._caliperModel.getCurrentGap().getVector())) {
			if (rotation % 180 === 90) {
				this.setCursor("cursor-resizeE");
			} else {
				this.setCursor("cursor-resizeN");
			}
		} else if (this._caliperModel.isMeasuringWidth(this._caliperModel.getCurrentGap().getVector())) {
			if (rotation % 180 === 90) {
				this.setCursor("cursor-resizeN");
			} else {
				this.setCursor("cursor-resizeE");
			}
		}
		return true;
	} else {
		this.setCursor(this.getDefaultCursor());
		return false;
	}
};

/**
 * Check if the given worldPoint is close to the line through worldStartorEndPoint according to measuring heights/widths
 * @param {ShapeUtils.Point} startOrEndPoint - world start or end point
 * @param {ShapeUtils.Point} worldPoint - world point to check
 * @returns {Boolean} True if the worldPoint is close to the line through worldStartorEndPoint according to the current gap vector, false otherwise
 */
CaliperTool.prototype._isCloseToLine = function(worldStartorEndPoint, worldPoint) {
	var zoom = this._widget.getWorld().getZoom();
	var dist = Math.abs(this._getDistanceToLine(worldStartorEndPoint, worldPoint));
	return dist < this.CLOSETODISTANCE / zoom;
};

/**
 * Get the distance of the worldPoint to the line through worldStartorEndPoint according to measuring heights/widths
 * @param {ShapeUtils} worldStartorEndPoint - world start or end point
 * @param {ShapeUtils} worldPoint - world point to check
 * @returns {Number} distance of the worldPoint to the line through worldStartorEndPoint according to measuring heights/widths
 */
CaliperTool.prototype._getDistanceToLine = function(worldStartorEndPoint, worldPoint) {
	var vector = this._caliperModel.getCurrentGap().getVector();
	if (this._caliperModel.isMeasuringHeight(vector)) {
		return worldStartorEndPoint.y - worldPoint.y;
	} else if (this._caliperModel.isMeasuringWidth(vector)) {
		return worldStartorEndPoint.x - worldPoint.x;
	}
};

/**
 * @override
 */
CaliperTool.prototype.canBeInterrupted = function() {
	return this.getCurrentState().canBeInterrupted;
};

/**
 * Get the caliper model. Used to give it to the sub tool bar component
 * @returns {CaliperModel} The caliper model of the tool.
 */
CaliperTool.prototype.getCaliperModel = function() {
	return this._caliperModel;
};

/**
 * @see AbstractMeasureTool or MeasureModule
 * Can the tool be used on the currently active page?
 *
 * @returns {Boolean}
 */
CaliperTool.prototype.canBeUsedOnActivePage = function() {
	// can be used on all active pages
	return true;
};

/**
 * @see AbstractMeasureTool or MeasureModule
 * Update the data in the tool if needed
 *
 * @returns {Boolean}
 */
CaliperTool.prototype.updateIfNeeded = function(forceUpdate) {
	// do nothing
};

/**
 * @override
 */
CaliperTool.prototype.getDefaultCursor = function() {
	return "rulerCursor";
};
/**
 * Override, so activating the tool triggers a model update
 * @override
 */
CaliperTool.prototype.doActivation = function() {
	this._caliperLayerModel.modelChanged(false);
	AbstractTool.prototype.doActivation.call(this);
};

/**
 * Get the state of the measure tool
 * @see AbstractMeasureTool
 * @returns {Object}
 */
CaliperTool.prototype.getToolState = function() {
	var currentGap = this._caliperModel.getCurrentGap();
	var state = {
		// We cannot rely on the ID being right because of random ordering, so we don't save the gap, but rather the style and vector separately and search it back in the set state method
		caliperStyle: currentGap.getStyle(),
		caliperGapVector: currentGap.getVector(),
		startPoint: this._caliperLayerModel.getStartPoint(),
		isActive: this.isActive()
	};
	return state;
};

/**
 * Set the state of the measure tool
 * @see AbstractMeasureTool
 * @param {Object} state
 */
CaliperTool.prototype.setToolState = function(state) {
	if (state.isActive) {
		// first activate tool, and then disable interruption
		this.f.emit("toolManager.activateTool", this);
		this.f.emit("toolManager.interruptTools", true);
		var style = state.caliperStyle;
		var vector = state.caliperGapVector;
		if (style !== undefined && vector !== undefined && state.startPoint !== undefined) {
			// Find back the right caliper gap
			// We cannot rely on the ID being right because of random ordering, so we search it ourself in the model

			// Set the gap by id and vector
			var setGap = _.bind(function(id, vector) {
				this._caliperModel.setCurrentGapById(id);
				this._caliperModel.setCurrentGapVector(vector);
			}, this);

			if (style === MeasureConstants.CaliperStyle.Freeform) {
				// freeform
				var caliperGap = this._caliperModel.getFreeformGap();
				setGap(caliperGap.getId(), vector);
			} else {
				// fixed or preset
				var caliperGap = this._caliperModel.getFixedGap();
				setGap(caliperGap.getId(), vector);

				// first set fixed, then find right similar preset
				if (style === MeasureConstants.CaliperStyle.Preset) {
					var preset = this._caliperModel.getSimilarPreset(this._caliperModel.getCurrentGap());
					if (preset !== undefined) {
						// set to the right preset gap
						this._caliperModel.setCurrentGapById(preset.getId());
					}
				}
			}

			// Set start point
			this._caliperLayerModel.setStartPoint(state.startPoint);
			// Show popover
			this._caliperLayerModel.showPopover();
		}
	}
};

export default CaliperTool;
