import _ from "underscore";
import $ from "jquery";
import AbstractMeasureTool from "@esko/web-viewer/measure/AbstractMeasureTool";
import MeasureConstants from "@esko/web-viewer/measure/MeasureConstants";
import EyeDropperPopoverModel from "@esko/web-viewer/measure/EyeDropperPopoverModel";
import ShapeUtils from "@esko/shapeutils";
import ViewerUtils from "@esko/web-viewer/utils/ViewerUtils";
import CoreConstants from "@esko/web-viewer/core/CoreConstants";
/**
 * Eye dropper tool, You can see the color density of your selected area/point
 *
 * @exports EyeDropperMeasureTool
 * @memberof MeasureModule
 * @class EyeDropperMeasureTool
 * @augments Measure.AbstractMeasureTool
 *
 * @param {Core.ViewerFacade} f
 */
var EyeDropperMeasureTool = function(f) {
	AbstractMeasureTool.call(this, f, CoreConstants.TOOLS.MEASURE_EYEDROPPER);

	this.model.setRulerType(MeasureConstants.MeasureType.EyeDropper);

	//override the enterState event for the clickState
	this.registerState(
		this.clickState,
		_.defaults(
			{
				enterState: _.bind(function() {
					var worldResolution = this.getWorldResolution();
					if (worldResolution !== null) {
						var pixelSize = MeasureConstants.ClickSize.NORMAL;
						if (this.f.getActivePage().isDigitalFilmPage()) {
							pixelSize = MeasureConstants.ClickSize.DIGITAL_FILM;
						}
						var sizeX = pixelSize / worldResolution.x; //get the amount of px per mm for the document, not for the screen
						var sizeY = pixelSize / worldResolution.y;
						this.model.setStartPoint(
							new ShapeUtils.Point(this.model.ruler.startPoint.x - sizeX / 2, this.model.ruler.startPoint.y - sizeY / 2)
						);
						this.model.setEndPoint(new ShapeUtils.Point(this.model.ruler.startPoint.x + sizeX, this.model.ruler.startPoint.y + sizeY));

						this.calculate();
						this.changeState(this.defaultState);
					}
				}, this)
			},
			this.getState(this.drawingState)
		)
	);

	// parsed channel info in %
	this._channels = [];

	// listen to preference changes
	this.f.getCurrentUser().on("options.changed.PREFERENCES_UNIT_EYEDROPPER", _.bind(this._onEyeDropperUnitChanged, this));

	// listen to unit size for adapting eye dropper units
	this.f.getCurrentUser().on(
		"options.changed.PREFERENCES_UNIT_SIZE",
		_.bind(function() {
			var currentSizeUnit = this.f.getOption("PREFERENCES_UNIT_SIZE");

			// sync with size units
			if (currentSizeUnit !== CoreConstants.SizeUnits.POINTS) {
				// safety measure
				this.f.setOption("PREFERENCES_UNIT_EYEDROPPER", currentSizeUnit, true);
			}
		}, this)
	);
};

// extend from the AbstractTool
EyeDropperMeasureTool.prototype = Object.create(AbstractMeasureTool.prototype);
EyeDropperMeasureTool.prototype.constructor = EyeDropperMeasureTool;

/**
 * @override
 *
 * Gets the information from the to object converted xml and creates a MeasurePopoverModel that shows the information and tells the AbstractModel to add the popover
 *
 * @param {Object} object
 */
EyeDropperMeasureTool.prototype.showInfo = function(object) {
	this._channels = [];
	if (object) {
		if (object.status.info.dotpercentage) {
			if (!_.isArray(object.status.info.dotpercentage)) {
				object.status.info.dotpercentage = [object.status.info.dotpercentage];
			}

			_.each(
				object.status.info.dotpercentage,
				function(channel) {
					this._channels[channel.color] = ViewerUtils.parseFloat(channel.dotvalue);
				},
				this
			);
		}
	}
	this._onEyeDropperUnitChanged(true);
};

/**
 * Converts the values to the right unit
 * @param {Boolean} forceShow
 */
EyeDropperMeasureTool.prototype._onEyeDropperUnitChanged = function(forceShow) {
	if ((this.popover && this.isPopoverShown) || forceShow) {
		var ruler = this.model.ruler;
		if (ruler && ruler.startPoint && ruler.endPoint) {
			var size = ruler.startPoint.sizeTo(ruler.endPoint);

			var currentUnit = CoreConstants.EyeDropperUnits.PERCENT; // default for non-digital films
			if (this.f.getActivePage().isDigitalFilmPage()) {
				currentUnit = this.f.getOption("PREFERENCES_UNIT_EYEDROPPER");
			}

			var width = size.width; // in mm
			var height = size.height; // in mm

			var area = 0;

			switch (currentUnit) {
				case CoreConstants.EyeDropperUnits.SQMM:
					area = width * height;
					break;
				case CoreConstants.EyeDropperUnits.SQCM:
					area =
						ViewerUtils.parseFloat(ViewerUtils.convertFromMmValue(width, CoreConstants.SizeUnits.CM)) *
						ViewerUtils.parseFloat(ViewerUtils.convertFromMmValue(height, CoreConstants.SizeUnits.CM));
					break;
				case CoreConstants.EyeDropperUnits.SQINCH:
					area =
						ViewerUtils.parseFloat(ViewerUtils.convertFromMmValue(width, CoreConstants.SizeUnits.INCH)) *
						ViewerUtils.parseFloat(ViewerUtils.convertFromMmValue(height, CoreConstants.SizeUnits.INCH));
					break;
				case CoreConstants.EyeDropperUnits.PIXELS:
					var worldResolution = this.getWorldResolution();
					if (worldResolution !== null) {
						area = width * worldResolution.x * (height * worldResolution.y);
					}
					break;
				case CoreConstants.EyeDropperUnits.PERCENT:
					area = 100;
					break;
			}

			// convert
			var convertedChannels = [];
			_.each(
				this._channels,
				function(value, key) {
					if (value) {
						convertedChannels[key] = "" + (value / 100) * area; // graphics model parsing expects strings
					}
				},
				this
			);

			// some extra properties
			convertedChannels.makeTotalCount = true;
			convertedChannels.unit = this.f.i18n.t(CoreConstants.EyeDropperUnitsShortLocNames[currentUnit]);
			convertedChannels.unitFix = CoreConstants.EyeDropperUnitsFixes[currentUnit];

			this.f.emit("channels.informationAdded", convertedChannels);

			this.showPopover(convertedChannels);
		}
	}
};

/**
 * @override
 *
 * Sets the required options for the call to the viewserver
 */
EyeDropperMeasureTool.prototype.calculate = function() {
	this.getInfo({});
};

/**
 * @override
 * @returns {Boolean}
 */
EyeDropperMeasureTool.prototype.canBeUsedOnActivePage = function() {
	var page = this.f.getActivePage();
	return page && page.hasGraphics();
};

/**
 * @override
 * Does this tool need a total count
 * @returns {Boolean} True
 */
EyeDropperMeasureTool.prototype._needsTotalCount = function() {
	return true;
};

/**
 * @override
 */
EyeDropperMeasureTool.prototype.createPopoverModel = function(channelInfo, area) {
	return new EyeDropperPopoverModel(
		this.f,
		area,
		"Channels",
		$("<i>").addClass("icon icon-normal icon-WCR_Measure-Colorpicker"),
		channelInfo,
		this._needsTotalCount()
	);
};

export default EyeDropperMeasureTool;
