import _ from "underscore";
import $ from "jquery";
import LayerModel from "@esko/web-viewer/viewport/model/ViewportLayerModel";
import Ruler from "@esko/web-viewer/measure/Ruler";
import XMLUtils from "@esko/xmlutils";
import ShapeUtils from "@esko/shapeutils";
import MeasureConstants from "@esko/web-viewer/measure/MeasureConstants";
import SnapIndicator from "@esko/web-viewer/utils/SnapIndicator";
import RulerSnapIndicatorPopoverModel from "@esko/web-viewer/measure/RulerSnapIndicatorPopoverModel";
/**
 * @exports MeasureLayerModel
 * @memberof MeasureModule
 * @constructor
 */
var MeasureLayerModel = function(f, ID) {
	LayerModel.call(this, f, ID);

	this.f = f;

	this.ruler = new Ruler(f);
	this._drawType = MeasureConstants.MeasureDrawType.RECTANGLE;

	this._allowStartPointSnapping = false;
	this.snapIndicator = null;
	this.snapIndicatorModel = new RulerSnapIndicatorPopoverModel(f, "RulerSnapIndicatorPopover", this.snapIndicator);
};

// inherit
MeasureLayerModel.prototype = Object.create(LayerModel.prototype);
// correct the constructor pointer
MeasureLayerModel.prototype.constructor = MeasureLayerModel;

/**
 * @override
 */
MeasureLayerModel.prototype.getLinework = function(animated, rotationanimation) {
	var rulerLines = _.clone(this.ruler.getLinework());
	if (this.snapIndicator) {
		rulerLines.push(this.snapIndicator);
	}
	return rulerLines;
};

/**
 * Set the start point of the measure tool
 * @param {ShapeUtils.Point} startPoint
 * @param {Boolean} snap - if true, snap to the closest angle
 */
MeasureLayerModel.prototype.setStartPoint = function(startPoint, snap) {
	this.snapIndicator = null;

	this.ruler.startPoint = startPoint;
	if (snap && this._allowStartPointSnapping) {
		this.ruler.snapToClosestAngle(true);
	}
	this.repaint();
	this.modelChanged();
};

/**
 * Set the endpoint of the measure tool
 * @param {ShapeUtils.Point} endPoint - the endpoint of the ruler/measure tool
 * @param {Boolean} snap - if true, snap to the closest angle
 */
MeasureLayerModel.prototype.setEndPoint = function(endPoint, snap) {
	var previousEndPoint = this.ruler.endPoint ? this.ruler.endPoint.copy() : null;

	this.ruler.endPoint = endPoint;
	if (snap) {
		this.ruler.snapToClosestAngle();
	}
	if (!previousEndPoint || !previousEndPoint.equals(this.ruler.endPoint)) {
		this.repaint();

		if (this.ruler.rulerType == MeasureConstants.MeasureType.Ruler) {
			this.calculate();
		}
	}
	this.modelChanged();
};

/**
 * Get the start point of the measure tool
 * @returns {ShapeUtils.Point}
 */
MeasureLayerModel.prototype.getStartPoint = function() {
	return this.ruler.startPoint;
};

/**
 * Get the end point of the measure tool
 * @returns {ShapeUtils.Point}
 */
MeasureLayerModel.prototype.getEndPoint = function() {
	return this.ruler.endPoint;
};

MeasureLayerModel.prototype.repaint = function() {
	this.modelChanged();
};

MeasureLayerModel.prototype.reset = function() {
	this.ruler.reset();
	this.modelChanged();
};

MeasureLayerModel.prototype.hidePopover = function() {
	this.removePopover(this.name);
};

MeasureLayerModel.prototype.calculate = function(options) {
	if (this.ruler.endPoint && this.ruler.startPoint) {
		if (this.ruler.rulerType == MeasureConstants.MeasureType.Ruler) {
			var object = {};

			object.x1 = this.ruler.startPoint.x;
			object.y1 = this.ruler.startPoint.y;
			object.x2 = this.ruler.endPoint.x;
			object.y2 = this.ruler.endPoint.y;

			object.distance = this.ruler.getLength();
			object.angle = this.ruler.getAngle();

			var size = this.ruler.startPoint.sizeTo(this.ruler.endPoint);
			object.width = size.width;
			object.height = size.height;
		}
	}
};

MeasureLayerModel.prototype.setRulerType = function(type) {
	this.ruler.rulerType = type;
	this.reset();
};

/**
 * Set the point that should indicate that the user can snap
 * @param {ShapeUtils.Point} point
 */
MeasureLayerModel.prototype.setSnapIndicatorPoint = function(point) {
	var redraw = false; //only redraw if the situation changed
	if (point) {
		if (!this._allowStartPointSnapping && this.ruler.startPoint && point.equals(this.ruler.startPoint)) {
			if (this.snapIndicator) {
				this.snapIndicator = null;
				redraw = true;
			}
		} else {
			if (!this.snapIndicator) {
				this.snapIndicator = new SnapIndicator(point);
				redraw = true;
			} else {
				if (!point.equals(this.snapIndicator.rect.center())) {
					this.snapIndicator.setCenter(point);
					redraw = true;
				}
			}
		}
	} else {
		if (this.snapIndicator) {
			this.snapIndicator = null;
			redraw = true;
		}
	}

	if (redraw) {
		var indicatorIsSameAsEndpoint = false; //check if the indicator is the same as the endpoint
		var indicatorIsSameAsStartpoint = false;
		if (this.ruler && this.snapIndicator) {
			if (this.ruler.endPoint && this.snapIndicator.rect.center().equals(this.ruler.endPoint)) {
				indicatorIsSameAsEndpoint = true;
			}
			if (this.ruler.startPoint && this.snapIndicator.rect.center().equals(this.ruler.startPoint)) {
				indicatorIsSameAsStartpoint = true;
			}
		}

		var showSnapIndicator = this.snapIndicator !== null && !indicatorIsSameAsEndpoint && !indicatorIsSameAsStartpoint;
		if (showSnapIndicator && !this.getPopoverModel(this.snapIndicatorModel.getID())) {
			this.snapIndicatorModel.snapIndicator = this.snapIndicator;
			this.addPopoverModel(this.snapIndicatorModel);
		} else if (!showSnapIndicator && this.getPopoverModel(this.snapIndicatorModel.getID())) {
			this.removePopoverModel(this.snapIndicatorModel.getID());
		}

		this.snapIndicatorModel.render();
		this.modelChanged();
	}
};

MeasureLayerModel.prototype.clearSnapIndicator = function() {
	this.setSnapIndicatorPoint(null);
};

/**
 * Set whether or not start point can be snapped
 * @param {Boolean} b
 */
MeasureLayerModel.prototype.setAllowStartPointSnapping = function(b) {
	this._allowStartPointSnapping = b;
};

/**
 * Set the type of the selection that will be made (Rectangle, Line, Width, Height)
 * @param {MeasureConstants.MeasureType} drawType
 */
MeasureLayerModel.prototype.setDrawType = function(drawType) {
	this._drawType = drawType;
};

/**
 * Override the modelchanged, so a selection changed event is sent out
 * @param {boolean} deactivate - true when tool using this model gets deactivated
 * @override
 */
MeasureLayerModel.prototype.modelChanged = function(deactivate) {
	var selectionObject = null;

	if (!deactivate) {
		if (this.ruler && this.ruler.startPoint && this.ruler.endPoint) {
			var size = this.ruler.startPoint.sizeTo(this.ruler.endPoint);
			selectionObject = {
				mode: this._drawType,
				w: size.width,
				h: size.height
			};
			if (this._drawType === MeasureConstants.MeasureDrawType.RECTANGLE) {
				selectionObject.x = this.ruler.startPoint.x;
				selectionObject.y = this.ruler.startPoint.y;
			} else if (this._drawType === MeasureConstants.MeasureDrawType.LINE) {
				selectionObject.d = this.ruler.getLength();
				selectionObject.a = this.ruler.getAngle();
			}
		}
	}

	// structure of selectionObject: see {DocumentInfoModel.setSelection}
	this.f.emit("selection.changed", selectionObject);

	// make sure the normal model changed behavior is still executed
	LayerModel.prototype.modelChanged.call(this);
};

export default MeasureLayerModel;
