import ViewerUtils from "@esko/web-viewer/utils/ViewerUtils";
import ViewportPopoverModel from "@esko/web-viewer/viewport/model/ViewportPopoverModel";
import tpl from "@esko/web-viewer/measure/MeasurePopoverTemplate.tpl.html";
import Template from "@esko/template";
import _ from "underscore";
import $ from "jquery";
/**
 * PoveroverModel for the measure tools, defines the layout (uses the the MeasurePopoverTemplate) of the Popover for measure tools
 *
 * @exports MeasurePopoverModel
 * @memberof MeasureModule
 * @class MeasurePopoverModel
 * @augments Viewport.Model.ViewportPopoverModel
 * @constructor
 *
 * @param {Core.ViewerFacade} f
 * @param {LineworkWrapper} area
 * @param {String} name
 * @param {String | JQuery.Element} title - Title to be shown in the header
 * @param {Array<Number> | Array<String>} channels
 * @param {Boolean} [makeTotalCount] Defines if a total of the different channel values needs to be calculated, only possible if the
 * 										values of the channels array are numbers
 * @param {String | JQuery.Element} headerExtra - A jQuery component or string that is displayed in upper-right corner of this popver
 */
var MeasurePopoverModel = function(f, area, name, title, channels, makeTotalCount, headerExtra) {
	ViewportPopoverModel.call(this, f, name);

	this._content = $("<div/>").addClass("measureInfo");

	// Header
	this._header = $("<div />").addClass("measureInfoHeader");
	this._headerTitle = $("<div />")
		.addClass("headerTitle")
		.html(title);
	this._headerExtra = $("<div />")
		.addClass("headerExtra")
		.html(headerExtra);
	this._header.append(this._headerTitle).append(this._headerExtra);
	this._content.append(this._header);

	// Body
	this._body = $("<div />");
	this._content.append(this._body);

	// initialize template
	this.t = new Template(f.i18n, tpl);
	this.t.setContext(this._body);
	this.makeTotalCount = makeTotalCount;
	this._channelsInfo = channels;
	this.channels = [];

	this.t.bind(
		"click",
		".addAnnotation",
		_.bind(function(source, event) {
			//Create a settings XML document
			var content = "<p>";
			_.each(
				this.channels,
				function(channel, i) {
					content += (i > 0 ? "<br />" : "") + "<b>" + channel.name + "</b>: " + channel.value;
				},
				this
			);
			content += "</p>";

			this.f.emit("module.annotation.createAnnotation", this.area.rect, content);
			this.f.emit("toolManager.deactivateTool");
		}, this)
	);

	// load language if needed
	this.f.i18n.loadBundle(
		"viewer",
		_.bind(function() {
			this.t.render();
		}, this)
	);

	// Fixed Template vars
	this.t.d.channels = [];
	this.t.d.makeTotalCount = makeTotalCount;

	this.area = area;

	this.update();

	this.getFrame = function() {
		if (this.area) {
			return this.area.rect;
		} else {
			return null;
		}
	};

	this.getContent = function() {
		return this._content;
	};

	this.getRestrictBorder = function() {
		return 0;
	};

	/**
	 * Is this an interactive popover
	 * @returns {Boolean} True
	 */
	this.isInteractive = function() {
		return true;
	};
};

// inherit
MeasurePopoverModel.prototype = Object.create(ViewportPopoverModel.prototype);
// correct the constructor pointer
MeasurePopoverModel.prototype.constructor = MeasurePopoverModel;

MeasurePopoverModel.prototype.getChannelsInformation = function() {
	return this._channelsInfo;
};

/**
 * Update the popover
 */
MeasurePopoverModel.prototype.update = function() {
	this.f.requestModule(
		"graphics",
		_.bind(function(graphicsModule) {
			var graphicsModel = graphicsModule.getGraphicsModel();
			this.channels = [];
			this.t.d.channels = [];
			this.t.d.addAnnotation = !!this.f.getModule("annotation");
			var parsedInfo = graphicsModel.getParsedChannelsInfo(this._channelsInfo);

			_.each(
				parsedInfo.channels,
				function(channel) {
					if (!graphicsModel.isInDifferenceCompareMode() || !channel.isInPreviousAndMatchedWithCurrent()) {
						this.channels.push(channel);
						this.t.d.channels.push(channel);
					}
				},
				this
			);

			if (this.makeTotalCount) {
				this.t.d.totalCount = parsedInfo.total;
			} else {
				this.t.d.totalCount = "";
			}

			this.t.render();
		}, this)
	);
};

/**
 * Update the channels info
 * @param channelInfo
 */
MeasurePopoverModel.prototype.setChannelsInfo = function(channelInfo) {
	this._channelsInfo = channelInfo;
	this.update();
};

/**
 * Set the area of the popover
 * @param {LineworkWrapper} area
 */
MeasurePopoverModel.prototype.setArea = function(area) {
	this.area = area;
};

/**
 * Replaces/sets the extra header with this new element
 * @param {jQuery.Element} el The new extraHeader component
 */
MeasurePopoverModel.prototype.setHeaderExtra = function(el) {
	this._headerExtra.html(el);
};

/**
 * Replaces/sets the title header with this new element
 * @param {jQuery.Element} el The new header component
 */
MeasurePopoverModel.prototype.setHeader = function(el) {
	this._headerTitle.html(el);
};

export default MeasurePopoverModel;
