import LineworkWrapper from "@esko/web-viewer/utils/LineworkWrapper";
import LineworkPoint from "@esko/web-viewer/utils/LineworkPoint";
import ShapeUtils from "@esko/shapeutils";
import ViewerUtils from "@esko/web-viewer/utils/ViewerUtils";
import DrawUtils from "@esko/web-viewer/utils/DrawUtils";
import MeasureConstants from "@esko/web-viewer/measure/MeasureConstants";
import _ from "underscore";
/**
 * @exports Ruler
 * @memberof MeasureModule
 * @constructor
 */
var Ruler = function(f) {
	this.f = f;

	this.rulerType = 1; // 0 - line, 1 - square

	this.startPoint = null;
	this.endPoint = null;
};

/**
 * Reset ruler; remove both end points
 */
Ruler.prototype.reset = function() {
	this.startPoint = null;
	this.endPoint = null;
};

/**
 * Snap the ruler to the closest angle
 * @param snapStart - controls whether you want to snap the startpoint or the end point
 */
Ruler.prototype.snapToClosestAngle = function(snapStart) {
	if (this.rulerType === 0) {
		var radius = this.getLength();
		var angle = ViewerUtils.closestAngle(this.getAngle(), 8);
		if (snapStart) {
			this.startPoint.x = this.endPoint.x - radius * Math.cos((angle * Math.PI) / 180);
			this.startPoint.y = this.endPoint.y - radius * Math.sin((angle * Math.PI) / 180);
		} else {
			this.endPoint.x = radius * Math.cos((angle * Math.PI) / 180) + this.startPoint.x;
			this.endPoint.y = radius * Math.sin((angle * Math.PI) / 180) + this.startPoint.y;
		}
	}
};

/**
 * Get length of ruler in mm
 * @returns {Number}
 */
Ruler.prototype.getLength = function() {
	if (this.endPoint && this.startPoint) {
		var w = Math.abs(this.endPoint.x - this.startPoint.x);
		var h = Math.abs(this.endPoint.y - this.startPoint.y);
		return Math.sqrt(w * w + h * h);
	} else {
		return 0;
	}
};

/**
 * Get the angle of the ruler in degrees
 * @returns {Number}
 */
Ruler.prototype.getAngle = function() {
	return (Math.atan2(this.endPoint.y - this.startPoint.y, this.endPoint.x - this.startPoint.x) * 180) / Math.PI;
};

/**
 * Get the linework of the ruler tool
 * @returns
 */
Ruler.prototype.getLinework = function() {
	if (this.endPoint && this.startPoint) {
		if (this.rulerType === 0) {
			return this.getRulerLinework();
		} else {
			return this.getSquareLinework();
		}
	} else {
		return [];
	}
};

/**
 * Get square linework for popover
 * @returns {Array.<LineworkWrapper>}
 */
Ruler.prototype.getSquareLinework = function() {
	var linework = [];

	if (this.startPoint && this.endPoint) {
		var ref = new LineworkWrapper();
		ref.fillcolor = "rgba(0, 0, 0, 0.30)";
		ref.color = "rgba(0, 0, 0, 1.00)";
		ref.linewidth = 1;
		ref.type = DrawUtils.LWTYPE_RECT;

		var startpoint = new LineworkPoint();
		startpoint.makeMoveTo(this.startPoint.x, this.startPoint.y);
		ref.pointArray[0] = startpoint;

		var endpoint = new LineworkPoint();
		endpoint.makeLineTo(this.endPoint.x, this.endPoint.y);
		ref.pointArray[1] = endpoint;

		ref.rect = new ShapeUtils.Rect(this.startPoint, this.startPoint.sizeTo(this.endPoint));
		linework.push(ref);

		// dash rectangle
		ref = new LineworkWrapper();
		ref.fillcolor = "rgba(0, 0, 0, 0.0)";
		ref.color = "rgba(220, 220, 220, 0.9)";
		ref.linewidth = 1;
		ref.type = DrawUtils.LWTYPE_DASHEDBOX;
		ref.dash = [4, 3];

		var startpoint = new LineworkPoint();
		startpoint.makeMoveTo(this.startPoint.x, this.startPoint.y);
		ref.pointArray[0] = startpoint;

		var endpoint = new LineworkPoint();
		endpoint.makeLineTo(this.endPoint.x, this.endPoint.y);
		ref.pointArray[1] = endpoint;

		ref.rect = new ShapeUtils.Rect(this.startPoint, this.startPoint.sizeTo(this.endPoint));

		linework.push(ref);
	}

	return linework;
};

/**
 * Get the linework for drawing the ruler
 *  (use cached version if possible)
 * @returns {Array.<LineworkWrapper>}
 */
Ruler.prototype.getRulerLinework = function() {
	var color = "black";
	if (this.f) {
		color = this.f.getCurrentUser().getColor("PREFERENCES_RULER_DISPLAY_COLOR");
	}

	if (this.startPoint && this.endPoint) {
		if (
			this._cachedRulerLinework &&
			this._cachedRulerLinework.startPoint.equals(this.startPoint) &&
			this._cachedRulerLinework.endPoint.equals(this.endPoint) &&
			this._cachedRulerLinework.color == color
		) {
			// return cached version
			return this._cachedRulerLinework.linework;
		} else {
			return this.generateRulerLinework(this.startPoint, this.endPoint, color);
		}
	}
};

/**
 * Generate the linework for drawing the ruler
 * @param origStartPoint {ShapeUtils.Point} - start point
 * @param origEndPoint {ShapeUtils.Point} - end point
 * @param color {String} - color of dashes
 * @returns {Array.<LineworkWrapper>}
 */
Ruler.prototype.generateRulerLinework = function(origStartPoint, origEndPoint, color) {
	var linework = [];

	var startpoint = new LineworkPoint();
	startpoint.makeMoveTo(origStartPoint.x, origStartPoint.y);

	var endpoint = new LineworkPoint();
	endpoint.makeLineTo(origEndPoint.x, origEndPoint.y);

	// to get the double colored dashed line, we draw a colored dashed line on top of a white full line
	var whiteLine = new LineworkWrapper();
	whiteLine.pointArray.push(startpoint);
	whiteLine.pointArray.push(endpoint);
	whiteLine.color = "white";
	whiteLine.type = DrawUtils.LWTYPE_LINE;
	linework.push(whiteLine);

	var colorLine = new LineworkWrapper();
	colorLine.pointArray.push(startpoint);
	colorLine.pointArray.push(endpoint);
	colorLine.color = color;
	colorLine.dash = [10, 10];
	colorLine.type = DrawUtils.LWTYPE_LINE;
	linework.push(colorLine);

	// generate anchors
	var startPoint = this.getCursor(origStartPoint);
	var endPoint = this.getCursor(origEndPoint);
	linework.push(startPoint);
	linework.push(endPoint);

	// store in cache
	this._cachedRulerLinework = {
		startPoint: origStartPoint.copy(),
		endPoint: origEndPoint.copy(),
		color: color,
		linework: linework
	};

	return linework;
};

/**
 * Generate the crosshair to put at the end points of the ruler
 * @param point {ShapeUtils.Point} - point on which cursor should be applied
 * @returns {LineworkWrapper}
 */
Ruler.prototype.getCursor = function(point) {
	var cursor = new LineworkWrapper();
	cursor.type = DrawUtils.LWTYPE_IMAGE;
	cursor.linewidth = 1;
	cursor.imageRef = MeasureConstants.RulerPoint["DEFAULT"];
	cursor.staticsize = true;
	cursor.setRect(new ShapeUtils.Rect(new ShapeUtils.Point(point.x, point.y)));
	cursor.imageOffset = new ShapeUtils.Size(MeasureConstants.RulerPoint.OFFSET.width, MeasureConstants.RulerPoint.OFFSET.height);
	return cursor;
};

export default Ruler;
