import _ from "underscore";
import $ from "jquery";
import ViewerUtils from "@esko/web-viewer/utils/ViewerUtils";
import ViewportPopoverModel from "@esko/web-viewer/viewport/model/ViewportPopoverModel";
import ShapeUtils from "@esko/shapeutils";
import Template from "@esko/template";
import tpl from "@esko/web-viewer/measure/RulerPopoverTemplate.tpl.html";
import CoreConstants from "@esko/web-viewer/core/CoreConstants";
/**
 * @exports RulerPopoverModel
 * @memberof MeasureModule
 * @constructor
 */
var RulerPopoverModel = function(f, name, ruler, end, cssClass) {
	ViewportPopoverModel.call(this, f, name);

	var div = $("<div/>").addClass("rulerMeasurement");

	this.ruler = ruler;
	this.end = end;
	this.fix = 3;
	this.f = f;
	if (ruler) {
		this.point = ruler.startPoint;
	}

	// initialize template
	this.t = new Template(f.i18n, tpl);

	// load language if needed
	f.i18n.loadBundle(
		"viewer",
		_.bind(function() {
			this.t.render();
		}, this)
	);

	// listen to units change
	this.f.getCurrentUser().on("options.changed.PREFERENCES_UNIT_SIZE", _.bind(this.render, this));

	this.t.setContext(div);

	this.t.bind(
		"mousedown",
		".addAnnotation",
		function(source, event) {
			event.preventDefault();
			event.stopPropagation();
		},
		this
	);

	this.t.bind(
		"click",
		".addAnnotation",
		function(source, event) {
			this._emit("createRulerAnnotation");
			event.preventDefault();
			event.stopPropagation();
		},
		this
	);

	this.t.render();

	this.getFrame = function() {
		if (this.ruler) {
			if (this.end && ruler.endPoint) {
				return new ShapeUtils.Rect(ruler.endPoint.copy(), null);
			} else if (ruler.startPoint) {
				return new ShapeUtils.Rect(ruler.startPoint.copy(), null);
			}
		} else if (this.point) {
			return new ShapeUtils.Rect(this.point.copy(), null);
		}
		return null;
	};

	/**
	 * @override
	 */
	this.getContent = function() {
		return div;
	};

	/**
	 * @override
	 */
	this.getPreferredPosition = function() {
		if (this.ruler.startPoint && this.ruler.endPoint) {
			if (this.ruler.startPoint.x.toFixed(this.fix) == this.ruler.endPoint.x.toFixed(this.fix)) {
				if (this.ruler.startPoint.y <= this.ruler.endPoint.y) {
					return !this.end ? "top" : "bottom";
				} else {
					return !this.end ? "bottom" : "top";
				}
			} else {
				if (this.ruler.startPoint.x <= this.ruler.endPoint.x) {
					return this.end ? "right" : "left";
				} else {
					return this.end ? "left" : "right";
				}
			}
		} else {
			return "bottom";
		}
	};

	/**
	 * @override
	 */
	this.getRestrictBorder = function() {
		return 0;
	};

	/**
	 * Get the context for the popover
	 */
	this.getContext = _.bind(function() {
		return this.f.getContext().find(".viewport");
	}, this);

	/**
	 * Get the class for the popover
	 */
	this.getClass = function() {
		return cssClass;
	};

	/**
	 * Is this an interactive popover
	 * @returns {Boolean} True
	 */
	this.isInteractive = function() {
		return true;
	};
};

// inherit
RulerPopoverModel.prototype = Object.create(ViewportPopoverModel.prototype);
// correct the constructor pointer
RulerPopoverModel.prototype.constructor = RulerPopoverModel;

/**
 * is the popover model empty
 * @returns {Boolean}
 */
RulerPopoverModel.prototype.isEmpty = function() {
	if (!this.end) {
		return !(this.f.getOption("PREFERENCES_RULER_DISPLAY_MODE_XY") || (this.point ? true : false));
	} else {
		return !(
			this.f.getOption("PREFERENCES_RULER_DISPLAY_MODE_DA") ||
			this.f.getOption("PREFERENCES_RULER_DISPLAY_MODE_XY") ||
			this.f.getOption("PREFERENCES_RULER_DISPLAY_MODE_WH")
		);
	}
};

/**
 * Renders the contents of the ruler popover
 */
RulerPopoverModel.prototype.render = function() {
	if (this.ruler) {
		if (!this.ruler.startPoint || !this.ruler.endPoint) {
			return;
		}
	} else {
		if (!this.point) {
			return;
		}
	}

	//add the data for the template to the data object
	_.extend(this.t.d, this.getMeasureData());
	this.t.render();
};

/**
 * Get the measurement information so it can be used by the template
 * @returns {Object}
 */
RulerPopoverModel.prototype.getMeasureData = function() {
	var measureData = {};

	measureData.unitsloc = this.f.getUnitsLocName();
	this.fix = this.f.getUnitsFix();

	var units = this.f.getUnits();

	var rect;
	var resX = null,
		resY = null,
		distanceRes = null;
	if (this.ruler) {
		rect = new ShapeUtils.Rect(this.ruler.startPoint.copy(), this.ruler.startPoint.sizeTo(this.ruler.endPoint));
		if (units === CoreConstants.SizeUnits.PIXELS) {
			var page = this.f.getActivePage();
			resX = ViewerUtils.parseFloat(page.getParam("resolutionx"));
			resY = ViewerUtils.parseFloat(page.getParam("resolutiony"));
		}
	} else if (this.point) {
		rect = new ShapeUtils.Rect(this.point);
	} else {
		rect = new ShapeUtils.Rect();
	}

	measureData.end = this.end;

	if (!this.end) {
		if (this.ruler && !this.ruler.startPoint) {
			return;
		}
		measureData.showXY = this.f.getOption("PREFERENCES_RULER_DISPLAY_MODE_XY") || (this.point ? true : false);
		measureData.showDA = false;
		measureData.showWH = false;

		// convert to correct coordinate system
		var origin = new ShapeUtils.Point(rect.origin.x, rect.origin.y);
		origin = this._coordinates.getTransformedPoint(origin);

		measureData.x = ViewerUtils.convertFromMmValue(origin.x, units, this.fix, resX);
		measureData.y = ViewerUtils.convertFromMmValue(origin.y, units, this.fix, resY);
	} else {
		if (!this.ruler.endPoint) {
			return;
		}
		measureData.createAnnotation = !!this.f.getModule("annotation");
		measureData.showDA = this.f.getOption("PREFERENCES_RULER_DISPLAY_MODE_DA");
		measureData.showXY = this.f.getOption("PREFERENCES_RULER_DISPLAY_MODE_XY");
		measureData.showWH = this.f.getOption("PREFERENCES_RULER_DISPLAY_MODE_WH");

		// convert to correct coordinate system
		var posPoint = new ShapeUtils.Point(rect.right(), rect.bottom());
		posPoint = this._coordinates.getTransformedPoint(posPoint);

		measureData.x = ViewerUtils.convertFromMmValue(posPoint.x, units, this.fix, resX);
		measureData.y = ViewerUtils.convertFromMmValue(posPoint.y, units, this.fix, resY);

		if (units === CoreConstants.SizeUnits.PIXELS) {
			// calculate resolution for diagonal distance
			var angleRad = (this.ruler.getAngle() * Math.PI) / 180;
			distanceRes = Math.sqrt(
				Math.cos(angleRad) * Math.cos(angleRad) * resX * resX + Math.sin(angleRad) * Math.sin(angleRad) * resY * resY
			);
		}
		measureData.distance = ViewerUtils.convertFromMmValue(this.ruler.getLength(), units, this.fix, distanceRes);
		measureData.angle = this.f.getOption("PREFERENCES_RULER_ANGLE_MODE")
			? -this.ruler.getAngle().toFixed(2)
			: this.ruler.getAngle().toFixed(2);

		measureData.width = ViewerUtils.convertFromMmValue(rect.size.width, units, this.fix, resX);
		measureData.height = ViewerUtils.convertFromMmValue(rect.size.height, units, this.fix, resY);
	}

	return measureData;
};

export default RulerPopoverModel;
