import _ from "underscore";
import AbstractMeasureTool from "@esko/web-viewer/measure/AbstractMeasureTool";
import MeasureConstants from "@esko/web-viewer/measure/MeasureConstants";
import ListenerUtil from "@esko/web-viewer/utils/ListenerUtil";
import RulerPopoverModel from "@esko/web-viewer/measure/RulerPopoverModel";
import Template from "@esko/template";
import tpl from "@esko/web-viewer/measure/RulerToAnnotationText.tpl.html";
import AnnotationConstants from "@esko/web-viewer/annotation/AnnotationConstants";
import ShapeUtils from "@esko/shapeutils";
import CoreConstants from "@esko/web-viewer/core/CoreConstants";
/**
 * The RulerTool enables the user to measure a distance on the viewport
 *
 * @class RulerTool the ruler tool is an interactive ruler
 * @memberof Measure
 * @constructor
 * @param {Core.ViewerFacade}
 */
var RulerTool = function RulerTool(f) {
	AbstractMeasureTool.call(this, f, CoreConstants.TOOLS.MEASURE_RULER);
	ListenerUtil.mixin(this, f);

	this.hoverState = "hover";
	this.draggingState = "dragging";
	this.drawnState = "drawn";

	// initialize template used to generate annotation with
	this.t = new Template(f.i18n, tpl);

	this.model.setRulerType(MeasureConstants.MeasureType.Ruler);
	this.model.setAllowStartPointSnapping(true);

	this.model.setDrawType(MeasureConstants.MeasureDrawType.LINE);

	this.startPopover = new RulerPopoverModel(f, "startRulerPopover", this.model.ruler, false, "transparent");
	this.endPopover = new RulerPopoverModel(f, "endRulerPopover", this.model.ruler, true, "transparent");

	this._registerListener("createRulerAnnotation", this._createAnnotation, this, this.endPopover);

	this.getState(this.defaultState).canBeInterrupted = true;

	/**
	 * Drawn State --> Default state after ruler is drawn
	 */
	this.registerState(
		this.drawnState,
		_.defaults(
			{
				// check if we will go to hover mode
				mousemove: _.bind(function(e) {
					if (this._isHovering(e)) {
						this.changeState(this.hoverState);
					}
					this.checkForSnapPoint(e);
				}, this),
				mouseup: _.bind(function(e) {
					if (this.model.ruler.startPoint && !this.model.ruler.endPoint) {
						this.changeState(this.clickState);
					} else {
						this.changeState(this.drawnState);
						if (this.model.ruler.startPoint && this.model.ruler.endPoint) {
							this.calculate();
						}
					}
				}, this),
				mouseleave: _.bind(function(e) {
					this.changeState(this.drawnState);
				}, this),
				canBeInterrupted: true
			},
			this.getState(this.defaultState)
		)
	);

	/**
	 * Hover State --> Mouse is over one of the anchors.
	 * When in this state, we can start dragging anchors around.
	 */
	this.registerState(
		this.hoverState,
		_.defaults(
			{
				// start dragging
				mousedown: _.bind(function(e) {
					var distanceToPoints = this._getDistanceToPoints(e);
					if (distanceToPoints) {
						this._isMovingStartPoint = distanceToPoints.toStartPoint < distanceToPoints.toEndPoint;
						this.changeState("dragging");
					}
				}, this),
				mousemove: _.bind(function(e) {
					if (!this._isHovering(e)) {
						this.changeState(this.drawnState);
					}
					this.checkForSnapPoint(e);
				}, this),

				canBeInterrupted: false,
				canBeInterruptedForNavigation: true
			},
			this.getState(this.drawnState)
		)
	);

	/**
	 * Dragging State --> We're moving one of the end points.
	 */
	this.registerState(
		this.draggingState,
		_.defaults(
			{
				enterState: _.bind(function() {
					this.setCursor("cursor-none");
				}, this),
				// move end points
				mousemove: _.bind(function(e) {
					this.setCursor("cursor-none");
					if (this._isMovingStartPoint) {
						this.registerStartPoint(e);
					} else {
						this.registerEndPoint(e);
					}
					this.checkForSnapPoint(e);
				}, this),
				// when releasing, we are above an end point, so we will go to hover
				mouseup: _.bind(function(e) {
					this.changeState(this.hoverState);
				}, this),

				canBeInterrupted: false,
				canBeInterruptedForNavigation: true
			},
			this.getState(this.drawnState)
		)
	);

	/**
	 * Drawing State --> We're drawing the ruler
	 * Overwrites Drawing state of AbstractMeasureTool
	 */
	this.registerState(
		this.drawingState,
		_.defaults(
			{
				enterState: function() {
					f.emit("layout.disableTextSelection");
					f.emit("viewport.startInteractive");
				},
				leaveState: function() {
					setTimeout(function() {
						f.emit("layout.enableTextSelection");
					}, 500);
					f.emit("viewport.stopInteractive");
				},
				// update ruler
				mousemove: _.bind(function(e) {
					if (this.model.ruler.endPoint) {
						this.setCursor("cursor-none");
					}
					var size = this.keepPointWithinWorldBounds(this.getWorldPoint(e), true).sizeTo(this.touchPoint);
					if (this.model.ruler.endPoint || Math.abs(size.width) >= this.minimum || Math.abs(size.height) > this.minimum) {
						this.registerEndPoint(e);
					}
					this.checkForSnapPoint(e);
				}, this),
				canBeInterrupted: false,
				canBeInterruptedForNavigation: true
			},
			this.getState(this.drawnState)
		)
	);

	/**
	 * Click State --> We clicked, and will draw/update the ruler until we click again
	 */
	this.registerState(
		this.clickState,
		_.defaults(
			{
				enterState: function() {},
				mousedown: _.bind(function(e) {
					this.registerEndPoint(e);
					this.changeState(this.drawnState);
				}, this)
			},
			this.getState(this.drawingState)
		)
	);

	this._override("activate", function(interruptRequest) {
		this._super(interruptRequest);

		this.endPointSet();
	});

	f.getCurrentUser().on(
		"options.changed.PREFERENCES_RULER_ANGLE_MODE",
		_.bind(function(oldValue, newValue) {
			this.startPopover.render();
			this.endPopover.render();
			this.model.repaint();
		}, this)
	);
	f.getCurrentUser().on(
		"options.changed.PREFERENCES_RULER_DISPLAY_MODE",
		_.bind(function(oldValue, newValue) {
			if (this.startPopover.isEmpty()) {
				this.model.removePopoverModel(this.startPopover.getID());
			} else {
				if (this.model.getPopoverModel(this.startPopover.getID()) === undefined) {
					this.model.addPopoverModel(this.startPopover);
				}
				this.startPopover.render();
			}
			if (this.endPopover.isEmpty()) {
				this.model.removePopoverModel(this.endPopover.getID());
			} else {
				if (this.model.getPopoverModel(this.endPopover.getID()) === undefined) {
					this.model.addPopoverModel(this.endPopover);
				}
				this.endPopover.render();
			}
			this.model.repaint();
		}, this)
	);
	f.getCurrentUser().on(
		"options.changed.PREFERENCES_RULER_DISPLAY_COLOR",
		_.bind(function(oldValue, newValue) {
			this.startPopover.render();
			this.endPopover.render();
			this.model.repaint();
		}, this)
	);
};

// extend from the AbstractTool
// inherit
RulerTool.prototype = Object.create(AbstractMeasureTool.prototype);
RulerTool.prototype.constructor = RulerTool;

/**
 * Actions that need to be done when an endpoint is set
 * @override
 */
RulerTool.prototype.endPointSet = _.throttle(function() {
	if (this.model.ruler.startPoint && !this.model.getPopoverModel(this.startPopover.getID()) && !this.startPopover.isEmpty()) {
		this.model.addPopoverModel(this.startPopover);
	}
	if (this.model.ruler.endPoint && !this.model.getPopoverModel(this.endPopover.getID()) && !this.endPopover.isEmpty()) {
		this.model.addPopoverModel(this.endPopover);
	}

	this.startPopover.render();
	this.endPopover.render();
}, 100);

/**
 * Remove popover for start and end point
 * @override
 */
RulerTool.prototype.didReset = function() {
	this.model.removePopoverModel(this.startPopover.getID());
	this.model.removePopoverModel(this.endPopover.getID());
};

/**
 * Remove popover for start and end point
 * @override
 */
RulerTool.prototype.removePopover = function() {
	this.didReset();
};

/**
 * Get a point to snap to, if needed
 * @param {ShapeUtils.Point} refpoint - reference point
 * @param {Number} distance - maximum distance to the mouse to snap
 * @returns {ShapeUtils.Point}
 * @override
 */
RulerTool.prototype.getPointToSnapTo = function(e, distance) {
	if (!this.f.getActivePage().hasCad() && !this.f.getActivePage().hasLinkedPage()) {
		return null;
	}
	var v = this._widget.requestView(e);
	if (v) {
		var refpoint = this.getScenePoint(e);
		var point = v.getClosestLineworkPointFromPoint(refpoint, distance);
		if (point) {
			point = this.sceneToWorld(point);
		}
		return point;
	} else {
		return null;
	}
};

/**
 * Check whether or not we are hovering over the start or end point of the ruler by measuring the distance to the points.
 * @param e
 * @returns {Boolean}
 */
RulerTool.prototype._isHovering = function(e) {
	var distanceToPoints = this._getDistanceToPoints(e);
	if (
		distanceToPoints &&
		(distanceToPoints.toStartPoint < MeasureConstants.RulerPoint.CLOSETODISTANCE ||
			distanceToPoints.toEndPoint < MeasureConstants.RulerPoint.CLOSETODISTANCE)
	) {
		this.setCursor("cursor-moveObject");
		return true;
	} else {
		this.setCursor(this.getDefaultCursor());
		return false;
	}
};

/**
 * Get the distance from the current mouse location to the ruler's end points.
 * It only returns something if the anchors are already defined
 * @param e
 * @returns {Object} - contains distance to start and to end point
 */
RulerTool.prototype._getDistanceToPoints = function(e) {
	var scenePoint = this.getScenePoint(e);
	var worldStartPoint = this.model.ruler.startPoint;
	var worldEndPoint = this.model.ruler.endPoint;

	if (worldStartPoint && worldEndPoint) {
		var startPoint = this._widget.getWorld().getScenePointFromWorld(worldStartPoint);
		var endPoint = this._widget.getWorld().getScenePointFromWorld(worldEndPoint);
		var sizeToStart = scenePoint.sizeTo(startPoint);
		var distToStart = Math.sqrt(Math.abs(sizeToStart.width * sizeToStart.width + sizeToStart.height * sizeToStart.height));
		var sizeToEnd = scenePoint.sizeTo(endPoint);
		var distToEnd = Math.sqrt(Math.abs(sizeToEnd.width * sizeToEnd.width + sizeToEnd.height * sizeToEnd.height));

		return {
			toStartPoint: distToStart,
			toEndPoint: distToEnd
		};
	}

	return undefined;
};

/**
 * Create an annotation out of the ruler. emits an event to request an annotation to be added.
 */
RulerTool.prototype._createAnnotation = function() {
	var startPoint = this.startPopover.getMeasureData();
	this.t.d.fromX = startPoint.x;
	this.t.d.fromY = startPoint.y;
	_.extend(this.t.d, this.endPopover.getMeasureData());

	var content = this.t.getRenderedHtml();
	var rect = new ShapeUtils.Rect(this.model.ruler.startPoint, this.model.ruler.startPoint.sizeTo(this.model.ruler.endPoint));
	this.f.emit("module.annotation.createAnnotation", rect, content, AnnotationConstants.AnnotationType.ANNOT_ARROW);
	this.f.emit("toolManager.deactivateTool");
};

/**
 * @override
 */
RulerTool.prototype.canBeInterrupted = function(interruptor) {
	return (
		this.getCurrentState().canBeInterrupted ||
		(this.getCurrentState().canBeInterruptedForNavigation &&
			interruptor &&
			(interruptor.getIID() === CoreConstants.TOOLS.PAN ||
				interruptor.getIID() === CoreConstants.TOOLS.NAVIGATIONPAN ||
				interruptor.getIID() === CoreConstants.TOOLS.TOUCH ||
				interruptor.getIID() === CoreConstants.TOOLS.ZOOM))
	);
};

export default RulerTool;
