import _ from "underscore";
import $ from "jquery";
import AbstractMeasureTool from "@esko/web-viewer/measure/AbstractMeasureTool";
import MeasureConstants from "@esko/web-viewer/measure/MeasureConstants";
import ScreenRulingPopoverModel from "@esko/web-viewer/measure/ScreenRulingPopoverModel";
import ShapeUtils from "@esko/shapeutils";
import ViewerUtils from "@esko/web-viewer/utils/ViewerUtils";
import CoreConstants from "@esko/web-viewer/core/CoreConstants";
/**
 * Screen ruling tool. With this, you can see the angle and line density of a digital film document.
 *
 * @exports ScreenRulingMeasureTool
 * @memberof MeasureModule
 * @class ScreenRulingMeasureTool
 * @augments Measure.AbstractMeasureTool
 *
 * @param {Core.ViewerFacade} f
 */
var ScreenRulingMeasureTool = function(f) {
	AbstractMeasureTool.call(this, f, CoreConstants.TOOLS.MEASURE_SCREEN_RULING);

	this.model.setRulerType(MeasureConstants.MeasureType.ScreenRuling);

	// store results so it can be rerendered when units change
	this._storedResults = null;

	//override the enterState event for the clickState
	this.registerState(
		this.clickState,
		_.defaults(
			{
				enterState: _.bind(function() {
					var worldResolution = this.getWorldResolution();
					if (worldResolution !== null) {
						var sizeX = MeasureConstants.ClickSize.DIGITAL_FILM / worldResolution.x; // square of 64 world pixels
						var sizeY = MeasureConstants.ClickSize.DIGITAL_FILM / worldResolution.y;
						this.model.setStartPoint(
							new ShapeUtils.Point(this.model.ruler.startPoint.x - sizeX / 2, this.model.ruler.startPoint.y - sizeY / 2)
						);
						this.model.setEndPoint(new ShapeUtils.Point(this.model.ruler.startPoint.x + sizeX, this.model.ruler.startPoint.y + sizeY));
						this.calculate();
						this.changeState(this.defaultState);
					}
				}, this)
			},
			this.getState(this.drawingState)
		)
	);

	// rerender when unit settings change
	this.f.getCurrentUser().on("options.changed.PREFERENCES_UNIT_SCREENRULING", _.bind(this.showInfo, this, null));
	this.f.getCurrentUser().on("options.changed.PREFERENCES_UNIT_SCREEN", _.bind(this.showInfo, this, null));
	this.f.getCurrentUser().on("options.changed.PREFERENCES_SCREEN_ORIENTATION", _.bind(this.showInfo, this, null));
};

// extend from the AbstractTool
ScreenRulingMeasureTool.prototype = Object.create(AbstractMeasureTool.prototype);
ScreenRulingMeasureTool.prototype.constructor = ScreenRulingMeasureTool;

/**
 * @override
 * Can the tool be shown for this page?
 * @returns {Boolean} false
 */
ScreenRulingMeasureTool.prototype.canBeUsedOnActivePage = function() {
	var page = this.f.getActivePage();
	return page && page.hasGraphics() && page.isDigitalFilmPage();
};

/**
 * @override
 * Does this tool need a total count
 * @returns {Boolean} false
 */
ScreenRulingMeasureTool.prototype._needsTotalCount = function() {
	return false;
};

/**
 * @override
 * Sets the required options for the call to the viewserver
 */
ScreenRulingMeasureTool.prototype.calculate = function() {
	this.getInfo({});
};

/**
 * @override
 * Gets the information from the to object converted xml and creates a MeasurePopoverModel that shows the information and tells the AbstractModel to add the popover
 * @param {Object} object
 */
ScreenRulingMeasureTool.prototype.showInfo = function(object) {
	var channels = [];

	// store object for rerendering when units change
	if (object) {
		this._storedResults = object;
	} else {
		object = this._storedResults;
	}

	// read modes from preferences
	var mode = this.f.getOption("PREFERENCES_UNIT_SCREEN");
	var orientation = this.f.getOption("PREFERENCES_SCREEN_ORIENTATION");

	if (object) {
		if (object.status.info.screen) {
			if (!_.isArray(object.status.info.screen)) {
				object.status.info.screen = [object.status.info.screen];
			}

			_.each(
				object.status.info.screen,
				_.bind(function(channel) {
					var channelInfo = "";
					if (channel.found === "1") {
						if (mode === MeasureConstants.ScreenRulingMode.LINES || mode === MeasureConstants.ScreenRulingMode.BOTH) {
							channelInfo += ViewerUtils.convertResolutionFromPpmmValue(
								ViewerUtils.parseFloat(channel.ruling),
								this.f.getScreenUnits(),
								this.f.getScreenUnitsFix()
							);
							channelInfo += " " + this.f.i18n.t(this.f.getScreenUnitsLocName());
						}
						if (mode === MeasureConstants.ScreenRulingMode.BOTH) {
							channelInfo += " / ";
						}
						if (mode === MeasureConstants.ScreenRulingMode.ANGLE || mode === MeasureConstants.ScreenRulingMode.BOTH) {
							var angle = ViewerUtils.parseFloat(channel.angle);
							if (orientation === MeasureConstants.ScreenRulingOrientation.CCW) {
								angle = 90 - angle;
							}
							channelInfo += angle.toFixed(2) + "°";
						}
					}
					channels[channel.color] = channelInfo;
				}, this)
			);
		}
		this.f.emit("channels.informationAdded", channels);
	}
	this.showPopover(channels);
};

/**
 * @override
 */
ScreenRulingMeasureTool.prototype.createPopoverModel = function(channelInfo, area) {
	return new ScreenRulingPopoverModel(
		this.f,
		area,
		"Channels",
		$("<i>").addClass("icon icon-normal icon-WCR_Measure-Screen"),
		channelInfo,
		this._needsTotalCount()
	);
};

export default ScreenRulingMeasureTool;
