import DropdownMiniButton from "@esko/web-viewer/layout/component/DropdownMiniButton";
import CoreConstants from "@esko/web-viewer/core/CoreConstants";
import MeasureConstants from "@esko/web-viewer/measure/MeasureConstants";
import _ from "underscore";
/**
 * @exports ScreenRulingUnitSelector
 * @class The ScreenRulingUnitSelector has a button that allows setting the unit for screen ruling
 * @constructor
 * @param {Core.ViewerFacade} f
 */
var ScreenRulingUnitSelector = function(f) {
	this.f = f;

	/**
	 * Make the options dropdown
	 */
	this._optionButton = new DropdownMiniButton(f, "screenRulingUnitSettings", "icon-WCR_Settings-submenu");

	f.i18n.loadBundle(
		"viewer",
		function() {
			// Units for size
			function toggleScreenRulingUnits(unit, mode, toggle) {
				if (toggle) {
					this.setUnitAndMode(unit, mode);
				}
			}

			this._optionButton.addOption({
				name: "unitLpmm",
				content: this.f.i18n.t("UnitScreenmm"),
				onchange: _.bind(toggleScreenRulingUnits, this, CoreConstants.ResolutionUnits.PPMM, MeasureConstants.ScreenRulingMode.LINES),
				userCannotUncheck: true
			});

			this._optionButton.addOption({
				name: "unitLpcm",
				content: this.f.i18n.t("UnitScreencm"),
				onchange: _.bind(toggleScreenRulingUnits, this, CoreConstants.ResolutionUnits.PPCM, MeasureConstants.ScreenRulingMode.LINES),
				userCannotUncheck: true
			});

			this._optionButton.addOption({
				name: "unitLpi",
				content: this.f.i18n.t("UnitScreenInch"),
				onchange: _.bind(toggleScreenRulingUnits, this, CoreConstants.ResolutionUnits.PPI, MeasureConstants.ScreenRulingMode.LINES),
				userCannotUncheck: true
			});

			this._optionButton.addOption({
				name: "unitDeg",
				content: this.f.i18n.t("UnitDegree"),
				onchange: _.bind(toggleScreenRulingUnits, this, 0, MeasureConstants.ScreenRulingMode.ANGLE),
				userCannotUncheck: true
			});

			this._optionButton.addOption({
				name: "unitLpmmDeg",
				content: this.f.i18n.t("UnitDegreemm"),
				onchange: _.bind(toggleScreenRulingUnits, this, CoreConstants.ResolutionUnits.PPMM, MeasureConstants.ScreenRulingMode.BOTH),
				userCannotUncheck: true
			});

			this._optionButton.addOption({
				name: "unitLpcmDeg",
				content: this.f.i18n.t("UnitDegreecm"),
				onchange: _.bind(toggleScreenRulingUnits, this, CoreConstants.ResolutionUnits.PPCM, MeasureConstants.ScreenRulingMode.BOTH),
				userCannotUncheck: true
			});

			this._optionButton.addOption({
				name: "unitLpiDeg",
				content: this.f.i18n.t("UnitDegreeInch"),
				onchange: _.bind(toggleScreenRulingUnits, this, CoreConstants.ResolutionUnits.PPI, MeasureConstants.ScreenRulingMode.BOTH),
				userCannotUncheck: true
			});

			this._optionButton.addDivider();

			// angle CW or CCW
			function toggleScreenOrientation(mode, toggle) {
				if (toggle) {
					this.setOrientation(mode);
				}
			}

			this._optionButton.addOption({
				name: "scrOrientationCW",
				content: this.f.i18n.t("ScreenCW"),
				onchange: _.bind(toggleScreenOrientation, this, MeasureConstants.ScreenRulingOrientation.CW),
				userCannotUncheck: true
			});

			this._optionButton.addOption({
				name: "scrOrientationCCW",
				content: this.f.i18n.t("ScreenCCW"),
				onchange: _.bind(toggleScreenOrientation, this, MeasureConstants.ScreenRulingOrientation.CCW),
				userCannotUncheck: true
			});

			this._updateUnitOptions();
			this._updateOrientationOptions();
			this._optionButton.setTooltip("viewer", "UnitSettings");

			this.f.getCurrentUser().on("options.changed.PREFERENCES_UNIT_SCREENRULING", _.bind(this._updateUnitOptions, this));
			this.f.getCurrentUser().on("options.changed.PREFERENCES_UNIT_SCREEN", _.bind(this._updateUnitOptions, this));
			this.f.getCurrentUser().on("options.changed.PREFERENCES_SCREEN_ORIENTATION", _.bind(this._updateOrientationOptions, this));
		},
		this
	);
};

/**
 * Get the button element
 * @returns {JQueryObj} element of the button
 */
ScreenRulingUnitSelector.prototype.getButtonElement = function() {
	return this._optionButton.element;
};

/**
 * Set the unit and mode. The unit itself is not persisted, but the mode is.
 * @param {CoreConstants.ResolutionUnits} unit
 * @param {MeasureConstants.ScreenRulingMode} mode
 */
ScreenRulingUnitSelector.prototype.setUnitAndMode = function(unit, mode) {
	// set options and send out events
	if (this.f.getOption("PREFERENCES_UNIT_SCREENRULING") !== unit) {
		this.f.setOption("PREFERENCES_UNIT_SCREENRULING", unit, true);
	}
	if (this.f.getOption("PREFERENCES_UNIT_SCREEN") !== mode) {
		this.f.setOption("PREFERENCES_UNIT_SCREEN", mode, false);
	}
};

/**
 * Update the toggle options of the button for units
 */
ScreenRulingUnitSelector.prototype._updateUnitOptions = function() {
	var unit = this.f.getOption("PREFERENCES_UNIT_SCREENRULING");
	var mode = this.f.getOption("PREFERENCES_UNIT_SCREEN");

	this._optionButton.toggleOption(
		"unitLpmm",
		unit === CoreConstants.ResolutionUnits.PPMM && mode === MeasureConstants.ScreenRulingMode.LINES
	);
	this._optionButton.toggleOption(
		"unitLpcm",
		unit === CoreConstants.ResolutionUnits.PPCM && mode === MeasureConstants.ScreenRulingMode.LINES
	);
	this._optionButton.toggleOption(
		"unitLpi",
		unit === CoreConstants.ResolutionUnits.PPI && mode === MeasureConstants.ScreenRulingMode.LINES
	);
	this._optionButton.toggleOption("unitDeg", mode === MeasureConstants.ScreenRulingMode.ANGLE);
	this._optionButton.toggleOption(
		"unitLpmmDeg",
		unit === CoreConstants.ResolutionUnits.PPMM && mode === MeasureConstants.ScreenRulingMode.BOTH
	);
	this._optionButton.toggleOption(
		"unitLpcmDeg",
		unit === CoreConstants.ResolutionUnits.PPCM && mode === MeasureConstants.ScreenRulingMode.BOTH
	);
	this._optionButton.toggleOption(
		"unitLpiDeg",
		unit === CoreConstants.ResolutionUnits.PPI && mode === MeasureConstants.ScreenRulingMode.BOTH
	);
};

/**
 * Set the screen orientation.
 * @param {MeasureConstants.ScreenRulingOrientation} mode
 */
ScreenRulingUnitSelector.prototype.setOrientation = function(mode) {
	// set option and send out events
	this.f.setOption("PREFERENCES_SCREEN_ORIENTATION", mode, false);
};

/**
 * Update the toggle options of the button for orientation
 */
ScreenRulingUnitSelector.prototype._updateOrientationOptions = function() {
	var orientation = this.f.getOption("PREFERENCES_SCREEN_ORIENTATION");

	this._optionButton.toggleOption("scrOrientationCW", orientation === MeasureConstants.ScreenRulingOrientation.CW);
	this._optionButton.toggleOption("scrOrientationCCW", orientation === MeasureConstants.ScreenRulingOrientation.CCW);
};

export default ScreenRulingUnitSelector;
