/**
 * @namespace MessageModule
 */
/**
 * @namespace MessageModule
 */
import AbstractModule from "@esko/web-viewer/core/AbstractModule";

import MessageWidgetWrapper from "@esko/web-viewer/message/MessageWidgetWrapper";
import MessageComponent from "@esko/web-viewer/message/MessageComponent";
import _ from "underscore";
/**
 * EVENTS
 *
 * userMessage
 *
 * 		Listen to user messages to show.
 *
 * 		@param {Message|String} message - Message to show. If message is string, type is info.
 * 		OR
 * 		@param {String} type - Type of the message (error|warn|info|success).
 * 		@param {String} message - Message to show.
 * 		OR
 * 		@param {String} type - Type of the message (error|warn|info|success).
 * 		@param {String} message - Message to show.
 * 		@param {Boolean} autoHide - True if the message should be auto hidden, false otherwise
 * 		OR
 * 		@param {String} type - Type of the message (error|warn|info|success).
 * 		@param {String} message - Message to show.
 * 		@param {Function | String} action - Function executed when clicking on the message or if string URL accessed when clicking on the message.
 * 		OR
 * 		@param {String} type - Type of the message (error|warn|info|success).
 * 		@param {String} message - Message to show.
 * 		@param {Function | String} action - Function executed when clicking on the message or if string URL accessed when clicking on the message.
 * 		@param {Boolean} autoHide - True if the message should be auto hidden, false otherwise
 * 		OR
 * 		@param {String} type - Type of the message (error|warn|info|success).
 * 		@param {String} message - Message to show.
 *		@param {Function | String} action - Function executed when clicking on the message or if string URL accessed when clicking on the message.
 * 		@param {Boolean} autoHide - True if the message should be auto hidden, false otherwise
 * 		@param {String} tag - this is a label for the message, it can be used to make all messages with a certain label obsolete
 * 		OR
 * 		@param {String} type - Type of the message (error|warn|info|success).
 * 		@param {String} message - Message to show.
 * 		@param {Function | String} action - Function executed when clicking on the message or if string URL accessed when clicking on the message.
 * 		@param {Boolean} autoHide - True if the message should be auto hidden, false otherwise
 *		@param {String} tag - this is a label for the message, it can be used to make all messages with a certain label obsolete
 * 		@param {Boolean} doNotAddIfTagExists - do not add the message if another message with the same tag is being shown or is in the queue
 *
 *
 * userMessage.markObsolete
 * 		Mark all messages having a certain tag as obsolete
 * 		@param {String} tag
 *
 * error.SEVERE
 *
 * 		Listen to severe error messages to show
 * 		@param {String} message - Message of the error
 *
 *
 * error.WARNING
 *
 * 		Listen to warning error messages to show
 * 		@param {String} message - Message of the warning
 */

/**
 * MessageModule is responsible for notifying the user on certain events (errors, warnings, info etc).
 * @exports MessageModule
 * @memberof MessageModule
 * @constructor
 * @augments Core.AbstractModule
 * @param {Core.ViewerFacade} f
 * @param {Boolean} silenced
 */
var MessageModule = function MessageModule(f, silenced) {
	AbstractModule.call(this, f, silenced, "message");

	this._messageModel = new MessageComponent.MessageModel();

	// Message widget
	this._publishWidget(
		MessageWidgetWrapper.widgetName,
		_.bind(function(layoutType) {
			this._messageWidgetWrapper = new MessageWidgetWrapper(f, this._messageModel);
			return this._messageWidgetWrapper;
		}, this)
	);

	/**
	 * Listen to userMessage events
	 */
	this.f.on("userMessage", _.bind(this._handleMessageEvent, this));

	/**
	 * Listen to markAsObsolete events
	 */
	this.f.on("userMessage.markObsolete", _.bind(this._markObsolete, this));

	/**
	 * Listen to severe error messages to show
	 * @param {String} message - Message of the error
	 */
	this.f.on(
		"error.SEVERE",
		_.bind(function(message) {
			this._handleMessageEvent("error", message);
		}, this)
	);

	/**
	 * Listen to warning error messages to show
	 * @param {String} message - Message of the warning
	 * @param {Boolean} shouldAutoHide - should the warning message be auto hidden
	 */
	this.f.on(
		"error.WARNING",
		_.bind(function(message, shouldAutoHide) {
			this._handleMessageEvent("warn", message, undefined, shouldAutoHide);
		}, this)
	);
};

MessageModule.prototype = Object.create(AbstractModule.prototype);
MessageModule.prototype.constructor = MessageModule;

/**
 * Add to Message Queue and alert the widget that we have a new message
 * @param {Message} message - Message to add to the queue
 */
MessageModule.prototype._addToMessageQueue = function(message) {
	this.f.debug.log("Message received", message);
	this._messageModel.addToMessageQueue(message);
};

/**
 * @param {Message|String} message - Message to show. If message is string, type is info.
 * OR
 * @param {String} type - Type of the message (error|warn|info|success).
 * @param {String} message - Message to show.
 * OR
 * @param {String} type - Type of the message (error|warn|info|success).
 * @param {String} message - Message to show.
 * @param {Boolean} autoHide - True if the message should be auto hidden, false otherwise
 * OR
 * @param {String} type - Type of the message (error|warn|info|success).
 * @param {String} message - Message to show.
 * @param {Function | String} action - Function executed when clicking on the message or if string URL accessed when clicking on the message.
 * OR
 * @param {String} type - Type of the message (error|warn|info|success).
 * @param {String} message - Message to show.
 * @param {Function | String} action - Function executed when clicking on the message or if string URL accessed when clicking on the message.
 * @param {Boolean} autoHide - True if the message should be auto hidden, false otherwise
 * OR
 * @param {String} type - Type of the message (error|warn|info|success).
 * @param {String} message - Message to show.
 * @param {Function | String} action - Function executed when clicking on the message or if string URL accessed when clicking on the message.
 * @param {Boolean} autoHide - True if the message should be auto hidden, false otherwise
 * @param {String} tag - this is a label for the message, it can be used to make all messages with a certain label obsolete
 * OR
 * @param {String} type - Type of the message (error|warn|info|success).
 * @param {String} message - Message to show.
 * @param {Function | String} action - Function executed when clicking on the message or if string URL accessed when clicking on the message.
 * @param {Boolean} autoHide - True if the message should be auto hidden, false otherwise
 * @param {String} tag - this is a label for the message, it can be used to make all messages with a certain label obsolete
 * @param {Boolean} doNotAddIfTagExists - do not add the message if another message with the same tag is being shown or is in the queue
 */
MessageModule.prototype._handleMessageEvent = function() {
	var nrOfArgs = arguments.length;
	var msgSpan = document.createElement("span");
	if (_.isString(arguments[0])) {
		msgSpan.dataset.sahiId = "web-viewer-message-" + arguments[0].toLowerCase();
	} else {
		msgSpan.dataset.sahiId = "'web-viewer-message-" + arguments[0].name.toLowerCase();
	}
	msgSpan.innerHTML = arguments[1];
	if (nrOfArgs === 1) {
		if (_.isString(arguments[0])) {
			msgSpan.dataset.sahiId = "web-viewer-message-" + arguments[0].toLowerCase();
			msgSpan.innerHTML = arguments[0];
			this._addToMessageQueue(new MessageComponent.Message("info", msgSpan));
		} else {
			this._addToMessageQueue(arguments[0]);
		}
		return;
	} else if (nrOfArgs === 2) {
		this._addToMessageQueue(new MessageComponent.Message(arguments[0], msgSpan));
	} else if (nrOfArgs === 3) {
		if (_.isBoolean(arguments[2])) {
			this._addToMessageQueue(new MessageComponent.Message(arguments[0], msgSpan, false, arguments[2]));
		} else {
			this._addToMessageQueue(new MessageComponent.Message(arguments[0], msgSpan, arguments[2], false));
		}
	} else if (nrOfArgs === 4) {
		this._addToMessageQueue(new MessageComponent.Message(arguments[0], msgSpan, arguments[2], arguments[3]));
	} else if (nrOfArgs === 5) {
		this._addToMessageQueue(new MessageComponent.Message(arguments[0], msgSpan, arguments[2], arguments[3], arguments[4]));
	} else if (nrOfArgs === 6) {
		var tag = arguments[4];
		var doNotAddIfTagExists = arguments[5];
		// first check no message with the same tag exists in the queue or is shown in the widget
		if (
			!doNotAddIfTagExists ||
			(!_.some(this._messageQueue, function(msg) {
				return msg.tag === tag;
			}) &&
				(!this._messageWidget || !this._messageWidget.isShownMessageWithTag(tag)))
		) {
			this._addToMessageQueue(new MessageComponent.Message(arguments[0], msgSpan, arguments[2], arguments[3], arguments[4]));
		}
	} else {
		// unrecognized event
		this.f.debug.debug("Unrecognized userMessage event", arguments);
	}
};

/**
 * Mark all messages with "tag" as tag, obsolete
 * @param tag
 */
MessageModule.prototype._markObsolete = function(tag) {
	this._messageModel.markObsolete(tag);
};

export default MessageModule;
