/**
 * @namespace NavigationModule
 */
/**
 * @namespace NavigationModule
 */
import _ from "underscore";

import AbstractModule from "@esko/web-viewer/core/AbstractModule";
import PanTool from "@esko/web-viewer/navigation/PanTool";
import TouchTool from "@esko/web-viewer/navigation/TouchTool";
import TouchPreventerTool from "@esko/web-viewer/navigation/TouchPreventerTool";
import ZoomTool from "@esko/web-viewer/navigation/ZoomTool";
import NavigatorWidget from "@esko/web-viewer/navigation/NavigatorWidget";
import NavigationConstants from "@esko/web-viewer/navigation/NavigationConstants";
import AdjustMonitorResolutionPopup from "@esko/web-viewer/navigation/AdjustMonitorResolutionPopup";
import ToolbarComponent from "@esko/toolbar/ToolbarComponent";
import ButtonGroup from "@esko/toolbar/ButtonGroup";
import ViewerToolbarButton from "@esko/web-viewer/layout/component/ViewerToolbarButton";
import DropdownToolbarButton from "@esko/web-viewer/layout/component/DropdownToolbarButton";
import LayoutConstants from "@esko/web-viewer/layout/LayoutConstants";
import ViewerUtils from "@esko/web-viewer/utils/ViewerUtils";
import Mousetrap from "@esko/mousetrap";
/**
 * Create the NavigationModule
 *
 * @exports NavigationModule
 * @memberof NavigationModule
 * @class NavigationModule - responsible for publishing the correct tools and toolbarcomponents enabling navigation
 * @augments Core.AbstractModule
 * @param {Core.ViewerFacade} f
 * @param {Boolean} silenced
 */
var NavigationModule = function(f, silenced, overrideClasses) {
	AbstractModule.call(this, f, silenced, "navigation");

	this.requiresViewportAccess = true;

	this._extraRotation = 0;

	this._initializeTools(f);

	let NavigatorWidgetImplementation = NavigatorWidget;
	if (overrideClasses && overrideClasses.NavigatorWidget) {
		NavigatorWidgetImplementation = overrideClasses.NavigatorWidget;
	}

	this._publishWidget(
		NavigatorWidgetImplementation.widgetName,
		_.bind(function(layoutType) {
			this._navigatorWidget = new NavigatorWidgetImplementation(f);
			return this._navigatorWidget;
		}, this)
	);
};

// inherit
NavigationModule.prototype = Object.create(AbstractModule.prototype);
// correct the constructor pointer
NavigationModule.prototype.constructor = NavigationModule;

/**
 * Initializes the tools, depending on the layout type.
 * @param {Core.ViewerFacade} f
 */
NavigationModule.prototype._initializeTools = function(f) {
	var layoutType = this.f.getLayoutType();

	if (layoutType === LayoutConstants.LayoutType.MOBILE) {
		var touchTool = new TouchTool(f);
		this._publishTool(touchTool);
	} else {
		var panTool = new PanTool(f);
		var zoomTool = new ZoomTool(f);
		var noTouchTool = new TouchPreventerTool(f);
		this._publishTool(zoomTool);
		this._publishTool(panTool);
		this._publishTool(noTouchTool);

		this._publishToolbarComponent(
			LayoutConstants.ToolbarComponent.NAVIGATE,
			_.bind(function(layoutType) {
				/**
				 * Component: Navigate
				 */
				var navigateComponent = new ToolbarComponent(LayoutConstants.ToolbarComponent.NAVIGATE, f);

				/**
				 * Group: zoom + pan
				 */
				var navigateGroup = new ButtonGroup("NavigateGroup");
				navigateGroup.setTitle(f.i18n, "viewer", "NavBtnGroupTitle");
				navigateComponent.addButtonGroup(navigateGroup);

				// Zoom button
				var zoomButton = new ViewerToolbarButton(f, "ZoomButton", "icon icon-WCR_Zoom", false);
				zoomButton.setTooltip("viewer", "ZoomBtnTT");
				zoomButton.listenToTool(zoomTool, true);
				zoomButton.element.on("dblclick", function(e) {
					f.emit("viewport.zoomToFit", {
						animated: true
					});
					e.stopPropagation();
				});
				navigateGroup.addButton(zoomButton);

				// ZoomIn button
				var zoomInButton = new ViewerToolbarButton(f, "ZoomInButton", "icon icon-WCR_WF-ZoomIn", false, function() {
					f.emit("viewport.transform", {
						zoom: 1,
						animated: true
					});
				});
				zoomInButton.setTooltip("viewer", "ZoomInCursor");
				navigateGroup.addButton(zoomInButton);

				// ZoomOut button
				var zoomOutButton = new ViewerToolbarButton(f, "ZoomOutButton", "icon icon-WCR_WF-ZoomOut", false, function() {
					f.emit("viewport.transform", {
						zoom: -1,
						animated: true
					});
				});
				zoomOutButton.setTooltip("viewer", "ZoomOutCursor");
				navigateGroup.addButton(zoomOutButton);

				// ZoomToFit button
				var zoomToFitButton = new DropdownToolbarButton(f, "ZoomToFitButton", "icon icon-WCR_WF-ZoomFit", false, function() {
					f.emit("viewport.zoomToFit", {
						animated: true
					});
				});
				zoomToFitButton.setTooltip("viewer", "ZoomFitToPage");
				navigateGroup.addButton(zoomToFitButton);

				// ZoomOneToOne button
				this._zoomOneToOneButton = new DropdownToolbarButton(
					f,
					"ZoomOneToOne",
					"icon icon-WCR_Zoom1",
					false,
					_.bind(this._zoomOneToOne, this)
				);
				navigateGroup.addButton(this._zoomOneToOneButton);

				// Pan button
				var panButton = new ViewerToolbarButton(f, "PanButton", "icon icon-WCR_Pan", false);
				panButton.setTooltip("viewer", "PanBtnTT");
				panButton.listenToTool(panTool, true);
				navigateGroup.addButton(panButton);

				/**
				 * Group: rotate + flip
				 */
				var rotFlipGroup = new ButtonGroup("RotFlipGroup");
				rotFlipGroup.setTitle(f.i18n, "viewer", "RotateBtnGroupTitle");
				navigateComponent.addButtonGroup(rotFlipGroup);

				// Rotate button
				this._rotateButton = new ViewerToolbarButton(
					f,
					"RotateButton",
					"icon icon-WCR_Orientation-Rotate1",
					true,
					_.bind(function(event) {
						var rotation = 90;
						if (event.shiftKey) {
							rotation = -90;
						}
						this.f.emit("viewport.transform", {
							rotate: rotation,
							animated: {
								interrupt: true
							}
						});
					}, this)
				);
				rotFlipGroup.addButton(this._rotateButton);

				// Flip button
				this._flipButton = new ViewerToolbarButton(
					f,
					"FlipButton",
					"icon icon-WCR_Orientation-Mirror",
					true,
					_.bind(function() {
						this._doToggleFlip();
					}, this)
				);
				rotFlipGroup.addButton(this._flipButton);

				this.f.loadBundle(
					"viewer",
					function() {
						/**
						 * Zoom to Fit: submenu
						 */
						// zoomFitWidth
						zoomToFitButton.addAction({
							name: "zoomFitWidth",
							content: f.i18n.t("ZoomFitWidth"),
							icon: "icon icon-WCR_Zoom-FitWidth",
							action: function() {
								f.emit("viewport.zoomToFitWidth", {
									animated: true
								});
							}
						});
						// zoomFitWHeight
						zoomToFitButton.addAction({
							name: "zoomFitHeight",
							content: f.i18n.t("ZoomFitHeight"),
							icon: "icon icon-WCR_Zoom-FitHeight",
							action: function() {
								f.emit("viewport.zoomToFitHeight", {
									animated: true
								});
							}
						});

						/**
						 * Zoom One To One: submenu
						 */
						this._zoomOneToOneButton.addAction({
							name: "zoomOneToOnePixel",
							content: f.i18n.t("ZoomOneToOnePixel"),
							icon: "icon icon-WCR_ZoomPx",
							action: _.bind(this._zoomOneToOnePixel, this)
						});
						this._zoomOneToOneButton.addAction({
							name: "zoomOneToOneRealSize",
							content: f.i18n.t("ZoomOneToOneRealSize"),
							icon: "icon icon-WCR_Zoom1",
							action: _.bind(this._zoomOneToOneRealSize, this)
						});
						this._zoomOneToOneButton.addDivider();
						this._zoomOneToOneButton.addAction({
							name: "adjustMonitorResolution",
							content: f.i18n.t("ZoomOneToOneAdjustRes"),
							action: _.bind(this._adjustMonitorResolution, this)
						});

						this._updateZoomOneToOneButton();

						// for further updates
						this.f.getCurrentUser().on("options.changed.PREFERENCES_ZOOM_REALSIZE", _.bind(this._updateZoomOneToOneButton, this));
						this.f.getDocumentController().whenActivePageChanged(_.bind(this._updateZoomOneToOneButton, this));

						this._updateRotateButton();
						this.f.on("transform.viewportWidget.rotated", _.bind(this._updateRotateButton, this));

						this._updateFlipButton();
						this.f.on("transform.viewportWidget.flipped", _.bind(this._updateFlipButton, this));
					},
					this
				);

				return navigateComponent;
			}, this)
		);

		// Keyboard shortcuts
		Mousetrap.bind(
			"ctrl+m",
			_.bind(function(e) {
				this._doToggleFlip();

				e.preventDefault();
				e.stopPropagation();
			}, this),
			"keydown"
		);

		Mousetrap.bind(
			["ctrl+alt+plus", "ctrl+alt+="],
			_.bind(function(e) {
				var rotation = 90;

				this.f.emit("viewport.transform", {
					rotate: rotation,
					animated: {
						interrupt: true
					}
				});

				e.preventDefault();
				e.stopPropagation();
			}, this)
		);

		Mousetrap.bind(
			"ctrl+alt+-",
			_.bind(function(e) {
				var rotation = -90;

				this.f.emit("viewport.transform", {
					rotate: rotation,
					animated: {
						interrupt: true
					}
				});

				e.preventDefault();
				e.stopPropagation();
			}, this)
		);
	}
};

/**
 * Toggle the flip state of the viewport
 */
NavigationModule.prototype._doToggleFlip = function() {
	var viewportModule = this.f.getModule("viewport");
	if (viewportModule) {
		var rotation = viewportModule.getRotation();
		if (rotation % 180 === 0) {
			// not rotated
			this.f.emit("viewport.transform", {
				flip: true,
				animated: false
			});
		} else {
			// rotated
			this.f.emit("viewport.transform", {
				rotate: 180,
				flip: true,
				animated: false
			});
		}
	}
};

/**
 * Update the rotation button on rotate events
 */
NavigationModule.prototype._updateRotateButton = function() {
	var rotationClasses = [
		"icon icon-WCR_Orientation-Rotate1",
		"icon icon-WCR_Orientation-Rotate2",
		"icon icon-WCR_Orientation-Rotate3",
		"icon icon-WCR_Orientation-Rotate4"
	];
	var rotationTTs = ["Rotate90TT", "Rotate180TT", "Rotate270TT", "Rotate0TT"];
	var viewportModule = this.f.getModule("viewport");
	if (viewportModule && this._rotateButton) {
		var rotation = viewportModule.getRotation() + this._extraRotation;
		var i = (rotation / 90) % 4;
		this._rotateButton.setIcon(rotationClasses[i]);
		this._rotateButton.setTooltip("viewer", rotationTTs[i]);
	}
};

/**
 * Set an extra rotation to the icon in the top toolbar
 * @param {Number} extraRotation
 */
NavigationModule.prototype.setExtraRotation = function(extraRotation) {
	this._extraRotation = extraRotation;
	this._updateRotateButton();
};

/**
 * Reset the extra rotation for the rotation icon
 */
NavigationModule.prototype.resetExtraRotation = function() {
	this._extraRotation = 0;
	this._updateRotateButton();
};

/**
 * Update the flip button on flip events
 */
NavigationModule.prototype._updateFlipButton = function() {
	var viewportModule = this.f.getModule("viewport");
	if (viewportModule && this._flipButton) {
		if (viewportModule.isFlip()) {
			// on
			this._flipButton.setIcon("icon icon-WCR_Orientation-MirrorReverse");
			this._flipButton.setTooltip("viewer", "MirrorTT2");
		} else {
			// off
			this._flipButton.setIcon("icon icon-WCR_Orientation-Mirror");
			this._flipButton.setTooltip("viewer", "MirrorTT");
		}
	}
};

/**
 * Check whether we have to zoom one to one pixel or not
 * @returns {Boolean}
 */
NavigationModule.prototype._isZoomOneToOnePixel = function() {
	// only valid for graphics and if we need to
	return (
		this.f.getActivePage().hasGraphics() && this.f.getOption("PREFERENCES_ZOOM_REALSIZE") === NavigationConstants.ZoomOneToOneClick.AUTO
	);
};

/**
 * Called when the zoom one to one button behavior should be updated
 */
NavigationModule.prototype._updateZoomOneToOneButton = function() {
	if (this._zoomOneToOneButton) {
		if (this._isZoomOneToOnePixel()) {
			this._zoomOneToOneButton.setTooltip("viewer", "ZoomOneToOnePixel");
			this._zoomOneToOneButton.setIcon("icon icon-WCR_ZoomPx");
		} else {
			this._zoomOneToOneButton.setTooltip("viewer", "ZoomOneToOneRealSize");
			this._zoomOneToOneButton.setIcon("icon icon-WCR_Zoom1");
		}
		if (this.f.getActivePage().hasGraphics()) {
			this._zoomOneToOneButton.setActionEnabled("zoomOneToOnePixel", true);
		} else {
			this._zoomOneToOneButton.setActionEnabled("zoomOneToOnePixel", false);
		}
	}
};

/**
 * Called when user wants to adjust monitor resolution
 */
NavigationModule.prototype._adjustMonitorResolution = function() {
	if (this._adjustMonitorResolutionPopup === undefined) {
		this._adjustMonitorResolutionPopup = new AdjustMonitorResolutionPopup(this.f, "adjustMonitorResulutionPopup", this);
	}
	this._adjustMonitorResolutionPopup.show();
};

/**
 * Zoom 1:1: called when toolbar button clicked
 */
NavigationModule.prototype._zoomOneToOne = function() {
	if (this._isZoomOneToOnePixel()) {
		this._zoomOneToOnePixel();
	} else {
		this._zoomOneToOneRealSize();
	}
};

/**
 * Zoom 1:1 Pixel: called when button clicked in menu/toolbar
 */
NavigationModule.prototype._zoomOneToOnePixel = function() {
	if (this.f.getActivePage().hasGraphics()) {
		// Shouldn't be possible for CAD (disabled), but extra check
		this.f.setOption("PREFERENCES_ZOOM_REALSIZE", NavigationConstants.ZoomOneToOneClick.AUTO);
	}
	this.f.emit("viewport.zoomOneToOne", {
		animated: true
	});
};

/**
 * Zoom 1:1 Real Size: called when button clicked in menu/toolbar
 */
NavigationModule.prototype._zoomOneToOneRealSize = function() {
	if (this.f.getActivePage().hasGraphics()) {
		// Only save the preference for Graphic files
		this.f.setOption("PREFERENCES_ZOOM_REALSIZE", NavigationConstants.ZoomOneToOneClick.REALSIZE);
	}
	this.f.emit("viewport.zoomOneToOne", {
		animated: true,
		worldResolution: this.getMonitorResolution()
	});
};

/**
 * The monitor resolution in ppmm
 * @returns {Number}
 */
NavigationModule.prototype.getMonitorResolution = function() {
	if (this._monitorResolution !== undefined) {
		return this._monitorResolution;
	} else {
		return ViewerUtils.parseFloat(this.f.getOption("PREFERENCE_MONITOR_RES"));
	}
};

/**
 * Set the new monitor resolution (in ppmm)
 * @param {Number} monitorResolution
 */
NavigationModule.prototype.setMonitorResolution = function(monitorResolution) {
	// keep it temporary available and only save on OK
	this._monitorResolution = monitorResolution;
};

/**
 * Called when the ok button of the monitor resolution popup is clicked
 */
NavigationModule.prototype.okMonitorResolution = function() {
	this.f.setOption("PREFERENCE_MONITOR_RES", this.getMonitorResolution() + "");
	this._monitorResolution = undefined;
	this._zoomOneToOneRealSize();
	this._adjustMonitorResolutionPopup.hide();
};

/**
 * Called when the cancel button of the monitor resolution popup is clicked
 */
NavigationModule.prototype.cancelMonitorResolution = function() {
	this._monitorResolution = undefined;
};

export default NavigationModule;
