import _ from "underscore";
import AbstractTool from "@esko/web-viewer/viewport/tools/AbstractTool";
import ShapeUtils from "@esko/shapeutils";
import CoreConstants from "@esko/web-viewer/core/CoreConstants";
/**
 * @exports NavigatorPanTool
 * @memberof NavigationModule
 * @augments Viewport.AbstractTool
 *
 * @class
 * The NavigatorPanTool tool enables the user to pan in the navigator widget by moving the highlighted area.
 * @constructor
 * @param {Core.ViewerFacade} f
 * @param {ViewportModule} viewport
 */
var NavigatorPanTool = function NavigatorPanTool(f, viewport) {
	AbstractTool.call(this, f, CoreConstants.TOOLS.NAVIGATIONPAN);

	/**
	 * Start point of the panning, in pixels
	 */
	this._startPoint = new ShapeUtils.Point();

	/**
	 * Offset of the currently active highlight in the miniature image, in world coordinates (mm)
	 */
	this._startOffset = new ShapeUtils.Size();

	/**
	 * Current center of the highlight, in pixels
	 */
	this._currentCenter = new ShapeUtils.Size();

	/**
	 * The viewport module instance
	 */
	this._viewport = viewport;

	/**
	 * Save the last mouse position while panning to trigger a rerender when leaving panning state
	 */
	this._lastMousePosition = null;

	/**
	 * DEFAULT STATE
	 */
	this.registerState("default", {
		mousedown: _.bind(function(e) {
			var visibleRect = this._viewport.getVisibleRect();
			if (!this._viewport.getWorldBounds().rectEquals(visibleRect)) {
				var offset = this._widget.getNavigatorImageOffset(); //get the offset of the image in the browser window, to get the location in the miniature image
				var coordinates = new ShapeUtils.Point(e.pageX - offset.left, e.pageY - offset.top); //coordinates of the click point in the canvas
				var worldCoordinates = this.convertScenePointToWorldPoint(coordinates); //the coordinates in the world (in mm)
				this._startOffset = this._viewport.getWorld().getOffsetSize(); //set the start offset of the currently active selection (highlighted area), in mm
				this._startPoint = coordinates; //the startpoint of the panning
				if (visibleRect.containsPoint(worldCoordinates)) {
					//cursor location is in the current seletion (highlighted area)
					this.changeState("panning"); //change the state to panning
				} else if (this._viewport.getWorldBounds().containsPoint(worldCoordinates)) {
					//if not move the visible rect to the clicked location, enable panning
					this._currentCenter = visibleRect.center(); //get the center of the current selection (highlighted area)
					this._updateViewport(coordinates, true); //update the viewport, move the selection to the clicked area
					this._startPoint = this._widget.getNavigatorTransformation().transformedPoint(this._currentCenter); //transform the world coordinates back to scene coordinates
					this.changeState("panning"); //set the state to panning
				}
			}
		}, this),

		mouseup: function() {},

		mousemove: function() {},

		mouseout: function() {},

		mouseleave: function() {},

		/**
		 * on right click don't do anything
		 */
		contextmenu: _.bind(function(e) {
			e.preventDefault();
			e.stopPropagation();
		}, this)
	});

	/**
	 * PANNING STATE
	 */
	this.registerState(
		"panning",
		_.defaults(
			{
				/**
				 * when starting to drag, notify viewport on interactive behavior and disable text selection
				 */
				enterState: _.bind(function() {
					this.setCursor("cursor-moveObject"); //set the cursor to a move icon
					this.f.emit("viewport.startInteractive");
					this.f.emit("viewport.startPanning");
					this.f.emit("layout.disableTextSelection");
				}, this),

				/**
				 * When dragging has stopped, notify viewport and enable text selection again
				 */
				leaveState: _.bind(function() {
					this.setCursor("cursor-default");
					this.f.emit("viewport.stopInteractive");
					this.f.emit("viewport.stopPanning");
					this.f.emit("layout.enableTextSelection");
					if (this._lastMousePosition) {
						//trigger a forced update of the viewport after panning has stopped to trigger a rerender of the viewport
						this._updateViewport(this._lastMousePosition, false, true); //force an update
					}
				}, this),

				/**
				 * On mouse up, panning has stopped , update the world a last time, and go back to the default state
				 */
				mouseup: _.bind(function(e) {
					e.preventDefault();
					this._lastMousePosition = null; //set to null to make sure a click will not trigger a forced update viewport
					this.changeState("default");
				}, this),

				/**
				 * Mouse is moving, update panning/world
				 */
				mousemove: _.bind(function(e) {
					var offset = this._widget.getNavigatorImageOffset(); //get the correct offset
					var coordinates = new ShapeUtils.Point(e.pageX - offset.left, e.pageY - offset.top); //calculate the correct coordinates
					this._lastMousePosition = coordinates; //set the last mouse position for a rerender if needed.
					this._updateViewport(coordinates, false); //update the viewport
				}, this),

				/**
				 * panning has stopped, update world, go back to default state
				 */
				mouseout: _.bind(function(e) {
					e.preventDefault();
					this.changeState("default");
				}, this),

				/**
				 * panning has stopped, update world, go back to default state
				 */
				mouseleave: _.bind(function(e) {
					e.preventDefault();
					this.changeState("default");
				}, this)
			},
			this.getState("default")
		)
	);
};

// inherit
NavigatorPanTool.prototype = Object.create(AbstractTool.prototype);
// correct the constructor pointer
NavigatorPanTool.prototype.constructor = NavigatorPanTool;

/**
 * Initialize the tool
 * Only accepts NavigatorWidget!!
 * @param {NavigatorWidget} widget
 */
NavigatorPanTool.prototype.initialize = function(widget, toolManager) {
	AbstractTool.prototype.initialize.call(this, widget, toolManager);

	// register the needed events
	this.registerActiveEvents(["mousedown", "mousemove", "mouseup", "mouseout", "mouseleave", "contextmenu"]);
};

/**
 * Returns the default cursor, takes the current state into account
 * @returns {String}
 */
NavigatorPanTool.prototype.getDefaultCursor = function() {
	if (this.getCurrentStateName() != "panning") {
		return "cursor-default";
	} else {
		return "cursor-moveObject";
	}
};

/**
 * The tool can be interrupted when not in "panning" state, will not be used as the tool is managed by Navigatorwidget
 * @returns {Boolean}
 */
NavigatorPanTool.prototype.canBeInterrupted = function() {
	return this.getCurrentStateName() != "panning";
};

/**
 * Do the activation of the tool.
 */
NavigatorPanTool.prototype.doActivation = function() {
	AbstractTool.prototype.doActivation.call(this);
	this.changeState("default");
};

/**
 * Do the deactivation of the tool
 * @param {function} interruptRequest
 */
NavigatorPanTool.prototype.deactivate = function(interruptRequest) {
	AbstractTool.prototype.deactivate.call(this, interruptRequest);
	this.changeState("default");
};

/**
 * Update the world
 *
 * @param {ShapeUtils.Point} to - to where has the user dragged, or where has the user clicked
 * @param {Boolean} useCurrentCenter - should the selection be moved to 'to' from the current center
 * @param {Boolean} [forceUpdate] - if set it will force an update
 */
NavigatorPanTool.prototype._updateViewport = function(to, useCurrentCenter, forceUpdate) {
	var draggedSize = this._startPoint.sizeTo(to);

	if (Math.abs(draggedSize.width) > 0 || Math.abs(draggedSize.height) > 0 || useCurrentCenter || forceUpdate) {
		var worldOffset = new ShapeUtils.Size();
		if (useCurrentCenter) {
			worldOffset = this._currentCenter.sizeTo(this.convertScenePointToWorldPoint(to));
		} else {
			worldOffset = this.convertScenePointToWorldPoint(this._startPoint).sizeTo(this.convertScenePointToWorldPoint(to));
		}

		worldOffset = new ShapeUtils.Size(-worldOffset.width, -worldOffset.height);
		worldOffset.add(this._startOffset);
		this._viewport.getWorld().setOffsetSize(worldOffset);
	}
};

/**
 * convert the scene point to world coordinates
 * @param {ShapeUtils.Point} point
 * @returns {ShapeUtils.Point}
 */
NavigatorPanTool.prototype.convertScenePointToWorldPoint = function(point) {
	return this._widget
		.getNavigatorTransformation()
		.createInverse()
		.transformedPoint(point);
};

export default NavigatorPanTool;
