import AbstractWidget from "@esko/web-viewer/layout/AbstractWidget";
import LayoutConstants from "@esko/web-viewer/layout/LayoutConstants";
import ZoomBar from "@esko/web-viewer/navigation/ZoomBar";
import $ from "jquery";
import _ from "underscore";
import HTTPUtils from "@esko/httputils";
import Fragment from "@esko/web-viewer/utils/Fragment";
import Transformation from "@esko/transformation";
import DrawUtils from "@esko/web-viewer/utils/DrawUtils";
import ShapeUtils from "@esko/shapeutils";
import ListenerUtil from "@esko/web-viewer/utils/ListenerUtil";
import NavigatorPanTool from "@esko/web-viewer/navigation/NavigatorPanTool";
import CompareConstants from "@esko/web-viewer/compare/CompareConstants";
/**
 * @const the max height of navigator image, to prevent the widget will grow to large
 */
var MAX_HEIGHT = 200;

/**
 * @exports NavigatorWidget
 * @memberof NavigationModule
 * @augments Layout.AbstractWidget
 * @class The NavigatorWidget shows the position of the user in the document and allows the user to navigate through the document
 * @constructor
 * @param {Core.ViewerFacade} f
 */
var NavigatorWidget = function NavigatorWidget(f) {
	AbstractWidget.call(this, f, LayoutConstants.Widget.NAVIGATOR);
	ListenerUtil.mixin(this, f);

	//the container all the layers (canvas-elements) will be added to, makes it possible to center them
	this.navigatorImage = $("<div>").addClass("navigatorImage");

	//define a layer for all the graphics that will be drawn in the navigator
	this._graphicsLayer = $("<canvas>").css({
		position: "absolute", //makes it possible to place multiple layer on top of each other
		top: 0,
		left: 0
	});

	//define a layer for all the CAD lines that will be drawn in the navigator
	this._cadLayer = $("<canvas>").css({
		position: "absolute",
		top: 0,
		left: 0
	});

	//define a layer for the interactive selection rectangle
	this._interactiveLayer = $("<canvas>").css({
		position: "absolute",
		top: 0,
		left: 0
	});

	//last requested fragment will be saved to this property
	this._fragment = null;

	//instance of the viewport, so we don't need to request it all the time
	this._viewport = null;

	this._worldTranformation = null;

	this.f.loadBundle(
		"viewer",
		function() {
			this.setTitle(this.f.t("NavigatorTitle"));
		},
		this
	);

	this.setIcon("icon icon-WCR_Navigator");

	//make the render navigator image debounced, so it doesn't render all the time
	var debouncedRender = _.debounce(this._renderNavigatorImage, 200);

	//only renders the overlay of the current visible rect in the navigator widget
	var throttledRectRender = _.throttle(
		_.bind(function() {
			if (this._viewport) {
				var rotation = this._viewport.getWorld().getRotation() % 360;
				var worldBounds = this._viewport.getWorldBounds();

				//if the rotation switches width and height of the world, change them for the world as we'll use this to calculate
				if (rotation == 90 || rotation == 270) {
					worldBounds.size = new ShapeUtils.Size(worldBounds.size.height, worldBounds.size.width);
				}
				if (worldBounds.size.width !== 0 && worldBounds.size.height !== 0) {
					//render the visible rect on to the navigator image
					this._renderVisibleRect(this._interactiveLayer[0].getContext("2d"), worldBounds, rotation);
				}
			}
		}, this),
		10
	);

	this.resized = function() {
		//when the widgets gets resized render the widget again
		debouncedRender.call(this);
	};

	// Zoom bar
	this._zoomBar = null;

	//register all the listeners for events that can happen in the viewport
	this._registerListener("transform.viewportWidget.worldBoundsChanged", debouncedRender, this);
	this._registerListener("module.compare.changed.diffMode", debouncedRender, this);
	this._registerListener("module.viewport.finishedLoadingImageData", debouncedRender, this);
	this._registerListener("cadModel.*.modelChanged", debouncedRender, this);
	this._registerListener("transform.viewportWidget.rotated", debouncedRender, this);
	this._registerListener("transform.viewportWidget.flipped", debouncedRender, this);
	this._registerListener("transform.viewportWidget.flippedViewFromBack", debouncedRender, this);
	this._registerListener("transform.viewportWidget.transformChanged", throttledRectRender, this);
	this._registerListener("modalResize", debouncedRender, this, this);
	this._registerListener("collapsed", debouncedRender, this, this);
	this._registerListener("visible", debouncedRender, this, this);
};

//inherit
NavigatorWidget.prototype = Object.create(AbstractWidget.prototype);
NavigatorWidget.prototype.constructor = NavigatorWidget;
NavigatorWidget.widgetName = LayoutConstants.Widget.NAVIGATOR;

/**
 * @override
 */
NavigatorWidget.prototype.setContext = function(context, headerExtra) {
	AbstractWidget.prototype.setContext.call(this, context);

	context.append(this.navigatorImage);
	this.navigatorImage.append(this._graphicsLayer);
	this.navigatorImage.append(this._cadLayer);
	this.navigatorImage.append(this._interactiveLayer);

	// initial render or when viewport becomes available
	this.f.whenModuleAvailable(
		"viewport",
		function(viewportModule) {
			this._viewport = viewportModule;

			this._panTool = new NavigatorPanTool(this.f, this._viewport);
			this._panTool.initialize(this);
			this._panTool.activate();

			this._zoomBar = new ZoomBar(this.f);
			this.context.addClass("navigatorWidget").append(this._zoomBar.element);
			this._renderNavigatorImage();
		},
		this
	);
};

/**
 * Render the navigator image.
 */
NavigatorWidget.prototype._renderNavigatorImage = function() {
	if (!this.isCollapsed() && this.isVisible() && this._viewport && this.context) {
		var page = this.f.getActivePage();
		var cadOverlay = this.f.getDocumentController().getCadOverlay();
		var size = new ShapeUtils.Size(
			this.context.innerWidth() || null,
			(this.context.innerHeight() || null) - (this._zoomBar.element.outerHeight() || null)
		); //get the current size of the widget

		var rotation = this._viewport.world.getRotation() % 360;
		var worldBounds = this._viewport.getWorldBounds();

		if (worldBounds.size.width !== 0 && worldBounds.size.height !== 0) {
			//if the rotation switches width and height of the world, change them for the world as we'll use this to calculate
			if (rotation == 90 || rotation == 270) {
				worldBounds.size = new ShapeUtils.Size(worldBounds.size.height, worldBounds.size.width);
			}

			//calculate the correct width/height of the layers so they fit into the widget, to a maximum of 200px height and the current width of the widget
			this._pageRes = size.width / worldBounds.size.width; //first calculate the resolution of the width given the current widget width
			var newWidth = size.width; //the width stays the same for now, we'll have to check if the height isn't more than 200px
			var newHeight = worldBounds.size.height * this._pageRes; //calculate the height with the resolution of the width
			var widgetHeight = MAX_HEIGHT;
			if (this._isModal) {
				widgetHeight = size.height;
			}
			//check if the new height of the layers isn't higher than 200px or height (if modal)
			if (newHeight > widgetHeight) {
				newHeight = widgetHeight; //we'll restrict the dimensions of the navigator image to the maximum height
				this._pageRes = newHeight / worldBounds.size.height; //new resolution is needed given the fixed height of 200px
				newWidth = worldBounds.size.width * this._pageRes; //calculate the width of the layer with this resolution
			}

			this.navigatorImage.height(Math.round(newHeight)); //set the width and height of the navigator images to make sure the widget will resize
			this.navigatorImage.width(Math.round(newWidth));
			this._fireResized();

			var newSize = new ShapeUtils.Size(newWidth, newHeight); //make sizes that take the rotation into account

			var rotatedSize = new ShapeUtils.Size(newWidth, newHeight); //make a size that doesn't take the rotation into account
			if (rotation == 90 || rotation == 270) {
				//if the image is rotated, still make sure the width represents the original image width (only now in pixels)
				rotatedSize = new ShapeUtils.Size(newHeight, newWidth);
			}

			if (page.hasGraphics() || page.isBlankPage() || page.hasSvg()) {
				//if there are graphics shown
				this._graphicsLayer.show();
				this._graphicsLayer[0].width = newWidth; //set the width and height of the canvas to the correct dimensions
				this._graphicsLayer[0].height = newHeight;
				if (page.hasGraphics()) {
					// For graphic files we can copy the image from the canvas
					this._renderGraphicsImage(this._graphicsLayer[0].getContext("2d"), this._pageRes, newSize, rotatedSize, worldBounds); //render the graphics on the given canvas
				} else {
					// for svg we  have to get the thumbnail from the server
					this._renderSVGImage(this._graphicsLayer[0].getContext("2d"), this._pageRes, newSize, rotatedSize, worldBounds, page);
				}
			} else {
				//if there are no graphics, hide the graphics layer
				this._graphicsLayer.hide();
			}
			if (page.hasCad() || cadOverlay) {
				//if there is CAD shown in the viewport
				this._cadLayer.show();
				this._cadLayer[0].width = newWidth;
				this._cadLayer[0].height = newHeight;

				if (this.f.getModule("cad")) {
					if (page.hasCad()) {
						//there is only CAD, no CAD+Graphics
						this._renderCAD(this._cadLayer[0].getContext("2d"), page, this._pageRes, worldBounds, null, rotatedSize);
					} else {
						//there is CAD+Graphics
						//big difference is the extra offset between CAD and graphics
						this._renderCAD(this._cadLayer[0].getContext("2d"), cadOverlay, this._pageRes, worldBounds, true, rotatedSize);
					}
				}
			} else {
				this._cadLayer.hide();
			}
			this._interactiveLayer[0].width = newWidth;
			this._interactiveLayer[0].height = newHeight;
			this._renderVisibleRect(this._interactiveLayer[0].getContext("2d"), worldBounds, rotation);
		}
	}
};

/**
 * Render the Graphics that are shown in the viewport onto the canvas of the navigator widget
 * @param {CanvasRenderingContext2D} ctx - context of the canvas to drawn the image on
 * @param {Number} pageRes - resolution of the image, will decide the size in pixels, as this says how many pixel per mm there will be rendered
 * @param {ShapeUtils.Size} newSize - new size (in pixels) of the canvas (as visible and rotated)
 * @param {ShapeUtils.Size} rotatedSize - new size (in pixels) of the canvas (not rotated, 90,270 degrees won't show the 'visbile' width/height)
 * @param {ShapeUtils.Rect} worldBounds - the world bounds of the visible image (viewport) in mm
 */
NavigatorWidget.prototype._renderGraphicsImage = function(ctx, pageRes, newSize, rotatedSize, worldBounds) {
	throw "Not Implemented: NavigatorWidget._renderGraphicsImage";
};

/**
 * Renders the navigator widget by getting a thumbnail image from the server
 * @param {CanvasRenderingContext2D} ctx - context of the canvas to drawn the image on
 * @param {Number} pageRes - resolution of the image, will decide the size in pixels, as this says how many pixel per mm there will be rendered
 * @param {ShapeUtils.Size} newSize - new size (in pixels) of the canvas (as visible and rotated)
 * @param {ShapeUtils.Size} rotatedSize - new size (in pixels) of the canvas (not rotated, 90,270 degrees won't show the 'visbile' width/height)
 * @param {ShapeUtils.Rect} worldBounds - the world bounds of the visible image (viewport) in mm
 * @param {AbstractDocumentPage} page
 */
NavigatorWidget.prototype._renderSVGImage = function(ctx, pageRes, newSize, rotatedSize, worldBounds, page) {
	//draw a white rectangle for blank pages
	ctx.setTransform(1, 0, 0, 1, 0, 0); //reset the previous transformation of the graphics layer (canvas-context)
	ctx.clearRect(0, 0, this._graphicsLayer.width() || null, this._graphicsLayer.height() || null); //make sure the canvas is clear before drawing new image on it
	this._transformCanvas(ctx, newSize, null, null); //calculate the new transformation for the graphics layer
	ctx.fillStyle = "rgba(255, 255, 255, 1)"; //make the fill color white
	DrawUtils.drawRect(ctx, 0, 0, rotatedSize.width, rotatedSize.height, 0, true, false); //draw the image, it will be rotated by the transformation to draw in rotate indepented size

	var svgElement = page.getSVGElement();

	if (svgElement) {
		// Create string from the svg element
		var svgString = new XMLSerializer().serializeToString(page.getSVGElement());

		// Check if the page is readerspread, if so, we also need the difference page svg
		var diffSvgString;
		var diffPage;
		var diffSVG;
		if (page.isReaderSpread()) {
			diffPage = this.f.getDocumentController().getDifferencePage();
			diffSVG = diffPage && diffPage.getSVGElement();
			if (diffSVG) {
				diffSvgString = new XMLSerializer().serializeToString(diffSVG);
			}
		}

		ctx.setTransform(1, 0, 0, 1, 0, 0); //reset the previous transformation of the graphics layer (canvas-context)
		ctx.clearRect(0, 0, this._graphicsLayer.width() || null, this._graphicsLayer.height() || null); //make sure the canvas is clear before drawing new image on it
		this._transformCanvas(ctx, newSize, null, null); //calculate the new transformation for the graphics layer

		// If the svg is loaded, draw it onto the canvas
		var drawOnCanvas = _.bind(function(page, img) {
			var originX = page.isReaderSpread() && page.get("isRightPage") ? rotatedSize.width / 2 : 0;
			var pageSize = page.getPageSize();
			var width = pageSize.width * pageRes;
			var height = pageSize.height * pageRes;
			ctx.fillRect(originX, 0, width, height);
			ctx.drawImage(img, originX, 0, width, height); //draw the image, it will be rotated by the transformation to draw in rotate indepented size
		}, this);

		// draw the active page
		this._getSVGImage(svgString, _.partial(drawOnCanvas, page));
		if (diffSvgString) {
			// if there is a diff page, draw it on the canvas as well
			this._getSVGImage(diffSvgString, _.partial(drawOnCanvas, diffPage));
		}
	}
};

/**
 * Greates and image of an svg string and passes it as argument to the onload function
 * @param {String} svgString
 * @param {Function} onload
 */
NavigatorWidget.prototype._getSVGImage = function(svgString, onload) {
	var DOMURL = window.URL || window.webkitURL || window;

	var img = new Image();

	// Create a blob for the svg for which we can create a URL
	var svgBlob = new Blob([svgString], {
		type: "image/svg+xml;charset=utf-8"
	});
	var url = DOMURL.createObjectURL(svgBlob);

	img.onload = function() {
		onload(img);
		DOMURL.revokeObjectURL(url);
	};
	img.src = url;
};

/**
 * Renders CAD linework in the navigator widget, on the provided canvas-context
 * @param {CanvasRenderingContext2D} ctx - context of the canvas to drawn the image on
 * @param {AbstractDocumentPage} page - page that has the CAD linework information, CAD can be requested from this page
 * @param {Number} pageRes - the resolution of the CAD, will be used to set the correct scale to the canvas-context
 * @param {ShapeUtils.Rect} worldBounds - bounds of the world, size of the CAD-file (or combination of CAD-files in compare)
 * @param {Boolean} isCadOverlay - Does the overlay need to be rendered or not
 * @param {ShapeUtils.Size} canvasSize - new size (in pixels) of the canvas (not rotated, 90,270 degrees won't show the 'visbile' width/height)
 */
NavigatorWidget.prototype._renderCAD = function(ctx, page, pageRes, worldBounds, isCadOverlay, canvasSize) {
	page.requestCad(
		_.bind(function() {
			//to make sure the CAD is present when trying to render it in the navigator widget
			var cadModule = this.f.getModule("cad"); //request the CAD-module
			//we need CadModel of both the current and reference (if any) page
			var currentCadModel;
			if (isCadOverlay) {
				currentCadModel = cadModule.getCadOverlayModel();
			} else {
				currentCadModel = cadModule.getCurrentPageCadModel();
			}
			var referenceCadModel = cadModule.getReferencePageCadModel();
			var visibleRect = this._viewport.getWorldBounds(); //the currently visible rect (represents the rectangle of the entire CAD-document)
			if (currentCadModel !== null && currentCadModel.getCad()) {
				//get the origin of the active page's CAD linework
				var origin = currentCadModel.getCad().getOrigin();
				//make sure the visibleRect is translated to the (0,0)-point (the origin) of the CAD file
				visibleRect.origin.x += origin.x; //add the x-coordinate of the origin to it
				visibleRect.origin.y -= origin.y; //as y-axis starts for the bottom (not the top), subtract the y-coordinate of the origin

				var cadLines = currentCadModel.getVisibleCadLines(); //return all the visible CAD lines

				var offset = new ShapeUtils.Size();
				if (!isCadOverlay) {
					offset = new ShapeUtils.Size(-origin.x - worldBounds.origin.x, origin.y - worldBounds.origin.y); //this is an extra offset needed for CAD
				} else {
					offset = new ShapeUtils.Size(-worldBounds.origin.x, -worldBounds.origin.y); //this is an extra offset needed for CAD
				}

				//reset the transform of the canvas-context
				ctx.setTransform(1, 0, 0, 1, 0, 0);
				ctx.clearRect(0, 0, this._cadLayer.width() || null, this._cadLayer.height() || null); //clear the canvas before rendering the CAD again on the canvas

				var t = this._transformCanvas(ctx, worldBounds.size, offset, pageRes); //calculate the correct transformation (include pageRes as we will drawn CAD in mm, not pixels)

				var overlayTransform = null;
				if (isCadOverlay) {
					overlayTransform = cadModule.getOverlayTransform();
					if (overlayTransform) {
						ctx.transform(
							overlayTransform.a,
							overlayTransform.b,
							overlayTransform.c,
							overlayTransform.d,
							overlayTransform.e,
							overlayTransform.f
						);
					}
				} else {
					// draw background
					ctx.save();
					ctx.fillStyle = currentCadModel.getBackgroundString();
					DrawUtils.drawRect(
						ctx,
						visibleRect.origin.x,
						visibleRect.origin.y,
						visibleRect.size.width,
						visibleRect.size.height,
						0,
						true,
						true
					);
					ctx.restore();
				}

				var newTransform = t.copy();

				if (cadModule.compareDiffMode !== CompareConstants.diffMode.DIFFVERSION || isCadOverlay) {
					if (overlayTransform) {
						newTransform.concatenate(overlayTransform);
					}
					var drawRect = this.f
						.getModule("viewport")
						.getWorld()
						.getWorldRectFromScene(
							new ShapeUtils.Rect(null, new ShapeUtils.Size(this._cadLayer.width() || null, this._cadLayer.height() || null)),
							newTransform
						);
					//draw the CAD-linework on the canvas
					DrawUtils.drawOverlay(
						this.f,
						ctx,
						cadLines,
						drawRect,
						function() {},
						canvasSize.width / visibleRect.size.width,
						t,
						currentCadModel.getCadColor(),
						currentCadModel.getAlpha()
					);
				}

				//if there is CAD-linework in the reference cadmodel, CAD is in compare mode
				if (referenceCadModel !== null && referenceCadModel.getCad() !== null && cadModule.getReferenceTransform()) {
					var referenceLines = referenceCadModel.getVisibleCadLines();

					t = cadModule.getReferenceTransform();
					t.preTranslate(origin.x, -origin.y);
					//transform the canvas
					ctx.transform(t.a, t.b, t.c, t.d, t.e, t.f);

					newTransform.concatenate(t);
					var referenceDrawRect = this.f
						.getModule("viewport")
						.getWorld()
						.getWorldRectFromScene(
							new ShapeUtils.Rect(null, new ShapeUtils.Size(this._cadLayer.width() || null, this._cadLayer.height() || null)),
							newTransform
						);
					//draw the reference document page CAD linework
					DrawUtils.drawOverlay(
						this.f,
						ctx,
						referenceLines,
						referenceDrawRect,
						function() {},
						canvasSize.width / visibleRect.size.width,
						t,
						referenceCadModel.getCadColor(),
						referenceCadModel.getAlpha()
					);
				}
			}
		}, this)
	);
};

/**
 * Renders the currently visible rect of the viewport on to the navigator image
 * @param {CanvasRenderingContext2D} ctx - context of the canvas to drawn overlay onto
 * @param {ShapeUtils.Rect} worldBounds - bounds of the world, size of the CAD-file/Graphics-file (or combination of CAD-files/Graphics-files in compare)
 * @param {Number} rotation - current rotation of the viewport
 */
NavigatorWidget.prototype._renderVisibleRect = function(ctx, worldBounds, rotation) {
	var visibleRect = this._viewport.getVisibleRect(); //get the currently visible rectangle
	ctx.setTransform(1, 0, 0, 1, 0, 0); //first reset the transformation
	ctx.clearRect(0, 0, this._interactiveLayer.width() || null, this._interactiveLayer.height() || null); //clear the canvas
	//transform the canvas, add an offset for the origin of CAD, this makes the transformation from scene to world coordinates correct
	this._worldTranformation = this._transformCanvas(
		ctx,
		worldBounds.size,
		new ShapeUtils.Size(-worldBounds.origin.x, -worldBounds.origin.y),
		this._pageRes
	);

	// draw a greyed out overlay
	ctx.beginPath();
	ctx.strokeStyle = "rgba(0, 0, 0, 1.0)";
	ctx.lineWidth = 1 / this._pageRes;
	ctx.fillStyle = "rgba(0, 0, 0, 0.50)"; //fill the rectangle
	if (rotation == 90 || rotation == 270) {
		//make sure the worldbounds have the correct orientation when rotated
		ctx.rect(worldBounds.origin.x, worldBounds.origin.y, worldBounds.size.height, worldBounds.size.width);
	} else {
		ctx.rect(worldBounds.origin.x, worldBounds.origin.y, worldBounds.size.width, worldBounds.size.height);
	}
	ctx.fill();
	ctx.stroke();

	// cut out the selection area
	ctx.clearRect(visibleRect.origin.x, visibleRect.origin.y, visibleRect.size.width, visibleRect.size.height);

	// draw dashed rectangle
	ctx.beginPath();
	ctx.strokeStyle = "rgba(255,255,255,1)";
	ctx.lineWidth = 1 / this._pageRes;
	ctx.strokeStyle = "rgba(0,0,0,1)";
	if (this._pageRes) {
		DrawUtils.drawDashedRect(ctx, visibleRect, [2, 5], this._pageRes);
	}

	ctx.stroke();
};

/**
 * Transforms the canvas-context to have the correct rotation, and flip of the viewport
 * @param {CanvasRenderingContext2D} ctx
 * @param {ShapeUtils.Size} worldSize - size of the world
 * @param {ShapeUtils.Size} [offset] - offset to translate the transformation
 * @param {Number} [pageRes] - resolution of the page, used to scale the canvas, provide this if you want to drawn in world-coordinates (not pixels)
 * @returns {Transformation}
 */
NavigatorWidget.prototype._transformCanvas = function(ctx, worldSize, offset, pageRes) {
	var isFlipped = this._viewport.world.isFlip(); //check in the viewport if the world is flipped or rotated
	var isFlipViewFromBack = this._viewport.world.isFlipViewFromBack();
	var rotation = this._viewport.world.getRotation() % 360;

	var t = new Transformation();

	if (pageRes !== null) {
		//scale the transformation if needed
		t.scale(pageRes, pageRes);
	}
	t.translate(worldSize.width / 2, worldSize.height / 2); //first translate to the center to rotate and flip the contents of the canvas
	t.rotate(rotation * (Math.PI / 180)); //rotate
	if (isFlipped) {
		t.scale(-1, 1);
	}
	if (isFlipViewFromBack) {
		if (rotation % 180 === 0) {
			// not rotated
			t.scale(-1, 1);
		} else {
			// rotated, but flip horizontally
			t.scale(1, -1);
		}
	}
	if (rotation == 90 || rotation == 270) {
		//if the rotation is 90 or 270 (width and heights are swapped) calculate extra translate distance
		t.translate(
			-worldSize.width / 2 + (worldSize.width - worldSize.height) / 2,
			-worldSize.height / 2 - (worldSize.width - worldSize.height) / 2
		);
	} else {
		t.translate(-worldSize.width / 2, -worldSize.height / 2);
	}
	if (offset !== null) {
		//if an offset is provide, translate over this offset
		t.translate(offset.width, offset.height);
	}
	//perform the translation on the canvas context
	ctx.transform(t.a, t.b, t.c, t.d, t.e, t.f);

	return t;
};

/**
 * Get the transformation for the navigator (used for the interactive layer (and CAD))
 * @returns {Transformation}
 */
NavigatorWidget.prototype.getNavigatorTransformation = function() {
	return this._worldTranformation;
};

/**
 * Returns the minimum size for the widget
 * @returns {ShapeUtils.Size}
 */
NavigatorWidget.prototype.getMinimumSize = function() {
	return new ShapeUtils.Size(200, (this.navigatorImage.height() || null) + (this._zoomBar.element.height() || null));
};

/**
 * Get the offset of the navigatorImage in the viewer
 * @returns {ShapeUtils.Size}
 */
NavigatorWidget.prototype.getNavigatorImageOffset = function() {
	return this.navigatorImage.offset();
};

export default NavigatorWidget;
