import _ from "underscore";
import AbstractTool from "@esko/web-viewer/viewport/tools/AbstractTool";
import ShapeUtils from "@esko/shapeutils";
import GEOUtils from "@esko/web-viewer/utils/GEOUtils";
import CoreConstants from "@esko/web-viewer/core/CoreConstants";
/**
 * @exports PanTool
 * @memberof NavigationModule
 * @augments Viewport.AbstractTool
 *
 * @class
 * The pan tool enables the user to pan in the document. Scrolling and double click/right click can be used to zoom
 * @constructor
 * @param {Core.ViewerFacade} f
 */
var PanTool = function Pan(f) {
	AbstractTool.call(this, f, CoreConstants.TOOLS.PAN);

	/**
	 * Start point, set when dragging is started
	 */
	var startPoint = new ShapeUtils.Point(0, 0);
	/**
	 * the offset needed for dragging
	 */
	var startOffset = new ShapeUtils.Size();

	var deactivateAfterPanAction = false;

	/**
	 * Update the world
	 *
	 * @param {ShapeUtils.Point} to - to where has the user dragged?
	 * @param {Boolean} isEndOfInteraction
	 */
	var updateWorld = _.bind(function(to, isEndOfInteraction) {
		var draggedSize = startPoint.sizeTo(to);

		if (Math.abs(draggedSize.width) > 0 || Math.abs(draggedSize.height) > 0 || isEndOfInteraction) {
			var worldOffset = this.sceneToWorld(startPoint).sizeTo(this.sceneToWorld(to));
			worldOffset.add(startOffset);

			var animation = undefined;
			if (isEndOfInteraction) {
				animation = {
					duration: 200,
					type: "ease-out"
				};
				if (deactivateAfterPanAction) {
					f.emit("toolManager.deactivateTool", this);
					deactivateAfterPanAction = false;
				}
			}
			this._widget.getWorld().setOffsetSize(worldOffset, animation);
		}
	}, this);

	/**
	 * DEFAULT STATE
	 */
	this.registerState("default", {
		/*
         * on mouse down, start dragging
         */
		mousedown: _.bind(function(e) {
			if (e.which === 1) {
				startPoint = this.getScenePoint(e);
				startOffset = this._widget.getWorld().getOffsetSize();
				this.changeState("panning");

				e.preventDefault();
			}
		}, this),

		mouseup: function() {},

		mousemove: function() {},

		mouseout: function() {},

		mouseleave: function() {},

		/**
		 * on double click zoom in
		 */
		dblclick: _.bind(function(e) {
			this.changeState("default");
			if (this._widget.getWorld().isSmallerThanFullView()) {
				this._widget.getWorld().zoomToFit(true);
			} else {
				// zoom in
				this.zoom(1, this.getWorldPoint(e), true);
			}
		}, this),

		/**
		 * on right click zoom out
		 */
		contextmenu: _.bind(function(e) {
			e.preventDefault();
			e.stopPropagation();

			//zoom out
			this.zoom(-1, this.getWorldPoint(e), true);
		}, this)
	});

	/**
	 * a reference to timer for the re-enabling of text selection
	 */
	var textSelectionTimeOut = undefined;
	/**
	 * last mouse position is used to handle mouseleave and mouseout correctly
	 */
	var lastMousePosition = new ShapeUtils.Point(0, 0);

	this.registerState(
		"panning",
		_.defaults(
			{
				/**
				 * when starting to drag, notify viewport on interactive behavior and disable text selection
				 */
				enterState: function() {
					if (textSelectionTimeOut) {
						clearTimeout(textSelectionTimeOut);
						textSelectionTimeOut = null;
					}
					f.emit("viewport.startInteractive");
					f.emit("viewport.startPanning");
					f.emit("layout.disableTextSelection");
				},

				/**
				 * When dragging has stopped, notify viewport and enable text selection again
				 */
				leaveState: function() {
					startPoint = new ShapeUtils.Point(0, 0);
					f.emit("viewport.stopInteractive");
					f.emit("viewport.stopPanning");
					textSelectionTimeOut = setTimeout(function() {
						f.emit("layout.enableTextSelection");
					}, 500);
				},

				/**
				 * On mouse up, panning has stopped , update the world a last time, and go back to the default state
				 */
				mouseup: _.bind(function(e) {
					e.preventDefault();
					updateWorld(this.getScenePoint(e), true);
					this.changeState("default");
				}, this),

				/**
				 * Mouse is moving, update panning/world
				 */
				mousemove: _.bind(function(e) {
					lastMousePosition = this.getScenePoint(e);
					updateWorld(lastMousePosition, false);
				}, this),

				/**
				 * panning has stopped, update world, go back to default state
				 */
				mouseout: _.bind(function(e) {
					e.preventDefault();
					updateWorld(lastMousePosition, true);
					this.changeState("default");
				}, this),

				/**
				 * panning has stopped, update world, go back to default state
				 */
				mouseleave: _.bind(function(e) {
					e.preventDefault();
					updateWorld(lastMousePosition, true);
					this.changeState("default");
				}, this)
			},
			this.getState("default")
		)
	);

	// define some key events
	this.passiveKeyEvents = [];
	this.activeKeyEvents = [];

	/**
	 * Initialize the tool
	 */
	this._override("initialize", function(widget, toolManager) {
		this._super(widget, toolManager);

		// register the needed events
		this.registerActiveEvents(["mousedown", "mousemove", "mouseup", "mouseout", "mouseleave", "contextmenu", "dblclick"]);

		/**
		 * Passive events
		 */
		var requestedInterrupt = false;
		// on space down request interruption
		this.passiveKeyEvents.push([
			"space",
			_.bind(function(e) {
				if (!requestedInterrupt) {
					requestedInterrupt = true;
					f.emit("toolManager.request.activation", this);
				}
			}, this),
			"keydown"
		]);

		// on space up request interruption
		this.passiveKeyEvents.push([
			"space",
			function(e) {
				requestedInterrupt = false;
			},
			"keyup"
		]);

		/**
		 * Active events
		 */
		// on space up, and in interruption mode, request deactivation
		this.activeKeyEvents.push([
			"space",
			_.bind(function(e) {
				if (requestedInterrupt) {
					requestedInterrupt = false;
					f.emit("toolManager.request.deactivation", this);
				}
			}, this),
			"keyup"
		]);

		// always activate the tool when pressing H
		this.passiveKeyEvents.push([
			"h",
			_.bind(function(e) {
				e.preventDefault();
				f.emit("toolManager.activateTool", this);
			}, this),
			"keypress"
		]);

		this.passiveKeyEvents.push([
			"mod+up",
			_.bind(function(e) {
				e.preventDefault();
				e.stopPropagation();
				this._setPanStart();
				var sceneSize = this._widget.getWorld().getSceneSize();
				var toPoint = new ShapeUtils.Point(startPoint.x, startPoint.y + sceneSize.height / 4);
				updateWorld(toPoint, true);
			}, this),
			"keypress"
		]);

		this.passiveKeyEvents.push([
			"mod+down",
			_.bind(function(e) {
				e.preventDefault();
				e.stopPropagation();
				this._setPanStart();
				var sceneSize = this._widget.getWorld().getSceneSize();
				var toPoint = new ShapeUtils.Point(startPoint.x, startPoint.y - sceneSize.height / 4);
				updateWorld(toPoint, true);
			}, this),
			"keypress"
		]);

		this.passiveKeyEvents.push([
			"mod+right",
			_.bind(function(e) {
				e.preventDefault();
				e.stopPropagation();
				this._setPanStart();
				var sceneSize = this._widget.getWorld().getSceneSize();
				var toPoint = new ShapeUtils.Point(startPoint.x - sceneSize.width / 4, startPoint.y);
				updateWorld(toPoint, true);
			}, this),
			"keypress"
		]);

		this.passiveKeyEvents.push([
			"mod+left",
			_.bind(function(e) {
				e.preventDefault();
				e.stopPropagation();
				this._setPanStart();
				var sceneSize = this._widget.getWorld().getSceneSize();
				var toPoint = new ShapeUtils.Point(startPoint.x + sceneSize.width / 4, startPoint.y);
				updateWorld(toPoint, true);
			}, this),
			"keypress"
		]);

		// activate pan tool until one pan action is done
		this.passiveKeyEvents.push([
			"f7",
			_.bind(function(e) {
				e.preventDefault();
				e.stopPropagation();
				deactivateAfterPanAction = true;
				f.emit("toolManager.activateTool", this, true);
			}, this)
		]);
	});

	this._setPanStart = function() {
		var sceneSize = this._widget.getWorld().getSceneSize();
		startPoint = new ShapeUtils.Point(sceneSize.width / 2, sceneSize.height / 2);
		startOffset = this._widget.getWorld().getOffsetSize();
	};

	// always show the hand cursor when active
	this.getDefaultCursor = function() {
		return "cursor-move";
	};

	/**
	 * The tool can be interrupted when not in "panning" state
	 */
	this.canBeInterrupted = function() {
		return this.getCurrentStateName() != "panning";
	};

	this._override("doActivation", function() {
		this._super();
		this.changeState("default");
	});

	this._override("deactivate", function(interruptRequest) {
		this._super(interruptRequest);
		this.changeState("default");
	});

	/**
	 * Zoom in/out
	 *
	 * @param step {int} if positive zoom in, if negative zoom out
	 * @param worldMiddle {Point} what is the center of the zoom action
	 * @param animated {boolean} show animation
	 */
	this.zoom = function(step, worldMiddle, animated) {
		this._widget.getWorld().transform({
			zoom: step,
			middle: worldMiddle,
			animated: animated
		});
	};
};

// inherit
PanTool.prototype = Object.create(AbstractTool.prototype);
// correct the constructor pointer
PanTool.prototype.constructor = PanTool;

export default PanTool;
