import _ from "underscore";
import AbstractTool from "@esko/web-viewer/viewport/tools/AbstractTool";
import ViewerUtils from "@esko/web-viewer/utils/ViewerUtils";
import TouchUtil from "@esko/web-viewer/utils/TouchUtil";
import CoreConstants from "@esko/web-viewer/core/CoreConstants";
/**
 * @exports TouchTool
 * @memberof NavigationModule
 * @augments Viewport.AbstractTool
 *
 * @class
 * The touch tool enables the user to zoom, pan and rotate on the document using touch
 * @constructor
 * @param {Core.ViewerFacade} f
 */
var TouchTool = function Touch(f) {
	AbstractTool.call(this, f, CoreConstants.TOOLS.TOUCH);

	// create touch util within viewport context
	this._touchUtil = null;

	// Allows rotation or not (only rotate when threshold is passed)
	this._allowRotation = true;
	this._unlockRotation = false;
	this._unlockRotationAngle = 25;

	this._initZoom = undefined;
	this._initRotate = undefined;
	this._initTouchCenter = undefined;
};

// extend from the AbstractTool
TouchTool.prototype = Object.create(AbstractTool.prototype);
TouchTool.prototype.constructor = TouchTool;

/**
 * Action the will be called when an action is started
 * @param {Event} e - launched event
 */
TouchTool.prototype.startAction = function(e) {
	this._touchUtil.preventDefault(e);

	this.f.emit("viewport.startInteractive");
	this._unlockRotation = false;

	var world = this._widget.getWorld();
	this._initZoom = world.getZoom();
	this._initRotate = world.getRotation();
	var touchPoint = this._touchUtil.getScenePointFromTouch(e, this._widget);
	this._initTouchCenter = this.sceneToWorld(touchPoint.point);
	this._currentView = touchPoint.view;
};

/**
 * Action the will be called when an action is stopped
 * @param {Event} e - launched event
 */
TouchTool.prototype.stopAction = function(e) {
	this._touchUtil.preventDefault(e);
	delete this._currentView;
	this.f.emit("viewport.stopInteractive");
	this.releaseCanvasLayer();

	this._unlockRotation = false;
	this._widget.getWorld().transform({
		rubber: false,
		animated: {
			duration: 100,
			type: "ease-out"
		},
		constrainRotation: true
	});
};

/**
 * Handle double tap: zoom in to point if fit to screen, otherwise do a fit to screen
 * @param {Event} e - launched event
 */
TouchTool.prototype.doubleTap = function(e) {
	this._touchUtil.preventDefault(e);

	var atpoint = this.getScenePointFromTouch(e);
	var worldatpoint = this._widget.getWorld().getWorldPointFromScene(atpoint);
	var world = this._widget.getWorld();
	if (!world.isFullView()) {
		world.zoomToFit(true);
	} else {
		world.transform({
			zoom: 3,
			middle: worldatpoint,
			animated: true
		});
	}
};

/**
 * Handle pinch move action
 * @param {Event} e - launched event
 */
TouchTool.prototype.pinchMove = function(e) {
	this._touchUtil.preventDefault(e);

	// Get the center point, zoom and rotation
	var zoom = this._touchUtil.getScale(e);
	var rotation = this._touchUtil.getRotation(e);
	var center = this.getScenePointFromTouch(e);

	this.updateWorld(center, zoom, rotation);
};

/**
 * Handle pan move action
 * @param {Event} e - launched event
 */
TouchTool.prototype.panAction = function(e) {
	this._touchUtil.preventDefault(e);

	var center = this.getScenePointFromTouch(e);
	this.updateWorld(center);
};

/**
 * Update the world.
 * @param {ShapeUtils.Point} center - center point of touch (touchpoint for pan, between fingers for pinch)
 * @param {Number} zoom - New zoom level
 * @param {Number} rotation - New rotation (in degrees)
 */
TouchTool.prototype.updateWorld = function(center, zoom, rotation) {
	// Mechanism:
	// - first, do rotation and zoom, with center point as middle
	// - afterwards, do a translation, so the "world center point" (this._initTouchCenter) is between the fingers again

	var world = this._widget.getWorld();
	var options = {};

	// Set zoom
	if (zoom) {
		var worldZoom = this._initZoom * zoom;
		options.fixedzoom = worldZoom;
	}

	// Calculate new rotation (if allowed)
	if (rotation && this._allowRotation) {
		rotation = parseInt(rotation, 10) % 360;
		// If the rotation is not unlocked yet, check if it should be unlocked
		if (!this._unlockRotation) {
			// If the new rotation is bigger than the unlock angle, unlock the rotation
			var newRotation = rotation - ViewerUtils.closestAngle(rotation, 4);
			if (Math.abs(newRotation) > this._unlockRotationAngle) {
				this._unlockRotation = true;
			}
		}

		if (this._unlockRotation) {
			// Only rotate if the rotation is unlocked
			options.fixedrotation = rotation + this._initRotate;
		}
	}

	// use center of touches as center point of transformation
	options.middle = this._initTouchCenter;

	// Transform, but without redrawing, we still need to shift the center point back
	world.transform(options, true);

	// shift back, as we want to center point to stay on the same page
	var offsetSizeInScene = world
		.getSceneBounds()
		.center()
		.sizeTo(center);
	var offsetSizeInWorld = world.getWorldSizeFromScene(offsetSizeInScene);
	options = {};
	options.offsetX = offsetSizeInWorld.width;
	options.offsetY = offsetSizeInWorld.height;
	world.transform(options);
};

/**
 * Initialize the touch tool
 * @param {AbstractWidget} widget - the widget on which the tool is activated
 */
TouchTool.prototype.initialize = function(widget, toolManager) {
	AbstractTool.prototype.initialize.apply(this, arguments);
	// note this only works on the viewportwidget
	this._touchUtil = new TouchUtil(widget.getElement()[0]);
};

/**
 * override the deactivation process, so all listeners are removed
 * @override
 */
TouchTool.prototype.doDeactivation = function() {
	AbstractTool.prototype.doDeactivation.call(this);

	// unregister events
	this._touchUtil.removeAllListeners();
};

/**
 * override the activation process, so all listeners are added
 * @override
 */
TouchTool.prototype.doActivation = function() {
	AbstractTool.prototype.doActivation.call(this);

	// register events
	this._touchUtil.addPinchStartListener(
		_.bind(function(e) {
			this.f.emit("viewport.startZooming");
			this.startAction(e);
		}, this)
	);
	this._touchUtil.addPinchMoveListener(_.bind(this.pinchMove, this));
	this._touchUtil.addPinchEndListener(
		_.bind(function(e) {
			this.f.emit("viewport.stopZooming");
			this.stopAction(e);
		}, this)
	);

	this._touchUtil.addDoubleTapListener(_.bind(this.doubleTap, this));

	this._touchUtil.addPanStartListener(
		_.bind(function(e) {
			this.f.emit("viewport.startPanning");
			this.startAction(e);
		}, this)
	);
	this._touchUtil.addPanMoveListener(_.bind(this.panAction, this));
	this._touchUtil.addPanEndListener(
		_.bind(function(e) {
			this.f.emit("viewport.stopPanning");
			this.stopAction(e);
		}, this)
	);
};

/**
 * Get a scene point from a touch event
 * if a view is set with the startAction, the scene point is calculated for that view
 * @param {Event} event - the touch event
 * @returns {ShapeUtils.Point} the ScenePoint
 */
TouchTool.prototype.getScenePointFromTouch = function(event) {
	return this._touchUtil.getScenePointFromTouch(event, this._widget, this._currentView).point;
};

export default TouchTool;
