import _ from "underscore";
import AbstractTool from "@esko/web-viewer/viewport/tools/AbstractTool";
import ShapeUtils from "@esko/shapeutils";
import GEOUtils from "@esko/web-viewer/utils/GEOUtils";
import Mousewheel from "@esko/mousewheel";
import DrawUtils from "@esko/web-viewer/utils/DrawUtils";
import CoreConstants from "@esko/web-viewer/core/CoreConstants";
/**
 * @exports ZoomTool
 * @memberof NavigationModule
 * @augments Viewport.AbstractTool
 *
 * @class
 * The zoomtool enables the user to zoom in and out on the document
 *
 * Several possible interaction are possible:
 * 	zoom in: ctrl + OR left mouse click OR start dragging rectangle
 *  zoom out: ctrl - OR right mouseclick
 *  zoom to fit: ctrl 0
 *
 *  pressing alt when the tool is active changes the behavior (left click zoom out, right click zoom in)
 *
 *  activate zoom tool: z
 * @constructor
 * @param {Core.ViewerFacade} f
 */
var ZoomTool = function Zoom(f) {
	AbstractTool.call(this, f, CoreConstants.TOOLS.ZOOM);

	/* Options */
	var selectionThreshold = 16;
	var defaultZoom = 1;
	var selection;
	var mouseUpsToIgnore = 0;
	var previousMouseLocation = new ShapeUtils.Point();
	var deactiveToolAfterZoomAction = false;

	/**
	 * Action to be done after zoom
	 */
	var afterZoomAction = _.bind(function() {
		if (deactiveToolAfterZoomAction) {
			f.emit("toolManager.deactivateTool", this);
			deactiveToolAfterZoomAction = false;
		}
	}, this);

	this._zoomOnScroll = true;
	this.f.on(
		"zoomTool.zoomOnScroll",
		function(zoomOnScroll) {
			this._zoomOnScroll = zoomOnScroll;
		}.bind(this)
	);

	/**
	 * This is the mode when the tool is active,
	 * it will zoom on click and go the couldStartDraggingMode if left mousedown is clicked
	 */
	this.registerState(
		"default",
		_.bind(function() {
			/**
			 * helper function, to change the zoom mode if altkey is pressed
			 */
			var updateZoomMode = _.bind(function(e) {
				if (e.altKey) {
					defaultZoom = -1;
				} else {
					defaultZoom = 1;
				}

				this.setCursor(defaultZoom == -1 ? "cursor-zoomOut" : "cursor-zoomIn");
			}, this);

			return {
				mousedown: _.bind(function(e) {
					selection = new ShapeUtils.Rect(this.getScenePoint(e));

					updateZoomMode(e);

					// if left click, go to couldStartDragging mode
					if (e.which === 1) {
						this.changeState("couldStartDragging");
					}

					e.preventDefault();
				}, this),
				mouseup: _.bind(function(e) {
					// mouseup when not dragging --> zoom in or out
					updateZoomMode(e);

					if (mouseUpsToIgnore > 0) {
						mouseUpsToIgnore--;
					} else {
						this.zoom((e.which === 1 ? 1 : -1) * defaultZoom, this.getWorldPoint(e), true);

						this.changeState("default");
						afterZoomAction();
						e.preventDefault();
					}
				}, this),
				mousemove: function(e) {
					// mousemove when not dragging --> just update the zoom mode (cursor)
					updateZoomMode(e);
				},
				contextmenu: function(e) {
					// prevent the context menu
					e.preventDefault();
					e.stopPropagation();
				},
				mouseleave: _.bind(function() {
					// when the mouse goes out of the viewport, revert to default state
					this.changeState("default");
				}, this),
				esc: _.bind(function() {
					this.changeState("default");
				}, this),
				spacedown: function() {},
				spaceup: function() {},
				canBeInterrupted: true
			};
		}, this)()
	);

	/**
	 * In this mode, we check if the selection threshold is surpassed to see if we should start showing selection rectangle
	 * if so go to draggingMode
	 */
	this.registerState(
		"couldStartDragging",
		_.defaults(
			{
				mousemove: _.bind(function(e) {
					// mouse if moving, if the dragged distance is far enough, start the dragging mode of the selection box
					var p = this.getScenePoint(e);

					selection.update(p);

					if (Math.abs(selection.size.width) > selectionThreshold || Math.abs(selection.size.height) > selectionThreshold) {
						this.requestCanvasLayer(e);
						this.changeState("dragging");
					}
				}, this),
				mouseleave: _.bind(function() {
					// mouse left viewport, go back to default state
					this.changeState("default");
				}, this),
				esc: _.bind(function() {
					mouseUpsToIgnore++;
					this.changeState("default");
				}, this),
				canBeInterrupted: false
			},
			this.getState("default")
		)
	);

	/**
	 * release the canvas
	 */
	var cleanupCanvas = _.bind(function() {
		this.releaseCanvasLayer();
	}, this);

	/**
	 * Render the selection rectangle
	 */
	var renderSelection = _.throttle(
		_.bind(function() {
			var ctx = this.getCanvasCtx();
			if (ctx) {
				var csize = this.getCanvasSize();
				// first clear all previous drawings
				ctx.clearRect(0, 0, csize.width, csize.height);

				// draw a greyed out overlay
				ctx.beginPath();
				ctx.strokeStyle = "rgba(0, 0, 0, 1.0)";
				ctx.lineWidth = 1;
				ctx.fillStyle = "rgba(0, 0, 0, 0.40)";
				ctx.rect(0, 0, csize.width, csize.height);
				ctx.fill();
				ctx.stroke();

				// cut out the selection area
				ctx.clearRect(selection.origin.x, selection.origin.y, selection.size.width, selection.size.height);

				// draw dashed rectangle
				ctx.beginPath();
				ctx.strokeStyle = "rgba(210,210,210,0.9)";
				ctx.lineWidth = 1;
				DrawUtils.drawDashedRect(ctx, selection, [4, 3]);
				ctx.stroke();
			}
		}, this),
		40
	);

	/**
	 * The selection rectangle is shown, the tool can't be interrupted now
	 */
	this.registerState(
		"dragging",
		_.defaults(
			{
				enterState: function(previousState) {
					f.emit("layout.disableTextSelection");
					f.emit("viewport.startInteractive");
				},
				leaveState: function(nextState) {
					if (nextState != "panningSelection" || nextState != "dragging") {
						setTimeout(function() {
							f.emit("layout.enableTextSelection");
						}, 500);
						f.emit("viewport.stopInteractive");
					}
				},
				mousedown: function(e) {
					// increment mouseUp counter to ignore this right/middle click
					mouseUpsToIgnore++;
				},
				mousemove: _.bind(function(e) {
					// update rectangle and render (if needed)
					var p = this.getScenePoint(e);

					previousMouseLocation = p;
					selection.update(p);
					renderSelection();
				}, this),
				mouseleave: _.bind(function(e) {
					// mouse left viewport, go back to default state
					this.changeState("default");
					cleanupCanvas();
				}, this),
				mouseup: _.bind(function(e) {
					// the mouse button is released, if no clicks need to be ignored, zoom in
					if (mouseUpsToIgnore > 0) {
						mouseUpsToIgnore--;
					} else {
						var p = this.getScenePoint(e);

						previousMouseLocation = p;
						selection.update(p);
						this.zoomToSelection(selection);
						this.changeState("default");
						afterZoomAction();
						cleanupCanvas();
					}
				}, this),
				esc: _.bind(function(e) {
					mouseUpsToIgnore++;
					this.changeState("default");
					cleanupCanvas();
				}, this),
				spacedown: _.bind(function() {
					this.changeState("panningSelection");
				}, this),
				canBeInterrupted: false
			},
			this.getState("default")
		)
	);

	this.registerState(
		"panningSelection",
		_.defaults(
			{
				mousemove: _.bind(function(e) {
					// update rectangle and render (if needed)
					var p = this.getScenePoint(e);

					var move = previousMouseLocation.sizeTo(p);
					previousMouseLocation = p;

					selection.origin.x += move.width;
					selection.origin.y += move.height;
					renderSelection();
				}, this),
				spacedown: function() {},
				spaceup: _.bind(function() {
					this.changeState("dragging");
				}, this)
			},
			this.getState("dragging")
		)
	);

	// define some key events
	this.passiveKeyEvents = [];
	this.activeKeyEvents = [];
	this.persistentKeyEvents = [];

	this._override("initialize", function(widget, toolManager) {
		this._super(widget, toolManager);

		this.registerActiveEvents(["mousedown", "mousemove", "mouseup", "mouseleave", "contextmenu"]);

		/**
		 * PERSISTENT EVENT
		 */

		// ZOOM IN
		this.persistentKeyEvents.push([
			["ctrl+plus", "f5", "ctrl+="],
			_.bind(function(e) {
				e.preventDefault();
				e.returnValue = false; // preventDefault() for IE
				this.zoom(1, undefined, true);
			}, this)
		]);

		// ZOOM OUT
		this.persistentKeyEvents.push([
			["ctrl+-", "f6"],
			_.bind(function(e) {
				e.preventDefault();
				e.returnValue = false; // preventDefault() for IE
				this.zoom(-1, undefined, true);
			}, this)
		]);

		this.persistentKeyEvents.push([
			["ctrl+0", "f9"],
			function(e) {
				if (!e.shiftKey) {
					e.preventDefault();
					f.emit("viewport.zoomToFit", {
						animated: true
					});
				}
			}
		]);

		/**
		 * PASSIVE EVENTS
		 */

		// activate zoom tool
		this.passiveKeyEvents.push([
			"z",
			_.bind(function(e) {
				e.preventDefault();
				f.emit("toolManager.activateTool", this);
			}, this),
			"keypress"
		]);

		// activate zoom tool until one zoom action is done
		this.passiveKeyEvents.push([
			"f8",
			_.bind(function(e) {
				e.preventDefault();
				deactiveToolAfterZoomAction = true;
				f.emit("toolManager.activateTool", this, true);
			}, this)
		]);

		/**
		 * ACTIVE EVENTS
		 */
		// cancel zoom in action
		this.activeKeyEvents.push([
			"esc",
			_.bind(function(e) {
				e.preventDefault();
				this.getCurrentState()["esc"](e);
			}, this),
			"keyup"
		]);

		// spacebar, drag zoom rectangle
		this.activeKeyEvents.push([
			"space",
			_.bind(function(e) {
				this.getCurrentState()["spacedown"](e);
			}, this),
			"keydown"
		]);

		this.activeKeyEvents.push([
			"space",
			_.bind(function(e) {
				this.getCurrentState()["spaceup"](e);
			}, this),
			"keyup"
		]);
	});

	/**
	 * Zoom in/out
	 *
	 * @param step {int} if positive zoom in, if negative zoom out
	 * @param worldMiddle {Point} what is the center of the zoom action
	 * @param animated {boolean} show animation
	 */
	this.zoom = function(step, worldMiddle, animated) {
		this._widget.getWorld().transform({
			zoom: step,
			middle: worldMiddle,
			animated: animated
		});
	};

	/**
	 * Zoom to fit to a selection
	 *
	 * @param selection {Rect}
	 */
	this.zoomToSelection = function(selection) {
		this._widget.getWorld().zoomToSceneRect(selection, true);
	};

	/**
	 * Can the tool be interrupted ?
	 */
	this.canBeInterrupted = function() {
		return this.getCurrentState().canBeInterrupted;
	};

	this.getDefaultCursor = function() {
		if (defaultZoom == -1) {
			return "cursor-zoomOut";
		} else {
			return "cursor-zoomIn";
		}
	};

	/**
	 * @override
	 */
	this._override("doActivation", function() {
		this._super();
		this.changeState("default");
	});

	/**
	 * @override
	 */
	this._override("deactivate", function(interruptRequest) {
		this._super(interruptRequest);
		this.changeState("default");
		cleanupCanvas();
	});
};

// extend from the AbstractTool
ZoomTool.prototype = Object.create(AbstractTool.prototype);
ZoomTool.prototype.constructor = ZoomTool;

/**
 * Get the icon offset of the zoom tool
 * @override
 */
ZoomTool.prototype.getIconOffset = function() {
	return new ShapeUtils.Size(6, 6);
};

/**
 * override the passivation process, register a persistent handle scroll wheel listener
 * @override
 */
ZoomTool.prototype.doPassivation = function() {
	AbstractTool.prototype.doPassivation.call(this);

	// add persistent mousewheel listener
	// zoom in/out when scrolling
	Mousewheel.addListener(_.throttle(_.bind(this.handleScrollWheel, this), 150));
};

/**
 * Handle scroll wheel events
 * @param {EventObj} event
 * @param {Number} delta
 */
ZoomTool.prototype.handleScrollWheel = function(event, delta) {
	// only react if scrolling in viewport
	if (this._widget.context.has(event.target).length > 0 && this._zoomOnScroll) {
		this.zoom(delta < 0 ? -1 : 1, this.getWorldPoint(event), true);
		event.preventDefault();
	}
};

export default ZoomTool;
