import ToolbarComponent from "@esko/toolbar/ToolbarComponent";
import ButtonGroup from "@esko/toolbar/ButtonGroup";
import ViewerToolbarButton from "@esko/web-viewer/layout/component/ViewerToolbarButton";
import LayoutConstants from "@esko/web-viewer/layout/LayoutConstants";
import CompareConstants from "@esko/web-viewer/compare/CompareConstants";
import $ from "jquery";
import _ from "underscore";
var RENDER_SPEED = 40;

/**
 * Toolbar component for selecting pages for Page List/Multipage documents. It provides buttons for previous/next page, as well as a text field allowing a user to type in a page.
 * If the active document is a page list, a Mark Page button is shown.
 *
 * @exports PageSelectionToolbarComponent
 * @memberof PagelistModule
 * @constructor
 * @param {Core.ViewerFacade} f
 * @param {PagelistModule} pagelistModule
 */
var PageSelectionToolbarComponent = function(f, pagelistModule) {
	ToolbarComponent.call(this, LayoutConstants.ToolbarComponent.PAGESELECTION, f);

	this.f = f;

	this._pagelistModule = pagelistModule;

	this._context = $("<div />").addClass("pageSelection");
	this.addObject({
		element: this._context
	});

	// Previous Page
	this._previousPageOrSpreadBtn = new ViewerToolbarButton(
		this.f,
		"previousPageOrSpread",
		"icon-WCR_Page-Previous",
		true,
		_.bind(function() {
			this._pagelistModule.previousPageOrSpread();
		}, this)
	);
	this._previousPageOrSpreadBtn.element.removeClass("btn");
	this._previousPageOrSpreadBtn.element.addClass("mini-btn");
	this._context.append(this._previousPageOrSpreadBtn.element);

	// page input
	this._currentPage = $('<input type="text" size="4" pattern="[0-9]*" />');
	this._currentPage.on("keydown", _.bind(this._currentPageChanged, this));
	this._currentPage.on(
		"focus",
		_.bind(function() {
			var parsedValue = this._readPageValue(this._currentPage);
			if (parsedValue < 0) {
				// if an invalid number was avialable, set to empty string
				parsedValue = "";
			}
			this._currentPage.val(parsedValue);
			this._currentPage.trigger("select");
		}, this)
	);
	this._context.append(this._currentPage);

	// Next Page
	this._nextPageOrSpreadBtn = new ViewerToolbarButton(
		this.f,
		"nextPageOrSpread",
		"icon-WCR_Page-Next",
		true,
		_.bind(function() {
			this._pagelistModule.nextPageOrSpread();
		}, this)
	);
	this._nextPageOrSpreadBtn.element.removeClass("btn");
	this._nextPageOrSpreadBtn.element.addClass("mini-btn");
	this._context.append(this._nextPageOrSpreadBtn.element);

	this.f.i18n.loadBundle(
		"viewer",
		function() {
			// merk current page
			this._toggleMarkPageBtn = new ViewerToolbarButton(
				this.f,
				"toggleMarkPage",
				"icon-WCR_Mark",
				true,
				_.bind(function() {
					this._pagelistModule.togglePageMarked();
				}, this)
			);

			this._context.append(this._toggleMarkPageBtn.element);

			// render buttons
			this.updateButtons = _.throttle(_.bind(this._updateButtons, this), RENDER_SPEED);

			this._currentPage.on("blur", this.updateButtons);

			// listen to Mark changed
			this.f.on(
				"doc.**.pagelistMarked.changed",
				_.bind(function(page) {
					// we only reflect the active page marked state here
					if (page === this.f.getDocumentController().getActivePage()) {
						this.updateButtons();
					}
				}, this)
			);

			// Page changes
			this.f.getDocumentController().whenActivePageChanged(this.updateButtons, this);
		},
		this
	);
};

PageSelectionToolbarComponent.prototype = Object.create(ToolbarComponent.prototype);
PageSelectionToolbarComponent.prototype.constructor = PageSelectionToolbarComponent;
PageSelectionToolbarComponent.toolbarName = LayoutConstants.ToolbarComponent.PAGESELECTION;

/**
 * Update the state of the buttons
 * Throttled version available via this.updateButtons()
 */
PageSelectionToolbarComponent.prototype._updateButtons = function() {
	var activePage = this.f.getDocumentController().getActivePage();
	var currentPageNum = activePage.getPagenum();

	var firstPageNum = activePage
		.getDocumentVersion()
		.getFirstPage()
		.getPagenum();
	var lastPageNum = activePage
		.getDocumentVersion()
		.getLastPage()
		.getPagenum();

	this._previousPageOrSpreadBtn.setEnabled(currentPageNum > firstPageNum);
	this._nextPageOrSpreadBtn.setEnabled(currentPageNum < lastPageNum);

	if (activePage.isPagelistMarked()) {
		this._toggleMarkPageBtn.setIcon("icon-WCR_Mark");
	} else {
		this._toggleMarkPageBtn.setIcon("icon-WCR_Unmark");
	}

	this._currentPage.val(this.f.i18n.t("PageNumberOfCount", activePage.getPagenum(), activePage.getDocumentVersion().getPageCount()));
};

/**
 * Show the mark button or not
 * @param {Boolean} show - True to show, false to hide
 */
PageSelectionToolbarComponent.prototype.showMarkButton = function(show) {
	if (show) {
		this._toggleMarkPageBtn.show();
	} else {
		this._toggleMarkPageBtn.hide();
	}

	this._emit("changed.content");
};

/**
 * Read the page number of an input field
 * @param {Element} element - input field\
 * @returns {Number}
 */
PageSelectionToolbarComponent.prototype._readPageValue = function(element) {
	var value = $(element).val();
	var parsedValue = parseInt(value, 10);
	if (!isNaN(parsedValue)) {
		// valid number
		var activeDocumentVersion = this.f.getDocumentController().getActiveDocumentVersion();
		// check boundaries
		if (parsedValue < 1) {
			parsedValue = 1;
		} else if (parsedValue > activeDocumentVersion.getPageCount()) {
			parsedValue = activeDocumentVersion.getPageCount();
		}

		return parsedValue;
	} else {
		return -1;
	}
};

/**
 * Change page when input was changed
 * @param event
 */
PageSelectionToolbarComponent.prototype._currentPageChanged = function(event) {
	if (event.which == 13 || event.keyCode == 13) {
		var parsedValue = this._readPageValue(event.target);
		if (parsedValue >= 0) {
			// change page
			var activeDocumentVersion = this.f.getDocumentController().getActiveDocumentVersion();
			var newPage = activeDocumentVersion.getPage(parsedValue);
			if (newPage && newPage !== this.f.getDocumentController().getActivePage()) {
				this.f.getDocumentController().changeActivePage(newPage);
			} else {
				if (!newPage) {
					this.f.emit("userMessage", "warn", this.f.i18n.t("WarningGetNonexistentPage"), true);
				}
				// re-render to reset
				this._updateButtons();
			}
		} else {
			// not valid number, re-render to reset
			this._updateButtons();
		}

		event.preventDefault();
		$(event.target).trigger("blur");
	}
};

export default PageSelectionToolbarComponent;
