/**
 *  @namespace Pagelist
 */
/**
 *  @namespace Pagelist
 */
import AbstractModule from "@esko/web-viewer/core/AbstractModule";

import PagelistWidget from "@esko/web-viewer/pagelist/PagelistWidget";
import PagelistToolbarComponent from "@esko/web-viewer/pagelist/PagelistToolbarComponent";
import PageSelectionToolbarComponent from "@esko/web-viewer/pagelist/PageSelectionToolbarComponent";
import PagelistDataExtensions from "@esko/web-viewer/pagelist/PagelistDataExtensions";
import ListenerUtil from "@esko/web-viewer/utils/ListenerUtil";
import CompareConstants from "@esko/web-viewer/compare/CompareConstants";
import _ from "underscore";
/**
 * This module is responsible for page lists and multi page files
 *
 * @exports PagelistModule
 * @constructor
 * @augments Core.AbstractModule
 * @param {Core.ViewerFacade} f
 * @param {Boolean} silenced
 */
var PagelistModule = function(f, silenced) {
	AbstractModule.call(this, f, silenced, "pagelist");
	ListenerUtil.mixin(this, f);

	/**
	 * Inject pagelist functionality in data objects
	 */
	PagelistDataExtensions(f);

	this._publishWidget(
		PagelistWidget.widgetName,
		_.bind(function(layoutType) {
			this._widget = this._createPageListWidget();
			this._activePageChanged(null, this.f.getActivePage());
			return this._widget;
		}, this)
	);

	this._toolbarComponent = null;
	this._pageSelectionToolbarComponent = null;

	this._publishToolbarComponent(
		PagelistToolbarComponent.toolbarName,
		_.bind(function(layoutType) {
			this._toolbarComponent = new PagelistToolbarComponent(this.f, this);
			this._activePageChanged(null, this.f.getActivePage());
			return this._toolbarComponent;
		}, this)
	);

	/**
	 * Toolbar Component: Browse through pages
	 */
	this._publishToolbarComponent(
		PageSelectionToolbarComponent.toolbarName,
		_.bind(function(layoutType) {
			this._pageSelectionToolbarComponent = new PageSelectionToolbarComponent(this.f, this);
			this._activePageChanged(null, this.f.getActivePage());
			return this._pageSelectionToolbarComponent;
		}, this)
	);

	this.f.getDocumentController().whenActivePageChanged(_.bind(this._activePageChanged, this));

	this._showInReaderSpread = false;
};

// inherit
PagelistModule.prototype = Object.create(AbstractModule.prototype);
// correct the constructor pointer
PagelistModule.prototype.constructor = PagelistModule;

/**
 * When the active page changes, check if we have a multipage document or not
 * @param {AbstractDocumentPage} oldPage - the oldPage, undefined if there was no previous page
 * @param {AbstractDocumentPage} newPage - the new active page
 */
PagelistModule.prototype._activePageChanged = function(oldPage, newPage) {
	if (this._widget) {
		this._widget.setEnabled(newPage.getDocumentVersion().isMultiPage());
	}
	if (this._toolbarComponent) {
		if (newPage.getDocumentVersion().isMultiPage()) {
			// pagelist are also multipage
			if (newPage.getDocumentVersion().isPageList()) {
				this._toolbarComponent.showMarkButton(true);
			} else {
				this._toolbarComponent.showMarkButton(false);
			}
			// this one will fire toolbar component changed visibility events
			this._toolbarComponent.show();
		} else {
			this._toolbarComponent.toggleReaderSpreadBtn(false);
			this._toolbarComponent.hide();
		}
	}
	if (this._pageSelectionToolbarComponent) {
		if (newPage.getDocumentVersion().isMultiPage()) {
			// pagelist are also multipage
			if (newPage.getDocumentVersion().isPageList()) {
				this._pageSelectionToolbarComponent.showMarkButton(true);
			} else {
				this._pageSelectionToolbarComponent.showMarkButton(false);
			}
		}
	}
};

/**
 * Toggle reader spread on or off in the compare module
 * @param {Boolean} activate - should reader spread be activated or not
 */
PagelistModule.prototype.toggleReaderspread = function(activate) {
	var compareModule = this.f.getModule("compare");
	if (compareModule) {
		if (activate) {
			compareModule.activateModeByName(CompareConstants.mode.READERSPREAD);
		} else {
			compareModule.activateModeByName();
		}
	}
};

/**
 * Is Reader spread active
 * @returns {Boolean}
 */
PagelistModule.prototype.isReaderSpreadActive = function() {
	var compareModule = this.f.getModule("compare");
	if (compareModule) {
		return compareModule.isReaderSpreadActive();
	} else {
		return false;
	}
};

/**
 * Change to first page
 */
PagelistModule.prototype.firstPage = function() {
	var activeDocumentPage = this.f.getDocumentController().getActivePage();
	if (
		activeDocumentPage.getPagenum() >
		activeDocumentPage
			.getDocumentVersion()
			.getFirstPage()
			.getPagenum()
	) {
		var firstPage = activeDocumentPage.getDocumentVersion().getFirstPage();
		this.f.getDocumentController().changeActivePage(firstPage);
	}
};

/**
 * Change to previous page
 */
PagelistModule.prototype.previousPage = function() {
	var activeDocumentPage = this.f.getDocumentController().getActivePage();
	if (
		activeDocumentPage.getPagenum() >
		activeDocumentPage
			.getDocumentVersion()
			.getFirstPage()
			.getPagenum()
	) {
		var previousPage = activeDocumentPage.getPreviousPage(undefined, this._isCompareActive());
		this.f.getDocumentController().changeActivePage(previousPage);
	}
};

/**
 * Change to next page
 */
PagelistModule.prototype.nextPage = function() {
	var activeDocumentPage = this.f.getDocumentController().getActivePage();

	if (
		activeDocumentPage.getPagenum() <
		activeDocumentPage
			.getDocumentVersion()
			.getLastPage()
			.getPagenum()
	) {
		var nextPage = activeDocumentPage.getNextPage(undefined, this._isCompareActive());
		this.f.getDocumentController().changeActivePage(nextPage);
	}
};

/**
 * Change to last page
 */
PagelistModule.prototype.lastPage = function() {
	var activeDocumentPage = this.f.getDocumentController().getActivePage();
	if (
		activeDocumentPage.getPagenum() <
		activeDocumentPage
			.getDocumentVersion()
			.getLastPage()
			.getPagenum()
	) {
		var lastPage = activeDocumentPage.getDocumentVersion().getLastPage();
		this.f.getDocumentController().changeActivePage(lastPage);
	}
};

/**
 * Change to previous page or spread
 */
PagelistModule.prototype.previousPageOrSpread = function() {
	var activePage = this.f.getDocumentController().getActivePage();

	if (this.isReaderSpreadActive()) {
		// previous spread
		var currentPageNum = activePage.getPagenum();
		if (currentPageNum % 2 === 1) {
			// for odd pages (on the right), do previous previous because we want to go to the previous spread
			var previousPreviousPage = activePage.getPreviousPage(2, this._isCompareActive());
			this.f.getDocumentController().changeActivePage(previousPreviousPage);
			return;
		}
	}

	// default previous page
	var previousPage = activePage.getPreviousPage(undefined, this._isCompareActive());
	this.f.getDocumentController().changeActivePage(previousPage);
};

/**
 * Change to next page or spread
 */
PagelistModule.prototype.nextPageOrSpread = function() {
	var activePage = this.f.getDocumentController().getActivePage();

	if (this.isReaderSpreadActive()) {
		// next spread
		var currentPageNum = activePage.getPagenum();
		if (currentPageNum % 2 === 0) {
			// for even pages (on the left), do next next because we want to go to the next spread
			var nextNextPage = activePage.getNextPage(2, this._isCompareActive());
			this.f.getDocumentController().changeActivePage(nextNextPage);
			return;
		}
	}

	// default next page
	var nextPage = activePage.getNextPage(undefined, this._isCompareActive());
	this.f.getDocumentController().changeActivePage(nextPage);
};

/**
 * Toggle the mark of a page
 * @param {AbstractDocumentPage} [page] - Page to toggle mark off. If not defined, it takes the current active page.
 */
PagelistModule.prototype.togglePageMarked = function(page) {
	if (page === undefined) {
		page = this.f.getDocumentController().getActivePage();
	}
	page.setPagelistMarked(!page.isPagelistMarked());
};

/**
 * Get all pages that are marked of the current active doc version.
 * @returns {Array<AbstractDocumentPage>} Array of all marked pages of the current active doc version.
 */
PagelistModule.prototype.getAllMarkedPages = function() {
	var activeDocVersion = this.f.getDocumentController().getActiveDocumentVersion();
	if (activeDocVersion !== null) {
		var allPages = activeDocVersion.getAllPages();
		return _.filter(allPages, function(page) {
			return page.isPagelistMarked();
		});
	} else {
		return [];
	}
};

/**
 * Clear the marks on all pages of the current acitve doc version.
 */
PagelistModule.prototype.clearAllMarkedPages = function() {
	_.invoke(this.getAllMarkedPages(), "setPagelistMarked", false);
};

/**
 * Mark all pages of the current active doc version
 */
PagelistModule.prototype.markAllPages = function() {
	var activeDocVersion = this.f.getDocumentController().getActiveDocumentVersion();
	if (activeDocVersion !== null) {
		var allPages = activeDocVersion.getAllPages();
		_.invoke(allPages, "setPagelistMarked", true);
	}
};

/**
 * Will check the compare module if comparing is active
 * @returns {Boolean}
 */
PagelistModule.prototype._isCompareActive = function() {
	var compareModule = this.f.getModule("compare");

	if (compareModule && compareModule.getActiveMode() !== null && !compareModule.isReaderSpreadActive()) {
		return true;
	} else {
		return false;
	}
};

/**
 * Returns the marked pages as string that is readable to the user.
 * E.g. pages 1,2,3,5,7,8,9 --> 1-3,5,7-9,
 * @returns {String} the marked pages as string
 */
PagelistModule.prototype.getAllMarkedPagesAsString = function() {
	var markedPages = this.getAllMarkedPages();
	var pagesString = "";
	var lastNumber = null;
	var openDash = false;
	for (var i = 0; i < markedPages.length; i++) {
		var page = markedPages[i].getPagenum();
		if (lastNumber === null) {
			pagesString += page;
		} else if (page === lastNumber + 1) {
			if (!openDash) {
				pagesString += "-";
				openDash = true;
			}
		} else {
			if (openDash) {
				pagesString += lastNumber;
			}
			pagesString += ", " + page;
			openDash = false;
		}
		lastNumber = page;
	}
	if (openDash) {
		pagesString += lastNumber;
	}
	return pagesString;
};

PagelistModule.prototype._createPageListWidget = function() {
	return new PagelistWidget(this.f, this);
};

export default PagelistModule;
