import _ from "underscore";
import ToolbarComponent from "@esko/toolbar/ToolbarComponent";
import ButtonGroup from "@esko/toolbar/ButtonGroup";
import ViewerToolbarButton from "@esko/web-viewer/layout/component/ViewerToolbarButton";
import DropdownToolbarButton from "@esko/web-viewer/layout/component/DropdownToolbarButton";
import LayoutConstants from "@esko/web-viewer/layout/LayoutConstants";
import CompareConstants from "@esko/web-viewer/compare/CompareConstants";

var RENDER_SPEED = 40;

/**
 * Toolbar component for pagelists/Multipages
 *
 * @exports PagelistToolbarComponent
 * @memberof PagelistModule
 * @constructor
 * @param {Core.ViewerFacade} f
 * @param {PagelistModule} pagelistModule
 */
var PagelistToolbarComponent = function(f, pagelistModule) {
	ToolbarComponent.call(this, LayoutConstants.ToolbarComponent.PAGELIST, f);

	this.f = f;

	this._pagelistModule = pagelistModule;

	this._buttonGroup = new ButtonGroup("pagelist");
	this._buttonGroup.setTitle(f.i18n, "viewer", "PagelistTitle");
	this.addButtonGroup(this._buttonGroup);

	// Page list buttons
	this._readerSpreadBtn = new ViewerToolbarButton(
		this.f,
		"readerSpreadView",
		"icon-WCR_ReaderSpread",
		true,
		_.bind(function() {
			this.toggleReaderSpread();
		}, this)
	);
	this._readerSpreadBtn.setTooltip("viewer", "ReaderSpreadTT");
	this._buttonGroup.addButton(this._readerSpreadBtn);

	this._previousPageOrSpreadBtn = new ViewerToolbarButton(
		this.f,
		"previousPageOrSpread",
		"icon-WCR_Page-Previous",
		true,
		_.bind(function() {
			this.previousPageOrSpread();
		}, this)
	);
	this._previousPageOrSpreadBtn.setTooltip("viewer", "PreviousPageTT");
	this._buttonGroup.addButton(this._previousPageOrSpreadBtn);

	this._nextPageOrSpreadBtn = new ViewerToolbarButton(
		this.f,
		"nextPageOrSpread",
		"icon-WCR_Page-Next",
		true,
		_.bind(function() {
			this.nextPageOrSpread();
		}, this)
	);
	this._nextPageOrSpreadBtn.setTooltip("viewer", "NextPageTT");
	this._buttonGroup.addButton(this._nextPageOrSpreadBtn);

	/**
	 * Mark	Pages
	 */
	this._toggleMarkPageBtn = new DropdownToolbarButton(
		this.f,
		"toggleMarkPage",
		"toggleMarkPage",
		true,
		_.bind(function() {
			this.toggleMarkPage();
		}, this)
	);

	this.f.i18n.loadBundle(
		"viewer",
		function() {
			// Mark
			this._toggleMarkPageBtn.addAction({
				name: "markOne",
				content: this.f.i18n.t("MarkedOn"),
				enabled: true,
				action: _.bind(function() {
					this._pagelistModule.togglePageMarked();
				}, this)
			});

			// Mark all pages
			this._toggleMarkPageBtn.addAction({
				name: "markAll",
				content: this.f.i18n.t("MarkedOnAll"),
				enabled: true,
				action: _.bind(function() {
					this._pagelistModule.markAllPages();
				}, this)
			});

			// Clear marks of all pages
			this._toggleMarkPageBtn.addAction({
				name: "markClearAll",
				content: this.f.i18n.t("MarkedOffAll"),
				enabled: true,
				action: _.bind(function() {
					this._pagelistModule.clearAllMarkedPages();
				}, this)
			});

			this._toggleMarkPageBtn.setTooltip("viewer", "MarkedOn");
			this._buttonGroup.addButton(this._toggleMarkPageBtn);

			// render buttons
			this.updateButtons = _.throttle(_.bind(this._updateButtons, this), RENDER_SPEED);

			/**
			 * Start listening if we have everything setup
			 */

			// Mark changed
			this.f.on(
				"doc.**.pagelistMarked.changed",
				_.bind(function(page) {
					// we only reflect the active page marked state here
					if (page === this.f.getDocumentController().getActivePage()) {
						this.updateButtons();
					}
				}, this)
			);

			// Compare mode changed
			this.f.on(
				"module.compare.changed.mode",
				_.bind(function(oldMode, newMode) {
					this.toggleReaderSpreadBtn(newMode && newMode.getName() === CompareConstants.mode.READERSPREAD);
				}, this)
			);

			// Page changes
			this.f.getDocumentController().whenActiveOrDifferencePageChangedAndParametersAvailableOrChange(this.updateButtons, this);
		},
		this
	);
};

PagelistToolbarComponent.prototype = Object.create(ToolbarComponent.prototype);
PagelistToolbarComponent.prototype.constructor = PagelistToolbarComponent;
PagelistToolbarComponent.toolbarName = LayoutConstants.ToolbarComponent.PAGELIST;

/**
 * Toggle reader spread mode on/off
 * @param {Boolean} activate
 */
PagelistToolbarComponent.prototype.toggleReaderSpread = function(activate) {
	if (activate === undefined) {
		activate = !this._readerSpreadBtn.isActive();
	}
	this.toggleReaderSpreadBtn(activate);
	this._pagelistModule.toggleReaderspread(activate);
	this.updateButtons();
};

/**
 * Toggle the reader spread button, will set the button active
 * @param {Boolean} activate
 */
PagelistToolbarComponent.prototype.toggleReaderSpreadBtn = function(activate) {
	this._readerSpreadBtn.setActive(activate);
};

/**
 * Is reader spread mode active
 * @returns {Boolean}
 */
PagelistToolbarComponent.prototype.isReaderSpreadActive = function() {
	return this._readerSpreadBtn.isActive();
};

/**
 * Update the state of the buttons
 * Throttled version available via this.updateButtons()
 */
PagelistToolbarComponent.prototype._updateButtons = function() {
	var activePage = this.f.getDocumentController().getActivePage();
	var currentPageNum = activePage.getPagenum();

	var firstPageNum = activePage
		.getDocumentVersion()
		.getFirstPage()
		.getPagenum();
	var lastPageNum = activePage
		.getDocumentVersion()
		.getLastPage()
		.getPagenum();

	this._previousPageOrSpreadBtn.setEnabled(currentPageNum > firstPageNum);
	this._nextPageOrSpreadBtn.setEnabled(currentPageNum < lastPageNum);

	if (this.isReaderSpreadActive()) {
		this._nextPageOrSpreadBtn.setTooltip("viewer", "NextSpreadTT");
		this._previousPageOrSpreadBtn.setTooltip("viewer", "PreviousSpreadTT");
	} else {
		this._nextPageOrSpreadBtn.setTooltip("viewer", "NextPageTT");
		this._previousPageOrSpreadBtn.setTooltip("viewer", "PreviousPageTT");
	}

	if (activePage.isPagelistMarked()) {
		this._toggleMarkPageBtn.setActionText("markOne", this.f.i18n.t("MarkedOff"));
		this._toggleMarkPageBtn.setTooltip("viewer", "MarkedOff");
		this._toggleMarkPageBtn.setIcon("icon-normal icon-WCR_Mark");
	} else {
		this._toggleMarkPageBtn.setActionText("markOne", this.f.i18n.t("MarkedOn"));
		this._toggleMarkPageBtn.setTooltip("viewer", "MarkedOn");
		this._toggleMarkPageBtn.setIcon("icon-normal icon-WCR_Unmark");
	}
};

/**
 * Change to previous page or spread
 */
PagelistToolbarComponent.prototype.previousPageOrSpread = function() {
	this._pagelistModule.previousPageOrSpread();
};

/**
 * Change to next page or spread
 */
PagelistToolbarComponent.prototype.nextPageOrSpread = function() {
	this._pagelistModule.nextPageOrSpread();
};

/**
 * Show the mark button or not
 * @param {Boolean} show - True to show, false to hide
 */
PagelistToolbarComponent.prototype.showMarkButton = function(show) {
	if (show) {
		this._toggleMarkPageBtn.show();
	} else {
		this._toggleMarkPageBtn.hide();
	}

	this._emit("changed.content");
};

/**
 * Toggle the current active page mark
 */
PagelistToolbarComponent.prototype.toggleMarkPage = function() {
	this._pagelistModule.togglePageMarked();
};

export default PagelistToolbarComponent;
