import AbstractWidget from "@esko/web-viewer/layout/AbstractWidget";
import Template from "@esko/template";
import DropdownMiniButton from "@esko/web-viewer/layout/component/DropdownMiniButton";
import LayoutConstants from "@esko/web-viewer/layout/LayoutConstants";
import mainTpl from "@esko/web-viewer/pagelist/PagelistWidget.tpl.html";
import subTpl from "@esko/web-viewer/pagelist/PagelistBlock.tpl.html";
import PagelistConstants from "@esko/web-viewer/pagelist/PagelistConstants";
import PagelistWidgetPageContextMenu from "@esko/web-viewer/pagelist/PagelistWidgetPageContextMenu";
import ListenerUtil from "@esko/web-viewer/utils/ListenerUtil";
import ShapeUtils from "@esko/shapeutils";
import $ from "jquery";
import _ from "underscore";
/**
 * @class A PageBlock is a small datastructure used internally for the template rendering. It  holds a left and right page
 * Keeps track if the block is unveiled or not (if it has a thumbnail or not)
 * and can give a thumbnail url
 * @constructor
 * @param {data.AbstractDocumentPage} leftPage - the left page of the block (or null/undefined)
 * @param {data.AbstractDocumentPage} rightPage - the right page of the block (or null/undefined)
 * @param {Number} leftPageNumber - the page number of the first page
 * @param {Number} rightPageNumber - the page number of the second page
 */
function PageBlock(leftPage, rightPage, leftPageNumber, rightPageNumber) {
	this.left = leftPage;
	this.right = rightPage;
	this.leftPageNumber = leftPageNumber;
	this.rightPageNumber = rightPageNumber;
	this.resetThumbnail();
}

/**
 * Get the thumbnail for a certain page
 * @param {String} which - for which page you want the thumbnail url 'left' or 'right'
 * @returns {String}
 */
PageBlock.prototype.getThumb = function(which) {
	if (!this.isUnveiled) {
		return require("@esko/web-viewer/layout/img/Empty_page.png");
	} else {
		return this.thumb[which];
	}
};
/**
 * Unveil the pageblock: set the thumbnail urls
 * @param {String} leftThumbURL - URL for the left thumbnail
 * @param {String} rightThumbURL - URL for the right thumbnail
 */
PageBlock.prototype.unveil = function(leftThumbURL, rightThumbURL) {
	this.isUnveiled = true;
	this.thumb = {
		left: leftThumbURL,
		right: rightThumbURL
	};
};

/**
 * Reset the thumbnails for this block
 */
PageBlock.prototype.resetThumbnail = function() {
	this.isUnveiled = false;
	this.thumb = {};
};

/**
 * @exports PagelistWidget
 * @augments Layout.AbstractWidget
 * @class The Pagelist widget shows an overview of pages in a document.
 * @constructor
 * @memberof Pagelist
 * @mixes {Utils.ListenerUtilMixin}
 * @param {Core.ViewerFacade} f
 * @param {Pagelist.PagelistModule} pagelistModule
 */
var PagelistWidget = function PagelistWidget(f, pagelistModule) {
	AbstractWidget.call(this, f, LayoutConstants.Widget.PAGELIST);
	ListenerUtil.mixin(this, f);
	this.setIcon("icon icon-WCR_Pagelist");
	this._pagelistModule = pagelistModule;

	this._currentDocumentVersion = null;
	this._thumbnailSize = null;
	this._pageBlocks = [];

	this.unveil = _.debounce(this._unveil, 100);

	// initialize template
	this.t = new Template(f.i18n, mainTpl);
	this.t.registerTemplate("web-viewer/pagelist/PagelistBlock", subTpl);
	// open a page when clicking on a thumb
	this.t.bind("click", "img.thumb", this._openPage, this);
	// context menu of a page block
	this.t.bind("contextmenu", "div.pageBlock", this._contextMenu, this);

	// some helper methods for the template
	this.t.d.getPageActiveClass = _.bind(this._getPageActiveClass, this);
	this.t.d.getPageBlocks = _.bind(this._getPageBlocks, this);
	this.t.d.getPageHeaderClass = _.bind(this._getPageHeaderClass, this);
	this.t.d.getPageFooterClass = _.bind(this._getPageFooterClass, this);
	this.t.d.getPageHeaderIconClass = _.bind(this._getPageHeaderIconClass, this);

	// option button
	this.optionButton = new DropdownMiniButton(this.f, "pagelistSettings", "icon-WCR_Settings-submenu");

	// page context menu (for pagelists to mark pages)
	this._pageContextMenu = null;

	// render when page has changed
	this.f
		.getDocumentController()
		.whenActiveOrDifferencePageChangedAndParametersAvailable(function(activePage, referencePage, oldPage, isActiveChanged) {
			if (isActiveChanged && oldPage !== undefined && activePage.getDocumentVersion() !== oldPage.getDocumentVersion()) {
				this.update(undefined, true);
			} else {
				this.update();
			}
			this.scrollToActivePage();
		}, this);

	// re-render when the page got/removed annotations
	this.f.on(
		"doc.**.hasAnnotations.changed",
		_.bind(function(oldvalue, newvalue, page) {
			this.update(page);
		}, this)
	);

	// re-render when a page list mark has changed
	this.f.on(
		"doc.**.pagelistMarked.changed",
		_.bind(function(page) {
			this.update(page);
		}, this)
	);

	this.on(
		"visible",
		_.bind(function() {
			if (this.isVisible()) {
				this._unveil();
			}
		}, this)
	);

	//  update thumbnails when size has changed
	f.getCurrentUser().on("options.changed.PREFERENCES_PAGELIST_ICON_SIZE", _.bind(this.resetThumbnails, this));

	// load bundle and define options for thumbnail size
	this.f.loadBundle(
		"viewer",
		function() {
			this.setTitle(this.f.t("PagelistTitle"));

			function toggleThumbnailSize(size, toggle) {
				if (toggle) {
					this.setThumbnailSize(size);
				}
			}

			this.optionButton.addOption({
				name: "PagelistIconSizeSmall",
				content: this.f.i18n.t("PagelistIconSizeSmall"),
				onchange: _.bind(toggleThumbnailSize, this, PagelistConstants.viewerThumbSize.SMALL),
				selected: this.f.getOption("PREFERENCES_PAGELIST_ICON_SIZE") == PagelistConstants.viewerThumbSize.SMALL,
				userCannotUncheck: true
			});
			this.optionButton.addOption({
				name: "PagelistIconSizeMedium",
				content: this.f.i18n.t("PagelistIconSizeMedium"),
				onchange: _.bind(toggleThumbnailSize, this, PagelistConstants.viewerThumbSize.MEDIUM),
				selected: this.f.getOption("PREFERENCES_PAGELIST_ICON_SIZE") == PagelistConstants.viewerThumbSize.MEDIUM,
				userCannotUncheck: true
			});
			this.optionButton.addOption({
				name: "PagelistIconSizeLarge",
				content: this.f.i18n.t("PagelistIconSizeLarge"),
				onchange: _.bind(toggleThumbnailSize, this, PagelistConstants.viewerThumbSize.LARGE),
				selected: this.f.getOption("PREFERENCES_PAGELIST_ICON_SIZE") == PagelistConstants.viewerThumbSize.LARGE,
				userCannotUncheck: true
			});

			this.optionButtonInitialized = true;
		},
		this
	);
};

PagelistWidget.prototype = Object.create(AbstractWidget.prototype);
PagelistWidget.prototype.constructor = PagelistWidget;
PagelistWidget.widgetName = LayoutConstants.Widget.PAGELIST;

/**
 * Reset all thumbnails
 */
PagelistWidget.prototype.resetThumbnails = function() {
	this._calculateThumbSize();

	_.each(this._pageBlocks, function(block) {
		block.resetThumbnail();
	});

	this.render();
};

/**
 * @override
 */
PagelistWidget.prototype.setContext = function(context, headerExtra) {
	AbstractWidget.prototype.setContext.call(this, context, headerExtra);
	this.context.addClass("pagelistWidget");
	this.t.setContext(this.context);

	// unveil thumbnails as we are scrolling
	this.context.on("scroll", _.bind(this.unveil, this));

	// add buttons to the widget header
	this.headerExtra.append(this.optionButton.element);

	this.update(undefined, true);
};

/**
 * Update the info displayed in the widget
 * @param {AbstractDocumentPage} [page] - The page that is changed. If not defined, it assumes it needs to update for the active page
 * @param {Boolean} [forceAll] - Force an update of all page blocks or not
 */
PagelistWidget.prototype.update = function(page, forceAll) {
	var docVersion = this.f.getActivePage().getDocumentVersion();
	if (docVersion !== this._currentDocumentVersion || forceAll) {
		this._currentDocumentVersion = docVersion;
		this.f.getActivePage().whenParametersAvailable(
			_.bind(function() {
				this._pageBlocks = [];

				this._calculateThumbSize();

				var binding = docVersion.getDocument().get("binding");
				// First page
				var firstPage = docVersion.getPage(1);

				if (binding === PagelistConstants.binding.LEFT || binding === undefined) {
					if (firstPage) {
						this._pageBlocks.push(new PageBlock(null, firstPage));
					} else {
						this._pageBlocks.push(new PageBlock(null, undefined, null, 1));
					}
				} else if (binding === PagelistConstants.binding.RIGHT) {
					if (firstPage) {
						this._pageBlocks.push(new PageBlock(firstPage, null));
					} else {
						this._pageBlocks.push(new PageBlock(undefined, null, 1, null));
					}
				}
				var pageCount = docVersion.getPageCount();

				// Other pages
				for (var i = 2; i <= pageCount; i = i + 2) {
					firstPage = docVersion.getPage(i);
					var secondPage = docVersion.getPage(i + 1); // for the last page this will return undefined
					var firstPageNumber = i;
					var secondPageNumber = null;
					if (i + 1 <= docVersion.getPageCount()) {
						secondPageNumber = i + 1;
					}

					// if undefined, assume left binding, otherwise check only the firstPage binding (secondPage just follows that binding)
					if (binding === undefined || binding === PagelistConstants.binding.LEFT) {
						this._pageBlocks.push(new PageBlock(firstPage, secondPage, firstPageNumber, secondPageNumber));
					} else if (binding === PagelistConstants.binding.RIGHT) {
						this._pageBlocks.push(new PageBlock(secondPage, firstPage, secondPageNumber, firstPageNumber));
					}
				}
				this.render();
			}, this)
		);
	} else if (this.context) {
		// update the page without rerendering the whole damn thing
		if (page === undefined) {
			page = this.f.getActivePage();
			this.context.find(".page.active").toggleClass("active", false);
		}
		var pageBlock = this._getPageBlock(page);
		// only update if we have all the needed information
		var pageBlockElement = this.context.find(".pageBlock[data-bind='" + pageBlock.index + "']");
		if (pageBlockElement.length > 0 && pageBlock && pageBlock.pageBlock) {
			pageBlockElement.html(this.t.d.callTemplate("web-viewer/pagelist/PagelistBlock", pageBlock.pageBlock, "block"));
		}
	}
};

/**
 * Set the thumbnail size
 * @param size
 */
PagelistWidget.prototype.setThumbnailSize = function(size) {
	this.f.setOption("PREFERENCES_PAGELIST_ICON_SIZE", size);
};

/**
 * Calculate the thumbnail size based on the active page
 */
PagelistWidget.prototype._calculateThumbSize = function() {
	var thumbSize = _.invert(PagelistConstants.viewerThumbSize)[this.f.getOption("PREFERENCES_PAGELIST_ICON_SIZE")];
	this.t.d.pagelistSize = thumbSize.toLowerCase();
	var width = PagelistConstants.thumbWidth[thumbSize];
	var pageSize = this.f.getActivePage().getPageSize();
	if (pageSize) {
		this._thumbnailSize = new ShapeUtils.Size(width, Math.round((pageSize.height * width) / pageSize.width));
	} else {
		this._thumbnailSize = new ShapeUtils.Size(width, width);
	}
	this.t.d.thumbSize = this._thumbnailSize;

	if (this.optionButtonInitialized) {
		this.optionButton.toggleOption(
			"PagelistIconSizeSmall",
			this.f.getOption("PREFERENCES_PAGELIST_ICON_SIZE") == PagelistConstants.viewerThumbSize.SMALL
		);
		this.optionButton.toggleOption(
			"PagelistIconSizeMedium",
			this.f.getOption("PREFERENCES_PAGELIST_ICON_SIZE") == PagelistConstants.viewerThumbSize.MEDIUM
		);
		this.optionButton.toggleOption(
			"PagelistIconSizeLarge",
			this.f.getOption("PREFERENCES_PAGELIST_ICON_SIZE") == PagelistConstants.viewerThumbSize.LARGE
		);
	}
};

/**
 * Get the list of pageblocks
 * @private
 * @returns {Array}
 */
PagelistWidget.prototype._getPageBlocks = function() {
	if (this._currentDocumentVersion === null) {
		this.update();
	}
	return this._pageBlocks;
};

/**
 * Get the page block + block index based on the given page
 * @private
 * @param {AbstractDocumentPage} page - The page to get the page block + block index of
 * @returns {Object} result.pageBlock = The page block the page number belongs to, result.index = The index of the page block
 */
PagelistWidget.prototype._getPageBlock = function(page) {
	var pageBlocks = this._getPageBlocks();
	var index = Math.floor(page.getPagenum() / 2);
	return {
		pageBlock: pageBlocks[index],
		index: index
	};
};

/**
 * Get the thumbnail size (and calculate if necessary)
 * @returns {ShapeUtils.Size}
 */
PagelistWidget.prototype.getThumbnailSize = function() {
	if (this._currentDocumentVersion === null) {
		this.update();
	}
	return this._thumbnailSize;
};

/**
 * Render the widget
 */
PagelistWidget.prototype.render = function() {
	this.t.render();
	this.unveil();
};

/**
 * This is a willfill widget :)
 * @returns {Boolean}
 */
PagelistWidget.prototype.willFill = function() {
	return true;
};

/**
 * Open a page
 * @param el
 * @param event
 */
PagelistWidget.prototype._openPage = function(el, event) {
	var pageNumber = $(el)
		.parent(".page")
		.attr("data-bind-page");
	var page = this.f
		.getActivePage()
		.getDocumentVersion()
		.getPage(pageNumber);
	if (page) {
		this.f.getDocumentController().changeActivePage(page);
	}
};

/**
 * Unveil all visible thumbnails
 */
PagelistWidget.prototype._unveil = function() {
	if (this.context) {
		// first we need to determine which pageblocks are visible at the moment
		var scrollTop = this.context.scrollTop();

		// we limit the innerheight to the window innerheight.
		// in some situations the widgetwrapper didn't set a height yet to the context of the widget when _unveil is called
		// we never want to load everything in one go, so only load thumbnails that could be visible on a fullscreen widget
		var innerHeight = Math.min(this.context.innerHeight() || null, window.innerHeight);
		var pageBlockSize = this.getPageBlockSize();

		var columns = 1;
		if (innerHeight > 0 && pageBlockSize.width > 0) {
			columns = Math.max(1, Math.floor((this.context.innerWidth() || null) / pageBlockSize.width));

			// first calculate the start row and determine which pageblock is the first in this row (based on # columns)
			var startRow = Math.floor(scrollTop / pageBlockSize.height);
			var startIndex = startRow * columns;

			// now calculate the end row and determine which pageblock is the last in this row (based on # columns)
			var endRow = Math.ceil((scrollTop + innerHeight) / pageBlockSize.height);
			var endIndex = endRow * columns;

			// unveil every visible page block
			var blockUnveilingPromises = [];
			_.range(Math.max(0, startIndex), Math.min(endIndex, this._pageBlocks.length)).forEach(
				function(i) {
					var block = this._pageBlocks[i];
					if (!block.isUnveiled) {
						blockUnveilingPromises.push(
							this._unveilPageBlock(block).then(
								function() {
									// find the html element of the pageblock and update it
									var blockElement = this.context.find(".pageBlock[data-bind='" + i + "']");
									blockElement.find(".left img.thumb").attr("src", block.getThumb("left"));
									blockElement.find(".right img.thumb").attr("src", block.getThumb("right"));
								}.bind(this)
							)
						);
					}
				}.bind(this)
			);

			Promise.all(blockUnveilingPromises).then(
				function() {
					this._emit("unveiled", { startIndex: startIndex, endIndex: endIndex });
				}.bind(this)
			);
		}
	}
};

/**
 * Unveil a specific pageblock (not this is only on logical level, luckily the _unveil also updates the html)
 * @param block
 */
PagelistWidget.prototype._unveilPageBlock = function(block) {
	if (!block.isUnveiled) {
		var leftRightThumbURL = this._getLeftRightThumbURL(block);
		var leftThumbURL = Promise.resolve(leftRightThumbURL.leftThumbURL || "");
		var rightThumbURL = Promise.resolve(leftRightThumbURL.rightThumbURL || "");

		return Promise.all([leftThumbURL, rightThumbURL]).then(function(thumbs) {
			block.unveil(thumbs[0], thumbs[1]);
		});
	}
};

/**
 * Generates img urls for a given block
 *
 * @param {PageBlock} block
 * @returns object with properties leftThumbURL and rightThumbURL
 */
PagelistWidget.prototype._getLeftRightThumbURL = function(block) {
	return {
		leftThumbURL: "",
		rightThumbURL: ""
	};
};

/**
 * When the widget is resized, unveil thumbnails
 */
PagelistWidget.prototype.resized = function() {
	this.unveil();
};

/**
 * Check if a page is active
 * @param page
 * @private
 * @returns
 */
PagelistWidget.prototype._getPageActiveClass = function(page) {
	return this.f.getActivePage() === page ? "active" : "";
};

/**
 * Get the page block header class
 * @param {AbstractDocumentPage} page
 * @returns {String} class (e.g. Pending)
 */
PagelistWidget.prototype._getPageHeaderClass = function(page) {
	return "";
};

/**
 * Get the page block footer class
 * @param {AbstractDocumentPage} page
 * @returns {String} class (e.g. Pending)
 */
PagelistWidget.prototype._getPageFooterClass = function(page) {
	return "";
};

/**
 * Gets the page block header class used to render icons
 *
 * @param {any} page
 * @returns
 */
PagelistWidget.prototype._getPageHeaderIconClass = function(page) {
	return "";
};

/**
 * Scroll to the active page if needed
 */
PagelistWidget.prototype.scrollToActivePage = function() {
	if (this.context) {
		var page = this.context.find("div.page[data-bind-page=" + this.f.getActivePage().getPagenum() + "]");
		if (page.length > 0) {
			var pageTop = page.offset().top;
			var pageBottom = pageTop + (page.outerHeight(true) || null);
			var contextTop = this.context.offset().top;
			var contextBottom = contextTop + (this.context.innerHeight() || null);

			// only scroll if page is out of view
			if (pageTop < contextTop || pageBottom > contextBottom) {
				this.context.animate({
					scrollTop: this.context.scrollTop() + (pageTop - contextTop) - 10
				});
			}
		}
	}
};

/**
 * We need a faster dimension calculation for huge pagelists/multipage documents, take the size of 1 pageblock and calculate dimensions from there
 * @override
 * @returns {ShapeUtils.Size}
 */
PagelistWidget.prototype.getPreferredSize = function() {
	var pageBlockSize = this.getPageBlockSize();
	var columns = 1;
	if (this.context) {
		columns = Math.floor((this.context.innerWidth() || null) / pageBlockSize.width);
	}
	return new ShapeUtils.Size(pageBlockSize.width * columns, pageBlockSize.height * this._pageBlocks.length);
};

/**
 * Get the current size of a PageBlock
 * @returns {ShapeUtils.Size}
 */
PagelistWidget.prototype.getPageBlockSize = function() {
	var size = new ShapeUtils.Size();
	if (this.context) {
		var pageBlock = this.context.find(".pageBlock").first();
		if (pageBlock.length > 0) {
			size.width = $(pageBlock).outerWidth(true) || null;
			size.height = $(pageBlock).outerHeight(true) || null;
		}
	}

	return size;
};

/**
 * Called when context menu opened on page block
 * @param el - Element that triggered the contextmenu event
 * @param event - Event
 */
PagelistWidget.prototype._contextMenu = function(el, event) {
	// don't show context menu for empty pages, but still prevent default
	// the page instead of the target element
	el = $(event.target).parent(".page");
	if (el && !el.hasClass("empty")) {
		var docVersion = this.f.getDocumentController().getActiveDocumentVersion();
		if (docVersion !== null && docVersion.isPageList()) {
			// load bundle first
			this.f.i18n.loadBundle(
				"viewer",
				function() {
					// check page
					var pageNum = el.attr("data-bind-page");
					if (pageNum !== undefined) {
						var page = docVersion.getPage(pageNum);
						if (page !== undefined) {
							if (this._pageContextMenu === null) {
								this._pageContextMenu = new PagelistWidgetPageContextMenu(this.f, this._pagelistModule);
							}
							this._pageContextMenu.show(page, new ShapeUtils.Point(event.clientX, event.clientY));
						}
					}
				},
				this
			);
		}
	}
	event.stopPropagation();
	event.preventDefault();
};

export default PagelistWidget;
