import PrivateChannel from "@esko/web-viewer/core/PrivateChannel";
import PrintConstants from "@esko/web-viewer/print/PrintConstants";
import AbstractModel from "@esko/web-viewer/core/AbstractModel";
import DataConstants from "@esko/web-viewer/data/DataConstants";
import ViewerUtils from "@esko/web-viewer/utils/ViewerUtils";
import $ from "jquery";
import "underscore";
/**
 * @exports PrintModel
 * @memberof PrintModule
 * @class PrintModel holds all the information for printing. It saves all printing options and persists them to user preferences.
 * @constructor
 *
 * @param {Core.ViewerFacade} f - Viewer facade
 */
var PrintModel = function(f) {
	AbstractModel.call(this, f, "printModel");
	this.f = f;
	/**
	 * @member {PrintConstants.PrintType} - print mode used
	 * @private
	 */
	this._printType = this.f.getCurrentUser().getOption("PREFERENCES_PRINT_TYPE");

	/**
	 * @member {PrintConstants.PrintArea} - what image area is used for print
	 * @private
	 */
	this._printArea = this.f.getCurrentUser().getOption("PREFERENCES_PRINT_AREA");

	/**
	 * @member {PrintConstants.PrintChannels} - print all channels or only selected channels
	 * @private
	 */
	this._printChannels = this.f.getCurrentUser().getOption("PREFERENCES_PRINT_CHANNELS");

	/**
	 * @member {Number} the resolution to print
	 * @private
	 */
	this._printResolution = ViewerUtils.parseFloat(this.f.getCurrentUser().getOption("PREFERENCES_PRINT_RESOLUTION"));

	/**
	 * @member {CoreConstants.ResolutionUnits}
	 */
	this._printResolutionUnit = this.f.getCurrentUser().getOption("PREFERENCES_UNIT_RESOLUTION");

	/**
	 * @member {PrintConstants.PrintAnnotationsSource}
	 * @private
	 */
	this._printAnnotationsSource = PrintConstants.PrintAnnotationsSource.PRINTANNOTATIONSCURRENT;

	/**
	 * @member {Boolean} - is the viewport zoomed in, for print area the option current view area is added
	 * @private
	 */
	this._isViewportZoomed = false;

	/**
	 * @member {Boolean} - is there a selection (measure tools), for print area the option current selection is added
	 * @private
	 */
	this._hasViewportActiveSelection = false;

	/**
	 * @member {Boolean} - are all the channels selected, options to print all channels or only selected channels is added
	 * @private
	 */
	this._areAllChannelsSelected = true;

	/**
	 * @member {DataConstants.documentPageType}
	 * @private
	 */
	this._fileType = null;

	/**
	 * @member {Boolean} - Is reader spread active
	 * @private
	 */
	this._readerSpreadActive = false;
};

/**
 * Get the print type
 * @returns {PrintConstants.PrintType}
 */
PrintModel.prototype.getPrintType = function() {
	return this._printType;
};

/**
 * Get the print area option
 * @returns {PrintConstants.PrintArea}
 */
PrintModel.prototype.getPrintArea = function() {
	return this._printArea;
};

/**
 * Get the option for print channels
 * @returns {PrintConstants.PrintChannels}
 */
PrintModel.prototype.getPrintChannels = function() {
	return this._printChannels;
};

/**
 * Get the resolution to print
 * @returns {Number} - in PPMM
 */
PrintModel.prototype.getResolution = function() {
	return this._printResolution;
};

/**
 * Get the unit for the print resolution
 * @returns {CoreConstants.ResolutionUnits}
 */
PrintModel.prototype.getResolutionUnit = function() {
	return this._printResolutionUnit;
};

/**
 * Get the annotation source to print
 * @returns {PrintConstants.PrintAnnotationsSource}
 */
PrintModel.prototype.getPrintAnnotationsSource = function() {
	return this._printAnnotationsSource;
};

/**
 * Does the view have to show the options for print area
 * @returns {Boolean}
 */
PrintModel.prototype.showPrintArea = function() {
	return this._printType === PrintConstants.PrintType.PRINTIMAGEAREA && (this._hasViewportActiveSelection || this._isViewportZoomed);
};

/**
 * Does the view have to show the options to print all channels or only the selected channels
 * @returns {Boolean}
 */
PrintModel.prototype.showPrintChannels = function() {
	return !this._areAllChannelsSelected && this._printType !== PrintConstants.PrintType.PRINTANNOTATIONOVERVIEW;
};

/**
 * Does the view have to show the option to specify the resolution
 * @returns {Boolean}
 */
PrintModel.prototype.showResolution = function() {
	return this._printType === PrintConstants.PrintType.PRINTIMAGEAREA && this._fileType === DataConstants.documentPageType.GRAPHICS;
};

/**
 * Set the print mode
 * @param {PrintConstants.PrintType} printType
 */
PrintModel.prototype.setPrintType = function(printType) {
	this._printType = printType;
	this.f.getCurrentUser().setOption("PREFERENCES_PRINT_TYPE", this._printType);
};

/**
 * Set the printing area
 * @param {PrintConstants.PrintArea} printArea
 */
PrintModel.prototype.setPrintArea = function(printArea) {
	this._printArea = printArea;
	this.f.getCurrentUser().setOption("PREFERENCES_PRINT_AREA", this._printArea);
};

/**
 * Set the print channels option
 * @param {PrintConstants.PrintChannels} printChannels
 */
PrintModel.prototype.setPrintChannels = function(printChannels) {
	this._printChannels = printChannels;
	this.f.getCurrentUser().setOption("PREFERENCES_PRINT_CHANNELS", this._printChannels);
};

/**
 * Set the resolution to print in
 * @param {Number} resolution -  in PPMM
 */
PrintModel.prototype.setPrintResolution = function(resolution) {
	this._printResolution = resolution;
	this.f.getCurrentUser().setOption("PREFERENCES_PRINT_RESOLUTION", this._printResolution + "");
};

/**
 * Set the resolution unit to print
 * @param {CoreConstants.ResolutionUnits} resolutionUnit
 */
PrintModel.prototype.setPrintResolutionUnit = function(resolutionUnit) {
	this._printResolutionUnit = resolutionUnit;
};

/**
 * Set the annotation source to print
 * @param {PrintConstants.PrintAnnotationsSource} printAnnotationsSource
 */
PrintModel.prototype.setPrintAnnotationsSource = function(printAnnotationsSource) {
	this._printAnnotationsSource = printAnnotationsSource;
};

/**
 * Set the if the viewport is zoomed in, to show options for image area printing mode
 * @param {Boolean} isZoomed
 */
PrintModel.prototype.setViewportZoomed = function(isZoomed) {
	this._isViewportZoomed = isZoomed;
};

/**
 * set hasViewportActiveSelection, for when a selection (measure tools) is made or not
 * @deprecated
 * @param {Boolean} hasViewportActiveSelection
 */
PrintModel.prototype.setViewportActiveSelection = function(hasViewportActiveSelection) {
	this._hasViewportActiveSelection = hasViewportActiveSelection;
};

/**
 * set if all channels are selected
 * @param {Boolean} areAllChannelsSelected
 */
PrintModel.prototype.setAreAllChannelsSelected = function(areAllChannelsSelected) {
	this._areAllChannelsSelected = areAllChannelsSelected;
};

/**
 * Show the print annotation source options
 * @returns {Boolean} True if it needs to be shown, false otherwise
 */
PrintModel.prototype.showPrintAnnotationsSource = function() {
	return (
		this._printType === PrintConstants.PrintType.PRINTANNOTATIONOVERVIEW &&
		this.f.getDocumentController().getDifferencePage() &&
		!this.isReaderSpreadActive()
	);
	// give the choice if there is a diff page too and it's not in readerspread
};

/**
 * Set the page type that's being printed
 * @param {DataConstants.documentPageType} pageType
 */
PrintModel.prototype.setPageType = function(fileType) {
	this._fileType = fileType;
};

/**
 * Get the type of the file that's being printed
 * @returns {DataConstants.documentPageType}
 */
PrintModel.prototype.getPageType = function() {
	return this._fileType;
};

/**
 * Set whether reader spread is active
 * @param {Boolean} readerSpreadActive
 */
PrintModel.prototype.setReaderSpreadActive = function(readerSpreadActive) {
	this._readerSpreadActive = readerSpreadActive;
};

/**
 * Get whether reader spread is active
 * @returns {Boolean} True if active, false otherwise
 */
PrintModel.prototype.isReaderSpreadActive = function() {
	return this._readerSpreadActive;
};

export default PrintModel;
