/**
 * @namespace PrintModule
 */
/**
 * @namespace PrintModule
 */
import AbstractModule from "@esko/web-viewer/core/AbstractModule";

import ListenerUtil from "@esko/web-viewer/utils/ListenerUtil";
import PrintConstants from "@esko/web-viewer/print/PrintConstants";
import PrintModel from "@esko/web-viewer/print/PrintModel";
import PrintPopup from "@esko/web-viewer/print/PrintPopup";
import AbstractPrintDataAdapter from "@esko/web-viewer/print/AbstractPrintDataAdapter";
import ViewerUtils from "@esko/web-viewer/utils/ViewerUtils";
import DataConstants from "@esko/web-viewer/data/DataConstants";
import _ from "underscore";
import Template from "@esko/template";
import PrintImageAreaTemplate from "./PrintImageArea.tpl.html";
import PrintCurrentViewTemplate from "./PrintCurrentView.tpl.html";
/**
 * Initialize the Print Module
 *
 * @exports PrintModule
 * @memberof PrintModule
 * @class
 * The print module enables printing in the viewer.
 * @constructor
 * @augments {Core.AbstractModule}
 * @param {Core.ViewerFacade} f
 * @param {Boolean} silenced - if true, the module shouldn't publish tools/widgets/toolbarcomponents
 */
var PrintModule = function(f, silenced, dataAdapter) {
	AbstractModule.call(this, f, silenced, "print", dataAdapter || new AbstractPrintDataAdapter(f));
	ListenerUtil.mixin(this, f);
	this.f = f;
	/**
	 * @member {PrintModule.PrintModel}
	 */
	this._printModel = null;

	/**
	 * @member {PrintModule.PrintPopup}
	 */
	this._printPopup = null;

	this.f.on("ready.printImageURL", _.bind(this._handlePrintUrl, this));
	this.f.on("ready.printPageURL", function(url) {
		window.open(url, "_blank");
	});
};

PrintModule.prototype = Object.create(AbstractModule.prototype);
PrintModule.prototype.constructor = PrintModule;

/**
 * Opens the popup for print
 */
PrintModule.prototype.openPrintPopup = function() {
	this._initializePrintModel();
	this._initializePrintPopup();

	this._printPopup.show();

	this._checkForUnsavedAnnotations();
};

/**
 * initializes the print model with correct options which it gets from the state of other modules
 */
PrintModule.prototype._initializePrintModel = function() {
	var isZoomed = false;
	var areAllChannelsSelected = true;

	if (this._printModel === null) {
		this._printModel = new PrintModel(this.f);
	}

	var state;
	//get state of the viewport
	var viewportModule = this.f.getModule("viewport");
	if (viewportModule !== undefined) {
		state = viewportModule.getState();
		isZoomed = !state.world.isFullView;
		this._printModel.setViewportZoomed(isZoomed);
	}

	//get state of the graphics
	var graphicsModule = this.f.getModule("graphics");
	if (graphicsModule) {
		state = graphicsModule.getState();
		var channels = state.graphicsModel.channels;
		areAllChannelsSelected = !_.any(channels, function(channel) {
			return !channel.enabled;
		});
		this._printModel.setAreAllChannelsSelected(areAllChannelsSelected);
	}

	//for when viewed file is CAD, no channels!
	var activePage = this.f.getDocumentController().getActivePage();
	if (activePage.hasCad()) {
		//the active page has CAD
		this._printModel.setPageType(DataConstants.documentPageType.CAD);
		this._printModel.setAreAllChannelsSelected(true); //to make sure the option to print all/selected annotations is not shown
	} else if (activePage.hasSvg && activePage.hasSvg()) {
		this._printModel.setPageType(DataConstants.documentPageType.SVG);
	} else if (activePage.hasVideo && activePage.hasVideo()) {
		this._printModel.setPageType(DataConstants.documentPageType.VIDEO);
	} else {
		this._printModel.setPageType(DataConstants.documentPageType.GRAPHICS);
	}

	// get if reader spread is active
	var compareModule = this.f.getModule("compare");
	if (compareModule) {
		this._printModel.setReaderSpreadActive(compareModule.isReaderSpreadActive());
	}
};

/**
 * initializes the print popup and registers all the listeners if the popup was not yet initialized
 */
PrintModule.prototype._initializePrintPopup = function() {
	if (this._printPopup === null) {
		this._printPopup = new PrintPopup(this.f, "printPopup", this._printModel);

		//Register all the listeners to the channelMatchingPopup
		this._registerListener("print", this._showPrintPreview, this, this._printPopup);

		this._registerListener("printType", this._printTypeChanged, this, this._printPopup);
		this._registerListener("printArea", this._printAreaChanged, this, this._printPopup);
		this._registerListener("printChannels", this._printChannelsChanged, this, this._printPopup);
		this._registerListener("resolutionChanged", this._printResolutionChanged, this, this._printPopup);
		this._registerListener("printAnnotationsSource", this._printAnnotationsSourceChanged, this, this._printPopup);
		this._registerListener("resolutionUnitChanged", this._printResolutionUnitChanged, this, this._printPopup);

		this.f.getCurrentUser().on(
			"options.changed.PREFERENCES_UNIT_RESOLUTION",
			_.bind(function(oldValue, newValue) {
				this._printResolutionUnitChanged(newValue);
			}, this)
		);
	}
};

/**
 * Action for when the print type changes, sets the correct value in the print model
 * @param {PrintConstants.PrintType} printType
 */
PrintModule.prototype._printTypeChanged = function(printType) {
	this._printModel.setPrintType(printType);
	this._printPopup.render();
};

/**
 * Action for when the print area option changes, sets the correct value in the print model
 * @param {Boolean} printCurrentView - print only the current view
 */
PrintModule.prototype._printAreaChanged = function(printCurrentView) {
	if (printCurrentView) {
		this._printModel.setPrintArea(PrintConstants.PrintArea.PRINTCURRENTVIEWAREA);
	} else {
		this._printModel.setPrintArea(PrintConstants.PrintArea.PRINTENTIREIMAGE);
	}
};

/**
 * Action for when the print channel option changes, sets the correct value in the print model
 * @param {Boolean} printOnlySelectedChannels - print only the selected channels
 */
PrintModule.prototype._printChannelsChanged = function(printOnlySelectedChannels) {
	if (printOnlySelectedChannels) {
		this._printModel.setPrintChannels(PrintConstants.PrintChannels.PRINTSELECTEDCHANNELS);
	} else {
		this._printModel.setPrintChannels(PrintConstants.PrintChannels.PRINTALLCHANNELS);
	}
};

/**
 * Action for when the print annotations source option changes, sets the correct value in the print model
 * @param {PrintConstants.PrintAnnotationsSource} printAnnotations
 */
PrintModule.prototype._printAnnotationsSourceChanged = function(printAnnotationsSource) {
	this._printModel.setPrintAnnotationsSource(printAnnotationsSource);
};

/**
 * Action for when the print resolution changes, sets the correct value in the print model
 * @param {Number} resolution - in either ppi, ppcm or ppmm
 */
PrintModule.prototype._printResolutionChanged = function(resolution) {
	var resInPpmm = ViewerUtils.convertResolutionToPpmmValue(resolution, this._printModel.getResolutionUnit());
	this._printModel.setPrintResolution(ViewerUtils.parseFloat(resInPpmm));
};

/**
 * Action for when the print resolution unit changed
 * @param {CoreConstants.ResolutionUnits} resolutionUnit
 */
PrintModule.prototype._printResolutionUnitChanged = function(resolutionUnit) {
	this._printModel.setPrintResolutionUnit(resolutionUnit);
	if (this._printPopup && this._printPopup.isVisible()) {
		this._printPopup.render();
	}
};

/**
 * Check for unsaved annotations and asks the user to save them if needed.
 */
PrintModule.prototype._checkForUnsavedAnnotations = function() {
	// check whether there are unsaved annotations
	var annotationModule = this.f.getModule("annotation");
	if (annotationModule !== undefined) {
		var hasUnsavedItems = annotationModule.promptToSaveUnsavedAnnotations(
			_.bind(function() {
				this._printPopup.show(); //show the printpopup, the annotations are saved
			}, this),
			function() {
				//don't do anything, cancel will not show the print popup and won't save the annotations
			},
			_.bind(function() {
				annotationModule.reload(); //Don't save the annotations, reload them
				this._printPopup.show(); //now show the print popup
			}, this)
		);

		if (hasUnsavedItems) {
			// hide the printpopup while we are prompting for saving unsaved annotations
			this._printPopup.hide();
		}
	}
};

/**
 * Show the print preview, it will check which print type is selected to open the correct print preview
 */
PrintModule.prototype._showPrintPreview = function() {
	//get the document information from the documentcontroller
	var activePage = this.f.getDocumentController().getActivePage();
	if (this._printModel.getPrintType() === PrintConstants.PrintType.PRINTANNOTATIONOVERVIEW) {
		this._printAnnotations();
	} else if (this._printModel.getPrintType() === PrintConstants.PrintType.PRINTIMAGEAREA && !activePage.hasCad()) {
		this._printImageArea();
	} else {
		//print what is currently visible, and image area for CAD
		this._printCurrentView();
	}
};

/**
 * Print the annotations overview
 * @private
 */
PrintModule.prototype._printAnnotations = function() {
	var dataAdapter = this.f.getDataAdapterForModule("print");
	dataAdapter.printAnnotations(this._printModel);
};

/**
 * Print the image area
 * @private
 */
PrintModule.prototype._printImageArea = function() {
	var dataAdapter = this.f.getDataAdapterForModule("print");
	dataAdapter.printImageArea(this._printModel);
};

/**
 * Print the image area using the obtained URL
 * @private
 */
PrintModule.prototype._handlePrintUrl = function(url) {
	//get the document information from the documentcontroller
	var activePage = this.f.getDocumentController().getActivePage();
	//open a new window (or tab, browsers choice)
	var printPreview = window.open("", "_blank");

	var viewerState = this.f.getViewerState(); //get the state of the viewer
	var worldState = viewerState.viewport.world;

	var t = new Template(this.f.i18n, PrintImageAreaTemplate);
	//create the img tag to display the requested image, it will resize to the worldSize of the document (or only the visible part)
	// but it still prints in the requested resolution! The browser decides how to translate mm to pixels
	if (this._printModel.showPrintArea() && this._printModel.getPrintArea() === PrintConstants.PrintArea.PRINTCURRENTVIEWAREA) {
		//if its a viewarea, only the size of the viewarea is needed (in world coordinates)
		t.d.imgHeight = worldState.visibleRect.size.height;
		t.d.imgWidth = worldState.visibleRect.size.width;
	} else {
		t.d.imgHeight = worldState.worldBounds.size.height;
		t.d.imgWidth = worldState.worldBounds.size.width;
	}
	t.d.imgUrl = url;
	t.d.docName = activePage.getDocument().get("docname");

	this.f.i18n.loadBundle(
		"viewer",
		_.bind(function() {
			printPreview.document.open();
			var html = t.getRenderedHtml();
			printPreview.document.write(html);
			printPreview.document.close();
		}, this)
	);
};

/**
 * Print the current view
 * @private
 */
PrintModule.prototype._printCurrentView = function() {
	//first check if not all channels are selected
	if (this._printModel.showPrintChannels()) {
		var printAllChannels = this._printModel.getPrintChannels(); //get the setting for printing all or only selected channels
		if (printAllChannels === PrintConstants.PrintChannels.PRINTALLCHANNELS && this.f.getModule("graphics")) {
			//print all channels
			// enable all channels before showing the print popup
			// set channels
			var graphicsModule = this.f.getModule("graphics");

			var selectionBefore = graphicsModule.getChannelSelection();
			var allSelected = {};
			_.each(graphicsModule.getChannelSelection(), function(selected, ID) {
				allSelected[ID] = true;
			});

			graphicsModule.setChannelSelection(allSelected);

			this._openWindowForPrintCurrentView(function() {
				graphicsModule.setChannelSelection(selectionBefore);
			});

			return;
		}
	}

	// no need to wait for stuff, just open the print popup
	this._openWindowForPrintCurrentView();
};

/**
 * This opens the popup for printing the current view
 * Once all image data has been loaded, the canvas data is copied to the print popup canvases
 * @param {Function} afterInitialization - callback called when the printpopup is correctly initialized and the canvas data is transferred
 */
PrintModule.prototype._openWindowForPrintCurrentView = function(afterInitialization) {
	var activePage = this.f.getDocumentController().getActivePage();
	var referencePage = this.f.getDocumentController().getDifferencePage();
	var docNames = activePage.getDocument().get("docname");
	if (referencePage !== undefined) {
		docNames += "-" + referencePage.getDocument().get("docname");
	}

	var printPreview = window.open("", "_blank");

	// it is safe to get the viewport like this, otherwise we wouldn't be printing :)
	var viewportModule = this.f.getModule("viewport");
	var viewportElement = viewportModule.widget.getElement();

	// generate the HTML for the print popup
	var t = new Template(this.f.i18n, PrintCurrentViewTemplate);
	// Dirty hack to include css as a string, ideally we don't want to include webviewer css
	t.d.cssUrl = require("!to-string-loader!css-loader!less-loader!../layout/print/less/print.less");
	t.d.docNames = docNames;
	t.d.loaderImgUrl = require("../layout/img/loaderTransparent.gif");
	t.d.sceneHeight = viewportElement.height() || null;
	t.d.sceneWidth = viewportElement.width() || null;
	var html = t.getRenderedHtml();

	printPreview.document.open();
	printPreview.document.write(html);
	printPreview.document.close();

	setTimeout(
		_.bind(function() {
			// wait for update of channel selection (this is debounced at 40ms)
			// wait until all fragments have finished loading
			viewportModule.whenFinishedLoadingImageData(function() {
				setTimeout(
					_.bind(function() {
						// wait for the rendering of fragments (rendering is throttled at 40ms)
						// copy over the HTML of the viewport element
						printPreview.document.getElementById("viewport").innerHTML = viewportElement.html();

						// if there are multiple views shown, we want to viewerapp to have border
						if (printPreview.document.getElementsByClassName("view").length > 1) {
							printPreview.document.getElementById("viewerApp").className = "viewerApp withBorder";
						}

						// now we want to transfer all canvas data to the popup
						try {
							// copy the data for every canvas
							viewportElement.find("canvas").each(function(number, canvas) {
								var destinationCanvas = printPreview.document.getElementById(canvas.id);
								if (destinationCanvas) {
									var ctx = destinationCanvas.getContext("2d");
									ctx.drawImage(canvas, 0, 0);
								}
							});
						} catch (err) {
							// IE doesn't support drawing canvasdata from another window... but we do have a fix :)
							// replace every canvas with an image with the imagedata from the original canvas
							viewportElement.find("canvas").each(function(number, canvas) {
								var image = printPreview.document.createElement("img");

								var oldCanvas = printPreview.document.getElementById(canvas.id);
								if (oldCanvas) {
									// initialize the image with correct settings (and copy styling)
									image.alt = "";
									image.id = oldCanvas.id;
									oldCanvas.id = "";
									image.style.cssText = oldCanvas.style.cssText;
									image.width = oldCanvas.width;
									image.height = oldCanvas.height;

									// replace the canvas with the image
									oldCanvas.parentNode.replaceChild(image, oldCanvas);

									// copy over the imageData from the original canvas
									image.src = canvas.toDataURL("image/png");
								}
							});
						} finally {
							// we finished copying the image data

							// hide loading indicator
							var loadingIndicator = printPreview.document.getElementById("printLoadingIndicator");
							loadingIndicator.parentNode.removeChild(loadingIndicator);

							// call callbacks if needed
							if (afterInitialization) {
								afterInitialization();
							}
						}
					}, this),
					100
				);
			}, this);
		}, this),
		100
	);
};

export default PrintModule;
