import PrivateChannel from "@esko/web-viewer/core/PrivateChannel";
import ModalPopup from "@esko/web-viewer/layout/component/ModalPopup";
import PrintConstants from "@esko/web-viewer/print/PrintConstants";
import DataConstants from "@esko/web-viewer/data/DataConstants";
import Template from "@esko/template";
import tpl from "@esko/web-viewer/print/PrintPopup.tpl.html";
import ViewerUtils from "@esko/web-viewer/utils/ViewerUtils";
import $ from "jquery";
import _ from "underscore";
/**
 * @exports PrintPopup
 * @memberof PrintModule
 * @augments layout.component.ModelPopup
 * @class Popup for printing, shows all the options for printing and
 * @constructor
 *
 * @param {Core.ViewerFacade} f - Viewer facade
 * @param {String} id - Id of the popup. It is used in its private channel.
 * @param {PrintModel} printModel
 */
var PrintPopup = function(f, id, printModel) {
	ModalPopup.call(this, f, id, f.getContext());
	PrivateChannel.installTo(this, f, ["printPopup"]);

	f.i18n.loadBundle(
		"viewer",
		_.bind(function() {
			this.setHeader(
				$("<span/>")
					.addClass("popupTitle")
					.html(f.i18n.t("PrintSetupTitle"))
			);

			this.setOkBtnText(f.i18n.t("PrintPrintButton"));
			this.setOkAction(
				_.bind(function() {
					this._emit("print");
				}, this)
			);
		}, this)
	);

	this.element.addClass("printPopup");

	/**
	 * @private
	 * @member {PrintModel} - model used to construct the popup
	 */
	this._printModel = printModel;

	/**
	 * @member {Template} template for the popup
	 */
	this.t = new Template(this.f.i18n, tpl);
	this.t.setContext(this._popupContent);

	//action for the radio buttons to change the print type, emits printPopup.printType
	this.t.bind(
		"change",
		"[name=printType]",
		function(source, event) {
			this._emit("printType", parseInt(source.value, 10));
		},
		this
	);

	//action for the checkbox to change the print area, emits printPopup.printArea
	this.t.bind(
		"change",
		"[name=printArea]",
		function(source, event) {
			this._emit("printArea", source.checked);
		},
		this
	);

	//action for the checkbox to change the channels to be printed, emits printPopup.printChannels
	this.t.bind(
		"change",
		"[name=printChannels]",
		function(source, event) {
			this._emit("printChannels", source.checked);
		},
		this
	);

	//action for the input field for the resolution, emits printPopup.resolutionChanged
	this.t.bind(
		"input",
		"[name=printResolution]",
		_.debounce(
			_.partial(function(that, source, event) {
				var value = ViewerUtils.parseFloat(source.value);
				if (isNaN(value)) {
					// if nothing is filled in or only letter(s), behave like entered the minimum
					value = 0;
				}
				source.value = value; // by doing this letters added to the end will be filtered out
				that._emit("resolutionChanged", value);
			}, this),
			400
		)
	);

	this.t.bind(
		"change",
		"[name=printResolutionUnit]",
		function(source, event) {
			this._emit("resolutionUnitChanged", parseInt(source.value, 10));
		},
		this
	);

	//action for the radio buttons to select the annotations of the current/reference
	this.t.bind(
		"change",
		"[name=printAnnotationsSource]",
		function(source, event) {
			this._emit("printAnnotationsSource", parseInt(source.value, 10));
		},
		this
	);
};

PrintPopup.prototype = Object.create(ModalPopup.prototype);
PrintPopup.prototype.constructor = PrintPopup;

/**
 * Renders the popup for channel matching, it's a throttled function
 */
PrintPopup.prototype.render = _.throttle(function() {
	this.t.d.printType = this._printModel.getPrintType();
	this.t.d.showResolution = this._printModel.showResolution();
	this.t.d.printResolutionUnit = this._printModel.getResolutionUnit();
	this.t.d.showPrintImageArea = this._printModel.getPageType() === DataConstants.documentPageType.GRAPHICS;
	this.t.d.printResolution = ViewerUtils.convertResolutionFromPpmmValue(
		this._printModel.getResolution(),
		this._printModel.getResolutionUnit(),
		this.f.getResolutionUnitsFix()
	);
	this.t.d.showPrintArea = this._printModel.showPrintArea();
	this.t.d.printArea = this._printModel.getPrintArea();
	this.t.d.showPrintChannels = this._printModel.showPrintChannels();
	this.t.d.printChannels = this._printModel.getPrintChannels();
	this.t.d.showPrintAnnotationsSource = this._printModel.showPrintAnnotationsSource();
	this.t.d.printAnnotationsSource = this._printModel.getPrintAnnotationsSource();
	this.t.d.showPrintAnnotationOverview = this.f.getDataAdapterForModule("print").showPrintAnnotationOverview();
	this.t.d.showPrintCurrentView = this._printModel.getPageType() !== DataConstants.documentPageType.VIDEO;

	this.t.render();
	var shouldSelectInput = this.t.getContext().find("[name=printType][value=" + this._printModel.getPrintType() + "]");
	if (shouldSelectInput.length === 0) {
		shouldSelectInput = this.t
			.getContext()
			.find("[name=printType]")
			.first();
	}
	if (!shouldSelectInput.prop("checked")) {
		shouldSelectInput.prop("checked", true);
		shouldSelectInput.trigger("change");
	}
	if (this._isInitialized) {
		this.recalculateSize(300);
	}
}, 40);

/**
 * Show the modal popup
 */
PrintPopup.prototype.show = function() {
	this._printModel.setPrintResolutionUnit(this.f.getOption("PREFERENCES_UNIT_RESOLUTION"));
	this.render();
	ModalPopup.prototype.show.call(this);
	this.recalculateSize(300);
};

export default PrintPopup;
