import ViewportLayerModel from "@esko/web-viewer/viewport/model/ViewportLayerModel";
import ShapeUtils from "@esko/shapeutils";
import ListenerUtil from "@esko/web-viewer/utils/ListenerUtil";
/**
 * @exports SvgLayerModel
 * @augments ViewportLayerModel
 * @class This layer model is responsible to draw svgs
 * @constructor
 * @mixes Utils.ListenerUtilMixin
 *
 * @param {Core.ViewerFacade} f - Viewer facade
 * @param {String} ID - the ID of the layer model
 * @param {AbstractDocumentPage} page
 */

var SvgLayerModel = function(f, ID, page) {
	ViewportLayerModel.call(this, f, ID);
	ListenerUtil.mixin(this, f);

	this.f.whenModuleAvailable(
		"viewport",
		function(viewportModule) {
			this._viewport = viewportModule;
		},
		this
	);

	this._page = page;

	this._registerListener(
		"doc.parameters.changed",
		function() {
			this._emit("repaint");
		},
		this
	);
};

// inherit
SvgLayerModel.prototype = Object.create(ViewportLayerModel.prototype);
// correct the constructor pointer
SvgLayerModel.prototype.constructor = SvgLayerModel;

/**
 * Gets the SVG element
 */
SvgLayerModel.prototype.getSVGElement = function() {
	if (this._page && this._page.hasSvg() && this._page.isSVGLoaded()) {
		return this._page.getSVGElement();
	} else {
		return null;
	}
};

/**
 * Remove all listeners
 */
SvgLayerModel.prototype.destroy = function() {
	ViewportLayerModel.prototype.destroy.call(this);

	this._unregisterAllListeners();
};

/**
 * returns the size of the svg
 * @return ShapeUtils.Size
 */
SvgLayerModel.prototype.getSVGSize = function() {
	var activePage = this.f.getActivePage();
	if (this._page && this._page.isSVGLoaded()) {
		return this._page.getSVGSize();
	} else if (!this._page && activePage && activePage.hasSvg() && activePage.isSVGLoaded() && activePage.isReaderSpread()) {
		return activePage.getSVGSize();
	}
	return null;
};

SvgLayerModel.prototype.setPage = function(page) {
	this._page = page;
	this._emit("pageChanged");
};

SvgLayerModel.prototype.getPage = function() {
	return this._page;
};

/**
 * @return {String}
 */
SvgLayerModel.prototype.getType = function() {
	return "svg";
};

SvgLayerModel.prototype.getReaderSpreadOffset = function() {
	var activePage = this.f.getActivePage();
	var diffPage = this.f.getDocumentController().getDifferencePage();

	if (activePage === this._page && this._page.isReaderSpread() && this._page.get("isRightPage")) {
		// Active page is right page
		return new ShapeUtils.Point(diffPage ? diffPage.getSVGSize().width : activePage.getSVGSize().width, 0);
	}
	if (activePage !== this._page && activePage && activePage.isReaderSpread() && !activePage.get("isRightPage")) {
		return new ShapeUtils.Point(activePage.getSVGSize().width, 0);
	}
	return new ShapeUtils.Point();
};

export default SvgLayerModel;
