/**
 * @namespace SvgModule
 */
/**
 * @namespace SvgModule
 */
import AbstractModule from "@esko/web-viewer/core/AbstractModule";

import CompareConstants from "@esko/web-viewer/compare/CompareConstants";
import SvgLayerModel from "./SvgLayerModel";
import SvgDiffLayerModel from "./SvgDiffLayerModel";
import TextSelectionTool from "./TextSelectionTool";
import TextSelectionModel from "./TextSelectionModel";
import LayoutConstants from "@esko/web-viewer/layout/LayoutConstants";
import ToolbarComponent from "@esko/toolbar/ToolbarComponent";
import ButtonGroup from "@esko/toolbar/ButtonGroup";
import ViewerToolbarButton from "@esko/web-viewer/layout/component/ViewerToolbarButton";
import _ from "underscore";
/**
 * Module for enable viewing of svg files
 * @param {ViewerFacade} f
 * @param {Boolean} silenced
 */
var SvgModule = function(f, silenced) {
	AbstractModule.call(this, f, silenced, "svg");

	/**
	 * @member {SvgLayerModel} SvgLayerModel
	 * @private
	 */
	this._layerModel = null;

	/**
	 * @member {SvgDiffLayerModel} SvgDiffLayerModel
	 * @private
	 */
	this._diffLayerModel = null;

	f.getDocumentController().whenActiveOrDifferencePageChanged(_.bind(this._setSVGLayerModels, this));

	this._textSelectionModel = new TextSelectionModel(f, [this.getPrivateChannelName(), "textSelection", "model"]);

	// Add text selection tool to the toolbar
	this._textSelectionTool = new TextSelectionTool(f, this._textSelectionModel);
	this._publishTool(this._textSelectionTool);

	this._textSelectionButtonGroup = null;

	this._publishToolbarComponent(
		LayoutConstants.ToolbarComponent.TEXTSELECTION,
		_.bind(function(layoutType) {
			var textSelectionButton = new ViewerToolbarButton(f, "TextSelectionButton", "icon icon-WCR_Viewer-TextSelection", false);
			textSelectionButton.setTooltip("viewer", "TextSelectionTT");
			textSelectionButton.listenToTool(this._textSelectionTool, true);
			/**
			 * Component: Textselection
			 */
			var textSelectionToolbar = new ToolbarComponent(LayoutConstants.ToolbarComponent.TEXTSELECTION, f);

			/**
			 * Group
			 */
			var selectionGroup = new ButtonGroup("selectionGroup");
			selectionGroup.setTitle(f.i18n, "viewer", "TextBtnGroupTitle");
			textSelectionToolbar.addButtonGroup(selectionGroup);

			this._textSelectionButtonGroup = selectionGroup;

			// select button
			selectionGroup.addButton(textSelectionButton);

			this.f.on(
				"module.compare.changed.mode",
				function() {
					if (this._textSelectionTool.canBeActivated()) {
						textSelectionButton.setEnabled(true);
					} else {
						if (this._textSelectionTool.isActive()) {
							this.f.emit("toolManager.deactivateTool", this._textSelectionTool);
						}
						textSelectionButton.setEnabled(false);
					}
				}.bind(this)
			);
			return textSelectionToolbar;
		}, this)
	);

	// Toggle the diff layer
	this.f.on(
		"module.compare.changed.diffMode",
		_.bind(function(oldMode, newMode) {
			var diffPage = this.f.getDocumentController().getDifferencePage();
			var activePage = this.f.getActivePage();

			// Do not display difference layer for 'Current Version' and 'Undefined'
			this.f
				.getModule("viewport")
				.toggleLayer(
					"SVGDifferenceLayer",
					diffPage &&
						diffPage.hasSvg() &&
						newMode !== CompareConstants.diffMode.UNDEFINED &&
						newMode !== CompareConstants.diffMode.CURRENTVERSION
				);
			// Do not display current layer for 'Difference Version'
			this.f
				.getModule("viewport")
				.toggleLayer("SVGLayer", activePage && activePage.hasSvg() && newMode !== CompareConstants.diffMode.DIFFVERSION);
		}, this)
	);
};

// inherit
SvgModule.prototype = Object.create(AbstractModule.prototype);
// correct the constructor pointer
SvgModule.prototype.constructor = SvgModule;

/**
 * set the correct pages for the svg layer models
 * @param {AbstractDocumentPage} page
 */
SvgModule.prototype._setSVGLayerModels = function() {
	var page = this.f.getActivePage();
	var hasSVG = page.hasSvg();
	var diffPage = this.f.getDocumentController().getDifferencePage();
	var hasDiffSVG = diffPage && diffPage.hasSvg();
	if (hasSVG) {
		if (!this._layerModel) {
			this._layerModel = new SvgLayerModel(this.f, "SVGLayer", page);
		} else {
			this._layerModel.setPage(page);
		}

		if (this._textSelectionButtonGroup) {
			this._textSelectionButtonGroup.show();
		}
	} else {
		if (this._textSelectionButtonGroup) {
			this._textSelectionButtonGroup.hide();
		}
		if (this._textSelectionTool) {
			this._textSelectionTool.deactivate();
		}
	}

	if (hasDiffSVG) {
		// could be that diffpage is null -> no problem
		if (!this._diffLayerModel) {
			this._diffLayerModel = new SvgDiffLayerModel(this.f, "SVGDifferenceLayer", diffPage);
		}
	}

	// toggle svg layers
	this.f.whenModuleAvailable(
		"viewport",
		function(viewportModule) {
			viewportModule.whenWidgetModelAvailable(function() {
				if (hasSVG) {
					viewportModule.addLayer("SVGLayer", this._layerModel);
					viewportModule.toggleLayer("SVGLayer", true);
				} else {
					viewportModule.toggleLayer("SVGLayer", false);
				}

				if (hasDiffSVG) {
					viewportModule.addLayer("SVGDifferenceLayer", this._diffLayerModel);
					viewportModule.toggleLayer("SVGDifferenceLayer", true);
				} else {
					viewportModule.toggleLayer("SVGDifferenceLayer", false);
				}
			}, this);
		},
		this
	);
};

export default SvgModule;
