import ViewerUtils from "@esko/web-viewer/utils/ViewerUtils";
import ViewportPopoverModel from "@esko/web-viewer/viewport/model/ViewportPopoverModel";
import tpl from "@esko/web-viewer/svg/TextSelectionPopoverTemplate.tpl.html";
import Template from "@esko/template";
import ShapeUtils from "@esko/shapeutils";
import _ from "underscore";
import $ from "jquery";
import AnnotationConstants from "@esko/web-viewer/annotation/AnnotationConstants";
/**
 * PoveroverModel for text selection tool
 *
 * @exports TextSelectionPopoverModel
 * @memberof SvgModule
 * @class TextSelectionPopoverModel
 * @augments Viewport.Model.ViewportPopoverModel
 * @constructor
 *
 * @param {Core.ViewerFacade} f
 * @param {String} name
 */
var TextSelectionPopoverModel = function(f, name) {
	ViewportPopoverModel.call(this, f, name);

	this._content = $("<div/>").addClass("TextSelectionPopover");
	this.textContent = "";
	this._popoverLocation = null;
	this._worldTextFrames = null;
	this._boundingWorldFrame = null;

	// initialize template
	this.t = new Template(f.i18n, tpl);
	this.t.setContext(this._content);

	// Add a regular annotation
	this.t.bind(
		"mousedown",
		".addTextAnnotation",
		_.bind(function(source, event) {
			this.f.requestModule(
				"annotation",
				_.bind(function(annotationModule) {
					annotationModule.promptIfAnnotationOutsidePage(
						this._boundingWorldFrame,
						_.bind(function() {
							// add annotation
							this.f.emit("module.annotation.createAnnotation", this._boundingWorldFrame, this._textContent);
							this._emit("addAnnotation");
						}, this),
						function() {
							// do nothing when 'No' is pressed
						},
						false,
						true
					);
				}, this)
			);
			this._emit("popupClosed");
		}, this)
	);

	// Copy selection
	this.t.bind(
		"mousedown",
		".copyText",
		_.bind(function(source, event) {
			document.execCommand("copy");
			this._emit("popupClosed");
		}, this)
	);

	// Add multiple linked highlight annotations
	this.t.bind(
		"mousedown",
		".addTextHighlightAnnotation",
		_.bind(function(source, event) {
			this.f.requestModule(
				"annotation",
				_.bind(function(annotationModule) {
					annotationModule.promptIfAnnotationOutsidePage(
						this._boundingWorldFrame,
						_.bind(function() {
							annotationModule.addLinkedAnnotations(
								this._worldTextFrames,
								this._textContent,
								AnnotationConstants.AnnotationType.ANNOT_HIGHLIGHT
							);

							this._emit("addAnnotation");
						}, this),
						function() {
							// do nothing when 'No' is pressed
						},
						false,
						true
					);
				}, this)
			);
			this._emit("popupClosed");
		}, this)
	);

	// load language if needed
	this.f.i18n.loadBundle(
		"viewer",
		_.bind(function() {
			this.t.render();
		}, this)
	);

	this.getFrame = function() {
		if (this._popoverLocation) {
			return this._popoverLocation;
		} else {
			return new ShapeUtils.Rect(new ShapeUtils.Point(0, 0), new ShapeUtils.Size(1, 1));
		}
	};

	this.getContent = function() {
		return this._content;
	};

	this.getPreferredPosition = function() {
		// override
		return "right";
	};

	this.getRestrictBorder = function() {
		return 0;
	};

	this.update();
};

// inherit
TextSelectionPopoverModel.prototype = Object.create(ViewportPopoverModel.prototype);
TextSelectionPopoverModel.prototype.constructor = TextSelectionPopoverModel;

/**
 * Update the popover
 */
TextSelectionPopoverModel.prototype.update = function() {
	this.t.render();
};

/**
 * Set the list of areas of the highlight popover
 * @param {Array.<ShapeUtils.Rect>} The array of areas that overlap the individual text tags
 */
TextSelectionPopoverModel.prototype.setWorldTextFrames = function(worldTextFrames) {
	this._worldTextFrames = worldTextFrames;
	// Calculate the bounding rect which contains all these text frames
	this._boundingWorldFrame = this._calculateBoundingWorldFrame(worldTextFrames);
};

/**
 * Set the location of the popover
 * @param {ShapeUtils.Point} point for which the popover should be shown
 */
TextSelectionPopoverModel.prototype.setPopoverLocation = function(point) {
	this._popoverLocation = new ShapeUtils.Rect(point, new ShapeUtils.Size(0.001, 0.001));
};

/**
 * Set the area of the popover
 * @param {ShapeUtils.Rect} area for which the popover should be shown
 */
TextSelectionPopoverModel.prototype.setTextContent = function(content) {
	this._textContent = content;
};

/**
 * Get the bounding world frame, based on given the array of world text rects
 * @private
 * @param   {Array.<ShapeUtils.Rect>} worldTextFrames - the array of individual rects
 * @returns {ShapeUtils.Rect} the bounding world rect
 */
TextSelectionPopoverModel.prototype._calculateBoundingWorldFrame = function(worldTextFrames) {
	var boundingWorldFrame = null;
	if (worldTextFrames) {
		_.each(
			worldTextFrames,
			_.bind(function(textFrame) {
				if (!boundingWorldFrame) {
					boundingWorldFrame = textFrame;
				} else {
					boundingWorldFrame = boundingWorldFrame.calculateArea(textFrame);
				}
			}, this)
		);
	}
	return boundingWorldFrame;
};

export default TextSelectionPopoverModel;
