import _ from "underscore";
import $ from "jquery";
import AbstractTool from "@esko/web-viewer/viewport/tools/AbstractTool";
import CompareConstants from "@esko/web-viewer/compare/CompareConstants";
import CoreConstants from "@esko/web-viewer/core/CoreConstants";
import ConversionConstants from "@esko/web-viewer/utils/ConversionConstants";
import TextSelectionLayerModel from "./TextSelectionLayerModel";
import rbush from "rbush";
import ShapeUtils from "@esko/shapeutils";
import ListenerUtil from "@esko/web-viewer/utils/ListenerUtil";
import rbushknn from "rbush-knn";
var DEFAULTSTATE = "default";
var DRAGGINGSTATE = "dragging";
var SELECTEDSTATE = "selected";

/**
 * @exports TextSelectionTool
 * @memberof SvgModule
 * @class The TextSelectionTool allows to select text in SVG files
 * @constructor
 * @augments {Core.AbstractTool}
 *
 * @param {Core.ViewerFacade} f - Viewer facade
 * @param {SvgModule.TextSelectionModel}
 */
var TextSelectionTool = function(f, textSelectionModel) {
	AbstractTool.call(this, f, CoreConstants.TOOLS.TEXTSELECTION);
	ListenerUtil.mixin(this, f);

	// listen tot popover interaction to change state
	this.currentPopoverListener = null;

	/**
	 * @member {TextSelectionModel}
	 * @private
	 */
	this._textSelectionModel = textSelectionModel;

	/**
	 * @member {TextSelectionLayerModel}
	 * @private
	 */
	this._textSelectionLayerModel = new TextSelectionLayerModel(f, this.getIID(), this._textSelectionModel);

	this.f.whenModuleAvailable(
		"viewport",
		function(viewportModule) {
			this._viewport = viewportModule;
			if (this.isActive()) {
				this._viewport
					.getFirstViewOccurrence("Main")
					.getViewContext()
					.addClass("enableTextSelection");
			}
		},
		this
	);

	/**
	 * @member {Rbush}
	 * @private
	 */
	this._rbush = null;

	f.getDocumentController().whenActiveOrDifferencePageChangedAndParametersAvailableOrChange(_.bind(this._onPageChange, this));

	var startElement = null;
	var startCharIndex = null;
	var startElementIndex = null;

	/**
	 * DEFAULT STATE -> nothing set
	 */
	this.registerState(DEFAULTSTATE, {
		enterState: _.bind(function() {
			this._textSelectionModel.setContent("");
			this._textSelectionModel.setStartPoint(null);
			this._textSelectionModel.setEndPoint(null);
		}, this),

		leaveState: _.bind(function() {}, this),

		mousedown: _.bind(function(e) {
			document.activeElement.blur();
			if (!this.isActive()) {
				// This is only triggered in the SELECTED state (destroys popup and text selection)
				this.changeState(DEFAULTSTATE);
			} else {
				var curLoc = this.keepPointWithinWorldBounds(this.getWorldPoint(e), false);
				this._textSelectionModel.setStartPoint(curLoc);

				var result = rbushknn(this._rbush, curLoc.x, curLoc.y, 1);

				if (result.length > 0) {
					startElement = result[0].element;
					startCharIndex = result[0].charIndex;
					startElementIndex = result[0].elementIndex;
					this.changeState(DRAGGINGSTATE);
				}
			}
		}, this),

		mouseup: function() {},

		mousemove: function() {},

		mouseout: function() {},

		mouseleave: function() {},

		/**
		 * no context menu!
		 */
		contextmenu: _.bind(function(e) {
			e.preventDefault();
			e.stopPropagation();
		}, this),

		canBeInterrupted: true
	});

	/**
	 * Dragging STATE -> selecting text
	 */
	this.registerState(
		DRAGGINGSTATE,
		_.defaults(
			{
				mousedown: function() {},

				mouseup: _.bind(function(e) {
					var curLoc = this.keepPointWithinWorldBounds(this.getWorldPoint(e), false);
					this._textSelectionModel.setEndPoint(curLoc);

					if (curLoc.equals(this._textSelectionModel.getStartPoint())) {
						this.changeState(DEFAULTSTATE);
					} else {
						this.changeState(SELECTEDSTATE);
					}
				}, this),

				mousemove: _.bind(function(e) {
					var curLoc = this.keepPointWithinWorldBounds(this.getWorldPoint(e), false);
					// update selected text with current location

					if (!curLoc.equals(this._textSelectionModel.getStartPoint())) {
						var result = rbushknn(this._rbush, curLoc.x, curLoc.y, 1);

						if (result.length > 0) {
							var endElement = result[0].element;
							var endCharIndex = result[0].charIndex;
							var endElementIndex = result[0].elementIndex;

							var selection = window.getSelection();
							selection.removeAllRanges();
							var range = document.createRange();
							if (startElementIndex < endElementIndex) {
								range.setStart(startElement.firstChild, startCharIndex);
								range.setEnd(endElement.firstChild, endCharIndex + 1);
							} else if (startElementIndex > endElementIndex) {
								range.setStart(endElement.firstChild, endCharIndex);
								range.setEnd(startElement.firstChild, startCharIndex + 1);
							} else {
								if (endCharIndex < startCharIndex) {
									range.setStart(startElement.firstChild, endCharIndex);
									range.setEnd(endElement.firstChild, startCharIndex);
								} else {
									range.setStart(startElement.firstChild, startCharIndex);
									range.setEnd(endElement.firstChild, endCharIndex + 1);
								}
							}
							selection.addRange(range);
						}
					}
				}, this)
			},
			this.getState(DEFAULTSTATE)
		)
	);

	/**
	 * selected STATE -> text is selected and annotation balloon is shown
	 */
	this.registerState(
		SELECTEDSTATE,
		_.defaults(
			{
				enterState: _.bind(function() {
					var selection = window.getSelection();
					if (selection.rangeCount > 0) {
						// Calculate the text world rects of the selected text
						var range = selection.getRangeAt(0);
						var startTextTag = range.startContainer.parentNode;
						var endTextTag = range.endContainer.parentNode;
						var textWorldFrames = [];

						if (range.startContainer === range.endContainer) {
							// Calculate bounding box of the single text tag
							textWorldFrames.push(this._calculateBoundingBoxOfText(startTextTag, range.startOffset, range.endOffset));
						} else {
							// Calculate the bounding box of the first text tag between position range.startOffset <-> startTextTag.getNumberOfChars()
							if (range.startOffset < startTextTag.getNumberOfChars()) {
								textWorldFrames.push(this._calculateBoundingBoxOfText(startTextTag, range.startOffset, startTextTag.getNumberOfChars()));
							}

							// Calculate the bounding boxes of the text tags between position 0 <-> textElement.getNumberOfChars()
							_.each(
								this._getElementsInsideRange(range),
								_.bind(function(element) {
									$(element)
										.find("text")
										.each(
											_.bind(function(index, textElement) {
												textWorldFrames.push(this._calculateBoundingBoxOfText(textElement, 0, textElement.getNumberOfChars()));
											}, this)
										);
								}, this)
							);

							// Calculate the bounding box of the last text tag between position 0 <-> range.endOffset
							if (range.endOffset > 0) {
								textWorldFrames.push(this._calculateBoundingBoxOfText(endTextTag, 0, range.endOffset));
							}
						}

						// Populate models & show popover
						this._textSelectionLayerModel.setWorldTextFrames(textWorldFrames);
						// encode the string and set it as contents
						this._textSelectionModel.setContent(
							$("<div/>")
								.text(selection.toString())
								.html()
						);
						this._textSelectionLayerModel.showPopover();

						//register the listener to the model to create global annotations
						this.currentPopoverListener = this._registerListener(
							"addAnnotation",
							function() {
								this.changeState(DEFAULTSTATE);
							},
							this,
							this._textSelectionLayerModel.getPopoverModel()
						);
					}
				}, this),

				leaveState: _.bind(function() {
					this._textSelectionLayerModel.hidePopover();
					window.getSelection().removeAllRanges();
					if (this.currentPopoverListener) {
						this._unregisterListener(this.currentPopoverListener);
					}
				}, this)
			},
			this.getState(DEFAULTSTATE)
		)
	);

	this.registerActiveEvents(["mousedown", "mousemove", "mouseup", "mouseleave", "contextmenu"]);
	this.registerPassiveEvents(["mousedown"]);

	// Esc allows to reset selection
	this.activeKeyEvents.push([
		"esc",
		_.bind(function(e) {
			this.changeState(DEFAULTSTATE);
			e.preventDefault();
			e.stopPropagation();
		}, this),
		"keyup"
	]);

	/**
	 * @override
	 */
	this._override("initialize", function(widget, toolManager) {
		this._super(widget, toolManager);
		this.changeState(DEFAULTSTATE);
	});

	/**
	 * @override
	 */
	this._override("activate", function(interruptRequest) {
		this._super(interruptRequest);
		if (this._viewport) {
			this._viewport
				.getFirstViewOccurrence("Main")
				.getViewContext()
				.addClass("enableTextSelection");
		}
		this._initializeRbush();
		this.f.getModule("viewport").addLayer(this.getIID(), this._textSelectionLayerModel);
	});

	/**
	 * @override
	 */
	this._override("deactivate", function(interruptRequest) {
		this._super(interruptRequest);
		if (!interruptRequest) {
			this._rbush = null;
			this.changeState(DEFAULTSTATE);
			// The SVGLayer should be first in the dom
			this.f.emit("viewport.viewmodel.Main.layers.swapIfIndexIsHigher", "SVGLayer", "SVGDifferenceLayer");
			this._textSelectionLayerModel.modelChanged(true);
			this.f.getModule("viewport").removeLayer(this.getIID());
			if (this._viewport) {
				this._viewport
					.getFirstViewOccurrence("Main")
					.getViewContext()
					.removeClass("enableTextSelection");
			}
		}
	});

	// activate text selection tool
	this.passiveKeyEvents.push([
		"t",
		_.bind(function(e) {
			if (this.f.getActivePage().hasSvg && this.f.getActivePage().hasSvg()) {
				e.preventDefault();
				f.emit("toolManager.activateTool", this);
			}
		}, this),
		"keypress"
	]);
};

// extend from the AbstractTool
// inherit
TextSelectionTool.prototype = Object.create(AbstractTool.prototype);
TextSelectionTool.prototype.constructor = TextSelectionTool;

TextSelectionTool.prototype._onPageChange = function() {
	if (this.isActive()) {
		this.changeState(DEFAULTSTATE);
		this._rbush = null;
		this._initializeRbush();
	}
};

/**
 * Returns the default cursor for the tool
 * @returns {String}
 */
TextSelectionTool.prototype.getDefaultCursor = function() {
	return "cursor-textSelect";
};

/**
 * If the world is animating, cancel it and wait for the svglayer to have been repainted
 */
TextSelectionTool.prototype._initializeRbush = function() {
	if (this._rbush) {
		return;
	}
	var svgLayer;
	if (this._widget.getWorld().isAnimating()) {
		// As there is a lot of asynchronicity in the viewer, we have to make sure the SVG's transform is the same as the viewer's transform
		// To make sure this is the case we first have to listen to the a repaint of the view.
		// This will trigger all layers to rerender. This also happens asynchronous. So we should then listen to the new repaint of the svg layer
		svgLayer = this._viewport.getFirstViewOccurrence("Main").getLayer("SVGLayer");
		var viewListener = this._registerListener(
			"repainted",
			function() {
				// The widget has been repainted, now we have to wait to make sure the svg's transformation is updated as well
				var layerListener = this._registerListener(
					"repainted",
					function() {
						this._calculateRbush();
						this._unregisterListener(layerListener);
					},
					this,
					svgLayer
				);
				this._unregisterListener(viewListener);
			},
			this,
			this._widget
		);

		this._widget.getWorld().cancelAnimation();
	} else {
		this._calculateRbush();
	}
};

/**
 * Initializes the Rbush with the text elements in the svg
 */
TextSelectionTool.prototype._calculateRbush = function() {
	this._rbush = new rbush();
	var rbushElems = [];
	var svgLayer = this._viewport
		.getFirstViewOccurrence("Main")
		.getLayer("SVGLayer")
		.getLayerElement();
	if (this.f.getDocumentController().hasDifferenceDocumentPage()) {
		var svgDifferenceLayer = this._viewport
			.getFirstViewOccurrence("Main")
			.getLayer("SVGDifferenceLayer")
			.getLayerElement();
		if (!this.f.getActivePage().get("isRightPage")) {
			// For the text selection to work in readerspread mode the left page should be the first in the dom.
			this.f.emit("viewport.viewmodel.Main.layers.swapIfIndexIsHigher", "SVGLayer", "SVGDifferenceLayer");

			this._calculateRbushElements($(svgLayer), rbushElems);
			this._calculateRbushElements($(svgDifferenceLayer), rbushElems);
		} else {
			// For the text selection to work in readerspread mode the left page should be the first in the dom.
			this.f.emit("viewport.viewmodel.Main.layers.swapIfIndexIsHigher", "SVGDifferenceLayer", "SVGLayer");
			this._calculateRbushElements($(svgDifferenceLayer), rbushElems);
			this._calculateRbushElements($(svgLayer), rbushElems);
		}
	} else {
		this._calculateRbushElements($(svgLayer), rbushElems);
	}

	try {
		this._rbush.load(rbushElems);
	} catch (e) {
		console.error("rbush - could not insert character bounding boxes");
	}

	if (process.env.NODE_ENV === "development") {
		// this._renderRbush(); //debug
	}
};

/**
 * @param {Node} svgElement
 * @param {Array.<Object>} rbushElems
 */
TextSelectionTool.prototype._calculateRbushElements = function(svgElement, rbushElems) {
	var self = this;
	svgElement.find("text").each(function(index) {
		var el = $(this);
		var length = el[0].getNumberOfChars();

		for (var j = 0; j < length; j++) {
			var coords = self._convertCoords(j, el[0]);
			var obj = {
				minX: coords.origin.x,
				minY: coords.origin.y,
				maxX: coords.origin.x + coords.size.width,
				maxY: coords.origin.y + coords.size.height,
				charIndex: j,
				element: el[0],
				elementIndex: rbushElems.length
			};
			rbushElems.push(obj);
		}
	});
};

/**
 * @param {Integer} index of character in text element
 * @param {Node} element - the text element
 */
TextSelectionTool.prototype._convertCoords = function(index, element, svg) {
	var extent = element.getExtentOfChar(index);
	var matrix = element.getScreenCTM();
	svg = this._viewport
		.getFirstViewOccurrence("Main")
		.getLayer("SVGLayer")
		.getLayerElement();

	var pt = svg[0].createSVGPoint();
	pt.x = extent.x;
	pt.y = extent.y;
	pt = pt.matrixTransform(matrix);

	var pt2 = svg[0].createSVGPoint();
	pt2.x = extent.x + extent.width;
	pt2.y = extent.y + extent.height;
	pt2 = pt2.matrixTransform(matrix);

	var parent = this._viewport
		.getFirstViewOccurrence("Main")
		.getViewContext()
		.offset();
	var scenePoint = new ShapeUtils.Point(pt.x - parent.left, pt.y - parent.top);
	var sceneSize = new ShapeUtils.Size(pt2.x - pt.x, pt2.y - pt.y);

	return this.sceneToWorldRect(new ShapeUtils.Rect(scenePoint, sceneSize));
};

TextSelectionTool.prototype.canBeInterrupted = function(tool) {
	return this.getCurrentStateName() !== DRAGGINGSTATE;
};

if (process.env.NODE_ENV === "development") {
	/**
	 * Renders rectangles around the text elements
	 * @private
	 */
	TextSelectionTool.prototype._renderRbush = function() {
		this.requestCanvasLayer("Main");
		var ctx = this.getCanvasCtx();
		if (!this._interval) {
			this._interval = setInterval(
				_.bind(function() {
					if (this._rbush) {
						var allElements = this._rbush.all();
						ctx.setTransform(1, 0, 0, 1, 0, 0);
						ctx.clearRect(0, 0, ctx.canvas.width, ctx.canvas.height);
						ctx.beginPath();
						_.each(
							allElements,
							function(obj) {
								var worldOrigin = new ShapeUtils.Point(obj.minX, obj.minY);
								var worldSize = new ShapeUtils.Size(obj.maxX - obj.minX, obj.maxY - obj.minY);
								var sceneOrigin = this.worldToScenePoint(worldOrigin);
								var sceneSize = this.worldToSceneSize(worldSize);
								ctx.rect(sceneOrigin.x, sceneOrigin.y, sceneSize.width, sceneSize.height);
							},
							this
						);

						ctx.stroke();
					}
				}, this),
				1000 / 30
			);
		}
	};
}

/**
 * Gets all the elements inside the range, start and end node are not included
 * @private
 * @param   {Range}   range - the range object
 * @returns {Array.<Node>} an array of nodes inside the given range
 */
TextSelectionTool.prototype._getElementsInsideRange = function(range) {
	var start = range.startContainer.parentNode;
	var end = range.endContainer.parentNode;
	var ancestor = range.commonAncestorContainer;
	var el;
	var before = [];
	while (start.parentNode !== ancestor) {
		el = start;
		while (el.nextSibling) {
			before.push((el = el.nextSibling));
		}
		start = start.parentNode;
	}

	var after = [];
	while (end.parentNode !== ancestor) {
		el = end;
		while (el.previousSibling) {
			after.push((el = el.previousSibling));
		}
		end = end.parentNode;
	}
	after.reverse();

	while ((start = start.nextSibling) !== end) {
		before.push(start);
	}
	return before.concat(after);
};

/**
 * Calculates the bounding box of a svg text tag with a given start and end offset
 * @private
 * @param   {Node} text  svg text tag
 * @param   {Integer} start Char position of start
 * @param   {Integer} end   Char position of end
 * @returns {ShapeUtils.Rect} The world bounding box of the given text tag with start and end offset
 */
TextSelectionTool.prototype._calculateBoundingBoxOfText = function(text, start, end) {
	var boundingBox = null;
	for (var i = start; i < end; i++) {
		if (!boundingBox) {
			// Start boundingBox with the first char of the text
			boundingBox = this._convertCoords(i, text);
		} else {
			// Keep expanding the boundingBox with the other chars
			boundingBox = boundingBox.calculateArea(this._convertCoords(i, text));
		}
	}
	return boundingBox;
};

/**
 * Determines if this tool can be activated
 *
 * @returns  {Boolean}
 */
TextSelectionTool.prototype.canBeActivated = function() {
	var compareModule = this.f.getModule("compare");
	if (compareModule) {
		var mode = compareModule.getActiveMode();
		var allowedModes = [CompareConstants.mode.SHOW_CURRENT, CompareConstants.mode.READERSPREAD];
		return !mode || allowedModes.indexOf(mode.getName()) !== -1;
	}
	return true;
};

export default TextSelectionTool;
