import _ from "underscore";
import ShapeUtils from "@esko/shapeutils";
import ViewerUtils from "@esko/web-viewer/utils/ViewerUtils";
import Cache from "@esko/cache";
import DashUtils from "@esko/dashutils";
import CadConstants from "@esko/web-viewer/cad/CadConstants";
import ConversionConstants from "@esko/web-viewer/utils/ConversionConstants";
import Bezier from "bezier-js";
var debugOverlay = false;

/**
 * Image cache used for caching loaded image information
 */
var imageCache = new Cache(30, false);
imageCache._loadItem = function(name) {
	var img = new Image();
	img.onload = _.bind(function() {
		this._registerItem(name, img);
	}, this);
	img.src = name;
};

/**
 * Tinted icon cache
 *
 * name --> url|r|g|b
 *
 * loads image with imageCache, and then makes a tinted version of it
 */
var iconCache = new Cache(30, false);
iconCache._loadItem = function(name) {
	var icon = name.split("|");
	var color = {
		r: icon[1],
		g: icon[2],
		b: icon[3]
	};

	var generateIcon = _.bind(function(img, color) {
		var result = new Image();

		result.onload = _.bind(function() {
			this._registerItem(name, result);
		}, this);

		generateTintImage(img, color.r, color.g, color.b, result);
	}, this);

	// get image from cache
	var image = imageCache.get(icon[0], function(img) {
		generateIcon(img, color);
	});

	if (image) {
		// the image was in the cache, generate icon
		generateIcon(image, color);
	}
};

/**
 * Make a tinted version of an image
 *
 * @param {Image} img - the source image
 * @param {Number} red - red tint
 * @param {Number} green - green tint
 * @param {Number} blue - blue tint
 * @param {Image} [resultImage] - result image, if provided, this image will be used to store the tinted image in (useful to bind callbacks if needed)
 */
function generateTintImage(img, red, green, blue, resultImage) {
	var w = img.width;
	var h = img.height;

	var canvas = document.createElement("canvas");
	canvas.width = w;
	canvas.height = h;

	var ctx = canvas.getContext("2d");
	ctx.drawImage(img, 0, 0, w, h);

	var pixels = ctx.getImageData(0, 0, w, h);

	for (var i = 0; i < pixels.data.length; i += 4) {
		pixels.data[i] = pixels.data[i] * (red / 255); //R
		pixels.data[i + 1] = pixels.data[i + 1] * (green / 255); //G
		pixels.data[i + 2] = pixels.data[i + 2] * (blue / 255); //B
		// we don't need to do anything for alpha
	}

	ctx.putImageData(pixels, 0, 0);

	if (resultImage === undefined) {
		resultImage = new Image();
	}

	resultImage.src = canvas.toDataURL();

	return resultImage;
}

/**
 * @memberof Utils
 * @exports DrawUtils
 */
var DrawUtils = {
	LWTYPE_RECT: -1,
	LWTYPE_ELLIPSE: -2,
	LWTYPE_LINEARRAY: -3,
	LWTYPE_LINE: -4,
	LWTYPE_ARROW: -5,
	LWTYPE_IMAGE: -7,
	LWTYPE_TEXT: -21,
	LWTYPE_DASHEDBOX: -100,
	LWTYPE_DASHEDLINE: -101,
	LWANCHOR_TOPLEFT: 1,
	LWANCHOR_TOPRIGHT: 2,
	LWANCHOR_TOPCENTER: 3,
	LWANCHOR_BOTTOMLEFT: 4,
	LWANCHOR_BOTTOMRIGHT: 5,
	LWANCHOR_BOTTOMCENTER: 6,
	LWANCHOR_CENTERLEFT: 7,
	LWANCHOR_CENTERRIGHT: 8,
	LWANCHOR_CENTERCENTER: 9,
	LWANCHOR_CENTERFREE: 10,
	LWANCHOR_HORIZONTAL: 11,
	LWANCHOR_VERTICAL: 12,
	LWANCHOR_START: 13,
	LWANCHOR_END: 14,

	/**
	 * Draw an overlay of linework
	 * @param {Core.ViewerFacade} f
	 * @param {CanvasRenderingContext2D} ctx - canvas context
	 * @param {Array<LineworkWrapper>} linework - array of linework
	 * @param {ShapeUtils.Rect} viewport - rectangle representing viewport
	 * @param {Function} imagecallback - callback called everytime a linework item is of type LWTYPE_IMAGE and when item becomes available (e.g. after loading if not in cache)
	 * @param {Number} [zoom = 1] - current zoom level
	 * @param {Transformation} t - transformation
	 * @param {String} [color = linework[i].color] - The linework color "#RGB"
	 * @param {Number} [alpha = 1] - transparency of the overlay
	 */
	drawOverlay: function(f, ctx, linework, viewport, imagecallback, zoom, t, color, alpha) {
		var lineworksDrawn = 0;
		if (!zoom) zoom = 1;

		ctx.save();

		if (viewport !== undefined) {
			//set the start location so we will draw on pixels, and make sure to translate by +0.5 pixels
			var left = viewport.left(); //mm, distance between the left side of the document and left bound of the visible rect
			var top = viewport.top(); //mm, distance between the top side of the document and top bound of the visible rect
			//first get the mm to pixels with zoom (ppmm) and only get the part that's not a full pixel, to draw lines with linewidth 1 on exactly
			//1 pixel make sure to draw at +0.5 pixel. The result is the amount of pixels the coordinates in the canvas have to be shifted

			var leftPx = left * zoom; //translate the mm coordinates to pixels
			var topPx = top * zoom;
			if (Math.abs(leftPx % 1) < 0.000001 || Math.abs(1 - (leftPx % 1)) < 0.000001) {
				//if there's a small difference (rounding error), round to nearest integer (happens on graphics files and makes the grid corrupt)
				leftPx = Math.round(leftPx);
			}
			if (Math.abs(topPx % 1) < 0.000001 || Math.abs(1 - (topPx % 1)) < 0.000001) {
				topPx = Math.round(topPx);
			}

			var xShift = 0.5 + (leftPx % 1); //pixels, amount the canvas needs to shift, compensate the rest of a hole pixel and 0.5 pixel
			var yShift = 0.5 + (topPx % 1); //pixels, amount the canvas needs to shift, compensate the rest of a hole pixel and 0.5 pixel

			//since coordinate system is working in mm and not pixels, convert the pixel back to mm
			ctx.translate(xShift / zoom, yShift / zoom);
		}

		if (alpha) {
			ctx.globalAlpha = alpha;
		} else {
			ctx.globalAlpha = 1;
		}
		for (var i = 0; i < linework.length; i++) {
			var wrap = linework[i];
			if (wrap == null) {
				continue;
			}
			if (debugOverlay) {
				ctx.save();
				ctx.lineWidth = 0.01;
				ctx.fillStyle = "red";
				this.drawRect(ctx, wrap.rect.origin.x, wrap.rect.origin.y, wrap.rect.size.width, wrap.rect.size.height, 0, "red", "red");
				ctx.restore();
			}

			var viewrect = wrap.rect.copy();
			if (wrap.staticsize) {
				var center = viewrect.center();
				var size = new ShapeUtils.Size(viewrect.size.width / zoom, viewrect.size.height / zoom);
				viewrect = new ShapeUtils.Rect(new ShapeUtils.Point(center.x - size.width / 2, center.y - size.height / 2), size);
			}

			if (wrap.NODRAW) {
				// this wrapper should not be drawn
				continue;
			}

			if (viewport && viewrect && !ShapeUtils.rectIsNull(viewrect)) {
				//check if the rect is in the current viewport. If not, don't worry about drawing it
				if (!viewport.rectIntersect(wrap.rect) && wrap.type != this.LWTYPE_TEXT) {
					continue;
				}
			}

			lineworksDrawn++;

			ctx.save();

			var linethickness = Math.floor(wrap.linewidth);
			if (linethickness === 0) {
				linethickness = 1;
			}

			if (!wrap.staticwidth) {
				if (zoom > 1) {
					linethickness /= Math.ceil(zoom);
				} else {
					linethickness /= zoom;
				}
			}

			var strokecolor = wrap.color;
			var fillcolor = wrap.fillcolor;
			if (color !== null && color !== undefined) {
				strokecolor = color;
				if (fillcolor) {
					fillcolor = color;
				}
			}

			var fill = fillcolor ? true : false;
			var stroke = strokecolor ? true : false;
			if (strokecolor) {
				ctx.strokeStyle = strokecolor;
			}
			if (fill) {
				if (color) {
					//general color for entire overlay was defined, use this color with an alpha
					var colorObj = ViewerUtils.rgbColor(fillcolor);
					ctx.fillStyle = "rgba(" + colorObj.r + ", " + colorObj.g + ", " + colorObj.b + ", 0.5)";
				} else {
					ctx.fillStyle = fillcolor;
				}
			}
			if (linethickness) {
				ctx.lineWidth = linethickness;
			}

			switch (wrap.type) {
				case this.LWTYPE_RECT: //rect
					var radius;
					if (wrap.cornerRadius) {
						radius = wrap.cornerRadius;
					} else {
						radius = 0;
					}
					this.drawRect(ctx, viewrect.origin.x, viewrect.origin.y, viewrect.size.width, viewrect.size.height, radius, fill, stroke);

					break;
				case this.LWTYPE_ELLIPSE: //elipse
					this.drawEllipse(ctx, viewrect.origin.x, viewrect.origin.y, viewrect.size.width, viewrect.size.height, fill, stroke);
					break;
				case this.LWTYPE_IMAGE:
					var ref = wrap.imageRef;
					var img = imageCache.get(ref, imagecallback);
					if (!img) continue;
					var tintedImg = undefined;

					if (wrap.color) {
						var rgb = wrap.rgbColor();
						tintedImg = iconCache.get([ref, rgb.r, rgb.g, rgb.b].join("|"), imagecallback);
					}

					var imgx = viewrect.origin.x;
					var imgy = viewrect.origin.y;

					imgy -= img.naturalHeight / zoom;

					if (viewrect.size.width == 0 || viewrect.size.height == 0) {
						ctx.save();
						// no width/height is defined in the linework wrapper, draw the image in its real pixel size

						var transformedRect = viewrect;
						if (t) {
							var tmpT = t.copy();
							tmpT.translate(xShift / zoom, yShift / zoom);
							transformedRect = tmpT.transformedRect(transformedRect);
						}

						ctx.setTransform(1, 0, 0, 1, 0, 0);

						// if .imageOffset is defined, add the offest to the image (in pixels)
						var offsetX = 0,
							offsetY = 0;
						if (wrap.imageOffset) {
							offsetX = wrap.imageOffset.width;
							offsetY = wrap.imageOffset.height;
						}

						// make sure to draw on pixels (Math.floor)
						ctx.drawImage(
							tintedImg ? tintedImg : img,
							Math.ceil(transformedRect.origin.x - 0.5) - offsetX,
							Math.ceil(transformedRect.origin.y - 0.5) - offsetY
						);
						ctx.restore();
					} else {
						ctx.drawImage(tintedImg ? tintedImg : img, imgx, imgy, viewrect.size.width, viewrect.size.height);
					}

					break;
				case this.LWTYPE_TEXT: //text
					if (!fillcolor) {
						ctx.fillStyle = strokecolor;
					}
					var textH = ViewerUtils.parseFloat(wrap.textoptions["font-size"]);
					if (wrap.staticsize) {
						textH *= zoom;
					}
					var fs = textH + "px " + wrap.textoptions["font-family"];
					ctx.font = fs;

					var text = ViewerUtils.trim(wrap.textoptions["text"]);
					ctx.textAlign = wrap.textoptions["textAlignment"];

					if (viewport) {
						var textW = ctx.measureText(text).width;
						var textSize = new ShapeUtils.Size(textW, textH);
						var textRect = ShapeUtils.rectMake(
							viewrect.origin.x - textSize.width / 2,
							viewrect.origin.y - textSize.height / 2,
							textSize.width,
							textSize.height
						);
						if (!viewport.rectIntersect(textRect)) {
							//continue;
						}
					}

					ctx.save();
					if (wrap.staticsize) {
						var transformedRect = viewrect;
						if (t) {
							transformedRect = t.transformedRect(transformedRect);
						}

						ctx.setTransform(1, 0, 0, 1, 0, 0);

						ctx.fillText(text, transformedRect.origin.x, transformedRect.origin.y);

						ctx.restore();
					} else {
						//Compensate the rotation
						ctx.translate(viewrect.origin.x, viewrect.origin.y);
						ctx.rotate((wrap.textoptions.rotate * Math.PI) / 180);
						ctx.translate(-viewrect.origin.x, -viewrect.origin.y);

						//Fill the text
						ctx.fillText(text, viewrect.origin.x, viewrect.origin.y);
					}
					ctx.restore();

					break;
				case this.LWTYPE_DASHEDBOX:
					var dash = wrap.dash;

					ctx.save();
					//					    	if(t){
					//					    		viewrect = t.transformedRect(viewrect);
					//					    	}
					//					    	ctx.setTransform(1,0,0,1,0,0);
					ctx.lineWidth = linethickness;
					DashUtils.drawDashedRect(ctx, viewrect, dash, !wrap.staticwidth ? zoom : null);

					ctx.restore();

					break;
				case this.LWTYPE_DASHEDLINE:
					var dash = wrap.dash;

					ctx.beginPath();
					if (viewrect.size.width == 0) {
						//Vertical
						DashUtils.drawDash(
							ctx,
							viewrect.origin.x,
							viewrect.origin.y,
							viewrect.origin.x,
							viewrect.origin.y + viewrect.size.height,
							dash
						);
					} else {
						//Horizontal
						DashUtils.drawDash(ctx, viewrect.origin.x, viewrect.origin.y, viewrect.origin.x + viewrect.size.width, viewrect.origin.y, dash);
					}

					ctx.stroke();

					break;
				default:
					var arrow = wrap.type == this.LWTYPE_ARROW ? true : false;
					var points = wrap.pointArray;
					var dash = wrap.dash;
					ctx.beginPath();
					if (points.length > 0) {
						this.drawLine(f, ctx, points, arrow, t, zoom, fill, dash, wrap.lineDecorator);
					}
					break;
			}

			if (wrap.showIndicator) {
				/**
				 * Show an indicator of the object
				 * By default, this is placed on the top left corner of the linework
				 *
				 * Unless:
				 * 	- Array of points: show the indicator at the first point
				 *  - Ellipse: top center
				 */

				ctx.save();
				ctx.setTransform(1, 0, 0, 1, 0, 0);

				var textH = 14;
				ctx.font = textH + "px Arial";
				ctx.textAlign = "center";

				var text = wrap.indicatorText;

				var boxwidth = ctx.measureText(text).width;
				var textHeight = (textH * 72) / 96;
				var boxheight = textHeight;

				var offset = boxheight * 0.5;
				boxwidth += offset;
				boxheight += offset;

				var indicationRect = viewrect;
				if (t) {
					indicationRect = t.transformedRect(indicationRect);
				}
				var indicationPoint = indicationRect.origin; // new ShapeUtils.Point(indicationRect.left(), indicationRect.top());
				if (wrap.type == this.LWTYPE_ELLIPSE) {
					indicationPoint.x = indicationRect.center().x - boxwidth / 2;
				} else if (wrap.pointArray && wrap.pointArray.length > 0) {
					indicationPoint = t.transformedPoint(wrap.pointArray[0]);
				}

				var textBox = new ShapeUtils.Rect(
					new ShapeUtils.Point(indicationPoint.x, indicationPoint.y),
					new ShapeUtils.Size(boxwidth, boxheight)
				);

				ctx.fillStyle = wrap.indicatorColor;
				ctx.strokeStyle = "black";
				ctx.lineWidth = 2;

				var radius = textBox.size.height / 2;
				radius = 0;
				this.drawRect(ctx, textBox.origin.x, textBox.origin.y, textBox.size.width, textBox.size.height, radius, true, true);

				ctx.fillStyle = wrap.indicatorTextColor;
				ctx.fillText(text, textBox.center().x, indicationPoint.y + boxheight / 2 + textHeight / 2);

				ctx.restore();
			}
			ctx.restore();
		}
		ctx.restore();
		if (debugOverlay) console.log("endOverlay (" + lineworksDrawn + ")");
	},
	generateTintImage: generateTintImage,
	/**
	 * Draws a rounded rectangle using the current state of the canvas.
	 * If you omit the last three params, it will draw a rectangle
	 * outline with a 5 pixel border radius
	 * @param {CanvasRenderingContext2D} ctx - Canvas context
	 * @param {Number} x - The top left x coordinate
	 * @param {Number} y - The top left y coordinate
	 * @param {Number} width - The width of the rectangle
	 * @param {Number} height - The height of the rectangle
	 * @param {Number} radius - The corner radius. Defaults to 5;
	 * @param {Boolean} fill - Whether to fill the rectangle. Defaults to false.
	 * @param {Boolean} stroke - Whether to stroke the rectangle. Defaults to true.
	 */
	drawRect: function(ctx, x, y, width, height, radius, fill, stroke) {
		if (typeof stroke == "undefined") {
			stroke = true;
		}
		if (typeof radius === "undefined") {
			radius = 5;
		}
		ctx.save();
		ctx.beginPath();
		ctx.moveTo(x + radius, y);
		ctx.lineTo(x + width - radius, y);
		ctx.quadraticCurveTo(x + width, y, x + width, y + radius);
		ctx.lineTo(x + width, y + height - radius);
		ctx.quadraticCurveTo(x + width, y + height, x + width - radius, y + height);
		ctx.lineTo(x + radius, y + height);
		ctx.quadraticCurveTo(x, y + height, x, y + height - radius);
		ctx.lineTo(x, y + radius);
		ctx.quadraticCurveTo(x, y, x + radius, y);
		ctx.closePath();
		if (stroke) {
			ctx.stroke();
		}
		if (fill) {
			ctx.fill();
		}
		ctx.restore();
	},
	/**
	 * Draw a line between points
	 * @param {ViewerFacade} f - Viewer facade
	 * @param {CanvasRenderingContext2D} ctx - Canvas context
	 * @param {Array<ShapeUtils.Point>} points - Array of points to draw lines in between
	 * @param {Boolean} arrow - True to make arrow between the first two points
	 * @param {Transformation} t - Transformation
	 * @param {Number} zoom - Zoom level
	 * @param {Boolean} fill - True to fill area in between the lines
	 * @param {Boolean} dash - True to dash line, false otherwise
	 */
	drawLine: function(f, ctx, points, arrow, t, zoom, fill, dash, decorator) {
		var debugdrawing = false;
		ctx.save();
		ctx.beginPath();

		for (var i = 0; i < points.length; i++) {
			var pointObject = points[i];

			var p = new ShapeUtils.Point(pointObject.x, pointObject.y);

			//round the points so we are drawing right on the pixels to avoid that linewidth 1 spreads over 2 pixels
			var x = Math.ceil(pointObject.x * zoom) / zoom;
			var y = Math.ceil(pointObject.y * zoom) / zoom;

			if (this.dashedLineWarning === undefined) {
				this.dashedLineWarning = _.once(function() {
					f.i18n.loadBundle("viewer", function() {
						f.emit("error.WARNING", f.i18n.t("DashedLineSupport"));
					});
				});
			}

			var dashSizes = dash;

			if (dash && zoom) {
				//if zoom is defined, draw dashsizes in pixel instead of mm
				dashSizes = _.map(dash, function(value) {
					//map all dashsizes from mm to pixels
					return Math.round(value * zoom) / zoom / zoom;
				});
			}

			var supportsLineDash = true; //are the lines dashed or not
			if (dashSizes) {
				//if there are dashed lines
				if (ctx.setLineDash !== undefined) {
					//check if setLineDash exists and dashed lines are supported
					ctx.setLineDash(dashSizes);
				} else if (ctx.mozDash !== undefined) {
					//for older versions of firefox
					ctx.mozDash = dashSizes;
				} else {
					supportsLineDash = false; //dashed lines are not supported
				}
			}

			switch (pointObject.type) {
				case 0: // MOVE TO
					ctx.moveTo(x, y);
					break;
				case 1: // LINE
					if (supportsLineDash) {
						//if dashed lines are supported make use of this feature
						ctx.lineTo(x, y);
					} else {
						//dashed lines are not supported draws the dashed lines
						var prPoint = points[i - 1];
						var previousPoint = new ShapeUtils.Point(prPoint.x, prPoint.y);
						DashUtils.drawDash(ctx, previousPoint.x, previousPoint.y, x, y, dashSizes);
					}
					if (debugdrawing) ctx.stroke();
					break;
				case 2: // BEZIER CUBE
					var p2 = new ShapeUtils.Point(pointObject.v1, pointObject.v2);
					var p3 = new ShapeUtils.Point(pointObject.v3, pointObject.v4);

					if (!supportsLineDash) {
						//show the dashed line warning if the browser doesn't support them
						this.dashedLineWarning();
					}

					ctx.bezierCurveTo(p.x, p.y, p2.x, p2.y, p3.x, p3.y);
					if (debugdrawing) ctx.stroke();
					break;
				case 3: // ARC
					var prPoint = points[i - 1];
					var previousPoint = new ShapeUtils.Point(prPoint.x, prPoint.y);
					var rx = pointObject.v1;
					var angle = pointObject.v3;
					var largeArcFlag = pointObject.v4;
					var sweepFlag = pointObject.v5;

					if (!supportsLineDash) {
						this.dashedLineWarning();
					}

					this.drawArcFrom(ctx, previousPoint, p, rx, largeArcFlag, sweepFlag, angle);

					if (debugdrawing) ctx.stroke();
					break;
				case 4: //NICK
					//First point of the nick
					var p = new ShapeUtils.Point(pointObject.x, pointObject.y);

					//Second point of the nick is the same as the first point, minus the coordinates given in the nick definition
					var p2 = p.copy();
					p2.x -= pointObject.v3 * 10;
					p2.y -= pointObject.v2 * 10;

					//The text should be a bit away from the second point
					var p3 = p2.copy();
					p3.x -= pointObject.v3 * 5;
					p3.y -= pointObject.v2 * 5;

					//The rotation is the rotation between point 1 and 2 + 90 degrees
					var rotation = Math.atan2(p2.y - p.y, p2.x - p.x) + Math.PI / 2;

					//Move
					ctx.moveTo(p.x, p.y);
					//Line
					ctx.lineTo(p2.x, p2.y);

					//Save the context and rotate to draw the text
					ctx.save();

					ctx.translate(p3.x, p3.y);
					ctx.rotate(rotation);
					ctx.translate(-p3.x, -p3.y);

					var textH = 8;
					ctx.font = textH + "px Arial";
					ctx.textAlign = "center";
					ctx.fillStyle = "blue";
					var text = ViewerUtils.parseFloat(pointObject.v1).toFixed(2);
					ctx.fillText(text, p3.x, p3.y);

					//Restore the context
					ctx.restore();

					break;
				default:
					console.log("a line that is not supported: " + pointObject.type);
					break;
			}
		}
		if (!debugdrawing && !fill) {
			ctx.stroke();
		}
		if (arrow && points.length > 1) {
			ctx.beginPath();
			// Find out an exellent tutorial: http://www.dbp-consulting.com/tutorials/canvas/CanvasArrow.html
			// Find out an exellent tutorial: http://www.dbp-consulting.com/tutorials/canvas/CanvasArcTo.html
			var startP = points[0];
			var endP = points[1];

			var endPoint;
			var startPoint;
			if (endP.type == 2) {
				startPoint = new ShapeUtils.Point(endP.v1, endP.v2);
				endPoint = new ShapeUtils.Point(endP.v3, endP.v4);
			} else {
				startPoint = new ShapeUtils.Point(startP.x, startP.y);
				endPoint = new ShapeUtils.Point(endP.x, endP.y);
			}
			var start = startPoint;
			var end = endPoint;
			if (t) {
				start = t.transformedPoint(start);
				end = t.transformedPoint(end);
				ctx.save();
				ctx.setTransform(1, 0, 0, 1, 0, 0);
			}

			var u = new ShapeUtils.Point(end.x - start.x, end.y - start.y);
			var length = Math.sqrt(u.x * u.x + u.y * u.y);

			if (length === 0) {
				u.x = 0;
				u.y = 0;
			} else {
				u.x /= length;
				u.y /= length;
			}

			var v = new ShapeUtils.Point(-u.y, u.x);

			var arrowSize = 8;

			u.x *= arrowSize;
			u.y *= arrowSize;
			v.x *= arrowSize / 1.0;
			v.y *= arrowSize / 1.0;
			var p = new ShapeUtils.Point(end.x, end.y);
			p.x += -u.x + v.x;
			p.y += -u.y + v.y;
			var p2 = new ShapeUtils.Point(p.x, p.y);
			p2.x -= 2 * v.x;
			p2.y -= 2 * v.y;

			ctx.moveTo(p.x, p.y);
			ctx.lineTo(end.x, end.y);
			ctx.lineTo(p2.x, p2.y);

			if (debugdrawing) ctx.stroke();

			if (t) {
				ctx.restore();
			}
		}
		if (fill) {
			ctx.fill();
		} else if (!debugdrawing) {
			ctx.stroke();
		}

		ctx.restore();

		if (decorator) {
			this.drawDecorators(f, ctx, points, arrow, t, zoom, fill, dash, decorator);
		}
	},

	drawDecorators: function(f, ctx, points, arrow, t, zoom, fill, dash, decoratorType) {
		//ctx.beginPath();

		var decorator = CadConstants.LineDecoratorMap[decoratorType];
		if (!decorator) {
			return;
		}
		// we keep track of the current offset, as paths are combined, the dashpattern runs accros multiple lines/curves/arcs
		var currentOffset = decorator.offset;
		// lenght of the patter (this is the space betwee two decorators)
		var decoratorLength = decorator.length;
		// size of the decorator mark
		var decoratorMarkSize = decorator.length - decorator.offset;

		ctx.save();

		// we draw in Pixels
		ctx.scale(1 / zoom, 1 / zoom);

		// keep track of the previous point
		var previousPoint = new ShapeUtils.Point();
		for (var i = 0; i < points.length; i++) {
			var pointObject = points[i];

			//round the points so we are drawing right on the pixels to avoid that linewidth 1 spreads over 2 pixels
			var x = Math.ceil(pointObject.x * zoom);
			var y = Math.ceil(pointObject.y * zoom);
			var p = new ShapeUtils.Point(x, y);

			switch (pointObject.type) {
				case 0: // MOVE TO
					// a move, disconnects the path, the dash pattern doesn't run further
					previousPoint = p;
					currentOffset = decorator.offset;
					break;
				case 1: // LINE
					var length = previousPoint.getDistance(p);

					ctx.save();

					var dx = p.x - previousPoint.x;
					var dy = p.y - previousPoint.y;
					// angle for decorators
					var theta = Math.atan2(dy, dx) - 0.5 * Math.PI;
					ctx.translate(previousPoint.x, previousPoint.y);
					ctx.rotate(theta);

					ctx.translate(0, currentOffset);

					var position = currentOffset;
					while (length > decoratorMarkSize && position < length) {
						this.drawDecoratorInstance(ctx, decorator, 0);
						// uncomment for some debugging info
						// ctx.fillText("#" + position + " LIN", 0, 0);
						segment++;
						ctx.translate(0, decoratorLength);
						position += decoratorLength;
					}

					ctx.restore();
					currentOffset = Math.floor(position + 3 * decoratorLength - length) % decoratorLength;
					previousPoint = p;
					break;
				case 2: // BEZIER CUBE
					var p2 = new ShapeUtils.Point(Math.ceil(pointObject.v1 * zoom), Math.ceil(pointObject.v2 * zoom));
					var p3 = new ShapeUtils.Point(Math.ceil(pointObject.v3 * zoom), Math.ceil(pointObject.v4 * zoom));

					var curve = new Bezier(previousPoint.x, previousPoint.y, p.x, p.y, p2.x, p2.y, p3.x, p3.y);
					var curveLength = curve.length();
					var curvePosition = currentOffset;

					var segment = 1,
						ti = curvePosition / curveLength,
						arrivedPosition = null;
					while (curvePosition < curveLength && curveLength > decoratorMarkSize && ti <= 1) {
						// first guess of the position
						if (arrivedPosition !== null) {
							ti += ((1 - ti) * decoratorLength) / (curveLength - arrivedPosition);
						} else {
							ti = curvePosition / curveLength;
						}

						// step size for binary search for correct position
						var step = 0.5 * (decoratorLength / curveLength);

						var split = curve.split(ti);
						arrivedPosition = split.left.length();

						var iterations = 0;

						ti = Math.max(0, Math.min(1, ti));
						// as long as we are not close enough the the correct distance, make a new estimate (maximum 15 guesses)
						while (Math.abs(arrivedPosition - curvePosition) > 0.5 && iterations < 15) {
							ti = ti + Math.sign(curvePosition - arrivedPosition) * step;
							step /= 2;
							var split = curve.split(ti);
							arrivedPosition = split.left.length();
							iterations++;
						}

						if (ti >= 0 && ti <= 1) {
							var point = curve.get(ti);
							var normal = curve.normal(ti);
							var theta = Math.PI + Math.atan2(normal.y, normal.x);

							ctx.save();
							ctx.translate(point.x, point.y);
							// uncomment for some debugging info
							// ctx.rotate(theta);
							// ctx.fillText("#" + segment + " BEZ -- " + iterations, 0, 0);
							// ctx.rotate(-theta);
							this.drawDecoratorInstance(ctx, decorator, theta);
							ctx.restore();
						}
						curvePosition = Math.round(arrivedPosition + decoratorLength);

						segment++;
					}

					currentOffset = Math.floor((curvePosition + 3 * decoratorLength - curveLength) % decoratorLength);
					previousPoint = p3;

					break;
				case 3: // ARC
					var radius = pointObject.v1 * zoom;
					var anticlockwise = pointObject.v5; // clockwise
					var from = previousPoint;
					var to = p;

					var distance = to.getDistance(from);
					var middle = new ShapeUtils.Point((to.x + from.x) / 2, (to.y + from.y) / 2); //middle between to and from
					var squareRoot = Math.sqrt(radius * radius - (distance / 2) * (distance / 2)); //tmp var for the calculation of the centre of the circle arc
					if (!anticlockwise) {
						//when not left, invert x
						squareRoot = -squareRoot;
					}
					//Calculate the center of the circle arc
					var center = new ShapeUtils.Point(
						middle.x - (squareRoot * (from.y - to.y)) / distance,
						middle.y - (squareRoot * (to.x - from.x)) / distance
					);
					var startCos = (from.x - center.x) / radius; //cosinus of the angle of the starting point of the circle arc
					var endCos = (to.x - center.x) / radius; //cosinus of the end angle of the circle arc
					startCos = Math.min(1, Math.max(-1, startCos));
					endCos = Math.min(1, Math.max(-1, endCos));
					var startAngle = Math.acos(startCos);
					var endAngle = Math.acos(endCos);
					if (from.y < center.y) {
						//make sure to choose the correct angle if in doubt (- or +)
						startAngle = -startAngle;
					}
					if (to.y < center.y) {
						endAngle = -endAngle;
					}

					if (anticlockwise) {
						if (startAngle < endAngle) {
							startAngle += 2 * Math.PI;
						}
					} else {
						// clock wise
						if (endAngle < startAngle) {
							endAngle += 2 * Math.PI;
						}
					}

					var fullAngle = endAngle - startAngle;
					if (anticlockwise) {
						fullAngle = startAngle - endAngle;
					}
					var length = Math.abs(fullAngle * radius);
					ctx.save();

					// FUTURE we actually need to find out if the browser connects to the previous path or not... happens sometimes (if there are not too much rounding errors I suppose)
					var currentPosition = currentOffset;
					var direction = 1;
					if (anticlockwise) {
						direction = -1;
					}
					var prev = {
						x: 0,
						y: 0
					};

					var segment = 0;
					while (currentPosition <= length && length > decoratorMarkSize) {
						var angle = startAngle + (direction * currentPosition) / radius;

						var point = {
							x: center.x + Math.cos(angle) * radius,
							y: center.y + Math.sin(angle) * radius
						};
						ctx.translate(point.x - prev.x, point.y - prev.y);

						if (anticlockwise) {
							angle += Math.PI;
						}
						this.drawDecoratorInstance(ctx, decorator, angle);
						// uncomment for some debugging info
						// ctx.rotate(angle);
						// ctx.fillText("#" + segment + "  ARC " + currentPosition, 0, 0);
						// ctx.rotate(-angle);
						segment++;
						prev = point;
						currentPosition += decoratorLength;
					}
					// arctTo doesnt connect nicely always...   so current offset might be decorator offset
					//currentOffset = decorator.offset;
					currentOffset = Math.floor(currentPosition + 3 * decoratorLength - length) % decoratorLength;

					ctx.restore();
					previousPoint = to;
					break;
				case 4: //NICK
					// IGNORE

					break;
				default:
					console.log("a line that is not supported: " + pointObject.type);
					break;
			}
		}
		ctx.restore();
	},
	/**
	 * Draw a decorator at (0,0) in a specified angle
	 * @param {CanvasRenderingContext2D} ctx - Canvas context}
	 * @param {Object} decorator - decorator description (see CadConstants)
	 * @param {Number} angle
	 */
	drawDecoratorInstance: function(ctx, decorator, angle) {
		if (angle) {
			ctx.rotate(angle);
		}
		ctx.beginPath();
		switch (decorator.name) {
			case "CROSS":
			case "CROSSGAP":
				ctx.lineWidth = 1;
				ctx.moveTo(-4, 1);
				ctx.lineTo(3, 7);
				ctx.moveTo(4, 1);
				ctx.lineTo(-3, 7);
				break;
			case "LINE":
				ctx.lineWidth = 1;
				ctx.moveTo(-4, 0);
				ctx.lineTo(4, 0);
				break;
			case "HEARTBEAT":
				ctx.lineWidth = 1;
				ctx.moveTo(0, 0);
				ctx.lineTo(-3, 2);
				ctx.lineTo(3, 4);
				ctx.lineTo(0, 7);
				break;
		}

		ctx.stroke();
		if (angle) {
			ctx.rotate(-angle);
		}
	},
	/**
	 * Draw an arc
	 * @param {CanvasRenderingContext2D} ctx - Canvas context
	 * @param {ShapeUtils.Point} from - connecting begin point
	 * @param {ShapeUtils.Point} to - connecting end point
	 * @param {Number} radius - radius of the circle
	 * @param {Boolean} bigArc - ?
	 * @param {Boolean} left - True if left, false otherwise
	 * @param {Number} angle - angle of the arc
	 */
	drawArcFrom: function(ctx, from, to, radius, bigArc, left, angle) {
		//Temp solution: when the 2 X values are the same (with accuracy of 2), just draw a line instead of an arch
		var xDiff = Math.abs(from.x - to.x).toFixed(2);
		if (xDiff == 0) {
			ctx.lineTo(to.x, to.y);
		} else {
			var pointDiff = new ShapeUtils.Point(to.x - from.x, to.y - from.y);
			var pointDiffSquare = pointDiff.x * pointDiff.x + pointDiff.y * pointDiff.y;
			var distance = Math.sqrt(pointDiffSquare); //distance between to and from
			var middle = new ShapeUtils.Point((to.x + from.x) / 2, (to.y + from.y) / 2); //middle between to and from
			var squareRoot = Math.sqrt(radius * radius - (distance / 2) * (distance / 2)); //tmp var for the calculation of the centre of the circle arc
			if (!left) {
				//when not left, invert x
				squareRoot = -squareRoot;
			}
			//Calculate the center of the circle arc
			var center = new ShapeUtils.Point(
				middle.x - (squareRoot * (from.y - to.y)) / distance,
				middle.y - (squareRoot * (to.x - from.x)) / distance
			);
			var startCos = (from.x - center.x) / radius; //cosinus of the angle of the starting point of the circle arc
			var endCos = (to.x - center.x) / radius; //cosinus of the end angle of the circle arc
			startCos = startCos > 1 ? 1 : startCos; //cosinus cannot be bigger than 1 or smaller than -1
			startCos = startCos < -1 ? -1 : startCos;
			endCos = endCos > 1 ? 1 : endCos;
			endCos = endCos < -1 ? -1 : endCos;
			var startAngle = Math.acos(startCos);
			var endAngle = Math.acos(endCos);
			if (from.y < center.y) {
				//make sure to choose the correct angle if in doubt (- or +)
				startAngle = -startAngle;
			}
			if (to.y < center.y) {
				endAngle = -endAngle;
			}

			ctx.lineTo(from.x, from.y); //make the move to the starting point
			ctx.arc(center.x, center.y, radius, startAngle, endAngle, left); //draw the arc
			ctx.lineTo(to.x, to.y); //make the move to the end point, to make sure we restart the path from the correct position
		}
	},

	/**
	 * Draw a dashed line between two points
	 * @param {CanvasRenderingContext2D} ctx - Canvas context
	 * @param {Number} x - x coordinate of the first point
	 * @param {Number} y - y coordinate of the first point
	 * @param {Number} x2 - x coorindate of the second point
	 * @param {Number} y2 - x coordinate of the second point
	 * @param {Array<Number>} dashArray - dash lenghts
	 */
	drawDash: DashUtils.drawDash,
	/**
	 * Draw an ellipse
	 * @param {CanvasRenderingContext2D} ctx - Canvas context
	 * @param {Number} x - x coordinate (top left of the ellipse)
	 * @param {Number} y - y coordinate (top left of the ellipse)
	 * @param {Number} w - width of the ellipse
	 * @param {Number} h - height of the ellipse
	 * @param {Boolean} fill - True to fill the ellipse
	 * @param {Boolean} stroke - True to stroke the ellipse
	 */
	drawEllipse: function(ctx, x, y, w, h, fill, stroke) {
		var kappa = 0.5522848;
		var ox = (w / 2) * kappa; // control point offset horizontal
		var oy = (h / 2) * kappa; // control point offset vertical
		var xe = x + w; // x-end
		var ye = y + h; // y-end
		var xm = x + w / 2; // x-middle
		var ym = y + h / 2; // y-middle

		ctx.beginPath();
		ctx.moveTo(x, ym);
		ctx.bezierCurveTo(x, ym - oy, xm - ox, y, xm, y);
		ctx.bezierCurveTo(xm + ox, y, xe, ym - oy, xe, ym);
		ctx.bezierCurveTo(xe, ym + oy, xm + ox, ye, xm, ye);
		ctx.bezierCurveTo(xm - ox, ye, x, ym + oy, x, ym);
		ctx.closePath();
		if (fill) {
			ctx.fill();
		}
		if (stroke) {
			ctx.stroke();
		}
	},
	/**
	 * Calculate the distance from a point to a segment with a max of maxDistance
	 * @param {ShapeUtils.Point} pnt - Point
	 * @param {ShapeUtils.Point} from - Begin point of segment
	 * @param {ShapeUtils.Point} to - End point of segment
	 * @param {Number} maxDistance - maximum distance between point and segment (will return maxDistance + 1 if exceeded)
	 * @returns {Number} the distance between the point and the segment with a max of maxDistance + 1
	 */
	distanceFromPointToSegment: function(pnt, from, to, maxDistance) {
		// Quick bounding box check
		if (from.x > to.x) {
			if (pnt.x < to.x - maxDistance) return maxDistance + 1;
			if (pnt.x > from.x + maxDistance) return maxDistance + 1;
		} else {
			if (pnt.x < from.x - maxDistance) return maxDistance + 1;
			if (pnt.x > to.x + maxDistance) return maxDistance + 1;
		}
		if (from.y > to.y) {
			if (pnt.y < to.y - maxDistance) return maxDistance + 1;
			if (pnt.y > from.y + maxDistance) return maxDistance + 1;
		} else {
			if (pnt.y < from.y - maxDistance) return maxDistance + 1;
			if (pnt.y > to.y + maxDistance) return maxDistance + 1;
		}

		// Distance point to infinite line
		var s = new ShapeUtils.Point();
		s.x = to.x - from.x;
		s.y = to.y - from.y;
		var p = new ShapeUtils.Point();
		p.x = pnt.x - from.x;
		p.y = pnt.y - from.y;
		if (s.x != 0 || s.y != 0) {
			var segmentLength2 = s.x * s.x + s.y * s.y;
			// Test position along segment
			var dotS = s.x * p.x + s.y * p.y;
			if (dotS > 0 && dotS < segmentLength2) {
				// We are inside the segment
				var dotN = s.y * p.x - s.x * p.y;
				if (dotN * dotN > maxDistance * maxDistance * segmentLength2) return maxDistance + 1; // too far away from infinite line
				return Math.abs(dotN / Math.sqrt(segmentLength2));
			}
		}
		var dFrom = p.x * p.x + p.y * p.y;
		if (dFrom > maxDistance * maxDistance) return maxDistance + 1;
		else return Math.sqrt(dFrom);
		var p2 = new ShapeUtils.Point();
		p2.x = point.x - to.x;
		p2.y = point.y - to.y;
		var dTo = p2.x * p2.x + p2.y * p2.y;
		if (dTo > maxDistance * maxDistance) return maxDistance + 1;
		else return sqrt(dTo);

		return maxDistance + 1; // just to avoid compiler warnings
	},
	/**
	 * Calculate the distance from a point to an ellipse
	 * @param {ShapeUtils.Point} p - point
	 * @param {ShapeUtils.Size} r - rectangle around the ellipse
	 * @param {Number} max - maximum distance (will return max + 1 if exceeded)
	 * @returns {Number} distance from point to an ellipse
	 */
	distanceFromPointToEllipse: function(p, r, max) {
		if (r.size.width == 0) return max + 1;
		if (r.size.height == 0) return max + 1;

		// Simple bounding box test
		if (p.x < r.origin.x - max) return max + 1;
		if (p.y < r.origin.y - max) return max + 1;
		if (p.x > r.origin.x + r.size.width + max) return max + 1;
		if (p.y > r.origin.y + r.size.height + max) return max + 1;

		var ctr = new ShapeUtils.Point(0, 0);
		ctr.x = r.origin.x + r.size.width / 2.0;
		ctr.y = r.origin.y + r.size.height / 2.0;

		// This is not mathematically correct, but accurate enough for this purpose
		var d = new ShapeUtils.Point();
		d.x = p.x - ctr.x;
		d.y = p.y - ctr.y;
		d.x *= 2.0 / r.size.width;
		d.y *= 2.0 / r.size.height;
		var length = Math.sqrt(d.x * d.x + d.y * d.y);
		if (length < 0.0001) return 0;
		d.x *= (r.size.width * (1 - length)) / (2.0 * length);
		d.y *= (r.size.height * (1 - length)) / (2.0 * length);
		length = d.x * d.x + d.y * d.y;
		if (length > max * max) return max + 1;
		return Math.sqrt(length);
	},
	/**
	 * Calculate distance between lineworkwrapper and point
	 * @param {LineworkWrapper} lineworkWrapper
	 * @param {ShapeUtils.Point} point
	 * @param {Transformation} transform
	 * @param {Number} [fatfinger = 6] - max distance
	 * @returns {Number} distance between lineworkwrapper and point
	 */
	distanceOfScenePoint: function(lineworkWrapper, point, transform, fatfinger) {
		if (!fatfinger) {
			fatfinger = 6;
		}
		var g = fatfinger;

		var dist = -1;

		var points = lineworkWrapper.pointArray;

		// check by points
		if (points && points.length != 0) {
			if (points.length == 1) {
				//only one point
				var lwPoint = transform.transformedPoint(points[0].point);
				var refPoint = new ShapeUtils.Point(lwPoint.x, lwPoint.y);
				if (Math.abs(refPoint.x - point.x) < g && Math.abs(refPoint.y - point.y) < g) {
					dist = 999;
				}
			} else {
				for (var j = 0; j < points.length - 1; j++) {
					var lwPoint = transform.transformedPoint(points[j]);
					var lwPoint2 = transform.transformedPoint(points[j + 1]);
					var refPoint = new ShapeUtils.Point(lwPoint.x, lwPoint.y);
					var refPoint2 = new ShapeUtils.Point(lwPoint2.x, lwPoint2.y);
					var refdist = this.distanceFromPointToSegment(point, refPoint, refPoint2, g);
					if (refdist <= g) {
						dist = refdist;
					}
				}
			}
		} else {
			// check by rect
			var r = transform.transformedRect(lineworkWrapper.rect);
			if (lineworkWrapper.staticsize) {
				//because it's a static size, use the actual size of the rectangle
				var center = r.center();
				var size = new ShapeUtils.Size(lineworkWrapper.rect.size.width, lineworkWrapper.rect.size.height);
				r = new ShapeUtils.Rect(new ShapeUtils.Point(center.x - size.width / 2, center.y - size.height / 2), size);
			}

			var outsideRect = new ShapeUtils.Rect(
				new ShapeUtils.Point(r.origin.x - g, r.origin.y - g),
				new ShapeUtils.Size(r.size.width + g * 2, r.size.height + g * 2)
			);

			r.normalize();
			outsideRect.normalize();

			var refdist;
			if (lineworkWrapper.type == this.LWTYPE_ELLIPSE) {
				if (lineworkWrapper.fillcolor) {
					if (r.containsPoint(point)) {
						dist = 1;
					} else if (outsideRect.containsPoint(point)) {
						dist = 999;
					}
				} else {
					refdist = this.distanceFromPointToEllipse(point, r, g);
					if (refdist <= g) {
						dist = refdist;
					}
				}
			} else {
				if (lineworkWrapper.type == this.LWTYPE_IMAGE && lineworkWrapper.imagetouchspot) {
					//image
					var touchSpot = lineworkWrapper.imagetouchspot;
					var tX = touchSpot.origin.x;
					var tY = touchSpot.origin.y;
					var tWidth = touchSpot.size.width;
					var tHeight = touchSpot.size.height;
					outsideRect = new ShapeUtils.Rect(
						new ShapeUtils.Point(outsideRect.origin.x + tX, outsideRect.origin.y + tY),
						new ShapeUtils.Size(tWidth + g * 2, tHeight + g * 2)
					);
				}

				if (outsideRect.containsPoint(point)) {
					if ((lineworkWrapper.fillcolor || lineworkWrapper.type == this.LWTYPE_IMAGE) && !lineworkWrapper.treatAsOutline) {
						dist = 999;
					} else {
						var insideRect = new ShapeUtils.Rect(
							new ShapeUtils.Point(r.origin.x + g, r.origin.y + g),
							new ShapeUtils.Size(r.size.width - g * 2, r.size.height - g * 2)
						);
						if (!insideRect.containsPoint(point)) {
							//rectangle
							dist = 999;
						}
					}
				}
			}
		}
		return dist;
	},
	/**
	 * Draw a dashed line
	 * @param {Context2D} ctx - canvas context
	 * @param {Number} x - start x
	 * @param {Number} y - start y
	 * @param {Number} x2 - end x
	 * @param {Number} y2 - end y
	 * @param {Array.<Number>} dashArray - dash pattern [px draw, px empty, px draw, px empty, ...]
	 */
	drawDashedLine: DashUtils.drawDashedLine,
	/**
	 * Draw a dashed rectangle
	 * @param {Context2D} ctx - canvas context
	 * @param {ShapeUtils.Rect} rect - rectangle to draw
	 * @param {Array.<Number>} dashArray - dash pattern
	 * @param {Number} [zoom] - which zoomlevel is used (use this if you want to define the dasharray in pixels and not world coordinates)
	 */
	drawDashedRect: DashUtils.drawDashedRect
};
export default DrawUtils;
