import _ from "underscore";
import $ from "jquery";

// empty image
var emptyImage = new Image();

/**
 * @exports Fragment
 * @memberof Utils
 * @class
 * A fragment is a dynamically loaded image
 * @constructor
 * @param {ShapeUtils.Rect} viewrect - rectangle of the fragment
 * @param {Number} z - zoom level of the fragment
 */
var Fragment = function(viewrect, z) {
	this.rect = viewrect;
	this.zoom = z;
	this.image = new Image();

	this.fetched = false;
	this.dirty = false;
};

/**
 * Start fetching the fragment if needed
 * @param {String} url - the viewserver url of the image
 * @param {Function} onfinish - callback when the image was loaded successfully
 * @param {Function} onerror - callback when there was an error while loading the image
 */
Fragment.prototype.fetchImage = function(fragmentCache, url, onfinish, onerror) {
	//temporarily put in an empty image
	this.image = emptyImage;

	// fetch the image from the cache, if already in cache, callbacks will be called immediatly
	fragmentCache.get(
		url,
		_.bind(function(img) {
			if (img === undefined || img.error) {
				// something went wrong while loading image, remove it from the cache
				onerror.call(undefined);
			} else {
				// everything went ok, set the fragment as fetched and execute callback
				this.image = img;
				this.fetched = true;
				onfinish.call(img);
			}
		}, this)
	);
};

/**
 * get a reference to the image object
 * @returns {Image}
 */
Fragment.prototype.getImage = function() {
	return this.image;
};

/**
 * Set the image for the fragment
 * @param {Image} image
 */
Fragment.prototype.setImage = function(image) {
	this.image = $("<img>").attr("src", image)[0];
	this.fetched = true;
};

/**
 * @returns {Boolean} is the fragment fetched already?
 */
Fragment.prototype.isFetched = function() {
	return this.fetched;
};

/**
 * @returns {Boolean} is the fragment dirty? (should it be thrown away if there are new fragments)
 */
Fragment.prototype.isDirty = function() {
	return this.dirty;
};

/**
 * Mark the fragment as dirty
 */
Fragment.prototype.makeDirty = function() {
	this.dirty = true;
};

/**
 * Get the rectangle of the fragment
 * @returns {ShapeUtils.Rect}
 */
Fragment.prototype.getRect = function() {
	return this.rect;
};

/**
 * Get the zoom level of the fragment
 * @returns {Number}
 */
Fragment.prototype.getZoom = function() {
	return this.zoom;
};

/**
 * Get the real width (in pixels) of the image
 * @returns {Number} width in pixels
 */
Fragment.prototype.getNaturalWidth = function() {
	return this.image.naturalWidth;
};

/**
 * Get the real height (in pixels) of the image
 * @returns {Number} height in pixels
 */
Fragment.prototype.getNaturalHeight = function() {
	return this.image.naturalHeight;
};

export default Fragment;
