import _ from "underscore";
import ShapeUtils from "@esko/shapeutils";
import Extendable from "@esko/web-viewer/core/Extendable";
import DrawUtils from "@esko/web-viewer/utils/DrawUtils";
import LineworkPoint from "@esko/web-viewer/utils/LineworkPoint";
/**
 * @memberof Utils
 * @exports LineworkWrapper
 * @class
 * A lineworkwrapper is used to draw linework
 * @constructor
 */
var LineworkWrapper = function() {
	Extendable.installTo(this);

	this.color = null;
	this.c = null;
	this.fillcolor = null;
	this.pointArray = [];
	this.rect = new ShapeUtils.Rect();
	this.type = 0;
	this.linewidth = 1;
	this.imagehotspot = null;
	this.imagetouchspot = null;
	this.imageRef = null;

	this.treatAsOutline = false; //treat the object as an outline object for selecting

	this.textoptions = {};
	this.selected = false;

	this.showIndicator = false;
	this.indicatorText = null;
	this.indicatorTextColor = null;
	this.indicatorColor = null;

	this.snapable = false; //if snapable, the user can snap to the points of the wrapper
};

/**
 * Move the lineworkwrapper over a size
 * @param {ShapeUtils.Size} size
 */
LineworkWrapper.prototype.move = function(size) {
	this.rect.origin.x += size.width;
	this.rect.origin.y += size.height;
	if (this.pointArray && this.pointArray.length > 0) {
		for (var j = 0; j < this.pointArray.length; j++) {
			var point = this.pointArray[j];
			point.x += size.width;
			point.y += size.height;
		}
	}
};

/**
 * Sets the rectangle of the lineworkwrapper
 * @param {ShapeUtils.Rect} rect
 * @param {Boolean} updateStartEnd
 */
LineworkWrapper.prototype.setRect = function(rect, updateStartEnd) {
	if (updateStartEnd) {
		if (this.pointArray && this.pointArray.length == 2) {
			var p1 = rect.topleft();
			var p2 = rect.bottomright();

			if (
				(this.pointArray[0].x == p1.x && this.pointArray[0].y == p1.y) ||
				(this.pointArray[0].x == this.pointArray[1].x && this.pointArray[0].y == this.pointArray[1].y)
			) {
				this.pointArray[0].x = p1.x;
				this.pointArray[0].y = p1.y;

				this.pointArray[1].x = p2.x;
				this.pointArray[1].y = p2.y;
			} else {
				this.pointArray[1].x = p1.x;
				this.pointArray[1].y = p1.y;

				this.pointArray[0].x = p2.x;
				this.pointArray[0].y = p2.y;
			}
		}
	}
	this.rect = rect;
};

/**
 * Set the start and the end point for the lineworkwrapper
 * @param {ShapeUtils.Point} start
 * @param {ShapeUtils.Point} end
 */
LineworkWrapper.prototype.setStartEnd = function(start, end) {
	if (this.pointArray && this.pointArray.length == 2) {
		if (start) {
			this.pointArray[0].x = start.x;
			this.pointArray[0].y = start.y;
		} else {
			start = new ShapeUtils.Point(this.pointArray[0].x, this.pointArray[0].y);
		}
		if (end) {
			this.pointArray[1].x = end.x;
			this.pointArray[1].y = end.y;
		} else {
			end = new ShapeUtils.Point(this.pointArray[1].x, this.pointArray[1].y);
		}

		this.setRect(new ShapeUtils.Rect(start, start.sizeTo(end)));
	}
};

/**
 * Get the rgba color of the line work with alpha
 * @param {Number} alpha
 * @returns {String}
 */
LineworkWrapper.prototype.rgbaColor = function(alpha) {
	var rgb = this.rgbColor();
	return "rgba(" + rgb.r + "," + rgb.g + "," + rgb.b + "," + alpha + ")";
};

/**
 * Get the rgb color
 * @returns {Object}
 */
LineworkWrapper.prototype.rgbColor = function() {
	var result = /^#?([a-f\d]{2})([a-f\d]{2})([a-f\d]{2})$/i.exec(this.color);
	return result
		? {
				r: parseInt(result[1], 16),
				g: parseInt(result[2], 16),
				b: parseInt(result[3], 16)
		  }
		: null;
};

/**
 * Calculate the rectangle from all the points of the point array of the linework
 */
LineworkWrapper.prototype.calculateRectFromPointArray = function() {
	var rectx1 = null;
	var rectx2 = null;
	var recty1 = null;
	var recty2 = null;

	for (var i = 0; i < this.pointArray.length; i++) {
		var p = this.pointArray[i];

		if (!rectx1 || p.x < rectx1) {
			rectx1 = p.x;
		}
		if (!rectx2 || p.x > rectx2) {
			rectx2 = p.x;
		}
		if (!recty1 || p.y < recty1) {
			recty1 = p.y;
		}
		if (!recty2 || p.y > recty2) {
			recty2 = p.y;
		}
	}
	var r = new ShapeUtils.Rect(new ShapeUtils.Point(rectx1, recty1), new ShapeUtils.Size(rectx2 - rectx1, recty2 - recty1));
	r.normalize();
	this.rect = r;
};

/**
 * Set the selection anchor point
 * @param {ShapeUtils.Point} point
 * @param fillColor
 * @param strokeColor
 * @param subtype
 * @returns {LineworkWrapper}
 */
LineworkWrapper.prototype.selectionAnchorPoint = function(point, fillColor, strokeColor, subtype) {
	var anchor = new LineworkWrapper();

	anchor.type = DrawUtils.LWTYPE_ELLIPSE;
	anchor.color = strokeColor;
	anchor.fillcolor = fillColor;
	anchor.staticsize = true;
	anchor.linewidth = 1;
	var size = 10;
	anchor.rect = new ShapeUtils.Rect(new ShapeUtils.Point(point.x - size / 2, point.y - size / 2), new ShapeUtils.Size(size, size));
	anchor.anchor = subtype;
	anchor.linework = this;

	return anchor;
};

/**
 * Get the selection box of the linework wrapper
 * @returns {Array.<LineworkWrapper>}
 */
LineworkWrapper.prototype.getSelectionBox = function() {
	var selectionFillColor = "rgb(153,152,187)";
	var selectionStrokeColor = "white";

	if (this.pointArray && this.pointArray.length >= 2) {
		return [
			this.selectionAnchorPoint(
				new ShapeUtils.Point(this.pointArray[0].x, this.pointArray[0].y),
				selectionFillColor,
				selectionStrokeColor,
				DrawUtils.LWANCHOR_START
			),
			this.selectionAnchorPoint(
				new ShapeUtils.Point(this.pointArray[this.pointArray.length - 1].x, this.pointArray[this.pointArray.length - 1].y),
				selectionFillColor,
				selectionStrokeColor,
				DrawUtils.LWANCHOR_END
			)
		];
	} else {
		var anchors = [
			this.selectionAnchorPoint(this.rect.topleft(), selectionFillColor, selectionStrokeColor, DrawUtils.LWANCHOR_TOPLEFT),
			this.selectionAnchorPoint(this.rect.topright(), selectionFillColor, selectionStrokeColor, DrawUtils.LWANCHOR_TOPRIGHT),
			this.selectionAnchorPoint(this.rect.bottomleft(), selectionFillColor, selectionStrokeColor, DrawUtils.LWANCHOR_BOTTOMLEFT),
			this.selectionAnchorPoint(this.rect.bottomright(), selectionFillColor, selectionStrokeColor, DrawUtils.LWANCHOR_BOTTOMRIGHT),
			this.selectionAnchorPoint(this.rect.topcenter(), selectionFillColor, selectionStrokeColor, DrawUtils.LWANCHOR_TOPCENTER),
			this.selectionAnchorPoint(this.rect.bottomcenter(), selectionFillColor, selectionStrokeColor, DrawUtils.LWANCHOR_BOTTOMCENTER),
			this.selectionAnchorPoint(this.rect.centerleft(), selectionFillColor, selectionStrokeColor, DrawUtils.LWANCHOR_CENTERLEFT),
			this.selectionAnchorPoint(this.rect.centerright(), selectionFillColor, selectionStrokeColor, DrawUtils.LWANCHOR_CENTERRIGHT)
		];

		return _.union(anchors);
	}
};

/**
 * Scale the linework wrapper with a given factor (both horizontal as vertical)
 * @param {ShapeUtils.Point} scaleFactor
 */
LineworkWrapper.prototype.scale = function(scaleFactor) {
	if (scaleFactor) {
		_.each(this.pointArray, function(point) {
			point.x *= scaleFactor.x;
			point.y *= scaleFactor.y;
		});

		this.rect.origin.x *= scaleFactor.x;
		this.rect.origin.y *= scaleFactor.y;
		this.rect.size.width *= scaleFactor.x;
		this.rect.size.height *= scaleFactor.y;
	}
};

export default LineworkWrapper;
