import _ from "underscore";
/**
 * @memberof Utils
 * @exports ListenerUtil
 * @class ListenerUtil
 */
var ListenerUtil = {
	/**
	 * Install ListenerUtilMixin in to obj
	 * @param {Any} obj - object which should be extended
	 * @param {Core.ViewerFacade} f
	 */
	mixin: function(obj, f) {
		/**
		 * This mixin provides utility methods for listener registration
		 * @memberof Utils
		 * @exports ListenerUtilMixin
		 * @mixin
		 */
		var ListenerUtilMixin = {
			/**
			 * Register a listener to an object (or the default event bus)
			 * @param {String|Array.<String>} channel - the channel on which the listener should be installerd
			 * @param {Function} callback - the callback function for the listener
			 * @param {Any} [binding] - if provided, the callback function will be bound to this object
			 * @param {Any} object - if provided, the listener is registered on this object
			 * @returns {ListenerObject} this object can be used for the _unregisterListener
			 */
			_registerListener: function(channel, callback, binding, object) {
				var finalCallback = null;
				// bind the callback
				if (binding) {
					finalCallback = _.bind(callback, binding);
				} else {
					finalCallback = callback;
				}

				// define on which object the listener should be registered
				if (object === undefined) {
					object = f;
				}

				var lObj = {
					object: object,
					channel: channel,
					callback: finalCallback
				};

				// register the listener
				lObj.object.on(lObj.channel, lObj.callback);

				this._registeredListeners.push(lObj);

				return lObj;
			},
			/**
			 * Unregister a certain listener
			 * @param {ListenerObject} listenerObject - the object that was returned in _registerListener
			 * @param {Boolean} [_removeListener} - should the listener be removed from registered list (only used internally!)
			 */
			_unregisterListener: function(listenerObject, _removeListener) {
				listenerObject.object.off(listenerObject.channel, listenerObject.callback);

				if (_removeListener === undefined || _removeListener === true) {
					this._registeredListeners = _.without(this._registeredListeners, listenerObject);
				}
			},

			/**
			 * Unregister all previously registered listeners
			 */
			_unregisterAllListeners: function() {
				_.each(
					this._registeredListeners,
					function(listenerObject) {
						this._unregisterListener(listenerObject, false);
					},
					this
				);
				this._registeredListeners = [];
			},

			/**
			 * Unregister all listeners registered on a certain object
			 * @param {Any} object - object on which the listeners were registered
			 */
			_unregisterAllListenersForObject: function(object) {
				var removed = [];
				_.each(
					this._registeredListeners,
					function(listenerObj) {
						if (listenerObj.object === object) {
							this._unregisterListener(listenerObj, false);
							removed.push(listenerObj);
						}
					},
					this
				);
				this._registeredListeners = _.difference(this._registeredListeners, removed);
			},

			/**
			 * @private
			 */
			_registeredListeners: []
		};

		// install the mixing methods
		_.each(ListenerUtilMixin, function(method, name) {
			obj[name] = method;
		});
	}
};

export default ListenerUtil;
