import _ from "underscore";
import Hammer from "hammerjs";
import ShapeUtils from "@esko/shapeutils";
/**
 * Handle touch events with the use of HammerJS
 * @memberof Utils
 * @exports TouchUtil
 * @class TouchUtil
 * @param {HTMLElement} ctx - context in which the touch events are registered
 */
var TouchUtil = function(ctx) {
	this.ctx = ctx;
	// Initialize HammeJS and enable pinch, double tap and pan
	var touchMgr = new Hammer.Manager(ctx, { preventDefault: true });
	var pinch = new Hammer.Pinch();
	var singleTap = new Hammer.Tap({
		event: "singletap"
	});
	var doubleTap = new Hammer.Tap({
		event: "doubletap",
		taps: 2
	});
	var pan = new Hammer.Pan({ direction: Hammer.DIRECTION_ALL });
	touchMgr.add([pinch, singleTap, doubleTap, pan]);
	pinch.recognizeWith(pan);
	pinch.requireFailure(pan);
	doubleTap.recognizeWith(singleTap);

	// Keep track of all listeners
	this.listeners = {};

	// register events HammerJS should listen to
	touchMgr.on("singletap doubletap pinchstart pinchmove pinchend panstart panmove panend pancancel", _.bind(this.handleTouchEvent, this));
};

TouchUtil.prototype.handleTouchEvent = function(event) {
	var type = event.type;
	if (this.listeners[type] && this.listeners[type].length > 0) {
		_.each(this.listeners[type], function(callback) {
			callback(event);
		});
	}
};

// translate adding of listeners to the HammerJS library events
TouchUtil.prototype.addSingleTapListener = function(callback) {
	this._addListener("singletap", callback);
};

TouchUtil.prototype.addDoubleTapListener = function(callback) {
	this._addListener("doubletap", callback);
};

TouchUtil.prototype.addPinchStartListener = function(callback) {
	this._addListener("pinchstart", callback);
};

TouchUtil.prototype.addPinchMoveListener = function(callback) {
	this._addListener("pinchmove", callback);
};

TouchUtil.prototype.addPinchEndListener = function(callback) {
	this._addListener("pinchend", callback);
	this._addListener("pinchcancel", callback);
};

TouchUtil.prototype.addPanStartListener = function(callback) {
	this._addListener("panstart", callback);
};

TouchUtil.prototype.addPanMoveListener = function(callback) {
	this._addListener("panmove", callback);
};

TouchUtil.prototype.addPanEndListener = function(callback) {
	this._addListener("panend", callback);
	this._addListener("pancancel", callback);
};

// translate removing of listeners to the HammerJS library events
TouchUtil.prototype.removeDoubleTapListener = function(callback) {
	this._removeListener("doubletap", callback);
};

TouchUtil.prototype.removePinchStartListener = function(callback) {
	this._removeListener("pinchstart", callback);
};

TouchUtil.prototype.removePinchMoveListener = function(callback) {
	this._removeListener("pinchmove", callback);
};

TouchUtil.prototype.removePinchEndListener = function(callback) {
	this._removeListener("pinchcancel", callback);
	this._removeListener("pinchend", callback);
};

TouchUtil.prototype.removePanStartListener = function(callback) {
	this._removeListener("panstart", callback);
};

TouchUtil.prototype.removePanMoveListener = function(callback) {
	this._removeListener("panmove", callback);
};

TouchUtil.prototype.removePanEndListener = function(callback) {
	this._removeListener("panend", callback);
	this._removeListener("pancancel", callback);
};

/**
 * Remove all listeners
 */
TouchUtil.prototype.removeAllListeners = function() {
	this.listeners = {};
};

/**
 * Add listener of a specific event type
 * @param {String} Type - type of event, e.g. pinchstart, pinchmove
 * @param {Function} callback - method that will be called when event is triggered
 */
TouchUtil.prototype._addListener = function(type, callback) {
	if (!this.listeners[type]) {
		this.listeners[type] = [];
	}
	if (_.indexOf(this.listeners[type], callback) < 0) {
		this.listeners[type].push(callback);
	}
};

/**
 * Remove specific listener for a type of event
 * @param {String} Type - type of event, e.g. pinchstart, pinchmove
 * @param {Function} callback - method that will be removed from listener
 */
TouchUtil.prototype._removeListener = function(type, callback) {
	if (this.listeners[type]) {
		this.listeners[type] = _.without(this.listeners[type], callback);
	}
};

/**
 * Data Getters and Event handler actions
 * We place these here, because we don't want the TouchTool to know anything about the object obtained by HammerJS
 */
TouchUtil.prototype.preventDefault = function(e) {
	e.preventDefault();
};
TouchUtil.prototype.getScale = function(e) {
	return e.scale;
};
TouchUtil.prototype.getRotation = function(e) {
	return e.rotation;
};

/**
 * Get the Center of a Touch Event
 *
 * Note! the doesn't take in to account multiple views in the widget... It is safer to use getScenePointFromTouch
 *
 * @param {TouchEvent} event - the touch event
 * @param {Object} offset - {top: topOffset, left: leftOffset{
 * @returns {ShapeUtils.Point} center point
 */
TouchUtil.prototype.getCenter = function(e, offset) {
	var center = new ShapeUtils.Point();
	center.x = e.center.x;
	center.y = e.center.y;
	if (offset) {
		center.x -= offset.left;
		center.y -= offset.top;
	}
	return center;
};

/**
 * Get the center scene point of a TouchEvent
 * @param {TouchEvent} event - the touch event
 * @param {ViewportWidget} widget - the ViewportWidget
 * @param {ViewportView} view - the view in which the ScenePoint should be determined
 * @returns {Object} {view: reference to the View, point: scenePoint}
 */
TouchUtil.prototype.getScenePointFromTouch = function(event, widget, view) {
	var tapPoint = this.getCenter(event);
	if (!view) {
		view = widget.requestView(tapPoint);
	}

	if (view) {
		var offset = view.getViewContext().offset();
		tapPoint.offsetXY(-offset.left, -offset.top);
	}

	return {
		view: view,
		point: tapPoint
	};
};

export default TouchUtil;
