import ShapeUtils from "@esko/shapeutils";
import GEOUtils from "@esko/web-viewer/utils/GEOUtils";
import CoreConstants from "@esko/web-viewer/core/CoreConstants";
import ConversionConstants from "@esko/web-viewer/utils/ConversionConstants";
import _ from "underscore";
import $ from "jquery";
var ViewerUtils = {
	/**
	 * Convert rgb to hex
	 * @param {Number} R - red value in RGB [0-255]
	 * @param {Number} G - green value in RGB [0-255]
	 * @param {Number} B - blue value in RGB [0-255]
	 * @returns {String} rgb in hex format
	 */
	rgbToHex: function(R, G, B) {
		return this.toHex(R) + this.toHex(G) + this.toHex(B);
	},
	/**
	 * Convert a number to hex
	 * @param {Number} n - number to convert to hex format
	 * @returns {String} n in hex format
	 */
	toHex: function(n) {
		n = parseInt(n, 10);
		if (isNaN(n)) {
			return "00";
		}
		n = Math.max(0, Math.min(n, 255));
		return "0123456789ABCDEF".charAt((n - (n % 16)) / 16) + "0123456789ABCDEF".charAt(n % 16);
	},
	/**
	 * Convert a number encoded RGB color to hex format
	 * @param {Number} color - Number encoded color to convert to hex format
	 * @returns {String} hex format of the number encoded RGB color
	 */
	hexFromInt: function(color) {
		var r = (color >> 16) & 0xff;
		var g = (color >> 8) & 0xff;
		var b = (color >> 0) & 0xff;
		return this.rgbToHex(r, g, b);
	},
	/**
	 * Convert a hex RGB color to a number encoded RGB color
	 * @param {String} hex - Hex encoded RGB color
	 * @returns {Number} number encoded RGB color of the given hex RGB color
	 */
	intFromHex: function(hex) {
		var rgb = this.rgbColor(hex);
		return (
			((rgb.r & 0xff) << 16) | //red
			((rgb.g & 0xff) << 8) | //green
			((rgb.b & 0xff) << 0)
		); //blue
	},
	/**
	 * Convert an RGB number encoded color to BGR number encoded color
	 * @param {Number} color - RGB number encoded color
	 * @returns {Number} BGR number encoded color from the RGB number encoded color
	 */
	rgbIntToBgrInt: function(color) {
		var r = this.toHex((color >> 16) & 0xff);
		var g = this.toHex((color >> 8) & 0xff);
		var b = this.toHex((color >> 0) & 0xff);
		return parseInt(b + g + r, 16);
	},
	/**
	 * Convert an RGB hex color String to an object
	 * @param {String} string - hex formatted color string
	 * @returns {Object} o.r, o.g, o.b contains respectively the R, G and B value. Returns null if parsing failed.
	 */
	rgbColor: function(string) {
		var result = /^#?([a-f\d]{2})([a-f\d]{2})([a-f\d]{2})$/i.exec(string);
		return result
			? {
					r: parseInt(result[1], 16),
					g: parseInt(result[2], 16),
					b: parseInt(result[3], 16)
			  }
			: null;
	},
	/**
	 * Trim a String
	 * @param {String} s - String to trim
	 * @returns {String} trimmed string
	 */
	trim: function(s) {
		s = s.replace(/(^\s*)|(\s*$)/gi, "");
		s = s.replace(/[ ]{2,}/gi, " ");
		s = s.replace(/\n /, "\n");
		return s;
	},
	/**
	 * Convert value in mm to another unit with a fixed number of digits
	 * @param {Number} valueInMM - Value to convert in mm
	 * @param {CoreConstants.SizeUnits} units - Unit to convert to (0 for mm, 1 for cm, 2 for inches).
	 * @param {Number} [fix] - Number of digits
	 * @param {Number} [resolution = 72] - resolution for pixel conversion. Defaults to 72 ppi.
	 * @returns {String} Converted value of valueInMM to units with fix number of digits
	 */
	convertFromMmValue: function(valueInMM, units, fix, resolution) {
		var value = valueInMM;
		if (units == CoreConstants.SizeUnits.INCH) {
			//inch
			value = valueInMM / ConversionConstants.MM_TO_INCH;
		} else if (units == CoreConstants.SizeUnits.CM) {
			//cm
			value = valueInMM / ConversionConstants.MM_TO_CM;
		} else if (units == CoreConstants.SizeUnits.PIXELS || units == CoreConstants.SizeUnits.POINTS) {
			if (!isNaN(resolution)) {
				value = valueInMM * resolution;
			} else {
				value = valueInMM * ConversionConstants.MM_TO_POINT_RATIO;
			}
		}
		if (fix !== undefined) {
			return value.toFixed(fix);
		} else {
			return value.toString();
		}
	},
	/**
	 * Convert value in mm to another unit with a fixed number of digits
	 * @param {Number} valueInMM - Value to convert in mm
	 * @param {CoreConstants.ResolutionUnits} units - Unit to convert from (0 for mm, 1 for cm, 2 for inches).
	 * @param {Number} [fix] - Number of digits
	 * @param {Number} [resolution = 72] - resolution for pixel conversion. Defaults to 72 ppi.
	 * @returns {String} Converted value of valueInMM to units with fix number of digits
	 */
	convertToMmValue: function(valueInUserUnit, units, fix, resolution) {
		var value = valueInUserUnit;
		if (units == CoreConstants.SizeUnits.INCH) {
			//inch
			value = valueInUserUnit * ConversionConstants.MM_TO_INCH;
		} else if (units == CoreConstants.SizeUnits.CM) {
			//cm
			value = valueInUserUnit * ConversionConstants.MM_TO_CM;
		} else if (units == CoreConstants.SizeUnits.PIXELS || units == CoreConstants.SizeUnits.POINTS) {
			if (!isNaN(resolution)) {
				value = valueInUserUnit / resolution;
			} else {
				value = valueInUserUnit * ConversionConstants.POINTS_TO_MM_RATIO;
			}
		}
		if (fix !== undefined) {
			return value.toFixed(fix);
		} else {
			return value.toString();
		}
	},
	/**
	 * Convert from inches to mm
	 * @param {Number} valueInInch - Value to convert in inches
	 * @param {Number} fix - Number of digits
	 * @returns {String} Converted value of valueInInch to mm with fix number of digits
	 */
	convertFromInchToMm: function(valueInInch, fix) {
		var value = valueInInch;
		return (value * ConversionConstants.MM_TO_INCH).toFixed(fix);
	},
	/**
	 * Convert resolution value in ppmm to another unit with a fixed number of digits
	 * @param {Number} value - Value to convert in ppmm
	 * @param {CoreConstants.ResolutionUnits} toMode - Unit to convert to (0 for ppmm, 1 for ppcm, 2 for ppi).
	 * @param {Number} [fix] - Number of digits
	 * @returns {String} Converted value in ppmm to units with fixed number of digits
	 */
	convertResolutionFromPpmmValue: function(value, toMode, fix) {
		if (toMode === CoreConstants.ResolutionUnits.PPMM) {
			value = value * 1; // safety for when we get numbers as strings
		} else if (toMode === CoreConstants.ResolutionUnits.PPCM) {
			value = value * ConversionConstants.MM_TO_CM;
		} else if (toMode === CoreConstants.ResolutionUnits.PPI) {
			value = value * ConversionConstants.MM_TO_INCH;
		}
		if (fix !== undefined) {
			return value.toFixed(fix);
		} else {
			return value.toString();
		}
	},
	/**
	 * Convert resolution value of another unit to ppmm with a fixed number of digits
	 * @param {Number} value - Value to convert to ppmm
	 * @param {CoreConstants.ResolutionUnits} fromMode - Unit to convert from (0 for ppmm, 1 for ppcm, 2 for ppi).
	 * @param {Number} [fix] - Number of digits
	 * @returns {String} Converted value to ppmm with fixed number of digits
	 */
	convertResolutionToPpmmValue: function(value, fromMode, fix) {
		if (fromMode === CoreConstants.ResolutionUnits.PPMM) {
			value = value / 1; // safety for when we get numbers as strings
		} else if (fromMode === CoreConstants.ResolutionUnits.PPCM) {
			value = value / ConversionConstants.MM_TO_CM;
		} else if (fromMode === CoreConstants.ResolutionUnits.PPI) {
			value = value / ConversionConstants.MM_TO_INCH;
		}
		if (fix !== undefined) {
			return value.toFixed(fix);
		} else {
			return value.toString();
		}
	},
	/**
	 * Find the closest angle of a section of the given angle if you divide 360 in numSections
	 * @param {Number} angle - Angle in degrees of which you want to lookup the section
	 * @param {Number} numSections - Number of sections to divide 360 in
	 * @returns {Number} Closest angle of a section of the given angle if you divide 360 in numSections
	 */
	closestAngle: function(angle, numSections) {
		var closest = (Math.round((angle / 360) * numSections) * 360) / numSections;
		if (closest >= 360) {
			closest -= 360;
		}
		return closest;
	},
	/**
	 * Get the point on a circle distance
	 * @param {Number} radius - radius
	 * @param {Number} angle - angle in degrees
	 * @param {ShapeUtils.Point} center - center of the circle
	 * @returns {ShapeUtils.Point} point on a circle distance
	 */
	pointAtLength: function(radius, angle, center) {
		var a = (angle * Math.PI) / 180;

		var x = center.x + radius * Math.cos(a);
		var y = center.y + radius * Math.sin(a);

		return new ShapeUtils.Point(x, y);
	},

	/**
	 * Check if the string starts with a certain other string
	 * @param {String} string - String to check
	 * @param {String} startString - String string should start with
	 * @returns {Boolean} True if string starts with startString, false if not
	 */
	startsWith: function(string, startString) {
		return string.indexOf(startString) === 0;
	},
	/**
	 * Check if the string ends with a certain other string
	 * @param {String} string    - The string to check
	 * @param {String} endString - String string should end with
	 * @returns {Boolean} True if string ends with endString, false if not
	 */
	endsWith: function(string, endString) {
		var position = string.length - endString.length;
		var lastIndex = string.indexOf(endString, position);
		return lastIndex !== -1 && lastIndex === position;
	},
	/**
	 * Generate a unique DOM ID with the given prefix
	 * e.g. "input_45610" if prefix was "input_"
	 * @param {String} prefix - Prefix of the unique DOM ID
	 * @returns {String} unique DOM ID with the given prefix
	 */
	generateUniqueDOMId: function generateUniqueDOMId(prefix) {
		var nr = Math.floor(Math.random() * 10000);
		var id = prefix + nr;
		while (document.getElementById(id)) {
			return generateUniqueDOMId(prefix);
		}
		return id;
	},

	/**
	 * Convert the  date string to an actual date object
	 * @param {String} date - XFDF date string
	 * @returns {Date} date object of the xfdf date string
	 */
	dateFromString: function(date) {
		//D:20121016171755+02'00'
		var year = date.substring(2, 6);
		var month = date.substring(6, 8);
		var day = date.substring(8, 10);
		var hour = date.substring(10, 12);
		var minute = date.substring(12, 14);
		var second = date.substring(14, 16);
		var offsethour = date.substring(16, 19);
		var offsetminute = date.substring(20, 22);

		//"2012-01-17T12:55:00.000+01:00"
		return new Date(year + "-" + month + "-" + day + "T" + hour + ":" + minute + ":" + second + ".000" + offsethour + ":" + offsetminute);
	},
	/**
	 * < n will be converted to 0n
	 * @param {Number} n - Number to convert
	 * @returns formatted number to two digits
	 */
	twoDigits: function(n) {
		return n > 9 ? "" + n : "0" + n;
	},
	/**
	 * Get a XFDF date String from a Date object
	 * @param {Date} date - date object to convert
	 * @returns {String} XFDF date String from the given date object
	 */
	stringFromDate: function(date) {
		//D:20121016171755+02'00'

		var offset = date.getTimezoneOffset();
		var timezoneString = offset <= 0 ? "+" : "-";

		offset = Math.abs(offset);
		var timezonehours = Math.floor(offset / 60);
		var timezoneminutes = offset - timezonehours * 60;
		timezoneString += this.twoDigits(timezonehours) + "'" + this.twoDigits(timezoneminutes) + "'";

		return (
			"D:" +
			date.getUTCFullYear() +
			this.twoDigits(date.getUTCMonth() + 1) +
			this.twoDigits(date.getUTCDate()) +
			this.twoDigits(date.getHours()) +
			this.twoDigits(date.getMinutes()) +
			this.twoDigits(date.getUTCSeconds()) +
			timezoneString
		);
	},
	/**
	 * Generate a random string of length characters in [0-9a-zA-Z]
	 * @param {Number} length - Number of characters
	 * @returns {String} random string of length characters in [0-9a-zA-Z]
	 */
	randomString: function(length) {
		var chars = "0123456789abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ";
		var result = "";
		for (var i = length; i > 0; --i) {
			result += chars[Math.round(Math.random() * (chars.length - 1))];
		}
		return result;
	},
	/**
	 * Convert a ShapeUtils.Rect to a String "left, top, right, bottom" in PS points
	 * @param {ShapeUtils.Rect} rect - Rect (in mm) to convert to String
	 * @returns {String} a String "left, top, right, bottom" in PS points
	 */
	rectToString: function(rect) {
		var x1 = GEOUtils.convertMMToPSPoints(rect.origin.x);
		var y1 = GEOUtils.convertMMToPSPoints(rect.origin.y);
		var x2 = GEOUtils.convertMMToPSPoints(rect.right());
		var y2 = GEOUtils.convertMMToPSPoints(rect.bottom());

		return x1 + "," + y1 + "," + x2 + "," + y2;
	},

	/**
	 * Parse a float regardless language settings
	 * @param {String | Number} string - String to parse
	 * @returns {Number} the original number it was already a number, otherwise a parsed number (or NaN) if it was a string, else undefined
	 */
	parseFloat: function(string) {
		if (_.isString(string)) {
			string = string.trim();
			if (string) {
				return +string.replace(/,/, ".");
			}
		} else if (_.isNumber(string)) {
			return string;
		}
		return undefined;
	},
	/**
	 * Parse the first found float regardless language settings
	 * @param {String | Number} string - String to parse
	 * @returns {Number} the original number it was already a number, otherwise a parsed number (or NaN) if it was a string, else undefined
	 */
	parseFirstFloat: function(string) {
		if (_.isString(string)) {
			string = string.trim();
			if (string) {
				return parseFloat(string.replace(/,/, "."));
			}
		} else if (_.isNumber(string)) {
			return string;
		}
		return undefined;
	},
	/**
	 * Is the number close to another number with a given precision
	 * @param   {Number} number
	 * @param   {Number} closeTo
	 * @param   {Number} precision
	 * @returns {Boolean}
	 */
	isCloseTo: function(number, closeTo, precision) {
		return Math.abs(closeTo - number) < Math.pow(10, -precision) / 2;
	}
};

export default ViewerUtils;
