/**
 * @namespace VideoModule
 */
/**
 * @namespace VideoModule
 */
import KeyframeWidget from "@esko/web-viewer/video/KeyframeWidget";

import VideoLayerModel from "@esko/web-viewer/video/VideoLayerModel";
import VideoToolbar from "@esko/web-viewer/video/VideoToolbar";
import VideoTool from "@esko/web-viewer/video/VideoTool";
import AbstractModule from "@esko/web-viewer/core/AbstractModule";
import LayoutConstants from "@esko/web-viewer/layout/LayoutConstants";
import CoreConstants from "@esko/web-viewer/core/CoreConstants";
import AnnotationConstants from "@esko/web-viewer/annotation/AnnotationConstants";
import _ from "underscore";
import $ from "jquery";
/**
 * Module for enable viewing of video files
 */
var VideoModule = function(f, silenced) {
	AbstractModule.call(this, f, silenced, "Video");

	this._video = null;
	this._videoCallbacks = [];
	this._videoTool = new VideoTool(f, this);
	this._videoLayerModel = new VideoLayerModel(f, "Video", this);

	this._publishToolbarComponent(
		VideoToolbar.toolbarName,
		_.bind(function(layout) {
			this._videoToolbar = new VideoToolbar(f, this);
			this._updateVideoToolbarStatus(undefined, f.getDocumentController().getActivePage());
			return this._videoToolbar;
		}, this)
	);
	this._publishWidget(
		KeyframeWidget.widgetName,
		function() {
			return new KeyframeWidget(f, this);
		}.bind(this)
	);

	this.f.getDocumentController().whenActivePageChanged(this._updateVideoToolbarStatus.bind(this));

	var eventsAndCallbacks = {
		"tool.annotationdrawtool.activated": this._onAnnotationToolActivated.bind(this),
		"tool.annotationnotetool.activated": this._onAnnotationToolActivated.bind(this),
		"tool.annotationdrawtool.state.changed.drawing": this._onAnnotationToolActivated.bind(this),
		"tool.annotationTouchDrawTool.activated": this._onAnnotationToolActivated.bind(this),
		"annotationModel.annotation.selectedAnnotation": this._onAnnotationSelected.bind(this),
		"annotation.modified": _.debounce(_.bind(this._onAnnotationChange, this), 500)
	};
	this._registerListeners(eventsAndCallbacks);

	// Re-render annotation layer when time changes in video
	this.whenVideoAvailable(
		function(video) {
			video.addEventListener(
				"timeupdate",
				function() {
					this.f.emit("annotationModel.annotation.modelChanged");
				}.bind(this)
			);
		}.bind(this)
	);
};

VideoModule.prototype = Object.create(AbstractModule.prototype);
VideoModule.prototype.constructor = VideoModule;

/**
 * Register events
 */
VideoModule.prototype._registerListeners = function(callbacks) {
	this.f.getDocumentController().whenActivePageChanged(function(oldPage, newPage) {
		if (oldPage && oldPage.hasVideo() && !newPage.hasVideo()) {
			this._publish("layout.switchLayoutMode");
			this._unpublishTool(this._videoTool);
			_.each(
				callbacks,
				function(callback, event) {
					this.f.off(event, callback);
				},
				this
			);
		} else if (newPage.hasVideo() && (!oldPage || !oldPage.hasVideo())) {
			this._publish("layout.switchLayoutMode", LayoutConstants.LayoutMode.VIDEO);
			this._publishTool(this._videoTool);
			_.each(
				callbacks,
				function(callback, event) {
					this.f.on(event, callback);
				},
				this
			);
		}
	}, this);
};

/**
 * Internal event listener that listens to "whenActivePageChanged" events
 *
 * @param {AbstractDocumentPage} oldPage
 * @param {AbstractDocumentPage} newPage
 */
VideoModule.prototype._updateVideoToolbarStatus = function(oldPage, newPage) {
	this.f.whenModuleAvailable(
		"viewport",
		_.bind(
			function(viewportModule) {
				viewportModule.whenWidgetModelAvailable(
					function() {
						if (newPage.hasVideo && newPage.hasVideo()) {
							if (this._videoToolbar) {
								this._videoToolbar.show();
							}
							viewportModule.addLayer("Video", this._videoLayerModel);
						} else {
							if (this._videoToolbar) {
								this._videoToolbar.hide();
							}
							viewportModule.removeLayer("Video");
						}
					}.bind(this)
				);
			}.bind(this)
		)
	);
};

/**
 * Returns the current instance of the video object or waits untill it becomes available
 *
 * @param {Function} callback
 */
VideoModule.prototype.whenVideoAvailable = function(callback) {
	this._videoCallbacks.push(callback);
	if (this._video && this._video.readyState > 0) {
		callback(this._video);
	}
};

/**
 * Returns the current video object, can be null or undefined when not yet loaded
 *
 * @returns video element
 */
VideoModule.prototype.getVideo = function() {
	return this._video;
};

/**
 * Call this function to initialize a new video. Make use of 'whenVideoAvailable' to fetch an instance of this video object
 *
 * @param {DOM element} video
 */
VideoModule.prototype.fireNewVideo = function(video) {
	if (video.readyState > 0) {
		this._fireWhenVideoAvailableCallbacks(video);
	} else {
		$(video).one(
			"loadedmetadata",
			function() {
				this._fireWhenVideoAvailableCallbacks(video);
			}.bind(this)
		);
	}
};

/**
 * Internal function to fire all callbacks with a specified video object
 *
 * @param {DOM element} video
 */
VideoModule.prototype._fireWhenVideoAvailableCallbacks = function(video) {
	this._video = video;
	_.each(
		this._videoCallbacks,
		function(callback) {
			callback(this._video);
		}.bind(this)
	);
};

/**
 * Listener that is fired when one of the annotation tools are activated
 *
 */
VideoModule.prototype._onAnnotationToolActivated = function() {
	if (this._video) {
		this._video.pause();
	}
};

/**
 * Listener that is fired when an annotation is selected
 *
 * @param {Annotation} annot
 */
VideoModule.prototype._onAnnotationSelected = function(annot) {
	if (this._video) {
		this._video.pause();
		if (this._video.currentTime < annot.startTime || this._video.currentTime > annot.endTime) {
			this._video.currentTime = annot.startTime;
		}
	}
};

VideoModule.prototype._onAnnotationChange = function(annot, what) {
	if (what === "THUMBNAIL" || what === "DETAILEDTHUMBNAIL") {
		return;
	}

	if (annot.startTime != null && this._video) {
		this._createAnnotationThumbnail(annot);
	}
};

/**
 * Modifies the thumbnail properties of the given annot
 *
 * @param {Annotation} annot
 */
VideoModule.prototype._createAnnotationThumbnail = function(annot) {
	//To be implemented in inheriting module
};

/**
 * Gets the name of the transcoding that should be used initially. Depends on the user preference and the ability to play the video in the browser
 *
 * @returns {String}
 */
VideoModule.prototype.getPreferredTranscoding = function() {
	var preferredTranscoding = this.f.getOption("PREFERENCES_VIDEO_TRANSCODING");
	var groupedTranscodings = this.f.getActivePage().get("transcodings");
	// Check if transcoding can be played or if a preferred transcoding was not yet previously selected
	if (
		(preferredTranscoding && (!groupedTranscodings[preferredTranscoding] || !groupedTranscodings[preferredTranscoding].canPlayType)) ||
		!preferredTranscoding
	) {
		_.each(groupedTranscodings, function(transcodings, groupName) {
			if (transcodings.canPlayType) {
				preferredTranscoding = groupName;
			}
		});
	}

	// If no available transcoding was selected then just pick any transcoding.
	// It will fail since it can not be played by the browser.
	// This is on purpose to generate an error message.
	if (!preferredTranscoding || !groupedTranscodings[preferredTranscoding]) {
		preferredTranscoding = Object.keys(groupedTranscodings)[0];
	}
	return preferredTranscoding;
};

export default VideoModule;
