import AbstractTool from "@esko/web-viewer/viewport/tools/AbstractTool";
import CoreConstants from "@esko/web-viewer/core/CoreConstants";
import FullscreenUtil from "@esko/web-viewer/utils/FullscreenUtil";

/**
 * Tools that registers key events and mouse clicks for video pages.
 * @param {ViewerFacade} f
 * @param {VideoModule} module
 */
var VideoTool = function(f, module) {
	AbstractTool.call(this, f, CoreConstants.TOOLS.VIDEOTOOL);
	this._registerListeners();

	/**
	 * @override
	 */
	this._override("doActivation", function() {
		this._super();
	});

	var keyEvents = [];
	var lastPress;
	this.changeState("default");

	// space to pause is on keyup to not interfere with the panning tool.
	keyEvents.push([
		"space",
		function(e) {
			// only pause/play the video of the spacebar was pressed 300ms ago
			var diff = Date.now() - lastPress;
			lastPress = null;
			if (
				this.f
					.getDocumentController()
					.getActivePage()
					.hasVideo() &&
				diff < 300
			) {
				module.whenVideoAvailable(function(video) {
					video.paused ? video.play() : video.pause();
				});
			}
		}.bind(this),
		"keyup"
	]);

	keyEvents.push([
		"space",
		function(e) {
			if (!lastPress) {
				lastPress = Date.now();
			}
			e.preventDefault();
		}.bind(this),
		"keydown"
	]);

	// go back 5 seconds
	keyEvents.push([
		"left",
		function(e) {
			if (
				this.f
					.getDocumentController()
					.getActivePage()
					.hasVideo()
			) {
				module.whenVideoAvailable(function(video) {
					video.currentTime = Math.max(video.currentTime - 5, 0);
				});
			}
		}.bind(this),
		"keydown"
	]);

	// forwards 5 seconds
	keyEvents.push([
		"right",
		function(e) {
			if (
				this.f
					.getDocumentController()
					.getActivePage()
					.hasVideo()
			) {
				module.whenVideoAvailable(function(video) {
					video.currentTime = Math.min(video.currentTime + 5, video.duration);
				});
			}
		}.bind(this),
		"keydown"
	]);

	// Keys can be used both when the video tool is active or not.
	this.activeKeyEvents = keyEvents;
	this.passiveKeyEvents = keyEvents;
};

VideoTool.prototype = Object.create(AbstractTool.prototype);
VideoTool.prototype.constructor = VideoTool;

/**
 * register click listeners
 */
VideoTool.prototype._registerListeners = function() {
	this.registerActiveEvents(["dblclick"]);
	var showAboveContent = true;
	this.f.emit("layout.toggleAboveContent", true);
	var toggleAboveContent = function() {
		showAboveContent = !showAboveContent || !!this._selectedAnnotation;
		this.f.emit("layout.toggleAboveContent", showAboveContent);
	}.bind(this);

	var defaultState = {
		dblclick: function() {
			if (FullscreenUtil.fullscreenElement) {
				FullscreenUtil.exitFullscreen();
			} else {
				this.f
					.getLayoutManager()
					.getLayout()
					.fullscreenViewport();
			}
		}.bind(this)
	};

	if (window.PointerEvent) {
		this.registerPassiveEvents(["pointerup"]);
		defaultState.pointerup = toggleAboveContent;
	} else {
		this.registerPassiveEvents(["touchend"]);
		defaultState.touchend = toggleAboveContent;
	}

	this.registerState("default", defaultState);

	this.f.on(
		"annotationModel.annotation.selectedAnnotation",
		function(annotation) {
			this._selectedAnnotation = annotation;
			toggleAboveContent();
		}.bind(this)
	);

	this.f.on(
		"annotationModel.annotation.unselectedAnnotation",
		function(annotation) {
			this._selectedAnnotation = null;
		}.bind(this)
	);
};

VideoTool.prototype.getShortcuts = function() {
	return {
		videoShortcuts: {
			playPauseShortcut: "Space",
			videoForward5Sec: "&rarr;",
			videoBackward5Sec: "&larr;"
		}
	};
};

export default VideoTool;
