import DropdownMenu from "@esko/dropdownmenu/DropdownMenu";
import Slider from "@esko/slider/Slider";
import ToolbarComponent from "@esko/toolbar/ToolbarComponent";
import ButtonGroup from "@esko/toolbar/ButtonGroup";
import ViewerToolbarButton from "@esko/web-viewer/layout/component/ViewerToolbarButton";
import SubtoolbarButton from "@esko/web-viewer/layout/component/SubtoolbarButton";
import LayoutConstants from "@esko/web-viewer/layout/LayoutConstants";
import FullscreenUtil from "@esko/web-viewer/utils/FullscreenUtil";
import TimelineSlider from "@esko/web-viewer/video/TimelineSlider";
import _ from "underscore";
import $ from "jquery";
/**
 * @exports VideoToolbar
 * @memberof VideoModule
 * @constructor
 * @param {Core.ViewerFacade} f - Viewer facade
 * @param {VideoModule} videoModule - Video Module
 */
var VideoToolbar = function(f, videoModule) {
	ToolbarComponent.call(this, LayoutConstants.ToolbarComponent.VIDEO);

	this.f = f;
	this.module = videoModule;

	this.buttonGroup = new ButtonGroup("Video");
	this.buttonGroup.setTitle(f.i18n, "viewer", "VideoBtnGroupTitle");
	this.addButtonGroup(this.buttonGroup);

	// Listen for a new video source -> generate new video controls
	this.module.whenVideoAvailable(
		function(video) {
			if (this._slider) {
				this._slider.destroy();
			}
			this.buttonGroup.btngroup.empty();
			this.buttonGroup.buttons = [];
			clearInterval(this._interval);
			this._addButtons(video, this.f.isMobile());
		}.bind(this)
	);

	this.f.on(
		"annotationModel.annotation.selectedAnnotation",
		function(annotation) {
			this._selectedAnnotation = annotation;
			this._togglePauseButtonEnabled();
			annotation.on("modified.startEndTime", this._togglePauseButtonEnabled);
		}.bind(this)
	);
	this.f.on(
		"annotationModel.annotation.unselectedAnnotation",
		function(annotation) {
			annotation.off("modified.startEndTime", this._togglePauseButtonEnabled);
			this._selectedAnnotation = null;
			this._togglePauseButtonEnabled();
		}.bind(this)
	);

	this._togglePauseButtonEnabled = function() {
		if (this._selectedAnnotation && this._selectedAnnotation.startTime === this._selectedAnnotation.endTime) {
			this.module.whenVideoAvailable(function(video) {
				video.pause();
			});
			this._pauseButton.setEnabled(false);
		} else {
			this._pauseButton.setEnabled(true);
		}
	}.bind(this);

	this.f.getCurrentUser().on(
		"options.changed.PREFERENCES_VIDEO_LOOP",
		function(oldValue, newValue) {
			this.module.whenVideoAvailable(
				function(video) {
					if (newValue === 1 && !this._selectedAnnotation) {
						$(video).prop("loop", true);
						video.play();
					} else {
						$(video).prop("loop", false);
					}
				}.bind(this)
			);
		}.bind(this)
	);
};

VideoToolbar.prototype = Object.create(ToolbarComponent.prototype);
VideoToolbar.prototype.constructor = VideoToolbar;
VideoToolbar.toolbarName = LayoutConstants.ToolbarComponent.VIDEO;

/**
 * Generates and adds buttons to the toolbar
 *
 * @param {Video} video
 */
VideoToolbar.prototype._addButtons = function(video, forMobile) {
	var initialIcon = video.paused || video.ended ? "icon-WCR_Video-Play" : "icon-WCR_Video-Pause";
	var pauseButton = new ViewerToolbarButton(this.f, "PauseButton", initialIcon, forMobile, function() {
		video.paused || video.ended ? video.play() : video.pause();
	});
	this.buttonGroup.addButton(pauseButton);
	this._pauseButton = pauseButton;

	var oldVolume = video.volume;
	if (!forMobile) {
		var muteButton = new ViewerToolbarButton(
			this.f,
			"MuteButton",
			video.volume > 0 ? "icon-WCR_Video-Volume" : "icon-WCR_Video-Volume-Muted",
			forMobile,
			function() {
				if (video.volume > 0) {
					oldVolume = video.volume;
					video.volume = 0;
				} else {
					video.volume = oldVolume > 0 ? oldVolume : 1;
				}
				volumeSlider.setValue(this._volumeToSlider(video.volume));
			}.bind(this)
		);

		muteButton.setTooltip(this.f.i18n.t("VideoMute"), null, null, {
			prefPosition: "top",
			container: this.buttonGroup.btngroup
		});
		muteButton.element.addClass("btn-mute");

		var volumeSlider = new Slider({
			class: "volume-slider",
			single: {
				value: this._volumeToSlider(video.volume)
			},
			max: 100,
			on: {
				"value.change": function(newValue) {
					video.volume = this._sliderToVolume(newValue);
				}.bind(this)
			}
		});

		this.buttonGroup.btngroup.append(muteButton.element).append(volumeSlider.element);
	}
	this._slider = new TimelineSlider(this.f, video, this.module);

	this.buttonGroup.addButton({
		element: this._slider.element,
		moved: function() {}
	});

	var settingsDropdown = new DropdownMenu(
		this.f.i18n,
		this.buttonGroup.btngroup,
		this.f.isMobile(),
		"top",
		"wcr-dropdown",
		undefined,
		undefined,
		this.f.getContext()
	);
	var rates = [0.25, 0.5, 1, 2, 8];

	rates.forEach(
		function(rate) {
			settingsDropdown.addOption({
				name: "changeRate" + rate,
				content: rate === 1 ? "<b>" + this.f.i18n.t("VideoDefaultPlaybackRate") + "</b>" : rate + "x",
				selected: rate === 1,
				onchange: function() {
					video.playbackRate = rate;
				},
				userCannotUncheck: true
			});
		}.bind(this)
	);

	// group transcodings by resolution
	var groupedTranscodings = this.f.getActivePage().get("transcodings");
	if (Object.keys(groupedTranscodings).length > 1) {
		settingsDropdown.addDivider();
		var preferredTranscoding = this.module.getPreferredTranscoding();
		_.each(
			groupedTranscodings,
			function(transcodings, groupName) {
				settingsDropdown.addOption({
					name: "transcoding" + groupName,
					content: groupName,
					enabled: transcodings.canPlayType,
					selected: preferredTranscoding == groupName,
					onchange: function() {
						_.each(groupedTranscodings, function(transcodings, otherGroupName) {
							settingsDropdown.toggleOption("transcoding" + otherGroupName, otherGroupName == groupName);
						});

						this.f.setOption("PREFERENCES_VIDEO_TRANSCODING", groupName);
					}.bind(this),
					userCannotUncheck: true
				});
			}.bind(this)
		);
	}

	settingsDropdown.addDivider();

	if (!this.f.isMobile()) {
		settingsDropdown.addOption({
			name: "autoPlay",
			content: this.f.i18n.t("VideoSettingAutoplay"),
			onchange: function() {
				var isAutoPlay = this.f.getOption("PREFERENCES_VIDEO_AUTO_PLAY") === 1;
				this.f.setOption("PREFERENCES_VIDEO_AUTO_PLAY", isAutoPlay ? 0 : 1);
			}.bind(this),
			selected: this.f.getOption("PREFERENCES_VIDEO_AUTO_PLAY") === 1
		});
	}

	settingsDropdown.addOption({
		name: "loop",
		content: this.f.i18n.t("VideoSettingLoop"),
		onchange: function() {
			var isLoop = this.f.getOption("PREFERENCES_VIDEO_LOOP") === 1;
			this.f.setOption("PREFERENCES_VIDEO_LOOP", isLoop ? 0 : 1);
		}.bind(this),
		selected: this.f.getOption("PREFERENCES_VIDEO_LOOP") === 1
	});

	var settingsButton = new ViewerToolbarButton(this.f, "SettingsButton", "icon-WCR_Settings", forMobile, function() {
		settingsDropdown.show(settingsButton.element);
	});
	settingsButton.setTooltip(this.f.i18n.t("VideoPlayerSettings"), null, null, {
		prefPosition: "top",
		container: this.buttonGroup.btngroup
	});
	this.buttonGroup.addButton(settingsButton);

	if (FullscreenUtil.fullscreenEnabled && !forMobile) {
		var fullscreenButton = new ViewerToolbarButton(this.f, "FullscreenButton", "icon-WCR_Video-FullScreen", false, function() {
			if (FullscreenUtil.fullscreenElement) {
				FullscreenUtil.exitFullscreen();
			} else {
				this.f
					.getLayoutManager()
					.getLayout()
					.fullscreenViewport();
			}
		});
		fullscreenButton.setTooltip(this.f.i18n.t("VideoFullscreen"), null, null, {
			prefPosition: "top",
			container: this.buttonGroup.btngroup
		});
		this.buttonGroup.addButton(fullscreenButton);

		FullscreenUtil.onfullscreenchange = function() {
			if (FullscreenUtil.fullscreenElement) {
				fullscreenButton.setIcon("icon-WCR_WF-Minimize");
				fullscreenButton.setTooltip(this.f.i18n.t("VideoUnfullscreen"));
			} else {
				fullscreenButton.setIcon("icon-WCR_Video-FullScreen");
				fullscreenButton.setTooltip(this.f.i18n.t("VideoFullscreen"));
			}
		}.bind(this);
	}

	/*********************************************************
	 * ADD ALL LISTENERS TO VIDEO TO UPDATE PLAYER CONTROLS *
	 *********************************************************/

	$(video).on(
		"play pause ended playing",
		function() {
			pauseButton.setIcon(video.paused || video.ended ? "icon-WCR_Video-Play" : "icon-WCR_Video-Pause");
		}.bind(this)
	);

	$(video).on(
		"volumechange",
		function() {
			if (video.volume > 0) {
				muteButton.setTooltip(this.f.i18n.t("VideoMute"));
				muteButton.setIcon("icon-WCR_Video-Volume");
			} else {
				muteButton.setTooltip(this.f.i18n.t("VideoUnmute"));
				muteButton.setIcon("icon-WCR_Video-Volume-Muted");
			}
			this.f.setOption("PREFERENCES_VIDEO_VOLUME", "" + video.volume);
		}.bind(this)
	);

	$(video).on("ratechange", function() {
		rates.forEach(function(rate) {
			settingsDropdown.toggleOption("changeRate" + rate, rate === video.playbackRate);
		});
	});
};

/**
 * Utility function used to convert volume values from 0-1 to a slider value of 0-100
 *
 * @param {Number} volume between 0-1
 * @returns {Number} value between 0-100
 */
VideoToolbar.prototype._volumeToSlider = function(volume) {
	return Math.pow(volume, 1 / 3) * 100;
};

/**
 * Utility function used to convert slider values from 0-100 to a volume 0-1
 *
 * @param {Number} sliderValue
 * @returns {Number} value between 0-1
 */
VideoToolbar.prototype._sliderToVolume = function(sliderValue) {
	return Math.pow(sliderValue / 100, 3);
};

export default VideoToolbar;
