import DataConstants from "@esko/web-viewer/data/DataConstants";
import AbstractWidget from "@esko/web-viewer/layout/AbstractWidget";
import MiniButton from "@esko/web-viewer/layout/component/MiniButton";
import LayoutConstants from "@esko/web-viewer/layout/LayoutConstants";
import ShapeUtils from "@esko/shapeutils";
import $ from "jquery";
import _ from "underscore";
/**
 * @exports
 * @memberof Compare
 * @augments Layout.AbstractWidget
 * @class The View3DWidget shows the 3D Viewer inside a widget of the HTML5 Viewer.
 * @constructor
 *
 * @param {Core.ViewerFacade} f
 */
var AbstractView3DWidget = function AbstractView3DWidget(f, name) {
	AbstractWidget.call(this, f, name);

	/**
	 * @member {Core.ViewerFacade} the viewer facade
	 */
	this.f = f;

	this.setIcon("icon icon-WCR_Viewer-View-3D");

	this._3DContent = $("<div>", {
		class: "View3DContent"
	});

	// intialize the open in new window button
	this._openNewWindow = new MiniButton(
		this.f,
		"open3D",
		"icon-WCR_FullScreen",
		true,
		_.bind(function() {
			var compareModule = this.f.getModule("compare");
			if (compareModule) {
				compareModule.activateModeByName();
			}

			this.f.emit("layout.switchLayoutMode", LayoutConstants.LayoutMode.DESKTOP_3D);
		}, this)
	);
	this._openNewWindow.setTooltip("viewer", "expand3DView");

	this.f.i18n.loadBundle(
		"viewer",
		_.bind(function() {
			// set the widget title
			this.setTitle(this.f.t("view3DTitle"));
		}, this)
	);
};

AbstractView3DWidget.prototype = Object.create(AbstractWidget.prototype);
AbstractView3DWidget.prototype.constructor = AbstractView3DWidget;

/**
 * Shows the 3D widget and starts to fetch the 3D model
 * @param {Object} [position] - the initial camera position. See setCameraPosition.
 */
AbstractView3DWidget.prototype.show = function(position, requestedDocVerId) {
	this.whenContextAvailable(function() {
		this.makeVisible();

		var docVersionId;
		if (requestedDocVerId) {
			docVersionId = requestedDocVerId;
		} else {
			var page = this._getPage();
			if (page) {
				var threeDViews = page.getDocument().getLinkedDocumentPages(DataConstants.LINKEDDOCUMENTTYPES.THREEDVIEW);
				if (threeDViews.length > 0) {
					// there are 3DView linked documents, used that one to generate the 3D widget.
					docVersionId = threeDViews[0].page.getDocumentVersion().getDocumentVersionID();
				} else if (page.get("canGenerate3D")) {
					// Use original 3D page to render
					docVersionId = page.getDocumentVersion().getDocumentVersionID();
				}
			}
		}

		if (!docVersionId) {
			return; // Could not find a docVerId to render 3D OR we're trying to fetch the same docVersionId again
		}
		this._currentDocVersionId = docVersionId;

		var finalCallback = _.bind(function(success, error) {
			if (success) {
				this._on3DPerspectiveGenerated();
				this._emit("loaded");
			}

			if (error) {
				this._currentDocVersionId = undefined;
			}

			this.resized();

			if (!success) {
				this.f.loadBundle(
					"viewer",
					function() {
						if (!error) {
							error = this.f.t("errorLoading3D");
						}
						this.f.emit("userMessage", "error", error, undefined, false, "3DERROR", false);
						this.setEnabled(false);
					},
					this
				);
			} else {
				this.f.emit("view3d.widget.loaded");
			}
		}, this);

		//make sure the 3D Viewer is not initialized with a height smaller than the minimum widget height
		var minimumSize = this.getMinimumSize();
		if ((this.context.innerHeight() || null) < minimumSize.height) {
			this.context.innerHeight(minimumSize.height);
		}

		// start loading the 3D view
		if (!this._3DLauncher) {
			this.f.whenModuleAvailable("view3D", this._create3DViewerLauncher.bind(this, docVersionId, finalCallback, position));
		} else {
			this._3DLauncher.load3DForDocVersion(docVersionId, finalCallback, position);
		}
	}, this);
};

/**
 * Creates the 3DViewer launcher for a specifed docVersionID.
 * This function will be called can be called only once. Use this._3DLauncher afterwards
 */
AbstractView3DWidget.prototype._create3DViewerLauncher = function(docVersionId, finalCallback, position, mod) {
	if (!this._3DLauncher) {
		this._3DLauncher = mod.get3DViewerLauncher(this._3DContent, docVersionId, finalCallback, position);
		this._3DLauncher.setEndCameraInteractionCallback(
			function() {
				this._emit("camerainteractionend");
			}.bind(this)
		);
	}
};

/**
 * @override
 */
AbstractView3DWidget.prototype.setContext = function(context, headerExtra) {
	// call the old implementation of setContext() (the implementation of AbstractWidget.js)
	AbstractWidget.prototype.setContext.call(this, context, headerExtra);

	this.context.addClass("view3DWidget");
	this.context.append(this._3DContent);

	headerExtra.append(this._openNewWindow.element);
};

/**
 * Get the minimum size for the View3DWidget
 * @returns {ShapeUtils.Size}
 */
AbstractView3DWidget.prototype.getMinimumSize = function() {
	return new ShapeUtils.Size(250, 300);
};

/**
 * Get the prefered size for the View3DWidget
 * @returns {ShapeUtils.Size}
 */
AbstractView3DWidget.prototype.getPreferredSize = function() {
	return new ShapeUtils.Size(250, 300);
};

/**
 * Call when the widget is resizing and the contents needs to adapt
 */
AbstractView3DWidget.prototype.resizing = function() {
	this.resized();
};

/**
 * This method notifies the widget that it has been resized
 */
AbstractView3DWidget.prototype.resized = function() {
	if (this._3DLauncher) {
		var view3DModule = this.f.getModule("view3D");
		if (view3DModule && view3DModule.isUsingAnnotations()) {
			// Using annotations -> Resize and keep camera distance equal
			var originalDistance = this._3DLauncher.getCameraPosition().distance;
			this._3DLauncher.resize(); // TODO STU-6914, wait for new @esko/3d-web-viewer version that includes this._viewer.Resize() with redraw flag to fix unnecessary re-render
			if (originalDistance > 0) {
				this._3DLauncher.setCameraPosition({ distance: originalDistance });
			}
		} else {
			// Not using annotations -> Resize
			this._3DLauncher.resize();
		}
	}
};

/**
 * Get the current camera position if the 3D viewer is loaded
 * @returns {Object} @example
 *                   {scale: 1.5,
 *                   azimuth: 46.6,
 *                   elevation: -52}
 */
AbstractView3DWidget.prototype.getCameraPosition = function() {
	if (this._3DLauncher) {
		return this._3DLauncher.getCameraPosition();
	}
	return null;
};

/**
 * Set the current camera position of the 3D widget if the 3D viewer is loaded
 * @param {Object} position -
 *                 @example
 *                    {scale: 1.5,
 *                    azimuth: 46.6,
 *                    elevation: -52}
 */
AbstractView3DWidget.prototype.setCameraPosition = function(position) {
	if (this._3DLauncher) {
		this._3DLauncher.setCameraPosition(position);
	}
};

/**
 * Gets the AbstractDocumentPage that is used to render
 *
 */
AbstractView3DWidget.prototype._getPage = function() {
	throw new Error("AbstractView3DWidget.prototype._getPage is not implemented!");
};

/**
 * Is called whenever the 3D perspective is succesfully generated
 *
 */
AbstractView3DWidget.prototype._on3DPerspectiveGenerated = function() {
	// Can be implemented in the subclass
};

/**
 * Get a snapshot of the current viewer state as specified in format
 *
 * @function GetSnapshot
 * @param {string} [format = "png"] - image format (accepted: "png, jpeg")
 * @param {number} [width = context canvas width or 1] - image width in pixels (accepted: [1:2048] higher depend on graphics card and browser)
 * @param {number} [height = context canvas height or 1] - image height in pixels (accepted: [1:2048] higher depend on graphics card and browser)
 * @returns {Object} image data
 */
AbstractView3DWidget.prototype.getSnapShot = function(format, width, height) {
	return this._3DLauncher.getSnapShot(format, width, height);
};

/**
 * Callbacks passed to this function will be called once when the 3d viewer is or becomes available
 *
 * @param {Function} callback
 * @param {Object} binding
 */
AbstractView3DWidget.prototype.when3DContentAvailable = function(callback, binding) {
	if (binding) {
		callback = _.bind(callback, binding);
	}

	if (this._3DLauncher != null) {
		callback();
	} else {
		this.once("loaded", callback);
	}
};

/**
 * @returns JQuery Element that wraps the 3d viewer
 */
AbstractView3DWidget.prototype.get3DContent = function() {
	return this._3DContent;
};

export default AbstractView3DWidget;
