/**
 * @namespace View3DModule
 */
/**
 * @namespace View3DModule
 */
import _ from "underscore";

import $ from "jquery";
import ShapeUtils from "@esko/shapeutils";
import Transformation from "@esko/transformation";
import Viewer3DLauncher from "@esko/3d-viewer-launcher/3DViewerLauncher";
import AnnotationConstants from "@esko/web-viewer/annotation/AnnotationConstants";
import AbstractModule from "@esko/web-viewer/core/AbstractModule";
import CoreConstants from "@esko/web-viewer/core/CoreConstants";
import View3DWidget from "@esko/web-viewer/view3D/View3DWidget";
import CompareView3DWidget from "@esko/web-viewer/view3D/CompareView3DWidget";
import View3DToolbarComponent from "@esko/web-viewer/view3D/View3DToolbarComponent";
import AbstractView3DDataAdapter from "@esko/web-viewer/view3D/AbstractView3DDataAdapter";
var ANNOTATION_TOOLS = [
	CoreConstants.TOOLS.ANNOTATION_TOUCH_DRAW,
	CoreConstants.TOOLS.ANNOTATION_TOUCH,
	CoreConstants.TOOLS.ANNOTATION_DRAW,
	CoreConstants.TOOLS.ANNOTATION_NOTE_DRAW,
	CoreConstants.TOOLS.ANNOTATION_INDICATOR
];

/**
 * Creates the View3DModule
 *
 * @exports View3DModule
 * @memberof View3DModule
 * @class
 * Responsible for publishing a 3D Widget which can show the 3D model of ARD or Graphics if it's linked to an ARD
 * @augments Core.AbstractModule
 * @param {Core.ViewerFacade} f
 * @param {Boolean} silenced
 */
var View3DModule = function(f, silenced, adapter) {
	AbstractModule.call(this, f, silenced, "view3D", adapter || new AbstractView3DDataAdapter(f));

	this._3DWidget = null;
	this._toolbarComponent = null;

	var boundUpdateToolbarVisibility = _.bind(this._updateToolbarVisibility, this);
	var boundShow3DIfAvailable = function() {
		var newPage = this.f.getDocumentController().getActivePage();
		if (newPage) {
			var linkedPage = newPage.getLinkedPage();
			if (linkedPage) {
				if (this.f.getDocumentController().getCadOverlay() != null) {
					this.show3D(null, linkedPage.get("docversionid"));
				}
			} else {
				if (this._3DWidget) {
					this._3DWidget.setEnabled(false);
				}
				this.f.emit("layout.switchLayoutMode");
			}
		}
		boundUpdateToolbarVisibility();
	}.bind(this);

	this.f.getDocumentController().whenActivePageChangedAndParametersAvailable(function(oldPage, newPage) {
		if (this._3DWidget && !newPage.has3D()) {
			//hide the widget when the active document page changed
			this._3DWidget.setEnabled(false);
			this._emit("stop3D");
		}

		var diffPage = this.f.getDocumentController().getDifferencePage();
		if (diffPage && !diffPage.has3D() && this._difference3DWidget) {
			this._difference3DWidget.setEnabled(true);
		}

		// Auto-show 3DView widget if the page is 3D or CAD (rely on overlay change event to load when there is only a 3D perspective)
		if (newPage.has3D()) {
			this.show3D();
		}

		// remove old listeners from old page
		if (oldPage) {
			oldPage.off("property.has3DPerspective.changed", boundUpdateToolbarVisibility);
		}

		// listen to has3DPerspective property changes
		newPage.on("property.has3DPerspective.changed", boundUpdateToolbarVisibility);

		this._updateToolbarVisibility();
	}, this);

	this.f.getDocumentController().whenDocumentsLoaded(
		_.once(
			function() {
				this.f.on("doc.*.changed.linkedDocument.*", boundShow3DIfAvailable);
			}.bind(this)
		)
	);

	this.f.getDocumentController().whenDifferencePageChanged(
		function(oldPage, newPage) {
			if (newPage && newPage.has3D()) {
				this._showDifference3D();
			}
		}.bind(this)
	);

	this.f.getDocumentController().whenCadOverlayChanged(
		function() {
			this._updateToolbarVisibility();
			if (this._3DWidget) {
				this._3DWidget.setEnabled(this.has3DPerspective() && this.f.getDocumentController().getCadOverlay() != null);
			}
		}.bind(this)
	);

	if (!this.f.isMobile()) {
		this.setLockOn(false);
	}

	this.f.on(
		"changed.lock",
		_.bind(function(value) {
			this.setLockOn(value);
		}, this)
	);

	this.f.on("toolManager.activated", this._onPossiblyAnnotationInteractionChange.bind(this));
	this.f.on("toolManager.deactivated", this._onPossiblyAnnotationInteractionChange.bind(this));
	this.f.on("annotationModel.annotation.selectedAnnotation", this._onPossiblyAnnotationInteractionChange.bind(this));
	this.f.on("annotationModel.annotation.unselectedAnnotation", this._onPossiblyAnnotationInteractionChange.bind(this));
	this.f.on("annotation.modified", _.debounce(this._onAnnotationChange.bind(this), 500));

	this.on(
		"usingAnnotations",
		function(usingAnnotations) {
			if (usingAnnotations) {
				this.f.emit("viewport.viewmodel.Main.layers.moveLayerToBack", "View3D");
			} else {
				this.f.emit("viewport.viewmodel.Main.layers.restoreLayer", "View3D");
			}
		}.bind(this)
	);
};

// inherit
View3DModule.prototype = Object.create(AbstractModule.prototype);
// correct the constructor pointer
View3DModule.prototype.constructor = View3DModule;

View3DModule.prototype.get3DViewerLauncher = function(threeDContent, docVersionId, finalCallback, position) {
	return new Viewer3DLauncher(threeDContent, docVersionId, finalCallback, position);
};

/**
 * Shows the 3D widget with the HTML5 3D Collada viewer
 */
View3DModule.prototype.show3D = function(position, docVersionId) {
	this._emit("start3D");
	if (!this._3DWidget) {
		this._publishWidget(
			View3DWidget.widgetName,
			_.bind(function(layoutType) {
				this._3DWidget = new View3DWidget(this.f);
				this._3DWidget.show(position);

				// React on enabled changes
				this._3DWidget.on(
					"enabled",
					_.bind(function(enabled) {
						this._emit("3DWidget.enabled", enabled);
					}, this)
				);
				this._3DWidget.on("camerainteractionend", this.onCameraInteractionEnd.bind(this));
				this._3DWidget.when3DContentAvailable(
					function() {
						this.prevCur = this._3DWidget.getCameraPosition();
					}.bind(this)
				);
				this._emit("3DWidget.enabled", this._3DWidget.isEnabled());

				this._emit("3Dstarted");

				return this._3DWidget;
			}, this)
		);
	} else {
		this._3DWidget.show(position, docVersionId);
		this._emit("3Dstarted");
	}
};

View3DModule.prototype.set3DDifferenceEnabled = function(enabled) {
	if (this._difference3DWidget && !enabled) {
		this._difference3DWidget.setEnabled(enabled);
	} else if (enabled) {
		this._showDifference3D();
	}
};

View3DModule.prototype._showDifference3D = function() {
	if (!this._difference3DWidget) {
		this._publishWidget(
			CompareView3DWidget.widgetName,
			_.bind(function(layoutType) {
				this._difference3DWidget = new CompareView3DWidget(this.f);
				this._difference3DWidget.on("camerainteractionend", this.onCameraInteractionEnd.bind(this));
				this._difference3DWidget.show();
				this._difference3DWidget.when3DContentAvailable(
					function() {
						this.prevDiff = this._difference3DWidget.getCameraPosition();
					}.bind(this)
				);
				return this._difference3DWidget;
			}, this)
		);
	} else {
		this._difference3DWidget.setEnabled(true);
		this._difference3DWidget.show();
	}
};

/**
 * Is the 3D widget currently enabled
 * @returns {Boolean} True if enabled, false if not enabled
 */
View3DModule.prototype.is3DWidgetEnabled = function() {
	if (this._3DWidget) {
		return this._3DWidget.isEnabled();
	}
	return false;
};

/**
 * Initialize and shows the 3D toolbar component.
 */
View3DModule.prototype._show3DToolbarButton = function() {
	if (!this._toolbarComponent) {
		this._publishToolbarComponent(
			View3DToolbarComponent.toolbarName,
			_.bind(function(layoutType) {
				this._toolbarComponent = new View3DToolbarComponent(this.f, this);

				return this._toolbarComponent;
			}, this)
		);
	} else {
		this._toolbarComponent.show();
	}
};

/**
 * Update the visibility of the 3D toolbar component
 */
View3DModule.prototype._updateToolbarVisibility = function() {
	if (this.has3DPerspective()) {
		this._show3DToolbarButton();
	} else {
		if (this._toolbarComponent) {
			this._toolbarComponent.hide();
		}
	}
	this._emit("has3DPerspective", this.has3DPerspective());
};

/**
 * Is there a 3D perspective available? should the 3D view toolbarbutton or mode be available?
 * @returns {Boolean}
 */
View3DModule.prototype.has3DPerspective = function() {
	return this.f.getActivePage().has3D();
};

/**
 * Get the current camera position of the 3D widget if it is enabled or null otherwise
 * @returns {Object} @example
 *                   {scale: 1.5,
 *                   azimuth: 46.6,
 *                   elevation: -52}
 */
View3DModule.prototype.getCameraPosition = function() {
	if (this._3DWidget && this._3DWidget.isEnabled()) {
		return this._3DWidget.getCameraPosition();
	} else {
		return null;
	}
};

/**
 * Set the current camera position of the 3D widget if it is enabled
 * Otherwise it will start loading the 3d widget
 * @param {Object} position -
 *                 @example
 *                    {scale: 1.5,
 *                    azimuth: 46.6,
 *                    elevation: -52}
 */
View3DModule.prototype.setCameraPosition = function(position) {
	if (this._3DWidget && this._3DWidget.isEnabled()) {
		this._3DWidget.setCameraPosition(position);
	} else {
		this.show3D(position);
	}
};

/**
 * Set whether or now Documnts should be locked on their positions
 * When on, you can move current and difference documents separately in the respective widgets
 * When off, documents will continuously sync.
 * @param {Broolean} value
 */
View3DModule.prototype.setLockOn = function(value) {
	this.lockOn = value;

	var fn = _.bind(function() {
		if (this.f.getOption("PREFERENCES_VIEW3D_CAMERA_LOCK_LIVE") === 1) {
			this.syncLatestChanges();
		}

		if (!this.lockOn) {
			window.setTimeout(fn, 120);
		}
	}, this);

	if (!this.lockOn) {
		if (this._3DWidget) {
			this.prevCur = this._3DWidget.getCameraPosition();
		}
		if (this._difference3DWidget) {
			this.prevDiff = this._difference3DWidget.getCameraPosition();
		}
		fn();
	}
};

/**
 * Set the Camera position of the different to the same as the current document.
 */
View3DModule.prototype.syncCurrentWithDifference = function() {
	if (this._3DWidget && this._difference3DWidget) {
		var cur = this._3DWidget.getCameraPosition();
		this._difference3DWidget.setCameraPosition(cur);
		if (this.prevDiff) {
			this.prevDiff = cur;
		}
	}
};

/**
 * Syncs both 3d widgets (main and diff) based upon the most recent changes.
 * The last interacted widget will serve as the new position for both widgets
 *
 */
View3DModule.prototype.syncLatestChanges = function() {
	if (!this.lockOn && this._3DWidget && this._difference3DWidget) {
		var cur = this._3DWidget.getCameraPosition();
		var diff = this._difference3DWidget.getCameraPosition();

		if (cur && diff) {
			if (!this.prevCur) {
				this.prevCur = cur;
			}
			if (!this.prevDiff) {
				this.prevDiff = diff;
			}

			var multiplicative = ["scale"];
			var additive = ["azimuth", "elevation", "panx", "pany"];
			var env = ["_x", "_y", "_z", "_w"];

			var hasDifference = function(obj1, obj2, arr) {
				var hasDiff = false;
				_.each(arr, function(option) {
					if (obj1[option] != obj2[option]) {
						hasDiff = true;
					}
				});
				return hasDiff;
			};
			var propagateDifference = function(obj1, obj2, target, multiplicative, additive) {
				_.each(multiplicative, function(option) {
					target[option] = (target[option] / obj1[option]) * obj2[option];
				});
				_.each(additive, function(option) {
					target[option] = target[option] - obj1[option] + obj2[option];
				});
			};

			var needsUpdate = function(obj1, obj2) {
				var diff = hasDifference(obj1, obj2, _.flatten([multiplicative, additive]));
				var diff2 = hasDifference(obj1.environment, obj2.environment, env);
				return diff || diff2;
			};

			var update = function(obj1, obj2, target) {
				propagateDifference(obj1, obj2, target, multiplicative, additive);
				propagateDifference(obj1.environment, obj2.environment, target.environment, [], env);
			};

			if (needsUpdate(cur, this.prevCur)) {
				update(this.prevCur, cur, this.prevDiff);
				this.prevCur = cur;
				this._difference3DWidget.setCameraPosition(this.prevDiff);
			} else if (needsUpdate(diff, this.prevDiff)) {
				update(this.prevDiff, diff, this.prevCur);
				this.prevDiff = diff;
				this._3DWidget.setCameraPosition(this.prevCur);
			}
		}
	}
};

/**
 * Check whether of not synchronization between document is possible
 * @returns {Boolean}
 */
View3DModule.prototype.canSyncWidgets = function() {
	return !!(this._3DWidget && this._difference3DWidget);
};

/**
 * This function is called whenever the interaction with the main 3d widget or difference 3d widget ends
 */
View3DModule.prototype.onCameraInteractionEnd = function() {
	if (this.f.getOption("PREFERENCES_VIEW3D_CAMERA_LOCK_LIVE") !== 1) {
		this.syncLatestChanges();
	}
};

/**
 * Gets the main 3d widget
 *
 * @returns {Abstract3DWidget}
 */
View3DModule.prototype.get3DWidget = function() {
	return this._3DWidget;
};

/**
 * Callback used to trigger a "usingAnnotations" event, see isUsingAnnotations.
 *
 * @returns {boolean}
 */
View3DModule.prototype._onPossiblyAnnotationInteractionChange = function() {
	var isUsingAnnotations = this.isUsingAnnotations();
	if (this._previouslyUsingAnnotations !== isUsingAnnotations) {
		this._previouslyUsingAnnotations = isUsingAnnotations;
		this._emit("usingAnnotations", this._previouslyUsingAnnotations);
	}
};

/**
 * Determines if the user is currently creating/changing annotations or if an annotation is selected.
 *
 * @returns {boolean}
 */
View3DModule.prototype.isUsingAnnotations = function() {
	var viewportModule = this.f.getModule("viewport");
	if (viewportModule) {
		var isUsingAnnotations = false;
		var annotationModule = this.f.getModule("annotation");
		if (annotationModule) {
			isUsingAnnotations = !_.isEmpty(annotationModule.annotationModel.getSelectedAnnotations());
		}
		var activeTool = viewportModule.widget.getToolManager().getActiveTool();
		if (activeTool) {
			isUsingAnnotations = isUsingAnnotations || ANNOTATION_TOOLS.indexOf(activeTool.getIID()) !== -1;
		}
		return isUsingAnnotations;
	}
	return false;
};

/**
 * Triggered whenever an annotation changes
 *
 * @param {Annotation} annot
 * @param {String} what reason for change
 */
View3DModule.prototype._onAnnotationChange = function(annot, what) {
	if (what === "THUMBNAIL" || what === "DETAILEDTHUMBNAIL") {
		return;
	}

	if (this.f.getActivePage().is3D()) {
		this._createAnnotationThumbnail(annot);
	}
};

/**
 * Modifies the thumbnail & detailed-thumbnail properties of the annotation
 *
 * @param {Annotation} annot
 */
View3DModule.prototype._createAnnotationThumbnail = function(annot) {
	var thumbSize = 400;
	var marginFraction = 0.2;
	var base64img = this._3DWidget.getSnapShot("png", thumbSize, thumbSize);
	annot.setThumbnail(base64img);

	var worldBounds = this.f.getModule("viewport").getWorldBounds();
	var canvas = $('<canvas width="' + thumbSize + '" height="' + thumbSize + '" />')[0];
	var ctx = canvas.getContext("2d");
	var rect = annot.rect;

	if (annot.annottype === AnnotationConstants.AnnotationType.ANNOT_PIN) {
		var noteCompensation = Math.max(worldBounds.size.height, worldBounds.size.width) * 0.1;
		var noteSize = 2 * noteCompensation;
		rect = {
			size: {
				width: noteSize,
				height: noteSize
			},
			origin: {
				x: annot.rect.origin.x - noteCompensation,
				y: annot.rect.origin.y - noteCompensation
			}
		};
	}

	const graphicsSize = 100;

	var image = new Image();
	image.onload = function() {
		var x = ((rect.origin.x + graphicsSize / 2) / graphicsSize) * thumbSize;
		var y = ((graphicsSize - (rect.origin.y + graphicsSize / 2) - Math.abs(rect.size.height)) / graphicsSize) * thumbSize;
		var width = (Math.abs(rect.size.width) / graphicsSize) * thumbSize;
		var height = (Math.abs(rect.size.height) / graphicsSize) * thumbSize;
		var middleX = x + width / 2;
		var middleY = y + height / 2;
		var size = Math.max(width, height);
		size += size * marginFraction * 2;
		ctx.drawImage(image, middleX - size / 2, middleY - size / 2, size, size, 0, 0, thumbSize, thumbSize);
		annot.setDetailedThumbnail(canvas.toDataURL());
		this.f.emit("annotationModel.annotation.modelChanged");
	}.bind(this);
	image.src = base64img;
};

View3DModule.prototype.isAngleEqual = function(cameraAngle1, cameraAngle2) {
	const properties = ["azimuth", "elevation", "distance", "panx", "pany"];
	return properties.every(prop => {
		return cameraAngle1[prop] === cameraAngle2[prop];
	});
};

export default View3DModule;
