import ToolbarComponent from "@esko/toolbar/ToolbarComponent";
import ButtonGroup from "@esko/toolbar/ButtonGroup";
import ViewerToolbarButton from "@esko/web-viewer/layout/component/ViewerToolbarButton";
import LayoutConstants from "@esko/web-viewer/layout/LayoutConstants";
import CompareConstants from "@esko/web-viewer/compare/CompareConstants";
import $ from "jquery";
import _ from "underscore";
/**
 * @exports View3DToolbarComponent
 * @memberof View3DModule
 * @constructor
 * @param {Core.ViewerFacade} f - Viewer facade
 * @param {View3DModule} viewController
 */
var View3DToolbarComponent = function(f, viewController) {
	ToolbarComponent.call(this, LayoutConstants.ToolbarComponent.VIEW3D, f);

	this.f = f;

	this._viewController = viewController;

	var view3DBtnGroup = new ButtonGroup("View3DGroup");
	view3DBtnGroup.setTitle(this.f.i18n, "viewer", "Toggle2D3DBtnGroupTitle");
	this.addButtonGroup(view3DBtnGroup);

	this._to2D = new ViewerToolbarButton(
		this.f,
		"to2DButton",
		"icon icon-WCR_Viewer-View-2D-min",
		false,
		_.partial(this._toggleLayoutMode, LayoutConstants.LayoutMode.VIEW2D).bind(this)
	);
	view3DBtnGroup.addButton(this._to2D);

	this._to3D = new ViewerToolbarButton(
		this.f,
		"to3DButton",
		"icon icon-WCR_Viewer-View-3D-min",
		false,
		_.partial(this._toggleLayoutMode, LayoutConstants.LayoutMode.DESKTOP_3D).bind(this)
	);
	view3DBtnGroup.addButton(this._to3D);

	// Function that will activate/deactivate the 2D/3D buttons
	var updateButtonActive = function() {
		var compareModule = this.f.getModule("compare");
		var mode;
		if (compareModule) {
			mode = compareModule.getActiveMode();
		}
		if (mode) {
			mode = mode.getName();
		} else {
			mode = this.f.getLayoutMode().getName();
		}
		// Layout or compare modes in which the 2D button should be active
		var twoDActive = [LayoutConstants.LayoutMode.VIEW2D, LayoutConstants.LayoutMode.DESKTOP_2D3D];

		twoDActive = _.union(twoDActive, _.keys(CompareConstants.modes2D));

		// Layout or compare modes in which the 3D button should be active
		var threeDActive = [LayoutConstants.LayoutMode.DESKTOP_3D, LayoutConstants.LayoutMode.DESKTOP_2D3D];
		threeDActive = _.union(threeDActive, _.keys(CompareConstants.modes3D));

		this._to2D.setActive(
			_.find(twoDActive, function(availableMode) {
				return mode === availableMode;
			}) != null
		);
		this._to3D.setActive(
			_.find(threeDActive, function(availableMode) {
				return mode === availableMode;
			}) != null
		);
	}.bind(this);

	// Function that will enable/disable the 2D/3D buttons
	var updateButtonEnabled = function() {
		var activePage = this.f.getDocumentController().getActivePage();
		var diffPage = this.f.getDocumentController().getDifferencePage();

		var compareMode;
		var is3DCompareModeAvailable = true;
		var is2DCompareModeAvailable = true;

		var compareModule = this.f.getModule("compare");
		if (compareModule) {
			var availableCompareModes = activePage.getAvailableCompareModes(diffPage);
			compareMode = compareModule.getActiveMode() && compareModule.getActiveMode().getName();
			is3DCompareModeAvailable =
				!CompareConstants.modes2D[compareMode] ||
				(CompareConstants.map2Dto3D[compareMode] && availableCompareModes.indexOf(CompareConstants.map2Dto3D[compareMode]) !== -1);
			is2DCompareModeAvailable =
				!CompareConstants.modes3D[compareMode] ||
				(CompareConstants.map3Dto2D[compareMode] && availableCompareModes.indexOf(CompareConstants.map3Dto2D[compareMode]) !== -1);
		}

		this._to2D.setEnabled(!activePage.is3D() && (!diffPage || !diffPage.is3D()) && is2DCompareModeAvailable);
		this._to3D.setEnabled(activePage.has3D() && (!diffPage || diffPage.has3D()) && is3DCompareModeAvailable);
	}.bind(this);

	this.f.on("layout.*.changed.layoutMode", updateButtonEnabled);
	this.f.on("module.compare.changed.mode", updateButtonEnabled);
	this.f.getDocumentController().whenActiveOrDifferencePageChanged(updateButtonEnabled);
	this.f.on("doc.*.changed.linkedDocument.*", updateButtonEnabled);
	this.on("changed.visibility", updateButtonEnabled);
	this.f.getDocumentController().whenActiveOrDifferencePageChanged(updateButtonActive);
	this.f.on("layout.*.changed.layoutMode", updateButtonActive);
	this.f.on("module.compare.changed.mode", updateButtonActive);
	this.on("changed.visibility", updateButtonActive);
	updateButtonActive();
	updateButtonEnabled();
};

View3DToolbarComponent.prototype = Object.create(ToolbarComponent.prototype);
View3DToolbarComponent.prototype.constructor = View3DToolbarComponent;
View3DToolbarComponent.toolbarName = LayoutConstants.ToolbarComponent.VIEW3D;

/**
 * This is called when the user wants to toggle between 2D/3D in center pane.
 *
 * @param {String} layout the layout you which to toggle
 */
View3DToolbarComponent.prototype._toggleLayoutMode = function(layout) {
	var currentLayout = this.f.getLayoutMode().getName();
	var compareModule = this.f.getModule("compare");
	var currentDiffMode;
	if (compareModule) {
		currentDiffMode = compareModule.getActiveMode();
	}
	if (currentDiffMode != null) {
		this._toggleLayoutModeBasedOnDiffMode(currentDiffMode.getName(), layout);
	} else {
		this._toggleLayoutModeBasedOnLayout(currentLayout, layout);
	}
};

/**
 * toggle layout mode based on the current layout (this fn is called if no compare mode is active)
 */
View3DToolbarComponent.prototype._toggleLayoutModeBasedOnLayout = function(currentLayout, layout) {
	var newLayout;
	if (currentLayout === LayoutConstants.LayoutMode.DESKTOP_2D3D && layout === LayoutConstants.LayoutMode.VIEW2D) {
		// currently in 2D3D layout and 2D is toggled -> go to 3D
		newLayout = LayoutConstants.LayoutMode.DESKTOP_3D;
	} else if (currentLayout === LayoutConstants.LayoutMode.DESKTOP_2D3D && layout === LayoutConstants.LayoutMode.DESKTOP_3D) {
		// currently in 2D3D layout and 3D is toggled -> go to 2D
		newLayout = LayoutConstants.LayoutMode.VIEW2D;
	} else if (currentLayout !== layout) {
		// if you are in basic or 3D and toggle the other go to 2D3D
		newLayout = LayoutConstants.LayoutMode.DESKTOP_2D3D;
	} else {
		// If you deselect the only active mode, switch to the other one.
		newLayout = layout === LayoutConstants.LayoutMode.VIEW2D ? LayoutConstants.LayoutMode.DESKTOP_3D : LayoutConstants.LayoutMode.VIEW2D;
	}
	if (newLayout !== undefined) {
		this.f.emit("layout.switchLayoutMode", newLayout);
	}
};

/**
 * toggle layout mode based on the current compare mode
 */
View3DToolbarComponent.prototype._toggleLayoutModeBasedOnDiffMode = function(diffMode, layout) {
	var activePage = this.f.getDocumentController().getActivePage();
	var diffPage = this.f.getDocumentController().getDifferencePage();
	var availableCompareModes = _.intersection(activePage.getAvailableCompareModes(diffPage), diffPage.getAvailableCompareModes(activePage));

	if (CompareConstants.map2Dto3D[diffMode] && layout === LayoutConstants.LayoutMode.DESKTOP_3D) {
		var name = CompareConstants.map2Dto3D[diffMode];
		if (availableCompareModes.indexOf(name) !== -1) {
			this.f.getModule("compare").activateModeByName(name);
		}
	} else if (CompareConstants.map3Dto2D[diffMode] && layout === LayoutConstants.LayoutMode.VIEW2D) {
		var name = CompareConstants.map3Dto2D[diffMode];

		if (availableCompareModes.indexOf(name) !== -1) {
			this.f.getModule("compare").activateModeByName(name);
		}
	}
};

export default View3DToolbarComponent;
