import PrivateChannel from "@esko/web-viewer/core/PrivateChannel";
import ViewerToolbarButton from "@esko/web-viewer/layout/component/ViewerToolbarButton";
import ColorPicker from "@esko/web-viewer/layout/component/ColorPicker";
import ListenerUtil from "@esko/web-viewer/utils/ListenerUtil";
import ViewerUtils from "@esko/web-viewer/utils/ViewerUtils";
import $ from "jquery";
import _ from "underscore";
/**
 * @exports BackgroundColorSelector
 * @memberof ViewControlsModule
 * @class The BackgroundColorSelector allows to change the image and window background color via color pickers from a tool bar when clicking the change background color button.
 * @constructor
 * @param {Core.ViewerFacade} f
 * @param {Function} Callback called when change background button is clicked
 */
var BackgroundColorSelector = function(f, viewControlsModel) {
	ListenerUtil.mixin(this, f);
	PrivateChannel.installTo(this, f, "backgroundColorSelector");

	this.f = f;
	this._viewControlsModel = viewControlsModel;

	/**
	 * @member {ViewerToolbarButton}
	 */
	this._changeBackgroundButton = new ViewerToolbarButton(
		this.f,
		"changeBackgroundButton",
		"changeBackgroundColor",
		true,
		_.bind(this._changeBackgroundButtonClicked, this)
	);
	this._changeBackgroundButton.setTooltip("viewer", "ToolBackgroundColorTT");

	/**
	 * @member {JQueryObj}
	 */
	this._toolbar = $("<div />").addClass("toolbar");
	/**
	 * @member {JQueryObj}
	 */
	this._toolbarTitle = null;

	/**
	 * @member {ColorPicker}
	 */
	this._changeImageBackgroundColorPicker = null;
	/**
	 * @member {ViewerToolbarButton}
	 */
	this._changeImageBackgroundColorButton = null;

	/**
	 * @member {ColorPicker}
	 */
	this._changeWindowBackgroundColorPicker = null;
	/**
	 * @member {ViewerToolbarButton}
	 */
	this._changeWindowBackgroundColorButton = null;

	/**
	 * @member {ColorPicker}
	 */
	this._changeBothBackgroundColorPicker = null;
	/**
	 * @member {ViewerToolbarButton}
	 */
	this._changeBothBackgroundColorButton = null;

	/**
	 * @member {ViewerToolbarButton}
	 */
	this._resetBackgroundColorButton = null;

	this._registerListener("changed.backgroundColorActive", this._onBackgroundColorActiveChanged, this, this._viewControlsModel);
	this._onBackgroundColorActiveChanged();
};

/*
 * INIT
 */

/**
 * Init the toolbar components
 */
BackgroundColorSelector.prototype._initToolbarComponents = function() {
	// Init title
	if (this._toolbarTitle === null) {
		this._toolbarTitle = $("<p />").addClass("toolbarTitle");

		this.f.i18n.loadBundle(
			"viewer",
			function() {
				this._toolbarTitle.html(this.f.i18n.t("BackgroundTitle"));
			},
			this
		);

		this._toolbar.append(this._toolbarTitle);
	}

	// Init color pickers
	if (this._changeImageBackgroundColorPicker === null) {
		this._changeImageBackgroundColorButton = new ViewerToolbarButton(
			this.f,
			"changeImageBackgroundColorButton",
			"icon-WCR_Color-Img",
			true,
			_.bind(function() {}, this)
		);
		this._changeImageBackgroundColorButton.setTooltip("viewer", "BackgroundImage");

		this._changeImageBackgroundColorPicker = new ColorPicker(
			this.f,
			_.bind(this._changeImageBackgroundColor, this),
			this._changeImageBackgroundColorButton.element
		);

		this._toolbar.append(this._changeImageBackgroundColorButton.element);

		// initial update and further updates
		this._registerListener("changed.imageBackgroundColor", this._onImageBackgroundColorChanged, this, this._viewControlsModel);
		this._updateImageBackgroundColorPicker();
	}
	if (this._changeWindowBackgroundColorPicker === null) {
		this._changeWindowBackgroundColorButton = new ViewerToolbarButton(
			this.f,
			"changeWindowBackgroundColorButton",
			"icon-WCR_Color-Bg",
			true,
			_.bind(function() {}, this)
		);
		this._changeWindowBackgroundColorButton.setTooltip("viewer", "BackgroundWindow");

		this._changeWindowBackgroundColorPicker = new ColorPicker(
			this.f,
			_.bind(this._changeWindowBackgroundColor, this),
			this._changeWindowBackgroundColorButton.element
		);

		this._toolbar.append(this._changeWindowBackgroundColorButton.element);

		// initial update and further updates
		this._registerListener("changed.windowBackgroundColor", this._onWindowBackgroundColorChanged, this, this._viewControlsModel);
		this._updateWindowBackgroundColorPicker();
	}
	if (this._changeBothBackgroundColorPicker === null) {
		// icon
		this._changeBothBackgroundColorButtonImageIcon = $("<i/>")
			.addClass("icon subIcon")
			.addClass("icon-WCR_Color-Img");
		this._changeBothBackgroundColorButtonWindowIcon = $("<i/>")
			.addClass("icon containerIcon")
			.addClass("icon-WCR_Color-Bg");
		this._changeBothBackgroundColorButtonWindowIcon.append(this._changeBothBackgroundColorButtonImageIcon);

		this._changeBothBackgroundColorButton = new ViewerToolbarButton(
			this.f,
			"changeBothBackgroundColorButton",
			this._changeBothBackgroundColorButtonWindowIcon,
			true,
			_.bind(function() {}, this)
		);
		this._changeBothBackgroundColorButton.setTooltip("viewer", "BackgroundBoth");

		var imageHex, windowHex;
		this._changeBothBackgroundColorPicker = new ColorPicker(
			this.f,
			_.bind(this._changeBothBackgroundColor, this),
			this._changeBothBackgroundColorButton.element,
			{
				show: function() {
					imageHex = this._viewControlsModel.getImageBackgroundColor();
					windowHex = this._viewControlsModel.getWindowBackgroundColor();
				}.bind(this),
				revert: function() {
					this._viewControlsModel.setImageBackgroundColor(imageHex);
					this._viewControlsModel.setWindowBackgroundColor(windowHex);
				}.bind(this)
			}
		);

		this._toolbar.append(this._changeBothBackgroundColorButton.element);

		// initial update, further updates: see image/background changed methods
		this._updateBothBackgroundColorPicker();
	}
	if (this._resetBackgroundColorButton === null) {
		this._resetBackgroundColorButton = new ViewerToolbarButton(
			this.f,
			"resetBackgroundColor",
			"icon icon-WCR_WF-Undo",
			true,
			_.bind(this._resetBackgroundColor, this)
		);

		this._resetBackgroundColorButton.setTooltip("viewer", "BackgroundReset");

		this._toolbar.append(this._resetBackgroundColorButton.element);

		// initial update
		this._updateResetBackgroundButton();
	}
};

/*
 * RENDERING
 */

/**
 * Re-render the main change background button
 */
BackgroundColorSelector.prototype._updateChangeBackgroundButton = function() {
	// set state of button
	this._changeBackgroundButton.setActive(this._viewControlsModel.isBackgroundColorActive());
};

/**
 * Re-render the color picker of image background
 */
BackgroundColorSelector.prototype._updateImageBackgroundColorPicker = function() {
	if (this._changeImageBackgroundColorPicker !== null) {
		// color picker
		var hexColor = ViewerUtils.hexFromInt(this._viewControlsModel.getImageBackgroundColor());
		this._changeImageBackgroundColorPicker.setColor(hexColor);

		// button
		this._changeImageBackgroundColorButton.element.css({
			color: "#" + hexColor
		});
	}
};

/**
 * Re-render the color picker of window background
 */
BackgroundColorSelector.prototype._updateWindowBackgroundColorPicker = function() {
	if (this._changeWindowBackgroundColorPicker !== null) {
		// color picker
		var hexColor = ViewerUtils.hexFromInt(this._viewControlsModel.getWindowBackgroundColor());
		this._changeWindowBackgroundColorPicker.setColor(hexColor);

		// button
		this._changeWindowBackgroundColorButton.element.css({
			color: "#" + hexColor
		});
	}
};

/**
 * Re-render the color picker of both background colors
 */
BackgroundColorSelector.prototype._updateBothBackgroundColorPicker = function() {
	if (this._changeBothBackgroundColorPicker !== null) {
		// color picker
		var hexImageColor = ViewerUtils.hexFromInt(this._viewControlsModel.getImageBackgroundColor());
		var hexWindowColor = ViewerUtils.hexFromInt(this._viewControlsModel.getWindowBackgroundColor());

		// by default, take color of window
		this._changeBothBackgroundColorPicker.setColor(hexWindowColor);

		// button
		this._changeBothBackgroundColorButtonImageIcon.css({
			color: "#" + hexImageColor
		});
		this._changeBothBackgroundColorButtonWindowIcon.css({
			color: "#" + hexWindowColor
		});
	}
};

/**
 * Re-render the reset background button
 */
BackgroundColorSelector.prototype._updateResetBackgroundButton = function() {
	if (this._resetBackgroundColorButton !== null) {
		var isDirty = !this._viewControlsModel.isImageBackgroundColorDefault() || !this._viewControlsModel.isWindowBackgroundColorDefault();
		this._resetBackgroundColorButton.setEnabled(isDirty);
	}
};

/*
 * CONTROLLING
 */

/**
 * Called when the changeBackgroundColorButton is clicked.
 */
BackgroundColorSelector.prototype._changeBackgroundButtonClicked = function() {
	// just toggle the button
	this._toggleActive();
};

/**
 * Toggle the active state of this selector
 * @param {Boolean} [toggle] - If defined, take that active state, otherwise toggle the current state.
 */
BackgroundColorSelector.prototype._toggleActive = function(toggle) {
	if (toggle === undefined) {
		// if undefined, toggle the active state
		toggle = !this._viewControlsModel.isBackgroundColorActive();
	}

	this._viewControlsModel.setBackgroundColorActive(toggle);
};

/**
 * Called when the image background color changed in the model
 */
BackgroundColorSelector.prototype._onImageBackgroundColorChanged = function() {
	this._updateImageBackgroundColorPicker();
	this._updateBothBackgroundColorPicker();
	this._updateResetBackgroundButton();
};

/**
 * Called when the window background color changed in the model
 */
BackgroundColorSelector.prototype._onWindowBackgroundColorChanged = function() {
	this._updateWindowBackgroundColorPicker();
	this._updateBothBackgroundColorPicker();
	this._updateResetBackgroundButton();
};

/**
 * Called on the active state change of the background color
 */
BackgroundColorSelector.prototype._onBackgroundColorActiveChanged = function() {
	this._updateChangeBackgroundButton();

	var active = this._viewControlsModel.isBackgroundColorActive();
	if (active) {
		// we need to show toolbar

		// init if needed
		this._initToolbarComponents();
		this._toolbar.show();
	} else {
		// we need to hide toolbar

		this._toolbar.hide();
	}
	this._emit("resized");
};

/**
 * Called when color picker changed color for image background
 * @param {String} hexColor
 */
BackgroundColorSelector.prototype._changeImageBackgroundColor = function(hexColor) {
	var intColor = ViewerUtils.intFromHex(hexColor);
	this._viewControlsModel.setImageBackgroundColor(intColor);
};

/**
 * Called when color picker changed color for window background
 * @param {String} hexColor
 */
BackgroundColorSelector.prototype._changeWindowBackgroundColor = function(hexColor) {
	var intColor = ViewerUtils.intFromHex(hexColor);
	this._viewControlsModel.setWindowBackgroundColor(intColor);
};

/**
 * Called when color picker changed color for both image and window background
 * @param {String} hexColor
 */
BackgroundColorSelector.prototype._changeBothBackgroundColor = function(hexColor) {
	// update both settings
	this._changeImageBackgroundColor(hexColor);
	this._changeWindowBackgroundColor(hexColor);
};

/**
 * Reset the background color of image and window to its original values
 */
BackgroundColorSelector.prototype._resetBackgroundColor = function() {
	this._viewControlsModel.resetBackgroundColors();
	// switch off after reset
	this._viewControlsModel.setBackgroundColorActive(false);
};

/*
 * Public methods
 */

/**
 * Get the button element to change the background color
 * @returns {JQueryObj}
 */
BackgroundColorSelector.prototype.getButtonElement = function() {
	return this._changeBackgroundButton.element;
};

/**
 * Get the toolbar
 * @returns {JQueryObj}
 */
BackgroundColorSelector.prototype.getToolbar = function() {
	return this._toolbar;
};

export default BackgroundColorSelector;
