import PrivateChannel from "@esko/web-viewer/core/PrivateChannel";
import ViewerToolbarButton from "@esko/web-viewer/layout/component/ViewerToolbarButton";
import ColorPicker from "@esko/web-viewer/layout/component/ColorPicker";
import ListenerUtil from "@esko/web-viewer/utils/ListenerUtil";
import ViewerUtils from "@esko/web-viewer/utils/ViewerUtils";
import Mousetrap from "@esko/mousetrap";
import $ from "jquery";
import _ from "underscore";
/**
 * @exports OverprintHighlightColorSelector
 * @memberof ViewControlsModule
 * @class The OverprintHighlightColorSelector allows to change the overprint highlight color from a tool bar when clicking the set overprint highlight color button.
 * It also listens to CTRL+H to switch between non-highlight and highlight color view.
 * @constructor
 * @param {Core.ViewerFacade} f
 * @param {ViewControlsModel} viewControlsModel - The view controls model to interact with.
 */
var OverprintHighlightColorSelector = function(f, viewControlsModel) {
	ListenerUtil.mixin(this, f);
	PrivateChannel.installTo(this, f, "overprintHighlightColorSelector");

	this.f = f;
	this._viewControlsModel = viewControlsModel;

	/**
	 * @member {ViewerToolbarButton}
	 */
	this._setOverprintHighlightColorButton = new ViewerToolbarButton(
		this.f,
		"setOverprintHighlightColor",
		"setOverprintHighlightColorOff",
		true,
		_.bind(this._setOverprintHighlightColorButtonClicked, this)
	);

	/**
	 * @member {JQueryObj}
	 */
	this._toolbar = $("<div />").addClass("toolbar");
	/**
	 * @member {JQueryObj}
	 */
	this._toolbarTitle = null;

	/**
	 * @member {ColorPicker}
	 */
	this._changeOverprintHighlightColorPicker = null;
	/**
	 * @member {ViewerToolbarButton}
	 */
	this._changeOverprintHighlightColorButton = null;

	// Keyboard short cuts
	this._listeningToShortcuts = false;
	this._onCtrlHBound = _.bind(function(e) {
		this._onCtrlH();
		e.preventDefault();
		e.stopPropagation();
	}, this);
	this._listenToShortcuts(true);

	// Update initial state and listen to future changes
	this._registerListener(
		"changed.overprintHighlightColorActive",
		this._onOverprintHighlightColorActiveChanged,
		this,
		this._viewControlsModel
	);
	this._onOverprintHighlightColorActiveChanged();
};

/*
 * INIT
 */

/**
 * Init the toolbar components
 */
OverprintHighlightColorSelector.prototype._initToolbarComponents = function() {
	// Init title
	if (this._toolbarTitle === null) {
		this._toolbarTitle = $("<p />").addClass("toolbarTitle");

		this.f.i18n.loadBundle(
			"viewer",
			function() {
				this._toolbarTitle.html(this.f.i18n.t("OverprintTitle"));
			},
			this
		);

		this._toolbar.append(this._toolbarTitle);
	}

	// Init color picker
	if (this._changeOverprintHighlightColorPicker === null) {
		this._changeOverprintHighlightColorPicker = new ColorPicker(this.f, _.bind(this._changeOverprintHighlightColor, this));
		this._changeOverprintHighlightColorPicker.element.addClass("icon icon-normal");

		this._changeOverprintHighlightColorButton = new ViewerToolbarButton(
			this.f,
			"changeOverprintHighlightColorButton",
			this._changeOverprintHighlightColorPicker.element,
			true,
			_.bind(function() {
				this._changeOverprintHighlightColorPicker.show();
			}, this)
		);
		this._changeOverprintHighlightColorButton.setTooltip("viewer", "OverprintColorTT");

		this._registerListener(
			"changed.overprintHighlightColor",
			this._updateChangeOverprintHighlightColorPicker,
			this,
			this._viewControlsModel
		);

		this._toolbar.append(this._changeOverprintHighlightColorButton.element);

		// initial update
		this._updateChangeOverprintHighlightColorPicker();
	}
};

/*
 * RENDERING
 */
/**
 * Re-render the set overprint highlight color button
 */
OverprintHighlightColorSelector.prototype._updateSetOverprintHighlightColorButton = function() {
	var active = this._viewControlsModel.isOverprintHighlightColorActive();
	if (active) {
		this._setOverprintHighlightColorButton.setIcon("setOverprintHighlightColorOn");
		this._setOverprintHighlightColorButton.setTooltip("viewer", "OverprintTT2");
	} else {
		this._setOverprintHighlightColorButton.setIcon("setOverprintHighlightColorOff");
		this._setOverprintHighlightColorButton.setTooltip("viewer", "OverprintTT");
	}
	this._setOverprintHighlightColorButton.setActive(active);
};

/**
 * Re-render the color picker of overprint highlight color
 */
OverprintHighlightColorSelector.prototype._updateChangeOverprintHighlightColorPicker = function() {
	if (this._changeOverprintHighlightColorPicker !== null) {
		// color picker
		var overprintHighlightColor = this._viewControlsModel.getOverprintHighlightColor();
		var hexColor = ViewerUtils.hexFromInt(overprintHighlightColor);
		this._changeOverprintHighlightColorPicker.setColor(hexColor);
	}
};

/*
 * CONTROLLING
 */
/**
 * Called when the setOverprintHighlightColorButton is clicked.
 */
OverprintHighlightColorSelector.prototype._setOverprintHighlightColorButtonClicked = function() {
	// toggle active mode
	this._toggleActive();
};

/**
 * Called when color picker changed color for overprint highlight color
 * @param {String} hexColor
 */
OverprintHighlightColorSelector.prototype._changeOverprintHighlightColor = function(hexColor) {
	var intColor = ViewerUtils.intFromHex(hexColor);
	this._viewControlsModel.setOverprintHighlightColor(intColor);
};

/**
 * Toggle the active state of this selector
 * @param {Boolean} [toggle] - If defined, take that active state, otherwise toggle the current state.
 */
OverprintHighlightColorSelector.prototype._toggleActive = function(toggle) {
	if (toggle === undefined) {
		// if undefined, toggle the active state
		toggle = !this._viewControlsModel.isOverprintHighlightColorActive();
	}

	this._viewControlsModel.setOverprintHighlightColorActive(toggle);
};

/**
 * Called when the overprint highlight color active state changes in the model.
 */
OverprintHighlightColorSelector.prototype._onOverprintHighlightColorActiveChanged = function() {
	this._updateSetOverprintHighlightColorButton();

	var active = this._viewControlsModel.isOverprintHighlightColorActive();
	if (active) {
		// init if needed
		this._initToolbarComponents();
		this._toolbar.show();

		// show the color picker if default color (no color set in past)
		if (this._viewControlsModel.isOverprintHighlightColorDefault()) {
			this._changeOverprintHighlightColorPicker.show();
		}
	} else {
		this._toolbar.hide();
	}
	this._emit("resized");
};

/**
 * Called on key combo CTRL+H
 */
OverprintHighlightColorSelector.prototype._onCtrlH = function() {
	this._toggleActive();
};

/**
 * Start/stop listening to short cuts
 * @param {Boolean} listen - True to listen to shortcuts, false to stop listening
 */
OverprintHighlightColorSelector.prototype._listenToShortcuts = function(listen) {
	if (listen) {
		if (!this._listeningToShortcuts) {
			// if we were not listening yet
			Mousetrap.bind("ctrl+h", this._onCtrlHBound, "keydown");
			this._listeningToShortcuts = true;
		}
	} else {
		if (this._listeningToShortcuts) {
			// if we were listening
			Mousetrap.unbind("ctrl+h", this._onCtrlHBound, "keydown");
			this._listeningToShortcuts = false;
		}
	}
};

/*
 * Public methods
 */

/**
 * Get the button element to set the overprint highlight color
 * @returns {JQueryObj}
 */
OverprintHighlightColorSelector.prototype.getButtonElement = function() {
	return this._setOverprintHighlightColorButton.element;
};

/**
 * Get the toolbar
 * @returns {JQueryObj}
 */
OverprintHighlightColorSelector.prototype.getToolbar = function() {
	return this._toolbar;
};

/**
 * Show/hide the overprint highlight color selector
 * @param {Boolean} toggle - True to show the button, false to hide and disable it
 */
OverprintHighlightColorSelector.prototype.toggle = function(toggle) {
	if (toggle) {
		this._listenToShortcuts(true);
		this._setOverprintHighlightColorButton.show();
	} else {
		this._listenToShortcuts(false);
		this._toggleActive(false);
		this._setOverprintHighlightColorButton.hide();
	}
};

export default OverprintHighlightColorSelector;
