import DropdownToolbarButton from "@esko/web-viewer/layout/component/DropdownToolbarButton";
import ListenerUtil from "@esko/web-viewer/utils/ListenerUtil";
import ColorPicker from "@esko/web-viewer/layout/component/ColorPicker";
import DoubleColorBlock from "@esko/web-viewer/layout/component/DoubleColorBlock";
import ViewControlsConstants from "@esko/web-viewer/viewControls/ViewControlsConstants";
import _ from "underscore";
/**
 * @exports TrimboxDropdownToolbarButton
 * @memberof ViewControlsModule
 * @class Extends the dropdownToolbarButton, makes is possible to show/hide the trimbox of the document page and the dropdown lets the user change
 * the color of the trimbox for the active document page and the reference document page
 * @augments {DropdownToolbarButton}
 * @constructor
 * @param {Core.ViewerFacade} f
 * @param {String} name - the name of the button
 * @param {String | JQueryObj } icon - The class which is used for displaying the icon (css class for 16px * 16px icon), or your own jQuery element
 * @param {Boolean} isShowingTrimbox - does the button needs to be activated by default
 * @param {ViewControlsConstants.trimboxMode} trimboxMode - the trimbox mode to initialize the dropdown with
 */
var TrimboxDropdownToolbarButton = function(f, name, icon, isShowingTrimbox, trimboxMode) {
	DropdownToolbarButton.call(this, f, name, icon, false, this._toggleTrimbox);
	ListenerUtil.mixin(this, f);

	/**
	 * @member {Core.ViewerFacade}
	 */
	this.f = f;

	//set the button to active if it needs to be active by default
	this.setActive(isShowingTrimbox);

	/**
	 * @member {ViewControlsConstants.trimboxMode}
	 * @private
	 */
	this._trimboxMode = trimboxMode;

	/**
	 * @member {ColorPicker} - the color picker for the color of the trimbox of the active document page
	 * @private
	 */
	this._trimboxColorPicker = new ColorPicker(f, _.bind(this._changeTrimboxColor, this));
	this._trimboxColorPicker.setColor(this.f.getCurrentUser().getColor("PREFERENCES_HIGHLIGHTCOLORS_MARGINS")); //get the color from the user settings (is persisted)

	/**
	 * @member {ColorPicker} -  the color picker to change the color of the trimbox of the reference document page
	 * @private
	 */
	this._secondaryTrimboxColorPicker = new ColorPicker(f, _.bind(this._changeSecondaryTrimboxColor, this));
	this._secondaryTrimboxColorPicker.setColor(this.f.getCurrentUser().getColor("PREFERENCES_HIGHLIGHTCOLORS_MARGINS_DIFF")); //get the color from the user settings

	/**
	 * @member {DoubleColorBlock} makes it possible to change both the color of the active document pages' trimbox as the trimbox of the reference pages' trimbox
	 * @private
	 */
	this._changeBothTrimboxColorBlock = new DoubleColorBlock(
		this.f.getCurrentUser().getColor("PREFERENCES_HIGHLIGHTCOLORS_MARGINS"),
		this.f.getCurrentUser().getColor("PREFERENCES_HIGHLIGHTCOLORS_MARGINS_DIFF")
	);

	/**
	 * @member {ColorPicker} the color picker to change both colors
	 */
	this._changeBothTrimboxColor = new ColorPicker(f, _.bind(this._changeBothTrimboxColor, this), this._changeBothTrimboxColorBlock.element);
	this._changeBothTrimboxColor.setColor(this.f.getCurrentUser().getColor("PREFERENCES_HIGHLIGHTCOLORS_MARGINS")); //set the color from the active document page as the default selected color on opening the color picker
	//make sure the language bundle is loaded before initializing the dropdown menu
	this.f.loadBundle(
		"viewer",
		function() {
			this._initializeDropdown();
		},
		this
	);
};

//inherit
TrimboxDropdownToolbarButton.prototype = Object.create(DropdownToolbarButton.prototype);
TrimboxDropdownToolbarButton.prototype.constructor = TrimboxDropdownToolbarButton;

/**
 * Toggle the show trimbox button, sets the trimbox mode to the graphics module
 * @private
 * @param {Boolean} toggle - show or hide the trimbox
 */
TrimboxDropdownToolbarButton.prototype._toggleTrimbox = function(toggle) {
	var graphicsModule = this.f.getModule("graphics"); //get the graphics module, the setting is set to this module
	this.setActive(toggle); //set the button active/not-active
	if (graphicsModule !== undefined) {
		//only if the graphicsmodule is available
		graphicsModule.setShowTrimbox(toggle, this._trimboxMode);
	} else {
		//wait for the module to be loaded if this was not yet the case
		this.f.whenModuleAvailable(
			"graphics",
			_.bind(function(graphicsModule) {
				graphicsModule.setShowTrimbox(this.isActive(), this._trimboxMode);
			}, this)
		);
	}
	//also set the settings of the users to persist the trimbox settings
	this.f.setOption("PREFERENCES_DISPLAY_MARGINS", toggle ? 1 : 0);
	this.f.setOption("PREFERENCES_MARGIN_DISPLAY_MODE", this._trimboxMode);
};

/**
 * initializes the dropdown menu
 * @private
 */
TrimboxDropdownToolbarButton.prototype._initializeDropdown = function() {
	this._initializeShowTrimboxOptions(); //first add the trimbox options
	this.addDivider(); //add a divider
	this._initializeColorOptions(); //add the color options for the trimbox
};

/**
 * initializes the trimbox options in the dropdown menu
 * @private
 */
TrimboxDropdownToolbarButton.prototype._initializeShowTrimboxOptions = function() {
	this.addOption({
		//add as options the user cannot uncheck, either one of the two options needs to be active
		name: "onCurrentPage", //the trimbox is only shown on the current page (only valid in reader spread mode)
		content: this.f.i18n.t("ToolMarginsShowCurrent"),
		onchange: _.bind(function() {
			this._trimboxMode = ViewControlsConstants.trimboxMode.ONCURRENTPAGE; //set to the correct trimbox mode
			this._toggleTrimbox(this.isActive()); //make sure to toggle the show trimbox to trigger the setting to be set in the graphics module
			this.toggleOption("bothPagesInReaderSpread", false); //toggle the other option
		}, this),
		selected: this._trimboxMode === ViewControlsConstants.trimboxMode.ONCURRENTPAGE, //should it be selected on initialization
		userCannotUncheck: true
	});

	this.addOption({
		name: "bothPagesInReaderSpread",
		content: this.f.i18n.t("ToolMarginsShowBoth"),
		onchange: _.bind(function() {
			this._trimboxMode = ViewControlsConstants.trimboxMode.ONBOTHINREADERSPREAD;
			this._toggleTrimbox(this.isActive());
			this.toggleOption("onCurrentPage", false);
		}, this),
		selected: this._trimboxMode === ViewControlsConstants.trimboxMode.ONBOTHINREADERSPREAD,
		userCannotUncheck: true
	});
};

/**
 * initializes the color options in the dropdown menu
 * @private
 */
TrimboxDropdownToolbarButton.prototype._initializeColorOptions = function() {
	this.addAction({
		//add as actions, and add the colorblock as icon in the dropdown menu
		name: "trimboxColor",
		content: this.f.i18n.t("ToolMarginsColorPrompt"),
		action: _.bind(function() {
			this._trimboxColorPicker.show(); //show the colorpicker on click
		}, this),
		enabled: true,
		icon: this._trimboxColorPicker.element
	});

	this.addAction({
		name: "secondaryTrimboxColor",
		content: this.f.i18n.t("ToolMarginsColorPromptDiff"),
		action: _.bind(function() {
			this._secondaryTrimboxColorPicker.show();
		}, this),
		enabled: true,
		icon: this._secondaryTrimboxColorPicker.element
	});

	this.addAction({
		name: "changeBothTrimboxColor",
		content: this.f.i18n.t("ToolMarginsColorPromptBoth"),
		action: _.bind(function() {
			this._changeBothTrimboxColor.show();
		}, this),
		enabled: true,
		icon: this._changeBothTrimboxColor.element
	});

	this.addAction({
		name: "revertTrimboxColors",
		content: this.f.i18n.t("ToolMarginsColorPromptReset"),
		action: _.bind(function() {
			this.f.getCurrentUser().resetOption("PREFERENCES_HIGHLIGHTCOLORS_MARGINS");
			this.f.getCurrentUser().resetOption("PREFERENCES_HIGHLIGHTCOLORS_MARGINS_DIFF");
			var activeColor = this.f.getCurrentUser().getColor("PREFERENCES_HIGHLIGHTCOLORS_MARGINS");
			var referenceColor = this.f.getCurrentUser().getColor("PREFERENCES_HIGHLIGHTCOLORS_MARGINS_DIFF");
			this._changeBothTrimboxColorBlock.setLeftColor(activeColor);
			this._changeBothTrimboxColorBlock.setRightColor(referenceColor);
			this._trimboxColorPicker.setColor(activeColor); //set the colors in the color pickers of the active and reference document pages
			this._secondaryTrimboxColorPicker.setColor(referenceColor);
		}, this),
		enabled: true,
		icon: "resetColor"
	});
};

/**
 * Change only the trimbox color of the active document page
 * @param {String} color - as RGB
 */
TrimboxDropdownToolbarButton.prototype._changeTrimboxColor = function(color) {
	this.f.getCurrentUser().setColor("PREFERENCES_HIGHLIGHTCOLORS_MARGINS", color);
	this._changeBothTrimboxColorBlock.setLeftColor("#" + color); //set the color to the double color block
};

/**
 * Change only the trimbox color of the reference document page
 * @param {String} color - as RGB
 */
TrimboxDropdownToolbarButton.prototype._changeSecondaryTrimboxColor = function(color) {
	this.f.getCurrentUser().setColor("PREFERENCES_HIGHLIGHTCOLORS_MARGINS_DIFF", color);
	this._changeBothTrimboxColorBlock.setRightColor("#" + color);
};

/**
 * Change only the trimbox color of both the active and the reference document page
 * @param {String} color - as RGB
 */
TrimboxDropdownToolbarButton.prototype._changeBothTrimboxColor = function(color) {
	this._changeTrimboxColor(color);
	this._changeSecondaryTrimboxColor(color);
	this._trimboxColorPicker.setColor(color); //set the colors in the color pickers of the active and reference document pages
	this._secondaryTrimboxColorPicker.setColor(color);
};

export default TrimboxDropdownToolbarButton;
