import _ from "underscore";
import PrivateChannel from "@esko/web-viewer/core/PrivateChannel";
import AbstractModel from "@esko/web-viewer/core/AbstractModel";
import ViewerUtils from "@esko/web-viewer/utils/ViewerUtils";
/**
 * @exports ViewControlsModel
 * @memberof ViewControlsModule
 * @class ViewControlsModule holds all the information for the view controls.
 * @constructor
 *
 * @param {Core.ViewerFacade} f - Viewer facade
 */
var ViewControlsModel = function(f) {
	AbstractModel.call(this, f, "ViewControlsModel");
	PrivateChannel.installTo(this, f, "viewControlsModel");

	/**
	 * @member {Number} - horizontal scale, only for digital film (if no scaling this will be 1)
	 */
	this._hscale = 1;

	/**
	 * @member {Number} - vertical scale, only for digital film (if no scaling this will be 1)
	 */
	this._vscale = 1;

	/**
	 * @member {Number} - horizontal distortion, only for digital film (if no distortion this will be 1)
	 */
	this._hdist = 1;

	/**
	 * @member {Number} - vertical distortion, only for digital film (if no distortion this will be 1)
	 */
	this._vdist = 1;

	/**
	 * @member {Number} - horizontal scale, only for digital film (if no scaling this will be 1), for reference page
	 */
	this._diffhscale = 1;

	/**
	 * @member {Number} - vertical scale, only for digital film (if no scaling this will be 1) for reference page
	 */
	this._diffvscale = 1;

	/**
	 * @member {Number} - horizontal distortion, only for digital film (if no distortion this will be 1) for reference page
	 */
	this._diffhdist = 1;

	/**
	 * @member {Number} - vertical distortion, only for digital film (if no distortion this will be 1) for reference page
	 */
	this._diffvdist = 1;

	/**
	 * @member {Boolean} is the distortion compensated or not
	 */
	this._isDistCompensated = false;

	/**
	 * @member {Boolean} is the scaling compensated or not
	 */
	this._isScaleCompensated = false;

	/**
	 * @member {Boolean} is the inverse view mode enabled
	 */
	this._isNegated = false;

	/**
	 * @member {Boolean} is view from back or not
	 */
	this._viewFromBack = false;

	/*
     * Overprint
     */
	/**
	 * @member {Boolean} is OverprintHighlightColor active or not
	 */
	this._overprintHighlightColorActive = false;
	/**
	 * @member {Number} Overprint highlight color (rgb int)
	 */
	this._overprintHighlightColor = this.f.getOption("PREFERENCES_HIGHLIGHTCOLORS");
	this.f.getCurrentUser().on(
		"options.changed.PREFERENCES_HIGHLIGHTCOLORS",
		_.bind(function(oldValue, newValue) {
			this.setOverprintHighlightColor(newValue);
		}, this)
	);

	/*
     * Background color 
     */
	/**
	 * @member {Boolean}
	 */
	this._backgroundColorActive = false;

	/**
	 * @member {Number}
	 */
	this._imageBackgroundColor = this.f.getCurrentUser().getOption("PREFERENCES_IMAGE_BACKGROUNDCOLOR");
	/**
	 * @member {Number}
	 */
	this._windowBackgroundColor = this.f.getCurrentUser().getOption("PREFERENCES_WINDOW_BACKGROUNDCOLOR");
};

ViewControlsModel.prototype = Object.create(AbstractModel.prototype);
ViewControlsModel.prototype.constructor = ViewControlsModel;

/**
 * Set the scale for the currently active digital film file
 * @param {String|Number} hscale - horizontal scale
 * @param {String|Number} vscale - vertical scale
 * @param {Boolean} forCurrentPage - for the current page
 */
ViewControlsModel.prototype.setScale = function(hscale, vscale, forCurrentPage) {
	if (hscale && vscale) {
		//check if both horizontal as vertical scale are defined
		if (forCurrentPage) {
			this._hscale = ViewerUtils.parseFloat(hscale); //parse the hscale and vscale
			this._vscale = ViewerUtils.parseFloat(vscale);
		} else {
			this._diffhscale = ViewerUtils.parseFloat(hscale); //parse the hscale and vscale
			this._diffvscale = ViewerUtils.parseFloat(vscale);
		}
	} else {
		//set to null if not both are defined
		if (forCurrentPage) {
			this._hscale = 1;
			this._vscale = 1;
		} else {
			this._diffhscale = 1;
			this._diffvscale = 1;
		}
	}
	this._emit("scaleOrDistortionChanged");
};

/**
 * Reset the scale for the currently active page (set to null), also reset the is compensating
 * @param {Boolean} forCurrentPage - for the current page
 */
ViewControlsModel.prototype.resetScale = function(forCurrentPage) {
	if (forCurrentPage) {
		this._hscale = 1;
		this._vscale = 1;
	} else {
		this._diffhscale = 1;
		this._diffvscale = 1;
	}
	this._isScaleCompensated = false;
	this._emit("scaleOrDistortionChanged");
};

/**
 * Set the distortion for the currently active digital film file
 * @param {String|Number} hdist - horizontal distortion
 * @param {String|Number} vdist - vertical distortion
 */
ViewControlsModel.prototype.setDistortion = function(hdist, vdist, forCurrentPage) {
	if (hdist && vdist) {
		if (forCurrentPage) {
			this._hdist = ViewerUtils.parseFloat(hdist);
			this._vdist = ViewerUtils.parseFloat(vdist);
		} else {
			this._diffhdist = ViewerUtils.parseFloat(hdist);
			this._diffvdist = ViewerUtils.parseFloat(vdist);
		}
	} else {
		if (forCurrentPage) {
			this._hdist = 1;
			this._vdist = 1;
		} else {
			this._diffhdist = 1;
			this._diffvdist = 1;
		}
	}
	this._emit("scaleOrDistortionChanged");
};

/**
 * Reset the distortion for the currently active page (set to null), also reset the is compensating
 */
ViewControlsModel.prototype.resetDistortion = function(forCurrentPage) {
	if (forCurrentPage) {
		this._hdist = 1;
		this._vdist = 1;
	} else {
		this._diffhdist = 1;
		this._diffvdist = 1;
	}
	this._isDistCompensated = false;
	this._emit("scaleOrDistortionChanged");
};

/**
 * Has the currently active page a distortion (vdist and hdist are not 100%)
 * @returns {Boolean}
 */
ViewControlsModel.prototype.hasPageDistortion = function() {
	return this._hdist !== undefined && this._vdist !== undefined && (this._hdist !== 1 || this._vdist !== 1);
};

/**
 * Has the reference page a distortion (vdist and hdist are not 100%)
 * @returns {Boolean}
 */
ViewControlsModel.prototype.hasDiffPageDistortion = function() {
	return this._diffhdist !== undefined && this._diffvdist !== undefined && (this._diffhdist !== 1 || this._diffvdist !== 1);
};

/**
 * Has the currently active page a scaling (vscale and hscale are not 100%)
 * @returns {Boolean}
 */
ViewControlsModel.prototype.hasPageScale = function() {
	return this._hscale !== undefined && this._vscale !== undefined && (this._hscale !== 1 || this._vscale !== 1);
};

/**
 * Has the reference page a scaling (vscale and hscale are not 100%)
 * @returns {Boolean}
 */
ViewControlsModel.prototype.hasDiffPageScale = function() {
	return this._diffhscale !== undefined && this._diffvscale !== undefined && (this._diffhscale !== 1 || this._diffvscale !== 1);
};

/**
 * Get the horizontal scale of the currently active page
 * @param {Boolean} forCurrentPage - for the current page
 * @returns {Number} - 1 is not scaling (or 100%)
 */
ViewControlsModel.prototype.getHorizontalScale = function(forCurrentPage) {
	if (forCurrentPage) {
		return this._hscale;
	} else {
		return this._diffhscale;
	}
};

/**
 * Get the vertical scale of the currently active page
 * @param {Boolean} forCurrentPage - for the current page
 * @returns {Number} - 1 is not scaling (or 100%)
 */
ViewControlsModel.prototype.getVerticalScale = function(forCurrentPage) {
	if (forCurrentPage) {
		return this._vscale;
	} else {
		return this._diffvscale;
	}
};

/**
 * Get the horizontal distortion of the currently active page
 * @param {Boolean} forCurrentPage - for the current page
 * @returns {Number} - 1 is not scaling (or 100%)
 */
ViewControlsModel.prototype.getHorizontalDistortion = function(forCurrentPage) {
	if (forCurrentPage) {
		return this._hdist;
	} else {
		return this._diffhdist;
	}
};

/**
 * Get the vertical distortion of the currently active page
 * @param {Boolean} forCurrentPage - for the current page
 * @returns {Number} - 1 is not scaling (or 100%)
 */
ViewControlsModel.prototype.getVerticalDistortion = function(forCurrentPage) {
	if (forCurrentPage) {
		return this._vdist;
	} else {
		return this._diffvdist;
	}
};

/**
 * Set if the scale is currently compensated or not
 * @param {Boolean} isCompensated
 */
ViewControlsModel.prototype.setIsScaleCompensated = function(isCompensated) {
	this._isScaleCompensated = isCompensated;
	this._emit("scaleOrDistortionChanged");
};

/**
 * Set if the distortion is currently compensated or not
 * @param {Boolean} isCompensated
 */
ViewControlsModel.prototype.setIsDistCompensated = function(isCompensated) {
	this._isDistCompensated = isCompensated;
	this._emit("scaleOrDistortionChanged");
};

/**
 * Enable/disable the inverse view option
 * @param {Boolean} isNegated
 */
ViewControlsModel.prototype.setIsNegated = function(isNegated) {
	var oldValue = this._isNegated;
	this._isNegated = isNegated;
	this._fireChange(false, ["negated"], oldValue, isNegated);
};

/**
 * Is the distortion compensated or not
 * @returns {Boolean}
 */
ViewControlsModel.prototype.isDistCompensated = function() {
	return this._isDistCompensated;
};

/**
 * Is the scaling compensated or not
 * @returns {Boolean}
 */
ViewControlsModel.prototype.isScaleCompensated = function() {
	return this._isScaleCompensated;
};

/**
 * Is the inverse view option enabled or not
 * @returns {Boolean}
 */
ViewControlsModel.prototype.isNegated = function() {
	return this._isNegated;
};

/**
 * Get the state of the viewControls
 * @returns {Object} state object
 */
ViewControlsModel.prototype.getState = function() {
	return {
		negated: this.isNegated(),
		viewFromBack: this.isViewFromBack(),
		overprintHighlightColorActive: this.isOverprintHighlightColorActive(),
		overprintHighlightColor: this.getOverprintHighlightColor(),
		backgroundColorActive: this.isBackgroundColorActive(),
		imageBackgroundColor: this.getImageBackgroundColor(),
		windowBackgroundColor: this.getWindowBackgroundColor()
	};
};

/**
 * Return the default state
 * @returns {Object}
 */
ViewControlsModel.prototype.getDefaultState = function() {
	// defaults
	return {
		negated: false,
		viewFromBack: false,
		overprintHighlightColorActive: false,
		overprintHighlightColor: this.f.getOption("PREFERENCES_HIGHLIGHTCOLORS"),
		backgroundColorActive: false,
		imageBackgroundColor: this.f.getOption("PREFERENCES_IMAGE_BACKGROUNDCOLOR"),
		windowBackgroundColor: this.f.getOption("PREFERENCES_WINDOW_BACKGROUNDCOLOR")
	};
};

/**
 * Set the state of the viewControls
 * @param {Object} state - state object (if undefined, default settings are loaded)
 */
ViewControlsModel.prototype.setState = function(state) {
	if (!state) {
		state = this.getDefaultState();
	}

	this.setIsNegated(state.negated);
	this.setViewFromBack(state.viewFromBack);
	// order matters: in viewcontrolsmodule, there is first an active check before appending URL parameters
	this.setOverprintHighlightColorActive(state.overprintHighlightColorActive);
	this.setOverprintHighlightColor(state.overprintHighlightColor);
	// order matters: in viewcontrolsmodule, there is first an active check before appending URL parameters
	this.setBackgroundColorActive(state.backgroundColorActive);
	this.setImageBackgroundColor(state.imageBackgroundColor);
	this.setWindowBackgroundColor(state.windowBackgroundColor);
};

/**
 * Set view from back or not
 * @param {Boolean} viewFromBack
 */
ViewControlsModel.prototype.setViewFromBack = function(viewFromBack) {
	var oldValue = this._viewFromBack;
	this._viewFromBack = viewFromBack;
	this._fireChange(false, ["viewFromBack"], oldValue, viewFromBack);
};

/**
 * Is view from back on or not
 * @returns {Boolean}
 */
ViewControlsModel.prototype.isViewFromBack = function() {
	return this._viewFromBack;
};

/**
 * Set the overprint high light color active or not
 * @param {Boolean} toggle
 */
ViewControlsModel.prototype.setOverprintHighlightColorActive = function(toggle) {
	var oldValue = this._overprintHighlightColorActive;
	this._overprintHighlightColorActive = toggle;
	this._fireChange(false, ["overprintHighlightColorActive"], oldValue, toggle);
};

/**
 * Is overprint highlight color active or not
 * @returns {Boolean} True if active, false if not active
 */
ViewControlsModel.prototype.isOverprintHighlightColorActive = function() {
	return this._overprintHighlightColorActive;
};

/**
 * Set the overprint highlight color
 * @param {Number} rgbInt - Color as RGB encoded number
 */
ViewControlsModel.prototype.setOverprintHighlightColor = function(rgbInt) {
	var oldValue = this._overprintHighlightColor;
	this._overprintHighlightColor = rgbInt;
	// save the setting
	this.f.setOption("PREFERENCES_HIGHLIGHTCOLORS", rgbInt);
	this._fireChange(false, ["overprintHighlightColor"], oldValue, rgbInt);
};

/**
 * Get the overprint highlight color
 * @returns {Number} The current overprint highlight color: rgb encoded number
 */
ViewControlsModel.prototype.getOverprintHighlightColor = function() {
	return this._overprintHighlightColor;
};

/**
 * Check whether the current overprint highlight color is the same as the default value
 * @returns {Boolean} True if default, false otherwise
 */
ViewControlsModel.prototype.isOverprintHighlightColorDefault = function() {
	return this._overprintHighlightColor === this.f.getCurrentUser().getOptionDefault("PREFERENCES_HIGHLIGHTCOLORS");
};

/**
 * Is Background color active or not
 * @returns {Boolean} True if active, false if not active
 */
ViewControlsModel.prototype.isBackgroundColorActive = function() {
	return this._backgroundColorActive;
};

/**
 * Set the background colors active or not
 * @param {Boolean} toggle - True for active, false for deactivation
 */
ViewControlsModel.prototype.setBackgroundColorActive = function(toggle) {
	var oldValue = this._backgroundColorActive;
	this._backgroundColorActive = toggle;
	this._fireChange(false, ["backgroundColorActive"], oldValue, toggle);
};

/**
 * Set the window background color
 * @param {Number} rgbInt - RGB number encoded color
 */
ViewControlsModel.prototype.setWindowBackgroundColor = function(rgbInt) {
	var oldValue = this._windowBackgroundColor;
	this._windowBackgroundColor = rgbInt;
	// save the setting
	this.f.setOption("PREFERENCES_WINDOW_BACKGROUNDCOLOR", rgbInt);
	this._fireChange(false, ["windowBackgroundColor"], oldValue, rgbInt);
};

/**
 * Get the window background color
 * @returns {Number} The current window background color: rgb encoded number
 */
ViewControlsModel.prototype.getWindowBackgroundColor = function() {
	return this._windowBackgroundColor;
};

/**
 * Set the image background color
 * @param {Number} rgbInt - RGB number encoded color
 */
ViewControlsModel.prototype.setImageBackgroundColor = function(rgbInt) {
	var oldValue = this._imageBackgroundColor;
	this._imageBackgroundColor = rgbInt;
	// save the setting
	this.f.setOption("PREFERENCES_IMAGE_BACKGROUNDCOLOR", rgbInt);
	this._fireChange(false, ["imageBackgroundColor"], oldValue, rgbInt);
};

/**
 * Get the image background color
 * @returns {Number} The current image background: rgb encoded number
 */
ViewControlsModel.prototype.getImageBackgroundColor = function() {
	return this._imageBackgroundColor;
};

/**
 * Check whether the current image background color is the same as the default value
 * @returns {Boolean} True if default, false otherwise
 */
ViewControlsModel.prototype.isImageBackgroundColorDefault = function() {
	return this._imageBackgroundColor === this.f.getCurrentUser().getOptionDefault("PREFERENCES_IMAGE_BACKGROUNDCOLOR");
};

/**
 * Check whether the current window background color is the same as the default value
 * @returns {Boolean} True if default, false otherwise
 */
ViewControlsModel.prototype.isWindowBackgroundColorDefault = function() {
	return this._windowBackgroundColor === this.f.getCurrentUser().getOptionDefault("PREFERENCES_WINDOW_BACKGROUNDCOLOR");
};

/**
 * Reset the background colors to its default values
 */
ViewControlsModel.prototype.resetBackgroundColors = function() {
	this.setImageBackgroundColor(this.f.getCurrentUser().getOptionDefault("PREFERENCES_IMAGE_BACKGROUNDCOLOR"));
	this.setWindowBackgroundColor(this.f.getCurrentUser().getOptionDefault("PREFERENCES_WINDOW_BACKGROUNDCOLOR"));
};

export default ViewControlsModel;
