/**
 * @namespace ViewControlsModule
 */
/**
 * @namespace ViewControlsModule
 */
import AbstractModule from "@esko/web-viewer/core/AbstractModule";

import ViewControlsWidget from "@esko/web-viewer/viewControls/ViewControlsWidget";
import ViewControlsModel from "@esko/web-viewer/viewControls/ViewControlsModel";
import ViewControlsConstants from "@esko/web-viewer/viewControls/ViewControlsConstants";
import ListenerUtil from "@esko/web-viewer/utils/ListenerUtil";
import ViewerUtils from "@esko/web-viewer/utils/ViewerUtils";
import _ from "underscore";
/**
 * @class
 * This module is responsible for displaying the view widget.
 * Also acts as the controller for the compensate distortion button, and compensates the scale and distortion for digital film files.
 *
 * @exports ViewControlsModule
 * @constructor
 * @augments Core.AbstractModule
 * @param {Core.ViewerFacade} f
 * @param {Boolean} silenced
 */
var ViewControlsModule = function ViewControlsModule(f, silenced) {
	AbstractModule.call(this, f, silenced, "viewControls");
	ListenerUtil.mixin(this, f);
	/**
	 * @member {ViewControlsModel} - model for the view controls, saves data needed for some of the controls
	 */
	this._viewControlsModel = new ViewControlsModel(f);

	/**
	 * @member {ViewControlsWidget} - the view widget, all view controls are accesable from this widget
	 */
	this._viewControlsWidget = null;
	this._publishWidget(
		ViewControlsWidget.widgetName,
		_.bind(function(layoutType) {
			this._viewControlsWidget = new ViewControlsWidget(f, this._viewControlsModel, this);
			return this._viewControlsWidget;
		}, this)
	);

	//listen to when the active or difference pages changes and be sure the parameters are available
	this.f.getDocumentController().whenActiveOrDifferencePageChangedAndParametersAvailable(
		_.bind(function(activePage, diffPage, oldPage, isActivePage) {
			var isScaleCompensated = this._viewControlsModel.isScaleCompensated();
			var isDistCompensated = this._viewControlsModel.isDistCompensated();

			this._viewControlsModel.resetScale(true); //be sure to reset the scale and distortion saved in the model
			this._viewControlsModel.resetDistortion(true);
			this._viewControlsModel.resetScale(false); //be sure to reset the scale and distortion saved in the model
			this._viewControlsModel.resetDistortion(false);
			if (activePage.isDigitalFilmPage()) {
				//check if the file is a digital film
				this._viewControlsModel.setScale(activePage.get("hscale"), activePage.get("vscale"), true); //set the scale parameters of the page in the model
				this._viewControlsModel.setDistortion(activePage.get("hdist"), activePage.get("vdist"), true); //set the distortion parameters of the page in the model
			}
			if (diffPage) {
				//for the diff page no imagetype-parameter is present, always add the distortion/scaling, view is smart enough to not show the button
				this._viewControlsModel.setScale(diffPage.get("hscale"), diffPage.get("vscale"), false); //set the scale parameters of the page in the model
				this._viewControlsModel.setDistortion(diffPage.get("hdist"), diffPage.get("vdist"), false); //set the distortion parameters of the page in the model
			}

			if (activePage.isDigitalFilmPage() || (diffPage && diffPage.isDigitalFilmPage())) {
				if (isScaleCompensated && isDistCompensated) {
					this.compensateScaleAndDistortion();
				} else if (isScaleCompensated) {
					this.compensateScale();
				} else if (isDistCompensated) {
					this.compensateDistortion();
				}
			}

			// Save/Restore state when not print
			if (isActivePage) {
				// save state in old page
				if (oldPage !== undefined) {
					oldPage.set("viewControls", this._viewControlsModel.getState());
				}

				var state = activePage.get("viewControls");
				if (state !== undefined) {
					var defaultState = this._viewControlsModel.getDefaultState();
					// fall back on defaults for highlight color to reset it on page change
					state.overprintHighlightColor = defaultState.overprintHighlightColor;
					state.overprintHighlightColorActive = defaultState.overprintHighlightColorActive;
				}

				// restore state
				this._viewControlsModel.setState(state);
			}
		}, this)
	);

	// catch requests for fragments to add parameters to the request
	this.f.on("intercept.fragmentURL", _.bind(this._interceptFragmentURL, this));
	this.f.on("intercept.parameters", _.bind(this._fillCompensationParameters, this));

	this._registerListener("changed.backgroundColorActive", this._backgroundColorActiveChanged, this, this._viewControlsModel);
	this._registerListener("changed.imageBackgroundColor", this._imageBackgroundColorChanged, this, this._viewControlsModel);
	this._registerListener("changed.windowBackgroundColor", this._windowBackgroundColorChanged, this, this._viewControlsModel);
	this._registerListener("changed.overprintHighlightColorActive", this._viewControlsModelChanged, this, this._viewControlsModel);
	this._registerListener("changed.overprintHighlightColor", this._viewControlsModelChanged, this, this._viewControlsModel);
	this._registerListener("changed.negated", this._viewControlsModelChanged, this, this._viewControlsModel);
	this._registerListener("changed.viewFromBack", this._viewFromBackChanged, this, this._viewControlsModel);
	this._registerListener("options.changed.PREFERENCES_MARGIN_CLIP_MODE", this._viewControlsModelChanged, this, this.f.getCurrentUser());

	this.f.whenModuleAvailable("cad", this._cadBackgroundColorChanged, this);
};

// inherit
ViewControlsModule.prototype = Object.create(AbstractModule.prototype);
// correct the constructor pointer
ViewControlsModule.prototype.constructor = ViewControlsModule;

/**
 * Compensate distortion for digital film files, only possible when active page is digital film (and 'compensate distortion'-button is active).
 */
ViewControlsModule.prototype.compensateDistortion = function() {
	this.f.getDocumentController().willNeedParameterUpdate("COMPENSTATE_DISTORTION");
	if (this._viewControlsModel.isDistCompensated()) {
		//check if the distortion is currently compensated, if so undo the compensation
		this.f.getActivePage().resetStateParam("compensateDist"); //reset the parameter of the active page, we'll refetch them since we are changing dimension of the page again
		this._viewControlsModel.setIsDistCompensated(false); //the compensation is no longer active, set this correctly in the model
	} else {
		//we want to compensate distortion!
		this.f.getActivePage().setStateParam("compensateDist", ViewControlsConstants.compensateDist.DISTONLY); //set a parameter of the currently active document page to compensate distortion
		this._viewControlsModel.setIsDistCompensated(true); //the distortion is not compensated
	}
	this.f.getDocumentController().readyForParameterUpdate(
		"COMPENSTATE_DISTORTION",
		function() {
			this.f.emit("viewControls.compensationChanged");
		},
		this
	);
};

/**
 * Compensate the scale of the active digital film page.
 */
ViewControlsModule.prototype.compensateScale = function() {
	//implementation is ~same as compensateDistortion
	this.f.getDocumentController().willNeedParameterUpdate("COMPENSTATE_SCALE");
	if (this._viewControlsModel.isScaleCompensated()) {
		this.f.getActivePage().resetStateParam("compensateDist");
		this._viewControlsModel.setIsScaleCompensated(false);
	} else {
		this.f.getActivePage().setStateParam("compensateDist", ViewControlsConstants.compensateDist.SCALEONLY);

		this._viewControlsModel.setIsScaleCompensated(true);
	}
	this.f.getDocumentController().readyForParameterUpdate(
		"COMPENSTATE_SCALE",
		function() {
			this.f.emit("viewControls.compensationChanged");
		},
		this
	);
};

/**
 * Compensate the scale and distortion of the digital film page
 * @param {Boolean} needsBothCompensated - do both the scaling and distortion need to be enabled before disabling them
 */
ViewControlsModule.prototype.compensateScaleAndDistortion = function(needsBothCompensated) {
	this.f.getDocumentController().willNeedParameterUpdate("COMPENSTATE_SCALE_AND_DISTORTION");
	if (
		(needsBothCompensated && this._viewControlsModel.isScaleCompensated() && this._viewControlsModel.isDistCompensated()) ||
		(!needsBothCompensated && (this._viewControlsModel.isScaleCompensated() || this._viewControlsModel.isDistCompensated()))
	) {
		this.f.getActivePage().resetStateParam("compensateDist");
		this._viewControlsModel.setIsDistCompensated(false);
		this._viewControlsModel.setIsScaleCompensated(false);
	} else {
		this.f.getActivePage().setStateParam("compensateDist", ViewControlsConstants.compensateDist.DISTANDSCALE);
		this._viewControlsModel.setIsDistCompensated(true);
		this._viewControlsModel.setIsScaleCompensated(true);
	}
	this.f.getDocumentController().readyForParameterUpdate(
		"COMPENSTATE_SCALE_AND_DISTORTION",
		function() {
			this.f.emit("viewControls.compensationChanged");
		},
		this
	);
};

ViewControlsModule.prototype.getImageBackgroundColor = function() {
	return this._viewControlsModel.getImageBackgroundColor();
};

ViewControlsModule.prototype.isBackgroundColorActive = function() {
	return this._viewControlsModel.isBackgroundColorActive();
};

/**
 * The viewControlsModel has changed, invalidate fragments and rerender
 */
ViewControlsModule.prototype._viewControlsModelChanged = function() {
	this.f.emit("viewport.viewserverParametersChanged");
};

/**
 * Fill in the parameters for scale and distortion compensation
 * @param c
 */
ViewControlsModule.prototype._fillCompensationParameters = function(c) {};

/**
 * Method for interception the request for new fragment, will add if the distortion or scale needs to be compensated
 * @private
 * @param {Object} c
 */
ViewControlsModule.prototype._interceptFragmentURL = function(c) {};

/**
 * Control other modules for set view from back
 */
ViewControlsModule.prototype._viewFromBackChanged = function() {
	var isViewFromBack = this._viewControlsModel.isViewFromBack();
	var graphicsModule = this.f.getModule("graphics");
	if (graphicsModule) {
		var graphicsModel = graphicsModule.getGraphicsModel();
		graphicsModel.setReverseChannelOrder(isViewFromBack);
	}
	this.f.emit("viewport.transform", {
		flipViewFromBack: isViewFromBack,
		animated: false
	});

	this._viewControlsModelChanged();
};

/**
 * Called when the background color active state changed
 * -> update other modules
 */
ViewControlsModule.prototype._backgroundColorActiveChanged = function() {
	this._imageBackgroundColorChanged();
	this._windowBackgroundColorChanged();
};

/**
 * Control other modules for cad background color
 */
ViewControlsModule.prototype._cadBackgroundColorChanged = function() {
	// cad
	var cadModule = this.f.getModule("cad");
	if (cadModule) {
		if (this._viewControlsModel.isBackgroundColorActive()) {
			// the chosen one
			var hexColor = ViewerUtils.hexFromInt(this._viewControlsModel.getImageBackgroundColor());
			cadModule.setCadBackgroundColor(hexColor);
		} else {
			// none
			cadModule.setCadBackgroundColor();
		}
	}
};

/**
 * Control other modules for image background color
 */
ViewControlsModule.prototype._imageBackgroundColorChanged = function() {
	// for cad
	this._cadBackgroundColorChanged();
	// for graphics
	this._viewControlsModelChanged();
};

/**
 * Control other modules for window background color
 */
ViewControlsModule.prototype._windowBackgroundColorChanged = function() {
	var viewportModule = this.f.getModule("viewport");
	if (viewportModule) {
		if (this._viewControlsModel.isBackgroundColorActive()) {
			// the chosen one
			var hexColor = ViewerUtils.hexFromInt(this._viewControlsModel.getWindowBackgroundColor());
			viewportModule.setWidgetBackgroundColor(hexColor);
		} else {
			// none
			viewportModule.setWidgetBackgroundColor();
		}
	}
};

/**
 * @override
 */
ViewControlsModule.prototype.getState = function() {
	return {
		modelState: this._viewControlsModel.getState(),
		isDistCompensated: this._viewControlsModel.isDistCompensated(),
		isScaleCompensated: this._viewControlsModel.isScaleCompensated(),
		hscale: this._viewControlsModel.getHorizontalScale(true),
		vscale: this._viewControlsModel.getVerticalScale(true),
		hdist: this._viewControlsModel.getHorizontalDistortion(true),
		vdist: this._viewControlsModel.getVerticalDistortion(true),
		diffhscale: this._viewControlsModel.getHorizontalScale(),
		diffvscale: this._viewControlsModel.getVerticalScale(),
		diffhdist: this._viewControlsModel.getHorizontalDistortion(),
		diffvdist: this._viewControlsModel.getVerticalDistortion()
	};
};

/**
 * @Override
 * @param {Object} state
 */
ViewControlsModule.prototype.setState = function(state) {
	if (state) {
		this._viewControlsModel.setState(state.modelState);
		this._viewControlsModel.setScale(state.hscale, state.vscale, true); //set the scale parameters of the page in the model
		this._viewControlsModel.setDistortion(state.hdist, state.vdist, true); //set the distortion parameters of the page in the model
		this._viewControlsModel.setScale(state.diffhscale, state.diffvscale, false); //set the scale parameters of the page in the model
		this._viewControlsModel.setDistortion(state.diffhdist, state.diffvdist, false); //set the distortion parameters of the page in the model
		if (state.isDistCompensated && state.isScaleCompensated) {
			this.compensateScaleAndDistortion();
		} else if (state.isDistCompensated) {
			this.compensateDistortion();
		} else if (state.isScaleCompensated) {
			this.compensateScale();
		}
	}
};

export default ViewControlsModule;
