import AbstractWidget from "@esko/web-viewer/layout/AbstractWidget";
import LayoutConstants from "@esko/web-viewer/layout/LayoutConstants";
import DropdownToolbarButton from "@esko/web-viewer/layout/component/DropdownToolbarButton";
import ViewerToolbarButton from "@esko/web-viewer/layout/component/ViewerToolbarButton";
import BackgroundColorSelector from "@esko/web-viewer/viewControls/BackgroundColorSelector";
import OverprintHighlightColorSelector from "@esko/web-viewer/viewControls/OverprintHighlightColorSelector";
import TrimboxDropdownToolbarButton from "@esko/web-viewer/viewControls/TrimboxDropdownToolbarButton";
import TrimboxClippingDropdownToolbarButton from "@esko/web-viewer/viewControls/TrimboxClippingDropdownToolbarButton";
import ListenerUtil from "@esko/web-viewer/utils/ListenerUtil";
import _ from "underscore";
import $ from "jquery";
/**
 * @exports ViewControlsWidget
 * @memberof ViewControlsModule
 * @augments Layout.AbstractWidget
 * @class The View widget shows several advanced options and tools.
 * @constructor
 * @param {Core.ViewerFacade} f
 */
var ViewControlsWidget = function ViewControlsWidget(f, viewControlsModel, viewControlsModule) {
	AbstractWidget.call(this, f, LayoutConstants.Widget.VIEWCONTROLS);
	ListenerUtil.mixin(this, f);

	/**
	 * @member {ViewControlsModle} model for view controls
	 */
	this._viewControlsModel = viewControlsModel;

	/**
	 * @member {ViewControlsModule} module, acts as a controller for the view controls
	 */
	this._viewControlsModule = viewControlsModule;

	/**
	 * @member {Button} this button enables/disables the inverse view
	 */
	this._negateButton = new ViewerToolbarButton(
		this.f,
		"negate",
		"icon-WCR_Invert",
		false,
		_.bind(function() {
			this._viewControlsModel.setIsNegated(!this._viewControlsModel.isNegated());
		}, this)
	);

	/**
	 * @member {ViewerToolbarButton} this button enables/disables the view from back view
	 */
	this._viewFromBackButton = new ViewerToolbarButton(
		this.f,
		"viewFromBackButton",
		"viewFromBackOff",
		false,
		_.bind(function() {
			this._viewControlsModel.setViewFromBack(!this._viewControlsModel.isViewFromBack());
		}, this)
	);

	/**
	 * @member {DropdownToolbarButton} dropdown button to compensate distortion and scaling in digital film files,
	 * will only show the dropdown if both distortion and scaling are in the file
	 */
	this._distortionButton = new DropdownToolbarButton(f, "distortion", "distortionOffIcon", false, _.bind(function() {}, this));

	this.setIcon("icon icon-WCR_View");

	/**
	 * @member {DropdownToolbarButton} dropdown button to enable disabling trimbox clipping
	 */
	this._marginClippingButton = new TrimboxClippingDropdownToolbarButton(this.f, "marginClipping");
	this._marginClippingButton.setTooltip("viewer", "TrimboxClippingTT");
	this.toggleCompensateDistortionButton(false);

	this._trimboxDropdown = new TrimboxDropdownToolbarButton(
		this.f,
		"showTrimboxDropdown",
		"showTrimboxDropdown",
		this.f.getOption("PREFERENCES_DISPLAY_MARGINS") === 1,
		this.f.getOption("PREFERENCES_MARGIN_DISPLAY_MODE")
	);
	this._trimboxDropdown.setTooltip("viewer", "TrimboxEnableColorTT");

	this.f.loadBundle(
		"viewer",
		function() {
			this.setTitle(this.f.t("ViewControlsTitle"));

			// add options to the 'compensate distortion'-button
			this._distortionButton.addOption({
				//first option to only compensate distortion
				name: "onlyDistortion",
				content: f.i18n.t("FileDistChoiceDistOnly"),
				onchange: _.bind(function() {
					//when the option is clicked, enable/disable compensate distortion
					var isOtherActive = this._distortionButton.getOptionValue("onlyScale") || this._distortionButton.getOptionValue("both");
					this._distortionButton.toggleOption("onlyScale", false); //toggle the other two option to false only the distortion is altered
					this._distortionButton.toggleOption("both", false);
					if (this._viewControlsModel.isScaleCompensated()) {
						//if the scale was compensated, undo this only distortion gets altered
						this._viewControlsModel.setIsScaleCompensated(false);
					}
					if (isOtherActive && this._viewControlsModel.isDistCompensated()) {
						//only if another option was active disable distortion to get the correct data for only distortion
						this._viewControlsModel.setIsDistCompensated(false);
					}
					this._viewControlsModule.compensateDistortion(); //compensate distortion, will turn distortion off if it was on before (option was selected)
				}, this),
				userCannotUncheck: false //it is possible to disable the compensate distortion again
			});

			this._distortionButton.addOption({
				//only compensate scale option
				name: "onlyScale",
				content: f.i18n.t("FileDistChoiceScaleOnly"),
				onchange: _.bind(function() {
					//same as the 'only distortion'-option but for compensate scaling
					var isOtherActive = this._distortionButton.getOptionValue("onlyDistortion") || this._distortionButton.getOptionValue("both");
					this._distortionButton.toggleOption("onlyDistortion", false); //disable other options
					this._distortionButton.toggleOption("both", false);
					if (this._viewControlsModel.isDistCompensated()) {
						//if distortion was compensated, undo this compensation
						this._viewControlsModel.setIsDistCompensated(false);
					}
					if (isOtherActive && this._viewControlsModel.isScaleCompensated()) {
						//only when the option wasn't active disable to get data for only scaling
						this._viewControlsModel.setIsScaleCompensated(false);
					}
					this._viewControlsModule.compensateScale(); //enable/disable compensate scaling
				}, this),
				userCannotUncheck: false
			});

			this._distortionButton.addOption({
				//compensate both distortion and scaling
				name: "both",
				content: f.i18n.t("FileDistChoiceBoth"),
				onchange: _.bind(function() {
					//when clicked set other option to false and trigger compensation for both distortion and scaling
					this._distortionButton.toggleOption("onlyDistortion", false);
					this._distortionButton.toggleOption("onlyScale", false);
					this._viewControlsModule.compensateScaleAndDistortion(true);
				}, this),
				userCannotUncheck: false
			});

			this._distortionButton.disableDropdown(); //disable the dropdown by default

			this._updateNegateButton();
			this._updateViewFromBackButton();
		},
		this
	);

	// Change background color for image and window
	this._backgroundColorSelector = new BackgroundColorSelector(this.f, this._viewControlsModel);
	this._registerListener(
		"resized",
		function() {
			this._fireResized();
		},
		this,
		this._backgroundColorSelector
	);

	// Set overprint highlight color
	this._overprintHighlightColorSelector = new OverprintHighlightColorSelector(this.f, this._viewControlsModel);
	this._registerListener(
		"resized",
		function() {
			this._fireResized();
		},
		this,
		this._overprintHighlightColorSelector
	);

	//make the distortion button renderer debounced, might get trigger by both changing scaling and distortion (in this case it gets called twice)
	var debouncedDistortionButtonRenderer = _.debounce(this._updateDistortionButton, 50);
	this._registerListener("scaleOrDistortionChanged", debouncedDistortionButtonRenderer, this, this._viewControlsModel);
	this._registerListener("changed.negated", this._updateNegateButton, this, this._viewControlsModel);
	this._registerListener("changed.viewFromBack", this._updateViewFromBackButton, this, this._viewControlsModel);

	var updateClippingButton = _.bind(function() {
		this._marginClippingButton.update();
	}, this);

	this._registerListener(
		"module.compare.changed.diffMode",
		function() {
			_.defer(updateClippingButton);
		},
		this._marginClippingButton
	); // defer the update, to make sure the modules already did the needed changes after activepage changed (i.e. setting isReaderspread)

	this.f.getDocumentController().whenActivePageChanged(_.bind(this._updateButtons, this));
};

//inherit
ViewControlsWidget.prototype = Object.create(AbstractWidget.prototype);
ViewControlsWidget.prototype.constructor = ViewControlsWidget;
ViewControlsWidget.widgetName = LayoutConstants.Widget.VIEWCONTROLS;

/**
 * Set the context of the widget
 * @override
 * @param {JQueryObj} context
 * @param {JQueryObj} header
 */
ViewControlsWidget.prototype.setContext = function(context, header) {
	AbstractWidget.prototype.setContext.call(this, context, header);

	var buttons = $("<div/>").addClass("buttons"); //make a wrapper for all buttons (or group button in the future)
	buttons.append(this._distortionButton.element);
	buttons.append(this._backgroundColorSelector.getButtonElement());
	buttons.append(this._viewFromBackButton.element);
	buttons.append(this._overprintHighlightColorSelector.getButtonElement());
	buttons.append(this._negateButton.element);
	buttons.append(this._trimboxDropdown.element);
	buttons.append(this._marginClippingButton.element);

	this.context.append(buttons);
	this.context.append(this._backgroundColorSelector.getToolbar());
	this.context.append(this._overprintHighlightColorSelector.getToolbar());

	this.context.addClass("viewWidget");
};

/**
 * Render the distortion button (and show/hide its dropdown), change the default action for the toolbarbutton
 */
ViewControlsWidget.prototype._updateDistortionButton = function() {
	//if there is no distortion and no scaling for both the active and reference document pages be sure to hide the distortion button
	if (
		!this._viewControlsModel.hasPageDistortion() &&
		!this._viewControlsModel.hasPageScale() &&
		(!this._viewControlsModel.hasDiffPageDistortion() && !this._viewControlsModel.hasDiffPageScale())
	) {
		this.toggleCompensateDistortionButton(false);
	} else {
		//render the distortion button correctly
		this.toggleCompensateDistortionButton(true);
		this.f.loadBundle(
			"viewer",
			function() {
				//load language strings for the tooltip of the distortion button
				this._distortionButton.toggleOption("both", false);
				this._distortionButton.toggleOption("onlyDistortion", false);
				this._distortionButton.toggleOption("onlyScale", false);
				if (this._viewControlsModel.isScaleCompensated() || this._viewControlsModel.isDistCompensated()) {
					//if there's a compensation active
					if (this._viewControlsModel.isScaleCompensated() && this._viewControlsModel.isDistCompensated()) {
						this._distortionButton.toggleOption("both", true);
					} else if (this._viewControlsModel.isScaleCompensated()) {
						this._distortionButton.toggleOption("onlyScale", true);
					} else {
						this._distortionButton.toggleOption("onlyDistortion", true);
					}
					this._distortionButton.setActive(true); //render the button as being active
					this._distortionButton.setIcon("distortionOnIcon"); //and change the icon on the toolbarbutton
				} else {
					//compensation for distortion or scaling is not active at this moment
					this._distortionButton.setActive(false);
					this._distortionButton.setIcon("distortionOffIcon");
				}
				//if either the active or reference page has both distortion and scaling, active the dropdown of the toolbarbutton
				if (
					(this._viewControlsModel.hasPageDistortion() && this._viewControlsModel.hasPageScale()) ||
					(this._viewControlsModel.hasDiffPageDistortion() && this._viewControlsModel.hasDiffPageScale())
				) {
					this._distortionButton.enableDropdown(); //enables the dropdown
				} else {
					this._distortionButton.disableDropdown();
				}
				//start rendering the tooltip for the 'compensate distortion'-button
				var tooltip = "";
				//if the active or reference page has distortion, show the distortion in the tooltip
				if (this._viewControlsModel.hasPageDistortion() || this._viewControlsModel.hasDiffPageDistortion()) {
					//set the default (left click) action for the 'compensate distortion'-button to only compensate distortion
					this._distortionButton.setAction(_.bind(this._viewControlsModule.compensateDistortion, this._viewControlsModule));
					var referenceHDist = "";
					var referenceVDist = "";
					if (this._viewControlsModel.hasDiffPageDistortion()) {
						//if the reference page has distortion show this in the tooltip
						referenceHDist = " / " + (this._viewControlsModel.getHorizontalDistortion(false) * 100).toFixed(1);
						referenceVDist = " / " + (this._viewControlsModel.getVerticalDistortion(false) * 100).toFixed(1);
					}
					if (this._viewControlsModel.isDistCompensated()) {
						//if compensation for distortion is active or not, render a different tooltip
						tooltip += this.f.t("FileCompensateDistOnTT");
					} else {
						tooltip += this.f.t("FileCompensateDistOffTT");
					}
					tooltip +=
						"<br/>&nbsp;&nbsp;" +
						this.f.t("FileVertDistTT", (this._viewControlsModel.getVerticalDistortion(true) * 100).toFixed(1) + referenceVDist) +
						" %";
					tooltip +=
						"<br/>&nbsp;&nbsp;" +
						this.f.t("FileHorDistTT", (this._viewControlsModel.getHorizontalDistortion(true) * 100).toFixed(1) + referenceHDist) +
						" %";
				}
				//if the active or reference page has scaling, show the scaling in the tooltip
				if (this._viewControlsModel.hasPageScale() || this._viewControlsModel.hasDiffPageScale()) {
					if (tooltip !== "") {
						//if there is already a tooltip constructed for the distortion, add a <br>
						tooltip += "<br/>";
					}
					//set the default (left click) action for the button to compensate only the scaling of the active and reference document page
					this._distortionButton.setAction(_.bind(this._viewControlsModule.compensateScale, this._viewControlsModule));
					var referenceHScale = "";
					var referenceVScale = "";
					if (this._viewControlsModel.hasDiffPageScale()) {
						//if there is reference page add the scaling to the tooltip
						referenceHScale = " / " + (this._viewControlsModel.getHorizontalScale(false) * 100).toFixed(1);
						referenceVScale = " / " + (this._viewControlsModel.getVerticalScale(false) * 100).toFixed(1);
					}
					if (this._viewControlsModel.isScaleCompensated()) {
						tooltip += this.f.t("FileCompensateScaleOnTT");
					} else {
						tooltip += this.f.t("FileCompensateScaleOffTT");
					}
					tooltip +=
						"<br/>&nbsp;&nbsp;" +
						this.f.t("FileVertScaleTT", (this._viewControlsModel.getVerticalScale(true) * 100).toFixed(1) + referenceVScale) +
						" %";
					tooltip +=
						"<br/>&nbsp;&nbsp;" +
						this.f.t("FileHorScaleTT", (this._viewControlsModel.getHorizontalScale(true) * 100).toFixed(1) + referenceHScale) +
						" %";
				}
				//if the reference or active page has both distortion and scaling, change the default action for the button to compensate both the distortion and the scaling
				if (
					(this._viewControlsModel.hasPageDistortion() && this._viewControlsModel.hasPageScale()) ||
					(this._viewControlsModel.hasDiffPageDistortion() && this._viewControlsModel.hasDiffPageScale())
				) {
					this._distortionButton.setAction(
						_.bind(function() {
							//change the action
							if (this._distortionButton.isActive()) {
								this._viewControlsModule.compensateScaleAndDistortion(); //compensate both distortion and scaling
							} else {
								this._distortionButton.showDropdown();
							}
						}, this)
					);
				}

				this._distortionButton.setTooltip(tooltip); //set the tooltip of the toolbarbutton
			},
			this
		);
	}
};

/**
 * Render the view from back button
 */
ViewControlsWidget.prototype._updateViewFromBackButton = function() {
	if (this.f.getActivePage().hasGraphics()) {
		this._viewFromBackButton.show();
		if (this._viewControlsModel.isViewFromBack()) {
			this._viewFromBackButton.setIcon("viewFromBackOn");
			this._viewFromBackButton.setTooltip("viewer", "ToolBackTT2");
			this._viewFromBackButton.setActive(true);
		} else {
			this._viewFromBackButton.setIcon("viewFromBackOff");
			this._viewFromBackButton.setTooltip("viewer", "ToolBackTT");
			this._viewFromBackButton.setActive(false);
		}
	} else {
		this._viewFromBackButton.hide();
	}
};

/**
 * Toggle the 'compensate distortion'-button, shows or hides the button
 * @param {Boolean} toggle
 */
ViewControlsWidget.prototype.toggleCompensateDistortionButton = function(toggle) {
	if (toggle) {
		this._distortionButton.element.show();
	} else {
		this._distortionButton.element.hide();
	}
};

/**
 * Toggle the overprint highlight color button, shows or hides the button
 * @param {Boolean} toggle
 */
ViewControlsWidget.prototype.toggleOverprintHighlightColorButton = function(toggle) {
	this._overprintHighlightColorSelector.toggle(toggle);
};

/**
 * Update the display of the inverse view button
 */
ViewControlsWidget.prototype._updateNegateButton = function() {
	if (this.f.getActivePage().hasGraphics()) {
		this._negateButton.show();
		var isNegated = this._viewControlsModel.isNegated();
		this._negateButton.setActive(isNegated);
		this._negateButton.setIcon(isNegated ? "icon-WCR_InvertReverse" : "icon-WCR_Invert");
		this._negateButton.setTooltip("viewer", isNegated ? "ToolInvertTT2" : "ToolInvertTT");
	} else {
		this._negateButton.hide();
	}
};

/**
 * Update the 'show trimbox dropdown'-button, shows or hides the button
 */
ViewControlsWidget.prototype._updateShowTrimboxDropdownToolbarButton = function() {
	if (this.f.getActivePage().hasGraphics() || this.f.getActivePage().isBlankPage()) {
		this._trimboxDropdown.show();
	} else {
		this._trimboxDropdown.hide();
	}
};

/**
 * Update the overprint highlight color button
 */
ViewControlsWidget.prototype._updateOverprintHighlightColorButton = function() {
	this._overprintHighlightColorSelector.toggle(this.f.getActivePage().hasGraphics());
};

/**
 * Update the buttons after a active page change
 */
ViewControlsWidget.prototype._updateButtons = function() {
	this._marginClippingButton.update();
	this._updateShowTrimboxDropdownToolbarButton();
	this._updateNegateButton();
	this._updateOverprintHighlightColorButton();
	this._updateViewFromBackButton();
};

export default ViewControlsWidget;
