import $ from "jquery";
import _ from "underscore";
import DrawUtils from "@esko/web-viewer/utils/DrawUtils";
import ShapeUtils from "@esko/shapeutils";
import Transformation from "@esko/transformation";
import ListenerUtil from "@esko/web-viewer/utils/ListenerUtil";
import Fragment from "@esko/web-viewer/utils/Fragment";
import ConversionConstants from "@esko/web-viewer/utils/ConversionConstants";
import ViewportFragmentManager from "@esko/web-viewer/viewport/model/ViewportFragmentManager";
import HTTPUtils from "@esko/httputils";
import Popover from "@esko/popover/Popover";
/**
 * Rendering speed
 * @constant
 */
var RENDER_SPEED = 40;

/**
 * Create a ViewportLayer
 *
 * @exports ViewportLayer
 * @memberof ViewportModule
 * @class A layer is responsible to render the content of one specific part of a view
 * @constructor
 * @mixes Utils.ListenerUtilMixin
 *
 * @param {Core.ViewerFacade} f
 * @param {ViewportModule.viewportView} viewportView - view of the layer
 * @param {ViewportModule.ViewportWorld} world - the world of the Viewport
 * @param {Transformation} viewTransform - transformation dictated by the view
 * @param {String} id - the ID of the Layer
 */
var ViewportLayer = function(f, viewportView, world, viewTransform, id) {
	this.f = f;
	ListenerUtil.mixin(this, f);

	/**
	 * Unique ID of the canvas
	 * this also the ID in the DOM
	 * @protected
	 */
	this._id = id;

	/**
	 * Main canvas object that is used for rendering
	 * @protected
	 */
	this._canvas = $("<canvas>")
		.css({
			position: "absolute",
			top: 0,
			left: 0
		})
		.attr("id", this._id);

	/**
	 * The layerModel keeps track of the content of the layer
	 * @protected
	 * @type {ViewportModule.ViewportLayerModel}
	 */
	this._layerModel = null;

	/**
	 * The view this is layer is part of
	 * @private
	 */
	this._view = viewportView;

	/**
	 * @protected
	 * @type {Boolean}
	 */
	this._visible = true;
	this._canvasIsShown = true;

	/**
	 * This keeps track of all listeners registered to the model
	 */
	this._listeners = [];

	/**
	 * The world for the ViewPort
	 * @private
	 */
	this._world = world;

	/**
	 * The currently used transformation
	 * @private
	 */
	this._transform = viewTransform;

	/**
	 * Is the buffer enabled?
	 * NOTE this buffer doesn't work correctly at the moment, should be repaired if it is necessary to have it enabled
	 * @private
	 */
	this._enableBuffer = false;

	/**
	 * Should the current linework be buffered (when the buffer is enabled)
	 * @private
	 */
	this._shouldBuffer = false;

	/**
	 * Buffer image canvas
	 * @private
	 */
	this._buffer = null;

	/**
	 * Debug the buffer functionality?
	 * @private
	 */
	this._debugBuffer = false;

	/**
	 * Used to now whether some ImageDataRequest should be send out (this is to now whethere all fragments are loaded)
	 * @protected {Boolean}
	 */
	this._dirty = false;

	/**
	 * Interactive mode
	 * @private {Boolean}
	 */
	this._interactive = false;
	/**
	 * Render the layer in a throttled way
	 */
	this.throttledRender = _.throttle(_.bind(this.render, this), RENDER_SPEED);

	/**
	 * Repaint the layer in a throttled way
	 */
	this.debouncedRepaint = _.debounce(
		_.bind(function() {
			var wasDirty = this._dirty;
			this._dirty = false;
			this.preloadViewportFragment();
			this.throttledRender();

			if (wasDirty) {
				this.f.emit("viewportLayer.finishedImageDataRequest", this._id);
			}
		}, this),
		40
	);

	/**
	 * This manages the fragments shown in
	 * @protected
	 */
	this._fragmentManager = new ViewportFragmentManager(
		this.f,
		{
			getURLForFragment: _.bind(function(fragment) {
				return this.generateFragmentURL(fragment);
			}, this),
			fragmentsChanged: this.throttledRender
		},
		this._world
	);

	// extra overwrite options on top of url parameters
	this._overwriteOptions = {};
};

/**
 * Get the layerModel of the layer
 * @returns {ViewportModel.ViewportLayerModel}
 */
ViewportLayer.prototype.getLayerModel = function() {
	return this._layerModel;
};

/**
 * Set the layermodel. This will register listeners to the model and unregister listeners to the old model
 * @param layerModel
 */
ViewportLayer.prototype.setLayerModel = function(layerModel) {
	this._unregisterAllListeners();

	this._layerModel = layerModel;

	this._registerListener("modelChanged", this.onLayerModelChange, this, this._layerModel);
	this._registerListener("popoverAdded", this.onPopoverAdded, this, this._layerModel);
	this._registerListener("popoverRemoved", this.onPopoverRemoved, this, this._layerModel);
	this._registerListener("repaint", this.repaint, this, this._layerModel);
	this._registerListener("fragmentsDirty", this.markAllFragmentsAsDirty, this, this._layerModel);
	this._registerListener("cleanupFragments", this.cleanupFragments, this, this._layerModel);
	this._registerListener("overwriteOption.**", this.markAllFragmentsAsDirty, this, this._layerModel);
	this._registerListener(
		"changed.visible",
		function() {
			this.repaint();
		},
		this,
		this._layerModel
	);

	this.repaint();
};

/**
 * Destroy the layer (unregister listeners, remove popups/popovers
 */
ViewportLayer.prototype.destroy = function() {
	this._unregisterAllListeners();

	this.hideAllPopovers();
	this._canvas.remove();
};

/**
 * When the layerModel has changed, rerender the layer
 */
ViewportLayer.prototype.onLayerModelChange = function() {
	this.throttledRender();
};

/**************************************************************
 *LAYER SETTINGS
 **************************************************************/

/**
 * Start interactive mode
 * All fetching is disabled, buffering is used (if enabled)
 * All popovers and popups are hidden
 */
ViewportLayer.prototype.startInteractive = function() {
	this._shouldBuffer = true;
	this._interactive = true;
	this.hideAllNonInteractivePopovers();
	this.hideAllPopups();
};

/**
 * Stop interactive mode
 * Re-enable fragment fetching
 * Stop buffering
 * repaint
 */
ViewportLayer.prototype.stopInteractive = function() {
	this._shouldBuffer = false;
	this._interactive = false;

	if (!this._world.isAnimating()) {
		this.drawPopovers();
		this.drawPopups();
		this.preloadViewportFragment();
	}
};

/**
 * Hide/show the layer
 * @param {Boolean} hide - hide/show the layer
 */
ViewportLayer.prototype.setVisible = function(visible) {
	this._visible = visible;
	this.repaint();
};

/**
 * Set the size for the canvas
 * @param {ShapeUtils.Size} dim - dimensions of the canvas
 */
ViewportLayer.prototype.setSize = function(dim) {
	if (this._canvas[0].width != dim.width || this._canvas[0].height != dim.height) {
		this._canvas[0].width = dim.width;
		this._canvas[0].height = dim.height;

		this.render(); // we have to render immediately otherwise there could be flickering (when size of canvas is set, all pixels are reset)
	}
};

/**
 * @returns {JQueryObj} the canvas
 */
ViewportLayer.prototype.getLayerElement = function() {
	return this._canvas;
};

/**
 * Get the drawing context of the layer
 * @returns {Context2D}
 */
ViewportLayer.prototype.getCanvasContext = function() {
	return this._canvas[0].getContext("2d");
};

/**
 * Set the overwrite options for the layer (not the model)
 * @param {Object}  options
 */
ViewportLayer.prototype.setOverwriteOptions = function(options) {
	this._overwriteOptions = options;
	this.markAllFragmentsAsDirty();
};

/**
 * Is the layer visible (this checks both the setting in the layer and in the model)
 * @returns {Boolean}
 */
ViewportLayer.prototype.isVisible = function() {
	return this._visible && (!this._layerModel || this._layerModel.isVisible());
};

/**************************************************************
 * CALCULATIONS/MEASUREMENTS
 **************************************************************/

/**
 * Calculate the closest linework given a certain point
 * @param {ShapeUtils.Point} point - Point that needs to be verified
 * @param {Number} [fatfinger] - Optional parameter to be able to work with fat fingers :) it searches less strict (with an diameter of fatfinger)
 * @returns {LineworkWrapper}  linework - Linework that is closest to the point. null if no object was found
 */
ViewportLayer.prototype.getClosestLineworkFromPoint = function(point, fatfinger) {
	if (this._layerModel) {
		var closest = null;
		var closestDist = -1;
		var t = this.getTransformMatrix();
		var viewRect = this._world.getWorldRectFromScene(this._world.getSceneBounds(), t);

		var linework = this._layerModel.getLinework();
		for (var i = 0; i < linework.length; i++) {
			var wrap = linework[i];
			if (wrap.rect && wrap.rect.size.width > 0 && wrap.rect.size.height > 0) {
				//check if the rect is actually in the viewport
				if (!viewRect.rectIntersect(wrap.rect)) {
					continue;
				}
			}
			var dist = DrawUtils.distanceOfScenePoint(wrap, point, t, fatfinger);
			if (dist > 0 && (closestDist < 0 || dist <= closestDist)) {
				closest = wrap;
				closestDist = dist;
			}
		}
		return closest;
	}
};

/**
 * Calculate the closest point given a certain point
 * @param {ShapeUtils.Point} refpoint - Point that needs to be verified
 * @returns {ShapeUtils.Point} point - Point that is closest to the point. null if no object was found
 */
ViewportLayer.prototype.getClosestLineworkPointFromPoint = function(refpoint, distance) {
	var closest = null;
	if (this._layerModel !== null) {
		var transform = this.getTransformMatrix();
		var viewRect = this._world.getWorldRectFromScene(this._world.getSceneBounds(), transform);
		var size = new ShapeUtils.Size(distance, distance);

		var referenceRect = new ShapeUtils.Rect(new ShapeUtils.Point(refpoint.x - distance / 2, refpoint.y - distance / 2), size);
		var referencePointInWorld = this.getWorldPointFromScene(refpoint);
		var referenceRectInWorld = this.getWorldRectFromScene(referenceRect);

		var linework = this._layerModel.getLinework();
		for (var i = 0; i < linework.length; i++) {
			var wrap = linework[i];

			if (wrap.snapable && wrap.pointArray && wrap.pointArray.length > 0) {
				// the wrapper is snappable and there is a pointarray with some points in it

				// if there is a rectangle, we only want to find the closest point if the rectangle is in the visible area
				if (!(wrap.rect && wrap.rect.size.width > 0 && wrap.rect.size.height > 0 && !viewRect.rectIntersect(wrap.rect))) {
					var points = wrap.pointArray;
					for (var j = 0; j < points.length; j++) {
						var lwPoint = points[j];
						if (referenceRectInWorld.containsPoint(lwPoint)) {
							var candidate = new ShapeUtils.Point(lwPoint.x, lwPoint.y);
							if (!closest || candidate.getDistance(referencePointInWorld) < closest.getDistance(referencePointInWorld)) {
								closest = candidate;
							}
						}
					}
				}
			}
		}
	}

	if (closest) {
		return this.getScenePointFromWorld(closest);
	}
	return null;
};

/**
 * Get the World-coordinates for a scene point
 * @param {ShapeUtils.Point} point - Scene point
 * @returns {ShapeUtils.Point} World point
 */
ViewportLayer.prototype.getWorldPointFromScene = function(point) {
	return this._world.getWorldPointFromScene(point, this.getTransformMatrix());
};

/**
 * Get the World-coordinates for a scene rect
 * @param {ShapeUtils.Rect} rect - Scene rect
 * @returns {ShapeUtils.Rect} World rect
 */
ViewportLayer.prototype.getWorldRectFromScene = function(rect) {
	return this._world.getWorldRectFromScene(rect, this.getTransformMatrix());
};

/**
 * Get the Scene-coordinates for a world rect
 * @param {ShapeUtils.Rect} rect - World rect
 * @returns {ShapeUtils.Rect} Scene rect
 */
ViewportLayer.prototype.getSceneRectFromWorld = function(rect) {
	return this._world.getSceneRectFromWorld(rect, this.getTransformMatrix());
};

/**
 * Get the Scene-coordinates for a world point
 * @param {ShapeUtils.Point} point - World point
 * @returns {ShapeUtils.Point} Scene point
 */
ViewportLayer.prototype.getScenePointFromWorld = function(point) {
	return this._world.getScenePointFromWorld(point, this.getTransformMatrix());
};

/**
 * Function to get the layer transformation within the world
 * @returns {Transformation}
 */
ViewportLayer.prototype.getLayerTransform = function() {
	var l = this._layerModel.getTransform().copy();

	var t = new Transformation();

	var offsetX = this._world.getWorldSize().width / 2 - l.e;
	var offsetY = this._world.getWorldSize().height / 2 - l.f;

	t.translate(offsetX, offsetY);
	t.concatenate(l);
	t.translate(-offsetX, -offsetY);

	return t;
};

/**
 * Function to convert a rect from a layer to world coordinates
 * @returns {ShapeUtils.Rect}
 */
ViewportLayer.prototype.convertLayerRectToWorld = function(rect) {
	return this.getLayerTransform().transformedRect(rect);
};

/**
 * Function to convert a world rect to layer coordinates
 * @returns {ShapeUtils.Rect}
 */
ViewportLayer.prototype.convertWorldRectToLayer = function(rect) {
	return this.getLayerTransform()
		.createInverse()
		.transformedRect(rect);
};

/**
 * Function to convert a point from a layer to world coordinates
 * @returns {ShapeUtils.Point}
 */
ViewportLayer.prototype.convertLayerPointToWorld = function(point) {
	return this.getLayerTransform().transformedPoint(point);
};

/**
 * Function to convert a world point to layer coordinates
 * @returns {ShapeUtils.Point}
 */
ViewportLayer.prototype.convertWorldPointToLayer = function(point) {
	return this.getLayerTransform()
		.createInverse()
		.transformedPoint(point);
};

/**
 * Calculate the matrix that is needed to draw objects on the canvas
 * @returns {Transformation} the transformation that needs to be applied for this layer
 */
ViewportLayer.prototype.getTransformMatrix = function() {
	var t = new Transformation();

	if (this._layerModel) {
		/* Layer transform */
		var layerworld = this._layerModel.getTransform();

		t.preConcatenate(layerworld);
	}

	/* View transform */
	var offsetX = this._world.getWorldSize().width / 2 - this._transform.e;
	var offsetY = this._world.getWorldSize().height / 2 - this._transform.f;

	t.preTranslate(-offsetX, -offsetY);
	t.preConcatenate(this._transform);
	t.preTranslate(offsetX, offsetY);

	/* World transform */
	t.preConcatenate(this._world.getTransformation());

	return t;
};

/**
 * @returns {float} the current zoom of the layer
 */
ViewportLayer.prototype.getZoom = function() {
	var t = this._world.getTransformation();
	return Math.sqrt(t.c * t.c + t.d * t.d);
};

/**************************************************************
 * FRAGMENT management
 **************************************************************/

/**
 * Add/preload a fragment on the viewportfragmenetmanager
 * @param {ShapeUtils.Rect} viewrect - Rect to fetch from the viewserver
 * @param {float} zoom - zoom/resolution of the fragment to fetch
 */
ViewportLayer.prototype.addFragment = function(bounds, zoom) {
	if (isNaN(zoom)) {
		return;
	}

	var maxr = this._layerModel.getMaximumResolution();
	if (maxr >= 0) {
		zoom = Math.min(maxr, zoom);
	}

	this._fragmentManager.preloadFragment(bounds, zoom, true);
};

/**
 * Generate the url parameters for a certain fragment
 * @param {Utils.Fragment} fragment - the fragment for which the URL is generated
 * @returns {String}
 */
ViewportLayer.prototype.generateFragmentURL = function(fragment) {
	var context = this._layerModel.getContextForFragment(fragment);
	return this.f.getDataAdapterForModule("viewport").generateFragmentURL(context, this._overwriteOptions);
};

/**
 * Get fragments that need to be rendered
 * @param {ShapeUtils.Rect} viewrect - frame that needs to be fetched
 * @param {float} resolution - resolution of the layer
 * @returns the fragments that are relevant for the viewrect and the resolution
 */
ViewportLayer.prototype.getApplicableFragments = function(viewrect, resolution) {
	return this._fragmentManager.getApplicableFragments(viewrect, resolution);
};

/**
 * Mark all loaded fragments as dirty
 */
ViewportLayer.prototype.markAllFragmentsAsDirty = function() {
	if (!this._dirty) {
		this._dirty = true;
		if (this.canFetchFragments()) {
			this.f.emit("viewportLayer.startImageDataRequest", this._id);
		}
	}
	this._fragmentManager.markAllAsDirty();
	this.repaint();
};

/**
 * Cleanup all fragments
 */
ViewportLayer.prototype.cleanupFragments = function() {
	this._fragmentManager.cleanupFragments();
	this.repaint();
};

/**************************************************************
 * RENDERING
 **************************************************************/

/**
 * Repaint the layer, it is possible to set a new world and transformation
 * @param {ViewportModule.ViewportWorld} [world] - the world of the Viewport
 * @param {Transformation} [viewTransform] - transformation dictated by the view
 */
ViewportLayer.prototype.repaint = function(world, transform) {
	if (world) {
		this._world = world;
	}

	if (transform) {
		this._transform = transform;
	}

	if (this._canvasIsShown != this.isVisible()) {
		this._canvasIsShown = this.isVisible();
		this.getLayerElement().toggle(this._canvasIsShown);
	}

	this.debouncedRepaint();
	this.throttledRender();
};

/**
 * Is the layer allowed to fetch fragments?
 * @returns {Boolean}
 */
ViewportLayer.prototype.canFetchFragments = function() {
	return (
		this._layerModel !== null &&
		!this._interactive &&
		this.isVisible() &&
		this._world !== undefined &&
		this._world !== null &&
		this._layerModel.hasGraphics() &&
		this._world.getZoom() > 0 &&
		this.f.canFetchFragments()
	);
};

/**
 * Preload the fragments to draw the document with the current world settingss
 */
ViewportLayer.prototype.preloadViewportFragment = function() {
	if (this.canFetchFragments()) {
		var customLoader = this.f.getDataAdapterForModule("viewport").getCustomFragmentLoader();
		if (customLoader) {
			customLoader(
				this._world,
				this._layerModel,
				_.bind(this.addFragment, this),
				_.bind(this._fragmentManager.updateInitialFragment, this._fragmentManager)
			);
		} else {
			// default implementation: load complete viewport
			var zoom = this._world.getZoom();
			var targetViewport = this._world.getAnimationDestinationTargetViewPort();

			// make sure we are requesting an image with integer values for width/height in the pixel domain
			targetViewport.size.width = Math.floor(targetViewport.size.width * zoom) / zoom;
			targetViewport.size.height = Math.floor(targetViewport.size.height * zoom) / zoom;

			if (!this._layerModel.shouldAutoFetchFragments()) {
				// let the model load the fragment
				this._layerModel.preloadFragment(targetViewport, zoom);
			} else {
				// autoload the fragment in the Layer
				this.addFragment(targetViewport, zoom);

				if (!this._world.isAnimating()) {
					this._fragmentManager.updateInitialFragment();
				}
			}
		}
	}
};

/**
 * Zooming started, this is used for interactive zooming (pinching)
 */
ViewportLayer.prototype.zoomingStarted = function() {
	this._isZooming = true;
};

/**
 * Zooming stopped, used for interactive zooming (pinching)
 */
ViewportLayer.prototype.zoomingStopped = function() {
	this._isZooming = false;
};

/**
 * Calculate the buffer for the linework (if any) of this layer if panning started
 */
ViewportLayer.prototype.panningStarted = function() {
	if (
		this._layerModel !== null &&
		this._layerModel.hasLinework(
			this._view
				.getPanel()
				.getModel()
				.getName()
		) &&
		!this._layerModel.hasBuffer(
			this._view
				.getPanel()
				.getModel()
				.getName()
		) &&
		this._canvasIsShown
	) {
		var z = this._world.getZoom();
		var rect = this._world.getTargetViewPort(true);
		var rotation = this._world.getRotation();
		var isFlipped = this._world.isFlip();
		var isViewFromBack = this._world.isFlipViewFromBack(); // is flipped view from back
		if (isViewFromBack) {
			if (rotation === 90 || rotation === 270) {
				rotation = (rotation + 180) % 360;
			}
			isFlipped = !isFlipped;
		}

		rotation = (360 + rotation) % 360;

		//we'll have to calculate a new origin when the file is rotated from 0 to 90 degrees
		//and the size of the rect of the fragment needs to change when rotation is 90 or 270
		var width = rect.size.width;
		var height = rect.size.height;
		if (rotation % 360 === 270) {
			if (height > width && !isFlipped) {
				rect.origin.x -= height - width;
			}
			rect.size = new ShapeUtils.Size(height, width);
		} else if (rotation % 360 === 90) {
			if (width > height) {
				rect.origin.y += height - width; //needs to be added to origin
			} else if (height > width && isFlipped) {
				rect.origin.x -= height - width;
			}
			rect.size = new ShapeUtils.Size(height, width);
		}

		//construct fragment for linework
		var fragment = new Fragment(rect, z);
		var sceneBounds = this._world.getSceneBounds();
		//we'll rotate the image so the world and the layer can use the correct rotation this way the transformation don't have to be altered
		//first create a canvas element and make sure there is enough space to rotate the image (nothing gets lost)
		var canvas = document.createElement("canvas");
		var squareSize = Math.max(sceneBounds.size.width, sceneBounds.size.height);
		if (rotation % 360 === 90 || rotation % 360 === 270) {
			canvas.width = squareSize;
			canvas.height = squareSize;
		} else {
			canvas.width = sceneBounds.size.width;
			canvas.height = sceneBounds.size.height;
		}

		//do the rotation
		var context = canvas.getContext("2d");
		context.translate(canvas.width / 2, canvas.height / 2);
		if (isFlipped) {
			context.scale(-1, 1);
		}
		context.rotate(((360 - (rotation % 360)) / 180) * Math.PI);

		context.translate(-canvas.width / 2, -canvas.height / 2);
		context.drawImage(this._canvas[0], 0, 0); //draw the canvas, prevent performance issues by not creating images

		//set the canvas in the fragment as image that will be used as buffer
		fragment.image = canvas;
		fragment.fetched = true;

		if (this._layerModel.setBuffer !== undefined) {
			this._layerModel.setBuffer(fragment);
		}
	}
};

/**
 * Remove the buffer for lineworks (if any)
 */
ViewportLayer.prototype.panningStopped = function() {
	if (
		this._layerModel !== null &&
		this._layerModel.hasLinework(
			this._view
				.getPanel()
				.getModel()
				.getName()
		) &&
		this._layerModel.setBuffer !== undefined
	) {
		this._layerModel.setBuffer(null);
		this.throttledRender();
	}
};

/**
 * Render the layer
 * !! note this method should be avoided, use throttledRender() instead!
 */
ViewportLayer.prototype.render = function() {
	if (
		!this._layerModel ||
		!this._world ||
		!this._transform ||
		(this._world.getWorldSize().width === 0 && this._world.getWorldSize().height === 0) ||
		!this.isVisible()
	) {
		// we do not want to draw if there is no this._layerModel, this._world, this._transform
		// or if the worldsize is 0x0
		// or if we are hidden

		return;
	}

	// hide all popovers/popups when an animation is ongoing
	if (this._world.isAnimating()) {
		this.hideAllPopovers();
		this.hideAllPopups();
	}

	// get the canvas drawing context
	var ctx = this.getCanvasContext();

	// set the default for the context
	ctx.setTransform(1, 0, 0, 1, 0, 0);

	// clear canvas
	ctx.clearRect(0, 0, this._canvas.width() || null, this._canvas.height() || null);

	var t = this.getTransformMatrix();
	ctx.transform(t.a, t.b, t.c, t.d, t.e, t.f);

	//get the current zoom level, depends on the state of the transformation
	var zoom = t.getScaleX();

	// calculate the rectangle in which can be drawn
	var drawrect = this._world.getWorldRectFromScene(this._world.getSceneBounds(), t);

	// render
	this._renderBackground(ctx);
	this._renderFragments(ctx, drawrect, zoom);
	//make sure the zoom is depending on the real zoom so make sure the initial zoom isn't involved
	this._renderLinework(ctx, drawrect, zoom, t);

	this._layerModel.render(ctx, drawrect, zoom, t);
	// show popups if the world is not animating
	if (!this._world.isAnimating()) {
		this.drawPopovers();
		this.drawPopups();
	}
};

/**
 * Draw the background as it is defined in the canvas
 * @param ctx
 */
ViewportLayer.prototype._renderBackground = function(ctx) {
	var bgoptions = this._layerModel.getBackgroundOptions();
	if (_.size(bgoptions) > 0) {
		ctx.save();

		for (var k in bgoptions) {
			if (bgoptions.hasOwnProperty(k)) {
				ctx[k] = bgoptions[k];
			}
		}
		var bounds = this._world.getWorldBounds();
		var tr = this._layerModel.getTransform();
		var layerTransform = tr.createInverse();
		bounds = layerTransform.transformedRect(bounds);
		ctx.fillRect(bounds.origin.x, bounds.origin.y, bounds.size.width, bounds.size.height);

		ctx.restore();
	}
};

/**
 * Render all fragments on the canvas
 * @param {Context2D} ctx
 * @param {ShapeUtils.Rect} drawrect
 * @param {Number} zoom
 */
ViewportLayer.prototype._renderFragments = function(ctx, drawrect, zoom) {
	var fragments;
	if (this._layerModel.hasGraphics()) {
		if (!this._layerModel.shouldAutoFetchFragments()) {
			// rely on the model for the fragments
			fragments = this._layerModel.fragments(drawrect, zoom);
		} else {
			// auto fetching is enabled, get the fragments from the layer
			fragments = this.getApplicableFragments(drawrect, zoom);
		}

		// pixel perfect drawing, requires us to draw in pixel domain, not in the document domain (mm)

		// first store the current transformation
		ctx.save();

		// scale to pixel coordinates
		zoom = this._world.getZoom();
		ctx.scale(1 / zoom, 1 / zoom);

		// draw fragments in pixel domain
		for (var i = 0; i < fragments.length; i++) {
			var fragment = fragments[i];

			if (fragment.isFetched()) {
				var fragmentRect = fragment.getRect();

				var x = Math.floor(fragmentRect.origin.x * zoom);
				var y = Math.floor(fragmentRect.origin.y * zoom);
				var width = Math.ceil(fragmentRect.size.width * zoom);
				var height = Math.ceil(fragmentRect.size.height * zoom);

				if (Math.abs(fragment.getZoom() - zoom) < 0.0005) {
					// we are drawing in the same zoomlevel, draw at the natural height/width of the fragment
					width = fragment.getNaturalWidth();
					height = fragment.getNaturalHeight();
				}
				ctx.drawImage(fragment.image, x, y, width, height);
			}
		}

		// restore the transformation
		ctx.restore();

		/* un-uncomment for drawing rectangles around fragments (can be used easily see which fragments are drawn) 
     * normal fragments have red border + fill
     * perfect fit is shown in green
     * dirty fragments have blue border + fill
     * in the corner of every fragment the draw sequence number is shown
     *
    for(var i=0; i<fragments.length; i++){
        var fragment = fragments[i];
        if (fragment.isFetched()){
            
            var fragmentRect = fragment.getRect();
            ctx.beginPath();
            ctx.rect(fragmentRect.origin.x, fragmentRect.origin.y, fragmentRect.size.width, fragmentRect.size.height);
            ctx.lineWidth = 2/zoom;
            if(fragment.isInitial){
                ctx.strokeStyle = 'yellow';
                ctx.fillStyle = "rgba(0, 255, 255, 0.2)";
            } else {
                 if (fragment.perfectFit){
                    ctx.strokeStyle = 'green';
                    ctx.fillStyle = "rgba(0, 255, 0, 0.2)";
                    fragment.perfectFit = false;
                } else if (fragment.isDirty()){
                    ctx.strokeStyle = 'blue';
                    ctx.fillStyle = "rgba(0, 0, 255, 0.1)";
                } else {
                    ctx.strokeStyle = 'red';
                    ctx.fillStyle = "rgba(255, 0, 0, 0.1)";
                }
            }
            
           
            
            ctx.fill();
            ctx.stroke();	 
            ctx.font = (10 / zoom) +"pt Verdana";
            ctx.fillStyle = "black";
            ctx.fillText(i, fragmentRect.origin.x + (2 / zoom)  , fragmentRect.origin.y + (12 / zoom));
            
        }
    }
    /* */
	}
};

/**
 * Render all the linework (use a buffer if needed)
 * @param {Context2D} ctx
 * @param {ShapeUtils.Rect} drawrect
 * @param {Number} zoom
 * @param {Transformation} t - transformation for this layer
 */
ViewportLayer.prototype._renderLinework = function(ctx, drawrect, zoom, t) {
	if (
		!this._layerModel.hasBuffer(
			this._view
				.getPanel()
				.getModel()
				.getName()
		) ||
		!this._layerModel.hasLinework(
			this._view
				.getPanel()
				.getModel()
				.getName()
		)
	) {
		//only don't draw linework if CAD and panning

		var linework = this._layerModel.getLinework(
			this._world.isAnimating(),
			this._world.isAnimatingRotation(),
			drawrect,
			this._view.getModel().getID(),
			this._view
				.getPanel()
				.getModel()
				.getName(),
			this._isZooming
		);

		if (linework && linework.length > 0) {
			// reset the buffer
			this._buffer = null;

			DrawUtils.drawOverlay(
				this.f,
				ctx,
				linework,
				drawrect,
				_.bind(this.throttledRender, this),
				zoom,
				t,
				this._layerModel.getLineworkColor(),
				this._layerModel.getAlpha()
			);

			if (process.env.NODE_ENV === "development") {
				if (this._debugBuffer) {
					ctx.fillStyle = "rgba(1, 0, 0, 0.4)";
					ctx.fillRect(drawrect.origin.x, drawrect.origin.y, drawrect.size.width, drawrect.size.height);
				}
			}
		}
	} else if (
		this._layerModel.hasLinework(
			this._view
				.getPanel()
				.getModel()
				.getName()
		) &&
		this._layerModel.hasBuffer(
			this._view
				.getPanel()
				.getModel()
				.getName()
		)
	) {
		//if there is Cad and a a buffer render the buffer
		this._renderBuffer(ctx, zoom, t);
	}
};

/**
 * Renders the buffer of the layer if there is a buffer (used for buffer of lineworks)
 * @param {Context2D} ctx
 * @param {Number} zoom
 * @param {Transformation} t - transformation for this layer
 */
ViewportLayer.prototype._renderBuffer = function(ctx, zoom, t) {
	var fragment = this._layerModel.getBuffer();

	if (fragment !== null) {
		ctx.save();
		zoom = this._world.getZoom();

		t = this._world.getTransformation();
		ctx.setTransform(1, 0, 0, 1, 0, 0);
		ctx.transform(t.a, t.b, t.c, t.d, t.e, t.f);
		ctx.scale(1 / zoom, 1 / zoom);

		if (fragment.isFetched()) {
			var fragmentRect = fragment.getRect();

			var x = fragmentRect.origin.x * zoom;
			var y = fragmentRect.origin.y * zoom;
			var width = fragmentRect.size.width * zoom;
			var height = fragmentRect.size.height * zoom;

			var imageCanvas = fragment.getImage();
			if (Math.abs(fragment.getZoom() - zoom) < 0.0005) {
				// we are drawing in the same zoomlevel, draw at the natural height/width of the fragment
				width = imageCanvas.width;
				height = imageCanvas.height;
			}
			ctx.drawImage(imageCanvas, x, y, width, height);
		}

		ctx.restore();
	}
};

/**************************************************************
 * POPOVERS
 **************************************************************/

/**
 * A popover was added
 * @param {String} id - the id of the popover
 */
ViewportLayer.prototype.onPopoverAdded = function(id) {
	var popoverModel = this._layerModel.getPopoverModel(id);

	//first check if the the popover is already added to the layer model:
	if (this._layerModel.getPopover(id) === undefined) {
		var popoverObj = {};

		// Create a div for the popover
		var popoverDiv = $("<div>").css("pointer-events", "none"); // ignore mouse events for the dummy popover div

		// Add the div to the viewport
		this.f
			.getContext()
			.find(".viewport")
			.append(popoverDiv);
		popoverObj.popoverDiv = popoverDiv;

		// Create the popover object
		var popover = new Popover(
			popoverDiv,
			popoverModel.getContext(),
			this._canvas,
			popoverModel.getPreferredPosition(),
			"#fff",
			popoverModel.getClass(),
			popoverModel.getRestrictBorder()
		);

		//For IE9 and IE10: no support for pointer-events css!! hide the popover element and emit the event on the element below popover
		popover.element.on("mouseenter", function(e) {
			$(this).hide();
			var elementBelow = document.elementFromPoint(e.clientX, e.clientY);
			$(this).show();
			$(elementBelow).trigger("mouseenter");
		});

		// Add events
		if (popoverModel.onAddedToContainer) {
			popover.onAddedToContainer(popoverModel.onAddedToContainer);
		}
		if (popoverModel.onDestroy) {
			popover.onDestroy(popoverModel.onDestroy);
		}

		// Set content
		popover.setContent(popoverModel.getContent());

		popoverObj.popover = popover;

		// Store it
		this._layerModel.savePopover(popoverObj, id);

		this.drawPopovers();
	}
};

/**
 * Remove a popover (object)
 * @param {String} id - the id of the popover
 */
ViewportLayer.prototype.onPopoverRemoved = function(id) {
	var popoverObj = this._layerModel.getPopover(id);
	if (popoverObj) {
		popoverObj.popover.hide();
		popoverObj.popover.destroy();
		popoverObj.popoverDiv.remove();

		this._layerModel.removePopover(id);
	}
};

/**
 * Hide all popovers
 */
ViewportLayer.prototype.hideAllPopovers = function() {
	if (this._layerModel !== null) {
		_.each(this._layerModel.getAllPopovers(), function(popoverObj) {
			popoverObj.popover.hide();
			popoverObj.popoverDiv.hide();
		});
	}
};

/**
 * Hide all popovers (divs and popover objects) that are not interactive
 */
ViewportLayer.prototype.hideAllNonInteractivePopovers = function() {
	if (this._layerModel !== null) {
		_.each(
			this._layerModel.getAllPopovers(),
			function(popoverObj, id) {
				var popoverModel = this._layerModel.getPopoverModel(id);
				if (!popoverModel.isInteractive()) {
					popoverObj.popover.hide();
					popoverObj.popoverDiv.hide();
				}
			},
			this
		);
	}
};

/**
 * Redraw all the popovers that are needed to be drawn
 * (typically because the world changed and the popover needs to be displayed from another rect)
 */
ViewportLayer.prototype.drawPopovers = function() {
	// only draw popups if the model is set and the layer is not hidden and the view of this layer is active
	if (this._layerModel && this.isVisible() && this._view.isActive()) {
		_.each(
			this._layerModel.getAllPopovers(),
			function(popoverObj, id) {
				popoverObj.popover.restrict = this._canvas; //make sure to restrict the popover to the current view
				popoverObj.popoverDiv.show();

				var popoverModel = this._layerModel.getPopoverModel(id);
				var frame = popoverModel.getFrame();
				if (frame) {
					var r = this._world.getSceneRectFromWorld(frame, this.getTransformMatrix());

					var viewLocation = this._view.getViewLocation(); //get the location of the view in the viewport
					var titleHeight = 0;
					if (this._view.isTitleVisible()) {
						titleHeight = this._view.getTitleHeight();
					}
					/**
					 * Set the position of the popover div, so the popover can be attached to it
					 */
					popoverObj.popoverDiv.css({
						position: "absolute",
						width: r.size.width + 10 + "px", //10 to make the popover not overlap the tool
						height: r.size.height + 10 + "px",
						top: r.origin.y + viewLocation.y + titleHeight - 5 + "px", //take the box of 10px into account
						left: r.origin.x + viewLocation.x - 5 + "px"
					});
					popoverObj.popover.setPreferredPlacement(popoverModel.getPreferredPosition());

					popoverObj.popover.show();
				}
			},
			this
		);
	}
};

/**************************************************************
 * POPUPS
 **************************************************************/
/**
 * Hide all popups
 */
ViewportLayer.prototype.hideAllPopups = function() {
	// only do stuff when we have a layermodel
	if (this._layerModel) {
		_.each(
			this._layerModel.getPopupModels(),
			function(popupModel) {
				popupModel.getPopup(this).hide();
			},
			this
		);
	}
};

/**
 * Display all popups
 */
ViewportLayer.prototype.drawPopups = function() {
	// only do stuff when we have a layermodel and the layer is shown
	if (this._layerModel && this._canvasIsShown) {
		// get all popup models
		_.each(
			this._layerModel.getPopupModels(),
			function(popupModel) {
				var popup = popupModel.getPopup(this);
				if (popupModel.isShown() && !this._interactive) {
					// Convert the world rect to scene rect.
					var sceneOrigin = this.getScenePointFromWorld(popupModel.getWorldOrigin());
					var sceneSize = popupModel.getWorldSize().copy();

					// 96.0/25.4 is a constant to convert world to pixels (for a standard screen PPI)
					sceneSize.width *= ConversionConstants.MM_TO_PX_RATIO;
					sceneSize.height *= ConversionConstants.MM_TO_PX_RATIO;

					// Is there an offset defined in scene pixels?
					var sceneOffset = popupModel.getSceneOffset();
					if (sceneOffset) {
						// add the offset
						sceneOrigin.x += sceneOffset.width;
						sceneOrigin.y += sceneOffset.height;
					}

					popup.show();
					popup.setRect(new ShapeUtils.Rect(sceneOrigin, sceneSize));
					// only add once
					if (popup.listeners("resized").length === 0) {
						popup.on(
							"resized",
							_.bind(function(sceneRect, autoAdapted) {
								var sceneOffset = popupModel.getSceneOffset();
								if (sceneOffset) {
									// we have to compensate the defined offset
									sceneRect.origin.x -= sceneOffset.width;
									sceneRect.origin.y -= sceneOffset.height;
								}

								// convert the scene rect to world rect
								var worldRect = this.getWorldRectFromScene(sceneRect);
								worldRect.size.width = sceneRect.size.width * ConversionConstants.PX_TO_MM_RATIO;
								worldRect.size.height = sceneRect.size.height * ConversionConstants.PX_TO_MM_RATIO;
								var worldPoint = this.getWorldPointFromScene(sceneRect.origin);
								var worldSize = sceneRect.size.copy();
								worldSize.width *= ConversionConstants.PX_TO_MM_RATIO;
								worldSize.height *= ConversionConstants.PX_TO_MM_RATIO;

								popupModel.onResized(new ShapeUtils.Rect(worldPoint, worldSize), autoAdapted);
							}, this)
						);
					}
					// only add once
					if (popup.listeners("dragged").length === 0) {
						popup.on(
							"dragged",
							_.bind(function(scenePoint) {
								var sceneOffset = popupModel.getSceneOffset();
								if (sceneOffset) {
									// we have to compensate the defined offset
									scenePoint.x -= sceneOffset.width;
									scenePoint.y -= sceneOffset.height;
								}

								popupModel.onDragged(this.getWorldPointFromScene(scenePoint));
							}, this)
						);
					}
				} else {
					popup.hide();
				}
			},
			this
		);
	}
};

export default ViewportLayer;
