/**
 *  @namespace ViewportModule
 */
/**
 *  @namespace ViewportModule
 */
import AbstractModule from "@esko/web-viewer/core/AbstractModule";

import ViewportWidget from "@esko/web-viewer/viewport/ViewportWidget";
import ViewportWidgetModel from "@esko/web-viewer/viewport/model/ViewportWidgetModel";
import AbstractViewportDataAdapter from "@esko/web-viewer/viewport/AbstractViewportDataAdapter";
import ViewportWorld from "@esko/web-viewer/viewport/ViewportWorld";
import ViewerUtils from "@esko/web-viewer/utils/ViewerUtils";
import ShapeUtils from "@esko/shapeutils";
import StringSet from "@esko/stringset";
import $ from "jquery";
import Cache from "@esko/cache";
import ListenerUtil from "@esko/web-viewer/utils/ListenerUtil";
import _ from "underscore";
/**
 * @constant
 */
var IMAGE_CACHE_SIZE = 75;

/**
 * @exports ViewportModule
 * @memberof ViewportModule
 * @augments Core.AbstractModule
 * @constructor
 * @param {Core.ViewerFacade} f
 * @param {Boolean} silenced
 */
var ViewportModule = function(f, silenced, adapter) {
	AbstractModule.call(this, f, silenced, "viewport", adapter || new AbstractViewportDataAdapter());
	ListenerUtil.mixin(this, f);
	this._initializeImageCache();

	this.widgetModel = null;

	this._disabledAnimationSources = [];

	// how many image requests are ongoing?
	this._ongoingImageDataLoads = 0;
	// how many layers might start image requests?
	this._pendingImageRequests = new StringSet();

	this.world = new ViewportWorld(this.f, "viewportWidget", f.getProperty("initialZoom"));
	this.widget = new ViewportWidget(f, this.world);
	f.requestModule("featureDetector", function(featureDetector) {
		// Check if canvas available
		if (featureDetector.hasCanvas()) {
			f.debug.log("Canvas feature available");
		} else {
			f.debug.log("Canvas feature NOT available");
			// Show error message with possibility to go to java viewer
			f.i18n.loadBundle("viewer", function() {
				// to be sure we have the message available
				f.emit("userMessage", "error", f.i18n.t("CanvasNotAvailable"), f.getProperty("javaViewerURL"));
			});
		}
	});

	// viewport widget is always created, otherwise we might run in the concurrency problems...
	this._publishWidget(
		ViewportWidget.widgetName,
		_.bind(function() {
			return this.widget;
		}, this)
	);

	// listen to viewport. should be deprecated...
	this._registerListener("viewport.transform", this.transform, this);
	this._registerListener("viewport.toggleAnimation", this.toggleAnimations, this);
	this._registerListener(
		"viewport.zoomToFit",
		function(options) {
			options = options || {};
			this.zoomToFit(options.animated);
		},
		this
	);
	this._registerListener(
		"viewport.zoomToFitWidth",
		function(options) {
			this.zoomToFitWidth(options.animated);
		},
		this
	);
	this._registerListener(
		"viewport.zoomToFitHeight",
		function(options) {
			this.zoomToFitHeight(options.animated);
		},
		this
	);
	this._registerListener(
		"viewport.zoomOneToOne",
		function(options) {
			this.zoomOneToOne(options.animated, options.worldResolution);
		},
		this
	);
	this._registerListener(
		"viewport.fitInRect",
		function(options) {
			this.fitInRect(options.rect, options.animated, options.constrain);
		},
		this
	);
	this._registerListener(
		"viewport.repaint",
		function() {
			if (this.widget) {
				this.widget.repaint();
			}
		},
		this
	);

	this._registerListener(
		"viewportLayer.startImageDataRequest",
		function(id) {
			this._pendingImageRequests.add(id);

			if (this._ongoingImageDataLoads === 0) {
				this._emit("startLoadingImageData");
			}
		},
		this
	);

	this._registerListener(
		"viewportLayer.finishedImageDataRequest",
		function(id) {
			var isRemoved = this._pendingImageRequests.remove(id);

			if (isRemoved && this._pendingImageRequests.size() === 0 && this._ongoingImageDataLoads === 0) {
				this._emit("finishedLoadingImageData");
			}
		},
		this
	);

	this._registerListener(
		"viewport.getContext",
		function(callback) {
			callback(this.widget.getContext());
		}.bind(this)
	);

	this.f.getDocumentController().whenActivePageChanged(function() {
		if (this.world) {
			this.world.resetOrientation();
		}
	}, this);
	this.f.getDocumentController().whenActiveOrDifferenceOrCadOverlayChangedAndParametersAvailable(function() {
		this.world.zoomToFit(false);
	}, this);
};

ViewportModule.prototype = Object.create(AbstractModule.prototype);
ViewportModule.prototype.constructor = ViewportModule;

/**
 * Sets the viewportwidget model
 * @param {ViewportWidgetModel} widgetModel
 */
ViewportModule.prototype.setWidgetModel = function(widgetModel) {
	this.widget.setWidgetModel(widgetModel);
	this.widgetModel = widgetModel;

	if (this.widgetModel) {
		this._emit("widgetModelAvailable");
	}
};

/**
 * Returns the widget model
 * @returns {ViewportWidgetModel}
 */
ViewportModule.prototype.getWidgetModel = function() {
	return this.widgetModel;
};

/**
 * Initialize the image cache
 */
ViewportModule.prototype._initializeImageCache = function() {
	this._imageCache = new Cache(IMAGE_CACHE_SIZE, true);

	// in readerspread mode we don't want to reload an image if the active/diff page have switched position
	this._imageCache.getAlternativeName = this.f.getDataAdapterForModule("viewport").getAlternativeName;

	/**
	 * Load an item if it is not yet in the cache
	 */
	this._imageCache._loadItem = _.partial(this.f.getDataAdapterForModule("viewport").loadImageFragment, this.f, _, this);
};

/**
 * Indicate the start of loading an image
 * it sends a loading event (and starTLoadingImageData event)
 * @param {String} loadingID - unique string for the loading request
 */
ViewportModule.prototype._startLoadingImageData = function(loadingID) {
	this.f.emit("loading.start", loadingID);
	if (this._ongoingImageDataLoads === 0 && this._pendingImageRequests.size() > 0) {
		this._emit("startLoadingImageData");
	}

	this._ongoingImageDataLoads++;
};

/**
 * Indicate that the loading of an image has finsihed (or failed)
 * @param {String} loadingID - unique string for the loading request
 */
ViewportModule.prototype._stopLoadingImageData = function(loadingID) {
	this.f.emit("loading.stop", loadingID);
	this._ongoingImageDataLoads--;

	if (this._ongoingImageDataLoads === 0) {
		this._emit("finishedLoadingImageData");
	}
};

/**
 * Execute a callback when all image loading has finished
 * @param {Function} callback  - called when no long fetching imag edata
 * @param {Any}      [binding] - binding for the callback function
 */
ViewportModule.prototype.whenFinishedLoadingImageData = function(callback, binding) {
	if (binding) {
		callback = _.bind(callback, binding);
	}
	if (this._ongoingImageDataLoads === 0 && this._pendingImageRequests.size() === 0) {
		callback();
	} else {
		this.once("finishedLoadingImageData", callback);
	}
};

/**
 * Get a reference to the ImageCache for the viewport
 * @returns {Utils.Cache} the ImageChache (a cache of name->Image)
 */
ViewportModule.prototype.getImageCache = function() {
	return this._imageCache;
};

/**
 * Create a new view with options
 * @borrows ViewportModule.ViewportWidgetModel~createView
 * @param {String} id - the id of the view
 * @param {String} title - the title of the view
 * @param {Boolean} isDefaultView - is the view the default view of the viewport (used to show popovers if no view is active)
 */
ViewportModule.prototype.createView = function(viewID, title, isDefaultView) {
	if (this.widgetModel) {
		this.widgetModel.createView(viewID, title, isDefaultView);
	}
};

/**
 * Removes a view with a certain id
 * @borrows ViewportModule.ViewportWidgetModel~removeView as removeView
 * @param {String} viewID - id of the view
 */
ViewportModule.prototype.removeView = function(viewID) {
	if (this.widgetModel) {
		this.widgetModel.removeView(viewID);
	}
};

/**
 * Get the first occurence from a view with a certain name
 * @borrows ViewportModule.ViewportWidgetModel~getFirstViewOccurrence
 * @param {String} viewName
 * @returns {ViewportModule.ViewportView}
 */
ViewportModule.prototype.getFirstViewOccurrence = function(viewName) {
	if (this.widget) {
		return this.widget.getFirstViewOccurrence(viewName);
	}
};

/**
 * Hide or show the view with the specified id
 * @borrows ViewportModule.ViewportWidgetModel~toggleView
 * @param {String} id - Id of the view
 * @param {Boolean} show - True if the view should be shown, false otherwise
 */
ViewportModule.prototype.toggleView = function(viewID, toggle) {
	if (this.widgetModel) {
		this.widgetModel.toggleView(viewID, toggle);
	}
};

/**
 * Set the order of the views
 * @borrows ViewportModule.ViewportWidgetModel~setViewsOrder
 * @param {Array.<String>} ordered keys
 */
ViewportModule.prototype.setViewsOrder = function(orderedKeys) {
	if (this.widgetModel) {
		this.widgetModel.setViewsOrder(orderedKeys);
	}
};
/**
 * Set a new layout mode for the views
 * @borrows ViewportModule.ViewportWidgetModel~setViewsLayoutMode
 * @param {ViewportConstants.ViewsLayoutMode} viewsLayoutMode - The new views layout mode
 */
ViewportModule.prototype.setViewsLayoutMode = function(viewsLayoutMode) {
	if (this.widgetModel) {
		this.widgetModel.setViewsLayoutMode(viewsLayoutMode);
	}
};

/**
 * Add a layer (this can be for all views or for a specific view)
 * @borrows ViewportModule.ViewportWidgetModel~addLayer
 * @param {String} layerName - name of the layer
 * @param {LayerModel} layerModel - model of the layer
 * @param {String} [viewID] - if specified, only add the layer to the view with this id. If not specified, add it to all views (and all future views)
 */
ViewportModule.prototype.addLayer = function(layerName, layerModel, viewID) {
	if (this.widgetModel) {
		this.widgetModel.addLayer(layerName, layerModel, viewID);
	}
};

/**
 * Remove a layer from the views
 * @borrows ViewportModule.ViewportWidgetModel~removeLayer
 * @param {String} layerName - name of the layer
 * @param {String} [viewID] - if specified, only remove the layer from the view with this id. If not specified, remove it from all views
 */
ViewportModule.prototype.removeLayer = function(layerName, viewID) {
	if (this.widgetModel) {
		this.widgetModel.removeLayer(layerName, viewID);
	}
};

/**
 * @borrows ViewportModule.ViewportWidgetModel~hideLayer
 * @param {String} layerName
 * @param {Boolean} visible
 * @param {String} [viewName]
 */
ViewportModule.prototype.toggleLayer = function(layerName, visible, viewName) {
	if (this.widgetModel) {
		this.widgetModel.toggleLayer(layerName, visible, viewName);
	}
};

/**
 * @borrows ViewportModule.ViewportWorld~transform
 * @param {Object} options
 */
ViewportModule.prototype.transform = function(options) {
	this.world.transform(options);
};

/**
 * @borrows ViewportModule.ViewportWorld~toggleAnimations
 * @param toggle
 */
ViewportModule.prototype.toggleAnimations = function(source, toggle) {
	if (!toggle) {
		this._disabledAnimationSources.push(source);
	} else {
		this._disabledAnimationSources = _.without(this._disabledAnimationSources, source);
	}
	this.world.toggleAnimation(this._disabledAnimationSources.length === 0);
};

/**
 * @borrows ViewportModule.ViewportWorld~zoomToFit
 * @param {Boolean} animated
 */
ViewportModule.prototype.zoomToFit = function(animated) {
	this.world.zoomToFit(animated);
};

/**
 * @borrows ViewportModule.ViewportWorld~zoomToFitWidth
 * @param {Boolean} animated
 */
ViewportModule.prototype.zoomToFitWidth = function(animated) {
	this.world.zoomToFitWidth(animated);
};

/**
 * @borrows ViewportModule.ViewportWorld~zoomToFitHeight
 * @param {Boolean} animated
 */
ViewportModule.prototype.zoomToFitHeight = function(animated) {
	this.world.zoomToFitHeight(animated);
};

/**
 * @borrows ViewportModule.ViewportWorld~zoomOneToOne
 * @param {Boolean} animated
 * @param {Number} [worldResolution]
 */
ViewportModule.prototype.zoomOneToOne = function(animated, worldResolution) {
	this.world.zoomOneToOne(animated, worldResolution);
};

/**
 * @borrows ViewportModule.ViewportWorld~getZoomPercentage
 * @returns {Number} The current zoom percentage
 */
ViewportModule.prototype.getZoomPercentage = function() {
	return this.world.getZoomPercentage();
};

/**
 * @borrows ViewportModule.ViewportWorld~getNumberOfZoomLevels
 * @returns {Number} The number of zoom levels
 */
ViewportModule.prototype.getNumberOfZoomLevels = function() {
	return this.world.getNumberOfZoomLevels();
};

/**
 * @borrows ViewportModule.ViewportWorld~getMinZoom
 * @returns {Number} The min zoom level of the world
 */
ViewportModule.prototype.getMinZoom = function() {
	return this.world.getMinZoom();
};

/**
 * @borrows ViewportModule.ViewportWorld~getMaxZoom
 * @returns {Number} The max zoom level of the world
 */
ViewportModule.prototype.getMaxZoom = function() {
	return this.world.getMaxZoom();
};

/**
 * @borrows ViewportModule.ViewportWorld~getResolution
 * @returns {Number}
 */
ViewportModule.prototype.getResolution = function() {
	return this.world.getZoom();
};

/**
 * @borrows ViewportModule.ViewportWorld~getRotation
 * @returns {Number}
 */
ViewportModule.prototype.getRotation = function() {
	return this.world.getRotation();
};

/**
 * @borrows ViewportModule.ViewportWorld~isFlip
 * @returns {Boolean}
 */
ViewportModule.prototype.isFlip = function() {
	return this.world.isFlip();
};

/**
 * @borrows ViewportModule.ViewportWorld~fitInRect
 * @param rect
 * @param animated
 * @param constrain
 */
ViewportModule.prototype.fitInRect = function(rect, animated, constrain) {
	this.world.fitInRect(rect, animated, constrain);
};

/**
 * @borrows ViewportModule.ViewportWidgetModel~setPanelsLayoutMode
 */
ViewportModule.prototype.setPanelsLayoutMode = function(layoutMode) {
	if (this.widgetModel) {
		this.widgetModel.setPanelsLayoutMode(layoutMode);
	}
};

/**
 * @borrows ViewportModule.ViewportWidgetModel~getPanelModel
 */
ViewportModule.prototype.getPanelModel = function(name) {
	if (this.widgetModel) {
		return this.widgetModel.getPanelModel(name);
	}
};

/**
 * @borrows ViewportModule.ViewportWidgetModel~addPanel
 */
ViewportModule.prototype.addPanel = function(name, overwriteOptions, disabledLayers) {
	if (this.widgetModel) {
		return this.widgetModel.addPanel(name, overwriteOptions, disabledLayers);
	}
};

/**
 * @borrows ViewportModule.ViewportWidgetModel~removePanel
 */
ViewportModule.prototype.removePanel = function(name) {
	if (this.widgetModel) {
		this.widgetModel.removePanel(name);
	}
};

/**
 * Returns the visible rectangle of the viewport.
 * @returns {ShapeUtils.Rect}
 */
ViewportModule.prototype.getVisibleRect = function() {
	return this.world.getTargetViewPort();
};

/**
 * Returns the world bounds
 * @returns {ShapeUtils.Rect}
 */
ViewportModule.prototype.getWorldBounds = function() {
	return this.world.getWorldBounds();
};

/**
 * Get the middle parameter of the viewport
 * @returns {ShapeUtils.Point} middle
 */
ViewportModule.prototype.getMiddle = function() {
	return this.world.getMiddle();
};

/**
 * Get the zoom level of the viewport
 * @returns {Number} zoomlevel
 */
ViewportModule.prototype.getZoom = function() {
	return this.world.getZoom();
};

/**
 * @override
 */
ViewportModule.prototype.getState = function() {
	var state = {
		// world
		world: {
			rotation: this.world.getRotation(),
			offsetSize: this.world.getOffsetSize(),
			zoom: this.world.getZoom(),
			isFullView: this.world.isFullView(),
			sceneSize: this.widget.getSize(),
			worldBounds: this.world.getWorldBounds(),
			flip: this.world.isFlip(),
			flipViewFromBack: this.world.isFlipViewFromBack(),
			visibleRect: this.world.getTargetViewPort()
		}
	};
	return state;
};

/**
 * @override
 * @param state
 */
ViewportModule.prototype.setState = function(state) {
	this.f.getDocumentController().whenActivePageChangedAndParametersAvailable(
		_.bind(function() {
			this.f.whenModuleAvailable(
				"viewport",
				function() {
					// This module is ready
					if (state !== undefined) {
						// World
						var worldState = state.world;
						if (worldState !== undefined) {
							this.world.setWorldBounds(worldState.worldBounds);
							this.world.setInitialTransformation({
								fixedrotation: worldState.rotation,
								fixedoffset: worldState.offsetSize,
								fixedzoom: worldState.zoom,
								flip: worldState.flip,
								flipViewFromBack: worldState.flipViewFromBack
							});
							this.world.setSceneSize(new ShapeUtils.Size(worldState.visibleRect.size.width, worldState.visibleRect.size.height));
						}
					}
				},
				this
			);
		}, this)
	);
};

/**
 * Get the viewportworld
 * @returns {ViewportWorld}
 */
ViewportModule.prototype.getWorld = function() {
	return this.world;
};

/**
 * Set the background color of the widget
 * @borrows ViewportModule.ViewportWidget~setBackgroundColor
 * @param {String} [hexColor] - New background color in hex format without # or none to set none
 */
ViewportModule.prototype.setWidgetBackgroundColor = function(hexColor) {
	this.widget.setBackgroundColor(hexColor);
};

/**
 * The callback is called when the widgetmodel is set and the viewport can be used...
 * @param {Function} callback - called when the viewport is ready
 * @param {Object}   [binding]  - binding for the callback
 */
ViewportModule.prototype.whenWidgetModelAvailable = function(callback, binding) {
	if (binding) {
		callback = _.bind(callback, binding);
	}

	if (this.widgetModel) {
		callback();
	} else {
		this.once("widgetModelAvailable", callback);
	}
};

export default ViewportModule;
