import ListenerUtil from "@esko/web-viewer/utils/ListenerUtil";
import ShapeUtils from "@esko/shapeutils";
import ViewportView from "@esko/web-viewer/viewport/ViewportView";
import ViewportConstants from "@esko/web-viewer/viewport/ViewportConstants";
import _ from "underscore";
import $ from "jquery";
/**
 * A viewport panel is a representation of a bunch of viewportviewmodels. It creates view objects for every viewmodel is should show
 * ViewportPanel is responsible for the layout of its panel
 * @mixes Utils.ListenerUtilMixin
 * @param {Core.ViewerFacade} f
 * @param {ViewportModule.ViewportPanelModel} panelModel - the model for the panel
 * @param {ViewPortModule.ViewportWidgetModel} widgetModel - the widgetmodel
 * @param {ViewportModule.ViewportWorld} viewportWorld - the viewport world
 */
var ViewportPanel = function(f, panelModel, widgetModel, viewportWorld) {
	this.f = f;
	ListenerUtil.mixin(this, f);

	/**
	 * @private a map of the existing ViewObjects
	 */
	this._viewObjects = {};
	/**
	 * @private
	 */
	this._widgetModel = widgetModel;
	/**
	 * @private
	 */
	this._panelModel = panelModel;
	/**
	 * @private
	 */
	this._viewportWorld = viewportWorld;

	/**
	 * This is the element in which the panel will draw itself
	 */
	this.element = $("<div/>")
		.attr("name", panelModel.getName())
		.css({
			top: 0,
			left: 0
		})
		.addClass("viewportPanel");
	this.container = $("<div />").addClass("viewportPanelContainer");
	this.element.append(this.container);

	/**
	 * Debounced version of _layout()
	 * Only .layout() should be used
	 * @private
	 */
	this._debouncedLayout = _.debounce(this._layout, 50);
	this.updateOverwriteOptions = _.debounce(this._updateOverwriteOptions, 20);

	// start listening to the correct events
	this._registerListener("changed.overwriteOption.**", this.updateOverwriteOptions, this, this._panelModel);
	this._registerListener("changed.disabledLayers", this.updateDisabledLayers, this, this._panelModel);
	this._registerListener("changed.headerText", this.updatePanelHeader, this, this._panelModel);

	this._registerListener("changed.view.toggle.*", this.layout, this, this._widgetModel);

	this._registerListener("changed.view.order", this.layout, this, this._widgetModel);
	// changed.view.add is not necessary as a "order" is also fired
	this._registerListener(
		"changed.view.remove",
		function(viewModel) {
			this._destroyView(viewModel.getID());
		},
		this,
		this._widgetModel
	);

	this._registerListener("changed.layout.*", this.layout, this, this._widgetModel);

	//Listen to when panning is started
	this._registerListener(
		"viewport.startPanning",
		_.bind(function() {
			this._isPanning = true;
			_.each(this._viewObjects, function(view) {
				view.panningStarted();
			});
		}, this)
	);

	// Listen to when panning is stopped
	this._registerListener(
		"viewport.stopPanning",
		_.bind(function() {
			this._isPanning = false;
			_.each(this._viewObjects, function(view) {
				view.panningStopped();
			});
		}, this)
	);

	//
	this._registerListener(
		"viewport.startZooming",
		_.bind(function() {
			this._isZooming = true;
			_.each(this._viewObjects, function(view) {
				view.zoomingStarted();
			});
		}, this)
	);

	//
	this._registerListener(
		"viewport.stopZooming",
		_.bind(function() {
			this._isZooming = false;
			_.each(this._viewObjects, function(view) {
				view.zoomingStopped();
			});
		}, this)
	);

	this.repaint = _.throttle(
		_.bind(function() {
			this._repaint();
		}, this),
		40
	);
};

/**
 * Destroy the panel
 * This removes all listeners and DOM elements
 */
ViewportPanel.prototype.destroy = function() {
	this._unregisterAllListeners();

	// destroy all views
	_.each(this._viewObjects, function(view) {
		view.destroy();
	});

	this.element.remove();
};

/**
 * Is the viewport widget active, is the mouse cursor in the viewport at this moment.
 * @returns {Boolean}
 */
ViewportPanel.prototype.isViewportWidgetActive = function() {
	return this._widgetModel.isWidgetActive();
};

/**
 * Update the layout of the panel
 */
ViewportPanel.prototype.layout = function() {
	if (!this.hasBeenLayouted) {
		this._layout();
	} else {
		this._debouncedLayout();
	}
};

/**
 * Get the model of the panel
 * @returns {ViewportPanelModel}
 */
ViewportPanel.prototype.getModel = function() {
	return this._panelModel;
};
/**
 * Update the layout of the panel
 *
 * layout will also create view objects if no object exists yet
 * !! only use ViewportPanel.layout();
 * @private
 */
ViewportPanel.prototype._layout = function() {
	var viewsLayoutMode = this._widgetModel.getResolvedViewsLayoutMode();
	var viewKeys = this._widgetModel.getShownViewKeys();
	this.viewCount = {
		total: viewKeys.length
	};

	if (viewsLayoutMode === ViewportConstants.ViewsLayoutMode.HORIZONTAL) {
		this._layoutViewsHorizontal(viewKeys);
	} else if (viewsLayoutMode === ViewportConstants.ViewsLayoutMode.VERTICAL) {
		this._layoutViewsVertical(viewKeys);
	} else {
		this._layoutViewsAutomatic(viewKeys);
	}

	if (this.viewCount.total > 0) {
		this.hasBeenLayouted = true;
	}

	this.updatePanelHeader();
};

/**
 * Layout the views vertically (left to right)
 * @param {Array.<String>} viewKeys
 * @private
 */
ViewportPanel.prototype._layoutViewsVertical = function(viewKeys) {
	var divSize = new ShapeUtils.Size(this.container.width() || null, this.container.height() || null);

	var x = viewKeys.length;
	var y = 1;

	this.viewCount.x = x;
	this.viewCount.y = y;

	var viewWidth = Math.ceil(divSize.width / x);
	var viewHeight = Math.ceil(divSize.height / y);

	var viewSize = new ShapeUtils.Size(viewWidth, viewHeight);
	this._widgetModel.setSizeForView(viewSize);
	_.each(
		viewKeys,
		function(viewID, index) {
			var view = this.getView(viewID);

			// show a border on the right for all views, except the last one
			view.setBorderVisibility(false, index < viewKeys.length - 1);

			view.setRect(ShapeUtils.rectMake(viewWidth * index, 0, viewWidth, viewHeight));
		},
		this
	);
};

/**
 * Layout the views horizontally (top to bottom)
 * @param {Array.<String>} viewKeys
 * @private
 */
ViewportPanel.prototype._layoutViewsHorizontal = function(viewKeys) {
	var divSize = new ShapeUtils.Size(this.container.width() || null, this.container.height() || null);

	var x = 1;
	var y = viewKeys.length;

	this.viewCount.x = x;
	this.viewCount.y = y;

	var viewWidth = Math.ceil(divSize.width / x);
	var viewHeight = Math.ceil(divSize.height / y);

	var viewSize = new ShapeUtils.Size(viewWidth, viewHeight);
	this._widgetModel.setSizeForView(viewSize);
	_.each(
		viewKeys,
		function(viewID, index) {
			var view = this.getView(viewID);
			// in vertical mode we don't want to see a border
			view.setBorderVisibility(false, false);

			view.setRect(ShapeUtils.rectMake(0, viewHeight * index, viewWidth, viewHeight));
		},
		this
	);
};
/**
 * Layout the views automatically
 * @param {Array.<String>} viewKeys
 * @private
 */
ViewportPanel.prototype._layoutViewsAutomatic = function(viewKeys) {
	var divSize = new ShapeUtils.Size(this.container.width() || null, this.container.height() || null);

	var viewsCount = viewKeys.length;

	var x = viewsCount;
	var y = 1;

	// Math.floor for integer dividing
	if (viewsCount > 1) {
		x = Math.floor(Math.round((Math.sqrt(viewsCount) * divSize.width) / divSize.height));
		if (x < 1) {
			x = 1;
		}
		y = Math.floor((viewsCount + x - 1) / x);
		x = Math.floor((viewsCount + y - 1) / y); // calculate again for best possible spacing
		var empty = x * y - viewsCount;
		if (x > 1 && y > 1 && empty > 0) {
			// try to adjust by +1 to reduce white space
			var x1 = x + 1;
			var y1 = Math.floor((viewsCount + x1 - 1) / x1);
			var empty1 = x1 * y1 - viewsCount;
			var y2 = y + 1;
			var x2 = Math.floor((viewsCount + y2 - 1) / y2);
			var empty2 = x2 * y2 - viewsCount;
			if (Math.min(empty1, empty2) < empty) {
				if (empty1 <= empty2) {
					x = x1;
					y = y1;
				} else {
					x = x2;
					y = y2;
				}
			}
		}
	}

	this.viewCount.x = x;
	this.viewCount.y = y;

	var viewWidth = Math.ceil(divSize.width / x);
	var viewHeight = Math.ceil(divSize.height / y);
	var viewSize = new ShapeUtils.Size(viewWidth, viewHeight);
	this._widgetModel.setSizeForView(viewSize);
	_.each(
		viewKeys,
		function(viewID, index) {
			var view = this.getView(viewID);
			// show a border on the right for every view (except the last one of a row)
			view.setBorderVisibility(false, index % x < x - 1);
			view.setRect(ShapeUtils.rectMake(viewWidth * Math.floor(index % x), viewHeight * Math.floor(index / x), viewWidth, viewHeight));
		},
		this
	);
};

/**
 * Trigger a repaint for all the subviews, one of the overwrite options is changes
 * @private
 */
ViewportPanel.prototype._repaint = function() {
	_.each(
		this._widgetModel.getShownViewKeys(),
		function(viewID, index) {
			var view = this.getView(viewID);
			view.repaint(this._viewportWorld);
		},
		this
	);
};

/**
 * Update the overwrite Options for all views
 * After that repaint all views
 * @private
 */
ViewportPanel.prototype._updateOverwriteOptions = function() {
	var overwriteOptions = this._panelModel.getOverwriteOptions();
	_.each(
		this._widgetModel.getShownViewKeys(),
		function(viewID, index) {
			var view = this.getView(viewID);
			view.setOverwriteOptions(overwriteOptions);
		},
		this
	);
};

/**
 * Update all layer visibilities
 * @param {Array.<String>} previouslyDisabled - the list of previously disabled layers
 * @param {Array.<String>} currentlyDisabled - the list of currently disabled layers
 */
ViewportPanel.prototype.updateDisabledLayers = function(previouslyDisabled, currentlyDisabled) {
	var toEnable = _.difference(previouslyDisabled, currentlyDisabled);
	var toDisable = currentlyDisabled;

	_.each(this._viewObjects, function(view, key) {
		view.toggleLayers(toEnable, true);
		view.toggleLayers(toDisable, false);
	});
};

/**
 * Get the ViewportView object for a certain view
 * @param {String} viewID
 * @returns {ViewportModule.ViewportView}
 */
ViewportPanel.prototype.getView = function(viewID) {
	if (this._viewObjects[viewID]) {
		// view object exists
		return this._viewObjects[viewID];
	} else {
		// create a view object
		return this._createView(viewID);
	}
};
/**
 * Create viewportview object this also injects the view element in to the DOM
 * @private
 * @param {String} viewID - The ID of the view
 * @return {ViewportModule.ViewportView}
 */
ViewportPanel.prototype._createView = function(viewID) {
	var view = new ViewportView(this.f, this._viewportWorld, this._panelModel.getName() + "_" + viewID, this);

	view.setModel(this._widgetModel.getViewModel(viewID));

	view.setOverwriteOptions(this._panelModel.getOverwriteOptions());
	view.toggleLayers(this._panelModel.getDisabledLayers(), false);
	this._viewObjects[viewID] = view;
	this.container.append(view.element);

	return view;
};

/**
 * Destroy a view object as it is no longer needed
 * Note: this function also removes the element from the DOM
 * @private
 * @param {String} viewID
 */
ViewportPanel.prototype._destroyView = function(viewID) {
	if (this._viewObjects[viewID]) {
		this._viewObjects[viewID].destroy();
		delete this._viewObjects[viewID];
	}
};

/**
 * Get a view at a certain point
 * @param {ShapeUtils.Point} point
 * @returns {ViewportView}
 */
ViewportPanel.prototype.getViewAt = function(point) {
	point = point.copy();
	var offset = this.element.offset();
	point.x -= offset.left;
	point.y -= offset.top;

	var viewsLayoutMode = this._widgetModel.getResolvedViewsLayoutMode();

	var viewSize = this._widgetModel.getSizeForView();
	if (viewSize.width === 0 || viewSize.height === 0) {
		return undefined;
	}

	var viewIndex = 0;

	if (viewsLayoutMode === ViewportConstants.ViewsLayoutMode.HORIZONTAL) {
		viewIndex = Math.floor(point.y / viewSize.height);
	} else if (viewsLayoutMode === ViewportConstants.ViewsLayoutMode.VERTICAL) {
		viewIndex = Math.floor(point.x / viewSize.width);
	} else {
		// automatic
		var row = Math.floor(point.y / viewSize.height);
		var column = Math.floor(point.x / viewSize.width);
		viewIndex = row * this.viewCount.x + column;
	}

	return this._viewObjects[this._widgetModel.getIDForShownViewsAtIndex(viewIndex)];
};

/**
 * Get a view if it already exists
 * @param {String} viewName
 * @returns {ViewportModule.Viewportview}
 */
ViewportPanel.prototype.getViewIfExists = function(viewName) {
	return this._viewObjects[viewName];
};

/**
 * Set the  bounds of the viewport panel
 * @param position
 * @param size
 */
ViewportPanel.prototype.setBounds = function(position, size) {
	this.element.css({
		left: position.x,
		top: position.y,
		width: size.width,
		height: size.height
	});
	this.position = position;
	this.layout();
};

/**
 * Get the position of the panel in the viewport.
 * @returns {ShapeUtils.Point}
 */
ViewportPanel.prototype.getPosition = function() {
	if (this.position !== undefined) {
		return this.position.copy();
	} else {
		return new ShapeUtils.Point();
	}
};

/**
 * If the header text has changed, update the header of the first view to show the header text if needed
 */
ViewportPanel.prototype.updatePanelHeader = function() {
	var headerText = this._panelModel.getHeaderText();
	var shownViewCount = this._widgetModel.getShownViewKeys().length;
	var viewID = this._widgetModel.getIDForViewAtIndex(0);
	if (headerText !== null && shownViewCount > 0) {
		// set the header of the first view to the headerText
		this._viewObjects[viewID].setTitle(headerText);
	} else if (shownViewCount > 1) {
		// Show the composite title
		this.f.i18n.loadBundle(
			"viewer",
			function() {
				this._viewObjects[viewID].setTitle(this.f.i18n.t("CaptionComposite"));
			},
			this
		);
	} else if (shownViewCount === 1) {
		this._viewObjects[viewID].setTitle(null);
	}
};

export default ViewportPanel;
