import ViewportLayer from "./ViewportLayer";
import PrivateChannel from "@esko/web-viewer/core/PrivateChannel";
import ConversionConstants from "@esko/web-viewer/utils/ConversionConstants";
import $ from "jquery";
var SVGNS = "http://www.w3.org/2000/svg";

/**
 * Create a ViewportLayer that displays svg documents
 *
 * @exports ViewportLayer
 * @memberof ViewportModule
 * @class A layer is responsible to render the content of one specific part of a view
 * @constructor
 * @mixes Utils.ListenerUtilMixin
 *
 * @param {Core.ViewerFacade} f
 * @param {ViewportModule.viewportView} viewportView - view of the layer
 * @param {ViewportModule.ViewportWorld} world - the world of the Viewport
 * @param {Transformation} viewTransform - transformation dictated by the view
 * @param {String} id - the ID of the Layer
 */

/**
 * Helper function to traverse an xml element and all its children
 */
function traverse(fn, elem) {
	fn(elem);
	if (elem.childNodes) {
		Array.prototype.forEach.call(elem.childNodes, traverse.bind(undefined, fn));
	}
}

var ViewportSvgLayer = function(f, viewportView, world, viewTransform, id) {
	ViewportLayer.call(this, f, viewportView, world, viewTransform, id);
	PrivateChannel.installTo(this, f, "viewport.svgLayer");

	this._svg = $(document.createElementNS(SVGNS, "svg")).css({
		position: "absolute",
		top: 0,
		left: 0,
		bottom: 0,
		right: 0,
		width: "100%",
		height: "100%"
	});

	/**
	 * @member {JQueryObject} SVG Element used for transforming the svg
	 * @private
	 */
	this._svgTransform = $(document.createElementNS(SVGNS, "g")).appendTo(this._svg);

	/**
	 * @member {JQueryObject} SVG Element used for offsetting if readerspread
	 * @private
	 */
	this._readerSpreadTransform = $(document.createElementNS(SVGNS, "g")).appendTo(this._svgTransform);

	/**
	 * @member {JQueryObject} SVG Element used to give the SVG a standard white background
	 * @private
	 */
	this._svgRect = $(document.createElementNS(SVGNS, "rect")).appendTo(this._readerSpreadTransform);

	this._svgContainer = $(document.createElementNS(SVGNS, "g")).appendTo(this._readerSpreadTransform);

	f.on("viewControlsModel.changed.imageBackgroundColor", this._updateSVGRect.bind(this));
	f.on("viewControlsModel.changed.backgroundColorActive", this._updateSVGRect.bind(this));
};

ViewportSvgLayer.prototype = Object.create(ViewportLayer.prototype);
ViewportSvgLayer.prototype.constructor = ViewportSvgLayer;

/**
 * @override
 */
ViewportSvgLayer.prototype.getLayerElement = function() {
	return this._svg;
};

/**
 * @override
 * render the svg in the viewport
 */
ViewportSvgLayer.prototype.render = function() {
	if (
		!this._layerModel ||
		(this._layerModel.getPage() && !this._layerModel.getPage().hasSvg()) ||
		!this._world ||
		!this._transform ||
		(this._world.getWorldSize().width === 0 && this._world.getWorldSize().height === 0) ||
		!this.isVisible()
	) {
		// we do not want to draw if there is no this._layerModel, this._world, this._transform
		// or if the worldsize is 0x0
		// or if we are hidden

		return;
	}
	this._setSvgElement();

	// hide all popovers/popups when an animation is ongoing
	if (this._world.isAnimating()) {
		this.hideAllPopovers();
		this.hideAllPopups();
	}

	this._updateReaderSpreadTransform();
	this._updateSVGRect();
	this._updateTransformation();

	// show popups if the world is not animating
	if (!this._world.isAnimating()) {
		this.drawPopovers();
		this.drawPopups();
	}
	this._emit("repainted");
};

/**
 * @override
 */
ViewportSvgLayer.prototype.destroy = function() {
	this._svg.remove();
	ViewportLayer.prototype.destroy.call(this);
};

/**
 * @override
 */
ViewportSvgLayer.prototype.setLayerModel = function(layerModel) {
	ViewportLayer.prototype.setLayerModel.call(this, layerModel);

	this._svgContainer.empty();
	this._appendSvgElement();

	this._registerListener(
		"changed.visible",
		function() {
			if (this.isVisible()) {
				this._svgTransform.appendTo(this._svg);
			} else {
				this._svgTransform.detach();
			}
			this.repaint();
		},
		this,
		this._layerModel
	);

	this._registerListener("pageChanged", this._setSvgElement, this, this._layerModel);
};

/**
 * update the g element to transform the svg
 */
ViewportSvgLayer.prototype._updateTransformation = function() {
	var matrix = this.getTransformMatrix();

	// svg transformation happens on a pixel basis while viewer transformations are in mm
	var scale = ConversionConstants.PX_TO_MM_RATIO;
	matrix.scale(scale, scale);

	var matrixString = "matrix(" + [matrix.a, matrix.b, matrix.c, matrix.d, matrix.e, matrix.f].join(",") + ")";

	this._svgTransform[0].setAttribute("transform-origin", "0 0 0");
	this._svgTransform[0].setAttribute("transform", matrixString);
};

/**
 * set the size of the svg
 */
ViewportSvgLayer.prototype.setSize = function(dim) {
	this._svg[0].setAttributeNS(SVGNS, "width", dim.width);
	this._svg[0].setAttributeNS(SVGNS, "height", dim.height);
	this._svg[0].setAttributeNS(SVGNS, "viewBox", "0 0 " + dim.width + " " + dim.height);

	this.render(); // we have to render immediately otherwise there could be flickering (when size of canvas is set, all pixels are reset)
};

/**
 * Update the white background rect
 */
ViewportSvgLayer.prototype._updateSVGRect = function() {
	var dim = this._layerModel.getSVGSize();
	if (dim) {
		this._svgRect[0].setAttribute("x", 0);
		this._svgRect[0].setAttribute("y", 0);
		this._svgRect[0].setAttribute("width", dim.width / ConversionConstants.PX_TO_MM_RATIO);
		this._svgRect[0].setAttribute("height", dim.height / ConversionConstants.PX_TO_MM_RATIO);
		var viewControls = this.f.getModule("viewControls");
		var fill = "fff";
		if (viewControls && viewControls.isBackgroundColorActive()) {
			var backgroundImageColor = viewControls.getImageBackgroundColor();
			if (backgroundImageColor !== -1) {
				fill = backgroundImageColor.toString(16);
			}
		}
		this._svgRect[0].setAttribute("fill", "#" + fill);
	}
};

/**
 * Updates the transformation that will translate the page if this is viewed in readerspread mode
 */
ViewportSvgLayer.prototype._updateReaderSpreadTransform = function() {
	var offset = this._layerModel.getReaderSpreadOffset();
	var offsetString =
		"translate(" + offset.x / ConversionConstants.PX_TO_MM_RATIO + "," + offset.y / ConversionConstants.PX_TO_MM_RATIO + ")";
	this._readerSpreadTransform[0].setAttribute("transform", offsetString);
};

/**
 * detaches / attaches the correct svg element
 */
ViewportSvgLayer.prototype._setSvgElement = function() {
	if (this._svgElement !== this._layerModel.getSVGElement()) {
		if (this._svgElement) {
			$(this._svgContainer).empty();
		}

		this._appendSvgElement();
	}
};
var counter = 0;
ViewportSvgLayer.prototype._appendSvgElement = function() {
	if (this._layerModel.getSVGElement()) {
		this._svgElement = this._layerModel.getSVGElement();
		var svgClone = $(this._svgElement).clone();
		var dim = this._layerModel.getSVGSize();
		svgClone[0].setAttribute("width", dim.width / ConversionConstants.PX_TO_MM_RATIO);
		svgClone[0].setAttribute("height", dim.height / ConversionConstants.PX_TO_MM_RATIO);
		var ids = [];

		var uniqueCounter = counter++;

		// traverse the xml and add all ids to an array. Then change the id to something unique on the document
		traverse(function(elem) {
			if (elem.hasAttribute && elem.hasAttribute("id")) {
				var id = elem.getAttribute("id");
				ids.push(id);
				elem.setAttribute("id", "WCR_uniq_" + uniqueCounter + "_" + id);
			}
		}, svgClone[0]);

		// traverse the svg element again and replace all references to the saved ids with the new unique id.
		traverse(function(elem) {
			if (elem.hasAttribute) {
				// check if we are dealing with an xml node and not text
				Array.prototype.forEach.call(elem.attributes, function(attr) {
					var attrVal = attr.value;
					ids.forEach(function(id) {
						if (attrVal.indexOf("#" + id) > 0) {
							attr.value = attrVal.replace("#" + id, "#" + "WCR_uniq_" + uniqueCounter + "_" + id);
						}
					});
				});
			}
		}, svgClone[0]);
		this._svgContainer.append(svgClone);
	}
};

export default ViewportSvgLayer;
