import ViewportLayer from "@esko/web-viewer/viewport/ViewportLayer";
import ConversionConstants from "@esko/web-viewer/utils/ConversionConstants";
import $ from "jquery";
import _ from "underscore";
var ViewportVideoLayer = function() {
	ViewportLayer.apply(this, arguments);

	this._container = $("<div>")
		.css({
			position: "absolute",
			top: 0,
			left: 0
		})
		.addClass("enableTextSelection VideoViewportWidget");

	this._whenActivePageChanged = function() {
		this.inited = false;
		this.render();
	}.bind(this);

	this.f.getDocumentController().whenActivePageChanged(this._whenActivePageChanged);

	this.f.getCurrentUser().on(
		"options.changed.PREFERENCES_VIDEO_TRANSCODING",
		function(oldTranscodingName, transcodingName) {
			var groupedTranscodings = this.f.getActivePage().get("transcodings");
			var transcodings = groupedTranscodings[transcodingName];
			this._setTranscodings(transcodings, true);
		}.bind(this)
	);
};

ViewportVideoLayer.prototype = Object.create(ViewportLayer.prototype);
ViewportVideoLayer.prototype.constructor = ViewportVideoLayer;

/**
 * Render the widget
 */
ViewportVideoLayer.prototype.render = function() {
	if (this._container) {
		if (!this.f.getActivePage().hasVideo()) {
			this._container.empty();
			var loader = $("<span/>").text("Loading");
			this._container.append(loader);
			this.inited = false;
		} else {
			if (!this.inited) {
				this._container.empty();

				this._videoElement = $("<video>", {
					attr: {
						preload: "auto"
					},
					on: {
						error: this._handleError.bind(this)
					}
				}).appendTo(this._container);

				this.f.whenModuleAvailable(
					"video",
					function(videoModule) {
						if (this.f.getOption("PREFERENCES_VIDEO_AUTO_PLAY") === 1 || this.f.isMobile()) {
							this._videoElement.prop("autoplay", true);
						}

						if (this.f.getOption("PREFERENCES_VIDEO_LOOP") === 1) {
							this._videoElement.prop("loop", true);
						}

						if (!this.f.isMobile()) {
							this._videoElement[0].volume = Math.min(1, parseFloat(this.f.getOption("PREFERENCES_VIDEO_VOLUME"), 10));
						}

						var groupedTranscodings = this.f.getActivePage().get("transcodings");
						var transcodings = groupedTranscodings[videoModule.getPreferredTranscoding()];
						this._setTranscodings(transcodings, false);

						this._videoElement
							.find("source")
							.last()
							.on("error", this._handleError.bind(this));

						this._videoElement[0].load();

						videoModule.fireNewVideo(this._videoElement[0]);
					}.bind(this)
				);

				this.inited = true;
			}
			this._updateTransformation();
		}
	}
};

ViewportVideoLayer.prototype.destroy = function() {
	ViewportLayer.prototype.destroy.call(this);
	this.f.getDocumentController().off("page.active", this._whenActivePageChanged);
	this._container.remove();
};

/**
 * Adds the source elements to a video element based on a transcoding object
 */
ViewportVideoLayer.prototype._setTranscodings = function(transcodings, restorePreviousState) {
	var time = this._videoElement[0].currentTime;
	var wasPlaying = !this._videoElement[0].paused && !this._videoElement[0].ended;
	this._videoElement[0].pause();
	this._videoElement.empty();
	_.each(transcodings, this._addSourceElements.bind(this));

	this._videoElement.one(
		"loadedmetadata",
		function() {
			this.f.emit("viewport.zoomToFit", {
				animated: false
			});
			if (restorePreviousState) {
				this._videoElement[0].currentTime = time;
				if (wasPlaying) {
					this._videoElement[0].play();
				} else {
					this._videoElement[0].pause();
				}
			}
		}.bind(this)
	);
	this._videoElement[0].load();
};

/**
 * Adds the source elements to a video element based on a transcoding object
 */
ViewportVideoLayer.prototype._addSourceElements = function(transcoding) {
	var url = transcoding.url;
	var mimeindex = url.indexOf("&mime=");
	var src = $("<source>", {
		attr: {
			src: url
		}
	});
	if (transcoding.mime) {
		src.attr("type", transcoding.mime);
	} else if (mimeindex != -1) {
		var mimestr = url.substr(mimeindex + "&mime=".length).split("&")[0];
		src.attr("type", mimestr);
	}
	this._videoElement.prepend(src);
};

ViewportVideoLayer.prototype.getLayerElement = function() {
	return this._container;
};

ViewportVideoLayer.prototype.setSize = function(dim) {
	this._container.css({
		height: dim.height,
		width: dim.width
	});
};

/**
 * update the g element to transform the svg
 */
ViewportVideoLayer.prototype._updateTransformation = function() {
	if (this._videoElement) {
		var matrix = this.getTransformMatrix();

		// video transformation happens on a pixel basis while viewer transformations are in mm
		var scale = ConversionConstants.PX_TO_MM_RATIO;
		matrix.scale(scale, scale);

		var pageWidth =
			this.f
				.getDocumentController()
				.getActivePage()
				.get("size").width / ConversionConstants.PX_TO_MM_RATIO;
		var pageHeight =
			this.f
				.getDocumentController()
				.getActivePage()
				.get("size").height / ConversionConstants.PX_TO_MM_RATIO;
		var transcodingWidth = this._videoElement.width() || null;
		var transcodingHeight = this._videoElement.height() || null;

		matrix.scale(pageWidth / transcodingWidth, pageHeight / transcodingHeight);

		var matrixString = "matrix(" + [matrix.a, matrix.b, matrix.c, matrix.d, matrix.e, matrix.f].join(",") + ")";

		this._videoElement.css({
			"transform-origin": "0 0 0",
			transform: matrixString
		});
	}
};

/**
 * Handles any errors thrown by video or source elements
 *
 * @param {JQueryEvent} e
 */
ViewportVideoLayer.prototype._handleError = function(e) {
	var msg;
	if (e.target.error && e.target.error.code) {
		switch (e.target.error.code) {
			case e.target.error.MEDIA_ERR_ABORTED:
				msg = this.f.i18n.t("MEDIA_ERR_ABORTED");
				break;
			case e.target.error.MEDIA_ERR_NETWORK:
				msg = this.f.i18n.t("MEDIA_ERR_NETWORK");
				break;
			case e.target.error.MEDIA_ERR_DECODE:
				msg = this.f.i18n.t("MEDIA_ERR_DECODE");
				break;
			case e.target.error.MEDIA_ERR_SRC_NOT_SUPPORTED:
				msg = this.f.i18n.t("MEDIA_ERR_SRC_NOT_SUPPORTED");
				break;
			default:
				msg = this.f.i18n.t("MEDIA_ERR_DEFAULT_MESSAGE");
				break;
		}
	} else {
		msg = this.f.i18n.t("MEDIA_ERR_DEFAULT_MESSAGE");
	}
	this.f.emit("error.SEVERE", msg, e);
};

export default ViewportVideoLayer;
