import _ from "underscore";
import ViewportLayer from "@esko/web-viewer/viewport/ViewportLayer";
import ViewportSVGLayer from "./ViewportSvgLayer";
import ViewportVideoLayer from "./ViewportVideoLayer";
import Viewport3DLayer from "./Viewport3DLayer";
import ShapeUtils from "@esko/shapeutils";
import ListenerUtil from "@esko/web-viewer/utils/ListenerUtil";
import $ from "jquery";
/**
 * @exports ViewportView
 * @memberof ViewportModule
 * @class A view is responsible for showing different layers within one div-object
 * The view is mainly steered by its model {ViewportViewModel}
 * @constructor
 * @param {Core.ViewerFacade} f - Facade
 * @param {ViewportWorld} w - the viewport world
 * @param {String} id
 */
var ViewportView = function(f, w, id, viewportPanel) {
	/**
	 * Reference to the world of the viewport
	 */
	this._world = w;
	this.f = f;
	this.id = id;
	ListenerUtil.mixin(this, f);

	/**
	 * Main div for the view
	 */
	this._viewWrapper = $("<div />")
		.addClass("view")
		.attr("name", this.id);
	this.element = this._viewWrapper;

	/**
	 * Override title of model
	 */
	this._title = null;
	this._viewTitle = $("<div />").addClass("viewTitle");
	this._viewWrapper.append(this._viewTitle);
	this._titleShown = false;
	this._viewDiv = $("<div />").css({
		position: "relative"
	});

	this._viewWrapper.append(this._viewDiv);

	// which borders should be shown
	this.borders = {
		left: false,
		right: false
	};

	/**
	 * is the mouse cursor in the view element
	 * @private
	 */
	this._isActive = false;

	/**
	 * location and size of the view in the viewportPanel
	 * @private
	 */
	this._rect = null;

	/**
	 * Panel
	 * @private
	 */
	this._viewportPanel = viewportPanel;

	/**
	 * Layers
	 */
	this.layers = {};
	this._disabledLayers = [];

	this._isPanning = false;

	/**
	 * Model for the view
	 */
	this._viewportViewModel = undefined;

	this._overwriteOptions = {};

	/**
	 * Render for the view with throttle
	 */
	this._throttledRender = _.throttle(this._renderView, 40);
	this.recalculateViewSize();

	this._registerListener(
		"viewport.startInteractive",
		function() {
			_.each(this.layers, function(layer) {
				layer.startInteractive();
			});
		},
		this
	);

	this._registerListener(
		"viewport.stopInteractive",
		function() {
			_.each(this.layers, function(layer) {
				layer.stopInteractive();
			});
		},
		this
	);

	//register listeners to detect if the view is active or not
	this.element.on(
		"mouseenter",
		_.bind(function() {
			this._isActive = true;
		}, this)
	);

	this.element.on(
		"mouseleave",
		_.bind(function() {
			this._isActive = false;
		}, this)
	);
};

/**
 * Get a layer
 * @param {String} name - name of the layer
 * @return {Layer} layer with the given name
 */
ViewportView.prototype.getLayer = function(name) {
	return this.layers[name];
};

/**
 * Checks if the view is currently active
 * @returns {Boolean}
 */
ViewportView.prototype.isActive = function() {
	//if the view is not active, check if the view is the default view and no other view is active (mouse cursor is not in the viewport)
	return this._isActive || (this.isDefaultView() && !this._viewportPanel.isViewportWidgetActive());
};

/**
 * Calculate the closest point given a certain point
 * @param {ShapeUtils.Point} refpoint - Point that needs to be verified (in scene coordinates)
 * @param {Number} distance - snap distance (in px)
 * @param {Array.<String>} [searchFilter] - list of layernames in which should be searched
 * @returns {Object} point - Point that is closest to the point. null if no object was found (in scene coordinates)
 */
ViewportView.prototype.getClosestLineworkPointFromPoint = function(refpoint, distance, searchFilter) {
	var closest = null;

	var layers;
	if (searchFilter) {
		layers = _.filter(this.layers, function(layer) {
			if (layer.getLayerModel()) {
				return _.contains(searchFilter, layer.getLayerModel().getName());
			} else {
				return false;
			}
		});
	} else {
		layers = this.layers;
	}

	_.each(layers, function(layer) {
		var point = layer.getClosestLineworkPointFromPoint(refpoint, distance);
		if (point) {
			//convert the point to the actual world
			if (!closest || point.getDistance(refpoint) < closest.getDistance(refpoint)) {
				closest = point;
				return true;
			}
		}
	});

	return closest;
};

/**
 * Get the closest linework from a certain point (or area if you use fatfinger)
 * !NOTE It will return the first linework found close to the refpoint when searching multiple layers)
 * @param {ShapeUtils.Point} refpoint - point in scene coordinates
 * @param {Array.<String>} [searchFilter] - list of layernames in which should be searched
 * @param {Number} [fatfinger] - search area around refpoint (in px)
 * @returns {LineworkWrapper} the closest linwwork to refpoint
 */
ViewportView.prototype.getClosestLineworkFromPoint = function(refpoint, searchFilter, fatfinger) {
	var linework = null;

	var layers;
	if (searchFilter) {
		layers = _.filter(this.layers, function(layer) {
			if (layer.getLayerModel()) {
				return _.contains(searchFilter, layer.getLayerModel().getName());
			} else {
				return false;
			}
		});
	} else {
		layers = this.layers;
	}

	_.each(layers, function(layer) {
		if (!linework) {
			linework = layer.getClosestLineworkFromPoint(refpoint, fatfinger);
		}
	});

	return linework;
};

/**
 * Recalculate the viewDiv size
 */
ViewportView.prototype.recalculateViewSize = function() {
	var css = {
		width: this._viewWrapper.innerWidth() || null,
		height: (this._viewWrapper.innerHeight() || null) - (this._viewTitle.outerHeight(true) || null)
	};

	// draw a border on the left side if needed
	if (this.borders.left) {
		css.width--;
		css["border-left"] = "1px solid";
	}

	// draw a border on the right side if needed
	if (this.borders.right) {
		css.width--;
		css["border-right"] = "1px solid";
	}
	this._viewDiv.css(css);
};

/**
 * Hide a layer of this view
 * @param {String} layerName - Name of the layer
 * @param {Boolean} hide - True if hide, false if show
 */
ViewportView.prototype.hideLayer = function(layerName, hide) {
	if (this.layers[layerName]) {
		this.layers[layerName].setVisible(!hide);
	}
};

/**
 * Hide all layers of this view
 * @param {Boolean} hide - True if hide, false if show
 */
ViewportView.prototype.hideAllLayers = function(hide) {
	_.each(
		this.layers,
		function(layer, layerName) {
			this.hideLayer(layerName, hide);
		},
		this
	);
};

/**
 * @param {ShapeUtils.Rect} rect - Rectangle
 */
ViewportView.prototype.setRect = function(rect) {
	this._viewWrapper.css({
		position: "absolute",
		width: rect.size.width,
		height: rect.size.height,
		top: rect.top(),
		left: rect.left()
	});
	this._rect = rect;
	this.resizeLayers();
	this.repaint();
};

/**
 * Get the location of the view in the viewport
 * @returns {ShapeUtils.Point}
 */
ViewportView.prototype.getViewLocation = function() {
	var location = new ShapeUtils.Point();
	if (this._rect !== null) {
		location = this._rect.origin.copy(); //only saves the position in the panel (not in the viewport)
	}
	var panelLocation = this._viewportPanel.getPosition(); //get the location of the panel
	if (panelLocation !== undefined) {
		//add the location of the panel to the view location
		location.x += panelLocation.x;
		location.y += panelLocation.y;
	}
	return location;
};

/**
 * Get the div of the view wrapper (including title)
 * @returns {JQueryObj} Div of the view wrapper
 */
ViewportView.prototype.getWrapperContext = function() {
	return this._viewWrapper;
};

/**
 * Get the div of the view only (without title)
 * @returns {JQueryObj} Div of the view only
 */
ViewportView.prototype.getViewContext = function() {
	return this._viewDiv;
};

/**
 * Resize all the layers
 */
ViewportView.prototype.resizeLayers = function() {
	this.recalculateViewSize();
	_.each(
		this.layers,
		_.bind(function(layer, name) {
			this._resizeLayer(layer);
		}, this)
	);
};

/**
 * Resize a layer
 * @param {Layer} layer - layer to be resized
 */
ViewportView.prototype._resizeLayer = function(layer) {
	layer.setSize(new ShapeUtils.Size($(this._viewDiv).width() || null, $(this._viewDiv).height() || null));
};

/**
 * Insert a layer at a certain index
 * @param {Number} index
 * @param {JQueryObject} element
 */
ViewportView.prototype._insertLayerAtIndex = function(index, element) {
	var lastIndex = this._viewDiv.children().length;
	if (index < 0) {
		// the title sits at index 1
		index = Math.max(0, lastIndex + 1 + index);
	}
	if (index < lastIndex) {
		this._viewDiv
			.children()
			.eq(Math.max(0, index))
			.before(element);
	} else {
		this._viewDiv.append(element);
	}
};

/**
 * Add a layer to the view
 * @param {String} name - name of the layer (fetched from the model)
 */
ViewportView.prototype.addLayer = function(name) {
	var layer = this.layers[name];
	var layerModel = this._viewportViewModel.modelForLayer(name);

	if (!layer) {
		if (layerModel && layerModel.getType && layerModel.getType() === "Video") {
			layer = new ViewportVideoLayer(this.f, this, this._world, this._viewportViewModel.getTransform(), this.id + "_" + name);
		} else if (layerModel && layerModel.getType && layerModel.getType() === "svg") {
			layer = new ViewportSVGLayer(this.f, this, this._world, this._viewportViewModel.getTransform(), this.id + "_" + name);
		} else if (layerModel && layerModel.getType && layerModel.getType() === "View3D") {
			layer = new Viewport3DLayer(this.f, this, this._world, this._viewportViewModel.getTransform(), this.id + "_" + name);
		} else {
			layer = new ViewportLayer(this.f, this, this._world, this._viewportViewModel.getTransform(), this.id + "_" + name);
		}
		layer.setOverwriteOptions(this._getOverwriteOptions());
		var index = this._viewportViewModel.indexForLayer(name, this.layers);
		this._insertLayerAtIndex(index, layer.getLayerElement());
		layer.setVisible(this.isLayerVisible(name));
		this.layers[name] = layer;
	}

	if (layerModel) {
		layer.setLayerModel(layerModel);
	}

	this.recalculateViewSize();
	this._resizeLayer(layer);

	return layer;
};

/**
 * Returns the overwrite options
 * @returns {Object}
 */
ViewportView.prototype._getOverwriteOptions = function getOverwriteOptions() {
	return this._overwriteOptions;
};

/**
 * Repaint the view
 * @param {World} [w] - the world
 * @param {Layer} [layer] - if specified, only this layer gets repainted
 */
ViewportView.prototype.repaint = function(w, layer) {
	if (this._viewportViewModel.isShown()) {
		if (w !== undefined) {
			this._world = w;
		}
		this._throttledRender(layer);
	}
};

/**
 * Set the model of the view.
 *
 * Subscribes to the events that the model might send
 * Unsubscribes to the old model (if set)
 */
ViewportView.prototype.setModel = function(viewportViewModel) {
	/*
     * Unsubscribe the old channels on viewportViewModel
     */
	if (this._viewportViewModel) {
		this._unregisterAllListenersForObject(this._viewportViewModel);
	}
	// NOTE probably layers also need to be removed?

	/*
     * Set the new model
     */
	this._viewportViewModel = viewportViewModel;

	/*
     * Subscribe on the new events
     */
	this._registerListener("layerAdded", this._addLayerToView, this, this._viewportViewModel);
	this._registerListener("layerRemoved", this.removeLayer, this, this._viewportViewModel);
	this._registerListener("titleChanged", this._repaintTitle, this, this._viewportViewModel);
	this._registerListener("showChanged", this._showHide, this, this._viewportViewModel);
	this._registerListener("layerIndexChanged", this._layerIndexChanged, this, this._viewportViewModel);

	/*
     * Add the current situation of views
     */
	for (var i = 0; i < this._viewportViewModel.numberOfLayers(); i++) {
		var name = this._viewportViewModel.nameForLayerAtIndex(i);
		this.addLayer(name);
	}
};

/**
 * Get the model for the viewport view
 * @returns {ViewportViewModel}
 */
ViewportView.prototype.getModel = function() {
	return this._viewportViewModel;
};

/**
 * Removes the layer from the view, when a layer in the model was deleted
 */
ViewportView.prototype.removeLayer = function(layername) {
	if (this.layers[layername]) {
		this.layers[layername].destroy();
		delete this.layers[layername];
	} else {
		this.f.debug.debug("Trying to remove a layer that does not exists: " + layername);
	}
};

/**
 * The index of layers have been changed.
 * We remove all layers out of the DOM and add them back in the correct order
 */
ViewportView.prototype._layerIndexChanged = function() {
	var layerNames = _.keys(this.layers);
	_.each(
		layerNames,
		function(layerName) {
			var layer = this.layers[layerName];
			layer.getLayerElement().detach();
		},
		this
	);

	layerNames = this._viewportViewModel.sortLayers(layerNames);

	_.each(
		layerNames,
		function(layerName, i) {
			this._insertLayerAtIndex(i, this.layers[layerName].getLayerElement());
		},
		this
	);
};

/**
 * Add a layer on the view, when a layer in the model was added
 */
ViewportView.prototype._addLayerToView = function(layerName) {
	this.addLayer(layerName);
	this._throttledRender(layerName);
};

/**
 * Show or hide this view
 */
ViewportView.prototype._showHide = function() {
	var show = this._viewportViewModel.isShown();
	if (show) {
		this._viewWrapper.show();
		this.hideAllLayers(false);
	} else {
		this.hideAllLayers(true);
		this._viewWrapper.hide();
	}
};

/**
 * get the height of the title
 * @returns {Number}
 */
ViewportView.prototype.getTitleHeight = function() {
	return $(this._viewTitle).outerHeight(true) || null;
};

/**
 *  Shows this view a title
 *  @returns {Boolean}
 */
ViewportView.prototype.isTitleVisible = function() {
	return this._titleShown;
};

/**
 * Repaint the title based on its value in the model
 */
ViewportView.prototype._repaintTitle = function() {
	var title = this._viewportViewModel.getTitle();
	if (this._title !== null) {
		title = this._title;
	}

	if (title) {
		this._viewTitle.text(title);
		this._viewTitle.attr("title", title);
		if (!this._titleShown) {
			this._viewTitle.show();
			this._titleShown = true;
			this.recalculateViewSize();
			this.resizeLayers();
		}
	} else {
		this._viewTitle.empty();
		if (this._titleShown) {
			this._viewTitle.hide();
			this._titleShown = false;
			this.recalculateViewSize();
			this.resizeLayers();
		}
	}
};

/**
 * Render the view again
 * @param {Layer} [layer] - if specified, only this layer gets repainted
 */
ViewportView.prototype._renderView = function(layer) {
	this._repaintTitle();
	if (layer) {
		var actuallayer = this.layers[layer];
		if (actuallayer) {
			actuallayer.repaint(this._world, this._viewportViewModel.getTransform());
		}
	} else {
		_.each(
			this.layers,
			function(actuallayer, name) {
				actuallayer.repaint(this._world, this._viewportViewModel.getTransform());
			},
			this
		);
	}
};

/**
 * is panning active
 * @returns {Boolean}
 */
ViewportView.prototype.isPanning = function() {
	return this._isPanning;
};

/**
 * Show/hide borders
 * You should call recalculateViewSize (or a function that calls it for you like setRect)
 * after you set this border visibility to update the CSS.
 * @param {Boolean} left - Show border on the left
 * @param {Boolean} right - Show border on the right
 */
ViewportView.prototype.setBorderVisibility = function(left, right) {
	this.borders.left = left;
	this.borders.right = right;
};

/**
 * Destroy the view
 * @param {Boolean} destroyModel - if true also destroy the model
 */
ViewportView.prototype.destroy = function(destroyModel) {
	this._unregisterAllListeners();

	// destroy all layers
	_.each(this.layers, function(layer) {
		layer.destroy();
	});
	// delete all created HTML elements
	this.element.remove();
};

/**
 * Get the id of the view
 * @returns {String}
 */
ViewportView.prototype.getID = function() {
	return this.id;
};

/**
 * Set overwrite options on all layers
 * @param {Object} options
 */
ViewportView.prototype.setOverwriteOptions = function(options) {
	this._overwriteOptions = options;
	_.each(this.layers, function(layer) {
		layer.setOverwriteOptions(options);
	});
};

/**
 * Toggle a specific layer
 * @param {String} layerName - name of the layer
 * @param {Boolean} toggle
 */
ViewportView.prototype.toggleLayer = function(layerName, toggle) {
	var layer = this.layers[layerName];
	if (layer) {
		layer.setVisible(toggle);
	}

	if (toggle) {
		this._disabledLayers = _.without(this._disabledLayers, layerName);
	} else {
		this._disabledLayers.push(layerName);
	}
};
/**
 * Toggle a set of layers
 * @param {Array.<String>} layers - names of the layers to toggle
 * @param {Boolean} toggle
 */
ViewportView.prototype.toggleLayers = function(layers, toggle) {
	_.each(
		layers,
		function(layer) {
			this.toggleLayer(layer, toggle);
		},
		this
	);
};

/**
 * is a specific layer visible?
 * @param {String} layerName
 * @returns {Boolean}
 */
ViewportView.prototype.isLayerVisible = function(layerName) {
	return _.indexOf(this._disabledLayers, layerName) < 0;
};

/**
 * Override the title defined in the model
 * @param {String|null} title
 */
ViewportView.prototype.setTitle = function(title) {
	this._title = title;
	this._repaintTitle();
};

/**
 * is the view the default view
 * @returns {Boolean}
 */
ViewportView.prototype.isDefaultView = function() {
	// note: overwriteoptions.isDefaultView is used only internally
	return this._viewportViewModel.isDefaultView() && this._overwriteOptions.isDefaultView;
};

/**
 * The viewport started panning, tell all the layers that panning is started
 */
ViewportView.prototype.panningStarted = function() {
	this._isPanning = true;
	_.each(this.layers, function(layer) {
		layer.panningStarted();
	});
};

/**
 * The viewport stopped panning, tell all the layers that panning is stopped
 */
ViewportView.prototype.panningStopped = function() {
	this._isPanning = true;
	_.each(this.layers, function(layer) {
		layer.panningStopped();
	});
};

/**
 * The viewport started panning, tell all the layers that panning is started
 */
ViewportView.prototype.zoomingStarted = function() {
	this._isZooming = true;
	_.each(this.layers, function(layer) {
		layer.zoomingStarted();
	});
};

/**
 * The viewport started panning, tell all the layers that panning is started
 */
ViewportView.prototype.zoomingStopped = function() {
	this._isZooming = false;
	_.each(this.layers, function(layer) {
		layer.zoomingStopped();
	});
};

/**
 * Returns the panel this view is contained in
 * @returns {ViewportPanel}
 */
ViewportView.prototype.getPanel = function() {
	return this._viewportPanel;
};

export default ViewportView;
