import AbstractWidget from "@esko/web-viewer/layout/AbstractWidget";
import ViewportWorld from "@esko/web-viewer/viewport/ViewportWorld";
import ShapeUtils from "@esko/shapeutils";
import ToolManager from "@esko/web-viewer/viewport/tools/ToolManager";
import ViewportConstants from "@esko/web-viewer/viewport/ViewportConstants";
import ViewportPanel from "@esko/web-viewer/viewport/ViewportPanel";
import OrderedObject from "@esko/orderedobject";
import ListenerUtil from "@esko/web-viewer/utils/ListenerUtil";
import LayoutConstants from "@esko/web-viewer/layout/LayoutConstants";
import MiniButton from "@esko/web-viewer/layout/component/MiniButton";
import $ from "jquery";
import _ from "underscore";
/**
 *  @exports ViewportWidget
 *  @memberof ViewportModule
 *  @class The viewport widget is responsible to draw the content of the viewer
 *
 * Channels that it listens to:
 *  - "viewport.transform" : applies a transform to the world
 *  - "viewport.zoomToFit" : Fits the content in the widget, so it can be viewed full view (or 90%)
 *  - "viewport.fitInRect" : Fit the viewport to a certain rectangle. E.g. zoom to an annotation
 *  - "viewport.toggleAnimation : enable or disable animations
 *
 *  @param {Core.ViewerFacade} f - Facade
 *  @param {ViewportWorld} world - the viewport world
 */
var ViewportWidget = function ViewPortWidget(f, world) {
	AbstractWidget.call(this, f, LayoutConstants.Widget.VIEWPORT);
	ListenerUtil.mixin(this, f);

	/**
	 * @private
	 */
	this._toolManager = new ToolManager(this.f, this);

	this.setIcon("icon icon-WCR_Viewer-View-2D");
	/**
	 * @private
	 */
	this._world = world;

	/**
	 * @private
	 */
	this._widgetModel = null;

	/**
	 * @private
	 */
	this._panelObjects = {};
	/**
	 * @private
	 */
	this._element = $("<div />").addClass("viewport");

	/**
	 * Repaint the widget
	 */
	this.repaint = _.throttle(
		_.bind(function() {
			this._repaint();
		}, this),
		40
	);

	this.layout = _.debounce(
		_.bind(function() {
			this._layout();
		}, this),
		40
	);

	/**
	 * Did the viewer already zoom to fit
	 * @private {Boolean}
	 */
	this._didZoomToFit = false;

	this._openNewWindow = new MiniButton(
		this.f,
		"open3D",
		"icon-WCR_FullScreen",
		true,
		_.bind(function() {
			this.f.emit("layout.switchLayoutMode", LayoutConstants.LayoutMode.VIEW2D);
		}, this)
	);
	this._openNewWindow.setTooltip("viewer", "expand2DView");

	// listen to transform changed of world
	this._registerListener("transformChanged", this.repaint, this, this._world);
	this.f.getDocumentController().whenActivePageChanged(function(oldPage, newPage) {
		if (oldPage !== newPage) {
			this._didZoomToFit = false;
		}
	}, this);
	this.f.getDocumentController().whenDifferencePageChanged(function(oldPage, newPage) {
		if (oldPage !== newPage) {
			this._didZoomToFit = false;
		}
	}, this);

	this._registerListener("parameters.endUpdate", this._verifyWorld, this, this.f.getDocumentController());
	this.f.on(
		"module.loaded.*",
		function() {
			this._verifyWorld();
		}.bind(this),
		this
	);
	this._verifyWorld();

	//register listeners to detect if the mouse cursor is in the viewport
	this._element.on(
		"mouseenter",
		_.bind(function() {
			this._widgetModel.setWidgetActive(true);
		}, this)
	);

	this._element.on(
		"mouseleave",
		_.bind(function() {
			this._widgetModel.setWidgetActive(false);
		}, this)
	);

	this.f.i18n.loadBundle(
		"viewer",
		_.bind(function() {
			// set the widget title
			this.setTitle(this.f.t("ViewportWidgetTitle"));
		}, this)
	);
};

ViewportWidget.prototype = Object.create(AbstractWidget.prototype);
ViewportWidget.prototype.constructor = ViewportWidget;
ViewportWidget.widgetName = LayoutConstants.Widget.VIEWPORT;

//------------ viewport specific  ------------------------------------------------------------------------------
/**
 * Set the WidgetModel
 * @param {ViewportModule.ViewportWidgetModel} model
 */
ViewportWidget.prototype.setWidgetModel = function(model) {
	if (this._widgetModel) {
		// unregister all listeners registered on _widgetModel
		this._unregisterAllListenersForObject(this._widgetModel);
	}

	this._widgetModel = model;

	this._registerListener("changed.panel.order", this.layout, this, this._widgetModel);
	this._registerListener("changed.panel.remove", this._destroyPanel, this, this._widgetModel);
	this._registerListener("changed.view.size", this._verifyWorld, this, this._widgetModel);
	// we don't need to listen to changed.panel.add, as this will also trigger a "order" update

	this._registerListener("changed.layout.*", this.layout, this, this._widgetModel);
	// trigger a repaint
	this.layout();
	this.repaint();
};

ViewportWidget.prototype.getContext = function() {
	return this._element;
};

/**
 * Get the size of the viewport widget
 * @returns {ShapeUtils.Size}
 */
ViewportWidget.prototype.getSize = function() {
	return new ShapeUtils.Size(this._element.innerWidth() || null, this._element.innerHeight() || null);
};

/**
 * Get the view located at a certain mouse position
 * @param {Event|ShapeUtils.Point|String}
 * @returns {ViewportModule.ViewportView|undefined} The view on the given mouse position, or undefined if not found.
 */
ViewportWidget.prototype.requestView = function(e) {
	if (_.isString(e)) {
		return this.getFirstViewOccurrence(e);
	}

	var point = null;
	if (e instanceof ShapeUtils.Point) {
		point = e.copy();
	} else {
		if (e.pageX) {
			point = new ShapeUtils.Point(e.pageX, e.pageY);
		} else {
			point = new ShapeUtils.Point(e.clientX, e.clientY);
		}
	}

	var panel = this.getPanelAt(point);
	if (panel) {
		return panel.getViewAt(point);
	} else {
		return undefined;
	}
};

//------------ world  ------------------------------------------------------------------------------
/**
 * Update the world if needed, this checks if all world and scene sizes are still correct, changes them if needed
 * @returns {Boolean} true if the world has changed
 */
ViewportWidget.prototype._verifyWorld = function() {
	if (!this._widgetModel) {
		//make sure widgetModel is set
		return false;
	}
	//defines the size of the world, this is a ShapeUtils.Rect
	var worldBounds = this._widgetModel.getWorldBounds();
	if (!worldBounds) {
		return false;
	}

	//Check if the world is still OK, checks if the scene size is correct
	var changed = false;
	var sizeForView = this._widgetModel.getSizeForView();
	if (!sizeForView.equals(this._world.getSceneSize())) {
		this._world.setSceneSize(sizeForView); //set the new correct scene size
		changed = true;
	}

	// set the resolution of the world on the world (in this way we can zoom in to the same levels as the viewser generates rasterized images)
	var worldResolution = this._widgetModel.getWorldResolution();
	if (worldResolution !== undefined) {
		this._world.setWorldResolution(worldResolution);
	}

	if (!worldBounds.rectEquals(this._world.getWorldBounds())) {
		this._world.setWorldBounds(worldBounds);
		changed = true;
	}

	var worldSizeNotNull = this._world.getWorldSize().width !== 0 && this._world.getWorldSize().height !== 0;
	var sceneSizeNotNull = this._world.getSceneSize().width !== 0 && this._world.getSceneSize().height !== 0;

	if ((!this._didZoomToFit || (changed && !this._world.didChangeTransformSinceLastZoomToFit())) && worldSizeNotNull && sceneSizeNotNull) {
		this._world.zoomToFit(false);
		if (sizeForView.width > 0 && sizeForView.height > 0 && !this.f.isParameterUpdateOngoing()) {
			this._didZoomToFit = true;
		}
	}

	if (changed && this._world.getInitialTransformation()) {
		this._world.transform(this._world.getInitialTransformation());
	}

	return changed;
};

/**
 * Get the world used in the viewport
 * @returns {ViewportModule.ViewportWorld}
 */
ViewportWidget.prototype.getWorld = function() {
	return this._world;
};

//------------ widget specific ------------------------------------------------------------------------------
/**
 * Set the context of the viewport widget
 * @override
 * @param {JQueryObj} context
 */
ViewportWidget.prototype.setContext = function(context, headerExtra) {
	AbstractWidget.prototype.setContext.call(this, context);
	headerExtra.append(this._openNewWindow.element);

	this.context.append(this._element);
};

/**
 * The widget was resized, update all panels/views --> repaint
 * @override
 */
ViewportWidget.prototype.resized = function() {
	this.layout();
};

/**
 * Get the prefered size for the View3DWidget
 * @returns {ShapeUtils.Size}
 */
ViewportWidget.prototype.getPreferredSize = function() {
	return new ShapeUtils.Size(250, 300);
};

/**
 * Get the prefered size for the View3DWidget
 * @returns {ShapeUtils.Size}
 */
ViewportWidget.prototype.getMinimumSize = function() {
	return new ShapeUtils.Size(250, 300);
};

//------------ Panels ----------------------------------------------------------------------
/**
 * Get a Panel Object with a certain name
 * If the object is not existing, create it
 * @param {String} name - name of the panel
 * @returns {ViewportModule.ViewportPanel}
 */
ViewportWidget.prototype.getPanel = function(name) {
	if (this._panelObjects[name] !== undefined) {
		return this._panelObjects[name];
	} else {
		return this._createPanel(name);
	}
};

/**
 * Create a panel for a certain Panel Model
 * @private
 * @param {String} name - name of the panel
 * @returns {ViewportModule.ViewportPanel}
 */
ViewportWidget.prototype._createPanel = function(name) {
	var panel = new ViewportPanel(this.f, this._widgetModel.getPanelModel(name), this._widgetModel, this._world);
	this._panelObjects[name] = panel;
	this._element.append(panel.element);
	return panel;
};

/**
 * Destroy a panel (remove all listeners + DOM elements)
 * @param {ViewportModule.ViewportPanel} panel - the panel to destroy
 */
ViewportWidget.prototype._destroyPanel = function(panel) {
	var name = panel.getName();
	if (this._panelObjects[name] !== undefined) {
		this._panelObjects[name].destroy();
		delete this._panelObjects[name];
		this.layout();
	}
};

/**
 * Get the panel sitting at specific point
 * @param {ShapeUtils.Point} point
 * @returns {ViewportModule.ViewportPanel} the panel at the requested point
 */
ViewportWidget.prototype.getPanelAt = function(point) {
	point = point.copy();
	var offset = this._element.offset();
	point.x -= offset.left;
	point.y -= offset.top;

	var panelsLayoutMode = this._widgetModel.getResolvedPanelsLayoutMode();
	var panelIndex = 0;
	if (panelsLayoutMode === ViewportConstants.PanelsLayoutMode.HORIZONTAL) {
		panelIndex = Math.floor(
			point.y / (Math.ceil((this._element.height() || null) / this._widgetModel.getPanelCount()) + ViewportConstants.PanelBorderWidth)
		);
	} else {
		panelIndex = Math.floor(
			point.x / (Math.ceil((this._element.width() || null) / this._widgetModel.getPanelCount()) + ViewportConstants.PanelBorderWidth)
		);
	}

	return this._panelObjects[this._widgetModel.getPanelNameAtIndex(panelIndex)];
};

/**
 * Layout the panels respecting the settings in the WidgetModel
 * @private
 */
ViewportWidget.prototype._layout = function() {
	if (this._widgetModel !== undefined && this._widgetModel.getWorldBounds()) {
		var panelOrder = this._widgetModel.getPanelOrder();
		var panelCount = panelOrder.length;
		if (panelCount > 0) {
			var panelsLayoutMode = this._widgetModel.getResolvedPanelsLayoutMode();

			if (panelsLayoutMode === ViewportConstants.PanelsLayoutMode.HORIZONTAL) {
				// vertical tiling
				this._layoutPanelsHorizontal(panelOrder, panelCount);
			} else {
				// horizontal tiling
				this._layoutPanelsVertical(panelOrder, panelCount);
			}
		}
	}
};

/**
 * Layout the panels in vertical tiled mode (left to right)
 * @private
 * @param {Array.<String>} panelOrder - order of the panels (with names in the array)
 * @param {Number} panelCount - total number of panels
 */
ViewportWidget.prototype._layoutPanelsVertical = function(panelOrder, panelCount) {
	var panelSize = new ShapeUtils.Size(
		Math.ceil(((this._element.width() || null) - (panelCount - 1) * ViewportConstants.PanelBorderWidth) / panelCount),
		this._element.height() || null
	);

	_.each(
		panelOrder,
		function(panelName, index) {
			var panel = this.getPanel(panelName);
			panel.setBounds(new ShapeUtils.Point((panelSize.width + ViewportConstants.PanelBorderWidth) * index, 0), panelSize);
			if (index < panelCount - 1) {
				panel.element.css("border-right-width", ViewportConstants.PanelBorderWidth + "px");
				panel.element.css("border-right-style", "solid");
			}
			panel.element.css("border-bottom", 0);
		},
		this
	);
};

/**
 * Layout the panels in horizontal tiled mode (top to bottom)
 * @private
 * @param {Array.<String>} panelOrder - order of the panels (with names in the array)
 * @param {Number} panelCount - total number of panels
 */
ViewportWidget.prototype._layoutPanelsHorizontal = function(panelOrder, panelCount) {
	var panelSize = new ShapeUtils.Size(
		this._element.width() || null,
		Math.ceil(((this._element.height() || null) - (panelCount - 1) * ViewportConstants.PanelBorderWidth) / panelCount)
	);

	_.each(
		panelOrder,
		function(panelName, index) {
			var panel = this.getPanel(panelName);
			var point = new ShapeUtils.Point(0, (panelSize.height + ViewportConstants.PanelBorderWidth) * index);
			panel.setBounds(point, panelSize);
			if (index < panelCount - 1) {
				panel.element.css("border-bottom-width", ViewportConstants.PanelBorderWidth + "px");
				panel.element.css("border-bottom-style", "solid");
			}
			panel.element.css("border-right", 0);
		},
		this
	);
};

/**
 * Repaint all panels/views/layers (private! use .repaint())
 * @private
 */
ViewportWidget.prototype._repaint = function() {
	if (this._widgetModel && this._widgetModel.getWorldBounds()) {
		this._verifyWorld();

		_.each(
			this._widgetModel.getPanelOrder(),
			function(panelName, index) {
				var panel = this.getPanel(panelName);
				panel.repaint();
			},
			this
		);
		this._emit("repainted");
	}
};

/**
 * Get the first occurrence of a certain view
 * @param {String} viewName - name if the view
 * @returns {ViewportView}
 */
ViewportWidget.prototype.getFirstViewOccurrence = function(viewName) {
	var panelOrder = this._widgetModel.getPanelOrder();
	for (var i = 0; i < panelOrder.length; i++) {
		var panel = this._panelObjects[panelOrder[i]];
		if (panel) {
			var view = panel.getViewIfExists(viewName);
			if (view) {
				return view;
			}
		}
	}
	return null;
};

/**
 * Get the element of the ViewportWidget
 * @returns {JQueryObj} the element
 */
ViewportWidget.prototype.getElement = function() {
	return this._element;
};

/**
 * Set the background color of the viewport widget
 * @param {String} [hexColor] - New background color in hex format without # or none to set none
 */
ViewportWidget.prototype.setBackgroundColor = function(hexColor) {
	if (hexColor === undefined) {
		hexColor = "";
	} else {
		hexColor = "#" + hexColor;
	}
	this._element.css({
		"background-color": hexColor
	});
};

/**
 * Gets the tool manager
 *
 * @returns ToolManager instance
 */
ViewportWidget.prototype.getToolManager = function() {
	return this._toolManager;
};

export default ViewportWidget;
