import _ from "underscore";
import Transformation from "@esko/transformation";
import ShapeUtils from "@esko/shapeutils";
import PrivateChannel from "@esko/web-viewer/core/PrivateChannel";
import ViewerUtils from "@esko/web-viewer/utils/ViewerUtils";
import AnimationOptions from "@esko/animation-options";
/**
 * @exports ViewportWorld
 * @memberof ViewportModule
 * @class The world keeps track of the actual size (WorldSize) of the viewport content  (in its own units, e.g. mm) and the size of the viewport within the screen (SceneSize)
 * @constructor
 * @param {Core.ViewerFacade} f - Viewer Facade
 * @param {String} IID - internal id
 * @param {Number} [iz=1] - initial zoom percentage (e.g. "0.9" will show the first view at 90%)
 */
var ViewportWorld = function(f, IID, iz) {
	PrivateChannel.installTo(this, f, ["transform", IID]);

	/**
	 * @member {Object} - see transform options
	 */
	this._initialTransformation = undefined;
	this._transform = undefined;
	this._maxZoom = undefined;
	this._minZoom = undefined;
	//is initial zoom, Minimum of SceneX/WorldX and SceneY/WorldY when the document is completely visible
	this._initialStartZoom = undefined;
	//initialZoom is the initial zoom percentage when view is created
	this._initialZoom = iz ? iz : 1;

	/* Animation Options */
	this._allowAnimation = true;
	this._animationFramerate = 45;
	this._defaultAnimationDuration = 250;
	this._currentAnimation = null;

	/* Transformation Options */
	this._rotation = 0; // in degrees
	this._offsetSize = new ShapeUtils.Size();
	this._zoom = 1;
	this._flip = false;
	this._flipViewFromBack = false;

	/* Sizes */
	this._worldSize = new ShapeUtils.Size();
	this._sceneSize = new ShapeUtils.Size();

	this._worldOrigin = new ShapeUtils.Point();

	// this is the resolution of the document (ppmm)
	this._worldResolution = {
		x: 0,
		y: 0
	};

	this._didChangeSinceLastZoomToFit = false;

	// timer used while animating
	this._animationTimer = null;
};

/**
 * Recalculate the initial start zoom of the transformation and return it immediately
 * @returns {Number} the start zoom of the transformation, based on the world/scene size
 */
ViewportWorld.prototype.recalculateInitialStartZoom = function() {
	//if the worldSize is 0 dont calculate the startzoom ==> will give Infinity
	if (this._worldSize.width === 0 || this._worldSize.height === 0) {
		return this._zoom;
	}

	//rotation needs to be fixed
	var rotate = true;
	if (this._rotation % 180 === 0) {
		rotate = false;
	}
	var fullWidthZoom = (rotate ? this._sceneSize.height : this._sceneSize.width) / this._worldSize.width;
	var fullHeightZoom = (rotate ? this._sceneSize.width : this._sceneSize.height) / this._worldSize.height;

	var z = Math.min(fullWidthZoom, fullHeightZoom);
	z *= this._initialZoom;

	var minZoom = this.getIntegerZoom(-this.getNumberOfMinZoomLevels());
	var maxZoom = this.getIntegerZoom(this.getNumberOfMaxZoomLevels());
	if (this._minZoom !== minZoom || this._maxZoom !== maxZoom) {
		this._minZoom = minZoom;
		this._maxZoom = maxZoom;
		this._emit("minMaxZoomChanged", this, this._minZoom, this._maxZoom);
	}

	this._initialStartZoom = z;
	return z;
};

/**
 * Check whether the current zoom is full view
 * @returns {Boolean} True if the current zoom is the initial start zoom (= full view)
 */
ViewportWorld.prototype.isFullView = function() {
	return this._zoom === this._initialStartZoom;
};

/**
 * Check whether the current zoom is smaller than the full view
 * @returns {Boolean} True if the current view is smaller than the full view (so zoomed out)
 */
ViewportWorld.prototype.isSmallerThanFullView = function() {
	return this._zoom < this._initialStartZoom;
};

/**
 * Get the initial start zoom
 * This is the initial zoom as minimum of SceneX/WorldX and SceneY/WorldY when the document is completely visible
 * @returns {Number} The initial start zoom
 */
ViewportWorld.prototype.getInitialStartZoom = function() {
	return this._initialStartZoom;
};

/**
 * Get the initial zoom
 * It is the initial zoom percentage when viewport is created
 * @returns {Number} the initial zoom of the viewport world
 */
ViewportWorld.prototype.getInitialZoom = function() {
	return this._initialZoom;
};

/**
 * Zoom the world to fit in the world
 * @param {Boolean} animated - True if it should be animated
 */
ViewportWorld.prototype.zoomToFit = function(animated) {
	this._offsetSize = new ShapeUtils.Size();
	this.transform({
		fixedzoom: this.recalculateInitialStartZoom(),
		animated: animated
	});

	this._didChangeSinceLastZoomToFit = false;
};

/**
 * Zoom the world to fit width in the world
 * @param {Boolean} animated - True if it should be animated
 */
ViewportWorld.prototype.zoomToFitWidth = function(animated) {
	this._reportChange();
	var rotate = true;
	if (this._rotation % 180 === 0) {
		rotate = false;
	}

	var denominator = null;
	if (rotate) {
		if (this._worldSize.height !== 0) {
			denominator = this._worldSize.height;
		}
	} else {
		if (this._worldSize.width !== 0) {
			denominator = this._worldSize.width;
		}
	}

	this._offsetSize = new ShapeUtils.Size();

	if (denominator !== null) {
		var fullWidthZoom = this._sceneSize.width / denominator;
		fullWidthZoom *= this._initialZoom;
		this.transform({
			fixedzoom: fullWidthZoom,
			animated: animated
		});
	} else {
		this.updateTransformation(animated);
	}
};

/**
 * Zoom the world to fit height in the world
 * @param {Boolean} animated - True if it should be animated
 */
ViewportWorld.prototype.zoomToFitHeight = function(animated) {
	this._reportChange();
	var rotate = true;
	if (this._rotation % 180 === 0) {
		rotate = false;
	}

	var denominator = null;
	if (rotate) {
		if (this._worldSize.width !== 0) {
			denominator = this._worldSize.width;
		}
	} else {
		if (this._worldSize.height !== 0) {
			denominator = this._worldSize.height;
		}
	}

	this._offsetSize = new ShapeUtils.Size();

	if (denominator !== null) {
		var fullHeightZoom = this._sceneSize.height / denominator;
		fullHeightZoom *= this._initialZoom;
		this.transform({
			fixedzoom: fullHeightZoom,
			animated: animated
		});
	} else {
		this.updateTransformation(animated);
	}
};

/**
 * Zoom the world to 1:1
 * @param {Boolean} animated - True if it should be animated
 * @param {Number} [worldResolution] - Pixels per mm. Define if you want to use another resolution than the default world resolution
 */
ViewportWorld.prototype.zoomOneToOne = function(animated, worldResolution) {
	this._reportChange();

	var zoom = worldResolution;
	if (zoom === undefined) {
		if (this._worldResolution && this._worldResolution.x !== 0) {
			zoom = this._worldResolution.x;
		}
	}

	this.transform({
		fixedzoom: zoom,
		animated: animated
	});
};

/**
 * Get the zoom value for the transform
 * @returns {Number} the current zoom level
 */
ViewportWorld.prototype.getZoom = function() {
	return this._zoom;
};

/**
 * Get the current zoom percentage
 * @returns {Number} The current zoom percentage
 */
ViewportWorld.prototype.getZoomPercentage = function() {
	if (this._worldResolution && this._worldResolution.x !== 0) {
		return (this._zoom / this._worldResolution.x) * 100;
	}
	return 0;
};

/**
 * Get the number of zoom levels
 * @returns {Number} Number of zoom levels
 */
ViewportWorld.prototype.getNumberOfZoomLevels = function() {
	return this.getNumberOfMinZoomLevels() + this.getNumberOfMaxZoomLevels() + 1;
};

/**
 * Get the number of zoom levels to zoom out
 * @returns {Number}
 */
ViewportWorld.prototype.getNumberOfMinZoomLevels = function() {
	// until 65536:1
	return 16;
};

/**
 * Get the number of zoom levels to zoom in
 * @returns {Number}
 */
ViewportWorld.prototype.getNumberOfMaxZoomLevels = function() {
	// until 1:512
	return 9;
};

/**
 * Get the max zoom of the world
 * @returns {Number}
 */
ViewportWorld.prototype.getMaxZoom = function() {
	return this._maxZoom;
};

/**
 * Get the min zoom of the world
 * @returns {Number}
 */
ViewportWorld.prototype.getMinZoom = function() {
	return this._minZoom;
};

/**
 * Zoom to a scene rectangle, takes a rotation of the document page into account
 * @param {ShapeUtils.Rect} rect - Rectangle to zoom to
 * @param {Boolean} animated - True if it should be animated
 */
ViewportWorld.prototype.zoomToSceneRect = function(rect, animated) {
	this._reportChange();
	var worldRect = this.getWorldRectFromScene(rect);
	var fullWidthZoom = Math.abs(this._sceneSize.width / worldRect.size.width);
	var fullHeightZoom = Math.abs(this._sceneSize.height / worldRect.size.height);
	var rotation = this.getRotation() % 360;
	if (rotation === 90 || rotation === 270) {
		fullWidthZoom = Math.abs(this._sceneSize.width / worldRect.size.height);
		fullHeightZoom = Math.abs(this._sceneSize.height / worldRect.size.width);
	}

	this.transform({
		fixedzoom: Math.min(fullWidthZoom, fullHeightZoom),
		middle: worldRect.center(),
		animated: animated
	});
};

/**
 * Set the size of the scene
 * @param {ShapeUtils.Size} s - size of the scene (screen)
 */
ViewportWorld.prototype.setSceneSize = function(s) {
	this._sceneSize = s;
	this.updateTransformation(false, true);
};

/**
 * Get the current scene size
 * @returns {ShapeUtils.Size} the size of the scene
 */
ViewportWorld.prototype.getSceneSize = function() {
	return this._sceneSize;
};

/**
 * Get the bounds of the scene
 * @returns {ShapeUtils.Rect} the bounds of the scene
 */
ViewportWorld.prototype.getSceneBounds = function() {
	return new ShapeUtils.Rect(null, this._sceneSize.copy());
};

/**
 * Get the current size of the world
 * @returns {ShapeUtils.Size} the size of the world
 */
ViewportWorld.prototype.getWorldSize = function() {
	return this._worldSize;
};

/**
 * Get the current bounds of the world
 * @returns {ShapeUtils.Rect} the bounds of the world
 */
ViewportWorld.prototype.getWorldBounds = function() {
	return new ShapeUtils.Rect(this._worldOrigin.copy(), this._worldSize.copy());
};

/**
 * Set the world bounds.
 * This updates the world size and world origin.
 * @param {ShapeUtils.Rect} bounds - New world bounds
 */
ViewportWorld.prototype.setWorldBounds = function(bounds) {
	this._worldSize = bounds.size.copy();
	this._worldOrigin = bounds.origin.copy();
	this.updateTransformation(false);
	this._emit("worldBoundsChanged");
};

/**
 * Check whether the world is busy with an animation
 * @returns {Boolean} True if the world is doing an animation
 */
ViewportWorld.prototype.isAnimating = function() {
	return this._currentAnimation != null && !this._currentAnimation.done;
};

/**
 * Check whether the world is busy with a rotation animation
 * @returns {Boolean} True if the world is doing a rotation animation
 */
ViewportWorld.prototype.isAnimatingRotation = function() {
	var t1 = this.getTransformation();
	var t2 = this.calculateCurrentTransformation();

	var rotation1 = t1.getRotation();
	var rotation2 = t2.getRotation();

	return rotation1 != rotation2;
};

/**
 * Gets the current rotation (in degrees)
 * @returns {Number} the current rotation in degrees
 */
ViewportWorld.prototype.getRotation = function() {
	return this._rotation;
};

/**
 * Get the transformation Matrix
 * @returns {Transformation} the transformation of the world
 */
ViewportWorld.prototype.getTransformation = function() {
	if (!this._transform) {
		this.updateTransformation();
	}
	return this._transform.copy();
};

/**
 * Calculate the rubber value
 * @param {Number} value - current value
 * @param {Number} rest - overflow
 * @param {Number} max - maximum rubber
 * @returns {Number} converted number based on the rubber value
 */
ViewportWorld.prototype.rubber = function(value, rest, max) {
	var stretch = (value - rest) / (max - rest);
	if (stretch <= 0) {
		return value;
	}
	stretch = stretch / (1 + stretch);
	return rest + (max - rest) * stretch;
};

/**
 * Constrain the zoom
 * @param {Boolean} rubberBand - if true, it will allow a rubber band, which will give a hint that the zoom is being constrained
 * @returns {Boolean} if the zoom was constrained
 */
ViewportWorld.prototype.constrainZoom = function(rubberBand) {
	if (this._zoom > this._maxZoom) {
		if (rubberBand) {
			this._zoom = this.rubber(this._zoom, this._maxZoom, this._maxZoom * 1.6);
		} else {
			this._zoom = this._maxZoom;
		}
		return true;
	} else if (this._zoom < this._minZoom) {
		if (rubberBand) {
			this._zoom = this.rubber(this._zoom, this._minZoom, this._minZoom / 1.6);
		} else {
			this._zoom = this._minZoom;
		}
		return true;
	} else {
		return false;
	}
};

/**
 * Constrain the rotation
 * @param {Boolean} hard - if hard is true, it will put a hard 90 degrees constrain
 */
ViewportWorld.prototype.constrainRotation = function(hard) {
	if (hard && this._rotation % 90 !== 0) {
		this._rotation = parseInt(this._rotation, 10) % 360;
		if (this._rotation < 0) {
			this._rotation += 360;
		}

		this._rotation = ViewerUtils.closestAngle(this._rotation, 4);
	}
};

/**
 * Constrain panning
 * @param {Boolean} rubberBand - if true, it will allow a rubber band, which will give a hint that the zoom is being constrained
 * @returns {Boolean} if the pan was constrained
 */
ViewportWorld.prototype.constrainPan = function(rubberBand) {
	var newMiddle = this.getMiddle();
	//TODO: implementation is not as it should be, newMiddle has rotated coordinates => not correct! see known issues
	if (this.isSmallerThanFullView()) {
		//Don't allow panning when the view is smaller than the start view
		this._offsetSize = new ShapeUtils.Size();
		return;
	} else {
		var cSize = this._worldSize.copy();
		var isRotated = false;
		//checks if the document has been rotated
		if (this._rotation % 180 === 90) {
			isRotated = true;
		}
		var izoom = this.recalculateInitialStartZoom(); //is initial zoom, Minimum of SceneX/WorldX and SceneY/WorldY when the document is completly visible
		if (this._zoom < izoom) {
			//zoom is the value for the current zoom
			izoom = this._zoom * this._initialZoom; //initialZoom is the initial zoom percentage when view is created
		}
		//original calculations to constrain pan
		//this._sceneSize.width / izoom is the width of the document+the background-border at zoom-to fit of the document
		var left = (cSize.width - this._sceneSize.width / izoom + this._sceneSize.width / this._zoom) * 0.5;
		var right = (cSize.width + this._sceneSize.width / izoom - this._sceneSize.width / this._zoom) * 0.5;
		var top = (cSize.height - this._sceneSize.height / izoom + this._sceneSize.height / this._zoom) * 0.5;
		var bottom = (cSize.height + this._sceneSize.height / izoom - this._sceneSize.height / this._zoom) * 0.5;
		//
		if (isRotated) {
			//when rotated the bounds are different
			left = (cSize.height - this._sceneSize.width / izoom - this._sceneSize.width / this._zoom) * 0.5;
			right = (cSize.height + this._sceneSize.width / izoom + this._sceneSize.width / this._zoom) * 0.5;
			top = (cSize.width - this._sceneSize.height / izoom - this._sceneSize.height / this._zoom) * 0.5;
			bottom = (cSize.width + this._sceneSize.height / izoom + this._sceneSize.height / this._zoom) * 0.5;
		}
		//the coordinates of newMiddle are rotated!! convert to normal coordinates with origin in top left corner
		var tempMiddle = newMiddle.copy();
		//first make sure we don't include the origin of the world in our calculations (this would make a lot harder!)
		newMiddle.offset(new ShapeUtils.Size(-this._worldOrigin.x, -this._worldOrigin.y));
		tempMiddle.offset(new ShapeUtils.Size(-this._worldOrigin.x, -this._worldOrigin.y));

		if (this._rotation % 360 === 90) {
			tempMiddle = new ShapeUtils.Point(cSize.height - newMiddle.y, newMiddle.x);
		} else if (this._rotation % 360 === 180) {
			tempMiddle = new ShapeUtils.Point(cSize.width - newMiddle.x, cSize.height - newMiddle.y);
		} else if (this._rotation % 360 === 270) {
			tempMiddle = new ShapeUtils.Point(newMiddle.y, cSize.width - newMiddle.x);
		}

		//checks if the middle point lays in the bounds
		if (tempMiddle.x < left) {
			tempMiddle.x = left;
		}
		if (tempMiddle.x > right) {
			tempMiddle.x = right;
		}
		if (tempMiddle.y < top) {
			tempMiddle.y = top;
		}
		if (tempMiddle.y > bottom) {
			tempMiddle.y = bottom;
		}

		// also check if the document is still within the view bounds
		if (isRotated) {
			left = 0;
			right = cSize.height;
			top = 0;
			bottom = cSize.width;
		} else {
			left = 0;
			right = cSize.width;
			top = 0;
			bottom = cSize.height;
		}

		if (tempMiddle.x < left) {
			tempMiddle.x = left;
		}
		if (tempMiddle.x > right) {
			tempMiddle.x = right;
		}
		if (tempMiddle.y < top) {
			tempMiddle.y = top;
		}
		if (tempMiddle.y > bottom) {
			tempMiddle.y = bottom;
		}
		//convert the coordinates back to set middle if needed
		newMiddle = tempMiddle;
		if (this._rotation % 360 === 90) {
			newMiddle = new ShapeUtils.Point(tempMiddle.y, cSize.height - tempMiddle.x);
		} else if (this._rotation % 360 === 180) {
			newMiddle = new ShapeUtils.Point(cSize.width - tempMiddle.x, cSize.height - tempMiddle.y);
		} else if (this._rotation % 360 === 270) {
			newMiddle = new ShapeUtils.Point(cSize.width - tempMiddle.y, tempMiddle.x);
		}

		newMiddle.offset(new ShapeUtils.Size(this._worldOrigin.x, this._worldOrigin.y));
	}

	var middle;
	if (!rubberBand) {
		middle = this.getMiddle();
		if (!middle.equals(newMiddle)) {
			this.setMiddle(newMiddle);
			return true;
		} else {
			return false;
		}
	} else {
		middle = this.getMiddle();
		var delta = new ShapeUtils.Point();
		delta.x = Math.round(middle.x - newMiddle.x);
		delta.y = Math.round(middle.y - newMiddle.y);
		middle = newMiddle;
		var stretch = Math.sqrt(delta.x * delta.x + delta.y * delta.y);

		if (stretch < 0.01) {
			return false;
		}

		var newStretch = this.rubber(stretch, 0, 200 / this._zoom);
		delta.x *= newStretch / stretch;
		delta.y *= newStretch / stretch;

		middle.x += delta.x;
		middle.y += delta.y;
		this.setMiddle(middle);
		return true;
	}
};

/**
 * Sets the middle of the world
 * @param {ShapeUtils.Point} middle - the new middle
 */
ViewportWorld.prototype.setMiddle = function(middle) {
	this._reportChange();
	var center = this.getWorldBounds().center();

	this._offsetSize = new ShapeUtils.Size(center.x - middle.x, center.y - middle.y);
};

/**
 * Get the middle of the world
 * @returns {ShapeUtils.Point} - the middle of the world
 */
ViewportWorld.prototype.getMiddle = function() {
	var center = this.getWorldBounds().center();
	return new ShapeUtils.Point(center.x - this._offsetSize.width, center.y - this._offsetSize.height);
};

/**
 * Update the transformation matrix
 * @param {Boolean | Object} animation - either "false" or "true"
 * @param hardrotation - hard rotation constrain
 * @param {Boolean} [tmpDisableConstraints] - if true constraints are disabled temporarily
 */
ViewportWorld.prototype.updateTransformation = function(animation, hardrotation, tmpDisableConstraints) {
	if (!tmpDisableConstraints) {
		this.constrainZoom(false);
		this.constrainPan(false); //rubber doesn't work correctly with current constrainPan

		this.constrainRotation(hardrotation);
	}

	var newtransform = this.calculateCurrentTransformation();
	if (this._transform === undefined) {
		this._transform = newtransform;
	}

	if (newtransform.equals(this._transform)) {
		//transforms are the same. Do not animate
		animation = false;
	}

	if (!this.isAnimationAllowed()) {
		animation = false;
	}

	if (animation) {
		if (_.isBoolean(animation)) {
			animation = {};
		}
		if (!animation.duration) {
			animation.duration = this._defaultAnimationDuration;
		}
		//animation.duration=5000; // uncomment this line for slower animations (makes it easier to debug) (slow animationspeed)
		var gain = 100 / (animation.duration / this._animationFramerate) / 100;

		if (!animation.type) {
			animation.type = "ease-in-out";
		}
		var animationOption = new AnimationOptions(animation.type);

		animationOption.from =
			animation.interrupt && this._currentAnimation && !this._currentAnimation.done ? this._currentAnimation.to : this._transform;
		animationOption.to = newtransform;
		animationOption.gain = gain;
		animationOption.progress = 0;
		animationOption.done = false;
		this.updateTransformationAnimated(animationOption);

		if (this._currentAnimation) {
			this._currentAnimation.done = true;
		}
		this._currentAnimation = animationOption;
	} else {
		if (this.isAnimating()) {
			this.cancelAnimation(true);
		}
		this._transform = newtransform;
		this.fireTransformChanged();
	}
};

/**
 * Create the current transformation matrix
 * @returns {Transformation} the current transformation matrix
 */
ViewportWorld.prototype.calculateCurrentTransformation = function() {
	var worldSizeWidth = this._zoom !== 0 ? this._sceneSize.width / this._zoom : 0;
	var worldSizeHeight = this._zoom !== 0 ? this._sceneSize.height / this._zoom : 0;
	var worldSceneBounds = new ShapeUtils.Rect(null, new ShapeUtils.Size(worldSizeWidth, worldSizeHeight));
	var offsetX = (worldSceneBounds.size.width - this._worldSize.width) / 2;
	var offsetY = (worldSceneBounds.size.height - this._worldSize.height) / 2;

	var t = new Transformation();

	t.scale(this._zoom, this._zoom);

	t.translate(worldSceneBounds.size.width / 2, worldSceneBounds.size.height / 2);
	t.rotate(this._rotation * (Math.PI / 180));
	if (this.isFlip()) {
		t.scale(-1, 1);
	}
	if (this.isFlipViewFromBack()) {
		if (this._rotation % 180 === 0) {
			// not rotated
			t.scale(-1, 1);
		} else {
			// rotated, but flip horizontally
			t.scale(1, -1);
		}
	}
	t.translate(-worldSceneBounds.size.width / 2, -worldSceneBounds.size.height / 2);

	t.translate(offsetX, offsetY);
	if (this._worldOrigin) {
		t.translate(-this._worldOrigin.x, -this._worldOrigin.y);
	}
	t.translate(this._offsetSize.width, this._offsetSize.height);

	// round to integer values so we only draw on pixels
	t.e = Math.floor(t.e);
	t.f = Math.floor(t.f);

	return t;
};

/**
 * Update the transformation
 * @param {AnimationOption} animationOptions
 */
ViewportWorld.prototype.updateTransformationAnimated = _.throttle(function(animationOption) {
	if (animationOption.done) {
		return; //not longer the latest animation
	}
	if (animationOption.progress > 0.9999 || animationOption.progress >= 1) {
		animationOption.progress = 1;
	}
	if (animationOption.from.equals(animationOption.to)) {
		animationOption.progress = 1;
	}

	var p = animationOption.get(animationOption.progress);
	var newtr = new Transformation();

	newtr.a = animationOption.from.a * (1 - p) + animationOption.to.a * p;
	newtr.b = animationOption.from.b * (1 - p) + animationOption.to.b * p;
	newtr.c = animationOption.from.c * (1 - p) + animationOption.to.c * p;
	newtr.d = animationOption.from.d * (1 - p) + animationOption.to.d * p;
	newtr.e = animationOption.from.e * (1 - p) + animationOption.to.e * p;
	newtr.f = animationOption.from.f * (1 - p) + animationOption.to.f * p;

	this._transform = newtr;

	if (animationOption.progress >= 1) {
		animationOption.done = true;
		if (this._currentAnimation == animationOption) {
			this._currentAnimation = null;
		}
		this.fireTransformChanged();
	} else {
		this.fireTransformChanged();

		this._animationTimer = setTimeout(
			_.bind(function() {
				animationOption.progress += animationOption.gain;
				this.updateTransformationAnimated(animationOption);
			}, this),
			this._animationFramerate
		);
	}
}, 0);

/**
 * Set the transform
 * @param {Object} options
 * 	- rotate: add a rotation to the current rotation
 *  - offsetX: add offset to the X axis
 *  - offsetY: add offset to the Y axis
 *  - flip: flip the world horizontally
 *  - flipViewFromBack: toggle flip for view from back (without interfering with the normal flip setting)
 *  - zoom: add a zoom to the current zoom in steps (1 is one zoomlevel, 2 is two zoomlevels, -1 is zoom out a zoomlevel)
 *  - fixedzoom: set the zoom to a fixed zoom
 *  - fixedoffset: set the offset to a fixed size
 *  - fixedrotation: set the rotation to a fixed number
 *  - middle: set the middle to a fixed middle
 *  - animated: identify if the transformation needs to be animated
 *  - rubber: constrain the zoom/pan with a rubber (this will give an indication if the zoom/pan goes beyond the bounds)
 *  - constrainRotation: constrains the rotation to steps of 90 degrees (so 0,90,180,270)
 * @param {boolean} skipUpdate - means we're in a temporary state in which we don't want to do an update
 */
ViewportWorld.prototype.transform = function(options, skipUpdate) {
	this._reportChange();

	if (options.rotate && !this._disableRotation) {
		this._rotation = (this._rotation + options.rotate + 360) % 360;
		this._emit("rotated");
	}
	if (options.fixedrotation) {
		this._rotation = options.fixedrotation % 360;
		this._emit("rotated");
	}
	if (options.flip && !this._disableFlip) {
		this._flip = !this._flip;
		this._emit("flipped");
	}

	if ((options.offsetX || options.offsetY) && !this._disablePan) {
		var ox = options.offsetX ? options.offsetX : 0;
		var oy = options.offsetY ? options.offsetY : 0;
		this._offsetSize.width += ox;
		this._offsetSize.height += oy;
	}

	if (options.flipViewFromBack !== undefined && !this._disableFlip) {
		this._flipViewFromBack = options.flipViewFromBack;
		this._emit("flippedViewFromBack");
	}

	if (options.zoom && !this._disableZoom) {
		if (this._worldResolution && this._worldResolution.x !== 0) {
			// we want to zoom to the levels for which a rasterized image exists
			// the zoomlevels should be in this row:
			// x = worldResolution * n
			// or
			// x = wordResolution / (2^n)
			//
			// zoomInZoomOut = -1 for zoom out, +1 or +2 for zoom in
			// worldResolution * 2 ^ (round(log2(zoom/worldResolution)) + zoomInZoomOut)
			var integerZoomLevel = Math.round(Math.log(this._zoom / this._worldResolution.x) / Math.log(2));
			this._zoom = this.getIntegerZoom(integerZoomLevel + options.zoom);
		} else {
			// just divide by two or multiply by 2
			this._zoom *= Math.pow(2, options.zoom);
		}
	}

	if (options.fixedzoom) {
		this._zoom = options.fixedzoom;
	}

	//if the viewport shouldn't animate anymore, cancel the current animation
	if (!this.isAnimationAllowed()) {
		this.cancelAnimation();
	}

	if (options.fixedoffset) {
		this._offsetSize = options.fixedoffset;
	}

	if (options.middle && !this._disableMiddle) {
		this.setMiddle(options.middle);
	}

	if (!skipUpdate) {
		this.updateTransformation(options.animated, options.constrainRotation, options.tmpDisableConstraints);
	}
};

/**
 * Get the zoom for the given integer zoom level
 * @return {Number}
 */
ViewportWorld.prototype.getIntegerZoom = function(integerZoomLevel) {
	return this._worldResolution.x * Math.pow(2, integerZoomLevel);
};

/**
 * Set the initial transformation for the viewport world (used for setting initial state)
 * @param {Object} transform
 */
ViewportWorld.prototype.setInitialTransformation = function(transform) {
	this._reportChange();
	this._initialTransformation = transform;
};

/**
 * Get the initial transformation of the viewport world
 * @returns {Object}
 */
ViewportWorld.prototype.getInitialTransformation = function() {
	return this._initialTransformation;
};

/**
 * Fit in a rectangle
 * @param {ShapeUtils.Rect} rect - the rectangle to fit into
 * @param {Boolean} animated - if the transition needs to be animated
 * @param {Boolean} constrain - true to constrain the zoom
 */
ViewportWorld.prototype.fitInRect = function(rect, animated, constrain) {
	this._reportChange();
	var fullWidthZoom = Math.abs(this._sceneSize.width / rect.size.width);
	var fullHeightZoom = Math.abs(this._sceneSize.height / rect.size.height);
	var zoom = Math.min(fullWidthZoom, fullHeightZoom);

	if (constrain) {
		zoom *= 0.8;
		zoom = Math.min(zoom, this._maxZoom);
	}

	this.transform({
		fixedzoom: zoom,
		middle: rect.center(),
		animated: animated
	});
};

/**
 * Gets the offset size of the viewport world
 * @returns {ShapeUtils.Size}
 */
ViewportWorld.prototype.getOffsetSize = function() {
	return this._offsetSize.copy();
};

/**
 * Set the offset size of the world
 * @param {ShapeUtils.Size} os - offset size
 * @param {Boolean} animated - true for animation
 */
ViewportWorld.prototype.setOffsetSize = function(os, animated) {
	this._reportChange();
	this._offsetSize = os;
	this.updateTransformation(animated);
};

/**
 * Get the viewport displayed with the current transform
 * @param {Boolean} donotconstrain - True for not constraining
 * @returns {ShapeUtils.Rect} the viewport rect
 */
ViewportWorld.prototype.getTargetViewPort = function(donotconstrain) {
	var rect = this.getWorldRectFromScene(this.getSceneBounds(), this.calculateCurrentTransformation());
	if (!donotconstrain) {
		rect = rect.calculateOverlap(this.getWorldBounds());
	}

	return rect;
};

/**
 * Get the rect of the viewport for when the animation is finished.
 *
 * if no animation is running, just return the targetviewport
 * @returns {ShapeUtils.Rect} the viewport rect for after the animation
 */
ViewportWorld.prototype.getAnimationDestinationTargetViewPort = function() {
	if (this.isAnimating()) {
		var rect = this.getWorldRectFromScene(this.getSceneBounds(), this._currentAnimation.to);
		rect = rect.calculateOverlap(this.getWorldBounds());
		return rect;
	} else {
		return this.getTargetViewPort();
	}
};
/**
 * Converts the size of a scene to the world
 * @param {ShapeUtils.Size} sceneSize - the scene size to convert to world
 * @param {Transformation} [t] - optionally the transformation
 * @returns {ShapeUtils.Size} the converted size in world coordinates
 */
ViewportWorld.prototype.getWorldSizeFromScene = function(sceneSize, t) {
	if (!t) {
		t = this.getTransformation();
	}
	var p1 = this.getWorldPointFromScene(new ShapeUtils.Point(), t);
	var p2 = this.getWorldPointFromScene(new ShapeUtils.Point(sceneSize.width, sceneSize.height), t);
	var worldSize = new ShapeUtils.Size(p2.x - p1.x, p2.y - p1.y);
	return worldSize;
};

/**
 * Converts the point of a scene to the world
 * @param {ShapeUtils.Point} scenePoint - the scene point to convert
 * @param {Transformation} [t] - optionally the transformation
 * @returns {ShapeUtils.Point} the converted point in world coordinates
 */
ViewportWorld.prototype.getWorldPointFromScene = function(scenePoint, t) {
	if (!t) {
		t = this.getTransformation();
	}
	return t.createInverse().transformedPoint(scenePoint);
};

/**
 * Converts the rect of a scene to the world
 * @param {ShapeUtils.Rect} sceneRect - the scene rect to convert to world
 * @param {Transformation} t - optionally the transformation
 * @returns {ShapeUtils.Rect} the converted rect in world coordinates
 */
ViewportWorld.prototype.getWorldRectFromScene = function(sceneRect, t) {
	if (!t) {
		t = this.getTransformation();
	}
	var worldRect = t.createInverse().transformedRect(sceneRect);
	var zoom = this.getZoom();
	//Due to floating point calculations in JavaScript, round origin points of the visible rectangle (caued alignment issues in Grid)
	if (1 - worldRect.origin.y * zoom < 0.000001) {
		//if there is a small epsilon difference (~.99999999998) add an larger epsilon value to make sure it rounds correctly after calculations
		worldRect.origin.y += 0.000001 / zoom; // add 0.000001 pixel to the origin
	}
	if (1 - worldRect.origin.x * zoom < 0.000001) {
		worldRect.origin.x += 0.000001 / zoom;
	}
	return worldRect;
};

/**
 * Converts the rect of a world to scene coordinates
 * @param {ShapeUtils.Rect} worldRect - the world rect to convert to scene
 * @param {Transformation} t - optionally the transformation
 * @returns {ShapeUtils.Rect} the converted rect in scene coordinates
 */
ViewportWorld.prototype.getSceneRectFromWorld = function(worldRect, t) {
	if (!t) {
		t = this.getTransformation();
	}
	return t.transformedRect(worldRect);
};

/**
 * Converts the point of the world to the scene
 * @param {ShapeUtils.Point} worldPoint - the world point to convert
 * @param {Transformation} t - optionally the transformation
 * @returns {ShapeUtils.Point} the converted point in scene coordinates
 */
ViewportWorld.prototype.getScenePointFromWorld = function(worldPoint, t) {
	if (!t) {
		t = this.getTransformation();
	}
	return t.transformedPoint(worldPoint);
};

/**
 * Converts the size of the world to the scene
 * @param {ShapeUtils.Size} worldPoint - the world size to convert
 * @param {Transformation} t - optionally the transformation
 * @returns {ShapeUtils.Size} the converted size in scene coordinates
 */
ViewportWorld.prototype.getSceneSizeFromWorld = function(worldSize, t) {
	if (!t) {
		t = this.getTransformation();
	}
	return t.transformedSize(worldSize);
};

/**
 * Alert the listeners that the world changed
 */
ViewportWorld.prototype.fireTransformChanged = function() {
	this._emit("transformChanged", this);
};

/**
 * Can animations be used?
 * @param {Boolean} toggle - True if animations can be used, false otherwise
 */
ViewportWorld.prototype.toggleAnimation = function(toggle) {
	this._allowAnimation = toggle;
	if (!toggle) {
		this.cancelAnimation();
	}
};

/**
 * Is flipped horizontally?
 * @returns {Boolean} True if flipped horizontally
 */
ViewportWorld.prototype.isFlip = function() {
	return this._flip;
};

/**
 * Check if flipviewfromback is set
 * @returns {Boolean}
 */
ViewportWorld.prototype.isFlipViewFromBack = function() {
	return this._flipViewFromBack;
};

/**
 * Set the world resolution (the resolution of the original document)
 * @param {Object} newWorldResolution - the resolution of the document
 * {x: {Number} x-resolution,
 *  y: {Number} y-resolution }
 */
ViewportWorld.prototype.setWorldResolution = function(newWorldResolution) {
	this._worldResolution = newWorldResolution;
};

/**
 * Get the resolution of the document
 * @returns {Object}
 * {x: {Number} x-resolution,
 *  y: {Number} y-resolution }
 */
ViewportWorld.prototype.getWorldResolution = function() {
	return this._worldResolution;
};

/**
 * Is the zoom level higher than 1-16
 * @returns {Boolean} True if higher than 1-16 zoom
 */
ViewportWorld.prototype.isDeepZoom = function() {
	return this._zoom / this._worldResolution.x >= 16 || this._zoom / this._worldResolution.y >= 16;
};

/**
 * Internal method for marking the world as changed since the last zoom to fit
 */
ViewportWorld.prototype._reportChange = function() {
	this._didChangeSinceLastZoomToFit = true;
};
/**
 * Is the transform of the world changed since the last zoom to fit operation?
 * @returns {Boolean}
 */
ViewportWorld.prototype.didChangeTransformSinceLastZoomToFit = function() {
	return this._didChangeSinceLastZoomToFit;
};

/**
 * Cancel the currently ongoing animation
 * @param {Boolean} [cancelResult] - if true, the destination transformation of the animation will be ignored, by default, the destation transformation of the animation is set as transform of the viewport
 */
ViewportWorld.prototype.cancelAnimation = function(cancelResult) {
	if (this._animationTimer && this._currentAnimation) {
		clearTimeout(this._animationTimer);

		if (!this._currentAnimation.done && !cancelResult) {
			this._transform = this._currentAnimation.to;
			this.fireTransformChanged();
		}

		this._currentAnimation.done = true;
		this._currentAnimation = null;
	}
};

/**
 * Is animation of the viewport allowed?
 * @returns {Boolean}
 */
ViewportWorld.prototype.isAnimationAllowed = function() {
	return this._allowAnimation && this.getZoomPercentage() < 1000;
};

/**
 * Reset the current orientation options of the world (flipped, rotated and view from back)
 */
ViewportWorld.prototype.resetOrientation = function() {
	this._flip = false;
	this._emit("flipped");
	this._rotation = 0;
	this._emit("rotated");
	this._flipViewFromBack = false;
	this._emit("flippedViewFromBack");
};

/**
 * Allow rotation transformatons
 *
 * @param {Boolean} enabled True if allowed to perform this transformation, false otherwise
 */
ViewportWorld.prototype.allowRotation = function(enabled) {
	this._disableRotation = !enabled;
};

/**
 * Allow pan transformatons
 *
 * @param {Boolean} enabled True if allowed to perform this transformation, false otherwise
 */
ViewportWorld.prototype.allowPan = function(enabled) {
	this._disablePan = !enabled;
};

/**
 * Allow zoom transformatons
 *
 * @param {Boolean} enabled True if allowed to perform this transformation, false otherwise
 */
ViewportWorld.prototype.allowZoom = function(enabled) {
	this._disableZoom = !enabled;
};

/**
 * Allow flip transformatons
 *
 * @param {Boolean} enabled True if allowed to perform this transformation, false otherwise
 */
ViewportWorld.prototype.allowFlip = function(enabled) {
	this._disableFlip = !enabled;
};

/**
 * Allow middle transformatons
 *
 * @param {Boolean} enabled True if allowed to perform this transformation, false otherwise
 */
ViewportWorld.prototype.allowMiddle = function(enabled) {
	this._disableMiddle = !enabled;
};

export default ViewportWorld;
