import _ from "underscore";
import ShapeUtils from "@esko/shapeutils";
import Fragment from "@esko/web-viewer/utils/Fragment";
var MAX_FRAGMENTS_COUNT = 5;
var DEBUG = false;
var COMPARE_THRESHOLD = 0.00001;

function thresholdEquals(value1, value2) {
	return Math.abs(value1 - value2) <= COMPARE_THRESHOLD;
}

/**
 * @exports ViewportFragmentManager
 * @memberof ViewportModule
 * @class ViewportFragmentManager is repsonsible for managing all frames shown in a certain view. When a frame is request, the fragmentCache is used
 * @param {Core.ViewerFacade} f - the ViewerFacade
 * @param {FragmentManagerDelegate} delegate - the delegate needed for the fragmentmanager
 * @example
 *
 * The FragmentManagerDelegate type has following methods:
 * - {String} getURLForFragment({Utils.Fragment}) - generate a URL for the a specific fragment
 * - {void} fragmentsChanged() - the fragment definition has changed, re render
 *
 */
var ViewportFragmentManager = function(f, delegate, world) {
	this.f = f;
	this.delegate = delegate;
	this.world = world;

	// a list of available fragments, ordered by their zoom level
	this.fragments = [];
	this.initialFragment = null;
	this.dirtyInitialFragment = null;
	this.fragmentCache = this.f.getModule("viewport").getImageCache();
};

/**
 * Get a list of all fragments
 * @returns {Array<Utils.Fragment>}
 */
ViewportFragmentManager.prototype.getAllFragments = function() {
	var array = this.fragments.slice(0);
	array.prepend(this.intialFragment);
	return array;
};

/**
 * Get the initial fragment
 *
 * @returns {Utils.Fragment|null} the initial fragment (fragment with initial zoom) or null if no fragment is available
 */
ViewportFragmentManager.prototype.getInitialFragment = function() {
	if (!this.initialFragment || !this.initialFragment.isFetched()) {
		return this.dirtyInitialFragment;
	}
	return this.initialFragment;
};

/**
 * Check if a rectangle is the same or covered/contained by referenceRect
 * @param {ShapeUtils.Rect} rect
 * @param {ShapeUtils.Rect} referenceRect
 * @returns {Boolean}
 */
ViewportFragmentManager.prototype.isRectCoveredBy = function(rect, referenceRect) {
	return rect.rectEquals(referenceRect) || referenceRect.containsRect(rect);
};

/**
 * Check if a fragment has the same resolution and is covered/contained by another fragment
 * @param {Utils.Fragment} fragment
 * @param {Utils.Fragment} reference
 * @returns {Boolean}
 */
ViewportFragmentManager.prototype.isFragmentCoveredBy = function(fragment, reference) {
	return thresholdEquals(fragment.getZoom(), reference.getZoom()) && this.isRectCoveredBy(fragment.getRect(), reference.getRect());
};

/**
 * Request to include a fragment. The function will see if the fetch is worthwhile (if no other fragment is already there for the current frame + resolution)
 *
 * @param {ShapeUtils.Rect} viewrect - Rect to fetch from the viewserver
 * @param {float} zoom - zoom/resolution of the fragment to fetch
 */
ViewportFragmentManager.prototype.preloadFragment = function(fragmentRect, zoomLevel) {
	if (fragmentRect.size.width === 0 || fragmentRect.size.height === 0) {
		// don't preload empty fragments
		return;
	}

	var newFragment = new Fragment(fragmentRect, zoomLevel);

	var initialFragment = this.getInitialFragment();
	if (initialFragment && !initialFragment.isDirty() && this.isFragmentCoveredBy(newFragment, initialFragment)) {
		// we are requesting something similar to the original fragment... ignore :)
		return;
	}

	var i;
	// first check if we really need to fetch this fragment
	for (i = 0; i < this.fragments.length; i++) {
		var reference = this.fragments[i];
		if (!reference.isDirty() && this.isFragmentCoveredBy(newFragment, reference)) {
			if (process.env.NODE_ENV === "development") {
				if (DEBUG) {
					this.f.debug.debug("Fragment available, shouldn't fetch new data");
				}
			}
			return;
		}
	}

	// add the fragment at the correct location
	i = 0;
	while (i < this.fragments.length && newFragment.getZoom() >= this.fragments[i].getZoom()) {
		i++;
	}

	this.fragments.splice(i, 0, newFragment);

	if (process.env.NODE_ENV === "development") {
		if (DEBUG) {
			this.f.debug.debug("Fetch for view port: " + fragmentRect.toString() + " and zoom: " + zoomLevel + " insert at index: " + i);
		}
	}

	// a new fragment starts at age 0
	newFragment.age = 0;
	newFragment.fetchImage(
		this.fragmentCache,
		this.getUrlForFragment(newFragment),
		_.bind(this._fragmentFetched, this),
		_.bind(this._removeFragment, this, newFragment)
	);
};

/**
 * Get applicable fragments, these are the fragments that should be drawn in the viewport
 * note, this function will also try to remove unnecessary fragments
 * @param {ShapeUtils.Rect} viewRectangle
 * @param {Number} resolution
 * @returns {Array.<Utils.Fragment>}
 */
ViewportFragmentManager.prototype.getApplicableFragments = function(viewRectangle, resolution) {
	var finalFragments = [];
	var i;
	var perfectFit = null;
	var initialFragment = this.getInitialFragment();

	if (initialFragment && initialFragment.isFetched()) {
		if (
			!initialFragment.isDirty() &&
			viewRectangle.rectEquals(initialFragment.getRect()) &&
			thresholdEquals(initialFragment.getZoom(), resolution)
		) {
			initialFragment.perfectFit = true;
			finalFragments.push(initialFragment);
			return finalFragments;
		} else {
			initialFragment.perfectFit = false;
		}
	} else {
		initialFragment = null;
	}

	// cleanFragments contains the applicable fragments that are clean
	var cleanFragments = [];
	// dirtyFragments contains the applicable fragments that are dirty
	var dirtyFragments = [];
	var totalFetchedFrames = 0;
	var totalDirtyFrames = 0;
	// inline helper function to add the fragment to the correct array (dirty or non-dirty)
	var addFragment = function(fragment) {
		if (fragment.isDirty()) {
			dirtyFragments.push(fragment);
		} else {
			cleanFragments.push(fragment);
		}
		// once a fragment is applicable in some way, it gets 1 (year) younger
		fragment.age--;
	};

	// calculate the overlap of the document with the view rectangle, this is needed to detect a perfect fit
	var overlap = viewRectangle.calculateOverlap(this.world.getWorldBounds());

	// run through all frames
	for (i = 0; i < this.fragments.length; i++) {
		var fragment = this.fragments[i];
		// everytime we draw, a frame gets 3 (years) older
		fragment.age += 3;
		if (!perfectFit) {
			if (fragment.isFetched()) {
				totalFetchedFrames++;

				if (fragment.isDirty()) {
					totalDirtyFrames++;
				}

				if (fragment.getZoom() <= resolution) {
					// the fragment is of lower (or equal) resolution than needed
					if (ShapeUtils.rectIntersection(fragment.getRect(), viewRectangle)) {
						// there is some kind of overlap, this fragment will be drawn
						if (!initialFragment || !(fragment.getZoom() < initialFragment.getZoom() && initialFragment.getZoom() > resolution)) {
							// if there is an initial fragment, we only draw fragments that have a higher resolution than the initialfragment, unless the required drawing resolution is lower than the the initial fragments resolution
							if (thresholdEquals(fragment.getZoom(), resolution) && this.isRectCoveredBy(overlap, fragment.getRect())) {
								// we have a perfect fit (the fragment covers the view rectangleoverlap and is in the same resolution
								fragment.perfectFit = true;
								if (!fragment.isDirty()) {
									perfectFit = fragment;

									if (process.env.NODE_ENV === "development") {
										if (DEBUG) {
											this.f.debug.debug("Perfect (nondirty) was found!");
										}
									}
								}
							} else {
								fragment.perfectFit = false;
							}

							addFragment(fragment);
						}
					}
				} else if (this.world.isAnimating() && fragment.getZoom() * 3 <= resolution) {
					// during animations we can also draw fragments that are sharper than needed
					if (ShapeUtils.rectIntersection(fragment.getRect(), viewRectangle)) {
						addFragment(fragment);
					}
				}
			}
		}
	}

	if (perfectFit) {
		// we found a perfect fit, so only draw the initial fragment (if any) and the perfect fit
		if (initialFragment) {
			finalFragments.push(initialFragment);
		}
		// the perfect fit frame gets 3 year younger
		perfectFit.age = -3;
		finalFragments.push(perfectFit);
	} else {
		// build up the list of fragments to draw, first draw dirty fragments till we have clean fagments
		var dirtyI = 0,
			cleanI = 0;
		var currentDirty = dirtyFragments[dirtyI];
		var currentClean = cleanFragments[cleanI];
		while (currentDirty && (!currentClean || currentDirty.getZoom() < currentClean.getZoom())) {
			//if a fragment is drawn it gets one year younger
			currentDirty.age -= 1;
			finalFragments.push(currentDirty);
			currentDirty = dirtyFragments[++dirtyI];
		}
		// draw all the clean fragments, but don't draw fragments for which the viewable area is completily overdrawn by one the following frame
		while (cleanI < cleanFragments.length) {
			var nextFragment = cleanFragments[cleanI + 1];
			if (
				!nextFragment ||
				!this.isRectCoveredBy(viewRectangle.calculateOverlap(cleanFragments[cleanI].getRect()), nextFragment.getRect())
			) {
				// if a fragment gets drawn it gets 1 year younger
				cleanFragments[cleanI].age -= 1;
				finalFragments.push(cleanFragments[cleanI]);
			}
			cleanI++;
		}
	}

	// if there are no fragments and an initial fragment or we are zoomed in more than the initial fragment, add the initial fragment
	if (initialFragment && (finalFragments.length == 0 || initialFragment.getZoom() <= resolution)) {
		// find the correct location for the initial fragment
		i = 0;
		while (i < finalFragments.length && finalFragments[i].getZoom() < initialFragment.getZoom()) {
			i++;
		}

		finalFragments.splice(i, 0, initialFragment);
	}

	// now we need to do some clean up
	var toRemove = totalFetchedFrames - totalDirtyFrames - MAX_FRAGMENTS_COUNT;
	var before = this.fragments.length;

	if (toRemove > 0) {
		// we store to many clean/fetched fragments, clean up!

		// order the fragments by age (and moment added (more recent fragments first))
		var orderedByAge = _.sortBy(this.fragments, function(fragment, index, list) {
			return 100 * fragment.age + (list.length - index);
		});

		if (process.env.NODE_ENV === "development") {
			if (DEBUG) {
				this.f.debug.debug(
					"Removing fragment, available ages:",
					_.map(this.fragments, function(fragment, index, list) {
						return 100 * fragment.age + (list.length - index);
					})
				);
				this.f.debug.debug(
					"age ordering",
					_.map(orderedByAge, function(fragment, index, list) {
						return 100 * fragment.age + index;
					})
				);
			}
		}

		// now lets remove, try the oldest fragments first
		for (i = orderedByAge.length - 1; i >= 0 && toRemove > 0; i--) {
			let fragment = orderedByAge[i];
			if (fragment !== initialFragment && fragment.isFetched() && !fragment.isDirty() && !_.contains(finalFragments, fragment)) {
				if (process.env.NODE_ENV === "development") {
					if (DEBUG) {
						this.f.debug.debug("REMOVE fragment with age", fragment.age);
					}
				}
				this._removeFragment(fragment);
				toRemove--;
			}
		}
	}

	if (process.env.NODE_ENV === "development") {
		if (DEBUG) {
			if (before > this.fragments.length) {
				this.f.debug.debug("CLEANED fragments: #", before - this.fragments.length);
			}

			this.f.debug.debug("drawing fragments: #", finalFragments.length);
		}
	}

	return finalFragments;
};

/**
 * Get the URL for a fragment
 * @param {Utils.Fragment} fragment - the fragment for which you want the url
 * @returns {String}
 */
ViewportFragmentManager.prototype.getUrlForFragment = function(fragment) {
	var url = this.delegate.getURLForFragment(fragment);

	if (process.env.NODE_ENV === "development") {
		if (DEBUG) {
			this.f.debug.debug("Fetching url: " + url);
		}
	}

	return url;
};

/**
 * A fragment was fetched, clean up dirty frames and inform the delegate
 */
ViewportFragmentManager.prototype._fragmentFetched = function() {
	this.cleanupDirtyFragments();
	this.delegate.fragmentsChanged();
};

/**
 * Remove a fragment
 * @private
 * @param {Utils.Fragment} fragmentToBeRemoved
 */
ViewportFragmentManager.prototype._removeFragment = function(fragmentToBeRemoved) {
	var index = _.indexOf(this.fragments, fragmentToBeRemoved);
	if (index >= 0) {
		this.removeFragmentAtIndex(index);
	}
};

/**
 * Remove a fragment at a certain index
 * @param index
 */
ViewportFragmentManager.prototype.removeFragmentAtIndex = function(index) {
	this.fragments.splice(index, 1);
};

/**
 * Clean out all dirty fragments if there is at least 1 fetched non-dirty fragment
 */
ViewportFragmentManager.prototype.cleanupDirtyFragments = function() {
	var cleanFragments = _.filter(this.fragments, function(fragment) {
		return !fragment.isDirty();
	});

	// if the cleanFragments and fragments differ in length, we should try to clean
	if (cleanFragments.length != this.fragments.length) {
		// only clean up if there is at least one fetched clean fragment available
		if (
			_.some(cleanFragments, function(fragment) {
				return fragment.isFetched();
			})
		) {
			this.fragments = cleanFragments;
		}
	}
	this.dirtyInitialFragment = null;
};

/**
 * Mark all available fragments as dirty
 */
ViewportFragmentManager.prototype.markAllAsDirty = function() {
	//first clean up
	this.cleanupDirtyFragments();
	_.each(this.fragments, function(fragment) {
		fragment.makeDirty();
	});
	this.dirtyInitialFragment = this.initialFragment;
	if (this.dirtyInitialFragment) {
		this.dirtyInitialFragment.makeDirty();
	}

	this.initialFragment = null;
};

/**
 * Cleanup all fragments
 */
ViewportFragmentManager.prototype.cleanupFragments = function() {
	this.fragments = [];
	this.initialFragment = null;
	this.dirtyInitialFragment = null;
};

/**
 * Get the document rectangle, but make sure the width and height are integer numbers in the pixel domain
 * @returns {ShapeUtils.Rect} the document rect
 */
ViewportFragmentManager.prototype.calculateRoundedWorldBounds = function() {
	var worldBounds = this.world.getWorldBounds();
	var zoom = this.world.getInitialStartZoom();
	// make sure width/height are integer values in the pixel domain
	worldBounds.size.width = Math.floor(worldBounds.size.width * zoom) / zoom;
	worldBounds.size.height = Math.floor(worldBounds.size.height * zoom) / zoom;
	return worldBounds;
};

/**
 * Update the initial fragment if the world has changed or no initial fragment is set
 */
ViewportFragmentManager.prototype.updateInitialFragment = function() {
	var worldBounds = this.calculateRoundedWorldBounds();
	if (
		this.world.getInitialStartZoom() > 0 &&
		(!this.initialFragment ||
			(!thresholdEquals(this.initialFragment.getZoom(), this.world.getInitialStartZoom()) ||
				!this.initialFragment.getRect().rectEquals(worldBounds)))
	) {
		if (process.env.NODE_ENV === "development") {
			if (DEBUG) {
				this.f.debug.debug("Fetching initial fragment");
			}
		}
		this.dirtyInitialFragment = this.initialFragment;

		this.initialFragment = new Fragment(worldBounds, this.world.getInitialStartZoom());
		this.initialFragment.isInitial = true;
		this.initialFragment.fetchImage(
			this.fragmentCache,
			this.getUrlForFragment(this.initialFragment),
			_.bind(this._fragmentFetched, this),
			_.bind(function() {
				this.initialFragment = null;
			}, this)
		);
	}
};

export default ViewportFragmentManager;
