import _ from "underscore";
import AbstractModel from "@esko/web-viewer/core/AbstractModel";
import Transformation from "@esko/transformation";
import ShapeUtils from "@esko/shapeutils";
import PrivateChannel from "@esko/web-viewer/core/PrivateChannel";
import "@esko/web-viewer/viewport/model/ViewportPopoverModel";
/**
 * @exports ViewportLayerModel
 * @memberof ViewportModule
 * @augments {Core.AbstractModel}
 * @class The layer model keeps track of the content of the layer
 * It manages the linework or the image fetcher and keeps a transform of the layer
 * @constructor
 *
 * @param {Core.ViewerFacade} f
 * @param {String} IID - unique identifier for the layer
 */
var ViewportLayerModel = function(f, IID) {
	AbstractModel.call(this, f, ["layerModel", IID]); // TODO in the future we need a better naming here... (but for now it doesn't give any problem)

	this.f = f;

	this._name = IID;

	/**
	 * {Transformation}
	 * @private
	 */
	this.transform = new Transformation();

	/**
	 * All popover models
	 * @private
	 */
	this.popoverModels = {};

	/**
	 * All popup models
	 * @private
	 */
	this.popupModels = {};

	// url parameters can be overriden by putting new values in here
	this.fragmentUrlOverwriteOptions = {};

	/**
	 * Is the layer visible
	 * @private {Boolean}
	 */
	this._visible = true;

	/**
	 * Saved popovers for the layers with this layermodel
	 * Map popover id -> object containing the popoverDiv (where the popover is attached to), and the popover itself
	 * @private {Map.<String,Object>}
	 */
	this._popovers = {};
};

ViewportLayerModel.prototype = Object.create(AbstractModel.prototype);
ViewportLayerModel.prototype.constructor = ViewportLayerModel;

/**
 * Get the name of the layer
 * @returns {String}
 */
ViewportLayerModel.prototype.getName = function() {
	return this._name;
};

/**
 * Save a popover in the layermodel, this popover can be used by all the layers that use this layermodel.
 * @param {Object} popover
 * @param {String} id
 */
ViewportLayerModel.prototype.savePopover = function(popover, id) {
	this._popovers[id] = popover;
};

/**
 * Get a saved popover with given id.
 * @param {String} id
 * @returns {Object}
 */
ViewportLayerModel.prototype.getPopover = function(id) {
	return this._popovers[id];
};

/**
 * Returns the collection of all the saved popovers.
 * @returns {Map.<String,Object>}
 */
ViewportLayerModel.prototype.getAllPopovers = function() {
	return this._popovers;
};

/**
 * Remove a popover from the layer model.
 * @param {String} id
 */
ViewportLayerModel.prototype.removePopover = function(id) {
	delete this._popovers[id];
};

/**
 * Add a popover model to the layer model
 * @param {ViewportPopoverModel} model of the popover
 */
ViewportLayerModel.prototype.addPopoverModel = function(popoverModel) {
	if (!this.popoverModels[popoverModel.getID()]) {
		this.popoverModels[popoverModel.getID()] = popoverModel;
		this._emit("popoverAdded", popoverModel.getID());
	}
};

/**
 * Remove a popover from the layer
 * @param {String} popoverID - ID of the popover model to remove
 */
ViewportLayerModel.prototype.removePopoverModel = function(popoverID) {
	delete this.popoverModels[popoverID];
	this._emit("popoverRemoved", popoverID);
};

/**
 * Get the popover with id.
 * @param {String} popoverID
 */
ViewportLayerModel.prototype.getPopoverModel = function(popoverID) {
	return this.popoverModels[popoverID];
};

/**
 * Add popup model
 * @param {ViewportPopupModel} popupModel - Popup model to add.
 */
ViewportLayerModel.prototype.addPopupModel = function(popupModel) {
	this.popupModels[popupModel.getId()] = popupModel;
};

/**
 * Remove a popup model
 * @param {String} popupModelId - Id of the popup model.
 */
ViewportLayerModel.prototype.removePopupModel = function(popupModelId) {
	var popupModel = this.popupModels[popupModelId];
	popupModel.onDestroy();
	delete this.popupModels[popupModelId];
};

/**
 * Remove all the popup models
 */
ViewportLayerModel.prototype.removeAllPopupModels = function() {
	_.each(
		this.popupModels,
		function(popupModel) {
			this.removePopupModel(popupModel.getId());
		},
		this
	);
};

/**
 * Get the popup model for the given id
 * @param {String} popupModelId - Id of the popup model
 * @returns {ViewportPopupModel} - Popup model with the given Id or undefined if not found
 */
ViewportLayerModel.prototype.getPopupModel = function(popupModelId) {
	return this.popupModels[popupModelId];
};

/**
 * Get all the popup models.
 * @returns {Object} Object with id -> popup model mapping.
 */
ViewportLayerModel.prototype.getPopupModels = function() {
	return this.popupModels;
};

/**
 * Get the linework that needs to be rendered on the layer
 * @param {Boolean} [animated] - indicate if the layer is being animated at the time
 * @param {Boolean} [rotationanimation] - indicate if the layer is being animated with a rotation animation
 * @param {ShapeUtils.Rect} [drawrect] - the rectangle for which the layermodel needs to return linework, the visible rectangle
 * @param {String} [viewName] - name of the view to get the grid from
 * @param {String} [panelName] - the name of the panel this layer is part of
 * @returns {Array} [empty]
 *
 */
ViewportLayerModel.prototype.getLinework = function(animated, rotationanimation, drawrect, viewName, panelName) {
	return [];
};

/**
 * Returns background options for the canvas
 * @returns {Object} object containing the background options
 */
ViewportLayerModel.prototype.getBackgroundOptions = function() {
	return {};
};

/**
 * Get the transform
 * @returns {Transformation}
 */
ViewportLayerModel.prototype.getTransform = function() {
	return this.transform;
};

/**
 * Get the rotation of the layer model (needs to implemented by specific layer model)
 * @return {Number}
 */
ViewportLayerModel.prototype.getRotation = function() {
	return 0;
};

/**
 * Get the flipped state of the layer model (needs to implemented by specific layer model)
 * @return {Boolean}
 */
ViewportLayerModel.prototype.isFlipped = function() {
	return false;
};

/**
 * Reset the transformation
 */
ViewportLayerModel.prototype.resetTransformation = function() {
	this.transform = new Transformation();
};

/**
 * Get the maximum resolution
 * @returns {Number} [-1]
 */
ViewportLayerModel.prototype.getMaximumResolution = function() {
	return -1;
};

/**
 * Emit this model changed
 */
ViewportLayerModel.prototype.modelChanged = function() {
	this._emit("modelChanged");
};

/**
 * Preload a fragment (only needed when shouldAutoLoad returns false
 */
ViewportLayerModel.prototype.preloadFragment = function(vp, zoom) {};

/**
 * Request the url for a fragment
 * @param fragment
 * @returns {Object} context for the fragment (rectangle, resolution, servlet and params)
 */
ViewportLayerModel.prototype.getContextForFragment = function(fragment) {
	var context = {
		rectangle: fragment.getRect(),
		resolution: fragment.getZoom(),
		// the servlet that should be used to fetch the image information
		servlet: "",
		// the parameters used for requesting the image
		params: {
			x1: 0,
			x2: 0,
			y1: 0,
			y2: 0
		}
	};

	// ask all interceptors to update the context. All listeners have access to the context object and can adapt the information that is in the object
	// there is no predefined order in which these interceptors are called, their actions thus should not overlap!
	this.f.emit("intercept.fragmentURL", context);

	// override paramers
	this.f.getDataAdapterForModule("viewport").applyViewportLayerOverwriteOptions(context, this.fragmentUrlOverwriteOptions);

	return context;
};

/**
 * Returns the color of the cad linework
 */
ViewportLayerModel.prototype.getLineworkColor = function() {
	return null;
};

/**
 * Returns the transparency (alpha) of the layer
 */
ViewportLayerModel.prototype.getAlpha = function() {};

/**
 * Set an overwrite option
 * @param {String} option - option to be overwritten
 * @param {String} value - value for the option
 */
ViewportLayerModel.prototype.setOverwriteOption = function(option, value) {
	var oldValue = this.fragmentUrlOverwriteOptions[option];

	this.fragmentUrlOverwriteOptions[option] = value;
	this._fireChange(false, ["overwriteOption", option], oldValue, value);
};

/**
 * Set overwrite options
 * @param {Dictionary} options - options to be overwritten
 */
ViewportLayerModel.prototype.setOverwriteOptions = function(options) {
	var oldValue = this.fragmentUrlOverwriteOptions.slice(0);

	this.fragmentUrlOverwriteOptions = options;
	this._fireChange(false, ["overwriteOption"], oldValue, options.slice(0));
};

/**
 * Get overwrite options
 * For DEBUGGING
 */
ViewportLayerModel.prototype.getOverwriteOptions = function() {
	return this.fragmentUrlOverwriteOptions;
};

/**
 * Remove an overwrite option
 * @param {String} option to be removed
 */
ViewportLayerModel.prototype.removeOverwriteOption = function(option) {
	var oldValue = this.fragmentUrlOverwriteOptions[option];
	delete this.fragmentUrlOverwriteOptions[option];

	this._fireChange(false, ["overwriteOption", option], oldValue, undefined);
};

/**
 * Destroy all listeners
 */
ViewportLayerModel.prototype.destroy = function() {};

/**
 * Should the Layer fetch the fragments himself?
 * @returns {Boolean}
 */
ViewportLayerModel.prototype.shouldAutoFetchFragments = function() {
	return false;
};

/**
 * @returns {Boolean} if the layer contains graphics (and these should be autofetched
 */
ViewportLayerModel.prototype.hasGraphics = function() {
	return false;
};

/**
 * Check if the layer model has Linework
 * @returns {Boolean}
 */
ViewportLayerModel.prototype.hasLinework = function() {
	return false;
};

/**
 * Does the layer model have a buffer
 * @returns {Boolean}
 */
ViewportLayerModel.prototype.hasBuffer = function() {
	return false;
};

/**
 * Returns the buffer, if any needs to be implemented, returns null if no buffer
 * @returns {null|Fragment}
 */
ViewportLayerModel.prototype.getBuffer = function() {
	return null;
};

/**
 * Mark all frames as dirty
 */
ViewportLayerModel.prototype._markAllAsDirty = function() {
	this._emit("fragmentsDirty");
};

/**
 * Clean up all fragments/frames
 */
ViewportLayerModel.prototype._cleanupFragments = function() {
	this._emit("cleanupFragments");
};

/**
 * Hide/show layer
 * @param toggle
 */
ViewportLayerModel.prototype.toggle = function(toggle) {
	var oldValue = this._visible;
	this._visible = toggle;
	this._fireChange(false, ["visible"], oldValue, toggle);
};

/**
 * Is the layer visible
 * @returns {Boolean}
 */
ViewportLayerModel.prototype.isVisible = function() {
	return this._visible;
};

/**
 * returns the type of the viewport
 */
ViewportLayerModel.prototype.getType = function() {
	return "viewport";
};

/**
 * Render the model on the canvas
 * @param {Context2D} ctx
 * @param {ShapeUtils.Rect} drawrect
 * @param {Number} zoom
 * @param {Transformation} t - transformation for this layer
 */
ViewportLayerModel.prototype.render = function(ctx, drawrect, zoom, t) {
	// empty implementation by default
};

/**
 * Calculates the transformation the layer model, takes into account that document page can been flipped, rotated or has an offset.
 * @param {ShapeUtils.Rect} worldBounds - bounds of the world
 * @param {ShapeUtils.Size} pageSize - size of the document page
 * @param {ShapeUtils.Size} offset - offset for the annotations
 * @param {Number} rotation - the rotation of the layer (0,90,180 or 270)
 * @param {Boolean} isFlipped - is the layer for the annotations flipped
 * @param {Number} postRotation - the rotation of the layer, after offset has been added to the transformation (rotates around world middle)
 * @param {Boolean} isPostFlipped - is the layer for the annotation flipped after the offset has been added
 * @param {ShapeUtils.Point} [rotationPoint] - rotation point for the transformation, if not provided (0,0) is used to rotate
 * @param {ShapeUtils.Point} [flipPoint] - point to flip around
 */
ViewportLayerModel.prototype.calculateTransformation = function(
	worldBounds,
	pageSize,
	offset,
	rotation,
	isFlipped,
	postRotation,
	isPostFlipped,
	rotationPoint,
	flipPoint
) {
	if (worldBounds !== undefined && worldBounds.size !== undefined && pageSize !== undefined) {
		this.offset = offset; //set the offset for the layer
		//rotate the transformation around 0,0 by moving the transformation centrum to 0,0
		this.getTransform().rotate(-rotation * (Math.PI / 180), worldBounds.size.width / 2, -worldBounds.size.height / 2);

		if (isFlipped) {
			//if the reference document page is flipped, flip the layerModel
			//first translate the layer to the 0,0 of the transformation
			this.getTransform().translate(worldBounds.size.width / 2, -worldBounds.size.height / 2);
			this.getTransform().flip(0, 0); //flip
		}

		var isRotated = rotation % 180 === 90;
		if (isRotated && isFlipped) {
			//the origin seems to have been inverted both for x- and y-axis, so invert all the translations
			this.getTransform().translate(worldBounds.size.width / 2, -worldBounds.size.height / 2);
			this.getTransform().translate(-this.offset.width, -this.offset.height); //do the translate for the offset
		} else if (isFlipped) {
			//redo the translations
			this.getTransform().translate(-worldBounds.size.width / 2, worldBounds.size.height / 2);
			this.getTransform().translate(this.offset.width, this.offset.height); //do the translate for the offset
		} else {
			//not flipped so only do the translation of the offset
			this.getTransform().translate(this.offset.width, this.offset.height);
		}

		//calculate the offset which the already constructed transformation will use to transform all to the correct position
		var transformOffset = new ShapeUtils.Size(
			worldBounds.size.width / 2 - this.getTransform().e,
			worldBounds.size.height / 2 - this.getTransform().f
		);
		this.getTransform().translate(-transformOffset.width, -transformOffset.height);
		this.getTransform().preTranslate(transformOffset.width, transformOffset.height);

		//if the layer needs to be flipped or rotated after an offset was added use preFlip and preRotate
		if (isPostFlipped) {
			if (flipPoint) {
				this.getTransform().preTranslate(-flipPoint.x, -flipPoint.y);
			}
			this.getTransform().preFlip(0, 0);
			if (flipPoint) {
				this.getTransform().preTranslate(flipPoint.x, flipPoint.y);
			}
		}

		if (postRotation) {
			//currently rotates around the middle of the view
			if (!rotationPoint) {
				this.getTransform().preRotate(postRotation * (Math.PI / 180), 0, 0);
			} else {
				this.getTransform().preRotate(postRotation * (Math.PI / 180), rotationPoint.x, rotationPoint.y);
			}
		}
	}
};

export default ViewportLayerModel;
