/**
 * @namespace ViewportModule
 */
/**
 * @namespace ViewportModule
 */
import AbstractModel from "@esko/web-viewer/core/AbstractModel";

import _ from "underscore";
/**
 * @exports ViewportPanelModel
 * @memberof ViewportModule
 * @class ViewportPanelModel - a model for panels
 * @augments Core.AbstractModel
 * @constructor
 * @param {Core.ViewerFacade} f
 * @param {String} name - the name of the panel
 * @param {Object} [overwriteOptions] - key-value pairs with overwrite options
 * @param {Array.<String>} [disabledLayers] - layers to disable
 */
var ViewportPanelModel = function(f, name, overwriteOptions, disabledLayers) {
	AbstractModel.call(this, f, ["module", "viewport", "panel", name]);
	/**
	 * @private
	 * @member {String} name of the panel
	 */
	this._name = name;

	/**
	 * @private
	 */
	this._overwriteOptions = {};
	if (overwriteOptions !== undefined) {
		this._overwriteOptions = _.clone(overwriteOptions);
	}

	/**
	 * @private
	 */
	this._disabledLayers = [];
	if (disabledLayers !== undefined) {
		this._disabledLayers = disabledLayers.slice(0);
	}

	/**
	 * @private
	 * @member {String|null} Text of the header of the panel
	 */
	this._headerText = null;
};

ViewportPanelModel.prototype = Object.create(AbstractModel.prototype);
ViewportPanelModel.prototype.constructor = ViewportPanelModel;

/**
 * Get the name of the panel
 * @returns {String}
 */
ViewportPanelModel.prototype.getName = function() {
	return this._name;
};

/**
 * Toggle a layer on all views (the layer will be hidden)
 * @param {String} name
 * @param {Boolean} toggle
 */
ViewportPanelModel.prototype.toggleLayer = function(name, toggle) {
	var oldValue = this._disabledLayers.slice(0);

	if (toggle) {
		// show the layer
		this._disabledLayers = _.without(this._disabledLayers, name);
	} else {
		if (_.indexOf(this._disabledLayers, name) < 0) {
			this._disabledLayers.push(name);
		}
	}

	this._fireChange(false, ["disabledLayers"], oldValue, this._disabledLayers.slice(0));
};

/**
 * Check if a layer is enabled
 * @param {String} name - the name of the layer
 * @returns {Boolean}
 */
ViewportPanelModel.prototype.isLayerEnabled = function(name) {
	return !this.isLayerDisabled(name);
};

/**
 * Check if a layer is disabled
 * @param name - the name of the layer
 * @returns {Boolean}
 */
ViewportPanelModel.prototype.isLayerdisabled = function(name) {
	return _.indexOf(this._disabledLayers, name) >= 0;
};

/**
 * Get a list of disabled layers
 * @returns {Array.<String>}
 */
ViewportPanelModel.prototype.getDisabledLayers = function() {
	return this._disabledLayers.slice(0);
};

/**
 * Get all the overwrite options for this panel
 *
 * By default a copy of these options is returned, for performance reasons "doNotClone" can be used
 * @param {Boolean} doNotClone - if true the options are passed as reference, they should not be adapted in this case!!
 * @returns {object} the overwrite options as a dictionary
 */
ViewportPanelModel.prototype.getOverwriteOptions = function(doNotClone) {
	if (doNotClone) {
		return this._overwriteOptions;
	} else {
		return _.clone(this._overwriteOptions);
	}
};

/**
 * Get the value for a specific overwrite option
 * @param {String} name - name of the overwite option
 * @returns {Any} the value of the overwrite option
 */
ViewportPanelModel.prototype.getOverwriteOption = function(name) {
	return this._overwriteOptions[name];
};

/**
 * Set a specific overwrite option
 * @param {String} name - name of the overwrite option
 * @param {Any} value - value of the overwrite option (probably a String value)
 */
ViewportPanelModel.prototype.setOverwriteOption = function(name, value) {
	var oldValue = this.getOverwriteOption(name);
	this._overwriteOptions[name] = value;

	this._fireChange(false, ["overwriteOption", name], oldValue, value);
};

/**
 * Set multiple overwrite options in one go
 * @param {Object} options - object with key/value pairs
 */
ViewportPanelModel.prototype.setOverwriteOptions = function(options) {
	_.each(
		options,
		function(value, name) {
			this.setOverwriteOption(name, value);
		},
		this
	);
};

/**
 * Delete a specific overwrite option
 * @param {String} name - name of the overwrite option to delete
 */
ViewportPanelModel.prototype.deleteOverwriteOption = function(name) {
	var oldValue = this.getOverwriteOption(name);
	delete this._overwriteOptions[name];

	// intentional use of undefined!
	this._fireChange(false, ["overwriteOption", name], oldValue, undefined);
};

/**
 * Get the header text for the panel
 * @returns {String|null} the header text or null if not defined
 */
ViewportPanelModel.prototype.getHeaderText = function() {
	return this._headerText;
};

/**
 * Change the header text
 * @param {String|null} headerText
 */
ViewportPanelModel.prototype.setHeaderText = function(headerText) {
	var oldValue = this._headerText;

	this._headerText = headerText;
	this._fireChange(false, ["headerText"], oldValue, headerText);
};

export default ViewportPanelModel;
