import _ from "underscore";
import PrivateChannel from "@esko/web-viewer/core/PrivateChannel";
/**
 * Construct a new viewport popup model
 * @exports ViewportPopupModel
 * @memberof ViewportModule
 * @class
 * This model is responsible for the state (shown/not shown) of the popup.
 * It also contains the popup itself.
 * @param {Core.ViewerFacade} f - Viewerfacade
 * @param {String} id - Id of the viewport popup model
 */
var ViewportPopupModel = function(f, id) {
	this.f = f;
	this.id = id;
	this.shown = true;
};

/**
 * Get the id of the popup model
 * @returns {String} The id of the popup model
 */
ViewportPopupModel.prototype.getId = function() {
	return this.id;
};

/**
 * Indicate that the popup should be shown.
 */
ViewportPopupModel.prototype.show = function() {
	this.shown = true;
};

/**
 * Indicate that the popup should be hidden.
 */
ViewportPopupModel.prototype.hide = function() {
	this.shown = false;
};

/**
 * Check whether the popup is shown.
 * @returns {Boolean} True if the popup is shown, false otherwise/
 */
ViewportPopupModel.prototype.isShown = function() {
	return this.shown;
};

/**
 * Get the popup belonging to this model.
 * Override this function.
 * @param {ViewportLayer} viewportLayer - The layer where the popups should be shown.
 * @returns {Popup} The popup belonging to this model.
 */
ViewportPopupModel.prototype.getPopup = function(viewportLayer) {
	// override
	return undefined;
};

/**
 * Called when the popup model is destroyed.
 * Override this function.
 * Useful to do some cleanup.
 */
ViewportPopupModel.prototype.onDestroy = function() {
	// override to do some cleanup
};

/**
 * Return the world origin of the popup
 * Override this function.
 * @returns {ShapeUtils.Point} - Origin of the popup.
 */
ViewportPopupModel.prototype.getWorldOrigin = function() {
	// override
	return undefined;
};

/**
 * Returns the world size of the popup
 * Override this function.
 * @returns {ShapeUtils.Size} - Size of the popup.
 */
ViewportPopupModel.prototype.getWorldSize = function() {
	// override
	return undefined;
};

/**
 * When there needs to be added an offset to the popup location (in scene pixels), return the offset here
 * @returns {ShapeUtils.Size} offset in pixels that needs to be added to the popuplocation
 */
ViewportPopupModel.prototype.getSceneOffset = function() {
	return undefined;
};

/**
 * Function called by the layer after translating the scene rectangle to world coordinates.
 * Override this function.
 * @param {ShapeUtils.Rect} worldRect - Rectangle in world coordinates
 * @param {Boolean} autoAdapted - True if the event is fired because the content was auto adapted (e.g. from a makeMeVisible trigger from an element in the content), false otherwise
 */
ViewportPopupModel.prototype.onResized = function(worldRect, autoAdapted) {
	// override
};

/**
 * Function called by the layer after translating the scene point to world coordinates.
 * Override this function.
 * @param {ShapeUtils.Point} worldPoint - Point in world coordinates
 */
ViewportPopupModel.prototype.onDragged = function(worldPoint) {
	// override
};

export default ViewportPopupModel;
