import Transformation from "@esko/transformation";
import PrivateChannel from "@esko/web-viewer/core/PrivateChannel";
import _ from "underscore";
import "@esko/web-viewer/viewport/model/ViewportLayerModel";
/**
 * @exports ViewportViewModel
 * @memberof ViewportModule
 * @class A view model class is responsible for keeping track of the layers of one view
 * @param {Core.ViewerFacade} f - Facade
 * @param {String} IID
 * @param {String} title
 * @param {Boolean} isDefaultView
 */
var ViewportViewModel = function(f, IID, title, isDefaultView) {
	PrivateChannel.installTo(this, f, ["viewport", "viewmodel", IID]);

	/**
	 * Layer models
	 * {ViewportLayerModel}
	 */
	this._layers = {};

	/**
	 * The title of the view
	 * @private
	 */
	this._title = title;

	/**
	 * ID of the view
	 * @private
	 */
	this._IID = IID;

	/**
	 * Indicates whether the view is shown
	 * @private
	 */
	this._shown = true;

	/**
	 * Indicates whether the this view is the default view
	 * @private
	 */
	this._isDefaultView = isDefaultView;

	/**
	 * {Transformation}
	 * @private
	 */
	this._transform = new Transformation();
};

/**
 * @returns {Object.<String,ViewportLayerModel}
 */
ViewportViewModel.prototype.getLayers = function() {
	return this._layers;
};

/**
 * Add a layer model to the view model
 * @param {String} name
 * @param {ViewportLayerModel} layer
 */
ViewportViewModel.prototype.addLayer = function(name, layer) {
	this._layers[name] = layer;
	this._emit("layerAdded", name);
};

/**
 * removes a layer model from the view model
 * @param {String} name
 */
ViewportViewModel.prototype.removeLayer = function(name) {
	delete this._layers[name];
	this._emit("layerRemoved", name);
};

/**
 * get a layer model
 * @param {String} name
 * @returns {ViewportLayerModel}
 */
ViewportViewModel.prototype.getLayer = function(name) {
	return this._layers[name];
};

/**
 * Get the number of layer models
 * @returns {Number}
 */
ViewportViewModel.prototype.numberOfLayers = function() {
	return _.keys(this._layers).length;
};

/**
 * returns the name of layer at a given index
 * @param {Number} index
 * @returns {String}
 */
ViewportViewModel.prototype.nameForLayerAtIndex = function(index) {
	return _.keys(this._layers)[index];
};

/**
 * Get the model of a layer
 * @param {String} name
 * @returns {ViewportLayerModel}
 */
ViewportViewModel.prototype.modelForLayer = function(name) {
	return this._layers[name];
};

/**
 * Get the layer model of layer at a certain index
 * @param {Number} index
 * @returns {ViewportLayerModel}
 */
ViewportViewModel.prototype.modelForLayerAtIndex = function(index) {
	return this._layers[this.nameForLayerAtIndex(index)];
};

/**
 * Transform the layer
 * @param layername
 * @param tr
 */
ViewportViewModel.prototype.transformLayer = function(layername, tr) {
	var layer = this._layers[layername];
	if (layer) {
		layer.getTransform().transform(tr);
	}
};

/**
 * Get the unique identifier of the view
 * @returns {String}
 */
ViewportViewModel.prototype.getID = function() {
	return this._IID;
};

/**
 * Returns the index for a layer
 */
ViewportViewModel.prototype.indexForLayer = function(layername) {
	return _.indexOf(_.keys(this._layers), layername);
};

/**
 * Returns the transformation of the view
 * @returns {Transformation}
 */
ViewportViewModel.prototype.getTransform = function() {
	return this._transform;
};

/**
 * Set the title of the view
 * @param {String} title - Title of the view
 */
ViewportViewModel.prototype.setTitle = function(title) {
	if (this._title !== title) {
		this._title = title;
		this._emit("titleChanged");
	}
};

/**
 * Get the title of the view
 * @returns {String} Title of the view
 */
ViewportViewModel.prototype.getTitle = function() {
	return this._title;
};

/**
 * Set whether the view is shown or not
 * @param {Boolean} shown - True if the view should be shown, false otherwise
 */
ViewportViewModel.prototype.setShown = function(shown) {
	if (this._shown !== shown) {
		this._shown = shown;
		this._emit("showChanged");
	}
};

/**
 * Get whether the view is shown or not
 * @returns {Boolean} True if the view is shown, false otherwise
 */
ViewportViewModel.prototype.isShown = function() {
	return this._shown;
};

/**
 * Enable/disable a layer (if it is in the ViewModel)
 * @param {String} layerName - name of the layer
 * @param {Boolean} visible
 */
ViewportViewModel.prototype.toggleLayer = function(layerName, visible) {
	var layer = this.getLayer(layerName);
	if (layer !== undefined) {
		layer.toggle(visible);
	}
};

/**
 * is this view the default view.
 * @returns {Boolean}
 */
ViewportViewModel.prototype.isDefaultView = function() {
	return this._isDefaultView;
};

export default ViewportViewModel;
