import AbstractModel from "@esko/web-viewer/core/AbstractModel";
import ShapeUtils from "@esko/shapeutils";
import ViewportConstants from "@esko/web-viewer/viewport/ViewportConstants";
import OrderedObject from "@esko/orderedobject";
import ViewportPanelModel from "@esko/web-viewer/viewport/model/ViewportPanelModel";
import _ from "underscore";
/**
 * @exports ViewportWidgetModel
 * @memberof ViewportModule
 * @class A widget model is an abstract class that manages the views on one widget
 * @constructor
 * @param {Core.ViewerFacade} f - Facade
 * @param {String} id - unique id
 */
var ViewportWidgetModel = function(f, id) {
	AbstractModel.call(this, f, ["viewport", "widget", "model"]);

	// id of the widget
	this.id = id;

	/**
	 * collection of ViewportViewModel
	 * @private
	 * @type {OrderedObject.<ViewportModule.ViewportViewModel>}
	 */
	this._viewportViewModels = new OrderedObject();

	/**
	 * Keep track of the global layers that can be added to newly created views
	 * @private
	 */
	this._globalLayers = {};

	/**
	 * The current views layout mode
	 * @private
	 */
	this._viewsLayoutMode = ViewportConstants.ViewsLayoutMode.AUTOMATIC;
	this._panelsLayoutMode = ViewportConstants.PanelsLayoutMode.AUTOMATIC;
	this._sizeForView = new ShapeUtils.Size();

	/**
	 * The panels in the widget, by default there is only 1 panel with no special options
	 * This is a list of PanelModel!
	 * @private
	 * @type {OrderedObject.<ViewportModule.ViewportPanelModel>}
	 */
	this._panels = new OrderedObject();
	this.addPanel("default", {
		isDefaultView: true
	});

	/**
	 * Is the mouse active in the viewport at this moment.
	 */
	this._isActive = false;
};

ViewportWidgetModel.prototype = Object.create(AbstractModel.prototype);
ViewportWidgetModel.prototype.constructor = ViewportWidgetModel;

//------------ general ------------------------------------------------------------------------------

/**
 * @returns {Number} the size of the World
 */
ViewportWidgetModel.prototype.getWorldSize = function() {
	return undefined;
};

/**
 * @returns {ShapeUtils.Rect} the rect of the World
 */
ViewportWidgetModel.prototype.getWorldBounds = function() {
	return undefined;
};

/**
 * Get the resolution of the loaded document in the world
 * @returns {Object} resolutions of the document in ppmm
 * {x: {Number} x-resolution,
 *  y: {Number} y-resolution }
 */
ViewportWidgetModel.prototype.getWorldResolution = function() {
	return {
		x: 0,
		y: 0
	};
};

/**
 * Set if the mouse is currently active in the viewport widget
 * @param isActive
 */
ViewportWidgetModel.prototype.setWidgetActive = function(isActive) {
	this._isActive = isActive;
};

/**
 * Is the viewport widget active, it the mouse cursor in the viewport
 * @returns {Boolean}
 */
ViewportWidgetModel.prototype.isWidgetActive = function() {
	return this._isActive;
};

//------------ views ------------------------------------------------------------------------------

/**
 * Create a new view with options
 * @param {String} id - the id of the view
 * @param {String} title - the title of the view
 * @param {Dictionary} isDefaultView - are the view with this viewmodel default views
 */
ViewportWidgetModel.prototype.createView = function(id, title, isDefaultView) {
	var model = this.createModelForView(id, title, isDefaultView);
	this._addView(id, model);
};

/**
 * Add a view to the model
 * (private function: use "createView" to generate one)
 */
ViewportWidgetModel.prototype._addView = function(viewID, model) {
	// If it already exists
	if (this._viewportViewModels.getItemByKey(viewID)) {
		// update the view
		throw "Trying to add view with already existing name";
	} else {
		// Add it
		var oldOrder = this.getViewsOrder();
		this._viewportViewModels.addItem(viewID, model);

		// Add all global layers
		_.each(
			this._globalLayers,
			function(layer, layername) {
				this.addLayer(layername, layer, viewID);
			},
			this
		);

		this._fireChange(false, ["view", "add"], undefined, model);
		this._fireChange(false, ["view", "order"], oldOrder, this.getViewsOrder());
	}
};

/**
 * Hide or show the view with the specified id
 * @param {String} id - Id of the view
 * @param {Boolean} show - True if the view should be shown, false otherwise
 */
ViewportWidgetModel.prototype.toggleView = function(id, show) {
	if (this._viewportViewModels.getItemByKey(id) !== undefined) {
		var oldValue = this._viewportViewModels.getItemByKey(id).isShown();

		this._viewportViewModels.getItemByKey(id).setShown(show);
		this._fireChange(false, ["view", "toggle", id], oldValue, show);
	}
};

/**
 * Removes a view with a certain id
 * @param {String} viewID - id of the view
 */
ViewportWidgetModel.prototype.removeView = function(viewID) {
	var oldOrder = this.getViewsOrder();
	var model = this._viewportViewModels.removeByKey(viewID);
	this._fireChange(false, ["view", "remove"], model, undefined);
	this._fireChange(false, ["view", "order"], oldOrder, this.getViewsOrder());
};

/**
 * @returns {Number} number of viewportViewModels in the model
 */
ViewportWidgetModel.prototype.getViewCount = function() {
	return this._viewportViewModels.getLength();
};

/**
 * @returns {ViewportModule.ViewportViewModel} a view with a certain id
 */
ViewportWidgetModel.prototype.getViewModel = function(viewID) {
	return this._viewportViewModels.getItemByKey(viewID);
};

/**
 * Note, if you are only interested in visible views use getIDForShownViewsAtIndex
 * @returns {String} id of the view at a certain index
 */
ViewportWidgetModel.prototype.getIDForViewAtIndex = function(index) {
	return this._viewportViewModels.getKeyByIndex(index);
};

/**
 * @returns {Viewportmodule.ViewportViewModel} the model for a view at a certain index
 */
ViewportWidgetModel.prototype.getModelForViewAtIndex = function(index) {
	return this._viewportViewModels.getItemByIndex(index);
};

/**
 * Get all available views as a list
 * @returns {Array.<ViewportModule.ViewportViewModel>}
 */
ViewportWidgetModel.prototype.getViews = function() {
	return this._viewportViewModels.getOrderedObjectList();
};

/**
 * Get the keys of all the views that are visible
 * @returns {Array.<String>}
 */
ViewportWidgetModel.prototype.getShownViewKeys = function() {
	return _.filter(
		this._viewportViewModels.getKeys(),
		function(key) {
			return this._viewportViewModels.getItemByKey(key).isShown();
		},
		this
	);
};

/**
 * This funciton will return the key of the view at a certain index
 * This method only scans through visible views (use getIDForViewAtIndex if also hidden views should be taken into account)
 * @param index
 * @returns
 */
ViewportWidgetModel.prototype.getIDForShownViewsAtIndex = function(index) {
	// only use visible views to determine the key of the  view at position index
	var shownViewKeys = this.getShownViewKeys();
	return shownViewKeys[index];
};

/**
 * Create a model for a view (Method should be overwritten by the implementation) ViewportViewModel
 * @param {String} id - id of the view
 * @param {String} title - Title of the view
 * @param {Boolean} isDefaultView - are the view with this view model default views
 */
ViewportWidgetModel.prototype.createModelForView = function(id, title, isDefaultView) {
	throw "Not Implemented";
};

/**
 * Set the order of the views
 * Probably this method should be moved to ordered object (where it can be unit tested!)
 *
 * @param {Array.<String>} ordered keys
 */
ViewportWidgetModel.prototype.setViewsOrder = function(orderedKeys) {
	var oldOrder = this.getViewsOrder();

	var originalKeys = this._viewportViewModels.getKeys();

	var nonMentionedKeys = _.difference(originalKeys, orderedKeys, ["Main"]);

	if (_.indexOf(originalKeys, "Main") >= 0) {
		orderedKeys = _.without(orderedKeys, "Main");
		orderedKeys.unshift("Main");
	}

	orderedKeys = orderedKeys.concat(nonMentionedKeys);

	this._viewportViewModels.setOrder(orderedKeys);

	this._fireChange(false, ["view", "order"], oldOrder, this.getViewsOrder());
};

/**
 * Get the order of the views (in keys)
 * @returns {Array.<String>}
 */
ViewportWidgetModel.prototype.getViewsOrder = function() {
	return this._viewportViewModels.getKeys();
};

/**
 * Set the currently used size for a view
 * @param size
 */
ViewportWidgetModel.prototype.setSizeForView = function(size) {
	var oldSize = this._sizeForView;
	this._sizeForView = size;
	this._fireChange(false, ["view", "size"], oldSize, size);
};

/**
 * Get the currently set view size (this is the same for all the views)
 * @returns {ShapeUtils.Size}
 */
ViewportWidgetModel.prototype.getSizeForView = function() {
	return this._sizeForView;
};

//------------ views layout mode ------------------------------------------------------------------------------

/**
 * Get the current views layout mode (this is how views are shown inside a panel)
 *
 * @returns {ViewportConstants.ViewsLayoutMode} Current views layout mode
 */
ViewportWidgetModel.prototype.getViewsLayoutMode = function() {
	return this._viewsLayoutMode;
};

/**
 * Set a new layout mode for the views
 * @param {ViewportConstants.ViewsLayoutMode} viewsLayoutMode - The new views layout mode
 */
ViewportWidgetModel.prototype.setViewsLayoutMode = function(viewsLayoutMode) {
	var oldValue = this._viewsLayoutMode;

	this._viewsLayoutMode = viewsLayoutMode;
	this._fireChange(false, ["layout", "views"], oldValue, this._viewsLayoutMode);
};

/**
 * Get the current panels layout mode (With AUTOMATIC resolved to the correct mode)
 * (this depends on panelslayout mode and count of panels, if there is 1 panel just use automatic (or other setting)
 * if more than one panel choose the best option
 */
ViewportWidgetModel.prototype.getResolvedViewsLayoutMode = function() {
	if (this.getPanelCount() > 1 && this._viewsLayoutMode === ViewportConstants.PanelsLayoutMode.AUTOMATIC) {
		if (this._panelsLayoutMode === ViewportConstants.PanelsLayoutMode.HORIZONTAL) {
			return ViewportConstants.ViewsLayoutMode.VERTICAL;
		} else {
			return ViewportConstants.ViewsLayoutMode.HORIZONTAL;
		}
	} else {
		return this._viewsLayoutMode;
	}
};

//------------ panels layout mode ------------------------------------------------------------------------------
/**
 * Get the current panels layout mode (this is how panels are shown inside the viewport )
 * @returns {ViewportConstants.PanelsLayoutMode} Current panels layout mode
 */
ViewportWidgetModel.prototype.getPanelsLayoutMode = function() {
	return this._panelsLayoutMode;
};

/**
 * Get the current panels layout mode (With AUTOMATIC resolved to the correct mode)
 * @returns {ViewportConstants.PanelsLayoutMode}
 */
ViewportWidgetModel.prototype.getResolvedPanelsLayoutMode = function() {
	var mode = this._panelsLayoutMode;
	if (mode === ViewportConstants.PanelsLayoutMode.AUTOMATIC) {
		if (this._viewsLayoutMode === ViewportConstants.ViewsLayoutMode.VERTICAL) {
			mode = ViewportConstants.PanelsLayoutMode.HORIZONTAL;
		} else {
			// automatic views layout or vertical
			mode = ViewportConstants.PanelsLayoutMode.VERTICAL;
		}
	}
	return mode;
};

/**
 * Set a new layout mode for the panels
 * @param {ViewportConstants.PanelsLayoutMode} panelsLayoutMode - The new panels layout mode
 */
ViewportWidgetModel.prototype.setPanelsLayoutMode = function(panelsLayoutMode) {
	var oldValue = this._panelsLayoutMode;

	this._panelsLayoutMode = panelsLayoutMode;
	this._fireChange(false, ["layout", "panels"], oldValue, this._panelsLayoutMode);
};

//------------ panels  ------------------------------------------------------------------------------
/**
 * Add panel, the panel is added in the end of the list
 * if the panel already exists, the panel is moved to the end
 * @param {String} name - name of the panel to add.
 * @param {Object} [overwriteOptions] - key-value pairs with overwrite options
 * @param {Array.<String>} [disabledLayers] - layers to disable
 * @returns {ViewportModule.ViewportPanelModel}
 */
ViewportWidgetModel.prototype.addPanel = function(name, overwriteOptions, disabledLayers) {
	var oldOrder = this._panels.getKeys();
	var panel = this._panels.getItemByKey(name);
	if (panel !== undefined) {
		// move the item
		this._panels.moveItemByKeyToEnd(name);
	} else {
		panel = new ViewportPanelModel(this.f, name, overwriteOptions, disabledLayers);
		this._panels.addItem(name, panel);
		this._fireChange(false, ["panel", "add"], undefined, panel);
	}

	this._fireChange(false, ["panel", "order"], oldOrder, this._panels.getKeys());
	return panel;
};

/**
 * Get the model for a specific panel
 * @param {String} name - name of the panel
 * @returns {ViewportModule.ViewportPanelModel}
 */
ViewportWidgetModel.prototype.getPanelModel = function(name) {
	return this._panels.getItemByKey(name);
};

/**
 * Remove a panel
 * @param {String} name - name of the panel to remove
 */
ViewportWidgetModel.prototype.removePanel = function(name) {
	var oldOrder = this._panels.getKeys();

	var panel = this._panels.removeByKey(name);

	this._fireChange(false, ["panel", "remove"], panel, undefined);
	this._fireChange(false, ["panel", "order"], oldOrder, this._panels.getKeys());
};

/**
 * Set the order of the panels
 * !! all panel keys should be in the new order!!
 * @param {Array.<String>} newOrder - order defined as a list of panel names
 */
ViewportWidgetModel.prototype.setPanelOrder = function(newOrder) {
	var oldOrder = this._panels.getKeys();
	this._panels.setOrder(newOrder);
	this._fireChange(false, ["panel", "order"], oldOrder, this._panels.getKeys());
};

/**
 * Get the order of the panels
 * @returns {Array.<String>}
 */
ViewportWidgetModel.prototype.getPanelOrder = function() {
	return this._panels.getKeys();
};

/**
 * Get the amount of active panels
 * @returns {Number}
 */
ViewportWidgetModel.prototype.getPanelCount = function() {
	return this._panels.getLength();
};

/**
 * Get the panel (name) at a certain index
 * @param {Number} index
 * @returns {String} the name of the panel at position "index"
 */
ViewportWidgetModel.prototype.getPanelNameAtIndex = function(index) {
	return this._panels.getKeyByIndex(index);
};

/**
 * Get the panel model at a certain index
 * @param {Number} index
 * @returns {ViewportModule.PanelModel} panel model for the panel at position "index"
 */
ViewportWidgetModel.prototype.getPanelAtIndex = function(index) {
	return this._panels.getItemByIndex(index);
};

//------------ layers ------------------------------------------------------------------------------

/**
 * Add a layer to the views
 * @param {String} layerName - name of the layer
 * @param {LayerModel} layerModel - model of the layer
 * @param {String} [viewID] - if specified, only add the layer to the view with this id. If not specified, add it to all views (and all future views)
 */
ViewportWidgetModel.prototype.addLayer = function(layerName, layerModel, viewID) {
	if (viewID !== undefined) {
		// Only add it to the specified view
		if (this._viewportViewModels.getItemByKey(viewID)) {
			this._viewportViewModels.getItemByKey(viewID).addLayer(layerName, layerModel);
		}
	} else {
		// Add it already to the known views
		_.each(this._viewportViewModels.getItemsAsObject(), function(view) {
			view.addLayer(layerName, layerModel);
		});

		// Save it to add it later to new views
		this._globalLayers[layerName] = layerModel;
	}
};

/**
 * Remove a layer from the views
 * @param {String} layerName - name of the layer
 * @param {String} [viewID] - if specified, only remove the layer from the view with this id. If not specified, remove it from all views
 */
ViewportWidgetModel.prototype.removeLayer = function(layerName, viewID) {
	if (viewID !== undefined) {
		if (this._viewportViewModels.getItemByKey(viewID)) {
			this._viewportViewModels.getItemByKey(viewID).removeLayer(layerName);
		}
	} else {
		_.each(this._viewportViewModels.getItemsAsObject(), function(view) {
			view.removeLayer(layerName);
		});
		delete this._globalLayers[layerName];
	}
};

/**
 * Show/hide a layer
 * If view is provided, only the layer in that view will be toggled
 * @param {String} layerName
 * @param {Boolean} visible
 * @param {String} [viewID] - if provided, only the layer in this view will be toggled
 */
ViewportWidgetModel.prototype.toggleLayer = function(layerName, visible, viewID) {
	if (viewID !== undefined) {
		if (this._viewportViewModels.getItemByKey(viewID)) {
			this._viewportViewModels.getItemByKey(viewID).toggleLayer(layerName);
		}
	} else {
		var layer = this._globalLayers[layerName];
		if (layer) {
			layer.toggle(visible);
		} else {
			// it is not a global layer, try to toggle it by looking in all views
			_.each(this._viewportViewModels.getItemsAsObject(), function(view) {
				view.toggleLayer(layerName, visible);
			});
		}
	}
};

export default ViewportWidgetModel;
