import _ from "underscore";
import ShapeUtils from "@esko/shapeutils";
import Extendable from "@esko/web-viewer/core/Extendable";
import PrivateChannel from "@esko/web-viewer/core/PrivateChannel";
import Mousetrap from "@esko/mousetrap";
/**
 * @exports AbstractTool
 * @memberof ViewportModule
 * @class An abstract tool can interact with the user on a widget (presumably the viewport widget)
 *
 * Tools are managed by the toolmanager, and can be passivated. If a tool is passivated, it can listen to events and request activation based on these events
 *
 * @param f {Core.ViewerFacade}
 * @param IID {String} internal unique ID
 */
var AbstractTool = function AbstractTool(f, IID) {
	Extendable.installTo(this);
	PrivateChannel.installTo(this, f, ["tool", IID]);

	this.f = f;

	var active = false;

	/**
	 * @returns true if the tool is active
	 */
	this.isActive = function() {
		return active;
	};

	/**
	 * @returns the unique internal id of the tool
	 */
	this.getIID = function() {
		return IID;
	};

	/**
	 * Ask the toolmanager to be activated
	 */
	this.requestActivation = function() {
		f.emit("toolManager.request.activation", this);
	};

	/**
	 * Ask the toolmanager to be deactivated
	 */
	this.requestDeactivation = function() {
		f.emit("toolManager.request.deactivation", this);
	};

	/**
	 * It can be very straightforward to define the needed eventlisteners in these two objects,
	 * The events are registered and unregistered at the correct moments
	 */

	/**
	 * events needed when passivated
	 */
	this.passiveEvents = {};
	// structure: per key combo array with two elements ["combo", function(){}] or [[combo1, combo2], function(){}]
	this.passiveKeyEvents = [];

	/**
	 * events needed when activated;
	 */
	this.activeEvents = {};
	this.activeKeyEvents = [];

	// these keypress are always active (both in passive and active mode)
	this.persistentKeyEvents = [];

	/**
	 * This method is called when the tool is activated
	 * @param {Boolean} interruptRequest - set to true if the activation is requested by an interrupt event
	 */
	this.activate = function(interruptRequest) {
		active = true;

		//only activate the tool when the widget is actived
		this._widget.whenContextAvailable(
			_.bind(function() {
				this.doActivation();
				if (this.isActive()) {
					this._emit("activated", interruptRequest);
				}
			}, this)
		);
	};

	/**
	 * Deactivate the tool
	 *
	 * !!should only be called by the toolmanager
	 * @param {Boolean} interruptRequest - set to true if the activation is requested by an interrupt event
	 */
	this.deactivate = function(interruptRequest) {
		this.doDeactivation();
		active = false;
		this._emit("deactivated", interruptRequest);
	};

	/**
	 * State management
	 *
	 * A state is a way of handling events. At any time a tool can change state, which will be in charge of handling the events
	 */
	var states = {}; // list of registered states (by calling registerState)
	var currentState; //name of the current state

	/**
	 * Register all possible active events for this tool
	 */
	this.registerActiveEvents = function(events) {
		_.each(
			events,
			function(event) {
				this.activeEvents[event] = _.bind(function(e) {
					this.getState(currentState)[event](e);
				}, this);
			},
			this
		);
	};

	/**
	 * Register all possible active events for this tool
	 */
	this.registerPassiveEvents = function(events) {
		_.each(
			events,
			function(event) {
				this.passiveEvents[event] = _.bind(function(e) {
					this.getState(currentState)[event](e);
				}, this);
			},
			this
		);
	};

	/**
	 * Register a state with a name and a function
	 */
	this.registerState = function(stateName, stateFunction) {
		states[stateName] = stateFunction;
	};

	/**
	 * Change the state of the tool to a new one
	 */
	this.changeState = function(stateName) {
		if (currentState !== stateName) {
			var previousState = currentState;
			var nextState = stateName;

			// execute leave state if set
			if (states[previousState] && states[previousState].leaveState) {
				states[previousState].leaveState(nextState);
			}
			currentState = nextState;
			this._emit('State changed to "' + nextState + '"');
			this._emit("state.changed." + nextState);
			// execute enter state if set
			if (states[nextState] && states[nextState].enterState) {
				states[nextState].enterState(previousState);
			}
			this.cursorShouldChange();
		}
	};

	/**
	 * Get the function of a state-name
	 */
	this.getState = function(stateName) {
		return states[stateName];
	};

	/**
	 * Get the current state
	 */
	this.getCurrentState = function() {
		return this.getState(currentState);
	};

	/**
	 * Get the current statename
	 */
	this.getCurrentStateName = function() {
		return currentState;
	};
};

/**
 * Convert a point in the widget (scene) to a point in the world
 */
AbstractTool.prototype.sceneToWorld = function(point) {
	if (this._widget) {
		return this._widget.getWorld().getWorldPointFromScene(point);
	} else {
		this.f.emit("error.SEVERE", "Requesting world point with no widget present");
		return point;
	}
};

/**
 * Converts a world size to a scene size
 * @param   {ShapeUtils.Size} size - the world size
 * @returns {ShapeUtils.Size} the scene size
 */
AbstractTool.prototype.worldToSceneSize = function(size) {
	if (this._widget) {
		return this._widget.getWorld().getSceneSizeFromWorld(size);
	} else {
		this.f.emit("error.SEVERE", "Requesting world size with no widget present");
		return size;
	}
};

/**
 * Converts a world point to a scene point
 * @param   {ShapeUtils.Point} point - the world point
 * @returns {ShapeUtils.Point} the scene point
 */
AbstractTool.prototype.worldToScenePoint = function(point) {
	if (this._widget) {
		return this._widget.getWorld().getScenePointFromWorld(point);
	} else {
		this.f.emit("error.SEVERE", "Requesting world point with no widget present");
		return point;
	}
};

/**
 * Converts a world rect to a scene rect
 * @param   {ShapeUtils.Rect} rect - the world rect
 * @returns {ShapeUtils.Rect} the scene rect
 */
AbstractTool.prototype.worldToSceneRect = function(rect) {
	if (this._widget) {
		return this._widget.getWorld().getSceneRectFromWorld(rect);
	} else {
		this.f.emit("error.SEVERE", "Requesting world rect with no widget present");
		return rect;
	}
};

/**
 * Convert a rectangle from scene to world
 * @param {ShapeUtils.Rect} rect
 * @returns
 */
AbstractTool.prototype.sceneToWorldRect = function(rect) {
	if (this._widget) {
		return this._widget.getWorld().getWorldRectFromScene(rect);
	} else {
		this.f.emit("error.SEVERE", "Requesting world rectangle with no widget present");
		return null;
	}
};

/**
 * Convert a size from scene to world
 * @param {ShapeUtils.Size} size
 * @returns
 */
AbstractTool.prototype.sceneToWorldSize = function(size) {
	if (this._widget) {
		return this._widget.getWorld().getWorldSizeFromScene(size);
	} else {
		this.f.emit("error.SEVERE", "Requesting world rectangle with no widget present");
		return null;
	}
};

/**
 * Get the world resolution
 * @returns Null or {Object}
 * {x: {Number} x-resolution,
 *  y: {Number} y-resolution }
 */
AbstractTool.prototype.getWorldResolution = function() {
	if (this._widget) {
		return this._widget.getWorld().getWorldResolution();
	} else {
		this.f.emit("error.SEVERE", "Requesting world resolution with no widget present");
		return null;
	}
};

/**
 * Converts a point to the closest point inside the world
 *
 * @param {ShapeUtils.Point} point - expects a worldpoint
 * @param {Boolean} returnScenePoint - is a worldPoint or a scenePoint returned
 * @returns {ShapeUtils.Point} a scenepoint/worldpoint
 */
AbstractTool.prototype.keepPointWithinWorldBounds = function(point, returnScenePoint) {
	var worldRect = this._widget.getWorld().getWorldBounds();
	return this.keepPointWithinRect(point, worldRect, returnScenePoint);
};

/**
 * Keeps a world rect inside the bounding box of the world
 *
 * @param {ShapeUtils.Rect} rect - expects a worldrect
 * @returns {ShapeUtils.Rect} a worldrect
 */
AbstractTool.prototype.keepRectWithinWorldBounds = function(rect) {
	var worldRect = this._widget.getWorld().getWorldBounds();
	return worldRect.calculateOverlap(rect);
};

/**
 * Converts a point to the closest point inside a rect (in world coordinates)
 *
 * @param {ShapeUtils.Point} point - expects a worldpoint
 * @param {ShapeUtils.Rect} rect - the rectangle to keep the point in
 * @param {Boolean} returnScenePoint - is a worldPoint or a scenePoint returned
 * @returns {ShapeUtils.Point} a scenepoint/worldpoint
 */
AbstractTool.prototype.keepPointWithinRect = function(point, rect, returnScenePoint) {
	var newPointX = 0;
	//check if x is outside of bounds
	if (point.x < rect.left() || point.x > rect.right()) {
		//if outside check is falling left or right world bounds
		newPointX = point.x < rect.left() ? rect.left() : rect.right();
	} else {
		//inside bounds
		newPointX = point.x;
	}
	var newPointY = 0;
	if (point.y < rect.top() || point.y > rect.bottom()) {
		//if outside check is falling above or beneath world bounds
		newPointY = point.y < rect.top() ? rect.top() : rect.bottom();
	} else {
		//inside bounds
		newPointY = point.y;
	}
	if (returnScenePoint) {
		//convert worldpoint to scenepoint
		return this._widget.getWorld().getScenePointFromWorld(new ShapeUtils.Point(newPointX, newPointY));
	} else {
		return new ShapeUtils.Point(newPointX, newPointY);
	}
};

/**
 * Keep a move size within the bounds of the rectangle
 * @param   {ShapeUtils.Rect} boudingRect - rectangle the move will be bounded to
 * @param   {ShapeUtils.Rect} movingRect - the rectangle that will be moved over a given size
 * @param   {ShapeUtils.Size}   movesize - the size the rectangle will be moved
 * @returns {ShapeUtils.Size}
 */
AbstractTool.prototype.keepMoveSizeWithinRect = function(boundingRect, movingRect, movesize) {
	var left = movingRect.left() + movesize.width;
	var right = movingRect.right() + movesize.width;
	var top = movingRect.top() + movesize.height;
	var bottom = movingRect.bottom() + movesize.height;

	var newMoveSize = new ShapeUtils.Size();
	if (
		Math.min(left, right) >= boundingRect.left() &&
		Math.max(left, right) <= boundingRect.right() &&
		Math.min(top, bottom) >= boundingRect.top() &&
		Math.max(top, bottom) <= boundingRect.bottom()
	) {
		newMoveSize.width = movesize.width;
		newMoveSize.height = movesize.height;
	}
	return newMoveSize;
};

/**
 * When the tool is active, and another tool requests activation, the ToolManager will first check if the current active tool can be interrupted
 *
 * If no interruption is allowed, return false
 * @param {AbstractTool} tool - The tool that wants to interrupt the current tool
 * @returns {Boolean}
 */
AbstractTool.prototype.canBeInterrupted = function(tool) {
	return true;
};

/**
 * A tool should always allow activation, but sometimes it has some conditions to be activated
 * 	e.g.:
 * 		- the annotation draw tool can only be activated if it follows the annotationmode logic
 * 		- licensing
 *
 * The tool can't be activated, return false
 * @returns {Boolean}
 */
AbstractTool.prototype.canBeActivated = function() {
	return true;
};

/**
 * Change the cursor
 * @param cursor
 */
AbstractTool.prototype.setCursor = function(cursor) {
	if (this._widget) {
		this._widget.setCursor(cursor);
	}
};

/**
 * Request the cursor to change
 */

AbstractTool.prototype.cursorShouldChange = function() {
	if (this.isActive()) {
		this.setCursor(this.getDefaultCursor());
	}
};

/**
 * Returns the default cursor for the tool
 * @returns {String}
 */
AbstractTool.prototype.getDefaultCursor = function() {
	return "cursor-default";
};

/**
 * When the tool is initialized, the passivate function is executed
 * 

 */
AbstractTool.prototype.passivate = function() {
	this._widget.whenContextAvailable(
		_.bind(function() {
			this.doPassivation();
			this._emit("passivated");
		}, this)
	);
};

/**
 * In this function the tool can register itself
 * (preferably on the context of the widget)
 *
 * if you use this.passiveEvents there's no need to override this method
 */
AbstractTool.prototype.doPassivation = function() {
	this._context = this._widget.context;
	this._context.on(this.passiveEvents);
	this.bindKeys(this.persistentKeyEvents);
	this.bindKeys(this.passiveKeyEvents);
};

/**
 * When the tool is depassivated, it should no longer listen to events and unregister itself
 *
 * if you use this.passiveEvents there's no need to override this method
 */
AbstractTool.prototype.depassivate = function() {
	if (this._context) {
		this._context.off(this.passiveEvents);
	}
	this.unbindKeys(this.passiveKeyEvents);

	this._emit("depassivated");
};

/**
 * Register for the needed events to be activated
 */
AbstractTool.prototype.doActivation = function() {
	this._context = this._widget.context;

	this._context.off(this.passiveEvents);
	this._context.on(this.activeEvents);
	this.unbindKeys(this.passiveKeyEvents);
	this.bindKeys(this.activeKeyEvents);
};

/**
 * Unregister all active events, and possible start passive events
 */
AbstractTool.prototype.doDeactivation = function() {
	if (this._context) {
		this._context.off(this.activeEvents);
		this._context.on(this.passiveEvents);
	}
	this.unbindKeys(this.activeKeyEvents);
	this.bindKeys(this.passiveKeyEvents);
};

/**
 * Request the current view based on the mouseposition
 * @param event
 * @returns
 */
AbstractTool.prototype.requestView = function(event) {
	return this._widget.requestView(event);
};

/**
 * Request a canvas layer to draw on
 * @param event
 */
AbstractTool.prototype.requestCanvasLayer = function(event) {
	this._viewForActiveCanvasLayer = this.requestView(event);
	if (this._viewForActiveCanvasLayer) {
		this._layer = this._viewForActiveCanvasLayer.addLayer(this.getIID());
	}
};

/**
 * release the requested canvaslayer
 */
AbstractTool.prototype.releaseCanvasLayer = function() {
	if (this._layer) {
		this._viewForActiveCanvasLayer.removeLayer(this.getIID());
		this._layer = undefined;
	}
};

/**
 * Returns the drawing context of the canvas for the tool
 * @returns
 */
AbstractTool.prototype.getCanvasCtx = function() {
	if (this._layer) {
		return this._layer.getCanvasContext();
	}
};

/**
 * Get the size of the Canvas
 * @returns {ShapeUtils.Size}
 */
AbstractTool.prototype.getCanvasSize = function() {
	if (this._layer) {
		return new ShapeUtils.Size(this._layer.getLayerElement()[0].width, this._layer.getLayerElement()[0].height);
	}
};

/**
 * Get the scenepoint given a (mouse) event. This method takes the icon offset into account as returned by getIconOffset (by default nothing)
 * @param {Object} e - Event object
 * @returns {ShapeUtils.Point}
 */
AbstractTool.prototype.getScenePoint = function(e) {
	var iconOffset = this.getIconOffset();

	var mousePoint;
	if (e.pageX !== undefined) {
		if (iconOffset) {
			mousePoint = new ShapeUtils.Point(e.pageX + iconOffset.width, e.pageY + iconOffset.height);
		} else {
			mousePoint = new ShapeUtils.Point(e.pageX, e.pageY);
		}
	} else if (e.clientX !== undefined) {
		if (iconOffset) {
			mousePoint = new ShapeUtils.Point(e.clientX + iconOffset.width, e.clientY + iconOffset.height);
		} else {
			mousePoint = new ShapeUtils.Point(e.clientX, e.clientY);
		}
	} else {
		mousePoint = new ShapeUtils.Point();
	}

	var view = this._widget.requestView(mousePoint);
	if (view) {
		var offset = view.getViewContext().offset();
		mousePoint.offsetXY(-offset.left, -offset.top);
		return mousePoint;
	} else {
		return new ShapeUtils.Point();
	}
};

/**
 * Get the icon offset of the tool
 * @returns {ShapeUtils.Size} Icon offset of the tool
 */
AbstractTool.prototype.getIconOffset = function() {
	// To be implemented
	return null;
};

/**
 * Get the worldpoint given a (mouse) event
 * @param e
 * @returns {ShapeUtils.Point}
 */
AbstractTool.prototype.getWorldPoint = function(e) {
	var point = this.sceneToWorld(this.getScenePoint(e));

	return point;
};

/**
 * Bind keys using MouseTrap
 * @param {Array} array - array of arrays [keycombo, function, keyup|keydown|keypress]
 */
AbstractTool.prototype.bindKeys = function(array) {
	for (var i = 0; i < array.length; i++) {
		Mousetrap.bind(array[i][0], array[i][1], array[i][2]);
	}
};

/**
 * Unbind keys
 * @param array
 */
AbstractTool.prototype.unbindKeys = function(array) {
	for (var i = 0; i < array.length; i++) {
		Mousetrap.unbind(array[i][0], array[i][1], array[i][2]);
	}
};

/**
 * Get the panel at a certain (mouse) event)
 * @param event
 * @returns {ViewportPanel}
 */
AbstractTool.prototype.getPanelAt = function(event) {
	var point;

	if (event.pageX) {
		point = new ShapeUtils.Point(event.pageX, event.pageY);
	} else {
		point = new ShapeUtils.Point(event.clientX, event.clientY);
	}
	return this._widget.getPanelAt(point);
};

/**
 * Initialize the tool for the given widget
 * @param widget {Layout.AbstractWidget} the widget on which the tool is activated
 * @param toolManager {ToolManager} the tool manager on which the tool is added
 */
AbstractTool.prototype.initialize = function(widget, toolManager) {
	this._widget = widget;
	this._toolManager = toolManager;
};

/**
 * In this method you can add shortcuts which will be displayed in ShortcutsPopup
 */
AbstractTool.prototype.getShortcuts = function() {
	return {};
};

export default AbstractTool;
