import PrivateChannel from "@esko/web-viewer/core/PrivateChannel";
import InterruptTool from "@esko/web-viewer/viewport/tools/InterruptTool";
import CoreConstants from "@esko/web-viewer/core/CoreConstants";
import LayoutConstants from "@esko/web-viewer/layout/LayoutConstants";
import _ from "underscore";
/**
 * @exports ToolManager
 * @memberof ViewportModule
 * @class The tool manager is responsible to manage the tools used on the viewport
 * It keeps track of the current tools and has a stack of previous tools (in case a tool deactivates itself, the previous one can be restored)
 * Communication is done via events.
 * @example
 * It reacts on the following events:
 * 	"toolManager.addTool": Add a tool to the tool manager.
 * 		,{AbstractTool} tool - Tool to add
 *
 *  "toolManager.request.activation": Request to activate the given tool. This request can be denied if the active tool is in a "Do not interrupt"-state.
 *   	,{String | AbstractTool} tool - Tool name or tool to activate
 *
 *  "toolManager.request.deactivation" - Request to deactivate the given tool or the current active tool if no tool specified.
 *   	,{String | AbstractTool} [tool] - Tool name or tool to deactivate.
 *
 *  "toolManager.activateTool": Activate the given tool without asking permission to the active tool
 *   	,{String | AbstractTool} tool - Tool name or tool to activate
 *
 *  "toolManager.deactivateTool": Deactivate the given tool or the current active tool if no tool specified.
 *   	,{String | AbstractTool} [tool] - Tool name or tool to deactivate.
 *
 *  "toolManager.activeTool": Get the current active tool.
 *   	,callback {AbstractTool} The current active tool or undefined if no tool is active.
 *
 *  "toolManager.requestTool": Request the tool with the given name.
 *   	,{String} tool - Tool name.
 *   	,callback {AbstractTool} - The tool with the given name or undefined if not found.
 *
 *  "toolManager.interruptTools":  Interrupt active tools.
 * 		,{Boolean} interrupt - True if you want to interrupt the active tools, false if you want to restore them.
 *
 * @param {Core.ViewerFacade} f - Viewer facade
 * @param {Layout.AbstractWidget} widget - The widget where the tools of this tool manager are activated on.
 */
var ToolManager = function(f, widget) {
	PrivateChannel.installTo(this, f, "toolManager");

	this.f = f;

	this._widget = widget;

	this._loadedTools = {};
	this._toolStack = [];

	this._allowedTools = null;
	this._disallowedTools = null;

	/**
	 * Create the interrupt tool and add it to the tool manager
	 */
	this._interruptTool = new InterruptTool(f);
	this._addTool(this._interruptTool);

	/**
	 * Add a tool to the tool manager.
	 * @param {AbstractTool} tool - Tool to add
	 */
	this.on(
		"addTool",
		_.bind(function(tool) {
			this._addTool(tool);
		}, this)
	);

	/**
	 * Remove a tool from the tool manager
	 * @param {String | AbstractTool} tool - IID of the tool
	 */
	this.on(
		"removeTool",
		_.bind(function(tool) {
			this._removeTool(tool);
		}, this)
	);

	/**
	 * Request to activate the given tool. This request can be denied if the active tool is in a "Do not interrupt"-state.
	 * @param {String | AbstractTool} tool - Tool IID or tool to activate
	 */
	this.on(
		"request.activation",
		_.bind(function(tool) {
			this._requestActivation(tool);
		}, this)
	);

	/**
	 * Request to deactivate the given tool or the current active tool if no tool specified.
	 * @param {String | AbstractTool} [tool] - Tool IID or tool to deactivate.
	 */
	this.on(
		"request.deactivation",
		_.bind(function(tool) {
			this._deactivateToolImpl(tool, false, true, false);
		}, this)
	);

	/**
	 * Deactivate the given tool or the current active tool if no tool specified.
	 * @param {String | AbstractTool} [tool] - Tool IID or tool to deactivate.
	 */
	this.on(
		"deactivateTool",
		_.bind(function(tool) {
			this._deactivateToolImpl(tool, false, true);
		}, this)
	);

	/**
	 * Get the current active tool
	 * @param {Function} callback - Callback executed with parameter {AbstractTool} tool - The current active tool or undefined if no tool is active.
	 */
	this.on(
		"activeTool",
		_.bind(function(callback) {
			callback(this.getActiveTool());
		}, this)
	);

	/**
	 * Get the shortcuts for all tools
	 */
	this.on(
		"shortcuts",
		_.bind(function(callback) {
			var shortcuts = _.reduce(
				this._loadedTools,
				function(memo, tool) {
					return _.extend(memo, tool.getShortcuts());
				},
				{}
			);
			callback(shortcuts);
		}, this)
	);

	/**
	 * Activate the given tool without asking permission to the active tool
	 * @param {String | AbstractTool} tool - Tool IID or tool to activate
	 * @param {Boolean} [keepHistory] - If true, the current active tool stays on the tool stack, otherwise it's popped (default)
	 */
	this.on(
		"activateTool",
		_.bind(function(tool, keepHistory) {
			if (keepHistory === undefined) {
				keepHistory = false;
			}

			this._activateToolImpl(tool, keepHistory, false);
		}, this)
	);

	/**
	 * Request the tool with the given IID.
	 * @param {String} tool - Tool IID.
	 * @param {Function} callback - Callback executed with parameter {AbstractTool} tool - The tool with the given IID or undefined if not found.
	 */
	this.on(
		"requestTool",
		_.bind(function(tool, callback) {
			callback(this._loadedTools[tool]);
		}, this)
	);

	/**
	 * Interrupt active tools.
	 * @param {Boolean} interrupt - True if you want to interrupt the active tools, false if you want to restore them.
	 */
	this.on(
		"interruptTools",
		_.bind(function(interrupt) {
			if (interrupt) {
				this._activateInterruptTool();
			} else {
				this._deactivateInterruptTool();
			}
		}, this)
	);

	this.on(
		"request.allowAllTools",
		_.bind(function() {
			this.allowAllToolsExcept(null);
		}, this)
	);

	this.on(
		"request.allowAllToolsExcept",
		_.bind(function(disallowedTools) {
			this.allowAllToolsExcept(disallowedTools);
		}, this)
	);

	this.on(
		"request.disallowAllToolsExcept",
		_.bind(function(allowedTools) {
			this.disallowAllToolsExcept(allowedTools);
		}, this)
	);
};

ToolManager.prototype._getDefaultTool = function() {
	try {
		return this.f
			.getLayoutManager()
			.getLayout()
			.getDefaultTool();
	} catch (ex) {
		return this.f.getLayoutType() === LayoutConstants.LayoutType.MOBILE ? CoreConstants.TOOLS.TOUCH : CoreConstants.TOOLS.ZOOM;
	}
};

/**
 * Add a tool to the toolmanager. The tool will be "passivated".
 * @param {AbstractTool} tool - Tool to add.
 */
ToolManager.prototype._addTool = function(tool) {
	if (tool) {
		this._loadedTools[tool.getIID()] = tool;
		tool.initialize(this._widget, this);

		tool.passivate();

		// if the tool is the default tool, activate it
		if (tool.getIID() == this._getDefaultTool() && this._isToolAllowed(this._getDefaultTool())) {
			this._activateTool(tool);
		}
	}
};

/**
 * Remove a tool from the tool manager
 * @param {String|AbstractTool} tool - IID of the tool
 */
ToolManager.prototype._removeTool = function(tool) {
	if (_.isString(tool)) {
		tool = this._loadedTools[tool];

		if (!tool) {
			this.f.emit("error.WARNING", "Trying to deactivate a tool which isn't loaded");
			return;
		}
	}

	if (tool.isActive()) {
		this._deactivateToolImpl(tool, false, true, false);
	}

	this._toolStack = _.without(this._toolStack, tool);

	tool.depassivate();
	delete this._loadedTools[tool.getIID()];
};

/**
 * Activate the given tool. It will not keep the previous tool on the stack.
 * @param {String|AbstractTool} tool - Tool IID or tool to activate.
 */
ToolManager.prototype._activateTool = function(tool) {
	this._activateToolImpl(tool, false, false);
};

/**
 * Activate a tool
 * @param {String | AbstractTool} tool - Tool IID or tool to activate
 * @param {Boolean} keepHistory - If true, the current active tool stays on the tool stack, otherwise it's popped (default)
 * @param {Boolean} skipDeactivation - If true, the deactivation of the previous tool is skipped.
 * @param {Boolean} interruptRequest - If true, the activation is requested by an interrupt event
 */
ToolManager.prototype._activateToolImpl = function(tool, keepHistory, skipDeactivation, interruptRequest) {
	if (!this._isToolAllowed(tool)) {
		this._emit("request.denied", tool);
		return;
	}

	if (_.isString(tool)) {
		tool = this._loadedTools[tool];

		if (!tool) {
			this.f.emit("error.WARNING", "Trying to activate a tool which isn't loaded");
			return;
		}
	}

	if (tool) {
		if (!tool.canBeActivated()) {
			this.f.debug.warn("Trying to activate tool which refuses to activate", tool);
			return;
		}

		var currentActive = this.getActiveTool();
		if (currentActive && !skipDeactivation) {
			if (currentActive == tool) {
				return; // the tool is already active!
			}

			// deactivate the current activated tool
			this._deactivateToolImpl(undefined, keepHistory, false, interruptRequest);
		}

		// make sure the last tool was not the same tool...
		if (_.last(this._toolStack) != tool) {
			this._toolStack.push(tool);
		}

		tool.activate(interruptRequest);

		if (this.getActiveTool() === tool) {
			this._widget.setCursor(tool.getDefaultCursor());
			this.f.emit("toolManager.activated", tool);
		}
	} else {
		this.f.emit("error.SEVERE", "Can't activate tool.", tool);
	}
};

/**
 * Deactivate a tool
 * @param {String|AbstractTool} tool - Tool IID or tool to deactivate.
 * @param {Boolean} keepHistory : If true, the current active tools stays on the stack (note do not call deactiveTool with stack = true from outside!)
 * @param {Boolean} restorePreviousTool - If true, the previous tools needs to be re-activated
 * @param {Boolean} interruptRequest - If true, the activation is requested by an interrupt event
 */
ToolManager.prototype._deactivateToolImpl = function(tool, keepHistory, restorePreviousTool, interruptRequest) {
	if (_.isString(tool)) {
		tool = this._loadedTools[tool];

		if (!tool) {
			this.f.emit("error.WARNING", "Trying to deactivate a tool which isn't loaded");
			return;
		}
	}

	if (!tool) {
		tool = this.getActiveTool();
	} else if (tool != this.getActiveTool()) {
		this.f.emit("error.WARNING", "Trying to deactive a tool which is not the current active tool", tool);
		return;
	}

	if (tool) {
		tool.deactivate(interruptRequest);

		if (!keepHistory) {
			this._toolStack.pop();
		}

		this.f.emit("toolManager.deactivated", tool);

		if (restorePreviousTool) {
			var toolToRestore = _.last(this._toolStack);

			while (toolToRestore && !this._isToolAllowed(toolToRestore)) {
				this._toolStack.pop();
				toolToRestore = _.last(this._toolStack);
			}
			// make sure we remove the tool from the stack before activating it again
			this._toolStack.pop();

			if (!toolToRestore) {
				// we are at the end of the toolstack, try the default tool
				if (this._isToolAllowed(this._getDefaultTool())) {
					toolToRestore = this._loadedTools[this._getDefaultTool()];
				}
			}

			if (toolToRestore) {
				// activate the tool
				this._activateToolImpl(toolToRestore, false, true);
			} else {
				this._widget.setCursor(undefined);
			}
		}
	}
};

/**
 * Activate interrupt tool
 */
ToolManager.prototype._activateInterruptTool = function() {
	this._activateToolImpl(this._interruptTool, true, false, true);
};

/**
 * Deactivate interrupt tool
 */
ToolManager.prototype._deactivateInterruptTool = function() {
	this._deactivateToolImpl(this._interruptTool, false, true, false);
};

/**
 * Get the current active tool.
 * @returns {AbstractTool} The current active tool or undefined if no tool is active.
 */
ToolManager.prototype.getActiveTool = function() {
	return _.last(this._toolStack);
};

/**
 * Request to activate the given tool. This request can be denied if the active tool is in a "Do not interrupt"-state.
 * @param {String | AbstractTool} tool - Tool IID or tool to activate
 */
ToolManager.prototype._requestActivation = function(tool) {
	// check if tool can be activated
	if (!this._isToolAllowed(tool)) {
		// the tool is not allowed
		this._emit("request.denied", tool);
		return;
	}
	//when we are requesting activation, first check if the current active tool can be interrupted
	var activeTool = this.getActiveTool();
	if (activeTool && activeTool != tool) {
		if (!activeTool.canBeInterrupted(tool)) {
			this._emit("request.denied", tool);
			return;
		}
	}
	this._activateToolImpl(tool, true, false, true);
};

/**
 * Check if a tool is allowed to be activated
 * @param {String | AbstractTool} tool - Tool IID or tool to activate
 * @returns {Boolean}
 */
ToolManager.prototype._isToolAllowed = function(tool) {
	if (!_.isString(tool)) {
		tool = tool.getIID();
	}

	// interrupt tool is always allowed
	if (tool === CoreConstants.TOOLS.INTERRUPT) {
		return true;
	}

	// by default allow the tool
	var allowed = true;

	if (this._allowedTools) {
		// only these tools are allowed
		allowed = _.contains(this._allowedTools, tool);
	}

	if (this._disallowedTools) {
		// all these tools are disallowed
		if (_.contains(this._disallowedTools, tool)) {
			// the tool is disallowed
			allowed = false;
		}
	}

	return allowed;
};

/**
 * Allow all tools, but disallow a specific list of tools
 * @param {Array.<String>} disallowedTools - the list of toolnames that should be disallowedZ
 */
ToolManager.prototype.allowAllToolsExcept = function(disallowedTools) {
	this._allowedTools = null;
	this._disallowedTools = disallowedTools;

	this._emit("allowedAllToolsExcept", disallowedTools);
	this._enforceToolAllowance();
};

/**
 *
 * @param allowedTools
 */
ToolManager.prototype.disallowAllToolsExcept = function(allowedTools) {
	this._allowedTools = allowedTools;
	this._disallowedTools = null;
	this._emit("disallowedAllToolsExcept", allowedTools);
	this._enforceToolAllowance();
};

/**
 * Internal function to run over all tools and all
 */
ToolManager.prototype._enforceToolAllowance = function() {
	// run over all loaded tools and enforce the allow/disallow
	var activeTool = this.getActiveTool();
	if (activeTool && !this._isToolAllowed(activeTool)) {
		this._deactivateToolImpl(activeTool, false, true, false);
	}

	// restore default tool if no tool is active
	var defaultTool = this._getDefaultTool();
	if (!activeTool && this._isToolLoaded(defaultTool) && this._isToolAllowed(defaultTool)) {
		this._activateTool(defaultTool);
	}
};

/**
 * Determines if a given tool is loaded and ready to be activated
 *
 * @param {String} toolIID
 * @returns True if the tool is loaded (added), false otherwise
 */
ToolManager.prototype._isToolLoaded = function(toolIID) {
	return this._loadedTools[toolIID] != null;
};

export default ToolManager;
