/**
 * @namespace WebviewerModule
 */
/**
 * @namespace WebviewerModule
 */
import _ from "underscore";

import AbstractModule from "@esko/web-viewer/core/AbstractModule";
import WebviewerWidgetModel from "@esko/web-viewer/webviewer/WebviewerWidgetModel";
import LayoutConstants from "@esko/web-viewer/layout/LayoutConstants";
/**
 * @exports WebviewerModule
 * @memberof WebviewerModule
 * @class
 * The WebviewerModule, this module installs the correct models for the viewport and load Graphics or Cad module
 * @constructor
 * @augments Core.AbstractModule
 * @param {Core.ViewerFacade} f
 * @param {Boolean} silenced
 */
var WebviewerModule = function(f, silenced) {
	AbstractModule.call(this, f, silenced, "webviewer");

	this.webviewerWidgetModel = new WebviewerWidgetModel(f);
	f.getModule("viewport").setWidgetModel(this.webviewerWidgetModel);

	var docController = this.f.getDocumentController();

	this._pageChanged(undefined, this.f.getActivePage());
	this._pageChanged(undefined, docController.getCadOverlay());

	// react when the active page has changed
	docController.whenActivePageChanged(this._pageChanged, this);
	docController.whenCadOverlayChanged(this._pageChanged, this);
	docController.whenDifferencePageChanged(this._pageChanged, this);
};

// inherit
WebviewerModule.prototype = Object.create(AbstractModule.prototype);
// correct the constructor pointer
WebviewerModule.prototype.constructor = WebviewerModule;

/**
 * The active page or CAD overlay has changed, load the needed modules, and create a new view
 *
 * @param {Data.AbstractDocumentPage} oldPage
 * @param {Data.AbstractDocumentPage} newPage
 */
WebviewerModule.prototype._pageChanged = function(oldPage, newPage) {
	var mainView = this.webviewerWidgetModel.getViewModel("Main");

	if (mainView === undefined || mainView === null) {
		this.webviewerWidgetModel.createView("Main", "", true);
	}

	if (newPage) {
		// create main view
		if (newPage.hasCad() || newPage.hasLinkedPage()) {
			this.f.requestModule("cad", undefined, this.silenced);
		}

		if (newPage.hasSvg()) {
			this.f.requestModule("svg", undefined, this.silenced);
		} else if (newPage.hasVideo() || newPage.isBlankPage()) {
			this.f.requestModule("video", undefined, this.silenced);
			this.launchInitialLayout(newPage);
		} else if (newPage.hasGraphics() || newPage.isBlankPage()) {
			this.f.requestModule("graphics", undefined, this.silenced);
		}

		// load 3D module if there is a 3D perspective or a 3d perspective can be generated
		if (newPage.has3D() || this.f.getProperty("show3D")) {
			this.f.requestModule("view3D", undefined, this.silenced);
			// Show 3D Model immediately if 'show3D' config is set to true
			this.launchInitialLayout(this.f.getActivePage());
		}

		newPage.whenParametersAvailable(
			function() {
				if ((newPage.hasGraphics() || newPage.isBlankPage() || newPage.hasSvg()) && newPage.getDocumentVersion().isMultiPage()) {
					this.f.requestModule("pagelist", undefined, this.silenced);
				}
			}.bind(this)
		);
	}
};

WebviewerModule.prototype.launchInitialLayout = _.once(function(newPage) {
	var startupLayoutMode;
	if (newPage.is3D()) {
		startupLayoutMode = LayoutConstants.LayoutMode.THREEDONLY;
	} else if (this.f.getProperty("show3D")) {
		startupLayoutMode = this.f.isMobile() ? LayoutConstants.LayoutMode.THREEDONLY : LayoutConstants.LayoutMode.DESKTOP_3D;
		if (!newPage.has3D()) {
			this.f.whenModuleAvailable("view3D", function(view3D) {
				view3D.show3D();
			});
		}
	} else if (newPage.hasVideo()) {
		startupLayoutMode = LayoutConstants.LayoutMode.VIDEO;
	}
	if (startupLayoutMode) {
		this.f.on(
			"layout.initialized",
			function() {
				this.f.emit("layout.switchLayoutMode", startupLayoutMode);
			}.bind(this)
		);
		this.f.emit("layout.switchLayoutMode", startupLayoutMode);
	}
});

/**
 *	Overrride the setSilenced method to also set the SVG Module as non silenced, indicating that it can publish its tools etc.
 * 	@override
 */
WebviewerModule.prototype.setSilenced = function(isSilenced) {
	AbstractModule.prototype.setSilenced.call(this, isSilenced);
	if (!isSilenced) {
		this._getManagedModules().forEach(
			function(mod) {
				this.f.whenModuleAvailable(mod, function(mod) {
					mod.setSilenced(false);
				});
			}.bind(this)
		);
	}
};

/**
 * Generates a list of modules that can be silenced by this module
 *
 * @returns {Array<String>}
 */
WebviewerModule.prototype._getManagedModules = function() {
	return ["svg", "cad", "graphics", "view3D", "pagelist", "video"];
};

export default WebviewerModule;
