import ViewportViewModel from "@esko/web-viewer/viewport/model/ViewportViewModel";
import _ from "underscore";
/**
 * @memberof WebviewerModule
 * @exports ViewserverViewModel
 * @class
 * The viewserver view model is responsible to manage the layers on a view
 * @constructor
 * @augments Viewport.ViewportViewModel
 * @param {Core.ViewerFacade} f - Facade
 * @param {String} ID
 * @param {String} title
 * @param {Boolean} isDefaultView
 */
var WebviewerViewModel = function(f, ID, title, isDefaultView) {
	ViewportViewModel.call(this, f, ID, title, isDefaultView);

	var layerOrder = [
		"Graphics",
		"Video",
		"SVGLayer",
		"SVGDifferenceLayer",
		"Cad",
		"currentDiffCad",
		"referenceDiffCad",
		"referenceCad",
		"CadOverlay",
		"Dyncontent",
		"Dyncontentpopups",
		"Annotations",
		"AnnotationPopups",
		"View3D"
	];

	var originalPositions = {};

	/**
	 * Returns the index for a layer
	 * TODO: improve this implementation
	 */
	this.indexForLayer = function(layername, layers) {
		//Use the intersection of the layers and the layerOrder, so the index of the layers can be returned
		var currentLayers = _.keys(layers);
		currentLayers.push(layername);
		var layerUnion = _.intersection(layerOrder, currentLayers);
		return _.indexOf(layerUnion, layername);
	};

	/**
	 * Sorts an array of layername strings based on the current order
	 *
	 * @param {Array<String>} layers
	 * @returns {Array<String>} the sorted layers
	 */
	this.sortLayers = function(layers) {
		return _.sortBy(layers, function(layerName) {
			return layerOrder.indexOf(layerName);
		});
	};

	/**
	 * swaps two layers if the index of the first layer is higher than the second
	 *
	 */
	this.swapLayers = function(layer, swap) {
		var index = layerOrder.indexOf(layer);
		var index2 = layerOrder.indexOf(swap);
		if (index > index2) {
			layerOrder[index2] = layer;
			layerOrder[index] = swap;
			this._emit("layerIndexChanged");
		}
	};
	this.on("layers.swapIfIndexIsHigher", _.bind(this.swapLayers, this));

	/**
	 * Moves a layer all the way to the back,
	 * send an event "layers.restoreLayer" to bring it back to its original position
	 *
	 * @param {String} layerName
	 */
	this.moveLayerToBack = function(layerName) {
		var index = layerOrder.indexOf(layerName);
		if (index !== -1 && !originalPositions[layerName]) {
			originalPositions[layerName] = index;
			layerOrder.splice(index, 1);
			layerOrder.unshift(layerName);
			this._emit("layerIndexChanged");
		}
	};
	this.on("layers.moveLayerToBack", this.moveLayerToBack.bind(this));

	/**
	 * Restores a layer to its original position when it has been temporarily moved to an other position
	 *
	 * @param {String} layerName
	 */
	this.restoreLayer = function(layerName) {
		var index = layerOrder.indexOf(layerName);
		if (index !== -1 && originalPositions[layerName] >= 0) {
			layerOrder.splice(index, 1);
			layerOrder.splice(originalPositions[layerName], 0, layerName);
			delete originalPositions[layerName];
			this._emit("layerIndexChanged");
		}
	};
	this.on("layers.restoreLayer", this.restoreLayer.bind(this));
};

// inherit
WebviewerViewModel.prototype = Object.create(ViewportViewModel.prototype);
// correct the constructor pointer
WebviewerViewModel.prototype.constructor = WebviewerViewModel;

export default WebviewerViewModel;
