import _ from "underscore";
import WebviewerViewModel from "@esko/web-viewer/webviewer/WebviewerViewModel";
import ViewportWidgetModel from "@esko/web-viewer/viewport/model/ViewportWidgetModel";
import ShapeUtils from "@esko/shapeutils";
import ViewerUtils from "@esko/web-viewer/utils/ViewerUtils";
import ConversionConstants from "@esko/web-viewer/utils/ConversionConstants";
/**
 * Create a new WebviewerWidgetModel
 * @memberof WebviewerModule
 * @exports WebviewerWidgetModel
 * @augments Viewport.ViewportWidgetModel
 * @class
 * The WebviewerWidget model is responsible for loading CAD or Graphics module and setting the correct properties for the viewport widget
 * @constructor
 * @param {Core.ViewerFacade} f
 */
var WebviewerWidgetModel = function(f) {
	ViewportWidgetModel.call(this, f, "webviewer");
	this.f = f;
};

// inherit
WebviewerWidgetModel.prototype = Object.create(ViewportWidgetModel.prototype);
// correct the constructor pointer
WebviewerWidgetModel.prototype.constructor = WebviewerWidgetModel;

/**
 * @override
 */
WebviewerWidgetModel.prototype.createModelForView = function(ID, title, isDefaultView) {
	return new WebviewerViewModel(this.f, ID, title, isDefaultView);
};

/**
 * @override
 */
WebviewerWidgetModel.prototype.getWorldBounds = function() {
	var activePage = this.f.getActivePage();
	var referencePage = this.f.getDocumentController().getDifferencePage();
	var cadOverlay = this.f.getDocumentController().getCadOverlay();

	var rect = function() {
		if (activePage !== undefined && referencePage !== undefined && activePage.hasCad() && referencePage.hasCad()) {
			if (!this.f.getModule("cad")) {
				// CAD Module isn ot yet loaded
				return undefined;
			}

			// compare with CAD documents
			var activeOrigin = activePage.getOrigin(true);
			var referenceOrigin = referencePage.getOrigin(true);

			var activePageSize = activePage.getPageSize();
			var referencePageSize = referencePage.getPageSize();

			if (activeOrigin !== null && referenceOrigin !== null) {
				if (referencePage.getCad() && referencePage.getCad().getLocation()) {
					referenceOrigin = referencePage.getCad().getLocation().origin;
					referencePageSize = referencePage.getCad().getLocation().size;
				} else {
					referenceOrigin = new ShapeUtils.Point(
						referenceOrigin.x - activeOrigin.x,
						activeOrigin.y + activePageSize.height - (referenceOrigin.y + referencePageSize.height)
					);
				}
				var currentRect = new ShapeUtils.Rect(null, activePageSize);
				var referenceRect = new ShapeUtils.Rect(referenceOrigin, referencePageSize);

				return currentRect.calculateArea(referenceRect);
			} else {
				return undefined;
			}
		} else {
			if (activePage && activePage.hasVideo()) {
				var pageSize = activePage.getPageSize().copy();
				return new ShapeUtils.Rect(new ShapeUtils.Point(), pageSize);
			} else if (activePage && activePage.is3D()) {
				return new ShapeUtils.Rect(new ShapeUtils.Point(), activePage.getPageSize().copy());
			}
			if (
				(activePage !== undefined && activePage.hasSvg() && activePage.isSVGLoaded && activePage.isSVGLoaded()) ||
				(referencePage && referencePage.hasSvg() && referencePage.isSVGLoaded())
			) {
				// SVG file
				var svgSize = activePage.getPageSize().copy();
				var diffPageSVGSize = referencePage && referencePage.getPageSize();
				if (activePage.isReaderSpread()) {
					if (diffPageSVGSize) {
						svgSize.width += diffPageSVGSize.width;
						svgSize.height = Math.max(diffPageSVGSize.height, svgSize.height);
					} else {
						svgSize.width *= 2;
					}
				} else if (referencePage && diffPageSVGSize) {
					svgSize.width = Math.max(svgSize.width, diffPageSVGSize.width);
					svgSize.height = Math.max(diffPageSVGSize.height, svgSize.height);
				}
				return new ShapeUtils.Rect(new ShapeUtils.Point(), svgSize);
			} else if (activePage && activePage.hasCad()) {
				// just 1 CAD file is loaded
				return new ShapeUtils.Rect(activePage.getOrigin(), activePage.getPageSize());
			} else if (activePage) {
				//Graphics (in compare and with or without cad overlay)
				// we could be in graphical compare (doesn't realy matter, viewserver gave us correct sizex and sizey)
				var graphicsWorldRect = new ShapeUtils.Rect(
					activePage.getOrigin(),
					new ShapeUtils.Size(ViewerUtils.parseFloat(activePage.getParam("sizex")), ViewerUtils.parseFloat(activePage.getParam("sizey")))
				);

				if (activePage.isBlankPage() && (!referencePage || referencePage.isBlankPage())) {
					var margins = activePage.getMargins();
					graphicsWorldRect.size.height += margins.top + margins.bottom;
					graphicsWorldRect.size.width += margins.left + margins.right;
					if (activePage.isReaderSpread()) {
						graphicsWorldRect.size.width *= 2;
					}
				}

				return graphicsWorldRect;
			}
		}
	}.bind(this)();

	if (cadOverlay && cadOverlay.getCad && cadOverlay.getCad()) {
		// there is a CAD overlay
		var cadRect;
		if (cadOverlay.getCad() && cadOverlay.getCad().getLocation()) {
			cadRect = cadOverlay.getCad().getLocation();
		} else {
			var cadOrigin = cadOverlay.getCad().getOrigin();
			var cadOffset = new ShapeUtils.Size(
				ViewerUtils.parseFloat(cadOverlay.get("offsetx")),
				ViewerUtils.parseFloat(cadOverlay.get("offsety"))
			);
			var cadSize = cadOverlay.getPageSize();
			cadRect = new ShapeUtils.Rect(
				new ShapeUtils.Point(-cadOffset.width + cadOrigin.x, cadOffset.height - (cadSize.height + cadOrigin.y)),
				cadSize
			);
		}
		return cadRect.calculateArea(rect);
	}
	return rect;
};

/**
 * @override
 */
WebviewerWidgetModel.prototype.getWorldResolution = function() {
	var worldResolution = {
		x: 0,
		y: 0
	};
	var activePage = this.f.getActivePage();
	if (activePage !== undefined && activePage.isParametersAvailable()) {
		var resolutionx = activePage.getParam("resolutionx");
		var resolutiony = activePage.getParam("resolutiony");
		if (resolutionx !== undefined) {
			worldResolution.x = ViewerUtils.parseFloat(resolutionx);
		} else {
			worldResolution.x = ConversionConstants.DEFAULT_POSTSCRIPT_PPI;
		}
		if (resolutiony !== undefined) {
			worldResolution.y = ViewerUtils.parseFloat(resolutiony);
		} else {
			worldResolution.y = ConversionConstants.DEFAULT_POSTSCRIPT_PPI;
		}
		return worldResolution;
	}
};

export default WebviewerWidgetModel;
