import XDropPanel from "@esko/xdrop/panel/XDropPanel";
import XDropDisplay from "@esko/xdrop/display/XDropDisplay";
import ToggleController from "@esko/xdrop/controllers/ToggleController";
import LabelRenderer from "@esko/xdrop/renderers/LabelRenderer";
import ErrorRenderer from "@esko/xdrop/renderers/ErrorRenderer";
import BlackListDataProvider from "@esko/xdrop/dataproviders/BlackListDataProvider";
import _ from "underscore";
import Observer from "@esko/observer";
import Observable from "@esko/observable";
/**
 * Xtendible Dropdown
 *
 * @param {XDropDataProvider} dataProvider
 * @param {Object}            renderFactories - map of result type to factory
 * @param {Object}             options         extra options
 * {
 *    absoluteXDropPanel: false (add results absolute in body if true)
 *    allowClear: false (add a button to clear)
 *
 * }
 * @param {Object}             paneloptions         extra options for the XDroppanel
 */
var XDrop = function(dataProvider, renderFactories, options, panelOptions) {
	Observer.mixin(this);
	Observable.mixin(this, "XDrop");

	this._options = _.extend(
		{
			allowClear: false,
			placeholder: "",
			blacklistSelectedEntry: false,
			isMobile: false
		},
		options
	);

	this._panelOptions = _.extend(
		{
			active: false,
			focusFilterOnActivate: !this._options.isMobile,
			multiSelect: false
		},
		panelOptions
	);

	/**
	 * Used to create renderers for resultentries
	 *  we can provide some default renderers
	 */
	renderFactories = _.extend(
		{
			label: LabelRenderer,
			error: ErrorRenderer
		},
		renderFactories
	);

	if (this._options.blacklistSelectedEntry) {
		dataProvider = new BlackListDataProvider(dataProvider);
		this.on("selected", dataProvider.addBlacklistedEntry.bind(dataProvider));
		this.on("deselected", dataProvider.removeBlacklistedEntry.bind(dataProvider));
	}

	this._init(dataProvider, renderFactories, this._options, this._panelOptions);

	this.getElement()[0].XDrop = this;

	["selected", "deselected", "changed"].forEach(
		function(event) {
			this._observe(this._panel, event, this._emit.bind(this, event));
		}.bind(this)
	);
};

/**
 * init behavior
 */
XDrop.prototype._init = function(dataProvider, renderFactories, options, panelOptions) {
	this._panel = new XDropPanel(dataProvider, renderFactories, panelOptions);
	this._display = new XDropDisplay(renderFactories, this._panel, options);
	this._toggleController = new ToggleController(this._display, this._panel, options);
};

/**
 * Get the main element of the XDrop
 * @returns {JQueryObj}
 */
XDrop.prototype.getElement = function() {
	return this._display.getElement();
};

/**
 * Is the XDrop opened?
 * @returns {Boolean}
 */
XDrop.prototype.isOpen = function() {
	return !this._toggleController || this._toggleController.isOpen();
};

/**
 * Update the placeholder of the XDrop and refresh the value to display the new placeholder if needed
 * @param {String} placeholder
 */
XDrop.prototype.setPlaceHolder = function(placeholder) {
	this._display.setPlaceHolder(placeholder);
};

/**
 * Get the value of the XDrop
 * @returns {ResultEntry}
 */
XDrop.prototype.getValue = function() {
	if (this._panel.getValue().length > 0) {
		return this._panel.getValue()[0];
	} else {
		return null;
	}
};

/**
 * Set the value of the XDrop
 * @param   {ResultEntry} resultEntry
 */
XDrop.prototype.setValue = function(resultEntry) {
	return this._panel.setValue(_.compact(_.flatten([resultEntry])));
};

/**
 * Clear the selection of the XDrop
 */
XDrop.prototype.clear = function() {
	this._panel.deselect();
};

/**
 * Enable or disable the XDrop
 *
 * @param disabled
 */
XDrop.prototype.toggleEnabled = function(enabled) {
	this._display.toggleEnabled(enabled);
};

/**
 * Make the XDrop read only. (not the same as "disabled" => see toggleEnabled(false))
 *
 * To enable it again, use "toggleEnabled(true)".
 *
 * Making the XDrop read only will make it appear as if there is no input at all... (no borders and such)
 *
 * @param disabled
 */
XDrop.prototype.makeReadOnly = function() {
	this.toggleEnabled(false);
	this._display
		.getElement()
		.removeClass("disabled")
		.addClass("readonly");
};

/**
 * Set the query of the dropdown panel
 * @param {String} query
 */
XDrop.prototype.setQuery = function(query) {
	this._panel.setQuery(query);
};

/**
 * Whether to show or hide the filter input
 *
 * @param {Boolean} show - true when we want to show the filter input
 */
XDrop.prototype.setShowFilterField = function(show) {
	this._panel.setShowFilterField(show);
};

/**
 * Get the DOM element of the results panel
 * (this can be used for finding/clicking results in SAHI
 * @returns {HTMLNode}
 */
XDrop.prototype.getPanelElement = function() {
	return this._panel.getElement()[0];
};

/**
 * Get the XDropPanel
 * @returns {XDropPanel}
 */
XDrop.prototype.getPanel = function() {
	return this._panel;
};

/**
 * Call this function when the XDrop component is moved to update its panel position
 * This is especially useful if absoluteXDropPanel = true
 */
XDrop.prototype.moved = function() {
	this._toggleController.moved();
};

/**
 * Call to destroy the XDrop
 */
XDrop.prototype.destroy = function() {
	this._toggleController && this._toggleController.destroy();
	this._display.destroy();
};

/**
 * Toggle
 */
XDrop.prototype.toggle = function(isOpen) {
	if (this._toggleController) {
		this._toggleController.toggle(isOpen);
	}
};

export default XDrop;
