import _ from "underscore";
import $ from "jquery";
import XDropRelativePanelWrapper from "@esko/xdrop/panel/XDropRelativePanelWrapper";
import XDropAbsolutePanelWrapper from "@esko/xdrop/panel/XDropAbsolutePanelWrapper";
import XDropModalPanelWrapper from "@esko/xdrop/panel/XDropModalPanelWrapper";
import Observer from "@esko/observer";
/**
 * Toggle Controller mixes in the toggle behavior between a xdrop display and an xdrop panel.
 * @param {XDropDisplay} xDropDisplay the display to add click handlers to
 * @param {XDropPanel} xDropPanel the xdrop panel to show on click
 * @param {Object} options - see constructor for all available options
 */
var ToggleController = function(xDropDisplay, xDropPanel, options) {
	Observer.mixin(this);

	this._options = _.extend(
		{
			absoluteXDropPanel: false,
			modalXDropPanel: false,
			preferredMinPanelHeight: 300,
			isMobile: false,
			showToggleButton: true
		},
		options
	);

	/**
	 * @member {XDropDisplay}
	 * @private
	 */
	this._display = xDropDisplay;

	/**
	 * @member {XDropPanel}
	 * @private
	 */
	this._panel = xDropPanel;

	/**
	 * @member {Boolean} is the dropdown panel open
	 */
	this._isOpen = false;

	/**
	 * Open/Close the xdrop panel if needed
	 */
	this._toggleButton = $("<a>", {
		class: "dropdown-toggle"
	}).append(
		$("<span>", {
			class: "caret"
		})
	);
	this._clickOverlay = $("<div>", {
		class: "click-overlay"
	});

	this._toggleButton.on(
		"click",
		function() {
			this.toggle(!this.isOpen());
		}.bind(this)
	);

	// initialize the correct wrapper
	var Wrapper = XDropRelativePanelWrapper;
	if (this._options.isMobile || this._options.modalXDropPanel) {
		Wrapper = XDropModalPanelWrapper;
	} else if (this._options.absoluteXDropPanel) {
		Wrapper = XDropAbsolutePanelWrapper;
	}
	this._panelWrapper = new Wrapper(this._panel, this._options, this.toggle.bind(this));

	this._panel.setVisible(false);

	this._observe(this._panel, "closeAndFocusNext", function() {
		this.toggle(false);

		var selector = ".form-control:visible,.XDrop:visible,.sdrop:visible,.note-editable";
		var tabables = $(selector).not(this._display.getElement().find(selector));
		var index = tabables.index(this._display.getElement());
		tabables.eq((index + 1) % tabables.length).trigger("focus");
	});

	this._observe(this._panel, "close", this.toggle.bind(this, false));
	this._observe(this._panel, "cleared", this.toggle.bind(this, false));
	this._observe(this._display, "disabled", this.toggle.bind(this, false));

	this._panelWrapper.addToDOM(this._display.getElement());
	this._init();
};

/**
 * Toggle the xdrop open/close
 * @param {Boolean} isOpen
 */
ToggleController.prototype.toggle = function(isOpen) {
	if (!this._display.isEnabled() && isOpen) {
		return;
	}

	if (this._isOpen !== isOpen) {
		this._isOpen = isOpen;

		this._display.getElement().toggleClass("open", isOpen);

		if (this._isOpen) {
			// When opening the xdrop, we want to set the zindex of the wrapper to the zindex of the display + 1 to ensure that the panel is always visible.
			var elem = this._display.getElement(),
				position,
				value;
			var zIndex = 0;
			while (elem.length && elem[0] !== document) {
				// Ignore z-index if position is set to a value where z-index is ignored by the browser
				// This makes behavior of this function consistent across browsers
				// WebKit always returns auto if the element is positioned
				position = elem.css("position");
				if (position === "absolute" || position === "relative" || position === "fixed") {
					// IE returns 0 when zIndex is not specified
					// other browsers return a string
					// we ignore the case of nested elements with an explicit value of 0
					// <div style="z-index: -10;"><div style="z-index: 0;"></div></div>
					value = parseInt(elem.css("zIndex"), 10);
					if (!isNaN(value) && value !== 0) {
						zIndex = value;
						break;
					}
				}
				elem = elem.parent();
			}

			this._panelWrapper.getElement().css("z-index", Math.max(zIndex + 1, 1000));
		}
	}
	this._panelWrapper.toggle(isOpen);
};

/**
 * Creates the element
 * @private
 */
ToggleController.prototype._init = function() {
	var element = this._display.getElement();
	element.on("focus", this.toggle.bind(this, true));
	this._clickOverlay.on("click", this.toggle.bind(this, true));
	if (this._options.showToggleButton) {
		element.append(this._toggleButton);
	}
	element.append(this._clickOverlay);
};

/**
 * Is the XDrop opened?
 * @returns {Boolean}
 */
ToggleController.prototype.isOpen = function() {
	return this._isOpen;
};

/**
 * Call to destroy the XDrop
 */
ToggleController.prototype.destroy = function() {
	this._panelWrapper.destroy();
	this._stopObserveAll();
};

/**
 * Call this function when the XDrop component is moved to update its panel position
 * This is especially useful if absoluteXDropPanel = true
 */
ToggleController.prototype.moved = function() {
	this._panelWrapper._updatePosition(this.isOpen());
};

/**
 * Get the toggle button
 * @return {JQueryObj}
 */
ToggleController.prototype.getToggleButton = function() {
	return this._toggleButton;
};

export default ToggleController;
