import CompositeDataProvider from "@esko/xdrop/dataproviders/CompositeDataProvider";
import ResultEntry from "@esko/xdrop/dataproviders/ResultEntry";
import _ from "underscore";
import i18n from "@esko/i18n/webi18n-loader?bundle=xdrop/locales!";
/**
 * A Category DataProvider shows all sub dataproviders in a collapsable category
 * The results are put in the order of how the dataproviders are defined
 * @class
 * @augments CompositeDataProvider
 * @mixes Observer
 * @param {Object} options
 * @param {XDropDataProvider...,Array.<XDropDataProvider>} dataproviders - list of data provider, this can be by providing a sequence of parameters, or arrays with datproviders
 */
var CategoryDataProvider = function(options, dataproviders) {
	CompositeDataProvider.apply(this, _.rest(arguments));

	this._options = _.extend(
		{
			allowCollapse: true,
			collapsedIfNoQuery: false,
			expandWhenQueryIsSet: false,
			isCollapsed: false,
			hideIfNoResults: true,
			noMatchesLabel: i18n.t("NoMatches"),
			title: "Category",
			className: "XDrop-category",
			allowFilteringOnTitle: false,
			indicateResultCount: false
		},
		options
	);

	var self = this;
	this._categoryHeaderEntry = new ResultEntry("categoryTitle", {
		getTitle: function() {
			return self._options.title;
		},
		allowCollapse: function() {
			return self._options.allowCollapse;
		},
		isCollapsed: function() {
			return self.isCollapsed();
		},
		getClassName: function() {
			return self._options.className;
		},
		toggle: function() {
			if (self._options.allowCollapse) {
				self.setCollapsed(!self._options.isCollapsed);
				self._options.expandWhenQueryIsSet = false;
				self._updateResultSet();
			}
		},
		allowFilteringOnTitle: function() {
			return self._options.allowFilteringOnTitle;
		},
		showCount: function() {
			return self._options.indicateResultCount;
		},
		getCount: function() {
			var size = 0;
			_.each(self._providers, function(provider) {
				size += provider.provider.getResultSet().getSize();
			});
			return size;
		},
		getBadgeColor: function() {
			return self._options.badgeColorClass;
		}
	});

	this._categoryHeaderEntry.setDefaultSelectionIndex(0); // we really don't want this to be selected as default
};

CategoryDataProvider.prototype = Object.create(CompositeDataProvider.prototype);
CategoryDataProvider.prototype.constructor = CategoryDataProvider;

/**
 * @Override
 */
CategoryDataProvider.prototype.setQuery = function(query) {
	this._query = query;

	if (query.length > 0 && this._options.expandWhenQueryIsSet) {
		this._options.expandWhenQueryIsSet = false;
		this.setCollapsed(false);
	} else {
		this._pushMaxResults();
	}

	CompositeDataProvider.prototype.setQuery.apply(this, arguments);
};

/**
 * @Override
 */
CategoryDataProvider.prototype.setMaxResults = function(maxResults) {
	this._maxResults = maxResults;
	this._pushMaxResults();
};

/**
 * Push the correct maxResults value to the (sub)dataproviders
 * if collapsed: 0, else the max results number
 * @private
 */
CategoryDataProvider.prototype._pushMaxResults = function() {
	var maxResults = this._maxResults;
	if (this.isCollapsed()) {
		maxResults = 0;
	}

	this._callMethodOnAllProviders("setMaxResults", maxResults);
};

/**
 * Collapse/expand the category
 * @param {Boolean} isCollapsed
 */
CategoryDataProvider.prototype.setCollapsed = function(isCollapsed) {
	if (this._options.isCollapsed !== isCollapsed) {
		this._options.isCollapsed = isCollapsed;
		this._pushMaxResults();
		this._updateResultSet();

		this._emit("change." + (isCollapsed ? "collapse" : "expand"));
	}
};

/**
 * Is the category collapsed?
 * @returns {Boolean} [[Description]]
 */
CategoryDataProvider.prototype.isCollapsed = function() {
	return this._options.allowCollapse && (this._options.isCollapsed || (this._options.collapsedIfNoQuery && !this._query));
};

/**
 * Merge the resultsets of the sub dataproviders
 * @protected
 */
CategoryDataProvider.prototype._updateResultSetImpl = function() {
	this._currentResultList = this._generateResultList();

	var list = [];

	if (this._currentResultList.length > 0 || this.isCollapsed() || this.isLoading() || !this._options.hideIfNoResults) {
		if (!this.isCollapsed()) {
			list = this._currentResultList.slice(0);

			if (list.length === 0) {
				var label = this._options.noMatchesLabel;
				if (this.isLoading()) {
					label = i18n.t("Searching");
				}
				// show no results indication
				var nomatches = new ResultEntry("label", label);
				nomatches.setAllowSelection(false);

				list.push(nomatches);
			}
		}

		list.unshift(this._categoryHeaderEntry);
	}

	this._resultSet.setEntries(list);
	this._resultSet.commitChanges();
};

export default CategoryDataProvider;
