import XDropDataProvider from "@esko/xdrop/dataproviders/XDropDataProvider";
import Observer from "@esko/observer";
import _ from "underscore";
/**
 * A Composite DataProvider merges the results from all sub dataproviders
 * The results are put in the order of how the dataproviders are defined
 * @class
 * @augments XDropDataProvider
 * @mixes Observer
 * @param {XDropDataProvider...,Array.<XDropDataProvider>} dataproviders - list of data provider, this can be by providing a sequence of parameters, or arrays with datproviders
 */
var CompositeDataProvider = function(dataproviders) {
	XDropDataProvider.call(this);
	Observer.mixin(this);

	/**
	 * @member {Array.<Object>} for every subprovider a POJO is added with two keys: name (unique id in the composite provider), provider: reference to the subprovider
	 */
	this._providers = [];
	this._query = null;

	this._providerIdx = 0;

	this._updateResultSet = _.debounce(this._updateResultSetImpl, 20);

	this.setDataProviders.apply(this, arguments);
};

CompositeDataProvider.prototype = Object.create(XDropDataProvider.prototype);
CompositeDataProvider.prototype.constructor = CompositeDataProvider;

/**
 * @Override
 */
CompositeDataProvider.prototype.setQuery = function(text) {
	this._query = text;

	this._callMethodOnAllProviders("setQuery", text);
};

/**
 * @Override
 */
CompositeDataProvider.prototype.setMaxResults = function(maxResults) {
	this._maxResults = maxResults;

	this._callMethodOnAllProviders("setMaxResults", maxResults);
};

/**
 * Push down active state to composite providers
 */
CompositeDataProvider.prototype.setActive = function(isActive) {
	XDropDataProvider.prototype.setActive.call(this, isActive);

	this._callMethodOnAllProviders("setActive", isActive);
};

/**
 * Utility method to call a method on all registered dataproviders. You can pass the arguments for the method call after the method name
 * @private
 * @param {String} method - method name
 */
CompositeDataProvider.prototype._callMethodOnAllProviders = function(method) {
	var parameters = _.rest(arguments);

	_.each(
		this._providers,
		function(providerEntry) {
			providerEntry.provider[method].apply(providerEntry.provider, parameters);
		},
		this
	);
};
/**
 * Set the list of dataproviders
 * @param {XDropDataProvider...,Array.<XDropDataProvider>} dataproviders - list of data provider, this can be by providing a sequence of parameters, or arrays with datproviders
 */
CompositeDataProvider.prototype.setDataProviders = function() {
	var providers = _.flatten(arguments, true);

	var previousProviders = this._providers;
	this._providers = [];
	_.each(
		providers,
		function(provider) {
			var existingProviderEntry = _.findWhere(previousProviders, {
				provider: provider
			});

			if (existingProviderEntry) {
				// the provider was already being used
				previousProviders = _.without(previousProviders, existingProviderEntry);
				this._providers.push(existingProviderEntry);
			} else {
				// new provider
				this.addDataProvider(provider);
			}
		},
		this
	);

	_.each(previousProviders, this._cleanProviderEntry, this);

	this._updateResultSet();
};

/**
 * Add a single dataprovider
 * (it is possible to add a dataprovider twice, we have no protection against this)
 * @param {XDropDataProvider} provider
 */
CompositeDataProvider.prototype.addDataProvider = function(provider) {
	var providerEntry = {
		name: "provider_" + this._providerIdx++,
		provider: provider
	};

	this._observe(provider, "loading", function(isLoading) {
		if (isLoading) {
			this._startLoading(name);
		} else {
			this._stopLoading(name);
		}
	});

	this._observe(provider.getResultSet(), "changed", this._updateResultSet);

	this._providers.push(providerEntry);

	if (this._query) {
		provider.setQuery(this._query);
	}

	if (this._maxResults > -1) {
		provider.setMaxResults(this._maxResults);
	}

	provider.setActive(this._active);

	this._updateResultSet();
};

/**
 * Remove a dataprovide
 * @param {XDropDataProvider} provider
 */
CompositeDataProvider.prototype.removeDataProvider = function(provider) {
	var existingProviderEntry = _.findWhere(this._providers, {
		provider: provider
	});
	if (existingProviderEntry) {
		this._providers = _.without(this._providers, existingProviderEntry);
		this._cleanProviderEntry(existingProviderEntry);

		this._updateResultSet();
	}
};

/**
 * Internal method to clean up and stop observing a dataprovider
 * @private
 * @param {object} providerEntry
 */
CompositeDataProvider.prototype._cleanProviderEntry = function(providerEntry) {
	this._stopObserveObject(providerEntry.provider);
	this._stopObserveObject(providerEntry.provider.getResultSet());
	this._stopLoading(providerEntry.name);
};

/**
 * Compose the merged results of the sub dataprovider
 * @private
 * @returns {Array.<ResultEntry>}
 */
CompositeDataProvider.prototype._generateResultList = function() {
	return _.flatten(
		_.map(this._providers, function(providerEntry) {
			return providerEntry.provider.getResultSet().getEntriesImmutable();
		}),
		true
	);
};

/**
 * Merge the resultsets of the sub dataproviders
 * @protected
 */
CompositeDataProvider.prototype._updateResultSetImpl = function() {
	var results = this._generateResultList();
	this._resultSet.setEntries(results);
	this._resultSet.commitChanges();
};

export default CompositeDataProvider;
