import CompositeDataProvider from "@esko/xdrop/dataproviders/CompositeDataProvider";
import ResultEntry from "@esko/xdrop/dataproviders/ResultEntry";
import _ from "underscore";
import i18n from "@esko/i18n/webi18n-loader?bundle=xdrop/locales!";
/**
 * A LimitResultsDataProvider merges the results from all sub dataproviders and limits the number of results
 * If there are more results, a button is shown to show more results
 * @class
 * @augments CompositeDataProvider
 * @mixes Observer
 * @param {Object} options
 * @param {XDropDataProvider...,Array.<XDropDataProvider>} dataproviders - list of data provider, this can be by providing a sequence of parameters, or arrays with datproviders
 */
var LimitResultsDataProvider = function(options, dataproviders) {
	CompositeDataProvider.call(this);

	// first set options
	this._options = _.extend(
		{
			maxResults: 50,
			extraEntriesShownOnClick: 20,
			showMoreLabel: i18n.t("LoadMoreResults")
		},
		options
	);
	this.setMaxResults(this._options.maxResults);

	// add data providers
	this.setDataProviders.apply(this, _.rest(arguments));

	var self = this;
	this._lasKnownSize = this._maxResults;
	this._showMoreEntry = new ResultEntry("button", {
		getLabel: function() {
			return self._options.showMoreLabel;
		},
		getClassNames: function() {
			return "btn btn-showmore";
		},
		doAction: function(event, api) {
			self._lasKnownSize = self._maxResults + self._options.extraEntriesShownOnClick;
			self.setMaxResults(self._maxResults + self._options.extraEntriesShownOnClick);
			self._updateResultSet();
		}
	});

	this._showMoreEntry.setKeepSelected(false);
	this._showMoreEntry.setDefaultSelectionIndex(70); // we don't want this to be selected as default
};

LimitResultsDataProvider.prototype = Object.create(CompositeDataProvider.prototype);
LimitResultsDataProvider.prototype.constructor = LimitResultsDataProvider;

/**
 * @override reset the limit when query changes
 * @param {String} query
 */
LimitResultsDataProvider.prototype.setQuery = function(query) {
	this.setMaxResults(this._options.maxResults);

	CompositeDataProvider.prototype.setQuery.apply(this, arguments);
};

/**
 * @override
 */
LimitResultsDataProvider.prototype.setMaxResults = function(maxResults) {
	if (maxResults === -1) {
		// when reset to unlimited, pick the last known
		maxResults = this._lasKnownSize;
	}
	CompositeDataProvider.prototype.setMaxResults.call(this, maxResults);
};

/**
 * Merge the resultsets of the sub dataproviders
 * @protected
 */
LimitResultsDataProvider.prototype._updateResultSetImpl = function() {
	this._currentResultList = this._generateResultList();
	if (this._maxResults === -1) {
		this._currentResultList.push(this._showMoreEntry);
	} else if (this._currentResultList.length >= this._maxResults) {
		var itemsToShow = this._maxResults - 1;
		var itemsToRemove = this._currentResultList.length - itemsToShow;
		this._currentResultList.splice(itemsToShow, itemsToRemove, this._showMoreEntry);
	}

	this._resultSet.setEntries(this._currentResultList);
	this._resultSet.commitChanges();
};

export default LimitResultsDataProvider;
