import XDropDataProvider from "@esko/xdrop/dataproviders/XDropDataProvider";
import ResultEntry from "@esko/xdrop/dataproviders/ResultEntry";
import _ from "underscore";
import i18n from "@esko/i18n/webi18n-loader?bundle=xdrop/locales!";
/**
 * A PagedDataProvider allows displaying paged data in a dataprovider
 * If there are more results, a button is shown to show more results
 *
 * @author PJPR
 * @class
 * @mixes Observer
 * @augments XDropDataProvider
 * @param {Object} options
 */
var PagedDataProvider = function(options) {
	XDropDataProvider.apply(this);
	this._options = _.extend(
		{
			showMoreLabel: i18n.t("LoadMoreResults"),
			loadingLabel: i18n.t("Searching"),
			debounce: 250
		},
		options
	);

	this._hasMore = false;
	this._autoload = false;
	this._currentPage = -1;
	this._operationIdx = 0;
	this._ongoingOperation = Promise.resolve();
	this._pagesToLoad = 1;

	this._data = [];
	this._filtered = null;
	this._query = null;
	this._debouncedQuery = "";

	if (this._options.debounce) {
		this._debouncedSetQuery = _.debounce(this._setQuery.bind(this), this._options.debounce);
	} else {
		this._debouncedSetQuery = this._setQuery;
	}

	var self = this;

	this._showMoreEntry = new ResultEntry("button", {
		getLabel: function() {
			return self.isLoading() ? self._options.loadingLabel : self._options.showMoreLabel;
		},
		getClassNames: function() {
			return "btn btn-showmore";
		},
		doAction: function(event, api) {
			if (!self.isLoading()) {
				api.focusResultEntry(self._showMoreEntry);
				self._loadMore(event, api);
			}
		}
	});

	this._showMoreEntry.setKeepSelected(false);

	this._showMoreEntry.setDefaultSelectionIndex(70); // we don't want this to be selected as default

	this._setQuery(""); // start loading data
};

PagedDataProvider.prototype = Object.create(XDropDataProvider.prototype);
PagedDataProvider.prototype.constructor = PagedDataProvider;

/**
 * Implement the page loading logic in this method
 * The current query is available as this._query
 *
 * The returned promise should resolve to an object containing the resultEntries for the requested page and a boolean indicating whether or not there are more pages available
 *
 * Override this method!
 *
 * @private
 * @param   {Number} pageIdx - the page that is to be loaded
 * @param 	{Number} pagesToLoad - the number of pages that need to be loaded (defaults to one)
 * @returns {Promise.<Object>} Resolved to result object containing data (Array.<ResultEntry>) and hasMore property)
 */
PagedDataProvider.prototype._loadPage = function(pageIdx, pagesToLoad) {
	return Promise.resolve({
		data: [],
		hasMore: false
	});
};

/**
 * Reload all data in the dataprovider
 * @param {Boolean} [keepPages] - if true, the dataprovider will try to reload an equal amount of pages when reloading
 * @returns {Promise} resolved once complete
 */
PagedDataProvider.prototype.reload = function(keepPages) {
	this._cancelOngoingRequests();

	if (keepPages) {
		this._pagesToLoad = Math.max(1, this._currentPage + 1, this._pagesToLoad);
	} else {
		this._pagesToLoad = 1;
	}

	this._currentPage = -1;
	this._instantFilter();

	return this._loadMore();
};

/**
 * @Overide
 * @param {Number} maxResults
 */
PagedDataProvider.prototype.setMaxResults = function(maxResults) {
	this._maxResults = maxResults;
};

/**
 * @override reset the limit when query changes
 * @param {String} query
 */
PagedDataProvider.prototype.setQuery = function(query) {
	this._debouncedQuery = query;
	if (this._debouncedQuery != query) {
		this._instantFilter();
	}
	this._debouncedSetQuery();
};

/**
 * After debounce, set the query
 * @private
 */
PagedDataProvider.prototype._setQuery = function() {
	if (this._query !== this._debouncedQuery) {
		// query was changed
		this._query = this._debouncedQuery;
		this.reload();
	}
};

/**
 * Apply query to current data
 * @private
 */
PagedDataProvider.prototype._instantFilter = function() {
	this._filtered = this._filterCurrentDataSet();
	this._updateResultSet();
};

/**
 * Check if the given operation is not cancelled yet
 * @private
 * @param   {Number}  operationIdx
 * @returns {Boolean}
 */
PagedDataProvider.prototype._isValidOperation = function(operationIdx) {
	return this._operationIdx === operationIdx;
};

/**
 * Cancel ongoing requests
 * You could override this method to really cancel your ajax requests
 * @private
 */
PagedDataProvider.prototype._cancelOngoingRequests = function() {
	this._operationIdx++;
	this._ongoingOperation = Promise.resolve();
};

/**
 * if you want live feedback, override this method and filter the current data set
 * in this way the filtered results of the currently available data will be shown instantly
 * @private
 */
PagedDataProvider.prototype._filterCurrentDataSet = function() {
	// by default we just clear the data
	return [];
};

/**
 * Load more information (no need to override this method)
 * @private
 */
PagedDataProvider.prototype._loadMore = function() {
	var operationIdx = -1;
	var pagesToLoad = this._pagesToLoad;

	this._ongoingOperation = this._ongoingOperation
		.then(
			function() {
				this._operationIdx++;
				operationIdx = this._operationIdx;
				this._startLoading("loadMore" + operationIdx);
				this._updateResultSet();
				return this._loadPage(this._currentPage + 1, pagesToLoad);
			}.bind(this)
		)
		.then(
			function(result) {
				if (this._isValidOperation(operationIdx)) {
					if (this._currentPage === -1) {
						// clear previous data
						this._data = [];
					}
					this._filtered = null;
					this._data = this._data.concat(result.data);
					this._hasMore = result.hasMore;
					this._currentPage += pagesToLoad;
					this._pagesToLoad = 1;
					this._updateResultSet();
				}
			}.bind(this)
		)
		.catch(function(e) {
			if (console && console.log) {
				console.log("XDropPagedDataProvider: Error while fetching more items", e);
			}
		})
		.then(
			function() {
				this._stopLoading("loadMore" + operationIdx);
				this._updateResultSet();
			}.bind(this)
		);

	return this._ongoingOperation;
};

/**
 * Pick the correct dataset and update the resutlset
 * @private
 */
PagedDataProvider.prototype._updateResultSet = function() {
	this._resultSet.setEntries(this._buildEntryList());
	this._resultSet.commitChanges();
};

/**
 * This can be overriden to alter the resultset with extra data if needed
 * @protected
 * @returns {List.<ResultEntry>}
 */
PagedDataProvider.prototype._buildEntryList = function() {
	var data = this._filtered !== null ? this._filtered : this._data;
	var results = data.slice(0);

	if (this._hasMore) {
		results.push(this._showMoreEntry);
	}

	return results;
};

/**
 * React to onScroll event from xdroppanel (used for autoloadonscroll)
 * @private
 * @param {ResultEntry} top    - entry displayed in top of panel
 * @param {ResultEntry} bottom - entry displayed in bottom of panel
 */
PagedDataProvider.prototype._onScroll = function(top, bottom) {
	var entries = this._resultSet.getEntriesImmutable();
	if (this._hasMore && !this.isLoading()) {
		var bottomIndex = entries.indexOf(bottom);
		if (bottomIndex >= 0 && entries.length - bottomIndex < 3) {
			this._loadMore();
		}
	}
};

/**
 * Enable autoload on scroll. When scrolling the XDropPanel, the dataprovider will automatically fetch extra information if needed
 * @param {XDropPanel} xdropPanel
 */
PagedDataProvider.prototype.enableAutoLoadOnScroll = function(xdropPanel) {
	this._autoload = true;
	xdropPanel.getRenderAPI().onScroll(_.throttle(this._onScroll.bind(this), 40, { leading: false }));
};

/**
 * Has more results to load
 * @returns {boolean}
 */
PagedDataProvider.prototype.hasMore = function() {
	return this._hasMore;
};

export default PagedDataProvider;
