import _ from "underscore";
var GUIDCounter = 0;

/**
 * A ResultEntry is a simple container that can contain a data object
 * @class
 * @param {String} type - type of the result entry
 * @param {Object} data - data value of the result entry
 */
var ResultEntry = function(type, data) {
	this._id = "R" + GUIDCounter++;

	this._type = type;
	this._data = data;

	this._allowSelection = true;

	this._keepSelected = true;

	this._triggers = {};

	/**
	 * the higher the value, the higher the priority to select as default selected element
	 * It is best to to make this higher than the default 100
	 */
	this._defaultSelectionIndex = 100;
};

/**
 * Every Result Entry has a system wide unique ID
 * @returns {String}
 */
ResultEntry.prototype.getUniqueId = function() {
	return this._id;
};

/**
 * Get the Type of the entry
 * @returns {String}
 */
ResultEntry.prototype.getType = function() {
	return this._type;
};

/**
 * Get the data object
 * @returns {Object}
 */
ResultEntry.prototype.getData = function() {
	return this._data;
};

/**
 * Can the result entry be selected/focused?
 * @returns {Boolean}
 */
ResultEntry.prototype.allowSelection = function() {
	return this._allowSelection;
};

/**
 * allow/disallow selection/focus of the result
 * @param {Boolean} allowSelection
 */
ResultEntry.prototype.setAllowSelection = function(allowSelection) {
	this._allowSelection = allowSelection;
};

/**
 * Change the behavior when the item is selected and the resultset updates
 * @param {Boolean} keepSelected
 */
ResultEntry.prototype.setKeepSelected = function(keepSelected) {
	this._keepSelected = keepSelected;
};

/**
 * If the item is selected when the resultset updated, should the selection be kept?
 * @returns {Boolean} [[Description]]
 */
ResultEntry.prototype.keepSelected = function() {
	return this._keepSelected;
};

/**
 * Get the default Selection index.
 * The higher the value, the higher the priority to select as default selected element
 * @returns {Number}
 */
ResultEntry.prototype.getDefaultSelectionIndex = function() {
	return this._defaultSelectionIndex;
};

/**
 * Change the default selection behavior of the result entry
 * It is best to to make this higher than the default 100
 * @param {Number} defaultSelectionIndex
 */
ResultEntry.prototype.setDefaultSelectionIndex = function(defaultSelectionIndex) {
	this._defaultSelectionIndex = defaultSelectionIndex;
};

/**
 * Trigger a data event on the result entry
 * (will be called by renderer)
 * @param {String} event - which event needs to be triggered
 * @param {Any...} [args] - optional arguments for the event
 *
 * @returns Array.<Any> an array with all results of the listeners
 */
ResultEntry.prototype.trigger = function(event, args) {
	return _.map(
		this._triggers[event],
		function(fn) {
			return fn.apply(this, args);
		},
		this
	);
};

/**
 * Register listener for resultentry event
 * @param {String} event - name of the event
 * @param {Function} fn - callback function (can return a value which will be passed when triggered)
 */
ResultEntry.prototype.on = function(event, fn) {
	if (!this._triggers[event]) {
		this._triggers[event] = [fn];
	} else {
		this._triggers[event].push(fn);
	}
};

/**
 * Are two result entry equal?
 * Implement an equals method in the data element. If not implemented _.isEqual will be used
 * @param {ResultEntry} other
 */
ResultEntry.prototype.equals = function(other) {
	if (other) {
		if (other.getType() === this.getType()) {
			if (other === this || other.getData() === this.getData()) {
				return true;
			} else if (_.isFunction(this.getData().equals)) {
				return this.getData().equals(other.getData());
			} else {
				return _.isEqual(this.getData(), other.getData());
			}
		}
	}

	return false;
};

export default ResultEntry;
