import _ from "underscore";
import Observable from "@esko/observable";
/**
 * A resultset is an ordered list of results
 * @class
 * @mixes Observable
 */
var ResultSet = function() {
	Observable.mixin(this, "ResultSet");

	this._pendingCommit = null;

	this._list = [];
};

/**
 * Clear the resultset
 */
ResultSet.prototype.clear = function() {
	this._list = [];
	this._deferredCommit();
};

/**
 * Get all the entries in the result set
 * @returns {Array.<ResultEntry>}
 */
ResultSet.prototype.getEntries = function() {
	return this._list.slice(0);
};

/**
 * Get the list of entries.
 * !!!WARNING!!! this is not a copy, the entrylist should never be modified!!
 * If you want a list which you can directly modify, use getEntries()
 * @returns {Array.<ResultEntry>}
 */
ResultSet.prototype.getEntriesImmutable = function() {
	return this._list;
};

/**
 * Get the entry at index "pos"
 * @param   {Number}      pos
 * @returns {ResultEntry}
 */
ResultSet.prototype.getEntry = function(pos) {
	return this._list[pos];
};

/**
 * Replace the result set with a new List
 * @param {Array.<ResultEntry>} list [[Description]]
 */
ResultSet.prototype.setEntries = function(list) {
	this._list = list.slice(0);
	this._deferredCommit();
};

/**
 * Append an item, or multiple items to the list
 * @param {ResultEntry} entry...
 */
ResultSet.prototype.append = function() {
	this._list.push.apply(this._list, arguments);
	this._deferredCommit();
};

/**
 * Append an array of results
 * @param {Array.<ResultEntry>} arr
 */
ResultSet.prototype.appendAll = function(arr) {
	this.append.apply(this, arr);
};

/**
 * Prepend an item to the list
 * @param {ResultEntry} result
 */
ResultSet.prototype.prepend = function(result) {
	this._list.unshift(result);
	this._deferredCommit();
};

/**
 * Get the size of the result set
 * @returns {Number}
 */
ResultSet.prototype.getSize = function() {
	return this._list.length;
};

/**
 * Get the index of a resultentry
 * @param   {ResultEntry} resultEntry
 * @returns {Number}
 */
ResultSet.prototype.indexOf = function(resultEntry) {
	return _.indexOf(this._list, resultEntry);
};

/**
 * Similar as Array.prototype.splice
 * @param {Number} start
 * @param {Number} deleteCount
 * ...
 */
ResultSet.prototype.splice = function(start, deleteCount) {
	var returnValue = this._list.splice.apply(this._list, arguments);
	this._deferredCommit();

	return returnValue;
};

/**
 * When executing actions, we always want to commit these in a deffered way
 * @private
 */
ResultSet.prototype._deferredCommit = function() {
	if (this._pendingCommit === null) {
		this._pendingCommit = setTimeout(this.commitChanges.bind(this), 10);
	}
};

/**
 * When you have done all modifications to the ResultSet
 * You can trigger an instance change event.
 * If you don't commit changes, a deffered commit will occur
 */
ResultSet.prototype.commitChanges = function() {
	if (this._pendingCommit !== null) {
		clearTimeout(this._pendingCommit);
		this._pendingCommit = null;
	}

	this._emit("changed");
};

export default ResultSet;
