import ResultSet from "@esko/xdrop/dataproviders/ResultSet";
import StringSet from "@esko/stringset";
import Observable from "@esko/observable";
import _ from "underscore";
/**
 * An XDropData provider is responsible for providing data to the XDRop
 * @class
 */
var XDropDataProvider = function XDropDataProvider(options) {
	Observable.mixin(this, "dataprovider");

	this._resultSet = new ResultSet();

	this._options = _.extend(
		{
			maxResults: -1 // -1 means no limit of records.
		},
		options
	);

	this._active = false;
	this._initialized = false;

	this._maxResults = this._options.maxResults;

	this._ongoingLoads = new StringSet();
};

/**
 * Get the resultset of the dataprovider
 * @returns {ResultSet}
 */
XDropDataProvider.prototype.getResultSet = function() {
	return this._resultSet;
};

/**
 * Filter the results of the dataprovider
 * !WARNING! This call is not throttled: to improve performance you should throttle the handling of changing the query
 * @param {String} queryString
 */
XDropDataProvider.prototype.setQuery = function(queryString) {
	if (process.env.NODE_ENV === "development") {
		/** PLEASE IMPLEMENT IN SUBCLASS **/
		console.warn("Not implemented");
	}
};

/**
 * Limit the number of results returned by the dataprovider
 *
 * @param {Number} maxResults - if maxResults = -1 --> no limit
 */
XDropDataProvider.prototype.setMaxResults = function(maxResults) {
	this._maxResults = maxResults;
	if (process.env.NODE_ENV === "development") {
		/** PLEASE IMPLEMENT IN SUBCLASS **/
		console.warn("Not implemented");
	}
};

XDropDataProvider.prototype.getMaxResults = function(maxResults) {
	if (this._maxResults < 0) {
		return this._options.maxResults;
	} else if (this._options.maxResults >= 0) {
		return Math.min(this._maxResults, this._options.maxResults);
	} else {
		return -1;
	}
};

//// LOADING
/**
 * Mark the start of a load operation
 * @private
 * @param {String} key - unique identification of the load operation
 */
XDropDataProvider.prototype._startLoading = function(key) {
	var wasLoading = this.isLoading();
	this._ongoingLoads.add(key);
	if (wasLoading != this.isLoading()) {
		this._emit("loading", this.isLoading());
	}
};

/**
 * Is the dataprovider currently loading data?
 * @returns {Boolean}
 */
XDropDataProvider.prototype.isLoading = function() {
	return this._ongoingLoads.size() > 0;
};

/**
 * Mark the end of a load operation
 * @private
 * @param {String} key - unique identification of the load operation (same as used for _startLoading() call)
 */
XDropDataProvider.prototype._stopLoading = function(key) {
	var wasLoading = this.isLoading();
	this._ongoingLoads.remove(key);
	if (wasLoading != this.isLoading()) {
		this._emit("loading", this.isLoading());
	}
};

/**
 * Override to react on the (de)activation
 *
 * @param {Boolean} isActive
 */
XDropDataProvider.prototype.setActive = function(isActive) {
	this._active = isActive;
	if (isActive && !this._initialized) {
		this._initialize();
		this._initialized = true;
	}
};

/**
 * Called on first initialization
 */
XDropDataProvider.prototype._initialize = function() {
	// do nothing
};

export default XDropDataProvider;
