import $ from "jquery";
import XDropDisplay from "./XDropDisplay";
import Observer from "@esko/observer";
import _ from "underscore";
/**
 * A display for XDrop that displays multiple selected values and allows for deselecting them.
 * @param {Object.<Renderer>} renderers
 * @param {XDropPanel} xDropPanel
 * @param {Object} options
 */
var MultiSelectXDropDisplay = function(renderers, xDropPanel, options) {
	Observer.mixin(this);

	this._options = _.extend(
		{
			placeholder: "",
			showInput: true,
			emptyPlaceholderOnSelect: true
		},
		options
	);

	this._representationLayer = $("<div>", {
		class: "XDropResult representation-layer"
	});

	this._input = $("<input>", {
		class: "form-control"
	})
		.attr("placeholder", this._options.placeholder)
		.attr("autocomplete", "off");
	if (this._options.label) {
		this._representationLayer.append(
			$("<div>", {
				class: "inline-label"
			}).append(options.label + ": ")
		);
	}
	this._representationLayer.append(this._input);
	this._representationLayer.append($("<input>").attr("style", "opacity: 0;height: 0;")); // this input will catch the autocomplete in chrome

	if (!this._options.showInput) {
		this._input.hide();
	}
	this._tokenMap = {};

	XDropDisplay.call(this, renderers, xDropPanel, this._options);
	this._hasFocus = false;

	xDropPanel.getValue().forEach(
		function(resultEntry) {
			var token = this._createToken(resultEntry);
			this._input.before(token);
		}.bind(this)
	);

	this._observe(this._panel, "cleared", this._clear);
	this._observe(
		this._panel,
		"loading",
		function(isLoading) {
			this._element.toggleClass("is-loading", isLoading);
		}.bind(this)
	);
};

MultiSelectXDropDisplay.prototype = Object.create(XDropDisplay.prototype);
MultiSelectXDropDisplay.prototype.constructor = MultiSelectXDropDisplay;

/**
 * Clear the input
 */
MultiSelectXDropDisplay.prototype._clear = function() {
	this._input.val("");
	this._panel.setQuery(this._input.val());
	this._input.attr("placeholder", this._options.placeholder);
};

/**
 * @override
 */
MultiSelectXDropDisplay.prototype._init = function() {
	this._element.addClass("multi-select");
	this._element.append(this._representationLayer, this._toggleButton);

	if (this._options.maxHeight) {
		this._representationLayer.css({
			"max-height": this._options.maxHeight,
			overflow: "auto"
		});
	}

	/////
	// Register listeners
	////

	//emit focus
	this._input.on(
		"focus",
		_.bind(function() {
			this._hasFocus = true;
			this._emit("focus");
			if (this._options.isMobile) {
				this._input.trigger("blur");
				this._hasFocus = true;
			}
		}, this)
	);
	this._input.on(
		"blur",
		_.bind(function() {
			this._hasFocus = false;
		}, this)
	);
	this._input.on(
		"input",
		_.bind(function() {
			if (this._hasFocus) {
				//fix for IE
				this._removeSelection();
				if (this._input.val()) {
					//fix for IE11: IE fires input event when setting a placeholder, caused unwanted opening of the xdrop
					this._emit("input");
				}
				this._panel.setQuery(this._input.val());
			}
		}, this)
	);
	this._input.on(
		"click",
		function() {
			if (this._options.openOnClick) {
				this._emit("open");
			}
		}.bind(this)
	);

	// make focus work correctly
	var ignoreElementFocus = false; // we need this extra mechanism because of the IE async focus behavior fix. (this breaks FF so an extra guard is needed)
	this._element.on(
		"focus",
		function() {
			if (!ignoreElementFocus) {
				ignoreElementFocus = true;
				var openOnFocus = this._options.openOnFocus;
				this._options.openOnFocus = false;
				this.focusInput();
				// defer is needed for IE, since it handles focus async
				_.defer(
					function() {
						this._options.openOnFocus = openOnFocus;
						ignoreElementFocus = false;
					}.bind(this)
				);
			}
		}.bind(this)
	);

	//When element selected add token
	this._panel.on(
		"selected",
		_.bind(function(resultEntry) {
			if (this._options.emptyPlaceholderOnSelect) {
				this._input.attr("placeholder", ""); //Empty placeholder
			}
			//create and add token
			var token = this._createToken(resultEntry);
			this._input.before(token);
			//clear input and focus
			this._input.val("");
			this._panel.setQuery(this._input.val());

			var display = this._representationLayer.get(0);
			if (display.scrollHeight > display.clientHeight) {
				// If there is a scrollbar on the display, make sure the input is in view.
				this._input.get(0).scrollIntoView();
			}
			if (!this._options.isMobile) {
				this._input.trigger("focus");
			}
		}, this)
	);

	//When element is deselected => remove token
	this._panel.on(
		"deselected",
		_.bind(function(resultEntry) {
			this._removeToken(resultEntry);
		}, this)
	);

	//Focus on input when whereever clicked in the representation layer
	this._representationLayer.on(
		"click",
		_.bind(function() {
			this.focusInput();
		}, this)
	);

	//Handle special key events
	this._input.on(
		"keydown",
		function(event) {
			this._emit("keydown", event);
			if (!event.originalEvent.defaultPrevented) {
				this._processKeyPress(event.keyCode, event);
			}
		}.bind(this)
	);

	//Deselect tokens when no focus
	this._element.on(
		"focusout",
		_.bind(function() {
			this._removeSelection();
		}, this)
	);
};

/**
 * Creates a token
 * @private
 * @param   {ResultEntry} resultEntry
 * @returns {JQueryObj}
 */
MultiSelectXDropDisplay.prototype._createToken = function(resultEntry) {
	var element;
	var rendererFactory = this._renderFactories[resultEntry.getType()];
	if (rendererFactory) {
		var renderer = rendererFactory.make($("<div>"), resultEntry, this._panel.getRenderAPI());
		element = renderer.getToken();

		this._tokenMap[resultEntry.getUniqueId()] = element;

		element.on(
			"mousedown",
			_.bind(function(event) {
				event.stopPropagation();
				event.preventDefault(); // prevents focus event, which would open the xdrop
				this._removeSelection();
				this._selectToken(element);
			}, this)
		);
	}

	return element;
};

/**
 * Focus the input field on non-mobile devices
 */
MultiSelectXDropDisplay.prototype.focusInput = function() {
	if (!this._options.isMobile) {
		this._input.trigger("focus");
	}
};

/**
 * Removes a token
 * @private
 * @param {JQueryObj} element
 */
MultiSelectXDropDisplay.prototype._removeToken = function(resultEntry) {
	var element = this._tokenMap[resultEntry.getUniqueId()];
	element.remove();
	delete this._tokenMap[resultEntry.getUniqueId()];

	if (this._panel.getValue().length === 0) {
		this._input.attr("placeholder", this._options.placeholder);
	}
};

/**
 * Processes key presses in the input field
 * @private
 * @param {integer} keyCode
 * @param {object}  event
 */
MultiSelectXDropDisplay.prototype._processKeyPress = function(keyCode, event) {
	var selectedToken, resultEntry;

	switch (keyCode) {
		case 8: //BACKSPACE
			if (this._input.val() === "") {
				var toBeDeleted = this._getSelectedTokens();
				if (toBeDeleted.length === 0) {
					toBeDeleted = $(".token", this._representationLayer).last();
				}
				if (toBeDeleted.length !== 0) {
					if (!this._isSelected(toBeDeleted)) {
						this._selectToken(toBeDeleted);
					} else {
						resultEntry = toBeDeleted.data("resultEntry");
						this._panel.getRenderAPI().deselectResultEntry(resultEntry);
					}
				}
			}
			break;
		case 46: //DELETE
			if (this._input.val() === "") {
				selectedToken = this._getSelectedTokens();
				resultEntry = selectedToken.data("resultEntry");
				this._panel.getRenderAPI().deselectResultEntry(resultEntry);
			}
			break;
		case 37: //ARROW LEFT
			if (this._input.val() === "") {
				selectedToken = this._getSelectedTokens();
				var toBeSelected;
				if (selectedToken.length === 0) {
					toBeSelected = $(".token", this._representationLayer).last();
				} else {
					this._deselectToken(selectedToken);
					toBeSelected = selectedToken.prev();
				}
				this._selectToken(toBeSelected);
			}
			break;
		case 39: //ARROW RIGHT
			if (this._input.val() === "") {
				selectedToken = this._getSelectedTokens();
				this._deselectToken(selectedToken);
				this._selectToken(selectedToken.next());
			}
			break;
		default:
			this._panel._processKeyPress(event.keyCode, event);
			break;
	}
};

/**
 * Get the currently selected tokens
 * @private
 * @returns {JQueryObj}
 */
MultiSelectXDropDisplay.prototype._getSelectedTokens = function() {
	return $(".selected", this._representationLayer);
};

/**
 * Deselect all selected tokens
 * @private
 */
MultiSelectXDropDisplay.prototype._removeSelection = function() {
	this._getSelectedTokens().removeClass("selected");
};

/**
 * Selects the given token
 * @private
 * @param {JQueryObj} token
 */
MultiSelectXDropDisplay.prototype._selectToken = function(token) {
	token.addClass("selected");
};

/**
 * Deselects the given token
 * @private
 * @param {JQueryObj} token
 */
MultiSelectXDropDisplay.prototype._deselectToken = function(token) {
	token.removeClass("selected");
};

/**
 * Is the token selected
 * @private
 * @param   {JQueryObj} token
 * @returns {boolean}
 */
MultiSelectXDropDisplay.prototype._isSelected = function(token) {
	return token.hasClass("selected");
};

/**
 * @override
 */
MultiSelectXDropDisplay.prototype._displaySelectedEntry = function() {
	//NOT IMPLEMENTED
};

/**
 * Renders the Xdrop in a loading state (e.g. when retrieving the selected values takes some time)
 * @param {Boolean} isLoading
 * @param {String} placeholder
 */
MultiSelectXDropDisplay.prototype.loading = function(isLoading, placeholder) {
	this._element.toggleClass("is-loading", isLoading);
	if (isLoading) {
		this._input.attr("placeholder", placeholder);
	} else {
		this._input.attr("placeholder", this._options.placeholder);
	}
};

export default MultiSelectXDropDisplay;
