import $ from "jquery";
import _ from "underscore";
import Observer from "@esko/observer";
import Observable from "@esko/observable";
/**
 * This view displays the selected element from the given XDropPanel
 * @param {Object.<Renderer>} renderers
 * @param {XDropPanel} xDropPanel
 * @param {Object} options
 */
var XDropDisplay = function(renderers, xDropPanel, options) {
	Observer.mixin(this);
	Observable.mixin(this, "XDropDisplay");
	this._options = _.extend(
		{
			allowClear: false,
			placeholder: "",
			isMobile: false
		},
		options
	);

	/**
	 * @member {XDropPanel}
	 * @private
	 */
	this._panel = xDropPanel;

	/**
	 * Used to create renderers for resultentries
	 *  we can provide some default renderers
	 */
	this._renderFactories = renderers;

	/**
	 * Main element of the XDrop
	 */
	this._element = $("<div>", {
		class: "XDrop",
		tabindex: 0
	});

	this._value = $("<div>", {
		class: "XDropResult"
	});

	/**
	 * Have a way to clear the value
	 */
	this._clearButton = $("<a>", {
		class: "clear-button"
	}).append(
		$("<i>", {
			class: "icon icon-WCR_Remove12 icon-small"
		})
	);
	this._clearButton.on(
		"mousedown touchstart",
		function(e) {
			this._panel.deselect();
			e.stopPropagation();
			e.preventDefault();
		}.bind(this)
	);

	this._observe(this._panel, "changed", function() {
		this._displaySelectedEntry();
	});

	this._disabled = false;

	this._displaySelectedEntry();
	this._init();
};

/**
 * Creates the element
 * @private
 */
XDropDisplay.prototype._init = function() {
	this._element.append(this._value);
	if (this._options.allowClear) {
		this._element.append(this._clearButton);
	}
};

/**
 * Get the main element of the XDrop
 * @returns {JQueryObj}
 */
XDropDisplay.prototype.getElement = function() {
	return this._element;
};

/**
 * Display the correct value
 * @private
 */
XDropDisplay.prototype._displaySelectedEntry = function() {
	var oldValue;
	var resultEntry = this._panel.getValue()[0];

	if (resultEntry && this._options.allowClear && !this._disabled) {
		this._clearButton.show();
	} else {
		this._clearButton.hide();
	}

	if (resultEntry) {
		var renderFactory = this._renderFactories[resultEntry.getType()];
		if (renderFactory) {
			var renderer = renderFactory.make($("<div>"), resultEntry, this._panel.getRenderAPI());
			oldValue = this._value;
			this._value = renderer.getElement();

			if (this._value != oldValue) {
				oldValue.replaceWith(this._value);
			}

			return;
		}
	}

	oldValue = this._value;

	this._value = $("<div>", {
		class: "XDropResult placeholder",
		text: this._options.placeholder
	});
	oldValue.replaceWith(this._value);
};

/**
 * Update the placeholder of the XDrop and refresh the value to display the new placeholder if needed
 * @param {String} placeholder
 */
XDropDisplay.prototype.setPlaceHolder = function(placeholder) {
	this._options.placeholder = placeholder;
	this._displaySelectedEntry();
};

/**
 * Enable or disable the XDrop
 *
 * @param disabled
 */
XDropDisplay.prototype.toggleEnabled = function(enabled) {
	this._disabled = !enabled;
	if (!enabled) {
		this._emit("disabled");
	}
	this._element.removeClass("readonly").toggleClass("disabled", !enabled);
};

/**
 * Whether or not the dispaly is enabled
 */
XDropDisplay.prototype.isEnabled = function() {
	return !this._disabled;
};

/**
 * Call to destroy the XDrop
 */
XDropDisplay.prototype.destroy = function() {
	this._element.remove();
	this._stopObserveAll();
};
export default XDropDisplay;
