import AnimationFrameRenderer from "@esko/animationframerenderer";
import Observable from "@esko/observable";
import ResultEntry from "@esko/xdrop/dataproviders/ResultEntry";
import LabelRenderer from "@esko/xdrop/renderers/LabelRenderer";
import $ from "jquery";
import _ from "underscore";
import i18n from "@esko/i18n/webi18n-loader?bundle=xdrop/locales!";
/**
 * The XDropPanel acts as a ViewController for the XDrop functionality
 * It contains the dataprovider and displays the ResultSet of the dataprovider using renderers
 *
 * @class
 * @param {XDropDataProvider} dataProvider
 * @param {Object}            renderFactories - map of result type to factory
 */
var XDropPanel = function(dataProvider, renderFactories, options) {
	AnimationFrameRenderer.mixin(this);
	Observable.mixin(this, "XDropPanel");

	this._options = _.extend(
		{
			filterPlaceholder: i18n.t("FilterPlaceholder"),
			noMatchesLabel: i18n.t("NoMatches"),
			noMatchesLoadingLabel: i18n.t("Searching"),
			focusFilterOnActivate: true,
			showNoMatches: true,
			comparator: function(entryA, entryB) {
				return entryA === entryB || (entryA && entryA.equals(entryB));
			},
			active: true,
			asDropDown: true
		},
		options
	);

	// we provide some default renderers
	renderFactories = _.extend(
		{
			label: LabelRenderer
		},
		renderFactories
	);

	this._focusIndex = 0;

	this._selectedEntries = [];

	this._currentSearch = "";
	this._forceFocus = false;

	this._element = $("<div>", {
		class: this._options.asDropDown ? "XDropPanel XDropPanelAsDropDown" : "XDropPanel"
	});

	// The filter field allows the user to filter down results
	// we need to handle keypresses and update dataproviders with new queries if the input is changed
	this._input = $("<input>", {
		class: "form-control XDrop-filter",
		placeholder: this._options.filterPlaceholder
	})
		.on(
			"keydown",
			function(event) {
				this._processKeyPress(event.keyCode, event);
			}.bind(this)
		)
		.on(
			"focus",
			function() {
				this.setActive(true);
			}.bind(this)
		)
		.on(
			"input",
			function() {
				this.setQuery(this._input.val());
			}.bind(this)
		);
	this._element
		.append(
			$("<div>", {
				class: "XDropResult XDrop-filter"
			}).append(
				$("<div>", {
					class: "filter"
				}).append(this._input)
			)
		)
		.on("focusin", function(e) {
			e.stopPropagation(); // JQuery UI modal use focusin event to focus something inside the modal, however the xdrop panel could be added inside the body so stop propagation to ensure the xdrop panel is still focussed.
		});

	/**
	 * All results will be added in this div
	 */
	this._results = $("<div>", {
		class: "results"
	});
	this._element.append(this._results);

	/**
	 * Store the dataprovider and its resultset
	 */
	this._dataProvider = dataProvider;
	this._resultSet = dataProvider.getResultSet();

	// when the result set changes, we need to fix the focused result and render ourselves
	this._resultSet.on(
		"changed",
		function() {
			var wasForceFocus = this._forceFocus;
			// update focus index
			if (this._focusEntry && this._focusEntry.keepSelected()) {
				var newIndex = this._resultSet.indexOf(this._focusEntry);
				if (newIndex >= 0) {
					this.setFocusIndex(this.findFirstFocusableEntry(newIndex));
				} else {
					this.resetFocusIndex();
				}
			} else if (this._focusEntry) {
				this.setFocusIndex(this.findFirstFocusableEntry(this._focusIndex, true));
			} else {
				this.resetFocusIndex();
			}
			this._forceFocus = wasForceFocus;
			this._render();
			this._emit("results.changed", this._resultSet.getSize());
		}.bind(this)
	);

	this._dataProvider.on(
		"loading",
		function(isLoading) {
			this._element.toggleClass("is-loading", isLoading);
			this._emit("loading", isLoading);
		}.bind(this)
	);

	/**
	 * @member {ResultEntry} the currently focused result
	 */
	this._focusEntry = null;
	this.resetFocusIndex();

	this._isVisible = true;
	this._isActive = this._options.active;

	this._initRender(renderFactories);

	this._render();
};

/**
 * Get the element of the XDropPanel
 * @returns {JQueryObj}
 */
XDropPanel.prototype.getElement = function() {
	return this._element;
};

/**
 * Init the render engine for the XDropPanel
 * @private
 * @param {Object}            renderFactories - map of result type to factory
 */
XDropPanel.prototype._initRender = function(renderFactories) {
	this._renderFactories = renderFactories;
	this._renderMap = {};
	this._usedRenderMap = {};
	this._enableRender = true;

	var self = this;

	/**
	 * The render API can be used by the renderers  to communicate with the XDrop
	 */
	this._renderAPI = {
		/**
		 * This method will request the XDrop to be closed (will only work if supported by the implementation)
		 */
		closeXDrop: function() {
			self._emit("close");
		},

		/**
		 * Change the currently selected result for the XDrop
		 * Depending on the multiselect setting (if true) the resultEntry is added to the set of selected
		 * entries or (if false) the entry is set as the selected entry
		 *
		 * @param {ResultEntry} resultEntry
		 */
		selectResultEntry: function(resultEntry) {
			if (self._options.multiSelect) {
				self.select(resultEntry);
			} else {
				self.setValue([resultEntry]);
			}
		},

		/**
		 * Deselects the given resultEntry
		 * @param {ResultEntry} resultEntry
		 */
		deselectResultEntry: function(resultEntry) {
			var selectedEntries = self.getValue();
			var index = selectedEntries.indexOf(resultEntry);
			selectedEntries.splice(index, 1);
			self.setValue(selectedEntries);
		},

		/**
		 * Set the focus to a specific result entry
		 * Only trigger this based on a user action
		 * @param {ResultEntry} resultEntry
		 */
		focusResultEntry: function(resultEntry) {
			var focusIndex = self._resultSet.indexOf(resultEntry);
			if (focusIndex >= 0) {
				self.setFocusIndex(focusIndex);
			}
		},

		/**
		 * Returns the value of the XDropPanel
		 * @returns {Array.<ResultEntry>}
		 */
		getValue: function() {
			return self.getValue();
		},

		/**
		 * Set the value of the XDropPanel
		 */
		setValue: function(value) {
			self.setValue(value);
		},

		/**
		 * returns if the XDropPanel used in MultiSelectXDrop
		 * @returns {boolean}
		 */
		isMultiSelect: function() {
			return self._options.multiSelect;
		},

		/**
		 * Listen to scroll events of the xdroppanel
		 * @param {Function} callback - called with two arguments: first visible resultentry, last visible resultentry
		 */
		onScroll: function(callback) {
			if (!self._listeningToScroll) {
				self._listenomgToScroll = true;
				self._results.on(
					"scroll",
					_.throttle(
						function() {
							var clientRect = self._results[0].getBoundingClientRect();
							var el = document.elementFromPoint(clientRect.left + 1, clientRect.top + 1);
							var top = null;
							if (el && self._renderMap[el.getAttribute("data-resultentry")]) {
								top = self._renderMap[el.getAttribute("data-resultentry")].getResultEntry();
							}

							el = document.elementFromPoint(clientRect.left + 1, clientRect.bottom - 1);
							var bottom = null;

							if (el && self._renderMap[el.getAttribute("data-resultentry")]) {
								bottom = self._renderMap[el.getAttribute("data-resultentry")].getResultEntry();
							}

							self._emit("scroll", top, bottom);
						},
						{
							leading: false
						}
					)
				);
			}

			self.on("scroll", callback);
		},

		/**
		 * Temporarily pause rendering of the XDrop, this could be used when doing drag and drop from a xdroppanel
		 * Don't use if yo udon't know what you're doing.
		 * @param {Boolean} isPaused
		 */
		pauseRender: function(isPaused) {
			self._enableRender = !isPaused;
			self._render();
		}
	};
};

/**
 * Set the filter query of the XDropPanel
 * @param {String} queryString
 */
XDropPanel.prototype.setQuery = function(queryString) {
	if (queryString != this._currentSearch) {
		if (this._input.val() !== queryString) {
			this._input.val(queryString);
		}

		this._currentSearch = queryString;

		// reset focus
		this.resetFocusIndex();

		// update the dataprovider
		this._dataProvider.setQuery(queryString);
		this._render();
	}
};

/**
 * Handle keypresses in the filter field
 * @private
 * @param {Number} keyCode
 * @param {Event}  event   - browser event
 */
XDropPanel.prototype._processKeyPress = function(keyCode, event) {
	if (this._isActive) {
		switch (keyCode) {
			case 13: // ENTER --> activate the entry
				var entry = this.getFocusedEntry();
				if (entry && this._renderMap[entry.getUniqueId()]) {
					this._renderMap[entry.getUniqueId()].activate();
				}
				event.stopPropagation();
				event.preventDefault();
				break;
			case 38: // ARROW UP --> change focus
				this.setFocusIndex(this.findFirstFocusableEntry(this._focusIndex - 1, true, true));
				break;
			case 40: // ARROW DOWN --> change focus
				this.setFocusIndex(this.findFirstFocusableEntry(this._focusIndex + 1, true));
				break;
			case 33: // PAGE UP	--> change focus
				this.setFocusIndex(this.findFirstFocusableEntry(this._focusIndex - 10, true, true));
				event.stopPropagation();
				event.preventDefault();
				break;
			case 34: // PAGE DOWN --> change focus
				this.setFocusIndex(this.findFirstFocusableEntry(this._focusIndex + 10, true));
				event.stopPropagation();
				event.preventDefault();
				break;
			case 9: // TAB --> deactivate
				this.setActive(false);
				event.stopPropagation();
				event.preventDefault();
				this._emit("closeAndFocusNext");
				break;
			case 27: // ESCAPE --> deactivate
				this.setActive(false);
				this._emit("close");
				break;
		}
	}
};

/**
 * Find the first entry that can be focused and selected
 * @param   {Number}  [startPos=0]           - start position
 * @param 	{Boolean} forNonDefaultSelection - if true, also elements which shouldn't be selected as default are matched
 * @returns {Number}  focus index
 */
XDropPanel.prototype.findFirstFocusableEntry = function(startPos, forNonDefaultSelection, upwards) {
	if (!startPos || isNaN(startPos + 0)) {
		startPos = 0;
	}
	var resultList = this._resultSet.getEntriesImmutable();
	var i, pos, resultEntry;
	var bestDefaultSelectionIndex;
	var bestPosition = startPos;

	var step = 1;
	if (upwards) {
		step = -1;
	}

	for (i = 0; i < resultList.length; i++) {
		pos = this._normalizeFocusIndex(startPos + step * i);
		resultEntry = resultList[pos];
		if (resultEntry && resultEntry.allowSelection()) {
			if (forNonDefaultSelection) {
				return pos;
			} else {
				if (resultEntry.getDefaultSelectionIndex() > bestDefaultSelectionIndex || _.isUndefined(bestDefaultSelectionIndex)) {
					bestDefaultSelectionIndex = resultEntry.getDefaultSelectionIndex();
					bestPosition = pos;
				}
			}
		}
	}

	return bestPosition;
};

/**
 * Normalize a focus Index (make sure it is in the range [0, resultSetSize.size()]
 * @private
 * @param   {Number} index
 * @returns {Number}
 */
XDropPanel.prototype._normalizeFocusIndex = function(index) {
	var resultSetSize = this._resultSet.getSize();
	if (resultSetSize === 0) {
		return 0;
	} else {
		return ((index % resultSetSize) + resultSetSize) % resultSetSize;
	}
};

/**
 * Reset the currently focused element
 */
XDropPanel.prototype.resetFocusIndex = function() {
	this.setFocusIndex(this.findFirstFocusableEntry());
};

/**
 * Change the currently focused element
 * @param {Number} index
 */
XDropPanel.prototype.setFocusIndex = function(index) {
	this._forceFocus = true;
	this._focusIndex = this._normalizeFocusIndex(index);
	this._focusEntry = this._resultSet.getEntry(this._focusIndex);
	this._render();
};

/**
 * Get the focused entry (based on the focusindex)
 * @returns {ResultEntry}
 */
XDropPanel.prototype.getFocusedEntry = function() {
	return this._resultSet.getEntry(this._focusIndex);
};

/**
 * Get the number of entries in the panel.
 * @returns {ResultEntry}
 */
XDropPanel.prototype.getNumEntries = function() {
	return this._resultSet.getSize();
};

/**
 * Get the render API of the XDropPanel
 * @returns {Object}
 */
XDropPanel.prototype.getRenderAPI = function() {
	return this._renderAPI;
};

/**
 * Update the rendering of the XDropPanel
 * @private
 */
XDropPanel.prototype._doRender = function() {
	if (!this._isVisible || !this._enableRender) {
		return; // no render when invisible or rendering is blocked
	}

	var domEl = this._results.get(0);
	var entries = this._resultSet.getEntriesImmutable();
	var panelClientRect, focusedClientRect;
	var scrollTop;
	var focusedRenderer = null;

	if (!this._forceFocus && this._focusedElement) {
		var focusedClientRect = this._focusedElement.getBoundingClientRect();
		var panelClientRect = domEl.getBoundingClientRect();

		// if focused element currently visible it should be focused again
		this._forceFocus =
			(focusedClientRect.top > panelClientRect.top && focusedClientRect.top < panelClientRect.bottom) ||
			(focusedClientRect.bottom < panelClientRect.bottom && focusedClientRect.bottom > panelClientRect.top);
	}

	if (entries.length === 0 && this._options.showNoMatches) {
		// show no results indication
		var label = this._options.noMatchesLabel;
		if (this._dataProvider.isLoading()) {
			label = this._options.noMatchesLoadingLabel;
		}
		var nomatches = new ResultEntry("label", label);
		nomatches.setAllowSelection(false);
		entries = [nomatches];
	}
	for (var i = 0; i < entries.length; i++) {
		var renderer = this._getRenderer(entries[i]);
		var renderEl = renderer.getElement()[0];
		if (domEl.children[i] != renderEl) {
			domEl.insertBefore(renderEl, domEl.children[i] || null); // IE10 can't handle undefined --> || null
		}

		// update the state of the renderers
		var isFocused = i === this._focusIndex;

		renderer.setFocused(isFocused);
		if (isFocused) {
			focusedRenderer = renderer;
		}
		renderer.setSelected(_.findIndex(this._selectedEntries, entry => this._options.comparator(entries[i], entry)) !== -1);

		renderer.render();
		renderer.setQuery(this._currentSearch);
	}

	this._cleanRenderers();

	if (this._forceFocus && focusedRenderer) {
		// scroll the highlighted result in to the view
		this._focusedElement = focusedRenderer.getElement()[0];
		var focusedClientRect = this._focusedElement.getBoundingClientRect();
		var panelClientRect = domEl.getBoundingClientRect();

		var diff;
		// +1 -1 is needed for borders
		if (focusedClientRect.top + 1 < panelClientRect.top) {
			diff = panelClientRect.top - focusedClientRect.top;
			this._results.scrollTop(this._results.scrollTop() - diff);
		} else if (focusedClientRect.bottom - 1 > panelClientRect.bottom) {
			diff = panelClientRect.bottom - focusedClientRect.bottom;
			this._results.scrollTop(this._results.scrollTop() - diff);
		}
		this._forceFocus = false;
	}

	this._emit("render");
};

/**
 * Set the value of the XDrop
 * Currently only single value selection is supported
 * @param {Array.<ResultEntry>} resultEntries - the result to be selected
 */
XDropPanel.prototype.setValue = function(resultEntries) {
	if (resultEntries) {
		if (!(resultEntries instanceof Array)) {
			throw "SetValue requires an array";
		}
		var oldSelectedEntries = this._selectedEntries.slice();

		var changed = false;

		//deselect the entries that are not in the new array
		_.each(
			this._selectedEntries,
			function(selectedEntry) {
				if (_.findIndex(resultEntries, entry => this._options.comparator(selectedEntry, entry)) === -1) {
					// Use comparator to find renderers which match the deselected entry
					_.each(this._renderMap, renderer => {
						if (this._options.comparator(selectedEntry, renderer.getResultEntry())) {
							renderer.setSelected(false);
						}
					});
					this._emit("deselected", selectedEntry);

					changed = true; // we removed
				}
			},
			this
		);

		this._selectedEntries = resultEntries;

		// select the entries that are new
		_.each(
			this._selectedEntries,
			function(resultEntry) {
				if (_.findIndex(oldSelectedEntries, entry => this._options.comparator(resultEntry, entry)) === -1) {
					// Use comparator to find renderers which match the selected entry
					_.each(this._renderMap, renderer => {
						if (this._options.comparator(resultEntry, renderer.getResultEntry())) {
							renderer.setSelected(true);
						}
					});

					this._emit("willSelectAndFocus", resultEntry);

					this._renderAPI.focusResultEntry(resultEntry);

					this._emit("selected", resultEntry);

					changed = true; // we added
				}
			},
			this
		);

		if (changed) {
			this._emit("changed");
		}
		if (resultEntries.length === 0) {
			this._emit("cleared");
		}
	}
};

/**
 * Is an entry already selected?
 * Uses the comparator function
 * @param   {ResultEntry} resultEntry
 * @returns {Boolean}
 */
XDropPanel.prototype.isSelected = function(resultEntry) {
	return !!_.find(
		this._selectedEntries,
		function(selected) {
			return this._options.comparator(resultEntry, selected);
		},
		this
	);
};

/**
 * Adds an entry to the selected entries
 * @private
 * @param {ResultEntry} resultEntry
 */
XDropPanel.prototype.select = function(resultEntry) {
	if (!this.isSelected(resultEntry)) {
		var selectedEntries = this.getValue();
		selectedEntries.push(resultEntry);
		this.setValue(selectedEntries);
	}
};

/**
 * Deselects the selected entries.
 */
XDropPanel.prototype.deselect = function() {
	this.setValue([]);
};

/**
 * Get the value of the XDrop
 * @returns {ResultEntry}
 */
XDropPanel.prototype.getValue = function() {
	return this._selectedEntries.slice(); //return a copy
};

/**
 * Get the renderer for a resultentry
 * @private
 * @param   {ResultEntry} resultEntry
 * @returns {Renderer}
 */
XDropPanel.prototype._getRenderer = function(resultEntry) {
	var renderer = this._renderMap[resultEntry.getUniqueId()];

	if (renderer) {
		delete this._renderMap[resultEntry.getUniqueId()];
	} else {
		var factory = this._renderFactories[resultEntry.getType()];
		if (!factory) {
			throw new Error("Trying to render item of unknown type");
			// future make fallback  renderer
		}

		renderer = factory.make(
			$("<div>", {
				"data-resultentry": resultEntry.getUniqueId()
			}),
			resultEntry,
			this._renderAPI
		);

		if (_.findIndex(this._selectedEntries, entry => this._options.comparator(resultEntry, entry)) !== -1) {
			renderer.setSelected(true);
		}
	}

	this._usedRenderMap[resultEntry.getUniqueId()] = renderer;

	return renderer;
};

/**
 * Clean out the no longer used renderers
 * @private
 */
XDropPanel.prototype._cleanRenderers = function() {
	_.each(this._renderMap, function(renderer) {
		renderer.destroy();
	});

	this._renderMap = this._usedRenderMap;
	this._usedRenderMap = {};
};

/**
 * Change the visibility of the XDropPanel
 * @param {Boolean} isVisible
 */
XDropPanel.prototype.setVisible = function(isVisible) {
	if (this._isVisible != isVisible) {
		this._isVisible = isVisible;
		this._element.toggleClass("hidden", !isVisible);

		this._render();
	}
};

/**
 * Activate or deactivate the component
 * @param {Boolean} isActive
 */
XDropPanel.prototype.setActive = function(isActive) {
	if (this._isActive != isActive) {
		this._isActive = isActive;
		this._dataProvider.setActive(isActive);

		if (this._options.focusFilterOnActivate) {
			this._input.trigger("focus").trigger("select");
		}

		if (isActive) {
			if (this._selectedEntries.length > 0) {
				var selectedIndex = this._resultSet.indexOf(_.last(this._selectedEntries));
				if (selectedIndex >= 0) {
					this.setFocusIndex(selectedIndex);
					this._render();
				}
			} else {
				this.setFocusIndex(0);
			}
		} else {
			this._emit("close");
		}
	} else if (isActive) {
		if (this._options.focusFilterOnActivate) {
			_.defer(
				function() {
					if (!this._input.is(":focus")) {
						this._input.trigger("focus").trigger("select");
					}
				}.bind(this)
			);
		}
	}
};

/**
 * Whether to show or hide the filter input
 *
 * @param {Boolean} show - true when we want to show the filter input
 */
XDropPanel.prototype.setShowFilterField = function(show) {
	this._element.toggleClass("nofilter", !show);
};

export default XDropPanel;
