import $ from "jquery";
import _ from "underscore";
import "@esko/jquery-highlight-text";
/**
 * Default renderer implementation
 * @class
 * @param {JQueryObj}   div
 * @param {ResultEntry} resultEntry
 * @param {Object}      api         -  render API
 * @param {Object}      options {
 *                                   highlightingPerWord: (does the highlighting for each word in the query separately)
 *                              }
 */
var Renderer = function(div, resultEntry, api, options) {
	this._element = div;
	this._tokenContent = this._element;

	this._element.addClass("XDropResult");
	this._resultEntry = resultEntry;

	this._hasFocus = false;
	this._isSelected = false;

	this._api = api;

	this._options = _.extend(
		{
			highlightingPerWord: false
		},
		options
	);

	this._element.on("click", this.click.bind(this));

	this._elementsToHighlightMatches = [this._element];
};

/**
 * Get the result entry of the renderer
 * @returns {ResultEntry}
 */
Renderer.prototype.getResultEntry = function() {
	return this._resultEntry;
};

/**
 * Get the element of the renderer
 * @returns {JQueryObj}
 */
Renderer.prototype.getElement = function() {
	return this._element;
};

/**
 * Get the token element
 * @returns {JQueryObj}
 */
Renderer.prototype.getToken = function() {
	var closeBtn = $("<a>", {
		class: "btn token-close"
	}).append(
		$("<i>", {
			class: "icon icon-WCR_Close12"
		})
	);
	closeBtn.on(
		"click",
		_.bind(function(e) {
			this._deselectResultEntry(this._resultEntry);
			e.stopPropagation();
		}, this)
	);

	this._token = $("<div>", {
		class: "token"
	}).append(closeBtn);

	this._token.data("resultEntry", this._resultEntry);

	return this._token;
};

/**
 * Mark the element as focused
 * @param {Boolean} hasFocus
 */
Renderer.prototype.setFocused = function(hasFocus) {
	if (this._hasFocus != hasFocus) {
		this._element.toggleClass("has-focus", hasFocus);
		this._hasFocus = hasFocus;
	}
};

/**
 * Mark the element as selected
 * @param {Boolean} isSelected
 */
Renderer.prototype.setSelected = function(isSelected) {
	if (this._isSelected != isSelected) {
		this._element.toggleClass("is-selected", isSelected);
		this._isSelected = isSelected;
	}
};

/**
 * Change which elements should highlight matches
 * Can be used if you don't like to overwrite setQuery
 * @private
 * @param {Array.<JQueryObj>} elements
 */
Renderer.prototype.setElementsToHighlightMatches = function(elements, dontClean) {
	if (!dontClean) {
		_.each(this._elementsToHighlightMatches, function(element) {
			element.removeHighlightText("xdrop-hightlightMatch");
		});
	}

	this._elementsToHighlightMatches = elements;
};

/**
 * Higlight the query text in the element
 * @param {String} query
 */
Renderer.prototype.setQuery = function(query) {
	if (this._previousQuery !== query && this._resultEntry.allowSelection()) {
		_.each(
			this._elementsToHighlightMatches,
			function(element) {
				element.removeHighlightText("xdrop-hightlightMatch");
				if (!this._options.highlightingPerWord) {
					element.highlightText(query, "xdrop-hightlightMatch");
				} else {
					_.each(query.split(" "), function(queryWord) {
						element.highlightText(queryWord, "xdrop-hightlightMatch");
					});
				}
			}.bind(this)
		);
		this._previousQuery = query;
	}
};

/**
 * Called whenever ResultSet was updated
 */
Renderer.prototype.render = function() {
	// implement this method to update view when resultset was modified
};

/**
 * A click event occured on the renderer, by default we will activate
 */
Renderer.prototype.click = function(event) {
	this.activate();
	event.preventDefault();
};

/**
 * Default activate behavior is to select the current result entry and close the dropdown
 */
Renderer.prototype.activate = function() {
	if (this._resultEntry.allowSelection()) {
		this._focusResultEntry(this._resultEntry);
		this._selectResultEntry(this._resultEntry);
		this._closeXDrop();
	}
};

/**
 * Select a ResultEntry
 * (wrapper around the XDrop Renderer API)
 * @private
 * @param {ResultEntry} resultEntry - the result entry that should be selected
 */
Renderer.prototype._selectResultEntry = function(resultEntry) {
	this._api.selectResultEntry(resultEntry);
};

/**
 * Deselect a ResultEntry
 * (wrapper around the XDrop Renderer API)
 * @private
 * @param {ResultEntry} resultEntry - the result entry that should be selected
 */
Renderer.prototype._deselectResultEntry = function(resultEntry) {
	this._api.deselectResultEntry(resultEntry);
};

/**
 * Focus a ResultEntry
 * @private
 * @param {ResultEntry} resultEntry
 */
Renderer.prototype._focusResultEntry = function(resultEntry) {
	this._api.focusResultEntry(resultEntry);
};

/**
 * Notify the XDrop that is can be closed
 * (wrapper around the XDrop Renderer API)
 * @private
 */
Renderer.prototype._closeXDrop = function() {
	this._api.closeXDrop();
};

/**
 * Trigger a "data" event on the ResultEntry
 * @private
 */
Renderer.prototype._triggerEvent = function() {
	return this._resultEntry.trigger.apply(this._resultEntry, arguments);
};

/**
 * Destroy the element
 */
Renderer.prototype.destroy = function() {
	this._element.remove();
};

export default Renderer;
